// Option.cpp: implementation of the Option class.
//
//////////////////////////////////////////////////////////////////////

#ifdef _WIN32
#include <windows.h>
#pragma warning(disable:4786) // supress the annoying windows warning
#endif

#include "Option.h"

//////////////////////////////////////////////////////////////////////
// Class Option
//////////////////////////////////////////////////////////////////////
Option::~Option()
{
    if (! optionList.empty()){
        OptionList::const_iterator iter;
        for ( iter = optionList.begin(); iter != optionList.end(); ++iter ) {
            delete(iter->second);
        } // for
    } // if
};
bool Option::addOption(const std::string & keyword, 
                       const std::string& value, 
                       const InputType & type, 
                       const std::string & comment)
{
    Argument *argument = new Argument( value, type, comment);
    optionList.insert( OptionList::value_type(keyword, argument ));
    return true;
};

bool Option::setValue(const std::string& index, const std::string& value)
{
    OptionList::iterator  where= optionList.find(index);
    if (where != optionList.end() ){ 
        where->second->setValue(value);
        return true;
    }
    return false;
}

bool Option::parseOptions(int& argc, char**& argv)
{
    OptionList::iterator where;
    
    // No non-default input, we are done.
    if(argc == 0) return false;
    
    // if we have option other than default
    int i = 0;
    int j = 0;
    while(i < argc){
        std::string arg(argv[i]);
        //looking for marker 
#if defined(_WIN32) //wired win32 and unix difference
        if(arg.compare(0, m_marker.length(),m_marker) !=0){
#else
        if(arg.compare(m_marker, 0, m_marker.length()) !=0){
#endif
            i++;
            j++;
            continue;
        }else{
            //looking for valid option
            where = optionList.find(arg.substr(m_marker.length(), arg.length()));
            // we didn't find the option
            if(where == optionList.end())
                return false;       
            //set value in this argument
            Argument *argument = where->second;
            i++;
            j = 0;
            // if this is a FLAG, set it to true
            if(argument->getType() == FLAG_TYPE) 
                argument->setValue(true_string);	
            else if(i < argc){
                arg = std::string(argv[i]);
#if defined(_WIN32)  //wired win32 and unix difference
                if(arg.compare(0, m_marker.length(),m_marker) !=0){
#else
                if(arg.compare(m_marker, 0, m_marker.length()) !=0){
#endif
                    // there  is a value for this option, set it
                    argument->setValue(arg);
                    i++;
                } 
            }
        } 
    } // while
    argc -= i - j;
    argv += i - j;
    return true;
}
const bool Option::getArgument(const std::string& index, Argument& arg)
{ 
    OptionList::iterator  where= optionList.find(index);
    if ( where != optionList.end() ){ 
        arg = *(where->second);
        return true;
    }else 
        return false;
}
void Option::printHelp()
{ 
    std::cout << m_header.data()<< std::endl;
    if (! optionList.empty()){
        OptionList::const_iterator iter;
        for ( iter = optionList.begin(); iter != optionList.end(); ++iter ) {
            std::cout << m_marker.data() << iter->first << '\t' << (*(iter->second))<< std::endl;
        } // for
    } // if
} // printHelp

//////////////////////////////////////////////////////////////////////
// Class Option
//////////////////////////////////////////////////////////////////////
Argument::~Argument() {;}


/////////////////////////////////////////////////////////////////////
// Friend function
////////////////////////////////////////////////////////////////////
std::ostream& operator<<(std::ostream& s, const Argument& a)
{
    std::string typeStr;
    switch (a.m_type) {
    case STRING_TYPE:
        typeStr.append("[string]");
        break;
    case INT32_TYPE:
        typeStr.append("[long int]");
        break;
    case FLAG_TYPE:
        typeStr.append("[flag]");
        break;
    case FLOAT32_TYPE:
        typeStr.append("[float]");
    default:
        typeStr.append("[na]");
    }
    if(a.m_type != FLAG_TYPE)
	s<< a.m_comment << std::endl << "\tDefault [" << a.m_dfvalue <<"], Type " 
        << typeStr;
    else
	s<< a.m_comment;
    return s;
}
