/* Option.h: interface for the Option class. */
/* */
/*//////////////////////////////////////////////////////////////////// */

#if !defined(AFX_OPTION_H__6D394BA7_4670_4036_8B6F_DA7C2A589E9F__INCLUDED_)
#define AFX_OPTION_H__6D394BA7_4670_4036_8B6F_DA7C2A589E9F__INCLUDED_

#include <map>
#include <string>
#include <iostream>
#include <strstream>
#include <sstream>

/*using namespace std; */

enum InputType{
    NO_TYPE  = -1,
    UINT8_TYPE   = 0,
    INT16_TYPE   = 1,
    INT32_TYPE   = 2,
    UINT32_TYPE  = 3,
    FLOAT32_TYPE = 4,
    FLOAT64_TYPE = 5,
    FLAG_TYPE    = 6,
    STRING_TYPE  = 7
};

const std::string true_string("1");
const std::string false_string("0");

class Argument{
public:
    /* Construtors and Destructors */
    Argument() : m_value(), m_comment(), m_dfvalue(), m_type(NO_TYPE){}   
    Argument(const std::string& value, const InputType & type, const std::string & comment) : 
    m_value(value), m_comment(comment), m_dfvalue(value),m_type(type){}
    virtual ~Argument(); 
    
    /*Get values */
    const std::string getComment()	{ return m_comment;};
    const int getType()	{ return m_type;};
    template<class T> bool getValue(T& value) {
    /* use istream_iterator as an adapter to convert m_vallue from string to T */
    std::stringstream ss;
    ss.rdbuf()->str(m_value);
    std::istream is(ss.rdbuf());

    /* create an iterator adapter */
    std::istream_iterator<T> it(is);
    std::istream_iterator<T> itEOF;

    if (it != itEOF){
        value = *it;
        return true;
    }

    return false;
    }
    
    /*Set values */
    bool setValue(const std::string& value){
        return(setInternalValue(value, m_value));
    }
    bool setDfValue(const std::string& value){
        return(setInternalValue(value, m_dfvalue));
    }
    
    /*Opterator overloading */
    friend std::ostream& operator<<(std::ostream& s, const Argument& a);
    
protected:
    
    std::string		m_value;
    std::string		m_comment;
    std::string		m_dfvalue;
    InputType		m_type;
    
    bool setInternalValue(const std::string& value, std::string& storedValue){
        storedValue = value;		
        return true;
    };
};

typedef std::map<std::string, Argument *> OptionList;

class Option  
{
public:

    Option(const  std::string & header, const std::string & marker) : 
       m_header(header), m_marker(marker){};
    Option() : m_header(), m_marker("-"){};
    virtual ~Option();

    bool addOption(const std::string & keyword, const std::string& value, const InputType & type, const std::string & comment);
    bool parseOptions(int& argc, char**& argv);

    bool setValue(const std::string& index, const std::string& value);
    template<class T> bool getValue(const std::string& index, T& value) {
        Argument arg;
        if (getArgument(index, arg)){
            if(arg.getValue(value)) return true;
        }
        return false;
    }
    const bool getArgument(const std::string& index, Argument& arg);

    void printHelp();

protected:
    OptionList optionList;
    std::string m_header;
    std::string m_marker;
    
};
#endif /* !defined(AFX_OPTION_H__6D394BA7_4670_4036_8B6F_DA7C2A589E9F__INCLUDED_) */


