/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Author:  ?, ?
**
**  Filename:  DgpDecode_main.c
**
**  Front end for DgpDecode.
**
**
**	Modification history:
**	    Dillon H. Tracy esq. / March 2000
**		o  Auto-implemented dgp types.
**		o  Search on 'dhtesq' for this coder's comments & changes.
**      August 2000
**      o  As prep for FX release, implemented a subset (mono) of cmd line options
**      with libcmdln.
**  
**************************************************************************/

#if 0
/* Now that xiff is gone, I think this is no longer necessary... */
#ifdef _WIN32
/* libipw defines cPlanes which is used in winsock2
 * so if a user include readXiff3.h and then winsock2.h they
 * get a really nasty error.  So we'll just include it first 
 * so this doesn't happen.
 */
#include <winsock2.h>
#endif
#endif


/* */
#include <math.h>
#include <time.h>
#include "dgp_err_handle.h"
#include "dgp_cl.h"         /* Standardized command-line interface definitions. */
#include "image.h"
#include "silx_decode.h"
#include "image_defs.h"
#include "PsgFix.h"

static Dgp_Float64 
get_utime()
{
	return (Dgp_Float64)clock()/CLOCKS_PER_SEC;
}

/*
 * Default values for command line arguments.
 */
#define DEF_OUTFILENAME "input file name + outtype"
#define DEF_VERBOSE     DGP_FALSE
#define DEF_OUTPUT_RES  0
#define DEF_OUTPUT_RES_STR  "same as input"
#define DEF_FORCE_COLOR DGP_FALSE
#define DEF_FORCE_MONO  DGP_FALSE
#define DEF_JBIG2_ONLY	DGP_FALSE
#define DEF_IMAGES_ONLY DGP_FALSE
#define DEF_OUTTYPE     "tif"
#define DEF_ORIENT      "RD"
#define DEF_ROTATION	0
#define DEF_ROTATION_STR "0"
#define DEF_START_PAGE  0
#define DEF_START_PAGE_STR  "0"
#define DEF_STOP_PAGE   0
#define DEF_STOP_PAGE_STR "end of document"
#define DEF_STRIPSIZE   0
#define DEF_STRIPSIZE_STR "page height"
#define DEF_NOOUT       DGP_FALSE
/*#define DEF_CMYK        DGP_FALSE */
#define OUTNAMEENTRY 1


typedef struct {
  Dgp_Int8 *out_filename;
  Dgp_Bool verbose;
  Dgp_Int32 output_res;
  Dgp_Bool force_color;
  Dgp_Bool force_mono;
  Dgp_Int8 *outtype;
  Dgp_Int8 *orientation;
  Dgp_Int32 rotation;
  Dgp_Int32 start_page;
  Dgp_Int32 stop_page;
  Dgp_Int32 stripsize;
  Dgp_Bool no_output;
  Dgp_Bool jbig2Only;
  Dgp_Bool imagesOnly;
/*  Dgp_Bool do_cmyk; */
} CLOpts;

static CLOpts cl_opts = {   DEF_OUTFILENAME, 
                            DEF_VERBOSE, 
                            DEF_OUTPUT_RES, 
                            DEF_FORCE_COLOR, 
                            DEF_FORCE_MONO,
                            DEF_OUTTYPE,
                            DEF_ORIENT,
							DEF_ROTATION,
                            DEF_START_PAGE,
                            DEF_STOP_PAGE,
                            DEF_STRIPSIZE,
                            DEF_NOOUT,
							DEF_JBIG2_ONLY,
							DEF_IMAGES_ONLY
                            /*DEF_CMYK*/ };

/* Displayed when showing usage information. */
static const Dgp_Int8 *cl_header = "Usage: DgpDecode [options] foo.tfx [bar.tfx...]\n \
    \tDecompresses tiff-fx documents to the specified format.";

/*
 * The definition of the command line options.
 */
static DgpCl_Option cl_cmdln_options[] = {
  DGPCL_HELP_CMD,
  DGPCL_OUTFNAME_CMD(cl_opts.out_filename, DEF_OUTFILENAME),
  DGPCL_VERBOSE_CMD(cl_opts.verbose),
  DGPCL_RES_CMD(cl_opts.output_res, DEF_OUTPUT_RES_STR),
  DGPCL_COLOR_CMD(cl_opts.force_color),
  DGPCL_MONO_CMD(cl_opts.force_mono),
  DGPCL_NEW_STRING_CMD("outtype", cl_opts.outtype, 
    "Output format (PNM or TIF). Extension appended to fname", DEF_OUTTYPE),
  DGPCL_NEW_INT_CMD("rot", cl_opts.rotation, "Rotate output clockwise by this multiple of pi/2", DEF_ROTATION_STR),
  DGPCL_NEW_INT_CMD("start", cl_opts.start_page, "Starting page for decompression, counting from zero", 
    DEF_START_PAGE_STR),
  DGPCL_NEW_INT_CMD("stop", cl_opts.stop_page, "Stopping page for decompression, counting from zero", 
    DEF_STOP_PAGE_STR),
  DGPCL_NEW_INT_CMD("stripsize", cl_opts.stripsize, "Generate each page in bands of this height", 
    DEF_STRIPSIZE_STR),
  DGPCL_NEW_FLAG_CMD("noout", cl_opts.no_output, "Decompress to memory;  no file output"),
  DGPCL_NEW_FLAG_CMD("jbig2", cl_opts.jbig2Only, "Just output jbig2 plane (text only)"),
  DGPCL_NEW_FLAG_CMD("images", cl_opts.imagesOnly, "Just output non-jbig2 planes (images only)"), 
  /*DGPCL_NEW_FLAG_CMD("cmyk", cl_opts.do_cmyk, "Decompress to cmyk separations"), */
  {NULL}
};


/*************************************************************************/
/* Argument validator... */
/* Resolve as many exclusive options as we can here.  Last option wins. */
static ErrCode
validateArgs(Dgp_State *dgp_state,
             Dgp_Int32 unparsed_argc, 
             Dgp_Int8 **unparsed_argv,
             FILE *err_out,
             void *validate_cookie)
{
  DGP_PROLOGUE("validateArgs()", DGP_PERRSTATE);
  CLOpts *my_opts = (CLOpts *)validate_cookie;

  unparsed_argc = unparsed_argc;
  unparsed_argv = unparsed_argv;

  if(my_opts->output_res < 0) {
    DGPCL_PRINT_ERRF0(err_out, "Positive resolutions only...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->start_page < 0) {
    DGPCL_PRINT_ERRF0(err_out, "Positive pages only...");
    DGP_ERR(cErrInvalidCmdLine);
  }
  if(my_opts->stop_page < my_opts->start_page) {
    my_opts->stop_page = my_opts->start_page;
  }
  if(my_opts->force_color){
      my_opts->force_mono = DGP_FALSE;
  }
  if(my_opts->force_mono){
      my_opts->force_color = DGP_FALSE;
  }
  /* Let -jbig2 and -images coexist... */
  /*if(my_opts->jbig2Only)... */
  /*
  /* cmyk support limitations... */
  if(my_opts->do_cmyk){
      my_opts->outtype = "PNM";
  }
  */  
  NORMAL_RETURN;
}
/*************************************************************************/


int
main(int argc, Dgp_Int8 **argv){
  
  Dgp_Int32             docStart;
  SilxDecCookie          sOpt;
  Dgp_State             *dgp_state;
  FILE                  *fp=NULL;
  void                  *image = NULL;
  Dgp_Float64			t0, t1;


  DGP_MAIN_PROLOGUE("main()", stderr, Dgp_fileErrPrinter);
  if(argc == 1){
      DgpCl_printHelp(stderr, cl_header, cl_cmdln_options);
      MAIN_NORMAL_RETURN;
  }

  /* Decode timer... */
  t0 = get_utime();

  DGP_ERR_PASS(Dgp_initDgpState(&dgp_state, 
                                NULL,
                                NULL,
                                stderr, 
                                Dgp_fileErrPrinter,
                                DGP_TRUE, 
                                DGP_FALSE));
  DGP_ERR_CHECK_NO_MSG(DgpCl_parse(dgp_state, 
      (Dgp_Int32 *)&argc, argv, 
      cl_cmdln_options, stderr, 
      cl_header, 
      validateArgs, 
      &cl_opts)){
      if(DGP_ERROR_IS(cErrInvalidCmdLine)){
          MAIN_NORMAL_RETURN;
      }
  }

    /* Concoct an output filename if the user didn't specify one... */
    if((cl_opts.out_filename == NULL) || (cl_cmdln_options[OUTNAMEENTRY].found == DGP_FALSE)) {
        cl_opts.out_filename = NULL;
        DGP_CHECK_ALLOC_CLEANUP(cl_opts.out_filename = DGP_MALLOC((strlen(argv[0]) + 1) * sizeof(Dgp_Int8)));
        cl_opts.out_filename[0] = '\0'; 
        sprintf(cl_opts.out_filename, "%s", argv[0]);
    }
  DGP_ERR_PASS_CLEANUP(silx_initDecState(dgp_state, &sOpt));

 /* set the values in SlixDecOption, use NULL if you want to use default setting. */
 DGP_ERR_PASS_CLEANUP(silx_setDecState(
                dgp_state,       
                sOpt,                      /* point to SilxDecopt. Must be init first */
                &(cl_opts.force_color),     /* TRUE to force color output */
                &(cl_opts.force_mono),      /* TRUE to force mono output */
                &(cl_opts.rotation),
                &(cl_opts.start_page),
                &(cl_opts.stop_page),
                &(cl_opts.verbose),
                &(cl_opts.stripsize),
                &(cl_opts.jbig2Only),       /* only return jbig2 text */
                &(cl_opts.imagesOnly),      /* only return back ground */
                &(cl_opts.output_res),      /* output resolution, same in x and y direction */
                &(cl_opts.no_output),     /* TRUE for out to file or out to stdout, FALSE for out to buffer  */
                &(cl_opts.outtype),         /* ("TIF"/"BMP"/"PNM")         only valid when outToFile is TRUE and outFileName is not stdout. */
                &(cl_opts.out_filename)));  /* ("stdout"/ other file name) only valid when outToFile is chosen */

 
  /* If going to stdout, don't process multiple files.  Just take the first one. */
  for(docStart = 0; 
        docStart < ((!cl_opts.no_output && ! strcmp(cl_opts.out_filename, "stdout"))? 1:argc); docStart++){
      fp = fopen(argv[docStart], "rb");
      DGP_ERR_GEN_CLEANUP((!fp), cErrFileNotFound); 
      silx_decDocumentAll(dgp_state, sOpt, fp);
      fclose(fp);
  }
  if(cl_cmdln_options[OUTNAMEENTRY].found == DGP_FALSE){
     DGP_FREE(cl_opts.out_filename);
     cl_opts.out_filename = NULL;
  }
  silx_cleanDecState(dgp_state, &sOpt);
  t1 = get_utime();
  DGP_REPORTF1(cl_opts.verbose, "Decoding time:  %.3f secs.", (t1 - t0));
  Dgp_cleanupDgpState(&dgp_state);	

  MAIN_NORMAL_RETURN;

ERR_CLEANUP:
 if(cl_cmdln_options[OUTNAMEENTRY].found == DGP_FALSE && cl_opts.out_filename){
     DGP_FREE(cl_opts.out_filename);
     cl_opts.out_filename = NULL;
 }
  if(fp) fclose(fp);
  if(image) DGPIMAGE_FREE(image);
  if(sOpt) silx_cleanDecState(dgp_state, &sOpt);
  if(dgp_state)Dgp_cleanupDgpState(&dgp_state);	

  MAIN_ERR_RETURN;
}
/*************************************************************************/
