
#include "StdAfx.h"

#include <string>
#include <iostream>
#include <sstream>

#include "PDFReader.h"

static bool FlateToPixr ( ipcore::PIXR** ret_image, const char* buffer, Asd::Int32 buf_size, 
                         Asd::Int32 width, Asd::Int32 height, const bool inverse, const char* filename);

void PDFReader::handleFlate (objInfo& oInfo, BitFilter filter)
{

    std::string subKey;
    int width, height;
    bool hasKey;
    bool inverse = false;

    // get height
    subKey = "/Height";
    hasKey = hasKeyword (subKey, &height);
    
    // and width of image
    subKey = "/Width";
    hasKey = hasKeyword (subKey, &width);
    
    inverse = lookForDecode ();

    std::cout << "Flate Image " << oInfo.objNr << "  " << oInfo.pos << "  " << oInfo.size << "  " << height;
    std::cout << "  " << width << "  " << inverse << '\n';
    
    // read all the data of the object
    char *begin, *buf = new char [oInfo.size];
    m_file.seekg (oInfo.pos
#ifdef _WIN32
, std::ios_base::beg);
#else
, std::ios::beg);
#endif

    m_file.read (buf, oInfo.size);

    begin = strstr (buf, "stream");
    begin += 6;
    while (*begin == '\015' || *begin == '\012') begin++;
    int offset = (int) begin - (int) buf;

    // create imInfo struct
    imInfo *info = new imInfo;
    info->width = width;
    info->height = height;
    info->comp = Flate;
    info->jb2Encoded = false;
    oInfo.info = info;

    // check if we need to decode a pre filter
    if (filter == base85)
    {
        int newSize =  decodeAB85 (begin, oInfo.size - offset);
        if (newSize == 0 || newSize > oInfo.size - offset)
        {
            std::cout << "Decoding of ASCIIBase85 failed\n";
            delete [] buf;
            return;
        }
    }   

    ipcore::PIXR* ret_image;
    if (!FlateToPixr ( &ret_image, begin, oInfo.size - offset, 
                    width, height, inverse, NULL )) 
    {
        delete [] buf;
        std::cout << "Decoding of Flate failed\n";
        return;
    }
    delete [] buf;
    begin = NULL;

    // feed pixr to tokenizer
    bool status = m_pJB2->addPage (ret_image);
    
    freePixr (&ret_image);

    if (status) info->jb2Encoded = true;

    return;
}

bool PDFReader::lookForDecode ()
{
    std::string buffer;
    std::istringstream ist (m_tBuf);
    int find;

    while (ist >> buffer)
    {
        find = buffer.find ("/Decode");
        if (find == 0)
        {
            ist >> buffer;
            find = buffer.find ("[");
            if (find == 0)
            {
                int temp;
                ist >> temp;
                if (temp == 1) return true;
                if (temp == 0) return false;
            }
        }
    }
    return false;
}

// IPCORE G4 libs
namespace ipcore
{
#include "rcs.pub"
#include "types.pub"
#include "libutilw.pub"
#include "pixr.h"

#include "frames.pub"
#include "shrcmprs.pub"
#include "ccitt.pub"
#include "mmr.pub"

#include "geomadt.h"
#include "props.pub"

#include "rast.pub"

#include "reduce.pub"

#include "prxform.pub"

#include "rescale.pub"

#include "shrpixr.pub"

// for writeTiff
#include "imagerep.h"
#include "shrcc.pub"
#include "iotiff.pub"
#include "io.pub"

const int pixrRes = 300;

#include "zlib.h"

}

static bool FlateToPixr ( ipcore::PIXR** ret_image, const char* buffer, Asd::Int32 buf_size, 
                         Asd::Int32 width, Asd::Int32 height, const bool inverse, const char* filename)
{

    *ret_image = NULL;
    const Asd::Int32 framePix = 32;

    // ipcore setup
    ipcore::IPWState *pIPWState = NULL;
    Asd::Int32 status = ipcore::w_IPWStateInit (NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, 0, NULL, NULL, &pIPWState);

    // allocate temporary PIXR
    ipcore::PIXR* retPixr = NULL;
    Asd::UInt8 *data;
    Asd::Int32 bpl;
    status += ipcore::w_createPixr (pIPWState, width, height, 1, 
        ipcore::pixrRes, ipcore::pixrRes, framePix, &retPixr);
    status += ipcore::w_pixrGetImage (pIPWState, retPixr, &data);
    status += ipcore::w_pixrGetBpl (pIPWState, retPixr, &bpl);

    ipcore::z_stream Zstruct;
    Zstruct.next_in = (unsigned char*) buffer;  /* next input byte */
    Zstruct.avail_in = buf_size;  /* number of bytes available at next_in */
    Zstruct.next_out = data;    /* next output byte should be put there */
    Zstruct.avail_out = width / 8; /* remaining free space at next_out */

    Zstruct.zalloc = Z_NULL;  /* used to allocate the internal state */
    Zstruct.zfree = Z_NULL;   /* used to free the internal state */
    Zstruct.opaque = Z_NULL;  /* private data object passed to zalloc and zfree */
    Zstruct.data_type = Z_BINARY;  /* best guess about the data type: ascii or binary */

    // initialize the struct
    ipcore::inflateInit_ (&Zstruct, ZLIB_VERSION, sizeof(Zstruct));   

    for (int i = 0; i < height; i++)
    {
        Zstruct.next_out = data; /* next output byte should be put there */
        Zstruct.avail_out = (width + 7) / 8; /* remaining free space at next_out */

        ipcore::inflate (&Zstruct, Z_SYNC_FLUSH);

        // advance data pointer to next line in image
        data += bpl;
    }

    // free all memory in the decoder
    ipcore::inflateEnd (&Zstruct);

    // for some reason image is inverted and byte-flipped ...
    if (!inverse) status += ipcore::w_invert (pIPWState, retPixr, &retPixr);
    status += ipcore::w_flip (pIPWState, retPixr);

    // write image
    if (filename)
    status += ipcore::w_writeTiff (pIPWState, retPixr, 
        filename, COMPRESSION_CCITTFAX4 ,cTrue, 0);

    // cleanup
    status += ipcore::w_IPWStateTerm (&pIPWState);

    // check if there were any errors
    if (status) return false;

    *ret_image = retPixr;

    return true;

}
