
#include "StdAfx.h"

#include "Jb2Encode.h"

namespace ipcore
{
#include "rcs.pub"
#include "types.pub"
#include "libutilw.pub"
#include "pixr.h"

#include "frames.pub"

#include "geomadt.h"
#include "props.pub"

#include "rast.pub"
#include "prxform.pub"

#include "shrpixr.pub"

#include "dgp_err_printers.h"

extern "C"
{
extern ErrCode Dgp_addXeroxLogo (PIXR *image, Dgp_State *dgp_state);
}
}


Jb2Encode::~Jb2Encode ()
{
    ipcore::DgpApi_close (m_handle);
    ipcore::Dgp_cleanupDgpState((ipcore::Dgp_State **)(&(m_llTokOpt.dgp_state)));
}

Jb2Encode::Jb2Encode (bool arithmetic)
{
    ipcore::Dgp_Int32 ret = cSuccessful;
    ipcore::Dgp_State* dgp_state = NULL;

const ipcore::Dgp_UInt8 DEF_VERBOSE =    0;
const ipcore::Dgp_Float32 DEF_FRACHAUSD  = (ipcore::Dgp_Float32) 0.97;
const ipcore::Dgp_Int32 DEF_WIDTHDIFF  = 1;
const ipcore::Dgp_Int32 DEF_HEIGHTDIFF = 1;
const ipcore::Dgp_Int32 DEF_MAXDENSITY = 3;
const ipcore::Dgp_Int32 DEF_COMMITTHRESH =10000;
const ipcore::Dgp_Int32 DEF_FREEZEHISTTHRESH =10000;
const ipcore::Dgp_Int32 DEF_MAKEHISTTHRESH =10;
const ipcore::Dgp_Int32 DEF_RETURNHISTTHRESH =0;
const ipcore::Dgp_Int32 DEF_DILATIONRADIUS =1;
const ipcore::Dgp_Int32 DEF_SIZETHRESH = 1;
const ipcore::Dgp_Int32 DEF_WRITETHRESH =10;
const ipcore::Dgp_Int32 DEF_NEARESTNEIGHBORS =8;
const ipcore::Dgp_Int32 DEF_XSHIFT   =   3;
const ipcore::Dgp_Int32 DEF_XSLOPECHANGE= 6;
const ipcore::Dgp_Float32 DEF_XNEWSLOPE =  2.0;
const ipcore::Dgp_Int32 DEF_LOWERDILATIONLIMIT = 6;
const ipcore::Dgp_Int32 DEF_BASELINEORIENT =BO_DOWN;
const ipcore::Dgp_Int32 DEF_DESKEWACCURACY = 3;        // [0...5]
const ipcore::Dgp_UInt8 DEF_DESKEW= 0;
const ipcore::Dgp_Int32 DEF_FORCE_RES   =   -1;
const ipcore::Dgp_Int32 DEF_ROTATION          =  0;
const ipcore::Dgp_UInt8 DEF_G4TOKENS  =  0;
const ipcore::Dgp_UInt8 DEF_TOKENSIZELIMIT =1;
//const ipcore::Dgp_Int32 DEF_JBIG2PROFILE= 3;

    /* Init dgp state */    
    ret = ipcore::Dgp_initDgpState( &dgp_state, NULL, NULL, NULL, 
        ipcore::Dgp_fileErrPrinter, 1, 1);

    /*fill in DgpApi_Options*/
    m_opts.sizeThresh    =   DEF_SIZETHRESH;
    m_opts.classThresh   =   DEF_WRITETHRESH;
    m_opts.grayHalftone      =   0;
    m_opts.deskew        =   DEF_DESKEW;
    m_opts.forceRes		=   DEF_FORCE_RES;
    m_opts.baselineOrient=   DEF_BASELINEORIENT;
    m_opts.skewAccuracy  =   DEF_DESKEWACCURACY;
    m_opts.rotation      =   DEF_ROTATION;
	if (arithmetic)		 
		m_opts.jbig2Profile  =   3;
	else
		m_opts.jbig2Profile = 2;
    m_opts.g4_toks		=   DEF_G4TOKENS;
    m_opts.bitmapSizeLimit=  DEF_TOKENSIZELIMIT;  //limit token size be 10% of page width
    m_opts.verbose		=   DEF_VERBOSE;
    m_opts.membond       =   -1;             // default turn off mem upper limit
    m_opts.writeclasses  =   0;
    m_opts.instrument    =   0;
    m_opts.pdfOut        =   1;  // output jbig2 in pdf
    
    /* Fill in lossy tokenizer options */
    m_tokopt.dgp_state			 = dgp_state;
    m_tokopt.fracHausd			 = DEF_FRACHAUSD;
    m_tokopt.maxDense             = DEF_MAXDENSITY;
    m_tokopt.widthDiff			 = DEF_WIDTHDIFF;
    m_tokopt.heightDiff		     = DEF_HEIGHTDIFF;
    m_tokopt.commitThresh       = DEF_COMMITTHRESH;
    m_tokopt.freezeHistThresh   = DEF_FREEZEHISTTHRESH;
    m_tokopt.makeHistThresh     = DEF_MAKEHISTTHRESH;
    m_tokopt.returnHistThresh   = DEF_RETURNHISTTHRESH;
    m_tokopt.nneigh             = DEF_NEARESTNEIGHBORS;
    m_tokopt.xshift             = DEF_XSHIFT;
    m_tokopt.xslopechange       = DEF_XSLOPECHANGE;
    m_tokopt.newslope           = DEF_XNEWSLOPE;
    m_tokopt.dilationRadius     = DEF_DILATIONRADIUS;
    m_tokopt.lowerDilationLimit = DEF_LOWERDILATIONLIMIT;
    m_tokopt.flags              = TOK_OPT_ALIGN | TOK_OPT_RETURNOFFSET;
    m_tokopt.mergeCallback      = NULL;
    m_tokopt.callbackCookie     = NULL;
    
    /* Fill in lossless tokenizer options */
    m_llTokOpt.dgp_state		 = dgp_state;
    m_llTokOpt.flags			 = 0;
    m_llTokOpt.verbose			 = m_opts.verbose;
    m_llTokOpt.ascii			 = 0;
    m_llTokOpt.color			 = 0;
    m_llTokOpt.halftone		 = 0;
    m_llTokOpt.lossy			 = 0;
    m_llTokOpt.maxSamptokH		 = -1;
    m_llTokOpt.maxSamptokW		 = -1;
    m_llTokOpt.noascii			 = 1;
    m_llTokOpt.numStrips		     = 1;
    m_llTokOpt.sizeThresh		 = DEF_SIZETHRESH;
    m_llTokOpt.washThresh		 = 301;
    m_llTokOpt.tokenizerParams	 = NULL;
    m_llTokOpt.colorTokenizerParams = NULL;
    
    ret = ipcore::DgpApi_init (&m_handle, &(dgp_state->a), &(dgp_state->sm_a), 
        &m_opts, &m_tokopt, &m_llTokOpt, NULL);

    // set state and count to correct values
    m_addState = false;
    m_imCount = 0;
}

bool Jb2Encode::addPage (ipcore::PIXR* page)
{

    ipcore::DgpApi_PageData			pd;
    ipcore::Dgp_Int32               bpp;
    ipcore::IPWState *pIPWState = m_tokopt.dgp_state->ipwstate;
    
    // check if we are ready to add new images
    if (!(m_addState || startSection())) return false;

    //memory limit
    ipcore::Dgp_UInt8 overflow = 0;
    ipcore::Dgp_Int32 ret = 0;
    
    // Get information from PIXR
    ret +=  ipcore::w_pixrGetWidth(pIPWState, page, &pd.width);
    ret +=  ipcore::w_pixrGetHeight(pIPWState, page, &pd.height);
    ret +=  ipcore::w_pixrGetBpl(pIPWState, page, &pd.bpl);
    ret +=  ipcore::w_pixrGetResolutionDPI(pIPWState, page, &pd.xRes, &pd.yRes);
    ret +=  ipcore::w_pixrGetDepth(pIPWState, page, &bpp);

    // Add Xerox Logo, if necessary (for external release)
    ret += ipcore::Dgp_addXeroxLogo (page, m_tokopt.dgp_state);

    /* get image from pixr */
    ret += ipcore::w_pixrGetImageAll(pIPWState, page, &(pd.image));    

    // monochrome image
    ret += ipcore::DgpApi_addMonoPage (m_handle, &pd, &overflow);
    
    if (ret == 0 && overflow == 0)
    {
        m_imCount++;
        return true;
    }

    return false;
}

bool Jb2Encode::endSectionGetDict (char** data, int* data_size)
{
    ipcore::DgpApi_Jbig2Data		dict;
    ipcore::Dgp_Int32 ret = 0;

    // we can only end the section if we are in addState
    if (!m_addState) return false;
    endSection ();

    ret += ipcore::DgpApi_getDict(m_handle, 1, &dict);

    *data = dict.data;
    *data_size = dict.data_size;

    if (ret == 0) return true;

    return false;
}

bool Jb2Encode::getNextPosBlock (char** data, int* data_size)
{
    ipcore::DgpApi_Jbig2Data posBlock;
    ipcore::Dgp_Int32 ret = 0;

    *data = NULL;
    *data_size = 0;

    if (!m_addState && m_imCount > 0)
    {
        ret += ipcore::DgpApi_getNextPosBlock (m_handle, &posBlock);
        *data = posBlock.data;
        *data_size = posBlock.data_size;
        m_imCount--;
        if (ret == 0) return true;
    }

    return false;
}

bool Jb2Encode::startSection ()
{
    // check if we have retrieved all compressed data
    if (m_imCount > 0) return false;

    ipcore::Dgp_Int32 ret = ipcore::DgpApi_startSection (m_handle);

    m_imCount = 0;
    m_addState = true;

    if (ret == 0) return true;
    return false;
}

bool Jb2Encode::endSection ()
{
    ipcore::Dgp_Int32 ret = ipcore::DgpApi_endSection(m_handle, 0);
    // turn addState off
    m_addState = false;

    if (ret == 0) return true;
    return false;
}
