
#include "StdAfx.h"

#include "PdfOutput.h"
#include "PDFReader.h"

#include <string>
#include <iostream>
#include <sstream>
#include <algorithm>

#ifdef _WIN32
#define IOS ios_base
#else
#define IOS ios
#endif

PDFReader::~PDFReader ()
{
    if (m_pJB2) delete m_pJB2;
    if (m_tBufSize > 0) delete [] m_tBuf;
}

int PDFReader::getObjects ()
{
    int nextPos, nObj = 0;
    if (!m_file) return nObj;
        
    // find the end of the input file    
    if (m_file)    
    {    
        m_file.seekg (0, std::IOS::end);        
        m_fileEnd = m_file.tellg ();        
    }
    
    // find and parse the xref table
    nextPos = findXref();
    if (nextPos == 0) return nObj;

    parseXref (nextPos);
    nObj = cleanXref();
    return nObj;
}

bool PDFReader::findAndEncodeBinaryImages ()
{
    const int large = 1024;  // at least 1 kb to be interesting ...
    const int imHeaderSize = 1024; // to be save that all of the header is read

    if (!m_file) return false;

    // find binary images, and then send them to tokenizer
    int value;
    std::string key = "/BitsPerComponent";
    std::string antiKey = "/ImageMask";
    bool hasMask = false;

    bool hasKey;
    bool hasChanges = false;
    std::vector<objInfo>::iterator iter = m_objs.begin(), last = m_objs.end();
    while (iter != last)
    {
        if ((*iter).size > large)
        {
            m_file.seekg ((*iter).pos, std::IOS::beg);
            readBuffer (imHeaderSize);
            hasKey = hasKeyword (antiKey, &hasMask);
            hasKey = hasKeyword (key, &value);
            if (hasKey && value == 1)
            {
                BitFilter bitFilter = NONE;
                std::string subKey = "/ASCIIHexDecode";
                // see if it is  encoded
                if (hasKeyword (subKey, (int*) NULL))
                {
                    std::cout << "IMPORTANT EXAMPLE !! Please send PDF file to mkrocken@parc.com\n";
                }
                subKey = "/ASCII85Decode";
                // see if it is  encoded
                if (hasKeyword (subKey, (int*) NULL))
                {
                    bitFilter = base85;
                }

                subKey = "/CCITTFaxDecode";
                // see if it is G3/4 encoded
                if (hasKeyword (subKey, (int*) NULL))
                {
                    handleG4 (*iter, bitFilter);
                    if ((*iter).info && (*iter).info->jb2Encoded)
                        hasChanges = true;
                    if ((*iter).info && hasMask)
                        (*iter).info->mask = true;
                    if ((*iter).info && !hasMask)
                        (*iter).info->mask = false;
                }

                subKey = "/JBIG2Decode";
                // see if it is JBIG2 encoded
                if (hasKeyword (subKey, (int*) NULL))
                {
                    std::cout << "JBIG2 encoded\n";
                }
                subKey = "/LZWDecode";
                // see if it is  encoded
                if (hasKeyword (subKey, (int*) NULL))
                {
                    std::cout << "IMPORTANT EXAMPLE !! Please send PDF file to mkrocken@parc.com\n";
                }
                subKey = "/FlateDecode";
                // see if it is  encoded
                if (hasKeyword (subKey, (int*) NULL))
                {
                    handleFlate (*iter, bitFilter);
                    if ((*iter).info && (*iter).info->jb2Encoded)
                        hasChanges = true;
                    if ((*iter).info && hasMask)
                        (*iter).info->mask = true;
                    if ((*iter).info && !hasMask)
                        (*iter).info->mask = false;
                //    (*iter).info->jb2Encoded = false;
                }
                subKey = "/RunLengthDecode";
                // see if it is  encoded
                if (hasKeyword (subKey, (int*) NULL))
                {
                    std::cout << "IMPORTANT EXAMPLE !! Please send PDF file to mkrocken@parc.com\n";
                }   
            }
        }
        ++iter;
    }
    return hasChanges;
}

bool PDFReader::generateOutputPDF(const char* outfilename)
{

    // check if there is any JBIG2 data
    if (!m_pJB2 || !m_pJB2->inAddState() || m_pJB2->getImgCount() == 0)
    {
        std::cout << "No JBIG2 data for new file found\n";        
        return false;        
    }

    // create the output file    
    PdfOutput* pPDFOut = new PdfOutput;    
    bool pdfRet = pPDFOut->PdfStartDoc (outfilename);    
    if (!pdfRet)     
    {    
        std::cout << "Output file cannot be opened\n";        
        return false;        
    }
    if (!m_file) return false;

    // end the JBIG2 section and get the dictionary
    char *dict, *posBlock;
    int dict_size, posBlock_size;
    m_pJB2->endSectionGetDict (&dict, &dict_size);
    int dictID = m_Size;
    std::cout << "Dict size  " << dict_size << '\n';

    // create the output file by copying/replacing the objects
    std::vector<objInfo>::iterator iter = m_objs.begin(), last = m_objs.end();
    ++iter; // don't copy the 0th element
    while (iter != last)
    {
        if ((*iter).info)
        {
            imInfo *info = (*iter).info;
            bool ret;
            if (info->jb2Encoded)
            {
                posBlock = NULL; posBlock_size = 0;
                ret = m_pJB2->getNextPosBlock (&posBlock, &posBlock_size);
                pPDFOut->writeJBIG2 (posBlock, posBlock_size, (*iter).objNr, info->height, info->width, dictID, info->mask);
                m_pJB2->freeData (posBlock);
                std::cout << "got image " << (*iter).objNr << "  " << posBlock_size << "  " << ret << '\n';
            }
            else
            {
                if ((*iter).size > 0)
                {
                    // copy object
                    char* data = new char [(*iter).size];
                    m_file.seekg ((*iter).pos, std::IOS::beg);
                    m_file.read (data, (*iter).size);
                    pPDFOut->writeObj (data, (*iter).size, (*iter).objNr, (*iter).version);
                    delete [] data;
                }
            }
            // delete the info block
            delete info;
        }
        else
        {
            if ((*iter).size > 0)
            {
                // copy object
                char* data = new char [(*iter).size];
                m_file.seekg ((*iter).pos, std::IOS::beg);
                m_file.read (data, (*iter).size);
                pPDFOut->writeObj (data, (*iter).size, (*iter).objNr, (*iter).version);
                delete [] data;
            }
        }
        ++iter;
    }

    // add the dictionary object
    pPDFOut->writeJBIG2Dict (dict, dict_size, dictID);
    m_pJB2->freeData (dict);

    // finish the PDF file
    pPDFOut->writeTrailer (m_Root, m_Info);
    int outSize = pPDFOut->PdfEndDoc ();
    std::cout << "Compression : " << (float) m_fileEnd / (float) outSize << '\n';
    delete pPDFOut;

    return true;
}


int PDFReader::findXref()
{
    int XrefPos = 0;
    
    // find position of xref table
    int endOffset = 100;

    while (XrefPos == 0 && endOffset < 2000)
    {
    // seek to xref table
    m_file.seekg (-endOffset, std::IOS::end);
    readBuffer (endOffset);

    hasKeyword ("startxref", &XrefPos);

    endOffset += 100;
    }

    return XrefPos;
}

void PDFReader::parseXref (int pos)
{
    std::string buffer;
    int nObj = 0;
    const int unused = 100000;

    // seek to xref table
    m_file.seekg (pos, std::IOS::beg);

    m_file >> buffer;
    if (buffer != "xref") return;
    m_xrefPos.push_back (pos);

    while (buffer != "trailer")
    {
        int first, num;
        
        if (buffer == "xref")
            m_file >> first;
        else
            first = atoi(buffer.c_str());
        
        m_file >> num;
        nObj = num + first;
        
        // allocate a vector to hold the entries
        if (m_xrefVec.size() < nObj) 
        {
            m_xrefVec.resize(nObj);
            int j = m_version.size();
            m_version.resize(nObj);
            for ( ; j < nObj; j++)
                m_version[j] = unused;
        }
        
        // parse the table
        for (int i = first; i < nObj; i++)
        {
            int temp;
            int version;
            m_file >> temp;
            m_file >> version; 
            m_file >> buffer;
            if (m_version[i] == unused)
            {
                if (buffer == "n")
                {
                    m_xrefVec[i] = temp;
                    m_version[i] = version;
                }
                else
                {
                    m_xrefVec[i] = 0;
                    m_version[i] = -1 * version;
                }
            }
        }
        
        // look for trailer
        m_file >> buffer;
    }

    // read the trailer
    readBuffer (1024);

    int retVal;
    // look for catalog, size, info only in main table
    if (!m_moreXref)
    {

        if (hasKeyword ("/Encrypt", &retVal)) return; // don't deal with encrypted docs
        if (hasKeyword ("/Size", &retVal)) m_Size = retVal;
        if (hasKeyword ("/Root", &retVal)) m_Root = retVal;
        if (hasKeyword ("/Info", &retVal)) m_Info = retVal;
    }

    // if there are more xref tables, parse them
    if (hasKeyword ("/Prev", &retVal)) 
    {
        m_moreXref = true;
        parseXref (retVal);
    }
    
    return;
}

int PDFReader::cleanXref ()
{

    // now we have all entries. Sort them and add sizes
    const int vecSize = m_xrefVec.size();
    if (vecSize == 0) return 0;
    m_objs.resize (vecSize);

    int i;
    for (i = 0; i < vecSize; i++)
    {
        m_objs[i].objNr = i;
        m_objs[i].pos = m_xrefVec[i];
        m_objs[i].info = NULL;
        m_objs[i].version = m_version[i];
    }

    std::sort (m_objs.begin(), m_objs.end());

    for (i = 1; i < vecSize; i++)
    {
        m_objs[i-1].size = m_objs[i].pos - m_objs[i-1].pos;
//        std::cout << m_objs[i-1].objNr << "  " << m_objs[i-1].pos << "  " << m_objs[i-1].size << '\n';
    }
    m_objs[vecSize - 1].size = m_fileEnd - m_objs[vecSize - 1].pos;

    // fix sizes with the known positions of xRef tables
    std::list<int>::iterator iter = m_xrefPos.begin(), end = m_xrefPos.end();
    while (iter != end)
    {
        i = 0;
        while (i < vecSize && m_objs[i].pos < (*iter)) i++;
        if (i > 0 && m_objs[i-1].pos < (*iter))
            m_objs[i-1].size = (*iter) - m_objs[i-1].pos;
        ++iter;
    }
        
/*    for (i = 0; i < vecSize; i++)
    {
        std::cout << m_objs[i].objNr << "  " << m_objs[i].pos << "  " << m_objs[i].size << "  " << m_objs[i].version << '\n';
    } */

    // look if the PDF was linearized. If so, remove the linearization objects
    m_file.seekg (m_objs[0].pos, std::IOS::beg);
    readBuffer (1024);
    if (hasKeyword ("/Linearized", (int*) NULL))
    {
        m_objs[0].size = -1;

        std::cout << "is linearized\n";

        // looks for the hints table and remove it
        int linobj;
        hasKeyword ("[", &linobj);

        std::cout << linobj << "   ";
        i = 0;
        while (m_objs[i].pos < linobj) i++;
        if (m_objs[i].pos == linobj)
        {
            std::cout << m_objs[i].objNr << '\n';
            m_objs[i].size = -1;
        }
    }

    // clean up the temp vars
    m_xrefPos.clear();
    m_version.clear();
    m_xrefVec.clear();

    return vecSize;
}

int PDFReader::readBuffer (const int size)
{
    if (m_tBufSize < size)
    {
        delete [] m_tBuf;
        m_tBuf = new char [size];
        if (!m_tBuf) 
        {
            m_tBufSize = 0;
            return 0;
        }
        m_tBufSize = size;
    }
    
    int distToEnd = m_fileEnd - m_file.tellg();

    if (distToEnd >= size)
    {
        m_file.read (m_tBuf, size);
        return size;
    }

    // not enough bytes in the file
    memset (m_tBuf, 0, size);
    int small_size = distToEnd;
    m_file.read (m_tBuf, small_size);
    return small_size;

}


bool PDFReader::hasKeyword (const std::string& keyword, int* value)
{
    std::string buffer;
    std::istringstream ist (m_tBuf);
    int find;

    while (ist >> buffer)
    {
        find = buffer.find (keyword);
        if (find >= 0)
        {
            if (value != NULL)
                ist >> *value;
            return true;
        }
    }

    return false;
}

bool PDFReader::hasKeyword (const std::string& keyword, bool* value)
{
    std::string buffer;
    std::istringstream ist (m_tBuf);
    int find, findt, findf;

    while (ist >> buffer)
    {
        find = buffer.find (keyword);
        if (find >= 0)
        {
            ist >> buffer;
            findt = buffer.find ("true");
            findf = buffer.find ("false");
            if (findt >= 0) 
                *value = true;
            else if (findf >= 0) 
                *value = false;
            else return false;
            return true;
        }
    }

    return false;
}

