
#include <fstream>
#include <list>
#include <vector>
#include <string>
#include "AsdTypes.h"
#include "Jb2Encode.h"

#ifndef PDFREADER_H
#define PDFREADER_H

enum Compression { G4, JBIG2, Flate, None };

enum BitFilter { base85, NONE };

struct imInfo
{
    bool jb2Encoded;
    bool mask;
    int width, height;
    Compression comp;
};

struct objInfo
{
    int objNr;
    int size;
    int pos;
    imInfo *info;
    int version;
    int operator< (const objInfo& other) const { return (pos < other.pos); }
};

class PDFReader
{
public:

    PDFReader (const char* in_filename, bool arithmetic) : 
      m_file(in_filename
#ifdef _WIN32
, std::ios_base::binary)
#else
)
#endif
     , m_moreXref (false), m_xrefVec (), m_xrefPos (), m_version(), m_tBuf(NULL), m_tBufSize(0)
    {
        m_pJB2 = new Jb2Encode(arithmetic);
    }
    ~PDFReader ();

    int getObjects ();
    bool findAndEncodeBinaryImages ();
    bool generateOutputPDF(const char* out_filename);

private:

    /* fills m_objs vector */
    int findXref ();
    void parseXref (int pos);
    int cleanXref ();

    /* buffers and keywords */
    char *m_tBuf;
    int m_tBufSize;
    int readBuffer (const int size);
    bool hasKeyword (const std::string& keyword, int* value);
    bool hasKeyword (const std::string& keyword, bool* value);

    /* decode G4 and Flate */
    void handleG4 (objInfo& oInfo, BitFilter filter);
    void handleFlate (objInfo& oInfo, BitFilter filter);
    void PDFReader::freePixr ( ipcore::PIXR** ret_image );
    bool lookForDecode ();

    /* decode ASCIIBase85  */
    int decodeAB85 (char* inbuf, int buf_size);

    std::ifstream m_file;
    std::vector<objInfo> m_objs;
    int m_fileEnd;
    int m_Size, m_Root, m_Info;

    Jb2Encode* m_pJB2;

    /* some temp storage for recursive parseXref() func */
    std::list<int> m_xrefPos;
    std::vector<int> m_xrefVec;
    std::vector<int> m_version;
    bool m_moreXref;

};

#endif
