
#include "StdAfx.h"


#include <string>
#include <iostream>
#include <sstream>

#include "PDFReader.h"

static bool g4ToPixr ( ipcore::PIXR** ret_image, const char* buffer, Asd::Int32 buf_size, 
                         Asd::Int32 width, Asd::Int32 height, Asd::Int32 k, bool inverse, const char* filename);


// decode ASCIIBase85. Returns the size of the buffer on output
int PDFReader::decodeAB85 (char* inbuf, int buf_size)
{
    const int p[5] = {52200625, 614125, 7225, 85, 1};
    
    const int pp1 = 256;
    const int pp2 = 65536;
    const int pp3 = 16777216;
    
    unsigned int total = 0, temp;
    
    int val, count = 0;
    char *outBuf = new char [buf_size];
    char *outP = outBuf;
    char *begin = inbuf;
    int outcount = 0;
    
    while (1) // infinite loop
    {
        while ((val = (int)*begin++) < 33) // get a new val, skip all white chars
        {
            if (begin - inbuf > buf_size) return 0;
        }
        if (val == 126) // ~
        {
            // output last count - 1 bytes
            if ( count ) count--;
            if ( count )
            {
                outcount++;
                if (outcount > buf_size) return outcount;

                temp = total / pp3;
                *outP++ = (char) temp;
                total -= temp * pp3;
                count--;
            }
            
            if ( count )
            {
                outcount++;
                if (outcount > buf_size) return outcount;

                temp = total / pp2;
                *outP++ = (char) temp;
                total -= temp * pp2;
                count--;
            }
            
            if ( count )
            {
                outcount++;
                if (outcount > buf_size) return outcount;

                temp = total / pp1;
                *outP++ = (char) temp;
                total -= temp * pp1;
            }
            break; // we are done
        }
        
        // a valid character ...
        if (val <= 117)
        {
            total += (val - 33) * p[count];
            count ++;
        }
        else if (val == 122) // z
        {
            outcount += 4;
            if (outcount > buf_size) return outcount;
            *outP++ = 0;
            *outP++ = 0;
            *outP++ = 0;
            *outP++ = 0;
        }
        else
            return false;
        
        if ( count == 5 )
        {
            outcount += 4;
            if (outcount > buf_size) return outcount;

            temp = total / pp3;
            *outP++ = (char) temp;
            total -= temp * pp3;
            
            temp = total / pp2;
            *outP++ = (char) temp;
            total -= temp * pp2;
            
            temp = total / pp1;
            *outP++ = (char) temp;
            total -= temp * pp1;
            
            *outP++ = (char) total;
            count = 0;
            total = 0;
        }
    }

    // copy the decoded data into the buffer
    memcpy (inbuf, outBuf, buf_size);
    
    delete [] outBuf;
    return outcount;

}


void PDFReader::handleG4 (objInfo& oInfo, BitFilter filter)
{

    std::string subKey;
    int width, height, k = 0, cols = 1728;
    bool hasKey;
    bool inverse = false;

    // get height
    subKey = "/Height";
    hasKey = hasKeyword (subKey, &height);
    
    // and width of image
    subKey = "/Width";
    hasKey = hasKeyword (subKey, &width);
    
    subKey = "/K";
    hasKey = hasKeyword (subKey, &k);

    subKey = "/Columns";
    hasKey = hasKeyword (subKey, &cols);

    subKey = "/BlackIs1";
    hasKey = hasKeyword (subKey, &inverse);
    inverse ^= lookForDecode ();

    std::cout << "Binary G3/4 Image " << oInfo.objNr << "  " << oInfo.pos << "  " << oInfo.size << "  " << height;
    std::cout << "  " << width << "  " << k << "  " << cols << "  " << inverse << '\n';
    
    // read all the data of the object
    char *begin, *buf = new char [oInfo.size];
    m_file.seekg (oInfo.pos, 
#ifdef _WIN32
std::ios_base::beg);
#else
std::ios::beg);
#endif

    m_file.read (buf, oInfo.size);

    begin = strstr (buf, "stream");
    begin += 6;
    while (*begin == '\015' || *begin == '\012') begin++;
    int offset = (int) begin - (int) buf;

    // create imInfo struct
    imInfo *info = new imInfo;
    info->width = width;
    info->height = height;
    info->comp = G4;
    info->jb2Encoded = false;
    oInfo.info = info;

    // check if we need to decode a pre filter
    if (filter == base85)
    {
        int newSize =  decodeAB85 (begin, oInfo.size - offset);
        if (newSize == 0 || newSize > oInfo.size - offset)
        {
            std::cout << "Decoding of ASCIIBase85 failed\n";
            delete [] buf;
            return;
        }
    }   

    // decode G3/4
    ipcore::PIXR* ret_image = NULL;    
    if (!g4ToPixr ( &ret_image, begin, oInfo.size - offset, 
        width, height, k, inverse, NULL )) 
    {
        delete [] buf;
        std::cout << "Decoding of G4 failed\n";
        return;
    }
    delete [] buf;
    begin = NULL;

    // feed pixr to tokenizer
    bool status = m_pJB2->addPage (ret_image);
    
    freePixr (&ret_image);

    if (status) info->jb2Encoded = true;

    return;
}


// IPCORE G4 libs
namespace ipcore
{
#include "rcs.pub"
#include "types.pub"
#include "libutilw.pub"
#include "pixr.h"

#include "frames.pub"
#include "shrcmprs.pub"
#include "ccitt.pub"
#include "mmr.pub"

#include "geomadt.h"
#include "props.pub"

#include "rast.pub"

#include "reduce.pub"

#include "prxform.pub"

#include "rescale.pub"

#include "shrpixr.pub"

// for writeTiff
#include "imagerep.h"
#include "shrcc.pub"
#include "iotiff.pub"

const int pixrRes = 300;

}

void PDFReader::freePixr ( ipcore::PIXR** ret_image )
{
    // ipcore setup
    ipcore::IPWState *pIPWState = NULL;
    Asd::Int32 status = ipcore::w_IPWStateInit (NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, 0, NULL, NULL, &pIPWState);

    ipcore::w_destroyPixr (pIPWState, ret_image);

    status += ipcore::w_IPWStateTerm (&pIPWState);

}

static bool g4ToPixr ( ipcore::PIXR** ret_image, const char* buffer, Asd::Int32 buf_size, 
                         Asd::Int32 width, Asd::Int32 height, Asd::Int32 k, bool inverse, const char* filename)
{

    *ret_image = NULL;
    const Asd::Int32 framePix = 32;

    // ipcore setup
    ipcore::IPWState *pIPWState = NULL;
    Asd::Int32 status = ipcore::w_IPWStateInit (NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, 0, NULL, NULL, &pIPWState);

    // allocate temporary PIXR
    ipcore::PIXR* retPixr = NULL;
    Asd::UInt8 *data;
    Asd::Int32 bpl;
    status += ipcore::w_createPixr (pIPWState, width, height, 1, 
        ipcore::pixrRes, ipcore::pixrRes, framePix, &retPixr);
    status += ipcore::w_pixrGetImage (pIPWState, retPixr, &data);
    status += ipcore::w_pixrGetBpl (pIPWState, retPixr, &bpl);

    // find first and last row of the band that we need
    Asd::UInt32 firstRow = 0;
    Asd::UInt32 nRow = height;

	// allocate decode structures
    Asd::Int32 stateBlockSize;
    status += ipcore::i_getSizeOfCCITTDecompressionInfo( pIPWState,
	    width + 2 * framePix, &stateBlockSize);
    void *compInfo = (void*) new Asd::UInt8 [stateBlockSize];
    status += ipcore::i_initializeCCITTDecompressionInfo(
	    pIPWState,
	    compInfo, width + 2 * framePix, k < 0,
	    cPadToByte, stateBlockSize);

    // clear the image
    status += ipcore::w_clear (pIPWState, retPixr);

    // set up for decode
    Asd::UInt32 imgExh = 0, cmpExh = 0;
    Asd::UInt8 *iPtr = data - framePix / 8;
    Asd::UInt8 *cPtr = (Asd::UInt8*) buffer;
    Asd::Int32 bytesLeft = buf_size, bytesUsed, pct;
    bool contImg = 0, contCmp = 0;
    
    // call the decoder
    do {
        if (k < 0)
        status += ipcore::i_decompressRowsMMR (pIPWState, compInfo,
	    iPtr, firstRow,
        nRow, bpl, cFalse, cPtr, bytesLeft, contImg,
	    contCmp, 1, &imgExh, &cmpExh, &pct, &bytesUsed);
        else
        status += ipcore::i_decompressRowsMR (pIPWState, compInfo,
	    iPtr, firstRow,
        nRow, bpl, cFalse, cPtr, bytesLeft, contImg,
	    contCmp, 1, &imgExh, &cmpExh, &pct, &bytesUsed);

        if (status) return false;

        bytesLeft -= bytesUsed;

	    if (imgExh == 1 && cmpExh == 0) {

	        /* Okay, we can live with it, but it's wrong */

	    } else if (cmpExh == 1 && imgExh == 0) {

            return false;
        }
    	contImg = 1;
	    contCmp = 1;
    } while (imgExh == 0 && cmpExh == 0 && bytesLeft > 0);

    if (inverse) status += ipcore::w_invert (pIPWState, retPixr, &retPixr);
    // write image
    if (filename)
    status += ipcore::w_writeTiff (pIPWState, retPixr, 
        filename, COMPRESSION_CCITTFAX4 ,cTrue, 0);

    // cleanup
    delete [] compInfo; 
    status += ipcore::w_IPWStateTerm (&pIPWState);

    // check if there were any errors
    if (status) return false;

    *ret_image = retPixr;

    return true;
}
