
#include "StdAfx.h"
#include "AdapterFactory.h"
#include "IPCoreImgBuf.h"

#ifndef REALOS
#include "IntelImgBuf.h"
#else
#include "SilxImgBuf.h"
#endif

AdapterFactory::~AdapterFactory ()
{
    // delete all ImageAdapters
    for (std::list<ImageAdapter*>::iterator adapter = m_adapList.begin(); 
         adapter != m_adapList.end(); adapter++)
    {
        delete *adapter;
    }
}


ImageAdapter* AdapterFactory::createInstance (const ColorDepth & depth, const Asd::Rect & region, 
                                    ImageBuf * buf)
{
    ImageAdapter *ibuf = NULL;

    if (depth == MONOCHROME || depth == COLOR_8_BIT)
    {
        // IPCore handles those
        ibuf = new IPCoreImgBuf(region.width(), region.height(), depth, buf);
    }
    else
    {
#ifndef REALOS
        // INTEL for Windows
        ibuf = new IntelImgBuf (region.width(), region.height(), depth, buf);
#else
        // SILX for Unix
        ibuf = new SilxImgBuf (region.width(), region.height(), depth, buf);
#endif
    }

    if (ibuf) m_adapList.push_back(ibuf);
    return ibuf;

}

ImageAdapter* AdapterFactory::createSub (Asd::Rect & region, ImageAdapter * adapt)
{
#ifndef REALOS
    IntelImgBuf* intel = dynamic_cast<IntelImgBuf*> (adapt);
#else
    SilxImgBuf* silx = dynamic_cast<SilxImgBuf*> (adapt);
#endif
    IPCoreImgBuf* ipcore = dynamic_cast<IPCoreImgBuf*> (adapt);

    ImageAdapter* ibuf = NULL;
#ifndef REALOS
    if (intel)
        ibuf = new IntelImgBuf (region, *intel);
#else
    if (silx)
        ibuf = new SilxImgBuf (region, *silx);
#endif
    if (ipcore)
        ibuf = new IPCoreImgBuf (region, *ipcore);

    if (ibuf) m_adapList.push_back(ibuf);
    return ibuf;
}
