#ifndef __DibImgBuf_h_Included__
#define __DibImgBuf_h_Included__

#include "AsdTypes.h"
#include "ImageBuf.h"

#ifdef _WIN32
#include <Windows.h>
#else

typedef struct tagRGBQUAD 
{
    Asd::UInt8    rgbBlue; 
    Asd::UInt8    rgbGreen; 
    Asd::UInt8    rgbRed; 
    Asd::UInt8    rgbReserved; 
}
RGBQUAD;
 
typedef struct tagBITMAPINFOHEADER
{
  Asd::UInt32 biSize; 
  Asd::Int32  biWidth; 
  Asd::Int32  biHeight; 
  Asd::UInt16 biPlanes; 
  Asd::UInt16 biBitCount; 
  Asd::UInt32 biCompression; 
  Asd::UInt32 biSizeImage; 
  Asd::Int32 biXPelsPerMeter; 
  Asd::Int32 biYPelsPerMeter; 
  Asd::UInt32 biClrUsed; 
  Asd::UInt32 biClrImportant; 
}
BITMAPINFOHEADER, *PBITMAPINFOHEADER; 

typedef struct tagBITMAPINFO
{ 
  BITMAPINFOHEADER bmiHeader; 
  RGBQUAD          bmiColors[1]; 
}
BITMAPINFO, *PBITMAPINFO;

#define BI_RGB 0;

#endif
    

class DibImgBuf : public ImageBuf
{
public:
	
	DibImgBuf( void );

    virtual Asd::UInt32 width( void ) const;
	virtual Asd::UInt32 height( void ) const;

	virtual ColorDepth  depth( void ) const;

    const BITMAPINFO & getBmpInfo( void ) const;

    /* By default, resize makes a bitmap top-down. */
    virtual void resize( const Asd::Point32 &, ColorDepth, Asd::UInt32 );
    
    bool isTopDown ( void ) const { return bmih().biHeight < 0; }
    bool isBottomUp( void ) const { return !isTopDown(); }

    void setTopDown( void )
    {
        bmih().biHeight = -Asd::Int32(height());
    }

    void setBottomUp( void )
    {
        bmih().biHeight =  Asd::Int32(height());
    }

    void flipData( void );

private:

    std::valarray<Asd::UInt8> m_headerBuf;

    void setColorDepth( ColorDepth );

    BITMAPINFO & bmi( void )
    {
        return *reinterpret_cast<BITMAPINFO*>(&m_headerBuf[0]);
    }

    BITMAPINFOHEADER & bmih( void )
    {
        return *reinterpret_cast<BITMAPINFOHEADER*>(&m_headerBuf[0]);
    }

    const BITMAPINFOHEADER & bmih( void ) const;

    static Asd::UInt32 calcBpl( const Asd::UInt32 width, 
        const Asd::UInt32 bpp, const Asd::UInt32 align );

    void fixWidth( void );
};
#endif
