#include "StdAfx.h"
#include "IPCoreImgBuf.h"

/*when compiling with IPCORE-PSG these guys
  show up as undefined for some reason*/

#ifndef COMPRESSION_CCITTFAX4
#define COMPRESSION_CCITTFAX4 4
#endif

#ifndef COMPRESSION_LZW 
#define COMPRESSION_LZW 5
#endif

#ifdef PSG
#undef cTrue //was formerly defined as ((Bool) 1) which caused a parse error
#define cTrue 1
#endif

// ipcore libs
namespace ipcore
{
#include "geomadt.h"
#include "props.pub"

// for writeTiff
#include "imagerep.h"
#include "shrcc.pub"
#include "iotiff.pub"
#include "io.pub"
#include "rescale.pub"
#include "frame.pub"
#include "shrpixr.pub"
#include "prxform.pub"
}


// empty image, 1 bit per pixel
IPCoreImgBuf::IPCoreImgBuf () : ImageAdapter(NULL), m_pixr(NULL)
{
    createState ();
    Asd::Int32 status = ipcore::w_createPixr (m_IPWState, 0, 0, 1, ipcore::pixrRes, 
        ipcore::pixrRes, ipcore::framePix, &m_pixr);
}


// wrap a pixr around an existing ImageBuf
IPCoreImgBuf::IPCoreImgBuf (Asd::UInt32 width, Asd::UInt32 height, ColorDepth depth, 
                            ImageBuf* pBuf) : ImageAdapter (pBuf), m_pixr(NULL)
{

    assert (depth == MONOCHROME || depth == COLOR_8_BIT);
    createState ();

    // set size to (1,1), because there are some crashes on (0,0)
    if (height == 0) height = 1;
    if (width == 0) width = 1;

    // check that we got the right size
    Asd::Int32 bpl, bufSize;
    Asd::UInt32 bitsPerPixel = depth;
    // Asd::Int32 status = 0;
    Asd::Int32 status = ipcore::w_pixrImageProperties(m_IPWState, width, height, bitsPerPixel, ipcore::framePix, &bpl, &bufSize);

    // pad by one, since IPCore likes larger buffers
    if (bpl > 0)
    {
        Asd::Int32 ipcHeight = bufSize / bpl + 1;
        Asd::Int32 ipcWidth = bpl * 8 / depth;
        // resize pBuf to accomodate pixr frame
        pBuf->resize (Asd::Point32(ipcWidth, ipcHeight), depth, 4);
        assert (bpl == pBuf->bpl() && bufSize <= pBuf->dataSize());
    }
    
    // generate the pixr header
    Asd::UInt8* dataWithFrame = pBuf->data();
    status += ipcore::w_createPixrStructOnly (m_IPWState, width, height, bitsPerPixel, 
        ipcore::pixrRes, ipcore::pixrRes, ipcore::framePix, dataWithFrame, &m_pixr);
    status += ipcore::w_pixrSetBpl (m_IPWState, m_pixr, pBuf->bpl ());
}

// create a new header for the region specified by roi
IPCoreImgBuf::IPCoreImgBuf (Asd::Rect & roi, ImageAdapter & buf) : 
ImageAdapter (buf.getIBuf()), m_pixr(NULL)
{
    createState ();

    // get a pointer to the beginning of the region, adjusted for the frame
    Asd::UInt8* dataWithFrame = buf.data() + roi.top() * buf.bpl() + roi.left() * buf.depth() / 8;
    dataWithFrame -= (ipcore::framePix / 8) * (buf.bpl() + 1);

    Asd::UInt32 status = ipcore::w_createPixrStructOnly (m_IPWState, roi.width(), roi.height(),
        buf.depth(), ipcore::pixrRes, ipcore::pixrRes, ipcore::framePix, dataWithFrame, &m_pixr);

    // fix the bytes per line
    status += ipcore::w_pixrSetBpl (m_IPWState, m_pixr, buf.bpl());

}

// Destructor
IPCoreImgBuf::~IPCoreImgBuf () 
{ 
    // remove pointers to data
    m_pixr->pixrData.plane.image = NULL;
    m_pixr->pixrData.plane.imageBuffer = NULL;

    // free all memory
	Asd::Int32 status = ipcore::w_destroyPixr(m_IPWState, &m_pixr); 
    status = ipcore::w_IPWStateTerm (&m_IPWState);
}

// dimensions
Asd::UInt32 IPCoreImgBuf::width ( void ) const
{
    Asd::Int32 width = 0;
    Asd::Int32 status = ipcore::w_pixrGetWidth (m_IPWState, m_pixr, &width);
    return (Asd::UInt32) width;
}

Asd::UInt32 IPCoreImgBuf::height ( void ) const
{
    Asd::Int32 height = 0;
    Asd::Int32 status = ipcore::w_pixrGetHeight (m_IPWState, m_pixr, &height);
    return (Asd::UInt32) height;
}

Asd::Point32 IPCoreImgBuf::size ( void ) const
{
    Asd::Int32 width = 0, height = 0;
    Asd::Int32 status = ipcore::w_pixrGetWidth (m_IPWState, m_pixr, &width);
    status = ipcore::w_pixrGetHeight (m_IPWState, m_pixr, &height);
    return Asd::Point32 (width, height);
}


// depth
ColorDepth IPCoreImgBuf::depth( void ) const
{
    Asd::Int32 depth = 0;
    Asd::Int32 status = ipcore::w_pixrGetDepth (m_IPWState, m_pixr, &depth);

    assert( (depth == 1) || (depth == 8) || (depth == 24) );

    return ColorDepth( depth );
}

// data
Asd::UInt8* IPCoreImgBuf::data( void )
{
    Asd::UInt8* retVal = NULL;
    Asd::Int32 status = ipcore::w_pixrGetImage (m_IPWState, m_pixr, &retVal);

    return retVal;
}

const Asd::UInt8* IPCoreImgBuf::data( void ) const
{
    Asd::UInt8* retVal = NULL;
    Asd::Int32 status = ipcore::w_pixrGetImage (m_IPWState, m_pixr, &retVal);

    return retVal;
}

Asd::UInt32 IPCoreImgBuf::dataSize( void ) const
{
    Asd::Int32 bytes = 0;
    Asd::Int32 status = ipcore::w_pixrImageBytes (m_IPWState, m_pixr, &bytes);
    return (Asd::UInt32) bytes;
}

Asd::UInt32 IPCoreImgBuf::bpl( void ) const
{
    Asd::Int32 bpl = 0;
    Asd::Int32 status = ipcore::w_pixrGetBpl (m_IPWState, m_pixr, &bpl);
    return (Asd::UInt32) bpl;
}

void IPCoreImgBuf::createState () 
{
#ifndef PSG
    m_IPWState = NULL;
    Asd::Int32 status = ipcore::w_IPWStateInit (NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, 0, NULL, NULL, &m_IPWState);

    return;
#else //IPCORE-PSG passes 17 arguments, ours only does 16.
    m_IPWState = NULL;
    Asd::Int32 status = ipcore::w_IPWStateInit (NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, 0, NULL, NULL, NULL, &m_IPWState);

    return;
#endif
}

bool IPCoreImgBuf::save(const char *fileName, const char *ext, const Asd::Int32 & xRes, const Asd::Int32 & yRes){ 
	if(!fileName) return false;
	if(!ext) return false;
	if(strncmp(ext, "tif", strlen(ext)) && strncmp(ext, "tiff", strlen(ext))){
		char* newname = new char [strlen(fileName) +5];
		sprintf(newname, "%s.tif", fileName);
		this->saveAsTiff (newname, xRes, yRes);
		delete [] newname;
	}else 
		this->saveAsTiff (fileName, xRes, yRes);
	return true;
	
}

void IPCoreImgBuf::saveAsTiff (const char* filename, const Asd::Int32 & xRes, const Asd::Int32 & yRes)
{
	Asd::Int32 frame;

	ipcore::w_pixrGetFrame (m_IPWState, m_pixr, &frame);
	if(frame == 0){
		ipcore::PIXR* framePixr = NULL;
		ipcore::w_addFrame(m_IPWState, m_pixr, &framePixr);
        ipcore::w_pixrSetResolutionDPI (m_IPWState, framePixr, Int32_TO_ScaledInt1616(xRes), Int32_TO_ScaledInt1616(yRes));

		if( depth() == MONOCHROME )
			Asd::Int32 status = ipcore::w_writeTiff (m_IPWState, framePixr, filename, COMPRESSION_CCITTFAX4 ,cTrue, 0);
		else
			Asd::Int32 status = ipcore::w_writeTiff (m_IPWState, framePixr, filename, COMPRESSION_LZW ,cTrue, 0);
		ipcore::w_destroyPixr(m_IPWState, &framePixr);

	}else{
		if( depth() == MONOCHROME )

			Asd::Int32 status = ipcore::w_writeTiff (m_IPWState, m_pixr, filename, COMPRESSION_CCITTFAX4 ,cTrue, 0);
		else
			Asd::Int32 status = ipcore::w_writeTiff (m_IPWState, m_pixr, filename, COMPRESSION_LZW ,cTrue, 0);
	}
    return;
}

void IPCoreImgBuf::flipBytes ()
{
    Asd::Int32 status = ipcore::w_flip (m_IPWState, m_pixr);
    return;
}


bool IPCoreImgBuf::scaleToSize (ImageAdapter* output)
{
    IPCoreImgBuf* out = dynamic_cast<IPCoreImgBuf*> (output);
    if (!out) return false;

    if (out->width() == 0 || out->height() == 0) return true;
    if (out->depth() != 8) return false;

    ipcore::PIXR* pixr = NULL, *framed = NULL;

    Asd::Int32 status = ipcore::w_addFrame (m_IPWState, m_pixr, &framed);
    
    if (depth() != 8)
    {
        status += ipcore::w_depth1To8 (m_IPWState, framed, cTrue, &pixr);
        status += ipcore::w_destroyPixr (m_IPWState, &framed);
        framed = pixr;
        pixr = NULL;
    }

    status += ipcore::w_scaleLinearSize (m_IPWState, framed, out->width(),
                out->height(), &pixr);
    status += ipcore::w_destroyPixr (m_IPWState, &framed);

    // copy the data into the output buffer
    ipcore::PIXR* ttempPixr = out->pixr ();

    Asd::Int32 ws, hs, depth, bpls, bpld;
    Asd::UInt8 *pImages, *pImaged;
    status += ipcore::w_pixrGetDepth(m_IPWState, pixr, &depth);
    status += ipcore::w_pixrFramedWidth(m_IPWState, pixr, &ws);
	status += ipcore::w_pixrFramedHeight(m_IPWState, pixr, &hs);
	status += ipcore::w_pixrGetImageAll(m_IPWState, pixr, &pImages);
	status += ipcore::w_pixrGetImage(m_IPWState, ttempPixr, &pImaged);
	status += ipcore::w_pixrGetBpl(m_IPWState, pixr, &bpls);
	status += ipcore::w_pixrGetBpl(m_IPWState, ttempPixr, &bpld);

	/* blit the inner part of the src, exclusive of cFrameBits,
	 * into the dest.  */
	status += ipcore::i_removeFrame (m_IPWState, pImages,
					 ws, hs, depth, bpls,
					 pImaged, bpld);


    status += ipcore::w_destroyPixr (m_IPWState, &pixr);

    if (status) return false;
    return true;
}
