
#include "stdafx.h"
#include <math.h>
#include "imageadapter.h"
#include "AsdRgb.h"

/*****************************************************************************
 * Used for the RGB->LAB conversion:
 ****************************************************************************/
#define N_BITS_F	12	/* Works best at 14, but costs extra time */
#define N_F		(1 << N_BITS_F)
#define KZERO_BITS	12	/* If you change, see srgbLUT[] */
#define KONE_BITS	14
#define KTWO_BITS	3
#define KZERO		(1 << KZERO_BITS)
#define KONE		(1 << KONE_BITS)
#define KTWO		(1 << KTWO_BITS)
#define EIGHTBITFROMINT(x)	(unsigned char)((x) < 0? 0: (x) > 255? 255: ((x)))
#define SIGNEDEIGHTBITFROMINT(x) (unsigned char)((x) < 0? (((x) < -128)? 128: 256 + (x)): (x) > 127? 127: ((x)))
#define INTLAB(x)	((x) >> (KONE_BITS + KTWO_BITS))
#define KONMO		(1 << (KZERO_BITS + KONE_BITS - N_BITS_F - 1))
#define f(x)		(ctab[((x) + KONMO) >> (KZERO_BITS + KONE_BITS - N_BITS_F)])

/*****************************************************************************
 * Used for the LAB->RGB conversion:
 ****************************************************************************/
#define N_BITS_F_INV	10
#define N_F_INV		(1 << N_BITS_F_INV)
#define KTHREE_BITS	22
#define KFOUR_BITS	14U
#define KFIVE_BITS	9U
#define KSIX_BITS	10  /* works best at 14, but costs extra time */
#define KTHREE		(1 << KTHREE_BITS)
#define KFOUR		(1 << KFOUR_BITS)
#define KFIVE		(1 << KFIVE_BITS)
#define KSIX		(1 << KSIX_BITS)

#define EIGHTBITRGB(x)	(unsigned char)((x) < 0? 0: (x) > (KFOUR * KFIVE * 255)? 255: \
			(x) >> (KFOUR_BITS + KFIVE_BITS))
/* If negative, use the first array element.
 * If greater than the maximum value less the added rounding factor, use the
 * last array element.
 * Otherwise, round, shift to KSIX elements, and use the LUT.
 */
#define ARRAY_MASK_INV	~(0xffffffff << (KTHREE_BITS))
#define f_inv(x)	(((x) < 0)? \
	ctab[0]: \
	((unsigned)(x) > (ARRAY_MASK_INV)? ctab[N_F_INV]: \
	ctab[(x) >> (KTHREE_BITS - N_BITS_F_INV)]))

/* If negative, use the first array element.
 * If greater than the maximum value, less the added rounding factor, use the
 * last array element.
 * Otherwise, round, and shift to KSIX elements, and use the LUT.
 */
#define KSMO		(1 << (KSIX_BITS - 1))
#define ARRAY_MASK_SRGB	~(0xffffffff << (KFOUR_BITS + KFIVE_BITS))
#define sRGBp2sRGB8(x)	(((x) < 0)? ptab[0]: \
	((unsigned)(x) > (ARRAY_MASK_SRGB - KSMO)? ptab[KSIX - 1]: \
	ptab[((x) + KSMO) >> (KFOUR_BITS + KFIVE_BITS - KSIX_BITS)]))


typedef struct LabRGBTables {
    Asd::Int32 *lab_rgb_tab;
    Asd::UInt8 *srp_sr8_tab;
} LabRGBTables;


static LabRGBTables *
lab_rgb_tables(void)
{
    Asd::Int32 i;
    float f;
    static Asd::Int32 lab_rgb_tab[N_F_INV+1];
    static Asd::UInt8 srp_sr8_tab[KSIX];
    static LabRGBTables tables = {lab_rgb_tab, srp_sr8_tab};
    static int initialized = 0;

    if (! initialized) {
	for (i = 0; i < N_F_INV+1; i++) {
	    f = (float)i / (float)(N_F_INV - 1);
	    if (f < 0.20689303f) {
		f = ((f) - 16.0f / 116.0f) / 7.787f;
	    } else {
		f = f * f * f;
	    }
	    lab_rgb_tab[i] = (Asd::Int32)(f * (float)KFOUR);
	}

	for (i = 0; i < KSIX; i++) {
	    f = (float)i / (float)(KSIX - 1);
	    if (f <= 0.00313) {
		f = f * 12.92f * 256.0f;
	    } else {
		f = ((float)pow((double)f, 1.0 / 2.4) * 1.055f - 0.055f) * 256.0f;
	    }
	    if (f > 255.0f) {
		srp_sr8_tab[i] = 255;
	    } else if (f < 0.0f) {
		srp_sr8_tab[i] = 0;
	    } else {
		srp_sr8_tab[i] = (Asd::UInt8)f;
	    }
	}
	initialized = 1;
    }

    return &tables;
}


void ImageAdapter::ITULab2RGB (Asd::Float64 Lmin, Asd::Float64 Lmax,
                     Asd::Float64 Amin, Asd::Float64 Amax,
                     Asd::Float64 Bmin, Asd::Float64 Bmax)
{
    Asd::UInt8* data;
    Asd::RgbPtr rgb;
    Asd::UInt32 row, col;
    Asd::UInt8 prevL, prevA, prevB;
    Asd::Int32 L, a, b;
    Asd::Int32 X, Y, Z, R, G, B;
    Asd::Int32 c, t;
    Asd::Int32 r1, r2, r3, g1, g2, g3, b1, b2, b3;
    const Asd::Float64 Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;	/* White point, D65 */
    Asd::Int32 const1, const2, const3, const4, const5, const6;
    LabRGBTables *tables;
    /* f_inv() LUT */
    Asd::Int32 *ctab;
    /* sRGB to 8-bit sRGB power function LUT */
    Asd::UInt8 *ptab;

    /* Precompute the power function */
    tables = lab_rgb_tables();
    ctab = tables->lab_rgb_tab;
    ptab = tables->srp_sr8_tab;

    /* Precompute the constants */
    const1 = (Asd::Int32)(((Lmax - Lmin) / (255.0f * 116.0f)) * KTHREE);
    const2 = (Asd::Int32)(((Lmin + 16.0f) / 116.0f) * KTHREE);
    const3 = (Asd::Int32)(((Amax - Amin) / (255.0f * 500.0f)) * KTHREE);
    const4 = (Asd::Int32)((Amin / 500.0f) * KTHREE);
    const5 = (Asd::Int32)(((Bmax - Bmin) / (255.0f * 200.0f)) * KTHREE);
    const6 = (Asd::Int32)((-Bmin / 200.0f) * KTHREE);

    r1 = (Asd::Int32)(( 3.2410f * Xn) * KFIVE);
    r2 = (Asd::Int32)((-1.5374f * Yn) * KFIVE);
    r3 = (Asd::Int32)((-0.4986f * Zn) * KFIVE);
    g1 = (Asd::Int32)((-0.9692f * Xn) * KFIVE);
    g2 = (Asd::Int32)(( 1.8760f * Yn) * KFIVE);
    g3 = (Asd::Int32)(( 0.0416f * Zn) * KFIVE);
    b1 = (Asd::Int32)(( 0.0556f * Xn) * KFIVE);
    b2 = (Asd::Int32)((-0.2040f * Yn) * KFIVE);
    b3 = (Asd::Int32)(( 1.0570f * Zn) * KFIVE);

    /* Convert the image, a pixel at a time */
    data = this->data();
    for (row = 0; row < this->height(); row++) 
    {
        // initialize out of gammut
        prevL = 255;
        prevA = 0;
        prevB = 0;
        rgb.reset (data + row * bpl());
    
        for (col = 0; col < this->width(); col++) 
        {

            // check if we have the same value as before
            if (prevL == rgb.r() && prevA == rgb.g() && prevB == rgb.b())
            {
                rgb.setRGB (B, G, R);
                ++rgb;
                continue;
            }

            // read Lab value from image
            prevL = rgb.r();
	    	prevA = rgb.g();
		    prevB = rgb.b();
		    L = (Asd::Int32) prevL;
    		a = (Asd::Int32) prevA;
	    	b = (Asd::Int32) prevB;

	        /* Convert from ITULAB -> XYZ: */
    	    c = const1 * L + const2;
	        t = const3 * a + const4 + c;
	        X = f_inv(t);
    	    Y = f_inv(c);
	        t = c - b * const5 + const6;
	        Z = f_inv(t);

    	    /* Note: real XYZ values are these divided by KFOUR (=2^14 =16384) */

	        /* Convert from XYZ to RGB */
	        R = r1 * X + r2 * Y + r3 * Z;
    	    G = g1 * X + g2 * Y + g3 * Z;
	        B = b1 * X + b2 * Y + b3 * Z;

	        /* Note: realy RGB values are these divided by KFOUR*KFIVE (=2^23 =8388608) */
    	    /* Convert from linear-RGB to sRGB' to 8-bit sRGB */
	        /* Rail and integerize to 8-bit */
            R = sRGBp2sRGB8(R);
            G = sRGBp2sRGB8(G);
            B = sRGBp2sRGB8(B);

            // overwrite it with BGR and advance to the next one
            rgb.setRGB (B, G, R);
            ++rgb;
        }
    }

    return;
}



