// ImageAdapter.cpp: implementation of the ImageAdapter class.
//
//////////////////////////////////////////////////////////////////////
#include "StdAfx.h"
#include "MiscConst.h"
#include "ImageAdapter.h"
#include "masktable.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

bool ImageAdapter::alphaBlend(ImageAdapter *selBuf, ImageAdapter *fgBuf){
    
    // first we need to get the data, bpl, with and height of all the buffers
    Asd::Int32 fgWidth = fgBuf->width();
    Asd::Int32 fgHeight = fgBuf->height();
    Asd::Int32 fgBpl = fgBuf->bpl();
    Asd::UInt8 * fgData = fgBuf->data();
    Asd::Int32 bgWidth = this->width();
    Asd::Int32 bgHeight = this->height();
    Asd::Int32 bgBpl = this->bpl();
    Asd::UInt8 * bgData = this->data();
    Asd::Int32 selWidth = selBuf->width();
    Asd::Int32 selHeight = selBuf->height();
    Asd::Int32 selBpl = selBuf->bpl();
    Asd::UInt8 *selData = (Asd::UInt8 *)selBuf->data();
    Asd::UInt8  * bgptr, *fgptr,*selptr;
    Asd::UInt32 row, col;
    register Asd::UInt8 bgsel, fgsel;
    
    for(row = 0; row < fgHeight; row++){
        bgptr  = bgData;
        selptr = selData;
        fgptr  = fgData;
        for(col = 0;  col < fgWidth; col += 4){
            // note ipcore selector is alwas big endian
            for(int j =3; j >=0; j--){
                bgsel = selptr[j];
                fgsel = 255 - bgsel;
                *bgptr = (*fgptr++ * fgsel + *bgptr * bgsel ) >> 8;
                bgptr ++;
                *bgptr = (*fgptr++ * fgsel + *bgptr * bgsel ) >> 8;
                bgptr ++;
                *bgptr = (*fgptr++ * fgsel + *bgptr * bgsel ) >> 8;
                bgptr ++;
            }
            selptr +=4;
        }
        bgData += bgBpl;
        selData += selBpl;
        fgData += fgBpl;
    }
    return true;
}

bool ImageAdapter::maskedCopy(ImageAdapter *selBuf, ImageAdapter *fgBuf)
{
    // We have a mask... 
    // fgBuf, selBuf, and bgBuf are all cropped to the right size, and same resolution, so we don't need to do cropping or zooming here
    
    
    // first we need to get the data, bpl, with and height of all the buffers
    Asd::Int32 fgWidth = fgBuf->width();
    Asd::Int32 fgHeight = fgBuf->height();
    Asd::Int32 fgBpl = fgBuf->bpl();
    Asd::UInt8 * fgData = fgBuf->data();
    Asd::Int32 bgWidth = this->width();
    Asd::Int32 bgHeight = this->height();
    Asd::Int32 bgBpl = this->bpl();
    Asd::UInt8 * bgData = this->data();
    Asd::Int32 selWidth = selBuf->width();
    Asd::Int32 selHeight = selBuf->height();
    Asd::Int32 selBpl = selBuf->bpl();
    Asd::UInt32 *selData = (Asd::UInt32 *)selBuf->data();
    
    if(fgWidth != bgWidth || fgWidth != selWidth) return false;
    if(fgHeight!= bgHeight|| fgHeight != selHeight) return false;
    
    // selector layer DWORD per line
    Asd::UInt32 maskWpl = selBpl / sizeof(Asd::UInt32);
    
    register int selCol, row = 0;  
    register Asd::UInt32 maskvalue;
    register Asd::UInt8  *fgptr, *bgptr;
    register Asd::UInt32 *selptr;
    register Asd::UInt32 wStart, shift;
    
    for(row=0; row < fgHeight; row++){
        bgptr  = bgData ;
        selptr = selData ;
        fgptr  = fgData;
        for(selCol = 0; selCol < selWidth; selCol += 4){
            // get maskvalue in unit of 32 bit
            // We assume a 1-bpp decompressed mask in the same orientation as the jpeg.
            // The masks coming out of the print driver seem to be big-endian by construction.
            shift  = 28 - selCol % 32;
            wStart = selCol / 32;
            maskvalue = (selptr[wStart] >> shift) & 0x0f;
            mask_pixel_assign1[maskvalue](bgptr, fgptr); 
            bgptr += 4 * RGB_PIXELSIZE;
            fgptr += 4 * RGB_PIXELSIZE;
        } //end of for( selCol...)
        bgData  += bgBpl;
        selData += maskWpl;
        fgData  += fgBpl;
    }//end of for row =...
    return true;
}

