
#ifndef ImageAdapter_h_Included
#define ImageAdapter_h_Included

#include "ImageBuf.h"

class ImageAdapter
{
public:
	
    ImageAdapter( ImageBuf* pBuf ) { m_piBuf = pBuf; }
    virtual ~ImageAdapter () {}

    virtual Asd::UInt32  width ( void ) const { return m_piBuf->width(); }
	virtual Asd::UInt32  height( void ) const { return m_piBuf->height(); }
    
    Asd::Point32 size( void ) const {return Asd::Point32(width(), height());}

	virtual ColorDepth depth( void ) const { return m_piBuf->depth(); }

    /* Allows us to access the data.  Remember, the data in an ImageBuf */
    /* is always an flat, uncompressed array. */
    virtual Asd::UInt8 * data( void ) { return m_piBuf->data(); }
	virtual const Asd::UInt8 * data( void ) const { return m_piBuf->data(); }

    /* The number of bytes one needs to advance a pointer to reach the */
    /* same column in the next row. */
    virtual Asd::UInt32 bpl( void ) const { return m_piBuf->bpl(); }

    ImageBuf* getIBuf () { return m_piBuf; }

    /* some image processing */
    virtual bool scaleUp (const Asd::Point32& scaleFac, ImageAdapter* output) { 
        return false; 
    }
    virtual bool scaleToSize (ImageAdapter* output) { 
        return false; 
    }
    virtual void flipBytes() {return;}

    /* copy a masked image on top of the buffer */
    virtual bool maskedCopy (ImageAdapter* mask, ImageAdapter* fg);

    /* alpha blend a fg image onto the buffer */
    virtual bool alphaBlend (ImageAdapter* mask, ImageAdapter* fg);
    virtual bool save(const char *outname, const char * ext, const Asd::Int32 & xRes, const Asd::Int32 & yRes) = 0;

    /* color space conversion, initialize variables to default ITULAB settings */
    void ITULab2RGB (Asd::Float64 Lmin = 0., Asd::Float64 Lmax = 100.,
                     Asd::Float64 Amin = -85.33333, Asd::Float64 Amax = 84.66666,
                     Asd::Float64 Bmin = -75.29412, Asd::Float64 Bmax = 124.70588);

    /* convert to BW for printing */
    virtual void convertToBW (ImageBuf* inBuf) {}

    void fillWhite () {memset ((void*) m_piBuf->data(), 0xFF, m_piBuf->dataSize()); }
    void fillBlack () {memset ((void*) m_piBuf->data(), 0x00, m_piBuf->dataSize()); }

protected:

    ImageBuf *m_piBuf;

};

#endif
/*#ifndef ImageAdapter_h_Included */
