#include "StdAfx.h"

#include "ImageBuf.h"
#include "AsdRgb.h"
#include "OddPtr.h"

using namespace Asd;

ImageBuf::~ImageBuf( void )
{
}

ImageBuf::ImageBuf( void )
:
 m_align( sizeof(UInt32) ) // DIB are 32
{
}


UInt8 * ImageBuf::data( void )
{
    return &(m_data[0]);
}

const UInt8 * ImageBuf::data( void ) const
{
    ImageBuf * me = const_cast<ImageBuf *>(this);

    return &(me->m_data[0]);
}

void ImageBuf::clear( void )
{
    this->resize( Point32(0,0), this->depth(), this->m_align );
}

void ImageBuf::copyFrom( const ImageBuf & from )
{
	this->resize( from.size(), from.depth(), from.m_align );

	assert( this->dataSize() == from.dataSize() );

	memcpy( this->data(), from.data(), this->dataSize() );
}


void ImageBuf::
copyRotate( Orientation orient, ImageBuf * buf, const Rect & region )
{
    assert( buf );

    if( buf && OrientIsValid(orient) )
    {
        if( orient == Orient_0 )
        {
            assert( NULL );
        }
        else
        {
            switch( this->depth() )
            {
                case MONOCHROME:
                {
                    copyRotate1Bit( orient, buf, region );
                }
                break;

                case COLOR_8_BIT:
                {
                    copyRotate8Bit( orient, buf, region );
                }
                break;

                case COLOR_24_BIT:
                {
                    copyRotate24Bit( orient, buf, region );
                }
                break;

                default: assert( NULL ); break;
            }
        }
    }
}

void ImageBuf::
copyRotate1Bit( Orientation orient, ImageBuf * buf, const Rect & region )
{
    assert( buf );
    assert( orient != Orient_0     );
    assert( OrientIsValid(orient) );
    assert( this->depth() == MONOCHROME );

    // If we are not in debug mode, we don't hit this assertions, so we
    // should just return if buf is NULL or we do not have 24-bit color.
    
    if( !buf ) { return; }
    
    if( this->depth() != MONOCHROME ) { return; }

    if( orient == Orient_180 )
    {
        buf->resize( region.size(), this->depth(), this->m_align );

        for( Int32 row = region.top(); row <= region.bottom(); ++row )
        {
            BitPtr srcPtr( this->data() + (this->bpl() * row) );
            srcPtr += region.left();

            BitPtr dstPtr( buf->data() + (buf->bpl() * (buf->height() - 1 - row + region.top())) );
            dstPtr += (buf->width() - 1);

            for( Int32 col = region.left(); col <= region.right(); ++col )
            {
                dstPtr.set( *srcPtr );
                --dstPtr;
                ++srcPtr;
            }
        }
    }
    else if( orient == Orient_270 )
    {
        const Point32 bufSize( region.size().getY(), region.size().getX() );

        buf->resize( bufSize, this->depth(), this->m_align );

        for( Int32 row = region.top(); row <= region.bottom(); ++row )
        {
            BitPtr srcPtr( this->data() + (this->bpl() * row) );
            srcPtr += region.right();

            BitPtr dstPtr( buf->data() );
            dstPtr += (row - region.top());

            for( Int32 col = region.left(); col <= region.right(); ++col )
            {
                dstPtr.set( *srcPtr );
                dstPtr.offset( buf->bpl() );
                --srcPtr;
            }
        }
    }
    else if( orient == Orient_90 )
    {
        const Point32 bufSize( region.size().getY(), region.size().getX() );

        buf->resize( bufSize, this->depth(), this->m_align );

        for( Int32 row = region.top(); row <= region.bottom(); ++row )
        {
            BitPtr srcPtr( this->data() + (this->bpl() * row) );
            srcPtr += region.left();

            BitPtr dstPtr( buf->data() );
            dstPtr += region.bottom() - row;

            for( Int32 col = region.left(); col <= region.right(); ++col )
            {
                dstPtr.set( *srcPtr );
                dstPtr.offset( buf->bpl() );
                ++srcPtr;
            }
        }
    }
}


void ImageBuf::
copyRotate8Bit( Orientation orient, ImageBuf * buf, const Rect & region )
{
    assert( buf );
    assert( orient != Orient_0     );
    assert( OrientIsValid(orient) );
    assert( this->depth() == COLOR_8_BIT );

    // If we are not in debug mode, we don't hit this assertions, so we
    // should just return if buf is NULL or we do not have 24-bit color.
    
    if( !buf ) { return; }
    
    if( this->depth() != COLOR_8_BIT ) { return; }

    if( orient == Orient_180 )
    {
        buf->resize( region.size(), this->depth(), this->m_align );

        for( Int32 row = region.top(); row <= region.bottom(); ++row )
        {
            UInt8 * srcPtr = this->data() + (this->bpl() * row);
            srcPtr += region.left();

            UInt8 * dstPtr = buf->data() + (buf->bpl() * (buf->height() - 1 - row + region.top()) );
            dstPtr += (buf->width() - 1);

            for( Int32 col = region.left(); col <= region.right(); ++col )
            {
                *dstPtr = *srcPtr;
                --dstPtr;
                ++srcPtr;
            }
        }
    }
    else if( orient == Orient_270 )
    {
        const Point32 bufSize( region.size().getY(), region.size().getX() );

        buf->resize( bufSize, this->depth(), this->m_align );

        for( Int32 row = region.top(); row <= region.bottom(); ++row )
        {
            UInt8 * srcPtr = this->data() + (this->bpl() * row);
            srcPtr += region.right();

            UInt8 * dstPtr = buf->data();
            dstPtr += (row - region.top());

            for( Int32 col = region.left(); col <= region.right(); ++col )
            {
                *dstPtr = *srcPtr;
                dstPtr += buf->bpl();
                --srcPtr;
            }
        }
    }
    else if( orient == Orient_90 )
    {
        const Point32 bufSize( region.size().getY(), region.size().getX() );

        buf->resize( bufSize, this->depth(), this->m_align );

        for( Int32 row = region.top(); row <= region.bottom(); ++row )
        {
            UInt8 * srcPtr = this->data() + (this->bpl() * row);
            srcPtr += region.left();

            UInt8 * dstPtr = buf->data();
            dstPtr += region.bottom() - row;

            for( Int32 col = region.left(); col <= region.right(); ++col )
            {
                *dstPtr = *srcPtr;
                dstPtr += buf->bpl();
                ++srcPtr;
            }
        }
    }
}


void ImageBuf::
copyRotate24Bit( Orientation orient, ImageBuf * buf, const Rect & region )
{
    assert( buf );
    assert( OrientIsValid(orient) );
    assert( orient != Orient_0 );
    assert( this->depth() == COLOR_24_BIT );

    // If we are not in debug mode, we don't hit this assertions, so we
    // should just return if buf is NULL or we do not have 24-bit color.
    
    if( !buf ) { return; }
    
    if( this->depth() != COLOR_24_BIT ) { return; }

    if( orient == Orient_180 )
    {
        buf->resize( region.size(), this->depth(), this->m_align );

        for( Int32 row = region.top(); row <= region.bottom(); ++row )
        {
            RgbPtr srcPtr( this->data() + (this->bpl() * row) );
            srcPtr += region.left();

            RgbPtr dstPtr( buf->data() + (buf->bpl() * (buf->height() - 1 - row + region.top()) ) );
            dstPtr += (buf->width() - 1);


            for( Int32 col = region.left(); col <= region.right(); ++col )
            {
                dstPtr.copyDataFrom( srcPtr );
                --dstPtr;
                ++srcPtr;
            }
        }
    }
    else if( orient == Orient_270 )
    {
        const Point32 bufSize( region.size().getY(), region.size().getX() );

        buf->resize( bufSize, this->depth(), this->m_align );

        for( Int32 row = region.top(); row <= region.bottom(); ++row )
        {
            RgbPtr srcPtr( this->data() + (this->bpl() * row) );
            srcPtr += region.right();

            RgbPtr dstPtr( buf->data() );
            dstPtr += (row - region.top());

            for( Int32 col = region.left(); col <= region.right(); ++col )
            {
                dstPtr.copyDataFrom( srcPtr );
                dstPtr.offset( buf->bpl() );
                --srcPtr;
            }
        }
    }
    else if( orient == Orient_90 )
    {
        const Point32 bufSize( region.size().getY(), region.size().getX() );

        buf->resize( bufSize, this->depth(), this->m_align );

        for( Int32 row = region.top(); row <= region.bottom(); ++row )
        {
            RgbPtr srcPtr( this->data() + (this->bpl() * row) );
            srcPtr += region.left();

            RgbPtr dstPtr( buf->data() );
            dstPtr += region.bottom() - row;

            for( Int32 col = region.left(); col <= region.right(); ++col )
            {
                dstPtr.copyDataFrom( srcPtr );
                dstPtr.offset( buf->bpl() );
                ++srcPtr;
            }
        }
    }
}
