#ifndef __ImageBuf_h_Included__
#define __ImageBuf_h_Included__

/* This doen because some of the headers we use unwisely define min */
/* and max. */

#ifdef max
#undef max
#endif

#ifdef min
#undef min
#endif

#include <valarray>
#include "MiscConst.h"
#include "AsdRect.h"

/*

  ImageBuf is an absrtact base class represennting an uncompressed,
  2D raster image.

  resize() must be called to make sure that the buffer is properly
  allocated.

  Once allocated, image data is maniupulated throught he data poitner
  obtained by calling the data() method.  The pointer returned by data
  points to the upper-left corner of the image.  bpl() returns the number
  of bytes required to advance a pointer to the same pixel in the
  following row.  The value returned by bpl() may not be equal to the
  product of width() and depth() because of alignment issues.

*/

class ImageBuf
{
public:
	
    ImageBuf( void );
    virtual ~ImageBuf();

    /* width(), height(), and size() are all in pixels.  These */
    /* values say nothing about total memory needed for which we */
    /* need to know the byte alignment and bits per pixel.  If the */
    /* product of width() and depth() does not lie on an alignment */
    /* boundary, for example, width() will differ from bpl(). */

    virtual Asd::UInt32  width ( void ) const = 0;
	virtual Asd::UInt32  height( void ) const = 0;
    
    Asd::Point32 size( void ) const {return Asd::Point32(width(), height());}

	virtual ColorDepth depth( void ) const = 0;

    /* Allows us to access the data.  Remember, the data in an ImageBuf */
    /* is always an flat, uncompressed array. */
    virtual       Asd::UInt8 * data( void );
	virtual const Asd::UInt8 * data( void ) const;

    /* The number of bytes one needs to advance a pointer to reach the */
    /* same column in the next row. */
    Asd::UInt32 bpl( void ) const { if (height() > 0) return dataSize() / height();
                                    return 0; }

    /* The total amount of memory needed.  I think this will always be */
    /* the product of bpl() and height(), but I am not 100% sure.  If */
    /* I ever become sure, this will not have to be virtual. */
    Asd::UInt32 dataSize( void ) const { return m_data.size(); }

    /* Allocate or reallocate the buffer for an image with these dimensions */
    /* and padding.  The internal bookkeeping variables will obviosly always */
    /* be updated, but implementations of this method should strive to avoid */
    /* unnecessary memory reallocation.  That is, no matter what the new */
    /* values of size, color depth, and alignment are, if dataSize() does */
    /* not change, no memory reallocation should occur. */
    /* */
    /* Note, however, that regardless of attempts to avoid unneeded memory */
    /* reallocation, calling resize invalidates any old copies of the data */
    /* pointer.  Should the data pointer "accidentally" remain unchanged, */
    /* the data to which it points is undefined. */
    virtual void resize( const Asd::Point32 & size, ColorDepth, 
        Asd::UInt32 align ) = 0;
    
	void clear( void );

	void copyFrom( const ImageBuf & );

    void copyRotate( Orientation, ImageBuf * buf, const Asd::Rect & region );

protected:

    std::valarray<Asd::UInt8> m_data;
    Asd::UInt32 m_align;

private:

    void copyRotate1Bit( Orientation, ImageBuf * buf, const Asd::Rect & region );
    void copyRotate8Bit( Orientation, ImageBuf * buf, const Asd::Rect & region );
    void copyRotate24Bit( Orientation, ImageBuf * buf, const Asd::Rect & region );
};


#endif
/*#ifndef ImageBuf_h_Included */
