#include "StdAfx.h"

#include <stdio.h>

#include "IntelImgBuf.h"
#include "IPCoreImgBuf.h"
#include "DibImgBuf.h"
#include <stdio.h>

// Constructors
/*

  iplCreateImageHeader(int nChannels,
int alphaChannel, int depth, char* colorModel,
char* channelSeq, int dataOrder, int origin, int align,
int width, int height, IplROI* roi, IplImage* maskROI,
void* imageID, IplTileInfo* tileInfo);

*/

 
// empty image
IntelImgBuf::IntelImgBuf () : ImageAdapter(NULL)
{ 
    // set size to (1,1), because intel crashes on (0,0)
	m_image = intel::iplCreateImageHeader( 1, 0, IPL_DEPTH_8U, "RGB",
		"BGR", IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_QWORD, 
		1, 1, NULL, NULL, 
		NULL, NULL); 
}

// default values (TopLeft coords, RGB, white) for a given size
IntelImgBuf::IntelImgBuf (Asd::UInt32 width, Asd::UInt32 height, ColorDepth depth, ImageBuf* pBuf) :
ImageAdapter(pBuf)
{ 

    Asd::UInt32 nChannel = 1;
    if (depth == COLOR_24_BIT) nChannel = 3;

    // set size to (1,1), because intel crashes on (0,0)
    if (height == 0) height = 1;
    if (width == 0) width = 1;

    // align buf to 64 bits
    m_piBuf->resize (Asd::Point32(width, height), depth, 8); 

    if (nChannel == 3)
    	m_image = intel::iplCreateImageHeader( 3, 0, IPL_DEPTH_8U, "RGB",
		"BGR", IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_QWORD, 
		width, height, NULL, NULL, 
		NULL, NULL); 
    if (nChannel == 1)
    {
        if (depth == COLOR_8_BIT)
    	m_image = intel::iplCreateImageHeader( 1, 0, IPL_DEPTH_8U, "Gray",
		"Gray", IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_QWORD, 
		width, height, NULL, NULL, 
		NULL, NULL); 
        else
    	m_image = intel::iplCreateImageHeader( 1, 0, IPL_DEPTH_1U, "Gray",
		"Gray", IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_QWORD, 
		width, height, NULL, NULL, 
		NULL, NULL); 
    }

    m_image->imageData = reinterpret_cast<char*> (m_piBuf->data());
    m_image->imageDataOrigin = reinterpret_cast<char*> (m_piBuf->data());
    m_image->widthStep = m_piBuf->bpl();
    m_image->imageSize =  m_piBuf->dataSize();
}
	
// copy header
IntelImgBuf::IntelImgBuf (Asd::Rect & roi, ImageAdapter & buf) : ImageAdapter (buf.getIBuf())
{
    Asd::UInt32 nChannel = 1;
    if (buf.depth() == COLOR_24_BIT) nChannel = 3;

    Asd::UInt32 width = roi.width();
    Asd::UInt32 height = roi.height();
    // set size to (1,1), because intel crashes on (0,0)
    if (height == 0) height = 1;
    if (width == 0) width = 1;

    if (nChannel == 3)
    	m_image = intel::iplCreateImageHeader( 3, 0, IPL_DEPTH_8U, "RGB",
		"BGR", IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_QWORD, 
		width, height, NULL, NULL, NULL, NULL); 
    if (nChannel == 1)
    {
        if (buf.depth() == COLOR_8_BIT)
    	m_image = intel::iplCreateImageHeader( 1, 0, IPL_DEPTH_8U, "Gray",
		"Gray", IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_QWORD, 
		width, height, NULL, NULL, 
		NULL, NULL); 
        else
    	m_image = intel::iplCreateImageHeader( 1, 0, IPL_DEPTH_1U, "Gray",
		"Gray", IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_QWORD, 
		width, height, NULL, NULL, 
		NULL, NULL); 
    }

    m_image->imageData = reinterpret_cast<char*> (buf.data() + roi.top() * buf.bpl() + roi.left() * nChannel);
    m_image->imageDataOrigin = m_image->imageData;
    m_image->widthStep = buf.bpl();
}


// Destructor
IntelImgBuf::~IntelImgBuf () 
{ 
    invalidateData();
    intel::iplDeallocate(m_image, IPL_IMAGE_HEADER); 
}

bool IntelImgBuf::save(const char *fileName, const char * ext, const Asd::Int32 & xRes, const Asd::Int32 & yRes)
{
	if(!fileName) return false;
	if(!ext) return false;
	if(strncmp(ext, "pnm", strlen(ext))){
		char* newname = new char [strlen(fileName) +5];
		sprintf(newname, "%s.pnm", fileName);
		this->saveAsPnm (newname);
		delete [] newname;
	}else 
		this->saveAsPnm (fileName);
	return true;
}
void IntelImgBuf::saveAsPnm (const char* fileName)
{
    FILE* fp = fopen (fileName, "wb");

    // header
    fprintf (fp, "P6\n");
    fprintf (fp, "%d %d\n", width(), height());
    fprintf (fp, "255\n");

    Asd::UInt8* buffer = data();

    // data
    for (int i = 0; i < height(); i++)
    {
        fwrite (buffer, 3, width(), fp);
        buffer += bpl();
    }

    fclose (fp);
}

bool IntelImgBuf::scaleUp (const Asd::Point32& scaleFac, ImageAdapter* output) 
{ 

    IntelImgBuf* large = dynamic_cast<IntelImgBuf*> (output);
    if (!large) return FALSE;

    intel::iplZoom(ipl(), large->ipl (), scaleFac.getX(), 1, 
        scaleFac.getY(), 1, IPL_INTER_NN);
    
    return TRUE; 
}

bool IntelImgBuf::scaleToSize (ImageAdapter* output) 
{ 

    IntelImgBuf* out = dynamic_cast<IntelImgBuf*> (output);
    if (!out) return FALSE;

    intel::iplResizeFit( ipl(), out->ipl(), IPL_INTER_LINEAR );
    
    return TRUE; 
}

// copy a masked image on top of the buffer
bool IntelImgBuf::maskedCopy (ImageAdapter* mask, ImageAdapter* fgBuf) 
{ 
    IntelImgBuf* pFG;

    IPCoreImgBuf* ipcore = dynamic_cast<IPCoreImgBuf*> (mask);        
    if (!ipcore) return FALSE;
    ipcore->flipBytes ();

    // convert from an IPCore adapter to Intel
    Asd::Rect bandRoi(ipcore->width(), ipcore->height());
    IntelImgBuf selIntel (bandRoi, *ipcore);

    // use the selector as a mask
    setMask (selIntel.ipl());

    pFG = dynamic_cast<IntelImgBuf*> (fgBuf);
        
    intel::iplCopy (pFG->ipl(), ipl ());

    // clear the mask
    clearMask ();
    return TRUE; 
}

    // alpha blend a fg image onto the buffer
bool IntelImgBuf::alphaBlend (ImageAdapter* mask, ImageAdapter* fg) 
{ 
    IntelImgBuf* pFG;
    
    IPCoreImgBuf* ipcore = dynamic_cast<IPCoreImgBuf*> (mask);
    if (!ipcore) return FALSE;
    ipcore->flipBytes ();

    // convert from an IPCore adapter to Intel
    Asd::Rect bandRoi(ipcore->width(), ipcore->height());    
    IntelImgBuf selIntel (bandRoi, *ipcore);
    
    pFG = dynamic_cast<IntelImgBuf*> (fg);
    if (!pFG) return FALSE;
    // alpha blend together foreground and background 
    intel::iplAlphaComposite(ipl(), pFG->ipl(), ipl(), IPL_COMPOSITE_PLUS,    
        selIntel.ipl(), NULL, NULL, FALSE, FALSE);

    return TRUE; 
}

void IntelImgBuf::convertToBW (ImageBuf* inBuf)
{

    const Asd::Int32 noise = 0; // in percent

    // check if we need to do anything
    if (depth() == MONOCHROME) return;

    // wrap an intel header around it
    IntelImgBuf bw (width(), height(), MONOCHROME, inBuf);

    // create a mask buffer
    DibImgBuf maskBuf;
    IntelImgBuf tempMask (width(), height(), MONOCHROME, &maskBuf);

    if (depth() == COLOR_24_BIT)
    {

        // create a temporary gray buffer
        DibImgBuf tempBuf;
        IntelImgBuf tempGray (width(), height(), COLOR_8_BIT, &tempBuf);

        // convert to gray
        intel::iplColorToGray(ipl(), tempGray.ipl());

        // convert to BW
        intel::iplReduceBits(tempGray.ipl(), bw.ipl(), noise, IPL_DITHER_FS, 2);
        
        // generate a mask for all white pixels
        intel::iplThreshold(tempGray.ipl(), tempMask.ipl(), 254);
    }

    if (depth() == COLOR_8_BIT)
    {
        // convert to BW
        intel::iplReduceBits(ipl(), bw.ipl(), noise, IPL_DITHER_FS, 2);

        // generate a mask
        intel::iplThreshold(ipl(), tempMask.ipl(), 254);
    }

    // apply mask
    intel::iplOr(bw.ipl(), tempMask.ipl(), bw.ipl());
    // switch to our convention
    intel::iplNot(bw.ipl(), bw.ipl());

    return;
}
