#ifndef IntelImgBuf_h_Included
#define IntelImgBuf_h_Included

#include "ImageAdapter.h"

/* INTEL imaging library */
namespace intel
{
#include "ipl.h"
}

/* Image Object represented by Intel Imaging Library */
class IntelImgBuf : public ImageAdapter
{
public:
	
	/* Constructors */
	/* empty image */
	IntelImgBuf ();
	/* wrap an intel header around an existing ImageBuf */
	IntelImgBuf (Asd::UInt32 width, Asd::UInt32 height, ColorDepth depth, ImageBuf* pBuf);
    /* create a new header for the region specified by roi */
    IntelImgBuf (Asd::Rect & roi, ImageAdapter & buf);
	/* Destructor */
	~IntelImgBuf ();

	/* dimensions */
	virtual Asd::UInt32 width( void ) const
    {
        return Asd::UInt32(m_image->width);
    }

	virtual Asd::UInt32 height( void ) const
    {
        return Asd::UInt32(m_image->height);
    }
	
    virtual Asd::Point32 size( void ) const
    {
        return Asd::Point32 (m_image->width, m_image->height);
    }

	/* depth */
    virtual ColorDepth depth( void ) const
	{ 
		if (m_image->nChannels == 3 &&
			m_image->depth == IPL_DEPTH_8U)
			return  COLOR_24_BIT;

		if (m_image->nChannels == 1 && 
			m_image->depth == IPL_DEPTH_1U)
			return MONOCHROME;

		if (m_image->nChannels == 1 && 
			m_image->depth == IPL_DEPTH_8U)
			return COLOR_8_BIT;

        return COLOR_DEPTH_ERROR;
	}

	/* data */
    virtual  Asd::UInt8 * data( void )
    {
        /* Does this need a cast at all? */
        return reinterpret_cast<Asd::UInt8*>(m_image->imageData);
    }
	
    virtual const Asd::UInt8 * data( void ) const
    {
        /* Does this need a cast at all? */
        return reinterpret_cast<const Asd::UInt8*>(m_image->imageData);
    }

    virtual Asd::UInt32 dataSize( void ) const
    {
        return Asd::UInt32(m_image->imageSize);
    }

	virtual Asd::UInt32 bpl( void ) const
    {
        return Asd::UInt32(m_image->widthStep);
    }

    /* some image processing */
    virtual bool scaleUp (const Asd::Point32& scaleFac, ImageAdapter* output);
    virtual bool scaleToSize (ImageAdapter* output);

    /* copy a masked image on top of the buffer */
    virtual bool maskedCopy (ImageAdapter* mask, ImageAdapter* fg);

    /* alpha blend a fg image onto the buffer */
    virtual bool alphaBlend (ImageAdapter* mask, ImageAdapter* fg);
    void saveAsPnm (const char* fileName);

    virtual bool save(const char *fileName, const char * ext,const Asd::Int32 & xRes, const Asd::Int32 & yRes);

    intel::IplImage* ipl() { return m_image;}

    void offsetDataPtr (const Asd::UInt32 offset) { m_image->imageData += offset; }

    void setMask (intel::IplImage* mask) {m_image->maskROI = mask;}
    void clearMask () {m_image->maskROI = NULL;}

    virtual void convertToBW (ImageBuf* inBuf);

protected:

    void invalidateData ()
    {
        m_image->imageData = NULL;
        m_image->imageDataOrigin = NULL;
        m_image->widthStep = 0;
        m_image->imageSize =  0;
    }

private:

	/* actual data */
	intel::IplImage* m_image;

};


#endif
/*#ifndef IntelImgBuf_h_Included */
