#include "StdAfx.h"

// SilxImgBuf.cpp: implementation of the SilxImgBuf class.
//
//////////////////////////////////////////////////////////////////////

#include "SilxImgBuf.h"

#include "AsdSilx.h" 
namespace AsdSilx
{
#include "scaleBlit.h"
#include "resolution_t.h"
#include "dgp_err_printers.h"
};
using namespace AsdSilx;
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// empty image
SilxImgBuf::SilxImgBuf () : ImageAdapter(NULL), dgp_state(NULL), m_image(NULL)
{ 
        Dgp_initDgpState((Dgp_State **) &dgp_state, 
                                NULL,
                                NULL,
                                stderr, 
                                Dgp_fileErrPrinter,
                                DGP_FALSE, 
                                DGP_FALSE);
		DgpImage_newAligned ((Dgp_State *)dgp_state, DGPIMAGE_RGB, 0, 0, &m_image);
}

// Only copy header
SilxImgBuf::SilxImgBuf (const Asd::Rect & roi, ImageAdapter & buf) : 
ImageAdapter (buf.getIBuf()), dgp_state(NULL), m_image(NULL)
{
    DgpImage_Type type;
    Dgp_UInt32 ret_val;
    void * image;
    ret_val = Dgp_initDgpState((Dgp_State **)&dgp_state, 
                                NULL,
                                NULL,
                                stderr, 
                                Dgp_fileErrPrinter,
                                DGP_FALSE, 
                                DGP_FALSE);
    if(ret_val != cSuccessful) dgp_state = NULL;

    //convert colorDepth to image type;
    Asd::UInt32 * data = NULL;
    switch(buf.depth()){
    case   MONOCHROME :
        type = DGPIMAGE_PACKED;
        data = (Asd::UInt32 *) (buf.data() + roi.top() * buf.bpl() + roi.left() / 8);
        break;
    case COLOR_24_BIT:
        type = DGPIMAGE_RGB;
        data = (Asd::UInt32 *) (buf.data() + roi.top() * buf.bpl() + roi.left() * 3);
        break;
    case COLOR_8_BIT:
        type = DGPIMAGE_GRAY;
        data = (Asd::UInt32 *)(buf.data() + roi.top() * buf.bpl() + roi.left());
        break;
    default:
        type = 0;
    }
    if( type > 0 && dgp_state) {
        ret_val = DgpImage_wrapAll((Dgp_State *)dgp_state, data, type, roi.width(), roi.height(), buf.bpl(), &image);
        if(ret_val == cSuccessful) {
            m_image = image;
        }else m_image = NULL;
    }else m_image = NULL;
}
// Convert Any ImageBuf to SilxImgBuf
SilxImgBuf::SilxImgBuf (Asd::UInt32 width, Asd::UInt32 height, ColorDepth colorDepth, ImageBuf* pBuf) : 
ImageAdapter(pBuf), dgp_state(NULL), m_image(NULL)
{ 
    DgpImage_Type type;
    Dgp_UInt32 ret_val, diWidth = 0;
    void * image;
    // create dgp_state
    ret_val = Dgp_initDgpState((Dgp_State **)&dgp_state, 
                                NULL,
                                NULL,
                                stderr, 
                                Dgp_fileErrPrinter,
                                DGP_FALSE, 
                                DGP_FALSE);
    if(ret_val != cSuccessful) dgp_state = NULL;

    //convert colorDepth to dgp image type;
    switch(colorDepth){
    case   MONOCHROME :
        type = DGPIMAGE_PACKED;
        diWidth = (width + 31 ) / 32 * 32; // 32 bit aligned byte perline
        break;
    case COLOR_24_BIT:
        type = DGPIMAGE_RGB;
        diWidth = (width + 3)/ 4 * 4; // 4 byte aligned
        break;
    case COLOR_8_BIT:
        type = DGPIMAGE_GRAY;
        diWidth = (width + 3) / 4 * 4;
        break;
    default:
        type = 0;
    }

    if( type > 0 && dgp_state) {
        // Ask the image buffer to resize if it is not already 32 bit aligned
        m_piBuf->resize (Asd::Point32(diWidth, height), colorDepth, 4); // 4 means 4 bytes, 32 bit align       
        // wrap a DgpImage header arround the buffer data
        ret_val = DgpImage_wrapAll((Dgp_State *)dgp_state, (Asd::UInt32 *)m_piBuf->data(), type, width, height, m_piBuf->bpl(), &image);
        if(ret_val == cSuccessful) {
            m_image = image;
        }else m_image = NULL;
    }else m_image = NULL;

}

// Destructor
SilxImgBuf::~SilxImgBuf () 
{ 
    if(m_image){
            Asd::UInt32 *retval;
            DgpImage_unwrapAll (m_image, &retval);
    }
    m_image = NULL;
    if(dgp_state) Dgp_cleanupDgpState((Dgp_State **)(&dgp_state));
    dgp_state = NULL;

}


// copy
void * SilxImgBuf::copyImage () 
{ 
    void * image = NULL;
    if(m_image)	DgpImage_dup (m_image, DGP_TRUE, &image);
    return(image);

}


bool SilxImgBuf::save(const char *fileName, const char * ext,const Asd::Int32 & xRes, const Int32 & yRes)

{

	if(!fileName) return false;
	if(!ext) return false;
	if(strncmp(ext, "pnm", strlen(ext))){
		char* newname = new char [strlen(fileName) +5];
		sprintf(newname, "%s.pnm", fileName);
		this->saveAsPnm (newname);
		delete [] newname;
	}else 
		this->saveAsPnm (fileName);
	return true;
}



void SilxImgBuf::saveAsPnm (const char* fileName)
{
    char name[100];
    strncpy(name, fileName, strlen(fileName) + 1);
    if(m_image) DgpImage_save(m_image, name);
}

// dimensions, cast the image to it's lowest class DgpImage_Packed
Asd::UInt32 SilxImgBuf::width( void ) const
{
    if(m_image)    
        return Asd::UInt32(DGPIMG_WIDTH((DgpImage_Packed)m_image));
    else return 0;
}

Asd::UInt32 SilxImgBuf::height( void ) const
{
    if(m_image)
        return Asd::UInt32(DGPIMG_HEIGHT((DgpImage_Packed)m_image));
    else return 0;
}

Asd::Point32 SilxImgBuf::size( void ) const
{
    if(m_image)
        return Asd::Point32 (DGPIMG_WIDTH((DgpImage_Packed)m_image),DGPIMG_HEIGHT((DgpImage_Packed)m_image));
    else return Asd::Point32(0,0);
}

// depth
ColorDepth SilxImgBuf::depth( void ) const
{ 
    if(m_image){
        switch (DGPIMG_TYPE((DgpImage_Packed)m_image)) {
        case DGPIMAGE_RGB:
		    return COLOR_24_BIT;
            break;
        case DGPIMAGE_PACKED:
            return MONOCHROME;
            break;
        case DGPIMAGE_GRAY:
            return COLOR_8_BIT;
        default:
            return COLOR_DEPTH_ERROR;
        }
    } else return COLOR_DEPTH_ERROR;
}

// data
 Asd::UInt8 * SilxImgBuf::data( void )
{
    // Does this need a cast at all?
    if(m_image)
        return (Asd::UInt8*)DGPIMG_PIXPTR((DgpImage_Gray)m_image, 0, 0);
    else return NULL;
}

const Asd::UInt8 * SilxImgBuf::data( void ) const
{
    // Does this need a cast at all?
    if(m_image)
        return (const Asd::UInt8*) DGPIMG_PIXPTR((DgpImage_Gray)m_image, 0, 0);
    else return NULL;
}

Asd::UInt32 SilxImgBuf::dataSize( void ) const
{
    if(m_image) {
        if(DGPIMG_TYPE((DgpImage_Packed)m_image) != DGPIMAGE_PACKED)
        return Asd::UInt32(DGPIMG_GETROWWIDTH((DgpImage_Gray)m_image)  * DGPIMG_HEIGHT((DgpImage_Packed)m_image));
        else 
            return Asd::UInt32(DGPIMG_GETROWWIDTH((DgpImage_Packed)m_image) * sizeof(Dgp_Int32) * DGPIMG_HEIGHT((DgpImage_Packed)m_image));
    }else return 0;
}

Asd::UInt32 SilxImgBuf::bpl( void ) const
{
    if(m_image){
        if(DGPIMG_TYPE((DgpImage_Packed)m_image) != DGPIMAGE_PACKED)
            return Asd::UInt32(DGPIMG_GETROWWIDTH((DgpImage_Gray)m_image));
        else 
            return Asd::UInt32(DGPIMG_GETROWWIDTH((DgpImage_Packed)m_image) * sizeof(Dgp_Int32));
    }else return 0;
}


bool SilxImgBuf::scaleUp (const Asd::Point32& scaleFac, ImageAdapter* output) 
{ 
    DgpCoord_Resolution res;
    res.xRes = DgpCoord_eqi(scaleFac.getX());
    res.yRes = DgpCoord_eqi(scaleFac.getY());

    switch (DGPIMG_TYPE((DgpImage_Packed)m_image)) {
	case DGPIMAGE_RGB:
		DgpBlat_scaleBlitRGB((DgpImage_RGB *)this->data(), (Dgp_Int32)this->bpl(),0, 0, // src 
			(DgpImage_RGB *)output->data(), output->bpl(), 0, 0, 0, 0, output->width(), output->height(), //dst
			&res );  // scale
		break;
	case DGPIMAGE_PACKED:	
	case DGPIMAGE_GRAY:
		return false;
	}
    return true;
}

 
bool SilxImgBuf::scaleToSize (ImageAdapter* output) 
{ 
    DgpCoord_Resolution res;
    res.xRes = DgpCoord_eq(output->width(), this->width());
    res.yRes = DgpCoord_eq(output->height(), this->height());

    switch (DGPIMG_TYPE((DgpImage_Packed)m_image)) {
	case DGPIMAGE_RGB:
		DgpBlat_scaleBlitRGB((DgpImage_RGB *)this->data(), (Dgp_Int32)this->bpl(),0, 0, // src 
			(DgpImage_RGB *)output->data(), output->bpl(), 0, 0, 0, 0, output->width(), output->height(), //dst
			&res );  // scale

		break;
	case DGPIMAGE_PACKED:	
	case DGPIMAGE_GRAY:
		return false;

	}

	
    return true;
}
