#include "Jb2Decode.h"
#include "dgp_state.h"
#include "uncomp-jbig.h"
#include "dgp_err_printers.h"
#include "SilxErr.h"

namespace silx
{   
    // silx header
    #include "image.h"
    #include "image_defs.h"
    #include "simple_utils.h"
	#include"g4.h"
    #include "uncomp-jbig.h"
}

using namespace silx;

Jb2Decode::Jb2Decode() : m_jb2Info(NULL),m_state(NULL){
	Dgp_Int32 ret = cSuccessful;
	ret = Dgp_initDgpState((Dgp_State **)(&m_state), 
                            NULL, NULL,
                            stderr, Dgp_fileErrPrinter, 
                            DGP_TRUE, DGP_TRUE);
    if(ret != cSuccessful) {
		if(m_state)		Dgp_cleanupDgpState((Dgp_State **)(&m_state));
	}
	ret = initJb2Info(  (jb2info **)(&m_jb2Info),
                        DGP_FALSE, NULL, NULL, DGP_TRUE, 
                        0xffff, DGP_TRUE, DGP_FALSE,
                        NULL, NULL,//use default alloc function
                        (Dgp_State*)m_state);
    if(ret != cSuccessful) {
		if(m_state)		Dgp_cleanupDgpState((Dgp_State **)(&m_state));
		if(m_jb2Info) 	cleanupJb2Info((jb2info **)(&m_jb2Info));
	}
}

Jb2Decode::~Jb2Decode(){
	if(m_jb2Info) 	
		cleanupJb2Info((jb2info **)(&m_jb2Info));
	if(m_state)		
		Dgp_cleanupDgpState((Dgp_State **)(&m_state));
}

bool Jb2Decode::decodeSymDict( char *dictData, long dataSize)
{
	//int status;
    // validation
    SILX_NULL_CHECK(dictData);
	SILX_NULL_CHECK(m_state);
	SILX_NULL_CHECK(m_jb2Info);

	// Get the dictionary data
	SILX_STATUS_CHECK(
		processDictFromData((jb2info*)m_jb2Info, (Dgp_UInt8 *)dictData, dataSize));
	return true;

ERR_CLEANUP:
	return false;
}
bool Jb2Decode::decodePage2G4(
						   char *pageData, 
						   long pageDataSize, 
						   char *dictData, 
						   long dictDataSize, 
						   long width,
						   long height,
						   char **g4Image,
						   long *datasize)
{
	jb2info * pageJb2Info = NULL;

	SILX_NULL_CHECK(m_state);
	SILX_NULL_CHECK(m_jb2Info);

    // First we need to decode the global dictionary if it is not already decoded
	if(dictData && dictDataSize !=0)
		SILX_ERR_CHECK( decodeSymDict(dictData, dictDataSize));

	// create an SilxImageBuf and passed it to jb2 processor to render on
	SILX_ERR_CHECK( 
		initJb2Info((jb2info**)&pageJb2Info,
						DGP_FALSE, 
						NULL, 
						NULL, 
						DGP_TRUE, 
						0xffff,
						DGP_TRUE, 
						DGP_TRUE, 
                        NULL, 
						NULL, 
						(Dgp_State*) m_state));
	
	//Copy segments (dictionary) from m_jb2Info to pageJb2Info)
	SILX_ERR_CHECK(
			copySegmentList(pageJb2Info, (jb2info*)m_jb2Info));

	// Process all segments
    SILX_ERR_CHECK(
		processSegmentsFromData(  pageJb2Info, (Dgp_UInt8 *)pageData, pageDataSize, NULL));
	DgpImage_save(pageJb2Info->curpage, "h:\\temp.pnm");
	SILX_STATUS_CHECK(encodeG4(pageJb2Info->curpage, g4Image, datasize));
	if(pageJb2Info) cleanupJb2Info(&pageJb2Info);
	return true;
ERR_CLEANUP:
	if(pageJb2Info) cleanupJb2Info(&pageJb2Info);
	return false;
}

bool Jb2Decode::encodeG4(void *image, char **data, long* datasize){
	long bpl = 0, width = 0, height = 0;
	UInt32 *pImage;
	UInt8 *encBuf = NULL;
	DgpImage_Packed di = (DgpImage_Packed) image;
	Int32 bytesWritten = 0;

	SILX_NULL_CHECK(di);
	width =  DGPIMG_WIDTH (di);
    height = DGPIMG_HEIGHT (di);
	pImage = DGPIMG_PIXPTR(di, 0 ,0);
	bpl = DGPIMG_GETROWWIDTH(di);

	//alloc encode buffer
	try{
          encBuf = new UInt8[height * bpl];
	}catch (...){
		  delete [] encBuf;
          return false;
	}

	// feed image to g4 encoder
	bytesWritten = height * bpl;
	SILX_ERR_CHECK(g4Encode(pImage, width, height, bpl/4, encBuf, (size_t *)&bytesWritten, DGP_FALSE, (Dgp_State*)m_state));
	
	/*
	Int32 retlen = 0;
	UInt8 *decBuf;
	try{
          decBuf = new UInt8[height * bpl];
	}catch (...){
		  delete [] decBuf;
          return false;
	}
	SILX_ERR_CHECK(
		g4Decode((Dgp_UInt32 *)decBuf, width, height, (width+31)/32, 0, width-1, height-1, 
				encBuf, bytesWritten, DGP_FALSE, (size_t*)&retlen, (Dgp_State*)m_state)); */
	*datasize = (long) bytesWritten; 
	*data = (char *)encBuf;
	return true;
ERR_CLEANUP:
   if(encBuf) delete [] encBuf;
   return false;
}
