#include <string>
#include <iostream>
#include <sstream>

#include "PDFReader.h"
#include "Jb2Decode.h"


void PDFReader::handleJBIG2 (objInfo& oInfo, BitFilter filter)
{

    std::string subKey;
    int width = 0, height=0, dictObjID=0;
    bool hasKey;
    filter = filter;

    // get height
    subKey = "/Height";
    hasKey = hasKeyword (subKey, &height);
    
    // and width of image
    subKey = "/Width";
    hasKey = hasKeyword (subKey, &width);
    
	subKey = "/JBIG2Globals";
	hasKey = hasKeyword(subKey, &dictObjID); 

	std::cout << "Binary JBIG2 Image " << oInfo.objNr << "  " << oInfo.pos << "  " << oInfo.size << "  " << height;
    std::cout << "  " << width << '\n';

	// find out exact data size
	int pageDataSize = 0;
	subKey = "/Length";
	hasKey = hasKeyword(m_tBuf, subKey, &pageDataSize);

    // read all data of the page obj
    char *begin;
	my_ptr<char> buf;
	char *bufPtr = NULL;
	readBuffer( oInfo, &bufPtr);
	buf.reset(bufPtr);
	begin = strstr (buf.get(), "stream");
	begin += 6;
	while (*begin == 0x0A) begin++;
	char *pageData = begin;

	// read all the data of the dictionary object
	char *dictData = NULL;
	int dictDataSize = 0;
	my_ptr<char> bufDict;
	if(dictObjID){
		char * bufDictPtr = NULL;
		std::vector<objInfo>::iterator iter = m_objs.begin(), last = m_objs.end();
		bool dictFound = false;
		while (iter != last){
			if ((*iter).objNr == dictObjID){
				dictFound = true;
				break;
			}
			iter++;
		}
		// if this page data refer to a global dictionary.
		// See if this dictioanry is already decoded
		if(dictFound == true){		
			std::vector<int>::iterator iter2 = m_dictIDs.begin(), last2 = m_dictIDs.end();
			bool newDict = true;
			while (iter2 != last2){
				if ((*iter2) == dictObjID){
					newDict = false;
					break;
				}
				iter2++;
			}
			if(newDict == true) {	
				// only read dictData if this is a new dict
				m_dictIDs.push_back(dictObjID);
				readBuffer(*iter, &bufDictPtr);
				bufDict.reset(bufDictPtr);
				begin = strstr (bufDict.get(), "stream");
				begin += 6;
				while (*begin == 0x0A) begin++;
				dictData = begin;
				
				// find out exact dict data size
				subKey = "/Length";
				hasKey = hasKeyword(bufDict.get(), subKey, &dictDataSize); 
			}
		}		
	}

	// create imInfo struct
    imInfo *info = NULL;
	try{ 
		info = new imInfo;
	}catch(...){
		delete info;
		return;
	}
    info->width = width;
    info->height = height;
    info->comp = JBIG2;
	info->dataSize = 0;


    // decode jbig2 and feed the image to G4 encoder
	char* encBuf = NULL;
	long encDataSize = 0;

    if (! m_pDecJB2->decodePage2G4 (pageData, pageDataSize, dictData, dictDataSize, 
									width, height,  &encBuf, &encDataSize) )
    {
        std::cout << "Decoding of JBIG2 failed\n";
		if(info) delete info;
        return;
    }
 
	oInfo.info = info;
	info->comp = G4;
	info->converted = true;
	info->data.reset(encBuf);
	info->dataSize = encDataSize;

	return;
}



