
#include <fstream>
#include <list>
#include <vector>
#include <string>
#include "AsdTypes.h"

#ifndef PDFREADER_H
#define PDFREADER_H

class Jb2Encode;
class Jb2Decode;


/* create a smart pointer to handle resource */
typedef void (*FreePtr)(void **) ;
template<class T>
class my_ptr{
public:
	my_ptr( T * p = NULL, void (*freePtr)( void **)= NULL): m_ptr(p), m_freePtr(freePtr){}
	~my_ptr() { 
		freePtr();
	}
	T* get(){ return m_ptr;}
	void reset( T* p){ 
		if(m_ptr && m_ptr != p) 
			freePtr();
		m_ptr = p; 
	}
	void freePtr() {
		if(m_freePtr){ 
			if(m_ptr)
				m_freePtr((void **) &m_ptr);
				m_ptr = NULL;
		} else { 
			if(m_ptr)
				delete [] m_ptr;
			m_ptr = NULL;
		}
	}
private:
	T *m_ptr;
	FreePtr m_freePtr;
};



enum Compression { G4, JBIG2, Flate, None };

enum BitFilter { base85, NONE };

struct imInfo
{
    bool mask;
    int width, height;
    Compression comp;
    my_ptr<char> data;
    long dataSize;
    bool converted;
};

struct objInfo
{
    int objNr;
    int size;
    int pos;
    imInfo *info;
    int version;
    int operator< (const objInfo& other) const { return (pos < other.pos); }
};

class PDFReader
{
public:

    PDFReader (const char* in_filename, const bool & jb2g4, const bool & arith);
    ~PDFReader ();

    int getObjects ();
    bool findAndEncodeBinaryImages ();
    bool generateOutputPDF(const char* out_filename);

private:

    /* fills m_objs vector */
    int findXref ();
    void parseXref (int pos);
    int cleanXref ();

    /* buffers and keywords */
    char *m_tBuf;
    int m_tBufSize;
    int readBuffer (const int size);
	int readBuffer(const objInfo& oInfo, char **buf);

    bool hasKeyword (const std::string& keyword, int* value);
    bool hasKeyword (const std::string& keyword, bool* value);
	bool PDFReader::hasKeyword (char * inBuf, const std::string& keyword, int* value);

    /* decode G4 */
    void handleG4 (objInfo& oInfo, BitFilter filter);
    /* decode JBIG2 */
    void handleJBIG2 (objInfo& oInfo, BitFilter filter);
	/* decode Flate */
    void handleFlate (objInfo& oInfo, BitFilter filter);

    bool lookForDecode ();

    /* decode ASCIIBase85  */
    int decodeAB85 (char* inbuf, int buf_size);

    std::ifstream m_file;
    std::vector<objInfo> m_objs;
    int m_fileEnd;
    int m_Size, m_Root, m_Info;

	/* JB2 encode and decode class */
    Jb2Encode* m_pJB2;
	Jb2Decode* m_pDecJB2;

    /* some temp storage for recursive parseXref() func */
    std::list<int> m_xrefPos;
    std::vector<int> m_xrefVec;
    std::vector<int> m_version;
	std::vector<int> m_dictIDs;
    bool m_moreXref;
	bool m_jb2g4;

};

void freePixr(void **);

#endif
