
//#include "StdAfx.h"

#include "PdfOutput.h"
#include <algorithm>



bool PdfOutput::writeTrailer (unsigned int root, unsigned int info)
{
    unsigned int xrefPos = m_offset;
    unsigned int i, objgap, j;

	/* Sort xref objects. */ 
    m_xRef.resize (m_arrayFill);
    std::sort (m_xRef.begin(), m_xRef.end());

    /* Write xref table & trailer. */
	m_offset += fprintf(m_fp, "\
xref\n\
0 %d\n\
0000000000 65535 f \n", m_xRef[m_arrayFill - 1].obj + 1);


	for (i = 0; i < m_arrayFill; i++) {
        if(i > 0){
            objgap = m_xRef[i].obj - m_xRef[i-1].obj;
            for(j = 1; j< objgap; j++){
                	m_offset += fprintf(m_fp, "0000000000 00000 f \n");
            }
        }
		m_offset += fprintf(m_fp, "%010d %05d n \n", m_xRef[i].offset, m_xRef[i].version);
	}

	m_offset += fprintf(m_fp, "\
trailer\n\
<<\n\
/Size %d\n\
/Root %d 0 R\n", m_xRef[m_arrayFill-1].obj + 1, root);

    if (info > 0) m_offset += fprintf(m_fp, "\
/Info %d 0 R\n", info);

	m_offset += fprintf(m_fp, "\
>>\n\
startxref\n\
%d\n\
%%%%EOF\n", xrefPos);

    return true;
}



void
PdfOutput::addXref(unsigned int obj, int version)
{

    if (m_arrayFill == m_xRef.size())
    {
        m_xRef.resize( m_xRef.size() + 50);
    }

    m_xRef[m_arrayFill].obj = obj;
    m_xRef[m_arrayFill].offset = m_offset;
    m_xRef[m_arrayFill].version = version;
    m_arrayFill++;

    return;
}

bool PdfOutput::writeJBIG2 (char * data, unsigned int data_size, unsigned int objNr, unsigned int height, unsigned int width, unsigned int dictObj, bool mask)
{
	addXref(objNr);
    m_offset += fprintf(m_fp, "\
%d 0 obj\n\
<<\n\
/Type /XObject\n\
/Subtype /Image\n\
/Width %d\n\
/Height %d\n\
/BitsPerComponent 1\n", objNr, width, height);

    if (mask) 
        m_offset += fprintf (m_fp, "/ImageMask true\n");
    else
        m_offset += fprintf (m_fp, "/ColorSpace /DeviceGray\n");
    m_offset += fprintf (m_fp, "/Filter /JBIG2Decode\n");
    if (dictObj > 0) 
        m_offset += fprintf (m_fp, "\
/DecodeParms << /JBIG2Globals %d 0 R >>\n", dictObj);
    m_offset += fprintf (m_fp, "\
/Length %d\n\
>>\n\
stream\n", data_size);
    fwrite (data, data_size, 1, m_fp);
    m_offset += data_size;
    m_offset += fprintf(m_fp, "\n\
endstream\n\
endobj\n\n");

    return true;
}

bool PdfOutput::writeG4 (char * data, unsigned int data_size, unsigned int objNr, unsigned int height, unsigned int width)
{
	if(!data) return false;
	addXref(objNr);
    m_offset += fprintf(m_fp, "\
%d 0 obj\n\
<<\n\
/Type /XObject\n\
/Subtype /Image\n\
/Width %d\n\
/Height %d\n\
/ColorSpace /DeviceGray\n\
/BitsPerComponent 1\n\
/Filter /CCITTFaxDecode\n", objNr, width, height);
 m_offset += fprintf (m_fp, "\
/DecodeParms <</K -1 /Columns %d /Rows %d /BlackIs1 false  >>\n", width, height);
    m_offset += fprintf (m_fp, "\
/Length %d\n\
>>\n\
stream\n", data_size);
    fwrite (data, data_size, 1, m_fp);
    m_offset += data_size;
    m_offset += fprintf(m_fp, "\n\
endstream\n\
endobj\n\n");

    return true;
}

bool PdfOutput::writeJBIG2Dict (char * data, unsigned int data_size, unsigned int dictObj)
{
	addXref(dictObj);
	m_offset += fprintf(m_fp, "\
%d 0 obj\n\
<< /Length %d\n\
>>\n\
stream\n", dictObj, data_size); 
    fwrite (data, data_size, 1, m_fp);
    m_offset += data_size; 
    m_offset += fprintf(m_fp, "\n\
endstream\n\
endobj\n\n");

    return true;
}

void PdfOutput::writeObj (char * data, unsigned int data_size, unsigned int objNr, int version)
{
    if (version >= 0)
    {
        addXref (objNr, version);
        if(data_size > 0 && data)
			fwrite (data, data_size, 1, m_fp);
        m_offset += data_size;
    }
}

PdfOutput::PdfOutput () : m_xRef(50), m_arrayFill(0), m_offset(0), m_fp (NULL)
{
}

bool PdfOutput::PdfStartDoc (const char* filename)
{
    if (m_fp) return false;
    m_fp = fopen (filename, "wb");
    if (!m_fp) return false;

	/* write header */
	m_offset += fprintf(m_fp, "\
%%PDF-1.4\n%%\xac\xad\xae\xaf\n");

    return true;
}

int PdfOutput::PdfEndDoc ()
{

    fclose (m_fp);
    m_fp = NULL;

    return m_offset;
}
