
//#include "StdAfx.h"

#include <string>
#include <iostream>
#include <sstream>

#include "PDFReader.h"
#include "Jb2Encode.h"
#include "SilxErr.h"

#ifdef PSG
#undef cFalse
#define cFalse 0
#undef cTrue
#define cTrue 1
#endif

#ifndef COMPRESSION_CCITTFAX4
#define COMPRESSION_CCITTFAX4 4
#endif

static bool g4ToPixr ( ipcore::PIXR** ret_image, const char* buffer, Asd::Int32 buf_size, 
                         Asd::Int32 width, Asd::Int32 height, Asd::Int32 k, bool inverse, const char* filename);


// decode ASCIIBase85. Returns the size of the buffer on output
int PDFReader::decodeAB85 (char* inbuf, int buf_size)
{
    const int p[5] = {52200625, 614125, 7225, 85, 1};
    
    const int pp1 = 256;
    const int pp2 = 65536;
    const int pp3 = 16777216;
    
    unsigned int total = 0, temp;
    
    int val, count = 0;
    my_ptr<char> outBuf (new char [buf_size], NULL);
    char *outP = outBuf.get();
    char *begin = inbuf;
    int outcount = 0;
    
    while (1) // infinite loop
    {
        while ((val = (int)*begin++) < 33) // get a new val, skip all white chars
        {
            if (begin - inbuf > buf_size) return 0;
        }
        if (val == 126) // ~
        {
            // output last count - 1 bytes
            if ( count ) count--;
            if ( count )
            {
                outcount++;
                if (outcount > buf_size) return outcount;

                temp = total / pp3;
                *outP++ = (char) temp;
                total -= temp * pp3;
                count--;
            }
            
            if ( count )
            {
                outcount++;
                if (outcount > buf_size) return outcount;

                temp = total / pp2;
                *outP++ = (char) temp;
                total -= temp * pp2;
                count--;
            }
            
            if ( count )
            {
                outcount++;
                if (outcount > buf_size) return outcount;

                temp = total / pp1;
                *outP++ = (char) temp;
                total -= temp * pp1;
            }
            break; // we are done
        }
        
        // a valid character ...
        if (val <= 117)
        {
            total += (val - 33) * p[count];
            count ++;
        }
        else if (val == 122) // z
        {
            outcount += 4;
            if (outcount > buf_size) return outcount;
            *outP++ = 0;
            *outP++ = 0;
            *outP++ = 0;
            *outP++ = 0;
        }
        else
            return false;
        
        if ( count == 5 )
        {
            outcount += 4;
            if (outcount > buf_size) return outcount;

            temp = total / pp3;
            *outP++ = (char) temp;
            total -= temp * pp3;
            
            temp = total / pp2;
            *outP++ = (char) temp;
            total -= temp * pp2;
            
            temp = total / pp1;
            *outP++ = (char) temp;
            total -= temp * pp1;
            
            *outP++ = (char) total;
            count = 0;
            total = 0;
        }
    }

    // copy the decoded data into the buffer
    memcpy (inbuf, outBuf.get(), buf_size);
    return outcount;

}


void PDFReader::handleG4 (objInfo& oInfo, BitFilter filter)
{

    std::string subKey;
    int width, height, k = 0, cols = 1728;
    bool hasKey;
    bool inverse = false;

    // get height
    subKey = "/Height";
    hasKey = hasKeyword (subKey, &height);
    
    // and width of image
    subKey = "/Width";
    hasKey = hasKeyword (subKey, &width);
    
    subKey = "/K";
    hasKey = hasKeyword (subKey, &k);

    subKey = "/Columns";
    hasKey = hasKeyword (subKey, &cols);

    subKey = "/BlackIs1";
    hasKey = hasKeyword (subKey, &inverse);
    inverse ^= lookForDecode ();

    std::cout << "Binary G3/4 Image " << oInfo.objNr << "  " << oInfo.pos << "  " << oInfo.size << "  " << height;
    std::cout << "  " << width << "  " << k << "  " << cols << "  " << inverse << '\n';
    
    // read all the data of the object
    char *begin;
	my_ptr<char> buf(new char [oInfo.size]);
    m_file.seekg (oInfo.pos, 
#ifdef _WIN32
std::ios_base::beg);
#else
std::ios::beg);
#endif

    m_file.read (buf.get(), oInfo.size); 

    begin = strstr (buf.get(), "stream");
    begin += 6;
    while (*begin == '\015' || *begin == '\012') begin++;
    int offset = (int) begin - (int) buf.get();

    // create imInfo struct
    imInfo *info = NULL;
	try{
		info = new imInfo;
	}catch(...){
		delete info;
		return;
	}
    info->width = width;
    info->height = height;
    info->comp = G4;
	info->converted = false;

    // check if we need to decode a pre filter
    if (filter == base85)
    {
        int newSize =  decodeAB85 (begin, oInfo.size - offset);
        if (newSize == 0 || newSize > oInfo.size - offset)
        {	delete info;
            std::cout << "Decoding of ASCIIBase85 failed\n";
            return;
        }
    }   

    // decode G3/4
    ipcore::PIXR* ret_image = NULL;    
    if (!g4ToPixr ( &ret_image, begin, oInfo.size - offset, 
        width, height, k, inverse, NULL )) 
    {
		delete info;
        std::cout << "Decoding of G4 failed\n";
        return;
    }
    begin = NULL;
    oInfo.info = info;
    // feed pixr to tokenizer
    bool status = m_pJB2->addPage (ret_image);
	if (status) {
		info->comp = JBIG2;
		info->converted = true;
	}
    freePixr ((void **)&ret_image);


    return;
}


// IPCORE G4 libs
namespace ipcore
{
#include "rcs.pub"
#include "types.pub"
#include "pixr.h"

#include "frames.pub"
#include "shrcmprs.pub"
#include "ccitt.pub"
#include "mmr.pub"

#include "geomadt.h"
#include "props.pub"

#include "rast.pub"

#include "reduce.pub"

#include "prxform.pub"

#include "rescale.pub"

#include "shrpixr.pub"

// for writeTiff
#include "imagerep.h"
#include "shrcc.pub"
#include "iotiff.pub"
#include "g4.h"

const int pixrRes = 300;
const int framePix = 32;
}

void freePixr ( void **ret_image )
{
   ipcore::IPWState *pIPWState = NULL;

#ifdef PSG
    // ipcore setup
    Asd::Int32 status = ipcore::w_IPWStateInit (NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, 0, NULL, NULL, NULL, &pIPWState);
#else
    Asd::Int32 status = ipcore::w_IPWStateInit (NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, 0, NULL, NULL, &pIPWState);
#endif

    ipcore::w_destroyPixr (pIPWState, (ipcore::PIXR **)ret_image);

    status += ipcore::w_IPWStateTerm (&pIPWState);

}

static bool g4ToPixr ( ipcore::PIXR** ret_image, const char* buffer, Asd::Int32 buf_size, 
                         Asd::Int32 width, Asd::Int32 height, Asd::Int32 k, bool inverse, const char* filename)
{

    *ret_image = NULL;
    Asd::UInt8 *compInfo = NULL;

    //Initialize varaibles

	//ipcore
    ipcore::IPWState *pIPWState = NULL;
    ipcore::PIXR* retPixr = NULL;
    Asd::UInt8 *data = NULL;
    Asd::Int32 bpl = 0;

	// set up for decode
    Asd::UInt32 imgExh = 0, cmpExh = 0;
    Asd::UInt8 *iPtr = NULL;
    Asd::UInt8 *cPtr = NULL;
    Asd::Int32 bytesLeft = 0, bytesUsed = 0, pct =0;
    bool contImg = 0, contCmp = 0;
    Asd::UInt32 firstRow = 0;
    Asd::UInt32 nRow = 0;

	// ipcore setup
#ifdef PSG
    SILX_ERR_CHECK(ipcore::w_IPWStateInit( NULL, NULL, NULL, NULL, NULL,
											NULL, NULL, NULL, NULL, NULL,
											NULL, NULL, 0, NULL, NULL, NULL, &pIPWState));
#else
    SILX_ERR_CHECK( ipcore::w_IPWStateInit (	NULL, NULL, NULL, NULL, NULL,
									NULL, NULL, NULL, NULL, NULL,
									NULL, NULL, 0, NULL, NULL, &pIPWState));
#endif

    // allocate temporary PIXR
    SILX_ERR_CHECK(ipcore::w_createPixr ( pIPWState, width, height, 1, 
        ipcore::pixrRes, ipcore::pixrRes, ipcore::framePix, &retPixr));
    SILX_ERR_CHECK( ipcore::w_pixrGetImage (pIPWState, retPixr, &data));
    SILX_ERR_CHECK( ipcore::w_pixrGetBpl (pIPWState, retPixr, &bpl));

    // find first and last row of the band that we need
    firstRow = 0;
    nRow = height;

	// allocate decode structures
    Asd::Int32 stateBlockSize;
    SILX_ERR_CHECK(ipcore::i_getSizeOfCCITTDecompressionInfo( pIPWState,
						width + 2 * ipcore::framePix, &stateBlockSize));

	try{
		 compInfo = new Asd::UInt8 [stateBlockSize];
	}catch(...){
		delete [] compInfo;
	}

	SILX_ERR_CHECK( ipcore::i_initializeCCITTDecompressionInfo(
							pIPWState,
							compInfo, width + 2 * ipcore::framePix, k < 0,
							cPadToByte, stateBlockSize));

    // clear the image
    SILX_ERR_CHECK( ipcore::w_clear (pIPWState, retPixr));

	// set up for decode
    iPtr = data - ipcore::framePix / 8;
    cPtr = (Asd::UInt8*) buffer;
    bytesLeft = buf_size;
    
    // call the decoder
    do {
#ifndef PSG
		if (k < 0){
			SILX_ERR_CHECK(ipcore::i_decompressRowsMMR (pIPWState, compInfo,
												iPtr, firstRow,
												nRow, bpl, cFalse, cPtr, bytesLeft, contImg,
												contCmp, 1, &imgExh, &cmpExh, &pct, &bytesUsed));
		}
		else {
			SILX_ERR_CHECK(ipcore::i_decompressRowsMR (pIPWState, compInfo,
												iPtr, firstRow,
												nRow, bpl, cFalse, cPtr, bytesLeft, contImg,
												contCmp, 1, &imgExh, &cmpExh, &pct, &bytesUsed));
		}
#else //IPCore-PSG defines Bool as int instead of UInt32
		if (k < 0)
			SILX_ERR_CHECK(ipcore::i_decompressRowsMMR (pIPWState, compInfo,
							iPtr, firstRow,
							nRow, bpl, cFalse, cPtr, bytesLeft, contImg,
							contCmp, 1, ((int *)&imgExh), ((int *)&cmpExh), &pct, &bytesUsed));
        else
			SILX_ERR_CHECK(ipcore::i_decompressRowsMR (pIPWState, compInfo,
							iPtr, firstRow,
							nRow, bpl, cFalse, cPtr, bytesLeft, contImg,
							contCmp, 1, ((int *)&imgExh), ((int *)&cmpExh), &pct, &bytesUsed));
#endif
        bytesLeft -= bytesUsed;

	    if (imgExh == 1 && cmpExh == 0) {
	        /* Okay, we can live with it, but it's wrong */
	    } else 
			SILX_STATUS_CHECK( cmpExh == 1 && imgExh == 0);

    	contImg = 1;
	    contCmp = 1;
    } while (imgExh == 0 && cmpExh == 0 && bytesLeft > 0);

    if (inverse) SILX_ERR_CHECK( ipcore::w_invert (pIPWState, retPixr, &retPixr));
    // write image
    if (filename)
		SILX_ERR_CHECK(ipcore::w_writeTiff (pIPWState, retPixr, 
			filename, COMPRESSION_CCITTFAX4 ,cTrue, 0));

    // cleanup
    SILX_ERR_CHECK(ipcore::w_IPWStateTerm (&pIPWState));
    *ret_image = retPixr;
	delete []  compInfo;
    return true;

ERR_CLEANUP:
    if(retPixr) ipcore::w_destroyPixr(pIPWState, &retPixr);
    ipcore::w_IPWStateTerm (&pIPWState);
	delete []  compInfo;
    return false;


}

