#ifdef _WIN32
#include <windows.h>
#pragma warning(disable:4786) // supress the annoying windows warning
#endif

#include "PDFReader.h"
#include <iostream>
#include "Option.h"

using namespace std;

void 
addOptions(Option& op, const string& filename){
        op.addOption("o", filename, STRING_TYPE, "Output File Name");
		op.addOption("arith", false_string, FLAG_TYPE, "Use arithmetic encoding");
		op.addOption("jb2g4", false_string, FLAG_TYPE, "Convert JBIG2 Encoded PDF to G4 encoded PDF");
        op.addOption("help", false_string, FLAG_TYPE, "Print help");
}
bool parseArguments(Option &op, int& argc, char**& argv){

	// parse options with marker
    if(!op.parseOptions(--argc, ++argv)) return false;
    //first looking for help option
    bool helpFlag = false;
    op.getValue("help", helpFlag);
    if(helpFlag) return false;
    return true;
}

int main (int argc, char** argv)
{
	const string dfOutFile("output.pdf");

    //create header string
    string header("PDFConverter --- Convert G4 encoded PDF to JBIG2 encoded PDF or vice versa \n");
    header.append("PDFConverter [-options] input.pdf  \n");

	// create Option and add custom options
    Option op(header, "-");   
    addOptions(op, dfOutFile);

	//parse Option
    if(! parseArguments(op, argc, argv)){
            op.printHelp();
            return 1;
    } 
    
	//parse input file name 
    if(argc ==0) {
                op.printHelp();
            return 1;
    }

   //get outfile name
   string outfile;
   op.getValue("o", outfile);

   bool jb2g4 = false;
   op.getValue("jb2g4", jb2g4);

	bool arithmetic = false;
   op.getValue("arith", arithmetic);

   // last argument has to be infile
   string infile(argv[0]);
   // if no user setup for output, modify default outfile
   if(!outfile.compare(dfOutFile)) {
	   if(infile.find_last_of(".") != string::npos)
			outfile = infile.substr(0,infile.find_last_of("."));
		else
			outfile = infile;
	   if(jb2g4)
		   outfile += "_g4.pdf";
	   else
		   outfile +="_jb2.pdf";
   }

   PDFReader finder (infile.data(), jb2g4, arithmetic);

   finder.getObjects ();
   if (finder.findAndEncodeBinaryImages ())
       finder.generateOutputPDF(outfile.data());
   else
       std::cout << "No output created, because no images need to be changed in this PDF file\n";

    return 1;
}