#if !defined(AFX_SILXSCREENCTL_H__A70845DB_C709_42AA_964B_E78F84FAD21B__INCLUDED_)
#define AFX_SILXSCREENCTL_H__A70845DB_C709_42AA_964B_E78F84FAD21B__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif /* _MSC_VER > 1000 */

/* SilxScreenCtl.h : Declaration of the CSilxScreenCtrl ActiveX Control class. */

/*/////////////////////////////////////////////////////////////////////////// */
/* CSilxScreenCtrl : See SilxScreenCtl.cpp for implementation. */

class TfxFile;
class TfxPage;

class CSilxScreenCtrl : public COleControl
{
	DECLARE_DYNCREATE(CSilxScreenCtrl)

/* Constructor */
public:
	CSilxScreenCtrl();

    /*********************************************************************

    Page represents a page in our file.  We use this extra level of
    inderiection between our control and TfxPage for three reasons:

    1.  There is some functionality for which the Page class makes sense
        as a name space.
    2.  We have caching information that we would like to store with page.
    3.  We would like to be able to use pages without actually creating
        the TfxPage object if we can avoid it.  Page knows when to go
        to the TfxFile with a page number and when to use TfxPage.
    
    *********************************************************************/
    
    class Page
    {
    public:

        Page( Asd::UInt32 pageNum, TfxFile * );

        virtual ~Page( void ) { /*Do Nothing*/ }

		void clear( void );
	
		Asd::Float64 approxBytesNeeded( Asd::Float64 zoom, ColorDepth ) const;
        
		Asd::Float64 currentBytesNeeded( void ) const;

        TfxPage * getTfxPage( void ) const;

		Asd::Point32 getSize( void ) const;

        /* The following data is public because we need to have nearly total */
		/* control over it.  Basically, it is in an object because we want to */
		/* have it grouped.  We do not want to hide it. */

		/* The image we have rendered for this page.  This does not have to */
		/* be the whole page. */
		DibImgBuf m_img;

        /* The zoom factor at which m_img was rendered. */
		Asd::Float64 m_zoom;

        /* What portion of the page has been rendered in zoomed coordinates. */
        Asd::Rect m_renderedRegion;

        /* The layer that was rendered. */
		Asd::Int32 m_layer;

		/* What color was the image rendered at? */
		ColorDepth m_colorDepth;

    private:

        /* This data is private because it is part of our Page abstraction. */
		/* */
		/* Note that we do not actually store a pointer to a TfxPage. */
		/* When we need one, we use m_file and m_pageNum to fetch it. */
		/* We possible, however, we just call methods on m_file using */
		/* m_pageNum. */
		TfxFile * m_file; /* We don't own this. */
        Asd::UInt32 m_pageNum;
    };


/* Overrides */
	/* ClassWizard generated virtual function overrides */
	/*{{AFX_VIRTUAL(CSilxScreenCtrl) */
	public:
	virtual void OnDraw(CDC* pdc, const CRect& rcBounds, const CRect& rcInvalid);
	virtual void DoPropExchange(CPropExchange* pPX);
	virtual void OnResetState();
	virtual DWORD GetControlFlags();
	/*}}AFX_VIRTUAL */

/* Implementation */
protected:
	~CSilxScreenCtrl();

	DECLARE_OLECREATE_EX(CSilxScreenCtrl)    /* Class factory and guid */
	DECLARE_OLETYPELIB(CSilxScreenCtrl)      /* GetTypeInfo */
	DECLARE_PROPPAGEIDS(CSilxScreenCtrl)     /* Property page IDs */
	DECLARE_OLECTLTYPE(CSilxScreenCtrl)		/* Type name and misc status */

/* Message maps */
	/*{{AFX_MSG(CSilxScreenCtrl) */
	afx_msg BOOL OnMouseWheel(UINT nFlags, short zDelta, CPoint pt);
	/*}}AFX_MSG */
	DECLARE_MESSAGE_MAP()

/* Dispatch maps */
	/*{{AFX_DISPATCH(CSilxScreenCtrl) */
	long m_fullPageCacheThreshold;
	afx_msg void OnFullPageCacheThresholdChanged();
	long m_pageBreakSpace;
	afx_msg void OnPageBreakSpaceChanged();
	long m_extraPageCacheSize;
	afx_msg void OnExtraPageCacheSizeChanged();
	long m_viewMode;
	afx_msg void OnViewModeChanged();
	afx_msg BSTR GetFileName();
	afx_msg void SetFileName(LPCTSTR lpszNewValue);
	afx_msg double GetZoomFactor();
	afx_msg void SetZoomFactor(double newValue);
	afx_msg long GetCurrentPage();
	afx_msg void SetCurrentPage(long nNewValue);
	afx_msg long GetLayer();
	afx_msg void SetLayer(long nNewValue);
	afx_msg BOOL GetIsTiff();
	afx_msg long GetOrientation();
	afx_msg void SetOrientation(long nNewValue);
	afx_msg BOOL GetIsSilx();
	afx_msg long FirstPage();
	afx_msg long LastPage();
	afx_msg long NextPage();
	afx_msg long PreviousPage();
	afx_msg long NumPages();
	afx_msg void SetPosition(long x, long y);
	afx_msg long GetYPosition();
	afx_msg long GetXPosition();
	afx_msg void PurgeAllData();
	afx_msg void SetXPosition(long x);
	afx_msg void SetYPosition(long y);
	afx_msg long GetModalHeight();
	afx_msg long GetModalWidth();
	afx_msg long GetTotalHeight();
	afx_msg long GetTotalWidth();
	afx_msg long GetCurrentPageWidth();
	afx_msg long GetCurrentPageHeight();
	afx_msg void Pan(long dx, long dy);
	afx_msg long GetModalXPosition();
	afx_msg long GetModalYPosition();
	afx_msg void SetHandCursor();
	afx_msg void SetGrabCursor();
	afx_msg void SetDefaultCursor();
	afx_msg void SetZoomToolCursor();
	afx_msg long CopyToClipboardTLBR(long top, long left, long bottom, long right);
	afx_msg long CopyPageToClipboard(long page);
	afx_msg void SetModalYPosition(long y);
	afx_msg long GetPageX();
	afx_msg long GetPageY();
	afx_msg void SetZoomAndModalPosition(double zoom, long x, long y);
	afx_msg long GetPageForPosition(long x, long y);
	afx_msg long LayersOnCurrentPage();
	afx_msg BSTR GetFileVersion();
	afx_msg BSTR GetBuildDate();
	afx_msg long PrintLayerFromTo(long hdc, long layer, long fromPage, long toPage);
	afx_msg long SavePageAsBmp(LPCTSTR fileName, long page, long layer, double zoom);
	afx_msg long LayersOnPage(long page);
	afx_msg long ConvertTiff2Tfx(LPCTSTR TiffFile, LPCTSTR TfxFile);
	afx_msg void SetOrientationAndPagePosition(short orientation, long x, long y);
	/*}}AFX_DISPATCH */
	DECLARE_DISPATCH_MAP()

	afx_msg void AboutBox();

/* Event maps */
	/*{{AFX_EVENT(CSilxScreenCtrl) */
	void FireMouseWheel(long nFlags, short zDelta, long x, long y)
		{FireEvent(eventidMouseWheel,EVENT_PARAM(VTS_I4  VTS_I2  VTS_I4  VTS_I4), nFlags, zDelta, x, y);}
	/*}}AFX_EVENT */
	DECLARE_EVENT_MAP()

/* Dispatch and event IDs */
public:

    enum ViewMode
    {
        ViewContinuous = 0,
        ViewOnePage    = 1,
        MaxViewMode    = 1
    };

	enum {
	/*{{AFX_DISP_ID(CSilxScreenCtrl) */
	dispidFileName = 5L,
	dispidZoomFactor = 6L,
	dispidCurrentPage = 7L,
	dispidFullPageCacheThreshold = 1L,
	dispidPageBreakSpace = 2L,
	dispidExtraPageCacheSize = 3L,
	dispidViewMode = 4L,
	dispidLayer = 8L,
	dispidIsTiff = 9L,
	dispidOrientation = 10L,
	dispidIsSilx = 11L,
	dispidFirstPage = 12L,
	dispidLastPage = 13L,
	dispidNextPage = 14L,
	dispidPreviousPage = 15L,
	dispidNumPages = 16L,
	dispidSetPosition = 17L,
	dispidGetYPosition = 18L,
	dispidGetXPosition = 19L,
	dispidPurgeAllData = 20L,
	dispidSetXPosition = 21L,
	dispidSetYPosition = 22L,
	dispidGetModalHeight = 23L,
	dispidGetModalWidth = 24L,
	dispidGetTotalHeight = 25L,
	dispidGetTotalWidth = 26L,
	dispidGetCurrentPageWidth = 27L,
	dispidGetCurrentPageHeight = 28L,
	dispidPan = 29L,
	dispidGetModalXPosition = 30L,
	dispidGetModalYPosition = 31L,
	dispidSetHandCursor = 32L,
	dispidSetGrabCursor = 33L,
	dispidSetDefaultCursor = 34L,
	dispidSetZoomToolCursor = 35L,
	dispidCopyToClipboardTLBR = 36L,
	dispidCopyPageToClipboard = 37L,
	dispidSetModalYPosition = 38L,
	dispidGetPageX = 39L,
	dispidGetPageY = 40L,
	dispidSetZoomAndModalPosition = 41L,
	dispidGetPageForPosition = 42L,
	dispidLayersOnCurrentPage = 43L,
	dispidGetFileVersion = 44L,
	dispidGetBuildDate = 45L,
	dispidPrintLayerFromTo = 46L,
	dispidSavePageAsBmp = 47L,
	dispidLayersOnPage = 48L,
	dispidConvertTiff2Tfx = 49L,
	dispidSetOrientationAndPagePosition = 50L,
	eventidMouseWheel = 1L,
	/*}}AFX_DISP_ID */
	};

private:

	/* The TfxFile object that we are viewing.  This pointer may be NULL. */
	/* It just means we do not currently have an open file. */
    TfxFile * m_file;
        
    /* The name of the file we are viewing.  This is used only to */
    /* persist our object and to name it when printing. */
    CString m_fileName;

    /* What orientation do we want to show? */
    Orientation m_orientation;

    /* The amount that we are zoomed as compared to the raw image pixels. */
    /* At 1.0 --> 1 pixel on the screen is 1 pixel in the raw image. */
    /* At 0.5 --> 1 pixel on the screen is 2x2 pixels in the raw image. */
    /* At 2.0 --> 1 pixel in the raw image occuppies 2x2 on the screen. */
    /* We restrict m_zoom to the range 10 >= m_zoom >= 0.01. */
    Asd::Float64 m_zoom;
	
    /* The current page.  This is the page the currently overlaps the */
    /* upper-left-hand corner of the control. */
    Asd::UInt32 m_currentPage;
    
    /* The location in the image (in raw image pixels) that is in the */
    /* upper-left-hand corner of the control.  The position is relative */
	/* to the current page. */
    Asd::Point32 m_position;

    /* Array of Page objects representing the pages in the document.  These */
    /* may be NULL rigth up to the time they are first used.  If you want to */
    /* use a page, you should call getPage().  Basically, unless you are */
	/* actually modifying methods like getPage(), clear(), etc., the only,  */
	/* reason you should use m_pgs directly is to find out if the page has */
	/* been allocated or not.  getPage() will create the page if it does not */
	/* exist.  That is not always desireable. */
    std::valarray<Page *> m_pgs;

    /* What layer are we currently rendering?  0 indicates we are supposed */
    /* to render the composite image.  1 is the background.  2 is the first */
	/* selector.  3 is the first foreground, etc. */
    Asd::Int32 m_layer;
	 
    Page * getPage( Asd::UInt32 page );
	Page * getCurrentPage( void ) { return getPage( m_currentPage ); }

    Asd::Point32 getOrientedPageSize( Asd::UInt32 page, Orientation );

	bool isPageAllocated( Asd::UInt32 page );
    
    void clearPage( Asd::UInt32 page );
      
	void Draw( CDC* , const Asd::Rect& bounds, const
        Asd::Rect& invalid, Asd::UInt32 page, const Asd::Point32 & position,
        Asd::Float64 zoom, Orientation, bool fill, Asd::Int32 layer,
        ViewMode );

    void DrawWash( CDC *, const Asd::Rect & bounds, const Asd::Rect & invalid );
    
	void DrawZoomedIn( CDC*, const Asd::Rect& bounds, const
        Asd::Rect& invalid, Asd::UInt32 page, const Asd::Point32 & position,
        Asd::Float64 zoom, Orientation, bool fill, Asd::Int32 layer,
        ViewMode );

    void DrawZoomedOut( CDC*, const Asd::Rect& bounds, const
        Asd::Rect& invalid, Asd::UInt32 page, const Asd::Point32 & position,
        Asd::Float64 zoom, Orientation, bool fill, Asd::Int32 layer,
        ViewMode );

    Asd::Rect DrawZoomedInPage( CDC*, const Asd::Rect & bounds,
        const Asd::Rect & invalid, Asd::UInt32 page,
        const Asd::Point32 & topLeft, Asd::Float64 zoom, Orientation,
        bool fill, Asd::Int32 layer,ViewMode );

    Asd::Rect DrawZoomedOutPage( CDC*, const Asd::Rect & bounds,
        const Asd::Rect & invalid, Asd::UInt32 page,
        const Asd::Point32 & topLeft, Asd::Float64 zoom, Orientation,
        bool fill, Asd::Int32 layer, ViewMode );

    void loadFile( const std::string & fileName );

    void DrawFilledRect( CDC *, const Asd::Rect &, const COLORREF & );

    void FillBackRect( CDC *, const Asd::Rect & bounds );

    void PurgeUnneededPages( Asd::UInt32 before, Asd::UInt32 after );

    static Asd::Float64 forceZoomInBounds( Asd::Float64 );

    Asd::Int32 convertModalScreenYToY( const Asd::Int32 y, Asd::UInt32 * pageVal );
};

/*{{AFX_INSERT_LOCATION}} */
/* Microsoft Visual C++ will insert additional declarations immediately before the previous line. */

#endif /* !defined(AFX_SILXSCREENCTL_H__A70845DB_C709_42AA_964B_E78F84FAD21B__INCLUDED) */
