#include "StdAfx.h"

#include "dgp_err_handle.h"
#include "silx_encode.h"
#include "DgpEncode.h"
#include <time.h>
#include <limits.h>


ErrCode tiff2Tfx(Dgp_Int8* infname, Dgp_Int8* outfname, Dgp_State *dgp_state)
{
    // For error & status reports...
    Dgp_Int32               retval = 0;
    FILE                    *outfp = NULL;
    SilxEncCookie           sCookie;

    DGP_PROLOGUE("tiff2Tfx()", DGP_PERRSTATE);
    /*check input parameter*/
    DGP_CHECK_NULL_CLEANUP(infname);
    DGP_CHECK_NULL_CLEANUP(outfname);

    DGP_SYS_ERR_GEN_CLEANUP(IS_NULL(outfp = fopen (outfname, "wb")));
    DGP_ERR_PASS_CLEANUP(silx_initEncState(dgp_state, &sCookie));
    
    /*start encodig*/
    DGP_ERR_PASS_CLEANUP(silx_encStart(dgp_state, sCookie, (void *)outfp, NULL)); 
    retval = (silx_encDocument(dgp_state, sCookie, infname, DGP_TRUE));
    if(retval == cTFXDictionaryLimitErr){
            DGP_ERR_PRINTERF0("Terminating encoder because TIFF-FX dictionary limit of 5 has been reached.");
	}else DGP_ERR_PASS_CLEANUP(retval);

    DGP_ERR_PASS_CLEANUP(silx_cleanEncState(dgp_state, &sCookie));
    if(outfp) fclose(outfp);       
    Dgp_cleanupDgpState(&dgp_state);

    NORMAL_RETURN;
    
ERR_CLEANUP:
    if(outfp) fclose(outfp);
    if(sCookie) silx_cleanEncState(dgp_state, &sCookie);
    ERR_RETURN;
}

/* test */
void Tiff2Tfx( const std::string & tiff, const std::string & tfx )
{
    Dgp_Int8 * infname  = NULL;
    Dgp_Int8 * outfname = NULL;

    // We will need to do a const cast, so we make copies that we will
    // immediately discard.
    std::string tiffCopy( tiff );
    std::string tfxCopy( tfx );

    // std::strings are not always null-terminated.
    tiffCopy += '\0';
    tfxCopy  += '\0';

    infname  = const_cast<Dgp_Int8 *>(tiffCopy.data());
    outfname = const_cast<Dgp_Int8 *>(tfxCopy. data());


    Dgp_Int32 ret = cSuccessful;
    Dgp_State* dgp_state=NULL;

    /* Init dgp state */    
    ret = Dgp_initDgpState( &dgp_state, NULL, NULL, stderr,
        Dgp_fileErrPrinter, DGP_TRUE, DGP_TRUE );

    if(dgp_state)
    {
        ret+= tiff2Tfx(infname, outfname,dgp_state);
    }

    if(ret != cSuccessful)
    {
        if(dgp_state)
        {
            Dgp_cleanupDgpState((Dgp_State **)(&dgp_state));
        }
    }
}