#include "bit_array.h"
#include "PsgFix.h"
/* For VxWorks/Tornado, malloc is in stdlib.h / dhtesq. */
#ifdef VXWORKS
#include <stdlib.h>
#else
#include <malloc.h>
#endif

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libadt/bit_array.c,v 1.10 2003/02/10 22:53:18 mkrocken Exp $");

/* macros for going from index into array to index of uchar
   and offset in that uchar */
#define UCHARS_NEEDED(x) ((x + DGP_BITS_UINT8 - 1) / DGP_BITS_UINT8)

#define UCHAR_HOLDING(i) (i / DGP_BITS_UINT8)

#define UCHAR_OFFSET(i) (i % DGP_BITS_UINT8)

/* macro for looking at the j'th bit of u */
#define UCHAR_MASK(u,j) (u & (1 << j))


/* macro returning u with the j'th bit set */
#define UCHAR_W_SET(u,j) (u | (1 << j))

/* macro returning u with the j'th bit cleared */
#define UCHAR_W_CLEAR(u,j) (u & ~(1 << j))


/* Create a new DgpAdt_BitArray which can hold num_bits bits. num_bits need not
   be a multiple of the length of a byte. This function uses calloc (or the
   user-defined calloc) to allocate the packed array (so the elements of
   the array will be initialized to 0). */
ErrCode 
DgpAdt_baNew(DgpAdt_BitArray *_ba, Dgp_Int32 num_bits, Dgp_State *dgp_state) 
{
  DGP_PROLOGUE("baNew()", DGP_PERRSTATE);
  Dgp_Allocators *a = DGP_PALLOCATORS;
  DgpAdt_BitArray ba = DGP_NULLBITARRAY;

  DGP_CHECK_NULL(_ba);

  DGP_ERR_GEN_CLEANUP((num_bits < 1), cErrInvalidParm);

  /* allocate the structure */
  DGP_CHECK_ALLOC_CLEANUP(ba = (DgpAdt_BitArray)(a->calloc)((size_t)1,sizeof(*ba)));
  ba->dgp_state = dgp_state;
  
  /* allocate the actual array */
  DGP_CHECK_ALLOC_CLEANUP(ba->data = 
	  (Dgp_UInt8 *)(a->calloc)((size_t)1, (size_t)UCHARS_NEEDED(num_bits)));
  ba->num_bits = num_bits;
  ba->first_set = -1;

  *_ba = ba;
  ba = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(ba != NULL) {
    if(ba->data != NULL) (a->free)(ba->data);
    (a->free)(ba);
  }

  ERR_RETURN;
}
  
/* Free the memory associated with a DgpAdt_BitArray */
void
DgpAdt_baFree(DgpAdt_BitArray ba) {
  Dgp_Free tmp_free;

  if(ba == NULL) return;
  tmp_free = ba->DGP_FREE;
  if(ba->data) (tmp_free)(ba->data);
  (tmp_free)(ba);
}

/* Return TRUE iff the index'th bit of the DgpAdt_BitArray is set */
Dgp_Bool
DgpAdt_baIsSet(DgpAdt_BitArray ba, Dgp_Int32 index) {
  Dgp_UInt8 u;

  u = ba->data[UCHAR_HOLDING(index)];
  
  return (UCHAR_MASK(u,UCHAR_OFFSET(index)) != 0);
}

/* set the index'th bit of the DgpAdt_BitArray */
void
DgpAdt_baSet(DgpAdt_BitArray ba, Dgp_Int32 index) {
  Dgp_UInt8 u;
  Dgp_Int32 uh;

  uh = UCHAR_HOLDING(index);
  u = ba->data[uh];
  
  ba->data[uh] = UCHAR_W_SET(u,UCHAR_OFFSET(index));
}

/* clear the index'th bit of the DgpAdt_BitArray */
void
DgpAdt_baClear(DgpAdt_BitArray ba, Dgp_Int32 index) {
  Dgp_UInt8 u;
  Dgp_Int32 uh;

  uh = UCHAR_HOLDING(index);
  u = ba->data[uh];
  
  ba->data[uh] = UCHAR_W_CLEAR(u,UCHAR_OFFSET(index));
}
