/* bit_array.h
   Aaron Stump, Fall '96

   This module implements a bit array, which is a packed Dgp_Bool 1D array
   (ie, a 1D array where each element is Dgp_Bool and is stored in one bit).

   This implementation of a bit array is optimized for sparse bit arrays.

   */
   
/* $Header: /project/icons/cvsroot/digipaper/src/libadt/bit_array.h,v 1.7 2000/03/04 00:22:15 mkrocken Exp $ */

#ifndef BIT_ARRAY_H
#define BIT_ARRAY_H

#include <sys/types.h>
#include <stddef.h>
#include "util.h"
#include "dgp_state.h"

#ifdef __cplusplus
extern "C"
{
#endif

typedef struct {
  Dgp_UInt8 *data;
  Dgp_Int32 num_bits;
  Dgp_Int32 first_set, last_set;
  Dgp_State *dgp_state;
} *DgpAdt_BitArray;

#define DGP_NULLBITARRAY (DgpAdt_BitArray)NULL

DGP_LIBEXPORT ErrCode 
DgpAdt_baNew(DgpAdt_BitArray *ba, Dgp_Int32 num_elts, Dgp_State *dgp_state);

DGP_LIBEXPORT void 
DgpAdt_baFree(DgpAdt_BitArray ba);

DGP_LIBEXPORT Dgp_Bool 
DgpAdt_baIsSet(DgpAdt_BitArray ba, Dgp_Int32 index);

DGP_LIBEXPORT void 
DgpAdt_baSet(DgpAdt_BitArray ba, Dgp_Int32 index);

DGP_LIBEXPORT void
DgpAdt_baClear(DgpAdt_BitArray ba, Dgp_Int32 index);

/* some compilers doesn't like braces in these macros... */
#ifndef __GNUC__
#define DGPADT_BAGET_NUMBITS(ba) ((ba)->num_bits)

/* BAGETFIRSTSET() returns the index of the */
#define DGPADT_BAGET_FIRSTSET(ba) ((ba)->first_set)
#define DGPADT_BAGET_LASTSET(ba) ((ba)->last_set)

/* ...but gcc is cool with it */
#else
#define DGPADT_BAGET_NUMBITS(ba) ({(ba)->num_bits;})

/* BAGETFIRSTSET() returns the index of the */
#define DGPADT_BAGET_FIRSTSET(ba) ({(ba)->first_set;})
#define DGPADT_BAGET_LASTSET(ba) ({(ba)->last_set;})
#endif

#ifdef __cplusplus
}
#endif

#endif
