/* Routines to access DgpAdt_BitFiles */

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libadt/bitfile.c,v 1.16 2003/01/27 21:48:28 awilliam Exp $
 *
 */

#ifdef VXWORKS
#include <stdlib.h>
#else
#include <malloc.h>
#endif

#include "bitfile.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libadt/bitfile.c,v 1.16 2003/01/27 21:48:28 awilliam Exp $");

/* buflen is the number of bytes to use in buf */
ErrCode
DgpAdt_bfNew(DgpAdt_BitFile *_f, Dgp_UInt8 *buf, size_t buflen, Dgp_State *dgp_state) 
{
  DGP_PROLOGUE("bfNew()", DGP_PERRSTATE);
  DgpAdt_BitFile f = NULL;

  DGP_CHECK_NULL(_f);

  DGP_CHECK_ALLOC_CLEANUP(f = (DgpAdt_BitFile)(DGP_MALLOC)(sizeof(*f)));

  f->dgp_state = dgp_state;

  f->buflen = f->iobufleft = buflen;
  f->buf = 0;
  f->bitsinbuf = 0;
  /* cur points to the current location in the buffer.
     iobuf points to the start of the buffer. */
  f->cur = f->iobuf = (Dgp_UInt8 *)buf;
  
  *_f = f;
  f = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(f != NULL) (DGP_FREE)(f);

  ERR_RETURN;
}

void 
DgpAdt_bfFree(DgpAdt_BitFile f, size_t *buf_used) {
  Dgp_Free tmp_free = f->DGP_FREE;
  /* return how much of the buffer we actually used (we'll free the
     whole buffer ourselves later). */  
  *buf_used = f->buflen - f->iobufleft;
  (tmp_free)((void *)f);
}

/* bfNewBuffer()

   Plug in a new buffer to this DgpAdt_BitFile. 

   Parameters:
   f          -- the DgpAdt_BitFile
   new_buf    -- a pointer to the new buffer
   new_buflen -- the size of the new buffer

   NOTE that you can use this function together with realloc() to
   realloc a DgpAdt_BitFile. 
   NOTE that the new buffer could possibly be smaller than the
   old one and everything should still work. */
void 
DgpAdt_bfNewBuffer(DgpAdt_BitFile f, Dgp_UInt8 *new_buf, size_t new_buflen) 
{
  size_t num_new_bytes;

  /* set cur and iobuf.  NOTE that since cur points to the current location
     in the old buffer, we have to do a little work to get it to poDgp_Int32 to
     the corresponding location in the new buffer. */
  f->cur = &new_buf[f->cur - f->iobuf];
  f->iobuf = new_buf;

  /* set buflen and iobufleft */
  num_new_bytes = new_buflen - f->buflen;
  f->iobufleft += num_new_bytes;
  f->buflen = new_buflen;
}

/*
 * Write the given code, held in the *low-order* codelen bits of code
 * to the given file. Return TRUE on success, FALSE on failure. codelen
 * can be up to 32. code is written MSB-to-LSB. The non-significant
 * bits of code (those not the codelen LSBs) must be zero.
 */
ErrCode
DgpAdt_bfWriteCode(Dgp_UInt32 code, Dgp_Int32 codelen, DgpAdt_BitFile f) 
{
  DGP_PROLOGUE("bfWriteCode()", f->DGP_PERRSTATE);
  DGP_ASSERT(f != (DgpAdt_BitFile)NULL);

  /* we won't be able to fit this code into our buffer */
  DGP_ERR_GEN((f->iobufleft < (Dgp_UInt32)((codelen + DGP_BITS_UINT8 - 1) / DGP_BITS_UINT8)), cErrIO_WriteError);

  while (codelen) {

    /* Can we fit into the current buf? */
    if (f->bitsinbuf + codelen <= DGP_BITS_UINT8) {
      f->buf |= (code << (DGP_BITS_UINT8 - f->bitsinbuf - codelen));
      f->bitsinbuf += codelen;
      codelen = 0;
    }
    else {
      /* Put as much as we can into the buffer */
      f->buf |= (code >> (codelen - (DGP_BITS_UINT8 - f->bitsinbuf)));
      codelen -= DGP_BITS_UINT8 - f->bitsinbuf;
      f->bitsinbuf = DGP_BITS_UINT8;
      /* Mask off what we've got left */
      code &= (1ul << codelen) - 1;
    }

    if (f->bitsinbuf == DGP_BITS_UINT8) {

      DGP_ERR_GEN((!f->iobufleft), cErrIO_WriteError);

      *f->cur++ = f->buf;
      f->iobufleft--;
      f->buf = 0;
      f->bitsinbuf = 0;
    }
  }

  NORMAL_RETURN;
}

/*
 * Write out anything in the buffer.
 */
ErrCode
DgpAdt_bfFlush(DgpAdt_BitFile f)
{
  DGP_PROLOGUE("bfFlush()", f->DGP_PERRSTATE);
    DGP_ASSERT(f != (DgpAdt_BitFile )NULL);

    if (f->bitsinbuf) {
	    DGP_ERR_GEN((!f->iobufleft), cErrIO_WriteError);
	    *f->cur++ = f->buf;
	    f->iobufleft--;
	  }
    f->buf = 0;
    f->bitsinbuf = 0;

    NORMAL_RETURN;
}

/*
 * Write out a value to the file. The encoding is:
 * Value	Code
 * 0-127	0+(value as 7 bits)
 * 128-1151	10+(value-128 as 10 bits)
 * 1152-32767	11+(value as 15 bits)
 * 32768-	1100000+(value as 32 bits)
 */
ErrCode
DgpAdt_bfWriteValue(Dgp_UInt32 x, DgpAdt_BitFile f)
{
  DGP_PROLOGUE("DgpAdt_bfWriteValue()", f->DGP_PERRSTATE);
    /* DGP_DEBUG("wv(%08x)\n", x); */
  if (x < 128) {
#ifdef	INSTRUMENT
	*(f->nbits) += 8;
#endif
	  DGP_ERR_PASS(DgpAdt_bfWriteCode(0, 1, f));		/* 0 */
	  DGP_ERR_PASS(DgpAdt_bfWriteCode(x, 7, f));		/* value as 7 bits */
	}
  else if (x < 1152) {
#ifdef	INSTRUMENT
	*(f->nbits) += 12;
#endif
	  DGP_ERR_PASS(DgpAdt_bfWriteCode(2, 2, f));		    /* 10 */
	  DGP_ERR_PASS(DgpAdt_bfWriteCode(x-128, 10, f));	/* value-128 as 10 bits */
	}
  else if (x < 32768) {
#ifdef	INSTRUMENT
	*(f->nbits) += 17;
#endif
	  DGP_ERR_PASS(DgpAdt_bfWriteCode(3, 2, f));		  /* 11 */
	  DGP_ERR_PASS(DgpAdt_bfWriteCode(x, 15, f));		/* value as 15 bits */
	}
  else {
#ifdef	INSTRUMENT
	*(f->nbits) += 39;
#endif
	  DGP_ERR_PASS(DgpAdt_bfWriteCode(0x60, 7, f));		/* 1100000 */
	  DGP_ERR_PASS(DgpAdt_bfWriteCode(x, 32, f));		  /* value as 32 bits */
	}

  NORMAL_RETURN;
}

/*
 * Read a few bits from a DgpAdt_BitFile. Pack them into the codelen LSBs
 * of *code. Return TRUE on success, FALSE on failure.
 *
 * The f->bitsinbuf LSBs of f->buf are the valid ones, since it gets
 * read out from the MSB down.
 * 
 * 22.8.2000 / dhtesq.
 * Profiling indicates this is a good place for an inline.
 *
 */
#ifdef _WIN32
DGP_LIBEXPORT
#endif
INLINE ErrCode
DgpAdt_bfReadCode(Dgp_UInt32 *code, Dgp_Int32 codelen, DgpAdt_BitFile f)
{
  DGP_PROLOGUE("bfReadCode()", f->DGP_PERRSTATE);
    DGP_ASSERT(code != (Dgp_UInt32 *)NULL);
    DGP_ASSERT(f != (DgpAdt_BitFile )NULL);

    *code = 0;
    while (codelen) {
	if (f->bitsinbuf >= codelen) {
	    /* Enough in the buffer for us. */
	    *code |= (f->buf >> (f->bitsinbuf - codelen)) &
		((1ul << codelen) - 1);
	    f->bitsinbuf -= codelen;
	    NORMAL_RETURN;
	    }
	/* Eat what's in the buffer and get more */
	if (f->bitsinbuf) {
	    *code |= (f->buf & ((1ul << f->bitsinbuf) - 1)) <<
		(codelen - f->bitsinbuf);
	    codelen -= f->bitsinbuf;
	    }
	DGP_ERR_GEN((!f->iobufleft), cErrIO_ReadError);
	f->iobufleft--;
	f->buf = *f->cur++;
	f->bitsinbuf = DGP_BITS_UINT8;
	}
    NORMAL_RETURN;
}

/* Read n bits into *v, bailing in case of failure */
#define	readChecking(v, n, f)				\
do {						\
  DGP_ERR_PASS(DgpAdt_bfReadCode((v), (n), (f)));\
} while(0)

/*
 * Read a value from the file.
 */
ErrCode
DgpAdt_bfReadValue(Dgp_UInt32 *x, DgpAdt_BitFile f)
{
  DGP_PROLOGUE("bfReadValue()", f->DGP_PERRSTATE);
  Dgp_UInt32 v;

  readChecking(&v, 1, f);
  if (v == 0) {				/* 0 */
	  readChecking(&v, 7, f);			/* value as 7 bits */

	  *x = v;
#ifdef  INSTRUMENT
    *(f->nbits) += 8;
#endif
	}
  else {
	  readChecking(&v, 1, f);
	  if (!v) {				/* 10 */
	      readChecking(&v, 10, f);		/* value-127 as 10 bits */
	      *x = v + 128;
#ifdef  INSTRUMENT
        *(f->nbits) += 12;
#endif
	  }
	  else {					/* 11 */
	    readChecking(&v, 15, f);		/* value as 15 bits */
	    
	    if (v < 1024) {
		    /*
		     * The file really contains a 32-bit value. We just read
		     * the top 10 bits of it (the first 5 of the 15 we got were
		     * zero to flag us to this).
		     */
		    *x = v << 22;
		    readChecking(&v, 22, f);
		    *x |= v;
#ifdef  INSTRUMENT
        *(f->nbits) += 39;
#endif
		  }
	    else {
		    *x = v;
#ifdef  INSTRUMENT
        *(f->nbits) += 17;
#endif
		  }
	  }
	}

    /* DGP_DEBUG("rv(%08x)\n", *x); */
    NORMAL_RETURN;
}
