/* cache_table.h
   
   Aaron Stump, Summer '97

   This file contains the specification for the DgpAdt_CacheTable data structure.
   The DgpAdt_CacheTable caches data (somewhat cryptically referred to as elements
   in much of the code) in memory.  The DgpAdt_CacheTable has a mechanism for
   allowing the user to define a correspondence between keys and data
   to be cached.  The user asks a DgpAdt_CacheTable for the data associated (via
   the user's own correspondence function) with a given key.  If that data
   is already in memory, the DgpAdt_CacheTable returns it.  Otherwise, it calls
   a user-specified function to get the data. There is also a mechanism
   for flushing data from the cache if it is not referenced often enough
   (where again, "often enough" is user-specified). */

/* $Header: /project/icons/cvsroot/digipaper/src/libadt/cache_table.h,v 1.9 2000/03/04 00:22:16 mkrocken Exp $ */

#ifndef CACHE_TABLE_H
#define CACHE_TABLE_H

#include <stddef.h>
#include "hash_table.h"
#include "list.h"
#include "dgp_state.h"

#ifdef __cplusplus
extern "C"
{
#endif

typedef void * (*DgpAdt_ctElementGetter)(void *, void *);
typedef void   (*DgpAdt_ctElementDisposer)(void *, void *);

typedef struct {
  DgpAdt_HashTable ht;
  DgpAdt_List elts;    /* a DgpAdt_List of elements in the DgpAdt_HashTable */

  DgpAdt_htComparator isKeyEqual;
  DgpAdt_htHash hashKey;
  DgpAdt_htLastCall last_call;
  DgpAdt_ctElementGetter getter;
  DgpAdt_ctElementDisposer disposer;

  void *user_cookie;
  Dgp_Int32 flush_after;
  Dgp_State *dgp_state;
} *DgpAdt_CacheTable;

DGP_LIBEXPORT ErrCode 
DgpAdt_ctNew(DgpAdt_CacheTable *ct,
      void *user_cookie, 
      DgpAdt_htComparator isKeyEqual, 
	  DgpAdt_htHash hashKey,
      DgpAdt_htLastCall last_call, 
	  DgpAdt_ctElementGetter getter,
      DgpAdt_ctElementDisposer disposer,
      Dgp_Int32 flush_after,
      Dgp_State *dgp_state);

DGP_LIBEXPORT void       
DgpAdt_ctFree(DgpAdt_CacheTable ct);

DGP_LIBEXPORT ErrCode
DgpAdt_ctGetElement(DgpAdt_CacheTable ct, void *key, void **key_in_table, void **element);

DGP_LIBEXPORT void
DgpAdt_ctTock(DgpAdt_CacheTable ct);
  

#ifdef __cplusplus
}
#endif

#endif
