
/* */
/* NOTE: REMOVING FROM THIS HASHTABLE DOES NOT WORK. */
/* I'll fix it if I get around to it. For now, noone does  */
/* removes anyway. Actually, I don't think anyone even uses */
/* this code except my halftone stuff. We have no */
/* idea who originally wrote this buggy hash-table. */
/* */
/* */
/* 1.8.2000 / dhtesq. */
/* I'm changing all the shorts in here to longs, particularly for the result */
/* of the hash operation, so I can hash with 32-bit pointers... */


#include "hash_table.h"
#include "sm_image.h"
#include "PsgFix.h"

#ifdef VXWORKS
#include <stdlib.h>
#else
#include <malloc.h>
#endif

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libadt/hash_table.c,v 1.18 2003/02/10 22:53:18 mkrocken Exp $");

static Dgp_Bool 
generic_isequal(void *cookie, DgpAdt_ht_elt e1, DgpAdt_ht_elt e2) {

  (void)cookie;

  return (e1 == e2);
}

static Dgp_UInt16
generic_hash(void *cookie, DgpAdt_ht_elt e) {
  Dgp_UInt32 l = (Dgp_UInt32)e;

  (void)cookie;

  return (Dgp_UInt16)(l % ((DgpAdt_HashTable)cookie)->expected_size);
}

/* htNew()
   
   Allocate a new DgpAdt_HashTable.

   Params:
   void *user_cookie    -- a state cookie that will be handed to isequal()
                           every time it's called, and to hash() every time
			   it's called (this is the user's own cookie, which
			   can be whatever s/he wants, including NULL).
   htComparator isequal -- a function that returns TRUE iff two DgpAdt_ht_elt's
                           are to be considered equal.
   htHash hash          -- a function that returns an Dgp_UInt32 that
                           gives the hash for a given DgpAdt_ht_elt.
   htLastCall last_call -- a function to be called exactly once by htFree().
                           The argument to last_call() will be the 
			   user_cookie.

   Any or all of the above functions can be NULL.  If isequal or hash
   is NULL, generic htComparator and htHash functions will be supplied.
   If last_call is NULL, no such function will be called by htFree().
*/
ErrCode 
DgpAdt_htNew(DgpAdt_HashTable *_ht, void *user_cookie, DgpAdt_htComparator isequal, DgpAdt_htHash hash, 
      DgpAdt_htLastCall last_call, Dgp_UInt16 expected_size, Dgp_State *dgp_state) 
{      
  DGP_PROLOGUE("htNew()", DGP_PERRSTATE);
  DgpAdt_HashTable ht = NULL;
  
  DGP_CHECK_NULL(_ht);

  DGP_CHECK_ALLOC_CLEANUP(ht = (DgpAdt_HashTable)(DGP_CALLOC)(1, sizeof(*ht)));

  ht->dgp_state = dgp_state;
  
  if(hash == (DgpAdt_htHash)NULL) {
    ht->hash = generic_hash;
    ht->hash_cookie = ht;
  } else {
    ht->hash = hash;
    ht->hash_cookie = user_cookie;
  }

  ht->isequal = ((isequal == (DgpAdt_htComparator)NULL) ? generic_isequal : isequal);

  ht->last_call = last_call;
  ht->cookie = user_cookie;

  ht->expected_size = ((expected_size <= 0) ? 100 : expected_size);

  DGP_ERR_PASS_CLEANUP(DgpAdt_smNew(&(ht->sm), DGP_TRUE, ht->expected_size, dgp_state));
  
  DgpAdt_htResetGetNextElement(ht);

  *_ht = ht;
  ht = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(ht != NULL) {
    if(ht->sm != NULL) DgpAdt_smFree(ht->sm);
    (DGP_FREE)(ht);
  }

  ERR_RETURN;
}

void
DgpAdt_htFree(DgpAdt_HashTable ht) {
  Dgp_Free tmp_free;
  
  if(ht == NULL) return;

  tmp_free = ht->DGP_FREE;

  DgpAdt_smFree(ht->sm);
  if (ht->last_call)
    ht->last_call(ht->cookie);
  (tmp_free)((void *)ht);
}

/* htIsInTable()

   Is the given piece of data in the given DgpAdt_HashTable?

   Parameters:
   DgpAdt_HashTable ht      -- the DgpAdt_HashTable in which to look for the given piece of
                        data
   DgpAdt_ht_elt data       -- the data to look for
   DgpAdt_ht_elt *in_table  -- if an element e of the given table is equal to data
                        (according to the isequal function associated by the
			user with this DgpAdt_HashTable) in the appropriate hash bin
			(as specified by the hash function associated by the
			user with this DgpAdt_HashTable), *in_table will be filled in
			with e. Otherwise, *in_table will be set to NULL.
			
   Guarantee:
   In a call to htIsInTable(), the following will hold:
   1.   ht->hash() will be called no more than once, 
   2.   any calls to ht->isequal() will follow a call to ht->hash(), and
   3.   DgpAdt_ht_elt data (the param) will always be the first argument to
        ht->isequal().

   Return value: 
   This function returns FALSE iff there was some kind of internal error
   (failure of memory allocation in the underlying SMImage, eg).

  2.8.2000 / dhtesq.
  Modified to return ErrCode...
 */
ErrCode
DgpAdt_htIsInTable(DgpAdt_HashTable ht, DgpAdt_ht_elt data, DgpAdt_ht_elt *in_table) {
  DGP_PROLOGUE("htIsInTable()", ht->DGP_PERRSTATE);
  DgpAdt_SMImage sm = ht->sm;
  Dgp_UInt16 y,x;
  DgpAdt_sm_elt_type e;

  DGP_STOP_WHINE;
  DGP_ASSERT(ht != (DgpAdt_HashTable)NULL);
  DGP_ASSERT(data != (DgpAdt_ht_elt)NULL); /* we're not allowing users to add NULL
				   to the table */

  y = ht->hash(ht->hash_cookie, data);
  x = 0;
  
  *in_table = NULL;
  for(;;) {
    /* We're guaranteed to bail from this loop eventually, because we cause no
       additional memory to be allocated by the underlying sm_image module
       and so eventually, as we increment x, we'll hit a part of this
       bin which has not been allocated . */

    if (!DgpAdt_smIsAllocated(sm,x,y))
      /* we haven't allocated this part of this bin of the hash table,
	 so we certainly haven't stuck this piece of data in it! */
      NORMAL_RETURN;
    DGP_ERR_CHECK_NO_MSG(DgpAdt_smGet(sm,x,y,&e)) {
      ERR_RETURN;
    }
    if (e == NULL)
      /* all bins have non-NULL values for pieces of data added by
	 the user, and since all the user's data is stored in the locations
	 between that of x == 0 and some x_end, a NULL value in a
	 bin indicates that we can quit searching this bin */
      NORMAL_RETURN;
    if (ht->isequal(ht->cookie,data,e)) {
      /* we found the user's data! */
      *in_table = e;
      NORMAL_RETURN;
    }
    x++;
  }
  
  ERR_RETURN; /* we should never get to this line,
		   but gcc will complain if we don't return
		   something at the end of the function! */
}

/* htAddToTable()
   
   Add the given piece of user's data to the table (if an element in the
   table matches it, do nothing).
   
   Parameters:
   DgpAdt_HashTable ht      -- the DgpAdt_HashTable to add the data to
   DgpAdt_ht_elt data       -- the data to add.
   DgpAdt_ht_elt *in_table  -- if an element e of the given table is equal to data
                        (according to the isequal function associated by the
			user with this DgpAdt_HashTable) in the appropriate hash bin
			(as specified by the hash function associated by the
			user with this DgpAdt_HashTable), *in_table will be filled in
			with e. Otherwise, *in_table will be set to NULL.
   Guarantee:
   In a call to htAddToTable(), the following will hold:
   1.   ht->hash() will be called no more than once, and
   2.   any calls to ht->isequal() will follow a call to ht->hash().
   3.   DgpAdt_ht_elt data (the param) will always be the first argument to
        ht->isequal().

   Return value:
   FALSE iff there was some kind of internal error.
*/
ErrCode
DgpAdt_htAddToTable(DgpAdt_HashTable ht, DgpAdt_ht_elt data, DgpAdt_ht_elt *in_table) 
{
  DGP_PROLOGUE("htAddToTable()", ht->DGP_PERRSTATE);
  DgpAdt_SMImage sm = ht->sm;
  Dgp_UInt16 y,x;
  DgpAdt_sm_elt_type e;

  DGP_ASSERT(ht   != (DgpAdt_HashTable)NULL);
  DGP_ASSERT(data != (DgpAdt_ht_elt)NULL); /* we're not allowing users to add NULL
				   to the table */
  y = ht->hash(ht->hash_cookie, data);
  x = 0;
  
  *in_table = NULL;
  
  for(;;) {
    /* We're guaranteed to bail from this loop eventually, because we cause no
       additional memory to be allocated by the underlying sm_image module
       and so eventually, as we increment x, we'll hit a part of this
       bin which has not been allocated . */
    
    /* If the appropriate part of the table hasn't even been allocated
       yet, stick the data there. */
    if (!DgpAdt_smIsAllocated(sm,x,y)) {
      /* we haven't allocated this part of this bin of the hash table,
	       so we certainly haven't stuck this piece of data in it! */
      DGP_ERR_PASS(DgpAdt_smSet(sm,x,y,data));
      NORMAL_RETURN;
    }

    /* If the appropriate part of the table has been allocated but has
       a NULL DgpAdt_ht_elt for its value, stick the data there. */
    DGP_ERR_PASS(DgpAdt_smGet(sm,x,y,&e));
    if (e == NULL) {
      /* all bins have non-NULL values for pieces of data added by
	 the user, and since all the user's data is stored in the locations
	 between that of x == 0 and some x_end, a NULL value in a
	 bin indicates that we can quit searching this bin */
      DGP_ERR_PASS(DgpAdt_smSet(sm,x,y,data));
      NORMAL_RETURN;
    }
    
    /* If the data matches the value at the current position in the table,
       we can quit searching. */
    if (ht->isequal(ht->cookie,data,e)) {
      /* we found something matching the user's data! */
      *in_table = e;
      NORMAL_RETURN;
    }
    x++;
  }

  ERR_RETURN; /* we should never get to this line,
		   but gcc will complain if we don't return
		   something at the end of the function! */
}

/* htRemoveFromTable()
   
   Remove the element matching the given piece of user's data from the table
   (if there is no matching element, do nothing).
   
   Parameters:
   DgpAdt_HashTable ht      -- the DgpAdt_HashTable to remove the data from.
   DgpAdt_ht_elt data       -- remove the element matching DgpAdt_ht_elt data.
   DgpAdt_ht_elt *in_table  -- iff an element e of the given table is equal to data
                        (according to the isequal function associated by the
			user with this DgpAdt_HashTable) in the appropriate hash bin
			(as specified by the hash function associated by the
			user with this DgpAdt_HashTable), *in_table will be filled in
			with e. Otherwise, *in_table will be set to NULL.
   Guarantee:
   In a call to htRemoveFromTable(), the following will hold:
   1.   ht->hash() will be called no more than once, and
   2.   any calls to ht->isequal() will follow a call to ht->hash().
   3.   DgpAdt_ht_elt data (the param) will always be the first argument to
        ht->isequal().

   Return value:
   FALSE iff there was some kind of internal error.
*/
ErrCode
DgpAdt_htRemoveFromTable(DgpAdt_HashTable ht, DgpAdt_ht_elt data, DgpAdt_ht_elt *in_table) 
{
  DGP_PROLOGUE("htRemoveFromTable()", ht->DGP_PERRSTATE);
  DgpAdt_SMImage sm = ht->sm;
  Dgp_UInt16 y,x;
  DgpAdt_sm_elt_type e;

  DGP_ASSERT(ht   != (DgpAdt_HashTable)NULL);
  DGP_ASSERT(data != (DgpAdt_ht_elt)NULL); /* we're not allowing users to add NULL
				   to the table */
  y = ht->hash(ht->hash_cookie, data);
  x = 0;
  
  *in_table = NULL;
  
  for(;;) {
    /* We're guaranteed to bail from this loop eventually, because we cause no
       additional memory to be allocated by the underlying sm_image module
       and so eventually, as we increment x, we'll hit a part of this
       bin which has not been allocated . */
    
    /* If the appropriate part of the table hasn't even been allocated
       yet, stick the data there. */
    if (!DgpAdt_smIsAllocated(sm,x,y)) 
      /* we haven't allocated this part of this bin of the hash table,
	       so we certainly haven't stuck this piece of data in it! */
      NORMAL_RETURN;

    DGP_ERR_PASS(DgpAdt_smGet(sm,x,y,&e));

    if (e == NULL) 
      /* all bins have non-NULL values for pieces of data added by
	       the user, and since all the user's data is stored in the locations
	       between that of x == 0 and some x_end, a NULL value in a
	       bin indicates that we can quit searching this bin */
      NORMAL_RETURN;
    
    /* If the data matches the value at the current position in the table,
       we can quit searching. */
    if (ht->isequal(ht->cookie,data,e)) {
      /* we found something matching the user's data! */
      
      *in_table = e;
      /* remove the reference to the user's data in the table */
      DGP_ERR_PASS(DgpAdt_smSet(sm,x,y,NULL));
      NORMAL_RETURN;
    }
    x++;
  }
  ERR_RETURN; /* we should never get to this line,
		   but gcc will complain if we don't return
		   something at the end of the function! */
}

/* htResetGetNextElement()
   
   After calling this function, calling htGetNextElement() until
   it returns NULL will give you all the elements in the DgpAdt_HashTable.

   This function is called by ntNew(), so the first time you try stepping
   through the DgpAdt_HashTable with htGetNextElement(), you'll be able to
   get all the elements.  After you step through it at least partway, though,
   you'll need to call htResetGetNextElement() to be able to step all the way
   through it again.
*/
void
DgpAdt_htResetGetNextElement(DgpAdt_HashTable ht) {
  ht->next_y_to_get = 0;
  ht->next_x_to_get = 0;
}

/* htGetNextElement()
   
   Each time this function is called, it gets the next piece of data
   which has not been returned by a call to htGetNextElement() since the
   last call to htResetGetNextElement() (or htNew(), since htNew() calls
   htResetGetNextElement().)  It passes back that piece of data in *in_table.
   If there is no such next piece of data, it passes back NULL in *in_table.

   Return value:
   FALSE if there was an internal error, and TRUE otherwise. */
ErrCode 
DgpAdt_htGetNextElement(DgpAdt_HashTable ht, DgpAdt_ht_elt *in_table) 
{
  DGP_PROLOGUE("htGetNextElement()", ht->DGP_PERRSTATE);

#define goto_next_line() \
  ht->next_y_to_get = DgpAdt_smGetNextY(sm,(Dgp_UInt16)ht->next_y_to_get);\
  if (ht->next_y_to_get > ymax)\
    /* we've gone through all the elements we have (smGetNextY() returns\
       the same*/\
    NORMAL_RETURN;\
  ht->next_x_to_get = 0;\

  DgpAdt_SMImage sm = ht->sm;
  DgpAdt_sm_elt_type e = NULL;
  int ymax;


  *in_table = NULL;
  ymax = DGPADT_SMGET_YMAX(sm);

  for (;;) {
    /* This loop is guaranteed to terminate, since there is only a finite
       number of slots in our DgpAdt_SMImage which we haven't yet checked for
       user's data, and this loop checks one such slot (and advances to the
       next) every time it runs through completely. */

    if (!DgpAdt_smIsAllocated(sm,(Dgp_UInt16)ht->next_x_to_get,(Dgp_UInt16)ht->next_y_to_get)) {
      /* nothing at this (x,y) position */
      goto_next_line()
    }  
    else {
      /* this position is allocated in our DgpAdt_SMImage */
      DGP_ERR_PASS(DgpAdt_smGet(sm,(Dgp_UInt16)ht->next_x_to_get,(Dgp_UInt16)ht->next_y_to_get,&e));
      if (e != NULL) {
	      /* we found the next item of user's data! */
	      *in_table = e;
	      ++ht->next_x_to_get;
	      NORMAL_RETURN;
      }

      /* e == NULL, so hop to the next line */
      goto_next_line();

    } /* end smIsAllocated() */
  } /* end for (;;) */

  /* we should never get here */
  ERR_RETURN; 
}
