/*
*
*	$Header: /project/icons/cvsroot/digipaper/src/libadt/list.c,v 1.6 2003/01/27 21:48:29 awilliam Exp $
*
* Copyright (c) 1990, 1991, 1992, 1993 Cornell University.  All Rights
* Reserved.
*
* Copyright (c) 1991, 1992 Xerox Corporation.  All Rights Reserved.
*
* Use, reproduction and preparation of derivative works of this software is
* permitted.  Any copy of this software or of any derivative work must
* include both the above copyright notices of Cornell University and Xerox
* Corporation and this paragraph.  This software is made available AS IS, and
* XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING
* WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
* FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED
* HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
* EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
* NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED OF
* THE POSSIBILITY OF SUCH DAMAGES.
*/
/*
* Xerox may use this code without restriction.
*/

/*
*	list.c - Yet Another DgpAdt_List Module
*
* How many times have you seen this implemented? I know I've implemented it
* a few too many times...
*
* This module handles generic lists.
*
* Exports:
*	type DgpAdt_List
*	type DgpAdt_ListNode
*
*	constant DgpAdt_List DGP_NULLLIST
*	constant DgpAdt_ListNode DGP_NULLLISTNODE
*
*	DgpAdt_List liNew(void) - create a new, empty list
*
*	void liFree(DgpAdt_List l) - free list l
*
*	void liClear(DgpAdt_List l) - free all elements of list l (but not the list itself)
*
*	DgpAdt_ListNode liFirst(DgpAdt_List l) - return the first entry in list l
*
*	DgpAdt_ListNode liLast(DgpAdt_List l) - return the last entry in list l
*
*	DgpAdt_ListNode liNext(DgpAdt_ListNode ln) - return the next entry after
*		ln in whatever list ln is in
*
*	DgpAdt_ListNode liPrev(DgpAdt_ListNode ln) - return the previous entry before
*		ln in whatever list ln is in
*
*	DgpAdt_ListNode liAdd(DgpAdt_List l, void *userdata) - append a node containing
*		userdata to list l
*
*	DgpAdt_ListNode liAddAfter(DgpAdt_List l, DgpAdt_ListNode ln, void *userdata) - add a node
*		containing userdata to l's after ln.
*
*	void liRem(DgpAdt_ListNode ln) - remove node ln from whatever list it
*		is in
*
*	DgpAdt_ListNode liRemAndNext(DgpAdt_ListNode ln) - remove node ln from whatever list
*		it is in, and return the DgpAdt_ListNode which followed it.
*
*	void *liGet(DgpAdt_ListNode ln) - return the userdata of node ln.
*
*	void liPut(DgpAdt_ListNode ln, void *userdata) - change the userdata of
*		node ln.
*
*	DgpAdt_ListNode liIsIn(DgpAdt_List l, void *userdata) - search l for a node
*		containing that userdata value
*
*	int liLen(DgpAdt_List l) - return the length of l
*
*	void liApp(DgpAdt_List l1, DgpAdt_List l2) - append l2 to l1
*
*	macro foreach(ln, l) - iterate over each node in a list
*
*	macro forafter(ln1, ln2) - iterate over each node in a list after
*		a given node
*
*	macro forfrom(ln1, ln2) - iterate over each node in a list after
*		(and including) a given node
*
* An object of type DgpAdt_List is used as a handle on a list, which consists of
*	multiple DgpAdt_ListNodes. Each DgpAdt_ListNode contains a void *
*	field, which is used for whatever data the user wishes to attach
*	to that node. It is standard C idiom that (void *) pointers can
*	be cast to and from any other type of pointer without error or loss.
*
* liNew creates an empty list. If this is not possible, it returns NULLLIST.
*
* liFree frees the storage associated with a DgpAdt_List and the DgpAdt_ListNodes it
*	contains. Note that it does not free the userdata; the user
*	is responsible for doing this.
*
* liClear frees storage for all the nodes in the list, but does not free the list
* itself.  Again, note that it does not free the userdata; the user
*	is responsible for doing this.
*
* liFirst returns the first entry of the given list. If the list is
*	empty, it returns DGP_NULLLISTNODE.
*
* liLast returns the last entry of the given list. If the list is
*	empty, it returns DGP_NULLLISTNODE.
*
* liNext returns the DgpAdt_ListNode after the given one. A call to liFirst
*	followed by multiple calls to liNext will return all entries
*	in that list, in the order in which they were added. When liNext
*	is called on the last entry of a list, it returns DGP_NULLLISTNODE.
*
* liPrev returns the DgpAdt_ListNode before the given one.  When liPrev
*	is called on the first entry of a list, it returns DGP_NULLLISTNODE.
*
* liAdd appends a node to the given list. The userdata pointer is
*	stored in that node. If it is not possible to append to that
*	list, it returns DGP_NULLLISTNODE.
*
* liAddAfter adds a node to the given list after the given node. If
*	ln is DGP_NULLLISTNODE, then it adds to the beginning of the list.
*	If it can't add the node, it returns DGP_NULLLISTNODE; otherwise, it
*	returns the new node.
*
* liRem removes a node from the list it is in.
*
* liRemAndNext removes a node from the list it is in, and returns the
*	DgpAdt_ListNode that followed it. If the DgpAdt_ListNode was the last in its list,
*	liRemAndNext returns DGP_NULLLISTNODE. This is really a convenience
*	function: it makes it possible to step through a list removing some
*	elements without having two pointers, or worrying about accessing
*	freed memory.
*
* liGet gets the userdata associated with the given node.
*
* liPut modifies the userdata associated with the given node.
*
* liIsIn searches the userdata fields of a list and returns where
*	that field appeared in the list. This is the only routine which
*	does anything to the userdata fields other than copy them. liIsIn
*	compares them for equality. If the field does not appear in the list,
*	it returns DGP_NULLLISTNODE.
*
* liLen counts the number of nodes which are on the given list.
*
* liApp appends the list l2 to the list l1. They must not be the same list.
*	l2 becomes invalid after the call.
*
* foreach is a replacement for the C "for" construct which successively
*	assigns ln to each node in l.
*
* forafter is a replacement for the C "for" construct which successively
*	assigns ln1 to each node in the list containing ln2, starting with
*	the successor of ln2 (liNext(ln2)). ln1 and ln2 may be the same
*	variable.
*
* forfrom is like forafter, except that the loop starts at ln2, rather than
*	its successor.
*
* Note that all DgpAdt_ListNodes must be part of some list. When a DgpAdt_ListNode is
* removed from a list, or when its list is freed, it becomes invalid.
* When a DgpAdt_List is freed, it becomes invalid.
*
* DgpAdt_ListNodes are persistent: once one is created, it will refer to the
* same element for the life of that element, even if the list containing
* that element is merged with another list (via liApp); it becomes invalid
* only as a result of a liFree or liRem (or liRemAndNext).
*
* Some of these functions may be implemented as macros for speed reasons.
*/

#include "util.h"
#include "list.h"
#include <stdlib.h>
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libadt/list.c,v 1.6 2003/01/27 21:48:29 awilliam Exp $");

/* the small pool ids for the list allocations */
#define LIST_POOL_ID DGP_SMALLPOOL_MIN16

#define LIST_SEC_MALLOC(_ppl) \
  DGP_ERR_PASS(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, LIST_POOL_ID, (void **)(_ppl)))
#define LIST_FREE(_pl) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, LIST_POOL_ID, (void *)(_pl)))

#define LN_POOL_ID DGP_SMALLPOOL_MIN16

#define LN_SEC_MALLOC(_ppln) \
  DGP_ERR_PASS(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, LN_POOL_ID, (void **)(_ppln)))
#define LN_FREE(_pln) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, LN_POOL_ID, (void *)(_pln)))


ErrCode
DgpAdt_liNew(DgpAdt_List *retlist, Dgp_State *dgp_state)
{
  DgpAdt_List newList;
  
  DGP_PROLOGUE ("DgpAdt_liNew", DGP_PERRSTATE);
  
  /* check small pool usage */
  DGP_SMALLPOOL_VALIDATE(struct _DgpAdt_List, LIST_POOL_ID);
  DGP_SMALLPOOL_VALIDATE(struct _DgpAdt_ListNode, LN_POOL_ID);
  
  LIST_SEC_MALLOC(&newList);
  
  newList->first = newList->last = DGP_NULLLISTNODE;
  newList->dgp_state = dgp_state;
  newList->nelt = 0;
  *retlist = newList;
  NORMAL_RETURN;
}


ErrCode
DgpAdt_liClear(DgpAdt_List l)
{
  DgpAdt_ListNode ln, nextln;
  Dgp_State *dgp_state = l->dgp_state;
  
  DGP_PROLOGUE ("DgpAdt_liFree", DGP_PERRSTATE);
  DGP_STOP_WHINE;
  
  /* Free all the DgpAdt_ListNodes this list owns */
  for (ln = l->first; ln != DGP_NULLLISTNODE; ln = nextln) {
    DGP_ASSERT(ln->owner == l);
    
    nextln = ln->next;
    LN_FREE(ln);
  }

  /* reset the list */
  l->first = l->last = DGP_NULLLISTNODE;
  l->nelt = 0;
  
  NORMAL_RETURN;
}


ErrCode
DgpAdt_liFree(DgpAdt_List l)
{
  DgpAdt_ListNode ln, nextln;
  Dgp_State *dgp_state = l->dgp_state;
  
  DGP_PROLOGUE ("DgpAdt_liFree", DGP_PERRSTATE);
  DGP_STOP_WHINE;
  
  /* Free all the DgpAdt_ListNodes this list owns */
  for (ln = l->first; ln != DGP_NULLLISTNODE; ln = nextln) {
    DGP_ASSERT(ln->owner == l);
    
    nextln = ln->next;
    LN_FREE(ln);
  }
  
  /* and the DgpAdt_List structure itself */
  LIST_FREE(l);
  
  NORMAL_RETURN;
}

ErrCode
DgpAdt_liAdd(DgpAdt_List l, void *userdata, DgpAdt_ListNode *retnode)
{
  DgpAdt_ListNode newln;
  Dgp_State *dgp_state = l->dgp_state;
  
  DGP_PROLOGUE ("DgpAdt_liAdd", DGP_PERRSTATE);
  
  DGP_ASSERT(l != DGP_NULLLIST);
  
  /* Get a new DgpAdt_ListNode */
  LN_SEC_MALLOC(&newln);
  
  /* and fill it out */
  newln->next = DGP_NULLLISTNODE;
  newln->prev = l->last;
  newln->owner = l;
  newln->userdata = userdata;
  
  /* Hook it in */
  if (l->last != DGP_NULLLISTNODE) {
    l->last->next = newln;
    l->last = newln;
  }
  else {
    DGP_ASSERT(l->first == DGP_NULLLISTNODE);
    l->first = l->last = newln;
  }
  
  l->nelt++;
  if(retnode != NULL) *retnode = newln;
  
  NORMAL_RETURN;
}

ErrCode
DgpAdt_liAddAfter(DgpAdt_List l, DgpAdt_ListNode ln, void *userdata, DgpAdt_ListNode *retnode)
{
  DgpAdt_ListNode newln;
  Dgp_State *dgp_state = l->dgp_state;
  
  DGP_PROLOGUE ("DgpAdt_liAddAfter", DGP_PERRSTATE);
  
  DGP_ASSERT(l != DGP_NULLLIST);
  
  /* Get a new DgpAdt_ListNode */
  LN_SEC_MALLOC(&newln);
  
  /* Hook it in */
  if (ln == DGP_NULLLISTNODE) {
    newln->next = l->first;
    l->first = newln;
    newln->prev = DGP_NULLLISTNODE;
    if (l->last == DGP_NULLLISTNODE) {
      DGP_ASSERT(newln->next == DGP_NULLLISTNODE);
      l->last = newln;
    }
    else {
      DGP_ASSERT(newln->next != DGP_NULLLISTNODE);
      newln->next->prev = newln;
    }
  }
  else {
    DGP_ASSERT(ln->owner == l);
    newln->next = ln->next;
    ln->next = newln;
    newln->prev = ln;
    if (newln->next != DGP_NULLLISTNODE) {
      newln->next->prev = newln;
    }
    else {
      DGP_ASSERT(ln == l->last);
      l->last = newln;
    }
  }
  
  /* and fill it out */
  newln->owner = l;
  newln->userdata = userdata;
  
  l->nelt++;
  if(retnode != NULL) *retnode = newln;
  
  NORMAL_RETURN;
}

ErrCode
DgpAdt_liRem(DgpAdt_ListNode ln)
{
  Dgp_State *dgp_state = ln->owner->dgp_state;
  
  DGP_PROLOGUE ("DgpAdt_liRem", DGP_PERRSTATE);
  
  DGP_STOP_WHINE;
  DGP_ASSERT(ln != DGP_NULLLISTNODE);
  DGP_ASSERT(ln->owner != DGP_NULLLIST);
  
  /* Unhook the DgpAdt_ListNode */
  if (ln->owner->first == ln) {
    ln->owner->first = ln->next;
    DGP_ASSERT(ln->prev == DGP_NULLLISTNODE);
  }
  else {
    ln->prev->next = ln->next;
  }
  if (ln->owner->last == ln) {
    ln->owner->last = ln->prev;
    DGP_ASSERT(ln->next == DGP_NULLLISTNODE);
  }
  else {
    ln->next->prev = ln->prev;
  }
  /* and get rid of it */
  ln->owner->nelt--;
  LN_FREE(ln);
  
  NORMAL_RETURN;
}

ErrCode
DgpAdt_liRemAndNext(DgpAdt_ListNode ln, DgpAdt_ListNode *retnode)
{
  DgpAdt_ListNode nextln;
  Dgp_State *dgp_state = ln->owner->dgp_state;
  
  DGP_PROLOGUE ("DgpAdt_liRemAndNext", DGP_PERRSTATE);
  
  DGP_ASSERT(ln != DGP_NULLLISTNODE);
  DGP_ASSERT(ln->owner != DGP_NULLLIST);
  
  nextln = ln->next;
  DGP_ERR_PASS (DgpAdt_liRem(ln));
  *retnode = nextln;
  NORMAL_RETURN;
}


ErrCode
DgpAdt_liIsIn(DgpAdt_List l, void *userdata, DgpAdt_ListNode *retnode)
{
  DgpAdt_ListNode ln;
  Dgp_State *dgp_state = l->dgp_state;
  
  DGP_PROLOGUE ("DgpAdt_liIsIn", DGP_PERRSTATE);
  DGP_STOP_WHINE;
  
  DGP_ASSERT(l != DGP_NULLLIST);
  
  *retnode = DGP_NULLLISTNODE;
  
  for (ln = l->first; ln != DGP_NULLLISTNODE; ln = ln->next) {
    if (ln->userdata == userdata) {
      *retnode = ln;
    }
  }
  
  NORMAL_RETURN;    
}

ErrCode
DgpAdt_liApp(DgpAdt_List l1, DgpAdt_List l2)
{
  DgpAdt_ListNode ln;
  Dgp_State *dgp_state = l1->dgp_state;
  
  DGP_PROLOGUE ("DgpAdt_liApp", DGP_PERRSTATE);
  DGP_STOP_WHINE;
  
  DGP_ASSERT(l1 != DGP_NULLLIST);
  DGP_ASSERT(l2 != DGP_NULLLIST);
  DGP_ASSERT(l1 != l2);
  
  /* Patch up the owners from l2 */
  for (ln = l2->first; ln != DGP_NULLLISTNODE; ln = ln->next) {
    DGP_ASSERT(ln->owner == l2);
    ln->owner = l1;
  }
  
  /* and hook l2 onto the end of l1 */
  if (l1->last != DGP_NULLLISTNODE) {
    DGP_ASSERT(l1->last->next == DGP_NULLLISTNODE);
    
    l1->last->next = l2->first;
    if (l2->first != DGP_NULLLISTNODE) {
      DGP_ASSERT(l2->first->prev == DGP_NULLLISTNODE);
      l2->first->prev = l1->last;
      
      DGP_ASSERT(l2->last != DGP_NULLLISTNODE);
      l1->last = l2->last;
    }
  }
  else {
    DGP_ASSERT(l1->first == DGP_NULLLISTNODE);
    l1->last = l2->last;
    l1->first = l2->first;
  }
  
  /* The DgpAdt_List l2 is now no longer valid. */
  l1->nelt += l2->nelt;
  LIST_FREE(l2);
  
  NORMAL_RETURN;
}

#if 0
DgpAdt_ListNode
liFirst(DgpAdt_List l)
{
  assert(l != DGP_NULLLIST);
  
  return(l->first);
}

DgpAdt_ListNode
liNext(DgpAdt_ListNode ln)
{
  assert(ln != DGP_NULLLISTNODE);
  
  return(ln->next);
}

static void *
liGet(DgpAdt_ListNode ln)
{
  assert(ln != DGP_NULLLISTNODE);
  
  return(ln->userdata);
}

static int
liLen(DgpAdt_List l)
{
  assert(l != DGP_NULLLIST);
  
  return(l->nelt);
    }

#endif
