/*
 *
 *	$Header: /project/icons/cvsroot/digipaper/src/libadt/list.h,v 1.3 2001/03/13 00:43:54 mkrocken Exp $
 *
 * Copyright (c) 1990, 1991, 1992, 1993 Cornell University.  All Rights
 * Reserved.
 *
 * Copyright (c) 1991, 1992 Xerox Corporation.  All Rights Reserved.
 *
 * Use, reproduction and preparation of derivative works of this software is
 * permitted.  Any copy of this software or of any derivative work must
 * include both the above copyright notices of Cornell University and Xerox
 * Corporation and this paragraph.  This software is made available AS IS, and
 * XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING
 * WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED
 * HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
 * EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
 * NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGES.
 */
/*
 * list.h - header file for the list module
 */
#ifndef LIST_H
#define LIST_H

#include "util.h"
#include "dgp_state.h"

/*
 * Types. These types are opaque. You can't see these, really...
 */
/*
 * Complain, complain: recursive types are messy in C.
 */
struct _DgpAdt_List {
    struct _DgpAdt_ListNode *first, *last;
	Dgp_State *dgp_state;
    Dgp_UInt32 nelt;
    };
typedef struct _DgpAdt_List *DgpAdt_List;

struct _DgpAdt_ListNode {
    struct _DgpAdt_ListNode *next, *prev;
    DgpAdt_List owner;
    void *userdata;
    };
typedef struct _DgpAdt_ListNode *DgpAdt_ListNode;

/*
 * Constants
 */
#define	DGP_NULLLIST	((DgpAdt_List)NULL)
#define	DGP_NULLLISTNODE	((DgpAdt_ListNode)NULL)

/*
 * Functions
 */
DGP_LIBEXPORT ErrCode 
DgpAdt_liNew(DgpAdt_List *retlist, Dgp_State *dgp_state);

DGP_LIBEXPORT ErrCode 
DgpAdt_liFree(DgpAdt_List l);

DGP_LIBEXPORT ErrCode 
DgpAdt_liClear(DgpAdt_List l);

DGP_LIBEXPORT ErrCode 
DgpAdt_liAdd(DgpAdt_List l, void *userdata, DgpAdt_ListNode *retnode);

DGP_LIBEXPORT ErrCode 
DgpAdt_liAddAfter(DgpAdt_List l, DgpAdt_ListNode ln, void *userdata, DgpAdt_ListNode *retnode);

DGP_LIBEXPORT ErrCode 
DgpAdt_liRem(DgpAdt_ListNode ln);

DGP_LIBEXPORT ErrCode 
DgpAdt_liRemAndNext(DgpAdt_ListNode ln, DgpAdt_ListNode *retnode);

DGP_LIBEXPORT ErrCode 
DgpAdt_liIsIn(DgpAdt_List l, void *userdata, DgpAdt_ListNode *retnode);

DGP_LIBEXPORT ErrCode 
DgpAdt_liApp(DgpAdt_List l1, DgpAdt_List l2);


/*
 * Macros
 */
#define	DGPADT_LIFIRST(l)	((l)->first)
#define	DGPADT_LILAST(l)	((l)->last)
#define	DGPADT_LINEXT(ln)	((ln)->next)
#define	DGPADT_LIPREV(ln)	((ln)->prev)
#define DGPADT_LIGET(ln)	((ln)->userdata)
#define DGPADT_LIPUT(ln, val)	((ln)->userdata = (val))
#define	DGPADT_LILEN(l)	((l)->nelt)
#define DGPADT_FOREACH(ln, l) for((ln) = DGPADT_LIFIRST((l)); (ln) != DGP_NULLLISTNODE; (ln) = DGPADT_LINEXT((ln)))
#define DGPADT_FORAFTER(ln1, ln2) for ((ln1) = DGPADT_LINEXT(ln2); (ln1) != DGP_NULLLISTNODE; (ln1) = DGPADT_LINEXT((ln1)))
#define DGPADT_FORFROM(ln1, ln2) for ((ln1) = (ln2); (ln1) != DGP_NULLLISTNODE; (ln1) = DGPADT_LINEXT((ln1)))

#endif
