/*
 *
 *	$Header: /project/icons/cvsroot/digipaper/src/libadt/priq.h,v 1.2 2000/05/16 21:24:46 jahlborn Exp $
 *
 * Copyright (c) 1990, 1991, 1992, 1993 Cornell University.  All Rights
 * Reserved.
 *
 * Copyright (c) 1991, 1992 Xerox Corporation.  All Rights Reserved.
 *
 * Use, reproduction and preparation of derivative works of this software is
 * permitted.  Any copy of this software or of any derivative work must
 * include both the above copyright notices of Cornell University and Xerox
 * Corporation and this paragraph.  This software is made available AS IS, and
 * XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING
 * WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED
 * HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
 * EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
 * NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGES.
 */
/*
 * priq.h - header file for the priority queue module
 */
#ifndef PRIQ_H
#define PRIQ_H

#include "util.h"
#include "dgp_state.h"

/*
 * Types. These types are opaque. You can't see these, really...
 */
typedef struct _DgpAdt_PriQ {
    struct _DgpAdt_PriQIntNode *root, *last;
    Dgp_Bool (*compareGT)(void *a, void *b);
	Dgp_State *dgp_state;
    Dgp_Int32 count;
    } *DgpAdt_PriQ;

typedef struct _DgpAdt_PriQIntNode {
    struct _DgpAdt_PriQIntNode *left, *right, *parent, *next, *prev;
    struct _DgpAdt_PriQNode *owner;
    void *userdata;
    } *DgpAdt_PriQIntNode;

typedef struct _DgpAdt_PriQNode {
    DgpAdt_PriQIntNode val;
    DgpAdt_PriQ owner;
    } *DgpAdt_PriQNode;

/*
 * Constants
 */
#define	DGP_NULLPRIQ	((DgpAdt_PriQ)NULL)
#define	DGP_NULLPRIQNODE	((DgpAdt_PriQNode)NULL)

/*
 * Functions
 */
DGP_LIBEXPORT ErrCode 
DgpAdt_pqNew(Dgp_Bool (*compareGT)(void *a, void *b), Dgp_State *dgp_state, DgpAdt_PriQ *retpriq);

DGP_LIBEXPORT ErrCode
DgpAdt_pqFree(DgpAdt_PriQ p);

DGP_LIBEXPORT ErrCode
DgpAdt_pqAdd(DgpAdt_PriQ p, void *userdata, DgpAdt_PriQNode *retpriqn);

DGP_LIBEXPORT ErrCode 
DgpAdt_pqRem(DgpAdt_PriQNode pqn);

DGP_LIBEXPORT Dgp_Bool 
(*DgpAdt_pqGetCmpFunction(DgpAdt_PriQ p))(void *, void *);

/*
 * Macros
 */
#define DGPADT_PQFIRST(p)	((p)->root ? (p)->root->owner : DGP_NULLPRIQNODE)
#define	DGPADT_PQGET(pqn)	((pqn)->val->userdata)
#define	DGPADT_PQLEN(p)	((p)->count)

#endif
