/* seekable_buffer.h

   Aaron Stump, Summer '97

   This file defines an ADT that can be plugged in to libxiff instead
   of (FILE *).  The data structure is just a buffer in memory.  
   I define functions that perform the analogs of fseek(), ftell(),
   fread(), and fwrite() for such a buffer.  
*/

/* $Header: /project/icons/cvsroot/digipaper/src/libadt/seekable_buffer.h,v 1.8 2001/05/08 00:42:54 dtracy Exp $ */

#ifndef SEEKABLE_BUFFER_H
#define SEEKABLE_BUFFER_H

#include "util.h"
#include "dgp_allocators.h"
#include <stddef.h>
#include "dgp_state.h"

#ifdef __cplusplus
extern "C"
{
#endif

typedef struct {
  Dgp_Int8 *buf;
  Dgp_Int8 *cur;
  Dgp_Int8 *max_cur;
  Dgp_UInt32 buflen;
  Dgp_Bool do_realloc;
  Dgp_State *dgp_state;
} *DgpAdt_SeekableBuffer;

DGP_LIBEXPORT ErrCode
DgpAdt_sbNew(DgpAdt_SeekableBuffer *s, Dgp_Int8 *buf, Dgp_UInt32 buflen, Dgp_Bool do_realloc, Dgp_State *dgp_state);

DGP_LIBEXPORT void
DgpAdt_sbFree(DgpAdt_SeekableBuffer s);

DGP_LIBEXPORT void
DgpAdt_sbGetBuffer(DgpAdt_SeekableBuffer s, Dgp_Int8 **buf, Dgp_UInt32 *buflen, 
	    Dgp_UInt32 *len_written);


/* Functions that work like fseek(), ftell(), fread(), and fwrite() */

DGP_LIBEXPORT Dgp_Int32
DgpAdt_sbSeek(DgpAdt_SeekableBuffer s, Dgp_Int32  offset, Dgp_Int32 origin);

DGP_LIBEXPORT Dgp_Int32  
DgpAdt_sbTell(DgpAdt_SeekableBuffer s);

DGP_LIBEXPORT size_t
DgpAdt_sbRead(void *ptr, size_t size, size_t nobj, DgpAdt_SeekableBuffer s);

DGP_LIBEXPORT size_t
DgpAdt_sbWrite(void *ptr, size_t size, size_t nobj, DgpAdt_SeekableBuffer s);

#ifdef __cplusplus
}
#endif

#endif
