/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 4.2.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename: stack.h.
**
**  Types, constants, prototypes, & statics for the stack adt.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling;  impacts the prototypes for some functions, but
**		   I understand currently only libscancolor uses this adt.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**		o  Move to libadt? for neatness...
**
**  
**************************************************************************/

#ifndef STACK_H
#define STACK_H

#include "dgp_err_control.h"

/* 
 * Types 
 */
/* The following typedef is for convenience and to allow
   easy modification later; the user shouldn't need to
   use _DgpAdt_Stacknode. */
typedef void *_DgpAdt_Stacknode;

typedef struct {
  _DgpAdt_Stacknode *data;
  _DgpAdt_Stacknode *top;
  Dgp_Int32 max_num_nodes;
} *DgpAdt_Stack;

/* 
 * Constants 
 */
#define DGP_NULLSTACK ((DgpAdt_Stack)NULL)

/*
 * Functions 
 */
ErrCode DgpAdt_staNew(Dgp_State *dgp_state, Dgp_Int32 n, DgpAdt_Stack *_news);
void DgpAdt_staFree(Dgp_State *dgp_state, DgpAdt_Stack s);
void DgpAdt_staClear(DgpAdt_Stack s);
ErrCode DgpAdt_staRealloc(Dgp_State *dgp_state, DgpAdt_Stack s, Dgp_Int32 n);
Dgp_Int32 DgpAdt_staLen(DgpAdt_Stack s);
Dgp_Int32 DgpAdt_staMaxNumNodes(DgpAdt_Stack s);

/*************************************************************************
**
**  Function:  static INLINE Dgp_Int32 DgpAdt_staEmpty(DgpAdt_Stack s)
**
**
**  Returns DGP_TRUE if there are no elements in s;  DGP_FALSE otherwise.
**
**
**	Arguments:
**		o  s is the stack to be queried.
**
**	Return value:
**		o  Integer (Dgp_Bool).
**
**	Modification history:
**		o  None;  use Dgp_Bool data type? / dhtesq.
**  
**/

static INLINE Dgp_Int32 
DgpAdt_staEmpty(DgpAdt_Stack s)
{ 
  return (s->top < s->data); 
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static INLINE Dgp_Int32 DgpAdt_staFull(DgpAdt_Stack s)
**
**
**  Returns DGP_TRUE if s is holding the maximum number of nodes allowed by the last
**	allocation, and DGP_FALSE otherwise.
**
**	Arguments:
**		o  s is the stack to be queried.
**
**	Return value:
**		o  Integer (Dgp_Bool).
**
**	Modification history:
**		o  None;  use Dgp_Bool data type? / dhtesq.
**  
**/

static INLINE Dgp_Int32 DgpAdt_staFull(DgpAdt_Stack s) 
{
  return (s->top - s->data + 1 == s->max_num_nodes);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static INLINE void *DgpAdt_staPop(DgpAdt_Stack s)
**
**
**  DgpAdt_staPop() pops the top element of the DgpAdt_Stack s, and returns the data it had
**  contained.  If s is empty, DgpAdt_staPop() returns (void *)NULL.
**
**	Arguments:
**		o  s is the stack to be queried.
**
**	Return value:
**		o  Generic pointer;  typically recast by caller.
**
**	Modification history:
**		o  None.
**  
**/
static INLINE void
*DgpAdt_staPop(DgpAdt_Stack s)
{
  if (DgpAdt_staEmpty(s))
    return (void *)NULL;

  return *(s->top--);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static INLINE Dgp_Int32 DgpAdt_staPush(DgpAdt_Stack s, void *userdata)
**
**
**  DgpAdt_staPush() pushes userdata onto the DgpAdt_Stack s.  It returns DGP_TRUE if there
**  was enough room on the stack, and DGP_FALSE otherwise.  (Note that
**  even if s was created by DgpAdt_staNew with realloc_if_nec given as DGP_TRUE,
**  DgpAdt_staPush might still return DGP_FALSE if it tries to reallocate s's memory
**  and can't.)
**
**  (I think this refers to some other implementation; no
**	realloc_if_nec on ours -- dhtesq.)
**
**  Note that if the user doesn't know in advance how many items she'll want
**  to put on a given DgpAdt_Stack s, she can check the return value of DgpAdt_staPush each
**  time she tries to push something onto s and call DgpAdt_staRealloc with s and 
**  some number greater than DgpAdt_staMaxNumNodes(s) whenever that return value is 
**  DGP_FALSE.
**
**
**	Arguments:
**		o  s is the stack to be filled.
**		o  userdata
**
**	Return value:
**		o  Integer (Dgp_Bool).
**
**	Modification history:
**		o  None;  use Dgp_Bool for return type? / dhtesq.
**  
**/

static INLINE Dgp_Int32
DgpAdt_staPush(DgpAdt_Stack s, void *userdata)
{
  if (DgpAdt_staFull(s))
    return 0;

  *(++(s->top)) = userdata;
  return 1;
}
/*************************************************************************/

#endif
