
/* 
 *  This program generate random number of character with random location on the page.
 *  random width and height, random word length and number of words, in order to test
 *  the encoding and decoding of libascii.
 */
#ifdef _WIN32
#include <wchar.h>
#include "ascii_exts.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include <malloc.h>
#include <time.h>
#include "windows.h"
#include "mmcompress.h"
#include "extension_tags.h"
#include "dgp_icr.h"
#include "icrpub.h"
#include "zlib.h"
#include "zconf.h"
#include "dgp_sys_err_control.h"

typedef struct {
	WCHAR c;
	RECT box;
} CharInfo;

typedef struct {
	int numChars;
	CharInfo *cInfo;
} DgpTextInfo;

/* structure for doing error/report output */
typedef struct {
  FILE *errfile;
  FILE *rptfile;
} AnnTest_statusOutputFiles;

#ifdef DGP_DBUG

/* function to output debug errors to a file */
void 
AnnTest_errPrinter(void *cookie, const Dgp_Int8 *file_name, const Dgp_Int8 *funct_name,
                  Dgp_Int32 line_number, ErrCode error, const Dgp_Int8 *format, ...)
{
  va_list(arglist);
  FILE *outfile;

  if(cookie) {
    va_start(arglist, format);
    if(error == DGP_REPORT) {
      outfile = ((AnnTest_statusOutputFiles *)cookie)->rptfile;
    } else {
      outfile = ((AnnTest_statusOutputFiles *)cookie)->errfile;
    }

    /* we're done if the outfile is NULL */
    if(outfile == NULL) return;

    if(error == DGP_REPORT) {
      fprintf(outfile, "Report -- ");
    } else if(error == DGP_WARNING) {
      fprintf(outfile, "Warning -- ");
    } else {
      fprintf(outfile, "Error %d, \"%s\", line %d, %s -- ", (Dgp_Int32)error, file_name, line_number, funct_name);
    }
    if(format != NULL) vfprintf(outfile, format, arglist);
    fprintf(outfile, "\n");
  }
}

#else /* DGP_DBUG is not defined */

/* function to output nondebug errors to a file */
void 
AnnTest_errPrinter(void *cookie, ErrCode error, const Dgp_Int8 *format, ...)
{
  va_list(arglist);
  FILE *outfile;

  if(cookie) {
    va_start(arglist, format);
    if(error == DGP_REPORT) {
      outfile = ((AnnTest_statusOutputFiles *)cookie)->rptfile;
    } else {
      outfile = ((AnnTest_statusOutputFiles *)cookie)->errfile;
    }

    /* we're done if the outfile is NULL */
    if(outfile == NULL) return;

    if(error == DGP_REPORT) {
      fprintf(outfile, "Report -- ");
    } else if(error == DGP_WARNING) {
      fprintf(outfile, "Warning -- ");
    } else {
      fprintf(outfile, "Error %d -- ", (int)error);
    }
    if(format != NULL) vfprintf(outfile, format, arglist);
    fprintf(outfile, "\n");
  }
}

#endif


main(int argc, Dgp_Int8 **argv)

{
	AnnTest_statusOutputFiles stat_files = {stderr, NULL};
	DGP_MAIN_PROLOGUE("main()", (void *)&stat_files, AnnTest_errPrinter);
	Dgp_State *dgp_state;
	XDOCData *xdocData = NULL, *curLet;
	Dgp_Int32 i;
	CharInfo *curChar, *charInfo=NULL;
	TextInfo text;
	Dgp_Int32 numWords, curWordLength;
	Dgp_Int32 *wordLengths = NULL;
	wchar_t c;
	Dgp_Int32 buflen;
	char *textExtension;
	const Dgp_Int32 numChars = 1000;
	const Dgp_Int32 pageWidth = 2400;
	const Dgp_Int32 pageHeight = 3300;
	const Dgp_Int32 charWidth = 500;    /*char width will be any random number < charWidth */
	const Dgp_Int32 charMax = 127;      /*(0~127 ASCII, 128~255 Europe, 256~65535 Unicode) */
	FILE *out_fp;
	FILE *out_fp_pos;

	DGP_SYS_ERR_GEN_CLEANUP(IS_NULL(out_fp = fopen("extout.txt", "wb")));
	DGP_SYS_ERR_GEN_CLEANUP(IS_NULL(out_fp_pos = fopen("posout.txt", "wb")));
	
	DGP_ERR_PASS_CLEANUP(Dgp_initDgpState(&dgp_state, NULL, NULL, (void *)(&stat_files),
					AnnTest_errPrinter, DGP_TRUE, DGP_TRUE));

	DGP_SEC_MALLOC(charInfo, (CharInfo *), sizeof(CharInfo)* numChars);
	
	/* Seed the random-number generator with current time so that
	* the numbers will be different every time we run.
	*/
	srand( (unsigned)time( NULL ) );

	/* Initialize the chars */
	curChar = charInfo;
	for(i = 0; i < numChars-1; i++){
		curChar->c = rand() * charMax / RAND_MAX;
		curChar->box.left = rand() * pageWidth / RAND_MAX ;
		curChar->box.right = curChar->box.left + charWidth*rand() / RAND_MAX ;
		curChar->box.top = rand() * pageHeight /RAND_MAX ;
		curChar->box.bottom = curChar->box.top + charWidth*rand() / RAND_MAX;
		curChar ++;
	}
	
    curChar->c = L'\0';

	DGP_SEC_MALLOC(xdocData, (XDOCData *), sizeof(XDOCData) * numChars);

	DGP_SEC_MALLOC(wordLengths, (Dgp_Int32 *), sizeof(Dgp_Int32) * numChars);

	curLet = xdocData;
	curChar = charInfo;
	numWords = 0;
	curWordLength = 0;
	for(i = 0; i < numChars; i++) {
		/* Convert the character */
	/*	c = __toascii(curChar->c); */
		c = curChar->c;
		/* Check if the char is a spacing char */
		if((c == L'\0') || (c == L' ')) {

			/* Check if this is the first space after a word */
			if(curWordLength != 0) {

				/* save word data */
				wordLengths[numWords] = curWordLength;
				numWords++;
			}

			/* reset word length */
			curWordLength = 0;

		} else {
			/* This is a real letter save its info */

			curLet->word_info.start_of_word = (curWordLength == 0);
			curLet->word_info.word_toks[0] = c;
			curLet->co_ords.upper_left_x = curChar->box.left;
			curLet->co_ords.upper_left_y = curChar->box.top;
			curLet->co_ords.lower_right_x = curChar->box.right;
			curLet->co_ords.lower_right_y = curChar->box.bottom;
			
			curLet++;
			curWordLength++;
		}

		curChar++;
	}

	text.xdoc_data = xdocData;
	text.num_chars = curLet - xdocData;
	text.num_words = numWords;
	text.word_lengths = wordLengths;

	DGP_ERR_PASS_CLEANUP( ZCreateAsciiExtension(	dgp_state, 
													&text, 
													&buflen, 
													&textExtension));
	/* done with this stuff */
	if(xdocData)	DGP_FREE(xdocData);
	xdocData = NULL;
	if(wordLengths) DGP_FREE(wordLengths);
	wordLengths = NULL;
	if(charInfo) DGP_FREE(charInfo);
	charInfo = NULL;


	DGP_ERR_PASS_CLEANUP( DecodeTextExtension( dgp_state,
											 buflen,
											 textExtension,
											 out_fp,
											 out_fp_pos));
	fclose(out_fp);
	fclose(out_fp_pos);

    if(textExtension) DGP_FREE(textExtension);
	Dgp_cleanupDgpState(&dgp_state);
    dgp_state = NULL;

    textExtension = NULL;

    MAIN_NORMAL_RETURN;
  
ERR_CLEANUP:

	if(xdocData)	DGP_FREE(xdocData);
	if(wordLengths) DGP_FREE(wordLengths);
	if(charInfo) DGP_FREE(charInfo);
    if(textExtension) DGP_FREE(textExtension);

	if(out_fp ) fclose(out_fp);
	if(out_fp_pos) fclose(out_fp_pos);
    if(dgp_state != NULL)  Dgp_cleanupDgpState(&dgp_state);
  
  MAIN_ERR_RETURN;
}
#endif
