/*////////////////////////////////////////////////////////////////////////// */
/* */
/*  DigiPaper Codebase */
/* */
/*  Ranjani Ramamurthy, September 24 1997 */
/*  Feb 08, 1998: Added all the libicr calls to interface with the */
/*            Textbridge code. RR */
/* */
/*  Filename: create_ascii_exts.c */
/* */
/*  Description of whats implemented in this file: */
/* */
/* */
/* */
/* */
/* */
/*///////////////////////////////////////////////////////////////////////// */

#ifdef _WIN32
#include <winsock2.h>
#endif

#include "dgp_rcs.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libascii/create_ascii_exts.c,v 1.23 2003/02/11 01:31:35 mkrocken Exp $");

#ifdef _WIN32
#include <wchar.h>
#include "ascii_exts.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include <malloc.h>
#include "mmcompress.h"
#include "extension_tags.h"
#include "dgp_icr.h"
#include "icrpub.h"
#include "zlib.h"
#include "zconf.h"
#include "dgp_sys_err_control.h"


/* Functions defined in ascii_exts_utils.c */
#define ZCOMPRESS_ERROR   -1
#define ZCOMPRESS_LEVEL    9

/* Some local structures for encapsulation. They are */
/* used only within this fine, and hence, are not defined */
/* in a header file */

typedef struct{
  Dgp_Int32     *vals;
  Dgp_Int32     num_vals;
  Dgp_Int8      *comp_array;
  Dgp_Int32     comp_length;
} encodestruct;

typedef struct{
  encodestruct *word_lengths;
  encodestruct *dlhsx;
  encodestruct *dlhsy;
  encodestruct *width;
  encodestruct *height;
  Dgp_Int8     *word_arr;
  Dgp_Int32    length;
}finaldata;

/*
extern PosExtAsciiCookie init_pext_ascii_cookie(void *(*new_malloc)(size_t),
  void (*new_free)(void *), void * (*new_calloc)(size_t, size_t),
  void * (*new_realloc)(void *, size_t),
  Dgp_Int8* zstr, Dgp_Int32 bytelen);
*/

#ifdef OCR

DGP_LIBEXPORT Dgp_Bool ReadXDOCFile(TextInfo *textInfo, Dgp_Int8 *filter){
  Dgp_Int32 ret;

  ret = ParseXDOCFile(filter, &(textInfo->xdoc_data),
    &(textInfo->num_chars), &(textInfo->num_words),
    &(textInfo->word_lengths));

  if(ret == DGP_XDOC_EOF_NOT_REACHED){
    return DGP_FALSE; /* Some error occured in the parse */
  }

  return DGP_TRUE;
}

#endif

/* Function defined in encode_ascii.c */
extern void
EncodeElements(Dgp_Int32* vals, Dgp_Int32 nel,
         Dgp_Int32* ret_len, Dgp_Int8** encoded_arr);

/*////////////////// BEGINNING OF ALL STATIC FUNCTIONS//////////// */
/*//////////////////////////////////////////////////////////////// */
/* function name: free_encodestruct */
/* */
/*//////////////////////////////////////////////////////////////// */
static void
free_encodestruct(encodestruct *input, Dgp_State *dgp_state)
{
   if(input->vals)
     (DGP_FREE)(input->vals);
   if(input->comp_array)
     (DGP_FREE)(input->comp_array);
   return;
}/*end of function free_encodestruct */
/*//////////////////////////////////////////////////////////////// */
/* function name: free_alldata */
/* */
/*//////////////////////////////////////////////////////////////// */
static void
free_alldata(finaldata *final_dat, Dgp_State *dgp_state)
{

  if(final_dat->word_arr)
    (DGP_FREE)(final_dat->word_arr);
  if(final_dat->dlhsx)
    free_encodestruct(final_dat->dlhsx, dgp_state);
  if(final_dat->dlhsy)
    free_encodestruct(final_dat->dlhsy, dgp_state);
  if(final_dat->width)
    free_encodestruct(final_dat->width, dgp_state);
  if(final_dat->height)
    free_encodestruct(final_dat->height, dgp_state);

  return;
} /* end of function free_alldata */

/*/////////////////////////////////////////////////////////////////////////// */
/* static function: set_compressed_data */
/* */
/*/////////////////////////////////////////////////////////////////////////// */
static void  set_compressed_data(Dgp_Int8** in_final_array,
  Dgp_Int32* offset,Dgp_Int32 copy_length,Dgp_Int8* array_to_be_copied)
{

  Dgp_Int8* final_array ;
#ifdef DEBUG_EXTENSION
  Dgp_Int32 val = 0;
#endif

  final_array = *(in_final_array);

  final_array[(*offset)] = (copy_length >> 8) & 0xff;
  final_array[ (*offset)+1] = (copy_length) & 0xff;

#ifdef DEBUG_EXTENSION
   val = (final_array[*offset] << 8) | (final_array[*offset+1]);
#endif

  (*offset) += DGP_INFO_SIZE;
  if(array_to_be_copied != NULL){
    memcpy(final_array + (*offset),
    array_to_be_copied, copy_length);
    (*offset) += copy_length;
  }
}
/*////////////////////////////////////////////////////////////////////// */
/* function : static function free_all.. */
/* */
/*////////////////////////////////////////////////////////////////////// */
static void
free_all(Dgp_Int32 *arr1, Dgp_Int32 *arr2, Dgp_Int32 *arr3, Dgp_Int32 *arr4, Dgp_State *dgp_state)
{
  if(arr1)
    (DGP_FREE)(arr1);
  if(arr2)
    (DGP_FREE)(arr2);
  if(arr3)
    (DGP_FREE)(arr3);
  if(arr4)
    (DGP_FREE)(arr4);
  return;
}

/*////////////////////////////////////////////////////////////////////////// */
/* */
/* function name: z_compress_data */
/* Will call appropriate functions from the zlib set in order to compress */
/* the data. */
/* */
/*////////////////////////////////////////////////////////////////////////// */
static ErrCode
z_compress_data(Dgp_Int8* source, Dgp_Int32 sourcelen, Dgp_State *dgp_state, Dgp_Int32* destlen,
                Dgp_Int8 **_ret_array)
{
  DGP_PROLOGUE("z_compress_data()", DGP_PERRSTATE);
  Dgp_Int8 *ret_array = NULL;
  uLong ret_array_len;

  ret_array_len = sourcelen + (Dgp_Int32)(0.4 * (Dgp_Float32)sourcelen) + 12;

  DGP_CHECK_ALLOC_CLEANUP(ret_array = (Dgp_Int8*)calloc(ret_array_len,1));

  DGP_ERR_GEN_CLEANUP((compress2((Byte *)ret_array,&ret_array_len,
                     (/*const*/ Bytef*)source,
                      sourcelen,ZCOMPRESS_LEVEL) != Z_OK), cErrCompressionError);

  *destlen = ret_array_len;
  *_ret_array = ret_array;
  ret_array = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(ret_array != NULL) (DGP_FREE)(ret_array);
  *destlen = 0;
  *_ret_array = NULL;

  ERR_RETURN;
} /*end of function z_compress_data */

/*//////////////////////////////////////////////////////////////// */
static ErrCode
encode_and_compress(encodestruct *in_struct, Dgp_State *dgp_state, FILE *fp)
{
  DGP_PROLOGUE("encode_and_compress()", DGP_PERRSTATE);
  Dgp_Int8 *encoded_array = NULL;
  Dgp_Int32  encoded_length;

#ifndef DEBUG_EXTENSION
  /* it's used! */
  (void)fp;
#endif

  /*First, allocate the encode_array */
  DGP_CHECK_ALLOC(encoded_array = (DGP_CALLOC)(in_struct->num_vals,
                 sizeof(Dgp_Int32)));

  /* This function is defined in encode_ascii.c */
  EncodeElements(in_struct->vals, in_struct->num_vals,
               &encoded_length,&encoded_array);
#ifdef DEBUG_EXTENSION
  fprintf(fp,"is %d",encoded_length);
  fflush(fp);
#endif

  /*Now compress by interfacing with zlib */
  DGP_ERR_PASS_CLEANUP(z_compress_data(encoded_array,encoded_length,
                  dgp_state, &(in_struct->comp_length), &(in_struct->comp_array)));

  if(encoded_array != NULL) (DGP_FREE)(encoded_array);
  encoded_array = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(encoded_array != NULL) (DGP_FREE)(encoded_array);

  ERR_RETURN;
} /*end of function encode_and_compress */

/*////////////////////////////////////////////////////////////// */
/* function name: malloc_int_arrays */
/* */
/*////////////////////////////////////////////////////////////// */
static Dgp_Bool
malloc_int_arrays(Dgp_Int32 **arr1, Dgp_Int32 **arr2, Dgp_Int32 **arr3,
          Dgp_Int32 **arr4, Dgp_Int32 in_size, Dgp_State *dgp_state)
{
  (*arr1) = (DGP_CALLOC)(in_size,sizeof(Dgp_Int32));
  (*arr2) = (DGP_CALLOC)(in_size,sizeof(Dgp_Int32));
  (*arr3) = (DGP_CALLOC)(in_size,sizeof(Dgp_Int32));
  (*arr4) = (DGP_CALLOC)(in_size,sizeof(Dgp_Int32));
  if( !(*arr1) || !(*arr2) || !(*arr3) || !(*arr4) ){
      free_all(*arr1,*arr2,*arr3,*arr4, dgp_state);
    return DGP_FALSE;
  }
  return DGP_TRUE;
}

/*/////////////////////////////////////////////////////////// */
/* function name: init_encodestructs */
/* */
/*//////////////////////////////////////////////////////////// */
static ErrCode
init_encodestructs(encodestruct *dlhsx, encodestruct *dlhsy,
  encodestruct *width, encodestruct *height, Dgp_Int32 num_chars,
  XDOCData *xdoc_data, Dgp_State *dgp_state, FILE* fp)
{
  DGP_PROLOGUE("init_encodestructs()", DGP_PERRSTATE);
  Dgp_Int32 i;
  Dgp_Int32 prev_x, prev_y, prev_width;


#ifndef DEBUG_EXTENSION
  /* it's used! */
  (void)fp;
#endif

  prev_x = prev_y = prev_width = 0;
  dlhsx->comp_length = 0;
  dlhsy->comp_length = width->comp_length = 0;
  height->comp_length = 0;

  dlhsx->num_vals = dlhsy->num_vals = width->num_vals =
                   height->num_vals = num_chars;

  /*Create the integer arrays for all the other values.. */
  /*dlhsx, dlhsy, width, height */
  DGP_ERR_GEN(!malloc_int_arrays(&dlhsx->vals,&dlhsy->vals,&width->vals,
    &height->vals,num_chars, dgp_state), cErrNomem);

  for(i = 0; i < num_chars; i++){
    dlhsx->vals[i] =
    xdoc_data[i].co_ords.upper_left_x
     - (prev_x + prev_width);
  dlhsy->vals[i] =
      xdoc_data[i].co_ords.upper_left_y - prev_y;
    width->vals[i] = xdoc_data[i].co_ords.lower_right_x -
    xdoc_data[i].co_ords.upper_left_x;
  height->vals[i] = xdoc_data[i].co_ords.lower_right_y -
    xdoc_data[i].co_ords.upper_left_y;

  prev_x = xdoc_data[i].co_ords.upper_left_x;
  prev_y = xdoc_data[i].co_ords.upper_left_y;
  prev_width = xdoc_data[i].co_ords.lower_right_x - xdoc_data[i].co_ords.upper_left_x;

  }

#ifdef DEBUG_EXTENSION
    prev_x = prev_y = prev_width = 0;
    for(i = 0; i < num_chars; i++){
    fprintf(fp,"LHS <%d, %d > ::",
     xdoc_data[i].co_ords.upper_left_x,
     xdoc_data[i].co_ords.upper_left_y);
    fprintf(fp,"RHS <%d, %d > ::",
     xdoc_data[i].co_ords.lower_right_x,
     xdoc_data[i].co_ords.lower_right_y);
    fprintf(fp,"W %d and H %d",
         (xdoc_data[i].co_ords.lower_right_x - xdoc_data[i].co_ords.upper_left_x),
         (xdoc_data[i].co_ords.lower_right_y - xdoc_data[i].co_ords.upper_left_y));

    fprintf(fp,"%d::Delta<x> is %d and delta y is %d",i,
          (xdoc_data[i].co_ords.upper_left_x - (prev_x + prev_width)),
        (xdoc_data[i].co_ords.upper_left_y - prev_y));
    prev_x = xdoc_data[i].co_ords.upper_left_x;
    prev_y = xdoc_data[i].co_ords.upper_left_y;
    prev_width = xdoc_data[i].co_ords.lower_right_x
             - xdoc_data[i].co_ords.upper_left_x;
  }

#endif

  NORMAL_RETURN;
}
/*///////////////////////////////////////////////////////// */
/* function name: comp_word_array */
/* */
/*/////////////////////////////////////////////////////////// */
static ErrCode
comp_word_array(XDOCData *xdoc_data,Dgp_Int32 num_chars,
        Dgp_State *dgp_state, Dgp_Int32* comp_word_len, Dgp_Int8 **ret_array)
{
  DGP_PROLOGUE("comp_word_array()", DGP_PERRSTATE);
  Dgp_Int32 i;
  Dgp_UInt8 *word_array = NULL;

  DGP_CHECK_ALLOC(word_array = (DGP_CALLOC)(num_chars * 2,sizeof(Dgp_UInt8)));
  
  /*put low byte value first then high byte value to achieve better compression for ascii in wchar mode */
  for(i = 0; i < num_chars; i++){
    word_array[i]  = (Dgp_UInt8)(xdoc_data[i].word_info.word_toks[0] & 0xff);
    word_array[i + num_chars] = (Dgp_UInt8)(xdoc_data[i].word_info.word_toks[0] >> 8);
  }

  DGP_ERR_PASS_CLEANUP(z_compress_data((Dgp_Int8 *)word_array, 2 * num_chars, dgp_state, comp_word_len, ret_array));

  if(word_array != NULL) (DGP_FREE)(word_array);
  word_array = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(word_array != NULL) (DGP_FREE)(word_array);

  ERR_RETURN;
} /*end of function compress_word_array */
/*///////////////////////////////////////////////////////////// */
/* function name: concatenate_all */
/* */
/*///////////////////////////////////////////////////////////// */
static ErrCode
concatenate_all(finaldata encaps, Dgp_Int32 num_words,
        Dgp_Int32 num_chars, Dgp_State *dgp_state, FILE* fp, Dgp_Int32 *final_len, 
        Dgp_Int8 **_final_array)
{
  DGP_PROLOGUE("concatenate_all()", DGP_PERRSTATE);
  Dgp_Int32 final_array_offset,final_ext_len = 0;
  Dgp_Int8 *final_array = NULL;

#ifndef DEBUG_EXTENSION
  /* it's used! */
  (void)fp;
#endif

  final_ext_len = 1 + (7 * DGP_INFO_SIZE)
                + encaps.word_lengths->comp_length
          + encaps.length
          + encaps.dlhsx->comp_length
          + encaps.dlhsy->comp_length
          + encaps.width->comp_length
          + encaps.height->comp_length;

  DGP_CHECK_ALLOC_CLEANUP(final_array = (DGP_CALLOC)(final_ext_len,sizeof(Dgp_Int8)));
  
  final_array_offset = 0;
  /*Put in a revision number */
  final_array[0] = TEXT_COMPRESS_VERSION;
  final_array_offset = 1;

  /*Put in the number of words.. */
  set_compressed_data(&final_array,&final_array_offset,
     num_words,NULL);

#ifdef DEBUG_EXTENSION
  fprintf(fp,"full length = %d",final_ext_len);
  fprintf(fp,"Num Dgp_Int8s is %d",num_chars);
  fprintf(fp,"num words %d & offset %d ",num_words,final_array_offset);
  fflush(fp);
#endif

  /* Now start putting in all the compressed data.. */
  set_compressed_data(&final_array,&final_array_offset,
     encaps.word_lengths->comp_length,
     encaps.word_lengths->comp_array);

#ifdef DEBUG_EXTENSION
  fprintf(fp,"word length arr_sz %d & offset %d",
      encaps.word_lengths->comp_length,final_array_offset);
#endif

  set_compressed_data(&final_array,&final_array_offset,
    encaps.length,encaps.word_arr);

#ifdef DEBUG_EXTENSION
  fprintf(fp,"words arr_sz %d & offset %d",
       encaps.length,final_array_offset);
#endif

  set_compressed_data(&final_array,&final_array_offset,
    encaps.dlhsx->comp_length,encaps.dlhsx->comp_array);

#ifdef DEBUG_EXTENSION
  fprintf(fp,"dlhsx arr_sz %d & offset %d",
       encaps.dlhsx->comp_length,final_array_offset);
#endif

  set_compressed_data(&final_array,&final_array_offset,
    encaps.dlhsy->comp_length,encaps.dlhsy->comp_array);

#ifdef DEBUG_EXTENSION
  fprintf(fp,"dlhsy arr_sz %d & offset %d",
       encaps.dlhsy->comp_length,final_array_offset);
#endif

  set_compressed_data(&final_array,&final_array_offset,
    encaps.width->comp_length,encaps.width->comp_array);


#ifdef DEBUG_EXTENSION
  fprintf(fp,"width arr_sz %d & offset %d",
       encaps.width->comp_length,final_array_offset);
#endif

  set_compressed_data(&final_array,&final_array_offset,
    encaps.height->comp_length,encaps.height->comp_array);

#ifdef DEBUG_EXTENSION
  fprintf(fp,"height arr_sz %d & offset %d",
       encaps.height->comp_length,final_array_offset);
#endif

  /*Sanity check.. */
  DGP_ASSERT(final_array_offset == final_ext_len);

  *final_len = final_ext_len;
  *_final_array = final_array;
  final_array = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(final_array != NULL) (DGP_FREE)(final_array);

  ERR_RETURN;
} /*end of function concatenate_all */

/*////////////// END OF ALL STATIC FUNCTIONS ///////////////////// */

/*////////////////////////////////////////////////////////////////////// */
/* function : TextExtensionBuilder */
/* */
/*////////////////////////////////////////////////////////////////////// */

Dgp_Bool TextExtensionBuilder(tokpos** sorted_tokpos,
    Dgp_Int32 num_pos, void* extcookie, void** extension,
    size_t *ext_len){

  PosExtAsciiCookie pcookie =
    (PosExtAsciiCookie)extcookie;

  if(pcookie->bytelen > 0){
      *extension = (void*)pcookie->zstr;
    *ext_len = pcookie->bytelen;

  } else{
      *extension = NULL;
    *ext_len = 0;
  }
    return DGP_TRUE;
}
/*////////////////////////////////////////////////////////////// */
/* function name: CreatePosExtAsciiCookie */
/* */
/*////////////////////////////////////////////////////////////// */
/*
PosExtAsciiCookie
CreatePosExtAsciiCookie(Dgp_Int8* zstr, Dgp_Int32 bytelen)
{

  PosExtAsciiCookie local_cookie;

    local_cookie = init_pext_ascii_cookie(malloc,
    free,calloc,realloc,zstr,bytelen);

  if(!local_cookie){
    return NULL;
  }
  return local_cookie;
} */
/*////////////////////////////////////////////////////////////// */
/* function name: CreateAsciiPosExtensions */
/* */
/*////////////////////////////////////////////////////////////// */
/*
void
CreateAsciiPosExtensions(PosExtension* pos_ext,
                PosExtAsciiCookie cookie)
{

// Create the Position Block Extension 
  pos_ext->extdata = NULL;
  pos_ext->builder =  TextExtensionBuilder;
  pos_ext->cookie = (void*)cookie;
  pos_ext->tag = TEXT_EXT_TAG;

  return;
  } //end of CreateDictExtensions

*/
/*////////////////////////////////////////////////////////////////////////// */
/* */
/* function name: ZCreateAsciiExtension: */
/* */
/* This function actually takes the XDOC data and puts it in a format */
/* such that it can be stored as a position extension. This function */
/* calls the zlib libraries in order to compress the various 'strings' */
/* of information (i.e. word-lengths, words, lhs x co-ord of character */
/* bounding box, lhs y co-ord of character bounding box, rhs x-co-ord */
/* of character bounding box, rhs y co-ord of character bounding box. */
/* */
/* An explanation of how things are stored here.. */
/* */
/* Consider an XDOC data structure holding the following information: */
/* */
/*  {the word 'see' stored as} */
/*  's' '1' <1,2,3,4> */
/*  'e' '0' <5,6,7,8> */
/*  'e' '0' <9,10,11,12> */
/*  {the word 'vishnu' stored as} */
/*  'v'  '1' <13,14,15,16> */
/*  'i'  '0' <17,18,19,20> */
/*  's'  '0' <21,22,23,24> */
/*  'h'  '0' <25,26,27,28> */
/*  'n'  '0' <29,30,31,32> */
/*  'u'  '0' <33,34,35,36> */
/*  {the word 'play' stored as} */
/*  'p'  '1'  <37,38,39,40> */
/*  'l'  '0'  <41,42,43,44> */
/*  'a'  '0'  <45,46,47,48> */
/*  'y'  '0'  <49,50,51,52> */
/* */
/*  num_words = 3; num_chars = 12 */
/* */
/*  word_lengths = {3,5,4} respectively. */
/* */
/*  Note that for every piece of information, we use 2 bytes to hold that */
/*  information.. */
/* */
/*  Here is how a string returned by ZCreateAsciiExtension will look: */
/* */
/*  03 | <03> |{354}|<12>| compressed{seevishnuplay} | */
/*      |<12>| {...}| <12>| {..} | <12>|{..} */
/*      |<12>| {..} */
/* */
/*  An explanation of what they are:: */
/*  03 -> Number of words in the file */
/*  <03> -> say, length of the compressed word lengths */
/*  {354} -> the actual word lengths (note that this will be compressed) */
/*  <12> => the length of the compressed words. I have written this in brackets */
/*  as in this case, it's only as many bytes - as I am giving an example where */
/*  the word data is not compressed. */
/* */
/*  {seevishnuplay} -> the actual words (will be later on compressed */
/*              by zlib */
/*  <12> => length of the compressed delta-LHS x co-ords */
/*  {..} => the compressed delta-LHS x-co-ords */
/*  <12> => length of the compressed delta-LHS y co-ords */
/*  {..} => the compressed delta-LHS y-co-ords */
/*  <12> => length of the compressed widths */
/*  {..} => the compressed widths */
/*  <12> => length of the compressed heights */
/*  {..} => the compressed heights */
/*  The length of the entire position extension could be computed as */
/*  follows. */
/* */
/*  We use 2 bytes to store any piece of relevant information - be it */
/*  the number of words or the length of any compressed segment. Going */
/*  by that account, the length of the position extension for the example */
/*  given would be: */
/* */
/*  [2 bytes] + [ 2 bytes] + (3 bytes) + [ 2 bytes ] + (12 bytes) + */
/*  4 * { [ 2 bytes ] + (12 bytes) } = 77 bytes */
/* */
/*  Note that all 'information' is in [] brackets where as compressed */
/*  data lengths is in () brackets. */
/* */
/*  if DGP_INFO_SIZE = 2 (bytes), the ascii position extension length would */
/*  be :: */
/*  DGP_INFO_SIZE * 7 + {length in bytes of compressed word length} */
/*            + { length in bytes of compressed words} */
/*              + { length in bytes of compressed LHS x co-ords} */
/*              + { length in bytes of compressed LHS y co-ords} */
/*              + { length in bytes of compressed RHS x co-ords} */
/*              + { length in bytes of compressed RHS y co-ords} */
/* */
/*  2* 7 + (3) + (12) + (12) +(12) +(12) +(12) = 14 + 3 + 5 * 12 = 77 */
/*///////////////////////////////////////////////////////////////////////////// */
DGP_LIBEXPORT ErrCode
ZCreateAsciiExtension(Dgp_State *dgp_state, TextInfo *textInfo, Dgp_Int32* bytelen, Dgp_Int8 **_final_str)
{
  DGP_PROLOGUE("ZCreateAsciiExtension()", DGP_PERRSTATE);
  XDOCData* xdoc_data = textInfo->xdoc_data;
  Dgp_Int32 num_chars = textInfo->num_chars;
  Dgp_Int32 num_words = textInfo->num_words;
  Dgp_Int32 *word_len_array = textInfo->word_lengths;

  encodestruct word_lengths, dlhsx, dlhsy,width,height;
  Dgp_Int8 *word_array,*final_str = NULL;
  Dgp_Int32 comp_word_len;
  finaldata encaps;
  Dgp_Int32 final_len;
  FILE* fp = NULL;

  DGP_CHECK_NULL(_final_str);

#ifdef DEBUG_EXTENSION
  if("extn.dat")
    remove("extn.dat");
  DGP_SYS_ERR_GEN_CLEANUP(fp = fopen("extn.dat","w"));

#endif

  /*Initialize the data in the encodestructs. */
  word_lengths.num_vals = num_words;
  word_lengths.vals = word_len_array;
  word_lengths.comp_length = comp_word_len = final_len =0;

  DGP_ERR_PASS_CLEANUP(init_encodestructs(
    &dlhsx,&dlhsy,&width,&height,num_chars,
      xdoc_data, dgp_state, fp));

  /*Encode and Compress the word_length_array */
#ifdef DEBUG_EXTENSION
  fprintf(fp,"Length of encoded word lengths ");
#endif
  DGP_ERR_PASS_CLEANUP(encode_and_compress(&word_lengths, dgp_state, fp));

  /*Encode and Compress the dlhsx */
#ifdef DEBUG_EXTENSION
  fprintf(fp,"Length of encoded dlhsx ");
#endif
  DGP_ERR_PASS_CLEANUP(encode_and_compress(&dlhsx, dgp_state, fp));

  /*Encode and Compress the dlhsy */
#ifdef DEBUG_EXTENSION
  fprintf(fp,"Length of encoded dlhsy ");
#endif
  DGP_ERR_PASS_CLEANUP(encode_and_compress(&dlhsy, dgp_state, fp));

  /*Encode and Compress the width */
#ifdef DEBUG_EXTENSION
  fprintf(fp,"Length of encoded width ");
#endif
  DGP_ERR_PASS_CLEANUP(encode_and_compress(&width, dgp_state, fp));

  /*Encode and Compress the height */
#ifdef DEBUG_EXTENSION
  fprintf(fp,"Length of encoded height ");
#endif
  DGP_ERR_PASS_CLEANUP(encode_and_compress(&height, dgp_state, fp));

  /*Now compress the actual words.. */
  DGP_ERR_PASS_CLEANUP(comp_word_array(xdoc_data, num_chars, dgp_state, &comp_word_len, 
    &word_array));

  encaps.word_arr = word_array;
  encaps.length = comp_word_len;
  encaps.word_lengths = &word_lengths;
  encaps.dlhsx = &dlhsx;
  encaps.dlhsy = &dlhsy;
  encaps.width = &width;
  encaps.height = &height;

  DGP_ERR_PASS_CLEANUP(concatenate_all(encaps,num_words,
              num_chars, dgp_state, fp, (Dgp_Int32*)&final_len, &final_str));

  free_alldata(&encaps, dgp_state);

  *bytelen = final_len;
  *_final_str = final_str;
  final_str = NULL;

#ifdef DEBUG_EXTENSION
  fflush(fp);
  fclose(fp);
#endif

  NORMAL_RETURN;

ERR_CLEANUP:

  if(final_str != NULL) (DGP_FREE)(final_str);

  ERR_RETURN;
} /* end of function ZCreateExtension */
/*////////////////////////////////////////////////////////////// */
/* function name: CleanupTextBuilder */
/*////////////////////////////////////////////////////////////// */
/*
void
CleanupTextBuilder(PosExtAsciiCookie pcookie)
{
  void (*localFree)(void*);

  localFree = pcookie->free;
  if(pcookie->zstr)
    localFree(pcookie->zstr);
  localFree(pcookie);
  return;
} */
/*//////////////////////////////////////////////////////////////////////////// */

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}



#endif
