/*///////////////////////////////////////////////////////////////////////// */
/* */
/*  DigiPaper Codebase */
/* */
/*  Ranjani Ramamurthy, Nov 2, 1998 */
/* */
/*  Filename: decode_ascii.c */
/* */
/*  Description of what's implemented in this file: */
/* */
/*  A modification of Table 1 of the xiff_3.0 spec. This version produces */
/*  encodings that are always byte aligned. We use this to encode all */
/*  values in the ascii position extensions and feed them into zlib to */
/*  generate the actual compressed extensions. */
/* */
/*  Here is what the table represents:: */
/* */
/*  if (value >= 0) { */
/*    if (value < 128) */
/*  write "0" plus 7 bits encoding value */
/*    else if (value < 128 + 8192) */
/*  write "100" plus 13 bits encoding (value - 128) */
/*    else */
/*  write "1010" plus 28 bits encoding (value - (128 + 8192)) */
/*    } */
/*  else { */
/*    if (value >= -64) */
/*  write "11" plus 6 bits encoding (value + 64) */
/*    else */
/*        write "1011" plus 28 bits encoding (-value - 65) */
/*    } */
/* */
/* Note that this always produces 1, 2, or 4 bytes of output for every input */
/* value; the most common values should also always be 1 byte.  Also, if the */
/* value is between -64 and +127, its encoding is one byte that is equal to */
/* the low 8 bits of the actual value. */
/* */
/* Modification History: */
/*    Dillon H. Tracy esq. / March 2000 */
/*    o  Threw in some directives for VxWorks port (e.g. malloc.h -> stdlib.h). */
/*///////////////////////////////////////////////////////////////////////// */
#ifndef LYNX
#include <wchar.h>
#endif
#include "ascii_exts.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#if defined(VXWORKS)
#include <stdlib.h>
#else
#include <malloc.h>
#endif

#include <zlib.h>
#include <zconf.h>
#include "dgp_sys_err_control.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libascii/decode_ascii.c,v 1.32 2003/02/10 22:53:19 mkrocken Exp $");

#define ZDECOMPRESS_ERROR  -1

extern Dgp_Int32 get_bit(Dgp_Int8 val, Dgp_Int32 i) ;

/* Statics for bounding box decode code. */
static ErrCode
free_int_arrays(Dgp_State *dgp_state, Dgp_Int32 *arr1, Dgp_Int32 *arr2, Dgp_Int32 *arr3,
          Dgp_Int32 *arr4);

static ErrCode
malloc_int_arrays(Dgp_State *dgp_state, Dgp_Int32 **arr1, Dgp_Int32 **arr2, Dgp_Int32 **arr3,
          Dgp_Int32 **arr4, Dgp_Int32 arr_sz);
/*////////////////////////////////////// */

/*///////////////////////////////////////////////////////////////////////////// */
/* function name: z_uncomp_array */
/*///////////////////////////////////////////////////////////////////////////// */

static ErrCode
z_uncomp_array(Dgp_Int8* src_array, Dgp_Int32 src_sz, Dgp_State *dgp_state,
             Dgp_Int8** dest_array, Dgp_Int32 est_uncomp_sz, Dgp_Int32* dest_sz)
{
  DGP_PROLOGUE("z_uncomp_array()", DGP_PERRSTATE);
  Dgp_Int8* ret_array = NULL;
  uLong ret_array_len;

  ret_array_len = est_uncomp_sz;
  DGP_CHECK_ALLOC_CLEANUP(ret_array = (DGP_CALLOC)(ret_array_len,1));

  DGP_ERR_GEN_CLEANUP((uncompress(ret_array,&ret_array_len,src_array,src_sz) != Z_OK),
    cErrCompressionError);

  *(dest_array) = ret_array;
  ret_array = NULL;
  *dest_sz = ret_array_len;
  
  NORMAL_RETURN;

ERR_CLEANUP:

  if(ret_array) (DGP_FREE)(ret_array);

  ERR_RETURN;
} /*end of function z_uncomp_array */

/*/////////////////////////////////////////////////////////// */
/* */
/* function name: decode_elements */
/* */
/*//////////////////////////////////////////////////////////// */
static void
decode_elements(Dgp_Int8 *input,Dgp_Int32 in_len,Dgp_Int32 num_elems,
                Dgp_Int32 **integer_vals)
{
  DGP_VOID_PROLOGUE("decode_elements()", NULL);
  Dgp_Int32 counter,bit,number;
  Dgp_Int8 cur_char;
  Dgp_Int8 *tmp_array;
  Dgp_Int32 val_cnt = 0;
  Dgp_Int32 *return_int_array;

#ifdef DEBUG_EXTENSION
FILE *fp;
#endif

  /* remember that the input is written in big endian.. */

  counter = bit = number = 0;
  tmp_array = (Dgp_Int8*)&number;
    return_int_array = *(integer_vals);
    val_cnt = 0;

#ifdef DEBUG_EXTENSION
  if("int.vals")
    remove("int.vals");
  fp = fopen("int.vals","w");
  assert(fp != (FILE*)NULL);
#endif


  while( counter < in_len){
    number = 0;
    cur_char = input[counter];
     bit = get_bit(cur_char,7); /*get the MSB */
    if(bit == 0){
        return_int_array[val_cnt] = (Dgp_Int32)cur_char;
      counter+=ONEBYTE;
    } else {
      bit = get_bit(cur_char,6);
      if(bit == 0){
        bit = get_bit(cur_char,5);
        if(bit == 0){ /*100 (2 bytes) */
         cur_char &= 0x1f; /*set the first three bits to 0 */
          tmp_array[BYTELSB] = input[counter+1];
          tmp_array[BYTEBEFLSB] = cur_char;
          tmp_array[BYTEBEFMSB] = 0;
          tmp_array[BYTEMSB] = 0;
          number += 128;
            return_int_array[val_cnt] = number;
        counter += TWOBYTES;
        } else {
          bit = get_bit(cur_char,4);
          if(bit == 0){  /* "1010" (4 bytes) */
           cur_char &= 0x0f; /*set the first four bits to 0 */

            tmp_array[BYTELSB] = input[counter+3];
            tmp_array[BYTEBEFLSB] = input[counter+2];
            tmp_array[BYTEBEFMSB] = input[counter+1];
            tmp_array[BYTEMSB] = cur_char;

            number += 8320;
        return_int_array[val_cnt] = number;
          counter += FOURBYTES;
          } else { /* "1011" (4 bytes) */
           cur_char &= 0x0f; /*set the first four bits to 0 */

            tmp_array[BYTELSB] = input[counter+3];
            tmp_array[BYTEBEFLSB] = input[counter+2];
            tmp_array[BYTEBEFMSB] = input[counter+1];
            tmp_array[BYTEMSB] = cur_char;

            number = -(number + 65);
        return_int_array[val_cnt] = number;
          counter += FOURBYTES;
          }
        }
      } else { /* 11 */
       /* This fits in one byte and is bet -1 and -64 */
       cur_char &= 0x3f;
       cur_char -= 64;
     return_int_array[val_cnt] = (Dgp_Int32)cur_char;
       counter += ONEBYTE;
      }
    }
#ifdef DEBUG_EXTENSION
    fprintf(fp,"integer[%d] is %d\n",
          val_cnt,return_int_array[val_cnt]);
    fflush(fp);
#endif
    val_cnt++; /*attacking the next number */
  }
  /*sanity check */
  DGP_VOID_ASSERT(val_cnt == num_elems);
  /* stop compiler whining */
#ifndef DGP_DBUG
  (void)num_elems;
#endif

#ifdef DEBUG_EXTENSION
  if(fp){
    fflush(fp);
    fclose(fp);
  }
#endif

  return;
} /*end of function decode_elements */
/*///////////////////////////////////////////////////////////// */
/* */
/* function name: data_length */
/* */
/* */
/* */
/*////////////////////////////////////////////////////////////// */
static Dgp_Int32
data_length(Dgp_UInt8* array, Dgp_Int32* curr_off){
Dgp_Int32 INFO_SIZE = 2;
Dgp_Int32 ret_val;
   ret_val = (array[*curr_off] << 8) | (array[*curr_off+1]);
   *(curr_off) += INFO_SIZE;
   return ret_val;
} /*end of function data_length */



/*///////////////////////////////////////////////////////////// */
/* function name: uncomp_and_decode */
/* */
/*///////////////////////////////////////////////////////////// */
static ErrCode
uncomp_and_decode(Dgp_State *dgp_state, Dgp_Int8 *ext,Dgp_Int32 *curr_offset,Dgp_Int32 **int_array,
                  Dgp_Int32 num_elems)
{
  DGP_PROLOGUE("uncomp_and_decode()", DGP_PERRSTATE);
  Dgp_Int32 comp_data_sz, uncomp_length;
  Dgp_Int8 *uncomp_array = NULL;
  Dgp_Int32 est_uncomp_sz;
  comp_data_sz = uncomp_length = 0;
  /*read the size of the compressed data item */

  comp_data_sz = data_length((Dgp_UInt8*)ext,curr_offset);
  est_uncomp_sz = sizeof(Dgp_Int32) * num_elems;
  DGP_ERR_PASS_CLEANUP(z_uncomp_array(ext+(*curr_offset),comp_data_sz,
    dgp_state, &uncomp_array, est_uncomp_sz,&uncomp_length));
  (*curr_offset) += comp_data_sz;

  /*Decode the integer numbers and fill the int array */
  decode_elements(uncomp_array,uncomp_length,num_elems,
                  int_array);

  if(uncomp_array) (DGP_FREE)(uncomp_array);
  uncomp_array = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(uncomp_array) (DGP_FREE)(uncomp_array);

  ERR_RETURN;
}/*end of function uncomp_and_decode */

/*#ifdef DEBUG_EXTENSION */
/*///////////////////////////////////////////////////////////// */
/* function name: malloc_int_arrays */
/* */
/*///////////////////////////////////////////////////////////// */
static ErrCode
malloc_int_arrays(Dgp_State *dgp_state, Dgp_Int32 **arr1, Dgp_Int32 **arr2, Dgp_Int32 **arr3,
          Dgp_Int32 **arr4, Dgp_Int32 arr_sz)
{
  DGP_PROLOGUE("malloc_int_arrays()", DGP_PERRSTATE);
  Dgp_Int32 *l1, *l2, *l3, *l4;
  l1 = l2 = l3 = l4 = NULL;

  DGP_CHECK_ALLOC_CLEANUP(l1 = DGP_CALLOC(arr_sz, sizeof(Dgp_Int32)));
  DGP_CHECK_ALLOC_CLEANUP(l2 = DGP_CALLOC(arr_sz, sizeof(Dgp_Int32)));
  DGP_CHECK_ALLOC_CLEANUP(l3 = DGP_CALLOC(arr_sz, sizeof(Dgp_Int32)));
  DGP_CHECK_ALLOC_CLEANUP(l4 = DGP_CALLOC(arr_sz, sizeof(Dgp_Int32)));

  *(arr1) = l1; *(arr2) = l2; *(arr3) = l3; *(arr4) = l4;

  NORMAL_RETURN;
ERR_CLEANUP:
  DGP_ERR_PASS(free_int_arrays(dgp_state, l1, l2, l3, l4));
  ERR_RETURN;
} /* end of function malloc_int_arrays */
/*///////////////////////////////////////////////////////////// */
/* function name: free_int_arrays */
/* */
/*///////////////////////////////////////////////////////////// */
static ErrCode
free_int_arrays(Dgp_State *dgp_state, Dgp_Int32 *arr1, Dgp_Int32 *arr2, Dgp_Int32 *arr3,
          Dgp_Int32 *arr4)
{
  DGP_PROLOGUE("free_int_arrays()", DGP_PERRSTATE);
  DGP_STOP_WHINE;
  if(arr1) DGP_FREE(arr1);
  if(arr2) DGP_FREE(arr2);
  if(arr3) DGP_FREE(arr3);
  if(arr4) DGP_FREE(arr4);

  NORMAL_RETURN;
} /*end of function free_int_arrays */

/*//////////////////////////////////////////////////////////// */
/* function name: print_char_bb_info */
/* */
/*///////////////////////////////////////////////////////////// */
static void
print_char_info(   Dgp_State    *dgp_state, 
                   Dgp_Int32    *lhsx_vals, 
                   Dgp_Int32    *lhsy_vals,
                   Dgp_Int32    *uncomp_width_array,
                   Dgp_Int32    *uncomp_height_array,		    
                   Dgp_Int8     *uncomp_array,
                   FILE         *outfile,
                   Dgp_Int32    *word_lengths, 
                   Dgp_Int32    num_words)
{
    DGP_PROLOGUE("print_char_bb_info()", DGP_PERRSTATE);

    Dgp_Int32 prev_rt_x, prev_rt_y, prev_width;
    Dgp_Int32 x1_val, y1_val, x2_val, y2_val;
    Dgp_UInt8 first = 1; 
    wchar_t c;    

    Dgp_Int32 i,j,offset;
    Dgp_Int32 num_chars;
    Dgp_Int32 total_chars = 0;

    DGP_STOP_WHINE;

    prev_rt_x = prev_rt_y = prev_width = 0;
    x1_val = y1_val = x2_val = y2_val = 0;
    offset = 0;
    
    for(i = 0; i< num_words; i++){
        total_chars += word_lengths[i];
    }
    for(i = 0; i < num_words; i++) {
        num_chars = word_lengths[i];
        for (j = 0; j < num_chars; j++){ 
            c = (Dgp_UInt8)uncomp_array[offset + j] + 
                (((Dgp_UInt8)uncomp_array[offset + total_chars +j]) << 8);
            if (c == L' ') {
                j++;
                c = (Dgp_UInt8)uncomp_array[offset + j] + 
                    (((Dgp_UInt8)uncomp_array[offset + total_chars +j]) << 8);
            }
            if (first) {
                /* Format: x of upper left <tab> y of upper left <tab> width <tab> height */
                fprintf(outfile, "%x\t%x\t%d\t%d\t%d\t%d\n", (int)(c >> 8), (int)(c & 0xff),(int)lhsx_vals[0], 
                    (int)lhsy_vals[0], (int)uncomp_width_array[0], (int)uncomp_height_array[0]);
                prev_rt_x = lhsx_vals[0] + uncomp_width_array[0];
                prev_rt_y = lhsy_vals[0];
                prev_width = uncomp_width_array[0];
                first = 0;
            } else {
                x1_val = prev_rt_x + lhsx_vals[offset + j];
                y1_val = prev_rt_y + lhsy_vals[offset + j];
                x2_val = x1_val + (Dgp_Int32)uncomp_width_array[offset + j];
                y2_val = y1_val + (Dgp_Int32)uncomp_height_array[offset + j];
                fprintf(outfile, "%x\t%x\t%d\t%d\t%d\t%d\n", (int)(c>>8), (int)(c & 0xff), (int)x1_val, (int)y1_val, 
                    (int)uncomp_width_array[offset + j], (int)uncomp_height_array[offset + j]);
                prev_rt_x = x2_val;
                prev_rt_y = y1_val;
            }         
        }
        offset += word_lengths[i];
    } /*end of the for loop */
    fflush(outfile);  
    
    /*fclose(fp); */
    return;
}/*end of function print_char_info */

/*///////////////////////////////////////////////////////////// */
/* */
/* function name: dump_text_file */
/* */
/* */
/* */
/*////////////////////////////////////////////////////////////// */
static ErrCode
dump_text_file(Dgp_State    *dgp_state, 
               Dgp_Int8     *uncomp_array,
               FILE         *outfile,
               Dgp_Int32    *word_lengths, 
               Dgp_Int32    num_words, 
               Dgp_Int32    *_num_chars)
{
  DGP_PROLOGUE("dump_text_file()", DGP_PERRSTATE);
  Dgp_Int32 i,cr,offset;
  Dgp_Int32 num_chars;

  num_chars = cr = 0; 
  offset = 0;
  for(i = 0; i < num_words; i++) {
    num_chars += word_lengths[i];
    DGP_SYS_ERR_GEN_CLEANUP(fwrite(uncomp_array+offset, word_lengths[i], 1, outfile) < 1);
    fprintf(outfile, " ");
    offset += word_lengths[i];
    cr++;
    if(cr == 10){ /*This is just to make the text output readable */
      fprintf(outfile, "\n");
      fflush(outfile);
        cr = 0;
    }
  } /*end of the for loop */
  
  *_num_chars = num_chars;

  fprintf(outfile, "\n");

  NORMAL_RETURN;

ERR_CLEANUP:

  *_num_chars = num_chars - word_lengths[i];

  ERR_RETURN;
} /*end of function dump_text_file */

/*///////////////////////////////////////////////////////////// */
/* */
/* function name: uncompress_all */
/* */
/* */
/* */
/*////////////////////////////////////////////////////////////// */
static ErrCode
uncompress_all(Dgp_State    *dgp_state, 
               Dgp_Int8     *ext, 
               FILE         *fp,
               FILE         *fp_asc,
               WordCoOrd    **char_info)
{
  DGP_PROLOGUE("uncompress_all()", DGP_PERRSTATE);
  Dgp_Int32 curr_offset, num_words, comp_data_sz, num_chars;
  Dgp_Int8 *uncomp_array = NULL; Dgp_Int32 *word_lengths = NULL, uncomp_length;
  Dgp_Int32 index, this_offset = 0;
  Dgp_Int32 *dlhsx, *dlhsy, *width, *height;
  Dgp_Int32 est_uncomp_length;

#ifndef DEBUG_EXTENSION
  /* it's used! */
  char_info = char_info;
#endif

  DGP_CHECK_NULL(dgp_state);
  DGP_CHECK_NULL(fp);
  DGP_CHECK_NULL(ext);

  curr_offset = num_words = comp_data_sz = 0;
  uncomp_length = num_chars = 0;

  /*skip the first byte, which is the version number */
  curr_offset = 1;

  /*Read the number of words in the file */
  num_words = data_length((Dgp_UInt8*)ext, &curr_offset);
  DGP_CHECK_ALLOC_CLEANUP(word_lengths = (DGP_CALLOC)(num_words, sizeof(Dgp_Int32)));

  DGP_ERR_PASS_CLEANUP(uncomp_and_decode(dgp_state, ext, &curr_offset, 
    &word_lengths, num_words));
  for (index = 0; index < num_words; index++)
    num_chars += word_lengths[index];

  /*Read the length of the compressed words */
  comp_data_sz = data_length((Dgp_UInt8*)ext, &curr_offset);
  /* adding estimated uncomp length */
  est_uncomp_length = num_chars *2 ;
  DGP_ERR_PASS_CLEANUP(z_uncomp_array(ext + curr_offset, comp_data_sz, dgp_state,
    &((Dgp_Int8*)uncomp_array), est_uncomp_length,&uncomp_length));
  curr_offset += comp_data_sz;


  DGP_ERR_PASS(malloc_int_arrays(dgp_state, &dlhsx, &dlhsy, &width, 
				 &height, num_chars));

  ext += curr_offset;

  DGP_ERR_PASS_CLEANUP(uncomp_and_decode(dgp_state, ext, &this_offset, &dlhsx, num_chars));
  DGP_ERR_PASS_CLEANUP(uncomp_and_decode(dgp_state, ext, &this_offset, &dlhsy, num_chars));
  DGP_ERR_PASS_CLEANUP(uncomp_and_decode(dgp_state, ext, &this_offset, &width, num_chars));
  DGP_ERR_PASS_CLEANUP(uncomp_and_decode(dgp_state, ext, &this_offset, &height, num_chars));

  print_char_info(dgp_state, dlhsx, dlhsy, width, height,
		         uncomp_array, fp, word_lengths, num_words);
  if(fp_asc)  /* if we want ascii text  */
       DGP_ERR_PASS_CLEANUP(dump_text_file(dgp_state, uncomp_array, fp_asc, word_lengths, num_words, &num_chars));
  DGP_ERR_PASS(free_int_arrays(dgp_state, dlhsx, dlhsy, width, height));

  if(uncomp_array) (DGP_FREE)(uncomp_array);
  uncomp_array = NULL;
  if(word_lengths) (DGP_FREE)(word_lengths);
  word_lengths = NULL;


  NORMAL_RETURN;

ERR_CLEANUP:

  if(uncomp_array) DGP_FREE(uncomp_array);
  if(word_lengths) DGP_FREE(word_lengths);  

  ERR_RETURN;
}/*end of static function uncompress_all */




/*///////////////////////////////////////////////////////////// */
/* */
/* function name: DecodeTextExtension */
/* */
/* called from file decode_annot.c in libtc_util */
/* */
/*////////////////////////////////////////////////////////////// */
DGP_LIBEXPORT ErrCode
DecodeTextExtension(Dgp_State *dgp_state, Dgp_Int32 length, void *extension, FILE *fp, FILE *fp_asc)
{
  DGP_PROLOGUE("DecodeTextExtension()", DGP_PERRSTATE);
  WordCoOrd *char_info = NULL;


  if(length == 0) NORMAL_RETURN;

  DGP_CHECK_NULL(fp);
  DGP_CHECK_NULL(extension);

  /*First we need to uncompress the lot..then we can */
  /*decode it. */

  /*uncompress_all interfaces with zlib and uncompresses */
  /*the text data. it returns only the character co-ord */
  /*information - it dumps the text out into a file, */
  /*as it is being uncompressed. there is no need to store */
  /*the text in a data structure. we actually dont need */
  /*the co-ordinate information in the c code. we use it */
  /*only for debugging purposes. */

/*  DGP_ERR_PASS(uncompress_all_old(dgp_state, (Dgp_Int8*)extension, fp, fp_asc, &char_info)); */
  DGP_ERR_PASS(uncompress_all (dgp_state, (Dgp_Int8*)extension, fp, fp_asc, &char_info));

  NORMAL_RETURN;
} /* end of function DecodeTextExtension */

/*//////////////////////////////////////////////////////////////// */


