/*//////////////////////////////////////////////////////////////////////////  */
/*  */
/*  DigiPaper Codebase  */
/*      */
/*  Ranjani Ramamurthy, Nov 2, 1998  */
/*  */
/*  Filename: encode_ascii.c  */
/*  */
/*  Description of whats implemented in this file:  */
/*  */
/*  A modification of Table 1 of the xiff_3.0 spec. This version produces  */
/*  encodings that are always byte aligned. We use this to encode all  */
/*  values in the ascii position extensions and feed them into zlib to  */
/*  generate the actual compressed extensions.  */
/*    */
/*  Here is what the table represents::  */
/*  */
/*  if (value >= 0) {  */
/*    if (value < 128)  */
/*      write "0" plus 7 bits encoding value  */
/*    else if (value < 128 + 8192)  */
/*      write "100" plus 13 bits encoding (value - 128)  */
/*    else  */
/*      write "1010" plus 28 bits encoding (value - (128 + 8192))  */
/*    }  */
/*  else {  */
/*    if (value >= -64)  */
/*      write "11" plus 6 bits encoding (value + 64)  */
/*    else  */
/*      write "1011" plus 28 bits encoding (-value - 65)  */
/*    }  */
/*  */
/* Note that this always produces 1, 2, or 4 bytes of output for every input  */
/* value; the most common values should also always be 1 byte.  Also, if the  */
/* value is between -64 and +127, its encoding is one byte that is equal to  */
/* the low 8 bits of the actual value.  */
/*  */
/* Modification History: */
/*    Dillon H. Tracy esq. / March 2000 */
/*    o  Threw in some directives for VxWorks port (e.g. malloc.h -> stdlib.h).  */
/*/////////////////////////////////////////////////////////////////////////  */
#ifndef LYNX
#include <wchar.h> 
#endif
#include "ascii_exts.h" 
#include <stdio.h> 
#include <stdlib.h> 

#if defined(VXWORKS)
#include <stdlib.h>
#else
#include <malloc.h>
#endif

#include <assert.h> 

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libascii/encode_ascii.c,v 1.11 2003/02/10 22:53:19 mkrocken Exp $");

/*///////////////////////////////////////////////////////////  */
/*  */
/* function name: EncodeElements  */
/*  */
/*///////////////////////////////////////////////////////////  */
void 
EncodeElements(Dgp_Int32* vals, Dgp_Int32 nel,  
         Dgp_Int32* ret_len, Dgp_Int8** encoded_arr) 
{ 
Dgp_Int32 i, tmp_val; 
Dgp_Int32 ptr_to_input = 0; 
Dgp_Int8 inval; 
Dgp_Int8* ret_array; 
 
  /*We are going to write the encoded array  out in   */
  /*explicit big endian format  */
 
  /* the encoded array has already been allocated in the  */
  /* calling program.  */
 
    ret_array = *(encoded_arr); 
 
  for(i = 0; i < nel; i++){ 
    inval = 0; tmp_val = 0; 
    if(vals[i] >= 0){ 
      if( vals[i] < 128 ) { /*will fit in 1 byte  */
        inval = get_byte(vals[i],BYTELSB);  
        ret_array[ptr_to_input] = inval; 
        ptr_to_input += ONEBYTE; 
      } else if ( vals[i] < 8320){ /*will fit in 2 bytes  */
        tmp_val = vals[i] - 128; 
        inval = get_byte(tmp_val,BYTEBEFLSB); 
        inval |= 0x80; /*set the first three bits to 100  */
        ret_array[ptr_to_input] = inval; 
        ret_array[ptr_to_input+1] = get_byte(tmp_val,BYTELSB);  
        ptr_to_input += TWOBYTES; 
      } else { /*will fit in 4 bytes  */
        tmp_val = vals[i] - 8320; 
        inval = get_byte(tmp_val,BYTEMSB); 
        inval |= 0xa0; /*set the first 4 bits to '1010'  */
        ret_array[ptr_to_input] = inval; 
        ret_array[ptr_to_input+1] = get_byte(tmp_val,BYTEBEFMSB); 
        ret_array[ptr_to_input+2] = get_byte(tmp_val,BYTEBEFLSB); 
        ret_array[ptr_to_input+3] = get_byte(tmp_val,BYTELSB); 
        ptr_to_input += FOURBYTES; 
      } 
    } else { /*val is < 0  */
      if( vals[i] >= -64 ) { 
        tmp_val = vals[i] + 64; 
        inval = get_byte(tmp_val,BYTELSB);  
        inval |= 0xc0; /*set the first 2 bits to '11'  */
        ret_array[ptr_to_input] = inval; 
        ptr_to_input += ONEBYTE ; 
      } else { /*will fit in 4 bytes  */
        tmp_val = -(vals[i]) - 65; 
        inval = get_byte(tmp_val,BYTEMSB); 
        inval |= 0xb0; /* set the first four bits to '1011'  */
        ret_array[ptr_to_input] = inval; 
        ret_array[ptr_to_input+1] = get_byte(tmp_val,BYTEBEFMSB); 
        ret_array[ptr_to_input+2] = get_byte(tmp_val,BYTEBEFLSB); 
        ret_array[ptr_to_input+3] = get_byte(tmp_val,BYTELSB); 
        ptr_to_input += FOURBYTES; 
      } 
    } /*end of else for val < 0  */
  } /*end of the for loop  */
 
  *ret_len = ptr_to_input; 
  return; 
 
} /* end of function EncodeElements()  */
 
/*////////////////////////////////////////////////////////////////////  */
 
