/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libblat/blit.c,v 1.9 2003/02/10 22:53:19 mkrocken Exp $
 *
 */

/*
 * Module blit
 *
 * Exports:
 *	void blitBitsNoClip(Dgp_UInt32 *dest, Dgp_Int32 destlpl,
 *			    Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 srcw, Dgp_Int32 srch,
 *			    Dgp_Int32 xpos, Dgp_Int32 ypos) - draw src into dest at
 *		(xpos, ypos).  No clipping is done.  Pad bits of src must
 *		be zero.  xpos,ypos must be >= 0.  This routine doesn't
 *		know the dest size, so src had better fit into dest.
 *
 *	void blitBitsClipLR(Dgp_UInt32 *dest, Dgp_Int32 destlpl,
 *			    Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 minx,
 *			    Dgp_Int32 srcw, Dgp_Int32 srch, Dgp_Int32 xpos, Dgp_Int32 ypos) -
 *		draw src into dest such that the pixel at (minx, 0) of
 *		src is placed at (xpos, ypos).  Bits of src outside
 *		(minx, minx + width - 1) are clipped.  No vertical clipping
 *		is done.  This routine doesn't know the dest size, so the
 *		clipped version of src had better fit in the dest.
 *		xpos,ypos must be >= 0.
 *
 *	void blitBits(Dgp_UInt32 *dest, Dgp_Int32 destlpl,
 *		      Dgp_Int32 destminx, Dgp_Int32 destminy, Dgp_Int32 destmaxx, Dgp_Int32 destmaxy,
 *		      Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 srcw, Dgp_Int32 srch,
 *		      Dgp_Int32 xpos, Dgp_Int32 ypos) - draw src into dest at (xpos, ypos).
 *		Full clipping is done.  Pad bits of src must be zero.  Negative
 *		values of xpos and ypos are allowed.  dest(min|max)[xy] provide
 *		bounds on the region of dest that can be touched (inclusive
 *		on min, exclusive on max).
 *
 *	void blitPacked(DgpImage_Packed dest, DgpImage_Packed src,
 *			Dgp_Int32 xpos, Dgp_Int32 ypos) - draw src into dest so that
 *		src's (0,0) pixel is placed at (xpos, ypos).  Full clipping
 *		is done.  Pad bits of src must be zero.  Offsets of src
 *		and dest may be non-zero; all of the data area of src is
 *		drawn in any case.
 *
 *	void blitBitsOp(Dgp_UInt32 *dest, Dgp_Int32 destlpl,
 *		        Dgp_Int32 destminx, Dgp_Int32 destminy, Dgp_Int32 destmaxx, Dgp_Int32 destmaxy,
 *		        Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 srcw, Dgp_Int32 srch,
 *		        Dgp_Int32 xpos, Dgp_Int32 ypos, Dgp_Int32 blitop) - draw src into dest
 *		at (xpos, ypos).
 *		Full clipping is done.  Pad bits of src must be zero.  Negative
 *		values of xpos and ypos are allowed.  Use the provided
 *		operator.  If the operator is DGPBLAT_OP_OR, this is equivalent
 *		to blitBits.
 *
 *	void blitPackedOp(DgpImage_Packed dest, DgpImage_Packed src,
 *			  Dgp_Int32 xpos, Dgp_Int32 ypos, Dgp_Int32 blitop) - draw src into dst
 *		using the provided operator.  If the operator is DGPBLAT_OP_OR,
 *		this is equivalent to blitPacked.
 *
 *	void blitFillPacked(DgpImage_Packed dest) - fill all the data area of
 *		dest with '1'.
 */

#include "util.h"
#include "image.h"
#include "image_defs.h"
#include "blit.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libblat/blit.c,v 1.9 2003/02/10 22:53:19 mkrocken Exp $");

void
DgpBlat_blitBitsNoClip(Dgp_UInt32 *dest, Dgp_Int32 destlpl,
	       Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 srcw, Dgp_Int32 srch,
	       Dgp_Int32 xpos, Dgp_Int32 ypos, Dgp_State *dgp_state)
{
    Dgp_Int32 l, i, j;
    Dgp_UInt32 *dptr, *sptr, *drowptr, *srowptr, carry;
    Dgp_Int32 rshift, lshift;
    Dgp_Bool extraword;

    /* Unused..  /dhtesq. */
    dgp_state = dgp_state;
  
    l = xpos / DGP_BITS_UINT32;	/* which long to start at */
    rshift = (xpos % DGP_BITS_UINT32);
    lshift = DGP_BITS_UINT32 - rshift;
    if (srcw + rshift - srclpl * DGP_BITS_UINT32 > 0) {
	extraword = DGP_TRUE;
	}
    else {
	extraword = DGP_FALSE;
	}
  
    if (rshift) {
	for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl + l;
	     j < srch;
	     j++, srowptr += srclpl, drowptr += destlpl) {
	    dptr = drowptr;
	    sptr = srowptr;
      
	    carry = 0;
	    for (i = 0; i < srclpl; i++) {
		*dptr++ |= ((*sptr >> rshift) | carry);
		carry = *sptr++ << lshift;
		}
      
	    if (extraword) {
		/* Still some bits in carry */
		*dptr |= carry;
		}
	    }
	}
  
    else {
	/* No shifting! */
	for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl + l;
	     j < srch;
	     j++, srowptr += srclpl, drowptr += destlpl) {
	    dptr = drowptr;
	    sptr = srowptr;
      
	    for (i = 0; i < srclpl; i++)
		*(dptr++) |= *(sptr++);
	    }
	}
    }

/*
 * Blit, clipping on the L and R.  Don't worry about clipping above and below.
 * The region to be drawn starts at minx and
 * is srcw pixels wide.  The pixel at (minx, 0) gets put at (xpos, ypos).
 *
 * 21.8.2000 / dhtesq.
 * Decoder profiling suggests this is a good place for an inline...
 *
 */
#ifdef _WIN32
DGP_LIBEXPORT
#endif
INLINE void 
DgpBlat_blitBitsClipLR(Dgp_UInt32 *dest, Dgp_Int32 destlpl,
	       Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 minx, Dgp_Int32 srcw, Dgp_Int32 srch,
	       Dgp_Int32 xpos, Dgp_Int32 ypos, Dgp_State *dgp_state)
{
    Dgp_Int32 l, i, j;
    Dgp_UInt32 *dptr, *sptr, *drowptr, *srowptr, carry;
    Dgp_Int32 rshift, lshift;
    Dgp_Bool extraword, skipfirst;
    Dgp_UInt32 leftmask, rightmask;
    Dgp_Int32 validlongsperline;

    DGP_PROLOGUE ("DgpBlat_blitBitsClipLR", DGP_PERRSTATE);
    DGP_STOP_WHINE;

    /* Trim off words */
    src += minx / DGP_BITS_UINT32;
    minx &= (DGP_BITS_UINT32 - 1);

    /* How many source longs contain valid data? */
    validlongsperline = (minx + srcw - 1) / DGP_BITS_UINT32 + 1;
  
    /*
     * The idea is that this code will parallel the code above, except
     * that it will apply a mask to the first and last source longs,
     * and may skip the first destination long if it knows that no bits
     * will be written to it.
     */

    /*
     * Deselect the first minx bits.
     * Boundaries: minx = 0 -> all 1s
     * 		   minx = 31 -> 1 in LSB; rest 0
     */
    leftmask = (~0ul) >> minx;
    /*
     * Deselect bits past (and including) minx+srcw
     * Boundaries: (minx+srcw)%32 == 0 -> all 1s (this is the previous long)
     * 		   (minx+srcw)%32 == 1 -> 1 in MSB; rest 0
     *		   (minx+srcw)%32 == 31 -> 0 in LSB; rest 1
     */
    if (((minx + srcw) & (DGP_BITS_UINT32 - 1)) == 0) {
	rightmask = ~0ul;
	}
    else {
	rightmask = (~0ul) <<
	    (DGP_BITS_UINT32 - 1 - ((minx + srcw - 1) & (DGP_BITS_UINT32 - 1)));
	}
	
    /* Do rightmask and leftmask apply to the same word? */
    if (validlongsperline == 1) {
	/* Yes - AND them together to get the only-word mask */
	leftmask &= rightmask;
	}

    /* Can we skip the first dest long? */
    if ((xpos - minx < 0) ||
	(((xpos - minx) / DGP_BITS_UINT32) != (xpos / DGP_BITS_UINT32))) {
	/* Yes, we can skip it */
	skipfirst = DGP_TRUE;
	}
    else {
	skipfirst = DGP_FALSE;
	}

    /*
     * Which dest long do we start at?  This has already had skipfirst
     * applied to it (it's the first dest long we actually touch).
     */
    l = xpos / DGP_BITS_UINT32;	/* which dest long to start at */

    rshift = (xpos - minx) & (DGP_BITS_UINT32 - 1);
    lshift = DGP_BITS_UINT32 - rshift;

    /*
     * Does the last source longword straddle two dest longwords?
     * The source has (srcw + minx) active bits, ending in longword
     * number (srcw + minx - 1) / DGP_BITS_UINT32.  That last longword has
     * (srcw + minx - 1) % DGP_BITS_UINT32 + 1 good bits in it
     * (32 if minx+srcw%32 == 0, 1 if minx+srcw%32=1).  The first bit in
     * that longword is source bit number (minx + srcw - 1) & ~31;
     * the last bit is source bit number (minx + srcw - 1).  So the
     * question is, does the dest bit at (minx + srcw - 1 + xpos - minx)
     * lie in the same longword as the dest bit at
     * (((minx + srcw - 1) & ~31) + xpos - minx)?
     */
    if (((((minx + srcw - 1) & ~(DGP_BITS_UINT32 - 1)) + xpos - minx) / DGP_BITS_UINT32) ==
	((srcw - 1 + xpos) / DGP_BITS_UINT32)) {
	/* They're in the same word - no extra needed. */
	extraword = DGP_FALSE;
	}
    else {
	extraword = DGP_TRUE;
	}
  
    if (rshift) {
	for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl + l;
	     j < srch;
	     j++, srowptr += srclpl, drowptr += destlpl) {
	    dptr = drowptr;
	    sptr = srowptr;
      
	    /*
	     * First word - apply leftmask.  If skipfirst is DGP_TRUE, then
	     * we don't store the left side anywhere, and don't need to
	     * increment dptr either; just ignore those left bits.
	     */
	    if (!skipfirst) {
		*dptr++ |= ((*sptr & leftmask) >> rshift);
		}
	    carry = (*sptr++ & leftmask) << lshift;

	    /* Subsequent words, not including last one */
	    for (i = 1; i < validlongsperline - 1; i++) {
		*dptr++ |= ((*sptr >> rshift) | carry);
		carry = *sptr++ << lshift;
		}
	    if (validlongsperline > 1) {
		/* Still a word to go */
		*dptr++ |= (((*sptr & rightmask) >> rshift) | carry);
		carry = (*sptr & rightmask) << lshift;
		}
	    else if (skipfirst) {
		/*
		 * There's only one valid dest long, and the bits in
		 * carry haven't been stored anywhere.  Better store
		 * them now.
		 */
		*dptr |= carry;
		}
      
	    if (extraword) {
		/* Still some bits in carry */
		*dptr |= carry;
		}
	    }
	}
    else {
	/* No shifting! */
	DGP_VOID_ASSERT(!skipfirst);
	DGP_VOID_ASSERT(!extraword);

	for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl + l;
	     j < srch;
	     j++, srowptr += srclpl, drowptr += destlpl) {
	    dptr = drowptr;
	    sptr = srowptr;
      
	    *(dptr++) |= *(sptr++) & leftmask;
	    for (i = 1; i < validlongsperline - 1; i++) {
		*(dptr++) |= *(sptr++);
		}
	    if (validlongsperline > 1) {
		*(dptr++) |= *(sptr++) & rightmask;
		}
	    }
	}
    }

/*
 * Blit, clipping on the L and R.  Don't worry about clipping above and below.
 * The region to be drawn starts at minx and
 * is srcw pixels wide.  The pixel at (minx, 0) gets put at (xpos, ypos).
 * Replace the pixels of dest with the corresponding pixels of src
 * where src pixels are valid.
 */
static void
DgpBlat_blitBitsClipLRREPLACE(Dgp_UInt32 *dest, Dgp_Int32 destlpl,
		      Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 minx,
		      Dgp_Int32 srcw, Dgp_Int32 srch,
		      Dgp_Int32 xpos, Dgp_Int32 ypos, Dgp_State *dgp_state)
{
    Dgp_Int32 i, j;
    Dgp_UInt32 *dptr, *sptr, *drowptr, *srowptr, carry;
    Dgp_Int32 rshift, lshift;
    Dgp_Bool extraword;
    Dgp_UInt32 leftmask, rightmask;
    Dgp_Int32 destlongstouched, srclongstouched;

    /* Unused... / dhtesq. */
    dgp_state = dgp_state;

    /*
     * Here's how this code works.
     * We have a bunch of bits in each row of src (bits number
     * minx ... minx + srcw - 1) that we want to insert into the destination
     * row, bits number xpos ... xpos + srcw - 1.  Sounds easy, right?
     *
     * We can adjust the src pointer and dest pointer so that the long
     * containing the first active bit is always the first in the row.
     */
    src += minx / DGP_BITS_UINT32;
    minx &= DGP_BITS_UINT32 - 1;

    dest += xpos / DGP_BITS_UINT32;
    xpos &= DGP_BITS_UINT32 - 1;

    /*
     * Now minx < DGP_BITS_UINT32, xpos < DGP_BITS_UINT32.
     * Our loop is based on the number of destination longs that get
     * touched per row.  That is equal to
     * (xpos + srcw - 1) / DGP_BITS_UINT32 - xpos / DGP_BITS_UINT32 + 1,
     * since those are the indices of the longwords containing the first and
     * last longs touched.  Of course xpos / DGP_BITS_UINT32 = 0 now.
     */
    destlongstouched = (xpos + srcw - 1) / DGP_BITS_UINT32 + 1;

    /*
     * We're going to need some masks to protect the parts of the destination
     * that shouldn't be touched.  We'll have two masks for this:
     * leftmask, which is a mask that has a 1 where the dest bits in the
     * first dest long should be touched
     * rightmask, which has a 1 for every touched bit in the last dest long
     *
     * If the dest only has one long, these are combined into leftmask.
     */
    
    /*
     * leftmask has the first xpos bits deselected, the rest selected.
     * Boundaries: xpos = 0 -> all 1s
     * 		   xpos = 31 -> 1 in LSB; rest 0
     */
    leftmask = (~0ul) >> xpos;
    /*
     * Deselect bits past (and including) xpos+srcw, select up to that.
     * Boundaries: (xpos+srcw)%32 == 0 -> all 1s (this is the previous long)
     * 		   (xpos+srcw)%32 == 1 -> 1 in MSB; rest 0
     *		   (xpos+srcw)%32 == 31 -> 0 in LSB; rest 1
     */
    if (((xpos + srcw) & (DGP_BITS_UINT32 - 1)) == 0) {
	rightmask = ~0ul;
	}
    else {
	rightmask = (~0ul) <<
	    (DGP_BITS_UINT32 - 1 - ((xpos + srcw - 1) & (DGP_BITS_UINT32 - 1)));
	}

    /* Is there only one valid dest long? */
    if (destlongstouched == 1) {
	/* Yes - combine the masks */
	leftmask &= rightmask;
	}

    /*
     * OK - now those basics are set up.  We need to iterate over the
     * dest longs, sticking the correct bits of the source into it.
     * The problem is that every dest long can be affected by two source
     * longs, and the number of source longs to be considered might be
     * one greater or one smaller than destlongstouched.
     */
    srclongstouched = (minx + srcw - 1) / DGP_BITS_UINT32 + 1;

    /*
     * How many bits RIGHT do you need to shift the bit at minx so that
     * it lands at xpos?  Note that if minx > xpos, you can't do this with
     * a right shift; however, we compute this value anyway, as we'll
     * be shifting from a previous long.  This value is always going to
     * be 0...DGP_BITS_UINT32 - 1.
     */
    rshift = (xpos - minx) & (DGP_BITS_UINT32 - 1);
    /*
     * Now, how many bits LEFT do you need to shift the bit at minx so
     * that it lands at xpos?
     */
    lshift = DGP_BITS_UINT32 - rshift;

    /*
     * There are two cases here:
     * - rshift != 0: we need to do some shifting on everything.
     * - rshift == 0: xpos == minx, and we're perfectly happy with no
     *   shifting.
     */
    if (rshift) {
	/*
	 * The inner loop looks kinda funky because it's really two
	 * cases folded together.
	 * If minx <= xpos:
	 *	Load up the long at *sptr, shift it right (with rshift)
	 *	so that the bit at minx is shifted to xpos, and put the
	 *	result into *dptr (protecting with leftmask).
	 *	Take the remaining bits of *sptr into carry (shifted
	 *	with lshift); bump sptr.
	 *	Then, for each of the remaining destlongstouched - 2
	 *	"normal" dest longs,
	 *	Take *sptr >> rshift, merge with carry, store into *dptr
	 *	Take *sptr << lshift, stash in carry.
	 *	Note that as we go through this loop, we've always touched
	 *	the same number of source longs and dest long.
	 *	Finally, the last dest long: do the same thing, except
	 *	protecting the store using rightmask; if srclongstouched
	 *	is less than destlongstouched, there's nothing to load,
	 *	just the stuff in carry.
	 * If minx > xpos:
	 *	Load up the long at *sptr, shift it left (with lshift),
	 *	and put it into carry; bump sptr.
	 *	Use carry and *sptr to write into *dptr (protected with
	 *	leftmask); if srclongstouched is 1, just use carry.
	 *	Stash *sptr << lshift in carry; bump sptr.
	 *	For each of the remaining destlongstouched - 2 "normal"
	 *	dest longs,
	 *	Take *sptr >> rshift, merge with carry, store into *dptr
	 *	Take *sptr << lshift, stash in carry.
	 *	Note that as we go through this loop, we've always touched
	 *	one more source long than dest long.
	 *	Finally, the last dest long: do the same thing, except
	 *	protecting the store using rightmask; if srclongstouched
	 *	is *equal* to destlongstouched, just use carry.
	 *
	 * The condition that says whether we read a source long at the
	 * time that we process the last dest long is stored in
	 * extraword.
	 */
	if (((minx > xpos) && (srclongstouched > destlongstouched)) ||
	    ((minx <= xpos) && (srclongstouched == destlongstouched))) {
	    extraword = DGP_TRUE;
	    }
	else {
	    extraword = DGP_FALSE;
	    }

	for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl;
	     j < srch;
	     j++, srowptr += srclpl, drowptr += destlpl) {
	    dptr = drowptr;
	    sptr = srowptr;
      
	    /*
	     * First word - load up the long at *sptr, shift right with
	     * rshift, and combine with the dest long, protected with
	     * leftmask.  Set carry for the remaining longs
	     */
	    if (minx > xpos) {
		carry = *sptr++ << lshift;
		if (srclongstouched == 1) {
		    /* That's all, folks */
		    *dptr = (*dptr & ~leftmask) | (carry & leftmask);
		    continue;
		    }
		*dptr = (*dptr & ~leftmask) |
		    (((*sptr >> rshift) | carry) & leftmask);
		carry = *sptr++ << lshift;
		dptr++;
		}
	    else {
		*dptr = (*dptr & ~leftmask) |
		    ((*sptr >> rshift) & leftmask);
		carry = *sptr++ << lshift;
		dptr++;
		}

	    /* Subsequent words, not including last one */
	    for (i = 1; i < destlongstouched - 1; i++) {
		/* Clobber dest */
		*dptr++ = ((*sptr >> rshift) | carry);
		carry = *sptr++ << lshift;
		}
	    if (destlongstouched > 1) {
		if (extraword) {
		    *dptr = (*dptr & ~rightmask) |
			(((*sptr >> rshift) | carry) & rightmask);
		    }
		else {
		    *dptr = (*dptr & ~rightmask) | (carry & rightmask);
		    }
		}
	    }
	}
    else {
	/* No shifting! */
	for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl;
	     j < srch;
	     j++, srowptr += srclpl, drowptr += destlpl) {
	    dptr = drowptr;
	    sptr = srowptr;
      
	    /* Copy the first long, clobbering only the bits in leftmask */
	    *dptr = (*dptr & ~leftmask) | (*sptr++ & leftmask);
	    dptr++;
	    /* Copy all the remaining longs except the last */
	    for (i = 1; i < destlongstouched - 1; i++) {
		*dptr++ = *sptr++;
		}
	    /* Copy the last long, if there's >1, protecting with rightmask */
	    if (destlongstouched > 1) {
		*dptr = (*dptr & ~rightmask) | (*sptr & rightmask);
		}
	    }
	}
    }


/*
 * Same logic as REPLACE, but using AND to combine with the dest.
 */
static void
DgpBlat_blitBitsClipLRAND(Dgp_UInt32 *dest, Dgp_Int32 destlpl,
		  Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 minx,
		  Dgp_Int32 srcw, Dgp_Int32 srch,
		  Dgp_Int32 xpos, Dgp_Int32 ypos, Dgp_State *dgp_state)
{
    Dgp_Int32 i, j;
    Dgp_UInt32 *dptr, *sptr, *drowptr, *srowptr, carry;
    Dgp_Int32 rshift, lshift;
    Dgp_Bool extraword;
    Dgp_UInt32 leftmask, rightmask;
    Dgp_Int32 destlongstouched, srclongstouched;

    /* Unused... / dhtesq. */
    dgp_state = dgp_state;

    src += minx / DGP_BITS_UINT32;
    minx &= DGP_BITS_UINT32 - 1;

    dest += xpos / DGP_BITS_UINT32;
    xpos &= DGP_BITS_UINT32 - 1;

    destlongstouched = (xpos + srcw - 1) / DGP_BITS_UINT32 + 1;

    leftmask = (~0ul) >> xpos;

    if (((xpos + srcw) & (DGP_BITS_UINT32 - 1)) == 0) {
	rightmask = ~0ul;
	}
    else {
	rightmask = (~0ul) <<
	    (DGP_BITS_UINT32 - 1 - ((xpos + srcw - 1) & (DGP_BITS_UINT32 - 1)));
	}

    if (destlongstouched == 1) {
	leftmask &= rightmask;
	}

    srclongstouched = (minx + srcw - 1) / DGP_BITS_UINT32 + 1;

    rshift = (xpos - minx) & (DGP_BITS_UINT32 - 1);
    lshift = DGP_BITS_UINT32 - rshift;

    if (rshift) {
	if (((minx > xpos) && (srclongstouched > destlongstouched)) ||
	    ((minx <= xpos) && (srclongstouched == destlongstouched))) {
	    extraword = DGP_TRUE;
	    }
	else {
	    extraword = DGP_FALSE;
	    }

	for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl;
	     j < srch;
	     j++, srowptr += srclpl, drowptr += destlpl) {
	    dptr = drowptr;
	    sptr = srowptr;
      
	    if (minx > xpos) {
		carry = *sptr++ << lshift;
		if (srclongstouched == 1) {
		    *dptr = (*dptr & ~leftmask) | ((*dptr & carry) & leftmask);
		    continue;
		    }
		*dptr = (*dptr & ~leftmask) |
		    ((*dptr & ((*sptr >> rshift) | carry)) & leftmask);
		carry = *sptr++ << lshift;
		dptr++;
		}
	    else {
		*dptr = (*dptr & ~leftmask) |
		    ((*dptr & (*sptr >> rshift)) & leftmask);
		carry = *sptr++ << lshift;
		dptr++;
		}

	    for (i = 1; i < destlongstouched - 1; i++) {
		*dptr++ &= ((*sptr >> rshift) | carry);
		carry = *sptr++ << lshift;
		}
	    if (destlongstouched > 1) {
		if (extraword) {
		    *dptr = (*dptr & ~rightmask) |
			((*dptr & ((*sptr >> rshift) | carry)) & rightmask);
		    }
		else {
		    *dptr = (*dptr & ~rightmask) |
			((*dptr & carry) & rightmask);
		    }
		}
	    }
	}
    else {
	for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl;
	     j < srch;
	     j++, srowptr += srclpl, drowptr += destlpl) {
	    dptr = drowptr;
	    sptr = srowptr;
      
	    *dptr = (*dptr & ~leftmask) | ((*dptr & *sptr++) & leftmask);
	    dptr++;
	    for (i = 1; i < destlongstouched - 1; i++) {
		*(dptr++) &= *(sptr++);
		}
	    if (destlongstouched > 1) {
		*dptr = (*dptr & ~rightmask) | ((*dptr & *sptr) & rightmask);
		}
	    }
	}
    }

/*
 * Same logic as REPLACE, but using AND to combine the ~src with the dest.
 */
static void
DgpBlat_blitBitsClipLRCOVER(Dgp_UInt32 *dest, Dgp_Int32 destlpl,
		  Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 minx,
		  Dgp_Int32 srcw, Dgp_Int32 srch,
		  Dgp_Int32 xpos, Dgp_Int32 ypos, Dgp_State *dgp_state)
{
    Dgp_Int32 i, j;
    Dgp_UInt32 *dptr, *sptr, *drowptr, *srowptr, carry;
    Dgp_Int32 rshift, lshift;
    Dgp_Bool extraword;
    Dgp_UInt32 leftmask, rightmask;
    Dgp_Int32 destlongstouched, srclongstouched;

    /* Unused... / dhtesq. */
    dgp_state = dgp_state;

    src += minx / DGP_BITS_UINT32;
    minx &= DGP_BITS_UINT32 - 1;

    dest += xpos / DGP_BITS_UINT32;
    xpos &= DGP_BITS_UINT32 - 1;

    destlongstouched = (xpos + srcw - 1) / DGP_BITS_UINT32 + 1;

    leftmask = (~0ul) >> xpos;

    if (((xpos + srcw) & (DGP_BITS_UINT32 - 1)) == 0) {
	rightmask = ~0ul;
	}
    else {
	rightmask = (~0ul) <<
	    (DGP_BITS_UINT32 - 1 - ((xpos + srcw - 1) & (DGP_BITS_UINT32 - 1)));
	}

    if (destlongstouched == 1) {
	leftmask &= rightmask;
	}

    srclongstouched = (minx + srcw - 1) / DGP_BITS_UINT32 + 1;

    rshift = (xpos - minx) & (DGP_BITS_UINT32 - 1);
    lshift = DGP_BITS_UINT32 - rshift;

    if (rshift) {
        if (((minx > xpos) && (srclongstouched > destlongstouched)) ||
            ((minx <= xpos) && (srclongstouched == destlongstouched))) {
            extraword = DGP_TRUE;
        }
        else {
            extraword = DGP_FALSE;
        }

        for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl; j < srch; j++, srowptr += srclpl, drowptr += destlpl) {
            dptr = drowptr;
            sptr = srowptr;
      
            if (minx > xpos) {
                carry = (~*sptr++ << lshift);
                if (srclongstouched == 1) {
                    *dptr = (*dptr & ~leftmask) | ((*dptr & carry) & leftmask);
                    continue;
                }
                *dptr = (*dptr & ~leftmask) | ((*dptr & ((~*sptr >> rshift) | carry)) & leftmask);
                carry = (~*sptr++ << lshift);
                dptr++;
            }
            else {
                *dptr = (*dptr & ~leftmask) | ((*dptr & (~*sptr >> rshift)) & leftmask);
                carry = (~*sptr++ << lshift);
                dptr++;
            }

            for (i = 1; i < destlongstouched - 1; i++) {
                *dptr++ &= ((~*sptr >> rshift) | carry);
                carry = (~*sptr++ << lshift);
            }
            if (destlongstouched > 1) {
                if (extraword) {
                    *dptr = (*dptr & ~rightmask) | ((*dptr & ((~*sptr >> rshift) | carry)) & rightmask);
                } else {
                    *dptr = (*dptr & ~rightmask) | ((*dptr & carry) & rightmask);
                }
            }
        }
    }
    else {
        for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl; j < srch; j++, srowptr += srclpl, drowptr += destlpl) {
            dptr = drowptr;
            sptr = srowptr;
            
            *dptr = (*dptr & ~leftmask) | ((*dptr & ~*sptr++) & leftmask);
            dptr++;
            for (i = 1; i < destlongstouched - 1; i++) {
                *(dptr++) &= ~*(sptr++);
            }
            if (destlongstouched > 1) {
                *dptr = (*dptr & ~rightmask) | ((*dptr & ~*sptr) & rightmask);
            }
        }
    }

    return;
}

/*
 * Same logic as REPLACE, but using XOR to combine with the dest.
 */
static void
DgpBlat_blitBitsClipLRXOR(Dgp_UInt32 *dest, Dgp_Int32 destlpl,
		  Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 minx,
		  Dgp_Int32 srcw, Dgp_Int32 srch,
		  Dgp_Int32 xpos, Dgp_Int32 ypos, Dgp_State *dgp_state)
{
    Dgp_Int32 i, j;
    Dgp_UInt32 *dptr, *sptr, *drowptr, *srowptr, carry;
    Dgp_Int32 rshift, lshift;
    Dgp_Bool extraword;
    Dgp_UInt32 leftmask, rightmask;
    Dgp_Int32 destlongstouched, srclongstouched;

    /* Unused... / dhtesq. */
    dgp_state = dgp_state;

    src += minx / DGP_BITS_UINT32;
    minx &= DGP_BITS_UINT32 - 1;

    dest += xpos / DGP_BITS_UINT32;
    xpos &= DGP_BITS_UINT32 - 1;

    destlongstouched = (xpos + srcw - 1) / DGP_BITS_UINT32 + 1;

    leftmask = (~0ul) >> xpos;

    if (((xpos + srcw) & (DGP_BITS_UINT32 - 1)) == 0) {
	rightmask = ~0ul;
	}
    else {
	rightmask = (~0ul) <<
	    (DGP_BITS_UINT32 - 1 - ((xpos + srcw - 1) & (DGP_BITS_UINT32 - 1)));
	}

    if (destlongstouched == 1) {
	leftmask &= rightmask;
	}

    srclongstouched = (minx + srcw - 1) / DGP_BITS_UINT32 + 1;

    rshift = (xpos - minx) & (DGP_BITS_UINT32 - 1);
    lshift = DGP_BITS_UINT32 - rshift;

    if (rshift) {
	if (((minx > xpos) && (srclongstouched > destlongstouched)) ||
	    ((minx <= xpos) && (srclongstouched == destlongstouched))) {
	    extraword = DGP_TRUE;
	    }
	else {
	    extraword = DGP_FALSE;
	    }

	for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl;
	     j < srch;
	     j++, srowptr += srclpl, drowptr += destlpl) {
	    dptr = drowptr;
	    sptr = srowptr;
      
	    if (minx > xpos) {
		carry = *sptr++ << lshift;
		if (srclongstouched == 1) {
		    *dptr = (*dptr & ~leftmask) | ((*dptr ^ carry) & leftmask);
		    continue;
		    }
		*dptr = (*dptr & ~leftmask) |
		    ((*dptr ^ ((*sptr >> rshift) | carry)) & leftmask);
		carry = *sptr++ << lshift;
		dptr++;
		}
	    else {
		*dptr = (*dptr & ~leftmask) |
		    ((*dptr ^ (*sptr >> rshift)) & leftmask);
		carry = *sptr++ << lshift;
		dptr++;
		}

	    for (i = 1; i < destlongstouched - 1; i++) {
		*dptr++ ^= ((*sptr >> rshift) | carry);
		carry = *sptr++ << lshift;
		}
	    if (destlongstouched > 1) {
		if (extraword) {
		    *dptr = (*dptr & ~rightmask) |
			((*dptr ^ ((*sptr >> rshift) | carry)) & rightmask);
		    }
		else {
		    *dptr = (*dptr & ~rightmask) |
			((*dptr ^ carry) & rightmask);
		    }
		}
	    }
	}
    else {
	for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl;
	     j < srch;
	     j++, srowptr += srclpl, drowptr += destlpl) {
	    dptr = drowptr;
	    sptr = srowptr;
      
	    *dptr = (*dptr & ~leftmask) | ((*dptr ^ *sptr++) & leftmask);
	    dptr++;
	    for (i = 1; i < destlongstouched - 1; i++) {
		*(dptr++) ^= *(sptr++);
		}
	    if (destlongstouched > 1) {
		*dptr = (*dptr & ~rightmask) | ((*dptr ^ *sptr) & rightmask);
		}
	    }
	}
    }

/*
 * Same logic as REPLACE, but using XNOR to combine with the dest.
 */
static void
DgpBlat_blitBitsClipLRXNOR(Dgp_UInt32 *dest, Dgp_Int32 destlpl,
		   Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 minx,
		   Dgp_Int32 srcw, Dgp_Int32 srch,
		   Dgp_Int32 xpos, Dgp_Int32 ypos, Dgp_State *dgp_state)
{
    Dgp_Int32 i, j;
    Dgp_UInt32 *dptr, *sptr, *drowptr, *srowptr, carry;
    Dgp_Int32 rshift, lshift;
    Dgp_Bool extraword;
    Dgp_UInt32 leftmask, rightmask;
    Dgp_Int32 destlongstouched, srclongstouched;

    DGP_PROLOGUE ("DgpBlat_blitBitsClipLRXNOR", DGP_PERRSTATE);
    DGP_STOP_WHINE;
    DGP_VOID_ASSERT(dgp_state);

    src += minx / DGP_BITS_UINT32;
    minx &= DGP_BITS_UINT32 - 1;

    dest += xpos / DGP_BITS_UINT32;
    xpos &= DGP_BITS_UINT32 - 1;

    destlongstouched = (xpos + srcw - 1) / DGP_BITS_UINT32 + 1;

    leftmask = (~0ul) >> xpos;

    if (((xpos + srcw) & (DGP_BITS_UINT32 - 1)) == 0) {
	rightmask = ~0ul;
	}
    else {
	rightmask = (~0ul) <<
	    (DGP_BITS_UINT32 - 1 - ((xpos + srcw - 1) & (DGP_BITS_UINT32 - 1)));
	}

    if (destlongstouched == 1) {
	leftmask &= rightmask;
	}

    srclongstouched = (minx + srcw - 1) / DGP_BITS_UINT32 + 1;

    rshift = (xpos - minx) & (DGP_BITS_UINT32 - 1);
    lshift = DGP_BITS_UINT32 - rshift;

    if (rshift) {
	if (((minx > xpos) && (srclongstouched > destlongstouched)) ||
	    ((minx <= xpos) && (srclongstouched == destlongstouched))) {
	    extraword = DGP_TRUE;
	    }
	else {
	    extraword = DGP_FALSE;
	    }

	for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl;
	     j < srch;
	     j++, srowptr += srclpl, drowptr += destlpl) {
	    dptr = drowptr;
	    sptr = srowptr;
      
	    if (minx > xpos) {
		carry = *sptr++ << lshift;
		if (srclongstouched == 1) {
		    *dptr = (*dptr & ~leftmask) |
			((*dptr ^ ~carry) & leftmask);
		    continue;
		    }
		*dptr = (*dptr & ~leftmask) |
		    ((*dptr ^ ~((*sptr >> rshift) | carry)) & leftmask);
		carry = *sptr++ << lshift;
		dptr++;
		}
	    else {
		*dptr = (*dptr & ~leftmask) |
		    ((*dptr ^ ~(*sptr >> rshift)) & leftmask);
		carry = *sptr++ << lshift;
		dptr++;
		}

	    for (i = 1; i < destlongstouched - 1; i++) {
		*dptr++ ^= ~((*sptr >> rshift) | carry);
		carry = *sptr++ << lshift;
		}
	    if (destlongstouched > 1) {
		if (extraword) {
		    *dptr = (*dptr & ~rightmask) |
			((*dptr ^ ~((*sptr >> rshift) | carry)) & rightmask);
		    }
		else {
		    *dptr = (*dptr & ~rightmask) |
			((*dptr ^ ~carry) & rightmask);
		    }
		}
	    }
	}
    else {
	for (j = 0, srowptr = src, drowptr = dest + ypos * destlpl;
	     j < srch;
	     j++, srowptr += srclpl, drowptr += destlpl) {
	    dptr = drowptr;
	    sptr = srowptr;
      
	    *dptr = (*dptr & ~leftmask) | ((*dptr ^ ~*sptr++) & leftmask);
	    dptr++;
	    for (i = 1; i < destlongstouched - 1; i++) {
		*(dptr++) ^= ~*(sptr++);
		}
	    if (destlongstouched > 1) {
		*dptr = (*dptr & ~rightmask) | ((*dptr ^ ~*sptr) & rightmask);
		}
	    }
	}
    }

void
DgpBlat_blitBits(Dgp_UInt32 *dest, Dgp_Int32 destlpl, Dgp_Int32 destminx, Dgp_Int32 destminy,
	 Dgp_Int32 destmaxx, Dgp_Int32 destmaxy,
	 Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 srcw, Dgp_Int32 srch,
	 Dgp_Int32 xpos, Dgp_Int32 ypos, Dgp_State *dgp_state)
{
    if ((xpos >= destminx) && (ypos >= destminy) &&
	(xpos + srcw <= destmaxx) && (ypos + srch <= destmaxy)) {
	/* Simple - no clipping */
	DgpBlat_blitBitsNoClip(dest, destlpl, src, srclpl, srcw, srch, xpos, ypos, dgp_state);
	}
    else {
	/* Do a complex blit, clipping on all four edges */
	/*
	 * What are the clipped bounds of the symbol (in source coordinates?)
	 * Inclusive on min, exclusive on max
	 */
	Dgp_Int32 minx = MAX(destminx - xpos, 0);
	Dgp_Int32 miny = MAX(destminy - ypos, 0);
	Dgp_Int32 maxx = MIN(srcw, destmaxx - xpos);
	Dgp_Int32 maxy = MIN(srch, destmaxy - ypos);

	if ((minx >= maxx) || (miny >= maxy)) {
	    /* Nothing to do - entirely clipped */
	    return;
	    }

	/*
	 * The clipped bounds of the blit (in source coordinates)
	 * are [minx, maxx) X [miny, maxy).
	 * In our call to blitBitsClipLR, we know that the pixel at (minx, 0)
	 * gets put at (xpos, ypos) where minx, xpos, ypos are the parameters
	 * the blitBitsClipLR, not the current values of those variables.  For
	 * the rest, of the discussion, we're talking about our variables,
	 * not those parameters.
	 *
	 * In X, the solution is easy: we want the pixel at (minx, miny)
	 * to be put at (xpos + minx, ypos + miny) in true dest coordinates.
	 * Therefore, we pass in minx=minx, xpos=xpos+minx.  Of course, the
	 * width of the bit is maxx-minx and the height is maxy-miny.
	 *
	 * In Y, the solution is a bit more complicated because we're going
	 * to do the Y clip right here, by adjusting the pointers.
	 * We add miny to the source pointer, to clip off the top part,
	 * and pass in maxy-miny for the source height.  Now, the pixel
	 * in the top row of the adjusted source (row miny in the actual
	 * source) needs to go into the dest at ypos + miny.
	 */
	DgpBlat_blitBitsClipLR(dest, destlpl,
		       src + srclpl * miny, srclpl,
		       minx, maxx - minx, maxy - miny,
		       xpos + minx,
		       ypos + miny, dgp_state);
	}
    }

void
DgpBlat_blitBitsOp(Dgp_UInt32 *dest, Dgp_Int32 destlpl, Dgp_Int32 destminx, Dgp_Int32 destminy,
	   Dgp_Int32 destmaxx, Dgp_Int32 destmaxy,
	   Dgp_UInt32 *src, Dgp_Int32 srclpl, Dgp_Int32 srcw, Dgp_Int32 srch,
	   Dgp_Int32 xpos, Dgp_Int32 ypos, Dgp_Int32 blitop, Dgp_State *dgp_state)
{
    /*
     * For all but DGPBLAT_OP_OR: same logic as blitBits, but we're not
     * implementing the "simple" case of no clipping because you need
     * left and right masks anyway.
     */

    Dgp_Int32 minx = MAX(destminx - xpos, 0);
    Dgp_Int32 miny = MAX(destminy - ypos, 0);
    Dgp_Int32 maxx = MIN(srcw, destmaxx - xpos);
    Dgp_Int32 maxy = MIN(srch, destmaxy - ypos);

    if ((minx >= maxx) || (miny >= maxy)) {
	/* Nothing to do - entirely clipped */
	return;
	}

    switch(blitop) {
      case DGPBLAT_OP_XOR: {
	/* We can do the Y clip for free by adjusting the pointer. */
	DgpBlat_blitBitsClipLRXOR(dest, destlpl,
			  src + srclpl * miny, srclpl,
			  minx, maxx - minx, maxy - miny,
			  xpos + minx,
			  ypos + miny, dgp_state);
	break;
	}

      case DGPBLAT_OP_AND: {
	DgpBlat_blitBitsClipLRAND(dest, destlpl,
			  src + srclpl * miny, srclpl,
			  minx, maxx - minx, maxy - miny,
			  xpos + minx,
			  ypos + miny, dgp_state);
	break;
	}

      case DGPBLAT_OP_XNOR: {
	DgpBlat_blitBitsClipLRXNOR(dest, destlpl,
			   src + srclpl * miny, srclpl,
			   minx, maxx - minx, maxy - miny,
			   xpos + minx,
			   ypos + miny, dgp_state);
	break;
	}

      case DGPBLAT_OP_REPLACE: {
	DgpBlat_blitBitsClipLRREPLACE(dest, destlpl,
			      src + srclpl * miny, srclpl,
			      minx, maxx - minx, maxy - miny,
			      xpos + minx,
			      ypos + miny, dgp_state);
	break;
	}

      case DGPBLAT_OP_COVER: {
	DgpBlat_blitBitsClipLRCOVER(dest, destlpl,
			      src + srclpl * miny, srclpl,
			      minx, maxx - minx, maxy - miny,
			      xpos + minx,
			      ypos + miny, dgp_state);
	break;
	}

      case DGPBLAT_OP_OR: {
	if ((xpos >= destminx) && (ypos >= destminy) &&
	    (xpos + srcw <= destmaxx) && (ypos + srch <= destmaxy)) {
	    /* Simple - no clipping */
	    DgpBlat_blitBitsNoClip(dest, destlpl, src, srclpl, srcw, srch,
			   xpos, ypos, dgp_state);
	    }
	else {
	    /* Do a complex blit, clipping on all four edges */
	    /* The "all clipped" check has been done */

	    DgpBlat_blitBitsClipLR(dest, destlpl,
			   src + srclpl * miny, srclpl,
			   minx, maxx - minx, maxy - miny,
			   xpos + minx,
			   ypos + miny, dgp_state);
	    }
	break;
	}
      }
    }

/*
 * Basically, just strip off the structure and call blitBits.  You have
 * to be a little careful as src and dest might have non-zero offsets.
 */
void
DgpBlat_blitPacked(DgpImage_Packed dest, DgpImage_Packed src, Dgp_Int32 xpos, Dgp_Int32 ypos)
{
    /*
     * We're going to pass in the true memory areas of src and dest
     * down to blitBits.  Call these srcbits and destbits.
     *
     * The (0,0) pixel of src is actually in srcbits row -DGPIMG_YBASE(src).
     * The (xpos,ypos) pixel of dest is in destbits row -DGPIMG_YBASE(dest).
     * Thus, row src(0) is really row srcbits(-ybase(src)), and that
     * needs to be drawn onto row destbits(ypos-ybase(dest)).  Row
     * srcbits(0) therefore needs to be drawn onto row
     * destbits(ypos-ybase(dest)+ybase(src)).
     * For example, if ybase(src) is -2, ybase(dest) is +1, and ypos is 9,
     * then row 0 of src (i.e., row +2 of srcbits) should go onto row 9
     * of dest (row 8 of destbits); row 0 of srcbits should go onto row
     * 6 of destbits.
     *
     * xpos isn't taken into account in any of the pointers, so there's no
     * pointer juggling.  Pixel (0,0) of src is in bit number -xbase(src)
     * of row src(0); pixel (xpos,ypos) of dest is in bit number
     * xpos - xbase(dest) of row dest(ypos).  Thus, pixel (xbase,0) of src
     * (the first true pixel) should be drawn onto pixel
     * (xpos+xbase(src)-xbase(dest),ypos) of dest.
     * For example, if xbase(src) is -2, xbase(dest) is +1, and xpos is 9,
     * pixel (0,0) of src (bit number 2 in row 0) should go to bit number 8
     * of row ypos; bit number 0 of row 0 (the first true bit) should go to
     * bit number 6 of row ypos.
     */
    DgpBlat_blitBits(DGPIMG_GETROW(dest, DGPIMG_YBASE(dest)),
	     DGPIMG_GETROWWIDTH(dest) / (int)sizeof(Dgp_UInt32), 0, 0,
	     DGPIMG_WIDTH(dest), DGPIMG_HEIGHT(dest),
	     DGPIMG_GETROW(src, DGPIMG_YBASE(src)),
	     DGPIMG_GETROWWIDTH(src) / (int)sizeof(Dgp_UInt32),
	     DGPIMG_WIDTH(src), DGPIMG_HEIGHT(src),
	     xpos + DGPIMG_XBASE(src) - DGPIMG_XBASE(dest),
	     ypos + DGPIMG_YBASE(src) - DGPIMG_YBASE(dest), src->dgp_state);
    }

void
DgpBlat_blitPackedOp(DgpImage_Packed dest, DgpImage_Packed src, Dgp_Int32 xpos, Dgp_Int32 ypos, 
					Dgp_Int32 blitop)
{
    Dgp_Int32 minx = MAX(-xpos, 0);
    Dgp_Int32 miny = MAX(-ypos, 0);
    Dgp_Int32 maxx = MIN(DGPIMG_WIDTH(src), DGPIMG_WIDTH(dest) - xpos);
    Dgp_Int32 maxy = MIN(DGPIMG_HEIGHT(src), DGPIMG_HEIGHT(dest) - ypos);
    Dgp_State *dgp_state = src->dgp_state;

    DGP_PROLOGUE ("DgpBlat_blitPackedOp", DGP_PERRSTATE);
    DGP_STOP_WHINE;
    DGP_VOID_ASSERT(dgp_state);

    if ((minx >= maxx) || (miny >= maxy)) {
	/* Nothing to do - entirely clipped */
	return;
	}

    switch(blitop) {
      case DGPBLAT_OP_XOR: {
	/* We can do the Y clip for free by adjusting the pointer. */
	DgpBlat_blitBitsClipLRXOR(DGPIMG_GETROW(dest, 0),
			  DGPIMG_GETROWWIDTH(dest) / sizeof(Dgp_UInt32),
			  DGPIMG_GETROW(src, miny),
			  DGPIMG_GETROWWIDTH(src) / sizeof(Dgp_UInt32),
			  minx, maxx - minx, maxy - miny,
			  xpos + minx,
			  ypos + miny, dgp_state);
	break;
	}

      case DGPBLAT_OP_AND: {
	/* We can do the Y clip for free by adjusting the pointer. */
	DgpBlat_blitBitsClipLRAND(DGPIMG_GETROW(dest, 0),
			  DGPIMG_GETROWWIDTH(dest) / sizeof(Dgp_UInt32),
			  DGPIMG_GETROW(src, miny),
			  DGPIMG_GETROWWIDTH(src) / sizeof(Dgp_UInt32),
			  minx, maxx - minx, maxy - miny,
			  xpos + minx,
			  ypos + miny, dgp_state);
	break;
	}

      case DGPBLAT_OP_XNOR: {
	/* We can do the Y clip for free by adjusting the pointer. */
	DgpBlat_blitBitsClipLRXNOR(DGPIMG_GETROW(dest, 0),
			   DGPIMG_GETROWWIDTH(dest) / sizeof(Dgp_UInt32),
			   DGPIMG_GETROW(src, miny),
			   DGPIMG_GETROWWIDTH(src) / sizeof(Dgp_UInt32),
			   minx, maxx - minx, maxy - miny,
			   xpos + minx,
			   ypos + miny, dgp_state);
	break;
	}

      case DGPBLAT_OP_REPLACE: {
	/* We can do the Y clip for free by adjusting the pointer. */
	DgpBlat_blitBitsClipLRREPLACE(DGPIMG_GETROW(dest, 0),
			      DGPIMG_GETROWWIDTH(dest) / sizeof(Dgp_UInt32),
			      DGPIMG_GETROW(src, miny),
			      DGPIMG_GETROWWIDTH(src) / sizeof(Dgp_UInt32),
			      minx, maxx - minx, maxy - miny,
			      xpos + minx,
			      ypos + miny, dgp_state);
	break;
	}

      case DGPBLAT_OP_OR: {
	DgpBlat_blitPacked(dest, src, xpos, ypos);
	break;
	}

      default: {
	DGP_PANIC("bad blitop");
	break;
	}
      }
    }

void
DgpBlat_blitFillPacked(DgpImage_Packed dest)
{
    Dgp_UInt32 *p;
    Dgp_Int32 nlongs, longsperline;
    Dgp_Bool lastempty;
    Dgp_UInt32 lastlong;
    Dgp_Int32 x, y;
    
    /* Cache the row stride */
    longsperline = DGPIMG_GETROWWIDTH(dest) / sizeof(Dgp_UInt32);
    
    /* How many full longs per row? */
    nlongs = DGPIMG_WIDTH(dest) / DGP_BITS_UINT32;

    /* and are there any spillover bits past the last one? */
    x = DGPIMG_WIDTH(dest) - nlongs * DGP_BITS_UINT32;
    if (x > 0) {
	/* Set the leftmost x bits of lastlong */
	lastlong = (~0ul) << (DGP_BITS_UINT32 - x);
	lastempty = DGP_FALSE;
	}
    else {
	lastlong = 0;
	lastempty = DGP_TRUE;
	}

    /* Do the fill */
    for (y = 0, p = DGPIMG_GETROW(dest, DGPIMG_YBASE(dest));
	 y < DGPIMG_HEIGHT(dest); y++, p += longsperline) {
	for (x = 0; x < nlongs; x++) {
	    p[x] = ~0ul;
	    }

	if (!lastempty) {
	    p[x] = lastlong;
	    }
	}
    }

#if 0

extern long random(void);
#include <stdlib.h>

const unsigned char blitops[] = {
    DGPBLAT_OP_OR,
    DGPBLAT_OP_AND,
    DGPBLAT_OP_XOR,
    DGPBLAT_OP_XNOR,
    DGPBLAT_OP_REPLACE
    };

void
fillim(DgpImage_Packed im)
{
    Dgp_Int32 w, h;
    Dgp_Int32 x, y;
    long randval = 0;
    Dgp_Int32 nb = 0;
    
    w = DGPIMG_WIDTH(im);
    h = DGPIMG_HEIGHT(im);

    for (y = 0; y < h; y++) {
	for (x = 0; x < w; x++) {
	    if (nb == 0) {
		randval = random();
		nb = 31;
		}
	    if (randval & 1) {
		DGPIMG_SETBIT(im, x, y);
		}
	    randval >>= 1;
	    nb--;
	    }
	}
    }

main(Dgp_Int32 argc, char **argv)
{
    Dgp_Int32 n = (argc > 1) ? atoi(argv[1]) : 10000;
    Dgp_Int32 minx, maxx, miny, maxy;
    Dgp_Int32 minxt, maxxt, minyt, maxyt;
    Dgp_UInt32 *p, *q;
    Dgp_Int32 i;
    DgpImage_Packed src, dest, dest2;
    Dgp_Int32 sw, sh, dw, dh;
    Dgp_Int32 xpos, ypos;
    Dgp_Int32 blitop;
    Dgp_Int32 x, y, sb, db, rb, shouldab;

    for (i = 0; i < n; i++) {
	/* Pick widths and heights - uniform in 1-129 */
	sw = random() % 129 + 1;
	sh = random() % 129 + 1;
	dw = random() % 129 + 1;
	dh = random() % 129 + 1;
	
	/* Pick a window in the destination - uniform in 0-(dw,dh) */
	minx = random() % dw;
	maxx = random() % (dw - minx) + minx + 1;
	miny = random() % dh;
	maxy = random() % (dh - miny) + miny + 1;
	
	/* Pick an operation */
	blitop = blitops[random() % 5];

	/* Pick a destination offset - uniform in 0-(dw,dh) */
	xpos = random() % dw;
	ypos = random() % dh;

	/* Allocate and fill the src and dest */
	src = (DgpImage_Packed)imNew(IMAGE_PACKED, sw, sh);
	dest = (DgpImage_Packed)imNew(IMAGE_PACKED, dw, dh);

	fillim(src);
	fillim(dest);

	/* Make a clone of the dest so we can verify */
	dest2 = (DgpImage_Packed)imDup(dest);

	/* Do the blit */
	blitBitsOp(DGPIMG_GETROW(dest, 0),
		   DGPIMG_GETROWWIDTH(dest) / sizeof(Dgp_UInt32),
		   minx, miny, maxx, maxy,
		   DGPIMG_GETROW(src, 0),
		   DGPIMG_GETROWWIDTH(src) / sizeof(Dgp_UInt32),
		   sw, sh, xpos, ypos, blitop);

	/* Now check it */
#if 1
	/* What parts should have been touched? */
	minxt = MAX(minx, xpos);
	maxxt = MIN(maxx, xpos + sw);
	minyt = MAX(miny, ypos);
	maxyt = MIN(maxy, ypos + sh);

	if ((minxt >= maxxt) || (minyt >= maxyt)) {
	    /* Nothing should have been touched */
	    for (y = 0; y < dh; y++) {
		p = DGPIMG_GETROW(dest, y);
		q = DGPIMG_GETROW(dest2, y);
		for (x = 0; x < DGPIMG_GETROWWIDTH(dest) / (int)sizeof(Dgp_UInt32); x++) {
		    if (*p++ != *q++) {
			printf("Problem: corruption around (%d, %d) in blit %d, params\n",
			       x * DGP_BITS_UINT32, y, i);
			printf(" dw = %d  dh = %d  sw = %d  sh = %d\n",
			       dw, dh, sw, sh);
			printf(" minx = %d  miny = %d  maxx = %d  maxy = %d\n",
			       minx, miny, maxx, maxy);
			printf(" xpos = %d  ypos = %d  blitop = %x\n",
			       xpos, ypos, blitop);

			imSave(src, "/tmp/src.pbm");
			imSave(dest, "/tmp/result.pbm");
			imSave(dest2, "/tmp/dest.pbm");
			exit(1);
			}
		    }
		}
	    }
	else {
	    /* Rows 0...minyt - 1 shouldn't have been touched */
	    for (y = 0; y < minyt; y++) {
		p = DGPIMG_GETROW(dest, y);
		q = DGPIMG_GETROW(dest2, y);
		for (x = 0; x < DGPIMG_GETROWWIDTH(dest) / (int)sizeof(Dgp_UInt32); x++) {
		    if (*p++ != *q++) {
			printf("Problem: corruption around (%d, %d) in blit %d, params\n",
			       x * DGP_BITS_UINT32, y, i);
			printf(" dw = %d  dh = %d  sw = %d  sh = %d\n",
			       dw, dh, sw, sh);
			printf(" minx = %d  miny = %d  maxx = %d  maxy = %d\n",
			       minx, miny, maxx, maxy);
			printf(" xpos = %d  ypos = %d  blitop = %x\n",
			       xpos, ypos, blitop);

			imSave(src, "/tmp/src.pbm");
			imSave(dest, "/tmp/result.pbm");
			imSave(dest2, "/tmp/dest.pbm");
			exit(1);
			}
		    }
		}

	    /* Rows minyt ... maxyt - 1 should be checked */
	    for (; y < maxyt; y++) {
		/* First minxt columns should be untouched */
		for (x = 0; x < minxt; x++) {
		    db = DGPIMG_GETBIT01(dest2, x, y);
		    rb = DGPIMG_GETBIT01(dest, x, y);
		    if (db != rb) {
			printf("Problem: corruption at (%d, %d) in blit %d, params\n",
			       x, y, i);
			printf(" dw = %d  dh = %d  sw = %d  sh = %d\n",
			       dw, dh, sw, sh);
			printf(" minx = %d  miny = %d  maxx = %d  maxy = %d\n",
			       minx, miny, maxx, maxy);
			printf(" xpos = %d  ypos = %d  blitop = %x\n",
			       xpos, ypos, blitop);

			imSave(src, "/tmp/src.pbm");
			imSave(dest, "/tmp/result.pbm");
			imSave(dest2, "/tmp/dest.pbm");
			exit(1);
			}
		    }

		/* Check columns minxt ... maxxt - 1 */
		for (; x < maxxt; x++) {
		    db = DGPIMG_GETBIT01(dest2, x, y);
		    rb = DGPIMG_GETBIT01(dest, x, y);
		    sb = DGPIMG_GETBIT01(src, x - xpos, y - ypos);
		    shouldab = ((blitop >> (sb * 2)) >> (db)) & 1;

		    if (rb != shouldab) {
			/* A problem. */
			printf("Problem: mismatch at (%d, %d) in blit %d, params\n",
			       x, y, i);
			printf(" dw = %d  dh = %d  sw = %d  sh = %d\n",
			       dw, dh, sw, sh);
			printf(" minx = %d  miny = %d  maxx = %d  maxy = %d\n",
			       minx, miny, maxx, maxy);
			printf(" xpos = %d  ypos = %d  blitop = %x\n",
			       xpos, ypos, blitop);

			imSave(src, "/tmp/src.pbm");
			imSave(dest, "/tmp/result.pbm");
			imSave(dest2, "/tmp/dest.pbm");
			exit(1);
			}
		    }

		/* Columns maxxt up should be untouched */
		for (; x < dw; x++) {
		    db = DGPIMG_GETBIT01(dest2, x, y);
		    rb = DGPIMG_GETBIT01(dest, x, y);
		    if (db != rb) {
			printf("Problem: corruption at (%d, %d) in blit %d, params\n",
			       x, y, i);
			printf(" dw = %d  dh = %d  sw = %d  sh = %d\n",
			       dw, dh, sw, sh);
			printf(" minx = %d  miny = %d  maxx = %d  maxy = %d\n",
			       minx, miny, maxx, maxy);
			printf(" xpos = %d  ypos = %d  blitop = %x\n",
			       xpos, ypos, blitop);

			imSave(src, "/tmp/src.pbm");
			imSave(dest, "/tmp/result.pbm");
			imSave(dest2, "/tmp/dest.pbm");
			exit(1);
			}
		    }
		}

	    /* Same for rows maxyt on up */
	    for (; y < dh; y++) {
		p = DGPIMG_GETROW(dest, y);
		q = DGPIMG_GETROW(dest2, y);
		for (x = 0; x < DGPIMG_GETROWWIDTH(dest) / (int)sizeof(Dgp_UInt32); x++) {
		    if (*p++ != *q++) {
			printf("Problem: corruption around (%d, %d) in blit %d, params\n",
			       x * DGP_BITS_UINT32, y, i);
			printf(" dw = %d  dh = %d  sw = %d  sh = %d\n",
			       dw, dh, sw, sh);
			printf(" minx = %d  miny = %d  maxx = %d  maxy = %d\n",
			       minx, miny, maxx, maxy);
			printf(" xpos = %d  ypos = %d  blitop = %x\n",
			       xpos, ypos, blitop);

			imSave(src, "/tmp/src.pbm");
			imSave(dest, "/tmp/result.pbm");
			imSave(dest2, "/tmp/dest.pbm");
			exit(1);
			}
		    }
		}
	    }

#else
	for (y = 0; y < dh; y++) {
	    for (x = 0; x < dw; x++) {
		db = DGPIMG_GETBIT01(dest2, x, y);
		rb = DGPIMG_GETBIT01(dest, x, y);

		if ((x >= minx) && (y >= miny) &&
		    (x < maxx) && (y < maxy) &&
		    (x - xpos >= 0) && (x - xpos < sw) &&
		    (y - ypos >= 0) && (y - ypos < sh)) {
		    /* In the area that should have been affected */
		    sb = DGPIMG_GETBIT01(src, x - xpos, y - ypos);
		    shouldab = ((blitop >> (sb * 2)) >> (db)) & 1;

		    if (rb != shouldab) {
			/* A problem. */
			printf("Problem: mismatch at (%d, %d) in blit %d, params\n",
			       x, y, i);
			printf(" dw = %d  dh = %d  sw = %d  sh = %d\n",
			       dw, dh, sw, sh);
			printf(" minx = %d  miny = %d  maxx = %d  maxy = %d\n",
			       minx, miny, maxx, maxy);
			printf(" xpos = %d  ypos = %d  blitop = %x\n",
			       xpos, ypos, blitop);

			imSave(src, "/tmp/src.pbm");
			imSave(dest, "/tmp/result.pbm");
			imSave(dest2, "/tmp/dest.pbm");
			exit(1);
			}
		    }
		else {
		    /* Shouldn't have been touched */
		    if (db != rb) {
			printf("Problem: corruption at (%d, %d) in blit %d, params\n",
			       x, y, i);
			printf(" dw = %d  dh = %d  sw = %d  sh = %d\n",
			       dw, dh, sw, sh);
			printf(" minx = %d  miny = %d  maxx = %d  maxy = %d\n",
			       minx, miny, maxx, maxy);
			printf(" xpos = %d  ypos = %d  blitop = %x\n",
			       xpos, ypos, blitop);

			imSave(src, "/tmp/src.pbm");
			imSave(dest, "/tmp/result.pbm");
			imSave(dest2, "/tmp/dest.pbm");
			exit(1);
			}
		    }
		}
	    }
#endif

	if (i % 1000 == 0) {
	    printf("%d\n", i);
	    imSave(src, "/tmp/src.pbm");
	    imSave(dest, "/tmp/result.pbm");
	    imSave(dest2, "/tmp/dest.pbm");
	    }

	imFree(src);
	imFree(dest);
	imFree(dest2);
	}

    return(0);
    }

#endif
