#include "blit_byte.h"
#include "blitTables.h"
#include <string.h>
/* blit_byte.c
   wbell 10.20.99
   rgb interleaved blitting into rgb interleaved destination

  25.8.2000 / dhtesq.
  ...or into grayscale images / color separations...
 */

/******************************************************************/
static INLINE void 
blitBinaryGrayRow(Dgp_UInt8     *dstCur, 
                  Dgp_UInt32    *srcCur,
                  Dgp_Int32     curSrcBitsLeft, 
                  Dgp_Int32     dstW,
                  Dgp_UInt8     value,
                  const Dgp_UInt32 valueWord[2]);


static INLINE void 
blitBinaryGrayRow(Dgp_UInt8     *dstCur, 
                  Dgp_UInt32    *srcCur,
                  Dgp_Int32     curSrcBitsLeft, 
                  Dgp_Int32     dstW,
                  Dgp_UInt8     value,
                  const Dgp_UInt32 valueWord[2])
{
  Dgp_Int32 x = 0, temp;
  Dgp_Int32 wb;
  Dgp_UInt32 curSrcWord;
  Dgp_Int32 srcBitsUsed;
  Dgp_UInt32 *dstLong;
  const Dgp_UInt32 *dstLookup;

  /* Do pixel by pixel until dstCur is on a word boundary. */
  wb = (Dgp_Int32)dstCur % sizeof(Dgp_UInt32);
  /* Make sure this is less than the blatting width. */
  wb = MIN(wb, dstW);

  /* We can need to do either 0,1,2,3 pixels. We just special */
  /* case them all. */
  /* Clear out all the bits we're not interested in. */
  curSrcWord = *(srcCur++) << (DGP_BITS_UINT32 - curSrcBitsLeft);
  srcBitsUsed = 0;

  /* Make sure we actually have enough bits in curSrcWord to */
  /* to do these few writes. */
  if (curSrcBitsLeft < wb){
      /* This assumes srcBitsUsed = 0... */
      curSrcWord = curSrcWord | (*srcCur >> curSrcBitsLeft);
      srcBitsUsed = DGP_BITS_UINT32 - curSrcBitsLeft;
      curSrcBitsLeft = DGP_BITS_UINT32;
  }

  if(wb == 1){
      if(curSrcWord >> 31){
          *dstCur = value;
      }
      dstCur++;
      curSrcBitsLeft--;
      curSrcWord = curSrcWord << 1;
      x++;
  }
  else if(wb == 2){
      if(curSrcWord >> 31){
          *dstCur = value;
      }
      dstCur++;
      curSrcWord = curSrcWord << 1;

      if(curSrcWord >> 31){
          *dstCur = value;
      }
      dstCur++;
      curSrcWord = curSrcWord << 1;

      curSrcBitsLeft -= 2;
      x += 2;
  }
  else if(wb == 3){
      if(curSrcWord >> 31){
          *dstCur = value;
      }
      dstCur++;
      curSrcWord = curSrcWord << 1;

      if(curSrcWord >> 31){
          *dstCur = value;
      }
      dstCur++;
      curSrcWord = curSrcWord << 1;

      if(curSrcWord >> 31){
          *dstCur = value;
      }
      dstCur++;
      curSrcWord = curSrcWord << 1;

      curSrcBitsLeft -= 3;
      x += 3;
  }

  /* Now we should be word aligned in the destination, or we should be done. */

  dstLong = (Dgp_UInt32 *)dstCur;
  /* Now we're going to do this 32 pixels at a time in blocks of 4. */
  while((x + DGP_BITS_UINT32) <= dstW){
      /* Fill the word up with 32 pixels, we'll grab what we can from srcCur, */
      /* and if we need more, we'll move srcCur forward. */
      if(curSrcBitsLeft != DGP_BITS_UINT32){
          if((curSrcBitsLeft + (DGP_BITS_UINT32 - srcBitsUsed)) < DGP_BITS_UINT32){
              /* ...spanning a word. */
              /* We'll take as much as is left of this */
              /* one, and get the rest from */
              /* the next. */
              if(srcBitsUsed < DGP_BITS_UINT32){
                  curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
                  curSrcBitsLeft += (DGP_BITS_UINT32 - srcBitsUsed);
              }
              
              srcCur++;
              srcBitsUsed = 0;
          }
          
          /* ...not spanning a word. */
          curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
          srcBitsUsed += (DGP_BITS_UINT32 - curSrcBitsLeft);
          curSrcBitsLeft = DGP_BITS_UINT32;
      }

      /* We've got enough, now let's do the 4 8-bit blats. */
      dstLookup = monoToGray8bit[(Dgp_UInt8)(curSrcWord >> 24)];
      dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & valueWord[0]);
      dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & valueWord[1]);
      dstLong += 2;

      dstLookup = monoToGray8bit[(Dgp_UInt8)(curSrcWord >> 16)];
      dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & valueWord[0]);
      dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & valueWord[1]);
      dstLong += 2;

      dstLookup = monoToGray8bit[(Dgp_UInt8)(curSrcWord >> 8)];
      dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & valueWord[0]);
      dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & valueWord[1]);
      dstLong += 2;

      dstLookup = monoToGray8bit[(Dgp_UInt8)curSrcWord];
      dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & valueWord[0]);
      dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & valueWord[1]);
      dstLong += 2;

      curSrcBitsLeft = 0;
      curSrcWord = 0;

      x += DGP_BITS_UINT32;
  }

  while((x + 8) <= dstW){
      /* Fill the word up with 8 pixels, we'll grab what we can from srcCur, */
      /* and if we need more, we'll move srcCur forward. */
      if(curSrcBitsLeft < 8){
          if((8 - curSrcBitsLeft) > (DGP_BITS_UINT32 - srcBitsUsed)){
              /* ...spanning a word. */
              /* We'll take as much as is left of this one, and get the rest from */
              /* the next. */
              if (srcBitsUsed < DGP_BITS_UINT32){
                  curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
                  curSrcBitsLeft += (DGP_BITS_UINT32 - srcBitsUsed);
              }
              
              srcCur++;
              srcBitsUsed = 0;
          }
          /* ....not spanning a word, just take as much as we can. */
          curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
          temp = MIN(DGP_BITS_UINT32 - srcBitsUsed, DGP_BITS_UINT32 - curSrcBitsLeft);
          srcBitsUsed += temp;
          curSrcBitsLeft += temp;
      }

      /* We've got enough, now let's do the 8-bit blat. */
      dstLookup = monoToGray8bit[(Dgp_UInt8)(curSrcWord >> 24)];
      dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & valueWord[0]);
      dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & valueWord[1]);
      dstLong += 2;

      curSrcWord = curSrcWord << 8;
      curSrcBitsLeft -= 8;
      x += 8;
  }

  dstCur = (Dgp_UInt8 *)dstLong;
  /* So now we could have at most 7 more pixels to do. */
  wb = (dstW - x);
  /* Make sure we have enough bits in curSrcWord to do it. */
  if(curSrcBitsLeft < wb){
      if((wb - curSrcBitsLeft) > (DGP_BITS_UINT32 - srcBitsUsed)){
          /* ...spanning a word. */
          /* We'll take as much as is left of this one, and get the rest from */
          /* the next. */
          if(srcBitsUsed < DGP_BITS_UINT32){
              curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
              curSrcBitsLeft += (DGP_BITS_UINT32 - srcBitsUsed);
          }
          
          srcCur++;
          srcBitsUsed = 0;
      }

      /* ...not spanning a word, just take as much as we can. */
      curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
      temp = MIN(DGP_BITS_UINT32 - srcBitsUsed, 32 - curSrcBitsLeft);
      srcBitsUsed += temp;
      curSrcBitsLeft += temp;
  }

  /* Now just special case them. */
  while(x < dstW){
      if((curSrcWord >> 31)){
          *dstCur = value;
      }
      dstCur++;
      curSrcWord = curSrcWord << 1;

      x++;
  }
  return;
}


/******************************************************************/
/* blitSetByte()
   set a portion of a byte image to a particular value.
   will respect outside image data and will not modify
   outside the given rectangle.
   
   val - the value you want it to be set to.
   dst - pointer to the destination at dstMinX, dstMinY.
   dstBpl - bytes per line in the dest.
   dstX, dstY - where you want set, the upper left hand corner.
   dstW, dstH - how much you want cleared or set, must be positive.
*/
void DgpBlat_blitSetByte(Dgp_UInt8 val, 
		Dgp_UInt8 * dst, Dgp_Int32 dstBpl,
		Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		Dgp_Int32 dstX, Dgp_Int32 dstY,
		Dgp_Int32 dstW, Dgp_Int32 dstH)
{
  Dgp_UInt8 * dstRow;
  Dgp_UInt8 * dstCur;
  Dgp_UInt8 * dstEnd;

  if (dstW <= 0 || dstH <= 0 || !dstBpl){
    return;
  }

  dstRow = dst + ((dstY - dstMinY) * dstBpl)
    + (dstX - dstMinX);
  dstEnd = dstRow + (dstH * dstBpl);
  
  for (dstCur = dstRow ; dstCur < dstEnd ; dstCur = dstCur + dstBpl){
    memset(dstCur, val, dstW);
  }
}
/******************************************************************/

/******************************************************************/
/*
25.8.2000 / dhtesq.
Sets a portion of an 8-bit grayscale image to the (single) passed value
at those locations where the passed source bitmap is true.

Robbed this from DgpBlat_colorBlitBinaryRGB() to blit tokens onto
cmyk separations.
*/
#ifdef _WIN32
DGP_LIBEXPORT 
#endif
INLINE void 
DgpBlat_blitBinaryGray( Dgp_UInt32       *src,
                        Dgp_Int32        srcBpl, 
                        Dgp_Int32        srcMinX, 
                        Dgp_Int32        srcMinY,
                        Dgp_Int8         *dst, 
                        Dgp_Int32        dstBpl, 
                        Dgp_Int32        dstMinX,
                        Dgp_Int32        dstMinY,
                        Dgp_Int32        dstX,
                        Dgp_Int32        dstY,
                        Dgp_Int32        dstW, 
                        Dgp_Int32        dstH, 
                        Dgp_UInt8        value)
{
  /* The general idea here is that we want to use an 8-bit lookup
   * table to convert 8-bits of packed image data to 8 bytes of
   * gray data, so we need to
   * for each row, find the first dest (gray) pixel that is word
   * aligned, and up to that point, the conversion one bit at a
   * time, and then afterwards do the conversion 8-bits at a time.
   */

  Dgp_UInt32 *srcRow, *srcCur;
  Dgp_UInt8 *dstRow, *dstCur;

  Dgp_Int32 dstOff, srcOff, srcLpl, curSrcBitsLeft;
  Dgp_Int32 y;

  Dgp_UInt32 valueWord[2];

  srcLpl = srcBpl / sizeof(Dgp_UInt32);

  dstOff = ((dstY - dstMinY) * dstBpl) + ((dstX - dstMinX) * sizeof(Dgp_UInt8));
  dstRow = (Dgp_UInt8 *)((Dgp_Int8 *)dst + dstOff);

  valueWord[0] = value | (value << 8) | (value << 16) | (value << 24);
  valueWord[1] = valueWord[0];

  /* This gets us the beginning of the row of the source, but */
  /* now we need to find the first word we're concerned with, and */
  /* which pixel (bit) in that word is the first to be converted. */
  srcOff = (srcLpl * (dstY - srcMinY));
  srcOff += ((dstX - srcMinX) / 32);
  srcRow = src + srcOff;

  for(y = 0; y < dstH; y ++){
      dstCur = dstRow;
      srcCur = srcRow;
      curSrcBitsLeft = DGP_BITS_UINT32 - (dstX - srcMinX) % DGP_BITS_UINT32;

      blitBinaryGrayRow(dstCur, srcCur, curSrcBitsLeft, dstW, value, valueWord);

      dstRow = (Dgp_UInt8 *)((Dgp_Int8 *)dstRow + dstBpl);
      srcRow = srcRow + srcLpl;
    }
}
/******************************************************************/

