#ifndef BLIT_MONO_H
#define BLIT_MONO_H

#ifdef __cplusplus
extern "C"
{
#endif

#include "dgp_types.h"
#include "util.h"

/* blit_mono.h
   wbell 10.19.99
   packed mono blitting into packed mono destination
 */



/* blitSetBinary()
   set a portion of a binary image to a particular value (0) or (1).
   will respect outside image data (like partial words) and will not modify
   outside the given rectangle.
   
   val - the value you want it to be set to, should be either 0 or 1.
   dst - pointer to the destination at dstMinX, dstMinY.
   dstBpl - bytes per line in the dest.
   dstX, dstY - where you want set, the upper left hand corner.
   dstW, dstH - how much you want cleared or set, must be positive
*/
extern void 
DgpBlat_blitSetBinary(Dgp_UInt8 val, 
		   Dgp_UInt32 * dst, Dgp_Int32 dstBpl,
		   Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		   Dgp_Int32 dstX, Dgp_Int32 dstY,
		   Dgp_Int32 dstW, Dgp_Int32 dstH);


/* blitBinaryOrExact()
   blits a binary image into a binary destination image. Source image must
   have no extra bytes on the end of the row (rows must only be padded to 
   32-bits) and padding must be set to 0. 
   Code stolen from Russ Atkinson's OrBlt
   dst - pointer to first word of destination.
   off - what bit of dst to put down first bit of src (31 is leftmost bit)
   dstBpl - bytes per line in the destination
   src - pointer to first word of source
   srcH - how many lines is the source token?
*/
#ifdef _WIN32
DGP_LIBEXPORT
#endif 
INLINE void 
DgpBlat_blitBinaryOrExact(Dgp_UInt32 * dst,
		       Dgp_UInt32 off,
		       Dgp_Int32 dstBpl,
		       Dgp_UInt32 *src,
		       Dgp_UInt32 srcW,
		       Dgp_Int32 srcH);


#ifdef __cplusplus
}
#endif
#endif
