#include "blit_rgb.h"
#include <string.h>
/* blit_rgb.c
   wbell 10.20.99
   rgb interleaved blitting into rgb interleaved destination
 */

/* blitSetDgpImage_RGB()
   set a portion of a rgb image to a particular value.
   will respect outside image data and will not modify
   outside the given rectangle.
   
   val - the value you want it to be set to.
   dst - pointer to the destination at dstMinX, dstMinY.
   dstBpl - bytes per line in the dest.
   dstX, dstY - where you want set, the upper left hand corner.
   dstW, dstH - how much you want cleared or set, must be positive
*/

void DgpBlat_blitSetRGB(DgpImage_RGB val, 
		DgpImage_RGB * dst, Dgp_Int32  dstBpl,
		Dgp_Int32  dstMinX, Dgp_Int32  dstMinY,
		Dgp_Int32  dstX, Dgp_Int32  dstY,
		Dgp_Int32  dstW, Dgp_Int32  dstH)
{
  DgpImage_RGB * dstRow;
  DgpImage_RGB * dstCur;
  DgpImage_RGB * dstEnd;
  Dgp_Int32 rsize;
  Dgp_Int32 filler_row_length = 0;
  Dgp_Int32 num_rows = 0;

  if (dstW <= 0 || dstH <= 0 || !dstBpl){
    return;
  }

  /*
  20.6.2000 / dhtesq.  Fixed bug.
  For some orientations, dstMinX > dstX or dstMinY > dstY, resulting in _subtraction_ of rows.
  Fixed...

  */
  filler_row_length = (dstX >= dstMinX ? (dstX - dstMinX) : (dstX));
  num_rows = (dstY >= dstMinY ? (dstY - dstMinY) : (dstY));
  dstRow = (DgpImage_RGB *)((Dgp_Int8 *)dst + (num_rows * dstBpl)
		   + filler_row_length * sizeof(DgpImage_RGB));
  dstEnd = (DgpImage_RGB *)((Dgp_Int8 *)dstRow + (dstH * dstBpl));

  rsize = dstW * sizeof(DgpImage_RGB);

  /* do the first row. */
  /* 8.6.2000 / dhtesq. */
  /* This used to write up to dstRow + dstW, filling up only one third of the row for an rgb image. */
  /* Fixed... */
  for (dstCur = dstRow; dstCur < (DgpImage_RGB *)((Dgp_Int8 *)dstRow + rsize); dstCur ++) {
    *dstCur = val;
  }
  
  for (dstCur = (DgpImage_RGB *)((Dgp_Int8 *)dstRow + dstBpl); dstCur < dstEnd; 
       dstCur = (DgpImage_RGB *)((Dgp_Int8 *)dstCur + dstBpl)){
    memcpy(dstCur, dstRow, rsize);
  }
}









