#include "blitTables.h"


#include "image.h"

/* inline definition for diff compilers */
#include "util.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libblat/colorBlit.c,v 1.12 2003/02/10 22:53:19 mkrocken Exp $");

#include "endianness.h"

static INLINE void colorBlitBinaryRGBRow(DgpImage_RGB * dstCur, Dgp_UInt32 *srcCur,
				  Dgp_Int32 curSrcBitsLeft, Dgp_Int32 dstW, DgpImage_RGB color,
				  const Dgp_UInt32 colorWord[3]);

static INLINE void colorBlitRGBRGBRow(DgpImage_RGB * dstCur, DgpImage_RGB * srcCur, 
			       Dgp_UInt32 *maskCur, 
			       Dgp_Int32 curMaskBitsLeft, Dgp_Int32 dstW);

/*
21.8.2000 / dhtesq.
Decoder profiling suggests this is a good place for an inline...
*/
#ifdef _WIN32
DGP_LIBEXPORT 
#endif 
INLINE void 
DgpBlat_colorBlitBinaryRGB(Dgp_UInt32 * src,
			Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, DgpImage_RGB color)
{
  /* The general idea here is that we want to use an 8-bit lookup
   * table to convert 8-bits of packed image data to 24 bytes of
   * RGB data (which is 8 rgbs, which is 6 words), so we need to
   * for each row, find the first dest (RGB) pixel that is word
   * aligned, and up to that point, the conversion one bit at a
   * time, and then afterwards do the conversion 8-bits at a time.
	 */

  Dgp_UInt32 *srcRow,*srcCur;
  DgpImage_RGB * dstRow,*dstCur;

  Dgp_Int32 dstOff,srcOff,srcLpl,curSrcBitsLeft;
  Dgp_Int32 y;

  Dgp_UInt32 colorWord[3];

  srcLpl=srcBpl/sizeof(Dgp_UInt32);

  dstOff=((dstY-dstMinY)*dstBpl)+((dstX-dstMinX)*sizeof(DgpImage_RGB));
  dstRow=(DgpImage_RGB *)((Dgp_Int8 *)dst+dstOff);

		/* build colors for each state */
#ifdef _LITTLE_ENDIAN
  colorWord[0] = ((Dgp_UInt32)color.r << 24) |
    ((Dgp_UInt32)color.b << 16) |
    ((Dgp_UInt32)color.g << 8) | (Dgp_UInt32)color.r;
  colorWord[1] = ((Dgp_UInt32)color.g << 24) |
    ((Dgp_UInt32)color.r << 16) |
    ((Dgp_UInt32)color.b << 8) | (Dgp_UInt32)color.g;
  colorWord[2] = ((Dgp_UInt32)color.b << 24) |
    ((Dgp_UInt32)color.g << 16) |
    ((Dgp_UInt32)color.r << 8) | (Dgp_UInt32)color.b;
#else
  colorWord[0] = ((Dgp_UInt32)color.r << 24) |
    ((Dgp_UInt32)color.g << 16) |
    ((Dgp_UInt32)color.b << 8) | (Dgp_UInt32)color.r;
  colorWord[1] = ((Dgp_UInt32)color.g << 24) |
    ((Dgp_UInt32)color.b << 16) |
    ((Dgp_UInt32)color.r << 8) | (Dgp_UInt32)color.g;
  colorWord[2] = ((Dgp_UInt32)color.b << 24) |
    ((Dgp_UInt32)color.r << 16) |
    ((Dgp_UInt32)color.g << 8) | (Dgp_UInt32)color.b;
#endif


  /* this gets us the beginning of the row of the source, but */
  /* now we need to find the first word we're concerned with, and */
  /* which pixel (bit) in that word is the first to be converted. */
  srcOff=(srcLpl*(dstY-srcMinY));
  srcOff+=((dstX-srcMinX)/32);
  srcRow=src+srcOff;

  for (y = 0; y < dstH ; y ++)
    {
      dstCur = dstRow;
      srcCur = srcRow;
      curSrcBitsLeft = DGP_BITS_UINT32-(dstX-srcMinX) % DGP_BITS_UINT32;

      colorBlitBinaryRGBRow(dstCur, srcCur, curSrcBitsLeft, 
			    dstW, color, colorWord);

      dstRow = (DgpImage_RGB *)((Dgp_Int8 *)dstRow + dstBpl);
      srcRow = srcRow + srcLpl;
    }
}


static INLINE void colorBlitBinaryRGBRow(DgpImage_RGB * dstCur, Dgp_UInt32 *srcCur,
				  Dgp_Int32 curSrcBitsLeft, Dgp_Int32 dstW,
				  DgpImage_RGB color, const Dgp_UInt32 colorWord[3])
{
  Dgp_Int32 x = 0,temp;
  Dgp_Int32 wb;
  Dgp_UInt32 curSrcWord;
  Dgp_Int32 srcBitsUsed;
  Dgp_UInt32 * dstLong;
  const Dgp_UInt32 *dstLookup;


  /* do pixel by pixel until dstCur is on a word boundary. */
  wb = (Dgp_Int32)dstCur % sizeof(Dgp_UInt32);
  /* make sure this is less than the blatting width. */
  wb = MIN(wb,dstW);

  /* we can need to do either 0,1,2,3 pixels. We just special */
  /* case them all. */
  /* clear out all the bits we're not interested in. */
  curSrcWord = *(srcCur++) << (DGP_BITS_UINT32 - curSrcBitsLeft);
  srcBitsUsed = 0;

  /* make sure we actually have enough bits in curSrcWord to */
  /* to do these few writes. */
  if (curSrcBitsLeft < wb)
    {
      /* this assumes srcBitsUsed = 0; */
      curSrcWord = curSrcWord | (*srcCur >> curSrcBitsLeft);
      srcBitsUsed = DGP_BITS_UINT32 - curSrcBitsLeft;
      curSrcBitsLeft = DGP_BITS_UINT32;
    }

  if (wb == 1)
    {
      if ((curSrcWord >> 31))
	{
	  *dstCur=color;
	}
      dstCur++;
      curSrcBitsLeft--;
      curSrcWord = curSrcWord << 1;
      x++;
    }
  else if (wb == 2)
    {
      if ((curSrcWord >> 31))
	{
	  *dstCur=color;
	}
      dstCur++;
      curSrcWord = curSrcWord << 1;

      if ((curSrcWord >> 31))
	{
	  *dstCur=color;
	}
      dstCur++;
      curSrcWord = curSrcWord << 1;

      curSrcBitsLeft -= 2;
      x+=2;
    }
  else if (wb == 3)
    {
      if ((curSrcWord >> 31))
	{
	  *dstCur=color;
	}
      dstCur++;
      curSrcWord = curSrcWord << 1;

      if ((curSrcWord >> 31))
	{
	  *dstCur=color;
	}
      dstCur++;
      curSrcWord = curSrcWord << 1;

      if ((curSrcWord >> 31))
	{
	  *dstCur=color;
	}
      dstCur++;
      curSrcWord = curSrcWord << 1;

      curSrcBitsLeft -= 3;
      x+=3;
    }

  /* now we should be word aligned in the destination, or we should be done. */
  /* // assert(!((Dgp_Int32)dstCur % sizeof(Dgp_UInt32)) || x==dstW); */

  dstLong = (Dgp_UInt32 *)dstCur;
  /* now we're going to do this 32 pixels at a time in blocks of 4. */
  while ((x + DGP_BITS_UINT32) <= dstW)
    {
      /* fill the word up with 32 pixels, we'll grab what we can from srcCur, */
      /* and if we need more, we'll move srcCur forward. */
      if (curSrcBitsLeft != DGP_BITS_UINT32)
	{
	  if ((curSrcBitsLeft + (DGP_BITS_UINT32 - srcBitsUsed)) < DGP_BITS_UINT32)
	    {
	      /* spanning a word. */
	      /* we'll take as much as is left of this */
	      /* one, and get the rest from */
	      /* the next. */
	      if (srcBitsUsed < DGP_BITS_UINT32)
		{
		  curSrcWord = curSrcWord | 
		    ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
		  curSrcBitsLeft += (DGP_BITS_UINT32 - srcBitsUsed);
		}

	      srcCur ++;
	      srcBitsUsed = 0;
	    }

				/* no spanning a word. */
	  curSrcWord = curSrcWord |
	    ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
	  srcBitsUsed += (DGP_BITS_UINT32 - curSrcBitsLeft);
	  curSrcBitsLeft = DGP_BITS_UINT32;

	  /* assert(srcBitsUsed<=32 && srcBitsUsed>=0); */
	  /* assert(curSrcBitsLeft<=32 && curSrcBitsLeft>=0); */
	}

      /* we've got enough, now let's do the 4 8-bit blats. */
      dstLookup = monoToRGB8bit[(Dgp_UInt8)(curSrcWord >> 24)];
      dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & colorWord[0]);
      dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & colorWord[1]);
      dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & colorWord[2]);
      dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & colorWord[0]);
      dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & colorWord[1]);
      dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & colorWord[2]);
      dstLong += 6;

      dstLookup = monoToRGB8bit[(Dgp_UInt8)(curSrcWord >> 16)];
      dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & colorWord[0]);
      dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & colorWord[1]);
      dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & colorWord[2]);
      dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & colorWord[0]);
      dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & colorWord[1]);
      dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & colorWord[2]);
      dstLong += 6;

      dstLookup = monoToRGB8bit[(Dgp_UInt8)(curSrcWord >> 8)];
      dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & colorWord[0]);
      dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & colorWord[1]);
      dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & colorWord[2]);
      dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & colorWord[0]);
      dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & colorWord[1]);
      dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & colorWord[2]);
      dstLong += 6;

      dstLookup = monoToRGB8bit[(Dgp_UInt8)curSrcWord];
      dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & colorWord[0]);
      dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & colorWord[1]);
      dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & colorWord[2]);
      dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & colorWord[0]);
      dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & colorWord[1]);
      dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & colorWord[2]);
      dstLong += 6;

      curSrcBitsLeft = 0;
      curSrcWord = 0;

      x += DGP_BITS_UINT32;
    }


  while ((x + 8) <= dstW)
    {
      /* fill the word up with 8 pixels, we'll grab what we can from srcCur, */
      /* and if we need more, we'll move srcCur forward. */
      if (curSrcBitsLeft < 8)
	{
	  if ((8 - curSrcBitsLeft) > (DGP_BITS_UINT32 - srcBitsUsed))
	    {
				/* spanning a word. */
				/* we'll take as much as is left of this one, and get the rest from */
				/* the next. */
	      if (srcBitsUsed < DGP_BITS_UINT32)
		{
		  curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
		  curSrcBitsLeft += (DGP_BITS_UINT32 - srcBitsUsed);
		}

	      srcCur ++;
	      srcBitsUsed = 0;
	    }
	  /* no spanning a word, just take as much as we can. */
	  curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
	  temp = MIN(DGP_BITS_UINT32-srcBitsUsed,DGP_BITS_UINT32-curSrcBitsLeft);
	  srcBitsUsed += temp;
	  curSrcBitsLeft += temp;

	  /* assert(srcBitsUsed<=32 && srcBitsUsed>=0); */
	  /* assert(curSrcBitsLeft<=32 && curSrcBitsLeft>=0); */
	}

      /* we've got enough, now let's do the 8-bit blat. */
      dstLookup = monoToRGB8bit[(Dgp_UInt8)(curSrcWord >> 24)];
      dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & colorWord[0]);
      dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & colorWord[1]);
      dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & colorWord[2]);
      dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & colorWord[0]);
      dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & colorWord[1]);
      dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & colorWord[2]);
      dstLong += 6;

      curSrcWord = curSrcWord << 8;
      curSrcBitsLeft -= 8;
      x += 8;
    }

  dstCur = (DgpImage_RGB *)dstLong;
  /* so now we could have at most 7 more pixels to do. */
  wb = (dstW - x);
  /* make sure we have enough bits in curSrcWord to do it. */
  if (curSrcBitsLeft < wb)
    {
      if ((wb - curSrcBitsLeft) > (DGP_BITS_UINT32 - srcBitsUsed))
	{
				/* spanning a word. */
				/* we'll take as much as is left of this one, and get the rest from */
				/* the next. */
	  if (srcBitsUsed < DGP_BITS_UINT32)
	    {
	      curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
	      curSrcBitsLeft += (DGP_BITS_UINT32 - srcBitsUsed);
	    }

	  srcCur ++;
	  srcBitsUsed = 0;
	}

      /* no spanning a word, just take as much as we can. */
      curSrcWord = curSrcWord | ((*srcCur << srcBitsUsed) >> curSrcBitsLeft);
      temp = MIN(DGP_BITS_UINT32-srcBitsUsed,32-curSrcBitsLeft);
      srcBitsUsed += temp;
      curSrcBitsLeft += temp;

      /* assert(curSrcBitsLeft<=32 && curSrcBitsLeft>=0); */
    }

  /* now just special case them. */
  while (x < dstW)
    {
      if ((curSrcWord >> 31))
	{
	  *dstCur=color;
	}
      dstCur++;
      curSrcWord = curSrcWord << 1;

      x++;
    }
}


void DgpBlat_colorBlitRGBRGB(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		     Dgp_UInt32 * mask, Dgp_Int32 maskBpl,
		     DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		     Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH)
{
  Dgp_UInt32 *maskRow;
  DgpImage_RGB * dstRow,*dstCur,*srcRow,*srcCur;

  Dgp_Int32 dstOff,srcOff,maskOff,maskLpl,curMaskBitsLeft;
  Dgp_Int32 y;

  maskLpl=maskBpl/sizeof(Dgp_UInt32);

  dstOff=((dstY-dstMinY)*dstBpl)+((dstX-dstMinX)*sizeof(DgpImage_RGB));
  dstRow=(DgpImage_RGB *)((Dgp_Int8 *)dst+dstOff);

  srcOff=(srcMinY*srcBpl)+(srcMinX*sizeof(DgpImage_RGB));
  srcRow=(DgpImage_RGB *)((Dgp_Int8 *)src+srcOff);

  maskOff=(maskLpl*srcMinY);
  maskOff+=(srcMinX/32);
  maskRow=mask+maskOff;

  for (y = 0; y < dstH ; y ++)
    {
      dstCur = dstRow;
      srcCur = srcRow;
      /*curMaskBitsLeft = DGP_BITS_UINT32-(dstX-srcMinX) % DGP_BITS_UINT32; */
      curMaskBitsLeft = DGP_BITS_UINT32-srcMinX % DGP_BITS_UINT32;

      colorBlitRGBRGBRow(dstCur, srcCur, maskRow, curMaskBitsLeft, dstW);

      dstRow = (DgpImage_RGB *)((Dgp_Int8 *)dstRow + dstBpl);
      srcRow = (DgpImage_RGB *)((Dgp_Int8 *)srcRow + srcBpl);
      maskRow = maskRow + maskLpl;
    }
}


static INLINE void 
colorBlitRGBRGBRow(DgpImage_RGB * dstCur, DgpImage_RGB * srcCur, Dgp_UInt32 *maskCur, 
                   Dgp_Int32 curMaskBitsLeft, Dgp_Int32 dstW)
{
  Dgp_Int32 x = 0,temp;
  Dgp_Int32 wb, curSrcAlign;
  Dgp_UInt32 curMaskWord = 0;
  Dgp_Int32 maskBitsUsed;
  Dgp_UInt32 * dstLong,*srcLong,lastSrcLong = 0;
  const Dgp_UInt32 *dstLookup;


  /* do pixel by pixel until dstCur is on a word boundary. */
  wb = (Dgp_Int32)dstCur % sizeof(Dgp_UInt32);
  /* make sure this is less than the blatting width. */
  wb = MIN(wb,dstW);

  /* keep track of where in the src word the current pixel starts. */
  curSrcAlign = (Dgp_Int32)srcCur % sizeof(Dgp_UInt32);

  /* we can need to do either 0,1,2,3 pixels. We just special */
  /* case them all. */
  /* clear out all the bits we're not interested in. */
  curMaskWord = *(maskCur++) << (DGP_BITS_UINT32 - curMaskBitsLeft);
  maskBitsUsed = 0;

  /* make sure we actually have enough bits in curSrcWord to */
  /* to do these few writes. */
  if (curMaskBitsLeft < wb)
    {
      /* this assumes maskBitsUsed = 0; */
      curMaskWord = curMaskWord | (*maskCur >> curMaskBitsLeft);
      maskBitsUsed = DGP_BITS_UINT32 - curMaskBitsLeft;
      curMaskBitsLeft = DGP_BITS_UINT32;
    }

  if (wb == 1)
    {
      if ((curMaskWord >> 31))
	{
	  *dstCur=*srcCur;
	}
      dstCur++;
      srcCur++;
      curMaskBitsLeft--;
      curMaskWord = curMaskWord << 1;
      curSrcAlign = (curSrcAlign + sizeof(DgpImage_RGB)) % sizeof(Dgp_UInt32);
      x++;
    }
  else if (wb == 2)
    {
      if ((curMaskWord >> 31))
	{
	  *dstCur=*srcCur;
	}
      dstCur++;
      srcCur++;

      curMaskWord = curMaskWord << 1;

      if ((curMaskWord >> 31))
	{
	  *dstCur=*srcCur;
	}
      dstCur++;
      srcCur++;
      curMaskWord = curMaskWord << 1;

      curMaskBitsLeft -= 2;
      curSrcAlign = (curSrcAlign + 2 * sizeof(DgpImage_RGB)) % sizeof(Dgp_UInt32);
      x+=2;
    }
  else if (wb == 3)
    {
      if ((curMaskWord >> 31))
	{
	  *dstCur=*srcCur;
	}
      dstCur++;
      srcCur++;
      curMaskWord = curMaskWord << 1;

      if ((curMaskWord >> 31))
	{
	  *dstCur=*srcCur;
	}
      dstCur++;
      srcCur++;
      curMaskWord = curMaskWord << 1;

      if ((curMaskWord >> 31))
	{
	  *dstCur=*srcCur;
	}
      dstCur++;
      srcCur++;
      curMaskWord = curMaskWord << 1;

      curMaskBitsLeft -= 3;
      curSrcAlign = (curSrcAlign + 3 * sizeof(DgpImage_RGB)) % sizeof(Dgp_UInt32);
      x+=3;
    }

  /* now we should be word aligned in the destination, or we should be done. */
  /* assert(!((Dgp_Int32)dstCur % sizeof(Dgp_UInt32)) || x==dstW); */

  dstLong = (Dgp_UInt32 *)dstCur;

  /* this doesn't mean that we are word aligned in the source, however. */
  srcLong = (Dgp_UInt32 *)((Dgp_Int32)srcCur - curSrcAlign);

  /* moreover we should advance the srcLong by one long if the src is not word aligned */
  /* be warned that srcLong may not poDgp_Int32 to a valid pixel! so we have to be very careful */
  /* below about dereferencing it. */
  if(curSrcAlign)
    {
      lastSrcLong = *srcLong;
      srcLong++;
    }

  /* now we're going to do this 32 pixels at a time in blocks of 4. */
  while ((x + DGP_BITS_UINT32) <= dstW)
    {
      /* fill the word up with 32 pixels, we'll grab what we can from srcCur, */
      /* and if we need more, we'll move srcCur forward. */
      if (curMaskBitsLeft != DGP_BITS_UINT32)
	{
	  if ((curMaskBitsLeft + (DGP_BITS_UINT32 - maskBitsUsed)) < DGP_BITS_UINT32)
	    {
				/* spanning a word. */
				/* we'll take as much as is left of this one, and get the rest from */
				/* the next. */
	      if (maskBitsUsed < DGP_BITS_UINT32)
		{
		  curMaskWord = curMaskWord | ((*maskCur << maskBitsUsed) >> curMaskBitsLeft);
		  curMaskBitsLeft += (DGP_BITS_UINT32 - maskBitsUsed);
		}

	      maskCur ++;
	      maskBitsUsed = 0;
	    }

				/* no spanning a word. */
	  curMaskWord = curMaskWord | ((*maskCur << maskBitsUsed) >> curMaskBitsLeft);
	  maskBitsUsed += (DGP_BITS_UINT32 - curMaskBitsLeft);
	  curMaskBitsLeft = DGP_BITS_UINT32;

	  /* assert(maskBitsUsed<=32 && maskBitsUsed>=0); */
	  /* assert(curMaskBitsLeft<=32 && curMaskBitsLeft>=0); */
	}

      /* we've got enough, now let's do the 4 8-bit blats. */
      /* we need to special case each of the four possible alignments for the src. */
      switch( curSrcAlign)
	{
	case 0:	 /* no shifting! yahoo! */
	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 24)];
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & srcLong[0]);
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & srcLong[1]);
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & srcLong[2]);
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & srcLong[3]);
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & srcLong[4]);
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & srcLong[5]);
	  dstLong += 6;
	  srcLong += 6;

	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 16)];
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & srcLong[0]);
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & srcLong[1]);
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & srcLong[2]);
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & srcLong[3]);
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & srcLong[4]);
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & srcLong[5]);
	  dstLong += 6;
	  srcLong += 6;

	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 8)];
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & srcLong[0]);
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & srcLong[1]);
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & srcLong[2]);
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & srcLong[3]);
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & srcLong[4]);
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & srcLong[5]);
	  dstLong += 6;
	  srcLong += 6;

	  dstLookup = monoToRGB8bit[(Dgp_UInt8)curMaskWord];
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & srcLong[0]);
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & srcLong[1]);
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & srcLong[2]);
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & srcLong[3]);
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & srcLong[4]);
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & srcLong[5]);
	  dstLong += 6;
	  srcLong += 6;
	  break;

	case 1:
	  /* the src pixel starts on the second byte of the word, so every word */
	  /* of the dst actually depends on _two_ words from the src. ugly. */
	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 24)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 24 | lastSrcLong >> 8));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 24 | srcLong[0] >> 8));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 24 | srcLong[1] >> 8));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 24 | srcLong[2] >> 8));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 24 | srcLong[3] >> 8));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 24 | srcLong[4] >> 8));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 24 | lastSrcLong << 8));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 24 | srcLong[0] << 8));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 24 | srcLong[1] << 8));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 24 | srcLong[2] << 8));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 24 | srcLong[3] << 8));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 24 | srcLong[4] << 8));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;

	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 16)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 24 | lastSrcLong >> 8));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 24 | srcLong[0] >> 8));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 24 | srcLong[1] >> 8));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 24 | srcLong[2] >> 8));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 24 | srcLong[3] >> 8));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 24 | srcLong[4] >> 8));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 24 | lastSrcLong << 8));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 24 | srcLong[0] << 8));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 24 | srcLong[1] << 8));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 24 | srcLong[2] << 8));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 24 | srcLong[3] << 8));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 24 | srcLong[4] << 8));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;


	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 8)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 24 | lastSrcLong >> 8));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 24 | srcLong[0] >> 8));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 24 | srcLong[1] >> 8));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 24 | srcLong[2] >> 8));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 24 | srcLong[3] >> 8));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 24 | srcLong[4] >> 8));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 24 | lastSrcLong << 8));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 24 | srcLong[0] << 8));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 24 | srcLong[1] << 8));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 24 | srcLong[2] << 8));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 24 | srcLong[3] << 8));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 24 | srcLong[4] << 8));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;

	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 24 | lastSrcLong >> 8));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 24 | srcLong[0] >> 8));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 24 | srcLong[1] >> 8));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 24 | srcLong[2] >> 8));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 24 | srcLong[3] >> 8));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 24 | srcLong[4] >> 8));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 24 | lastSrcLong << 8));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 24 | srcLong[0] << 8));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 24 | srcLong[1] << 8));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 24 | srcLong[2] << 8));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 24 | srcLong[3] << 8));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 24 | srcLong[4] << 8));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;
	  break;

	case 2:
	  /* as above, but now the current pixel starts in the third byte of the srcWord */
	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 24)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 16 | lastSrcLong >> 16));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 16 | srcLong[0] >> 16));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 16 | srcLong[1] >> 16));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 16 | srcLong[2] >> 16));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 16 | srcLong[3] >> 16));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 16 | srcLong[4] >> 16));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 16 | lastSrcLong << 16));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 16 | srcLong[0] << 16));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 16 | srcLong[1] << 16));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 16 | srcLong[2] << 16));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 16 | srcLong[3] << 16));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 16 | srcLong[4] << 16));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;

	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 16)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 16 | lastSrcLong >> 16));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 16 | srcLong[0] >> 16));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 16 | srcLong[1] >> 16));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 16 | srcLong[2] >> 16));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 16 | srcLong[3] >> 16));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 16 | srcLong[4] >> 16));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 16 | lastSrcLong << 16));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 16 | srcLong[0] << 16));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 16 | srcLong[1] << 16));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 16 | srcLong[2] << 16));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 16 | srcLong[3] << 16));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 16 | srcLong[4] << 16));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;

	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 8)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 16 | lastSrcLong >> 16));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 16 | srcLong[0] >> 16));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 16 | srcLong[1] >> 16));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 16 | srcLong[2] >> 16));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 16 | srcLong[3] >> 16));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 16 | srcLong[4] >> 16));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 16 | lastSrcLong << 16));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 16 | srcLong[0] << 16));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 16 | srcLong[1] << 16));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 16 | srcLong[2] << 16));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 16 | srcLong[3] << 16));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 16 | srcLong[4] << 16));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;

	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 16 | lastSrcLong >> 16));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 16 | srcLong[0] >> 16));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 16 | srcLong[1] >> 16));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 16 | srcLong[2] >> 16));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 16 | srcLong[3] >> 16));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 16 | srcLong[4] >> 16));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 16 | lastSrcLong << 16));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 16 | srcLong[0] << 16));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 16 | srcLong[1] << 16));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 16 | srcLong[2] << 16));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 16 | srcLong[3] << 16));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 16 | srcLong[4] << 16));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;
	  break;

	case 3:
	  /* last one, this time the current pixel starts on the fourth byte of srcWord */
	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 24)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 8 | lastSrcLong >> 24));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 8 | srcLong[0] >> 24));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 8 | srcLong[1] >> 24));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 8 | srcLong[2] >> 24));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 8 | srcLong[3] >> 24));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 8 | srcLong[4] >> 24));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 8 | lastSrcLong << 24));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 8 | srcLong[0] << 24));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 8 | srcLong[1] << 24));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 8 | srcLong[2] << 24));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 8 | srcLong[3] << 24));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 8 | srcLong[4] << 24));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;

	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 16)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 8 | lastSrcLong >> 24));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 8 | srcLong[0] >> 24));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 8 | srcLong[1] >> 24));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 8 | srcLong[2] >> 24));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 8 | srcLong[3] >> 24));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 8 | srcLong[4] >> 24));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 8 | lastSrcLong << 24));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 8 | srcLong[0] << 24));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 8 | srcLong[1] << 24));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 8 | srcLong[2] << 24));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 8 | srcLong[3] << 24));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 8 | srcLong[4] << 24));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;

	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 8)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 8 | lastSrcLong >> 24));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 8 | srcLong[0] >> 24));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 8 | srcLong[1] >> 24));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 8 | srcLong[2] >> 24));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 8 | srcLong[3] >> 24));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 8 | srcLong[4] >> 24));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 8 | lastSrcLong << 24));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 8 | srcLong[0] << 24));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 8 | srcLong[1] << 24));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 8 | srcLong[2] << 24));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 8 | srcLong[3] << 24));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 8 | srcLong[4] << 24));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;

	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 8 | lastSrcLong >> 24));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 8 | srcLong[0] >> 24));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 8 | srcLong[1] >> 24));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 8 | srcLong[2] >> 24));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 8 | srcLong[3] >> 24));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 8 | srcLong[4] >> 24));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 8 | lastSrcLong << 24));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 8 | srcLong[0] << 24));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 8 | srcLong[1] << 24));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 8 | srcLong[2] << 24));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 8 | srcLong[3] << 24));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 8 | srcLong[4] << 24));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;
	  break;

	default:
	  /* should never hit this case... */
	  /* assert(0); */
		;
	}

      curMaskBitsLeft = 0;
      curMaskWord = 0;
      x += DGP_BITS_UINT32;
    } /* end of word by word blatting */


  while ((x + 8) <= dstW)
    {
      /* fill the word up with 8 pixels, we'll grab what we can from maskCur, */
      /* and if we need more, we'll move maskCur forward. */
      if (curMaskBitsLeft < 8)
	{
	  if ((8 - curMaskBitsLeft) > (DGP_BITS_UINT32 - maskBitsUsed))
	    {
				/* spanning a word. */
				/* we'll take as much as is left of this one, and get the rest from */
				/* the next. */
	      if (maskBitsUsed < DGP_BITS_UINT32)
		{
		  curMaskWord = curMaskWord | ((*maskCur << maskBitsUsed) >> curMaskBitsLeft);
		  curMaskBitsLeft += (DGP_BITS_UINT32 - maskBitsUsed);
		}

	      maskCur ++;
	      maskBitsUsed = 0;
	    }
	  /* no spanning a word, just take as much as we can. */
	  curMaskWord = curMaskWord | ((*maskCur << maskBitsUsed) >> curMaskBitsLeft);
	  temp = MIN(DGP_BITS_UINT32-maskBitsUsed,DGP_BITS_UINT32-curMaskBitsLeft);
	  maskBitsUsed += temp;
	  curMaskBitsLeft += temp;

	  /* assert(maskBitsUsed<=32 && maskBitsUsed>=0); */
	  /* assert(curMaskBitsLeft<=32 && curMaskBitsLeft>=0); */
	}

      /* we've got enough, now let's do the 8-bit blat. */
      switch(curSrcAlign)
	{
	case 0:
	  /* no shifting! whoppee! */
	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 24)];
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & srcLong[0]);
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & srcLong[1]);
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & srcLong[2]);
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & srcLong[3]);
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & srcLong[4]);
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & srcLong[5]);

	  dstLong += 6;
	  srcLong += 6;
	  break;

	case 1:
	  /* the src pixel starts on the second byte of the word, so every word */
	  /* of the dst actually depends on _two_ words from the src. ugly. */
	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 24)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 24 | lastSrcLong >> 8));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 24 | srcLong[0] >> 8));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 24 | srcLong[1] >> 8));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 24 | srcLong[2] >> 8));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 24 | srcLong[3] >> 8));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 24 | srcLong[4] >> 8));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 24 | lastSrcLong << 8));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 24 | srcLong[0] << 8));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 24 | srcLong[1] << 8));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 24 | srcLong[2] << 8));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 24 | srcLong[3] << 8));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 24 | srcLong[4] << 8));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;
	  break;

	case 2:
	  /* as above, but now the current pixel starts in the third byte of the srcWord */
	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 24)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 16 | lastSrcLong >> 16));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 16 | srcLong[0] >> 16));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 16 | srcLong[1] >> 16));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 16 | srcLong[2] >> 16));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 16 | srcLong[3] >> 16));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 16 | srcLong[4] >> 16));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 16 | lastSrcLong << 16));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 16 | srcLong[0] << 16));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 16 | srcLong[1] << 16));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 16 | srcLong[2] << 16));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 16 | srcLong[3] << 16));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 16 | srcLong[4] << 16));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;
	  break;

	case 3:
	  /* last one, this time the current pixel starts on the fourth byte of srcWord */
	  dstLookup = monoToRGB8bit[(Dgp_UInt8)(curMaskWord >> 24)];
#ifdef _LITTLE_ENDIAN
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] << 8 | lastSrcLong >> 24));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] << 8 | srcLong[0] >> 24));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] << 8 | srcLong[1] >> 24));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] << 8 | srcLong[2] >> 24));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] << 8 | srcLong[3] >> 24));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] << 8 | srcLong[4] >> 24));
#else /* _BIG_ENDIAN */
	  dstLong[0] = (dstLookup[0] & dstLong[0]) | (~dstLookup[0] & (srcLong[0] >> 8 | lastSrcLong << 24));
	  dstLong[1] = (dstLookup[1] & dstLong[1]) | (~dstLookup[1] & (srcLong[1] >> 8 | srcLong[0] << 24));
	  dstLong[2] = (dstLookup[2] & dstLong[2]) | (~dstLookup[2] & (srcLong[2] >> 8 | srcLong[1] << 24));
	  dstLong[3] = (dstLookup[3] & dstLong[3]) | (~dstLookup[3] & (srcLong[3] >> 8 | srcLong[2] << 24));
	  dstLong[4] = (dstLookup[4] & dstLong[4]) | (~dstLookup[4] & (srcLong[4] >> 8 | srcLong[3] << 24));
	  dstLong[5] = (dstLookup[5] & dstLong[5]) | (~dstLookup[5] & (srcLong[5] >> 8 | srcLong[4] << 24));
#endif
	  lastSrcLong = srcLong[5];
	  dstLong += 6;
	  srcLong += 6;
	  break;

	default:
	  /* should never hit this case... */
	  /* assert(0); */
		;
	  }

      curMaskWord = curMaskWord << 8;
      curMaskBitsLeft -= 8;
      x += 8;
    } /* end of byte by byte blatting */

  dstCur = (DgpImage_RGB *)dstLong;
  /* so now we could have at most 7 more pixels to do. */
  wb = (dstW - x);

  /* also we need to find the current src- if it's not aligned then it's in just before srcLong */
  if(curSrcAlign)
    {
      srcCur = (DgpImage_RGB *)((Dgp_Int32)srcLong + curSrcAlign - sizeof(Dgp_UInt32));
    }
  else
    {
      srcCur = (DgpImage_RGB *)srcLong;
    }

  /* make sure we have enough bits in curSrcWord to do it. */
  if (curMaskBitsLeft < wb)
    {
      if ((wb - curMaskBitsLeft) > (DGP_BITS_UINT32 - maskBitsUsed))
	{
				/* spanning a word. */
				/* we'll take as much as is left of this one, and get the rest from */
				/* the next. */
	  if (maskBitsUsed < DGP_BITS_UINT32)
	    {
	      curMaskWord = curMaskWord | ((*maskCur << maskBitsUsed) >> curMaskBitsLeft);
	      curMaskBitsLeft += (DGP_BITS_UINT32 - maskBitsUsed);
	    }

	  maskCur++;
	  maskBitsUsed = 0;
	}

      /* no spanning a word, just take as much as we can. */
      curMaskWord = curMaskWord | ((*maskCur << maskBitsUsed) >> curMaskBitsLeft);
      temp = MIN(DGP_BITS_UINT32-maskBitsUsed,DGP_BITS_UINT32-curMaskBitsLeft);
      maskBitsUsed += temp;
      curMaskBitsLeft += temp;

      /* assert(curMaskBitsLeft<=32 && curMaskBitsLeft>=0); */
    }

  /* now just special case them. */
  while (x < dstW)
    {

      if ((curMaskWord >> 31))
	{
	  *dstCur=*srcCur;
	}
      dstCur++;
      srcCur++;
      curMaskWord = curMaskWord << 1;

      x++;
    }
}
