#ifndef __SCALEBLIT_H__
#define __SCALEBLIT_H__

#include "resolution_t.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* libwjr for DgpImage_RGB */
#include "image.h"

/* DGP_allocators */
#include "dgp_allocators.h"

#include "dgp_types.h"

/* If you already have the usual definitions for BGR, just define this
   out.
 */
#ifndef SCALEBLIT_NO_DEF_BGR
typedef struct
	{
	Dgp_UInt8 b;
	Dgp_UInt8 g;
	Dgp_UInt8 r;
} DgpBlat_BGR;
#endif

/* These let you know how much src you need to generate a particular amount of
	dst, and how much dst is affected by a certain amount of source.
 */
extern void 
DgpBlat_scaleBlitCalcNeeded(Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
			    DgpCoord_Resolution * res,
			    Dgp_Int32 srcImgW, Dgp_Int32 srcImgH,
			    Dgp_Int32 * srcX, Dgp_Int32 * srcY, Dgp_Int32 * srcW, Dgp_Int32 * srcH);
  
extern void 
DgpBlat_scaleBlitCalcAffected(Dgp_Int32 srcX, Dgp_Int32 srcY, Dgp_Int32 srcW, Dgp_Int32 srcH,
			   DgpCoord_Resolution * res,
			   Dgp_Int32 dstImgW, Dgp_Int32 dstImgH,
			   Dgp_Int32 * dstX, Dgp_Int32 * dstY, Dgp_Int32 * dstW, Dgp_Int32 * dstH);

/* All named similarly:
   if the function scales, it's in the name, else it's just a blit
   the input type is the first word, the output type is the second.
   
   All functions take params in the same way:
   src is a pointer to the source at pixel (srcMinX, srcMinY)
   dst is a pointer to the dest at (dstMinX, dstMinY)
   you want to generate (dstX, dstY) to (dstX + dstW - 1, dstY + dstH - 1)
   of the destination, assuming that the full source image is
   the full dest image at some other scale. You can tweak with
   this to get the behaviour you want for blatting tokens or 
   whatnot.
   
   You're in charge of making sure everythings alright. If you're not
   sure, call the calc functions, and make sure.
   
   if the functions return a value, it's because they need a temp
   buffer (they need malloc/free too)
   
   returns 1 if successful
   0 if not (the memory allocation failed)
*/
  
/* no scaling */
extern void 
DgpBlat_blitRGBBGR(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH);
  
extern void 
DgpBlat_blitRGB(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
	     DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
	     Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH);
  
/* uses the 256x6 8 bit lookup table in blitTables.h, */
extern void 
DgpBlat_blitBinaryRGB(Dgp_UInt32 * src,
		   Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		   DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		   Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH);
  
  
/* scales by subsampling. All scales handled. */
extern Dgp_Int32 
DgpBlat_scaleBlitBinaryRGB(Dgp_UInt32 * src,
		       Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		       DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		       Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		       DgpCoord_Resolution * res, Dgp_Allocators *a);
  
/* these two handle general subsampling, */
/* and do special case the 1.0 scale case. */
extern void 
DgpBlat_scaleBlitRGBBGR(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		     DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		     Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		     DgpCoord_Resolution * res);
  
extern void 
DgpBlat_scaleBlitRGB(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		  DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		  Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		  DgpCoord_Resolution * res);


/* if you have 2 regions in the same block of memory, this'll copy one to */
/* the other without messing things up. */
extern void 
DgpBlat_copyRGBOverlapNoClip(DgpImage_RGB * baseMem, Dgp_Int32 baseBpl, Dgp_Int32 w, Dgp_Int32 h,
			  Dgp_Int32 srcX, Dgp_Int32 srcY, Dgp_Int32 dstX, Dgp_Int32 dstY);

/*#if 0   */
/* What's wrong?  Resurrecting for santa;  August 2000. / dhtesq. */

/* this one's kinda sketchy. It approximates a binary to binary scale for */
/* cases from > 0% to <=400%. It's fast, but only generates scales accurate to */
/* 3.125%, so make sure your scale is either a multiple of 3.125% */
/* or you have enough source around so that it can use the amount */
/* needed for the 3.125% increment below your scale. */
/* */
/* Since no one else uses this function, I'm changing the internals to make it more friendly */
/* to token operations.  The other scaling functions assume that the src buffer fits into */
/* the (entire )destintation buffer at some scale, whereas when you start with a token bitmap */
/* you want the src buffer to fit into the *destination region* at the desired scale. */
/* / dhtesq. */
void scaleBlitBinaryApprox(Dgp_UInt32       *src,
                           Dgp_Int32        srcBpl, 
                           Dgp_Int32        srcMinX, 
                           Dgp_Int32        srcMinY,
                           Dgp_UInt32       *dst, 
                           Dgp_Int32        dstBpl,
                           Dgp_Int32        dstMinX, 
                           Dgp_Int32        dstMinY,
                           Dgp_Int32        dstX, 
                           Dgp_Int32        dstY, 
                           Dgp_Int32        dstW, 
                           Dgp_Int32        dstH,
                           DgpCoord_Resolution *res);
/*#endif // 0 */
#ifdef __cplusplus
}
#endif
#endif /*__SCALEBLIT_H__ */
