#include "scaleBlitAvgColor.h"
#include "scaleBlit.h"

#include "resolution_t.h"

#include "assert.h"
#include <stdlib.h>

#if !defined(VXWORKS) && !defined(EQUATOR_CA_ENV)
#include <memory.h>
#endif    /* (!VXWORKS) */

#include <math.h>

/* libwjr for MIN/MAX */
/* inline definition for different compilers */
#include "util.h"

/* The code for average by 2,3,4,5,8 are all very similar. Optimization/fixes to
	 one should be applied to all 
 */

/* The more general stuff is very similar to that in scaleBlitMono
	apply changes there as well 
 */

INLINE void DgpBlat_scaleRGBGRAvgby2Row(DgpImage_RGB * srcRow, Dgp_Int32 srcBpl, Dgp_Int32 * avgBuf, Dgp_Int32 rgbDstW);
INLINE void DgpBlat_scaleRGBGRAvgby3Row(DgpImage_RGB * srcRow, Dgp_Int32 srcBpl, Dgp_Int32 * avgBuf, Dgp_Int32 rgbDstW);
INLINE void DgpBlat_scaleRGBGRAvgby4Row(DgpImage_RGB * srcRow, Dgp_Int32 srcBpl, Dgp_Int32 * avgBuf, Dgp_Int32 rgbDstW);
INLINE void DgpBlat_scaleRGBGRAvgby5Row(DgpImage_RGB * srcRow, Dgp_Int32 srcBpl, Dgp_Int32 * avgBuf, Dgp_Int32 rgbDstW);
INLINE void DgpBlat_scaleRGBGRAvgby8Row(DgpImage_RGB * srcRow, Dgp_Int32 srcBpl, Dgp_Int32 * avgBuf, Dgp_Int32 rgbDstW);

Dgp_Int32 DgpBlat_scaleRGBBGRAvgby2(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a)
{
  /* We want to geneDGPCOORD_Rate (dstX,dstY) to (dstX+dstW-1,dstY+dstH-1) at 1/2
     scale by averaging pixels.
	 */
  DgpImage_RGB * srcRow;
  DgpBlat_BGR * dstRow;
  Dgp_Int32 * avgBuf;
  Dgp_Int32 srcX,srcY;
  Dgp_Int32 y,x,x3;
  Dgp_Int32 rgbDstW;

  srcX = (dstX * 2);
  srcY = (dstY * 2);

  rgbDstW = dstW * 3;

  avgBuf = (Dgp_Int32 *)a->malloc(3*dstW*sizeof(Dgp_Int32));
  if(!avgBuf)
    return 0;

  srcRow = (DgpImage_RGB *)((char *)src + (((srcY-srcMinY) * srcBpl) + ((srcX-srcMinX) * sizeof(DgpImage_RGB))));
  dstRow = (DgpBlat_BGR *)((char *)dst + (((dstY-dstMinY) * dstBpl) + ((dstX-dstMinX) * sizeof(DgpBlat_BGR))));
  for (y = 0; y < dstH; y++)
    {
      DgpBlat_scaleRGBGRAvgby2Row(srcRow,srcBpl, avgBuf, rgbDstW);

      for (x = 0, x3 = 0 ; x < dstW ; x++, x3+=3)
	{
	  dstRow[x].r=(int)(avgBuf[x3]/4);
	  dstRow[x].g=(int)(avgBuf[x3+1]/4);
	  dstRow[x].b=(int)(avgBuf[x3+2]/4);
	}

      srcRow = (DgpImage_RGB *)((char *)srcRow + (2*srcBpl));
      dstRow = (DgpBlat_BGR *)((char *)dstRow + dstBpl);
    }
	

  a->free(avgBuf);
  return 1;
}


Dgp_Int32 DgpBlat_scaleRGBBGRAvgby3(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a)
{

  /* We want to geneDGPCOORD_Rate (dstX,dstY) to (dstX+dstW-1,dstY+dstH-1) at 1/3
     scale by averaging pixels.
  */
  DgpImage_RGB * srcRow;
  DgpBlat_BGR * dstRow;
  Dgp_Int32 * avgBuf;
  Dgp_Int32 srcX,srcY;
  Dgp_Int32 y,x,x3;
  Dgp_Int32 rgbDstW;

  srcX = (dstX * 3);
  srcY = (dstY * 3);

  rgbDstW = dstW * 3;

  avgBuf = (Dgp_Int32 *)a->malloc(3*dstW*sizeof(Dgp_Int32));
  if(!avgBuf)
    return 0;

  srcRow = (DgpImage_RGB *)((char *)src + (((srcY-srcMinY) * srcBpl) + ((srcX-srcMinX) * sizeof(DgpImage_RGB))));
  dstRow = (DgpBlat_BGR *)((char *)dst + (((dstY-dstMinY) * dstBpl) + ((dstX-dstMinX) * sizeof(DgpBlat_BGR))));
  for (y = 0; y < dstH; y++)
    {
      DgpBlat_scaleRGBGRAvgby3Row(srcRow,srcBpl, avgBuf, rgbDstW);
		
      for (x = 0, x3 = 0 ; x < dstW ; x++, x3+=3)
	{
	  dstRow[x].r=(int)(avgBuf[x3]/9);
	  dstRow[x].g=(int)(avgBuf[x3+1]/9);
	  dstRow[x].b=(int)(avgBuf[x3+2]/9);
	}

      srcRow = (DgpImage_RGB *)((char *)srcRow + (3*srcBpl));
      dstRow = (DgpBlat_BGR *)((char *)dstRow + dstBpl);
    }
	

  a->free(avgBuf);
  return 1;
}


Dgp_Int32 DgpBlat_scaleRGBBGRAvgby4(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a)
{

  /* We want to geneDGPCOORD_Rate (dstX,dstY) to (dstX+dstW-1,dstY+dstH-1) at 1/4
     scale by averaging pixels.
  */
  DgpImage_RGB * srcRow;
  DgpBlat_BGR * dstRow;
  Dgp_Int32 * avgBuf;
  Dgp_Int32 srcX,srcY;
  Dgp_Int32 y,x,x3;
  Dgp_Int32 rgbDstW;

  srcX = (dstX * 4);
  srcY = (dstY * 4);

  rgbDstW = dstW * 3;

  avgBuf = (Dgp_Int32 *)a->malloc(3*dstW*sizeof(Dgp_Int32));
  if(!avgBuf)
    return 0;

  srcRow = (DgpImage_RGB *)((char *)src + (((srcY-srcMinY) * srcBpl) + ((srcX-srcMinX) * sizeof(DgpImage_RGB))));
  dstRow = (DgpBlat_BGR *)((char *)dst + (((dstY-dstMinY) * dstBpl) + ((dstX-dstMinX) * sizeof(DgpBlat_BGR))));
  for (y = 0; y < dstH; y++)
    {
      DgpBlat_scaleRGBGRAvgby4Row(srcRow,srcBpl, avgBuf, rgbDstW);
	
      for (x = 0, x3 = 0 ; x < dstW ; x++, x3+=3)
	{
	  dstRow[x].r=(int)(avgBuf[x3]/16);
	  dstRow[x].g=(int)(avgBuf[x3+1]/16);
	  dstRow[x].b=(int)(avgBuf[x3+2]/16);
	}

      srcRow = (DgpImage_RGB *)((char *)srcRow + (4*srcBpl));
      dstRow = (DgpBlat_BGR *)((char *)dstRow + dstBpl);
    }
	

  a->free(avgBuf);
  return 1;
}



Dgp_Int32 DgpBlat_scaleRGBBGRAvgby5(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a)
{

  /* We want to geneDGPCOORD_Rate (dstX,dstY) to (dstX+dstW-1,dstY+dstH-1) at 1/5
     scale by averaging pixels.
  */
  DgpImage_RGB * srcRow;
  DgpBlat_BGR * dstRow;
  Dgp_Int32 * avgBuf;
  Dgp_Int32 srcX,srcY;
  Dgp_Int32 y,x,x3;
  Dgp_Int32 rgbDstW;

  srcX = (dstX * 5);
  srcY = (dstY * 5);

  rgbDstW = dstW * 3;

  avgBuf = (Dgp_Int32 *)a->malloc(3*dstW*sizeof(Dgp_Int32));
  if(!avgBuf)
    return 0;

  srcRow = (DgpImage_RGB *)((char *)src + (((srcY-srcMinY) * srcBpl) + ((srcX-srcMinX) * sizeof(DgpImage_RGB))));
  dstRow = (DgpBlat_BGR *)((char *)dst + (((dstY-dstMinY) * dstBpl) + ((dstX-dstMinX) * sizeof(DgpBlat_BGR))));
  for (y = 0; y < dstH; y++)
    {
      DgpBlat_scaleRGBGRAvgby5Row(srcRow,srcBpl, avgBuf, rgbDstW);
	
      for (x = 0, x3 = 0 ; x < dstW ; x++, x3+=3)
	{
	  dstRow[x].r=(int)(avgBuf[x3]/25);
	  dstRow[x].g=(int)(avgBuf[x3+1]/25);
	  dstRow[x].b=(int)(avgBuf[x3+2]/25);
	}

      srcRow = (DgpImage_RGB *)((char *)srcRow + (5*srcBpl));
      dstRow = (DgpBlat_BGR *)((char *)dstRow + dstBpl);
    }
	

  a->free(avgBuf);
  return 1;
}


Dgp_Int32 DgpBlat_scaleRGBBGRAvgby8(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a)
{

  /* We want to geneDGPCOORD_Rate (dstX,dstY) to (dstX+dstW-1,dstY+dstH-1) at 1/8
     scale by averaging pixels.
  */
  DgpImage_RGB * srcRow;
  DgpBlat_BGR * dstRow;
  Dgp_Int32 * avgBuf;
  Dgp_Int32 srcX,srcY;
  Dgp_Int32 y,x,x3;
  Dgp_Int32 rgbDstW;

  srcX = (dstX * 8);
  srcY = (dstY * 8);

  rgbDstW = dstW * 3;

  avgBuf = (Dgp_Int32 *)a->malloc(3*dstW*sizeof(Dgp_Int32));
  if(!avgBuf)
    return 0;

  srcRow = (DgpImage_RGB *)((char *)src + (((srcY-srcMinY) * srcBpl) + ((srcX-srcMinX) * sizeof(DgpImage_RGB))));
  dstRow = (DgpBlat_BGR *)((char *)dst + (((dstY-dstMinY) * dstBpl) + ((dstX-dstMinX) * sizeof(DgpBlat_BGR))));
  for (y = 0; y < dstH; y++)
    {
      DgpBlat_scaleRGBGRAvgby8Row(srcRow,srcBpl, avgBuf, rgbDstW);
	
      for (x = 0, x3 = 0 ; x < dstW ; x++, x3+=3)
	{
	  dstRow[x].r=(int)(avgBuf[x3]/64);
	  dstRow[x].g=(int)(avgBuf[x3+1]/64);
	  dstRow[x].b=(int)(avgBuf[x3+2]/64);
	}

      srcRow = (DgpImage_RGB *)((char *)srcRow + (8*srcBpl));
      dstRow = (DgpBlat_BGR *)((char *)dstRow + dstBpl);
    }
	

  a->free(avgBuf);
  return 1;
}



Dgp_Int32 DgpBlat_scaleRGBBGRAvgSubGen(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			 DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			 Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
			 DgpCoord_Resolution * res, Dgp_Allocators * a)
{
  /* We need to find the best averaging and then scale down a little from
     there, note that we have to use the max(scaleX,scaleY) as to determine
     which to use 
  */
  DgpCoord_Rat maxScale;
  Dgp_Int32 avgFactor = 0;

  Dgp_Int32 x,x3,y;

  DgpImage_RGB *srcRow = src;
  /* srcOff is just a junk var used to compute offsets for the src. */
  Dgp_Int32 srcOff;

  double scaleRatioX,invScaleY;
  Dgp_Int32 avgFacSq;

  DgpBlat_BGR *dstRow;
  Dgp_Int32 * avgBuf;

  Dgp_Int32 rgbDstW;

  /* dstRow is the first pixel in the first row to be blitted. */
  /* get the top row. */
  Dgp_Int32 dstOff=((dstY-dstMinY) * dstBpl);
  /* advance the x position. */
  dstOff += ((dstX-dstMinX)*sizeof(DgpBlat_BGR));
		
  dstRow = (DgpBlat_BGR *)((char *)dst + dstOff);
  if (DgpCoord_g(res->xRes, res->yRes)){
    maxScale = res->xRes;
  }
  else {
    maxScale = res->yRes;
  }
  /* You shouldn't be calling this function if there's no averaging that 
     needs to be done.
  */
  assert(DgpCoord_le(maxScale,DgpCoord_eq(1,2)));

  if (DgpCoord_le(maxScale, DgpCoord_eq(1, 2))){
    avgFactor = 2;
  }
  if (DgpCoord_le(maxScale, DgpCoord_eq(1, 3))){
    avgFactor = 3;
  }
  if (DgpCoord_le(maxScale, DgpCoord_eq(1, 4))){
    avgFactor = 4;
  }
  if (DgpCoord_le(maxScale, DgpCoord_eq(1, 5))){
    avgFactor = 5;
  }
  if (DgpCoord_le(maxScale, DgpCoord_eq(1, 8))){
    avgFactor = 8;
  }

  avgFacSq = avgFactor * avgFactor;

  scaleRatioX=((double)1/avgFactor)/DgpCoord_float(res->xRes);
  invScaleY = (double)res->yRes.den / res->yRes.num;
  rgbDstW = (int)(dstW*scaleRatioX)*3;

  avgBuf = (Dgp_Int32 *)a->malloc(rgbDstW*sizeof(Dgp_Int32));
  if (!avgBuf)
    return 0;

  for (y=0; y<dstH; y++)
    {
      srcOff = ((int)((dstY + y) * invScaleY)) - srcMinY;
      srcOff *= srcBpl;
      /* dunno if this is right. */
      srcOff += (int)((dstX / scaleRatioX) - srcMinX);
      srcRow = (DgpImage_RGB *)((char *)src + srcOff);

      if (avgFactor == 2)
	{
	  DgpBlat_scaleRGBGRAvgby2Row(srcRow, srcBpl, avgBuf, rgbDstW);
	}
      else if (avgFactor == 3)
	{
	  DgpBlat_scaleRGBGRAvgby3Row(srcRow, srcBpl, avgBuf, rgbDstW);
	}
      else if (avgFactor == 4)
	{
	  DgpBlat_scaleRGBGRAvgby4Row(srcRow, srcBpl, avgBuf, rgbDstW);
	}
      else if (avgFactor == 5)
	{
	  DgpBlat_scaleRGBGRAvgby5Row(srcRow, srcBpl, avgBuf, rgbDstW);
	}
      else /* avgFactor == 8 */
	{
	  DgpBlat_scaleRGBGRAvgby8Row(srcRow, srcBpl, avgBuf, rgbDstW);
	}
      /* Now we need to pick the pixels of this row that we actually want.
       */
		
      for (x = 0; x < dstW ; x++)
	{
	  x3 = 3*(int)(x * scaleRatioX);
	  dstRow[x].r=(int)(avgBuf[x3]/avgFacSq);
	  dstRow[x].g=(int)(avgBuf[x3+1]/avgFacSq);
	  dstRow[x].b=(int)(avgBuf[x3+2]/avgFacSq);
	}
      dstRow = (DgpBlat_BGR *)((char *)dstRow + dstBpl);
    }


  a->free(avgBuf);
  return 1;
}


/* srcRow - ptr to upper left boundary of first src mxm block to be scaled.
 * srcBpl -	bytes per line of the source.
 * dstRow - ptr to upper left corner of first destination pixel in upper left nxn block 
 * dstBpl - bytes per line in dst.
 * n	  - size of dst scaled block
 * startBlockRow - what row of the block (0..n-1) goes into dstRow
 * endBlockRow - what row of the block (startBlockRow..n) is the last row of the block
 *		[note exclusive]
 * firstStart - what column of the first block goes into dstRow (0..n-1)
 * firstEnd - what is the last column of the first block [note exclusive]
 *	 (normally n, but you could be blatting a really small width < n)
 *  lastEnd - what is the last pixel of the src row that should be accumulated
 *		(normally dstX+dstW, but could be different if srcImgWidth % m != 0)
 */
static void DgpBlat_scaleRGBBGRAvgBlockRow(DgpImage_RGB * srcRow, Dgp_Int32 srcBpl, 
			    DgpBlat_BGR * dstRow, Dgp_Int32 dstBpl, Dgp_Int32 n, 
			    Dgp_Int32 startBlockRow, Dgp_Int32 endBlockRow,
			    Dgp_Int32 firstStart, Dgp_Int32 firstEnd, Dgp_Int32 lastEnd,
			    Dgp_Int32 * nsrc, Dgp_Int32 * buf, Dgp_Int32 bufLen)
{
  Dgp_Int32 y2, y3;
  Dgp_Int32 dy2;
  Dgp_Int32 dx, dx2;
  Dgp_Int32 dx2rgb;
  Dgp_Int32 blockPix;
  DgpImage_RGB * neededEnd;

  DgpImage_RGB * data;
  DgpBlat_BGR * dstMap;

  Dgp_Int32 lastSize;
  Dgp_Int32 lastStart;

	
  /* when does the full block stuff end? */
  /* so normally, it's just lastEnd % n */
  /* but if it's a really short blat (only does a partial first block) */
  /* we need to make sure the other loops don't run, so we set it to lastEnd */
  lastStart = lastEnd;
  if (firstEnd == n)
    lastStart = lastEnd - (lastEnd % n);
	
  /* precompute the size of the last block */
  lastSize = lastEnd - lastStart;

  /* offset the source row the number of rows it would have moved */
  /* if we were doing the block from the beginning. */
  for(y2 = 0, y3 = 0;  y3 < startBlockRow; y3++)
    {
      y2 += nsrc[y3];
      srcRow = (DgpImage_RGB *)((char *)srcRow + (nsrc[y3] * srcBpl));
    }

  /* same for the x */
  for (dx2 = 0; dx2 < firstStart ; dx2++)
    srcRow += nsrc[dx2];

  /* y is the block number in y */
  /* y2 is the start of the source needed for this dst block line */
  /* y3 is the position of the source while accumulating for a dst block line. */
  /* iteDGPCOORD_Rate over dest rows */

  /* don't set y2 to zero, we've calculated the number of src rows that */
  /* starting on this destination row would be. */
  for (dy2 = startBlockRow; dy2 < endBlockRow; y2 += nsrc[dy2], dy2++)
    {
      /* Clear the accumulation buffers */
      memset(buf,0,bufLen*sizeof(int));

      /* IteDGPCOORD_Rate over the source rows corresponding to this dest row */
      for (y3 = 0; y3 < nsrc[dy2]; y3++)
	{
	  data = srcRow;
	  dx2rgb = firstStart * 3;
	  for (dx2 = firstStart; dx2 < firstEnd; dx2++)
	    {
	      for (neededEnd = data + nsrc[dx2]; data < neededEnd; data++)
		{
		  buf[dx2rgb] += data->r;
		  buf[dx2rgb + 1] += data->g;
		  buf[dx2rgb + 2] += data->b;
		}
	      dx2rgb += 3;
	    }

	  /* now do the general x block as far as we can. */
	  /* dx is block position (dst) */
	  for (dx = firstEnd; dx < lastStart; dx += n)
	    {
				/* dx2 is the position in that block (dst) */
	      dx2rgb = dx * 3;
	      for (dx2 = 0; dx2 < n; dx2++)
		{
		  for (neededEnd = data + nsrc[dx2]; data < neededEnd; data++)
		    {
		      buf[dx2rgb] += data->r;
		      buf[dx2rgb + 1] += data->g;
		      buf[dx2rgb + 2] += data->b;
		    }
		  dx2rgb += 3;
		}
				
	    }

	  /* We're out of full blocks, now for some more fun;
	   * we want to accumulate only part of the last block
	   * this covers 2 cases: 
	   *	(1) we don't actually want all of the last block
	   *		and there exists enough source
	   *	(2) there's not enough source (the srcImg width 
	   *		need not be a multiple of m), and we want to
	   *		generate the right edge
	   */
	  dx = lastStart;
	  dx2rgb = dx * 3;
	  for (dx2 = 0; dx2 < lastSize; dx2++)
	    {
	      for (neededEnd = data + nsrc[dx2]; data < neededEnd; data++)
		{
		  buf[dx2rgb] += data->r;
		  buf[dx2rgb + 1] += data->g;
		  buf[dx2rgb + 2] += data->b;
		}

	      dx2rgb += 3;
	    }
	  srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
	}

      /* Do the mapping from buf to dest pixels */
      dstMap = dstRow;
				
      /* once again, we need to split this up into 3 blocks
       * (1) partial first block
       * (2) full blocks
       * (3) partial last block
       */
      dx = 0;
      dx2rgb = firstStart * 3;
      for (dx2 = firstStart ; dx2 < firstEnd ; dx2++)
	{
	  blockPix = nsrc[dy2] * nsrc[dx2];
	  dstMap->r = (Dgp_UInt8) (buf[dx2rgb] / blockPix);
	  dstMap->g = (Dgp_UInt8) (buf[dx2rgb + 1] / blockPix);
	  dstMap->b = (Dgp_UInt8) (buf[dx2rgb + 2] / blockPix);
	  dstMap ++;
	  dx2rgb += 3;
	}

      for (dx = firstEnd ; dx < lastStart ; dx += n) 
	{
	  dx2rgb = dx * 3;
	  for (dx2 = 0 ; dx2 < n; dx2++)
	    {
	      blockPix = nsrc[dy2] * nsrc[dx2];
	      dstMap->r = (Dgp_UInt8) (buf[dx2rgb] / blockPix);
	      dstMap->g = (Dgp_UInt8) (buf[dx2rgb + 1] / blockPix);
	      dstMap->b = (Dgp_UInt8) (buf[dx2rgb + 2] / blockPix);
	      dstMap ++;
	      dx2rgb += 3;
	    }
	}

      dx = lastStart;
      dx2rgb = dx * 3;
      for (dx2 = 0 ; dx2 < lastSize ; dx2++)
	{
	  blockPix = nsrc[dy2] * nsrc[dx2];
	  dstMap->r = (Dgp_UInt8) (buf[dx2rgb] / blockPix);
	  dstMap->g = (Dgp_UInt8) (buf[dx2rgb + 1] / blockPix);
	  dstMap->b = (Dgp_UInt8) (buf[dx2rgb + 2] / blockPix);
	  dstMap ++;
	  dx2rgb += 3;
	}

		
      dstRow = (DgpBlat_BGR *)((char *)dstRow + dstBpl);
    }



}



static Dgp_Int32 DgpBlat_scaleRGBBGRAvg(DgpImage_RGB * src, Dgp_Int32 srcBpl,
		   Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		   DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		   Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		   Dgp_Int32 m, Dgp_Int32 n, Dgp_Allocators * a)
{
  /*
   * The source image can be considered to be tiled by squares
   * of m*m pixels, each of which maps to an n*n square in the
   * destination.  As we walk across each square, different numbers
   * of source pixels are averaged to produce each destination pixel.
   * For example, if m==3, n==2, the 3x3 square in the source maps
   * to the dest as follows:
   * 1 1 2
   * 1 1 2
   * 3 3 4
   * where
   * 1 2
   * 3 4
   * is the dest 2x2 square.  We need to compute nsrc[0...n-1] so that
   * nsrc[i] is the number of source pixels going into dest pixel i,
   * both horizontally and vertically (dest pixel (i,j) came from
   * (nsrc[i], nsrc[j]) source pixels).
   */

  Dgp_Int32 i,y;
  Dgp_Int32 * nsrc = NULL;
  Dgp_Int32 dstStartX, dstStartY;
  Dgp_Int32 srcStartX, srcStartY;
  Dgp_Int32 bufLen;
  Dgp_Int32 * buf = (Dgp_Int32 *)NULL;
  DgpImage_RGB * srcRow;
  DgpBlat_BGR * dstRow;
  Dgp_Int32 startBlockRow,endBlockRow;
  Dgp_Int32 firstStart,firstEnd;
  Dgp_Int32 lastEnd;

  nsrc = (Dgp_Int32 *)a->malloc(n*sizeof(int));
  if(!nsrc) goto bailout;

  for (i = 0; i < n; i++) {
    nsrc[i] = (m * (i + 1) + n / 2) / n - (m * i + n / 2) / n;
  }	

  dstStartX = dstX - dstX % n;
  dstStartY = dstY - dstY % n;
	
  srcStartX = (dstStartX / n)*m;
  srcStartY = (dstStartY / n)*m;

  /* A buffer to hold the accumulation of pixels */
  bufLen=(int)((ceil((float)(dstX + dstW) / n) * n) - dstStartX)*sizeof(DgpImage_RGB);
  buf = (Dgp_Int32 *)a->malloc(bufLen*sizeof(int));
  if(!buf) 
    goto bailout;

	/* we have 3 loops here
	 * (1) to handle needing only a partial top dst nxn block row (clipped both)
	 * (2) general case blocked rows
	 * (3) handle a partial bottom row (clipped at the bottom)
	 */

  srcRow = (DgpImage_RGB *)((char *)src + ((srcStartY - srcMinY) * srcBpl));
  srcRow += (srcStartX - srcMinX);
		
  dstRow = (DgpBlat_BGR *)((char *)dst + (dstY - dstMinY) * dstBpl);
  dstRow += (dstX - dstMinX);

  startBlockRow = dstY - dstStartY;
  endBlockRow = MIN(n, startBlockRow + dstH);

  firstStart = dstX - dstStartX;
  firstEnd = MIN(n, firstStart + dstW);

  lastEnd = firstStart + dstW;
	
  DgpBlat_scaleRGBBGRAvgBlockRow(srcRow, srcBpl, 
			 dstRow, dstBpl, n, 
			 startBlockRow, endBlockRow,
			 firstStart, firstEnd, lastEnd,
			 nsrc, buf, bufLen);
  srcRow = (DgpImage_RGB *)((char *)srcRow + (m * srcBpl));
  dstRow = (DgpBlat_BGR *)((char *)dstRow + (endBlockRow - startBlockRow) * dstBpl);

  for ( y = endBlockRow-startBlockRow ; y+n <= dstH ; y += n)
    {
      DgpBlat_scaleRGBBGRAvgBlockRow(srcRow, srcBpl, 
			     dstRow, dstBpl, n, 
			     0, n, firstStart, firstEnd, lastEnd,
			     nsrc, buf, bufLen);

      srcRow = (DgpImage_RGB *)((char *)srcRow + (m * srcBpl));
      dstRow = (DgpBlat_BGR *)((char *)dstRow + (n * dstBpl));
    }

  if ( y < dstH)
    {
      DgpBlat_scaleRGBBGRAvgBlockRow(srcRow, srcBpl, 
			     dstRow, dstBpl, n, 
			     0,  dstH - y , firstStart, firstEnd, lastEnd,
			     nsrc, buf, bufLen);
    }

  a->free(nsrc);
  a->free(buf);


  return 1;

 bailout:
  if (nsrc)
    {
      a->free(nsrc);
    }
  if (buf)
    {
      a->free(buf);
    }

  return 0;
}


Dgp_Int32 DgpBlat_scaleRGBBGRAvgSub(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		      DgpCoord_Resolution * res, Dgp_Allocators * a)
{
  /* We're going to do a generic scaling down by scaling down 
     by 2,3,4,5,or 8 (simultaneous in x and y) and then doing
     sub-sampling from there. But we're going to do this all
     at once, deciding which line buffers of 2,3,4,5,or 8 to geneDGPCOORD_Rate
     and then sub-sampling those. So first, should be scale down
     by 2,3,4,5,or 8?
	 */
	
  /* If there's no averaging to be done (scales <=1.0  >.5) we'll just
     call the straight subsampling function.
  */
  if(DgpCoord_equal(res->xRes, DgpCoord_eq(1,1)) && DgpCoord_equal(res->yRes, DgpCoord_eq(1,1)))
    {
      DgpBlat_blitRGBBGR(src, srcBpl, srcMinX, srcMinY,
		 dst, dstBpl, dstMinX, dstMinY,
		 dstX, dstY, dstW, dstH);
      return 1;
    }
  /*
	if(scaleX>.5f || scaleY>.5f)
		{		
		scaleBlitRGBBGR(src,srcBpl,srcMinX,srcMinY,
						dst,dstBpl,dstMinX,dstMinY,
						dstX,dstY,dstW,dstH,
						scaleX,scaleY);
		return 1;
		}
*/		
	/* If this is just averaging, we'll call the special case functions
	   that just do averaging of 2,3,4,5,8
	 */
  if(DgpCoord_equal(res->xRes, res->yRes))
    {
      if (DgpCoord_equal(res->xRes, DgpCoord_eq(1, 2))){
	return DgpBlat_scaleRGBBGRAvgby2(src, srcBpl, srcMinX, srcMinY,
				 dst, dstBpl, dstMinX, dstMinY,
				 dstX, dstY, dstW, dstH, a);
      }
      if (DgpCoord_equal(res->xRes, DgpCoord_eq(1, 3))){
	return DgpBlat_scaleRGBBGRAvgby3(src, srcBpl, srcMinX, srcMinY,
				 dst, dstBpl, dstMinX, dstMinY,
				 dstX, dstY, dstW, dstH, a);
      }
      if (DgpCoord_equal(res->xRes, DgpCoord_eq(1, 4))){
	return DgpBlat_scaleRGBBGRAvgby4(src, srcBpl, srcMinX, srcMinY,
				 dst, dstBpl, dstMinX, dstMinY,
				 dstX, dstY, dstW, dstH, a);
      }
      if (DgpCoord_equal(res->xRes, DgpCoord_eq(1, 5))){
	return DgpBlat_scaleRGBBGRAvgby5(src, srcBpl, srcMinX, srcMinY,
				 dst, dstBpl, dstMinX, dstMinY,
				 dstX, dstY, dstW, dstH, a);
      }
      if (DgpCoord_equal(res->xRes, DgpCoord_eq(1, 8))){
	return DgpBlat_scaleRGBBGRAvgby8(src, srcBpl, srcMinX, srcMinY,
				 dst, dstBpl, dstMinX, dstMinY,
				 dstX, dstY, dstW, dstH, a);
      }
      else{
	return DgpBlat_scaleRGBBGRAvg(src, srcBpl, srcMinX, srcMinY,
			      dst, dstBpl, dstMinX, dstMinY,
			      dstX, dstY, dstW, dstH,
			      res->xRes.den,res->xRes.num, a);
      }
    }
  
  /* So since none of the special cases work, we'll just find the
     the larger average and scale downward.
  */
  return DgpBlat_scaleRGBBGRAvgSubGen(src, srcBpl, srcMinX, srcMinY,
			      dst, dstBpl, dstMinX, dstMinY,
			      dstX, dstY, dstW, dstH, res, a);
  
}





INLINE void DgpBlat_scaleRGBGRAvgby2Row(DgpImage_RGB * srcRow, Dgp_Int32 srcBpl, Dgp_Int32 * avgBuf, Dgp_Int32 rgbDstW)
{
  DgpImage_RGB * srcCur = srcRow;
  Dgp_Int32 * avgCur;
  Dgp_Int32 * avgEnd = avgBuf + rgbDstW;
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] = srcCur->r;	
      avgCur[1] = srcCur->g;
      avgCur[2] = srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;	
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
	
  srcCur = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;	
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;	
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
}

INLINE void DgpBlat_scaleRGBGRAvgby3Row(DgpImage_RGB * srcRow, Dgp_Int32 srcBpl, Dgp_Int32 * avgBuf, Dgp_Int32 rgbDstW)
{
  DgpImage_RGB * srcCur = srcRow;
  Dgp_Int32 * avgCur;
  Dgp_Int32 * avgEnd = avgBuf + rgbDstW;
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] = srcCur->r;
      avgCur[1] = srcCur->g;
      avgCur[2] = srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* These 2 the same */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
}


INLINE void DgpBlat_scaleRGBGRAvgby4Row(DgpImage_RGB * srcRow, Dgp_Int32 srcBpl, Dgp_Int32 * avgBuf, Dgp_Int32 rgbDstW)
{
  DgpImage_RGB * srcCur = srcRow;
  Dgp_Int32 * avgCur;
  Dgp_Int32 * avgEnd = avgBuf + rgbDstW;
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] = srcCur->r;
      avgCur[1] = srcCur->g;
      avgCur[2] = srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* These 3 are the same */		
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
}


INLINE void DgpBlat_scaleRGBGRAvgby5Row(DgpImage_RGB * srcRow, Dgp_Int32 srcBpl, Dgp_Int32 * avgBuf, Dgp_Int32 rgbDstW)
{
  DgpImage_RGB * srcCur = srcRow;
  Dgp_Int32 * avgCur;
  Dgp_Int32 * avgEnd = avgBuf + rgbDstW;
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] = srcCur->r;
      avgCur[1] = srcCur->g;
      avgCur[2] = srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* These 4 are the same */		
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
}

INLINE void DgpBlat_scaleRGBGRAvgby8Row(DgpImage_RGB * srcRow, Dgp_Int32 srcBpl, Dgp_Int32 * avgBuf, Dgp_Int32 rgbDstW)
{
  DgpImage_RGB * srcCur = srcRow;
  Dgp_Int32 * avgCur;
  Dgp_Int32 * avgEnd = avgBuf + rgbDstW;
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] = srcCur->r;
      avgCur[1] = srcCur->g;
      avgCur[2] = srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* These 7 are the same */		
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
  srcCur = srcRow = (DgpImage_RGB *)((char *)srcRow + srcBpl);
  for (avgCur = avgBuf; avgCur < avgEnd; avgCur += 3)
    {
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
		
      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;

      avgCur[0] += srcCur->r;
      avgCur[1] += srcCur->g;
      avgCur[2] += srcCur->b;
      srcCur++;
    }
  /* */
}

