#ifndef __SCALEBLITAVGCOLOR_H__
#define __SCALEBLITAVGCOLOR_H__

#include "resolution_t.h"

#ifdef __cplusplus
extern "C"
{
#endif

#include "scaleBlit.h"

/* Here we have functions for down scaling RGB to BGR by averaging, RGB to RGB
	versions can easily be made out of these should there be need. 
 */

/* In all cases, the idea is this function generates a specific portion
   of a destination scaled image [(dstX,dstY) to (dstX+dstW-1,dstY+dstH-1)]
   given the appropriate portion of the source image. 
   src is a pointer to a point in the source image
   srcBpl is the bytes per line for the source image
   srcMinX,srcMinY is the point in the source image that src corresponds to
   dst is a pointer to a point in the dest image.
   dstBpl is the bytes per line for the dest image
   dstMinX,dstMinY is the point in the dest image that dst corresponds to
   dstX,dstY,dstW,dstH is a rectangle in the dest image that you wish 
   to generate.
   scaleX and scaleY are the x and y scale factors. Note, this only does
   downsampling, so both must be <=1.0
 */

/* Calls malloc / free for a temporary line buffer 
	returns 1 if successful
			0 if not (the memory allocation failed)
 */

/* scaleRGBBGRAvgSub is the normal entry point; it case dispatches the 
	scale factors and calls the special case versions or the general 
	case
 */
extern Dgp_Int32 
DgpBlat_scaleRGBBGRAvgSub(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		      DgpCoord_Resolution * res, Dgp_Allocators * a);
  
/* We've got averaging for scale factors of 1/2,1/3,1/4,1/5,1/8. These functions
     all come from the same code, so changes / optimizations to one 
	should be applied to the rest. 
 */
extern Dgp_Int32 
DgpBlat_scaleRGBBGRAvgby2(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

extern Dgp_Int32 
DgpBlat_scaleRGBBGRAvgby3(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

extern Dgp_Int32 
DgpBlat_scaleRGBBGRAvgby4(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

extern Dgp_Int32 
DgpBlat_scaleRGBBGRAvgby5(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

extern Dgp_Int32 
DgpBlat_scaleRGBBGRAvgby8(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

/* If you want some scale factor (or scaleX and scaleY are different),
	this general purpose function will find the closest (but above) one
	of the average factors above, and then scale down. It does this in
	one pass, not by doing the full average and then subsampling.
 */
extern Dgp_Int32 
DgpBlat_scaleRGBBGRAvgSubGen(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			 DgpBlat_BGR * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			 Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
			 DgpCoord_Resolution * res, Dgp_Allocators * a);


/*/////////////////////////////////// */
/* Same functionality, RGB->RGB... */
extern Dgp_Int32 
DgpBlat_scaleRGBRGBAvgSub(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		      DgpCoord_Resolution * res, Dgp_Allocators * a);
  
extern Dgp_Int32 
DgpBlat_scaleRGBRGBAvgby2(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

extern Dgp_Int32 
DgpBlat_scaleRGBRGBAvgby3(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

extern Dgp_Int32 
DgpBlat_scaleRGBRGBAvgby4(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

extern Dgp_Int32 
DgpBlat_scaleRGBRGBAvgby5(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

extern Dgp_Int32 
DgpBlat_scaleRGBRGBAvgby8(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

extern Dgp_Int32 
DgpBlat_scaleRGBRGBAvgSubGen(DgpImage_RGB * src, Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			 DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			 Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
			 DgpCoord_Resolution * res, Dgp_Allocators * a);


#ifdef __cplusplus
}
#endif

#endif /*__SCALEBLITAVGCOLOR_H__ */
