#include "scaleBlitAvgMono.h"
#include "blitTables.h"
#include "resolution_t.h"

#include "assert.h"
#include <stdlib.h>

#if !defined(VXWORKS) && !defined(EQUATOR_CA_ENV)
#include <malloc.h>
#include <memory.h>
#endif  /* (!VXWORKS) */

#include <math.h>

/* ...for MIN/MAX */
/* inline definition for different compilers */
#include "util.h"

#define MAKE_RGB(x)	{ (char)(x), (char)(x), (char)(x) }

/* the range of intensities for the different averaging factors */
static const DgpImage_RGB scaled2[] = 
{MAKE_RGB(0xff), MAKE_RGB(0xc0), MAKE_RGB(0x80),MAKE_RGB(0x40),
 MAKE_RGB(0x00)};

static const DgpImage_RGB scaled4[] =
{MAKE_RGB(0xff), MAKE_RGB(0xef), MAKE_RGB(0xdf), MAKE_RGB(0xcf), 
 MAKE_RGB(0xbf), MAKE_RGB(0xaf), MAKE_RGB(0x9f), MAKE_RGB(0x8f),
 MAKE_RGB(0x7f), MAKE_RGB(0x70), MAKE_RGB(0x60), MAKE_RGB(0x50), 
 MAKE_RGB(0x40), MAKE_RGB(0x30), MAKE_RGB(0x20), MAKE_RGB(0x10), 
 MAKE_RGB(0x00)}; 

static const DgpImage_RGB scaled8[] = 
{MAKE_RGB(0xff), MAKE_RGB(0xfc), MAKE_RGB(0xf8), MAKE_RGB(0xf4), 
 MAKE_RGB(0xf0), MAKE_RGB(0xec), MAKE_RGB(0xe8), MAKE_RGB(0xe4), 
 MAKE_RGB(0xe0), MAKE_RGB(0xdc), MAKE_RGB(0xd8), MAKE_RGB(0xd4), 
 MAKE_RGB(0xd0), MAKE_RGB(0xcc), MAKE_RGB(0xc8), MAKE_RGB(0xc4), 
 MAKE_RGB(0xc0), MAKE_RGB(0xbc), MAKE_RGB(0xb8), MAKE_RGB(0xb4), 
 MAKE_RGB(0xb0), MAKE_RGB(0xac), MAKE_RGB(0xa8), MAKE_RGB(0xa4), 
 MAKE_RGB(0xa0), MAKE_RGB(0x9c), MAKE_RGB(0x98), MAKE_RGB(0x94), 
 MAKE_RGB(0x90), MAKE_RGB(0x8c), MAKE_RGB(0x88), MAKE_RGB(0x84), 
 MAKE_RGB(0x80), MAKE_RGB(0x7c), MAKE_RGB(0x78), MAKE_RGB(0x74), 
 MAKE_RGB(0x70), MAKE_RGB(0x6c), MAKE_RGB(0x68), MAKE_RGB(0x64), 
 MAKE_RGB(0x60), MAKE_RGB(0x5c), MAKE_RGB(0x58), MAKE_RGB(0x54), 
 MAKE_RGB(0x50), MAKE_RGB(0x4c), MAKE_RGB(0x48), MAKE_RGB(0x44), 
 MAKE_RGB(0x40), MAKE_RGB(0x3c), MAKE_RGB(0x38), MAKE_RGB(0x34), 
 MAKE_RGB(0x30), MAKE_RGB(0x2c), MAKE_RGB(0x28), MAKE_RGB(0x24), 
 MAKE_RGB(0x20), MAKE_RGB(0x1c), MAKE_RGB(0x18), MAKE_RGB(0x14), 
 MAKE_RGB(0x10), MAKE_RGB(0x0c), MAKE_RGB(0x08), MAKE_RGB(0x04), 
 MAKE_RGB(0x00) }; 

#undef MAKE_RGB

static INLINE void scaleBinaryRGBAvgby2Row(Dgp_UInt32 * srcRow, 
				    Dgp_Int32 srcLpl, Dgp_Int32 srcW, 
				    unsigned char * grayBuf);
static INLINE void scaleBinaryRGBAvgby4Row(Dgp_UInt32 * srcRow, 
				    Dgp_Int32 srcLpl, Dgp_Int32 srcW, 
				    unsigned char * grayBuf);
static INLINE void scaleBinaryRGBAvgby8Row(Dgp_UInt32 * srcRow, 
				    Dgp_Int32 srcLpl, Dgp_Int32 srcW, 
				    unsigned char * grayBuf);


/* Note: the binary to gray stuff is pretty much carbon copied for
	the three versions. So any bug fixes / optimizations should be
	applied to them all.
 */

/* The more general stuff is very similar to that in scaleBlitAvgColor
	apply changes there as well 
 */

Dgp_Int32 DgpBlat_scaleBinaryRGBAvgby2(Dgp_UInt32 * src,
			 Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			 DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			 Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a)
{
  /* Given a destination BGR image and a source packed binary image
   * this routine will generate arbitrary portions of the dest image
   * at a scale factor of 1/2 in gray.
   * This takes in a destination BGR buffer where dst is the 
   * point (dstMinX, dstMinY) and takes a source image where 
   * src is the point (srcMinX,srcMinY) and draws the portion
   * (dstX,dstY) (dstX+dstW-1,dstY+dstH-1) by scaling to gray
   * by 2 in x and y
   */

  char * grayBuf;
  Dgp_UInt32 * srcRow;
	
  DgpImage_RGB * dstRow;
  Dgp_Int32 ii = 0;
  Dgp_Int32 x,y;
  Dgp_Int32 srcLpl;
  Dgp_Int32 dstMin;
	
  Dgp_Int32 srcX, srcY, srcW;
	
	/* since we're not guaranteed of starting or ending on word boundaries, */
	/* we're going to actually generate the intensities for the full first */
	/* word and the last word and only use those needed for the dest. */

	/* we need a temporary buffer for this opeDGPCOORD_Ration, we'll find the */
	/* intensities in an 8-bit buffer and then scale the intensities */
	/* into the RGB output. */
	
	/* what is the first pixel of src needed?  */
  srcX = dstX*2;
  /* what is the 32-bit boundary of src before srcMin? */
  srcX = (srcX - (srcX % 32));
  assert((srcMinX % 32)==0);

	/* the width in the source has to add any extra pixels we might */
	/* generate because we want a 32-bit boundary on the left */
  srcW = (dstW * 2) + ((dstX*2) - srcX);
  assert((srcW % 2)==0);

	/*srcX = (int)((srcX - srcMinX)/32); */
	/*srcX *= 32; */
  srcY = dstY * 2;
  dstMin = (int)(srcX / 2);

	
  srcLpl=srcBpl/sizeof(Dgp_UInt32);
	
  /* we know we can need at most 32 (if we have to generate the complete words before and after) */
  /* if the source is malaligned. */
  grayBuf=(char *)a->malloc((((dstW)+32)*sizeof(char)));
  if(!grayBuf)
    return 0;

	/* the 8 is for 8 bits per pixel. */
  srcRow = src + (((srcY-srcMinY) * srcLpl) + (int)((srcX - srcMinX)/ 32));
  dstRow = (DgpImage_RGB *)((char *)dst + (((dstY-dstMinY) * dstBpl) + ((dstX-dstMinX) * sizeof(DgpImage_RGB))));

  for (y = 0 ; y < dstH; y ++ )
    {
      scaleBinaryRGBAvgby2Row(srcRow, srcLpl, srcW, grayBuf);

      for( x=0, ii=(dstX-dstMin); x < dstW; x++, ii++)
	{
	  dstRow[x]=scaled2[(int)grayBuf[ii]];
	}
		
      srcRow = srcRow + (2*srcLpl);
      dstRow = (DgpImage_RGB *)((char *)dstRow + dstBpl);
    }

  a->free(grayBuf);
  return 1;
}
						



Dgp_Int32 DgpBlat_scaleBinaryRGBAvgby4(Dgp_UInt32 * src,
			 Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			 DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			 Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a)
{
  /* Given a destination BGR image and a source packed binary image
   * this routine will generate arbitrary portions of the dest image
   * at a scale factor of 1/4 in gray.
   * This takes in a destination RGB buffer where dst is the 
   * point (dstMinX, dstMinY) and takes a source image where 
   * src is the point (srcMinX,srcMinY) and draws the portion
   * (dstX,dstY) (dstX+dstW-1,dstY+dstH-1) by scaling to gray
   * by 4 in x and y
   */

  char * grayBuf;
  Dgp_UInt32 * srcRow;
	
  DgpImage_RGB * dstRow;
  Dgp_Int32 ii = 0;
  Dgp_Int32 x,y;
  Dgp_Int32 srcLpl;
  Dgp_Int32 dstMin;
	
  Dgp_Int32 srcX, srcY, srcW;
	
	/* since we're not guaranteed of starting or ending on word boundaries, */
	/* we're going to actually generate the intensities for the full first */
	/* word and the last word and only use those needed for the dest. */

	/* we need a temporary buffer for this opeDGPCOORD_Ration, we'll find the */
	/* intensities in an 8-bit buffer and then scale the intensities */
	/* into the RGB output. */
	
	/* what is the first pixel of src needed?  */
  srcX = dstX*4;
  /* what is the 32-bit boundary of src before srcMin? */
  srcX = (srcX - (srcX % 32));
  assert((srcMinX % 32)==0);

	/* the width in the source has to add any extra pixels we might */
	/* generate because we want a 32-bit boundary on the left */
  srcW = (dstW * 4) + ((dstX*4) - srcX);
  assert((srcW % 4)==0);

  srcY = dstY * 4;
  dstMin = (int)(srcX / 4);

	
  srcLpl=srcBpl/sizeof(Dgp_UInt32);
	
  /* we know we can need at most 16 (if we have to generate the complete words before and after) */
  /* if the source is malaligned. */
  grayBuf=(char *)a->malloc((((dstW)+16)*sizeof(char)));
  if(!grayBuf)
    return 0;

	/* the 8 is for 8 bits per pixel. */
  srcRow = src + (((srcY-srcMinY) * srcLpl) + (int)((srcX - srcMinX)/ 32));
  dstRow = (DgpImage_RGB *)((char *)dst + (((dstY-dstMinY) * dstBpl) + ((dstX-dstMinX) * sizeof(DgpImage_RGB))));

  for (y = 0 ; y < dstH; y ++ )
    {

      scaleBinaryRGBAvgby4Row(srcRow, srcLpl, srcW, grayBuf);
		
      for( x=0, ii=(dstX-dstMin); x < dstW; x++, ii++)
	{
	  dstRow[x]=scaled4[(int)grayBuf[ii]];
	}
		
      srcRow = srcRow + (4*srcLpl);
      dstRow = (DgpImage_RGB *)((char *)dstRow + dstBpl);
    }

  a->free(grayBuf);
  return 1;
}
						

Dgp_Int32 DgpBlat_scaleBinaryRGBAvgby8(Dgp_UInt32 * src,
			 Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			 DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			 Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a)
{
  /* Given a destination BGR image and a source packed binary image
   * this routine will generate arbitrary portions of the dest image
   * at a scale factor of 1/8 in gray.
   * This takes in a destination RGB buffer where dst is the 
   * point (dstMinX, dstMinY) and takes a source image where 
   * src is the point (srcMinX,srcMinY) and draws the portion
   * (dstX,dstY) (dstX+dstW-1,dstY+dstH-1) by scaling to gray
   * by 8 in x and y
   */

  char * grayBuf;
  Dgp_UInt32 * srcRow;
	
  DgpImage_RGB * dstRow;
  Dgp_Int32 ii = 0;
  Dgp_Int32 x,y;
  Dgp_Int32 srcLpl;
  Dgp_Int32 dstMin;
	
  Dgp_Int32 srcX, srcY, srcW;
	

	/* since we're not guaranteed of starting or ending on word boundaries, */
	/* we're going to actually generate the intensities for the full first */
	/* word and the last word and only use those needed for the dest. */

	/* we need a temporary buffer for this opeDGPCOORD_Ration, we'll find the */
	/* intensities in an 8-bit buffer and then scale the intensities */
	/* into the RGB output. */
	
	/* what is the first pixel of src needed?  */
  srcX = dstX*8;
  /* what is the 32-bit boundary of src before srcMin? */
  srcX = (srcX - (srcX % 32));
  assert((srcMinX % 32)==0);

	/* the width in the source has to add any extra pixels we might */
	/* generate because we want a 32-bit boundary on the left */
  srcW = (dstW * 8) + ((dstX*8) - srcX);
  assert((srcW % 8)==0);

  srcY = dstY * 8;
  dstMin = (int)(srcX / 8);

	
  srcLpl=srcBpl/sizeof(Dgp_UInt32);
	
  /* we know we can need at most 8 (if we have to generate the complete words before and after) */
  /* if the source is malaligned. */
  grayBuf=(char *)a->malloc((((dstW)+8)*sizeof(char)));
  if(!grayBuf)
    return 0;

	/* the 8 is for 8 bits per pixel. */
  srcRow = src + (((srcY-srcMinY) * srcLpl) + (int)((srcX - srcMinX)/ 32));
  dstRow = (DgpImage_RGB *)((char *)dst + (((dstY-dstMinY) * dstBpl) + ((dstX-dstMinX) * sizeof(DgpImage_RGB))));
  for (y = 0 ; y < dstH; y ++ )
    {
      scaleBinaryRGBAvgby8Row(srcRow, srcLpl, srcW, grayBuf);

      for( x=0, ii=(dstX-dstMin); x < dstW; x++, ii++)
	{
	  dstRow[x]=scaled8[(int)grayBuf[ii]];
	}
		
      srcRow = srcRow + (8 *srcLpl);
      dstRow = (DgpImage_RGB *)((char *)dstRow + dstBpl);
    }

  a->free(grayBuf);
  return 1;
}


Dgp_Int32 DgpBlat_scaleBinaryRGBAvgSubGen(Dgp_UInt32 * src, Dgp_Int32 srcBpl,
			    Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			    DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			    Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
			    DgpCoord_Resolution * res, Dgp_Allocators * a)
{
  /* We need to find the best averaging and then scale down a little from
     there, note that we have to use the max(scaleX,scaleY) as to determine
     which to use 
  */
  DgpCoord_Rat maxScale;
  Dgp_Int32 avgFactor = 0;

  Dgp_Int32 x,y;

  Dgp_UInt32 *srcRow = src;
  /* srcOff is just a junk var used to compute offsets for the src. */
  Dgp_Int32 srcOff,srcLpl;

  double scaleRatioX,invScaleY;
  Dgp_Int32 avgFacSq;

  DgpImage_RGB *dstRow;
  char * grayBuf;

  Dgp_Int32 dstMin, dstEnd;
	
  Dgp_Int32 srcX, srcY, srcW;
	
  /* dstRow is the first pixel in the first row to be blitted. */
  /* get the top row. */
  Dgp_Int32 dstOff=((dstY-dstMinY) * dstBpl);
  /* advance the x position. */
  dstOff += ((dstX-dstMinX)*sizeof(DgpImage_RGB));
  
  dstRow = (DgpImage_RGB *)((char *)dst + dstOff);
  
  srcLpl = srcBpl/sizeof(Dgp_UInt32);

  if (DgpCoord_g(res->xRes, res->yRes)){
    maxScale = res->xRes;
  }
  else {
    maxScale = res->yRes;
  }
  /* You shouldn't be calling this function if there's no averaging that 
     needs to be done.
  */
  assert(DgpCoord_le(maxScale,DgpCoord_eq(1,2)));


  if (DgpCoord_le(maxScale, DgpCoord_eq(1, 2))){
    avgFactor = 2;
  }
  if (DgpCoord_le(maxScale, DgpCoord_eq(1, 4))){
    avgFactor = 4;
  }
  if (DgpCoord_le(maxScale, DgpCoord_eq(1, 8))){
    avgFactor = 8;
  }

  avgFacSq = avgFactor * avgFactor;

  scaleRatioX=((double)1/avgFactor)/ DgpCoord_float(res->xRes);
  invScaleY = (1 / DgpCoord_float(res->yRes));

  /* since we're not guaranteed of starting or ending on word boundaries, */
  /* we're going to actually generate the intensities for the full first */
  /* word and the last word and only use those needed for the dest. */

  /* we need a temporary buffer for this opeDGPCOORD_Ration, we'll find the */
  /* intensities in an 8-bit buffer and then scale the intensities */
  /* into the RGB output. */
  
  /* what is the first pixel of src needed?  */
  srcX = (int)(dstX * (1 / DgpCoord_float(res->xRes)));
  /* what is the 32-bit boundary of src before srcMin? */
  srcX = (int)(srcX - (srcX % 32));
	
  srcY = (int)(dstY * (1 / DgpCoord_float(res->yRes)));

  dstMin = (int)(srcX * (1.0 / avgFactor));
	

  srcW = ((int)((dstW-1) * (1 / DgpCoord_float(res->xRes)))+1)
    +((int)(dstX * (1 / DgpCoord_float(res->xRes)))-srcX);
  srcW += (avgFactor - (srcW % avgFactor));
  assert((srcW % avgFactor) == 0);
  srcLpl=srcBpl/sizeof(Dgp_UInt32);

  grayBuf=(char *)a->malloc((((int)(srcW * 
				    ((float)1/avgFactor)+32)*sizeof(char))));
  if(!grayBuf)
    return 0;

  dstEnd = dstW+dstX;
  for (y=0; y<dstH; y++)
    {
      srcOff = (int)((dstY + y) * invScaleY) - srcMinY;
      srcOff *= srcLpl;
      srcOff += (int)((srcX - srcMinX) / 32);
      srcRow = src + srcOff;

      if (avgFactor == 2)
	{
	  scaleBinaryRGBAvgby2Row(srcRow, srcLpl, srcW, grayBuf);
	  for(x=dstX; x < dstEnd; x++)
	    dstRow[x-dstX]=scaled2[(int)grayBuf[(int)(x*scaleRatioX)-dstMin]];
	}
      else if (avgFactor == 4)
	{
	  scaleBinaryRGBAvgby4Row(srcRow, srcLpl, srcW, grayBuf);
	  for(x=dstX; x < dstEnd; x++)
	    dstRow[x-dstX]=scaled4[(int)grayBuf[(int)(x*scaleRatioX)-dstMin]];
	}
      else /* avgFactor == 8 */
	{
	  scaleBinaryRGBAvgby8Row(srcRow, srcLpl, srcW, grayBuf);
	  for(x=dstX; x < dstEnd; x++)
	    dstRow[x-dstX]=scaled8[(int)grayBuf[(int)(x*scaleRatioX)-dstMin]];
	}
	
		
      dstRow = (DgpImage_RGB *)((char *)dstRow + dstBpl);
    }
  a->free(grayBuf);
  return 1;
}



/* srcRow - ptr to upper left 32-bit boundary of first src mxm block to be scaled.
 * srcBitsLeft - how many bits are to the right of the bit the first bit of the first
 *		src mxm block (including the first bit)
 * srcLpl -	longs per line of the source.
 * dstRow - ptr to upper left corner of first destination pixel in upper left nxn block 
 * dstBpl - bytes per line in dst.
 * n	  - size of dst scaled block
 * startBlockRow - what row of the block (0..n-1) goes into dstRow
 * endBlockRow - what row of the block (startBlockRow..n) is the last row of the block
 *		[note exclusive]
 * firstStart - what column of the first block goes into dstRow (0..n-1)
 * firstEnd - what is the last column of the first block [note exclusive]
 *	 (normally n, but you could be blatting a really small width < n)
 *  lastEnd - what is the last pixel of the src row that should be accumulated
 *		(normally dstX+dstW, but could be different if srcImgWidth % m != 0)
 */
void DgpBlat_scaleBinaryRGBAvgBlockRow(Dgp_UInt32 * srcRow, Dgp_Int32 srcBitsLeft, Dgp_Int32 srcLpl, 
			       DgpImage_RGB * dstRow, Dgp_Int32 dstBpl, Dgp_Int32 n, 
			       Dgp_Int32 startBlockRow, Dgp_Int32 endBlockRow,
			       Dgp_Int32 firstStart, Dgp_Int32 firstEnd, Dgp_Int32 lastEnd,
			       Dgp_Int32 * nsrc, Dgp_Int32 * buf, Dgp_Int32 bufLen, unsigned char ** lookups)
{
  Dgp_Int32 y2, y3;
  Dgp_Int32 dy2;
  Dgp_Int32 dx, dx2;

  Dgp_UInt32 * data;
  DgpImage_RGB * dstMap;

  Dgp_UInt32 ourbits;
  Dgp_Int32 bitsneeded;
  Dgp_UInt32 curbits;
  Dgp_Int32 bitsleft;

  Dgp_Int32 lastSize;
  Dgp_Int32 lastStart;
	
  /* when does the full block stuff end? */
  /* so normally, it's just lastEnd % n */
  /* but if it's a really short blat (only does a partial first block) */
  /* we need to make sure the other loops don't run, so we set it to lastEnd */
  lastStart = lastEnd;
  if (firstEnd == n)
    lastStart = lastEnd - (lastEnd % n);
	
  /* precompute the size of the last block */
  lastSize = lastEnd - lastStart;

  /* offset the source row the number of rows it would have moved */
  /* if we were doing the block from the beginning. */
  for(y3 = 0;  y3 < startBlockRow; y3++)
    srcRow = srcRow + (nsrc[y3] * srcLpl);

  /* same for the x */

  bitsneeded=0;
  for (dx2 = 0; dx2 < firstStart ; dx2++)
    bitsneeded += nsrc[dx2];
	 
  if ((srcBitsLeft - bitsneeded) < 0)
    {
      srcRow += (int)ceil((float)(bitsneeded - srcBitsLeft) / 32);
      srcBitsLeft += (int)ceil((float)(bitsneeded - srcBitsLeft) / 32)*32;
    }

  srcBitsLeft -= bitsneeded;
  if(srcBitsLeft == 0)
    {
      srcRow++;
      srcBitsLeft = 32;
    }
  assert(srcBitsLeft <= 32 && srcBitsLeft > 0);

  /* y is the block number in y */
  /* y2 is the start of the source needed for this dst block line */
  /* y3 is the position of the source while accumulating for a dst block line. */
  /* iteDGPCOORD_Rate over dest rows */
  for (dy2 = startBlockRow, y2 = 0; dy2 < endBlockRow; y2 += nsrc[dy2], dy2++)
    {
      /* Clear the accumulation buffers */
      memset(buf,0,bufLen*sizeof(int));

      /* IteDGPCOORD_Rate over the source rows corresponding to this dest row */
      for (y3 = 0; y3 < nsrc[dy2]; y3++)
	{
	  /* Same set of loops over x, but here we're working on */
	  /* the bit level. */
	  /* where we're getting the current source from. */
	  data = srcRow;
			
	  /* a src buffer; valid bits at top */
	  curbits = *(data++) << (32 - srcBitsLeft);
	  bitsleft = srcBitsLeft; /* How many valid bits */

	  /* handle the first partial block...note, it could need to be clipped */
	  /* on both sides for a really small blat. */
			
	  /* buf[0] is the start of the first dest block in the row. */
	  dx = 0;
	  for (dx2 = firstStart; dx2 < firstEnd; dx2++)
	    {
	      bitsneeded = nsrc[dx2];

			
	      while (bitsleft < bitsneeded)
		{
		  if (curbits)
		    {
		      buf[dx + dx2] +=
			blat_bitcounts[curbits >> 24] +
			blat_bitcounts[(curbits >> 16) & 0xff] +
			blat_bitcounts[(curbits >> 8) & 0xff] +
			blat_bitcounts[curbits & 0xff];
		    }

		  curbits = *(data++);
		  bitsneeded -= bitsleft;
		  bitsleft = 32;
		}
				
				
				/* Now 8 >= bitsleft >= bitsneeded */
	      ourbits = (curbits >> (32 - bitsneeded));
	      if (ourbits)
		{
		  buf[dx + dx2] +=
		    blat_bitcounts[(ourbits >> 24)] +
		    blat_bitcounts[(ourbits >> 16) & 0xff] +
		    blat_bitcounts[(ourbits >> 8) & 0xff] +
		    blat_bitcounts[ourbits & 0xff];
		}

			
	      curbits <<= bitsneeded;
	      bitsleft -= bitsneeded;
	      if (!bitsleft)
		{
		  curbits = *(data++);
		  bitsleft = 32;
		}
	    }

			

	  /* now do the general x block as far as we can. */
	  /* dx is block position (dst) */
	  for (dx = firstEnd; dx < lastStart; dx += n)
	    {
				/* dx2 is the position in that block (dst) */
	      for (dx2 = 0; dx2 < n; dx2++)
		{
		  bitsneeded = nsrc[dx2];

		  while (bitsleft < bitsneeded)
		    {
		      if (curbits)
			{
			  buf[dx + dx2] +=
			    blat_bitcounts[curbits >> 24] +
			    blat_bitcounts[(curbits >> 16) & 0xff] +
			    blat_bitcounts[(curbits >> 8) & 0xff] +
			    blat_bitcounts[curbits & 0xff];
			}

		      curbits = *(data++);
		      bitsneeded -= bitsleft;
		      bitsleft = 32;
		    }
		  /* Now 32 >= bitsleft >= bitsneeded */
					
		  ourbits = (curbits >> (32 - bitsneeded));
		  if (ourbits)
		    {
		      buf[dx + dx2] +=
			blat_bitcounts[(ourbits >> 24)] +
			blat_bitcounts[(ourbits >> 16) & 0xff] +
			blat_bitcounts[(ourbits >> 8) & 0xff] +
			blat_bitcounts[ourbits & 0xff];
		      assert((ourbits >> 24) <= 255);
		    }
					
				
		  curbits <<= bitsneeded;
		  bitsleft -= bitsneeded;
		  if (!bitsleft)
		    {
		      curbits = *(data++);
		      bitsleft = 32;
		    }
		}
				
	    }


	  /* We're out of full blocks, now for some more fun;
	   * we want to accumulate only part of the last block
	   * this covers 2 cases: 
	   *	(1) we don't actually want all of the last block
	   *		and there exists enough source
	   *	(2) there's not enough source (the srcImg width 
	   *		need not be a multiple of m), and we want to
	   *		generate the right edge
	   */
	  dx = lastStart;
	  for (dx2 = 0; dx2 < lastSize; dx2++)
	    {
	      bitsneeded = nsrc[dx2];

	      while (bitsleft < bitsneeded)
		{
		  buf[dx + dx2] +=
		    blat_bitcounts[curbits >> 24] +
		    blat_bitcounts[(curbits >> 16) & 0xff] +
		    blat_bitcounts[(curbits >> 8) & 0xff] +
		    blat_bitcounts[curbits & 0xff];

		  curbits = *(data++);
		  bitsneeded -= bitsleft;
		  bitsleft = 32;
		}
				/* Now 8 >= bitsleft >= bitsneeded */
	      ourbits = (curbits >> (32 - bitsneeded));
	      buf[dx + dx2] +=
		blat_bitcounts[(ourbits >> 24)] +
		blat_bitcounts[(ourbits >> 16) & 0xff] +
		blat_bitcounts[(ourbits >> 8) & 0xff] +
		blat_bitcounts[ourbits & 0xff];

	      curbits <<= bitsneeded;
	      bitsleft -= bitsneeded;
	      assert(bitsleft>=0);
	    }

	  srcRow = srcRow + srcLpl;
	}

      /* Do the mapping from buf to dest pixels */
      dstMap = dstRow;
				
      /* once again, we need to split this up into 3 blocks
       * (1) partial first block
       * (2) full blocks
       * (3) partial last block
       */
      dx = 0;
      for (dx2 = firstStart ; dx2 < firstEnd ; dx2++)
	{
	  dstMap->b = dstMap->g = dstMap->r = 
	    lookups[nsrc[dy2] * nsrc[dx2]][buf[dx + dx2]];
	  dstMap ++;
	}

      for (dx = firstEnd ; dx < lastStart ; dx += n) 
	{
	  for (dx2 = 0 ; dx2 < n; dx2++)
	    {
	      dstMap->b = dstMap->g = dstMap->r = 
		lookups[nsrc[dy2] * nsrc[dx2]][buf[dx + dx2]];
	      dstMap ++;
	    }
	}

      dx = lastStart;
      for (dx2 = 0 ; dx2 < lastSize ; dx2++)
	{
	  dstMap->b = dstMap->g = dstMap->r = 
	    lookups[nsrc[dy2] * nsrc[dx2]][buf[dx + dx2]];
	  dstMap ++;
	}

		
      dstRow = (DgpImage_RGB *)((char *)dstRow + dstBpl);
    }


}

Dgp_Int32 DgpBlat_scaleBinaryRGBAvg(Dgp_UInt32 * src, Dgp_Int32 srcBpl,
		      Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		      Dgp_Int32 m, Dgp_Int32 n, Dgp_Allocators * a)
{
	

  /*
   * The source image can be considered to be tiled by squares
   * of m*m pixels, each of which maps to an n*n square in the
   * destination.  As we walk across each square, different numbers
   * of source pixels are averaged to produce each destination pixel.
   * For example, if m==3, n==2, the 3x3 square in the source maps
   * to the dest as follows:
   * 1 1 2
   * 1 1 2
   * 3 3 4
   * where
   * 1 2
   * 3 4
   * is the dest 2x2 square.  We need to compute nsrc[0...n-1] so that
   * nsrc[i] is the number of source pixels going into dest pixel i,
   * both horizontally and vertically (dest pixel (i,j) came from
   * (nsrc[i], nsrc[j]) source pixels).
   */
  Dgp_Int32 *nsrc = NULL;
  unsigned char **lookups = NULL;
  Dgp_Int32 * buf = NULL;

  Dgp_Int32 y,i,j,k,npix;
  Dgp_Int32 bufLen;
  Dgp_UInt32 *srcRow;
  Dgp_Int32 srcBitsLeft;
  Dgp_Int32 srcLpl;
  DgpImage_RGB * dstRow;
  Dgp_Int32 startBlockRow,endBlockRow;
  Dgp_Int32 firstStart, firstEnd, lastEnd;
	
  Dgp_Int32 maxval = 0;
	
  Dgp_Int32 srcStartY,srcStartX,dstStartY,dstStartX;


  nsrc = (Dgp_Int32 *)a->malloc(n*sizeof(int));
  if(!nsrc) goto bailout;

  for (i = 0; i < n; i++) {
    nsrc[i] = (m * (i + 1) + n / 2) / n - (m * i + n / 2) / n;
  }

  /*
   * Now, for each possible pair of values that appear in nsrc,
   * we're going to need a lookup table mapping a value from
   * 0..nsrc[i]*nsrc[j] to 0..255.  This corresponds to the different
   * sizes of the source regions corresponding to each dest region.
   */

  for (i = 0; i < n; i++) {
    if (maxval < nsrc[i] * nsrc[i]) {
      maxval = nsrc[i] * nsrc[i];
    }
  }
	
  lookups = (unsigned char **)a->malloc((maxval+1)*sizeof(unsigned char *));
  if(!lookups)
    goto bailout;
  for(i=0; i<maxval+1; i++)
    lookups[i]=NULL;

  /* This could be more efficient... */
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++) {
      npix = nsrc[i] * nsrc[j];
      if (lookups[npix] == NULL) {
	lookups[npix] = (unsigned char *)a->malloc((npix+1)*sizeof(unsigned char));
	if(!lookups[npix])
	  goto bailout;
	for (k = 0; k <= npix; k++) {
	  lookups[npix][k]  = (unsigned char)(255 - 255 * k / npix);
	}
      }
    }
  }
	
  dstStartX = dstX - dstX % n;
  dstStartY = dstY - dstY % n;
	
  srcStartX = (dstStartX / n)*m;
  srcStartY = (dstStartY / n)*m;

  /* A buffer to hold the accumulation of pixels */
  bufLen=(int)(ceil((float)(dstX + dstW) / n) * n) - dstStartX;
  buf = (Dgp_Int32 *)a->malloc(bufLen*sizeof(int));
  if(!buf) 
    goto bailout;

  /* we have 3 loops here
   * (1) to handle needing only a partial top dst nxn block row (clipped both)
   * (2) general case blocked rows
   * (3) handle a partial bottom row (clipped at the bottom)
   */

  srcLpl = srcBpl / sizeof(Dgp_UInt32);
  srcRow = src + ((srcStartY - srcMinY) * srcLpl + ((srcStartX - srcMinX) / 32));
  srcBitsLeft = 32 - ((srcStartX - srcMinX) % 32);
	
  dstRow = (DgpImage_RGB *)((char *)dst + (dstY - dstMinY) * dstBpl);
  dstRow += (dstX - dstMinX);

  startBlockRow = dstY - dstStartY;
  endBlockRow = MIN(n, startBlockRow + dstH);

  firstStart = dstX - dstStartX;
  firstEnd = MIN(n, firstStart + dstW);

  lastEnd = firstStart + dstW;
	
  DgpBlat_scaleBinaryRGBAvgBlockRow(srcRow, srcBitsLeft, srcLpl, 
			    dstRow, dstBpl, n, 
			    startBlockRow, endBlockRow,
			    firstStart, firstEnd, lastEnd,
			    nsrc, buf, bufLen, lookups);
  srcRow = srcRow + (m * srcLpl);
  dstRow = (DgpImage_RGB *)((char *)dstRow + (endBlockRow - startBlockRow) * dstBpl);

  for ( y = endBlockRow-startBlockRow ; y+n <= dstH ; y += n)
    {
      DgpBlat_scaleBinaryRGBAvgBlockRow(srcRow, srcBitsLeft, srcLpl, 
				dstRow, dstBpl, n, 
				0, n, firstStart, firstEnd, lastEnd,
				nsrc, buf, bufLen, lookups);

      srcRow = srcRow + (m * srcLpl);
      dstRow = (DgpImage_RGB *)((char *)dstRow + (n * dstBpl));
    }

  if ( y < dstH)
    {
      DgpBlat_scaleBinaryRGBAvgBlockRow(srcRow, srcBitsLeft, srcLpl, 
				dstRow, dstBpl, n, 
				0,  dstH - y , firstStart, firstEnd, lastEnd,
				nsrc, buf, bufLen, lookups);
    }

  a->free(nsrc);
  for (i = 0; i< maxval+1; i++)
    {
      if(lookups[i])
	{
	  a->free(lookups[i]);
	}
    }
  a->free(lookups);
  lookups = NULL;
  a->free(buf);


  return 1;

 bailout:
  if (nsrc)
    {
      a->free(nsrc);
    }
  if (lookups)
    {
      for (i = 0; i< maxval+1; i++)
	{
	  if(lookups[i])
	    {
	      a->free(lookups[i]);
	    }
	}
      a->free(lookups);
    }
  if (buf)
    a->free(buf);

  return 0;

}




Dgp_Int32 DgpBlat_scaleBinaryRGBAvgSub(Dgp_UInt32 * src, Dgp_Int32 srcBpl,
			 Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			 DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			 Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
			 DgpCoord_Resolution * res, Dgp_Allocators * a)
{
  /* We're going to do a generic scaling down by scaling down 
     by 2,4, or 8 (simultaneous in x and y) and then doing
     sub-sampling from there. But we're going to do this all
     at once, deciding which line buffers of 2,4,or 8 to generate
     and then sub-sampling those. So first, should be scale down
     by 2,4,or 8?
	 */
	
  /* If there's no averaging to be done (scales <=1.0  >.5) we'll just
     call the straight subsampling function.
  */
	
	
  if(DgpCoord_equal(res->xRes, DgpCoord_eq(1,1)) && DgpCoord_equal(res->yRes, DgpCoord_eq(1,1)))
    {
      DgpBlat_blitBinaryRGB(src, srcBpl, srcMinX, srcMinY,
		    dst, dstBpl, dstMinX, dstMinY,
		    dstX, dstY, dstW, dstH);
      return 1;
    }

  /*	if(scaleX>.5f || scaleY>.5f)
		{		
		return scaleBlitBinaryRGB(src, srcBpl, srcMinX, srcMinY,
			dst, dstBpl, dstMinX, dstMinY,
			dstX, dstY, dstW, dstH,
			scaleX, scaleY);
		}
*/		
	/* If this is just averaging, we'll call the special case functions
	   that just do averaging of 2,4,8
	 */
  if(DgpCoord_equal(res->xRes, res->yRes))
    {
      if (DgpCoord_equal(res->xRes, DgpCoord_eq(1, 2))){
          return DgpBlat_scaleBinaryRGBAvgby2(src, srcBpl, srcMinX, srcMinY,
              dst, dstBpl, dstMinX, dstMinY,
              dstX, dstY, dstW, dstH, a);
      }
      if (DgpCoord_equal(res->xRes, DgpCoord_eq(1, 4))){
          return DgpBlat_scaleBinaryRGBAvgby4(src, srcBpl, srcMinX, srcMinY,
              dst, dstBpl, dstMinX, dstMinY,
              dstX, dstY, dstW, dstH, a);
      }
      if (DgpCoord_equal(res->xRes, DgpCoord_eq(1, 8))){
          return DgpBlat_scaleBinaryRGBAvgby8(src, srcBpl, srcMinX, srcMinY,
              dst, dstBpl, dstMinX, dstMinY,
              dstX, dstY, dstW, dstH, a);
      }
      else {
          return DgpBlat_scaleBinaryRGBAvg(src, srcBpl, srcMinX, srcMinY,
              dst, dstBpl, dstMinX, dstMinY,
              dstX, dstY, dstW, dstH,
              res->xRes.den,res->xRes.num, a);
      }
    }
      
  
  /* So since none of the special cases work, we'll just find the
     the larger average and scale downward.
  */
  return DgpBlat_scaleBinaryRGBAvgSubGen(src, srcBpl, srcMinX, srcMinY,
				 dst, dstBpl, dstMinX, dstMinY,
				 dstX, dstY, dstW, dstH, res, a);
	
}







static INLINE void scaleBinaryRGBAvgby2Row(Dgp_UInt32 * srcRow, 
				    Dgp_Int32 srcLpl, Dgp_Int32 srcW, unsigned char * grayBuf)
{
  Dgp_UInt32 l;
  Dgp_Int32 i,x,j,v;
  for (i = 0, x = 0, j = 0; x + 32 <= srcW; x += 32, i++)
    {
      l=srcRow[i];
      grayBuf[j++] = blat_bitcounts[(int)((l >> 30) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 28) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 26) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 24) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 22) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 20) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 18) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 16) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 14) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 12) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 10) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 8) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 6) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 4) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 2) & 0x3)]; 
      grayBuf[j++] = blat_bitcounts[(int)(l & 0x3)]; 
    }

  if (x < srcW)
    {
      l=srcRow[i++];
      for (v = 0; x < srcW; x += 2, v++)
	{
	  grayBuf[j++] = blat_bitcounts[(int)((l>>(30 - 2*v)) & 0x3)];
	}
    }
	
  srcRow = srcRow + srcLpl;
  for (i = 0,x = 0, j = 0; x + 32 <= srcW; x += 32, i++)
    {
      l = srcRow[i];
      grayBuf[j++] += blat_bitcounts[(int)((l >> 30) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 28) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 26) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 24) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 22) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 20) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 18) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 16) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 14) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 12) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 10) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 8) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 6) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 4) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 2) & 0x3)]; 
      grayBuf[j++] += blat_bitcounts[(int)(l & 0x3)]; 
    } 

  if (x < srcW) { 
    Dgp_Int32 l = srcRow[i++]; 
    for (v = 0; x < srcW; x += 2, v++) 
      { 
	grayBuf[j++] += blat_bitcounts[(int)((l >> (30 - 2 * v)) & 0x3)]; 
      } 
  } 
}


static INLINE void scaleBinaryRGBAvgby4Row(Dgp_UInt32 * srcRow, 
				    Dgp_Int32 srcLpl, Dgp_Int32 srcW, 
				    unsigned char * grayBuf)

{
	
  Dgp_UInt32 l;
  Dgp_Int32 i,x,j,v;
  for (i=0, x = 0, j = 0; x + 32 <= srcW; x+=32, i++)
    {
      l=srcRow[i];
      grayBuf[j++] = blat_bitcounts[(int)((l >> 28) & 0xf)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 24) & 0xf)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 20) & 0xf)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 16) & 0xf)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 12) & 0xf)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 8) & 0xf)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 4) & 0xf)]; 
      grayBuf[j++] = blat_bitcounts[(int)(l & 0xf)]; 
    }
  if (x < srcW)
    {
      l=srcRow[i++];
      for (v = 0; x < srcW; x += 4, v++)
	{
	  grayBuf[j++] = blat_bitcounts[(int)((l>>(28 - 4*v)) & 0xf)];
	}
    }
  /* these 3 chunks are all the same */		
  srcRow = srcRow + srcLpl;
  for (i=0,x=0, j=0; x + 32 <= srcW; x += 32, i++)
    {
      l = srcRow[i];
      grayBuf[j++] += blat_bitcounts[(int)((l >> 28) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 24) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 20) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 16) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 12) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 8) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 4) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)(l & 0xf)]; 
    } 

  if (x < srcW) { 
    Dgp_Int32 l = srcRow[i++]; 
    for (v = 0; x < srcW; x += 4, v++) 
      { 
	grayBuf[j++] += blat_bitcounts[(int)((l >> (28 - 4 * v)) & 0xf)]; 
      } 
  } 
  /* */
  srcRow = srcRow + srcLpl;
  for (i=0,x=0, j=0; x + 32 <= srcW; x += 32, i++)
    {
      l = srcRow[i];
      grayBuf[j++] += blat_bitcounts[(int)((l >> 28) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 24) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 20) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 16) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 12) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 8) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 4) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)(l & 0xf)]; 
    } 

  if (x < srcW) { 
    Dgp_Int32 l = srcRow[i++]; 
    for (v = 0; x < srcW; x += 4, v++) 
      { 
	grayBuf[j++] += blat_bitcounts[(int)((l >> (28 - 4 * v)) & 0xf)]; 
      } 
  } 
  /* */
  srcRow = srcRow + srcLpl;
  for (i=0,x=0, j=0; x + 32 <= srcW; x += 32, i++)
    {
      l = srcRow[i];
      grayBuf[j++] += blat_bitcounts[(int)((l >> 28) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 24) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 20) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 16) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 12) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 8) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 4) & 0xf)]; 
      grayBuf[j++] += blat_bitcounts[(int)(l & 0xf)]; 
    } 

  if (x < srcW) { 
    Dgp_Int32 l = srcRow[i++]; 
    for (v = 0; x < srcW; x += 4, v++) 
      { 
	grayBuf[j++] += blat_bitcounts[(int)((l >> (28 - 4 * v)) & 0xf)]; 
      } 
  } 
  /* */
}


static INLINE void scaleBinaryRGBAvgby8Row(Dgp_UInt32 * srcRow, 
				    Dgp_Int32 srcLpl, Dgp_Int32 srcW, 
				    unsigned char * grayBuf)
{
  Dgp_UInt32 l;
  Dgp_Int32 i,x,j,v;
  for (i=0, x = 0, j = 0; x + 32 <= srcW; x+=32, i++)
    {
      l=srcRow[i];
      grayBuf[j++] = blat_bitcounts[(int)((l >> 24) & 0xff)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 16) & 0xff)]; 
      grayBuf[j++] = blat_bitcounts[(int)((l >> 8) & 0xff)]; 
      grayBuf[j++] = blat_bitcounts[(int)(l & 0xff)]; 
    }
  if (x < srcW)
    {
      l=srcRow[i++];
      for (v = 0; x < srcW; x += 8, v++)
	{
	  grayBuf[j++] = blat_bitcounts[(int)((l>>(24 - 8*v)) & 0xff)];
	}
    }

  /* these 7 chunks should all be the same. */		
  srcRow = srcRow + srcLpl;
  for (i=0,x=0, j=0; x + 32 <= srcW; x += 32, i++)
    {
      l = srcRow[i];
      grayBuf[j++] += blat_bitcounts[(int)((l >> 24) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 16) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 8) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)(l & 0xff)]; 
    } 

  if (x < srcW) { 
    Dgp_Int32 l = srcRow[i++]; 
    for (v = 0; x < srcW; x += 8, v++) 
      { 
	grayBuf[j++] += blat_bitcounts[(int)((l >> (24 - 8 * v)) & 0xff)]; 
      } 
  } 
  /* */
  srcRow = srcRow + srcLpl;
  for (i=0,x=0, j=0; x + 32 <= srcW; x += 32, i++)
    {
      l = srcRow[i];
      grayBuf[j++] += blat_bitcounts[(int)((l >> 24) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 16) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 8) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)(l & 0xff)]; 
    } 

  if (x < srcW) { 
    Dgp_Int32 l = srcRow[i++]; 
    for (v = 0; x < srcW; x += 8, v++) 
      { 
	grayBuf[j++] += blat_bitcounts[(int)((l >> (24 - 8 * v)) & 0xff)]; 
      } 
  } 
  /* */
  srcRow = srcRow + srcLpl;
  for (i=0,x=0, j=0; x + 32 <= srcW; x += 32, i++)
    {
      l = srcRow[i];
      grayBuf[j++] += blat_bitcounts[(int)((l >> 24) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 16) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 8) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)(l & 0xff)]; 
    } 

  if (x < srcW) { 
    Dgp_Int32 l = srcRow[i++]; 
    for (v = 0; x < srcW; x += 8, v++) 
      { 
	grayBuf[j++] += blat_bitcounts[(int)((l >> (24 - 8 * v)) & 0xff)]; 
      } 
  } 
  /* */
  srcRow = srcRow + srcLpl;
  for (i=0,x=0, j=0; x + 32 <= srcW; x += 32, i++)
    {
      l = srcRow[i];
      grayBuf[j++] += blat_bitcounts[(int)((l >> 24) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 16) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 8) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)(l & 0xff)]; 
    } 

  if (x < srcW) { 
    Dgp_Int32 l = srcRow[i++]; 
    for (v = 0; x < srcW; x += 8, v++) 
      { 
	grayBuf[j++] += blat_bitcounts[(int)((l >> (24 - 8 * v)) & 0xff)]; 
      } 
  } 
  /* */
  srcRow = srcRow + srcLpl;
  for (i=0,x=0, j=0; x + 32 <= srcW; x += 32, i++)
    {
      l = srcRow[i];
      grayBuf[j++] += blat_bitcounts[(int)((l >> 24) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 16) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 8) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)(l & 0xff)]; 
    } 

  if (x < srcW) { 
    Dgp_Int32 l = srcRow[i++]; 
    for (v = 0; x < srcW; x += 8, v++) 
      { 
	grayBuf[j++] += blat_bitcounts[(int)((l >> (24 - 8 * v)) & 0xff)]; 
      } 
  } 
  /* */
  srcRow = srcRow + srcLpl;
  for (i=0,x=0, j=0; x + 32 <= srcW; x += 32, i++)
    {
      l = srcRow[i];
      grayBuf[j++] += blat_bitcounts[(int)((l >> 24) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 16) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 8) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)(l & 0xff)]; 
    } 

  if (x < srcW) { 
    Dgp_Int32 l = srcRow[i++]; 
    for (v = 0; x < srcW; x += 8, v++) 
      { 
	grayBuf[j++] += blat_bitcounts[(int)((l >> (24 - 8 * v)) & 0xff)]; 
      } 
  } 
  /* */
  srcRow = srcRow + srcLpl;
  for (i=0,x=0, j=0; x + 32 <= srcW; x += 32, i++)
    {
      l = srcRow[i];
      grayBuf[j++] += blat_bitcounts[(int)((l >> 24) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 16) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)((l >> 8) & 0xff)]; 
      grayBuf[j++] += blat_bitcounts[(int)(l & 0xff)]; 
    } 

  if (x < srcW) { 
    Dgp_Int32 l = srcRow[i++]; 
    for (v = 0; x < srcW; x += 8, v++) 
      { 
	grayBuf[j++] += blat_bitcounts[(int)((l >> (24 - 8 * v)) & 0xff)]; 
      } 
  } 
  /* */
}





