#ifndef __SCALEBLITAVGMONO_H__
#define __SCALEBLITAVGMONO_H__

#include "scaleBlit.h"
#include "resolution_t.h"
#ifdef __cplusplus
extern "C"
{
#endif


/* Here's some functions for scaling a packed bits image to gray.
	We have special cases for factors of 2,4,8, and we subsample
	from a the closest (but higher) factor down to the desired scale.
	It looks pretty nice.
	*/

/* In all cases, the idea is this function generates a specific portion
   of a destination scaled image [(dstX,dstY) to (dstX+dstW-1,dstY+dstH-1)]
   given the appropriate portion of the source image. 
	src is a pointer to source data
	srcBpl is the number of bytes per line in the source buffer
	srcMinX, srcMinY is the (x,y) coordinate of the src image
		corresponding to the src pointer

	dst is a pointer to destination data (RGB)
	dstBpl is the number of bytes per line in the dest buffer
	dstMinX, dstMinY is the (x,y) coordinate of the dst image
		corresponding to the dest pointer.

	dstX,dstY
	dstW,dstH is the rectangle of the destination image that
		is desired to be filled in.
	scaleX,scaleY is the x/y scale (1/2 is a factor of 2, etc.)
 */

/* Calls malloc / free for a temporary line buffer 
	returns 1 if successful
			0 if not (the memory allocation failed)
 */


/* scaleBinaryRGBAvgSub is the normal entry point; it case dispatches the 
	scale factors and calls the special case versions or the general 
	case
 */
extern Dgp_Int32 
DgpBlat_scaleBinaryRGBAvgSub(Dgp_UInt32 * src, Dgp_Int32 srcBpl,
			 Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			 DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			 Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
			 DgpCoord_Resolution * res, Dgp_Allocators * a);


/* here's a generic m/n rational scale. the src is tiled by mxm blocks */
/* that go to nxn blocks in the dest. */
extern Dgp_Int32 
DgpBlat_scaleBinaryRGBAvg(Dgp_UInt32 * src, Dgp_Int32 srcBpl,
		      Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
		      DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
		      Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
		      Dgp_Int32 m, Dgp_Int32 n, Dgp_Allocators * a);

/* We've got averaging for scale factors of 1/2,1/4,1/8. These functions
	all come from the same code, so changes / optimizations to one 
	should be applied to the rest. 
 */
extern Dgp_Int32 
DgpBlat_scaleBinaryRGBAvgby2(Dgp_UInt32 * src,
			 Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			 DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			 Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

extern Dgp_Int32 
DgpBlat_scaleBinaryRGBAvgby4(Dgp_UInt32 * src,
			 Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			 DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			 Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

extern Dgp_Int32 
DgpBlat_scaleBinaryRGBAvgby8(Dgp_UInt32 * src,
			 Dgp_Int32 srcBpl, Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			 DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			 Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Allocators * a);

/* If you want some scale factor (or scaleX and scaleY are different),
	this general purpose function will find the closest (but above) one
	of the average factors above, and then scale down. It does this in
	one pass, not by doing the full average and then subsampling.
 */
extern Dgp_Int32 
DgpBlat_scaleBinaryRGBAvgSubGen(Dgp_UInt32 * src, Dgp_Int32 srcBpl,
			    Dgp_Int32 srcMinX, Dgp_Int32 srcMinY,
			    DgpImage_RGB * dst, Dgp_Int32 dstBpl, Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
			    Dgp_Int32 dstX, Dgp_Int32 dstY, Dgp_Int32 dstW, Dgp_Int32 dstH,
			    DgpCoord_Resolution * res, Dgp_Allocators * a);



#ifdef __cplusplus
}
#endif


#endif /*__SCALEBLITAVGMONO_H__ */
