#ifndef COLOR_EXTENSION_H
#define COLOR_EXTENSION_H

#include <image.h>
#include <samptok.h>
#include "dgp_state.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* we only write 8 bits per sample */
#define BITS_PER_SAMPLE 8

/* each color token is labeled with a color_tag which  */
/* indicates its color or sampled image index. */
typedef struct {
	union {
		DgpImage_RGB rgb_color;
		Dgp_UInt32 image_index;
		samptokClass samp_class;
	} value;
	Dgp_UInt8 color_type;
} color_tag;

/* possible types of a color_tag */
#define RGB_COLOR 1
#define IMAGE_INDEX 2
#define SAMP_CLASS 3

/* the cookie is shared by the encoder & decoder  */
typedef struct {
  void *colorbuf;
  Dgp_Bool report;
  Dgp_State *dgp_state;
} color_cookie;

/* the encoder & decoder shares cookie constructors & distructors */
extern ErrCode
initColors(color_cookie **ccookie, Dgp_Bool report, Dgp_State *dgp_state);

extern void
cleanupColors(color_cookie *ccookie);

#ifdef __cplusplus
}
#endif

#endif
