/*
   Pedro Felzenszwalb, Summer '98.
   New color extension to position block encoder & decoder.
*/

#include "encode-colors.h"
#include <huff.h>
#include <bitfile.h>
#include <stdlib.h>
#include <stdio.h>
#include "PsgFix.h"

#if !defined(VXWORKS) && !defined(EQUATOR_CA_ENV)
#include <memory.h>
#endif


DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcolcomp/decode-colors.c,v 1.25 2003/02/10 22:53:22 mkrocken Exp $");

/* each run in the color extension is a color tag + its length */
typedef struct {
	Dgp_Int32 tag;
	Dgp_Int32 len;
} tag_run;

static ErrCode
get_tag_index(DgpAdt_BitFile f, Dgp_UInt32 *firstcode, Dgp_Int32 **index, 
			  Dgp_UInt32 *tag) 
{
  DGP_PROLOGUE("get_tag_index()", NULL);
	Dgp_UInt32 tmp = 0; 
	Dgp_UInt32 v = 0, l = 0; 

	/* see the comment above huffBuildDecodeTables() in huff.c (libwjr/support) 
       for info on what this next snippet of code is doing (it's basically using
       some tables generated by huffBuildDecodeTables() to decode the
       Huffman code). */
	while (v < firstcode[l]) {
		tmp = 0;
		DGP_ERR_PASS(DgpAdt_bfReadCode(&tmp,1,f));
		v = (v << 1) | tmp;
		l++;
	} 

	/* again, you have to read William's comment above huffBuildDecodeTables()
	   to know what the following line is doing. */
	*tag = index[l][v - firstcode[l]];
	NORMAL_RETURN;
}

/* decode a color extension & returns a bunch of unique color tags together with an array of indices */
ErrCode
decodeColorsIndexed(color_cookie *ccookie, void *colorbuf, size_t colorbuflen, 
					Dgp_Int32 **_tag_index, Dgp_Int32 *_num_tag_index, color_tag **_unique_tags, Dgp_Int32 *_num_unique) 
{
  DGP_PROLOGUE("decodeColorsIndexed()", ccookie->DGP_PERRSTATE);
	DgpAdt_BitFile f = NULL;
	Dgp_Int32 num_unique_tags, num_unique_colors, num_unique_images;
	color_tag *unique_tags = NULL;
	Dgp_Int32 *tag_index = NULL;
	Dgp_Int32 num_tag_index;
	tag_run *runs = NULL;
	Dgp_Int32 num_runs;
	Dgp_UInt32 *firstcode = NULL;
	Dgp_Int32 **index = NULL;
	DgpEncode_Huff *vals = NULL;
	Dgp_UInt32 tmp_uDgp_Int32;
	Dgp_UInt32 tmp_int;
	Dgp_Int32 pos, i, j;
  Dgp_Bool index_built = DGP_FALSE;
  Dgp_Allocators *a = &(ccookie->DGP_ALLOCATORS);


	/* Set up the DgpAdt_BitFile */
	DGP_ERR_PASS_CLEANUP(DgpAdt_bfNew(&f, colorbuf, colorbuflen, ccookie->dgp_state));

	/* allocate two arrays that huffBuildDecodeTables() wants */
	DGP_CHECK_ALLOC_CLEANUP(index = (Dgp_Int32 **)(a->malloc)((DGP_MAX_CODELEN + 1) * sizeof(Dgp_Int32 *)));
	DGP_CHECK_ALLOC_CLEANUP(firstcode = (Dgp_UInt32 *)(a->malloc)((DGP_MAX_CODELEN + 1) * sizeof(Dgp_UInt32)));

	/* get the number of colors  */
	tmp_int = 0;
	DGP_ERR_PASS_CLEANUP(DgpAdt_bfReadValue(&tmp_int, f));
	num_unique_colors = tmp_int;
	
	/* get the number of images */
	tmp_int = 0;
	DGP_ERR_PASS_CLEANUP(DgpAdt_bfReadValue(&tmp_int, f));
	num_unique_images = tmp_int;

	num_unique_tags = num_unique_images + num_unique_colors;

	DGP_CHECK_ALLOC_CLEANUP(vals = (DgpEncode_Huff *)(a->malloc)(num_unique_tags*sizeof(DgpEncode_Huff)));

	DGP_CHECK_ALLOC_CLEANUP(unique_tags = (color_tag *)(a->malloc)(num_unique_tags*sizeof(color_tag)));

	/* read color tags  */
	DGP_REPORTF0(ccookie->report, "\nDistinct colors:");

	for (i = 0; i < num_unique_colors; i++) {
		unique_tags[i].color_type = RGB_COLOR;
		tmp_uDgp_Int32 = 0;
		DGP_ERR_PASS_CLEANUP(DgpAdt_bfReadCode(&tmp_uDgp_Int32, BITS_PER_SAMPLE, f));
		unique_tags[i].value.rgb_color.r = (Dgp_UInt8)tmp_uDgp_Int32;
		tmp_uDgp_Int32 = 0;
		DGP_ERR_PASS_CLEANUP(DgpAdt_bfReadCode(&tmp_uDgp_Int32, BITS_PER_SAMPLE, f));
		unique_tags[i].value.rgb_color.g = (Dgp_UInt8)tmp_uDgp_Int32;
		tmp_uDgp_Int32 = 0;
		DGP_ERR_PASS_CLEANUP(DgpAdt_bfReadCode(&tmp_uDgp_Int32, BITS_PER_SAMPLE, f));
		unique_tags[i].value.rgb_color.b = (Dgp_UInt8)tmp_uDgp_Int32;

		DGP_REPORTF3(ccookie->report, "(%d,%d,%d)", unique_tags[i].value.rgb_color.r, 
				unique_tags[i].value.rgb_color.g, unique_tags[i].value.rgb_color.b);

		tmp_int = 0;
		DGP_ERR_PASS_CLEANUP(DgpAdt_bfReadValue(&tmp_int, f));
		vals[i].codelen = tmp_int;
	}

	/* read image indices tags */
	DGP_REPORTF0(ccookie->report, "\nDistinct images:");

	for (j = 0; j < num_unique_images; j++, i++) {
		unique_tags[i].color_type = IMAGE_INDEX;
		tmp_int = 0;
		DGP_ERR_PASS_CLEANUP(DgpAdt_bfReadValue(&tmp_int, f));
		unique_tags[i].value.image_index = (Dgp_UInt8)tmp_int;

		DGP_REPORTF1(ccookie->report, "%d", (Dgp_Int32)unique_tags[i].value.image_index);

		tmp_int = 0;
		DGP_ERR_PASS_CLEANUP(DgpAdt_bfReadValue(&tmp_int, f));
		vals[i].codelen = tmp_int;
	}

	/* build the huffman codes */
	DGP_ERR_PASS_CLEANUP(DgpEncode_huffBuildDecodeTables(ccookie->dgp_state, vals, num_unique_tags, firstcode, index));
  index_built = DGP_TRUE;

	/* get runs */
	tmp_int = 0;
	DGP_ERR_PASS_CLEANUP(DgpAdt_bfReadValue(&tmp_int, f));
	num_runs = tmp_int;

	/* allocate space for runs */
	DGP_CHECK_ALLOC_CLEANUP(runs = (a->malloc)(sizeof(tag_run)*num_runs));

	/* read runs & count positions */
	num_tag_index = 0;
	for (i = 0; i < num_runs; i++) {
		DGP_ERR_PASS_CLEANUP(get_tag_index(f, firstcode, index, &tmp_uDgp_Int32));
		runs[i].tag = tmp_uDgp_Int32;
		tmp_int = 0;
		DGP_ERR_PASS_CLEANUP(DgpAdt_bfReadValue(&tmp_int, f));
		runs[i].len = tmp_int;
		num_tag_index += runs[i].len;
	}
	
	/* allocate space for tag indices */
	DGP_CHECK_ALLOC_CLEANUP(tag_index = (a->malloc)(sizeof(Dgp_Int32)*num_tag_index));

	/* stuff it */
	pos = 0;
	for (i = 0; i < num_runs; i++) {
		for (j = 0; j < runs[i].len; j++) {
			tag_index[pos] = runs[i].tag;
			pos++;
		}
	}
	DGP_ASSERT(pos == num_tag_index);
	
	/* free up */
	(a->free)(runs);
  runs = NULL;
	{ size_t dummy;  DgpAdt_bfFree(f,&dummy); }
	(a->free)((void *)vals);
  vals = NULL;
  for (i = 0; i <= DGP_MAX_CODELEN; i++) {
		if (index[i])
			(a->free)((void *)index[i]);
  }
	a->free((void *)index);
  index = NULL;
	a->free((void *)firstcode);
  firstcode = NULL;

	*_unique_tags = unique_tags;
  unique_tags = NULL;
	*_num_unique = num_unique_tags;
	*_tag_index = tag_index;
  tag_index = NULL;
	*_num_tag_index = num_tag_index;

	NORMAL_RETURN;

ERR_CLEANUP:

  if(index != NULL) {
    if(index_built) {
      for (i = 0; i <= DGP_MAX_CODELEN; i++) {
		    if (index[i])
			    (a->free)((void *)index[i]);
      }
    }
    (a->free)(index);
  }
  if(runs != NULL) (a->free)(runs);
  if(vals != NULL) (a->free)(vals);
  if(firstcode != NULL) (a->free)(firstcode);
  if(tag_index != NULL) (a->free)(tag_index);
  if(unique_tags != NULL) (a->free)(unique_tags);

	ERR_RETURN;
}

/* decodes a color extension & returns a bunch of color tags. it also returns the number of unique tags */
/* just for fun. */
ErrCode
decodeColors(color_cookie *ccookie, void *colorbuf, size_t colorbuflen, 
			 color_tag **_tags, Dgp_Int32 *_num_tags, Dgp_Int32 *_num_unique) 
{
  DGP_PROLOGUE("decodeColors()", ccookie->DGP_PERRSTATE);

	Dgp_Int32 i;
	color_tag *tags = NULL, *unique_tags = NULL;
	Dgp_Int32 *tag_index = NULL, num_tag_index, num_unique;

	DGP_ERR_PASS_CLEANUP(decodeColorsIndexed(ccookie, colorbuf, colorbuflen, &tag_index,
		&num_tag_index, &unique_tags, &num_unique));
					
	DGP_CHECK_ALLOC_CLEANUP(tags = (ccookie->DGP_MALLOC)(sizeof(color_tag)*num_tag_index));

	for (i = 0; i < num_tag_index; i++) {
		tags[i] = unique_tags[tag_index[i]];
	}

	/* free up */
	(ccookie->DGP_FREE)(unique_tags);
  unique_tags = NULL;
	(ccookie->DGP_FREE)(tag_index);
	tag_index = NULL;

	*_tags = tags;
  tags = NULL;
	*_num_tags = num_tag_index;
	*_num_unique = num_unique;

	NORMAL_RETURN;

ERR_CLEANUP:

  if(tags != NULL) (ccookie->DGP_FREE)(tags);
	if(unique_tags != NULL) (ccookie->DGP_FREE)(unique_tags);
	if(tag_index != NULL) (ccookie->DGP_FREE)(tag_index);
  
  ERR_RETURN;
}
