/*
   Pedro Felzenszwalb, Summer '98.
   New color extension to position block encoder & decoder.
*/

#include "encode-colors.h"
#include <huff.h>
#include <bitfile.h>
#include <stdlib.h>
#include <stdio.h>
#include "PsgFix.h"

#if !defined(VXWORKS) && !defined(EQUATOR_CA_ENV)
#include <memory.h>
#endif


DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcolcomp/decode-jbig2col.c,v 1.6 2003/02/10 22:53:22 mkrocken Exp $");


/* decodes a color extension & returns a bunch of color tags.  */
ErrCode
decodeJbig2Colors(color_cookie *ccookie, void *colorbuf, size_t colorbuflen, 
			 color_tag **_tags, Dgp_Int32 *_num_tags) 
{
    DGP_PROLOGUE("decodeColors()", ccookie->DGP_PERRSTATE);
    Dgp_UInt32 i,j;
    color_tag *tags = NULL, color;
    Dgp_UInt8 *buf;
	Dgp_UInt8 *bufend;
    Dgp_UInt32 nComp, compLen, nval=0;
    Dgp_UInt32 run_len;

    /* try to decode header */
    DGP_CHECK_NULL(colorbuf);
    
    if(colorbuflen <6) ERR_RETURN;

    buf= colorbuf;
	bufend = buf + colorbuflen;
    nComp = *(buf ++);           /* number of color component */
    compLen = *(buf ++);     /* component length */

    color.color_type = RGB_COLOR;

    /* try to decode runlength encoded color tag */
    if(nComp == 3 && compLen == 1){

        nval = ((*buf) << 24) + ((*(buf+1)) <<16) + ((*(buf+2)) <<8) + *(buf+3);
        buf += 4;
        DGP_CHECK_ALLOC_CLEANUP(tags = (ccookie->DGP_MALLOC)(sizeof(color_tag) * nval));
		i = 0;
		while ( ( buf < bufend +3) && ( i < nval)) {
			run_len = *(buf ++);
			if (run_len == 0) {
				run_len = *(buf ++);
				run_len = (run_len << 8) + *(buf ++);
			}
            color.value.rgb_color.r = *(buf++);
            color.value.rgb_color.g = *(buf++);
            color.value.rgb_color.b = *(buf++);
			for ( j = 0; j < run_len; j ++)
				tags[i ++] = color;
		}
        *_tags = tags;
        *_num_tags = nval;
    }else{ /* right now the digipaper library doesn't support other color scheme. */
        *_num_tags = 0;
        *_tags = NULL;
    }

	/* free up */

        tags = NULL;

	NORMAL_RETURN;

ERR_CLEANUP:

  if(tags != NULL) (ccookie->DGP_FREE)(tags);
    *_num_tags = 0;
    *_tags = NULL;
  ERR_RETURN;
}
