#ifndef __DICT_UTIL_H__
#define __DICT_UTIL_H__

#include "token-compress.h"
#include <tok-lossless.h>
#include <list.h>
#include <util.h>
#include <mmcompress.h>
#include <dgp_state.h>
#include "token-compress-priv.h"

#ifdef __cplusplus
extern "C"
{
#endif

#define SCALE_DIMENSION(_dim, _scale)  (Dgp_Int32)(((Dgp_Float32)(_dim) / (_scale)) + 0.5)
#define SCALE_RESOLUTION(_res, _scale) (Dgp_Int32)((Dgp_Float32)(_res) / (_scale))


typedef struct {
  Dgp_UInt16 xpos, ypos;
  samptokClass samptok;
  DgpImage_Packed mask;
  Dgp_Int32 maskOffset;
  Dgp_Int8 *buf;
  Dgp_Int32 size;
  Dgp_Int32 width, height;
} ImSubIFD;

#define SUBIFD_POOL_ID DGP_SMALLPOOL_MIN44

#define SUBIFD_SEC_MALLOC(_ppsubifd) \
  DGP_ERR_PASS_CLEANUP(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, SUBIFD_POOL_ID, (void **)(_ppsubifd)))
#define SUBIFD_FREE(_psubifd) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, SUBIFD_POOL_ID, (void *)(_psubifd)))


ErrCode
createTokensNew(Dgp_State		*dgp_state, 
				DgpAdt_List		tokOutput, 
				DgpTdc_SectionBlock *sb, 
				token			**toks,
				Dgp_Int32		*num_non_singletons, 
				Dgp_Bool		final,
				Dgp_Int32		nElemGlobalPadding,
				Dgp_Int32		*nElemPerPagePadding);

ErrCode
makeSubIFDs(Dgp_State *dgp_state, 
            tokpos* position, 
            Dgp_Int32 *num_positions, 
            color_tag *colors, 
            DgpAdt_List subIFDims);

ErrCode
compressRgbImageJpeg(Dgp_State *dgp_state,
                     DgpImage_ColorRGB im, 
                     Dgp_Int32 xres, 
                     Dgp_Int32 yres, 
                     Dgp_Float32 scaleFactor,
                     Dgp_Int32 quality,
                     void **_cinfo,
                     Dgp_Int8 **buf, 
                     Dgp_Int32 *size);

ErrCode
trimImageRGB(DgpCC_ColorCC* sampCC, 
              Dgp_State *dgp_state);

ErrCode
fillUnderImageMask(Dgp_State *dgp_state, 
                   DgpImage_ColorRGB im, 
                   DgpImage_Packed mask, 
                   Dgp_Int32 maskOffset);



ErrCode
handleLargeImages(tdoc_state *cookie, 
                  DgpCC_ColorCC **_cc, 
                  Dgp_Int32 width,
                  Dgp_Int32 height,
                  Dgp_Int32 flags);


typedef struct _tdoc_image
{
  Dgp_UInt32 starty;
  Dgp_UInt32 endy;
  DgpCC_ColorCC *pcc;
} tdoc_image;

ErrCode
createJpeg(tdoc_state *cookie, 
                  tdoc_image* im_array, 
                  Dgp_Int32 first_image,
                  Dgp_Int32 last_image);
                  
#ifdef __cplusplus
}
#endif

#endif /*__DICT_UTIL_H__ */
