/* include-ipcore */
#include <libiow.pub>
#include <librastw.pub>
#include <props.pub>
#include <frames.pub>

#include <image.h>
#include "image_defs.h"
#include "dict-util.h"
#include <colorCC.h>
#include <util.h>
#include <limits.h>
#include "token-compress-priv.h"
#include <insertbox.h>
#include "run2bitmap.h"
#include "simple_utils.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcompdoc/imgOps.c,v 1.20 2003/02/10 22:53:22 mkrocken Exp $");

/*
 * Try to trim all the excess background off the image
 * Making it more rectangular in the process
 *
 * The first step is finding the real bounding box of the image
 *
 * Then go through the runs and alter them so they start and end at
 * this bounding box.
 * 
 * Adust the offset of the cc to indicate this shift of the bounding box.
 * and adjust the image. (don't know how yet)
 */

#define DEFAULT_BG_COL_R 0xFF
#define DEFAULT_BG_COL_G 0xFF
#define DEFAULT_BG_COL_B 0xFF

ErrCode
trimImageRGB(DgpCC_ColorCC* sampCC, Dgp_State *dgp_state)
{
  DGP_PROLOGUE("trimImageRGB()", DGP_PERRSTATE);

  run_elt *runs = (run_elt *)sampCC->runList;
  Dgp_Int32 num_runs = sampCC->numRuns;
  DgpImage_ColorRGB im = sampCC->colorInfo.rgbIm;
  DgpImage_ColorRGB imTrimmed = (DgpImage_ColorRGB)NULL;
  DgpImage_RGB *trimPtr, *oldPtr;
  Dgp_Int32 runLeft, run_len, lastRunRight;
  Dgp_Int32 i,j;
  Dgp_Int32 x,x2,y, width, height;
  Dgp_Int32 ulX, ulY, lrX, lrY;
  DgpAdt_List deleteRuns = DGP_NULLLIST;
  DgpAdt_ListNode ln;
  Dgp_Int32 removedRun;
  Dgp_Int32 curIndex;


  width = DGPIMG_WIDTH(im);
  height = DGPIMG_HEIGHT(im);
  ulX = width;
  ulY = height;
  lrX = 0;
  lrY = 0;

  for(y=0; y<height; y++) {
    /* Scan right for a valid pixel */
    oldPtr = DGPIMG_PIXPTR(im, 0, y);
    for(x=0; x<width; x++) {
      if((oldPtr->r != DEFAULT_BG_COL_R) ||
         (oldPtr->g != DEFAULT_BG_COL_G) ||
         (oldPtr->b != DEFAULT_BG_COL_B)) {
        
        if(x < ulX) ulX = x;
        break;

      }
      oldPtr++;
    }

    if(x != width) {
      /* The line has a valid pixel in it */
      if(y < ulY) {
        ulY = y;
      } else if(y > lrY) {
        lrY = y;
      }

      /* Scan left for a corresponding valid pixel */
      oldPtr = DGPIMG_PIXPTR(im, width-1, y);
      for(x2=width-1; x2>=x; x2--) {
        if((oldPtr->r != DEFAULT_BG_COL_R) ||
           (oldPtr->g != DEFAULT_BG_COL_G) ||
           (oldPtr->b != DEFAULT_BG_COL_B)) {
        
          if(x2 > lrX) lrX = x2;
          break;

        }

        oldPtr--;
      }
    }
  }

  if((lrX == 0)  ||
    (lrY == 0)) {
    /* image has no valid pixels */
    /* delete it */
    sampCC->lrX = 0;
    sampCC->ulX = 1;
    sampCC->lrY = 0;
    sampCC->ulY = 1;

    NORMAL_RETURN;
  }

  lrX++;
  lrY++;
  if(ulX == 0 &&
     ulY == 0 &&
     (sampCC->lrX - sampCC->ulX + 1)  == lrX &&
     (sampCC->lrY - sampCC->ulY + 1) == lrY) {
    
    NORMAL_RETURN;
  }

  /* we need to trim the image */
  width = lrX - ulX;
  DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_RGB | DGPIMAGE_DONTCLEAR, 
    width, (lrY - ulY), (void **)&imTrimmed));

  trimPtr = DGPIMG_PIXPTR(imTrimmed, 0,0);
  for(y=ulY; y<lrY; y++) {
    oldPtr = DGPIMG_PIXPTR(im, ulX, y);
    memcpy(trimPtr, oldPtr, width * sizeof(DgpImage_RGB));
    trimPtr += width;
  }

#ifdef DGP_DBUG
    DgpImage_save (im, "untrimmed.pnm");
    DgpImage_save (imTrimmed, "trimmed.pnm");
#endif

  DGPIMAGE_FREE(im);
  im = NULL;
  sampCC->colorInfo.rgbIm = imTrimmed;

  /* we need to trim the runs */
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&deleteRuns, dgp_state));

  y = 0;
  lastRunRight = -1;
  removedRun = DGP_FALSE;
  for (i=0; i < num_runs; i++) {
    
    runLeft = RunX(runs,i);
    run_len = RunLen(runs,i);

    if( run_len == 0) {
      y = runLeft;
      lastRunRight = -1;
      if(y < ulY || y >= lrY) {
        removedRun = DGP_TRUE;
        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(deleteRuns, (void *)i, NULL));
      }
      RunX(runs,i) = (Dgp_UInt16)(y - ulY);
      
      continue;
    }

    if (runLeft <= lastRunRight) {
      y++;
    } 
    lastRunRight = runLeft + run_len;

    if(y < ulY || y >= lrY) {
      removedRun = DGP_TRUE;
      DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(deleteRuns, (void *)i, NULL));
      continue;
    }
    
    if(runLeft >= lrX) {
      removedRun = DGP_TRUE;
      DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(deleteRuns, (void *)i, NULL));
      continue;
    }

    if(lastRunRight <= ulX) {
      removedRun = DGP_TRUE;
      DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(deleteRuns, (void *)i, NULL));
      continue;
    }

    if(removedRun) {
      ln = deleteRuns->last;
      RunX(runs,(Dgp_Int32)DGPADT_LIGET(ln)) = (Dgp_UInt16)(y - ulY);
      RunLen(runs, (Dgp_Int32)DGPADT_LIGET(ln)) = 0;
      DgpAdt_liRem(ln);
      removedRun = DGP_FALSE;
    }
    
    if(runLeft < ulX)
      runLeft = ulX;
    
    if(lastRunRight > lrX)
      lastRunRight = lrX;
    
    /* set the new length */
    RunLen(runs, i) = (Dgp_UInt16)(lastRunRight - runLeft);
    
    /* adjust to the new offset; */
    RunX(runs, i) = (Dgp_UInt16)(runLeft - ulX);
    
  }
  sampCC->lrX = sampCC->ulX + lrX - 1;
  sampCC->lrY = sampCC->ulY + lrY - 1;

  sampCC->ulX += (Dgp_UInt16)ulX;
  sampCC->ulY += (Dgp_UInt16)ulY;

  /* Need to realloc runs and copy to the new. */
  ln = DGPADT_LIFIRST(deleteRuns);
  if(ln == DGP_NULLLISTNODE) {
    DgpAdt_liFree(deleteRuns);
    deleteRuns = DGP_NULLLIST;
    NORMAL_RETURN;
  }
  curIndex= (Dgp_Int32)DGPADT_LIGET(ln);

  i=curIndex;
  for(j = curIndex; j < num_runs; j++) {  
    if(j == curIndex) {
      ln = DGPADT_LINEXT(ln);
      if(ln == DGP_NULLLISTNODE) {
        curIndex = -1;
      } else {
        curIndex = (Dgp_Int32)DGPADT_LIGET(ln);
      }
      
      continue;
    }
    RunX(runs, i) = RunX(runs, j);
    RunLen(runs, i) = RunLen(runs, j);
    i++;
  }

  DgpAdt_liFree(deleteRuns);
  deleteRuns = DGP_NULLLIST;


  /* sometimes the runs cumming out of libsupercc are wrong.
     We try to catch those cases here with this special code.
     The solution is to create completely new runs. */
  height = DGPIMG_HEIGHT(imTrimmed);
  if (i >= height)
  {
  DGP_SEC_REALLOC(sampCC->runList, (Dgp_UInt16 *), (i*2*sizeof(Dgp_UInt16)));
  sampCC->numRuns = i;
  }
  else
  {
	  /* we need to create new runs.
	     this is the simplest possible set of runs ... */
	  width = DGPIMG_WIDTH(imTrimmed);
	  DGP_SEC_REALLOC(sampCC->runList, (Dgp_UInt16 *), (height*2*sizeof(Dgp_UInt16)));
	  sampCC->numRuns = height;
	  
	  for(y=0; y<height; y++) {
		  RunX (sampCC->runList, y) = (Dgp_UInt16) 0;
		  RunLen(sampCC->runList, y) = (Dgp_UInt16) width;
	  }
  }

  NORMAL_RETURN;

ERR_CLEANUP:

  if(imTrimmed != NULL) {
    DGPIMAGE_FREE(imTrimmed);
  }

  if(deleteRuns != DGP_NULLLIST) DgpAdt_liFree(deleteRuns);


  ERR_RETURN;
}

#define cmyk_c(color) \
  ((color & 0xFF000000) >> 24)
#define cmyk_m(color) \
  ((color & 0x00FF0000) >> 16)
#define cmyk_y(color) \
  ((color & 0x0000FF00) >> 8)
#define cmyk_k(color) \
  (color & 0x000000FF)

typedef DgpImage_RGB Color;

INLINE Color avgWeightedColor(Dgp_Int32 up, Color upCol, Dgp_Int32 down, Color downCol, 
                Dgp_Int32 right, Color rightCol, Dgp_Int32 left, Color leftCol,
                Dgp_Int32 total)
{
  Dgp_UInt32 edges[4];
  Color edge_cols[4];
  Color result = {0, 0, 0};
  Dgp_Int32 i, total2;
  Dgp_Int32 retCol[3];


  retCol[0] = retCol[1] = retCol[2] = 0;

  if(up == 0)
    up = INT_MAX;

  edges[0] = up;
  edge_cols[0] = upCol;
  edges[1] = down;
  edge_cols[1] = downCol;
  edges[2] = right;
  edge_cols[2] = rightCol;
  edges[3] = left;
  edge_cols[3] = leftCol;

  total2 = 0;
  for(i = 0; i < 4; i++) {
    retCol[0] += edge_cols[i].r * (total / edges[i]);
    retCol[1] += edge_cols[i].g * (total / edges[i]);
    retCol[2] += edge_cols[i].b * (total / edges[i]);
    total2 += total / edges[i];
  }

  if (total2 == 0)
    return result;

  result.r = ((retCol[0] / total2) & 0xFF);
  result.g = ((retCol[1] / total2) & 0xFF);
  result.b = ((retCol[2] / total2) & 0xFF);

  return result;
}

#define FRAME 32

typedef struct{
  Color col;
  Dgp_Int32 dist;
  Dgp_Int8 inc;
} ColorVect;

ErrCode
fillUnderImageMask(Dgp_State *dgp_state, 
                   DgpImage_ColorRGB im, 
                   DgpImage_Packed mask, 
                   Dgp_Int32 maskOffset)
{
  DGP_PROLOGUE("fillUnderImageMask()", DGP_PERRSTATE);

  Dgp_Int32 x, y, width, height;
  Dgp_Int32 up, down = 0, left, right;
  Color upCol, downCol, leftCol, rightCol;
  Dgp_Int32 xMaskStart, xMask, yMask;
  Dgp_Int32 yMaskUpperLimit;
  Dgp_Int32 xMaskUpperLimit;
  Dgp_Int32 offset;
  Dgp_Int32 incLeft, decRight;
  ColorVect *upperRecord = NULL, *lowerRecord = NULL;
  Dgp_Int32 maxTotDist;

  width = DGPIMG_WIDTH(im);
  height = DGPIMG_HEIGHT(im);
  maxTotDist = width + height + 2;
  xMaskStart = maskOffset + FRAME;
  yMaskUpperLimit = height + FRAME - 1;
  xMaskUpperLimit = width + xMaskStart - 1;
  DGP_SEC_CALLOC(upperRecord, (ColorVect *), width, sizeof(ColorVect));
  DGP_SEC_CALLOC(lowerRecord, (ColorVect *), width, sizeof(ColorVect));


  for(y=0, yMask = FRAME; y<height; y++, yMask++) {
    left = INT_MAX;
    incLeft = 0;
    right = 0;
    decRight = 1;
    for(x=0, xMask = xMaskStart; x<width; x++, xMask++) {
      
      if(right == 0) {
        /* The pixel just to the left is valid, we need to find the  */
        /* next valid right pixel */

        offset = xMask;
        while((offset <= xMaskUpperLimit) &&
          !DGPIMG_GETBIT(mask, offset, yMask)) {
          offset++;
        }

        right = offset - xMask;
        if(x+right < width) {
          rightCol = DGPIMG_REF(im, x + right, y);
        } else {
          /* If the are no more valid pixels on the right */
          /* set right to the max and make sure it doesn't change */
          decRight = 0;
          right = INT_MAX;
        }

      } else {
        right -= decRight; 
      }

      if(right == 0) {
        /* We are on a valid pixel */
        leftCol = DGPIMG_REF(im, x, y);
        left = 0;
        /* Don't start incrementing left until a real pixel 
         * has been seen
         */
        incLeft = 1;

        upperRecord[x].col = leftCol;
        upperRecord[x].dist = 0;
        upperRecord[x].inc = 1;

        continue;
      }
    
      if(lowerRecord[x].dist == 1) {
        offset = yMask;
        while((offset <= yMaskUpperLimit) &&
          !DGPIMG_GETBIT(mask, xMask, offset)) {
          offset++;
        }

        if(offset <= yMaskUpperLimit) {
          lowerRecord[x].dist = offset - yMask;
          lowerRecord[x].col = DGPIMG_REF(im, x, y+down);
        } else {
          lowerRecord[x].inc = 1;
          lowerRecord[x].dist = INT_MAX;
        }
      } else {
        lowerRecord[x].dist -= (1 - lowerRecord[x].inc);
      }

      left += incLeft;
      upperRecord[x].dist += upperRecord[x].inc;
      up = upperRecord[x].dist;
      upCol = upperRecord[x].col;
      downCol = lowerRecord[x].col;
      down = lowerRecord[x].dist;

      DGPIMG_REF(im, x, y) = avgWeightedColor(up, upCol, down, downCol, 
        right, rightCol, left, leftCol, maxTotDist);
    }
  }

  DGP_FREE(upperRecord);
  upperRecord = NULL;
  DGP_FREE(lowerRecord);
  lowerRecord = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(upperRecord != NULL) DGP_FREE(upperRecord);
  if(lowerRecord != NULL) DGP_FREE(lowerRecord);

  ERR_RETURN;

}


ErrCode
handleLargeImages(tdoc_state *cookie, 
                  DgpCC_ColorCC **_cc, 
                  Dgp_Int32 width,
                  Dgp_Int32 height,
                  Dgp_Int32 flags)
{
  Dgp_State *dgp_state = cookie->dgp_state; 
  DGP_PROLOGUE("handleLargeImages()", DGP_PERRSTATE);
  
  DgpCC_ColorCC *cc = *_cc;
  ImSubIFD *subIFD = NULL;
  Dgp_Int32 jpQuality;
  Dgp_Float32 scaleFactor;


  /* mine! */
  *_cc = NULL;

  if(cookie->curPage.largeImages == DGP_NULLLIST) {
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&(cookie->curPage.largeImages), dgp_state));
  }

  SUBIFD_SEC_MALLOC(&subIFD);
  subIFD->xpos = cc->ulX;
  subIFD->ypos = cc->lrY;
  subIFD->samptok = NULL;
  subIFD->mask = NULL;
  subIFD->buf = NULL;
  subIFD->width = width;
  subIFD->height = height;

  if(!(flags & SAMPTOK_IS_RECT)) {
    token tok;
    DgpImage_Packed maskImage;
    tokpos framePos = {NULL, NULL, cFrameBits, cFrameBits};
    Dgp_Int32 offset;

    tok.width = (Dgp_UInt16)width;
    tok.height = (Dgp_UInt16)height;
    tok.uses = 1;
    tok.longsperline = (width + DGP_BITS_UINT32 - 1)/DGP_BITS_UINT32;
    DGP_ERR_PASS_CLEANUP(RunsToUnframedBitmap(&(tok.bits),
      cc->runList, cc->numRuns, height, tok.longsperline,
      0, dgp_state));
   
    /* Need to make sure we take into account pixr alignment issues */

    DGP_ERR_PASS_CLEANUP(pimNewFrameAligned(dgp_state, width,
      height, &offset, &maskImage));
    subIFD->mask = maskImage;
    framePos.xpos += (Dgp_Int16)offset;
    framePos.tok = NULL;
    subIFD->maskOffset = offset;
    InsertBoxPacked(maskImage, &framePos, &tok, DGP_FALSE);
    DGP_FREE(tok.bits);
    DGP_ERR_PASS_CLEANUP(fillUnderImageMask(dgp_state, 
      cc->colorInfo.rgbIm, maskImage, offset));

  }          

  switch(cookie->sampCompType) {
  case DGPTDC_SAMP_COMPTYPE_JPEG:
#ifdef DGP_FAXMODE
    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else

    /* only scale if it's big */
    if((width > (Dgp_Int32)(cookie->curPage.width * LARGE_CC_PERCENT)) && 
       (height > (Dgp_Int32)(cookie->curPage.height * LARGE_CC_PERCENT))) {
      scaleFactor = ((DgpTdc_JpegCompInfo *)cookie->sampCompInfo)->scaleFactor;
    } else {
      scaleFactor = 1.0;
    }
    jpQuality = ((DgpTdc_JpegCompInfo *)cookie->sampCompInfo)->jpQuality;

    if(colorcc_isWash(cc)) {
      jpQuality = WASH_JPQUALITY(jpQuality);
    }

    DGP_ERR_PASS_CLEANUP(compressRgbImageJpeg(dgp_state, cc->colorInfo.rgbIm, 
      cookie->curPage.xRes, cookie->curPage.yRes, scaleFactor, jpQuality, 
      &(cookie->curBlock->cinfo), &(subIFD->buf), &(subIFD->size)));
#endif
    break;

  default:

    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
  
  }

  DGPIMAGE_FREE(cc->colorInfo.rgbIm);
  cc->colorInfo.rgbIm = NULL;

  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(cookie->curPage.largeImages, subIFD, NULL));
  subIFD = NULL;

  /* ditch the cc */
  DGP_FREE(cc->runList);
  cc->runList = NULL;
  COLORCC_FREE(cc);
  cc = NULL;


  NORMAL_RETURN;

ERR_CLEANUP:

  if(cc != NULL) colorcc_cleanup(cc);

  if(subIFD != NULL) {
    if(subIFD->mask != NULL) {
      DGPIMAGE_FREE(subIFD->mask);
    }
    if(subIFD->buf != NULL) {
      DGP_FREE(subIFD->buf);
    }
    SUBIFD_FREE(subIFD);
  }

  ERR_RETURN;
}

ErrCode
createJpeg(tdoc_state *cookie, 
                  tdoc_image *array, 
                  Dgp_Int32 first_image,
                  Dgp_Int32 last_image)
{
  Dgp_State *dgp_state = cookie->dgp_state; 
  DGP_PROLOGUE("handleLargeImages()", DGP_PERRSTATE);
  
  ImSubIFD *subIFD = NULL;
  Dgp_Int32 jpQuality;
  Dgp_Float32 scaleFactor;
  tdoc_image *pim;
  Dgp_Int32 count;
  Dgp_Int32 minX, maxX, minY, maxY;
  DgpCC_ColorCC *pcc;
  DgpImage_ColorRGB image;

    token tok;
    DgpImage_Packed maskImage;
    tokpos framePos = {NULL, NULL, cFrameBits, cFrameBits};
    Dgp_Int32 offset;

  minX = minY = 1000000;
  maxX = maxY = 0;

  /* get smallest box that holds all images */
  for (count = first_image; count <= last_image; count++)
  {
    pcc = array[count].pcc;
    if (pcc->ulX < minX) minX = pcc->ulX;
    if (pcc->ulY < minY) minY = pcc->ulY;
    if (pcc->lrX > maxX) maxX = pcc->lrX;
    if (pcc->lrY > maxY) maxY = pcc->lrY;
  }

  if(cookie->curPage.largeImages == DGP_NULLLIST) {
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&(cookie->curPage.largeImages), dgp_state));
  }


  SUBIFD_SEC_MALLOC(&subIFD);
  subIFD->xpos = (Dgp_UInt16) minX;
  subIFD->ypos = (Dgp_UInt16) maxY;
  subIFD->samptok = NULL;
  subIFD->buf = NULL;
  subIFD->width = maxX - minX + 1;
  subIFD->height = maxY - minY + 1;

  /* allocate mask image */
  DGP_ERR_PASS_CLEANUP(pimNewFrameAligned(dgp_state, subIFD->width,
      subIFD->height, &offset, &maskImage));
  subIFD->mask = maskImage;
  subIFD->maskOffset = offset;

  /* allocate combined image */
  DGP_ERR_PASS_CLEANUP(DgpImage_new (dgp_state, DGPIMAGE_RGB, subIFD->width, 
				     subIFD->height, (void **)&image));


  for (count = first_image; count <= last_image; count++)
  {
    pim = &array[count];

    tok.width = (Dgp_UInt16) (pim->pcc->lrX - pim->pcc->ulX + 1);
    tok.height = (Dgp_UInt16) (pim->pcc->lrY - pim->pcc->ulY + 1);
    tok.uses = 1;
    tok.longsperline = (tok.width + DGP_BITS_UINT32 - 1)/DGP_BITS_UINT32;

    /* create a binary mask, wrapped in a token struct */
    DGP_ERR_PASS_CLEANUP(RunsToUnframedBitmap(&(tok.bits),
      pim->pcc->runList, 
      pim->pcc->numRuns, tok.height, tok.longsperline,
      0, dgp_state));
   
    /* Need to make sure we take into account pixr alignment issues */

    framePos.xpos = (Dgp_Int16) (pim->pcc->ulX - minX);
    framePos.ypos = (Dgp_Int16) (pim->pcc->ulY - minY);
    framePos.tok = NULL;
    InsertSampledBox(image, pim->pcc->colorInfo.rgbIm, &framePos, &tok);

    /* mask alignment issues (PIXR frame etc.) */
    framePos.xpos += (Dgp_Int16) (offset + offset);
    framePos.ypos += (Dgp_Int16) offset;
    InsertBoxPacked(maskImage, &framePos, &tok, DGP_FALSE);
    DGP_FREE(tok.bits);


  DGPIMAGE_FREE(pim->pcc->colorInfo.rgbIm);
  pim->pcc->colorInfo.rgbIm = NULL;

  /* ditch the cc */
  DGP_FREE(pim->pcc->runList);
  pim->pcc->runList = NULL;
  COLORCC_FREE(pim->pcc);
  pim->pcc = NULL;

  }

#ifdef DGP_DBUG
    DgpImage_save (image, "image.pnm");
    DgpImage_save (maskImage, "mask.pnm");
#endif

    DGP_ERR_PASS_CLEANUP(fillUnderImageMask(dgp_state, 
      image, maskImage, offset));

#ifdef DGP_DBUG
    DgpImage_save (image, "filled_image.pnm");
#endif    

    /***********************
			   COMPRESSION
                                        ************************************/

  switch(cookie->sampCompType) {
  case DGPTDC_SAMP_COMPTYPE_JPEG:
#ifdef DGP_FAXMODE
    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else

    /* only scale if it's big */
    if((subIFD->width > (Dgp_Int32)(cookie->curPage.width * LARGE_CC_PERCENT)) && 
       (subIFD->height > (Dgp_Int32)(cookie->curPage.height * LARGE_CC_PERCENT))) {
      scaleFactor = ((DgpTdc_JpegCompInfo *)cookie->sampCompInfo)->scaleFactor;
    } else {
      scaleFactor = 1.0;
    }
    jpQuality = ((DgpTdc_JpegCompInfo *)cookie->sampCompInfo)->jpQuality;

    DGP_ERR_PASS_CLEANUP(compressRgbImageJpeg(dgp_state, 
      image, 
      cookie->curPage.xRes, cookie->curPage.yRes, scaleFactor, jpQuality, 
      &(cookie->curBlock->cinfo), &(subIFD->buf), &(subIFD->size)));

	DGPIMAGE_FREE(image);
    
#endif
	break;

  default:

    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
  
  }

  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(cookie->curPage.largeImages, subIFD, NULL));
  subIFD = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(array[first_image].pcc != NULL) colorcc_cleanup(array[first_image].pcc);

  if(subIFD != NULL) {
    if(subIFD->mask != NULL) {
      DGPIMAGE_FREE(subIFD->mask);
    }
    if(subIFD->buf != NULL) {
      DGP_FREE(subIFD->buf);
    }
    SUBIFD_FREE(subIFD);
  }

  ERR_RETURN;
}
