#ifndef __TOKEN_COMPRESS_PRIV_H__
#define __TOKEN_COMPRESS_PRIV_H__

#include "token-compress.h"
#include "iomrc.h"
#include "iowrap.h"
#include <dgp_state.h>
#include <tok-lossless.h>
#include <colorCC.h>
#if DGPTDC_DO_HALFTONE
#include "hamtok.h"
#endif

#define TIFFFX_MAX_DICT 5

typedef struct {
  void* stream;
  IO_INSTHANDLE inst;
  IO_DOCHANDLE  doc;
  IO_PAGEHANDLE page;
  Dgp_Int32 dict_list[TIFFFX_MAX_DICT];
  Dgp_Int32 n_dict;
} IO_State;

typedef struct {
  Dgp_State *dgp_state;
  Dgp_Int32 numPgs;
  void *dictblock;
  IO_State* IOState;
  void *jbig2State;
  Dgp_Bool ascii_ext;
} tout_state;

typedef struct {
  Dgp_State *dgp_state;
  
  Dgp_Bool lossy;
  Dgp_Bool color;
  Dgp_Bool noTokenImages;
  Dgp_Bool sepLargeImages;
  Dgp_Int32 halftone;

  Dgp_Int32 sizeThreshold;
  void *tokenizerParams;
  void *colorTokenizerParams;
  LlTokState llTokState;
  SampTokState samptokstate;
  Dgp_Int32 sampCompType;
  void *sampCompInfo;

#if DGPTDC_DO_HALFTONE
  pHalfTokState halfTokState;  /*the halftone tokenizer  */
#endif

  /* Section stuff */
  DgpTdc_PageInfo curPage;

  DgpCC_ColorCookie cccc;
  Dgp_UInt16 washThreshold;

  DgpTdc_SectionBlock *curBlock;

  /* 16.4.2001 / dhtesq. */
  /* Added... */
  LlTokOptions *llTokenizerOptions;
} tdoc_state;

/* the quality of wash images */
#define WASH_JPQUALITY(_qual) ((Dgp_Int32)((_qual)/(Dgp_Float32)2.5))


/* get colorCC states */
#define colorcc_isSampled(colorcc) ((colorcc)->state.sampled)
#define colorcc_isWash(colorcc) ((colorcc)->state.wash)
#define colorcc_isAbove(colorcc) ((colorcc)->state.above)
#define colorcc_isWashOrSampled(colorcc) (colorcc_isWash(colorcc) || colorcc_isSampled(colorcc))

#define COLORCC_FREE(_pnewcc) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, DGPCC_COLORCC_POOL, (void *)(_pnewcc)))

#define colorcc_cleanup(cc) \
do { \
  if((colorcc_isWashOrSampled(cc)) && ((cc)->colorInfo.rgbIm != NULL)) { \
    DGPIMAGE_FREE((cc)->colorInfo.rgbIm); \
    (cc)->colorInfo.rgbIm = NULL; \
  } \
  if((cc)->runList != NULL) { \
    DGP_FREE((cc)->runList); \
    (cc)->runList = NULL; \
  } \
  COLORCC_FREE(cc); \
  (cc) = NULL; \
} while(0)

#define LARGE_CC_PERCENT (Dgp_Float32)(0.1)


#endif /* __TOKEN_COMPRESS_PRIV_H__ */
