#ifndef __TOKEN_COMPRESS_H__
#define __TOKEN_COMPRESS_H__

#include "dgp_types.h"  /* digipaper data types defs */
#include <annotation.h> /* for the Annotation struct */
#include <mmcompress.h> /* for the tokpos and token structs */
#include <uds.h>  /* for uds struct passed to tout_init */
#include <color-extension.h> /* for the color_tag struct  */
#include <colorCC.h> /* for the input callback for color images */

/* include-libwjr */
#include <list.h> /* for the list struct */
#include "tokenCC.h"

#ifdef __cplusplus
extern "C"
{
#endif
 

/* at this point we don't want halftone stuff (it doesn't really work...) */
#define DGPTDC_DO_HALFTONE 0


#define DGPTDC_RECOMMEND_COLOR_POOLS \
  (DGP_USE_SMALLPOOL_MIN8 | DGP_USE_SMALLPOOL_MIN16 | DGP_USE_SMALLPOOL_MIN24 | \
   DGP_USE_SMALLPOOL_MIN44 | DGP_USE_SMALLPOOL_MIN128)

#define DGPTDC_RECOMMEND_MONO_POOLS \
  (DGP_USE_SMALLPOOL_MIN16 | DGP_USE_SMALLPOOL_MIN28 | DGP_USE_SMALLPOOL_MIN128)

#define DGPTDC_SAMPTOK_NEED_MASK  SAMPTOK_USER_FLAG1
#define DGPTDC_SAMPTOK_IS_TOKEN   SAMPTOK_USER_FLAG2
#define DGPTDC_SAMPTOK_IS_WASH    SAMPTOK_USER_FLAG3

#define DGPTDC_SAMP_COMPTYPE_JPEG 0

#define addLlPage(curPage, sb) \
do { \
  if((sb)->numPages >= (sb)->numPagesAlloced) { \
    Dgp_Int32 temp_int =  ((sb)->numPages * 3) / 2; \
    DGP_SEC_REALLOC((sb)->pages, (DgpTdc_PageInfo *), temp_int *sizeof(DgpTdc_PageInfo)); \
    (sb)->numPagesAlloced = temp_int; \
  } \
  (curPage).index = (sb)->numPages; \
  (sb)->pages[(sb)->numPages] = (curPage); \
  (sb)->numPages++; \
} while(0)


typedef struct {
  Dgp_Float32 scaleFactor;
  Dgp_Int32 jpQuality;
} DgpTdc_JpegCompInfo;


typedef struct {
  /* General */
  Dgp_Int32 pageNum;
  Dgp_Int32 index;
  Dgp_Int32 numBoxes;
  tokpos *boxes;
  Dgp_Int32 width, height;
  Dgp_Int32 xRes, yRes;

  /* Color info */
  Dgp_Bool color;
  color_tag *colors;
  DgpAdt_List largeImages;

  /* halftone info */
  tokpos *halftone_boxes;
  Dgp_Int32 halftone_numBoxes;
  DgpAdt_List outliers;
  Dgp_Int32 cell_size;
 
  /*Annotation information  */
  Dgp_Int32 num_annotations; 
  Annotation *annotations; 
 
  /* Per page dictionary info */
  token *toks;
  Dgp_Int32 numToks;
  Dgp_Int32 numToksAlloced;

  Dgp_Int32 memUsage;
  Dgp_Int32 tokMem;
  Dgp_Int32 pageTokMem;
} DgpTdc_PageInfo;

typedef struct {

  /* filled by tdoc */
  Dgp_Int32 numPages;
  Dgp_Int32 numPagesAlloced;
  DgpTdc_PageInfo *pages;

  DgpCC_PosInfoBlock *pBlock;

  /* Dictionary stuff */
  DgpCC_TokInfoBlock *tBlock;  /*Block containing the TokInfos pointed to by the positions */
  Dgp_Int32 numOldToks;  /* Number of TokInfos in above block that are from previously written sections */
  DgpAdt_List classesFound;  /* list of tokClass structs */

  /* Even elements are token arrays of previously 
       written tokens.  Odd elements are the size of
     those arrays.
     */
  DgpAdt_List tokBlockList;  

  /* The Images,  */
  Dgp_Int32 sampCompType;
  void *sampCompInfo;
  
  /* compress struct for jpeg compression */
  void *cinfo;

  /* Halftone info */
  DgpAdt_List halfOutput;
  
  Dgp_Int32 tokMem;
  Dgp_Bool final;

} DgpTdc_SectionBlock;

typedef void *DgpTdc_Cookie;
typedef void *DgpTdc_WriteCookie;

/*
 * ---TDOC---
 * The token finding and token matching part of the interface
 *
 * All fuctions have a report arg.  If this is true some trace info
 * will be printed with the rptPrinter in the tc_cookie
 */

/*
 * Initialize TDOC
 *
 * Setting lossy, color, or halftone to DGP_TRUE,
 * allows the possible use of these qualities.
 * Currently none of the params are used
 *
 * PRE-  
 *   dgp_state needs to initialized with DgpInitState
 *   _cookie needs to point to a allocated DgpTdc_Cookie     
 *   
 * 
 */
DGP_LIBEXPORT ErrCode
DgpTdc_initDoc(Dgp_State *dgp_state,  
               DgpTdc_Cookie *cookie, 
               Dgp_Bool lossy, 
               Dgp_Bool color, 
               Dgp_Int32 halftone,
               Dgp_Int32 sizeThreshold,  
               Dgp_UInt16 washThreshold,
               Dgp_Int32 sampCompType, 
               void *sampCompInfo, 
               void *tokenizerParams, 
               void *colorTokenizerParams, 
               Dgp_Bool report);


/*
 * Begin a section
 *   Sections are blocks of pages that have a common dictionary
 * 
 * Use null for prev the first time this is called.
 * If it is called more than once, the old DgpTdc_SectionBlock should be passed in.
 * (Especially if the previous section had a partial dictionary)
 *
 *
 * PRE-
 *    cookie needs to be initialized by tdoc_init
 *    prev needs to be NULL or point to an allocated initialized section block
 *       that has been passed to tout_writeSection
 */
DGP_LIBEXPORT ErrCode
DgpTdc_startSection(DgpTdc_Cookie cookie, 
                    DgpTdc_SectionBlock *prev, 
                    Dgp_Bool report);


/*
 * Add a page to the section
 * as usual image must be padded by 32 pixels on each side
 *
 * this is the only function that needs to be called to add a monochrome page
 *
 * PRE -
 *   cookie needs to be initialized by tdoc_init
 *   image needs to be the data of an initialized ipcore binary image. 
 */
DGP_LIBEXPORT ErrCode
DgpTdc_addMonoPage(DgpTdc_Cookie cookie, 
                   Dgp_Int32 pageNumber, 
                   UInt8 *image, 
                   Dgp_Int32 pageWidth,
                   Dgp_Int32 pageHeight, 
                   Dgp_Int32 bpl, 
                   Dgp_Int32 xRes, 
                   Dgp_Int32 yRes, 
                   Dgp_Bool report);

/*
 * Prepare for a color page
 *
 * compType determines the compression used for sampled images
 *   The only option right now is jpeg
 *
 * compInfo is the information for the compression type.
 *
 * readRun is a function used by the CC finder to return a run of pixels
 *    the definition is in colorCC.h  (The cookie passed to it is given 
 *    in the next function: addStrip)
 * PRE -
 *   cookie needs to be initialized by tdoc_init
 *   compInfo (see above)
 */
DGP_LIBEXPORT ErrCode
DgpTdc_startColorPage(DgpTdc_Cookie cookie, 
                      Dgp_Int32 pageNumber, 
                      Dgp_Int32 pageWidth,
                      Dgp_Int32 pageHeight, 
                      Dgp_Int32 xRes, 
                      Dgp_Int32 yRes,
                      Dgp_Bool noTokenImages, 
                      Dgp_Bool sepLargeImages,
                      Dgp_Bool report);


/*
 * Send a strip of image data for the readRun function
 * 
 *  stripData -- will be passed to the readRun function passed
 *     to tdoc_startColorPage.
 *
 *
 * PRE -
 *   cookie needs to be initialized by tdoc_init
 *   pageNumber tdoc_addColorPage must be called with this same page number
 *   stripData must be setup according to how the readRun function wants it
 *
*/
DGP_LIBEXPORT ErrCode
DgpTdc_addColorStrip(DgpTdc_Cookie cookie,
                     Dgp_Int32 pageNumber, 
                     Dgp_Int32 numLines, 
                     DgpCC_PFNReadRun readRun, 
                     void *stripData,
                     Dgp_Bool report);

/*
 * Complete the color page 
 *
 *
 * PRE -
 *   tdoc_startPageColor be called with the same page number
 *   all the lines of the page must be add by calling tdoc_addStrip
 *
*/
DGP_LIBEXPORT ErrCode
DgpTdc_endColorPage(DgpTdc_Cookie cookie, 
                    Dgp_Int32 pageNumber, 
                    Dgp_Bool report);

/*
 *
 *  
 *  PRE -
 *    the most recent tdoc_addMonoPage or tdoc_startColorPage must be
 *    with the same pgNum.  On of these two functions must be called first
 *
 *    tag must be a valid Xiff tag
 *    data must be in the correct XIFF format for the secified "tag" 
 */
DGP_LIBEXPORT ErrCode
DgpTdc_addAnnotation(DgpTdc_Cookie cookie, 
                     Dgp_Int32 pgNum, 
                     void *data, 
                     Dgp_Int32 size,
                     Dgp_Int32 tag,
                     Dgp_Bool report);


/* Finish the section: this will run tokPartial or tokFinal
 * and put all the necessary information for the section in a 
 * "portable" block
 *
 * That can be passed to TOUT
 *
 * PRE-
 *   retSb must be a pointer to an allocated DgpTdc_SectionBlock *
 *
 * POST -
 *    *retSb will now point to a DgpTdc_SectionBlock which contains
 *  the "token compression" data for each page that was 
 *  added, since the last tdoc_startSection call.  
 *  (See DgpTdc_SectionBlock struct for more info about this data)
 */
DGP_LIBEXPORT ErrCode
DgpTdc_endSection(DgpTdc_Cookie cookie, 
                  DgpTdc_SectionBlock **retSb, 
                  Dgp_Bool final, 
                  Dgp_Bool report);

/* Finish the section and rollback one page: 
 * this will run tokPartial and put all the necessary information 
 * for the section in a "portable" block. Then go through the list
 * and pull tokens on the current page out from the list. 
 * Free the current page content and Make previous
 * page the current page
 * That can be passed to TOUT
 *
 * PRE-
 *   retSb must be a pointer to an allocated DgpTdc_SectionBlock *
 *
 * POST -
 *    *retSb will now point to a DgpTdc_SectionBlock which contains
 *  the "token compression" data for each page that was 
 *  added, since the last tdoc_startSection call.  
 *  (See DgpTdc_SectionBlock struct for more info about this data)
 */
DGP_LIBEXPORT ErrCode
DgpTdc_endSectionRollBack(DgpTdc_Cookie cookie, 
                  DgpTdc_SectionBlock **retSb, 
                  Dgp_Bool report);

/* Finish the document:
 * 
 * Free all the TDOC state
 * if sb is not NULL it will also try
 * to free the section block state.
 *
 * PRE -
 *   sb must be NULL or an allocated DgpTdc_SectionBlock
 *
 * POST -
 *   cookie will no longer be valid
 *   sb will be freed
 */
ErrCode
DgpTdc_endDoc(DgpTdc_Cookie *cookie, 
              DgpTdc_SectionBlock *sb);


/*
 * ---TOUT---
 * The Xiff file outputting part of the interface
 *
 * All fuctions have a report arg.  If this is true some trace info
 * will be printed with the rptPrinter in the tc_cookie
 */

/*
 * Prepare TOUT for writting a section
 * The numDicts can be an estimate, but numPgs must be accurate
 * The stream is the where the output will be written.
 *
 * PRE-  same as tdoc_init
 */
ErrCode
DgpTdc_writeInit(Dgp_State *dgp_state, 
                 DgpTdc_WriteCookie *cookie, 
                 Dgp_Int32 numPgs, 
                 Dgp_Int32 numDicts, 
                 uds *stream, 
                 Dgp_Bool ascii_ext, 
                 void *params);


/*
 * Write pos blocks and dictblocks to file
 * The DgpTdc_SectionBlock will be altered a bit after this function
 * The resulting sb should be passed back to the TDOC for correct cleanup
 *
 * PRE -
 *   sb must point to an allocated valid DgpTdc_SectionBlock.  The data in 
 *     this DgpTdc_SectionBlock will be xiff compressed and written to the
 *     stream passed to tout_init.
 *
 * POST -
 *   sb will be modifed.  The sb->tBlock that was passed in will be
 *     reordered.  So the userdata field in the corresponding classes
 *     in the sb->classesFound will be updated.   If the sb->final is 
 *     true all tokClass structs will be removed from classesFound and
 *     deleted.  If the sb->final is false, all single use classes will
 *     be removed from the classesFound list and their structure deleted.
 *     Also if sb->final is false an array of token structures on for
 *     for each non single use class will be added along with its length
 *     to the tokBlockList
 */
DGP_LIBEXPORT ErrCode
DgpTdc_writeSection(DgpTdc_WriteCookie cookie, 
                    DgpTdc_SectionBlock *sb, 
                    Dgp_Bool report);

/*
 * Close the file and free any state
 */
ErrCode
DgpTdc_writeEnd(DgpTdc_WriteCookie *cookie);


#ifdef __cplusplus
}
#endif

#endif /*__TOKEN_COMPRESS_H__ */
