#include <simple_utils.h>  /* for some reason this needs to be first?? */
#include "token-compress.h"
#include "token-compress-priv.h"
#include "dict-util.h"
#include "image_defs.h"

/* include-libwjr */
#include <util.h>

/* include-ipcore */
#include <librastw.pub>
#include <props.pub>
#include <frames.pub>
#include "PsgFix.h"

#include <tok-lossless.h>
#include <colorCC.h>
#include <cc2box.h>
#include <insertbox.h>
#include <tokenCC.h>
#include <run2bitmap.h>

#if DGPTDC_DO_HALFTONE
#include "hamtok.h"
#include "halfutil.h"
#endif

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcompdoc/token-document.c,v 1.76 2003/02/10 22:53:23 mkrocken Exp $");


ErrCode
DgpTdc_initDoc(Dgp_State *dgp_state,  
               DgpTdc_Cookie *_cookie, 
               Dgp_Bool lossy, 
               Dgp_Bool color, 
               Dgp_Int32 halftone,
               Dgp_Int32 sizeThreshold,  
               Dgp_UInt16 washThreshold,
               Dgp_Int32 sampCompType, 
               void *sampCompInfo, 
               void *tokenizerParams, 
               void *colorTokenizerParams, 
               Dgp_Bool report)
{
  DGP_STATE_PROLOGUE2("DgpTdc_initDoc()");

  tdoc_state *cookie = NULL;

  /* validate params */
  DGP_ERR_BEGIN(dgp_state);
  DGP_CHECK_NULL(_cookie);
  *_cookie = NULL;

  /* validate small pool usage */
  DGP_SMALLPOOL_VALIDATE(ImSubIFD, SUBIFD_POOL_ID);

  DGP_SEC_CALLOC(cookie, (tdoc_state *), 1, sizeof(tdoc_state));
  report = DGP_FALSE;

  cookie->dgp_state = dgp_state;
  cookie->lossy = lossy;
  cookie->color = color;
  cookie->sizeThreshold = sizeThreshold;
  cookie->washThreshold = washThreshold;
  cookie->llTokenizerOptions = (LlTokOptions *) tokenizerParams;
  cookie->colorTokenizerParams = colorTokenizerParams;

  cookie->sampCompType = sampCompType;
  cookie->sampCompInfo = sampCompInfo;

  cookie->halftone = halftone;
  cookie->llTokState = NULLLLTOKSTATE;

  /*DGP_REPORTF0(report, "starting document"); */

  cookie->cccc = NULL;
  if(cookie->color) {
    /* set up the color component cookie */
    DGP_ERR_PASS_CLEANUP(DgpCC_init(dgp_state, &(cookie->cccc)));
  }


  *_cookie = cookie;
  cookie = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(cookie != NULL) DgpTdc_endDoc((DgpTdc_Cookie *)&cookie, NULL);

  ERR_RETURN;
}


/* Return TRUE if the mask represented by the given runs is rectangular;
   return FALSE otherwise. */
Dgp_Bool
isRect(run_elt *runs, Dgp_Int32 num_runs) {
  Dgp_Int32 runLeft, runLength, oldRunLeft, oldRunLength;
  Dgp_Int32 i;
  Dgp_Int32 y;

  oldRunLeft   = RunX(runs,0);
  oldRunLength = RunLen(runs,0);
  if(oldRunLength == 0 && oldRunLeft != 0) {
    return DGP_FALSE;
  }

  y =0;
  for (i=1; i < num_runs; i++) {
    runLeft   = RunX(runs,i);
    runLength = RunLen(runs,i);
    if(runLength == 0) {
      if(runLeft != y)
        return DGP_FALSE;
      else
        continue;
    }

    if ((runLeft != oldRunLeft) || (runLength != oldRunLength))
      return DGP_FALSE;

    y++;
  }
  return DGP_TRUE;
}



/* how big to allocate some of our internal arrays initially (they'll
   be reallocated if we get more than this many pages) */
#define INIT_NUM_PAGES 15 
#define INIT_NUM_TOKS  50
#define INIT_NUM_POS   100


ErrCode
DgpTdc_startSection(DgpTdc_Cookie _cookie, DgpTdc_SectionBlock *prev, Dgp_Bool report)
{
    DGP_STATE_PROLOGUE2("DgpTdc_startSection()");
    tdoc_state *cookie;
    Dgp_State *dgp_state;
    
    DgpTdc_SectionBlock *sb;
    LlTokOptions *tok_opt = NULL;
    Dgp_Int32 i;
    
    
    /* validate params */
    DGP_CHECK_NULL(_cookie);
    cookie = (tdoc_state *)_cookie;
    sb = cookie->curBlock;
    dgp_state = cookie->dgp_state;
    tok_opt = cookie->llTokenizerOptions;
    report = DGP_FALSE;
    DGP_ERR_BEGIN(dgp_state);
    
    
    /*DGP_REPORTF0(report, "Starting a new section..."); */
    
    /* Start the tokenizer if it isn't started because:
    1. This is the first section
    2. The last section was final
    */
    if(cookie->llTokState == NULLLLTOKSTATE) {
        tok_opt->flags = 0;
        tok_opt->dgp_state = dgp_state;
        /* fire up the main tokenizer */
        DGP_ERR_PASS_CLEANUP(DgpLlTokInit(&(cookie->llTokState), tok_opt));
        
    } else {
        /* This there might be tokens in the output */
        /* so pass them back to the tokenizer */
        if(prev != NULL) {
            
            DGP_ERR_PASS_CLEANUP(DgpLlTokResume(cookie->llTokState, prev->classesFound));
            prev->classesFound = DGP_NULLLIST;
            
            /* this is the situation where the last page of the prev section caused a  */
            /* tokMem overflow and been rolled back. We need to do a tokFinal so that  */
            /* the global dictionary be freed and start over  */
            if(prev->final == DGP_TRUE){
                DGP_ERR_PASS_CLEANUP(DgpLlTokFinal(cookie->llTokState, &(prev->classesFound)));
                /* Free all the sectionblock stuff */
                if(sb != NULL) {
                    if(sb->classesFound != DGP_NULLLIST) {
                        DgpLlTokFreeOutput(sb->classesFound, cookie->dgp_state);
                    }
                    sb->classesFound = DGP_NULLLIST;
                }
                cookie->llTokState = NULLLLTOKSTATE;
                tok_opt->flags = 0;
                tok_opt->dgp_state = dgp_state;
                
                /* fire up the main tokenizer */
                DGP_ERR_PASS_CLEANUP(DgpLlTokInit(&(cookie->llTokState), tok_opt));
                prev->tokMem = 0;
            }  
        } else {
        /* There is a non null llTokState in the cookie
        * but the user didn't pass us back the sectionBlock that 
        * resulted from that llTokState.  Tell 'em and fail.
            */
            
            DGP_ERR_CLEANUPF0(cErrInternal, "Need a previous section to restart the tokenizer.");
        }
    }
    
#if DGPTDC_DO_HALFTONE
    if(cookie->halftone) {
        if (cookie->halfTokState != NULL) { 
        /* 
        * Kill it if it is still alive  
        * (We should actually resume, but I don't know if it would work)
            */
            EndHalfTokenizer(cookie->halfTokState);
            if(prev != NULL) {
                prev->halfOutput = DGP_NULLLIST;
            }
        }
        
        /*Start it up (maybe again)  */
        cookie->halfTokState = InitHalfTokenizer(DGP_PALLOCATORS);
        DGP_CHECK_ALLOC(cookie->halfTokState);
    }    
#endif
    
    
    if(prev == NULL) {
        
        DGP_SEC_CALLOC(sb, (DgpTdc_SectionBlock *), 1, sizeof(DgpTdc_SectionBlock));
        cookie->curBlock = sb;
        
        DGP_SEC_MALLOC(sb->pages, (DgpTdc_PageInfo *), sizeof(DgpTdc_PageInfo) * INIT_NUM_PAGES);
        
        sb->numPagesAlloced = INIT_NUM_PAGES; 
        sb->halfOutput = DGP_NULLLIST;
        
        DGP_ERR_PASS_CLEANUP(DgpCC_initTokBlock(dgp_state, &(sb->tBlock), INIT_NUM_TOKS));
        
        sb->numOldToks = 0;
        DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&(sb->tokBlockList), dgp_state));
        
        /* set sampled compression info */
        sb->sampCompType = cookie->sampCompType;
        sb->sampCompInfo = cookie->sampCompInfo;
        sb->tokMem = 0;
    } else {
        
        sb = cookie->curBlock = prev;
        /* 24.4.2001 / dhtesq. */
        /* Leaking pBlocks... */
        /* ...and poss... */
        if(sb->pBlock){
            DGP_ERR_PASS_CLEANUP(DgpCC_cleanupPosBlock(dgp_state, sb->pBlock));
        }
        
        /*
        * The pages array will stay allocated, but 
        * the arrays in each page are freed
        */
        for(i = 0; i < sb->numPages; i++) {
            if(sb->pages[i].color && sb->pages[i].colors) {
                DGP_FREE(sb->pages[i].colors);
                sb->pages[i].colors = NULL;
            }
            
            if(sb->pages[i].boxes != NULL) {
                DGP_FREE(sb->pages[i].boxes);
                sb->pages[i].boxes = NULL;
            }
        }
        
        /* What about classesFound?
        * if the llTokState was null (returned from a final, or starting new)
        *  it is possible that there are classesFound in prev.
        *  However that means the user messed up.
        */
        DGP_ERR_GEN_CLEANUPF0((sb->classesFound != DGP_NULLLIST), cErrInternal, 
            "Previous section block isn't in sync with cookie");
        
            /* What about halfOutput?
            *  Currently the halftoneTokenizer can't frees its output in stages
            *  Each token is freed (which actually just frees its runs)
            *  then when endHalfTokenizer is called the list, nodes, and classes are freed
            *  This is because of some chunk mumbo jumbo.
            * Therefore it should be NULL by this point.
        */
        DGP_ERR_GEN_CLEANUPF0((cookie->halftone && (sb->halfOutput != DGP_NULLLIST)), 
            cErrInternal, "Previous section block isn't in sync with cookie");
        
        
        /* reallocate this array to leave (INIT_NUM_TOKS) space  */
        if(sb->tBlock->toks == NULL) {
            sb->tBlock->numAlloced = INIT_NUM_TOKS;
            DGP_SEC_MALLOC(sb->tBlock->toks, (DgpCC_TokInfo *), 
                sizeof(DgpCC_TokInfo) * INIT_NUM_TOKS);
            sb->numOldToks = 0;
        } else {
            sb->tBlock->numAlloced = INIT_NUM_TOKS + sb->numOldToks;
            DGP_SEC_REALLOC(sb->tBlock->toks, (DgpCC_TokInfo *), 
                sizeof(DgpCC_TokInfo) * sb->tBlock->numAlloced);
            sb->tBlock->num = sb->numOldToks;
        }
        
        if(sb->tokBlockList == DGP_NULLLIST) {
            /* create new tokblocklist */
            DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&(sb->tokBlockList), dgp_state));
        }
        if(sb->final == DGP_TRUE) sb->tokMem = 0;

        
    }
    
    /* initialize the block */
    sb->numPages = 0;
    DGP_ERR_PASS_CLEANUP(DgpCC_initPosBlock(dgp_state, &(sb->pBlock), INIT_NUM_POS));
    
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    /* Need to add frees for all allocations */
    
    ERR_RETURN;
    
}

Dgp_Int32 
tdoc_computeMemUsage(DgpTdc_Cookie _cookie)
{
  tdoc_state *cookie = (tdoc_state *)_cookie;
  Dgp_Int32 i,j;
  Dgp_Int32 totalUsage = 0;
  Dgp_Int32 memUsage;
  DgpTdc_PageInfo *curPage;
    
  for(i=0; i<cookie->curBlock->numPages; i++) {
    curPage = &(cookie->curBlock->pages[i]);

    if(curPage->memUsage != -1) {
      totalUsage += curPage->memUsage;

      continue;
    }

    memUsage = 0;
    memUsage += (curPage->numBoxes + curPage->halftone_numBoxes) * sizeof(tokpos);
    if(curPage->color) {
      memUsage +=  curPage->numBoxes * sizeof(color_tag);
    }


    for(j=0; j < curPage->num_annotations; j++) {
      memUsage += sizeof(Annotation);
      memUsage += curPage->annotations[j].size;
    }

    curPage->memUsage = memUsage;
    totalUsage += memUsage + sizeof(DgpTdc_PageInfo);
  }

  /* 
   * Now the fun part figuring out how much memory is used up by the 
   * section block this will include what is stored by the 
   * tokenizers (mono, sampled and halftone) the function are not exact
   * they only measure the size of the tokens.  I need to update them 
   * include the storage size some how.
   *
   * I could check classesFound and sampOutput, and halfOutput, but there is
   * no reason to call this function if tdoc_endSection has been called.
   */
  totalUsage += DgpLlTokMemUsage(cookie->llTokState);

  return totalUsage;
}


/* monochrome start page */
ErrCode
DgpTdc_addMonoPage(DgpTdc_Cookie _cookie, Dgp_Int32 pageNumber, 
         UInt8 *image, Dgp_Int32 pageWidth, Dgp_Int32 pageHeight,
         Dgp_Int32 bpl, Dgp_Int32 xRes, Dgp_Int32 yRes, Dgp_Bool report)
{
  DGP_STATE_PROLOGUE2("DgpTdc_addMonoPage()");
  tdoc_state *cookie;
  Dgp_State *dgp_state;

  DgpTdc_PageInfo *curPage;
  DgpTdc_SectionBlock *sb;

  UInt8 *w_image = image;
  PIXR *readPixr = NULL;
#if DGPTDC_DO_HALFTONE
  UInt8 cell_size=0;  
#endif


  /* validate params */
  DGP_CHECK_NULL(_cookie);
  cookie = (tdoc_state *)_cookie;
  sb = cookie->curBlock;
  curPage = &(cookie->curPage);
  dgp_state = cookie->dgp_state;
  DGP_ERR_BEGIN(cookie->dgp_state);
  DGP_CHECK_NULL(dgp_state->ipwstate);
  
  curPage->pageNum = pageNumber;
  curPage->index = -1;
  curPage->width = pageWidth;
  curPage->height = pageHeight;
  curPage->xRes = xRes;
  curPage->yRes = yRes;
  curPage->color = DGP_FALSE;
  curPage->colors = NULL;
  curPage->annotations = NULL;
  curPage->num_annotations = 0;
  curPage->halftone_boxes = NULL;
  curPage->halftone_numBoxes = 0;
  curPage->numToks = 0;
  curPage->toks = NULL;
  curPage->numBoxes = 0;
  curPage->boxes = NULL;

#if DGPTDC_DO_HALFTONE
  if (cookie->halftone) {
    DoHalftoneFinding(cookie->halfTokState, dgp_state, dgp_state->ipwstate, &w_image, 
      pageWidth, pageHeight, bpl, xRes, yRes, &readPixr, &cell_size, 
      &(curPage->halftone_boxes), &(curPage->halftone_numBoxes), report);
    
    if (cookie->halftone==2) curPage->cell_size = cell_size;
  }
#endif

  /*err=w_write(ipwstate, "c:\\images2\\text.pbm", readPixr, PNM); */
  /*DGP_ASSERT(err==cSuccessful);  */
  
  DGP_REPORTF1(report, "About to find and tokenize CCs for page %4d.",curPage->pageNum);

  /* Get the boxes. The tokenizing via tokAddToken() happens during
   * these calls to GetBoxes(). 
   *
   * NOTE: at this point, no outliers will ever be returned here (we set the
   * max size to page width and page height) everything goes to the tokenizer right now.
   */
  DGP_ERR_PASS_CLEANUP(DgpCC_getBoxes(dgp_state, cookie->llTokState, w_image, pageWidth, pageHeight, 
    pageWidth, pageHeight, bpl, &(curPage->numBoxes), &(curPage->boxes), cookie->sizeThreshold,
    &(curPage->outliers), sb->tBlock, sb->pBlock, sb->numPages, report));


  DGP_REPORTF3(report, 
      "Found %5d CCs  (%5d boxes and %5d outliers).",
      DGPADT_LILEN(curPage->outliers) + curPage->numBoxes,
      curPage->numBoxes, 
      DGPADT_LILEN(curPage->outliers));
  DGP_REPORTF1(report, 
      "Found %5d halftone boxes.",
      curPage->halftone_numBoxes);
  
  /* we aren't ever expecting any outliers (right now anyway)... */
  DGP_ASSERT(DGPADT_LILEN(curPage->outliers) == 0);

  /* ditch the list */
  DgpAdt_liFree(curPage->outliers);
  curPage->outliers = DGP_NULLLIST;

  if(readPixr != NULL) w_destroyPixr(dgp_state->ipwstate, &readPixr);
  readPixr = NULL;

  addLlPage(cookie->curPage, sb);

  NORMAL_RETURN;

ERR_CLEANUP:
  if(readPixr != NULL) w_destroyPixr(dgp_state->ipwstate, &readPixr);

  ERR_RETURN;
}

#ifndef DGP_FAXMODE
ErrCode
DgpTdc_startColorPage(DgpTdc_Cookie _cookie, 
                      Dgp_Int32 pageNumber, 
                      Dgp_Int32 pageWidth,
                      Dgp_Int32 pageHeight, 
                      Dgp_Int32 xRes, 
                      Dgp_Int32 yRes,
                      Dgp_Bool noTokenImages, 
                      Dgp_Bool sepLargeImages,
                      Dgp_Bool report)
{
  DGP_STATE_PROLOGUE2("DgpTdc_startColorPage()");
  tdoc_state *cookie;


  /* validate params */
  DGP_CHECK_NULL(_cookie);
  cookie = (tdoc_state *)_cookie;
  DGP_ERR_BEGIN(cookie->dgp_state);
  
  /* validate color compression params */
  DGP_ERR_GEN(((cookie->sampCompType < DGPTDC_SAMP_COMPTYPE_JPEG) || 
               (cookie->sampCompType > DGPTDC_SAMP_COMPTYPE_JPEG)), cErrInvalidParm);
  DGP_CHECK_NULL(cookie->sampCompInfo);
  
  DGP_REPORTF1(report, "About to find and tokenize color CCs for page %4d.", pageNumber);

  cookie->curPage.color = DGP_TRUE;
  cookie->curPage.pageNum = pageNumber;
  cookie->curPage.index = -1;
  cookie->curPage.width = pageWidth;
  cookie->curPage.height = pageHeight;
  cookie->curPage.xRes = xRes;
  cookie->curPage.yRes = yRes;
  cookie->curPage.annotations = NULL;
  cookie->curPage.num_annotations = 0;
  cookie->curPage.halftone_boxes = NULL;
  cookie->curPage.halftone_numBoxes = 0;
  cookie->curPage.largeImages = NULL;
  cookie->curPage.numToks = 0;
  cookie->curPage.toks = NULL;
  cookie->curPage.colors = NULL;
  cookie->curPage.numBoxes = 0;
  cookie->curPage.boxes = NULL;

  cookie->noTokenImages = noTokenImages;
  cookie->sepLargeImages = sepLargeImages;


  DGP_ERR_PASS_CLEANUP(DgpCC_initPage(cookie->cccc, pageWidth, pageHeight, 
    cookie->washThreshold, 1));

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;

}


/* some macros for the function below */
#define pos_init(_pos, _xpos, _ypos, _extdata) \
do { \
  (_pos).xpos = (_xpos); \
  (_pos).ypos = (_ypos); \
  (_pos).extdata = (_extdata); \
} while(0)

#define rtok_init(_rtok, _runlist, _nruns, _width, _height, _userdata) \
do { \
  (_rtok).runs = (_runlist); \
  (_rtok).nruns = (_nruns); \
  (_rtok).width = (Dgp_Int16)(_width); \
  (_rtok).height = (Dgp_Int16)(_height); \
  (_rtok).userdata = (void *)(_userdata); \
} while(0)

#define poss_setInfo(_pblock, _pinfo, _pagenum, _index) \
do { \
  (_pinfo) = &((_pblock)->poss[(_pblock)->num]); \
  (_pinfo)->page = (_pagenum); \
  (_pinfo)->posOffset = (_index); \
  (_pblock)->num++; \
} while(0)

static int
cmpfunc(const void *p1, const void *p2) 
{ 
    const tdoc_image *i1 = (const tdoc_image *)p1; 
    const tdoc_image *i2 = (const tdoc_image *)p2; 
    if (i1->starty > i2->starty)
    {
      return(1); 
	  } 
    if (i1->starty == i2->starty)
    {
      return (0);
    }
	  return(-1); 
} 

ErrCode
DgpTdc_addColorStrip(DgpTdc_Cookie _cookie, 
                     Dgp_Int32 pageNumber, 
                     Dgp_Int32 numLines, 
                     DgpCC_PFNReadRun readRun, 
                     void *stripData,
                     Dgp_Bool report)
{
  Dgp_State *dgp_state;
  DGP_STATE_PROLOGUE2("DgpTdc_addColorStrip()");
  tdoc_state *cookie;
  DgpTdc_SectionBlock *sb;

  tokpos *pos;
  DgpAdt_ListNode ln;
  Dgp_Int32 width, height;
  DgpCC_ColorCC *cur_cc = NULL;
  Dgp_Int32 cur_pos_index;
  Dgp_Int32 flags = 0;
  DgpCC_PosInfoBlock *pBlock;
  DgpCC_PosInfo *pInfo;
  LlRunToken rTok;
  Dgp_Int32 numNewBoxes;
  DgpAdt_List new_ccs = DGP_NULLLIST, new_sampled_ccs = DGP_NULLLIST;
  tdoc_image *array;
  Dgp_Int32 count, total_count, first_image, last_image;
  Dgp_UInt32 maxy;

  /* validate params */
  DGP_CHECK_NULL(_cookie);
  cookie = (tdoc_state *)_cookie;
  dgp_state = cookie->dgp_state;
  DGP_ERR_BEGIN(dgp_state);
    
  DGP_ERR_GEN_CLEANUPF0((cookie->curPage.pageNum != pageNumber), 
    cErrInternal, "Can't add a strip to a non-current page");

  /* create the lists */
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&new_ccs, dgp_state));
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&new_sampled_ccs, dgp_state));

  /* process the color strip */
  DGP_ERR_PASS_CLEANUP(DgpCC_doStrip(cookie->cccc, readRun, stripData, numLines, 
      new_ccs, new_sampled_ccs));

  /* get number of new ccs generated */
  numNewBoxes = DGPADT_LILEN(new_ccs) + DGPADT_LILEN(new_sampled_ccs);

  if(numNewBoxes == 0) {

    /* free lists */
    DgpAdt_liFree(new_ccs);
    DgpAdt_liFree(new_sampled_ccs);

    /* we're done */
    NORMAL_RETURN;
  }

  DGP_REPORTF1(report, "Found %d more ccs", numNewBoxes);

  sb = cookie->curBlock;
  pBlock = sb->pBlock;

  cur_pos_index = cookie->curPage.numBoxes;
  
  /* total number of boxes on the page so far... */
  cookie->curPage.numBoxes += numNewBoxes;

  /* reallocate the array for holding the colors */
  DGP_SEC_REALLOC(cookie->curPage.colors, (color_tag *), 
    (sizeof(color_tag)*(cookie->curPage.numBoxes)));

  /* reallocate the array of tokpos's and convert the CCs to tokpos's. (We
    want to convert the list of CCs to tokpos's ASAP, because tokpos's
    are smaller than CCs) */
  pos = cookie->curPage.boxes;
  DGP_SEC_REALLOC(pos, (tokpos *), 
    (sizeof(tokpos)*(cookie->curPage.numBoxes)));
  cookie->curPage.boxes = pos;
    
  if((pBlock->num + numNewBoxes) >= pBlock->numAlloced) {
    pBlock->numAlloced += numNewBoxes;
    DGP_SEC_REALLOC(pBlock->poss, (DgpCC_PosInfo *), 
      pBlock->numAlloced * sizeof(DgpCC_PosInfo));
  }

  /* put the ccs where they need to be */
  DGPADT_FOREACH(ln, new_ccs) {

    /* get next cc */
    cur_cc = (DgpCC_ColorCC *)DGPADT_LIGET(ln);
    DGPADT_LIPUT(ln, NULL);
  
    width = cur_cc->lrX - cur_cc->ulX + 1;
    height = cur_cc->lrY - cur_cc->ulY + 1;
    
    DGP_ASSERT(!colorcc_isWashOrSampled(cur_cc));

    /* this colorCC is not sampled at all */
    cookie->curPage.colors[cur_pos_index].color_type = RGB_COLOR;
    cookie->curPage.colors[cur_pos_index].value.rgb_color = cur_cc->colorInfo.rgbColor;

    /* Fill in the fields in the corresponding tokpos (throw the array of 
       runs at the tokenizer to get the tok ptr). */
    pos_init(pos[cur_pos_index], cur_cc->ulX, cur_cc->lrY, NULL);

    rtok_init(rTok, cur_cc->runList, cur_cc->numRuns, width, height, pBlock->num);
    cur_cc->runList = NULL;

    poss_setInfo(pBlock, pInfo, sb->numPages, cur_pos_index);

    DGP_ERR_PASS_CLEANUP(DgpCC_tokenAddCC(dgp_state, cookie->llTokState, sb->tBlock, 
      &rTok, &(pos[cur_pos_index].tok)));
    cur_pos_index++;

    /* done with this colorcc */
    COLORCC_FREE(cur_cc);
    cur_cc = NULL;
  }

  /* ditch the list */
  DgpAdt_liFree(new_ccs);
  new_ccs = DGP_NULLLIST;

  /* update the number of boxes (there may be fewer than we originally 
     thought because some images may have become subIFDs, not tokens) */
  cookie->curPage.numBoxes = cur_pos_index;

  if (new_sampled_ccs->nelt > 0)
  {
  /* put the sampled ccs where they need to be */
  array = DGP_MALLOC (new_sampled_ccs->nelt * sizeof (tdoc_image));
  count = 0;
  DGPADT_FOREACH(ln, new_sampled_ccs) {

    /* get next cc */
    cur_cc = (DgpCC_ColorCC *)DGPADT_LIGET(ln);
    DGPADT_LIPUT(ln, NULL);
  
    width = cur_cc->lrX - cur_cc->ulX + 1;
    height = cur_cc->lrY - cur_cc->ulY + 1;
    
    DGP_ASSERT(colorcc_isWashOrSampled(cur_cc));
      
    /* If it is not on top, see if we can trim it */
    if(!colorcc_isWash(cur_cc)) {

      DGP_ERR_PASS_CLEANUP(trimImageRGB(cur_cc, dgp_state));
      width = cur_cc->lrX - cur_cc->ulX + 1;
      height = cur_cc->lrY - cur_cc->ulY + 1;

      if(width == 0) {

        /* the image has no necessary pixels so it was
           deleted by the trimming code. */
        colorcc_cleanup(cur_cc);

        /* done with this colorcc */
        continue;
      }

    }

    array[count].endy = cur_cc->lrY;
    array[count].starty = cur_cc->ulY;
    array[count].pcc = cur_cc;
    count++;

    flags = 0;

    /* set info about the image */
    if(isRect(cur_cc->runList, cur_cc->numRuns)) {
      flags |= SAMPTOK_IS_RECT;
    } 

    if(colorcc_isWash(cur_cc)) {
      flags |= DGPTDC_SAMPTOK_IS_WASH;
    }
  
    DGP_ASSERT(width <= DGPIMG_WIDTH(cur_cc->colorInfo.rgbIm));
    DGP_ASSERT(height <= DGPIMG_HEIGHT(cur_cc->colorInfo.rgbIm));
    
  }

  total_count = count;

  /* ditch the list */
  DgpAdt_liFree(new_sampled_ccs);
  new_sampled_ccs = DGP_NULLLIST;

  /* sort the images by starty */
  qsort (array, total_count, sizeof(tdoc_image), cmpfunc);

  /* combine images and Jpeg compress them */
  first_image = 0;
  last_image = 0;
  count = 0;
  maxy = array[first_image].endy;
  while (count + 1 < total_count)
  {
	/* check if we need to combine in the next image */
    if (count + 1 < total_count && 
		array[count + 1].starty < maxy)
    {
      count++;
	  last_image = count;
      if (array[count].endy > maxy)
        maxy = array[count].endy;
    }
	else   
	{
		if (first_image == last_image)
		{
		/* Write out single image */
			width = array[first_image].pcc->lrX - array[first_image].pcc->ulX + 1;
			height = array[first_image].pcc->lrY - array[first_image].pcc->ulY + 1;
			DGP_ERR_PASS_CLEANUP(handleLargeImages(cookie, &array[first_image].pcc, 
				width, height, flags));
		}
		else
		{
			/* write out a set of images that need to be combined */
			DGP_ERR_PASS_CLEANUP(createJpeg(cookie, array, first_image, last_image));
		}
		count++;
		first_image = last_image = count;
        maxy = array[count].endy;
	}
  }

  /* Write out set of images */
  if (total_count > 0)
  {
		if (first_image == last_image)
		{
			width = array[first_image].pcc->lrX - array[first_image].pcc->ulX + 1;
			height = array[first_image].pcc->lrY - array[first_image].pcc->ulY + 1;
			DGP_ERR_PASS_CLEANUP(handleLargeImages(cookie, &array[first_image].pcc, width, height,
				flags));
		}
		else
		{
			DGP_ERR_PASS_CLEANUP(createJpeg(cookie, array, first_image, last_image));
		}
  }

  DGP_FREE (array);

  }

		
		NORMAL_RETURN;

ERR_CLEANUP:

  if(cur_cc != NULL) colorcc_cleanup(cur_cc);

  /* ditch lists if they exist */
  if(new_ccs != DGP_NULLLIST) {
    DGPADT_FOREACH(ln, new_ccs) {
      cur_cc = DGPADT_LIGET(ln);
      if(ln != NULL) colorcc_cleanup(cur_cc);
    }
    DgpAdt_liFree(new_ccs);
  }
  if(new_sampled_ccs != DGP_NULLLIST) {
    DGPADT_FOREACH(ln, new_sampled_ccs) {
      cur_cc = DGPADT_LIGET(ln);
      if(ln != NULL) colorcc_cleanup(cur_cc);
    }
    DgpAdt_liFree(new_sampled_ccs);
  }

  ERR_RETURN;
}
 

ErrCode
DgpTdc_endColorPage(DgpTdc_Cookie _cookie, 
                    Dgp_Int32 pageNumber, 
                    Dgp_Bool report)
{
  DGP_STATE_PROLOGUE2("DgpTdc_endColorPage()");
  tdoc_state *cookie;
  Dgp_State *dgp_state; 

  report = report;

  /* validate params */
  DGP_CHECK_NULL(_cookie);
  cookie = (tdoc_state *)_cookie;
  dgp_state = cookie->dgp_state;
  DGP_ERR_BEGIN(dgp_state);  

  DGP_ERR_GEN_CLEANUPF0((cookie->curPage.pageNum != pageNumber), 
    cErrInternal, "Can't end non-current page");

  addLlPage(cookie->curPage, cookie->curBlock);

  NORMAL_RETURN;
ERR_CLEANUP:
  ERR_RETURN;
}

 
ErrCode
DgpTdc_addAnnotation(DgpTdc_Cookie _cookie, 
                     Dgp_Int32 pgNum, 
                     void *data, 
                     Dgp_Int32 size,
                     Dgp_Int32 tag,
                     Dgp_Bool report)
{
  DGP_STATE_PROLOGUE2("DgpTdc_addAnnotation()");
  tdoc_state *cookie;
  Dgp_State *dgp_state;

  DgpTdc_PageInfo *page;
  Dgp_Int32 numAnn;  


  /* validate params */
  DGP_CHECK_NULL(_cookie);
  cookie = (tdoc_state *)_cookie;
  dgp_state = cookie->dgp_state;
  DGP_ERR_BEGIN(dgp_state);  

  DGP_ERR_GEN_CLEANUPF0((cookie->curPage.pageNum != pgNum), 
    cErrInternal, "Can't add an annotation to a non-current page");

  /* Use the index to see if the curPage has been added to the curBlock */
  if(cookie->curPage.index != -1) {
    /* It has been add so modify the page in the block */
    page = &(cookie->curBlock->pages[cookie->curPage.index]);
  } else {
    /* It hasn't been added yet, so modify the curPage */
    page = &(cookie->curPage);
  }

  numAnn = page->num_annotations;
  DGP_SEC_REALLOC(page->annotations, (Annotation *), 
    (page->num_annotations + 1) * sizeof(Annotation));

  page->annotations[numAnn].data = data;
  page->annotations[numAnn].size = size;
  page->annotations[numAnn].tag = tag;
  page->num_annotations++;


  DGP_REPORTF2(report,
    "Added annotation %d to page %d", numAnn, pgNum);

  NORMAL_RETURN;

ERR_CLEANUP:
    
  ERR_RETURN;
}
#endif

/*
 * Pull tokens on the current page back from the tokenizer
 * Free the current page. Make previous page as current page
 */
ErrCode
DgpTdc_endSectionRollBack(
                          DgpTdc_Cookie _cookie, 
                          DgpTdc_SectionBlock **retSb, 
                          Dgp_Bool report)
{
    DGP_STATE_PROLOGUE2("DgpTdc_endSectionRollBack()");
    tdoc_state *cookie;
    Dgp_State *dgp_state;
    DgpAdt_ListNode nextClass, nextMember;
    Dgp_Bool isNewClass = DGP_TRUE;
    DgpTdc_SectionBlock *sb;
    DgpTdc_PageInfo curPage;
    LlTokClass c;
    DgpCC_PosInfo cc;
    Dgp_Int32 page, offset;

    /* validate params */
    DGP_CHECK_NULL(_cookie);
    cookie = (tdoc_state *)_cookie;
    sb = cookie->curBlock;
    curPage = cookie->curPage;
    page = sb->numPages -1;
    dgp_state = cookie->dgp_state;
    report = DGP_FALSE;

    DGP_ERR_BEGIN(dgp_state);
    DGP_CHECK_NULL(retSb);
    *retSb = NULL;
    
    /* get the classes from the tokenizer */
    sb->final = DGP_TRUE;
    DGP_ERR_PASS_CLEANUP(DgpLlTokPartial(cookie->llTokState, &(sb->classesFound)));
    nextClass = DGPADT_LIFIRST(sb->classesFound);
    while(nextClass != NULL){
        isNewClass = DGP_TRUE;
        c = (LlTokClass)DGPADT_LIGET(nextClass);
        if(c->members != DGP_NULLLIST){
		    nextMember = DGPADT_LIFIRST(c->members);
            while (nextMember != DGP_NULLLISTNODE) {
                offset = (Dgp_Int32)DGPADT_LIGET(nextMember);
                cc = sb->pBlock->poss[offset];
                if ( cc.page == page) {
                   DGP_ERR_PASS_CLEANUP(
                                 DgpAdt_liRemAndNext(nextMember, &nextMember));
                } else {
                    isNewClass = DGP_FALSE;
			        nextMember = DGPADT_LINEXT(nextMember);
			    }
		    }/* end nextMember != NULL */
            if(isNewClass == DGP_TRUE){ 
                /* if this class is only relavent to this page */
                /* free this class */
		        DGP_ASSERT(DGPADT_LILEN( c->members) == 0);
		        DgpAdt_liFree(c->members);
                DgpAdt_liRemAndNext(nextClass,&nextClass);
                DgpLlTokFreeClass(c, dgp_state);
            }else{ 
                nextClass = DGPADT_LINEXT(nextClass);
            }
        }else nextClass = DGPADT_LINEXT(nextClass);
    } /* end DGPADT_FOREACH class */
    

    sb->final = DGP_TRUE;
    /* make the previous page to be the current page */
    sb->numPages--;
    DGP_ASSERT(sb->numPages > 0);
    cookie->curPage = sb->pages[sb->numPages-1];
    sb->tokMem = 0;
    /* clean up the page. Note the bit are owned by tokenizer */
    /* and no outlier is generated this time */
    if(curPage.color && curPage.colors) {
        DGP_FREE(curPage.colors);
        curPage.colors = NULL;
      }

    if(curPage.boxes != NULL) {
        DGP_FREE(curPage.boxes);
        curPage.boxes = NULL;
    }
	*retSb = sb;
   
    NORMAL_RETURN;
    
ERR_CLEANUP:
    ERR_RETURN;
}

ErrCode
DgpTdc_endSection(DgpTdc_Cookie _cookie, 
                  DgpTdc_SectionBlock **retSb, 
                  Dgp_Bool final, 
                  Dgp_Bool report)
{
  DGP_STATE_PROLOGUE2("DgpTdc_endSection()");
  tdoc_state *cookie;
  Dgp_State *dgp_state;

  DgpTdc_SectionBlock *sb;
#if DGPTDC_DO_HALFTONE
  DgpTdc_PageInfo *pages;
  Dgp_Int32  max_cell_size = 0;
  Dgp_Int32  the_cell_size = 0;
  Dgp_Int32  *cell_size_histo;
  Dgp_Int32 i;
  Dgp_Int32  max_index     = 1;
#endif


  /* validate params */
  DGP_CHECK_NULL(_cookie);
  cookie = (tdoc_state *)_cookie;
  sb = cookie->curBlock;
  dgp_state = cookie->dgp_state;
  report = DGP_FALSE;

#if DGPTDC_DO_HALFTONE
  pages = sb->pages;
#endif
  DGP_ERR_BEGIN(dgp_state);
  DGP_CHECK_NULL(retSb);
  *retSb = NULL;

  /* get the classes from the tokenizer (DgpLlTokFinal() frees the state cookie) */
  sb->final = final;
  if(final) {    
    if (cookie->llTokState)
      {
	DGP_ERR_PASS_CLEANUP(DgpLlTokFinal(cookie->llTokState, &(sb->classesFound)));
	cookie->llTokState = NULLLLTOKSTATE;
      }
  } else {
    DGP_ERR_PASS_CLEANUP(DgpLlTokPartial(cookie->llTokState, &(sb->classesFound)));
  }
    
#if DGPTDC_DO_HALFTONE
  if(cookie->halftone) {
    /*All the trees hanging off the   */
    /*state has been freed, but the structure */
    /*itself has not been, because the pointer to */
    /*the memory chunk that contains the classes */
    /*is in there. I'll free it in cleanupCompressDocument. */
    sb->halfOutput = HalfTokFinal(cookie->halfTokState); 

    if (cookie->halfTokState->next_toknum && (!(sb->halfOutput))) {
      DGP_ERR_GEN_CLEANUPF0(tc_errOut, 
        "Problem getting output from halftone tokenizer.");
      goto bailout;
    }

    /*lets determine if there are some halftones that  */
    /*we are supposed to lossyly tokenize. */
    
    /*pass through the array to find the maximum */
    /*cell_size therein */
    for (i=0; i < sb->numPages; i++) {
      if (pages[i].cell_size > max_cell_size) max_cell_size = pages[i].cell_size;
    }
    
    /*now, only set up the hamming tokenizer if the max_cell_size is > 0 */
    if (max_cell_size) {
      /*allocate the histogram */
      cell_size_histo = (Dgp_Int32 *)tc_calloc((max_cell_size + 1), sizeof(Dgp_Int32));
      
      /*run through the reported cell_sizes again to count up */
      for (i=0; i < sb->numPages; i++) {
        cell_size_histo[pages[i].cell_size]++;
      }
      
      /*find the maximum, excepting zero. */
      for (i=1; i <= max_cell_size; i++) {
        if (cell_size_histo[i] > cell_size_histo[max_index]) max_index = i;
      }
      
      the_cell_size = max_index;
      
      free(cell_size_histo);
    }    
    
    /*lets replace the tokens */
    if (the_cell_size) {
      if(!HammingReplace(tcc, the_cell_size, 1, the_cell_size, 3, 
        sb->halfOutput, pages, sb->numPages, -1, report)) {
        goto bailout;
      }
    }

  }
#endif  

  *retSb = sb;

  NORMAL_RETURN;

ERR_CLEANUP:
  ERR_RETURN;
}

/* Finish the document:
 * free the llTokState if 
 * 
 * if endSection was not called with final=true 
 * return false, 
 * Or perhaps this should return a 
 * "portable" document block that contains output from 
 * all unfinalized tokstates
 */
ErrCode
DgpTdc_endDoc(DgpTdc_Cookie *_cookie, 
              DgpTdc_SectionBlock *sb)
{
  DGP_STATE_PROLOGUE2("DgpTdc_endDoc()");
  tdoc_state *cookie;
  Dgp_Int32 i;
  Dgp_State *dgp_state;
  DgpAdt_List dummyList;

  DGP_STOP_WHINE;

  if(_cookie == NULL) NORMAL_RETURN;
  
  cookie = *_cookie;

  if(cookie == NULL) NORMAL_RETURN;
  *_cookie = NULL;

  dgp_state = cookie->dgp_state;

  /* 27.4.2001 / dhtesq. */
  /* It's possible to end on an incremental dictionary or something and end the */
  /* document without freeing internal tokenizer state -- force finality here. */
  if(cookie->llTokState){
	  DGP_ERR_PASS(DgpLlTokFinal(cookie->llTokState, &dummyList));
	  if(dummyList) DgpAdt_liFree(dummyList);
	  cookie->llTokState = NULLLLTOKSTATE;
  }
 
  /* Free the options pointer */
  if (cookie->llTokenizerOptions)
    {
      DGP_FREE (cookie->llTokenizerOptions);
      cookie->llTokenizerOptions = NULL;
    }

  /* Free all the sectionblock stuff */
  if(sb != NULL) {
    if(sb->classesFound != DGP_NULLLIST) {
      DgpLlTokFreeOutput(sb->classesFound, cookie->dgp_state);
    }
    sb->classesFound = DGP_NULLLIST;
  }

  if(cookie->cccc != NULL) DGP_ERR_PASS(DgpCC_cleanup(&(cookie->cccc)));

#if DGPTDC_DO_HALFTONE
  if(cookie->halftone) {
    if (cookie->halfTokState != NULL) { 
      /* 
       * Kill it if it is still alive  
       * (We should actually resume, but I don't know if it would work)
       */
      EndHalfTokenizer(cookie->halfTokState);
    }
  }
#endif 
  
  if(sb != NULL) {

    /* ditch the jpeg compression info if it exists */
    if(sb->cinfo != NULL) DGP_FREE(sb->cinfo);
    
    for(i = 0; i < sb->numPages; i++) {
      if(sb->pages[i].color && sb->pages[i].colors) {
        DGP_FREE(sb->pages[i].colors);
      }

      if(sb->pages[i].boxes != NULL) {
        DGP_FREE(sb->pages[i].boxes);
      }
    }

    DGP_FREE(sb->pages);

    if(sb->tBlock != NULL) {
      DgpCC_cleanupTokBlock(cookie->dgp_state, sb->tBlock);
    }
    if(sb->pBlock != NULL) {
      DgpCC_cleanupPosBlock(cookie->dgp_state, sb->pBlock);
    }

    if(sb->tokBlockList) {
	  DgpAdt_liFree(sb->tokBlockList);
    }

    DGP_FREE(sb);
  }
  
  DGP_FREE(cookie);

  NORMAL_RETURN;
} 

