#include "token-compress.h"
#include "token-compress-priv.h"
#include "dict-util.h"
#include "halftone.h"
#if DGPTDC_DO_HALFTONE
#include "hamtok.h"
#endif

/* include/shared, adt... */
#include <util.h>
#include <list.h>

/* include/include-ipcore */
#include <libiow.pub>
#include <libutilw.pub>
#include <libpixrw.pub>
#include <libscalw.pub>
#include "PsgFix.h"

/* include */
#include "jbig2apiCB.h"
#include "g4.h"
#include "iowrap.h"
#include <mmcompress.h>
#include <tok-lossless.h>
#include <ascii_exts.h>
#include <simple_utils.h>
#include <limits.h>
#include <color-extension.h>
#include <encode-colors.h>
#include <colorpos.h>

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcompdoc/token-output.c,v 1.61 2003/02/10 22:53:23 mkrocken Exp $");

ErrCode
DgpTdc_writeInit(Dgp_State *dgp_state, DgpTdc_WriteCookie *_cookie, Dgp_Int32 numPgs, Dgp_Int32 numDicts, 
                 uds *stream, Dgp_Bool ascii_ext, void *params)
{
  DGP_STATE_PROLOGUE2("DgpTdc_toutInit()");
  tout_state *cookie = NULL;
  IO_State* io_state;
  /* TIFF FX writer */
  Dgp_UInt32 myInstanceID = 0;
  DOCTYPE outDocType = DOCTYPE_TIFFFX_M1;
  Dgp_UInt8 version = 0;
  Dgp_UInt32 options = 0;
  
  /* not used, for now...? */
  (void)params;
  
  /* validate params */
  DGP_ERR_BEGIN(dgp_state);
  DGP_CHECK_NULL(_cookie);
  *_cookie = NULL;
  DGP_ERR_GEN((numPgs < 0), cErrInvalidParm);
  DGP_ERR_GEN((numDicts < 0), cErrInvalidParm);
  DGP_CHECK_NULL(stream);
  /* validate parts of the stream */
  DGP_ERR_GEN(((stream->fseek == NULL) || (stream->ftell == NULL) || 
    (stream->fread == NULL) || (stream->fwrite == NULL)), cErrInvalidParm);
  
  
  DGP_SEC_CALLOC(cookie, (tout_state *), 1, sizeof(tout_state));
  DGP_SEC_CALLOC(io_state, (IO_State *), 1, sizeof(IO_State));
  
  io_state->n_dict = 0;
  cookie->dgp_state = dgp_state;
  cookie->numPgs = numPgs;
  cookie->ascii_ext = ascii_ext;
  cookie->IOState = io_state;

  /* create jbig2 state for lossless token encoding */
  DGP_ERR_PASS_CLEANUP(
    createJb2StateAPICB(stream, dgp_state, 2, numPgs, numDicts, /* jbig2 profile */
		     &(cookie->jbig2State), NULL));
  
  /* Write the file header */ 
  
  io_state->stream = (void*) stream->fp;
  DGP_ERR_PASS_CLEANUP(IO_InitInstance(myInstanceID, &io_state->inst));
  DGP_ERR_PASS_CLEANUP(IO_CreateDocument(io_state->inst, 
					 outDocType, 
					 (Dgp_Int32) io_state->stream,
					 version, 
					 options, 
					 (IO_READFUNC)ibufRead, 
					 (IO_WRITEFUNC)ibufWrite, 
					 ibufSeek, 
					 ibufFileSize, 
					 ibufTell,
					 &io_state->doc));

  *_cookie = cookie;
  cookie = NULL;
  
  NORMAL_RETURN;
  
ERR_CLEANUP:
  
  if(cookie != NULL) DgpTdc_writeEnd((DgpTdc_WriteCookie *)&cookie);
  
  ERR_RETURN;
}

/*
* Need to reconsider this
*/
void
freeTokens(Dgp_State *dgp_state, token *toks, Dgp_Int32 ntoks)
{
  token *cur_tok, *end_tok;
  
  for(cur_tok = toks, end_tok = toks + ntoks; cur_tok < end_tok; cur_tok++) {
    DGP_FREE((void *)cur_tok->bits);
  }
  DGP_FREE(toks);
  
  return;
}

ErrCode
writeHalftoneDictionary(Dgp_State *dgp_state, void **ret,
                        void *jbig2State, 
                        DgpAdt_List half_tokOutput, Dgp_Bool report)
{
  DGP_PROLOGUE("writeHalftoneDictionary()", DGP_PERRSTATE);
  
  void  *half_dictblock = NULL;
  token *halftone_toks = NULL;
  Dgp_Int32 half_num_singletons = 0, half_num_non_singletons = 0;
  Dgp_Int32 num_halftone_toks = 0;
  /*Dgp_Int32 last_half_tok_written = -1; */
  
  /* used in both */
  DgpAdt_ListNode classi;
  LlTokClass c;
  
  /*ok, we got some halftone tokens. They are going into */
  /*their own global halftone dictionary.  */
#if 0
  /*lets get a token array of the kind writeXiffDictionary likes. */
  if (!createTokens(tcc, half_tokOutput,&halftone_toks,&half_num_non_singletons,
    &half_num_singletons, &last_half_tok_written,1,
    INT_MAX, -1, NULL, report)) {
    goto ERR_CLEANUP;
  }
#endif
  
  /*At this point, we don't need the list of tokens from  */
  /*the halftone tokenizer anymore. Free their runs, but let  */
  /*the chDestroy in cleanupCompressoDocument take care of the actual  */
  /*classes. The reason we are doing this here is not to carry  */
  /*around any additional memory. I know its weird to free  */
  /*the runs here and the actual classes later, but I don't */
  /*have access to the halftone tokenizer cookie chunk here... */
  DGPADT_FOREACH(classi, half_tokOutput) {
    c = (LlTokClass)DGPADT_LIGET(classi);
    DGP_FREE(c->runs);
  }
  
  /*after this point, the chunk that contains the classes can be destroyed */
  /*because all the class info has been copied. Furthermore, */
  /*the whole hafltone tok state can now be freed     */
  num_halftone_toks = half_num_non_singletons + half_num_singletons;
  
  
  DGP_REPORTF0(report, 
    "About to write the halfone dictionary.");
  
  /*now write out the halftone dictionary and get the dictblock */
  /*for it */
  
  DGP_ERR_PASS_CLEANUP(
      writeJb2DictionaryCB(halftone_toks, num_halftone_toks,
      half_dictblock, &half_dictblock, NULL, 0, jbig2State, NULL));
  
  freeTokens(dgp_state, halftone_toks, num_halftone_toks);
  
  
  *ret = half_dictblock;
  
  NORMAL_RETURN;
  
ERR_CLEANUP:
  
  ERR_RETURN;
  
}

#ifdef DBUG_SAMPIMS
Dgp_Int32 pageNum = 0;
Dgp_Int32 count = 0;
#endif

ErrCode
writeImageSubIFD(Dgp_State *dgp_state, 
                 DgpTdc_PageInfo *page, 
                 IO_State *IOState, 
                 ImSubIFD *imageIFD, 
                 Dgp_Int32 sampCompType,
                 void *sampCompInfo, 
                 void **cinfo,
                 Dgp_Bool report)
{
  DGP_PROLOGUE("writeImageSubIFD()", DGP_PERRSTATE);
  
  IPWState *ipwstate = dgp_state->ipwstate;
  Dgp_UInt8 *comp_mask;
  Dgp_UInt32 comp_mask_size;
  PIXR *maskPixr = NULL;
  PIXR *tempPixr = NULL;
  Dgp_Int32 jpQuality = -1;
  Dgp_Float32 scaleFactor = 0.0;
  Dgp_Int32 width, height, xRes, yRes;
  Dgp_Int32 plane;

  /* IO stuff */
  IO_IMGHANDLE outIoImg;
  /* 19.3.2001 / dhtesq. */
  /* Changed colorspace from default to itulab. */
  IO_COLORSPACE colorSpace = IO_COLORSPACE_ITULAB;
  IO_COMPRESSION compression = IO_COMP_JPEG;
  IO_COLORTYPE ioColorType = IO_COLORTYPE_COLOR24;
  Dgp_UInt32 quality, bytes_written;

#ifdef DBUG_SAMPIMS
  Dgp_Int8 temp_str[40];
  FILE *fp;
#endif

  report = report;  
  
  /* adjust to the top left corner */
  imageIFD->ypos -= (imageIFD->height - 1);
  
  
  width = imageIFD->width;
  height = imageIFD->height;
  xRes = page->xRes;
  yRes = page->yRes;

  /* get scale factor (if it's a big image) */
  if((width > (Dgp_Int32)(page->width * LARGE_CC_PERCENT)) && 
     (height > (Dgp_Int32)(page->height * LARGE_CC_PERCENT))) {

    switch(sampCompType) {
    case DGPTDC_SAMP_COMPTYPE_JPEG:
      /* only scale if it's big */
        scaleFactor = ((DgpTdc_JpegCompInfo *)(sampCompInfo))->scaleFactor;
      break;
    default:
      DGP_ERRF0(cErrInternal, "invalid image compression type");
    }

    /* scale image dimensions */
    width = SCALE_DIMENSION(width, scaleFactor);
    height = SCALE_DIMENSION(height, scaleFactor);

    /* scale the resolution as well */
    xRes = SCALE_RESOLUTION(page->xRes, scaleFactor);
    yRes = SCALE_RESOLUTION(page->yRes, scaleFactor);

  } else {

    scaleFactor = 1.0;

  }

  if(imageIFD->mask != NULL) {

    /* we haven't filled under samptoks yet  */
    /* (we didn't know until now if they were singletons) */
    if(imageIFD->samptok != NULL) {
      DGP_ERR_PASS_CLEANUP(fillUnderImageMask(dgp_state, 
        imageIFD->samptok->im, imageIFD->mask, imageIFD->maskOffset));
    }
    
    /* turn into a pixr */
    DGP_ERR_PASS_CLEANUP(packedtopixr(dgp_state, imageIFD->mask, 32, &maskPixr));

    /* scale if necessary  */
    if(scaleFactor != 1.0) {

      DGP_ERR_PASS_CLEANUP(
        w_scaleToSize(ipwstate, maskPixr, (Dgp_Int32)(width),
        (Dgp_Int32)(height), &tempPixr));
    
      /* destroy unscaled pixr */
      DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipwstate, &maskPixr));
      maskPixr = tempPixr;
      tempPixr = NULL;

    }

    /* set resolution */
    DGP_ERR_PASS_CLEANUP(
      w_pixrSetResolutionDPI(ipwstate, maskPixr, 
        Int32_TO_ScaledInt1616(xRes), Int32_TO_ScaledInt1616(yRes)));

    comp_mask_size = 0;
    comp_mask = NULL;
    DGP_ERR_PASS_CLEANUP(
      w_g4_enc_pixr (ipwstate, maskPixr, &comp_mask, &comp_mask_size));
    DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipwstate, &maskPixr));

  }
  
  /* Figure out which image plane to put it in. */
  if(imageIFD->mask) {
    plane = 5;
  } else {
    plane = BACKGROUND_PLANE;
  }
  
  /* samptoks haven't been compressed yet */
  if(imageIFD->samptok != NULL) {

    DGP_ASSERT(imageIFD->buf == NULL);

    /* we shouldn't be doing a samptok that has multiple instances */
    DGP_ASSERT(imageIFD->samptok->num_instances == 1);

    /* compress it, scaling as necessary */
    switch(sampCompType) {
    case DGPTDC_SAMP_COMPTYPE_JPEG:

      jpQuality = ((DgpTdc_JpegCompInfo *)(sampCompInfo))->jpQuality;
    
      if(imageIFD->samptok->flags & DGPTDC_SAMPTOK_IS_WASH) {
        jpQuality = WASH_JPQUALITY(jpQuality);
      }
    
      DGP_ERR_PASS_CLEANUP(compressRgbImageJpeg(dgp_state, imageIFD->samptok->im, 
        page->xRes, page->yRes, scaleFactor, jpQuality, 
        cinfo, &(imageIFD->buf), &(imageIFD->size)));

      break;

    default:
      DGP_ERRF0(cErrInternal, "invalid image compression type");
    }
    
    (imageIFD->samptok->num_instances)--;
    
    /* free original image */
    DGPIMAGE_FREE(imageIFD->samptok->im);
    imageIFD->samptok->im = NULL;

  }
  
#ifdef DBUG_SAMPIMS
  if(pageNum != page->num) {
    count = 0;
    pageNum = page->num;
  }
  
  sprintf(temp_str,"subIFD_p%.3d_%.2d.jpg",pageNum, count);
  fp = fopen(temp_str, "wb");
  fwrite(imageIFD->buf, 1, imageIFD->size, fp);
  fclose(fp);
  
  if(maskPixr != NULL) {
    sprintf(temp_str,"subIFD_mask_p%.3d_%.2d.pbm",pageNum, count);
    w_write(ipwstate, temp_str, maskPixr, PNM);
  }
  
  count++;
#endif
  
  /* by now it should be compressed. (and scaled) */
  switch(sampCompType) {
  case DGPTDC_SAMP_COMPTYPE_JPEG:

    if (imageIFD->mask)
    {

  IO_COLORSPACE g4_colorSpace = IO_COLORSPACE_DEFAULT;
  IO_COMPRESSION g4_compression = IO_COMP_MMR;
  IO_COLORTYPE g4_ioColorType = IO_COLORTYPE_BINARY;

      DGP_ERR_PASS_CLEANUP(IO_AddImage(IOState->page,
				     plane - 1, g4_ioColorType,
				     imageIFD->xpos, imageIFD->ypos, 
				     xRes, yRes, width, height,
				     height /*BestStripHeight*/, NULL, &outIoImg));
      DGP_ERR_PASS_CLEANUP(IO_ImageWriteStart(outIoImg, 0, &g4_compression,
						    &quality, &g4_colorSpace));

      DGP_ERR_PASS_CLEANUP(IO_CompressedWriteStrip (outIoImg, 0 /* row offset */,
				     height /* strip height */, comp_mask, 
				     comp_mask_size, &bytes_written));

      DGP_ERR_PASS_CLEANUP(IO_ImageWriteFinish (outIoImg));

      DGP_FREE (comp_mask);
    }

    DGP_ERR_PASS_CLEANUP(IO_AddImage(IOState->page,
				     plane, ioColorType,
				     imageIFD->xpos, imageIFD->ypos, 
				     xRes, yRes, width, height,
				     height /*BestStripHeight*/, NULL, &outIoImg));
    DGP_ERR_PASS_CLEANUP(IO_ImageWriteStart(outIoImg, 0, &compression,
						    &quality, &colorSpace));

    DGP_ERR_PASS_CLEANUP(IO_CompressedWriteStrip (outIoImg, 0 /* row offset */,
				     height /* strip height */, imageIFD->buf, 
				     imageIFD->size, &bytes_written));

    DGP_ERR_PASS_CLEANUP(IO_ImageWriteFinish (outIoImg));

/*    DGP_ERR_PASS_CLEANUP(
      writeXiffJpegMaskCompressed(imageIFD->xpos, imageIFD->ypos, xRes, 
        yRes, plane, imageIFD->buf, imageIFD->size, 
        width, height, maskPixr, xiffState)); */
    break;

  default:
    DGP_ERRF0(cErrInternal, "invalid image compression type");
  }

  DGP_FREE(imageIFD->buf);
  imageIFD->buf = NULL;


  if(maskPixr != NULL) {
    DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipwstate, &maskPixr));
    maskPixr = NULL;
  }
  
  NORMAL_RETURN;
  
ERR_CLEANUP:
  
  if(maskPixr != NULL) w_destroyPixr(dgp_state->ipwstate, &maskPixr);
  if(tempPixr != NULL) w_destroyPixr(dgp_state->ipwstate, &tempPixr);

  ERR_RETURN;
  
}

Dgp_Bool
overlaps(ImSubIFD *sub1, 
         ImSubIFD *sub2)
{
  Dgp_Int32 xpos1 = sub1->xpos;
  Dgp_Int32 xpos2 = sub2->xpos;
  Dgp_Int32 ypos1 = sub1->ypos - sub1->height + 1;
  Dgp_Int32 ypos2 = sub2->ypos - sub2->height + 1;
  
  return((xpos1 - sub2->width < xpos2) &&
    (xpos1 + sub1->width > xpos2) &&
    (ypos1 - sub2->height < ypos2) &&
    (ypos1 + sub1->height > ypos2));
}

ErrCode
writeImageSubIFDs(Dgp_State *dgp_state,
                  DgpTdc_PageInfo *page, 
                  IO_State *IOState, 
                  DgpAdt_List subIFDims, 
                  Dgp_Int32 sampCompType,
                  void *sampCompInfo, 
                  void **cinfo,
                  Dgp_Bool report) 
{
  DGP_PROLOGUE("writeImageSubIFDs()", DGP_PERRSTATE);
  
  ImSubIFD *subIFD;
  DgpAdt_ListNode ln;

  ln = DGPADT_LIFIRST(subIFDims);
  while(ln != DGP_NULLLISTNODE) {
    subIFD = (ImSubIFD *)DGPADT_LIGET(ln);
    
    /* Add it */
    /* If it has a mask then the mask will get added as well */
    DGP_ERR_PASS_CLEANUP(writeImageSubIFD(dgp_state, page, 
      IOState, subIFD, sampCompType, sampCompInfo, cinfo, report));
  
    SUBIFD_FREE(subIFD);
    subIFD = NULL;

    DgpAdt_liRemAndNext(ln, &ln);
  }
    
  NORMAL_RETURN;
  
ERR_CLEANUP:
  
  ERR_RETURN;
}

/*
* Write pos blocks and dictblocks to file
* set the "marked for death" flag in in the classesFound
* of the section block, if a class isn't put in the
* the dictionary
*/ 
ErrCode
DgpTdc_writeSection(DgpTdc_WriteCookie _cookie, DgpTdc_SectionBlock *sb, 
                    Dgp_Bool report) 
                    
{
  DGP_STATE_PROLOGUE2("DgpTdc_toutWriteSection()");
  tout_state *cookie;
  Dgp_State *dgp_state;
  Dgp_Allocators *a;
  Dgp_Int32 numPgs = sb->numPages;  
  DgpTdc_PageInfo *pages = sb->pages;
  DgpAdt_List tokOutput = sb->classesFound;
  DgpTdc_PageInfo *page;
  PosExtension posExt;
  /* halftone variables */
  void  *half_dictblock = NULL;
  
  /* writeCompressed variables */
  Dgp_Int32 page_num, exts, i = 0;
  token *toks = NULL;
  Dgp_Int32 num_non_singletons;
  Dgp_Int32 numToks;
  DictExtension dict_ext;
  Dgp_Int32  num_dict_ext=0;
  void  *tmp_dictblock = NULL;
  DgpAdt_ListNode ln;
  ImSubIFD *subIFD;
  Dgp_Int32 original_num_toks=0;

#if defined(DBUG_SAMPIMS) || defined(DBUG_COLORS)
  Dgp_Int32 j;
#endif
  
  /* variables for samp images */
  DgpAdt_List subIFDims = DGP_NULLLIST;
  colorDictInfo *color_dictinfo = NULL;

  /* IO stuff */
  IO_IMGHANDLE outIoImg, newFGImg;
  /* 19.3.2001 / dhtesq. */
  /* Changed colorspace from default to itulab. */
  IO_COLORSPACE colorSpace = IO_COLORSPACE_ITULAB;
  IO_COMPRESSION compression = IO_COMP_JBIG2;
  IO_COLORTYPE ioColorType = IO_COLORTYPE_BINARY;
  Dgp_UInt32 quality, imageLayer = 2, offset = 0, dictID, buf_size, dict_size;
  Dgp_UInt8 *buf, *dict;
  void * extdata = NULL;
  /* validate params */
  DGP_CHECK_NULL(_cookie);
  cookie = (tout_state *)_cookie;
  dgp_state = cookie->dgp_state;
  a = DGP_PALLOCATORS;
  DGP_ERR_BEGIN(dgp_state);

    DGP_REPORTF0(report, "Starting DgpTdc_writeSection.");

  /* check if we have room for another incremental dictionary */
  if (sb->final == DGP_FALSE && cookie->IOState->n_dict >= TIFFFX_MAX_DICT - 2)
  {
    sb->final = DGP_TRUE;
  }

  /* Create the array of tokens from the output from the tokenizer */
  original_num_toks = DGPADT_LILEN(tokOutput);
  
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&subIFDims, dgp_state));
  
    DGP_REPORTF0(report, "Creating Tokens.");

  DGP_ERR_PASS_CLEANUP(createTokensNew(dgp_state, tokOutput, sb, &toks, 
    &num_non_singletons, sb->final, 0, NULL));
  
  dict_ext.extdata= NULL;
  num_dict_ext = 0;
  
  /* write the main dictionary for this section 
     if there are some global tokens */
  if (num_non_singletons > 0) {   
    
    DGP_REPORTF0(report, "About to create the main dictionary.");
    
    tmp_dictblock = cookie->dictblock;
    DGP_ERR_PASS_CLEANUP(
      writeJb2DictionaryCB(
        toks, num_non_singletons, tmp_dictblock, &(cookie->dictblock),
	&dict_ext, num_dict_ext, cookie->jbig2State, NULL));

    /* free old dict block */
    if((tmp_dictblock != NULL) && 
      (tmp_dictblock != cookie->dictblock)) {
      DGP_REPORTF0(report,
        "About to free old dictblock.");
      
      freeJb2DictblockCB(tmp_dictblock, cookie->jbig2State, NULL);
      tmp_dictblock = NULL;
    }  
  }
  
  /* Write the halftone dictionary */
  if ((sb->halfOutput != DGP_NULLLIST) && 
    (DGPADT_LILEN(sb->halfOutput) > 0)) {
    
    DGP_ERR_PASS_CLEANUP(writeHalftoneDictionary(dgp_state, 
      &half_dictblock, cookie->jbig2State, sb->halfOutput, report));
    
  }
  
  /* Now write each page's private dictionary */
  for(page_num = 0; page_num < numPgs; page_num++) {
    page = &(pages[page_num]);

    DGP_REPORTF0(report, "Adding TIFF-FX page");
    
/* DGP_ERR_PASS_CLEANUP(newXiffPage(cookie->xiffState, cookie->ascii_ext)); */
    /* MRCIO calls to write it out */
    DGP_ERR_PASS_CLEANUP(IO_AddPageStart(cookie->IOState->doc,
				       -1 /* Append to the end */,
				       &cookie->IOState->page));
    DGP_ERR_PASS_CLEANUP(IO_AddImage(cookie->IOState->page,
				     imageLayer, ioColorType,
				     offset, offset, page->xRes,
				     page->yRes, page->width, page->height,
				     page->height /*BestStripHeight*/, NULL, &outIoImg));
    DGP_ERR_PASS_CLEANUP(IO_ImageWriteStart(outIoImg, 0, &compression,
						    &quality, &colorSpace));

    /* save the global dictionary with the first page */
    if (page_num == 0 && num_non_singletons > 0)
    {

    DGP_REPORTF0(report, "About to write the main dictionary.");
      
      DGP_ERR_PASS_CLEANUP(getJb2CompressedData (cookie->jbig2State, 
        &dict, &dict_size));
      DGP_ERR_PASS_CLEANUP(
        IO_StoreSymbolDictionary (cookie->IOState->doc, dict,
        dict_size, dict_size, &dictID));
      DGP_FREE (dict);

      cookie->IOState->dict_list[cookie->IOState->n_dict] = dictID;
      cookie->IOState->n_dict++;    
    }

    DGP_REPORTF1(report,
      "About to write page info for page %d.",page_num);

    /* start with Page Info segment */
    DGP_ERR_PASS_CLEANUP (writeJb2PageInfo(
					   page->width,
					   page->height,
					   page->xRes,
					   page->yRes,
					   cookie->jbig2State,
					   NULL));
    
    /* 
    cookie->dictblock is the dict block associated with the main dictionary.
    We need to merge the main dictionary with the dictionary for
    this page and get a dictblock for that union (since the positions
    for this page make reference to the main dictionary and the
    dictionary for this page both). 
    */
    tmp_dictblock = cookie->dictblock;
    if(page->numToks > 0) {

      DGP_REPORTF0(report, 
        "About to write per-page dictionary.");
      
      dict_ext.extdata = NULL;
      num_dict_ext = 0;

      DGP_ERR_PASS_CLEANUP(
        writeJb2PageDictionaryCB(page->toks, page->numToks, 
        page->width, page->height, page->xRes, page->yRes, cookie->ascii_ext,
	cookie->dictblock, &tmp_dictblock, &dict_ext, num_dict_ext, 
        cookie->jbig2State, NULL));

      
    } 
	posExt.extdata = &extdata;
	posExt.extlen = 0;

    if (!page->color) { 
            
    /*
    * This is a bit tricky:
    * We need to write a pos block on every page regardless of whether 
    * there are positions or not.
    * Look at the report prints for the types of pos blocks written.
    * If half_dictblock is not null then a halftone pos block will
    * be written, so we don't need to write an extra empty pos block
      */
      if (page->numToks > 0 || cookie->dictblock || !half_dictblock) {
        
        if(page->numToks > 0) {
          DGP_REPORTF0(report, 
            "About to write position block dependent on global and per page dicts.");
        } else if (cookie->dictblock) {
          /* notice this could be a blank page */
          DGP_REPORTF0(report,
            "About to write position block dependent on global dict only.");
        } else { /* !half_dictblock must be true */
          DGP_REPORTF0(report,
            "About to write empty position block dependent on no dictionaries.");
        }                  

        DGP_ERR_PASS_CLEANUP(
          writeJb2PosBlockCB(page->boxes, page->numBoxes,
          page->width, page->height, page->xRes, page->yRes,
          NULL, 0, tmp_dictblock, DGP_FALSE, cookie->jbig2State, NULL));

        freeTokens(dgp_state, page->toks, page->numToks);
        page->toks = NULL;
        page->numToks = 0;
        
      }
      /*
      * Write halftone pos every page (even if its empty) so Walt 
      * doesn't throw away the halftone dictionary    
      */
      if (half_dictblock) {
        DGP_REPORTF0(report,
          "About to write position block for the halftone image(s).");
        
        DGP_ERR_PASS_CLEANUP(
          writeJb2PosBlockCB(page->halftone_boxes, page->halftone_numBoxes,
          page->width, page->height, page->xRes, page->yRes,
          NULL, 0, half_dictblock, DGP_FALSE, cookie->jbig2State, NULL));
      }
    } else {  /* color */
      
#ifdef DBUG_COLORS
      if(report) {
        for(j=0; j<page->numBoxes; j++) {
          if(page->colors[j].color_type != SAMP_CLASS) {
            DGP_REPORTF1(report,
              "Token Color%X", page->colors[j].value.rgb_color);
          }
        }
      }
#endif 
      
#ifdef DBUG_SAMPIMS
      if(report) {
        for(j=0; j<page->numBoxes; j++) {
          if(page->colors[j].color_type == SAMP_CLASS) {
            DGP_REPORTF1(report,
              "Image Index: %d", page->colors[j].value.samp_class->samptoknum);
          }
        }
      }
#endif
      
      /* write the positions and any extensions to the positions.
      * NOTE that we're telling writePosBlock that boxes[i] has 
      * numBoxes[i] + num_pos_singletons[i] elements in it because
      * it has numBoxes[i] non-singletons followed by num_pos_singletons[i]
      * elements in it. 
      */
      if(page->numBoxes != 0) {
/*        DGP_ERR_PASS_CLEANUP(
          writeXiffColorPosBlock(page->boxes,page->numBoxes,page->width, 
          page->height, page->xRes, page->yRes, 
          NULL, 0, tmp_dictblock, DGP_FALSE, page->colors, color_dictinfo, 
          cookie->xiffState)); */
          posExt.builder = (PosExtBuilder)encodeJbig2Colors;
          /* link boxes extdata with color_tags */
          for(i=0; i<page->numBoxes; i++){
            page->boxes[i].extdata = &page->colors[i];
          }
          DGP_ERR_PASS_CLEANUP(initColors((color_cookie **)(&(posExt.cookie)), report, dgp_state));
          DGP_ERR_PASS_CLEANUP(
          writeJb2PosBlockCB(page->boxes, page->numBoxes,
                              page->width, page->height, page->xRes, page->yRes,
                              &posExt, 0, tmp_dictblock, DGP_FALSE, cookie->jbig2State, NULL));

            freeTokens(dgp_state, page->toks, page->numToks);
            cleanupColors (posExt.cookie);
            page->toks = NULL;
            page->numToks = 0;
      } else {
        /* write an empty block for the page */
        
        DGP_ERR_PASS_CLEANUP(
          writeJb2PosBlockCB(page->boxes, page->numBoxes,
          page->width, page->height, page->xRes, page->yRes,
          NULL, 0, tmp_dictblock, DGP_FALSE, cookie->jbig2State, NULL));
      }
            
    } /*end of else(color) */
        
    /* free per-page dictblock */
    if((tmp_dictblock != NULL) && 
      (tmp_dictblock != cookie->dictblock)) {
      DGP_REPORTF0(report,
        "About to free per page dictblock.");
      
      freeJb2DictblockCB(tmp_dictblock, cookie->jbig2State, NULL);
      tmp_dictblock = NULL;
    }  

    /* write the annotations if there are any */
    if(cookie->ascii_ext != DGP_FALSE) {
      
      /* make it a jbig2 binary extension */
      /* this should be done on the printer driver level ! FIXME */
      if (page->annotations)
      {
        for (exts = 0; exts < page->num_annotations; exts++)
        {
          if (page->annotations[exts].tag == 22 /*XIF TEXT_ANNOTATION_TAG*/)
          {
            page->annotations[exts].tag = DGP_CBOX_EXT;
          }
        }
      }
      DGP_REPORTF2(report,
        "About to write %d annotation on page %d", page->num_annotations, i);
      DGP_ERR_PASS_CLEANUP(writeJb2AnnotationsCB(page->num_annotations, 
        page->annotations, cookie->jbig2State, NULL));
    }
    
    
    /* finish up */
    /*    DGP_ERR_PASS_CLEANUP(finishXiffPage(cookie->xiffState)); */
    DGP_ERR_PASS_CLEANUP(finishJb2PageCB(cookie->jbig2State, NULL));  
    DGP_ERR_PASS_CLEANUP(getJb2CompressedData (cookie->jbig2State, 
					  &buf, &buf_size));
            /* allocate output data buffers */
    if(posExt.extlen > 0 && *(posExt.extdata) != NULL) {
        DGP_CHECK_ALLOC (buf = (Dgp_UInt8*) a->realloc (buf, buf_size + posExt.extlen));
        memcpy(buf+ buf_size, *(posExt.extdata), posExt.extlen);
    }else{
        posExt.extlen =0;
        *(posExt.extdata) = NULL;
    }
    DGP_ERR_PASS_CLEANUP(IO_StorePositionBlock (outIoImg, buf,
		                    buf_size + posExt.extlen, cookie->IOState->dict_list, 
                            (Dgp_UInt16) cookie->IOState->n_dict));
    
    DGP_ERR_PASS_CLEANUP(IO_ImageWriteFinish(outIoImg));
    if (page->color)
    {
      DGP_ERR_PASS_CLEANUP(IO_ShareColorTokenImage(outIoImg, buf_size, IO_COLORTYPE_COLOR24, &newFGImg));
      if(*(posExt.extdata) != NULL) a->free(*(posExt.extdata));
      *(posExt.extdata) = NULL;
      posExt.extdata = NULL;
      /* Write the subIFDs for the page */
      if(page->largeImages != DGP_NULLLIST) {
        DgpAdt_liApp(subIFDims, page->largeImages);
        page->largeImages = DGP_NULLLIST;
      }
      
      DGP_ERR_PASS_CLEANUP(writeImageSubIFDs(dgp_state, 
        page, cookie->IOState, subIFDims, 
        sb->sampCompType, sb->sampCompInfo, &(sb->cinfo), report));
    }

    DGP_ERR_PASS_CLEANUP(IO_AddPageFinish(cookie->IOState->page));

    DGP_FREE ((void*) buf);
    if (i == 0 && num_non_singletons > 0){
		DGP_FREE (dict);
	}

    if(page->color && page->colors) {
      DGP_FREE(page->colors);
      page->colors = NULL;
    }
    
    if(page->boxes != NULL) {
      DGP_FREE(page->boxes);
      page->boxes = NULL;
    }
    
    if(page->halftone_boxes != NULL) {
      DGP_FREE(page->halftone_boxes);
      page->halftone_boxes = NULL;
    }    
    
    if(page->num_annotations > 0) {
      DGP_FREE (page->annotations->data);
      DGP_FREE (page->annotations);
      page->annotations = NULL;      
      page->num_annotations = 0;
    }

  } /* end of the for loop */
  
  /* free color_dictinfo */
  if(color_dictinfo != NULL) {
    /*    freeXiffColorDictinfo(color_dictinfo, cookie->xiffState); */
    color_dictinfo = NULL;
  }
  
  if(half_dictblock != NULL) {
    freeJb2DictblockCB(half_dictblock, cookie->jbig2State, NULL);
    half_dictblock = NULL;
  }
  
  if(sb->pBlock != NULL) {
    if(sb->pBlock->poss != NULL) {
      DGP_FREE(sb->pBlock->poss);
    }
    
    DGP_FREE(sb->pBlock);
    sb->pBlock = NULL;
  }
  
  
  /*
  * Add global tokens to the section block list
  * We could almost free their runs here
  * Except if there is an image that matches a written
  * token it will need the runs to make its mask
  * so we must keep them unless we improve this mask stuff
  */
  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(sb->tokBlockList, (void *)toks, NULL));
  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(sb->tokBlockList, (void *)num_non_singletons, NULL));
  
  if(sb->final) {  
 /*
  * The token finder has been stopped
  * thus we free the tokens in the dictionaries referenced by dictblock
  * will not be used later.  Therefore we ought to free this
  * dictblock.
  */
    if(cookie->dictblock != NULL) {
      freeJb2DictblockCB(cookie->dictblock,cookie->jbig2State, NULL);
      cookie->dictblock = NULL;
    }
    
    if(sb->classesFound != DGP_NULLLIST) {
      DgpLlTokFreeOutput(sb->classesFound, dgp_state);
    }
    sb->classesFound = DGP_NULLLIST;
    
    /* All arrays on each page have been freed */
    /* so set the num of pages to zero */
    sb->numPages = 0;
    
    ln = DGPADT_LIFIRST(sb->tokBlockList);
    while(ln != DGP_NULLLISTNODE) {
      toks = (token *)DGPADT_LIGET(ln);
      DgpAdt_liRemAndNext(ln, &ln);
      numToks = (Dgp_Int32)DGPADT_LIGET(ln);
      DgpAdt_liRemAndNext(ln, &ln);
      
      freeTokens(dgp_state, toks, numToks);
    }
    DgpAdt_liFree(sb->tokBlockList);
    sb->tokBlockList = DGP_NULLLIST;

    /* free dictblock and set the dictcount to 0 */
    freeJb2DictblockCB (NULL, cookie->jbig2State, NULL);
    cookie->IOState->n_dict = 0;
    
  }
  
  /* cleanup subifds if necessary */
  DGPADT_FOREACH(ln, subIFDims) {
    subIFD = (ImSubIFD *)DGPADT_LIGET(ln);
    DGPADT_LIPUT(ln, NULL);

    if(subIFD != NULL) {
      if(subIFD->mask != NULL) {
        DGPIMAGE_FREE(subIFD->mask);
      }
      if(subIFD->buf != NULL) {
        DGP_FREE(subIFD->buf);
      }
      SUBIFD_FREE(subIFD);
    }
  }
  DgpAdt_liFree(subIFDims);
  

  NORMAL_RETURN;
  
ERR_CLEANUP:
  
  DGP_REPORTF0(report, "Something wrong in write Section");

  /* give up?!?! */
  if(*(posExt.extdata) != NULL) a->free(*(posExt.extdata));
  *(posExt.extdata) = NULL;
  posExt.extdata = NULL;
  ERR_RETURN;
} 

/*
* Close the file
*/
ErrCode
DgpTdc_writeEnd(DgpTdc_WriteCookie *_cookie)
{
  DGP_STATE_PROLOGUE2("DgpTdc_writeEnd()");
  tout_state *cookie;
  Dgp_Free tmp_free;
  
  if(_cookie == NULL) NORMAL_RETURN;
  
  cookie = *_cookie;
  
  if(cookie == NULL) NORMAL_RETURN;
  *_cookie = NULL;
  
  tmp_free = cookie->DGP_FREE;
  
  if(cookie->IOState != NULL) {
    DGP_ERR_PASS (IO_CloseDocument (cookie->IOState->doc));
    DGP_ERR_PASS (IO_EndInstance   (cookie->IOState->inst));
    tmp_free (cookie->IOState);
  }

  if(cookie->jbig2State != NULL) {
    DGP_ERR_PASS(closeJb2StateAPICB(cookie->jbig2State, NULL));
  }

  /* free the cookie */
  tmp_free(cookie);
  
  NORMAL_RETURN;
  
}
