/* cc_box.c
-- original GetBoxes() by ???; modified by Aaron Stump, summer '96 and
summer '97.
-- GetColorBoxes() by Aaron Stump, summer '96; modified by Scott
Cytacki, summer '97

  This file contains a function, GetBoxes(), for calling the IPCore
  CC-finder on a monochrome image, and a function, GetColorBoxes(),
  for calling Scott Cytacki's color CC finder on a qdr file.  
  
	pff '98: modified to build new type of color extension to position block
*/


/* includes, in alphabetical order */
#ifdef _WIN32
/* this is foul (for some reason, MSDEV is including <stdarg.h> here, and
<starg.h>'s definitions are conflicting with those in <varargs.h>; the 
following yecch kludge makes it happy). */
#define _INC_STDARG
#endif

#include "types.pub"
typedef struct StrelInfo *StrelInfoPtr;
#include "memory.pub"
#include "shros.pub"
#include "ipstate.pub"
#define _IPWSTATE_DEFINED_

#include "cc2box.h"
#include "colorCC.h"
#include "chunk.h"
#include "list.h"
#include "run2bitmap.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include "uds.h"
#include <color-extension.h>
#include "simple_utils.h"
#include "ipserrs.pub"
#include "shrutils.pub"
#include "iaerror.pub"
#include "shrcc.pub"
#include "dgp_utils.h"
#include "tokenCC.h"

/* We have to get the right include file for 
htons(), htonl() */
#ifndef _WIN32
/* #include "undecl-stdio.h" */
#endif

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcomponent/cc2box.c,v 1.18 2003/02/10 22:53:23 mkrocken Exp $");

DgpImage_RGB defaultBgColor={255, 255, 255};

#define Bit(b)         (1 << (b))
#define BitOn(v, b)    ((v) |= (0x80000000 >> (b)))

typedef struct _CCData {
	int count;
	TokState tokstate;
	tokpos *boxes;
	Dgp_Int32 num_boxes;
	Dgp_Int32 num_boxes_alloced;
	DgpAdt_List outliers;
	int maxWidth;
	int maxHeight;
	Dgp_State *dgp_state;
	TokInfoBlock *tBlock;
	PosInfoBlock *pBlock;
	int pageIndex;
} CCData;

#define INIT_SIZE_BOXES 5000
#define INIT_SIZE_SAMPBOXES 10

DGP_CHFUNCTIONS(cc_);

/*
* This is the callback function for the ipshared function 
* i_findConnectedComponentRuns.
*
* For big CCs, this function converts the runs to another flavor of runs,
* and throws a descriptor onto a list of outliers (inside (void *)data).
*
* For CCs small enough to tokenize, it throws them straight at the tokenizer
* (libtok-lossless lite), which gives it right back a token number.  Then
* a descriptor smaller than the one required for a CC is thrown onto a 
* different list (again, inside (void *)data).
*/
static Int32 
MakeModelsFromRuns(IPState *ipstate, void *data, Int32 x, Int32 y,
				   Int32 w, Int32 h, Int32 m, Int32 nRuns,
				   TCScanLineRun *pRuns)
{
	CCData *ccData = (CCData *)data;
	Dgp_State *dgp_state = ccData->dgp_state;
	DGPERR_INIT(MakeModelsFromRuns);
	TCScanLineRun *pRunsIndex;
	PosInfo *pInfo;
	PosInfoBlock *pBlock = ccData->pBlock;
	cc_ *newcc = (cc_ *)NULL;
	run_elt *tmpruns = (run_elt *)NULL;
	
	(void)m;
	(void)ipstate;
	/*
	* If this model is "too big" then don't build the bitmap -- keep the runs. 
	* Make one big area for them - more efficient. Just store 2 Int16's for 
	* each run - left and right.
	*/
	if ((w > ccData->maxWidth) || (h > ccData->maxHeight)) {
		Int16 *run;
		DgpAdt_ListNode retnode;
		
		DGP_ERR_PASS_CLEANUP(DgpAdt_malloc_cc_(dgp_state, &newcc));
		newcc->runs = NULL;
		
		newcc->bitmap.width = (Dgp_UInt16)w;
		newcc->bitmap.height = (Dgp_UInt16)h;
		newcc->bitmap.Dgp_Int32sperline =
			(newcc->bitmap.width + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
		newcc->bitmap.bits = (Dgp_UInt32 *)NULL;
		newcc->bitmap.userdata = newcc;
		
		newcc->runs = (Dgp_Int16 *)NULL;
		newcc->nruns = 0;
		/*
		* Where is this box? Not where it ought to be... there's a magic offset
		* of 32 because we're dealing with an ipshared routine here and sometimes
		* ipshared thinks there's a frame of 32 pixels at the top and left. Yay.
		*/
		newcc->x = x - 32;
		newcc->y = y - 32;
		
		DGP_SEC_MALLOC(newcc->runs, (Dgp_Int16 *), (2 * sizeof(*newcc->runs) * nRuns));
		newcc->nruns = nRuns;
		
		/*
		* Read out the runs. They were generated bottom-to-top, so reverse
		* them.
		*/
		for (run = newcc->runs + 2 * (nRuns - 1), pRunsIndex = pRuns;
	       pRunsIndex;
		   run -= 2, pRunsIndex = pRunsIndex->pNext) {
			   run[0] = (Int16)(pRunsIndex->runLeft - x);
			   run[1] = (Int16)(pRunsIndex->runRight - x);
		   }
		   /* Hook it onto the list of outliers */
		   DGP_ERR_PASS (DgpAdt_liAdd(ccData->outliers, (void *)newcc, &retnode));
		   newcc = NULL;
		   ccData->count++;
	}
	else {
		/* This is something we can tokenize */
		tokpos *newpos;
		runToken rTok;
		run_elt *run;
		
		rTok.width = (Dgp_Int16)w;
		rTok.height = (Dgp_Int16)h;
		
		if (ccData->num_boxes >= ccData->num_boxes_alloced) {
			ccData->num_boxes_alloced += ccData->num_boxes_alloced / 2;
			DGP_SEC_REALLOC(ccData->boxes, (tokpos *), (sizeof(tokpos) * ccData->num_boxes_alloced));
		}
		newpos = &(ccData->boxes[ccData->num_boxes]);
		
		/*
		* Where is this box? Not where it ought to be... there's a magic offset
		* of 32 because we're dealing with an ipshared routine here and sometimes
		* ipshared thinks there's a frame of 32 pixels at the top and left. Yay.
		*/
		newpos->xpos = x - 32;
		newpos->ypos = y - 32 + h - 1; /* we're going to shift this box's coordinates
									   from specifying the UL corner to the LL
		corner at this point in the execution */
		newpos->extdata = NULL;
		
		DGP_SEC_MALLOC(tmpruns, (run_elt *), (2 * sizeof(run_elt) * nRuns));
		rTok.runs = tmpruns;
		rTok.nruns = nRuns;
		
		/*
		* Read out the runs. They were generated bottom-to-top, so reverse
		* them.
		*/
		for (run = rTok.runs + 2 * (nRuns - 1), pRunsIndex = pRuns;
	       pRunsIndex;
		   run -= 2, pRunsIndex = pRunsIndex->pNext) {
			   run[0] = (Int16)(pRunsIndex->runLeft - x); /* normalized x-location */
			   run[1] = (Int16)(pRunsIndex->runRight - pRunsIndex->runLeft + 1); /* length */
		   }
		   
		   if(pBlock->num >= pBlock->numAlloced){
			   pBlock->numAlloced = pBlock->numAlloced * 3 / 2;
			   DGP_SEC_REALLOC(pBlock->poss, (PosInfo *), (pBlock->numAlloced * sizeof(PosInfo)));
		   }
		   
		   pInfo = &(pBlock->poss[pBlock->num]);
		   pInfo->page = ccData->pageIndex;
		   pInfo->posOffset = ccData->num_boxes;
		   
		   
		   rTok.userdata = (void *)pBlock->num;
		   
		   tokenAddCC_err((dgp_state, ccData->tokstate, ccData->tBlock, &rTok, &(newpos->tok)));
		   tmpruns = NULL;
		   
		   pBlock->num++;
		   ccData->num_boxes++;
		   
		   ccData->count++;
	}
	
	NORMAL_RETURN;
	
ERR_CLEANUP:
	
	if(newcc != NULL) {
		if(newcc->runs != NULL) DGP_FREE(newcc->runs);
		DgpAdt_free_cc_(dgp_state, newcc);
	}
	if(tmpruns != NULL) DGP_FREE(tmpruns);
	
	ERR_RETURN;
}


/* GetBoxes()

  Attempt to read a TIFF file and get the CCs from the resulting image.
  Throw small CCs at the (lossless) tokenizer, and store large CCs as
  outliers.
  
	Parameters:
	Dgp_State *dgp_state	-- a state cookie very similar to an ipstate, it contains
	the memory allocation functions, and error and report printing functions
	TokState tokstate    -- a state cookie acquired from a call to tokInit().
	UInt8 *image         -- a pointer to the (framed) image to find CCs in.
	Int32 xSize          -- the width of the image (not including frame)
	Int32 ySize          -- the height of the image (not including frame)
	Int32 bpl            -- the number of bytes per line (including frame)
	int *numBoxes        -- the number of tokpos's found will be passed back
	in *numBoxes.
	tokpos **boxes       -- the cc's found.  Null if no cc's where found
	int sizeThresh       -- CCs smaller than sizeThresh x sizeThresh will be
	neither tokenized nor put on the DgpAdt_List of outliers. 
	DgpAdt_List *outliers       -- the outliers (CCs which are too big to tokenize) will
	be passed back in *outliers (each outlier will be
	given by a cc struct -- see compress.h for the
	definition of a cc struct).  outliers may be NULL
	(so no DgpAdt_List will be returned in *outliers).
	TokInfoBlock *tBlock -- A vector of TokInfo structs.  Call initTokBlock (tokenCC.h)
	to get this.  GetBoxes puts an offset into the tBlock->toks array in the
	userdata field of each of the classes added to the tokenizer.  The TokInfo 
	at this offset, contains a pointer to the token if it is known, otherwise 
	NULL.  The token will only be known if you set it eariler.  (see the 
	comment in tokenCC.h for more info)
	PosInfoBlock *pBlock -- A vector of PosInfo structs.  Call initPosBlock (tokenCC.h)
	to get start one.  GetBoxes adds an offset into pBlock->poss to the members 
	list of each added class for each cc.  In otherwords, when a new cc is found 
	a new PosInfo is made in the pBlock.  The offset to this new PosInfo is added
	to the members list of the class in the tokenizer.  The PosInfo contains the 
	page index and an offset into the boxes array.
	int pageIndex		-- this is value is added to each PosInfo.  It is there so you 
	can have mutilple pages in one pBlock, that coorespond to a different array 
	of positions for each page.
	Dgp_Bool report		-- If true the report printer of the tcc is used to give some
	info about this function.
	
	  Return values:
	  return DGP_TRUE on success, and DGP_FALSE on failure.
	  */
#define INIT_NUM_TOKS 50
	  
	  ErrCode
		  GetBoxes(Dgp_State *dgp_state, TokState tokstate, UInt8 *image, Int32 xSize, Int32 ySize,
		  Int32 bpl, int *numBoxes, tokpos **boxes, int sizeThresh, DgpAdt_List *outliers, 
		  TokInfoBlock *tBlock, PosInfoBlock *pBlock, int pageIndex, Dgp_Bool report)
	  {
		  DGPERR_INIT(GetBoxes);
		  
		  Int32 returnValue;
		  CCData ccData;
		  IPState *ipstate = dgp_state->ipstate;
		  
#ifdef DBG
		  DGP_REPORTF3(report, "GetBoxes() starts with \n\
														   UInt8 *image = %ld\n\
														   Int32 xSize  = %d\n\
														   Int32 ySize  = %d",image,xSize,ySize); 
														   DGP_REPORTF1(report, "\
															   Int32 bpl    = %d", bpl);
#else
														   DGP_REPORTF0(report, "GetBoxes() starts.");
#endif
														   
														   /* Add token cookie */
														   ccData.dgp_state = dgp_state;
														   
														   /* init tokBlock */
														   ccData.tBlock = tBlock;
														   ccData.pBlock = pBlock;
														   
														   ccData.pageIndex = pageIndex;
														   
														   if (outliers != NULL)  
															   *outliers = NULL;
														   ccData.boxes = NULL; 
														   
														   DGP_ERR_PASS (DgpAdt_liNew(&ccData.outliers, dgp_state));
														   
														   ccData.tokstate = tokstate;
														   ccData.count = 0;
														   
														   ccData.num_boxes_alloced = INIT_SIZE_BOXES;
														   DGP_SEC_MALLOC(ccData.boxes, (tokpos *), 
															   sizeof(tokpos) * ccData.num_boxes_alloced);
														   
														   ccData.num_boxes = 0;
														   
														   DGP_ASSERT(image != (UInt8 *)NULL); 
														   
														   /* Decide what the cutoff for conversion to bitmap is */
														   ccData.maxWidth  = xSize;	/* 100% of page width */
														   ccData.maxHeight = ySize;	/* 100% of page height */
														   
																						/* NOTE that we just effectively turned off the feature whereby
																						only those bitmaps smaller than a certain size are eligible
														   to be tokenized.  So we'll be trying to tokenize CCs of any size. */
														   
														   DGP_REPORTF0(report,  "About to call ipcore CC finder.");
														   
														   /* Get all the CCs */
														   DGP_ERR_PASS_CLEANUP(i_findConnectedComponentRuns(ipstate,
															   image, (unsigned)bpl, 32, 32,
															   xSize, ySize,
															   (unsigned)sizeThresh, (unsigned)sizeThresh,
															   1,
															   (void *)&ccData,
															   MakeModelsFromRuns, 
															   &returnValue));
														   
														   DGP_REPORTF1(report,  "Found %d ccs", ccData.count);
														   
														   /* shrink the array to the min. size required (it might have more tokpos's
														   in it than just ccData.num_boxes of them) */
														   if (ccData.num_boxes > 0){
															   DGP_SEC_REALLOC(ccData.boxes, (tokpos *), ccData.num_boxes * sizeof(tokpos));
															   
														   } else {
															   DGP_FREE(ccData.boxes);
															   ccData.boxes = NULL;
														   }
														   
														   *numBoxes = ccData.num_boxes;
														   if (outliers != NULL)  
															   *outliers = ccData.outliers;
														   
														   DGP_REPORTF0(report, "GetBoxes() ends (ok).");
														   *boxes = ccData.boxes;
														   
														   NORMAL_RETURN;
														   
ERR_CLEANUP:
														   if (ccData.boxes)
															   DGP_FREE((void *)ccData.boxes);
														   
														   if (ccData.outliers)
															   DgpAdt_liFree(ccData.outliers);
														   
														   ERR_RETURN;
	  }
	  
	  /* Return DGP_TRUE if the mask represented by the given runs is rectangular;
	  return DGP_FALSE otherwise. */
		  static Dgp_Bool
			  isRectangular(run_elt *runs, int num_runs) {
			  int runLeft, runLength, oldRunLeft, oldRunLength;
			  int i;
			  
			  oldRunLeft   = RunX(runs,0);
			  oldRunLength = RunLen(runs,0);
			  
			  for (i=1; i < num_runs; i++) {
				  runLeft   = RunX(runs,i);
				  runLength = RunLen(runs,i);
				  if (runLeft != oldRunLeft)
					  return DGP_FALSE;
				  if (runLength != oldRunLength)
					  return DGP_FALSE;
			  }
			  return DGP_TRUE;
		  }
		  
