/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libcomponent/cc2box.h,v 1.7 2000/03/07 22:38:31 mkrocken Exp $
 *
 */

#ifndef CC2BOX_H
#define CC2BOX_H

#include "bit_array.h"
#include "list.h"
#include "mmcompress.h"
#include "samptok.h"
#include "tok-lossless.h"
#include "types.pub"
#include "uds.h"
#include <color-extension.h>
#include "dgp_utils.h"
#include "tokenCC.h"

#ifdef __cplusplus
extern "C"
{
#endif

extern DgpImage_RGB defaultBgColor;

/* tokBitmap used to be defined and used by the module giving the lossless 
   tokenizer, but now that tokenizer works exclusively with runs, so we
   need to define tokBitmap here ourselves. */
typedef struct {
    Dgp_UInt16 width, height;
    Dgp_UInt32 *bits;
    Dgp_UInt16 Dgp_Int32sperline;
    void *userdata;
    } tokBitmap;

typedef struct {
    tokBitmap bitmap;
    int x, y;
    Dgp_Int16 *runs;
    int nruns;
    int toknum;
    } cc_;

extern ErrCode DgpAdt_malloc_cc_(Dgp_State *dgp_state, cc_ **cc);
extern void DgpAdt_free_cc_(Dgp_State *dgp_state, cc_ *cc);

/* GetBoxes()
   
   Attempt to read a TIFF file and get the CCs from the resulting image.
   Throw small CCs at the (lossless) tokenizer, and store large CCs as
   outliers.
   
   Parameters:
   TokState tokstate -- a state cookie acquired from a call to tokInit().
   UInt8 *image      -- a pointer to the (framed) image to find CCs in.
   Int32 xSize       -- the width of the image (not including frame)
   Int32 ySize       -- the height of the image (not including frame)
   Int32 bpl         -- the number of bytes per line (including frame)
   int *numBoxes     -- the number of tokpos's found will be passed back
                        in *numBoxes.
   tokpos **boxes    -- the cc's found.  Null if no cc's where found
   int sizeThresh    -- CCs smaller than sizeThresh x sizeThresh will be
                        neither tokenized nor put on the DgpAdt_List of outliers. 
   DgpAdt_List *outliers    -- the outliers (CCs which are too big to tokenize) will
                        be passed back in *outliers (each outlier will be
			given by a cc struct -- see compress.h for the
			definition of a cc struct).  outliers may be NULL
			(so no DgpAdt_List will be returned in *outliers).
   
   Return values:
   return DGP_TRUE on success, and DGP_FALSE on failure.
 */
#define GetBoxes_err(args) \
	DGP_ERR_PASS_CLEANUP(GetBoxes##args)

ErrCode
GetBoxes(Dgp_State *dgp_state, TokState tokstate, UInt8 *image, Int32 xSize, Int32 ySize,
	 Int32 bpl, int *numBoxes, tokpos **boxes, int sizeThresh, DgpAdt_List *outliers, 
	 TokInfoBlock *tBlock, PosInfoBlock *pBlock, int pageIndex, Dgp_Bool report);

#ifdef __cplusplus
}
#endif

#endif
