#include <util.h>
#include "colorCC.h"


typedef struct _Set_ {
  int rank;
  struct _Set_ *parent;
  void *data;
} Set;

typedef struct {
  Dgp_Int8 sampled;
  Dgp_Int8 wash;
  Dgp_Bool added;
  Dgp_Bool onTop;
  Dgp_Bool fillAll;
  Dgp_Bool added2wash;
  Dgp_Bool active;
  Dgp_Bool isRect;
  Dgp_UInt8 del;
  union {
	  DgpImage_Long cmyk;
	  DgpImage_ColorRGB rgb;
  } sampling;
  union {
	  Dgp_UInt32 cmyk;
	  DgpImage_RGB rgb;
  } color;
  Dgp_UInt16 ulX, ulY, lrX, lrY;
  Dgp_UInt16 *runDgpAdt_List;
  Dgp_UInt16 *currPos;
  Dgp_UInt32 size;
  Dgp_UInt32 number;
  Dgp_UInt32 area;
  int numHoles;
  Dgp_UInt16 lastY, lastX;
  DgpAdt_List washConnections;
  Set *washLabel;
} CCold;

typedef struct _Run {
  cccRun cccrun;
  Dgp_Int8 wash;
  Dgp_UInt8 matchesBelow;
  Dgp_UInt16 xPos;
  Set *label;
  struct _Run *lowerRun;
} Run;


#ifdef RGB_FORMAT
#define	CSPACE			rgb


#define addCCToDgpAdt_List		addCCToDgpAdt_List_rgb
#define labelRun		labelRun_rgb
#define ccc_doStrip		ccc_doStrip_rgb
#define fillAndAdd		fillAndAdd_rgb
#define labelRun		labelRun_rgb
#define readAndLabel	readAndLabel_rgb
#define sampling_		sampling.rgb

#define runsMatch(run1, run2) \
	(((run1)->cccrun.sampled == (run2)->cccrun.sampled) && \
     ((run1)->wash == (run2)->wash) && \
	  ((run1)->cccrun.color == (run2)->cccrun.color))

#define colorsMatch(col1, col2) \
  (((col1).r == (col2).r) && ((col1).g == (col2).g) && ((col1).b == (col2).b))

/*						             ffrrggbb */
#define DEFAULT_BG_COLOR 0x00FFFFFF

#define frgb_f(color) \
	(((color >> 24)) & 0xFF000000)
#define frgb_r(color) \
	(((color >> 16)) & 0x00FF0000)
#define frgb_g(color) \
	(((color >> 8)) & 0x0000FF00)
#define frgb_b(color) \
	((color) & 0x000000FF)

#define copyToSampling(sampling, color) \
	(sampling).r = (Dgp_UInt8)(frgb_r(color)); \
	(sampling).g = (Dgp_UInt8)(frgb_g(color)); \
	(sampling).b = (Dgp_UInt8)(frgb_b(color))


#define SAMP_IMG_TYPE DGPIMAGE_RGB
typedef DgpImage_RGB Color;

extern INLINE int colorDist(Dgp_UInt32 color1, Dgp_UInt32 color2);


#else /*!RGB_FORMAT */
#define CSPACE			cmyk

#define addCCToDgpAdt_List		addCCToDgpAdt_List_cmyk
#define labelRun		labelRun_cmyk
#define ccc_doStrip		ccc_doStrip_cmyk
#define fillAndAdd		fillAndAdd_cmyk
#define labelRun		labelRun_cmyk
#define readAndLabel	readAndLabel_cmyk
#define sampling_		sampling.cmyk

#define runsMatch(run1, run2) \
	(((run1)->cccrun.sampled == (run2)->cccrun.sampled) && \
     ((run1)->wash == (run2)->wash) && \
	  ((run1)->cccrun.color == (run2)->cccrun.color))

#define colorsMatch(col1, col2) \
  ((col1) == (col2))

/*						   ccmmyykk */
#define DEFAULT_BG_COLOR 0x00000000
#define cmyk_c(color) \
	((color & 0xFF000000) >> 24)
#define cmyk_m(color) \
	((color & 0x00FF0000) >> 16)
#define cmyk_y(color) \
	((color & 0x0000FF00) >> 8)
#define cmyk_k(color) \
	(color & 0x000000FF)

#define copyToSampling(sampling, color) \
	 (sampling) = (color)

typedef Dgp_UInt32 Color;

#define colorDistNat colorDist
#define SAMP_IMG_TYPE DGPIMAGE_LONG

extern INLINE int colorDist(Color color1, Color color2);

#endif /*!RGB_FORMAT */


#define getWashLabel(run) \
	(((CCstate*)(findSet((run)->label)->data))->washLabel)


typedef struct {
  Dgp_Bool sampled;
  Dgp_Bool wash;
  Dgp_Bool active;
  Dgp_Bool added2wash;
  Dgp_UInt16 ulX, ulY, lrX, lrY;
  int number;
  int area;
  int numHoles;
  DgpAdt_List washConnections;
  Set *washLabel;
} CCstate;

typedef struct {
	int currLine;
	int currX;
	Run nextRun;
} *read_cookie; 

typedef struct {
	int length;
	int offset;
} RunRow;

typedef struct {
	int width, height;
	int size;
	RunRow *rows;
	Run *runs;
	Dgp_Allocators *a;
} *RunImage;

#define rimGetRow(im, y) ((im)->runs + (im)->rows[(y)].offset)

/* rows has an extra space so we don't hit invalid memory here  */
#define rimEndRow(im, y, numRuns) \
	(im)->rows[y].length = (numRuns); \
	(im)->rows[(y) + 1].offset = (im)->rows[y].offset + (numRuns)


extern INLINE Run *rimStartRow(RunImage im, int y);

typedef struct{
	Dgp_State *dgp_state;
	PFNReadRun readRun;
	int width, height;
	DgpAdt_List ccSets;
	DgpAdt_List washSets;
	Run defaultRun;
	int curLine;
	RunImage im;
	int minSampSize;
	int wThresh;
	read_cookie rCook;
	Set *curNode;
	Set *nodeBlock;
	int nBlockSize;

  int maxMemUsage;  /* memory usage suggestion */

} ccc_state, *ccc_cookie;


/* colorCCfill */
int fillAndAdd(ccc_cookie cookie, DgpAdt_List ccDgpAdt_List, DgpAdt_List sampCCDgpAdt_List,
			   DgpAdt_List washCCDgpAdt_List);

/* colorCCread.c */
ErrCode readRunNoWash(read_cookie rCook, Run *currRun, ccc_cookie cookie, RunCookie runCook);
ErrCode readRunWash(read_cookie rCook, Run *currRun, ccc_cookie cookie, RunCookie runCook);
read_cookie initReadRuns(ccc_cookie cookie);

/* colorCCrim.c */
RunImage rDgpImage_new(int width, int height, Dgp_Allocators *a);
void rDgpImage_free(RunImage im);

/* colorCCwash.c */
ErrCode readAndLabelWash(ccc_cookie cookie, RunCookie runCook, int numLines);
void findHardEdges(ccc_cookie cookie);

extern INLINE int runsMatchWash(Run *run1, Run *run2);
extern INLINE int runsMatchWashNoDiagonal(ccc_cookie cookie, Run *run1, Run *run2);

#define initRun(run, x) \
	(run).xPos = x; \
	(run).lowerRun = NULL; \
	(run).label = NULL; \
	(run).wash = 0; \
	(run).matchesBelow = DGP_FALSE; \
	if(!(run).cccrun.sampled){ \
		(run).cccrun.sampling_ = NULL; \
	} 


#define mineq(x,y) ((x)=(x)<(y)?(x):(y))
#define maxeq(x,y) ((x)=(x)>(y)?(x):(y))
#define MAX_NUM_HOLES_TO_DISCARD 4
#define MAX_NUM_HOLES_CHAR 4

/* colorCC.c */
extern INLINE Set *makeSet (ccc_cookie cookie);

extern INLINE Set *findSet(Set *x);

extern INLINE Set *linkSetWash(Set *x, Set *y);


#define linkCCtoCC(ccX, ccY) \
		mineq((ccX)->ulX, (ccY)->ulX); \
		mineq((ccX)->ulY, (ccY)->ulY); \
		maxeq((ccX)->lrX, (ccY)->lrX); \
		maxeq((ccX)->lrY, (ccY)->lrY); \
		(ccX)->area += (ccY)->area; \
		(ccX)->number += (ccY)->number; \
		(ccX)->numHoles += (ccY)->numHoles; \
		(ccX)->wash = (ccX)->wash || (ccY)->wash; \
		DgpAdt_liApp((ccX)->washConnections, (ccY)->washConnections); \
		linkSetWash(findSet((ccX)->washLabel), findSet((ccY)->washLabel)); \
		free(ccY)


#define linkCCtoRun(cc, run, y) \
		(cc)->number++; \
		(cc)->area += (run)->cccrun.length; \
		mineq((cc)->ulX, (run)->xPos); \
		maxeq((cc)->lrX, ((run)->xPos + (run)->cccrun.length - 1)); \
		(cc)->lrY = (y); \
		(cc)->wash = (cc)->wash || (run)->wash

extern INLINE Set *linkSetCC(Set *x, Set *y);

extern INLINE CCstate *makeCC (ccc_cookie cookie, Run *currRun, int y);
extern INLINE CC *makeCCnew(ccc_cookie cookie, Run *currRun, int y);
