/*
@(#)$Header: /project/icons/cvsroot/digipaper/src/libcomponent/colorCC.c,v 1.20 2003/02/10 22:53:23 mkrocken Exp $ 
*/
#include "colorCC-priv.h"
#include <stdlib.h>
#include <sys/types.h>
#include <image.h>
#include <list.h>
#include "colorCC.h"
#include "dgp_utils.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcomponent/colorCC.c,v 1.20 2003/02/10 22:53:23 mkrocken Exp $");

/* We have to get the right include file for 
htons(), htonl() */
#ifndef _WIN32
/* #include "undecl-stdio.h" */
#endif

INLINE Run *rimStartRow(RunImage im, int y)
{
  if((im->size - im->rows[y].offset)  < im->width){ 
    im->size += im->width * 10; 
    im->runs = im->a->realloc(im->runs, im->size * sizeof(Run)); 
    if(im->runs == NULL){ 
      return NULL;
    } 
  } 
	
  return rimGetRow(im, y);
}

INLINE CC *convertCC (ccc_cookie cookie, CCstate *ccState)
{
  CC *cc;
  int width, height;

  cc = (CC *)cookie->DGP_MALLOC(sizeof(CC));
  if(cc == NULL){
    return cc;
  }

  width = ccState->lrX - ccState->ulX + 1;
  height = ccState->lrY - ccState->ulY + 1;

  cc->ulX = ccState->ulX;
  cc->ulY = ccState->ulY;
  cc->lrX = ccState->lrX;
  cc->lrY = ccState->lrY;

  cc->lastY = 0;
  cc->lastX = -1;
  cc->runDgpAdt_List = NULL;
  cc->numRuns = 0;
  cc->added = DGP_FALSE;
  cc->holes = ccState->numHoles > 0;
  cc->large = width > (cookie->width / 10) || height > (cookie->height / 10);
  cc->save = !cc->large && cc->holes && 
    (ccState->numHoles < MAX_NUM_HOLES_CHAR);
  cc->del = DGP_FALSE;
  cc->runDgpAdt_ListLen = ccState->number;
  cc->del = DGP_FALSE;

  cc->sampling = NULL;
  if(ccState->sampled || ccState->wash){
    cc->sampling = (SampState*)cookie->DGP_MALLOC(sizeof(SampState));
    if(cc->sampling == NULL){
      cookie->DGP_FREE(cc);
      return NULL;
    }
	
    cc->sampling->above = DGP_FALSE;
    cc->sampling->below = DGP_FALSE;
    cc->sampling->split = ((width < cookie->minSampSize) ||
			   (height < cookie->minSampSize));
    cc->sampling->wash = ccState->wash;
    cc->sampling->image.cmyk = NULL;
    /*
     * Need to fill filledRunLength, real* stuff
     */
  }

  cookie->DGP_FREE(ccState);	

  return cc;
}

INLINE CC *makeCCnew(ccc_cookie cookie, Run *currRun, int y)
{
  CC *cc;

  cc = (CC *)cookie->DGP_MALLOC(sizeof(CC));
  if(cc == NULL){
    return cc;
  }

  cc->holes = DGP_FALSE;
  cc->save = DGP_TRUE;
  cc->large = DGP_FALSE;
  cc->added = DGP_FALSE;
  cc->del = DGP_FALSE;
  cc->ulX = currRun->xPos;
  cc->lrX = currRun->xPos + currRun->cccrun.length - 1;
  cc->ulY = y;
  cc->lrY = y;
#ifdef RGB_FORMAT
  cc->color.rgb.r = currRun->cccrun.color >> 16 && 0xFF;
  cc->color.rgb.g = currRun->cccrun.color >> 8 && 0xFF;
  cc->color.rgb.b = currRun->cccrun.color && 0xFF;
#else 
  cc->color.cmyk = currRun->cccrun.color;
#endif
  cc->lastY = 0;
  cc->lastX = -1;
  cc->runDgpAdt_List = NULL;
  cc->runDgpAdt_ListLen = 1;
  cc->numRuns = 0;
  cc->sampling = NULL;

  return cc;
}


INLINE CCstate *makeCC (ccc_cookie cookie, Run *currRun, int y) 
{
  CCstate *cc;

  cc = (CCstate *)cookie->DGP_MALLOC(sizeof(CCstate));
  if(cc == NULL){
    return cc;
  }

  cc->area = currRun->cccrun.length;
  cc->number = 1;
  cc->ulX = currRun->xPos;
  cc->lrX = currRun->xPos + currRun->cccrun.length - 1;
  cc->ulY = y;
  cc->lrY = y;
  cc->sampled = currRun->cccrun.sampled;
  cc->wash = currRun->wash;
  cc->numHoles = 0;
  DgpAdt_liNew(&cc->washConnections, cookie->dgp_state);
  cc->washLabel = NULL;
  cc->added2wash = DGP_FALSE;

  return cc;
}

INLINE Set *makeSet (ccc_cookie cookie)
{
  Set *node;
  DgpAdt_ListNode retnode;
  
  if((cookie->curNode - cookie->nodeBlock) == cookie->nBlockSize){
    cookie->curNode = cookie->nodeBlock = (Set *)cookie->DGP_CALLOC(sizeof(Set) * cookie->nBlockSize, 1);
    if(cookie->curNode == NULL){
      return NULL;
    }
	DgpAdt_liAdd(cookie->ccSets, cookie->nodeBlock, &retnode);
    if(retnode == DGP_NULLLISTNODE){
      return NULL;
    }
  }

  node = cookie->curNode;
  node->rank = 0;
  node->parent = node;

  node->data = NULL;

  cookie->curNode++;	
  return node;
}

INLINE Set *findSet(Set *x)
{
  Set *y = x;
  while(y != y->parent){
    y = y->parent;
  }
  x->parent = y;
  return(y);
}
INLINE Set *linkSetCC(Set *x, Set *y)
{
  if(x == y) {
    return x;
  }
  if(x->rank > y->rank) {
    y->parent = x;
    linkCCtoCC((CCstate*)x->data, (CCstate*)y->data);
    y->data = NULL;
    return x;
  } else {
    x->parent = y;
    linkCCtoCC((CCstate*)y->data, (CCstate*)x->data);
    x->data = NULL;

    if(x->rank == y->rank) {
      (y->rank)++;
    }
    return y;
  }
}

/*
* labelRuns()
* This funct looks at currRun and the row above it to determine
* the label of currRun.  It also compares and sets maxArea and
* maxLabel, and it updates the list sets so all the labels can
* be freed later on.
*
* This is a tricky function.  So look at it carefully.
* Remember we are looking for 8-connected color components.
* Two runs are the same color if the are solid and their  32-bit cmyk
* match, or if they a both sampled.
* If a run is sampled its color has no meaning.
*
* We are using disjoint sets to label the runs.  (basically trees
* with the pointers going in the other direction).
* To start we make the assumption that the currTopNode is pointing
* to the first run in the row above that is 8-connected to us.
* We compare this run to the currRun and if currRun has no label yet
* we give it the label of the run above.
* When we give it the label above, we must update the bounding box
* that is stored in this label.  If the run already has a label
* we unite the two labels.  unionSet takes care if the two labels are
* equal.
* We advance currTopNode to the next run and repeat, until the
* start of the next run is past the end of currRun, or there is
* no next run.  We keep track
* of the previous top run so currTopNode can be set up for the next
* call to labelRuns.  Because this is 8-connected we still have one
* more case to check for: if the begining of the top run is on the
* pixel immediately after the end of the currRun.  This must be
* a seperate case because in this case the previous top run should
* not be advanced.
*
* To finish up, if the label of the currRun is still not set then
* this is a new CCC, so a label is created.  Also the
* area of the label of the current run is compaired to the maxArea
* and maxLabel is set accordingly.
*/
INLINE void labelRun(ccc_cookie cookie, Run **currTopRun, Run *lastTopRun,
		     Run *currRun, int y)
{
  Run *topRun = *currTopRun;
  Set *label = currRun->label;
  CCstate *cc;
  int lastXplus1 = currRun->xPos + currRun->cccrun.length;
  Set *topSet;

  /*
   * Go through all the top runs that could be 8 connected
   * to the currRun
   *
   * exit the loop with currTopNode pointing to the 
   * last possible 8 connected run, save the previous
   * in case of a diagonal connection
   *
   * notice, The top node can only become null if 
   * the currRun is the last on the line, in which case
   * the value of currTopNode is not used by the caller
   */
  while (topRun != lastTopRun) {
    if(runsMatch(currRun, topRun)) {
      /* Link runs for our simple closed loop finding stuff */
      topSet = findSet(topRun->label);
      if(label == (Set *)NULL) {
	label = topSet;
	cc = (CCstate*)label->data;
	linkCCtoRun(cc, currRun, y);
      } else {
	if(label == topSet){
	  /* We should have just closed a loop
	     Check by looking at the topRun and the previous topRun color
	     the previous topRun has to be valid because there is no other way
	     for the label to be non-null
	     */
	  if(!runsMatch(topRun, (topRun -1))){
	    /* We have a closed loop */
	    ((CCstate*)(topSet->data))->numHoles++;
	  } else {
	    /* should combine the top runs */
	    label = label;
	  }
					

	} else {
	  /*  otherwise we just connected two branches 
	      link up the sets,
	      (and maybe for pass and a half add runlists together, and delete the 
	      extra cc from the list of cc's
	      */
	  label = linkSetCC(label, topSet);

	}
      }
    }

    if((topRun->xPos + topRun->cccrun.length) > lastXplus1){
      break;
    }

    topRun++;
  } 

  /* if the last overlapping run is is only touching by  */
  /* the corner, roll back the *currTopNode for the next call */
  if ((topRun != NULL) &&
      (topRun < lastTopRun) && 
      (lastXplus1 == topRun->xPos)) {
    topRun--;
  }

  /* If the label is still null the run didn't match anything */
  /* make a new set */
  if(label == (Set *)NULL) {
    label = makeSet(cookie);
    label->data = makeCC(cookie, currRun, y);
  }

  *currTopRun = topRun;
  currRun->label = label;
}



/*
* readAndLabel()
* This funct fills rows with one list of runs for each row
* The runs are labeled using disjoint sets so that the parent
* label of each run is the same iff the runs are in the same
* CCC.
*
* We read in each run, and have labelRun decide what to do with
* the label.
*
* labelRun also tells us the label of the CCC with the max area.
* We set the max area in the begining to be the length of the Dgp_Int32est
* run in the top row.  Area is defined as the number of pixels in the
* CCC.
*
* See readRun, labeRuns
*/
ErrCode readAndLabel(ccc_cookie cookie, RunCookie runCook, int numLines)
{
  Dgp_State *dgp_state = cookie->dgp_state;
  DGPERR_INIT(readAndLabel);
  int y=0;
  Run *pRuns, *topRun, *lastTopRun, *firstRun, currRun;
  int numRuns;
  int endLine = cookie->curLine + numLines;
  RunImage im = cookie->im;
  int width = cookie->width;

  if(cookie->curLine == 0){
    topRun = NULL;
    lastTopRun = NULL;
  } else {
    topRun = rimGetRow(im, (cookie->curLine - 1));
    lastTopRun = rimGetRow(im, cookie->curLine);
  }

  for(y = cookie->curLine; y<endLine; y++) {

    /* increase the size of our run pool if necessary. */
    pRuns = firstRun = rimStartRow(im, y);
    DGP_CHECK_ALLOC(firstRun);

    numRuns = 0;

    do{
      DGP_ERR_PASS_CLEANUP(readRunNoWash(cookie->rCook, &currRun, cookie, runCook));		
			
      labelRun(cookie, &topRun, lastTopRun, &currRun, y);
      *pRuns = currRun;
      pRuns++;
      numRuns++;
    } while((currRun.xPos + currRun.cccrun.length) < width);
				
    topRun = firstRun;
    lastTopRun = firstRun + numRuns;
    rimEndRow(im, y, numRuns);

  } 
	
  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}





/*
* See the comment in colorCC.h to see what this black box
* is supposed to do.
*
* Here is what it really does.
* A .qdr file consists of two Dgp_Int16 indicating the height and
* with of the image.  Then it contains lists of (header, data)
* pairs seperated by end of line markers.  See readRun which
* describes the encoding of these pairs and markers.
* Each pair is a run of pixels and a color.  The color can
* be either solid or sampled.  The solid color is a 32-bit cmyk color.
* The sampled color is a list of 32-bit cmyk colors for the length
* of the run.
*
* These runs are read from the file and given a label.
* all run's which are touching by 8-connectedness have the sample parent
* label.  The label also contains the the number of runs and the
* bounding box of all runs that are labeled with it.
* The runs of each row are stored in a list.
* If the run is sampled an array of 32-bit values contains its color.
* Also on the is first pass through the image a label is return that
* is pointed to the most number of pixels
*
* The creation of CCC's and the placement of runs in CCC's is done on
* row by row basis.
* First each run of a row is converted:
*  If the label of the run does not have a CCC associated with it
*   a new CCC is created, and added to the list of CCC's.
*   If the run is sampled then the CCC contains
*   an image to store the sampled colors.
*  The length, cc, and a flag of the run are put into an array for the
*   the row.
*  If the run is sampled, the sampled colors are put into the CCC image.
* Second the runs in the row's array are stored into the CCC's that they
*  point to.  If a run meets the filling criteria it is extended.
* The list of CCC's is returned.
*
* modification (stump, 9/21/96): the number of sampled images found will
* be returned in *num_sampled_images.
*
* modification (scytacki, 8/1/97): allow input of image array 
* by generalizing input to void * and stucture that contains
* functions for returning runs.
*/
ErrCode 
ccc_doStrip(colorCC_cookie _cookie, RunCookie runCookie, int numLines, 
	    DgpAdt_List ccDgpAdt_List, DgpAdt_List sampCClist)
{
  ccc_cookie cookie = (ccc_cookie) _cookie;
  Dgp_State *dgp_state = cookie->dgp_state;
  DgpAdt_ListNode retnode;
  DGPERR_INIT(ccc_doStrip);
	

  DgpAdt_List mergeDgpAdt_List, washCClist;
  Set *curSet, *topSet, *firstSet;
  DgpAdt_ListNode ln, washLn, ln2;
  /*int count;*/

  DGP_ERR_PASS (DgpAdt_liNew(&mergeDgpAdt_List, dgp_state));
  DGP_ERR_PASS (DgpAdt_liNew(&washCClist, dgp_state));

  cookie->defaultRun.cccrun.sampled = DGP_FALSE;
  cookie->defaultRun.wash = DGP_FALSE;
  cookie->defaultRun.cccrun.color = DEFAULT_BG_COLOR;

  /* Pass 1 */
  if(cookie->wThresh == 0){
    DGP_ERR_PASS_CLEANUP(readAndLabel(cookie, runCookie, numLines));
  } else {
    DGP_ERR_PASS_CLEANUP(readAndLabelWash(cookie, runCookie, numLines));
  }

  /*	
	count = 0;
	connectionCount = 0;
	ln = DGPADT_LIFIRST(cookie->ccSets);
	while(ln != DGP_NULLLISTNODE){
	/* Need to free CC's that aren't added */
	curSet = (Set *)DGPADT_LIGET(ln);
		

	if(curSet->data != NULL){
	count++;
	cc = (CC*)(curSet->data);
	connectionCount += DGPADT_LILEN(cc->washConnections);
	}
		
	ln = DGPADT_LINEXT(ln);
	}
	*/

  cookie->curLine += numLines;

  /* Last strip on the page */
  /* so finish up */
  if(cookie->curLine == cookie->height){		

    /* Pass 2 */
    findHardEdges(cookie);

    washLn = DGPADT_LIFIRST(cookie->washSets);
    while(washLn != DGP_NULLLISTNODE) {
      curSet = (Set *)DGPADT_LIGET(washLn);

      curSet = findSet(curSet);

      if(curSet->data == NULL){
	washLn = DGPADT_LINEXT(washLn);
	continue;
      }

      /*
       * This is an attempt to go through all the 
       * connected cc's and merge them into wash cc's
       * this could be done with recursion and it might
       * look nicer, but this ought to be faster.
       */
      ln = DGPADT_LIFIRST((DgpAdt_List)(curSet->data));
      DGP_ASSERT(ln != DGP_NULLLISTNODE);
      while (ln != DGP_NULLLISTNODE){
	topSet = findSet(DGPADT_LIGET(ln));
	ln2 = DGPADT_LIFIRST(((CCstate*)(topSet->data))->washConnections);
	if((ln2 != DGP_NULLLISTNODE) &&
	   (findSet(DGPADT_LIGET(ln2)) == topSet)){
	  /* we've already merged this top node */
	  DgpAdt_liRemAndNext(ln, &ln);
	  continue;
	}

	while(ln2 != DGP_NULLLISTNODE){
	  DgpAdt_liAdd(mergeDgpAdt_List, findSet(DGPADT_LIGET(ln2)), &retnode);
	  DgpAdt_liRemAndNext(ln2, &ln2);
	}

	ln2 = DGPADT_LIFIRST(mergeDgpAdt_List);
	while(ln2 != DGP_NULLLISTNODE){
	  ((CCstate*)(topSet->data))->wash = DGP_TRUE;
	  topSet = linkSetCC(findSet(DGPADT_LIGET(ln2)), topSet);
	  DgpAdt_liRemAndNext(ln2, &ln2);
	}

	DgpAdt_liRemAndNext(ln, &ln);
      }

			
      DgpAdt_liFree((DgpAdt_List)(curSet->data));
      curSet->data = NULL;

      washLn = DGPADT_LINEXT(washLn);
    }

    /* We have finished merging the 
     * sets so go through them all
     * compress them and convert the data pointers
     * we'll use the rank member of the sets
     * to tell if we've already converted a
     * data pointer or not
     */
    ln = DGPADT_LIFIRST(cookie->ccSets);
    while(ln != DGP_NULLLISTNODE){
      Set *topSet;

      curSet = firstSet = (Set *)DGPADT_LIGET(ln);

      for(;curSet < (firstSet + cookie->nBlockSize); curSet++){
	if(curSet->parent == NULL){
	  continue;
	}

	topSet = curSet;
	while(topSet != topSet->parent){
	  topSet = topSet->parent;
	}
	curSet->parent = topSet;
	if(topSet->rank != -1 &&
	   topSet->data != NULL){
	  /* convert the CC */
	  topSet->data = convertCC(cookie, (CCstate *)topSet->data);
	  topSet->rank = -1;
	} 
	curSet->data = topSet->data;

      }

      ln = DGPADT_LINEXT(ln);
    }


    /* Pass 3 */
    DGP_ERR_GEN_CLEANUP(!fillAndAdd(cookie, ccDgpAdt_List, sampCClist, washCClist),
			cErrInternal);

    ln = DGPADT_LIFIRST(cookie->ccSets);
    while(ln != DGP_NULLLISTNODE){
      curSet  = (Set *)DGPADT_LIGET(ln);
      cookie->DGP_FREE(curSet);
      DgpAdt_liRemAndNext(ln, &ln);
    }

    rDgpImage_free(cookie->im);
    cookie->im = NULL;
		
    DgpAdt_liFree(cookie->washSets);
  }


  DgpAdt_liFree(mergeDgpAdt_List);

  NORMAL_RETURN;
ERR_CLEANUP:
  ERR_RETURN;
	
}


