/*
   @(#)$Header: /project/icons/cvsroot/digipaper/src/libcomponent/colorCC.h,v 1.10 2001/05/08 00:42:54 dtracy Exp $ 
*/

#ifndef _COLORCC_H_
#define _COLORCC_H_

#include "image.h"
#include "list.h"
#include "nt_compat.h"
#include "util.h"
#include <dgp_allocators.h>
#include <dgp_state.h>
#include "cccrun_t.h"

#ifdef __cplusplus
extern "C"
{
#endif

typedef struct {
  union {
	  DgpImage_Long cmyk;
	  DgpImage_ColorRGB rgb;
  } image;
  Dgp_UInt16 realXoff, realYoff;
  Dgp_UInt16 realWidth, realHeight;
  Dgp_UInt16 *filledRunDgpAdt_List;
  int runDgpAdt_ListLen;
  int runDgpAdt_ListSize;
  Dgp_Bool wash;
  Dgp_Bool above;
  Dgp_Bool below;
  Dgp_Bool split;
} SampState;

typedef struct {
  Dgp_Bool holes;
  Dgp_Bool save;
  Dgp_Bool large;
  Dgp_Bool added;
  Dgp_Bool del;
  Dgp_UInt16 ulX, ulY, lrX, lrY;
  union {
	  Dgp_UInt32 cmyk;
	  DgpImage_RGB rgb;
  } color;
  Dgp_UInt16 lastY, lastX;
  Dgp_UInt16 *runDgpAdt_List;
  int numRuns;
  int runDgpAdt_ListLen;
  SampState *sampling;
} CC;

/*
 * Types and a structure to allow any type of input
 * source.  
 */
typedef void *RunCookie;
typedef ErrCode (TC_CDECL *PFNReadRun)(cccRun *, Dgp_Bool *eol, RunCookie, Dgp_State *);


typedef void *colorCC_cookie;

/*
 * Given an input reader structure findCC() makes a list of CCC's.
 * The maxCC is not included in the list of CCC's.
 * Currently the maxCC is the CCC(Color Connected Component),
 * with the most pixels.  Its ulX,ulY,lrX,lrY are correct.
 * It's runDgpAdt_List is NULL.
 * We define a CCC as a 8-connected component.
 * So if we are looking at pixel "O" all          X X X
 * the "X"'s are canidates for being in the       X O X
 * same component.                                X X X
 *
 * The CCC's returned by this function are also filled in.
 * The intension of filling is to cut down on duplicate
 * shape information.  This duplication occurs whenever
 * two CCC's touch.  We don't want to remove all the duplication
 * because we use the shape information to match one CCC with
 * another.  So first the criteria then the outcome.
 * Criteria- we want to elimate duplicate shape info when:
 *  o The bounding box of one CCC is completely inside of another.
 *   and
 *  o The outside CCC is present both to the left and the right
 *     of a run of pixels in the inside CCC.
 *  
 * Outcome - when a run of pixels of the inner CCC meets these
 *  criteria, the run of pixels is added to the outer CCC,
 *  the inner CCC remains the same.
 * We call this a fill.
 *
 * No layering information is necessary to reconstruct the image
 *  from these CCC's.  This is because of the bounding box
 *  stipulation.  The reconstructer only needs to put down
 *  CCC's that contain other CCC's before the contained CCC's.
 *  This can be acomplished by sorting the CCC's by a bounding
 *  box corner and then drawing the CCC's starting from that
 *  same corner of the page.
 * 
 *  modification (stump, 9/21/96): the number of sampled images findCC finds
 *  will be returned as *num_sampled_images.
 */
#define ccc_init_err(args) \
	DGP_ERR_PASS_CLEANUP(ccc_init##args)

extern ErrCode TC_CDECL
ccc_init(Dgp_State *dgp_state, colorCC_cookie *_cookie, int width, int height, int washThreshold,
					int minSampSize, PFNReadRun readRun);
extern void TC_CDECL
ccc_freeCookie(colorCC_cookie cookie);

#define ccc_doStrip_cmyk_err(args) \
	DGP_ERR_PASS_CLEANUP(ccc_doStrip_cmyk##args)

extern ErrCode TC_CDECL
ccc_doStrip_cmyk(colorCC_cookie cookie, RunCookie runCookie, int numLines, 
				 DgpAdt_List ccDgpAdt_List, DgpAdt_List sampCClist);

#define ccc_doStrip_rgb_err(args) \
	DGP_ERR_PASS_CLEANUP(ccc_doStrip_rgb##args)

extern ErrCode TC_CDECL
ccc_doStrip_rgb(colorCC_cookie cookie, RunCookie runCookie, int numLines, 
				DgpAdt_List ccDgpAdt_List, DgpAdt_List sampCClist);


#ifdef __cplusplus
}
#endif

#endif

