#include "colorCC.h"
#include "colorCC-priv.h"
#include <util.h>
#include "image_defs.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcomponent/colorCCfill.c,v 1.13 2003/02/10 22:53:23 mkrocken Exp $");

#ifdef RGB_FORMAT
INLINE void avgColor(RGB *retCol, RGB color1, RGB color2)
{
	retCol->r = ((int)color1.r + (int)color2.r)/2;
	retCol->g = ((int)color1.g + (int)color2.g)/2;
	retCol->b = ((int)color1.b + (int)color2.b)/2;
}


#else /*LONG_DGPIMAGE_FORMAT */

INLINE void avgColor(Color *retCol, Color color1, Color color2)
{
	*retCol = 0;
	
	*retCol |= ((cmyk_c(color1) + cmyk_c(color2)) /2) << 24;
	*retCol |= ((cmyk_m(color1) + cmyk_m(color2)) /2) << 16;
	*retCol |= ((cmyk_y(color1) + cmyk_y(color2)) /2) << 8;
	*retCol |= ((cmyk_k(color1) + cmyk_k(color2)) /2);
	
}


#endif



/*
* addCCToDgpAdt_List
* 
* modification (stump, 9/21/96):  *num_sampled_images will contain the
* number of sampled images found up to this point; this function will
* update *num_sampled_images if it finds a sampled image.
* 
*/
INLINE int addCCToDgpAdt_List(ccc_cookie cookie, CC *cc, DgpAdt_List ccDgpAdt_List, 
					   DgpAdt_List sampCCDgpAdt_List, DgpAdt_List washCCDgpAdt_List)
{
	DgpAdt_ListNode retnode;
	if(cc->del){
		/* Opps, we thought we wouldn't be adding this CC. */
		/* so some runs of the CC have been eliminated, so its bounding  */
		/* box is no Dgp_Int32er the same  Not a big deal if */
		/* we don't do it very often. */
	}
	
  DgpAdt_liAdd(ccDgpAdt_List, (void *)cc, &retnode); 
  if(retnode == DGP_NULLLISTNODE) {
    return 0;
  }
	
	cc->runDgpAdt_List =
		(Dgp_UInt16 *)cookie->DGP_MALLOC(sizeof(Dgp_UInt16) * 2 * cc->runDgpAdt_ListLen);
	if(cc->runDgpAdt_List == NULL){
		return 0;
	}

	cc->numRuns = 0;
	if(cc->sampling != NULL){
		DgpAdt_liAdd(sampCCDgpAdt_List, (void *)cc, &retnode); 
		if(retnode == DGP_NULLLISTNODE) {
      return 0;
    }
		 DgpImage_new(cookie->dgp_state, SAMP_IMG_TYPE,
			cc->lrX - cc->ulX + 1,
			cc->lrY - cc->ulY + 1, (void**) &cc->sampling->image.CSPACE);
		if(cc->sampling->image.CSPACE == NULL){
			if(cc->runDgpAdt_List != NULL){
				cookie->DGP_FREE(cc->runDgpAdt_List);
			}
			return 0;
		}
	} 
	
	cc->added = DGP_TRUE;
	
	return(1);
}

INLINE void addSampling(ccc_cookie cookie, CC *cc, Run *currRun, int y)
{
	DGP_VOID_PROLOGUE("addSampling()", NULL);
	Dgp_UInt32 color;
	Color *sampPtr, *endPtr, *sampPtr2;
	
	if(currRun->cccrun.sampling_ == NULL){
		DGP_VOID_ASSERT(cc->sampling != NULL && 
			cc->sampling->wash && !currRun->wash);
		color = currRun->cccrun.color;
		
		sampPtr = &(DGPIMG_REF(cc->sampling->image.CSPACE, currRun->xPos - cc->ulX, (y - cc->ulY)));
		endPtr = sampPtr + currRun->cccrun.length;
		/* This will be slow for RGB (probably should special case this) */
		for(; sampPtr < endPtr; sampPtr++){ 
			copyToSampling(*sampPtr, color);
		}
	} else {		
		sampPtr = &(DGPIMG_REF(cc->sampling->image.CSPACE, currRun->xPos - cc->ulX, (y - cc->ulY)));
		sampPtr2 = &(currRun->cccrun.sampling_[0]);
		endPtr = sampPtr2 + currRun->cccrun.length;
		
		for(; sampPtr2 < endPtr; sampPtr2++,sampPtr++){
			*sampPtr = *sampPtr2;
		}
		
	}
}

#define deleteSampling(cookie, currRun) \
	if((currRun)->cccrun.sampling_ != (Color *)NULL){ \
	(cookie)->DGP_FREE((currRun)->cccrun.sampling_); \
	(currRun)->cccrun.sampling_ = NULL; \
	}


#define addRunElement(cc, xPos, length) \
	if((cc)->runDgpAdt_ListLen <= (cc)->numRuns){ \
	(cc)->runDgpAdt_ListLen = 3 * (cc)->runDgpAdt_ListLen / 2; \
	(cc)->runDgpAdt_List = \
	(Dgp_UInt16 *)cookie->DGP_REALLOC((cc)->runDgpAdt_List, sizeof(Dgp_UInt16) * 2 * cc->runDgpAdt_ListLen); \
	}\
	(cc)->runDgpAdt_List[(cc)->numRuns * 2] = (xPos); \
	(cc)->runDgpAdt_List[(cc)->numRuns * 2 + 1] = (length); \
(cc)->numRuns++


INLINE void addRun(ccc_cookie cookie, CC *cc, int xPos, int yPos, int length) 
{
	DGP_VOID_PROLOGUE("addRun()", NULL);
	
	if((yPos) > ((cc)->lastY + 1)){ 
		addRunElement(cc, yPos, 0); 
	} else if (yPos == (cc)->lastY) {
		DGP_VOID_ASSERT(!((cc->lastX != (Dgp_UInt16)-1) &&
			(xPos <= cc->lastX)));
	} else if (yPos == (cc->lastY + 1)) {
		DGP_VOID_ASSERT(xPos <= cc->lastX);
	} else {
		DGP_VOID_ASSERT(0);
	}
	addRunElement(cc, xPos, length); 
	(cc)->lastY = (yPos); 
	(cc)->lastX = (xPos) + length;
}

void
splitAndAdd(ccc_cookie cookie, Run *currRun, int y, DgpAdt_List ccDgpAdt_List, 
			DgpAdt_List sampCCDgpAdt_List, DgpAdt_List washCCDgpAdt_List)
{
	DGP_VOID_PROLOGUE("splitAndAdd()", NULL);
	Run currSubRun;
	CC *subCC, *cc;
	Color currColor;				
	int currSubRunX = 0;
	int x;
	
	cc = (CC*)(currRun->label->data);
	
	if(currRun->cccrun.sampling_ == NULL){
		/* This is solid run in a wash cc */
		DGP_VOID_ASSERT(!currRun->wash && (cc->sampling != NULL));
		
		currRun->cccrun.sampled = DGP_FALSE;
		subCC = makeCCnew(cookie, currRun, y);
		
		/* Add this cc to the list */
		if(!addCCToDgpAdt_List(cookie, subCC, ccDgpAdt_List, sampCCDgpAdt_List, washCCDgpAdt_List)){
			goto bailout;
		}
		
		/* then add the run to this cc */
		addRun(cookie, subCC, 0, 0, currRun->cccrun.length);
		
	} else {
		currSubRun.cccrun.sampled = DGP_FALSE;
		currSubRun.wash = DGP_FALSE;
		
		currColor = currRun->cccrun.sampling_[0];
		x = 1;
		while(1){
			if((x == currRun->cccrun.length) ||
        (!colorsMatch(currColor, currRun->cccrun.sampling_[x]))) {
        /*(currColor != currRun->cccrun.sampling_[x])){ */
				
				/* Setup new Run to init cc */
				currSubRun.cccrun.length = x - currSubRunX;
				currSubRun.xPos = currRun->xPos + currSubRunX;
				
				/* so we need to create a new cc for each new run in this */
				/* sampled one */
				subCC = makeCCnew(cookie, &currSubRun, y);
				subCC->color.CSPACE = currColor;
				
				/* Add this cc to the list */
				if(!addCCToDgpAdt_List(cookie, subCC, ccDgpAdt_List, sampCCDgpAdt_List, washCCDgpAdt_List)){
					goto bailout;
				}
				
				/* then add the run to this cc */
				addRun(cookie, subCC, 0, 0, currSubRun.cccrun.length);
				
				if(x == currRun->cccrun.length){
					break;
				} else {
					currSubRunX = x;
					currColor = currRun->cccrun.sampling_[x];
				}
			}
			
			x++;
		}
		cookie->DGP_FREE(currRun->cccrun.sampling_);
		
	}
	
bailout:
	return;
}

int fillAndAdd(ccc_cookie cookie, DgpAdt_List ccDgpAdt_List, DgpAdt_List sampCCDgpAdt_List,
			   DgpAdt_List washCCDgpAdt_List)
{
	DGP_VOID_PROLOGUE("fillAndAdd()", NULL);
	int y, xOffset;
	CC *cc, *rAcc, *lRcc;
	Run *currRun, *runsAfter;
	Dgp_UInt16 ulX, ulY, lrX, lrY;
	Dgp_UInt16 length;
	int height = cookie->height;
	Run *firstRun, *lastRun;
	RunImage im = cookie->im;
	
	
	for(y=0;y<height;y++) {
		firstRun = rimGetRow(im, y);
		lastRun = rimGetRow(im, y+1);
		
		for (currRun = firstRun; currRun != lastRun; currRun++) {
			
			cc = (CC*)currRun->label->data;
			
			if(currRun->matchesBelow) { 
				
				if(currRun->cccrun.sampling_ != NULL){
					deleteSampling(cookie, currRun);
				}
				cc->runDgpAdt_ListLen--;
				if(cc->runDgpAdt_ListLen == 0){
					cookie->DGP_FREE(cc);
					currRun->label->data = NULL;
				}
				continue;
			}
			
			/*
			* If this cc is sampled and its size is too 
			* small, we need to split each sampled run into
			* mutliple ccs.  (fun!)
			*/
			if((cc->sampling != NULL)  &&
				cc->sampling->split){
				splitAndAdd(cookie, currRun, y, ccDgpAdt_List, sampCCDgpAdt_List, washCCDgpAdt_List);
				
				/* don't execute the rest of the loop just move to the next run */
				continue;
			}
			
			if(!cc->added){	
				if(!addCCToDgpAdt_List(cookie, cc, ccDgpAdt_List, sampCCDgpAdt_List, washCCDgpAdt_List)){
					goto bailout;
				}
				
				/* Need to set the color here */
#ifdef RGB_FORMAT
				cc->color.rgb.r = currRun->cccrun.color >> 16 && 0xFF;
				cc->color.rgb.g = currRun->cccrun.color >> 8 && 0xFF;
				cc->color.rgb.b = currRun->cccrun.color && 0xFF;
#else 
				cc->color.cmyk = currRun->cccrun.color;
#endif
				
				
			}
			
			ulX = cc->ulX;
			ulY = cc->ulY;
			lrX = cc->lrX;
			lrY = cc->lrY;
			
			length = currRun->cccrun.length;
			if(cc->sampling != NULL){
				addSampling(cookie, cc, currRun, y);
				deleteSampling(cookie, currRun);
			}
			
			xOffset = currRun->xPos;
			
			if((cc->sampling != NULL) && cc->large &&
				!cc->sampling->above){
				
				if(cc->lastY != y && 
					ulX != xOffset){
					/* we should scan left to make this rectangular */
				}
				runsAfter = currRun + 1;
				
				while((runsAfter != lastRun) &&
					(lrX >= (xOffset + length))){
					rAcc = (CC*)(runsAfter->label->data);
					
					if(cc == rAcc) {
						length += runsAfter->cccrun.length;
						addSampling(cookie, cc, runsAfter, y);
						
						runsAfter->matchesBelow = DGP_TRUE;
						runsAfter->lowerRun = currRun;
					} else {
						cc->sampling->below = DGP_TRUE;
						
						
						if(runsAfter->matchesBelow){
						/* we will want to do something special here
						* I don't know what yet
							*/
						}
						
						/* 
						* we want to fill this CC so that is rectangular if we 
						* can.  
						*
						* Also we want to save the shape of the stuff we
						* fill/ or don't fill
						* so that we can make the image smooth under the 
						* mask or cc's that cover it
						*/
						length += runsAfter->cccrun.length;
						
						if((length + xOffset) > (lrX + 1)){
							length = lrX - xOffset + 1;
							/*
							* We are creating a interesting case here
							* we just filled under a run, but we didn't completly
							* fill under the run.  So we need to say the run
							* doesn't match below
							*/
							runsAfter->matchesBelow = DGP_FALSE;
							
						} else {
							
						/*
						* Since we just changed this run's lowerRun see if it the lowerRun
						* matches the run
							*/
							
							/*
							* INSERT matching test here.  Use length and xOffset not currRun
							* values.
							*/
							runsAfter->matchesBelow = runsMatchWash(currRun, runsAfter);
							
							/*  A little hack until runsMatchWash is correct */
							if((rAcc->sampling != NULL) && 
								rAcc->sampling->wash && cc->sampling->wash){
								runsAfter->matchesBelow = DGP_TRUE;
							}
							
							runsAfter->lowerRun = currRun;
							
						}
						
						/*
						* Now we should set the filled run list of the sampling
						* to this filled in area.  Lets do that in a little bit
						*/
						if(rAcc->sampling != NULL){
							rAcc->sampling->above = DGP_TRUE;
						}
						
					}
					
					runsAfter++;
				}
			} else if (cc->holes && 
				((cc->sampling == NULL) || !cc->sampling->above)){
				/* 
				* This cc has holes in it, so we have two choices.
				* 1. We can save the cc's in holes and fill this cc in underneath. 
				* 2. If this cc is also inside a hole that has been filled in,  
				*    (Ex. The background is a big cc and the letters are holes in them)
				*    then the holes inside this cc can be thrown away.  IF they
				*    are the same color as the cc underneath this one.
				*/
				runsAfter = currRun + 1;
				/* 
				*  Starting at the run after the current one 
				*  look at all runs that could be inside of this cc
				*  So runs before the left bound of the cc and before the edge 
				*  of the image
				*/
				while((runsAfter != lastRun) &&
					(lrX >=(xOffset + length))){
					
					rAcc = (CC*)(runsAfter->label->data);  /*cc of the run after */
					
					if(cc == rAcc) {
						
					/* 
					* This cc matches the next run cc
					* increase the length of the run we are saving
					* and add the sampling to the cc
						*/
						length += runsAfter->cccrun.length;
						/* Put the sampling in the cc */
						if(cc->sampling != NULL){
							addSampling(cookie, cc, runsAfter, y);
						}
						
						/* mark this run so we don't add it twice */
						runsAfter->matchesBelow = DGP_TRUE;
						runsAfter->lowerRun = currRun;
						
					} else if (ulX < rAcc->ulX &&
						ulY < rAcc->ulY &&
						lrX > rAcc->lrX &&
						lrY > rAcc->lrY) {
						/* This is a rAcc that is "inside" cc */
						
						if(((rAcc->sampling != NULL) || cc->save) &&
							runsAfter->matchesBelow)
						{
							/* Add the sampling from this run to the lower one */
							if(rAcc->sampling != NULL){
								lRcc = (CC*)(runsAfter->lowerRun->label->data);
								addSampling(cookie, lRcc, runsAfter, y);
								/* The sampling will get deleted when when  */
								/* the runAfter becomes the current run */
								
								/* We need to recheck whether it really matches */
								/* the matchesBelow bit was set based on whether it was */
								/* a wash or not.  Now we have more info about this hole */
								/* and wash, so use this to check if this wash really */
								/* matches */
								
								/* We need to update the runDgpAdt_List for the holes in */
								/* the lower CC */
							}
							
							/*
							* The run after cc is a wash or sampled cc (which are
							* expensive to store so we try to eliminate as many as possible).
							* 
							* We want to save the shape of the current
							* cc.  and...
							*
							* This runAfter is the same color as the run below it
							*/
							
							/* mark the cc so we know we deleted a run */
							rAcc->del = 1;
							
							/* 
							* We have been building up a run as we filled underneath cc's 
							* this run will have a break in it because we want the 
							* run under this one to be seen
							*
							* So save this run
							*/
							if(length != 0){
								addRun(cookie, cc, xOffset - ulX, y - ulY, length);
							}
							
							/* Start a new run  */
							xOffset = xOffset + length + runsAfter->cccrun.length;
							length = 0;
							
						} else if (rAcc->sampling == NULL || 
							!rAcc->sampling->below){
							/*
							* The runAfter cc is inside of this cc 
							* but we are not discarding the run After cc
							* We can fill in this cc because the run After cc 
							* will be blatted on top
							*/
							runsAfter->lowerRun = currRun;
							
							length += runsAfter->cccrun.length;
							
							/*
							* Since we just changed this run's lowerRun see if it the lowerRun
							* matches the run
							*/
							/*
							* INSERT matching test here.  Use length and xOffset not currRun
							* values.
							*/
							runsAfter->matchesBelow = runsMatchWash(currRun, runsAfter);
							
							if(rAcc->sampling != NULL){
								rAcc->sampling->above = DGP_TRUE;
								
								/*  A little hack until runsMatchWash is correct */
								if(rAcc->sampling->wash && 
									(cc->sampling != NULL) && cc->sampling->wash){
									runsAfter->matchesBelow = DGP_TRUE;
								}
							}
							
							/* if the run is a wash or sampled image set the 
							* filled in bits to an appropriate color, or the run
							* above if it matchesBelow and set the fill distance
							*/
							if(cc->sampling != NULL){
								int x;
								int end;
								/* fill image/wash */
								end = runsAfter->xPos + runsAfter->cccrun.length;
								
								cc->sampling->below = DGP_TRUE;
								/* Need to update runlist for holes of the cc */
								/* Soon we shouldn't do the filling here we will */
								/* do it latter */
								
								
								if(cc->ulY == y){
									for(x = runsAfter->xPos; x < end; x++){ 
										DGPIMG_REF(cc->sampling->image.CSPACE , x - cc->ulX, (y - cc->ulY)) =
											DGPIMG_REF(cc->sampling->image.CSPACE, runsAfter->xPos - cc->ulX - 1, (y - cc->ulY));
									}
								} else {
									Color tempCol;
									
									for(x = runsAfter->xPos; x < end; x++){ 
										avgColor(&tempCol, DGPIMG_REF(cc->sampling->image.CSPACE, x - cc->ulX, (y - cc->ulY) - 1),
											DGPIMG_REF(cc->sampling->image.CSPACE, x - cc->ulX-1, (y - cc->ulY)));
										DGPIMG_REF(cc->sampling->image.CSPACE , x - cc->ulX, (y - cc->ulY)) = tempCol;												
									}
								}
								
							}
							
						}
	  } else {
		  break;
	  }
	  
	  runsAfter++;
	}
      } 
	  
      else if((cc->sampling != NULL) && 
		  cc->sampling->wash){
		  /* check for consequtive runs */
		  runsAfter = currRun + 1;
		  while((runsAfter != lastRun) &&
			  (lrX >=(xOffset + length))){
			  rAcc = (CC*)(runsAfter->label->data);  /*run after cc */
			  
			  /*	  DGP_VOID_ASSERT(cc != rAcc); */
			  /*	  break; */
			  
			  if(cc == rAcc) {
				  
			  /* 
			  * This cc matches the next run cc
			  * increase the length of the run we are saving
			  * and add the sampling to the cc
				  */
				  length += runsAfter->cccrun.length;
				  /* Put the sampling in the cc */
				  addSampling(cookie, cc, runsAfter, y);
				  deleteSampling(cookie, runsAfter);
				  
				  
				  /* mark this run so we don't add it twice */
				  runsAfter->matchesBelow = DGP_TRUE;
				  runsAfter->lowerRun = currRun;
			  } else {
				  break;
			  }
			  
			  
			  runsAfter++;
			  
		  }
      }
	  
      if(length != 0){
		  addRun(cookie, cc, xOffset - ulX, y - ulY, length);
      }	
    }
  }	
  
  return DGP_TRUE;
bailout:
  return DGP_FALSE;
}

