#include "colorCC.h"
#include <util.h>
#include "colorCC-priv.h"
#include "dgp_utils.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcomponent/colorCCread.c,v 1.14 2003/02/10 22:53:23 mkrocken Exp $");


#ifdef RGB_FORMAT

INLINE int colorDist(Dgp_UInt32 color1, Dgp_UInt32 color2)
{
  int tempColor;
  int total;

  tempColor = frgb_r(color1) - frgb_r(color2);
  total = tempColor * tempColor;

  tempColor = frgb_g(color1) - frgb_g(color2);
  total += tempColor * tempColor;

  tempColor = frgb_b(color1) - frgb_b(color2);
  total += tempColor * tempColor;

  return total;
}


#else
INLINE int colorDist(Color color1, Color color2)
{
  int tempColor;
  int total;

  tempColor = cmyk_c(color1) - cmyk_c(color2);
  total = tempColor * tempColor;

  tempColor = cmyk_m(color1) - cmyk_m(color2);
  total += tempColor * tempColor;

  tempColor = cmyk_y(color1) - cmyk_y(color2);
  total += tempColor * tempColor;

  tempColor = cmyk_k(color1) - cmyk_k(color2);
  total += tempColor * tempColor;

  return total;
}
#endif

/* Need to define "i" to make this work */
#define convertToWash(run, _matchesBelow, _length) \
(run).wash = DGP_TRUE; \
(run).matchesBelow = _matchesBelow; \
if(!(_matchesBelow)) \
(run).lowerRun = &(cookie->defaultRun); \
(run).cccrun.sampling_ = cookie->DGP_MALLOC(sizeof(Color) * (_length)); \
for(i=0; i<(run).cccrun.length; i++){ \
				 copyToSampling((run).cccrun.sampling_[i], (run).cccrun.color); \
										    }


read_cookie initReadRuns(ccc_cookie cookie){
  read_cookie retCookie;
	
  retCookie = (read_cookie)cookie->DGP_MALLOC(sizeof(*retCookie));
  if(retCookie == NULL){
    return NULL;
  }

  retCookie->currLine = 0;
  retCookie->currX = 0;

  return retCookie;
}

ErrCode readRunNoWash(read_cookie rCook, Run *currRun, ccc_cookie cookie, RunCookie runCook)
{
  Dgp_State *dgp_state = cookie->dgp_state;
  DGPERR_INIT(readRunNoWash);
  Run *nextRun;
  int x,y;
  Dgp_Bool eol;
  cccRun retRun;

  x = rCook->currX;
  y = rCook->currLine;
  nextRun = &(rCook->nextRun);

  if(x == 0){
    /* This is the first run of the row */
    DGP_ERR_PASS((cookie->readRun)(&retRun, &eol, runCook, cookie->dgp_state));
    currRun->cccrun = retRun;
    initRun(*currRun, x);
    x += currRun->cccrun.length;
  } else {
    *currRun = *nextRun;
  }

  currRun->lowerRun = &(cookie->defaultRun);
  currRun->matchesBelow = runsMatch(&(cookie->defaultRun), currRun);
  eol = DGP_FALSE;

  while(!eol){		
    /* Read and paritally initialize the next run */
    DGP_ERR_PASS((cookie->readRun)(&retRun, &eol, runCook, cookie->dgp_state));
    nextRun->cccrun = retRun;
    initRun(*nextRun, x);

    if(eol){
      /* we are reading the last run in the row */
      rCook->currLine++;
      rCook->currX = 0;
      NORMAL_RETURN;
    }

    x += nextRun->cccrun.length;
    if(!runsMatch(currRun, nextRun)){
      /*
       * it is a run that doesn't match the next run
       * Thus we save the next run and return
       */
      rCook->currX = x;
      NORMAL_RETURN;
    }
				
				
    /*
     * The nextRun did match the currRun.
     * so we skip the next run and just add the 
     * two together
     */
    currRun->cccrun.length += nextRun->cccrun.length;
		
    /* need to add the sampling as well */
    if(currRun->cccrun.sampled){
      currRun->cccrun.sampling_ = cookie->DGP_REALLOC(currRun->cccrun.sampling_, 
					     sizeof(Color) * currRun->cccrun.length);
    }
				
  }

  if(eol){
    /* we read the last run in the row */
    rCook->currLine++;
    rCook->currX = 0;
    NORMAL_RETURN;
  } 

  ERR_RETURN;
	
}

ErrCode readRunWash(read_cookie rCook, Run *currRun, ccc_cookie cookie, RunCookie runCook)
{	
  Dgp_State *dgp_state = cookie->dgp_state;
  DGPERR_INIT(readRunWash);
  int colorDistance;
  Run *nextRun;
  int x,y;
  int newLength;
  int i;
  Dgp_Bool eol;
  cccRun retRun;

  x = rCook->currX;
  y = rCook->currLine;
  nextRun = &(rCook->nextRun);

  if(x == 0){
    /* This is the first run of the row */
    DGP_ERR_PASS((cookie->readRun)(&retRun, &eol, runCook, cookie->dgp_state));
    currRun->cccrun = retRun;
    initRun(*currRun, x);
    x += currRun->cccrun.length;
  } else {
    *currRun = *nextRun;
  }

  currRun->lowerRun = &(cookie->defaultRun);
  currRun->matchesBelow = runsMatchWash(&(cookie->defaultRun), currRun);
  eol = DGP_FALSE;

  while(!eol){
    /* Read and paritally initialize the next run */
    DGP_ERR_PASS((cookie->readRun)(&retRun, &eol, runCook, cookie->dgp_state));
    nextRun->cccrun = retRun;
    initRun(*nextRun, x);
    x += nextRun->cccrun.length;
		
    if(eol) {
      /* this is the last run in a row */
      /* we are reading the last run in the row */
      rCook->currLine++;
      rCook->currX = 0;
      NORMAL_RETURN;
    }

    /*
     * Get the run's color separation
     */
    if(currRun->cccrun.sampled == nextRun->cccrun.sampled){
      /*
       * There is a little trickiness here:
       * if the runs are both sampled their colors will be the same default value
       *
       * If the currRun has been marked wash, then its color
       * will be the color of the last run added to the wash
       * so the function will be comparing the nextRun.color to the color
       * of the last pixel in the wash run
       */
      colorDistance = colorDist(currRun->cccrun.color, nextRun->cccrun.color);
    } else {
      colorDistance = cookie->wThresh + 1;
    }
		
    if(colorDistance > cookie->wThresh){
      /*
       * the runs don't match at all  
       */
      rCook->currX = x;
      NORMAL_RETURN;

    } else if(colorDistance != 0) {
			
      /*
       * The distance is below the threshold so the runs could combined
       * into a wash
       */
      newLength = currRun->cccrun.length + nextRun->cccrun.length;
			
      /* make a new wash run if necessary */
      if(!currRun->wash){
	      convertToWash(*currRun, DGP_FALSE, newLength);
				
      } else {
	      currRun->cccrun.sampling_ = cookie->DGP_REALLOC(currRun->cccrun.sampling_, 
					       sizeof(Color) * newLength);
      }
			
      /* fill in the new wash bits */
      for(i = currRun->cccrun.length; i < newLength; i++){
	      copyToSampling(currRun->cccrun.sampling_[i], nextRun->cccrun.color); 
      }
			
      /* set the run color to the color of the left most pixel */
      currRun->cccrun.color = nextRun->cccrun.color;
			
      currRun->cccrun.length = newLength;
			
    } else {
      /*
       * The nextRun did match the currRun.
       * so we skip the next run and just add the 
       * two together
       */
      currRun->cccrun.length += nextRun->cccrun.length;
			
      /* need to add the sampling as well */
      if(currRun->cccrun.sampled){
	      currRun->cccrun.sampling_ = cookie->DGP_REALLOC(currRun->cccrun.sampling_, 
					       sizeof(Color) * currRun->cccrun.length);
      }
    }


  }
					
  if(eol){
    /* we read the last run in the row */
    rCook->currLine++;
    rCook->currX = 0;
    NORMAL_RETURN;
  } 

  ERR_RETURN;
			
}
