#include "colorCC.h"
#include "colorCC-priv.h"
#include <util.h>
#include "dgp_utils.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcomponent/colorCCwash.c,v 1.15 2003/02/10 22:53:23 mkrocken Exp $");


#ifdef	RGB_FORMAT
INLINE int colorDistNat(RGB color1, RGB color2)
{
	int tempColor;
	int total;
	
	tempColor = color1.r - color2.r;
	total = tempColor * tempColor;
	
	tempColor = color1.g - color2.g;
	total += tempColor * tempColor;
	
	tempColor = color1.b - color2.b;
	total += tempColor * tempColor;
	
	return total;
}
#endif


INLINE int runsMatchWash(Run *run1, Run *run2)
{
	if(run1->wash && run2->wash){
		return DGP_TRUE;
	} else {
		return runsMatch(run1, run2);
	}
}

INLINE int runsMatchWashDiagonal(ccc_cookie cookie, Run *run1, Run *run2)
{
	int i;
	Color *run1color, *run2color, color;
	int length, offset;
	int run1length, run2length;
	Run *runs[2];
	
	if(run1->cccrun.sampled || run2->cccrun.sampled){
		return DGP_FALSE;
	}
	
	if(!run1->wash && !run2->wash) {
		return (colorDist(run1->cccrun.color, run2->cccrun.color) <= cookie->wThresh);
		
	} 
	
	/* compute offsets and length of overlap
	* run1* is the run that starts before the other run
	* offset is how much before
	* and runs[0] goes with run1*
	* and runs[1] goes with run2*
	*/
	if(run1->xPos > run2->xPos){
		offset = run1->xPos - run2->xPos;
		run1color = run2->cccrun.sampling_ + offset;
		run1length = run2->cccrun.length - offset;
		runs[0] = run2;
		runs[1] = run1;
		run2color = run1->cccrun.sampling_;
		run2length = run1->cccrun.length;
	} else {
		offset = run2->xPos - run1->xPos;
		run1color = run1->cccrun.sampling_ + offset;
		run1length = run1->cccrun.length - offset;
		runs[0] = run1;
		runs[1] = run2;
		run2color = run2->cccrun.sampling_;
		run2length = run2->cccrun.length;
	}
	
	if(run1length > run2length){
		length = run2length;
	} else {
		length = run1length;
	}
	
	if(!offset && !length){
		return DGP_FALSE;
	}
	
	if(run1->wash && run2->wash){
		if(offset && (runs[0] != run2)){
			/* check the diagonal */
			if(colorDistNat(*(run1color-1), *run2color) > cookie->wThresh){
				return DGP_FALSE;
			}
		} 
		
		for(i=0; i<length; i++){
			if(colorDistNat(*run1color, *run2color) > cookie->wThresh){
				return DGP_FALSE;
			}
			
			run1color++;
			run2color++;
		}
		
		if(run1length > length){
			/* check the diagonal */
			if(colorDistNat(*run1color, *(run2color-1)) > cookie->wThresh){
				return DGP_FALSE;
			}
		} else if(run2length > length){
			/* check the diagonal */
			if(colorDistNat(*(run1color-1), *run2color) > cookie->wThresh){
				return DGP_FALSE;
			}
		} else if(length == 0){
			return DGP_FALSE;
		}
		
		return DGP_TRUE;
	}
	
	if(runs[0]->wash){
		copyToSampling(color, runs[1]->cccrun.color);
		if(offset && (runs[0] != run2)){
			run1color--;
			length++;
		}
		
		if(run1length > length){
			length++;
		}
		
	} else {
		run1color = run2color;
		copyToSampling(color, runs[0]->cccrun.color);
		
		if(run2length > length){
			length++;
		}
	}
	
	if(length == 0){
		return DGP_FALSE;
	}
	
	for(i=0; i<length; i++){
		if(colorDistNat(*run1color, color) > cookie->wThresh){
			return DGP_FALSE;
		}
		run1color++;
	}
	
	/* if we made it here.  Then we must have a wash and a non-wash that matchup */
	return DGP_TRUE;
}


INLINE int runsMatchWashNoDiagonal(ccc_cookie cookie, Run *run1, Run *run2)
{
	DGP_VOID_PROLOGUE("runsMatchWashNoDiagonal()", NULL);
	Color *firstColor, *secondColor, *lastColor, color;
	int length, offset;
	int firstLength, secondLength;
	Run *first, *second;
	
	if(run1->cccrun.sampled || run2->cccrun.sampled){
		return DGP_FALSE;
	}
	
	/* first is the run that starts before the other run
	*/
	if(run1->xPos > run2->xPos){
		first = run2;
		second = run1;
	} else {
		first = run1;
		second = run2;
	}
	
	offset = second->xPos - first->xPos;
	
	firstLength = first->cccrun.length - offset;
	secondLength = second->cccrun.length;
	if(firstLength > secondLength){
		length = secondLength;
	} else {
		length = firstLength;
	}
	
	DGP_VOID_ASSERT(length >= 0);
	
	if(length == 0){
		return DGP_FALSE;
	}
	
	if(!first->wash && !second->wash) {
		return (colorDist(first->cccrun.color, second->cccrun.color) <= cookie->wThresh);
	} 
	
	firstColor = first->cccrun.sampling_ + offset;
	secondColor = second->cccrun.sampling_;
	
	if(first->wash && second->wash){
		lastColor = firstColor + length;
		while(firstColor < lastColor){
			if(colorDistNat(*firstColor, *secondColor) > cookie->wThresh){
				return DGP_FALSE;
			}
			
			firstColor++;
			secondColor++;
		}
		
		return DGP_TRUE;
	}
	
	/*
	* One run is a wash and the other is not
	*
	* Put the solid run's color in "color"
	* and put the pointer to the wash's sampling
	* in firstColor
	*/
	if(first->wash){
		copyToSampling(color, second->cccrun.color);
	} else {
		firstColor = secondColor;
		copyToSampling(color, first->cccrun.color);
		
	}
	
	lastColor = firstColor + length;
	while(firstColor < lastColor){
		if(colorDistNat(*firstColor, color) > cookie->wThresh){
			return DGP_FALSE;
		}
		firstColor++;
	}
	
	/* if we made it here.  Then we must have a wash and a non-wash that matchup */
	return DGP_TRUE;
}

INLINE Set *linkSetWash(Set *x, Set *y)
{
	if(x == y) {
		return x;
	}
	if(x->rank > y->rank) {
		y->parent = x;
		/* Merge label lists of each wash */
		DgpAdt_liApp((DgpAdt_List)(x->data), (DgpAdt_List)(y->data));
		y->data = NULL;
		return x;
	} else {
		x->parent = y;
		/* Merge label lists of each wash */
		DgpAdt_liApp((DgpAdt_List)(y->data), (DgpAdt_List)(x->data));
		
		x->data = NULL;
		if(x->rank == y->rank) {
			(y->rank)++;
		}
		
		return y;
	}
}



INLINE void labelRunWash(ccc_cookie cookie, Run **currTopRun, Run *lastTopRun,
						 Run *currRun, int y)
{
	Run *topRun = *currTopRun;
	Set *label = currRun->label;
	CCstate *cc;
	int lastXplus1 = currRun->xPos + currRun->cccrun.length;
	Set *topSet;
	DgpAdt_List nonMatches;
	DgpAdt_ListNode ln, retnode;
	
	DgpAdt_liNew(&nonMatches, cookie->dgp_state);

	/*
	* Go through all the top runs that could be 8 connected
	* to the currRun
	*
	* link the run to matching runs
	*/
	while (topRun != lastTopRun) {
		if(runsMatchWash(currRun, topRun)) {
			/* Link runs for our simple closed loop finding stuff */
			topSet = findSet(topRun->label);
			if(label == (Set *)NULL) {
				label = topSet;
				cc = (CCstate*)label->data;
				linkCCtoRun(cc, currRun, y);
				
			} else {
				if(label == topSet){
				/* We should have just closed a loop
					*/
					((CCstate*)(topSet->data))->numHoles++;
					
				} else {
				/*  otherwise we just connected two branches 
				link up the sets,
				(and maybe for pass and a half add runlists together, and delete the 
				extra cc from the list of cc's
					*/
					label = linkSetCC(label, topSet);
					
				}
			}
		} else {
			/* add non-matching runs to wash-check list */
			DgpAdt_liAdd(nonMatches, topRun, &retnode);
		}
		
		
		if((topRun->xPos + topRun->cccrun.length) > lastXplus1){
			break;
		}
		
		topRun++;
	} 
	
	/* If the label is still null the run didn't match anything */
	/* make a new set */
	if(label == (Set *)NULL) {
		label = makeSet(cookie);
		label->data = makeCC(cookie, currRun, y);
		
	}
	currRun->label = label;
	((CCstate*)(label->data))->active = DGP_TRUE;
	
	/* if the last overlapping run is is only touching by  */
	/* the corner, roll back the *currTopNode for the next call */
	if ((topRun != NULL) &&
		(topRun < lastTopRun) && 
		(lastXplus1 == topRun->xPos)) {
		topRun--;
	}
	
	*currTopRun = topRun;
	
	/*
	* Second loop through
	* 
	* Check for wash matches
	* If there is match then merge labels
	* and update lists of "wash"-connected cc's
	*/
	label = getWashLabel(currRun);
	
	DGPADT_FOREACH(ln, nonMatches){
		topRun = DGPADT_LIGET(ln);
		
		if(runsMatchWashNoDiagonal(cookie, currRun, topRun)) {
			/* Link runs for our simple closed loop finding stuff */
			topSet = findSet(getWashLabel(topRun));
			if(label == (Set *)NULL) {
				label = topSet;
				
				/* Store the labels of each in each others wash connection list */
				DgpAdt_liAdd(((CCstate*)(findSet(topRun->label)->data))->washConnections, findSet(currRun->label), &retnode);
				DgpAdt_liAdd(((CCstate*)(findSet(currRun->label)->data))->washConnections, findSet(topRun->label), &retnode);
				
			} else {
				label = findSet(label);
				if(label == topSet){
				/* We should have just closed a loop
					*/
					
					/* Might need to store some info about this */
					
				} else {
				/*  otherwise we just connected two branches 
				link up the sets,
					*/
					label = linkSetWash(findSet(label), findSet(topSet));
					
					/* add the topRun->cc to the currRun->cc's touching DgpAdt_List */
					DgpAdt_liAdd(((CCstate*)(findSet(topRun->label)->data))->washConnections,  findSet(currRun->label), &retnode);
					DgpAdt_liAdd(((CCstate*)(findSet(currRun->label)->data))->washConnections,  findSet(topRun->label), &retnode);
					
				}
			}
		}
		
	} 
	
	/* If the label is still null the run didn't match or "wash"-match anything */
	/* make a new wash set */
	if(label == (Set *)NULL) {
		label = makeSet(cookie);
		DgpAdt_liNew((DgpAdt_List*) &label->data, cookie->dgp_state);
		DgpAdt_liAdd(cookie->washSets, (void *)label, &retnode);
		
		
	}
	getWashLabel(currRun) = label;
	
	/* exit the loop with currTopNode pointing to the 
	* last possible 8 connected run, save the previous
	* in case of a diagonal connection
	*
	* notice, The top node can only become null if 
	* the currRun is the last on the line, in which case
	* the value of currTopNode is not used by the caller
	*/
	
	DgpAdt_liFree(nonMatches);
}

/*
* readAndLabel()
* The runs are labeled using disjoint sets so that the parent
* label of each run is the same iff the runs are in the same
* CCC.
*
* We read in each run, and have labelRun decide what to do with
* the label.
*
* labelRun also tells us the label of the CCC with the max area.
* We set the max area in the begining to be the length of the Dgp_Int32est
* run in the top row.  Area is defined as the number of pixels in the
* CCC.
*
* See readRun, labeRuns
*/
ErrCode readAndLabelWash(ccc_cookie cookie, RunCookie runCook, int numLines)
{	
  Dgp_State *dgp_state = cookie->dgp_state;
  DGPERR_INIT(readRunWash);
	int y=0;
	Run *pRuns, *topRun, *lastTopRun, *firstRun, currRun;
	int numRuns;
	int endLine = cookie->curLine + numLines;
	RunImage im = cookie->im;
	int width = cookie->width;
	DgpAdt_ListNode ln, ln2;
	Set *label, *curSet;
	CCstate *cc, *cc2;
	Run *run;
	DgpAdt_List ccMergeDgpAdt_List;
	DgpAdt_ListNode retnode;
	
	DgpAdt_liNew(&ccMergeDgpAdt_List, cookie->dgp_state);
	
	y = cookie->curLine;
	
	if(y == 0){
		/* Set them equal so no checking is done in the labeler */
		topRun = NULL;
		lastTopRun = NULL;
		pRuns = firstRun = rimStartRow(im, y);
		numRuns = 0;
		do{
			DGP_ERR_PASS_CLEANUP(readRunWash(cookie->rCook, &currRun, cookie, runCook));
			
			labelRunWash(cookie, &topRun, lastTopRun, &currRun, y);
			*pRuns = currRun;
			pRuns++;
			numRuns++;
		} while((currRun.xPos + currRun.cccrun.length) < width);
		
		rimEndRow(im, y, numRuns);
		y++;
	} 
	
	for(;y < endLine; y++){
		lastTopRun = pRuns = firstRun = rimStartRow(im, y);
		DGP_CHECK_ALLOC(pRuns);
		topRun = rimGetRow(im, y -1);
		
		for(run = topRun; run != lastTopRun; run++) {
			((CCstate*)(findSet(run->label)->data))->active = DGP_FALSE;
			
		}
		
		numRuns = 0;
		do{
			DGP_ERR_PASS_CLEANUP(readRunWash(cookie->rCook, &currRun, cookie, runCook));
			
			labelRunWash(cookie, &topRun, lastTopRun, &currRun, y);
			*pRuns = currRun;
			pRuns++;
			numRuns++;
		} while((currRun.xPos + currRun.cccrun.length) < width);
		
		for(run = rimGetRow(im, y - 1); run != lastTopRun; run++) {
			label = findSet(run->label);
			cc = (CCstate*)(label->data);
			if(!cc->active){
			/* this cc is finished
			*
			* Check to see if it is has been targeted for a wash
			* but it isn't "wash"like
				*/
				if(!cc->wash &&
					((cc->lrY - cc->ulY) > (cookie->height / 20))){
					
					if(DGPADT_LILEN(cc->washConnections) > 0){
						/* give this cc a new wash label */
						cc->washLabel = makeSet(cookie);
						DgpAdt_liNew((DgpAdt_List*) &cc->washLabel->data, cookie->dgp_state);
						DgpAdt_liAdd(cookie->washSets, (void *)cc->washLabel, &retnode);
					}
					
					/* Remove all the two way wash connections */
					ln = DGPADT_LIFIRST(cc->washConnections);
					while(ln != DGP_NULLLISTNODE){
						curSet = findSet((Set *)DGPADT_LIGET(ln));
						if(curSet == label) {
							/* we found this cc's connection to itself */
							ln = DGPADT_LINEXT(ln);
							
						} else {
							cc2 = (CCstate*) (curSet->data);
							ln2 = DGPADT_LIFIRST(cc2->washConnections);
							while(ln2 != DGP_NULLLISTNODE){
								curSet = findSet((Set *)DGPADT_LIGET(ln2));
								if(curSet == label){
									/* we found the connecting cc's link to this one */
									DgpAdt_liRemAndNext(ln2, &ln2);
								} else {
									ln2 = DGPADT_LINEXT(ln2);
								}
							}
							
							DgpAdt_liRemAndNext(ln, &ln);
						}
						
					}
					
				}

				/* Check wash connections and see if we can merge any */
				ln = DGPADT_LIFIRST(cc->washConnections);
				while(ln != DGP_NULLLISTNODE){
					curSet = findSet((Set *)DGPADT_LIGET(ln));
					cc2 = (CCstate*) (curSet->data);
					if((!cc2->active) &&
						(cc2->lrY < cc->ulY)){
						/* add cc to merge list */
						DgpAdt_liAdd(ccMergeDgpAdt_List, curSet, &retnode);
						cc->wash = DGP_TRUE;
						
						/* Remove the Dgp_Float64 link */
						ln2 = DGPADT_LIFIRST(cc2->washConnections);
						while(ln2 != DGP_NULLLISTNODE){
							curSet = findSet((Set *)DGPADT_LIGET(ln2));
							if(curSet == label){
								/* we found the connecting cc's link to this one */
								DgpAdt_liRemAndNext(ln2, &ln2);
							} else {
								ln2 = DGPADT_LINEXT(ln2);
							}
						}
						
						DgpAdt_liRemAndNext(ln, &ln);
					} else {
						ln = DGPADT_LINEXT(ln);
					}
				}
				
				ln = DGPADT_LIFIRST(ccMergeDgpAdt_List);
				while(ln !=DGP_NULLLISTNODE){
					label = linkSetCC(label, (Set *)DGPADT_LIGET(ln));
					DgpAdt_liRemAndNext(ln, &ln);
				}
			}
			
		}
		
		rimEndRow(im, y, numRuns);
		
  } 

  DgpAdt_liFree(ccMergeDgpAdt_List);

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

void
removeRunFromWash(Set *washLabel, Set *maxCCLabelDel, Set *maxCCLabelKeep)
{
	DgpAdt_List washDgpAdt_List;
	Set  *curSet;
	DgpAdt_ListNode ln;
	CCstate *ccDel, *ccKeep;
	
	washDgpAdt_List = (DgpAdt_List)(washLabel->data);
	
	ln = DGPADT_LIFIRST(washDgpAdt_List);
	while(ln != DGP_NULLLISTNODE){
		curSet = (Set *)DGPADT_LIGET(ln);
		if(findSet(curSet) == maxCCLabelDel){
			DgpAdt_liRemAndNext(ln, &ln);
		} else {
			ln = DGPADT_LINEXT(ln);
		}
	}
	
	ccDel = (CCstate*) (maxCCLabelDel->data);
	
	
	ln = DGPADT_LIFIRST(ccDel->washConnections);
	while(ln != DGP_NULLLISTNODE){
		curSet = (Set *)DGPADT_LIGET(ln);
		if(findSet(curSet) == maxCCLabelDel ||
			findSet(curSet) == maxCCLabelKeep){
			DgpAdt_liRemAndNext(ln, &ln);
		} else {
			ln = DGPADT_LINEXT(ln);
		}
	}
	
	ccKeep = (CCstate*) maxCCLabelKeep->data;
	
	ln = DGPADT_LIFIRST(ccKeep->washConnections);
	while(ln != DGP_NULLLISTNODE){
		curSet = (Set *)DGPADT_LIGET(ln);
		if(findSet(curSet) == maxCCLabelDel){
			DgpAdt_liRemAndNext(ln, &ln);
		} else {
			ln = DGPADT_LINEXT(ln);
		}
	}
	
}

void
findHardEdges(ccc_cookie cookie)
{
	int height = cookie->height;
	Run *firstRun, *lastRun;
	int y;
	Run *currRun, *runAfter;
	RunImage im = cookie->im;
	CCstate *currCC, *afterCC;
	Set *washLabel, *currLabel, *afterLabel;
	DgpAdt_ListNode retnode;
	
	for(y=0;y<height;y++) {
		firstRun = rimGetRow(im, y);
		lastRun = rimGetRow(im, y+1);
		
		/* Add the first run if necessary */
		currRun = firstRun;
		currLabel = findSet(currRun->label);
		currCC = (CCstate*)(currLabel->data);
		washLabel = findSet(currCC->washLabel);
		if(!currCC->added2wash){
			DgpAdt_liAdd((DgpAdt_List)(washLabel->data), currLabel, &retnode);	
			currCC->added2wash = DGP_TRUE;
		}
		
		
		for (runAfter = firstRun + 1; runAfter != lastRun; runAfter++) {
			
			afterLabel = findSet(runAfter->label);
			afterCC = (CCstate*)(afterLabel->data);
			washLabel = findSet(afterCC->washLabel);
			
			if(!afterCC->added2wash){
				DgpAdt_liAdd((DgpAdt_List)(washLabel->data), afterLabel, &retnode);	
				afterCC->added2wash = DGP_TRUE;
			}
			
			
			if((currCC != afterCC) && 
				(washLabel == findSet(currCC->washLabel))){
				
				/*
				*  We found a hard edge.  These two runs are targeted to merged into 
				* one wash cc, but the edge between them is greater than our wash threshold
				* so pick on of the CC's associated with the wash and remove it
				*/
				if(runAfter->wash && !currRun->wash){
					/* remove currRun */
					removeRunFromWash(washLabel, currLabel, afterLabel);
				} else if (currRun->wash && !runAfter->wash){
					/* remove runAfter */
					removeRunFromWash(washLabel, afterLabel, currLabel);
				} else {
				/*
				* Either they are both washes or both solid runs
					*/
					if(DGPADT_LILEN(currCC->washConnections) >
						DGPADT_LILEN(afterCC->washConnections)){
						/* remove runAfter */
						removeRunFromWash(washLabel, afterLabel, currLabel);
					}else {
						/* remove currRun */
						removeRunFromWash(washLabel, currLabel, afterLabel);
					}
				}
			}
			
			/* Update the current run */
			currRun = runAfter;
			currLabel = afterLabel;
			currCC = afterCC;
		}
		
		
	}
}
