#include "tokenCC.h"
#include <tok-lossless.h>

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libcomponent/tokenCC.c,v 1.10 2000/03/04 00:22:20 mkrocken Exp $");


ErrCode 
initTokBlock(Dgp_State *dgp_state, TokInfoBlock **block, int numToks)
{
  DGPERR_INIT(initTokBlock);

  TokInfoBlock *tBlock;

  DGP_SEC_MALLOC(tBlock, (TokInfoBlock *), sizeof(TokInfoBlock));

  DGP_SEC_MALLOC(tBlock->toks, (TokInfo *), sizeof(TokInfo) * numToks);

  tBlock->numAlloced = numToks;
  tBlock->num = 0;

  *block = tBlock;

  NORMAL_RETURN;

ERR_CLEANUP:
  ERR_RETURN;
}

void
cleanupTokBlock(Dgp_State *dgp_state, TokInfoBlock *block)
{
  Dgp_Free tmp_free;

  if(block == NULL) return;

  tmp_free = DGP_FREE;

  if(block->toks != NULL) tmp_free(block->toks);
  tmp_free(block);
}

ErrCode
initPosBlock(Dgp_State *dgp_state, PosInfoBlock **block, int num)
{
  DGPERR_INIT(initPosBlock);

  PosInfoBlock *pBlock;

  DGP_SEC_MALLOC(pBlock, (PosInfoBlock *), sizeof(PosInfoBlock));

  DGP_SEC_MALLOC(pBlock->poss, (PosInfo *), sizeof(PosInfo) * num);

  pBlock->numAlloced = num;
  pBlock->num = 0;
	
  *block = pBlock;

  NORMAL_RETURN;

ERR_CLEANUP:
  ERR_RETURN;
}


void
cleanupPosBlock(Dgp_State *dgp_state, PosInfoBlock *block)
{
  Dgp_Free tmp_free;

  if(block == NULL) return;

  tmp_free = DGP_FREE;

  if(block->poss != NULL) tmp_free(block->poss);
  tmp_free(block);
}



ErrCode
tokenAddCC(Dgp_State *dgp_state, TokState tokstate, TokInfoBlock *tBlock, runToken *rTok, 
           token **tok)
{
  DGPERR_INIT(tokenAddCC);

  int oldOffset;

  if(tBlock->num >= tBlock->numAlloced){
    tBlock->numAlloced = tBlock->numAlloced * 3 / 2;
    DGP_SEC_REALLOC(tBlock->toks, (TokInfo *), tBlock->numAlloced * sizeof(TokInfo));
  }
	
  /* 
   * This is a bit of a hack
   * oldOffset can be one of three things:
   *
   * 1. if the token definately matches a previous token then oldOffset will be
   * the userdata set when that token was added
   *
   * 2. if the token definately doesn't match anything then oldOffset will be 
   * the the userdata set on this call
   *
   * 3. if the tokenizer can't say right now whether the token matches or doesn't
   * match then oldOffset will be 0
   *
   * This code makes sense for the first two cases, but its a bit silly for the
   * last case.  In reality the tokAddToken called for a given tokBlock will
   * have some predictable behavior:  
   * One version of tokAddToken will only do the first 2 cases.
   * The other version will only do the last case.  
   * 
   * If the last case is the only one executed for a given tok block, 
   * there will be only one valid tokenInfo for the block [0].  It will
   * have a null token* thus giving all the positions null token* which is
   * just what we want
   */

  tBlock->toks[tBlock->num].token = NULL;

  DGP_ERR_PASS_CLEANUP(
		       tokAddToken(tokstate, rTok, (void *)tBlock->num, (void **)&oldOffset));
		
  if(oldOffset == tBlock->num){
    tBlock->num++;
  } 

  *tok = tBlock->toks[oldOffset].token;
	
  NORMAL_RETURN;
ERR_CLEANUP:
  ERR_RETURN;
}
