#include "coord.h"
#include "orientation.h"
#include "PsgFix.h"
/* coord.c
   wbell 10/12/99
   some coordinate system manipulation functions.
 */

/* transforms from ORIENT_RD to another DgpCoord_Orientation . */
/* 21.6.2000 / dhtesq. */
/* Certain terms in certain cases were buggy;  fixed. */
void 
DgpCoord_coordOrientTransformNative(Dgp_Int32 xbase, Dgp_Int32 ybase, Dgp_Int32 wbase, Dgp_Int32 hbase,
				Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h, 
				DgpCoord_Orientation  o2, Dgp_Int32 * pfo, Dgp_Int32 * pso,
				Dgp_Int32 * pfs, Dgp_Int32 * pss, Dgp_Int32 * pfw, Dgp_Int32 * psw)
{
  Dgp_Int32 fo = -1, so = -1, fs = -1, ss = -1, fw = -1, sw = -1;
  DGP_VOID_PROLOGUE("coordOrientTransformNative()",NULL);

  switch(o2){
  case ORIENT_RD:
    fo = xbase;
    so = ybase;
    fs = x;
    ss = y;
    fw = w;
    sw = h;
    break;
  case ORIENT_LD:
    fo = xbase + (wbase - 1);
    so = ybase;
    fs = fo - (x + (w - 1));    /* fixed... */
    ss = y;
    fw = w;
    sw = h;
    break;
  case ORIENT_RU:
    fo = xbase;
    so = ybase + (hbase - 1);
    fs = x;
    ss = y + (h - 1);
    fw = w;
    sw = h;
    break;
  case ORIENT_LU:
    fo = xbase + (wbase - 1);
    so = ybase + (hbase - 1);
    fs = fo - (x + (w - 1));    /* fixed... */
    ss = so - (y + (h - 1));    /* fixed... */
    fw = w;
    sw = h;
    break;
  case ORIENT_UR:
    fo = ybase + (hbase - 1);
    so = xbase;
    fs = fo - (y + (h - 1));    /* fixed... */
    ss = x;
    fw = h;
    sw = w;
    break;
  case ORIENT_DR:
    fo = ybase;
    so = xbase;
    fs = y;
    ss = x;
    fw = h;
    sw = w;
    break;
  case ORIENT_UL:
    fo = ybase + (hbase - 1);
    so = xbase + (wbase - 1);
    fs = y + (h - 1);
    ss = x + (w - 1);
    fw = h;
    sw = w;
    break;
  case ORIENT_DL:
    fo = ybase;
    so = xbase + (wbase - 1);
    fs = y;
    ss = so - (x + (w - 1));    /* fixed... */
    fw = h;
    sw = w;
    break;
  default:
    DGP_VOID_ASSERT(DGP_FALSE);
  }
  
  DGP_VOID_ASSERT(fo >= 0);
  DGP_VOID_ASSERT(so >= 0);
  DGP_VOID_ASSERT(fs >= 0);
  DGP_VOID_ASSERT(ss >= 0);
  DGP_VOID_ASSERT(fw >= 0);
  DGP_VOID_ASSERT(sw >= 0);

  if (pfo){
    *pfo = fo;
  }
  if (pso){
    *pso = so;
  }
  if (pfs){
    *pfs = fs;
  }
  if (pss){
    *pss = ss;
  }
  if (pfw){
    *pfw = fw;
  }
  if (psw){
    *psw = sw;
  }
}

/* coordOrientTransform()
   Transforms a coordinate system in ORIENT_RD to another DgpCoord_Orientation .

   Say you have an image that occupies some space in a larger coordinate 
   system, and its location is (xbase, ybase),
   (xbase + wbase-1, ybase + hbase-1) in DgpCoord_Orientation ORIENT_RD.
   You have some set of coordinates in that larger space that is a portion
   of your image (x1, y1), (x1 + w1-1, y1 + h1-1) and you want to know
   where this set of coordinates maps to in the image if the image is
   actually stored in DgpCoord_Orientation o2. 
   
   We will return you a normalized set of coordinates in that frame that
   this rectangle occupies in the new coordinate system 
   (meaning that we'll give you a rectangle that is normalized with relation
    to fast / slow coordinate system, so that fastwidth is >=0 and 
    slowwidth >= 0)
   as well as giving you the mapping of x1, y1 into that normalized coordinate
   space.

   xbase, ybase, wbase, hbase - the image in some larger coordinate system.
   x1, y1, w1, h1 - the subrectangle of that image (in the larger system)
     that you wish to have transformed.
   o2 - which DgpCoord_Orientation  you wish to transform into (which DgpCoord_Orientation 
     your raster is in)
   fo, so - where x1, y1 is in the normalized raster coordinate system
     ( >= 0)
   fs, ss - the upper-left (read as min fast, min slow) point of the 
    rectangle in raster coordinate system (>= 0)
   fw, sw - the width of this rectangle in both the fast and slow directions.
 */

ErrCode
DgpCoord_coordOrientTransform(Dgp_Int32 xbase, Dgp_Int32 ybase, Dgp_Int32 wbase, Dgp_Int32 hbase,
		     Dgp_Int32 x1, Dgp_Int32 y1, Dgp_Int32 w1, Dgp_Int32 h1,
		     DgpCoord_Orientation  o2,
		     Dgp_Int32 * fo, Dgp_Int32 * so,
		     Dgp_Int32 * fs, Dgp_Int32 * ss,
		     Dgp_Int32 * fw, Dgp_Int32 * sw,
		     Dgp_State * dgp_state)
{
  DGP_STATE_PROLOGUE("coordOrientTransform()", dgp_state);
  DGP_ERR_GEN(!dgp_state, cErrInvalidParm);
  DGP_ERR_GEN(!ISVALIDORIENT(o2), cErrInvalidParm);
  DGP_ERR_GEN(wbase < 0 || hbase < 0, cErrInvalidParm);
  DGP_ERR_GEN(x1 < xbase || y1 < ybase, cErrInvalidParm);
  DGP_ERR_GEN((x1 + w1) > (xbase + wbase) || (y1 + h1) > (ybase + hbase),
	      cErrInvalidParm);
  
  /* we need to get xbase, ybase, wbase, hbase back into native coords. */
  DgpCoord_coordOrientTransformNative(xbase, ybase, wbase, hbase,
			     x1, y1, w1, h1,
			     o2, fo, so, fs, ss, fw, sw);
  
  NORMAL_RETURN;
}




