#ifndef COORD_H
#define COORD_H

#include "dgp_state.h"
#include "orientation_t.h"

/* coord.h
   wbell 10/12/99
   some coordinate system manipulation functions.
 */


#ifdef __cplusplus
extern "C"
{
#endif

/* coordOrientTransform()
   transforms a coordinate system in ORIENT_RD to another DgpCoord_Orientation .

   say you have an image that occupies some space in a larger coordinate 
   system, and it's location is (xbase, ybase),
   (xbase + wbase-1, ybase + hbase-1) in DgpCoord_Orientation  ORIENT_RD.
   you have some set of coordinates in that larger space that is a portion
   of your image (x1, y1), (x1 + w1-1, y1 + h1-1) and you want to know
   where this set of coordinates maps to in the image if the image is
   actually stored in DgpCoord_Orientation  o2. 
   
   We will return you a normalized set of coordinates in that frame that
   this rectangle occupies in the new coordinate system 
   (meaning that we'll give you a rectangle that is normalized with relation
    to fast / slow coordinate system, so that fastwidth is >=0 and 
    slowwidth >= 0)
   as well as giving you the mapping of x1,y1 into that normalized coordinate
   space.

   xbase, ybase, wbase, hbase - the image in some larger coordinate system.
   x1, y1, w1, h1 - the subrectangle of that image (in the larger system)
     that you wish to have transformed.
   o2 - which DgpCoord_Orientation  you wish to transform into (which DgpCoord_Orientation 
     your raster is in)
   fo, so - where x1, y1 is in the normalized raster coordinate system
     ( >= 0)
   fs, ss - the upper-left (read as min fast, min slow) point of the 
    rectangle in raster coordinate system (>= 0)
   fw, sw - the width of this rectangle in both the fast and slow directions.
 */

ErrCode
DgpCoord_coordOrientTransform(Dgp_Int32 xbase, Dgp_Int32 ybase, Dgp_Int32 wbase, Dgp_Int32 hbase,
		     Dgp_Int32 x1, Dgp_Int32 y1, Dgp_Int32 w1, Dgp_Int32 h1,
		     DgpCoord_Orientation  o2,
		     Dgp_Int32 * fo, Dgp_Int32 * so,
		     Dgp_Int32 * fs, Dgp_Int32 * ss,
		     Dgp_Int32 * fw, Dgp_Int32 * sw,
		     Dgp_State * dgp_state);

#ifdef __cplusplus
}
#endif
#endif
