#include "orientation.h"
#include "util.h"
#include "dgp_types.h"
#include "ctype.h"
#include "dgp_err_control.h"
#include "PsgFix.h"

static INLINE Dgp_Bool isVert(DgpCoord_Orientation o)
{
  DGP_VOID_PROLOGUE("isVert()", NULL);
  switch(o){
  case ORIENT_RD:
  case ORIENT_LD:
  case ORIENT_RU:
  case ORIENT_LU:
    return DGP_FALSE;
    break;
  case ORIENT_DL:
  case ORIENT_DR:
  case ORIENT_UL:
  case ORIENT_UR:
    return DGP_TRUE;
    break;
  }
  DGP_VOID_ASSERT(DGP_FALSE);
  return DGP_FALSE;
}


/* determines if 2 orientations are rotated from one another (that some  */
/* series of transformations will get from one to another contains a  */
/* rotation) */
Dgp_Bool 
DgpCoord_isRotated(DgpCoord_Orientation o1, DgpCoord_Orientation o2)
{
  Dgp_Bool o1vert = DGP_FALSE, o2vert = DGP_FALSE;
  DGP_VOID_PROLOGUE("isRotated()", NULL);
  DGP_VOID_ASSERT(ISVALIDORIENT(o1));
  DGP_VOID_ASSERT(ISVALIDORIENT(o2));

  /* we're going to determine if they are vertical or horizontal with */
  /* respect to orient_rd, and then see if the two are the same. */
  o1vert = isVert(o1);
  o2vert = isVert(o2);

  if (o1vert != o2vert){
    return DGP_TRUE;
  }
  else{
    return DGP_FALSE;
  }
}



/* given a two character string such as "RD", returns the orientation  */
/* associated with those chars. Will handle case. */
ErrCode
DgpCoord_strOrient(Dgp_Int8 str[2], DgpCoord_Orientation * po, Dgp_State * dgp_state)
{
  Dgp_Int8 l1,l2;
  DgpCoord_Orientation out;

  DGP_STATE_PROLOGUE("strOrient()", dgp_state);
  DGP_ERR_GEN(!po, cErrInvalidParm);
  DGP_ERR_GEN(!str, cErrInvalidParm);

  l1 = toupper(str[0]);
  l2 = toupper(str[1]);

  if (l1 == 'R' && l2 =='D'){
    out = ORIENT_RD;
  }
  else if (l1 == 'R' && l2 == 'U'){
    out=ORIENT_RU;
  }
  else if (l1 == 'L' && l2 == 'D'){
    out = ORIENT_LD;
  }
  else if (l1 == 'L' && l2 == 'U'){
    out = ORIENT_LU;
  }
  else if (l1 == 'U' && l2 == 'L'){
    out = ORIENT_UL;
  }
  else if (l1 == 'U' && l2 == 'R'){
    out = ORIENT_UR;
  }
  else if (l1 == 'D' && l2 == 'L'){
    out = ORIENT_DL;
  }
  else if (l1 == 'D' && l2 == 'R'){
    out = ORIENT_DR;
  }
  else {
    DGP_ERR(cErrInvalidParm);
  }

  *po = out;

  NORMAL_RETURN;
}

/* given an orientation and a 2 character buffer, this will fill in those */
/* to chars, such as "RD" for ORIENT_RD */
ErrCode
DgpCoord_orientStr(DgpCoord_Orientation o, Dgp_Int8 str[2], Dgp_State * dgp_state)
{
  DGP_STATE_PROLOGUE("strOrient()", dgp_state);
  DGP_ERR_GEN(!str, cErrInvalidParm);
  DGP_ERR_GEN(!ISVALIDORIENT(o), cErrInvalidParm);

  switch(o){
  case ORIENT_RD:
    str[0] = 'R';
    str[1] = 'D';
    break;
  case ORIENT_RU:
    str[0] = 'R';
    str[1] = 'U';
    break;
  case ORIENT_LD:
    str[0] = 'L';
    str[1] = 'D';
    break;
  case ORIENT_LU:
    str[0] = 'L';
    str[1] = 'U';
    break;
  case ORIENT_DR:
    str[0] = 'D';
    str[1] = 'R';
    break;
  case ORIENT_DL:
    str[0] = 'D';
    str[1] = 'L';
    break;
  case ORIENT_UR:
    str[0] = 'U';
    str[1] = 'R';
    break;
  case ORIENT_UL:
    str[0] = 'U';
    str[1] = 'L';
    break;
  }

  NORMAL_RETURN;
}













