#include <stdlib.h>                /* VxWorks abs() implementation here.  / dhtesq. */
#include <math.h>
#include "resolution_t.h"


Dgp_Int32 DgpCoord_greatest_common_div(Dgp_Int32 m, Dgp_Int32 n)
{
  Dgp_Int32 t;
  if(!m){
    return(1);
  }
  do {
    t = n % m;
    n = m;
    m = t;
  } while (m);
  
  return(n);
}


Dgp_Int32 DgpCoord_least_common_mul(Dgp_Int32 i, Dgp_Int32 j)
{
  return (ABS( (i * j) / DgpCoord_greatest_common_div(i, j)));
}

DgpCoord_Rat DgpCoord_eqi(Dgp_Int32 i)
{
  DgpCoord_Rat ret;
  ret.num=i;
  ret.den=1;
  return ret;
}

DgpCoord_Rat DgpCoord_invert(DgpCoord_Rat r)
{
  DgpCoord_Rat ret;
  ret.num=r.den;
  ret.den=r.num;
  return ret;
}


DgpCoord_Rat DgpCoord_eq(Dgp_Int32 num, Dgp_Int32 den)
{
  DgpCoord_Rat x;
  Dgp_Int32 gcd=DgpCoord_greatest_common_div(num,den);
  x.num=num/gcd;
  x.den=den/gcd;
  return x;
}

void DgpCoord_normalize(DgpCoord_Rat *r)
{
  Dgp_Int32 gcd=DgpCoord_greatest_common_div(r->num,r->den);
  r->num/=gcd;
  r->den/=gcd;
}

DgpCoord_Rat DgpCoord_muli(DgpCoord_Rat r, Dgp_Int32 x)
{
  r.num*=x;

  DgpCoord_normalize(&r);

  return r;
}

DgpCoord_Rat DgpCoord_mul(DgpCoord_Rat r1, DgpCoord_Rat r2)
{
  DgpCoord_Rat ret;

  ret.num=(r1.num*r2.num);
  ret.den=(r1.den*r2.den);

  DgpCoord_normalize(&ret);

  return ret;
}


DgpCoord_Rat DgpCoord_add(DgpCoord_Rat r1, DgpCoord_Rat r2)
{
  DgpCoord_Rat ret;
	
  ret.num=(r1.num*r2.den)+(r2.num*r1.den);
  ret.den=(r1.den*r2.den);

  DgpCoord_normalize(&ret);

  return ret;
}

DgpCoord_Rat DgpCoord_div(DgpCoord_Rat r1, DgpCoord_Rat r2)
{
  DgpCoord_Rat ret;

  ret.num=(r1.num*r2.den);
  ret.den=(r1.den*r2.num);

  DgpCoord_normalize(&ret);

  return ret;
}

DgpCoord_Rat DgpCoord_divi(DgpCoord_Rat r, Dgp_Int32 x)
{
  r.den*=x;

  DgpCoord_normalize(&r);

  return r;
}

DgpCoord_Rat DgpCoord_idiv(Dgp_Int32 x, DgpCoord_Rat r)
{
  DgpCoord_Rat ret;

  ret.num=x*r.den;
  ret.den=r.num;

  DgpCoord_normalize(&ret);

  return ret;
}

Dgp_Int32 DgpCoord_floor(DgpCoord_Rat r)
{
  return (Dgp_Int32)(((Dgp_Float32)r.num)/((Dgp_Float32)r.den));
}

Dgp_Int32 DgpCoord_ceil(DgpCoord_Rat r)
{
  return ((Dgp_Int32)ceil(((Dgp_Float32)r.num)/((Dgp_Float32)r.den)));
}

DgpCoord_Rat DgpCoord_sub(DgpCoord_Rat r1,DgpCoord_Rat r2)
{
  DgpCoord_Rat ret;
	
  ret.num=(r1.num*r2.den)-(r2.num*r1.den);
  ret.den=(r1.den*r2.den);

  DgpCoord_normalize(&ret);

  return ret;
}

Dgp_Bool DgpCoord_gei(DgpCoord_Rat r, Dgp_Int32 x)
{
  if((Dgp_Int32)((Dgp_Float32)r.num/(Dgp_Float32)r.den)>=x){
    return DGP_TRUE;
  }
  else{
    return DGP_FALSE;
  }
}


Dgp_Bool DgpCoord_equal(DgpCoord_Rat r1, DgpCoord_Rat r2)
{
  if((r1.num*r2.den)==(r2.num*r1.den)){
    return DGP_TRUE;
  }
  else{
    return DGP_FALSE;
  }
}

Dgp_Bool DgpCoord_ge(DgpCoord_Rat r1, DgpCoord_Rat r2)
{
  if((r1.num*r2.den)>=(r2.num*r1.den)){
    return DGP_TRUE;
  }
  else{
    return DGP_FALSE;
  }
}

Dgp_Bool DgpCoord_g(DgpCoord_Rat r1, DgpCoord_Rat r2)
{
  if((r1.num*r2.den)>(r2.num*r1.den)){
    return DGP_TRUE;
  }
  else{
    return DGP_FALSE;
  }
}

Dgp_Bool DgpCoord_l(DgpCoord_Rat r1, DgpCoord_Rat r2)
{
  if((r1.num*r2.den)<(r2.num*r1.den)){
    return DGP_TRUE;
  }
  else{
    return DGP_FALSE;
  }
}


Dgp_Bool DgpCoord_le(DgpCoord_Rat r1, DgpCoord_Rat r2)
{
  if((r1.num*r2.den)<=(r2.num*r1.den)){
    return DGP_TRUE;
  }
  else{
    return DGP_FALSE;
  }
}

Dgp_Float32 DgpCoord_float(DgpCoord_Rat r)
{
  return ((Dgp_Float32)r.num/(Dgp_Float32)r.den);
}

Dgp_Float64 DgpCoord_double(DgpCoord_Rat r)
{
  return ((Dgp_Float64)r.num/(Dgp_Float64)r.den);
}

