


#include "deskew_bmp.h"
/*

#include <io.pub>

*/
#include "libiow.pub"
#include "libpixrw.pub"
#include <rotatop.pub>
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libdeskew/deskew_bmp.c,v 1.6 2003/01/27 21:48:30 awilliam Exp $");

#define DGP_BITS_ULONG 32
#define DGP_BITS_ULONG_2   ((unsigned int)DGP_BITS_ULONG >> 1)
#define DGP_BITS_ULONG_4   ((unsigned int)DGP_BITS_ULONG >> 2)
#define DGP_BITS_ULONG_8   ((unsigned int)DGP_BITS_ULONG >> 3)
#define DGP_BITS_ULONG_16  ((unsigned int)DGP_BITS_ULONG >> 4)
#define DGP_BITS_ULONG_32  ((unsigned int)DGP_BITS_ULONG >> 5)
#define DGP_BITS_ULONG_64  ((unsigned int)DGP_BITS_ULONG >> 6)


static void
find_edges(IPWState * ipwstate, PIXR *pixr, Dgp_Int32 *new_x, Dgp_Int32 *new_y, 
	   Dgp_Int32 *new_w, Dgp_Int32 *new_h);

static int
bit_num_from_right(Dgp_UInt32 val);

/* deskew a bitmap given:
     ipwstate --  assumed to be valid (already initialized).
     old_bmp,
     old_w,
     old_h   -- old bitmap, width and height.  The original bitmap is
                left untouched by this function.
     new_bmp,
     new_w,
     new_h,
     new_lpl -- new bitmap, width, height, and longsperline.
                NOTE: You should call scSetAllocators if you want to change
		the malloc used to create the new bitmap.  (free may also be
		used, but only if the function fails).
     skew_angle -- need i say?
     err_out -- stream to which errors should be printed, iff not NULL.
     a       -- pointer to structure as defined in Dgp_Allocators.h of memory
                Dgp_Allocators to be used.
     max_w,
     max_h   -- maximum dimensions for new bmp.
*/

ErrCode 
DgpDeskew_bmp(Dgp_UInt32 *old_bmp, Dgp_Int32 old_w, Dgp_Int32 old_h,
	   Dgp_Int32 old_lpl, Dgp_UInt32 **new_bmp, Dgp_Int32 *new_w, Dgp_Int32 *new_h, Dgp_Int32 *new_lpl,
	   Dgp_Float32 skew_angle, Dgp_Int32 max_w, Dgp_Int32 max_h, Dgp_State *dgp_state) 
{
  DGP_PROLOGUE("deskew_bmp()", DGP_PERRSTATE);

  PIXR *rot_pixr1 = (PIXR *)NULL, *rot_pixr2 = (PIXR *)NULL;
  Dgp_Int32 new_x = 0, new_y = 0, i, bpl = 0, tmp_bpl = 0, tmp_lpl = 0;
  UInt8 *ptr = NULL;
  Dgp_UInt32 *bits = (Dgp_UInt32)NULL;
  IPWState *ipwstate = DGP_PIPWSTATE;
  Dgp_Allocators *a = DGP_PALLOCATORS;


  DGP_CHECK_NULL(new_bmp);
  /* pointer to new bmp should be NULL */
  DGP_CHECK_NOT_NULL(*new_bmp);
  
  /* create new pixr for rotating */
  DGP_ERR_PASS_CLEANUPF0(w_createPixr(ipwstate, (Int32)(old_w), (Int32)(old_h), 1, 0, 0, cFrameBits,
		  &rot_pixr1),
      "pixr creation failed for token rotation");
  
  /* get new pixr info */
  w_pixrGetImage(ipwstate, rot_pixr1, &ptr);
  w_pixrGetBpl(ipwstate, rot_pixr1, (Int32 *)&bpl);
  
  /* get old bitmap info */
  bits = old_bmp;
  tmp_bpl = old_lpl * sizeof( Dgp_UInt32);

  /* copy old bitmap into pixr */
  for(i = 0; i < old_h; i++) {
    memcpy((char *)ptr, (char *)bits, tmp_bpl);
    bits += old_lpl;
    ptr += bpl;
  }
  
  /* rotate pixr, writing it into a new pixr */
  DGP_ERR_PASS_CLEANUPF0(w_rotateArb(ipwstate, rot_pixr1, (Dgp_Float32)skew_angle, cFalse,
		 cFalse, &rot_pixr2),
     "rotation of token failed");

  /* free unrotated pixr with frame */
  w_destroyPixr(ipwstate, &rot_pixr1);

  /* find any padding around rotated pixr */
  find_edges(ipwstate, rot_pixr2, &new_x, &new_y, new_w, new_h);

  /* make sure new bmp is not bigger than maximum dimensions */
  if(*new_w > max_w) *new_w = max_w;
  if(*new_h > max_h) *new_h = max_h;
  
  /* crop rotated token in place */
  DGP_ERR_PASS_CLEANUPF0(w_cutRectIP(ipwstate, new_x, new_y, *new_w, *new_h, rot_pixr2),
    "cropping of rotated token failed");
  
  /* get info from rotated pixr */
  w_pixrGetBpl(ipwstate, rot_pixr2, (Int32 *)&bpl);

  /* i shouldn't have to get width and height, but that's the bug i found... */
  w_pixrGetWidth(ipwstate, rot_pixr2, (Int32 *)new_w);
  w_pixrGetHeight(ipwstate, rot_pixr2, (Int32 *)new_h);

  /* compute new bitmap info */
  tmp_lpl = (int)((*new_w + DGP_BITS_ULONG - 1) / DGP_BITS_ULONG);
  tmp_bpl = tmp_lpl * sizeof(Dgp_UInt32);
  *new_lpl = tmp_lpl;
  
  /* create new bitmap */
  DGP_CHECK_ALLOC_CLEANUP(*new_bmp = (Dgp_UInt32 *)a->malloc(tmp_bpl * (*new_h)));

  /* get ptrs to the two bitmaps so that we can copy the pixr (with frame)
     to the bitmap (without frame) */
  bits = *new_bmp;
  w_pixrGetImage(ipwstate, rot_pixr2, &ptr);

  /* copy rotated pixr into new token */
  for(i = 0; i < *new_h; i++) {
    memcpy((char *)bits, (char *)ptr, tmp_bpl);
    bits += tmp_lpl;
    ptr += bpl;
  }

  /* free rotated pixr with frame */
  w_destroyPixr(ipwstate, &rot_pixr2);


  NORMAL_RETURN;

ERR_CLEANUP:

  if(*new_bmp) {
    (a->free)(*new_bmp);
    *new_bmp = NULL;
  }

  if(rot_pixr1) w_destroyPixr(ipwstate, &rot_pixr1);
  if(rot_pixr2) w_destroyPixr(ipwstate, &rot_pixr2);

  ERR_RETURN;
}


/* find the new edges of a rotated token, get rid of any padding. */
static void
find_edges(IPWState * ipwstate, PIXR *pixr, Dgp_Int32 *new_x, Dgp_Int32 *new_y, 
	   Dgp_Int32 *new_w, Dgp_Int32 *new_h)
{
  Dgp_Int32 x = 0, y = 0, w, h, i, lpl;
  Dgp_UInt32 *lptr, val;

  w_pixrGetWidth(ipwstate, pixr, (Int32 *)&w);
  w_pixrGetHeight(ipwstate, pixr, (Int32 *)&h);

  *new_x = x; *new_y = y; *new_w = w; *new_h = h;

  lpl = (w + DGP_BITS_ULONG - 1) / DGP_BITS_ULONG;

  val = 0;
  while(!val) {
    w_getLine(ipwstate, pixr, (y++), (UInt8 **)&lptr);
    for(i = 0; ((i < lpl) && (!val)); i++) val |= *(lptr++);
  }
  y--;

  val = 0;
  while(!val) {
    w_getLine(ipwstate, pixr, (--h), (UInt8 **)&lptr);
    for(i = 0; ((i < lpl) && (!val)); i++) val |= *(lptr++);
  }
  h = h - y + 1;

  val = 0;
  while(!val) {
    for(i = 0; i < *new_h; i++) {
      w_getLine(ipwstate, pixr, i, (UInt8 **)&lptr);
      val |= lptr[x];
    }
    x++;
  }
  x = (x - 1) * DGP_BITS_ULONG;

  x += DGP_BITS_ULONG - bit_num_from_right(val) - 1;

  w = lpl;
  val = 0;
  while(!val) {
    w--;
    for(i = 0; i < *new_h; i++) {
      w_getLine(ipwstate, pixr, i, (UInt8 **)&lptr);
      val |= lptr[w];
    }
  }
  val &= (-((long)val));
  w = (w + 1) * DGP_BITS_ULONG;

  w -= (bit_num_from_right(val) + x);

  *new_x = x; *new_y = y; *new_w = w; *new_h = h;
}


/* stupid function to find left-most bit in a long. */
static int
bit_num_from_right(Dgp_UInt32 val)
{
  Dgp_Int32 n = 0;

  if(val >> DGP_BITS_ULONG_2) {
    n += DGP_BITS_ULONG_2;
    val = val >> DGP_BITS_ULONG_2;
  }
  if(val >> DGP_BITS_ULONG_4) {
    n+= DGP_BITS_ULONG_4;
    val = val >> DGP_BITS_ULONG_4;
  }
  if(val >> DGP_BITS_ULONG_8) {
    n+= DGP_BITS_ULONG_8;
    val = val >> DGP_BITS_ULONG_8;
  }
  if(val >> DGP_BITS_ULONG_16) {
    n+= DGP_BITS_ULONG_16;
    val = val >> DGP_BITS_ULONG_16;
  }
  if(val >> DGP_BITS_ULONG_32) {
    n+= DGP_BITS_ULONG_32;
    val = val >> DGP_BITS_ULONG_32;
  }
#if 0
  if(val >> DGP_BITS_ULONG_64) {
    n+= DGP_BITS_ULONG_64;
    val = val >> DGP_BITS_ULONG_64;
  }
#endif

  return(n);
}

