#include <stdio.h>
#include <libpixrw.pub>
#include <libjpegw.pub>
#include <rescale.pub>

#include "dgp_state.h"
#include "image.h"
#include "jtokapi.h"
#include "mrc2file.h"
#include "scancolor.h"
#include "silx_encode_priv.h"
#include "PsgFix.h"

#ifndef JC_DEFAULT
#define JC_DEFAULT 0
#endif

typedef struct {
  Dgp_State *dgp_state;
  Dgp_Int8 *buf;
  Dgp_UInt32 size;
  Dgp_UInt32 alloced;
} jcookie;

#define DEFAULT_DPI 300

#ifndef PSG
static ErrCode 
jpeg_Compress_callback(IPWState *ipwstate, 
                              void *cookie, 
                              Dgp_UInt8 *buf, 
                              Dgp_UInt32 nbytes,
                              Dgp_UInt32 *nwritten)
{
  jcookie *jc = (jcookie *)cookie;
  DGP_PROLOGUE("jpeg_Compress_callback()", jc->DGP_PERRSTATE);
  Dgp_UInt32 size;
  Dgp_Int8 *tmpptr;

  (void)ipwstate;

  size = jc->size + nbytes;
  if(size > jc->alloced){
    jc->alloced = (size * 3) / 2;
    DGP_CHECK_ALLOC(tmpptr = (Dgp_Int8 *)(jc->DGP_REALLOC)(jc->buf, jc->alloced));
    jc->buf = tmpptr;
  }

  memcpy(jc->buf + jc->size, buf, nbytes);
  jc->size += nbytes;

  *nwritten = nbytes;

  NORMAL_RETURN;
}
#else
static ErrCode 
jpeg_Compress_callback(IPWState *ipwstate, 
                              void *cookie, 
                              Dgp_UInt8 *buf, 
                              Dgp_UInt32 nbytes)
{
  jcookie *jc = (jcookie *)cookie;
  DGP_PROLOGUE("jpeg_Compress_callback()", jc->DGP_PERRSTATE);
  Dgp_UInt32 size;
  Dgp_Int8 *tmpptr;

  (void)ipwstate;

  size = jc->size + nbytes;
  if(size > jc->alloced){
    jc->alloced = (size * 3) / 2;
    DGP_CHECK_ALLOC(tmpptr = (Dgp_Int8 *)(jc->DGP_REALLOC)(jc->buf, jc->alloced));
    jc->buf = tmpptr;
  }

  memcpy(jc->buf + jc->size, buf, nbytes);
  jc->size += nbytes;

  NORMAL_RETURN;
}
#endif

ErrCode
Dgp_segmentation (PIXR* read, 
                  Dgp_CompressedPage *cPage, 
				  Dgp_Int32 jpeg_dpi, 
                  Dgp_Int32 jpeg_qual, 
                  Dgp_Int32 sizeThresh, 
                  sc_cc **text_cc, 
                  Dgp_UInt32 *numccs,
                  DgpImage_Gray *grayImage)
{
  Dgp_State	        *dgp_state = cPage->dgp_state;
  PIXR		         *background = NULL, *scaled_bg = NULL;
  void		        *ipstate = (void*) dgp_state->ipstate;
  Dgpsc_cookie	        *scCookie;
  ScaledInt1616           tmp_xres, tmp_yres;
  Dgp_Int32 xres, yres;

  jcookie jc;
  Dgp_CompressedImage *jpeg;
  Dgp_UInt8 *tmpptr;
  Dgp_Int32 width, height, res_factor;

  DGP_PROLOGUE ("Dgp_segmentation()", DGP_PERRSTATE);
 
  /* init scan color pipeline */
  DGP_ERR_PASS_CLEANUP(DgpSC_init(dgp_state, &scCookie));
  
  /* set scan color parameters based on input image resolution */
  DGP_ERR_PASS_CLEANUP(w_pixrGetResolutionDPI(ipstate, read, &tmp_xres, &tmp_yres));
  xres = ScaledInt1616_TO_Int32(tmp_xres);
  yres = ScaledInt1616_TO_Int32(tmp_yres);	
  if (xres == 0)
    xres = DEFAULT_DPI;
  if (yres == 0)
    yres = DEFAULT_DPI;
  DGP_ERR_PASS_CLEANUP(DgpSC_defaultOptions(scCookie, xres, yres));
  scCookie->segOpt.sizeThresh = sizeThresh;
  
  /* process this image */
  DGP_ERR_PASS_CLEANUP(DgpSC_process(scCookie, read));

  /* This call implicitly allocates for text_ccs;  we'll free them later. */
  DGP_ERR_PASS_CLEANUP(DgpSC_getTextComponents(scCookie, text_cc, numccs));
  if(grayImage)
    DGP_ERR_PASS_CLEANUP(DgpSC_getGray(scCookie, grayImage));
  /* create a low res background image */
  DGP_ERR_PASS_CLEANUP(DgpSC_getBackground(scCookie, xres, &background));
  DGP_ERR_PASS_CLEANUP(DgpSC_done(scCookie));

  /* find a possible scaling factor */
  /* TIFF-FX requires the page resolution to be a multiple of the jpeg resolution  */
  res_factor = xres / jpeg_dpi;
  if (res_factor < 1) res_factor = 1;
  while (xres % res_factor != 0) res_factor--;

  /* rescale the background */
  if (res_factor != 1)
    {
  DGP_ERR_PASS_CLEANUP(w_pixrGetWidth (ipstate, background, &width));
  width /= res_factor;
  DGP_ERR_PASS_CLEANUP(w_pixrGetHeight (ipstate, background, &height));
  height /= res_factor;
  DGP_ERR_PASS_CLEANUP(w_scaleToSize (ipstate, background, width, height, &scaled_bg));
  DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipstate, &background));
  background = scaled_bg;
    }

  /* compress background image */
  DGP_CHECK_ALLOC (jpeg = DGP_MALLOC (sizeof(*jpeg)));
  
  DGP_ERR_PASS_CLEANUP(w_pixrGetWidth(ipstate, background, &(jpeg->width)));
  DGP_ERR_PASS_CLEANUP(w_pixrGetHeight(ipstate, background, &(jpeg->height)));
/*  DGP_ERR_PASS_CLEANUP(w_pixrGetResolutionDPI(ipstate, background, &tmp_xres, &tmp_yres)); */
  jpeg->xres = xres / res_factor; /*ScaledInt1616_TO_Int32(tmp_xres); */
  jpeg->yres = yres / res_factor; /*ScaledInt1616_TO_Int32(tmp_yres);	 */
  jpeg->x = 0;
  jpeg->y = 0;
  jpeg->quality = jpeg_qual;

  /* set up compression cookie */
  jc.dgp_state = dgp_state;
  jc.size = 0;
  jc.alloced = jpeg->width * jpeg->height / 10;
  DGP_CHECK_ALLOC (jc.buf = DGP_MALLOC(jc.alloced));

  /* 19.3.2001 / dhtesq. */
  /* Changed colorspace from default to itulab... */
  DGP_ERR_PASS_CLEANUP(
      w_jCompress_callback(DGP_PIPWSTATE, background, jpeg_Compress_callback, 
        (void *)&jc, jpeg_qual, JC_DEFAULT));

  /* free images */
  DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipstate, &background));
  
  if (jc.alloced > jc.size){
    DGP_CHECK_ALLOC(tmpptr = (Dgp_Int8 *)DGP_REALLOC(jc.buf, jc.size));
    jc.buf = tmpptr;
  }

  /* save background */
  jpeg->layer = 1;
  jpeg->data = jc.buf;
  jpeg->data_size = jc.size;
  jpeg->extData_size = 0; 
  
  DGP_ERR_PASS_CLEANUP(
    Dgp_storeJpegImage (cPage, jpeg));

  NORMAL_RETURN;

ERR_CLEANUP:

  if(grayImage){
      if(*grayImage)   DGPIMAGE_FREE(*grayImage);
      *grayImage = NULL;
  }
  ERR_RETURN;
}
