#ifndef __SILX_ENCODE_H__
#define __SILX_ENCODE_H__

#include "dgp_state.h"

#ifdef __cplusplus
extern "C"
{
#endif

typedef struct SilxEncState *SilxEncCookie;

DGP_LIBEXPORT ErrCode
silx_initEncState (
Dgp_State              *dgp_state,
SilxEncCookie          *cookie);

DGP_LIBEXPORT ErrCode
silx_encDocument(
    Dgp_State *dgp_state,
    SilxEncCookie cookie,
    Dgp_Int8  *inFile,
    Dgp_Bool  lastFile);

DGP_LIBEXPORT ErrCode
silx_encStart(
    Dgp_State *dgp_state,
    SilxEncCookie cookie,
    void     *outfp,
    Dgp_Int8 *errfname);

DGP_LIBEXPORT ErrCode
silx_encPage(
             Dgp_State *dgp_state,
             SilxEncCookie cookie,
             Dgp_UInt8     *pgBuffer,
             Dgp_Int32     bufWidth,
             Dgp_Int32     bufHeight,
             Dgp_Int32     bpl,
             Dgp_Int32     bpp,
             Dgp_Int32     xRes,
             Dgp_Int32     yRes,
             Dgp_Bool      lastPage);

DGP_LIBEXPORT ErrCode
silx_cleanEncState (
    Dgp_State *dgp_state,
    SilxEncCookie *sCookie);

DGP_LIBEXPORT void 
silx_setTokenOptions(
    SilxEncCookie cookie,
    Dgp_Int32 *writeThresh,
    Dgp_Float32 *fracHausd,
    Dgp_Int32 *maxDensity,
    Dgp_Int32 *widthDiff,
    Dgp_Int32 *heightDiff,
    Dgp_Int32 *commitThresh,
    Dgp_Int32 *freezeHistThresh,
    Dgp_Int32 *makeHistThresh,
    Dgp_Int32 *returnHistThresh,
    Dgp_Int32 *nearestNeighbors,
    Dgp_Int32 *xShift,
    Dgp_Int32 *SlopeChange,
    Dgp_Float32 *xNewSlope,
    Dgp_Int32 *dilationRadius,
    Dgp_Int32 *lowerDilationLimit,
    Dgp_Int32 *ccSizeThresh);

DGP_LIBEXPORT void
silx_setJpegOptions(
             SilxEncCookie          cookie,
             Dgp_Int32              *jpegQuality,
             Dgp_Int32              *jpegRes,
             Dgp_Float32            *jpegScaleFactor);
DGP_LIBEXPORT void 
silx_setGeneralEncOptions(
             SilxEncCookie    cookie,
             Dgp_Bool  *grayHalftone,
             Dgp_Bool  *deskew,
             Dgp_Int32 *deskewAccuracy,
             Dgp_Int32 *forceRes,
             Dgp_Int8  **baselineOrient,
             Dgp_Int32 *rotation,
             Dgp_Int32 *jbig2Profile,
             Dgp_Bool  *g4Tokens,
             Dgp_Bool  *tokenSizeLimit,
             Dgp_Bool  *verbose,
             Dgp_Int32 *membond,
             Dgp_Bool  *no_err_out,
             Dgp_Bool  *lossless,
             Dgp_Int32 *incrementalDict,
             Dgp_Int32 *clearDict,
             Dgp_Bool  *redTokNoise,
             Dgp_Bool  *jb2Only,
	     Dgp_Bool  *pdfOut);

DGP_LIBEXPORT ErrCode
silx_encMonoBufInPage(
             Dgp_State *dgp_state,
             SilxEncCookie cookie,
             Dgp_UInt8     *buffer,
             Dgp_Int32     bufWidth,
             Dgp_Int32     bufHeight,
             Dgp_Int32     bpl,
             Dgp_Int32     pgHeight,
             Dgp_Int32     bufIndex,
             Dgp_Bool      lastBufInPage);
DGP_LIBEXPORT ErrCode
silx_encMonoBufGetPGData(
     Dgp_State *dgp_state,
     SilxEncCookie cookie,
     Dgp_UInt8**   dictData,
     Dgp_Int32*    dictDataSize,
     Dgp_UInt8**   pgData,
     Dgp_Int32*    pgDataSize);


#ifdef __cplusplus
}
#endif
#endif


