/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libencode/mqcoder.c,v 1.5 2002/03/12 00:23:38 dabercro Exp $
 *
 */

#include "mqcoder.h"
#include "util.h"

#ifndef DGP_QCODER_DBUG
#define DGP_QCODER_DBUG DGP_FALSE
#endif

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libencode/mqcoder.c,v 1.5 2002/03/12 00:23:38 dabercro Exp $");

/*
 * Implemented according to N825
 */
static Dgp_UInt16 Qe[46] = {
    0x5601, 0x3401, 0x1801, 0x0ac1, 0x0521, 0x0221, 0x5601, 0x5401,
    0x4801, 0x3801, 0x3001, 0x2401, 0x1c01, 0x1601, 0x5601, 0x5401,
    0x5101, 0x4801, 0x3801, 0x3401, 0x3001, 0x2801, 0x2401, 0x2201,
    0x1c01, 0x1801, 0x1601, 0x1401, 0x1201, 0x1101, 0x0ac1, 0x09c1,
    0x08a1, 0x0521, 0x0441, 0x02a1, 0x0221, 0x0141, 0x0111, 0x0085,
    0x0049, 0x0025, 0x0015, 0x0009, 0x0005, 0x0001
    };

static Dgp_UInt8 nmps[46] = {
      1,   2,   3,   4,   5,  38,   7,   8,
      9,  10,  11,  12,  13,  29,  15,  16,
     17,  18,  19,  20,  21,  22,  23,  24,
     25,  26,  27,  28,  29,  30,  31,  32,
     33,  34,  35,  36,  37,  38,  39,  40,
     41,  42,  43,  44,  45,  45
    };

/*
 * Least significant 7 bits (well, 6) contain the NLPS value; the
 * MSB contains the "switch" value.
 */
static Dgp_UInt8 nlps[46] = {
    129,   6,   9,  12,  29,  33, 134,  14,
     14,  14,  17,  18,  20,  21, 142,  14,
     15,  16,  17,  18,  19,  19,  20,  21,
     22,  23,  24,  25,  26,  27,  28,  29,
     30,  31,  32,  33,  34,  35,  36,  37,
     38,  39,  40,  41,  42,  43
    };

void
DgpEncode_mqeInit(DgpEncode_MQEncoder *encoder,
	Dgp_Bool (*byte_out)(Dgp_UInt8 b, void *userdata),
	void *userdata, Dgp_State *dgp_state)
{
    encoder->a = 0x8000;
    encoder->c = 0;
    encoder->ct = 12;
    encoder->b = 0;
    encoder->byte_out = byte_out;
    encoder->state = MQE_START;
    encoder->userdata = userdata;
	encoder->dgp_state = dgp_state;
    }

static Dgp_Bool
byteout(DgpEncode_MQEncoder *encoder)
{
  Dgp_State *dgp_state = encoder->dgp_state;
  DGP_PROLOGUE ("byteout", DGP_PERRSTATE);
  DGP_STOP_WHINE;
    
  /* The first byte should get skipped. */
    DGP_REPORTF3(DGP_QCODER_DBUG,"mqe byteout a=%04x c=%08lx b=%02x\n", encoder->a, encoder->c, encoder->b);
    if (encoder->state != MQE_START) {
	if (encoder->b == 0xff) {
	    DGP_REPORTF1(DGP_QCODER_DBUG,"mqe byteout(%02x)\n", encoder->b);
	    if (!(*encoder->byte_out)(encoder->b, encoder->userdata)) {
		return(DGP_FALSE);
		}
	    encoder->b = (Dgp_UInt8) (encoder->c >> 20);
	    encoder->c &= 0xfffff;
	    encoder->ct = 7;
	    DGP_REPORTF3(DGP_QCODER_DBUG,"mqe byteout a=%04x c=%08lx b=%02x\n", encoder->a, encoder->c, encoder->b);
	    return(DGP_TRUE);
	    }
	else if (encoder->c > 0x7ffffff) {
	    encoder->b++;
	    if (encoder->b == 0xff) {
		encoder->c &= 0x7ffffff;
		DGP_REPORTF1(DGP_QCODER_DBUG,"mqe byteout(%02x)\n", encoder->b);
		if (!(*encoder->byte_out)(encoder->b, encoder->userdata)) {
		    return(DGP_FALSE);
		    }
		encoder->b = (Dgp_UInt8) (encoder->c >> 20);
		encoder->c &= 0xfffff;
		encoder->ct = 7;
		DGP_REPORTF3(DGP_QCODER_DBUG,"mqe byteout a=%04x c=%08lx b=%02x\n", encoder->a, encoder->c, encoder->b);
		return(DGP_TRUE);
		}
	    }
	
	DGP_REPORTF1(DGP_QCODER_DBUG,"mqe byteout(%02x)\n", encoder->b);
	if (!(*encoder->byte_out)(encoder->b, encoder->userdata)) {
	    return(DGP_FALSE);
	    }
	}
    else {
	encoder->state = MQE_OK;
	}

    encoder->b = (Dgp_UInt8) (encoder->c >> 19);
    encoder->c &= 0x7ffff;
    encoder->ct = 8;
    DGP_REPORTF3(DGP_QCODER_DBUG,"mqe byteout a=%04x c=%08lx b=%02x\n", encoder->a, encoder->c, encoder->b);
    return(DGP_TRUE);
    }

Dgp_Bool
DgpEncode_mqeEncode(DgpEncode_MQEncoder *encoder, Dgp_UInt8 *context, Dgp_Int32 bit)
{
  Dgp_State *dgp_state = encoder->dgp_state;
  DGP_PROLOGUE ("mqeEncode", DGP_PERRSTATE);
  Dgp_Int32 cx = *context & 0x7f;
  DGP_STOP_WHINE;

    DGP_REPORTF3(DGP_QCODER_DBUG,"mqe encode cx=%02x mps=%d bit=%d\n", cx, (*context & 0x80), bit);
    DGP_REPORTF2(DGP_QCODER_DBUG,"mqe encode a=%04x c=%08lx\n", encoder->a, encoder->c);
    /* Is it an MPS? */
    if (bit == (*context >> 7)) {
	/* Yes */
	encoder->a -= Qe[cx];
	if (encoder->a >= 0x8000) {
	    encoder->c += Qe[cx];
	    DGP_REPORTF2(DGP_QCODER_DBUG,"mqe encode mps fast a=%04x c=%08lx\n", encoder->a, encoder->c);
	    return(DGP_TRUE);
	    }

	if (encoder->a < Qe[cx]) {
	    encoder->a = Qe[cx];
	    DGP_REPORTF2(DGP_QCODER_DBUG,"mqe encode mps->lps exchange a=%04x c=%08lx\n", encoder->a, encoder->c);
	    }
	else {
	    encoder->c += Qe[cx];
	    DGP_REPORTF2(DGP_QCODER_DBUG,"mqe encode mps a=%04x c=%08lx\n", encoder->a, encoder->c);
	    }

	*context = nmps[cx] | (*context & 0x80);
	}
    else {
	/* LPS */
	encoder->a -= Qe[cx];
	if (encoder->a < Qe[cx]) {
	    encoder->c += Qe[cx];
	    DGP_REPORTF2(DGP_QCODER_DBUG,"mqe encode lps->mps exchange a=%04x c=%08lx\n", encoder->a, encoder->c);
	    }
	else {
	    encoder->a = Qe[cx];
	    DGP_REPORTF2(DGP_QCODER_DBUG,"mqe encode lps a=%04x c=%08lx\n", encoder->a, encoder->c);
	    }

	*context = nlps[cx] ^ (*context & 0x80);
	}
    DGP_REPORTF2(DGP_QCODER_DBUG,"mqe encode cx=%02x mps=%d\n", (*context & 0x7f), (*context & 0x80));

    /* Renorm */
    do {
	encoder->a <<= 1;
	encoder->c <<= 1;
	encoder->ct--;
	DGP_REPORTF3(DGP_QCODER_DBUG,"mqe encode renorm a=%04x c=%08lx ct=%02x\n", encoder->a, encoder->c, encoder->ct);
	if (encoder->ct == 0) {
	    if (!byteout(encoder)) {
		return(DGP_FALSE);
		}
	    }
	} while (encoder->a < 0x8000);
    return(DGP_TRUE);
    }

Dgp_Bool
DgpEncode_mqeFlush(DgpEncode_MQEncoder *encoder)
{

	Dgp_State *dgp_state = encoder->dgp_state;
    Dgp_UInt32 t = encoder->c + encoder->a;
    Dgp_Bool wasff;

	DGP_VOID_PROLOGUE ("mqeFlush", DGP_PERRSTATE);

    /*
     * We need to end up with an FFAC.  Make as many bits in the C
     * register 1 as you can, then start outputting bytes.
     * Three possibilities:
     * - the first byte is an FF, and the second is a 7F.  In this case,
     *   overwrite the 7F with a AC.
     * - the first byte isn't an FF and the second byte is an FF.  In
     *   this case, output an additional AC.
     * - neither the first nor second byte is an FF.  In this case, output
     *   an FF AC.
     */
    encoder->c |= 0xffff;
    if (encoder->c >= t) {
	encoder->c -= 0x8000;
	}

    encoder->c <<= encoder->ct;
    if (!byteout(encoder)) {
	return(DGP_FALSE);
	}

    /* What byte was produced by that byteout? */
    if (encoder->b == 0xff) {
	wasff = DGP_TRUE;
	}
    else {
	wasff = DGP_FALSE;
	}

    /*
     * If the second byteout produces a FF, we're done; otherwise,
     * we need a third byteout.
     */
    encoder->c <<= encoder->ct;
    if (!byteout(encoder)) {
	return(DGP_FALSE);
	}
    
    if (encoder->b != 0xff) {
	if (wasff && (encoder->b == 0x7f)) {
	    /* Aha - can clobber the 7f with a ac */
	    if (!(*encoder->byte_out)(0xac, encoder->userdata)) {
		return(DGP_FALSE);
		}
	    return(DGP_TRUE);
	    }

	/* It didn't produce an FF - do another byteout */
	encoder->c <<= encoder->ct;
	if (!byteout(encoder)) {
	    return(DGP_FALSE);
	    }
	}

    /*
     * Now, the byte in encoder->b has officially been "produced" by the
     * encoded, but hasn't actually been stored to memory yet.
     * We know it's an FF.  We also want to write a byte after it that
     * isn't a legal result of encoding (FFxx where xx >= A0 can't be
     * produced by the encoder normally), so output two more bytes:
     * an FF and a AC.
     */
    DGP_VOID_ASSERT(encoder->b == 0xff);

    if (!(*encoder->byte_out)(0xff, encoder->userdata)) {
	return(DGP_FALSE);
	}
    
    if (!(*encoder->byte_out)(0xac, encoder->userdata)) {
	return(DGP_FALSE);
	}

    return(DGP_TRUE);
    }

/*
 * Initialise a decoder, pointing at the initial buffer.  Also,
 * when that buffer is exhausted, call the callback to get more.  It's
 * legal to pass in NULL for data_in.
 *
 * Notes on the states the decoder can be in:
 * - if it's seen a "end" marker (FF xx where xx >= A0) then it goes
 *   into an END state.  When in an end state, it pingpongs between
 *   hallucinating FF and 7F bytes alternately.
 * - Also, if it's tried to call data_in and gotten nothing back, it goes
 *   into a NODATA state, pingponging between FF and 7F.  This
 *   is distinguished from the END state because it's useful to know
 *   later on how we got there.
 * - When the decoder starts up, it reads three bytes.
 */
static void
bytein(DgpEncode_MQDecoder *decoder)
{
  Dgp_State *dgp_state = decoder->dgp_state;
  DGP_PROLOGUE ("bytein", DGP_PERRSTATE);

  Dgp_UInt8 b;

  DGP_STOP_WHINE;
    if ((decoder->state == MQD_END) || (decoder->state == MQD_NODATA)) {
	if (decoder->b == 0xff) {
	    b = 0x7f;
	    }
	else {
	    b = 0xff;
	    }
	}
    else if (decoder->bp == decoder->bp_end) {
	/* Out of data - try for some more */
	Dgp_UInt8 *newbuf;
	size_t newbuflen;

	if (decoder->bp == decoder->bp_end) {
	    if ((decoder->data_in ==
		 (Dgp_Bool (*)(Dgp_UInt8 **, size_t *, void *))NULL) ||
		(!(*decoder->data_in)(&newbuf, &newbuflen, decoder->userdata)) ||
		(newbuflen == 0)) {
		/* No more. */
		decoder->state = MQD_NODATA;
		if (decoder->b == 0xff) {
		    b = 0x7f;
		    }
		else {
		    b = 0xff;
		    }
		}
	    else {
		/* We got something - woo hoo! */
		decoder->bp = newbuf;
		decoder->bp_end = newbuf + newbuflen;
		b = *decoder->bp++;
		}
	    }
	}
    else {
	b = *decoder->bp++;
	}

    DGP_REPORTF3(DGP_QCODER_DBUG,"mqd bytein a=%04x c=%08lx b=%02x ", decoder->a, decoder->c, decoder->b);
	DGP_REPORTF2(DGP_QCODER_DBUG,"newb=%02x state=%d\n", b, decoder->state);
    if (decoder->b == 0xff) {
	/* Check for an END marker */
	if (b >= 0xa0) {
	    /* Yep. */
	    decoder->state = MQD_END;
	    b = 0x7f;
	    }

	decoder->c += b << 9;
	decoder->ct = 7;
	}
    else {
	decoder->c += b << 8;
	decoder->ct = 8;
	}

    decoder->b = b;
    DGP_REPORTF2(DGP_QCODER_DBUG,"mqd bytein a=%04x c=%08lx ", decoder->a, decoder->c);
	DGP_REPORTF2(DGP_QCODER_DBUG,"b=%02x state=%d\n", decoder->b, decoder->state);
    }

void
DgpEncode_mqdInit(DgpEncode_MQDecoder *decoder, Dgp_UInt8 *buffer, size_t buflen,
	Dgp_Bool (*data_in)(Dgp_UInt8 **buffer, size_t *buflen,
			   void *userdata),
	void *userdata, Dgp_State *dgp_state)
{
    Dgp_UInt8 b;
    Dgp_UInt8 *bp_end = buffer + buflen;

    decoder->state = MQD_OK;
    decoder->a = 0x8000;
    decoder->b = 0;
    decoder->data_in = data_in;
    decoder->userdata = userdata;
	decoder->dgp_state = dgp_state;

    if (buflen == 0) {
	/* Sometimes it's easier to just use the callback */
	if ((!(*decoder->data_in)(&buffer, &buflen, decoder->userdata)) ||
	    (buflen == 0)) {
	    /* No data and no callback.  Thanks a lot. */
	    decoder->state = MQD_NODATA;
	    b = 0;
	    bp_end = buffer = (Dgp_UInt8 *)NULL;
	    }
	else {
	    b = *buffer++;
	    }
	}
    else {
	b = *buffer++;
	}

    decoder->bp = buffer;
    decoder->bp_end = bp_end;

    decoder->c = b << 16;

    decoder->b = b;
    bytein(decoder);
    decoder->c <<= 7;

    decoder->ct -= 7;
    }

Dgp_Int32 
DgpEncode_mqdDecode(DgpEncode_MQDecoder *decoder, Dgp_UInt8 *context)
{
    Dgp_UInt8 cx = *context & 0x7f;
    Dgp_UInt8 mps = *context & 0x80;
    Dgp_Int32 d;
    Dgp_State *dgp_state = decoder->dgp_state;
    DGP_PROLOGUE ("mqdDecode", DGP_PERRSTATE);
    DGP_STOP_WHINE;

    DGP_REPORTF2(DGP_QCODER_DBUG,"mqd decode cx=%02x mps=%d\n", cx, mps);
    DGP_REPORTF2(DGP_QCODER_DBUG,"mqd decode a=%04x c=%08lx\n", decoder->a, decoder->c);

    decoder->a -= Qe[cx];
    if ((decoder->c >> 16) >= Qe[cx]) {
	decoder->c -= Qe[cx] << 16;
	if (decoder->a >= 0x8000) {
	    /* MPS, no renorm */
	    DGP_REPORTF3(DGP_QCODER_DBUG,"mqd decode mps fast a=%04x c=%08lx d=%d\n", decoder->a, decoder->c, (mps >> 7));
	    return(mps >> 7);
	    }
	
	/* MPS, renorm happening */
	if (decoder->a < Qe[cx]) {
	    /* It's really an LPS */
	    d = 1 - (mps >> 7);
	    *context = nlps[cx] ^ mps;
	    DGP_REPORTF3(DGP_QCODER_DBUG,"mqd decode mps->lps exchange a=%04x c=%08lx d=%d\n", decoder->a, decoder->c, d);
	    }
	else {
	    d = mps >> 7;
	    *context = nmps[cx] | mps;
	    DGP_REPORTF3(DGP_QCODER_DBUG,"mqd decode mps a=%04x c=%08lx d=%d\n", decoder->a, decoder->c, d);
	    }
	}
    else {
	/* LPS */
	if (decoder->a < Qe[cx]) {
	    /* It's really an MPS */
	    decoder->a = Qe[cx];
	    d = mps >> 7;
	    *context = nmps[cx] | mps;
	    DGP_REPORTF3(DGP_QCODER_DBUG,"mqd decode lps->mps exchange a=%04x c=%08lx d=%d\n", decoder->a, decoder->c, d);
	    }
	else {
	    decoder->a = Qe[cx];
	    d = 1 - (mps >> 7);
	    *context = nlps[cx] ^ mps;
	    DGP_REPORTF3(DGP_QCODER_DBUG,"mqd decode lps a=%04x c=%08lx d=%d\n", decoder->a, decoder->c, d);
	    }
	}

    DGP_REPORTF2(DGP_QCODER_DBUG,"mqd decode cx=%02x mps=%d\n", (*context & 0x7f), (*context & 0x80));
    /* Now renorm */
    do {
	if (decoder->ct == 0) {
	    bytein(decoder);
	    }
	decoder->a <<= 1;
	decoder->c <<= 1;
	decoder->ct--;
	DGP_REPORTF3(DGP_QCODER_DBUG,"mqd decode renorm a=%04x c=%08lx ct=%02x\n", decoder->a, decoder->c, decoder->ct);
	} while (decoder->a < 0x8000);
	    
    return(d);
    }

Dgp_Bool
DgpEncode_mqdEndSeen(DgpEncode_MQDecoder *decoder)
{
    return(decoder->state == MQD_END);
    }

/*
 * Return the pointer to the first unconsumed byte.  The decoder should
 * be sitting staring at the termination marker when this is called.
 *
 * There's a subtlety here: the decoder might not actually be in
 * the MQD_END state.  Instead, it could have read the FF byte, but
 * not gone beyond that and read the AC byte.  I'm not sure how this
 * case is supposed to work, so I just handle the simple cases (in
 * end state, or about to read AC byte); if anything else is
 * going on, it returns NULL.
 */
Dgp_UInt8 *
DgpEncode_mqdFinish(DgpEncode_MQDecoder *decoder)
{
    if (decoder->state == MQD_END) {
	return(decoder->bp);
	}
    else if (decoder->state == MQD_NODATA) {
	/* There really isn't a data pointer here */
	return((Dgp_UInt8 *)NULL);
	}
    else if (decoder->b == 0xFF) {
	/* We're assuming the next byte is an AC */
	/* Call bytein to get it; that forces us into an end state */
	bytein(decoder);

	/* Argh! */
	if (decoder->state != MQD_END) {
	    return((Dgp_UInt8 *)NULL);
	    }

	/* Made it through the minefield */
	return(decoder->bp);
	}
    else {
	/* We're really confused. */
	return((Dgp_UInt8 *)NULL);
	}
    }

/*-----------------------------------------------------------------------*/
/* Much the same stuff, but for the software conventions decoder */


/*
 * Initialise a decoder, pointing at the initial buffer.  Also,
 * when that buffer is exhausted, call the callback to get more.  It's
 * legal to pass in NULL for data_in.
 *
 * Notes on the states the decoder can be in:
 * - if it's seen a "end" marker (FF xx where xx >= A0) then it goes
 *   into an END state.  When in an end state, it never modifies C.
 * - Also, if it's tried to call data_in and gotten nothing back, it goes
 *   into a NODATA state.  This
 *   is distinguished from the END state because it's useful to know
 *   later on how we got there.
 * - When the decoder starts up, it reads three bytes.
 */
static void
byteinSW(DgpEncode_MQDecoder *decoder)
{
    Dgp_UInt8 b;
    Dgp_State *dgp_state = decoder->dgp_state;
    DGP_PROLOGUE ("byteinSW", DGP_PERRSTATE);
    DGP_STOP_WHINE;

    if ((decoder->state == MQD_END) || (decoder->state == MQD_NODATA)) {
	/* Nothing to do - just reset CT and keep going */
	decoder->ct = 8;
	return;
	}
    else if (decoder->bp == decoder->bp_end) {
	/* Out of data - try for some more */
	Dgp_UInt8 *newbuf;
	size_t newbuflen;

	if (decoder->bp == decoder->bp_end) {
	    if ((decoder->data_in ==
		 (Dgp_Bool (*)(Dgp_UInt8 **, size_t *, void *))NULL) ||
		(!(*decoder->data_in)(&newbuf, &newbuflen, decoder->userdata)) ||
		(newbuflen == 0)) {
		/* No more. */
		decoder->state = MQD_NODATA;
		if (decoder->b == 0xff) {
		    b = 0x7f;
		    }
		else {
		    b = 0xff;
		    }
		}
	    else {
		/* We got something - woo hoo! */
		decoder->bp = newbuf;
		decoder->bp_end = newbuf + newbuflen;
		b = *decoder->bp++;
		}
	    }
	}
    else {
	b = *decoder->bp++;
	}

    DGP_REPORTF3(DGP_QCODER_DBUG,"mqdsw bytein a=%04x c=%08lx b=%02x ", decoder->a, decoder->c, decoder->b); 
	DGP_REPORTF2(DGP_QCODER_DBUG,"newb=%02x state=%d\n", b, decoder->state);
    if (decoder->b == 0xff) {
	/* Check for an END marker */
	if (b >= 0xa0) {
	    /* Yep. */
	    decoder->state = MQD_END;
	    b = 0x7f;
	    }

	decoder->c += 0xfe00 - (b << 9);
/*	decoder->c += (b ^ 0x7f) << 9;*/
	decoder->ct = 7;
	}
    else {
	decoder->c += 0xff00 - (b << 8);
	decoder->ct = 8;
	}

    decoder->b = b;
    DGP_REPORTF2(DGP_QCODER_DBUG,"mqdsw bytein a=%04x c=%08lx ", decoder->a, decoder->c);
	DGP_REPORTF2(DGP_QCODER_DBUG,"b=%02x state=%d\n", decoder->b, decoder->state);	
    }

void
DgpEncode_mqdInitSW(DgpEncode_MQDecoder *decoder, Dgp_UInt8 *buffer, size_t buflen,
	  Dgp_Bool (*data_in)(Dgp_UInt8 **buffer, size_t *buflen,
			     void *userdata),
	  void *userdata, Dgp_State *dgp_state)
{
    Dgp_UInt8 b;
    Dgp_UInt8 *bp_end = buffer + buflen;

    decoder->state = MQD_OK;
    decoder->a = 0x8000;
    decoder->b = 0;
    decoder->data_in = data_in;
    decoder->userdata = userdata;
	decoder->dgp_state = dgp_state;

    if (buflen == 0) {
	/* Sometimes it's easier to just use the callback */
	if ((!(*decoder->data_in)(&buffer, &buflen, decoder->userdata)) ||
	    (buflen == 0)) {
	    /* No data and no callback.  Thanks a lot. */
	    decoder->state = MQD_NODATA;
	    b = 0;
	    bp_end = buffer = (Dgp_UInt8 *)NULL;
	    }
	else {
	    b = *buffer++;
	    }
	}
    else {
	b = *buffer++;
	}

    decoder->bp = buffer;
    decoder->bp_end = bp_end;

    decoder->c = (b ^ 0xff) << 16;

    decoder->b = b;
    byteinSW(decoder);
    decoder->c <<= 7;

    decoder->ct -= 7;
    }

Dgp_Int32 
DgpEncode_mqdDecodeSW(DgpEncode_MQDecoder *decoder, Dgp_UInt8 *context)
{
    Dgp_UInt8 cx = *context & 0x7f;
    Dgp_UInt8 mps = *context & 0x80;
    Dgp_Int32 d;
    Dgp_State *dgp_state = decoder->dgp_state;
    DGP_PROLOGUE ("mqdDecodeSW", DGP_PERRSTATE);
    DGP_STOP_WHINE;

    DGP_REPORTF2(DGP_QCODER_DBUG,"mqdsw decode cx=%02x mps=%d\n", cx, mps);
    DGP_REPORTF2(DGP_QCODER_DBUG,"mqdsw decode a=%04x c=%08lx\n", decoder->a, decoder->c);

    decoder->a -= Qe[cx];
    if ((decoder->c >> 16) < decoder->a) {
	if (decoder->a >= 0x8000) {
	    /* MPS, no renorm */
	    DGP_REPORTF3(DGP_QCODER_DBUG,"mqdsw decode mps fast a=%04x c=%08lx d=%d\n", decoder->a, decoder->c, (mps >> 7));
	    return(mps >> 7);
	    }
	
	/* MPS, renorm happening */
	if (decoder->a < Qe[cx]) {
	    /* It's really an LPS */
	    d = 1 - (mps >> 7);
	    *context = nlps[cx] ^ mps;
	    DGP_REPORTF3(DGP_QCODER_DBUG,"mqdsw decode mps->lps exchange a=%04x c=%08lx d=%d\n", decoder->a, decoder->c, d);
	    }
	else {
	    d = mps >> 7;
	    *context = nmps[cx] | mps;
	    DGP_REPORTF3(DGP_QCODER_DBUG,"mqdsw decode mps a=%04x c=%08lx d=%d\n", decoder->a, decoder->c, d);
	    }
	}
    else {
	/* LPS */
	decoder->c -= decoder->a << 16;

	if (decoder->a < Qe[cx]) {
	    /* It's really an MPS */
	    decoder->a = Qe[cx];
	    d = mps >> 7;
	    *context = nmps[cx] | mps;
	    DGP_REPORTF3(DGP_QCODER_DBUG,"mqdsw decode lps->mps exchange a=%04x c=%08lx d=%d\n", decoder->a, decoder->c, d);
	    }
	else {
	    decoder->a = Qe[cx];
	    d = 1 - (mps >> 7);
	    *context = nlps[cx] ^ mps;
	    DGP_REPORTF3(DGP_QCODER_DBUG,"mqdsw decode lps a=%04x c=%08lx d=%d\n", decoder->a, decoder->c, d);
	    }
	}

    DGP_REPORTF2(DGP_QCODER_DBUG,"mqdsw decode cx=%02x mps=%d\n", (*context & 0x7f), (*context & 0x80));
    /* Now renorm */
    do {
	if (decoder->ct == 0) {
	    byteinSW(decoder);
	    }
	decoder->a <<= 1;
	decoder->c <<= 1;
	decoder->ct--;
	DGP_REPORTF3(DGP_QCODER_DBUG,"mqdsw decode renorm a=%04x c=%08lx ct=%02x\n", decoder->a, decoder->c, decoder->ct);
	} while (decoder->a < 0x8000);
	    
    return(d);
    }

Dgp_Bool
DgpEncode_mqdEndSeenSW(DgpEncode_MQDecoder *decoder)
{
    return(decoder->state == MQD_END);
    }

/*
 * Return the pointer to the first unconsumed byte.  The decoder should
 * be sitting staring at the termination marker when this is called.
 *
 * There's a subtlety here: the decoder might not actually be in
 * the MQD_END state.  Instead, it could have read the FF byte, but
 * not gone beyond that and read the AC byte.  I'm not sure how this
 * case is supposed to work, so I just handle the simple cases (in
 * end state, or about to read AC byte); if anything else is
 * going on, it returns NULL.
 */
Dgp_UInt8 *
DgpEncode_mqdFinishSW(DgpEncode_MQDecoder *decoder)
{
    if (decoder->state == MQD_END) {
	return(decoder->bp);
	}
    else if (decoder->state == MQD_NODATA) {
	/* There really isn't a data pointer here */
	return((Dgp_UInt8 *)NULL);
	}
    else if (decoder->b == 0xFF) {
	/* We're assuming the next byte is an AC */
	/* Call byteinSW to get it; that forces us into an end state */
	byteinSW(decoder);

	/* Argh! */
	if (decoder->state != MQD_END) {
	    return((Dgp_UInt8 *)NULL);
	    }

	/* Made it through the minefield */
	return(decoder->bp);
	}
    else {
	/* We're really confused. */
	return((Dgp_UInt8 *)NULL);
	}
    }


