/*
   @(#)$Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_allocators.h,v 1.3 2001/06/08 20:26:43 mkrocken Exp $ 
*/

#ifndef DGP_ALLOCATORS_H
#define DGP_ALLOCATORS_H

#include <stddef.h>
#include "nt_compat.h"
#include "types.pub"
#include "dgp_types.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* these allow us to define our own memory allocators 
   the definitions are the same as the system calls */

typedef void * (TC_CDECL *Dgp_Malloc)(size_t size);
typedef void (TC_CDECL *Dgp_Free)(void *p);
typedef void * (TC_CDECL *Dgp_Calloc)(size_t nobj, size_t size);
typedef void * (TC_CDECL *Dgp_Realloc)(void *p, size_t size);

/* We'll just define a little struct for holding a suite of memory 
   allocators. */ 
typedef struct _Dgp_Allocators{
  Dgp_Malloc malloc;
  Dgp_Free free;
  Dgp_Calloc calloc;
  Dgp_Realloc realloc;
} Dgp_Allocators;

/* these are the small memory allocators */
typedef ErrCode (TC_CDECL *Dgp_MallocSmall)(void *small_cookie, Dgp_Int32 pool_id, void **retptr);
typedef void (TC_CDECL *Dgp_FreeSmall)(void *small_cookie, Dgp_Int32 pool_id, void *ptr);

/* the pool ids for the small memory allocators */
#define DGP_SMALLPOOL_MIN8    0
#define DGP_SMALLPOOL_MIN16   1
#define DGP_SMALLPOOL_MIN20   2
#define DGP_SMALLPOOL_MIN24   3
#define DGP_SMALLPOOL_MIN28   4
#define DGP_SMALLPOOL_MIN32   5
#define DGP_SMALLPOOL_MIN44   6
#define DGP_SMALLPOOL_MIN128  7

/* number of pools currently defined */
#define DGP_NUM_SMALLPOOLS 8

/* use this to make sure changes in struct sizes don't
   mess things up */
#define DGP_SMALLPOOL_VALIDATE(_struct, _pool_id) \
  DGP_ASSERT(sizeof(_struct) <= Dgp_getPoolSize(_pool_id))


/* our struct for holding small mem allocation */
typedef struct _Dgp_SmallAllocators{
  void *smallCookie;
  Dgp_MallocSmall mallocSmall;
  Dgp_FreeSmall freeSmall;
} Dgp_SmallAllocators;

#ifdef __cplusplus
}
#endif

#endif /* DGP_ALLOCATORS_H */
