/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, September 1 1999
**
**  Filename: dgp_err_control.c
**
**  Description of what's implemented in this file:
**    functions for use with error handling
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_err_control.c,v 1.31 2003/01/27 21:48:31 awilliam Exp $
 *
 */

#include "dgp_xif_err_control.h"
#include "dgp_sys_err_control.h"
#include "dgp_w32_err_control.h"
#include "dgp_err_control_priv.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_err_control.c,v 1.31 2003/01/27 21:48:31 awilliam Exp $");

#if defined EQUATOR_CA_ENV || defined PSG /* use PSG IPCORE */
#undef NORMAL_RETURN
#define NORMAL_RETURN do {if (1) return __error__;} while (0)
#endif

/*
 * in order to hide potentially informative messages from the end user, 
 * we have this def wrapped around those strings which could be considered
 * "sensitive info.
 */
#ifdef DGP_DBUG

#define DGP_SECRET_STR(str) str

#else

static const Dgp_Int8 dgp_secret_str[] = "Internal error encountered, bummer";
#define DGP_SECRET_STR(str) dgp_secret_str

#endif

/* From "errgen.pub" */
/*
 * Errors from 1-99 are IPShared "general" errors; they have no prefix
 */
#define NUM_DGP_IPS_ERRORS 38
static const Dgp_Int8 *DGP_IPS_ERRORS[NUM_DGP_IPS_ERRORS] = {
"No error", /* cSuccessful 0 */
"Unable to allocate memory", /*cErrNomem  	1*/
"Invalid parameter", /* cErrInvalidParm 2 */
DGP_SECRET_STR("Too few horizontal lines for skew detection"), /* cErrNoLines 3 */
DGP_SECRET_STR("Couldn't find next point in fgsrot"), /* cErrNoPoint 4 */
DGP_SECRET_STR("Stucture element was too big to generate accelerated code"), /* cErrSETooBig 5 */
"Function is not implemented", /* cErrNoImpl 6 */
"Buffer overflow", /* cErrBufferExceeded 7 */
"Operator does not support images with this number of bits/pixel", /* cErrDepthNotSupported 8 */
"Angle parameter is larger in absolute value than operator can handle", /* cErrAngleTooLarge 9, 
    * usually used in small 
    * angle rotation routines. */
"No on pixels in this source word", /* cErrNoOnPixel 10 */
"Two objects should have been the same length", /* cErrNotSameLength 11, e.g. TRCs */
"Internal error", /* cErrInternal 12, e.g. something
    * unexpected happend. please report
    * as bug to ipcore! */
"Client's callback routine returned error", /* cErrCallbackError 13 */
"Yield procedure signaled abort", /* cErrAborted 14 */
"Error during decode", /* cErrDecodeFailure 15 */
"Image is too wide", /*cErrImageTooWide 16, (to generate a
    * model set for Hausdorff routines)
    */
"Out of bounds error", /* ia_outOfBounds 17 */
"Error during read", /* ia_readError 18  */
"Error during write", /* ia_writeError 19 */
"Reading/scanning early end found", /* ia_earlyEndEncountered 20  */
"No image data found", /* ia_noImageData 21  */
"Illegal type", /* ia_badType 22 */
"Error during seek", /* ia_seekFailure 23 */
"Illegal version", /* ia_badVersion 24 */
"Error during de/compression", /* ia_compressionError 25 */
"Bad file length", /* ia_badFileLength 26  */
"Couldn't open data pipe", /* ia_createPipeFail 27 */
"Bytes per line not a multiple of 4", /* ia_bplNotMult4 28 */
"Multiple image planes not supported", /* ia_planesNotSupported 29 */
"Type not supported", /* ia_typeNotSupported 30, The combination of depth
 * (bits per sample and samples per
 * pixel), planes, color type, etc
 * is not supported 
 */
"Desired number of tiles not supported", /* ia_nTilesNotSupported 31 */
"Desired de/compression type not supported", /* ia_compressionNotSupported 32 */
"Desired frame size not supported", /* ia_frameNotSupported 33 */
"Desired number of channels not supported", /* ia_channelsNotSupported 34 */
"Dimensions required to be equal", /* ia_dimensNotEqual 35 */
"Depths required to be equal", /* ia_depthsNotEqual 36  */
"Channels required to be equal" /* ia_channelsNotEqual 37 */
};


/* From "ipserrs.pub" */
/* 
 * Errors from 100-999 are IPCore "general" errors; they have no prefix
 */
#define NUM_DGP_IPC_ERRORS 23
static const Dgp_Int8 *DGP_IPC_ERRORS[NUM_DGP_IPC_ERRORS] = {
"Out of bounds error", /* cErrOutOfBounds cGen+0 */
"No image data found", /* cErrNoImageData cGen+1  */
"Illegal type", /* cErrBadType cGen+2 */
"Illegal version", /* cErrBadVersion cGen+3 */
"Bytes per line not a multiple of 4", /* cErrBplNotMult4 cGen+4 */
"Multiple image planes not supported", /* cErrPlanesNotSupported cGen+5 */
"Type not supported", /* cErrTypeNotSupported cGen+6, The combination of depth
 * (bits per sample and samples per
 * pixel), planes, color type, etc
 * is not supported 
 */
"Error during de/compression", /* cErrCompressionError cGen+7 */
"Desired de/compression type not supported", /* cErrCompressionNotSupported cGen+8 */
"Desired frame size not supported", /* cErrFrameNotSupported CGen+9 */
"Desired number of channels not supported", /* cErrChannelsNotSupported cGen+10 */
"Channels required to be equal", /* cErrChannelsNotEqual cGen+11 */
"Image Dimensions are incorrect", /* cErrDestDimensIncorrect cGen+12 */
"Dimensions required to be equal", /* cErrSrcDimensNotEqual cGen+13 */
"Regression test failed", /* cErrTestFailure cGen+14  */
"Illegal reduction factor", /* cErrBadReduction cGen+15 */
"Image dimensions not correct ratio", /* cErrBadRatio cGen+16 */
"Desired expansion not supported", /* cErrExpansionNotSupported cGen+17 */
"Unexpected NULL pointer", /* cErrNullPtr cGen+18 */
"Fixed-length array is full", /* cErrArrayFull cGen+19 */
"Expected NULL pointer", /* cErrPtrNotNull cGen+20 */
"Pointer to pIPWState was NULL", /* cErrNullIPWState cGen+21 */
/* ERROR in "ipserrs.pub" ???, line 114 */
"Pointer to pIPState was NULL" /* cErrNullIPState cGen+22 */
};


/* From "ipwerrs.pub" */
/* Errors from 1000-1999 are PIXR format errors */
#define NUM_DGP_IPW_PIXR_ERRORS 7
static const Dgp_Int8 *DGP_IPW_PIXR_ERRORS[NUM_DGP_IPW_PIXR_ERRORS] = {
"PIXR: Incorrect byte order", /* cErrPixr_byteOrder cPixr+0 */
"PIXR: Depth not supported", /* cErrPixr_depth cPixr+1 */
"PIXR: Wrong number of channels", /* cErrPixr_channels cPixr+2 */
"PIXR: NULL PIXR", /* cErrPixr_null cPixr+3 */
"PIXR: Framed image required", /* cErrPixr_frame cPixr+4d */
"PIXR: Resolution not set", /* cErrPixr_resolution cPixr+5 */
"PIXR: Invalid colormap" /* cErrPixr_colormap cPixr+6 */
};

/* Errors from 2000-2999 are "IO" errors */
#define NUM_DGP_IPW_IO_ERRORS 9
static const Dgp_Int8 *DGP_IPW_IO_ERRORS[NUM_DGP_IPW_IO_ERRORS] = {
"Error during read", /* cErrIO_ReadError cIO+0 */
"Error during write", /* cErrIO_WriteError cIO+1 */
"Error during seek", /* cErrIO_SeekFailure cIO+2 */
"Bad file length", /* cErrIO_BadFileLength cIO+3 */
"Couldn't open data pipe", /* cErrIO_CreatePipeFail cIO+4*/
"Reading/scanning early end found", /* cErrIO_EarlyEndEncountered cIO+5*/
"Desired number of tiles not supported", /* cErrIO_NTilesNotSupported cIO+6*/
"File is not CCITT Group 4", /* cErrIO_NotGroup4 cIO+7 */
"TIFF strip not filled" /* cErrIO_StripNotFilled cIO+8 */
};

/* Errors from 3000-3999 are DataGlyph errors */
#define NUM_DGP_IPW_DGLYPH_ERRORS 36
static const Dgp_Int8 *DGP_IPW_DGLYPH_ERRORS[NUM_DGP_IPW_DGLYPH_ERRORS] = {
"DGlyph: Invalid error code",
"DGlyph: Unable to find center of glyph", /* cErrDG_CenterNotValid cDataGlyph+1 */
"DGlyph: Unable to find direction of glyph", /* cErrDG_DirNotValid	cDataGlyph+2 */
"DGlyph: Not enough glyphs horizontally", /* cErrDG_NoHorizontalRuns cDataGlyph+3 */
"DGlyph: Not enough glyphs vertically", /* cErrDG_NoVerticalRuns cDataGlyph+4 */
"DGlyph: Expected glyph not found", /* cErrDG_NoGlyph cDataGlyph+5 */
"DGlyph: Expected glyph rectangle not found", /* cErrDG_NoRect cDataGlyph+6 */
"DGlyph: Unable to read data file", /* cErrDG_FileMissing cDataGlyph+7 */
"DGlyph: Data file corrupted", /* cErrDG_FileCorrupted cDataGlyph+8 */
"DGlyph: Insufficient synchronization bits found", /* cErrDG_NoSynch cDataGlyph+9 */
"DGlyph: Rotation angle too large", /* cErrDG_AngleTooLarge cDataGlyph+10 */
"DGlyph: CRC incorrect", /* cErrDG_CRCFailure cDataGlyph+11 */
"DGlyph: Decode failed", /* cErrDG_DecodeFailure cDataGlyph+12 */
"DGlyph: Unable to find vertical synch", /* cErrDG_VertSynchNotFound cDataGlyph+13 */
"DGlyph: Containing rectangle too small", /* cErrDG_RectTooSmall cDataGlyph+14 */
"DGlyph: Image too small", /* cErrDG_ImageTooSmall cDataGlyph+15 */
"DGlyph: Image rotated 90 degrees", /* cErrDG_ImageRotated90 cDataGlyph+16 */
"DGlyph: Image rotated 180 degrees", /* cErrDG_ImageRotated180 cDataGlyph+17 */
"DGlyph: Image rotated 270 degrees", /* cErrDG_ImageRotated270 cDataGlyph+18 */
"DGlyph: Rectangle supplied not found", /* cErrDG_IllegalRect cDataGlyph+19 */
"DGlyph: Invalid layout handle", /* cErrDG_InvalidLayout cDataGlyph+20 */
"DGlyph: Invalid region handle", /* cErrDG_InvalidRegion cDataGlyph+21 */
"DGlyph: invalid ECC level", /* cErrDG_InvalidECCLevel cDataGlyph+22 */
"DGlyph: Occlusion already specified", /* cErrDG_OcclusionAlreadySet cDataGlyph+23 */
"DGlyph: Occlusion size too small", /* cErrDG_OcclusionTooSmall cDataGlyph+24 */
"DGlyph: Occlusion spans rectangle", /* cErrDG_OcclusionSpansRect cDataGlyph+25 */
"DGlyph: Required rectangle larger than given dimensions", /* cErrDG_LayoutOverflow cDataGlyph+26 */
"DGlyph: Desired alignment not supported", /* cErrDG_NotSupported cDataGlyph+27 */
"DGlyph: Required rectangle too large", /* cErrDG_RectOverflow cDataGlyph+28 */
"DGlyph: Client allocation region handle too small", /* cErrDG_AllocationTooSmall cDataGlyph+29 */
"DGlyph: Invalid byte count", /* cErrDG_InvalidByteCount cDataGlyph+30 */
"DGlyph: Invalid glyph size", /* cErrDG_InvalidGlyphSize cDataGlyph+31 */
"DGlyph: Too many occlusions" /* cErrDG_TooManyOcclusions cDataGlyph+32 */
"DGlyph: Occlusion not found", /* cErrDG_OcclusionNotFound cDataGlyph+33 */
"DGlyph: Multiple non-contiguous regions in glyph", /* cErrDG_MulipleLiveGlyphRegions cDataGlyph+34 */
"DGlyph: Illegal glyph block dimensions" /* cErrDG_BlockDimsIllegal cDataGlyph+35 */
};

/* Errors from 4000-4999 are Strel errors */
#define NUM_DGP_IPW_STREL_ERRORS 5
static const Dgp_Int8 *DGP_IPW_STREL_ERRORS[NUM_DGP_IPW_STREL_ERRORS] = {
DGP_SECRET_STR("Strel: Invalid error code"),
DGP_SECRET_STR("Strel: Mgr not initialized"), /* cErrStrel_MgrNotInitted cStrel+1 */
DGP_SECRET_STR("Strel: Not 2D seperable"), /* cErrStrel_Not2DSeparable cStrel+2 */
DGP_SECRET_STR("Strel: Invalid ID"), /* cErrStrel_InvalidID cStrel+3 */
DGP_SECRET_STR("Strel: ID not initialized") /* cErrStrel_IDNotInitted cStrel+4 */
};

/* Errors from 5000-5999 are "Heur" errors, i.e. heuristic failures, etc. */
#define NUM_DGP_IPW_HEUR_ERRORS 2
static const Dgp_Int8 *DGP_IPW_HEUR_ERRORS[NUM_DGP_IPW_HEUR_ERRORS] = {
"Nothing found", /* cErrHeuristic_foundNothing cHeuristic+0, Didn't find what we were
 looking for, e.g. image to
 crop or find halftones on */
"Not enough points" /* cErrHeuristic_notEnoughPoints cHeuristic+1 */
};

/* Errors from 6000-6999 are "adt" errors, i.e. abstract data type errors */
#define NUM_DGP_IPW_ADT_ERRORS 1
static const Dgp_Int8 *DGP_IPW_ADT_ERRORS[NUM_DGP_IPW_ADT_ERRORS] = {
"Invalid rectangle count" /* cErrADT_invalidRectcount cADT+0 */
};

/* Errors from 7000-7999 are "apps" errors */
#define NUM_DGP_IPW_APPS_ERRORS 7
static const Dgp_Int8 *DGP_IPW_APPS_ERRORS[NUM_DGP_IPW_APPS_ERRORS] = {
"Invalid error code",
"Bad reduction factor", /* cErrApps_redFactor cApps+1 */
"Bad reduction", /* cErrApps_badReduction cApps+2 */
"Invalid resolution", /* cErrApps_invalidResolution cApps+3 */
"Reduction not supported", /* cErrApps_cannotReduce cApps+4 */
"Wrong dimensions specified", /* cErrApps_differentDimensions cApps+5 */
"Too much data" /* cErrApps_tooMuchData cApps+6 */
};


/* 
 * Errors from 10,000 - 20,000 are reserved for digipaper
 */
#define NUM_DGP_ERRORS 17

static const Dgp_Int8 *DGP_ERRORS[NUM_DGP_ERRORS] = {
"File not found", /* cFileNotFound DGP_BASE_GEN_ERR + 0 */
"Internal buffer overflow", /* cIntBufOverflow DGP_BASE_GEN_ERR + 1 */
"Error within Ipcore", /* cInternalIpcoreErr DGP_BASE_GEN_ERR + 2 */
"Dgp assert failure", /* cErrAssertFailure DGP_BASE_GEN_ERR + 3*/
"Invalid command line", /* cErrInvalidCmdLine DGP_BASE_GEN_ERR + 4 */
"Insufficient Memory", /* cErrInsufficientMem DGP_BASE_GEN_ERR + 5 */
"Burma! Internal consistency check failure", /* cErrPanic DGP_BASE_GEN_ERR + 6*/

/*	Two exceptions for stack type in libadt.  Added 7.2.00 / dhtesq. */
"ADT:  Null stack requested",										/* cADTNullStackRequested;  cDgpGen + 7 */
"ADT:  Smaller stack requested",									/* cADTSmallerStackRequested;  cDgpGen + 8 */

/*  Three procedural exceptions for libscancolor.  Added 7.2.2000 / dhtesq. */
"ScanColor:  x and y resolutions differ;  can't scale background.",	/* cSCCantScaleBackground;  cDgpGen + 9 */
"ScanColor:  Call outside of initialization stage.",				/* cSCNotInStageInit;  cDgpGen + 10 */
"ScanColor:  Call outside of find text stage.",						/* cSCNotInStageFindText;  cDgpGen + 11 */

/* Exception for libjpeg. Added 3.30.2000 , ranjani */
"Jpeg6b : Corrupt input data. ",  /* cJpegInternalPanicErr: DGP_BASE_GEN_ERR + 12*/
"Jpeg6b : Out of Memory error. ", /* cJpegOutOfMemoryErr : DGP_BASE_GENERR + 13 */
"Jpeg6b : Not a JPEG file at all.",/* cJpegNotAJpegFile : DGP_BASE_GEN + 14 */
"Jpeg6b : File IO Error. ",         /* cJpegIOErr : DGP_BASE_GEN + 15 */
"TIFF-FX dictionary limit of 5 has been reached" /*cTFXDictionaryLimitErr*/

/* To add an error string here, see below... */
};

/* 
 * ATTENTION: Please read the section on New Error Codes in "Error Handling.htm" 
 *  before adding your own, thanks. 
 */


/* constants to map xiff errors to contiguous numbers */
#define DGP_XIF_NOERROR		0
#define DGP_XIF_NOMEMORY		1
#define DGP_XIF_BADPARAMETER	2
#define DGP_XIF_NOSUPPORT		3
#define DGP_XIF_BUFFEREXCEEDED	4
#define DGP_XIF_INTERNAL_ERROR	5
#define DGP_XIF_CLIENTABORT		6
#define DGP_XIF_TIFFEMPTY		7
#define DGP_XIF_NOTTIFF		8
#define DGP_XIF_NOTXIFF		9
#define DGP_XIF_SEQUENCE		10
#define DGP_XIF_BADFILEMODE		11
#define DGP_XIF_BADFILEFORMAT	12
#define DGP_XIF_READ_ERR		13
#define DGP_XIF_WRITE_ERR		14
#define DGP_XIF_SEEK_ERR		15
#define DGP_XIF_FILELEN_ERR		16

static const Dgp_Int8 *DGP_XIF_ERRORS[] = {
"Xif: No error", /* XIF_NOERROR 0 */
"Xif: Unable to allocate memory", /* XIF_NOMEMORY 1 */
"Xif: Invalid parameter", /* XIF_BADPARAMETER 2 */
"Xif: Unsupported feature", /* XIF_NOSUPPORT 6 */
"Xif: Buffer overflow", /* XIF_BUFFEREXCEEDED 7 */
"Xif: Internal Error", /* XIF_INTERNAL_ERROR 12 */
"Xif: Client requested abort", /* XIF_CLIENTABORT 14 */
"Xif: 0-page TIFF", /* XIF_TIFFEMPTY 18, (okay for XIFF) */
"Xif: File does not comform to TIFF", /* XIF_NOTTIFF 19 */
"Xif: File does not comform to XIFF", /* XIF_NOTXIFF 20 */
"Xif: Function called out-of-sequence", /* XIF_SEQUENCE 21 */
"Xif: File not opened in BINARY/READ/WRITE", /* XIF_BADFILEMODE 22 */
"Xif: File format/data is corrupted", /* XIF_BADFILEFORMAT 23 */
"Xif: Error reading", /* XIF_READ_ERR 2000 */
"Xif: Error writing", /* XIF_WRITE_ERR 2001 */
"Xif: Error seeking", /* XIF_SEEK_ERR 2002 */
"Xif: Error seeking to EOF" /* XIF_FILELEN_ERR 2003 */
};


/* return the error string for a given digipaper, ipcore, or xif error code */
const Dgp_Int8 *
Dgp_getErrStr(ErrCode error)
{
  Dgp_Int32 error_index;

  if(error < 0) {
    /* cErrUndefined -1, Undefined error; shouldn't happen */
    if(error == -1) return ("Undefined error");
  } else if(error < 100) {

    /* 1-99 are IPShared "general" errors */
    if(error < NUM_DGP_IPS_ERRORS) return DGP_IPS_ERRORS[error];

  } else if(error < 1000) {

    /* 100-999 are IPCore "general" errors */
    error_index = error - cGen;
    if(error_index < NUM_DGP_IPC_ERRORS) return DGP_IPC_ERRORS[error_index];

  } else if(error < 2000) {

    /* 1000-1999 are PIXR format errors */
    error_index = error - cPixr;
    if(error_index < NUM_DGP_IPW_PIXR_ERRORS) return DGP_IPW_PIXR_ERRORS[error_index];

  } else if(error < 3000) {

    /* 2000-2999 are "IO" errors */
    error_index = error - cIO;
    if(error_index < NUM_DGP_IPW_IO_ERRORS) return DGP_IPW_IO_ERRORS[error_index];

  } else if(error < 4000) {

    /* 3000-3999 are DataGlyph errors */
    error_index = error - cDataGlyph;
    if(error_index < NUM_DGP_IPW_DGLYPH_ERRORS) return DGP_IPW_DGLYPH_ERRORS[error_index];

  } else if(error < 5000) {

    /* 4000-4999 are Strel errors */
    error_index = error - cStrel;
    if(error_index < NUM_DGP_IPW_STREL_ERRORS) return DGP_IPW_STREL_ERRORS[error_index];

  } else if(error < 6000) {

    /* 5000-5999 are "Heur" errors, i.e. heuristic failures, etc. */
    error_index = error - cHeuristic;
    if(error_index < NUM_DGP_IPW_HEUR_ERRORS) return DGP_IPW_HEUR_ERRORS[error_index];

  } else if(error < 7000) {

    /* 6000-6999 are "adt" errors, i.e. abstract data type errors */
    error_index = error - cADT;
    if(error_index < NUM_DGP_IPW_ADT_ERRORS) return DGP_IPW_ADT_ERRORS[error_index];

  } else if(error < 8000) {

    /* 7000-7999 are "apps" errors */
#if defined (EQUATOR_CA_ENV) || defined (PSG)
    error_index = error - cApps;
#else
    error_index = error - cErrApps;
#endif
    if(error_index < NUM_DGP_IPW_APPS_ERRORS) return DGP_IPW_APPS_ERRORS[error_index];

  } else if(error <= DGP_MAX_GEN_ERR) {

    /* 10000-10999 are Digipaper errors */
    error_index = error - DGP_BASE_GEN_ERR;
    if(error_index < NUM_DGP_ERRORS) return DGP_ERRORS[error_index];

  } else if(error <= maxDgpXif) {

    /* 11000-11999 are Digipaper wrapped xif errors */
    error_index = error - cDgpXif;
    switch(error_index) {
    case XIF_NOERROR:
      return DGP_XIF_ERRORS[DGP_XIF_NOERROR];
    case XIF_NOMEMORY:
      return DGP_XIF_ERRORS[DGP_XIF_NOMEMORY];
    case XIF_BADPARAMETER:
      return DGP_XIF_ERRORS[DGP_XIF_BADPARAMETER];
    case XIF_NOSUPPORT:
      return DGP_XIF_ERRORS[DGP_XIF_NOSUPPORT];
    case XIF_BUFFEREXCEEDED:
      return DGP_XIF_ERRORS[DGP_XIF_BUFFEREXCEEDED];
    case XIF_INTERNAL_ERROR:
      return DGP_XIF_ERRORS[DGP_XIF_INTERNAL_ERROR];
    case XIF_CLIENTABORT:
      return DGP_XIF_ERRORS[DGP_XIF_CLIENTABORT];
    case XIF_TIFFEMPTY:
      return DGP_XIF_ERRORS[DGP_XIF_TIFFEMPTY];
    case XIF_NOTTIFF:
      return DGP_XIF_ERRORS[DGP_XIF_NOTTIFF];
    case XIF_NOTXIFF:
      return DGP_XIF_ERRORS[DGP_XIF_NOTXIFF];
    case XIF_SEQUENCE:
      return DGP_XIF_ERRORS[DGP_XIF_SEQUENCE];
    case XIF_BADFILEMODE:
      return DGP_XIF_ERRORS[DGP_XIF_BADFILEMODE];
    case XIF_BADFILEFORMAT:
      return DGP_XIF_ERRORS[DGP_XIF_BADFILEFORMAT];
    case XIF_READ_ERR:
      return DGP_XIF_ERRORS[DGP_XIF_READ_ERR];
    case XIF_WRITE_ERR:
      return DGP_XIF_ERRORS[DGP_XIF_WRITE_ERR];
    case XIF_SEEK_ERR:
      return DGP_XIF_ERRORS[DGP_XIF_SEEK_ERR];
    case XIF_FILELEN_ERR:
      return DGP_XIF_ERRORS[DGP_XIF_FILELEN_ERR];
    }

  } else if(error <= maxDgpSys) {

    /* 12000-12999 are Digipaper wrapped system errors */
    error_index = error - cDgpSys;

    return strerror(error_index);

  } else if(error <= maxDgpW32) {

    /* 12000-12999 are Digipaper wrapped win32 errors */
    return "Win32 error";

  }

  return "Invalid error code";
}


/* Function to create an error state. */
ErrCode
Dgp_initDgpErrState(Dgp_ErrState **_dgp_err, Dgp_Allocators *a, 
                void *userErrCookie, Dgp_ErrPrinterType errPrinter)
{
  Dgp_ErrState *dgp_err = NULL;

#ifndef EQUATOR_CA_ENV
  DGP_MAIN_PROLOGUE("Dgp_InitDgpErrState", userErrCookie, errPrinter);
#else
  Dgp_ErrState *__PERR_STATE__, temp;
  PROLOGUE;
  Dgp_UInt8* __FUNCT_NAME__ = "Dgp_InitDgpErrState";
  temp.userCookie = userErrCookie;
  temp.errPrinter = errPrinter;
  __PERR_STATE__ = &temp;
#endif

  DGP_CHECK_ALLOC_CLEANUP(dgp_err = (Dgp_ErrState *)(a->malloc)(sizeof(*dgp_err)));

  dgp_err->userCookie = userErrCookie;
  dgp_err->errPrinter = errPrinter;

  *_dgp_err = dgp_err;
  dgp_err = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  Dgp_cleanupDgpErrState(&dgp_err, a);

  ERR_RETURN;
}

void
Dgp_voidAssertPrinter(const Dgp_Int8 *file_name, const Dgp_Int8 *funct_name,
                      Dgp_Int32 line_number, const Dgp_Int8 *cond)
{
  fprintf(stderr, "Assert Failure, \"%s\", line %d, %s -- (%s)\n",
        file_name, (int)line_number, funct_name, cond);
  exit(EXIT_FAILURE);
}

/* Function to cleanup an error state */
void
Dgp_cleanupDgpErrState(Dgp_ErrState **_dgp_err, Dgp_Allocators *a)
{
  Dgp_ErrState *dgp_err = NULL;

  if(_dgp_err == NULL) return;

  dgp_err = *_dgp_err;
  *_dgp_err = NULL;

  if(dgp_err == NULL) return;

  if(dgp_err != NULL) (a->free)(dgp_err);
}
