/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, February 1 1998
**
**  Filename: dgp_err_printersl.h
**
**  Description of what's implemented in this file:
**    default printers to do error output
**  
**************************************************************************/

/*
 * RCS Header
 *
 * $Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_err_printers.h,v 1.14 2003/01/31 19:15:43 mkrocken Exp $
 */

#ifndef __DGP_ERR_PRINTERS_H
#define __DGP_ERR_PRINTERS_H

#include "types.pub"

/*  VxWorks needs explicit architecture definition for stdarg.h to work. / dhtesq. */
#ifdef VXWORKS
#include <vxWorks.h>
#endif  /* (VXWORKS) */

#include <stdarg.h>
#include "dgp_allocators.h"

#ifdef __cplusplus
extern "C"
{
#endif

/*
 * Default functions to send error output to a file.  The
 * error cookie must be a "FILE *" (it may be NULL, in which case no output
 * is written.
 */
#ifdef DGP_DBUG

/* default function to output debug errors to a file */
DGP_LIBEXPORT void 
Dgp_fileErrPrinter(void *cookie, const Dgp_Int8 *file_name, const Dgp_Int8 *funct_name,
                          Dgp_Int32 line_number, ErrCode error, const Dgp_Int8 *format, ...);

#else /* DGP_DBUG is not defined */

/* default function to output nondebug errors to a file */
DGP_LIBEXPORT void 
Dgp_fileErrPrinter(void *cookie, ErrCode error, const Dgp_Int8 *format, ...);

#endif /* DGP_DBUG */


/*
 * Default functions to send error output to a string.  The
 * error cookie must be a "Dgp_Int8 *" of sufficient size (it may be NULL, in which
 * case no output is written.
 */
#ifdef DGP_DBUG

/* default function to output debug errors to a file */
DGP_LIBEXPORT void 
Dgp_strErrPrinter(void *cookie, const Dgp_Int8 *file_name, const Dgp_Int8 *funct_name,
                          Dgp_Int32 line_number, ErrCode error, const Dgp_Int8 *format, ...);

#else /* DGP_DBUG is not defined */

/* default function to output nondebug errors to a file */
DGP_LIBEXPORT void 
Dgp_strErrPrinter(void *cookie, ErrCode error, const Dgp_Int8 *format, ...);

#endif /* DGP_DBUG */


/*
 * Default functions to send error output to a "safe" string.  The
 * error cookie must be a "Dgp_Int8 *" of sufficient size (it may be NULL, in which
 * case no output is written.
 */
typedef struct {
  Dgp_Int32 length;
  Dgp_Int8 str[1];
} Dgp_SafeErrString;

/* cheap function to create a "safe" error string of a given string length */

#ifdef DGP_DBUG

/* default function to output debug errors to a file */
DGP_LIBEXPORT void 
Dgp_safeStrErrPrinter(void *cookie, const Dgp_Int8 *file_name, const Dgp_Int8 *funct_name,
                          Dgp_Int32 line_number, ErrCode error, const Dgp_Int8 *format, ...);

#else /* DGP_DBUG is not defined */

/* default function to output nondebug errors to a file */
DGP_LIBEXPORT void 
Dgp_safeStrErrPrinter(void *cookie, ErrCode error, const Dgp_Int8 *format, ...);

#endif /* DGP_DBUG */


#ifdef __cplusplus
}
#endif

#endif /* __DGP_ERR_PRINTERS_H */







