/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, September 1 1999
**
**  Filename: dgp_file_err.c
**
**  Description of what's implemented in this file:
**    functions for default error ouput to a file
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_file_err.c,v 1.11 2001/12/19 22:30:39 jlin Exp $
 *
 */


#include "dgp_err_control.h"
#include "dgp_err_printers.h"
#include "stdio.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_file_err.c,v 1.11 2001/12/19 22:30:39 jlin Exp $");

#ifdef DGP_DBUG

/* default function to output debug errors to a file */
DGP_LIBEXPORT void 
Dgp_fileErrPrinter(void *cookie, const Dgp_Int8 *file_name, const Dgp_Int8 *funct_name,
                  Dgp_Int32 line_number, ErrCode error, const Dgp_Int8 *format, ...)
{
  va_list(arglist);

  if(cookie) {
    va_start(arglist, format);
    if(error == DGP_REPORT) {
      fprintf((FILE *)cookie, "Report -- ");
    } else if(error == DGP_WARNING) {
      fprintf((FILE *)cookie, "Warning -- ");
    } else {
      fprintf((FILE *)cookie, "Error %d, \"%s\", line %d, %s -- ", (int)error, file_name, (int)line_number, funct_name);
    }
    if(format != NULL) vfprintf((FILE *)cookie, format, arglist);
    fprintf((FILE *)cookie, "\n");
  }
}

#else /* DGP_DBUG is not defined */

/* default function to output nondebug errors to a file */
DGP_LIBEXPORT void 
Dgp_fileErrPrinter(void *cookie, ErrCode error, const Dgp_Int8 *format, ...)
{
  va_list(arglist);

  if(cookie) {
    va_start(arglist, format);
    if(error == DGP_REPORT) {
      fprintf((FILE *)cookie, "Report -- ");
    } else if(error == DGP_WARNING) {
      fprintf((FILE *)cookie, "Warning -- ");
    } else {
      fprintf((FILE *)cookie, "Error %d -- ", (int)error);
    }
    if(format != NULL) vfprintf((FILE *)cookie, format, arglist);
    fprintf((FILE *)cookie, "\n");
  }
}

#endif
