/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, September 1 1999
**
**  Filename: dgp_safe_str_err.c
**
**  Description of what's implemented in this file:
**    functions for default error ouput to a "safe" string
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_safe_str_err.c,v 1.16 2003/02/10 22:53:24 mkrocken Exp $
 *
 */

#include "dgp_err_control.h"
#include "dgp_err_printers.h"
#ifdef __sparc__
/* 30.10.2000 / dhtesq. */
/* Our parc gcc 2.7.2 installation evidently has no prototypes for these... */
int snprintf(char *s, size_t n, const char *format,...);
int vsnprintf(char *s, size_t n, const char *format, va_list ap);
#endif 
#include <stdio.h>
#include <stdarg.h>
#include "dgp_allocators.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_safe_str_err.c,v 1.16 2003/02/10 22:53:24 mkrocken Exp $");

#ifdef _WIN32
#define snprintf _snprintf
#define vsnprintf _vsnprintf
#endif

/* Have to dig for snprintf, vsnprintf on VxWorks... */
/* 14.3.2000 / dhtesq. */
#ifdef VXWORKS
#include "private/stdioP.h"
#endif /* (VXWORKS) */

/* 30.10.2000 / dhtesq. */
/* Moved these statics from dgp_err_printers.h...ifdeffed out since at present */
/* they are doing nothing for the code base except issuing compiler warnings. */
#if 0
/* cheap function to create a "safe" error string of a given string length */
static INLINE Dgp_SafeErrString *
Dgp_createSafeStr(Dgp_Malloc malloc_func, Dgp_Int32 str_len)
{
  Dgp_SafeErrString *tmp_str = NULL;

  if((tmp_str = (Dgp_SafeErrString *)
      (malloc_func)(sizeof(Dgp_SafeErrString) + (str_len * sizeof(Dgp_Int8))))
     != NULL) {
    tmp_str->length = str_len;
  }
  
  return tmp_str;
}

/* cheap function to free a "safe" error string */
static INLINE void
Dgp_freeSafeStr(Dgp_Free free_func, Dgp_SafeErrString *str)
{
  free_func(str);
}
#endif

#ifdef DGP_DBUG

/* default function to output debug errors to a string */
void 
Dgp_safeStrErrPrinter(void *_cookie, const Dgp_Int8 *file_name, const Dgp_Int8 *funct_name,
                 Dgp_Int32 line_number, ErrCode error, const Dgp_Int8 *format, ...)
{
  Dgp_SafeErrString *cookie = (Dgp_SafeErrString *)_cookie;
  va_list(arglist);
  Dgp_Int32 len = 0, max = cookie->length - 1, tmp;

  if(cookie) {
    va_start(arglist, format);
    if(error == DGP_REPORT) {
      tmp = snprintf(cookie->str, max, "Report -- ");
    } else if(error == DGP_WARNING) {
      tmp = snprintf(cookie->str, max, "Warning -- ");
    } else {
      tmp = snprintf(cookie->str, max, "Error %d, \"%s\", line %d, %s -- ", (int)error, file_name, (int)line_number, funct_name);
    }
    if(tmp < 0) goto filled;
    len += tmp;
    if(format != NULL) tmp = vsnprintf((cookie->str + len), (max - len), format, arglist);
    if(tmp < 0) goto filled;
    len += tmp;
    if(len == max) goto filled;
  }

  return;

filled:
  cookie->str[cookie->length-1] = '\0';
}

#else /* DGP_DBUG is not defined */

/* default function to output debug errors to a string */
void 
Dgp_safeStrErrPrinter(void *_cookie, ErrCode error, const Dgp_Int8 *format, ...)
{
  Dgp_SafeErrString *cookie = (Dgp_SafeErrString *)_cookie;
  va_list(arglist);
  Dgp_Int32 len = 0, max = cookie->length - 1, tmp;

  if(cookie) {
    va_start(arglist, format);
    if(error == DGP_REPORT) {
      tmp = snprintf(cookie->str, max, "Report -- ");
    } else if(error == DGP_WARNING) {
      tmp = snprintf(cookie->str, max, "Warning -- ");
    } else {
      tmp = snprintf(cookie->str, max, "Error %d -- ", (int)error);
    }
    if(tmp < 0) goto filled;
    len += tmp;
    if(format != NULL) tmp = vsnprintf((cookie->str + len), (max - len), format, arglist);
    if(tmp < 0) goto filled;
    len += tmp;
    if(len == max) goto filled;
  }

  return;

filled:
  cookie->str[cookie->length-1] = '\0';
}

#endif
