
#include "dgp_sm_alloc.h"
#include "dgp_utils.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_sm_alloc.c,v 1.10 2003/02/11 01:31:41 mkrocken Exp $");


/*
  These are the current chunk sizes:

	// list stuff 
X    void *lnChunk;   = 16 | 16
X    void *listChunk; = 16 | 16

	// priority queue stuff (libadt/priq.c) 
X    void *pqChunk; = 20   | 20
X    void *pqnChunk; = 8   | 8
X    void *pqinChunk; = 28 | 28


	// lossy tokenizer stuff (libtok-lossy/miscalloc.c) 
X	void *tokCModelInfoChunk; = 24 | 24
X	void *tokCImageInfoChunk; = 20 | 20
X	void *pixelSumChunk; = 17 | 20
X	void *classChunk;  = 20 | 20
X	void *tokCBoxTypeChunk; = 28 | 28

	// color encoding stuff (libcolcomp/encode_colors.c) 
X	void *unique_tagChunk; = 12 | 16 
X	void *tag_runChunk; = 8 | 8

  // lossless token compressor (libtok-lossless/miscalloc.c) 
X  void *tokClassChunk; = 28 | 28

  // libscancolor (libscancolor/chains.c edges.c) 
X  void *chainChunk; = 128  | 128
X  void *edgeChunk;  = 8 | 8

  // libsamptok 
X  void *samptokClassChunk; = 36

  // libtclydoc & tcomp-scan-color (tcomp-scan-color/getBoxes.c) (libtclydoc/tiff_cc2box.c) 
X  void *ccChunk; = 38
cc 44 | 128

  // libcomponent 
X  void *DgpCC_CCChunk; = 34

  // libsupercc  
X  void *scrs_chunk; 5 + (4 * 4) = 21 | 24
X  void *scxrs_chunk; ... + (4 * 4) = 37 | 40
colorCC   ?? | 24
CCstate   44

*/

#if defined EQUATOR_CA_ENV || defined PSG /* use PSG IPCORE */
#undef NORMAL_RETURN
#define NORMAL_RETURN do {if (1) return __error__;} while (0)
#endif

#define SMALLPOOL_NUM_POOLS 8

/* init pools = 8, 16, 24, 44, 128 */
#define SMALLPOOL_INIT_POOLS \
  (DGP_USE_SMALLPOOL_MIN8 | DGP_USE_SMALLPOOL_MIN16 | DGP_USE_SMALLPOOL_MIN24 | \
   DGP_USE_SMALLPOOL_MIN44 | DGP_USE_SMALLPOOL_MIN128)

/* How many objects to allocate at once */
#define	SMALLPOOL_QUANTUM_SMALL	1000
#define	SMALLPOOL_QUANTUM_LARGE	100

static Dgp_UInt16 pool_quantums[SMALLPOOL_NUM_POOLS] =
  {SMALLPOOL_QUANTUM_SMALL,
   SMALLPOOL_QUANTUM_SMALL,
   SMALLPOOL_QUANTUM_SMALL,
   SMALLPOOL_QUANTUM_SMALL,
   SMALLPOOL_QUANTUM_SMALL,
   SMALLPOOL_QUANTUM_SMALL,
   SMALLPOOL_QUANTUM_SMALL,
   SMALLPOOL_QUANTUM_LARGE};

static Dgp_UInt32 pool_select_masks[SMALLPOOL_NUM_POOLS] =
  {DGP_USE_SMALLPOOL_MIN8,
   DGP_USE_SMALLPOOL_MIN16,
   DGP_USE_SMALLPOOL_MIN20,
   DGP_USE_SMALLPOOL_MIN24,
   DGP_USE_SMALLPOOL_MIN28,
   DGP_USE_SMALLPOOL_MIN32,
   DGP_USE_SMALLPOOL_MIN44,
   DGP_USE_SMALLPOOL_MIN128};

typedef struct _Dgp_ChunkBit {
    struct _Dgp_ChunkBit *next;
} Dgp_ChunkBit;

/*
11.5.2001 / dhtesq.
Here's a good one.  Users will cast chunks of this pool into whatever they like -- typically
structures involving a mix of data types.  Apparently on Solaris (Sparc?) double-precision 
storage has to aligned to (a multiple of) 64 bits.  32 won't cut it;  you'll get bus errors 
when you dereference the memory.

To avoid this, we have to make sure this header size is a multiple of 8 bytes.  Some will be
unused padding.
*/
#define BLOCKHEADER_SIZE (sizeof(Dgp_ChunkBit) + 8 - (sizeof(Dgp_ChunkBit) % 8))

typedef struct {
  Dgp_ChunkBit *free_list;
  Dgp_Int32 chunk_size;
} Dgp_PoolInfo;


typedef struct {
  Dgp_PoolInfo *mapped_pools[DGP_NUM_SMALLPOOLS];
  Dgp_PoolInfo actual_pools[DGP_NUM_SMALLPOOLS];
  Dgp_ChunkBit *block_list;
  Dgp_UInt32 used_pools;
  Dgp_State *dgp_state;
} Dgp_SmallPools;


ErrCode
dgp_mallocSmall(void *small_cookie, 
                Dgp_Int32 pool_id, 
                void **retptr)
{
  Dgp_State *dgp_state = ((Dgp_SmallPools *)small_cookie)->dgp_state;
  DGP_PROLOGUE("dgp_mallocSmall()", DGP_PERRSTATE);

  void *ptr, *tmpptr;
  Dgp_Int32 i;
  Dgp_PoolInfo *cur_pool;

  DGP_ASSERT(small_cookie != NULL);
  DGP_ASSERT((pool_id >= 0) && (pool_id < DGP_NUM_SMALLPOOLS));
  DGP_ASSERT(retptr != NULL);

  cur_pool = ((Dgp_SmallPools *)small_cookie)->mapped_pools[pool_id];

  /* this pool had better exist! */
  DGP_ASSERT(cur_pool != NULL);

  /* the fast case (we hope) */
  if(cur_pool->free_list != NULL) {
    
    *retptr = (void *)cur_pool->free_list;
    cur_pool->free_list = cur_pool->free_list->next;

    NORMAL_RETURN;

  }

	/* Allocate a blob */
  DGP_CHECK_ALLOC(ptr = DGP_MALLOC((cur_pool->chunk_size * pool_quantums[pool_id]) +
                                   BLOCKHEADER_SIZE));

  /* save a pointer to this new block */
	((Dgp_ChunkBit *)ptr)->next = ((Dgp_SmallPools *)small_cookie)->block_list;
	((Dgp_SmallPools *)small_cookie)->block_list = (Dgp_ChunkBit *)ptr;

  /* remember that we've used this pool (and, make sure we record
     the *actual* pool this block is in, not the initial pool_id
     pool, because it may be mapped to a larger pool in reality) */
  ((Dgp_SmallPools *)small_cookie)->used_pools |= 
    pool_select_masks[cur_pool - ((Dgp_SmallPools *)small_cookie)->actual_pools];

  /* skip past this header */
  ptr = (void *)(((Dgp_Int8 *)ptr) + BLOCKHEADER_SIZE);

	/* Thread the list through it */
	for (i = 1, tmpptr = (void *)(((Dgp_Int8 *)ptr) + cur_pool->chunk_size);
		   i < pool_quantums[pool_id] - 1;
		   i++, tmpptr = (void *)(((Dgp_Int8 *)tmpptr) + cur_pool->chunk_size)) {
		((Dgp_ChunkBit *)tmpptr)->next = 
      (Dgp_ChunkBit *)(((Dgp_Int8 *)tmpptr) + cur_pool->chunk_size);
	}
	/* Terminate the list */
	((Dgp_ChunkBit *)tmpptr)->next = (Dgp_ChunkBit *)NULL;
	
	/* Point the freelist at the second element */
	cur_pool->free_list = (Dgp_ChunkBit *)(((Dgp_Int8 *)ptr) + cur_pool->chunk_size);

	/* and return the first */
	*retptr = (void *)ptr;


  NORMAL_RETURN;
}

void
dgp_freeSmall(void *small_cookie,
              Dgp_Int32 pool_id, 
              void *ptr)
{
  Dgp_State *dgp_state = ((Dgp_SmallPools *)small_cookie)->dgp_state;
  DGP_VOID_PROLOGUE("dgp_freeSmall()", DGP_PERRSTATE);

  Dgp_PoolInfo *cur_pool;

  DGP_VOID_ASSERT(dgp_state != NULL);
  DGP_VOID_ASSERT((pool_id >= 0) && (pool_id < DGP_NUM_SMALLPOOLS));
  DGP_VOID_ASSERT(ptr != NULL);

  cur_pool = ((Dgp_SmallPools *)small_cookie)->mapped_pools[pool_id];

  ((Dgp_ChunkBit  *)ptr)->next = cur_pool->free_list;
  cur_pool->free_list = (Dgp_ChunkBit  *)ptr;

}


void
dgp_selectPools(void *small_cookie, 
                Dgp_UInt32 selected_pools)
{
  Dgp_Int32 i = DGP_NUM_SMALLPOOLS - 1;
  
  /* set initial pool to the largest */
  Dgp_PoolInfo *cur_pool = &(((Dgp_SmallPools *)small_cookie)->actual_pools[i]);
    
  /* work from largest to smallest */
  for(; i >= 0; i--) {

    /* we select this pool if the user selected it *or* it's already been
       used by someone else.  we can't get rid of a pool if someone else
       has already allocated from it... BAD! */
    if(((selected_pools & pool_select_masks[i]) != 0) ||
      ((((Dgp_SmallPools *)small_cookie)->used_pools & pool_select_masks[i]) != 0)) {
      cur_pool = &(((Dgp_SmallPools *)small_cookie)->actual_pools[i]);
    }
    ((Dgp_SmallPools *)small_cookie)->mapped_pools[i] = cur_pool;
  }

}




ErrCode
dgp_initSmallMem(Dgp_State *dgp_state, 
                 void **_cookie)
{
  DGP_STATE_PROLOGUE2("dgp_initSmallMem()");
  Dgp_Int32 i;
  Dgp_SmallPools *cookie = NULL;

  DGP_ERR_BEGIN(dgp_state);
  DGP_CHECK_NULL(_cookie);

  /* make sure our struct fits in the smallest pool */
  DGP_SMALLPOOL_VALIDATE(Dgp_ChunkBit, DGP_SMALLPOOL_MIN8);

  /* make sure this code is up to date, if DGP_NUM_SMALLPOOLS changes,
     the arrays above need to change! */
  DGP_ASSERT(SMALLPOOL_NUM_POOLS == DGP_NUM_SMALLPOOLS);

  DGP_SEC_CALLOC(cookie, (Dgp_SmallPools *), 1, sizeof(Dgp_SmallPools));
  cookie->dgp_state = dgp_state;

  /* set sizes */
  for(i = 0; i < DGP_NUM_SMALLPOOLS; i++) {
    cookie->actual_pools[i].chunk_size = Dgp_getPoolSize(i);
  }

  /* now, set the pools we are using */
  dgp_selectPools(cookie, SMALLPOOL_INIT_POOLS);

  /* return the new cookie */
  *_cookie = cookie;
  cookie = NULL;

  NORMAL_RETURN;
ERR_CLEANUP:
  ERR_RETURN;
}

void
dgp_cleanupSmallMem(void **_cookie)
{
  void *next;
  Dgp_SmallPools *small_pools;
  Dgp_Free tmp_free;

  if(_cookie == NULL) return;

  small_pools = (Dgp_SmallPools *)(*_cookie);
  *_cookie = NULL;

  if(small_pools == NULL) return;

  tmp_free = small_pools->DGP_FREE;


  while(small_pools->block_list != NULL) {
    next = small_pools->block_list->next;

    tmp_free(small_pools->block_list);

    small_pools->block_list = next;
  }

  tmp_free(small_pools);

}
