/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 1, 2000
**
**  Filename: dgp_utils.h
**
**  Description of what's implemented in this file:
**    utility macros to simplify error handling
**  
**************************************************************************/

/*
 * $Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_utils.h,v 1.8 2000/04/25 22:29:04 jahlborn Exp $
 */


#ifndef __DGP_UTILS_H
#define __DGP_UTILS_H

#include <dgp_state.h>

#ifdef __cplusplus
extern "C"
{
#endif

#define DGPERR_INIT(name) DGP_PROLOGUE(#name "()", DGP_PERRSTATE)

/* allocator calls (must have "Dgp_State *dgp_state" defined) */
#define DGP_SEC_MALLOC(ret, type, size) \
	DGP_CHECK_ALLOC_CLEANUP((ret) = type (DGP_MALLOC)((size)))
#define DGP_SEC_CALLOC(ret, type, num, size) \
	DGP_CHECK_ALLOC_CLEANUP((ret) = type (DGP_CALLOC)((num), (size)))
#define DGP_SEC_REALLOC(ret, type, size) \
  do { \
    void *__tmp_ptr__; \
  	DGP_CHECK_ALLOC_CLEANUP(__tmp_ptr__ = type (DGP_REALLOC)((ret), (size))); \
    (ret) = __tmp_ptr__; \
  } while(0)


/* allocator calls (must have "Dgp_Allocators *a" defined 
   (less pointer chasing than above macros)) */
#define DGP_SEC_AMALLOC(ret, type, size) \
	DGP_CHECK_ALLOC_CLEANUP((ret) = type (a->malloc)((size)))
#define DGP_SEC_ACALLOC(ret, type, num, size) \
	DGP_CHECK_ALLOC_CLEANUP((ret) = type (a->calloc)((num), (size)))
#define DGP_SEC_AREALLOC(ret, type, size) \
  do { \
    void *__tmp_ptr__; \
  	DGP_CHECK_ALLOC_CLEANUP(__tmp_ptr__ = type (a->realloc)((ret), (size))); \
    (ret) = __tmp_ptr__; \
  } while(0)

#define DGP_AFREE(ptr) (a->free)(ptr)

#ifdef __cplusplus
}
#endif

#endif /* __DGP_UTILS_H */
