/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, February 1 1998
**
**  Filename: dgp_w32_err_cotrol.h
**
**  Description of what's implemented in this file:
**    macros to simplify error reporting of win32 errors.
**  
**************************************************************************/

/*
 * RCS Header
 *
 * $Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_w32_err_control.h,v 1.6 2000/05/16 21:13:20 jahlborn Exp $
 */

#ifndef __DGP_W32_ERR_CONTROL_H
#define __DGP_W32_ERR_CONTROL_H

#include <stdio.h>
#include <string.h>
#include "dgp_err_control.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* 
 * These macros wrap win32 errors.
 */

/* 
 * Errors from 13,000 - 13,999 will be digipaper wrapped win32 errors. 
 */
#define cDgpW32 13000
#define maxDgpW32 13999


#ifdef _WIN32

/* utility macros for the macros defined below */
#ifdef DGP_DBUG

#define DGP_W32_ERR_PRINTER \
  if(__PERR_STATE__ != NULL) Dgp_w32ErrPrinter(__PERR_STATE__, __FILE__, __FUNCT_NAME__, __LINE__, __error__)

#else /* DGP_DBUG is not defined */

#define DGP_W32_ERR_PRINTER \
  if(__PERR_STATE__ != NULL) Dgp_w32ErrPrinter(__PERR_STATE__, __error__)

#endif


/*
 * DGP_W32_ERR_GEN(F*)
 *   Generate a win32 error. "cond" is some condition.
 */ 
#define DGP_W32_ERR_GEN(cond)\
do { \
  if (cond) \
  {\
    __error__ = GetLastError() + cDgpW32;\
    DGP_W32_ERR_PRINTER;\
    return __error__;\
  }\
} while(0)

#define DGP_W32_ERR_GENF0(cond, err_str)\
do { \
  if(cond) \
  {\
    __error__ = GetLastError() + cDgpW32;\
    DGP_ERR_PRINTERF0(err_str);\
    return __error__;\
  }\
} while(0)


/*
 * DGP_W32_ERR_GEN_NO_MSG
 *   Generate a win32 error. "cond" is some condition.
 */ 
#define DGP_W32_ERR_GEN_NO_MSG(cond)\
do { \
  if (cond) \
  {\
    __error__ = GetLastError() + cDgpW32; \
    DGP_NO_PRINTER; \
    return __error__; \
  }\
} while(0)


/*
 * DGP_W32_ERR_GEN_CLEANUP(F*)
 *   Win32 error generation: instead of returning, as in DGP_ERR_GEN, go to a cleanup
 *   action at the end of the function. 
 */
#define DGP_W32_ERR_GEN_CLEANUP(cond)\
do { \
  if (cond) \
  {\
    __error__ = GetLastError() + cDgpW32;\
    DGP_W32_ERR_PRINTER;\
    goto ERR_CLEANUP;\
  }\
} while(0)

#define DGP_W32_ERR_GEN_CLEANUPF0(cond, err_str)\
do { \
  if(cond) \
  {\
    __error__ = GetLastError() + cDgpW32;\
    DGP_ERR_PRINTERF0(err_str);\
    goto ERR_CLEANUP;\
  }\
} while(0)

#define DGP_W32_ERR_GEN_CLEANUPF1(cond, err_str, arg1)\
do { \
  if(cond) \
  {\
    __error__ = GetLastError() + cDgpW32;\
    DGP_ERR_PRINTERF1((err_str), (arg1));\
    goto ERR_CLEANUP;\
  }\
} while(0)

#define DGP_W32_ERR_GEN_CLEANUPF2(cond, err_str, arg1, arg2)\
do { \
  if(cond) \
  {\
    __error__ = GetLastError() + cDgpW32;\
    DGP_ERR_PRINTERF2((err_str), (arg1), (arg2));\
    goto ERR_CLEANUP;\
  }\
} while(0)

#define DGP_W32_ERR_GEN_CLEANUPF3(cond, err_str, arg1, arg2, arg3)\
do { \
  if(cond) \
  {\
    __error__ = GetLastError() + cDgpW32;\
    DGP_ERR_PRINTERF3((err_str), (arg1), (arg2), (arg3));\
    goto ERR_CLEANUP;\
  }\
} while(0)


/*
 * DGP_W32_ERR_GEN_CLEANUP_NO_MSG
 *   Generate a system error. "cond" is some condition and don't output
 *   anything.  This must be followed by braces {}.
 */ 
#define DGP_W32_ERR_GEN_CLEANUP_NO_MSG(cond)\
do { \
  if (cond) \
  {\
    __error__ = GetLastError() + cDgpW32;\
    DGP_NO_PRINTER; \
    goto ERR_CLEANUP; \
  }\
} while(0)


#ifdef DGP_DBUG

DGP_LIBEXPORT void
Dgp_w32ErrPrinter(Dgp_ErrState *dgp_err_state, const Dgp_Int8 *file_name, const Dgp_Int8 *funct_name,
                    Dgp_Int32 line_number, ErrCode error);

#else /* DGP_DBUG is not defined */

DGP_LIBEXPORT void
Dgp_w32ErrPrinter(Dgp_ErrState *dgp_err_state, ErrCode error);

#endif /* DGP_DBUG */


#endif /* _WIN32 */

#ifdef __cplusplus
}
#endif

#endif /* __DGP_W32_ERR_CONTROL_H */







