/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, September 1 1999
**
**  Filename: dgp_xif_err_cotrol.h
**
**  Description of what's implemented in this file:
**    macros to simplify error reporting.
**  
**************************************************************************/

/*
 * RCS Header
 *
 * $Header: /project/icons/cvsroot/digipaper/src/liberr/dgp_xif_err_control.h,v 1.12 2000/03/04 00:22:22 mkrocken Exp $
 */

#ifndef __DGP_XIF_ERR_CONTROL_H
#define __DGP_XIF_ERR_CONTROL_H

#include "dgp_err_control.h"
#include "xiff.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* 
 * These macros wrap the xiff errors.  The reason there is a need for this is 
 * because there is a conflict between the error codes of xiff and those of
 * ipcore, DOH!
 */

/* 
 * Errors from 11,000 - 11,999 will be digipaper wrapped xif errors. 
 */
#define cDgpXif 11000
#define maxDgpXif 11999


/*
 * DGP_XIF_ERR_PASS(F*)
 *   This is an error check on a function call. It assumes that the
 *   function returns an error code, and if the error code is non-zero,
 *   returns from the function (after possibly printing an error message).
 */
#define DGP_XIF_ERR_PASS(functionCall)\
do {\
  __error__ = (functionCall);\
  if (__error__ != XIF_NOERROR)\
  {\
    __error__ += cDgpXif; \
    DGP_ERR_PASS_PRINTER;\
    return __error__;\
  }\
} while (0)

#define DGP_XIF_ERR_PASSF0(functionCall, err_str)\
do {\
  __error__ = (functionCall);\
  if (__error__ != XIF_NOERROR)\
  {\
    __error__ += cDgpXif; \
    DGP_ERR_PASS_PRINTERF0(err_str);\
    return __error__;\
  }\
} while (0)


/*
 * DGP_XIF_ERR_PASS_CLEANUP(F*)
 *   This is an error check on a function call. It is like DGP_XIF_ERR_PASS except
 *   that rather than returning from the function, it sets the local
 *   "__error__" variable, and does a "goto" to some cleanup code. The
 *   cleanup code is labeled "ERR_CLEANUP". The cleanup should have an
 *   invocation of the "ERR_RETURN" macro at the end of it.
 */
#define DGP_XIF_ERR_PASS_CLEANUP(functionCall)\
do {\
  __error__ = (functionCall);\
  if (__error__ != XIF_NOERROR)\
  {\
    __error__ += cDgpXif; \
    DGP_ERR_PASS_PRINTER;\
    goto ERR_CLEANUP;\
  }\
} while (0)

#define DGP_XIF_ERR_PASS_CLEANUPF0(functionCall, err_str)\
do {\
  __error__ = (functionCall);\
  if (__error__ != XIF_NOERROR)\
  {\
    __error__ += cDgpXif; \
    DGP_ERR_PASS_PRINTERF0(err_str);\
    goto ERR_CLEANUP;\
  }\
} while (0)

#define DGP_XIF_ERR_PASS_CLEANUPF1(functionCall, err_str, arg1)\
do {\
  __error__ = (functionCall);\
  if (__error__ != XIF_NOERROR)\
  {\
    __error__ += cDgpXif; \
    DGP_ERR_PASS_PRINTERF1((err_str), (arg1));\
    goto ERR_CLEANUP;\
  }\
} while (0)

#define DGP_XIF_ERR_PASS_CLEANUPF2(functionCall, err_str, arg1, arg2)\
do {\
  __error__ = (functionCall);\
  if (__error__ != XIF_NOERROR)\
  {\
    __error__ += cDgpXif; \
    DGP_ERR_PASS_PRINTERF2((err_str), (arg1), (arg2));\
    goto ERR_CLEANUP;\
  }\
} while (0)

#define DGP_XIF_ERR_PASS_CLEANUPF3(functionCall, err_str, arg1, arg2, arg3)\
do {\
  __error__ = (functionCall);\
  if (__error__ != XIF_NOERROR)\
  {\
    __error__ += cDgpXif; \
    DGP_ERR_PASS_PRINTERF3((err_str), (arg1), (arg2), (arg3));\
    goto ERR_CLEANUP;\
  }\
} while (0)


/*
 * DGP_XIF_ERR_CHECK_NO_MSG
 *   This is an even simpler variant of an error check on a function call.
 *   It expects a statement or block to follow it, and will execute
 *   that statement or block if an error ocurred in the function call.
 *   It differs from ERR_CHECK in that no error message is printed out.
 *   This macro is used when we haven't decided whether there's really
 *   an error.
 */
#define DGP_XIF_ERR_CHECK_NO_MSG(functionCall)\
    __error__ = (functionCall);\
    if (__error__ != XIF_NOERROR)


/*
 * DGP_WRAP_XIF_ERR
 *   This macro translates a xiff error code to a set of error codes
 *   which do not overlap with ipcore.
 */
#define DGP_WRAP_XIF_ERR(error) ((error) + cDgpXif)

#ifdef __cplusplus
}
#endif

#endif /* __DGP_XIF_ERR_CONTROL_H */







