/* $Header: /project/icons/cvsroot/digipaper/src/libg4/CCITTFaxCompress.c,v 1.25 2003/02/10 22:53:24 mkrocken Exp $ */

/*
 * -- CCITTFaxCompressImpl.mesa
 *   -- Copyright (c) 1998 by Xerox Corporation.  All rights reserved.
 *   -- Russ Atkinson (RRA) October 17, 1998 10:19 pm PDT
 *
 * Translated to C by James Ahlborn in January 1999.
 */

#if defined(_WIN32) || defined (EQUATOR_CA_ENV)
#define _LITTLE_ENDIAN
#else
#if defined(VXWORKS) || defined(X86_LINUX) || defined(LYNX)
#include "endianness.h"
#else
#include <sys/isa_defs.h>
#endif  /* (VXWORKS) */
#endif  /* (_WIN32) */

#include <stdio.h>
#include <stdlib.h>
#include "util.h"
#include "dgp_err_control.h"
#include "dgp_allocators.h"
#include "g4.h"
#include "g4private.h"
#include "PsgFix.h"


/* set canDebug to 1 to enable debugging output */
#define canDebug  0
#define eliDebug  0
#define eliDebug2 0

static TabEntry *shortTabs[2] = {whiteShortCodeTab, blackShortCodeTab};
static TabEntry *longTabs[2] = {whiteLongCodeTab, blackLongCodeTab};


/* 
   Resets the cd object to the initial state.  This is useful when desiring
   to use the same compression handle over several pages, since the storage
   can be reused.
   -- Internal buffers will be allocated according to the bpl parameter for
      outBuffer, transOld, and transNew.
   */
ErrCode
EncodeInit(CompressData *cd, ULONG bpl, Dgp_Bool invertColor)
{
  DGP_PROLOGUE("EncodeInit", cd->DGP_PERRSTATE);

  Dgp_Allocators *a = &(cd->DGP_ALLOCATORS);
  ULONG lim = bpl + 4;
  Trans *tnew = cd->transNew, *told = cd->transOld;


  if(bpl > cd->old_bpl) {
    
    if(tnew->data) (a->free)(tnew->data);

    DGP_CHECK_ALLOC_CLEANUP(tnew->data = 
	    (ULONG *)(a->malloc)(lim * sizeof(*(tnew->data))));
    tnew->max = lim;

    if(told->data) (a->free)(told->data);

    DGP_CHECK_ALLOC_CLEANUP(told->data = 
	   (ULONG *)(a->malloc)(lim * sizeof(*(told->data))));
    told->max = lim;

    if(cd->outBuffer.data) (a->free)(cd->outBuffer.data);

    DGP_CHECK_ALLOC_CLEANUP(cd->outBuffer.data = 
	   (Dgp_Int8 *)(a->malloc)(sizeof(*(cd->outBuffer.data)) * (bpl+10)));
    cd->outBuffer.maxLength = bpl+10;
    cd->old_bpl = bpl;
  }

  cd->bpl = bpl;
  tnew->len = 0;
  told->len = 0;
  tnew->data[0] = tnew->data[1] = bpl;
  told->data[0] = told->data[1] = bpl;

  cd->outBuffer.length = 0;

  cd->outMod = 0;
  cd->outWord = 0;

  if (!invertColor)
    cd->sense = 0;  /* White is default */
  else
    cd->sense = 1;

 
  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


/* cleanup memory stored in comress data struct */
void
EncodeCleanUp(void **_cd)
{
  CompressData *cd;
  Dgp_Free tmp_free;
  Trans *trans;

  if(_cd == NULL) return;

  cd = *_cd;
  *_cd = NULL;

  if(cd == NULL) return;

  tmp_free = cd->DGP_FREE;

  trans = cd->transNew;
  if(trans) {
    if(trans->data) (tmp_free)(trans->data);
    trans->data = NULL;
    (tmp_free)(trans);
  }
  cd->transNew = NULL;

  trans = cd->transOld;
  if(trans) {
    if(trans->data) (tmp_free)(trans->data);
    trans->data = NULL;
    (tmp_free)(trans);
  }
  cd->transOld = NULL;

  if(cd->outBuffer.data) (tmp_free)(cd->outBuffer.data);
  cd->outBuffer.data = NULL;

  (tmp_free)(cd);
}


/*
  -- Creates a new handle for compression.
  */
ErrCode
EncodeCreate(void **_cd, Dgp_State *dgp_state, FILE *db_out) 
{
  DGP_PROLOGUE("EncodeCreate", DGP_PERRSTATE);

  CompressData *cd = NULL;
  Dgp_Allocators *a = &(dgp_state->a);

  
  *_cd = NULL;

  DGP_CHECK_ALLOC_CLEANUP(cd = (CompressData *)(a->calloc)(1, sizeof(*cd)));

  DGP_CHECK_ALLOC_CLEANUP(cd->transNew = (Trans *)(a->calloc)(1, sizeof(*(cd->transNew))));

  DGP_CHECK_ALLOC_CLEANUP(cd->transOld = (Trans *)(a->calloc)(1, sizeof(*(cd->transOld))));

  cd->dgp_state = dgp_state;
  cd->debug = db_out;

  *_cd = cd;

  NORMAL_RETURN;

ERR_CLEANUP:

  EncodeCleanUp((void **)&cd);

  ERR_RETURN;
}


/*
  -- Converts a line of raw data to transitions.  All lines begin with the
  first (leftmost) bit of the word addressed by ptr.
  -- The transitions in transNew are moved to those in transOld before the
  conversion.
  -- The new transitions are then put into transNew.
  */
ErrCode
LineToTrans(CompressData *cd, ULONG *ptr)
{
  DGP_PROLOGUE("LineToTrans", cd->DGP_PERRSTATE);
  ULONG sense = cd->sense, pos = 0, w = 0, len = (ULONG)cd->bpl, rem, xor, tmp;
/*  ULONG sense = 0, pos = 0, w = 0, len = (ULONG)cd->bpl, rem, xor, tmp;*/
  Trans *trans = cd->transOld;
  ULONG *dst = trans->data;
  ULONG *dp = dst;

#define AddRun(k)				\
  do {sense = ~sense;				\
  dp[0] = pos;					\
  dp++;						\
  pos += (k);} while(0)

#define AddRun2(k1, k2)			       		\
  do {dp[0] = pos;						\
  pos = pos + (k1);					\
  dp[1] = pos;						\
  dp += 2;				 		\
  pos += (k2);} while(0)

  cd->transOld = cd->transNew;
  cd->transNew = trans;

  /* Do whole words in this loop */
  while(len >= bpw) {
    rem = bpw;
    w = ptr[0];
    ptr++;
    len -= rem;
    
    /* First of all, take care of Dgp_Int32 runs */
    if(sense != w) {
      /* Quick test for the entire word being OK */
      while(rem >= 4) {
	xor = w ^ sense;
	tmp = xor >> (bpw-4);
	if(tmp == 0) {
	  /* Quick test for the remainder being a single run */
	  if((xor >> (UCHAR)(bpw-rem)) == 0) {
	    break;
	  }
	  pos += 4;
	} else if(tmp == 1) {
	  pos += 3;
	  AddRun(1);
	} else if(tmp == 2) {
	  pos += 2;
	  AddRun2(1, 1);
	} else if(tmp == 3) {
	  pos += 2;
	  AddRun(2);
	} else if(tmp == 4) {
	  pos += 1;
	  AddRun2(1, 2);
	} else if(tmp == 5) {
	  pos += 1;
	  AddRun2(1, 1);
	  AddRun(1);
	} else if(tmp == 6) {
	  pos += 1;
	  AddRun2(2, 1);
	} else if(tmp == 7) {
	  pos += 1;
	  AddRun(3);
	} else {
	  /* In wrong sense, so cut current run Dgp_Int16 and try again */
	  AddRun(0);
	  continue;
	}
	w *= 16;
	rem -= 4;
      }
    }
    pos += rem;
  }

  if(len != 0) {
    /* Finish up remainder of line, bit by bit */
    w = ptr[0];
    while(len > 0) {
      if(((Dgp_Int32)(w ^ sense)) < 0) {
	AddRun(1);
      } else {
	pos++;
      }
      len -= 1;
      w += w;
    }
  }
  
  DGP_ERR_GEN_CLEANUPF0((pos != cd->bpl), cErrCompressionError, "invalid data");

  dp[0] = pos;
  dp++;

  /* Compute the length and terminate with 2 entries */
  trans->len = ((ULONG)(dp-dst)) /*/ sizeof(ULONG)*/;
  dp[0] = dp[1] = pos;

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


/* FlipSense  (Added by Eli)
 * Sense is the color (black/white) of the run of same-colored pixels */
void
FlipSense(CompressData *cd)
{
  cd->sense = ~cd->sense;
}


/* Takes nb bits, right-justified in w, and puts them into the output bit
   stream. */
#define PBtmp(w, n) do {       						\
  ULONG _w = (w);							\
  UCHAR _n = (UCHAR)(n);	       						\
  UCHAR _oc = outMod;							\
  outMod = (_n + outMod) % bpw;	       					\
  if(outMod > _oc) {							\
    /* Just storing a partial field */					\
    _w = _w << ((UCHAR)(bpw - outMod));					\
    outWord |=  _w;							\
  } else if(_n == 0) {							\
    /* No bits to write */						\
  } else {								\
    /* Crossing a word boundary */					\
    ULONG _newlen = rt->length + sizeof(ULONG);	       			\
    if(outMod == 0) {							\
      DGP_ERR_GEN_CLEANUP((_newlen > rt->maxLength), cErrIntBufOverflow);	\
      /* Filling a word exactly */					\
      wp[0] = outWord | _w;						\
      outWord = 0;							\
    } else {								\
      DGP_ERR_GEN_CLEANUP((_newlen >= rt->maxLength), cErrIntBufOverflow);	\
      /* There are bits left over for the next word */			\
      wp[0] = outWord | (_w >> outMod);					\
      outWord = _w << ((UCHAR)(bpw - outMod));				\
    }									\
    wp++;								\
    rt->length = _newlen;				       		\
  }									\
} while(0)

#if canDebug

#define PB(w, n) do {							\
  if(db) fprintf(db, "PB[%lx, %lu]\n", (ULONG)(w), (ULONG)(n));		\
  PBtmp((w), (n));     							\
} while(0)

#else

#define PB(w, n) PBtmp((w), (n))

#endif


#define PutEnt(ent) PB((ent)->val, (ent)->len)

#define PutHLenTmp(len, state) do {			       		    \
  ULONG _len = (len);							    \
  UCHAR _state = (state);						    \
  TabEntry *_ent;							    \
  _ent = &(longTabs[(Dgp_Int32)_state][(codeLongSplit / codeSplit) % codeSplit]); \
  while(_len >= codeLongSplit) {					    \
    PutEnt(_ent);							    \
    _len -= codeLongSplit;						    \
  }									    \
  if(_len >= codeSplit) {						    \
    _ent = &(longTabs[(Dgp_Int32)_state][(_len / codeSplit) % codeSplit]);	    \
    PutEnt(_ent);							    \
  }									    \
  _ent = &(shortTabs[(Dgp_Int32)_state][_len % codeSplit]);			    \
  PutEnt(_ent);								    \
} while(0)


#if canDebug								    

#define PutHLen(len, state) do {       					     \
  if(db) fprintf(db, "PH [%lu, %d]\n", (ULONG)(len), (Dgp_Int32)(state));          \
  PutHLenTmp((len), (state));						     \
} while(0)

#else

#define PutHLen(len, state) PutHLenTmp((len), (state))

#endif


/* 
   Takes n bits, right-justified in w, and puts them into the output bit
   stream. 
   */
static ErrCode
PutBits(CompressData *cd, ULONG w, UCHAR n)
{
  DGP_PROLOGUE("PutBits", cd->DGP_PERRSTATE);

  UCHAR oc = cd->outMod;
  UCHAR outMod = (n+oc) % bpw;
  ULONG outWord = cd->outWord;
  RefText *rt;
  Dgp_Int32 oldlen;
  ULONG *wp;
  ULONG newlen;

#if canDebug
  if(cd->debug) fprintf(cd->debug, "PB[%lx, %lu]\n", (ULONG)w, (ULONG)n);
#endif

  if(outMod > oc) {
    /* Just storing a partial field */
    w = w << ((UCHAR)(bpw-outMod));
    outWord |= w;
  } else if(n == 0) {
    /* No bits to write */
  } else {
    /* Crossing a word boundary */
    rt = &(cd->outBuffer);
    oldlen = rt->length;
    wp = (ULONG *)(rt->data + oldlen);
    newlen = oldlen + sizeof(ULONG);
    
    if(outMod == 0) {
      DGP_ERR_GEN_CLEANUP((newlen > rt->maxLength), cErrIntBufOverflow);
      /* Filling a word exactly */
      wp[0] = outWord | w;
      outWord = 0;
    } else {
      DGP_ERR_GEN_CLEANUP((newlen >= rt->maxLength), cErrIntBufOverflow);
      /* There are bits left over for the next word */
      wp[0] = outWord | (w >> outMod);
      outWord = w << ((UCHAR)(bpw - outMod));
    }
    rt->length = newlen;
  }
  
  cd->outMod = outMod;
  cd->outWord = outWord;

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


/*  -- Wipes out the compressed data and substitutes uncompressed data for
       the line.
    -- Note that the G3/G4 standards permit running uncompressed data
       across scan lines, but that this implementation does not do so.
       */
static ErrCode
UseUncompressed(CompressData *cd, Dgp_Int32 dim, UCHAR initMod, Dgp_Int32 initPos)
{
  DGP_PROLOGUE("UseUncompressed", cd->DGP_PERRSTATE);

  ULONG *wp = (ULONG *)(cd->outBuffer.data + initPos);
  ULONG m = ~(WORD_LAST >> initMod);
  ULONG bpl = cd->bpl;
  ULONG tx = 0;
  ULONG *trans = (ULONG *)(cd->transNew->data);
  ULONG pos = 0, next, d;

  /* First, restore the output state to the initial state, which positions us
     just after any EOL codes, but before any compressed data. */
  wp[0] = cd->outWord = wp[0] & m;
  cd->outWord = initMod;
  cd->outBuffer.length = initPos;

  /* Emit an entry to uncompressed mode.  The number of bits for the K=0 case
     is still unclear. */

  switch(dim) {
  case 0:
    /* -- ??, K = 0 ?? */
    DGP_ERR_PASS_CLEANUPF0(PutBits(cd, 15, 12), "failed writing bits to output buffer");
    break;
  case 1:
    /* 1-dim line */
    DGP_ERR_PASS_CLEANUPF0(PutBits(cd, 15, 12), "failed writing bits to output buffer");
    break;
  case 2:
    /* 2-dim line, or K < 0 */
    DGP_ERR_PASS_CLEANUPF0(PutBits(cd, 15, 10), "failed writing bits to output buffer");
    break;
  }

  while(DGP_TRUE) {
    next = trans[tx];
    d = next - pos;
    
    if(next == pos) {
      /* Zero-length run, so no code */
    } else if((tx % 2) == 1) {
      /* Encoding a black run, so just put out d 1's */
      while(DGP_TRUE) {
	if(d == 0) {
	  break;
	} else if(d < bpw) {
	  DGP_ERR_PASS_CLEANUPF0(PutBits(cd, (1 << d) - 1, (UCHAR)d), "failed writing bits to output buffer");
	  break;
	} else {
	  DGP_ERR_PASS_CLEANUPF0(PutBits(cd, WORD_LAST, bpw), "failed writing bits to output buffer");
	  d -= bpw;
	}
      }
    } else {
      /* Encoding a white run */
      while(DGP_TRUE) {
	if((next == (ULONG)bpl) && (d <= 4)) {
	  /* This exits uncompressed mode (encoding d zero bits), next 
	     state = white*/
	  DGP_ERR_PASS_CLEANUPF0(PutBits(cd, 2, (UCHAR)(8+d)), "failed writing bits to output buffer");
	  break;
	}
	if(d <= 4) {
	  DGP_ERR_PASS_CLEANUPF0(PutBits(cd, 0, (UCHAR)d), "failed writing bits to output buffer");
	  break;
	} else {
	  DGP_ERR_PASS_CLEANUPF0(PutBits(cd, 1, (UCHAR)6), "failed writing bits to output buffer");
	  d -= 5;
	}
      }
    }
    pos = next;
    if(pos == (ULONG)bpl) {
      /* This exits uncompressed mode, next state = white */
      DGP_ERR_PASS_CLEANUPF0(PutBits(cd, 2, 8), "failed writing bits to output buffer");
      break;
    }
    tx++;
  }

  if(cd->outMod != 0) {
    /* Flush partial bits to the buffer, but don't update the state */
    wp = (ULONG *)(cd->outBuffer.data + cd->outBuffer.length);
    wp[0] = cd->outWord;
  }

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


/*
  Converts the transitions in transNew to G4 coding in the outBuffer.
  */
ErrCode
TransToG4Code(CompressData *cd, Dgp_Bool uncompressed)
{
  DGP_PROLOGUE("TransToG4Code", cd->DGP_PERRSTATE);
  UCHAR outMod = cd->outMod;
  ULONG outWord = cd->outWord;
  RefText *rt = &(cd->outBuffer);
  UCHAR initMod = outMod;
  Dgp_Int32 initPos = rt->length;
  ULONG *wp = (ULONG *)(rt->data + initPos);
  ULONG bpl = cd->bpl;
  ULONG rx = 0;
  ULONG *runs = cd->transNew->data;
  ULONG *lag = cd->transOld->data;
  ULONG lagLen = cd->transOld->len;
  ULONG lagx = 0, a0 = 0, b1 = lag[lagx], b2 = 0, d = 0, a1, a2, w;
  UCHAR color,  nb;
#if canDebug
  FILE *db = (cd->debug);
#endif

  while(a0 < bpl) {
    a1 = runs[rx];
    b2 = lag[lagx+1];
    if(a1 > b2) {
      /* Pass mode, skip a transition on the lagging line
         The clause "AND (lagx+8 > lagLen OR a1 < lag[lagx+8])" can be
	 appended to the above test "a1 > b2" to improve the compression
	 ratio (and decompression speed), but some G4 decompressors can't
	 handle the generality of skipping over several lagging transitions
	 and demand the pass codes instead. Both DocuPrint and PostScript
	 can handle the missing pass codes.
	 */
#if canDebug
      if(db) fprintf(db, "P, a1: %lu, b2: %lu, ", (ULONG)a1, (ULONG)b2);
#endif

      /* pass code */
      PB(1, 4);

      a0 = b2;
      lagx += 2;
      if(lagx > lagLen) {
	lagx = lagLen;
      }
      /* We do NOT have to scan to find the right lagx, since the color is
	 already OK and we know that the next b1 > a0.  This can save a bit
	 of time. */
      b1 = lag[lagx];
      continue;
    }
    d = b1 - a1 + 3;

    if(d < 7) {
      /* Vertical mode */
      w = 1;
      nb = 1;
      
      switch(d) {
      case 0:
	/* -- a1 = b1 + 3 */
	w = 3;
	nb = 7;
	break;
      case 1:
	/* -- a1 = b1 + 2 */
	w = 3;
	nb = 6;
	break;
      case 2:
	/* -- a1 = b1 + 1 */
	w = 3;
	nb = 3;
	break;
      case 4:
	/* -- a1 = b1 + 2 */
	w = 2;
	nb = 3;
	break;
      case 5:
	/* -- a1 = b1 + 3 */
	w = 2;
	nb = 6;
	break;
      case 6:
	/* -- a1 = b1 + 3 */
	w = 2;
	nb = 7;
	break;
      }
#if canDebug
      if(db) fprintf(db, "V, a1: %lu, b1: %lu, ", (ULONG)a1, (ULONG)b1);
#endif
      PB(w, nb);
      rx++;
      a0 = a1;
    } else {
      /* Horizontal Mode */
      a2 = runs[rx + 1];
      color = (UCHAR)(rx % 2);
#if canDebug
      if(db) fprintf(db, "H, a1: %lu, a2: %lu, ", (ULONG)a1, (ULONG)a2);
#endif
      /* horizontal code */
      PB(1, 3);

      PutHLen((a1-a0), color);
      PutHLen((a2-a1), 1-color);
      rx = rx+2;
      a0 = a2;
    }

    /* At this point the value of b1 (based on lag[lagx]) is probably
       wrong, so we have to update it to correspond to the new a0. */
    if(lagx > 0) {
      lagx -= 1;
    }

    while(DGP_TRUE) {
      b1 = lag[lagx];
      if((b1 > a0) && ((rx % 2) == (lagx %2))) break;
      if(lagx >= lagLen) break;
      lagx++;
    }
  }

  cd->outMod = outMod;
  cd->outWord = outWord;

  /* Flush partial bits to the buffer, but don't update the state */
  if(outMod != 0) {
    wp[0] = outWord;
  }

  DGP_ERR_GEN_CLEANUPF0((a0 != bpl), cErrUndefined, "a0 not equal to bpl");
  
  if(uncompressed) {
    if(((ULONG)(rt->length - initPos) * 8) > ((bpl*4) + bpl)) {
      DGP_ERR_PASS_CLEANUPF0(UseUncompressed(cd, 2, initMod, initPos), 
        "failed writing uncompressed data");
    }
  }

  NORMAL_RETURN;
  
ERR_CLEANUP:

  ERR_RETURN;
}


/*
  -- Converts a sequence of transitions (created by LineToTrans) to G3 coding.
  -- dim = 0 => 1-D coding, no leading tag bit
  -- dim = 1 => 1-D coding, leading tag bit is 1
  -- dim = 2 => 1-D coding, leading tag bit is 0
  -- uncompressed => permits uncompressed coding of the line if the line would
be encoded more efficiently with the uncompressed encoding [note: although not
required by G3 spec, we encode the whole line as uncompressed]
*/
ErrCode
TransToG3Code(CompressData *cd, Dgp_Int32 dim, Dgp_Bool uncompressed)
{
  DGP_PROLOGUE("TransToG3Code", cd->DGP_PERRSTATE);
  UCHAR outMod = cd->outMod;
  ULONG outWord = cd->outWord;
  RefText *rt = &(cd->outBuffer);
  UCHAR initMod = outMod;
  Dgp_Int32 initPos = rt->length;
  ULONG *wp = (ULONG *)(rt->data + initPos);
  ULONG bpl = cd->bpl;
  ULONG rx = 0;
  ULONG *runs = cd->transNew->data;
  /*ULONG *lag = cd->transOld.data;
  ULONG lagLen = cd->transOld.len;*/
  ULONG a0 = 0, /*lagx = 0, b1 = lag[lagx], b2 = 0, d = 0,*/ a1, a2;
  UCHAR color;
#if canDebug
  FILE *db =  (cd->debug);
#endif

  switch(dim) {
  case 1:
    /* 1-dim tag */
    PB(1, 1);
    break;
  case 2:
    /* 2-dim tag */
    PB(0, 1);
    cd->outMod = outMod;
    cd->outWord = outWord;
    DGP_ERR_PASS_CLEANUPF0(TransToG4Code(cd, uncompressed), "g4 encoding failed");
    NORMAL_RETURN;
    break;
  default:
    /* no tag */
    break;
  }

  /* Save the state here (after any EOL and tag bits) */
  initMod = outMod;
  initPos = rt->length;
  
  /* Emit pairs of length codes */
  while(a0 < bpl) {
    a1 = runs[rx];
    a2 = runs[rx+1];
    color = (UCHAR)rx % 2;

#if canDebug
    if(db) fprintf(db, "H, a1: %lu, a2: %lu, ", (ULONG)a1, (ULONG)a2);
#endif

    PutHLen((a1-a0), color);
    if(a1 < bpl) {
      PutHLen((a2 - a1), (1-color));
    }
    rx += 2;
    a0 = a2;
  }

  cd->outMod = outMod;
  cd->outWord = outWord;

  /* Flush partial bits to the buffer, but don't update the state */
  if(outMod != 0) {
    wp[0] = outWord;
  }
  DGP_ERR_GEN_CLEANUPF0((a0 != bpl), cErrCompressionError, "incorrect value for a0");
  if(uncompressed) {
    if((ULONG)((rt->length-initPos) * 8) > ((bpl * 4) + bpl)) {
      DGP_ERR_PASS_CLEANUPF0(UseUncompressed(cd, dim, initMod, initPos), 
        "failed writing uncompressed data");
    }
  }

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


/*
  Puts out a G3/G4 EOL code (no tag bits).
  */
ErrCode
PutEol(CompressData *cd)
{
  DGP_PROLOGUE("PutEol", cd->DGP_PERRSTATE);

  DGP_ERR_PASS_CLEANUPF0(PutBits(cd, eolCode, eolBits), "failed writing bits to output buffer");

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

/*
  Puts out a G3/G4 terminating code in the outBuffer, determined by the K
  parameter (see DoPage description).
  */
ErrCode
Terminate(CompressData *cd, Dgp_Int32 k)
{
  DGP_PROLOGUE("Terminate", cd->DGP_PERRSTATE);

  Dgp_Int32 i;
  Dgp_Int32 times = ((k < 0) ? 2 : 6);
  Dgp_Int32 tCode = ((k > 0) ? ((eolCode * 2) + 1) : eolCode);
  Dgp_Int32 tLen = ((k > 0) ? (eolBits + 1) : eolBits);
  
  for(i = 0; i < times; i++) {
    DGP_ERR_PASS_CLEANUPF0(PutBits(cd, tCode, (UCHAR)tLen), "failed writing bits to output buffer");      
  }

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


/*
  -- Emits a "page" of G3/G4 data from the given number of lines of raw bitmap.
  The pitch is the number of addressing units to add to ptr to get to the next
  line.
  -- The K parameter is consistent with the Adobe CCITTFaxDecode usage:
    -- K < 0 =>
      -- G4 coding, no tag bits
      -- If eof, then terminator is 2 EOL codes, no tag bits
    -- K = 0 =>
      -- G3 coding using only 1-dim, no tag bits
      -- If eof, then terminator is 6 EOL codes, no tag bits
    -- K > 0 =>
      -- G3 coding using mixed 1-dim/2-dim, 1 tag bit at start of line
      -- Each run of K lines will start with a 1-dim line and will be followed
      by K-1 2-dim lines.
      -- If eof, then terminator is 6 EOL codes, with tag bits = 1
  -- If eol, then a prefix EOL code will be put out at the start of each line.
  -- If uncompressed, then uncompressed coding will be used if it saves space.
  This mode will usually cause compression to be slower.  Note that many
  decoders do not support uncompressed coding, since both G3 and G4 standards
  treat it as optional.
  -- If pad, then each line is prefixed with zero bits so that the encoding for
  the line starts on a byte boundary.  The padding is inserted before EOL codes
  (if any).  The tag bit (if any) is on the byte boundary.  Note that the G3
  standard permits padding bits before the EOL code, so this approach allows
  conforming G3 decoders to ignore the padding.
  -- If flush, remaining bits in cd->outWord are flushed to output.
  */
ErrCode
DoPage(CompressData* cd, UCHAR *outBuf, ULONG *_outBufLen, ULONG *ptr,
       ULONG pitch, ULONG lines, Dgp_Int32 k, Dgp_Bool eol, Dgp_Bool eof,
       Dgp_Bool uncompressed, Dgp_Bool pad, Dgp_Bool flush)
{
  DGP_PROLOGUE("DoPage", cd->DGP_PERRSTATE);

  ULONG i, j, outBufMax = *_outBufLen, outBufLen = 0, tmp_ulong, tmpBufLen, rem = 0;
#ifdef _LITTLE_ENDIAN
  ULONG *lbuf;
#endif


  *_outBufLen = 0;

  /* Flush any accumulated stuff in the output buffer now */
  if(cd->outBuffer.length != 0) {

    /* output old data */
    tmpBufLen = cd->outBuffer.length;
    DGP_ERR_GEN_CLEANUP_NO_MSG(((tmpBufLen + outBufLen) > outBufMax), cErrBufferExceeded);
    
#ifdef _LITTLE_ENDIAN

    lbuf = (ULONG *)(cd->outBuffer.data);
    for(j = 0; j < tmpBufLen; j+=sizeof(ULONG)) {
      tmp_ulong = *(lbuf++);
#if (DGP_BITS_UINT32 == 64) 
      {
	      *(outBuf++) = (UCHAR)((tmp_ulong >> 56) & 0xFF);
	      *(outBuf++) = (UCHAR)((tmp_ulong >> 48) & 0xFF);
	      *(outBuf++) = (UCHAR)((tmp_ulong >> 40) & 0xFF);
	      *(outBuf++) = (UCHAR)((tmp_ulong >> 32) & 0xFF);
      }
#endif
      *(outBuf++) = (UCHAR)((tmp_ulong >> 24) & 0xFF);
      *(outBuf++) = (UCHAR)((tmp_ulong >> 16) & 0xFF);
      *(outBuf++) = (UCHAR)((tmp_ulong >> 8) & 0xFF);
      *(outBuf++) = (UCHAR)(tmp_ulong & 0xFF);
    }

#else

    /* copy output buffer to 'outBuf' */
    memcpy(outBuf, cd->outBuffer.data, tmpBufLen);
    outBuf += tmpBufLen;

#endif

    outBufLen += tmpBufLen;
    cd->outBuffer.length = 0;
  }

  for(i = 0; i < lines; i++) {
    DGP_ERR_PASS_CLEANUPF1(LineToTrans(cd, ptr), "could not compress line %lu", i);
    if(pad) {
      if(eol) {
	      DGP_ERR_PASS_CLEANUPF0(PutBits(cd, 0, (UCHAR)(((bpw * 2) - (cd->outMod + eolBits)) % 8)),
          "failed writing bits to output buffer");
      } else {
	      DGP_ERR_PASS_CLEANUPF0(PutBits(cd, 0, (UCHAR)((bpw - cd->outMod) % 8)),
          "failed writing bits to output buffer");
      }
    }
    if(eol) {
      DGP_ERR_PASS_CLEANUPF0(PutBits(cd, eolCode, eolBits),
        "failed writing bits to output buffer");
    }

    if(k < 0) {
      /* G4 encoding */
      DGP_ERR_PASS_CLEANUPF0(TransToG4Code(cd, uncompressed),
        "g4 encoding failed");
    } else if(k == 0) {
      /* Always use 1-dim coding, and leave off the tag bit */
      DGP_ERR_PASS_CLEANUPF0(TransToG3Code(cd, 0, uncompressed),
        "g3 (1-dim, no tag bit) encoding failed ");
    } else if(rem == 0) {
      /* This time, use 1-dim coding and reset the remainder */
      DGP_ERR_PASS_CLEANUPF0(TransToG3Code(cd, 1, uncompressed),
        "g3 (1-dim, reset rem) encoding failed ");
      rem = k - 1;
    } else {
      /* This time, use 2-dim coding */
      DGP_ERR_PASS_CLEANUPF0(TransToG3Code(cd, 2, uncompressed),
        "g3 (2-dim) encoding failed ");
      rem--;
    }

    /* output new data */
    tmpBufLen = cd->outBuffer.length;
    DGP_ERR_GEN_CLEANUP_NO_MSG(((tmpBufLen + outBufLen) > outBufMax), cErrBufferExceeded);
    
#ifdef _LITTLE_ENDIAN

    lbuf = (ULONG *)(cd->outBuffer.data);
    for(j = 0; j < tmpBufLen; j+=sizeof(ULONG)) {
      tmp_ulong = *(lbuf++);
#if (DGP_BITS_UINT32 == 64)
      {
	      *(outBuf++) = (UCHAR)((tmp_ulong >> 56) & 0xFF);
	      *(outBuf++) = (UCHAR)((tmp_ulong >> 48) & 0xFF);
	      *(outBuf++) = (UCHAR)((tmp_ulong >> 40) & 0xFF);
	      *(outBuf++) = (UCHAR)((tmp_ulong >> 32) & 0xFF);
      }
#endif
      *(outBuf++) = (UCHAR)((tmp_ulong >> 24) & 0xFF);
      *(outBuf++) = (UCHAR)((tmp_ulong >> 16) & 0xFF);
      *(outBuf++) = (UCHAR)((tmp_ulong >> 8) & 0xFF);
      *(outBuf++) = (UCHAR)(tmp_ulong & 0xFF);
    }

#else

    /* copy output buffer to 'outBuf' */
    memcpy(outBuf, cd->outBuffer.data, tmpBufLen);
    outBuf += tmpBufLen;

#endif

    outBufLen += tmpBufLen;
    cd->outBuffer.length = 0;
    ptr += pitch;
  }

  if(eof) {
    DGP_ERR_PASS_CLEANUPF0(Terminate(cd, k), "failed writing terminate code");

    flush = DGP_TRUE;
  }

  /* flush remaining bytes if desired */
  if((flush) && (cd->outMod > 0)) {
    rem = (cd->outMod + DGP_BITS_UINT8 - 1) / DGP_BITS_UINT8;
    tmp_ulong = cd->outWord;

    DGP_ERR_GEN_CLEANUP_NO_MSG(((outBufLen + rem) > outBufMax), cErrBufferExceeded);
    
    for(j = 0; j < rem; j++) {
      *(outBuf++) = (UCHAR)((tmp_ulong >> ((DGP_BITS_UINT32 - DGP_BITS_UINT8) -
					   (j * DGP_BITS_UINT8))) & 0xFF);
    }

    outBufLen += rem;

    cd->outWord = 0;
    cd->outMod = 0;
  }

  *_outBufLen = outBufLen;

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


/* for our purposes, no uncompressed data */
#define DO_UNCOMP DGP_FALSE     

/*
 * Take the input image, which is packed with MSB=left, width
 * by height bits, and each row having longsPerRow Dgp_Int32s, and encode it
 * using G4. 0 bits are white, and 1 bits are black.
 *
 * output points to the location that the result must be put into.
 * Initially, *outlen gives the size of the output buffer. At return, *outlen
 * should be set to the number of bytes actually written into the buffer.
 * Return DGP_TRUE on success, DGP_FALSE on failure. Running out of buffer space
 * is a failure.
 *
 * No EOL codes are used, either between lines or at the end. Encoded rows
 * are not padded; the only padding done is on the final byte.
 *
 * If eofb is DGP_TRUE, an EOFB (two EOL codes) is written at the end.
 */

/* Whereas g4Encode uses the passed valeu of outlen to determine the maxlength, 
 * g4Encodedoes not.  */
ErrCode 
g4Encode(Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
	 Dgp_Int32 longsPerRow, Dgp_UInt8 *output,
	 size_t *outlen, Dgp_Bool eofb, Dgp_State *dgp_state)
{
  DGP_PROLOGUE("g4Encode", DGP_PERRSTATE);

  Dgp_Int32 i;
  CompressData *cd = NULL;
  ULONG outBufMax = *outlen, outBufLen = 0, tmp_ulong, tmpBufLen, j, rem;
#ifdef _LITTLE_ENDIAN
  ULONG *lbuf;
#endif

  /* create compress data struct */
  DGP_ERR_PASS_CLEANUPF0(EncodeCreate((void **)&cd, dgp_state, NULL),
    "failed creating compress data struct");

  /* initialize the compress data struct */
  DGP_ERR_PASS_CLEANUPF0(EncodeInit(cd, width, DGP_FALSE),
    "failed initializing compress data struct");

  *outlen = 0;

  /* process each line of input image */
  for(i = 0; i < height; i++) {

    /* convert lines to transitions */
    DGP_ERR_PASS_CLEANUPF1(LineToTrans(cd, image), 
      "could not compress line %lu", i);

    /* G4 encoding */
    DGP_ERR_PASS_CLEANUPF0(TransToG4Code(cd, DO_UNCOMP),
      "g4 encoding failed");

    /* output new data */
    tmpBufLen = cd->outBuffer.length;
    DGP_ERR_GEN_CLEANUP_NO_MSG(((tmpBufLen + outBufLen) > outBufMax), cErrBufferExceeded);
    
#ifdef _LITTLE_ENDIAN

    lbuf = (ULONG *)(cd->outBuffer.data);
    for(j = 0; j < tmpBufLen; j+=sizeof(ULONG)) {
      tmp_ulong = *(lbuf++);
#if (DGP_BITS_UINT32 == 64) 
      {
	      *(output++) = (UCHAR)((tmp_ulong >> 56) & 0xFF);
	      *(output++) = (UCHAR)((tmp_ulong >> 48) & 0xFF);
	      *(output++) = (UCHAR)((tmp_ulong >> 40) & 0xFF);
	      *(output++) = (UCHAR)((tmp_ulong >> 32) & 0xFF);
      }
#endif
      *(output++) = (UCHAR)((tmp_ulong >> 24) & 0xFF);
      *(output++) = (UCHAR)((tmp_ulong >> 16) & 0xFF);
      *(output++) = (UCHAR)((tmp_ulong >> 8) & 0xFF);
      *(output++) = (UCHAR)(tmp_ulong & 0xFF);
    }

#else

    /* copy output buffer to 'output' */
    memcpy(output, cd->outBuffer.data, tmpBufLen);
    output += tmpBufLen;

#endif

    outBufLen += tmpBufLen;
    cd->outBuffer.length = 0;
    image += longsPerRow;
  }


  /* Write optional EOFB */
  if(eofb) {
    DGP_ERR_PASS_CLEANUPF0(Terminate(cd, G4_K_VAL),
      "failed writing terminate code");

    /* output new data */
    tmpBufLen = cd->outBuffer.length;
    DGP_ERR_GEN_CLEANUP_NO_MSG(((tmpBufLen + outBufLen) > outBufMax), cErrBufferExceeded);
    
#ifdef _LITTLE_ENDIAN

    lbuf = (ULONG *)(cd->outBuffer.data);
    for(j = 0; j < tmpBufLen; j+=sizeof(ULONG)) {
      tmp_ulong = *(lbuf++);
#if (DGP_BITS_UINT32 == 64) 
      {
	      *(output++) = (UCHAR)((tmp_ulong >> 56) & 0xFF);
	      *(output++) = (UCHAR)((tmp_ulong >> 48) & 0xFF);
	      *(output++) = (UCHAR)((tmp_ulong >> 40) & 0xFF);
	      *(output++) = (UCHAR)((tmp_ulong >> 32) & 0xFF);
      }
#endif
      *(output++) = (UCHAR)((tmp_ulong >> 24) & 0xFF);
      *(output++) = (UCHAR)((tmp_ulong >> 16) & 0xFF);
      *(output++) = (UCHAR)((tmp_ulong >> 8) & 0xFF);
      *(output++) = (UCHAR)(tmp_ulong & 0xFF);
    }

#else

    /* copy output buffer to 'output' */
    memcpy(output, cd->outBuffer.data, tmpBufLen);
    output += tmpBufLen;

#endif

    outBufLen += tmpBufLen;
    cd->outBuffer.length = 0;
  }

  /* flush remaining bytes if necessary */
  if(cd->outMod > 0) {
    rem = (cd->outMod + DGP_BITS_UINT8 - 1) / DGP_BITS_UINT8;
    tmp_ulong = cd->outWord;

    DGP_ERR_GEN_CLEANUP_NO_MSG(((outBufLen + rem) > outBufMax), cErrBufferExceeded);
    
    for(j = 0; j < rem; j++) {
      *(output++) = (UCHAR)((tmp_ulong >> ((DGP_BITS_UINT32 - DGP_BITS_UINT8) - (j * DGP_BITS_UINT8))) & 0xFF);
    }

    outBufLen += rem;

    cd->outWord = 0;
    cd->outMod = 0;
  }

  /* return output data size */
  *outlen = outBufLen;

  /* cleanup */
  EncodeCleanUp((void **)&cd);

  NORMAL_RETURN;

ERR_CLEANUP:

  /* cleanup */
  EncodeCleanUp((void **)&cd);

  ERR_RETURN;
}


/*
 * This function is the same as the above function, except without the
 * creation and deletion of the CompressData struct each time.
 */
ErrCode 
g4EncodeC(void *_cd, Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
	  Dgp_Int32 longsPerRow, Dgp_UInt8 *output,
	  size_t *outlen, Dgp_Bool eofb)
{
  CompressData *cd = (CompressData *)_cd;
  DGP_PROLOGUE("g4EncodeC", cd->DGP_PERRSTATE);

  Dgp_Int32 i;
  ULONG outBufMax = *outlen, outBufLen = 0, tmp_ulong, tmpBufLen, j, rem;
#ifdef _LITTLE_ENDIAN
  ULONG *lbuf;
#endif

  /* initialize the compress data struct */
  DGP_ERR_PASS_CLEANUPF0(EncodeInit(cd, width, DGP_FALSE),
    "failed initializing compress data struct");

  *outlen = 0;

  /* process each line of input image */
  for(i = 0; i < height; i++) {

    /* convert lines to transitions */
    DGP_ERR_PASS_CLEANUPF1(LineToTrans(cd, image),
      "could not compress line %lu", i);

    /* G4 encoding */
    DGP_ERR_PASS_CLEANUPF0(TransToG4Code(cd, DO_UNCOMP),
      "g4 encoding failed");

    /* output new data */
    tmpBufLen = cd->outBuffer.length;
    DGP_ERR_GEN_CLEANUP_NO_MSG(((tmpBufLen + outBufLen) > outBufMax), cErrBufferExceeded);

#ifdef _LITTLE_ENDIAN
    
    lbuf = (ULONG *)(cd->outBuffer.data);
    for(j = 0; j < tmpBufLen; j+=sizeof(ULONG)) {
      tmp_ulong = *(lbuf++);
#if (DGP_BITS_UINT32 == 64) 
      {
	      *(output++) = (UCHAR)((tmp_ulong >> 56) & 0xFF);
	      *(output++) = (UCHAR)((tmp_ulong >> 48) & 0xFF);
	      *(output++) = (UCHAR)((tmp_ulong >> 40) & 0xFF);
	      *(output++) = (UCHAR)((tmp_ulong >> 32) & 0xFF);
      }
#endif
      *(output++) = (UCHAR)((tmp_ulong >> 24) & 0xFF);
      *(output++) = (UCHAR)((tmp_ulong >> 16) & 0xFF);
      *(output++) = (UCHAR)((tmp_ulong >> 8) & 0xFF);
      *(output++) = (UCHAR)(tmp_ulong & 0xFF);
    }

#else

    /* copy output buffer to 'output' */
    memcpy(output, cd->outBuffer.data, tmpBufLen);
    output += tmpBufLen;

#endif

    outBufLen += tmpBufLen;
    cd->outBuffer.length = 0;
    image += longsPerRow;
  }


  /* Write optional EOFB */
  if(eofb) {
    DGP_ERR_PASS_CLEANUPF0(Terminate(cd, G4_K_VAL),
      "failed writing terminate code");
  }

  /* flush remaining bytes if necessary */
  if(cd->outMod > 0) {
    rem = (cd->outMod + DGP_BITS_UINT8 - 1) / DGP_BITS_UINT8;
    tmp_ulong = cd->outWord;

    DGP_ERR_GEN_CLEANUP_NO_MSG(((outBufLen + rem) > outBufMax), cErrBufferExceeded);
    
    for(j = 0; j < rem; j++) {
      *(output++) = (UCHAR)((tmp_ulong >> ((DGP_BITS_UINT32 - DGP_BITS_UINT8) -
					   (j * DGP_BITS_UINT8))) & 0xFF);
    }

    outBufLen += rem;

    cd->outWord = 0;
    cd->outMod = 0;
  }

  /* return output data size */
  *outlen = outBufLen;

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

/***************************************************************/
/* ---- ADDED BY ELI FOR CHUNK-BY-CHUNK ENCODING ---- */
/* Four functions support chunk-by-chunk encoding:  startDoc, g4EncodeChunk,
 * endDoc, and g4EncodeChunkCleanUp.
 */

/* ---- ADDED BY ELI FOR CHUNK-BY-CHUNK ENCODING ---- */
/* startDoc
 * Creates and initializes CompressData struct and output buffer to be used by
 * g4EncodeChunk.
 */
ErrCode
startDoc(void **_cd, Dgp_Int32 width, Dgp_Int32 height,
         Dgp_Int32 maxNumLines, Dgp_State *dgp_state, Dgp_Bool invertColor)
{
  DGP_PROLOGUE("startDoc", DGP_PERRSTATE);
  CompressData *cd = NULL;
  
  Dgp_Int32 outBufAllocLen;

  DGP_REPORTF0(((maxNumLines < 0) || (maxNumLines > height)), 
               "Invalid parameter:  maxNumLines"); 

  /* create compress data struct */
  DGP_ERR_PASS_CLEANUPF0(EncodeCreate((void **)&cd, dgp_state, NULL),
                         "failed creating compress data struct");

  /* initialize the compress data struct */
  DGP_ERR_PASS_CLEANUPF0(EncodeInit(cd, width, invertColor),
                         "failed initializing compress data struct");

  cd->maxNumLines = maxNumLines;

  /* allocate outBuf */
  outBufAllocLen = (cd->maxNumLines*(width)*10)/8;
      /* 10*width chosen arbitrarily. g4 codes are all 7 bits or shorter. */
  cd->chunkBuffer.data = (Dgp_UInt8 *) malloc(outBufAllocLen);

  if (cd->chunkBuffer.data) {
    DGP_ERR_CLEANUP(cErrNomem);
  }

  cd->chunkBuffer.maxLength = outBufAllocLen;
  cd->chunkBuffer.length = 0;


  *_cd = cd;

  NORMAL_RETURN;

ERR_CLEANUP:	

  ERR_RETURN;
}





/* ---- ADDED BY ELI FOR CHUNK-BY-CHUNK ENCODING ---- */
/* g4EncodeChunk
 * This function is the same as Encode and EncodeC,
 * except that this function does not create and delete the CompressData
 * struct each time it is called (as g4Encode does) and it does not
 * initialize the CompressData struct each time it is called (as
 * g4EncodeC does).
 *
 * CompressData struct will store state for chunk-by-chunk encoding.
 * cd stores transOld, the transitions data from the last line of
 * image input.  
 * Assume the image is not encoded all at once, and is
 * instead encoded in several chunks.  The first line of the next
 * chunk needs to know the transitions data of the line above it in
 * the image input, which is the last line of the previous chunk. 
 * These transitions are stored in cd->transOld.
 * Without these transitions, g4 encoding can still be done, but it
 * will result in horizontal mode encoding (essentially run-length
 * encoding, which is in most cases less compact, I think).
 */

ErrCode
g4EncodeChunk(void *_cd, Dgp_UInt32 *image, Dgp_Int32 numLines,
              Dgp_Int32 longsPerRow, Dgp_UInt8 **_output, 
              size_t *outBufUsedLen)
{
  CompressData *cd = (CompressData *)_cd;
  DGP_PROLOGUE("g4EncodeChunk", cd->DGP_PERRSTATE);

  Dgp_UInt32 outBufMaxLen;
  Dgp_Int32 i;
  ULONG tmpBufLen;
  Dgp_UInt8 *output = (Dgp_UInt8 *)NULL;
#ifdef _LITTLE_ENDIAN
  ULONG tmp_ulong;
  ULONG *lbuf;
  Dgp_UInt32 j;

#endif

  DGP_REPORTF0(((numLines< 0) || (numLines>cd->maxNumLines)), 
               "Invalid parameter:  numLines");

  /* Move output pointer to the buffer allocated by startDoc */
  output = cd->chunkBuffer.data;
  *outBufUsedLen = cd->chunkBuffer.length;  /* init to 0 in startDoc */
  outBufMaxLen = cd->chunkBuffer.maxLength;

  /* process each line of input image */
  for(i = 0; i < numLines; i++) {

    /* convert lines to transitions */
    DGP_ERR_PASS_CLEANUPF1(LineToTrans(cd, image),
      "could not compress line %lu", i);

    /* G4 encoding */
    DGP_ERR_PASS_CLEANUPF0(TransToG4Code(cd, DO_UNCOMP), "g4 encoding failed");

    /* output new data */
    tmpBufLen = cd->outBuffer.length;
    if ((tmpBufLen + (ULONG)*outBufUsedLen) > outBufMaxLen)
    {
       DGP_ERRF0(cErrBufferExceeded, "chunk-by-chunk encoding buffer too full");
    }

#ifdef _LITTLE_ENDIAN
    lbuf = (ULONG *)(cd->outBuffer.data);
    for(j = 0; j < tmpBufLen; j+=sizeof(ULONG)) {
      tmp_ulong = *(lbuf++);
#if (DGP_BITS_UINT32 == 64)
      {
              *(output++) = (UCHAR)((tmp_ulong >> 56) & 0xFF);
              *(output++) = (UCHAR)((tmp_ulong >> 48) & 0xFF);
              *(output++) = (UCHAR)((tmp_ulong >> 40) & 0xFF);
              *(output++) = (UCHAR)((tmp_ulong >> 32) & 0xFF);
      }
#endif
      *(output++) = (UCHAR)((tmp_ulong >> 24) & 0xFF);
      *(output++) = (UCHAR)((tmp_ulong >> 16) & 0xFF);
      *(output++) = (UCHAR)((tmp_ulong >> 8) & 0xFF);
      *(output++) = (UCHAR)(tmp_ulong & 0xFF);
    }

#else

    /* copy output buffer to 'output'; size checked by above if statement */
    memcpy(output, cd->outBuffer.data, tmpBufLen);
    output += tmpBufLen;

#endif

    *outBufUsedLen += tmpBufLen;
    cd->outBuffer.length = 0;
    image += longsPerRow;  /* Caller must also do this (price of control) */
  } /* for */

  *_output = cd->chunkBuffer.data; /* 'output' now pting to end of data */

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}  /* g4EncodeChunk() */




/* ---- ADDED BY ELI FOR CHUNK-BY-CHUNK ENCODING ---- */
/* endDoc */

ErrCode
endDoc(void *_cd, Dgp_Int32 longsPerRow, 
       Dgp_UInt32 *flushedBytes, Dgp_Int8 *numFlushedBytes)
{
  Dgp_Free edFree;
  CompressData *cd = (CompressData *)_cd;
  
#if eliDebug
  Dgp_Int8 *j = 0;
#endif

  DGP_PROLOGUE("endDoc", cd->DGP_PERRSTATE);
  longsPerRow = longsPerRow;
  /* flush remaining bytes if necessary */
  if(cd->outMod > 0)
  {
     DGP_ERR_GEN_CLEANUP_NO_MSG((cd->outMod > DGP_BITS_UINT32), 
                                cErrBufferExceeded);
     *flushedBytes = cd->outWord;
     *numFlushedBytes = (cd->outMod + DGP_BITS_UINT8 - 1) / DGP_BITS_UINT8; 
                        /* why -1? */
#if eliDebug
  printf("\nendDoc\tflushed:%x\tnumflushed:%d",*flushedBytes,*numFlushedBytes); 
#endif

#if eliDebug2
  printf("\tbytes:  ");
  for (j=flushedBytes; j < flushedBytes + 4; j++) printf("%d",(Dgp_Int8 *)(j));
#endif

  } /* if */

  /* Clean Up chunk buffer */
  edFree = cd->DGP_FREE;
  (edFree)(cd->chunkBuffer.data);
  EncodeCleanUp((void **)&cd);

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}




/* ---- ADDED BY ELI FOR CHUNK-BY-CHUNK ENCODING ---- */
  /* FOR TESTING ELI'S CODE ONLY */
  /* Receives a ptr to an allocated output buffer.  Allocates tmpBuf */

  /* NB:  User of g4EncodeChunk should never have to manipulate cd directly.
   * But user of g4EncodeChunk should declare and pass cd, because cd contains
   * state of the encoding process.
   */
ErrCode
TestEncodeChunk(Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
         Dgp_Int32 numLines, Dgp_Int32 longsPerRow, Dgp_UInt8 *toutput,
         size_t *toutlen, Dgp_Bool eofb, Dgp_State *dgp_state)
  /* *toutput, not **toutput because want to leave caller's pointer at beginning
   * of buffer
   */
{
  DGP_PROLOGUE("TestEncodeChunk", DGP_PERRSTATE);
  void *cd;
  Dgp_Int32 i;
  size_t ttmpBufUsedLen = 0, toutBufMaxLen;
  Dgp_UInt8 *ttmpBuf = (Dgp_UInt8 *)NULL;
  Dgp_Int8 numFlushedBytes = 0; 
  Dgp_UInt32 flushedBytes = 0;

#if eliDebug
  Dgp_UInt8 *j;
  Dgp_Int32 k;
#endif

#ifdef _LITTLE_ENDIAN
  Dgp_UInt8 *flipbuf = (Dgp_UInt8 *)NULL;
#endif


  eofb = eofb;
  toutBufMaxLen = *toutlen;
  *toutlen = 0;      /* Initialize used space counter of toutput buffer */


  /* Allocate toutput buffer, create and initialize cd*/
  DGP_ERR_PASS_CLEANUPF0(startDoc(&cd, width, height, numLines, dgp_state, 
                                  DGP_FALSE), 
                         "startDoc failed to start encoding");
  /* Should pass to startDoc the max num of lines to be encoded per call to 
   * g4EncodeChunk.  startDoc uses numLines to allocate the chunk buffer.  
   */

#if eliDebug
  printf("\nOutput: ");
#endif

  /* Encode the image in chunks numLines-lines-long */
  for(i = height; i > 0; i -= numLines) {
    if (i < numLines) 
    {
      DGP_ERR_PASS_CLEANUPF0(g4EncodeChunk(cd, image, i, longsPerRow, 
                                           &ttmpBuf, &ttmpBufUsedLen),
                             "failed to Encode Chunk");
    } else {
      DGP_ERR_PASS_CLEANUPF0(g4EncodeChunk(cd, image, numLines, longsPerRow, 
                                           &ttmpBuf, &ttmpBufUsedLen),
                             "failed to Encode Chunk");
    }  /* if */


    /* Append the encoded lines to the total output buffer */
    if ((*toutlen + ttmpBufUsedLen) >= toutBufMaxLen)
      DGP_ERRF0(cErrBufferExceeded,
                "toutput buffer too small; remainder not written.");
   
    memcpy(toutput, ttmpBuf, ttmpBufUsedLen); 
    toutput += ttmpBufUsedLen;    /* Advance output pointer */
    *toutlen += ttmpBufUsedLen;

    image += longsPerRow * numLines;  /* Advance input pointer */

#if eliDebug
    for (j = toutput - ttmpBufUsedLen; j < toutput; j++) printf("%x",*(j));
#endif
  }  /* for */

  /* eofb not supported; flush remaining bytes if necessary */
  DGP_ERR_PASS_CLEANUPF0(endDoc(cd, longsPerRow, 
                         &flushedBytes, &numFlushedBytes), 
                         "endDoc failed");

  /* Copy flushed bytes to output buffer */
  DGP_ERR_GEN_CLEANUP_NO_MSG(((*toutlen + numFlushedBytes ) > toutBufMaxLen), cErrBufferExceeded);

#ifdef _LITTLE_ENDIAN

  flipbuf = (Dgp_UInt8 *) malloc(4);

  *(flipbuf) = (UCHAR)((flushedBytes >> 24) & 0xFF);
  *(flipbuf+1) = (UCHAR)((flushedBytes >> 16) & 0xFF);
  *(flipbuf+2) = (UCHAR)((flushedBytes >> 8) & 0xFF);
  *(flipbuf+3) = (UCHAR)(flushedBytes & 0xFF);

  memcpy(toutput, flipbuf, numFlushedBytes);

  free(flipbuf);

#else

  /* copy output buffer to 'output' */
  memcpy(toutput, &flushedBytes, numFlushedBytes);

#endif

  toutput += numFlushedBytes;
  *toutlen += numFlushedBytes;

#if eliDebug
  printf("\tTestEncLns flushed bytes: ");
  for (j = toutput - numFlushedBytes; j < toutput; j++) printf("%x",*(j));
  printf("\n");
#endif

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
} /* TestEncodeChunk */

