/* $Header: /project/icons/cvsroot/digipaper/src/libg4/CCITTFaxExpand.c,v 1.19 2003/01/27 21:48:31 awilliam Exp $ */

/*
 * -- CCITTFaxExpandImpl.mesa
 *   -- Copyright (c) 1998 by Xerox Corporation.  All rights reserved.
 *   -- Russ Atkinson (RRA) October 17, 1998 10:19 pm PDT
 *
 * Translated to C by James Ahlborn in January 1999.
 */

#include <stdio.h>
#include <stdlib.h>
#include "g4.h"
#include "util.h"
#include "g4private.h"
#include "PsgFix.h"


#define COL_WHITE 0
#define COL_BLACK 1

/*
  -- This set of definitions is the interface to the CCITT group 3 (T.4) and
group 4 (T.6) fax expansion code.  For convenience, these are referred to as G3
and G4.
  --
  -- To read an understand this code you MUST have access to the G3 and G4
standards.
  --
  -- In keeping with various other pieces of code, including the Adobe
CCITTFaxDecode filter, the parameter K (ed.k) has the following interpretation:
    -- K < 0 => G4 coding
    -- K = 0 => G3 coding, pure 1-D
    -- K > 0 => G3 coding, mixed 1-D and 2-D
      -- (nominally there are at most K lines of 2-D data, but for decoding we
permit any number of lines of 2-D data, since the EOL codes have a trailing tag
bit that sets the ed.oneDimTag)
  --
  -- A comment of the following form:
    -- Q: should we exit uncompressed mode?
  -- indicates a place in the code where there is a question about the right
thing to do.
  --
  */

#define maxNeed (bpw - DGP_BITS_UINT8)

#define uncompBits1d 12
#define uncompBits2d 10
#define uncompCode   15

/*
  useDebug and innerDebug are not currently implemented in this code,
  although the cedar code for them is still included in comments.
  (in case i ever decide to add this code).

  useDebug: BOOL = DGP_FALSE;
    -- use "useDebug DGP_FALSE" for slightly higher performance
    -- use "useDebug DGP_TRUE" for debugging

  innerDebug: BOOL = DGP_FALSE;
    -- use "innerDebug DGP_FALSE" for slightly higher performance
    -- use "innerDebug DGP_TRUE" for really detailed debugging

  useFastFetch: BOOL = DGP_TRUE;
    -- use "useFastFetch 0" for slightly less (and slower) code
    -- use "useFastFetch 1" for slightly faster code

    */

#define useDebug     DGP_FALSE
#define innerDebug   DGP_FALSE
#define useFastFetch 1
#define debugFile NULL

static TabEntry whiteLongRunTab[runSplit];
static TabEntry whiteShortRunTab[runSplit];
static TabEntry blackLongRunTab[runSplit];
static TabEntry blackShortRunTab[runSplit];

/*
  colorNames: ARRAY Color OF ROPE = ["white", "black"];

*/
/*static Dgp_Int8 *colorNames[2] = {"white", "black"};*/

static TabEntry *runTabs[4] = {whiteShortRunTab, blackShortRunTab,
			       whiteLongRunTab, blackLongRunTab};

/*
  -- When examining x, the value of the current 13 leading bits, if the
  leading 4 bits are zeros (=> x >= 512) then we use the index directly into
  the "Dgp_Int32" table.  Otherwise we divide by 2**4 (= 16), add the mod, and
  index the Dgp_Int16 table.
    */
#define runTableBits  13
#define runTableZeros  4
#define runTableMod ((Dgp_Int32)1 << (runTableBits-runTableZeros))
#define runTableDiv ((Dgp_Int32)1 << runTableZeros)
#define runTableSplit 64

#define MAX_BYTE 256

static UCHAR firstOneTab[MAX_BYTE];


/*
  Returns the run-to-code table for the given color and length selection.
  */
static TabEntry *
GetCodeTab(Dgp_Int32 color, Dgp_Int32 which)
{
  if(color == TAB_COL_WHITE) {
    if(which == TAB_LEN_SHORT) {
      return whiteShortCodeTab;
    } else if(which == TAB_LEN_LONG) {
      return whiteLongCodeTab;
    }
  } else if(color == TAB_COL_BLACK) {
    if(which == TAB_LEN_SHORT) {
      return blackShortCodeTab;
    } else if(which == TAB_LEN_LONG) {
      return blackLongCodeTab;
    }
  }

  return NULL;
}


static void
GenRunTab(Dgp_Int32 color, Dgp_Int32 which, TabEntry *new_arr)
{
  Dgp_Int32 cw, i;
  TabEntry *ct = NULL, *ent;
  Dgp_Int32 bits, shift;
  Dgp_UInt32 rem, base, len;

  for(i = 0; i < runSplit; i++) {
    new_arr[i].val = 0;
    new_arr[i].len = 0;
  }

  for(cw = TAB_LEN_SHORT; cw <= TAB_LEN_LONG; cw++) {
    ct = GetCodeTab(color, cw);
    for(i = 0; i < codeSplit; i++) {
      ent = ct + i;
      bits = ent->len;
      shift = maxCodeBits - bits;
      rem = 1 << shift;
      base = ent->val << shift;
      len = i;
      
      if(cw == TAB_LEN_LONG) len = i * codeSplit;

      if(base == 0) {
	/* do nothing */
      } else if((base >= runSplit) && (which == TAB_LEN_SHORT)) {
	base /= 16;
	rem /= 16;
	while(rem != 0) {
	  rem--;
	  new_arr[base + rem].val = (USHORT)len;
	  new_arr[base + rem].len = (USHORT)bits;
	}
      } else if((base < runSplit) && (which == TAB_LEN_LONG)) {
	while(rem != 0) {
	  rem--;
	  new_arr[base+rem].val = (USHORT)len;
	  new_arr[base+rem].len = (USHORT)bits;
	}
      }
    }
  }
}


/* generate the run tables.  this only needs to be called once per run of the
   program */
void
InitRunTabs()
{
  GenRunTab(TAB_COL_WHITE, TAB_LEN_SHORT, whiteShortRunTab);
  GenRunTab(TAB_COL_WHITE, TAB_LEN_LONG, whiteLongRunTab);
  GenRunTab(TAB_COL_BLACK, TAB_LEN_SHORT, blackShortRunTab);
  GenRunTab(TAB_COL_BLACK, TAB_LEN_LONG, blackLongRunTab);
}


/* generate the first one table.  this only needs to be called once per
   run of the program */
void
InitFirstOneTab()
{
  Dgp_Int32 b = 1;
  UCHAR bb, c;
  firstOneTab[0] = 8;

  while(DGP_TRUE) {
    bb = (UCHAR)b;
    c = 0;
    while(bb < 128) {
      bb += bb;
      c++;
    }
    firstOneTab[b] = c;
    if(b == 255) break;
    b++;
  }
}


static UCHAR
FindFirstOne(ULONG w)
{
  ULONG ret = 0;
  UCHAR b;

  if(w == 0) return(bpw);

  while(DGP_TRUE) {
    b = (UCHAR)(w / ((Dgp_Int32)1 << (bpw - 8)));
    if(b != 0) return((UCHAR)((ret + firstOneTab[b]) % bpw));

    ret += 8;
    w *= 256;
  }
}


/*
  This information is for keeping statistics, it's not currently implemented.

  keepStats: BOOL = DGP_FALSE;
  stats: Stats := ALL[0];

  statsIndexToName: ARRAY StatsIndex OF ATOM := [
    $nLines,
      $mgb, $g4Fast,
    $g4Entries,
      $g4V0, $g4V1, $g4V2, $g4V3, $g4H, $g4P, $g4U,
    $g3Entries,
      $g3H,
    $uncompEntries,
    $nErrors
    ];

  BumpStat: PROC [which: StatsIndex] = INLINE {
    IF keepStats THEN stats[which] := stats[which] + 1;
    };
-- Exported procedures
--
  GetStats: PUBLIC PROC RETURNS [Stats] = {
    RETURN [stats];
    };

  StatsIndexToName: PUBLIC PROC [index: StatsIndex] RETURNS [ATOM] = {
    RETURN [statsIndexToName[index]];
    };
    */

#define keepStats DGP_FALSE



/* Utility procedures & macros */

#define InvertColor(c) (1-(c))
 
#define GetRunEntry(color, x)						      \
 (((x) >= runTableMod) ? (&runTabs[color][((x) / runTableDiv) % runTableMod]) \
  : (&runTabs[2+(color)][(x) % runTableMod]))

/* be careful, the argument 'x' is used twice (not good if x invloves
   some computation) */
#define ClipZero(x) ((Dgp_Int32)(((ULONG)(x)) & ((((ULONG)(x)) >> (bpw-1)) - 1)))

#define GetTransition(trans_s, i) ((trans_s)[(i)])


#define SetTransition(lineTransitions, i, j) 	\
 do {(((lineTransitions)->s)[(i)]) = (j);} while(0)


/*
  These functions are used in the debugging, not currently implemented.

  PutGoodBits: PROC [st: STREAM, w: WORD, bits: NAT] = {
    THROUGH [0..MIN[bits, bpw]) DO
      IO.PutChar[st, '0+Basics.BITRSHIFT[w, bpw-1]];
      w := w + w;
      ENDLOOP;
    };

  PutB: PROC [st: STREAM, w: WORD, bitCount: Dgp_Int32, nl: BOOL := DGP_FALSE] = {
    IF st # NIL THEN {
      WHILE bitCount > 0 DO
        shift: BitPos = LOOPHOLE[bitCount-1, WORD] MOD bpw;
        c: CHAR := '0 + (Basics.BITRSHIFT[w, shift] MOD 2);
        IO.PutChar[st, c];
        bitCount := shift;
        ENDLOOP;
      IF nl THEN IO.PutChar[st, '\n];
      };
    };
  LogError: PROC [data: ExpandData, error: ATOM, bufferedBits: NAT] = {
    IF data.error = NIL THEN {
      data.error := error;
      IF data.debug # NIL THEN
        IO.PutF1[data.debug, "\nLogError: %g\n", [atom[error]]];
      };
    };

  ErrorEOF: PROC [data: ExpandData] = {
    -- Called on end-of-stream, which should never happen with valid data.
    IF data.error = NIL THEN data.error := $eof;
    data.end := DGP_TRUE;
    BumpStat[nErrors];
    };
*/


/*
  Sets up the expander with a given input stream, k, bpl
  -- DecodeInit must be called prior to calling NextLine or ConsumeEndOfBlock.
  */
ErrCode
DecodeInit(ExpandData *ed, UCHAR *in, ULONG inLen, ULONG bpl, Dgp_Int32 k)
{
  DGP_PROLOGUE("DecodeInit", ed->DGP_PERRSTATE);
  Dgp_Allocators *a = &(ed->DGP_ALLOCATORS);
  ULONG tSize = bpl+4;
  IndexSequence *ref = ed->referenceTransitions;
  IndexSequence *line = ed->lineTransitions;

  if(ref->size < tSize) {
    if(ref->s) (a->free)(ref->s);

    DGP_CHECK_ALLOC_CLEANUP((ref->s = (Dgp_Int32 *)(a->malloc)(tSize * sizeof(*(ref->s)) )));
    ref->size = tSize;

    if(line->s) (a->free)(line->s);

    DGP_CHECK_ALLOC_CLEANUP(line->s = (Dgp_Int32 *)(a->malloc)(tSize * sizeof(*(line->s))));
    line->size = tSize;
  }

  ed->bpl = bpl;
  ed->k = k;
  ed->referenceTransitions = ref;
  ed->lineTransitions = line;
  ed->in = in;
  ed->inLength = inLen;
  ed->inIndex = 0;
  ed->bitBuffer = 0;
  ed->goodBits = 0;
  ed->uncompressed = DGP_FALSE;
  ed->color = COL_WHITE;
  ed->end = DGP_FALSE;

  ed->oneDimTag = (k == 0);

  ref->s[0] = -1;
  ref->s[1] = bpl;
  ref->s[2] = bpl;
  ref->end = 3;

  line->s[0] = -1;
  line->s[1] = bpl;
  line->s[2] = bpl;
  line->end = 3;

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


/* clean up the expand data struct */
void
DecodeCleanUp(void **_ed)
{
  ExpandData *ed;
  Dgp_Free tmp_free;
  IndexSequence *lt;

  if(_ed == NULL) return;

  ed = *_ed;
  *_ed = NULL;

  if(ed == NULL) return;

  tmp_free = ed->DGP_FREE;

  lt = ed->referenceTransitions;
  if(lt) {
    if(lt->s) (tmp_free)(lt->s);
    lt->s = NULL;
    (tmp_free)(lt);
  }
  ed->referenceTransitions = NULL;

  lt = ed->lineTransitions;
  if(lt) {
    if(lt->s) (tmp_free)(lt->s);
    lt->s = NULL;
    (tmp_free)(lt);
  }
  ed->lineTransitions = NULL;
  
  (tmp_free)(ed);
}


/* 
   Creates a new handle for expansion.
   -- DecodeInit must be called to process lines.
   */
ErrCode
DecodeCreate(void **_ed, Dgp_State *dgp_state, FILE *db_out)
{
  DGP_PROLOGUE("DecodeCreate", DGP_PERRSTATE);
  ExpandData *ed = NULL;
  static Dgp_Bool init_tabs = DGP_FALSE;
  Dgp_Allocators *a = &(dgp_state->a);

  *_ed = NULL;

  /* initialize tables if that hasn't been done yet */
  if(!init_tabs) {
    InitRunTabs();
    InitFirstOneTab();
    init_tabs = DGP_TRUE;
  }

  DGP_CHECK_ALLOC_CLEANUP(ed = (ExpandData *)(a->calloc)(1, sizeof(*ed)));

  DGP_CHECK_ALLOC_CLEANUP(ed->referenceTransitions = 
       (IndexSequence *)(a->calloc)(1, sizeof(*(ed->referenceTransitions))));

  DGP_CHECK_ALLOC_CLEANUP(ed->lineTransitions = (IndexSequence *)(a->calloc)(1, 
      sizeof(*(ed->lineTransitions))));

  ed->k = -1;

  ed->dgp_state = dgp_state;
  ed->debug = db_out;

  *_ed = ed;

  NORMAL_RETURN;

ERR_CLEANUP:

  DecodeCleanUp((void **)&ed);

  ERR_RETURN;
}


#define EatBits1(n) do {       			\
  goodBits = (UCHAR)(goodBits - (n));		\
  bitBuffer <<= (n);				\
} while(0)

#if useFastFetch

#define PullBits() do {							  \
  index = ed->inIndex;							  \
  if((index+sizeof(ULONG)) <= length) {					  \
    while(DGP_TRUE) {							  \
      goodBits += 8;							  \
      bitBuffer = bitBuffer + (in[index++] << ((UCHAR)(bpw - goodBits))); \
      if(goodBits >= maxNeed) break;					  \
    }									  \
  } else {								  \
    UCHAR _b;								  \
    while(goodBits < maxNeed) {						  \
      _b = ((index < length) ? (in[index++]) : 0);			  \
      goodBits = (UCHAR)(goodBits + DGP_BITS_UINT8);			  \
      bitBuffer += _b << ((UCHAR)(bpw - goodBits));			  \
    }									  \
    ed->goodBits = goodBits;						  \
  }									  \
  ed->inIndex = index;							  \
} while(0)

#else

#define PullBits() do {					\
  UCHAR _b;						\
  index = ed->inIndex;					\
  while(goodBits < maxNeed) {				\
    _b = ((index < length) ? (in[index++]) : 0);	\
    goodBits = (UCHAR)(goodBits + DGP_BITS_UINT8);		\
    bitBuffer += _b << ((UCHAR)(bpw - goodBits));	\
  }							\
  ed->goodBits = goodBits;				\
  ed->inIndex = index;					\
} while(0)

#endif



#define Peek(n) ((ULONG)(bitBuffer >> ((UCHAR)(bpw - (n)))))


/*
  Consumes the EOB (end of block) pattern (if any).  The EOB pattern is
  nominally 2 EOL codes for G4 and 6 EOL codes for G3.  Returns the number o
  EOL codes actually consumed.
  */
Dgp_Int32
ConsumeEndOfBlock(ExpandData *ed)
{
  Dgp_Int32 eolSeen = 0, repeat, i;
  ULONG bitBuffer, index, length;
  UCHAR goodBits, needBits, b;
  UCHAR *in = ed->in;

  if((in != NULL) && !(ed->end)) {
    /* buffered input bits */
    bitBuffer = ed->bitBuffer;
    /* number of good bits, left justified in bitBuffer */
    goodBits = ed->goodBits;
    needBits = eolBits;
    repeat = 2;

    if(ed->k >= 0) {
      /* G3 coding requires 6 EOL codes for an EOF */
      repeat = 6;

      /* if k > 0 then we also need a tag bit */
      if(ed->k > 0) {
	needBits++;
      }
    }

    index = ed->inIndex;
    length = ed->inLength;

    for(i = 0; i < repeat; i++) {
      while((goodBits < needBits) && (bitBuffer == 0)) {
	b = ((index < length) ? (in[index++]) : 0);

	goodBits += 8;
	bitBuffer += b << (bpw - goodBits);
      }
      
      if(goodBits < needBits) break;
      /*  No EOL code, so exit now */
      if(Peek(eolBits) != eolCode) break;

      eolSeen++;
      bitBuffer <<= eolBits;
      
      if(needBits != eolBits) {
	/* For G3 ed with K > 0 there is an extra bit */
	ed->oneDimTag = (Dgp_Bool)(bitBuffer / ((Dgp_Int32)1 << (bpw-1)));
	bitBuffer += bitBuffer;
      }

      goodBits -= needBits;
    }

    ed->inIndex = index;

    ed->bitBuffer = bitBuffer;
    ed->goodBits = goodBits;
  }

  return(eolSeen);
}


/*
  DGP_TRUE if an EOB was seen or the input stream has no more bits
*/
ErrCode
EndSeen(ExpandData *ed)
{
  return(((ed->bitBuffer == 0) && (ed->inIndex >= ed->inLength)) || (ed->end));
}


/*
  Uses the results of the most recent call to NextLine to inclusive OR
  bits in the range of [min..max], where the source bit labelled with min
  corresponds to the dst bit indexed by dstBitIndex.
  */
/*
17.8.2000 / dhtesq.
Decoder profiling suggests this is a good place for an inline...
*/

INLINE void
OrBltLine(ExpandData *ed, ULONG *dstBase, ULONG dstBitIndex, ULONG min,
	  ULONG max)
{
  ULONG *rp, *p0, lim, dstStart, *p, w, words, dstLim, start, fill;
  UCHAR c0, dstMod;
  

  if(min < max) {
    rp = (ULONG *)((ed->referenceTransitions->s) + 1);
    fill = WORD_LAST;
    p0 = dstBase + (dstBitIndex / bpw);
    c0 = (UCHAR)(dstBitIndex % bpw);

    /* toss leading runs */
    while(rp[1] <= min) rp += 2;
    
    /* clip leading visible run */
    start = MAX(rp[0], min);
    
    while(start < max) {
      lim = MIN(rp[1], max);

      if(lim > start) {
	dstStart = (start-min) + c0;
	p = p0 + (dstStart / bpw);
	dstMod = (UCHAR)(dstStart % bpw);
	dstLim = dstMod + (lim - start);
	w = fill >> dstMod;
	words = dstLim / bpw;

	if(words != 0) {
	  *(p++) |= w;
	  w = fill;

	  if(words > 1) {
	    words--;
	    while(words >= 4) {
	      p[0] = w;
	      p[1] = w;
	      p[2] = w;
	      p[3] = w;
	      p += 4;
	      words -= 4;
	    }

	    if(words == 0) goto wordless;
	    *(p++) = w;
	    if(words == 1) goto wordless;
	    *(p++) = w;
	    if(words == 2) goto wordless;
	    *(p++) = w;
	    
	  }
	
	wordless:
	  
	  if((dstLim % bpw) == 0) goto noRem;
	}

	w -= (fill >> (dstLim % bpw));
	p[0] |= w;
	
      }

    noRem:

      rp += 2;
      start = rp[0];
    }
  }
}

/*
  Uses the results of the most recent call NextLine to to move bits in the
  range of [min..max], where the source bit labelled with min corresponds to
  the dst bit indexed by dstBitIndex.  If invert, then the sense of the bits
  in the source is inverted.
*/
void
MoveLine(ExpandData *ed, ULONG *dstBase, ULONG dstBitIndex, ULONG min,
	 ULONG max, Dgp_Bool invert)
{
  ULONG *rp, *p, w, fill, newMin, outLim, words, mask;
  UCHAR shift;

  if(min < max) {
    rp = (ULONG *)((ed->referenceTransitions->s) + 1);
    shift = (UCHAR)(dstBitIndex % bpw);
    p = dstBase + (dstBitIndex / bpw);
    w = 0;
    fill = (invert ? WORD_LAST : 0);

    while(rp[0] <= min) {
      rp++;
      fill = ~fill;
    }

    if(shift != 0) {
      /* Load w with the initial bits */
      w = p[0];
      w -= ((w << shift) >> shift);
    }

    while(min < max) {
      newMin = MIN(max, rp[0]);
      outLim = (newMin - min) + shift;
      words = outLim / bpw;
      w += fill >> shift;

      if(words != 0) {
	*(p++) = w;
	w = fill;

	if(words > 1) {
	  words--;

	  while(words >= 4) {
	    p[0] = w;
	    p[1] = w;
	    p[2] = w;
	    p[3] = w;
	    p += 4;
	    words -= 4;
	  }

	  if(words == 0) goto wordless;
	  *(p++) = w;
	  if(words == 1) goto wordless;
	  *(p++) = w;
	  if(words == 2) goto wordless;
	  *(p++) = w;
	  
	}
      }

    wordless:
      
      shift = (UCHAR)(outLim % bpw);
      w -= fill >> shift;
      rp++;
      min = newMin;
      fill = ~fill;
    }

    if(shift != 0) {
      /* Take care of the trailing bits */
      mask = WORD_LAST >> shift;
      w -= (w & mask);
      p[0] = w + (p[0] & mask);
    }
  }
}
	

/* expand uncompressed data */
static ErrCode
Uncompressed(ExpandData *ed, Dgp_Int32 j, ULONG bitBuffer, UCHAR goodBits,
	     Dgp_Int32 *ret)
{
  DGP_PROLOGUE("Uncompressed", ed->DGP_PERRSTATE);

  Dgp_Int32 bpl = ed->bpl;
  UCHAR color = ed->color;
  ULONG pk = 0, index, length = ed->inLength;
  UCHAR nb, nw;
  UCHAR *in = ed->in;

  
#define Flip() do {						\
  color = InvertColor(color);					\
  if(j < bpl) {						\
    IndexSequence *_lineTransitions = ed->lineTransitions;	\
    ULONG _end = _lineTransitions->end;				\
    								\
    DGP_ERR_GEN_CLEANUPF0((_end >= _lineTransitions->size),	cErrCompressionError, \
      "error in line transitions");			\
    								\
    SetTransition(_lineTransitions, _end, j);			\
    _lineTransitions->end = _end + 1;				\
								\
    /*if((innerDebug) && (ed->debug != NULL)) {			\
        IF goodBits < eolBits THEN PullBits[];		\
          IO.PutRope[debug, "Uncompressed flip, "];		\
          IO.PutF1[debug, "j: %g, ", [integer[j]]];		\
          PutGoodBits[debug, bitBuffer, goodBits];		\
          IO.PutChar[debug, '\n]; 				\
    }*/								\
  }								\
} while(0)

  /*Table of uncompressed codes
      -- code	bits	notes
      -- 1	1
      -- 01	01
      -- 001	001
      -- 0001	0001
      -- 00001	00001
      -- 000001	00000
      -- 0000001T		exit, T is new color
      -- 00000001T	0	exit, T is new color
      -- 000000001T	00	exit, T is new color
      -- 0000000001T	000	exit, T is new color
      -- 00000000001T	0000	exit, T is new color
      */
  /*IF useDebug AND debug # NIL THEN {
      IF goodBits < eolBits THEN PullBits[];
      IO.PutRope[debug, "Uncompressed entry, "];
      PutGoodBits[debug, bitBuffer, goodBits];
      IO.PutChar[debug, '\n];
      };
    BumpStat[uncompEntries];*/

  j = ClipZero(j);
  while(DGP_TRUE) {
    /* longest code has 12 bits, same as EOL length */
    if(goodBits < eolBits) {
      PullBits();
      continue;
    }

    pk = Peek(8);
    if(pk < 127) {
      /* we have some black bits, which we can consume up to 8 bits at a time*/
      nb = firstOneTab[255 ^ pk];
      if((j + nb) > bpl) nb = (UCHAR)(bpl - j);
      
      EatBits1(nb);

      /* flip to black if we are in white */
      if((nb != 0) && (color == COL_WHITE)) {
	Flip();
      }

      j += nb;
      
      continue;
    }

    if(pk > 3) {
      /* we have some white bits, but we are NOT exiting uncompressed mode
         -- (codes 01, 001, 0001, 00001, 000001) */
      nw = firstOneTab[pk]; /* # of white bits */
      
      if((j + nw) > bpl) nw = (UCHAR)(bpl - j);
      if((nw != 0) && (color == COL_BLACK)) {
	Flip();
      }
      j += nw;

      if(nw == 5) {
	EatBits1(6);
      } else {
	EatBits1(nw);
      }
      
      if(j == bpl) break;

      continue;
    }
    
    pk = Peek(eolBits);

    if(pk == eolCode) {
      /* EOL code, leave it in the stream
	 -- Q: should we exit uncompressed mode? Most likely a coding error. */
      /* LogError[ed, $coding, goodBits]; */
      ed->uncompressed = DGP_FALSE;
      break;
    } else if(pk == 0) {
      /* coding error, too many zeros
	 LogError[ed, $coding, goodBits]; */
      break;
    } else {
      /* try to exit uncompressed mode
            -- (codes 0000001T, 00000001T, 000000001T, 0000000001T,
	        00000000001T) */
      nw = firstOneTab[pk*4]; /* # of white bits */
      
      if((j + nw) > bpl) {
	/* continued over a line, so we just eat a few bits */
	nw = (UCHAR)(bpl - j);
	EatBits1(nw);
	if((nw != 0) && (color == COL_BLACK)) {
	  Flip();
	}
	  
	j = bpl;
	break;
      }

      /* flip to black if we are in white */
      if(color == COL_BLACK) Flip();
      
      EatBits1(7+nw);

      color = (UCHAR)(Peek(1));
      EatBits1(1);
      ed->uncompressed = DGP_FALSE;
    }
  }

  ed->color = color;
  ed->bitBuffer = bitBuffer;
  ed->goodBits = goodBits;
  
  /* IF useDebug AND debug # NIL THEN {
      IF goodBits < eolBits THEN PullBits[];
      IO.PutRope[debug, "Uncompressed exit, "];
      PutGoodBits[debug, bitBuffer, goodBits];
      IO.PutChar[debug, '\n];
      }; */

  *ret = j;

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


ErrCode
FastScan1D(ExpandData *ed, Dgp_Int32 j, ULONG bitBuffer, UCHAR goodBits, Dgp_Int32 *ret)
{
  DGP_PROLOGUE("FastScan1D", ed->DGP_PERRSTATE);

  Dgp_Int32 nj = ClipZero(j);
  UCHAR needBits = maxNeed;
  Dgp_Int32 bpl = ed->bpl;
  UCHAR color = ed->color, shift;
  ULONG x, padding, index = ed->inIndex, length = ed->inLength;
  TabEntry *e;
  Dgp_Int32 v;
  UCHAR *in = ed->in;

#define Fill1D() do {							     \
  Dgp_Int32 _jb = ClipZero(j);						     \
  IndexSequence *_lineTransitions = ed->lineTransitions;		     \
  ULONG _end = _lineTransitions->end;					     \
  if((nj > _jb) && (color == (UCHAR)(_end % 2))) {	       		     \
    DGP_ERR_GEN_CLEANUPF0((_end >= _lineTransitions->size), cErrCompressionError, \
      "error in line transitions");				     \
    /* IF innerDebug AND ed.debug # NIL THEN {				     \
       IO.PutF1[ed.debug, "Fast scan emit transition: %g, ", [integer[jb]]]; \
       IO.PutRope[ed.debug, colorNames[color]];				     \
       IO.PutChar[ed.debug, '\n];					     \
       }; */								     \
    SetTransition(_lineTransitions, _end, _jb);				     \
    _lineTransitions->end = _end + 1;					     \
  }									     \
  j = nj;								     \
  } while(0)

#define EatBits() do {				\
    goodBits = (UCHAR)(goodBits - needBits);	\
    bitBuffer <<= needBits;			\
  } while(0)

  /*IF useDebug AND debug # NIL THEN {
    IF goodBits < runTableBits THEN PullBits[];
    IO.PutRope[debug, "FastScan1D entry, "];
    PutGoodBits[debug, bitBuffer, goodBits];
    IO.PutChar[debug, '\n];
    };
    BumpStat[g3Entries];*/

  while(j < bpl) {
    if(goodBits < runTableBits) PullBits();

    /* Check for a non-run */
    if(Peek(8) == 0) break;

    nj = ClipZero(j);

    while(DGP_TRUE) {
      x = bitBuffer >> (bpw-runTableBits);
      e = GetRunEntry(color, x);
      v = (Dgp_Int32)e->val;
      needBits = (UCHAR)(e->len);
      
      if(needBits == 0) break;

      /*BumpStat[g3H];*/
      EatBits();
      nj += v;
      if(v < runTableSplit) break;
      if(goodBits < runTableBits) PullBits();
    }

    if(nj > bpl) {
      /* LogError[ed, $Dgp_Int32scan, goodBits]; */
      nj = bpl;
    }

    /* IF innerDebug AND debug # NIL THEN {
       IO.PutF1[debug, "FastScan1D fill: %g, ", [integer[nj]] ];
       IO.PutRope[debug, colorNames[color]];
       IO.PutChar[debug, '\n];
       }; */

    /* nj, color; j := nj */
    Fill1D();
  }

  if(j > bpl) {
    /* LogError[ed, $Dgp_Int32scan, goodBits]; */
    j = bpl;
  }

  if(goodBits < MAX(eolBits, uncompBits1d)) PullBits();

  if((j < bpl) && (Peek(uncompBits1d) == uncompCode)) {
    /* switching to uncompressed during a 1-D encoded line */
    ed->uncompressed = DGP_TRUE;
  } else if(Peek(eolBits) == 0) {
    /* We have fill bits that were put in to accomodate a minimum
       transmission time.  The idea here is to scan for an EOL and eat up
       zero bits relatively quickly.  We have to leave the EOL in the stream
       to be able to handle further state transtitions between 1-D and 2-D
       encoding. */
    padding = 0;
    
    while(DGP_TRUE) {
      if(goodBits < eolBits) PullBits();
      if(bitBuffer != 0) {
	/* At this point we definitely have an EOL (or an encoding error)
	   in the bitBuffer. */
	shift = FindFirstOne(bitBuffer) + 1; /*places to shift to consume EOL*/
	if(shift >= eolBits) {
	  /* We have some final zero bits to consume; also consume the EOL */
	  needBits = shift - eolBits;
	  padding += needBits;
	  EatBits();
	}
	break;
      }
      
      /* Discard leading zero bits, but leave enough to find EOL */
      if(length == index) break;
      needBits = goodBits - (eolBits - 1);
      padding += needBits;
      EatBits();
    }

    /*IF innerDebug AND debug # NIL THEN {
      IO.PutF1[debug, "FastScan1D pad: %g\n", [cardinal[padding]] ];
      }; */
  }

  /* IF useDebug AND debug # NIL THEN {
     IO.PutF1[debug, "FastScan1D exit: %g, ", [integer[j]] ];
     PutGoodBits[debug, bitBuffer, goodBits];
     IO.PutChar[debug, '\n];
     }; */

  ed->bitBuffer = bitBuffer;
  ed->goodBits = goodBits;

  *ret = j;

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


/*
17.8.2000 / dhtesq.
Decoder profiling suggests this is a good place for an inline...
*/
static INLINE void
FastScan2D(ExpandData *ed, Dgp_Int32 j, ULONG bitBuffer, UCHAR goodBits, Dgp_Int32 *ret)
{
  /* Table of 2D codes
      -- code	action
      -- 1	vertical, d=0
      -- 010	vertical, d=-1
      -- 011	vertical, d=+1
      -- 001	horizontal
      -- 0001	pass
      -- 000010	vertical, d=-2
      -- 000011	vertical, d=+2
      -- 0000010	vertical, d=-3
      -- 0000011	vertical, d=+3 */
  /* # of bits needed to determine mode (pass, horizontal, vertical, or
     orther) */
  Dgp_Int32 fastScanBits = 7;
  UCHAR color = ed->color;
  Dgp_Int32 nj = 0, d, v;
  ULONG ri = 0, ni, x;
  UCHAR needBits = maxNeed;
  Dgp_Int32 bpl = ed->bpl;
  IndexSequence *ref = ed->referenceTransitions;
  ULONG riLim = ref->end - 2, index, length = ed->inLength;
  Dgp_Int32 *ref_s = ref->s;
  UCHAR startColor;
  TabEntry *e;
  Dgp_Bool bit;
  UCHAR *in = ed->in;


#define Fill2D() do {							      \
  if(nj < bpl) {							      \
    IndexSequence *_lineTransitions = ed->lineTransitions;		      \
    ULONG _end = _lineTransitions->end;					      \
    /* IF innerDebug AND ed.debug # NIL THEN {				      \
       IO.PutF[ed.debug,", %g: %g", [rope[colorNames[color]]],[integer[nj]]]; \
       }; */								      \
    if(_end < _lineTransitions->size) {					      \
      SetTransition(_lineTransitions, _end, nj);			      \
      _lineTransitions->end = _end + 1;					      \
    }									      \
  }									      \
  color = InvertColor(color);						      \
  } while(0)

  /* BumpStat[g4Entries]; */

  /* Perform the initial scan to get ri on the right index */
  ri += (1 + color + ri) % 2;

  while(((nj = GetTransition(ref_s, ri)) <= j) && (ri < riLim)) ri +=2;

  if(ri > riLim) ri = riLim;

  j = ClipZero(j);

  if(goodBits < fastScanBits) PullBits();
  /* IF useDebug AND ed.debug # NIL THEN {
     IO.PutF[ed.debug, "FastScan2D entry: j=%g, nj=%g, ri=%g, ",
     [integer[j]], [integer[nj]], [integer[ri]] ];
     PutGoodBits[ed.debug, bitBuffer, goodBits];
     IO.PutChar[ed.debug, '\n];
     }; */

  while(DGP_TRUE) {
    if(((Dgp_Int32)(bitBuffer)) < 0) {

      /* Vertical, d = 0; special case!  Very common (over 50% of the time) */
      /* IF innerDebug AND ed.debug # NIL THEN {
	 IO.PutRope[ed.debug, "FastScan2D V 0, bits: 1"];
	 };
	 BumpStat[g4V0]; */
      EatBits1(1);
      Fill2D(); /* nj, color */
      j = nj;
      /* IF innerDebug AND ed.debug # NIL THEN {
	 IO.PutF1[ed.debug, ", j: %g\n", [integer[j]]];
	 }; */
      if(j >= bpl) break;

      /* Only need to advance to next transition, since the next transition
	 is known to be of the opposite color already, and transition
	 sequences must be increasing. */
      ri++;
      if(ri > riLim) ri = riLim;
      nj = GetTransition(ref_s, ri);
      continue;

    } else if(goodBits < fastScanBits) {

      /* May need more bits, can test this after the first test because
	 the first test only needs a single 1 bit to win. */
      goto pull;

    } else if(((Dgp_Int32)(bitBuffer + bitBuffer)) < 0) {

      /* Vertical, ABS[d] = 1; fairly common */
      /* BumpStat[g4V1]; */
      d = 1;
      needBits = 3;
      goto scanV;

    } else if(((Dgp_Int32)(bitBuffer * 4)) < 0) {

      /* Horizontal case has two transitions encoded */
      startColor = color;
      /* IF innerDebug AND ed.debug # NIL THEN {
	 IO.PutRope[ed.debug, "FastScan2D H, bits: 001"];
	 };
	 BumpStat[g4H]; */

      EatBits1(3);
      nj = j;
      while(DGP_TRUE) {
	while(DGP_TRUE) {
	  if(goodBits < runTableBits) PullBits();
	  
	  x = Peek(runTableBits);
	  e = GetRunEntry(color, x);
	  v = (Dgp_Int32)e->val;
	  needBits = (UCHAR)(e->len);
	  /* IF innerDebug AND ed.debug # NIL THEN {
	     IO.PutF1[ed.debug, ", v: %g", [integer[v]] ];
	     IO.PutRope[ed.debug, ", bits: "];
	     PutGoodBits[ed.debug, bitBuffer, needBits];
	     }; */
	  if(needBits == 0) goto coding;
	  goodBits = (UCHAR)(goodBits - needBits);
	  bitBuffer <<= needBits;
	  nj += v;

	  if(v < runTableSplit) break;
	}
	if(nj > bpl) {
	  /* LogError[ed, $Dgp_Int32scan, goodBits]; */
	  nj = bpl;
	}
	Fill2D(); /* nj, color */
	if(color == startColor) break;
      }
      ni = ri;
      /* Flow through to common scanning code */

    } else if(((Dgp_Int32)(bitBuffer * 8)) < 0) {

      /* Pass, skip over 2 reference transitions */
      j = GetTransition(ref_s, ri + 1);
      /* BumpStat[g4P]; */
      EatBits1(4);
      ri += 2;
      
      if(ri > riLim) ri = riLim;
      nj = GetTransition(ref_s, ri);
      
      /*IF innerDebug AND ed.debug # NIL THEN {
	IO.PutRope[ed.debug, "FastScan2D P, bits: 0001"];
	IO.PutF1[ed.debug, ", nj: %g", [integer[nj]] ];
	IO.PutF1[ed.debug, ", j: %g\n", [integer[j]] ];
	}; */
      if(j < bpl) continue;
      break;

    } else if(((Dgp_Int32)(bitBuffer * 16)) < 0) {
      
      /* Vertical, ABS[d] = 2 */
      /* BumpStat[g4V2]; */
      d = 2;
      needBits = 6;
      goto scanV;

    } else if(((Dgp_Int32)(bitBuffer * 32)) < 0) {

      /* Vertical, ABS[d] = 3 */
      /* BumpStat[g4V3]; */
      d = 3;
      needBits = 7;
      goto scanV;

    } else if(goodBits < eolBits) {

      /* Now we need enough bits to test for EOL */
      goto pull;

    } else if(Peek(uncompBits2d) == uncompCode) {

      /* enter uncompressed mode */
      /* BumpStat[g4U];*/
      ed->uncompressed = DGP_TRUE;
      EatBits1(10);
      break;
    } else if(Peek(eolBits) == eolCode) {

      /* A valid EOL code, but we leave it for the exterior code to consume */
      break;

    } else {

      /* Coding error of some kind (or unsupported feature) */
      goto coding;

    }

    /* this code can only be entered through 'goto' */
    if(DGP_FALSE) {
    pull:
      PullBits();
      continue;

    coding:
      /* LogError[ed, $coding, goodBits]; */
      break;
      
    scanV:
      /* Common exit for vertical codes, d != 0 */
      bit = (UCHAR)(Peek(needBits) % 2);
      ni = ri + bit + bit - 1;
      
      if(bit == 0) d = -d;
      /* IF innerDebug AND ed.debug # NIL THEN {
	 IO.PutF1[ed.debug, "FastScan2D V %g, bits: ", [integer[d]] ];
	 PutB[ed.debug, Peek[needBits], needBits, DGP_FALSE];
	 }; */
      nj += d;
      if(nj < j) {
	/* LogError[ed, $backup, goodBits]; */
	break;
      }
      if(nj > bpl) {
	/* LogError[ed, $Dgp_Int32scan, goodBits]; */
	nj = bpl;
      }
      EatBits1(needBits);
      Fill2D(); /* nj, color */
      
      /* this is supposed to fall through here... i think */
    }      

    /* Common scanning code for horizontal and most vertical cases */
    j = nj;
    /* IF innerDebug AND ed.debug # NIL THEN {
       IO.PutF1[ed.debug, ", j: %g\n", [integer[j]]];
       }; */
    if(j >= bpl) break;
    ri = ni;
    
    while((nj = GetTransition(ref_s, ri)) <= j) ri += 2;

    if(ri > riLim) ri = riLim;
  }

  /* IF useDebug AND ed.debug # NIL THEN {
     IO.PutF1[ed.debug, "FastScan2D exit: %g, ", [integer[j]] ];
     IF goodBits < runTableBits THEN PullBits[];
     PutGoodBits[ed.debug, bitBuffer, goodBits];
     IO.PutChar[ed.debug, '\n];
     }; */
  ed->bitBuffer = bitBuffer;
  ed->goodBits = goodBits;

  *ret = j;
}

/*
  Processes the next scan line of input.  For 2D data we assume that the
  previous line was setup by Init or a previous call to NextLine.  The
  resulting transitions are placed in ed->referenceTransitions.

21.8.2000 / dhtesq.
Decoder profiling suggests this is a good place for an inline...
*/

INLINE ErrCode
NextLine(ExpandData *ed)
{
  DGP_PROLOGUE("NextLine", ed->DGP_PERRSTATE);

  ULONG bitBuffer = ed->bitBuffer;
  UCHAR goodBits = ed->goodBits;
  Dgp_Int32 bpl = ed->bpl;
  ULONG eolSeen = 0, tmpPeek, index, length = ed->inLength;
  IndexSequence *lineTransitions = ed->lineTransitions;
  Dgp_Int32 j;
  UCHAR *in = ed->in;


  SetTransition(lineTransitions, 0, -1);
  lineTransitions->end = 1;

  /*BumpStat(nLines); */

  if(!ed->end) {
    j = -1;
    
    if(!ed->uncompressed) {
      if((ed->referenceTransitions->s[1] == bpl) &&
	       (Peek(1) == 1)) {
	      /* We are copying the line above exactly */
	      /*BumpStat(g4Fast); */
	      EatBits1(1);
	      goto skipAhead;
            }

            /* Handle the EOL codes (if any) BEFORE the line ed.  For the K > 0
	       case we need this to establish the 1D tag for the remaining ed.
	       All EOL codes are optional. */
            while(DGP_TRUE) {
	      if(goodBits < (eolBits + 1)) PullBits();
	      /* ensure enough for EOL+tag */
	      if(Peek(eolBits) != eolCode) break;

	      EatBits1(eolBits);

	      if(ed->k > 0) {
	        /* For G3 ed with K > 0 there is an extra bit
                    -- 0 => 2-D mode
                    -- 1 => 1-D mode */
	        ed->oneDimTag = (Dgp_Bool)(Peek(1));
	        EatBits1(1);
	      }

	      /*if(ed->debug) {
	        fprintf(ed->debug, "Consume EOL");
	        if(ed->k > 0) {
	          fprintf(ed->debug, " (%d)", (Dgp_Bool)(ed->oneDimTag));
	        }
	        fprintf(ed->debug, "\n");
	      }*/
	      
	      eolSeen++;
	      
	      if(((ed->k < 0) && (eolSeen == 2)) || (eolSeen == 6)) {
	        ed->end = DGP_TRUE;
	        goto skipAhead;
	      }
      }
      ed->color = COL_WHITE;
    }
      
    while(DGP_TRUE) {
      if(ed->end) {
	      break;
      } else if(ed->uncompressed) {
	      DGP_ERR_PASS_CLEANUPF0(Uncompressed(ed, j, bitBuffer, goodBits, &j),
          "failed reading uncompressed line");
      } else if(ed->oneDimTag) {
	      DGP_ERR_PASS_CLEANUPF0(FastScan1D(ed, j, bitBuffer, goodBits, &j),
          "failed doing 1D scan");
      } else {
	      FastScan2D(ed, j, bitBuffer, goodBits, &j);
      }

      bitBuffer = ed->bitBuffer;
      goodBits = ed->goodBits;
      
      if(j == bpl) break;
      
      if(goodBits < eolBits) PullBits();
      
      tmpPeek = Peek(eolBits);
	
      if((tmpPeek == 0) || (tmpPeek == eolCode)) break;
    }

  }

skipAhead:

  ed->bitBuffer = bitBuffer;
  ed->goodBits = goodBits;

  DGP_ERR_GEN_CLEANUPF0(((lineTransitions->end + 2) > lineTransitions->size),
    cErrCompressionError, "error in line transitions");

  SetTransition(lineTransitions, lineTransitions->end, bpl);
  SetTransition(lineTransitions, lineTransitions->end + 1, bpl);
  lineTransitions->end += 2;
  ed->lineTransitions = ed->referenceTransitions;
  ed->referenceTransitions = lineTransitions;

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


#define DO_INVERT DGP_FALSE

/*
 * Take the compressed stream and output it to the given image. The
 * image size had better match what the compressed stream contains.
 * The image should have previously been cleared to zero.
 *
 * It will consume an EOFB if there is one present at the end of input and
 * eatEOFB is DGP_TRUE.  If consumed is not NULL, on return *consumed will contain
 * the number of bytes decoded.
 */
ErrCode 
g4Decode(Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
	 Dgp_Int32 longsPerRow, Dgp_Int32 xoffset, Dgp_Int32 xmax, Dgp_Int32 ymax,
	 Dgp_UInt8 *input, size_t inlen,
	 Dgp_Bool eatEOFB, size_t *consumed,
	 Dgp_State *dgp_state)
{
  DGP_PROLOGUE("g4Encode", DGP_PERRSTATE);

  ExpandData *ed = NULL;
  Dgp_Int32 line = 0;
  Dgp_Int32 stopy, stopx;


  stopy = MIN(height, ymax + 1);
  stopx = (MIN(xoffset + width, xmax + 1)) - xoffset;

  /* create and initialize expand data struct */
  DGP_ERR_PASS_CLEANUPF0(DecodeCreate((void **)&ed, dgp_state, debugFile),
    "failed creating expand data struct");

  DGP_ERR_PASS_CLEANUPF0(DecodeInit(ed, input, inlen, width, G4_K_VAL),
    "failed intializing expand data struct");

  /* expand a line and write it to the image */
  while(line < stopy) {
    /* expand a line of data */
    DGP_ERR_PASS_CLEANUPF1(NextLine(ed),
      "failed expanding line %d of image", line);

    /* write expanded line to image */
    OrBltLine(ed, image, xoffset, 0, stopx);

    image += longsPerRow;
    line++;
  }
  
  if (eatEOFB) {
      (void)ConsumeEndOfBlock(ed);
      }

  if (consumed != (size_t *)NULL) {
      /* How many bytes are completely read? */
      *consumed = ed->inIndex;
      /*
       * It may have pre-read some bytes - account for them.
       * Any completely good bytes in the buffer should be
       * counted as unused.  However, be careful: goodBits might include
       * some zero bits that were hallucinated from past the end of the
       * input.  Therefore, if goodBits pushes you over ed->inLength,
       * ignore them; they're fake.
       *
       * N.B.  If the actual G4-encoded data is equal to the length of
       * the buffer (inLength) minus a byte or two, then some goodBits might
       * be real and some might be hallucinated - *and there's no way to
       * tell them apart*.  Therefore, this routine might indicate
       * that all the data was consumed when actually a byte or two wasn't.
       * FIX.
       */
      if (ed->inIndex != ed->inLength) {
	  *consumed -= ed->goodBits / DGP_BITS_UINT8;
	  }
      }

  /* cleanup */
  DecodeCleanUp((void **)&ed);

  NORMAL_RETURN;

ERR_CLEANUP:

  DecodeCleanUp((void **)&ed);

  ERR_RETURN;
}


/*
 * This function is the same as the above function, except without the
 * creation and deletion of the ExpandData struct each time.
 *
 * 21.8.2000 / dhtesq.
 * Decoder profiling suggests this is a good place for an inline...
 *
 */
#ifdef _WIN32
DGP_LIBEXPORT 
#endif 
INLINE ErrCode 
g4DecodeC(void *_ed, Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
	  Dgp_Int32 longsPerRow, Dgp_Int32 xoffset, Dgp_Int32 xmax, Dgp_Int32 ymax,
	  Dgp_UInt8 *input, size_t inlen,
	  Dgp_Bool eatEOFB, size_t *consumed)
{
  ExpandData *ed = (ExpandData *)_ed;
  DGP_PROLOGUE("g4EncodeC", ed->DGP_PERRSTATE);

  Dgp_Int32 line = 0;
  Dgp_Int32 stopy, stopx;

  stopy = MIN(height, ymax + 1);
  stopx = (MIN(xoffset + width, xmax + 1)) - xoffset;

  /* initialize expand data struct */
  DGP_ERR_PASS_CLEANUPF0(DecodeInit(ed, input, inlen, width, G4_K_VAL),
    "failed intializing expand data struct");

  /* expand a line and write it to the image */
  while(line < stopy) {
    /* expand a line of data */
    DGP_ERR_PASS_CLEANUPF1(NextLine(ed),
      "failed expanding line %d of image", line);

    /* write expanded line to image */
    OrBltLine(ed, image, xoffset, 0, stopx);

    image += longsPerRow;
    line++;
  }

  if (eatEOFB) {
      (void)ConsumeEndOfBlock(ed);
      }

  if (consumed != (size_t *)NULL) {
      *consumed = ed->inLength - ed->inIndex;
      *consumed += ed->goodBits / DGP_BITS_UINT8;
      }

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}







/* 
 * Following code added by Eli
 *
 * Ripped off from g4Decode
 */

ErrCode 
g3_1DDecode(Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
	 Dgp_Int32 longsPerRow, Dgp_Int32 xoffset, Dgp_Int32 xmax, Dgp_Int32 ymax,
	 Dgp_UInt8 *input, size_t inlen,
	 Dgp_Bool eatEOFB, size_t *consumed,
	 Dgp_State *dgp_state)
{
  DGP_PROLOGUE("g3Encode", DGP_PERRSTATE);

  ExpandData *ed = NULL;
  Dgp_Int32 line = 0;
  Dgp_Int32 stopy, stopx;


  stopy = MIN(height, ymax + 1);
  stopx = (MIN(xoffset + width, xmax + 1)) - xoffset;

  /* create and initialize expand data struct */
  DGP_ERR_PASS_CLEANUPF0(DecodeCreate((void **)&ed, dgp_state, debugFile),
    "failed creating expand data struct");

  DGP_ERR_PASS_CLEANUPF0(DecodeInit(ed, input, inlen, width, G3_1D_K_VAL),
    "failed intializing expand data struct");

  /* expand a line and write it to the image */
  while(line < stopy) {
    /* expand a line of data */
    DGP_ERR_PASS_CLEANUPF1(NextLine(ed),
      "failed expanding line %d of image", line);

    /* write expanded line to image */
    OrBltLine(ed, image, xoffset, 0, stopx);

    image += longsPerRow;
    line++;
  }
  
  if (eatEOFB) {
      (void)ConsumeEndOfBlock(ed);
      }

  if (consumed != (size_t *)NULL) {
      /* How many bytes are completely read? */
      *consumed = ed->inIndex;
      /*
       * It may have pre-read some bytes - account for them.
       * Any completely good bytes in the buffer should be
       * counted as unused.  However, be careful: goodBits might include
       * some zero bits that were hallucinated from past the end of the
       * input.  Therefore, if goodBits pushes you over ed->inLength,
       * ignore them; they're fake.
       *
       * N.B.  If the actual G4-encoded data is equal to the length of
       * the buffer (inLength) minus a byte or two, then some goodBits might
       * be real and some might be hallucinated - *and there's no way to
       * tell them apart*.  Therefore, this routine might indicate
       * that all the data was consumed when actually a byte or two wasn't.
       * FIX.
       */
      if (ed->inIndex != ed->inLength) {
	  *consumed -= ed->goodBits / DGP_BITS_UINT8;
	  }
      }

  /* cleanup */
  DecodeCleanUp((void **)&ed);

  NORMAL_RETURN;

ERR_CLEANUP:

  DecodeCleanUp((void **)&ed);

  ERR_RETURN;
}


/*
 * Added by Eli
 *
 * Also ripped off from g4 version:  g4DecodeC
 *
 * This function is the same as the above function, except without the
 * creation and deletion of the ExpandData struct each time.
 *
 */
#ifdef _WIN32
DGP_LIBEXPORT 
#endif 
INLINE ErrCode 
g3_1DDecodeC(void *_ed, Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
	  Dgp_Int32 longsPerRow, Dgp_Int32 xoffset, Dgp_Int32 xmax, Dgp_Int32 ymax,
	  Dgp_UInt8 *input, size_t inlen,
	  Dgp_Bool eatEOFB, size_t *consumed)
{
  ExpandData *ed = (ExpandData *)_ed;
  DGP_PROLOGUE("g3EncodeC", ed->DGP_PERRSTATE);

  Dgp_Int32 line = 0;
  Dgp_Int32 stopy, stopx;

  stopy = MIN(height, ymax + 1);
  stopx = (MIN(xoffset + width, xmax + 1)) - xoffset;

  /* initialize expand data struct */
  DGP_ERR_PASS_CLEANUPF0(DecodeInit(ed, input, inlen, width, G3_1D_K_VAL),
    "failed intializing expand data struct");

  /* expand a line and write it to the image */
  while(line < stopy) {
    /* expand a line of data */
    DGP_ERR_PASS_CLEANUPF1(NextLine(ed),
      "failed expanding line %d of image", line);

    /* write expanded line to image */
    OrBltLine(ed, image, xoffset, 0, stopx);

    image += longsPerRow;
    line++;
  }

  if (eatEOFB) {
      (void)ConsumeEndOfBlock(ed);
      }

  if (consumed != (size_t *)NULL) {
      *consumed = ed->inLength - ed->inIndex;
      *consumed += ed->goodBits / DGP_BITS_UINT8;
      }

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


/* 
 * Following code added by Eli
 *
 * Ripped off from g4Decode
 */

ErrCode 
g3_2DDecode(Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
	 Dgp_Int32 longsPerRow, Dgp_Int32 xoffset, Dgp_Int32 xmax, Dgp_Int32 ymax,
	 Dgp_UInt8 *input, size_t inlen,
	 Dgp_Bool eatEOFB, size_t *consumed,
	 Dgp_State *dgp_state)
{
  DGP_PROLOGUE("g3Encode", DGP_PERRSTATE);

  ExpandData *ed = NULL;
  Dgp_Int32 line = 0;
  Dgp_Int32 stopy, stopx;


  stopy = MIN(height, ymax + 1);
  stopx = (MIN(xoffset + width, xmax + 1)) - xoffset;

  /* create and initialize expand data struct */
  DGP_ERR_PASS_CLEANUPF0(DecodeCreate((void **)&ed, dgp_state, debugFile),
    "failed creating expand data struct");

  DGP_ERR_PASS_CLEANUPF0(DecodeInit(ed, input, inlen, width, G3_2D_K_VAL),
    "failed intializing expand data struct");

  /* expand a line and write it to the image */
  while(line < stopy) {
    /* expand a line of data */
    DGP_ERR_PASS_CLEANUPF1(NextLine(ed),
      "failed expanding line %d of image", line);

    /* write expanded line to image */
    OrBltLine(ed, image, xoffset, 0, stopx);

    image += longsPerRow;
    line++;
  }
  
  if (eatEOFB) {
      (void)ConsumeEndOfBlock(ed);
      }

  if (consumed != (size_t *)NULL) {
      /* How many bytes are completely read? */
      *consumed = ed->inIndex;
      /*
       * It may have pre-read some bytes - account for them.
       * Any completely good bytes in the buffer should be
       * counted as unused.  However, be careful: goodBits might include
       * some zero bits that were hallucinated from past the end of the
       * input.  Therefore, if goodBits pushes you over ed->inLength,
       * ignore them; they're fake.
       *
       * N.B.  If the actual G4-encoded data is equal to the length of
       * the buffer (inLength) minus a byte or two, then some goodBits might
       * be real and some might be hallucinated - *and there's no way to
       * tell them apart*.  Therefore, this routine might indicate
       * that all the data was consumed when actually a byte or two wasn't.
       * FIX.
       */
      if (ed->inIndex != ed->inLength) {
	  *consumed -= ed->goodBits / DGP_BITS_UINT8;
	  }
      }

  /* cleanup */
  DecodeCleanUp((void **)&ed);

  NORMAL_RETURN;

ERR_CLEANUP:

  DecodeCleanUp((void **)&ed);

  ERR_RETURN;
}


/*
 * Added by Eli
 *
 * Also ripped off from g4 version:  g4DecodeC
 *
 * This function is the same as the above function, except without the
 * creation and deletion of the ExpandData struct each time.
 *
 */
#ifdef _WIN32
DGP_LIBEXPORT 
#endif 
INLINE ErrCode 
g3_2DDecodeC(void *_ed, Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
	  Dgp_Int32 longsPerRow, Dgp_Int32 xoffset, Dgp_Int32 xmax, Dgp_Int32 ymax,
	  Dgp_UInt8 *input, size_t inlen,
	  Dgp_Bool eatEOFB, size_t *consumed)
{
  ExpandData *ed = (ExpandData *)_ed;
  DGP_PROLOGUE("g3EncodeC", ed->DGP_PERRSTATE);

  Dgp_Int32 line = 0;
  Dgp_Int32 stopy, stopx;

  stopy = MIN(height, ymax + 1);
  stopx = (MIN(xoffset + width, xmax + 1)) - xoffset;

  /* initialize expand data struct */
  DGP_ERR_PASS_CLEANUPF0(DecodeInit(ed, input, inlen, width, G3_2D_K_VAL),
    "failed intializing expand data struct");

  /* expand a line and write it to the image */
  while(line < stopy) {
    /* expand a line of data */
    DGP_ERR_PASS_CLEANUPF1(NextLine(ed),
      "failed expanding line %d of image", line);

    /* write expanded line to image */
    OrBltLine(ed, image, xoffset, 0, stopx);

    image += longsPerRow;
    line++;
  }

  if (eatEOFB) {
      (void)ConsumeEndOfBlock(ed);
      }

  if (consumed != (size_t *)NULL) {
      *consumed = ed->inLength - ed->inIndex;
      *consumed += ed->goodBits / DGP_BITS_UINT8;
      }

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

