/* $Header: /project/icons/cvsroot/digipaper/src/libg4/g4.h,v 1.13 2001/10/29 23:36:43 ekim Exp $ */

#ifndef __G4_H
#define __G4_H

#include <stdio.h>

#include "util.h"
#include "dgp_allocators.h"
#include "dgp_err_control.h"
#include "dgp_state.h"

#include "pixr.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* 
 * These two functions are the original interface to g4 encoding and
 * decoding 
 */
ErrCode 
g4Encode(Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
	  Dgp_Int32 longsPerRow, Dgp_UInt8 *output,
	  size_t *outlen, Dgp_Bool eofb, Dgp_State *dgp_state);

ErrCode 
g4Decode(Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
	  Dgp_Int32 longsPerRow, Dgp_Int32 xoffset, Dgp_Int32 xmax, Dgp_Int32 ymax,
	  Dgp_UInt8 *input, size_t inlen,
	 Dgp_Bool eatEOFB, size_t *consumed,
	 Dgp_State *dgp_state);

/* 
 * Alternately, if you are doing a lot of encoding and/or decoding at one
 * time, you can use the functions below to minimize memory reallocation.
 * When Encoding(Decoding) call Encode(Decode)Create() to create an
 * encode(decode) cookie.  Pass this cookie in to the encode(decode)
 * function each time you need to encode(decode) something.  When you are
 * finished, you can delete the encode(decode) cookie by calling
 * Encode(Decode)CleanUp().
 */

ErrCode
EncodeCreate(void **encode_cookie, Dgp_State *dgp_state, FILE *db_out);

void
EncodeCleanUp(void **encode_cookie);

ErrCode 
g4EncodeC(void *encode_cookie, Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
	  Dgp_Int32 longsPerRow, Dgp_UInt8 *output, size_t *outlen,
	  Dgp_Bool eofb);

ErrCode
DecodeCreate(void **decode_cookie, Dgp_State *dgp_state, FILE *db_out);

void
DecodeCleanUp(void **decode_cookie);

/* 29 OCTOBER 01--ELI
 * As another option, to encode n lines of each token at a
 * time (rather than the entire token at once), use the below functions.  
 * These four functions make up the interface added by Eli Kim for 
 * chunk-by-chunk g4 encoding and decoding 
*/
ErrCode
startDoc(void **_cd, Dgp_Int32 width, Dgp_Int32 height,
         Dgp_Int32 maxNumLines, Dgp_State *dgp_state, Dgp_Bool invertColor);

ErrCode
g4EncodeChunk(void *_cd, Dgp_UInt32 *img_ptr, Dgp_Int32 numLines,
              Dgp_Int32 longsPerRow, Dgp_UInt8 **output, 
              size_t *outBufUsedLen);


ErrCode
endDoc(void *cd, Dgp_Int32 longsPerRow, 
       Dgp_UInt32 *flushedBytesBuf, Dgp_Int8 *numFlushedBytes);

/*
21.8.2000 / dhtesq.
Decoder profiling suggests this is a good place for an inline...
*/
#ifdef _WIN32
DGP_LIBEXPORT 
#endif
INLINE ErrCode 
g4DecodeC(void *decode_cookie, Dgp_UInt32 *image, Dgp_Int32 width, Dgp_Int32 height,
	  Dgp_Int32 longsPerRow, Dgp_Int32 xoffset, Dgp_Int32 xmax, Dgp_Int32 ymax,
	  Dgp_UInt8 *input, size_t inlen,
	  Dgp_Bool eatEOFB, size_t *consumed);

  /* encode a pixr */
DGP_LIBEXPORT ErrCode CDECL
w_g4_enc_pixr (IPWState *pIPWState, PIXR* pPixr, UInt8** ppDst, 
	       UInt32* pdwBytesInDst);

#ifdef __cplusplus
}
#endif

#endif /* __G4_H */
