/* $Header: /project/icons/cvsroot/digipaper/src/libg4/g4private.h,v 1.10 2001/10/29 23:36:28 ekim Exp $ */


#ifndef __G4PRIVATE_H
#define __G4PRIVATE_H

#include <stdio.h>
#include "util.h"
#include "dgp_allocators.h"
#include "dgp_err_control.h"
#include "dgp_state.h"

#ifdef __cplusplus
extern "C"
{
#endif

#define ULONG  Dgp_UInt32
#define UCHAR  Dgp_UInt8
#define USHORT Dgp_UInt16

#define bpw DGP_BITS_UINT32

#define WORD_LAST ((ULONG)0xFFFFFFFF)

#define eolBits 12
#define eolCode 1
  
#define codeSplit        64
#define codeLongSplit  2560

#define maxCodeBits 13
#define runSplit    512

#define TAB_COL_WHITE 0
#define TAB_COL_BLACK 1
#define TAB_LEN_SHORT 0
#define TAB_LEN_LONG  1

#define G4_K_VAL    -1
#define G3_1D_K_VAL  0
#define G3_2D_K_VAL  1

typedef struct _TabEntry{
  USHORT val;
  USHORT len;
} TabEntry;

extern TabEntry whiteShortCodeTab[];
extern TabEntry whiteLongCodeTab[];
extern TabEntry blackShortCodeTab[];
extern TabEntry blackLongCodeTab[];


/* compression structures */
typedef struct _RefText {
  Dgp_Int8 *data;
  ULONG length;
  ULONG maxLength;
} RefText;

typedef  struct _Trans{
  ULONG len;
  ULONG max;
  ULONG *data;
} Trans;

typedef struct _CompressData{
  ULONG old_bpl;

  ULONG bpl;
  RefText outBuffer;
  UCHAR outMod;
  ULONG outWord;
  Trans *transOld;
  Trans *transNew;
  FILE *debug;
  Dgp_State *dgp_state;
  ULONG sense;
  RefText chunkBuffer;
  Dgp_Int32 maxNumLines;
} CompressData;

/* compression functions */
ErrCode
EncodeInit(CompressData *cd, ULONG bpl, Dgp_Bool invertColor);

ErrCode
LineToTrans(CompressData *cd, ULONG *ptr);

void
FlipSense(CompressData *cd);

ErrCode
TransToG4Code(CompressData *cd, Dgp_Bool uncompressed);

ErrCode
TransToG3Code(CompressData *cd, Dgp_Int32 dim, Dgp_Bool uncompressed);

ErrCode
DoPage(CompressData* cd, UCHAR *outBuf, ULONG *_outBufLen, ULONG *ptr,
       ULONG pitch, ULONG lines, Dgp_Int32 k, Dgp_Bool eol, Dgp_Bool eof,
       Dgp_Bool uncompressed, Dgp_Bool pad, Dgp_Bool flush);

ErrCode
PutEol(CompressData *cd);

ErrCode
Terminate(CompressData *cd, Dgp_Int32 k);


/* expansion structures */
typedef struct _IndexSequence{
  ULONG end;
  ULONG size;
  Dgp_Int32 *s;
} IndexSequence;

typedef struct _ExpandData{
  ULONG bpl;
  UCHAR *in;
  ULONG inLength;
  ULONG inIndex;
  ULONG bitBuffer;
  UCHAR goodBits;
  IndexSequence *referenceTransitions;
  IndexSequence *lineTransitions;
  Dgp_Int32 k;
  Dgp_Bool uncompressed;
  Dgp_Bool oneDimTag;
  UCHAR color;
  Dgp_Bool end;
  UCHAR *error;
  FILE *debug;
  Dgp_State *dgp_state;
} ExpandData;

/* expansion functions */
ErrCode
DecodeInit(ExpandData *ed, UCHAR *in, ULONG inLen, ULONG bpl, Dgp_Int32 k);

ErrCode
NextLine(ExpandData *ed);

void
OrBltLine(ExpandData *ed, ULONG *dstBase, ULONG dstBitIndex, ULONG min,
	  ULONG max);

void
MoveLine(ExpandData *ed, ULONG *dstBase, ULONG dstBitIndex, ULONG min,
	 ULONG max, Dgp_Bool invert);

Dgp_Int32
ConsumeEndOfBlock(ExpandData *ed);

ErrCode
EndSeen(ExpandData *ed);

#ifdef __cplusplus
}
#endif

#endif /* __G4PRIVATE_H */
