
#include "halftone.h"
#include "image.h"
#include "image_defs.h"
#include "enchtdict.h"
#include "enchtreg.h"

#include "rast.pub"
#include "PsgFix.h"

static ErrCode
pixr_to_short_gray(Dgp_State *dgp_state, PIXR *pixr, DgpImage_Short *_im);


ErrCode
encodeHalftoneJBig2 (DgpAdt_List imageSize_List, DgpAdt_List halftone_pxr, 
                     jb2File jb2f, Dgp_Int32 section, Dgp_Int32 pgnum, Dgp_State *dgp_state)
{
    
    DgpAdt_ListNode ln, lnpic;
    pHALFTONEPIC ppic;
    PIXR *htpixr = NULL;
    DgpImage_Short sim;
    /*DgpImage_Gray gim; */
    jb2Segment htdict, htreg;
    jb2Pattern *htpats;
    jb2Segment *refsegs[2];
    Dgp_Int32 refsegnums[2], i;
    Dgp_Bool refretain[2];
    
    DGP_PROLOGUE ("encodeHalftoneJBig2", DGP_PERRSTATE);
    
    /* chech Params */
    DGP_ASSERT (halftone_pxr != NULL);
    DGP_ASSERT (imageSize_List != NULL);

    for(ln = DGPADT_LIFIRST(halftone_pxr), lnpic = DGPADT_LIFIRST(imageSize_List); 
        ln != NULL && lnpic != NULL; 
        ln = DGPADT_LINEXT(ln),lnpic = DGPADT_LINEXT(lnpic)){
        htpixr = (PIXR*)DGPADT_LIGET(ln) ;
        ppic   = (pHALFTONEPIC)DGPADT_LIGET(lnpic);
        if(ppic->cell_size > 0 && ppic->section == section){ /* if this is a valid halftone */

            DGP_ERR_PASS (pixr_to_short_gray(dgp_state, htpixr, &sim));
            htpixr = NULL;
            /*DgpImage_save (sim, "outfile.pbm"); */
        
            /*
            * Have to fill in a halftone dictionary segment.
            */
            DGP_CHECK_ALLOC (htpats = DGP_MALLOC (
                ppic->npattern * sizeof (jb2Pattern)));
        
            DGP_ERR_PASS (makeSpiralHalftonePattern (dgp_state, ppic->cell_size, 
                ppic->npattern, DGP_TRUE, htpats));
            htdict.pagenum = pgnum;
            htdict.data.htdict.npatterns = ppic->npattern;
            htdict.data.htdict.pat_w = (Dgp_UInt16) ppic->cell_size;
            htdict.data.htdict.pat_h = (Dgp_UInt16) ppic->cell_size;
            htdict.data.htdict.patterns = htpats;
        
            DGP_ERR_PASS (jb2EncodeHTDict(jb2f, &htdict));
            for (i = 0; i < ppic->npattern; i++)
                DGP_FREE (htpats[i].bits);
            DGP_FREE (htpats);
        
            /* Fake up a halftone region segment */
            htreg.segtype = JB2SEGTYPE_HTREG_IMMLL;
            htreg.pagenum = pgnum;
            htreg.nrefsegs = 1;
            refsegs[0] = &htdict;
            refsegnums[0] = htdict.segnum;
            refretain[0] = DGP_FALSE;
            htreg.refsegs = refsegnums;
            htreg.refretain = refretain;
            htreg.retain = DGP_FALSE;
            htreg.defer_nonretain = DGP_FALSE;
            htreg.data.htreg.cells = sim;
        
            htreg.data.htreg.hgx = 0;
            htreg.data.htreg.hgy = 0;
            htreg.data.htreg.hrx = ppic->cell_size * 256;
            htreg.data.htreg.hry = 0;
            htreg.data.htreg.int_combop = COMBOP_OR;
            htreg.data.htreg.defpixel = 0;
            htreg.data.htreg.dictseg = &htdict;
        
            htreg.reginfo.regxpos = ppic->x;
            htreg.reginfo.regypos = ppic->y;
            htreg.reginfo.regwidth = ppic->w;
            htreg.reginfo.regheight = ppic->h;
            htreg.reginfo.combop = COMBOP_OR;
        
            DGP_ERR_PASS (jb2EncodeHTRegion(jb2f, &htreg, refsegs));
            DgpImage_free (sim);
/*            DGP_FREE (ppic);
            ppic = NULL;
        }else {
             // if this is a outlier
            ln = DGPADT_LINEXT(ln);
            lnpic = DGPADT_LINEXT(lnpic); */
        }
    }
    
    NORMAL_RETURN;
}



ErrCode
pixr_to_short_gray(Dgp_State *dgp_state, PIXR *pixr, DgpImage_Short *_im) {
    
    DGP_PROLOGUE("pixr_to_short_gray()", DGP_PERRSTATE);
    DgpImage_ShortPixPtr		dptr;	
    Dgp_Int32			width, height, depth, frame, value;
    Dgp_Int32			i, j, pixelcount = 0;
    IPWState		        *ipstate = dgp_state->ipwstate;
    DgpImage_Short		im;
    
    DGP_ASSERT (pixr != NULL);
    DGP_ASSERT (_im != NULL);
    
    DGP_ERR_PASS(w_pixrGetWidth(ipstate, pixr, &width));
    DGP_ERR_PASS(w_pixrGetHeight(ipstate, pixr, &height));
    DGP_ERR_PASS(w_pixrGetDepth(ipstate, pixr, &depth));
    DGP_ERR_PASS(w_pixrGetFrame(ipstate, pixr, &frame));
    
    DGP_ERR_PASS(DgpImage_new(dgp_state, DGPIMAGE_SHORT, width, height, 
        (void **)&im));
    
    dptr = DGPIMG_PIXPTR(im, 0, 0);
    
    for(j = 0; j < height; j++) {
    /*    DGP_ERR_PASS(w_getLine(ipstate, pixr, j, &pptr));
        pptr += frame * depth / 8; */ /* skip the frame at the beginning of the line */
        for(i = 0; i < width; i++, dptr++, pixelcount++) {
        /*      val = *pptr;
        if (depth == 8)
        *dptr = (Dgp_Int16) val;
        else if (depth == 4)
        {
        *dptr = (Dgp_Int16) (val & 0xF0) >> 4;
        dptr++; i++; pixelcount++;
        *dptr = (Dgp_Int16) val & 0x0F;
        } */
            DGP_ERR_PASS (w_pixrGetPixel (ipstate, pixr, i, j, (void *) &value));
            *dptr = (Dgp_Int16) value;
        }
    }
    
    *_im = im;
    NORMAL_RETURN;
}

