
#include "dgp_types.h"
#include "image.h"
#include "image_defs.h"
#include "dgp_state.h"
#include "dgp_err_control.h"
#include "jb2halftone.h"
#include "PsgFix.h"

ErrCode
makeSpiralHalftonePattern (Dgp_State *dgp_state, Dgp_UInt32 grid_size, 
                           Dgp_UInt32 range, Dgp_Bool inverse, jb2Pattern *cells)
{
  
  Dgp_UInt32 start_pos;
  Dgp_UInt32 start_len;
  Dgp_Bool even = DGP_FALSE, sign = DGP_TRUE;

  Dgp_UInt32 count, count2, pattern_count = 0, pixel_count = 0;
  Dgp_UInt32 sum = 1;
  Dgp_Int32 offset;
  
  Dgp_UInt32 n_drop;
  Dgp_UInt32 *drop_vals;
  Dgp_UInt32 drop_index, *data_ptr;
  Dgp_Float32 fraction;

  DgpImage_Packed pim;
  Dgp_UInt16 longsperline, bytesperline;
  Dgp_UInt32 rowcount;

  DGP_PROLOGUE ("makeSpiralHalftonePattern", DGP_PERRSTATE);
  
  /* check Params */
  DGP_ASSERT (grid_size > 0);
  DGP_ASSERT (range <= grid_size * grid_size + 1);
  DGP_ASSERT (range >= 2);
  DGP_ASSERT (cells != NULL);

  /* create packed image, buffers for bitmaps */
DGP_ERR_PASS (DgpImage_new (dgp_state, DGPIMAGE_PACKED, grid_size, grid_size, (void **)&pim));
  longsperline = (Dgp_UInt16) grid_size / DGP_BITS_UINT32 + 1;
  bytesperline = longsperline * sizeof (Dgp_UInt32);
  for (count = 0; count < range; count++)
  {
DGP_CHECK_ALLOC (cells[count].bits = (Dgp_UInt32*) DGP_MALLOC (longsperline * 
                 grid_size * sizeof (Dgp_UInt32)));
    cells[count].longsperline = longsperline;
  }
  
  /* which patterns are not going to be exported */
  n_drop = grid_size * grid_size + 1 - range;
DGP_CHECK_ALLOC (drop_vals = (Dgp_UInt32*) DGP_MALLOC (n_drop * sizeof (Dgp_UInt32)));
  fraction = ((Dgp_Float32) grid_size * grid_size) / ((Dgp_Float32)n_drop + 1);
  drop_index = 0;
  /* printf ("FRACTION %f\n", fraction); */
  for (count = 1; count <= n_drop; count++)
    {
      drop_vals[count-1] = (Dgp_UInt32) (fraction * (Dgp_Float32) count + 0.5);
      /*      printf ("DROP VAL %d\n", drop_vals[count-1]); */
    }

  /* odd or even */
  if ((grid_size % 2) == 0) even = DGP_TRUE;

  /* set the starting pixel and the length of the smallest square */
  if (even) 
    {
      start_pos = (grid_size / 2) - 1;
      start_len = 2;
    }
  else
    { 
      start_pos = ((grid_size + 1) / 2) - 1;
      start_len = 1;
    }

  /* empty pattern */
  data_ptr = cells[pattern_count].bits;
  for (rowcount = 0; rowcount < grid_size; rowcount++)
  { 
    memcpy (data_ptr, DGPIMG_GETROW (pim, rowcount), bytesperline);
    data_ptr += longsperline;
  }
#ifdef DGP_DBUG
  printf ("PATTERN %d    %d pixel\n", (int)pattern_count,
    (int)pixel_count);
#endif
  pattern_count++;

  /* loop over increasing squares */
  for (count = start_len; count <= grid_size; count += 2, start_pos--)
    {

      
      if (count > 1)
	{
	  /* set offset, which is the position of the first pixel 
	     on the square */
	  if (even)
	    offset = (count  / 2) - 1;
	  else
	    offset = (count - 1)  / 2;

	  sum = 1;
	  
	  /* loop over all pixels on the square */
	  for (count2 = 1; count2 < count; count2++)
	    {

	      /*	      printf ("\nDEBUG %d  %d  %d     %d %u\n", 
			      start_pos, offset, count, sum, sign); */

	      /* top line */
#ifdef DGP_DBUG
      printf ("Add pix x %d   y %d  \n", 
		      (int)(start_pos + offset), (int)start_pos);
#endif
        DGPIMG_SETBIT (pim, start_pos + offset, start_pos);
	      pixel_count++;

        if ( n_drop == 0 || pixel_count != drop_vals[drop_index])
        {
          data_ptr = cells[pattern_count].bits;
          for (rowcount = 0; rowcount < grid_size; rowcount++)
          { 
            memcpy (data_ptr, DGPIMG_GETROW (pim, rowcount), bytesperline);
            data_ptr += longsperline;
          }
#ifdef DGP_DBUG
          printf ("PATTERN %d    %d pixel\n", (int)pattern_count,
            (int)pixel_count);
#endif
          pattern_count++;
          
        }
        else
	{
          if (drop_index < n_drop - 1) drop_index++;
	}

	      /* right line */
#ifdef DGP_DBUG
        printf ("Add pix x %d   y %d  \n", 
		      (int)(start_pos + count - 1), (int)(start_pos + offset));
#endif
        DGPIMG_SETBIT (pim, start_pos + count - 1, start_pos + offset);
	      pixel_count++;
        if ( n_drop == 0 || pixel_count != drop_vals[drop_index])
		{
          data_ptr = cells[pattern_count].bits;
          for (rowcount = 0; rowcount < grid_size; rowcount++)
          { 
            memcpy (data_ptr, DGPIMG_GETROW (pim, rowcount), bytesperline);
            data_ptr += longsperline;
          }
#ifdef DGP_DBUG
          printf ("PATTERN %d    %d pixel\n", (int)pattern_count,
            (int)pixel_count);
#endif
          pattern_count++;
		}
	      else
	{
          if (drop_index < n_drop - 1) drop_index++;
	}


	      /* bottom line */
#ifdef DGP_DBUG
	      printf ("Add pix x %d   y %d  \n", 
		      (int)(start_pos + count - offset - 1), (int)(start_pos + count - 1));
#endif
        DGPIMG_SETBIT (pim, start_pos + count - offset - 1, start_pos + count - 1);
	      pixel_count++;
        if ( n_drop == 0 || pixel_count != drop_vals[drop_index])
		{
          data_ptr = cells[pattern_count].bits;
          for (rowcount = 0; rowcount < grid_size; rowcount++)
          { 
            memcpy (data_ptr, DGPIMG_GETROW (pim, rowcount), bytesperline);
            data_ptr += longsperline;
          }
#ifdef DGP_DBUG
          printf ("PATTERN %d    %d pixel\n", (int)pattern_count,
            (int)pixel_count);
#endif
          pattern_count++;
		}
	      else
	{
          if (drop_index < n_drop - 1) drop_index++;
	}


	      /* left line */
#ifdef DGP_DBUG
	      printf ("Add pix x %d   y %d  \n", 
		      (int)start_pos, (int)(start_pos + count - offset - 1));
#endif
        DGPIMG_SETBIT (pim, start_pos, start_pos + count - offset - 1);
	      pixel_count++;
        if ( n_drop == 0 || pixel_count != drop_vals[drop_index])
		{
          data_ptr = cells[pattern_count].bits;
          for (rowcount = 0; rowcount < grid_size; rowcount++)
          { 
            memcpy (data_ptr, DGPIMG_GETROW (pim, rowcount), bytesperline);
            data_ptr += longsperline;
          }
#ifdef DGP_DBUG
          printf ("PATTERN %d    %d pixel\n", (int)pattern_count,
            (int)pixel_count);
#endif
          pattern_count++;
		}
	      else
	{
          if (drop_index < n_drop - 1) drop_index++;
	}


	      /* change offset to move to the next set of pixels 
		 on the square */
	      if (sign)
		{
		  offset += sum;
		  sign = DGP_FALSE;
		}
	      else
		{
		  offset -= sum;
		  sign = DGP_TRUE;
		}
	      sum++;
	    }
	}
      else
	{
	  /* special case of only single pixel square */
#ifdef DGP_DBUG
	  printf ("Add pix x %d   y %d  \n", 
		  (int)start_pos, (int)start_pos);
#endif
    DGPIMG_SETBIT (pim, start_pos, start_pos);
	  pixel_count++;
        if ( n_drop == 0 || pixel_count != drop_vals[drop_index])
    {
          data_ptr = cells[pattern_count].bits;
          for (rowcount = 0; rowcount < grid_size; rowcount++)
          { 
            memcpy (data_ptr, DGPIMG_GETROW (pim, rowcount), bytesperline);
            data_ptr += longsperline;
          }
#ifdef DGP_DBUG
          printf ("PATTERN %d    %d pixel\n", (int)pattern_count,
            (int)pixel_count);
#endif
          pattern_count++;
    }
    else
	{
          if (drop_index < n_drop - 1) drop_index++;
	}


	}

    }      

  /* invert if desired */
  if (inverse)
  {
    for (count = 0; count < range; count++)
    {
      if (count < range - count - 1)
      {
        data_ptr = cells[count].bits; 
        cells[count].bits = cells[range - count - 1].bits;
        cells[range - count - 1].bits = data_ptr;
      }
      else
      {
        break;
      }
    }
  }


  /* clean up */
  DgpImage_free (pim);
  DGP_FREE (drop_vals);
  NORMAL_RETURN;
}

