#ifndef HALFTONE_H
#define HALFTONE_H

#include "list.h"
#include "libiow.pub"
#include "libutilw.pub"
#include "libpixrw.pub"
#include "dgp_allocators.h"
#include "jb2file.h"
#include "segment.h"
#include <dgp_state.h>

#ifdef __cplusplus
extern "C"
{
#endif

/*how many subcells big we want to tokenize at */
#define SIZEFAC 2

/*return 0 on success, 1 on failure */

/*This fucntion takes a page, and attempts to segment it into halftone and text, */
/*while figurng out relevant halftoning information such as phase and cell_size */
/*It returns a list of images and a mask for the areas they cover. */
/*for examples of usage, see compress-document.c */
/*Bug: sometimes (rarely), the IPcore presegmenter that we use does not */
/*do a good enough job for Russ's algorithm to successfully find  */
/*the halftone cell sizes. (ex. zebra document, pg 152) */
/*in this case, we won't get as accurate image detection and classification */
/*and hence, less compression for the page.  */

typedef struct {
    Int32 x; 
    Int32 y;	
    Int32 w; 
    Int32 h;
    Int32 phase_x;
    Int32 phase_y;
    Int32 cell_size;
    Int32 npattern;
	Int32 section;
} HALFTONEPIC;

typedef HALFTONEPIC* pHALFTONEPIC;

DGP_LIBEXPORT 
ErrCode GetAllHalftoneInfo( Dgp_State *dgp_state, 
                            PIXR *pbS, 
                            PIXR **ppMask, 
                            DgpAdt_List *pics, 
                            UInt8 *cell_size,
                            Dgp_Bool *reverseVedio); 

DGP_LIBEXPORT 
ErrCode compressHalftoneToGray(
	Dgp_State *dgp_state,
   	PIXR *halftone_images,
        DgpAdt_List pics,
        Dgp_Int32 cell_size,
        DgpAdt_List gray_images
        );

DGP_LIBEXPORT 
ErrCode makeListofHalftoneRegions(
	Dgp_State *dgp_state,
   	PIXR *halftone_images,
        DgpAdt_List pics,
        DgpAdt_List gray_images
        );

ErrCode
convert4bppTo1bppPlanes(
        Dgp_State *dgp_state,
        PIXR *gray_images,
        DgpAdt_List pics,
        DgpAdt_List *planes
    );

ErrCode
convert8bppTo1bppPlanes(
        Dgp_State *dgp_state,
        PIXR *gray_images,
        DgpAdt_List pics,
        DgpAdt_List *planes
    );

DGP_LIBEXPORT ErrCode
encodeHalftoneJBig2 (DgpAdt_List imageSize, DgpAdt_List halftones, 
		     jb2File jb2f, Dgp_Int32 section, Dgp_Int32 pgnum, Dgp_State *dgp_state);

ErrCode
makeSpiralHalftonePattern (Dgp_State *dgp_state, Dgp_UInt32 grid_size, 
                           Dgp_UInt32 range, Dgp_Bool inverse, jb2Pattern *cells);


#ifdef __cplusplus
}
#endif

#endif
