/* */
/*   */
/* halfutil.c Roman Yelensky Wed, June 24, 1998 */
/*  Some functions for tokenizing halftones. */
/*  Note that halftone detection and analysis lives in halftone.c */
/* */
/* */

/* */
/* Take a specified rectangular region in a PIXR */
/* and return the kind of run_elts libtok_lossless  */
/* likes.  */
/* */

#include "libpixrw.pub"
#include "librastw.pub"
#include "mmcompress.h"
#include "tok-lossless.h" 
#include "halftone.h"
#include "image.h"
#include "hamtok.h"
#include "token-compress.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libhalf/halfutil.c,v 1.7 2003/02/10 22:53:25 mkrocken Exp $");

/* */
/*  Implements No, not really lossless tokenization of halftones. */
/* */
Dgp_Bool HammingReplace(Dgp_State *dgp_state, 
					   int cell_size,		/*work on tokens (SIZEFAC * cell_size) x (SIZEFAC * cell_size)  */
		       int start_hamdist,	/*the beginning hamming distance to start searching at */
		       int end_hamdist,		/*the end hamming distance  */
		       unsigned int border_usecount, /* > this goes in the tokenizer, <= potentially replaces */
		       DgpAdt_List tokOutput,		/*token list, this gets edited to rmove gone tokens */
		       PageInfo *pages,     /* boxes, and numBoxes for page */
		       int numPgs,		/*how many pages */
		       int last_token_written,	/*<= this number tokens don't get inserted or replaced */
			   Dgp_Bool report) {
	
    int i,j;
    pBTree pOurTree=NULL;
    pBTree *OurTrees=NULL;
    int side_len_w, side_len_h;
    int bit_string_len;
    int *bit_string=NULL;
    int *bit_string2=NULL;
    tokClass c;
    tokClass tokclass;
    DgpAdt_ListNode classi;
    int toknum;
    int max_toknum = -1; 
    int	min_toknum = 0x0FFFFFFF;
    int counter_inserted=0;
    int counter_replaced=0;
    int map_size;
    int *map;
    int mapto, answer, mass, how_many_masses, max_mass, temp, hamdist;
    int check_first, check_second;
    int new_tok, tok_dummy, fake_next_toknum;
    
    /*global variable for memory usage counting... */
    answer = 0; mem_alloced=0;
	
    /*first of all, return if we got called with 0 cell_size */
    if (cell_size <= 0) return DGP_TRUE;
	
    /*now we know we have something to do. */
    side_len_w	    = SIZEFAC * cell_size; /*we are going to be working with tokens side_len x side_len */
    side_len_h	    = SIZEFAC * cell_size;
    bit_string_len  = side_len_w * side_len_h; /*our tokens represented as bit strings will be this long */
    bit_string	    = (int *)tc_malloc(bit_string_len * sizeof(int));
	if (!bit_string) return DGP_FALSE;
    bit_string2	    = (int *)tc_malloc(bit_string_len * sizeof(int));
	if (!bit_string2) return DGP_FALSE;
    how_many_masses = bit_string_len + 1;
    max_mass	    = bit_string_len;
	
    OurTrees  = (pBTree *)tc_malloc(how_many_masses * sizeof(pBTree));
	if (!OurTrees) return DGP_FALSE;
	
    for (i=0; i < how_many_masses; i++) { 
		pOurTree = NewTree(&(tcc->a), bit_string_len);
		if (!pOurTree) return DGP_FALSE;
		OurTrees[i] = pOurTree;	
    }
    pOurTree=NULL;
	
    /*first step is to shove all the eligible tokens into our	 */
    /*hamming tokenizer and figure out the maximum and minimum toknum that  */
    /*we see */
    DGPADT_FOREACH (classi, tokOutput) {
		c = (tokClass)liGet(classi);
		assert(c != (tokClass)NULL);
		
		toknum = c->toknum;
		if (toknum > max_toknum) max_toknum = toknum;
		if (toknum < min_toknum) min_toknum = toknum;
		
		if( (toknum > last_token_written) 
			&& (c->height == side_len_h) && (c->width == side_len_w) && (c->num_instances > border_usecount) ) {
			/*expand the token into a bit string */
			RunsToBitString(c->runs, c->num_runs, bit_string, side_len_w, side_len_h);  
			
			/*hash it to its mass, so we know what tree to  */
			/*put it into */
			for (mass=0, i = 0; i < c->num_runs; i++) {
				mass+=RunLen(c->runs, i);	
			}
			
			/*put it in, with the pointer to the class struct as the user data */
			
			/*TInsert also works as an underlying tokenizer, so it writes a toknum */
			/*into the class thats its given. We'll just fool it here by having */
			/*it write the same toknum as the class already has.  */
			fake_next_toknum = c->toknum;   
			
			tok_dummy=TInsert(OurTrees[mass], bit_string, c, &new_tok, &fake_next_toknum);
			if (tok_dummy == -1) return DGP_FALSE;
			
			assert(new_tok); /*all these are unique */
			
			counter_inserted++; /*just to keep track of stuff. */
		}
		
    }
	
    if (report) { 
		tc_rptPrint(tc_rptOut, "Inserted %d tokens into the hamming tokenizer.", counter_inserted);
    }
	
    /*now that we know our toknum range, lets build a replacement */
    /*map, initialized to no replacements. */
    map_size = max_toknum - min_toknum + 1;
    map	     = (int *)tc_malloc(map_size * sizeof(int));
	if (!map) return DGP_FALSE;
	
    for(i=0, mapto = min_toknum; i < map_size; i++, mapto++) {
		map[i] = mapto;
    }
	
	
    /*lets try to replace every eligible token. */
    /*at this step, we remove it from the token list and and mark in our map */
    /*so the tokposes may be bashed.  */
	
    check_first  = 1;
    check_second = -1;
    
    classi=liFirst(tokOutput);
    while (classi) {
		c = (tokClass)liGet(classi);
		assert(c != (tokClass)NULL);
		toknum = c->toknum;
		
		/*if the token is eligble for replacement */
		if( (toknum > last_token_written) 
			&& (c->height == side_len_h) && (c->width == side_len_w) && (c->num_instances <= border_usecount) ) {
			
			/*expand the token into a bit string */
			RunsToBitString(c->runs, c->num_runs, bit_string, side_len_w, side_len_h);  
			
			/*hash to the mass */
			for (mass=0, i = 0; i < c->num_runs; i++) {
				mass+=RunLen(c->runs, i);	
			}
			
			/*now lets be a little tricky. I want to alternate the order */
			/*of the mass buckets I look for each token, because  */
			/*I don't want to tend to replace with either lighter or darker */
			/*tokens all the time. */
			
			/*there is only one token each of maxmass and 0, so just leave 'em alone. */
			/*(and we don't ever have mass 0 tokens anyway. */
			if (mass > 0 && mass < max_mass) {
				
				tokclass=NULL;
				
				for (hamdist = start_hamdist; hamdist<=end_hamdist; hamdist++) {
					if (hamdist%2 == 0) {
						/*we check only its own mass bucket */
						answer=SearchHam(OurTrees[mass], bit_string, bit_string2, hamdist, (void *)&tokclass);
						if (answer) break;
					} else {
						/*we check both the +1 and -1 buckets */
						answer=SearchHam(OurTrees[mass + check_first], bit_string, bit_string2, hamdist, (void *)&tokclass);
						if (answer) break;
						answer=SearchHam(OurTrees[mass + check_second], bit_string, bit_string2, hamdist, (void *)&tokclass);
						if (answer) break;
					}
				}
				
				if (answer) {
					/*Eureka, we've found something to replace this token with */
					
					/*tokclass now points to the token thats close enough to this one, */
					/*map the old one to the replacement. */
					map[toknum - min_toknum] = tokclass->toknum;
					
					/*change the order in which we look at mass buckets; */
					temp=check_first;
					check_first=check_second;
					check_second=temp;
					
					/*free the runs of this class, the actual class will be */
					/*freed when its chunk is destroyed. This  */
					/*really kindof sucks, because we don't know that */
					/*the runs were allocated from the same heap as this */
					/*free is working with. (they were, which is why this works) */
					tc_free(c->runs);
					
					/*this token is history, remove it from the token list and move on */
					classi=liRemAndNext(classi);
					
					counter_replaced++;
					
				} else {
					/*we didn't  */
					classi=liNext(classi);
				}
				
			} else {
				/*just move on to the next one */
				classi=liNext(classi);
			}
		} else {
			/*just move on to the next one */
			classi=liNext(classi);
		}
		
    }
	
    if (report) {
		tc_rptPrint(tc_rptOut, "Replaced %d tokens.", counter_replaced);
    }
	
	
    /*after all this hard work, lets go through all the tokposes */
    /*for all the pages and remap the toknums */
    for (i=0; i<numPgs; i++) {
		for (j=0; j<pages[i].halftone_numBoxes; j++) {
			mapto = pages[i].halftone_boxes[j].toknum - min_toknum;
			if ((mapto >= 0) && (mapto < max_toknum - min_toknum + 1)) {
				pages[i].halftone_boxes[j].toknum = map[mapto];
			}
		}
    }
	
	
    /*free stuff */
    tc_free(map);
    tc_free(bit_string);
    tc_free(bit_string2);
    for (i=0; i < how_many_masses; i++) { 
		FreeTree(OurTrees[i]);	
    }
    tc_free(OurTrees);
	
    return DGP_TRUE;
}


UInt8 BitmapToRuns(Dgp_Allocators *a,
		   DgpImage_Packed pbS, 
		   Int32 x,
		   Int32 y,
		   Int32 w,
		   Int32 h,
		   run_elt **runs, 
		   int *num_runs,
		   run_elt *temp_runs) {

    Int32 x_coord,y_coord;
    Int32 x_lim=x+w;
    Int32 y_lim=y+h;
    Int32 val;
    Int32 x_norm=0, y_norm;
    Int32 run_end=w;
    Int32 last_run_y=-1;
    /*run_elt *temp_runs; */
    int temp_num=0;
    int recording=0;
    /*int runs_alloced; */

    /*allocate much more than maximum. its too slow to realloc on the fly */
    /*temp_runs=(run_elt *)malloc(2*w*h*sizeof(run_elt));
    if(!temp_runs) {
	return 1;
    }
    runs_alloced=2*w*h;
    */

    for (y_coord=y; y_coord<y_lim; y_coord++) {
	
	y_norm=y_coord-y;

	for (x_coord=x; x_coord<x_lim; x_coord++) {
	    
	    val=pimGetBit(pbS, x_coord, y_coord);
	    
	    /*everything having to do with runs is done in coords */
	    /*with origin in top left of bounding box. */
	    x_norm=x_coord-x;
	
	    /*make more room if we are about to run out */
	    /*if (temp_num+3>runs_alloced) {
		temp_runs=(run_elt *)realloc((void *)temp_runs, ((runs_alloced*2)+(h*2))*sizeof(run_elt));
		if(!temp_runs) {
		    return 1;
		}
		runs_alloced+=h;
	    }*/

	    if(val && !recording) {
		/*we are at the start of a new run. */
		
		/*we need to use the explicit skip  */
		/*if either 1 or more lines above are blank */
		/*-or- the x_norm of this run is not <= */
		/*the last run's end + 1 and this is the first run on this y_norm */
		if ((y_norm-last_run_y>1) || 
		    ((x_norm>run_end+1) && (y_norm>last_run_y))) {
		
		    if (last_run_y==-1) last_run_y=0;
		    
		    RunLen(temp_runs, temp_num)=0;
		    RunX(temp_runs, temp_num)=(unsigned short)(y_norm-last_run_y);
		    temp_num++;
	
		
		}
			
		
		recording=1;
		last_run_y=y_norm;
		RunX(temp_runs, temp_num)=(run_elt)x_norm;
		RunLen(temp_runs, temp_num)=1;
	    } else if (val && recording) {
		/*we are in a run, just increment its length */
		RunLen(temp_runs, temp_num)++;
	    } else if (!val && recording) {
		/*a run just terminated */
		temp_num++;
		recording=0;
		run_end=x_norm-1;
	    
	    }  /*and the other case is ofcourse if  */
	       /*val==0 and we are not recording, which we don't care */
	       /*about */
	}
	/*now, we need to move down a row. If we are still recording a run, */
	/*we need to end it. The thing is that the convention */
	/*specified for run_elts arrays in tok.c says that if */
	/*you are moving down 1 row, you can't use the 0,newy  */
	/*method because you can use the other, implicit  */
	/*way of the new x-coord being less than the end of the  */
	/*previous run+1. This doesn't work for us though, */
	/*because the x-coord of the next run may in fact be more  */
	/*since a single bitmap may be multiple CCs. However, due to the */
	/*way RunsToUnframedBitmap is coded, if we can use the imlicit */
	/*line jump, we must. So, use the implicit jump if you can, */
	/*and the explicit if you can't */
	
	if (recording) {
	    recording=0;
	    run_end=x_norm-1;
	    temp_num++; 
	}
	
    }
    
    /*return */
    /*but shrink the returned array to the right size */
    /**runs=(run_elt *)realloc((void *)temp_runs, temp_num*2*sizeof(run_elt)); */
    *runs=(run_elt *)a->malloc(temp_num*2*sizeof(run_elt));
    
    if (temp_num && !*runs) {
	return 1;
    }
    
    memcpy(*runs, temp_runs, temp_num*2*sizeof(run_elt)); 
    
    *num_runs=temp_num;
    return 0;
}

UInt8 GoThroughHalftone(Dgp_Allocators *a,
			IPWState *pIPWState, 
			PIXR *pbS, 
			pHalfTokState halftokstate,
			pHALFTONEPIC pinfo,
			tokpos **tps,
			int *tpsnum) {
    
    Int32 top_left_x  =	pinfo->x; 
    Int32 top_left_y  =	pinfo->y;	
    Int32 bot_right_x =	top_left_x+pinfo->w; 
    Int32 bot_right_y =	top_left_y+pinfo->h;
    Int32 ht_box_size =	pinfo->cell_size;
    Int32 ht_box_x    = (((top_left_x/ht_box_size) + 1) * ht_box_size) + pinfo->phase_x;
    Int32 ht_box_y    = (((top_left_y/ht_box_size) + 1) * ht_box_size) + pinfo->phase_y;
    Int32 width, height;

    Int32 box_x, box_y;
    run_elt *runs;
    run_elt *temp_runs;
    int num_runs;
    int toknum;
    tokpos *tokposes;
    int tokposnum=0;
    int tokphase;
    int temp_width;
    int new_tok;
    
    /*for the hamming tokenizer (well, here, its not really hamming, its just a fast */
    /*tokenizer) */
    int bit_string[MAXSTRLEN];
    int bit_string_len;

    /*setup a libwjr packed image for fast access to individual pixels */
    DgpImage_Packed pSource;
    Int32 imWidth, imHeight;
    UInt8 *ppImage, *line1, *line2;
    int longsperline;

    w_pixrGetHeight(pIPWState, pbS, &imHeight);
    w_pixrGetWidth(pIPWState, pbS, &imWidth);

    w_getLine(pIPWState, pbS, 0, &line1);
    w_getLine(pIPWState, pbS, 1, &line2);
    longsperline = ((int)line2 - (int)line1)/sizeof(long);

    w_pixrGetImage(pIPWState, pbS, &ppImage);

    pSource=(DgpImage_Packed)pimWrap((unsigned long *)ppImage, imWidth, imHeight, longsperline);
    if (pSource==(DgpImage_Packed)NULL) {
	return 1;
    }
    
    /*just for kicks, lets make sure pimWrap works */
    /*imSave(pSource, "t:\\images\\wrapped.pbm"); */


    /*roll back the start to cover the whole region */
    while (ht_box_y>top_left_y) ht_box_y-=ht_box_size;
    while (ht_box_x>top_left_x) ht_box_x-=ht_box_size;

    /*sanity checks: since I've shrunk the region to conform be perfectly on */
    /*the phase grid, all these calculations here are not necessary, */
    /*but I am just doing a sanity check */
    assert(ht_box_x==top_left_x);
    assert(ht_box_y==top_left_y);
    assert(top_left_x % ht_box_size == pinfo->phase_x);
    assert(top_left_y % ht_box_size == pinfo->phase_y);
    assert(pinfo->w % (SIZEFAC*pinfo->cell_size)==0);
    assert(pinfo->h % (SIZEFAC*pinfo->cell_size)==0);

    width=SIZEFAC * ht_box_size;
    height=SIZEFAC * ht_box_size;

    /*allocate the array for tokposes. make slightly more than we'll need, */
    /*we'll realloc later */
    tokposes=(tokpos *)a->malloc((pinfo->w/width+3)*(pinfo->h/height+3)*sizeof(tokpos));
    if(!tokposes) {
	return 1;
    }    

    /*now that we know the width and height, we can set up the funky  */
    /*tokenizer */
    bit_string_len = width * height;
    assert(bit_string_len <= MAXSTRLEN);	/*the hamming tokenizer has a hard limit */
						/*on the length of strings it can accept. */
						/*This is equal to 16x16=256, which is */
						/*the maximum size of a halftone token. */
						/*(GetAllHalftone info will not return */
						/*anything more than 8 for the subcell size) */



    /*give BitmaptoRuns some scratch space so it doesn't have to  */
    /*do a malloc and a realloc. We'll do a malloc and a memcpy instead */
    temp_runs=(run_elt *)a->malloc(2*width*height*sizeof(run_elt));
    if (!temp_runs) {
	return 1;
    }

    
    tokphase=0;	    /*if we ever go back to the brick tiling. */
    temp_width = width/SIZEFAC;
    
    /*this outer loop will iterate over all the boxes that cover the region */
    for (box_y=ht_box_y; box_y+height<=bot_right_y; box_y+=height) {
	
	/*in the case that we did a phase shift to conserve tokens, */
	/*we need to still get the little token that was left to the */
	/*left of the shifted row */
	if (tokphase) {
		
	    if(BitmapToRuns(a, pSource, ht_box_x, box_y, temp_width, height, &runs, &num_runs, temp_runs)) {
		return 1;
	    }

	    /*also, get the bit_string we need for the hamming tokenizer */
	    RunsToBitString(runs, num_runs, bit_string, temp_width, height);

	    /*if its a blank token, just skip it, we're not interested */
	    if (!num_runs)  { 
		a->free(runs);
	    } else {
		
		toknum = AddHalfTok(halftokstate, runs, num_runs, 
				     bit_string, temp_width, height, &new_tok);
		if (toknum==-1) {
		    return 1;
		}

		if (!new_tok) halftokstate->a->free(runs); 
		/*its not a new token, its runs are already in the tokenizer. */
		/*so get rid of this extra copy. */
		
		
		/*record the tokpos */
		tokposes[tokposnum].xpos=(unsigned short)ht_box_x;
		tokposes[tokposnum].ypos=(unsigned short)box_y+height-1;
		tokposes[tokposnum].toknum=(unsigned int)toknum;
		tokposnum++;
	    }
	} 
	
	
	for (box_x=ht_box_x + tokphase; box_x+width<=bot_right_x; box_x+=width) {

	    /*another sanity check */
	   assert(box_x % ht_box_size == pinfo->phase_x);
	   assert(box_y % ht_box_size == pinfo->phase_y);
	    
	    /*get out bitmap converted to what the lossless tokenizer likes */
	    if(BitmapToRuns(a, pSource, box_x, box_y, width, height, &runs, &num_runs, temp_runs)) {
		return 1;
	    }
	    
	    /*also, get the bit_string we need for the hamming tokenizer */
	    RunsToBitString(runs, num_runs, bit_string, width, height);

	    /*if its a blank token, just skip it, we're not interested */
	    if (!num_runs)  { 
		a->free(runs);
		continue;
	    }
	    
	    toknum = AddHalfTok(halftokstate, runs, num_runs, 
				 bit_string, width, height, &new_tok);

	    if (toknum==-1) {
		return 1;
	    }
	    
	    
	    if (!new_tok) halftokstate->a->free(runs); 
	    /*its not a new token, its runs are already in the tokenizer. */
	    /*so get rid of this extra copy. */
	    
	    /*record the tokpos */
	    tokposes[tokposnum].xpos=(unsigned short)box_x;
	    tokposes[tokposnum].ypos=(unsigned short)box_y+height-1;
	    tokposes[tokposnum].toknum=(unsigned int)toknum;
	    tokposnum++;   

	    /*note that the actual token representation is sitting in tokstate, */
	    /*waiting for the tokfinal */

	}
	
	/*The phased row stopped one little box short. Lets get it */
	if (tokphase) {
	    if(BitmapToRuns(a, pSource, box_x, box_y, temp_width, height, &runs, &num_runs, temp_runs)) {
		return 1;
	    }
		
	    /*also, get the bit_string we need for the hamming tokenizer */
	    RunsToBitString(runs, num_runs, bit_string, temp_width, height);

	    /*if its a blank token, just skip it, we're not interested */
	    if (!num_runs)  { 
		a->free(runs);
	    } else {
		toknum = AddHalfTok(halftokstate, runs, num_runs, 
				     bit_string, temp_width, height, &new_tok);

		if (toknum==-1) {
		    return 1;
		}

		if (!new_tok) halftokstate->a->free(runs); 
		/*its not a new token, its runs are already in the tokenizer. */
		/*so get rid of this extra copy. */
		
		/*record the tokpos */
		tokposes[tokposnum].xpos=(unsigned short)box_x;
		tokposes[tokposnum].ypos=(unsigned short)box_y+height-1;
		tokposes[tokposnum].toknum=(unsigned int)toknum;
		tokposnum++;
	    }
	} 


	/*tokphase+=ht_box_size; */
	/*tokphase%=2*ht_box_size; */
    
    }

    /*free */
    pimUnwrap(pSource);
    a->free(temp_runs);

    /*return */
    *tps=(tokpos *)a->realloc((void *)tokposes, tokposnum*sizeof(tokpos));
    if(!*tps) {
	return 1;
    } 
    *tpsnum=tokposnum;    
    return 0;
}

Dgp_Bool
DoHalftoneFinding(pHalfTokState halfTokState, Dgp_State *dgp_state, IPWState *ipstate, UInt8 **image,
	Int32 xSize, Int32 ySize, Int32 bpl, int xRes, int yRes, PIXR **readPixr, 
	unsigned char *cell_size, tokpos **boxes, int *numBoxes, Dgp_Bool report)
{
	
	DgpAdt_ListNode ln;
	Int32 newbpl;
	UInt8 *w_image;
	PIXR *mask = NULL;
	DgpAdt_List pics = NULL;
	ErrCode err;
	tokpos *tokposes;
	int tokposnum;
	
	/*I need to make a copy of the user's input buffer because if there are haftone pictures, */
	/*I need to mask them out after I've tokenized them to pass the remainder (text) further to  */
	/*GetBoxes. Obviously I can't do it in the user's buffer.  */
	
	
	err=w_createPixr(ipstate, xSize, ySize, 1, Int32_TO_ScaledInt1616(xRes), Int32_TO_ScaledInt1616(yRes), 
		32, readPixr);
	if (err!=cSuccessful) goto bailout;
	
	w_pixrGetBpl(ipstate, *readPixr, &newbpl);
	
	/*oh, I hope this works... */
	assert(newbpl==bpl);
	if (newbpl!=bpl) goto bailout;
	/*we're f***ed... */
	
	err=w_pixrGetImageAll(ipstate, *readPixr, &w_image);
	assert(err==cSuccessful);
	if (err!=cSuccessful) goto bailout;
	
	/*there is some sort of bug in purify (so we've determined) */
	/*that causes it to puke on this memcpy inside ntdll */
	/*when you try  to do a big (600dpi) image */
	/*after several other images. This works though! */
	memcpy(w_image, *image, (ySize + 64) * bpl);
	
	/* return this image so it can be used some more */
	*image = w_image;  

	/*lets see it. */
	/*err=w_write(ipstate, "t:\\images\\fakepixr.pbm", readPixr, PNM); */
	/*assert(err==cSuccessful);  */
	
	/*there is no way now that readPixr doesn't exist, so we can assert on that */
	assert(*readPixr);
	
	/*lets do some halftoning. I'd like to note that */
	/*I had nothing to do with libcompdoc and its problems (some of which I fixed), */
	/*I am just putting in the halftoning stuff 'cause I am such a nice guy. (roman)	 */
	
	if (report){
		tc_rptPrint(tc_rptOut,  "Segmenting halftone and text and analyzing.");
	}
	
	if(GetAllHalftoneInfo(&(tcc->a), ipstate, *readPixr, &mask, &pics, cell_size)) {
		/*returns 1 if failed */
		goto bailout;
	}
	
	if (report){
		tc_rptPrint(tc_rptOut, "GetAllHalftoneInfo Done.");
	}
	/*If pics is null, that means that there wasn't enough */
	/*information to determine a cell-size  */
	if (!pics) {
		if (report){
			tc_rptPrint(tc_rptOut, "Find HT Size caused a fast exit from GetAllHalftone Info. (not an error!)");
		}
	} else if (liLen(pics)==0) {
		/*this means that FindHTSize did find enough information */
		/*to determine a cell size, but the individual images (or junk) */
		/*were too small. We let this case pass to GetBoxes. */
		if (report){
			tc_rptPrint(tc_rptOut, "Images too small or junk. Letting this pass to GetBoxes. (not an error!)");
		}
	} else {
		/*ok, this is the case we deal with. There are at least one halftone images */
		/*on our list. */
		
		/*save the cell_size in the cookie for later if we are supposed to  */
		/*try to lossyly tokenize halftones. If we don't save it, */
		/*they all be 0, which will tell WriteCompressed not  */
		/*to try to lossyly tokenize them. */
		
		DGPADT_FOREACH(ln, pics) {
			if(GoThroughHalftone(&(tcc->a), ipstate, *readPixr, halfTokState, 
				(pHALFTONEPIC)liGet(ln), &tokposes, &tokposnum)) goto bailout;
			/*GoThroughHalftone returns 1 on failure */
			
			/*so now all the tokens corresponding to this */
			/*halftone picture have been placed into the */
			/*tokenizer. we have the tokposes. we need to */
			/*append it to the growing list in cookie->halftone_boxes[curPg][i] */
			
			/*allocate for it */
			*boxes= (tokpos *)tc_realloc(*boxes, (*numBoxes+tokposnum)*sizeof(tokpos));
			if (!*boxes) goto bailout;
			
			/*copy  */
			memcpy(&((*boxes)[*numBoxes]), tokposes, tokposnum*sizeof(tokpos));
			(*numBoxes) += tokposnum;
			tc_free(tokposes);
			
			if (report)
				tc_rptPrint(tc_rptOut, "Tokenized halftone picture. cz: %d, x: %d, y: %d, w: %d, h: %d, phx: %d, phy: %d",
				(int)((pHALFTONEPIC)liGet(ln))->cell_size,
				(int)((pHALFTONEPIC)liGet(ln))->x,
				(int)((pHALFTONEPIC)liGet(ln))->y,
				(int)((pHALFTONEPIC)liGet(ln))->w,
				(int)((pHALFTONEPIC)liGet(ln))->h,
				(int)((pHALFTONEPIC)liGet(ln))->phase_x,
				(int)((pHALFTONEPIC)liGet(ln))->phase_y);
			
			/*we don't need the node storing the picture information. */
			tc_free((pHALFTONEPIC)liGet(ln));
		}
		
		
		/*now that we have all the pictures, mask 'em out and run the  */
		/*regular tokenizer. */
		err=w_rasterop(ipstate, 0, 0, mask, PIX_NOT(PIX_SRC)&PIX_DST, 0, 0, xSize, ySize, *readPixr);
		if (err!=cSuccessful) goto bailout;
		
	}
	
	if (pics) liFree(pics);
	pics=NULL;
	if (mask) w_destroyPixr(ipstate, &mask);

	return DGP_TRUE;

bailout:
	if (mask) w_destroyPixr(ipstate, &mask);

	return DGP_FALSE;
}



