#include "hamtok.h"
#include "assert.h"
#include "memory.h"
#include "list.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libhalf/hamtok.c,v 1.4 2003/02/10 22:53:25 mkrocken Exp $");

#define setBit(num, bit)  (num)|=(0x80000000 >> (bit))
#define getBit(num, bit)  ((num)&(0x80000000 >> (bit))) /* ? 1 : 0 */
#define getMultiLongBit(bit_string, bit) getBit(((bit_string)[(bit)/32]), (bit))
#define setMultiLongBit(bit_string, bit) setBit(((bit_string)[(bit)/32]), (bit))



__inline void setTNode(pTNode pNode, pTNode pOne, pTNode pZero, pTNode pParent, 
		       char prefix, int prefix_len, tokClass pData) {

    pNode->pOne = pOne;
    pNode->pZero = pZero;
    pNode->pParent = pParent;
    memset(pNode->prefix, prefix, LONGSIBS * 4);
    pNode->prefix_len = prefix_len;
    pNode->pData = pData;

}


pBTree NewTree(Dgp_Allocators *a, int len) {

    pBTree tmp;

    tmp = (pBTree)a->malloc(sizeof(BTree));
    if (!tmp) return NULL;

    
    /*initialize our chunk */
    tmp->MyMemChunk = DgpAdt_chNew(sizeof(TNode));
    if (!tmp->MyMemChunk) return NULL;
    
    tmp->pRoot = (pTNode)DgpAdt_chAlloc(tmp->MyMemChunk);
    if (!tmp->pRoot) return NULL; 
    
    
    /*initialize the root node */
    setTNode(tmp->pRoot, NULL, NULL, NULL, 0, 0, NULL);

    tmp->a = a;
    
    /*all the strings in this tree will be of this length */
    tmp->len = len;


    return tmp;
}


void TreeWalk(pTNode *ppNode, int *bit_string, int *pindex, int *pcurr_bit, int len) {
    
    int curr_bit = 0;
    int index = *pindex;
    pTNode pNode = *ppNode;
    int their_bit, not_their_bit;
    int our_bit, not_our_bit;
    unsigned long *prefix;
    int prefix_len;
    pTNode pOne, pZero;

    for (;index < len;) {
    	
	pOne	  = pNode->pOne;
	pZero	  = pNode->pZero;
	prefix_len= pNode->prefix_len;
	prefix	  = pNode->prefix;

	/*in this type of tree a node has either no children */
	/*or both of them. we should assert on this */
	assert((pOne && pZero) || ((!pOne) && (!pZero))); 

	if (pNode==*ppNode) curr_bit = *pcurr_bit; 
	    else curr_bit = 0;
	/*start with the initial value provided if we are on the first node, */
	/*if not, 0 */
	
	/*ok, lets roll along our bitstring in the current node, */
	/*consuming as much of their bitstring as we can */
	for (; curr_bit < prefix_len; curr_bit++, index++) {
	   their_bit	 = bit_string[index];
	   not_their_bit = !their_bit;
	   our_bit	 = getMultiLongBit(prefix, curr_bit);
	   not_our_bit	 = !our_bit; 
	
	   if (their_bit && our_bit) continue; 
	   else if (not_their_bit && not_our_bit) continue; 
	     else break;
	}

	/*index is now pointing to the first unconsumed bit, */
	/*curr_bit is pointing to first discordant bit (or past the end) */
	/*if children are available and only if we are past the  */
	/*last bit of the prefix, descend  */

	if ((curr_bit == prefix_len) && (index < len)) {
	    
	    their_bit	 = bit_string[index];
	    not_their_bit= !their_bit;

	    if (pOne && their_bit) pNode=pOne;
	    else if (pZero && not_their_bit) pNode=pZero;
		else break;
	
	    /*if not, we've consumed as much of the string as posible. */
	} else {
	    /*we can't descend anymore , and we can't */
	    /*match anymore bits  */
	    break;
	}
    }


    /*return the modified variables to the caller */
    *ppNode = pNode;
    *pindex = index;
    *pcurr_bit = curr_bit;
}

int TInsert(pBTree pTree, int *bit_string, tokClass user_data, int *new_tok, int *next_toknum) {

    /*Dgp_Allocators *a = pTree->a; */
    int len	 = pTree->len;
    pTNode pNode = pTree->pRoot;
    int index	 = 0;
    pTNode temp,temp2;
    int curr_bit = 0; 
    int	temp_bit, temp_bit2;
    int return_toknum=0;
    tokClass c = user_data;

    /*first, consume as much of the string as possible. */
    /*at the end of this loop, the index will point to */
    /*the first unconsumed bit.  */

    TreeWalk(&pNode, bit_string, &index, &curr_bit, len);
   

    /*Ok, the following things are true now: */
    /* 1) index points to the next bit to be consumed or past end, in which case, go on */
    /* 2) pNode points to the last node we descended to */
    /*	    a) this is an empty tree with no strings inserted into it. */
    /*	       in this case, index=0, curr_bit==0==prefix_len, and no */
    /*	       children of pNode. We just write the string into  */
    /*	       into the first node. */
    /* */
    /*	    b) This node does not have all of its prefix length consumed */
    /*	       and has no children */
    /*	       - We have to split the tree here at the */
    /*		 discordant bit. This involves creating both children, */
    /*		 putting each of the suffixes in the right child, moving the userdata */
    /*		 of the old bitstring to the correct child and setting up the */
    /*		 userdata of the new child.     */
    /* */
    /*	    c) This node does not have all of its prefix length consumed and has */
    /*	       both children. This means that the discord in the bitstrings comes */
    /*	       before the branching of this node. This means that we need to */
    /*	       insert a new node in the tree that contains the common prefix. */
    /*	       and points to the old node and a node containing the newest suffix. */
    /*	  */
    /*	 */

    /*now insert the rest */
   if(index < len) {
	
	/*we know we are inserting a new token, */
        /*lets assign its toknum */
	
	if (c) {
	    c->toknum = return_toknum = *next_toknum;
	    (*next_toknum)++; 
	}

	*new_tok = 1;


	/*check for case a */
	if (pNode->prefix_len==0) {
	    
	    assert(curr_bit==0);
	    assert(index==0);
    
	    /*record the bitstring  */
	    for (;index < len; index++, curr_bit++) {
		if (bit_string[index]) setMultiLongBit(pNode->prefix, curr_bit);
	    }

	    pNode->pData = user_data;
	    pNode->prefix_len = index;

	} else if ((curr_bit < pNode->prefix_len) && (!pNode->pOne) && (!pNode->pZero)) {
	    
	    /*case b */
	    
	    /*first create the two nodes */
	    temp = (pTNode)DgpAdt_chAlloc(pTree->MyMemChunk);
	    if (!temp) return -1;
	    
	    setTNode(temp, NULL, NULL, NULL, 0, 0, NULL);

	    temp2 = (pTNode)DgpAdt_chAlloc(pTree->MyMemChunk);
	    if (!temp2) return -1;
	    
	    setTNode(temp2, NULL, NULL, NULL, 0, 0, NULL);

	    mem_alloced+=(2*sizeof(TNode));
	
	    /*now split the information contained in the original node  */
	   
	    /*first, place the suffix of the original string in temp */
	    for (temp_bit=0, temp_bit2=curr_bit; temp_bit2 < pNode->prefix_len; temp_bit2++, temp_bit++) {
		if (getMultiLongBit(pNode->prefix, temp_bit2)) setMultiLongBit(temp->prefix, temp_bit);				 
	    }
	    temp->prefix_len = temp_bit2 - curr_bit;
	    temp->pData = pNode->pData;	    /*move the user data to the new final node */
					    /*of the original string */
	    temp->pParent = pNode;

	    /*second, place the suffix of the new string in temp2 */
	    for (temp_bit=0, temp_bit2=index; temp_bit2 < len; temp_bit2++, temp_bit++) {
		if (bit_string[temp_bit2]) setMultiLongBit(temp2->prefix, temp_bit);				 
	    }
	    temp2->prefix_len = temp_bit2 - index;
	    temp2->pData      = user_data;   /*place the new user data in the final  */
					     /*node of the new string. */
	    
	    temp2->pParent = pNode;

	    /*hookup the new child nodes to the parent */
	    if (getMultiLongBit(pNode->prefix, curr_bit)) {
		pNode->pOne = temp;
		pNode->pZero = temp2;
	    } else {
		pNode->pOne = temp2;
		pNode->pZero = temp;
	    }
	
	    /*change the length of the parent to indicate that it has shortened */
	    /*because it now has suffixes */
	    pNode->prefix_len = curr_bit;
	    pNode->pData      = NULL;	/*and null out this user data. */
					/*its in a better place now. */
	
	} else if ((curr_bit < pNode->prefix_len) && (pNode->pOne) && (pNode->pZero)) {
	
	    /*first create the two nodes */
	    temp = (pTNode)DgpAdt_chAlloc(pTree->MyMemChunk);
	    if (!temp) return -1;
	    
	    setTNode(temp, NULL, NULL, NULL, 0, 0, NULL);

	    temp2 = (pTNode)DgpAdt_chAlloc(pTree->MyMemChunk);
	    if (!temp2) return -1;
	    
	    setTNode(temp2, NULL, NULL, NULL, 0, 0, NULL);

	    mem_alloced+=(2*sizeof(TNode));

	    /*now, temp is going to get the original suffix */
	    for (temp_bit=0, temp_bit2=curr_bit; temp_bit2 < pNode->prefix_len; temp_bit2++, temp_bit++) {
		if (getMultiLongBit(pNode->prefix, temp_bit2)) setMultiLongBit(temp->prefix, temp_bit);				 
	    }
	    temp->prefix_len = temp_bit2 - curr_bit;

	    /*the original parent is going to become the new */
	    /*parent of the split tree. temp is going to take over  */
	    /*the old nodes children */
	    temp->pParent = pNode;
	    temp->pOne = pNode->pOne;
	    temp->pZero = pNode->pZero;
	    pNode->pOne->pParent = temp;
	    pNode->pZero->pParent = temp;

	    /*there is no userdata to copy, becaue it is in the old nodes children anyway */
	    assert(!pNode->pData);
	
	    /*temp2 is going to get the new suffix, which is longer than  */
	    /*old, place the suffix of the new string in temp2 */
	    for (temp_bit=0, temp_bit2=index; temp_bit2 < len; temp_bit2++, temp_bit++) {
		if (bit_string[temp_bit2]) setMultiLongBit(temp2->prefix, temp_bit);				 
	    }

	    temp2->prefix_len = temp_bit2 - index;
	    temp2->pData      = user_data; 
	    temp2->pParent    = pNode;

	    /*assert on the length issue */
	    assert(temp2->prefix_len > temp->prefix_len);
	
	    /*now, shorten the original node's prefix length */
	    pNode->prefix_len = curr_bit;
	
	    /*set up the new suffixes as children of the  */
	    /*of the original node. The original nodes children */
    	    /*are now children of temp */
	    if (getMultiLongBit(pNode->prefix, curr_bit)) {
		pNode->pOne = temp;
		pNode->pZero = temp2;
	    } else {
		pNode->pOne = temp2;
		pNode->pZero = temp;
	    }	

	} else {
	   
	    /*this should never happen, duh... */
	    assert(DGP_FALSE);

	}
	
	
    } else {
	
	/*this is an old token... */
	c = pNode->pData;
	if (c) {
	    c->num_instances++;
	    return_toknum = c->toknum;
	}

	*new_tok = 0;

    }
	
  
    return return_toknum;
}
    
/*A lower level function that SearchHam calls */
pTNode seaham(pTNode pNode, int curr_bit, int *search_for, int index, int len, int dist); 


/*search for a string a certain hamming distance away from given  */
/*string. Return 0 if none found. If found, return 1 and fill in the  */
/*buffer with the found string */

int SearchHam(pBTree pTree, int *search_for, int *found, int ham_dist, void **user_data) {
    
    int len	 = pTree->len;
    pTNode pNode = pTree->pRoot;
    int index	 = 0;
    int curr_bit;
    pTNode temp;
    
    /*Ok, call seaham on the root node. */
    temp = seaham(pNode, 0, search_for, 0, len, ham_dist);

    /*temp will either be null or will point to the last node */
    /*of the found string or null. */

    if (temp) {
	
	/*give the user back the data stored in the last node. */
	*user_data = temp->pData;
	
	/*fill in the found buffer with the string we found  */
	for (index = len - 1; index > -1 ;) {
	    
	    curr_bit = temp->prefix_len - 1;

	    while (curr_bit > -1) {
		if (getMultiLongBit(temp->prefix, curr_bit)) found[index]=1;
		    else found[index]=0;
		curr_bit--;
		index--;
	    }

	    temp = temp->pParent;
	}

	return 1;

    } else {
	*user_data=NULL;
	return 0;
    }

}

/*This recursive (so sue me) function gets a called on a node  */
/*with a certain hamming distance and index into our search */
/*for string. see comments inside for explanation of how it works  */

pTNode seaham(pTNode pNode, int curr_bit, int *search_for, int place, int len, int dist) {

    pTNode tmp = pNode;
    pTNode curr_node = pNode;
    int index = place;
    int curr_b = curr_bit;
    int branch_or_bit;
    int the_bit, not_the_bit;
    int search_bit, not_search_bit;
    

    /*the base case is very simple. If the dist == 0 */
    /*this means that we allow no more variation from  */
    /*the search string (which implies that  */
    /*dist was decremented somewhere above in the  */
    /*tree) So we just roll on down the  */
    /*tree trying to match the string exactly. */
    /*if we can, we return the pointer to the last node. */
    /*If we can't, we return NULL */
    if (dist == 0) {	
	   
	TreeWalk(&tmp, search_for, &index, &curr_b, len);

	if (index == len) return tmp;
	    else return NULL;
    }
    
    /*now say the distance is > 1. That means that if the right branch */
    /*is available, (the bit we're at in the string corresponds to an existing */
    /*bit in the tree) we take that branch without decrementing the dist. But, if the other child */
    /*is also available (the wrong one) we decrement the dist and take that branch as well. */
    
    /*lets precompute a couple of possibilites */
    if (curr_b < curr_node->prefix_len) {
	branch_or_bit = 1; 
	/*there is still at least one bit left at the current node, so we can advance */
	/*consuming this bit. */
    } else if (curr_node->pOne && curr_node->pZero) {
	branch_or_bit = 0;
	/*there are no bits, but we do have the branches available */
    } else {
	/*no more bits, no more branches, we're are at a terminal */
	/*node and dist!=0. The search fails. */
	return NULL;
    }

    if (branch_or_bit == 1) {
	/*work the bit */
	search_bit = search_for[place];
	not_search_bit = !search_bit;
	the_bit =  getMultiLongBit(curr_node->prefix, curr_b);
	not_the_bit = !the_bit;

	if ((search_bit && the_bit) || (not_search_bit && not_the_bit)) {
	    tmp = seaham(curr_node, curr_b + 1, search_for, place + 1, len, dist);
	    return tmp;

	} else {
	    /*we only have the wrong branch, take it */
	    tmp = seaham(curr_node, curr_b + 1, search_for, place + 1, len, dist - 1);
	    return tmp;
	}
    } else {
	/*we don't have anymore bits in the current node and we have both */
	/*children (as always), so there are more strings. */

	/*just recurse to the above case, let it deal with it  */
	
	tmp = seaham(curr_node->pOne, 0, search_for, place, len, dist);
	if (tmp) return tmp;

	tmp = seaham(curr_node->pZero, 0, search_for, place, len, dist);
	return tmp;
    }


    
    assert(DGP_FALSE);
    return NULL;
    
}
	

void RunsToBitString(run_elt *runs, int num_runs, int *bit_string, int side_length_w, int side_length_h) {

    int i,j;
    int *prow = bit_string;
    int *pbit;
    int max_index = side_length_w * side_length_h;

    /*first, init to 0; */
    for (i=0; i < max_index; i++) {
	bit_string[i]=0;
    }

    for (i=0; i < num_runs; i++) {
	
	/*if this is a skip, then, by all means, skip */
	if (RunLen(runs,i)==0) {
	    prow += side_length_w * RunX(runs, i);
	} else {
	    /*its a real run, but possibly also an implicit skip (only if we didn't have an  */
	    /*explicit skip immediately preceding.) */
	    if (i && (RunX(runs, i) <= RunX(runs, i-1) + RunLen(runs, i-1)) && RunLen(runs,i-1)) {
		prow += side_length_w;
	    }

	    /*start out on the current row */
	    pbit=prow;
	    
	    /*move to X location of the run  */
	    pbit+=RunX(runs, i);

	    /*fill in the ones */
	    for (j=0; j < RunLen(runs,i); j++) {
		*pbit=1;
		 pbit++;
	    }
	}
    }

}

    

void FreeTree(pBTree pTree) {

    /*get rid of the chunk */
    DgpAdt_chDestroy(pTree->MyMemChunk);

    /*now get rid of the actual tree structure. */
    pTree->a->free(pTree);

}

    
/* */
/* AddHalfTok, -1 on error, otherwise the toknum */
/* */

int AddHalfTok(pHalfTokState p_state, run_elt *runs, int num_runs, 
	       int *bit_string, int side_len_w, int side_len_h, int *new_tok) {
    
    int toknum;
    tokClass c;
    int bit_str_len = side_len_h * side_len_w;

    assert(bit_str_len <= MAXSTRLEN);

    /*first, create the class in case it will be inserted */
    c = (tokClass)DgpAdt_chAlloc(p_state->ClassMemChunk);
    if (!c) return -1;

    c->runs = runs;
    c->num_runs = num_runs;
    c->width = side_len_w;
    c->height = side_len_h;
    c->num_instances = 1;
    c->toknum = -1;

   /*it is possible that this is the first element  */
   /*of this h x w that we've seen, in case it is, */
   /*we need to create the tree */
   if (!p_state->Trees[bit_str_len]) {
	
	p_state->Trees[bit_str_len] = NewTree(p_state->a, bit_str_len);
	if (!p_state->Trees[bit_str_len]) return -1;
   
   }

    /*now try to insert it into our tree. If it inserts, */
    /*it'll come back with a new toknum, if not, it'll increment */
    /*the instances of the old token and come back with the old toknum */
   toknum = TInsert(p_state->Trees[bit_str_len], bit_string, c, new_tok, &(p_state->next_toknum)); 
    
   if (toknum == -1) return -1;	    /*error in TInsert */
	
   if (!*new_tok) DgpAdt_chFree(p_state->ClassMemChunk, c);
				    /*this wasn't a new token. We don't need */
				    /*the class we made. The user will also */
				    /*get the information that this isn't */
				    /*a new token, so he can free the runs if he */
				    /*wants to. */

   return toknum;

}

void lFinal(pTNode pN, DgpAdt_List l) {

    if (pN->pOne && pN->pZero) { 
	/*this node has both children, descend */
	lFinal(pN->pOne, l);
	lFinal(pN->pZero, l);
    } else if ((!(pN->pOne)) && (!(pN->pZero))) {
	/*This node has no children, which means */
	/*that it is a leaf node containing a class */
	assert(pN->pData);
	liAdd(l, pN->pData);
    } else {
	assert(DGP_FALSE);
    }
}



DgpAdt_List HalfTokFinal(pHalfTokState p_state) {

    DgpAdt_List l;
    int i;
    
    l = liNew();
    if (!l) return NULL;

    /*call the lower level function on all the trees to put all the classes onto the list */
    for (i=0; i <= MAXSTRLEN; i++) {
	if (p_state->Trees[i]) {
	    lFinal(p_state->Trees[i]->pRoot, l);
	    FreeTree(p_state->Trees[i]);	    	    
	    p_state->Trees[i]=NULL;
	}	
    }

    return l;
}


pHalfTokState InitHalfTokenizer(Dgp_Allocators *a) {

    pHalfTokState p_state;

    p_state = (pHalfTokState)a->malloc(sizeof(HalfTokState));
    if (!p_state) return NULL;

    p_state->a = a;

     /*initialize our chunk for classes */
    p_state->ClassMemChunk = DgpAdt_chNew(sizeof(struct _tokClass));
    if (!p_state->ClassMemChunk) return NULL;
    
    /*set all the pointers to NULL */
    memset(p_state->Trees, 0, (MAXSTRLEN+1) * sizeof(pBTree));

    p_state->next_toknum=0;

    return p_state;
}


void FreeHalfTokClass(pHalfTokState ps, tokClass c) {

    /*free the runs first */
    ps->a->free(c->runs);

    /*now, free the class from the chunk. */
    DgpAdt_chFree(ps->ClassMemChunk, c);
}


void EndHalfTokenizer(pHalfTokState ps) {

    /*remove the chunk that contained the classes. */
    DgpAdt_chDestroy(ps->ClassMemChunk);
    
    /*the individual trees should have been freed  */
    /*by a preceeding HalfTokFinal */

    /*now, just blow away the struct */
    ps->a->free(ps);
}

