/* */
/*  This module implements the prefix/suffix binary search tree, */
/*  and the hamming distance tokenizer. */
/* */
/* */


#ifndef HAMTOK_H
#define HAMTOK_H

#define LONGSIBS    8	/*the number of longs in the bit_string, meaning we can  */
			/*store LONGSIBS x 32 length strings */
#define MAXSTRLEN   (LONGSIBS * 32)

#include "dgp_allocators.h"
/*#include "chunk.h" */
#include "tok-lossless.h"


#ifdef __cplusplus
extern "C"
{
#endif

/*the tree node  */

typedef struct tnode {
    struct tnode* pOne;		/*branch 1 */
    struct tnode* pZero;	/*branch 2 */
    struct tnode* pParent;
    unsigned long prefix[LONGSIBS];	/*the common prefix stored at this node.  */
    int		  prefix_len;		/*the length of the common prefix stored at this node */
    tokClass      pData;
} TNode;

typedef TNode* pTNode;

/*the tree is essentially just a pointer to root */
/*and our Dgp_Allocators.  */

typedef struct {
    pTNode pRoot;
    Dgp_Allocators *a;
    DgpAdt_Chunk MyMemChunk;
    int len;
} BTree;

typedef BTree* pBTree;


typedef struct {
    int next_toknum;	    /*the toknum for the tokenizer */
    DgpAdt_Chunk ClassMemChunk;    /*from where will we allocate classes */
    Dgp_Allocators *a;
    pBTree Trees[MAXSTRLEN+1];	
	/*We are going to have a tree for 1..MAXSTRLEN */
	/*length (w x h) halftone tokens. Usually, */
	/*just one of these trees will be populated */
	/*with classes, but in a document, */
	/*it is conceivable that an incorrect judgment */
	/*with regards to halftones will create another cell_size, */
	/*which we'll need to tokenize. Also, different tiling */
	/*designs mat cause another token size at the borders of the */
	/*halftone image */
} HalfTokState;

typedef HalfTokState* pHalfTokState;


long mem_alloced;



pBTree NewTree(Dgp_Allocators *a, int len);
int TInsert(pBTree pTree, int *bit_string, tokClass user_data, int *new_tok, int *next_toknum);
int SearchHam(pBTree pTree, int *search_for, int *found, int ham_dist, void **user_data);
void RunsToBitString(run_elt *runs, int num_runs, int *bit_string, int side_length_w, int side_length_h);
void FreeTree(pBTree pTree);
int AddHalfTok(pHalfTokState p_state, run_elt *runs, int num_runs, 
	       int *bit_string, int side_len_w, int side_len_h, int *new_tok);
DgpAdt_List HalfTokFinal(pHalfTokState p_state);
pHalfTokState InitHalfTokenizer(Dgp_Allocators *a);
void EndHalfTokenizer(pHalfTokState ps);
void FreeHalfTokClass(pHalfTokState ps, tokClass c);

#ifdef __cplusplus
}
#endif

#endif
