
#include <stdio.h>
#include "intellibuffer.h"
#include "PsgFix.h"

/* temp include */
#include "dgp_err_printers.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libibuf/intellibuffer.c,v 1.16 2003/02/10 22:53:25 mkrocken Exp $");

#define IBUF_DEF_BLOCK_SIZE  65536
#define IBUF_DEF_NUM_BLOCKS 25

#define IBUF_ERROR -1

/* internal flags */
#define IBUF_TMPFILE        0x00000100
#define IBUF_UDSEOF         0x00000200

#define ibuf_read(_stream) ((((_stream)->flags) & DGPIBUF_READ) != 0)
#define ibuf_write(_stream) ((((_stream)->flags) & DGPIBUF_WRITE) != 0)
#define ibuf_limitUds(_stream) ((((_stream)->flags) & DGPIBUF_LIMITUDS) != 0)
#define ibuf_limitMem(_stream) ((((_stream)->flags) & DGPIBUF_LIMITMEM) != 0)

#define ibuf_isTmpFile(_stream) ((((_stream)->flags) & IBUF_TMPFILE) != 0)
#define ibuf_udsEOF(_stream) ((((_stream)->flags) & IBUF_UDSEOF) != 0)

#define cur_mem_usage(_num_blocks, _max_num_blocks, _block_size)

/* minumum memory allowance */
#define IBUF_MIN_MEM (sizeof(ibuf_stream) + sizeof(Dgp_Int8 *) + sizeof(Dgp_Int8))

#define cur_filepos(_block_num, _offset, _block_size) \
  (((_block_num) * (_block_size)) + (_offset))

typedef struct {
  Dgp_State *dgp_state;

  /* size of each block */
  Dgp_UInt32 block_size;

  /* current memory  */
  Dgp_Int32 first_block_num;
  Dgp_Int32 last_block_num;
  Dgp_UInt32 last_block_offset;
  Dgp_Int8 **blocks;

  /* current filepos */
  Dgp_Int32 cur_block_num;
  Dgp_UInt32 cur_block_offset;

  /* internal uds */
  uds u;

  /* init info */
  Dgp_UInt32 flags;

  /* resource limits */
  Dgp_Int32 max_mem_blocks;
  Dgp_Int32 max_uds_blocks;
} ibuf_stream;




static Dgp_Bool
ibuf_flushFirstBlock(ibuf_stream *stream, Dgp_Int8 **_block)
{
  Dgp_UInt32 tmp, filepos;


  /* write the old buffer if necessary and/or possible */
  if(ibuf_write(stream) && 
     (!ibuf_limitUds(stream) || (stream->max_uds_blocks != DGPIBUF_NOUDS))) {

    /* figure out where we need to write it */
    if(!ibuf_limitUds(stream)) {
      tmp = stream->first_block_num;
    } else {
      tmp = (Dgp_UInt32)(stream->first_block_num % stream->max_uds_blocks);
    }

    filepos = cur_filepos(tmp, 0, stream->block_size);

    /* write it out */
    stream->u.fseek(stream->u.fp, filepos, SEEK_SET);
    tmp = stream->u.fwrite(stream->blocks[stream->first_block_num % stream->max_mem_blocks],
      sizeof(Dgp_Int8), stream->block_size, stream->u.fp);

    /* didn't write it all? */
    if(tmp != stream->block_size) return DGP_FALSE;

  }

  *_block = stream->blocks[stream->first_block_num % stream->max_mem_blocks];

  stream->first_block_num++;


  return DGP_TRUE;
}


static Dgp_Bool
ibuf_loadCurBlocks(ibuf_stream *stream, Dgp_UInt32 num_bytes, Dgp_Bool doing_fread)
{
  Dgp_State *dgp_state = stream->dgp_state;
  DGP_PROLOGUE("ibuf_loadCurBlocks()", DGP_PERRSTATE);
  Dgp_Int32 num_blocks_alloced = stream->last_block_num - stream->first_block_num + 1;
  Dgp_UInt32 utmp;
  Dgp_Int32 tmp, num_blocks;
  Dgp_Int8 *curblock = NULL;
  Dgp_UInt32 filepos;

  Dgp_Int32 new_last_block_num;
  Dgp_UInt32 new_last_block_offset;


  /* figure out where the new mem eof is */
  filepos = cur_filepos(stream->cur_block_num, stream->cur_block_offset, stream->block_size)
    + num_bytes;
  new_last_block_num = (Dgp_UInt32)(filepos / stream->block_size);
  new_last_block_offset = (Dgp_UInt32)(filepos % stream->block_size);

  /* if we have no mem limit, make sure the buffer is big enough */
  if((!ibuf_limitMem(stream)) && 
     ((new_last_block_num + 1) > stream->max_mem_blocks)) {
    
    tmp = stream->max_mem_blocks + (Dgp_UInt32)(stream->max_mem_blocks / 2);
    if(tmp < (new_last_block_num + 1)) tmp = new_last_block_num + 1;

    DGP_SEC_REALLOC(stream->blocks, (Dgp_Int8 **), (sizeof(Dgp_Int8 *) * tmp));
    stream->max_mem_blocks = tmp;

  } else if((ibuf_limitMem(stream)) && 
    ((new_last_block_num - stream->cur_block_num + 1) > stream->max_mem_blocks)) {

    /* uh-oh, too much, need to shave it down for now */
    new_last_block_num = stream->cur_block_num + stream->max_mem_blocks - 1;
    new_last_block_offset = stream->block_size;

  }


  if(doing_fread) {

    /* set the uds filepos */
    filepos = cur_filepos(stream->last_block_num + 1, 0, stream->block_size);
    stream->u.fseek(stream->u.fp, filepos, SEEK_SET);

    /* if we are reading, we need to go one block at a time (we need to  */
    /* stop at uds EOF) */
    utmp = 0;
    while(stream->last_block_num < new_last_block_num) {
    
      if(num_blocks_alloced == stream->max_mem_blocks) {

        /* we need to flush an old block */
        if(!ibuf_flushFirstBlock(stream, &curblock)) goto ERR_CLEANUP;

        /* we should reset the filepos just to be safe */
        stream->u.fseek(stream->u.fp, filepos, SEEK_SET);
      
      } else {

        /* we need to allocate a new block */
        DGP_SEC_MALLOC(curblock, (Dgp_Int8 *), (sizeof(Dgp_Int8) * stream->block_size));
        num_blocks_alloced++;

      }

      stream->last_block_num++;
      stream->blocks[stream->last_block_num % stream->max_mem_blocks] = curblock;
      utmp = stream->u.fread(curblock, sizeof(Dgp_Int8), stream->block_size,
        stream->u.fp);
      curblock = NULL;
      filepos += utmp;

      if(utmp < stream->block_size) {
        stream->flags |= IBUF_UDSEOF;
    
        /* stop cause we're at uds EOF */
        break;
      }

    }

    stream->last_block_offset = utmp;

  } else {

    /* we can go whole hog! */
    num_blocks = new_last_block_num - stream->last_block_num;
    if(num_blocks > stream->max_mem_blocks) num_blocks = stream->max_mem_blocks;

    /* flush as many as necessary */
    tmp = stream->last_block_num + 1;
    while(num_blocks > 0) {

      if(num_blocks_alloced < stream->max_mem_blocks) {

        /* we need to allocate a new block */
        DGP_SEC_MALLOC(stream->blocks[tmp % stream->max_mem_blocks], (Dgp_Int8 *), 
          (sizeof(Dgp_Int8) * stream->block_size));
        num_blocks_alloced++;

      } else {

        /* otherwise, flush an old one */
        if(!ibuf_flushFirstBlock(stream, &(stream->blocks[tmp % stream->max_mem_blocks]))) 
          goto ERR_CLEANUP;

      }
      
      tmp++;
      num_blocks--;
    }

    /* if we're at uds EOF, we're done */
    if(ibuf_udsEOF(stream)) goto DONE_READ;

    /* set the uds filepos */
    filepos = cur_filepos(stream->last_block_num + 1, 0, stream->block_size);
    stream->u.fseek(stream->u.fp, filepos, SEEK_SET);

    /* now we should have all the blocks we need, load as necessary */
    while(stream->last_block_num < new_last_block_num) {

      stream->last_block_num++;
      utmp = stream->u.fread(stream->blocks[stream->last_block_num % stream->max_mem_blocks], 
        sizeof(Dgp_Int8), stream->block_size, stream->u.fp);

      if(utmp < stream->block_size) {
        stream->flags |= IBUF_UDSEOF;

        /* we don't need to load any more "old data" */
        break;
      }
    }


DONE_READ:

    stream->last_block_num = new_last_block_num;
    stream->last_block_offset = new_last_block_offset;

  }


  return DGP_TRUE;

ERR_CLEANUP:

  if(curblock != NULL) DGP_FREE(curblock);

  return DGP_FALSE;
}



ErrCode
DgpIbuf_fopen(void **_stream, 
              uds *u,
              Dgp_UInt32 mem_block_size,
              Dgp_UInt32 flags, 
              Dgp_UInt32 max_mem_usage,
              Dgp_UInt32 max_disk_usage, 
              Dgp_State *dgp_state)
{
  DGP_STATE_PROLOGUE2("DgpIbuf_init()");
  ibuf_stream *stream = NULL;

  DGP_ERR_BEGIN(dgp_state);
  DGP_CHECK_NULL(u);
  DGP_ERR_GENF0(((flags & DGPIBUF_LIMITMEM) && (max_mem_usage < IBUF_MIN_MEM)), 
    cErrInvalidParm, "max mem setting is too small");

  *_stream = NULL;

  DGP_SEC_CALLOC(stream, (ibuf_stream *), 1, sizeof(ibuf_stream));
  stream->dgp_state = dgp_state;
  stream->flags = flags;
  stream->u = *u;

  if(mem_block_size == DGPIBUF_DEFAULTBLOCKSIZE) {
    stream->block_size = IBUF_DEF_BLOCK_SIZE;
  } else {
    stream->block_size = mem_block_size;
  }

  if(ibuf_limitUds(stream) && (max_disk_usage != DGPIBUF_NOUDS)) {
    /* figure out how many blocks we can have on disk */
    stream->max_uds_blocks = (Dgp_UInt32)(max_disk_usage / stream->block_size);
  }

  if(ibuf_limitMem(stream)) {

    /* account for struct overhead */
    max_mem_usage -= sizeof(ibuf_stream);

    /* figure out how many blocks we can have in memory */
    stream->max_mem_blocks = (Dgp_UInt32)(max_mem_usage / (stream->block_size + sizeof(Dgp_Int8 *)));
  
    if(stream->max_mem_blocks == 0) {
      stream->max_mem_blocks = 1;
      stream->block_size = max_mem_usage - sizeof(Dgp_Int8) - sizeof(ibuf_stream);
    }

  } else {

    stream->max_mem_blocks = IBUF_DEF_NUM_BLOCKS;

  }

  /* allocate the block array */
  DGP_SEC_MALLOC(stream->blocks, (Dgp_Int8 **), (stream->max_mem_blocks *
    sizeof(Dgp_Int8 *)));

  /* indicate no blocks alloced */
  stream->last_block_num = -1;

  *_stream = stream;
  stream = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  DgpIbuf_fclose((void **)&stream);

  ERR_RETURN;
}


ErrCode
DgpIbuf_mopen(void **_stream, 
	      Dgp_UInt8* buffer,
              Dgp_UInt32 buf_size,
              Dgp_UInt32 flags, 
              Dgp_State *dgp_state)
{
  DGP_STATE_PROLOGUE2("DgpIbuf_mopen()");
  ibuf_stream *stream = NULL;

  DGP_ERR_BEGIN(dgp_state);
  DGP_CHECK_NULL(buffer);
  DGP_CHECK_NULL(dgp_state);

  *_stream = NULL;

  DGP_SEC_CALLOC(stream, (ibuf_stream *), 1, sizeof(ibuf_stream));
  stream->dgp_state = dgp_state;
  stream->flags = flags;
  stream->u.fp = NULL;

  stream->block_size = buf_size;
  stream->max_uds_blocks = 0;
  stream->max_mem_blocks = 2;

  /* create 2 blocks to make everybody happy */
  DGP_SEC_MALLOC(stream->blocks, (Dgp_Int8 **), (stream->max_mem_blocks *
    sizeof(Dgp_Int8 *)));
  stream->blocks[0] = buffer;

  /* indicate there are 2 blocks, even though only the first one has data */
  stream->cur_block_offset = 0;
  stream->last_block_offset = 0;
  stream->cur_block_num = 0;
  stream->first_block_num = 0;
  stream->last_block_num = 1;

  stream->flags |= IBUF_TMPFILE | IBUF_UDSEOF;

  *_stream = stream;
  stream = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  DgpIbuf_mclose((void **)&stream);

  ERR_RETURN;
}


ErrCode
DgpIbuf_tmpfile(void **_stream, 
                Dgp_UInt32 mem_block_size,
                Dgp_UInt32 flags, 
                Dgp_UInt32 max_mem_usage,
                Dgp_UInt32 max_disk_usage, 
                Dgp_State *dgp_state)
{
  DGP_STATE_PROLOGUE2("DgpIbuf_tmpfile()");
  uds u ;


  DGP_ERR_BEGIN(dgp_state);
  Dgp_initUds(dgp_state, &u, NULL);
  DGP_ERR_GEN((flags & DGPIBUF_NOUDS), cErrInvalidParm);

  /* create an actual tmpfile if necessary */
  if((flags & DGPIBUF_LIMITMEM) != 0) {
    u.fp = tmpfile();
    DGP_CHECK_NULL(u.fp);
  }

  /* now do the real work of creating the memory buffers */
  DGP_ERR_PASS_CLEANUP(DgpIbuf_fopen(_stream, &u, mem_block_size, flags, max_mem_usage,
    max_disk_usage, dgp_state));

  /* remember that we are using an internal temp file */
  ((ibuf_stream *)(*_stream))->flags |= IBUF_TMPFILE | IBUF_UDSEOF;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(u.fp != NULL) fclose(u.fp);

  ERR_RETURN;
}


ErrCode
DgpIbuf_fclose(void **_stream)
{
  DGP_STATE_PROLOGUE2("DgpIbuf_fclose()");
  ibuf_stream *stream;
  Dgp_State *dgp_state;
  Dgp_Int8 *tmpblock;

  if(_stream == NULL) NORMAL_RETURN;

  stream = *_stream;
  *_stream = NULL;

  if(stream == NULL) NORMAL_RETURN;

  dgp_state = stream->dgp_state;

  /* flush buffers */
  if(stream->blocks != NULL) {
    if(ibuf_write(stream) && !ibuf_isTmpFile(stream)) {
      while(stream->first_block_num <= stream->last_block_num) {
        DGP_ERR_GEN(!ibuf_flushFirstBlock(stream, &tmpblock), cErrInternal);
        DGP_FREE(tmpblock);
      }
    } else {   
      while(stream->first_block_num <= stream->last_block_num) {
        DGP_FREE(stream->blocks[stream->first_block_num % stream->max_mem_blocks]);
        stream->first_block_num++;
      }
    }
  
    DGP_FREE(stream->blocks);
    stream->blocks = NULL;
  }


  /* close tmp file */
  if((ibuf_isTmpFile(stream)) && (stream->u.fp != NULL)) {
    fclose(stream->u.fp);
    stream->u.fp = NULL;
  }
  
  /* free struct */
  DGP_FREE(stream);

  NORMAL_RETURN;
}

ErrCode
DgpIbuf_mclose(void **_stream)
{
  DGP_STATE_PROLOGUE2("DgpIbuf_mclose()");
  ibuf_stream *stream;
  Dgp_State *dgp_state;

  DGP_STOP_WHINE;

  if(_stream == NULL) NORMAL_RETURN;

  stream = *_stream;
  *_stream = NULL;

  if(stream == NULL) NORMAL_RETURN;

  dgp_state = stream->dgp_state;

  /* delete buffer */
  if (stream->blocks[0] != NULL)
    DGP_FREE(stream->blocks[0]);
  DGP_FREE(stream->blocks);
  
  /* free struct */
  DGP_FREE(stream);

  NORMAL_RETURN;
}

Dgp_Int32
DgpIbuf_fseek(void *_stream, 
              Dgp_Int32 offset, 
              Dgp_Int32 origin)
{
  ibuf_stream *stream;
  Dgp_Int32 tmp;

  if(_stream == NULL) return IBUF_ERROR;

  stream = (ibuf_stream *)_stream;

  /* make offset an absolute position */
  switch(origin) {
  case SEEK_CUR:

    /* offset from current pos */
    offset += cur_filepos(stream->cur_block_num, stream->cur_block_offset, stream->block_size);
    break;


  case SEEK_END:

    /* offset from end (first figure out where that is) */
    tmp = (Dgp_Int32)cur_filepos(stream->last_block_num, stream->last_block_offset, 
      stream->block_size);
    if(!ibuf_udsEOF(stream)) {
      stream->u.fseek(stream->u.fp, 0, SEEK_END);
      if(tmp < stream->u.ftell(stream->u.fp)) tmp = stream->u.ftell(stream->u.fp);
    }
    offset += tmp;
    break;

  }

  if(offset < 0) return IBUF_ERROR;

  /* change current position */
  stream->cur_block_num = (Dgp_Int32)(offset / stream->block_size);
  stream->cur_block_offset = (Dgp_Int32)(offset % stream->block_size);

  return 0;
}

Dgp_Int32 
DgpIbuf_ftell(void *_stream)
{
  ibuf_stream *stream;

  if(_stream == NULL) {
    return IBUF_ERROR;
  }

  stream = (ibuf_stream *)_stream;

  return cur_filepos(stream->cur_block_num, stream->cur_block_offset, stream->block_size);
}


static size_t
ibuf_ufread(ibuf_stream *stream, 
            Dgp_Int8 *buffer, 
            Dgp_UInt32 bytes_to_read)
{
  Dgp_UInt32 bytes_read;
  Dgp_UInt32 tmp, filepos_start, filepos_end;
  
  /* check if we still have the stuff */
  if((stream->u.fp == NULL) ||
     (ibuf_limitUds(stream) && 
      (stream->first_block_num - stream->cur_block_num > stream->max_uds_blocks))) return 0;;

  /*printf("hit uds! cur %d %d, first %d %d, last %d %d\n",
    stream->cur_block_num, stream->cur_block_offset,
    stream->first_block_num, 0,
    stream->last_block_num, stream->last_block_offset);*/

  /* figure out where it is in the file */
  if(!ibuf_limitUds(stream)) {

    /* simple case */
    filepos_start = cur_filepos(stream->cur_block_num, stream->cur_block_offset, 
      stream->block_size);
    stream->u.fseek(stream->u.fp, filepos_start, SEEK_SET);
    bytes_read = stream->u.fread(buffer, sizeof(Dgp_Int8), bytes_to_read, stream->u.fp);

  } else {

    /* trickier case (the disk space is limited, so */
    /* the buffers may be wrapped around in the file) */
    filepos_start = cur_filepos(stream->cur_block_num % stream->max_uds_blocks, 
      stream->cur_block_offset, stream->block_size);
    filepos_end = cur_filepos(stream->max_uds_blocks, 0, stream->block_size);

    tmp = filepos_end - filepos_start;
    stream->u.fseek(stream->u.fp, filepos_start, SEEK_SET);
    bytes_read = stream->u.fread(buffer, sizeof(Dgp_Int8), tmp, stream->u.fp);

    if(bytes_read < tmp) goto DONE_READ;

    buffer += bytes_read;

    /* the wrap around case (we need to do some  */
    /* more reading from the beginning of the file) */
    if(bytes_read < bytes_to_read) {

      tmp = bytes_to_read - bytes_read;
      stream->u.fseek(stream->u.fp, 0, SEEK_SET);
      bytes_read += stream->u.fread(buffer, sizeof(Dgp_Int8), tmp, stream->u.fp);

    }
  }
  
DONE_READ:

  tmp = cur_filepos(stream->cur_block_num, stream->cur_block_offset, 
      stream->block_size) + bytes_read;

  /* move current filepos */
  stream->cur_block_num = tmp / stream->block_size;
  stream->cur_block_offset = tmp % stream->block_size;


  return(bytes_read);
}


size_t 
DgpIbuf_fread(void *buffer, 
              size_t size, 
              size_t count, 
              void *_stream)
{
  Dgp_State *dgp_state;
  ibuf_stream *stream;
  Dgp_UInt32 tmp;
  Dgp_UInt32 bytes_to_read, num_bytes;
  
  if((_stream == NULL) || (buffer == NULL) || (size == 0) || (count == 0)) return 0;

  stream = (ibuf_stream *)_stream;

  /* check access */
  if(!ibuf_read(stream)) return 0;

  bytes_to_read = (size * count);
  dgp_state = stream->dgp_state;
  
  /* see if we need to read anything old from our uds */
  if(stream->cur_block_num < stream->first_block_num) {

    num_bytes = cur_filepos(stream->first_block_num, 0, stream->block_size) - 
      cur_filepos(stream->cur_block_num, stream->cur_block_offset, stream->block_size);
    if(num_bytes > bytes_to_read) num_bytes = bytes_to_read;
    tmp = ibuf_ufread(stream, buffer, num_bytes);
    bytes_to_read -= tmp;
    buffer = ((Dgp_Int8 *)buffer) + tmp;

    /* make sure we got what we needed */
    if(tmp != num_bytes) goto DONE_READ;

  }

  /* now read from memory */
  while(bytes_to_read > 0) {

    if(stream->cur_block_offset == stream->block_size) {
      /* end of block, move to next */
      stream->cur_block_num++;
      stream->cur_block_offset = 0;
    }

    if(stream->cur_block_num > stream->last_block_num) {
      /* this buffer doesn't exist right now... */
      if(ibuf_udsEOF(stream) ||
         !ibuf_loadCurBlocks(stream, bytes_to_read, DGP_TRUE)) goto DONE_READ;
    }

    /* copy over as many bytes as we can from this block */
    if(stream->cur_block_num < stream->last_block_num) {
      num_bytes = stream->block_size - stream->cur_block_offset;
    } else {
      num_bytes = stream->last_block_offset - stream->cur_block_offset;
      
      /* have we reached memory EOF? */
      if(num_bytes == 0) goto DONE_READ;
    }

    if(num_bytes > bytes_to_read) num_bytes = bytes_to_read;
    memcpy(buffer, (void *)(stream->blocks[stream->cur_block_num % stream->max_mem_blocks] + 
      stream->cur_block_offset), num_bytes);
    buffer = ((Dgp_Int8 *)buffer) + num_bytes;
    bytes_to_read -= num_bytes;
    stream->cur_block_offset += num_bytes;

  }

DONE_READ:

  return(((size * count) - bytes_to_read) / size);

}


static size_t
ibuf_ufwrite(ibuf_stream *stream, 
             Dgp_Int8 *buffer, 
             Dgp_UInt32 bytes_to_write)
{
  Dgp_UInt32 bytes_written;
  Dgp_UInt32 tmp, filepos_start, filepos_end;
  
  /* check if we still have the stuff */
  if((stream->u.fp == NULL) ||
     (ibuf_limitUds(stream) && 
      (stream->first_block_num - stream->cur_block_num > stream->max_uds_blocks))) return 0;;

  /* figure out where it is in the file */
  if(!ibuf_limitUds(stream)) {

    /* simple case */
    filepos_start = cur_filepos(stream->cur_block_num, stream->cur_block_offset, 
      stream->block_size);
    stream->u.fseek(stream->u.fp, filepos_start, SEEK_SET);
    bytes_written = stream->u.fwrite(buffer, sizeof(Dgp_Int8), bytes_to_write, stream->u.fp);

  } else {

    /* trickier case (the disk space is limited, so */
    /* the buffers may be wrapped around in the file) */
    filepos_start = cur_filepos(stream->cur_block_num % stream->max_uds_blocks, 
      stream->cur_block_offset, stream->block_size);
    filepos_end = cur_filepos(stream->max_uds_blocks, 0, stream->block_size);

    tmp = filepos_end - filepos_start;
    stream->u.fseek(stream->u.fp, filepos_start, SEEK_SET);
    bytes_written = stream->u.fwrite(buffer, sizeof(Dgp_Int8), tmp, stream->u.fp);

    if(bytes_written < tmp) goto DONE_WRITE;

    buffer += bytes_written;

    /* the wrap around case (we need to do some  */
    /* more writing from the beginning of the file) */
    if(bytes_written < bytes_to_write) {

      tmp = bytes_to_write - bytes_written;
      stream->u.fseek(stream->u.fp, 0, SEEK_SET);
      bytes_written += stream->u.fwrite(buffer, sizeof(Dgp_Int8), tmp, stream->u.fp);

    }
  }
  
DONE_WRITE:

  tmp = cur_filepos(stream->cur_block_num, stream->cur_block_offset, 
      stream->block_size) + bytes_written;

  /* move current filepos */
  stream->cur_block_num = tmp / stream->block_size;
  stream->cur_block_offset = tmp % stream->block_size;


  return(bytes_written);
}


size_t 
DgpIbuf_fwrite(void *buffer, 
               size_t size, 
               size_t count, 
               void *_stream)
{
  Dgp_State *dgp_state;
  ibuf_stream *stream;
  Dgp_UInt32 tmp;
  Dgp_UInt32 bytes_to_write, num_bytes;
  
  if((_stream == NULL) || (buffer == NULL) || (size == 0) || (count == 0)) return 0;

  stream = (ibuf_stream *)_stream;

  /* check access */
  if(!ibuf_write(stream)) return 0;

  bytes_to_write = (size * count);
  dgp_state = stream->dgp_state;
  
  /* see if we need to write anything old to our uds */
  if(stream->cur_block_num < stream->first_block_num) {

    num_bytes = cur_filepos(stream->first_block_num, 0, stream->block_size) - 
      cur_filepos(stream->cur_block_num, stream->cur_block_offset, stream->block_size);
    if(num_bytes > bytes_to_write) num_bytes = bytes_to_write;
    tmp = ibuf_ufwrite(stream, buffer, num_bytes);
    bytes_to_write -= tmp;
    buffer = ((Dgp_Int8 *)buffer) + tmp;

    /* make sure we got what we needed */
    if(tmp != num_bytes) goto DONE_WRITE;

  }

  /* now write to memory */
  while(bytes_to_write > 0) {

    /* end of block, move to next */
    if(stream->cur_block_offset == stream->block_size) {
      stream->cur_block_num++;
      stream->cur_block_offset = 0;
    }
    
    if(stream->cur_block_num > stream->last_block_num) {
      /* this buffer doesn't exist right now... */
      if(!ibuf_loadCurBlocks(stream, bytes_to_write, DGP_FALSE)) goto DONE_WRITE;
    }

    /* copy over as many bytes as we can from this block */
    num_bytes = stream->block_size - stream->cur_block_offset;
    if (num_bytes > bytes_to_write)
    {
      num_bytes = bytes_to_write;
    }

    memcpy((void *)(stream->blocks[stream->cur_block_num % stream->max_mem_blocks] + 
      stream->cur_block_offset), buffer, num_bytes);
    buffer = ((Dgp_Int8 *)buffer) + num_bytes;
    bytes_to_write -= num_bytes;
    stream->cur_block_offset += num_bytes;

  }

DONE_WRITE:

    /* reset eof if necessary */
    if((stream->cur_block_num == stream->last_block_num) &&
      (stream->cur_block_offset > stream->last_block_offset)) {
      stream->last_block_offset = stream->cur_block_offset;
    }


  return(((size * count) - bytes_to_write) / size);

}


ErrCode
DgpIbuf_fcopy(uds *u_dst, 
              void *_stream_src)
{
  DGP_STATE_PROLOGUE2("DgpIbuf_fcopy()");
  ibuf_stream *stream;
  Dgp_UInt32 filepos;
  Dgp_UInt32 tmp;
  Dgp_UInt32 num_bytes;
  Dgp_Int8 *tmpblock = NULL;

  
  DGP_CHECK_NULL(_stream_src);
  stream = (ibuf_stream *)_stream_src;
  DGP_ERR_BEGIN(stream->dgp_state);
  DGP_CHECK_NULL(u_dst);


  /* "fseek" to beginnig */
  stream->cur_block_num = 0;
  stream->cur_block_offset = 0;
  filepos = 0;

  /* reset dst */
  u_dst->fseek(u_dst->fp, 0, SEEK_SET);

  /* see if we need to copy anything from our */
  /* internal uds */
  if(stream->cur_block_num < stream->first_block_num) {

    /* see if we can get a mem block */
    if((ibuf_limitMem(stream)) && 
       ((stream->last_block_num - stream->first_block_num + 1) == stream->max_mem_blocks)) {
      /* doh!  need to flush one first... */
      DGP_ERR_GEN((!ibuf_flushFirstBlock(stream, &tmpblock)), cErrInternal);
    } else {
      /* just allocate a block */
      DGP_CHECK_ALLOC_CLEANUP(tmpblock = 
        (Dgp_Int8 *)(stream->DGP_MALLOC)(sizeof(Dgp_Int8) * stream->block_size));
    }

    /* see if we've lost any disk blocks... */
    if(ibuf_limitUds(stream) &&
      ((stream->first_block_num - stream->cur_block_num) > stream->max_uds_blocks)) {
      stream->cur_block_num = stream->first_block_num - stream->max_uds_blocks;
      tmp = stream->max_uds_blocks;
    } else {
      tmp = stream->first_block_num + 1;
    }

    filepos = cur_filepos((stream->cur_block_num % tmp), 0, stream->block_size);
    stream->u.fseek(stream->u.fp, filepos, SEEK_SET);
    while(stream->cur_block_num < stream->first_block_num) {

      /* read from our uds */
      num_bytes = stream->u.fread(tmpblock, sizeof(Dgp_Int8), stream->block_size, stream->u.fp);
      
      /* write to dest uds */
      num_bytes = u_dst->fwrite(tmpblock, sizeof(Dgp_Int8), num_bytes, u_dst->fp);

      stream->cur_block_num++;

      /* if we read less than we thought, we've reached EOF */
      if(num_bytes < stream->block_size) break;

      if((Dgp_UInt32)(stream->cur_block_num % tmp) == 0) {
        /* wrap around, need to move to beginning of our uds */
        stream->u.fseek(stream->u.fp, 0, SEEK_SET);
      }
    }

    /* we should be at first_block_num now... */
    DGP_ERR_GEN((stream->cur_block_num < stream->first_block_num), cErrInternal);
  }


  /* now copy any memory we have */
  while(stream->cur_block_num <= stream->last_block_num) {

    if(stream->cur_block_num != stream->last_block_num) {
      num_bytes = stream->block_size;
    } else {
      num_bytes = stream->last_block_offset;
    }
    
    tmp = u_dst->fwrite(stream->blocks[stream->cur_block_num % stream->max_mem_blocks], 
      sizeof(Dgp_Int8), num_bytes, u_dst->fp);

    stream->cur_block_num++;

    DGP_ERR_GEN_CLEANUP((tmp != num_bytes), cErrInternal);

  }
  
  /* could there still be something in the internal uds? */
  if(!ibuf_udsEOF(stream)) {

    /* get a buffer if we need it */
    if(tmpblock == NULL) {
      DGP_ERR_GEN((!ibuf_flushFirstBlock(stream, &tmpblock)), cErrInternal);
    }
 
    /* finally, copy over anything else still in uds */
    filepos = cur_filepos(stream->last_block_num, stream->last_block_offset, stream->block_size);
    stream->u.fseek(stream->u.fp, filepos, SEEK_SET);
    while(DGP_TRUE) {

      /* read from our uds */
      num_bytes = stream->u.fread(tmpblock, sizeof(Dgp_Int8), stream->block_size, stream->u.fp);

      if(num_bytes == 0) break;

      /* write to dest uds */
      tmp = u_dst->fwrite(tmpblock, sizeof(Dgp_Int8), num_bytes, u_dst->fp);

      DGP_ERR_GEN_CLEANUP((tmp != num_bytes), cErrInternal);

    }

  }

  if(tmpblock != NULL) (stream->DGP_FREE)(tmpblock);

  NORMAL_RETURN;

ERR_CLEANUP:

  if(tmpblock != NULL) (stream->DGP_FREE)(tmpblock);

  ERR_RETURN;
}
