/*
*
* $Header: /project/icons/cvsroot/digipaper/src/libimage/findcc.c,v 1.3 2003/01/27 21:48:32 awilliam Exp $
*
*/


#include <stdlib.h>
#include <sys/types.h>
#ifdef _WIN32
#include <memory.h>
#endif
#include "util.h"
#include "list.h"
#include "image.h"
#include "image_defs.h"
#include "imconv.h"
#include "dgp_rcs.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libimage/findcc.c,v 1.3 2003/01/27 21:48:32 awilliam Exp $");

static ErrCode
dgpimage_find_cc(Dgp_State *dgp_state, DgpImage_Binary im, Dgp_Bool eightconnected, DgpAdt_List *rlist);
static ErrCode
dgpimage_pullout(Dgp_State *dgp_state, DgpImage_Binary im, int x, int y,
		Dgp_Bool eightconnected, DgpImage_Component **rcomp);

/* static void *myrealloc(Dgp_State *dgp_state, void *addr, size_t oldsize, size_t newsize); */

/*
* Neighbour lists. Must be in circular CW order (CW in computer-Cartesian
* coords, that is).
*/
static DgpImage_Point fourneigh[4] = {
    {1, 0}, {0, 1}, {-1, 0}, {0, -1}
};

/* Additional constraint: diagonals should be in odd positions. */
static DgpImage_Point eightneigh[8] = {
    {1, 0}, {1, 1}, {0, 1}, {-1, 1}, {-1, 0}, {-1, -1}, {0, -1}, {1, -1}
};

/*
* Pick chunks out of the image. Do this by looking at a dilation of the
* image (actually, any image which is a superset will do; can be the
* same as the original image), picking connected components out of
* this, and reporting all the original pixels that lie under one of these
* connected components as a single chunk. The returned things may not
* themselves be connected, but if you look at their extension in dilated,
* that will be. If im == dilated, this is just a simple pick connected
* components routine.
*
* The return value is a List which contains one entry for each chunk; the
* entry's userdata is a Component describing the extracted component.
* The offset field describes what location in the original image (im)
* the (0,0) position of the Component's DgpImage_Binary used to be. The
* DgpImage_Binary's offset is (0,0).
*
* If eightconnected is TRUE, it finds 8-connected components; if it is FALSE,
* it finds 4-connected components.
*/
ErrCode
DgpImage_pick_cc(DgpImage_Binary im, DgpImage_Binary dilated, 
			   Dgp_Bool eightconnected, DgpAdt_List *rlist)
{
    DgpAdt_List gllist = DGP_NULLLIST;
    DgpAdt_ListNode ln;
    DgpAdt_ListNode newln;
    DgpImage_Component *newg;
    DgpImage_Component *g;
    DgpAdt_List cclist = DGP_NULLLIST;
    DgpImage_Binary dilatedcc;
    DgpImage_Binary cc;
    int minx = 0, miny = 0, maxx = 0, maxy = 0;
    int x, y;
    int lowx, lowy, highx, highy;
	Dgp_State *dgp_state = im->dgp_state;
	
	DGP_PROLOGUE ("DSUTIL_pick_cc", DGP_PERRSTATE);

    DGP_ASSERT(im != (DgpImage_Binary)NULL);
    DGP_ASSERT(dilated != (DgpImage_Binary)NULL);
	DGP_ASSERT(rlist != NULL);

    /* Find connected components in the *dilated* image */
    DGP_ERR (dgpimage_find_cc(dgp_state, dilated, eightconnected, &cclist));
	
    if (dilated == im) {
		/* All done. */
		*rlist = cclist;
		NORMAL_RETURN;
	}
	
    DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&gllist, dgp_state));
	
    /*
	* Now remangle the connected components to reveal the original bits
	* under each.
	*/
    DGPADT_FOREACH (ln, cclist) {
		g = (DgpImage_Component *)DGPADT_LIGET(ln);
		DGP_ASSERT(g != (DgpImage_Component *)NULL);
		DGP_ASSERT(g->im != (DgpImage_Binary)NULL);
		dilatedcc = g->im;
		
		/* Scan the image underneath this thing */
		/* Find the bounds of the underlying stuff */
		/* Find the intersection of this box with the original image */
		/* Note that these are *inclusive* bounds */
		lowx = MAX(DGPIMG_XBASE(dilatedcc), DGPIMG_XBASE(im) - g->offset.x);
		lowy = MAX(DGPIMG_YBASE(dilatedcc), DGPIMG_YBASE(im) - g->offset.y);
		highx = MIN(DGPIMG_XMAX(dilatedcc), DGPIMG_XMAX(im) - g->offset.x);
		highy = MIN(DGPIMG_YMAX(dilatedcc), DGPIMG_YMAX(im) - g->offset.y);
		/*
		* Since the bits in the box are a dilation of something in the
		* original image, it's likely that there are bits close to the edge.
		* In order to find the bounding box of the underlying bits, we
		* scan in from the edges.
		*/
		
		/* From the top */
		for (y = lowy; y <= highy; y++) {
			for (x = lowx; x <= highx; x++) {
				if (DGPIMG_REF(dilatedcc, x, y) && DGPIMG_REF(im, x + g->offset.x,
					y + g->offset.y)) {
					miny = y;
					break;
				}
			}
			if (x != highx + 1) {
				break;
			}
		}
		if (y == highy + 1) {
		/*
		* This was a connected component in dilated which had
		* no underlying original bits at all.
		* Nothing to do here...
			*/
			DGPIMAGE_FREE(dilatedcc);
			(DGP_FREE)((void *)g);
			continue;		/* And on to the next component */
		}
		
		DGP_ASSERT(y != highy + 1);	/* Must be something there */
		
		/* From the bottom */
		for (y = highy; y >= lowy; y--) {
			for (x = lowx; x <= highx; x++) {
				if (DGPIMG_REF(dilatedcc, x, y) && DGPIMG_REF(im, x + g->offset.x,
					y + g->offset.y)) {
					maxy = y;
					break;
				}
			}
			if (x != highx + 1) {
				break;
			}
		}
		DGP_ASSERT(y != lowy - 1);	/* Must be something there */
		
		/* From the left - only scan the strip where stuff might live */
		for (x = lowx; x <= highx; x++) {
			for (y = miny; y <= maxy; y++) {
				if (DGPIMG_REF(dilatedcc, x, y) && DGPIMG_REF(im, x + g->offset.x,
					y + g->offset.y)) {
					minx = x;
					break;
				}
			}
			if (y != maxy + 1) {
				break;
			}
		}
		DGP_ASSERT(x != highx + 1);
		
		/* From the right - ditto */
		for (x = highx; x >= lowx; x--) {
			for (y = miny; y <= maxy; y++) {
				if (DGPIMG_REF(dilatedcc, x, y) && DGPIMG_REF(im, x + g->offset.x,
					y + g->offset.y)) {
					maxx = x;
					break;
				}
			}
			if (y != maxy + 1) {
				break;
			}
		}
		DGP_ASSERT(x != lowx - 1);
		
		/*
		* The component occupies (minx..maxx, miny..maxy) in the dilatedcc
		* image, and so we can pull it out.
		*/
		DGP_ERR_PASS_CLEANUP (DGPIMAGE_NEW(dgp_state, DGPIMAGE_BINARY,
			(maxx - minx + 1),
			(maxy - miny + 1), (void**) &cc));
		
		newg = (DgpImage_Component *)(DGP_MALLOC)(sizeof(*newg));
		if (newg == (DgpImage_Component *)NULL) {
			DGPIMAGE_FREE(cc);
			goto ERR_CLEANUP;
		}
		
		newg->npts = 0;
		for (y = miny; y <= maxy; y++) {
			for (x = minx; x <= maxx; x++) {
				if (DGPIMG_REF(dilatedcc, x, y) && DGPIMG_REF(im, x + g->offset.x,
					y + g->offset.y)) {
					DGPIMG_REF(cc, x - minx, y - miny) = 1;
					newg->npts++;
				}
			}
		}
		
		newg->im = cc;
		newg->offset.x = minx + g->offset.x;	/* Offset from */
		newg->offset.y = miny + g->offset.y;	/* dilatedcc as well */
		
		DGP_ERR_CHECK (DgpAdt_liAdd(gllist, (void *)newg, &newln)) {
			/* Sigh */
			DGPIMAGE_FREE(cc);
			(DGP_FREE)((void *)newg);
			goto ERR_CLEANUP;
		}
		
		DGPIMAGE_FREE(dilatedcc);
		(DGP_FREE)((void *)g);
	}
	
	DGP_ERR_PASS_CLEANUP (DgpAdt_liFree(cclist));
	
	*rlist = gllist;
    NORMAL_RETURN;
	
ERR_CLEANUP:
	
    if (gllist != DGP_NULLLIST) {
		DGPADT_FOREACH (ln, gllist) {
			DGP_ASSERT(DGPADT_LIGET(ln) != (void *)NULL);
			DGPIMAGE_FREE(((DgpImage_Component *)DGPADT_LIGET(ln))->im);
			(DGP_FREE)(DGPADT_LIGET(ln));
		}
		DgpAdt_liFree(gllist);
	}
	
    if (cclist != DGP_NULLLIST) {
		DGPADT_FOREACH (ln, cclist) {
			DGP_ASSERT(DGPADT_LIGET(ln) != (void *)NULL);
			DGPIMAGE_FREE(((DgpImage_Component *)DGPADT_LIGET(ln))->im);
			(DGP_FREE)(DGPADT_LIGET(ln));
		}
		DgpAdt_liFree(cclist);
	}
	
    ERR_RETURN;
	
}

/*
* Find connected components in the given DgpImage_Binary.
* Return a List, with one entry for each component. The entry will contain
* a Component for that connected component. Don't do any fancy stuff like
* adding borders etc.
*/
static ErrCode
dgpimage_find_cc(Dgp_State *dgp_state, DgpImage_Binary im, Dgp_Bool eightconnected, DgpAdt_List *rlist)
{
	DgpAdt_List cclist = DGP_NULLLIST;
	DgpAdt_ListNode newln;
	DgpAdt_ListNode ln;
	DgpImage_Component *newg;
	int x, y;
	DgpImage_Binary imcopy = (DgpImage_Binary)NULL;
	int topx;
	
	DGP_PROLOGUE ("dgpimage_find_cc", DGP_PERRSTATE);

	DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&cclist, dgp_state));
	
	/* Make a copy we can trash, with a sanity border */
	DGP_ERR_PASS_CLEANUP (DGPIMAGE_NEWOFFSET(dgp_state, DGPIMAGE_BINARY,
		DGPIMG_WIDTH(im) + 2, DGPIMG_HEIGHT(im) + 2,
		DGPIMG_XBASE(im) - 1, DGPIMG_YBASE(im) - 1, (void**) &imcopy));
	
	for (y = DGPIMG_YBASE(im); y <= DGPIMG_YMAX(im); y++) {
		for (x = DGPIMG_XBASE(im); x <= DGPIMG_XMAX(im); x++) {
			DGPIMG_REF(imcopy, x, y) = DGPIMG_REF(im, x, y);
		}
	}
	
	/* Scan for something that hasn't been pulled out yet */
	topx = DGPIMG_XMAX(im);
	for (y = DGPIMG_YBASE(im); y <= DGPIMG_YMAX(im); y++) {
		for (x = DGPIMG_XBASE(im); x <= topx; x++) {
			if (DGPIMG_REF(imcopy, x, y)) {
				/* Got something - pull out everything it's connected to */
				DGP_ERR_PASS_CLEANUP (dgpimage_pullout(dgp_state, imcopy, x, y, 
						eightconnected, &newg));
				
				DGP_ASSERT(newg->im != (DgpImage_Binary)NULL);
				
				DGP_ERR_CHECK (DgpAdt_liAdd(cclist, (void *)newg, &newln)) {
					DGPIMAGE_FREE(newg->im);
					(DGP_FREE)((void *)newg);
					goto ERR_CLEANUP;
				}
			}
		}
	}
	
	DGPIMAGE_FREE(imcopy);
	
	*rlist = cclist;
	NORMAL_RETURN;
	
ERR_CLEANUP:
	if (imcopy != (DgpImage_Binary)NULL) {
		DGPIMAGE_FREE(imcopy);
	}
	
	if (cclist != DGP_NULLLIST) {
		DGPADT_FOREACH (ln, cclist) {
			DGP_ASSERT(DGPADT_LIGET(ln) != (void *)NULL);
			DGPIMAGE_FREE(((DgpImage_Component *)DGPADT_LIGET(ln))->im);
			(DGP_FREE)(DGPADT_LIGET(ln));
		}
		DgpAdt_liFree(cclist);
	}
	
	ERR_RETURN;
}

typedef struct {
	DgpImage_Point pos;
	int delta;
} DgpImage_ptdelta;

/* Initial size of workspace */
#define	NWORK	1000

/*
* Pull out a connected component which contains (x,y).
* Set the pixels in im which correspond to this component to 0.
* Return a Component which contains a DgpImage_Binary showing the component, and
* also where that image's box would land in the original image.
* (The DgpImage_Binary's internal offset will be (0,0) - I could have had it
* have the same coordinates as the original image, but that seemed as though
* it would cause problems in later processing).
* We assume that the image we're passed has a border of empty pixels at
* least one pixel wide all the way around (i.e. no object touches the edge).
* Funny, the image that's created for us to mangle has that property...
*/
static ErrCode
dgpimage_pullout(Dgp_State *dgp_state, DgpImage_Binary im, int x, int y, Dgp_Bool eightconnected,
		DgpImage_Component **rcomp)
{
	static void *vptchunk = (void *)NULL;
	DgpImage_ptdelta *ptchunk;
	static int nbig = 0;
	int nfree = 0;
	int nalloc = 0;
	DgpImage_Component *newg = (DgpImage_Component *)NULL;
	DgpImage_Binary newim = (DgpImage_Binary)NULL;
	int minx, miny, maxx, maxy;
	int lowx, lowy, highx, highy;
	DgpImage_ptdelta *newpt;
	DgpImage_ptdelta *pt;
	int ptnum;
	int i;
	int nx, ny;
	int delta;
	int minn, maxn;
	int nneighb = (eightconnected ? 8 : 4);
	DgpImage_Point *neighbours = (eightconnected ? eightneigh : fourneigh);
	
	DGP_PROLOGUE ("dgpimage_pullout", DGP_PERRSTATE);

	DGP_ASSERT(im != (DgpImage_Binary)NULL);
	
	/* Pull out the borders for easy reference */
	lowx = DGPIMG_XBASE(im);
	lowy = DGPIMG_YBASE(im);
	highx = DGPIMG_XMAX(im);
	highy = DGPIMG_YMAX(im);
	DGP_ASSERT((x >= lowx) && (x < highx) && (y >= lowy) && (y < highy));
	DGP_ASSERT(DGPIMG_REF(im, x, y) != 0);
	
	if (vptchunk == (void *)NULL) {
		/* Need to allocate some work space */
		DGP_CHECK_ALLOC (vptchunk = DGP_MALLOC (NWORK * sizeof(*ptchunk)));
		nfree = NWORK;
		nalloc = 0;
		nbig = NWORK;
		/* Register it */
/*		DGP_ERR_PASS (DSUTIL_reRegister(&vptchunk, DGP_FREE)); */
		/* FIX ME !!!!!!!!!!!!!!!! */
	}
	else {
		nfree = nbig;
		nalloc = 0;
	}
	ptchunk = (DgpImage_ptdelta *)vptchunk;
	
	DGP_CHECK_ALLOC_CLEANUP (newg = 
		(DgpImage_Component *)(DGP_MALLOC)(sizeof(*newg)));
	
	minx = maxx = x;
	miny = maxy = y;
	
	/*
	* Scan through the image, using ptchunk as a BFS queue.
	* Keep track of the bounds in (min,max)(x,y).
	* Once the component has been scanned, ptchunk should contain
	* all its points. Create a DgpImage_Binary of the correct size and read
	* them all out into it.
	*/
	
	/* Start off the queue */
	newpt = ptchunk;
	nfree--;
	nalloc++;
	
	newpt->pos.x = x;
	newpt->pos.y = y;
	newpt->delta = -100;		/* Something out-of-range */
	DGPIMG_REF(im, x, y) = 0;		/* Pull out the initial point */
	
	for (ptnum = 0; ptnum < nalloc; ptnum++) {
		pt = &(ptchunk[ptnum]);
		
		x = pt->pos.x;
		y = pt->pos.y;
		delta = pt->delta;
		
		DGP_ASSERT(DGPIMG_REF(im, x, y) == 0);	/* Already processed */
		
		minx = MIN(minx, x);
		miny = MIN(miny, y);
		maxx = MAX(maxx, x);
		maxy = MAX(maxy, y);
		
		/*
		* We can prune the search based on the side we came in from: we
		* know which of our neighbours fingered us, and so we don't need to
		* look at any of our neighbours which are also their neighbours.
		*
		* delta is set to the neighbour index that got us to this point.
		* Set minn and maxn to the smallest and largest indices of our
		* neighbours that need to be checked. Don't worry about wrap-around;
		* treat the indices as circular.
		*/
		if (eightconnected) {
			if (delta < 0) {
				/* First neighbour - can't prune anything */
				minn = 0;
				maxn = nneighb - 1;
			}
			else if (delta & 1) {
			/*
			* Came in from a diagonal neighbour
			*
			* Suppose that delta is the index of the (1, 1) entry. Then
			* the picture looks like:
			*	* + x
			*	+ X x
			*	x x x
			* where * is the pixel we came in from, + are its neighbours,
			* and X is us. We therefore have to scan from delta-2 to
			* delta+2 (remember, delta is the direction they took
			* to get to us, it refers to our down-right neighbour.
				*/
				minn = delta - 2;
				maxn = delta + 2;
			}
			else {
			/*
			* Came in from a side neighbour
			* The picture is:
			*	+ + x
			*	* X x
			*	+ + x
			*	so we need to scan from delta-1 to delta+1.
				*/
				minn = delta - 1;
				maxn = delta + 1;
			}
		}
		else {
			if (delta < 0) {
				/* First neighbour - can't prune anything */
				minn = 0;
				maxn = nneighb - 1;
			}
			else {
			/*
			* Came in from a side neighbour (there are no other kind)
			* Picture is:
			*	+ x @
			*	* X x
			*	+ x @
			* (where @ means we don't scan there). Check from
			* delta-1 to delta+2.
				*/
				minn = delta - 1;
				maxn = delta + 1;
			}
		}
		
		/* Scan our neighbours for other turned-on points */
		for (i = minn; i <= maxn; i++) {
			nx = x + neighbours[i & (nneighb - 1)].x;
			ny = y + neighbours[i & (nneighb - 1)].y;
			if (DGPIMG_REF(im, nx, ny) != 0) {
				/* Pull it out */
				DGPIMG_REF(im, nx, ny) = 0;
				/* Add it to the queue */
				if (nfree == 0) {
					/* Need more queue */
					DGP_CHECK_ALLOC (vptchunk = DGP_REALLOC (vptchunk,
						(size_t)(nbig * 2 * sizeof(*ptchunk))));
/*					vptchunk = myrealloc(dgp_state, vptchunk,
						(size_t)(nbig * sizeof(*ptchunk)),
						(size_t)(nbig * 2 *
						sizeof(*ptchunk))); 
					if (vptchunk == (void *)NULL) {
						goto ERR_CLEANUP;
					} */
					ptchunk = (DgpImage_ptdelta *)vptchunk;
					nfree = nbig;
					nbig *= 2;
				}
				newpt = &(ptchunk[nalloc]);
				nalloc++;
				nfree--;
				
				newpt->pos.x = nx;
				newpt->pos.y = ny;
				newpt->delta = i;
			}
		}
	}
	
    /* Now have a list of points and a bounding box */
    DGP_ERR_PASS_CLEANUP (DGPIMAGE_NEW(dgp_state, DGPIMAGE_BINARY,
		(maxx - minx + 1),
		(maxy - miny + 1), (void**) &newim));
    newg->im = newim;
    newg->offset.x = minx;
    newg->offset.y = miny;
    newg->npts = nalloc;
	
    /* Read the points off the queue and stick them into the image */
    for (ptnum = 0; ptnum < nalloc; ptnum++) {
		pt = &(ptchunk[ptnum]);
		
		DGPIMG_REF(newim, pt->pos.x - minx, pt->pos.y - miny) = 1;
	}
	
	*rcomp = newg;
    
	/* clean up the work space */
	if (vptchunk) 
		DGP_FREE (vptchunk);

	NORMAL_RETURN;
	
ERR_CLEANUP:
    if (newg != (DgpImage_Component *)NULL) {
		(DGP_FREE)((void *)newg);
	}
	
    if (newim != (DgpImage_Binary)NULL) {
		DGPIMAGE_FREE(newim);
	}
	
	/* clean up the work space */
	if (vptchunk)
		DGP_FREE (vptchunk);

    ERR_RETURN;
}
	
/*
* A realloc for ourselves, because the libwjr globals don't give us one.
*
* If we're using the standard malloc, use the standard realloc. Otherwise,
* allocate a new block, then copy and free the old block.

static void *
myrealloc(Dgp_State *dgp_state, void *addr, size_t oldsize, size_t newsize)
{
	void *newblock;
	
	if (DGP_MALLOC == malloc) {
		return(realloc(addr, newsize));
	}
	
	newblock = (DGP_MALLOC)(newsize);
	if (newblock == (void *)NULL) {
		return((void *)NULL);
	}
	
	memcpy(newblock, addr, oldsize);
	(DGP_FREE)(addr);
	return(newblock);
}
*/
