/*
*
* $Header: /project/icons/cvsroot/digipaper/src/libimage/image.c,v 1.22 2003/02/10 22:53:25 mkrocken Exp $
*
* Copyright (c) 1990, 1991, 1992, 1993 Cornell University.  All Rights
* Reserved.
*
* Copyright (c) 1991, 1992 Xerox Corporation.  All Rights Reserved.
*  
* Use, reproduction, preparation of derivative works, and distribution
* of this software is permitted.  Any copy of this software or of any
* derivative work must include both the above copyright notices of
* Cornell University and Xerox Corporation and this paragraph.  Any
* distribution of this software or derivative works must comply with all
* applicable United States export control laws.  This software is made
* available AS IS, and XEROX CORPORATION DISCLAIMS ALL WARRANTIES,
* EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE,
* AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
* LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
* EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
* NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
* OF THE POSSIBILITY OF SUCH DAMAGES.
*/

/*
*	image.c - Images.
*
* An image is just a rectangular array of some type (several types are
* implemented here).
*
* Exports:
*	type struct { Dgp_UInt8 r, g, b } RGB
*
*	type DgpImage_Type
*
*	type DgpImage_Gray
*	type DgpImage_Float
*	type DgpImage_ColorRGB
*	type DgpImage_Double
*	type DgpImage_Binary
*	type DgpImage_Long
*	type DgpImage_Ptr
*	type DgpImage_Short
*	type DgpImage_ColorRGBFloat
*
*	type GrayPixPtr
*	type FloatPixPtr
*	type RGBPixPtr
*	type DoublePixPtr
*	type DgpImage_BinaryPixPtr
*	type LongPixPtr
*	type PtrPixPtr
*	type ShortPixPtr
*	type RGBFloatPixPtr
*
*	constant DgpImage_Type DGPIMAGE_GRAY
*	constant DgpImage_Type DGPIMAGE_FLOAT
*	constant DgpImage_Type DGPIMAGE_RGB
*	constant DgpImage_Type DGPIMAGE_DOUBLE
*	constant DgpImage_Type DGPIMAGE_BINARY
*	constant DgpImage_Type DGPIMAGE_LONG
*	constant DgpImage_Type DGPIMAGE_PTR
*	constant DgpImage_Type DGPIMAGE_SHORT
*	constant DgpImage_Type DGPIMAGE_RGBFLOAT
*
*	constant DgpImage_Type DGPIMAGE_DONTCLEAR
*
*	constant Dgp_Int32 DGPIMG_COLRNG
*
*	void *imNew(DgpImage_Type type, Dgp_Int32 width, Dgp_Int32 height) - create
*		a new image of the appropriate type. The pointer returned
*		will be a DGPIMAGE_Gray, DGPIMAGE_Float, DGPIMAGE_ColorRGB, DGPIMAGE_Double, 
*		DGPIMAGE_Binary or WhateverelseImage as appropriate.
*
*	void *imNewOffset(DgpImage_Type type, Dgp_Int32 width, Dgp_Int32 height,
*			  Dgp_Int32 xbase, Dgp_Int32 ybase) - create a new image of
*		the appropriate type whose addressing begins
*		at (xbase, ybase).
*		imNew is equivalent to xbase = ybase = 0.
*
*	void imFree(void *im) - free an image.
*
*	void imSetOffset(void *im, Dgp_Int32 newxbase, Dgp_Int32 newybase) - modify
*		an image so that its addressing begins at
*		(newxbase, newybase).
*
*	void *imLoad(DgpImage_Type type, Dgp_Int8 *filename) - load an image from
*		a file. The pointer returned will be of the appropriate
*		type. Its base will be (0,0).
*
*	void *imLoadF(DgpImage_Type type, FILE *inFile) - load an image
*		from a stream. The pointer returned will be of the
*		appropriate type.
*
*	Dgp_Int32 imSave(void *im, Dgp_Int8 *filename) - save an image into a file.
*
*	Dgp_Int32 imSaveF(void *im, FILE *outFile) - save an image to a stream.
*
*	void *imDup(void *im) - duplicate an image
*
*	void *imPromote(void *im, DgpImage_Type promoteto) - promote an image to
*		some higher type.
*
*	void imSetPromotion(DgpImage_Type promotefrom, DgpImage_Type promoteto,
*			    double slope, double offset) - change the method
*		of image promotion.
*
*	void *imDemote(void *im, DgpImage_Type demoteto) - demote an image to some
*		lower type.
*
*	void *imConvert(void *im, DgpImage_Type convertto) - convert (promote
*		or demote as appropriate) an image to some other type.
*
*	Dgp_Int32 DGPIMG_WIDTH(void *im) - get the width of an image.
*
*	Dgp_Int32 DGPIMG_HEIGHT(void *im) - get the height of an image.
*
*	Dgp_Int32 DGPIMG_XBASE(void *im) - get the x-base of an image.
*
*	Dgp_Int32 DGPIMG_YBASE(void *im) - get the y-base of an image.
*
*	Dgp_Int32 DGPIMG_XMAX(void *im) - get the maximum valid X value of an image.
*
*	Dgp_Int32 DGPIMG_YMAX(void *im) - get the maximum valid Y value of an image.
*
*	DgpImage_Type DGPIMG_TYPE(void *im) - get the type of an image.
*
*	Dgp_Int8 *DGPIMG_COMMENT(void *im) - get the comment attached to an image.
*
*	Dgp_Bool imSetComment(void *im, Dgp_Int8 *comment) - set the comment
*		attached to an image.
*
*	macro DGPIMG_REF(void *im, Dgp_Int32 x, Dgp_Int32 y) - access an element
*		of some form of Image.
*
*	macro DGPIMG_PIXPTR(void *im, Dgp_Int32 x, Dgp_Int32 y) - get a handle on a pixel of
*		some form of Image.
*
*	macro DGPIMG_PTRREF(void *im, *PixPtr) - dereference a PixPtr of some type.
*
*	macro DGPIMG_PTRDUP(void *im, *PixPtr) - duplicate a PixPtr of some type.
*
*	void DGPIMG_PTRUP(void *im, *PixPtr) - move a PixPtr up in the image.
*
*	void DGPIMG_PTRDOWN(void *im, *PixPtr) - move a PixPtr down in the image.
*
*	void DGPIMG_PTRLEFT(void *im, *PixPtr) - move a PixPtr left in the image.
*
*	void DGPIMG_PTRRIGHT(void *im, *PixPtr) - move a PixPtr right in the image.
*
*	void DGPIMG_PTRUPN(void *im, *PixPtr, Dgp_Int32 n)
*	void DGPIMG_PTRDOWNN(void *im, *PixPtr, Dgp_Int32 n)
*	void DGPIMG_PTRRIGHTN(void *im, *PixPtr, Dgp_Int32 n)
*	void DGPIMG_PTRLEFTN(void *im, *PixPtr, Dgp_Int32 n)
*
*	Dgp_Bool DGPIMG_PTREQ(void *im, *PixPtr1, *PixPtr2) - check two PixPtrs
*		for equality.
*
* An image is a rectangular storage area, with each cell being of some
* type. In this case, the types are Dgp_UInt8, float, RGB, double,
* binary, long, void * and short. Float is large enough for most purposes for
* image processing; for those for which it is not, DGPIMAGE_Double is provided;
* the space costs are much higher for this, so it should be avoided unless
* necessary. Binary images can store 0 or 1 in each location. By convention,
* 0 is white and 1 is black in binary images.
* DGPIMAGE_Ptrs can store a void * pointer at each location. These images may
* not be saved or loaded.
* DGPIMG_COLRNG is the number of discrete gray levels in a DGPIMAGE_Gray, and the number
* of levels in each component of an DGPIMAGE_ColorRGB. DGPIMG_COLRNG-1 is the brightest
* (most white) DGPIMAGE_Gray value.
*
* N.B. Older versions of this datatype had a different definition for RGB,
* namely:
*	typedef Dgp_UInt8 RGB[3];
* Due to C's weirdness with arrays, this proved troublesome, so the new
* definition was installed. However, since some code relies on the old
* definition, a backwards compatibility mode is available. To get it,
* #define DGPIMAGE_OLD_RGB
* before including image.h.
*
* imNew creates an Image of the given size, of the appropriate type,
*	returning (void *)NULL on failure. The Image returned
*	contains all zeros. The return value should be cast to
*	the appropriate type. If you OR the supplied DgpImage_Type with
*	DGPIMAGE_DONTCLEAR, then the image is *not* cleared to zero; this can
*	be useful if you're going to write to every location and want to
*	save the time that might be spent clearing it.
*
* imNewOffset creates an image of the given size, of the appropriate type,
*	returning (void *)NULL on failure. The valid range of addresses
*	for the new image is (xbase..xbase+width-1, ybase..ybase+height-1).
*	Addressing an image with a shifted base may be more efficient than
*	performing the base shift for each access.
*
* imFree frees the given Image.
*
* imSetOffset modifies the given image so that its range of addresses
*	is (newxbase..newxbase+width-1, newybase..newybase+height-1).
*	The image data is not changed. The pixel which was previously
*	addressed as DGPIMG_REF(im, DGPIMG_XBASE(im), DGPIMG_YBASE(im)) will now
*	be addressed as DGPIMG_REF(im, newxbase, newybase). DGPIMG_XBASE(im) and
*	DGPIMG_YBASE(im) will be updated to newxbase and newybase.
*
* imLoad attempts to load an Image from a file. It checks to see that
*	what is stored in the file has the appropriate type, and if not
*	returns NULL. It then attempts to load the image from the file,
*	and if successful returns the (.*)Image as appropriate. If it
*	fails, it returns NULL. The base of the loaded image is (0,0),
*	regardless of what it was when it was saved. The file is closed in
*	all cases.
*	If the file does not contain an image of the appropriate type, but
*	does contain an image of a type which may be promoted to that type
*	(for example, if it contains a DGPIMAGE_Binary and imLoad is attempting
*	to load a DGPIMAGE_Gray), the image will be read and silently promoted
*	to the requested type (see imPromote). If such a promotion is not
*	possible, NULL is returned.
*	Any comments present in the original file are read in and stored
*	as the comment associated with the image (see DGPIMG_COMMENT).
*	If more than one comment line was present in the file, then the
*	contents of the lines will be separated by newlines. There is no
*	trailing newline; thus, the image read from a file with a single
*	comment line will contain the contents of that line with no extra
*	newline. There is a maximum limit on the amount of comments which
*	will be read from a file of 8192 bytes (including newlines).
*
* imLoadF attempts to load an Image from a stdio stream. It performs the
*	same operations as imLoad, except that it assumes the stream is
*	open and correctly positioned. After loading, the stream is positioned
*	to just after the image; it is not closed. If the image cannot be
*	loaded (i.e. if NULL is returned), the position of the stream (and
*	what has been written to it) is undefined.
*
* imSave attempts to write the given Image out to a file. If successful,
*	it returns 0. If it fails, it returns -1. Files written by imSave
*	from DGPIMAGE_FLOAT, DGPIMAGE_DOUBLE, DGPIMAGE_LONG, DGPIMAGE_SHORT, and
*	DGPIMAGE_RGBFLOAT images are stored with each of the values
*	in big-endian format.  This means that a file written on one
*	sort of machine should be able to be read on another sort, as
*	long as both use compatible floating-point formats.  Luckily, this
*	is often the case.
*	It also writes out the comment associated with the image
*	as a comment in the file. The rules for this are:
*	- Every newline in the comment string causes a new comment line to
*	  be begun in the file.
*	- If any line would be longer than 70 characters, it is broken up.
*	  If possible, it will be broken at a whitespace character.
*	If imSave fails (returns -1), the contents of the file are
*	undefined.
*
* imSaveF performs the same function as imSave, except that it writes
*	to a previously-opened stream. After the save, the stream
*	is positioned after the newly-written data.
*
* imDup makes an exact duplicate of an Image - another Image having
*	the same size, type and offsets, the same image data, and the same
*	comment. It returns (void *)NULL if it cannot create the duplicate.
*
* imPromote converts an image to a "higher" type of image. The arguments are
*	the image to be converted and the type into which it should be
*	converted. It returns (void *)NULL if it cannot create the new image.
*	Only certain types of conversions are possible with this routine.
*	They are:
*	DGPIMAGE_Binarys can be promoted to:
*		DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_Float, DGPIMAGE_Double,
*		DGPIMAGE_ColorRGB, DGPIMAGE_ColorRGBFloat
*	DGPIMAGE_Grays can be promoted to:
*		DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_Float, DGPIMAGE_Double, DGPIMAGE_ColorRGB,
*		DGPIMAGE_ColorRGBFloat
*	DGPIMAGE_Shorts can be promoted to:
*		DGPIMAGE_Long, DGPIMAGE_Float, DGPIMAGE_Double, DGPIMAGE_ColorRGBFloat
*	DGPIMAGE_Longs can be promoted to:
*		DGPIMAGE_Double
*	DGPIMAGE_Floats can be promoted to:
*		DGPIMAGE_Double, DGPIMAGE_ColorRGBFloat
*	The rules of the conversions are:
*	DGPIMAGE_Binary->DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Long:	(1-DGPIMG_COLRNG,DGPIMG_COLRNG-1)
*		0 maps to DGPIMG_COLRNG - 1
*		1 maps to 0
*	DGPIMAGE_Binary->DGPIMAGE_Float, DGPIMAGE_Double:		(1,0)
*		0 maps to 0
*		1 maps to 1
*	DGPIMAGE_Binary->DGPIMAGE_ColorRGB:				(1-DGPIMG_COLRNG,DGPIMG_COLRNG-1)
*		0 maps to {DGPIMG_COLRNG-1, DGPIMG_COLRNG-1, DGPIMG_COLRNG-1}
*		1 maps to {0, 0, 0}
*	DGPIMAGE_Binary->DGPIMAGE_ColorRGBFloat:			(1-DGPIMG_COLRNG,DGPIMG_COLRNG-1)
*		0 maps to {DGPIMG_COLRNG-1, DGPIMG_COLRNG-1, DGPIMG_COLRNG-1}
*		1 maps to {0, 0, 0}
*	DGPIMAGE_Gray->DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_Float, DGPIMAGE_Double: (1,0)
*		a pixel with value v in the original image has value v
*		in the converted image. (i.e. v maps to v)
*	DGPIMAGE_Gray->DGPIMAGE_ColorRGB:				(1,0)
*		v maps to {v, v, v}
*	DGPIMAGE_Gray->DGPIMAGE_ColorRGBFloat:			(1,0)
*		v maps to {v, v, v}
*	DGPIMAGE_Short->DGPIMAGE_Long, DGPIMAGE_Float, DGPIMAGE_Double:	(1,0)
*		v maps to v
*	DGPIMAGE_Short->DGPIMAGE_ColorRGBFloat:			(1,0)
*		v maps to {v, v, v}
*	DGPIMAGE_Long->DGPIMAGE_Double:				(1,0)
*		v maps to v
*	DGPIMAGE_Float->DGPIMAGE_Double:			(1,0)
*		v maps to v
*	DGPIMAGE_Float->DGPIMAGE_ColorRGBFloat:			(1,0)
*		v maps to {v, v, v}
*	The meaning of the numbers in parentheses are: if a conversion
*	is labelled (a,b), then a pixel with input value v will be mapped
*	to av+b. These values are the default ones; they may be changed at
*	any time with imSetPromotion. Note that these default values are not
*	transitive: promoting (for example) a DGPIMAGE_Binary to a DGPIMAGE_Gray then
*	to a DGPIMAGE_Float will give different results than promoting the
*	DGPIMAGE_Binary directly to a DGPIMAGE_Float.
*	When converting to an DGPIMAGE_ColorRGB, all three components are set to the
*	same value.
*	These conversions are all those which are strictly "promotions": no
*	information is lost during the conversion.
*
* imSetPromotion alters the parameters of one of the promotions. The
*	"promotefrom" and "promoteto" DgpImage_Types must correspond to a
*	valid promotion. The slope and offset parameters will then be used
*	for all future promotions from images of type "promotefrom" to images
*	of type "promoteto". This includes the implicit promotions performed
*	by imLoad and imLoadF.
*
* imDemote converts an Image to an Image of lower type. The conversions
*	which this routine may perform are:
*	DGPIMAGE_Doubles can be demoted to:
*		DGPIMAGE_Binary, DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_Float,
*		DGPIMAGE_ColorRGB, DGPIMAGE_ColorRGBFloat
*	DGPIMAGE_Floats can be demoted to:
*		DGPIMAGE_Binary, DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_ColorRGB
*	DGPIMAGE_Longs can be demoted to:
*		DGPIMAGE_Binary, DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Float, DGPIMAGE_ColorRGB,
*		DGPIMAGE_ColorRGBFloat
*	DGPIMAGE_Shorts can be demoted to:
*		DGPIMAGE_Binary, DGPIMAGE_Gray, DGPIMAGE_ColorRGB
*	DGPIMAGE_Grays can be demoted to:
*		DGPIMAGE_Binary
*	DGPIMAGE_ColorRGBs can be demoted to:
*		DGPIMAGE_Binary, DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_Float,
*		DGPIMAGE_Double
*	DGPIMAGE_ColorRGBFloats can be demoted to:
*		DGPIMAGE_Binary, DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_Float,
*		DGPIMAGE_Double, DGPIMAGE_ColorRGB
*	All these conversions have the property that information may be lost.
*	Note that some conversions may lose information in both directions
*	(for example DGPIMAGE_Double<->DGPIMAGE_ColorRGB).
*	The rules of the conversions are:
*	DGPIMAGE_Double->DGPIMAGE_Float:
*		v maps to v, if v lies in the range of floats
*		undefined if v does not
*	DGPIMAGE_Double->DGPIMAGE_Long, DGPIMAGE_Short, DGPIMAGE_Gray:
*		v maps to floor(v) if floor(v) is representable in the
*		destination
*		undefined if floor(v) is not
*	DGPIMAGE_Double->DGPIMAGE_Binary:
*		0 maps to 0
*		non-zero maps to 1
*	DGPIMAGE_Double->DGPIMAGE_ColorRGB
*		v maps to {floor(v), floor(v), floor(v)} if floor(v) is
*		representable in the destination
*		undefined if floor(v) is not
*	DGPIMAGE_Double->DGPIMAGE_ColorRGBFloat
*		v maps to {v, v, v} if v lies in the range of floats
*		undefined if v does not
*	DGPIMAGE_Float->DGPIMAGE_Long, DGPIMAGE_Short, DGPIMAGE_Gray:
*		v maps to floor(v) if floor(v) is representable in the
*		destination
*		undefined if floor(v) is not
*	DGPIMAGE_Float->DGPIMAGE_Binary:
*		0 maps to 0
*		non-zero maps to 1
*	DGPIMAGE_Float->DGPIMAGE_ColorRGB
*		v maps to {floor(v), floor(v), floor(v)} if floor(v) is
*		representable in the destination
*		undefined if floor(v) is not
*	DGPIMAGE_Long->DGPIMAGE_Short, DGPIMAGE_Gray
*		v maps to v if v is representable in the destination
*		undefined if v is not
*	DGPIMAGE_Long->DGPIMAGE_Float
*		v maps to v (note that accuracy may be lost)
*	DGPIMAGE_Long->DGPIMAGE_Binary
*		0 maps to 1
*		non-zero maps to 0
*	DGPIMAGE_Long->DGPIMAGE_ColorRGB
*		v maps to {v, v, v} if v is representable in the destination
*		undefined if v is not
*	DGPIMAGE_Long->DGPIMAGE_ColorRGBFloat
*		v maps to {v, v, v} if v is representable in the destination
*		undefined if v is not
*	DGPIMAGE_Short->DGPIMAGE_Gray
*		v maps to v if v is representable in the destination
*		undefined if v is not
*	DGPIMAGE_Short->DGPIMAGE_Binary
*		0 maps to 1
*		non-zero maps to 0
*	DGPIMAGE_Short->DGPIMAGE_ColorRGB
*		v maps to {v, v, v} if v is representable in the destination
*		undefined if v is not
*	DGPIMAGE_Gray->DGPIMAGE_Binary
*		0 maps to 1
*		non-zero maps to 0
*	DGPIMAGE_ColorRGB->DGPIMAGE_Binary
*		If all components are zero, then 1
*		0 otherwise
*	DGPIMAGE_ColorRGB->anything else
*		The R, G and B components are first combined by
*		v = .299R + .587G + .114B
*		The resulting value is then assigned to the new image pixel.
*	DGPIMAGE_ColorRGBFloat->DGPIMAGE_Binary
*		If all components are zero, then 1
*		0 otherwise
*	DGPIMAGE_ColorRGBFloat->DGPIMAGE_ColorRGB
*		{r, g, b} maps to {floor(r), floor(g), floor(b)}
*		if all those values are representable.
*		undefined otherwise
*	DGPIMAGE_ColorRGBFloat->anything else
*		The R, G and B components are first combined by
*		v = floor(.299R + .587G + .114B)
*		The resulting value is then assigned to the new image pixel.
*		If the result is not representable in the destination,
*		the result is undefined.
*	These conversions are as closely as possibly the inverses of the
*	conversion routines used by imPromote (i.e. a promotion followed by a
*	demotion will give the original image back). Note that this makes them
*	non-transitive, and that there are some conversions which are
*	demotions in both directions. None of these conversions may be
*	altered; if another type of conversion is desired, a special routine
*	may be written.
*
* imConvert converts (promotes or demotes) an image to an image of another
*	type. Any two types of image may be converted between, except for
*	DGPIMAGE_Ptrs. This routine is a convenience routine which calls
*	imPromote or imDemote as necessary. As a special case,
*	converting an image to its original type has the same effect as
*	imDup EXCEPT that the comment is not copied (no other conversion copies
*	the comment, so this behaviour is consistent). Note that this form of
*	conversion is neither a promotion nor a demotion, and so will be
*	rejected by both imPromote and imDemote.
*
* DGPIMG_WIDTH returns the width of the given Image.
*
* DGPIMG_HEIGHT returns the height of the given Image.
*
* DGPIMG_XBASE returns the X base of the given Image. This is the lowest
*	valid X value.
*
* DGPIMG_YBASE returns the Y base of the given Image. This is the lowest
*	valid Y value.
*
* DGPIMG_XMAX returns the highest valid X value of an Image. This is equal to
*	DGPIMG_XBASE(im)+DGPIMG_WIDTH(im)-1.
*
* DGPIMG_YMAX returns the highest valid Y value of an Image. This is equal to
*	DGPIMG_YBASE(im)+DGPIMG_HEIGHT(im)-1.
*
* DGPIMG_TYPE returns the DgpImage_Type of the given Image.
*
* DGPIMG_COMMENT returns the comment associated with the Image, or NULL if
*	there is none. This is the comment which was present in the file
*	from which the image was loaded, if any, or a comment associated
*	with the image later using imSetComment. Note that the memory
*	pointed at by this pointer is owned by the Image module, and should
*	not be freed; it will remain valid until the Image is freed, or the
*	comment is changed.
*
* imSetComment assigns the given string as the comment of the given Image.
*	It returns DGP_TRUE if it is successful, DGP_FALSE if it is not. It makes
*	a copy of the string; the original string is not referred to after
*	this, and may be freed or altered without affecting the Image's
*	comment. You can pass in NULL as the new comment. If the comment
*	could not be changed (i.e. DGP_FALSE is returned), then the old comment
*	remains.
*
* DGPIMG_REF gives access to an entry in an Image of some form (it works
*	for all defined forms of Image). It gives an lvalue: it is valid
*	to say DGPIMG_REF(i, x, y) = foo;
*
* The PixPtr types allow efficient access to contiguous regions of an Image.
* You get a PixPtr by calling DGPIMG_PIXPTR, giving the location in the image
* you would like a handle on. You can then dereference this using
* DGPIMG_PTRREF. DGPIMG_REF(im, x, y) gives the same value as
* DGPIMG_PTRREF(DGPIMG_PIXPTR(im, x, y)); note that this is an lvalue and may be
* assigned to.
* The advantage of PixPtrs is that they can be efficiently moved through
* the image array. For example, if im is a DGPIMAGE_Gray:
* for (x = 0; x < w; x++) {
*     DGPIMG_REF(im, x, y) = 123;
*     }
* can be replaced by:
* GrayPixPtr pp;
* pp = DGPIMG_PIXPTR(im, 0, y);
* for (x = 0; x < w; x++) {
*     DGPIMG_PTRREF(im, pp) = 123;
*     DGPIMG_PTRRIGHT(im, pp);
*     }
* which may run more efficiently.
*
* DGPIMG_PIXPTR gets a handle on the (x,y) pixel in the image im, giving a
*	PixPtr of the appropriate type.
*
* DGPIMG_PTRREF dereferences a PixPtr of some type, giving an lvalue of the
*	appropriate type.
*
* DGPIMG_PTRDUP duplicates a PixPtr of some type, returning a PixPtr of the same
*	type, initially referring to the same location in the image.
*
* DGPIMG_PTRUP, DGPIMG_PTRDOWN, DGPIMG_PTRLEFT and DGPIMG_PTRRIGHT move a PixPtr through an
*	Image. DGPIMG_PTRUP effectively decrements the y component of the pixel
*	to which the pointer refers; DGPIMG_PTRDOWN increments it. DGPIMG_PTRLEFT and
*	DGPIMG_PTRRIGHT decrement and increment respectively the x component.
*	imPtr*N do the same thing but change by the specified amount.
*
* DGPIMG_PTREQ compares two pointers for equality, returning DGP_TRUE if they are,
*	DGP_FALSE if they are not. The pointers must refer to pixels in the
*	same Image. The use of this allows more efficient image operations.
*
* Some of these are implemented partially as macros for strange reasons.
* As a result, they may not be entirely safety-checked. If you want to
* have safety-checking in place, at a cost in speed and code size, put
* -DDGPIMAGE_CHECKALL on your gcc line, or add
* #define DGPIMAGE_CHECKALL
* to your code somewhere before you include image.h.
* This only works if you're compiling with gcc.
*
*/

#include "util.h"
#include "image.h"
#include "image_defs.h"
#include <ctype.h>
#include <stdio.h>
/*#include "undecl-stdio.h"*/
#include <stdlib.h>
#include <string.h>
/*#include "undecl-string.h"*/

#include "pbmstuff.h"
#include "image-private.h"
#include "rawimage.h"
#include "endianness.h"
#include "PsgFix.h"

/* VxWorks has no support for strdup();  hack.  Use dgp allocator? */
/* ...run-time implementation does not. */
/* 14.3.2000 / dhtesq. */
#ifdef VXWORKS
#define strdup(s) strcpy((Dgp_Int8 *)malloc(strlen((s))), (s))
#endif /* (VXWORKS) */

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libimage/image.c,v 1.22 2003/02/10 22:53:25 mkrocken Exp $");

/* Declare internal routines */
static Dgp_UInt32 **allocPackedData(Dgp_Int32 width, Dgp_Int32 height,
									Dgp_Int32 xbase, Dgp_Int32 ybase,
									Dgp_Bool dontclear, Dgp_State *dgp_state);
static Dgp_UInt32 **allocPackedDataPointers(Dgp_Int32 width, Dgp_Int32 height,
											Dgp_UInt32 *data,
											Dgp_Int32 longsperline,
											Dgp_Int32 xbase, Dgp_Int32 ybase, Dgp_State *dgp_state);
static DgpImage_Binary packedToBinary(DgpImage_Packed im);
static DgpImage_Packed binaryToPacked(DgpImage_Binary im);
ErrCode
DgpImage_pDemote (void *im, DgpImage_Type demoteto, void** image);
ErrCode
DgpImage_pPromote (void *im, DgpImage_Type promoteto, void** image);

static void getAreaAndSize(void *im, DgpImage_Header *header,
	void ***areaPtr, size_t *el_size, Dgp_State *dgp_state);
static void
freeImage(Dgp_State *dgp_state, void **ptr, Dgp_Int32 xbase, Dgp_Int32 ybase, size_t el_size);

#ifndef DGP_FAXMODE
static void *readBinaryCooked(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader);
static void *readGrayCooked(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader);
static void *readRGBCooked(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader);
static void *readBinary(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader);
static void *readGray(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader);
static void *readRGB(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader);
static void *readFloat(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader);
static void *readDouble(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader);
static void *readLong(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader);
static void *readShort(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader);
static void *readRGBFloat(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader);

static DgpImage_Packed readRaw(FILE *inFile, Dgp_State *dgp_state);
static DgpImage_Packed readCooked(FILE *inFile, Dgp_State *dgp_state);

/*
* Here begins file magic: list of tags with associated file types and
* reading functions.
*/
static struct {
    Dgp_Int8 *filetag;
    DgpImage_Type filetype;
    void * (*readfunc)(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader);
} fileReaders[] = {
	{ TAG_BINARY_COOKED, DGPIMAGE_BINARY, readBinaryCooked },
	{ TAG_GRAY_COOKED, DGPIMAGE_GRAY, readGrayCooked },
	{ TAG_RGB_COOKED, DGPIMAGE_RGB, readRGBCooked },
	{ TAG_BINARY, DGPIMAGE_BINARY, readBinary },
	{ TAG_GRAY, DGPIMAGE_GRAY, readGray },
	{ TAG_RGB, DGPIMAGE_RGB, readRGB },
	{ TAG_FLOAT, DGPIMAGE_FLOAT, readFloat },
	{ TAG_DOUBLE, DGPIMAGE_DOUBLE, readDouble },
	{ TAG_LONG, DGPIMAGE_LONG, readLong },
	{ TAG_SHORT, DGPIMAGE_SHORT, readShort },
	{ TAG_RGBFLOAT, DGPIMAGE_RGBFLOAT, readRGBFloat }
};
#define	NREADERS	((Dgp_Int32)(sizeof(fileReaders) / sizeof(fileReaders[0])))
#endif

/* Here are all the conversion routines */
static void *pG2F(void *im) , *pG2R(void *im) , *pG2D(void *im) ;
static void *dG2B(void *im) , *pG2L(void *im) , *pG2S(void *im) , *pG2RF(void *im) ;
static void *dF2G(void *im) , *dF2R(void *im) , *pF2D(void *im) ;
static void *dF2B(void *im) , *dF2L(void *im) , *dF2S(void *im) , *pF2RF(void *im) ;
static void *dR2G(void *im) , *dR2F(void *im) , *dR2D(void *im) ;
static void *dR2B(void *im) , *dR2L(void *im) , *dR2S(void *im) , *pR2RF(void *im) ;
static void *dD2G(void *im) , *dD2F(void *im) , *dD2R(void *im)                  ;
static void *dD2B(void *im) , *dD2L(void *im) , *dD2S(void *im) , *dD2RF(void *im) ;
static void *pB2G(void *im) , *pB2F(void *im) , *pB2R(void *im) , *pB2D(void *im) ;
static void *pB2L(void *im) , *pB2S(void *im) , *pB2RF(void *im) ;
static void *dL2G(void *im) , *dL2F(void *im) , *dL2R(void *im) , *pL2D(void *im) ;
static void *dL2B(void *im) , *dL2S(void *im) , *dL2RF(void *im) ;
static void *dS2G(void *im) , *pS2F(void *im) , *dS2R(void *im) , *pS2D(void *im) ;
static void *dS2B(void *im) , *pS2L(void *im) ,  *pS2RF(void *im) ;
static void *dRF2G(void *im) , *dRF2F(void *im) , *dRF2R(void *im) , *dRF2D(void *im) ;
static void *dRF2B(void *im) , *dRF2L(void *im) , *dRF2S(void *im)                ;

/*
* Here's the table of conversions.
* convertTable[type1][type2] tells how we convert from an image of type1
* to an image of type2.
*/
static struct {
    Dgp_Bool isPromotion;
    double slope;
    double offset;
    void * (*convertFunc)(void *im) ;
} convertTable[DGPIMAGE_RGBFLOAT + 1][DGPIMAGE_RGBFLOAT + 1] = {
	{	/* Invalid image type */
		{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL},
		{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL},
		{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL},
		{ DGP_FALSE, 0, 0, NULL}},
		
		{	/* DGPIMAGE_Gray */
			{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL}, { DGP_TRUE , 1, 0, pG2F},
			{ DGP_TRUE , 1, 0, pG2R}, { DGP_TRUE , 1, 0, pG2D}, { DGP_FALSE, 0, 0, dG2B},
			{ DGP_TRUE , 1, 0, pG2L}, { DGP_FALSE, 0, 0, NULL}, { DGP_TRUE , 1, 0, pG2S},
			{ DGP_TRUE , 1, 0, pG2RF}},
			
			{	/* DGPIMAGE_Float */
				{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dF2G}, { DGP_FALSE, 0, 0, NULL},
				{ DGP_FALSE, 0, 0, dF2R}, { DGP_TRUE , 1, 0, pF2D}, { DGP_FALSE, 0, 0, dF2B},
				{ DGP_FALSE, 0, 0, dF2L}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dF2S},
				{ DGP_TRUE , 1, 0, pF2RF}},
				
				{	/* DGPIMAGE_ColorRGB */
					{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dR2G}, { DGP_FALSE, 0, 0, dR2F},
					{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dR2D}, { DGP_FALSE, 0, 0, dR2B},
					{ DGP_FALSE, 0, 0, dR2L}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dR2S},
					{ DGP_TRUE,  1, 0, pR2RF}},
					
					{	/* DGPIMAGE_Double */
						{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dD2G}, { DGP_FALSE, 0, 0, dD2F},
						{ DGP_FALSE, 0, 0, dD2R}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dD2B},
						{ DGP_FALSE, 0, 0, dD2L}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dD2S},
						{ DGP_FALSE, 0, 0, dD2RF}},
						
						{	/* DGPIMAGE_Binary */
							{ DGP_FALSE, 0, 0, NULL}, { DGP_TRUE , 1-DGPIMG_COLRNG, DGPIMG_COLRNG-1, pB2G}, { DGP_TRUE , 1, 0, pB2F},
							{ DGP_TRUE , 1-DGPIMG_COLRNG, DGPIMG_COLRNG-1, pB2R}, { DGP_TRUE , 1, 0, pB2D}, { DGP_FALSE, 0, 0, NULL},
							{ DGP_TRUE , 1-DGPIMG_COLRNG, DGPIMG_COLRNG-1, pB2L}, { DGP_FALSE, 0, 0, NULL}, { DGP_TRUE , 1-DGPIMG_COLRNG, DGPIMG_COLRNG-1, pB2S},
							{ DGP_TRUE , 1-DGPIMG_COLRNG, DGPIMG_COLRNG-1, pB2RF}},
							
							{	/* DGPIMAGE_Long */
								{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dL2G}, { DGP_FALSE, 0, 0, dL2F},
								{ DGP_FALSE, 0, 0, dL2R}, { DGP_TRUE , 1, 0, pL2D}, { DGP_FALSE, 0, 0, dL2B},
								{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dL2S},
								{ DGP_FALSE, 0, 0, dL2RF}},
								
								{	/* DGPIMAGE_Ptr */
									{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL},
									{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL},
									{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL},
									{ DGP_FALSE, 0, 0, NULL}},
									
									{	/* DGPIMAGE_Short */
										{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dS2G}, { DGP_TRUE , 1, 0, pS2F},
										{ DGP_FALSE, 0, 0, dS2R}, { DGP_TRUE , 1, 0, pS2D}, { DGP_FALSE, 0, 0, dS2B},
										{ DGP_TRUE , 1, 0, pS2L}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, NULL},
										{ DGP_TRUE , 1, 0, pS2RF}},
										
										{	/* DGPIMAGE_ColorRGBFloat */
											{ DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dRF2G}, { DGP_FALSE, 0, 0, dRF2F},
											{ DGP_FALSE, 0, 0, dRF2R}, { DGP_FALSE, 0, 0, dRF2D}, { DGP_FALSE, 0, 0, dRF2B},
											{ DGP_FALSE, 0, 0, dRF2L}, { DGP_FALSE, 0, 0, NULL}, { DGP_FALSE, 0, 0, dRF2S},
											{ DGP_FALSE, 0, 0, NULL}}
};

/*
* This routine allocates a rectangular area of width*height*el_size bytes
* (suitable for holding width*height elements of el_size size), and
* initialises it to zero (unless dontclear is DGP_TRUE). It then allocates and
* sets up an array of height pointers into this, each one pointing to a
* row (row_width*el_size bytes).
*
* Each pointer is offset by xbase elements, and the returned pointer is
* offset by ybase rows; this means that the first element in this array
* is accessed by array[ybase][xbase]. Note that xbase and ybase may be
* negative.
*
* Note that this isn't really type-safe: it takes this (void **) pointer
* which will get cast, for example, into a (float **) pointer, which isn't
* really legit. On machines where all pointers have the same format,
* it will work. The problem is also that dereferencing a (float **)
* should get you a (float *), but will actually poDgp_Int32 to the bit-pattern
* of a (void *), which could be cast into that (float *) with no problems,
* but which might not work if it is simply *treated* as a (float *)
* without the cast. Bah. Doing it right would be too horrendous for words.
* C just can't handle generics very well...
* Only work for non-packed image
*/
static void **
allocImage(Dgp_State *dgp_state, Dgp_Int32 row_width, Dgp_Int32 height, Dgp_Int32 xbase, Dgp_Int32 ybase, size_t el_size, Dgp_Bool dontclear)
{
    register Dgp_Int32 i;
    register void *mainArea;
    register void **ptrArea;
    register Dgp_Int8 *p;
    register void **q;
	
    /* Allocate the actual image data area */
    if (!dontclear) {
		mainArea = (DGP_CALLOC)((size_t)height, row_width);
    }
    else {
		mainArea = (DGP_MALLOC)(height * row_width);
	}
    if (mainArea == (void *)NULL) {
		return((void **)NULL);
    }
	

    /* Allocate a pointer for each row */
    ptrArea = (void **)(DGP_MALLOC)(height * sizeof(void *));
    if (ptrArea == (void **)NULL) {
		(DGP_FREE)(mainArea);
		return((void **)NULL);
	}
	
    /* Set up the row pointers to point to the (offset) rows */
    for (i = 0, q = ptrArea, p = (Dgp_Int8 *)mainArea; i < height; i++, q++, p += row_width) {
		*q = (void *)(p - xbase * (Dgp_Int32)el_size);
	}
	
    /* and return the (offset) row pointer area */
    return(ptrArea - ybase);
}

ErrCode
DgpImage_free (void *im)
{
	DgpImage_Packed pim = (DgpImage_Packed) im;
	Dgp_State *dgp_state = pim->dgp_state;
	DgpImage_Header *header;
	DGP_PROLOGUE ("DgpImage_free", DGP_PERRSTATE);
        DGP_STOP_WHINE;

	if(pim == (void *)NULL) NORMAL_RETURN;
	header = &pim->header;
	
	if (header->comment != (Dgp_Int8 *)NULL) {
		(DGP_FREE)((void *)header->comment);
	}
	
    if (header->tag != DGPIMAGE_PACKED) {
		void **areaPtr;
		size_t el_size;
		Dgp_Int32 xbase, ybase;
		/*DgpImage_Gray gimage = (DgpImage_Gray) im; */
		
		if(pim->data) {
		    /* Pull out the data area and the size of the element */
		    /* and get rid of all of them */
	    	    xbase = header->xbase;
		    ybase = header->ybase;
                    getAreaAndSize(im, header, &areaPtr, &el_size, dgp_state);
		    freeImage (dgp_state, areaPtr, xbase, ybase, el_size);
                }
		
		(DGP_FREE)((void *)im);
    } else {
		
		if(pim->data != (Dgp_UInt32 **)NULL){
		
		    /* Free the data area, the pointer block, the header, and the image */
		    (DGP_FREE)((void *)pim->data[pim->header.ybase]);
		    (DGP_FREE)((void *)(pim->data + pim->header.ybase));
                }
		(DGP_FREE)((void *)pim);
    }
	NORMAL_RETURN;

}

/*
* This routine gets rid of a memory block allocated by allocImage above.
*/
static void
freeImage(Dgp_State *dgp_state, 
                    void **ptr, 
                    Dgp_Int32 xbase, 
                    Dgp_Int32 ybase, 
                    size_t el_size)
{
    /* Free the memory block. This should give the first element. */
    (DGP_FREE)((void *)((Dgp_Int8 *)(*(ptr + ybase)) + xbase * (Dgp_Int32)el_size));
    /* and the pointer block */
    (DGP_FREE)((void *)(ptr + ybase));
}


/*
 * Aligned version of allocating image
 */
ErrCode
DgpImage_newOffsetAligned(Dgp_State *dgp_state, 
                    DgpImage_Type type, 
                    Dgp_Int32 width, 
                    Dgp_Int32 height,
                    Dgp_Int32 xbase, 
                    Dgp_Int32 ybase, 
                    Dgp_Bool align, 
                    void** image)
{
	void** areaPtr;
	DgpImage_Header *newHeader;
	DgpImage_Packed newPacked;
	Dgp_Bool dontclear = DGP_FALSE;
	
	DGP_PROLOGUE ("DgpImage_newOffsetAligned", DGP_PERRSTATE);
	
	if (type & DGPIMAGE_DONTCLEAR) {
		type &= (~DGPIMAGE_DONTCLEAR);
		dontclear = DGP_TRUE;
	}
	
       /*
	* Allocate the data area for the appropriate type of image, 
	* allocate and initialise the actual *Image and return it.
	* This exists in one form for each type of Image.
        *(( Note row_width for NON packed image is now defined as bpl
        * For packed image, row width is DOWRD per line
	*/
#define	ALLOC_INIT(imtype) \
    { \
	imtype newim; \
	DGP_CHECK_ALLOC (newim = (imtype)(DGP_MALLOC)(sizeof(* newim))); \
	newHeader = &newim->header; \
        if (align) \
            newHeader->row_width = ( sizeof(**newim->data) * width + sizeof(Dgp_Int32) -1) / sizeof(Dgp_UInt32) * sizeof(Dgp_UInt32); \
        else \
            newHeader->row_width = sizeof(**newim->data) * width; \
        if(width !=0 && height != 0){ \
	    areaPtr = allocImage(dgp_state, newHeader->row_width, height, \
                             xbase, ybase, \
	                     sizeof(**newim->data), dontclear); \
	    if (areaPtr == (void **)NULL) { \
	        (DGP_FREE)((void *)newim); \
	        DGP_ERR (cErrInsufficientMem); \
	    } \
        }else \
            areaPtr = NULL; /* allow empty image*/\
	newim->data = (void *)areaPtr; /* A better cast is possible */ \
	newim->dgp_state = dgp_state;		\
	*image = (void *)newim;  \
    }
	
	switch(type) {
	case DGPIMAGE_GRAY:		ALLOC_INIT(DgpImage_Gray);		break;
	case DGPIMAGE_FLOAT:	ALLOC_INIT(DgpImage_Float);		break;
	case DGPIMAGE_RGB:		ALLOC_INIT(DgpImage_ColorRGB);	break;
	case DGPIMAGE_DOUBLE:   ALLOC_INIT(DgpImage_Double);	break;
	case DGPIMAGE_BINARY:   ALLOC_INIT(DgpImage_Binary);	break;
	case DGPIMAGE_LONG:		ALLOC_INIT(DgpImage_Long);		break;
	case DGPIMAGE_PTR:		ALLOC_INIT(DgpImage_Ptr);		break;
	case DGPIMAGE_SHORT:	ALLOC_INIT(DgpImage_Short);		break;
	case DGPIMAGE_RGBFLOAT:	ALLOC_INIT(DgpImage_ColorRGBFloat);	break;
	case DGPIMAGE_PACKED:
		DGP_CHECK_ALLOC (newPacked = (DgpImage_Packed)(DGP_MALLOC)(sizeof(*newPacked)));
		newHeader = &newPacked->header;
		newPacked->dgp_state = dgp_state;
                if(width !=0 && height != 0){
		    newPacked->data = allocPackedData(width, height, xbase, ybase, dontclear, dgp_state);
		    if (newPacked->data == (Dgp_UInt32 **)NULL) {
			(DGP_FREE)((void *)newPacked);
			DGP_ERR (cErrInsufficientMem);
		    }    
                }else newPacked->data = (Dgp_UInt32 **)NULL; /*allow empty */
		*image = (void*) newPacked;
	        newHeader->row_width = DGPIMG_ROWLEN(width);
		break;
	default:    DGP_ERR	(cErrInvalidCmdLine);
	}
	
#undef	ALLOC_INIT

	/* Initialise the header */
    newHeader->width = width;
	newHeader->height = height;
	newHeader->xbase = xbase;
	newHeader->ybase = ybase;
	newHeader->tag = type;
	newHeader->comment = (Dgp_Int8 *)NULL;
	
	NORMAL_RETURN;
}

/*
 * Unaligned version of allocating image
 */
ErrCode
DgpImage_newOffset(Dgp_State *dgp_state, 
                   DgpImage_Type type, 
                   Dgp_Int32 width, 
                   Dgp_Int32 height, 
                   Dgp_Int32 xbase, 
                   Dgp_Int32 ybase, 
                   void** image)
{
	DGP_PROLOGUE ("DgpImage_newOffset", DGP_PERRSTATE);
	DGP_ERR_PASS (DgpImage_newOffsetAligned(dgp_state, type, width, height, xbase, ybase, DGP_FALSE, image));
	NORMAL_RETURN;
}
ErrCode
DgpImage_new(Dgp_State *dgp_state, 
                   DgpImage_Type type, 
                   Dgp_Int32 width, 
                   Dgp_Int32 height, 
                   void** image)
{
	
	DGP_PROLOGUE ("DgpImage_new", DGP_PERRSTATE);
	
	DGP_ERR_PASS (DgpImage_newOffset(dgp_state, type, width, height, 0, 0, image));
	NORMAL_RETURN;
}
ErrCode
DgpImage_newAligned(Dgp_State *dgp_state, 
                    DgpImage_Type type, 
                    Dgp_Int32 width, 
                    Dgp_Int32 height, 
                    void** image)
{
	
	DGP_PROLOGUE ("DgpImage_newAligned", DGP_PERRSTATE);
	
	DGP_ERR_PASS (DgpImage_newOffsetAligned(dgp_state, type, width, height, 0, 0, DGP_TRUE,image));
	NORMAL_RETURN;
}

/*
* Allocate and set up a block of memory to hold the image, plus
* a block of pointers into it. Offset the row pointer array by ybase;
* don't worry about xbase.
*/
static Dgp_UInt32 **
allocPackedData(Dgp_Int32 width, 
                   Dgp_Int32 height, 
                   Dgp_Int32 xbase, 
                   Dgp_Int32 ybase,
                   Dgp_Bool dontclear, 
                   Dgp_State *dgp_state)
{
	Dgp_UInt32 *mainArea;
	Dgp_UInt32 **ptrArea;
	Dgp_Int32 row_width = DGPIMG_ROWLEN(width);
	Dgp_Int32 lrow_width = row_width / sizeof(Dgp_UInt32);
	
	/* Allocate the image area */
	if (!dontclear) {
		mainArea =
			(Dgp_UInt32 *)(DGP_CALLOC)((size_t)height,
			lrow_width *
			sizeof(*mainArea));
	}
	else {
		mainArea =
			(Dgp_UInt32 *)(DGP_MALLOC)(height * lrow_width *
			sizeof(*mainArea));
	}
	if (mainArea == (Dgp_UInt32 *)NULL) {
		return((Dgp_UInt32 **)NULL);
	}
	
	/* Allocate a pointer for each row */
	ptrArea = allocPackedDataPointers(width, height, mainArea, lrow_width,
		xbase, ybase, dgp_state);
	if (ptrArea == (Dgp_UInt32 **)NULL) {
		(DGP_FREE)(mainArea);
		return((Dgp_UInt32 **)NULL);
	}
	
	return(ptrArea);
}

static Dgp_UInt32 **
allocPackedDataPointers(Dgp_Int32 width, 
                   Dgp_Int32 height, 
                   Dgp_UInt32 *data,
                   Dgp_Int32 longsperline, 
                   Dgp_Int32 xbase, 
                   Dgp_Int32 ybase, 
                   Dgp_State *dgp_state)
{
	Dgp_UInt32 **ptrArea;
	Dgp_Int32 i;
	Dgp_UInt32 *p, **q;
	
	/* kill a couple of warnings */
	width = width;
	xbase = xbase;
	
	/* Allocate the row pointers */
	ptrArea = (Dgp_UInt32 **)(DGP_MALLOC)(height *
		sizeof(*ptrArea));
	
	/* Set up the row pointers to point to the rows */
	for (i = 0, q = ptrArea, p = data; i < height; i++, q++, p += longsperline) {
		*q = p;
	}
	
	/* and return the (offset) row pointer area */
	return(ptrArea - ybase);
}

/* Wrap DgpImage around a buffer. Only work for DgpImage_Packed */
ErrCode
DgpImage_wrap(Dgp_State *dgp_state, 
              Dgp_UInt32 *data, 
              Dgp_Int32 width, 
              Dgp_Int32 height, 
              Dgp_Int32 longsperline, 
              DgpImage_Packed *newim)
{
	DgpImage_Header *newHeader;
	
	DGP_PROLOGUE ("DgpImage_wrap", DGP_PERRSTATE);
	
	DGP_ASSERT(data != (Dgp_UInt32 *)NULL);
	DGP_ASSERT(longsperline * (Dgp_Int32)sizeof(Dgp_UInt32) >= DGPIMG_ROWLEN(width));
	
	DGP_CHECK_ALLOC (*newim = (DgpImage_Packed)(DGP_MALLOC)(sizeof(**newim)));
	
	newHeader = &(*newim)->header;
	(*newim)->data = allocPackedDataPointers(width, height, data, longsperline,
		0, 0, dgp_state);
	if ((*newim)->data == (Dgp_UInt32 **)NULL) {
		(DGP_FREE)((void *) *newim);
		DGP_ERR (cErrNullPtr);
	}
	
	newHeader->width = width;
	newHeader->row_width = longsperline * sizeof(Dgp_UInt32);
	newHeader->height = height;
	newHeader->xbase = 0;
	newHeader->ybase = 0;
	newHeader->tag = DGPIMAGE_PACKED;
  newHeader->comment = (Dgp_Int8 *)NULL;
	
    /*  Added read-in of dgp_state... / dhtesq. */
    (*newim)->dgp_state = dgp_state;

	NORMAL_RETURN;
}

ErrCode
DgpImage_unwrap(DgpImage_Packed im, Dgp_UInt32 **retval)
{
	
	Dgp_State *dgp_state = im->dgp_state;
	
	DGP_PROLOGUE ("DgpImage_unwrap", DGP_PERRSTATE);
        DGP_STOP_WHINE;
	
	DGP_ASSERT(im != (DgpImage_Packed)NULL);
	DGP_ASSERT(im->data != (Dgp_UInt32 **)NULL);
	
	/* Free the pointer block and the image, and return the data area */
	*retval = im->data[im->header.ybase];
	(DGP_FREE)((void *)(im->data + im->header.ybase));
    /*  I get a debug error at run time when trying to free this in tcomp-scan-color. -- ? / dhtesq. */
    (DGP_FREE)((void *)im);
	
	NORMAL_RETURN;
}
ErrCode
DgpImage_save (void *im, Dgp_Int8 *filename)
{
	FILE *outFile;
	DgpImage_Packed pim = (DgpImage_Packed) im;
	Dgp_State *dgp_state = pim->dgp_state;
	DgpImage_Header *header;
	
	DGP_PROLOGUE ("DgpImage_pSave_", DGP_PERRSTATE);
	
	DGP_ASSERT(filename != (Dgp_Int8 *)NULL);
	DGP_ASSERT(im != (void *)NULL);
	header = &pim->header;
	DGP_ASSERT(header != (DgpImage_Header *)NULL);
	
	/* Open the file, imSaveF on it, then close it */
	outFile = fopen(filename, "wb");
	DGP_CHECK_NULL_CLEANUP (outFile);
	
	/* Write the image */
	DGP_ERR_PASS_CLEANUP (DgpImage_pSaveF (im, outFile)); 
	
	/* and close the file */
	if (fclose(outFile) == EOF) {
		/* Bah. */
		DGP_ERR_CLEANUP(cErrFileNotFound);	
	}
	
	NORMAL_RETURN;
ERR_CLEANUP:
    if(outFile) fclose(outFile);
    ERR_RETURN;

}

/* 23.3.2001 / dhtesq. */
/* Added newlines before maxval;  some third-party utilities were choking. */
ErrCode
DgpImage_pSaveF (void *im, FILE *outFile)
{
	
	DgpImage_Packed pim = (DgpImage_Packed) im;
	Dgp_State *dgp_state = pim->dgp_state;
	DgpImage_Header *header;
	DgpImage_Type tag;
	DGP_PROLOGUE ("DgpImage_pSaveF_", DGP_PERRSTATE);
	
	DGP_ASSERT(outFile != (FILE *)NULL);
	DGP_ASSERT(im != (void *)NULL);
	header = &pim->header;
	DGP_ASSERT(header != (DgpImage_Header *)NULL);
	tag = header->tag;

	if (tag != DGPIMAGE_PACKED) {
#ifdef DGP_FAXMODE
        DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else		
        Dgp_Int8 *startString;
		Dgp_Int8 *startTag;
		Dgp_Bool maxvalPresent;
		Dgp_Int32 i;
		Dgp_Int8 *p;
		Dgp_Int8 commentBuf[PBM_MAX_LINELEN];
		DgpImage_Gray gimage = (DgpImage_Gray) im;
		Dgp_State *dgp_state = gimage->dgp_state;
		
		/*
		* Get a pointer to the data area, the size of each element, and
		* the format of the header. The header may include a dummy entry
		* (some of the p*m formats have a maxval entry in the header - we
		* just use the maximum value we can represent).
		*/
#define	GET_SAVE_INFO(imtype, tag, string, maxval) \
		{ \
		startTag = tag; \
		startString = string; \
		maxvalPresent = maxval; \
		}
		
		switch (header->tag) {
		case DGPIMAGE_GRAY:	GET_SAVE_INFO(DgpImage_Gray, TAG_GRAY, "%d %d\n%d\n", DGP_TRUE);
			break;
			
		case DGPIMAGE_FLOAT:	GET_SAVE_INFO(DgpImage_Float, TAG_FLOAT, "%d %d\n", DGP_FALSE);
			break;
			
		case DGPIMAGE_RGB:	GET_SAVE_INFO(DgpImage_ColorRGB, TAG_RGB, "%d %d\n%d\n", DGP_TRUE);
			break;
			
		case DGPIMAGE_DOUBLE:	GET_SAVE_INFO(DgpImage_Double, TAG_DOUBLE,
									"%d %d\n", DGP_FALSE);
			break;
			
		case DGPIMAGE_BINARY:	GET_SAVE_INFO(DgpImage_Binary, TAG_BINARY,
									"%d %d\n", DGP_FALSE);
			break;
			
		case DGPIMAGE_LONG:	GET_SAVE_INFO(DgpImage_Long, TAG_LONG, "%d %d\n", DGP_FALSE);
			break;
			
		case DGPIMAGE_SHORT:	GET_SAVE_INFO(DgpImage_Short, TAG_SHORT, "%d %d\n", DGP_FALSE);
			break;
			
		case DGPIMAGE_RGBFLOAT:	GET_SAVE_INFO(DgpImage_Short, TAG_RGBFLOAT, "%d %d\n", DGP_FALSE);
			break;
			
		default: {
			DGP_ERR ( cErrInvalidCmdLine );
				 }
		}
		
#undef	GET_SAVE_INFO
		
		/* Write out the tag, and comments if any */
		if (fprintf(outFile, "%s\n", startTag) == EOF) {
			DGP_ERR (cErrFileNotFound);
		}
		
		if (header->comment != (Dgp_Int8 *)NULL) {
			p = header->comment;
			while (*p != '\0') {
				/* Copy a line into commentBuf */
				for (i = 0;
				(i < PBM_MAX_LINELEN) && (p[i] != '\n') && (p[i] != '\0');
				i++) {
					commentBuf[i] = p[i];
				}
				if (i == PBM_MAX_LINELEN) {
					/* The line was too long - search back for whitespace */
					for (i = PBM_MAX_LINELEN - 1;
					(i >= 0) && (!isspace(commentBuf[i]));
					i--) {
					}
					if (i < 0) {
						/* There was no whitespace. Break it how we like it... */
						commentBuf[PBM_MAX_LINELEN - 1] = '\0';
						i = PBM_MAX_LINELEN - 1;
					}
					else {
						/* Break at the whitespace */
						commentBuf[i] = '\0';
						i++;
					}
				}
				else {
					commentBuf[i] = '\0';
				}
				/*
				* At this point, commentBuf contains i characters to be
				* printed.
				*/
				if (fprintf(outFile, "#%s\n", commentBuf) == EOF) {
					DGP_ERR (cErrFileNotFound);
				}
				/* Advance the pointer */
				p += i;
				if (*p == '\n') {
					p++;
				}
			}
			
		}
		
		/* Write out the header info */
		if (maxvalPresent) {
			if (fprintf(outFile, startString, header->width, header->height,
				DGPIMG_COLRNG - 1) == EOF) {
				DGP_ERR (cErrFileNotFound);
			}
		}
		else {
			if (fprintf(outFile, startString, header->width,
				header->height) == EOF) {
				DGP_ERR (cErrFileNotFound);
			}
		}
		
		
		/* Write out the data block; force it to be bigendian */
		DGP_ERR_PASS (DgpImage_saveRawF_(im, header->tag, outFile, DGPIMAGE_BIGENDIAN, dgp_state));
#endif	
    }    
	else 
    {
		
		/* First, spew the header */
		if (fprintf(outFile, "%s\n%d %d\n",
			TAG_BINARY, (int)DGPIMG_WIDTH((DgpImage_Packed)im),
			(int)DGPIMG_HEIGHT((DgpImage_Packed)im)) == EOF) {
			DGP_ERR (cErrFileNotFound);	
		}
		
		if (!DgpImage_savePackedRaw((DgpImage_Packed)im, outFile, 0)) {
			DGP_ERR (cErrFileNotFound);	
		}
    }
	
	NORMAL_RETURN;

}

ErrCode
DgpImage_dup (void *im, Dgp_Bool copycomment, void** image)
{
	DgpImage_Packed dup;
	Dgp_UInt32 *p, *q;
	Dgp_Int32 y;
	Dgp_Int32 orowwidth, nrowwidth;
	DgpImage_Packed pim = (DgpImage_Packed) im;
	Dgp_State *dgp_state = pim->dgp_state;
	DgpImage_Header *header;
	DgpImage_Type tag;
	
	DGP_PROLOGUE ("DgpImage_dup_", DGP_PERRSTATE);
	
	DGP_ASSERT(im != (void *)NULL);
	header = &pim->header;
	DGP_ASSERT(header != (DgpImage_Header *)NULL);
    tag = header->tag;

	if (tag != DGPIMAGE_PACKED) {
#ifdef DGP_FAXMODE
            DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else

		void *dup;
		size_t el_size;
		Dgp_Int8 *oldbase;
		Dgp_Int8 *newbase;
		DgpImage_Header *newheader;
		DgpImage_Gray gimage = (DgpImage_Gray) im;
		Dgp_State *dgp_state = gimage->dgp_state;
		
		DGP_PROLOGUE ("DgpImage_unp_Dup_", DGP_PERRSTATE);
		
		DGP_ASSERT(im != (void *)NULL);
		
		/* Make a copy. */
		DGP_ERR_PASS (DgpImage_newOffset(dgp_state, 
                                         (DgpImage_Type)(header->tag | DGPIMAGE_DONTCLEAR),
                                         header->width, 
                                         header->height,
                                         header->xbase, 
                                         header->ybase, 
                                         &dup));

#define	GET_DUP_INFO(imtype) \
		{ \
		oldbase = \
		(void *)&(((imtype)im)->data[header->ybase][header->xbase]); \
		newbase = \
		(void *)&(((imtype)dup)->data[header->ybase][header->xbase]); \
		el_size = sizeof(**((imtype)im)->data); \
		newheader = &((imtype)im)->header; \
		}
		
		switch(header->tag) {
		case DGPIMAGE_GRAY:		GET_DUP_INFO(DgpImage_Gray);	break;
		case DGPIMAGE_FLOAT:		GET_DUP_INFO(DgpImage_Float);	break;
		case DGPIMAGE_RGB:		GET_DUP_INFO(DgpImage_ColorRGB);		break;
		case DGPIMAGE_DOUBLE:	GET_DUP_INFO(DgpImage_Double);	break;
		case DGPIMAGE_BINARY:	GET_DUP_INFO(DgpImage_Binary);	break;
		case DGPIMAGE_LONG:		GET_DUP_INFO(DgpImage_Long);	break;
		case DGPIMAGE_PTR:		GET_DUP_INFO(DgpImage_Ptr);		break;
		case DGPIMAGE_SHORT:		GET_DUP_INFO(DgpImage_Short);	break;
		case DGPIMAGE_RGBFLOAT:	GET_DUP_INFO(DgpImage_ColorRGBFloat);	break;
		default: {
			DGP_ERR (cErrInvalidCmdLine );
				 }
		}
		
#undef	GET_DUP_INFO
		
		/* Copy the comment, if any, and if we're supposed to */
		if (copycomment && (header->comment != (Dgp_Int8 *)NULL)) {
			newheader->comment = strdup(header->comment);
			if (newheader->comment == (Dgp_Int8 *)NULL) {
				/* Failed to copy */
				DgpImage_free (dup);
				DGP_ERR ( cErrInsufficientMem );
			}
		}
		
		/* Do the copy */
        /* change width to row_width (byte per line) to cope with DWORD alignment */
		(void)memcpy(newbase, oldbase,
			(el_size * header->row_width * header->height));
		
		*image = dup;
#endif
	}
	else
    {
		
		DGP_ERR_PASS (DgpImage_newOffset(dgp_state, DGPIMAGE_PACKED | DGPIMAGE_DONTCLEAR,
			DGPIMG_WIDTH(pim), DGPIMG_HEIGHT(pim),
			DGPIMG_XBASE(pim), DGPIMG_YBASE(pim), (void**) &dup));
		
		/* They might have passed in a padded image to imWrap... */
		if (pim->header.row_width != DGPIMG_ROWLEN(DGPIMG_WIDTH(pim))) {
			/* Sigh... */
			orowwidth = pim->header.row_width / sizeof(Dgp_UInt32);
			nrowwidth = DGPIMG_ROWLEN(DGPIMG_WIDTH(pim)) / sizeof(Dgp_UInt32);
			for (y = DGPIMG_YBASE(pim), p = dup->data[y], q = pim->data[y];
			y <= DGPIMG_YMAX(pim);
			y++, p += nrowwidth, q += orowwidth) {
				(void)memcpy(p, q, nrowwidth * sizeof(Dgp_UInt32));
			}
		}
		else {
			/* Aaah... */
			(void)memcpy(dup->data[DGPIMG_YBASE(dup)], pim->data[DGPIMG_YBASE(pim)],
				DGPIMG_GETROWWIDTH(pim) * DGPIMG_HEIGHT(pim));
		}
		
		*image = (void *)dup;
    }
	
	NORMAL_RETURN;

}
ErrCode
DgpImage_setOffset (void *im, Dgp_Int32 newxbase, Dgp_Int32 newybase)
{
	Dgp_Int32 oldybase;
	DgpImage_Packed pim = (DgpImage_Packed) im;
	Dgp_State *dgp_state = pim->dgp_state;
	DgpImage_Header *header;
	DgpImage_Type tag;
	
	DGP_PROLOGUE ("DgpImage_pSetOffset_", DGP_PERRSTATE);
	DGP_ASSERT(pim != (DgpImage_Packed)NULL);
	header = &pim->header;
	DGP_ASSERT(header != (DgpImage_Header *)NULL);
	tag = header->tag;
	
	if (tag != DGPIMAGE_PACKED) {
		
		Dgp_Int32 y;
		Dgp_Int32 oldxbase;
		Dgp_Int32 oldybase;
		/* DgpImage_Gray gimage = (DgpImage_Gray) im; */
		
		oldxbase = header->xbase;
		oldybase = header->ybase;
		
#define	FIX_PTRS(imtype) \
		{ \
		imtype vim = (imtype)(im); \
		for (y = DGPIMG_YBASE(vim); y <= DGPIMG_YMAX(vim); y++) { \
		vim->data[y] += (oldxbase - newxbase); \
		} \
		vim->data += (oldybase - newybase); \
		}
		
		switch(header->tag) {
		case DGPIMAGE_GRAY:		FIX_PTRS(DgpImage_Gray);	break;
		case DGPIMAGE_FLOAT:		FIX_PTRS(DgpImage_Float);	break;
		case DGPIMAGE_RGB:		FIX_PTRS(DgpImage_ColorRGB);	break;
		case DGPIMAGE_DOUBLE:	FIX_PTRS(DgpImage_Double);	break;
		case DGPIMAGE_BINARY:	FIX_PTRS(DgpImage_Binary);	break;
		case DGPIMAGE_LONG:		FIX_PTRS(DgpImage_Long);	break;
		case DGPIMAGE_PTR:		FIX_PTRS(DgpImage_Ptr);	break;
		case DGPIMAGE_SHORT:		FIX_PTRS(DgpImage_Short);	break;
		case DGPIMAGE_RGBFLOAT:	FIX_PTRS(DgpImage_ColorRGBFloat);break;
		default: {
			DGP_ERR	(cErrInvalidCmdLine);
				 }
		}
		
#undef	FIX_PTRS
		
    }
	else
    {
		DGP_ASSERT(pim->data != (Dgp_UInt32 **)NULL);
		
		/*
		* Since the row pointers aren't offset, just fix the pointer
		* block pointer and update the header.
		*/
		oldybase = pim->header.ybase;
		pim->data += (oldybase - newybase);
    }
	
	header->xbase = newxbase;
	header->ybase = newybase;
	
	NORMAL_RETURN;
}

ErrCode
DgpImage_convert (void *im, DgpImage_Type convertto, void** image)
{
	
	DgpImage_Packed pim = (DgpImage_Packed) im;
	Dgp_State *dgp_state = pim->dgp_state;
	DgpImage_Type tag;
	DgpImage_Header *header;
	
	DGP_PROLOGUE ("DgpImage_pConvert_", DGP_PERRSTATE);
	
	DGP_ASSERT(im != (void *)NULL);
	header = &pim->header;
	DGP_ASSERT(header != (DgpImage_Header *)NULL);
	tag = header->tag;	
	
	if ((tag != DGPIMAGE_PACKED) && (convertto != DGPIMAGE_PACKED)) {
        DgpImage_Type origtype;		
		origtype = header->tag;
		if (origtype == convertto) {
			/* They really just want a copy */
			DGP_ERR_PASS (DgpImage_dup (im, DGP_FALSE, image));
			NORMAL_RETURN;
		}
		else if (convertTable[origtype][convertto].isPromotion) {
			/* It's a promotion */
			DGP_ERR_PASS (DgpImage_pPromote (im, convertto, image));
			NORMAL_RETURN;
		}
		else {
			/* It's a demotion, or illegal - let imDemote handle it */
			DGP_ERR_PASS (DgpImage_pDemote (im, convertto, image));
			NORMAL_RETURN;
		}
	}
	else {
		
		if ((tag == DGPIMAGE_BINARY) || (convertto == DGPIMAGE_BINARY)) {
			DGP_ERR_PASS (DgpImage_pPromote (im, convertto, image));
			NORMAL_RETURN;
		}
		else if ((tag == DGPIMAGE_PACKED) && (convertto == DGPIMAGE_PACKED)) {
			DGP_ERR_PASS (DgpImage_dup (im, DGP_FALSE, image));
			NORMAL_RETURN;
		}
		else {
			DGP_ERR_PASS (DgpImage_pDemote (im, convertto, image));
			NORMAL_RETURN;
		}
	}
		
	NORMAL_RETURN;
}

ErrCode
DgpImage_getHeader (void *im, DgpImage_Header **header)
{
	DgpImage_Gray gim = (DgpImage_Gray) im;
	Dgp_State *dgp_state = gim->dgp_state;

	DGP_PROLOGUE ("DgpImage_getHeader", DGP_PERRSTATE);
        DGP_STOP_WHINE;

	DGP_ASSERT (header != NULL);

	*header = &gim->header;

	NORMAL_RETURN;
}
	
ErrCode
DgpImage_pPromote (void *im, DgpImage_Type promoteto, void** image)
{
	DgpImage_Binary tmp;
	DgpImage_Packed pim = (DgpImage_Packed) im;
	Dgp_State *dgp_state = pim->dgp_state;
	DgpImage_Header *header;
	DgpImage_Type tag;
	
	DGP_PROLOGUE ("DgpImage_pPromote_", DGP_PERRSTATE);
	
	DGP_ASSERT(im != (void *)NULL);
	header = &pim->header;
	DGP_ASSERT(header != (DgpImage_Header *)NULL);
	tag = header->tag;

	if ((tag != DGPIMAGE_PACKED) && (promoteto != DGPIMAGE_PACKED)) {
		DgpImage_Type origtype;
		origtype = header->tag;
		if (!convertTable[origtype][promoteto].isPromotion) {
			/* They just requested a non-promotion. Bad... */
			DGP_PANIC("non-promotion request made to imPromote()");
		}
		DGP_ASSERT(convertTable[origtype][promoteto].convertFunc !=
			(void * (*)(void *))NULL);
		
		*image = (*convertTable[origtype][promoteto].convertFunc)(im);
	}
	else {
		if (promoteto == DGPIMAGE_PACKED) {
			if (tag == DGPIMAGE_BINARY) {
				*image = (void *)binaryToPacked((DgpImage_Binary)im);
				NORMAL_RETURN;
			}
			else {
				DgpImage_Header *header;
				
				DGP_ASSERT(tag != DGPIMAGE_PACKED);
				
				/* Try to make it into a DgpImage_Binary. */
				DGP_ERR_PASS (DgpImage_getHeader(im, &header));
				DGP_ERR_PASS (DgpImage_convert(im,
					DGPIMAGE_BINARY, (void**) &tmp));
				
				*image = (void *)binaryToPacked(tmp);
				DgpImage_free (tmp);
				NORMAL_RETURN;
			}
		}
		else if (promoteto == DGPIMAGE_BINARY) {
			DGP_ASSERT(tag == DGPIMAGE_PACKED);
			*image = (void *)packedToBinary((DgpImage_Packed)im);
			NORMAL_RETURN;
		}
		else {
			DGP_ASSERT(tag == DGPIMAGE_PACKED);
			
			tmp = packedToBinary((DgpImage_Packed)im);
			DGP_CHECK_NULL (tmp);
			
			DgpImage_convert (tmp, promoteto, image);
			DgpImage_free (tmp);
			NORMAL_RETURN;
		}
	}
	
	NORMAL_RETURN;
}

ErrCode
DgpImage_pDemote (void *im, DgpImage_Type demoteto, void** image)
{
	DgpImage_Binary tmp;
	DgpImage_Header *header;
	DgpImage_Packed pim = (DgpImage_Packed) im;
	Dgp_State *dgp_state = pim->dgp_state;
	DgpImage_Type tag;
	
	DGP_PROLOGUE ("DgpImage_pDemote_", DGP_PERRSTATE);
	DGP_ASSERT(im != (void *)NULL);
	header = &pim->header;
	DGP_ASSERT(header != (DgpImage_Header *)NULL);
	tag = header->tag;
	
	if ((tag != DGPIMAGE_PACKED) && (demoteto != DGPIMAGE_PACKED)) {
		DGP_ERR_PASS (DgpImage_getHeader(im, &header));
		DGP_ERR_PASS (DgpImage_pDemote (im, demoteto, image));
		NORMAL_RETURN;
	}
	
	if (demoteto == DGPIMAGE_PACKED) {
		DGP_ASSERT((tag != DGPIMAGE_PACKED) && (tag != DGPIMAGE_BINARY));
		
		/* Try to make it into a DgpImage_Binary. */
		DGP_ERR_PASS (DgpImage_getHeader(im, &header));
		DGP_ERR_PASS (DgpImage_pDemote(im, DGPIMAGE_BINARY, (void**) &tmp));
		
		*image = (void *)binaryToPacked(tmp);
		DgpImage_free (tmp);
		NORMAL_RETURN;
	}
	else {
		DGP_ASSERT(tag == DGPIMAGE_PACKED);
		DGP_ASSERT(demoteto != DGPIMAGE_BINARY);
		
		tmp = packedToBinary((DgpImage_Packed)im);
		DGP_CHECK_NULL (tmp);
		
		DGP_ERR_PASS (DgpImage_pDemote (tmp, demoteto, image));
		DgpImage_free (tmp);
		NORMAL_RETURN;
	}
}
static DgpImage_Binary
packedToBinary(DgpImage_Packed im)
{
	DgpImage_Binary bim;
	DgpImage_BinaryPixPtr bptr;
	Dgp_Int32 x, y;
	Dgp_UInt32 *p;
	Dgp_UInt32 c = 0;
	Dgp_Int32 bitsleft;
	Dgp_Int32 xend, yend;
	Dgp_Int32 xstart;
	Dgp_State *dgp_state = im->dgp_state;
	
	DGP_VOID_PROLOGUE ("packedToBinary", DGP_PERRSTATE);
	DGP_VOID_ASSERT(im != (DgpImage_Packed)NULL);
	
	DgpImage_newOffset(dgp_state, DGPIMAGE_BINARY,
		DGPIMG_WIDTH(im), DGPIMG_HEIGHT(im),
		DGPIMG_XBASE(im), DGPIMG_YBASE(im), (void**) &bim);
	if (bim == (DgpImage_Binary)NULL) {
		return((DgpImage_Binary)NULL);
	}
	
	xstart = DGPIMG_XBASE(im);
	xend = DGPIMG_XMAX(im);
	yend = DGPIMG_YMAX(im);
	
	if (DGPIMG_WIDTH(im) == 0) {
		/* Silly case, but should be handled. */
		return(bim);
	}
	
	for (y = DGPIMG_YBASE(im); y <= yend; y++) {
		bitsleft = 0;
		p = DGPIMG_GETROW(im, y);
		for (x = xstart, bptr = DGPIMG_PIXPTR(bim, xstart, y);
		x <= xend;
		x++, DGPIMG_PTRRIGHT(bim, bptr)) {
			if (bitsleft == 0) {
				c = *p++;
				bitsleft = DGP_BITS_UINT32;
			}
			/* High bit first */
			if (c & (1ul << (DGP_BITS_UINT32 - 1))) {
				DGPIMG_PTRREF(bim, bptr) = 1;
			}
			else if (c == 0) {
				/* It's already zero by default, and we've run out of bits */
				DGPIMG_PTRRIGHTN(im, bptr, bitsleft - 1);
				x += bitsleft - 1;
				bitsleft = 0;
				continue;
			}
			bitsleft--;
			c <<= 1;
		}
	}
	
	return(bim);
}

static DgpImage_Packed
binaryToPacked(DgpImage_Binary im)
{
	DgpImage_Packed pim;
	Dgp_Int32 x, y, bits, endx;
	Dgp_UInt32 *p;
	Dgp_UInt32 b;
	DgpImage_BinaryPixPtr bp;
	Dgp_State *dgp_state = im->dgp_state;
	
	DGP_VOID_PROLOGUE ("binaryToPacked", DGP_PERRSTATE);
	
	DGP_VOID_ASSERT(im != (DgpImage_Binary)NULL);
	
	DgpImage_newOffset(dgp_state, DGPIMAGE_PACKED | DGPIMAGE_DONTCLEAR,
		DGPIMG_WIDTH(im), DGPIMG_HEIGHT(im),
		DGPIMG_XBASE(im), DGPIMG_YBASE(im), (void**) &pim);
	if (pim == (DgpImage_Packed)NULL) {
		return((DgpImage_Packed)NULL);
	}
	
	p = DGPIMG_GETROW(pim, DGPIMG_YBASE(im));
	endx = DGPIMG_XMAX(im);
	for (y = DGPIMG_YBASE(im); y <= DGPIMG_YMAX(im); y++) {
		for (x = DGPIMG_XBASE(im), bits = DGP_BITS_UINT32 - 1, b = 0,
			bp = DGPIMG_PIXPTR(im, DGPIMG_XBASE(im), y);
		x <= endx;
		x++, bits--, DGPIMG_PTRRIGHT(im, bp)) {
			if (DGPIMG_PTRREF(im, bp)) {
				b |= 1;
			}
			if (bits == 0) {
				*p++ = b;
				bits = DGP_BITS_UINT32;
				b = 0;
			}
			else {
				b <<= 1;
			}
		}
		if (bits != DGP_BITS_UINT32 - 1) {
			b <<= bits;
			*p++ = b;
		}
	}
	
	return(pim);
}

static void
getAreaAndSize(void *im, 
               DgpImage_Header *header,
               void ***areaPtr, 
               size_t *el_size, 
               Dgp_State *dgp_state)
{
	
    DGP_VOID_PROLOGUE ("getAreaAndSize", DGP_PERRSTATE);
	
    DGP_VOID_ASSERT(im != (void *)NULL);
    DGP_VOID_ASSERT(header != (DgpImage_Header *)NULL);
    DGP_VOID_ASSERT(areaPtr != (void ***)NULL);
    DGP_VOID_ASSERT(el_size != (size_t *)NULL);
	
#define	GET_AREA_SIZE(imtype) \
    { \
	imtype vim = (imtype)im; \
	*areaPtr = (void **)vim->data; \
	*el_size = sizeof(**vim->data); \
	}
	
    switch(header->tag) {
	case DGPIMAGE_GRAY:		GET_AREA_SIZE(DgpImage_Gray);	break;
	case DGPIMAGE_FLOAT: 	GET_AREA_SIZE(DgpImage_Float);	break;
	case DGPIMAGE_RGB:		GET_AREA_SIZE(DgpImage_ColorRGB);break;
	case DGPIMAGE_DOUBLE:	GET_AREA_SIZE(DgpImage_Double);	break;
	case DGPIMAGE_BINARY:	GET_AREA_SIZE(DgpImage_Binary);	break;
	case DGPIMAGE_LONG:		GET_AREA_SIZE(DgpImage_Long);	break;
	case DGPIMAGE_SHORT:    GET_AREA_SIZE(DgpImage_Short);	break;
	case DGPIMAGE_RGBFLOAT:	GET_AREA_SIZE(DgpImage_ColorRGBFloat);	break;
    case DGPIMAGE_PTR:		GET_AREA_SIZE(DgpImage_Ptr);	break;
	default: {
		DGP_PANIC("bad image tag");
			 }
	}
	
#undef	GET_AREA_SIZE
}


#ifndef DGP_FAXMODE
/*
* Root through the image's header to find its type, and get the appropriate
* base pointer of the data area, and the element size.
*/




ErrCode
DgpImage_load(Dgp_State *dgp_state, 
              DgpImage_Type type, 
              Dgp_Int8 *filename, 
              void **image)
{
	FILE *inFile;
	
	DGP_PROLOGUE ("DgpImage_pLoad", DGP_PERRSTATE);
	
	/* Open the file, imLoadF on it, then close it */
	DGP_ASSERT(filename != (Dgp_Int8 *)NULL);
	DGP_CHECK_NULL (inFile = fopen(filename, "rb"));
	
	DGP_ERR_PASS (DgpImage_pLoadF(dgp_state, type, inFile, image));
	
	if (fclose(inFile) == EOF) {
    /*
	* It would be a strange day that fclose on a read-only file
	* gave an error, but...
		*/
		if (*image) {
			DgpImage_free (*image);
		}
		DGP_ERR (cErrFileNotFound);
	}
	
	NORMAL_RETURN;
}

ErrCode
DgpImage_pLoadF(Dgp_State *dgp_state, 
              DgpImage_Type type, 
              FILE *inFile, 
              void** image)
{
	Dgp_Int8 fileHeader[PBM_MAX_LINELEN];
	Dgp_Int32 c;
	Dgp_Int32 i;
	
	DGP_PROLOGUE ("DgpImage_pLoadF", DGP_PERRSTATE);
	
	DGP_ASSERT(inFile != (FILE *)NULL);
	
	/*
	* Read up to the first whitespace in order to find the file's header.
	*/
	i = 0;
	do {
		c = getc(inFile);
		if (c == EOF) {
			DGP_ERR (cErrFileNotFound);
		}
		fileHeader[i++] = (Dgp_Int8) c;
	} while (!isspace(c));
	fileHeader[--i] = '\0';		/* Clobber the whitespace */
	
	if (type != DGPIMAGE_PACKED) 
    {
		DgpImage_Type filetype = (DgpImage_Type)0;
		void * (*readfunc)(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader) = NULL;
		void *readim;
		DgpImage_Header *readimHeader;
		
		/* Determine who is supposed to handle this sort of file */
		for (i = 0; i < NREADERS; i++) {
			if (strcmp(fileHeader, fileReaders[i].filetag) == 0) {
				/* Found it */
				filetype = fileReaders[i].filetype;
				readfunc = fileReaders[i].readfunc;
				break;
			}
		}
		if (i == NREADERS) {
			/* Didn't match any of the ones in the table - unknown type. */
			DGP_ERR (cErrFileNotFound);
		}
				
		/* Read in the file */
		readim = (*readfunc)(dgp_state, inFile, &readimHeader);
		if (readim == (void *)NULL) {
			/* Didn't load for whatever reason */
			DGP_ERR (cErrFileNotFound);
		}
		
		if (filetype == type) {
			/* The file was of the expected type - no conversion required */
			*image = readim;
			NORMAL_RETURN;
		}
		
		/*
		* OK. We read in a file, and it wasn't what we expected. We get to
		* promote it to the expected image type, if that's defined.
		*
		* If that conversion isn't a promotion, kick out.
		*/
		if (!convertTable[filetype][type].isPromotion) {
			DgpImage_free(readim);
			DGP_ERR (cErrInvalidCmdLine);
		}
		
		/* Convert... */
		DGP_ERR_CHECK (DgpImage_convert(readim, type, image)) {
			/* Didn't work. */
			DgpImage_free(readim);
			ERR_RETURN;
		}
		
		/* Kill the image we read in, and return the promoted one. */
		DgpImage_free(readim);
		
	}
	else {
		
		if (!strcmp(fileHeader, TAG_BINARY)) {
			*image = readRaw(inFile, dgp_state);
		}
		else if (!strcmp(fileHeader, TAG_BINARY_COOKED)) {
			*image = readCooked(inFile, dgp_state);
		}
		else {
			/* Not a PBM file */
			DGP_ERR (cErrFileNotFound);	
		}
	}
	
	NORMAL_RETURN;
}



/*
17.7.2000 / dhtesq.
This is just DgpImage_pSaveF() without the data write.  I came across the need for this
while implementing incremental, strip-by-strip writes.  The file header has to contain
the page values for width, height, etc., but the data writes must have strip dimensions.
*/
ErrCode
DgpImage_pSaveHeaderF (void *im, FILE *outFile)
{
	
	DgpImage_Packed pim = (DgpImage_Packed) im;
	Dgp_State *dgp_state = pim->dgp_state;
	DgpImage_Header *header;
	DgpImage_Type tag;
	DGP_PROLOGUE ("DgpImage_pSaveF_", DGP_PERRSTATE);
	
	DGP_ASSERT(outFile != (FILE *)NULL);
	DGP_ASSERT(im != (void *)NULL);
	header = &pim->header;
	DGP_ASSERT(header != (DgpImage_Header *)NULL);
	tag = header->tag;

	if (tag != DGPIMAGE_PACKED) 
    {
		Dgp_Int8 *startString;
		Dgp_Int8 *startTag;
		Dgp_Bool maxvalPresent;
		Dgp_Int32 i;
		Dgp_Int8 *p;
		Dgp_Int8 commentBuf[PBM_MAX_LINELEN];
		/* DgpImage_Gray gimage = (DgpImage_Gray) im; */
		/* Dgp_State *dgp_state = gimage->dgp_state; */
		
		/*
		* Get a pointer to the data area, the size of each element, and
		* the format of the header. The header may include a dummy entry
		* (some of the p*m formats have a maxval entry in the header - we
		* just use the maximum value we can represent).
		*/
#define	GET_SAVE_INFO(imtype, tag, string, maxval) \
		{ \
		startTag = tag; \
		startString = string; \
		maxvalPresent = maxval; \
		}
		
		switch (header->tag) {
		case DGPIMAGE_GRAY:	GET_SAVE_INFO(DgpImage_Gray, TAG_GRAY, "%d %d %d\n", DGP_TRUE);
			break;
			
		case DGPIMAGE_FLOAT:	GET_SAVE_INFO(DgpImage_Float, TAG_FLOAT, "%d %d\n", DGP_FALSE);
			break;
			
		case DGPIMAGE_RGB:	GET_SAVE_INFO(DgpImage_ColorRGB, TAG_RGB, "%d %d %d\n", DGP_TRUE);
			break;
			
		case DGPIMAGE_DOUBLE:	GET_SAVE_INFO(DgpImage_Double, TAG_DOUBLE,
									"%d %d\n", DGP_FALSE);
			break;
			
		case DGPIMAGE_BINARY:	GET_SAVE_INFO(DgpImage_Binary, TAG_BINARY,
									"%d %d\n", DGP_FALSE);
			break;
			
		case DGPIMAGE_LONG:	GET_SAVE_INFO(DgpImage_Long, TAG_LONG, "%d %d\n", DGP_FALSE);
			break;
			
		case DGPIMAGE_SHORT:	GET_SAVE_INFO(DgpImage_Short, TAG_SHORT, "%d %d\n", DGP_FALSE);
			break;
			
		case DGPIMAGE_RGBFLOAT:	GET_SAVE_INFO(DgpImage_Short, TAG_RGBFLOAT, "%d %d\n", DGP_FALSE);
			break;
			
		default: {
			DGP_ERR ( cErrInvalidCmdLine );
				 }
		}
		
#undef	GET_SAVE_INFO
		
		/* Write out the tag, and comments if any */
		if (fprintf(outFile, "%s\n", startTag) == EOF) {
			DGP_ERR (cErrFileNotFound);
		}
		
		if (header->comment != (Dgp_Int8 *)NULL) {
			p = header->comment;
			while (*p != '\0') {
				/* Copy a line into commentBuf */
				for (i = 0;
				(i < PBM_MAX_LINELEN) && (p[i] != '\n') && (p[i] != '\0');
				i++) {
					commentBuf[i] = p[i];
				}
				if (i == PBM_MAX_LINELEN) {
					/* The line was too long - search back for whitespace */
					for (i = PBM_MAX_LINELEN - 1;
					(i >= 0) && (!isspace(commentBuf[i]));
					i--) {
					}
					if (i < 0) {
						/* There was no whitespace. Break it how we like it... */
						commentBuf[PBM_MAX_LINELEN - 1] = '\0';
						i = PBM_MAX_LINELEN - 1;
					}
					else {
						/* Break at the whitespace */
						commentBuf[i] = '\0';
						i++;
					}
				}
				else {
					commentBuf[i] = '\0';
				}
				/*
				* At this point, commentBuf contains i characters to be
				* printed.
				*/
				if (fprintf(outFile, "#%s\n", commentBuf) == EOF) {
					DGP_ERR (cErrFileNotFound);
				}
				/* Advance the pointer */
				p += i;
				if (*p == '\n') {
					p++;
				}
			}
			
		}
		
		/* Write out the header info */
		if (maxvalPresent) {
			if (fprintf(outFile, startString, header->width, header->height,
				DGPIMG_COLRNG - 1) == EOF) {
				DGP_ERR (cErrFileNotFound);
			}
		}
		else {
			if (fprintf(outFile, startString, header->width,
				header->height) == EOF) {
				DGP_ERR (cErrFileNotFound);
			}
		}
				
    }    
	else 
    {
		
		/* First, spew the header */
		if (fprintf(outFile, "%s\n%d %d\n",
			TAG_BINARY, (int)DGPIMG_WIDTH((DgpImage_Packed)im),
			(int)DGPIMG_HEIGHT((DgpImage_Packed)im)) == EOF) {
			DGP_ERR (cErrFileNotFound);	
		}
    }
	
	NORMAL_RETURN;
}



#if 0
static ErrCode
dgpimage_unp_demote (void *im, DgpImage_Header *header, DgpImage_Type demoteto, void **image)
{
    DgpImage_Type origtype;
	DgpImage_Gray gimage = (DgpImage_Gray) im;
	Dgp_State *dgp_state = gimage->dgp_state;
	
	DGP_PROLOGUE ("DgpImage_unp_Demote_", DGP_PERRSTATE);
	
    DGP_ASSERT(im != (void *)NULL);
    DGP_ASSERT(header != (DgpImage_Header *)NULL);
	
    origtype = header->tag;
    if (convertTable[origtype][demoteto].isPromotion) {
		/* They just requested a promotion. Bad... */
		DGP_PANIC("promotion request made to imDemote()");
	}
    else if (convertTable[origtype][demoteto].convertFunc == (void * (*)(void *))NULL) {
		DGP_PANIC("impossible request made to imDemote()");
	}
	
    *image = (*convertTable[origtype][demoteto].convertFunc)(im);
	
    NORMAL_RETURN;
}
#endif /* 0 */

/* 11.1.2001 / dhtesq. */
/* Added these two color wrap/unwrap functions...needed for java-based viewer. */
ErrCode
DgpImage_wrapRGB(Dgp_State *dgp_state, 
              Dgp_UInt32 *data, 
              Dgp_Int32 width, 
              Dgp_Int32 height, 
              Dgp_Int32 bpl, 
              DgpImage_ColorRGB *newim)
{
	DgpImage_Header *newHeader;
    register Dgp_Int32 i;
    register void **ptrArea = NULL;
    register Dgp_Int8 *p;
    register void **q;
	
	DGP_PROLOGUE ("DgpImage_wrapRGB", DGP_PERRSTATE);
	
	DGP_ASSERT(data != (Dgp_UInt32 *)NULL);
	/*DGP_ASSERT(longsperline * (Dgp_Int32)sizeof(Dgp_UInt32) >= DGPIMG_ROWLEN(width)); */
	
    /* Allocate for image structure. */
	DGP_CHECK_ALLOC (*newim = (DgpImage_ColorRGB)(DGP_MALLOC)(sizeof(**newim)));	
	
    /* Allocate a pointer for each row. */
    DGP_CHECK_ALLOC(ptrArea = (void **)(DGP_MALLOC)(height * sizeof(void *)));
	
    /* Set up the row pointers to point to the (offset) rows... */
    for (i = 0, q = ptrArea, p = (Dgp_Int8 *)data; i < height; i++, q++, p += bpl) {
		*q = (void *)p;
	}
	
	(*newim)->data = (DgpImage_RGB **)ptrArea;
	if ((*newim)->data == (DgpImage_RGB **)NULL) {
		(DGP_FREE)((void *) *newim);
		DGP_ERR (cErrNullPtr);
	}

	newHeader = &(*newim)->header;
    
	newHeader->width = width;
	newHeader->row_width = bpl;/*longsperline * sizeof(Dgp_UInt32); */
	newHeader->height = height;
	newHeader->xbase = 0;
	newHeader->ybase = 0;
	newHeader->tag = DGPIMAGE_RGB;
    newHeader->comment = (Dgp_Int8 *)NULL;
	
    /*  Added read-in of dgp_state... / dhtesq. */
    (*newim)->dgp_state = dgp_state;

	NORMAL_RETURN;
}

ErrCode
DgpImage_unwrapRGB(DgpImage_ColorRGB im, Dgp_UInt32 **retval)
{	
	Dgp_State *dgp_state = im->dgp_state;
	
	DGP_PROLOGUE ("DgpImage_unwrapRGB", DGP_PERRSTATE);
        DGP_STOP_WHINE;
	
	DGP_ASSERT(im != (DgpImage_ColorRGB)NULL);
	DGP_ASSERT(im->data != (DgpImage_RGB **)NULL);
	
	/* Free the pointer block and the image, and return the data area. */
	*retval = (Dgp_UInt32 *)(im->data[im->header.ybase]);
	(DGP_FREE)((void *)(im->data + im->header.ybase));
    (DGP_FREE)((void *)im);
	
	NORMAL_RETURN;
}
/*//////////////// */

ErrCode
DgpImage_wrapAll( Dgp_State *dgp_state, 
              Dgp_UInt32 *data, 
              DgpImage_Type type,
              Dgp_Int32 width, 
              Dgp_Int32 height, 
              Dgp_Int32 bpl, 
              void **newimage)
{
#define	ALLOC_WRAP(imtype) \
{ \
        register Dgp_Int32 i; \
        register void **ptrArea; \
        register Dgp_Int8 *p; \
        register void **q; \
	imtype newim; \
        /*Allocate for image structure*/ \
	DGP_CHECK_ALLOC( newim = (imtype)(DGP_MALLOC)(sizeof(* newim))); \
        /* Allocate a pointer for each row. */ \
        ptrArea = (void **)(DGP_MALLOC)(height * sizeof(void *)); \
	if (!ptrArea) { \
            (DGP_FREE)((void *) newim); \
	    DGP_ERR (cErrInsufficientMem); \
        } \
        /* Set up the row pointers to point to the (offset) rows... */ \
        for (i = 0, q = ptrArea, p = (Dgp_Int8 *)data; i < height; i++, q++, p += bpl) {\
		*q = (void *)p; \
	}\
    newim->data = (void *) ptrArea; \
	newim->header.width = width; \
	newim->header.row_width = bpl; \
	newim->header.height = height; \
	newim->header.xbase = 0; \
	newim->header.ybase = 0; \
	newim->header.tag = type; \
    newim->header.comment = (Dgp_Int8 *)NULL; \
    /*Added read-in of dgp_state... */ \
    newim->dgp_state = dgp_state; \
    *newimage = (void *) newim; \
}


	DGP_PROLOGUE ("DgpImage_wrapALL", DGP_PERRSTATE);
        DGP_CHECK_NULL(data);
        DGP_CHECK_NULL(newimage);
        switch(type) {
        case DGPIMAGE_GRAY:     ALLOC_WRAP(DgpImage_Gray);      break;
        case DGPIMAGE_FLOAT:    ALLOC_WRAP(DgpImage_Float);     break;
        case DGPIMAGE_RGB:      ALLOC_WRAP(DgpImage_ColorRGB);  break;
        case DGPIMAGE_DOUBLE:   ALLOC_WRAP(DgpImage_Double);    break;
        case DGPIMAGE_BINARY:   ALLOC_WRAP(DgpImage_Binary);    break;
        case DGPIMAGE_LONG:     ALLOC_WRAP(DgpImage_Long);      break;
        case DGPIMAGE_PTR:      ALLOC_WRAP(DgpImage_Ptr);       break;
        case DGPIMAGE_SHORT:    ALLOC_WRAP(DgpImage_Short);     break;
        case DGPIMAGE_RGBFLOAT: ALLOC_WRAP(DgpImage_ColorRGBFloat);     break;
        case DGPIMAGE_PACKED:
            DGP_ERR_PASS(DgpImage_wrap(dgp_state, data, width, height, 
                bpl / sizeof(Dgp_Int32), (DgpImage_Packed *) newimage)); break;
        default:    DGP_ERR     (cErrInvalidCmdLine);
        }
#undef ALLOC_WRAP

       NORMAL_RETURN;

}

ErrCode
DgpImage_unwrapAll(void *_im, Dgp_UInt32 **retval)
{
    DgpImage_Packed im = (DgpImage_Packed)_im;
	Dgp_State *dgp_state = im->dgp_state;
	
	DGP_PROLOGUE ("DgpImage_unwrapAll", DGP_PERRSTATE);
	
	DGP_CHECK_NULL(im);
	DGP_CHECK_NULL(im->data);
	
	/* Free the pointer block and the image, and return the data area. */
	*retval = (Dgp_UInt32 *)(im->data[im->header.ybase]);
	(DGP_FREE)((void *)(im->data + im->header.ybase));
        (DGP_FREE)((void *)im);
	NORMAL_RETURN;
}
/*
* Read in a raw (non-printable) PBM file. Someone has already eaten the
* file tag.
*/
DgpImage_Packed
readRaw(FILE *inFile, Dgp_State *dgp_state)
{
	Dgp_Int32 width, height;
	Dgp_Bool failed;
	DgpImage_Packed im;
	
	/*
	* File format:
	* P4 <width> <height>
	* binary image, packed, high bit is leftmost; rows begin on byte
	* boundaries.
	*/
	width = DgpImage_pbmGetToken(inFile, &failed, (Dgp_Int8 *)NULL, (Dgp_Int32 *)NULL, dgp_state);
	if (failed) {
		return((DgpImage_Packed)NULL);
	}
	height = DgpImage_pbmGetToken(inFile, &failed, (Dgp_Int8 *)NULL, (Dgp_Int32 *)NULL, dgp_state);
	if (failed) {
		return((DgpImage_Packed)NULL);
	}
	
	/* Allocate it; no need to clear it. */
	DgpImage_new(dgp_state, DGPIMAGE_PACKED | DGPIMAGE_DONTCLEAR, width, height, (void**) &im);
	if (im == (DgpImage_Packed)NULL) {
		return((DgpImage_Packed)NULL);
	}
	
	/* call out to rawimage.c for the rest */
	if (!DgpImage_readPackedRaw(im, width, height, inFile, 0)) {
		DgpImage_free(im);
		return((DgpImage_Packed)NULL);
	}
	
	return(im);
}

/*
* Read in a cooked (printable) PBM file.
*/
static DgpImage_Packed
readCooked(FILE *inFile, Dgp_State *dgp_state)
{
	Dgp_Int32 width, height;
	Dgp_Bool failed;
	DgpImage_Packed im;
	Dgp_Int32 x, y;
	
	/*
	* File format:
	* P1 <width> <height>
	* bunches of 0[<ws>] and 1[<ws>], suitable for reading with
	* DgpImage_pbmGetPBMToken.
	*/
	width = DgpImage_pbmGetToken(inFile, &failed, (Dgp_Int8 *)NULL, (Dgp_Int32 *)NULL, dgp_state);
	if (failed) {
		return((DgpImage_Packed)NULL);
	}
	height = DgpImage_pbmGetToken(inFile, &failed, (Dgp_Int8 *)NULL, (Dgp_Int32 *)NULL, dgp_state);
	if (failed) {
		return((DgpImage_Packed)NULL);
	}
	
	/* Allocate and clear it. */
	DgpImage_new(dgp_state, DGPIMAGE_PACKED, width, height, (void**) &im);
	if (im == (DgpImage_Packed)NULL) {
		return((DgpImage_Packed)NULL);
	}
	
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			/* Can't use DgpImage_pbmGetToken - may not be separating whitespace */
			if (DgpImage_pbmGetPBMToken(inFile, &failed, (Dgp_Int8 *)NULL, (Dgp_Int32 *)NULL, dgp_state)) {
				DGPIMG_SETBIT(im, x, y);
			}
			if (failed) {
				DgpImage_free(im);
				return((DgpImage_Packed)NULL);
			}
		}
	}
	
	return(im);
}


ErrCode
DgpImage_setComment (void *im, Dgp_Int8 *newcomment)
{
    Dgp_Int8 *newcopy;
    DgpImage_Gray gimage = (DgpImage_Gray) im;
    Dgp_State *dgp_state = gimage->dgp_state;
    DgpImage_Header *header;
    
    DGP_PROLOGUE ("DgpImage_setComment_", DGP_PERRSTATE);
	
    /* stop compiler from whining */
    (void)im;
	
    DGP_ASSERT(im != (void *)NULL);
    header = &gimage->header;
    DGP_ASSERT(header != (DgpImage_Header *)NULL);
	
    if (newcomment != (Dgp_Int8 *)NULL) {
		newcopy = strdup(newcomment);
		if (newcopy == (Dgp_Int8 *)NULL) {
			DGP_ERR (cErrInsufficientMem);
		}
	}
    else {
		newcopy = (Dgp_Int8 *)NULL;
	}
	
    if (header->comment != (Dgp_Int8 *)NULL) {
		/* Free the old comment */
		(DGP_FREE)((void *)header->comment);
	}
	
    header->comment = newcopy;
    NORMAL_RETURN;
}

/*
* Here are the actual file-reading routines.
*
* Each one is passed a FILE * and an DgpImage_Header **. It knows that the file
* stream is positioned immediately after a P?M (or similar) magic number,
* and implicitly what that magic number was (i.e. what form the file will
* take). It then reads the file, creates an Image of the appropriate type,
* and returns it as its return value, putting a pointer to its DgpImage_Header
* into *readimHeader.
*/

static void *
readBinaryCooked(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader)
{
    Dgp_Int32 width, height;
    Dgp_Bool failed;
    DgpImage_Binary im;
    Dgp_Int32 x, y;
	Dgp_Int8 comment[LOAD_COMMENT_MAX];
	Dgp_Int32 commentsize = 0;
	
	DGP_VOID_PROLOGUE ("readBinaryCooked", DGP_PERRSTATE);

    /*
	* File format:
	* P1 <width> <height>
	* bunches of 0<ws> and 1<ws>, suitable for reading with pbmGetPBMToken.
	*/
    width = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    height = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    
    DgpImage_new(dgp_state, DGPIMAGE_BINARY | DGPIMAGE_DONTCLEAR, width, height, (void**) &im);
    if (im == (DgpImage_Binary)NULL) {
		return((void *)NULL);
	}
	
    for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			/* Can't use DgpImage_pbmGetToken - may not be separating whitespace */
			DGPIMG_REF(im, x, y) = (Dgp_Int8) DgpImage_pbmGetPBMToken(inFile, &failed,
				comment, &commentsize, dgp_state);
			if (failed) {
				DgpImage_free (im);
				return((void *)NULL);
			}
		}
	}
	
	if (commentsize > 0) {
		DGP_VOID_ASSERT(comment[commentsize - 1] == '\n');
		comment[commentsize] = '\0';
		DgpImage_setComment ((void*) im, comment);
	}

    *readimHeader = &im->header;
    return(im);
}

static void *
readGrayCooked(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader)
{
    Dgp_Int32 width, height;
    Dgp_Int32 maxval;
    Dgp_Bool failed;
    DgpImage_Gray im;
    Dgp_Int32 x, y;
	Dgp_Int8 comment[LOAD_COMMENT_MAX];
	Dgp_Int32 commentsize = 0;

	DGP_VOID_PROLOGUE ("readGrayCooked", DGP_PERRSTATE);
	
    /*
	* File format:
	* P2 <width> <height> <maxval>
	* bunches of <value><ws> where 0 <= <value> <= maxval, suitable for
	* reading with DgpImage_pbmGetToken.
	* If maxval is not 255, we'll have to rescale the values; this will
	* be slow.
	*/
    width = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    height = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    maxval = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    
    DgpImage_new(dgp_state, DGPIMAGE_GRAY | DGPIMAGE_DONTCLEAR, width, height, (void**) &im);
    if (im == (DgpImage_Gray)NULL) {
		return((void *)NULL);
	}
	
    if (maxval == DGPIMG_COLRNG - 1) {
		/* Special case the nice case */
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				DGPIMG_REF(im, x, y) = (Dgp_UInt8) DgpImage_pbmGetToken(inFile, &failed,
					comment, &commentsize, dgp_state);
				if (failed) {
					DgpImage_free (im);
					return((void *)NULL);
				}
			}
		}
	}
    else {
		/* Have to rescale. */
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				DGPIMG_REF(im, x, y) = (Dgp_UInt8) (
					(DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state) *
					(DGPIMG_COLRNG - 1)) / maxval);
				if (failed) {
					DgpImage_free (im);
					return((void *)NULL);
				}
			}
		}
	}
	
	if (commentsize > 0) {
		DGP_VOID_ASSERT(comment[commentsize - 1] == '\n');
		comment[commentsize] = '\0';
		DgpImage_setComment ((void*) im, comment);
	}

    *readimHeader = &im->header;
    return(im);
}

static void *
readRGBCooked(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader)
{
    Dgp_Int32 width, height;
    Dgp_Int32 maxval;
    Dgp_Bool failed;
    DgpImage_ColorRGB im;
    Dgp_Int32 x, y;
	Dgp_Int8 comment[LOAD_COMMENT_MAX];
	Dgp_Int32 commentsize = 0;

	DGP_VOID_PROLOGUE ("readRGBCooked", DGP_PERRSTATE);
	

    /*
	* File format:
	* P3 <width> <height> <maxval>
	* bunches of <rvalue><ws><gvalue><ws><bvalue><ws> where
	* 0 <= <rvalue>, <gvalue>, <bvalue> <= maxval, suitable for
	* reading with DgpImage_pbmGetToken.
	* If maxval is not 255, we'll have to rescale the values; this will
	* be slow.
	*/
    width = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    height = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    maxval = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    
    DgpImage_new (dgp_state, DGPIMAGE_RGB | DGPIMAGE_DONTCLEAR, width, height, (void**) &im);
    if (im == (DgpImage_ColorRGB)NULL) {
		return((void *)NULL);
	}
	
    if (maxval == DGPIMG_COLRNG - 1) {
		/* Special case the nice case */
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				DGPIMG_REF(im, x, y).r = (Dgp_UInt8) DgpImage_pbmGetToken(inFile, &failed,
					comment, &commentsize, dgp_state);
				if (failed) {
					DgpImage_free (im);
					return((void *)NULL);
				}
				
				DGPIMG_REF(im, x, y).g = (Dgp_UInt8) DgpImage_pbmGetToken(inFile, &failed,
					comment, &commentsize, dgp_state);
				if (failed) {
					DgpImage_free (im);
					return((void *)NULL);
				}
				
				DGPIMG_REF(im, x, y).b = (Dgp_UInt8) DgpImage_pbmGetToken(inFile, &failed,
					comment, &commentsize, dgp_state);
				if (failed) {
					DgpImage_free (im);
					return((void *)NULL);
				}
			}
		}
	}
    else {
		/* Have to rescale. */
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				DGPIMG_REF(im, x, y).r = (Dgp_UInt8) (
					(DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state) *
					(DGPIMG_COLRNG - 1)) / maxval);
				if (failed) {
					DgpImage_free (im);
					return((void *)NULL);
				}
				
				DGPIMG_REF(im, x, y).g = (Dgp_UInt8) (
					(DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state) *
					(DGPIMG_COLRNG - 1)) / maxval);
				if (failed) {
					DgpImage_free (im);
					return((void *)NULL);
				}
				
				DGPIMG_REF(im, x, y).b = (Dgp_UInt8) (
					(DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state) *
					(DGPIMG_COLRNG - 1)) / maxval);
				if (failed) {
					DgpImage_free (im);
					return((void *)NULL);
				}
			}
		}
	}
	
	if (commentsize > 0) {
		DGP_VOID_ASSERT(comment[commentsize - 1] == '\n');
		comment[commentsize] = '\0';
		DgpImage_setComment ((void*) im, comment);
	}

    *readimHeader = &im->header;
    return(im);
}

static void *
readBinary(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader)
{
    Dgp_Int32 width, height;
    Dgp_Bool failed;
    DgpImage_Binary im;
    Dgp_Int32 x, y;
    Dgp_Int32 c, bitsleft;
    Dgp_UInt8 cc = 0;
	Dgp_Int8 comment[LOAD_COMMENT_MAX];
	Dgp_Int32 commentsize = 0;

	DGP_VOID_PROLOGUE ("readBinary", DGP_PERRSTATE);

    /*
	* File format:
	* P4 <width> <height>
	* binary image, packed, high bit is leftmost; rows begin on byte
	* boundaries.
	*/
    width = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    height = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    
    /* Make sure it's cleared to zero */
    DgpImage_new (dgp_state, DGPIMAGE_BINARY, width, height, (void**) &im);
    if (im == (DgpImage_Binary)NULL) {
		return((void *)NULL);
	}
	
    c = EOF;
    for (y = 0; y < height; y++) {
		bitsleft = 0;
		for (x = 0; x < width; x++) {
			if (bitsleft == 0) {
				c = getc(inFile);
				if (c == EOF) {
					DgpImage_free (im);
					return((void *)NULL);
				}
				cc = (Dgp_UInt8) c;
				bitsleft = 8;
			}
			/* High bit first */
			if (cc & (1 << 7)) {
				DGPIMG_REF(im, x, y) = 1;
			}
			else if (cc == 0) {
				/* It's already zero by default, and we've run out of bits */
				x += bitsleft - 1;
				bitsleft = 0;
				continue;
			}
			bitsleft--;
			cc <<= 1;
		}
	}
	
	if (commentsize > 0) {
		DGP_VOID_ASSERT(comment[commentsize - 1] == '\n');
		comment[commentsize] = '\0';
		DgpImage_setComment ((void*) im, comment);
	}

    *readimHeader = &im->header;
    return(im);
}

static void *
readGray(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader)
{
    Dgp_Int32 width, height;
    Dgp_Int32 maxval;
    Dgp_Bool failed;
    DgpImage_Gray im;
    Dgp_Int32 x, y;
	Dgp_Int8 comment[LOAD_COMMENT_MAX];
	Dgp_Int32 commentsize = 0;

	DGP_VOID_PROLOGUE ("readGray", DGP_PERRSTATE);
	
    /*
	* File format:
	* P5 <width> <height> <maxval>
	* image data in reading order, one byte per pixel.
	* If maxval is not 255, we'll have to rescale the values; this will
	* be slow.
	*/
    width = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    height = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    maxval = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    
    DgpImage_new (dgp_state, DGPIMAGE_GRAY | DGPIMAGE_DONTCLEAR, width, height, (void**) &im);
    if (im == (DgpImage_Gray)NULL) {
		return((void *)NULL);
	}
	
    /* Inhale everything. */
    if (fread((Dgp_Int8 *)(im->data[0]), sizeof(Dgp_UInt8),
		(size_t)(width * height), inFile) != (size_t)(width * height)) {
		DgpImage_free (im);
		return((void *)NULL);
	}
    if (maxval != DGPIMG_COLRNG - 1) {
		/* Have to rescale. */
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				DGPIMG_REF(im, x, y) = (Dgp_UInt8) (
					(DGPIMG_REF(im, x, y) * (DGPIMG_COLRNG - 1)) / maxval);
			}
		}
	}
	
	if (commentsize > 0) {
		DGP_VOID_ASSERT(comment[commentsize - 1] == '\n');
		comment[commentsize] = '\0';
		DgpImage_setComment ((void*) im, comment);
	}

    *readimHeader = &im->header;
    return(im);
}

static void *
readRGB(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader)
{
    Dgp_Int32 width, height;
    Dgp_Int32 maxval;
    Dgp_Bool failed;
    DgpImage_ColorRGB im;
    Dgp_Int32 x, y;
	Dgp_Int8 comment[LOAD_COMMENT_MAX];
	Dgp_Int32 commentsize = 0;

	DGP_VOID_PROLOGUE ("readRGB", DGP_PERRSTATE);
	
    /*
	* File format:
	* P3 <width> <height> <maxval>
	* image data, three bytes per pixel, in RGB order.
	* If maxval is not 255, we'll have to rescale the values; this will
	* be slow.
	*/
    width = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    height = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    maxval = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state);
    if (failed) {
		return((void *)NULL);
	}
    
    DgpImage_new (dgp_state, DGPIMAGE_RGB | DGPIMAGE_DONTCLEAR, width, height, (void**) &im);
    if (im == (DgpImage_ColorRGB)NULL) {
		return((void *)NULL);
	}
	
    /* Inhale everything. */
    if (fread((Dgp_Int8 *)(im->data[0]), sizeof(DgpImage_RGB),
		(size_t)(width * height), inFile) != (size_t)(width * height)) {
		DgpImage_free (im);
		return((void *)NULL);
	}
    if (maxval != DGPIMG_COLRNG - 1) {
		/* Have to rescale. */
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				DGPIMG_REF(im, x, y).r = (Dgp_UInt8) (
					(DGPIMG_REF(im, x, y).r * (DGPIMG_COLRNG - 1)) / maxval);
				DGPIMG_REF(im, x, y).g = (Dgp_UInt8) (
					(DGPIMG_REF(im, x, y).g * (DGPIMG_COLRNG - 1)) / maxval);
				DGPIMG_REF(im, x, y).b = (Dgp_UInt8) (
					(DGPIMG_REF(im, x, y).b * (DGPIMG_COLRNG - 1)) / maxval);
			}
		}
	}
	
	if (commentsize > 0) {
		DGP_VOID_ASSERT(comment[commentsize - 1] == '\n');
		comment[commentsize] = '\0';
		DgpImage_setComment ((void*) im, comment);
	}


    *readimHeader = &im->header;
    return(im);
}

/*
* Here's a macro that expands out to a function to read a simple file.
* Note that it calls over to _im_read*Raw with an endian type of
* DGPIMAGE_BIGENDIAN - that is, it assumes that the file being read was
* stored with all the values in bigendian form.
*/

#define READFUNC(funcname, imtype, imtag, eltype) \
	static void * \
	funcname(Dgp_State *dgp_state, FILE *inFile, DgpImage_Header **readimHeader) \
{ \
    Dgp_Int32 width, height; \
    Dgp_Bool failed; \
    imtype im; \
	Dgp_Int8 comment[LOAD_COMMENT_MAX]; \
	Dgp_Int32 commentsize = 0; \
	DGP_VOID_PROLOGUE (" funcname ", DGP_PERRSTATE); \
    width = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state); \
    if (failed) { \
	return((void *)NULL); \
	} \
    height = DgpImage_pbmGetToken(inFile, &failed, comment, &commentsize, dgp_state); \
    if (failed) { \
	return((void *)NULL); \
	} \
    DgpImage_new (dgp_state, imtag | DGPIMAGE_DONTCLEAR, width, height, (void**) &im); \
    if (im == (imtype)NULL) { \
	return((void *)NULL); \
	} \
    if (!DgpImage_##funcname##Raw(im, width, height, inFile, DGPIMAGE_BIGENDIAN)) { \
	DgpImage_free (im); \
	return((void *)NULL); \
	} \
	if (commentsize > 0) { \
		DGP_VOID_ASSERT(comment[commentsize - 1] == '\n'); \
		comment[commentsize] = '\0'; \
		DgpImage_setComment ((void*) im, comment); \
	} \
    *readimHeader = &im->header; \
    return(im); \
}

READFUNC(readFloat, DgpImage_Float, DGPIMAGE_FLOAT, float)

READFUNC(readDouble, DgpImage_Double, DGPIMAGE_DOUBLE, double)

READFUNC(readLong, DgpImage_Long, DGPIMAGE_LONG, long)

READFUNC(readShort, DgpImage_Short, DGPIMAGE_SHORT, short)

READFUNC(readRGBFloat, DgpImage_ColorRGBFloat, DGPIMAGE_RGBFLOAT, RGBFloat)

#undef	READFUNC
#endif


/*
* Here be dragons.
*
* Well, more like an enormous number of conversion routines... which are
* cooked up out of some fairly horrendous macros.
*/

#define CONVFUNC(funcname, fromimtype, fromimtag, fromptrtype, toimtype, toimtag, toptrtype, assignment) \
	static void * \
	funcname(void *im) \
{ \
    fromimtype fromim; \
    toimtype toim; \
    fromptrtype fromptr; \
    toptrtype toptr; \
    Dgp_Int32 x, y; \
	DgpImage_Gray gimage = (DgpImage_Gray) im;		\
	Dgp_State *dgp_state = gimage->dgp_state;		\
	DGP_VOID_PROLOGUE (" funcname ", DGP_PERRSTATE);		\
    DGP_VOID_ASSERT(im != (void *)NULL); \
    fromim = (fromimtype)im; \
    DgpImage_newOffset(dgp_state, toimtag | DGPIMAGE_DONTCLEAR, \
    	DGPIMG_WIDTH(fromim), DGPIMG_HEIGHT(fromim), \
	    DGPIMG_XBASE(fromim), DGPIMG_YBASE(fromim), (void**) &toim); \
    if (toim == (toimtype)NULL) { \
	    return((void *)NULL); \
	} \
    for (y = DGPIMG_YBASE(fromim); y <= DGPIMG_YMAX(fromim); y++) { \
    	fromptr = DGPIMG_PIXPTR(fromim, DGPIMG_XBASE(fromim), y); \
	    toptr = DGPIMG_PIXPTR(toim, DGPIMG_XBASE(toim), y); \
	for (x = DGPIMG_XBASE(fromim); x <= DGPIMG_XMAX(fromim); x++) { \
	    assignment; \
	    DGPIMG_PTRRIGHT(fromim, fromptr); \
	    DGPIMG_PTRRIGHT(toim, toptr); \
	    } \
	} \
    return((void *)toim); \
}

#define PROMFUNC(funcname, fromimtype, fromimtag, fromptrtype, toimtype, toimtag, toptrtype, assignment1, assignment2) \
	static void * \
	funcname(void *im) \
{ \
    fromimtype fromim; \
    toimtype toim; \
    fromptrtype fromptr; \
    toptrtype toptr; \
    Dgp_Float64 slope; \
    Dgp_Float64 offset; \
    Dgp_Int32 x, y; \
	DgpImage_Gray gimage = (DgpImage_Gray) im;		\
	Dgp_State *dgp_state = gimage->dgp_state;		\
	DGP_VOID_PROLOGUE (" funcname ", DGP_PERRSTATE);		\
    DGP_VOID_ASSERT(im != (void *)NULL); \
    fromim = (fromimtype)im; \
    DgpImage_newOffset(dgp_state, toimtag | DGPIMAGE_DONTCLEAR, \
    	DGPIMG_WIDTH(fromim), DGPIMG_HEIGHT(fromim), \
	    DGPIMG_XBASE(fromim), DGPIMG_YBASE(fromim), (void**) &toim); \
    if (toim == (toimtype)NULL) { \
    	return((void *)NULL); \
    } \
    slope = convertTable[fromimtag][toimtag].slope; \
    offset = convertTable[fromimtag][toimtag].offset; \
    if ((slope == 1.) && (offset == 0.)) { \
	    /* Special case - straight assignment */ \
    	for (y = DGPIMG_YBASE(fromim); y <= DGPIMG_YMAX(fromim); y++) { \
	        fromptr = DGPIMG_PIXPTR(fromim, DGPIMG_XBASE(fromim), y); \
        	toptr = DGPIMG_PIXPTR(toim, DGPIMG_XBASE(toim), y); \
        	for (x = DGPIMG_XBASE(fromim); x <= DGPIMG_XMAX(fromim); x++) { \
            	assignment1; \
            	DGPIMG_PTRRIGHT(fromim, fromptr); \
            	DGPIMG_PTRRIGHT(toim, toptr); \
        	} \
       	} \
	} \
    else { \
    	for (y = DGPIMG_YBASE(fromim); y <= DGPIMG_YMAX(fromim); y++) { \
        	fromptr = DGPIMG_PIXPTR(fromim, DGPIMG_XBASE(fromim), y); \
        	toptr = DGPIMG_PIXPTR(toim, DGPIMG_XBASE(toim), y); \
            for (x = DGPIMG_XBASE(fromim); x <= DGPIMG_XMAX(fromim); x++) { \
            	assignment2; \
            	DGPIMG_PTRRIGHT(fromim, fromptr); \
            	DGPIMG_PTRRIGHT(toim, toptr); \
        	} \
       	} \
    } \
    return((void *)toim); \
} 

/* A standard promotion function */
/* Added toDgpType argument to dodge compiler warning. / dhtesq. */
#define	STDPROM(cfrom, lfrom, ufrom, cto, lto, uto, toDgpType) \
    PROMFUNC(p##cfrom##2##cto, DgpImage_##lfrom, DGPIMAGE_##ufrom, DgpImage_##lfrom##PixPtr, \
	DgpImage_##lto, DGPIMAGE_##uto, DgpImage_##lto##PixPtr, \
	DGPIMG_PTRREF(toim, toptr) = DGPIMG_PTRREF(fromim, fromptr), \
DGPIMG_PTRREF(toim, toptr) = DGPIMG_PTRREF(fromim, fromptr) * (toDgpType)slope + (toDgpType)offset)

/* A standard promote-to-RGB function */
/* Added toDgpType argument to dodge compiler warning. / dhtesq. */
#define	STDPROMRGB(cfrom, lfrom, ufrom, cto, lto, uto, toDgpType) \
    PROMFUNC(p##cfrom##2##cto, DgpImage_##lfrom, DGPIMAGE_##ufrom, DgpImage_##lfrom##PixPtr, \
	DgpImage_##lto, DGPIMAGE_##uto, DgpImage_##lto##PixPtr, \
	DGPIMG_PTRREF(toim, toptr).r = DGPIMG_PTRREF(toim, toptr).g = DGPIMG_PTRREF(toim, toptr).b = DGPIMG_PTRREF(fromim, fromptr), \
DGPIMG_PTRREF(toim, toptr).r = DGPIMG_PTRREF(toim, toptr).g = DGPIMG_PTRREF(toim, toptr).b = DGPIMG_PTRREF(fromim, fromptr) * (toDgpType)slope + (toDgpType)offset)

/* A standard promote-from-binary function */
/* Added toDgpType argument to dodge compiler warning. / dhtesq. */
#define	STDPROMBIN(cto, lto, uto, toDgpType) \
    PROMFUNC(pB2##cto, DgpImage_Binary, DGPIMAGE_BINARY, DgpImage_BinaryPixPtr, \
	DgpImage_##lto, DGPIMAGE_##uto, DgpImage_##lto##PixPtr, \
	DGPIMG_PTRREF(toim, toptr) = DGPIMG_PTRREF(fromim, fromptr), \
DGPIMG_PTRREF(toim, toptr) = (DGPIMG_PTRREF(fromim, fromptr) ? (toDgpType)slope : (toDgpType)0) + (toDgpType)offset)

/* A standard demotion function: straight assignment. */
/* Added toDgpType argument to dodge compiler warning. / dhtesq. */
#define	STDDEM(cfrom, lfrom, ufrom, cto, lto, uto, toDgpType) \
    CONVFUNC(d##cfrom##2##cto, DgpImage_##lfrom, DGPIMAGE_##ufrom, DgpImage_##lfrom##PixPtr, \
	DgpImage_##lto, DGPIMAGE_##uto, DgpImage_##lto##PixPtr, \
DGPIMG_PTRREF(toim, toptr) = (toDgpType)DGPIMG_PTRREF(fromim, fromptr))

/* A standard demote-to-RGB function: straight assignment */
#define	STDDEMTORGB(cfrom, lfrom, ufrom, cto, lto, uto) \
    CONVFUNC(d##cfrom##2##cto, DgpImage_##lfrom, DGPIMAGE_##ufrom, DgpImage_##lfrom##PixPtr, \
	DgpImage_##lto, DGPIMAGE_##uto, DgpImage_##lto##PixPtr, \
DGPIMG_PTRREF(toim, toptr).r = DGPIMG_PTRREF(toim, toptr).g = DGPIMG_PTRREF(toim, toptr).b = (Dgp_UInt8)DGPIMG_PTRREF(fromim, fromptr))

/* A standard demote-from-RGB function: combine then assign */
/* Added toDgpType argument to dodge compiler warning. / dhtesq. */
#define	STDDEMFROMRGB(cfrom, lfrom, ufrom, cto, lto, uto, toDgpType) \
    CONVFUNC(d##cfrom##2##cto, DgpImage_##lfrom, DGPIMAGE_ufrom##, DgpImage_##lfrom##PixPtr, \
	DgpImage_##lto, DGPIMAGE_##uto, DgpImage_##lto##PixPtr, \
DGPIMG_PTRREF(toim, toptr) = (toDgpType)(.299 * DGPIMG_PTRREF(fromim, fromptr).r + .587 * DGPIMG_PTRREF(fromim, fromptr).g + .114 * DGPIMG_PTRREF(fromim, fromptr).b))

/* A demote-to-binary function: result 1 iff original value is 0 */
#define	DEMBINEQ(cfrom, lfrom, ufrom) \
    CONVFUNC(d##cfrom##2B, DgpImage_##lfrom, DGPIMAGE_##ufrom, DgpImage_##lfrom##PixPtr, \
	DgpImage_Binary, DGPIMAGE_BINARY, DgpImage_BinaryPixPtr, \
DGPIMG_PTRREF(toim, toptr) = (DGPIMG_PTRREF(fromim, fromptr) == 0))

/* A demote-to-binary function: result 1 iff original value is != 0 */
#define	DEMBINNE(cfrom, lfrom, ufrom) \
    CONVFUNC(d##cfrom##2B, DgpImage_##lfrom, DGPIMAGE_##ufrom, DgpImage_##lfrom##PixPtr, \
	DgpImage_Binary, DGPIMAGE_BINARY, DgpImage_BinaryPixPtr, \
DGPIMG_PTRREF(toim, toptr) = (DGPIMG_PTRREF(fromim, fromptr) != 0))

/*  Added seventh argument:  dgp type for destination image. */
STDPROM(G, Gray, GRAY, F, Float, FLOAT, Dgp_Float32)
STDPROMRGB(G, Gray, GRAY, R, ColorRGB, RGB, Dgp_UInt8)
STDPROM(G, Gray, GRAY, D, Double, DOUBLE, Dgp_Float64)
DEMBINEQ(G, Gray, GRAY)
STDPROM(G, Gray, GRAY, L, Long, LONG, Dgp_Int32)
STDPROM(G, Gray, GRAY, S, Short, SHORT, Dgp_Int16)
STDPROMRGB(G, Gray, GRAY, RF, ColorRGBFloat, RGBFLOAT, Dgp_Float32)

STDDEM(F, Float, FLOAT, G, Gray, GRAY, Dgp_UInt8)
STDDEMTORGB(F, Float, FLOAT, R, ColorRGB, RGB)
STDPROM(F, Float, FLOAT, D, Double, DOUBLE, Dgp_Float64)
DEMBINNE(F, Float, FLOAT)
STDDEM(F, Float, FLOAT, L, Long, LONG, Dgp_Int32)
STDDEM(F, Float, FLOAT, S, Short, SHORT, Dgp_Int16)
STDPROMRGB(F, Float, FLOAT, RF, ColorRGBFloat, RGBFLOAT, Dgp_Float32)

STDDEMFROMRGB(R, ColorRGB, RGB, G, Gray, GRAY, Dgp_UInt8)
STDDEMFROMRGB(R, ColorRGB, RGB, F, Float, FLOAT, Dgp_Float32)
STDDEMFROMRGB(R, ColorRGB, RGB, D, Double, DOUBLE, Dgp_Float64)
CONVFUNC(dR2B, DgpImage_ColorRGB, DGPIMAGE_RGB, DgpImage_ColorRGBPixPtr,
		 DgpImage_Binary, DGPIMAGE_BINARY, DgpImage_BinaryPixPtr,
		 DGPIMG_PTRREF(toim, toptr) = ((DGPIMG_PTRREF(fromim, fromptr).r == 0) && (DGPIMG_PTRREF(fromim, fromptr).g == 0) && (DGPIMG_PTRREF(fromim, fromptr).b == 0)))
STDDEMFROMRGB(R, ColorRGB, RGB, L, Long, LONG, Dgp_Int32)
STDDEMFROMRGB(R, ColorRGB, RGB, S, Short, SHORT, Dgp_Int16)
CONVFUNC(pR2RF, DgpImage_ColorRGB, DGPIMAGE_RGB, DgpImage_ColorRGBPixPtr,
DgpImage_ColorRGBFloat, DGPIMAGE_RGBFLOAT, DgpImage_ColorRGBFloatPixPtr,
(DGPIMG_PTRREF(toim, toptr).r = DGPIMG_PTRREF(fromim, fromptr).r,
DGPIMG_PTRREF(toim, toptr).g = DGPIMG_PTRREF(fromim, fromptr).g,
DGPIMG_PTRREF(toim, toptr).b = DGPIMG_PTRREF(fromim, fromptr).b))

STDDEM(D, Double, DOUBLE, G, Gray, GRAY, Dgp_UInt8)
STDDEM(D, Double, DOUBLE, F, Float, FLOAT, Dgp_Float32)
STDDEMTORGB(D, Double, DOUBLE, R, ColorRGB, RGB)
DEMBINNE(D, Double, DOUBLE)
STDDEM(D, Double, DOUBLE, L, Long, LONG, Dgp_Int32)
STDDEM(D, Double, DOUBLE, S, Short, SHORT, Dgp_Int16)
STDDEMTORGB(D, Double, DOUBLE, RF, ColorRGBFloat, RGBFLOAT)

STDPROMBIN(G, Gray, GRAY, Dgp_UInt8)
STDPROMBIN(F, Float, FLOAT, Dgp_Float32)
PROMFUNC(pB2R, DgpImage_Binary, DGPIMAGE_BINARY, DgpImage_BinaryPixPtr,
DgpImage_ColorRGB, DGPIMAGE_RGB, DgpImage_ColorRGBPixPtr,
DGPIMG_PTRREF(toim, toptr).r = DGPIMG_PTRREF(toim, toptr).g = DGPIMG_PTRREF(toim, toptr).b = DGPIMG_PTRREF(fromim, fromptr),
DGPIMG_PTRREF(toim, toptr).r = DGPIMG_PTRREF(toim, toptr).g = DGPIMG_PTRREF(toim, toptr).b = (DGPIMG_PTRREF(fromim, fromptr) ? (Dgp_UInt8)slope : (Dgp_UInt8)0) + (Dgp_UInt8)offset)
STDPROMBIN(D, Double, DOUBLE, Dgp_Float64)
STDPROMBIN(L, Long, LONG, Dgp_Int32)
STDPROMBIN(S, Short, SHORT, Dgp_Int16)
PROMFUNC(pB2RF, DgpImage_Binary, DGPIMAGE_BINARY, DgpImage_BinaryPixPtr,
DgpImage_ColorRGBFloat, DGPIMAGE_RGBFLOAT, DgpImage_ColorRGBFloatPixPtr,
DGPIMG_PTRREF(toim, toptr).r = DGPIMG_PTRREF(toim, toptr).g = DGPIMG_PTRREF(toim, toptr).b = DGPIMG_PTRREF(fromim, fromptr),
DGPIMG_PTRREF(toim, toptr).r = DGPIMG_PTRREF(toim, toptr).g = DGPIMG_PTRREF(toim, toptr).b = (DGPIMG_PTRREF(fromim, fromptr) ? (Dgp_Float32)slope : (Dgp_Float32)0) + (Dgp_Float32)offset)

STDDEM(L, Long, LONG, G, Gray, GRAY, Dgp_UInt8)
STDDEM(L, Long, LONG, F, Float, FLOAT, Dgp_Float32)
STDDEMTORGB(L, Long, LONG, R, ColorRGB, RGB)
STDPROM(L, Long, LONG, D, Double, DOUBLE, Dgp_Float64)
DEMBINEQ(L, Long, LONG)
STDDEM(L, Long, LONG, S, Short, SHORT, Dgp_Int16)
STDDEMTORGB(L, Long, LONG, RF, ColorRGBFloat, RGBFLOAT)

STDDEM(S, Short, SHORT, G, Gray, GRAY, Dgp_UInt8)
STDPROM(S, Short, SHORT, F, Float, FLOAT, Dgp_Float32)
STDDEMTORGB(S, Short, SHORT, R, ColorRGB, RGB)
STDPROM(S, Short, SHORT, D, Double, DOUBLE, Dgp_Float64)
DEMBINEQ(S, Short, SHORT)
STDPROM(S, Short, SHORT, L, Long, LONG, Dgp_Int32)
STDPROMRGB(S, Short, SHORT, RF, ColorRGBFloat, RGBFLOAT, Dgp_Float32)

STDDEMFROMRGB(RF, ColorRGBFloat, RGBFLOAT, G, Gray, GRAY, Dgp_UInt8)
STDDEMFROMRGB(RF, ColorRGBFloat, RGBFLOAT, F, Float, FLOAT, Dgp_Float32)
CONVFUNC(dRF2R, 
         DgpImage_ColorRGBFloat, 
         DGPIMAGE_RGBFLOAT, 
         DgpImage_ColorRGBFloatPixPtr,
         DgpImage_ColorRGB,
         DGPIMAGE_RGB,
         DgpImage_ColorRGBPixPtr,
         (DGPIMG_PTRREF(toim, toptr).r = (Dgp_UInt8)DGPIMG_PTRREF(fromim, fromptr).r,
         DGPIMG_PTRREF(toim, toptr).g = (Dgp_UInt8)DGPIMG_PTRREF(fromim, fromptr).g,
         DGPIMG_PTRREF(toim, toptr).b = (Dgp_UInt8)DGPIMG_PTRREF(fromim, fromptr).b))
STDDEMFROMRGB(RF, ColorRGBFloat, RGBFLOAT, D, Double, DOUBLE, Dgp_Float64)
CONVFUNC(dRF2B, DgpImage_ColorRGBFloat, DGPIMAGE_RGBFLOAT, DgpImage_ColorRGBFloatPixPtr,
DgpImage_Binary, DGPIMAGE_BINARY, DgpImage_BinaryPixPtr,
DGPIMG_PTRREF(toim, toptr) = ((DGPIMG_PTRREF(fromim, fromptr).r == 0) && (DGPIMG_PTRREF(fromim, fromptr).g == 0) && (DGPIMG_PTRREF(fromim, fromptr).b == 0)))
STDDEMFROMRGB(RF, ColorRGBFloat, RGBFLOAT, L, Long, LONG, Dgp_Int32)
STDDEMFROMRGB(RF, ColorRGBFloat, RGBFLOAT, S, Short, SHORT, Dgp_Int16)



