/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libimage/image.h,v 1.13 2003/02/05 22:21:08 mkrocken Exp $
 *
 * Copyright (c) 1994 Xerox Corporation.  All Rights Reserved.
 *  
 */

#ifndef	DgpImage_H
#define	DgpImage_H

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libimage/image.h,v 1.13 2003/02/05 22:21:08 mkrocken Exp $
 *
 * Copyright (c) 1990, 1991, 1992, 1993 Cornell University.  All Rights
 * Reserved.
 *
 * Copyright (c) 1991, 1992 Xerox Corporation.  All Rights Reserved.
 *  
 * Use, reproduction, preparation of derivative works, and distribution
 * of this software is permitted.  Any copy of this software or of any
 * derivative work must include both the above copyright notices of
 * Cornell University and Xerox Corporation and this paragraph.  Any
 * distribution of this software or derivative works must comply with all
 * applicable United States export control laws.  This software is made
 * available AS IS, and XEROX CORPORATION DISCLAIMS ALL WARRANTIES,
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE,
 * AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
 * LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
 * EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
 * NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGES.
 */
/*
 * Header file for image maintenance routines
 */

/* For FILE */
#include <stdio.h>
/* For Dgp_Bool */
#include "util.h"
#include "dgp_types.h"
/* For Error Codes */
#include "dgp_state.h"

#ifdef __cplusplus
extern "C"
{
#endif

#ifndef	IMAGE_OLD_RGB

typedef struct _DgpImage_Point{
    Dgp_Int32 x;
    Dgp_Int32 y;
    } DgpImage_Point;

typedef struct _DgpImage_RGB{
    Dgp_UInt8 r, g, b;
    } DgpImage_RGB;

/* 25.8.2000 / dhtesq. */
/* We don't support a cmyk image type, but for purposes of color conversion */
/* this structure is useful. */
typedef struct _DgpImage_CMYK {
    Dgp_UInt8 c, m, y, k;
} DgpImage_CMYK;

typedef struct _DgpImage_RGBFloat{
    Dgp_Float32 r, g, b;
    } DgpImage_RGBFloat;

#else

typedef Dgp_UInt8 DgpImage_RGB[3];

#endif

typedef Dgp_UInt16 DgpImage_Type;

/* The image types */
#define	DGPIMAGE_GRAY	((DgpImage_Type)1)
#define	DGPIMAGE_FLOAT	((DgpImage_Type)2)
#define	DGPIMAGE_RGB	((DgpImage_Type)3)
#define	DGPIMAGE_DOUBLE	((DgpImage_Type)4)
#define	DGPIMAGE_BINARY	((DgpImage_Type)5)
#define	DGPIMAGE_LONG	((DgpImage_Type)6)
#define	DGPIMAGE_PTR	((DgpImage_Type)7)
#define	DGPIMAGE_SHORT	((DgpImage_Type)8)
#define	DGPIMAGE_RGBFLOAT	((DgpImage_Type)9)
#define	DGPIMAGE_PACKED	((DgpImage_Type)100)

#define	DGPIMAGE_BIGENDIAN		((DgpImage_Type)256)
#define	DGPIMAGE_LITTLEENDIAN	((DgpImage_Type)512)

#define	DGPIMAGE_DONTCLEAR	((DgpImage_Type)128)

/* An invisible image header type */
typedef struct _DgpImage_Header {
    Dgp_Int32 width, height, row_width;
    Dgp_Int32 xbase, ybase;
    DgpImage_Type tag;
    Dgp_Int8 *comment;
} DgpImage_Header;

/* Several types of image */
typedef struct _DgpImage_Packed{
    Dgp_UInt32 **data;
    struct _DgpImage_Header header;
	Dgp_State *dgp_state;
    } *DgpImage_Packed;

typedef struct _DgpImage_Gray{
    Dgp_UInt8 **data;
    struct _DgpImage_Header header;
	Dgp_State *dgp_state;
    } *DgpImage_Gray;
typedef Dgp_UInt8 *	DgpImage_GrayPixPtr;

typedef struct _DgpImage_floatPixPtr{
    Dgp_Float32 **data;
    struct _DgpImage_Header header;
	Dgp_State *dgp_state;
    } *DgpImage_Float;
typedef Dgp_Float32 *	DgpImage_FloatPixPtr;

typedef struct _DgpImage_ColorRGB{
    DgpImage_RGB **data;
    struct _DgpImage_Header header;
	Dgp_State *dgp_state;
    } *DgpImage_ColorRGB;
typedef DgpImage_RGB * DgpImage_ColorRGBPixPtr;

typedef struct _DgpImage_Double{
    Dgp_Float64 **data;
    struct _DgpImage_Header header;
	Dgp_State *dgp_state;
    } *DgpImage_Double;
typedef Dgp_Float64 * DgpImage_DoublePixPtr;

typedef struct _DgpImageBinary{
    Dgp_Int8 **data;
    struct _DgpImage_Header header;
	Dgp_State *dgp_state;
    } *DgpImage_Binary;
typedef Dgp_Int8 *	DgpImage_BinaryPixPtr;

typedef struct _DgpImage_Long{
    Dgp_Int32 **data;
    struct _DgpImage_Header header;
	Dgp_State *dgp_state;
    } *DgpImage_Long;
typedef Dgp_Int32 *	DgpImage_LongPixPtr;

typedef struct _DgpImage_Ptr{
    void ***data;
    struct _DgpImage_Header header;
	Dgp_State *dgp_state;
    } *DgpImage_Ptr;
typedef void **	DgpImage_PtrPixPtr;

typedef struct _DgpImage_ShortPixPtr{
    Dgp_Int16 **data;
    struct _DgpImage_Header header;
	Dgp_State *dgp_state;
    } *DgpImage_Short;
typedef Dgp_Int16* DgpImage_ShortPixPtr;

typedef struct _DgpImage_ColorRGBFloat{
    DgpImage_RGBFloat **data;
    struct _DgpImage_Header header;
	Dgp_State *dgp_state;
    } *DgpImage_ColorRGBFloat;
typedef DgpImage_RGBFloat * DgpImage_ColorRGBFloatPixPtr;



/* declare our exported functions */
DGP_LIBEXPORT ErrCode
DgpImage_newAligned (Dgp_State *dpg_state, DgpImage_Type type, Dgp_Int32 width, Dgp_Int32 height, void** image);

DGP_LIBEXPORT ErrCode
DgpImage_new (Dgp_State *dpg_state, DgpImage_Type type, Dgp_Int32 width, Dgp_Int32 height, void** image);

DGP_LIBEXPORT ErrCode
DgpImage_newOffset(Dgp_State *dpg_state, DgpImage_Type type, Dgp_Int32 width, Dgp_Int32 height, 
					Dgp_Int32 xbase, Dgp_Int32 ybase, void** image);

DGP_LIBEXPORT ErrCode
DgpImage_free (void *im);

DGP_LIBEXPORT ErrCode 
DgpImage_setOffset (void *im, Dgp_Int32 newxbase, Dgp_Int32 newybase);

DGP_LIBEXPORT ErrCode
DgpImage_load (Dgp_State *dpg_state, DgpImage_Type type, Dgp_Int8 *filename, void** image);

DGP_LIBEXPORT ErrCode
DgpImage_save (void *im, Dgp_Int8 *filename);

DGP_LIBEXPORT ErrCode
DgpImage_pLoadF (Dgp_State *dpg_state, DgpImage_Type type, FILE *inFile, void** image);

DGP_LIBEXPORT ErrCode
DgpImage_pSaveF (void *im, FILE *outFile);

DGP_LIBEXPORT ErrCode
DgpImage_pSaveHeaderF (void *im, FILE *outFile);

DGP_LIBEXPORT ErrCode
DgpImage_dup (void *im, Dgp_Bool copycomment, void** image);

DGP_LIBEXPORT ErrCode
DgpImage_convert (void *im, DgpImage_Type convertto, void** image);

DGP_LIBEXPORT ErrCode 
DgpImage_wrap (Dgp_State *dgp_state, Dgp_UInt32 *data, Dgp_Int32 width, Dgp_Int32 height,
			      Dgp_Int32 longsperline, DgpImage_Packed *image);

DGP_LIBEXPORT ErrCode 
DgpImage_wrapAll (Dgp_State *dgp_state, Dgp_UInt32 *data, DgpImage_Type type, Dgp_Int32 width, Dgp_Int32 height, Dgp_Int32 bpl, void **image);

DGP_LIBEXPORT ErrCode
DgpImage_unwrap (DgpImage_Packed im, Dgp_UInt32 **retval);

DGP_LIBEXPORT ErrCode
DgpImage_unwrapAll (void *im, Dgp_UInt32 **retval);

/* 11.1.2001 / dhtesq. */
/* Added these two color wrap/unwrap functions for java-based viewer. */
DGP_LIBEXPORT ErrCode 
DgpImage_wrapRGB (Dgp_State *dpg_state, Dgp_UInt32 *data, Dgp_Int32 width, Dgp_Int32 height,
			      Dgp_Int32 bpl, DgpImage_ColorRGB *image);

DGP_LIBEXPORT ErrCode
DgpImage_unwrapRGB (DgpImage_ColorRGB im, Dgp_UInt32 **retval);

DGP_LIBEXPORT ErrCode 
DgpImage_read_pts_packed (FILE *file, Dgp_Int32 *xsize, Dgp_Int32 *ysize,
			      Dgp_Int32 *npts, DgpImage_Packed *pointim, DgpImage_Point **retpoint, Dgp_State *dgp_state);

DGP_LIBEXPORT ErrCode 
DgpImage_pts_to_pim (Dgp_Int32 width, Dgp_Int32 height,
			      Dgp_Int32 xbase, Dgp_Int32 ybase,
			      Dgp_Int32 npts, DgpImage_Point *pts, DgpImage_Packed *pim, Dgp_State *dgp_state);

DGP_LIBEXPORT ErrCode
DgpImage_getHeader (void *im, DgpImage_Header **header);

DGP_LIBEXPORT ErrCode
DgpImage_setComment (void *im, Dgp_Int8 *newcomment);

DGP_LIBEXPORT ErrCode 
DgpImage_read_pts (FILE *file, Dgp_Int32 *xsize, Dgp_Int32 *ysize,
		       Dgp_Int32 *npts, DgpImage_Binary *pointim, DgpImage_Point **retpoint, Dgp_State *dgp_state);

DGP_LIBEXPORT ErrCode 
DgpImage_pts_to_bim (Dgp_Int32 width, Dgp_Int32 height, Dgp_Int32 xbase, Dgp_Int32 ybase,
			    Dgp_Int32 npts, DgpImage_Point *pts, DgpImage_Binary *bim, Dgp_State *dgp_state);


/* this should go somewhere else. For now we keep it here */
DGP_LIBEXPORT ErrCode 
DgpImage_shuffle_pts (DgpImage_Point *pts, Dgp_Int32 npts, Dgp_State *dgp_state);


/*
 * These macros are strange - they attempt to make imFree etc somewhat
 * polymorphic by figuring out, in a context where the type is known (at
 * compile-time), the type of the image and passing that in. If it was
 * not done this way, imFree etc would not be able to figure out what
 * sort of Image they had been passed (since the structures may not match).
 */
#define	DGPIMAGE_NEW(s, t, w, h, r)			DgpImage_new((s), (t), (w), (h), (r))
#define	DGPIMAGE_NEWOFFSET(s, t, w, h, xb, yb, r)	DgpImage_newOffset((s), (t), (w), (h), (xb), (yb), (r))
#define	DGPIMAGE_FREE(i)					DgpImage_free((void *)(i))
#define	DGPIMAGE_SETOFFSET(i, xb, yb)		DgpImage_setOffset((void *)(i), (xb), (yb))
#define	DGPIMAGE_LOAD(s, t, f, r)			DgpImage_load((s), (t), (f), (r))
/* #define	DGPIMAGE_LOADF(s, t, f, r)			DgpImage_pLoadF((s), (t), (f), (r)) */
#define	DGPIMAGE_SAVE(i, f)					DgpImage_save((void *)(i), (f))
/* #define	DGPIMAGE_SAVEF(i, f)				DgpImage_pSaveF((void *)(i), (i)->header.tag, (f)) */
#define	DGPIMAGE_DUP(i, r)					DgpImage_dup((void *)(i), DGP_TRUE, (r))
/*#define	DGPIMAGE_PROMOTE(i, t, r)			DgpImage_pPromote((void *)(i), (i)->header.tag, (t), (r))
#define	DGPIMAGE_DEMOTE(i, t, r)			DgpImage_pDemote((void *)(i), (i)->header.tag, (t), (r)) */
#define	DGPIMAGE_CONVERT(i, t, r)			DgpImage_convert((void *)(i), (t), (r))
#define	DGPIMAGE_SETCOMMENT(i, c)			DgpImage_setComment((void *)(i), (c))


#ifdef __cplusplus
}
#endif

#endif
