/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libimage/image_new.c,v 1.1 2000/03/04 00:20:51 mkrocken Exp $
 *
 * Copyright (c) 1990, 1991, 1992, 1993 Cornell University.  All Rights
 * Reserved.
 *
 * Copyright (c) 1991, 1992 Xerox Corporation.  All Rights Reserved.
 *  
 * Use, reproduction, preparation of derivative works, and distribution
 * of this software is permitted.  Any copy of this software or of any
 * derivative work must include both the above copyright notices of
 * Cornell University and Xerox Corporation and this paragraph.  Any
 * distribution of this software or derivative works must comply with all
 * applicable United States export control laws.  This software is made
 * available AS IS, and XEROX CORPORATION DISCLAIMS ALL WARRANTIES,
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE,
 * AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
 * LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
 * EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
 * NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGES.
 */

static char rcsid[] = "@(#)$Header: /project/icons/cvsroot/digipaper/src/libimage/image_new.c,v 1.1 2000/03/04 00:20:51 mkrocken Exp $";

/*
 *	image.c - Images.
 *
 * An image is just a rectangular array of some type (several types are
 * implemented here).
 *
 * Exports:
 *	type struct { unsigned char r, g, b } RGB
 *
 *	type Dgpimage_Type
 *
 *	type Dgpimage_Gray
 *	type Dgpimage_Float
 *	type Dgpimage_ColorRGB
 *	type Dgpimage_Double
 *	type Dgpimage_Binary
 *	type Dgpimage_Long
 *	type Dgpimage_Ptr
 *	type Dgpimage_Short
 *	type Dgpimage_ColorRGBFloat
 *
 *	type GrayPixPtr
 *	type FloatPixPtr
 *	type RGBPixPtr
 *	type DoublePixPtr
 *	type Dgpimage_BinaryPixPtr
 *	type LongPixPtr
 *	type PtrPixPtr
 *	type ShortPixPtr
 *	type RGBFloatPixPtr
 *
 *	constant Dgpimage_Type DGPIMAGE_GRAY
 *	constant Dgpimage_Type DGPIMAGE_FLOAT
 *	constant Dgpimage_Type DGPIMAGE_RGB
 *	constant Dgpimage_Type DGPIMAGE_DOUBLE
 *	constant Dgpimage_Type DGPIMAGE_BINARY
 *	constant Dgpimage_Type DGPIMAGE_LONG
 *	constant Dgpimage_Type DGPIMAGE_PTR
 *	constant Dgpimage_Type DGPIMAGE_SHORT
 *	constant Dgpimage_Type DGPIMAGE_RGBFLOAT
 *
 *	constant Dgpimage_Type DGPIMAGE_DONTCLEAR
 *
 *	constant int COLRNG
 *
 *	void *imNew(Dgpimage_Type type, int width, int height) - create
 *		a new image of the appropriate type. The pointer returned
 *		will be a DGPIMAGE_Gray, DGPIMAGE_Float, DGPIMAGE_ColorRGB, DGPIMAGE_Double, 
 *		DGPIMAGE_Binary or WhateverelseImage as appropriate.
 *
 *	void *imNewOffset(Dgpimage_Type type, int width, int height,
 *			  int xbase, int ybase) - create a new image of
 *		the appropriate type whose addressing begins
 *		at (xbase, ybase).
 *		imNew is equivalent to xbase = ybase = 0.
 *
 *	void imFree(void *im) - free an image.
 *
 *	void imSetOffset(void *im, int newxbase, int newybase) - modify
 *		an image so that its addressing begins at
 *		(newxbase, newybase).
 *
 *	void *imLoad(Dgpimage_Type type, char *filename) - load an image from
 *		a file. The pointer returned will be of the appropriate
 *		type. Its base will be (0,0).
 *
 *	void *imLoadF(Dgpimage_Type type, FILE *inFile) - load an image
 *		from a stream. The pointer returned will be of the
 *		appropriate type.
 *
 *	int imSave(void *im, char *filename) - save an image into a file.
 *
 *	int imSaveF(void *im, FILE *outFile) - save an image to a stream.
 *
 *	void *imDup(void *im) - duplicate an image
 *
 *	void *imPromote(void *im, Dgpimage_Type promoteto) - promote an image to
 *		some higher type.
 *
 *	void imSetPromotion(Dgpimage_Type promotefrom, Dgpimage_Type promoteto,
 *			    double slope, double offset) - change the method
 *		of image promotion.
 *
 *	void *imDemote(void *im, Dgpimage_Type demoteto) - demote an image to some
 *		lower type.
 *
 *	void *imConvert(void *im, Dgpimage_Type convertto) - convert (promote
 *		or demote as appropriate) an image to some other type.
 *
 *	int imGetWidth(void *im) - get the width of an image.
 *
 *	int imGetHeight(void *im) - get the height of an image.
 *
 *	int imGetXBase(void *im) - get the x-base of an image.
 *
 *	int imGetYBase(void *im) - get the y-base of an image.
 *
 *	int imGetXMax(void *im) - get the maximum valid X value of an image.
 *
 *	int imGetYMax(void *im) - get the maximum valid Y value of an image.
 *
 *	Dgpimage_Type imGetType(void *im) - get the type of an image.
 *
 *	char *imGetComment(void *im) - get the comment attached to an image.
 *
 *	dgp_boolean imSetComment(void *im, char *comment) - set the comment
 *		attached to an image.
 *
 *	macro imRef(void *im, int x, int y) - access an element
 *		of some form of Image.
 *
 *	macro imGetPixPtr(void *im, int x, int y) - get a handle on a pixel of
 *		some form of Image.
 *
 *	macro imPtrRef(void *im, *PixPtr) - dereference a PixPtr of some type.
 *
 *	macro imPtrDup(void *im, *PixPtr) - duplicate a PixPtr of some type.
 *
 *	void imPtrUp(void *im, *PixPtr) - move a PixPtr up in the image.
 *
 *	void imPtrDown(void *im, *PixPtr) - move a PixPtr down in the image.
 *
 *	void imPtrLeft(void *im, *PixPtr) - move a PixPtr left in the image.
 *
 *	void imPtrRight(void *im, *PixPtr) - move a PixPtr right in the image.
 *
 *	void imPtrUpN(void *im, *PixPtr, int n)
 *	void imPtrDownN(void *im, *PixPtr, int n)
 *	void imPtrRightN(void *im, *PixPtr, int n)
 *	void imPtrLeftN(void *im, *PixPtr, int n)
 *
 *	dgp_boolean imPtrEq(void *im, *PixPtr1, *PixPtr2) - check two PixPtrs
 *		for equality.
 *
 * An image is a rectangular storage area, with each cell being of some
 * type. In this case, the types are unsigned char, float, RGB, double,
 * binary, long, void * and short. Float is large enough for most purposes for
 * image processing; for those for which it is not, DGPIMAGE_Double is provided;
 * the space costs are much higher for this, so it should be avoided unless
 * necessary. Binary images can store 0 or 1 in each location. By convention,
 * 0 is white and 1 is black in binary images.
 * DGPIMAGE_Ptrs can store a void * pointer at each location. These images may
 * not be saved or loaded.
 * COLRNG is the number of discrete gray levels in a DGPIMAGE_Gray, and the number
 * of levels in each component of an DGPIMAGE_ColorRGB. COLRNG-1 is the brightest
 * (most white) DGPIMAGE_Gray value.
 *
 * N.B. Older versions of this datatype had a different definition for RGB,
 * namely:
 *	typedef unsigned char RGB[3];
 * Due to C's weirdness with arrays, this proved troublesome, so the new
 * definition was installed. However, since some code relies on the old
 * definition, a backwards compatibility mode is available. To get it,
 * #define DGPIMAGE_OLD_RGB
 * before including image.h.
 *
 * imNew creates an Image of the given size, of the appropriate type,
 *	returning (void *)NULL on failure. The Image returned
 *	contains all zeros. The return value should be cast to
 *	the appropriate type. If you OR the supplied Dgpimage_Type with
 *	DGPIMAGE_DONTCLEAR, then the image is *not* cleared to zero; this can
 *	be useful if you're going to write to every location and want to
 *	save the time that might be spent clearing it.
 *
 * imNewOffset creates an image of the given size, of the appropriate type,
 *	returning (void *)NULL on failure. The valid range of addresses
 *	for the new image is (xbase..xbase+width-1, ybase..ybase+height-1).
 *	Addressing an image with a shifted base may be more efficient than
 *	performing the base shift for each access.
 *
 * imFree frees the given Image.
 *
 * imSetOffset modifies the given image so that its range of addresses
 *	is (newxbase..newxbase+width-1, newybase..newybase+height-1).
 *	The image data is not changed. The pixel which was previously
 *	addressed as imRef(im, imGetXBase(im), imGetYBase(im)) will now
 *	be addressed as imRef(im, newxbase, newybase). imGetXBase(im) and
 *	imGetYBase(im) will be updated to newxbase and newybase.
 *
 * imLoad attempts to load an Image from a file. It checks to see that
 *	what is stored in the file has the appropriate type, and if not
 *	returns NULL. It then attempts to load the image from the file,
 *	and if successful returns the (.*)Image as appropriate. If it
 *	fails, it returns NULL. The base of the loaded image is (0,0),
 *	regardless of what it was when it was saved. The file is closed in
 *	all cases.
 *	If the file does not contain an image of the appropriate type, but
 *	does contain an image of a type which may be promoted to that type
 *	(for example, if it contains a DGPIMAGE_Binary and imLoad is attempting
 *	to load a DGPIMAGE_Gray), the image will be read and silently promoted
 *	to the requested type (see imPromote). If such a promotion is not
 *	possible, NULL is returned.
 *	Any comments present in the original file are read in and stored
 *	as the comment associated with the image (see imGetComment).
 *	If more than one comment line was present in the file, then the
 *	contents of the lines will be separated by newlines. There is no
 *	trailing newline; thus, the image read from a file with a single
 *	comment line will contain the contents of that line with no extra
 *	newline. There is a maximum limit on the amount of comments which
 *	will be read from a file of 8192 bytes (including newlines).
 *
 * imLoadF attempts to load an Image from a stdio stream. It performs the
 *	same operations as imLoad, except that it assumes the stream is
 *	open and correctly positioned. After loading, the stream is positioned
 *	to just after the image; it is not closed. If the image cannot be
 *	loaded (i.e. if NULL is returned), the position of the stream (and
 *	what has been written to it) is undefined.
 *
 * imSave attempts to write the given Image out to a file. If successful,
 *	it returns 0. If it fails, it returns -1. Files written by imSave
 *	from DGPIMAGE_FLOAT, DGPIMAGE_DOUBLE, DGPIMAGE_LONG, DGPIMAGE_SHORT, and
 *	DGPIMAGE_RGBFLOAT images are stored with each of the values
 *	in big-endian format.  This means that a file written on one
 *	sort of machine should be able to be read on another sort, as
 *	long as both use compatible floating-point formats.  Luckily, this
 *	is often the case.
 *	It also writes out the comment associated with the image
 *	as a comment in the file. The rules for this are:
 *	- Every newline in the comment string causes a new comment line to
 *	  be begun in the file.
 *	- If any line would be longer than 70 characters, it is broken up.
 *	  If possible, it will be broken at a whitespace character.
 *	If imSave fails (returns -1), the contents of the file are
 *	undefined.
 *
 * imSaveF performs the same function as imSave, except that it writes
 *	to a previously-opened stream. After the save, the stream
 *	is positioned after the newly-written data.
 *
 * imDup makes an exact duplicate of an Image - another Image having
 *	the same size, type and offsets, the same image data, and the same
 *	comment. It returns (void *)NULL if it cannot create the duplicate.
 *
 * imPromote converts an image to a "higher" type of image. The arguments are
 *	the image to be converted and the type into which it should be
 *	converted. It returns (void *)NULL if it cannot create the new image.
 *	Only certain types of conversions are possible with this routine.
 *	They are:
 *	DGPIMAGE_Binarys can be promoted to:
 *		DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_Float, DGPIMAGE_Double,
 *		DGPIMAGE_ColorRGB, DGPIMAGE_ColorRGBFloat
 *	DGPIMAGE_Grays can be promoted to:
 *		DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_Float, DGPIMAGE_Double, DGPIMAGE_ColorRGB,
 *		DGPIMAGE_ColorRGBFloat
 *	DGPIMAGE_Shorts can be promoted to:
 *		DGPIMAGE_Long, DGPIMAGE_Float, DGPIMAGE_Double, DGPIMAGE_ColorRGBFloat
 *	DGPIMAGE_Longs can be promoted to:
 *		DGPIMAGE_Double
 *	DGPIMAGE_Floats can be promoted to:
 *		DGPIMAGE_Double, DGPIMAGE_ColorRGBFloat
 *	The rules of the conversions are:
 *	DGPIMAGE_Binary->DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Long:	(1-COLRNG,COLRNG-1)
 *		0 maps to COLRNG - 1
 *		1 maps to 0
 *	DGPIMAGE_Binary->DGPIMAGE_Float, DGPIMAGE_Double:		(1,0)
 *		0 maps to 0
 *		1 maps to 1
 *	DGPIMAGE_Binary->DGPIMAGE_ColorRGB:				(1-COLRNG,COLRNG-1)
 *		0 maps to {COLRNG-1, COLRNG-1, COLRNG-1}
 *		1 maps to {0, 0, 0}
 *	DGPIMAGE_Binary->DGPIMAGE_ColorRGBFloat:			(1-COLRNG,COLRNG-1)
 *		0 maps to {COLRNG-1, COLRNG-1, COLRNG-1}
 *		1 maps to {0, 0, 0}
 *	DGPIMAGE_Gray->DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_Float, DGPIMAGE_Double: (1,0)
 *		a pixel with value v in the original image has value v
 *		in the converted image. (i.e. v maps to v)
 *	DGPIMAGE_Gray->DGPIMAGE_ColorRGB:				(1,0)
 *		v maps to {v, v, v}
 *	DGPIMAGE_Gray->DGPIMAGE_ColorRGBFloat:			(1,0)
 *		v maps to {v, v, v}
 *	DGPIMAGE_Short->DGPIMAGE_Long, DGPIMAGE_Float, DGPIMAGE_Double:	(1,0)
 *		v maps to v
 *	DGPIMAGE_Short->DGPIMAGE_ColorRGBFloat:			(1,0)
 *		v maps to {v, v, v}
 *	DGPIMAGE_Long->DGPIMAGE_Double:				(1,0)
 *		v maps to v
 *	DGPIMAGE_Float->DGPIMAGE_Double:			(1,0)
 *		v maps to v
 *	DGPIMAGE_Float->DGPIMAGE_ColorRGBFloat:			(1,0)
 *		v maps to {v, v, v}
 *	The meaning of the numbers in parentheses are: if a conversion
 *	is labelled (a,b), then a pixel with input value v will be mapped
 *	to av+b. These values are the default ones; they may be changed at
 *	any time with imSetPromotion. Note that these default values are not
 *	transitive: promoting (for example) a DGPIMAGE_Binary to a DGPIMAGE_Gray then
 *	to a DGPIMAGE_Float will give different results than promoting the
 *	DGPIMAGE_Binary directly to a DGPIMAGE_Float.
 *	When converting to an DGPIMAGE_ColorRGB, all three components are set to the
 *	same value.
 *	These conversions are all those which are strictly "promotions": no
 *	information is lost during the conversion.
 *
 * imSetPromotion alters the parameters of one of the promotions. The
 *	"promotefrom" and "promoteto" Dgpimage_Types must correspond to a
 *	valid promotion. The slope and offset parameters will then be used
 *	for all future promotions from images of type "promotefrom" to images
 *	of type "promoteto". This includes the implicit promotions performed
 *	by imLoad and imLoadF.
 *
 * imDemote converts an Image to an Image of lower type. The conversions
 *	which this routine may perform are:
 *	DGPIMAGE_Doubles can be demoted to:
 *		DGPIMAGE_Binary, DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_Float,
 *		DGPIMAGE_ColorRGB, DGPIMAGE_ColorRGBFloat
 *	DGPIMAGE_Floats can be demoted to:
 *		DGPIMAGE_Binary, DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_ColorRGB
 *	DGPIMAGE_Longs can be demoted to:
 *		DGPIMAGE_Binary, DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Float, DGPIMAGE_ColorRGB,
 *		DGPIMAGE_ColorRGBFloat
 *	DGPIMAGE_Shorts can be demoted to:
 *		DGPIMAGE_Binary, DGPIMAGE_Gray, DGPIMAGE_ColorRGB
 *	DGPIMAGE_Grays can be demoted to:
 *		DGPIMAGE_Binary
 *	DGPIMAGE_ColorRGBs can be demoted to:
 *		DGPIMAGE_Binary, DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_Float,
 *		DGPIMAGE_Double
 *	DGPIMAGE_ColorRGBFloats can be demoted to:
 *		DGPIMAGE_Binary, DGPIMAGE_Gray, DGPIMAGE_Short, DGPIMAGE_Long, DGPIMAGE_Float,
 *		DGPIMAGE_Double, DGPIMAGE_ColorRGB
 *	All these conversions have the property that information may be lost.
 *	Note that some conversions may lose information in both directions
 *	(for example DGPIMAGE_Double<->DGPIMAGE_ColorRGB).
 *	The rules of the conversions are:
 *	DGPIMAGE_Double->DGPIMAGE_Float:
 *		v maps to v, if v lies in the range of floats
 *		undefined if v does not
 *	DGPIMAGE_Double->DGPIMAGE_Long, DGPIMAGE_Short, DGPIMAGE_Gray:
 *		v maps to floor(v) if floor(v) is representable in the
 *		destination
 *		undefined if floor(v) is not
 *	DGPIMAGE_Double->DGPIMAGE_Binary:
 *		0 maps to 0
 *		non-zero maps to 1
 *	DGPIMAGE_Double->DGPIMAGE_ColorRGB
 *		v maps to {floor(v), floor(v), floor(v)} if floor(v) is
 *		representable in the destination
 *		undefined if floor(v) is not
 *	DGPIMAGE_Double->DGPIMAGE_ColorRGBFloat
 *		v maps to {v, v, v} if v lies in the range of floats
 *		undefined if v does not
 *	DGPIMAGE_Float->DGPIMAGE_Long, DGPIMAGE_Short, DGPIMAGE_Gray:
 *		v maps to floor(v) if floor(v) is representable in the
 *		destination
 *		undefined if floor(v) is not
 *	DGPIMAGE_Float->DGPIMAGE_Binary:
 *		0 maps to 0
 *		non-zero maps to 1
 *	DGPIMAGE_Float->DGPIMAGE_ColorRGB
 *		v maps to {floor(v), floor(v), floor(v)} if floor(v) is
 *		representable in the destination
 *		undefined if floor(v) is not
 *	DGPIMAGE_Long->DGPIMAGE_Short, DGPIMAGE_Gray
 *		v maps to v if v is representable in the destination
 *		undefined if v is not
 *	DGPIMAGE_Long->DGPIMAGE_Float
 *		v maps to v (note that accuracy may be lost)
 *	DGPIMAGE_Long->DGPIMAGE_Binary
 *		0 maps to 1
 *		non-zero maps to 0
 *	DGPIMAGE_Long->DGPIMAGE_ColorRGB
 *		v maps to {v, v, v} if v is representable in the destination
 *		undefined if v is not
 *	DGPIMAGE_Long->DGPIMAGE_ColorRGBFloat
 *		v maps to {v, v, v} if v is representable in the destination
 *		undefined if v is not
 *	DGPIMAGE_Short->DGPIMAGE_Gray
 *		v maps to v if v is representable in the destination
 *		undefined if v is not
 *	DGPIMAGE_Short->DGPIMAGE_Binary
 *		0 maps to 1
 *		non-zero maps to 0
 *	DGPIMAGE_Short->DGPIMAGE_ColorRGB
 *		v maps to {v, v, v} if v is representable in the destination
 *		undefined if v is not
 *	DGPIMAGE_Gray->DGPIMAGE_Binary
 *		0 maps to 1
 *		non-zero maps to 0
 *	DGPIMAGE_ColorRGB->DGPIMAGE_Binary
 *		If all components are zero, then 1
 *		0 otherwise
 *	DGPIMAGE_ColorRGB->anything else
 *		The R, G and B components are first combined by
 *		v = .299R + .587G + .114B
 *		The resulting value is then assigned to the new image pixel.
 *	DGPIMAGE_ColorRGBFloat->DGPIMAGE_Binary
 *		If all components are zero, then 1
 *		0 otherwise
 *	DGPIMAGE_ColorRGBFloat->DGPIMAGE_ColorRGB
 *		{r, g, b} maps to {floor(r), floor(g), floor(b)}
 *		if all those values are representable.
 *		undefined otherwise
 *	DGPIMAGE_ColorRGBFloat->anything else
 *		The R, G and B components are first combined by
 *		v = floor(.299R + .587G + .114B)
 *		The resulting value is then assigned to the new image pixel.
 *		If the result is not representable in the destination,
 *		the result is undefined.
 *	These conversions are as closely as possibly the inverses of the
 *	conversion routines used by imPromote (i.e. a promotion followed by a
 *	demotion will give the original image back). Note that this makes them
 *	non-transitive, and that there are some conversions which are
 *	demotions in both directions. None of these conversions may be
 *	altered; if another type of conversion is desired, a special routine
 *	may be written.
 *
 * imConvert converts (promotes or demotes) an image to an image of another
 *	type. Any two types of image may be converted between, except for
 *	DGPIMAGE_Ptrs. This routine is a convenience routine which calls
 *	imPromote or imDemote as necessary. As a special case,
 *	converting an image to its original type has the same effect as
 *	imDup EXCEPT that the comment is not copied (no other conversion copies
 *	the comment, so this behaviour is consistent). Note that this form of
 *	conversion is neither a promotion nor a demotion, and so will be
 *	rejected by both imPromote and imDemote.
 *
 * imGetWidth returns the width of the given Image.
 *
 * imGetHeight returns the height of the given Image.
 *
 * imGetXBase returns the X base of the given Image. This is the lowest
 *	valid X value.
 *
 * imGetYBase returns the Y base of the given Image. This is the lowest
 *	valid Y value.
 *
 * imGetXMax returns the highest valid X value of an Image. This is equal to
 *	imGetXBase(im)+imGetWidth(im)-1.
 *
 * imGetYMax returns the highest valid Y value of an Image. This is equal to
 *	imGetYBase(im)+imGetHeight(im)-1.
 *
 * imGetType returns the Dgpimage_Type of the given Image.
 *
 * imGetComment returns the comment associated with the Image, or NULL if
 *	there is none. This is the comment which was present in the file
 *	from which the image was loaded, if any, or a comment associated
 *	with the image later using imSetComment. Note that the memory
 *	pointed at by this pointer is owned by the Image module, and should
 *	not be freed; it will remain valid until the Image is freed, or the
 *	comment is changed.
 *
 * imSetComment assigns the given string as the comment of the given Image.
 *	It returns TRUE if it is successful, FALSE if it is not. It makes
 *	a copy of the string; the original string is not referred to after
 *	this, and may be freed or altered without affecting the Image's
 *	comment. You can pass in NULL as the new comment. If the comment
 *	could not be changed (i.e. FALSE is returned), then the old comment
 *	remains.
 *
 * imRef gives access to an entry in an Image of some form (it works
 *	for all defined forms of Image). It gives an lvalue: it is valid
 *	to say imRef(i, x, y) = foo;
 *
 * The PixPtr types allow efficient access to contiguous regions of an Image.
 * You get a PixPtr by calling imGetPixPtr, giving the location in the image
 * you would like a handle on. You can then dereference this using
 * imPtrRef. imRef(im, x, y) gives the same value as
 * imPtrRef(imGetPixPtr(im, x, y)); note that this is an lvalue and may be
 * assigned to.
 * The advantage of PixPtrs is that they can be efficiently moved through
 * the image array. For example, if im is a DGPIMAGE_Gray:
 * for (x = 0; x < w; x++) {
 *     imRef(im, x, y) = 123;
 *     }
 * can be replaced by:
 * GrayPixPtr pp;
 * pp = imGetPixPtr(im, 0, y);
 * for (x = 0; x < w; x++) {
 *     imPtrRef(im, pp) = 123;
 *     imPtrRight(im, pp);
 *     }
 * which may run more efficiently.
 *
 * imGetPixPtr gets a handle on the (x,y) pixel in the image im, giving a
 *	PixPtr of the appropriate type.
 *
 * imPtrRef dereferences a PixPtr of some type, giving an lvalue of the
 *	appropriate type.
 *
 * imPtrDup duplicates a PixPtr of some type, returning a PixPtr of the same
 *	type, initially referring to the same location in the image.
 *
 * imPtrUp, imPtrDown, imPtrLeft and imPtrRight move a PixPtr through an
 *	Image. imPtrUp effectively decrements the y component of the pixel
 *	to which the pointer refers; imPtrDown increments it. imPtrLeft and
 *	imPtrRight decrement and increment respectively the x component.
 *	imPtr*N do the same thing but change by the specified amount.
 *
 * imPtrEq compares two pointers for equality, returning TRUE if they are,
 *	FALSE if they are not. The pointers must refer to pixels in the
 *	same Image. The use of this allows more efficient image operations.
 *
 * Some of these are implemented partially as macros for strange reasons.
 * As a result, they may not be entirely safety-checked. If you want to
 * have safety-checking in place, at a cost in speed and code size, put
 * -DDGPIMAGE_CHECKALL on your gcc line, or add
 * #define DGPIMAGE_CHECKALL
 * to your code somewhere before you include image.h.
 * This only works if you're compiling with gcc.
 *
 */

#include "util.h"
#include "image.h"
#include "image_defs.h"
#include <ctype.h>
#include <stdio.h>
/*#include "undecl-stdio.h"*/
#include <stdlib.h>
#include <string.h>
/*#include "undecl-string.h"*/

#include "pbmstuff.h"
#include "image-private.h"
#include "rawimage.h"
#include "endianness.h"

static char incomment[LOAD_COMMENT_MAX];
static int incommentsize;

static void *readBinaryCooked(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader);
static void *readGrayCooked(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader);
static void *readRGBCooked(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader);
static void *readBinary(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader);
static void *readGray(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader);
static void *readRGB(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader);
static void *readFloat(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader);
static void *readDouble(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader);
static void *readLong(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader);
static void *readShort(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader);
static void *readRGBFloat(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader);

/*
 * Here begins file magic: list of tags with associated file types and
 * reading functions.
 */
static struct {
    char *filetag;
    Dgpimage_Type filetype;
    void * (*readfunc)(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader);
    } fileReaders[] = {
	{ TAG_BINARY_COOKED, DGPIMAGE_BINARY, readBinaryCooked },
	{ TAG_GRAY_COOKED, DGPIMAGE_GRAY, readGrayCooked },
	{ TAG_RGB_COOKED, DGPIMAGE_RGB, readRGBCooked },
	{ TAG_BINARY, DGPIMAGE_BINARY, readBinary },
	{ TAG_GRAY, DGPIMAGE_GRAY, readGray },
	{ TAG_RGB, DGPIMAGE_RGB, readRGB },
	{ TAG_FLOAT, DGPIMAGE_FLOAT, readFloat },
	{ TAG_DOUBLE, DGPIMAGE_DOUBLE, readDouble },
	{ TAG_LONG, DGPIMAGE_LONG, readLong },
	{ TAG_SHORT, DGPIMAGE_SHORT, readShort },
	{ TAG_RGBFLOAT, DGPIMAGE_RGBFLOAT, readRGBFloat }
	};
#define	NREADERS	((int)(sizeof(fileReaders) / sizeof(fileReaders[0])))

/* Here are all the conversion routines */
static void                                        *pG2F(void *im) , *pG2R(void *im) , *pG2D(void *im) ;
static void *dG2B(void *im) , *pG2L(void *im) , *pG2S(void *im) , *pG2RF(void *im) ;
static void *dF2G(void *im) ,                                        *dF2R(void *im) , *pF2D(void *im) ;
static void *dF2B(void *im) , *dF2L(void *im) , *dF2S(void *im) , *pF2RF(void *im) ;
static void *dR2G(void *im) , *dR2F(void *im) ,                                         *dR2D(void *im) ;
static void *dR2B(void *im) , *dR2L(void *im) , *dR2S(void *im) , *pR2RF(void *im) ;
static void *dD2G(void *im) , *dD2F(void *im) , *dD2R(void *im)                  ;
static void *dD2B(void *im) , *dD2L(void *im) , *dD2S(void *im) , *dD2RF(void *im) ;
static void *pB2G(void *im) , *pB2F(void *im) , *pB2R(void *im) , *pB2D(void *im) ;
static void                  *pB2L(void *im) , *pB2S(void *im) , *pB2RF(void *im) ;
static void *dL2G(void *im) , *dL2F(void *im) , *dL2R(void *im) , *pL2D(void *im) ;
static void *dL2B(void *im) ,                  *dL2S(void *im) , *dL2RF(void *im) ;
static void *dS2G(void *im) , *pS2F(void *im) , *dS2R(void *im) , *pS2D(void *im) ;
static void *dS2B(void *im) , *pS2L(void *im) ,                  *pS2RF(void *im) ;
static void *dRF2G(void *im) , *dRF2F(void *im) , *dRF2R(void *im) , *dRF2D(void *im) ;
static void *dRF2B(void *im) , *dRF2L(void *im) , *dRF2S(void *im)                ;

/*
 * Here's the table of conversions.
 * convertTable[type1][type2] tells how we convert from an image of type1
 * to an image of type2.
 */
static struct {
    dgp_boolean isPromotion;
    double slope;
    double offset;
    void * (*convertFunc)(void *im) ;
    } convertTable[DGPIMAGE_RGBFLOAT + 1][DGPIMAGE_RGBFLOAT + 1] = {
	{	/* Invalid image type */
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL},
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL},
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL},
	{ FALSE, 0, 0, NULL}},

	{	/* DGPIMAGE_Gray */
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL}, { TRUE , 1, 0, pG2F},
	{ TRUE , 1, 0, pG2R}, { TRUE , 1, 0, pG2D}, { FALSE, 0, 0, dG2B},
	{ TRUE , 1, 0, pG2L}, { FALSE, 0, 0, NULL}, { TRUE , 1, 0, pG2S},
	{ TRUE , 1, 0, pG2RF}},

	{	/* DGPIMAGE_Float */
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, dF2G}, { FALSE, 0, 0, NULL},
	{ FALSE, 0, 0, dF2R}, { TRUE , 1, 0, pF2D}, { FALSE, 0, 0, dF2B},
	{ FALSE, 0, 0, dF2L}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, dF2S},
	{ TRUE , 1, 0, pF2RF}},

	{	/* DGPIMAGE_ColorRGB */
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, dR2G}, { FALSE, 0, 0, dR2F},
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, dR2D}, { FALSE, 0, 0, dR2B},
	{ FALSE, 0, 0, dR2L}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, dR2S},
	{ TRUE,  1, 0, pR2RF}},

	{	/* DGPIMAGE_Double */
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, dD2G}, { FALSE, 0, 0, dD2F},
	{ FALSE, 0, 0, dD2R}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, dD2B},
	{ FALSE, 0, 0, dD2L}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, dD2S},
	{ FALSE, 0, 0, dD2RF}},

	{	/* DGPIMAGE_Binary */
	{ FALSE, 0, 0, NULL}, { TRUE , 1-COLRNG, COLRNG-1, pB2G}, { TRUE , 1, 0, pB2F},
	{ TRUE , 1-COLRNG, COLRNG-1, pB2R}, { TRUE , 1, 0, pB2D}, { FALSE, 0, 0, NULL},
	{ TRUE , 1-COLRNG, COLRNG-1, pB2L}, { FALSE, 0, 0, NULL}, { TRUE , 1-COLRNG, COLRNG-1, pB2S},
	{ TRUE , 1-COLRNG, COLRNG-1, pB2RF}},
	
	{	/* DGPIMAGE_Long */
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, dL2G}, { FALSE, 0, 0, dL2F},
	{ FALSE, 0, 0, dL2R}, { TRUE , 1, 0, pL2D}, { FALSE, 0, 0, dL2B},
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, dL2S},
	{ FALSE, 0, 0, dL2RF}},

	{	/* DGPIMAGE_Ptr */
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL},
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL},
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL},
	{ FALSE, 0, 0, NULL}},

	{	/* DGPIMAGE_Short */
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, dS2G}, { TRUE , 1, 0, pS2F},
	{ FALSE, 0, 0, dS2R}, { TRUE , 1, 0, pS2D}, { FALSE, 0, 0, dS2B},
	{ TRUE , 1, 0, pS2L}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, NULL},
	{ TRUE , 1, 0, pS2RF}},

	{	/* DGPIMAGE_ColorRGBFloat */
	{ FALSE, 0, 0, NULL}, { FALSE, 0, 0, dRF2G}, { FALSE, 0, 0, dRF2F},
	{ FALSE, 0, 0, dRF2R}, { FALSE, 0, 0, dRF2D}, { FALSE, 0, 0, dRF2B},
	{ FALSE, 0, 0, dRF2L}, { FALSE, 0, 0, NULL}, { FALSE, 0, 0, dRF2S},
	{ FALSE, 0, 0, NULL}}
	};

/*
 * This routine allocates a rectangular area of width*height*el_size bytes
 * (suitable for holding width*height elements of el_size size), and
 * initialises it to zero (unless dontclear is TRUE). It then allocates and
 * sets up an array of height pointers into this, each one pointing to a
 * row (width*el_size bytes).
 *
 * Each pointer is offset by xbase elements, and the returned pointer is
 * offset by ybase rows; this means that the first element in this array
 * is accessed by array[ybase][xbase]. Note that xbase and ybase may be
 * negative.
 *
 * Note that this isn't really type-safe: it takes this (void **) pointer
 * which will get cast, for example, into a (float **) pointer, which isn't
 * really legit. On machines where all pointers have the same format,
 * it will work. The problem is also that dereferencing a (float **)
 * should get you a (float *), but will actually point to the bit-pattern
 * of a (void *), which could be cast into that (float *) with no problems,
 * but which might not work if it is simply *treated* as a (float *)
 * without the cast. Bah. Doing it right would be too horrendous for words.
 * C just can't handle generics very well...
 */
static void **
allocImage(DgpState *dgp_state, int width, int height, int xbase, int ybase,
	   size_t el_size, dgp_boolean dontclear)
{
    register int i;
    register void *mainArea;
    register void **ptrArea;
    register char *p;
    register void **q;

    /* Allocate the actual image data area */
    if (!dontclear) {
	mainArea = (dgp_Calloc)((size_t)height, width*el_size);
	}
    else {
	mainArea = (dgp_Malloc)(height * width * el_size);
	}
    if (mainArea == (void *)NULL) {
	return((void **)NULL);
	}

    /* Allocate a pointer for each row */
    ptrArea = (void **)(dgp_Malloc)(height * sizeof(void *));
    if (ptrArea == (void **)NULL) {
	(dgp_Free)(mainArea);
	return((void **)NULL);
	}

    /* Set up the row pointers to point to the (offset) rows */
    for (i = 0, q = ptrArea, p = (char *)mainArea;
	 i < height;
	 i++, q++, p += width*el_size) {
	*q = (void *)(p - xbase * (int)el_size);
	}

    /* and return the (offset) row pointer area */
    return(ptrArea - ybase);
    }

/*
 * This routine gets rid of a memory block allocated by allocImage above.
 */
static void
freeImage(DgpState *dgp_state, void **ptr, int xbase, int ybase, size_t el_size)
{
    /* Free the memory block. This should give the first element. */
    (dgp_Free)((void *)((char *)(*(ptr + ybase)) + xbase * (int)el_size));
    /* and the pointer block */
    (dgp_Free)((void *)(ptr + ybase));
    }

/*
 * Root through the image's header to find its type, and get the appropriate
 * base pointer of the data area, and the element size.
 */
static void
getAreaAndSize(void *im, Dgpimage_Header *header,
	       void ***areaPtr, size_t *el_size, DgpState *dgp_state)
{

	DGP_PROLOGUE ("getAreaAndSize", dgp_pErrState);

	DGP_ASSERT(im != (void *)NULL);
    DGP_ASSERT(header != (Dgpimage_Header *)NULL);
    DGP_ASSERT(areaPtr != (void ***)NULL);
    DGP_ASSERT(el_size != (size_t *)NULL);

#define	GET_AREA_SIZE(imtype) \
    { \
	imtype vim = (imtype)im; \
	*areaPtr = (void **)vim->data; \
	*el_size = sizeof(**vim->data); \
      }

    switch(header->tag) {
      case DGPIMAGE_GRAY:		GET_AREA_SIZE(Dgpimage_Gray);	break;
      case DGPIMAGE_FLOAT: 	GET_AREA_SIZE(Dgpimage_Float);	break;
      case DGPIMAGE_RGB:		GET_AREA_SIZE(Dgpimage_ColorRGB);	break;
      case DGPIMAGE_DOUBLE:	GET_AREA_SIZE(Dgpimage_Double);	break;
      case DGPIMAGE_BINARY:	GET_AREA_SIZE(Dgpimage_Binary);	break;
      case DGPIMAGE_LONG:		GET_AREA_SIZE(Dgpimage_Long);	break;
      case DGPIMAGE_PTR:		GET_AREA_SIZE(Dgpimage_Ptr);	break;
      case DGPIMAGE_SHORT:		GET_AREA_SIZE(Dgpimage_Short);	break;
      case DGPIMAGE_RGBFLOAT:	GET_AREA_SIZE(Dgpimage_ColorRGBFloat);	break;
      default: {
	DGP_PANIC("bad image tag");
	}
      }

#undef	GET_AREA_SIZE
    }

/*
 * Given an Image's type, find its header
 */
ErrCode
dgpimage_unp_getHeader(void *im, Dgpimage_Type type, Dgpimage_Header **header)
{
	Dgpimage_Gray gimage = (Dgpimage_Gray) im;
	DgpState *dgp_state = gimage->dgp_state;

	DGP_PROLOGUE ("dgpimage_unp_getHeader", dgp_pErrState);

    DGP_ASSERT(im != (void *)NULL);

#define	GET_HEADER(imtype) \
    { \
	*header = &((imtype)im)->header; \
	}

    switch(type) {
	case DGPIMAGE_GRAY:	GET_HEADER(Dgpimage_Gray);		break;
	case DGPIMAGE_FLOAT:	GET_HEADER(Dgpimage_Float);		break;
	case DGPIMAGE_RGB:		GET_HEADER(Dgpimage_ColorRGB);		break;
	case DGPIMAGE_DOUBLE:	GET_HEADER(Dgpimage_Double);	break;
	case DGPIMAGE_BINARY:	GET_HEADER(Dgpimage_Binary);	break;
	case DGPIMAGE_LONG:	GET_HEADER(Dgpimage_Long);		break;
	case DGPIMAGE_PTR:		GET_HEADER(Dgpimage_Ptr);		break;
	case DGPIMAGE_SHORT:	GET_HEADER(Dgpimage_Short);		break;
	case DGPIMAGE_RGBFLOAT:	GET_HEADER(Dgpimage_ColorRGBFloat);	break;
	case DGPIMAGE_PACKED:	GET_HEADER(Dgpimage_Packed);	break;
	default: {
	    DGP_ERR (cErrInvalidCmdLine);
	    }
	}

#undef	GET_HEADER

    NORMAL_RETURN;
}

/*
 * Take the comment which was accumulated from the input file, copy it
 * and tack it into this header.
 */
static dgp_boolean
setComment(Dgpimage_Header *header, DgpState *dgp_state)
{

	DGP_PROLOGUE ("setComment", dgp_pErrState);
    
	DGP_ASSERT(header != (Dgpimage_Header *)NULL);
		      
    if (incommentsize == 0) {
	/* There was no comment in the input file */
	header->comment = (char *)NULL;
	return(TRUE);
	}

    /* The last thing in the comment should be a newline - clobber it */
    DGP_ASSERT(incomment[incommentsize - 1] == '\n');
    incomment[incommentsize] = '\0';
    header->comment = strdup(incomment);
    if (header->comment == (char *)NULL) {
	return(FALSE);
	}
    return(TRUE);
    }

ErrCode
dgpimage_unp_New(DgpState *dgp_state, Dgpimage_Type type, int width, int height, void** image)
{

	DGP_PROLOGUE ("dgpimage_unp_New", dgp_pErrState);
    
	DGP_ERR_PASS (dgpimage_unp_NewOffset(dgp_state, type, width, height, 0, 0, image));

	NORMAL_RETURN;
}

ErrCode
dgpimage_unp_NewOffset(DgpState *dgp_state, Dgpimage_Type type, int width, int height,
	    int xbase, int ybase, void** image)
{
    void **areaPtr;
    Dgpimage_Header *newHeader;
    dgp_boolean dontclear = FALSE;

	DGP_PROLOGUE ("dgpimage_unp_NewOffset", dgp_pErrState);

    if (type & DGPIMAGE_DONTCLEAR) {
	type &= (~DGPIMAGE_DONTCLEAR);
	dontclear = TRUE;
	}

    /*
     * Allocate the data area for the appropriate type of image, 
     * allocate and initialise the actual *Image and return it.
     * This exists in one form for each type of Image.
     */
#define	ALLOC_INIT(imtype) \
    { \
	imtype newim; \
	DGP_CHECK_ALLOC (newim = (imtype)(dgp_Malloc)(sizeof(* newim))); \
	areaPtr = allocImage(dgp_state, width, height, xbase, ybase, \
			     sizeof(**newim->data), dontclear); \
	if (areaPtr == (void **)NULL) { \
	    (dgp_Free)((void *)newim); \
	    DGP_ERR (cErrInsufficientMem); \
	    } \
	newHeader = &newim->header; \
	newim->data = (void *)areaPtr; /* A better cast is possible */ \
	newim->dgp_state = dgp_state;		\
	*image = (void *)newim; \
	}

    switch(type) {
      case DGPIMAGE_GRAY:		ALLOC_INIT(Dgpimage_Gray);		break;
      case DGPIMAGE_FLOAT:		ALLOC_INIT(Dgpimage_Float);		break;
      case DGPIMAGE_RGB:		ALLOC_INIT(Dgpimage_ColorRGB);		break;
      case DGPIMAGE_DOUBLE:	ALLOC_INIT(Dgpimage_Double);	break;
      case DGPIMAGE_BINARY:	ALLOC_INIT(Dgpimage_Binary);	break;
      case DGPIMAGE_LONG:		ALLOC_INIT(Dgpimage_Long);		break;
      case DGPIMAGE_PTR:		ALLOC_INIT(Dgpimage_Ptr);		break;
      case DGPIMAGE_SHORT:		ALLOC_INIT(Dgpimage_Short);		break;
      case DGPIMAGE_RGBFLOAT:	ALLOC_INIT(Dgpimage_ColorRGBFloat);	break;
      default: {
		DGP_ERR	(cErrInvalidCmdLine);
	}
      }

#undef	ALLOC_INIT

    /* Initialise the header */
    newHeader->width = width;
    newHeader->height = height;
    newHeader->xbase = xbase;
    newHeader->ybase = ybase;
    newHeader->tag = type;
    newHeader->comment = (char *)NULL;

    NORMAL_RETURN;
}

ErrCode
dgpimage_unp_Free_(void *im, Dgpimage_Header *header)
{
    void **areaPtr;
    size_t el_size;
    int xbase, ybase;
	Dgpimage_Gray gimage = (Dgpimage_Gray) im;
	DgpState *dgp_state = gimage->dgp_state;

	DGP_PROLOGUE ("dgpimage_unp_Free_", dgp_pErrState);

    DGP_ASSERT(im != (void *)NULL);
    DGP_ASSERT(header != (Dgpimage_Header *)NULL);

    /* Pull out the data area and the size of the element */
    getAreaAndSize(im, header, &areaPtr, &el_size, dgp_state);

    /* and get rid of all of them */
    xbase = header->xbase;
    ybase = header->ybase;

    if (header->comment != (char *)NULL) {
	(dgp_Free)((void *)header->comment);
	}
    (dgp_Free)((void *)im);
    freeImage (dgp_state, areaPtr, xbase, ybase, el_size);

	NORMAL_RETURN;
}

ErrCode
dgpimage_unp_SetOffset_(void *im, Dgpimage_Header *header, int newxbase, int newybase)
{
    int y;
    int oldxbase;
    int oldybase;
	Dgpimage_Gray gimage = (Dgpimage_Gray) im;
	DgpState *dgp_state = gimage->dgp_state;

	DGP_PROLOGUE ("dgpimage_unp_SetOffset_", dgp_pErrState);

    DGP_ASSERT(im != (void *)NULL);
    DGP_ASSERT(header != (Dgpimage_Header *)NULL);

    oldxbase = header->xbase;
    oldybase = header->ybase;

#define	FIX_PTRS(imtype) \
    { \
	imtype vim = (imtype)(im); \
	for (y = imGetYBase(vim); y <= imGetYMax(vim); y++) { \
	    vim->data[y] += (oldxbase - newxbase); \
	    } \
	vim->data += (oldybase - newybase); \
	}

    switch(header->tag) {
      case DGPIMAGE_GRAY:		FIX_PTRS(Dgpimage_Gray);	break;
      case DGPIMAGE_FLOAT:		FIX_PTRS(Dgpimage_Float);	break;
      case DGPIMAGE_RGB:		FIX_PTRS(Dgpimage_ColorRGB);	break;
      case DGPIMAGE_DOUBLE:	FIX_PTRS(Dgpimage_Double);	break;
      case DGPIMAGE_BINARY:	FIX_PTRS(Dgpimage_Binary);	break;
      case DGPIMAGE_LONG:		FIX_PTRS(Dgpimage_Long);	break;
      case DGPIMAGE_PTR:		FIX_PTRS(Dgpimage_Ptr);	break;
      case DGPIMAGE_SHORT:		FIX_PTRS(Dgpimage_Short);	break;
      case DGPIMAGE_RGBFLOAT:	FIX_PTRS(Dgpimage_ColorRGBFloat);break;
      default: {
		DGP_ERR	(cErrInvalidCmdLine);
	}
      }

#undef	FIX_PTRS

    header->xbase = newxbase;
    header->ybase = newybase;

    NORMAL_RETURN;
}

ErrCode
dgpimage_unp_imLoad(DgpState *dgp_state, Dgpimage_Type type, char *filename, void **image)
{
    FILE *inFile;

	DGP_PROLOGUE ("dgpimage_unp_imLoad", dgp_pErrState);

    /* Open the file, imLoadF on it, then close it */
    DGP_ASSERT(filename != (char *)NULL);
    DGP_CHECK_NULL (inFile = fopen(filename, "rb"));

    DGP_ERR_PASS (dgpimage_unp_LoadF(dgp_state, type, inFile, image));

    if (fclose(inFile) == EOF) {
	/*
	 * It would be a strange day that fclose on a read-only file
	 * gave an error, but...
	 *
	 * Sigh - have to track down the image's header so we can free it.
	 */
	if (*image) {
	    Dgpimage_Header *header;
		dgpimage_unp_getHeader(*image, type, &header);
	    dgpimage_unp_Free_(*image, header);
	    }
	DGP_ERR (cErrFileNotFound);
	}

    NORMAL_RETURN;
}

ErrCode
dgpimage_unp_LoadF(DgpState *dgp_state, Dgpimage_Type type, FILE *inFile, void** image)
{
    char fileHeader[PBM_MAX_LINELEN];
    int c;
    int i;
    Dgpimage_Type filetype = (Dgpimage_Type)0;
    void * (*readfunc)(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader) = NULL;
    void *readim;
    Dgpimage_Header *readimHeader;

	DGP_PROLOGUE ("dgpimage_unp_imLoadF", dgp_pErrState);

    DGP_ASSERT(inFile != (FILE *)NULL);

    /*
     * Read up to the first whitespace in order to find the file's header.
     */
    i = 0;
    do {
	c = getc(inFile);
	if (c == EOF) {
	    DGP_ERR (cErrFileNotFound);
	    }
	fileHeader[i++] = c;
	} while (!isspace(c));
    fileHeader[--i] = '\0';		/* Clobber the whitespace */

    /* Determine who is supposed to handle this sort of file */
    for (i = 0; i < NREADERS; i++) {
	if (strcmp(fileHeader, fileReaders[i].filetag) == 0) {
	    /* Found it */
	    filetype = fileReaders[i].filetype;
	    readfunc = fileReaders[i].readfunc;
	    break;
	    }
	}
    if (i == NREADERS) {
	/* Didn't match any of the ones in the table - unknown type. */
	DGP_ERR (cErrFileNotFound);
	}

    /* Clear the comment buffer */
    incommentsize = 0;

    /* Read in the file */
    readim = (*readfunc)(dgp_state, inFile, &readimHeader);
    if (readim == (void *)NULL) {
	/* Didn't load for whatever reason */
	DGP_ERR (cErrFileNotFound);
	}

    if (filetype == type) {
	/* The file was of the expected type - no conversion required */
	if (!setComment(readimHeader, dgp_state)) {
	    /* Sigh - stumbled over the comment */
	    dgpimage_unp_Free_(readim, readimHeader);
	    DGP_ERR (cErrFileNotFound);;
	    }
	*image = readim;
	NORMAL_RETURN;
	}

    /*
     * OK. We read in a file, and it wasn't what we expected. We get to
     * promote it to the expected image type, if that's defined.
     *
     * If that conversion isn't a promotion, kick out.
     */
    if (!convertTable[filetype][type].isPromotion) {
	dgpimage_unp_Free_(readim, readimHeader);
	DGP_ERR (cErrInvalidCmdLine);
	}
    
    /* Convert... */
	DGP_ERR_CHECK (dgpimage_unp_Promote_(readim, readimHeader, type, image)) {
	/* Didn't work. */
	dgpimage_unp_Free_(readim, readimHeader);
	ERR_RETURN;
	}

    /* Kill the image we read in, and return the promoted one. */
    dgpimage_unp_Free_(readim, readimHeader);

    /* Stick the comment from the file into the header */
    dgpimage_unp_getHeader(*image, type, &readimHeader);
    if (!setComment(readimHeader, dgp_state)) {
	/* Died in the comment copy */
	dgpimage_unp_Free_(*image, readimHeader);
	DGP_ERR (cErrInvalidCmdLine);
	}

    NORMAL_RETURN;
}

ErrCode
dgpimage_unp_Save_(void *im, Dgpimage_Header *header, char *filename)
{
    FILE *outFile;
	Dgpimage_Gray gimage = (Dgpimage_Gray) im;
	DgpState *dgp_state = gimage->dgp_state;
    
	DGP_PROLOGUE ("dgpimage_unp_Save_", dgp_pErrState);

    DGP_ASSERT(filename != (char *)NULL);
    DGP_ASSERT(im != (void *)NULL);
    DGP_ASSERT(header != (Dgpimage_Header *)NULL);

    /* Open the file, imSaveF on it, then close it */
    DGP_CHECK_NULL (outFile = fopen(filename, "wb"));

    /* Write the image */
    DGP_ERR_PASS (dgpimage_unp_SaveF_(im, header, outFile));

    /* and close the file */
    if (fclose(outFile) == EOF) {
	/* Bah. */
	DGP_ERR (cErrFileNotFound);
	}

    NORMAL_RETURN;
}

ErrCode
dgpimage_unp_SaveF_(void *im, Dgpimage_Header *header, FILE *outFile)
{
    char *startString;
    char *startTag;
    dgp_boolean maxvalPresent;
    int i;
    char *p;
    char commentBuf[PBM_MAX_LINELEN];
	Dgpimage_Gray gimage = (Dgpimage_Gray) im;
	DgpState *dgp_state = gimage->dgp_state;

	DGP_PROLOGUE ("dgpimage_unp_SaveF_", dgp_pErrState);

    DGP_ASSERT(outFile != (FILE *)NULL);
    DGP_ASSERT(im != (void *)NULL);
    DGP_ASSERT(header != (Dgpimage_Header *)NULL);

    /*
     * Get a pointer to the data area, the size of each element, and
     * the format of the header. The header may include a dummy entry
     * (some of the p*m formats have a maxval entry in the header - we
     * just use the maximum value we can represent).
     */
#define	GET_SAVE_INFO(imtype, tag, string, maxval) \
    { \
	 startTag = tag; \
	 startString = string; \
	 maxvalPresent = maxval; \
	 }

    switch (header->tag) {
      case DGPIMAGE_GRAY:	GET_SAVE_INFO(Dgpimage_Gray, TAG_GRAY, "%d %d %d\n", TRUE);
			break;
	    
      case DGPIMAGE_FLOAT:	GET_SAVE_INFO(Dgpimage_Float, TAG_FLOAT, "%d %d\n", FALSE);
			break;
	    
      case DGPIMAGE_RGB:	GET_SAVE_INFO(Dgpimage_ColorRGB, TAG_RGB, "%d %d %d\n", TRUE);
			break;

      case DGPIMAGE_DOUBLE:	GET_SAVE_INFO(Dgpimage_Double, TAG_DOUBLE,
					      "%d %d\n", FALSE);
				break;
	    
      case DGPIMAGE_BINARY:	GET_SAVE_INFO(Dgpimage_Binary, TAG_BINARY,
					      "%d %d\n", FALSE);
				break;
	    
      case DGPIMAGE_LONG:	GET_SAVE_INFO(Dgpimage_Long, TAG_LONG, "%d %d\n", FALSE);
			break;

      case DGPIMAGE_SHORT:	GET_SAVE_INFO(Dgpimage_Short, TAG_SHORT, "%d %d\n", FALSE);
			break;

      case DGPIMAGE_RGBFLOAT:	GET_SAVE_INFO(Dgpimage_Short, TAG_RGBFLOAT, "%d %d\n", FALSE);
				break;

      default: {
	DGP_ERR ( cErrInvalidCmdLine );
	}
      }
    
#undef	GET_SAVE_INFO

    /* Write out the tag, and comments if any */
    if (fprintf(outFile, "%s\n", startTag) == EOF) {
	DGP_ERR (cErrFileNotFound);
	}

    if (header->comment != (char *)NULL) {
	p = header->comment;
	while (*p != '\0') {
	    /* Copy a line into commentBuf */
	    for (i = 0;
		 (i < PBM_MAX_LINELEN) && (p[i] != '\n') && (p[i] != '\0');
		 i++) {
		commentBuf[i] = p[i];
		}
	    if (i == PBM_MAX_LINELEN) {
		/* The line was too long - search back for whitespace */
		for (i = PBM_MAX_LINELEN - 1;
		     (i >= 0) && (!isspace(commentBuf[i]));
		     i--) {
		    }
		if (i < 0) {
		    /* There was no whitespace. Break it how we like it... */
		    commentBuf[PBM_MAX_LINELEN - 1] = '\0';
		    i = PBM_MAX_LINELEN - 1;
		    }
		else {
		    /* Break at the whitespace */
		    commentBuf[i] = '\0';
		    i++;
		    }
		}
	    else {
		commentBuf[i] = '\0';
		}
	    /*
	     * At this point, commentBuf contains i characters to be
	     * printed.
	     */
	    if (fprintf(outFile, "#%s\n", commentBuf) == EOF) {
		DGP_ERR (cErrFileNotFound);
		}
	    /* Advance the pointer */
	    p += i;
	    if (*p == '\n') {
		p++;
		}
	    }

	}

    /* Write out the header info */
    if (maxvalPresent) {
	if (fprintf(outFile, startString, header->width, header->height,
		    COLRNG - 1) == EOF) {
	    DGP_ERR (cErrFileNotFound);
	    }
	}
    else {
	if (fprintf(outFile, startString, header->width,
		    header->height) == EOF) {
	    DGP_ERR (cErrFileNotFound);
	    }
	}
	

    /* Write out the data block; force it to be bigendian */
    DGP_ERR_PASS (DGPIMAGE_SaveRawF_(im, header->tag, outFile, DGPIMAGE_BIGENDIAN, dgp_state));

    /* and successfully return */
    NORMAL_RETURN;
}

ErrCode
dgpimage_unp_Dup_(void *im, Dgpimage_Header *header, dgp_boolean copycomment, void** image)
{
    void *dup;
    size_t el_size;
    char *oldbase;
    char *newbase;
    Dgpimage_Header *newheader;
	Dgpimage_Gray gimage = (Dgpimage_Gray) im;
	DgpState *dgp_state = gimage->dgp_state;

	DGP_PROLOGUE ("dgpimage_unp_Dup_", dgp_pErrState);

    DGP_ASSERT(im != (void *)NULL);
    DGP_ASSERT(header != (Dgpimage_Header *)NULL);

    /* Make a copy. */
    DGP_ERR_PASS (dgpimage_unp_NewOffset(dgp_state, header->tag | DGPIMAGE_DONTCLEAR,
		      header->width, header->height,
		      header->xbase, header->ybase, &dup));

#define	GET_DUP_INFO(imtype) \
    { \
	oldbase = \
	    (void *)&(((imtype)im)->data[header->ybase][header->xbase]); \
	newbase = \
	    (void *)&(((imtype)dup)->data[header->ybase][header->xbase]); \
	el_size = sizeof(**((imtype)im)->data); \
	newheader = &((imtype)im)->header; \
	}
    
    switch(header->tag) {
      case DGPIMAGE_GRAY:		GET_DUP_INFO(Dgpimage_Gray);	break;
      case DGPIMAGE_FLOAT:		GET_DUP_INFO(Dgpimage_Float);	break;
      case DGPIMAGE_RGB:		GET_DUP_INFO(Dgpimage_ColorRGB);		break;
      case DGPIMAGE_DOUBLE:	GET_DUP_INFO(Dgpimage_Double);	break;
      case DGPIMAGE_BINARY:	GET_DUP_INFO(Dgpimage_Binary);	break;
      case DGPIMAGE_LONG:		GET_DUP_INFO(Dgpimage_Long);	break;
      case DGPIMAGE_PTR:		GET_DUP_INFO(Dgpimage_Ptr);		break;
      case DGPIMAGE_SHORT:		GET_DUP_INFO(Dgpimage_Short);	break;
      case DGPIMAGE_RGBFLOAT:	GET_DUP_INFO(Dgpimage_ColorRGBFloat);	break;
      default: {
	DGP_ERR (cErrInvalidCmdLine );
	}
      }

#undef	GET_DUP_INFO

    /* Copy the comment, if any, and if we're supposed to */
    if (copycomment && (header->comment != (char *)NULL)) {
	newheader->comment = strdup(header->comment);
	if (newheader->comment == (char *)NULL) {
	    /* Failed to copy */
	    dgpimage_unp_Free_(dup, newheader);
	    DGP_ERR ( cErrInsufficientMem );
	    }
	}

    /* Do the copy */
    (void)memcpy(newbase, oldbase,
		 (el_size * header->width * header->height));

    *image = dup;
	NORMAL_RETURN;
}

ErrCode
dgpimage_unp_Promote_(void *im, Dgpimage_Header *header, Dgpimage_Type promoteto, void **image)
{
    Dgpimage_Type origtype;
	Dgpimage_Gray gimage = (Dgpimage_Gray) im;
	DgpState *dgp_state = gimage->dgp_state;

	DGP_PROLOGUE ("dgpimage_unp_Promote_", dgp_pErrState);

    DGP_ASSERT(im != (void *)NULL);
    DGP_ASSERT(header != (Dgpimage_Header *)NULL);

    origtype = header->tag;
    if (!convertTable[origtype][promoteto].isPromotion) {
	/* They just requested a non-promotion. Bad... */
	DGP_PANIC("non-promotion request made to imPromote()");
	}
    DGP_ASSERT(convertTable[origtype][promoteto].convertFunc !=
	   (void * (*)(void *))NULL);

    *image = (*convertTable[origtype][promoteto].convertFunc)(im);

    NORMAL_RETURN;
}

ErrCode
dgpimage_unp_imSetPromotion(DgpState *dgp_state, Dgpimage_Type promotefrom, Dgpimage_Type promoteto,
	       double slope, double offset)
{

	DGP_PROLOGUE ("dgpimage_unp_imSetPromotion", dgp_pErrState);
    
	if (!convertTable[promotefrom][promoteto].isPromotion) {
	DGP_PANIC("attempt made to imSetPromotion a non-promotion conversion");
	}
    convertTable[promotefrom][promoteto].slope = slope;
    convertTable[promotefrom][promoteto].offset = offset;

	NORMAL_RETURN;
}

ErrCode
dgpimage_unp_Demote_(void *im, Dgpimage_Header *header, Dgpimage_Type demoteto, void **image)
{
    Dgpimage_Type origtype;
	Dgpimage_Gray gimage = (Dgpimage_Gray) im;
	DgpState *dgp_state = gimage->dgp_state;

	DGP_PROLOGUE ("dgpimage_unp_Demote_", dgp_pErrState);

    DGP_ASSERT(im != (void *)NULL);
    DGP_ASSERT(header != (Dgpimage_Header *)NULL);

    origtype = header->tag;
    if (convertTable[origtype][demoteto].isPromotion) {
	/* They just requested a promotion. Bad... */
	DGP_PANIC("promotion request made to imDemote()");
	}
    else if (convertTable[origtype][demoteto].convertFunc == (void * (*)(void *))NULL) {
	DGP_PANIC("impossible request made to imDemote()");
	}

    *image = (*convertTable[origtype][demoteto].convertFunc)(im);

    NORMAL_RETURN;
}

ErrCode
dgpimage_unp_Convert_(void *im, Dgpimage_Header *header, Dgpimage_Type convertto, void **image)
{
    Dgpimage_Type origtype;
	Dgpimage_Gray gimage = (Dgpimage_Gray) im;
	DgpState *dgp_state = gimage->dgp_state;

	DGP_PROLOGUE ("dgpimage_unp_Convert_", dgp_pErrState);

    DGP_ASSERT(im != (void *)NULL);
    DGP_ASSERT(header != (Dgpimage_Header *)NULL);

    origtype = header->tag;
    if (origtype == convertto) {
	/* They really just want a copy */
		DGP_ERR_PASS (dgpimage_unp_Dup_(im, header, FALSE, image));
		NORMAL_RETURN;
	}
    else if (convertTable[origtype][convertto].isPromotion) {
	/* It's a promotion */
		DGP_ERR_PASS (dgpimage_unp_Promote_(im, header, convertto, image));
		NORMAL_RETURN;
	}
    else {
	/* It's a demotion, or illegal - let imDemote handle it */
		DGP_ERR_PASS (dgpimage_unp_Demote_(im, header, convertto, image));
		NORMAL_RETURN;
	}
}

ErrCode
dgpimage_unp_SetComment_(void *im, Dgpimage_Header *header, char *newcomment)
{
    char *newcopy;
	Dgpimage_Gray gimage = (Dgpimage_Gray) im;
	DgpState *dgp_state = gimage->dgp_state;

	DGP_PROLOGUE ("dgpimage_unp_SetComment_", dgp_pErrState);

    /* stop compiler from whining */
    (void)im;

    DGP_ASSERT(im != (void *)NULL);
    DGP_ASSERT(header != (Dgpimage_Header *)NULL);

    if (newcomment != (char *)NULL) {
	newcopy = strdup(newcomment);
	if (newcopy == (char *)NULL) {
	    DGP_ERR (cErrInsufficientMem);
	    }
	}
    else {
	newcopy = (char *)NULL;
	}

    if (header->comment != (char *)NULL) {
	/* Free the old comment */
	(dgp_Free)((void *)header->comment);
	}

    header->comment = newcopy;
    NORMAL_RETURN;
}

/*
 * Here are the actual file-reading routines.
 *
 * Each one is passed a FILE * and an Dgpimage_Header **. It knows that the file
 * stream is positioned immediately after a P?M (or similar) magic number,
 * and implicitly what that magic number was (i.e. what form the file will
 * take). It then reads the file, creates an Image of the appropriate type,
 * and returns it as its return value, putting a pointer to its Dgpimage_Header
 * into *readimHeader.
 */

static void *
readBinaryCooked(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader)
{
    int width, height;
    dgp_boolean failed;
    Dgpimage_Binary im;
    int x, y;

    /*
     * File format:
     * P1 <width> <height>
     * bunches of 0<ws> and 1<ws>, suitable for reading with pbmGetPBMToken.
     */
    width = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    height = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    
    dgpimage_unp_New(dgp_state, DGPIMAGE_BINARY | DGPIMAGE_DONTCLEAR, width, height, (void**) &im);
    if (im == (Dgpimage_Binary)NULL) {
	return((void *)NULL);
	}
	
    for (y = 0; y < height; y++) {
	for (x = 0; x < width; x++) {
	    /* Can't use dgpimage_pbmGetToken - may not be separating whitespace */
	    imRef(im, x, y) = dgpimage_pbmGetPBMToken(inFile, &failed,
					     incomment, &incommentsize, dgp_state);
	    if (failed) {
		dgpimage_unp_Free_(im, &im->header);
		return((void *)NULL);
		}
	    }
	}

    *readimHeader = &im->header;
    return(im);
    }

static void *
readGrayCooked(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader)
{
    int width, height;
    int maxval;
    dgp_boolean failed;
    Dgpimage_Gray im;
    int x, y;

    /*
     * File format:
     * P2 <width> <height> <maxval>
     * bunches of <value><ws> where 0 <= <value> <= maxval, suitable for
     * reading with dgpimage_pbmGetToken.
     * If maxval is not 255, we'll have to rescale the values; this will
     * be slow.
     */
    width = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    height = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    maxval = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    
    dgpimage_unp_New(dgp_state, DGPIMAGE_GRAY | DGPIMAGE_DONTCLEAR, width, height, (void**) &im);
    if (im == (Dgpimage_Gray)NULL) {
	return((void *)NULL);
	}
	
    if (maxval == COLRNG - 1) {
	/* Special case the nice case */
	for (y = 0; y < height; y++) {
	    for (x = 0; x < width; x++) {
		imRef(im, x, y) = dgpimage_pbmGetToken(inFile, &failed,
					      incomment, &incommentsize, dgp_state);
		if (failed) {
		    dgpimage_unp_Free_(im, &im->header);
		    return((void *)NULL);
		    }
		}
	    }
	}
    else {
	/* Have to rescale. */
	for (y = 0; y < height; y++) {
	    for (x = 0; x < width; x++) {
		imRef(im, x, y) =
		    (dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state) *
		     (COLRNG - 1)) / maxval;
		if (failed) {
		    dgpimage_unp_Free_(im, &im->header);
		    return((void *)NULL);
		    }
		}
	    }
	}

    *readimHeader = &im->header;
    return(im);
    }

static void *
readRGBCooked(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader)
{
    int width, height;
    int maxval;
    dgp_boolean failed;
    Dgpimage_ColorRGB im;
    int x, y;

    /*
     * File format:
     * P3 <width> <height> <maxval>
     * bunches of <rvalue><ws><gvalue><ws><bvalue><ws> where
     * 0 <= <rvalue>, <gvalue>, <bvalue> <= maxval, suitable for
     * reading with dgpimage_pbmGetToken.
     * If maxval is not 255, we'll have to rescale the values; this will
     * be slow.
     */
    width = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    height = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    maxval = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    
    dgpimage_unp_New(dgp_state, DGPIMAGE_RGB | DGPIMAGE_DONTCLEAR, width, height, (void**) &im);
    if (im == (Dgpimage_ColorRGB)NULL) {
	return((void *)NULL);
	}
	
    if (maxval == COLRNG - 1) {
	/* Special case the nice case */
	for (y = 0; y < height; y++) {
	    for (x = 0; x < width; x++) {
		imRef(im, x, y).r = dgpimage_pbmGetToken(inFile, &failed,
						incomment, &incommentsize, dgp_state);
		if (failed) {
		    dgpimage_unp_Free_(im, &im->header);
		    return((void *)NULL);
		    }

		imRef(im, x, y).g = dgpimage_pbmGetToken(inFile, &failed,
						incomment, &incommentsize, dgp_state);
		if (failed) {
		    dgpimage_unp_Free_(im, &im->header);
		    return((void *)NULL);
		    }

		imRef(im, x, y).b = dgpimage_pbmGetToken(inFile, &failed,
						incomment, &incommentsize, dgp_state);
		if (failed) {
		    dgpimage_unp_Free_(im, &im->header);
		    return((void *)NULL);
		    }
		}
	    }
	}
    else {
	/* Have to rescale. */
	for (y = 0; y < height; y++) {
	    for (x = 0; x < width; x++) {
		imRef(im, x, y).r =
		    (dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state) *
		     (COLRNG - 1)) / maxval;
		if (failed) {
		    dgpimage_unp_Free_(im, &im->header);
		    return((void *)NULL);
		    }

		imRef(im, x, y).g =
		    (dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state) *
		     (COLRNG - 1)) / maxval;
		if (failed) {
		    dgpimage_unp_Free_(im, &im->header);
		    return((void *)NULL);
		    }

		imRef(im, x, y).b =
		    (dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state) *
		     (COLRNG - 1)) / maxval;
		if (failed) {
		    dgpimage_unp_Free_(im, &im->header);
		    return((void *)NULL);
		    }
		}
	    }
	}

    *readimHeader = &im->header;
    return(im);
    }

static void *
readBinary(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader)
{
    int width, height;
    dgp_boolean failed;
    Dgpimage_Binary im;
    int x, y;
    int c, bitsleft;
    unsigned char cc = 0;

    /*
     * File format:
     * P4 <width> <height>
     * binary image, packed, high bit is leftmost; rows begin on byte
     * boundaries.
     */
    width = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    height = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    
    /* Make sure it's cleared to zero */
    dgpimage_unp_New(dgp_state, DGPIMAGE_BINARY, width, height, (void**) &im);
    if (im == (Dgpimage_Binary)NULL) {
	return((void *)NULL);
	}
	
    c = EOF;
    for (y = 0; y < height; y++) {
	bitsleft = 0;
	for (x = 0; x < width; x++) {
	    if (bitsleft == 0) {
		c = getc(inFile);
		if (c == EOF) {
		    dgpimage_unp_Free_(im, &im->header);
		    return((void *)NULL);
		    }
		cc = c;
		bitsleft = 8;
		}
	    /* High bit first */
	    if (cc & (1 << 7)) {
		imRef(im, x, y) = 1;
		}
	    else if (cc == 0) {
		/* It's already zero by default, and we've run out of bits */
		x += bitsleft - 1;
		bitsleft = 0;
		continue;
		}
	    bitsleft--;
	    cc <<= 1;
	    }
	}

    *readimHeader = &im->header;
    return(im);
    }

static void *
readGray(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader)
{
    int width, height;
    int maxval;
    dgp_boolean failed;
    Dgpimage_Gray im;
    int x, y;

    /*
     * File format:
     * P5 <width> <height> <maxval>
     * image data in reading order, one byte per pixel.
     * If maxval is not 255, we'll have to rescale the values; this will
     * be slow.
     */
    width = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    height = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    maxval = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    
    dgpimage_unp_New(dgp_state, DGPIMAGE_GRAY | DGPIMAGE_DONTCLEAR, width, height, (void**) &im);
    if (im == (Dgpimage_Gray)NULL) {
	return((void *)NULL);
	}
	
    /* Inhale everything. */
    if (fread((char *)(im->data[0]), sizeof(unsigned char),
	      (size_t)(width * height), inFile) != (size_t)(width * height)) {
	dgpimage_unp_Free_(im, &im->header);
	return((void *)NULL);
	}
    if (maxval != COLRNG - 1) {
	/* Have to rescale. */
	for (y = 0; y < height; y++) {
	    for (x = 0; x < width; x++) {
		imRef(im, x, y) =
		    (imRef(im, x, y) * (COLRNG - 1)) / maxval;
		}
	    }
	}

    *readimHeader = &im->header;
    return(im);
    }

static void *
readRGB(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader)
{
    int width, height;
    int maxval;
    dgp_boolean failed;
    Dgpimage_ColorRGB im;
    int x, y;

    /*
     * File format:
     * P3 <width> <height> <maxval>
     * image data, three bytes per pixel, in RGB order.
     * If maxval is not 255, we'll have to rescale the values; this will
     * be slow.
     */
    width = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    height = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    maxval = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state);
    if (failed) {
	return((void *)NULL);
	}
    
    dgpimage_unp_New(dgp_state, DGPIMAGE_RGB | DGPIMAGE_DONTCLEAR, width, height, (void**) &im);
    if (im == (Dgpimage_ColorRGB)NULL) {
	return((void *)NULL);
	}
	
    /* Inhale everything. */
    if (fread((char *)(im->data[0]), sizeof(Dgpimage_RGB),
	      (size_t)(width * height), inFile) != (size_t)(width * height)) {
	dgpimage_unp_Free_(im, &im->header);
	return((void *)NULL);
	}
    if (maxval != COLRNG - 1) {
	/* Have to rescale. */
	for (y = 0; y < height; y++) {
	    for (x = 0; x < width; x++) {
		imRef(im, x, y).r =
		    (imRef(im, x, y).r * (COLRNG - 1)) / maxval;
		imRef(im, x, y).g =
		    (imRef(im, x, y).g * (COLRNG - 1)) / maxval;
		imRef(im, x, y).b =
		    (imRef(im, x, y).b * (COLRNG - 1)) / maxval;
		}
	    }
	}

    *readimHeader = &im->header;
    return(im);
    }

/*
 * Here's a macro that expands out to a function to read a simple file.
 * Note that it calls over to _im_read*Raw with an endian type of
 * DGPIMAGE_BIGENDIAN - that is, it assumes that the file being read was
 * stored with all the values in bigendian form.
 */

#define READFUNC(funcname, imtype, imtag, eltype) \
static void * \
funcname(DgpState *dgp_state, FILE *inFile, Dgpimage_Header **readimHeader) \
{ \
    int width, height; \
    dgp_boolean failed; \
    imtype im; \
    width = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state); \
    if (failed) { \
	return((void *)NULL); \
	} \
    height = dgpimage_pbmGetToken(inFile, &failed, incomment, &incommentsize, dgp_state); \
    if (failed) { \
	return((void *)NULL); \
	} \
    dgpimage_unp_New(dgp_state, imtag | DGPIMAGE_DONTCLEAR, width, height, (void**) &im); \
    if (im == (imtype)NULL) { \
	return((void *)NULL); \
	} \
    if (!dgpimage_##funcname##Raw(im, width, height, inFile, DGPIMAGE_BIGENDIAN)) { \
	dgpimage_unp_Free_(im, &im->header); \
	return((void *)NULL); \
	} \
    *readimHeader = &im->header; \
    return(im); \
    }

READFUNC(readFloat, Dgpimage_Float, DGPIMAGE_FLOAT, float)

READFUNC(readDouble, Dgpimage_Double, DGPIMAGE_DOUBLE, double)

READFUNC(readLong, Dgpimage_Long, DGPIMAGE_LONG, long)

READFUNC(readShort, Dgpimage_Short, DGPIMAGE_SHORT, short)

READFUNC(readRGBFloat, Dgpimage_ColorRGBFloat, DGPIMAGE_RGBFLOAT, RGBFloat)

#undef	READFUNC

/*
 * Here be dragons.
 *
 * Well, more like an enormous number of conversion routines... which are
 * cooked up out of some fairly horrendous macros.
 */

#define CONVFUNC(funcname, fromimtype, fromimtag, fromptrtype, toimtype, toimtag, toptrtype, assignment) \
static void * \
funcname(void *im) \
{ \
    fromimtype fromim; \
    toimtype toim; \
    fromptrtype fromptr; \
    toptrtype toptr; \
    int x, y; \
	Dgpimage_Gray gimage = (Dgpimage_Gray) im;		\
	DgpState *dgp_state = gimage->dgp_state;		\
	DGP_PROLOGUE (" funcname ", dgp_pErrState);		\
    DGP_ASSERT(im != (void *)NULL); \
    fromim = (fromimtype)im; \
    dgpimage_unp_NewOffset(dgp_state, toimtag | DGPIMAGE_DONTCLEAR, \
				 imGetWidth(fromim), imGetHeight(fromim), \
				 imGetXBase(fromim), imGetYBase(fromim), (void**) &toim); \
    if (toim == (toimtype)NULL) { \
	return((void *)NULL); \
	} \
    for (y = imGetYBase(fromim); y <= imGetYMax(fromim); y++) { \
	fromptr = imGetPixPtr(fromim, imGetXBase(fromim), y); \
	toptr = imGetPixPtr(toim, imGetXBase(toim), y); \
	for (x = imGetXBase(fromim); x <= imGetXMax(fromim); x++) { \
	    assignment; \
	    imPtrRight(fromim, fromptr); \
	    imPtrRight(toim, toptr); \
	    } \
	} \
    return((void *)toim); \
    }

#define PROMFUNC(funcname, fromimtype, fromimtag, fromptrtype, toimtype, toimtag, toptrtype, assignment1, assignment2) \
static void * \
funcname(void *im) \
{ \
    fromimtype fromim; \
    toimtype toim; \
    fromptrtype fromptr; \
    toptrtype toptr; \
    double slope; \
    double offset; \
    int x, y; \
	Dgpimage_Gray gimage = (Dgpimage_Gray) im;		\
	DgpState *dgp_state = gimage->dgp_state;		\
	DGP_PROLOGUE (" funcname ", dgp_pErrState);		\
    DGP_ASSERT(im != (void *)NULL); \
    fromim = (fromimtype)im; \
    dgpimage_unp_NewOffset(dgp_state, toimtag | DGPIMAGE_DONTCLEAR, \
				 imGetWidth(fromim), imGetHeight(fromim), \
				 imGetXBase(fromim), imGetYBase(fromim), (void**) &toim); \
    if (toim == (toimtype)NULL) { \
	return((void *)NULL); \
	} \
    slope = convertTable[fromimtag][toimtag].slope; \
    offset = convertTable[fromimtag][toimtag].offset; \
    if ((slope == 1.) && (offset == 0.)) { \
	/* Special case - straight assignment */ \
	for (y = imGetYBase(fromim); y <= imGetYMax(fromim); y++) { \
	    fromptr = imGetPixPtr(fromim, imGetXBase(fromim), y); \
	    toptr = imGetPixPtr(toim, imGetXBase(toim), y); \
	    for (x = imGetXBase(fromim); x <= imGetXMax(fromim); x++) { \
		assignment1; \
		imPtrRight(fromim, fromptr); \
		imPtrRight(toim, toptr); \
		} \
	    } \
	} \
    else { \
	for (y = imGetYBase(fromim); y <= imGetYMax(fromim); y++) { \
	    fromptr = imGetPixPtr(fromim, imGetXBase(fromim), y); \
	    toptr = imGetPixPtr(toim, imGetXBase(toim), y); \
	    for (x = imGetXBase(fromim); x <= imGetXMax(fromim); x++) { \
		assignment2; \
		imPtrRight(fromim, fromptr); \
		imPtrRight(toim, toptr); \
		} \
	    } \
	} \
    return((void *)toim); \
    }

/* A standard promotion function */
#define	STDPROM(cfrom, lfrom, ufrom, cto, lto, uto) \
    PROMFUNC(p##cfrom##2##cto, Dgpimage_##lfrom, DGPIMAGE_##ufrom, Dgpimage_##lfrom##PixPtr, \
	     Dgpimage_##lto, DGPIMAGE_##uto, Dgpimage_##lto##PixPtr, \
	     imPtrRef(toim, toptr) = imPtrRef(fromim, fromptr), \
	     imPtrRef(toim, toptr) = imPtrRef(fromim, fromptr) * slope + offset)
	     
/* A standard promote-to-RGB function */
#define	STDPROMRGB(cfrom, lfrom, ufrom, cto, lto, uto) \
    PROMFUNC(p##cfrom##2##cto, Dgpimage_##lfrom, DGPIMAGE_##ufrom, Dgpimage_##lfrom##PixPtr, \
	     Dgpimage_##lto, DGPIMAGE_##uto, Dgpimage_##lto##PixPtr, \
	     imPtrRef(toim, toptr).r = imPtrRef(toim, toptr).g = imPtrRef(toim, toptr).b = imPtrRef(fromim, fromptr), \
	     imPtrRef(toim, toptr).r = imPtrRef(toim, toptr).g = imPtrRef(toim, toptr).b = imPtrRef(fromim, fromptr) * slope + offset)
	     
/* A standard promote-from-binary function */
#define	STDPROMBIN(cto, lto, uto) \
    PROMFUNC(pB2##cto, Dgpimage_Binary, DGPIMAGE_BINARY, Dgpimage_BinaryPixPtr, \
	     Dgpimage_##lto, DGPIMAGE_##uto, Dgpimage_##lto##PixPtr, \
	     imPtrRef(toim, toptr) = imPtrRef(fromim, fromptr), \
	     imPtrRef(toim, toptr) = (imPtrRef(fromim, fromptr) ? slope : 0) + offset)
	     
/* A standard demotion function: straight assignment */
#define	STDDEM(cfrom, lfrom, ufrom, cto, lto, uto) \
    CONVFUNC(d##cfrom##2##cto, Dgpimage_##lfrom, DGPIMAGE_##ufrom, Dgpimage_##lfrom##PixPtr, \
	     Dgpimage_##lto, DGPIMAGE_##uto, Dgpimage_##lto##PixPtr, \
	     imPtrRef(toim, toptr) = imPtrRef(fromim, fromptr))

/* A standard demote-to-RGB function: straight assignment */
#define	STDDEMTORGB(cfrom, lfrom, ufrom, cto, lto, uto) \
    CONVFUNC(d##cfrom##2##cto, Dgpimage_##lfrom, DGPIMAGE_##ufrom, Dgpimage_##lfrom##PixPtr, \
	     Dgpimage_##lto, DGPIMAGE_##uto, Dgpimage_##lto##PixPtr, \
	     imPtrRef(toim, toptr).r = imPtrRef(toim, toptr).g = imPtrRef(toim, toptr).b = imPtrRef(fromim, fromptr))

/* A standard demote-from-RGB function: combine then assign */
#define	STDDEMFROMRGB(cfrom, lfrom, ufrom, cto, lto, uto) \
    CONVFUNC(d##cfrom##2##cto, Dgpimage_##lfrom, DGPIMAGE_ufrom##, Dgpimage_##lfrom##PixPtr, \
	     Dgpimage_##lto, DGPIMAGE_##uto, Dgpimage_##lto##PixPtr, \
	     imPtrRef(toim, toptr) = .299 * imPtrRef(fromim, fromptr).r + .587 * imPtrRef(fromim, fromptr).g + .114 * imPtrRef(fromim, fromptr).b)

/* A demote-to-binary function: result 1 iff original value is 0 */
#define	DEMBINEQ(cfrom, lfrom, ufrom) \
    CONVFUNC(d##cfrom##2B, Dgpimage_##lfrom, DGPIMAGE_##ufrom, Dgpimage_##lfrom##PixPtr, \
	     Dgpimage_Binary, DGPIMAGE_BINARY, Dgpimage_BinaryPixPtr, \
	     imPtrRef(toim, toptr) = (imPtrRef(fromim, fromptr) == 0))

/* A demote-to-binary function: result 1 iff original value is != 0 */
#define	DEMBINNE(cfrom, lfrom, ufrom) \
    CONVFUNC(d##cfrom##2B, Dgpimage_##lfrom, DGPIMAGE_##ufrom, Dgpimage_##lfrom##PixPtr, \
	     Dgpimage_Binary, DGPIMAGE_BINARY, Dgpimage_BinaryPixPtr, \
	     imPtrRef(toim, toptr) = (imPtrRef(fromim, fromptr) != 0))

STDPROM(G, Gray, GRAY, F, Float, FLOAT)
STDPROMRGB(G, Gray, GRAY, R, ColorRGB, RGB)
STDPROM(G, Gray, GRAY, D, Double, DOUBLE)
DEMBINEQ(G, Gray, GRAY)
STDPROM(G, Gray, GRAY, L, Long, LONG)
STDPROM(G, Gray, GRAY, S, Short, SHORT)
STDPROMRGB(G, Gray, GRAY, RF, ColorRGBFloat, RGBFLOAT)

STDDEM(F, Float, FLOAT, G, Gray, GRAY)
STDDEMTORGB(F, Float, FLOAT, R, ColorRGB, RGB)
STDPROM(F, Float, FLOAT, D, Double, DOUBLE)
DEMBINNE(F, Float, FLOAT)
STDDEM(F, Float, FLOAT, L, Long, LONG)
STDDEM(F, Float, FLOAT, S, Short, SHORT)
STDPROMRGB(F, Float, FLOAT, RF, ColorRGBFloat, RGBFLOAT)

STDDEMFROMRGB(R, ColorRGB, RGB, G, Gray, GRAY)
STDDEMFROMRGB(R, ColorRGB, RGB, F, Float, FLOAT)
STDDEMFROMRGB(R, ColorRGB, RGB, D, Double, DOUBLE)
CONVFUNC(dR2B, Dgpimage_ColorRGB, DGPIMAGE_RGB, Dgpimage_ColorRGBPixPtr,
	 Dgpimage_Binary, DGPIMAGE_BINARY, Dgpimage_BinaryPixPtr,
	 imPtrRef(toim, toptr) = ((imPtrRef(fromim, fromptr).r == 0) && (imPtrRef(fromim, fromptr).g == 0) && (imPtrRef(fromim, fromptr).b == 0)))
STDDEMFROMRGB(R, ColorRGB, RGB, L, Long, LONG)
STDDEMFROMRGB(R, ColorRGB, RGB, S, Short, SHORT)
CONVFUNC(pR2RF, Dgpimage_ColorRGB, DGPIMAGE_RGB, Dgpimage_ColorRGBPixPtr,
	 Dgpimage_ColorRGBFloat, DGPIMAGE_RGBFLOAT, Dgpimage_ColorRGBFloatPixPtr,
	 (imPtrRef(toim, toptr).r = imPtrRef(fromim, fromptr).r,
	  imPtrRef(toim, toptr).g = imPtrRef(fromim, fromptr).g,
	  imPtrRef(toim, toptr).b = imPtrRef(fromim, fromptr).b))

STDDEM(D, Double, DOUBLE, G, Gray, GRAY)
STDDEM(D, Double, DOUBLE, F, Float, FLOAT)
STDDEMTORGB(D, Double, DOUBLE, R, ColorRGB, RGB)
DEMBINNE(D, Double, DOUBLE)
STDDEM(D, Double, DOUBLE, L, Long, LONG)
STDDEM(D, Double, DOUBLE, S, Short, SHORT)
STDDEMTORGB(D, Double, DOUBLE, RF, ColorRGBFloat, RGBFLOAT)

STDPROMBIN(G, Gray, GRAY)
STDPROMBIN(F, Float, FLOAT)
PROMFUNC(pB2R, Dgpimage_Binary, DGPIMAGE_BINARY, Dgpimage_BinaryPixPtr,
	 Dgpimage_ColorRGB, DGPIMAGE_RGB, Dgpimage_ColorRGBPixPtr,
	 imPtrRef(toim, toptr).r = imPtrRef(toim, toptr).g = imPtrRef(toim, toptr).b = imPtrRef(fromim, fromptr),
	 imPtrRef(toim, toptr).r = imPtrRef(toim, toptr).g = imPtrRef(toim, toptr).b = (imPtrRef(fromim, fromptr) ? slope : 0) + offset)
STDPROMBIN(D, Double, DOUBLE)
STDPROMBIN(L, Long, LONG)
STDPROMBIN(S, Short, SHORT)
PROMFUNC(pB2RF, Dgpimage_Binary, DGPIMAGE_BINARY, Dgpimage_BinaryPixPtr,
	 Dgpimage_ColorRGBFloat, DGPIMAGE_RGBFLOAT, Dgpimage_ColorRGBFloatPixPtr,
	 imPtrRef(toim, toptr).r = imPtrRef(toim, toptr).g = imPtrRef(toim, toptr).b = imPtrRef(fromim, fromptr),
	 imPtrRef(toim, toptr).r = imPtrRef(toim, toptr).g = imPtrRef(toim, toptr).b = (imPtrRef(fromim, fromptr) ? slope : 0) + offset)

STDDEM(L, Long, LONG, G, Gray, GRAY)
STDDEM(L, Long, LONG, F, Float, FLOAT)
STDDEMTORGB(L, Long, LONG, R, ColorRGB, RGB)
STDPROM(L, Long, LONG, D, Double, DOUBLE)
DEMBINEQ(L, Long, LONG)
STDDEM(L, Long, LONG, S, Short, SHORT)
STDDEMTORGB(L, Long, LONG, RF, ColorRGBFloat, RGBFLOAT)

STDDEM(S, Short, SHORT, G, Gray, GRAY)
STDPROM(S, Short, SHORT, F, Float, FLOAT)
STDDEMTORGB(S, Short, SHORT, R, ColorRGB, RGB)
STDPROM(S, Short, SHORT, D, Double, DOUBLE)
DEMBINEQ(S, Short, SHORT)
STDPROM(S, Short, SHORT, L, Long, LONG)
STDPROMRGB(S, Short, SHORT, RF, ColorRGBFloat, RGBFLOAT)

STDDEMFROMRGB(RF, ColorRGBFloat, RGBFLOAT, G, Gray, GRAY)
STDDEMFROMRGB(RF, ColorRGBFloat, RGBFLOAT, F, Float, FLOAT)
CONVFUNC(dRF2R, Dgpimage_ColorRGBFloat, DGPIMAGE_RGBFLOAT, Dgpimage_ColorRGBFloatPixPtr,
	 Dgpimage_ColorRGB, DGPIMAGE_RGB, Dgpimage_ColorRGBPixPtr,
	 (imPtrRef(toim, toptr).r = imPtrRef(fromim, fromptr).r,
	  imPtrRef(toim, toptr).g = imPtrRef(fromim, fromptr).g,
	  imPtrRef(toim, toptr).b = imPtrRef(fromim, fromptr).b))
STDDEMFROMRGB(RF, ColorRGBFloat, RGBFLOAT, D, Double, DOUBLE)
CONVFUNC(dRF2B, Dgpimage_ColorRGBFloat, DGPIMAGE_RGBFLOAT, Dgpimage_ColorRGBFloatPixPtr,
	 Dgpimage_Binary, DGPIMAGE_BINARY, Dgpimage_BinaryPixPtr,
	 imPtrRef(toim, toptr) = ((imPtrRef(fromim, fromptr).r == 0) && (imPtrRef(fromim, fromptr).g == 0) && (imPtrRef(fromim, fromptr).b == 0)))
STDDEMFROMRGB(RF, ColorRGBFloat, RGBFLOAT, L, Long, LONG)
STDDEMFROMRGB(RF, ColorRGBFloat, RGBFLOAT, S, Short, SHORT)
