/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libimage/pbmstuff.c,v 1.3 2002/10/15 00:16:49 jlin Exp $
 *
 */

#include "pbmstuff.h"
#include "util.h"
#include <ctype.h>
#include <stdio.h>
/*#include "undecl-stdio.h"*/

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libimage/pbmstuff.c,v 1.3 2002/10/15 00:16:49 jlin Exp $");

/*
 * Get a single positive integer from f. Eat a single character of whitespace
 * following the integer. The integer must be terminated by whitespace or
 * EOF. Nothing else (except for a comment) is allowed. Accumulate comments
 * in incomment, if not NULL.
 */
Dgp_Int32 
DgpImage_pbmGetToken(FILE *f, Dgp_Bool *failed, Dgp_Int8 *incomment, Dgp_Int32 *incommentsize, Dgp_State *dgp_state)
{
    Dgp_Int32 val;
    Dgp_Int32 c;

    DGP_PROLOGUE ("pbmGetToken", DGP_PERRSTATE);
    DGP_STOP_WHINE;

    DGP_ASSERT(failed != (Dgp_Bool *)NULL);

    /* Skip over whitespace and comments */
    while (DGP_TRUE) {
	do {
	    c = getc(f);
	    if (c == EOF) {
		*failed = DGP_TRUE;
		return(0);
		}
	    } while (isspace(c));

	if (c == '#') {
	    /* A comment - skip to the newline */
	    do {
		c = getc(f);
		if (c == EOF) {
		    *failed = DGP_TRUE;
		    return(0);
		    }
		if ((incomment != (Dgp_Int8 *)NULL) &&
		    (*incommentsize < LOAD_COMMENT_MAX - 1)) {
		    incomment[(*incommentsize)++] = ( Dgp_Int8 ) c;
		    }
		} while (c != '\n');
	    }
	else {
	    break;
	    }
	}

    DGP_ASSERT(!isspace(c));
    /* Accumulate the digits */
    val = 0;
    while (isdigit(c)) {
	val = 10 * val + c - '0';
	c = getc(f);
	if (c == EOF) {
	    /* Number terminated by EOF */
	    *failed = DGP_FALSE;
	    return(val);
	    }
	}

    if (!isspace(c)) {
	/* Number terminated by something which isn't whitespace */
	*failed = DGP_TRUE;
	return(0);
	}

    /* Number terminated by whitespace */
    *failed = DGP_FALSE;
    return(val);
    }

/*
 * Get a 1 or 0 from f. Eat whitespace before the character, if any.
 * Nothing else (except for a comment) is allowed. Accumulate comments
 * in incomment.
 */
Dgp_Int32 
DgpImage_pbmGetPBMToken(FILE *f, Dgp_Bool *failed, Dgp_Int8 *incomment, Dgp_Int32 *incommentsize, Dgp_State *dgp_state)
{
    Dgp_Int32 c;

    DGP_PROLOGUE ("pbmGetPBMToken", DGP_PERRSTATE);
    DGP_STOP_WHINE;

    DGP_ASSERT(failed != (Dgp_Bool *)NULL);

    /* Skip over whitespace and comments */
    while (DGP_TRUE) {
	do {
	    c = getc(f);
	    if (c == EOF) {
		*failed = DGP_TRUE;
		return(0);
		}
	    } while (isspace(c));

	if (c == '#') {
	    /* A comment - skip to the newline */
	    do {
		c = getc(f);
		if (c == EOF) {
		    *failed = DGP_TRUE;
		    return(0);
		    }
		if ((incomment != (Dgp_Int8 *)NULL) &&
		    (*incommentsize < LOAD_COMMENT_MAX - 1)) {
		    incomment[(*incommentsize)++] = (Dgp_Int8) c;
		    }
		} while (c != '\n');
	    }
	else {
	    break;
	    }
	}

    DGP_ASSERT(!isspace(c));

    *failed = DGP_FALSE;
    if (c == '0') {
	return(0);
	}
    else if (c == '1') {
	return(1);
	}
    else {
	*failed = DGP_TRUE;
	return(0);
	}
}
