#include "types.pub"
#include "uds.h"
#include "iowrap.h"


Dgp_Int32 TC_CDECL
fileRead(
    Dgp_UInt32 fileID,
    Dgp_Int8 *buffer,
    Dgp_Int32 byteCount
)
{
    FileID *file = (FileID *)fileID;
    int bytes;
    uds *u = (uds *) file->u;
    Dgp_Int32 totalRead = 0;

    /* Do my own buffered read */
    while ((bytes = u->fread(buffer, sizeof(Dgp_Int8), byteCount, file->fp)) > 0) {
	byteCount -= bytes;
	buffer += bytes;
	totalRead += (Dgp_UInt32)bytes;
	if (! byteCount) break;
    }
    file->pos += totalRead;
    if (bytes < 0) {
	return bytes;	/* Error */
    }
    if (byteCount > 0) {
	return totalRead;	/* Error (not all read) put breakpoint here */
    }
    return totalRead;		
}


Dgp_Int32 TC_CDECL
fileWrite(
    Dgp_UInt32 fileID,
    Dgp_Int8 *buffer,
    Dgp_Int32 byteCount
)
{
    FileID *file = (FileID *)fileID;
    int bytes;
    uds *u = file->u;
    Dgp_Int32 totalWritten = 0;

    /* Do my own buffered write */
    while ((bytes = u->fwrite((Dgp_Int8 *)buffer, sizeof(Dgp_Int8), byteCount, file->fp)) > 0) {
	byteCount -= bytes;
	buffer += bytes;
	totalWritten += (Dgp_UInt32)bytes;
	if (! byteCount) break;
    }
    file->pos += totalWritten;
    if (bytes < 0) {
	return bytes;	/* Error */
    }
    return totalWritten;		
}


Dgp_Int32 TC_CDECL
fileSeek(
    Dgp_UInt32 fileID,
    Dgp_Int32 offset
)
{
    FileID *file = (FileID *)fileID;
    long retval;
    uds *u = file->u;

    retval = u->fseek(file->fp, offset, SEEK_SET);

    if (retval != 0) {
	return (Dgp_UInt32)0;
    }
    return file->pos = (Dgp_Int32)offset;
}


Dgp_Int32 TC_CDECL
fileTell(
    Dgp_UInt32 fileID
)
{
    FileID *file = (FileID *)fileID;

    return file->pos;
}


Dgp_Int32 TC_CDECL
fileFileSize(
    Dgp_UInt32 fileID
)
{
    FileID *file = (FileID *)fileID;
    Dgp_Int32 retval, newpos;
    uds *u = file->u;

    retval = u->fseek(file->fp, 0, SEEK_END);
    if (retval != 0) {
	return (Dgp_Int32)0;
    }

    newpos = u->ftell(file->fp);
    return (Dgp_Int32)file->pos = newpos;
}

