
#include "jbig2file.h"
#include "PsgFix.h"

/* create a JBIG2 file and return a pointer to the data structure */
ErrCode
Dgp_createJbig2Stream (Dgp_State *dgp_state, uds *u, 
			      Dgp_CompressedDocument **cDocument)
{
  Dgp_CompressedDocument *cDoc = NULL;
  DGP_PROLOGUE ("Dgp_createJbig2Stream", DGP_PERRSTATE);
  Dgp_Int32 i;
  FILE *fp ;

  /* The magic Jb2 file header: M-C-W JB2 CR LF C-Z LF */
  Dgp_UInt8 jb2header[] = { 0x97, 0x4a, 0x42, 0x32, 0x0d, 0x0a, 0x1a, 0x0a };

  DGP_ASSERT (u != NULL);
  fp = u->fp;
  DGP_ASSERT (fp != NULL);
  DGP_ASSERT (cDocument != NULL);

  /* create structure */
  DGP_CHECK_ALLOC (cDoc = DGP_MALLOC (sizeof (Dgp_CompressedDocument)));
  cDoc->dgp_state = dgp_state;
  cDoc->outfile.u = u;

  /* set default vals */
  cDoc->n_pages = 0;
  cDoc->currentSection = NULL;
  cDoc->n_dict = 0;

  /* Begin a JBIG2 data stream */
  cDoc->outfile.fp = fp;
  cDoc->outfile.pos = 0UL;

	for (i = 0; i < (Dgp_Int32)sizeof(jb2header); i++) {
		putc(jb2header[i], fp);
	}
	/* Write 0x00: random-access organisation, page count known */
	/* Write 0x01: sequential organisation, page count known */
	/* Write 0x02: random-access organisation, page count not known */
	/* Write 0x03: sequential organisation, page count not known */
	putc(0x03, fp);
	
#if 0
	/* Write the number of pages, bigendian */
	putc((numPgs >> 24) & 0xff, fp);
	putc((numPgs >> 16) & 0xff, fp);
	putc((numPgs >> 8) & 0xff, fp);
	putc(numPgs & 0xff, fp);
#endif

  /* create section structure */
  DGP_CHECK_ALLOC (cDoc->currentSection = 
   DGP_MALLOC (sizeof (Dgp_CompressedSection)));

  cDoc->currentSection->dict = NULL;
  cDoc->currentSection->dict_size = 0;
  DGP_ERR_PASS_CLEANUP (DgpAdt_liNew (&(cDoc->currentSection->pages), dgp_state));

  *cDocument = cDoc;
  NORMAL_RETURN;

 ERR_CLEANUP:
  if (cDoc)
    {
      if (cDoc->currentSection) DGP_FREE (cDoc->currentSection);
      DGP_FREE (cDoc);
    }
  ERR_RETURN;

}

/* close the TIFF-FX file and free all remaining memory */
ErrCode
Dgp_closeJbig2Stream (Dgp_CompressedDocument *cDocument)
{
  Dgp_State *dgp_state = cDocument->dgp_state;
  DGP_PROLOGUE ("Dgp_closeJbig2Stream", DGP_PERRSTATE);
  DGP_STOP_WHINE;

  /* close JBIG2 stream */
  fclose(cDocument->outfile.fp);

  /* should confirm that current section is empty */
  DGP_FREE (cDocument->currentSection);
  DGP_FREE (cDocument);

  NORMAL_RETURN;
}

/* write the JBIG2 compressed data of this section to file and free all memory associated with this section */
ErrCode
Dgp_writeSectionJbig2Only (Dgp_CompressedDocument *cDocument, Dgp_Bool clear_dict)
{

  Dgp_State *dgp_state = cDocument->dgp_state;
  DGP_PROLOGUE ("Dgp_writeSection", DGP_PERRSTATE);


  /* ****NEW**** */
  Dgp_CompressedSection *sec = cDocument->currentSection;
  Dgp_CompressedPage *page;
  Dgp_CompressedImage *im;
  DgpAdt_ListNode ln_page;

  DGP_STOP_WHINE;
  clear_dict = clear_dict;

  if (sec->dict_size > 0)
  {
    cDocument->outfile.u->fwrite (sec->dict, sec->dict_size, 1, cDocument->outfile.fp);
    DGP_FREE (sec->dict);
    sec->dict = NULL;
    sec->dict_size = 0; 
  }

  
  DGPADT_FOREACH (ln_page, sec->pages)
  {
    /* write out the page */
    page = (Dgp_CompressedPage*) ln_page->userdata;
    
    if (page->posBlock)
    {
      im = page->posBlock;
      
      
      cDocument->outfile.u->fwrite (im->data, im->data_size, 1, cDocument->outfile.fp);
      
      /* finish up token image */
      DGP_FREE (im->data);
      DGP_FREE (im);
      page->posBlock = NULL;
    }
    Dgp_freePage(cDocument, &page);
  }

  cDocument->n_pages += (Dgp_UInt16) sec->pages->nelt;
  DgpAdt_liClear (sec->pages);

  NORMAL_RETURN;
/*
ERR_CLEANUP:

  ERR_RETURN;
*/
}


