
#include <stdio.h>

#include "dgp_state.h"
#include "dgp_err_control.h"
#include "PsgFix.h"

#include "pdf_util.h"

/* adds an object to the xref table */
ErrCode
addXref(Dgp_Int32       obj, 
        Dgp_Int32       offset, 
        PdfCookie       *pdfCookie,
        Dgp_State       *dgp_state) {
    DGP_PROLOGUE("addXref()", DGP_PERRSTATE);

	if (pdfCookie->xrefPos == pdfCookie->xrefSize) {
		pdfCookie->xrefSize *= 2;
		DGP_CHECK_ALLOC_CLEANUP(pdfCookie->xref = DGP_REALLOC(pdfCookie->xref, pdfCookie->xrefSize * sizeof(xrefElt)));
	}
	pdfCookie->xref[pdfCookie->xrefPos].obj = obj;
	pdfCookie->xref[pdfCookie->xrefPos].offset = offset;
	pdfCookie->xrefPos++;

	NORMAL_RETURN;

ERR_CLEANUP:
	ERR_RETURN;
}

/* init a pdf cookie */
ErrCode
initPdfCookie(PdfCookie *pdfCookie, Dgp_State *dgp_state) {
    DGP_PROLOGUE("initPdfCookie()", DGP_PERRSTATE);
    Dgp_UInt32 XREF_SIZE = 20;

	pdfCookie->offset = 0;

    /* reserve object # 2 for the catalog, # 1 for pages, #3,4 for linearized files */
	pdfCookie->objCount = 3;
    pdfCookie->catalogObj = 2;
    pdfCookie->PagesObj = 1;
    
    DGP_CHECK_ALLOC_CLEANUP(pdfCookie->xref = DGP_MALLOC(sizeof(xrefElt) * XREF_SIZE));
	pdfCookie->xrefPos = 0;
    pdfCookie->DictObj = 0;
	pdfCookie->xrefSize = XREF_SIZE;
    pdfCookie->numPages = 0;
    pdfCookie->pageStart = 0;
    pdfCookie->ProcObj = 0;

    DGP_ERR_PASS (DgpAdt_liNew (&(pdfCookie->pageObjList), dgp_state));
    DGP_ERR_PASS (DgpAdt_liNew (&(pdfCookie->dictList), dgp_state));

    NORMAL_RETURN;

ERR_CLEANUP:
    DGP_FREE(pdfCookie->xref);
	ERR_RETURN;
}

void
cleanupPdfCookie(PdfCookie *pdfCookie, Dgp_State *dgp_state) {	
	if(pdfCookie->xref) DGP_FREE(pdfCookie->xref);
    pdfCookie->xref = NULL;
    DgpAdt_liClear (pdfCookie->pageObjList);
    DgpAdt_liClear (pdfCookie->dictList);
    DgpAdt_liFree (pdfCookie->pageObjList);
    DgpAdt_liFree (pdfCookie->dictList);
}   


/* write pdf prolog */
void
write_pdf_prolog(uds *u, PdfCookie *pdfCookie) {
	
	/* write header */
	pdfCookie->offset += fprintf(u->fp, "\
%%PDF-1.4\n%%%c%c%c%c\n", 250, 251, 252, 253);

    return;
}

/* write pdf catalog and pages */
ErrCode
write_pdf_catalog(PdfCookie     *pdfCookie, 
                  uds           *u,
                  Dgp_State     *dgp_state) 
{
    DGP_PROLOGUE("write_pdf_pages()", DGP_PERRSTATE);
    
    DgpAdt_ListNode ln;

    /* Write page object. */	
    DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->PagesObj, pdfCookie->offset, pdfCookie, dgp_state));

    pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<<\n\
/Type /Pages\n\
/Count %ld\n\
/Kids [ ", pdfCookie->PagesObj, pdfCookie->pageObjList->nelt);

        /* Write one reference for each object page. */
	    DGPADT_FOREACH(ln, pdfCookie->pageObjList) {
		    pdfCookie->offset += fprintf(u->fp, "%ld 0 R ", (Dgp_Int32) ln->userdata);
	    }
	    pdfCookie->offset += fprintf(u->fp, "\
]\n\
>>\n\
endobj\n\n");

     /* Catalog object, references the pages object. */
     DGP_ERR_PASS(addXref(pdfCookie->catalogObj, pdfCookie->offset, pdfCookie,
dgp_state));
     pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<<\n\
/Type /Catalog\n\
/Pages %ld 0 R\n\
>>\n\
endobj\n\n", pdfCookie->catalogObj, pdfCookie->PagesObj);

        NORMAL_RETURN;

ERR_CLEANUP:
        ERR_RETURN;
}


ErrCode
write_jbig2_page(PdfCookie     *pdfCookie,
                   Dgp_State     *dgp_state,
                   uds           *u,
                   Dgp_UInt8     *data,
                   Dgp_UInt32    dataSize,
                   Dgp_UInt32    width,
                   Dgp_UInt32    height,
                   Dgp_UInt32    pageDPI)
{

    DGP_PROLOGUE("write_jbig2_object()", DGP_PERRSTATE);
    Dgp_Int8        iname[64], sizeStr[256];
    Dgp_UInt32      pageObjNum, sizeStrLen;
    Dgp_UInt32      pdfWidth, pdfHeight;
    Dgp_UInt32      pdfDefaultDPI = 72;

    pdfWidth = width * pdfDefaultDPI / pageDPI;
    pdfHeight = height * pdfDefaultDPI / pageDPI;

    /* Page description */
	
    /* name the image */
    sprintf (iname, "IID%ld", pdfCookie->objCount);

    /* save the number of the page object in a variable and in the page list */
    pageObjNum = pdfCookie->objCount;
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd (pdfCookie->pageObjList, (void*) pageObjNum, NULL));

	DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount, pdfCookie->offset, pdfCookie, dgp_state));
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj <<\n\
/Type /Page\n\
/Resources <<\n\
/XObject <<\n\
/%s %ld 0 R\n\
>>\n\
/ProcSet %ld 0 R\n\
>>\n\
/Parent %ld 0 R\n\
/Contents %ld 0 R\n\
/MediaBox [ 0 0 %ld %ld ]\n\
>>\n\
endobj\n\n", pdfCookie->objCount++, iname, pageObjNum + 2, pdfCookie->ProcObj,
    pdfCookie->PagesObj, 
    pageObjNum + 1, pdfWidth, pdfHeight);

    /* Object size */
    sizeStrLen = sprintf (sizeStr, "q %ld 0 0 %ld 0 0 cm /%s Do Q", pdfWidth, pdfHeight, iname);
	DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount , pdfCookie->offset, pdfCookie, dgp_state));
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<<\n\
/Length %ld\n\
>>\n\
stream\n\
%s\n\
endstream\n\
endobj\n\n", pdfCookie->objCount++, sizeStrLen, sizeStr);

    
    /* JBIG2 Position Block as XObject/Image */
    DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount,
                        pdfCookie->offset, pdfCookie, dgp_state));

    pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<<\n\
/Type /XObject\n\
/Subtype /Image\n\
/Width %ld\n\
/Height %ld\n\
/ColorSpace /DeviceGray\n\
/BitsPerComponent 1\n\
/Name /%s\n\
/Filter /JBIG2Decode\n", pdfCookie->objCount++, width, height, iname);
    if (pdfCookie->DictObj > 0) pdfCookie->offset += fprintf (u->fp, "\
/DecodeParms << /JBIG2Globals %ld 0 R >>\n", pdfCookie->DictObj);
    pdfCookie->offset += fprintf (u->fp, "\
/Length %ld\n\
>>\n\
stream\n", dataSize);
    u->fwrite (data, dataSize, 1, u->fp);
      pdfCookie->offset += dataSize;
    pdfCookie->offset += fprintf(u->fp, "\n\
endstream\n\
endobj\n\n");
  
  NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

ErrCode
write_jbig2_global(PdfCookie     *pdfCookie,
                   Dgp_State     *dgp_state,
                   uds           *u,
                   Dgp_UInt8     *dict,
                   Dgp_UInt32    dictSize)
{

    DGP_PROLOGUE("write_jbig2_global()", DGP_PERRSTATE);
    
    if (dictSize > 0)
    {

    /* Dictionary */
	DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount, pdfCookie->offset,
        pdfCookie, dgp_state));
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd (pdfCookie->dictList, (void*) pdfCookie->objCount, NULL));

    pdfCookie->DictObj = pdfCookie->objCount;
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<< /Length %ld\n\
>>\n\
stream\n", pdfCookie->objCount++, dictSize); 
  u->fwrite (dict, dictSize, 1, u->fp);
    pdfCookie->offset += dictSize; 
    pdfCookie->offset += fprintf(u->fp, "\n\
endstream\n\
endobj\n\n");
    }
    else
    {
        pdfCookie->DictObj = 0;
    }

    if (pdfCookie->ProcObj == 0)
    {
    /* Object Type */
	DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount, pdfCookie->offset,
        pdfCookie, dgp_state));
    /* add this object to the list of dicts, because it is a shared object */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd (pdfCookie->dictList, (void*) pdfCookie->objCount, NULL));

    pdfCookie->ProcObj = pdfCookie->objCount;
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
[\n\
/PDF /ImageB\n\
]\n\
endobj\n\n", pdfCookie->objCount++);
    }

  NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/* Sort objects in xref table. */
static int 
sort_by_obj(const void *a, const void *b)
{
  const xrefElt *ea = a;
  const xrefElt *eb = b;

  return (ea->obj - eb->obj);
}


void
write_pdf_end(PdfCookie *pdfCookie, uds *u) 
{
	Dgp_Int32 xrefPos = pdfCookie->offset;
	Dgp_Int32 i, objgap, j;

	/* Sort xref objects. */ 
	qsort(pdfCookie->xref, pdfCookie->xrefPos, sizeof(xrefElt), sort_by_obj);
	/* Write xref table & trailer. */
	pdfCookie->offset += fprintf(u->fp, "\
xref\n\
0 %ld\n\
0000000000 65535 f \n", pdfCookie->xref[pdfCookie->xrefPos-1].obj + 1);


	for (i = 0; i < pdfCookie->xrefPos; i++) {
        if(i > 0){
            objgap = pdfCookie->xref[i].obj - pdfCookie->xref[i-1].obj;
            for(j = 1; j< objgap; j++){
                	pdfCookie->offset += fprintf(u->fp, "0000000000 00000 f \n");
            }

        }
		pdfCookie->offset += fprintf(u->fp, "%010ld 00000 n \n", pdfCookie->xref[i].offset);
        
	}

	pdfCookie->offset += fprintf(u->fp, "\
trailer\n\
<<\n\
/Size %ld\n\
/Root %ld 0 R\n\
>>\n\
startxref\n\
%ld\n\
%%%%EOF\n",  pdfCookie->xref[pdfCookie->xrefPos-1].obj + 1, pdfCookie->catalogObj, xrefPos);

    return;
}

static int startPos (PdfCookie *pdfCookie, int Obj);
static int endPos (PdfCookie *pdfCookie, int Obj);
static void copyFile (size_t start_pos, size_t end_pos,
               char* buffer, int buffer_size, int* objMap,
               FILE* in, FILE* out);
static int* remapObjects (PdfCookie* pdfCookie, Dgp_State *dgp_state);

void
linearize_pdf_file(PdfCookie *pdfCookie, uds *u, Dgp_State *dgp_state) 
{

    FILE*  in = u->fp;
    FILE* out = fopen ("c:\\test.tmp", "w+b");
    char buffer[16384];
    int buf_size = 16384;
    int page, next_page;
    size_t next_pos, size, end_pos;
    int* objectMap;
    int firstPageXref[10], *mainXref;
    int fXref = 1, mXref = 1;
    int i, mXref_pos, fXref_pos;

    objectMap = remapObjects (pdfCookie, dgp_state);
    mainXref = DGP_MALLOC (pdfCookie->objCount * sizeof (int));

    fprintf(out, "%%PDF-1.4\n%%%c%c%c%c\n", 250, 251, 252, 253);
    
    /* get the file into a well defined state */
    fflush (in);

    /* go to the beginning of the file */
    fseek (in, 0, SEEK_SET);

    /* generate space for linearized object */
    firstPageXref[fXref++] = ftell (out);
    fprintf (out, "%d 0 obj\nendobj\n\n", 
        objectMap[pdfCookie->catalogObj] - 1);

    /* generate space for 1st page XREF */
    fXref_pos = ftell (out);
    memset (buffer, 32, 1024);
    buffer[1023] = '\n';
    fwrite(buffer, 1, 1024, out);

    /* copy catalog */
    next_pos = startPos (pdfCookie, pdfCookie->catalogObj);
    size = 1024;
    fseek (in, next_pos, SEEK_SET);
    size = fread (buffer, 1, size, in);
    firstPageXref[fXref++] = ftell (out);
    copyFile (next_pos, next_pos + size, buffer, buf_size, objectMap, in, out);

    /* generate and write primary hint stream */
    firstPageXref[fXref++] = ftell (out);
    fprintf (out, "%d 0 obj\nendobj\n\n", 
        objectMap[pdfCookie->catalogObj] + 1);

    /* copy 3 page objects for first page */
    if (pdfCookie->pageObjList->nelt >= 1)
    {
        page = (int) pdfCookie->pageObjList->first->userdata;
        next_pos = startPos (pdfCookie, page);
        end_pos = endPos (pdfCookie, page++);
        firstPageXref[fXref++] = ftell (out);
        copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);

        next_pos = startPos (pdfCookie, page);
        end_pos = endPos (pdfCookie, page++);
        firstPageXref[fXref++] = ftell (out);
        copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);

        next_pos = startPos (pdfCookie, page);
        end_pos = endPos (pdfCookie, page);
        firstPageXref[fXref++] = ftell (out);
        copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);
    }

    /* copy dictionary and ProcSet */
    if (pdfCookie->dictList->nelt > 1)
    {
        page = (int) pdfCookie->dictList->first->userdata;
        next_pos = startPos (pdfCookie, page);
        end_pos = endPos (pdfCookie, page++);
        firstPageXref[fXref++] = ftell (out);
        copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);

        next_pos = startPos (pdfCookie, page);
        end_pos = endPos (pdfCookie, page);
        firstPageXref[fXref++] = ftell (out);
        copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);
    }

    /* All other pages */

    /* copy 3 page objects for all other pages page */
    if ( pdfCookie->pageObjList->nelt >= 2)
    {
        DgpAdt_ListNode ln = pdfCookie->pageObjList->first->next;
        while (ln)
        {
            page = (int) ln->userdata;
            next_pos = startPos (pdfCookie, page);
            end_pos = endPos (pdfCookie, page++);
            mainXref[mXref++] = ftell (out);
            copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);

            next_pos = startPos (pdfCookie, page);
            end_pos = endPos (pdfCookie, page++);
            mainXref[mXref++] = ftell (out);
            copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);

            next_pos = startPos (pdfCookie, page);
            end_pos = endPos (pdfCookie, page);
            mainXref[mXref++] = ftell (out);
            copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);
            
            ln = ln->next;
        }
    }
    
    /* copy other dictionaries */
    if ( pdfCookie->dictList->nelt >= 3)
    {
        DgpAdt_ListNode ln = pdfCookie->dictList->first->next->next;
        while (ln)
        {
            page = (int) ln->userdata;
            next_pos = startPos (pdfCookie, page);
            end_pos = endPos (pdfCookie, page);
            mainXref[mXref++] = ftell (out);
            copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);
            fprintf (stderr, " %d %d \n", mXref - 1, objectMap[page]);
            
            ln = ln->next;
        }
    }

    /* copy page tree */
    next_pos = startPos (pdfCookie, pdfCookie->PagesObj);
    end_pos = endPos (pdfCookie, pdfCookie->PagesObj);
    mainXref[mXref++] = ftell (out);
    copyFile (next_pos, end_pos, buffer, buf_size, objectMap, in, out);

    /* create main XRef */
    mXref_pos = ftell (out);
    fprintf (out, "xref\n0 %d\n0000000000 65535 f \n", mXref);
    for (i = 1; i < mXref; i++)
        fprintf (out, "%010ld 00000 n \n", mainXref[i]);
    fprintf (out, "trailer\n<< /Size %d >>\nstartxref\n%d\n%%%%EOF\n", mXref, fXref_pos);

    /* create first page XRef */
    fflush (out);
    fseek (out, fXref_pos, SEEK_SET);
    fprintf (out, "xref\n%d %d\n", mXref, fXref - 1);
    for (i = 1; i < fXref; i++)
        fprintf (out, "%010ld 00000 n \n", firstPageXref[i]);
    fprintf (out, "trailer\n<< /Size %d\n/Prev %d\n/Root %d 0 R >>\nstartxref\n0\n%%%%EOF\n", 
        mXref + fXref - 1, mXref_pos, objectMap[pdfCookie->catalogObj]);
    

    fclose (out);
    DGP_FREE (objectMap);
    DGP_FREE (mainXref);
}

static int startPos (PdfCookie *pdfCookie, int Obj)
{
    int i;
    size_t first_pos = 0;

    for (i = 0; i < pdfCookie->xrefPos; i++)
    {
        if (pdfCookie->xref[i].obj == Obj)
        {
            first_pos = pdfCookie->xref[i].offset;
            break;
        }
    }

    return first_pos;
}

static int endPos (PdfCookie *pdfCookie, int Obj)
{
    int i;
    size_t size = 0;

    for (i = 0; i < pdfCookie->xrefPos; i++)
    {
        if (pdfCookie->xref[i].obj == Obj)
        {
            if (i+1 < pdfCookie->xrefPos) 
            {
                size = pdfCookie->xref[i+1].offset;
                break;
            }
        }
    }

    return size;
}

static void copyFile (size_t start_pos, size_t end_pos,
               char* buffer, int buffer_size, int* objMap,
               FILE* in, FILE* out)
{
    Dgp_Bool check = DGP_TRUE;
    size_t temp = 1, size = end_pos - start_pos, size_used = buffer_size;

    fseek (in, start_pos, SEEK_SET);

    while (size > 0 && temp > 0)
    {
        if (buffer_size > size) size_used = size;
        temp = fread (buffer, 1, size_used, in);
        if (check)
        {
            char *bufptr, *stream, *nextpos, *endpos;
            int objnr = atoi (buffer);
    
            /* set the end marker */
            /* terminate the buffer to limit the search */
            if (temp < buffer_size) buffer[temp] = '\0';
            endpos = buffer + temp;

            /* renumber the object */
            if (objnr > 0 && objnr < 100000)
                    fprintf (out, "%d 0 obj", objMap[objnr]);
            bufptr = strstr (buffer, "obj");
            bufptr += 3;

            /* look for a stream, and terminate the buffer there */
            stream = strstr (bufptr, "stream");
            if (stream)
            {
                *stream = '\0';
                check = DGP_FALSE;
            }

            nextpos = strstr (bufptr, " 0 R");
            while (nextpos)
            {
                char* pChar = nextpos;

                /* go to the beginning of this reference */
                while (*pChar == ' ' || *pChar == '\n' || *pChar == '\r')
                    pChar--;
                while (*pChar != ' ' && *pChar != '\n' && *pChar != '\r')
                    pChar--;
                pChar++;
                objnr = atoi (pChar);

                fwrite (bufptr, 1, pChar - bufptr, out);
                fprintf (out, "%d 0 R", objMap[objnr]);

                bufptr = nextpos + 4;
                nextpos = strstr (bufptr, " 0 R");
            }

            if (stream) *stream = 's';
            fwrite (bufptr, 1, endpos - bufptr, out);
        } 
        else
            fwrite (buffer, 1, temp, out);

        size -= temp;
    }
    return;
}

static int* remapObjects (PdfCookie* pdfCookie, Dgp_State *dgp_state)
{
    DgpAdt_ListNode ln;
    int index = 1 ;
    int first;
    int* map = DGP_MALLOC (pdfCookie->objCount * sizeof (int));

    /* 3 objects per page */
    if (pdfCookie->pageObjList->nelt > 1)
    {
        ln = pdfCookie->pageObjList->first->next; 
        while (ln)
        {
            first = (int) ln->userdata;
            map[first++] = index++;
            map[first++] = index++;
            map[first] = index++;
            ln = ln->next;
        }
    }
    
    /* all dicts after the first and proc set */
    if (pdfCookie->dictList->nelt > 2)
    {
        ln = pdfCookie->dictList->first->next->next; 
        while (ln)
        {
            first = (int) ln->userdata;
            map[first] = index++;
            ln = ln->next;
        }
    }

    /* page tree object */
    map[pdfCookie->PagesObj] = index++;

    /**** first page objects */
    /* linerize object */
    index++;

    /* catalog */
    map[pdfCookie->catalogObj] = index++;

    /* hint stream */
    index++;

    /* 3 objects for the first page */
    ln = pdfCookie->pageObjList->first; 
    first = (int) ln->userdata;
    map[first++] = index++;
    map[first++] = index++;
    map[first] = index++;

    /* first page dict plus proc set */
    ln = pdfCookie->dictList->first;     
    first = (int) ln->userdata;
    map[first++] = index++;
    if (pdfCookie->dictList->nelt > 1)
        map[first] = index++;


    for (index = 1; index < pdfCookie->objCount; index++)
        printf ("%d  %d\n", index, map[index]);
    return map;
}
