
#include "dgp_types.h"
#include "iowrap.h"
#include "dgp_state.h"
#include "list.h"

#include "iomrc.h"

#ifndef MRC2FILE_H
#define MRC2FILE_H

#define DGP_MAXDICT 256

typedef struct _Dgp_CompressedImage
{
  Dgp_UInt32 width, height, x, y; /* coordinates and size */
  Dgp_UInt32 xres, yres;          /* resolution */
  Dgp_UInt8* data;
  Dgp_UInt32 data_size;
  Dgp_UInt32 extData_size;        /* extension data, color extensions for tokens */
  Dgp_UInt8 layer;
  Dgp_UInt32 quality;             /* jpeg quality */
} Dgp_CompressedImage;

typedef struct _Dgp_CompressedPage
{
  Dgp_State *dgp_state;

  Dgp_UInt32 width, height;

  /* Token position block */
  Dgp_CompressedImage *posBlock;

  /* Jpeg Images, list of Dgp_CompressedImages */
  DgpAdt_List jpegImages;
  DgpAdt_List fgImages;


} Dgp_CompressedPage;

typedef struct _Dgp_CompressedSection
{
  DgpAdt_List pages;

  /* compressed token dictionary */
  Dgp_UInt8* dict;
  Dgp_UInt32 dict_size;
  Dgp_UInt8 dicts_used;
} Dgp_CompressedSection;

typedef struct _Dgp_CompressedDocument
{
  Dgp_State *dgp_state;

  Dgp_UInt16 n_pages;
  Dgp_CompressedSection *currentSection;

  /* token dictionaries */
  Dgp_UInt32 dict_list[DGP_MAXDICT];
  Dgp_UInt16 n_dict;

  Dgp_UInt32 ioInstanceID;
  IO_DOCHANDLE outIoDoc;
  FileID outfile;

  void* userdata;
  
} Dgp_CompressedDocument;

/* create a TIFF-FX file and return a pointer to the data structure */
ErrCode
Dgp_createCompressedDocument (Dgp_State *dgp_state, uds *u, 
			      Dgp_CompressedDocument **cDocument);

/* close the TIFF-FX file and free all remaining memory */
ErrCode
Dgp_closeCompressedDocument (Dgp_CompressedDocument *cDocument);

/* write a section to file and free all memory associated with this section */
ErrCode
Dgp_writeSection (Dgp_CompressedDocument *cDocument, Dgp_Bool clear_dict);

/* create a new page */
ErrCode
Dgp_startNewPage (Dgp_CompressedDocument *cDocument, Dgp_UInt32 width, 
		  Dgp_UInt32 height, Dgp_CompressedPage **cPage);

/* create a new page without add it to cDocument*/
ErrCode
Dgp_createNewPage (Dgp_CompressedDocument *cDocument, Dgp_UInt32 width, 
		  Dgp_UInt32 height, Dgp_CompressedPage **cPage);
/* add a new page to cDocument*/
ErrCode
Dgp_addNewPage (Dgp_CompressedDocument *cDocument, Dgp_CompressedPage *cPage);

/* free a  page */
ErrCode
Dgp_freePage (Dgp_CompressedDocument *cDocument, Dgp_CompressedPage **cPage);
/* store a compressed position block */
ErrCode
Dgp_storeCompPosBlock (Dgp_CompressedPage *cPage, 
		       Dgp_CompressedImage *posBlock);

/* store a JPEG image */
ErrCode
Dgp_storeJpegImage (Dgp_CompressedPage *cPage, 
		    Dgp_CompressedImage *jpeg);
/* store a JPEG image */
ErrCode
Dgp_storeFgImage (Dgp_CompressedPage *cPage, 
		            DgpAdt_List fgImages);
/* store a token dictionary */
ErrCode
Dgp_storeCompTokDict (Dgp_CompressedDocument *cDocument, 
		      Dgp_UInt8 *dict, Dgp_UInt32 dict_size);

/* get a pointer to a compressed page in the current section */
ErrCode
Dgp_getPageOfSection (Dgp_CompressedDocument *cDocument, Dgp_UInt32 pagenum,
			   Dgp_CompressedPage **cPage);


#endif
