#include <stdio.h>

#include "dgp_state.h"
#include "dgp_err_control.h"

#include "pdf_util.h"
#include "mrc2pdf.h" 
#include "PsgFix.h"

ErrCode
write_mrc_jbgglobal(PdfCookie     *pdfCookie,
                   Dgp_State     *dgp_state,
                   uds           *u,
                   Dgp_UInt8     *dict,
                   Dgp_UInt32    dictSize)
{

    DGP_PROLOGUE("write_jbig2_global()", DGP_PERRSTATE);
    
    if (dictSize > 0)
    {

    /* Dictionary */
	DGP_ERR_PASS_CLEANUP(addXref(pdfCookie->objCount, pdfCookie->offset,
        pdfCookie, dgp_state));
    pdfCookie->DictObj = pdfCookie->objCount;
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<< /Length %ld\n\
>>\n\
stream\n", pdfCookie->objCount++, dictSize); 
    u->fwrite (dict, dictSize, 1, u->fp);
    pdfCookie->offset += dictSize; 
    pdfCookie->offset += fprintf(u->fp, "\n\
endstream\n\
endobj\n\n");
    }
  
  NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

ErrCode
write_mrc_info(    PdfCookie   *pdfCookie,
                   Dgp_State     *dgp_state,
                   uds           *u,
                   Dgp_UInt32    width,
                   Dgp_UInt32    height,
                   Dgp_UInt32    pageDPI,
                   Dgp_UInt32    *selObjId,
                   Dgp_UInt32    *fgObjId,
                   Dgp_UInt32    *bgObjId)
{
    DGP_PROLOGUE("write_mrc_background()", DGP_PERRSTATE);

    Dgp_UInt32      pageObj;
    Dgp_UInt32      pdfWidth, pdfHeight;
    Dgp_UInt32      pdfDefaultDPI = 72;
    Dgp_Int8        sizeStr[256];
    Dgp_UInt32      sizeStrLen;
/* Page description */
    pdfWidth = width * pdfDefaultDPI / pageDPI;
    pdfHeight = height * pdfDefaultDPI / pageDPI;
    
    pdfCookie->objCount += 9;
    pageObj = pdfCookie->objCount;

    /*content obj */
    sizeStrLen = sprintf (sizeStr, "q %ld 0 0 %ld 0 0 cm /I0 Do Q\n\
q %ld 0 0 %ld 0 0 cm /I2 Do Q",  pdfWidth, pdfHeight, pdfWidth, pdfHeight);
    /*sizeStrLen = sprintf (sizeStr, "q %ld. 0 0 %ld 0 0 cm /I2 Do Q",  pdfWidth, pdfHeight); */

	DGP_ERR_PASS_CLEANUP(addXref(pageObj  - 2, pdfCookie->offset, pdfCookie, dgp_state));
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<< /Length %ld >>\n\
stream\n\
%s\n\
endstream\n\
endobj\n\n", pageObj - 2, sizeStrLen, sizeStr);

    /* resource obj */
	DGP_ERR_PASS_CLEANUP(addXref(pageObj - 1, pdfCookie->offset, pdfCookie, dgp_state));
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<<\n\
/ProcSet [ /PDF /ImageC ]\n\
/XObject << /I0 %ld 0 R /I1 %ld 0 R /I2 %ld 0 R >>\n\
>>\n\
endobj\n\n", pageObj - 1,     pageObj - 7,  pageObj - 5,  pageObj - 3); 
    *bgObjId   = pageObj  - 7 ;
    *selObjId  = pageObj  - 5;
    *fgObjId   = pageObj  - 3;
 
    /* header obj */
	DGP_ERR_PASS_CLEANUP(addXref(pageObj, pdfCookie->offset, pdfCookie, dgp_state));
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd (pdfCookie->pageObjList, (void*) pageObj, NULL));
    pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj <<\n\
/Type /Page\n\
/Resources %ld 0 R\n\
/MediaBox [ 0 0 %ld %ld ]\n\
/CropBox [ 0 0 %ld %ld ]\n\
/Contents %ld 0 R\n\
/Parent %ld 0 R\n\
/Rotate 0\n\
>>\n\
endobj\n\n",    pdfCookie->objCount++, pageObj - 1, 
                pdfWidth, pdfHeight, pdfWidth, pdfHeight,
                pageObj - 2,  pdfCookie->PagesObj);

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

ErrCode  
write_mrc_background(  PdfCookie *pdfCookie,
                   Dgp_State     *dgp_state,
                   uds           *u,
                   Dgp_UInt8     *data,
                   Dgp_UInt32    dataSize,
                   Dgp_UInt32    bgWidth,
                   Dgp_UInt32    bgHeight,
                   Dgp_UInt8     depth,
                   Dgp_UInt32    bgObjId)
{

    DGP_PROLOGUE("write_mrc_background()", DGP_PERRSTATE);
   
    /*write data length obj */
    DGP_ERR_PASS_CLEANUP(addXref(bgObjId-1, pdfCookie->offset,
        pdfCookie, dgp_state)); 
    pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
%ld\n\
endobj\n\n",bgObjId -1, dataSize);
   
     /* write bg obj */
	DGP_ERR_PASS_CLEANUP(addXref(bgObjId, pdfCookie->offset,
        pdfCookie, dgp_state)); 
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<< /Type /XObject\n\
/Subtype /Image\n\
/Width %ld /Height %ld\n\
/BitsPerComponent %d\n\
/Name /I0\n\
/ColorSpace /DeviceRGB\n\
/Interpolate true /Filter [ /DCTDecode ]\n\
/DecodeParms [ << >> ]\n\
/Length %ld 0 R >>\n\
stream\n",  bgObjId, 
            bgWidth, bgHeight, 
            depth,
            bgObjId - 1); /*reference to lenght obj */
    
    /*write data */
    u->fwrite (data, dataSize, 1, u->fp);
    pdfCookie->offset += dataSize;

    /*write end obj */
    pdfCookie->offset += fprintf(u->fp, "\n\
endstream\n\
endobj\n\n");
  
  NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}
ErrCode  
write_mrc_selector(  PdfCookie     *pdfCookie,
                   Dgp_State     *dgp_state,
                   uds           *u,
                   Dgp_UInt8     *data,
                   Dgp_UInt32    dataSize,
                   Dgp_UInt32    selWidth,
                   Dgp_UInt32    selHeight,
                   Dgp_UInt8     depth,
                   Dgp_UInt32    selObjId)
{

    DGP_PROLOGUE("write_mrc_background()", DGP_PERRSTATE);
    
    /*write data length obj */
    DGP_ERR_PASS_CLEANUP(addXref(selObjId-1, pdfCookie->offset, pdfCookie, dgp_state)); 
    pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
%ld\n\
endobj\n\n",selObjId-1, dataSize);

/* write selector obj */
	DGP_ERR_PASS_CLEANUP(addXref(selObjId, pdfCookie->offset, pdfCookie, dgp_state)); 
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<< /Type /XObject\n\
/Subtype /Image\n\
/Width %ld /Height %ld\n\
/BitsPerComponent %d\n\
/Name /I1\n\
/ImageMask true\n\
/Filter /JBIG2Decode\n", selObjId, selWidth, selHeight, depth);
   
    /*reference to dict obj */
    if (pdfCookie->DictObj > 0) pdfCookie->offset += fprintf (u->fp, "\
/DecodeParms << /JBIG2Globals %ld 0 R >>\n", pdfCookie->DictObj);
    
    /*reference to length obj */
    pdfCookie->offset += fprintf (u->fp, "\
/Length %ld 0 R\n\
>>\n\
stream\n", selObjId-1);
    
    /*write data */
    u->fwrite (data, dataSize, 1, u->fp);
    pdfCookie->offset += dataSize;

    /*write end obj */
    pdfCookie->offset += fprintf(u->fp, "\n\
endstream\n\
endobj\n\n");

  NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

ErrCode  
write_mrc_forground(  PdfCookie     *pdfCookie,
                   Dgp_State     *dgp_state,
                   uds           *u,
                   Dgp_UInt8     *data,
                   Dgp_UInt32    dataSize,
                   Dgp_UInt32    fgWidth,
                   Dgp_UInt32    fgHeight,
                   Dgp_UInt8     depth,
                   Dgp_UInt32    fgObjId,
                   Dgp_UInt32    selObjId)
{

    DGP_PROLOGUE("write_mrc_background()", DGP_PERRSTATE);
   
    /*write data length obj */
    DGP_ERR_PASS_CLEANUP(addXref(fgObjId-1, pdfCookie->offset,
        pdfCookie, dgp_state)); 
    pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
%ld\n\
endobj\n\n",fgObjId-1, dataSize);
   
     /* write fg obj */
	DGP_ERR_PASS_CLEANUP(addXref(fgObjId, pdfCookie->offset, pdfCookie, dgp_state)); 
	pdfCookie->offset += fprintf(u->fp, "\
%ld 0 obj\n\
<< /Type /XObject\n\
/Subtype /Image\n\
/Width %ld /Height %ld\n\
/BitsPerComponent %d\n\
/Name /I2\n\
/Mask %ld 0 R\n\
/ColorSpace /DeviceRGB\n\
/Interpolate true /Filter [ /DCTDecode ]\n\
/DecodeParms [ << >> ]\n\
/Length %ld 0 R >>\n\
stream\n",  fgObjId, 
            fgWidth, fgHeight,  
            depth,
            selObjId,
            fgObjId-1); 
    
    /*write data */
    u->fwrite (data, dataSize, 1, u->fp);
    pdfCookie->offset += dataSize;

    /*write end obj */
    pdfCookie->offset += fprintf(u->fp, "\n\
endstream\n\
endobj\n\n");
  NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}
