
#include "jbig2file.h"
#include "pdf_util.h"
#include "mrc2pdf.h"
#include "g4.h"
#include "PsgFix.h"

/* create a JBIG2 file and return a pointer to the data structure */
ErrCode
Dgp_createPDFStream (Dgp_State *dgp_state, uds *u, 
			      Dgp_CompressedDocument **cDocument)
{
  Dgp_CompressedDocument *cDoc = NULL;
  DGP_PROLOGUE ("Dgp_createJbig2Stream", DGP_PERRSTATE);
  FILE *fp ;
  PdfCookie *pdfCookie;

  DGP_ASSERT (u != NULL);
  fp = u->fp;
  DGP_ASSERT (fp != NULL);
  DGP_ASSERT (cDocument != NULL);

    

  /* create structure */
  DGP_CHECK_ALLOC (cDoc = DGP_MALLOC (sizeof (Dgp_CompressedDocument)));
  cDoc->dgp_state = dgp_state;
  cDoc->outfile.u = u;

  /* set default vals */
  cDoc->n_pages = 0;
  cDoc->currentSection = NULL;
  cDoc->n_dict = 0;

  /* Begin a JBIG2 data stream */
  cDoc->outfile.fp = fp;
  cDoc->outfile.pos = 0UL;

  /* create section structure */
  DGP_CHECK_ALLOC (cDoc->currentSection = 
   DGP_MALLOC (sizeof (Dgp_CompressedSection)));

  cDoc->currentSection->dict = NULL;
  cDoc->currentSection->dict_size = 0;
  DGP_ERR_PASS_CLEANUP (DgpAdt_liNew (&(cDoc->currentSection->pages), dgp_state));

  /* init a PDF cookie */
  DGP_CHECK_ALLOC (pdfCookie = DGP_MALLOC (sizeof(PdfCookie)));
  DGP_ERR_PASS_CLEANUP (initPdfCookie(pdfCookie, dgp_state));
  cDoc->userdata = (void*) pdfCookie;

  /* write the PDF header */
  write_pdf_prolog(u, pdfCookie);

  *cDocument = cDoc;
  NORMAL_RETURN;

 ERR_CLEANUP:
  if (cDoc)
    {
      if (cDoc->currentSection) DGP_FREE (cDoc->currentSection);
      DGP_FREE (cDoc);
    }
  ERR_RETURN;

}

/* close the TIFF-FX file and free all remaining memory */
ErrCode
Dgp_closePDFStream (Dgp_CompressedDocument *cDocument)
{
  Dgp_State *dgp_state = cDocument->dgp_state;
  PdfCookie *pdfCookie = (PdfCookie*) cDocument->userdata;
  DGP_PROLOGUE ("Dgp_closeJbig2Stream", DGP_PERRSTATE);

  /* write PDF catalog */
  DGP_ERR_PASS_CLEANUP(
    write_pdf_catalog(pdfCookie, cDocument->outfile.u, dgp_state));


  /* write PDF trailer or linearize the file */
  if (0)
  /* call linearizer */
    linearize_pdf_file (pdfCookie, cDocument->outfile.u, dgp_state);
  else
      write_pdf_end(pdfCookie, cDocument->outfile.u);
  cleanupPdfCookie(pdfCookie, dgp_state);
  DGP_FREE (pdfCookie);

  /* close JBIG2 stream */
  fclose(cDocument->outfile.fp);

  /* should confirm that current section is empty */
  DGP_FREE (cDocument->currentSection);
  DGP_FREE (cDocument);

  NORMAL_RETURN;
 ERR_CLEANUP:
  ERR_RETURN;
}

/* write the JBIG2 compressed data of this section to file and free all memory associated with this section */
ErrCode
Dgp_writePDFData (Dgp_CompressedDocument *cDocument)
{

  Dgp_State *dgp_state = cDocument->dgp_state;
  DGP_PROLOGUE ("Dgp_PDFData", DGP_PERRSTATE);
  PdfCookie *pdfCookie = (PdfCookie*) cDocument->userdata;

  /* ****NEW**** */
  Dgp_CompressedSection *sec = cDocument->currentSection;
  Dgp_CompressedPage *page;
  Dgp_CompressedImage *im;
  DgpAdt_ListNode ln_page, ln_jpeg;

  DGP_ERR_PASS_CLEANUP(write_jbig2_global(pdfCookie, dgp_state, cDocument->outfile.u, sec->dict, sec->dict_size));
  if (sec->dict_size > 0)
  {
    DGP_FREE (sec->dict);
    sec->dict = NULL;
    sec->dict_size = 0; 
  }
  
  DGPADT_FOREACH (ln_page, sec->pages)
  {
      /* write out the page */
      page = (Dgp_CompressedPage*) ln_page->userdata;
      if (page->posBlock &&  DGPADT_LILEN(page->jpegImages) ==0) /* binary */
      {
          im = page->posBlock;
          
          DGP_ERR_PASS_CLEANUP(
              write_jbig2_page(pdfCookie, dgp_state, cDocument->outfile.u, im->data, im->data_size, 
              im->width, im->height, im->xres));
          
          /* finish up token image */
          DGP_FREE (im->data);
          DGP_FREE (im);
      }else{
          Dgp_UInt32 selObjId = 0, fgObjId = 0, bgObjId = 0;
          im = page->posBlock;
          DGP_ERR_PASS_CLEANUP(write_mrc_info(    pdfCookie,
              dgp_state,
              cDocument->outfile.u,
              im->width,
              im->height,
              im->xres,
              &selObjId,
              &fgObjId,
              &bgObjId));
          DGPADT_FOREACH (ln_jpeg, page->jpegImages)
          {
              im = (Dgp_CompressedImage*) ln_jpeg->userdata;
              DGP_ERR_PASS_CLEANUP(write_mrc_background(    pdfCookie,
                  dgp_state,
                  cDocument->outfile.u,
                  im->data,
                  im->data_size,
                  im->width,
                  im->height,
                  8,
                  bgObjId));
              DGP_FREE (im->data);
              DGP_FREE (im);

          }    
          DgpAdt_liFree (page->jpegImages);
          page->jpegImages = DGP_NULLLIST;

          if(page->posBlock){
              im = page->posBlock;
              DGP_ERR_PASS_CLEANUP(write_mrc_selector(    pdfCookie,
                  dgp_state,
                  cDocument->outfile.u,
                  im->data,
                  im->data_size,
                  im->width,
                  im->height,
                  1,
                  selObjId));
              /* finish up token image */
              DGP_FREE (im->data);
              DGP_FREE (im);
          }
          page->posBlock = NULL; 
          DGPADT_FOREACH (ln_jpeg, page->fgImages)
          {
            im = (Dgp_CompressedImage*) ln_jpeg->userdata;
            DGP_ERR_PASS_CLEANUP(write_mrc_forground(    pdfCookie,
                  dgp_state,
                  cDocument->outfile.u,
                  im->data,
                  im->data_size,
                  im->width,
                  im->height,
                  8,
                  fgObjId,
                  selObjId));
            DGP_FREE (im->data);
            DGP_FREE (im);

          }      
          DgpAdt_liFree (page->fgImages);
          page->fgImages = DGP_NULLLIST;
      }
      Dgp_freePage(cDocument, &page);
  }

  cDocument->n_pages += (Dgp_UInt16) sec->pages->nelt;
  DgpAdt_liClear (sec->pages);

  NORMAL_RETURN;
ERR_CLEANUP:
  ERR_RETURN;
}
