#include "jb2file.h"
#include "datasource.h"
#include "memorydatasink.h"
#include "segment.h"
#include "segment_util.h"
#include "decgenreg.h"
#include "mqcoder.h"
#include "blit.h"
#include "image_defs.h"
#include "util.h" 
#include "datasource.h" 
#include "segment.h" 
#include <sys/types.h> 
#include "decsymdict.h" 
#include "decsymreg.h" 
#include "dechtdict.h" 
#include "dechtreg.h" 
#include "decgenreg.h" 
#include "decgenrefreg.h" 
#include "dechufftable.h" 
#include "decpageinfo.h" 
#include "decextseg.h" 
#include <stdlib.h> 

/* The combop to minterm translation table - here's as good a place as any */ 
const Dgp_Int32 jb2Minterms[5] = { 
    DGPBLAT_OP_OR, 
    DGPBLAT_OP_AND, 
    DGPBLAT_OP_XOR, 
    DGPBLAT_OP_XNOR, 
    DGPBLAT_OP_REPLACE 
    }; 

#define DUMMYFREE(fname) static void fname(jb2Segment *segment, Dgp_State *dgp_state) {segment=segment; dgp_state=dgp_state;} 
DUMMYFREE(jb2FreeProfilesSegment) 


/* A table of freeing routines indexed by segment type */ 
static void (*freeers [64])(jb2Segment *, Dgp_State *) = { 
    jb2FreeSymDictSegment, NULL, NULL, NULL, 
    jb2FreeSymRegSegment, NULL, jb2FreeSymRegSegment, jb2FreeSymRegSegment, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    jb2FreeHTDictSegment, NULL, NULL, NULL, 
    jb2FreeHTRegSegment, NULL, jb2FreeHTRegSegment, jb2FreeHTRegSegment, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    jb2FreeGenRegSegment, NULL, jb2FreeGenRegSegment, jb2FreeGenRegSegment, 
    jb2FreeGenRefRegSegment, NULL, jb2FreeGenRefRegSegment, jb2FreeGenRefRegSegment, 
    NULL, NULL, NULL, NULL, 
 
    jb2FreePageInfoSegment, jb2FreeEOPageSegment, jb2FreeEOStripeSegment, jb2FreeEOFileSegment, 
    jb2FreeProfilesSegment, jb2FreeHuffTableSegment, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, jb2FreeExtensionSegment, NULL 
    }; 

/*
 * Return the number of contexts that must be allocated in order to
 * decode using those parameters.
 */
size_t
jb2ABContextsNeeded(arith_bitmap_params *params)
{
    Dgp_Int32 tempsize = 0;

    DGP_PROLOGUE ("jb2ABContextsNeeded", NULL);
    DGP_STOP_WHINE;

    /* How many pixels in the template? */
    switch(params->rtemplate) {
      case 0: {
        tempsize = 16;
        break;
        }

      case 1: {
        tempsize = 13;
        break;
        }

      case 2:
      case 3: {
        tempsize = 10;
        break;
        }

      default: {
        DGP_WARNF0("bad template size");
        }
      }

    /*
     * Eventually, we might want to make allowances for AT pixels overlapping
     * regular pixels.
     */
    return(((size_t)1) << tempsize);
    }

void
jb2FreeEOStripeSegment(jb2Segment *segment, Dgp_State *dgp_state)
{
  dgp_state = dgp_state;
    segment = segment;          /* Nothing to do */
    }
void
jb2FreeEOFileSegment(jb2Segment *segment, Dgp_State *dgp_state)
{
  dgp_state = dgp_state;
    segment = segment;          /* Nothing to do */
    }
void
jb2FreePageInfoSegment(jb2Segment *segment, Dgp_State *dgp_state)
{
  dgp_state = dgp_state;
    segment = segment;          /* Nothing to do */
    }
void
jb2FreeEOPageSegment(jb2Segment *segment, Dgp_State *dgp_state)
{
  dgp_state = dgp_state;
    segment = segment;          /* Nothing to do */
}
void
jb2FreeGenRefRegSegment(jb2Segment *segment, Dgp_State *dgp_state)
{
    (DGP_FREE)((void *)segment->data.genrefreg.data);
}

/*
 * Free everything hung off the symbol dictionary part of the segment;
 * if any pointers have been NULLed out, don't free them.
 *
 * 29.8.2000 / dhtesq.
 * Moving the free token bitmaps responsibility to santa/supercache/wherever.
 * Serious memory ownership issues here.
 *
 * 30.4.2001 / dhtesq.
 * I don't know what upstream changes have taken place, but purification shows
 * freeing the bits here is necessary again.  The bug we used to have
 * when rotating in a viewer, which necessitated freeing the bits in libsanta
 * in certain circumstances, seems to be gone.  But keep an eye out.
 */
void
jb2FreeSymDictSegment(jb2Segment *segment, Dgp_State *dgp_state)
{
  Dgp_Allocators *a = &(dgp_state->a);
    Dgp_Int32 i;
    jb2Symbol *cursym;

    if (segment->data.symdict.symbols != (jb2Symbol *)NULL) {
        for (i = 0; i < segment->data.symdict.nsymbols; i++) {
            cursym = &segment->data.symdict.symbols[i];

            if (cursym->bits != (Dgp_UInt32 *)NULL) {
                (a->free)((void *)cursym->bits);
                cursym->bits = NULL;
            }

        }
        (a->free)((void *)segment->data.symdict.symbols);
        segment->data.symdict.symbols = NULL;
    }

    if (segment->data.symdict.direct_contexts != (Dgp_UInt8 *)NULL) {
        (a->free)((void *)segment->data.symdict.direct_contexts);
        segment->data.symdict.direct_contexts = NULL;
    }

    if (segment->data.symdict.ref_contexts != (Dgp_UInt8 *)NULL) {
        (a->free)((void *)segment->data.symdict.ref_contexts);
        segment->data.symdict.ref_contexts = NULL;
        }
}
void
jb2FreeHuffTableSegment(jb2Segment *segment, Dgp_State *dgp_state)
{
    jhtCleanup(&segment->data.table.table, dgp_state);
    (DGP_FREE)((void *)segment->data.table.table.lines);
    }
/* Free any type-specific data allocated for this segment */
void
jb2FreeGenRegSegment(jb2Segment *segment, Dgp_State *dgp_state)
{
    (DGP_FREE)((void *)segment->data.genreg.data);
    }
/* Get rid of instances and anything hung off them */
void
jb2FreeInstances(jb2SymbolInstance *instances, Dgp_Int32 ninstances,
                 Dgp_State *dgp_state)
{
  Dgp_Allocators *a = &(dgp_state->a);
    Dgp_Int32 i;
    jb2SymbolInstance *curinst;

    for (i = 0; i < ninstances; i++) {
        curinst = &instances[i];
        if (curinst->refdata != (jb2SymbolInstanceRefined *)NULL) {
            (a->free)((void *)curinst->refdata->bits);
            (a->free)((void *)curinst->refdata);
            }
        }
    (a->free)((void *)instances);
    }

void
jb2FreeSymRegSegment(jb2Segment *segment, Dgp_State *dgp_state)
{
    if (segment->data.symreg.instances != (jb2SymbolInstance *)NULL) {
        jb2FreeInstances(segment->data.symreg.instances,
                         segment->data.symreg.ninstances,
                         dgp_state);
        }
    }
void
jb2FreeHTDictSegment(jb2Segment *segment, Dgp_State *dgp_state)
{
  Dgp_Allocators *a = &(dgp_state->a);
    Dgp_Int32 i;

    if (segment->data.htdict.patterns != (jb2Pattern *)NULL) {
        for (i = 0; i < segment->data.htdict.npatterns; i++) {
            if (segment->data.htdict.patterns[i].bits != (Dgp_UInt32 *)NULL) {

                (a->free)((void *)segment->data.htdict.patterns[i].bits);
                }
            }
        (a->free)((void *)segment->data.htdict.patterns);
        }
    }

void
jb2FreeHTRegSegment(jb2Segment *segment, Dgp_State *dgp_state)
{
    (void)dgp_state;
    if (segment->data.htreg.cells != (DgpImage_Short)NULL) {
        DgpImage_free(segment->data.htreg.cells);
        }
    }

/* Return ceil(log2(n)); n must be > 0 */
Dgp_Int32
ceillog2(Dgp_Int32 n)
{
  DGP_VOID_PROLOGUE("ceillog2()", NULL);

    Dgp_Int32 i, n2;

    if (n == 0) {
        /*
         * I'd rather not assert fail here, as it's possible for a valid
         * file to call this with zero; do something reasonable that
         * allows it to succeed.  The symDecoder that gets produced won't
         * be really usable, but if there are no symbols at all, it should
         * never get used.
         */
        return(0);
        }
    DGP_VOID_ASSERT(n > 0);             /* Negative N should be impossible. */


    /*
     * Shift n until there's nothing left.  This computes floor(log2(n)).
     * If n is 1, you shift it once, and the result should be 0.
     * If n is 2 or 3, you shift it twice, and the result should be 1
     */
    for (i = -1, n2 = n; n2 > 0; i++) {
        n2 >>= 1;
        }

    /*
     * We have floor(log2(n)); if n = 2^i, then log2(n) = floor(log2(n))
     * Otherwise, log2(n) > floor(log2(n)), so
     * ceil(log2(n)) = floor(log2(n)) + 1
     */
    if (n != (1 << i)) {
        i++;
        }

    return(i);
    }

/*
 * Free the common stuff, then dispatch to a type-specific free routine.
 * Don't free the actual segment.
 */
void
jb2FreeSegmentInfo(jb2Segment *segment, Dgp_State *dgp_state)
{
  Dgp_Allocators *a = &(dgp_state->a);

  if (segment->refsegs != (Dgp_Int32 *)NULL) {
          (a->free)((void *)segment->refsegs);
          segment->refsegs = NULL;
  }
  if (segment->refretain != (Dgp_Bool *)NULL) {
          (a->free)((void *)segment->refretain);
          segment->refretain = NULL;
  }
  if (isRegionSegtype(segment->segtype) && segment->reginfo.rendered) {
          DgpImage_free(segment->reginfo.im);
  }
  (*freeers[segment->segtype])(segment, dgp_state);
  return;
}

