/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/datasink.c,v 1.10 2003/01/27 21:48:32 awilliam Exp $ 
 * 
 */ 
 
 
#include <stdio.h> 
#include "datasink.h" 
#include <sys/types.h> 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/datasink.c,v 1.10 2003/01/27 21:48:32 awilliam Exp $"); 
 
/* 
 * Routines for dealing with DataSinks.  A DataSink is basically a block 
 * of data plus a callback.  Calling the callback it on 
 * the DataSink replaces the buffer with a new one, disposing of the contents 
 * as desired, and returns DGP_TRUE, or returns DGP_FALSE indicating that the 
 * operation failed. 
 */ 
 
/* 
 * dsPutc(f, c) puts c into f, returning DGP_TRUE on success.  It's 
 * implemented as a macro, calling _dsPutc as required. 
 */ 
ErrCode 
_dsPutc(DataSink f, Dgp_UInt8 c) 
{ 
  DGP_PROLOGUE("_dsPutc()", f->DGP_PERRSTATE); 
 
  DGP_ERR_GEN((f->more_data == (ErrCode (*)(DataSink))NULL), cErrNullPtr); 
  DGP_ERR_PASS((*f->more_data)(f)); 
  DGP_ERR_GEN((f->bufsize - f->bufpos == 0), cErrIO_BadFileLength); 
 
    f->buf[f->bufpos++] = c; 
    f->count++; 
 
    NORMAL_RETURN; 
    } 
 
/* 
 * dsWrite(f, nbytes, buf) writes up to nbytes from f into buf, returning 
 * the actual number of bytes written. 
 */ 
ErrCode 
dsWrite(DataSink f, size_t nbytes, Dgp_UInt8 *buf, size_t *bytes_written) 
{ 
  DGP_PROLOGUE("dsWrite()", f->DGP_PERRSTATE); 
    Dgp_UInt8 *readpos = buf; 
    size_t bytes_left = nbytes; 
 
    for (;;) { 
	/* Write as many bytes as you can */ 
 
	if (f->bufpos + bytes_left <= f->bufsize) { 
	    /* Enough in the file's buffer to satisfy the write */ 
	    (void)memcpy((void *)(f->buf + f->bufpos), (void *)readpos, 
			 bytes_left); 
	    f->bufpos += bytes_left; 
	    f->count += bytes_left; 
	    *bytes_written = nbytes; 
	    NORMAL_RETURN; 
	    } 
 
	/* Not enough - copy what's there and go back for more */ 
	(void)memcpy((void *)(f->buf + f->bufpos), (void *)readpos, 
		     f->bufsize - f->bufpos); 
	bytes_left -= f->bufsize - f->bufpos; 
	readpos += f->bufsize - f->bufpos; 
	f->count += f->bufsize - f->bufpos; 
	f->bufpos = f->bufsize; 
 
	DGP_ERR_GEN_CLEANUP((f->more_data ==  
		     (ErrCode (*)(DataSink))NULL), cErrNullPtr); 
	DGP_ERR_PASS_CLEANUP((*f->more_data)(f)); 
	} 
 
 ERR_CLEANUP: 
    *bytes_written = nbytes- bytes_left; 
 
    ERR_RETURN; 
    } 
 
/* 
 * These two routines allow a single mark to be placed in the DataSink, 
 * at the current location; later, you can come back and fill in the mark 
 * with actual data.  This is intended for use when writing out segment 
 * headers: you write out the segment header, set a 4-byte mark where 
 * the data length field should be, then generate the segment data.  After 
 * the segment data length is known, you do a dsFillMark with the data 
 * length. 
 * 
 * Each DataSink can only have one outstanding mark at a time.  The marks 
 * are implemented by the underlying abstraction layer (the one that takes 
 * care of full buffers), since there's no way for this layer to know 
 * what happens to buffers once they leave (they may be written to disk, or 
 * stored in memory, for example). 
 */ 
ErrCode 
dsMark(DataSink f, size_t nbytes) 
{ 
  DGP_PROLOGUE("dsMark()", f->DGP_PERRSTATE); 
 
    /* Set the mark */ 
  DGP_ERR_PASS((*f->mark)(f)); 
 
    /* Skip nbytes */ 
    while (f->bufpos + nbytes > f->bufsize) { 
	nbytes -= f->bufsize - f->bufpos; 
	f->bufpos = f->bufsize; 
 
	DGP_ERR_GEN((f->more_data ==  
		     (ErrCode (*)(DataSink))NULL), cErrNullPtr); 
	DGP_ERR_PASS((*f->more_data)(f)); 
 
	} 
    f->bufpos += nbytes; 
 
    NORMAL_RETURN; 
    } 
 
ErrCode 
dsFillMark(DataSink f, size_t nbytes, Dgp_UInt8 *buf) 
{ 
  DGP_PROLOGUE("dsFillMark()", f->DGP_PERRSTATE); 
 
  DGP_ERR_PASS((*f->fillmark)(f, nbytes, buf)); 
 
  NORMAL_RETURN; 
    } 
 
/* 
 * These three functions are intended to make it easy to measure how many 
 * bytes get written to a DataSink.  Call dsResetCount to set the count to 
 * zero; call dsCount to get the count of bytes written since the last 
 * dsResetCount; call dsAdjustCount to report how many bytes have 
 * been written without passing through _dsPutc or dsWrite (for example, 
 * in a jbitfile) 
 */ 
void 
dsResetCount(DataSink f) 
{ 
    f->count = 0; 
    } 
 
size_t 
dsCount(DataSink f) 
{ 
    return(f->count); 
    } 
     
void 
dsAdjustCount(DataSink f, size_t adjust) 
{ 
    f->count += adjust; 
    } 
 
/* 
 * These routines are for simple DataSinks (single buffer, no 
 * more_data callback) to allow for marks.  They use the DataSink's userdata 
 * field. 
 */ 
ErrCode 
jb2SimpleDataSinkMark(DataSink f) 
{ 
  DGP_PROLOGUE("jb2SimpleDataSinkMark()", f->DGP_PERRSTATE); 
 
  /* keep compiler happy */ 
  DGP_STOP_WHINE; 
    f->userdata = (void *)(f->buf + f->bufpos); 
    NORMAL_RETURN; 
    } 
 
ErrCode 
jb2SimpleDataSinkFillmark(DataSink f, size_t nbytes, Dgp_UInt8 *buf) 
{ 
  DGP_PROLOGUE("jb2SimpleDataSinkFillmark()", f->DGP_PERRSTATE); 
 
  /* keep compiler happy */ 
  DGP_STOP_WHINE; 
    memcpy(f->userdata, (void *)buf, nbytes); 
    NORMAL_RETURN; 
    } 
 
 
/* 
 * These routines make it easy to attach an DgpEncode_MQEncoder to a DataSink. 
 * Call dsInitDgpEncode_MQEncoder on the DataSink and the (uninitialised) DgpEncode_MQEncoder. 
 * Use the DgpEncode_MQEncoder (don't touch the DataSink while this is going on). 
 * When you're done, call dsFinalDgpEncode_MQEncoder. 
 */ 
 
/* Output callback for the MQ coder */ 
static Dgp_Bool 
byte_out(Dgp_UInt8 b, void *userdata) 
{ 
    /* userdata is a DataSink */ 
    DataSink data = (DataSink)userdata; 
    DGP_PROLOGUE("byte_out()", data->DGP_PERRSTATE); 
 
    DGP_ERR_PASS_CLEANUP(dsPutc(data, b)); 
 
    return(DGP_TRUE); 
 
 ERR_CLEANUP: 
 
    return(DGP_FALSE); 
    } 
 
#ifndef	NOMQ 
void 
dsInitDgpEncode_MQEncoder(DataSink f, DgpEncode_MQEncoder *mqenc) 
{ 
    DgpEncode_mqeInit(mqenc, byte_out, (void *)f, f->dgp_state); 
    } 
 
ErrCode 
dsFinalDgpEncode_MQEncoder(DataSink f, DgpEncode_MQEncoder *mqenc) 
{ 
  DGP_PROLOGUE("dsFinalDgpEncode_MQEncoder()", f->DGP_PERRSTATE); 
 
  DGP_ERR_GEN((!DgpEncode_mqeFlush(mqenc)), cErrIO_WriteError); 
 
  NORMAL_RETURN; 
    } 
#endif 
