/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/datasource.c,v 1.7 2003/01/27 21:48:32 awilliam Exp $ 
 * 
 */ 
 
#include <stdio.h> 
#include "datasource.h" 
#include <sys/types.h> 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/datasource.c,v 1.7 2003/01/27 21:48:32 awilliam Exp $"); 
 
/* 
 * Routines for dealing with DataSources.  A DataSource is basically a block 
 * of data plus a callback.  If the callback is NULL, then that block of 
 * data is all there is.  If the callback is not NULL, then calling it on 
 * the DataSource replaces the buffer with a new one and returns DGP_TRUE, or 
 * returns DGP_FALSE indicating that the operation failed. 
 */ 
 
/* 
 * dsGetc(f) gets the next Dgp_Int8acter from f.  It's 
 * implemented as a macro, calling _dsGetc as required. 
 */ 
ErrCode 
_dsGetc(DataSource f, Dgp_UInt8 *val) 
{ 
  DGP_PROLOGUE("_dsGetc()", f->DGP_PERRSTATE); 
 
  DGP_ERR_GEN((f->more_data ==  
		       (ErrCode (*)(DataSource))NULL), cErrNullPtr); 
  DGP_ERR_PASS((*f->more_data)(f)); 
  DGP_ERR_GEN((f->bufsize - f->bufpos == 0), cErrIO_BadFileLength); 
 
  *val = f->buf[f->bufpos++]; 
   
  NORMAL_RETURN; 
    } 
 
/* 
 * dsPeek returns the next Dgp_Int8acter in f (without actually advancing 
 * the buffer pointer), or EOF if there is no next Dgp_Int8acter. 
 */ 
ErrCode 
dsPeek(DataSource f, Dgp_Int32 *val) 
{ 
  DGP_PROLOGUE("dsPeek()", f->DGP_PERRSTATE); 
 
    if (f->bufpos == f->bufsize) { 
 
      DGP_ERR_GEN_CLEANUP((f->more_data ==  
			   (ErrCode (*)(DataSource))NULL), cErrNullPtr); 
      DGP_ERR_PASS_CLEANUP((*f->more_data)(f)); 
 
      if(f->bufsize - f->bufpos == 0) { 
	*val = EOF; 
	NORMAL_RETURN; 
      } 
	} 
  *val = f->buf[f->bufpos]; 
   
  NORMAL_RETURN; 
 
 ERR_CLEANUP: 
  *val = EOF; 
 
  ERR_RETURN; 
 
    } 
     
/* 
 * dsRead(f, nbytes, buf) reads up to nbytes from f into buf, returning 
 * the actual number of bytes read. 
 */ 
ErrCode 
dsRead(DataSource f, size_t nbytes, Dgp_UInt8 *buf, size_t *bytes_read) 
{ 
  DGP_PROLOGUE("dsRead()", f->DGP_PERRSTATE); 
    Dgp_UInt8 *writepos = buf; 
    size_t bytes_left = nbytes; 
 
    for (;;) { 
	/* Read as many bytes as you can */ 
 
	if (f->bufpos + bytes_left <= f->bufsize) { 
	    /* Enough in the file's buffer to satisfy the read */ 
	    (void)memcpy((void *)writepos, (void *)(f->buf + f->bufpos), 
			 bytes_left); 
	    f->bufpos += bytes_left; 
	    *bytes_read = nbytes; 
	    NORMAL_RETURN; 
	    } 
 
	/* Not enough - copy what's there and go back for more */ 
	(void)memcpy((void *)writepos, (void *)(f->buf + f->bufpos), 
		     f->bufsize - f->bufpos); 
	bytes_left -= f->bufsize - f->bufpos; 
	writepos += f->bufsize - f->bufpos; 
	f->bufpos = f->bufsize; 
 
	DGP_ERR_GEN_CLEANUP((f->more_data ==  
			     (ErrCode (*)(DataSource))NULL), cErrNullPtr); 
	DGP_ERR_PASS_CLEANUP((*f->more_data)(f)); 
 
      } 
 
 ERR_CLEANUP: 
    *bytes_read = nbytes - bytes_left; 
 
    ERR_RETURN; 
    } 
 
#ifndef	NOMQ 
/* 
 * A callback shim: convert a DataSource more_data callback into 
 * a DgpEncode_MQDecoder callback. 
 */ 
static Dgp_Bool 
data_in(Dgp_UInt8 **buffer, size_t *buflen, void *userdata) 
{ 
    DataSource data = (DataSource)userdata; 
    DGP_PROLOGUE("data_in()", data->DGP_PERRSTATE); 
 
    DGP_ERR_GEN_CLEANUP((data->more_data ==  
			 (ErrCode (*)(DataSource))NULL), cErrNullPtr); 
    DGP_ERR_PASS_CLEANUP((*data->more_data)(data)); 
 
    *buffer = data->buf; 
    *buflen = data->bufsize; 
    return(DGP_TRUE); 
 
 ERR_CLEANUP: 
 
    return(DGP_FALSE); 
    } 
 
/* 
 * Initialise the MQ decoder, and set it up to read from the DataSource. 
 * Don't use the DataSource while the MQ decoder is active, obviously. 
 * Call dsFinalDgpEncode_MQDecoder when you're done with the MQ decoder. 
 */ 
void 
dsInitDgpEncode_MQDecoder(DataSource f, DgpEncode_MQDecoder *mqdec) 
{ 
    DgpEncode_mqdInit(mqdec, f->buf + f->bufpos, 
	    f->bufsize - f->bufpos, 
	    data_in, (void *)f, f->dgp_state); 
    } 
 
ErrCode 
dsFinalDgpEncode_MQDecoder(DataSource f, DgpEncode_MQDecoder *mqdec) 
{ 
  DGP_PROLOGUE("dsFinalDgpEncode_MQDecoder()", f->DGP_PERRSTATE); 
    Dgp_UInt8 *bp; 
 
    /* 
     * We call mqdFinishSW to get the pointer to the first unconsumed byte. 
     */ 
    bp = DgpEncode_mqdFinishSW(mqdec); 
 
    /* Something is badly wrong */ 
    DGP_ERR_GEN((bp == (Dgp_UInt8 *)NULL), cErrNullPtr); 
 
    f->bufpos = bp - f->buf; 
 
    NORMAL_RETURN; 
    } 
#endif 
