/* 
* 
* $Header: /project/icons/cvsroot/digipaper/src/libjbig2/decextseg.c,v 1.12 2003/02/10 22:53:26 mkrocken Exp $ 
* 
*/ 
 
#include "jb2file.h" 
#include "segment.h" 
#include "datasource.h" 
#include "decextseg.h" 
#include "writejbigCB.h" 
#include "PsgFix.h"

#include <util.h> 
#include <list.h> 
 
#define DGP_ENC DGP_FALSE

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/decextseg.c,v 1.12 2003/02/10 22:53:26 mkrocken Exp $"); 
 
static ErrCode decASCIIComment(jb2File file, DataSource data, 
							   jb2Segment *segment, jb2Segment **refsegs); 
static void freeASCIIComment(jb2Segment *segment, Dgp_State *dgp_state); 

static ErrCode decBinComment(jb2File file, DataSource data, 
							   jb2Segment *segment, jb2Segment **refsegs); 
static void freeBinComment(jb2Segment *segment, Dgp_State *dgp_state); 
 
typedef struct { 
    Dgp_UInt32 type; 
    ErrCode (*decode)(jb2File file, DataSource data, 
		      jb2Segment *segment, jb2Segment **refsegs); 
    void (*free)(jb2Segment *segment, Dgp_State *dgp_state); 
} exthandler; 
 
static exthandler handlers[] = { 
  { 0x20000000, decASCIIComment, freeASCIIComment } ,     
    { DGP_CBOX_EXT, decBinComment, freeBinComment }     
}; 
 
#define	NHANDLERS	((int)(sizeof(handlers) / (sizeof(handlers[0])))) 
 
ErrCode 
jb2DecodeExtension(jb2File file, DataSource data, 
				   jb2Segment *segment, jb2Segment **refsegs) 
{ 
	DGP_PROLOGUE("jb2DecodeExtension()", jb2_pErrState(file)); 
	 
    Dgp_UInt8 lbuf[4]; 
    Dgp_UInt32 type; 
    Dgp_Bool necessary, dependent, reserved; 
    Dgp_Int32 i; 
    size_t bytes; 
	 
    /* Read the 4-byte type */ 
    DSREAD(data, 4, lbuf, &bytes); 
    type = (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
	 
    /* Decode the top few bits */ 
    necessary = (type & (1ul << 31)) ? DGP_TRUE : DGP_FALSE; 
    dependent = (type & (1ul << 30)) ? DGP_TRUE : DGP_FALSE; 
    reserved = (type & (1ul << 29)) ? DGP_TRUE : DGP_FALSE; 
	 
    /* Now go through our table of type<->decoding routines */ 
    segment->data.extension.type = type; 
    for (i = 0; i < NHANDLERS; i++) { 
		if (type == handlers[i].type) { 
			return(handlers[i].decode(file, data, segment, refsegs)); 
		} 
	} 
    /* Nothing caught it */ 
    FILEERR(cErrCompressionError, "unrecognised segment type"); 
	 
    NORMAL_RETURN; 
	 
ERR_CLEANUP: 
    ERR_RETURN; 
} 
 
void 
jb2FreeExtensionSegment(jb2Segment *segment, Dgp_State *dgp_state) 
{ 
  Dgp_Int32 i; 
	 
  DGP_VOID_PROLOGUE ("jb2FreeExtensionSegment", DGP_PERRSTATE); 
	 
  /* Find the appropriate freeing routine */ 
  for (i = 0; i < NHANDLERS; i++) { 
    if (segment->data.extension.type == handlers[i].type) { 
      handlers[i].free(segment, dgp_state); 
      return; 
    } 
  } 
  /* Nothing caught it */ 
  DGP_REPORTF0(DGP_ENC, "freeing unrecognized segment type"); 
} 
 
/* Utility: read an unbounded-length NUL-terminated string from data */ 
#define	CHUNKSIZE	32	/* most strings are Dgp_Int16 */ 
static ErrCode 
getString(jb2File file, DataSource data, Dgp_Int8 **_final) 
{ 
	DGP_PROLOGUE("getString()", jb2_pErrState(file)); 
	 
	Dgp_Allocators *a = jb2_pAllocators(file); 
    DgpAdt_List chunks = DGP_NULLLIST; 
    DgpAdt_ListNode ln; 
    Dgp_Int8 chunk[CHUNKSIZE]; 
    Dgp_Int8 *final = (Dgp_Int8 *)NULL; 
    Dgp_Int32 i; 
    Dgp_Int32 len = 0; 
    Dgp_Bool done = DGP_FALSE; 
	 
    *_final = final; 
	 
    while (!done) { 
		for (i = 0; i < CHUNKSIZE; i++) { 
			DSGETC(data, &(chunk[i])); 
			len++; 
			if (chunk[i] == '\0') { 
				done = DGP_TRUE; 
				break; 
			} 
		} 
		if (!done) { 
			DgpAdt_ListNode retnode; 
			/* Throw this chunk onto the list */ 
			if (chunks == DGP_NULLLIST) { 
				/* oops - no list yet! */ 
				DGP_ERR_PASS (DgpAdt_liNew(&chunks, jb2GetFileDgp_State(file))); 
			} 
			 
			/* Allocate a copy of the current buffer */ 
			final = (Dgp_Int8 *)(a->malloc)(CHUNKSIZE); 
			if (final == (Dgp_Int8 *)NULL) { 
				NOMEM; 
			} 
			 
			memcpy(final, chunk, CHUNKSIZE); 
			 
			DGP_ERR_PASS (DgpAdt_liAdd(chunks, (void *)final, &retnode)); 
			final = (Dgp_Int8 *)NULL; 
		} 
	} 
	 
    /* We now have a list of chunks and a partial final chunk. */ 
    final = (Dgp_Int8 *)(a->malloc)(len); 
    DGP_CHECK_ALLOC_CLEANUP(final); 
	 
    i = 0; 
    /* Copy in all the chunks, freeing as we go */ 
    if (chunks != DGP_NULLLIST) { 
		DGPADT_FOREACH (ln, chunks) { 
			memcpy(&final[i], (Dgp_Int8 *)DGPADT_LIGET(ln), CHUNKSIZE); 
			(a->free)(DGPADT_LIGET(ln)); 
			 
			i += CHUNKSIZE; 
		} 
		/* Done with the list */ 
		DgpAdt_liFree(chunks); 
	} 
	 
    /* Copy the rest up to the NUL */ 
    len = 0; 
    do { 
		final[i++] = chunk[len++]; 
	} while (chunk[len - 1] != '\0'); 
	 
    *_final = final; 
    final = (Dgp_Int8 *)NULL; 
     
    NORMAL_RETURN; 
	 
ERR_CLEANUP: 
    if (final != (Dgp_Int8 *)NULL) { 
		(a->free)((void *)final); 
	} 
	 
    if (chunks != DGP_NULLLIST) { 
		DGPADT_FOREACH (ln, chunks) { 
			(a->free)(DGPADT_LIGET(ln)); 
		} 
		DgpAdt_liFree(chunks); 
	} 
	 
    ERR_RETURN; 
} 
 
/* Create and fill out a jb2Extension_ASCIIComment structure */ 
static ErrCode 
decASCIIComment(jb2File file, DataSource data, 
				jb2Segment *segment, jb2Segment **refsegs) 
{ 
	DGP_PROLOGUE("decASCIIComment()", jb2_pErrState(file)); 
	 
	Dgp_Allocators *a = jb2_pAllocators(file); 
    DgpAdt_List pairs = DGP_NULLLIST; 
    DgpAdt_ListNode ln; 
    Dgp_Int8 **names = (Dgp_Int8 **)NULL; 
    Dgp_Int8 **values = (Dgp_Int8 **)NULL; 
    Dgp_Int8 *name = (Dgp_Int8 *)NULL; 
    Dgp_Int8 *value = (Dgp_Int8 *)NULL; 
    jb2Extension_ASCIIComment *comment = (jb2Extension_ASCIIComment *)NULL; 
    Dgp_Int32 npairs; 
    Dgp_Int32 i; 
     
    (void)refsegs; 
	 
    /* 
	* We'll build a list where each (name, value) pair is stored in 
	* two consecutive DgpAdt_ListNodes 
	*/ 
    DGP_ERR_PASS (DgpAdt_liNew(&pairs, jb2GetFileDgp_State(file))); 
	 
    comment = (jb2Extension_ASCIIComment *)(a->malloc)(sizeof(*comment)); 
    if (comment == (jb2Extension_ASCIIComment *)NULL) { 
		NOMEM; 
	} 
	 
    while (DGP_TRUE) { 
		DgpAdt_ListNode retnode; 
		DGP_ERR_PASS_CLEANUP(getString(file, data, &name)); 
		 
		if (name[0] == '\0') { 
			/* Terminating NUL */ 
			(a->free)((void *)name); 
			name = (Dgp_Int8 *)NULL; 
			break; 
		} 
		 
		DGP_ERR_PASS_CLEANUP(getString(file, data, &value)); 
		 
		DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd(pairs, (void *)name, &retnode)); 
		name = (Dgp_Int8 *)NULL; 
		 
		DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd(pairs, (void *)value, &retnode)); 
		value = (Dgp_Int8 *)NULL; 
	} 
	 
    /* How many pairs did we get? */ 
    npairs = DGPADT_LILEN(pairs) / 2; 
	 
    /* Allocate the collective arrays */ 
    names = (Dgp_Int8 **)(a->malloc)(npairs * sizeof(*names)); 
    if (names == (Dgp_Int8 **)NULL) { 
		NOMEM; 
	} 
	 
    values = (Dgp_Int8 **)(a->malloc)(npairs * sizeof(*values)); 
    if (values == (Dgp_Int8 **)NULL) { 
		NOMEM; 
	} 
	 
    /* and read them out.  NOTHING MAY FAIL BEYOND THIS POINT. */ 
    for (ln = DGPADT_LIFIRST(pairs), i = 0; i < npairs; i++) { 
		names[i] = (Dgp_Int8 *)DGPADT_LIGET(ln); 
		ln = DGPADT_LINEXT(ln); 
		DGP_ASSERT(ln != DGP_NULLLISTNODE); 
		 
		values[i] = (Dgp_Int8 *)DGPADT_LIGET(ln); 
		ln = DGPADT_LINEXT(ln); 
	} 
    DGP_ASSERT(ln == DGP_NULLLISTNODE); 
    DgpAdt_liFree(pairs); 
     
    comment->npairs = npairs; 
    comment->names = names; 
    comment->values = values; 
	 
    segment->data.extension.data = (void *)comment; 
	 
    NORMAL_RETURN; 
	 
ERR_CLEANUP: 
	 
    if (pairs != DGP_NULLLIST) { 
		/* pairs owns the stuff on it */ 
		DGPADT_FOREACH (ln, pairs) { 
			(a->free)(DGPADT_LIGET(ln)); 
		} 
		DgpAdt_liFree(pairs); 
	} 
	 
    if (names != (Dgp_Int8 **)NULL) { 
		/* names never owns its contents; they're owned by pairs */ 
		(a->free)((void *)names); 
	} 
	 
    if (values != (Dgp_Int8 **)NULL) { 
		(a->free)((void *)values); 
	} 
	 
    if (name != (Dgp_Int8 *)NULL) { 
		(a->free)((void *)name); 
	} 
	 
    if (value != (Dgp_Int8 *)NULL) { 
		(a->free)((void *)value); 
	} 
	 
    if (comment != (jb2Extension_ASCIIComment *)NULL) { 
		/* comments doesn't own its contents: names and values own themselves */ 
		(a->free)((void *)comment); 
	} 
	 
    ERR_RETURN; 
    } 
	 
	static void 
		freeASCIIComment(jb2Segment *segment, Dgp_State *dgp_state) 
	{ 
		Dgp_Allocators *a = &(dgp_state->a); 
		int i; 
		jb2Extension_ASCIIComment *comment; 
		 
		comment = (jb2Extension_ASCIIComment *)segment->data.extension.data; 
		 
		for (i = 0; i < comment->npairs; i++) { 
			(a->free)((void *)comment->names[i]); 
			(a->free)((void *)comment->values[i]); 
		} 
		(a->free)((void *)comment->names); 
		(a->free)((void *)comment->values); 
		 
		(a->free)((void *)comment); 
    } 


/* copy binary data to segment */ 
static ErrCode 
decBinComment(jb2File file, DataSource data, 
				jb2Segment *segment, jb2Segment **refsegs) 
{ 
	DGP_PROLOGUE("decBinComment()", jb2_pErrState(file)); 
	 
	Dgp_Allocators *a = jb2_pAllocators(file); 
  size_t size, bytes;
  Dgp_UInt8 *buffer = NULL;
  
    (void)refsegs; 
	 
    /* subtract the size of the header */
    size = segment->datalen - 4;
	 
    /* Allocate the buffer */
    buffer = (Dgp_UInt8 *)(a->malloc)(size);
    if (buffer == (Dgp_UInt8 *)NULL) { 
		NOMEM; 
	} 
	 
    /* Read the data */ 
    DSREAD(data, size, buffer, &bytes); 
	 
    segment->data.extension.data = (void *) buffer;
    /* hijack the type field to send back the size of the buffer */
	/* 1.5.2001 / dhtesq.  Added userdata fields to extension, so we don't have */
	/* to use the type field anymore.  This also fixes a memory leak. */
    /*segment->data.extension.type = size;  */
	segment->data.extension.apple = size;

    NORMAL_RETURN; 
	 
ERR_CLEANUP: 
	 
    if (buffer != NULL)
    {
      (a->free) (buffer);
    }

    ERR_RETURN; 
    } 
	 
	static void 
		freeBinComment(jb2Segment *segment, Dgp_State *dgp_state) 
	{ 
		Dgp_Allocators *a = &(dgp_state->a); 
		 		 
		(a->free)((void *) segment->data.extension.data); 
    } 
