/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/decgenrefreg.c,v 1.17 2003/01/27 21:48:33 awilliam Exp $ 
 * 
 */ 
 
#include "jb2file.h" 
#include "datasource.h" 
#include "segment.h" 
#include "segment_util.h" 
#include "decgenrefreg.h" 
#include <stdlib.h> 
#include "mqcoder.h" 
#include "blit.h" 
#include "image_defs.h"
#include "PsgFix.h"
 
#define DGP_ENC DGP_FALSE

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/decgenrefreg.c,v 1.17 2003/01/27 21:48:33 awilliam Exp $"); 
 
/* Internal routines */ 
static Dgp_Bool checkRABParams(arith_ref_params *params); 
#ifndef	NOMQ 
static void refineImagePacked(DgpImage_Packed dest, DgpImage_Packed ref, 
			      Dgp_Int32 refdx, Dgp_Int32 refdy, arith_ref_params *params, 
			      Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec); 
static void refineImageT13(DgpImage_Packed dest, DgpImage_Packed ref, 
			   Dgp_Int32 refdx, Dgp_Int32 refdy, arith_ref_params *params, 
			   Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec); 
static void refineImageT10(DgpImage_Packed dest, DgpImage_Packed ref, 
			   Dgp_Int32 refdx, Dgp_Int32 refdy, arith_ref_params *params, 
			   Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec); 
#endif 
 
/* 
 * Decode generic refinement regions. 
 */ 
ErrCode 
jb2DecodeGenRefReg(jb2File file, DataSource data, 
		   jb2Segment *segment, jb2Segment **refsegs) 
{ 
  DGP_PROLOGUE("jb2DecodeGenRefReg()", jb2_pErrState(file)); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
    Dgp_UInt8 flags; 
    size_t datalen; 
    Dgp_Int32 bytesread = 0; 
    size_t bytes; 
 
#ifdef	NOMQ 
    FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
 
    /* 
     * Check the refinement for validity: there must be 
     * either one segment referred to, and that must be an intermediate 
     * region segment, or no segments referred to at all. 
     */ 
    if (segment->segtype == JB2SEGTYPE_GENREFREG_INT) { 
	/* Must refer to one segment */ 
	if (segment->nrefsegs != 1) { 
	    FILEERR(cErrCompressionError, "intermediate refinement refers to != 1 segments"); 
	    } 
	} 
    else { 
	/* Must refer to zero or one segments */ 
	if ((segment->nrefsegs != 0) && (segment->nrefsegs != 1)) { 
	    FILEERR(cErrCompressionError, "immediate refinement refers to != (0,1) segments"); 
	    } 
	} 
 
    /* If there's a segment referred to, it must be an intermediate region segment */ 
    if (segment->nrefsegs == 1) { 
	if ((!isRegionSegtype(refsegs[0]->segtype)) || 
	    (isImmediateRegion(refsegs[0]->segtype))) { 
	    FILEERR(cErrCompressionError, "refinement refers to non-intermediate region segment"); 
	    } 
 
	/* Stash the pointer to the to-be-refined segment */ 
	segment->data.genrefreg.refseg = refsegs[0]; 
	} 
    else { 
	/* No segment to be refined */ 
	segment->data.genrefreg.refseg = (jb2Segment *)NULL; 
	} 
 
    /* Decode the region segment header */ 
    DGP_ERR_PASS_CLEANUP(jb2DecodeRegSegment(file, data, &segment->reginfo)); 
    bytesread += 17;		/* read 17 bytes worth of region segment header */ 
 
    if (segment->nrefsegs == 1) { 
	/* 
	 * If we refer to another segment, our size, position, 
	 * and external combination operators must match it. 
	 */ 
	if ((refsegs[0]->reginfo.regwidth != segment->reginfo.regwidth) || 
	    (refsegs[0]->reginfo.regheight != segment->reginfo.regheight) || 
	    (refsegs[0]->reginfo.regxpos != segment->reginfo.regxpos) || 
	    (refsegs[0]->reginfo.regypos != segment->reginfo.regypos) || 
	    (refsegs[0]->reginfo.combop != segment->reginfo.combop)) { 
	    FILEERR(cErrCompressionError, "refinement segment doesn't match segment it's refining"); 
	    } 
	} 
 
    /* One-byte flags field */ 
    DSREAD(data, 1, &flags, &bytes); 
    bytesread++; 
 
#ifdef	VALPRINT 
    fprintf(stderr, "GBRFLAGS %02x\n", flags); 
#endif 
 
    /* GBRTEMPLATE */ 
    segment->data.genrefreg.refparams.rtemplate = flags & 1; 
 
    /* TPRON */ 
    segment->data.genrefreg.refparams.tpron = (flags & 2) ? DGP_TRUE : DGP_FALSE; 
 
    /* Decode the AT parameters */ 
    DGP_ERR_PASS_CLEANUP(jb2DecodeRABParams(file, data, 
					    &segment->data.genrefreg.refparams, 
					    &bytesread)); 
 
    /* 
     * All we do now is return a pointer to the compressed data.  Since 
     * there's no way to ensure that data inside a DataSource is going 
     * to stick around, we have to allocate a buffer and copy it in. 
     * 
     * Its length is known from this segment's length. 
     */ 
 
    /* bytesread says how many bytes we've read so far */ 
    datalen = segment->datalen - bytesread; 
 
    segment->data.genrefreg.data = 
	(Dgp_UInt8 *)(a->malloc)(datalen); 
    DGP_CHECK_ALLOC_CLEANUP(segment->data.genrefreg.data); 
    segment->data.genrefreg.datalen = datalen; 
 
    DSREAD(data, datalen, segment->data.genrefreg.data, &bytes); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    ERR_RETURN; 
    } 
 
/* Rendering generic regions */ 
 
/* 
 * Render the segment.  You can only do this if it's refining another 
 * segment, in which case it should have a pointer to the segment it's 
 * refining (which must be an intermediate region segment).  If that segment 
 * hasn't been rendered yet, render it.  Then use the reference segment's 
 * rendered bitmap as the reference bitmap, and decode this segment into 
 * its own bitmap. 
 */ 
ErrCode 
jb2RenderGenRefReg(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2RenderGenRefReg()", jb2_pErrState(file)); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
    jb2RegSegment *reginfo; 
    DgpImage_Packed im = (DgpImage_Packed)NULL; 
    DgpEncode_MQDecoder dec; 
    Dgp_UInt8 *contexts = (Dgp_UInt8 *)NULL; 
 
    reginfo = &segment->reginfo; 
 
    /* Done already? */ 
    if (reginfo->rendered) { 
	NORMAL_RETURN; 
	} 
 
    /* The combop can be ignored since we're not combining here. */ 
 
    /* Render the segment being referred to */ 
    DGP_ASSERT(segment->data.genrefreg.refseg != (jb2Segment *)NULL); 
    DGP_ASSERT(isRegionSegtype(segment->data.genrefreg.refseg->segtype)); 
    DGP_ASSERT(!isImmediateRegion(segment->data.genrefreg.refseg->segtype)); 
    DGP_ERR_PASS_CLEANUP(jb2RenderRegion(file,  
					 segment->data.genrefreg.refseg)); 
 
    /* Allocate the buffer */ 
    DGP_ERR_PASS (DgpImage_new(jb2_pDgp_State(file), DGPIMAGE_PACKED, 
			    reginfo->regwidth, reginfo->regheight, (void**) &im)); 
 
    /* Create the context array */ 
    contexts = (Dgp_UInt8 *) 
	(a->calloc)(jb2RABContextsNeeded(&segment->data.genrefreg.refparams), 1); 
    if (contexts == (Dgp_UInt8 *)NULL) { 
	NOMEM; 
	} 
 
#ifndef	NOMQ 
    /* and the MQ decoder */ 
    DgpEncode_mqdInit(&dec, segment->data.genrefreg.data, 
	    segment->data.genrefreg.datalen, 
	    (Dgp_Bool (*)(Dgp_UInt8 **, size_t *, void *))NULL, 
	    (void *)NULL, im->dgp_state); 
 
    /* Do the decoding */ 
    refineImagePacked(im, segment->data.genrefreg.refseg->reginfo.im, 
		      0, 0, 
		      &segment->data.genrefreg.refparams, 
		      contexts, &dec); 
#endif 
 
    /* Stash the bitmap away */ 
    reginfo->im = im; 
    reginfo->rendered = DGP_TRUE; 
 
    (a->free)((void *)contexts); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (im != (DgpImage_Packed)NULL) { 
	DgpImage_free(im); 
	} 
 
    if (contexts != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)contexts); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Render the segment into the page.  You can only do this if it's an 
 * immediate (or immediate lossless) segment.  It might refer to no 
 * other segment (meaning that it's a refinement of the page), or to 
 * an intermediate segment, meaning that the refinement should be applied 
 * to that segment and the result drawn into the page. 
 * 
 * To make life easier, if it's refining the page directly, we 
 * clip out the region of the page being refined, apply the refinement, 
 * and then draw back into it. 
 */ 
ErrCode 
jb2RenderGenRefRegIntoPage(jb2File file, jb2Segment *segment, 
			   DgpImage_Packed page) 
{ 
  DGP_PROLOGUE("jb2RenderGenRefRegIntoPage()", jb2_pErrState(file)); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
    jb2RegSegment *reginfo = &segment->reginfo; 
    DgpImage_Packed pagebits = (DgpImage_Packed)NULL; 
    DgpImage_Packed im = (DgpImage_Packed)NULL; 
    DgpEncode_MQDecoder dec; 
    Dgp_UInt8 *contexts = (Dgp_UInt8 *)NULL; 
 
    /* The region must fit entirely on the page */ 
    if (((int)reginfo->regxpos + (int)reginfo->regwidth > page->header.width) || 
	((int)reginfo->regypos + (int)reginfo->regheight > page->header.height)) { 
	FILEERR(cErrCompressionError, "region extends off page"); 
	} 
 
    /* You can't render an intermediate segment into a page */ 
    DGP_ASSERT(segment->segtype != JB2SEGTYPE_GENREFREG_INT); 
 
    /* Does it apply to the page? */ 
    if (segment->nrefsegs == 0) { 
	DGP_ASSERT(segment->data.genrefreg.refseg == (jb2Segment *)NULL); 
 
	/* Verify the combination operator */ 
	if (reginfo->combop != COMBOP_REPLACE) { 
	    FILEERR(cErrCompressionError, "refinement segment applying to page needs REPLACE"); 
	    } 
 
	/* Clip a bit of the page out */ 
	DGP_ERR_PASS (DgpImage_new(jb2_pDgp_State(file), DGPIMAGE_PACKED, reginfo->regwidth, 
				      reginfo->regheight, (void**) &pagebits)); 
	DGP_CHECK_ALLOC_CLEANUP(pagebits); 
 
	/* 
	 * We can extract the region from the page image by blitting the 
	 * page image into the region, with the page origin going to 
	 * (-xpos, -ypos) in region coordinates (equivalent to the region 
	 * origin being at (xpos, ypos) in page coordinates), and letting 
	 * the blit clipper take care of the parts of the page outside 
	 * the region. 
	 */ 
	DgpBlat_blitPacked(pagebits, page, -(int)reginfo->regxpos, -(int)reginfo->regypos); 
 
	/* Do the refinement */ 
 
	/* Make the dest array */ 
	DGP_ERR_PASS (DgpImage_new(jb2_pDgp_State(file), DGPIMAGE_PACKED, reginfo->regwidth, 
				reginfo->regheight, (void**) &im)); 
	DGP_CHECK_ALLOC_CLEANUP(im); 
 
	/* Create the context array */ 
	contexts = (Dgp_UInt8 *) 
	    (a->calloc)(jb2RABContextsNeeded(&segment->data.genrefreg.refparams), 
		   1); 
	if (contexts == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
 
#ifndef	NOMQ 
	/* and the MQ decoder */ 
	DgpEncode_mqdInit(&dec, segment->data.genrefreg.data, 
		segment->data.genrefreg.datalen, 
		(Dgp_Bool (*)(Dgp_UInt8 **, size_t *, void *))NULL, 
		(void *)NULL, im->dgp_state); 
 
	/* Do the decoding */ 
	refineImagePacked(im, pagebits, 0, 0, 
			  &segment->data.genrefreg.refparams, 
			  contexts, &dec); 
#endif 
	/* Draw it back into the page buffer */ 
	DgpBlat_blitPackedOp(page, im, reginfo->regxpos, reginfo->regypos, 
		     DGPBLAT_OP_REPLACE); 
	DgpImage_free(im); 
	DgpImage_free(pagebits); 
	(a->free)((void *)contexts); 
	} 
    else { 
	/* It's a refinement of a previous region to be drawn into the page */ 
 
	/* Do the refinement */ 
      DGP_ERR_PASS_CLEANUP(jb2RenderGenRefReg(file, segment)); 
 
	/* Now we have segment->reginfo.im */ 
 
	/* Draw it into the page buffer */ 
	DgpBlat_blitPackedOp(page, segment->reginfo.im, 
		     reginfo->regxpos, reginfo->regypos, 
		     jb2Minterms[reginfo->combop]); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (pagebits != (DgpImage_Packed)NULL) { 
	DgpImage_free(pagebits); 
	} 
 
    if (im != (DgpImage_Packed)NULL) { 
	DgpImage_free(im); 
	} 
 
    if (contexts != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)contexts); 
	} 
 
    ERR_RETURN; 
    } 
 
/* Utility stuff */ 
 
/* 
 * Decode the AT pixel parameters.  Use params->rtemplate to say which 
 * ones to decode.  Check them as well. 
 * Increment *bytesread by the number of bytes read. 
 */ 
ErrCode 
jb2DecodeRABParams(jb2File file, DataSource data, 
		   arith_ref_params *params, Dgp_Int32 *bytesread) 
{ 
  DGP_PROLOGUE("jb2DecodeRABParams()", jb2_pErrState(file)); 
 
    Dgp_UInt8 cbuf; 
    Dgp_Int32 i, n; 
 
    if (params->rtemplate == 0) { 
	/* Two AT pixels */ 
	n = 2; 
	} 
    else { 
	/* No AT pixels */ 
	n = 0; 
	} 
 
    for (i = 0; i < n; i++) { 
	DSGETC(data, &cbuf); 
	params->ratx[i] = (Dgp_Int8)cbuf; 
	DSGETC(data, &cbuf); 
	params->raty[i] = (Dgp_Int8)cbuf; 
#ifdef	VALPRINT 
	fprintf(stderr, "RATX%d %d\n", i + 1, params->ratx[i]); 
	fprintf(stderr, "RATY%d %d\n", i + 1, params->raty[i]); 
#endif 
 
	*bytesread += 2; 
	} 
 
    /* Fill in remaining ones with 0 */ 
    for (; i < 2; i++) { 
	params->ratx[i] = 0; 
	params->raty[i] = 0; 
	} 
 
    if (!checkRABParams(params)) { 
	FILEERR(cErrInvalidParm, "bad RAT params"); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* 
 * Return the number of contexts that must be allocated in order to 
 * decode using those parameters. 
 */ 
size_t 
jb2RABContextsNeeded(arith_ref_params *params) 
{ 
  Dgp_Int32 tempsize = 0; 
 
  DGP_VOID_PROLOGUE ("jb2RABContextsNeeded", NULL); 
 
    /* How many pixels in the template? */ 
    switch(params->rtemplate) { 
      case 0: { 
	tempsize = 13; 
	break; 
	} 
       
      case 1: { 
	tempsize = 10; 
	break; 
	} 
       
      default: { 
	DGP_REPORTF0(DGP_ENC, "bad template size"); 
	} 
      } 
 
    /* 
     * Eventually, we might want to make allowances for AT pixels overlapping 
     * regular pixels. 
     */ 
    return(((size_t)1) << tempsize); 
    } 
 
/* 
 * Return DGP_TRUE if the params given are valid, DGP_FALSE if they aren't. 
 */ 
static Dgp_Bool 
checkRABParams(arith_ref_params *params) 
{ 
    Dgp_Int32 i; 
 
    if (params->rtemplate > 1) { 
	return(DGP_FALSE); 
	} 
 
    if (params->rtemplate == 0) { 
	/* First AT pixel location must be causal */ 
	if ((params->raty[0] > 0) || 
	    ((params->raty[0] == 0) && (params->ratx[0] >= 0))) { 
	    return(DGP_FALSE); 
	    } 
	/* Second AT pixel location is arbitrary */ 
	} 
    else { 
	/* No actual AT pixels; both locations should be (0, 0) */ 
	for (i = 0; i < 2; i++) { 
	    if ((params->ratx[i] != 0) || (params->raty[i] != 0)) { 
		return(DGP_FALSE); 
		} 
	    } 
	} 
 
    /* It checked out */ 
    return(DGP_TRUE); 
    } 
 
/* Inner routines */ 
#ifndef	NOMQ 
 
/* 
 * dec is an DgpEncode_MQDecoder containing some arithmetically-encoded data. 
 * Using reference as a reference bitmap, refine a width by height bitmap, 
 * and draw it into *bitmap. 
 * No clipping is done.  *bitmap is assumed to be clear initially.  The 
 * decoder MUST be using the software conventions. 
 * 
 * If contexts is not NULL, it must point to a block of contexts large 
 * enough to handle the template size given.  If contexts is NULL, this 
 * routine will allocate a block of contexts, clear it, use it, and discard 
 * it. 
 * 
 * For coding simplicity, it's best to wrap a DgpImage_Packed around 
 * the bitmaps and then use them. 
 */ 
ErrCode 
jb2ArithDecodeRefinement(Dgp_UInt32 *bitmap, Dgp_Int32 width, Dgp_Int32 height, 
			 Dgp_Int32 Dgp_Int32sperline, 
			 Dgp_UInt32 *reference, 
			 Dgp_Int32 refwidth, Dgp_Int32 refheight, 
			 Dgp_Int32 refDgp_Int32sperline, 
			 Dgp_Int32 refdx, Dgp_Int32 refdy, 
			 arith_ref_params *params, 
			 Dgp_UInt8 *contexts, 
			 jb2File file, DgpEncode_MQDecoder *dec) 
{ 
  DGP_PROLOGUE("jb2ArithDecodeRefinement()", jb2_pErrState(file)); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
    Dgp_Bool mycontexts = DGP_FALSE; 
    DgpImage_Packed dest = (DgpImage_Packed)NULL; 
    DgpImage_Packed ref = (DgpImage_Packed)NULL; 
    size_t cxsize; 
	Dgp_UInt32 *retval; 
 
    cxsize = jb2RABContextsNeeded(params); 
 
    /* Make sure we have a context block */ 
    if (contexts == (Dgp_UInt8 *)NULL) { 
	contexts = (Dgp_UInt8 *)(a->calloc)(cxsize, 1); 
	if (contexts == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
	mycontexts = DGP_TRUE; 
	} 
 
    /* Make our DgpImage_Packeds */ 
    DGP_ERR_PASS (DgpImage_wrap(jb2_pDgp_State(file), bitmap, width, height, Dgp_Int32sperline, &dest)); 
    DGP_ERR_PASS (DgpImage_wrap(jb2_pDgp_State(file), reference, refwidth, refheight, refDgp_Int32sperline, &ref)); 
 
    refineImagePacked(dest, ref, refdx, refdy, params, contexts, dec); 
 
    if (mycontexts) { 
	(a->free)((void *)contexts); 
	} 
 
    DgpImage_unwrap(dest, &retval); 
    DgpImage_unwrap(ref, &retval); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (mycontexts && (contexts != (Dgp_UInt8 *)NULL)) { 
	(a->free)((void *)contexts); 
	} 
 
    if (dest != (DgpImage_Packed)NULL) { 
	DgpImage_unwrap(dest, &retval); 
	} 
 
    if (ref != (DgpImage_Packed)NULL) { 
	DgpImage_unwrap(ref, &retval); 
	} 
 
    return(DGP_FALSE); 
    } 
 
/* Dispatch on the template. */ 
static void 
refineImagePacked(DgpImage_Packed dest, DgpImage_Packed ref, 
		  Dgp_Int32 refdx, Dgp_Int32 refdy, arith_ref_params *params, 
		  Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec) 
{ 
    if (params->rtemplate == 0) { 
	refineImageT13(dest, ref, refdx, refdy, params, contexts, dec); 
	} 
    else { 
	refineImageT10(dest, ref, refdx, refdy, params, contexts, dec); 
	} 
    } 
 
static void 
refineImageT13(DgpImage_Packed dest, DgpImage_Packed ref, 
	       Dgp_Int32 refdx, Dgp_Int32 refdy, arith_ref_params *params, 
	       Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec) 
{ 
    Dgp_UInt32 cx, tprcx; 
    Dgp_Int32 x, y; 
    Dgp_Int32 bit; 
    Dgp_Int32 dwidth = DGPIMG_WIDTH(dest), dheight = DGPIMG_HEIGHT(dest); 
    Dgp_Int32 rwidth = DGPIMG_WIDTH(ref), rheight = DGPIMG_HEIGHT(ref); 
    Dgp_Bool ltp = DGP_FALSE; 
    Dgp_State *dgp_state = ref->dgp_state; 
	 
    DGP_VOID_PROLOGUE ("refineImageT13", DGP_PERRSTATE); 
 
    /* 
     * The template looks like: 
     * 1XX		2XX 
     * X?		X?X 
     *			XXX 
     * where the left half is in the dest image and the right half is in 
     * the ref image.  Also, pixel (0,0) in dest corresponds to pixel 
     * (-refdx, -refdy) in ref). 
     * 
     * We're going to form a context that looks like 
     * 1aa2bbcdddeee 
     * where "aa" is the top line on L, "bb" is the top line on R, 
     * "c" is the middle line on L, "ddd" is the middle line on R, 
     * "eee" is the bottom line on R.  Updating the context involves 
     * shifting it left by 1, ANDing it with 00 1001 0011 0110 
     * and doing 7 bit loads into bits 12 (AT1), 10 (top L), 9 (AT2), 
     * 7 (top R), 6 (middle L), 3 (middle R) and 0 (bottom R). 
     * We load from (after shift coordinates) L(ATX1, ATY1), L(1,-1), 
     * R(ATX2, ATY2), R(1, -1), L(-1, 0), R(1, 0), R(1, 1). 
     * The L(-1, 0) load is special because it can be done after the decode 
     * of the previous bit. 
     * 
     * We're also going to keep track of a context for TPR, which is 
     * bbbcccddd 
     * where "bbb" includes the pixel marked "2" above, as well as the 
     * "bb" pixels.  The "2" pixel might not be at that exact location, 
     * since it's an AT pixel, so we can't just mask bits out of cx. 
     * For each bit, we shift tprcx and AND with 01 1011 0110, then 
     * load up 3 bits into positions 6 (R(1, -1)), 3 (R(1, 0)), 0(R(1, 1)). 
     */ 
 
    /* 
     * Given values in cx, x and y, and bit, shift and mask CX, 
     * store the bit, and put the bit in cx.  Also shift and mask tprcx. 
     */ 
#define	DOUPDATE \
    do { \
	DGP_REPORTF1(DGP_ENC, "%d", bit); \
	DGP_REPORTF1(DGP_ENC, "(%04lx)", cx); \
	cx = ((cx << 1) & 0x936) | (bit << 6); \
	tprcx = ((tprcx << 1) & 0x1b6); \
	DGPIMG_ORBIT(dest, x, y, bit); \
	} while(0) 
 
    for (y = 0; y < dheight; y++) { 
	/* Is TPR on? */ 
	if (params->tpron) { 
	    /* 
	     * Yes; decode a bit using context 
	     * 000		000 
	     * 0?		010 
	     *			000 
	     * which linearises to 0 0000 0001 0000 
	     */ 
	    bit = DgpEncode_mqdDecode(dec, &contexts[0x0010]); 
	    DGP_REPORTF1(DGP_ENC, "TPR%d", bit); 
	    ltp ^= bit; 
	    } 
 
	/* 
	 * Do the initial load.  The 7 bit loads happen immediately 
	 * after this, so we just need bits 0a00b00dd0ee0.  In fact, 
	 * some of the "b", "dd", "ee" bits may not exist (off one 
	 * edge or the other of ref). 
	 * For TPR, we want bb0dd0ee0. 
	 */ 
	cx = 0; 
	tprcx = 0; 
 
	if (y - 1 >= 0) { 
	    /* Load up "a" */ 
	    cx |= DGPIMG_GETBIT01(dest, 0, y - 1) << 11; 
	    } 
 
	if ((y - 1 - refdy >= 0) && (y - 1 - refdy < rheight)) { 
	    if ((-refdx - 1 >= 0) && (-refdx - 1 < rwidth)) { 
		/* Load up first "b" in TPR */ 
		tprcx |= DGPIMG_GETBIT01(ref, -refdx - 1, y - refdy - 1) << 9; 
		} 
	    if ((-refdx >= 0) && (-refdx < rwidth)) { 
		/* Load up "b" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx, y - 1 - refdy); 
		cx |= bit << 8; 
		tprcx |= bit << 8; 
		} 
	    } 
 
	if ((y - refdy >= 0) && (y - refdy < rheight)) { 
	    if ((-refdx - 1 >= 0) && (-refdx - 1 < rwidth)) { 
		/* Load up the first "d" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx - 1, y - refdy); 
		cx |= bit << 5; 
		tprcx |= bit << 5; 
		} 
	    if ((-refdx >= 0) && (-refdx < rwidth)) { 
		/* Load up the second "d" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx, y - refdy); 
		cx |= bit << 4; 
		tprcx |= bit << 4; 
		} 
	    } 
 
	if ((y + 1 - refdy >= 0) && (y + 1 - refdy < rheight)) { 
	    if ((-refdx - 1 >= 0) && (-refdx - 1 < rwidth)) { 
		/* Load up the first "e" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx - 1, y + 1 - refdy); 
		cx |= bit << 2; 
		tprcx |= bit << 2; 
		} 
	    if ((-refdx >= 0) && (-refdx < rwidth)) { 
		/* Load up the second "e" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx, y + 1 - refdy); 
		cx |= bit << 1; 
		tprcx |= bit << 1; 
		} 
	    } 
 
	for (x = 0; x < dwidth; x++) { 
	    /* Load the AT pixels */ 
	    if ((x + params->ratx[0] >= 0) && (x + params->ratx[0] < dwidth) && 
		(y + params->raty[0] >= 0)) { 
		cx |= DGPIMG_GETBIT01(dest, x + params->ratx[0], 
				  y + params->raty[0]) << 12; 
		} 
	    if ((x - refdx + params->ratx[1] >= 0) && 
		(x - refdx + params->ratx[1] < rwidth) && 
		(y - refdy + params->raty[1] >= 0) && 
		(y - refdy + params->raty[1] < rheight)) { 
		cx |= DGPIMG_GETBIT01(ref, x - refdx + params->ratx[1], 
				  y - refdy + params->raty[1]) << 9; 
		} 
 
	    /* Load top L pixel */ 
	    if ((x + 1 < dwidth) && (y - 1 >= 0)) { 
		cx |= DGPIMG_GETBIT01(dest, x + 1, y - 1) << 10; 
		} 
 
	    if ((x - refdx + 1 >= 0) && (x - refdx + 1 < rwidth)) { 
		/* Load top R pixel */ 
		if ((y - refdy - 1 >= 0) && (y - refdy - 1 < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx + 1, y - refdy - 1); 
		    cx |= bit << 7; 
		    tprcx |= bit << 6; 
		    } 
 
		/* Load middle R pixel */ 
		if ((y - refdy >= 0) && (y - refdy < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx + 1, y - refdy); 
		    cx |= bit << 3; 
		    tprcx |= bit << 3; 
		    } 
 
		/* Load bottom R pixel */ 
		if ((y - refdy + 1 >= 0) && (y - refdy + 1 < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx + 1, y - refdy + 1); 
		    cx |= bit; 
		    tprcx |= bit; 
		    } 
		} 
 
	    /* 
	     * Is this pixel typical? 
	     * If all the 9 pixels on the R side have the same value, 
	     * and ltp is DGP_TRUE, don't decode anything, just hallucinate 
	     * that shared value. 
	     */ 
	    if (ltp) { 
		if (tprcx == 0) { 
		    bit = 0; 
		    } 
		else if (tprcx == 0x1ff) { 
		    bit = 1; 
		    } 
		else { 
		    bit = DgpEncode_mqdDecode(dec, &contexts[cx]); 
		    } 
		} 
	    else { 
		bit = DgpEncode_mqdDecode(dec, &contexts[cx]); 
		} 
 
	    DOUPDATE; 
	    } 
	DGP_REPORTF1(DGP_ENC, "%c", '\n'); 
	} 
    } 
#undef	DOUPDATE 
 
static void 
refineImageT10(DgpImage_Packed dest, DgpImage_Packed ref, 
	       Dgp_Int32 refdx, Dgp_Int32 refdy, arith_ref_params *params, 
	       Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec) 
{ 
    Dgp_UInt32 cx, tprcx; 
    Dgp_Int32 x, y; 
    Dgp_Int32 bit; 
    Dgp_Int32 dwidth = DGPIMG_WIDTH(dest), dheight = DGPIMG_HEIGHT(dest); 
    Dgp_Int32 rwidth = DGPIMG_WIDTH(ref), rheight = DGPIMG_HEIGHT(ref); 
    Dgp_Bool ltp = DGP_FALSE; 
    Dgp_State *dgp_state = ref->dgp_state; 
 
    DGP_VOID_PROLOGUE("refineImageT10", DGP_PERRSTATE); 
 
    /* 
     * The template looks like: 
     * XXX		 X 
     * X?		X?X 
     *			 XX 
     * where the left half is in the dest image and the right half is in 
     * the ref image.  Also, pixel (0,0) in dest corresponds to pixel 
     * (-refdx, -refdy) in ref). 
     * 
     * We're going to form a context that looks like 
     * aaabcdddee 
     * where "aaa" is the top line on L, "b" is the top line on R, 
     * "c" is the middle line on L, "ddd" is the middle line on R, 
     * "ee" is the bottom line on R.  Updating the context involves 
     * shifting it left by 1, ANDing it with 011 0001 1010 
     * and doing 5 bit loads into bits 7 (top L), 6 (top R), 
     * 5 (middle L), 2 (middle R) and 0 (bottom R). 
     * We load from (after shift coordinates) L(1,-1), 
     * R(0, -1), L(-1, 0), R(1, 0), R(1, 1). 
     * The L(-1, 0) load is special because it can be done after the decode 
     * of the previous bit. 
     * 
     * We're also going to keep track of a context for TPR, which is 
     * bbbcccddd 
     * where "bbb" includes the pixel marked "2" above, as well as the 
     * "bb" pixels.  The "2" pixel might not be at that exact location, 
     * since it's an AT pixel, so we can't just mask bits out of cx. 
     * For each bit, we shift tprcx and AND with 01 1011 0110, then 
     * load up 3 bits into positions 6 (R(1, -1)), 3 (R(1, 0)), 0(R(1, 1)). 
     */ 
 
    /* 
     * Given values in cx, x and y, and bit, shift and mask CX, 
     * store the bit, and put the bit in cx.  Also shift and mask tprcx. 
     */ 
#define	DOUPDATE \
    do { \
	DGP_REPORTF1(DGP_ENC, "%d", bit); \
	cx = ((cx << 1) & 0x31a) | (bit << 5); \
	tprcx = ((tprcx << 1) & 0x1b6); \
	DGPIMG_ORBIT(dest, x, y, bit); \
	} while(0) 
 
    for (y = 0; y < dheight; y++) { 
	/* Is TPR on? */ 
	if (params->tpron) { 
	    /* 
	     * Yes; decode a bit using context 
	     * 000		 0  
	     * 0?		010 
	     *			 00 
	     * which linearises to 00 0000 1000 
	     */ 
	    bit = DgpEncode_mqdDecode(dec, &contexts[0x0008]); 
	    DGP_REPORTF1(DGP_ENC, "TPR%d", bit); 
	    ltp ^= bit; 
	    } 
 
	/* 
	 * Do the initial load.  The 5 bit loads happen immediately 
	 * after this, so we just need bits aa000dd0e0.  In fact, 
	 * the first "a" bit there doesn't exist (off the left edge of dest), 
	 * and some of the "dd", "e" bits may not exist (off one 
	 * edge or the other of ref). 
	 * For TPR, we want bb0dd0ee0. 
	 */ 
	cx = 0; 
	tprcx = 0; 
 
	if (y - 1 >= 0) { 
	    /* Load up second "a" */ 
	    cx |= DGPIMG_GETBIT01(dest, 0, y - 1) << 8; 
	    } 
 
	if ((y - 1 - refdy >= 0) && (y - 1 - refdy < rheight)) { 
	    if ((-refdx - 1 >= 0) && (-refdx - 1 < rwidth)) { 
		/* Load up first "b" in TPR */ 
		tprcx |= DGPIMG_GETBIT01(ref, -refdx - 1, y - refdy - 1) << 9; 
		} 
	    if ((-refdx >= 0) && (-refdx < rwidth)) { 
		/* Load up second "b" in TPR */ 
		bit = DGPIMG_GETBIT01(ref, -refdx, y - 1 - refdy); 
		tprcx |= bit << 8; 
		} 
	    } 
 
	if ((y - refdy >= 0) && (y - refdy < rheight)) { 
	    if ((-refdx - 1 >= 0) && (-refdx - 1 < rwidth)) { 
		/* Load up the first "d" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx - 1, y - refdy); 
		cx |= bit << 4; 
		tprcx |= bit << 5; 
		} 
	    if ((-refdx >= 0) && (-refdx < rwidth)) { 
		/* Load up the second "d" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx, y - refdy); 
		cx |= bit << 3; 
		tprcx |= bit << 4; 
		} 
	    } 
 
	if ((y + 1 - refdy >= 0) && (y + 1 - refdy < rheight)) { 
	    if ((-refdx - 1 >= 0) && (-refdx - 1 < rwidth)) { 
		/* Load up the first "e" in TPR */ 
		bit = DGPIMG_GETBIT01(ref, -refdx - 1, y + 1 - refdy); 
		tprcx |= bit << 2; 
		} 
	    if ((-refdx >= 0) && (-refdx < rwidth)) { 
		/* Load up the second "e" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx, y + 1 - refdy); 
		cx |= bit << 1; 
		tprcx |= bit << 1; 
		} 
	    } 
 
	for (x = 0; x < dwidth; x++) { 
	    /* Load top L pixel */ 
	    if ((x + 1 < dwidth) && (y - 1 >= 0)) { 
		cx |= DGPIMG_GETBIT01(dest, x + 1, y - 1) << 7; 
		} 
 
	    if ((x - refdx >= 0) && (x - refdx < rwidth)) { 
		/* Load top R pixel */ 
		if ((y - refdy - 1 >= 0) && (y - refdy - 1 < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx, y - refdy - 1); 
		    cx |= bit << 6; 
		    } 
		} 
 
	    if ((x - refdx + 1 >= 0) && (x - refdx + 1 < rwidth)) { 
		/* Load top R pixel for TPR */ 
		if ((y - refdy - 1 >= 0) && (y - refdy - 1 < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx + 1, y - refdy - 1); 
		    tprcx |= bit << 6; 
		    } 
 
		/* Load middle R pixel */ 
		if ((y - refdy >= 0) && (y - refdy < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx + 1, y - refdy); 
		    cx |= bit << 2; 
		    tprcx |= bit << 3; 
		    } 
 
		/* Load bottom R pixel */ 
		if ((y - refdy + 1 >= 0) && (y - refdy + 1 < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx + 1, y - refdy + 1); 
		    cx |= bit; 
		    tprcx |= bit; 
		    } 
		} 
 
	    /* 
	     * Is this pixel typical? 
	     * If all the 9 pixels on the R side have the same value, 
	     * and ltp is DGP_TRUE, don't decode anything, just hallucinate 
	     * that shared value. 
	     */ 
	    if (ltp) { 
		if (tprcx == 0) { 
		    bit = 0; 
		    } 
		else if (tprcx == 0x1ff) { 
		    bit = 1; 
		    } 
		else { 
		    bit = DgpEncode_mqdDecode(dec, &contexts[cx]); 
		    } 
		} 
	    else { 
		bit = DgpEncode_mqdDecode(dec, &contexts[cx]); 
		} 
 
	    DOUPDATE; 
	    } 
	DGP_REPORTF1(DGP_ENC, "%c", '\n'); 
	} 
    } 
#undef	DOUPDATE 
 
#endif 
