/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/decgenreg.c,v 1.32 2003/01/27 21:48:33 awilliam Exp $ 
 * 
 */ 
 
 
#include "jb2file.h" 
#include "datasource.h" 
#include "memorydatasink.h" 
#include "segment.h" 
#include "segment_util.h" 
#include "decgenreg.h" 
#include <stdlib.h> 
#include "mqcoder.h" 
#include "blit.h" 
#include "image_defs.h"
#include "PsgFix.h"
 
/* Stuff you need for the DigiPaper G4 routines */ 
#include "g4.h" 
 
#define DGP_ENC DGP_FALSE

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/decgenreg.c,v 1.32 2003/01/27 21:48:33 awilliam Exp $"); 
 
/* debug stuff */
#define DGP_MESSAGE DGP_FALSE

/* Internal routines */ 
static Dgp_Bool checkABParams(arith_bitmap_params *params); 
#ifndef	NOMQ 
static void arithDecodeImageT16(DgpImage_Packed dest, arith_bitmap_params *params, 
				Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec, 
				DgpImage_Packed mask); 
static void arithDecodeImageT13(DgpImage_Packed dest, arith_bitmap_params *params, 
				Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec, 
				DgpImage_Packed mask); 
static void arithDecodeImageT10L3(DgpImage_Packed dest, 
				  arith_bitmap_params *params, 
				  Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec, 
				  DgpImage_Packed mask); 
static void arithDecodeImageT10L2(DgpImage_Packed dest, 
				  arith_bitmap_params *params, 
				  Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec, 
				  DgpImage_Packed mask); 
#endif 
 
/* Decoding generic regions */ 
ErrCode 
jb2DecodeGenReg(jb2File file, DataSource data, 
		jb2Segment *segment, jb2Segment **refsegs) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2DecodeGenReg()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_UInt8 flags; 
    size_t datalen; 
    size_t pos; 
    size_t bytes; 
    Dgp_Int32 bytesread = 0; 
    Dgp_UInt8 byte; 
    Dgp_UInt8 lbuf[4]; 
    DataSink sink = (DataSink)NULL, tmpsink; 
    Dgp_Bool gotfirst; 
    DgpAdt_List buflist = DGP_NULLLIST; 
    DgpAdt_ListNode ln; 
    mds_membuf *buf; 
    Dgp_UInt32 realheight; 
 
    (void)refsegs; 
 
    /* Must be no referenced segments */ 
    if (segment->nrefsegs != 0) { 
	FILEERR(cErrCompressionError, "generic region segment can't refer to anything"); 
	} 
 
    /* Decode the region segment header */ 
    DGP_ERR_PASS_CLEANUP(jb2DecodeRegSegment(file, data, &segment->reginfo)); 
    bytesread += 17;		/* read 17 bytes worth of region segment header */ 
 
    /* One-byte flags field */ 
    DSREAD(data, 1, &flags, &bytes); 
    bytesread++; 
 
#ifdef	VALPRINT 
    fprintf(stderr, "GBFLAGS %02x\n", flags); 
#endif 
 
    /* MMR */ 
    segment->data.genreg.mmr = (flags & 1) ? DGP_TRUE : DGP_FALSE; 
#ifdef	NOMQ 
    if (!segment->data.genreg.mmr) { 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
	} 
#endif 
     
    /* GBTEMPLATE */ 
    segment->data.genreg.abparams.rtemplate = (flags >> 1) & 3; 
 
    /* TPON */ 
    segment->data.genreg.abparams.tpon = (flags & 8) ? DGP_TRUE : DGP_FALSE; 
 
    if (!segment->data.genreg.mmr) { 
	/* Decode the AT parameters */ 
      DGP_ERR_PASS_CLEANUP(jb2DecodeABParams(file, data,  
				 &segment->data.genreg.abparams, &bytesread)); 
	} 
 
    /* 
     * All we do now is return a pointer to the compressed data.  Since 
     * there's no way to ensure that data inside a DataSource is going 
     * to stick around, we have to allocate a buffer and copy it in. 
     * 
     * Its length is known from this segment's length. 
     */ 
 
    if (segment->datalen == 0xffffffff) { 
	/* 
	 * This is a "streaming" generic region. 
	 * What this means: the encoded didn't know the region's true height 
	 * or the number of bytes required to encode it when it started. 
	 * This is only legal for immediate generic regions. 
	 * The data is terminated by FF AC xx xx xx xx (where xxxxxxxx is the 
	 * true number of rows; must be <= number of rows specified in the 
	 * header) if it's arithmetically-coded, or 00 00 xx xx xx xx if 
	 * it's MMR-coded. 
	 */ 
	if ((segment->segtype != JB2SEGTYPE_GENREG_IMM) && 
	    (segment->segtype != JB2SEGTYPE_GENREG_IMMLL)) { 
	    FILEERR(cErrCompressionError, "non-immediate streaming generic region"); 
	    } 
 
	/* 
	 * We're going to read until we encounter the end-of-data 
	 * marker, put that into the buffer, and fix up the data length 
	 * and row count. 
	 */ 
 
	/* Need somewhere to stash the data we read. */ 
	DGP_ERR_PASS_CLEANUP(jb2NewMemoryDataSink(dgp_state, &sink)); 
 
	/* Two-state state machine to recognise the end */ 
	gotfirst = DGP_FALSE; 
	for (;;) { 
	    /* Get a byte */ 
	    DSGETC(data, &byte); 
	    /* and copy it over */ 
	    DSPUTC(sink, byte); 
 
	    /* Is it the end marker? */ 
	    if (segment->data.genreg.mmr) { 
		if (byte == 0x00) { 
		    if (gotfirst) { 
			/* Two 00 bytes in a row - end seen */ 
			break; 
			} 
		    else { 
			/* First 00 byte */ 
			gotfirst = DGP_TRUE; 
			} 
		    } 
		else { 
		    /* Not a 00 byte - reset */ 
		    gotfirst = DGP_FALSE; 
		    } 
		} 
	    else { 
		if (byte == 0xAC) { 
		    if (gotfirst) { 
			/* FF AC seen */ 
			break; 
			} 
		    else { 
			/* Random AC - ignore */ 
			} 
		    } 
		else if (byte == 0xFF) { 
		    /* FF - is AC next? */ 
		    gotfirst = DGP_TRUE; 
		    } 
		else { 
		    /* Random byte - reset */ 
		    gotfirst = DGP_FALSE; 
		    } 
		} 
	    } 
	/* We've now seen and copied up to the end.  Get the row count. */ 
	DSREAD(data, 4, lbuf, &bytes); 
	realheight = 
	    (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
 
	/* Stick the row count into the sink as well */ 
	DSWRITE(sink, 4, lbuf, &bytes); 
 
	/* Patch the height */ 
	segment->reginfo.regheight = realheight; 
 
	/* Pull the DataSink together */ 
	tmpsink = sink; 
	sink = (DataSink)NULL; 
	DGP_ERR_PASS_CLEANUP(jb2CloseMemoryDataSink(tmpsink, &buflist)); 
 
	/* Count up the total size */ 
	datalen = 0; 
	DGPADT_FOREACH (ln, buflist) { 
	    buf = (mds_membuf *)DGPADT_LIGET(ln); 
	    datalen += buf->bufsize; 
	    } 
	 
	/* Allocate the dest buffer */ 
	segment->data.genreg.data = (Dgp_UInt8 *)(a->malloc)(datalen); 
	DGP_CHECK_ALLOC_CLEANUP(segment->data.genreg.data); 
	segment->data.genreg.datalen = datalen; 
 
	/* Start copying the buffers, freeing as we go */ 
	pos = 0; 
	DGPADT_FOREACH (ln, buflist) { 
	    buf = (mds_membuf *)DGPADT_LIGET(ln); 
	    (void)memcpy(segment->data.genreg.data + pos, 
			 buf->buf, buf->bufsize); 
	    pos += buf->bufsize; 
	    (a->free)((void *)buf->buf); 
	    (a->free)((void *)buf); 
	    } 
	/* DgpAdt_List is all done */ 
	DgpAdt_liFree(buflist); 
 
	/* Patch the data length */ 
	segment->datalen = datalen + bytesread; 
	} 
    else { 
	/* bytesread says how many bytes we've read so far */ 
	datalen = segment->datalen - bytesread; 
 
	segment->data.genreg.data = (Dgp_UInt8 *)(a->malloc)(datalen); 
	DGP_CHECK_ALLOC_CLEANUP(segment->data.genreg.data); 
	segment->data.genreg.datalen = datalen; 
 
	DSREAD(data, datalen, segment->data.genreg.data, &bytes); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    if (sink != (DataSink)NULL) { 
	/* closing it may create a list; buflist should be clear */ 
	DGP_ASSERT(buflist == DGP_NULLLIST); 
	(void)jb2CloseMemoryDataSink(sink, &buflist); 
	} 
 
    if (buflist != DGP_NULLLIST) { 
	DGPADT_FOREACH (ln, buflist) { 
	    buf = (mds_membuf *)DGPADT_LIGET(ln); 
	    (a->free)((void *)buf->buf); 
	    (a->free)((void *)buf); 
	    } 
	DgpAdt_liFree(buflist); 
	} 
 
    ERR_RETURN; 
    } 
 
/* Rendering generic regions */ 
ErrCode 
jb2RenderGenReg(jb2File file, jb2Segment *segment) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2RenderGenReg()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    jb2RegSegment *reginfo; 
    DgpImage_Packed im = (DgpImage_Packed)NULL; 
    size_t datalen = segment->data.genreg.datalen; 
 
    reginfo = &segment->reginfo; 
 
    /* Done already? */ 
    if (reginfo->rendered) { 
	NORMAL_RETURN; 
	} 
 
    /* Allocate the buffer */ 
    DGP_ERR_PASS (DgpImage_new(dgp_state, DGPIMAGE_PACKED, 
			    reginfo->regwidth, reginfo->regheight, (void**) &im)); 
 
    if (segment->data.genreg.mmr) { 
	/* Render it using MMR */ 
	DGP_ERR_PASS_CLEANUPF0(jb2MMRDecodeImageMemory(DGPIMG_GETROW(im, 0), 
				     (int)reginfo->regwidth, 
				     (int)reginfo->regheight, 
				     DGPIMG_GETROWWIDTH(im) / (int)sizeof(Dgp_UInt32), 
				     0, 
				     (int)reginfo->regwidth - 1, 
				     (int)reginfo->regheight - 1, 
				     segment->data.genreg.data, 
				     &datalen, dgp_state), 
			  "can't render MMR-encoded generic region segment"); 
 
	if (datalen != 0) { 
	    /* There was stuff left over */ 
	    FILEERR(cErrCompressionError, "not all MMR-encoded data consumed"); 
	    } 
	} 
    else { 
	/* Arithmetic coded. */ 
#ifndef	NOMQ 
	DgpEncode_MQDecoder dec; 
	Dgp_UInt8 *contexts = (Dgp_UInt8 *)NULL; 
 
	/* Create the context array */ 
	contexts = 
	    (Dgp_UInt8 *) 
	    (a->calloc)(jb2ABContextsNeeded(&segment->data.genreg.abparams), 
		   1); 
	if (contexts == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
 
	/* and the MQ decoder */ 
	DgpEncode_mqdInit(&dec, segment->data.genreg.data, segment->data.genreg.datalen, 
		(Dgp_Bool (*)(Dgp_UInt8 **, size_t *, void *))NULL, 
		(void *)NULL, dgp_state); 
 
	/* Do the decoding */ 
	jb2ArithDecodeImagePacked(im, &segment->data.genreg.abparams, 
				  contexts, &dec, (DgpImage_Packed)NULL); 
 
	(a->free)((void *)contexts); 
#endif 
	} 
 
    /* Stash the bitmap away */ 
    reginfo->im = im; 
    reginfo->rendered = DGP_TRUE; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (im != (DgpImage_Packed)NULL) { 
	DgpImage_free(im); 
	} 
 
    ERR_RETURN; 
    } 
 
/* Render the generic region straight into the page */ 
ErrCode 
jb2RenderGenRegIntoPage(jb2File file, jb2Segment *segment, DgpImage_Packed page) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2RenderGenRegIntoPage()", DGP_PERRSTATE); 
 
    /* 
     * Two cases: 
     * - It's MMR encoded.  In this case, we can do the decoding right into 
     *   the page buffer, if we're drawing with OR. 
     * - It's arithmetically encoded.  In this case, rendering straight 
     *   into the page buffer is too much of a pain (got to get all the 
     *   edge cases right), so we'll call jb2RenderGenReg to render it into 
     *   a buffer, then draw it into the page with one of the blit routines. 
     */ 
    jb2RegSegment *reginfo = &segment->reginfo; 
    size_t datalen = segment->data.genreg.datalen; 
 
    /* The region must fit entirely on the page */ 
    if (((int)reginfo->regxpos + (int)reginfo->regwidth > DGPIMG_WIDTH(page)) || 
	((int)reginfo->regypos + (int)reginfo->regheight > DGPIMG_HEIGHT(page))) { 
	FILEERR(cErrCompressionError, "region extends off page"); 
	} 
 
    /* You can't render an intermediate segment into a page */ 
    DGP_ASSERT(segment->segtype != JB2SEGTYPE_GENREG_INT); 
 
    if ((segment->data.genreg.mmr) && (reginfo->combop == COMBOP_OR)) { 
	/* Optimise the OR case of MMR - draw it as we decode */ 
	DGP_ERR_PASS_CLEANUPF0(jb2MMRDecodeImageMemory(DGPIMG_GETROW(page, reginfo->regypos), 
				     (int)reginfo->regwidth, 
				     (int)reginfo->regheight, 
				     DGPIMG_GETROWWIDTH(page) / (int)sizeof(Dgp_UInt32), 
				     (int)reginfo->regxpos, 
				     DGPIMG_WIDTH(page), 
				     DGPIMG_HEIGHT(page) - reginfo->regypos, 
				     segment->data.genreg.data, 
				     &datalen, dgp_state), 
			 "can't render MMR-encoded generic region segment"); 
 
	if (datalen != 0) { 
	    /* There was stuff left over */ 
	    FILEERR(cErrCompressionError, "not all MMR-encoded data consumed"); 
	    } 
	} 
    else { 
	/* Arithmetic coded, or MMR and not drawn with OR. */ 
	/* Render it */ 
      DGP_ERR_PASS_CLEANUP(jb2RenderGenReg(file, segment)); 
 
	/* and blit it */ 
	DgpBlat_blitPackedOp(page, reginfo->im, 
		     (int)reginfo->regxpos, (int)reginfo->regypos, 
		     jb2Minterms[reginfo->combop]); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* Utility routines */ 
 
/* 
 * Decode the AT pixel parameters.  Use params->rtemplate to say which 
 * ones to decode.  Check them as well. 
 * Increment *bytesread by the number of bytes read. 
 */ 
ErrCode 
jb2DecodeABParams(jb2File file, DataSource data, 
		  arith_bitmap_params *params, Dgp_Int32 *bytesread) 
{ 
  DGP_PROLOGUE("jb2DecodeABParams()", jb2_pErrState(file)); 
 
    Dgp_UInt8 cbuf; 
    Dgp_Int32 i, n; 
 
    /* Read the AT pixel flags */ 
    if (params->rtemplate != 0) { 
	/* Always read GBATX1, GBATY1 */ 
	n = 1; 
	} 
    else { 
	/* Read GBATX2-4, GBATY2-4 if TEMPLATE is 0 */ 
	n = 4; 
	} 
 
    for (i = 0; i < n; i++) { 
	DSGETC(data, &cbuf); 
	params->atx[i] = (Dgp_Int8)cbuf; 
	DSGETC(data, &cbuf); 
	params->aty[i] = (Dgp_Int8)cbuf; 
#ifdef	VALPRINT 
	fprintf(stderr, "ATX%d %d\n", i + 1, params->atx[i]); 
	fprintf(stderr, "ATY%d %d\n", i + 1, params->aty[i]); 
#endif 
 
	*bytesread += 2; 
	} 
 
    /* Fill in remaining ones with 0 */ 
    for (; i < 4; i++) { 
	params->atx[i] = 0; 
	params->aty[i] = 0; 
	} 
 
    if (!checkABParams(params)) { 
	FILEERR(cErrInvalidParm, "bad AT params"); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
/* 
 * Return DGP_TRUE if the params given are valid, DGP_FALSE if they aren't. 
 */ 
static Dgp_Bool 
checkABParams(arith_bitmap_params *params) 
{ 
    Dgp_Int32 i; 
 
    if (params->rtemplate > 3) { 
	return(DGP_FALSE); 
	} 
 
    if (params->rtemplate == 0) { 
	/* Check all four AT positions: must be causal */ 
	for (i = 0; i < 4; i++) { 
	    if ((params->aty[i] > 0) || 
		((params->aty[i] == 0) && (params->atx[i] >= 0))) { 
		return(DGP_FALSE); 
		} 
	    } 
	} 
    else { 
	/* Check the first AT position; other 3 must be (0,0) */ 
	if ((params->aty[0] > 0) || 
	    ((params->aty[0] == 0) && (params->atx[0] >= 0))) { 
	    return(DGP_FALSE); 
	    } 
	for (i = 1; i < 4; i++) { 
	    if ((params->atx[i] != 0) || (params->aty[i] != 0)) { 
		return(DGP_FALSE); 
		} 
	    } 
	} 
 
    /* It checked out */ 
    return(DGP_TRUE); 
    } 
 
/* Inner routines */ 
 
#define	cDummyMarker	0x7FFFFFFF 
 
/* 
 * data is a DataSource containing some MMR encoded data. 
 * Decode a width by height bitmap from that, and draw it (using 
 * OR) into *bitmap, starting at xoffset.  Clip any parts that 
 * would go beyond (xmax, ymax). 
 */ 
ErrCode 
jb2MMRDecodeImage(Dgp_UInt32 *bitmap, Dgp_Int32 width, Dgp_Int32 height, 
		  Dgp_Int32 Dgp_Int32sperline, Dgp_Int32 xoffset, Dgp_Int32 xmax, Dgp_Int32 ymax, 
		  jb2File file, DataSource data, size_t datalen) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2MMRDecodeImage()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_UInt8 *coded_data = (Dgp_UInt8 *)NULL; 
    size_t bytes; 
 
    /* 
     * The underlying G4 code really wants some memory, not a DataSource. 
     */ 
 
    coded_data = (Dgp_UInt8 *)(a->malloc)(datalen); 
    if (coded_data == (Dgp_UInt8 *)NULL) { 
	NOMEM; 
	} 
 
    /* Copy the DataSource into it */ 
    DSREAD(data, datalen, coded_data, &bytes); 
     
    /* Now decode it */ 
 
    DGP_ERR_PASS_CLEANUPF0(jb2MMRDecodeImageMemory(bitmap, width, height, 
						 Dgp_Int32sperline, 
				 xoffset, xmax, ymax, 
				 coded_data, &datalen, dgp_state), 
			   "bad G4 data"); 
 
    if (datalen != 0) { 
	FILEERR(cErrCompressionError, "not all MMR-encoded data consumed"); 
	} 
 
    (a->free)((void *)coded_data); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (coded_data != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)coded_data); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Do MMR decoding.  The decoded bitmap will be ORed with bitmap. 
 * *datalen will be adjusted to reflect the amount of data left. 
 */ 
ErrCode 
jb2MMRDecodeImageMemory(Dgp_UInt32 *bitmap, Dgp_Int32 width, Dgp_Int32 height, 
			Dgp_Int32 Dgp_Int32sperline, Dgp_Int32 xoffset, Dgp_Int32 xmax, Dgp_Int32 ymax, 
			Dgp_UInt8 *data, size_t *datalen, 
			Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jb2MMRDecodeImageMemory()", DGP_PERRSTATE); 
 
    size_t consumed = 0; 
 
    DGP_ERR_PASS_CLEANUP(g4Decode(bitmap, width, height, Dgp_Int32sperline, 
				  xoffset, xmax, ymax, data, *datalen, 
				  DGP_TRUE, &consumed, dgp_state)); 
 
	*datalen -= consumed; 
 
	NORMAL_RETURN; 
 
ERR_CLEANUP: 
	*datalen -= consumed; 
 
	ERR_RETURN; 
    } 
 
#ifndef	NOMQ 
 
/* 
 * dec is an DgpEncode_MQDecoder containing some arithmetically-encoded data. 
 * Decode a width by height bitmap from that, and draw it into *bitmap. 
 * No clipping is done.  *bitmap is assumed to be clear initially.  The 
 * decoder MUST be using the software conventions. 
 * 
 * If contexts is not NULL, it must point to a block of contexts large 
 * enough to handle the template size given.  If contexts is NULL, this 
 * routine will allocate a block of contexts, clear it, use it, and discard 
 * it. 
 * 
 * For coding simplicity, it's best to wrap a DgpImage_Packed around 
 * the destination bitmap and then use that. 
 */ 
ErrCode 
jb2ArithDecodeImage(Dgp_UInt32 *bitmap, Dgp_Int32 width, Dgp_Int32 height, 
		    Dgp_Int32 Dgp_Int32sperline, arith_bitmap_params *params, 
		    Dgp_UInt8 *contexts, 
		    jb2File file, DgpEncode_MQDecoder *dec) 
{ 
  DGP_PROLOGUE("jb2ArithDecodeImage()", jb2_pErrState(file)); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
    Dgp_Bool mycontexts = DGP_FALSE; 
    DgpImage_Packed dest = (DgpImage_Packed)NULL; 
    size_t cxsize; 
	Dgp_UInt32 *retval; 
 
    cxsize = jb2ABContextsNeeded(params); 
 
    /* Make sure we have a context block */ 
    if (contexts == (Dgp_UInt8 *)NULL) { 
	contexts = (Dgp_UInt8 *)(a->calloc)(cxsize, 1); 
	if (contexts == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
	mycontexts = DGP_TRUE; 
	} 
 
    /* Make our DgpImage_Packed */ 
    DGP_ERR_PASS (DgpImage_wrap(jb2GetFileDgp_State(file), bitmap, width, height, Dgp_Int32sperline, &dest)); 
 
    jb2ArithDecodeImagePacked(dest, params, contexts, dec, (DgpImage_Packed)NULL); 
 
    if (mycontexts) { 
	(a->free)((void *)contexts); 
	} 
 
    DgpImage_unwrap(dest, &retval); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (mycontexts && (contexts != (Dgp_UInt8 *)NULL)) { 
	(a->free)((void *)contexts); 
	} 
 
    if (dest != (DgpImage_Packed)NULL) { 
	DgpImage_unwrap(dest, &retval); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * The same, but we start with contexts and a DgpImage_Packed. 
 * Also, we have the option of being passed a mask: where the mask is 1, 
 * we don't decode the corresponding pixel.  The mask may be NULL. 
 */ 
void 
jb2ArithDecodeImagePacked(DgpImage_Packed dest, arith_bitmap_params *params, 
			  Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec, 
			  DgpImage_Packed mask) 
{ 
    /* Call the inner routine to do the decoding */ 
    switch(params->rtemplate) { 
      case 0: { 
	arithDecodeImageT16(dest, params, contexts, dec, mask); 
	break; 
	} 
       
      case 1: { 
	arithDecodeImageT13(dest, params, contexts, dec, mask); 
	break; 
	} 
       
      case 2: { 
	arithDecodeImageT10L3(dest, params, contexts, dec, mask); 
	break; 
	} 
       
      case 3: { 
	arithDecodeImageT10L2(dest, params, contexts, dec, mask); 
	break; 
	} 
      } 
    } 
 
/* 
 * Do the 16-pixel template decode. 
 */ 
static void 
arithDecodeImageT16(DgpImage_Packed dest, arith_bitmap_params *params, 
		    Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec, 
		    DgpImage_Packed mask) 
{ 
    Dgp_UInt32 cx; 
    Dgp_Int32 x, y; 
    Dgp_Int32 minaty, minatx, maxatx; 
    Dgp_Int32 bit; 
    Dgp_Int32 width = DGPIMG_WIDTH(dest), height = DGPIMG_HEIGHT(dest); 
    Dgp_Bool lntp = DGP_TRUE; 
    Dgp_State *dgp_state = dec->dgp_state; 
 
    DGP_VOID_PROLOGUE ("arithDecodeImageT16", DGP_PERRSTATE); 
 
    /* 
     * How to do this efficiently, especially given AT pixels? 
     * 
     * The template looks like: 
     *   4XXX3 
     *  2XXXXX1 
     * XXXX? 
     * where 4321 are the "nominal" locations of the AT pixels.  If we 
     * wanted to, we could special-case those locations, and have that 
     * be faster than the general case.  Maybe sometime in the future... 
     * for now, we'll do the most general thing. 
     * 
     * We have a "context" value that looks like: 
     * 4aaa 32bb bbb1 cccc 
     * where 4321 are the AT pixel values, aaa are the (LtoR) values from the 
     * top row of the template, bbbbb are the values from the middle row, and 
     * cccc are the values from the bottom row.  To go to the next pixel, we 
     * shift the context left one bit, AND it with 0 0110 0011 1100 1110 
     * (clearing out the bits that landed in the wrong place), and perform 
     * 7 bit-loads into bits 0 (bottom row), 4 (AT1), 5 (middle), 10 (AT2), 
     * 11 (AT3), 12 (top), 15 (AT4) 
     * We load from (after shift coordinates) (-1,0), (2,-1), (1,-2), 
     * (ATX1, ATY1), ... 
     * 
     * Unfortunately, we have to store the data a bit at a time as we read 
     * it, or the AT pixel references may mess up.  This routine isn't going 
     * to be speedy. 
     * 
     * We do want to optimise a couple of special cases: when the inner loop 
     * doesn't need conditionals, don't put any in. 
     */ 
 
    /* What's the smallest Y offset used? */ 
    minaty = MIN(params->aty[0], params->aty[1]); 
    minaty = MIN(minaty, params->aty[2]); 
    minaty = MIN(minaty, params->aty[3]); 
    minaty = MIN(minaty, -2);	/* -2 is always used */ 
 
    /* smallest and largest X offset used? */ 
    minatx = MIN(params->atx[0], params->atx[1]); 
    minatx = MIN(minatx, params->atx[2]); 
    minatx = MIN(minatx, params->atx[3]); 
    minatx = MIN(minatx, 1);	/* 1 is always used; -1 is special case */ 
    maxatx = MAX(params->atx[0], params->atx[1]); 
    maxatx = MAX(maxatx, params->atx[2]); 
    maxatx = MAX(maxatx, params->atx[3]); 
    maxatx = MAX(maxatx, 2);	/* 2 is always used */ 
 
    /* 
     * Given values in cx, x and y, decode a bit, shift and mask CX, 
     * store the bit, and put the bit in CX 
     */ 
#define	DODECODE \
    do { \
	bit = DgpEncode_mqdDecode(dec, &contexts[cx]); \
	DGP_REPORTF1(DGP_MESSAGE,"@(%4d, ", x); \
	DGP_REPORTF1(DGP_MESSAGE,"%4d) ", y); \
	DGP_REPORTF1(DGP_MESSAGE,"%d", bit); \
	DGP_REPORTF1(DGP_MESSAGE,"(%04lx)", cx); \
	DGP_REPORTF1(DGP_MESSAGE,"[%02x]", contexts[cx]); \
	cx = ((cx << 1) & 0x63ce) | bit; \
	DGPIMG_ORBIT(dest, x, y, bit); \
	} while (0) 
 
#define	DODECODEMASK \
    do { \
	if ((mask == (DgpImage_Packed)NULL) || (DGPIMG_GETBIT01(mask, x, y) == 0)) { \
	    DODECODE; \
	    } \
	else { \
	    cx = ((cx << 1) & 0x63ce); \
	    } \
	} while (0) 
 
    for (y = 0; y < height; y++) { 
	/* Is TP on? */ 
	if (params->tpon) { 
	    /* 
	     * Yes; decode a bit using context 
	     *   10011 
	     *  0110010 
	     * 0101? 
	     * which linearises to 1001 1011 0010 0101 
	     */ 
	    bit = DgpEncode_mqdDecode(dec, &contexts[0x9b25]); 
	    DGP_REPORTF1(DGP_MESSAGE,"TP%d", bit); 
	    lntp ^= bit; 
 
	    if (lntp == 0) { 
		/* It's a line copy */ 
		Dgp_UInt32 *p, *q; 
		if (y > 0) { 
		    for (x = 0, p = DGPIMG_GETROW(dest, y), 
			     q = DGPIMG_GETROW(dest, y - 1); 
			 x < DGPIMG_GETROWWIDTH(dest) / (int)sizeof(Dgp_UInt32); 
			 x++) { 
			*p++ = *q++; 
			} 
		    } 
		DGP_REPORTF1(DGP_MESSAGE,"%c", '\n'); 
		continue;	/* Done with this line */ 
		} 
	    } 
 
	/* Do the initial load */ 
	if (y >= 2) { 
	    /* 
	     * Normal case: load up 00aa 0000 bbb0 0000 
	     * where the "a" and "b" come from the lines above: 3 pixels 
	     * from above line, 2 from line above that. 
	     */ 
	    cx = ((*DGPIMG_GETROW(dest, y - 1) >> (DGP_BITS_UINT32 - 3)) << 5) | 
		((*DGPIMG_GETROW(dest, y - 2) >> (DGP_BITS_UINT32 - 2)) << 12); 
	    } 
	else if (y == 1) { 
	    /* Only get the stuff from the line above; rest zeros */ 
	    cx = ((*DGPIMG_GETROW(dest, y - 1) >> (DGP_BITS_UINT32 - 3)) << 5); 
	    } 
	else { 
	    /* Only zeros */ 
	    cx = 0; 
	    } 
 
	if ((mask == (DgpImage_Packed)NULL) && (y + minaty >= 0)) { 
	    /* No conditionals needed for y */ 
	    for (x = 0; x < MIN(-minatx, width); x++) { 
		/* Need conditionals for x */ 
 
		/* Load the AT pixels */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				      y + params->aty[0]) << 4; 
		    } 
		if ((x + params->atx[1] >= 0) && (x + params->atx[1] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[1], 
				      y + params->aty[1]) << 10; 
		    } 
		if ((x + params->atx[2] >= 0) && (x + params->atx[2] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[2], 
				      y + params->aty[2]) << 11; 
		    } 
		if ((x + params->atx[3] >= 0) && (x + params->atx[3] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[3], 
				      y + params->aty[3]) << 15; 
		    } 
 
		/* Current row pixel is already loaded */ 
		/* Load previous row pixel */ 
		if (x + 2 < width) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 2, y - 1) << 5; 
		    } 
		/* Load two-back row pixel */ 
		if (x + 1 < width) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 1, y - 2) << 12; 
		    } 
 
		DODECODE; 
		} 
 
	    for ( ; x + maxatx < width; x++) { 
		/* x + minatx >= 0; x + maxatx < width; no conditionals needed */ 
		cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				  y + params->aty[0]) << 4; 
		cx |= DGPIMG_GETBIT01(dest, x + params->atx[1], 
				  y + params->aty[1]) << 10; 
		cx |= DGPIMG_GETBIT01(dest, x + params->atx[2], 
				  y + params->aty[2]) << 11; 
		cx |= DGPIMG_GETBIT01(dest, x + params->atx[3], 
				  y + params->aty[3]) << 15; 
		cx |= DGPIMG_GETBIT01(dest, x + 2, y - 1) << 5; 
		cx |= DGPIMG_GETBIT01(dest, x + 1, y - 2) << 12; 
		DODECODE; 
		} 
	    for ( ; x < width; x++) { 
		/* Need conditionals again */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				      y + params->aty[0]) << 4; 
		    } 
		if ((x + params->atx[1] >= 0) && (x + params->atx[1] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[1], 
				      y + params->aty[1]) << 10; 
		    } 
		if ((x + params->atx[2] >= 0) && (x + params->atx[2] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[2], 
				      y + params->aty[2]) << 11; 
		    } 
		if ((x + params->atx[3] >= 0) && (x + params->atx[3] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[3], 
				      y + params->aty[3]) << 15; 
		    } 
 
		/* Current row pixel is already loaded */ 
		/* Load previous row pixel */ 
		if (x + 2 < width) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 2, y - 1) << 5; 
		    } 
		/* Load two-back row pixel */ 
		if (x + 1 < width) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 1, y - 2) << 12; 
		    } 
 
		DODECODE; 
		} 
	    } 
	else { 
	    /* Need conditionals in y; might as well just do them everywhere */ 
	    for (x = 0; x < width; x++) { 
		if ((y + params->aty[0] >= 0) && (x + params->atx[0] >= 0) && 
		    (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				      y + params->aty[0]) << 4; 
		    } 
		if ((y + params->aty[1] >= 0) && (x + params->atx[1] >= 0) && 
		    (x + params->atx[1] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[1], 
				      y + params->aty[1]) << 10; 
		    } 
		if ((y + params->aty[2] >= 0) && (x + params->atx[2] >= 0) && 
		    (x + params->atx[2] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[2], 
				      y + params->aty[2]) << 11; 
		    } 
		if ((y + params->aty[3] >= 0) && (x + params->atx[3] >= 0) && 
		    (x + params->atx[3] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[3], 
				      y + params->aty[3]) << 15; 
		    } 
 
		/* Current row pixel is already loaded */ 
		/* Load previous row pixel */ 
		if ((y - 1 >= 0) && (x + 2 < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 2, y - 1) << 5; 
		    } 
		/* Load two-back row pixel */ 
		if ((y - 2 >= 0) && (x + 1 < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 1, y - 2) << 12; 
		    } 
 
		DODECODEMASK; 
		} 
	    } 
	DGP_REPORTF1(DGP_MESSAGE,"%c", '\n'); 
	} 
    } 
#undef	DODECODE 
#undef	DODECODEMASK 
 
/* 
 * Do the 13-pixel template decode. 
 */ 
static void 
arithDecodeImageT13(DgpImage_Packed dest, arith_bitmap_params *params, 
		    Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec, DgpImage_Packed mask) 
{ 
    Dgp_UInt32 cx; 
    Dgp_Int32 x, y; 
    Dgp_Int32 minaty, minatx, maxatx; 
    Dgp_Int32 bit; 
    Dgp_Int32 width = DGPIMG_WIDTH(dest), height = DGPIMG_HEIGHT(dest); 
    Dgp_Bool lntp = DGP_TRUE; 
    Dgp_State *dgp_state = dec->dgp_state; 
 
    DGP_VOID_PROLOGUE ("arithDecodeImageT13", DGP_PERRSTATE); 
     
	/* 
     * The template looks like: 
     *    XXXX 
     *   XXXXX1 
     *  XXX? 
     * 
     * We have a "context" value that looks like: 
     * a aaab bbbb 1ccc 
     * where 1 is the AT pixel value, aaaa are the (LtoR) values from the 
     * top row of the template, bbbbb are the values from the middle row, and 
     * ccc are the values from the bottom row.  To go to the next pixel, we 
     * shift the context left one bit, AND it with 01 1101 1110 0110 
     * (clearing out the bits that landed in the wrong place), and perform 
     * 4 bit-loads into bits 0 (bottom row), 3 (AT1), 4 (middle), 9 (top) 
     * We load from (after shift coordinates) (-1,0), (2,-1), (2,-2), 
     * (ATX1, ATY1) 
     */ 
 
    /* What's the smallest Y offset used? */ 
    minaty = MIN(params->aty[0], -2);	/* -2 is always used */ 
 
    /* smallest and largest X offset used? */ 
    minatx = MIN(params->atx[0], 2); /* 2 is always used; -1 is special case */ 
    maxatx = MAX(params->atx[0], 2); /* 2 is always used */ 
 
    /* 
     * Given values in cx, x and y, decode a bit, shift and mask CX, 
     * store the bit, and put the bit in CX 
     */ 
#define	DODECODE \
    do { \
	bit = DgpEncode_mqdDecode(dec, &contexts[cx]); \
	DGP_REPORTF1(DGP_MESSAGE,"@(%4d, ", x); \
	DGP_REPORTF1(DGP_MESSAGE,"%4d) ", y); \
	DGP_REPORTF1(DGP_MESSAGE,"%d", bit); \
	DGP_REPORTF1(DGP_MESSAGE,"(%04lx)", cx); \
	DGP_REPORTF1(DGP_MESSAGE,"[%02x]", contexts[cx]); \
	cx = ((cx << 1) & 0x1de6) | bit; \
	DGPIMG_ORBIT(dest, x, y, bit); \
	} while (0) 
 
#define	DODECODEMASK \
    do { \
	if ((mask == (DgpImage_Packed)NULL) || (DGPIMG_GETBIT01(mask, x, y) == 0)) { \
	    DODECODE; \
	    } \
	else { \
	    cx = ((cx << 1) & 0x1de6); \
	    } \
	} while (0) 
 
    for (y = 0; y < height; y++) { 
	/* Is TP on? */ 
	if (params->tpon) { 
	    /* 
	     * Yes; decode a bit using context 
	     *   0011 
	     *  110010 
	     * 101? 
	     * which linearises to 0 0111 1001 0101 
	     */ 
	    bit = DgpEncode_mqdDecode(dec, &contexts[0x0795]); 
	    DGP_REPORTF1(DGP_MESSAGE,"TP%d", bit); 
	    lntp ^= bit; 
 
	    if (lntp == 0) { 
		/* It's a line copy */ 
		Dgp_UInt32 *p, *q; 
		if (y > 0) { 
		    for (x = 0, p = DGPIMG_GETROW(dest, y), 
			     q = DGPIMG_GETROW(dest, y - 1); 
			 x < DGPIMG_GETROWWIDTH(dest) / (int)sizeof(Dgp_UInt32); 
			 x++) { 
			*p++ = *q++; 
			} 
		    } 
		DGP_REPORTF1(DGP_MESSAGE,"%c", '\n'); 
		continue;	/* Done with this line */ 
		} 
	    } 
 
	/* Do the initial load */ 
	if (y >= 2) { 
	    /* 
	     * Normal case: load up 0 aaa0 0bbb 0000 
	     * where the "a" and "b" come from the lines above: 3 pixels 
	     * from above line, 3 from line above that. 
	     */ 
	    cx = ((*DGPIMG_GETROW(dest, y - 1) >> (DGP_BITS_UINT32 - 3)) << 4) | 
		((*DGPIMG_GETROW(dest, y - 2) >> (DGP_BITS_UINT32 - 3)) << 9); 
	    } 
	else if (y == 1) { 
	    /* Only get the stuff from the line above; rest zeros */ 
	    cx = ((*DGPIMG_GETROW(dest, y - 1) >> (DGP_BITS_UINT32 - 3)) << 4); 
	    } 
	else { 
	    /* Only zeros */ 
	    cx = 0; 
	    } 
 
	if ((mask == (DgpImage_Packed)NULL) && (y + minaty >= 0)) { 
	    /* No conditionals needed for y */ 
	    for (x = 0; x < MIN(-minatx, width); x++) { 
		/* Need conditionals for x */ 
 
		/* Load the AT pixel1 */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				      y + params->aty[0]) << 3; 
		    } 
 
		/* Current row pixel is already loaded */ 
		if (x + 2 < width) { 
		    /* Load previous row pixel */ 
		    cx |= DGPIMG_GETBIT01(dest, x + 2, y - 1) << 4; 
		    /* Load two-back row pixel */ 
		    cx |= DGPIMG_GETBIT01(dest, x + 2, y - 2) << 9; 
		    } 
 
		DODECODE; 
		} 
 
	    for ( ; x + maxatx < width; x++) { 
		/* x + minatx >= 0; x + maxatx < width; no conditionals needed */ 
		cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				  y + params->aty[0]) << 3; 
		cx |= DGPIMG_GETBIT01(dest, x + 2, y - 1) << 4; 
		cx |= DGPIMG_GETBIT01(dest, x + 2, y - 2) << 9; 
		DODECODE; 
		} 
	    for ( ; x < width; x++) { 
		/* Need conditionals again */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				      y + params->aty[0]) << 3; 
		    } 
		if (x + 2 < width) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 2, y - 1) << 4; 
		    cx |= DGPIMG_GETBIT01(dest, x + 2, y - 2) << 9; 
		    } 
 
		DODECODE; 
		} 
	    } 
	else { 
	    /* Need conditionals in y; might as well just do them everywhere */ 
	    for (x = 0; x < width; x++) { 
		if ((y + params->aty[0] >= 0) && (x + params->atx[0] >= 0) && 
		    (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				      y + params->aty[0]) << 3; 
		    } 
 
		/* Current row pixel is already loaded */ 
		/* Load previous row pixel */ 
		if ((y - 1 >= 0) && (x + 2 < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 2, y - 1) << 4; 
		    } 
		/* Load two-back row pixel */ 
		if ((y - 2 >= 0) && (x + 2 < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 2, y - 2) << 9; 
		    } 
 
		DODECODEMASK; 
		} 
	    } 
	DGP_REPORTF1(DGP_MESSAGE,"%c", '\n'); 
	} 
    } 
#undef	DODECODE 
#undef	DODECODEMASK 
 
/* 
 * Do the 10-pixel three-line template decode. 
 */ 
static void 
arithDecodeImageT10L3(DgpImage_Packed dest, arith_bitmap_params *params, 
		      Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec, 
		      DgpImage_Packed mask) 
{ 
    Dgp_UInt32 cx; 
    Dgp_Int32 x, y; 
    Dgp_Int32 minaty, minatx, maxatx; 
    Dgp_Int32 bit; 
    Dgp_Int32 width = DGPIMG_WIDTH(dest), height = DGPIMG_HEIGHT(dest); 
    Dgp_Bool lntp = DGP_TRUE; 
    Dgp_State *dgp_state = dec->dgp_state; 
 
    DGP_VOID_PROLOGUE ("arithDecodeImageT10L3", DGP_PERRSTATE); 
 
    /* 
     * The template looks like: 
     *   XXX 
     *  XXXX1 
     *  XX? 
     * 
     * We have a "context" value that looks like: 
     * aa abbb b1cc 
     * where 1 is the AT pixel value, aaa are the (LtoR) values from the 
     * top row of the template, bbbb are the values from the middle row, and 
     * cc are the values from the bottom row.  To go to the next pixel, we 
     * shift the context left one bit, AND it with 011 0111 0010 
     * (clearing out the bits that landed in the wrong place), and perform 
     * 4 bit-loads into bits 0 (bottom row), 2 (AT1), 3 (middle), 7 (top) 
     * We load from (after shift coordinates) (-1,0), (1,-1), (1,-2), 
     * (ATX1, ATY1) 
     */ 
 
    /* What's the smallest Y offset used? */ 
    minaty = MIN(params->aty[0], -2);	/* -2 is always used */ 
 
    /* smallest and largest X offset used? */ 
    minatx = MIN(params->atx[0], 1); /* 1 is always used; -1 is special case */ 
    maxatx = MAX(params->atx[0], 1); /* 1 is always used */ 
 
    /* 
     * Given values in cx, x and y, decode a bit, shift and mask CX, 
     * store the bit, and put the bit in CX 
     */ 
#define	DODECODE \
    do { \
	bit = DgpEncode_mqdDecode(dec, &contexts[cx]); \
	DGP_REPORTF1(DGP_MESSAGE,"@(%4d, ", x); \
	DGP_REPORTF1(DGP_MESSAGE,"%4d) ", y); \
	DGP_REPORTF1(DGP_MESSAGE,"%d", bit); \
	DGP_REPORTF1(DGP_MESSAGE,"(%03lx)", cx); \
	DGP_REPORTF1(DGP_MESSAGE,"[%02x]", contexts[cx]); \
	cx = ((cx << 1) & 0x372) | bit; \
	DGPIMG_ORBIT(dest, x, y, bit); \
	} while (0) 
 
#define	DODECODEMASK \
    do { \
	if ((mask == (DgpImage_Packed)NULL) || (DGPIMG_GETBIT01(mask, x, y) == 0)) { \
	    DODECODE; \
	    } \
	else { \
	    cx = ((cx << 1) & 0x372); \
	    } \
	} while (0) 
 
    for (y = 0; y < height; y++) { 
	/* Is TP on? */ 
	if (params->tpon) { 
	    /* 
	     * Yes; decode a bit using context 
	     *  001 
	     * 11001 
	     * 01? 
	     * which linearises to 00 1110 0101 
	     */ 
	    bit = DgpEncode_mqdDecode(dec, &contexts[0x0e5]); 
	    DGP_REPORTF1(DGP_MESSAGE,"TP%d", bit); 
	    lntp ^= bit; 
 
	    if (lntp == 0) { 
		/* It's a line copy */ 
		Dgp_UInt32 *p, *q; 
		if (y > 0) { 
		    for (x = 0, p = DGPIMG_GETROW(dest, y), 
			     q = DGPIMG_GETROW(dest, y - 1); 
			 x < DGPIMG_GETROWWIDTH(dest) / (int)sizeof(Dgp_UInt32); 
			 x++) { 
			*p++ = *q++; 
			} 
		    } 
		DGP_REPORTF1(DGP_MESSAGE,"%c", '\n'); 
		continue;	/* Done with this line */ 
		} 
	    } 
 
	/* Do the initial load */ 
	if (y >= 2) { 
	    /* 
	     * Normal case: load up 0a a00b b000 
	     * where the "a" and "b" come from the lines above: 2 pixels 
	     * from above line, 2 from line above that. 
	     */ 
	    cx = ((*DGPIMG_GETROW(dest, y - 1) >> (DGP_BITS_UINT32 - 2)) << 3) | 
		((*DGPIMG_GETROW(dest, y - 2) >> (DGP_BITS_UINT32 - 2)) << 7); 
	    } 
	else if (y == 1) { 
	    /* Only get the stuff from the line above; rest zeros */ 
	    cx = ((*DGPIMG_GETROW(dest, y - 1) >> (DGP_BITS_UINT32 - 2)) << 3); 
	    } 
	else { 
	    /* Only zeros */ 
	    cx = 0; 
	    } 
 
	if ((mask == (DgpImage_Packed)NULL) && (y + minaty >= 0)) { 
	    /* No conditionals needed for y */ 
	    for (x = 0; x < MIN(-minatx, width); x++) { 
		/* Need conditionals for x */ 
 
		/* Load the AT pixel1 */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				      y + params->aty[0]) << 2; 
		    } 
 
		/* Current row pixel is already loaded */ 
		if (x + 1 < width) { 
		    /* Load previous row pixel */ 
		    cx |= DGPIMG_GETBIT01(dest, x + 1, y - 1) << 3; 
		    /* Load two-back row pixel */ 
		    cx |= DGPIMG_GETBIT01(dest, x + 1, y - 2) << 7; 
		    } 
 
		DODECODE; 
		} 
 
	    for ( ; x + maxatx < width; x++) { 
		/* x + minatx >= 0; x + maxatx < width; no conditionals needed */ 
		cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				  y + params->aty[0]) << 2; 
		cx |= DGPIMG_GETBIT01(dest, x + 1, y - 1) << 3; 
		cx |= DGPIMG_GETBIT01(dest, x + 1, y - 2) << 7; 
		DODECODE; 
		} 
	    for ( ; x < width; x++) { 
		/* Need conditionals again */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				      y + params->aty[0]) << 2; 
		    } 
		if (x + 1 < width) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 1, y - 1) << 3; 
		    cx |= DGPIMG_GETBIT01(dest, x + 1, y - 2) << 7; 
		    } 
 
		DODECODE; 
		} 
	    } 
	else { 
	    /* Need conditionals in y; might as well just do them everywhere */ 
	    for (x = 0; x < width; x++) { 
		if ((y + params->aty[0] >= 0) && (x + params->atx[0] >= 0) && 
		    (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				      y + params->aty[0]) << 2; 
		    } 
 
		/* Current row pixel is already loaded */ 
		/* Load previous row pixel */ 
		if ((y - 1 >= 0) && (x + 1 < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 1, y - 1) << 3; 
		    } 
		/* Load two-back row pixel */ 
		if ((y - 2 >= 0) && (x + 1 < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 1, y - 2) << 7; 
		    } 
 
		DODECODEMASK; 
		} 
	    } 
	DGP_REPORTF1(DGP_MESSAGE,"%c", '\n'); 
	} 
    } 
#undef	DODECODE 
#undef	DODECODEMASK 
 
/* 
 * Do the 10-pixel 2-line template decode. 
 */ 
static void 
arithDecodeImageT10L2(DgpImage_Packed dest, arith_bitmap_params *params, 
		      Dgp_UInt8 *contexts, DgpEncode_MQDecoder *dec, 
		      DgpImage_Packed mask) 
{ 
    Dgp_UInt32 cx; 
    Dgp_Int32 x, y; 
    Dgp_Int32 minaty, minatx, maxatx; 
    Dgp_Int32 bit; 
    Dgp_Int32 width = DGPIMG_WIDTH(dest), height = DGPIMG_HEIGHT(dest); 
    Dgp_Bool lntp = DGP_TRUE; 
 
    Dgp_State *dgp_state = dec->dgp_state; 
 
    DGP_VOID_PROLOGUE("arithDecodeImageT10L2", DGP_PERRSTATE); 
 
 
    /* 
     * The template looks like: 
     *   XXXXX1 
     *  XXXX? 
     * 
     * We have a "context" value that looks like: 
     * bb bbb1 cccc 
     * where 1 is the AT pixel value, bbbbb are the (LtoR) values from the 
     * top row of the template, and cccc are the values from the bottom row. 
     * To go to the next pixel, we 
     * shift the context left one bit, AND it with 011 1100 1110 
     * (clearing out the bits that landed in the wrong place), and perform 
     * 3 bit-loads into bits 0 (bottom row), 4 (AT1), 5 (top) 
     * We load from (after shift coordinates) (-1,0), (1,-1), (ATX1, ATY1) 
     */ 
 
    /* What's the smallest Y offset used? */ 
    minaty = MIN(params->aty[0], -1);	/* -1 is always used */ 
 
    /* smallest and largest X offset used? */ 
    minatx = MIN(params->atx[0], 1); /* 1 is always used; -1 is special case */ 
    maxatx = MAX(params->atx[0], 1); /* 1 is always used */ 
 
    /* 
     * Given values in cx, x and y, decode a bit, shift and mask CX, 
     * store the bit, and put the bit in CX 
     */ 
#define	DODECODE \
    do { \
	bit = DgpEncode_mqdDecode(dec, &contexts[cx]); \
	DGP_REPORTF1(DGP_MESSAGE,"@(%4d, ", x); \
	DGP_REPORTF1(DGP_MESSAGE,"%4d) ", y); \
	DGP_REPORTF1(DGP_MESSAGE,"%d", bit); \
	DGP_REPORTF1(DGP_MESSAGE,"(%03lx)", cx); \
	DGP_REPORTF1(DGP_MESSAGE,"[%02x]", contexts[cx]); \
	cx = ((cx << 1) & 0x3ce) | bit; \
	DGPIMG_ORBIT(dest, x, y, bit); \
	} while (0) 
 
#define	DODECODEMASK \
    do { \
	if ((mask == (DgpImage_Packed)NULL) || (DGPIMG_GETBIT01(mask, x, y) == 0)) { \
	    DODECODE; \
	    } \
	else { \
	    cx = ((cx << 1) & 0x3ce); \
	    } \
	} while (0) 
 
    for (y = 0; y < height; y++) { 
	/* Is TP on? */ 
	if (params->tpon) { 
	    /* 
	     * Yes; decode a bit using context 
	     *  011001 
	     * 0101? 
	     * which linearises to 01 1001 0101 
	     */ 
	    bit = DgpEncode_mqdDecode(dec, &contexts[0x195]); 
	    DGP_REPORTF1(DGP_MESSAGE,"TP%d", bit); 
	    lntp ^= bit; 
 
	    if (lntp == 0) { 
		/* It's a line copy */ 
		Dgp_UInt32 *p, *q; 
		if (y > 0) { 
		    for (x = 0, p = DGPIMG_GETROW(dest, y), 
			     q = DGPIMG_GETROW(dest, y - 1); 
			 x < DGPIMG_GETROWWIDTH(dest) / (int)sizeof(Dgp_UInt32); 
			 x++) { 
			*p++ = *q++; 
			} 
		    } 
		DGP_REPORTF1(DGP_MESSAGE,"%c", '\n'); 
		continue;	/* Done with this line */ 
		} 
	    } 
 
	/* Do the initial load */ 
	if (y >= 1) { 
	    /* 
	     * Normal case: load up 00 0bb0 0000 
	     * where the "b" come from the line above: 2 pixels 
	     * from above line. 
	     */ 
	    cx = ((*DGPIMG_GETROW(dest, y - 1) >> (DGP_BITS_UINT32 - 2)) << 5); 
	    } 
	else { 
	    /* Only zeros */ 
	    cx = 0; 
	    } 
 
	if ((mask == (DgpImage_Packed)NULL) && (y + minaty >= 0)) { 
	    /* No conditionals needed for y */ 
	    for (x = 0; x < MIN(-minatx, width); x++) { 
		/* Need conditionals for x */ 
 
		/* Load the AT pixel1 */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				      y + params->aty[0]) << 4; 
		    } 
 
		/* Current row pixel is already loaded */ 
		if (x + 1 < width) { 
		    /* Load previous row pixel */ 
		    cx |= DGPIMG_GETBIT01(dest, x + 1, y - 1) << 5; 
		    } 
 
		DODECODE; 
		} 
 
	    for ( ; x + maxatx < width; x++) { 
		/* x + minatx >= 0; x + maxatx < width; no conditionals needed */ 
		cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				  y + params->aty[0]) << 4; 
		cx |= DGPIMG_GETBIT01(dest, x + 1, y - 1) << 5; 
		DODECODE; 
		} 
	    for ( ; x < width; x++) { 
		/* Need conditionals again */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				      y + params->aty[0]) << 4; 
		    } 
		if (x + 1 < width) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 1, y - 1) << 5; 
		    } 
 
		DODECODE; 
		} 
	    } 
	else { 
	    /* Need conditionals in y; might as well just do them everywhere */ 
	    for (x = 0; x < width; x++) { 
		if ((y + params->aty[0] >= 0) && (x + params->atx[0] >= 0) && 
		    (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + params->atx[0], 
				      y + params->aty[0]) << 4; 
		    } 
 
		/* Current row pixel is already loaded */ 
		/* Load previous row pixel */ 
		if ((y - 1 >= 0) && (x + 1 < width)) { 
		    cx |= DGPIMG_GETBIT01(dest, x + 1, y - 1) << 5; 
		    } 
 
		DODECODEMASK; 
		} 
	    } 
	DGP_REPORTF1(DGP_MESSAGE,"%c", '\n'); 
	} 
    } 
#undef	DODECODE 
#undef	DODECODEMASK 
 
#endif 
