/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/dechtdict.c,v 1.12 2003/01/27 21:48:33 awilliam Exp $ 
 * 
 */ 
 
#include "util.h" 
#include "jb2file.h" 
#include "datasource.h" 
#include "segment.h" 
#include "dechtdict.h" 
#include "jb2halftone.h" 
#include "decsymdict.h" 
#include "decgenreg.h" 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/dechtdict.c,v 1.12 2003/01/27 21:48:33 awilliam Exp $"); 
 
/* 
 * Decode the halftone dictionary; fill in *segment 
 */ 
ErrCode 
jb2DecodeHTDict(jb2File file, DataSource data, 
		jb2Segment *segment, jb2Segment **refsegs) 
{ 
  DGP_PROLOGUE("jb2DecodeHTDict()", jb2_pErrState(file)); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
    Dgp_UInt8 lbuf[4]; 
    Dgp_UInt8 cbuf; 
    Dgp_Bool hdmmr; 
    arith_bitmap_params abparams; 
    Dgp_Int32 graymax; 
    Dgp_UInt8 *direct_contexts = (Dgp_UInt8 *)NULL; 
    DgpEncode_MQDecoder mqdec; 
    Dgp_UInt32 *collective = (Dgp_UInt32 *)NULL; 
    Dgp_Int32 i; 
    Dgp_Int32 collw, colllpl, colllen; 
    Dgp_Int32 pattlpl; 
    jb2Pattern *patterns = (jb2Pattern *)NULL; 
    size_t bytes; 
 
    (void)refsegs; 
 
    /* Parse the header */ 
 
    /* One-byte flags field */ 
    DSGETC(data, &cbuf); 
 
#ifdef	VALPRINT 
    fprintf(stderr, "HDFLAGS %02x\n", cbuf); 
#endif 
 
    /* Bit 0: HDMMR */ 
    if (cbuf & 1) { 
	hdmmr = DGP_TRUE; 
	} 
    else { 
#ifndef	NOMQ 
	hdmmr = DGP_FALSE; 
#else 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
	} 
 
    /* Bits 1-2: HDTEMPLATE */ 
    abparams.rtemplate = (cbuf >> 1) & 3; 
    if (hdmmr) { 
	if (abparams.rtemplate != 0) { 
	    FILEERR(cErrCompressionError, "MMR halftone dictionary has non-zero template"); 
	    } 
	} 
 
    /* Bits 3-7: reserved */ 
    if ((cbuf >> 3) != 0) { 
	FILEERR(cErrCompressionError, "reserved bits in halftone dictionary flags aren't zero"); 
	} 
 
    /* One byte HDPW field */ 
    DSGETC(data, &cbuf); 
    segment->data.htdict.pat_w = cbuf; 
#ifdef	VALPRINT 
    fprintf(stderr, "HDPW %d\n", segment->data.htdict.pat_w); 
#endif 
 
    /* One byte HDPH field */ 
    DSGETC(data, &cbuf); 
    segment->data.htdict.pat_h = cbuf; 
#ifdef	VALPRINT 
    fprintf(stderr, "HDPH %d\n", segment->data.htdict.pat_h); 
#endif 
 
    /* Four byte GRAYMAX field */ 
    DSREAD(data, 4, lbuf, &bytes); 
    graymax = 
	(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
#ifdef	VALPRINT 
    fprintf(stderr, "GRAYMAX %d\n", graymax); 
#endif 
    segment->data.htdict.npatterns = graymax + 1; 
     
    /* Allocate a bitmap that's npatterns*pat_w wide, pat_h high */ 
    collw = (segment->data.htdict.pat_w * segment->data.htdict.npatterns); 
    colllpl = (collw + (DGP_BITS_UINT32 - 1)) / DGP_BITS_UINT32; 
    collective = (Dgp_UInt32 *)(a->calloc)(segment->data.htdict.pat_h, 
					 colllpl * sizeof(Dgp_UInt32)); 
    if (collective == (Dgp_UInt32 *)NULL) { 
	NOMEM; 
	} 
 
    /* How many bytes is it? */ 
    colllen = segment->datalen - 7;	/* Seven bytes already consumed */ 
 
    /* Decode the bitmap */ 
    if (hdmmr) { 
	DGP_ERR_PASS_CLEANUP(jb2MMRDecodeImage(collective, collw,  
					       segment->data.htdict.pat_h, 
			       colllpl, 0, 
			       collw - 1, segment->data.htdict.pat_h - 1, 
					       file, data, colllen)); 
	} 
    else { 
#ifndef	NOMQ 
	/* Gotta set up a decoder and contexts */ 
	/* AT pixels */ 
 
	/* First AT is always at the corresponding pixel in the prev pattern */ 
	abparams.atx[0] = -segment->data.htdict.pat_w; 
	abparams.aty[0] = 0; 
 
	if (abparams.rtemplate == 0) { 
	    /* Other ATs at default locs */ 
	    abparams.atx[1] = -3; 
	    abparams.aty[1] = -1; 
	    abparams.atx[2] = 2; 
	    abparams.aty[2] = -2; 
	    abparams.atx[3] = -2; 
	    abparams.aty[3] = -2; 
	    } 
	else { 
	    for (i = 1; i < 4; i++) { 
		abparams.atx[i] = abparams.aty[i] = 0; 
		} 
	    } 
 
	abparams.tpon = DGP_FALSE; 
 
	/* Set up the MQ decoder */ 
	dsInitDgpEncode_MQDecoder(data, &mqdec); 
 
	/* and the contexts */ 
	direct_contexts = 
	    (Dgp_UInt8 *)(a->calloc)(jb2ABContextsNeeded(&abparams), 1); 
	if (direct_contexts == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
 
	DGP_ERR_PASS_CLEANUP(jb2ArithDecodeImage(collective, collw, 
				 segment->data.htdict.pat_h, 
				 colllpl, &abparams, direct_contexts, 
						 file, &mqdec)); 
 
	/* Done decoding */ 
	(a->free)((void *)direct_contexts); 
	direct_contexts = (Dgp_UInt8 *)NULL; 
#else 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
	} 
 
    /* Now do the readout */ 
    patterns = (jb2Pattern *)(a->malloc)(segment->data.htdict.npatterns * 
				    sizeof(*patterns)); 
    if (patterns == (jb2Pattern *)NULL) { 
	NOMEM; 
	} 
 
    /* Clear bits for predictable bailout */ 
    for (i = 0; i < segment->data.htdict.npatterns; i++) { 
	patterns[i].bits = (Dgp_UInt32 *)NULL; 
	} 
 
    pattlpl = (segment->data.htdict.pat_w + (DGP_BITS_UINT32 - 1)) / DGP_BITS_UINT32; 
    for (i = 0; i < segment->data.htdict.npatterns; i++) { 
	/* Allocate the pattern bitmap */ 
	patterns[i].bits = 
	    (Dgp_UInt32 *)(a->calloc)(segment->data.htdict.pat_h, 
				    pattlpl * sizeof(Dgp_UInt32)); 
	if (patterns[i].bits == (Dgp_UInt32 *)NULL) { 
	    NOMEM; 
	    } 
	patterns[i].longsperline = (Dgp_UInt16)pattlpl; 
 
	/* Read out the pattern */ 
	jb2ReadoutBits(collective, i * segment->data.htdict.pat_w, 
		       collw, segment->data.htdict.pat_h, 
		       patterns[i].bits, segment->data.htdict.pat_w); 
	} 
 
    /* Stash into the segment */ 
    segment->data.htdict.patterns = patterns; 
 
    /* Sync the DataSource back up */ 
#ifndef	NOMQ 
    if (!hdmmr) { 
      DGP_ERR_PASS_CLEANUPF0(dsFinalDgpEncode_MQDecoder(data, &mqdec), 
			     "problem syncing MQ decoder"); 
    } 
#endif 
 
    /* Free the collective bitmap */ 
    (a->free)((void *)collective); 
 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
 
    if (collective != (Dgp_UInt32 *)NULL) { 
	(a->free)((void *)collective); 
	} 
 
    if (direct_contexts != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)direct_contexts); 
	} 
 
    if (patterns != (jb2Pattern *)NULL) { 
	for (i = 0; i < segment->data.htdict.npatterns; i++) { 
	    if (patterns[i].bits != (Dgp_UInt32 *)NULL) { 
		(a->free)((void *)patterns[i].bits); 
		} 
	    } 
	(a->free)((void *)patterns); 
	} 
 
    ERR_RETURN; 
    } 
