/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/dechtreg.c,v 1.17 2003/01/27 21:48:33 awilliam Exp $ 
 * 
 */ 
 
#include "util.h" 
#include "jb2file.h" 
#include "datasource.h" 
#include "segment.h" 
#include "image.h" 
#include "segment_util.h" 
#include "symdec.h" 
#include "decgenreg.h" 
#include "blit.h" 
#include "image_defs.h"
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/dechtreg.c,v 1.17 2003/01/27 21:48:33 awilliam Exp $"); 
 
static ErrCode decodePlanes(jb2File file, DataSource data, size_t datalen, 
			    Dgp_Bool hmmr, 
			    Dgp_UInt8 htemplate, Dgp_Bool henableskip, 
			    Dgp_Int32 nplanes, 
			    Dgp_Int32 hgw, Dgp_Int32 hgh, Dgp_Int32 hpw, Dgp_Int32 hph, 
			    jb2Segment *segment); 

ErrCode 
jb2DecodeHTReg(jb2File file, DataSource data, 
	       jb2Segment *segment, jb2Segment **refsegs) 
{ 
  DGP_PROLOGUE("jb2DecodeHTReg()", jb2_pErrState(file)); 
 
    Dgp_UInt8 lbuf[4]; 
    Dgp_UInt8 cbuf; 
    Dgp_Int32 nplanes; 
    Dgp_Int32 hgw, hgh; 
    Dgp_Bool henableskip; 
    Dgp_Bool hmmr; 
    Dgp_UInt8 htemplate; 
    jb2Segment *dict; 
    size_t bytes; 
 
    /* Should refer to exactly one segment: a halftone dictionary */ 
    if ((segment->nrefsegs != 1) || 
	(refsegs == (jb2Segment **)NULL) || 
	(refsegs[0]->segtype != JB2SEGTYPE_HTDICT)) { 
	FILEERR(cErrCompressionError, "bad reference list for halftone region"); 
	} 
 
    /* Get a handle on our dictionary */ 
    dict = refsegs[0]; 
 
    /* nplanes is based on the number of patterns in dict */ 
    nplanes = ceillog2(dict->data.htdict.npatterns); 
 
    /* Decode the header */ 
 
    /* First is a region segment header */ 
    DGP_ERR_PASS_CLEANUP(jb2DecodeRegSegment(file, data, &segment->reginfo)); 
 
    /* One-byte halftone region segment flags */ 
    DSGETC(data, &cbuf); 
 
#ifdef	VALPRINT 
    fprintf(stderr, "HRFLAGS %02x\n", cbuf); 
#endif 
 
    /* Bit 0: HMMR */ 
    hmmr = (cbuf & 1) ? DGP_TRUE : DGP_FALSE; 
#ifdef	NOMQ 
    if (!hmmr) { 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
	} 
#endif 
 
    /* Bits 1-2: HTEMPLATE */ 
    htemplate = (cbuf >> 1) & 3; 
    if (hmmr && (htemplate != 0)) { 
	FILEERR(cErrCompressionError, "non-zero template in MMR halftone region segment"); 
	} 
 
    /* Bit 3: HENABLESKIP */ 
    henableskip = (cbuf >> 3) & 1; 
    if (hmmr && henableskip) { 
	FILEERR(cErrCompressionError, "skip enabled in MMR halftone region segment"); 
	} 
 
    /* Bits 4-6: HCOMBOP */ 
    segment->data.htreg.int_combop = (cbuf >> 4) & 7; 
    if (segment->data.htreg.int_combop > 4) { 
	FILEERR(cErrCompressionError, "bad combination operator in halftone region segment"); 
	} 
 
    /* Bit 7: HDEFPIXEL */ 
    segment->data.htreg.defpixel = (cbuf & 128) ? 1 : 0; 
 
    /* HGW */ 
    DSREAD(data, 4, lbuf, &bytes); 
    hgw = (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
 
#ifdef	VALPRINT 
    fprintf(stderr, "HGW %d\n", hgw); 
#endif 
 
    /* HGH */ 
    DSREAD(data, 4, lbuf, &bytes); 
    hgh = (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
 
#ifdef	VALPRINT 
    fprintf(stderr, "HGH %d\n", hgh); 
#endif 
 
    /* HGX */ 
    DSREAD(data, 4, lbuf, &bytes); 
    segment->data.htreg.hgx = 
	(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
 
#ifdef	VALPRINT 
    fprintf(stderr, "HGX %d\n", segment->data.htreg.hgx); 
#endif 
 
    /* HGY */ 
    DSREAD(data, 4, lbuf, &bytes); 
    segment->data.htreg.hgy = 
	(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
 
#ifdef	VALPRINT 
    fprintf(stderr, "HGY %d\n", segment->data.htreg.hgy); 
#endif 
 
    /* HRX */ 
    DSREAD(data, 2, lbuf, &bytes); 
    segment->data.htreg.hrx = (lbuf[0] << 8) | lbuf[1]; 
     
#ifdef	VALPRINT 
    fprintf(stderr, "HRX %d\n", segment->data.htreg.hrx); 
#endif 
 
    /* HRY */ 
    DSREAD(data, 2, lbuf, &bytes); 
    segment->data.htreg.hry = (lbuf[0] << 8) | lbuf[1]; 
 
#ifdef	VALPRINT 
    fprintf(stderr, "HRY %d\n", segment->data.htreg.hry); 
#endif 
 
    /* Need to stash a pointer to the dictionary */ 
    segment->data.htreg.dictseg = dict; 
 
    /* Done with the header - decode the planes */ 
    DGP_ERR_PASS_CLEANUP(decodePlanes(file, data, segment->datalen - 38, 
		      hmmr, htemplate, henableskip, nplanes, 
		      hgw, hgh, 
		      dict->data.htdict.pat_w, dict->data.htdict.pat_h, 
				      segment)); 
     
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    ERR_RETURN; 
    } 
 
/* 
 * Decode the planes; stick the results in segment->data.htreg.planes 
 * The remaining amount of data in the segment is datalen. 
 */ 
static ErrCode 
decodePlanes(jb2File file, DataSource data, size_t datalen, Dgp_Bool hmmr, 
	     Dgp_UInt8 htemplate, Dgp_Bool henableskip, 
	     Dgp_Int32 nplanes, 
	     Dgp_Int32 hgw, Dgp_Int32 hgh, Dgp_Int32 hpw, Dgp_Int32 hph, jb2Segment *segment) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("decodePlanes()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    jb2HTRegSegment *htreg = &segment->data.htreg; 
    DgpImage_Packed *planes = (DgpImage_Packed *)NULL; 
    DgpImage_Short cells = (DgpImage_Short)NULL; 
    DgpImage_Packed mask = (DgpImage_Packed)NULL; 
    Dgp_Int32 i; 
    Dgp_UInt8 *coded_data = (Dgp_UInt8 *)NULL; 
    Dgp_UInt8 *direct_contexts = (Dgp_UInt8 *)NULL; 
    DgpEncode_MQDecoder mqdec; 
    Dgp_UInt8 *ptr; 
    size_t dataleft; 
    Dgp_Int32 x, y; 
    arith_bitmap_params abparams; 
    Dgp_Int32 val, lastbit, curbit; 
    Dgp_Int32 xloc, yloc; 
    size_t bytes; 
 
    if (nplanes > 15) { 
	FILEERR(cErrCompressionError, "unimplemented number of planes"); 
	} 
 
    /* Allocate the array of planes */ 
    planes = (DgpImage_Packed *)(a->malloc)(nplanes * sizeof(*planes)); 
    if (planes == (DgpImage_Packed *)NULL) { 
	NOMEM; 
	} 
 
    /* Clear pointers */ 
    for (i = 0; i < nplanes; i++) { 
	planes[i] = (DgpImage_Packed)NULL; 
	} 
 
    /* 
     * Our core MMR routines don't like not knowing in advance how much 
     * data there is, if they're reading from a DataSource.  If they're 
     * reading from memory, they're happier.  Thus, we pull the entire 
     * segment's data into memory, and just decode in blocks from there. 
     */ 
    if (hmmr) { 
	coded_data = (Dgp_UInt8 *)(a->malloc)(datalen); 
	if (coded_data == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
 
	DSREAD(data, datalen, coded_data, &bytes); 
	} 
    else { 
	/* No need to pre-read the data; just initialise the MQ decoder */ 
#ifndef	NOMQ 
	dsInitDgpEncode_MQDecoder(data, &mqdec); 
 
	/* Set up the parameters */ 
	if (htemplate == 0) { 
	    abparams.atx[0] = 3; 
	    abparams.aty[0] = -1; 
	    abparams.atx[1] = -3; 
	    abparams.aty[1] = -1; 
	    abparams.atx[2] = 2; 
	    abparams.aty[2] = -2; 
	    abparams.atx[3] = -2; 
	    abparams.aty[3] = -2; 
	    } 
	else { 
	    if (htemplate == 1) { 
		abparams.atx[0] = 3; 
		abparams.aty[0] = -1; 
		} 
	    else { 
		abparams.atx[0] = 2; 
		abparams.aty[0] = -1; 
		} 
	    for (i = 1; i < 4; i++) { 
		abparams.atx[i] = abparams.aty[i] = 0; 
		} 
	    } 
	abparams.rtemplate = htemplate; 
	abparams.tpon = DGP_FALSE; 
 
	/* Of course, we need contexts... */ 
	direct_contexts = 
	    (Dgp_UInt8 *)(a->calloc)(jb2ABContextsNeeded(&abparams), 1); 
	if (direct_contexts == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
 
	/* We may need to compute a mask */ 
	if (henableskip) { 
	    /* Yes - need a mask */ 
	    DGP_ERR_PASS (DgpImage_new(dgp_state, DGPIMAGE_PACKED, hgw, hgh, (void**) &mask)); 
 
	    for (y = 0; y < hgh; y++) { 
		for (x = 0; x < hgw; x++) { 
		    /* 
		     * Does this pixel of the cells array have no effect on 
		     * the final region? 
		     */ 
		    /* Compute the high-res version of the position */ 
		    xloc = htreg->hgx + y * htreg->hry + x * htreg->hrx; 
		    yloc = htreg->hgy + y * htreg->hrx - x * htreg->hry; 
 
		    /* Shift them to low-res */ 
		    xloc >>= 8; 
		    yloc >>= 8; 
 
		    if ((xloc + hpw <= 0) || 
			(xloc >= (int)segment->reginfo.regwidth) || 
			(yloc + hph <= 0) || 
			(yloc >= (int)segment->reginfo.regheight)) { 
			/* Skip this pixel */ 
			DGPIMG_SETBIT(mask, x, y); 
			} 
		    } 
		} 
#ifdef	VALPRINT 
	    { 
		Dgp_Int8 name[100]; 
		sprintf(name, "/tmp/skipmask.pbm", i); 
		imSave(mask, name); 
		} 
#endif 
	    } 
#else 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
	} 
 
    ptr = coded_data; 
    for (i = nplanes - 1; i >= 0; i--) { 
	/* Allocate the plane */ 
	DGP_ERR_PASS (DgpImage_new(dgp_state, DGPIMAGE_PACKED, hgw, hgh, (void **) &planes[i])); 
 
	/* Decode the plane */ 
	if (hmmr) { 
	    /* MMR decode it */ 
	    dataleft = datalen; 
	    DGP_ERR_PASS_CLEANUPF0(jb2MMRDecodeImageMemory(DGPIMG_GETROW(planes[i], 
								   0), 
					 hgw, hgh, 
					 DGPIMG_GETROWWIDTH(planes[i]) / sizeof(Dgp_UInt32), 
					 0, hgw - 1, hgh - 1, 
					 ptr, &dataleft, dgp_state), 
				   "can't MMR-decode halftone region plane"); 
 
	    /* How much did that use? */ 
	    ptr += datalen - dataleft; 
	    datalen = dataleft; 
	    } 
	else { 
	    /* Arith decode it */ 
#ifndef	NOMQ 
	    jb2ArithDecodeImagePacked(planes[i], &abparams, 
				      direct_contexts, &mqdec, 
				      mask); 
#else 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
	    } 
#ifdef	VALPRINT 
	{ 
	    Dgp_Int8 name[100]; 
	    sprintf(name, "/tmp/plane%d.pbm", i); 
	    imSave(planes[i], name); 
	    } 
#endif 
	} 
 
    /* Now take the planes and build the cells array */ 
 
    DGP_ERR_PASS (DgpImage_new(dgp_state, DGPIMAGE_SHORT | DGPIMAGE_DONTCLEAR, hgw, hgh, (void **) &cells)); 
 
    /* We don't need to consult the mask - masked things are don't cares */ 
    for (y = 0; y < hgh; y++) { 
	for (x = 0; x < hgw; x++) { 
	    val = 0; 
	    lastbit = 0; 
 
	    for (i = nplanes - 1; i >= 0; i--) { 
		/* Read the bit */ 
		curbit = DGPIMG_GETBIT01(planes[i], x, y); 
		/* Undo the Gray coding */ 
		curbit ^= lastbit; 
 
		val |= curbit << i; 
		lastbit = curbit; 
		} 
	    DGPIMG_REF(cells, x, y) = (Dgp_Int16)val; 
	    } 
	} 
     
#ifdef	VALPRINT 
    { 
	DgpImage_Gray foo = imDemote(cells, DGPIMAGE_GRAY); 
	imSave(foo, "/tmp/cells.pgm"); 
	DgpImage_free(foo); 
	} 
#endif 
 
    htreg->cells = cells; 
 
    /* Sync the DataSource back up */ 
#ifndef	NOMQ 
    if (!hmmr) { 
      DGP_ERR_PASS_CLEANUPF0(dsFinalDgpEncode_MQDecoder(data, &mqdec), 
			     "problem syncing MQ decoder"); 
	} 
#endif 
 
    for (i = 0; i < nplanes; i++) { 
	DgpImage_free(planes[i]); 
	} 
    (a->free)((void *)planes); 
 
    if (mask != (DgpImage_Packed)NULL) { 
	DgpImage_free(mask); 
	} 
 
    if (coded_data != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)coded_data); 
	} 
 
    if (direct_contexts != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)direct_contexts); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
	 
    if (cells != (DgpImage_Short)NULL) { 
	DgpImage_free(cells); 
	} 
 
    if (planes != (DgpImage_Packed *)NULL) { 
	for (i = 0; i < nplanes; i++) { 
	    if (planes[i] != (DgpImage_Packed)NULL) { 
		DgpImage_free(planes[i]); 
		} 
	    } 
	(a->free)((void *)planes); 
	} 
 
    if (mask != (DgpImage_Packed)NULL) { 
	DgpImage_free(mask); 
	} 
 
    if (coded_data != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)coded_data); 
	} 
 
    if (direct_contexts != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)direct_contexts); 
	} 
 
    ERR_RETURN; 
    } 
 
ErrCode 
jb2RenderHTReg(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2RenderHTReg()", jb2_pErrState(file)); 
 
    jb2RegSegment *reginfo; 
    DgpImage_Packed im = (DgpImage_Packed)NULL; 
    jb2HTRegSegment *htreg; 
    jb2HTDictSegment *dictinfo; 
    Dgp_Int32 hpw, hph; 
    Dgp_Int32 x, y; 
    Dgp_Int32 xloc, yloc; 
    Dgp_Int32 val; 
 
    reginfo = &segment->reginfo; 
 
    /* Done already? */ 
    if (reginfo->rendered) { 
	NORMAL_RETURN; 
	} 
     
    htreg = &segment->data.htreg; 
    dictinfo = &segment->data.htreg.dictseg->data.htdict; 
    hpw = dictinfo->pat_w; 
    hph = dictinfo->pat_h; 
 
    /* Can't handle everything */ 
 
    /* Allocate the buffer */ 
    DGP_ERR_PASS (DgpImage_new(jb2GetFileDgp_State(file), DGPIMAGE_PACKED, 
			    reginfo->regwidth, reginfo->regheight, (void**) &im)); 
 
    if (segment->data.htreg.defpixel != 0) { 
	/* Do the initial fill with 1 */ 
	DgpBlat_blitFillPacked(im); 
	} 
 
    /* Do the rendering */ 
    for (y = 0; y < DGPIMG_HEIGHT(htreg->cells); y++) { 
	for (x = 0; x < DGPIMG_WIDTH(htreg->cells); x++) { 
	    /* Compute the high-res version of the position */ 
	    xloc = htreg->hgx + y * htreg->hry + x * htreg->hrx; 
	    yloc = htreg->hgy + y * htreg->hrx - x * htreg->hry; 
 
	    /* Shift them to low-res */ 
	    xloc >>= 8; 
	    yloc >>= 8; 
 
	    val = DGPIMG_REF(htreg->cells, x, y); 
	    /* Paranoia-check */ 
	    if (val >= dictinfo->npatterns) { 
		FILEERR(cErrCompressionError, "bad halftone index"); 
		} 
 
	    DgpBlat_blitBitsOp(DGPIMG_GETROW(im, 0), 
		       DGPIMG_GETROWWIDTH(im) / sizeof(Dgp_UInt32), 
		       0, 0, reginfo->regwidth, reginfo->regheight, 
		       dictinfo->patterns[val].bits, 
		       dictinfo->patterns[val].longsperline, 
		       hpw, hph, 
		       xloc, yloc, 
		       jb2Minterms[segment->data.symreg.int_combop], 
			   jb2GetFileDgp_State(file)); 
	    } 
	} 
     
    /* Stash the bitmap away */ 
    reginfo->im = im; 
    reginfo->rendered = DGP_TRUE; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (im != (DgpImage_Packed)NULL) { 
	DgpImage_free(im); 
	} 
 
    ERR_RETURN; 
    } 
 
ErrCode 
jb2RenderHTRegIntoPage(jb2File file, jb2Segment *segment, DgpImage_Packed page) 
{ 
  DGP_PROLOGUE("jb2RenderHTRegIntoPage()", jb2_pErrState(file)); 
 
    jb2RegSegment *reginfo; 
 
    /* You can't render an intermediate segment into a page */ 
    DGP_ASSERT(segment->segtype != JB2SEGTYPE_HTREG_INT); 
 
    reginfo = &segment->reginfo; 
 
    /* The region must fit entirely on the page */ 
    if (((int)reginfo->regxpos + (int)reginfo->regwidth > DGPIMG_WIDTH(page)) || 
	((int)reginfo->regypos + (int)reginfo->regheight > DGPIMG_HEIGHT(page))) { 
	FILEERR(cErrCompressionError, "region extends off page"); 
	} 
     
    /* Render it */ 
    DGP_ERR_PASS_CLEANUP(jb2RenderHTReg(file, segment)); 
 
    /* and blit it */ 
    DgpBlat_blitPackedOp(page, reginfo->im, 
		 (int)reginfo->regxpos, (int)reginfo->regypos, 
		 jb2Minterms[reginfo->combop]); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    ERR_RETURN; 
    } 
 
