/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/dechufftable.c,v 1.14 2003/01/27 21:48:33 awilliam Exp $ 
 * 
 */ 
 
#include "segment.h" 
#include "datasource.h" 
#include "fileerr.h" 
#include "list.h" 
#include "jbitfile.h"
 #include "PsgFix.h"
#include <stdlib.h> 
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/dechufftable.c,v 1.14 2003/01/27 21:48:33 awilliam Exp $"); 
 
/* 
 * Decode Huffman table segments 
 */ 
 
ErrCode 
jb2DecodeHuffTable(jb2File file, DataSource data, 
		   jb2Segment *segment, jb2Segment *refsegs) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2DecodeHuffTable()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_UInt8 cbuf; 
    Dgp_UInt8 lbuf[4]; 
    Dgp_Int32 htps, htrs; 
    jbitfile bf = NULLJBITFILE, tmpbf; 
    Dgp_Int32 v; 
    Dgp_Int32 lowval, highval; 
    DgpAdt_List lineslist = DGP_NULLLIST; 
    DgpAdt_ListNode ln, retnode; 
    table_line *lines = (table_line *)NULL; 
    Dgp_Int32 nlines; 
    table_line *newline = (table_line *)NULL; 
    table_line oob_line, lower_line, upper_line; 
    Dgp_Int32 i; 
    void **lineptrs = (void **)NULL; 
    size_t bytes; 
    Dgp_UInt32 tmpv; 
 
    (void)refsegs; 
 
    if (segment->nrefsegs > 0) { 
	FILEERR(cErrCompressionError, "tables segment referencing another segment"); 
	} 
 
    /* Read the flags byte */ 
    DSGETC(data, &cbuf); 
#ifdef	VALPRINT 
    fprintf(stderr, "HTFLAGS %02x\n", cbuf); 
#endif 
 
    /* Read the bits */ 
    segment->data.table.table.oob_valid = cbuf & 1 ? DGP_TRUE : DGP_FALSE; 
    htps = ((cbuf & 0xe) >> 1) + 1; 
    htrs = ((cbuf & 0x70) >> 4) + 1; 
 
    /* Read the lowest value */ 
    DSREAD(data, 4, lbuf, &bytes); 
    /* 
     * Be a little careful - lowval might not be 32 bits.  We can get the 
     * right sign on the result by casting lbuf[0] to a *signed* Dgp_Int8 before 
     * shifting it.  If its MSB is set, then that will sign-extend it 
     * properly. 
     */ 
    lowval = (((Dgp_Int8)lbuf[0]) << 24) | 
	(lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
#ifdef	VALPRINT 
    fprintf(stderr, "HTLOW %ld\n", lowval); 
#endif 
 
    /* Same for the highest value */ 
    DSREAD(data, 4, lbuf, &bytes); 
    highval = (((Dgp_Int8)lbuf[0]) << 24) | 
	(lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
#ifdef	VALPRINT 
    fprintf(stderr, "HTHIGH %ld\n", highval); 
#endif 
     
    /* OK - now morph the DataSource into a jbitfile and start reading bits. */ 
    DGP_ERR_PASS_CLEANUP(jbfOpenFromDataSource(data, &bf)); 
 
    /* 
     * We need to stash the values we're reading somewhere (don't know how 
     * many there will be to begin with), so make a DgpAdt_List. 
     */ 
    DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&lineslist, dgp_state)); 
 
    for (v = lowval; v < highval; ) { 
	newline = (table_line *)(a->malloc)(sizeof(*newline)); 
	if (newline == (table_line *)NULL) { 
	    NOMEM; 
	    } 
 
	/* Get the prefix length */ 
	DGP_ERR_PASS_CLEANUP(jbfReadBits(bf, htps, &(tmpv))); 
	newline->preflen = (Dgp_Int8)tmpv; 
 
	/* and the size of the range */ 
	DGP_ERR_PASS_CLEANUP(jbfReadBits(bf, htrs, &(tmpv))); 
	newline->rangelen = (Dgp_Int8)tmpv; 
 
	/* and we know the base */ 
	newline->base = v; 
#ifdef	VALPRINT 
	fprintf(stderr, "HTLINE %d %d\n", newline->preflen, newline->rangelen); 
#endif 
 
	/* We know the size - figure the base of the next range */ 
	v += 1l << newline->rangelen; 
 
	/* Hook newline onto the list we're growing */ 
	DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd(lineslist, (void *)newline, &retnode)); 
 
	/* It's hooked - delete the live reference */ 
	newline = (table_line *)NULL; 
	} 
     
    /* Now read the lower line */ 
    DGP_ERR_PASS_CLEANUP(jbfReadBits(bf, htps, &(tmpv))); 
    lower_line.preflen = (Dgp_Int8)tmpv; 
#ifdef	VALPRINT 
    fprintf(stderr, "HTLOWLINE %d\n", lower_line.preflen); 
#endif 
    if (lower_line.preflen == 0) { 
	/* Lower line not valid */ 
	segment->data.table.table.lower_valid = DGP_FALSE; 
	} 
    else { 
	segment->data.table.table.lower_valid = DGP_TRUE; 
	lower_line.base = lowval - 1; 
	lower_line.rangelen = 32; 
	} 
	 
    /* Same for upper line */ 
    DGP_ERR_PASS_CLEANUP(jbfReadBits(bf, htps, &(tmpv))); 
    upper_line.preflen = (Dgp_Int8)tmpv; 
#ifdef	VALPRINT 
    fprintf(stderr, "HTHIGHLINE %d\n", upper_line.preflen); 
#endif 
    if (upper_line.preflen == 0) { 
	/* Upper line not valid */ 
	segment->data.table.table.upper_valid = DGP_FALSE; 
	} 
    else { 
	segment->data.table.table.upper_valid = DGP_TRUE; 
	upper_line.base = highval; 
	upper_line.rangelen = 32; 
	} 
 
    /* If there's an OOB template, read it */ 
    if (segment->data.table.table.oob_valid) { 
	DGP_ERR_PASS_CLEANUP(jbfReadBits(bf, htps, &(tmpv))); 
	oob_line.preflen = (Dgp_Int8)tmpv; 
#ifdef	VALPRINT 
	fprintf(stderr, "HTOOBLINE %d\n", oob_line.preflen); 
#endif 
	oob_line.rangelen = 0; 
	oob_line.base = 0; 
	} 
 
    /* We're done with the jbitfile now */ 
    tmpbf = bf; 
    bf = NULLJBITFILE; 
    DGP_ERR_PASS_CLEANUP(jbfCloseFromDataSource(tmpbf)); 
 
    /* Now read all those lines out into table.lines */ 
 
    /* How many are there? */ 
    nlines = DGPADT_LILEN(lineslist) + 
	(segment->data.table.table.lower_valid ? 1 : 0) + 
	(segment->data.table.table.upper_valid ? 1 : 0) + 
	(segment->data.table.table.oob_valid ? 1 : 0); 
 
    /* Allocate them */ 
    lines = (table_line *)(a->malloc)(nlines * sizeof(*lines)); 
 
    /* and copy them */ 
    i = 0; 
    DGPADT_FOREACH (ln, lineslist) { 
	newline = (table_line *)DGPADT_LIGET(ln); 
	lines[i++] = *newline; 
	(a->free)((void *)newline); 
	} 
    /* don't need lineslist */ 
    DgpAdt_liFree(lineslist); 
    lineslist = DGP_NULLLIST; 
 
    /* Now add in the extra table lines */ 
    if (segment->data.table.table.lower_valid) { 
	lines[i++] = lower_line; 
	} 
     
    if (segment->data.table.table.upper_valid) { 
	lines[i++] = upper_line; 
	} 
     
    if (segment->data.table.table.oob_valid) { 
	lines[i++] = oob_line; 
	} 
     
    /* Copy everything into the segment's table */ 
    segment->data.table.table.nlines = nlines; 
    segment->data.table.table.lines = lines; 
 
    /* Now determine the actual Huffman codes */ 
    DGP_ERR_PASS_CLEANUPF0(jhtSetCodes(&segment->data.table.table, dgp_state), 
	"bad Huffman table"); 
 
    /* Compile the table; the pointer for each line is the line itself. */ 
    lineptrs = (void **)(a->malloc)(nlines * sizeof(*lineptrs)); 
    if (lineptrs == (void **)NULL) { 
	NOMEM; 
	} 
 
    for (i = 0; i < nlines; i++) { 
	lineptrs[i] = &lines[i]; 
	} 
 
    DGP_ERR_PASS_CLEANUPF0(jhtCompile(&segment->data.table.table, lineptrs, 
				      dgp_state), 
	"Huffman table didn't compile"); 
 
    (a->free)((void *)lineptrs); 
    lineptrs = (void **)NULL; 
 
    /* We're done - the spec is fleshed out. */ 
 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
    if (bf != NULLJBITFILE) { 
	(void)jbfCloseFromDataSource(bf); 
	} 
 
    if (newline != (table_line *)NULL) { 
	/* An active line */ 
	(a->free)((void *)newline); 
	} 
 
    if (lineslist != DGP_NULLLIST) { 
	DGPADT_FOREACH (ln, lineslist) { 
	    newline = (table_line *)DGPADT_LIGET(ln); 
	    (a->free)((void *)newline); 
	    } 
	DgpAdt_liFree(lineslist); 
	} 
 
    if (lines != (table_line *)NULL) { 
	(a->free)((void *)lines); 
	} 
 
    if (lineptrs != (void **)NULL) { 
	(a->free)((void *)lineptrs); 
	} 
 
    ERR_RETURN; 
    } 
 
