/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/decint.c,v 1.20 2003/01/27 21:48:33 awilliam Exp $ 
 * 
 */ 
 
/* 
 * Decoding integers; opposite of encint.c 
 */ 
 
#include "decint.h" 
#include "hufftable.h" 
#include <stdlib.h> 
#include "mqcoder.h" 
#include <string.h> 
#include "PsgFix.h"

 
#define DGP_ENC DGP_FALSE
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/decint.c,v 1.20 2003/01/27 21:48:33 awilliam Exp $"); 
 
struct _diDecoder { 
    Dgp_Bool mqcoder; 
    const hufftable_spec *hufftable; 
    table_line *lower_line; 
    table_line *upper_line; 
    table_line *oob_line; 
    Dgp_Int32 nnormallines; 
    jbitfile f; 
    DgpEncode_MQDecoder *mqdec; 
    Dgp_UInt8 *contexts; 
    Dgp_State *dgp_state; 
    }; 
 
/* 
 * Create an integer decoder from that Huffman table spec and the bitfile. 
 * The table spec (and the bitfile) must live at least as Dgp_Int32 as this 
 * decoder. 
 */ 
ErrCode 
diNewDecoder(const hufftable_spec *spec, jbitfile file, Dgp_State *dgp_state,  
	     diDecoder *_newd) 
{ 
  DGP_PROLOGUE("diNewDecoder()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    diDecoder newd = (diDecoder)(a->malloc)(sizeof(*newd)); 
    table_line *lines = (table_line *)NULL; 
    Dgp_Int32 i; 
 
    *_newd = NULLDIDECODER; 
 
    DGP_CHECK_ALLOC_CLEANUP(newd); 
 
    newd->dgp_state = dgp_state; 
    newd->hufftable = spec; 
    newd->mqcoder = DGP_FALSE; 
 
    newd->f = file; 
    lines = newd->hufftable->lines; 
     
    /* Set up some values for easy search */ 
 
    /* How many normal table lines? */ 
    newd->nnormallines = spec->nlines - 
	(spec->lower_valid ? 1 : 0) - 
	(spec->upper_valid ? 1 : 0) - 
	(spec->oob_valid ? 1 : 0); 
 
    i = newd->nnormallines; 
 
    /* Set up the special templates */ 
    if (spec->lower_valid) { 
	newd->lower_line = &lines[i++]; 
	} 
    else { 
	newd->lower_line = (table_line *)NULL; 
	} 
 
    if (spec->upper_valid) { 
	newd->upper_line = &lines[i++]; 
	} 
    else { 
	newd->upper_line = (table_line *)NULL; 
	} 
 
    if (spec->oob_valid) { 
	newd->oob_line = &lines[i++]; 
	} 
    else { 
	newd->oob_line = (table_line *)NULL; 
	} 
 
    /* 
     * In the future, we should do some compilation on the Huffman table 
     * to make the decoding easier. 
     */ 
 
    *_newd = newd; 
    newd = NULLDIDECODER; 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (newd != NULLDIDECODER) { 
	(a->free)((void *)newd); 
	} 
 
    ERR_RETURN; 
    } 
 
#ifndef	NOMQ 
/* 
 * Create an MQ-based integer decoder from the given DgpEncode_MQDecoder and set of 
 * contexts.  The context block should be 512 bytes Dgp_Int32; it and the 
 * DgpEncode_MQDecoder must live at least as Dgp_Int32 as this decoder. 
 */ 
ErrCode 
diNewDecoderMQ(DgpEncode_MQDecoder *mqdec, Dgp_UInt8 *contexts, Dgp_State *dgp_state,  
	     diDecoder *_newd) 
{ 
  DGP_PROLOGUE("diNewDecoderMQ()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    diDecoder newd = (diDecoder)(a->malloc)(sizeof(*newd)); 
     
    *_newd = NULLDIDECODER; 
 
    DGP_CHECK_ALLOC_CLEANUP(newd); 
 
    newd->dgp_state = dgp_state; 
    newd->mqcoder = DGP_TRUE; 
    newd->mqdec = mqdec; 
    newd->contexts = contexts; 
 
    *_newd = newd; 
    newd = NULLDIDECODER; 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (newd != NULLDIDECODER) { 
	(a->free)((void *)newd); 
	} 
 
    ERR_RETURN; 
    } 
#endif 
 
/* 
 * If huff is DGP_TRUE, create an integer decoder based on spec and file. 
 * If huff is DGP_FALSE, create an integer decoder based on mqdec and 
 * contexts. Clear contexts if huff is DGP_FALSE. 
 * If huff is DGP_TRUE, mqdec and contexts need not be valid; if huff is DGP_FALSE, 
 * spec and file need not be valid. 
 */ 
ErrCode 
diNewDecoderEither(Dgp_Bool huff, 
		   const hufftable_spec *spec, jbitfile file, 
		   DgpEncode_MQDecoder *mqdec, Dgp_UInt8 *contexts,  
		   Dgp_State *dgp_state, diDecoder *newd) 
{ 
  DGP_PROLOGUE("diNewDecoderEither()", DGP_PERRSTATE); 
 
    if (huff) { 
      DGP_ERR_PASS(diNewDecoder(spec, file, dgp_state, newd)); 
	} 
    else { 
#ifndef	NOMQ 
	memset(contexts, 0, DICONTEXTSIZE); 
	DGP_ERR_PASS(diNewDecoderMQ(mqdec, contexts, dgp_state, newd)); 
#else 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
	} 
 
    NORMAL_RETURN; 
 
#ifdef	NOMQ 
  ERR_CLEANUP: 
    ERR_RETURN; 
#endif 
    } 
 
ErrCode 
diDecode(diDecoder dec, Dgp_Bool *oob, Dgp_Int32 *val) 
{ 
  DGP_PROLOGUE("diDecode()", dec->DGP_PERRSTATE); 
 
    Dgp_Int32 v; 
    Dgp_UInt32 tmpv; 
 
    if (dec->mqcoder) { 
#ifndef	NOMQ 
	/* MQ-based decode */ 
#define	DECODE \
	do { \
	    bit = DgpEncode_mqdDecode(dec->mqdec, &dec->contexts[cx]); \
	    cx = ((cx & 0x100) | (cx << 1) | bit) & 0x1ff; \
	    DGP_REPORTF1(DGP_ENC, "%d", bit); \
	    } while (0) 
	int cx = 1; 
	Dgp_Bool s; 
	int bit; 
	int base, nbits; 
	int i; 
 
	/* 
	 * FlowDgp_Int8t is: 
	 * Decode sign bit 
	 * Decode a bit; if it's zero, decode 2 more bits and 
	 * end up with 0-3 
	 * Otherwise, decode a bit; if it's zero, decode 4 bits, add 4, 
	 * and end up with 4-19 
	 * Otherwise, decode a bit; if it's zero, decode 6 bits, add 20, 
	 * and end up with 20-83 
	 * Otherwise, decode a bit; if it's zero, decode 8 bits, add 84, 
	 * and end up with 84-339 
	 * Otherwise, decode a bit; if it's zero, decode 12 bits, add 340, 
	 * and end up with 340-4435 
	 * Otherwise, decode 32 bits and add 4436. 
	 * In any case, apply the sign bit (0 = +ve, 1 = -ve), and 
	 * interpret -0 as OOB. 
	 */ 
	DECODE; 
	s = bit; 
 
	DECODE; 
	if (bit) { 
	    /* >= 4 */ 
	    DECODE; 
	    if (bit) { 
		/* >= 20 */ 
		DECODE; 
		if (bit) { 
		    /* >= 84 */ 
		    DECODE; 
		    if (bit) { 
			/* >= 340 */ 
			DECODE; 
			if (bit) { 
			    /* >= 4436; 32 bits */ 
			    base = 4436; 
			    nbits = 32; 
			    } 
			else { 
			    /* < 4435; 12 bits */ 
			    base = 340; 
			    nbits = 12; 
			    } 
			} 
		    else { 
			/* < 340; 8 bits */ 
			base = 84; 
			nbits = 8; 
			} 
		    } 
		else { 
		    /* < 84; 6 bits */ 
		    base = 20; 
		    nbits = 6; 
		    } 
		} 
	    else { 
		/* < 20; 4 bits */ 
		base = 4; 
		nbits = 4; 
		} 
	    } 
	else { 
	    /* < 4; 2 bits */ 
	    base = 0; 
	    nbits = 2; 
	    } 
 
	v = 0; 
	for (i = 0; i < nbits; i++) { 
	    DECODE; 
	    v = (v << 1) | bit; 
	    } 
	v += base; 
 
	/* Now have sign and V */ 
	if (s) { 
	    /* Negative or OOB */ 
	    if (v == 0) { 
		*oob = DGP_TRUE; 
		} 
	    else { 
		*oob = DGP_FALSE; 
		} 
	    v = -v; 
	    } 
	else { 
	    *oob = DGP_FALSE; 
	    } 
 
	/* MQ decoder never reports EOF */ 
	*val = v; 
	NORMAL_RETURN; 
#else 
	*val = 0; 
	NORMAL_RETURN; 
#endif 
	} 
    else { 
	/* Huffman-based decode */ 
	table_line *line = (table_line *)NULL; 
 
	DGP_ERR_PASS_CLEANUP(jhtReadCode(dec->hufftable, dec->f, dec->dgp_state, 
					 (void **)&line)); 
 
	if (line == dec->lower_line) { 
	    DGP_ERR_PASS_CLEANUP(jbfReadBits(dec->f, 32, &tmpv)); 
	    v = tmpv; 
 
	    *oob = DGP_FALSE; 
	    *val = (dec->lower_line->base - v); 
	    NORMAL_RETURN; 
	    } 
 
	if (line == dec->upper_line) { 
	    DGP_ERR_PASS_CLEANUP(jbfReadBits(dec->f, 32, &tmpv)); 
	    v = tmpv; 
 
	    *oob = DGP_FALSE; 
	    *val = (v + dec->upper_line->base); 
	    NORMAL_RETURN; 
	    } 
 
	if (line == dec->oob_line) { 
	    *oob = DGP_TRUE; 
	    *val = 0; 
	    NORMAL_RETURN; 
	    } 
 
	DGP_ERR_PASS_CLEANUP(jbfReadBits(dec->f, line->rangelen, &tmpv)); 
	v = tmpv; 
 
	*oob = DGP_FALSE; 
	*val = (v + line->base); 
	NORMAL_RETURN; 
	} 
 
 ERR_CLEANUP: 
 
    *oob = DGP_FALSE; 
 
    ERR_RETURN; 
    } 
 
void 
diFreeDecoder(diDecoder dec) 
{ 
  Dgp_Free tmp_free = dec->DGP_FREE; 
    (tmp_free)((void *)dec); 
    } 
 
