/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/decpageinfo.c,v 1.11 2003/01/27 21:48:33 awilliam Exp $ 
 * 
 */ 
 
#include "util.h" 
#include "datasink.h" 
#include "segment.h" 
#include "jb2file.h" 
#include "decpageinfo.h" 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/decpageinfo.c,v 1.11 2003/01/27 21:48:33 awilliam Exp $"); 
 
/* 
 * This file contains routines for decoding a few of the "utility" segments. 
 * You can decode page info, end of page, end of stripe, and end of file 
 * segments.  It also contains the corresponding "free" routines (although 
 * none of them do anything). 
 */ 
 
ErrCode 
jb2DecodePageInfo(jb2File file, DataSource data, 
		  jb2Segment *segment, jb2Segment **refsegs) 
{ 
  DGP_PROLOGUE("jb2DecodePageInfo()", jb2_pErrState(file)); 
 
    Dgp_UInt8 lbuf[4]; 
    Dgp_UInt8 cbuf; 
    Dgp_UInt16 sbuf; 
    size_t bytes; 
 
    refsegs = refsegs; 
 
    if (segment->nrefsegs > 0) { 
	FILEERR(cErrCompressionError, "page info segment referencing another segment"); 
	} 
 
    /* Page bitmap width: 4 bytes */ 
    DSREAD(data, 4, lbuf, &bytes); 
    segment->data.page.width = (lbuf[0] << 24) | (lbuf[1] << 16) | 
	(lbuf[2] << 8) | lbuf[3]; 
 
#ifdef	VALPRINT 
    fprintf(stderr, "PAGEWIDTH %ld\n", segment->data.page.width); 
#endif 
 
    /* Page bitmap height: 4 bytes */ 
    DSREAD(data, 4, lbuf, &bytes); 
    segment->data.page.height = (lbuf[0] << 24) | (lbuf[1] << 16) | 
	(lbuf[2] << 8) | lbuf[3]; 
 
#ifdef	VALPRINT 
    fprintf(stderr, "PAGEHEIGHT %ld\n", segment->data.page.height); 
#endif 
 
    /* Page X resolution: 4 bytes */ 
    DSREAD(data, 4, lbuf, &bytes); 
    segment->data.page.xres = (lbuf[0] << 24) | (lbuf[1] << 16) | 
	(lbuf[2] << 8) | lbuf[3]; 
 
#ifdef	VALPRINT 
    fprintf(stderr, "PAGEXRES %ld\n", segment->data.page.xres); 
#endif 
 
    /* Page Y resolution: 4 bytes */ 
    DSREAD(data, 4, lbuf, &bytes); 
    segment->data.page.yres = (lbuf[0] << 24) | (lbuf[1] << 16) | 
	(lbuf[2] << 8) | lbuf[3]; 
 
#ifdef	VALPRINT 
    fprintf(stderr, "PAGEYRES %ld\n", segment->data.page.yres); 
#endif 
 
    /* Page segment flags: 1 byte */ 
    DSGETC(data, &cbuf); 
#ifdef	VALPRINT 
    fprintf(stderr, "PAGEFLAGS %02x\n", cbuf); 
#endif 
 
    /* Bit 0: page is eventually lossless */ 
    segment->data.page.lossless = (cbuf & 1) ? DGP_TRUE : DGP_FALSE; 
 
    /* Bit 1: page may contain refinements */ 
    segment->data.page.may_refine = (cbuf & 2) ? DGP_TRUE : DGP_FALSE; 
 
    /* Bit 2: Page default pixel value */ 
    segment->data.page.defpixel = (cbuf & 4) ? DGP_TRUE : DGP_FALSE; 
 
    /* Bits 3-4: Page default combination operator */ 
    segment->data.page.combop = (cbuf >> 3) & 3; 
 
    /* Bit 5: Page requires auxiliary bitmaps */ 
    segment->data.page.may_have_auxbuf = (cbuf & 0x20) ? DGP_TRUE : DGP_FALSE; 
 
    /* Bit 6: Page combination operator overridden */ 
    segment->data.page.may_override_combop = (cbuf & 0x40) ? DGP_TRUE : DGP_FALSE; 
 
    /* Bit 7: must be zero */ 
    if (cbuf & 0x80) { 
	FILEERR(cErrCompressionError, "reserved flag is 1 in page info segment"); 
	} 
 
    /* Page striping information: 2 bytes */ 
    DSREAD(data, 2, lbuf, &bytes); 
    sbuf = (lbuf[0] << 8) | lbuf[1]; 
#ifdef	VALPRINT 
    fprintf(stderr, "PAGESTRIPE %04x\n", sbuf); 
#endif 
 
    /* Bit 15: page is striped */ 
    if (sbuf & 0x8000) { 
	segment->data.page.striped = DGP_TRUE; 
	segment->data.page.max_stripe = sbuf & 0x7fff; 
	} 
    else { 
	segment->data.page.striped = DGP_FALSE; 
	segment->data.page.max_stripe = 0; 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
 
/* 
 * This is a bit of an exception: it doesn't fill in real_height in 
 * segment->data.eopage, as that information isn't actually encoded in 
 * the eopage segment, but in the preceding eostripe segment. 
 */ 
ErrCode 
jb2DecodeEOPage(jb2File file, DataSource data, 
		jb2Segment *segment, jb2Segment **refsegs) 
{ 
  DGP_PROLOGUE("jb2DecodeEOPage()", jb2_pErrState(file)); 
 
    data = data; 
    refsegs = refsegs; 
 
    if (segment->nrefsegs > 0) { 
	FILEERR(cErrCompressionError, "end of page segment referencing another segment"); 
	} 
 
    /* No data */ 
    if (segment->datalen > 0) { 
	FILEERR(cErrCompressionError, "end of page segment has data!"); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
 
ErrCode 
jb2DecodeEOStripe(jb2File file, DataSource data, 
		  jb2Segment *segment, jb2Segment **refsegs) 
{ 
  DGP_PROLOGUE("jb2DecodeEOStripe()", jb2_pErrState(file)); 
 
    Dgp_UInt8 lbuf[4]; 
    size_t bytes; 
 
    refsegs = refsegs; 
 
    if (segment->nrefsegs > 0) { 
	FILEERR(cErrCompressionError, "end of stripe segment referencing another segment"); 
	} 
 
    /* Last row: 4 bytes */ 
    DSREAD(data, 4, lbuf, &bytes); 
    segment->data.eostripe.last_row = (lbuf[0] << 24) | (lbuf[1] << 16) | 
	(lbuf[2] << 8) | lbuf[3]; 
 
#ifdef	VALPRINT 
    fprintf(stderr, "EOSTRIPE %ld\n", segment->data.eostripe.last_row); 
#endif 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
 
 
ErrCode 
jb2DecodeEOFile(jb2File file, DataSource data, 
		jb2Segment *segment, jb2Segment **refsegs) 
{ 
  DGP_PROLOGUE("jb2DecodeEOFile()", jb2_pErrState(file)); 
 
    data = data; 
    refsegs = refsegs; 
 
    if (segment->nrefsegs > 0) { 
	FILEERR(cErrCompressionError, "end of file segment referencing another segment"); 
	} 
 
    /* No data */ 
    if (segment->datalen > 0) { 
	FILEERR(cErrCompressionError, "end of file segment has data!"); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
