/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/decsegment.c,v 1.3 2003/02/10 22:53:26 mkrocken Exp $ 
 * 
 */ 
 
/* 
 * Routines that deal with JBIG2 segments.  These routines read 
 * segment headers, encapsulate the data obtained by decoding segments, 
 * deal with inter-segment references, and so on. 
 */ 
 
#include "util.h" 
#include "datasource.h" 
#include "segment.h" 
#include <sys/types.h> 
#include "decsymdict.h" 
#include "decsymreg.h" 
#include "dechtdict.h" 
#include "dechtreg.h" 
#include "decgenreg.h" 
#include "decgenrefreg.h" 
#include "dechufftable.h" 
#include "decpageinfo.h" 
#include "decextseg.h" 
#include <stdlib.h> 
#include "jb2file.h" 
#include "segment_util.h" 
#include "blit.h" 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/decsegment.c,v 1.3 2003/02/10 22:53:26 mkrocken Exp $"); 
 
 
 
/* Dummy declarations for the moment */ 
#define	DUMMY(fname) static ErrCode fname(jb2File file, DataSource data, jb2Segment *segment, jb2Segment **refsegs) {file=file;data=data;segment=segment;refsegs=refsegs;return(DGP_FALSE);} 
DUMMY(jb2DecodeProfiles) 
 
/* A table of decoding routines indexed by segment type */ 
static ErrCode (*decoders [64])(jb2File, DataSource, 
				jb2Segment*, jb2Segment**) = { 
    jb2DecodeSymDict, NULL, NULL, NULL, 
    jb2DecodeSymReg, NULL, jb2DecodeSymReg, jb2DecodeSymReg, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    jb2DecodeHTDict, NULL, NULL, NULL, 
    jb2DecodeHTReg, NULL, jb2DecodeHTReg, jb2DecodeHTReg, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    jb2DecodeGenReg, NULL, jb2DecodeGenReg, jb2DecodeGenReg, 
    jb2DecodeGenRefReg, NULL, jb2DecodeGenRefReg, jb2DecodeGenRefReg, 
    NULL, NULL, NULL, NULL, 
 
    jb2DecodePageInfo, jb2DecodeEOPage, jb2DecodeEOStripe, jb2DecodeEOFile, 
    jb2DecodeProfiles, jb2DecodeHuffTable, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, jb2DecodeExtension, NULL 
    }; 
 
 
/* A table of rendering routines indexed by segment type */ 
static ErrCode (*renderers [64])(jb2File, jb2Segment *) = { 
    NULL, NULL, NULL, NULL, 
    jb2RenderSymReg, NULL, jb2RenderSymReg, jb2RenderSymReg, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    jb2RenderHTReg, NULL, jb2RenderHTReg, jb2RenderHTReg, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    jb2RenderGenReg, NULL, jb2RenderGenReg, jb2RenderGenReg, 
    jb2RenderGenRefReg, NULL, jb2RenderGenRefReg, jb2RenderGenRefReg, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL 
    }; 
 
/* A table of render-into-page routines indexed by segment type */ 
static ErrCode (*intorenderers [64])(jb2File, jb2Segment *, DgpImage_Packed) = { 
    NULL, NULL, NULL, NULL, 
    jb2RenderSymRegIntoPage, NULL, jb2RenderSymRegIntoPage, jb2RenderSymRegIntoPage, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    jb2RenderHTRegIntoPage, NULL, jb2RenderHTRegIntoPage, jb2RenderHTRegIntoPage, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    jb2RenderGenRegIntoPage, NULL, jb2RenderGenRegIntoPage, jb2RenderGenRegIntoPage, 
    jb2RenderGenRefRegIntoPage, NULL, jb2RenderGenRefRegIntoPage, jb2RenderGenRefRegIntoPage, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL 
    }; 
 
/* 
 * Given the information already present in *segment, which will have 
 * been filled in by jb2DecodeSegmentHeader, decode the segment's data, 
 * as much as possible, to some intermediate form.  In some cases, this 
 * may not be much at all. 
 * 
 * This routine is basically a dispatcher: look at the segment type and 
 * call the appropriate function. 
 */ 
ErrCode 
jb2DecodeSegment(jb2File file, DataSource data, 
		 jb2Segment *segment, jb2Segment **refsegs) 
{ 
  DGP_PROLOGUE("jb2DecodeSegment()", jb2_pErrState(file)); 
 
    if (decoders[segment->segtype] == 
	(ErrCode (*)(jb2File, DataSource, jb2Segment *, jb2Segment **))NULL) { 
	/* Bad segment type */ 
	FILEERR(cErrCompressionError, "bad segment type"); 
	} 
 
    DGP_ERR_PASS((*decoders[segment->segtype])(file, data, segment, refsegs)); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* 
 * Read and parse a segment header.  Fill in *segment. 
 * Allocate refsegs and refretain as part of this. 
 * Return DGP_FALSE on failure. 
 */ 
ErrCode 
jb2DecodeSegmentHeader(jb2File file, DataSource data, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2DecodeSegmentHeader()", jb2_pErrState(file)); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
    Dgp_UInt8 lbuf[4]; 
    Dgp_UInt8 cbuf; 
    Dgp_Bool smallpage; 
    Dgp_Int32 i, j; 
    size_t bytes; 
 
    segment->refsegs = (Dgp_Int32 *)NULL; 
    segment->refretain = (Dgp_Bool *)NULL; 
     
    /* Start reading bits of the header */ 
 
    /* Read the 4-byte segment number */ 
    DSREAD(data, 4, lbuf, &bytes); 
 
    segment->segnum = 
	(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGNUM %d\n", segment->segnum); 
#endif 
 
    /* Read the one-byte segment header flags */ 
    DSGETC(data, &cbuf); 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGFLAGS %02x\n", cbuf); 
#endif 
 
    /* and split out the bit-level fields */ 
    /* 
     * Bits 0-5: segment type 
     * Bit 6: page association field is 1 or 4 bytes Dgp_Int32 
     * Bit 7: deferred non-retain 
     */ 
    segment->segtype = cbuf & 63; 
    smallpage = (cbuf & 64) ? DGP_FALSE : DGP_TRUE; 
    segment->defer_nonretain = (cbuf & 128) ? DGP_TRUE : DGP_FALSE; 
     
    /* Read the first byte of the referenced segment count and retain flags */ 
    DSGETC(data, &cbuf); 
 
    if ((cbuf >> 5) == 7) { 
	/* It's in Dgp_Int32 form */ 
	/* 
	 * First four bytes are referenced segment count | (7 << 29) 
	 * Next byte has 8 retention flags, with the LSB being the one 
	 * for this segment, the next-to-LSB being the one for the 
	 * first referenced segment, and so on. 
	 */ 
	/* Mask the three high bits from the count field */ 
	lbuf[0] = (Dgp_UInt8)(cbuf & ~(7 << 5)); 
 
	/* Get the rest of the count */ 
	DSREAD(data, 3, lbuf + 1, &bytes); 
 
	segment->nrefsegs =  
	    (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
 
	/* Allocate the retention bits */ 
	segment->refretain = (Dgp_Bool *)(a->malloc)(segment->nrefsegs * 
					      sizeof(*segment->refretain)); 
	if (segment->refretain == (Dgp_Bool *)NULL) { 
	    NOMEM; 
	    } 
 
	/* and read them.  Reference# -1 is this segment. */ 
	for (i = -1; i < segment->nrefsegs; i += 8) { 
	    DSGETC(data, &cbuf); 
 
	    for (j = i; j < MIN(i + 8, segment->nrefsegs); j++, cbuf >>= 1) { 
		if (j == -1) { 
		    /* Self-retention bit */ 
		    segment->retain = (cbuf & 1) ? DGP_TRUE : DGP_FALSE; 
		    } 
		else { 
		    /* Retain bit for some other segment */ 
		    segment->refretain[j] = (cbuf & 1) ? DGP_TRUE : DGP_FALSE; 
		    } 
		} 
	    } 
 
	} 
    else { 
	/* It's in Dgp_Int16 form */ 
	/* 
	 * Short form is: 
	 * Bits 5-7: count of referenced segments 
	 * Bits 0-4: retention bits 
	 */ 
	/* Get the count field */ 
	segment->nrefsegs = cbuf >> 5; 
 
	/* Allocate the retention bits */ 
	/* 30.4.2001 / dhtesq. */
	/* Inserted trap for 0 nrefsegs (else we "leak" 0 bytes)... */
	if(segment->nrefsegs){
		segment->refretain = (Dgp_Bool *)(a->malloc)(segment->nrefsegs * 
			sizeof(*segment->refretain)); 
		if (segment->refretain == (Dgp_Bool *)NULL) { 
			NOMEM; 
		} 
	}
 
	/* Read the self-retention bit */ 
	segment->retain = (cbuf & 1) ? DGP_TRUE : DGP_FALSE; 
	cbuf >>= 1; 
 
	for (j = 0; j < segment->nrefsegs; j++, cbuf >>= 1) { 
	    /* Retain bit for some other segment */ 
	    segment->refretain[j] = (cbuf & 1) ? DGP_TRUE : DGP_FALSE; 
	    } 
	} 
 
    /* Now we have the segment reference count - allocate space for the refs */ 
	/* 30.4.2001 / dhtesq. */
	/* Inserted trap for 0 nrefsegs (else we "leak" 0 bytes)... */
 	if(segment->nrefsegs){
		segment->refsegs = (Dgp_Int32 *)(a->malloc)(segment->nrefsegs * 
			sizeof(*segment->refsegs)); 
		if (segment->refsegs == (Dgp_Int32 *)NULL) { 
			NOMEM; 
		} 
	}
 
    /* Read all the segment references */ 
    for (i = 0; i < segment->nrefsegs; i++) { 
	/* 
	 * Each field is 1, 2 or 4 bytes Dgp_Int32. 
	 * If the current segment's number is <= 256, they're one byte Dgp_Int32. 
	 * If the current segment's number is <= 65536, they're two bytes Dgp_Int32. 
	 * Otherwise, they're 4 bytes Dgp_Int32. 
	 */ 
	if (segment->segnum <= 256) { 
	    DSGETC(data, &cbuf); 
 
	    segment->refsegs[i] = cbuf; 
	    } 
	else if (segment->segnum <= 65536) { 
	    DSREAD(data, 2, lbuf, &bytes); 
 
	    segment->refsegs[i] = (lbuf[0] << 8) | lbuf[1]; 
	    } 
	else { 
	    DSREAD(data, 4, lbuf, &bytes); 
 
	    segment->refsegs[i] = 
		(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
	    } 
#ifdef	VALPRINT 
	fprintf(stderr, "SEGREF %d\n", segment->refsegs[i]); 
#endif 
 
	if (segment->refsegs[i] >= segment->segnum) { 
	    /* Shouldn't happen! */ 
	    FILEERR(cErrCompressionError, "bad referenced segment number"); 
	    } 
	} 
 
    /* Read the segment page association field - 1 or 4 bytes */ 
    if (smallpage) { 
	/* One byte */ 
	DSGETC(data, &cbuf); 
 
	segment->pagenum = cbuf; 
	} 
    else { 
	DSREAD(data, 4, lbuf, &bytes); 
 
	segment->pagenum = 
	    (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
	} 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGPAGE %d\n", segment->pagenum); 
#endif 
	 
    /* Read the data length - 4 bytes */ 
    DSREAD(data, 4, lbuf, &bytes); 
 
    segment->datalen = 
	(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGDATALEN %ld\n", segment->datalen); 
#endif 
 
    /* Fill in the "is it rendered yet?" bit */ 
    segment->reginfo.rendered = DGP_FALSE; 
    segment->reginfo.im = (DgpImage_Packed)NULL; 
 
    NORMAL_RETURN; 
		 
  ERR_CLEANUP: 
    if (segment->refsegs != (Dgp_Int32 *)NULL) { 
	(a->free)((void *)segment->refsegs); 
	segment->refsegs = (Dgp_Int32 *)NULL; 
	} 
 
    if (segment->refretain != (Dgp_Bool *)NULL) { 
	(a->free)((void *)segment->refretain); 
	segment->refretain = (Dgp_Bool *)NULL; 
	} 
 
    ERR_RETURN; 
    } 
 
 
/* Render the region into a buffer */ 
ErrCode 
jb2RenderRegion(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2RenderRegion()", jb2_pErrState(file)); 
 
  DGP_ERR_PASS((*renderers[segment->segtype])(file, segment)); 
 
  NORMAL_RETURN; 
    } 
 
/* Render the region into the page */ 
ErrCode 
jb2RenderRegionIntoPage(jb2File file, jb2Segment *segment, DgpImage_Packed page) 
{ 
  DGP_PROLOGUE("jb2RenderRegion()", jb2_pErrState(file)); 
 
  DGP_ERR_PASS((*intorenderers[segment->segtype])(file, segment, page)); 
 
  NORMAL_RETURN; 
    } 
