/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/decsegment_util.c,v 1.2 2003/01/27 21:48:33 awilliam Exp $ 
 * 
 */ 
 
/* 
 * Convenience routines for dealing with segments. 
 */ 
 
#include "util.h" 
#include "segment.h" 
#include "segment_util.h" 
#include "hufftable.h" 
#include "list.h" 
#include "PsgFix.h"
#include <stdlib.h> 
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/decsegment_util.c,v 1.2 2003/01/27 21:48:33 awilliam Exp $"); 
 
/* 
 * Find all the Huffman table segments in refsegs, and assign them (in order) 
 * to *usertables.  The number of Huffman table segments in refsegs must 
 * be exactly nusertables. 
 */ 
ErrCode 
findUserTables(hufftable_spec ***usertables, Dgp_Int32 nusertables, 
	       jb2Segment **refsegs, Dgp_Int32 nrefsegs, Dgp_State *dgp_state) 
{ 
    DGP_PROLOGUE("findUserTables()", DGP_PERRSTATE); 
    Dgp_Int32 i; 
    Dgp_Int32 ntables = 0; 
 
    for (i = 0; i < nrefsegs; i++) { 
	if (refsegs[i]->segtype == JB2SEGTYPE_TABLE) { 
	  /* More referenced tables than needed */ 
	  DGP_ERR_GEN((ntables >= nusertables), cErrInvalidParm); 
	    *usertables[ntables++] = &refsegs[i]->data.table.table; 
	    } 
	} 
 
    /* Not enough referenced tables */ 
    DGP_ERR_GEN((ntables != nusertables), cErrInvalidParm); 
 
    NORMAL_RETURN; 
    } 
 
/* 
 * Decode the region segment information. 
 */ 
ErrCode 
jb2DecodeRegSegment(jb2File file, DataSource data, jb2RegSegment *reginfo) 
{ 
  DGP_PROLOGUE("jb2DecodeRegSegment()", jb2_pErrState(file)); 
 
    Dgp_UInt8 lbuf[4]; 
    Dgp_UInt8 cbuf; 
    size_t bytes; 
 
    DSREAD(data, 4, lbuf, &bytes); 
    reginfo->regwidth = 
	(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
 
    DSREAD(data, 4, lbuf, &bytes); 
    reginfo->regheight = 
	(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
 
    DSREAD(data, 4, lbuf, &bytes); 
    reginfo->regxpos = 
	(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
 
    DSREAD(data, 4, lbuf, &bytes); 
    reginfo->regypos = 
	(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
 
    DSGETC(data, &cbuf); 
 
    reginfo->combop = cbuf & 7; 
    if (reginfo->combop > 4) { 
	FILEERR(cErrCompressionError, "bad combination operator"); 
	} 
 
    if ((cbuf & (~7)) != 0) { 
	FILEERR(cErrCompressionError, "bad region flags"); 
	} 
 
#ifdef	VALPRINT 
    fprintf(stderr, "REGWIDTH %ld\n", reginfo->regwidth); 
    fprintf(stderr, "REGHEIGHT %ld\n", reginfo->regheight); 
    fprintf(stderr, "REGXPOS %ld\n", reginfo->regxpos); 
    fprintf(stderr, "REGYPOS %ld\n", reginfo->regypos); 
    fprintf(stderr, "REGFLAGS %02x\n", cbuf); 
#endif 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* 
 * Given a segment (header only), and a DgpAdt_List of jb2Segment * 
 * values, allocate an array of jb2Segment * pointing to those 
 * segments.  If segment references no other segments, 
 * it's OK to return NULL; otherwise NULL is an error return. 
 * 
 * The caller is responsible for freeing the returned array. 
 * 
 * The DgpAdt_List is assumed to be sorted in order of increasing segment number, 
 * but the segment numbers in segment's refsegs array can be in any 
 * order. 
 */ 
ErrCode 
jb2MarshalSegments(jb2File file, jb2Segment *segment, DgpAdt_List allsegs, 
		   jb2Segment ***_refsegs) 
{ 
  DGP_PROLOGUE("jb2MarshalSegments()", jb2_pErrState(file)); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
    DgpAdt_ListNode ln; 
    jb2Segment **refsegs = (jb2Segment **)NULL; 
    jb2Segment *curseg = (jb2Segment *)NULL; 
    Dgp_Int32 i; 
 
    *_refsegs = (jb2Segment **)NULL; 
     
    if (segment->nrefsegs == 0) { 
      NORMAL_RETURN; 
	} 
 
    refsegs = (jb2Segment **)(a->malloc)(segment->nrefsegs * sizeof(*refsegs)); 
    if (refsegs == (jb2Segment **)NULL) { 
      NOMEM; 
	} 
 
    ln = DGP_NULLLISTNODE; 
    for (i = 0; i < segment->nrefsegs; i++) { 
	/* Search for segment number refsegs[i] */ 
	if ((i == 0) || (segment->refsegs[i] < segment->refsegs[i - 1])) { 
	    /* Start from the beginning of the list */ 
	    ln = DGPADT_LIFIRST(allsegs); 
	    } 
	else { 
	    /* 
	     * It's greater than or equal to the previous one - 
	     * keep going from where we found the previous one 
	     */ 
	    } 
 
	while (ln != DGP_NULLLISTNODE) { 
	    curseg = (jb2Segment *)DGPADT_LIGET(ln); 
	    if (curseg->segnum == segment->refsegs[i]) { 
		break; 
		} 
	    ln = DGPADT_LINEXT(ln); 
	    } 
	DGP_CHECK_NULL_CLEANUP(ln); 
 
	refsegs[i] = curseg; 
	} 
 
    *_refsegs = refsegs; 
    refsegs = (jb2Segment **)NULL; 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (refsegs != (jb2Segment **)NULL) { 
	(a->free)((void *)refsegs); 
	refsegs = NULL;
	} 
 
    ERR_RETURN; 
    } 
 
 
