/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/decsymdict.c,v 1.31 2003/01/27 21:48:33 awilliam Exp $ 
 * 
 */ 
 
 
#include "segment.h" 
#include "decsymdict.h" 
#include "jbitfile.h" 
#include "datasource.h" 
#include "jb2file.h" 
#include "fileerr.h" 
#include "segment_util.h" 
#include <stdlib.h> 
#include "decint.h" 
#include "decgenreg.h" 
#include "decgenrefreg.h" 
#include "decsymreg.h" 
#include "hufftable.h" 
#include "stdhuff.h" 
#include <string.h> 
#include "mqcoder.h" 
#include "symdec.h" 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/decsymdict.c,v 1.31 2003/01/27 21:48:33 awilliam Exp $"); 
 
static ErrCode decodeSymbols(jb2File file, DataSource data, jbitfile bf, 
			     DgpEncode_MQDecoder *mqdec, arith_bitmap_params *abparams, 
			     Dgp_UInt8 *direct_contexts, 
			     arith_ref_params *rabparams, 
			     Dgp_UInt8 *ref_contexts, 
			     Dgp_Bool sdhuff, Dgp_Bool sdrefagg, 
			     diDecoder dh, diDecoder dw, diDecoder bmsize, 
			     diDecoder agginst, symDecoder symdec, 
			     Dgp_Int32 nnewsyms, jb2Symbol *newsyms); 
 
static ErrCode dupSymbols(jb2Symbol *dest, jb2Symbol *src, Dgp_Int32 nsyms, 
			  Dgp_State *dgp_state); 
static void copySymbols(jb2Symbol *dest, jb2Symbol *src, Dgp_Int32 nsyms); 
 
/* Read a value from an integer decoder */ 
#ifdef	VALPRINT 
#include <stdio.h> 
 
#define	READV(coder) \
    do { \
	DGP_ERR_PASS_CLEANUP(diDecode(coder, &oob, &v)); \
	if (oob) { \
	    fprintf(stderr, #coder " OOB\n"); \
	    } \
	else { \
	    fprintf(stderr, #coder " %d\n", v); \
	    } \
	} while(0) 
 
#else 
 
#define	READV(coder) \
    do { \
	DGP_ERR_PASS_CLEANUP(diDecode(coder, &oob, &v)); \
	} while(0) 
#endif 
 
/* 
 * Decode the symbol dictionary; fill in *segment. 
 */ 
ErrCode 
jb2DecodeSymDict(jb2File file, DataSource data, 
		 jb2Segment *segment, jb2Segment **refsegs) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2DecodeSymDict()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_UInt8 lbuf[4]; 
    Dgp_Bool sdhuff, sdrefagg; 
    Dgp_Bool context_used, context_retained; 
    hufftable_spec *sdhuffdh, *sdhuffdw, *sdhuffbmsize, *sdhuffagginst; 
    Dgp_Int32 nusertables = 0; 
    hufftable_spec **usertables[4]; 
    Dgp_Int32 nexsyms = 0, nnewsyms = 0; 
    jb2Symbol *newsyms = (jb2Symbol *)NULL; 
    Dgp_Int32 i, j, k; 
    jb2Symbol *exsyms = (jb2Symbol *)NULL; 
    Dgp_Int32 currefseg, currefpos; 
    Dgp_Int32 v, skip; 
    Dgp_Bool exporting, oob; 
    Dgp_UInt16 flags; 
    Dgp_Int32 dummy; 
    arith_bitmap_params abparams; 
    arith_ref_params rabparams; 
    jbitfile bf = NULLJBITFILE, tmpbf; 
    DgpEncode_MQDecoder mqdec; 
    /* Some context buffers */ 
    Dgp_UInt8 *direct_contexts = (Dgp_UInt8 *)NULL; 
    Dgp_UInt8 *ref_contexts = (Dgp_UInt8 *)NULL; 
    size_t cxsize_direct = 0, cxsize_ref = 0; 
    Dgp_UInt8 dhcontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 dwcontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 exflagscontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 agginstcontexts[DICONTEXTSIZE]; 
    /* Integer decoders */ 
    diDecoder dh = NULLDIDECODER; 
    diDecoder dw = NULLDIDECODER; 
    diDecoder bmsize = NULLDIDECODER; 
    diDecoder agginst = NULLDIDECODER; 
    diDecoder exflags = NULLDIDECODER; 
    /* Symbol decoder for symbols in refinement/aggregate coding */ 
    symDecoder symdec = (symDecoder)NULL; 
    /* and things we need to build it */ 
    Dgp_Int32 numsyms; 
    void **symptrs = (void **)NULL; 
    Dgp_Int32 *lengths = (Dgp_Int32 *)NULL; 
    Dgp_Int32 codelen; 
    size_t bytes; 
 
    /* Parse the header */ 
 
    /* Two-byte flags field */ 
    DSREAD(data, 2, lbuf, &bytes); 
    flags = (lbuf[0] << 8) | lbuf[1]; 
#ifdef	VALPRINT 
    fprintf(stderr, "SDFLAGS %04x\n", flags); 
#endif 
 
    /* SDHUFF */ 
    sdhuff = (flags & 1) ? DGP_TRUE : DGP_FALSE; 
 
    /* SDREFAGG */ 
    sdrefagg = (flags & 2) ? DGP_TRUE : DGP_FALSE; 
 
    /* SDHUFFDH selection */ 
    if ((!sdhuff) && (((flags >> 2) & 3) != 0)) { 
	FILEERR(cErrCompressionError, "non-zero Huffman table specifier in non-Huffman symbol dictionary"); 
	} 
 
    switch ((flags >> 2) & 3) { 
      case 0: { 
	sdhuffdh = &stdhuffDH1; 
	break; 
	} 
       
      case 1: { 
	sdhuffdh = &stdhuffDH2; 
	break; 
	} 
       
      case 3: { 
	sdhuffdh = (hufftable_spec *)NULL; 
	usertables[nusertables++] = &sdhuffdh; 
	break; 
	} 
       
      default: { 
	FILEERR(cErrCompressionError, "bad SDHUFFDH specification"); 
        } 
      } 
	
    /* SDHUFFDW selection */ 
    if ((!sdhuff) && (((flags >> 4) & 3) != 0)) { 
	FILEERR(cErrCompressionError, "non-zero Huffman table specifier in non-Huffman symbol dictionary"); 
	} 
 
    switch ((flags >> 4) & 3) { 
      case 0: { 
	sdhuffdw = &stdhuffDW1; 
	break; 
	} 
       
      case 1: { 
	sdhuffdw = &stdhuffDW2; 
	break; 
	} 
       
      case 3: { 
	sdhuffdw = (hufftable_spec *)NULL; 
	usertables[nusertables++] = &sdhuffdw; 
	break; 
	} 
       
      default: { 
	FILEERR(cErrCompressionError, "bad SDHUFFDW specification"); 
        } 
      } 
 
    /* SDHUFFBMSIZE selection */ 
    if ((!sdhuff) && ((flags & 64) != 0)) { 
	FILEERR(cErrCompressionError, "non-zero Huffman table specifier in non-Huffman symbol dictionary"); 
	} 
 
    if (flags & 64) { 
	sdhuffbmsize = (hufftable_spec *)NULL; 
	usertables[nusertables++] = &sdhuffbmsize; 
	} 
    else { 
	sdhuffbmsize = &stdhuffA; 
	} 
     
    /* SDHUFFAGGINST selection */ 
    if ((!sdhuff) && ((flags & 128) != 0)) { 
	FILEERR(cErrCompressionError, "non-zero Huffman table specifier in non-Huffman symbol dictionary"); 
	} 
 
    if (flags & 128) { 
	sdhuffagginst = (hufftable_spec *)NULL; 
	usertables[nusertables++] = &sdhuffagginst; 
	} 
    else { 
	sdhuffagginst = &stdhuffA; 
	} 
     
    /* Previous dictionary context used */ 
    context_used = (flags & 256) ? DGP_TRUE : DGP_FALSE; 
    /* Current context to be retained */ 
    context_retained = (flags & 512) ? DGP_TRUE : DGP_FALSE; 
     
    if (sdhuff && (context_used || context_retained)) { 
	FILEERR(cErrCompressionError, "context used or retained in Huffman symbol dictionary"); 
	} 
 
    /* SDTEMPLATE */ 
    abparams.rtemplate = (flags >> 10) & 3; 
    /* while we're at it: TPON is always zero within symbol dictionaries */ 
    abparams.tpon = DGP_FALSE; 
 
    /* SDRTEMPLATE */ 
    rabparams.rtemplate = (flags >> 12) & 1; 
 
    if (!sdhuff) { 
	/* Decode the SDAT* values */ 
	dummy = 0; 
	DGP_ERR_PASS_CLEANUP(jb2DecodeABParams(file, data, &abparams, &dummy)); 
	} 
 
    if (sdrefagg) { 
	/* Decode the SDRAT* values */ 
	dummy = 0; 
	DGP_ERR_PASS_CLEANUP(jb2DecodeRABParams(file, data, &rabparams, 
						&dummy)); 
	rabparams.tpron = DGP_FALSE; 
	} 
 
    /* SDNUMEXSYMS - number of symbols exported */ 
    DSREAD(data, 4, lbuf, &bytes); 
    nexsyms = (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
#ifdef	VALPRINT 
    fprintf(stderr, "SDNUMEXSYMS %d\n", nexsyms); 
#endif 
 
    /* SDNUMNEWSYMS - number of symbols defined */ 
    DSREAD(data, 4, lbuf, &bytes); 
    nnewsyms = (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
#ifdef	VALPRINT 
    fprintf(stderr, "SDNUMNEWSYMS %d\n", nnewsyms); 
#endif 
 
    /* 
     * Find the actual user Huffman tables.  We still call this even if 
     * if this is a non-Huffman segment, as it checks that there are no 
     * Huffman table segments referenced. 
     */ 
    DGP_ERR_PASS_CLEANUPF0(findUserTables(usertables, nusertables, refsegs, 
					  segment->nrefsegs, dgp_state), 
			  "wrong number of Huffman table segments referenced"); 
 
    if (sdhuff) { 
	/* Check that OOB is correctly available/not available */ 
	if ((!sdhuffdw->oob_valid) | sdhuffdh->oob_valid | 
	    sdhuffbmsize->oob_valid | sdhuffagginst->oob_valid) { 
	    FILEERR(cErrCompressionError, "wrong OOB capabilities"); 
	    } 
	} 
 
    /* Allocate the space for the new symbols */ 
    newsyms = (jb2Symbol *)(a->malloc)(nnewsyms * sizeof(*newsyms)); 
 
    /* Clear their bits pointers for predictable bailout */ 
    for (i = 0; i < nnewsyms; i++) { 
	newsyms[i].bits = (Dgp_UInt32 *)NULL; 
  newsyms[i].userdata = NULL;
  newsyms[i].userflags = (Dgp_UInt8) 0;
    } 
 
    /* 
     * Allocate contexts, and optionally fill them from a referenced 
     * segment, if this segment is non-Huffman. 
     */ 
    if (!sdhuff) { 
	cxsize_direct = jb2ABContextsNeeded(&abparams); 
	/* We need to allocate contexts in any case */ 
	direct_contexts = (Dgp_UInt8 *)(a->calloc)(cxsize_direct, 1); 
	if (direct_contexts == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
	} 
 
    /* Similar for refinement/aggregation */ 
    if (sdrefagg) { 
	cxsize_ref = jb2RABContextsNeeded(&rabparams); 
	/* We need to allocate contexts in any case */ 
	ref_contexts = (Dgp_UInt8 *)(a->calloc)(cxsize_ref, 1); 
	if (ref_contexts == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
	} 
 
    if (context_used) { 
	Dgp_Int32 i; 
 
	/* 
	 * Find the contexts we're supposed to use.  They come from 
	 * the last-referenced symbol dictionary segment 
	 */ 
	for (i = segment->nrefsegs - 1; i >= 0; i--) { 
	    if (refsegs[i]->segtype == JB2SEGTYPE_SYMDICT) { 
		break; 
		} 
	    } 
 
	if (i < 0) { 
	    FILEERR(cErrCompressionError, "contexts used, but no symbol dictionary referenced"); 
	    } 
 
	if (!sdhuff) { 
	    /* It should have been the last segment referenced */ 
	    if (i != segment->nrefsegs - 1) { 
		FILEERR(cErrCompressionError, "non-Huffman symbol dictionary referencing non-symbol dictionary segment"); 
		} 
	    } 
 
	/* Grab the direct contexts */ 
	if (!sdhuff) { 
	    /* It had better have kept its contexts */ 
	    if (refsegs[segment->nrefsegs - 1]->data.symdict.direct_contexts == 
		(Dgp_UInt8 *)NULL) { 
		FILEERR(cErrCompressionError, "symbol dictionary attempting to use non-retained contexts"); 
		} 
 
	    /* Copy them all over */ 
	    memcpy(direct_contexts, 
		   refsegs[segment->nrefsegs - 1]->data.symdict.direct_contexts, 
		   cxsize_direct); 
	    } 
 
	/* Grab the refinement contexts */ 
	if (sdrefagg) { 
	    /* It had better have kept its contexts */ 
	    if (refsegs[segment->nrefsegs - 1]->data.symdict.ref_contexts == 
		(Dgp_UInt8 *)NULL) { 
		FILEERR(cErrCompressionError, "symbol dictionary attempting to use non-retained contexts"); 
		} 
 
	    /* Copy them all over */ 
	    memcpy(ref_contexts, 
		   refsegs[segment->nrefsegs - 1]->data.symdict.ref_contexts, 
		   cxsize_ref); 
	    } 
	} 
 
    /* 
     * Do the decoding.  This returns when the last symbol has been 
     * decoded, before the export flags. 
     */ 
    if (sdhuff) { 
	/* First we need a jbitfile */ 
      DGP_ERR_PASS_CLEANUP(jbfOpenFromDataSource(data, &bf)); 
	} 
    else { 
#ifndef	NOMQ 
	dsInitDgpEncode_MQDecoder(data, &mqdec); 
#else 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
	} 
 
    /* Create the various integer decoders */ 
    DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sdhuff, sdhuffdh, bf, &mqdec, 
					    dhcontexts, dgp_state, &dh)); 
 
    DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sdhuff, sdhuffdw, bf, &mqdec, 
					    dwcontexts, dgp_state, &dw)); 
 
    if ((sdhuff) && (!sdrefagg)) { 
	/* Only Huffman needs bmsize, and then only when sdrefagg is DGP_FALSE */ 
	DGP_ERR_PASS_CLEANUP(diNewDecoder(sdhuffbmsize, bf, dgp_state, 
					  &bmsize)); 
	} 
 
    if (sdrefagg) { 
	/* Need agginst */ 
      DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sdhuff, sdhuffagginst, bf, 
					      &mqdec, agginstcontexts,  
					      dgp_state, &agginst)); 
 
	/* 
	 * Also need a symbol decoder.  We should set it up now, as 
	 * it's cleverly been designed so that it doesn't change as 
	 * the symbol dictionary decoding goes on: it can refer to any 
	 * of the symbols in our list of dictionaries, plus the new 
	 * symbols we're defining. 
	 */ 
 
	/* How many symbols are in it? */ 
	numsyms = 0; 
	for (i = 0; i < segment->nrefsegs; i++) { 
	    /* Is it a dictionary segment */ 
	    if (refsegs[i]->segtype == JB2SEGTYPE_SYMDICT) { 
		/* Yes - get the count */ 
		numsyms += refsegs[i]->data.symdict.nsymbols; 
		} 
	    else { 
		/* No - better be a table */ 
		if (refsegs[i]->segtype != JB2SEGTYPE_TABLE) { 
		    FILEERR(cErrCompressionError, "bad referenced segment type"); 
		    } 
		} 
	    } 
 
	numsyms += nnewsyms; 
 
	/* Allocate an array of void *s to point to the symbols */ 
	symptrs = (void **)(a->malloc)(numsyms * sizeof(*symptrs)); 
	if (symptrs == (void **)NULL) { 
	    NOMEM; 
	    } 
 
	/* Now set the pointers */ 
	j = 0; 
	/* All the things we refer to */ 
	for (i = 0; i < segment->nrefsegs; i++) { 
	    if (refsegs[i]->segtype == JB2SEGTYPE_SYMDICT) { 
		for (k = 0; k < refsegs[i]->data.symdict.nsymbols; k++, j++) { 
		    symptrs[j] = &refsegs[i]->data.symdict.symbols[k]; 
		    } 
		} 
	    } 
 
	/* and all the things we define */ 
	for (i = 0; i < nnewsyms; i++, j++) { 
	    symptrs[j] = &newsyms[i]; 
	    } 
 
	/* If we're not Huffman-based, we have all the info we need */ 
	if (!sdhuff) { 
#ifndef	NOMQ 
	    DGP_ERR_PASS_CLEANUP(symDecoderArith(symptrs, numsyms, &mqdec,  
						 dgp_state, &symdec)); 
 
	    /* That copied symptrs */ 
	    (a->free)((void *)symptrs); 
	    symptrs = (void **)NULL; 
#endif 
	    } 
	else { 
	    /* Need an array of lengths */ 
	    lengths = (Dgp_Int32 *)(a->malloc)(numsyms * sizeof(*lengths)); 
	    if (lengths == (Dgp_Int32 *)NULL) { 
		NOMEM; 
		} 
 
	    /* all set to ceil(log2(numsyms)) */ 
	    codelen = ceillog2(numsyms); 
	    for (i = 0; i < numsyms; i++) { 
		lengths[i] = codelen; 
		} 
 
	    /* Now create the decoder */ 
	    DGP_ERR_PASS_CLEANUP(symDecoderHuffman(symptrs, numsyms, bf,  
						lengths, dgp_state, &symdec)); 
 
	    /* No more need for symptrs or lengths */ 
	    (a->free)((void *)symptrs); 
	    symptrs = (void **)NULL; 
	    (a->free)((void *)lengths); 
	    lengths = (Dgp_Int32 *)NULL; 
	    } 
	} 
 
    DGP_ERR_PASS_CLEANUP(decodeSymbols(file, data, bf, &mqdec, &abparams, 
				       direct_contexts, &rabparams, 
				       ref_contexts, sdhuff, sdrefagg, 
				       dh, dw, bmsize, agginst, symdec, 
				       nnewsyms, newsyms)); 
 
    /* Produce the export list */ 
     
    /* First alloc the exported symbols array */ 
    exsyms = (jb2Symbol *)(a->malloc)(nexsyms * sizeof(*exsyms)); 
 
    /* Create an integer decoder */ 
    DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sdhuff, &stdhuffA, bf, &mqdec, 
				 exflagscontexts, dgp_state, &exflags)); 
 
    i = 0; 
    currefseg = 0; 
    currefpos = 0; 
    exporting = DGP_FALSE; 
    while (i < nexsyms) { 
	/* Decode a run */ 
	READV(exflags); 
	 
	if (exporting) { 
	    if (i + v > nexsyms) { 
		FILEERR(cErrCompressionError, "too many symbols exported"); 
		} 
	    } 
 
	/* Handle v symbols starting from currefpos in currefseg */ 
	while (v > 0) { 
	    /* Are we reading from a referenced segment? */ 
	    if (currefseg < segment->nrefsegs) { 
		/* Yes - is it a dictionary? */ 
		if (refsegs[currefseg]->segtype != JB2SEGTYPE_SYMDICT) { 
		    /* No; better be a table; skip to the next */ 
		    if (refsegs[currefseg]->segtype != JB2SEGTYPE_TABLE) { 
			FILEERR(cErrCompressionError, "bad referenced segment type"); 
			} 
 
		    currefseg++; 
		    currefpos = 0; 
		    } 
		else { 
		    /* Anything left? */ 
		    if (currefpos == 
			refsegs[currefseg]->data.symdict.nsymbols) { 
			/* Finished this segment */ 
			currefseg++; 
			currefpos = 0; 
			} 
		    else { 
			/* It's a dictionary - handle some symbols */ 
			skip = MIN(v, 
				   refsegs[currefseg]->data.symdict.nsymbols - 
				   currefpos); 
 
			if (exporting) { 
			    /* Copy skip symbols */ 
			    DGP_ERR_PASS_CLEANUP(dupSymbols(&exsyms[i], 
				&refsegs[currefseg]->data.symdict.symbols[currefpos], 
						  skip, dgp_state)); 
			    i += skip; 
			    } 
 
			v -= skip; 
			currefpos += skip; 
			} 
		    } 
		} 
	    else { 
		/* We're working from newsyms */ 
		if (currefpos + v > nnewsyms) { 
		    FILEERR(cErrCompressionError, "too many symbols exported"); 
		    } 
 
		if (exporting) { 
		    copySymbols(&exsyms[i], &newsyms[currefpos], v); 
 
		    /* 
		     * Clear the bitmap pointers in newsyms so that there's 
		     * only one copy of each pointer at a time 
		     */ 
		    for (j = 0; j < v; j++) { 
			newsyms[currefpos + j].bits = (Dgp_UInt32 *)NULL; 
			} 
 
		    i += v; 
		    } 
 
		currefpos += v; 
		v = 0; 
		} 
	    } 
 
	/* Flip state for the next run */ 
	exporting = !exporting; 
	} 
 
    /* Don't need stuff anymore */ 
    diFreeDecoder(exflags); 
 
    if (dh != NULLDIDECODER) { 
	diFreeDecoder(dh); 
	} 
 
    if (dw != NULLDIDECODER) { 
	diFreeDecoder(dw); 
	} 
 
    if (bmsize != NULLDIDECODER) { 
	diFreeDecoder(bmsize); 
	} 
 
    if (agginst != NULLDIDECODER) { 
	diFreeDecoder(agginst); 
	} 
 
    if (symdec != (symDecoder)NULL) { 
	symDecoderFree(symdec); 
	} 
 
    (a->free)((void *)newsyms); 
    if (bf != NULLJBITFILE) { 
      tmpbf = bf; 
      bf = NULLJBITFILE; 
	DGP_ERR_PASS_CLEANUP(jbfCloseFromDataSource(tmpbf)); 
	} 
 
    /* Update the DataSource's bufptr from the DgpEncode_MQDecoder */ 
#ifndef	NOMQ 
    if (!sdhuff) { 
      DGP_ERR_PASS_CLEANUPF0(dsFinalDgpEncode_MQDecoder(data, &mqdec), 
			     "problem syncing MQ decoder"); 
	} 
#endif 
 
    /* Stick the values into the segment */ 
    segment->data.symdict.nsymbols = nexsyms; 
    segment->data.symdict.symbols = exsyms; 
 
    if (context_retained) { 
	/* Stash the bitmap contexts */ 
	segment->data.symdict.direct_contexts = direct_contexts; 
	segment->data.symdict.ref_contexts = ref_contexts; 
	} 
    else { 
	/* Don't need the bitmap contexts - free them */ 
	segment->data.symdict.direct_contexts = (Dgp_UInt8 *)NULL; 
	if (direct_contexts != (Dgp_UInt8 *)NULL) { 
	    (a->free)((void *)direct_contexts); 
	    } 
 
	segment->data.symdict.ref_contexts = (Dgp_UInt8 *)NULL; 
	if (ref_contexts != (Dgp_UInt8 *)NULL) { 
	    (a->free)((void *)ref_contexts); 
	    } 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    /* Free up anything we might have created */ 
    if (bf != NULLJBITFILE) { 
	(void)jbfCloseFromDataSource(bf); 
	} 
 
    if (newsyms != (jb2Symbol *)NULL) { 
	for (i = 0; i < nnewsyms; i++) { 
	    if (newsyms[i].bits != (Dgp_UInt32 *)NULL) { 
		(a->free)((void *)newsyms[i].bits); 
		} 
	    } 
	(a->free)((void *)newsyms); 
	} 
 
    if (dh != NULLDIDECODER) { 
	diFreeDecoder(dh); 
	} 
 
    if (dw != NULLDIDECODER) { 
	diFreeDecoder(dw); 
	} 
 
    if (bmsize != NULLDIDECODER) { 
	diFreeDecoder(bmsize); 
	} 
 
    if (agginst != NULLDIDECODER) { 
	diFreeDecoder(agginst); 
	} 
 
    if (exflags != NULLDIDECODER) { 
	diFreeDecoder(exflags); 
	} 
 
    if (symdec != (symDecoder)NULL) { 
	symDecoderFree(symdec); 
	} 
 
    if (direct_contexts != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)direct_contexts); 
	} 
 
    if (ref_contexts != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)ref_contexts); 
	} 
 
    if (exsyms != (jb2Symbol *)NULL) { 
	for (i = 0; i < nexsyms; i++) { 
	    if (exsyms[i].bits != (Dgp_UInt32 *)NULL) { 
		(a->free)((void *)exsyms[i].bits); 
		} 
	    } 
	(a->free)((void *)exsyms); 
	} 
 
    if (symptrs != (void **)NULL) { 
	(a->free)((void *)symptrs); 
	} 
 
    if (lengths != (Dgp_Int32 *)NULL) { 
	(a->free)((void *)lengths); 
	} 
 
    ERR_RETURN; 
    } 
 
 
/* Decode all the height classes. */ 
static ErrCode 
decodeSymbols(jb2File file, DataSource data, jbitfile bf, DgpEncode_MQDecoder *mqdec, 
	      arith_bitmap_params *abparams, Dgp_UInt8 *direct_contexts, 
	      arith_ref_params *rabparams, Dgp_UInt8 *ref_contexts, 
	      Dgp_Bool sdhuff, Dgp_Bool sdrefagg, 
	      diDecoder dh, diDecoder dw, diDecoder bmsize, 
	      diDecoder agginst, symDecoder symdec, 
	      Dgp_Int32 nnewsyms, jb2Symbol *newsyms) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("decodeSymbols()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Bool oob; 
    Dgp_Int32 v; 
    Dgp_Int32 classwidth; 
    Dgp_Int32 classrow_width; 
    Dgp_Int32 symnum; 
    Dgp_Int32 classstart; 
    Dgp_Int32 bitoffset; 
    Dgp_Int32 oldheight; 
    Dgp_Int32 oldwidth; 
    Dgp_Int32 g4len; 
    Dgp_Int32 ninsts = 0; 
    Dgp_UInt32 *classbits = (Dgp_UInt32 *)NULL; 
    jb2SymbolInstance *insts = (jb2SymbolInstance *)NULL; 
    Dgp_Int32 i; 
    Dgp_UInt32 tmp; 
    /* Decoders; only used if sdrefagg is DGP_TRUE */ 
    diDecoder fs = NULLDIDECODER; 
    diDecoder ds = NULLDIDECODER; 
    diDecoder dt = NULLDIDECODER; 
    diDecoder t = NULLDIDECODER; 
    diDecoder ri = NULLDIDECODER; 
    diDecoder rdx = NULLDIDECODER; 
    diDecoder rdy = NULLDIDECODER; 
    diDecoder rdw = NULLDIDECODER; 
    diDecoder rdh = NULLDIDECODER; 
    diDecoder rsize = NULLDIDECODER; 
    /* Context buffers for the integer decoders */ 
    Dgp_UInt8 fscontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 dscontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 dtcontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 tcontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 ricontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 rdxcontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 rdycontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 rdwcontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 rdhcontexts[DICONTEXTSIZE]; 
 
    /* Temporary variables for special-case one-instance aggregations */ 
    Dgp_UInt8 *refdata = (Dgp_UInt8 *)NULL; 
    jb2SymbolInstanceRefined *ref = (jb2SymbolInstanceRefined *)NULL; 
    Dgp_UInt32 *bits = (Dgp_UInt32 *)NULL; 
    size_t bytes; 
 
 
    /* 
     * For refinement/aggregate coding, we need a pile of integer decoders. 
     * These are persistent, and get passed in to each refinement/aggregate 
     * coded symbol.  It's our job to create them - lovely. 
     * Thankfully, our caller created symdec for us, or we'd have had 
     * to grovel through all the referenced segments too. 
     */ 
    if (sdrefagg) { 
	DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sdhuff, &stdhuffFS1, bf, mqdec, 
						fscontexts, dgp_state, &fs)); 
 
	DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sdhuff, &stdhuffDS1, bf, mqdec, 
						dscontexts, dgp_state, &ds)); 
 
	DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sdhuff, &stdhuffDT1, bf, mqdec, 
						dtcontexts, dgp_state, &dt)); 
 
	DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sdhuff, &stdhuffRD2, bf, mqdec, 
						rdwcontexts, dgp_state, &rdw)); 
 
	DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sdhuff, &stdhuffRD2, bf, mqdec, 
						rdhcontexts, dgp_state, &rdh)); 
 
	DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sdhuff, &stdhuffRD2, bf, mqdec, 
						rdxcontexts, dgp_state, &rdx)); 
 
	DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sdhuff, &stdhuffRD2, bf, mqdec, 
						rdycontexts, dgp_state, &rdy)); 
 
	if (sdhuff) { 
	    /* Only Huffman needs RSIZE */ 
	    DGP_ERR_PASS_CLEANUP(diNewDecoder(&stdhuffA, bf, 
			   dgp_state, &rsize)); 
 
	    /* and there's no RI decoder */ 
	    ri = NULLDIDECODER; 
 
	    /* T is special in Huffman case */ 
	    t = NULLDIDECODER; 
	    } 
	else { 
	    /* Only arithmetic needs RI */ 
#ifndef	NOMQ 
	    memset(ricontexts, 0, DICONTEXTSIZE); 
	    DGP_ERR_PASS_CLEANUP(diNewDecoderMQ(mqdec, ricontexts, dgp_state, 
						&ri)); 
 
	    rsize = NULLDIDECODER; 
 
	    /* Arithmetic needs T */ 
	    memset(tcontexts, 0, DICONTEXTSIZE); 
	    DGP_ERR_PASS_CLEANUP(diNewDecoderMQ(mqdec, tcontexts, dgp_state, 
						&t)); 
#endif 
	    } 
	} 
 
    oldheight = 0; 
    symnum = 0; 
    while (symnum < nnewsyms) { 
	/* Decode a height class */ 
 
	/* Delta height from previous height class */ 
	READV(dh); 
 
	oldheight += v; 
	if (oldheight <= 0) { 
	    FILEERR(cErrCompressionError, "bad height class height"); 
	    } 
 
	/* Read the size info for the symbols */ 
	classstart = symnum; 
 
	oldwidth = 0; 
	classwidth = 0; 
	for (;;) { 
	    READV(dw); 
 
	    if (oob) { 
		/* End of height class */ 
		break; 
		} 
 
	    oldwidth += v; 
	    classwidth += oldwidth; 
 
	    /* Check for too many symbols */ 
	    if (symnum == nnewsyms) { 
		FILEERR(cErrCompressionError, "too many symbols"); 
		} 
 
	    /* Got the symbol info, apart from the bitmap */ 
	    newsyms[symnum].width = oldwidth; 
	    newsyms[symnum].height = oldheight; 
	    newsyms[symnum].longsperline = 
		(oldwidth + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32; 
 
	    if (sdrefagg || (!sdhuff)) { 
		/* Read the symbol's bitmap */ 
 
		newsyms[symnum].refsym = (jb2Symbol *)NULL; 
		newsyms[symnum].userdata = (void *)NULL; 
 
		newsyms[symnum].bits =  
		    (Dgp_UInt32 *)(a->calloc)(oldheight, 
					    newsyms[symnum].longsperline * 
					    sizeof(Dgp_UInt32)); 
		if (newsyms[symnum].bits == (Dgp_UInt32 *)NULL) { 
		  NOMEM; 
		    } 
 
#ifndef	NOMQ 
		if (sdrefagg) { 
		    /* First, decode the instance count. */ 
		    READV(agginst); 
		    if (oob) { 
			FILEERR(cErrCompressionError, "bad OOB"); 
			} 
 
		    ninsts = v; 
 
		    /* Allocate something to decode into */ 
		    insts = (jb2SymbolInstance *)(a->malloc)(ninsts * 
							sizeof(*insts)); 
		    if (insts == (jb2SymbolInstance *)NULL) { 
		      NOMEM; 
			} 
 
		    /* Clear pointers for good bailing */ 
		    for (i = 0; i < ninsts; i++) { 
			insts[i].refdata = (jb2SymbolInstanceRefined *)NULL; 
			} 
 
		    if (ninsts > 1) { 
			/* Now decode the instances */ 
			DGP_ERR_PASS_CLEANUP(jb2DecodeInstances(file, bf, 
								sdhuff, DGP_TRUE, 
						DGP_FALSE, JB2_TOPLEFT, 
						0, 0, 
						symdec, fs, ds, dt, t, ri, 
						rdw, rdh, rdx, rdy, rsize, 
						data, mqdec, rabparams, 
						ref_contexts, ninsts, insts)); 
			} 
		    else { 
			/* 
			 * We're in special-case land here - only one 
			 * instance.  We're supposed to read a symbol ID, 
			 * a delta X and a delta Y, a bitmap size (if 
			 * Huffman), and do the refinement decoding. 
			 */ 
			jb2Symbol *sym; 
			Dgp_Int32 currdx, currdy; 
			Dgp_Int32 reflpl; 
			size_t refdatalen; 
			DgpEncode_MQDecoder huffmq; 
 
			DGP_ERR_PASS_CLEANUP(symDecode(symdec, &sym)); 
 
			READV(rdx); 
			if (oob) { 
			    FILEERR(cErrCompressionError, "bad OOB"); 
			    } 
			currdx = v; 
 
			READV(rdy); 
			if (oob) { 
			    FILEERR(cErrCompressionError, "bad OOB"); 
			    } 
			currdy = v; 
 
			/* Allocate the bitmap to refine into */ 
			reflpl = (oldwidth + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32; 
			bits = (Dgp_UInt32 *)(a->calloc)(reflpl * sizeof(Dgp_UInt32), 
								   oldheight); 
			if (bits == (Dgp_UInt32 *)NULL) { 
			  NOMEM; 
			    } 
 
			if (sdhuff) { 
			    /* EEP - we need to switch to arithmetic coding! */ 
			    READV(rsize); 
			    if (oob) { 
				FILEERR(cErrCompressionError, "bad OOB"); 
				} 
			    refdatalen = v; 
 
			    /* 
			     * rsize tells us how many bytes we need. 
			     * Copy that much out of the DataSource. 
			     */ 
			    /* Suspend the bitfile first */ 
			    DGP_ERR_PASS_CLEANUP(jbfSuspend(bf)); 
			    /* then allocate and copy the bytes */ 
			    refdata = (Dgp_UInt8 *)(a->malloc)(refdatalen); 
			    if (refdata == (Dgp_UInt8 *)NULL) { 
			      NOMEM; 
				} 
 
			    DSREAD(data, refdatalen, refdata, &bytes); 
 
			    /* Now turn the data into an MQ coder */ 
			    DgpEncode_mqdInit(&huffmq, refdata, refdatalen, 
				    (Dgp_Bool (*)(Dgp_UInt8 **, size_t *, void *))NULL, 
				    (void *)NULL, dgp_state); 
			    mqdec = &huffmq; 
			    } 
 
			DGP_ERR_PASS_CLEANUP(jb2ArithDecodeRefinement(bits, oldwidth, oldheight, reflpl, 
						      sym->bits, 
						      sym->width, sym->height, 
						      sym->longsperline, 
						      currdx, 
						      currdy, 
						      rabparams, ref_contexts, 
								      file, mqdec)); 
 
#if	VERBLEVEL >= 100 
			{ 
			    Dgp_Int32 xx, yy; 
			    for (yy = 0; yy < sym->height; yy++) { 
				for (xx = 0; xx < sym->width; xx++) { 
				    if (sym->bits[yy * sym->longsperline + (xx / DGP_BITS_UINT32)] & 
					(1ul << (DGP_BITS_UINT32 - 1 - (xx & (DGP_BITS_UINT32 - 1))))) { 
					fputc('1', stderr); 
					} 
				    else { 
					fputc('0', stderr); 
					} 
				    } 
				fputc('\n', stderr); 
				} 
			    fprintf(stderr, "   ---> @(XXX, XXX)\n"); 
			    for (yy = 0; yy < oldheight; yy++) { 
				for (xx = 0; xx < oldwidth; xx++) { 
				    if (bits[yy * reflpl + (xx / DGP_BITS_UINT32)] & 
					(1ul << (DGP_BITS_UINT32 - 1 - (xx & (DGP_BITS_UINT32 - 1))))) { 
					fputc('1', stderr); 
					} 
				    else { 
					fputc('0', stderr); 
					} 
				    } 
				fputc('\n', stderr); 
				} 
			    }				 
#endif 
			/* Hook the decoded info into the symbol instance */ 
			ref = (jb2SymbolInstanceRefined *)(a->malloc)(sizeof(*ref)); 
			if (ref == (jb2SymbolInstanceRefined *)NULL) { 
			  NOMEM; 
			    } 
 
			ref->bits = bits; 
			ref->longsperline = reflpl; 
			ref->width = oldwidth; 
			ref->height = oldheight; 
 
			insts[0].xpos = insts[0].ypos = 0; 
			insts[0].sym = sym; 
			insts[0].refdata = ref; 
			insts[0].userdata = (void *)NULL; 
 
			/* Hooked in - clear redundant pointers */ 
			ref = (jb2SymbolInstanceRefined *)NULL; 
			bits = (Dgp_UInt32 *)NULL; 
 
			if (sdhuff) { 
			    /* Switch back to Huffman coding */ 
			    (a->free)((void *)refdata); 
			    refdata = (Dgp_UInt8 *)NULL; 
 
			    jbfResume(bf); 
			    } 
			} 
 
		    /* Now we need to draw the instances. */ 
		    jb2DrawInstances(newsyms[symnum].bits, 
				     newsyms[symnum].width, 
				     newsyms[symnum].height, 
				     newsyms[symnum].longsperline, 
				     insts, ninsts, 
				     COMBOP_OR, dgp_state); 
 
		    /* Get rid of the instances */ 
		    jb2FreeInstances(insts, ninsts, dgp_state); 
		    insts = (jb2SymbolInstance *)NULL; 
		    ninsts = 0; 
		    } 
		else { 
		    /* Straight arithmetic decode */ 
		    DGP_ERR_PASS_CLEANUP(jb2ArithDecodeImage(newsyms[symnum].bits, 
					     newsyms[symnum].width, 
					     newsyms[symnum].height, 
					     newsyms[symnum].longsperline, 
					     abparams, 
					     direct_contexts, file, 
							     mqdec)); 
		    } 
#endif 
		} 
 
	    symnum++; 
	    } 
	     
	if ((!sdrefagg) && sdhuff) { 
	    /* Decode the height class collective bitmap */ 
	 
	    /* Read in the g4 length */ 
	    READV(bmsize); 
	    g4len = v; 
 
	    /* 
	     * Suspend the jbitfile so that we can get at the underlying 
	     * DataSource 
	     */ 
	    DGP_ERR_PASS_CLEANUPF0(jbfSuspend(bf), "can't suspend jbitfile"); 
 
	    /* Allocate the bitmap to decode into */ 
	    classrow_width = (classwidth + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32; 
	    classbits = 
		(Dgp_UInt32 *)(a->calloc)((size_t)oldheight, 
					(size_t)classrow_width * 
					sizeof(*classbits)); 
	    if (classbits == (Dgp_UInt32 *)NULL) { 
	      NOMEM; 
		} 
 
	    /* and decode it */ 
	    if (g4len == 0) { 
		/* 
		 * It's stored as a bitmap. Simple... except that it was 
		 * stored for byte-by-byte access, and we want Dgp_Int32-by-Dgp_Int32... 
		 */ 
		Dgp_Int32 byte_width = (classwidth + DGP_BITS_UINT8 - 1) / DGP_BITS_UINT8; 
		for (i = 0; i < oldheight; i++) { 
		    Dgp_UInt32 *p = classbits + i * classrow_width; 
		    Dgp_UInt8 *p2; 
		    Dgp_UInt32 *q; 
		    Dgp_Int32 ix, ix2; 
 
		    /* Read the bits in */ 
		    DSREAD(data, (size_t)byte_width, (Dgp_UInt8 *)p, &bytes); 
 
		    /* Clear the pad bytes */ 
		    for (ix = byte_width, p2 = (Dgp_Int8 *)p + byte_width; 
			 ix < classrow_width * (int)sizeof(Dgp_UInt32); 
			 ix++) { 
			*p2++ = 0; 
			} 
 
		    /* Shuffle the bytes */ 
		    for (ix = 0, p2 = (Dgp_Int8 *)p, q = p; 
			 ix < classrow_width; 
			 ix++) { 
			tmp = 0; 
			for (ix2 = 0; 
			     ix2 < (int)sizeof(Dgp_UInt32); 
			     ix2++) { 
			    tmp |= (*p2++) << 
				(DGP_BITS_UINT32 - (ix2 + 1) * DGP_BITS_UINT8); 
			    } 
			*q++ = tmp; 
			} 
		    } 
		} 
	    else { 
		/* It's stored as G4-compressed */ 
		DGP_ERR_PASS_CLEANUP(jb2MMRDecodeImage(classbits, classwidth, 
						       oldheight, 
				       classrow_width, 0, classwidth - 1, 
						       oldheight - 1, file, data, g4len)); 
		} 
#if	(VERBLEVEL >= 50) 
	    for (i = 0; i < oldheight; i++) { 
		Dgp_UInt32 *p = classbits + i * classrow_width; 
		Dgp_Int32 j; 
		for (j = 0; j < classwidth; j++) { 
		    if (p[j / DGP_BITS_UINT32] & (1ul << ((DGP_BITS_UINT32 - 1) - (j % DGP_BITS_UINT32)))) { 
			VDEBUG("%d", 1, 50); 
			} 
		    else { 
			VDEBUG("%d", 0, 50); 
			} 
		    } 
		VDEBUG("%c", '\n', 50); 
		} 
#endif 
 
	    /* Go back to normal jbitfile access */ 
	    jbfResume(bf); 
 
	    /* Now we just need to read out all the symbols. */ 
	    bitoffset = 0; 
	    for (i = classstart; i < symnum; i++) { 
		/* No refinement */ 
		newsyms[i].refsym = (jb2Symbol *)NULL; 
		newsyms[i].userdata = (void *)NULL; 
 
		if (symnum == classstart + 1) { 
		    /* There's only one symbol; no need to copy */ 
		    DGP_ASSERT(newsyms[i].longsperline == classrow_width); 
		    DGP_ASSERT(classwidth == newsyms[i].width); 
 
		    newsyms[i].bits = classbits; 
		    classbits = (Dgp_UInt32 *)NULL; 
		    bitoffset = classwidth; 
		    } 
		else { 
		    newsyms[i].bits = 
			(Dgp_UInt32 *)(a->malloc)(oldheight * 
						newsyms[i].longsperline * 
						sizeof(Dgp_UInt32)); 
		    if (newsyms[i].bits == (Dgp_UInt32 *)NULL) { 
		      NOMEM; 
			} 
 
		    jb2ReadoutBits(classbits, bitoffset, classwidth, oldheight, 
				   newsyms[i].bits, newsyms[i].width); 
		    bitoffset += newsyms[i].width; 
		    } 
		} 
 
	    DGP_ASSERT(bitoffset == classwidth); 
 
	    /* Done with this height class */ 
	    if (classbits != (Dgp_UInt32 *)NULL) { 
			(a->free)((void *)classbits); 
			classbits = (Dgp_UInt32 *)NULL; 
		} 
	    } 
	} 
 
    /* All decoded */ 
    if (fs != NULLDIDECODER) { 
	diFreeDecoder(fs); 
	} 
 
    if (ds != NULLDIDECODER) { 
	diFreeDecoder(ds); 
	} 
 
    if (dt != NULLDIDECODER) { 
	diFreeDecoder(dt); 
	} 
 
    if (t != NULLDIDECODER) { 
	diFreeDecoder(t); 
	} 
 
    if (ri != NULLDIDECODER) { 
	diFreeDecoder(ri); 
	} 
 
    if (rdw != NULLDIDECODER) { 
	diFreeDecoder(rdw); 
	} 
 
    if (rdh != NULLDIDECODER) { 
	diFreeDecoder(rdh); 
	} 
 
    if (rdx != NULLDIDECODER) { 
	diFreeDecoder(rdx); 
	} 
 
    if (rdy != NULLDIDECODER) { 
	diFreeDecoder(rdy); 
	} 
 
    if (rsize != NULLDIDECODER) { 
	diFreeDecoder(rsize); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    if (classbits != (Dgp_UInt32 *)NULL) { 
	(a->free)((void *)classbits); 
	} 
 
    if (fs != NULLDIDECODER) { 
	diFreeDecoder(fs); 
	} 
 
    if (ds != NULLDIDECODER) { 
	diFreeDecoder(ds); 
	} 
 
    if (dt != NULLDIDECODER) { 
	diFreeDecoder(dt); 
	} 
 
    if (t != NULLDIDECODER) { 
	diFreeDecoder(t); 
	} 
 
    if (ri != NULLDIDECODER) { 
	diFreeDecoder(ri); 
	} 
 
    if (rdw != NULLDIDECODER) { 
	diFreeDecoder(rdw); 
	} 
 
    if (rdh != NULLDIDECODER) { 
	diFreeDecoder(rdh); 
	} 
 
    if (rdx != NULLDIDECODER) { 
	diFreeDecoder(rdx); 
	} 
 
    if (rdy != NULLDIDECODER) { 
	diFreeDecoder(rdy); 
	} 
 
    if (rsize != NULLDIDECODER) { 
	diFreeDecoder(rsize); 
	} 
 
    if (insts != (jb2SymbolInstance *)NULL) { 
	jb2FreeInstances(insts, ninsts, dgp_state); 
	} 
 
    if (refdata != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)refdata); 
	} 
 
    if (ref != (jb2SymbolInstanceRefined *)NULL) { 
	(a->free)((void *)ref); 
	} 
 
    if (bits != (Dgp_UInt32 *)NULL) { 
	(a->free)((void *)bits); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Read out destwidth bits from bitoffset bits into the image stored in 
 * src, and write them to dest (filling from the MSB as always). Everything 
 * should be height rows high, and the rows should be padded to the nearest 
 * Dgp_Int32. 
 * 
 * Both the source and destination are supposed to be accessed Dgp_Int32-by-Dgp_Int32. 
 */ 
void 
jb2ReadoutBits(Dgp_UInt32 *src, Dgp_Int32 bitoffset, Dgp_Int32 srcwidth, Dgp_Int32 height, 
	       Dgp_UInt32 *dest, Dgp_Int32 destwidth) 
{ 
    Dgp_Int32 i, j; 
    Dgp_Int32 src_row_width = (srcwidth + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32; 
    Dgp_Int32 dest_row_width = (destwidth + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32; 
    Dgp_Int32 Dgp_Int32offset; 
    Dgp_UInt32 *p; 
    Dgp_UInt32 *q; 
    Dgp_UInt32 left; 
    Dgp_Int32 bitsinlastDgp_Int32; 
    Dgp_UInt32 lastmask; 
 
    Dgp_Int32offset = bitoffset / DGP_BITS_UINT32; 
    bitoffset = bitoffset % DGP_BITS_UINT32; 
 
    /* 
     * How many bits are valid in the last Dgp_Int32 of dest? 
     * We want to make sure they're cleared in the output, so we mask them. 
     * Set lastmask to a value with the leftmost bitsinlastDgp_Int32 bits set. 
     */ 
    bitsinlastDgp_Int32 = destwidth % DGP_BITS_UINT32; 
    if (bitsinlastDgp_Int32 == 0) { 
	bitsinlastDgp_Int32 = DGP_BITS_UINT32; 
	lastmask = ~0ul; 
	} 
    else { 
	lastmask = 
	    ((1ul << bitsinlastDgp_Int32) - 1) << (DGP_BITS_UINT32 - bitsinlastDgp_Int32); 
	} 
 
    src += Dgp_Int32offset; 
    for (i = 0; i < height; i++) { 
	/* Find the row bases */ 
	p = src + i * src_row_width; 
	q = dest + i * dest_row_width; 
 
	/* Go through Dgp_Int32-by-Dgp_Int32 */ 
	if (bitoffset != 0) { 
	    left = *p++ << bitoffset; 
	    for (j = 0; j < dest_row_width - 1; j++) { 
		*q++ = left | (*p >> (DGP_BITS_UINT32 - bitoffset)); 
		left = *p++ << bitoffset; 
		} 
	    /* Do we need to touch another source Dgp_Int32? */ 
	    if (bitoffset + bitsinlastDgp_Int32 > DGP_BITS_UINT32) { 
		*q++ = (left | (*p >> (DGP_BITS_UINT32 - bitoffset))) & lastmask; 
		} 
	    else { 
		*q++ = left & lastmask; 
		} 
	    } 
	else { 
	    /* No shifting - yay! */ 
	    for (j = 0; j < dest_row_width; j++) { 
		*q++ = *p++; 
		} 
	    /* Mask off the last Dgp_Int32 properly */ 
	    *(q - 1) &= lastmask; 
	    } 
	} 
    } 
 
/* Duplicate the symbols, including the bitmaps */ 
static ErrCode 
dupSymbols(jb2Symbol *dest, jb2Symbol *src, Dgp_Int32 nsyms, Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("dupSymbols()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Int32 i; 
 
    for (i = 0; i < nsyms; i++) { 
	/* Copy the whole structure */ 
	dest[i] = src[i]; 
 
	dest[i].bits = (Dgp_UInt32 *)(a->malloc)(src[i].height * 
					       src[i].longsperline * 
					       sizeof(Dgp_UInt32)); 
	DGP_CHECK_ALLOC_CLEANUP(dest[i].bits); 
 
	(void)memcpy(dest[i].bits, src[i].bits, 
		     src[i].height * src[i].longsperline * 
		     sizeof(Dgp_UInt32)); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    /* 
     * Everything we've allocated has been hooked into dest already - don't 
     * free anything. 
     */ 
    ERR_RETURN; 
    } 
 
/* Copy the symbol structures, but don't copy the bitmaps */ 
static void 
copySymbols(jb2Symbol *dest, jb2Symbol *src, Dgp_Int32 nsyms) 
{ 
    (void)memcpy((void *)dest, (void *)src, nsyms * sizeof(*dest)); 
    } 
 
