/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/decsymreg.c,v 1.28 2003/01/27 21:48:33 awilliam Exp $ 
 * 
 */ 
 
 
#include "util.h" 
#include "datasource.h" 
#include "segment.h" 
#include "jb2file.h" 
#include "segment_util.h" 
#include "decsymreg.h" 
#include "decgenrefreg.h" 
#include "jbitfile.h" 
#include "decint.h" 
#include "hufftable.h" 
#include "stdhuff.h" 
#include <stdlib.h> 
#include <string.h> 
#include "fileerr.h" 
#include "mqcoder.h" 
#include "symdec.h" 
#include "blit.h" 
#include "image_defs.h" 
#include "PsgFix.h"

#define DGP_ENC DGP_FALSE

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/decsymreg.c,v 1.28 2003/01/27 21:48:33 awilliam Exp $"); 
 
static ErrCode decodeIDHuffTable(jb2File file, jbitfile bf, DgpEncode_MQDecoder *dec, 
				    Dgp_Bool sbhuff, 
				    jb2Segment *segment, jb2Segment **refsegs, 
				 symDecoder *symdec); 
 
#ifdef	VALPRINT 
#include <stdio.h> 
 
#define	READV(coder) \
    do { \
	DGP_ERR_PASS_CLEANUP(diDecode(coder, &oob, &v)); \
	if (oob) { \
	    fprintf(stderr, #coder " OOB\n"); \
	    } \
	else { \
	    fprintf(stderr, #coder " %d\n", v); \
	    } \
	} while(0) 
 
#else 
 
#define	READV(coder) \
    do { \
	DGP_ERR_PASS_CLEANUP(diDecode(coder, &oob, &v)); \
	} while(0) 
#endif 
 
ErrCode 
jb2DecodeSymReg(jb2File file, DataSource data, 
		jb2Segment *segment, jb2Segment **refsegs) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2DecodeSymReg()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_UInt8 lbuf[4]; 
    jb2SymbolInstance *instances = (jb2SymbolInstance *)NULL; 
    Dgp_Bool sbhuff, sbrefine, transposed; 
    Dgp_Int32 logsbstrips, sbstrips, refcorner, sbcombop, sbdefpixel, sbdsoffset; 
    hufftable_spec *sbhufffs, *sbhuffds, *sbhuffdt, *sbhuffrdw, *sbhuffrdh, 
	*sbhuffrdx, *sbhuffrdy, *sbhuffrsize; 
    diDecoder fs = NULLDIDECODER; 
    diDecoder ds = NULLDIDECODER; 
    diDecoder dt = NULLDIDECODER; 
    diDecoder t = NULLDIDECODER; 
    diDecoder ri = NULLDIDECODER; 
    diDecoder rdx = NULLDIDECODER; 
    diDecoder rdy = NULLDIDECODER; 
    diDecoder rdw = NULLDIDECODER; 
    diDecoder rdh = NULLDIDECODER; 
    diDecoder rsize = NULLDIDECODER; 
    jbitfile bf = NULLJBITFILE, tmpbf; 
    DgpEncode_MQDecoder mqdec; 
    symDecoder sbsymid = (symDecoder)NULL; 
    Dgp_Int32 nusertables = 0; 
    hufftable_spec **usertables[8]; 
    Dgp_UInt16 flags; 
    Dgp_Int32 sbnuminstances; 
    Dgp_Int32 sbrtemplate; 
    arith_ref_params refparams; 
    Dgp_UInt8 *refcontexts = (Dgp_UInt8 *)NULL; 
    Dgp_Int32 dummy = 0; 
    /* Context buffers for the integer decoders */ 
    Dgp_UInt8 fscontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 dscontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 dtcontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 tcontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 ricontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 rdxcontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 rdycontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 rdwcontexts[DICONTEXTSIZE]; 
    Dgp_UInt8 rdhcontexts[DICONTEXTSIZE]; 
    size_t bytes; 
 
    /* Decode the region segment header */ 
    DGP_ERR_PASS_CLEANUP(jb2DecodeRegSegment(file, data, &segment->reginfo)); 
 
    /* Decode the symbol region segment flags */ 
    DSREAD(data, 2, lbuf, &bytes); 
    flags = (lbuf[0] << 8) | lbuf[1]; 
#ifdef	VALPRINT 
    fprintf(stderr, "SYMREGFLAGS %04x\n", flags); 
#endif 
     
    /* SBHUFF */ 
    sbhuff = (flags & 1) ? DGP_TRUE : DGP_FALSE; 
 
    /* SBREFINE */ 
    sbrefine = (flags & 2) ? DGP_TRUE : DGP_FALSE; 
 
    /* SBSTRIPS */ 
    logsbstrips = ((flags >> 2) & 3); 
    sbstrips = 1 << logsbstrips; 
 
    /* REFCORNER */ 
    refcorner = (flags >> 4) & 3; 
 
    /* TRANSPOSED */ 
    transposed = (flags & 64) ? DGP_TRUE : DGP_FALSE; 
 
    /* SBCOMBOP */ 
    sbcombop = (flags >> 7) & 3; 
 
    /* SBDEFPIXEL */ 
    sbdefpixel = (flags >> 9) & 1; 
 
    /* SBDSOFFSET */ 
    sbdsoffset = (flags >> 10) & 31; 
    /* slam it around so that it sign extends */ 
    if (sbdsoffset & 16) { 
	/* should be negative - fill in lots of leading 1s */ 
	sbdsoffset = (~31) | sbdsoffset; 
	} 
 
    /* SBRTEMPLATE */ 
    sbrtemplate = (flags >> 15) & 1; 
    if (!sbrefine) { 
	if (sbrtemplate != 0) { 
	    FILEERR(cErrCompressionError, "bad template flag for non-refining symbol region"); 
	    } 
	} 
 
    /* Huffman table flags */ 
    if (sbhuff) { 
	DSREAD(data, 2, lbuf, &bytes); 
	flags = (lbuf[0] << 8) | lbuf[1]; 
#ifdef	VALPRINT 
	fprintf(stderr, "SYMREGHUFFFLAGS %04x\n", flags); 
#endif 
 
	/* SBHUFFFS selection */ 
	switch (flags & 3) { 
	  case 0: { 
	      sbhufffs = &stdhuffFS1; 
	      break; 
	      } 
 
	  case 1: { 
	      sbhufffs = &stdhuffFS2; 
	      break; 
	      } 
 
	  case 3: { 
	      sbhufffs = (hufftable_spec *)NULL; 
	      usertables[nusertables++] = &sbhufffs; 
	      break; 
	      } 
       
	  default: { 
	      FILEERR(cErrCompressionError, "bad SBHUFFFS specification"); 
	      } 
	    } 
 
	/* SBHUFFDS selection */ 
	switch ((flags >> 2) & 3) { 
	  case 0: { 
	      sbhuffds = &stdhuffDS1; 
	      break; 
	      } 
 
	  case 1: { 
	      sbhuffds = &stdhuffDS2; 
	      break; 
	      } 
 
	  case 2: { 
	      sbhuffds = &stdhuffDS3; 
	      break; 
	      } 
 
	  case 3: { 
	      sbhuffds = (hufftable_spec *)NULL; 
	      usertables[nusertables++] = &sbhuffds; 
	      break; 
	      } 
	    } 
 
	/* SBHUFFDT selection */ 
	switch ((flags >> 4) & 3) { 
	  case 0: { 
	      sbhuffdt = &stdhuffDT1; 
	      break; 
	      } 
 
	  case 1: { 
	      sbhuffdt = &stdhuffDT2; 
	      break; 
	      } 
 
	  case 2: { 
	      sbhuffdt = &stdhuffDT3; 
	      break; 
	      } 
 
	  case 3: { 
	      sbhuffdt = (hufftable_spec *)NULL; 
	      usertables[nusertables++] = &sbhuffdt; 
	      break; 
	      } 
	    } 
 
	/* SBHUFFRDW selection */ 
	switch ((flags >> 6) & 3) { 
	  case 0: { 
	      sbhuffrdw = &stdhuffRD1; 
	      break; 
	      } 
 
	  case 1: { 
	      sbhuffrdw = &stdhuffRD2; 
	      break; 
	      } 
 
	  case 3: { 
	      sbhuffrdw = (hufftable_spec *)NULL; 
	      usertables[nusertables++] = &sbhuffrdw; 
	      break; 
	      } 
 
	  default: { 
	      FILEERR(cErrCompressionError, "bad SBHUFFRDW specification"); 
	      } 
	    } 
 
	/* SBHUFFRDH selection */ 
	switch ((flags >> 8) & 3) { 
	  case 0: { 
	      sbhuffrdh = &stdhuffRD1; 
	      break; 
	      } 
 
	  case 1: { 
	      sbhuffrdh = &stdhuffRD2; 
	      break; 
	      } 
 
	  case 3: { 
	      sbhuffrdh = (hufftable_spec *)NULL; 
	      usertables[nusertables++] = &sbhuffrdh; 
	      break; 
	      } 
 
	  default: { 
	      FILEERR(cErrCompressionError, "bad SBHUFFRDH specification"); 
	      } 
	    } 
 
	/* SBHUFFRDX selection */ 
	switch ((flags >> 10) & 3) { 
	  case 0: { 
	      sbhuffrdx = &stdhuffRD1; 
	      break; 
	      } 
 
	  case 1: { 
	      sbhuffrdx = &stdhuffRD2; 
	      break; 
	      } 
 
	  case 3: { 
	      sbhuffrdx = (hufftable_spec *)NULL; 
	      usertables[nusertables++] = &sbhuffrdx; 
	      break; 
	      } 
 
	  default: { 
	      FILEERR(cErrCompressionError, "bad SBHUFFRDX specification"); 
	      } 
	    } 
 
	/* SBHUFFRDY selection */ 
	switch ((flags >> 12) & 3) { 
	  case 0: { 
	      sbhuffrdy = &stdhuffRD1; 
	      break; 
	      } 
 
	  case 1: { 
	      sbhuffrdy = &stdhuffRD2; 
	      break; 
	      } 
 
	  case 3: { 
	      sbhuffrdy = (hufftable_spec *)NULL; 
	      usertables[nusertables++] = &sbhuffrdy; 
	      break; 
	      } 
 
	  default: { 
	      FILEERR(cErrCompressionError, "bad SBHUFFRDY specification"); 
	      } 
	    } 
 
	/* SBHUFFRSIZE selection */ 
	if (flags & (1 << 14)) { 
	    sbhuffrsize = (hufftable_spec *)NULL; 
	    usertables[nusertables++] = &sbhuffrsize; 
	    } 
	else { 
	    sbhuffrsize = &stdhuffA; 
	    } 
 
	/* Find the user Huffman tables */ 
	DGP_ERR_PASS_CLEANUPF0(findUserTables(usertables, nusertables, refsegs, 
					      segment->nrefsegs, dgp_state), 
	    "wrong number of Huffman table segments referenced"); 
     
	/* Check that OOB is correctly available/not available */ 
	if ((!sbhuffds->oob_valid) || sbhufffs->oob_valid || 
	    sbhuffdt->oob_valid || sbhuffrdw->oob_valid || 
	    sbhuffrdh->oob_valid || 
	    sbhuffrdx->oob_valid || sbhuffrdy->oob_valid || 
	    sbhuffrsize->oob_valid) { 
	    FILEERR(cErrCompressionError, "wrong OOB capabilities"); 
	    } 
	} 
    else { 
	/* 
	 * Put in some dummy values so we don't pass uninitialised 
	 * data to diNewDecoderEither.  We can use NULL, as findUserTables 
	 * has already been called. 
	 */ 
	sbhufffs = (hufftable_spec *)NULL; 
	sbhuffds = (hufftable_spec *)NULL; 
	sbhuffdt = (hufftable_spec *)NULL; 
	sbhuffrdw = sbhuffrdh = sbhuffrdx = sbhuffrdy = (hufftable_spec *)NULL; 
	sbhuffrsize = (hufftable_spec *)NULL; 
	}	 
     
    /* Decode the refinement parameters */ 
    if (sbrefine) { 
	refparams.rtemplate = (Dgp_UInt8)sbrtemplate; 
	refparams.tpron = DGP_FALSE; 
 
	DGP_ERR_PASS_CLEANUP(jb2DecodeRABParams(file, data, &refparams, 
						&dummy)); 
 
	/* Allocate the contexts needed */ 
	refcontexts = (Dgp_UInt8 *)(a->calloc)(jb2RABContextsNeeded(&refparams), 
					      1); 
	} 
 
    /* SBNUMINSTANCES */ 
    DSREAD(data, 4, lbuf, &bytes); 
    sbnuminstances = 
	(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
#ifdef	VALPRINT 
    fprintf(stderr, "SBNUMINSTANCES %d\n", sbnuminstances); 
#endif 
 
    /* Create the instances */ 
    instances = 
	(jb2SymbolInstance *)(a->malloc)(sbnuminstances * sizeof(*instances)); 
    DGP_CHECK_ALLOC_CLEANUP(instances); 
 
    if (sbhuff) { 
	/* Create a bitfile for the remaining decoding */ 
      DGP_ERR_PASS_CLEANUP(jbfOpenFromDataSource(data, &bf)); 
	} 
    else { 
	/* Create an DgpEncode_MQDecoder for the remaining decoding */ 
#ifndef	NOMQ 
	dsInitDgpEncode_MQDecoder(data, &mqdec); 
#else 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
	} 
 
    /* 
     * Oh boy - now we get to decode the symbol ID Huffman table. 
     * If sbhuff is DGP_FALSE, there isn't one, but calling it will do the right 
     * thing anyway. 
     */ 
    DGP_ERR_PASS_CLEANUP(decodeIDHuffTable(file, bf, &mqdec, sbhuff, segment, 
					   refsegs, &sbsymid)); 
 
    /* Create the various integer decoders */ 
    DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sbhuff, sbhufffs, bf, &mqdec, 
					    fscontexts, dgp_state, &fs)); 
 
    DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sbhuff, sbhuffds, bf, &mqdec, 
					    dscontexts, dgp_state, &ds)); 
 
    DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sbhuff, sbhuffdt, bf, &mqdec, 
					    dtcontexts, dgp_state, &dt)); 
 
    if (sbrefine) { 
#ifndef	NOMQ 
	DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sbhuff, sbhuffrdw, bf, &mqdec, 
						rdwcontexts, dgp_state, &rdw)); 
 
	DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sbhuff, sbhuffrdh, bf, &mqdec, 
						rdhcontexts, dgp_state, &rdh)); 
 
	DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sbhuff, sbhuffrdx, bf, &mqdec, 
						rdxcontexts, dgp_state, &rdx)); 
 
	DGP_ERR_PASS_CLEANUP(diNewDecoderEither(sbhuff, sbhuffrdy, bf, &mqdec, 
						rdycontexts, dgp_state, &rdy)); 
 
	if (sbhuff) { 
	    /* Only Huffman needs RSIZE */ 
	    DGP_ERR_PASS_CLEANUP(diNewDecoder(sbhuffrsize, bf, dgp_state, 
					      &rsize)); 
 
	    /* and there's no RI decoder */ 
	    ri = NULLDIDECODER; 
	    } 
	else { 
	    /* Only arithmetic needs RI */ 
	    memset(ricontexts, 0, DICONTEXTSIZE); 
	    DGP_ERR_PASS_CLEANUP(diNewDecoderMQ(&mqdec, ricontexts, dgp_state, 
						&ri)); 
 
	    rsize = NULLDIDECODER; 
	    } 
#else 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
	} 
 
    if (sbhuff) { 
	/* T is special in Huffman case */ 
	t = NULLDIDECODER; 
	} 
    else { 
	/* Arithmetic needs T */ 
#ifndef	NOMQ 
	memset(tcontexts, 0, DICONTEXTSIZE); 
	DGP_ERR_PASS_CLEANUP(diNewDecoderMQ(&mqdec, tcontexts, dgp_state, &t)); 
#endif 
	} 
 
    /* Now decode the actual symbol region data */ 
    DGP_ERR_PASS_CLEANUP(jb2DecodeInstances(file, 
											bf, 
											sbhuff, 
											sbrefine, 
											transposed, 
											(Dgp_UInt8)refcorner, 
											logsbstrips, 
											sbdsoffset, 
											sbsymid, 
											fs, 
											ds, 
											dt, 
											t, 
											ri, 
											rdw, 
											rdh, 
											rdx, 
											rdy,
											rsize, 
											data, 
											&mqdec, 
											&refparams, 
											refcontexts, 
											sbnuminstances, 
											instances)); 
			  
    /* Fill in the fields in *segment */ 
    segment->data.symreg.ninstances = sbnuminstances; 
    segment->data.symreg.instances = instances; 
    segment->data.symreg.int_combop = (Dgp_UInt8)sbcombop; 
    segment->data.symreg.defpixel = (Dgp_UInt8)sbdefpixel; 
 
    /* Free up stuff */ 
    if (bf != NULLJBITFILE) { 
      tmpbf = bf; 
      bf = NULLJBITFILE; 
	DGP_ERR_PASS_CLEANUP(jbfCloseFromDataSource(tmpbf)); 
	} 
 
    /* Update the DataSource's bufptr from the DgpEncode_MQDecoder */ 
#ifndef	NOMQ 
    if (!sbhuff) { 
      DGP_ERR_PASS_CLEANUPF0(dsFinalDgpEncode_MQDecoder(data, &mqdec), 
	    "problem syncing MQ decoder"); 
	} 
#endif 
 
    /* Free the world */ 
    symDecoderFree(sbsymid); 
    diFreeDecoder(fs); 
    diFreeDecoder(ds); 
    diFreeDecoder(dt); 
    if (t != NULLDIDECODER) { 
	diFreeDecoder(t); 
	} 
    if (ri != NULLDIDECODER) { 
	diFreeDecoder(ri); 
	} 
    if (rdw != NULLDIDECODER) { 
	diFreeDecoder(rdw); 
	} 
    if (rdh != NULLDIDECODER) { 
	diFreeDecoder(rdh); 
	} 
    if (rdx != NULLDIDECODER) { 
	diFreeDecoder(rdx); 
	} 
    if (rdy != NULLDIDECODER) { 
	diFreeDecoder(rdy); 
	} 
    if (rsize != NULLDIDECODER) { 
	diFreeDecoder(rsize); 
	} 
    if (refcontexts != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)refcontexts); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    if (instances != (jb2SymbolInstance *)NULL) { 
	(a->free)((void *)instances); 
	} 
 
    if (bf != NULLJBITFILE) { 
	(void)jbfCloseFromDataSource(bf); 
	} 
 
    if (sbsymid != (symDecoder)NULL) { 
	symDecoderFree(sbsymid); 
	} 
 
    if (fs != NULLDIDECODER) { 
	diFreeDecoder(fs); 
	} 
    if (ds != NULLDIDECODER) { 
	diFreeDecoder(ds); 
	} 
    if (dt != NULLDIDECODER) { 
	diFreeDecoder(dt); 
	} 
    if (t != NULLDIDECODER) { 
	diFreeDecoder(t); 
	} 
    if (ri != NULLDIDECODER) { 
	diFreeDecoder(ri); 
	} 
    if (rdw != NULLDIDECODER) { 
	diFreeDecoder(rdw); 
	} 
    if (rdh != NULLDIDECODER) { 
	diFreeDecoder(rdh); 
	} 
    if (rdx != NULLDIDECODER) { 
	diFreeDecoder(rdx); 
	} 
    if (rdy != NULLDIDECODER) { 
	diFreeDecoder(rdy); 
	} 
    if (rsize != NULLDIDECODER) { 
	diFreeDecoder(rsize); 
	} 
    if (refcontexts != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)refcontexts); 
	} 
 
    ERR_RETURN; 
    } 
 
static ErrCode 
decodeIDHuffTable(jb2File file, jbitfile bf, DgpEncode_MQDecoder *dec, Dgp_Bool sbhuff, 
		  jb2Segment *segment, jb2Segment **refsegs, 
		  symDecoder *_symdec) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("decodeIDHuffTable()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    hufftable_spec runcodes; 
    table_line runcode_lines[35]; 
    Dgp_Int32 runcode_values[35]; 
    void *runcode_ptrs[35]; 
    void **symptrs = (void **)NULL; 
    Dgp_Int32 *lengths = (Dgp_Int32 *)NULL; 
    Dgp_Int32 i, j, k; 
    void *vp; 
    Dgp_Int32 v; 
    Dgp_Int32 lastcode; 
    Dgp_Int32 sbnumsyms; 
    symDecoder symdec = (symDecoder)NULL; 
    Dgp_UInt32 tmpv; 
 
    *_symdec = (symDecoder)NULL; 
 
    /* First off, how many symbols do we have access to */ 
    sbnumsyms = 0; 
    for (i = 0; i < segment->nrefsegs; i++) { 
	/* Is it a dictionary segment */ 
	if (refsegs[i]->segtype == JB2SEGTYPE_SYMDICT) { 
	    /* Yes - get the count */ 
	    sbnumsyms += refsegs[i]->data.symdict.nsymbols; 
	    } 
	else { 
	    /* No - better be a table */ 
	    if (refsegs[i]->segtype != JB2SEGTYPE_TABLE) { 
		FILEERR(cErrCompressionError, "bad referenced segment type"); 
		} 
	    } 
	} 
 
    /* Allocate an array of void *s to point to the symbols */ 
    symptrs = (void **)(a->malloc)(sbnumsyms * sizeof(*symptrs)); 
    if (symptrs == (void **)NULL) { 
	    NOMEM; 
	} 
 
    /* Now set the pointers */ 
    j = 0; 
    for (i = 0; i < segment->nrefsegs; i++) { 
	if (refsegs[i]->segtype == JB2SEGTYPE_SYMDICT) { 
	    for (k = 0; k < refsegs[i]->data.symdict.nsymbols; k++, j++) { 
		symptrs[j] = &refsegs[i]->data.symdict.symbols[k]; 
		} 
	    } 
	} 
 
    /* If we're not Huffman-based, we have all the info we need */ 
    if (!sbhuff) { 
#ifndef	NOMQ 
      DGP_ERR_PASS_CLEANUP(symDecoderArith(symptrs, sbnumsyms, dec, dgp_state, 
					   &symdec)); 
 
	/* That copied symptrs */ 
	(a->free)((void *)symptrs); 
	symptrs = (void **)NULL; 
 
	*_symdec = symdec; 
	symdec = (symDecoder)NULL; 
	NORMAL_RETURN; 
#endif 
	} 
 
    /* The rest of the routine is the Huffman-based case */ 
 
    /* 
     * Read 35 lengths, each 4 bits, for the 35 run codes.  The run code 
     * meanings are: 
     * Code 0...31: The code length is 0...31 
     * Code 32: Repeat the previous code length 3-6 times 
     *		(next two bits plus 3) 
     * Code 33: Repeat a code length of 0 for 3-10 times 
     *		(next three bits plus 3) 
     * Code 34: Repeat a code length of 0 for 11-138 times 
     *		(next 7 bits plus 11) 
     */ 
    for (i = 0; i < 35; i++) { 
	DGP_ERR_PASS_CLEANUP(jbfReadBits(bf, 4, &(tmpv))); 
	runcode_lines[i].preflen = (Dgp_Int8)tmpv; 
 
	runcode_lines[i].base = runcode_lines[i].rangelen = 0; 
	} 
 
    /* Set up the other values in the table spec */ 
    runcodes.lines = runcode_lines; 
    runcodes.lower_valid = runcodes.upper_valid = runcodes.oob_valid = DGP_FALSE; 
    runcodes.nlines = 35; 
 
    /* Assign codes for the 35 run codes */ 
    DGP_ERR_PASS_CLEANUPF0(jhtSetCodes(&runcodes, dgp_state),  
		"bad RUNCODE Huffman table"); 
 
    /* and compile the mess; set the void * for each line to point to an Dgp_Int32 */ 
    for (i = 0; i < 35; i++) { 
	runcode_values[i] = i; 
	runcode_ptrs[i] = (void *)&runcode_values[i]; 
	} 
 
    DGP_ERR_PASS_CLEANUP(jhtCompile(&runcodes, runcode_ptrs, dgp_state)); 
 
    /* 
     * Now read out the symbol ID code lengths themselves. 
     * We need somewhere to stick them; allocate an array of lengths 
     */ 
    lengths = (Dgp_Int32 *)(a->malloc)(sbnumsyms * sizeof(*lengths)); 
    if (lengths == (Dgp_Int32 *)NULL) { 
      NOMEM; 
	} 
 
    /* Start reading out the code lengths */ 
    lastcode = 0; 
    for (i = 0; i < sbnumsyms; ) { 
	DGP_ERR_PASS_CLEANUP(jhtReadCode(&runcodes, bf, dgp_state, &vp)); 
 
	/* What run code did we see? */ 
	v = *(Dgp_Int32 *)vp; 
	if (v < 32) { 
	    /* Straight code */ 
#ifdef	VALPRINT 
	    fprintf(stderr, "RC %d\n", v); 
#endif 
	    lastcode = lengths[i] = v; 
	    i++; 
	    } 
	else if (v == 32) { 
	    /* Repeat the previous code length 3-6 times */ 
	  DGP_ERR_PASS_CLEANUP(jbfReadBits(bf, 2, &tmpv)); 
	  v = tmpv + 3; 
		 
#ifdef	VALPRINT 
	    fprintf(stderr, "RC %d %d\n", 32, v - 3); 
#endif 
 
	    if (i + v > sbnumsyms) { 
		FILEERR(cErrCompressionError, "too many runcodes"); 
		} 
 
	    for (; v > 0; v--) { 
		lengths[i] = lastcode; 
		i++; 
		} 
	    } 
	else if (v == 33) { 
	    /* Repeat a code length of 0 for 3-10 times */ 
	    DGP_ERR_PASS_CLEANUP(jbfReadBits(bf, 3, &tmpv)); 
	    v = tmpv + 3; 
		 
#ifdef	VALPRINT 
	    fprintf(stderr, "RC %d %d\n", 33, v - 3); 
#endif 
 
	    if (i + v > sbnumsyms) { 
		FILEERR(cErrCompressionError, "too many runcodes"); 
		} 
 
	    for (; v > 0; v--) { 
		lengths[i] = 0; 
		i++; 
		} 
	    lastcode = 0; 
	    } 
	else if (v == 34) { 
	    /* Repeat a code length of 0 for 11-138 times */ 
	    DGP_ERR_PASS_CLEANUP(jbfReadBits(bf, 7, &tmpv)); 
	    v = tmpv + 11; 
		 
#ifdef	VALPRINT 
	    fprintf(stderr, "RC %d %d\n", 34, v - 11); 
#endif 
 
	    if (i + v > sbnumsyms) { 
		FILEERR(cErrCompressionError, "too many runcodes"); 
		} 
 
	    for (; v > 0; v--) { 
		lengths[i] = 0; 
		i++; 
		} 
	    lastcode = 0; 
	    } 
	else { 
	    DGP_REPORTF0(DGP_ENC, "bad runcode"); 
	    } 
	} 
 
    /* Don't need the runcodes Huffman table anymore */ 
    jhtCleanup(&runcodes, dgp_state); 
 
    /* We now have the symbol ID lengths - set up the symbol ID decoder */ 
    DGP_ERR_PASS_CLEANUP(symDecoderHuffman(symptrs, sbnumsyms, bf, lengths,  
					   dgp_state, &symdec)); 
 
    /* Done - free up symptrs and sync bf to the next byte. */ 
    jbfFlushByte(bf); 
 
    (a->free)((void *)symptrs); 
    (a->free)((void *)lengths); 
 
    *_symdec = symdec; 
    symdec = (symDecoder)NULL; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (symptrs != (void **)NULL) { 
	(a->free)((void *)symptrs); 
	} 
 
    if (lengths != (Dgp_Int32 *)NULL) { 
	(a->free)((void *)lengths); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Decode all the instances; use the nice integer decoders. 
 * The r* decoders are used only if sbrefine is DGP_TRUE; rsize is used only 
 * if sbrefine is DGP_TRUE and sbhuff is DGP_TRUE.  t is used only if sbhuff is DGP_FALSE. 
 * bf is used only if sbhuff is DGP_TRUE. 
 */ 
ErrCode 
jb2DecodeInstances(jb2File file, jbitfile bf, 
		   Dgp_Bool sbhuff, Dgp_Bool sbrefine, 
		   Dgp_Bool transposed, Dgp_UInt8 refcorner, 
		   Dgp_Int32 logsbstrips, Dgp_Int32 sbdsoffset, 
		   symDecoder symid, diDecoder fs, diDecoder ds, 
		   diDecoder dt, diDecoder t, diDecoder ri, 
		   diDecoder rdw, diDecoder rdh, 
		   diDecoder rdx, diDecoder rdy, 
		   diDecoder rsize, DataSource data, 
		   DgpEncode_MQDecoder *mqdec, arith_ref_params *refparams, 
		   Dgp_UInt8 *refcontexts, 
		   Dgp_Int32 sbnuminstances, 
		   jb2SymbolInstance *instances) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2DecodeInstances()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Int32 stript, firsts, curs, curt, curh, curw; 
    Dgp_Int32 curri = 0;		/* Current instance refined? */ 
    Dgp_Int32 currdw, currdh, currdx, currdy; 
    Dgp_Int32 ninst; 
    Dgp_Bool oob; 
    Dgp_Int32 v; 
    jb2Symbol *sym; 
    Dgp_UInt32 *bits = (Dgp_UInt32 *)NULL; 
    Dgp_Int32 reflpl; 
    jb2SymbolInstanceRefined *ref = (jb2SymbolInstanceRefined *)NULL; 
    Dgp_Bool firstsym; 
    DgpEncode_MQDecoder huffmq;		/* To be used when refining inside Huffman */ 
    Dgp_UInt8 *refdata = (Dgp_UInt8 *)NULL; 
    size_t refdatalen; 
    size_t bytes; 
    Dgp_UInt32 tmpv; 
 
    /* Don't need r* coders yet */ 
 
    /* Start decoding */ 
 
    /* Read the initial stript value */ 
    READV(dt); 
    if (oob) { 
	FILEERR(cErrCompressionError, "bad OOB"); 
	} 
 
    stript = -(v << logsbstrips); 
    curs = firsts = 0; 
    ninst = 0; 
 
    /* Decode strips */ 
    while (ninst < sbnuminstances) { 
	/* Delta t value */ 
	READV(dt); 
	if (oob) { 
	    FILEERR(cErrCompressionError, "bad OOB"); 
	    } 
 
	stript += v << logsbstrips; 
	firstsym = DGP_TRUE; 
 
	for (;;) { 
	    if (firstsym) { 
		/* First instance S coordinate */ 
		READV(fs); 
		if (oob) { 
		    FILEERR(cErrCompressionError, "bad OOB"); 
		    } 
 
		firsts += v; 
		curs = firsts; 
		firstsym = DGP_FALSE; 
		} 
	    else { 
		READV(ds); 
		if (oob) { 
		    /* End of strip */ 
		    break; 
		    } 
		if (ninst == sbnuminstances) { 
		    FILEERR(cErrCompressionError, "too many symbol instances"); 
		    } 
 
		curs += v + sbdsoffset; 
		} 
 
	    /* Read T */ 
	    if (logsbstrips > 0) { 
		if (sbhuff) { 
		    DGP_ERR_PASS_CLEANUP(jbfReadBits(bf, logsbstrips, &tmpv)); 
		    curt = tmpv; 
		    } 
		else { 
		    DGP_ERR_PASS_CLEANUP(diDecode(t, &oob, &curt)); 
		    if (oob) { 
			FILEERR(cErrCompressionError, "bad OOB"); 
			} 
		    } 
		} 
	    else { 
		curt = 0; 
		} 
#ifdef	VALPRINT 
	    if ((!sbhuff) || (logsbstrips > 0)) { 
		/* Only print if we actually read bits */ 
		fprintf(stderr, "t %d\n", curt); 
		} 
#endif 
 
	    curt += stript; 
 
	    /* What's the symbol used here? */ 
	    DGP_ERR_PASS_CLEANUP(symDecode(symid, &sym)); 
 
	    /* 
	     * It's possible we've been given a bogus symbol. 
	     * This can happen if we've been called from within a 
	     * symbol dictionary, and the symbol ID refers to a symbol 
	     * that hasn't been decoded yet (that's possible, since the 
	     * code space may not be full).  We need to do a quick sanity 
	     * check. 
	     */ 
	    if (sym->bits == (Dgp_UInt32 *)NULL) { 
		FILEERR(cErrCompressionError, "bad symbol ID"); 
		} 
 
	    /* Look for refinement */ 
	    if (sbrefine) { 
		/* Does this instance contain refinement? */ 
		if (sbhuff) { 
		    /* Read one bit */ 
		    DGP_ERR_PASS_CLEANUP(jbfReadBits(bf, 1, &tmpv)); 
		    curri = tmpv; 
#ifdef	VALPRINT 
		    fprintf(stderr, "ri %d\n", curri); 
#endif 
		    } 
		else { 
		    /* Use the RI decoder */ 
		    READV(ri); 
		    if ((oob) || ((v != 0) && (v != 1))) { 
			FILEERR(cErrCompressionError, "bad RI value"); 
			} 
		    curri = v; 
		    } 
		} 
 
	    if (curri) { 
		curw = sym->width; 
		curh = sym->height; 
 
		/* Delta width */ 
		READV(rdw); 
		if (oob) { 
		    FILEERR(cErrCompressionError, "bad OOB"); 
		    } 
		currdw = v; 
		curw += currdw; 
 
		/* Delta height */ 
		READV(rdh); 
		if (oob) { 
		    FILEERR(cErrCompressionError, "bad OOB"); 
		    } 
		currdh = v; 
		curh += currdh; 
 
		/* Delta X */ 
		READV(rdx); 
		if (oob) { 
		    FILEERR(cErrCompressionError, "bad OOB"); 
		    } 
		currdx = v; 
 
		/* Delta Y */ 
		READV(rdy); 
		if (oob) { 
		    FILEERR(cErrCompressionError, "bad OOB"); 
		    } 
		currdy = v; 
 
		/* Allocate the bitmap to refine into */ 
		reflpl = (curw + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32; 
		bits = (Dgp_UInt32 *)(a->calloc)(reflpl * sizeof(Dgp_UInt32), 
					       curh); 
		if (bits == (Dgp_UInt32 *)NULL) { 
		  NOMEM; 
		    } 
 
		/* Invoke the refinement */ 
#ifndef	NOMQ 
		if (sbhuff) { 
		    /* EEP - we need to switch to arithmetic coding! */ 
		    READV(rsize); 
		    if (oob) { 
			FILEERR(cErrCompressionError, "bad OOB"); 
			} 
		    refdatalen = v; 
 
		    /* 
		     * rsize tells us how many bytes we need. 
		     * Copy that much out of the DataSource. 
		     */ 
		    /* Suspend the bitfile first */ 
		    DGP_ERR_PASS_CLEANUP(jbfSuspend(bf)); 
		    /* then allocate and copy the bytes */ 
		    refdata = (Dgp_UInt8 *)(a->malloc)(refdatalen); 
		    if (refdata == (Dgp_UInt8 *)NULL) { 
		      NOMEM; 
			} 
 
		    DSREAD(data, refdatalen, refdata, &bytes); 
 
		    /* Now turn the data into an MQ coder */ 
		    DgpEncode_mqdInit(&huffmq, refdata, refdatalen, 
			    (Dgp_Bool (*)(Dgp_UInt8 **, size_t *, void *))NULL, 
			    (void *)NULL, dgp_state); 
		    mqdec = &huffmq; 
		    } 
 
		/* 
		 * Do the decoding.  Use refcontexts. 
		 * The spec says that refdx should be 
		 * floor(currdw / 2) + currdx, which (remembering that 
		 * currdw could be negative) is (currdw >> 1) + currdx. 
		 * Similarly for currdh. 
		 */ 
		DGP_ERR_PASS_CLEANUP(jb2ArithDecodeRefinement(bits, curw, curh, reflpl, 
					      sym->bits, 
					      sym->width, sym->height, 
					      sym->longsperline, 
					      (currdw >> 1) + currdx, 
					      (currdh >> 1) + currdy, 
					      refparams, refcontexts, 
							      file, mqdec)); 
		 
#if	VERBLEVEL >= 100 
		{ 
		    Dgp_Int32 xx, yy; 
		    for (yy = 0; yy < sym->height; yy++) { 
			for (xx = 0; xx < sym->width; xx++) { 
			    if (sym->bits[yy * sym->longsperline + (xx / DGP_BITS_UINT32)] & 
				(1ul << (DGP_BITS_UINT32 - 1 - (xx & (DGP_BITS_UINT32 - 1))))) { 
				fputc('1', stderr); 
				} 
			    else { 
				fputc('0', stderr); 
				} 
			    } 
			fputc('\n', stderr); 
			} 
		    fprintf(stderr, "   ---> @(%d, %d)\n", curs, curt); 
		    for (yy = 0; yy < curh; yy++) { 
			for (xx = 0; xx < curw; xx++) { 
			    if (bits[yy * reflpl + (xx / DGP_BITS_UINT32)] & 
				(1ul << (DGP_BITS_UINT32 - 1 - (xx & (DGP_BITS_UINT32 - 1))))) { 
				fputc('1', stderr); 
				} 
			    else { 
				fputc('0', stderr); 
				} 
			    } 
			fputc('\n', stderr); 
			} 
		    }				 
#endif 
		 
		/* Hook the decoded info into the symbol instance */ 
		ref = (jb2SymbolInstanceRefined *)(a->malloc)(sizeof(*ref)); 
		if (ref == (jb2SymbolInstanceRefined *)NULL) { 
		  NOMEM; 
		    } 
 
		ref->bits = bits; 
		ref->longsperline = reflpl; 
		ref->width = curw; 
		ref->height = curh; 
 
		instances[ninst].refdata = ref; 
 
		/* Hooked in - clear redundant pointers */ 
		ref = (jb2SymbolInstanceRefined *)NULL; 
		bits = (Dgp_UInt32 *)NULL; 
 
		if (sbhuff) { 
		    /* Switch back to Huffman coding */ 
		    (a->free)((void *)refdata); 
		    refdata = (Dgp_UInt8 *)NULL; 
 
		    jbfResume(bf); 
		    } 
#endif 
		} 
	    else { 
		/* No refinement - just the symbol */ 
		curh = sym->height; 
		curw = sym->width; 
		instances[ninst].refdata = (jb2SymbolInstanceRefined *)NULL; 
		} 
 
	    /* Place the symbol */ 
	    if (transposed) { 
		if ((refcorner == JB2_BOTTOMLEFT) || (refcorner == JB2_BOTTOMRIGHT)) { 
		    curs += curh - 1; 
		    } 
		} 
	    else { 
		if ((refcorner == JB2_TOPRIGHT) || (refcorner == JB2_BOTTOMRIGHT)) { 
		    curs += curw - 1; 
		    } 
		} 
 
	    /* We're going to normalise the symbol to top-left */ 
	    /* 
	     * (curs, curt) (or (curt, curs) if transposed is set) is the 
	     * location of the symbol's reference corner 
	     */ 
	    if (transposed) { 
		switch(refcorner) { 
		  case JB2_BOTTOMLEFT: { 
		    instances[ninst].xpos = curt; 
		    instances[ninst].ypos = curs - curh + 1; 
		    break; 
		    } 
	       
		  case JB2_TOPLEFT: { 
		    instances[ninst].xpos = curt; 
		    instances[ninst].ypos = curs; 
		    break; 
		    } 
 
		  case JB2_BOTTOMRIGHT: { 
		    instances[ninst].xpos = curt - curw + 1; 
		    instances[ninst].ypos = curs - curh + 1; 
		    break; 
		    } 
	       
		  case JB2_TOPRIGHT: { 
		    instances[ninst].xpos = curt - curw + 1; 
		    instances[ninst].ypos = curs; 
		    break; 
		    } 
		  } 
		} 
	    else { 
		switch(refcorner) { 
		  case JB2_BOTTOMLEFT: { 
		    instances[ninst].xpos = curs; 
		    instances[ninst].ypos = curt - curh + 1; 
		    break; 
		    } 
	       
		  case JB2_TOPLEFT: { 
		    instances[ninst].xpos = curs; 
		    instances[ninst].ypos = curt; 
		    break; 
		    } 
 
		  case JB2_BOTTOMRIGHT: { 
		    instances[ninst].xpos = curs - curw + 1; 
		    instances[ninst].ypos = curt - curh + 1; 
		    break; 
		    } 
	       
		  case JB2_TOPRIGHT: { 
		    instances[ninst].xpos = curs - curw + 1; 
		    instances[ninst].ypos = curt; 
		    break; 
		    } 
		  } 
		} 
 
	    /* Finish the update of curs */ 
	    if (transposed) { 
		if ((refcorner == JB2_TOPLEFT) || (refcorner == JB2_TOPRIGHT)) { 
		    curs += curh - 1; 
		    } 
		} 
	    else { 
		if ((refcorner == JB2_TOPLEFT) || (refcorner == JB2_BOTTOMLEFT)) { 
		    curs += curw - 1; 
		    } 
		} 
		 
	    /* Copy the rest of the info into instances */ 
	    instances[ninst].sym = sym; 
	    ninst++; 
	    } 
	} 
 
    /* All done. */ 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    if (bits != (Dgp_UInt32 *)NULL) { 
	(a->free)((void *)bits); 
	} 
 
    if (ref != (jb2SymbolInstanceRefined *)NULL) { 
	(a->free)((void *)ref); 
	} 
 
    if (refdata != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)refdata); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Blit bits into im at (xpos, ypos), clipping the blit to the region 
 * position and size. 
 */ 
static void 
blitOffset(DgpImage_Packed im, Dgp_UInt32 *bits, Dgp_Int32 w, Dgp_Int32 h, Dgp_Int32 longsperline, 
	   Dgp_Int32 xpos, Dgp_Int32 ypos, 
	   Dgp_Int32 regxpos, Dgp_Int32 regypos, Dgp_Int32 regwidth, Dgp_Int32 regheight) 
{ 
    /* Does it fit entirely in the region? */ 
    if ((xpos >= 0) && (ypos >= 0) && 
	(xpos + w <= regwidth) && 
	(ypos + h <= regheight)) { 
	/* It fits - no clipping needed */ 
	DgpBlat_blitBitsNoClip(DGPIMG_GETROW(im, 0), 
		       DGPIMG_GETROWWIDTH(im) / (int)sizeof(Dgp_UInt32), 
		       bits, longsperline, w, h, 
		       xpos + regxpos, ypos + regypos, 
			   im->dgp_state); 
	} 
    else { 
	/* Do a complex blit, clipping on all four edges */ 
	/* 
	 * What are the clipped bounds of the symbol (in symbol 
	 * coordinates?) Inclusive on min, exclusive on max 
	 */ 
	Dgp_Int32 minx = MAX(-xpos, 0); 
	Dgp_Int32 miny = MAX(-ypos, 0); 
	Dgp_Int32 maxx = MIN(w, regwidth - xpos); 
	Dgp_Int32 maxy = MIN(h, regheight - ypos); 
 
	if ((minx >= maxx) || (miny >= maxy)) { 
	    /* Nothing to do - entirely clipped */ 
	    return; 
	    } 
 
	/* 
	 * We can do the Y clip for free by adjusting the pointer 
	 * and height. 
	 */ 
	DgpBlat_blitBitsClipLR(DGPIMG_GETROW(im, 0), 
		       DGPIMG_GETROWWIDTH(im) / (int)sizeof(Dgp_UInt32), 
		       bits + miny * longsperline, 
		       longsperline, 
		       minx, maxx - minx, maxy - miny, 
		       xpos + regxpos + minx, 
		       ypos + regypos + miny, 
			   im->dgp_state); 
	} 
    } 
 
/* 
 * Blit bits into im at (xpos, ypos), clipping the blit to the region 
 * position and size.  Use the provided combop. 
 */ 
static void 
blitOffsetOp(DgpImage_Packed im, 
	     Dgp_UInt32 *bits, Dgp_Int32 w, Dgp_Int32 h, Dgp_Int32 longsperline, 
	     Dgp_Int32 xpos, Dgp_Int32 ypos, 
	     Dgp_Int32 regxpos, Dgp_Int32 regypos, Dgp_Int32 regwidth, Dgp_Int32 regheight, 
	     Dgp_Int32 combop) 
{ 
    /* 
     * We're going to be calling blitBits, which does the clipping 
     * checks.  The region had better fit entirely into the page bitmap. 
     */ 
    DgpBlat_blitBitsOp(DGPIMG_GETROW(im, 0), 
	       DGPIMG_GETROWWIDTH(im) / (int)sizeof(Dgp_UInt32), 
	       regxpos, regypos, regxpos + regwidth, regypos + regheight, 
	       bits, longsperline, w, h, 
	       xpos + regxpos, 
	       ypos + regypos, jb2Minterms[combop], 
		   im->dgp_state); 
    } 
 
/* 
 * Draw the instances into the buffer.  The buffer must have been pre-cleared 
 * to the appropriate defpixel. 
 */ 
void 
jb2DrawInstances(Dgp_UInt32 *bits, Dgp_Int32 w, Dgp_Int32 h, 
		 Dgp_Int32 longsperline, jb2SymbolInstance *instances, 
		 Dgp_Int32 ninstances, Dgp_Int32 combop, Dgp_State *dgp_state) 
{ 
    Dgp_Int32 i; 
    jb2SymbolInstance *curinst; 
 
    for (i = 0; i < ninstances; i++) { 
	curinst = &instances[i]; 
 
	if (curinst->refdata != (jb2SymbolInstanceRefined *)NULL) { 
	    /* Ignore the sym; use the refdata */ 
	    jb2SymbolInstanceRefined *refdata = curinst->refdata; 
 
	    DgpBlat_blitBitsOp(bits, longsperline, 0, 0, w, h, 
		       refdata->bits, refdata->longsperline, 
		       refdata->width, refdata->height, 
		       curinst->xpos, curinst->ypos, jb2Minterms[combop], 
			   dgp_state); 
	    } 
	else { 
	    /* Use the sym */ 
	    jb2Symbol *sym = curinst->sym; 
 
	    DgpBlat_blitBitsOp(bits, longsperline, 0, 0, w, h, 
		       sym->bits, sym->longsperline, sym->width, sym->height, 
		       curinst->xpos, curinst->ypos, jb2Minterms[combop], 
			   dgp_state); 
	    } 
	} 
    } 
 
/* Create a buffer and render into it */ 
ErrCode 
jb2RenderSymReg(jb2File file, jb2Segment *segment) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2RenderSymReg()", DGP_PERRSTATE); 
 
    jb2RegSegment *reginfo; 
    Dgp_Int32 ninstances; 
    jb2SymbolInstance *instances; 
    Dgp_Int32 i; 
    DgpImage_Packed im = (DgpImage_Packed)NULL; 
 
    ninstances = segment->data.symreg.ninstances; 
    instances = segment->data.symreg.instances; 
    reginfo = &segment->reginfo; 
 
    /* Done already? */ 
    if (reginfo->rendered) { 
	NORMAL_RETURN; 
	} 
     
    /* Allocate the buffer */ 
    DGP_ERR_PASS (DgpImage_new(dgp_state, DGPIMAGE_PACKED, 
			    reginfo->regwidth, reginfo->regheight, (void**) &im)); 
 
    /* Set to 1 if that's the defpixel */ 
    if (segment->data.symreg.defpixel != 0) { 
	DgpBlat_blitFillPacked(im); 
	} 
 
    /* Do the rendering */ 
    for (i = 0; i < ninstances; i++) { 
	if (instances[i].refdata != (jb2SymbolInstanceRefined *)NULL) { 
	    /* Ignore the sym; just use the refdata */ 
	    jb2SymbolInstanceRefined *refdata = instances[i].refdata; 
	    blitOffsetOp(im, refdata->bits, refdata->width, refdata->height, 
			 refdata->longsperline, 
			 instances[i].xpos, instances[i].ypos, 
			 0, 0, 
			 (int)reginfo->regwidth, (int)reginfo->regheight, 
			 segment->data.symreg.int_combop); 
	    } 
	else { 
	    /* Use the sym */ 
	    jb2Symbol *sym = instances[i].sym; 
	    blitOffsetOp(im, sym->bits, sym->width, sym->height, 
			 sym->longsperline, 
			 instances[i].xpos, instances[i].ypos, 
			 0, 0, 
			 (int)reginfo->regwidth, (int)reginfo->regheight, 
			 segment->data.symreg.int_combop); 
	    } 
	} 
 
    /* Stash the bitmap away */ 
    reginfo->im = im; 
    reginfo->rendered = DGP_TRUE; 
 
    NORMAL_RETURN; 
 
    } 
 
/* Render the region into the page buffer */ 
ErrCode 
jb2RenderSymRegIntoPage(jb2File file, jb2Segment *segment, DgpImage_Packed page) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2RenderSymRegIntoPage()", DGP_PERRSTATE); 
 
    jb2RegSegment *reginfo; 
    Dgp_Int32 ninstances; 
    jb2SymbolInstance *instances; 
    Dgp_Int32 i; 
 
    ninstances = segment->data.symreg.ninstances; 
    instances = segment->data.symreg.instances; 
    reginfo = &segment->reginfo; 
 
    /* You can't render an intermediate segment into a page */ 
    DGP_ASSERT(segment->segtype != JB2SEGTYPE_SYMREG_INT); 
 
    /* The region must fit entirely on the page */ 
    if (((int)reginfo->regxpos + (int)reginfo->regwidth > DGPIMG_WIDTH(page)) || 
	((int)reginfo->regypos + (int)reginfo->regheight > DGPIMG_HEIGHT(page))) { 
	FILEERR(cErrCompressionError, "region extends off page"); 
	} 
 
    /* Can't handle everything the same way */ 
    if ((reginfo->combop != COMBOP_OR) || 
	(segment->data.symreg.int_combop != COMBOP_OR) || 
	(segment->data.symreg.defpixel != 0)) { 
	/* Render the region, then combine it */ 
 
      DGP_ERR_PASS_CLEANUP(jb2RenderSymReg(file, segment)); 
 
	DgpBlat_blitPackedOp(page, reginfo->im, 
		     (int)reginfo->regxpos, (int)reginfo->regypos, 
		     jb2Minterms[reginfo->combop]); 
	} 
    else { 
	/* 
	 * Everything's using OR and defpixel=0, so we can optimise 
	 * by drawing straight into the page buffer 
	 */ 
 
	/* 
	 * Blit into the page bitmap, but clip according to the region bounds. 
	 * The xpos and ypos values are relative to the region origin 
	 * (reginfo->regxpos, reginfo->regypos). 
	 */ 
	for (i = 0; i < ninstances; i++) { 
	    if (instances[i].refdata != (jb2SymbolInstanceRefined *)NULL) { 
		/* Ignore the sym; just use the refdata */ 
		jb2SymbolInstanceRefined *refdata = instances[i].refdata; 
		blitOffset(page, refdata->bits, 
			   refdata->width, refdata->height, 
			   refdata->longsperline, 
			   instances[i].xpos, instances[i].ypos, 
			   (int)reginfo->regxpos, (int)reginfo->regypos, 
			   (int)reginfo->regwidth, (int)reginfo->regheight); 
		} 
	    else { 
		/* Use the sym */ 
		jb2Symbol *sym = instances[i].sym; 
		blitOffset(page, sym->bits, sym->width, sym->height, 
			   sym->longsperline, 
			   instances[i].xpos, instances[i].ypos, 
			   (int)reginfo->regxpos, (int)reginfo->regypos, 
			   (int)reginfo->regwidth, (int)reginfo->regheight); 
		} 
	    } 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
