/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/encgenrefreg.c,v 1.10 2003/01/27 21:48:33 awilliam Exp $ 
 * 
 */ 
 
#include "jb2file.h" 
#include "datasink.h" 
#include "segment.h" 
#include "segment_util.h" 
#include "encgenrefreg.h" 
#include "decgenrefreg.h" 
#include <stdlib.h> 
#include "mqcoder.h" 
#include "image_defs.h"
#include "PsgFix.h"
 
#define DGP_ENC DGP_FALSE

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/encgenrefreg.c,v 1.10 2003/01/27 21:48:33 awilliam Exp $"); 
 
/* Internal routines */ 
#ifndef	NOMQ 
static ErrCode encRefImagePacked(DgpImage_Packed orig, DgpImage_Packed ref, 
				 Dgp_Int32 refdx, Dgp_Int32 refdy, arith_ref_params *params, 
				 Dgp_UInt8 *contexts, DgpEncode_MQEncoder *enc, 
			      Dgp_State *dgp_state); 
static ErrCode encRefImageT13(DgpImage_Packed orig, DgpImage_Packed ref, 
			      Dgp_Int32 refdx, Dgp_Int32 refdy, arith_ref_params *params, 
			      Dgp_UInt8 *contexts, DgpEncode_MQEncoder *enc, 
			      Dgp_State *dgp_state); 
static ErrCode encRefImageT10(DgpImage_Packed orig, DgpImage_Packed ref, 
			      Dgp_Int32 refdx, Dgp_Int32 refdy, arith_ref_params *params, 
			      Dgp_UInt8 *contexts, DgpEncode_MQEncoder *enc, 
			      Dgp_State *dgp_state); 
#endif 
 
/* 
 * Encode the AT pixel parameters.  Use params->rtemplate to say which 
 * ones to encode. 
 */ 
ErrCode 
jb2EncodeRABParams(jb2File file, DataSink data, 
		   arith_ref_params *params) 
{ 
  DGP_PROLOGUE("jb2EncodeRABParams()", jb2_pErrState(file)); 
 
    Dgp_UInt8 cbuf; 
    Dgp_Int32 i, n; 
 
    if (params->rtemplate == 0) { 
	/* Two AT pixels */ 
	n = 2; 
	} 
    else { 
	/* No AT pixels */ 
	n = 0; 
	} 
 
    for (i = 0; i < n; i++) { 
	cbuf = (Dgp_UInt8)params->ratx[i]; 
	DSPUTC(data, cbuf); 
 
	cbuf = (Dgp_UInt8)params->raty[i]; 
	DSPUTC(data, cbuf); 
#ifdef	VALPRINT 
	fprintf(stderr, "RATX%d %d\n", i + 1, params->ratx[i]); 
	fprintf(stderr, "RATY%d %d\n", i + 1, params->raty[i]); 
#endif 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* Inner routines */ 
#ifndef	NOMQ 
 
/* 
 * enc is an DgpEncode_MQEncoder. 
 * Encode bitmap using reference as a reference bitmap. 
 * The encoder MUST be using the software conventions. 
 * 
 * If contexts is not NULL, it must point to a block of contexts large 
 * enough to handle the template size given.  If contexts is NULL, this 
 * routine will allocate a block of contexts, clear it, use it, and discard 
 * it. 
 * 
 * For coding simplicity, it's best to wrap a DgpImage_Packed around 
 * the bitmaps and then use them. 
 */ 
ErrCode 
jb2ArithEncodeRefinement(Dgp_UInt32 *bitmap, Dgp_Int32 width, Dgp_Int32 height, 
			 Dgp_Int32 Dgp_Int32sperline, 
			 Dgp_UInt32 *reference, 
			 Dgp_Int32 refwidth, Dgp_Int32 refheight, 
			 Dgp_Int32 refDgp_Int32sperline, 
			 Dgp_Int32 refdx, Dgp_Int32 refdy, 
			 arith_ref_params *params, 
			 Dgp_UInt8 *contexts, 
			 jb2File file, DgpEncode_MQEncoder *enc) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2ArithEncodeRefinement()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Bool mycontexts = DGP_FALSE; 
    DgpImage_Packed orig = (DgpImage_Packed)NULL; 
    DgpImage_Packed ref = (DgpImage_Packed)NULL; 
    size_t cxsize; 
	Dgp_UInt32 *retval; 
 
    cxsize = jb2RABContextsNeeded(params); 
 
    /* Make sure we have a context block */ 
    if (contexts == (Dgp_UInt8 *)NULL) { 
	contexts = (Dgp_UInt8 *)(a->calloc)(cxsize, 1); 
	if (contexts == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
	mycontexts = DGP_TRUE; 
	} 
 
    /* Make our DgpImage_Packeds */ 
    DGP_ERR_PASS_CLEANUP (DgpImage_wrap(dgp_state, bitmap, width, height, Dgp_Int32sperline, &orig)); 
 
    DGP_ERR_PASS_CLEANUP (DgpImage_wrap(dgp_state, reference, refwidth, refheight, refDgp_Int32sperline, &ref)); 
 
    DGP_ERR_PASS_CLEANUP(encRefImagePacked(orig, ref, refdx, refdy, params, 
					   contexts, enc, dgp_state)); 
 
    if (mycontexts) { 
	(a->free)((void *)contexts); 
	} 
 
    DgpImage_unwrap(orig, &retval); 
    DgpImage_unwrap(ref, &retval); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (mycontexts && (contexts != (Dgp_UInt8 *)NULL)) { 
	(a->free)((void *)contexts); 
	} 
 
    if (orig != (DgpImage_Packed)NULL) { 
	DgpImage_unwrap(orig, &retval); 
	} 
 
    if (ref != (DgpImage_Packed)NULL) { 
	DgpImage_unwrap(ref, &retval); 
	} 
 
    ERR_RETURN; 
    } 
 
/* Dispatch on the template. */ 
static ErrCode 
encRefImagePacked(DgpImage_Packed orig, DgpImage_Packed ref, 
		  Dgp_Int32 refdx, Dgp_Int32 refdy, arith_ref_params *params, 
		  Dgp_UInt8 *contexts, DgpEncode_MQEncoder *enc, 
		  Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("encRefImagePacked()", DGP_PERRSTATE); 
 
    if (params->rtemplate == 0) { 
      DGP_ERR_PASS(encRefImageT13(orig, ref, refdx, refdy, params, contexts, 
				  enc, dgp_state)); 
	} 
    else { 
      DGP_ERR_PASS(encRefImageT10(orig, ref, refdx, refdy, params, contexts, 
				  enc, dgp_state)); 
	} 
 
    NORMAL_RETURN; 
    } 
 
static ErrCode 
encRefImageT13(DgpImage_Packed orig, DgpImage_Packed ref, 
	       Dgp_Int32 refdx, Dgp_Int32 refdy, arith_ref_params *params, 
	       Dgp_UInt8 *contexts, DgpEncode_MQEncoder *enc, 
	       Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("encRefImageT13()", DGP_PERRSTATE); 
 
    Dgp_UInt32 cx, tprcx; 
    Dgp_Int32 x, y; 
    Dgp_Int32 bit; 
    Dgp_Int32 dwidth = DGPIMG_WIDTH(orig), dheight = DGPIMG_HEIGHT(orig); 
    Dgp_Int32 rwidth = DGPIMG_WIDTH(ref), rheight = DGPIMG_HEIGHT(ref); 
    Dgp_Bool ltp = DGP_FALSE; 
 
    /* 
     * The template looks like: 
     * 1XX		2XX 
     * X?		X?X 
     *			XXX 
     * where the left half is in the orig image and the right half is in 
     * the ref image.  Also, pixel (0,0) in orig corresponds to pixel 
     * (-refdx, -refdy) in ref). 
     * 
     * We're going to form a context that looks like 
     * 1aa2bbcdddeee 
     * where "aa" is the top line on L, "bb" is the top line on R, 
     * "c" is the middle line on L, "ddd" is the middle line on R, 
     * "eee" is the bottom line on R.  Updating the context involves 
     * shifting it left by 1, ANDing it with 00 1001 0011 0110 
     * and doing 7 bit loads into bits 12 (AT1), 10 (top L), 9 (AT2), 
     * 7 (top R), 6 (middle L), 3 (middle R) and 0 (bottom R). 
     * We load from (after shift coordinates) L(ATX1, ATY1), L(1,-1), 
     * R(ATX2, ATY2), R(1, -1), L(-1, 0), R(1, 0), R(1, 1). 
     * The L(-1, 0) load is special because it can be done after the encode 
     * of the previous bit. 
     * 
     * We're also going to keep track of a context for TPR, which is 
     * bbbcccddd 
     * where "bbb" includes the pixel marked "2" above, as well as the 
     * "bb" pixels.  The "2" pixel might not be at that exact location, 
     * since it's an AT pixel, so we can't just mask bits out of cx. 
     * For each bit, we shift tprcx and AND with 01 1011 0110, then 
     * load up 3 bits into positions 6 (R(1, -1)), 3 (R(1, 0)), 0(R(1, 1)). 
     */ 
 
    /* 
     * Given values in cx, x and y, and bit, 
     * shift and mask CX (including adding the bit).  Also shift and 
     * mask tprcx. 
     */ 
#define	DOUPDATE \
    do { \
	DGP_REPORTF1(DGP_ENC, "%d", bit); \
	DGP_REPORTF1(DGP_ENC, "(%04lx)", cx); \
	cx = ((cx << 1) & 0x936) | (bit << 6); \
	tprcx = ((tprcx << 1) & 0x1b6); \
	} while(0) 
 
    for (y = 0; y < dheight; y++) { 
	/* Is TPR on? */ 
	if (params->tpron) { 
	    /* 
	     * At this point we could check to see if the row is typical. 
	     * Right now, we just say "not typical" in every case. 
	     * WRITEME 
	     */ 
	    /* 
	     * Yes; encode a bit using context 
	     * 000		000 
	     * 0?		010 
	     *			000 
	     * which linearises to 0 0000 0001 0000 
	     */ 
	    bit = 0;		/* Always non-typical */ 
 
	    DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[0x0010],  
					    bit ^ ltp)), 
				cErrIO_WriteError); 
 
	    DGP_REPORTF1(DGP_ENC, "TPR%d", bit ^ ltp); 
	    ltp = (Dgp_Bool)bit; 
	} 
 
	/* 
	 * Do the initial load.  The 7 bit loads happen immediately 
	 * after this, so we just need bits 0a00b00dd0ee0.  In fact, 
	 * some of the "b", "dd", "ee" bits may not exist (off one 
	 * edge or the other of ref). 
	 * For TPR, we want bb0dd0ee0. 
	 */ 
	cx = 0; 
	tprcx = 0; 
 
	if (y - 1 >= 0) { 
	    /* Load up "a" */ 
	    cx |= DGPIMG_GETBIT01(orig, 0, y - 1) << 11; 
	    } 
 
	if ((y - 1 - refdy >= 0) && (y - 1 - refdy < rheight)) { 
	    if ((-refdx - 1 >= 0) && (-refdx - 1 < rwidth)) { 
		/* Load up first "b" in TPR */ 
		tprcx |= DGPIMG_GETBIT01(ref, -refdx - 1, y - refdy - 1) << 9; 
		} 
	    if ((-refdx >= 0) && (-refdx < rwidth)) { 
		/* Load up "b" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx, y - 1 - refdy); 
		cx |= bit << 8; 
		tprcx |= bit << 8; 
		} 
	    } 
 
	if ((y - refdy >= 0) && (y - refdy < rheight)) { 
	    if ((-refdx - 1 >= 0) && (-refdx - 1 < rwidth)) { 
		/* Load up the first "d" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx - 1, y - refdy); 
		cx |= bit << 5; 
		tprcx |= bit << 5; 
		} 
	    if ((-refdx >= 0) && (-refdx < rwidth)) { 
		/* Load up the second "d" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx, y - refdy); 
		cx |= bit << 4; 
		tprcx |= bit << 4; 
		} 
	    } 
 
	if ((y + 1 - refdy >= 0) && (y + 1 - refdy < rheight)) { 
	    if ((-refdx - 1 >= 0) && (-refdx - 1 < rwidth)) { 
		/* Load up the first "e" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx - 1, y + 1 - refdy); 
		cx |= bit << 2; 
		tprcx |= bit << 2; 
		} 
	    if ((-refdx >= 0) && (-refdx < rwidth)) { 
		/* Load up the second "e" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx, y + 1 - refdy); 
		cx |= bit << 1; 
		tprcx |= bit << 1; 
		} 
	    } 
 
	for (x = 0; x < dwidth; x++) { 
	    /* Load the AT pixels */ 
	    if ((x + params->ratx[0] >= 0) && (x + params->ratx[0] < dwidth) && 
		(y + params->raty[0] >= 0)) { 
		cx |= DGPIMG_GETBIT01(orig, x + params->ratx[0], 
				  y + params->raty[0]) << 12; 
		} 
	    if ((x - refdx + params->ratx[1] >= 0) && 
		(x - refdx + params->ratx[1] < rwidth) && 
		(y - refdy + params->raty[1] >= 0) && 
		(y - refdy + params->raty[1] < rheight)) { 
		cx |= DGPIMG_GETBIT01(ref, x - refdx + params->ratx[1], 
				  y - refdy + params->raty[1]) << 9; 
		} 
 
	    /* Load top L pixel */ 
	    if ((x + 1 < dwidth) && (y - 1 >= 0)) { 
		cx |= DGPIMG_GETBIT01(orig, x + 1, y - 1) << 10; 
		} 
 
	    if ((x - refdx + 1 >= 0) && (x - refdx + 1 < rwidth)) { 
		/* Load top R pixel */ 
		if ((y - refdy - 1 >= 0) && (y - refdy - 1 < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx + 1, y - refdy - 1); 
		    cx |= bit << 7; 
		    tprcx |= bit << 6; 
		    } 
 
		/* Load middle R pixel */ 
		if ((y - refdy >= 0) && (y - refdy < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx + 1, y - refdy); 
		    cx |= bit << 3; 
		    tprcx |= bit << 3; 
		    } 
 
		/* Load bottom R pixel */ 
		if ((y - refdy + 1 >= 0) && (y - refdy + 1 < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx + 1, y - refdy + 1); 
		    cx |= bit; 
		    tprcx |= bit; 
		    } 
		} 
 
	    /* 
	     * Is this pixel typical? 
	     * If all the 9 pixels on the R side have the same value, 
	     * and ltp is DGP_TRUE, don't encode anything, just hallucinate 
	     * that shared value. 
	     */ 
	    bit = DGPIMG_GETBIT01(orig, x, y); 
	    if (ltp) { 
		if (((tprcx == 0) && (bit == 0)) || 
		    ((tprcx == 0x1ff) && (bit == 1))) { 
		    /* Don't do anything */ 
		    } 
		else { 
		    DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[cx], bit)), 
				cErrIO_WriteError); 
		    } 
		} 
	    else { 
		DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[cx], bit)), 
				cErrIO_WriteError); 
		} 
 
	    DOUPDATE; 
	    } 
	DGP_REPORTF1(DGP_ENC, "%c", '\n'); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
#undef	DOUPDATE 
 
static ErrCode 
encRefImageT10(DgpImage_Packed orig, DgpImage_Packed ref, 
	       Dgp_Int32 refdx, Dgp_Int32 refdy, arith_ref_params *params, 
	       Dgp_UInt8 *contexts, DgpEncode_MQEncoder *enc, 
	       Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("encRefImageT10()", DGP_PERRSTATE); 
 
    Dgp_UInt32 cx, tprcx; 
    Dgp_Int32 x, y; 
    Dgp_Int32 bit; 
    Dgp_Int32 dwidth = DGPIMG_WIDTH(orig), dheight = DGPIMG_HEIGHT(orig); 
    Dgp_Int32 rwidth = DGPIMG_WIDTH(ref), rheight = DGPIMG_HEIGHT(ref); 
    Dgp_Bool ltp = DGP_FALSE; 
 
    /* 
     * The template looks like: 
     * XXX		 X 
     * X?		X?X 
     *			 XX 
     * where the left half is in the orig image and the right half is in 
     * the ref image.  Also, pixel (0,0) in orig corresponds to pixel 
     * (-refdx, -refdy) in ref). 
     * 
     * We're going to form a context that looks like 
     * aaabcdddee 
     * where "aaa" is the top line on L, "b" is the top line on R, 
     * "c" is the middle line on L, "ddd" is the middle line on R, 
     * "ee" is the bottom line on R.  Updating the context involves 
     * shifting it left by 1, ANDing it with 011 0001 1010 
     * and doing 5 bit loads into bits 7 (top L), 6 (top R), 
     * 5 (middle L), 2 (middle R) and 0 (bottom R). 
     * We load from (after shift coordinates) L(1,-1), 
     * R(0, -1), L(-1, 0), R(1, 0), R(1, 1). 
     * The L(-1, 0) load is special because it can be done after the encode 
     * of the previous bit. 
     * 
     * We're also going to keep track of a context for TPR, which is 
     * bbbcccddd 
     * where "bbb" includes the pixel marked "2" above, as well as the 
     * "bb" pixels.  The "2" pixel might not be at that exact location, 
     * since it's an AT pixel, so we can't just mask bits out of cx. 
     * For each bit, we shift tprcx and AND with 01 1011 0110, then 
     * load up 3 bits into positions 6 (R(1, -1)), 3 (R(1, 0)), 0(R(1, 1)). 
     */ 
 
    /* 
     * Given values in cx, x and y, and bit, shift and mask CX, 
     * (including adding the bit).  Also shift and mask tprcx. 
     */ 
#define	DOUPDATE \
    do { \
	DGP_REPORTF1(DGP_ENC, "%d", bit); \
	cx = ((cx << 1) & 0x31a) | (bit << 5); \
	tprcx = ((tprcx << 1) & 0x1b6); \
	} while(0) 
 
    for (y = 0; y < dheight; y++) { 
	/* Is TPR on? */ 
	if (params->tpron) { 
	    /* 
	     * At this point we could check to see if the row is typical. 
	     * Right now, we just say "not typical" in every case. 
	     * WRITEME 
	     */ 
	    /* 
	     * Yes; encode a bit using context 
	     * 000		 0  
	     * 0?		010 
	     *			 00 
	     * which linearises to 00 0000 1000 
	     */ 
	    bit = 0;		/* Always non-typical */ 
 
	    DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[0x0008], bit ^ ltp)), 
				cErrIO_WriteError); 
	    DGP_REPORTF1(DGP_ENC, "TPR%d", bit ^ ltp); 
	    ltp = (Dgp_Bool)bit; 
	    } 
 
	/* 
	 * Do the initial load.  The 5 bit loads happen immediately 
	 * after this, so we just need bits aa000dd0e0.  In fact, 
	 * the first "a" bit there doesn't exist (off the left edge of orig), 
	 * and some of the "dd", "e" bits may not exist (off one 
	 * edge or the other of ref). 
	 * For TPR, we want bb0dd0ee0. 
	 */ 
	cx = 0; 
	tprcx = 0; 
 
	if (y - 1 >= 0) { 
	    /* Load up second "a" */ 
	    cx |= DGPIMG_GETBIT01(orig, 0, y - 1) << 8; 
	    } 
 
	if ((y - 1 - refdy >= 0) && (y - 1 - refdy < rheight)) { 
	    if ((-refdx - 1 >= 0) && (-refdx - 1 < rwidth)) { 
		/* Load up first "b" in TPR */ 
		tprcx |= DGPIMG_GETBIT01(ref, -refdx - 1, y - refdy - 1) << 9; 
		} 
	    if ((-refdx >= 0) && (-refdx < rwidth)) { 
		/* Load up second "b" in TPR */ 
		bit = DGPIMG_GETBIT01(ref, -refdx, y - 1 - refdy); 
		tprcx |= bit << 8; 
		} 
	    } 
 
	if ((y - refdy >= 0) && (y - refdy < rheight)) { 
	    if ((-refdx - 1 >= 0) && (-refdx - 1 < rwidth)) { 
		/* Load up the first "d" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx - 1, y - refdy); 
		cx |= bit << 4; 
		tprcx |= bit << 5; 
		} 
	    if ((-refdx >= 0) && (-refdx < rwidth)) { 
		/* Load up the second "d" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx, y - refdy); 
		cx |= bit << 3; 
		tprcx |= bit << 4; 
		} 
	    } 
 
	if ((y + 1 - refdy >= 0) && (y + 1 - refdy < rheight)) { 
	    if ((-refdx - 1 >= 0) && (-refdx - 1 < rwidth)) { 
		/* Load up the first "e" in TPR */ 
		bit = DGPIMG_GETBIT01(ref, -refdx - 1, y + 1 - refdy); 
		tprcx |= bit << 2; 
		} 
	    if ((-refdx >= 0) && (-refdx < rwidth)) { 
		/* Load up the second "e" */ 
		bit = DGPIMG_GETBIT01(ref, -refdx, y + 1 - refdy); 
		cx |= bit << 1; 
		tprcx |= bit << 1; 
		} 
	    } 
 
	for (x = 0; x < dwidth; x++) { 
	    /* Load top L pixel */ 
	    if ((x + 1 < dwidth) && (y - 1 >= 0)) { 
		cx |= DGPIMG_GETBIT01(orig, x + 1, y - 1) << 7; 
		} 
 
	    if ((x - refdx >= 0) && (x - refdx < rwidth)) { 
		/* Load top R pixel */ 
		if ((y - refdy - 1 >= 0) && (y - refdy - 1 < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx, y - refdy - 1); 
		    cx |= bit << 6; 
		    } 
		} 
 
	    if ((x - refdx + 1 >= 0) && (x - refdx + 1 < rwidth)) { 
		/* Load top R pixel for TPR */ 
		if ((y - refdy - 1 >= 0) && (y - refdy - 1 < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx + 1, y - refdy - 1); 
		    tprcx |= bit << 6; 
		    } 
 
		/* Load middle R pixel */ 
		if ((y - refdy >= 0) && (y - refdy < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx + 1, y - refdy); 
		    cx |= bit << 2; 
		    tprcx |= bit << 3; 
		    } 
 
		/* Load bottom R pixel */ 
		if ((y - refdy + 1 >= 0) && (y - refdy + 1 < rheight)) { 
		    bit = DGPIMG_GETBIT01(ref, x - refdx + 1, y - refdy + 1); 
		    cx |= bit; 
		    tprcx |= bit; 
		    } 
		} 
 
	    /* 
	     * Is this pixel typical? 
	     * If all the 9 pixels on the R side have the same value, 
	     * and ltp is DGP_TRUE, don't encode anything, just hallucinate 
	     * that shared value. 
	     */ 
	    bit = DGPIMG_GETBIT01(orig, x, y); 
	    if (ltp) { 
		if (((tprcx == 0) && (bit == 0)) || 
		    ((tprcx == 0x1ff) && (bit == 1))) { 
		    /* Don't do anything */ 
		    } 
		else { 
		    DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[cx], bit)), 
				cErrIO_WriteError); 
		    } 
		} 
	    else { 
		DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[cx], bit)), 
				cErrIO_WriteError); 
		} 
 
	    DOUPDATE; 
	    } 
	DGP_REPORTF1(DGP_ENC, "%c", '\n'); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
#undef	DOUPDATE 
 
#endif 
