/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/encgenreg.c,v 1.21 2003/01/27 21:48:33 awilliam Exp $ 
 * 
 */ 
 
 
#include "util.h" 
#include "jb2file.h" 
#include "params.h" 
#include "encgenreg.h" 
#include "decgenreg.h" 
#include "image.h" 
#include "segment_util.h" 
#include "image_defs.h"
#include "PsgFix.h"
 
/* Stuff you need for the DigiPaper G4 routines */ 
#include "g4.h" 
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/encgenreg.c,v 1.21 2003/01/27 21:48:33 awilliam Exp $"); 
 
/* debug stuff */
#define DGP_MESSAGE DGP_FALSE

/* Internal routines */ 
#ifndef	NOMQ 
static ErrCode arithEncodeImageT16(DgpImage_Packed src, 
				   arith_bitmap_params *params, 
				   Dgp_UInt8 *contexts, DgpEncode_MQEncoder *dec, 
				   Dgp_State *dgp_state); 
static ErrCode arithEncodeImageT13(DgpImage_Packed src, 
				   arith_bitmap_params *params, 
				   Dgp_UInt8 *contexts, DgpEncode_MQEncoder *dec, 
				   Dgp_State *dgp_state); 
static ErrCode arithEncodeImageT10L3(DgpImage_Packed src, 
				     arith_bitmap_params *params, 
				     Dgp_UInt8 *contexts, DgpEncode_MQEncoder *dec, 
				     Dgp_State *dgp_state); 
static ErrCode arithEncodeImageT10L2(DgpImage_Packed src, 
				     arith_bitmap_params *params, 
				     Dgp_UInt8 *contexts, DgpEncode_MQEncoder *dec, 
				     Dgp_State *dgp_state); 
#endif 
 
/* 
 * Encode the segment as a generic region. 
 * 
 * There are two possibilities here: 
 * 1. The segment has already been encoded: genreg.data is the encoded 
 *    data block; length in genreg.datalen.  The parameters used to 
 *    encode the data block must be stored in genreg.mmr, genreg.abparams. 
 * 2. The segment has not been encoded; reginfo.rendered is DGP_TRUE and 
 *    reginfo.im contains the image to be encoded.  genreg.data must be 
 *    NULL (this is how case 2 is detected). 
 * 
 * In case 1, the routine simply writes out a header, then writes the 
 * encoded data block. 
 * 
 * In case 2, the routine encodes the image (using the parameters in 
 * genreg.mmr, genreg.abparams), and writes out the header and encoded 
 * image.  It does *not* store the encoded data in genreg.data. 
 *  
 * The fields in segment that must be filled in already are: 
 * - pagenum 
 * - segtype 
 * - data->genreg 
 * - reginfo (except for image fields, in case 1). 
 * 
 * The remaining fields will be filled in as follows: 
 * - segnum will be allocated 
 * - datalen will be set based on the encoded segment's data length 
 * - retain will be set to DGP_FALSE 
 * - defer_nonretain will be set to DGP_FALSE 
 * - nrefsegs will be set to 0 
 * - refsegs will be set to NULL 
 * - refretain will be set to NULL 
 * 
 * The encoded segment header will be written out to 
 * jb2GetFileHeaderDataSink(file); the encoded segment data will be 
 * written out to jb2GetFileDataSink(file). 
 */ 
ErrCode 
jb2EncodeGenRegion(jb2File file, jb2Segment *segment) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2EncodeGenRegion()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
    DataSink bodyds; 
    Dgp_UInt8 cbuf; 
    Dgp_UInt8 *contexts = (Dgp_UInt8 *)NULL; 
    DgpEncode_MQEncoder mqenc; 
    size_t bytes; 
 
    DGP_ASSERT((segment->segtype == JB2SEGTYPE_GENREG_INT) || 
	   (segment->segtype == JB2SEGTYPE_GENREG_IMM) || 
	   (segment->segtype == JB2SEGTYPE_GENREG_IMMLL)); 
 
    /* Fill in the fields as promised */ 
    segment->segnum = jb2FileNextSegmentNumber(file); 
    segment->nrefsegs = 0; 
    segment->refsegs = (Dgp_Int32 *)NULL; 
    segment->refretain = (Dgp_Bool *)NULL; 
    segment->retain = DGP_FALSE; 
    segment->defer_nonretain = DGP_FALSE; 
 
    /* All but the length is known - it's OK to encode the header */ 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeader(file, segment)); 
 
    /* Write out the segment data header */ 
    bodyds = jb2GetFileDataSink(file); 
    /* Reset the count on bodyds so we can get the data length */ 
    dsResetCount(bodyds); 
 
    /* First the region segment data header */ 
    DGP_ERR_PASS_CLEANUP(jb2EncodeRegSegment(file, bodyds, 
					     &segment->reginfo)); 
 
    /* One-byte flags field */ 
    cbuf = 0; 
    /* Bit 0: MMR */ 
    if (segment->data.genreg.mmr) { 
	cbuf |= 1; 
	} 
#ifdef	NOMQ 
    else { 
      FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
	} 
#endif 
     
    /* Arith params */ 
    if (!segment->data.genreg.mmr) { 
	/* Bits 1-2: GBTEMPLATE */ 
	cbuf |= segment->data.genreg.abparams.rtemplate << 1; 
 
	/* Bit 3: TPON */ 
	if (segment->data.genreg.abparams.tpon) { 
	    cbuf |= 8; 
	    } 
	} 
     
#ifdef	VALPRINT 
    fprintf(stderr, "GBFLAGS %02x\n", cbuf); 
#endif 
    DSPUTC(bodyds, cbuf); 
 
    if (!segment->data.genreg.mmr) { 
	/* Encode the AT parameters */ 
      DGP_ERR_PASS_CLEANUP(jb2EncodeABParams(file, bodyds, &segment->data.genreg.abparams)); 
	} 
 
    if (segment->data.genreg.data != (Dgp_UInt8 *)NULL) { 
	/* Encoded data all ready to go */ 
	DSWRITE(bodyds, segment->data.genreg.datalen, 
		segment->data.genreg.data, &bytes); 
	} 
    else { 
	/* We're now in case 2. */ 
	DGP_ASSERT((segment->reginfo.rendered) && 
	       (segment->reginfo.im != (DgpImage_Packed)NULL)); 
 
	if (segment->data.genreg.mmr) { 
	    DGP_ERR_PASS_CLEANUP(jb2MMREncodeDS(file, bodyds, 
				DGPIMG_GETROW(segment->reginfo.im, 0), 
				segment->reginfo.regwidth, 
				segment->reginfo.regheight, 
				DGPIMG_GETROWWIDTH(segment->reginfo.im) / sizeof(Dgp_UInt32), 
						DGP_FALSE)); 
	    } 
	else { 
	    /* Arithmetic encoding. */ 
#ifndef	NOMQ 
	    /* Get the context block */ 
	    contexts = (Dgp_UInt8 *) 
		(a->calloc)(jb2ABContextsNeeded(&segment->data.genreg.abparams), 
				   1); 
 
	    if (contexts == (Dgp_UInt8 *)NULL) { 
	      NOMEM; 
		} 
	     
	    /* Set up the MQ encoder  */ 
	    dsInitDgpEncode_MQEncoder(bodyds, &mqenc); 
 
	    /* Encode the image */ 
	    DGP_ERR_PASS_CLEANUP(jb2ArithEncodeImagePacked(segment->reginfo.im, 
					   &segment->data.genreg.abparams, 
					   contexts, &mqenc, dgp_state)); 
 
	    /* Don't need the contexts */ 
	    (a->free)((void *)contexts); 
	    contexts = (Dgp_UInt8 *)NULL; 
 
	    /* Finalise the MQ encoder */ 
	    DGP_ERR_PASS_CLEANUP(dsFinalDgpEncode_MQEncoder(bodyds, &mqenc)); 
#endif 
	    } 
	} 
 
    /* Fix up the data length in the header */ 
    segment->datalen = dsCount(bodyds); 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeaderDataLength(file, segment)); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (contexts != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)contexts); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Encode the data in bitmap into *data; the area at *data is *datalen 
 * bytes Dgp_Int32.  If the compressed data won't fit, return DGP_FALSE. 
 * If it fits, store how many bytes were used in *datalen and return DGP_TRUE. 
 * 
 * The input image is packed with MSB=left, width 
 * by height bits, and each row having Dgp_Int32sPerRow Dgp_Int32s. 
 * 0 bits are white, and 1 bits are black. 
 * No EOL codes are used, either between lines or at the end. Encoded rows 
 * are not padded; the only padding done is on the final byte. 
 * 
 * If writeEOFB is DGP_TRUE, then it writes an EOFB (two EOLs) at the end of 
 * the bitmap data. 
 */ 
ErrCode 
jb2MMREncode(Dgp_UInt32 *bitmap, Dgp_Int32 width, Dgp_Int32 height, Dgp_Int32 Dgp_Int32sperline, 
	     Dgp_UInt8 *data, size_t *datalen, Dgp_Bool writeEOFB, 
	     Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jb2MMREncode()", DGP_PERRSTATE); 
 
  DGP_ERR_CHECK_NO_MSG(g4Encode(bitmap, width, height, Dgp_Int32sperline, data, 
				datalen, writeEOFB, dgp_state)) { 
    if(DGP_ERROR_IS_NOT(cErrBufferExceeded)) { 
      DGP_ERR_PRINTER; 
    } 
    ERR_RETURN; 
  } 
 
    NORMAL_RETURN; 
 
    } 
 
/* 
 * Like jb2MMREncode, but write the results to the DataSink. 
 */ 
ErrCode 
jb2MMREncodeDS(jb2File file, DataSink data, 
	       Dgp_UInt32 *bitmap, Dgp_Int32 width, Dgp_Int32 height, 
	       Dgp_Int32 Dgp_Int32sperline, 
	       Dgp_Bool writeEOFB) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2MMREncodeDS()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    size_t buflen, bufused, bytes; 
    Dgp_UInt8 *databuf = (Dgp_UInt8 *)NULL; 
 
    /* 
     * MMR-encode the image. 
     * This is a bit harder than it ought to be due to the lack of 
     * integration of the MMR code into the jbitfile/DataSink way 
     * of life.  Currently, we guess an upper bound on the encoded 
     * data length, allocate a buffer that large, and try to 
     * encode into it.  If that fails, we increase the buffer 
     * size and try again, repeating until we succeed or run out 
     * of memory. 
     */ 
 
    /* 
     * Initial guess: the amount of memory needed to store the 
     * uncompressed bitmap. 
     */ 
    buflen = (size_t)(width * height / DGP_BITS_UINT8); 
 
    for (;;) { 
	databuf = (Dgp_UInt8 *)(a->malloc)(buflen); 
	if (databuf == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
 
	bufused = buflen; 
	 
	DGP_ERR_CHECK_NO_MSG(jb2MMREncode(bitmap, width, height, Dgp_Int32sperline, 
					  databuf, &bufused, writeEOFB, 
					  dgp_state)) { 
 
	  /* hmmmm,, we got an error, was it a real error? */ 
	  if(DGP_ERROR_IS_NOT(cErrBufferExceeded)) { 
	    DGP_ERR_PRINTER; 
	    DGP_GOTO_CLEANUP; 
	  } 
 
	  /* oops, ran out of room, try again... */ 
	  DGP_CLEAR_CUR_ERROR; 
 
	} else { 
	    /* Worked! */ 
	    break; 
	} 
 
	/* Expand and try again */ 
	(a->free)((void *)databuf); 
	databuf = (Dgp_UInt8 *)NULL; 
	buflen += buflen / 4 + 1; 
	} 
 
    /* Write out and dispose of the encoded buffer */ 
    DSWRITE(data, bufused, databuf, &bytes); 
    (a->free)((void *)databuf); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (databuf != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)databuf); 
	} 
 
    ERR_RETURN; 
    } 
 
/* Write out the 2-byte or 8-byte AT location field. */ 
ErrCode 
jb2EncodeABParams(jb2File file, DataSink data, arith_bitmap_params *params) 
{ 
  DGP_PROLOGUE("jb2EncodeABParams()", jb2_pErrState(file)); 
 
    Dgp_Int32 i, n; 
 
    if (params->rtemplate == 0) { 
	/* 8-byte */ 
	n = 4; 
	} 
    else { 
	/* 2-byte */ 
	n = 1; 
	} 
 
    for (i = 0; i < n; i++) { 
	DSPUTC(data, (Dgp_UInt8)params->atx[i]); 
	DSPUTC(data, (Dgp_UInt8)params->aty[i]); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
#ifndef	NOMQ 
/* 
 * Write out the arithmetically-encoded bitmap using the MQcoder enc 
 * and contexts. 
 * 
 * If contexts is not NULL, it must point to a block of contexts large 
 * enough to handle the template size given.  If contexts is NULL, this 
 * routine will allocate a block of contexts, clear it, use it, and discard 
 * it. 
 * 
 * For coding simplicity, it's best to wrap a DgpImage_Packed around 
 * the source bitmap and then use that. 
 */ 
extern ErrCode 
jb2ArithEncodeImage(Dgp_UInt32 *bitmap, 
		    Dgp_Int32 width, Dgp_Int32 height, Dgp_Int32 Dgp_Int32sperline, 
		    arith_bitmap_params *params, 
		    Dgp_UInt8 *contexts, 
		    jb2File file, DgpEncode_MQEncoder *enc) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2ArithEncodeImage()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Bool mycontexts = DGP_FALSE; 
    DgpImage_Packed src = (DgpImage_Packed)NULL; 
    size_t cxsize; 
	Dgp_UInt32 *retval; 
     
    cxsize = jb2ABContextsNeeded(params); 
 
    /* Make sure we have a context block */ 
    if (contexts == (Dgp_UInt8 *)NULL) { 
	contexts = (Dgp_UInt8 *)(a->calloc)(cxsize, 1); 
	if (contexts == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
	mycontexts = DGP_TRUE; 
	} 
 
    /* Make our DgpImage_Packed */ 
    DGP_ERR_PASS_CLEANUP (DgpImage_wrap(dgp_state, bitmap, width, height, Dgp_Int32sperline, &src)); 
 
    /* Call the inner routine to do the encoding */ 
    DGP_ERR_PASS_CLEANUP(jb2ArithEncodeImagePacked(src, params, contexts, enc, 
						   dgp_state)); 
 
    if (mycontexts) { 
	(a->free)((void *)contexts); 
	} 
 
    DgpImage_unwrap(src, &retval); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (mycontexts && (contexts != (Dgp_UInt8 *)NULL)) { 
	(a->free)((void *)contexts); 
	} 
 
    if (src != (DgpImage_Packed)NULL) { 
	DgpImage_unwrap(src, &retval); 
	} 
 
    ERR_RETURN; 
 
    } 
 
/* 
 * The same, but we start with contexts and a DgpImage_Packed. 
 */ 
ErrCode 
jb2ArithEncodeImagePacked(DgpImage_Packed src, arith_bitmap_params *params, 
			  Dgp_UInt8 *contexts, DgpEncode_MQEncoder *enc, 
			  Dgp_State *dgp_state)  
{ 
  DGP_PROLOGUE("jb2ArithEncodeImagePacked()", DGP_PERRSTATE); 
 
    switch(params->rtemplate) { 
      case 0: { 
	DGP_ERR_PASS(arithEncodeImageT16(src, params, contexts, enc,  
					 dgp_state)); 
	break; 
	} 
       
      case 1: { 
	DGP_ERR_PASS(arithEncodeImageT13(src, params, contexts, enc,  
					 dgp_state)); 
	break; 
	} 
       
      case 2: { 
	DGP_ERR_PASS(arithEncodeImageT10L3(src, params, contexts, enc,  
					   dgp_state)); 
	break; 
	} 
       
      case 3: { 
	DGP_ERR_PASS(arithEncodeImageT10L2(src, params, contexts, enc,  
					   dgp_state)); 
	break; 
	} 
      } 
    NORMAL_RETURN; 
    } 
 
/* 
 * Do the 16-pixel template encode 
 */ 
static ErrCode 
arithEncodeImageT16(DgpImage_Packed src, arith_bitmap_params *params, 
		    Dgp_UInt8 *contexts, DgpEncode_MQEncoder *enc, 
		    Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("arithEncodeImageT16()", DGP_PERRSTATE); 
 
    Dgp_UInt32 cx; 
    Dgp_Int32 x, y; 
    Dgp_Int32 minaty, minatx, maxatx; 
    Dgp_Int32 bit; 
    Dgp_Int32 width = DGPIMG_WIDTH(src), height = DGPIMG_HEIGHT(src); 
    Dgp_Bool lntp = DGP_TRUE; 
 
    /* 
     * The template looks like: 
     *   4XXX3 
     *  2XXXXX1 
     * XXXX? 
     * where 4321 are the "nominal" locations of the AT pixels. 
     * 
     * As in the decoder, we'll have a context value that looks like 
     * 4aaa 32bb bbb1 cccc 
     * where 4321 are the AT pixel values, aaa are the (LtoR) values from the 
     * top row of the template, bbbbb are the values from the middle row, and 
     * cccc are the values from the bottom row.  To go to the next pixel, we 
     * shift the context left one bit, AND it with 0 0110 0011 1100 1110 
     * (clearing out the bits that landed in the wrong place), and perform 
     * 7 bit-loads into bits 0 (bottom row), 4 (AT1), 5 (middle), 10 (AT2), 
     * 11 (AT3), 12 (top), 15 (AT4) 
     * 
     * We do want to optimise a couple of special cases: when the inner loop 
     * doesn't need conditionals, don't put any in. 
     */ 
 
    /* What's the smallest Y offset used? */ 
    minaty = MIN(params->aty[0], params->aty[1]); 
    minaty = MIN(minaty, params->aty[2]); 
    minaty = MIN(minaty, params->aty[3]); 
    minaty = MIN(minaty, -2);	/* -2 is always used */ 
 
    /* smallest and largest X offset used? */ 
    minatx = MIN(params->atx[0], params->atx[1]); 
    minatx = MIN(minatx, params->atx[2]); 
    minatx = MIN(minatx, params->atx[3]); 
    minatx = MIN(minatx, 1);	/* 1 is always used; -1 is special case */ 
    maxatx = MAX(params->atx[0], params->atx[1]); 
    maxatx = MAX(maxatx, params->atx[2]); 
    maxatx = MAX(maxatx, params->atx[3]); 
    maxatx = MAX(maxatx, 2);	/* 2 is always used */ 
 
    /* 
     * Given values in cx, x and y, write a bit, shift and mask CX, 
     * and put the bit in CX 
     */ 
#define	DOENCODE \
    do { \
	bit = DGPIMG_GETBIT01(src, x, y); \
	DGP_REPORTF1(DGP_MESSAGE,"%d", bit); \
	DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[cx], bit)), cErrIO_WriteError);\
	cx = ((cx << 1) & 0x63ce) | bit; \
	} while (0) 
 
    for (y = 0; y < height; y++) { 
	/* Is TP on? */ 
	if (params->tpon) { 
	    /* 
	     * We can check to see if this row is "typical" (i.e., equal to 
	     * the row above) 
	     */ 
	    Dgp_Bool diff = DGP_FALSE; 
	    Dgp_UInt32 *p, *q; 
 
	    if (y > 0) { 
		for (x = 0, p = DGPIMG_GETROW(src, y), q = DGPIMG_GETROW(src, y - 1); 
		     x < DGPIMG_GETROWWIDTH(src) / (int)sizeof(Dgp_UInt32); 
		     x++) { 
		    if (*p++ != *q++) { 
			diff = DGP_TRUE; 
			break; 
			} 
		    } 
		} 
	    else { 
		/* Implicit row above is all zero */ 
		for (x = 0, p = DGPIMG_GETROW(src, y); 
		     x < DGPIMG_GETROWWIDTH(src) / (int)sizeof(Dgp_UInt32); 
		     x++) { 
		    if (*p++ != 0ul) { 
			diff = DGP_TRUE; 
			break; 
			} 
		    } 
		} 
	    /* 
	     * Encode a bit using context 
	     *   10011 
	     *  0110010 
	     * 0101? 
	     * which linearises to 1001 1011 0010 0101. 
	     * The decoder will take the bit we send them and XOR it 
	     * with lntp; if the result is 0, they'll think the line is 
	     * typical; if the result is 1, they'll think the line is 
	     * non-typical. 
	     */ 
	    if (!diff) { 
		/* 
		 * "Typical" row: a copy of the previous one. 
		 * Encode lntp. 
		 */ 
		DGP_REPORTF1(DGP_MESSAGE,"TP%d\n", lntp); 
		DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[0x9b25], lntp)), 
			    cErrIO_WriteError); 
		lntp = DGP_FALSE; 
		continue;	/* Done with this line */ 
		} 
	    else { 
		/* Encode !lntp */ 
		DGP_REPORTF1(DGP_MESSAGE,"TP%d", !lntp); 
		DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[0x9b25], !lntp)), 
			    cErrIO_WriteError); 
		lntp = DGP_TRUE; 
		} 
	    } 
 
	/* Do the initial load */ 
	if (y >= 2) { 
	    /* 
	     * Normal case: load up 00aa 0000 bbb0 0000 
	     * where the "a" and "b" come from the lines above: 3 pixels 
	     * from above line, 2 from line above that. 
	     */ 
	    cx = ((*DGPIMG_GETROW(src, y - 1) >> (DGP_BITS_UINT32 - 3)) << 5) | 
		((*DGPIMG_GETROW(src, y - 2) >> (DGP_BITS_UINT32 - 2)) << 12); 
	    } 
	else if (y == 1) { 
	    /* Only get the stuff from the line above; rest zeros */ 
	    cx = ((*DGPIMG_GETROW(src, y - 1) >> (DGP_BITS_UINT32 - 3)) << 5); 
	    } 
	else { 
	    /* Only zeros */ 
	    cx = 0; 
	    } 
 
	if (y + minaty >= 0) { 
	    /* No conditionals needed for y */ 
	    for (x = 0; x < MIN(-minatx, width); x++) { 
		/* Need conditionals for x */ 
 
		/* Load the AT pixels */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				      y + params->aty[0]) << 4; 
		    } 
		if ((x + params->atx[1] >= 0) && (x + params->atx[1] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[1], 
				       y + params->aty[1]) << 10; 
		    } 
		if ((x + params->atx[2] >= 0) && (x + params->atx[2] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[2], 
				       y + params->aty[2]) << 11; 
		    } 
		if ((x + params->atx[3] >= 0) && (x + params->atx[3] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[3], 
				       y + params->aty[3]) << 15; 
		    } 
 
		/* Current row pixel is already loaded */ 
		/* Load previous row pixel */ 
		if (x + 2 < width) { 
		    cx |= DGPIMG_GETBIT01(src, x + 2, y - 1) << 5; 
		    } 
		/* Load two-back row pixel */ 
		if (x + 1 < width) { 
		    cx |= DGPIMG_GETBIT01(src, x + 1, y - 2) << 12; 
		    } 
 
		DOENCODE; 
		} 
 
	    for ( ; x + maxatx < width; x++) { 
		/* x + minatx >= 0; x + maxatx < width; no conditionals needed */ 
		cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				   y + params->aty[0]) << 4; 
		cx |= DGPIMG_GETBIT01(src, x + params->atx[1], 
				   y + params->aty[1]) << 10; 
		cx |= DGPIMG_GETBIT01(src, x + params->atx[2], 
				   y + params->aty[2]) << 11; 
		cx |= DGPIMG_GETBIT01(src, x + params->atx[3], 
				   y + params->aty[3]) << 15; 
		cx |= DGPIMG_GETBIT01(src, x + 2, y - 1) << 5; 
		cx |= DGPIMG_GETBIT01(src, x + 1, y - 2) << 12; 
		DOENCODE; 
		} 
	    for ( ; x < width; x++) { 
		/* Need conditionals again */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				       y + params->aty[0]) << 4; 
		    } 
		if ((x + params->atx[1] >= 0) && (x + params->atx[1] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[1], 
				       y + params->aty[1]) << 10; 
		    } 
		if ((x + params->atx[2] >= 0) && (x + params->atx[2] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[2], 
				       y + params->aty[2]) << 11; 
		    } 
		if ((x + params->atx[3] >= 0) && (x + params->atx[3] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[3], 
				       y + params->aty[3]) << 15; 
		    } 
 
		/* Current row pixel is already loaded */ 
		/* Load previous row pixel */ 
		if (x + 2 < width) { 
		    cx |= DGPIMG_GETBIT01(src, x + 2, y - 1) << 5; 
		    } 
		/* Load two-back row pixel */ 
		if (x + 1 < width) { 
		    cx |= DGPIMG_GETBIT01(src, x + 1, y - 2) << 12; 
		    } 
 
		DOENCODE; 
		} 
	    } 
	else { 
	    /* Need conditionals in y; might as well just do them everywhere */ 
	    for (x = 0; x < width; x++) { 
		if ((y + params->aty[0] >= 0) && (x + params->atx[0] >= 0) && 
		    (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				       y + params->aty[0]) << 4; 
		    } 
		if ((y + params->aty[1] >= 0) && (x + params->atx[1] >= 0) && 
		    (x + params->atx[1] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[1], 
				       y + params->aty[1]) << 10; 
		    } 
		if ((y + params->aty[2] >= 0) && (x + params->atx[2] >= 0) && 
		    (x + params->atx[2] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[2], 
				       y + params->aty[2]) << 11; 
		    } 
		if ((y + params->aty[3] >= 0) && (x + params->atx[3] >= 0) && 
		    (x + params->atx[3] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[3], 
				       y + params->aty[3]) << 15; 
		    } 
 
		/* Current row pixel is already loaded */ 
		/* Load previous row pixel */ 
		if ((y - 1 >= 0) && (x + 2 < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + 2, y - 1) << 5; 
		    } 
		/* Load two-back row pixel */ 
		if ((y - 2 >= 0) && (x + 1 < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + 1, y - 2) << 12; 
		    } 
 
		DOENCODE; 
		} 
	    } 
	DGP_REPORTF1(DGP_MESSAGE,"%c", '\n'); 
	} 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
#undef	DOENCODE 
 
/* 
 * Do the 13-pixel template encode. 
 */ 
static ErrCode 
arithEncodeImageT13(DgpImage_Packed src, arith_bitmap_params *params, 
		    Dgp_UInt8 *contexts, DgpEncode_MQEncoder *enc, 
		    Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("arithEncodeImageT13()", DGP_PERRSTATE); 
 
    Dgp_UInt32 cx; 
    Dgp_Int32 x, y; 
    Dgp_Int32 minaty, minatx, maxatx; 
    Dgp_Int32 bit; 
    Dgp_Int32 width = DGPIMG_WIDTH(src), height = DGPIMG_HEIGHT(src); 
    Dgp_Bool lntp = DGP_TRUE; 
 
    /* 
     * The template looks like: 
     *    XXXX 
     *   XXXXX1 
     *  XXX? 
     * 
     * We have a "context" value that looks like: 
     * a aaab bbbb 1ccc 
     * where 1 is the AT pixel value, aaaa are the (LtoR) values from the 
     * top row of the template, bbbbb are the values from the middle row, and 
     * ccc are the values from the bottom row.  To go to the next pixel, we 
     * shift the context left one bit, AND it with 01 1101 1110 0110 
     * (clearing out the bits that landed in the wrong place), and perform 
     * 4 bit-loads into bits 0 (bottom row), 3 (AT1), 4 (middle), 9 (top) 
     * We load from (after shift coordinates) (-1,0), (2,-1), (2,-2), 
     * (ATX1, ATY1) 
     */ 
 
    /* What's the smallest Y offset used? */ 
    minaty = MIN(params->aty[0], -2);	/* -2 is always used */ 
 
    /* smallest and largest X offset used? */ 
    minatx = MIN(params->atx[0], 2); /* 2 is always used; -1 is special case */ 
    maxatx = MAX(params->atx[0], 2); /* 2 is always used */ 
 
#define	DOENCODE \
    do { \
	bit = DGPIMG_GETBIT01(src, x, y); \
	DGP_REPORTF1(DGP_MESSAGE,"%d", bit); \
	DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[cx], bit)), cErrIO_WriteError);\
	cx = ((cx << 1) & 0x1de6) | bit; \
	} while (0) 
 
    for (y = 0; y < height; y++) { 
	/* Is TP on? */ 
	if (params->tpon) { 
	    /* 
	     * We can check to see if this row is "typical" (i.e., equal to 
	     * the row above) 
	     */ 
	    Dgp_Bool diff = DGP_FALSE; 
	    Dgp_UInt32 *p, *q; 
 
	    if (y > 0) { 
		for (x = 0, p = DGPIMG_GETROW(src, y), q = DGPIMG_GETROW(src, y - 1); 
		     x < DGPIMG_GETROWWIDTH(src) / (int)sizeof(Dgp_UInt32); 
		     x++) { 
		    if (*p++ != *q++) { 
			diff = DGP_TRUE; 
			break; 
			} 
		    } 
		} 
	    else { 
		/* Implicit row above is all zero */ 
		for (x = 0, p = DGPIMG_GETROW(src, y); 
		     x < DGPIMG_GETROWWIDTH(src) / (int)sizeof(Dgp_UInt32); 
		     x++) { 
		    if (*p++ != 0ul) { 
			diff = DGP_TRUE; 
			break; 
			} 
		    } 
		} 
	    /* 
	     * Encode a bit using context 
	     *   0011 
	     *  110010 
	     * 101? 
	     * which linearises to 0 0111 1001 0101 
	     * The decoder will take the bit we send them and XOR it 
	     * with lntp; if the result is 0, they'll think the line is 
	     * typical; if the result is 1, they'll think the line is 
	     * non-typical. 
	     */ 
	    if (!diff) { 
		/* 
		 * "Typical" row: a copy of the previous one. 
		 * Encode lntp. 
		 */ 
		DGP_REPORTF1(DGP_MESSAGE,"TP%d\n", lntp); 
		DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[0x0795], lntp)), 
			    cErrIO_WriteError); 
		lntp = DGP_FALSE; 
		continue;	/* Done with this line */ 
		} 
	    else { 
		/* Encode !lntp */ 
		DGP_REPORTF1(DGP_MESSAGE,"TP%d", !lntp); 
		DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[0x0795], !lntp)), 
			    cErrIO_WriteError); 
		lntp = DGP_TRUE; 
		} 
	    } 
 
	/* Do the initial load */ 
	if (y >= 2) { 
	    /* 
	     * Normal case: load up 0 aaa0 0bbb 0000 
	     * where the "a" and "b" come from the lines above: 3 pixels 
	     * from above line, 3 from line above that. 
	     */ 
	    cx = ((*DGPIMG_GETROW(src, y - 1) >> (DGP_BITS_UINT32 - 3)) << 4) | 
		((*DGPIMG_GETROW(src, y - 2) >> (DGP_BITS_UINT32 - 3)) << 9); 
	    } 
	else if (y == 1) { 
	    /* Only get the stuff from the line above; rest zeros */ 
	    cx = ((*DGPIMG_GETROW(src, y - 1) >> (DGP_BITS_UINT32 - 3)) << 4); 
	    } 
	else { 
	    /* Only zeros */ 
	    cx = 0; 
	    } 
 
	if (y + minaty >= 0) { 
	    /* No conditionals needed for y */ 
	    for (x = 0; x < MIN(-minatx, width); x++) { 
		/* Need conditionals for x */ 
 
		/* Load the AT pixel1 */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				      y + params->aty[0]) << 3; 
		    } 
 
		/* Current row pixel is already loaded */ 
		if (x + 2 < width) { 
		    /* Load previous row pixel */ 
		    cx |= DGPIMG_GETBIT01(src, x + 2, y - 1) << 4; 
		    /* Load two-back row pixel */ 
		    cx |= DGPIMG_GETBIT01(src, x + 2, y - 2) << 9; 
		    } 
 
		DOENCODE; 
		} 
 
	    for ( ; x + maxatx < width; x++) { 
		/* x + minatx >= 0; x + maxatx < width; no conditionals needed */ 
		cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				  y + params->aty[0]) << 3; 
		cx |= DGPIMG_GETBIT01(src, x + 2, y - 1) << 4; 
		cx |= DGPIMG_GETBIT01(src, x + 2, y - 2) << 9; 
		DOENCODE; 
		} 
	    for ( ; x < width; x++) { 
		/* Need conditionals again */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				      y + params->aty[0]) << 3; 
		    } 
		if (x + 2 < width) { 
		    cx |= DGPIMG_GETBIT01(src, x + 2, y - 1) << 4; 
		    cx |= DGPIMG_GETBIT01(src, x + 2, y - 2) << 9; 
		    } 
 
		DOENCODE; 
		} 
	    } 
	else { 
	    /* Need conditionals in y; might as well just do them everywhere */ 
	    for (x = 0; x < width; x++) { 
		if ((y + params->aty[0] >= 0) && (x + params->atx[0] >= 0) && 
		    (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				      y + params->aty[0]) << 3; 
		    } 
 
		/* Current row pixel is already loaded */ 
		/* Load previous row pixel */ 
		if ((y - 1 >= 0) && (x + 2 < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + 2, y - 1) << 4; 
		    } 
		/* Load two-back row pixel */ 
		if ((y - 2 >= 0) && (x + 2 < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + 2, y - 2) << 9; 
		    } 
 
		DOENCODE; 
		} 
	    } 
	DGP_REPORTF1(DGP_MESSAGE,"%c", '\n'); 
	} 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
#undef	DOENCODE 
 
/* 
 * Do the 10-pixel three-line template encode. 
 */ 
static ErrCode 
arithEncodeImageT10L3(DgpImage_Packed src, arith_bitmap_params *params, 
		      Dgp_UInt8 *contexts, DgpEncode_MQEncoder *enc, 
		    Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("arithEncodeImageT10L3()", DGP_PERRSTATE); 
 
    Dgp_UInt32 cx; 
    Dgp_Int32 x, y; 
    Dgp_Int32 minaty, minatx, maxatx; 
    Dgp_Int32 bit; 
    Dgp_Int32 width = DGPIMG_WIDTH(src), height = DGPIMG_HEIGHT(src); 
    Dgp_Bool lntp = DGP_TRUE; 
 
    /* 
     * The template looks like: 
     *   XXX 
     *  XXXX1 
     *  XX? 
     * 
     * We have a "context" value that looks like: 
     * aa abbb b1cc 
     * where 1 is the AT pixel value, aaa are the (LtoR) values from the 
     * top row of the template, bbbb are the values from the middle row, and 
     * cc are the values from the bottom row.  To go to the next pixel, we 
     * shift the context left one bit, AND it with 011 0111 0010 
     * (clearing out the bits that landed in the wrong place), and perform 
     * 4 bit-loads into bits 0 (bottom row), 2 (AT1), 3 (middle), 7 (top) 
     * We load from (after shift coordinates) (-1,0), (1,-1), (1,-2), 
     * (ATX1, ATY1) 
     */ 
 
    /* What's the smallest Y offset used? */ 
    minaty = MIN(params->aty[0], -2);	/* -2 is always used */ 
 
    /* smallest and largest X offset used? */ 
    minatx = MIN(params->atx[0], 1); /* 1 is always used; -1 is special case */ 
    maxatx = MAX(params->atx[0], 1); /* 1 is always used */ 
 
#define	DOENCODE \
    do { \
	bit = DGPIMG_GETBIT01(src, x, y); \
	DGP_REPORTF1(DGP_MESSAGE,"%d", bit); \
	DGP_REPORTF1(DGP_MESSAGE,"(%03lx)", cx); \
	DGP_REPORTF1(DGP_MESSAGE,"[%02x]", contexts[cx]); \
	DGP_REPORTF1(DGP_MESSAGE,"[%06lx ", enc->c); \
	DGP_REPORTF1(DGP_MESSAGE,"%04x]\n", enc->a); \
	DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[cx], bit)), cErrIO_WriteError);\
	cx = ((cx << 1) & 0x372) | bit; \
	} while (0) 
 
    for (y = 0; y < height; y++) { 
	/* Is TP on? */ 
	if (params->tpon) { 
	    /* 
	     * We can check to see if this row is "typical" (i.e., equal to 
	     * the row above) 
	     */ 
	    Dgp_Bool diff = DGP_FALSE; 
	    Dgp_UInt32 *p, *q; 
 
	    if (y > 0) { 
		for (x = 0, p = DGPIMG_GETROW(src, y), q = DGPIMG_GETROW(src, y - 1); 
		     x < DGPIMG_GETROWWIDTH(src) / (int)sizeof(Dgp_UInt32); 
		     x++) { 
		    if (*p++ != *q++) { 
			diff = DGP_TRUE; 
			break; 
			} 
		    } 
		} 
	    else { 
		/* Implicit row above is all zero */ 
		for (x = 0, p = DGPIMG_GETROW(src, y); 
		     x < DGPIMG_GETROWWIDTH(src) / (int)sizeof(Dgp_UInt32); 
		     x++) { 
		    if (*p++ != 0ul) { 
			diff = DGP_TRUE; 
			break; 
			} 
		    } 
		} 
	    /* 
	     * Encode a bit using context 
	     *  001 
	     * 11001 
	     * 01? 
	     * which linearises to 00 1110 0101 
	     * The decoder will take the bit we send them and XOR it 
	     * with lntp; if the result is 0, they'll think the line is 
	     * typical; if the result is 1, they'll think the line is 
	     * non-typical. 
	     */ 
	    if (!diff) { 
		/* 
		 * "Typical" row: a copy of the previous one. 
		 * Encode lntp. 
		 */ 
		DGP_REPORTF1(DGP_MESSAGE,"TP%d\n", lntp); 
		DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[0x0e5], lntp)), 
			    cErrIO_WriteError); 
		lntp = DGP_FALSE; 
		continue;	/* Done with this line */ 
		} 
	    else { 
		/* Encode !lntp */ 
		DGP_REPORTF1(DGP_MESSAGE,"TP%d", !lntp); 
		DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[0x0e5], !lntp)), 
			    cErrIO_WriteError); 
		lntp = DGP_TRUE; 
		} 
	    } 
 
	/* Do the initial load */ 
	if (y >= 2) { 
	    /* 
	     * Normal case: load up 0a a00b b000 
	     * where the "a" and "b" come from the lines above: 2 pixels 
	     * from above line, 2 from line above that. 
	     */ 
	    cx = ((*DGPIMG_GETROW(src, y - 1) >> (DGP_BITS_UINT32 - 2)) << 3) | 
		((*DGPIMG_GETROW(src, y - 2) >> (DGP_BITS_UINT32 - 2)) << 7); 
	    } 
	else if (y == 1) { 
	    /* Only get the stuff from the line above; rest zeros */ 
	    cx = ((*DGPIMG_GETROW(src, y - 1) >> (DGP_BITS_UINT32 - 2)) << 3); 
	    } 
	else { 
	    /* Only zeros */ 
	    cx = 0; 
	    } 
 
	if (y + minaty >= 0) { 
	    /* No conditionals needed for y */ 
	    for (x = 0; x < MIN(-minatx, width); x++) { 
		/* Need conditionals for x */ 
 
		/* Load the AT pixel1 */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				      y + params->aty[0]) << 2; 
		    } 
 
		/* Current row pixel is already loaded */ 
		if (x + 1 < width) { 
		    /* Load previous row pixel */ 
		    cx |= DGPIMG_GETBIT01(src, x + 1, y - 1) << 3; 
		    /* Load two-back row pixel */ 
		    cx |= DGPIMG_GETBIT01(src, x + 1, y - 2) << 7; 
		    } 
 
		DOENCODE; 
		} 
 
	    for ( ; x + maxatx < width; x++) { 
		/* x + minatx >= 0; x + maxatx < width; no conditionals needed */ 
		cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				  y + params->aty[0]) << 2; 
		cx |= DGPIMG_GETBIT01(src, x + 1, y - 1) << 3; 
		cx |= DGPIMG_GETBIT01(src, x + 1, y - 2) << 7; 
		DOENCODE; 
		} 
	    for ( ; x < width; x++) { 
		/* Need conditionals again */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				      y + params->aty[0]) << 2; 
		    } 
		if (x + 1 < width) { 
		    cx |= DGPIMG_GETBIT01(src, x + 1, y - 1) << 3; 
		    cx |= DGPIMG_GETBIT01(src, x + 1, y - 2) << 7; 
		    } 
 
		DOENCODE; 
		} 
	    } 
	else { 
	    /* Need conditionals in y; might as well just do them everywhere */ 
	    for (x = 0; x < width; x++) { 
		if ((y + params->aty[0] >= 0) && (x + params->atx[0] >= 0) && 
		    (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				      y + params->aty[0]) << 2; 
		    } 
 
		/* Current row pixel is already loaded */ 
		/* Load previous row pixel */ 
		if ((y - 1 >= 0) && (x + 1 < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + 1, y - 1) << 3; 
		    } 
		/* Load two-back row pixel */ 
		if ((y - 2 >= 0) && (x + 1 < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + 1, y - 2) << 7; 
		    } 
 
		DOENCODE; 
		} 
	    } 
	DGP_REPORTF1(DGP_MESSAGE,"%c", '\n'); 
	} 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
#undef	DOENCODE 
 
/* 
 * Do the 10-pixel 2-line template encode. 
 */ 
static ErrCode 
arithEncodeImageT10L2(DgpImage_Packed src, arith_bitmap_params *params, 
		      Dgp_UInt8 *contexts, DgpEncode_MQEncoder *enc, 
		    Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("arithEncodeImageT10L2()", DGP_PERRSTATE); 
 
    Dgp_UInt32 cx; 
    Dgp_Int32 x, y; 
    Dgp_Int32 minaty, minatx, maxatx; 
    Dgp_Int32 bit; 
    Dgp_Int32 width = DGPIMG_WIDTH(src), height = DGPIMG_HEIGHT(src); 
    Dgp_Bool lntp = DGP_TRUE; 
 
    /* 
     * The template looks like: 
     *   XXXXX1 
     *  XXXX? 
     * 
     * We have a "context" value that looks like: 
     * bb bbb1 cccc 
     * where 1 is the AT pixel value, bbbbb are the (LtoR) values from the 
     * top row of the template, and cccc are the values from the bottom row. 
     * To go to the next pixel, we 
     * shift the context left one bit, AND it with 011 1100 1110 
     * (clearing out the bits that landed in the wrong place), and perform 
     * 3 bit-loads into bits 0 (bottom row), 4 (AT1), 5 (top) 
     * We load from (after shift coordinates) (-1,0), (1,-1), (ATX1, ATY1) 
     */ 
 
    /* What's the smallest Y offset used? */ 
    minaty = MIN(params->aty[0], -1);	/* -1 is always used */ 
 
    /* smallest and largest X offset used? */ 
    minatx = MIN(params->atx[0], 1); /* 1 is always used; -1 is special case */ 
    maxatx = MAX(params->atx[0], 1); /* 1 is always used */ 
 
#define	DOENCODE \
    do { \
	bit = DGPIMG_GETBIT01(src, x, y); \
	DGP_REPORTF1(DGP_MESSAGE,"%d", bit); \
	DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[cx], bit)), cErrIO_WriteError);\
	cx = ((cx << 1) & 0x3ce) | bit; \
	} while (0) 
 
    for (y = 0; y < height; y++) { 
	/* Is TP on? */ 
	if (params->tpon) { 
	    /* 
	     * We can check to see if this row is "typical" (i.e., equal to 
	     * the row above) 
	     */ 
	    Dgp_Bool diff = DGP_FALSE; 
	    Dgp_UInt32 *p, *q; 
 
	    if (y > 0) { 
		for (x = 0, p = DGPIMG_GETROW(src, y), q = DGPIMG_GETROW(src, y - 1); 
		     x < DGPIMG_GETROWWIDTH(src) / (int)sizeof(Dgp_UInt32); 
		     x++) { 
		    if (*p++ != *q++) { 
			diff = DGP_TRUE; 
			break; 
			} 
		    } 
		} 
	    else { 
		/* Implicit row above is all zero */ 
		for (x = 0, p = DGPIMG_GETROW(src, y); 
		     x < DGPIMG_GETROWWIDTH(src) / (int)sizeof(Dgp_UInt32); 
		     x++) { 
		    if (*p++ != 0ul) { 
			diff = DGP_TRUE; 
			break; 
			} 
		    } 
		} 
	    /* 
	     * Encode a bit using context 
	     *  011001 
	     * 0101? 
	     * which linearises to 01 1001 0101 
	     * The decoder will take the bit we send them and XOR it 
	     * with lntp; if the result is 0, they'll think the line is 
	     * typical; if the result is 1, they'll think the line is 
	     * non-typical. 
	     */ 
	    if (!diff) { 
		/* 
		 * "Typical" row: a copy of the previous one. 
		 * Encode lntp. 
		 */ 
		DGP_REPORTF1(DGP_MESSAGE,"TP%d\n", lntp); 
		DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[0x195], lntp)), 
			    cErrIO_WriteError); 
		lntp = DGP_FALSE; 
		continue;	/* Done with this line */ 
		} 
	    else { 
		/* Encode !lntp */ 
		DGP_REPORTF1(DGP_MESSAGE,"TP%d", !lntp); 
		DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc, &contexts[0x195], !lntp)), 
			    cErrIO_WriteError); 
		lntp = DGP_TRUE; 
		} 
	    } 
 
	/* Do the initial load */ 
	if (y >= 1) { 
	    /* 
	     * Normal case: load up 00 0bb0 0000 
	     * where the "b" come from the line above: 2 pixels 
	     * from above line. 
	     */ 
	    cx = ((*DGPIMG_GETROW(src, y - 1) >> (DGP_BITS_UINT32 - 2)) << 5); 
	    } 
	else { 
	    /* Only zeros */ 
	    cx = 0; 
	    } 
 
	if (y + minaty >= 0) { 
	    /* No conditionals needed for y */ 
	    for (x = 0; x < MIN(-minatx, width); x++) { 
		/* Need conditionals for x */ 
 
		/* Load the AT pixel1 */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				      y + params->aty[0]) << 4; 
		    } 
 
		/* Current row pixel is already loaded */ 
		if (x + 1 < width) { 
		    /* Load previous row pixel */ 
		    cx |= DGPIMG_GETBIT01(src, x + 1, y - 1) << 5; 
		    } 
 
		DOENCODE; 
		} 
 
	    for ( ; x + maxatx < width; x++) { 
		/* x + minatx >= 0; x + maxatx < width; no conditionals needed */ 
		cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				  y + params->aty[0]) << 4; 
		cx |= DGPIMG_GETBIT01(src, x + 1, y - 1) << 5; 
		DOENCODE; 
		} 
	    for ( ; x < width; x++) { 
		/* Need conditionals again */ 
		if ((x + params->atx[0] >= 0) && (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				      y + params->aty[0]) << 4; 
		    } 
		if (x + 1 < width) { 
		    cx |= DGPIMG_GETBIT01(src, x + 1, y - 1) << 5; 
		    } 
 
		DOENCODE; 
		} 
	    } 
	else { 
	    /* Need conditionals in y; might as well just do them everywhere */ 
	    for (x = 0; x < width; x++) { 
		if ((y + params->aty[0] >= 0) && (x + params->atx[0] >= 0) && 
		    (x + params->atx[0] < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + params->atx[0], 
				      y + params->aty[0]) << 4; 
		    } 
 
		/* Current row pixel is already loaded */ 
		/* Load previous row pixel */ 
		if ((y - 1 >= 0) && (x + 1 < width)) { 
		    cx |= DGPIMG_GETBIT01(src, x + 1, y - 1) << 5; 
		    } 
 
		DOENCODE; 
		} 
	    } 
	DGP_REPORTF1(DGP_MESSAGE,"%c", '\n'); 
	} 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
#undef	DOENCODE 
 
#endif 
