/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/enchtdict.c,v 1.9 2003/01/27 21:48:34 awilliam Exp $ 
 * 
 */ 
 
#include "util.h" 
#include "jb2file.h" 
#include "segment.h" 
#include "enchtdict.h" 
#include "blit.h" 
#include "encgenreg.h" 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/enchtdict.c,v 1.9 2003/01/27 21:48:34 awilliam Exp $"); 
 
/* 
 * Encode the halftone dictionary, given the info stored in *segment. 
 * 
 * The fields in segment that must be filled in already are: 
 * - pagenum 
 * - data->htdict 
 * 
 * The remaining fields will be filled in as follows: 
 * - segnum will be allocated 
 * - datalen will be set based on the encoded segment's data length 
 * - nrefsegs will be set to 0 
 * - refsegs will be set to NULL 
 * - refretain will be set to NULL 
 * - segtype will be set to JB2SEGTYPE_HTDICT 
 * - retain will be set to DGP_TRUE 
 * - defer_nonretain will be set to DGP_FALSE 
 * 
 * The encoded segment header will be written out to 
 * jb2GetFileHeaderDataSink(file); the encoded segment data will be 
 * written out to jb2GetFileDataSink(file). 
 * 
 * This routine obeys the encoding parameters set in file's jb2FileParams. 
 */ 
ErrCode 
jb2EncodeHTDict(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2EncodeHTDict()", jb2_pErrState(file)); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
    DataSink bodyds; 
    Dgp_UInt8 cbuf, lbuf[4]; 
    jb2FileParams *params = jb2GetFileParams(file); 
    Dgp_UInt32 *collective = (Dgp_UInt32 *)NULL; 
    Dgp_Int32 collw, colllpl; 
    arith_bitmap_params abparams; 
    DgpEncode_MQEncoder mqenc; 
    Dgp_Int32 i; 
    size_t bytes; 
 
    /* Fill in the fields as promised */ 
    segment->segnum = jb2FileNextSegmentNumber(file); 
    segment->nrefsegs = 0; 
    segment->refsegs = (Dgp_Int32 *)NULL; 
    segment->refretain = (Dgp_Bool *)NULL; 
    segment->segtype = JB2SEGTYPE_HTDICT; 
    segment->retain = DGP_TRUE; 
    segment->defer_nonretain = DGP_FALSE; 
 
    /* All but the length is known - it's OK to encode the header */ 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeader(file, segment)); 
 
    /* Get the DataSink we're going to be writing to */ 
    bodyds = jb2GetFileDataSink(file); 
    /* 
     * and reset the count - all bytes written after this need to be 
     * accounted for in the segment data length 
     */ 
    dsResetCount(bodyds); 
 
    /* One-byte flags field */ 
    cbuf = 0; 
 
    /* Bit 0: HDMMR */ 
    if (!params->hdarith) { 
	cbuf |= 1; 
	} 
#ifdef	NOMQ 
    else { 
      FILEERROR(cErrCompressionNotSupported, "no MQ coder available"); 
	} 
#endif 
 
    /* Bits 1-2: HDTEMPLATE */ 
    if (params->hdarith) { 
	cbuf |= params->hdabtemplate << 1; 
	} 
 
    DSPUTC(bodyds, cbuf); 
 
    /* One-byte HDPW field */ 
    DSPUTC(bodyds, (Dgp_UInt8)segment->data.htdict.pat_w); 
 
    /* One-byte HDPH field */ 
    DSPUTC(bodyds, (Dgp_UInt8)segment->data.htdict.pat_h); 
 
    /* Four-byte GRAYMAX field */ 
    to4byte(lbuf, segment->data.htdict.npatterns - 1); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* Create the collective bitmap */ 
 
    /* Allocate a bitmap that's npatterns*pat_w wide, pat_h high */ 
    collw = (segment->data.htdict.pat_w * segment->data.htdict.npatterns); 
    colllpl = (collw + (DGP_BITS_UINT32 - 1)) / DGP_BITS_UINT32; 
    collective = (Dgp_UInt32 *)(a->calloc)(segment->data.htdict.pat_h, 
						     colllpl * sizeof(Dgp_UInt32)); 
    if (collective == (Dgp_UInt32 *)NULL) { 
      NOMEM; 
	} 
 
    /* Draw all the patterns into it */ 
    for (i = 0; i < segment->data.htdict.npatterns; i++) { 
	DgpBlat_blitBits(collective, colllpl, 
		 i * segment->data.htdict.pat_w, 0, 
		 (i + 1) * segment->data.htdict.pat_w, 
		 segment->data.htdict.pat_h, 
		 segment->data.htdict.patterns[i].bits, 
		 segment->data.htdict.patterns[i].longsperline, 
		 segment->data.htdict.pat_w, segment->data.htdict.pat_h, 
		 i * segment->data.htdict.pat_w, 0, 
		 jb2GetFileDgp_State(file)); 
	} 
 
    /* Encode the bitmap */ 
    if (params->hdarith) { 
	/* Make up the decoding spec */ 
#ifndef	NOMQ 
	abparams.rtemplate = (Dgp_UInt8)params->hdabtemplate; 
	abparams.tpon = DGP_FALSE; 
	/* First AT is always at the corresponding pixel in the prev pattern */ 
	abparams.atx[0] = -segment->data.htdict.pat_w; 
	abparams.aty[0] = 0; 
 
	if (abparams.rtemplate == 0) { 
	    /* Other ATs at default locs */ 
	    abparams.atx[1] = -3; 
	    abparams.aty[1] = -1; 
	    abparams.atx[2] = 2; 
	    abparams.aty[2] = -2; 
	    abparams.atx[3] = -2; 
	    abparams.aty[3] = -2; 
	    } 
	else { 
	    for (i = 1; i < 4; i++) { 
		abparams.atx[i] = abparams.aty[i] = 0; 
		} 
	    } 
 
	/* Set up the MQ encoder */ 
	dsInitDgpEncode_MQEncoder(bodyds, &mqenc); 
	DGP_ERR_PASS_CLEANUP(jb2ArithEncodeImage(collective, collw, segment->data.htdict.pat_h, 
				 colllpl, &abparams, (Dgp_UInt8 *)NULL, 
						 file, &mqenc)); 
 
	/* Finalise the MQ encoder */ 
	DGP_ERR_PASS_CLEANUP(dsFinalDgpEncode_MQEncoder(bodyds, &mqenc)); 
#endif 
	} 
    else { 
	/* MMR-encode the collective bitmap */ 
	DGP_ERR_PASS_CLEANUP(jb2MMREncodeDS(file, bodyds, 
			    collective, collw, segment->data.htdict.pat_h, 
					    colllpl, DGP_FALSE)); 
	} 
 
    (a->free)((void *)collective); 
    collective = (Dgp_UInt32 *)NULL; 
 
    /* Fix up the data length in the header */ 
    segment->datalen = dsCount(bodyds); 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeaderDataLength(file, segment)); 
 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
 
    if (collective != (Dgp_UInt32 *)NULL) { 
	(a->free)((void *)collective); 
	} 
 
    ERR_RETURN; 
    } 
