/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/enchtreg.c,v 1.9 2003/01/27 21:48:34 awilliam Exp $ 
 * 
 */ 
 
#include "util.h" 
#include "jb2file.h" 
#include "segment.h" 
#include "enchtreg.h" 
#include "encgenreg.h" 
#include "decgenreg.h" 
#include "segment_util.h" 
#include "symdec.h" 
#include "image_defs.h"
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/enchtreg.c,v 1.9 2003/01/27 21:48:34 awilliam Exp $"); 
 
/* 
 * Encode the symbol region, given the info stored in *segment and 
 * *refsegs.  There should be exactly one segment referenced, which 
 * should be a halftone dictionary segment. 
 * 
 * The fields in segment that must be filled in already are: 
 * - pagenum 
 * - nrefsegs 
 * - refsegs 
 * - refretain 
 * - data->htreg 
 * - reginfo (except for image fields) 
 * 
 * The remaining fields will be filled in as follows: 
 * - segnum will be allocated 
 * - datalen will be set based on the encoded segment's data length 
 * - retain will be set to DGP_FALSE 
 * - defer_nonretain will be set to DGP_FALSE 
 * 
 * The encoded segment header will be written out to 
 * jb2GetFileHeaderDataSink(file); the encoded segment data will be 
 * written out to jb2GetFileDataSink(file). 
 */ 
ErrCode 
jb2EncodeHTRegion(jb2File file, jb2Segment *segment, jb2Segment **refsegs) 
{ 
  DGP_PROLOGUE("jb2EncodeHTRegion()", jb2_pErrState(file)); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
    DataSink bodyds; 
    DgpEncode_MQEncoder mqenc; 
    jb2Segment *dict; 
    jb2HTRegSegment *htreg; 
    jb2FileParams *params = jb2GetFileParams(file); 
    DgpImage_Packed *planes = (DgpImage_Packed *)NULL; 
    Dgp_Int32 nplanes = 0; 
    Dgp_Int32 i, x, y, val; 
    Dgp_UInt8 *contexts = (Dgp_UInt8 *)NULL; 
    Dgp_UInt8 cbuf, lbuf[4]; 
    size_t bytes; 
 
    DGP_ASSERT ((segment->nrefsegs == 1) && 
	    (refsegs[0]->segtype == JB2SEGTYPE_HTDICT)); 
    DGP_ASSERT((segment->segtype == JB2SEGTYPE_HTREG_INT) || 
	   (segment->segtype == JB2SEGTYPE_HTREG_IMM) || 
	   (segment->segtype == JB2SEGTYPE_HTREG_IMMLL)); 
 
    /* Get a handle on some things */ 
    htreg = &segment->data.htreg; 
    dict = refsegs[0]; 
 
    /* Fill in the fields we promised to fill in */ 
    segment->segnum = jb2FileNextSegmentNumber(file); 
    segment->retain = DGP_FALSE; 
    segment->defer_nonretain = DGP_FALSE; 
 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeader(file, segment)); 
 
    /* Write out the segment data header */ 
    bodyds = jb2GetFileDataSink(file); 
    /* Reset the count on bodyds so we can get the data length */ 
    dsResetCount(bodyds); 
 
    /* First the region segment data header */ 
    DGP_ERR_PASS_CLEANUP(jb2EncodeRegSegment(file, bodyds, 
					     &segment->reginfo)); 
     
    /* Now the halftone region segment data header */ 
 
    /* One-byte halftone region segment flags */ 
    cbuf = 0; 
 
    /* Bit 0: HMMR */ 
    if (!params->srarith) { 
	cbuf |= 1; 
	} 
#ifdef	NOMQ 
    else { 
      FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
	} 
#endif 
 
    /* Bits 1-2: HTEMPLATE */ 
    if (params->srarith) { 
	cbuf |= params->hrabtemplate << 1; 
	} 
 
    /* Bit 3: HENABLESKIP */ 
    /* always 0 currently */ 
 
    /* Bits 4-6: HCOMBOP */ 
    cbuf |= htreg->int_combop << 4; 
 
    /* Bit 7: HDEFPIXEL */ 
    cbuf |= htreg->defpixel << 7; 
 
    DSPUTC(bodyds, cbuf); 
 
    /* HGW */ 
    to4byte(lbuf, DGPIMG_WIDTH(htreg->cells)); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* HGH */ 
    to4byte(lbuf, DGPIMG_HEIGHT(htreg->cells)); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* HGX */ 
    to4byte(lbuf, segment->data.htreg.hgx); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* HGY */ 
    to4byte(lbuf, segment->data.htreg.hgy); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* HRX */ 
    to2byte(lbuf, segment->data.htreg.hrx); 
    DSWRITE(bodyds, 2, lbuf, &bytes); 
     
    /* HRY */ 
    to2byte(lbuf, segment->data.htreg.hry); 
    DSWRITE(bodyds, 2, lbuf, &bytes); 
 
    /* Now on to the planes */ 
    nplanes = ceillog2(dict->data.htdict.npatterns); 
 
    /* Allocate all the planes */ 
    planes = (DgpImage_Packed *)(a->malloc)(nplanes * sizeof(*planes)); 
    if (planes == (DgpImage_Packed *)NULL) { 
      NOMEM; 
	} 
 
    for (i = 0; i < nplanes; i++) { 
	planes[i] = (DgpImage_Packed)NULL; 
	} 
 
    for (i = 0; i < nplanes; i++) { 
	DGP_ERR_PASS (DgpImage_new(jb2GetFileDgp_State(file),DGPIMAGE_PACKED, 
				       DGPIMG_WIDTH(htreg->cells), 
				       DGPIMG_HEIGHT(htreg->cells), (void**) &planes[i])); 
	} 
     
    /* Fill them */ 
    for (y = 0; y < DGPIMG_HEIGHT(htreg->cells); y++) { 
	for (x = 0; x < DGPIMG_WIDTH(htreg->cells); x++) { 
	    val = DGPIMG_REF(htreg->cells, x, y); 
	    /* Gray-code it */ 
	    val ^= val >> 1; 
 
	    for (i = 0; i < nplanes; i++) { 
		DGPIMG_ORBIT(planes[i], x, y, (val >> i) & 1); 
		} 
	    } 
	} 
 
    /* Encode them */ 
    if (params->srarith) { 
#ifndef	NOMQ 
	arith_bitmap_params abparams; 
 
	/* Make up an DgpEncode_MQEncoder */ 
	dsInitDgpEncode_MQEncoder(bodyds, &mqenc); 
 
	/* Put together the arith bitmap params */ 
	abparams.rtemplate = (Dgp_UInt8)params->hrabtemplate; 
	abparams.tpon = DGP_FALSE; 
 
	if (params->hrabtemplate == 0) { 
	    abparams.atx[0] = 3; 
	    abparams.aty[0] = -1; 
	    abparams.atx[1] = -3; 
	    abparams.aty[1] = -1; 
	    abparams.atx[2] = 2; 
	    abparams.aty[2] = -2; 
	    abparams.atx[3] = -2; 
	    abparams.aty[3] = -2; 
	    } 
	else { 
	    if (params->hrabtemplate == 1) { 
		abparams.atx[0] = 3; 
		abparams.aty[0] = -1; 
		} 
	    else { 
		abparams.atx[0] = 2; 
		abparams.aty[0] = -1; 
		} 
	    for (i = 1; i < 4; i++) { 
		abparams.atx[i] = abparams.aty[i] = 0; 
		} 
	    } 
 
	/* Make up some contexts */ 
	contexts = (Dgp_UInt8 *)(a->calloc)(jb2ABContextsNeeded(&abparams), 
						       1); 
	if (contexts == (Dgp_UInt8 *)NULL) { 
	    NOMEM; 
	    } 
 
	for (i = nplanes - 1; i >= 0; i--) { 
	    DGP_ERR_PASS_CLEANUP(jb2ArithEncodeImage(DGPIMG_GETROW(planes[i], 0), 
				     DGPIMG_WIDTH(planes[i]), 
				     DGPIMG_HEIGHT(planes[i]), 
				     DGPIMG_GETROWWIDTH(planes[i]) / (int)sizeof(Dgp_UInt32), 
			       &abparams, contexts, file, &mqenc)); 
	    } 
 
	/* Done with the contexts */ 
	(a->free)((void *)contexts); 
	contexts = (Dgp_UInt8 *)NULL; 
 
	/* and the DgpEncode_MQEncoder */ 
	DGP_ERR_PASS_CLEANUP(dsFinalDgpEncode_MQEncoder(bodyds, &mqenc)); 
#else 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
	} 
    else { 
	/* MMR encode the planes.  EOFB required. */ 
	for (i = nplanes - 1; i >= 0; i--) { 
	    DGP_ERR_PASS_CLEANUP(jb2MMREncodeDS(file, bodyds, 
				DGPIMG_GETROW(planes[i], 0), 
				DGPIMG_WIDTH(planes[i]), 
				DGPIMG_HEIGHT(planes[i]), 
				DGPIMG_GETROWWIDTH(planes[i]) / (int)sizeof(Dgp_UInt32), 
						DGP_TRUE)); 
	    } 
	} 
 
    /* Fix up the data length */ 
    segment->datalen = dsCount(bodyds); 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeaderDataLength(file, segment)); 
 
    /* Free the planes */ 
    for (i = 0; i < nplanes; i++) { 
	DgpImage_free(planes[i]); 
	} 
    (a->free)(planes); 
 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
 
    if (planes != (DgpImage_Packed *)NULL) { 
	for (i = 0; i < nplanes; i++) { 
	    if (planes[i] != (DgpImage_Packed)NULL) { 
		DgpImage_free(planes[i]); 
		} 
	    } 
	(a->free)((void *)planes); 
	} 
 
    if (contexts != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)contexts); 
	} 
 
    ERR_RETURN; 
    } 
 
