/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/enchufftable.c,v 1.11 2003/01/27 21:48:34 awilliam Exp $ 
 * 
 */ 
 
 
#include "util.h" 
#include "enchufftable.h" 
#include "datasink.h" 
#include "segment.h" 
#include "jb2file.h" 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/enchufftable.c,v 1.11 2003/01/27 21:48:34 awilliam Exp $"); 
 
/* 
 * Encode the Huffman table whose specification is in *segment. 
 * The fields in segment that must be filled in already are: 
 * - pagenum 
 * - data->table 
 * 
 * The remaining fields will be filled in as follows: 
 * - segnum will be allocated 
 * - datalen will be set based on the encoded segment's data length 
 * - nrefsegs will be set to 0 
 * - refsegs will be set to NULL 
 * - refretain will be set to NULL 
 * - segtype will be set to JB2SEGTYPE_TABLE 
 * - retain will be set to DGP_TRUE 
 * - defer_nonretain will be set to DGP_FALSE 
 * 
 * The encoded segment header will be written out to 
 * jb2GetFileHeaderDataSink(file); the encoded segment data will be 
 * written out to jb2GetFileDataSink(file). 
 */ 
ErrCode 
jb2EncodeHuffTable(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2EncodeHuffTable()", jb2_pErrState(file)); 
 
    DataSink bodyds; 
    jbitfile bf = NULLJBITFILE, tmpbf; 
    Dgp_Int32 i; 
    hufftable_spec *spec; 
    Dgp_Int32 nnormallines; 
    Dgp_Int32 maxrange, maxprefix; 
    Dgp_Int32 htps, htrs; 
    Dgp_UInt8 cbuf; 
    Dgp_UInt8 lbuf[4]; 
    size_t bytes; 
 
    /* Fill in the fields as promised */ 
    segment->segnum = jb2FileNextSegmentNumber(file); 
    segment->nrefsegs = 0; 
    segment->refsegs = (Dgp_Int32 *)NULL; 
    segment->refretain = (Dgp_Bool *)NULL; 
    segment->segtype = JB2SEGTYPE_TABLE; 
    segment->retain = DGP_TRUE; 
    segment->defer_nonretain = DGP_FALSE; 
 
    /* All but the length is known - it's OK to encode the header */ 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeader(file, segment)); 
 
    /* Now we can start writing. */ 
    spec = &segment->data.table.table; 
 
    /* Get the DataSink we're going to be writing to */ 
    bodyds = jb2GetFileDataSink(file); 
    /* 
     * and reset the count - all bytes written after this need to be 
     * accounted for in the segment data length 
     */ 
    dsResetCount(bodyds); 
 
    /* Find the maximum range length and prefix length */ 
 
    /* Scan only the normal lines for range */ 
    nnormallines = spec->nlines - 
	(spec->lower_valid ? 1 : 0) - 
	(spec->upper_valid ? 1 : 0) - 
	(spec->oob_valid ? 1 : 0); 
 
    maxrange = maxprefix = 0; 
    for (i = 0; i < nnormallines; i++) { 
	maxprefix = MAX(spec->lines[i].preflen, maxprefix); 
	maxrange = MAX(spec->lines[i].rangelen, maxrange); 
	} 
    /* prefix length needs to include all the lines */ 
    for (; i < spec->nlines; i++) { 
	maxprefix = MAX(spec->lines[i].preflen, maxprefix); 
	} 
 
    /* Now figure out the size required to encode a value 0...maxprefix */ 
    for (htps = 1, maxprefix >>= 1; maxprefix > 0; maxprefix >>= 1, htps++) { 
	} 
    /* ditto 0...maxrange */ 
    for (htrs = 1, maxrange >>= 1; maxrange > 0; maxrange >>= 1, htrs++) { 
	} 
    /* Note that htps and htrs are always >= 1 */ 
 
    /* One byte flags field */ 
     
    /* Bit 0: HTOOB */ 
    cbuf = spec->oob_valid ? 1 : 0; 
 
    /* Bits 1-3: HTPS-1 */ 
    cbuf |= (htps - 1) << 1; 
 
    /* Bits 4-6: HTRS - 1 */ 
    cbuf |= (htrs - 1) << 4; 
 
    DSPUTC(bodyds, cbuf); 
#ifdef	VALPRINT 
    fprintf(stderr, "HTFLAGS %02x\n", cbuf); 
#endif 
 
    if (nnormallines > 0) { 
	/* Four byte lowest value - look at the first line */ 
	to4byte(lbuf, spec->lines[0].base); 
	DSWRITE(bodyds, 4, lbuf, &bytes); 
#ifdef	VALPRINT 
	fprintf(stderr, "HTLOW %ld\n", spec->lines[0].base); 
#endif 
 
	/* Four byte highest value plus one - look at the last normal line */ 
	to4byte(lbuf, spec->lines[nnormallines - 1].base + 
		(1l << spec->lines[nnormallines - 1].rangelen)); 
	DSWRITE(bodyds, 4, lbuf, &bytes); 
#ifdef	VALPRINT 
	fprintf(stderr, "HTHIGH %ld\n", spec->lines[nnormallines - 1].base + 
		(1l << spec->lines[nnormallines - 1].rangelen)); 
#endif 
	} 
    else { 
	/* In theory, there could be a table with no normal lines. */ 
	to4byte(lbuf, 0); 
	DSWRITE(bodyds, 4, lbuf, &bytes); 
	DSWRITE(bodyds, 4, lbuf, &bytes); 
#ifdef	VALPRINT 
	fprintf(stderr, "HTLOW 0\n"); 
	fprintf(stderr, "HTHIGH 0\n"); 
#endif 
	} 
 
    /* Now write the normal lines */ 
 
    /* Create a jbitfile to stuff bits into */ 
    DGP_ERR_PASS_CLEANUP(jbfOpenFromDataSink(bodyds, &bf)); 
 
    for (i = 0; i < nnormallines; i++) { 
	/* HTPS bits of prefix length */ 
	DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, spec->lines[i].preflen, htps)); 
 
	/* HTRS bits of range length */ 
	DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, spec->lines[i].rangelen, htrs)); 
#ifdef	VALPRINT 
	fprintf(stderr, "HTLINE %d %d\n", spec->lines[i].preflen, 
		spec->lines[i].rangelen); 
#endif 
	} 
 
    /* Lower range line, if any */ 
    if (spec->lower_valid) { 
	/* HTPS bits of prefix from the next table line */ 
	DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, spec->lines[i].preflen, htps)); 
#ifdef	VALPRINT 
	fprintf(stderr, "HTLOWLINE %d\n", spec->lines[i].preflen); 
#endif 
	i++; 
	} 
    else { 
	/* Write HTPS bits of 0 */ 
	DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, 0, htps)); 
#ifdef	VALPRINT 
	fprintf(stderr, "HTLOWLINE %d\n", 0); 
#endif 
	} 
 
    /* Upper range line, if any */ 
    if (spec->upper_valid) { 
	/* HTPS bits of prefix from the next table line */ 
	DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, spec->lines[i].preflen, htps)); 
#ifdef	VALPRINT 
	fprintf(stderr, "HTHIGHLINE %d\n", spec->lines[i].preflen); 
#endif 
	i++; 
	} 
    else { 
	/* Write HTPS bits of 0 */ 
	DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, 0, htps)); 
#ifdef	VALPRINT 
	fprintf(stderr, "HTHIGHLINE %d\n", 0); 
#endif 
	} 
 
    /* OOB line, if any */ 
    if (spec->oob_valid) { 
	/* HTPS bits of prefix from the next table line */ 
	DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, spec->lines[i].preflen, htps)); 
#ifdef	VALPRINT 
	fprintf(stderr, "HTOOBLINE %d\n", spec->lines[i].preflen); 
#endif 
	i++; 
	} 
 
    /* Done writing.  Close the jbitfile */ 
    tmpbf = bf; 
    bf = NULLJBITFILE; 
    DGP_ERR_PASS_CLEANUP(jbfCloseFromDataSink(tmpbf)); 
 
    /* 
     * Check the DataSink's running count to find how many bytes that 
     * all took, and write that to the header field 
     */ 
    segment->datalen = dsCount(bodyds); 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeaderDataLength(file, segment)); 
 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
 
    if (bf != NULLJBITFILE) { 
	(void)jbfCloseFromDataSink(bf); 
	} 
 
    ERR_RETURN; 
    } 
