/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/encint.c,v 1.16 2003/01/27 21:48:34 awilliam Exp $ 
 * 
 */ 
 
/* 
 * Encoding integers. 
 */ 
 
#include "encint.h" 
#include "hufftable.h" 
#include <stdlib.h> 
#include "mqcoder.h" 
#include <string.h> 
#include "PsgFix.h"
 
#define DGP_ENC DGP_FALSE

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/encint.c,v 1.16 2003/01/27 21:48:34 awilliam Exp $"); 
 
 
struct _eiEncoder { 
    Dgp_Bool mqcoder; 
    hufftable_spec hufftable; 
    table_line *lower_line; 
    table_line *upper_line; 
    table_line *oob_line; 
    Dgp_Int32 lower_linenum, upper_linenum, oob_linenum; 
    Dgp_Int32 nnormallines; 
    Dgp_Int32 lownormal; 
    Dgp_Int32 highnormal; 
    jbitfile f; 
    DgpEncode_MQEncoder *mqenc; 
    Dgp_UInt8 *contexts; 
    Dgp_State *dgp_state; 
    }; 
 
/* 
 * It's legal to pass in NULL as jbitfile, as Dgp_Int32 as all you do 
 * with this eiEncoder is call eiCount and eiFreeEncoder. 
 */ 
ErrCode 
eiNewEncoder(hufftable_spec *spec, jbitfile file, Dgp_State *dgp_state, 
	     eiEncoder *_newe) 
{ 
  DGP_PROLOGUE("eiNewEncoder()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    eiEncoder newe = (eiEncoder)(a->malloc)(sizeof(*newe)); 
    table_line *lines = (table_line *)NULL; 
    Dgp_Int32 i; 
    Dgp_Int32 v; 
 
    *_newe = NULLEIENCODER; 
    DGP_CHECK_ALLOC_CLEANUP(newe); 
 
    newe->dgp_state = dgp_state; 
    newe->hufftable = *spec; 
    newe->mqcoder = DGP_FALSE; 
    /* Copy the lines array */ 
    lines = (table_line *)(a->malloc)(spec->nlines * sizeof(*lines)); 
    DGP_CHECK_ALLOC_CLEANUP(lines); 
 
    newe->hufftable.lines = lines; 
    (void)memcpy((void *)lines, (const void *)spec->lines, 
		 spec->nlines * sizeof(*lines)); 
 
    newe->f = file; 
 
    /* Set up some values for easy search */ 
 
    /* How many normal table lines? */ 
    newe->nnormallines = spec->nlines - 
	(spec->lower_valid ? 1 : 0) - 
	(spec->upper_valid ? 1 : 0) - 
	(spec->oob_valid ? 1 : 0); 
 
    i = newe->nnormallines; 
 
    /* Set up the special templates */ 
    if (spec->lower_valid) { 
	newe->lower_linenum = i; 
	newe->lower_line = &lines[i++]; 
	} 
    else { 
	newe->lower_line = (table_line *)NULL; 
	} 
 
    if (spec->upper_valid) { 
	newe->upper_linenum = i; 
	newe->upper_line = &lines[i++]; 
	} 
    else { 
	newe->upper_line = (table_line *)NULL; 
	} 
 
    if (spec->oob_valid) { 
	newe->oob_linenum = i; 
	newe->oob_line = &lines[i++]; 
	} 
    else { 
	newe->oob_line = (table_line *)NULL; 
	} 
 
    /* Now find out what the range of normal values is */ 
    v = newe->lownormal = lines[0].base; 
    for (i = 0; i < newe->nnormallines; i++) { 
	v += 1l << lines[i].rangelen; 
	} 
    newe->highnormal = v - 1; 
 
    *_newe = newe; 
    newe = NULLEIENCODER; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    if (newe != NULLEIENCODER) { 
	(a->free)((void *)newe); 
	} 
 
    if (lines != (table_line *)NULL) { 
	(a->free)((void *)lines); 
	} 
 
    ERR_RETURN; 
    } 
 
#ifndef	NOMQ 
/* 
 * Create an MQ-based integer encoder from the given DgpEncode_MQEncoder and set of 
 * contexts.  The context block should be 512 bytes Dgp_Int32; it and the 
 * DgpEncode_MQEncoder must live at least as Dgp_Int32 as this encoder. 
 */ 
ErrCode 
eiNewEncoderMQ(DgpEncode_MQEncoder *mqenc, Dgp_UInt8 *contexts, Dgp_State *dgp_state, 
	       eiEncoder *_newe) 
{ 
  DGP_PROLOGUE("eiNewEncoderMQ()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    eiEncoder newe = (eiEncoder)(a->malloc)(sizeof(*newe)); 
 
    *_newe = NULLEIENCODER; 
    DGP_CHECK_ALLOC_CLEANUP(newe); 
 
    newe->dgp_state = dgp_state; 
    newe->mqcoder = DGP_TRUE; 
    newe->mqenc = mqenc; 
    newe->contexts = contexts; 
 
    *_newe = newe; 
    newe = NULLEIENCODER; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    if (newe != NULLEIENCODER) { 
	(a->free)((void *)newe); 
	} 
 
    ERR_RETURN; 
    } 
#endif 
 
/* 
 * If huff is DGP_TRUE, create an integer encoder based on spec and file. 
 * If huff is DGP_FALSE, create an integer encoder based on mqenc and 
 * contexts. Clear contexts if huff is DGP_FALSE. 
 * If huff is DGP_TRUE, mqenc and contexts need not be valid; if huff is DGP_FALSE, 
 * spec and file need not be valid. 
 */ 
ErrCode 
eiNewEncoderEither(Dgp_Bool huff, 
		   hufftable_spec *spec, jbitfile file, 
		   DgpEncode_MQEncoder *mqenc, Dgp_UInt8 *contexts, 
		   Dgp_State *dgp_state, eiEncoder *newe) 
{ 
  DGP_PROLOGUE("eiNewEncoderEither()", DGP_PERRSTATE); 
 
    if (huff) { 
      DGP_ERR_PASS(eiNewEncoder(spec, file, dgp_state, newe)); 
	} 
    else { 
#ifndef	NOMQ 
	memset(contexts, 0, EICONTEXTSIZE); 
	DGP_ERR_PASS(eiNewEncoderMQ(mqenc, contexts, dgp_state, newe)); 
#else 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
	} 
	 
    NORMAL_RETURN; 
 
#ifdef	NOMQ 
  ERR_CLEANUP: 
    ERR_RETURN; 
#endif 
    } 
 
 
ErrCode 
eiEncode(eiEncoder enc, Dgp_Int32 val, Dgp_Bool oob) 
{ 
  DGP_PROLOGUE("eiEncode()", enc->DGP_PERRSTATE); 
 
    if (enc->mqcoder) { 
#ifndef	NOMQ 
	int cx = 1; 
	int nbits; 
	int i; 
 
#define	ENCODE(bit)	\
	do { \
	    /* DGP_REPORTF1("%d", bit); */ \
	    DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc->mqenc, &enc->contexts[cx], \
					    (bit))), cErrIO_WriteError); \
	    cx = ((cx & 0x100) | (cx << 1) | (bit)) & 0x1ff; \
	    } while(0) 
	/* 
	 * FlowDgp_Int8t is: 
	 * Treat OOB as -0. 
	 * Encode sign bit. 
	 * If ABS(val) < 4, encode 0vv (vv = ABS(val)) 
	 * If ABS(val) < 20, encode 10vvvv (vvvv = ABS(val) - 4) 
	 * If ABS(val) < 84, encode 110vvvvvv (vvvvvv = ABS(val) - 20) 
	 * If ABS(val) < 340, encode 1110vvvvvvvv (vvvvvvvv = ABS(val) - 84) 
	 * If ABS(val) < 4436, encode 11110vvvvvvvvvvvv 
	 *	(vvvvvvvvvvvv = ABS(val) - 340) 
	 * Encode 11111(32v) (32v = ABS(val) - 4436) 
	 */ 
	if (oob) { 
	    /* Encode 1000 - negative zero */ 
	    ENCODE(1); 
	    ENCODE(0); 
	    ENCODE(0); 
	    ENCODE(0); 
	    } 
	else { 
	    /* Sign bit */ 
	    if (val < 0) { 
		ENCODE(1); 
		val = -val; 
		} 
	    else { 
		ENCODE(0); 
		} 
 
	    if (val < 4) { 
		ENCODE(0); 
		nbits = 2; 
		} 
	    else { 
		ENCODE(1); 
		if (val < 20) { 
		    ENCODE(0); 
		    val -= 4; 
		    nbits = 4; 
		    } 
		else { 
		    ENCODE(1); 
		    if (val < 84) { 
			ENCODE(0); 
			val -= 20; 
			nbits = 6; 
			} 
		    else { 
			ENCODE(1); 
			if (val < 340) { 
			    ENCODE(0); 
			    val -= 84; 
			    nbits = 8; 
			    } 
			else { 
			    ENCODE(1); 
			    if (val < 4436) { 
				ENCODE(0); 
				val -= 340; 
				nbits = 12; 
				} 
			    else { 
				ENCODE(1); 
				val -= 4436; 
				nbits = 32; 
				} 
			    } 
			} 
		    } 
		} 
 
	    /* Encode the mantissa */ 
	    for (i = 0; i < nbits; i++) { 
		if (val & (1 << (nbits - 1 - i))) { 
		    ENCODE(1); 
		    } 
		else { 
		    ENCODE(0); 
		    } 
		} 
	    } 
#endif 
	NORMAL_RETURN; 
	} 
    else { 
	/* 
	 * In the future, the eiEncoder should be preprocessed to make this 
	 * search easy. 
	 */ 
	table_line *line = (table_line *)NULL; 
	int i; 
 
	if (oob) { 
	    DGP_ASSERT(enc->hufftable.oob_valid); 
	    line = enc->oob_line; 
	    /* DGP_REPORTF1("jht %lx", line->prefix); 
	    DGP_REPORTF1("(%d)\n", line->preflen);  */
	    DGP_ERR_PASS(jbfWriteBits(enc->f, line->prefix, line->preflen)); 
	    NORMAL_RETURN; 
	    } 
 
	if (val < enc->lownormal) { 
	    DGP_ASSERT(enc->hufftable.lower_valid); 
	    line = enc->lower_line; 
	    DGP_ASSERT(line->base == enc->lownormal - 1); 
	    /* DGP_REPORTF1("jht %lx", line->prefix); 
	    DGP_REPORTF1("(%d)\n", line->preflen);  */
	    DGP_ERR_PASS(jbfWriteBits(enc->f, line->prefix, line->preflen)); 
	    DGP_ERR_PASS(jbfWriteBits(enc->f, line->base - val, 32)); 
	    NORMAL_RETURN; 
	    } 
	else if (val > enc->highnormal) { 
	    DGP_ASSERT(enc->hufftable.upper_valid); 
	    line = enc->upper_line; 
	    DGP_ASSERT(line->base == enc->highnormal + 1); 
	    /* DGP_REPORTF1("jht %lx", line->prefix); 
	    DGP_REPORTF1("(%d)\n", line->preflen);  */
	    DGP_ERR_PASS(jbfWriteBits(enc->f, line->prefix, line->preflen)); 
	    DGP_ERR_PASS(jbfWriteBits(enc->f, val - line->base, 32)); 
	    NORMAL_RETURN; 
	    } 
 
	for (i = 0; i < enc->nnormallines; i++) { 
	    if ((val >= enc->hufftable.lines[i].base) && 
		(val < enc->hufftable.lines[i].base + 
		 (1l << enc->hufftable.lines[i].rangelen))) { 
		/* It's this line */ 
		line = &enc->hufftable.lines[i]; 
		break; 
		} 
	    } 
	if (i == enc->nnormallines) { 
	    DGP_REPORTF0(DGP_ENC, "bad value to encode!"); 
	    } 
	 
	/* DGP_REPORTF1("jht %lx", line->prefix); 
	DGP_REPORTF1("(%d)\n", line->preflen);  */
	DGP_ERR_PASS(jbfWriteBits(enc->f, line->prefix, line->preflen)); 
	DGP_ERR_PASS(jbfWriteBits(enc->f, val - line->base, line->rangelen)); 
	NORMAL_RETURN; 
	} 
 
 ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
void 
eiFreeEncoder(eiEncoder enc) 
{ 
  Dgp_Free tmp_free = enc->DGP_FREE; 
 
    if (!enc->mqcoder) { 
	(tmp_free)((void *)enc->hufftable.lines); 
	} 
    (tmp_free)((void *)enc); 
    } 
 
/* 
 * Figure out which table line this goes to and increment the 
 * count for that line.  Used when choosing/tuning Huffman tables. 
 */ 
void 
eiCount(eiEncoder enc, Dgp_Int32 val, Dgp_Bool oob, Dgp_Int32 *counts) 
{ 
  DGP_VOID_PROLOGUE("eiCount()", enc->DGP_PERRSTATE); 
    Dgp_Int32 i; 
 
    if (oob) { 
	DGP_VOID_ASSERT(enc->hufftable.oob_valid); 
	counts[enc->oob_linenum]++; 
	return; 
	} 
    else if (val < enc->lownormal) { 
	DGP_VOID_ASSERT(enc->hufftable.lower_valid); 
	counts[enc->lower_linenum]++; 
	return; 
	} 
    else if (val > enc->highnormal) { 
	DGP_VOID_ASSERT(enc->hufftable.upper_valid); 
	counts[enc->upper_linenum]++; 
	return; 
	} 
 
    for (i = 0; i < enc->nnormallines; i++) { 
	if ((val >= enc->hufftable.lines[i].base) && 
	    (val < enc->hufftable.lines[i].base + 
	     (1l << enc->hufftable.lines[i].rangelen))) { 
	    /* It's this line */ 
	    counts[i]++; 
	    break; 
	    } 
	} 
    if (i == enc->nnormallines) { 
	DGP_REPORTF0(DGP_ENC, "bad value to encode!"); 
	} 
    } 
 
 
 
 
 
 
 
