/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/encpageinfo.c,v 1.9 2003/02/10 22:53:26 mkrocken Exp $ 
 * 
 */ 
 
#include "util.h" 
#include "datasink.h" 
#include "segment.h" 
#include "jb2file.h" 
#include "encpageinfo.h" 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/encpageinfo.c,v 1.9 2003/02/10 22:53:26 mkrocken Exp $"); 
 
/* 
 * Encode the page information segment whose specification is in *segment. 
 * The fields in segment that must be filled in already are: 
 * - pagenum (must be > 0) 
 * - data->page 
 * 
 * The remaining fields will be filled in as follows: 
 * - segnum will be allocated 
 * - datalen will be set based on the encoded segment's data length 
 * - nrefsegs will be set to 0 
 * - refsegs will be set to NULL 
 * - refretain will be set to NULL 
 * - segtype will be set to JB2SEGTYPE_PAGEINFO 
 * - retain will be set to DGP_FALSE 
 * - defer_nonretain will be set to DGP_FALSE 
 * 
 * The encoded segment header will be written out to 
 * jb2GetFileHeaderDataSink(file); the encoded segment data will be 
 * written out to jb2GetFileDataSink(file). 
 */ 
ErrCode 
jb2EncodePageInfo(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2EncodePageInfo()", jb2_pErrState(file)); 
 
    Dgp_UInt8 lbuf[4]; 
    Dgp_UInt8 cbuf; 
    Dgp_UInt16 sbuf; 
    DataSink bodyds; 
    size_t bytes; 
 
    /* Fill in the fields as promised */ 
    segment->segnum = jb2FileNextSegmentNumber(file); 
    segment->nrefsegs = 0; 
    segment->refsegs = (Dgp_Int32 *)NULL; 
    segment->refretain = (Dgp_Bool *)NULL; 
    segment->segtype = JB2SEGTYPE_PAGEINFO; 
    segment->retain = DGP_FALSE; 
    segment->defer_nonretain = DGP_FALSE; 
 
    DGP_ASSERT(segment->pagenum > 0); 
 
    /* All but the length is known - it's OK to encode the header */ 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeader(file, segment)); 
 
    /* Get the DataSink we're going to be writing to */ 
    bodyds = jb2GetFileDataSink(file); 
    /* 
     * and reset the count - all bytes written after this need to be 
     * accounted for in the segment data length 
     */ 
    dsResetCount(bodyds); 
 
    /* Page bitmap width: 4 bytes */ 
    to4byte(lbuf, segment->data.page.width); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* Page bitmap height: 4 bytes */ 
    to4byte(lbuf, segment->data.page.height); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* Page X resolution: 4 bytes */ 
    to4byte(lbuf, segment->data.page.xres); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* Page Y resolution: 4 bytes */ 
    to4byte(lbuf, segment->data.page.yres); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* Page segment flags: 1 byte */ 
    cbuf = 0; 
    /* Bit 0: page is eventually lossless */ 
    if (segment->data.page.lossless) { 
	cbuf |= 1; 
	} 
 
    /* Bit 1: page may contain refinements */ 
    if (segment->data.page.may_refine) { 
	cbuf |= 2; 
	} 
 
    /* Bit 2: Page default pixel value */ 
    if (segment->data.page.defpixel) { 
	cbuf |= 4; 
	} 
 
    /* Bits 3-4: Page default combination operator */ 
    cbuf |= segment->data.page.combop << 3; 
 
    /* Bit 5: Page requires auxiliary bitmaps */ 
    if (segment->data.page.may_have_auxbuf) { 
	cbuf |= 0x20; 
	} 
 
    /* Bit 6: Page combination operator overridden */ 
    if (segment->data.page.may_override_combop) { 
	cbuf |= 0x40; 
	} 
    DSPUTC(bodyds, cbuf); 
 
    /* Page striping information: 2 bytes */ 
    sbuf = 0; 
    if (segment->data.page.striped) { 
	sbuf = (segment->data.page.max_stripe & 0x7fff) | 0x8000; 
	} 
    to2byte(lbuf, sbuf); 
    DSWRITE(bodyds, 2, lbuf, &bytes); 
 
    /* 
     * Check the DataSink's running count to find how many bytes that 
     * all took, and write that to the header field 
     */ 
    segment->datalen = dsCount(bodyds); 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeaderDataLength(file, segment)); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* 
 * Encode the end of page segment whose specification is in *segment. 
 * The fields in segment that must be filled in already are: 
 * - pagenum (must be > 0) 
 * - data->eopage 
 * 
 * The remaining fields will be filled in as follows: 
 * - segnum will be allocated 
 * - datalen will be set based on the encoded segment's data length (always 0) 
 * - nrefsegs will be set to 0 
 * - refsegs will be set to NULL 
 * - refretain will be set to NULL 
 * - segtype will be set to JB2SEGTYPE_EOPAGE 
 * - retain will be set to DGP_FALSE 
 * - defer_nonretain will be set to DGP_FALSE 
 * 
 * The encoded segment header will be written out to 
 * jb2GetFileHeaderDataSink(file). 
 */ 
ErrCode 
jb2EncodeEOPage(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2EncodeEOPage()", jb2_pErrState(file)); 
 
    /* Fill in the fields as promised */ 
    segment->segnum = jb2FileNextSegmentNumber(file); 
    segment->nrefsegs = 0; 
    segment->refsegs = (Dgp_Int32 *)NULL; 
    segment->refretain = (Dgp_Bool *)NULL; 
    segment->segtype = JB2SEGTYPE_EOPAGE; 
    segment->retain = DGP_FALSE; 
    segment->defer_nonretain = DGP_FALSE; 
    segment->datalen = 0; 
 
    /*DGP_ASSERT(segment->pagenum > 0);  */
 
    /* Everything is known - encode the header */ 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeader(file, segment)); 
 
    /* and the data length */ 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeaderDataLength(file, segment)); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* 
 * Encode the end of stripe segment whose specification is in *segment. 
 * The fields in segment that must be filled in already are: 
 * - pagenum (must be > 0) 
 * - data->eostripe 
 * 
 * The remaining fields will be filled in as follows: 
 * - segnum will be allocated 
 * - datalen will be set based on the encoded segment's data length 
 * - nrefsegs will be set to 0 
 * - refsegs will be set to NULL 
 * - refretain will be set to NULL 
 * - segtype will be set to JB2SEGTYPE_EOSTRIPE 
 * - retain will be set to DGP_FALSE 
 * - defer_nonretain will be set to DGP_FALSE 
 * 
 * The encoded segment header will be written out to 
 * jb2GetFileHeaderDataSink(file); the encoded segment data will be 
 * written out to jb2GetFileDataSink(file). 
 */ 
ErrCode 
jb2EncodeEOStripe(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2EncodeEOStripe()", jb2_pErrState(file)); 
 
    Dgp_UInt8 lbuf[4]; 
    DataSink bodyds; 
    size_t bytes; 
 
    /* Fill in the fields as promised */ 
    segment->segnum = jb2FileNextSegmentNumber(file); 
    segment->nrefsegs = 0; 
    segment->refsegs = (Dgp_Int32 *)NULL; 
    segment->refretain = (Dgp_Bool *)NULL; 
    segment->segtype = JB2SEGTYPE_EOSTRIPE; 
    segment->retain = DGP_FALSE; 
    segment->defer_nonretain = DGP_FALSE; 
 
    DGP_ASSERT(segment->pagenum > 0); 
 
    /* All but the length is known - it's OK to encode the header */ 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeader(file, segment)); 
 
    /* Get the DataSink we're going to be writing to */ 
    bodyds = jb2GetFileDataSink(file); 
    /* 
     * and reset the count - all bytes written after this need to be 
     * accounted for in the segment data length 
     */ 
    dsResetCount(bodyds); 
     
    /* 4 bytes of last_row */ 
    to4byte(lbuf, segment->data.eostripe.last_row); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* 
     * Check the DataSink's running count to find how many bytes that 
     * all took, and write that to the header field 
     */ 
    segment->datalen = dsCount(bodyds); 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeaderDataLength(file, segment)); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* 
 * Encode the end of page segment whose specification is in *segment. 
 * The fields in segment that must be filled in already are: 
 * - pagenum 
 * 
 * The remaining fields will be filled in as follows: 
 * - segnum will be allocated 
 * - datalen will be set based on the encoded segment's data length (always 0) 
 * - nrefsegs will be set to 0 
 * - refsegs will be set to NULL 
 * - refretain will be set to NULL 
 * - segtype will be set to JB2SEGTYPE_EOFILE 
 * - retain will be set to DGP_FALSE 
 * - defer_nonretain will be set to DGP_FALSE 
 * 
 * The encoded segment header will be written out to 
 * jb2GetFileHeaderDataSink(file). 
 */ 
ErrCode 
jb2EncodeEOFile(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2EncodeEOFile()", jb2_pErrState(file)); 
 
    /* Fill in the fields as promised */ 
    segment->segnum = jb2FileNextSegmentNumber(file); 
    segment->nrefsegs = 0; 
    segment->refsegs = (Dgp_Int32 *)NULL; 
    segment->refretain = (Dgp_Bool *)NULL; 
    segment->segtype = JB2SEGTYPE_EOFILE; 
    segment->retain = DGP_FALSE; 
    segment->defer_nonretain = DGP_FALSE; 
    segment->datalen = 0; 
 
    /* Everything is known - encode the header */ 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeader(file, segment)); 
 
    /* and the data length */ 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeaderDataLength(file, segment)); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
