/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/encsegment.c,v 1.2 2003/01/27 21:48:34 awilliam Exp $ 
 * 
 */ 
 
/* 
 * Routines that deal with JBIG2 segments.  These routines read 
 * segment headers, encapsulate the data obtained by decoding segments, 
 * deal with inter-segment references, and so on. 
 */ 
 
#include "util.h" 
#include "datasource.h" 
#include "segment.h" 
#include <sys/types.h> 
#include "decsymdict.h" 
#include "decsymreg.h" 
#include "dechtdict.h" 
#include "dechtreg.h" 
#include "decgenreg.h" 
#include "decgenrefreg.h" 
#include "dechufftable.h" 
#include "decpageinfo.h" 
#include "decextseg.h" 
#include <stdlib.h> 
#include "jb2file.h" 
#include "segment_util.h" 
#include "blit.h" 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/encsegment.c,v 1.2 2003/01/27 21:48:34 awilliam Exp $"); 
 
 
 
/* 
 * This routine writes out the segment header of the given segment. 
 * It uses the DataSink specified by jb2GetFileHeaderDataSink(file). 
 * It doesn't write the segment's data length because that's almost 
 * always unknown at the point where this is called.  However, it 
 * does use the mark facility of DataSink to set a mark at the place 
 * where the data length should go; the mark is filled in by 
 * jb2EncodeSegmentHeaderDataLength. 
 */ 
ErrCode 
jb2EncodeSegmentHeader(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2EncodeSegmentHeader()", jb2_pErrState(file)); 
 
    Dgp_UInt8 lbuf[4]; 
    Dgp_Int32 cbuf; 
    Dgp_Int32 i, j; 
    DataSink data = jb2GetFileHeaderDataSink(file); 
    size_t bytes; 
 
    /* Write the 4-byte segment number */ 
    to4byte(lbuf, segment->segnum); 
    DSWRITE(data, 4, lbuf, &bytes); 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGNUM %d\n", segment->segnum); 
#endif 
 
    /* Write the segment header flags */ 
    /* Bits 0-5: segment type */ 
    cbuf = (segment->segtype & 63); 
    /* Bit 6: page association >= 256? */ 
    cbuf |= segment->pagenum > 255 ? 64 : 0; 
    /* Bit 7: deferred non-retain? */ 
    cbuf |= segment->defer_nonretain ? 128 : 0; 
 
    DSPUTC(data, (Dgp_UInt8)cbuf); 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGFLAGS %02x\n", cbuf); 
#endif 
 
    /* Referenced segment count and retention flags */ 
    if (segment->nrefsegs <= 4) { 
	/* Short form referenced segment info */ 
 
	/* Bits 5-7: count */ 
	cbuf = (segment->nrefsegs << 5); 
	/* Bit 0: self-retain bit */ 
	cbuf |= segment->retain ? 1 : 0; 
 
	/* Bits 1-4: retain flags for referenced segments 0-3 */ 
	for (i = 0; i < segment->nrefsegs; i++) { 
	    cbuf |= segment->refretain[i] ? (1 << (i + 1)) : 0; 
	    } 
 
	DSPUTC(data, (Dgp_UInt8)cbuf); 
	} 
    else { 
	/* Long form referenced segment info */ 
 
	/* 4 byte count + (7 << 29) */ 
	to4byte(lbuf, segment->nrefsegs + (7 << 29)); 
	DSWRITE(data, 4, lbuf, &bytes); 
 
	/* First byte: bit 0 is self-retain */ 
	cbuf = segment->retain ? 1 : 0; 
	/* Bits 1-7: retain flags for referenced segments 0-6 */ 
	for (i = 0; i < MIN(7, segment->nrefsegs); i++) { 
	    cbuf |= segment->refretain[i] ? (1 << (i + 1)) : 0; 
	    } 
 
	DSPUTC(data, (Dgp_UInt8)cbuf); 
 
	/* Subsequent bytes: 8 retain flags each, filled LSB-to-MSB */ 
	for (i = 7; i < segment->nrefsegs; i += 8) { 
	    cbuf = 0; 
	    for (j = 0; j < MIN(8, segment->nrefsegs - i); j++) { 
			cbuf |= segment->refretain[i + j] ? (1 << j) : 0; 
		} 
		DSPUTC(data, (Dgp_UInt8)cbuf); 
	} 
	} 
 
    /* Referred-to segment numbers */ 
    if (segment->segnum <= 256) { 
	/* One byte each */ 
	for (i = 0; i < segment->nrefsegs; i++) { 
	    DGP_ASSERT(segment->refsegs[i] < segment->segnum); 
	    cbuf = segment->refsegs[i]; 
	    DSPUTC(data, (Dgp_UInt8)cbuf); 
#ifdef	VALPRINT 
	    fprintf(stderr, "SEGREF %d\n", segment->refsegs[i]); 
#endif 
	    } 
	} 
    else if (segment->segnum <= 65536) { 
	/* Two bytes each */ 
	for (i = 0; i < segment->nrefsegs; i++) { 
	    DGP_ASSERT(segment->refsegs[i] < segment->segnum); 
	    to2byte(lbuf, segment->refsegs[i]); 
	    DSWRITE(data, 2, lbuf, &bytes); 
#ifdef	VALPRINT 
	    fprintf(stderr, "SEGREF %d\n", segment->refsegs[i]); 
#endif 
	    } 
	} 
    else { 
	/* Four bytes each */ 
	for (i = 0; i < segment->nrefsegs; i++) { 
	    DGP_ASSERT(segment->refsegs[i] < segment->segnum); 
	    to4byte(lbuf, segment->refsegs[i]); 
	    DSWRITE(data, 4, lbuf, &bytes); 
#ifdef	VALPRINT 
	    fprintf(stderr, "SEGREF %d\n", segment->refsegs[i]); 
#endif 
	    } 
	} 
 
    /* Segment page association */ 
    if (segment->pagenum < 256) { 
	/* One byte */ 
	cbuf = segment->pagenum; 
	DSPUTC(data, (Dgp_UInt8)cbuf); 
	} 
    else { 
	/* Four bytes */ 
	to4byte(lbuf, segment->pagenum); 
	DSWRITE(data, 4, lbuf, &bytes); 
	} 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGPAGE %d\n", segment->pagenum); 
#endif 
 
    /* Segment data length - 4 bytes; value currently unknown */ 
    DGP_ERR_PASS_CLEANUP(dsMark(data, 4)); 
	 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* Fill in the 4-byte segment header data length, now known */ 
 
ErrCode 
jb2EncodeSegmentHeaderDataLength(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2EncodeSegmentHeaderDataLength()", jb2_pErrState(file)); 
    Dgp_UInt8 lbuf[4]; 
    DataSink data = jb2GetFileHeaderDataSink(file); 
 
    to4byte(lbuf, segment->datalen); 
 
    DGP_ERR_PASS_CLEANUP(dsFillMark(data, 4, lbuf)); 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGDATALEN %ld\n", segment->datalen); 
#endif 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
