/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/encsegment_util.c,v 1.2 2003/01/27 21:48:34 awilliam Exp $ 
 * 
 */ 
 
/* 
 * Convenience routines for dealing with segments. 
 */ 
 
#include "util.h" 
#include "segment.h" 
#include "segment_util.h" 
#include "hufftable.h" 
#include "list.h" 
#include "PsgFix.h"
#include <stdlib.h> 
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/encsegment_util.c,v 1.2 2003/01/27 21:48:34 awilliam Exp $"); 
 
 
/* 
 * Routines to assist in segment manipulation. 
 */ 
 
/* 
 * This routine fixes up the referenced segment ID number array, adding in 
 * the given segments.  It also increases the size of the refretain array, 
 * setting the new values to DGP_FALSE.  They may be changed if desired. 
 * It also adjusts segment->nrefsegs. 
 */ 
ErrCode 
jb2AddReferencedSegments(jb2File file, jb2Segment *segment, 
			 jb2Segment **newrefsegs, Dgp_Int32 nnewrefsegs) 
{ 
  DGP_PROLOGUE("jb2AddReferencedSegments()", jb2_pErrState(file)); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
  Dgp_Int32 i; 
  Dgp_Int32 *newrefsegnums = NULL; 
  Dgp_Bool *newrefretain = NULL; 
 
    if (segment->refsegs == (Dgp_Int32 *)NULL) { 
	DGP_ASSERT(segment->nrefsegs == 0); 
	segment->refsegs = newrefsegnums = 
	    (Dgp_Int32 *)(a->malloc)(nnewrefsegs * sizeof(*newrefsegnums)); 
	if (newrefsegnums == (Dgp_Int32 *)NULL) { 
	    NOMEM; 
	    } 
	} 
    else { 
	/* Do you feel lucky? */ 
	newrefsegnums = 
	    (Dgp_Int32 *)(a->realloc)(segment->refsegs, 
			   (segment->nrefsegs + nnewrefsegs) * 
			   sizeof(*newrefsegnums)); 
 
	if (newrefsegnums == (Dgp_Int32 *)NULL) { 
	    NOMEM; 
	    } 
	segment->refsegs = newrefsegnums; 
	} 
 
    for (i = 0; i < nnewrefsegs; i++) { 
	newrefsegnums[i + segment->nrefsegs] = newrefsegs[i]->segnum; 
	} 
 
    if (segment->refretain == (Dgp_Bool *)NULL) { 
	DGP_ASSERT(segment->nrefsegs == 0); 
	segment->refretain = newrefretain = 
	    (Dgp_Bool *)(a->malloc)(nnewrefsegs * sizeof(*newrefretain)); 
	if (newrefretain == (Dgp_Bool *)NULL) { 
	    NOMEM; 
	    } 
	} 
    else { 
	/* Do you feel lucky? */ 
	newrefretain = 
	    (Dgp_Bool *)(a->realloc)(segment->refretain, 
			       (segment->nrefsegs + nnewrefsegs) * 
			       sizeof(*newrefretain)); 
 
	if (newrefretain == (Dgp_Bool *)NULL) { 
	    NOMEM; 
	    } 
	segment->refretain = newrefretain; 
	} 
 
    for (i = 0; i < nnewrefsegs; i++) { 
	newrefretain[i + segment->nrefsegs] = DGP_FALSE; 
	} 
 
    segment->nrefsegs += nnewrefsegs; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    /* 
     * Turns out there's nothing we can do to clean up: 
     * it some of the reallocs/mallocs succeeded before we got here, 
     * they're already cleanly linked into segment 
     */ 
    ERR_RETURN; 
    } 
 
/* 
 * Encode the region segment information 
 */ 
ErrCode 
jb2EncodeRegSegment(jb2File file, DataSink data, jb2RegSegment *reginfo) 
{ 
  DGP_PROLOGUE("jb2EncodeRegSegment()", jb2_pErrState(file)); 
 
    Dgp_UInt8 lbuf[4]; 
    Dgp_Int32 cbuf; 
    size_t bytes; 
 
    /* Width */ 
    to4byte(lbuf, reginfo->regwidth); 
    DSWRITE(data, 4, lbuf, &bytes); 
 
    /* Height */ 
    to4byte(lbuf, reginfo->regheight); 
    DSWRITE(data, 4, lbuf, &bytes); 
     
    /* X offset */ 
    to4byte(lbuf, reginfo->regxpos); 
    DSWRITE(data, 4, lbuf, &bytes); 
     
    /* Y offset */ 
    to4byte(lbuf, reginfo->regypos); 
    DSWRITE(data, 4, lbuf, &bytes); 
     
    /* Flags */ 
    /* Bits 0-2: combination operator */ 
    DGP_ASSERT(reginfo->combop <= 4); 
    cbuf = reginfo->combop; 
 
    DSPUTC(data, (Dgp_UInt8)cbuf); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
