/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/encsymdict.c,v 1.24 2003/01/27 21:48:34 awilliam Exp $ 
 * 
 */ 
 
 
#include "segment.h" 
#include "encsymdict.h" 
#include "jbitfile.h" 
#include "datasink.h" 
#include "jb2file.h" 
#include "fileerr.h" 
#include "segment_util.h" 
#include <stdlib.h> 
#include "encint.h" 
#include "encgenreg.h" 
#include "decgenreg.h" 
#include "hufftable.h" 
#include "hufftable_util.h" 
#include "stdhuff.h" 
#include "enchufftable.h" 
#include <string.h> 
#include "PsgFix.h"
 
#define DGP_ENC DGP_FALSE

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/encsymdict.c,v 1.24 2003/01/27 21:48:34 awilliam Exp $"); 
 
static int symGT(const void *p1, const void *p2); 
static ErrCode decideSDTables(jb2FileParams *params, 
			      jb2Symbol **symptrs, Dgp_Int32 nsyms, 
			      hufftable_spec **dwtable, Dgp_Bool *dwnew, 
			      hufftable_spec **dhtable, Dgp_Bool *dhnew, 
			      hufftable_spec **bmsizetable, 
			      Dgp_Bool *bmsizenew, Dgp_State *dgp_state); 
static ErrCode writeMMRCollectiveBitmap(jb2File file, DataSink data, 
					jbitfile bf, 
					eiEncoder bmsize, 
					jb2Symbol **symbols, 
					Dgp_Int32 nsymbols); 
static void placeSymbol(jb2Symbol *sym, Dgp_UInt32 *bits, 
			   Dgp_Int32 row_width, Dgp_Int32 xpos); 
static void memcpy_shuffling(Dgp_UInt32 *bits, Dgp_UInt8 *buf, 
			     Dgp_Int32 row_width, Dgp_Int32 Dgp_Int8_row_width, Dgp_Int32 height); 
 
#ifdef	VALPRINT 
 
#include <stdio.h> 
 
#define	WRITEV(coder, v) \
    do { \
	DGP_ERR_PASS_CLEANUP(eiEncode(coder, (v), DGP_FALSE)); \
	fprintf(stderr, #coder " %d\n", v); \
	} while (0) 
#define	WRITEOOB(coder) \
    do { \
	DGP_ERR_PASS_CLEANUP(eiEncode(coder, 0, DGP_TRUE)); \
	fprintf(stderr, #coder " OOB\n"); \
	} while (0) 
 
#else 
 
#define	WRITEV(coder, v) \
    do { \
	DGP_ERR_PASS_CLEANUP(eiEncode(coder, (v), DGP_FALSE)); \
	} while (0) 
#define	WRITEOOB(coder) \
    do { \
	DGP_ERR_PASS_CLEANUP(eiEncode(coder, 0, DGP_TRUE)); \
	} while (0) 
 
#endif 
 
/* 
 * Arrays of possible Huffman tables 
 */ 
static hufftable_spec *dwtables[] = { 
    &stdhuffDW1, &stdhuffDW2 
    }; 
#define	NDWTABLES	((int)(sizeof(dwtables) / sizeof(dwtables[0]))) 
 
static hufftable_spec *dhtables[] = { 
    &stdhuffDH1, &stdhuffDH2 
    }; 
#define	NDHTABLES	((int)(sizeof(dhtables) / sizeof(dhtables[0]))) 
 
/* 
 * Encode the symbol dictionary, given the info stored in *segment and 
 * *refsegs.  Each refseg should be another symbol dictionary segment. 
 * 
 * In some cases, this routine will generate additional Huffman table 
 * segments.  In this case, the Huffman table segments will be written out 
 * to the jb2File before this segment.  This routine 
 * will call jb2AddReferencedSegments to add those segments to this 
 * segment's lists.  They will be flagged as "not retained beyond this 
 * segment". 
 * 
 * The fields in segment that must be filled in already are: 
 * - pagenum 
 * - nrefsegs 
 * - refsegs 
 * - refretain 
 * - data->symdict 
 * 
 * The remaining fields will be filled in as follows: 
 * - segnum will be allocated 
 * - datalen will be set based on the encoded segment's data length 
 * - retain will be set to DGP_TRUE 
 * - defer_nonretain will be set to DGP_FALSE 
 * 
 * The encoded segment header will be written out to 
 * jb2GetFileHeaderDataSink(file); the encoded segment data will be 
 * written out to jb2GetFileDataSink(file).  Neither of those will be 
 * examined until the Huffman table segments (if any) have been written 
 * out. 
 * 
 * This routine also can participate in building extensions.  Each jb2Symbol 
 * can contain a void * pointer (userdata); there can also be an 
 * "extbuilder" routine passed in.  This routine gets called with an 
 * array of jb2Symbol * pointers, pointing to the symbols in the order in 
 * which they will be written out.  Note that extbuilder might be called, 
 * only to have jb2EncodeSymDict later fail. 
 * 
 * A VERY IMPORTANT side effect of this routine is that if reorder is DGP_TRUE, 
 * the symbols in segment are re-ordered, so that the order that they are 
 * in when the routine returns is the order that they were written to the 
 * file.  This may be disruptive: if you have pointers directly into the 
 * symbols array of segment, then this re-ordering will make those pointers 
 * invalid. 
 * 
 * The re-ordering is done in-place, so the symbols array on return is the 
 * same piece of memory as it was when the routine was called. 
 * 
 * This routine obeys the encoding parameters set in file's jb2FileParams. 
 */ 
ErrCode 
jb2EncodeSymDict(jb2File file, 
		 jb2Segment *segment, jb2Segment **refsegs, 
		 ErrCode (*extbuilder)(jb2Symbol **sortedsyms, Dgp_Int32 nsyms, 
				       void *extcookie), 
		 void *extcookie, Dgp_Bool reorder) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2EncodeSymDict()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    DataSink bodyds; 
    jbitfile bf = NULLJBITFILE, tmpbf; 
    DgpEncode_MQEncoder mqenc; 
    jb2SymDictSegment *symdict; 
    jb2Symbol **symptrs = (jb2Symbol **)NULL; 
    jb2Symbol *tempsyms = (jb2Symbol *)NULL; 
    hufftable_spec *dwtable = (hufftable_spec *)NULL; 
    hufftable_spec *dhtable = (hufftable_spec *)NULL; 
    hufftable_spec *bmsizetable = (hufftable_spec *)NULL; 
    Dgp_Bool dwnew, dhnew, bmsizenew; 
    jb2Segment *newrefsegs[3]; 
    Dgp_Int32 newref; 
    Dgp_Int32 i; 
    jb2Segment dwseg, dhseg, bmsizeseg; 
    eiEncoder dw = NULLEIENCODER, dh = NULLEIENCODER, 
	bmsize = NULLEIENCODER, exflags = NULLEIENCODER; 
    Dgp_Int32 refsymcount; 
    Dgp_UInt16 sbuf; 
    Dgp_UInt8 lbuf[4]; 
    Dgp_Int32 oldheight, oldwidth, classstart; 
    jb2Symbol *cursym; 
    jb2FileParams *params; 
    /* Some context buffers */ 
    Dgp_UInt8 *direct_contexts = (Dgp_UInt8 *)NULL; 
    Dgp_UInt8 dwcx[EICONTEXTSIZE], dhcx[EICONTEXTSIZE], 
	exflagscx[EICONTEXTSIZE]; 
    size_t bytes; 
 
    DGP_ASSERT(segment->segtype == JB2SEGTYPE_SYMDICT); 
    symdict = &segment->data.symdict; 
    /* 
     * First, decide how we're going to do the encoding: which 
     * Huffman tables, and so on.  Create any Huffman table 
     * segments if desired. 
     * Second, write out the segment header (has to come after the 
     * routine has a chance to create the Huffman tables and their 
     * associated segments). 
     * Third, write out the data header describing the encoding choices. 
     * Finally, write the dictionary itself. 
     */ 
 
    /* Get the parameters we need to follow */ 
    params = jb2GetFileParams(file); 
 
    /* Sort the symbols by width and height. */ 
 
    /* Allocate the array of pointers */ 
    symptrs = (jb2Symbol **)(a->malloc)(symdict->nsymbols * sizeof(*symptrs)); 
    if (symptrs == (jb2Symbol **)NULL) { 
      NOMEM; 
	} 
 
    /* and fill them in to point to the symbols */ 
    for (i = 0; i < symdict->nsymbols; i++) { 
	symptrs[i] = &symdict->symbols[i]; 
	} 
		     
    /* Do the sort */ 
    qsort((void *)symptrs, (unsigned)symdict->nsymbols, sizeof(*symptrs), 
	  symGT); 
 
    /* At this point, we can call the extbuilder. */ 
    if (extbuilder != (ErrCode (*)(jb2Symbol **, Dgp_Int32, void *))NULL) { 
      DGP_ERR_PASS_CLEANUP((*extbuilder)(symptrs, symdict->nsymbols, 
					 extcookie)); 
	} 
 
    /* Now run through the sorted data, accumulating statistics */ 
    if (!params->sdarith) { 
	DGP_ERR_PASS_CLEANUP(decideSDTables(params, symptrs, symdict->nsymbols, 
			    &dwtable, &dwnew, &dhtable, &dhnew, 
					    &bmsizetable, &bmsizenew, 
					    dgp_state)); 
 
	/* That gave us a bunch of hufftable_spec */ 
 
	/* 
	 * For each new Huffman table, write out a segment for it. 
	 * Add the segment pointers to newrefsegs as we go.  The order to 
	 * add them in MUST be DH, then DW, then BMSIZE. 
	 */ 
	newref = 0; 
 
	if (dhnew) { 
	    /* Fill in the required fields */ 
	    dhseg.pagenum = segment->pagenum; 
	    dhseg.data.table.table = *dhtable; 
 
	    /* and write out the segment */ 
	    DGP_ERR_PASS_CLEANUP(jb2EncodeHuffTable(file, &dhseg)); 
 
	    newrefsegs[newref++] = &dhseg; 
	    /* That didn't allocate anything - nothing to free */ 
	    } 
 
	if (dwnew) { 
	    dwseg.pagenum = segment->pagenum; 
	    dwseg.data.table.table = *dwtable; 
	    DGP_ERR_PASS_CLEANUP(jb2EncodeHuffTable(file, &dwseg)); 
	    newrefsegs[newref++] = &dwseg; 
	    } 
 
	if (bmsizenew) { 
	    bmsizeseg.pagenum = segment->pagenum; 
	    bmsizeseg.data.table.table = *bmsizetable; 
	    DGP_ERR_PASS_CLEANUP(jb2EncodeHuffTable(file, &bmsizeseg)); 
	    newrefsegs[newref++] = &bmsizeseg; 
	    } 
 
	/* 
	 * Add the new segment references to this segment's information 
	 */ 
	if (newref > 0) { 
	    DGP_ERR_PASS_CLEANUP(jb2AddReferencedSegments(file, segment, 
							  newrefsegs, newref)); 
	    /* Retain bits already filled with NULL */ 
	    } 
	} 
    else { 
	/* No Huffman tables were created */ 
	newref = 0; 
	dwnew = dhnew = bmsizenew = DGP_FALSE; 
	} 
 
    /* Finally, we can start on encoding the symbol dictionary */ 
 
    /* Write out the segment header */ 
 
    /* 
     * Fill in the fields we promised to fill in.  We couldn't even 
     * allocate this segment's segment number before doing the Huffman table 
     * dance above, so start there. 
     */ 
    segment->segnum = jb2FileNextSegmentNumber(file); 
    segment->retain = DGP_TRUE; 
    segment->defer_nonretain = DGP_FALSE; 
 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeader(file, segment)); 
     
    /* Write out the segment data header */ 
    bodyds = jb2GetFileDataSink(file); 
    /* Reset the count on bodyds so we can get the data length */ 
    dsResetCount(bodyds); 
 
    /* Symbol dictionary flags: two bytes */ 
 
    /* Bit 0: SDHUFF */ 
    if (params->sdarith) { 
	/* Non-Huffman */ 
#ifndef	NOMQ 
	sbuf = 0; 
#else 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
	} 
    else { 
	sbuf = 1; 
	} 
     
    /* Bit 1: SDREFAGG */ 
    /* always 0 currently */ 
 
    /* 
     * Fill in Huffman table selections only if non-arith; if arith, 
     * leave them all 0 
     */ 
    if (!params->sdarith) { 
	/* Bits 2-3: SDHUFFDH selection */ 
	if (dhtable == &stdhuffDH1) { 
	    /* value of 0 */ 
	    } 
	else if (dhtable == &stdhuffDH2) { 
	    /* value of 1 */ 
	    sbuf |= (1 << 2); 
	    } 
	else { 
	    /* custom table; value of 3 */ 
	    sbuf |= (3 << 2); 
	    } 
 
	/* Bits 4-5: SDHUFFDW selection */ 
	if (dwtable == &stdhuffDW1) { 
	    /* value of 0 */ 
	    } 
	else if (dwtable == &stdhuffDW2) { 
	    /* value of 1 */ 
	    sbuf |= (1 << 4); 
	    } 
	else { 
	    /* custom table; value of 3 */ 
	    sbuf |= (3 << 4); 
	    } 
 
	/* Bit 6: SDHUFFBMSIZE selection */ 
	if (bmsizetable == &stdhuffA) { 
	    /* value of 0 */ 
	    } 
	else { 
	    /* custom table; value of 1 */ 
	    sbuf |= (1 << 6); 
	    } 
 
	/* Bit 7: SDHUFFAGGINST selection */ 
	/* always 0 - stdhuffA */ 
	} 
 
    /* Bit 8: Bitmap coding context used */ 
    /* always 0 currently */ 
 
    /* Bit 9: Bitmap coding context retained */ 
    /* always 0 currently */ 
 
    /* Bits 10-11: SDTEMPLATE */ 
    /* always 0 for Huffman coding, see params for arith */ 
    if (params->sdarith) { 
	sbuf |= params->sdabparams.rtemplate << 10; 
	} 
     
    /* Bit 12: SDRTEMPLATE */ 
    /* always 0 for non-refinement coding */ 
 
    to2byte(lbuf, sbuf); 
    DSWRITE(bodyds, 2, lbuf, &bytes); 
 
    /* Encode AT flags field if arith */ 
    if (params->sdarith) { 
      DGP_ERR_PASS_CLEANUP(jb2EncodeABParams(file, bodyds, 
					     &params->sdabparams)); 
	} 
 
    /* No refinement AT flags fields for non-refinement coding */ 
 
    /* SDNUMEXSYMS: all of the symbols */ 
    to4byte(lbuf, symdict->nsymbols); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* SDNUMNEWSYMS: all of the symbols */ 
    to4byte(lbuf, symdict->nsymbols); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* Now start writing the actual body */ 
 
    /* 
     * Get a bitfile/DgpEncode_MQEncoder to write to, 
     * and generate some integer encoders 
     */ 
    if (params->sdarith) { 
#ifndef	NOMQ 
	dsInitDgpEncode_MQEncoder(bodyds, &mqenc); 
#endif 
 
	/* Allocate and clear the direct bitmap coding contexts */ 
	direct_contexts = 
	    (Dgp_UInt8 *)(a->calloc)(jb2ABContextsNeeded(&params->sdabparams), 
				    1); 
	DGP_CHECK_ALLOC_CLEANUP(direct_contexts); 
	if (params->sdabparams.tpon) { 
	    FILEERR(cErrCompressionError, "symdict can't have TPD on"); 
	    } 
	} 
    else { 
      DGP_ERR_PASS_CLEANUP(jbfOpenFromDataSink(bodyds, &bf)); 
 
	/* Only Huffman needs bmsize */ 
	DGP_ERR_PASS_CLEANUP(eiNewEncoder(bmsizetable, bf, dgp_state, 
					  &bmsize)); 
	} 
 
    DGP_ERR_PASS_CLEANUP(eiNewEncoderEither((Dgp_Bool)!params->sdarith, dwtable, bf, 
					 &mqenc, dwcx, dgp_state, &dw)); 
 
    DGP_ERR_PASS_CLEANUP(eiNewEncoderEither((Dgp_Bool)!params->sdarith, dhtable, bf, 
					  &mqenc, dhcx, dgp_state, &dh)); 
 
    DGP_ERR_PASS_CLEANUP(eiNewEncoderEither((Dgp_Bool)!params->sdarith, &stdhuffA, bf, 
			       &mqenc, exflagscx, dgp_state, &exflags)); 
 
    /* Start the encoding */ 
    oldheight = oldwidth = 0; 
    classstart = 0; 
    for (i = 0; i < symdict->nsymbols; i++) { 
	cursym = symptrs[i]; 
 
	if (cursym->height != oldheight) { 
	    /* New height; end the old height class if any */ 
	    if (oldheight != 0) { 
		/* End the previous height class with an DW-OOB */ 
		WRITEOOB(dw); 
 
		if (!params->sdarith) { 
		    /* Generate and write the collective bitmap */ 
		    DGP_ERR_PASS_CLEANUP(writeMMRCollectiveBitmap(file, bodyds, bf, bmsize, 
						  symptrs + classstart, 
						  i - classstart)); 
		    } 
		} 
 
	    /* Write the delta height */ 
	    WRITEV(dh, cursym->height - oldheight); 
 
	    oldheight = cursym->height; 
	    oldwidth = 0; 
	    classstart = i; 
	    } 
 
	/* Write out the delta width */ 
	WRITEV(dw, cursym->width - oldwidth); 
	oldwidth = cursym->width; 
 
	/* Write the symbol bitmap */ 
	if (params->sdarith) { 
#ifndef	NOMQ 
	    DGP_ERR_PASS_CLEANUP(jb2ArithEncodeImage(cursym->bits, cursym->width, 
				     cursym->height, cursym->longsperline, 
				     &params->sdabparams, direct_contexts, 
						     file, &mqenc)); 
#endif 
	    } 
	} 
 
    /* Write out an end for the last height class */ 
    if (oldheight != 0) { 
	/* End the previous height class with an DW-OOB */ 
	WRITEOOB(dw); 
 
	/* Generate and write the collective bitmap */ 
	if (!params->sdarith) { 
	    DGP_ERR_PASS_CLEANUP(writeMMRCollectiveBitmap(file, bodyds, bf, bmsize, 
					  symptrs + classstart, 
					  i - classstart)); 
	    } 
	} 
 
    /* Now write out the export list. */ 
 
    /* 
     * Simple at the moment: it's a run of "not exported" for all 
     * of the dictionaries referenced by this segment, followed by a run 
     * of "exported" for all of the symbols in this dictionary. 
     */ 
    refsymcount = 0; 
    if (segment->nrefsegs > newref) { 
	/* We were passed in some segments in refsegs */ 
	DGP_ASSERT(refsegs != (jb2Segment **)NULL); 
	for (i = 0; i < segment->nrefsegs - newref; i++) { 
	    if (refsegs[i]->segtype != JB2SEGTYPE_SYMDICT) { 
		DGP_REPORTF0(DGP_ENC, "bad referenced segment type"); 
		} 
	    refsymcount += refsegs[i]->data.symdict.nsymbols; 
	    } 
	} 
 
    /* Encode a run of "refsymcount" non-exported symbols */ 
    WRITEV(exflags, refsymcount); 
 
    /* Encode a run of "segment->nsymbols" exported symbols */ 
    WRITEV(exflags, symdict->nsymbols); 
 
    /* 
     * Close down the jbitfile; free the eiEncoders and any new 
     * Huffman tables 
     */ 
    if (params->sdarith) { 
	/* Flush the DgpEncode_MQEncoder */ 
#ifndef	NOMQ 
      DGP_ERR_PASS_CLEANUP(dsFinalDgpEncode_MQEncoder(bodyds, &mqenc)); 
#endif 
	(a->free)((void *)direct_contexts); 
	direct_contexts = (Dgp_UInt8 *)NULL; 
	} 
    else { 
      tmpbf = bf; 
      bf = NULLJBITFILE; 
      DGP_ERR_PASS_CLEANUP(jbfCloseFromDataSink(tmpbf)); 
	} 
 
    /* If they asked us to, reorder the symbols */ 
    if (reorder) { 
	/* Copy the symbols array in order, then memcpy it back */ 
	tempsyms = (jb2Symbol *)(a->malloc)(symdict->nsymbols * sizeof(*tempsyms)); 
	if (tempsyms == (jb2Symbol *)NULL) { 
	    NOMEM; 
	    } 
 
	/* Copy into tempsyms */ 
	for (i = 0; i < symdict->nsymbols; i++) { 
	    tempsyms[i] = *symptrs[i]; 
	    } 
 
	/* Copy all of tempsyms back */ 
	memcpy((void *)symdict->symbols, (void *)tempsyms, 
	       symdict->nsymbols * sizeof(*tempsyms)); 
	/* and free it */ 
	(a->free)((void *)tempsyms); 
	tempsyms = (jb2Symbol *)NULL; 
	} 
 
    (a->free)((void *)symptrs); 
 
    /* 
     * Check the DataSink's running count to find how many bytes that 
     * all took, and write that to the header field 
     */ 
    segment->datalen = dsCount(bodyds); 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeaderDataLength(file, segment)); 
 
    eiFreeEncoder(dw); 
    eiFreeEncoder(dh); 
    if (!params->sdarith) { 
	eiFreeEncoder(bmsize); 
	} 
    eiFreeEncoder(exflags); 
 
    if (dwnew) { 
	jhtFree(dwtable, dgp_state); 
	} 
 
    if (dhnew) { 
	jhtFree(dhtable, dgp_state); 
	} 
 
    if (bmsizenew) { 
	jhtFree(bmsizetable, dgp_state); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (dw != NULLEIENCODER) { 
	eiFreeEncoder(dw); 
	} 
 
    if (dh != NULLEIENCODER) { 
	eiFreeEncoder(dh); 
	} 
 
    if (bmsize != NULLEIENCODER) { 
	eiFreeEncoder(bmsize); 
	} 
 
    if (exflags != NULLEIENCODER) { 
	eiFreeEncoder(exflags); 
	} 
 
    if (bf != NULLJBITFILE) { 
	(void)jbfCloseFromDataSink(bf); 
	} 
 
    if ((dwtable != (hufftable_spec *)NULL) && (dwnew)) { 
	jhtFree(dwtable, dgp_state); 
	} 
     
    if ((dhtable != (hufftable_spec *)NULL) && (dhnew)) { 
	jhtFree(dhtable, dgp_state); 
	} 
     
    if ((bmsizetable != (hufftable_spec *)NULL) && (bmsizenew)) { 
	jhtFree(bmsizetable, dgp_state); 
	} 
     
    if (symptrs != (jb2Symbol **)NULL) { 
	(a->free)((void *)symptrs); 
	} 
 
    if (tempsyms != (jb2Symbol *)NULL) { 
	(a->free)((void *)tempsyms); 
	} 
 
    if (direct_contexts != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)direct_contexts); 
	} 
 
    ERR_RETURN; 
} 

typedef const jb2Symbol *jb2SPtr; 
/* 
 * Compare the two symbols pointed at by *p1 and *p2. 
 * Primary key is height; secondary key is width; tertiary 
 * key is the symbol pointers themselves. 
 */ 
static int
symGT(const void *p1, const void *p2) 
{ 
    const jb2Symbol *s1 = *((const jb2SPtr *)p1); 
    const jb2Symbol *s2 = *((const jb2SPtr *)p2); 
 
    if ((s1->height > s2->height) || 
	((s1->height == s2->height) && (s1->width > s2->width)) || 
	((s1->height == s2->height) && (s1->width == s2->width) && 
	 (s1 > s2))) { 
	return(1); 
	} 
    else { 
	return(-1); 
	} 
    } 
 
/* 
 * Run through the data, producing lists of values; decide on the three 
 * tables that the segment will use. 
 * 
 * For each one, return the chosen table in *FOOtable, and whether this 
 * is a custom table in *FOOnew.  If it's not a custom table, then 
 * *FOOtable is a pointer to one of the standard tables. 
 * 
 * The caller is responsible for freeing any new tables; use jhtFree. 
 * 
 * Currently, it always uses the same, standard, table for bmsize, since 
 * it's a real pain to figure out the statistics of that one without 
 * doing a lot of useless work. 
 * 
 * If params->sdtables is DGP_FALSE, it will never generate a custom table. 
 */ 
static ErrCode 
decideSDTables(jb2FileParams *params, jb2Symbol **symptrs, Dgp_Int32 nsyms, 
	       hufftable_spec **dwtable, Dgp_Bool *dwnew, 
	       hufftable_spec **dhtable, Dgp_Bool *dhnew, 
	       hufftable_spec **bmsizetable, Dgp_Bool *bmsizenew, 
	       Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("decideSDTables()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Int32 i, j; 
    Dgp_Int32 *dwlengths[NDWTABLES], *dhlengths[NDHTABLES], 
	*dwcounts[NDWTABLES], *dhcounts[NDHTABLES], 
	dwnormal[NDWTABLES], dhnormal[NDHTABLES], 
	dwtuned[NDWTABLES], dhtuned[NDHTABLES]; 
    jb2Symbol *cursym; 
    eiEncoder dw[NDWTABLES], dh[NDWTABLES]; 
    Dgp_Int32 oldheight, oldwidth; 
 
    /* Allocate the table information */ 
 
    /* Clear for easy bailout */ 
    for (i = 0; i < NDWTABLES; i++) { 
	dwlengths[i] = (Dgp_Int32 *)NULL; 
	dwcounts[i] = (Dgp_Int32 *)NULL; 
	dw[i] = NULLEIENCODER; 
	} 
 
    for (i = 0; i < NDHTABLES; i++) { 
	dhlengths[i] = (Dgp_Int32 *)NULL; 
	dhcounts[i] = (Dgp_Int32 *)NULL; 
	dh[i] = NULLEIENCODER; 
	} 
    *dwtable = (hufftable_spec *)NULL; 
    *dhtable = (hufftable_spec *)NULL; 
    *dwnew = DGP_FALSE; 
    *dhnew = DGP_FALSE; 
 
    /* Allocate the count and length arrays, plus the eiEncoders */ 
    DGP_ERR_PASS_CLEANUP(jhtAllocTableInfo(dwtables, dwlengths, dwcounts, dw, 
					   NDWTABLES, dgp_state)); 
 
    DGP_ERR_PASS_CLEANUP(jhtAllocTableInfo(dhtables, dhlengths, dhcounts, dh, 
					   NDHTABLES, dgp_state)); 
     
    /* Fixed choice for bmsize */ 
    *bmsizetable = &stdhuffA; 
    *bmsizenew = DGP_FALSE; 
 
    oldheight = oldwidth = 0; 
    for (i = 0; i < nsyms; i++) { 
	cursym = symptrs[i]; 
 
	if (cursym->height != oldheight) { 
	    /* New height class */ 
	    if (oldheight != 0) { 
		/* Not the first height class - write an OOB for width */ 
		for (j = 0; j < NDWTABLES; j++) { 
		    eiCount(dw[j], 0, DGP_TRUE, dwcounts[j]); 
		    } 
		} 
 
	    /* Write the delta height */ 
	    for (j = 0; j < NDHTABLES; j++) { 
		eiCount(dh[j], cursym->height - oldheight, 
			DGP_FALSE, dhcounts[j]); 
		} 
 
	    oldheight = cursym->height; 
	    oldwidth = 0; 
	    } 
 
	/* Write the delta width */ 
	for (j = 0; j < NDWTABLES; j++) { 
	    eiCount(dw[j], cursym->width - oldwidth, DGP_FALSE, dwcounts[j]); 
	    } 
	oldwidth = cursym->width; 
	} 
    /* Write an OOB for width to end the last height class */ 
    if (oldheight != 0) { 
	for (j = 0; j < NDWTABLES; j++) { 
	    eiCount(dw[j], 0, DGP_TRUE, dwcounts[j]); 
	    } 
	} 
 
    /* We now have all the statistics.  Compute bit counts. */ 
    for (i = 0; i < NDWTABLES; i++) { 
	DGP_ERR_PASS_CLEANUP(jhtCountBits(dwtables[i], dwcounts[i],  
					  dwlengths[i], &dwnormal[i], 
					  &dwtuned[i], dgp_state)); 
	} 
    for (i = 0; i < NDHTABLES; i++) { 
	DGP_ERR_PASS_CLEANUP(jhtCountBits(dhtables[i], dhcounts[i], 
					  dhlengths[i], &dhnormal[i], 
					  &dhtuned[i], dgp_state)); 
	} 
 
    /* Our parameters might force the decision. */ 
    if (params->sdfirstonly) { 
	/* No decision - only allowed to choose the first */ 
	*dwtable = dwtables[0]; 
	*dwnew = DGP_FALSE; 
 
	*dhtable = dhtables[0]; 
	*dhnew = DGP_FALSE; 
	} 
    else { 
	/* Make the DW decision */ 
	DGP_ERR_PASS_CLEANUP(jhtDecideTable(dwtables, dwcounts, dwlengths, 
					    dwnormal, dwtuned, dwnew, dwtable, 
					    NDWTABLES, params->sdtables, 
					    dgp_state)); 
 
	/* Make the DH decision */ 
	DGP_ERR_PASS_CLEANUP(jhtDecideTable(dhtables, dhcounts, dhlengths, 
					    dhnormal, dhtuned, dhnew, dhtable, 
					    NDHTABLES, params->sdtables, 
					    dgp_state)); 
	} 
 
    /* Clean up our arrays */ 
    for (i = 0; i < NDWTABLES; i++) { 
	(a->free)((void *)dwlengths[i]); 
	(a->free)((void *)dwcounts[i]); 
	eiFreeEncoder(dw[i]); 
	} 
 
    for (i = 0; i < NDHTABLES; i++) { 
	(a->free)((void *)dhlengths[i]); 
	(a->free)((void *)dhcounts[i]); 
	eiFreeEncoder(dh[i]); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    for (i = 0; i < NDWTABLES; i++) { 
	if (dwlengths[i] != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)dwlengths[i]); 
	    } 
	if (dwcounts[i] != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)dwcounts[i]); 
	    } 
 
	if (dw[i] != NULLEIENCODER) { 
	    eiFreeEncoder(dw[i]); 
	    } 
	} 
 
    for (i = 0; i < NDHTABLES; i++) { 
	if (dhlengths[i] != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)dhlengths[i]); 
	    } 
	if (dhcounts[i] != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)dhcounts[i]); 
	    } 
 
	if (dh[i] != NULLEIENCODER) { 
	    eiFreeEncoder(dh[i]); 
	    } 
	} 
 
    if (*dwnew && (*dwtable != (hufftable_spec *)NULL)) { 
	jhtFree(*dwtable, dgp_state); 
	} 
 
    if (*dhnew && (*dhtable != (hufftable_spec *)NULL)) { 
	jhtFree(*dhtable, dgp_state); 
	} 
 
    ERR_RETURN; 
    } 
 
static ErrCode 
writeMMRCollectiveBitmap(jb2File file, DataSink data, jbitfile bf, 
			 eiEncoder bmsize, jb2Symbol **symbols, 
			 Dgp_Int32 nsymbols) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("writeMMRCollectiveBitmap()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Int32 i; 
    Dgp_Int32 classwidth; 
    Dgp_Int32 classheight; 
    Dgp_UInt32 *classbits = (Dgp_UInt32 *)NULL; 
    size_t Dgp_Int8_row_width; 
    Dgp_Int32 row_width; 
    Dgp_UInt8 *encbuf = (Dgp_UInt8 *)NULL; 
    Dgp_Int32 spos; 
    size_t encbuflen, bytes; 
     
    DGP_ASSERT(nsymbols >= 1); 
 
    classheight = symbols[0]->height; 
 
    /* Figure out the total width of the class */ 
    classwidth = 0; 
    for (i = 0; i < nsymbols; i++) { 
	DGP_ASSERT(symbols[i]->height == classheight); 
	classwidth += symbols[i]->width; 
	} 
    /* How many Dgp_Int32s? */ 
    row_width = (classwidth + (DGP_BITS_UINT32 - 1)) / DGP_BITS_UINT32; 
    /* and how many bytes if I only pad to the next byte? */ 
    Dgp_Int8_row_width = (classwidth + (DGP_BITS_UINT8 - 1)) / DGP_BITS_UINT8; 
     
    /* Allocate the plonking place */ 
    classbits = (Dgp_UInt32 *)(a->calloc)((size_t)classheight * row_width, 
					sizeof(*classbits)); 
    if (classbits == (Dgp_UInt32 *)NULL) { 
      NOMEM; 
	} 
     
    /* AND start plonking */ 
    spos = 0; 
    for (i = 0; i < nsymbols; i++) { 
	placeSymbol(symbols[i], classbits, row_width, spos); 
 
	spos += symbols[i]->width; 
	} 
 
    /* 
     * Allocate a buffer to do the MMR encoding into.  Ideally, we'd like 
     * to do it straight into the DataSink, but that's not possible because 
     * we have to precede it with its length.  Therefore, we allocate a 
     * buffer and compress into that.  How large to make the buffer? 
     * Well, if the bitmap expands upon being fed through MMR, we're allowed 
     * to store it uncompressed.  Thus, we make the buffer exactly as large 
     * as the (row-Dgp_Int8-aligned) image, and try compressing into that. 
     * If that doesn't work, we put the image itself into the buffer, and 
     * flag a "data length" of zero. 
     */ 
    encbuflen = Dgp_Int8_row_width * classheight; 
    encbuf = (Dgp_UInt8 *)(a->malloc)(encbuflen); 
    if (encbuf == (Dgp_UInt8 *)NULL) { 
      NOMEM; 
	} 
 
    DGP_ERR_CHECK_NO_MSG(jb2MMREncode(classbits, classwidth, classheight, row_width, 
		      encbuf, &encbuflen, DGP_FALSE, dgp_state)) { 
	/* 
	 * Either it wouldn't fit, or there was another problem.  If it 
	 * wouldn't fit, we copy the bitmap into encbuf directly; if 
	 * there was another problem, well, that's probably better 
	 * than failing completely... 
	 */ 
      DGP_CLEAR_CUR_ERROR; 
 
	/* Encoded length of 0 means uncompressed */ 
	encbuflen = 0; 
 
	/* Copy it across, shuffling Dgp_Int32s as required */ 
	memcpy_shuffling(classbits, encbuf, 
			 row_width, Dgp_Int8_row_width, classheight); 
    } 
 
    /* 
     * We now have an encoded length in encbuflen and an encoded buffer 
     * in encbuf. 
     */ 
 
    /* Write the encoded length */ 
    WRITEV(bmsize, (int)encbuflen); 
 
    /* 
     * Now copy in the encoded buffer.  This is a bit tricky as we 
     * have to suspend the bitfile, mess with the DataSink, and resume 
     * the bitfile. 
     */ 
    DGP_ERR_PASS_CLEANUP(jbfSuspend(bf)); 
 
    /* Copy in the bits */ 
    if (encbuflen == 0) { 
	/* Need the real length */ 
	encbuflen = Dgp_Int8_row_width * classheight; 
	} 
 
    DSWRITE(data, encbuflen, encbuf, &bytes); 
 
    /* Resume the bitfile */ 
    jbfResume(bf); 
 
    /* OK to free the encoded and raw buffers now */ 
    (a->free)((void *)classbits); 
    (a->free)((void *)encbuf); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (classbits != (Dgp_UInt32 *)NULL) { 
	(a->free)((void *)classbits); 
	} 
 
    if (encbuf != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)encbuf); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Plonk the given token into the given bit array, offset in x by the given 
 * amount. The destination part of the bit array should have been cleared. 
 * The destination bitmap's height should match t's height. row_width is 
 * the width of each row of the destination bitmap in Dgp_Int32s. 
 * 
 * We assume that the token's bitmap is zero outside its borders (the 
 * padding bits are clear). 
 */ 
static void 
placeSymbol(jb2Symbol *sym, Dgp_UInt32 *bits, Dgp_Int32 row_width, Dgp_Int32 xpos) 
{ 
    Dgp_Int32 i, j; 
    Dgp_UInt32 tmp; 
    Dgp_UInt32 *p; 
    Dgp_UInt32 *q; 
    Dgp_Int32 Dgp_Int32pos; 
    Dgp_Int32 mrow_width; 
    Dgp_Int32 bitsinlastDgp_Int32; 
 
    /* How many Dgp_Int32s in the model actually contain data? */ 
    mrow_width = (sym->width + (DGP_BITS_UINT32 - 1)) / DGP_BITS_UINT32; 
    bitsinlastDgp_Int32 = DGP_BITS_UINT32 - ((mrow_width * DGP_BITS_UINT32) - sym->width); 
 
    Dgp_Int32pos = xpos / DGP_BITS_UINT32; 
    xpos = xpos % DGP_BITS_UINT32; 
     
    for (i = 0; i < sym->height; i++) { 
	p = sym->bits + i * sym->longsperline; 
	q = bits + Dgp_Int32pos + i * row_width; 
 
	/* Go through Dgp_Int32-by-Dgp_Int32 */ 
	if (xpos != 0) { 
	    for (j = 0; j < mrow_width - 1; j++) { 
		tmp = *p++; 
		*q++ |= tmp >> xpos; 
		*q |= tmp << (DGP_BITS_UINT32 - xpos); 
		} 
	    tmp = *p++; 
	    if (xpos + bitsinlastDgp_Int32 > DGP_BITS_UINT32) { 
		*q++ |= tmp >> xpos; 
		*q |= tmp << (DGP_BITS_UINT32 - xpos); 
		} 
	    else { 
		*q |= tmp >> xpos; 
		} 
	    } 
	else { 
	    /* No shifting - yay! */ 
	    for (j = 0; j < mrow_width; j++) { 
		*q++ |= *p++; 
		} 
	    } 
	} 
    } 
 
/* 
 * bits is an array of height rows of Dgp_Int32words, each row being row_width 
 * Dgp_Int32s wide.  We want to copy this, byte-by-byte, into *buf, copying 
 * only Dgp_Int8_row_width from each row. 
 */ 
static void 
memcpy_shuffling(Dgp_UInt32 *bits, Dgp_UInt8 *buf, 
		 Dgp_Int32 row_width, Dgp_Int32 Dgp_Int8_row_width, Dgp_Int32 height) 
{ 
  DGP_VOID_PROLOGUE("memcpy_shuffling()", NULL); 
 
    Dgp_Int32 x, y, x2; 
    Dgp_UInt32 tmp; 
    Dgp_UInt32 *p; 
    Dgp_UInt8 *q; 
 
    DGP_VOID_ASSERT((int)sizeof(Dgp_UInt32) * row_width >= Dgp_Int8_row_width); 
    DGP_VOID_ASSERT((int)sizeof(Dgp_UInt32) * (row_width - 1) < Dgp_Int8_row_width); 
 
    for (y = 0; y < height; y++) { 
		p = bits + y * row_width; 
		q = buf + y * Dgp_Int8_row_width; 
		for (x = 0; x < row_width - 1; x++) { 
		    tmp = *p++; 
			for (x2 = 0; x2 < (int)sizeof(Dgp_UInt32); x2++) { 
				*q++ = (Dgp_UInt8)(tmp >> (DGP_BITS_UINT32 - (x2 + 1) * DGP_BITS_UINT8)); 
			} 
		} 
		tmp = *p++; 
		for (x2 = 0; x2 < Dgp_Int8_row_width - (int)sizeof(Dgp_UInt32) * (row_width - 1); x2++) { 
			*q++ = (Dgp_UInt8)(tmp >> (DGP_BITS_UINT32 - (x2 + 1) * DGP_BITS_UINT8)); 
		} 
	} 
} 
