/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/encsymreg.c,v 1.28 2003/02/10 22:53:26 mkrocken Exp $ 
 * 
 */ 
 
 
#include "util.h" 
#include "hufftable.h" 
#include "encsymreg.h" 
#include "symbol.h" 
#include "segment.h" 
#include "stdhuff.h" 
#include "encint.h" 
#include "jbitfile.h" 
#include <stdlib.h> 
#include "hufftable_util.h" 
#include "enchufftable.h" 
#include "segment_util.h" 
#include "symenc.h" 
#include <string.h> 
#include "PsgFix.h"
 
#define DGP_ENC DGP_FALSE

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/encsymreg.c,v 1.28 2003/02/10 22:53:26 mkrocken Exp $"); 
 
/* Reference corners */ 
#define	BOTTOMLEFT	0 
#define	TOPLEFT		1 
#define	BOTTOMRIGHT	2 
#define	TOPRIGHT	3 
 
/* 
 * Arrays of possible Huffman tables 
 */ 
static hufftable_spec *dstables[] = { 
    &stdhuffDS1, &stdhuffDS2, &stdhuffDS3 
    }; 
#define	NDSTABLES	((Dgp_Int32)(sizeof(dstables) / sizeof(dstables[0]))) 
 
static hufftable_spec *fstables[] = { 
    &stdhuffFS1, &stdhuffFS2 
    }; 
#define	NFSTABLES	((Dgp_Int32)(sizeof(fstables) / sizeof(fstables[0]))) 
 
static hufftable_spec *dttables[] = { 
    &stdhuffDT1, &stdhuffDT2, &stdhuffDT3, &stdhuffDT3_neg 
    }; 
#define	NDTTABLES	((Dgp_Int32)(sizeof(dttables) / sizeof(dttables[0]))) 
#define	NEGDTTABLE	(NDTTABLES - 1) 
 
/* Info we need to construct per symbol */ 
typedef struct { 
    Dgp_Int32 count; 
    Dgp_Int32 symnum; 
    } symInfo; 
 
/* Internal routines */ 
static ErrCode decideSREncoding(jb2FileParams *params, jb2Segment *segment, 
				Dgp_Bool orderPreserved, 
				jb2SymbolInstance ***instptrs, 
				hufftable_spec **dttable, Dgp_Bool *dtnew, 
				hufftable_spec **dstable, Dgp_Bool *dsnew, 
				hufftable_spec **fstable, Dgp_Bool *fsnew, 
				Dgp_Int32 *stripsize, Dgp_Int32 *logstripsize, 
				Dgp_UInt8 *refcorner, 
				Dgp_Bool *transposed, 
				Dgp_Int32 *dsoffset, Dgp_Int32 *lowt, 
				Dgp_State *dgp_state); 
static ErrCode decideTransposed(jb2Segment *segment, Dgp_Bool *transposed, 
				Dgp_State *dgp_state); 
static ErrCode decideStrips(jb2Segment *segment, Dgp_Bool orderPreserved, 
			    Dgp_Bool transposed, 
			    Dgp_UInt8 refcorner, Dgp_Int32 *stripsize, 
			    Dgp_Int32 *logstripsize, Dgp_Int32 *modeds, Dgp_Int32 *lowt, 
			    Dgp_State *dgp_state); 
static Dgp_Int32 dsApproxBits(Dgp_Int32 val); 
static Dgp_Int32 dtApproxBits(Dgp_Int32 val); 
static Dgp_Int32 fsApproxBits(Dgp_Int32 val); 
static ErrCode sortInstances(jb2Segment *segment, 
			     Dgp_Bool orderPreserved, 
			     Dgp_Bool transposed, 
			     Dgp_UInt8 refcorner, 
			     Dgp_Int32 stripsize, Dgp_State *dgp_state, 
			     jb2SymbolInstance ***_sortedsyms); 
static ErrCode decideTables(jb2FileParams *params, 
			    jb2SymbolInstance **instptrs, Dgp_Int32 ninstances, 
			    hufftable_spec **dttable, Dgp_Bool *dtnew, 
			    hufftable_spec **dstable, Dgp_Bool *dsnew, 
			    hufftable_spec **fstable, Dgp_Bool *fsnew, 
			    Dgp_Int32 logstripsize, 
			    Dgp_UInt8 refcorner, 
			    Dgp_Bool transposed, Dgp_Int32 dsoffset, Dgp_Int32 lowt, 
			    Dgp_State *dgp_state); 
static ErrCode writeIDHuffTable(jb2SymRegSegment *symreg, jbitfile bf, 
				symInfo *syminfo, Dgp_Int32 nrefsyms, 
				symEncoder *symenc, 
				Dgp_State *dgp_state); 
 
#ifdef	VALPRINT 
 
#include <stdio.h> 
 
#define	WRITEV(coder, v) \
    do { \
	DGP_ERR_PASS_CLEANUP(eiEncode(coder, (v), DGP_FALSE)); \
	fprintf(stderr, #coder " %d\n", v); \
	} while (0) 
#define	WRITEOOB(coder) \
    do { \
	DGP_ERR_PASS_CLEANUP(eiEncode(coder, 0, DGP_TRUE)); \
	fprintf(stderr, #coder " OOB\n"); \
	} while (0) 
 
#else 
 
#define	WRITEV(coder, v) \
    do { \
	DGP_ERR_PASS_CLEANUP(eiEncode(coder, (v), DGP_FALSE)); \
	} while (0) 
#define	WRITEOOB(coder) \
    do { \
	DGP_ERR_PASS_CLEANUP(eiEncode(coder, 0, DGP_TRUE)); \
	} while (0) 
 
#endif 
 
/* Eventually we'll do refinement... */ 
 
/* 
 * Encode the symbol region, given the info stored in *segment and 
 * *refsegs.  Each refseg should be a symbol dictionary segment. 
 * 
 * In some cases, this routine will generate additional Huffman table 
 * segments.  In this case, the Huffman table segments will be written out 
 * to the jb2File before this segment.  This routine 
 * will call jb2AddReferencedSegments to add those segments to this 
 * segment's lists.  They will be flagged as "not retained beyond this 
 * segment". 
 * 
 * The fields in segment that must be filled in already are: 
 * - pagenum 
 * - nrefsegs 
 * - refsegs 
 * - refretain 
 * - data->symreg 
 * - reginfo (except for image fields) 
 * 
 * The remaining fields will be filled in as follows: 
 * - segnum will be allocated 
 * - datalen will be set based on the encoded segment's data length 
 * - retain will be set to DGP_FALSE 
 * - defer_nonretain will be set to DGP_FALSE 
 * 
 * The encoded segment header will be written out to 
 * jb2GetFileHeaderDataSink(file); the encoded segment data will be 
 * written out to jb2GetFileDataSink(file).  Neither of those will be 
 * examined until the Huffman table segments (if any) have been written 
 * out. 
 * 
 * This routine also can participate in building extensions.  Each 
 * jb2SymbolInstance can contain a void * pointer (userdata); there can 
 * also be an "extbuilder" routine passed in.  This routine gets called 
 * with an array of jb2SymbolInstance * pointers, pointing to the symbols 
 * in the order in which they will be written out.  Note that extbuilder 
 * might be called, only to have jb2EncodeSymRegion later fail. 
 * 
 * The order of the jb2Symbols in the symbol dictionary segments referenced 
 * by this segment MUST match the order that the symbols were written 
 * out by jb2EncodeSymDict.  
 * 
 * If the orderPreserved flag is DGP_TRUE, then the symbol instances will be 
 * written out in exactly the order that they appear in the segment. 
 * In some cases, depending on the coding options chosen for the jb2File 
 * (specifically srfirstonly, srllonly, srnontransonly), this may not be 
 * possible.  If this happens, the encoding will fail. 
 */ 
ErrCode 
jb2EncodeSymRegion(jb2File file, 
		   jb2Segment *segment, jb2Segment **refsegs, 
		   ErrCode (*extbuilder)(jb2SymbolInstance **sortedinsts, 
					 Dgp_Int32 ninsts, void *extcookie), 
		   void *extcookie, Dgp_Bool orderPreserved) 
{ 
  Dgp_State *dgp_state = jb2_pDgp_State(file); 
  DGP_PROLOGUE("jb2EncodeSymRegion()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    DataSink bodyds; 
    jbitfile bf = NULLJBITFILE, tmpbf; 
    DgpEncode_MQEncoder mqenc; 
    jb2FileParams *params = jb2GetFileParams(file); 
    jb2SymRegSegment *symreg; 
    jb2SymbolInstance **instptrs = (jb2SymbolInstance **)NULL; 
    hufftable_spec *dttable = (hufftable_spec *)NULL; 
    hufftable_spec *dstable = (hufftable_spec *)NULL; 
    hufftable_spec *fstable = (hufftable_spec *)NULL; 
    Dgp_Bool dtnew, dsnew, fsnew; 
    jb2Segment *newrefsegs[3]; 
    Dgp_Int32 newref; 
    Dgp_Int32 i, j, k; 
    jb2Segment dtseg, dsseg, fsseg; 
    eiEncoder dtei = NULLEIENCODER, dsei = NULLEIENCODER, 
	fsei = NULLEIENCODER, tei = NULLEIENCODER; 
    symEncoder symenc = (symEncoder)NULL; 
    Dgp_UInt8 refcorner; 
    Dgp_Bool transposed; 
    Dgp_Int32 stripsize; 
    Dgp_Int32 logstripsize; 
    Dgp_Int32 dsoffset; 
    Dgp_Int32 lowt; 
    void **userdata_copy = (void **)NULL; 
    Dgp_Bool userdata_copied = DGP_FALSE; 
    Dgp_Int32 orig_nrefsegs = segment->nrefsegs; 
    Dgp_Int32 nrefsyms; 
    symInfo *syminfo = (symInfo *)NULL; 
    jb2Segment *curseg; 
    Dgp_Int32 stript, firsts, curs; 
    Dgp_Int32 insts, instt, instsdist; 
    symInfo *curinfo; 
    /* Context buffers for the integer encoders */ 
    Dgp_UInt8 fscontexts[EICONTEXTSIZE]; 
    Dgp_UInt8 dscontexts[EICONTEXTSIZE]; 
    Dgp_UInt8 dtcontexts[EICONTEXTSIZE]; 
    Dgp_UInt8 tcontexts[EICONTEXTSIZE]; 
    size_t bytes; 
 
    Dgp_UInt16 sbuf; 
    Dgp_UInt8 lbuf[4]; 
    jb2SymbolInstance *curinst; 
     
    DGP_ASSERT((segment->segtype == JB2SEGTYPE_SYMREG_INT) || 
	   (segment->segtype == JB2SEGTYPE_SYMREG_IMM) || 
	   (segment->segtype == JB2SEGTYPE_SYMREG_IMMLL)); 
    symreg = &segment->data.symreg; 
    /* 
     * First, decide how we're going to do the encoding: which 
     * Huffman tables, and so on.  Create any Huffman table 
     * segments if desired. 
     * Second, write out the segment header (has to come after the 
     * routine has a chance to create the Huffman tables and their 
     * associated segments). 
     * Third, write out the data header describing the encoding choices. 
     * Finally, write the symbol region itself. 
     */ 
 
    /* 
     * The first step is to decide on the Huffman tables, strip size, 
     * reference corner, and so on.  In order to do this, we need 
     * to have sorted the symbols into the various possibilities, and 
     * tried each of them.  Whee! 
     */ 
    DGP_ERR_PASS_CLEANUP(decideSREncoding(params, segment, orderPreserved, &instptrs, 
			  &dttable, &dtnew, &dstable, &dsnew, 
			  &fstable, &fsnew, 
			  &stripsize, &logstripsize, 
			&refcorner, &transposed, &dsoffset, &lowt, dgp_state)); 
 
    /* At this point, we can call the extbuilder. */ 
    if (extbuilder != (ErrCode (*)(jb2SymbolInstance **, Dgp_Int32, void *))NULL) { 
      DGP_ERR_PASS_CLEANUP((*extbuilder)(instptrs, symreg->ninstances, 
					 extcookie)); 
	} 
 
    if (!params->srarith) { 
	/* 
	 * For each new Huffman table, write out a segment for it. 
	 * Add the segment pointers to newrefsegs as we go.  The order to 
	 * add them in MUST be FS, then DS, then DT. 
	 */ 
	newref = 0; 
 
	if (fsnew) { 
	    /* Fill in the required fields */ 
	    fsseg.pagenum = segment->pagenum; 
	    fsseg.data.table.table = *fstable; 
 
	    /* and write out the segment */ 
	    DGP_ERR_PASS_CLEANUP(jb2EncodeHuffTable(file, &fsseg)); 
 
	    newrefsegs[newref++] = &fsseg; 
	    /* That didn't allocate anything - nothing to free */ 
	    } 
     
	if (dsnew) { 
	    dsseg.pagenum = segment->pagenum; 
	    dsseg.data.table.table = *dstable; 
	    DGP_ERR_PASS_CLEANUP(jb2EncodeHuffTable(file, &dsseg)); 
	    newrefsegs[newref++] = &dsseg; 
	    } 
     
	if (dtnew) { 
	    dtseg.pagenum = segment->pagenum; 
	    dtseg.data.table.table = *dttable; 
	    DGP_ERR_PASS_CLEANUP(jb2EncodeHuffTable(file, &dtseg)); 
	    newrefsegs[newref++] = &dtseg; 
	    } 
     
	/* 
	 * Add the new segment references to this segment's information 
	 */ 
	if (newref > 0) { 
	    DGP_ERR_PASS_CLEANUP(jb2AddReferencedSegments(file, segment, 
							  newrefsegs, newref)); 
	    /* Retain bits already filled with NULL */ 
	    } 
	} 
    /* Time to start encoding. */ 
 
    /* Write out the segment header */ 
 
    /* Fill in the fields we promised to fill in */ 
    segment->segnum = jb2FileNextSegmentNumber(file); 
    segment->retain = DGP_FALSE; 
    segment->defer_nonretain = DGP_FALSE; 
 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeader(file, segment)); 
 
    /* Write out the segment data header */ 
    bodyds = jb2GetFileDataSink(file); 
    /* Reset the count on bodyds so we can get the data length */ 
    dsResetCount(bodyds); 
 
    /* First the region segment data header */ 
    DGP_ERR_PASS_CLEANUP(jb2EncodeRegSegment(file, bodyds, 
					     &segment->reginfo)); 
     
    /* Now the symbol region segment data header */ 
 
    /* Two byte flags field */ 
    sbuf = 0; 
 
    /* Bit 0: SBHUFF */ 
    if (!params->srarith) { 
	sbuf |= 1; 
	} 
#ifdef	NOMQ 
    else { 
      FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
	} 
#endif 
 
    /* Bit 1: SBREFINE */ 
    /* always 0 currently */ 
 
    /* Bits 2-3: LOGSBSTRIPS */ 
    sbuf |= (logstripsize << 2); 
#ifdef	VALPRINT 
    fprintf(stderr, "SBLOGSTRIPSIZE %d\n", logstripsize); 
#endif 
 
    /* Bits 4-5: REFCORNER */ 
    sbuf |= (refcorner << 4); 
#ifdef	VALPRINT 
    fprintf(stderr, "SBREFCORNER %d\n", refcorner); 
#endif 
 
    /* Bit 6: TRANSPOSED */ 
    sbuf |= (transposed << 6); 
#ifdef	VALPRINT 
    fprintf(stderr, "SBTRANSPOSED %d\n", transposed); 
#endif 
 
    /* Bits 7-8: SBCOMBOP */ 
    /* always OR currently */ 
    sbuf |= COMBOP_OR << 7; 
 
    /* Bit 9: SDBEFPIXEL */ 
    /* always 0 currently */ 
 
    /* Bits 10-14: SBDSOFFSET */ 
    sbuf |= (dsoffset & 31) << 10; 
#ifdef	VALPRINT 
    fprintf(stderr, "SBDSOFFSET %d\n", dsoffset); 
#endif 
 
    /* Bit 15: SBRTEMPLATE */ 
    /* always 0 currently */ 
 
#ifdef	VALPRINT 
    fprintf(stderr, "SYMREGFLAGS %04x\n", sbuf); 
#endif 
    to2byte(lbuf, sbuf); 
    DSWRITE(bodyds, 2, lbuf, &bytes); 
 
    if (!params->srarith) { 
	/* Two byte Huffman flags field */ 
	sbuf = 0; 
 
	/* Bits 0-1: SBHUFFFS selection */ 
	if (fstable == &stdhuffFS1) { 
	    /* value of 0 */ 
	    } 
	else if (fstable == &stdhuffFS2) { 
	    /* value of 1 */ 
	    sbuf |= 1; 
	    } 
	else { 
	    /* custom table; value of 3 */ 
	    sbuf |= 3; 
	    } 
 
	/* Bits 2-3: SBHUFFDS selection */ 
	if (dstable == &stdhuffDS1) { 
	    /* value of 0 */ 
	    } 
	else if (dstable == &stdhuffDS2) { 
	    /* value of 1 */ 
	    sbuf |= (1 << 2); 
	    } 
	else if (dstable == &stdhuffDS3) { 
	    /* value of 2 */ 
	    sbuf |= (2 << 2); 
	    } 
	else { 
	    /* custom table; value of 3 */ 
	    sbuf |= (3 << 2); 
	    } 
 
	/* Bits 4-5: SBHUFFDT selection */ 
	if (dttable == &stdhuffDT1) { 
	    /* value of 0 */ 
	    } 
	else if (dttable == &stdhuffDT2) { 
	    /* value of 1 */ 
	    sbuf |= (1 << 4); 
	    } 
	else if (dttable == &stdhuffDT3) { 
	    /* value of 2 */ 
	    sbuf |= (2 << 4); 
	    } 
	else { 
	    /* custom table; value of 3 */ 
	    sbuf |= (3 << 4); 
	    } 
 
	/* Bits 6-7: SBHUFFRDW selection */ 
	/* always 0 currently */ 
 
	/* Bits 8-9: SBHUFFRDH selection */ 
	/* always 0 currently */ 
 
	/* Bits 10-11: SBHUFFRDX selection */ 
	/* always 0 currently */ 
 
	/* Bits 12-13: SBHUFFRDY selection */ 
	/* always 0 currently */ 
 
	/* Bit 13: SBHUFFRSIZE selection */ 
	/* always 0 currently */ 
 
#ifdef	VALPRINT 
	fprintf(stderr, "SYMREGHUFFFLAGS %04x\n", sbuf); 
#endif 
	to2byte(lbuf, sbuf); 
	DSWRITE(bodyds, 2, lbuf, &bytes); 
	} 
 
    /* no refinement AT flags because SBREFINE=0 */ 
 
    /* SBNUMINSTANCES - 4 bytes */ 
#ifdef	VALPRINT 
    fprintf(stderr, "SBNUMINSTANCES %d\n", symreg->ninstances); 
#endif 
    to4byte(lbuf, symreg->ninstances); 
    DSWRITE(bodyds, 4, lbuf, &bytes); 
 
    /* 
     * For the rest of the routine, we're going to hijack the 
     * userdata pointer of each of the symbols in the dictionaries 
     * referred to by this segment.  We'll use that pointer to point 
     * to a structure containing the Huffman code corresponding to that 
     * symbol, and some other useful data. 
     * 
     * Of course, we have to put the original userdata pointers back 
     * when we're done. 
     */ 
    /* How many symbols are referred to?  Check only the original refsegs */ 
    nrefsyms = 0; 
    for (i = 0; i < orig_nrefsegs; i++) { 
	if (refsegs[i]->segtype == JB2SEGTYPE_SYMDICT) { 
	    nrefsyms += refsegs[i]->data.symdict.nsymbols; 
	    } 
	} 
 
    /* Allocate a place to put all the copied userdata pointers */ 
    userdata_copy = (void **)(a->malloc)(nrefsyms * sizeof(*userdata_copy)); 
    DGP_CHECK_ALLOC_CLEANUP(userdata_copy); 
 
    /* and allocate our structures */ 
    syminfo = (symInfo *)(a->calloc)(nrefsyms, sizeof(*syminfo)); 
    DGP_CHECK_ALLOC_CLEANUP(syminfo); 
 
    /* 
     * Copy them in order; set the pointer to our structure as we copy 
     * Also set the symnum field of the syminfo structures. 
     */ 
    k = 0; 
    for (i = 0; i < orig_nrefsegs; i++) { 
	curseg = refsegs[i]; 
	if (curseg->segtype == JB2SEGTYPE_SYMDICT) { 
	    for (j = 0; j < curseg->data.symdict.nsymbols; j++, k++) { 
		userdata_copy[k] = curseg->data.symdict.symbols[j].userdata; 
		curseg->data.symdict.symbols[j].userdata = &syminfo[k]; 
		syminfo[k].symnum = k; 
		} 
	    } 
	} 
 
    /* The copy is complete.  Set the flag so we know to undo it on bailout */ 
    userdata_copied = DGP_TRUE; 
 
    /* Finally, we can actually start writing */ 
 
    /* Get all the stuff we need: the bitfile/DgpEncode_MQEncoder and the eiEncoders */ 
 
    if (params->srarith) { 
#ifndef	NOMQ 
	dsInitDgpEncode_MQEncoder(bodyds, &mqenc); 
 
	/* Only arith needs a specific T encoder */ 
	memset(tcontexts, 0, EICONTEXTSIZE); 
	DGP_ERR_PASS_CLEANUP(eiNewEncoderMQ(&mqenc, tcontexts, dgp_state, 
					    &tei)); 
#endif 
	} 
    else { 
      DGP_ERR_PASS_CLEANUP(jbfOpenFromDataSink(bodyds, &bf)); 
	} 
 
    DGP_ERR_PASS_CLEANUP(eiNewEncoderEither((Dgp_Bool)!params->srarith, dttable, bf, 
			      &mqenc, dtcontexts, dgp_state, &dtei)); 
 
    DGP_ERR_PASS_CLEANUP(eiNewEncoderEither((Dgp_Bool)!params->srarith, dstable, bf, 
			      &mqenc, dscontexts, dgp_state, &dsei)); 
 
    DGP_ERR_PASS_CLEANUP(eiNewEncoderEither((Dgp_Bool)!params->srarith, fstable, bf, 
			      &mqenc, fscontexts, dgp_state, &fsei)); 
 
    /* Now the symbol ID encoder */ 
    if (params->srarith) { 
#ifndef	NOMQ 
	DGP_ERR_PASS_CLEANUP(symEncoderArith(nrefsyms, &mqenc, dgp_state, 
					     &symenc)); 
#else 
	FILEERR(cErrCompressionNotSupported, "no MQ coder available"); 
#endif 
	} 
    else { 
	/* Create and write the Huffman table */ 
      DGP_ERR_PASS_CLEANUP(writeIDHuffTable(symreg, bf, syminfo, nrefsyms, &symenc, dgp_state)); 
	} 
 
    /* Pick the initial strip */ 
    stript = lowt << logstripsize; 
    firsts = 0; 
    curs = 0; 
 
    /* and encode it */ 
    WRITEV(dtei, -lowt); 
 
    for (i = 0; i < symreg->ninstances; i++) { 
	curinst = instptrs[i]; 
 
	/* 
	 * Get S and T based on refcorner and transposed. 
	 * The instance (xpos,ypos) is always top right pixel. 
	 */ 
	if (transposed) { 
	    if ((refcorner == TOPLEFT) || (refcorner == BOTTOMLEFT)) { 
		/* 
		 * Why is BOTTOMLEFT doing the same thing as TOPLEFT? 
		 * Well, when you decode BOTTOMLEFT, 
		 * there's this odd pre-positioning update that gets done: 
		 * the S position decoded 
		 * gets H-1 added to it before it is used for positioning. 
		 * Therefore, the S position we use for encoding is really 
		 * just ypos (since the H-1 addition to shift it to the 
		 * bottom happens in the decoder). 
		 */ 
		insts = curinst->ypos; 
		instt = curinst->xpos; 
		} 
	    else { 
		/* TOPRIGHT and BOTTOMRIGHT can share code for the same reason */ 
		insts = curinst->ypos; 
		instt = curinst->xpos + curinst->sym->width - 1; 
		} 
	    instsdist = curinst->sym->height; 
	    } 
	else { 
	    if ((refcorner == TOPLEFT) || (refcorner == TOPRIGHT)) { 
		insts = curinst->xpos; 
		instt = curinst->ypos; 
		} 
	    else { 
		insts = curinst->xpos; 
		instt = curinst->ypos + curinst->sym->height - 1; 
		} 
	    instsdist = curinst->sym->width; 
	    } 
 
	/* Got S and T - handle the strip state mechanism. */ 
	if ((i == 0) || ((instt >> logstripsize) != (stript >> logstripsize))) { 
	    /* If it's not the first strip, emit an OOB for DS */ 
	    if (i != 0) { 
		WRITEOOB(dsei); 
		} 
 
	    /* Start of a new strip - do the delta T coding */ 
	    WRITEV(dtei, (instt - stript) >> logstripsize); 
	    /* and the first S coding */ 
	    WRITEV(fsei, insts - firsts); 
 
	    /* Update stript and firsts */ 
 
	    /* stript must always be the top row of the strip */ 
	    stript = (instt >> logstripsize) << logstripsize; 
	    firsts = insts; 
	    } 
	else { 
	    /* Within a strip - just encode a DS */ 
	    WRITEV(dsei, insts - curs); 
	    } 
 
	/* Write the T-within-strip value */ 
	if (params->srarith) { 
#ifndef	NOMQ 
	    if (logstripsize > 0) { 
		WRITEV(tei, instt - stript); 
		} 
#endif 
	    } 
	else { 
	    if (logstripsize > 0) { 
		DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, instt - stript, logstripsize)); 
#ifdef	VALPRINT 
		fprintf(stderr, "t %d\n", instt - stript); 
#endif 
		} 
	    } 
 
	/* Write the symbol ID */ 
	curinfo = (symInfo *)curinst->sym->userdata; 
	DGP_ERR_PASS_CLEANUP(symEncode(symenc, curinfo->symnum)); 
#ifdef	VALPRINT 
	fprintf(stderr, "sn %d\n", curinfo->symnum); 
#endif 
 
	/* Update the current S position; include all the offsets here */ 
	curs = insts + instsdist - 1 + dsoffset; 
	} 
 
    /* Write an OOB to end the last strip */ 
    if (symreg->ninstances != 0) { 
	/* There were some instances */ 
	WRITEOOB(dsei); 
	} 
 
    /* Done! */ 
    if (params->srarith) { 
#ifndef NOMQ 
      DGP_ERR_PASS_CLEANUP(dsFinalDgpEncode_MQEncoder(bodyds, &mqenc)); 
#endif 
	} 
    else { 
      tmpbf = bf; 
      bf = NULLJBITFILE; 
      DGP_ERR_PASS_CLEANUP(jbfCloseFromDataSink(tmpbf)); 
	} 
 
    (a->free)((void *)instptrs); 
 
    /* 
     * Check the DataSink's running count to find how many bytes that 
     * all took, and write that to the header field 
     */ 
    segment->datalen = dsCount(bodyds); 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeaderDataLength(file, segment)); 
 
    eiFreeEncoder(dtei); 
    eiFreeEncoder(dsei); 
    eiFreeEncoder(fsei); 
 
    if (tei != NULLEIENCODER) { 
	eiFreeEncoder(tei); 
	} 
 
    /* Restore the userdata values */ 
    k = 0; 
    for (i = 0; i < orig_nrefsegs; i++) { 
	curseg = refsegs[i]; 
	if (curseg->segtype == JB2SEGTYPE_SYMDICT) { 
	    for (j = 0; j < curseg->data.symdict.nsymbols; j++, k++) { 
		curseg->data.symdict.symbols[j].userdata = userdata_copy[k]; 
		} 
	    } 
	} 
 
    /* free the copied values */ 
    (a->free)((void *)userdata_copy); 
    /* and the symInfos */ 
    (a->free)((void *)syminfo); 
 
    /* Free the new Huffman tables */ 
    if (dtnew) { 
	jhtFree(dttable, dgp_state); 
	} 
    if (dsnew) { 
	jhtFree(dstable, dgp_state); 
	} 
    if (fsnew) { 
	jhtFree(fstable, dgp_state); 
	} 
    symEncoderFree(symenc); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (bf != NULLJBITFILE) { 
	(void)jbfCloseFromDataSink(bf); 
	} 
    if (dtei != NULLEIENCODER) { 
	eiFreeEncoder(dtei); 
	} 
    if (dsei != NULLEIENCODER) { 
	eiFreeEncoder(dsei); 
	} 
    if (fsei != NULLEIENCODER) { 
	eiFreeEncoder(fsei); 
	} 
    if (tei != NULLEIENCODER) { 
	eiFreeEncoder(tei); 
	} 
    if ((dttable != (hufftable_spec *)NULL) && (dtnew)) { 
	jhtFree(dttable, dgp_state); 
	} 
    if ((dstable != (hufftable_spec *)NULL) && (dsnew)) { 
	jhtFree(dstable, dgp_state); 
	} 
    if ((fstable != (hufftable_spec *)NULL) && (fsnew)) { 
	jhtFree(fstable, dgp_state); 
	} 
 
    if (userdata_copied) { 
	/* Need to copy the userdata fields back */ 
	DGP_ASSERT(userdata_copy != (void **)NULL); 
	k = 0; 
	for (i = 0; i < orig_nrefsegs; i++) { 
	    curseg = refsegs[i]; 
	    if (curseg->segtype == JB2SEGTYPE_SYMDICT) { 
		for (j = 0; j < curseg->data.symdict.nsymbols; j++, k++) { 
		    curseg->data.symdict.symbols[j].userdata = userdata_copy[k]; 
		    } 
		} 
	    } 
	} 
 
    if (userdata_copy != (void **)NULL) { 
	(a->free)((void *)userdata_copy); 
	} 
 
    if (syminfo != (symInfo *)NULL) { 
	(a->free)((void *)syminfo); 
	} 
 
    if (symenc != (symEncoder)NULL) { 
	symEncoderFree(symenc); 
	} 
 
    if (instptrs != (jb2SymbolInstance **)NULL) { 
	(a->free)((void *)instptrs); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Make a lot of encoding choices.  Choose the strip size, reference 
 * corner, and transposition flag.  Pick (or build) Huffman tables for 
 * DT, DS and FS.  Fill in *instptrs with a pointer to an array of 
 * jb2SymbolInstance *, containing the instances in the order chosen. 
 * Also, fill in *lowt with the smallest T value seen, shifted right 
 * by *logstripsize, or -1, whichever is smaller (i.e., return -1 if 
 * the smallest T value is >= 0, or the smallest T value if it is < 0). 
 * 
 * If orderPreserved is DGP_TRUE, then whatever encoding we choose 
 * isn't allowed to change the order of the symbol instances: they have to 
 * be written into the file in the exact order that they appear 
 * in *segment. 
 */ 
static ErrCode 
decideSREncoding(jb2FileParams *params, 
		 jb2Segment *segment, Dgp_Bool orderPreserved, 
		 jb2SymbolInstance ***instptrs, 
		 hufftable_spec **dttable, Dgp_Bool *dtnew, 
		 hufftable_spec **dstable, Dgp_Bool *dsnew, 
		 hufftable_spec **fstable, Dgp_Bool *fsnew, 
		 Dgp_Int32 *stripsize, Dgp_Int32 *logstripsize, 
		 Dgp_UInt8 *refcorner, Dgp_Bool *transposed, 
		 Dgp_Int32 *dsoffset, Dgp_Int32 *lowt, Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("decideSREncoding()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    /* 
     * This is going to be fun.  There are four possible reference 
     * corners, four possible strip sizes, and a transposed/not transposed 
     * flag.  There are also 3 standard DT tables plus the option of a 
     * custom table; 3 standard DS tables + custom; 2 standard FS tables + 
     * custom.  We also need to pick an offset for DS and the starting 
     * strip T value.  Not to mention that in some cases, we need to 
     * choose all of these without affecting the order of the symbol 
     * instances. 
     * 
     * How to evaluate all of these choices?  Well, given a setting 
     * for strip size, transposed, refcorner, we can evaluate all the 
     * tables, plus the custom table options, in a single pass through 
     * the data.  We could run through all 32 possible choices for 
     * those settings, but each one requires a sort, which is expensive. 
     * 
     * Instead, we're going to cut the choices down as follows: 
     * - if transposed is DGP_FALSE, choose BOTTOMLEFT as the reference corner 
     * - if transposed is DGP_TRUE, choose TOPLEFT as the reference corner 
     * Choose "transposed" based on a projection histogram of the 
     * lower left corners (rationale: for the horizontal projection, 
     * the bottom edge matters; for the vertical projection, the 
     * left edge matters).  The projection histogram with the 
     * highest peak determines the transposition direction: highest peak in 
     * the vertical projection histogram means "transposed"; otherwise, 
     * "not transposed". 
     * 
     * These heuristics assume that the bitmap 
     * we're representing is in "paper order": the X direction runs 
     * physically rightwards (English text reading direction) 
     * and the Y direction runs physically downwards. 
     * If that isn't the case, then the choice of reference corner might 
     * not be optimal.  However, these heuristics work for English text 
     * and vertically-oriented Chinese or Japanese text. 
     * 
     * Once we've chosen "transposed", we now just have to evaluate the 
     * strip size.  We do this by sorting the symbol instances into strips 
     * of size 8 (the maximum).  We then run through the strips; each 
     * strip gets run through four different times, being divided up into 
     * sub-strips of size 1, 2, 4 or 8.  For each of those, a simple 
     * pseudo-table is used to count the bits: 
     * - for DS, 0-3 cost 4 bits each; 4-128 and -1 - -4 
     *   cost 8 bits each; other values cost 12 bits each.  OOB costs 2 bits. 
     * - for DT, 1 costs 1 bit, 2...3 costs 3 bits, 4-8 cost 5 bits; other 
     *   values cost 8 bits 
     * - for FS, -32 - 32 cost 8 bits; other values cost 12 bits. 
     * - T values cost 0, 1, 2 or 3 bits, depending on the strip size. 
     * These are approximations, but it's cheaper to do this than to try 
     * all the Huffman tables at each strip size.  And if it's using arithmetic 
     * coding, it's not easy to measure the #bits needed Dgp_Int16 of encoding 
     * everything in all possible ways and picking the Dgp_Int16est. 
     * 
     * Finally, the strip size that comes out cheapest is chosen; the 
     * Huffman tables for that strip size are all evaluated, and 
     * one of them (or a custom one) is chosen. 
     * 
     * dsoffset throws a slight monkey wrench into that, but we can make 
     * the assumption that the modal delta s value is the same no matter 
     * which strip size we choose, so we can pick it based on the size-8 
     * strips, and use it everywhere. 
     * 
     * OK, you say, but what if orderPreserved is DGP_TRUE?  Surely, then, we 
     * can't choose a strip size or transposition... 
     * Well, there's a few things going on here: presumably the encoder 
     * actually put these things in some rational order.  For the DigiPaper 
     * encoder, that order is likely to be sorted in reading order by the 
     * top left corner.  So we can accomodate that by setting transposed 
     * to DGP_FALSE, stripsize to 1, and the reference corner to TOPLEFT, then 
     * encoding in the order that the instances appear.  No problem, right? 
     * But what if that wasn't the order the encoder had in mind, or what 
     * if there were exceptions to that order?  Well, in that case we 
     * need to do something slightly tricky: it's possible to encode 
     * negative values for delta S and delta T.  Negative delta S is handled 
     * by the standard tables, but negative delta T isn't: you absolutely 
     * need a custom table.  Our current scheme only allows us to generate 
     * a custom table by tuning a pre-existing table.  The table 
     * stdhuffDT3_neg has been created for precisely this purpose: it 
     * was created by slightly modifying the DT3 table. 
     * 
     * We could do something fancier if orderPreserved==DGP_TRUE, involving 
     * trying all the possibilities we normally try, and then tracking 
     * which of those possibilities violate the requirement that the 
     * order must be preserved.  However, that's a fair amount of 
     * bookkeeping, and furthermore the possibility that is most likely 
     * with the current code base (instances are sorted by top left corner) 
     * isn't one of the ones we normally try. 
     */ 
 
    /* clear for predictable bailout */ 
    *instptrs = (jb2SymbolInstance **)NULL; 
 
    /* Decide on the transposition */ 
    if (params->srnontransonly || orderPreserved) { 
	/* It's forced - non-transposed only */ 
	*transposed = DGP_FALSE; 
	} 
    else { 
      DGP_ERR_PASS_CLEANUP(decideTransposed(segment, transposed, dgp_state)); 
	} 
 
    /* Decide on the reference corner */ 
    if (*transposed) { 
	DGP_ASSERT(!params->srllonly); /* We can't handle transposed and forced-LL */ 
	*refcorner = TOPLEFT; 
	} 
    else { 
	if (orderPreserved) { 
	  DGP_ERR_GEN_CLEANUP((params->srllonly), cErrInvalidParm); 
	   /* *refcorner = TOPLEFT;  */
      *refcorner = TOPLEFT;
	    } 
	else { 
	    *refcorner = BOTTOMLEFT; 
	    } 
	} 
 
    /* 
     * Now run through the strips, counting (approximate) bits for each one. 
     * Decide on the strip size and dsoffset. 
     */ 
    DGP_ERR_PASS_CLEANUP(decideStrips(segment, orderPreserved, *transposed, *refcorner, 
				      stripsize, logstripsize, dsoffset, lowt, dgp_state)); 
 
    /* 
     * Yay - we now have the information we need!  Use stripsize, 
     * transposed and refcorner to sort the instances. 
     */ 
    DGP_ERR_PASS_CLEANUP(sortInstances(segment, orderPreserved, *transposed, 
			  *refcorner, *stripsize, dgp_state, instptrs)); 
    /* 
     * *instptrs is now an array of pointers to the instances, in 
     * sorted order.  Decide (for real) on the Huffman tables. 
     */ 
    if (!params->srarith) { 
	DGP_ERR_PASS_CLEANUP(decideTables(params, *instptrs, 
			  segment->data.symreg.ninstances, dttable, dtnew, 
			  dstable, dsnew, fstable, fsnew, 
			  *logstripsize, *refcorner, 
			  *transposed, *dsoffset, *lowt, dgp_state)); 
	} 
    else { 
	/* Guaranteed there are no new Huffman tables */ 
	*dtnew = *dsnew = *fsnew = DGP_FALSE; 
	} 
 
    /* Everything is decided now. */ 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    /* Only thing that might need freeing is *instptrs */ 
    if (*instptrs != (jb2SymbolInstance **)NULL) { 
	(a->free)((void *)*instptrs); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Decide if the symbols here should be transposed or not.  Use 
 * the projection histogram described above. 
 */ 
static ErrCode 
decideTransposed(jb2Segment *segment, Dgp_Bool *transposed, 
		 Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("decideTransposed()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    jb2SymRegSegment *symreg = &segment->data.symreg; 
    Dgp_Int32 *verthist = (Dgp_Int32 *)NULL; 
    Dgp_Int32 *horizhist = (Dgp_Int32 *)NULL; 
    Dgp_Int32 i; 
    jb2SymbolInstance *curinst; 
    Dgp_Int32 curx, cury, maxv, maxh; 
 
    /* Do the projection histogram */ 
 
    /* Allocate and clear them */ 
    verthist = (Dgp_Int32 *)(a->calloc)((size_t)segment->reginfo.regwidth, 
			     sizeof(*verthist)); 
    DGP_CHECK_ALLOC_CLEANUP(verthist); 
 
    horizhist = (Dgp_Int32 *)(a->calloc)((size_t)segment->reginfo.regheight, 
			     sizeof(*horizhist)); 
    DGP_CHECK_ALLOC_CLEANUP(horizhist); 
 
    /* Run through and count */ 
    for (i = 0; i < symreg->ninstances; i++) { 
	curinst = &symreg->instances[i]; 
 
	/* Ignore out-of-bounds entries */ 
	curx = curinst->xpos; 
	if ((curx >= 0) && (curx < (Dgp_Int32)segment->reginfo.regwidth)) { 
	    verthist[curx]++; 
	    } 
 
	/* Offset to lower left */ 
	cury = curinst->ypos + curinst->sym->height - 1; 
	if ((cury >= 0) && (cury < (Dgp_Int32)segment->reginfo.regheight)) { 
	    horizhist[cury]++; 
	    } 
	} 
 
    maxv = maxh = 0; 
    for (i = 0; i < (Dgp_Int32)segment->reginfo.regwidth; i++) { 
	maxv = MAX(maxv, verthist[i]); 
	} 
    for (i = 0; i < (Dgp_Int32)segment->reginfo.regheight; i++) { 
	maxh = MAX(maxh, horizhist[i]); 
	} 
 
    if (maxv > maxh) { 
	/* The largest vertical projection was larger - choose transposed */ 
	*transposed = DGP_TRUE; 
	} 
    else { 
	*transposed = DGP_FALSE; 
	} 
 
    /* No Dgp_Int32er need the histograms */ 
    (a->free)((void *)verthist); 
    (a->free)((void *)horizhist); 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (verthist != (Dgp_Int32 *)NULL) { 
	(a->free)((void *)verthist); 
	} 
 
    if (horizhist != (Dgp_Int32 *)NULL) { 
	(a->free)((void *)horizhist); 
	} 
    ERR_RETURN; 
    } 
 
/* 
 * Abstracted information about a symbol instance: where it is; how 
 * far in the S direction the thing it's representing extends. 
 */ 
typedef struct { 
    Dgp_Int32 s, t, sdist; 
    } instInfo; 
 
/* 
 * Compare two instInfo pointers.  Return 1 if p1 is greater than p2, 
 * according to: 
 * - p1 lies in a later strip (larger t) than p2 
 * - p1 lies in the same strip as p2, but to its right (larger s) 
 * - p1 lies in the same strip, at the same location as p2, but the 
 *   sdist is greater (slight hack, and not always stable) 
 * Otherwise return -1. 
 * 
 * Use a strip size of 8 always. 
 */ 
static int
instCmp(const void *p1, const void *p2) 
{ 
    const instInfo *i1 = (const instInfo *)p1; 
    const instInfo *i2 = (const instInfo *)p2; 
    Dgp_Int32 t1 = i1->t >> 3;	/* 
				 * Shift right is correct and divide by 8 
				 * isn't: think negative coordinates; 
				 * coords -8...-1 should make up a single 
				 * strip. 
				 */ 
    Dgp_Int32 t2 = i2->t >> 3; 
 
    if ((t1 > t2) || 
	((t1 == t2) && (i1->s > i2->s)) || 
	((t1 == t2) && (i1->s == i2->s) && (i1->sdist > i2->sdist))) { 
	return(1); 
	} 
    else { 
	return(-1); 
	} 
    } 
 
/* 
 * Decide on the strip size (1, 2, 4 or 8), using the approximate bit 
 * counting method described above.  Also decide and report the modal 
 * delta S value and smallest negative T value. 
 * 
 * If we've been told to preserve the order, then we just compute 
 * the delta S value and smallest negative T value; the strip size 
 * is forced to 1. 
 */ 
static ErrCode 
decideStrips(jb2Segment *segment, Dgp_Bool orderPreserved, Dgp_Bool transposed, 
	     Dgp_UInt8 refcorner, Dgp_Int32 *stripsize, Dgp_Int32 *logstripsize, 
	     Dgp_Int32 *modeds, Dgp_Int32 *lowt, Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("decideStrips()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    instInfo *sortinst = (instInfo *)NULL; 
    Dgp_Int32 count[4]; 
    Dgp_Int32 firsts[4][8], oldt[4], curs[4][8], lastfirsts[4]; 
    Dgp_Bool empty[4][8]; 
    Dgp_Int32 curstrip, curt; 
    Dgp_Int32 i, j, k; 
    Dgp_Int32 dshist[32]; 
    Dgp_Int32 ds, dsmax; 
    Dgp_Int32 mincount; 
    instInfo *curinst, *previnst; 
    jb2SymbolInstance *curSymInst; 
    jb2SymRegSegment *symreg = &segment->data.symreg; 
 
    /* Build the instInfo array */ 
    sortinst = (instInfo *)(a->malloc)(symreg->ninstances * sizeof(*sortinst)); 
    DGP_CHECK_ALLOC_CLEANUP(sortinst); 
 
    /* Read out the instances. Do the lowt picking now. */ 
    *lowt = 0;		/* Never report anything positive */ 
    if (transposed) { 
	DGP_ASSERT(refcorner == TOPLEFT); 
	for (i = 0; i < symreg->ninstances; i++) { 
	    curSymInst = &symreg->instances[i]; 
	    sortinst[i].s = curSymInst->ypos; 
	    sortinst[i].t = curSymInst->xpos; 
	    sortinst[i].sdist = curSymInst->sym->height; 
	    *lowt = MIN(*lowt, sortinst[i].t); 
	    } 
	} 
    else { 
	if (refcorner == BOTTOMLEFT) { 
	    for (i = 0; i < symreg->ninstances; i++) { 
		curSymInst = &symreg->instances[i]; 
		sortinst[i].s = curSymInst->xpos; 
		sortinst[i].t = curSymInst->ypos + curSymInst->sym->height - 1; 
		sortinst[i].sdist = curSymInst->sym->width; 
		*lowt = MIN(*lowt, sortinst[i].t); 
		} 
	    } 
	else { 
	    DGP_ASSERT((refcorner == TOPLEFT) && orderPreserved); 
	    for (i = 0; i < symreg->ninstances; i++) { 
		curSymInst = &symreg->instances[i]; 
		sortinst[i].s = curSymInst->xpos; 
		sortinst[i].t = curSymInst->ypos; 
		sortinst[i].sdist = curSymInst->sym->width; 
		*lowt = MIN(*lowt, sortinst[i].t); 
		} 
	    } 
	} 
	 
    /* Sort them */ 
    if (!orderPreserved) { 
	qsort((void *)sortinst, (unsigned)symreg->ninstances, 
	      sizeof(*sortinst), instCmp); 
	} 
 
    /* Go through and compute the modal delta S value */ 
 
    /* Zero out the histogram */ 
    for (i = 0; i < 32; i++) { 
	dshist[i] = 0; 
	} 
 
    for (i = 1; i < symreg->ninstances; i++) { 
	curinst = &sortinst[i]; 
	previnst = &sortinst[i - 1]; 
 
	if ((curinst->t >> 3) == (previnst->t >> 3)) { 
	    /* 
	     * This one's in the same strip as the previous one - kick the 
	     * histogram. 
	     */ 
	    ds = curinst->s - (previnst->s + previnst->sdist - 1); 
	    /* Only search in the appropriate range */ 
	    if ((ds >= -16) && (ds <= 15)) { 
		dshist[ds + 16]++; 
		} 
	    } 
	} 
 
    /* Scan for the peak - that's the modal delta S value */ 
    dsmax = dshist[0]; 
    *modeds = -16; 
    for (i = 1; i < 32; i++) { 
	if (dshist[i] > dsmax) { 
	    dsmax = dshist[i]; 
	    *modeds = i - 16; 
	    } 
	} 
 
    /* 
     * OK - they're now sorted into strips of 8, and L-to-R within 
     * each strip.  We want to run through each strip and decide 
     * how many bits it would cost, given each of the 4 possible strip 
     * heights. 
     * 
     * We can do this without changing the order: for the height-8 
     * strips, it's easy (just run through them in the order until you 
     * fall off the end of the strip). 
     * For the height-4 strips: do the counting of the two height-4 substrips 
     * of the actual height-8 strip in parallel.  Ignore the firstS and 
     * deltaT for the moment; just remember the first S coordinate you 
     * run across in each of the substrips (and whether you run across any), 
     * and count delta S within each substrip.  Then once you're done, 
     * look at the first S and empty-substrip fields, and figure out 
     * the firstS and deltaT costs. 
     * 
     * oldt is the largest T value seen for each strip height (scaled 
     * by strip height, updated between strips).  firsts and empty hold 
     * information about each substrip of the current strip.  count holds 
     * the total bit count for each strip height.  lastfirsts holds 
     * the first S value from the preceding non-empty strip. 
     */ 
 
    if (orderPreserved) { 
	/* 
	 * Strip size is forced to 1.  Subtract 1 from lowt (see 
	 * below) to find the first empty strip.  Free everything and 
	 * return. 
	 */ 
	(*lowt)--; 
	(a->free)((void *)sortinst); 
	*stripsize = 1; 
	*logstripsize = 0; 
	NORMAL_RETURN; 
	} 
 
    for (i = 0; i < 4; i++) { 
	count[i] = 0; 
	oldt[i] = *lowt >> i; 
	lastfirsts[i] = 0; 
	} 
 
    i = 0; 
    curstrip = 0; 
    while (i < symreg->ninstances) { 
	/* Process a strip */ 
 
	/* Clear the strip state variables */ 
	for (j = 0; j < 4; j++) { 
	    for (k = 0; k < 8; k++) { 
		firsts[j][k] = 0; 
		empty[j][k] = DGP_TRUE; 
		} 
	    } 
 
	/* curstrip is the last height-8 nonempty strip T */ 
	for (; (i < symreg->ninstances) && (curstrip == (sortinst[i].t >> 3)); 
	     i++) { 
	    /* We're within a strip. */ 
 
	    /* T offset within the strip */ 
	    curt = sortinst[i].t - (curstrip << 3); 
 
	    /* Deal with the current instance 4 ways */ 
	    for (j = 0; j < 4; j++) { 
		if (empty[j][curt >> j]) { 
		    /* First instance in that substrip */ 
		    empty[j][curt >> j] = DGP_FALSE; 
		    firsts[j][curt >> j] = sortinst[i].s; 
		    } 
		else { 
		    /* Not the first - what's the ds? */ 
		    ds = sortinst[i].s - curs[j][curt >> j] - *modeds; 
		    count[j] += dsApproxBits(ds); 
		    } 
 
		/* This has updated the current S value for that substrip */ 
		curs[j][curt >> j] = 
		    sortinst[i].s + sortinst[i].sdist - 1; 
 
		/* Don't forget the instance T bits */ 
		count[j] += j; 
		} 
	    } 
 
	/* 
	 * That finished a strip.  For each strip size, 
	 * scan through the empty/firsts values, updating lastfirsts and 
	 * oldt for the non-empty strips. 
	 * 
	 * The code above can have updated firsts[0][0...7], 
	 * firsts[1][0...3], firsts[2][0...1] and firsts[3][0]. 
	 */ 
	for (j = 0; j < 4; j++) { 
	    for (k = 0; k < (1 << (3 - j)); k++) { 
		if (!empty[j][k]) { 
		    /* 
		     * Encode the delta t since oldt[j].  The oldt values 
		     * are all shifted right by j. 
		     * curstrip is shifted right by 3. 
		     */ 
		    count[j] += 
			dtApproxBits((curstrip << (3 - j)) + k - oldt[j]); 
		    oldt[j] = (curstrip << (3 - j)) + k; 
 
		    /* Encode the delta in firsts */ 
		    count[j] += fsApproxBits(firsts[j][k] - lastfirsts[j]); 
		    lastfirsts[j] = firsts[j][k]; 
 
		    /* and the OOB for ending the strip */ 
		    count[j] += 2; 
		    } 
		/* No action needed on empty substrips */ 
		} 
	    } 
 
	/* 
	 * Set things up for the next strip. 
	 */ 
	if (i < symreg->ninstances) { 
	    curstrip = sortinst[i].t >> 3; 
	    } 
	} 
 
    /* 
     * We now have counts[0...3], which we use to pick the strip size. 
     */ 
    *stripsize = 1; 
    *logstripsize = 0; 
    mincount = count[0]; 
    for (j = 1; j < 4; j++) { 
	if (count[j] < mincount) { 
	    mincount = count[j]; 
	    *stripsize = 1 << j; 
	    *logstripsize = j; 
	    } 
	} 
 
    /* 
     * Shift lowt by the strip size and subtract one: the strip that lowt 
     * refers to *must* be empty, as the standard tables can't encode a 
     * value of zero, and so if there was an instance in lowt's 
     * strip, we'd have problems since there's a delta t between lowt's 
     * strip and the first non-empty strip. 
     * 
     * Subtracting one also ensures that lowt will be negative: it was 
     * initialised to zero, and could only have gotten smaller from there. 
     */ 
    *lowt >>= *logstripsize; 
    (*lowt)--; 
 
    /* Free the sorted instances */ 
    (a->free)((void *)sortinst); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (sortinst != (instInfo *)NULL) { 
	(a->free)((void *)sortinst); 
	} 
 
    ERR_RETURN; 
    } 
 
/* Do the approximate delta S count described above */ 
static Dgp_Int32 
dsApproxBits(Dgp_Int32 val) 
{ 
    if (val < -4) { 
	return(12); 
	} 
    else if (val < 0) { 
	return(8); 
	} 
    else if (val < 4) { 
	return(4); 
	} 
    else if (val < 128) { 
	return(8); 
	} 
    else { 
	return(12); 
	} 
    } 
 
static Dgp_Int32 
dtApproxBits(Dgp_Int32 val) 
{ 
    if (val == 1) { 
	return(1); 
	} 
    else if (val <= 4) { 
	return(3); 
	} 
    else if (val <= 8) { 
	return(5); 
	} 
    else { 
	return(8); 
	} 
    } 
 
static Dgp_Int32 
fsApproxBits(Dgp_Int32 val) 
{ 
    if ((val <= 32) && (val >= -32)) { 
	return(8); 
	} 
    else { 
	return(12); 
	} 
    } 
 
/* 
 * Helper sorting routines.  There are eight of them:  for strip sizes 
 * 1, 2, 4, 8, and for transposed/not transposed.  It would be nice if 
 * qsort passed in a cookie :-) 
 * 
 * All of these are stable by virtue of comparing the actual jb2SymbolInstance 
 * pointers as the last key. 
 */ 
typedef const jb2SymbolInstance *jb2SIPtr; 
/* Strip size 1, transposed */ 
static int
instGT1T(const void *p1, const void *p2) 
{ 
    const jb2SymbolInstance *i1 = *((const jb2SIPtr *)p1); 
    const jb2SymbolInstance *i2 = *((const jb2SIPtr *)p2); 
    Dgp_Int32 s1 = i1->ypos; 
    Dgp_Int32 s2 = i2->ypos; 
    Dgp_Int32 t1 = i1->xpos; 
    Dgp_Int32 t2 = i2->xpos; 
 
    if ((t1 > t2) || 
	((t1 == t2) && (s1 > s2)) || 
	((t1 == t2) && (s1 == s2) && (i1 > i2))) { 
	return(1); 
	} 
    else { 
	return(-1); 
	} 
    } 
     
/* Strip size 2, transposed */ 
static int
instGT2T(const void *p1, const void *p2) 
{ 
    const jb2SymbolInstance *i1 = *((const jb2SIPtr *)p1); 
    const jb2SymbolInstance *i2 = *((const jb2SIPtr *)p2); 
    Dgp_Int32 s1 = i1->ypos; 
    Dgp_Int32 s2 = i2->ypos; 
    Dgp_Int32 t1 = i1->xpos >> 1; 
    Dgp_Int32 t2 = i2->xpos >> 1; 
 
    if ((t1 > t2) || 
	((t1 == t2) && (s1 > s2)) || 
	((t1 == t2) && (s1 == s2) && (i1 > i2))) { 
	return(1); 
	} 
    else { 
	return(-1); 
	} 
    } 
     
/* Strip size 4, transposed */ 
static int
instGT4T(const void *p1, const void *p2) 
{ 
    const jb2SymbolInstance *i1 = *((const jb2SIPtr *)p1); 
    const jb2SymbolInstance *i2 = *((const jb2SIPtr *)p2); 
    Dgp_Int32 s1 = i1->ypos; 
    Dgp_Int32 s2 = i2->ypos; 
    Dgp_Int32 t1 = i1->xpos >> 2; 
    Dgp_Int32 t2 = i2->xpos >> 2; 
 
    if ((t1 > t2) || 
	((t1 == t2) && (s1 > s2)) || 
	((t1 == t2) && (s1 == s2) && (i1 > i2))) { 
	return(1); 
	} 
    else { 
	return(-1); 
	} 
    } 
     
/* Strip size 8, transposed */ 
static int
instGT8T(const void *p1, const void *p2) 
{ 
    const jb2SymbolInstance *i1 = *((const jb2SIPtr *)p1); 
    const jb2SymbolInstance *i2 = *((const jb2SIPtr *)p2); 
    Dgp_Int32 s1 = i1->ypos; 
    Dgp_Int32 s2 = i2->ypos; 
    Dgp_Int32 t1 = i1->xpos >> 3; 
    Dgp_Int32 t2 = i2->xpos >> 3; 
 
    if ((t1 > t2) || 
	((t1 == t2) && (s1 > s2)) || 
	((t1 == t2) && (s1 == s2) && (i1 > i2))) { 
	return(1); 
	} 
    else { 
	return(-1); 
	} 
    } 
     
/* Strip size 1, not transposed */ 
static int
instGT1NT(const void *p1, const void *p2) 
{ 
    const jb2SymbolInstance *i1 = *((const jb2SIPtr *)p1); 
    const jb2SymbolInstance *i2 = *((const jb2SIPtr *)p2); 
    Dgp_Int32 s1 = i1->xpos; 
    Dgp_Int32 s2 = i2->xpos; 
    Dgp_Int32 t1 = i1->ypos + i1->sym->height - 1; 
    Dgp_Int32 t2 = i2->ypos + i2->sym->height - 1; 
 
    if ((t1 > t2) || 
	((t1 == t2) && (s1 > s2)) || 
	((t1 == t2) && (s1 == s2) && (i1 > i2))) { 
	return(1); 
	} 
    else { 
	return(-1); 
	} 
    } 
 
/* Strip size 2, not transposed */ 
static int
instGT2NT(const void *p1, const void *p2) 
{ 
    const jb2SymbolInstance *i1 = *((const jb2SIPtr *)p1); 
    const jb2SymbolInstance *i2 = *((const jb2SIPtr *)p2); 
    Dgp_Int32 s1 = i1->xpos; 
    Dgp_Int32 s2 = i2->xpos; 
    Dgp_Int32 t1 = (i1->ypos + i1->sym->height - 1) >> 1; 
    Dgp_Int32 t2 = (i2->ypos + i2->sym->height - 1) >> 1; 
 
    if ((t1 > t2) || 
	((t1 == t2) && (s1 > s2)) || 
	((t1 == t2) && (s1 == s2) && (i1 > i2))) { 
	return(1); 
	} 
    else { 
	return(-1); 
	} 
    } 
 
/* Strip size 4, not transposed */ 
static int
instGT4NT(const void *p1, const void *p2) 
{ 
    const jb2SymbolInstance *i1 = *((const jb2SIPtr *)p1); 
    const jb2SymbolInstance *i2 = *((const jb2SIPtr *)p2); 
    Dgp_Int32 s1 = i1->xpos; 
    Dgp_Int32 s2 = i2->xpos; 
    Dgp_Int32 t1 = (i1->ypos + i1->sym->height - 1) >> 2; 
    Dgp_Int32 t2 = (i2->ypos + i2->sym->height - 1) >> 2; 
 
    if ((t1 > t2) || 
	((t1 == t2) && (s1 > s2)) || 
	((t1 == t2) && (s1 == s2) && (i1 > i2))) { 
	return(1); 
	} 
    else { 
	return(-1); 
	} 
    } 
 
/* Strip size 8, not transposed */ 
static int
instGT8NT(const void *p1, const void *p2) 
{ 
    const jb2SymbolInstance *i1 = *((const jb2SIPtr *)p1); 
    const jb2SymbolInstance *i2 = *((const jb2SIPtr *)p2); 
    Dgp_Int32 s1 = i1->xpos; 
    Dgp_Int32 s2 = i2->xpos; 
    Dgp_Int32 t1 = (i1->ypos + i1->sym->height - 1) >> 3; 
    Dgp_Int32 t2 = (i2->ypos + i2->sym->height - 1) >> 3; 
 
    if ((t1 > t2) || 
	((t1 == t2) && (s1 > s2)) || 
	((t1 == t2) && (s1 == s2) && (i1 > i2))) { 
	return(1); 
	} 
    else { 
	return(-1); 
	} 
    } 
 
/* 
 * Sort the symbol instances according to transposed, refcorner, stripsize, 
 * and return an array of pointers to the symbol instances, 
 * in sorted order. 
 * 
 * The caller is responsible for freeing the pointer returned. 
 */ 
static ErrCode 
sortInstances(jb2Segment *segment, Dgp_Bool orderPreserved, Dgp_Bool transposed, 
	      Dgp_UInt8 refcorner, Dgp_Int32 stripsize, Dgp_State *dgp_state, 
	      jb2SymbolInstance ***_sortedsyms) 
{ 
  DGP_PROLOGUE("sortInstances()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    jb2SymRegSegment *symreg = &segment->data.symreg; 
    jb2SymbolInstance **sortedsyms = (jb2SymbolInstance **)NULL; 
    Dgp_Int32 i; 
    int (*cmpfunc)(const void *, const void *); 
    cmpfunc = NULL;
 
    *_sortedsyms = (jb2SymbolInstance **)NULL; 
    refcorner = refcorner;
     
    /* Allocate the sorted array */ 
    sortedsyms = (jb2SymbolInstance **)(a->malloc)(symreg->ninstances * 
					      sizeof(*sortedsyms)); 
    DGP_CHECK_ALLOC_CLEANUP(sortedsyms); 
 
    /* Read them into the array in the initial order */ 
    for (i = 0; i < symreg->ninstances; i++) { 
	sortedsyms[i] = &symreg->instances[i]; 
	} 
    if (!orderPreserved) { 
        /* Decide what sorting function to use */ 
        if (transposed) { 
	    DGP_ASSERT(refcorner == TOPLEFT); 
	    switch(stripsize) { 
	      case 1: { 
	        cmpfunc = instGT1T; 
	        break; 
	        } 
	      case 2: { 
	        cmpfunc = instGT2T; 
	        break; 
	        } 
	      case 4: { 
	        cmpfunc = instGT4T; 
	        break; 
	        } 
	      case 8: { 
	        cmpfunc = instGT8T; 
	        break; 
	        } 
	      default: { 
	        DGP_REPORTF0(DGP_ENC, "bad strip size"); 
	        } 
	      } 
	    } 
        else { 
	    DGP_ASSERT(refcorner == BOTTOMLEFT); 
	    switch(stripsize) { 
	      case 1: { 
	        cmpfunc = instGT1NT; 
	        break; 
	        } 
	      case 2: { 
	        cmpfunc = instGT2NT; 
	        break; 
	        } 
	      case 4: { 
	        cmpfunc = instGT4NT; 
	        break; 
	        } 
	      case 8: { 
	        cmpfunc = instGT8NT; 
	        break; 
	        } 
	      default: { 
	        DGP_REPORTF0(DGP_ENC, "bad strip size"); 
	        } 
	      } 
	    } 
 
        /* and sort them, if we're not prohibited to */ 
	    qsort((void *)sortedsyms, symreg->ninstances, sizeof(*sortedsyms), 
	          cmpfunc); 
	} 
 
    /* Done! */ 
    *_sortedsyms = sortedsyms; 
    sortedsyms = (jb2SymbolInstance **)NULL; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (sortedsyms != (jb2SymbolInstance **)NULL) { 
	(a->free)((void *)sortedsyms); 
	} 
    ERR_RETURN; 
    } 
 
/* 
 * Given all the information about how to do the encoding, 
 * figure out which tables to use. 
 * 
 * For each one, return the chosen table in *FOOtable, and whether this 
 * is a custom table in *FOOnew.  If it's not a custom table, then 
 * *FOOtable is a pointer to one of the standard tables. 
 * 
 * The caller is responsible for freeing any new tables; use jhtFree. 
 */  
static ErrCode 
decideTables(jb2FileParams *params, 
	     jb2SymbolInstance **instptrs, Dgp_Int32 ninstances, 
	     hufftable_spec **dttable, Dgp_Bool *dtnew, 
	     hufftable_spec **dstable, Dgp_Bool *dsnew, 
	     hufftable_spec **fstable, Dgp_Bool *fsnew, 
	     Dgp_Int32 logstripsize, Dgp_UInt8 refcorner, 
	     Dgp_Bool transposed, Dgp_Int32 dsoffset, Dgp_Int32 lowt, 
	     Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("decideTables()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Int32 i, j; 
    Dgp_Int32 *dtlengths[NDTTABLES], *dslengths[NDSTABLES], *fslengths[NFSTABLES]; 
    Dgp_Int32 *dtcounts[NDTTABLES], *dscounts[NDSTABLES], *fscounts[NFSTABLES]; 
    Dgp_Int32 dtnormal[NDTTABLES], dsnormal[NDSTABLES], fsnormal[NFSTABLES]; 
    Dgp_Int32 dttuned[NDTTABLES], dstuned[NDSTABLES], fstuned[NFSTABLES]; 
    jb2SymbolInstance *curinst; 
    eiEncoder dtei[NDTTABLES], dsei[NDSTABLES], fsei[NDSTABLES]; 
    Dgp_Int32 stript, firsts, insts, instt, instsdist, curs; 
    Dgp_Bool negT = DGP_FALSE; 
     
    /* Clear for easy bailout */ 
    for (i = 0; i < NDTTABLES; i++) { 
	dtlengths[i] = dtcounts[i] = (Dgp_Int32 *)NULL; 
	dtei[i] = NULLEIENCODER; 
    } 
    for (i = 0; i < NDSTABLES; i++) { 
	dslengths[i] = dscounts[i] = (Dgp_Int32 *)NULL; 
	dsei[i] = NULLEIENCODER; 
    } 
    for (i = 0; i < NFSTABLES; i++) { 
	fslengths[i] = fscounts[i] = (Dgp_Int32 *)NULL; 
	fsei[i] = NULLEIENCODER; 
    } 
    *dttable = *dstable = *fstable = (hufftable_spec *)NULL; 
    *dtnew = *dsnew = *fsnew = DGP_FALSE; 
     
    /* Allocate the work space we need */ 
    DGP_ERR_PASS_CLEANUP(jhtAllocTableInfo(dttables, dtlengths, dtcounts, dtei, 
					   NDTTABLES, dgp_state)); 
     
    DGP_ERR_PASS_CLEANUP(jhtAllocTableInfo(dstables, dslengths, dscounts, dsei, 
					   NDSTABLES, dgp_state)); 
     
    DGP_ERR_PASS_CLEANUP(jhtAllocTableInfo(fstables, fslengths, fscounts, fsei, 
					   NFSTABLES, dgp_state)); 
     
    /* Pick the initial strip */ 
    stript = lowt << logstripsize; 
    firsts = 0; 
    curs = 0; 
 
    /* and encode it */ 
    for (j = 0; j < NDTTABLES; j++) { 
	eiCount(dtei[j], -lowt, DGP_FALSE, dtcounts[j]); 
	} 
 
    for (i = 0; i < ninstances; i++) { 
	curinst = instptrs[i]; 
 
	/* 
	 * Get S and T based on refcorner and transposed. 
	 * The instance (xpos,ypos) is always top right pixel. 
	 */ 
	if (transposed) { 
	    if ((refcorner == TOPLEFT) || (refcorner == BOTTOMLEFT)) { 
		/* 
		 * Why is BOTTOMLEFT doing the same thing as TOPLEFT? 
		 * Well, when you decode BOTTOMLEFT, 
		 * there's this odd pre-positioning update that gets done: 
		 * the S position decoded 
		 * gets H-1 added to it before it is used for positioning. 
		 * Therefore, the S position we use for encoding is really 
		 * just ypos (since the H-1 addition to shift it to the 
		 * bottom happens in the decoder). 
		 */ 
		insts = curinst->ypos; 
		instt = curinst->xpos; 
		} 
	    else { 
		/* TOPRIGHT and BOTTOMRIGHT can share code for the same reason */ 
		insts = curinst->ypos; 
		instt = curinst->xpos + curinst->sym->width - 1; 
		} 
	    instsdist = curinst->sym->height; 
	    } 
	else { 
	    if ((refcorner == TOPLEFT) || (refcorner == TOPRIGHT)) { 
		insts = curinst->xpos; 
		instt = curinst->ypos; 
		} 
	    else { 
		insts = curinst->xpos; 
		instt = curinst->ypos + curinst->sym->height - 1; 
		} 
	    instsdist = curinst->sym->width; 
	    } 
 
	/* Got S and T - handle the strip state mechanism. */ 
	if ((i == 0) || ((instt >> logstripsize) != (stript >> logstripsize))) { 
	    if (instt < stript) { 
		/* The order-preserving case may require backtracking */ 
		/* 
		 * Only use the special negative-permitting delta T 
		 * table, which is the last in dttables. 
		 */ 
		negT = DGP_TRUE; 
		} 
 
	    /* Start of a new strip - do the delta T coding */ 
	    if (!negT) { 
		for (j = 0; j < NDTTABLES; j++) { 
		    eiCount(dtei[j], (instt - stript) >> logstripsize, DGP_FALSE, 
			    dtcounts[j]); 
		    } 
		} 
	    else { 
		/* Only check the last delta T table */ 
		eiCount(dtei[NEGDTTABLE], (instt - stript) >> logstripsize, 
			DGP_FALSE, dtcounts[NEGDTTABLE]); 
		} 
 
	    /* Do the first S coding */ 
	    for (j = 0; j < NFSTABLES; j++) { 
		eiCount(fsei[j], insts - firsts, DGP_FALSE, fscounts[j]); 
		} 
 
	    /* Update stript and firsts */ 
 
	    /* stript must always be the top row of the strip */ 
	    stript = (instt >> logstripsize) << logstripsize; 
	    firsts = insts; 
 
	    /* If it's not the first strip, count an OOB for DS */ 
	    if (i != 0) { 
		for (j = 0; j < NDSTABLES; j++) { 
		    eiCount(dsei[j], 0, DGP_TRUE, dscounts[j]); 
		    } 
		} 
	    } 
	else { 
	    /* Within a strip - just encode a DS */ 
	    for (j = 0; j < NDSTABLES; j++) { 
		eiCount(dsei[j], insts - curs, DGP_FALSE, dscounts[j]); 
		} 
	    } 
	/* Update the current S position */ 
	curs = insts + instsdist - 1 + dsoffset; 
	} 
 
    /* Write an OOB to end the last strip */ 
    if (ninstances != 0) { 
	/* There were some instances */ 
	for (j = 0; j < NDSTABLES; j++) { 
	    eiCount(dsei[j], 0, DGP_TRUE, dscounts[j]); 
	    } 
	} 
 
    /* We now have all the statistics.  Compute bit counts. */ 
    if (!negT) { 
	for (i = 0; i < NDTTABLES; i++) { 
	    DGP_ERR_PASS_CLEANUP(jhtCountBits(dttables[i], dtcounts[i], 
					      dtlengths[i], &dtnormal[i], 
					      &dttuned[i], dgp_state)); 
	    } 
	} 
    else { 
	/* 
	 * We saw a negative deltaT value - our only choice is 
	 * the special table. 
	 */ 
	DGP_ERR_PASS_CLEANUP(jhtCountBits(dttables[NEGDTTABLE], 
					  dtcounts[NEGDTTABLE], 
					  dtlengths[NEGDTTABLE], 
					  &dtnormal[NEGDTTABLE], 
					  &dttuned[NEGDTTABLE], dgp_state)); 
	} 
     
    for (i = 0; i < NDSTABLES; i++) { 
	DGP_ERR_PASS_CLEANUP(jhtCountBits(dstables[i], dscounts[i], 
					  dslengths[i], &dsnormal[i], 
					  &dstuned[i], dgp_state)); 
	} 
     
    for (i = 0; i < NFSTABLES; i++) { 
	DGP_ERR_PASS_CLEANUP(jhtCountBits(fstables[i], fscounts[i], 
					  fslengths[i], &fsnormal[i], 
					  &fstuned[i], dgp_state)); 
	} 
     
    /* and decide on the tables */ 
 
    /* Our parameters might force the decision. */ 
    if (params->srfirstonly) { 
	/* No decision - only allowed to choose the first */ 
 
      /* We're SOL - we can't do this */ 
      DGP_ERR_GEN_CLEANUP((negT), cErrInvalidParm); 
 
	*dttable = dttables[0]; 
	*dtnew = DGP_FALSE; 
 
	*dstable = dstables[0]; 
	*dsnew = DGP_FALSE; 
 
	*fstable = fstables[0]; 
	*fsnew = DGP_FALSE; 
	} 
    else { 
	if (negT) { 
	    /* 
	     * It's forced - we have no choice but to create a tuned 
	     * version of the special delta T table (the only table 
	     * we have available that allows negative delta T values). 
	     */ 
	    DGP_ERR_PASS_CLEANUP(jhtTuneStandardTable(dttables[NEGDTTABLE], 
					    dtcounts[NEGDTTABLE], 
					    dtlengths[NEGDTTABLE], 
					    dgp_state, dttable)); 
	    *dtnew = DGP_TRUE; 
#ifdef	VALPRINT 
	    fprintf(stderr, "dttables: XXX XXX XXX %d normal XXX XXX XXX %d tuned\n", dtnormal[3], dttuned[3]); 
#endif 
	    } 
	else { 
	    /* 
	     * We can't allow it to pick the special table - exclude it 
	     * by saying there are NDTTABLES - 1 choices. 
	     */ 
	    DGP_ERR_PASS_CLEANUP(jhtDecideTable(dttables, dtcounts, dtlengths, 
						dtnormal, dttuned, dtnew, 
						dttable, NDTTABLES - 1, 
						params->srtables, dgp_state)); 
#ifdef	VALPRINT 
	    fprintf(stderr, "dttables: %d %d %d %d normal %d %d %d %d tuned\n", dtnormal[0], dtnormal[1], dtnormal[2], dtnormal[3], dttuned[0], dttuned[1], dttuned[2], dttuned[3]); 
#endif 
	    } 
     
	/* DS decision */ 
	DGP_ERR_PASS_CLEANUP(jhtDecideTable(dstables, dscounts, dslengths, 
					    dsnormal, dstuned, dsnew, dstable, 
					    NDSTABLES, params->srtables, 
					    dgp_state)); 
#ifdef	VALPRINT 
	fprintf(stderr, "dstables: %d %d %d normal %d %d %d tuned\n", dsnormal[0], dsnormal[1], dsnormal[2], dstuned[0], dstuned[1], dstuned[2]); 
#endif 
     
	/* FS decision */ 
	DGP_ERR_PASS_CLEANUP(jhtDecideTable(fstables, fscounts, fslengths, 
					    fsnormal, fstuned, fsnew, fstable, 
					    NFSTABLES, params->srtables, 
					    dgp_state)); 
#ifdef	VALPRINT 
	fprintf(stderr, "fstables: %d %d normal %d %d tuned\n", fsnormal[0], fsnormal[1], fstuned[0], fstuned[1]); 
#endif 
	} 
 
    /* Clean up our arrays */ 
    for (i = 0; i < NDTTABLES; i++) { 
	(a->free)((void *)dtlengths[i]); 
	(a->free)((void *)dtcounts[i]); 
	eiFreeEncoder(dtei[i]); 
	} 
 
    for (i = 0; i < NDSTABLES; i++) { 
	(a->free)((void *)dslengths[i]); 
	(a->free)((void *)dscounts[i]); 
	eiFreeEncoder(dsei[i]); 
	} 
 
    for (i = 0; i < NFSTABLES; i++) { 
	(a->free)((void *)fslengths[i]); 
	(a->free)((void *)fscounts[i]); 
	eiFreeEncoder(fsei[i]); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    for (i = 0; i < NDTTABLES; i++) { 
	if (dtlengths[i] != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)dtlengths[i]); 
	    } 
	if (dtcounts[i] != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)dtcounts[i]); 
	    } 
 
	if (dtei[i] != NULLEIENCODER) { 
	    eiFreeEncoder(dtei[i]); 
	    } 
	} 
    if (*dtnew && (*dttable != (hufftable_spec *)NULL)) { 
	jhtFree(*dttable, dgp_state); 
	} 
     
    for (i = 0; i < NDSTABLES; i++) { 
	if (dslengths[i] != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)dslengths[i]); 
	    } 
	if (dscounts[i] != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)dscounts[i]); 
	    } 
 
	if (dsei[i] != NULLEIENCODER) { 
	    eiFreeEncoder(dsei[i]); 
	    } 
	} 
    if (*dsnew && (*dstable != (hufftable_spec *)NULL)) { 
	jhtFree(*dstable, dgp_state); 
	} 
     
    for (i = 0; i < NFSTABLES; i++) { 
	if (fslengths[i] != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)fslengths[i]); 
	    } 
	if (fscounts[i] != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)fscounts[i]); 
	    } 
 
	if (fsei[i] != NULLEIENCODER) { 
	    eiFreeEncoder(fsei[i]); 
	    } 
	} 
    if (*fsnew && (*fstable != (hufftable_spec *)NULL)) { 
	jhtFree(*fstable, dgp_state); 
	} 
     
    ERR_RETURN; 
    } 
 
/* 
 * Compute and write the symbol ID table.  Fill in *symenc. 
 * 
 * We can assume that every symbol pointed to by any instance has had 
 * its userdata field set to point to a symInfo structure. 
 * The count fields are assumed to be currently clear, and the symnum 
 * fields already filled in.  syminfo is the underlying array 
 * that all those pointers point into. 
 * 
 * Count how many times each symbol is used; build a Huffman table 
 * for encoding the symbols; write out a representation of the Huffman 
 * table. 
 * 
 * When this routine returns, the symInfo count fields have been filled 
 * in, and the symnum fields each give the correct index into the symEncoder 
 * that's been created and stored in *symenc. 
 */ 
static ErrCode 
writeIDHuffTable(jb2SymRegSegment *symreg, jbitfile bf, 
		 symInfo *syminfo, Dgp_Int32 nrefsyms, 
		 symEncoder *symenc, Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("writeIDHuffTable()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    Dgp_Int32 i, j; 
    Dgp_Int32 *counts = (Dgp_Int32 *)NULL; 
    Dgp_Int32 *lengths = (Dgp_Int32 *)NULL; 
    jb2Symbol *cursym; 
    jb2SymbolInstance *curinst; 
    symInfo *curSI; 
    hufftable_spec runtable; 
    table_line runlines[35]; 
    Dgp_Int32 runcounts[35], runlengths[35]; 
    Dgp_Int32 currun, curlen; 
 
    /* 
     * First: go through and put in the counts.  We don't care about 
     * the order here, so just use the order in the original instance list. 
     */ 
    for (i = 0; i < symreg->ninstances; i++) { 
	curinst = &symreg->instances[i]; 
	cursym = curinst->sym; 
	curSI = (symInfo *)cursym->userdata; 
	curSI->count++; 
	} 
     
    /* Now copy the counts into a side array for jhtGetCodeLengths */ 
    counts = (Dgp_Int32 *)(a->malloc)(nrefsyms * sizeof(*counts)); 
    DGP_CHECK_ALLOC_CLEANUP(counts); 
 
    lengths = (Dgp_Int32 *)(a->malloc)(nrefsyms * sizeof(*lengths)); 
    DGP_CHECK_ALLOC_CLEANUP(lengths); 
 
    for (i = 0; i < nrefsyms; i++) { 
	counts[i] = syminfo[i].count; 
	} 
 
    /* Call jhtGetCodeLengths to find the length for each symbol */ 
    DGP_ERR_PASS_CLEANUP(jhtGetCodeLengths(counts, lengths, nrefsyms,  
					   dgp_state)); 
 
    /* Create a symEncoder with those lengths */ 
    DGP_ERR_PASS_CLEANUP(symEncoderHuffman(nrefsyms, bf, lengths, dgp_state, 
					   symenc)); 
 
    /* 
     * OK - now transmit the Huffman table.  Actually, all the information 
     * we need is in the array "lengths", so that's what we send.  Run-length 
     * code it; build a Huffman table for the run-length codes; send 
     * that table (well, the lengths for it) followed by the run-length 
     * coded "lengths". 
     * 
     * The codes used in the run coding are: 
     * 0-31:		a use count code length (0 means "unused") 
     * 32+xx:		repeat the previous code length xx+3 times 
     * 33+xxx:		repeat a code length of 0 xxx+3 times 
     * 34+xxxxxxx:	repeat a code length of 0 xxxxxxx+11 times 
     */ 
 
    /* Do the run length coding once, just accumulating counts */ 
    for (i = 0; i < 35; i++) { 
	runcounts[i] = 0; 
	} 
 
    /* 
     * The run-encoding state machine is: 
     * Look at the next code length to be encoded. 
     * If it is not zero, and doesn't match the previous one, terminate the 
     * current run and write it out. 
     * If it is not zero and does match the previous one, start a new run 
     * if there isn't one active; if there is one active, add to it. 
     * 
     * Writing out a run may involve breaking it down into max-length 
     * subruns. 
     * 
     * We count the number of times each code in this run code occurs, 
     * build a Huffman table, and write out the Huffman table as 35 4-bit 
     * values (the code lengths in that table). 
     */ 
    currun = 0; 
    curlen = -1; 
    for (i = 0; i < nrefsyms; i++) { 
	if (lengths[i] == curlen) { 
	    /* Continue this run */ 
	    currun++; 
	    } 
	else { 
	    /* New one - close the old one */ 
	    if (currun > 0) { 
		/* There was an old one */ 
		if (curlen > 0) { 
		    /* 
		     * Not a run of zeros.  Emit the value, then 
		     * a bunch of (32, 3), then however many counts 
		     * are left over. 
		     */ 
		    runcounts[curlen]++; 
		    currun--; 
		    while (currun >= 6) { 
			/* Code 32 */ 
			runcounts[32]++; 
			currun -= 6; 
			} 
		    if (currun > 2) { 
			/* It used code 32 */ 
			runcounts[32]++; 
			} 
		    else { 
			/* 0-2 more instances of the value itself */ 
			runcounts[curlen] += currun; 
			} 
		    } 
		else { 
		    /* A run of zeros */ 
		    while (currun >= 138) { 
			/* Very Dgp_Int32 run */ 
			runcounts[34]++; 
			currun -= 138; 
			} 
 
		    if (currun > 2) { 
			if (currun > 10) { 
			    /* Large run of zeros */ 
			    runcounts[34]++; 
			    } 
			else { 
			    /* Short run of zeros */ 
			    runcounts[33]++; 
			    } 
			} 
		    else { 
			/* Very Dgp_Int16 run of zeros */ 
			runcounts[0] += currun; 
			} 
		    } 
		} 
 
	    /* In any case, we start a new run */ 
	    curlen = lengths[i]; 
	    currun = 1; 
	    } 
	} 
 
    /* Now just close off the last run */ 
    if (currun > 0) { 
	if (curlen > 0) { 
	    runcounts[curlen]++; 
	    currun--; 
	    while (currun >= 6) { 
		runcounts[32]++; 
		currun -= 6; 
		} 
	    if (currun > 2) { 
		runcounts[32]++; 
		} 
	    else { 
		runcounts[curlen] += currun; 
		} 
	    } 
	else { 
	    while (currun >= 138) { 
		runcounts[34]++; 
		currun -= 138; 
		} 
	    if (currun > 2) { 
		if (currun > 10) { 
		    runcounts[34]++; 
		    } 
		else { 
		    runcounts[33]++; 
		    } 
		} 
	    else { 
		runcounts[0] += currun; 
		} 
	    } 
	} 
	 
    /* Done - now we have the counts.  Build the lengths */ 
    DGP_ERR_PASS_CLEANUP(jhtGetCodeLengths(runcounts, runlengths, 35,  
					   dgp_state)); 
 
    /* Build the Huffman table */ 
    runtable.lower_valid = runtable.upper_valid = runtable.oob_valid = DGP_FALSE; 
    runtable.nlines = 35; 
    runtable.lines = runlines; 
    for (i = 0; i < 35; i++) { 
	runlines[i].preflen = (Dgp_Int8)runlengths[i]; 
	runlines[i].rangelen = 0; 
	runlines[i].base = i; 
	} 
    DGP_ERR_PASS_CLEANUP(jhtSetCodes(&runtable, dgp_state)); 
 
    /* Output the run-length table Huffman code lengths */ 
    for (i = 0; i < 35; i++) { 
	DGP_ASSERT(runlengths[i] <= 15);	/* We should actually ensure this... */ 
	DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlengths[i], 4)); 
	} 
 
    /* Now start writing runs of symbol code lengths */ 
    currun = 0; 
    curlen = -1; 
    for (i = 0; i < nrefsyms; i++) { 
	if (lengths[i] == curlen) { 
	    /* Continue this run */ 
	    currun++; 
	    } 
	else { 
	    /* New one - close the old one */ 
	    if (currun > 0) { 
		/* There was an old one */ 
		if (curlen > 0) { 
		    /* Not a run of zeros */ 
		    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[curlen].prefix, 
				      runlines[curlen].preflen)); 
#ifdef	VALPRINT 
		    DEBUG("jht %lx(%d)\n", runlines[curlen].prefix, runlines[curlen].preflen); 
		    fprintf(stderr, "RC %d\n", curlen); 
#endif 
		    currun--; 
 
		    while (currun >= 6) { 
			/* Code 32 plus 2 bits of 3 */ 
			DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[32].prefix, 
					  runlines[32].preflen)); 
			DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, 3, 2)); 
#ifdef	VALPRINT 
			DEBUG("jht %lx(%d)\n", runlines[32].prefix, runlines[32].preflen); 
			fprintf(stderr, "RC %d %d\n", 32, 3); 
#endif 
			currun -= 6; 
			} 
 
		    if (currun > 2) { 
			/* Code 32 plus 2 bits of currun - 3 */ 
			DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[32].prefix, 
					  runlines[32].preflen)); 
			DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, currun - 3, 2)); 
#ifdef	VALPRINT 
			DEBUG("jht %lx(%d)\n", runlines[32].prefix, runlines[32].preflen); 
			fprintf(stderr, "RC %d %d\n", 32, currun - 3); 
#endif 
			} 
		    else { 
			/* 0-2 more instances of the value itself */ 
			for (j = 0; j < currun; j++) { 
			    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[curlen].prefix, 
					      runlines[curlen].preflen)); 
#ifdef	VALPRINT 
			    DEBUG("jht %lx(%d)\n", runlines[curlen].prefix, runlines[curlen].preflen); 
			    fprintf(stderr, "RC %d\n", curlen); 
#endif 
			    } 
			} 
		    } 
		else { 
		    /* A run of zeros */ 
		    while (currun >= 138) { 
			/* Process the very Dgp_Int32 run until it's Dgp_Int16 enough */ 
			DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[34].prefix, 
					  runlines[34].preflen)); 
			DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, 127, 7)); 
#ifdef	VALPRINT 
			DEBUG("jht %lx(%d)\n", runlines[34].prefix, runlines[34].preflen); 
			fprintf(stderr, "RC %d %d\n", 34, 127); 
#endif 
			currun -= 138; 
			} 
		    if (currun > 2) { 
			if (currun > 10) { 
			    /* Large run of zeros */ 
			    /* Code 34 plus 7 bits of currun - 11 */ 
			    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[34].prefix, 
					      runlines[34].preflen)); 
			    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, currun - 11, 7)); 
#ifdef	VALPRINT 
			    DEBUG("jht %lx(%d)\n", runlines[34].prefix, runlines[34].preflen); 
			    fprintf(stderr, "RC %d %d\n", 34, currun - 11); 
#endif 
			    } 
			else { 
			    /* Short run of zeros */ 
			    /* Code 33 plus 3 bits of currun - 3 */ 
			    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[33].prefix, 
					      runlines[33].preflen)); 
			    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, currun - 3, 3)); 
#ifdef	VALPRINT 
			    DEBUG("jht %lx(%d)\n", runlines[33].prefix, runlines[33].preflen); 
			    fprintf(stderr, "RC %d %d\n", 33, currun - 3); 
#endif 
			    } 
			} 
		    else { 
			/* Very Dgp_Int16 run of zeros */ 
			/* 1 or 2 instances of 0 */ 
			for (j = 0; j < currun; j++) { 
			    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[0].prefix, 
					      runlines[0].preflen)); 
#ifdef	VALPRINT 
			    DEBUG("jht %lx(%d)\n", runlines[0].prefix, runlines[0].preflen); 
			    fprintf(stderr, "RC %d\n", 0); 
#endif 
			    } 
			} 
		    } 
		} 
 
	    /* In any case, we start a new run */ 
	    curlen = lengths[i]; 
	    currun = 1; 
	    } 
	} 
 
    /* Now just close off the last run */ 
    if (currun > 0) { 
	if (curlen > 0) { 
	    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[curlen].prefix, 
			      runlines[curlen].preflen)); 
#ifdef	VALPRINT 
	    DEBUG("jht %lx(%d)\n", runlines[curlen].prefix, runlines[curlen].preflen); 
	    fprintf(stderr, "RC %d\n", curlen); 
#endif 
	    currun--; 
 
	    while (currun >= 6) { 
		DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[32].prefix, 
				  runlines[32].preflen)); 
		DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, 3, 2)); 
#ifdef	VALPRINT 
		DEBUG("jht %lx(%d)\n", runlines[32].prefix, runlines[32].preflen); 
		fprintf(stderr, "RC %d %d\n", 32, 3); 
#endif 
		currun -= 6; 
		} 
 
	    if (currun > 2) { 
		DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[32].prefix, 
				  runlines[32].preflen)); 
		DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, currun - 3, 2)); 
#ifdef	VALPRINT 
		DEBUG("jht %lx(%d)\n", runlines[32].prefix, runlines[32].preflen); 
		fprintf(stderr, "RC %d %d\n", 32, currun - 3); 
#endif 
		} 
	    else { 
		for (j = 0; j < currun; j++) { 
		    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[curlen].prefix, 
				      runlines[curlen].preflen)); 
#ifdef	VALPRINT 
		    DEBUG("jht %lx(%d)\n", runlines[curlen].prefix, runlines[curlen].preflen); 
		    fprintf(stderr, "RC %d\n", curlen); 
#endif 
		    } 
		} 
	    } 
	else { 
	    while (currun >= 138) { 
		DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[34].prefix, 
				  runlines[34].preflen)); 
		DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, 127, 7)); 
#ifdef	VALPRINT 
		DEBUG("jht %lx(%d)\n", runlines[34].prefix, runlines[34].preflen); 
		fprintf(stderr, "RC %d %d\n", 34, 127); 
#endif 
		currun -= 138; 
		} 
	    if (currun > 2) { 
		if (currun > 10) { 
		    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[34].prefix, 
				      runlines[34].preflen)); 
		    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, currun - 11, 7)); 
#ifdef	VALPRINT 
		    DEBUG("jht %lx(%d)\n", runlines[34].prefix, runlines[34].preflen); 
		    fprintf(stderr, "RC %d %d\n", 34, currun - 11); 
#endif 
		    } 
		else { 
		    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[33].prefix, 
				      runlines[33].preflen)); 
		    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, currun - 3, 3)); 
#ifdef	VALPRINT 
		    DEBUG("jht %lx(%d)\n", runlines[33].prefix, runlines[33].preflen); 
		    fprintf(stderr, "RC %d %d\n", 33, currun - 3); 
#endif 
		    } 
		} 
	    else { 
		for (j = 0; j < currun; j++) { 
		    DGP_ERR_PASS_CLEANUP(jbfWriteBits(bf, runlines[0].prefix, 
				      runlines[0].preflen)); 
#ifdef	VALPRINT 
		    DEBUG("jht %lx(%d)\n", runlines[0].prefix, runlines[0].preflen); 
		    fprintf(stderr, "RC %d\n", 0); 
#endif 
		    } 
		} 
	    } 
	} 
 
    /* Done - flush the last byte */ 
    jbfFlushByte(bf); 
 
    /* Can free lengths and counts - info is copied into symenc */ 
    (a->free)((void *)counts); 
    (a->free)((void *)lengths); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (counts != (Dgp_Int32 *)NULL) { 
	(a->free)((void *)counts); 
	} 
 
    if (lengths != (Dgp_Int32 *)NULL) { 
	(a->free)((void *)lengths); 
	} 
 
    ERR_RETURN; 
    } 
