/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/jb2file.c,v 1.19 2003/01/27 21:48:34 awilliam Exp $ 
 * 
 */ 
 
#include "jb2file.h" 
#include "jb2fileprivate.h" 
#include "PsgFix.h"
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/jb2file.c,v 1.19 2003/01/27 21:48:34 awilliam Exp $"); 
 
/* 
 * Allocate a new segment number 
 */ 
Dgp_Int32
jb2FileNextSegmentNumber(jb2File file) 
{ 
    return(file->lastsegnum++); 
    } 
 
/* 
 * A jb2File has two DataSinks associated with it at any moment: one for 
 * the segment headers, one for the segment data.  These can be the same 
 * or may be different.  Also, they can change between encoding segments. 
 * This allows a very flexible set of output modalities: the segment 
 * headers and data parts can be separated at birth; different segments 
 * can be written to different DataSinks, and so on. 
 */ 
 
/* 
 * Return the DataSink used to output segment headers 
 */ 
DataSink 
jb2GetFileHeaderDataSink(jb2File file) 
{ 
    return(file->headersink); 
    } 
 
/* 
 * Return the DataSink used to output segment data 
 */ 
DataSink 
jb2GetFileDataSink(jb2File file) 
{ 
    return(file->bodysink); 
    } 
 
 
/* Create and fill in a jb2File structure */ 
ErrCode __cdecl
jb2CreateFile(DataSink headersink, DataSink bodysink, 
	      Dgp_State *dgp_state, jb2File *_f) 
{ 
  DGP_PROLOGUE("jb2CreateFile()", DGP_PERRSTATE); 
    jb2File f = NULL; 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    f = (jb2File)(a->malloc)(sizeof(*f)); 
    DGP_CHECK_ALLOC_CLEANUP(f); 
 
    f->lastsegnum = 0; 
    f->headersink = headersink; 
    f->bodysink = bodysink; 
    f->dgp_state = dgp_state; 
 
    /* Set default parameters */ 
    /* Huffman coding */ 
    f->params.sdarith = f->params.srarith = DGP_FALSE; 
    /* Custom Huffman tables allowed */ 
    f->params.sdtables = f->params.srtables = DGP_TRUE; 
    /* All Huffman tables allowed */ 
    f->params.sdfirstonly = f->params.srfirstonly = DGP_FALSE; 
    /* All transposition, corner choices allowed */ 
    f->params.srllonly = f->params.srnontransonly = DGP_FALSE; 
 
    /* 10-pixel 3-line default for symdicts when arithmetic */ 
    f->params.sdabparams.rtemplate = 2; 
    /* and nominal AT position */ 
    f->params.sdabparams.atx[0] = 2; 
    f->params.sdabparams.aty[0] = -1; 
    /* might as well fill in nominal for template 0 */ 
    f->params.sdabparams.atx[1] = -3; 
    f->params.sdabparams.aty[1] = -1; 
    f->params.sdabparams.atx[2] = 2; 
    f->params.sdabparams.aty[2] = -2; 
    f->params.sdabparams.atx[3] = -2; 
    f->params.sdabparams.aty[3] = -2; 
    /* No TPGD is possible */ 
    f->params.sdabparams.tpon = DGP_FALSE; 
 
    /* Halftone dictionaries */ 
    f->params.hdarith = DGP_FALSE; 
    f->params.hdabtemplate = 2; 
    /* Halftone regions */ 
    f->params.hrarith = DGP_FALSE; 
    f->params.hrabtemplate = 0; 
 
    *_f = f; 
    f = NULL; 
 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
 
    ERR_RETURN; 
    } 
 
/* Destroy a jb2File structure */ 
void 
jb2FreeFile(jb2File f) 
{ 
  Dgp_Free tmp_free = f->DGP_FREE; 
 
    (tmp_free)((void *)f); 
    } 
 
void 
jb2SetFileSegnum(jb2File f, Dgp_Int32 segnum) 
{ 
    f->lastsegnum = segnum; 
    } 
 
Dgp_Int32
jb2GetFileSegnum(jb2File f) 
{ 
    return(f->lastsegnum); 
    } 
 
void 
jb2SetFileParams(jb2File f, jb2FileParams *params) 
{ 
    f->params = *params; 
    } 
 
jb2FileParams * 
jb2GetFileParams(jb2File f) 
{ 
    return(&f->params); 
    } 
 
Dgp_State * 
jb2GetFileDgp_State(jb2File f) 
{ 
  return(f->dgp_state); 
} 
 
 
