/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/jb2mmcompress.c,v 1.24 2003/01/27 21:48:34 awilliam Exp $ 
 * 
 */ 
 
#include <sys/types.h> 
#include "util.h" 
#include "mmcompress.h" 
#include "jb2mmcompressprivate.h" 
#include "symbol.h" 
#include <stdlib.h> 
#include "encsymdict.h" 
#include "encsymreg.h" 
#include "datasink.h" 
#include "PsgFix.h"
 
#define DGP_ENC DGP_FALSE

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/jb2mmcompress.c,v 1.24 2003/01/27 21:48:34 awilliam Exp $"); 
 
 
static void internal_convertTokensToState(void *dstate, Dgp_Int32 newntok, 
					     jb2Symbol **sortedsyms); 
 
/* Some guesses at byte sizes for stuff */ 
#define	TOKG4COMP	2	/* Guess at G4 compression for tokens */ 
#define	TOKG4SIZE	2	/* How many bytes token size might take */ 
#define	TOKOVERHEAD	60	/* Overhead per dictionary block */ 
#define	POSSIZE		4	/* How many bytes per position */ 
#define	POSOVERHEAD	100	/* Overhead per position block */ 
 
/* 
 * This file contains shim code to interface the compression side of DigiPaper 
 * to the JBIG2 encoder.  It provides 
 * ErrCode jb2_encodeTokDict(void *dstate, token *tokens, Dgp_Int32 ntok, 
 *		   Dgp_Bool (*extbuilder)(token **sorted_toks, 
 *					 Dgp_Int32 ntok, void *extcookie, 
 *					 void **extension, 
 *					 size_t *extlen), 
 *		   void *extcookie, 
 *		   void *outbuf, size_t outbuflen) 
 * 
 * size_t jb2_estimateEncodedTokDictSize(void *dstate, token *tokens, 
 *					  Dgp_Int32 ntok) 
 * 
 * ErrCode jb2_encodeTokPositions(void *dstate, tokpos *positions, Dgp_Int32 npos, 
 *			Dgp_Int32 lowx, Dgp_Int32 lowy, Dgp_Int32 highx, Dgp_Int32 highy, 
 *			Dgp_Bool (*extbuilder)(tokpos **sorted_tokpos, 
 *					      Dgp_Int32 ntok, 
 *					      void *extcookie, 
 *					      void **extension, 
 *					      size_t *extlen), 
 *			void *extcookie, 
 *			void *outbuf, size_t outbuflen) 
 * 
 * size_t jb2_estimateEncodedTokPosSize(void *dstate, tokpos *positions, 
 *					 Dgp_Int32 npos) 
 * 
 * which behave like the versions in libtokcomp. 
 */ 
 
/* 
 * An extension builder callback; all it does is copy out the array of 
 * sorted symbols, as the callback can be called later, and we need 
 * them outside this routine anyway. 
 */ 
typedef struct { 
    jb2Symbol **sortedsyms; 
    Dgp_State *dgp_state; 
    } my_symcookie; 
 
static ErrCode 
my_symextbuilder(jb2Symbol **sortedsyms, Dgp_Int32 nsyms, void *extcookie) 
{ 
    my_symcookie *c = (my_symcookie *)extcookie; 
    DGP_PROLOGUE("my_symextbuilder()", c->DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(c->DGP_ALLOCATORS); 
    jb2Symbol **sorted_copy = (jb2Symbol **)NULL; 
 
    sorted_copy = (jb2Symbol **)(a->malloc)(nsyms * sizeof(*sorted_copy)); 
    DGP_CHECK_ALLOC_CLEANUP(sorted_copy); 
 
    memcpy(sorted_copy, sortedsyms, nsyms * sizeof(*sorted_copy)); 
    c->sortedsyms = sorted_copy; 
 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
    if (sorted_copy != (jb2Symbol **)NULL) { 
	(a->free)((void *)sorted_copy); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Write out a token dictionary. Organise them by size and write out all 
 * the ones of a certain height together etc. Ignore any token that is 
 * marked "committed". 
 * 
 * Update the dictionary state block. 
 */ 
ErrCode 
jb2_encodeTokDict(void *dstate, token *tokens, Dgp_Int32 ntok, 
		   ErrCode (*extbuilder)(token **sorted_toks, 
					 Dgp_Int32 ntok, void *extcookie, 
					 void **extension, 
					 size_t *extlen, Dgp_State *dgp_state), 
		   void *extcookie, 
		   void *outbuf, size_t *outbuflen, Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jb2_encodeTokDict()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    jb2Symbol *syms = (jb2Symbol *)NULL; 
    token **sorted_toks = (token **)NULL; 
    size_t dummy; 
    void *dummybits; 
    Dgp_Int32 i, j; 
    Dgp_Int32 newntok; 
    jb2File file = (jb2File)NULL; 
    struct _DataSink ds; 
    jb2Segment segment; 
    dictstate *state = (dictstate *)dstate; 
    my_symcookie c; 
    jb2Symbol **sortedsyms = (jb2Symbol **)NULL; 
 
    /* 
     * How many of the tokens are new? 
     */ 
    newntok = 0; 
    for (i = 0; i < ntok; i++) { 
	if (!tokens[i].committed) { 
	    newntok++; 
	    } 
	} 
     
    /* Allocate the jb2Symbol structures */ 
    syms = (jb2Symbol *)(a->malloc)(newntok * sizeof(*syms)); 
    DGP_CHECK_ALLOC_CLEANUP(syms); 
 
    /* 
     * Copy the information in there. 
     * The userdata field of all the symbols will point to the original 
     * token. 
     */ 
    for (i = j = 0; i < ntok; i++) { 
	if (!tokens[i].committed) { 
	    syms[j].width = tokens[i].width; 
	    syms[j].height = tokens[i].height; 
	    syms[j].longsperline = tokens[i].longsperline; 
	    syms[j].bits = tokens[i].bits; 
	    syms[j].refsym = (jb2Symbol *)NULL; 
	    syms[j].userdata = (void *)&tokens[i]; 
 
	    j++; 
	    } 
	} 
		 
    /* Make up a simple DataSink pointed at the buffer we've been given */ 
    ds.buf = outbuf; 
    ds.bufsize = *outbuflen; 
    ds.bufpos = 0; 
    ds.count = 0; 
    ds.more_data = (ErrCode (*)(DataSink))NULL; 
    ds.mark = &jb2SimpleDataSinkMark; 
    ds.fillmark = &jb2SimpleDataSinkFillmark; 
	 
    /* Make up a jb2File */ 
    DGP_ERR_PASS_CLEANUP(jb2CreateFile(&ds, &ds, dgp_state, &file)); 
    jb2SetFileSegnum(file, state->segnum); 
 
    /* Make up a segment */ 
    segment.segtype = JB2SEGTYPE_SYMDICT; 
    segment.pagenum = 0; 
    segment.nrefsegs = 0; 
    segment.refsegs = (Dgp_Int32 *)NULL; 
    segment.refretain = (Dgp_Bool *)NULL; 
    segment.data.symdict.nsymbols = newntok; 
    segment.data.symdict.symbols = syms; 
    segment.data.symdict.direct_contexts = (Dgp_UInt8 *)NULL; 
    segment.data.symdict.ref_contexts = (Dgp_UInt8 *)NULL; 
 
    /* Call the segment creator */ 
    c.dgp_state = dgp_state; 
    DGP_ERR_PASS_CLEANUP(jb2EncodeSymDict(file, &segment, (jb2Segment **)NULL, 
					  my_symextbuilder, &c, DGP_FALSE)); 
     
    /* That may have allocated segment.refsegs, segment.refretain */ 
    if (segment.refsegs != (Dgp_Int32 *)NULL) { 
	(a->free)((void *)segment.refsegs); 
	} 
    if (segment.refretain != (Dgp_Bool *)NULL) { 
	(a->free)((void *)segment.refretain); 
	} 
 
    /* Call their callback, if they gave us one.  */ 
    sortedsyms = c.sortedsyms;	/* Our callback gave us the sorted array */ 
    if (extbuilder != (ErrCode (*)(token **, Dgp_Int32, 
				   void *, void **, size_t *, Dgp_State *))NULL) { 
	/* Allocate the sorted token pointers */ 
	sorted_toks = (token **)(a->malloc)(newntok * 
						   sizeof(*sorted_toks)); 
	DGP_CHECK_ALLOC_CLEANUP(sorted_toks); 
 
	/* Copy out the stuff in the symbols' userdata fields */ 
	for (i = 0; i < newntok; i++) { 
	    sorted_toks[i] = (token *)sortedsyms[i]->userdata; 
	    } 
 
	DGP_ERR_PASS_CLEANUP((*extbuilder)(sorted_toks, newntok, extcookie, 
					   &dummybits, &dummy, dgp_state)); 
	DGP_ASSERT(dummy == 0);	/* We can't cope with them actually generating data */ 
	(a->free)((void *)sorted_toks); 
	sorted_toks = (token **)NULL; 
	} 
    /* 
     * Now update the dictionary state block. 
     */ 
 
    /* Pull out the new segment number */ 
    state->segnum = jb2GetFileSegnum(file); 
    /* Mark the tokens as committed, and remember their order */ 
    internal_convertTokensToState(dstate, newntok, sortedsyms); 
 
    /* Clean up all the stuff we built */ 
    (a->free)((void *)sortedsyms); 
    (a->free)((void *)syms); 
    jb2FreeFile(file); 
 
    /* How much space did we use? */ 
    *outbuflen = ds.bufpos; 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    if (sortedsyms != (jb2Symbol **)NULL) { 
	(a->free)((void *)sortedsyms); 
	} 
 
    if (sorted_toks != (token **)NULL) { 
	(a->free)((void *)sorted_toks); 
	} 
 
    if (syms != (jb2Symbol *)NULL) { 
	(a->free)((void *)syms); 
	} 
 
    if (file != (jb2File)NULL) { 
	jb2FreeFile(file); 
	} 
 
    /* Should guess how much more space they should feed us. FIX */ 
    DGP_REPORTF0(DGP_ENC, "encodeTokDict bailing for more memory\n"); 
    *outbuflen += 50000; 
    /* Should pass back any non-out-of-memory errors */ 
    ERR_RETURN; 
    } 
 
/* 
 * This routine takes a dictionary state and a bunch of tokens and figures out 
 * how much memory (in bytes) is required to encode those tokens. This is only 
 * an estimate - see encodeTokDict. 
 */ 
size_t 
jb2_estimateEncodedTokDictSize(void *dstate, token *tokens, Dgp_Int32 ntok) 
{ 
    Dgp_Int32 i; 
    size_t totsize = 0; 
 
    DGP_VOID_PROLOGUE ("jb2_estimateEncodedTokDictSize", NULL); 
 
    /* They're used, really */ 
    dstate = dstate; 
 
    for (i = 0; i < ntok; i++) { 
	if (!tokens[i].committed) { 
	    /* Going to encode this one. */ 
	    totsize += TOKG4SIZE + 
		tokens[i].width * tokens[i].height / DGP_BITS_UINT8 / TOKG4COMP; 
	    } 
	} 
 
    /* Account for some overhead */ 
    totsize += TOKOVERHEAD; 
 
    DGP_REPORTF1(DGP_ENC, "estimating %d bytes for dict\n", totsize); 
    return(totsize); 
    } 
 
/* 
 * An extension builder callback; all it does is copy out the array of 
 * sorted symbol instances, as the callback can be called later, and we need 
 * them outside this routine anyway. 
 */ 
typedef struct { 
    jb2SymbolInstance **sortedinsts; 
  Dgp_State *dgp_state; 
    } my_instcookie; 
 
static ErrCode 
my_instextbuilder(jb2SymbolInstance **sortedinsts, Dgp_Int32 ninsts, void *extcookie) 
{ 
    my_instcookie *c = (my_instcookie *)extcookie; 
    DGP_PROLOGUE("my_instextbuilder()", c->DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(c->DGP_ALLOCATORS); 
    jb2SymbolInstance **sorted_copy = (jb2SymbolInstance **)NULL; 
 
    sorted_copy = (jb2SymbolInstance **)(a->malloc)(ninsts * sizeof(*sorted_copy)); 
    DGP_CHECK_ALLOC_CLEANUP(sorted_copy); 
 
    memcpy(sorted_copy, sortedinsts, ninsts * sizeof(*sorted_copy)); 
    c->sortedinsts = sorted_copy; 
 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
    if (sorted_copy != (jb2SymbolInstance **)NULL) { 
	(a->free)((void *)sorted_copy); 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * Write out all the boxes. 
 * 
 * Special case: dstate can be NULL if npos is 0. 
 */ 
ErrCode 
jb2_encodeTokPositions(void *dstate, tokpos *positions, Dgp_Int32 npos, 
			Dgp_Int32 lowx, Dgp_Int32 lowy, Dgp_Int32 highx, Dgp_Int32 highy, 
			ErrCode (*extbuilder)(tokpos **sorted_tokpos, 
					      Dgp_Int32 ntok, 
					      void *extcookie, 
					      void **extension, 
					      size_t *extlen, 
					      Dgp_State *dgp_state), 
			void *extcookie, 
			void *outbuf, size_t *outbuflen, 
			Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jb2_encodeTokPositions()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    dictstate *state = dstate; 
    tokpos **sorted_tokpos = (tokpos **)NULL; 
    void **userdata_copy = (void **)NULL; 
    size_t dummy; 
    void *dummybits; 
    jb2File file = (jb2File)NULL; 
    struct _DataSink ds; 
    jb2Segment segment; 
    jb2Segment dictseg; 
    my_instcookie c; 
    jb2SymbolInstance **sortedinsts = (jb2SymbolInstance **)NULL; 
    jb2SymbolInstance *insts = (jb2SymbolInstance *)NULL; 
    Dgp_Int32 ntok; 
    jb2Symbol *toks; 
    Dgp_Int32 i; 
    Dgp_Int32 *refsegnums = (Dgp_Int32 *)NULL; 
    Dgp_Bool *refretain = (Dgp_Bool *)NULL; 
    jb2Segment *refsegs; 
 
    /* Allocate the jb2SymbolInstance structures */ 
    insts = (jb2SymbolInstance *)(a->malloc)(npos * sizeof(*insts)); 
    DGP_CHECK_ALLOC_CLEANUP(insts); 
 
    /* 
     * We have a mess of tokpos that we need to write out.  Each one 
     * includes a pointer to a token structure.  We need to transmogrify 
     * this into a mess of jb2SymbolInstance with each one containing a 
     * pointer to a jb2Symbol. 
     * 
     * In order to do this, we need some way to translate an arbitrary 
     * token * value into the corresponding jb2Symbol *.  However, we do 
     * know one thing: the token * values all appear in our dictionary state 
     * block.  We can hijack their userdata fields temporarily: 
     * - look up the "toks" block in the dictstate: this is a 
     *   block of jb2Symbols with all the right information in them already. 
     *   This includes the jb2Symbols userdata pointer, which points at 
     *   the relevant token structure. 
     * - for each token in the state block, copy off the userdata field 
     *   and replace it with a pointer to the jb2Symbol; make the jb2Symbol 
     *   carry the same data as the token. 
     * - create a block of jb2SymbolInstances. 
     * - walk through the tokpos array to create each jb2SymbolInstance, 
     *   indexing through the token * to the jb2Symbol.  Also make the 
     *   jb2SymbolInstance's userdata field point back at the tokpos; this 
     *   makes life easier when building extensions. 
     */ 
 
    if (dstate != (void *)NULL) { 
	ntok = state->ntok; 
	toks = state->toks; 
	} 
    else { 
	DGP_ASSERT(npos == 0); 
	ntok = 0; 
	toks = (jb2Symbol *)NULL; 
	} 
 
    userdata_copy = 
	(void **)(a->malloc)(ntok * sizeof(*userdata_copy)); 
    DGP_CHECK_ALLOC_CLEANUP(userdata_copy); 
 
    /* Read out the old userdata and write the new one. */ 
    for (i = 0; i < ntok; i++) { 
	token *curtok = (token *)toks[i].userdata; 
	userdata_copy[i] = curtok->extdata; 
	curtok->extdata = &toks[i]; 
	} 
 
    /* 
     * Now we can make up the jb2SymbolInstances. 
     * 
     * Don't forget that positions is using LL convention, and insts is using 
     * UL convention. 
     */ 
    for (i = 0; i < npos; i++) { 
	token *curtok = positions[i].tok; 
	jb2Symbol *cursym = (jb2Symbol *)curtok->extdata; 
 
	insts[i].xpos = positions[i].xpos; 
	insts[i].ypos = 
	    positions[i].ypos - cursym->height + 1; 
	insts[i].sym = cursym; 
	insts[i].refdata = (jb2SymbolInstanceRefined *)NULL; 
	insts[i].userdata = (void *)&positions[i]; 
	} 
 
    /* Make up a simple DataSink pointed at the buffer we've been given */ 
    ds.buf = outbuf; 
    ds.bufsize = *outbuflen; 
    ds.bufpos = 0; 
    ds.count = 0; 
    ds.more_data = (ErrCode (*)(DataSink))NULL; 
    ds.mark = &jb2SimpleDataSinkMark; 
    ds.fillmark = &jb2SimpleDataSinkFillmark; 
	 
    /* Make up a jb2File */ 
    DGP_ERR_PASS_CLEANUP(jb2CreateFile(&ds, &ds, dgp_state, &file)); 
    jb2SetFileSegnum(file, state->segnum); 
 
    /* 
     * Create a segment that is our "dictionary segment": we don't have 
     * to remember all the actual dictionary segments; we'll just fake 
     * one up.  The order of the symbols is correct, and that's all that 
     * matters. 
     * 
     * This isn't legit since the referenced segment numbers won't be 
     * correct, but we can't fix that until we get some hooks into the 
     * file IO level.  FIX. 
     */ 
    dictseg.segtype = JB2SEGTYPE_SYMDICT; 
    dictseg.segnum = 0; 
    dictseg.data.symdict.nsymbols = ntok; 
    dictseg.data.symdict.symbols = toks; 
    dictseg.data.symdict.direct_contexts = (Dgp_UInt8 *)NULL; 
    dictseg.data.symdict.ref_contexts = (Dgp_UInt8 *)NULL; 
 
    /* 
     * Make up some reference info arrays.  These can't be 
     * pointers into the stack because if more referenced segments 
     * get added (for example, Huffman table segments get generated), 
     * the existing arrays will be freed. 
     */ 
    refsegnums = (Dgp_Int32 *)(a->malloc)(sizeof(*refsegnums)); 
    DGP_CHECK_ALLOC_CLEANUP(refsegnums); 
    *refsegnums = 0;		/* 0 is the segnum for our fake dict seg */ 
 
    refretain = (Dgp_Bool *)(a->malloc)(sizeof(*refretain)); 
    DGP_CHECK_ALLOC_CLEANUP(refretain); 
    *refretain = DGP_TRUE;		/* might as well not tempt fate */ 
 
    /* Make up our symbol region segment */ 
    segment.segtype = JB2SEGTYPE_SYMREG_IMM; 
    segment.pagenum = state->pagenum + 1; 
    segment.nrefsegs = 1; 
    segment.refsegs = refsegnums; 
    refsegnums = (Dgp_Int32 *)NULL;	/* segment manager now owns it */ 
    segment.refretain = refretain; 
    refretain = (Dgp_Bool *)NULL; 
    segment.reginfo.regwidth = highx - lowx; 
    segment.reginfo.regheight = highy - lowy; 
    segment.reginfo.regxpos = lowx; 
    segment.reginfo.regypos = lowy; 
    segment.reginfo.combop = COMBOP_OR; 
    segment.data.symreg.ninstances = npos; 
    segment.data.symreg.instances = insts; 
    segment.data.symreg.int_combop = COMBOP_OR; 
    segment.data.symreg.defpixel = 0; 
 
    c.dgp_state = dgp_state; 
    /* Call the segment creator */ 
    if (dstate == (void *)NULL) { 
	/* Hack for the special case */ 
	segment.nrefsegs = 0; 
 
	DGP_ERR_PASS_CLEANUP(jb2EncodeSymRegion(file, &segment, (jb2Segment **)NULL, 
						my_instextbuilder, &c, DGP_FALSE)); 
	} 
    else { 
	refsegs = &dictseg; 
	DGP_ERR_PASS_CLEANUP(jb2EncodeSymRegion(file, &segment, &refsegs, 
						my_instextbuilder, &c, DGP_FALSE)); 
	} 
     
    /* 
     * Get rid of the segment info - refsegnums and refretain 
     * or whatever they were replaced by. 
     */ 
    if (segment.refsegs != (Dgp_Int32 *)NULL) { 
	(a->free)((void *)segment.refsegs); 
	} 
    if (segment.refretain != (Dgp_Bool *)NULL) { 
	(a->free)((void *)segment.refretain); 
	} 
 
    /* Call their callback, if they gave us one.  */ 
    sortedinsts = c.sortedinsts; 
    if (extbuilder != (ErrCode (*)(tokpos **, Dgp_Int32, 
				   void *, void **, size_t *, Dgp_State *))NULL) { 
	/* They gave us a callback - call it */ 
	sorted_tokpos = 
	    (tokpos **)(a->malloc)(npos * sizeof(*sorted_tokpos)); 
	DGP_CHECK_ALLOC_CLEANUP(sorted_tokpos); 
 
	/* Copy out the stuff in the symbols' userdata fields */ 
	for (i = 0; i < npos; i++) { 
	    sorted_tokpos[i] = (tokpos *)sortedinsts[i]->userdata; 
	    } 
 
	DGP_ERR_PASS_CLEANUP((*extbuilder)(sorted_tokpos, npos, extcookie, 
					   &dummybits, &dummy, dgp_state)); 
	DGP_ASSERT(dummy == 0); 
 
	(a->free)((void *)sorted_tokpos); 
	sorted_tokpos = (tokpos **)NULL; 
	} 
     
    /* Update the state - now after all the failure points */ 
    state->pagenum++; 
    state->segnum = jb2GetFileSegnum(file); 
 
    /* Put the userdata values back */ 
    for (i = 0; i < ntok; i++) { 
	token *curtok = (token *)toks[i].userdata; 
	curtok->extdata = userdata_copy[i]; 
	} 
 
    /* Clean up all the stuff we built */ 
    (a->free)((void *)sortedinsts); 
    (a->free)((void *)insts); 
    (a->free)((void *)userdata_copy); 
    jb2FreeFile(file); 
 
    /* How much space did we use? */ 
    *outbuflen = ds.bufpos; 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
    if (sortedinsts != (jb2SymbolInstance **)NULL) { 
	(a->free)((void *)sortedinsts); 
	} 
 
    if (sorted_tokpos != (tokpos **)NULL) { 
	(a->free)((void *)sorted_tokpos); 
	} 
 
    if (userdata_copy != (void **)NULL) { 
	(a->free)((void *)userdata_copy); 
	} 
 
    if (insts != (jb2SymbolInstance *)NULL) { 
	(a->free)((void *)insts); 
	} 
 
    if (refsegnums != (Dgp_Int32 *)NULL) { 
	(a->free)((void *)refsegnums); 
	} 
 
    if (refretain != (Dgp_Bool *)NULL) { 
	(a->free)((void *)refretain); 
	} 
 
    if (file != (jb2File)NULL) { 
	jb2FreeFile(file); 
	} 
 
    /* Should guess how much more space they should feed us. FIX */ 
    DGP_REPORTF0(DGP_ENC, "encodeTokPositions bailing for more memory\n"); 
    *outbuflen += 50000; 
    /* Should pass back any non-out-of-memory errors */ 
    ERR_RETURN; 
    } 
 
/* 
 * This routine takes a dictionary state (which must have been properly filled 
 * in by encodeTokDict; one which has just been initialised by 
 * initializeTokDictState isn't enough), and a bunch of token positions. It 
 * figures out an estimate of how much space is required to encode those 
 * positions and returns that. 
 */ 
size_t 
jb2_estimateEncodedTokPosSize(void *dstate, tokpos *positions, Dgp_Int32 npos) 
{ 
 
  DGP_VOID_PROLOGUE ("jb2_estimateEncodedTokPosSize", NULL); 
 
	/* Stop compiler complaints */ 
    dstate = dstate; 
    positions = positions; 
 
    DGP_REPORTF1(DGP_ENC, "estimating %d bytes for pos\n", POSOVERHEAD + POSSIZE * npos); 
    return(POSOVERHEAD + POSSIZE * npos); 
    } 
  
 
/* 
 * This actually does the work of tcmp_convertTokensToState; it's a separate 
 * routine because it's also called from jb2_encodeTokDict, where newntok is 
 * already known so it shouldn't be recomputed. Also the ordering of the 
 * *new* tokens is significant... a mapping for them can be found in 
 * ordered (i.e., ordered[i] is the i'th token in some canonical ordering, 
 * which might not be the ordering in the dstate->toks array). The ordering 
 * of the old tokens is not as hard, since they're written into the 
 * dictionary state block in the canonical ordering. 
 * 
 * It stores away a pointer to the token in the jb2Symbol's userdata field. 
 */ 
static void 
internal_convertTokensToState(void *dstate, Dgp_Int32 newntok, 
			      jb2Symbol **sortedsyms) 
{ 
    Dgp_Int32 i; 
    dictstate *state = dstate; 
    jb2Symbol *oldtoks = state->toks; 
    Dgp_Int32 oldntok; 
 
    /* Total number of tokens now. */ 
    oldntok = state->ntok; 
    state->ntok = oldntok + newntok; 
 
    /* Remember all the info we need, and mark these things as "committed" */ 
    for (i = 0; i < newntok; i++) { 
	token *curtok = (token *)sortedsyms[i]->userdata; 
	/* This is where we do the permutation by sortedsyms. */ 
	oldtoks[i + oldntok].longsperline = 0; 
	oldtoks[i + oldntok].width = curtok->width; 
	oldtoks[i + oldntok].height = curtok->height; 
	oldtoks[i + oldntok].bits = (Dgp_UInt32 *)NULL; 
	oldtoks[i + oldntok].refsym = (jb2Symbol *)NULL; 
	oldtoks[i + oldntok].userdata = (void *)curtok; 
	curtok->committed = DGP_TRUE; 
	} 
    } 


