/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/jb2mmtokdict.c,v 1.9 2001/07/24 16:26:42 mkrocken Exp $ 
 * 
 */ 
 
#include <sys/types.h> 
#include "util.h" 
#include "mmcompress.h" 
#include "jb2mmcompressprivate.h" 
#include "symbol.h" 
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/jb2mmtokdict.c,v 1.9 2001/07/24 16:26:42 mkrocken Exp $"); 

#if 0 
/* 
 * Given a dictionary state block ("oldstate"), which may be NULL, and a bunch 
 * of tokens, this routine figures out how much dictionary state memory (in 
 * bytes) is required to hold the new tokens *plus* the stuff in the old 
 * state. This is what it returns. 
 */ 
size_t 
tcmp_estimateTokDictStateSize(void *oldstate, token *tokens, Dgp_Int32 ntok) 
{ 
    Dgp_Int32 i, j; 
    dictstate *state = oldstate; 
     
    /* How many new tokens? */ 
    j = 0; 
    for (i = 0; i < ntok; i++) { 
	if (!tokens[i].committed) { 
	    j++; 
	    } 
	} 
 
    if (state != (dictstate *)NULL) { 
	j += state->ntok; 
	} 
 
    return(sizeof(dictstate) + j * sizeof(jb2Symbol)); 
    } 
 
/* 
 * Given a bunch of dictionary state blocks, this routine figures out how 
 * much dictionary state memory (in bytes) is required to hold the union 
 * of those previous states.  This is what it returns. 
 */ 
size_t 
tcmp_estimateMergedTokDictStateSize(void **states, Dgp_Int32 nstates) 
{ 
    Dgp_Int32 i; 
    Dgp_Int32 ntok = 0; 
    dictstate *state; 
     
    /* How many new tokens? */ 
    for (i = 0; i < nstates; i++) { 
	state = (dictstate *)states[i]; 
	ntok += state->ntok; 
	} 
 
    return(sizeof(dictstate) + ntok * sizeof(jb2Symbol)); 
    } 
 
/* 
 * Given a dictionary state block ("oldstate"), some newly 
 * allocated dictionary state memory, and the size of that memory block, 
 * initialise the new state block, and copy over all the stuff from the old 
 * state block. The old state block should then be discarded. 
 * 
 * oldstate and newstatemem can be the same pointer, but should not otherwise 
 * overlap. 
 * 
 * oldstate can be NULL. 
 */ 
void 
tcmp_initializeTokDictState(void *oldstate, void *newstatemem, size_t memsize) 
{ 
  DGP_VOID_PROLOGUE("tcmp_initializeTokDictState()", NULL); 
 
    dictstate *olddstate = oldstate; 
    dictstate *newdstate = newstatemem; 
    Dgp_Int32 ntokcapacity; 
    jb2Symbol *oldtoks; 
 
    /* 
     * How many tokens can the new state memory hold? 
     * This is its capacity, not the number it's currently holding; 
     * that number is olddstate->ntok. 
     */ 
    ntokcapacity = 
	(memsize - sizeof(dictstate)) / sizeof(jb2Symbol); 
    if (ntokcapacity * sizeof(jb2Symbol) + sizeof(dictstate) != 
	memsize) { 
	/* Where did they get that memory size? */ 
	DGP_REPORTF0("bad token state block size"); 
	} 
    /* Make sure all the old stuff still fits. */ 
    DGP_VOID_ASSERT((oldstate == (void *)NULL) || (ntokcapacity >= olddstate->capacity)); 
 
    /* Patch up the pointers */ 
    /* Cross your fingers and hope for alignment... */ 
    newdstate->toks = 
	(jb2Symbol *)((void *)((Dgp_Int8 *)newstatemem + sizeof(dictstate))); 
 
    if (oldstate == newstatemem) { 
	/* 
	 * The block may have moved... what probably happened is 
	 * that they did a realloc() and passed in the value returned 
	 * as both "old" and "new" (this is quite legal). realloc() 
	 * might have extended the memory area in place, or copied 
	 * it over to a different location. 
	 * We can't trust oldtoks, since they 
	 * could refer to the block's old location. 
	 * The token info block shouldn't have to be moved. 
	 */ 
	oldtoks = (jb2Symbol *)((void *)((Dgp_Int8 *)olddstate +  
					 sizeof(dictstate))); 
	DGP_VOID_ASSERT(oldtoks == newdstate->toks); 
	/* 
	 * The rest of the dictstate block header (ntok and firstcodes) 
	 * shouldn't need to be touched - realloc will have copied them. 
	 */ 
	} 
    else if (oldstate == (void *)NULL) { 
	/* Just do an init. */ 
	newdstate->ntok = 0; 
	newdstate->segnum = newdstate->pagenum = 0; 
	} 
    else { 
	/* Copy it across then patch pointers. */ 
	/* Can't overlap, so memcpy is safe. */ 
	/* Get the pointers to the old memory - they're still valid */ 
	oldtoks = olddstate->toks; 
	memcpy(newdstate->toks, oldtoks, olddstate->ntok * sizeof(jb2Symbol)); 
 
	newdstate->ntok = olddstate->ntok; 
	newdstate->segnum = olddstate->segnum; 
	newdstate->pagenum = olddstate->pagenum; 
	} 
 
    /* Record the new capacity. */ 
    newdstate->capacity = ntokcapacity; 
    } 
 
/* 
 * Given a bunch of dictionary state blocks, some newly 
 * allocated dictionary state memory, and the size of that memory block, 
 * initialise the new state block, and copy over all the stuff from the old 
 * state blocks. 
 */ 
ErrCode 
tcmp_mergeTokDictStates(void *newstate, size_t newsize, 
			void **states, Dgp_Int32 nstates, Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("tcmp_mergeTokDictStates()", DGP_PERRSTATE); 
 
    dictstate *olddstate; 
    dictstate *newdstate = newstate; 
    Dgp_Int32 ntokcapacity; 
    Dgp_Int32 ntok; 
    jb2Symbol *toks; 
    Dgp_Int32 i, j, toknum; 
    Dgp_Int32 maxsegnum = 0, maxpagenum = 0; 
 
    /* 
     * How many tokens can the new state memory hold? 
     */ 
    ntokcapacity = 
	(newsize - sizeof(dictstate)) / sizeof(jb2Symbol); 
 
    /* Where did they get that memory size?  bad token state block size.. */ 
    DGP_ERR_GEN((ntokcapacity * sizeof(jb2Symbol) + 
			 sizeof(dictstate) != newsize), cErrInvalidParm); 
 
    newdstate->capacity = ntokcapacity; 
 
    /* How many new tokens? */ 
    ntok = 0; 
    for (i = 0; i < nstates; i++) { 
	olddstate = (dictstate *)states[i]; 
	ntok += olddstate->ntok; 
	} 
 
    /* Make sure all the old stuff still fits. */ 
    DGP_ASSERT(ntokcapacity >= ntok); 
 
    /* Patch up the pointers */ 
    /* Cross your fingers and hope for alignment... */ 
    newdstate->toks = toks = 
	(jb2Symbol *)((void *)((Dgp_Int8 *)newstate + sizeof(dictstate))); 
 
    /* Copy all the tokens across to the new block */ 
    toknum = 0; 
    for (i = 0; i < nstates; i++) { 
	olddstate = (dictstate *)states[i]; 
 
	maxsegnum = MAX(maxsegnum, olddstate->segnum); 
	maxpagenum = MAX(maxpagenum, olddstate->pagenum); 
 
	for (j = 0; j < olddstate->ntok; j++, toknum++) { 
	    toks[toknum] = olddstate->toks[j]; 
	    } 
 
	} 
 
    /* Update count of tokens in state */ 
    newdstate->ntok = ntok; 
    newdstate->segnum = maxsegnum; 
    newdstate->pagenum = maxpagenum; 
 
    NORMAL_RETURN; 
    } 
 
/* 
 * Given a bunch of tokens and a dictionary state (that had better have 
 * room and be properly initialised), find the uncommitted tokens and 
 * record the pertinent bits about them in *dstate, and mark them committed. 
 * 
 * Don't change the order of the tokens; use their existing order. 
 * 
 * It stores away a pointer to the token in the jb2Symbol's userdata field. 
 */ 
ErrCode 
tcmp_convertTokensToState(token *tokens, Dgp_Int32 ntok, void *dstate, 
			  Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("tcmp_convertTokensToState()", DGP_PERRSTATE); 
 
    Dgp_Int32 i; 
    Dgp_Int32 oldntok, newntok; 
    dictstate *state = dstate; 
    jb2Symbol *oldtoks = state->toks; 
 
    /* whine, whine, whine */ 
    DGP_STOP_WHINE; 
 
    oldntok = state->ntok; 
 
    /* How many of the tokens are new? */ 
    newntok = 0; 
    for (i = 0; i < ntok; i++) { 
	if (!tokens[i].committed) { 
	    /* Remember stuff about this token */ 
	    token *curtok = &tokens[i]; 
 
	    oldtoks[newntok + oldntok].longsperline = 0; 
	    oldtoks[newntok + oldntok].width = curtok->width; 
	    oldtoks[newntok + oldntok].height = curtok->height; 
	    oldtoks[newntok + oldntok].bits = (Dgp_UInt32 *)NULL; 
	    oldtoks[newntok + oldntok].refsym = (jb2Symbol *)NULL; 
	    oldtoks[newntok + oldntok].userdata = (void *)curtok; 
	    curtok->committed = DGP_TRUE; 
 
	    newntok++; 
	    } 
	} 
 
    /* Total number of tokens now. */ 
    state->ntok = oldntok + newntok; 
 
    NORMAL_RETURN; 
    } 
 
#endif 
 
 
 
 
 
 
 
 
