/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/jb2mmuncompress.c,v 1.34 2003/02/10 22:53:26 mkrocken Exp $ 
 * 
 */ 
 
#include <sys/types.h> 
#include "util.h" 
#include "mmcompress.h" 
#include "jb2mmcompressprivate.h" 
#include "jb2fileprivate.h" 
#include "symbol.h" 
#include <stdlib.h> 
#include "decsymdict.h" 
#include "decsymreg.h"
#include "datasource.h" 
#include "list.h" 
#include "segment_util.h" 
#include "segment.h" 
#include "image_defs.h"
#include "PsgFix.h"
#include <stdio.h> 
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/jb2mmuncompress.c,v 1.34 2003/02/10 22:53:26 mkrocken Exp $"); 

#if 0 
static ErrCode 
jb2_addTokens(jb2Segment *curseg, Dgp_State *dgp_state) ;
#endif
 
/* 
 * This file contains shim code to interface the decompression side of 
 * DigiPaper to the JBIG2 encoder.  It provides 
 * 
 * Dgp_Int32 tcmp_getTokenCount(void *buffer, size_t buflen) 
 * 
 * Dgp_Bool tcmp_decodeTokDict(token *tokens, void *buffer, size_t buflen, 
 *		Dgp_Int32 firsttoknum, void *(*allocator)(size_t allocsize)) 
 * 
 * Dgp_Int32 tcmp_getTokPosCount(void *buffer, size_t buflen) 
 * 
 * Dgp_Bool tcmp_decodeTokPositions(jb2SymbolInstance *positions, 
 *			Dgp_Int32 lowx, Dgp_Int32 lowy, Dgp_Int32 highx, Dgp_Int32 highy, 
 *			void *buffer, size_t buflen, void *dstate) 
 * 
 * which behave like the versions in libtokcomp. 
 */ 
 
 
/* 
 * This routine takes an encoded dictionary block and returns the number of 
 * tokens encoded in it. 
 * 
 * It has to read through the (variable-length) segment header, then 
 * parse the (variable-length) segment data header to find the value it needs. 
 * 
 * Actually, it needs to skip over segments until it reaches a symdict 
 * segment; it may be preceded by Huffman table segments, and so on. 
 */ 
ErrCode 
jb2_getTokenCount(void *buffer, size_t buflen, Dgp_State *dgp_state, 
		   size_t *count) 
{ 
  DGP_PROLOGUE("tcmp_getTokenCount()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    jb2Segment segment; 
    struct _jb2File file; 
    struct _DataSource ds; 
    Dgp_UInt8 buf[8]; 
    Dgp_UInt16 flags; 
    size_t bytes; 
 
    /* Make up a jb2File */ 
    file.lastsegnum = 0; 
    file.headersink = file.bodysink = (DataSink)NULL; 
    file.dgp_state = dgp_state; 
 
    ds.buf = buffer; 
    ds.bufsize = buflen; 
    ds.bufpos = 0; 
    ds.more_data = (ErrCode (*)(DataSource))NULL; 
    ds.userdata = (void *)NULL; 
	ds.dgp_state = dgp_state;
 
    for (;;) { 
      DGP_ERR_PASS_CLEANUP(jb2DecodeSegmentHeader(&file, &ds, &segment)); 
	/* Free the unneeded stuff in segment */ 
	(a->free)((void *)segment.refsegs); 
	(a->free)((void *)segment.refretain); 
 
	if (segment.segtype != JB2SEGTYPE_SYMDICT) { 
	    /* Skip over the body */ 
	    ds.bufpos += segment.datalen; 
 
	    /* Ran out of segments without hitting a dictionary! */ 
	    DGP_ERR_GEN_CLEANUP((ds.bufpos >= ds.bufsize), cErrBufferExceeded); 
	    } 
	else { 
	    /* Found it */ 
	    break; 
	    } 
	} 
 
    /* 
     * If SDHUFF is 1, there's two bytes of flags.  If SDHUFF is 0, we also 
     * need to look at SDTEMPLATE; if SDREFAGG is 1, we need to look at 
     * SDRTEMPLATE: 
     * SDHUFF = 0 && SDTEMPLATE = 0 implies an 8-byte field 
     * SDHUFF = 0 && SDTEMPLATE != 0 implies a 2-byte field 
     * SDREFAGG != 0 && SDRTEMPLATE = 0 implies a 4-byte field 
     * SDREFAGG != 0 && SDRTEMPLATE != 0 implies a 0-byte field 
     */ 
    DSREAD(&ds, 2, buf, &bytes); 
    flags = (buf[0] << 8) | buf[1]; 
    if (flags & 1) { 
	/* SDHUFF is 1 - all done */ 
	} 
    else { 
	if (flags & (3 << 10)) { 
	    /* SDTEMPLATE is not 0 */ 
	    DSREAD(&ds, 2, buf, &bytes); 
	    } 
	else { 
	    DSREAD(&ds, 8, buf, &bytes); 
	    } 
	} 
 
    if (flags & 2) { 
	/* SDREFINE is 1 - check the refinement template */ 
	if (flags & (1 << 12)) { 
	    /* SDRTEMPLATE is not 0 */ 
	    } 
	else { 
	    DSREAD(&ds, 4, buf, &bytes); 
	    } 
	} 
 
    /* Next 4 bytes are what we want */ 
    DSREAD(&ds, 4, buf, &bytes); 
    *count = (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3]; 
 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
    /* No cleanup, at least... */ 
    *count = -1; 
    ERR_RETURN; 
    } 
 
/* 
 * Decode a token dictionary.  This involves: 
 * - Reading any segments that precede the symbol dictionary segment 
 * - Assembling those segments 
 * - Calling the symbol dictionary decoding procedure 
 * - Reformatting all the jb2Symbols into tokens. 
 */ 
ErrCode 
jb2_decodeTokDict(token *tokens, void *buffer, size_t buflen, 
		   Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("tcmp_decodeTokDict()", DGP_PERRSTATE); 
   
  Dgp_Allocators *a = &(dgp_state->a); 
    struct _jb2File file; 
    struct _DataSource ds; 
    DgpAdt_List refseglist = DGP_NULLLIST; 
    jb2Segment **refsegs = (jb2Segment **)NULL; 
    jb2Segment *curseg = (jb2Segment *)NULL; 
    jb2Symbol *syms, *cursym; 
    Dgp_Int32 i; 
    DgpAdt_ListNode ln, retnode; 
 
    /* A DgpAdt_List to put the segments we build on */ 
    DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&refseglist, dgp_state)); 
 
    /* Make up a jb2File */ 
    file.lastsegnum = 0; 
    file.headersink = file.bodysink = (DataSink)NULL; 
	file.dgp_state = dgp_state;
 
    /* and a DataSource */ 
    ds.buf = buffer; 
    ds.bufsize = buflen; 
    ds.bufpos = 0; 
    ds.more_data = (ErrCode (*)(DataSource))NULL; 
    ds.userdata = (void *)NULL; 
	ds.dgp_state = dgp_state;
 
    /* Read and decode all the segments */ 
    for (;;) { 
	curseg = (jb2Segment *)(a->malloc)(sizeof(*curseg)); 
	DGP_CHECK_ALLOC_CLEANUP(curseg); 
 
	DGP_ERR_PASS_CLEANUP(jb2DecodeSegmentHeader(&file, &ds, curseg)); 
 
	/* Collect all the segments this one references */ 
	DGP_ERR_PASS_CLEANUP(jb2MarshalSegments(&file, curseg, refseglist, 
						&refsegs)); 
 
	DGP_ERR_PASS_CLEANUP(jb2DecodeSegment(&file, &ds, curseg, refsegs)); 
 
	/* It's decoded - don't need refsegs */ 
	if (refsegs != (jb2Segment **)NULL) { 
	    (a->free)((void *)refsegs); 
	    refsegs = (jb2Segment **)NULL; 
	    } 
 
	/* Was it the one we wanted? */ 
	if (curseg->segtype == JB2SEGTYPE_SYMDICT) { 
	    /* Done */ 
	    break; 
	    } 
 
	/* Add it to the list of referenced segments */ 
	DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(refseglist, (void *)curseg, &retnode)); 
	/* Owned by the list now */ 
	curseg = (jb2Segment *)NULL; 
	} 
	 
    /* curseg is now the decoded symbol dictionary segment - copy sym info */ 
    syms = curseg->data.symdict.symbols; 
    for (i = 0; i < curseg->data.symdict.nsymbols; i++) { 
	cursym = &syms[i]; 
	tokens[i].width = cursym->width; 
	tokens[i].height = cursym->height; 
	tokens[i].bits = cursym->bits; 
	cursym->bits = (Dgp_UInt32 *)NULL; /* take ownership */ 
	tokens[i].longsperline = cursym->longsperline; 
	tokens[i].uses = 0; 
	tokens[i].committed = DGP_FALSE; 
	tokens[i].extdata = (void *)NULL; 
	cursym->userdata = &tokens[i];
	} 
 
    /* Free what we have to */ 
    /* need the bits from the symbols in curseg */ 
    (a->free)((void *)curseg->refsegs); 
    (a->free)((void *)curseg->refretain); 
    (a->free)((void *)curseg->data.symdict.symbols); 
    (a->free)((void *)curseg); 
    DGPADT_FOREACH (ln, refseglist) { 
	curseg = (jb2Segment *)DGPADT_LIGET(ln); 
	jb2FreeSegmentInfo((void *)curseg, dgp_state); 
	(a->free)((void *)curseg); 
	} 
    DgpAdt_liFree(refseglist); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    if (curseg != (jb2Segment *)NULL) { 
	/* Not safe to jb2FreeSegmentInfo on curseg, unfortunately */ 
	if (curseg->refsegs != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)curseg->refsegs); 
	    } 
	if (curseg->refretain != (Dgp_Bool *)NULL) { 
	    (a->free)((void *)curseg->refretain); 
	    } 
	(a->free)((void *)curseg); 
	} 
 
    if (refseglist != DGP_NULLLIST) { 
	DGPADT_FOREACH (ln, refseglist) { 
	    curseg = (jb2Segment *)DGPADT_LIGET(ln); 
	    jb2FreeSegmentInfo((void *)curseg, dgp_state); 
	    (a->free)((void *)curseg); 
	    } 
	DgpAdt_liFree(refseglist); 
	} 
 
    if (refsegs != (jb2Segment **)NULL) { 
	(a->free)((void *)refsegs); 
	} 
 
    ERR_RETURN;  
    } 

ErrCode 
jb2_decodeTokDictSeg (jb2Segment **segment, void *buffer, size_t buflen, 
		   Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("tcmp_decodeTokDict()", DGP_PERRSTATE); 
   
  Dgp_Allocators *a = &(dgp_state->a); 
    struct _jb2File file; 
    struct _DataSource ds; 
    DgpAdt_List refseglist = DGP_NULLLIST; 
    jb2Segment **refsegs = (jb2Segment **)NULL; 
    jb2Segment *curseg = (jb2Segment *)NULL; 
    /*jb2Symbol *syms, *cursym;  */
    /*Dgp_Int32 i;  */
    DgpAdt_ListNode ln, retnode; 
 	
    /* A DgpAdt_List to put the segments we build on */ 
    DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&refseglist, dgp_state)); 

    /* Make up a jb2File */ 
    file.headersink = file.bodysink = (DataSink)NULL; 
	file.dgp_state = dgp_state;
 
    /* and a DataSource */ 
    ds.buf = buffer; 
    ds.bufsize = buflen; 
    ds.bufpos = 0; 
    ds.more_data = (ErrCode (*)(DataSource))NULL; 
	ds.dgp_state = dgp_state;
    ds.userdata = (void *)NULL; 
 
    /* Read and decode all the segments */ 
    for (;;) { 
	curseg = (jb2Segment *)(a->malloc)(sizeof(*curseg)); 
	DGP_CHECK_ALLOC_CLEANUP(curseg); 
 
	DGP_ERR_PASS_CLEANUP(jb2DecodeSegmentHeader(&file, &ds, curseg)); 
 
	/* Collect all the segments this one references */ 
	DGP_ERR_PASS_CLEANUP(jb2MarshalSegments(&file, curseg, refseglist, 
						&refsegs)); 
 
	DGP_ERR_PASS_CLEANUP(jb2DecodeSegment(&file, &ds, curseg, refsegs)); 
 
	/* It's decoded - don't need refsegs */ 
	if (refsegs != (jb2Segment **)NULL) { 
	    (a->free)((void *)refsegs); 
	    refsegs = (jb2Segment **)NULL; 
	    } 
 
	/* Was it the one we wanted? */ 
	if (curseg->segtype == JB2SEGTYPE_SYMDICT) { 
	    /* Done */ 
	    break; 
	    } 
 
	/* Add it to the list of referenced segments */ 
	DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(refseglist, (void *)curseg, &retnode)); 
	/* Owned by the list now */ 
	curseg = (jb2Segment *)NULL; 
	} 

    *segment = curseg;
	 
    DGPADT_FOREACH (ln, refseglist) { 
	curseg = (jb2Segment *)DGPADT_LIGET(ln); 
	jb2FreeSegmentInfo((void *)curseg, dgp_state); 
	(a->free)((void *)curseg); 
	} 
    DgpAdt_liFree(refseglist); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    if (curseg != (jb2Segment *)NULL) { 
	    /* Not safe to jb2FreeSegmentInfo on curseg, unfortunately */ 
	    if (curseg->refsegs != (Dgp_Int32 *)NULL) { 
	        (a->free)((void *)curseg->refsegs); 
	    } 
	    if (curseg->refretain != (Dgp_Bool *)NULL) { 
	        (a->free)((void *)curseg->refretain); 
	    } 
	    (a->free)((void *)curseg); 
	} 
 
    if (refseglist != DGP_NULLLIST) { 
	DGPADT_FOREACH (ln, refseglist) { 
	    curseg = (jb2Segment *)DGPADT_LIGET(ln); 
	    jb2FreeSegmentInfo((void *)curseg, dgp_state); 
	    (a->free)((void *)curseg); 
	    } 
	DgpAdt_liFree(refseglist); 
	} 
 
    if (refsegs != (jb2Segment **)NULL) { 
	(a->free)((void *)refsegs); 
	} 
 
    ERR_RETURN; 
    } 

 
/* 
 * This routine takes an encoded position block and returns the number of 
 * symbol instances encoded in it. 
 * 
 * It has to read through the (variable-length) segment header, then 
 * parse the (variable-length) segment data header to find the value it needs. 
 * 
 * As above, we skip segments until we hit a symbol region segment. 
 */ 
ErrCode 
jb2_getTokPosCount(void *buffer, size_t buflen, Dgp_State *dgp_state, 
		    size_t *count) 
{ 
  DGP_PROLOGUE("tcmp_getTokPosCount()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    jb2Segment segment; 
    struct _jb2File file; 
    struct _DataSource ds; 
    Dgp_UInt8 buf[17]; 
    Dgp_UInt16 flags; 
    size_t bytes; 
 
    /* Make up a jb2File */ 
    file.lastsegnum = 0; 
    file.headersink = file.bodysink = (DataSink)NULL; 
    file.dgp_state = dgp_state; 
 
    ds.buf = buffer; 
    ds.bufsize = buflen; 
    ds.bufpos = 0; 
    ds.more_data = (ErrCode (*)(DataSource))NULL; 
    ds.userdata = (void *)NULL; 
	ds.dgp_state = dgp_state;
 
    for (;;) { 
		DGP_ERR_PASS_CLEANUP(jb2DecodeSegmentHeader(&file, &ds, &segment)); 
		/* Free the unneeded stuff in segment */ 
		(a->free)((void *)segment.refsegs); 
		(a->free)((void *)segment.refretain); 
		
		if ((segment.segtype != JB2SEGTYPE_SYMREG_INT) && 
			(segment.segtype != JB2SEGTYPE_SYMREG_IMM) && 
			(segment.segtype != JB2SEGTYPE_SYMREG_IMMLL)) { 
			/* Skip over the body */ 
			ds.bufpos += segment.datalen; 
			
			/* Ran out of segments without hitting a dictionary! */
			DGP_ERR_GEN_CLEANUP((ds.bufpos >= ds.bufsize), cErrBufferExceeded);
		} 
		else {
			/* Found it */
			break;
		} 
		
	}

	/* 
     * There's always 17 bytes of region segment data header.  Then 
     * 2 bytes of flags; if SBHUFF is 1, there's two more bytes of flags. 
     * If SBREFINE is 1 and SBRTEMPLATE is 0 there's four more bytes of flags. 
     */ 
    DSREAD(&ds, 17, buf, &bytes); 
    DSREAD(&ds, 2, buf, &bytes); 
    flags = (buf[0] << 8) | buf[1]; 
    if (flags & 1) { 
	/* SBHUFF is 1 */ 
	DSREAD(&ds, 2, buf, &bytes); 
	} 
    if (flags & 2) { 
	/* SBREFINE is 1 */ 
	if (flags & (1 << 15)) { 
	    /* SBRTEMPLATE is not 0 */ 
	    } 
	else { 
	    DSREAD(&ds, 4, buf, &bytes); 
	    } 
	} 
 
    /* Next 4 bytes are what we want */ 
    DSREAD(&ds, 4, buf, &bytes); 
    *count = (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3]; 
 
    return(cSuccessful); 
 
  ERR_CLEANUP: 
    /* No cleanup, at least... */ 
    *count = -1; 
    return(cErrBufferExceeded); 
    } 
 
/* 
 * Very similar to decodeTokDict: parse all the segments we're given. 
 * 
 * We have to first fake up a dictionary segment that includes all the 
 * symbols in dstate.  Fortunately, that's not hard. 
 */ 
ErrCode 
jb2_decodeTokPositions(jb2SymbolInstance **positions, 
			Dgp_Int32 lowx, Dgp_Int32 lowy, Dgp_Int32 highx, Dgp_Int32 highy, 
			void *buffer, size_t buflen, DgpAdt_List refseglist,Dgp_Int32 *nboxes, 
			Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jb2_decodeTokPositions()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    struct _jb2File file; 
    struct _DataSource ds; 
    jb2Segment **refsegs = (jb2Segment **)NULL; 
    jb2Segment *curseg = (jb2Segment *)NULL; 
    jb2Segment *posseg = (jb2Segment *)NULL; 
    Dgp_Int32 nhtseg=0, offset = 0, i;
    jb2Symbol *syms;
    jb2SymbolInstance *pos;
    /*jb2SymbolInstance *insts, *curinst;  */
    /*Dgp_Int32 i;  */
    DgpAdt_ListNode ln, retnode; 
 
    /* stop compiler whining */ 
    (void)lowx; 
    (void)lowy; 
    (void)highx; 
    (void)highy; 

    /* Make up a jb2File */ 
    file.lastsegnum = 0; 
    file.headersink = file.bodysink = (DataSink)NULL; 
    file.dgp_state = dgp_state; 
 
    /* and a DataSource */ 
    ds.buf = buffer; 
    ds.bufsize = buflen; 
    ds.bufpos = 0; 
    ds.more_data = (ErrCode (*)(DataSource))NULL; 
    ds.userdata = (void *)NULL; 
	ds.dgp_state = dgp_state;


 
    /* Read and decode all the segments */ 
    while(ds.bufpos < ds.bufsize) { 
	    curseg = (jb2Segment *)(a->malloc)(sizeof(*curseg)); 
	    DGP_CHECK_ALLOC_CLEANUP(curseg); 
 
	    DGP_ERR_PASS_CLEANUP(jb2DecodeSegmentHeader(&file, &ds, curseg)); 

        /* Collect all the segments this one references */ 
	    DGP_ERR_PASS_CLEANUP(jb2MarshalSegments(&file, curseg, refseglist, 
						    &refsegs)); 
 
	    DGP_ERR_PASS_CLEANUP(jb2DecodeSegment(&file, &ds, curseg, refsegs)); 
 
	    /* It's decoded - don't need refsegs */
	    if (refsegs != (jb2Segment **)NULL) { 
	        (a->free)((void *)refsegs); 
	        refsegs = (jb2Segment **)NULL; 
	        }
 
	    /* is it a symb dict ?
	       if so, add tokens 
	    if (curseg->segtype == JB2SEGTYPE_SYMDICT)
		    jb2_addTokens(curseg, dgp_state);	*/

	    /* Was it the one we wanted? */ 
	    if ((curseg->segtype == JB2SEGTYPE_SYMREG_INT) || 
	        (curseg->segtype == JB2SEGTYPE_SYMREG_IMM) || 
	        (curseg->segtype == JB2SEGTYPE_SYMREG_IMMLL)) { 
            /* curseg is now the decoded symbol region segment - copy inst info */ 

			/* second or higher stripe */
			if (offset > 0)
			{
				/* adjust positions if we are beyond the first stripe */
				for (i = 0; i < curseg->data.symreg.ninstances; i++)
				{
					curseg->data.symreg.instances[i].ypos += offset;
				}

        /* add number of instances to the positions */
        DGP_CHECK_ALLOC_CLEANUP( 
            pos =(jb2SymbolInstance *)(a->realloc)(*positions, 
                    (*nboxes + curseg->data.symreg.ninstances) * 
					sizeof(jb2SymbolInstance)));

			memcpy ((void*) (pos + *nboxes), (const void*) curseg->data.symreg.instances,
				curseg->data.symreg.ninstances * sizeof(jb2SymbolInstance));
			*positions = pos;
	        *nboxes += curseg->data.symreg.ninstances;

			/* need to free the instances */
	        posseg->data.symreg.instances = pos;
	        posseg->data.symreg.ninstances = *nboxes;

			}
			else
			{
            *positions = curseg->data.symreg.instances; 
            *nboxes = curseg->data.symreg.ninstances;
            posseg = curseg;
			}

            /* Free what we have to */ 
            /*(a->free)((void *)curseg);  */
			/* 30.4.2001 / dhtesq. */
			/* Leaking intstances associated with this segment. */
	        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(refseglist, (void *)curseg, &retnode)); 
        }else if(   (curseg->segtype == JB2SEGTYPE_HTREG_INT) ||
                    (curseg->segtype == JB2SEGTYPE_HTREG_IMM) ||
                    (curseg->segtype == JB2SEGTYPE_HTREG_IMMLL)){
            DGP_ERR_PASS_CLEANUP(jb2RenderRegion(&file, curseg));
            /*free the cells after decoded */
            if (curseg->data.htreg.cells != (DgpImage_Short)NULL) { 
	            DgpImage_free(curseg->data.htreg.cells); 
                curseg->data.htreg.cells = NULL;
	        } 
            nhtseg++;
	        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(refseglist, (void *)curseg, &retnode)); 
        }else if(curseg->segtype == JB2SEGTYPE_EOSTRIPE ){
			offset = curseg->data.eostripe.last_row;
			DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(refseglist, (void *)curseg, &retnode)); 
        }else if((curseg->segtype == JB2SEGTYPE_GENREG_INT) ||
                 (curseg->segtype == JB2SEGTYPE_GENREG_IMM) ||
                 (curseg->segtype == JB2SEGTYPE_GENREG_IMMLL)){
            DGP_ERR_PASS_CLEANUP(jb2RenderRegion(&file, curseg));
	        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(refseglist, (void *)curseg, &retnode)); 
            nhtseg++;
        }else if((curseg->segtype == JB2SEGTYPE_EOPAGE) ||
                 (curseg->segtype == JB2SEGTYPE_EOFILE)) {
            (a->free)((void *)curseg); 
            break;
        } else {
	        /* Add it to the list of referenced segments */ 
	        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(refseglist, (void *)curseg, &retnode)); 
        }
	    /* Owned by the list now */ 
	    curseg = (jb2Segment *)NULL; 
	} 
    if(nhtseg >0){
       
        /* add nhtseg number of instances to the positions */
        DGP_CHECK_ALLOC_CLEANUP( 
            pos =(jb2SymbolInstance *)(a->realloc)(*positions, 
                    (*nboxes + nhtseg) * sizeof(jb2SymbolInstance))
            );
        *positions = pos;
        posseg->data.symreg.instances = pos;
        pos += *nboxes;
        *nboxes += nhtseg;
        posseg->data.symreg.ninstances = *nboxes;

        /* filling the data for new symbols and new symbol instances */
        DGPADT_FOREACH (ln, refseglist) { 
            curseg = (jb2Segment *)DGPADT_LIGET(ln); 
            if(     (curseg->segtype == JB2SEGTYPE_HTREG_INT) ||
                    (curseg->segtype == JB2SEGTYPE_HTREG_IMM) ||
                    (curseg->segtype == JB2SEGTYPE_HTREG_IMMLL) ||        
                    (curseg->segtype == JB2SEGTYPE_GENREG_INT) ||
                    (curseg->segtype == JB2SEGTYPE_GENREG_IMM) ||
                    (curseg->segtype == JB2SEGTYPE_GENREG_IMMLL)){

		/*If there's the encoded version still around, free it */
		if ((curseg->segtype == JB2SEGTYPE_GENREG_INT) ||
                    (curseg->segtype == JB2SEGTYPE_GENREG_IMM) ||
                    (curseg->segtype == JB2SEGTYPE_GENREG_IMMLL)){
		    if (curseg->data.genreg.data) {
			(a->free)((void *)curseg->data.genreg.data); 
		    }
		}
                
                /*Make the ht region segment a symbol dictionary segment */
                curseg->segtype = 0;

                /*create a symbol and put it in the segment */
                DGP_CHECK_ALLOC_CLEANUP( 
                    syms =(jb2Symbol *)(a->calloc)(1,sizeof(jb2Symbol)));
                curseg->data.symdict.symbols = syms;
                curseg->data.symdict.nsymbols = 1;
                curseg->data.symdict.direct_contexts = NULL;
                curseg->data.symdict.ref_contexts = NULL;
                syms->width = curseg->reginfo.regwidth;
                syms->height = curseg->reginfo.regheight;
                syms->bits = DGPIMG_GETROW(curseg->reginfo.im,0); 
                syms->longsperline =  DGPIMG_GETROWWIDTH(curseg->reginfo.im)/4;
                syms->refsym = NULL;
                syms->userdata = NULL;

                /*free the unnecessary memory */
                (a->free)((void *)curseg->reginfo.im->data); 
                (a->free)((void *)curseg->reginfo.im); 
                
                /*point the new position to the symbol */
                pos->xpos = curseg->reginfo.regxpos;
                pos->ypos = curseg->reginfo.regypos;
                pos->refdata = NULL;
                pos->userdata = NULL;
                pos->sym = syms ;
                pos++;
            } 
        }
    }

 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    if (curseg != (jb2Segment *)NULL) { 
	/* Not safe to jb2FreeSegmentInfo on curseg, unfortunately */ 
	if (curseg->refsegs != (Dgp_Int32 *)NULL) { 
	    (a->free)((void *)curseg->refsegs); 
	    } 
	if (curseg->refretain != (Dgp_Bool *)NULL) { 
	    (a->free)((void *)curseg->refretain); 
	    } 
	(a->free)((void *)curseg); 
	} 
 
    if (refseglist != DGP_NULLLIST) { 
	DGPADT_FOREACH (ln, refseglist) { 
	    curseg = (jb2Segment *)DGPADT_LIGET(ln); 
		jb2FreeSegmentInfo((void *)curseg, dgp_state); 
		(a->free)((void *)curseg); 
	    } 
	DgpAdt_liFree(refseglist); 
	} 
    if (refsegs != (jb2Segment **)NULL) { 
	(a->free)((void *)refsegs); 
	} 
 
    ERR_RETURN; 
    } 
 

/* 
 * Add tokens to jb2 symbols
 * 
 * 
 * 
 */ 
#if 0
static ErrCode 
jb2_addTokens(jb2Segment *curseg, Dgp_State *dgp_state) 
{ 

    jb2Symbol *syms, *cursym; 
	Dgp_UInt32 ntoks, i;
	token *tokens;

	DGP_PROLOGUE("jb2_addTokens()", DGP_PERRSTATE); 

	/* Was it the one we wanted? */ 
	if (curseg->segtype != JB2SEGTYPE_SYMDICT) { 
		return cErrInvalidParm;   
	} 

	ntoks = curseg->data.symdict.nsymbols;
DGP_CHECK_ALLOC (
    tokens = (token *)DGP_MALLOC( ntoks * sizeof( token)));

    /* curseg is now the decoded symbol dictionary segment - copy sym info */ 
    syms = curseg->data.symdict.symbols; 
    for (i = 0; i < ntoks; i++) { 
	cursym = &syms[i]; 
	tokens[i].width = cursym->width; 
	tokens[i].height = cursym->height; 
	tokens[i].bits = cursym->bits; 
	tokens[i].longsperline = cursym->longsperline; 
	tokens[i].uses = 0; 
	tokens[i].committed = DGP_FALSE; 
	tokens[i].extdata = (void *)NULL;
	tokens[i].userflags = (Dgp_UInt8) 0;
	cursym->userdata = &tokens[i];
	} 

NORMAL_RETURN;

}
#endif

/* 
 * Decode a binary extension block and return the binary data 
 * 
 */ 
ErrCode 
jb2_decodeBinExtension(Dgp_UInt8** data, Dgp_UInt32* size, 
			void *buffer, size_t buflen, DgpAdt_List refseglist, 
			Dgp_State *dgp_state) 
{ 
  DGP_PROLOGUE("jb2_decodeBinExtensions()", DGP_PERRSTATE); 
 
    struct _jb2File file; 
    struct _DataSource ds; 
    jb2Segment curseg;
 
	/* default returns */
	*data = NULL;
	*size = 0;
  
    (void)refseglist;
    /* Make up a jb2File */ 
    file.lastsegnum = 0; 
    file.headersink = file.bodysink = (DataSink)NULL; 
    file.dgp_state = dgp_state; 
 
    /* and a DataSource */ 
    ds.buf = buffer; 
    ds.bufsize = buflen; 
    ds.bufpos = 0; 
    ds.more_data = (ErrCode (*)(DataSource))NULL; 
    ds.userdata = (void *)NULL; 
	ds.dgp_state = dgp_state;
	
    /* Read all headers and decode the annotation segment */ 
    while (ds.bufpos < ds.bufsize) { 
 
	DGP_ERR_PASS_CLEANUP(jb2DecodeSegmentHeader(&file, &ds, &curseg)); 
  
	/* Was it the one we wanted? */ 
	if (curseg.segtype == JB2SEGTYPE_EXTENSION) 
  { 
	    /* Found one */ 
	DGP_ERR_PASS_CLEANUP(jb2DecodeSegment(&file, &ds, &curseg, NULL)); 
    *data = (Dgp_UInt8*) curseg.data.extension.data;
    *size = curseg.data.extension.type;
	    break; 
	    }  
	else if (curseg.segtype == JB2SEGTYPE_EOPAGE)
  {
    /* check for end of page segment */
    break;
  }
  else
  {
		/* skip data section */
		ds.bufpos += curseg.datalen;
	}
	} 
	 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    ERR_RETURN; 
    } 
 

