/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/jb2mmuncompress.h,v 1.6 2002/05/30 02:42:46 jlin Exp $ 
 * 
 */ 
 
#include "list.h"
#include "mmcompress.h" 
#include "symbol.h"
#include "segment.h"
  
#ifdef __cplusplus
extern "C"
{
#endif

/* 
 * This file contains shim code to interface the decompression side of 
 * DigiPaper to the JBIG2 encoder.  It provides 
 * 
 * Dgp_Int32 jb2_getTokenCount(void *buffer, size_t buflen) 
 * 
 * Dgp_Bool jb2_decodeTokDict(token *tokens, void *buffer, size_t buflen, 
 *		int firsttoknum, void *(*allocator)(size_t allocsize)) 
 * 
 * Dgp_Int32 jb2_getTokPosCount(void *buffer, size_t buflen) 
 * 
 * Dgp_Bool jb2_decodeTokPositions(jb2SymbolInstance *positions, 
 *			int lowx, Dgp_Int32 lowy, Dgp_Int32 highx, Dgp_Int32 highy, 
 *			void *buffer, size_t buflen, void *dstate) 
 * 
 * which behave like the versions in libtokcomp. 
 */ 
 
 
/* 
 * This routine takes an encoded dictionary block and returns the number of 
 * tokens encoded in it. 
 * 
 * It has to read through the (variable-length) segment header, then 
 * parse the (variable-length) segment data header to find the value it needs. 
 * 
 * Actually, it needs to skip over segments until it reaches a symdict 
 * segment; it may be preceded by Huffman table segments, and so on. 
 */ 
ErrCode 
jb2_getTokenCount(void *buffer, size_t buflen, Dgp_State *dgp_state, 
		   size_t *count); 
		   
		   
/* 
 * Decode a token dictionary.  This involves: 
 * - Reading any segments that precede the symbol dictionary segment 
 * - Assembling those segments 
 * - Calling the symbol dictionary decoding procedure 
 * - Reformatting all the jb2Symbols into tokens. 
 */ 
ErrCode 
jb2_decodeTokDict(token *tokens, void *buffer, size_t buflen, 
		   Dgp_State *dgp_state) ;

ErrCode 
jb2_decodeTokDictSeg(jb2Segment **segment, void *buffer, size_t buflen, 
		   Dgp_State *dgp_state) ;



/* 
 * This routine takes an encoded position block and returns the number of 
 * symbol instances encoded in it. 
 * 
 * It has to read through the (variable-length) segment header, then 
 * parse the (variable-length) segment data header to find the value it needs. 
 * 
 * As above, we skip segments until we hit a symbol region segment. 
 */ 
ErrCode 
jb2_getTokPosCount(void *buffer, size_t buflen, Dgp_State *dgp_state, 
		    size_t *count) ;
 
/* 
 * Very similar to decodeTokDict: parse all the segments we're given. 
 * 
 * We have to first fake up a dictionary segment that includes all the 
 * symbols in dstate.  Fortunately, that's not hard. 
 */ 
ErrCode 
jb2_decodeTokPositions(jb2SymbolInstance **positions, 
	    Dgp_Int32 lowx, Dgp_Int32 lowy, Dgp_Int32 highx, Dgp_Int32 highy, 
	    void *buffer, size_t buflen, DgpAdt_List refseglist, Dgp_Int32 *nboxes,
	    Dgp_State *dgp_state) ;

ErrCode 
jb2_decodeBinExtension(Dgp_UInt8** data, Dgp_UInt32* size,
			void *buffer, size_t buflen, DgpAdt_List refseglist, 
			Dgp_State *dgp_state) ;

#ifdef __cplusplus
}
#endif
