/*
*
* $Header: /project/icons/cvsroot/digipaper/src/libjbig2/jbig2apiCB.c,v 1.15 2003/02/10 22:53:27 mkrocken Exp $
*
*/

#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif

#include <stdio.h>
#include "dgp_err_control.h"
#include "dgp_utils.h"
#include "iostructures.h"
#include "tok-lossy.h"
#include "symbol.h"
#include "segment.h"
#include "encsymdict.h"
#include "encsymreg.h"
#include "encgenreg.h"
#include "encpageinfo.h"
#include "memorydatasink.h"
#include "jbig2apiCB.h"
#include "libutilw.pub"
#include "libpixrw.pub"
#include "PsgFix.h"

#include <stdlib.h>
#include <sys/types.h>

/*#ifndef _WINDOWS */
/*#include <netinet/in.h> */
/*#endif */

#include <errno.h>
#include <limits.h>
#include "blit.h"

void
freeSegDgpAdt_List(
				   DgpAdt_List segDgpAdt_List, 
				   Dgp_State *dgp_state); 

void
freeNsymDgpAdt_List(
					DgpAdt_List nsymDgpAdt_List, 
					Dgp_State *dgp_state) ;

static ErrCode
resetOutputBuffer (Jb2State *jb2State)
{
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	DGPERR_INIT(api_resetOutputBuffer);
	DataSink datasink = NULL;
	jb2File jb2f = NULL, jb2f_temp;
	/*Dgp_Allocators *a = &(dgp_state->a); */
	/*void *userData = NULL; */
	jb2FileParams *params;
	Dgp_Int32 curSegNum;	

	curSegNum = jb2FileNextSegmentNumber(jb2State->jb2f);

	DGP_ERR_PASS_CLEANUP(jb2NewMemoryDataSink(dgp_state, &datasink));
	DGP_ERR_PASS_CLEANUP(jb2CreateFile(	datasink, 
		datasink, 
		dgp_state, 
		&jb2f));
	params = jb2GetFileParams(jb2f);

	switch (jb2State->jbig2Profile)
	  {
	  case 1:
	/*
	* Here are the parameters for proposed ITU profile 1:
	*/
	params->sdarith = params->srarith = DGP_FALSE;      /* No arith coding */
	params->sdtables = params->srtables = DGP_FALSE;    /* No custom tables */
	/* Only the first choice is allowed */
	params->sdfirstonly = params->srfirstonly = DGP_TRUE;
	/* Only LL corner, non-transposed */
	params->srllonly = params->srnontransonly = DGP_TRUE;
	break;
	  case 3:
	/*
	* Here are the parameters for proposed ITU profile 3:
	*/
	params->sdarith = params->srarith = DGP_TRUE;       /* Arith coding */
	params->sdabparams.rtemplate = 2;                /* 10-pixel, 3-line */
	params->sdabparams.atx[0] = 2;                  /* Nominal AT */
	params->sdabparams.aty[0] = -1;
	params->sdabparams.tpon = DGP_FALSE;                /* No TP */
	
	/* Only LL corner, non-transposed */
	params->srllonly = params->srnontransonly = DGP_TRUE;
	break;
	
	  case 2:
	  default:
	/*
	* Here are the parameters for proposed ITU profile 2:
	*/
	params->sdarith = params->srarith = DGP_FALSE;      /* No arith coding */
	params->sdtables = params->srtables = DGP_TRUE;     /* Custom tables */
	/* More than the first choice is allowed */
	params->sdfirstonly = params->srfirstonly = DGP_FALSE;
	/* Any corner, any transposed */
	params->srllonly = params->srnontransonly = DGP_FALSE;
	break;
	  }

	jb2SetFileParams(jb2f, params);
	jb2SetFileSegnum(jb2f, curSegNum);

	/* put new jb2File into jb2State */
	jb2f_temp = jb2State->jb2f;
	jb2State->jb2f = jb2f;
	
	/* and free old jb2File */
	DGP_FREE (jb2f_temp);

/*  datasink = jb2GetFileHeaderDataSink((jb2File)jb2State->jb2f);	
  DGP_ERR_PASS_CLEANUP( (jb2CloseMemoryDataSink(datasink, 
			&datalist)));
  if (datalist->first->userdata != NULL) 
		DGP_FREE (datalist->first->userdata);
  DgpAdt_liFree (datalist); */


	NORMAL_RETURN;

ERR_CLEANUP:

	ERR_RETURN;
}


/****************************************************************
* Create a Jb2 state cookie.  initialize all the state variables.
* The given page and dictionary counts are estimates only.
* use MemoryDataSink stuff to hold the segments and segment headers
* @param stream User defined stream
* @param dgp_state     Dgp error handling cookie
* @param numPgs        Estimated number of pages
* @param numDicts      Estimated number of dictionaries (not used)
* @return _jb2State    Jbig2 state cookie contains infor needed to
*                      write out Jbig2 by using the Callback interface
* @return userData     User defined data (not used)
********************************************************************/
ErrCode 
createJb2StateAPICB(	uds *stream, 
			Dgp_State *dgp_state, 
			Dgp_Int32 jbig2Profile,
			Dgp_Int32 numPgs,
			Dgp_Bool pdfOut,
			void **_jb2State,
			void **_userData)
{
	DGPERR_INIT(createJb2State);
	DataSink datasink = NULL;
	jb2File jb2f = NULL;
	Jb2State *jb2State=NULL;
	Dgp_Allocators *a = &(dgp_state->a);
	/*void *userData = NULL; */
	jb2FileParams *params;
	
	/* These parameters are not used */
	numPgs = numPgs;
	DGP_CHECK_NULL(_jb2State);
	
	/* initialize jb2State	*/
	DGP_CHECK_ALLOC_CLEANUP(jb2State = 
		(Jb2State *) a->malloc(sizeof(Jb2State)));
	jb2State->stream = stream;
	jb2State->jbig2Profile = jbig2Profile;
	jb2State->pgnum = 0;
	jb2State->nsyms =0;
	jb2State->pgnsyms =0;
    jb2State->pdfOut = pdfOut;
	jb2State->nsymDgpAdt_List = DGP_NULLLIST;
	jb2State->segDgpAdt_List = DGP_NULLLIST;
	jb2State->pgSegDgpAdt_List = DGP_NULLLIST;
	jb2State->pgNsymDgpAdt_List = DGP_NULLLIST;
	DGP_ERR_PASS_CLEANUP(jb2NewMemoryDataSink(dgp_state, &datasink));
	DGP_ERR_PASS_CLEANUP(jb2CreateFile(	datasink, 
		datasink, 
		dgp_state, 
		&jb2f));
	params = jb2GetFileParams(jb2f);

	switch (jb2State->jbig2Profile)
	  {
	  case 1:
	/*
	* Here are the parameters for proposed ITU profile 1:
	*/
	params->sdarith = params->srarith = DGP_FALSE;      /* No arith coding */
	params->sdtables = params->srtables = DGP_FALSE;    /* No custom tables */
	/* Only the first choice is allowed */
	params->sdfirstonly = params->srfirstonly = DGP_TRUE;
	/* Only LL corner, non-transposed */
	params->srllonly = params->srnontransonly = DGP_TRUE;
	break;
	  case 3:
	/*
	* Here are the parameters for proposed ITU profile 3:
	*/
	params->sdarith = params->srarith = DGP_TRUE;       /* Arith coding */
	params->sdabparams.rtemplate = 2;                /* 10-pixel, 3-line */
	params->sdabparams.atx[0] = 2;                  /* Nominal AT */
	params->sdabparams.aty[0] = -1;
	params->sdabparams.tpon = DGP_FALSE;                /* No TP */
	
	/* Only LL corner, non-transposed */
	params->srllonly = params->srnontransonly = DGP_TRUE;
	break;

	  case 2:
	  default:
/*
	* Here are the parameters for proposed ITU profile 2:
	*/
	params->sdarith = params->srarith = DGP_FALSE;      /* No arith coding */
	params->sdtables = params->srtables = DGP_TRUE;     /* Custom tables */
	/* More than the first choice is allowed */
	params->sdfirstonly = params->srfirstonly = DGP_FALSE;
	/* Any corner, any transposed */
	params->srllonly = params->srnontransonly = DGP_FALSE;
	break;
	  }

	jb2SetFileParams(jb2f, params);
	
	jb2State->jb2f = jb2f;
	*_jb2State = jb2State;
	jb2f = NULL;
	jb2State = NULL;
	if (_userData != NULL)
		*_userData = (void*) jb2f;
	
	NORMAL_RETURN;
	
ERR_CLEANUP:
	if(jb2State != NULL) a->free(jb2State);
	ERR_RETURN;
}
				 

/*****************************************************************
* Close the Jb2State and free up memory.
*
* It creates and writes out and EOF segment.
* It closes the two DataSinks, takes the buflists it gets back, 
* writes them out (headers then bodies), then frees them.
* It also free the current list of segments.
* @param _jb2State Jbig2 State cookie
* @param _userData Not used
*****************************************************************/
ErrCode 
closeJb2StateAPICB( 
				void *_jb2State,
				void *_userData)
{
	Jb2State *jb2State = (Jb2State *)_jb2State;
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	DGPERR_INIT(closeJb2State);
	DataSink datasink;
	DgpAdt_List datalist;
	jb2Segment eofseg;
	Dgp_Allocators *a = &(dgp_state->a);
	void *userData;
	mds_membuf *membuf;
	DgpAdt_ListNode ln;
	
	userData = _userData;
	eofseg.pagenum = 0;
	eofseg.segtype = JB2SEGTYPE_EOFILE;
	DGP_ERR_PASS_CLEANUP( (jb2EncodeEOFile( jb2State->jb2f, 
		&eofseg))); 
	freeSegDgpAdt_List(jb2State->segDgpAdt_List,dgp_state);
	freeNsymDgpAdt_List(jb2State->nsymDgpAdt_List,dgp_state);
	freeSegDgpAdt_List(jb2State->pgSegDgpAdt_List,dgp_state);
	freeNsymDgpAdt_List(jb2State->pgNsymDgpAdt_List,dgp_state);
	jb2State->segDgpAdt_List=DGP_NULLLIST;
	jb2State->nsymDgpAdt_List=DGP_NULLLIST;
	jb2State->pgSegDgpAdt_List=DGP_NULLLIST;
	jb2State->pgNsymDgpAdt_List=DGP_NULLLIST;
	
	datasink = jb2GetFileHeaderDataSink((jb2File)jb2State->jb2f);
	
	DGP_ERR_PASS_CLEANUP( (jb2CloseMemoryDataSink(datasink, &datalist)));

	/* 20.4.2001 / dhtesq. */
	/* Leak. */
/*	if (datalist->first->userdata != NULL) 
		DGP_FREE (datalist->first->userdata);	*/
	ln = DGPADT_LIFIRST(datalist);
	while(ln != DGP_NULLLISTNODE){
		membuf = (mds_membuf *)DGPADT_LIGET(ln);
		DGP_FREE(membuf->buf);
		DGP_FREE(membuf);
		DgpAdt_liRemAndNext(ln, &ln);
	}
	DgpAdt_liFree (datalist);

/*	DGP_ASSERT (	datasink = NULL ); */
	a->free(jb2State->jb2f);

	a->free(jb2State); 
	_jb2State=NULL;
	userData = NULL;
	NORMAL_RETURN;
	
ERR_CLEANUP:
	if(jb2State !=NULL) a->free(jb2State);
	ERR_RETURN;
}


ErrCode
getJb2CompressedData (Jb2State *jb2State, Dgp_UInt8** buf, 
		      Dgp_UInt32* buf_size)
{
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	DGPERR_INIT(api_writeBuffer);
	DgpAdt_ListNode ln;
	mds_membuf *curbuf;
	DataSink datasink;
	DgpAdt_List datalist;
	Dgp_Allocators *a = &(dgp_state->a);
	size_t sizeData;
	Dgp_Int8* pdata;
		

	/* consume the data in the memory data sink */
	datasink = jb2GetFileHeaderDataSink((jb2File)jb2State->jb2f);
	
	DGP_ERR_PASS_CLEANUP( (jb2CloseMemoryDataSink(datasink, 
		&datalist)));

	/* get the size of the data in bytes */
	sizeData = 0;
	DGPADT_FOREACH (ln, datalist) {
		curbuf = (mds_membuf *)DGPADT_LIGET(ln);
		sizeData += curbuf->bufsize;
	}

	/* allocate output data buffers */
	*buf_size = sizeData;
	if (sizeData) {		/* It's invalid to request 0 bytes */
	    DGP_CHECK_ALLOC (pdata = (Dgp_Int8*) a->malloc (sizeData));
	} else {
	    pdata = NULL;
	}
	*buf = pdata;

	DGPADT_FOREACH (ln, datalist) {
		curbuf = (mds_membuf *)DGPADT_LIGET(ln);
		memcpy (pdata, curbuf->buf, curbuf->bufsize);
		pdata += curbuf->bufsize;
		a->free(curbuf->buf);
		a->free((void *)curbuf);
	}
	DgpAdt_liFree(datalist);

	DGP_ERR_PASS (resetOutputBuffer(jb2State));

	NORMAL_RETURN;
	
ERR_CLEANUP:

	ERR_RETURN;
}






