/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/memorydatasink.c,v 1.9 2003/01/27 21:48:37 awilliam Exp $ 
 * 
 */ 
 
#include "datasink.h" 
#include "memorydatasink.h" 
#include "util.h" 
#include "list.h" 
#include "PsgFix.h"
#include <stdlib.h> 
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/memorydatasink.c,v 1.9 2003/01/27 21:48:37 awilliam Exp $"); 
 
typedef struct { 
    DgpAdt_List bufs; 
    size_t markpos;		/* where in the buffer the mark was */ 
    DgpAdt_ListNode markln;		/* the current list node at the time the mark is set */ 
    } mdsink_data; 
 
/* How large are the individual buffers going to be? */ 
#define	MDS_BUFSIZE	16384 
 
/* The callbacks we'll use */ 
static ErrCode mdsink_moredata(DataSink f); 
static ErrCode mdsink_mark(DataSink f); 
static ErrCode mdsink_fillmark(DataSink f, size_t nbytes, Dgp_UInt8 *buf); 
 
/* 
 * Create a DataSink that will generate a DgpAdt_List of memory buffers (mds_membuf), 
 * each containing the specified number of bytes. 
 * 
 * The DgpAdt_List is initially empty; each buffer is added to it as it is 
 * filled.  The final buffer is added by jb2CloseMemoryDataSink. 
 */ 
ErrCode 
jb2NewMemoryDataSink(Dgp_State *dgp_state, DataSink *_newds) 
{ 
  DGP_PROLOGUE("jb2NewMemoryDataSink()", DGP_PERRSTATE); 
  Dgp_Allocators *a = &(dgp_state->a); 
 
    DataSink newds = (DataSink)NULL; 
    Dgp_UInt8 *buf = (Dgp_UInt8 *)NULL; 
    DgpAdt_List buflist = DGP_NULLLIST; 
    mdsink_data *newdata = (mdsink_data *)NULL; 
 
    *_newds = (DataSink)NULL; 
 
    newds = (DataSink)(a->malloc)(sizeof(*newds)); 
    DGP_CHECK_ALLOC_CLEANUP(newds); 
    newds->dgp_state = dgp_state; 
 
    buf = (Dgp_UInt8 *)(a->malloc)(MDS_BUFSIZE); 
    DGP_CHECK_ALLOC_CLEANUP(buf); 
 
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&buflist, dgp_state)); 
 
    newdata = (mdsink_data *)(a->malloc)(sizeof(*newdata)); 
    DGP_CHECK_ALLOC_CLEANUP(newdata); 
 
    newdata->bufs = buflist; 
    newdata->markpos = 0; 
 
    newds->buf = buf; 
    newds->bufsize = MDS_BUFSIZE; 
    newds->bufpos = 0; 
    newds->count = 0; 
    newds->more_data = mdsink_moredata; 
    newds->mark = mdsink_mark; 
    newds->fillmark = mdsink_fillmark; 
    newds->userdata = (void *)newdata; 
 
    *_newds = newds; 
    newds = (DataSink)NULL; 
 
    NORMAL_RETURN; 
 
 ERR_CLEANUP: 
    if (newds != (DataSink)NULL) { 
	(a->free)((void *)newds); 
	} 
    if (buf != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)buf); 
	} 
    if (buflist != DGP_NULLLIST) { 
	DgpAdt_liFree(buflist); 
	} 
    if (newdata != (mdsink_data *)NULL) { 
	(a->free)((void *)newdata); 
	} 
    NORMAL_RETURN; 
    } 
 
 
/* 
 * Link the current buffer onto the list it's building; free everything; 
 * return the DgpAdt_List.  On failure, what data could be salvaged is stored 
 * in *buflist, but data may have been lost.  The DataSink is destroyed 
 * in any case. 
 */ 
ErrCode 
jb2CloseMemoryDataSink(DataSink ds, DgpAdt_List *buflist) 
{ 
  DGP_PROLOGUE("jb2CloseMemoryDataSink()", ds->DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(ds->DGP_ALLOCATORS); 
    DgpAdt_List bufs; 
    DgpAdt_ListNode retnode; 
    mdsink_data *data; 
    mds_membuf *newmbuf = (mds_membuf *)NULL; 
 
    DGP_ASSERT(ds->more_data == mdsink_moredata); /* Make sure it's ours */ 
 
    data = (mdsink_data *)ds->userdata; 
    bufs = data->bufs; 
 
    /* Create a new membuf */ 
    newmbuf = (mds_membuf *)(a->malloc)(sizeof(*newmbuf)); 
    DGP_CHECK_ALLOC_CLEANUP(newmbuf); 
 
    /* Only count the actual bytes written */ 
    newmbuf->bufsize = ds->bufpos; 
    newmbuf->buf = ds->buf; 
 
    /* Add it to the list */ 
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(bufs, (void *)newmbuf, &retnode)); 
 
    (a->free)((void *)data); 
	data = NULL;
    (a->free)((void *)ds); 
	ds = NULL;
    *buflist = bufs; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    *buflist = bufs; 
 
    if (newmbuf != (mds_membuf *)NULL) { 
	(a->free)((void *)newmbuf); 
	} 
    (a->free)((void *)ds->buf); 
    (a->free)((void *)data); 
    (a->free)((void *)ds); 
 
    ERR_RETURN; 
    } 
 
/* 
 * Hook the current buffer onto the list being grown, and return a 
 * new buffer.  On failure, nothing has changed. 
 */ 
static ErrCode 
mdsink_moredata(DataSink f) 
{ 
  DGP_PROLOGUE("mdsink_moredata()", f->DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(f->DGP_ALLOCATORS); 
    Dgp_UInt8 *newbuf = (Dgp_UInt8 *)NULL; 
    mds_membuf *newmbuf = (mds_membuf *)NULL; 
	DgpAdt_ListNode retnode; 
 
    newbuf = (Dgp_UInt8 *)(a->malloc)(MDS_BUFSIZE); 
    DGP_CHECK_ALLOC_CLEANUP(newbuf); 
 
    newmbuf = (mds_membuf *)(a->malloc)(sizeof(*newmbuf)); 
    DGP_CHECK_ALLOC_CLEANUP(newmbuf); 
    newmbuf->buf = f->buf; 
    newmbuf->bufsize = f->bufpos; /* They might have called us without completely filling it... */ 
 
    /* Try hooking it onto the list first */ 
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(((mdsink_data *)f->userdata)->bufs, 
				  (void *)newmbuf, &retnode)); 
 
    /* That worked - install the new buffer */ 
    f->buf = newbuf; 
    f->bufsize = MDS_BUFSIZE; 
    f->bufpos = 0; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (newbuf != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)newbuf); 
	} 
    if (newmbuf != (mds_membuf *)NULL) { 
	(a->free)((void *)newmbuf); 
	} 
    ERR_RETURN; 
    } 
 
/* 
 * Set a mark in the DataSink, to be filled in later by mdsink_fillmark 
 */ 
static ErrCode 
mdsink_mark(DataSink f) 
{ 
  DGP_PROLOGUE("mdsink_mark()", f->DGP_PERRSTATE); 
 
    mdsink_data *mds = ((mdsink_data *)f->userdata); 
 
  DGP_STOP_WHINE; 
 
    mds->markpos = f->bufpos; 
    mds->markln = DGPADT_LILAST(mds->bufs); 
    NORMAL_RETURN; 
    } 
 
/* 
 * Write nbytes to the mark position set by mdsink_mark 
 */ 
static ErrCode 
mdsink_fillmark(DataSink f, size_t nbytes, Dgp_UInt8 *buf) 
{ 
   DGP_PROLOGUE("mdsink_fillmark()", f->DGP_PERRSTATE); 
 
    mdsink_data *mds = ((mdsink_data *)f->userdata); 
    DgpAdt_ListNode ln = mds->markln; 
    size_t markpos = mds->markpos; 
    mds_membuf *curbuf; 
 
   DGP_STOP_WHINE;
    /* 
     * The DgpAdt_ListNode after markln (or the current buffer if that's 
     * the last DgpAdt_ListNode) represents the buffer that the mark was put into. 
     * 
     * Special case: if markln is DGP_NULLLISTNODE, then the thing we're looking 
     * for is the first thing on the bufs list (if any, or the current buffer). 
     */ 
    if (ln != DGP_NULLLISTNODE) { 
	/* Go one forward to the buffer the mark is in */ 
	ln = DGPADT_LINEXT(ln); 
	} 
    else { 
	/* Go the the first buffer, if any */ 
	ln = DGPADT_LIFIRST(mds->bufs); 
	} 
 
    /* ln is now the buffer that the mark is in, or DGP_NULLLISTNODE */ 
    while (nbytes > 0) { 
	if (ln == DGP_NULLLISTNODE) { 
	    /* 
	     * Write into the current buffer - there MUST be enough room 
	     * in the already-written data to hold the stuff we're trying 
	     * to put there. 
	     */ 
	    DGP_ASSERT(markpos + nbytes <= f->bufpos); 
 
	    memcpy(f->buf + markpos, buf, nbytes); 
	    nbytes = 0; 
	    } 
	else { 
	    /* Write into the buffer at ln; may not be enough room */ 
	    curbuf = (mds_membuf *)DGPADT_LIGET(ln); 
	    if (curbuf->bufsize - markpos < nbytes) { 
		/* Write what you can */ 
		memcpy(curbuf->buf + markpos, buf, curbuf->bufsize - markpos); 
		nbytes -= curbuf->bufsize - markpos; 
		buf += curbuf->bufsize - markpos; 
		markpos = 0;	/* start at the beginning in the next buffer */ 
		ln = DGPADT_LINEXT(ln); /* advance to the next buffer */ 
		} 
	    else { 
		/* Enough room to fit it */ 
		memcpy(curbuf->buf + markpos, buf, nbytes); 
		nbytes = 0; 
		} 
	    } 
	} 
 
    NORMAL_RETURN; 
    } 
 
 
 
 
 
 
 
 
 
 
 
