/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/segment.c,v 1.23 2003/02/10 22:53:27 mkrocken Exp $ 
 * 
 */ 
 
/* 
 * Routines that deal with JBIG2 segments.  These routines read 
 * segment headers, encapsulate the data obtained by decoding segments, 
 * deal with inter-segment references, and so on. 
 */ 
 
#include "util.h" 
#include "datasource.h" 
#include "segment.h" 
#include <sys/types.h> 
#include "decsymdict.h" 
#include "decsymreg.h" 
#include "dechtdict.h" 
#include "dechtreg.h" 
#include "decgenreg.h" 
#include "decgenrefreg.h" 
#include "dechufftable.h" 
#include "decpageinfo.h" 
#include "decextseg.h" 
#include <stdlib.h> 
#include "jb2file.h" 
#include "segment_util.h" 
#include "blit.h" 
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/segment.c,v 1.23 2003/02/10 22:53:27 mkrocken Exp $"); 
 
 
/* The combop to minterm translation table - here's as good a place as any */ 
const Dgp_Int32 jb2Minterms[5] = { 
    DGPBLAT_OP_OR, 
    DGPBLAT_OP_AND, 
    DGPBLAT_OP_XOR, 
    DGPBLAT_OP_XNOR, 
    DGPBLAT_OP_REPLACE 
    }; 
 
/* Dummy declarations for the moment */ 
#define	DUMMY(fname) static ErrCode fname(jb2File file, DataSource data, jb2Segment *segment, jb2Segment **refsegs) {file=file;data=data;segment=segment;refsegs=refsegs;return(DGP_FALSE);} 
DUMMY(jb2DecodeProfiles) 
 
#define DUMMYFREE(fname) static void fname(jb2Segment *segment, Dgp_State *dgp_state) {segment=segment; dgp_state=dgp_state;} 
DUMMYFREE(jb2FreeProfilesSegment) 
 
/* A table of decoding routines indexed by segment type */ 
static ErrCode (*decoders [64])(jb2File, DataSource, 
				jb2Segment*, jb2Segment**) = { 
    jb2DecodeSymDict, NULL, NULL, NULL, 
    jb2DecodeSymReg, NULL, jb2DecodeSymReg, jb2DecodeSymReg, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    jb2DecodeHTDict, NULL, NULL, NULL, 
    jb2DecodeHTReg, NULL, jb2DecodeHTReg, jb2DecodeHTReg, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    jb2DecodeGenReg, NULL, jb2DecodeGenReg, jb2DecodeGenReg, 
    jb2DecodeGenRefReg, NULL, jb2DecodeGenRefReg, jb2DecodeGenRefReg, 
    NULL, NULL, NULL, NULL, 
 
    jb2DecodePageInfo, jb2DecodeEOPage, jb2DecodeEOStripe, jb2DecodeEOFile, 
    jb2DecodeProfiles, jb2DecodeHuffTable, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, jb2DecodeExtension, NULL 
    }; 
 
/* A table of freeing routines indexed by segment type */ 
static void (*freeers [64])(jb2Segment *, Dgp_State *) = { 
    jb2FreeSymDictSegment, NULL, NULL, NULL, 
    jb2FreeSymRegSegment, NULL, jb2FreeSymRegSegment, jb2FreeSymRegSegment, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    jb2FreeHTDictSegment, NULL, NULL, NULL, 
    jb2FreeHTRegSegment, NULL, jb2FreeHTRegSegment, jb2FreeHTRegSegment, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    jb2FreeGenRegSegment, NULL, jb2FreeGenRegSegment, jb2FreeGenRegSegment, 
    jb2FreeGenRefRegSegment, NULL, jb2FreeGenRefRegSegment, jb2FreeGenRefRegSegment, 
    NULL, NULL, NULL, NULL, 
 
    jb2FreePageInfoSegment, jb2FreeEOPageSegment, jb2FreeEOStripeSegment, jb2FreeEOFileSegment, 
    jb2FreeProfilesSegment, jb2FreeHuffTableSegment, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, jb2FreeExtensionSegment, NULL 
    }; 
 
/* A table of rendering routines indexed by segment type */ 
static ErrCode (*renderers [64])(jb2File, jb2Segment *) = { 
    NULL, NULL, NULL, NULL, 
    jb2RenderSymReg, NULL, jb2RenderSymReg, jb2RenderSymReg, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    jb2RenderHTReg, NULL, jb2RenderHTReg, jb2RenderHTReg, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    jb2RenderGenReg, NULL, jb2RenderGenReg, jb2RenderGenReg, 
    jb2RenderGenRefReg, NULL, jb2RenderGenRefReg, jb2RenderGenRefReg, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL 
    }; 
 
/* A table of render-into-page routines indexed by segment type */ 
static ErrCode (*intorenderers [64])(jb2File, jb2Segment *, DgpImage_Packed) = { 
    NULL, NULL, NULL, NULL, 
    jb2RenderSymRegIntoPage, NULL, jb2RenderSymRegIntoPage, jb2RenderSymRegIntoPage, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    jb2RenderHTRegIntoPage, NULL, jb2RenderHTRegIntoPage, jb2RenderHTRegIntoPage, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    jb2RenderGenRegIntoPage, NULL, jb2RenderGenRegIntoPage, jb2RenderGenRegIntoPage, 
    jb2RenderGenRefRegIntoPage, NULL, jb2RenderGenRefRegIntoPage, jb2RenderGenRefRegIntoPage, 
    NULL, NULL, NULL, NULL, 
 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL 
    }; 
 
/* 
 * Given the information already present in *segment, which will have 
 * been filled in by jb2DecodeSegmentHeader, decode the segment's data, 
 * as much as possible, to some intermediate form.  In some cases, this 
 * may not be much at all. 
 * 
 * This routine is basically a dispatcher: look at the segment type and 
 * call the appropriate function. 
 */ 
ErrCode 
jb2DecodeSegment(jb2File file, DataSource data, 
		 jb2Segment *segment, jb2Segment **refsegs) 
{ 
  DGP_PROLOGUE("jb2DecodeSegment()", jb2_pErrState(file)); 
 
    if (decoders[segment->segtype] == 
	(ErrCode (*)(jb2File, DataSource, jb2Segment *, jb2Segment **))NULL) { 
	/* Bad segment type */ 
	FILEERR(cErrCompressionError, "bad segment type"); 
	} 
 
    DGP_ERR_PASS((*decoders[segment->segtype])(file, data, segment, refsegs)); 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* 
 * Read and parse a segment header.  Fill in *segment. 
 * Allocate refsegs and refretain as part of this. 
 * Return DGP_FALSE on failure. 
 */ 
ErrCode 
jb2DecodeSegmentHeader(jb2File file, DataSource data, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2DecodeSegmentHeader()", jb2_pErrState(file)); 
 
  Dgp_Allocators *a = jb2_pAllocators(file); 
    Dgp_UInt8 lbuf[4]; 
    Dgp_UInt8 cbuf; 
    Dgp_Bool smallpage; 
    Dgp_Int32 i, j; 
    size_t bytes; 
 
    segment->refsegs = (Dgp_Int32 *)NULL; 
    segment->refretain = (Dgp_Bool *)NULL; 
     
    /* Start reading bits of the header */ 
 
    /* Read the 4-byte segment number */ 
    DSREAD(data, 4, lbuf, &bytes); 
 
    segment->segnum = 
	(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGNUM %d\n", segment->segnum); 
#endif 
 
    /* Read the one-byte segment header flags */ 
    DSGETC(data, &cbuf); 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGFLAGS %02x\n", cbuf); 
#endif 
 
    /* and split out the bit-level fields */ 
    /* 
     * Bits 0-5: segment type 
     * Bit 6: page association field is 1 or 4 bytes Dgp_Int32 
     * Bit 7: deferred non-retain 
     */ 
    segment->segtype = cbuf & 63; 
    smallpage = (cbuf & 64) ? DGP_FALSE : DGP_TRUE; 
    segment->defer_nonretain = (cbuf & 128) ? DGP_TRUE : DGP_FALSE; 
     
    /* Read the first byte of the referenced segment count and retain flags */ 
    DSGETC(data, &cbuf); 
 
    if ((cbuf >> 5) == 7) { 
	/* It's in Dgp_Int32 form */ 
	/* 
	 * First four bytes are referenced segment count | (7 << 29) 
	 * Next byte has 8 retention flags, with the LSB being the one 
	 * for this segment, the next-to-LSB being the one for the 
	 * first referenced segment, and so on. 
	 */ 
	/* Mask the three high bits from the count field */ 
	lbuf[0] = (Dgp_UInt8)(cbuf & ~(7 << 5)); 
 
	/* Get the rest of the count */ 
	DSREAD(data, 3, lbuf + 1, &bytes); 
 
	segment->nrefsegs =  
	    (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
 
	/* Allocate the retention bits */ 
	segment->refretain = (Dgp_Bool *)(a->malloc)(segment->nrefsegs * 
					      sizeof(*segment->refretain)); 
	if (segment->refretain == (Dgp_Bool *)NULL) { 
	    NOMEM; 
	    } 
 
	/* and read them.  Reference# -1 is this segment. */ 
	for (i = -1; i < segment->nrefsegs; i += 8) { 
	    DSGETC(data, &cbuf); 
 
	    for (j = i; j < MIN(i + 8, segment->nrefsegs); j++, cbuf >>= 1) { 
		if (j == -1) { 
		    /* Self-retention bit */ 
		    segment->retain = (cbuf & 1) ? DGP_TRUE : DGP_FALSE; 
		    } 
		else { 
		    /* Retain bit for some other segment */ 
		    segment->refretain[j] = (cbuf & 1) ? DGP_TRUE : DGP_FALSE; 
		    } 
		} 
	    } 
 
	} 
    else { 
	/* It's in Dgp_Int16 form */ 
	/* 
	 * Short form is: 
	 * Bits 5-7: count of referenced segments 
	 * Bits 0-4: retention bits 
	 */ 
	/* Get the count field */ 
	segment->nrefsegs = cbuf >> 5; 
 
	/* Allocate the retention bits */ 
	/* 30.4.2001 / dhtesq. */
	/* Inserted trap for 0 nrefsegs (else we "leak" 0 bytes)... */
	if(segment->nrefsegs){
		segment->refretain = (Dgp_Bool *)(a->malloc)(segment->nrefsegs * 
			sizeof(*segment->refretain)); 
		if (segment->refretain == (Dgp_Bool *)NULL) { 
			NOMEM; 
		} 
	}
 
	/* Read the self-retention bit */ 
	segment->retain = (cbuf & 1) ? DGP_TRUE : DGP_FALSE; 
	cbuf >>= 1; 
 
	for (j = 0; j < segment->nrefsegs; j++, cbuf >>= 1) { 
	    /* Retain bit for some other segment */ 
	    segment->refretain[j] = (cbuf & 1) ? DGP_TRUE : DGP_FALSE; 
	    } 
	} 
 
    /* Now we have the segment reference count - allocate space for the refs */ 
	/* 30.4.2001 / dhtesq. */
	/* Inserted trap for 0 nrefsegs (else we "leak" 0 bytes)... */
 	if(segment->nrefsegs){
		segment->refsegs = (Dgp_Int32 *)(a->malloc)(segment->nrefsegs * 
			sizeof(*segment->refsegs)); 
		if (segment->refsegs == (Dgp_Int32 *)NULL) { 
			NOMEM; 
		} 
	}
 
    /* Read all the segment references */ 
    for (i = 0; i < segment->nrefsegs; i++) { 
	/* 
	 * Each field is 1, 2 or 4 bytes Dgp_Int32. 
	 * If the current segment's number is <= 256, they're one byte Dgp_Int32. 
	 * If the current segment's number is <= 65536, they're two bytes Dgp_Int32. 
	 * Otherwise, they're 4 bytes Dgp_Int32. 
	 */ 
	if (segment->segnum <= 256) { 
	    DSGETC(data, &cbuf); 
 
	    segment->refsegs[i] = cbuf; 
	    } 
	else if (segment->segnum <= 65536) { 
	    DSREAD(data, 2, lbuf, &bytes); 
 
	    segment->refsegs[i] = (lbuf[0] << 8) | lbuf[1]; 
	    } 
	else { 
	    DSREAD(data, 4, lbuf, &bytes); 
 
	    segment->refsegs[i] = 
		(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
	    } 
#ifdef	VALPRINT 
	fprintf(stderr, "SEGREF %d\n", segment->refsegs[i]); 
#endif 
 
	if (segment->refsegs[i] >= segment->segnum) { 
	    /* Shouldn't happen! */ 
	    FILEERR(cErrCompressionError, "bad referenced segment number"); 
	    } 
	} 
 
    /* Read the segment page association field - 1 or 4 bytes */ 
    if (smallpage) { 
	/* One byte */ 
	DSGETC(data, &cbuf); 
 
	segment->pagenum = cbuf; 
	} 
    else { 
	DSREAD(data, 4, lbuf, &bytes); 
 
	segment->pagenum = 
	    (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
	} 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGPAGE %d\n", segment->pagenum); 
#endif 
	 
    /* Read the data length - 4 bytes */ 
    DSREAD(data, 4, lbuf, &bytes); 
 
    segment->datalen = 
	(lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3]; 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGDATALEN %ld\n", segment->datalen); 
#endif 
 
    /* Fill in the "is it rendered yet?" bit */ 
    segment->reginfo.rendered = DGP_FALSE; 
    segment->reginfo.im = (DgpImage_Packed)NULL; 
 
    NORMAL_RETURN; 
		 
  ERR_CLEANUP: 
    if (segment->refsegs != (Dgp_Int32 *)NULL) { 
	(a->free)((void *)segment->refsegs); 
	segment->refsegs = (Dgp_Int32 *)NULL; 
	} 
 
    if (segment->refretain != (Dgp_Bool *)NULL) { 
	(a->free)((void *)segment->refretain); 
	segment->refretain = (Dgp_Bool *)NULL; 
	} 
 
    ERR_RETURN; 
    } 
 
/* 
 * This routine writes out the segment header of the given segment. 
 * It uses the DataSink specified by jb2GetFileHeaderDataSink(file). 
 * It doesn't write the segment's data length because that's almost 
 * always unknown at the point where this is called.  However, it 
 * does use the mark facility of DataSink to set a mark at the place 
 * where the data length should go; the mark is filled in by 
 * jb2EncodeSegmentHeaderDataLength. 
 */ 
ErrCode 
jb2EncodeSegmentHeader(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2EncodeSegmentHeader()", jb2_pErrState(file)); 
 
    Dgp_UInt8 lbuf[4]; 
    Dgp_Int32 cbuf; 
    Dgp_Int32 i, j; 
    DataSink data = jb2GetFileHeaderDataSink(file); 
    size_t bytes; 
 
    /* Write the 4-byte segment number */ 
    to4byte(lbuf, segment->segnum); 
    DSWRITE(data, 4, lbuf, &bytes); 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGNUM %d\n", segment->segnum); 
#endif 
 
    /* Write the segment header flags */ 
    /* Bits 0-5: segment type */ 
    cbuf = (segment->segtype & 63); 
    /* Bit 6: page association >= 256? */ 
    cbuf |= segment->pagenum > 255 ? 64 : 0; 
    /* Bit 7: deferred non-retain? */ 
    cbuf |= segment->defer_nonretain ? 128 : 0; 
 
    DSPUTC(data, (Dgp_UInt8)cbuf); 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGFLAGS %02x\n", cbuf); 
#endif 
 
    /* Referenced segment count and retention flags */ 
    if (segment->nrefsegs <= 4) { 
	/* Short form referenced segment info */ 
 
	/* Bits 5-7: count */ 
	cbuf = (segment->nrefsegs << 5); 
	/* Bit 0: self-retain bit */ 
	cbuf |= segment->retain ? 1 : 0; 
 
	/* Bits 1-4: retain flags for referenced segments 0-3 */ 
	for (i = 0; i < segment->nrefsegs; i++) { 
	    cbuf |= segment->refretain[i] ? (1 << (i + 1)) : 0; 
	    } 
 
	DSPUTC(data, (Dgp_UInt8)cbuf); 
	} 
    else { 
	/* Long form referenced segment info */ 
 
	/* 4 byte count + (7 << 29) */ 
	to4byte(lbuf, segment->nrefsegs + (7 << 29)); 
	DSWRITE(data, 4, lbuf, &bytes); 
 
	/* First byte: bit 0 is self-retain */ 
	cbuf = segment->retain ? 1 : 0; 
	/* Bits 1-7: retain flags for referenced segments 0-6 */ 
	for (i = 0; i < MIN(7, segment->nrefsegs); i++) { 
	    cbuf |= segment->refretain[i] ? (1 << (i + 1)) : 0; 
	    } 
 
	DSPUTC(data, (Dgp_UInt8)cbuf); 
 
	/* Subsequent bytes: 8 retain flags each, filled LSB-to-MSB */ 
	for (i = 7; i < segment->nrefsegs; i += 8) { 
	    cbuf = 0; 
	    for (j = 0; j < MIN(8, segment->nrefsegs - i); j++) { 
			cbuf |= segment->refretain[i + j] ? (1 << j) : 0; 
		} 
		DSPUTC(data, (Dgp_UInt8)cbuf); 
	} 
	} 
 
    /* Referred-to segment numbers */ 
    if (segment->segnum <= 256) { 
	/* One byte each */ 
	for (i = 0; i < segment->nrefsegs; i++) { 
	    DGP_ASSERT(segment->refsegs[i] < segment->segnum); 
	    cbuf = segment->refsegs[i]; 
	    DSPUTC(data, (Dgp_UInt8)cbuf); 
#ifdef	VALPRINT 
	    fprintf(stderr, "SEGREF %d\n", segment->refsegs[i]); 
#endif 
	    } 
	} 
    else if (segment->segnum <= 65536) { 
	/* Two bytes each */ 
	for (i = 0; i < segment->nrefsegs; i++) { 
	    DGP_ASSERT(segment->refsegs[i] < segment->segnum); 
	    to2byte(lbuf, segment->refsegs[i]); 
	    DSWRITE(data, 2, lbuf, &bytes); 
#ifdef	VALPRINT 
	    fprintf(stderr, "SEGREF %d\n", segment->refsegs[i]); 
#endif 
	    } 
	} 
    else { 
	/* Four bytes each */ 
	for (i = 0; i < segment->nrefsegs; i++) { 
	    DGP_ASSERT(segment->refsegs[i] < segment->segnum); 
	    to4byte(lbuf, segment->refsegs[i]); 
	    DSWRITE(data, 4, lbuf, &bytes); 
#ifdef	VALPRINT 
	    fprintf(stderr, "SEGREF %d\n", segment->refsegs[i]); 
#endif 
	    } 
	} 
 
    /* Segment page association */ 
    if (segment->pagenum < 256) { 
	/* One byte */ 
	cbuf = segment->pagenum; 
	DSPUTC(data, (Dgp_UInt8)cbuf); 
	} 
    else { 
	/* Four bytes */ 
	to4byte(lbuf, segment->pagenum); 
	DSWRITE(data, 4, lbuf, &bytes); 
	} 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGPAGE %d\n", segment->pagenum); 
#endif 
 
    /* Segment data length - 4 bytes; value currently unknown */ 
    DGP_ERR_PASS_CLEANUP(dsMark(data, 4)); 
	 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* Fill in the 4-byte segment header data length, now known */ 
 
ErrCode 
jb2EncodeSegmentHeaderDataLength(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2EncodeSegmentHeaderDataLength()", jb2_pErrState(file)); 
    Dgp_UInt8 lbuf[4]; 
    DataSink data = jb2GetFileHeaderDataSink(file); 
 
    to4byte(lbuf, segment->datalen); 
 
    DGP_ERR_PASS_CLEANUP(dsFillMark(data, 4, lbuf)); 
#ifdef	VALPRINT 
    fprintf(stderr, "SEGDATALEN %ld\n", segment->datalen); 
#endif 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    ERR_RETURN; 
    } 
 
/* 
 * Free the common stuff, then dispatch to a type-specific free routine. 
 * Don't free the actual segment. 
 */ 
void 
jb2FreeSegmentInfo(jb2Segment *segment, Dgp_State *dgp_state) 
{ 
  Dgp_Allocators *a = &(dgp_state->a); 
 
  if (segment->refsegs != (Dgp_Int32 *)NULL) { 
	  (a->free)((void *)segment->refsegs); 
	  segment->refsegs = NULL;
  } 
  if (segment->refretain != (Dgp_Bool *)NULL) { 
	  (a->free)((void *)segment->refretain); 
	  segment->refretain = NULL;
  } 
  if (isRegionSegtype(segment->segtype) && segment->reginfo.rendered) { 
	  DgpImage_free(segment->reginfo.im); 
  } 
  (*freeers[segment->segtype])(segment, dgp_state); 
  return;
} 
 
/* Render the region into a buffer */ 
ErrCode 
jb2RenderRegion(jb2File file, jb2Segment *segment) 
{ 
  DGP_PROLOGUE("jb2RenderRegion()", jb2_pErrState(file)); 
 
  DGP_ERR_PASS((*renderers[segment->segtype])(file, segment)); 
 
  NORMAL_RETURN; 
    } 
 
/* Render the region into the page */ 
ErrCode 
jb2RenderRegionIntoPage(jb2File file, jb2Segment *segment, DgpImage_Packed page) 
{ 
  DGP_PROLOGUE("jb2RenderRegion()", jb2_pErrState(file)); 
 
  DGP_ERR_PASS((*intorenderers[segment->segtype])(file, segment, page)); 
 
  NORMAL_RETURN; 
    } 
