/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/segment.h,v 1.23 2003/02/10 22:53:27 mkrocken Exp $
 *
 */

#ifndef	SEGMENT_H

#define	SEGMENT_H

#include "util.h"
#include "image.h"
#include "hufftable.h"
#include "symbol.h"
#include "jb2halftone.h"
#include "datasource.h"
#include "jb2file.h"
#include "params.h"

#ifdef __cplusplus
extern "C"
{
#endif

/*
 * Definitions of various segment types
 */

/* Combination operators */
#define	COMBOP_OR	0
#define	COMBOP_AND	1
#define	COMBOP_XOR	2
#define	COMBOP_XNOR	3
#define	COMBOP_REPLACE	4

/* Utility: map from COMBOPs to blit.h's minterms */
extern const Dgp_Int32 jb2Minterms[];

/* Shared information from region-type segments */
typedef struct _jb2RegSegment{
    Dgp_UInt32 regwidth, regheight, regxpos, regypos;
    Dgp_UInt8 combop;
    Dgp_Bool rendered;		/* Is im valid? */
    DgpImage_Packed im;		/* If it's been rendered, here it is */
    } jb2RegSegment;

/* Symbol dictionary */
typedef struct _jb2SymDicSegment{
    Dgp_Int32 nsymbols;
    jb2Symbol *symbols;
    Dgp_UInt8 *direct_contexts;
    Dgp_UInt8 *ref_contexts;
    } jb2SymDictSegment;

/* Decoded, not yet rendered, symbol region */
typedef struct _jb2SymRegSegment{
    Dgp_Int32 ninstances;
    jb2SymbolInstance *instances;
    Dgp_UInt8 int_combop, defpixel;
    } jb2SymRegSegment;

/* Halftone dictionary */
typedef struct _jb2HTDictSegment{
    Dgp_Int32 npatterns;
    Dgp_UInt16 pat_w, pat_h;
    jb2Pattern *patterns;
    } jb2HTDictSegment;

/* Decoded, not yet rendered, halftone region */
typedef struct jb2HTRegSegment{
    DgpImage_Short cells;
    Dgp_Int32 hgx, hgy, hrx, hry;
    Dgp_UInt8 int_combop, defpixel;
    struct _jb2Segment *dictseg; /* Our dictionary */
    } jb2HTRegSegment;

/* Not yet decoded generic region */
typedef struct _jb2GenRegSegment{
    Dgp_Bool mmr;
    arith_bitmap_params abparams;
    Dgp_UInt8 *data;
    size_t datalen;
    } jb2GenRegSegment;

/* Not yet decoded generic region refinement */
typedef struct _jb2GenRefSegment{
    arith_ref_params refparams;
    Dgp_UInt8 *data;
    size_t datalen;
    struct _jb2Segment *refseg;	/* Segment to be refined; NULL if page */
    } jb2GenRefRegSegment;

/* Page information */
typedef struct _jb2PageInfoSegment{
    Dgp_UInt32 width, height;
    Dgp_UInt32 xres, yres;
    Dgp_Bool lossless;
    Dgp_Bool may_refine;
    Dgp_UInt8 defpixel;
    Dgp_UInt8 combop;
    Dgp_Bool may_have_auxbuf;
    Dgp_Bool may_override_combop;
    Dgp_Bool striped;
    Dgp_UInt16 max_stripe;
    } jb2PageInfoSegment;

/* End of page information */
typedef struct _jb2EOPageSegment{
    Dgp_Int8 dummy;
    } jb2EOPageSegment;

/* End of stripe information */
typedef struct _jb2EOStripeSegment{
    Dgp_UInt32 last_row;
    } jb2EOStripeSegment;

/* End of file information */
typedef struct _jb2EOFileSegment{
    Dgp_Int8 dummy;
    } jb2EOFileSegment;

/* Profiles */
typedef struct _jb2ProfileSegment {
    Dgp_UInt32 nprofiles;
    Dgp_UInt32 *profiles;
    } jb2ProfileSegment;

/* Tables */
typedef struct _jb2TableSegment {
    hufftable_spec table;
    } jb2TableSegment;

/* Extension */
typedef struct _jb2ExtensionSegment{
    Dgp_UInt32 type;
    void *data;

	/* 1.5.2001 / dhtesq.  Added these userdata fields to fix */
	/* a memory leak associated with the unicode extensions (we had */
	/* been stealing the type field for unintended purposes). */
	Dgp_Int32 apple, orange;
	void *grape;
    } jb2ExtensionSegment;

/*
 * This is a universal segment, including all the header information.
 */
typedef struct _jb2Segment {
    Dgp_Int32 segnum;
    Dgp_UInt32 datalen;
    Dgp_Int32 pagenum;
    Dgp_Int32 nrefsegs;
    Dgp_Int32 *refsegs;
    Dgp_Bool *refretain;
    Dgp_UInt8 segtype;
    Dgp_Bool retain;
    Dgp_Bool defer_nonretain;
    jb2RegSegment reginfo;	/* Only valid for region segments */
    Dgp_Int32 refcount;
    union {
	jb2SymDictSegment symdict;
	jb2SymRegSegment symreg;
	jb2HTDictSegment htdict;
	jb2HTRegSegment htreg;
	jb2GenRegSegment genreg;
	jb2GenRefRegSegment genrefreg;
	jb2PageInfoSegment page;
	jb2EOPageSegment eopage;
	jb2EOStripeSegment eostripe;
	jb2EOFileSegment eofile;
	jb2ProfileSegment profiles;
	jb2TableSegment table;
	jb2ExtensionSegment extension;
	} data;
    } jb2Segment;

/*
 * Segment types
 */
#define	JB2SEGTYPE_SYMDICT	0
#define	JB2SEGTYPE_SYMREG_INT	4
#define	JB2SEGTYPE_SYMREG_IMM	6
#define	JB2SEGTYPE_SYMREG_IMMLL	7

#define	JB2SEGTYPE_HTDICT	16
#define	JB2SEGTYPE_HTREG_INT   	20
#define	JB2SEGTYPE_HTREG_IMM	22
#define	JB2SEGTYPE_HTREG_IMMLL	23

#define	JB2SEGTYPE_GENREG_INT	36
#define	JB2SEGTYPE_GENREG_IMM	38
#define	JB2SEGTYPE_GENREG_IMMLL	39

#define	JB2SEGTYPE_GENREFREG_INT	40
#define	JB2SEGTYPE_GENREFREG_IMM	42
#define	JB2SEGTYPE_GENREFREG_IMMLL	43

#define	JB2SEGTYPE_PAGEINFO	48
#define	JB2SEGTYPE_EOPAGE	49
#define	JB2SEGTYPE_EOSTRIPE	50
#define	JB2SEGTYPE_EOFILE	51
#define	JB2SEGTYPE_PROFILES	52
#define	JB2SEGTYPE_TABLE	53
#define	JB2SEGTYPE_EXTENSION	62

/*
 * Routines to decode segments
 */
extern ErrCode jb2DecodeSegment(jb2File file, DataSource data,
				jb2Segment *segment, jb2Segment **refsegs);
extern ErrCode jb2DecodeSegmentHeader(jb2File file, DataSource data,
				      jb2Segment *segment);
extern ErrCode jb2EncodeSegmentHeader(jb2File file, jb2Segment *segment);
extern ErrCode jb2EncodeSegmentHeaderDataLength(jb2File file,
						jb2Segment *segment);
extern void jb2FreeSegmentInfo(jb2Segment *segment, Dgp_State *dgp_state);
extern ErrCode jb2RenderRegion(jb2File file, jb2Segment *segment);
extern ErrCode jb2RenderRegionIntoPage(jb2File file, jb2Segment *segment,
				       DgpImage_Packed page);

/*
 * Useful macros
 */
#define	to4byte(lb, v) do { \
    (lb)[0] = (Dgp_UInt8)(((v) >> 24) & 0xff); \
    (lb)[1] = (Dgp_UInt8)(((v) >> 16) & 0xff); \
    (lb)[2] = (Dgp_UInt8)(((v) >> 8) & 0xff); \
    (lb)[3] = (Dgp_UInt8)((v) & 0xff); \
    } while(0)

#define	to2byte(sb, v) do { \
    (sb)[0] = ((v) >> 8) & 0xff; \
    (sb)[1] = (v) & 0xff; \
    } while(0)

#define	isRegionSegtype(t)	((((t) & 0x30) != 0x30) && (((t) & 0xc) != 0))
#define	isDictSegtype(t)	((((t) & 0x30) != 0x30) && (((t) & 0xc) == 0))
#define	isImmediateRegion(t)	((((t) & 0x30) != 0x30) && (((t) & 0xc) != 0) && (((t) & 0x2) != 0))
#define	isLosslessRegion(t)	((((t) & 0x30) != 0x30) && (((t) & 0xc) != 0) && (((t) & 0x1) != 0))
#define	isRefinementRegion(t)	((((t) & 0x30) != 0x30) && (((t) & 0xc) == 2))

#ifdef __cplusplus
}
#endif

#endif
