/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/symdec.c,v 1.19 2003/01/27 21:48:37 awilliam Exp $ 
 * 
 */ 
 
/* 
 * This module decodes symbol codes, whether arithmetic or Huffman encoded. 
 * There are two ways to build a symbol ID decoder: one arithmetic and 
 * one Huffman.  In either case, you pass in an array of void *, each of 
 * which points to a jb2Symbol.  The arithmetic one also needs an DgpEncode_mqdDecoder; 
 * the Huffman one needs a jbitfile and an array of code lengths. 
 * (It has to be an array of void *, not jb2Symbol *, because that makes it 
 * possible to generate a Huffman table directly from it). 
 * 
 * The array of void * passed in is not used after the symDecoder has 
 * been created, and can be freed as desired.  On the other hand, the 
 * DgpEncode_mqdDecoder or jbitfile are used as Dgp_Int32 as the symDecoder exists. 
 * 
 * Once you have a symDecoder, you just call symDecode(); it reads a symbol 
 * code and returns the corresponding jb2Symbol *.  It returns NULL if 
 * something goes wrong. 
 * 
 * symDecoderFree gets rid of a symDecoder. 
 */ 
 
#include "jbitfile.h" 
#include "symbol.h" 
#include "mqcoder.h" 
#include "hufftable.h" 
#include "symdec.h" 
#include "PsgFix.h"
#ifdef	VALPRINT 
#include <stdio.h> 
#endif 
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/symdec.c,v 1.19 2003/01/27 21:48:37 awilliam Exp $"); 
 
 
/* Internal state */ 
struct _symDecoder { 
    Dgp_Bool arith; 
    Dgp_State *dgp_state; 
    union { 
	struct { 
	    void **symptrs;	/* Pointers for the int->pointer map */ 
	    DgpEncode_MQDecoder *dec;	/* Arithmetic decoder */ 
	    Dgp_Int32 nsyms;		/* How many symbols are valid */ 
	    Dgp_Int32 codelen;	/* How many bits in a code */ 
	    Dgp_UInt8 *cx;	/* Contexts for decoding */ 
	    } arith; 
	struct { 
	    hufftable_spec tab;	/* Huffman table */ 
	    jbitfile bf;	/* and the bitfile to read from */ 
	    } huffman; 
	} info; 
    }; 
 
ErrCode 
symDecoderHuffman(void **symptrs, Dgp_Int32 nsyms, 
		  jbitfile bf, Dgp_Int32 *lengths,  
		  Dgp_State *dgp_state, symDecoder *_newdec) 
{ 
  DGP_PROLOGUE("symDecoderHuffman()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    symDecoder newdec = (symDecoder)NULL; 
    table_line *lines = (table_line *)NULL; 
    Dgp_Int32 i; 
 
    *_newdec = (symDecoder)NULL; 
     
    newdec = (symDecoder)(a->malloc)(sizeof(*newdec)); 
    DGP_CHECK_ALLOC_CLEANUP(newdec); 
    newdec->dgp_state = dgp_state; 
     
    /* Allocate the table lines for the Huffman table */ 
    lines = (table_line *)(a->malloc)(nsyms * sizeof(*lines)); 
    DGP_CHECK_ALLOC_CLEANUP(lines); 
  
    /* Fill in the table */ 
    newdec->info.huffman.tab.lines = lines; 
    newdec->info.huffman.tab.nlines = nsyms; 
    newdec->info.huffman.tab.lower_valid = 
	newdec->info.huffman.tab.upper_valid = 
	newdec->info.huffman.tab.oob_valid = DGP_FALSE; 
 
    /* including filling in the lines */ 
    for (i = 0; i < nsyms; i++) { 
	lines[i].preflen = (Dgp_Int8)lengths[i]; 
	lines[i].base = lines[i].rangelen = 0; 
	} 
 
    /* Assign the codes */ 
    DGP_ERR_PASS_CLEANUP(jhtSetCodes(&newdec->info.huffman.tab, 
				     dgp_state)); 
 
    /* 
     * Compile the table so that jhtReadCode will return a pointer to 
     * the symbol corresponding to the code that was read. 
     */ 
    DGP_ERR_PASS_CLEANUP(jhtCompile(&newdec->info.huffman.tab, symptrs, 
				    dgp_state)); 
 
    newdec->info.huffman.bf = bf; 
    newdec->arith = DGP_FALSE; 
 
    *_newdec = newdec; 
    newdec = (symDecoder)NULL; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (newdec != (symDecoder)NULL) { 
	(a->free)((void *)newdec); 
	} 
 
    if (lines != (table_line *)NULL) { 
	(a->free)((void *)lines); 
	} 
 
    ERR_RETURN; 
    } 
 
#ifndef	NOMQ 
ErrCode 
symDecoderArith(void **symptrs, Dgp_Int32 nsyms, DgpEncode_MQDecoder *dec,  
		Dgp_State *dgp_state, symDecoder *_newdec) 
{ 
  DGP_PROLOGUE("symDecoderArith()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    symDecoder newdec = (symDecoder)NULL; 
    void **ptrcopy = (void **)NULL; 
    Dgp_UInt8 *cx = (Dgp_UInt8 *)NULL; 
     
    *_newdec = (symDecoder)NULL; 
 
    newdec = (symDecoder)(a->malloc)(sizeof(*newdec)); 
    DGP_CHECK_ALLOC_CLEANUP(newdec); 
    newdec->dgp_state = dgp_state; 
     
    /* We need a copy of the pointers so our caller can free the originals */ 
    ptrcopy = (void **)(a->malloc)(nsyms * sizeof(*ptrcopy)); 
    DGP_CHECK_ALLOC_CLEANUP(ptrcopy); 
    (void)memcpy(ptrcopy, symptrs, nsyms * sizeof(*ptrcopy)); 
 
    newdec->info.arith.symptrs = ptrcopy; 
    newdec->info.arith.dec = dec; 
    newdec->info.arith.nsyms = nsyms; 
 
    /* How many bits do we need? */ 
    newdec->info.arith.codelen = ceillog2(nsyms); 
 
    /* We need 2^(codelen+1) bytes of context */ 
    cx = (Dgp_UInt8 *) 
	(a->calloc)((size_t)(1l << (newdec->info.arith.codelen + 1)), 1); 
    DGP_CHECK_ALLOC_CLEANUP(cx); 
 
    newdec->info.arith.cx = cx; 
 
    newdec->arith = DGP_TRUE; 
 
    *_newdec = newdec; 
    newdec = (symDecoder)NULL; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (newdec != (symDecoder)NULL) { 
	(a->free)((void *)newdec); 
	} 
 
    if (ptrcopy != (void **)NULL) { 
	(a->free)((void *)ptrcopy); 
	} 
 
    if (cx != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)cx); 
	} 
 
    ERR_RETURN; 
    } 
#endif 
 
ErrCode 
symDecode(symDecoder dec, jb2Symbol **symnum) 
{ 
  DGP_PROLOGUE("symDecode()", dec->DGP_PERRSTATE); 
 
    if (dec->arith) { 
#ifndef	NOMQ 
	/* Arithmetic method of decoding */ 
	Dgp_UInt32 prev = 1; 
	int i; 
	int bit; 
 
	/* Read codelen bits; context is previous bits plus a leading 1 */ 
	for (i = 0; i < dec->info.arith.codelen; i++) { 
	    bit = DgpEncode_mqdDecode(dec->info.arith.dec, &dec->info.arith.cx[prev]); 
	    /* DGP_REPORTF1("%d", bit);  */
	    prev = (prev << 1) | bit; 
	    } 
 
	/* Clear out the leading 1 */ 
	prev -= (1ul << dec->info.arith.codelen); 
 
#ifdef	VALPRINT 
	fprintf(stderr, "sn %ld\n", prev); 
#endif 
	/* That's the index into symptrs */ 
	if ((int)prev < dec->info.arith.nsyms) { 
	  *symnum = ((jb2Symbol *)dec->info.arith.symptrs[prev]); 
	    } 
	else { 
	  *symnum = ((jb2Symbol *)NULL); 
	    } 
	  NORMAL_RETURN; 
#endif 
	} 
    else { 
	/* Huffman method of decoding */ 
	jb2Symbol *sym; 
 
	DGP_ERR_PASS(jhtReadCode(&dec->info.huffman.tab, dec->info.huffman.bf, 
				 dec->dgp_state, (void **)&sym)); 
#ifdef	VALPRINT 
	{ 
	    /* 
	     * you're not really going to do this are you? 
	     * 
	     * in order to find out what the index of the code that was 
	     * read was, we need to do some serious grovelling through the 
	     * table structure.  first, we look through the "decoded" table, 
	     * searching for the entry that matches sym.  once we have that, 
	     * we know what code (and codelen) the decoder just saw.  then 
	     * we can scan the table lines until we find the line that 
	     * matches that code and codelen.  finally, we have the index of 
	     * that line. 
	     */ 
	    Dgp_Int32 ii, codelen; 
	    Dgp_UInt32 jj = 0, code; 
 
	    for (ii = dec->info.huffman.tab.minlen; 
		 ii < JHT_MAXCODELEN; 
		 ii++) { 
		for (jj = dec->info.huffman.tab.firstcode[ii]; 
		     jj <= dec->info.huffman.tab.lastcode[ii]; 
		     jj++) { 
		    if (dec->info.huffman.tab.decoded[ii][jj - dec->info.huffman.tab.firstcode[ii]] == (void *)sym) { 
			/* found it */ 
			break; 
			} 
		    } 
		if (jj <= dec->info.huffman.tab.lastcode[ii]) { 
		    break; 
		    } 
		} 
	    if (ii == JHT_MAXCODELEN) { 
		panic("what the hell are you doing"); 
		} 
	    codelen = ii; 
	    code = jj; 
	    for (ii = 0; ii < dec->info.huffman.tab.nlines; ii++) { 
		if ((dec->info.huffman.tab.lines[ii].preflen == codelen) && 
		    (dec->info.huffman.tab.lines[ii].prefix == code)) { 
		    fprintf(stderr, "sn %d\n", ii); 
		    break; 
		    } 
		} 
	    if (ii == dec->info.huffman.tab.nlines) { 
		panic("the hole is very deep now"); 
		} 
	    } 
#endif 
 
	*symnum = sym; 
	NORMAL_RETURN; 
	} 
    } 
 
extern void 
symDecoderFree(symDecoder dec) 
{ 
  Dgp_Free tmp_free = dec->DGP_FREE; 
 
    if (dec->arith) { 
	(tmp_free)((void *)dec->info.arith.symptrs); 
	(tmp_free)((void *)dec->info.arith.cx); 
	} 
    else { 
	jhtCleanup(&dec->info.huffman.tab, dec->dgp_state); 
	(tmp_free)((void *)dec->info.huffman.tab.lines); 
	} 
    (tmp_free)((void *)dec); 
    } 
 
	 
