/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/symenc.c,v 1.8 2003/01/27 21:48:37 awilliam Exp $ 
 * 
 */ 
 
#include "jbitfile.h" 
#include "symbol.h" 
#include "mqcoder.h" 
#include "hufftable.h" 
#include "symenc.h" 
#include "symdec.h"
#include "PsgFix.h"
#ifdef	VALPRINT 
#include <stdio.h> 
#endif 
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libjbig2/symenc.c,v 1.8 2003/01/27 21:48:37 awilliam Exp $"); 
 
 
/* Internal state */ 
struct _symEncoder { 
    Dgp_Bool arith; 
    Dgp_State *dgp_state; 
    union { 
	struct { 
	    DgpEncode_MQEncoder *enc;	/* Arithmetic encoder */ 
	    Dgp_Int32 nsyms;		/* How many symbols are valid */ 
	    Dgp_Int32 codelen;	/* How many bits in a code */ 
	    Dgp_UInt8 *cx;	/* Contexts for encoding */ 
	    } arith; 
	struct { 
	    hufftable_spec tab;	/* Huffman table */ 
	    jbitfile bf;	/* and the bitfile to write to */ 
	    } huffman; 
	} info; 
    }; 
 
ErrCode 
symEncoderHuffman(int nsyms, jbitfile bf, Dgp_Int32 *lengths, Dgp_State *dgp_state, 
		  symEncoder *_newenc) 
{ 
  DGP_PROLOGUE("symEncoderHuffman()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    symEncoder newenc = (symEncoder)NULL; 
    table_line *lines = (table_line *)NULL; 
    Dgp_Int32 i; 
	 
    *_newenc = (symEncoder)NULL; 
 
    newenc = (symEncoder)(a->malloc)(sizeof(*newenc)); 
    DGP_CHECK_ALLOC_CLEANUP(newenc); 
    newenc->dgp_state = dgp_state; 
 
    /* Allocate the table lines for the Huffman table */ 
    lines = (table_line *)(a->malloc)(nsyms * sizeof(*lines)); 
    DGP_CHECK_ALLOC_CLEANUP(lines); 
  
    /* Fill in the table */ 
    newenc->info.huffman.tab.nlines = nsyms; 
    newenc->info.huffman.tab.lines = lines; 
    newenc->info.huffman.tab.lower_valid = 
	newenc->info.huffman.tab.upper_valid = 
	newenc->info.huffman.tab.oob_valid = DGP_FALSE; 
 
    for (i = 0; i < nsyms; i++) { 
	newenc->info.huffman.tab.lines[i].preflen = (Dgp_Int8)lengths[i]; 
	newenc->info.huffman.tab.lines[i].rangelen = 0; 
	newenc->info.huffman.tab.lines[i].base = i; 
	} 
 
    /* Assign the Huffman codes */ 
    DGP_ERR_PASS_CLEANUP(jhtSetCodes(&newenc->info.huffman.tab, 
				     dgp_state)); 
 
    newenc->info.huffman.bf = bf; 
    newenc->arith = DGP_FALSE; 
 
    *_newenc = newenc; 
    newenc = (symEncoder)NULL; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (newenc != (symEncoder)NULL) { 
	(a->free)((void *)newenc); 
	} 
 
    if (lines != (table_line *)NULL) { 
	(a->free)((void *)lines); 
	} 
 
    ERR_RETURN; 
    } 
 
#ifndef	NOMQ 
ErrCode 
symEncoderArith(int nsyms, DgpEncode_MQEncoder *enc, Dgp_State *dgp_state,  
		symEncoder *_newenc) 
{ 
  DGP_PROLOGUE("symEncoderArith()", DGP_PERRSTATE); 
 
  Dgp_Allocators *a = &(dgp_state->a); 
    symEncoder newenc = (symEncoder)NULL; 
    Dgp_UInt8 *cx = (Dgp_UInt8 *)NULL; 
 
    *_newenc = (symEncoder)NULL; 
 
    newenc = (symEncoder)(a->malloc)(sizeof(*newenc)); 
    DGP_CHECK_ALLOC_CLEANUP(newenc); 
    newenc->dgp_state = dgp_state; 
 
    newenc->info.arith.enc = enc; 
    newenc->info.arith.nsyms = nsyms; 
 
    /* How many bits do we need? */ 
    newenc->info.arith.codelen = ceillog2(nsyms); 
 
    /* We need 2^(codelen+1) bytes of context */ 
    cx = (Dgp_UInt8 *) 
	(a->calloc)((size_t)(1l << (newenc->info.arith.codelen + 1)), 1); 
    DGP_CHECK_ALLOC_CLEANUP(cx); 
 
    newenc->info.arith.cx = cx; 
 
    newenc->arith = DGP_TRUE; 
 
    *_newenc = newenc; 
    newenc = (symEncoder)NULL; 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
    if (newenc != (symEncoder)NULL) { 
	(a->free)((void *)newenc); 
	} 
 
    if (cx != (Dgp_UInt8 *)NULL) { 
	(a->free)((void *)cx); 
	} 
 
    ERR_RETURN; 
    } 
#endif 
 
ErrCode 
symEncode(symEncoder enc, Dgp_Int32 symnum) 
{ 
  DGP_PROLOGUE("symEncode()", enc->DGP_PERRSTATE); 
 
    if (enc->arith) { 
#ifndef	NOMQ 
	/* Arithmetic method of encoding */ 
	Dgp_UInt32 prev = 1; 
	int i; 
	int bit; 
	 
	/* Write codelen bits; context is previous bits plus a leading 1 */ 
	for (i = 0; i < enc->info.arith.codelen; i++) { 
	    bit = (symnum >> (enc->info.arith.codelen - 1 - i)) & 1; 
	    DGP_ERR_GEN_CLEANUP((!DgpEncode_mqeEncode(enc->info.arith.enc, 
					    &enc->info.arith.cx[prev], bit)), 
				cErrIO_WriteError); 
	    prev = (prev << 1) | bit; 
	    } 
#endif 
	} 
    else { 
	/* Huffman method of encoding */ 
	table_line *curline = &enc->info.huffman.tab.lines[symnum]; 
	DGP_ERR_PASS(jbfWriteBits(enc->info.huffman.bf, 
			    curline->prefix, curline->preflen)); 
	} 
 
    NORMAL_RETURN; 
 
  ERR_CLEANUP: 
 
    ERR_RETURN; 
    } 
 
void 
symEncoderFree(symEncoder enc) 
{ 
  Dgp_Free tmp_free = enc->DGP_FREE; 
 
    if (enc->arith) { 
	/* Free contexts */ 
	(tmp_free)((void *)enc->info.arith.cx); 
	} 
    else { 
	/* Free table lines */ 
	(tmp_free)((void *)enc->info.huffman.tab.lines); 
	} 
    (tmp_free)((void *)enc); 
    } 
