/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/uncomp-jbig.c,v 1.12 2003/02/10 22:53:27 mkrocken Exp $
 *
 */

 /*
 * This is going to be a JBIG2 decompressor: read segments,
 * produce pages, write them out.
*/

#define DEF_FILE_ERROR_OUTPUT

#include <stdio.h>
#include <stdlib.h>
#include "list.h"
#include "blit.h"
#include "dgp_err_printers.h"
#include "memorydatasource.h"
#include "segment_util.h"
#include "image_defs.h"
#include "uncomp-jbig.h"
#include "PsgFix.h"

/* for extensions */
#include "decextseg.h"


static ErrCode processSimpleSegment(jb2Segment *segment, jb2info *info);
static ErrCode processSymRegSegment(jb2Segment *segment, jb2info *info);
static ErrCode processGenRegSegment(jb2Segment *segment, jb2info *info);
static ErrCode processHTRegSegment(jb2Segment *segment, jb2info *info);
static ErrCode processGenRefRegSegment(jb2Segment *segment, jb2info *info);
static ErrCode processPageInfoSegment(jb2Segment *segment, jb2info *info);
static ErrCode processEOPageSegment(jb2Segment *segmente, jb2info *info);
static ErrCode processEOStripeSegment(jb2Segment *segment, jb2info *info);
static ErrCode processExtensionSegment(jb2Segment *segment, jb2info *info);

/* Most of the segments can be handled simply */
#define	processSymDictSegment	processSimpleSegment
#define	processEOFileSegment	processSimpleSegment
#define	processHuffTableSegment	processSimpleSegment
#define	processHTDictSegment	processSimpleSegment
#define	processProfilesSegment	processSimpleSegment

/* Print out all the segment header information */
static void dumpheader(jb2Segment *curseg);

/* Segment type names */
unsigned char *typenames[] = {
    "Symbol dictionary", "INVALID", "INVALID", "INVALID",
    "Intermediate symbol region", "INVALID", "Immediate symbol region", "Immediate lossless symbol region",
    "INVALID", "INVALID", "INVALID", "INVALID",
    "INVALID", "INVALID", "INVALID", "INVALID",

    "Halftone dictionary", "INVALID", "INVALID", "INVALID",
    "Intermediate halftone region", "INVALID", "Immediate halftone region", "Immediate lossless halftone region",
    "INVALID", "INVALID", "INVALID", "INVALID",
    "INVALID", "INVALID", "INVALID", "INVALID",

    "INVALID", "INVALID", "INVALID", "INVALID",
    "Intermediate generic region", "INVALID", "Immediate generic region", "Immediate lossless generic region",
    "Intermediate generic refinement region", "INVALID", "Immediate generic refinement region", "Immediate lossless generic refinement region",
    "INVALID", "INVALID", "INVALID", "INVALID",

    "Page information", "End of page", "End of stripe", "End of file",
    "Profiles", "Table", "INVALID", "INVALID",
    "INVALID", "INVALID", "INVALID", "INVALID",
    "INVALID", "INVALID", "Extension", "INVALID"
    };


static ErrCode (*processors [64])(jb2Segment *, jb2info *) = {
    processSymDictSegment, NULL, NULL, NULL,
    processSymRegSegment, NULL, processSymRegSegment, processSymRegSegment,
    NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL,

    processHTDictSegment, NULL, NULL, NULL,
    processHTRegSegment, NULL, processHTRegSegment, processHTRegSegment,
    NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL,

    NULL, NULL, NULL, NULL,
    processGenRegSegment, NULL, processGenRegSegment, processGenRegSegment,
    processGenRefRegSegment, NULL, processGenRefRegSegment, processGenRefRegSegment,
    NULL, NULL, NULL, NULL,

    processPageInfoSegment, processEOPageSegment, processEOStripeSegment, processEOFileSegment,
    processProfilesSegment, processHuffTableSegment, NULL, NULL,
    NULL, NULL, NULL, NULL,
    NULL, NULL, processExtensionSegment, NULL
    };

/* default allocimage function */
ErrCode myFreeImage(  Dgp_State * dgp_state,
                    DgpImage_Packed* dgpImage)
{
    DGP_PROLOGUE("myFreeImage()",DGP_PERRSTATE);
    DGP_STOP_WHINE;
    if(*dgpImage) {
            DgpImage_free (*dgpImage);
            *dgpImage = NULL; 
    }
    NORMAL_RETURN;
}
ErrCode myAllocImage( 
                   Dgp_State * dgp_state,
                   Dgp_Int32 width, Dgp_Int32 height,
                   DgpImage_Packed* dgpImage, void** usrImage)
{
    DGP_PROLOGUE("allocImage()",DGP_PERRSTATE);
    DgpImage_Packed image = NULL;

    DGP_CHECK_NULL(dgp_state);
    DGP_CHECK_NULL(dgpImage);

    DGP_ERR_PASS_CLEANUP( DgpImage_newAligned( dgp_state,  DGPIMAGE_PACKED, 
                                            width, height, (void**)(&image) ));
    *dgpImage = image;
	if(usrImage) *usrImage = image;
    NORMAL_RETURN;
ERR_CLEANUP:
    if(image) {
            DgpImage_free(image);
            image=NULL;
    }
    ERR_RETURN;
}
/* Decode a simple (no action needed after decoding) segment */
ErrCode
processSimpleSegment(jb2Segment *segment, jb2info *info)
{
    DGP_PROLOGUE("processSimpleSegment()", info->DGP_PERRSTATE);
    
    Dgp_Allocators *a = &(info->DGP_ALLOCATORS);
    jb2Segment **refsegs = (jb2Segment **)NULL;
    int oldcount;
    
    DGP_REPORTF1((info->verbose), "decoding segment %d\n", segment->segnum);
    
    /* Marshal the referenced segments */
    DGP_ERR_PASS_CLEANUPF0(jb2MarshalSegments(info->f, segment,
        info->activesegs, &refsegs),
        "can't marshal segments\n");
    
    /* How many bytes consumed so far? */
    oldcount = info->data->bufpos;
    
    /* Call the decoding routine */
    DGP_ERR_PASS_CLEANUPF0(jb2DecodeSegment(info->f, info->data, segment,
        refsegs),
        "can't decode segment\n");
    
    /* Check bytes */
    DGP_ERR_GEN_CLEANUPF2(
		((info->data)->bufpos - oldcount != segment->datalen), 
		cErrDecodeFailure,
        "segment claimed %d bytes but consumed %d\n",
        (int)segment->datalen,
        (int)((info->data)->bufpos - oldcount));
    
    /* Don't need refsegs any more */
    if (refsegs != (jb2Segment **)NULL) {
        (a->free)((void *)refsegs);
    }
    
    /* That's all we have to do. */
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (refsegs != (jb2Segment **)NULL) {
        (a->free)((void *)refsegs);
    }
    ERR_RETURN;
}

ErrCode
renderRegion(jb2Segment *segment, jb2info *info) 
{
    DGP_PROLOGUE("renderRegion()", info->DGP_PERRSTATE);
    
    Dgp_Allocators *a = &(info->DGP_ALLOCATORS);
    jb2RegSegment *clone = (jb2RegSegment *)NULL;
    DgpImage_Packed cloneim = (DgpImage_Packed)NULL;
    
    DGP_ASSERT(isRegionSegtype(segment->segtype));
    
    /* Check the page flags - make sure this is valid */
    DGP_ERR_GEN_CLEANUPF0(((!info->pageinfo.may_have_auxbuf) &&
        (!isImmediateRegion(segment->segtype))),
        cErrDecodeFailure,
        "intermediate region on a page that can't have aux buffers\n");
    
    DGP_ERR_GEN_CLEANUPF0(((!info->pageinfo.may_refine) &&
        (isRefinementRegion(segment->segtype))),
        cErrDecodeFailure,
        "refinement region on a page that can't have refinement\n");
    
    DGP_ERR_GEN_CLEANUPF0(((!info->pageinfo.may_override_combop) &&
        (segment->reginfo.combop != info->pageinfo.combop)),
        cErrDecodeFailure,
        "region combination operator doesn't match page combination operator\non a page where they must match\n");
    
    /* Check the page striping */
    if (info->pageinfo.striped) {
        DGP_ERR_GEN_CLEANUPF0(((int)segment->reginfo.regypos < info->laststripe),
            cErrDecodeFailure,
            "region goes into previous stripe\n");
        
        DGP_ERR_GEN_CLEANUPF0(((int)(segment->reginfo.regypos +
            segment->reginfo.regheight - 1) >
            info->laststripe + (int)info->pageinfo.max_stripe + 1),
            cErrDecodeFailure,
            "region makes stripe too high\n");
        
        /* Remember the last row touched  */
        info->lastregion = MAX(info->lastregion,
            (int)(segment->reginfo.regypos +
            segment->reginfo.regheight - 1));
    }

    if (isImmediateRegion(segment->segtype)) {
        /* Render it into the page */
        
        if (info->dunnosize) {
            DgpAdt_ListNode retnode;
            
            /* Can't render into the page because we don't know its height */
            DGP_ERR_PASS_CLEANUP(jb2RenderRegion(info->f, segment));
            
            /*
            * Now clone and stash the rendered region so we're not at
            * the mercy of the segment GC logic
            */
            clone = (jb2RegSegment *)(a->malloc)(sizeof(*clone));
            DGP_CHECK_ALLOC_CLEANUP(clone);
            *clone = segment->reginfo;
            
            DGP_ERR_PASS_CLEANUP(DgpImage_dup(segment->reginfo.im, DGP_FALSE, (void **)&cloneim));
            
            clone->im = cloneim;
            if(info->regions == DGP_NULLLIST)
                DgpAdt_liNew(&info->regions, info->dgp_state);
            DGP_CHECK_ALLOC_CLEANUP(info->regions);

            DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(info->regions, (void *)clone, &retnode));
        }
        else {
            /* Just render it into the page */
            DGP_ERR_GEN_CLEANUP(!info->curpage, cErrDecodeFailure); 
            DGP_ERR_PASS_CLEANUP(jb2RenderRegionIntoPage(info->f, segment,info->curpage));
        }
    }
    else {
        /* Just render it. */
        
        DGP_ERR_PASS_CLEANUP(jb2RenderRegion(info->f, segment));
    }
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (clone != (jb2RegSegment *)NULL) {
        (a->free)((void *)clone);
    }
    
    if (cloneim != (DgpImage_Packed)NULL) {
        DgpImage_free(cloneim);
    }

    if(info->regions) DgpAdt_liFree(info->regions);
    
    ERR_RETURN;	
}

/*
* Process a symbol region segment: decode it, mess around with the
* appropriate buffers.
*/
ErrCode
processSymRegSegment(jb2Segment *segment, jb2info *info)
{
    DGP_PROLOGUE("processSymRegSegment()", info->DGP_PERRSTATE);
    
    /* Step one: just do the segment decode */
    DGP_ERR_PASS(processSimpleSegment(segment, info));

    if(info->render)
	    DGP_ERR_PASS(renderRegion(segment, info));
    NORMAL_RETURN;
}

/*
* Process a generic region segment: decode it, mess around with the
* appropriate buffers.
*/
ErrCode
processGenRegSegment(jb2Segment *segment, jb2info *info)
{
    DGP_PROLOGUE("processGenRegSegment()", info->DGP_PERRSTATE);
    
    /* Step one: just do the segment decode */
    DGP_ERR_PASS(processSimpleSegment(segment, info));
    
    if(info->render)
        DGP_ERR_PASS(renderRegion(segment, info));
    
    NORMAL_RETURN;
}

/*
* Process a halftone region segment
*/
ErrCode
processHTRegSegment(jb2Segment *segment, jb2info *info)
{
    DGP_PROLOGUE("processHTRegSegment()", info->DGP_PERRSTATE);
    
    /* Step one: just do the segment decode */
    DGP_ERR_PASS(processSimpleSegment(segment, info));
    if(info->render)
        DGP_ERR_PASS(renderRegion(segment, info));
    
    NORMAL_RETURN;
}

/*
* Process a generic region segment: decode it, mess around with the
* appropriate buffers.
*/
ErrCode
processGenRefRegSegment(jb2Segment *segment, jb2info *info)
{
    DGP_PROLOGUE("processGenRefRegSegment()", info->DGP_PERRSTATE);
    
    /* Step one: just do the segment decode */
    DGP_ERR_PASS(processSimpleSegment(segment, info));
    
    DGP_ERR_GENF0((info->pageinfo.striped), cErrInvalidParm,
        "unimplemented combo of refinement and striping\n");
    if(info->render)
        DGP_ERR_PASS(renderRegion(segment, info));
    
    NORMAL_RETURN;
}

/*
 * Process a page info segment: decode it, create the page image
 */
ErrCode
processPageInfoSegment(jb2Segment *segment, jb2info *info)
{
    DGP_PROLOGUE("processPageInfoSegment()", info->DGP_PERRSTATE);
    
    /* Step one: just do the segment decode */
    DGP_ERR_PASS_CLEANUP(processSimpleSegment(segment, info));
    
    if(info->render){
        /* Be careful - may not know page size yet */
        if ((segment->data.page.striped) &&
            (segment->data.page.height == 0xffffffff)) {
            /* Page height unknown */
            info->curpage = (DgpImage_Packed)NULL;
            info->dunnosize = DGP_TRUE;
            /* Need a list of regions */
            DgpAdt_liNew(&info->regions, info->dgp_state);
            DGP_CHECK_ALLOC_CLEANUP(info->regions);
        }
        else {
		    info->dunnosize = DGP_FALSE;
            info->allocImage(info->dgp_state,
                            segment->data.page.width, segment->data.page.height,
                            &info->curpage, &info->usrPixr);
		    DGP_CHECK_ALLOC_CLEANUP(info->curpage);
		    info->regions = DGP_NULLLIST;
		    if (segment->data.page.defpixel != 0) {
			    DgpBlat_blitFillPacked(info->curpage);
	        }
	    }
    }
    info->pagenum = segment->pagenum;
    info->pageinfo = segment->data.page;
    info->laststripe = -1;
    info->lastregion = -1;
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if(info->curpage) info->freeImage(info->dgp_state, &info->curpage);
    ERR_RETURN;
}

/*
* Process an end-of-page segment: decode it, write out the page image,
* free the image
*/
ErrCode
processEOPageSegment(jb2Segment *segment, jb2info *info)
{
    DGP_PROLOGUE("processEOPageSegment()", info->DGP_PERRSTATE);
    
    Dgp_Allocators *a = &(info->DGP_ALLOCATORS);
    char fname[1000];
    
    /* Step one: just do the segment decode */
    DGP_ERR_PASS_CLEANUP(processSimpleSegment(segment, info));
    
    if(info->render){
        if (info->dunnosize) {
		    DgpAdt_ListNode ln;
            Dgp_Int32 pageHeight = 0;
	        Dgp_Int32 pageWidth = 0;

            if(info->pageinfo.striped){
		        /*
		        * We didn't know the page size.  The last EOStripe segment's
		        * end row is the page end row.
		        */
		        info->allocImage(info->dgp_state,
                    info->pageinfo.width, info->laststripe + 1,
                    &info->curpage, &info->usrPixr);
            }else{
                /* dunnosize is the result of having no pageinfo segment */
                /* get the page size of all the regioninfo size */
                /* Now we want to draw the list of segments into the page */
                DGPADT_FOREACH (ln, info->regions) {
                    jb2RegSegment *curreg = (jb2RegSegment *)DGPADT_LIGET(ln);
                    DGP_ERR_GEN_CLEANUP(!(curreg->im), cErrDecodeFailure);
                    pageHeight += DGPIMG_HEIGHT(curreg->im);
                    pageWidth = DGPIMG_WIDTH(curreg->im);
                }
		        info->allocImage(info->dgp_state,
			        pageWidth, pageHeight, &info->curpage, &info->usrPixr);


            }
            /* Now we want to draw the list of segments into the page */
            DGPADT_FOREACH (ln, info->regions) {
                jb2RegSegment *curreg = (jb2RegSegment *)DGPADT_LIGET(ln);
                DgpBlat_blitPackedOp(info->curpage, curreg->im,
                    curreg->regxpos, curreg->regypos,
                    jb2Minterms[curreg->combop]);
                DgpImage_free(curreg->im);
                (a->free)((void *)curreg);
            }
            DgpAdt_liFree(info->regions);
            info->regions = DGP_NULLLIST;
        }	    
        /* Save and free info->curpage */
        if (!info->noout && info->curpage) {
            if (info->format) {
                sprintf(fname, info->format, info->pagenum - 1); 
            }
            else {
                sprintf(fname, "%s.p%04d.pbm", info->fname, info->pagenum - 1);
            }
        
            DGP_ERR_PASS_CLEANUP(DgpImage_save(info->curpage, fname));
		    info->freeImage(info->dgp_state, &info->curpage);
        }
    }
    NORMAL_RETURN;
    
ERR_CLEANUP:
    ERR_RETURN;
}

ErrCode
processEOStripeSegment(jb2Segment *segment, jb2info *info)
{
    DGP_PROLOGUE("processEOStripeSegment()", info->DGP_PERRSTATE);
    
    /* Should we even have one of these? */
    DGP_ERR_GEN_CLEANUPF0((!info->pageinfo.striped), cErrDecodeFailure,
        "non-striped page has EOStripe segment\n");
    
    /* Decode the stripe */
    DGP_ERR_PASS_CLEANUP(processSimpleSegment(segment, info));
    
    DGP_ERR_GEN_CLEANUPF0(((int)segment->data.eostripe.last_row <
        info->laststripe), cErrDecodeFailure,
        "stripe goes backwards\n");
    
    DGP_ERR_GEN_CLEANUPF0(((int)segment->data.eostripe.last_row -
        info->laststripe > info->pageinfo.max_stripe),
        cErrDecodeFailure, "stripe size is too large\n");
    
    DGP_ERR_GEN_CLEANUPF0(((int)segment->data.eostripe.last_row <
        info->lastregion),  cErrDecodeFailure,
        "stripe doesn't include all regions within stripe\n");
    
    info->laststripe = segment->data.eostripe.last_row;
    NORMAL_RETURN;
    
ERR_CLEANUP:
    ERR_RETURN;
}

ErrCode
processExtensionSegment(jb2Segment *segment, jb2info *info)
{
    DGP_PROLOGUE("processExtensionSegment()", info->DGP_PERRSTATE);
    
    /* Step one: just do the segment decode */
    DGP_ERR_PASS_CLEANUP(processSimpleSegment(segment, info));
    
    /* Handle anything we recognise */
    switch(segment->data.extension.type) {
    case EXT_ASCIICOMMENT: 
        if (info->verbose) {
            int i;
            jb2Extension_ASCIIComment *comment =
                (jb2Extension_ASCIIComment *)segment->data.extension.data;
            
            printf("ASCII comment in segment %ld\n", segment->segnum);
            for (i = 0; i < comment->npairs; i++) {
                printf("%s\t|%s\n", comment->names[i], comment->values[i]);
            }
        }
    }
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    ERR_RETURN;
}   


/* Allocate *newseg; read a header into it. */
ErrCode
readheader(jb2info *info, jb2Segment **newseg)
{
    DGP_PROLOGUE("readheader()", info->DGP_PERRSTATE);
    
    Dgp_Allocators *a = &(info->DGP_ALLOCATORS);
    
    /* Allocate a new segment */
    *newseg = (jb2Segment *)(a->malloc)(sizeof(**newseg));
    DGP_CHECK_ALLOC_CLEANUP(*newseg);
    (*newseg)->refsegs = NULL;
    (*newseg)->refretain = NULL;
    (*newseg)->reginfo.im = NULL;
    (*newseg)->refcount = 1;

    /* Read in a header */
    DGP_ERR_PASS_CLEANUP(jb2DecodeSegmentHeader(info->f, info->data, *newseg));
    
    /* Paranoia check */
    if ((*newseg)->pagenum > info->npages) {
        fprintf(stderr, "segment has pagenum %ld > %d\n", (*newseg)->pagenum,
            info->npages);
    }
    
    if (info->verbose) {
        dumpheader(*newseg);
    }
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    ERR_RETURN;
}



/* Process a single segment */
ErrCode
processSegment(jb2Segment *segment, jb2info *info)
{
    Dgp_State *dgp_state = info->dgp_state;
    DGP_PROLOGUE("processSegment()", DGP_PERRSTATE);
    
    Dgp_Allocators *a = &(dgp_state->a);
    int i;
    DgpAdt_ListNode ln, retnode;
    jb2Segment *curseg;
    
    /* Dispatch on the segment type */
    DGP_ERR_GEN_CLEANUPF1((processors[segment->segtype] ==
        (ErrCode (*)(jb2Segment *, jb2info *))NULL),
        cErrNullPtr,
        "bad segment type %d\n", segment->segtype);
    
    DGP_ERR_PASS_CLEANUP((*processors[segment->segtype])(segment, info));
    
    /*
    * Handle the retention info.  If the segment is flagged as "retained",
    * put it on the active segment list.  Next, look at the segment's
    * referenced segments list; for those that are flagged as "not retained",
    * remove them from the active list.
    */
    if(info->keepAll){
            DGP_REPORTF1((info->verbose), "KeepAll, adding segment %d to active list\n",
                segment->segnum);
            DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(info->activesegs, (void *)segment, &retnode));
    } else{
        if (segment->retain) {
            DGP_REPORTF1((info->verbose), "adding segment %d to active list\n",
                segment->segnum);
            DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(info->activesegs, (void *)segment, &retnode));
        }
        for (i = 0; i < segment->nrefsegs; i++) {
            if (!segment->refretain[i]) {
                /* Slow search; I don't much care */
                DGP_REPORTF1((info->verbose),
                    "deleting segment %d from active list\n",
                    segment->refsegs[i]);
        
                DGPADT_FOREACH (ln, info->activesegs) {
                    curseg = (jb2Segment *)DGPADT_LIGET(ln);
                    if (curseg->segnum == segment->refsegs[i]) {
                        DgpAdt_liRem(ln);
                        jb2FreeSegmentInfo(curseg, dgp_state);
                        (a->free)((void *)curseg);
                        break;
                    }
                }
                DGP_ERR_GEN_CLEANUPF1((ln == DGP_NULLLISTNODE), cErrNullPtr,
                    "couldn't locate segment %d on active list\n",
                    segment->refsegs[i]);
            }
        }
        if (!segment->retain) {
                /* Can free it now */
                jb2FreeSegmentInfo(segment, dgp_state);
                (a->free)((void *)segment);
        }
    }
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    
    ERR_RETURN;
}


ErrCode
initJb2Info(
			jb2info ** info,		/* jbig2 decode state */
			Dgp_Bool verbose,		
			Dgp_UInt8* fname,		/* output file name */
			Dgp_UInt8* format,		/* output file format */
			Dgp_Bool noout,			/* do not output to a file when this flag is set */
			Dgp_Int32 numPages,		/* number of pages in this document */
            Dgp_Bool keepAll,		/* keep all segments when set,  otherwise just add the ones marked as retained */
            Dgp_Bool render,		/* only render region when this flag is set */
            ErrCode (*allocImage)( 
                   Dgp_State * dgp_state,
                   Dgp_Int32 width, Dgp_Int32 height,
                   DgpImage_Packed* dgpImage, void** usrImage),
            ErrCode (*freeImage)(Dgp_State *dgp_state, 
                    DgpImage_Packed* dgpImage),
			Dgp_State *dgp_state)
{
    DGP_PROLOGUE("initJb2Info", DGP_PERRSTATE);
    Dgp_Allocators *a = &(dgp_state->a);
	jb2info *jb2Info = NULL;

	jb2Info = (jb2info *)(a->malloc)(sizeof(*jb2Info));
	DGP_CHECK_ALLOC_CLEANUP(jb2Info);
	DGP_CHECK_NULL(info);

	jb2Info->dgp_state = dgp_state;
	jb2Info->verbose = verbose;
	jb2Info->noout = noout;
	if(noout != DGP_TRUE){
		DGP_CHECK_NULL(fname);
		DGP_CHECK_NULL(format);
		jb2Info->fname = fname;
		jb2Info->format = format;
	}else{
		jb2Info->fname = NULL;
		jb2Info->format = NULL;

	}
	/* NULL out some pointers */
	jb2Info->segs = DGP_NULLLIST;
	jb2Info->f = NULL;
	jb2Info->activesegs = DGP_NULLLIST;
	jb2Info->segs = DGP_NULLLIST;
	jb2Info->npages = numPages; 
    jb2Info->pageinfo.striped = 0;
	/* create the active list */
	DgpAdt_liNew(&(jb2Info->activesegs), jb2Info->dgp_state);
    DGP_CHECK_ALLOC_CLEANUP(jb2Info->activesegs);

	/* create the seglist */
	DgpAdt_liNew(&(jb2Info->segs), jb2Info->dgp_state);
    DGP_CHECK_ALLOC_CLEANUP(jb2Info->segs);
	jb2Info->curpage = NULL;
    jb2Info->usrPixr = NULL;
    jb2Info->dunnosize = DGP_TRUE;
    jb2Info->regions = DGP_NULLLIST;
    jb2Info->laststripe = -1;
    jb2Info->lastregion = -1;
    jb2Info->keepAll = keepAll;
    jb2Info->render = render;
    if(allocImage)
        jb2Info->allocImage = allocImage;
    else 
        jb2Info->allocImage = myAllocImage;

    if(freeImage)
        jb2Info->freeImage = freeImage;
    else
        jb2Info->freeImage = myFreeImage;

	*info = jb2Info;
	NORMAL_RETURN;
    
ERR_CLEANUP:
	if(jb2Info) cleanupJb2Info(&jb2Info);
    ERR_RETURN;
	
}

void
cleanupJb2Info(jb2info ** info)
{
    DgpAdt_ListNode ln;
    Dgp_State *dgp_state ;
    jb2info *jb2Info = *info;
    jb2Segment* seg = NULL;

	dgp_state = jb2Info->dgp_state;
    if(!dgp_state) return;

	if(jb2Info->f) jb2FreeFile(jb2Info->f);
    jb2Info->f = NULL;

    if (jb2Info->activesegs != DGP_NULLLIST) {
        DGPADT_FOREACH (ln, jb2Info->activesegs) {
            seg = (jb2Segment *)DGPADT_LIGET(ln);
            if(seg && seg->refcount <= 1){
                jb2FreeSegmentInfo(seg, dgp_state);
                (DGP_FREE)(seg);
                seg = NULL;
            }else 
                (seg->refcount)--;
        }
        DgpAdt_liFree(jb2Info->activesegs);
    }
    if (jb2Info->segs != DGP_NULLLIST) {
        DgpAdt_liFree(jb2Info->segs);
        jb2Info->segs = DGP_NULLLIST;
    }
    if(jb2Info->regions != DGP_NULLLIST){
		DGPADT_FOREACH (ln, jb2Info->regions) {
            jb2RegSegment *curreg;
            curreg = (jb2RegSegment *)DGPADT_LIGET(ln);
            if(curreg){
                if(curreg->im)
                DgpImage_free(curreg->im);
                curreg->im = NULL;
			    (DGP_FREE)(seg);
                seg = NULL;
            }
		}
		DgpAdt_liFree(jb2Info->regions);
    }

    if(jb2Info->curpage) jb2Info->freeImage(dgp_state, &jb2Info->curpage);
	DGP_FREE(jb2Info);
	*info = NULL;
}

ErrCode
processDictFromData(
					   jb2info *info, 
					   Dgp_UInt8 *dictData, 
					   Dgp_Int32 dictSize)

{
					   
    /* decode the dictionary and put it on the segment list so it won't be decoded again */
    /* First look through the header and see if this segment has already been decoded */
    Dgp_State *dgp_state = info->dgp_state;
    DGP_PROLOGUE("processSegmentFromData", DGP_PERRSTATE);
   
	/* create jb2File and data */
	jb2File f = NULL;
    DataSource datasource = NULL;
    jb2Segment *curseg = NULL;
	Dgp_Bool segfound = DGP_FALSE;
	DgpAdt_ListNode ln, retnode;
    Dgp_Int32 segtype;
    
    DGP_ERR_PASS_CLEANUP(jb2NewMemoryDataSource(dictData, dictSize, dgp_state, &datasource));    
    DGP_ERR_PASS_CLEANUP(jb2CreateFile(NULL, NULL, dgp_state, &f));   

	info->f = f;
    info->data =  datasource;

    for(;;){
    	/* readheader info */
        DGP_ERR_PASS_CLEANUP(readheader(info, &curseg));
        segtype = curseg->segtype;
        /* add new segment onto the segs list */
	    DGPADT_FOREACH (ln, info->segs) {
	      jb2Segment *listseg = (jb2Segment *)DGPADT_LIGET(ln);
	      if(listseg->segnum == curseg->segnum) {
		      segfound = DGP_TRUE;
              if(curseg->refsegs) DGP_FREE(curseg->refsegs);
              if(curseg->refretain) DGP_FREE(curseg->refretain);
              DGP_FREE((void *)curseg);
              curseg = NULL;		      
              break;
	      }
	    }
	    if(segfound != DGP_TRUE){
            DgpAdt_liAdd(info->segs, (void *)curseg, &retnode);
		    DGP_ERR_PASS_CLEANUP(processSegment(curseg, info));
	    }
   	    /* Was it the one we wanted? */ 
	    if (segtype == JB2SEGTYPE_SYMDICT) { 
	        /* Done */ 
	         break; 
	    } 
    }
    
    jb2CloseMemoryDataSource(datasource);
    jb2FreeFile(f);
    info->f = NULL;
	NORMAL_RETURN;
    
ERR_CLEANUP:
    if(f){
        jb2FreeFile(f);
        info->f= NULL;
    }
    if(curseg){
        if(curseg->refsegs) DGP_FREE(curseg->refsegs);
        if(curseg->refretain) DGP_FREE(curseg->refretain);
        DGP_FREE((void *)curseg);
    }
    ERR_RETURN;
}


ErrCode
processSegmentsFromData(
					   jb2info *info, 
					   Dgp_UInt8 *dictData, 
					   Dgp_Int32 dictSize,
                       void** usrImage)
{
					   
    /* decode the dictionary and put it on the segment list so it won't be decoded again */
    /* First look through the header and see if this segment has already been decoded */
    Dgp_State *dgp_state = info->dgp_state;
    DGP_PROLOGUE("processSegmentsFromData", DGP_PERRSTATE);
   
	/* create jb2File and data */
	jb2File f = NULL;
    DataSource datasource = NULL;
    jb2Segment *curseg = NULL;
    Dgp_Int32 val;
    Dgp_Int32 segtype;

    DGP_ERR_PASS_CLEANUP(jb2NewMemoryDataSource(dictData, dictSize, dgp_state, &datasource));    
    DGP_ERR_PASS_CLEANUP(jb2CreateFile(NULL, NULL, dgp_state, &f));   

	info->f = f;
    info->data =  datasource;
    /* Sequential file - process one segment at a time */
	for (;;) {
		/* Check that we haven't reached EOF */
		(void)dsPeek(info->data, (Dgp_Int32 *)(&val));
		if (val == EOF) {
			/* Yes - we've hit EOF */
			break;
		}
		/* readheader info */
		DGP_ERR_PASS_CLEANUP(readheader(info, &curseg));
		segtype = curseg->segtype;
		/* End early if we saw an EOF segment */
		if (curseg->segtype == JB2SEGTYPE_EOFILE) {
			/* No point in processing it */
			(DGP_FREE)((void *)curseg);
			break;
		}
		/* Add it to the list */
		DGP_ERR_PASS_CLEANUP(processSegment(curseg, info));
		if(segtype == JB2SEGTYPE_EOPAGE){
			break;			
		}
	}
    jb2CloseMemoryDataSource(datasource);
    jb2FreeFile(f);
    info->f = NULL;
    
    if(usrImage) *usrImage = info->usrPixr;

	NORMAL_RETURN;
    
ERR_CLEANUP:

    if(f){
        jb2FreeFile(f);
        info->f = NULL;
    }
	ERR_RETURN;
}

ErrCode
copySegmentList(jb2info* curInfo, jb2info* passedInfo)
{
    /* Copy segmentlinst from passinfo to curinfo */
    Dgp_State *dgp_state = curInfo->dgp_state;
    DGP_PROLOGUE("copySegmentList", DGP_PERRSTATE);

	DgpAdt_ListNode ln, retnode;
    jb2Segment *seg;


	/* Now we want to draw the list of segments into the page */
	DGPADT_FOREACH (ln, passedInfo->activesegs) {
        seg = (jb2Segment*) DGPADT_LIGET(ln);
         DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(curInfo->activesegs, (void *)seg, &retnode));
         seg->refcount++;
	}
	NORMAL_RETURN;
    
ERR_CLEANUP:

	ERR_RETURN;

}

/* Print out all the segment header information */
static void
dumpheader(jb2Segment *curseg)
{
    int j;

    fprintf(stderr,
	    "segnum %2ld:\tdata length %4d\tpagenum %2ld\tnrefsegs %2ld (",
	    curseg->segnum, (int)curseg->datalen, curseg->pagenum,
	    curseg->nrefsegs);

    for (j = 0; j < curseg->nrefsegs; j++) {
	fprintf(stderr, "%ld%c", curseg->refsegs[j],
		curseg->refretain[j] ? 'T' : 'F');
	if (j != curseg->nrefsegs - 1) {
	    fprintf(stderr, " ");
	    }
	}

    fprintf(stderr, ")\tsegtype %2d (%s)\t%c%c\n",
	    curseg->segtype, typenames[curseg->segtype],
	    curseg->retain ? 'T' : 'F',
	    curseg->defer_nonretain ? 'T' : 'F');
    }


