/*
*
* $Header: /project/icons/cvsroot/digipaper/src/libjbig2/writejbigCB.c,v 1.39 2003/02/10 22:53:27 mkrocken Exp $
*
*/

#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif

#include <stdio.h>
#include "dgp_err_control.h"
#include "dgp_utils.h"
#include "tok-lossy.h"
#include "tokpos_t.h"
#include "symbol.h"
#include "segment.h"
#include "encsymdict.h"
#include "encsymreg.h"
#include "encgenreg.h"
#include "encpageinfo.h"
#include "memorydatasink.h"
#include "writejbigCB.h"
#include "halftone.h"
#include "libutilw.pub"
#include "libpixrw.pub"
#include "libiow.pub"
#include "PsgFix.h"

#include <stdlib.h>
#include <sys/types.h>

/*#ifndef _WINDOWS */
/*#include <netinet/in.h> */
/*#endif */

#include <errno.h>
#include <limits.h>
#include "blit.h"
#include "image_defs.h"

/* internal function to free seglist and nsymlist memeory */
void freeSegDgpAdt_List( DgpAdt_List segDgpAdt_List, Dgp_State *dgp_state) ;
void freeNsymDgpAdt_List( DgpAdt_List nsymDgpAdt_List, Dgp_State *dgp_state) ;
/*
* This is a list of call back functions to be able to output jbig2 data
* (either sequential or random-access) instead of a XIFF file, by using
* the call back mechanism defined in libtclydoc/tclyout.h
* The interfaces it has to emulate are:
* - createJb2StateCB
* - closeJb2StateCB
* - freeJb2DictBlockCB
* - writeJb2DictionaryCB
* - writeJb2PageDictionaryCB
* - writeJb2PosBlockCB
*/

/* The magic Jb2 file header: M-C-W JB2 CR LF C-Z LF */
Dgp_UInt8 jb2header[] = { 0x97, 0x4a, 0x42, 0x32, 0x0d, 0x0a, 0x1a, 0x0a };

/****************************************************************
* Create a Jb2 state cookie.  initialize all the state variables.
* The given page and dictionary counts are estimates only.
* use MemoryDataSink stuff to hold the segments and segment headers
* @param stream User defined stream
* @param dgp_state     Dgp error handling cookie
* @param numPgs        Estimated number of pages
* @param numDicts      Estimated number of dictionaries (not used)
* @return _jb2State    Jbig2 state cookie contains infor needed to
*                      write out Jbig2 by using the Callback interface
* @return userData     User defined data (not used)
********************************************************************/
ErrCode 
createJb2StateCB(	uds *stream, 
				 Dgp_State *dgp_state, 
			Dgp_Int32 jbig2Profile,
				 Dgp_Int32 numPgs,
				 Dgp_Bool pdfOut,
				 void **_jb2State,
				 void **_userData)
{
	DGPERR_INIT(createJb2State);
	DataSink headersink = NULL;
	DataSink bodysink = NULL;
	jb2File jb2f = NULL;
	Dgp_Int32 i;
	Jb2State *jb2State=NULL;
	Dgp_Allocators *a = &(dgp_state->a);
	void *userData = NULL;
	jb2FileParams *params;
	
	DGP_CHECK_NULL(_jb2State);
	
	/* initialize jb2State	*/
	DGP_CHECK_ALLOC_CLEANUP(jb2State = 
		(Jb2State *) a->malloc(sizeof(Jb2State)));
	jb2State->stream = stream;
	jb2State->jbig2Profile = jbig2Profile;
	jb2State->pgnum = 0;
	jb2State->nsyms =0;
	jb2State->pgnsyms =0;
	jb2State->pdfOut = pdfOut;
	jb2State->nsymDgpAdt_List = DGP_NULLLIST;
	jb2State->segDgpAdt_List = DGP_NULLLIST;
	jb2State->pgSegDgpAdt_List = DGP_NULLLIST;
	jb2State->pgNsymDgpAdt_List = DGP_NULLLIST;
	DGP_ERR_PASS_CLEANUP(jb2NewMemoryDataSink(dgp_state, &headersink));
	DGP_ERR_PASS_CLEANUP(jb2NewMemoryDataSink(dgp_state, &bodysink));
	DGP_ERR_PASS_CLEANUP(jb2CreateFile(	headersink, 
		bodysink, 
		dgp_state, 
		&jb2f));
	params = jb2GetFileParams(jb2f);

	switch (jb2State->jbig2Profile)
	  {
	  case 1:
	/*
	* Here are the parameters for proposed ITU profile 1:
	*/
	params->sdarith = params->srarith = DGP_FALSE;      /* No arith coding */
	params->sdtables = params->srtables = DGP_FALSE;    /* No custom tables */
	/* Only the first choice is allowed */
	params->sdfirstonly = params->srfirstonly = DGP_TRUE;
	/* Only LL corner, non-transposed */
	params->srllonly = params->srnontransonly = DGP_TRUE;
	break;
	  case 3:
	/*
	* Here are the parameters for proposed ITU profile 3:
	*/
	params->sdarith = params->srarith = DGP_TRUE;       /* Arith coding */
	params->sdabparams.rtemplate = 2;                /* 10-pixel, 3-line */
	params->sdabparams.atx[0] = 2;                  /* Nominal AT */
	params->sdabparams.aty[0] = -1;
	params->sdabparams.tpon = DGP_FALSE;                /* No TP */
	
	/* Only LL corner, non-transposed */
	params->srllonly = params->srnontransonly = DGP_TRUE;

	  case 2:
	  default:
/*
	* Here are the parameters for proposed ITU profile 2:
	*/
	params->sdarith = params->srarith = DGP_FALSE;      /* No arith coding */
	params->sdtables = params->srtables = DGP_TRUE;     /* Custom tables */
	/* More than the first choice is allowed */
	params->sdfirstonly = params->srfirstonly = DGP_FALSE;
	/* Any corner, any transposed */
	params->srllonly = params->srnontransonly = DGP_FALSE;
	break;
	  }
	
	jb2SetFileParams(jb2f, params);
	
	
	/* Write out the Jb2 header to the output stream*/
	for (i = 0; i < (int)sizeof(jb2header); i++) {
		putc(jb2header[i], (FILE *)stream->fp);
	}
	/* Write 0x00: random-access organisation, page count known */
	putc(0x00, (FILE *)stream->fp);
	
	/* Write the number of pages, bigendian */
	putc((numPgs >> 24) & 0xff, (FILE *)stream->fp);
	putc((numPgs >> 16) & 0xff, (FILE *)stream->fp);
	putc((numPgs >> 8) & 0xff, (FILE *)stream->fp);
	putc(numPgs & 0xff, (FILE *)stream->fp);
	
	jb2State->jb2f = jb2f;
	*_jb2State = jb2State;
	jb2f = NULL;
	jb2State = NULL;
	*_userData = userData;
	
	NORMAL_RETURN;
	
ERR_CLEANUP:
	if(jb2State != NULL) a->free(jb2State);
	ERR_RETURN;
}
				 
/*
* Get rid of any stored state.  
* Since we put seglist inside jb2State. This function is actually
* used to free seglist information in jb2State after each section
* has been written and the state is final. dictState is not
* used in this code.
*/
void  
freeJb2DictblockCB(
				   void *_dictState,
				   void *_jb2State,
				   void *_userData
				   )
{
	Jb2State *jb2State = (Jb2State *)_jb2State;
	DgpAdt_List seglist = (DgpAdt_List)jb2State->segDgpAdt_List;
	DgpAdt_List symlist = (DgpAdt_List)jb2State->nsymDgpAdt_List;
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	Dgp_Allocators *a = &(dgp_state->a);
	void  *userData;
	
	userData = _userData;
	
	/*
	* jb2FreeSegmentInfo takes care of everything, even the symbol bits.
	* Unfortunately, we don't own those, so we've previously NULLed them
	* out.
	* Can't DgpAdt_liFree and then DgpAdt_liNew because that could fail.
	*/
	freeSegDgpAdt_List(seglist, dgp_state);
	freeNsymDgpAdt_List(symlist, dgp_state);
	jb2State->segDgpAdt_List = DGP_NULLLIST;
	jb2State->nsymDgpAdt_List = DGP_NULLLIST;
	
	/* We put everything into Jb2State 
	* DictState is just a fake structure not used
	*/
	if(_dictState != (void *)NULL) a->free(_dictState);
	userData = NULL;
}

/*****************************************************************
* Close the Jb2State and free up memory.
*
* It creates and writes out and EOF segment.
* It closes the two DataSinks, takes the buflists it gets back, 
* writes them out (headers then bodies), then frees them.
* It also free the current list of segments.
* @param _jb2State Jbig2 State cookie
* @param _userData Not used
*****************************************************************/
ErrCode 
closeJb2StateCB( 
				void *_jb2State,
				void *_userData)
{
	Jb2State *jb2State = (Jb2State *)_jb2State;
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	DGPERR_INIT(closeJb2State);
	DgpAdt_ListNode ln;
	mds_membuf *curbuf;
	DataSink headersink, bodysink;
	DgpAdt_List headerlist, bodylist;
	jb2Segment eofseg;
	Dgp_Allocators *a = &(dgp_state->a);
	uds *u = jb2State->stream;
	void *userData;
	size_t sizeHeaderData, sizeBodyData;
	
	userData = _userData;
	eofseg.pagenum = 0;
	eofseg.segtype = JB2SEGTYPE_EOFILE;
	DGP_ERR_PASS_CLEANUP( (jb2EncodeEOFile( jb2State->jb2f, 
		&eofseg))); 
	freeSegDgpAdt_List(jb2State->segDgpAdt_List,dgp_state);
	freeNsymDgpAdt_List(jb2State->nsymDgpAdt_List,dgp_state);
	freeSegDgpAdt_List(jb2State->pgSegDgpAdt_List,dgp_state);
	freeNsymDgpAdt_List(jb2State->pgNsymDgpAdt_List,dgp_state);
	jb2State->segDgpAdt_List=DGP_NULLLIST;
	jb2State->nsymDgpAdt_List=DGP_NULLLIST;
	jb2State->pgSegDgpAdt_List=DGP_NULLLIST;
	jb2State->pgNsymDgpAdt_List=DGP_NULLLIST;
	
	headersink = jb2GetFileHeaderDataSink((jb2File)jb2State->jb2f);
	bodysink = jb2GetFileDataSink((jb2File)jb2State->jb2f);
	
	DGP_ERR_PASS_CLEANUP( (jb2CloseMemoryDataSink(headersink, 
		&headerlist)));
	DGP_ERR_PASS_CLEANUP( (jb2CloseMemoryDataSink(bodysink, 
		&bodylist)));
	headersink = NULL;
	bodysink = NULL;
	a->free(jb2State->jb2f);

	/* get the size of the header in bytes */
	sizeHeaderData = 0;
	DGPADT_FOREACH (ln, headerlist) {
		curbuf = (mds_membuf *)DGPADT_LIGET(ln);
		sizeHeaderData += curbuf->bufsize;
	}
	DGPADT_FOREACH (ln, headerlist) {
		curbuf = (mds_membuf *)DGPADT_LIGET(ln);
		DGP_ERR_GEN_CLEANUPF0(u->fwrite(curbuf->buf, 
			curbuf->bufsize, 
			1, 
			u->fp) != 1,
			cErrIO_WriteError, 
			"writing header buf failed");
		a->free(curbuf->buf);
		a->free((void *)curbuf);
	}
	DgpAdt_liFree(headerlist);
	
	/* get the size of the body in bytes */
	sizeBodyData = 0;
	DGPADT_FOREACH (ln, bodylist) {
		curbuf = (mds_membuf *)DGPADT_LIGET(ln);
		sizeBodyData += curbuf->bufsize;
	}
	DGPADT_FOREACH (ln, bodylist) {
		curbuf = (mds_membuf *)DGPADT_LIGET(ln);
		DGP_ERR_GEN_CLEANUPF0(u->fwrite(curbuf->buf, 
			curbuf->bufsize, 
			1,
			u->fp) != 1, 
			cErrIO_WriteError, 
			"writing header buf failed");
		a->free(curbuf->buf);
		a->free((void *)curbuf);
	}
	DgpAdt_liFree(bodylist);
	a->free(jb2State); 
	_jb2State=NULL;
	userData = NULL;
	NORMAL_RETURN;
	
ERR_CLEANUP:
	if(jb2State !=NULL) a->free(jb2State);
	ERR_RETURN;
}

/***************************************************************
* Write out the global dictionary 
* @param toks global token array
* @param ntok number of tokens
* @param _dictState input dictState
* @param Structure to build extension
* @param number of extension
* @param lossy indication of lossy compression (DGP_TRUE)
* @param _jb2State State cookie to store jb2 segment
*              and symbol info needed to write PosBlock
* @param _userData not used
* @return _newdictState output dictState (currently not used)
**************************************************************/
ErrCode 
writeJb2DictionaryCB(
					 token *toks,
					 Dgp_Int32 ntok,
					 void *_dictState,
					 void **_newDictState,
					 DictExtension *_dictExtension,
					 Dgp_Int32 numExts,
					 void *_jb2State,
					 void *_userData)
{
	Jb2State *jb2State = (Jb2State *)_jb2State;
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	DGPERR_INIT(writeJb2DictaryCB);
	jb2Segment *newseg = (jb2Segment *)NULL;
	Dgp_UInt32 i,j,nsyms, *nsyms2=NULL;
	Dgp_Allocators *a = &(dgp_state->a);
	DgpAdt_List segDgpAdt_List= DGP_NULLLIST;
	DgpAdt_List nsymDgpAdt_List=DGP_NULLLIST;
	DgpAdt_ListNode retnode;
	jb2Symbol *syms = NULL;
	token *c;
	void *newDictState = NULL, *dictState;
	DictExtension *dictExtension;
	Dgp_Int32 numexts;
	void *userData;
	
	numexts = numExts;
	userData = _userData;
	dictState = _dictState;
	dictExtension = _dictExtension;
	
	segDgpAdt_List = jb2State->segDgpAdt_List;
	nsymDgpAdt_List = jb2State->nsymDgpAdt_List;
	
	/*
	* How many of the tokens are new?
	*/
	nsyms = 0;
	for (i = 0; i < (Dgp_UInt32)ntok; i++) {
		if (!toks[i].committed) {
			nsyms++;
		}
	}
	if (nsyms == 0) NORMAL_RETURN;
	
	/* Make a segment */
	DGP_CHECK_ALLOC_CLEANUP(
		newseg = (jb2Segment *)a->malloc(sizeof(jb2Segment)));
	
	DGP_CHECK_ALLOC_CLEANUP(
		syms = (jb2Symbol *)(a->malloc)(nsyms * sizeof(jb2Symbol)));
		/*
		* Copy the information in there.
		* The userdata field of all the symbols will point to the original
		* token.
	*/
	for (i = j = 0; i < (Dgp_UInt32)ntok; i++) {
		if (!toks[i].committed) {
			syms[j].width = toks[i].width;
			syms[j].height = toks[i].height;
			syms[j].longsperline = toks[i].longsperline;
			syms[j].bits = toks[i].bits;
			syms[j].refsym = (jb2Symbol *)NULL;
			syms[j].userdata = (void *)&toks[i];
			j++;
		}
	}
	
	jb2State->nsyms += j;
	
	newseg->segtype = JB2SEGTYPE_SYMDICT;
	newseg->pagenum = 0;    /* This is a "global" dictionary */
	newseg->nrefsegs = 0;
	newseg->refsegs = (Dgp_Int32 *)NULL;
	newseg->refretain = (Dgp_Bool *)NULL;
	newseg->data.symdict.nsymbols = nsyms;
	newseg->data.symdict.symbols = syms;
	newseg->data.symdict.direct_contexts = (Dgp_UInt8 *)NULL;
	newseg->data.symdict.ref_contexts = (Dgp_UInt8 *)NULL;
	
	/* Write it out */
	DGP_ERR_PASS_CLEANUP(jb2EncodeSymDict(	jb2State->jb2f,
		newseg, 
		(jb2Segment **)NULL,
		(ErrCode (*)(jb2Symbol **, Dgp_Int32, void *))NULL,
		(void *)NULL, 
		DGP_TRUE));
	/* Hook it onto the list of segments */
	if(segDgpAdt_List == DGP_NULLLIST){
		DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&segDgpAdt_List, dgp_state));
	}
	if(nsymDgpAdt_List == DGP_NULLLIST){
		DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&nsymDgpAdt_List, dgp_state));
	}
	DGP_ERR_PASS_CLEANUP( DgpAdt_liAdd(segDgpAdt_List,(void *)newseg, &retnode));
	
	DGP_CHECK_ALLOC_CLEANUP (nsyms2 = 
		(Dgp_UInt32 *)a->malloc( sizeof(Dgp_UInt32)));
	*nsyms2 = nsyms;
	DGP_ERR_PASS_CLEANUP( DgpAdt_liAdd(nsymDgpAdt_List,(void *)nsyms2, &retnode));
	
	/* That may have allocated segment.refsegs, segment.refretain */
	if (newseg->refsegs != (Dgp_Int32 *)NULL) {
		(a->free)((void *)newseg->refsegs);
		newseg->refsegs = NULL;
	}
	if (newseg->refretain != (Dgp_Bool *)NULL) {
		(a->free)((void *)newseg->refretain);
		newseg->refretain = NULL;
	}
	
	/*
	* Point each class at its corresponding symbol
	* Also NULL out all the symbol bitmap pointers.
	*/
	for (i = 0; i < (Dgp_UInt32)nsyms; i++) {
		c = (token *)syms[i].userdata;
		c->committed = DGP_TRUE;
		syms[i].bits = (Dgp_UInt32 *)NULL;
	}
	
	jb2State->segDgpAdt_List = segDgpAdt_List;
	jb2State->nsymDgpAdt_List = nsymDgpAdt_List;
	newseg = NULL;
	nsyms2 = NULL;
	segDgpAdt_List = DGP_NULLLIST;
	nsymDgpAdt_List = DGP_NULLLIST;
	*_newDictState = newDictState;
	userData = NULL;
	dictExtension = NULL;
	NORMAL_RETURN;
	
ERR_CLEANUP:
	if (newseg != NULL) a->free(newseg);
	if (syms != NULL) a->free(syms);
	if (nsyms2 != NULL) a->free(nsyms2);
	
	ERR_RETURN;
	
}

/************************************************* 
* Write out the per page dictionary 
* Write out the Position block segment of the page
* @param toks Per-page token array
* @param ntok number of tokens 
* @param page width
* @param page height
* @param resolution in X direction
* @param resolution in Y direction
* @param _dictState input dictState
* @param Structure to build extension
* @param number of extension
* @param lossy indication of lossy compression (DGP_TRUE)
* @param _jb2State State cookie to store jb2 segment
*              and symbol info needed to write PosBlock
* @param _userData not used
* @return _newdictState output dictState (currently not used)
******************************************************/

ErrCode
writeJb2PageInfo(		 Dgp_Int32 pgWidth,
						 Dgp_Int32 pgHeight,
						 Dgp_Int32 xRes,
						 Dgp_Int32 yRes,
						 void *_jb2State,
						 void *_userData)
{
	Jb2State *jb2State = (Jb2State *)_jb2State;
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	DGPERR_INIT(writeJb2PageInfo);
	jb2Segment pageseg;
    Dgp_Bool isColor = *((Dgp_Bool*) _userData);

	/*
	* First create page info
	* xRes[i] is x resolution in dots per inch; Jb2 wants
	* pixels/metre.  An inch is 25.4mm, so 300dpi = 300pix/0.0254m
	* = 300*(1/0.254) pixels/metre.  A value of 0 is "unknown" in both.
	*/
	pageseg.segtype = JB2SEGTYPE_PAGEINFO;
	if (jb2State->pdfOut)
	  pageseg.pagenum = 1;
	else
	  pageseg.pagenum = jb2State->pgnum + 1;
	pageseg.data.page.width = pgWidth;
	pageseg.data.page.height = pgHeight;
	pageseg.data.page.xres = (int)(xRes / 0.0254 + 0.5);
	pageseg.data.page.yres = (int)(yRes / 0.0254 + 0.5);
	pageseg.data.page.lossless = DGP_FALSE;     /* Not known to be lossless */
	pageseg.data.page.may_refine = DGP_FALSE;   /* but no refinements */
	pageseg.data.page.defpixel = 0;         /* BG is white */
	pageseg.data.page.combop = COMBOP_OR; /* Segments combined with OR */
	pageseg.data.page.may_have_auxbuf = DGP_FALSE;      /* No aux buffers */
	pageseg.data.page.may_override_combop = DGP_FALSE;  /* Always use OR */
    if (isColor)
    {
        pageseg.data.page.striped = DGP_FALSE;      /* One big blob */
	    pageseg.data.page.max_stripe = 0;       /* No striping */
    } else {
        pageseg.data.page.striped = DGP_TRUE;      /* 2 stripes */
	    pageseg.data.page.max_stripe = pgHeight - 1; 
    }
	
	DGP_ERR_PASS_CLEANUP(jb2EncodePageInfo((jb2File)jb2State->jb2f,
		&pageseg)) ;
	jb2State->pgnum++;

	NORMAL_RETURN;
ERR_CLEANUP:
    ERR_RETURN;
}
ErrCode 
writeJb2PageDictionaryCB(
    token *toks,
    Dgp_Int32 ntok,
    Dgp_Int32 pgWidth,
    Dgp_Int32 pgHeight,
    Dgp_Int32 xRes,
    Dgp_Int32 yRes,
    Dgp_Bool ascii_ext,
    void *dictState,
    void **newdictState,
    DictExtension *dictExtension,
    Dgp_Int32  numExts,
    void *_jb2State,
    void *userData)
{
	Jb2State *jb2State = (Jb2State *)_jb2State;
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	DGPERR_INIT(writeJb2PageDictionaryCB);
	jb2Segment *newseg = (jb2Segment *)NULL;
	Dgp_UInt32 i,j,nsyms, *nsyms2 = NULL;
	Dgp_Allocators *a = &(dgp_state->a);
	DgpAdt_List segDgpAdt_List,nsymDgpAdt_List;
	DgpAdt_ListNode retnode;
	jb2Symbol *syms = NULL;
	token *c;
	
        pgWidth = pgWidth;
        pgHeight= pgHeight;
        xRes = xRes;
        yRes = yRes;
        ascii_ext = ascii_ext;
       (void)dictState;
       *newdictState = NULL;
       (void)dictExtension;
       numExts = numExts;
       (void)userData;
	
	segDgpAdt_List = jb2State->pgSegDgpAdt_List;
	nsymDgpAdt_List = jb2State->pgNsymDgpAdt_List;
	
	
	/*
	* How many of the tokens are new?
	*/
	nsyms = 0;
	for (i = 0; i < (Dgp_UInt32)ntok; i++) {
        if (!toks[i].committed) {
            nsyms++;
		}
	}
	if (nsyms == 0) NORMAL_RETURN;
	
	/* Make a segment */
	DGP_CHECK_ALLOC_CLEANUP(
		newseg = (jb2Segment *)a->malloc(sizeof(jb2Segment)));
	DGP_CHECK_ALLOC_CLEANUP(
		syms = (jb2Symbol *)(a->malloc)(nsyms * sizeof(jb2Symbol)));
	
		/*
		* Copy the information in there.
		* The userdata field of all the symbols will point to the original
		* token.
	*/
	for (i = j = 0; i < (Dgp_UInt32)ntok; i++) {
		if (!toks[i].committed) {
			syms[j].width = toks[i].width;
			syms[j].height = toks[i].height;
			syms[j].longsperline = toks[i].longsperline;
			syms[j].bits = toks[i].bits;
			syms[j].refsym = (jb2Symbol *)NULL;
			syms[j].userdata = (void *)&toks[i];
			j++;
		}
	}
	
	jb2State->pgnsyms = j;
	
	newseg->segtype = JB2SEGTYPE_SYMDICT;
	if (jb2State->pdfOut)
	  newseg->pagenum = 1;    /* This is a "local" dictionary */
	else
	  newseg->pagenum = 0;
	newseg->nrefsegs = 0;
	newseg->refsegs = (Dgp_Int32 *)NULL;
	newseg->refretain = (Dgp_Bool *)NULL;
   	newseg->data.symdict.nsymbols = nsyms;
	newseg->data.symdict.symbols = syms;
	newseg->data.symdict.direct_contexts = (Dgp_UInt8 *)NULL;
	newseg->data.symdict.ref_contexts = (Dgp_UInt8 *)NULL;
	
	/* Write it out */
	DGP_ERR_PASS_CLEANUP(jb2EncodeSymDict(jb2State->jb2f,
		newseg, 
		(jb2Segment **)NULL,
		(ErrCode (*)(jb2Symbol **, Dgp_Int32, void *))NULL,
		(void *)NULL, 
		DGP_TRUE));
	/* Hook it onto the list of segments */
	if(segDgpAdt_List == DGP_NULLLIST){
		DgpAdt_liNew(&segDgpAdt_List, dgp_state);
	}
	if(nsymDgpAdt_List == DGP_NULLLIST){
		DgpAdt_liNew(&nsymDgpAdt_List, dgp_state);
	}
	DGP_CHECK_ALLOC_CLEANUP( nsyms2 = 
		(Dgp_UInt32 *)a->malloc( sizeof(Dgp_UInt32)));
	*nsyms2 = nsyms;
	DGP_ERR_PASS_CLEANUP( DgpAdt_liAdd(segDgpAdt_List, (void *)newseg, &retnode));
	DGP_ERR_PASS_CLEANUP( DgpAdt_liAdd(nsymDgpAdt_List, (void *)nsyms2, &retnode));
	
   	/* That may have allocated newseg.refsegs, newseg.refretain */
	if (newseg->refsegs != (Dgp_Int32 *)NULL) {
		(a->free)((void *)newseg->refsegs);
		newseg->refsegs = NULL;
	}
	if (newseg->refretain != (Dgp_Bool *)NULL) {
		(a->free)((void *)newseg->refretain);
		newseg->refretain = NULL;
	}
	/*
	* Point each class at its corresponding symbol
	* Also NULL out all the symbol bitmap pointers.
	*/
	for (i = 0; i < nsyms; i++) {
		c = (token *)syms[i].userdata;
		c->committed = DGP_TRUE;
		syms[i].bits = (Dgp_UInt32 *)NULL;
	}
	
	jb2State->pgSegDgpAdt_List = segDgpAdt_List;
	jb2State->pgNsymDgpAdt_List = nsymDgpAdt_List;
	segDgpAdt_List = DGP_NULLLIST;
	nsymDgpAdt_List = DGP_NULLLIST;
	newseg = NULL;
	nsyms2 = NULL;
	NORMAL_RETURN;
	
ERR_CLEANUP:
    if (newseg != NULL) a->free(newseg);
    if (syms != NULL) a->free(syms);
    if (nsyms2 != NULL) a->free(nsyms2);
    ERR_RETURN;
	
}

/*
 * An extension builder callback; all it does is copy out the array of
 * sorted symbol instances, as the callback can be called later, and we need
 * them outside this routine anyway.
 */
typedef struct {
    jb2SymbolInstance **sortedinsts;
  Dgp_State *dgp_state;
} my_instcookie;

static ErrCode
my_instextbuilder(jb2SymbolInstance **sortedinsts, Dgp_Int32 ninsts, void *extcookie)
{
    my_instcookie *c = (my_instcookie *)extcookie;
    DGP_PROLOGUE("my_instextbuilder()", c->DGP_PERRSTATE);

	Dgp_Allocators *a = &(c->DGP_ALLOCATORS);
    jb2SymbolInstance **sorted_copy = (jb2SymbolInstance **)NULL;

    sorted_copy = (jb2SymbolInstance **)(a->malloc)(ninsts * sizeof(*sorted_copy));
    DGP_CHECK_ALLOC_CLEANUP(sorted_copy);

    memcpy(sorted_copy, sortedinsts, ninsts * sizeof(*sorted_copy));
    c->sortedinsts = sorted_copy;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (sorted_copy != (jb2SymbolInstance **)NULL) {
        (a->free)((void *)sorted_copy);
        }

    ERR_RETURN;
}

/********************************************** 
* Write out the new page segment
* @param pos postion(instances) array
* @param npos number of positions (instances)
* @param page width
* @param page height
* @param resolution in X direction
* @param resolution in Y direction
* @param Structure to build extension
* @param number of extension 
* @param dictstate
* @param lossy indication of lossy compression (DGP_TRUE)
* @param _jb2State State cookie to store jb2 segment 
*		and symbol info
* @param _userData not used
**********************************************/

ErrCode 
writeJb2PosBlockCB(
				   tokpos *pos,		
				   Dgp_UInt32 npos,	/* number of jb2SymbolInstance 			*/
				   Dgp_Int32 pgWidth,
				   Dgp_Int32 pgHeight,
				   Dgp_Int32 xRes,
				   Dgp_Int32 yRes,
				   PosExtension *_exts,
				   Dgp_Int32 numExts,
				   void *_dictState,
				   Dgp_Bool lossy,
				   void *_jb2State,
				   void *_userData)
{
	Jb2State *jb2State = (Jb2State *) _jb2State;
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	DGPERR_INIT(writeJb2PosBlockCB);
	jb2Segment regseg, *curseg;
	token *curtok;
	Dgp_UInt32 nsyms,pgnsyms,*curnsyms;
	jb2Symbol *cursym;
	jb2SymbolInstance *insts = (jb2SymbolInstance *)NULL;
   	void **userdata_copy = (void **)NULL;
	void *dictState = _dictState;
	DgpAdt_List segDgpAdt_List, pgSegDgpAdt_List,nsymDgpAdt_List,pgNsymDgpAdt_List;
	DgpAdt_ListNode ln,ls;
	Dgp_UInt32 segDgpAdt_ListLen, pgSegDgpAdt_ListLen;
	Dgp_UInt32 i,j, jj, *sectionHeight = (Dgp_UInt32*) _userData;
	Dgp_Allocators *a = &(dgp_state->a);
	jb2Segment **refsegs = (jb2Segment **)NULL;
	Dgp_Int32 xres, yres, numexts;
	void *userData;
	my_instcookie instcookie;
    jb2SymbolInstance **sortedinsts = (jb2SymbolInstance **)NULL;
    tokpos **sorted_tokpos = (tokpos **)NULL;
    
    /* 
	*Variable intialization
	*/
	xres = xRes;
	yres = yRes;
	numexts = numExts;
	userData = _userData;
	
	pgnsyms	= jb2State->pgnsyms;
	nsyms	= jb2State->nsyms;
  if (jb2State->segDgpAdt_List == DGP_NULLLIST)
  {
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew (&jb2State->segDgpAdt_List, dgp_state));
  }
  if (jb2State->pgSegDgpAdt_List == DGP_NULLLIST)
  {
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew (&jb2State->pgSegDgpAdt_List, dgp_state));
  }
  if (jb2State->nsymDgpAdt_List == DGP_NULLLIST)
  {
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew (&jb2State->nsymDgpAdt_List, dgp_state));
  }
  if (jb2State->pgNsymDgpAdt_List == DGP_NULLLIST)
  {
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew (&jb2State->pgNsymDgpAdt_List, dgp_state));
  }
	segDgpAdt_List = jb2State->segDgpAdt_List;
    pgSegDgpAdt_List = jb2State->pgSegDgpAdt_List;
	nsymDgpAdt_List = jb2State->nsymDgpAdt_List;
	pgNsymDgpAdt_List = jb2State->pgNsymDgpAdt_List;
	segDgpAdt_ListLen = DGPADT_LILEN(segDgpAdt_List);
	pgSegDgpAdt_ListLen = DGPADT_LILEN(pgSegDgpAdt_List);
	
	DGP_CHECK_ALLOC_CLEANUP( userdata_copy = 
		(void **)(a->malloc)((nsyms + pgnsyms) * 
		sizeof(*userdata_copy)));
	
	/* Read out the old userdata and write the new one. */
	for (	j = 0, ln = DGPADT_LIFIRST(segDgpAdt_List), ls = DGPADT_LIFIRST(nsymDgpAdt_List); 
	j < segDgpAdt_ListLen;
	j++, ln = DGPADT_LINEXT(ln), ls = DGPADT_LINEXT(ls)) {
		curseg = (jb2Segment *)DGPADT_LIGET(ln);
		curnsyms = (Dgp_UInt32 *) DGPADT_LIGET(ls);
		for (i=0; i < *curnsyms; i++){
			token *curtok = (token *)curseg->data.symdict.symbols[i].userdata;
			userdata_copy[i+j*segDgpAdt_ListLen] = curtok->extdata;
			curtok->extdata = &curseg->data.symdict.symbols[i];

			if(curtok->extdata == NULL){
				printf("uhoh");
			}
		}
	}
	for (	jj = 0, ln = DGPADT_LIFIRST(pgSegDgpAdt_List), ls = DGPADT_LIFIRST(pgNsymDgpAdt_List); 
	jj < pgSegDgpAdt_ListLen;
	jj++, ln = DGPADT_LINEXT(ln), ls = DGPADT_LINEXT(ls)) {
		curseg = (jb2Segment *)DGPADT_LIGET(ln);
		curnsyms = (Dgp_UInt32 *) DGPADT_LIGET(ls);
		for (i=0; i< *curnsyms; i++){
			token *curtok = (token *)curseg->data.symdict.symbols[i].userdata;
			userdata_copy[j + jj*pgSegDgpAdt_ListLen + i] = curtok->extdata;
			curtok->extdata = &curseg->data.symdict.symbols[i];
			
			if(curtok->extdata == NULL){
				printf("uhoh");
			}
		}
	}
	
	/* Allocate the jb2SymbolInstance structures */
	DGP_CHECK_ALLOC_CLEANUP ( insts = 
		(jb2SymbolInstance *)(a->malloc)(npos * sizeof(*insts)));
		/*
		* Now we can make up the jb2SymbolInstances.
		* The Coordinate has already be adjust to UL convention
	*/
	for (i = 0; i < npos; i++) {
		curtok = pos[i].tok;
		cursym = (jb2Symbol *)curtok->extdata;
		insts[i].xpos = pos[i].xpos;
		insts[i].ypos = pos[i].ypos - cursym->height + 1;
		insts[i].sym = cursym;
		insts[i].refdata = (jb2SymbolInstanceRefined *)NULL;
		insts[i].userdata = (void *)&pos[i];
	}
	
	/* Allocate the list of referenced segment numbers */
	/* It refers to all segments on the current dict list */
	
	regseg.nrefsegs = segDgpAdt_ListLen + pgSegDgpAdt_ListLen;
	
	DGP_CHECK_ALLOC(regseg.refsegs = 
		(Dgp_Int32 *)DGP_MALLOC(regseg.nrefsegs * sizeof(*regseg.refsegs)));
	
	/* and the array of referenced segments */
	DGP_CHECK_ALLOC(refsegs = 
		(jb2Segment **)DGP_MALLOC(regseg.nrefsegs * sizeof(*refsegs)));
	
	/* and retention bits */
	DGP_CHECK_ALLOC(regseg.refretain = 
		(Dgp_Bool *)DGP_MALLOC(regseg.nrefsegs * sizeof(*regseg.refretain)));
	
		/*
		* We'd like to tell the decoder if the global dictionaries are
		* unretained.  Fortunately, that's exactly what lastUse says.
	*/
	for (j = 0, ln = DGPADT_LIFIRST(segDgpAdt_List); j < segDgpAdt_ListLen;
	j++, ln = DGPADT_LINEXT(ln)) {
		curseg = (jb2Segment *)DGPADT_LIGET(ln);
		/* Global dictionaries are retained */
		regseg.refretain[j] = DGP_TRUE;
		/* Put in the segment number */
		regseg.refsegs[j] = curseg->segnum;
		/* and copy it into refsegs */
		refsegs[j] = curseg;
	}
	for (i = 0, ln = DGPADT_LIFIRST(pgSegDgpAdt_List); i < pgSegDgpAdt_ListLen;
	i++, ln = DGPADT_LINEXT(ln)) {
		curseg = (jb2Segment *)DGPADT_LIGET(ln);
		/* Page dictionaries are not retained */
		if (lossy)
		{
		regseg.refretain[j+i] = DGP_FALSE;
		}
		else
		{
		regseg.refretain[j+i] = DGP_TRUE;
		}
		regseg.refsegs[j+i] = curseg->segnum;
		refsegs[j+i] = curseg;
	}
	
	/* Next, the region segment itself */
	regseg.segtype = JB2SEGTYPE_SYMREG_IMM;
	if (jb2State->pdfOut)
	  regseg.pagenum = 1; 
	else
	  regseg.pagenum = numExts;
	regseg.data.symreg.ninstances = npos;
	regseg.data.symreg.instances = insts;
	regseg.data.symreg.int_combop = COMBOP_OR;
	regseg.data.symreg.defpixel = 0;
	regseg.reginfo.regwidth = pgWidth;
	regseg.reginfo.regheight = pgHeight;
	regseg.reginfo.regxpos = 0;
    if (sectionHeight)
	regseg.reginfo.regypos = *sectionHeight;
    else
	regseg.reginfo.regypos = 0;
	regseg.reginfo.combop = COMBOP_OR;
	
	/* Get an array of the referenced segments */
	
	/* Write it out
	 * If we don't have color extensions we can improve compression by
	 * sorting the positions */
    instcookie.dgp_state = dgp_state;
	if (npos > 0) DGP_ERR_PASS_CLEANUP(jb2EncodeSymRegion( 
						jb2State->jb2f,
						&regseg, 
						refsegs,
						my_instextbuilder,
						&instcookie,
						DGP_FALSE /* order preserved */)) ;


    /* Call their callback, if they gave us one.  */
    sortedinsts = instcookie.sortedinsts;
    if(_exts!= NULL && npos > 0){
        if (_exts->builder != (ErrCode (*)(tokpos **, Dgp_Int32,
                                       void *, void **, size_t *, Dgp_State *))NULL)
        {
            /* They gave us a callback - call it */
            sorted_tokpos =
                (tokpos **)(a->malloc)(npos * sizeof(*sorted_tokpos));
            DGP_CHECK_ALLOC_CLEANUP(sorted_tokpos);

            /* Copy out the stuff in the symbols' userdata fields */
            for (i = 0; i < npos; i++) {
                sorted_tokpos[i] = (tokpos *)sortedinsts[i]->userdata;
                }

            DGP_ERR_PASS_CLEANUP((*(_exts->builder))(sorted_tokpos, npos, _exts->cookie,
                                               _exts->extdata, &(_exts->extlen), dgp_state));

            (a->free)((void *)sorted_tokpos);
            sorted_tokpos = (tokpos **)NULL;
        }
    }
	/* Put the userdata values back */
	
	for (	j = 0, ln = DGPADT_LIFIRST(segDgpAdt_List), ls = DGPADT_LIFIRST(nsymDgpAdt_List); 
	j < segDgpAdt_ListLen;
	j++, ln = DGPADT_LINEXT(ln), ls = DGPADT_LINEXT(ls)) {
		curseg = (jb2Segment *)DGPADT_LIGET(ln);
		curnsyms = (Dgp_UInt32 *) DGPADT_LIGET(ls);
		for (i=0; i< *curnsyms; i++){
			curtok = (token *)curseg->data.symdict.symbols[i].userdata;
			curtok->extdata = userdata_copy[j*segDgpAdt_ListLen+i];
		}
	}

	for (	jj = 0, ln = DGPADT_LIFIRST(pgSegDgpAdt_List), ls = DGPADT_LIFIRST(pgNsymDgpAdt_List); 
	jj < pgSegDgpAdt_ListLen;
	jj++, ln = DGPADT_LINEXT(ln), ls = DGPADT_LINEXT(ls)) {
		curseg = (jb2Segment *)DGPADT_LIGET(ln);
		curnsyms = (Dgp_UInt32 *) DGPADT_LIGET(ls);
		for (i=0; i< *curnsyms; i++){
			curtok = (token *)curseg->data.symdict.symbols[i].userdata;
			curtok->extdata = userdata_copy[j + jj*pgSegDgpAdt_ListLen+i];
		}
	}

	/* Free the perpage dictionary segment	*/
	if (lossy)
	{
	freeSegDgpAdt_List(pgSegDgpAdt_List, dgp_state);
	freeNsymDgpAdt_List(pgNsymDgpAdt_List, dgp_state);
	jb2State->pgSegDgpAdt_List = DGP_NULLLIST;
	jb2State->pgNsymDgpAdt_List = DGP_NULLLIST;
	}

	/* Don't need refsegs any more */
	a->free((void *)refsegs);
	refsegs = (jb2Segment **)NULL;
    /* or regseg's arrays */
    a->free((void *)regseg.refsegs);
	regseg.refsegs = NULL;
    a->free((void *)regseg.refretain);
	regseg.refretain = NULL;
	a->free((void **)userdata_copy);
	a->free((void *)insts);
	insts = (jb2SymbolInstance *)NULL;
	
	/* 20.4.2001 / dhtesq. */
	/* Sortedinsts are leaking.  I don't know if this is the appropriate place to do this, */
	/* but as far as I can tell all references to this pointer go out of scope with this */
	/* function. */
	if (npos > 0) DGP_FREE(sortedinsts);

	dictState = NULL;
	userData = NULL;
	
    NORMAL_RETURN;
	
ERR_CLEANUP:
	if(refsegs) free((void *)refsegs);
	if(insts) free((jb2SymbolInstance *) insts) ;
	
    ERR_RETURN;
}

/*********************************************************
* Encode the endof page segment
* @param _jb2State cookie state
* @param userData not used
********************************************************/
ErrCode  
finishJb2PageCB(  
				void *_jb2State, 
				void *_userData)
{
	Jb2State *jb2State = (Jb2State *) _jb2State;
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	DGPERR_INIT(finishJb2Page);
	jb2Segment eopageseg;
	void * userData;
	
	userData = _userData;
	if (jb2State->pdfOut == DGP_FALSE)
	  {

	    /* One more segment: EOPage */
	    eopageseg.segtype = JB2SEGTYPE_EOPAGE;
	    eopageseg.pagenum = jb2State->pgnum;
	    DGP_ERR_PASS_CLEANUP(jb2EncodeEOPage(jb2State->jb2f, 
						 &eopageseg));
	  }

    NORMAL_RETURN;
	
ERR_CLEANUP:
    ERR_RETURN;
}

ErrCode 
writeJb2EOStripe(Dgp_Int32 line_number,
			Dgp_Int32 page_number,
		    	void *_jb2state)
{
	Jb2State *jb2State = (Jb2State *) _jb2state;
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	DGPERR_INIT(writeJb2EOStrip);
	jb2Segment eostripeseg;

				/* One more segment: eostripe */
	eostripeseg.data.eostripe.last_row = line_number;
	eostripeseg.segtype = JB2SEGTYPE_EOSTRIPE;
	if (jb2State->pdfOut)
	  eostripeseg.pagenum = 1; 
	else
	  eostripeseg.pagenum = page_number; 
	DGP_ERR_PASS_CLEANUP(jb2EncodeEOStripe(jb2State->jb2f, 
		&eostripeseg));
    NORMAL_RETURN;
	
ERR_CLEANUP:
    ERR_RETURN;
}

void
freeSegDgpAdt_List(
				   DgpAdt_List segDgpAdt_List, 
				   Dgp_State *dgp_state) 
{
	jb2Segment *curseg;
	Dgp_Allocators *a = &(dgp_state->a);
	DgpAdt_ListNode ln;
	if(segDgpAdt_List != DGP_NULLLIST){
		DGPADT_FOREACH (ln, segDgpAdt_List) {
			curseg = (jb2Segment *)DGPADT_LIGET(ln);
			jb2FreeSegmentInfo(curseg,dgp_state);
			a->free((void *)curseg);
		}
		DgpAdt_liFree(segDgpAdt_List);
	}
}

void
freeNsymDgpAdt_List(
					DgpAdt_List nsymDgpAdt_List, 
					Dgp_State *dgp_state) 
{
	Dgp_UInt32 *nsym;
	Dgp_Allocators *a = &(dgp_state->a);
	DgpAdt_ListNode ln;
	if(nsymDgpAdt_List != NULL){
		DGPADT_FOREACH (ln, nsymDgpAdt_List) {
			nsym = (Dgp_UInt32 *)DGPADT_LIGET(ln);
			a->free(nsym);
		}
		DgpAdt_liFree(nsymDgpAdt_List);
	}
}

ErrCode 
writeJb2BinaryPixrCB(DgpAdt_List imPos,
					 Dgp_Int32 layer,
					 Dgp_Int32 section,
					 DgpAdt_List imageList,
					 void *_jb2State,
					 void *_userData)
{
	Jb2State *jb2State = (Jb2State *) _jb2State;
	jb2FileParams *params = jb2GetFileParams ((jb2File) jb2State->jb2f);
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	DGPERR_INIT(writeJb2BinaryPixrCB);
	jb2Segment regseg;
	void *userData;
	Dgp_Int32 bpl;
	IPWState *ipwstate = DGP_PIPWSTATE;
	UInt8 *w_image = NULL;
	DgpImage_Packed resid = NULL;
	Dgp_Int32 layerr, longsperline, i;
	DgpAdt_ListNode ln, ln_pos;
	PIXR* readPixr;
	pHALFTONEPIC ppic;

	
	userData = _userData;
	layerr = layer;
	
	ln_pos = DGPADT_LIFIRST (imPos);

	ln = DGPADT_LIFIRST (imageList);
	while (ln != NULL && ln_pos != NULL) {
 
	readPixr = (PIXR*) ln->userdata;		
	ppic = (pHALFTONEPIC)DGPADT_LIGET(ln_pos);
	
	if (ppic->section == section)
	{
/*	w_writeTiff (ipwstate, readPixr, "halftonePixr.tif", COMPRESSION_NONE, cTrue, 0); */

	DGP_ERR_PASS_CLEANUP( w_pixrGetBpl(	ipwstate, 
		readPixr, 
		(Int32 *)&bpl));
	DGP_ERR_PASS_CLEANUP(w_pixrGetImage(ipwstate, readPixr, &w_image));

	longsperline = bpl / 4;
	DGP_ERR_PASS_CLEANUP (DgpImage_wrap(dgp_state, (Dgp_UInt32 *)w_image, 
		ppic->w, ppic->h, longsperline, &resid)); 
/*	DgpImage_save (resid, "halftoneDgpImage.pbm"); */
	DGP_ERR_GEN_CLEANUP( resid == NULL, cErrNomem);
	/* Fill in the segment */
	regseg.segtype = JB2SEGTYPE_GENREG_IMM;
	if (jb2State->pdfOut)
	  regseg.pagenum = 1; 
	else
	  regseg.pagenum = jb2State->pgnum;
	regseg.nrefsegs = 0;
	regseg.refsegs = (Dgp_Int32 *)NULL;
	regseg.refretain = (Dgp_Bool *)NULL;
	regseg.reginfo.regwidth = ppic->w;
	regseg.reginfo.regheight = ppic->h;
	regseg.reginfo.regxpos = ppic->x;
	regseg.reginfo.regypos = ppic->y;
	regseg.reginfo.combop = COMBOP_OR;
	regseg.reginfo.rendered = DGP_TRUE;
	regseg.reginfo.im = resid;
	if (params->sdarith)
	{
		/* arithmetic coding */
	regseg.data.genreg.mmr = DGP_FALSE;
	regseg.data.genreg.abparams.rtemplate = params->sdabparams.rtemplate;

    /* set the segment's AT position according to halftone cell size */
    if(ppic->cell_size >0){
	    params->sdabparams.atx[0] = -(ppic->cell_size *2);
	    params->sdabparams.aty[0] = 0;

    }


	for (i = 0; i < 3; i++)
	{
	regseg.data.genreg.abparams.atx[i] = params->sdabparams.atx[i];
	regseg.data.genreg.abparams.aty[i] = params->sdabparams.aty[i];
	}
	regseg.data.genreg.abparams.tpon = params->sdabparams.tpon;
	}
	else
	{
		/* Huffman coding */
	regseg.data.genreg.mmr = DGP_TRUE;
	}
	regseg.data.genreg.data = (Dgp_UInt8 *)NULL;
	regseg.data.genreg.datalen = 0;

	DGP_ERR_PASS_CLEANUP(jb2EncodeGenRegion((jb2File)jb2State->jb2f, 
		&regseg));

	DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipwstate, &readPixr));
	DGP_FREE (ppic);

	DgpAdt_liRemAndNext (ln, &ln);
	DgpAdt_liRemAndNext (ln_pos, &ln_pos);

	}
	else
	{
	ln = ln->next;
	ln_pos = ln_pos->next;
	}
	if( resid != NULL) DgpImage_unwrap(resid, (Dgp_UInt32**)&w_image);
	resid = NULL;	
	}
	
	/*	pimUnwrap(resid); */
	userData = NULL;
	NORMAL_RETURN;
	
ERR_CLEANUP:
	if( resid != NULL) DgpImage_unwrap(resid, (Dgp_UInt32**)&w_image);
    ERR_RETURN;
	
}

/* write one large cc into one jbig2 generic region
 * called by libtclydoc
 */
ErrCode 
writeJb2BinaryCC(    DgpImage_Packed resid,
					 Dgp_Int32 layer,
					 Dgp_Int32 x,
                     Dgp_Int32 y,
					 void *_jb2State,
					 void *_userData)
{
	Jb2State *jb2State = (Jb2State *) _jb2State;
	jb2FileParams *params = jb2GetFileParams ((jb2File) jb2State->jb2f);
	Dgp_State *dgp_state = jb2GetFileDgp_State(jb2State->jb2f);
	DGPERR_INIT(writeJb2BinaryCC);
	jb2Segment regseg;
	void *userData;
	Dgp_Int32 layerr, i;
	
	userData = _userData;
	layerr = layer;
	
    DGP_ERR_GEN_CLEANUP( resid == NULL, cErrNomem);

    /* Fill in the segment */
    regseg.segtype = JB2SEGTYPE_GENREG_IMM;
    if (jb2State->pdfOut)
      regseg.pagenum = 1;
    else
      regseg.pagenum = jb2State->pgnum;
    regseg.nrefsegs = 0;
    regseg.refsegs = (Dgp_Int32 *)NULL;
    regseg.refretain = (Dgp_Bool *)NULL;
    regseg.reginfo.regwidth = DGPIMG_WIDTH(resid);
    regseg.reginfo.regheight = DGPIMG_HEIGHT(resid);
    regseg.reginfo.regxpos = x;
    regseg.reginfo.regypos = y;
    regseg.reginfo.combop = COMBOP_OR;
    regseg.reginfo.rendered = DGP_TRUE;
    regseg.reginfo.im = resid;
    if (params->sdarith)
    {
	    /* arithmetic coding */
	    regseg.data.genreg.mmr = DGP_FALSE;
	    regseg.data.genreg.abparams.rtemplate = params->sdabparams.rtemplate;

        /* set the segment's AT position according to halftone cell size */
	    for (i = 0; i < 3; i++)
	    {
	        regseg.data.genreg.abparams.atx[i] = params->sdabparams.atx[i];
	        regseg.data.genreg.abparams.aty[i] = params->sdabparams.aty[i];
	    }
	    regseg.data.genreg.abparams.tpon = params->sdabparams.tpon;
    }else{
		    /* Huffman coding */
	    regseg.data.genreg.mmr = DGP_TRUE;
    }
    regseg.data.genreg.data = (Dgp_UInt8 *)NULL;
    regseg.data.genreg.datalen = 0;

    DGP_ERR_PASS_CLEANUP(jb2EncodeGenRegion((jb2File)jb2State->jb2f, &regseg));   
    resid = NULL;
    userData = NULL;
    NORMAL_RETURN;
	
ERR_CLEANUP:
    ERR_RETURN;
	
}

ErrCode 
writeJb2AnnotationsCB(	Dgp_Int32 num_annotations, 
			Annotation *annotations, 
			void *_jb2state,
			void *userData)
{
  Jb2State *jb2State = (Jb2State *) _jb2state;
  jb2File file = jb2State->jb2f;
  Dgp_State *dgp_state = jb2GetFileDgp_State(file);
  DGPERR_INIT(writeJb2AnnotationsCB);
  DataSink bodyds;

  jb2Segment extseg;
  Dgp_UInt8  lbuf[4];
  size_t bytes;
  Dgp_Int32 i;

  userData = userData;

  /* Fill in the fields as promised */ 
  if (jb2State->pdfOut)
    extseg.pagenum = 1;
  else
    extseg.pagenum = jb2State->pgnum;
  extseg.segtype = JB2SEGTYPE_EXTENSION;
  extseg.nrefsegs = 0; 
  extseg.refsegs = (Dgp_Int32 *)NULL; 
  extseg.refretain = (Dgp_Bool *)NULL; 
  extseg.retain = DGP_FALSE; 
  extseg.defer_nonretain = DGP_FALSE; 

  /* loop over all annotations */
  for (i = 0; i < num_annotations; i++)
    {

      /* create unique segment number */
      extseg.segnum = jb2FileNextSegmentNumber(file);  
      /* All but the length is known - it's OK to encode the header */ 
      DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeader(file, &extseg)); 
 
      bodyds = jb2GetFileDataSink(file); 
      /* Reset the count on bodyds so we can get the data length */ 
      dsResetCount(bodyds); 

      /* write the header word identifiing the type of comment */
      to4byte(lbuf, annotations[i].tag);
      DSWRITE(bodyds, 4, lbuf, &bytes); 
  
      /* fill the data into the segment */
      DSWRITE(bodyds, (size_t)annotations[i].size, annotations[i].data, &bytes); 
      
      /* Fix up the data length in the header */ 
      extseg.datalen = dsCount(bodyds); 
      DGP_ERR_PASS_CLEANUP(jb2EncodeSegmentHeaderDataLength(file, &extseg)); 
    } 

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

