
/**************************************************************
 * Call back interface for create Jb2 output
 * $Header: /project/icons/cvsroot/digipaper/src/libjbig2/writejbigCB.h,v 1.15 2002/10/11 21:42:23 jlin Exp $
 **************************************************************/


#ifndef WRITE_Jb2CB_H
#define WRITE_Jb2CB_H


#include "dgp_err_control.h"
#include "dgp_state.h"
#include "list.h"
#include "jb2file.h"
#include "uds.h"
#include "tokpos_t.h"
#include "iostructures.h"
#include <stdio.h>
#include <mmcompress.h>
#include <libipw.pub>
#include <dgp_allocators.h>
#include <annotation.h>

/* Our extension type for storing unicode and bounding boxes of characters */
#define DGP_CBOX_EXT 0x00584552

#ifdef __cplusplus
extern "C"
{
#endif

/******************************************************************
 * It has:
 *      ntok - number of symbols the state currently holds
 *      capacity - number of symbols the state can hold
 *      toks - a pointer into another part of the block, to a set of
 *             jb2Symbol structures
 *      segnum - the next segment number to be assigned
 *      pagenum - the next page number to be assigned
 *
 */

typedef struct{
    uds *stream;
    jb2File jb2f;
    Dgp_Int32 jbig2Profile;
    DgpAdt_List segDgpAdt_List;
    DgpAdt_List pgSegDgpAdt_List;
    DgpAdt_List nsymDgpAdt_List;
    DgpAdt_List pgNsymDgpAdt_List;
    Dgp_UInt32 nsyms;
    Dgp_UInt32 pgnsyms;
    Dgp_Int32 pgnum;
    Dgp_Bool pdfOut;
}Jb2State;


/****************************************************************
 * Create a Jb2 state cookie.  initialize all the state variables.
 * The given page and dictionary counts are estimates only.
 * use MemoryDataSink stuff to hold the segments and segment headers 
 * @param stream User defined stream
 * @param dgp_state 	Dgp error handling cookie
 * @param numPgs 	Estimated number of pages
 * @param numDicts 	Estimated number of dictionaries (not used)
 * @return _jb2State	Jbig2 state cookie contains infor needed to
 *			write out Jbig2 by using the Callback interface
 * @return userData 	User defined data (not used)
 ********************************************************************/
ErrCode 
createJb2StateCB(         uds *stream,
                        Dgp_State *dgp_state,
			  Dgp_Int32 jbig2Profile,
                        Dgp_Int32 numPgs,
                        Dgp_Bool pdfOut,
                        void **_jb2State,
                        void **_userData);

/*****************************************************************
 * Close the Jb2State and free up memory.
 *
 * It creates and writes out and EOF segment.
 * It closes the two DataSinks, takes the buflists it gets back,
 * writes them out (headers then bodies), then frees them.
 * It also free the current list of segments.
 * @param _jb2State Jbig2 State cookie
 * @param _userData Not used
 *****************************************************************/
ErrCode 
closeJb2StateCB(
        void *_jb2State,
        void *_userData);


/*********************************************************
 * Encode the endof page segment
 * @param _jb2State cookie state
 * @param userData not used
 ********************************************************/
ErrCode  
finishJb2PageCB(
        void *_jb2State,
        void *_userData);


ErrCode 
writeJb2AnnotationsCB(	Dgp_Int32 num_annotations, 
			Annotation *annotations, 
			void *_jb2state,
			void *userData);

/**********************************************
 * Write out the Position block segment of the page
 * @param pos postion(instances) array
 * @param npos number of positions (instances)
 * @param page width
 * @param page height
 * @param resolution in X direction
 * @param resolution in Y direction
 * @param Structure to build extension
 * @param number of extension
 * @param dictstate
 * @param lossy indication of lossy compression (TRUE)
 * @param _jb2State State cookie to store jb2 segment
 *              and symbol info
 * @param _userData not used
 **********************************************/

ErrCode 
writeJb2PosBlockCB(
       tokpos *pos,
       unsigned long npos,      /* number of jb2SymbolInstance                     */
       Dgp_Int32 pgWidth,
       Dgp_Int32 pgHeight,
       Dgp_Int32 xRes,
       Dgp_Int32 yRes,
       PosExtension *_exts,
       Dgp_Int32 numExts,
       void *_dictState,
       Dgp_Bool lossy,
       void *_jb2State,
       void *_userData);


/***************************************************************
 * Write out the global dictionary
 * @param toks global token array
 * @param ntok number of tokens
 * @param _dictState input dictState
 * @param Structure to build extension
 * @param number of extension
 * @param lossy indication of lossy compression (TRUE)
 * @param _jb2State State cookie to store jb2 segment
 *              and symbol info needed to write PosBlock
 * @param _userData not used
 * @return _newdictState output dictState (currently not used)
 **************************************************************/
ErrCode 
writeJb2DictionaryCB(
        token *toks,
        Dgp_Int32 ntok,
        void *_dictState,
        void **_newDictState,
        DictExtension *_dictExtension,
        Dgp_Int32 numExts,
        void *_jb2State,
        void *_userData);

/*************************************************
 * Write out the per page dictionary
 * Write out the new page segment
 * @param toks Per-page token array
 * @param ntok number of tokens
 * @param page width
 * @param page height
 * @param resolution in X direction
 * @param resolution in Y direction
 * @param _dictState input dictState
 * @param Structure to build extension
 * @param number of extension
 * @param lossy indication of lossy compression (TRUE)
 * @param _jb2State State cookie to store jb2 segment
 *              and symbol info needed to write PosBlock
 * @param _userData not used
 * @return _newdictState output dictState (currently not used)
 ******************************************************/
ErrCode 
writeJb2PageDictionaryCB(
    token *toks,
    Dgp_Int32 ntok,
    Dgp_Int32 pgWidth,
    Dgp_Int32 pgHeight,
    Dgp_Int32 xRes,
    Dgp_Int32 yRes,
    Dgp_Bool ascii_ext,
    void *_dictState,
    void **_newdictState,
    DictExtension *_dictExtension,
    Dgp_Int32  numExts,
    void *_jb2State,
    void *_userData);


ErrCode 
writeJb2BinaryPixrCB(	DgpAdt_List ImPos,
		    	Dgp_Int32 layer, 
				Dgp_Int32 section,
			DgpAdt_List image, 
		    	void *_jb2state,
			void *userData);

ErrCode 
writeJb2EOStripe(Dgp_Int32 line_number,
			Dgp_Int32 page_number,
		    	void *_jb2state);

ErrCode
writeJb2PageInfo(		 Dgp_Int32 pgWidth,
						 Dgp_Int32 pgHeight,
						 Dgp_Int32 xRes,
						 Dgp_Int32 yRes,
						 void *_jb2State,
						 void *_userData);

ErrCode 
writeJb2BinaryCC(    DgpImage_Packed resid,
					 Dgp_Int32 layer,
					 Dgp_Int32 x,
                     Dgp_Int32 y,
					 void *_jb2State,
					 void *_userData);


/*
 * Get rid of any stored state.
 */
void 
freeJb2DictblockCB(	void *_dictblock, 
			void *_jb2state,
			void *_userData);

#ifdef __cplusplus
}
#endif

#endif


/************************************************
 * Running Log
 * $Log: writejbigCB.h,v $
 * Revision 1.15  2002/10/11 21:42:23  jlin
 * remove gcc compile warnings
 *
 * Revision 1.14  2002/05/29 17:36:42  mkrocken
 * add JBIG2 in PDF as a command line option
 *
 *  Modified Files:
 *  	writejbigCB.c writejbigCB.h
 *
 * Revision 1.13  2001/12/19 00:51:18  mkrocken
 * changes for writing 2 stripe JBIG2
 *
 *  Modified Files:
 *  	writejbigCB.c writejbigCB.h
 *
 * Revision 1.12  2001/08/17 18:28:11  jlin
 * add writeJb2BinaryCC, fix memeroy leak in writeJb2BinaryPixrCB
 *
 * Revision 1.11  2001/06/19 22:26:40  mkrocken
 * fixed the generic region encoder interface to be compliant with the new
 * halftone detection scheme. Also enabled arithmetic encoding.
 *
 * Revision 1.10  2001/05/29 22:23:57  mkrocken
 * changes to make the profile selectable
 *
 * Revision 1.9  2001/05/08 00:42:55  dtracy
 * allocators.h -> dgp_allocators.h throughout code base.
 *
 * Revision 1.8  2001/04/10 23:46:27  dtracy
 * Cleared xiff3 dependencies from codebase. Jpeg-ls tweaks.
 *
 * Revision 1.7  2000/12/29 18:41:02  mkrocken
 * major changes to decoder. Reworked supercache to use JBIG2 data structures
 *
 * Revision 1.6  2000/11/08 18:18:49  mkrocken
 * adding Dillon's edits to reduce compiler warnings to the head
 *
 * Revision 1.5.2.1  2000/11/06 23:26:01  dtracy
 * No-warnings checkin for xif branch; almost frozen.
 *
 * Revision 1.5  2000/11/03 01:57:33  mkrocken
 * changed to a private extension tag for the compressed unicode and bounding
 * boxes, called DGP_CBOX_EXT. Its value is set to 0x00584552
 *
 * Revision 1.4  2000/05/16 23:25:26  mkrocken
 * initial rev
 *
 * Revision 1.3  2000/03/22 18:10:02  mkrocken
 * missed some booleans, now changed to Dgp_Bool
 *
 * Revision 1.2  2000/03/21 20:01:12  mkrocken
 *
 * changes to use jbig2 callbacks
 * removed some warnings
 *
 * Revision 1.1  2000/03/20 22:34:38  mkrocken
 *
 * initial revision, currently incompatible with rest of the code
 *
 * Revision 1.1  2000/02/21  22:29:10  jlin
 * Initial version of the jbig2 call back functions to hook into tcomp-lossy
 *
 *
 ***********************************************/
