/*************************************************************************
**  Copyright 1999-2000 Xerox Corporation.All rights reserved.
**
**  DigiPaper Codebase
**    
**  Ranjani Ramamurthy, April 2000
**
**  Filename: jpeg_comp.c
**
**  Description of what's implemented in this file:
**
**
**************************************************************************/

#include "jinclude.h"
#include "jpeg6blib.h"
#include "jerror.h"

typedef struct {
  struct jpeg_destination_mgr pub; /* public fields  */

  JOCTET*   buffer;		/* start of temporary buffer */

  JOCTET*   output_buffer;
  int       output_length;
  int       output_offset;

} DGP_JpegCompDstMgr;

#define DGP_BUFWRITE_SZ			4096
/*
 * Initialize destination --- called by jpeg_start_compress
 * before any data is actually written.
 */

GLOBAL(void)
DGP_BufWrite(JOCTET* to_buffer, const JOCTET* from_buffer,
             int copy_amt)
{
   memcpy(to_buffer, from_buffer, copy_amt); 
}
             

METHODDEF(void)
DGP_jcbuf_init_destination(j_compress_ptr cinfo)
{

 DGP_JpegCompDstMgr* dest = (DGP_JpegCompDstMgr*) cinfo->dest;

 /* Allocate the output buffer --- it will be released when done with image  */
 dest->buffer = (JOCTET *)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
			       DGP_BUFWRITE_SZ * SIZEOF(JOCTET));

 dest->pub.next_output_byte = dest->buffer;
 dest->pub.free_in_buffer = DGP_BUFWRITE_SZ;
}


/*
 * Empty the output buffer --- called whenever buffer fills up.
 *
 * In typical applications, this should write the entire output buffer
 * (ignoring the current state of next_output_byte & free_in_buffer),
 * reset the pointer & count to the start of the buffer, and return TRUE
 * indicating that the buffer has been dumped.
 *
 */

METHODDEF (boolean)
DGP_jcbuf_empty_output_buffer (j_compress_ptr cinfo)
{

 DGP_JpegCompDstMgr* dest = (DGP_JpegCompDstMgr*) cinfo->dest;

 if((dest->output_offset+DGP_BUFWRITE_SZ) > dest->output_length){
   ERREXIT(cinfo,JERR_DGP_BUF_WRITE);
   return FALSE;
 }

 DGP_BufWrite((dest->output_buffer+ dest->output_offset),
                             dest->buffer,DGP_BUFWRITE_SZ);

 dest->output_offset += DGP_BUFWRITE_SZ;
 dest->pub.next_output_byte = dest->buffer;
 dest->pub.free_in_buffer = DGP_BUFWRITE_SZ;

 return TRUE;
}


/*
 * Terminate destination --- called by jpeg_finish_compress
 * after all data has been written.  Usually needs to flush buffer.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */

METHODDEF (void)
DGP_jcbuf_term_destination (j_compress_ptr cinfo)
{
  DGP_JpegCompDstMgr* dest = (DGP_JpegCompDstMgr*) cinfo->dest;

  int datacount = DGP_BUFWRITE_SZ - dest->pub.free_in_buffer;

  /* Write any data remaining in the buffer */

  if((dest->output_offset+datacount) > dest->output_length){
    ERREXIT(cinfo,JERR_DGP_BUF_WRITE);
  }

  DGP_BufWrite((dest->output_buffer+ dest->output_offset),
                             dest->buffer,datacount);

  dest->output_offset += datacount;
  cinfo->dest->free_in_buffer = dest->output_offset;

}

GLOBAL (void) 
DGP_jpeg_buf_dest(j_compress_ptr cinfo,
                  DGP_JPEGOutputDef* jpeg_dest)
{
  DGP_JpegCompDstMgr * dest;

  if (cinfo->dest == NULL) {
    cinfo->dest = (struct jpeg_destination_mgr *)
      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				  SIZEOF(DGP_JpegCompDstMgr));
  }

  dest = (DGP_JpegCompDstMgr*) cinfo->dest;
  dest->pub.init_destination = DGP_jcbuf_init_destination;
  dest->pub.empty_output_buffer = DGP_jcbuf_empty_output_buffer;
  dest->pub.term_destination = DGP_jcbuf_term_destination;

  /* fill up any appln specific fields here */
  dest->output_buffer = jpeg_dest->buffer;
  dest->output_length = jpeg_dest->outLen;
  dest->output_offset = 0;
}

