/* SPMG/JPEG-LS IMPLEMENTATION V.2.1
   =====================================
   These programs are Copyright (c) University of British Columbia. All rights reserved.
   They may be freely redistributed in their entirety provided that this copyright
   notice is not removed. THEY MAY NOT BE SOLD FOR PROFIT OR INCORPORATED IN
   COMMERCIAL PROGRAMS WITHOUT THE WRITTEN PERMISSION OF THE COPYRIGHT HOLDER.
   Each program is provided as is, without any express or implied warranty,
   without even the warranty of fitness for a particular purpose.

   =========================================================
   THIS SOFTWARE IS BASED ON HP's implementation of jpeg-ls:
   =========================================================

   LOCO-I/JPEG-LS IMPLEMENTATION V.0.90
   -------------------------------------------------------------------------------
   (c) COPYRIGHT HEWLETT-PACKARD COMPANY, 1995-1999.
        HEWLETT-PACKARD COMPANY ("HP") DOES NOT WARRANT THE ACCURACY OR
   COMPLETENESS OF THE INFORMATION GIVEN HERE.  ANY USE MADE OF, OR
   RELIANCE ON, SUCH INFORMATION IS ENTIRELY AT USER'S OWN RISK.
        BY DOWNLOADING THE LOCO-I/JPEG-LS COMPRESSORS/DECOMPRESSORS
   ("THE SOFTWARE") YOU AGREE TO BE BOUND BY THE TERMS AND CONDITIONS
   OF THIS LICENSING AGREEMENT.
        YOU MAY DOWNLOAD AND USE THE SOFTWARE FOR NON-COMMERCIAL PURPOSES
   FREE OF CHARGE OR FURTHER OBLIGATION.  YOU MAY NOT, DIRECTLY OR
   INDIRECTLY, DISTRIBUTE THE SOFTWARE FOR A FEE, INCORPORATE THIS
   SOFTWARE INTO ANY PRODUCT OFFERED FOR SALE, OR USE THE SOFTWARE
   TO PROVIDE A SERVICE FOR WHICH A FEE IS CHARGED.
        YOU MAY MAKE COPIES OF THE SOFTWARE AND DISTRIBUTE SUCH COPIES TO
   OTHER PERSONS PROVIDED THAT SUCH COPIES ARE ACCOMPANIED BY
   HEWLETT-PACKARD'S COPYRIGHT NOTICE AND THIS AGREEMENT AND THAT
   SUCH OTHER PERSONS AGREE TO BE BOUND BY THE TERMS OF THIS AGREEMENT.
        THE SOFTWARE IS NOT OF PRODUCT QUALITY AND MAY HAVE ERRORS OR DEFECTS.
   THE JPEG-LS STANDARD IS STILL UNDER DEVELOPMENT. THE SOFTWARE IS NOT A
   FINAL OR FULL IMPLEMENTATION OF THE STANDARD.  HP GIVES NO EXPRESS OR
   IMPLIED WARRANTY OF ANY KIND AND ANY IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR PURPOSE ARE DISCLAIMED.
        HP SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
   OR CONSEQUENTIAL DAMAGES ARISING OUT OF ANY USE OF THE SOFTWARE.
   -------------------------------------------------------------------------------
*/

/*
 *
 * Initial code by Alex Jakulin,  Aug. 1995
 *
 * Modified and optimized: Gadiel Seroussi, October 1995
 *
 * Modified and added Restart marker and input tables by:
 * David Cheng-Hsiu Chu, and Ismail R. Ismail march 1999
 *
 * 20.3.2001 / dhtesq.
 * Added to digipaper code base for evaluation purposes.
 *
 */
#include "dgpjpegls.h"
#include "bitio.h"

/****************************************************************************
 *  INPUT ROUTINES
 *  note: some routines are implemented as preprocessor macros. See bitio.h.
 ****************************************************************************/

void 
bufiinit(DgpJpegLsContext *context) {
    context->fp = BUFSIZE;
    context->truebufsize = 0;
    context->foundeof = 0;
	return;
}

ErrCode
fillinbuff(DgpJpegLsContext *context, Dgp_UInt8 *c)
{
	Dgp_Int32 i;
	uds *in = context->in;
	DGP_STATE_PROLOGUE("fillinbuff()", context->dgp_state);
	DGP_ASSERT(c);

	/* Remember 4 last bytes of current buffer (for "undo"). */
	for ( i = -4; i < 0; i++ ){
		context->buff[i] = context->buff[context->fp + i];
	}

	context->truebufsize = in->fread((void *)context->buff, 1, BUFSIZE, in->fp);
	if ( context->truebufsize < BUFSIZE ) {
		if ( context->truebufsize <= 0 ) {
			/* Second attempt to read past EOF? */
			DGP_ERR_GENF2(context->foundeof, cErrIO_ReadError,
				"Premature EOF in compressed file;  wanted %d bytes but got %d.", BUFSIZE, context->truebufsize);

			/* One attempt to read past EOF is OK  */
			context->foundeof = 1;
	    }
	    /* fill buffer with zeros */
	    memset(context->buff + context->truebufsize, 0, (BUFSIZE - context->truebufsize) * sizeof(*context->buff));
	}

	context->fp = 1;
	*c = context->buff[0];
	NORMAL_RETURN;
}

/* Initializes the bit input routines */
ErrCode 
bitiinit(DgpJpegLsContext *context) {
	DGP_STATE_PROLOGUE("bitiinit()", context->dgp_state);
	context->bits = context->reg = 0;
    DGP_ERR_PASS(fillbuffer(context, 24));
	NORMAL_RETURN;
}

/* 
   Flush the input bit buffer TO A BYTE BOUNDARY. Return unused whole
   bytes to the byte buffer
 */
ErrCode
bitiflush(DgpJpegLsContext *context) {
    Dgp_Int32 filled, discard, dbytes, i, k, treg;
    Dgp_UInt8 *bp;
	DGP_STATE_PROLOGUE("bitiflush()", context->dgp_state);

    filled = 24 - context->bits;    /* how many bits at the MS part of reg 
			      have unused data. These correspond to 
			      at most filled+2 bits from the input
			      stream, as at most 2 '0' bits might have
			      been dropped by marker processing */

    dbytes = (filled + 2) / 8;  /* the coorrect number of bytes we need to
				"unget" is either dbytes or dbytes-1 */
    /* this solution is more general than what is required here: it 
       will work correctly even if the end of a scan is not followed
       by a marker, as will necessarily be the case with the standard JPEG
       format */
    for ( ; ; dbytes-- ) 
	{
		bp = context->buff + context->fp - dbytes;  /* back-in the buffer */
		treg = k = 0;
		for ( i = 0; i < dbytes; i++ ) 
		{
			if ( bp[i - 1] == 0xff && (bp[i] & 0x80) == 0 ) 
			{
				treg |= bp[i] << (BITBUFSIZE - 7 - k);
				k += 7;
			}
			else 
			{
				treg |= bp[i] << (BITBUFSIZE - 8 - k);
				k += 8;
			}
		}

		if ( k <= filled ) break;
    }
    /* check consistency */
	DGP_ERR_GENF3( filled - k > 7, cErrInvalidParm,
		"Inconsistent bits = %d, filled = %d, k = %d", context->bits, filled, k);

    discard = filled - k;
    if ( treg != (Dgp_Int32)(context->reg << discard) ) {
		Dgp_Int8 errString[256];
		sprintf(errString, "Inconsistent bits = %d, discard = %d, reg = %08x, treg = %08x", context->bits, discard, context->reg, treg);
		DGP_ERR_GENF0(DGP_TRUE, cErrInvalidParm, errString);
    }

	if( context->reg & (((1 << discard) - 1) << (BITBUFSIZE - discard)) ){
		DGP_WARNF3("Discarding nonzero bits; reg = %08x, bits = %d, discard = %d", context->reg, context->bits, discard);
	}

    context->fp -= dbytes;  /* do the unget */
    if ( context->buff[context->fp - 1] == 0xff && context->buff[context->fp] == 0 ) context->fp++;

    context->bits = 0;
    context->reg = 0;
	NORMAL_RETURN;
}

/* Unpad zeros from byte length */
void 
unpadzeros(DgpJpegLsContext *context)
{
	Dgp_Int32 z;

	z = context->bits % 8;
	
	if (context->bits != 24)
	{
		context->reg = (context->reg << z);
		context->bits = context->bits + z;
	}
	return;
}

/* creates the bit counting look-up table. */
/* 256 was hard-coded here as the table size when I arrived.  Should this go up to */
/* LUTMAP16 for higher bit depths? */
ErrCode
createZeroLUT(DgpJpegLsContext	*context)
{
	Dgp_Int32 i, j, k, l;
	Dgp_State *dgp_state = context->dgp_state;
	DGP_STATE_PROLOGUE("createZeroLUT()", dgp_state);

	DGP_CHECK_ALLOC(context->zeroLUT = (Dgp_Int32 *)DGP_MALLOC(sizeof(*context->zeroLUT) * LUTMAX8));

	j = k = 1; l = 8;
	for (i = 0; i < LUTMAX8; ++i) {
		context->zeroLUT[i] = l;
		--k;
		if (k == 0) {
			k = j;
			--l;
			j *= 2;
		}
	}
	NORMAL_RETURN;
}

INLINE ErrCode
DgpJpegLsGetC(	DgpJpegLsContext	*context,
				Dgp_UInt8			*c)
{
	static int i = 0;
	DGP_STATE_PROLOGUE("DgpJpegLsGetC()", context->dgp_state);
	DGP_ASSERT(c);

	if(context->fp >= BUFSIZE){
		DGP_ERR_PASS(fillinbuff(context, c));
	} else {
		*c = context->buff[context->fp];
		context->fp++;
		i++;
	}

	NORMAL_RETURN;
}

INLINE void
DgpJpegLsUngetC(DgpJpegLsContext	*context,
				Dgp_UInt8			c)
{
	context->buff[--context->fp] = c;
	return;
}

/* loads more data in the input buffer */
/* Used to be a macro... / dhtesq. */
INLINE ErrCode
fillbuffer(DgpJpegLsContext		*context,
		   Dgp_Int32			no)
{
	Dgp_UInt8 x;
	DGP_STATE_PROLOGUE("fillbuffer()", context->dgp_state);
	DGP_ASSERT(no + context->bits <= 24);
	
	context->reg <<= no;
	context->bits += no;
    while (context->bits >= 0) {
		DGP_ERR_PASS(DgpJpegLsGetC(context, &x));
		if ( x == 0xff ) {
			if ( context->bits < 8 ) {
				DgpJpegLsUngetC(context, 0xff);
				break;
			}
			else {
				DGP_ERR_PASS(DgpJpegLsGetC(context, &x));
				if ( !(x & 0x80) )  { /* non-marker: drop 0 */
					context->reg |= (0xff << context->bits) | ((x & 0x7f) << (context->bits - 7));
					context->bits -= 15;
				}
				else {
					/* marker: hope we know what we're doing */
					/* the "1" bit following ff is NOT dropped */
					context->reg |= (0xff << context->bits) | (x << (context->bits - 8));
					context->bits -= 16;
				}
				continue;
			}
		}
		context->reg |= x << context->bits;
		context->bits -= 8;
	}
	NORMAL_RETURN;
}
	
#if 0
INLINE Dgp_Int32
myfeof(DgpJpegLsContext		*context)
{
	return( (context->fp >= context->truebufsize) && context->aUds.ffeof(context->aUds.fp) ? 1:0 );
}
#endif

/*#define DgpJpegLsGetC(aUds) ((context->fp >= BUFSIZE) ? (fillinbuff(aUds)) : (context->buff[context->fp++])) */
/*#define DgpJpegLsUngetC(x, aUds) (context->buff[--context->fp] = x) */
/*#define myfeof(aUds) ((context->fp >= context->truebufsize) && aUds->feof(aUds->fp)) */


#ifdef DGPJPEGLS_IOFXNS
/* if using functions instead of macros */
dword
getbits(Dgp_Int32 no) 
{
        register dword temp;

        assert(no > 0 && no <= 18);
        temp = reg >> (32 - no);
        fillbuffer(no);
        return temp;
}
#	define GETBITS(n)	getbits(n)
#else
/* see GETBITS(x,n) in bitio.h */
#endif


/* MISSING FUNCTIONS... */
/* These appear to have been omitted from the library to keep us from building the encoder. */
/* Stubs.... / dhtesq. */

ErrCode
bitoinit(DgpJpegLsContext *c)
{
	DGP_STATE_PROLOGUE("bitoinit()", c->dgp_state);
	/*DGP_STOP_WHINE(); */
	NORMAL_RETURN;
}

void
bitoflush(void)
{
	return;
}

void
put_zeros(Dgp_Int32 value)
{
	value = value;
	return;
}

void 
putbits(Dgp_Int32	val0,
		Dgp_Int32	val1)
{
	val0 = val0;
	val1 = val1;
	return;
}
