/* SPMG/JPEG-LS IMPLEMENTATION V.2.1
   =====================================
   These programs are Copyright (c) University of British Columbia. All rights reserved.
   They may be freely redistributed in their entirety provided that this copyright
   notice is not removed.  THEY MAY NOT BE SOLD FOR PROFIT OR INCORPORATED IN
   COMMERCIAL PROGRAMS WITHOUT THE WRITTEN PERMISSION OF THE COPYRIGHT HOLDER.
   Each program is provided as is, without any express or implied warranty,
   without even the warranty of fitness for a particular purpose.

   =========================================================
   THIS SOFTWARE IS BASED ON HP's implementation of jpeg-ls:
   =========================================================

   LOCO-I/JPEG-LS IMPLEMENTATION V.0.90
   -------------------------------------------------------------------------------
   (c) COPYRIGHT HEWLETT-PACKARD COMPANY, 1995-1999.
        HEWLETT-PACKARD COMPANY ("HP") DOES NOT WARRANT THE ACCURACY OR
   COMPLETENESS OF THE INFORMATION GIVEN HERE.  ANY USE MADE OF, OR
   RELIANCE ON, SUCH INFORMATION IS ENTIRELY AT USER'S OWN RISK.
        BY DOWNLOADING THE LOCO-I/JPEG-LS COMPRESSORS/DECOMPRESSORS
   ("THE SOFTWARE") YOU AGREE TO BE BOUND BY THE TERMS AND CONDITIONS
   OF THIS LICENSING AGREEMENT.
        YOU MAY DOWNLOAD AND USE THE SOFTWARE FOR NON-COMMERCIAL PURPOSES
   FREE OF CHARGE OR FURTHER OBLIGATION.  YOU MAY NOT, DIRECTLY OR
   INDIRECTLY, DISTRIBUTE THE SOFTWARE FOR A FEE, INCORPORATE THIS
   SOFTWARE INTO ANY PRODUCT OFFERED FOR SALE, OR USE THE SOFTWARE
   TO PROVIDE A SERVICE FOR WHICH A FEE IS CHARGED.
        YOU MAY MAKE COPIES OF THE SOFTWARE AND DISTRIBUTE SUCH COPIES TO
   OTHER PERSONS PROVIDED THAT SUCH COPIES ARE ACCOMPANIED BY
   HEWLETT-PACKARD'S COPYRIGHT NOTICE AND THIS AGREEMENT AND THAT
   SUCH OTHER PERSONS AGREE TO BE BOUND BY THE TERMS OF THIS AGREEMENT.
        THE SOFTWARE IS NOT OF PRODUCT QUALITY AND MAY HAVE ERRORS OR DEFECTS.
   THE JPEG-LS STANDARD IS STILL UNDER DEVELOPMENT. THE SOFTWARE IS NOT A
   FINAL OR FULL IMPLEMENTATION OF THE STANDARD.  HP GIVES NO EXPRESS OR
   IMPLIED WARRANTY OF ANY KIND AND ANY IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR PURPOSE ARE DISCLAIMED.
        HP SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
   OR CONSEQUENTIAL DAMAGES ARISING OUT OF ANY USE OF THE SOFTWARE.
   -------------------------------------------------------------------------------
*/

/* constants.h -- Encapsulated miscellanea for the jpegls library.
 * Most everything in here was ugly stuff dragged from dgpjpegls.h.
 * / dhtesq.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "util.h"
#include "uds.h"
#include "endianness.h"
#include "dgp_state.h"

#ifndef JPEGLS_CONSTANTS_H
#define JPEGLS_CONSTANTS_H


/* Version number */
#define JPEGLSVERSION   "V.2.1"

/* Maximal number of components in the implementation*/
#define MAX_COMPONENTS	6
#define MAX_SCANS	MAX_COMPONENTS	

/* For 1st component of plane interl. mode */
#define FIRST		1

/* Different colour modes */
#define PLANE_INT	0
#define LINE_INT	1
#define PIXEL_INT	2

#define DEFAULT_COLOR_MODE PIXEL_INT

/* alphabet size */
#define MAXA8 (256)
#define MAXA16 (65536)
#define LUTMAX8 (256)
#define LUTMAX16 (4501)

/* The original implementors called this "BIG_ENDIAN", but it doesn't seem to have anything */
/* to do with platform, so I renamed it */
#define NOT_REALLY_BIG_ENDIAN	1

/****** Constants */

/* margins for scan lines */
#define	LEFTMARGIN	2
#define RIGHTMARGIN	1

/*
Dgp_Int8 *disclaimer = "\
This program is Copyright (c) University of British Columbia.\n\
All rights reserved. It may be freely redistributed in its\n\
entirety provided that this copyright notice is not removed.\n\
It may not be sold for profit or incorporated in commercial programs\n\
without the written permission of the copyright holder.\n\
";
*/

#ifdef FIXALPHA
#  ifndef ALPHA
#    define	ALPHA	256
#  endif
#  define HIGHMASK (-(ALPHA))
#  ifndef POW2
#	 define POW2
#  endif
#  if (ALPHA!=2) && (ALPHA!=4) && (ALPHA!=8) && (ALPHA!=16) && (ALPHA!=32) &&\
     (ALPHA!=64) && ( ALPHA!=128) && (ALPHA!=256) && (ALPHA!=512) &&\
     (ALPHA!=1024) && ( ALPHA!=2048) && (ALPHA!=4096) && (ALPHA!=8192) &&\
     (ALPHA!=16384) && ( ALPHA!=32768) && (ALPHA!=65536)
#   	error "Fixed alpha must be a power of 2"
#  endif
#  define  	CEIL_HALF_ALPHA (ALPHA/2)
#else
#endif

#define DEF_NEAR	0


/* Quantization threshold basic defaults */
/* These are the defaults for LOSSLESS, 8 bpp. Defaults for other
   cases are derived from these basic values */
#define	BASIC_T1	3
#define	BASIC_T2	7
#define	BASIC_T3	21
#define	BASIC_Ta	5

#define CREGIONS (9)    /* quantization regions for d-a, a-c, c-b */

/* run-length threshold */
#ifndef MAXRUN
#	define MAXRUN (64)
#endif

#define EOLINE	 1
#define NOEOLINE 0

/* number of different contexts */
#define CONTEXTS1 (CREGIONS*CREGIONS*CREGIONS)

#define CONTEXTS   ((CONTEXTS1+1)/2) /* all regions, with symmetric merging */


/* Mandatory for JPEG-LS: */
#define CLAMP
#define CLAMPB
#define CLAMPC


#define MAX_C 127
#define MIN_C -128


#define MAXCODE (N_R_L_ERROR)


/* Number of end-of-run contexts */
#define EOR_CONTEXTS 2


/* Total number of contexts */
#define TOT_CONTEXTS (CONTEXTS +  EOR_CONTEXTS)


/* index of first end-of-run context */
#define EOR_0	(CONTEXTS)


/* index of run state */
#define RUNSTATE 0



/*** offsets */

/* The longest code the bit IO can facilitate */
#define MAXCODELEN 24

/* The stat initialization values */
#define INITNSTAT 1
#define MIN_INITABSTAT 2    /* min init value for abstat[] */
#define INITABSLACK 6       /* init value for abstat is roughly 
				       2^(bpp-INITABSLACK) but not less than above */
#define INITBIASTAT 0

/* Limit for unary code */
#define LIMIT 23

/* reset values */
#define DEFAULT_RESET 64
#define MINRESET 3

#ifdef FIXRESET
#   ifndef RESET
#		define RESET     DEFAULT_RESET
#   endif
#else
#endif

#define RESRUN    256

#ifdef NOT_REALLY_BIG_ENDIAN
#    define ENDIAN8(x)   (x)
#    define ENDIAN16(x)   (x)
#else
#    define ENDIAN8(x) (x & 0x000000ff)
#    define ENDIAN16(x) ( ((x >> 8) | (x << 8)) & 0x0000ffff)
#endif

/* ENDIAN function to fix endian of PCs (for 8 bit pixels) 
#define ENDIAN8(x) (x&0x000000ff)*/


/* ENDIAN function to fix endian of PCs (for 16 bit pixels) 
#define ENDIAN16(x) ( ((x>>8)|(x<<8)) & 0x0000ffff )*/

/* clipping macro */
#ifdef POW2
#	define CLIP(x, alpha) \
	    if ( x & c->highmask ) {\
	      if(x < 0) \
			x = 0;\
	      else \
			x = alpha - 1;\
	    }
#else
#	define CLIP(x, alpha) \
	  if(x < 0)  \
	    x = 0; \
	  else if (x >= alpha) \
	    x = alpha - 1;
#endif  /* POW2 */ 



/* macro to predict Px */
#define PREDICT(Rb, Ra, Rc)	\
{	\
	register pixel minx;	\
	register pixel maxx;	\
	\
	if (Rb > Ra) {	\
		minx = Ra;	\
		maxx = Rb;	\
	} else {	\
		maxx = Ra;	\
		minx = Rb;	\
	}	\
	if (Rc >= maxx)	\
		Px = minx;	\
	else if (Rc <= minx)	\
		Px = maxx;	\
	else	\
		Px = Ra + Rb - Rc;	\
}


#endif
