/* SPMG/JPEG-LS IMPLEMENTATION V.2.1
   =====================================
   These programs are Copyright (c) University of British Columbia. All right reserved.
   They may be freely redistributed in their entirety provided that this copyright
   notice is not removed. THEY MAY NOT BE SOLD FOR PROFIT OR INCORPORATED IN
   COMMERCIAL PROGRAMS WITHOUT THE WRITTEN PERMISSION OF THE COPYRIGHT HOLDER.
   Each program is provided as is, without any express or implied warranty,
   without even the warranty of fitness for a particular purpose.

   =========================================================
   THIS SOFTWARE IS BASED ON HP's implementation of jpeg-ls:
   =========================================================

   LOCO-I/JPEG-LS IMPLEMENTATION V.0.90
   -------------------------------------------------------------------------------
   (c) COPYRIGHT HEWLETT-PACKARD COMPANY, 1995-1999.
        HEWLETT-PACKARD COMPANY ("HP") DOES NOT WARRANT THE ACCURACY OR
   COMPLETENESS OF THE INFORMATION GIVEN HERE.  ANY USE MADE OF, OR
   RELIANCE ON, SUCH INFORMATION IS ENTIRELY AT USER'S OWN RISK.
        BY DOWNLOADING THE LOCO-I/JPEG-LS COMPRESSORS/DECOMPRESSORS
   ("THE SOFTWARE") YOU AGREE TO BE BOUND BY THE TERMS AND CONDITIONS
   OF THIS LICENSING AGREEMENT.
        YOU MAY DOWNLOAD AND USE THE SOFTWARE FOR NON-COMMERCIAL PURPOSES
   FREE OF CHARGE OR FURTHER OBLIGATION.  YOU MAY NOT, DIRECTLY OR
   INDIRECTLY, DISTRIBUTE THE SOFTWARE FOR A FEE, INCORPORATE THIS
   SOFTWARE INTO ANY PRODUCT OFFERED FOR SALE, OR USE THE SOFTWARE
   TO PROVIDE A SERVICE FOR WHICH A FEE IS CHARGED.
        YOU MAY MAKE COPIES OF THE SOFTWARE AND DISTRIBUTE SUCH COPIES TO
   OTHER PERSONS PROVIDED THAT SUCH COPIES ARE ACCOMPANIED BY
   HEWLETT-PACKARD'S COPYRIGHT NOTICE AND THIS AGREEMENT AND THAT
   SUCH OTHER PERSONS AGREE TO BE BOUND BY THE TERMS OF THIS AGREEMENT.
        THE SOFTWARE IS NOT OF PRODUCT QUALITY AND MAY HAVE ERRORS OR DEFECTS.
   THE JPEG-LS STANDARD IS STILL UNDER DEVELOPMENT. THE SOFTWARE IS NOT A
   FINAL OR FULL IMPLEMENTATION OF THE STANDARD.  HP GIVES NO EXPRESS OR
   IMPLIED WARRANTY OF ANY KIND AND ANY IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR PURPOSE ARE DISCLAIMED.
        HP SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
   OR CONSEQUENTIAL DAMAGES ARISING OUT OF ANY USE OF THE SOFTWARE.
   -------------------------------------------------------------------------------
*/

/* decoder.c --- the main module, argument parsing, file I/O 
 *
 * Initial code by Alex Jakulin,  Aug. 1995
 *
 * Modified and optimized: Gadiel Seroussi, October 1995
 *
 * Color Enhancement: Guillermo Sapiro, August 1996
 *
 * Modified and added Restart marker and input tables by:
 * David Cheng-Hsiu Chu, and Ismail R. Ismail march 1999
 *
 * 20.3.2001 / dhtesq.
 * Added to digipaper code base for evaluation purposes.
 *
 */

#include "dgpjpegls.h"
#include <string.h>
#include "jpegmark.h"
#include "bitio.h"

static Dgp_Int8 *banner="\n\
=============================================\n\
SPMG/JPEG-LS DECOMPRESSOR " JPEGLSVERSION "\n\
=============================================\n\
These programs are Copyright (c) University of British Columbia.\n\
All rights reserved. They may be freely redistributed in their\n\
entirety provided that this copyright notice is not removed.\n\
They may not be sold for profit or incorporated in commercial\n\
programs without the written permission of the copyright holder.\n\
Each program is provided as is, without any express or implied\n\
warranty, without even the warranty of fitness for a particular\n\
purpose.\n\
\n\
=========================================================\n\
THIS SOFTWARE IS BASED ON HP's implementation of jpeg-ls:\n\
=========================================================\n\
(c) COPYRIGHT HEWLETT-PACKARD COMPANY, 1995-1999.\n";

/* define color mode strings */
static Dgp_Int8  *plane_int_string = "plane by plane",
				 *line_int_string = "line intlv",
				 *pixel_int_string = "sample intlv";


/* Write one row of pixel values */
INLINE ErrCode
write_one_line(DgpJpegLsContext			*context, 
			   pixel					*line, 
			   Dgp_Int32				cols)
{
	Dgp_Int32 i, index;
	Dgp_UInt32 *maptable;
	Dgp_State *dgp_state = context->dgp_state;
	jpeg_ls_header **head_scan = context->head_scan;
	uds *out = context->out;
	DGP_STATE_PROLOGUE("write_one_line()", dgp_state);
			
	/* No mapping tables used... */
	if(!(head_scan[0]->need_table)){
		if (context->bpp16 == DGP_FALSE){
			Dgp_UInt8 *line8;
			DGP_CHECK_ALLOC(line8 = (Dgp_UInt8 *)DGP_MALLOC(cols * sizeof(*line8)));
	
			for (i = 0; i < cols; i++){
				*(line8 + i) = (Dgp_UInt8)ENDIAN8(*(line + i));
			}
		
			out->fwrite((void *)line8, sizeof(*line8), cols, out->fp);
			DGP_FREE(line8);
		} else {
			out->fwrite((void *)line, sizeof(Dgp_UInt16), cols, out->fp);
		}
	}

	/* Mapping tables used... */
	else
	{
		if (context->bpp16 == DGP_FALSE)
		{
			/* Write one byte per table entry... */
			if (head_scan[0]->Wt == 1)
			{
				Dgp_UInt8 *line8;
				DGP_CHECK_ALLOC(line8 = (Dgp_UInt8 *)DGP_MALLOC(cols * sizeof(*line8)));

				maptable = head_scan[0]->TABLE[head_scan[0]->TID];

				for (i = 0; i < cols; i++){
					index = *(line + i);
					*(line8 + i) = (Dgp_UInt8)ENDIAN8(maptable[index]);
				}

				out->fwrite((void *)line8, sizeof(*line8), cols, out->fp);
				DGP_FREE(line8);
			}
			/* Write two bytes per table entry */
			else if (head_scan[0]->Wt == 2)
			{
				Dgp_UInt16 *line16;
				DGP_CHECK_ALLOC(line16 = (Dgp_UInt16 *)DGP_MALLOC(cols * sizeof(*line16)));
				
				maptable = head_scan[0]->TABLE[head_scan[0]->TID];

				for (i = 0; i < cols; i++)
				{
					index = *(line + i);
					*(line16 + i) = (Dgp_UInt16) maptable[index];
				}

				out->fwrite((void *)line16, sizeof(*line16), cols, out->fp);
				DGP_FREE(line16);
			}
			/* Write three bytes per table entry */
			else if (head_scan[0]->Wt == 3)
			{
				Dgp_UInt8 *line8_3;
				DGP_CHECK_ALLOC(line8_3 = (Dgp_UInt8 *)DGP_MALLOC(cols * sizeof(*line8_3) * 3));

				maptable = head_scan[0]->TABLE[head_scan[0]->TID];

				for (i = 0; i < cols; i++)
				{
					index = *(line + i);
					*(line8_3 + (i * 3)) =  (Dgp_UInt8) (maptable[index] >> 16);
					*(line8_3 + (i * 3) + 1) = (Dgp_UInt8) (maptable[index] >> 8);
					*(line8_3 + (i * 3) + 2) = (Dgp_UInt8) (maptable[index]);
				}

				out->fwrite((void *)line8_3, sizeof(*line8_3), cols * 3, out->fp);
				DGP_FREE(line8_3);
			}
					
		/* Can't do 16 bit index values... */
		}
		else {
			DGP_ERR_GENF0(DGP_TRUE, cErrInvalidParm,
				"Sorry, mapping tables are only supported for 8 bpp images in this implementation.");
		}
	}
	NORMAL_RETURN;
}


ErrCode
initbuffersForDecode(	DgpJpegLsContext	*context, 
						Dgp_Bool			multi,
						Dgp_Int32			comp)
{
	Dgp_Int32	i;
	Dgp_State	*dgp_state = context->dgp_state;
	DGP_STATE_PROLOGUE("initbuffersForDecode()", dgp_state);

	if(multi == DGP_TRUE){		/* Output to several files */		
		for (i = 0; i < comp; i++) {			
			DGP_CHECK_ALLOC(context->c_pscanl0[i] = (pixel *)DGP_CALLOC((context->c_columns[i] + LEFTMARGIN + RIGHTMARGIN), sizeof(pixel)));
			DGP_CHECK_ALLOC(context->c_cscanl0[i] = (pixel *)DGP_CALLOC((context->c_columns[i] + LEFTMARGIN + RIGHTMARGIN), sizeof(pixel)));

			/*  Adjust scan line pointers taking into account the margins,
				and also the fact that indexing for scan lines starts from 1
			*/
			context->c_pscanline[i] = context->c_pscanl0[i] + (LEFTMARGIN - 1);
			context->c_cscanline[i] = context->c_cscanl0[i] + (LEFTMARGIN - 1);
		}
	}
	else {			/* Output to a single file */
		DGP_CHECK_ALLOC(context->pscanl0 = (pixel *)DGP_CALLOC(context->components * (context->columns + LEFTMARGIN + RIGHTMARGIN), sizeof(pixel)));
		DGP_CHECK_ALLOC(context->cscanl0 = (pixel *)DGP_CALLOC(context->components * (context->columns + LEFTMARGIN + RIGHTMARGIN), sizeof(pixel)));
		
		/*  Adjust scan line pointers taking into account the margins,
			and also the fact that indexing for scan lines starts from 1
			(this will probably have to be changed in the future).
		*/
		context->pscanline = context->pscanl0 + context->components * (LEFTMARGIN - 1);
		context->cscanline = context->cscanl0 + context->components * (LEFTMARGIN - 1);
	}

    DGP_ERR_PASS(createZeroLUT(context));
	NORMAL_RETURN;
}

static void 
swaplines(DgpJpegLsContext *c)
{
	pixel *temp = c->pscanl0;
	c->pscanl0 = c->cscanl0;
	c->cscanl0 = temp;
	c->pscanline = c->pscanl0 + c->components * (LEFTMARGIN - 1);
	c->cscanline = c->cscanl0 + c->components * (LEFTMARGIN - 1);
	return;
}

static void
c_swaplines(DgpJpegLsContext *c, Dgp_Int32 i)
{
	pixel *temp = c->c_pscanl0[i];
	c->c_pscanl0[i] = c->c_cscanl0[i];
	c->c_cscanl0[i] = temp;
	c->c_pscanline[i] = c->c_pscanl0[i] + (LEFTMARGIN - 1);
	c->c_cscanline[i] = c->c_cscanl0[i] + (LEFTMARGIN - 1);
	return;
}


/* Formerly "closebuffers()"... */
ErrCode
DgpJpegLsDecodeCleanup(DgpJpegLsContext *c)
{
	Dgp_Int32   i;
	Dgp_State	*dgp_state = c->dgp_state;
	DGP_STATE_PROLOGUE("DgpJpegLsDecodeCleanup()", dgp_state);
	DGP_ASSERT(c);

	DGP_FREE(c->qdiv0);
	c->qdiv0 = NULL;
	DGP_FREE(c->qmul0);
	c->qmul0 = NULL;

	DGP_FREE(c->vLUT);
	c->vLUT = NULL;

	DGP_FREE(c->classmap);
	c->classmap = NULL;

	DGP_FREE(c->head_frame);
	c->head_frame = NULL;

	DGP_FREE(c->negbuff);
	c->negbuff = NULL;

	for (i = 0; i < c->number_of_scans; i++) {
		DGP_FREE(c->head_scan[i]);
		c->head_scan[i] = NULL;
	}
	
	if(c->multipleOutputFiles){
		for (i = 0; i < c->components; i++) {
			DGP_FREE(c->c_pscanl0[i]);
			c->c_pscanl0[i] = NULL;
			DGP_FREE(c->c_cscanl0[i]);
			c->c_cscanl0[i] = NULL;
		}
	} else {
		DGP_FREE(c->pscanl0);
		c->pscanl0 = NULL;
		DGP_FREE(c->cscanl0);
		c->cscanl0 = NULL;
	}
	DGP_FREE(c->A);
	c->A = NULL;
	DGP_FREE(c->B);
	c->B = NULL;
	DGP_FREE(c->C);
	c->C = NULL;
	DGP_FREE(c->N);
	c->N = NULL;
	DGP_ERR_PASS(close_process_run(c));
	DGP_FREE(c);

	NORMAL_RETURN;
}

ErrCode
DgpJpegLsInitializeDecode(DgpJpegLsContext	**context,
						  Dgp_State			*dgp_state,
						  uds				*in,
						  Dgp_Bool			verbose)
{
	Dgp_Int8 *color_mode_string;
	Dgp_Int32 i, max_samp_columns, max_samp_rows, mk, n_s,
		end_of_seek = 0,
		seek_return,
		alpha0,
		gotinf = 0,
		gotoutf = 0,
		out_files = 0;

	Dgp_Int32 pos = 0;   /* position in the file, after the header */
	DgpJpegLsContext *c = NULL;
	DGP_STATE_PROLOGUE("DgpJpegLsInitialize()", dgp_state);
	DGP_ASSERT(in);
	DGP_ASSERT(in->fp);
	DGP_ASSERT(context);

	DGP_CHECK_ALLOC_CLEANUP(c = (DgpJpegLsContext *)DGP_CALLOC(1, sizeof(*c)));
	c->in = in;
	out_files = 1;
	c->multipleOutputFiles = DGP_FALSE;
	c->dgp_state = dgp_state;
	c->verbose = verbose;
	c->negbuff = NULL;
	DGP_CHECK_ALLOC_CLEANUP(c->negbuff = (Dgp_UInt8 *)DGP_CALLOC(1, BUFSIZE + 4));
	c->buff = c->negbuff + 4;

	/* Read the compressed image frame header */

	/* (This call sets some internal buffer parameters, but doesn't do any actual "reading"...) */
	bufiinit(c);

	c->head_frame = NULL;
	DGP_CHECK_ALLOC_CLEANUP(c->head_frame = (jpeg_ls_header *)DGP_CALLOC(1, sizeof(jpeg_ls_header)));

	for (n_s = 0; n_s < MAX_SCANS; n_s++) {
		c->head_scan[n_s] = NULL;
	}
	for (n_s = 0; n_s < 1; n_s++) {
		DGP_CHECK_ALLOC_CLEANUP(c->head_scan[n_s] = (jpeg_ls_header *)DGP_CALLOC(1, sizeof(jpeg_ls_header)));

		c->head_scan[n_s]->T1 = c->head_scan[n_s]->T2 = c->head_scan[n_s]->T3 = 0;
		c->head_scan[n_s]->RES = DEFAULT_RESET;
	}

	/* Read SOI */
	DGP_ERR_PASS_CLEANUP(seek_marker(c, &mk, &seek_return));
	DGP_ERR_GEN_CLEANUPF0((seek_return == EOF), cErrIO_ReadError, "Premature End of File seeking SOI...");

	pos = seek_return;
	DGP_ERR_GEN_CLEANUPF2((mk != SOI), cErrIO_ReadError, 
		"Marker %04x found: first marker must be SOI (%04x) in this implementation.", mk, SOI);

	/* Read SOF_LS */
	DGP_ERR_PASS_CLEANUP(seek_marker(c, &mk, &seek_return));
	DGP_ERR_GEN_CLEANUPF0((seek_return == EOF), cErrIO_ReadError, "Premature End of File seeking SOF_LS...");

	pos += seek_return;
	DGP_ERR_GEN_CLEANUPF2((mk != SOF_LS), cErrIO_ReadError, 
		"Marker %04x found: second marker must be SOF_LS (%04x) in this implementation.", mk, SOF_LS);

	/* Read the frame header (SOF) */
	DGP_ERR_PASS_CLEANUP(read_jpegls_frame(c, c->head_frame, &seek_return));
	pos += seek_return;

	c->head_scan[0]->alp = c->head_frame->alp;    /* default alpha */


	/* LSE Extension header */
	/* This version supports only 2 extension headers, and 1 set of
	   parameters for all the scans. */
	c->got_lse = DGP_FALSE;
	c->got_restart = DGP_FALSE;
	c->head_scan[0]->need_table = 0;

	while (!end_of_seek){
		DGP_ERR_PASS_CLEANUP(seek_marker(c, &mk, &seek_return));
		DGP_ERR_GEN_CLEANUPF0((seek_return == EOF), cErrIO_ReadError, "Premature End of File seeking SOS, DRI, or LSE marker.");
		pos += seek_return;

		switch(mk){
		case LSE:
			DGP_ERR_PASS_CLEANUP(read_jpegls_extmarker(c, c->head_scan[0], &seek_return));
			pos += seek_return;
			c->got_lse = DGP_TRUE;
			break;
		
		case DRI:
			DGP_ERR_PASS_CLEANUP(read_jpegls_restartmarker(c, c->head_scan[0], &seek_return));
			pos += seek_return;
			c->got_restart = DGP_TRUE;
			c->restart_interval = c->head_scan[0]->restart_interval;
			break;
		
		case SOS:
			end_of_seek = 1;
			break;

		default:
			DGP_ERR_GEN_CLEANUPF1(DGP_TRUE, cErrInvalidParm,
				"Invalid marker type 0x%x.", mk);
		}
	}
	/* End of extension header. */

	/* Read the scan header. */
	DGP_ERR_PASS_CLEANUP(read_jpegls_scan(c, c->head_scan[0], &seek_return));

	pos += seek_return;

	c->shift = c->head_scan[0]->shift;
	DGP_ERR_GEN_CLEANUPF1(c->shift != 0, cErrNoImpl,
		"I got shift = %d, but values other than 0 are not implemented.", c->shift);

	c->nearLosslessError = c->head_scan[0]->nearLosslessError;
	c->color_mode = c->head_scan[0]->color_mode;
	c->columns = c->head_frame->columns;
	c->rows = c->head_frame->rows;
	alpha0 = c->head_scan[0]->alp;
	c->head_frame->alp = alpha0;
	c->components = c->head_frame->comp;

	if (c->color_mode == PLANE_INT) {
		c->number_of_scans = c->components;
		/* This is bizarre, but "multiple output files" seems to include the case of plane interleaved output */
		/* with any number of components, including one. */
		c->multipleOutputFiles = DGP_TRUE;
	}
	else{
		c->number_of_scans = 1;
	}
	for (n_s = 1; n_s < c->number_of_scans; n_s++) {
		DGP_CHECK_ALLOC_CLEANUP(c->head_scan[n_s] = (jpeg_ls_header *)DGP_CALLOC(1, sizeof(jpeg_ls_header)));

		c->head_scan[n_s]->T1 = c->head_scan[n_s]->T2 = c->head_scan[n_s]->T3 = 0;
		c->head_scan[n_s]->RES = DEFAULT_RESET;
	}

	set_thresholds(c->head_scan[0]->alp, c->head_scan[0]->nearLosslessError, &(c->head_scan[0]->T1), &(c->head_scan[0]->T2), &(c->head_scan[0]->T3));
	c->T1 = c->head_scan[0]->T1;
	c->T2 = c->head_scan[0]->T2;
	c->T3 = c->head_scan[0]->T3;

#ifndef FIXRESET
	c->theReset = c->head_scan[0]->RES;
#else
	DGP_ERR_GEN_CLEANUPF2((c->head_scan[0]->RES != DEFAULT_RESET), cErrInvalidParm,
		"This version was compiled for a fixed RESET parameter %d, but I got %d from file.", DEFAULT_RESET, c->head_scan[0]->RES);
	c->theReset = DEFAULT_RESET;
#endif

	/* Lossless or lossy? */
	if (c->nearLosslessError == 0){
		c->lossy = DGP_FALSE;
	}
	else{
		c->lossy = DGP_TRUE;
	}

	/* Check for 16 or 8 bit mode. */
	if (alpha0 <= MAXA16 && alpha0 > MAXA8)	{
		/* 16 bit */
		c->bpp16 = DGP_TRUE;
		c->lutmax = LUTMAX16;
	}
	else if (alpha0 <= MAXA8 && alpha0 >= 1)
	{
		/* 8 bit */
		c->bpp16 = DGP_FALSE;
		c->lutmax = LUTMAX8;
	}
	else { 
		DGP_ERR_GEN_CLEANUPF1(DGP_TRUE, cErrInvalidParm,
			"I got an invalid value for alpha, %d.", alpha0 + 1);
	}

	DGP_ERR_PASS_CLEANUP(check_compatibility(dgp_state, c->head_frame, c->head_scan[0], 0));

	for (i = 0; i < c->components; i++) {
		c->samplingx[i] = c->head_frame->samplingx[i];
		c->samplingy[i] = c->head_frame->samplingy[i];
		DGP_ERR_GEN_CLEANUPF0(((c->multipleOutputFiles == DGP_FALSE) && ((c->samplingx[i] != 1) || (c->samplingy[i] != 1)) && (c->components > 1)),
			cErrInvalidParm, "I can't do PPM output with subsampled compressed components.");
	}

	if((c->multipleOutputFiles == DGP_FALSE) && (c->color_mode == PLANE_INT)){
		DGP_ERR_GEN_CLEANUPF0((c->components > 1), cErrInvalidParm,
			"I can't do PPM output in plane intereleaved mode.");
	}
	else{
		c->multipleOutputFiles = DGP_TRUE;
	}

	if((c->multipleOutputFiles == DGP_TRUE) && (c->color_mode == PIXEL_INT)) {
		c->multipleOutputFiles = DGP_FALSE;   /* equivalent to specifying -P flag */
	}

	DGP_ERR_GEN_CLEANUPF0(c->multipleOutputFiles && (c->color_mode == PIXEL_INT), cErrInvalidParm,
		"Pixel interleaved mode uses only 1 output file.");

	DGP_ERR_GEN_CLEANUPF2(c->multipleOutputFiles && (out_files != c->components), cErrInvalidParm,
		"Number of files for output (%d) must be equal to number of image components (%d).", out_files, c->components);

	/* Compute the image size for the different components. */
	if (c->components == 1) {
		c->c_columns[0] = c->columns;
		c->c_rows[0] = c->rows;
		c->whose_max_size_rows = 0;
		c->samplingy[0] = 1;
	}
	else{
		max_samp_columns = 0;
		max_samp_rows = 0;
		for(i = 0; i < c->components; i++) {
			if (c->samplingx[i] > max_samp_columns) {
				max_samp_columns = c->samplingx[i];
				c->whose_max_size_columns = i;
			}
			if (c->samplingy[i] > max_samp_rows) {
				max_samp_rows = c->samplingy[i];
				c->whose_max_size_rows = i;
			}
		}

		c->c_columns[c->whose_max_size_columns] = c->columns;	
		c->c_rows[c->whose_max_size_rows] = c->rows;

		for(i = 0; i < c->components; i++) {
			if (i != c->whose_max_size_columns)	{
				c->c_columns[i] = c->c_columns[c->whose_max_size_columns] * c->samplingx[i];
				c->c_columns[i] /= c->samplingx[c->whose_max_size_columns];
			}
			if (i != c->whose_max_size_rows) {
				c->c_rows[i] = c->c_rows[c->whose_max_size_rows] * c->samplingy[i];
				c->c_rows[i] /= c->samplingy[c->whose_max_size_rows];
			}
		}
	}

	/* Files have been opened by caller... */

	/* Check that color mode is valid and pick color mode string. */
	switch ( c->color_mode ) {
	case PLANE_INT:
		color_mode_string = plane_int_string;
		break;

	case LINE_INT:
		color_mode_string = line_int_string;
		break;
	
    case PIXEL_INT:
		color_mode_string = pixel_int_string;
		break;

	default:
		DGP_ERR_GEN_CLEANUPF1(DGP_TRUE, cErrInvalidParm,
			"Invalid color mode %d", c->color_mode);
	}

	/*DGP_REPORTF0(verbose, banner); */

#ifdef FIXALPHA
	DGP_ERR_GEN_CLEANUPF2(( alpha0 != alpha ), cErrInvalidParm,
		"This version has been optimized for alphabet size %d, but I got %d from the file.", alpha, alpha0);
	c->alpha = alpha;		/* or? */
#else
	c->alpha = alpha0;
	c->ceil_half_alpha = (c->alpha + 1) / 2;
#endif

#ifdef POW2
	c->highmask = -c->alpha;
	/* check that alpha is a power of 2 */
	for ( alpha0 = c->alpha, i = -1; alpha0; alpha0 >>= 1, i++);
	DGP_ERR_GEN_CLEANUPF1(c->alpha != (1 << i), cErrInvalidParm,
		"This version has been optimized for a power of 2 alphabet size, but I got %d.", c->alpha);
#endif	
/*#else?... */

	if (c->lossy == DGP_TRUE) {
		/* compute auxiliary parameters for near-lossless (globals) */
		c->quant = 2 * c->nearLosslessError + 1;
		c->qbeta = (c->alpha + 2 * c->nearLosslessError + c->quant - 1 ) / c->quant;
		c->beta = c->quant * c->qbeta;
		c->ceil_half_qbeta = (c->qbeta + 1) / 2;
		c->negNearLosslessError = -c->nearLosslessError;
		c->alpha1eps = c->alpha - 1 + c->nearLosslessError;
		DGP_REPORTF3(verbose, "Lossy mode: nearLosslessError = %d, beta = %d, qbeta = %d", c->nearLosslessError, c->beta, c->qbeta);
	}

	/* compute bits per sample for input symbols */
	for ( c->bpp = 1; (1L << c->bpp) < c->alpha; c->bpp++ );

	/* compute bits per sample for unencoded prediction errors */
	if (c->lossy == DGP_TRUE){
		for ( c->qbpp = 1; (1L << c->qbpp) < c->qbeta; c->qbpp++ );
	}
	else{
		c->qbpp = c->bpp;
	}

	if ( c->bpp < 2 ) c->bpp = 2;

	/* limit for unary part of Golomb code */
	if ( c->bpp < 8 ){
	    c->limit = 2 * (c->bpp + 8) - c->qbpp - 1;
	}
	else{
	    c->limit = 4 * c->bpp - c->qbpp - 1;
	}

	/* print out parameters */
	DGP_REPORTF2(verbose && (c->multipleOutputFiles == DGP_FALSE), "Image:  %d cols, %d rows.", c->columns, c->rows);
	DGP_REPORTF4(verbose && (c->multipleOutputFiles == DGP_FALSE), 
		"Image:  alpha = %d, comp = %d, color_mode = %d (%s).", alpha0, c->components, c->color_mode, color_mode_string);
	DGP_REPORTF5(verbose && (c->multipleOutputFiles == DGP_FALSE), 
		"Parameters:  T1 = %d, T2 = %d, T3 = %d, RESET = %d, limit = %d.", c->T1, c->T2, c->T3, c->theReset, c->limit);

	if(c->verbose && c->multipleOutputFiles == DGP_TRUE){
		for (i = 0; i < c->components; i++){
			DGP_REPORTF3(DGP_TRUE, "Image %d:\t%d cols, %d rows.", i, c->c_columns[i], c->c_rows[i]);
		}
		DGP_REPORTF4(DGP_TRUE, "alpha =  %d, comp = %d, color_mode = %d (%s).", alpha0, c->components, c->color_mode, color_mode_string);
		DGP_REPORTF5(DGP_TRUE, "Parameters:\tT1 = %d, T2 = %d, T3 = %d, RESET = %d, limit = %d.", c->T1, c->T2, c->T3, c->theReset, c->limit);
	}		

	/* Allocate memory pools. */
	DGP_ERR_PASS_CLEANUP(initbuffersForDecode(c, c->multipleOutputFiles, c->components));

	/* Write in header size, in bytes */
	c->headerBytesRead = pos;
	*context = c;
	NORMAL_RETURN;
	
ERR_CLEANUP:
	if(c->head_frame) DGP_FREE(c->head_frame);
	for (n_s = 0; n_s < MAX_SCANS; n_s++) {
		if(c->head_scan[n_s]) DGP_FREE(c->head_scan[n_s]);
	}
	if(c->negbuff) DGP_FREE(c->negbuff);
	if(c) DGP_FREE(c);
	ERR_RETURN;
}


/* Main loop for decoding files */
/* Not all cases implemented:  I'm concentrating on pixel-interleaved output to a single file or buffer. / dhtesq. */
ErrCode
DgpJpegLsDecode (DgpJpegLsContext *c,
				 uds			*out,
				 Dgp_Float64	*decodeTime)
{				 
	Dgp_Int32		n, n_c, n_s, mk, seek_return;
	Dgp_Int32		found_EOF = 0, retval;
	Dgp_Float64		t0, t1, get_utime();
	Dgp_Int32		pos0, 
					pos1,
					tot_in = 0,
					tot_out = 0;
	pixel			*local_scanl0 = NULL, 
					*local_scanl1 = NULL,
					*local_pscanline = NULL,
					*local_cscanline = NULL;
	Dgp_Int32		MCUs_counted;	
	Dgp_Int32		dummyMarker;
	Dgp_Bool		multi, lossy;
	Dgp_Bool		verbose = c->verbose;
	Dgp_State		*dgp_state = c->dgp_state;
	uds				*in = c->in;
	DGP_STATE_PROLOGUE("DgpJpegLsDecode()", dgp_state);
	DGP_ASSERT(out);
	DGP_ASSERT(c);
	
	multi = c->multipleOutputFiles;
	lossy = c->lossy;
	pos0 = c->headerBytesRead;
	c->out = out;

	/* start timer (must be AFTER initialization) */
	t0 = get_utime();

	/* Initialize the scanline buffers */
	if (multi == DGP_FALSE){
		for (n = 0; n <= (c->columns + 1); n++){
			c->pscanline[n] = 0;
		}
	}
	else{
		for (n_c = 0; n_c < c->components; n_c++){
			for (n = 0; n <= (c->c_columns[n_c] + 1); n++){
				c->c_pscanline[n_c][n] = 0;
			}
		}
	}

	if ((c->components > 1) && (multi == DGP_FALSE)){
		/* OUTPUT PPM file, allocate auxiliary buffers */
		DGP_CHECK_ALLOC(local_scanl0 = (pixel *)DGP_CALLOC(c->columns + LEFTMARGIN + RIGHTMARGIN, sizeof (pixel)));
		DGP_CHECK_ALLOC(local_scanl1 = (pixel *)DGP_CALLOC(c->columns + LEFTMARGIN + RIGHTMARGIN, sizeof (pixel)));

		local_pscanline = local_scanl0 + LEFTMARGIN - 1;
		local_cscanline = local_scanl1 + LEFTMARGIN - 1;
	}

	for (n_s = 0; n_s < c->number_of_scans; n_s++){
		/* Repeat prediction/quantization/statistics scanline by scanline, for each scan. */

		/* Reset Restart Markers for every scan. */
		MCUs_counted = 0;
		
		/* This implementation supports parameters in 1st scan. */
		if (n_s == 0){
			/* Prepare the quantization LUTs */
			DGP_ERR_PASS_CLEANUP(prepareLUTs(	dgp_state,
												c->alpha,
												c->lutmax,
												c->T1,
												c->T2,
												c->T3,
												c->nearLosslessError,
												lossy,
												&c->classmap,
												&c->vLUT));

			if(lossy == DGP_TRUE){
				/* prepare div/mul tables for near-lossless quantization */
				DGP_ERR_PASS_CLEANUP(prepare_qtables(	dgp_state,
														c->alpha,
														c->nearLosslessError,
														&c->qdiv0, &c->qdiv,
														&c->qmul0, &c->qmul));
			}
			DGP_CHECK_ALLOC_CLEANUP(c->A = (Dgp_Int32 *)DGP_MALLOC(sizeof(*c->A) * TOT_CONTEXTS));
			DGP_CHECK_ALLOC_CLEANUP(c->B = (Dgp_Int32 *)DGP_MALLOC(sizeof(*c->B) * TOT_CONTEXTS));
			DGP_CHECK_ALLOC_CLEANUP(c->C = (Dgp_Int32 *)DGP_MALLOC(sizeof(*c->C) * TOT_CONTEXTS));
			DGP_CHECK_ALLOC_CLEANUP(c->N = (Dgp_Int32 *)DGP_MALLOC(sizeof(*c->N) * TOT_CONTEXTS));
		}
		else{	/* Read further scan headers */
			DGP_ERR_PASS_CLEANUP(seek_marker(c, &mk, &seek_return));
			DGP_ERR_GEN_CLEANUPF0((seek_return == EOF), cErrIO_ReadError,
				"Premature End of File seeking SOS marker.");

			if(seek_return > 2){
				DGP_WARNF1("%d extra bytes between end of scan and next marker.  Added to marker segment count.", seek_return - 2);
			}
			pos0 += seek_return;
			DGP_ERR_GEN_CLEANUPF2((mk != SOS), cErrIO_ReadError,
				"Expecting SOS (%04x), but I got %04x.", SOS, mk);

			/* Read the scan header*/
			DGP_ERR_PASS_CLEANUP(read_jpegls_scan(c, c->head_scan[n_s], &seek_return));

			pos0 += seek_return;

			DGP_ERR_GEN_CLEANUPF1(c->head_scan[n_s]->shift != 0, cErrNoImpl,
				"I got shift = %d != 0; not implemented.", c->head_scan[n_s]->shift);

			DGP_ERR_GEN_CLEANUPF2(c->head_scan[n_s]->nearLosslessError != c->nearLosslessError, cErrInvalidParm,
				"nearLosslessError went from %d to %d:  cannot change parameters between scans in this implementation.",
				c->head_scan[n_s]->nearLosslessError, c->nearLosslessError);

			DGP_ERR_GEN_CLEANUPF0(((c->head_scan[n_s]->color_mode != PLANE_INT) || (c->head_scan[n_s]->comp != 1) ||
				(c->head_scan[n_s]->comp_ids[0] != n_s + 1)), cErrNoImpl,
				"This implementation supports multiple scans only in PLANE INTERLEAVED mode.");
		}

		/* Initializations for each scan */
		DGP_ERR_PASS_CLEANUP(bitiinit(c));

		/* Start from 1st image row */
		n = 0;

        /* Initialize stats arrays */
		DGP_ERR_PASS_CLEANUP(init_stats(dgp_state, lossy ? c->qbeta:c->alpha, c->A, c->B, c->C, c->N));

		/* Initialize run processing */
		DGP_ERR_PASS_CLEANUP(init_process_run(c, c->components, MAXRUN));

		switch(c->color_mode){
		/* We won't normally be generating plane-interleaved files, but the codec defaults to this mode for single-component */
		/* grayscale. */
		/***********************************************************************/
		/*           Plane interleaved mode 								   */
		/***********************************************************************/
		case PLANE_INT:
			if (lossy == DGP_FALSE){			
				/* LOSSLESS MODE */
				while (++n <= c->c_rows[n_s]){
					/* 'extend' the edges */
					c->c_cscanline[n_s][-1] = c->c_cscanline[n_s][0] = c->c_pscanline[n_s][1];

					DGP_ERR_PASS_CLEANUPF2(lossless_undoscanline(c, c->c_pscanline[n_s], c->c_cscanline[n_s], c->c_columns[n_s], n_s ),
						"Premature EOF?  Expected %d rows, got %d.\n", c->rows, n - 1);

					DGP_ERR_PASS_CLEANUP(write_one_line(c, c->c_cscanline[n_s] + 1, c->c_columns[n_s]));

					tot_out += c->c_columns[n_s];

					/* extend the edges */
					c->c_cscanline[n_s][c->c_columns[n_s] + 1] = c->c_cscanline[n_s][c->c_columns[n_s]];

					/* make the current scanline the previous one */
					c_swaplines(c, n_s);

					/* Take out the Restart Markers */
					if (c->got_restart){
						/* Look for restart markers only after a restart interval */
						if ((MCUs_counted % c->restart_interval) == 0){
							DGP_ERR_PASS_CLEANUP(bitiflush(c));
							DGP_ERR_PASS_CLEANUP(read_n_bytes(c, 2, &dummyMarker));	/* read the RST marker */
							DGP_ERR_PASS_CLEANUP(bitiinit(c));
						}
						MCUs_counted++;						
					}					
				}	/* End of line loop in PLANE_INT */
			}
			else{

				/* LOSSY MODE */
				while (++n <= c->c_rows[n_s]){
					/* 'extend' the edges */
					c->c_cscanline[n_s][-1] = c->c_cscanline[n_s][0] = c->c_pscanline[n_s][1];

					DGP_ERR_PASS_CLEANUPF2(lossy_undoscanline(c, c->c_pscanline[n_s], c->c_cscanline[n_s], c->c_columns[n_s], n_s ),
						"Premature EOF? Expected %d rows, got %d.\n", c->rows, n - 1);

					DGP_ERR_PASS_CLEANUP(write_one_line(c, c->c_cscanline[n_s] + 1, c->c_columns[n_s]));

					tot_out += c->c_columns[n_s];

					/* extend the edges */
					c->c_cscanline[n_s][c->c_columns[n_s] + 1] = c->c_cscanline[n_s][c->c_columns[n_s]];

					/* make the current scanline the previous one */
					c_swaplines(c, n_s);

					/* Take out the Restart Markers */
					if (c->got_restart){
						/* Look for restart markers only after a restart interval */
						if ((MCUs_counted % c->restart_interval) == 0){
							DGP_ERR_PASS_CLEANUP(bitiflush(c));
							DGP_ERR_PASS_CLEANUP(read_n_bytes(c, 2, &dummyMarker));	/* read the RST marker */
							DGP_ERR_PASS_CLEANUP(bitiinit(c));
						}
						MCUs_counted++;						
					}
				}	/* End of line loop in PLANE_INT */
			}
			break;

		/****************************************************************************/
		/*           Pixel interleaved mode with single file (buffer) received		*/
		/****************************************************************************/
		case PIXEL_INT:

			if (lossy == DGP_FALSE){			
				/* LOSSLESS MODE */
				while (++n <= c->rows)	{
					/* 'extend' the edges */
					for (n_c = 0; n_c < c->components; n_c++){
						c->cscanline[-c->components + n_c] = c->cscanline[n_c] = c->pscanline[c->components + n_c];
					}
					
					DGP_ERR_PASS_CLEANUPF2(lossless_undoscanline_pixel(c, c->pscanline, c->cscanline, c->components * c->columns, &retval),
						"Premature EOF? Expected %d rows, but I got %d.", c->rows, n - 1);
					
					DGP_ERR_PASS_CLEANUP(write_one_line(c, c->cscanline + c->components, c->components * c->columns));
					
					tot_out += c->components * c->columns;
					
					/* extend the edges */
					for (n_c = 0; n_c < c->components; n_c++){
						c->cscanline[c->components * (c->columns + 1) + n_c] = c->cscanline[c->components * c->columns + n_c];
					}

					/* Make the current scanline the previous one. */
					swaplines(c);
					
					/* Take out the Restart Markers */
					if (c->got_restart){
						/* Look for restart markers only after a restart interval... */
						if ((MCUs_counted % c->restart_interval) == 0){
							DGP_ERR_PASS_CLEANUP(bitiflush(c));
							DGP_ERR_PASS_CLEANUP(read_n_bytes(c, 2, &dummyMarker));
							DGP_ERR_PASS_CLEANUP(bitiinit(c));
						}
						MCUs_counted++;					
					}				
				}	/* End of line loop for PIXEL_INT */
			}
			else{
				
				/* LOSSY MODE */
				while (++n <= c->rows){
					/* 'extend' the edges */
					for (n_c = 0; n_c < c->components; n_c++){
						c->cscanline[-c->components + n_c] = c->cscanline[n_c] = c->pscanline[c->components + n_c];
					}

					DGP_ERR_PASS_CLEANUPF2(lossy_undoscanline_pixel(c, c->pscanline, c->cscanline, c->components * c->columns, &retval),
						"Premature EOF? Expected %d rows, but I got %d.", c->rows, n - 1);

					DGP_ERR_PASS_CLEANUP(write_one_line(c, c->cscanline + c->components, c->components * c->columns));

					tot_out += c->components * c->columns;

					/* extend the edges */
					for (n_c = 0; n_c < c->components; n_c++){
						c->cscanline[c->components * (c->columns + 1) + n_c] = c->cscanline[c->components * c->columns + n_c];
					}
					
					/* make the current scanline the previous one */
					swaplines(c);

					/* Take out the Restart Markers */
					if (c->got_restart){
						/* Look for restart markers only after a restart interval... */
						if ((MCUs_counted % c->restart_interval) == 0){
							DGP_ERR_PASS_CLEANUP(bitiflush(c));
							DGP_ERR_PASS_CLEANUP(read_n_bytes(c, 2, &dummyMarker));
							DGP_ERR_PASS_CLEANUP(bitiinit(c));
						}
						MCUs_counted++;					
					}				
				} /* End of line loop for PIXEL_INT */
			}
			break;
		default:
			DGP_ERR_GENF1(DGP_TRUE, cErrNoImpl, "Color mode %d not implemented...", c->color_mode);
			break;
		} /* end switch{}... */
		DGP_ERR_PASS_CLEANUP(bitiflush(c));
	} /* End of loop for scans */

	mk = 0;

	/* Read EOI */
	DGP_ERR_PASS_CLEANUP(seek_marker(c, &mk, &seek_return));
	DGP_ERR_GEN_CLEANUPF0(seek_return == EOF, cErrIO_ReadError,
		"I didn't get an EOI marker at the end of the compressed image.");

	if ( seek_return > 2 ){
		DGP_WARNF1("%d extra bytes between end of scan and next marker added to marker segment count.", seek_return - 2);
	}

	pos0 += seek_return;

	DGP_ERR_GEN_CLEANUPF0((mk != EOI), cErrIO_ReadError,
		"In this implementation last marker must be EOI.");

	DGP_REPORTF2(verbose && c->head_scan[0]->need_table,
		"I used a mapping table which had %i entries of %i bytes each.", c->head_scan[0]->MAXTAB, c->head_scan[0]->Wt);

	DGP_REPORTF1(verbose && c->got_restart, 
		"I found restart markers with a a restart interval of %i.", c->restart_interval);

	DGP_REPORTF1(verbose, 
		"Marker segment bytes: %ld.",pos0);

	/* position in input file */
	pos1 = in->ftell(in->fp);
	/* tot_in = 8*(pos1-pos0); */

	/* size of compressed file read (bits), incl. headers. */
	tot_in = 8 * pos1;

	/* Close down */
	DGP_ERR_PASS_CLEANUP(DgpJpegLsDecodeCleanup(c));
	if(local_scanl0) DGP_FREE(local_scanl0);
	if(local_scanl1) DGP_FREE(local_scanl1);

	t1 = get_utime();
	DGP_REPORTF3(verbose,
		"Total bits in: %ld.  Symbols out: %ld.  %5.3lf bps.", tot_in, tot_out, tot_in / (Dgp_Float64)tot_out);
	DGP_REPORTF2(verbose,
		"Time = %1.3lf secs, %1.0lf KSymbols/sec.", t1 - t0, (tot_out) / (1024 * (t1 - t0)));

	if(decodeTime){
		*decodeTime = t1 - t0;
	}
	NORMAL_RETURN;

ERR_CLEANUP:
	if(c){
		if(c->qdiv0) DGP_FREE(c->qdiv0);
		if(c->qmul0) DGP_FREE(c->qmul0);
		if(c->vLUT) DGP_FREE(c->vLUT);
		if(c->classmap) DGP_FREE (c->classmap);
		if(c->head_frame) DGP_FREE(c->head_frame);
		if(c->negbuff) DGP_FREE(c->negbuff);
		for (n_s = 0; n_s < MAX_SCANS; n_s++) {
			if(c->head_scan[n_s]) DGP_FREE(c->head_scan[n_s]);
		}
		if(multi){
			for (n_c = 0; n_c < c->components; n_c++) {
				if(c->c_pscanl0[n_c]) DGP_FREE(c->c_pscanl0[n_c]);
				if(c->c_cscanl0[n_c]) DGP_FREE(c->c_cscanl0[n_c]);
			}
		} else {
			if(c->pscanl0) DGP_FREE(c->pscanl0);
			if(c->cscanl0) DGP_FREE(c->cscanl0);
		}
		if(c->A) DGP_FREE(c->A);
		if(c->B) DGP_FREE(c->B);
		if(c->C) DGP_FREE(c->C);
		if(c->N) DGP_FREE(c->N);
		DGP_ERR_PASS(close_process_run(c));
		DGP_FREE(c);
	}
	if(local_scanl0) DGP_FREE(local_scanl0);
	if(local_scanl1) DGP_FREE(local_scanl1);

	ERR_RETURN;
}

