/* SPMG/JPEG-LS IMPLEMENTATION V.2.1
   =====================================
   These programs are Copyright (c) University of British Columbia. All rights reserved.
   They may be freely redistributed in their entirety provided that this copyright
   notice is not removed.  THEY MAY NOT BE SOLD FOR PROFIT OR INCORPORATED IN
   COMMERCIAL PROGRAMS WITHOUT THE WRITTEN PERMISSION OF THE COPYRIGHT HOLDER.
   Each program is provided as is, without any express or implied warranty,
   without even the warranty of fitness for a particular purpose.

   =========================================================
   THIS SOFTWARE IS BASED ON HP's implementation of jpeg-ls:
   =========================================================

   LOCO-I/JPEG-LS IMPLEMENTATION V.0.90
   -------------------------------------------------------------------------------
   (c) COPYRIGHT HEWLETT-PACKARD COMPANY, 1995-1999.
        HEWLETT-PACKARD COMPANY ("HP") DOES NOT WARRANT THE ACCURACY OR
   COMPLETENESS OF THE INFORMATION GIVEN HERE.  ANY USE MADE OF, OR
   RELIANCE ON, SUCH INFORMATION IS ENTIRELY AT USER'S OWN RISK.
        BY DOWNLOADING THE LOCO-I/JPEG-LS COMPRESSORS/DECOMPRESSORS
   ("THE SOFTWARE") YOU AGREE TO BE BOUND BY THE TERMS AND CONDITIONS
   OF THIS LICENSING AGREEMENT.
        YOU MAY DOWNLOAD AND USE THE SOFTWARE FOR NON-COMMERCIAL PURPOSES
   FREE OF CHARGE OR FURTHER OBLIGATION.  YOU MAY NOT, DIRECTLY OR
   INDIRECTLY, DISTRIBUTE THE SOFTWARE FOR A FEE, INCORPORATE THIS
   SOFTWARE INTO ANY PRODUCT OFFERED FOR SALE, OR USE THE SOFTWARE
   TO PROVIDE A SERVICE FOR WHICH A FEE IS CHARGED.
        YOU MAY MAKE COPIES OF THE SOFTWARE AND DISTRIBUTE SUCH COPIES TO
   OTHER PERSONS PROVIDED THAT SUCH COPIES ARE ACCOMPANIED BY
   HEWLETT-PACKARD'S COPYRIGHT NOTICE AND THIS AGREEMENT AND THAT
   SUCH OTHER PERSONS AGREE TO BE BOUND BY THE TERMS OF THIS AGREEMENT.
        THE SOFTWARE IS NOT OF PRODUCT QUALITY AND MAY HAVE ERRORS OR DEFECTS.
   THE JPEG-LS STANDARD IS STILL UNDER DEVELOPMENT. THE SOFTWARE IS NOT A
   FINAL OR FULL IMPLEMENTATION OF THE STANDARD.  HP GIVES NO EXPRESS OR
   IMPLIED WARRANTY OF ANY KIND AND ANY IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR PURPOSE ARE DISCLAIMED.
        HP SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
   OR CONSEQUENTIAL DAMAGES ARISING OUT OF ANY USE OF THE SOFTWARE.
   -------------------------------------------------------------------------------
*/

/* dgpjpegls.h --- prototypes for functions and global variables 
 *
 * Initial code by Alex Jakulin,  Aug. 1995
 *
 * Modified and optimized: Gadiel Seroussi, October 1995
 *
 * Modified and added Restart marker and input tables by:
 * David Cheng-Hsiu Chu, and Ismail R. Ismail march 1999
 *
 * 20.3.2001 / dhtesq.
 * Added to digipaper code base for evaluation purposes.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "constants.h"
#include "util.h"
#include "uds.h"
#include "endianness.h"
#include "dgp_state.h"

#ifndef JPEGLS_PROTOTYPES_H
#define JPEGLS_PROTOTYPES_H

typedef Dgp_UInt16 pixel;

typedef struct	jpeg_ls {

	Dgp_Int32	columns,					/* The number of columns */
		rows,						/* Number of rows */
		alp,						/* alphabet size (Max+1) , 2 bytes*/
		comp,						/* number of components, 1 byte  */
		nearLosslessError,			/* near-lossless error, 1 byte  */
		color_mode,					/* indicates the color mode , 1 byte */
		need_lse,					/* Indicates non-default parameters */
		need_table,					/* Indicates use of mapping table */
		need_restart,				/* Indicates use of restart markers */
		restart_interval,			/* The number of MCU's between restart markers */
		shift,						/* for sparse images, 1 byte */
		T1,T2,T3,					/* Thresholds, 2 bytes each */
		RES,						/* reset value for counters, 2 bytes */
		samplingx[MAX_COMPONENTS],	/* col. sampling rates 1 byte each*/
		samplingy[MAX_COMPONENTS],	/* row sampling rates */
		comp_ids[MAX_COMPONENTS],	/* component id's */
		acc_size,					/* 1 byte */
		adds[MAX_COMPONENTS];		/* size given by acc_size */
		Dgp_UInt32	TID,			/* Table ID, 1 byte */
			        MAXTAB,			/* Maximum table index value */
				    Wt,				/* Width of each table entry, 1 byte */
					*TABLE[MAX_COMPONENTS];	/* The table(s) for each component */

} jpeg_ls_header;


/* 27.3.2001 / dhtesq. */
typedef struct {
	Dgp_Int8 *current;
	Dgp_Int8 *start;
	Dgp_Int32 sizeInBytes;
	Dgp_Int32 width, height;
	Dgp_Int32 lpl;
	Dgp_Int32 xOffset;
	Dgp_Int32 xMax, yMax;
	Dgp_Int32 stopX, stopY;
	Dgp_Int32 currentRow, currentColumn;
	Dgp_Int32 writeableBytesPerRow;
	Dgp_Int32 bpp;
} DgpJpegLsMemoryStream;


/* 28.3.2001 / dhtesq. */
/* General encode & decode context. */
typedef struct {
	Dgp_Int32			headerBytesRead;
	Dgp_UInt32			reg;		/* Bit buffer for i/o. */
	Dgp_Int32			bits;		/* number of bits free in bit buffer (on output) */

	Dgp_UInt8			*negbuff;
	Dgp_UInt8			*buff;
	Dgp_Int32			fp;			/* index into byte buffer */
	Dgp_Int32			truebufsize; /* true size of byte buffer ( <= BUFSIZE) */
	Dgp_Int32			foundeof;

	/* melcode vectors (for run-length)... */
	Dgp_Int32			*melcstate;			/* index to the state array */
	Dgp_Int32			*melclen;				/*	contents of the state array location
         										indexed by melcstate: the "expected"
												run length is 2^melclen, shorter runs are
												encoded by a 1 followed by the run length
												in binary representation, with a fixed length
												of melclen bits. */
	Dgp_UInt32			*melcorder;			/* 2^ melclen */
	/*// */

	/* Tables... */
	Dgp_Int32			*zeroLUT;			/* table to find out number of leading zeros */
	Dgp_Int32			*classmap;
	Dgp_Int32			(*vLUT)[2 * LUTMAX16];
	Dgp_Int32			lutmax;
	Dgp_Int32			*qdiv0, *qdiv,        /* quantization table (division via look-up) */
						*qmul0, *qmul;        /* dequantization table */
	/* statistics tables */
	Dgp_Int32			*N,
						*A,
						*B,
						*C;

	Dgp_Int32			bpp,				/* bits per sample */
						qbpp,				/* bits per sample for quantized prediction errors */
						limit,				/* limit for unary part of Golomb code */
						limit_reduce;		/* reduction on above for EOR states */

	uds					*in, *out;
	uds					*mappingTable;
	uds					multiUds[3];		/* For i/o involving interleaved planes, e.g. (not implemented yet). */
	Dgp_State			*dgp_state;


	/* Decode state... */
	jpeg_ls_header		*head_frame;
	jpeg_ls_header		*head_scan[MAX_SCANS];
	pixel				*pscanl0, 
						*cscanl0;
	pixel				*c_pscanl0[MAX_COMPONENTS],		
						*c_cscanl0[MAX_COMPONENTS];
	Dgp_Int32			components;
	Dgp_Bool			multipleOutputFiles;
	Dgp_Bool			verbose;
	Dgp_Bool			got_lse;			/* got an LSE marker */
	Dgp_Bool			got_restart;		/* got a restart marker indication */
	Dgp_Int32			shift;				/* Shift value for sparse images */

	Dgp_Int32			whose_max_size_columns, 
						whose_max_size_rows;
	Dgp_Int32			number_of_scans;
	Dgp_Int32			eor_limit;			/* Used to be statically defined in lossless_d.c */

	/* Encode state... */
	Dgp_Bool			multipleInputFiles;
	Dgp_Bool			need_lse;			 	/* if we need an LSE marker (non-default params) */
	Dgp_Int32 			need_table;					/* if we need an LSE marker (mapping table) */
	Dgp_Bool 			need_restart;				/* if we need to add restart markers */
	pixel				*scanl0;
	pixel				*scanl1;
	pixel				*c_scanl0[MAX_COMPONENTS];
	pixel			    *c_scanl1[MAX_COMPONENTS];
	Dgp_Int32			application_header;			/* application bytes written in the header (unused? / dhtesq) */
	Dgp_Int32			all_header;				/* all bytes of the header, including application bytes and jpeg-ls bytes */

	/* Common... */
	Dgp_Int32			color_mode;
	Dgp_Int32			restart_interval;	/* the restart interval */
	Dgp_Int32			T1, T2, T3, Ta;
	Dgp_Int32			theReset;			/* Watch FIXRESET preprocessor definition. */
	Dgp_Bool			lossy;				/* Indicates if in lossy mode or not */
	Dgp_Int32			nearLosslessError;
	Dgp_Bool			bpp16;						/* Indicates if 16 bits per pixel mode or not */
	Dgp_Int32			columns, rows;
	Dgp_Int32			samplingx[MAX_COMPONENTS], 
						samplingy[MAX_COMPONENTS];
	pixel				*pscanline, 
						*cscanline;
	pixel				*c_pscanline[MAX_COMPONENTS],	
						*c_cscanline[MAX_COMPONENTS];
	Dgp_Int32			c_columns[MAX_COMPONENTS], 
						c_rows[MAX_COMPONENTS];

	/* alphabet size */
/*#ifndef FIXALPHA */
	Dgp_Int32			alpha,     /* alphabet size */
						ceil_half_alpha; /* ceil(alpha/2) */
	Dgp_Int32			highmask;  /* for powers of 2, a mask for high bits (may not be used / dhtesq.) */
/*#endif */

	/* Lossy parameters... */
	Dgp_Int32			quant, 
						beta, 
						qbeta,
						ceil_half_qbeta,
						negNearLosslessError,
						alpha1eps;
} DgpJpegLsContext;


/****** Function prototypes */

/* lossless_d.c */
ErrCode
lossless_undoscanline(DgpJpegLsContext		*c,
					  pixel					*psl,			/* previous scanline */
					  pixel					*sl,			/* current scanline */
					  Dgp_Int32				no,				/* number of values in it */
					  Dgp_Int32				color);

ErrCode
lossless_undoscanline_pixel(DgpJpegLsContext	*c,
							pixel				*psl,		/* previous scanline */
							pixel				*sl,		/* current scanline */
							Dgp_Int32			no,			/* number of values in it */
							Dgp_Int32			*retval);	

/* lossy_d.c */
ErrCode
lossy_undoscanline(DgpJpegLsContext		*c,
				   pixel				*psl,			/* previous scanline */
				   pixel				*sl,			/* current scanline */
				   Dgp_Int32			no,				/* number of values in it */
				   Dgp_Int32			color);

ErrCode
lossy_undoscanline_pixel(DgpJpegLsContext		*c,	
						 pixel					*psl,		/* previous scanline */
						 pixel					*sl,		/* current scanline */
						 Dgp_Int32				no,			/* number of values in it */
						 Dgp_Int32				*retval);

/* lossless_e.c */
ErrCode
lossless_doscanline( DgpJpegLsContext	*c,
					 pixel				*psl,            /* previous scanline */
					 pixel				*sl,             /* current scanline */
					 Dgp_Int32			no,				 /* number of values in it */
					 Dgp_Int32			color);			 

ErrCode
lossless_doscanline_pixel(DgpJpegLsContext		*c,
						  pixel					*psl,            /* previous scanline */
						  pixel					*sl,             /* current scanline */
						  Dgp_Int32				no);             /* number of values in it */


/* lossy_e.c */
ErrCode
lossy_doscanline(DgpJpegLsContext		*c,
				 pixel					*psl,          /* previous scanline */
				 pixel					*sl,           /* current scanline */
				 Dgp_Int32				no,			   /* number of values in it */
				 Dgp_Int32				color);

ErrCode
lossy_doscanline_pixel(DgpJpegLsContext		*c,
					   pixel				*psl,            /* previous scanline */
					   pixel				*sl,             /* current scanline */
					   Dgp_Int32			no);             /* number of values in it */


/* melcode.c */
ErrCode
init_process_run(DgpJpegLsContext *context,
				 Dgp_Int32 components,
				 Dgp_Int32 maxrun);    /* maxrun is ignored when using MELCODE; kept here for function compatibility. */

ErrCode
close_process_run(DgpJpegLsContext *context);

ErrCode
process_run_dec(DgpJpegLsContext *context,
				Dgp_Int32		lineleft, 
				Dgp_Int32		color,
				Dgp_Int32		*retLineleft);

/* MISSING */
void
process_run(Dgp_Int32			runCount, 
			Dgp_Bool			isEndOfLine, 
			Dgp_Int32			color);
/*//// */

/* initialize.c */
ErrCode
prepareLUTs(Dgp_State		*dgp_state, 
			Dgp_Int32		alpha, 
			Dgp_Int32		lutmax, 
			Dgp_Int32		T1, 
			Dgp_Int32		T2, 
			Dgp_Int32		T3,
			Dgp_Int32		nearLosslessError,
			Dgp_Bool		lossy,
			Dgp_Int32		**classmap,
			Dgp_Int32		((**vLUT)[2 * LUTMAX16]));

ErrCode
prepare_qtables(Dgp_State		*dgp_state, 
				Dgp_Int32		absize, 
				Dgp_Int32		nearLosslessError,
				Dgp_Int32		**qdiv0,
				Dgp_Int32		**qdiv,
				Dgp_Int32		**qmul0,
				Dgp_Int32		**qmul);

ErrCode
init_stats(Dgp_State		*dgp_state, 
		   Dgp_Int32		absize,
		   Dgp_Int32		*A,
		   Dgp_Int32		*B,
		   Dgp_Int32		*C,
		   Dgp_Int32		*N);

/* global.c */
void
set_thresholds(Dgp_Int32 alfa, Dgp_Int32 nearLosslessError, Dgp_Int32 *T1p, Dgp_Int32 *T2p, Dgp_Int32 *T3p);

ErrCode
check_compatibility(Dgp_State *dgp_state, jpeg_ls_header *head_frame, jpeg_ls_header *head_scan, Dgp_Int32 n_s);

/*
These help callers use a uds interface, even if encoding or decoding to memory.
*/

/* Although we have a full uds input and output interface, callers */
/* will probably going to memory most of the time.  Call this before your decode */
/* to get a memory-resident uds.  So:  this is for decoding jpeg-ls into */
/* an image buffer. */
/* */
/* As it turns out the codec only uses the fwrite() in this case. */
ErrCode
DgpJpegLsFOpenImage(Dgp_State			*dgp_state,
					Dgp_Int8			*start,
					Dgp_Int32			width,
					Dgp_Int32			height,
					Dgp_Int32			lpl,
					Dgp_Int32			xOffset,
					Dgp_Int32			xMax,
					Dgp_Int32			yMax,
					Dgp_Int32			bpp,
					uds					*out);

/* Multi-byte quantities in the jls headers appear to be big-endian by construction, */
/* so a straight fread won't work.  Use this.  It only works for <= 4 bytes. */
/*
size_t
DgpJpegLsBigEndianFRead(void *buffer, size_t itemSize, size_t itemCount, uds *in);
*/

ErrCode
DgpJpegLsFCloseImage(Dgp_State		*dgp_state,
					 uds			*out);

/* N.B. This only works if fed one row of samples at a time.  Cannot write fractions of a byte. */
size_t
DgpJpegLsFWriteToImage(void *buffer, size_t itemSize, size_t itemCount, void *stream);

size_t
DgpJpegLs_fwriteToMemory(void *buffer, size_t itemSize, size_t itemCount, void *stream);

size_t
DgpJpegLs_freadFromMemory(void *buffer, size_t itemSize, size_t itemCount, void *stream);

Dgp_Int32
DgpJpegLs_ftellInMemory(void *stream);

Dgp_Int32
DgpJpegLs_feofInMemory(void *stream);

/* decode.c */
ErrCode
DgpJpegLsInitializeDecode(DgpJpegLsContext	**c,
						  Dgp_State			*dgp_state,
						  uds				*in,
						  Dgp_Bool			verbose);

ErrCode
DgpJpegLsDecode (DgpJpegLsContext *c,
				 uds			*out,
				 Dgp_Float64	*decodeTime);


#endif
