/* SPMG/JPEG-LS IMPLEMENTATION V.2.1
   =====================================
   These programs are Copyright (c) University of British Columbia. All rights reserved.
   They may be freely redistributed in their entirety provided that this copyright
   notice is not removed. THEY MAY NOT BE SOLD FOR PROFIT OR INCORPORATED IN
   COMMERCIAL PROGRAMS WITHOUT THE WRITTEN PERMISSION OF THE COPYRIGHT HOLDER.
   Each program is provided as is, without any express or implied warranty,
   without even the warranty of fitness for a particular purpose.

   =========================================================
   THIS SOFTWARE IS BASED ON HP's implementation of jpeg-ls:
   =========================================================

   LOCO-I/JPEG-LS IMPLEMENTATION V.0.90
   -------------------------------------------------------------------------------
   (c) COPYRIGHT HEWLETT-PACKARD COMPANY, 1995-1999.
        HEWLETT-PACKARD COMPANY ("HP") DOES NOT WARRANT THE ACCURACY OR
   COMPLETENESS OF THE INFORMATION GIVEN HERE.  ANY USE MADE OF, OR
   RELIANCE ON, SUCH INFORMATION IS ENTIRELY AT USER'S OWN RISK.
        BY DOWNLOADING THE LOCO-I/JPEG-LS COMPRESSORS/DECOMPRESSORS
   ("THE SOFTWARE") YOU AGREE TO BE BOUND BY THE TERMS AND CONDITIONS
   OF THIS LICENSING AGREEMENT.
        YOU MAY DOWNLOAD AND USE THE SOFTWARE FOR NON-COMMERCIAL PURPOSES
   FREE OF CHARGE OR FURTHER OBLIGATION.  YOU MAY NOT, DIRECTLY OR
   INDIRECTLY, DISTRIBUTE THE SOFTWARE FOR A FEE, INCORPORATE THIS
   SOFTWARE INTO ANY PRODUCT OFFERED FOR SALE, OR USE THE SOFTWARE
   TO PROVIDE A SERVICE FOR WHICH A FEE IS CHARGED.
        YOU MAY MAKE COPIES OF THE SOFTWARE AND DISTRIBUTE SUCH COPIES TO
   OTHER PERSONS PROVIDED THAT SUCH COPIES ARE ACCOMPANIED BY
   HEWLETT-PACKARD'S COPYRIGHT NOTICE AND THIS AGREEMENT AND THAT
   SUCH OTHER PERSONS AGREE TO BE BOUND BY THE TERMS OF THIS AGREEMENT.
        THE SOFTWARE IS NOT OF PRODUCT QUALITY AND MAY HAVE ERRORS OR DEFECTS.
   THE JPEG-LS STANDARD IS STILL UNDER DEVELOPMENT. THE SOFTWARE IS NOT A
   FINAL OR FULL IMPLEMENTATION OF THE STANDARD.  HP GIVES NO EXPRESS OR
   IMPLIED WARRANTY OF ANY KIND AND ANY IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR PURPOSE ARE DISCLAIMED.
        HP SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
   OR CONSEQUENTIAL DAMAGES ARISING OUT OF ANY USE OF THE SOFTWARE.
   -------------------------------------------------------------------------------
*/

/* encoder.c --- the main module, argument parsing, file I/O
 *
 * 
 * Initial code by Alex Jakulin,  Aug. 1995
 *
 * Modified and optimized: Gadiel Seroussi, October 1995
 *
 * Color Enhancement: Guillermo Sapiro, August 1996
 *
 * Modified and added Restart marker and input tables by:
 * David Cheng-Hsiu Chu, and Ismail R. Ismail march 1999
 *
 * 20.3.2001 / dhtesq.
 * Added to digipaper code base for evaluation purposes.
 *
 */

#include "dgpjpegls.h"
#include "string.h"
#include "jpegmark.h"
#include "bitio.h"
#include <math.h>

static Dgp_Int8 *banner="\n\
=============================================\n\
SPMG/JPEG-LS COMPRESSOR " JPEGLSVERSION "\n\
=============================================\n\
These programs are Copyright (c) University of British Columbia.\n\
All rights reserved. They may be freely redistributed in their\n\
entirety provided that this copyright  notice is not removed.\n\
They may not be sold for profit or incorporated in commercial\n\
programs without the written permission of the copyright holder.\n\
Each program is provided as is, without any express or implied\n\
warranty, without even the warranty of fitness for a particular\n\
purpose.\n\
\n\
=========================================================\n\
THIS SOFTWARE IS BASED ON HP's implementation of jpeg-ls:\n\
=========================================================\n\
(c) COPYRIGHT HEWLETT-PACKARD COMPANY, 1995-1999.\n";

/* define color mode strings */
static Dgp_Int8  *plane_int_string = "plane by plane",
				 *line_int_string = "line intlv",
				 *pixel_int_string = "sample intlv";


/* Read one row of pixel values. */
INLINE ErrCode
read_one_line(DgpJpegLsContext			*c,
			  pixel						*line, 
			  Dgp_Int32					cols)
{
	Dgp_UInt8 *line8;
	Dgp_Int32 i;
	Dgp_State *dgp_state = c->dgp_state;
	uds *in = c->in;
	DGP_STATE_PROLOGUE("read_one_line()", dgp_state);

	if (c->bpp16 == DGP_FALSE){
		DGP_CHECK_ALLOC(line8 = (Dgp_UInt8 *)DGP_MALLOC(cols));

		DGP_ERR_GENF0((in->fread((void *)line8, 1, cols, in->fp) != (Dgp_UInt32)cols), cErrIO_ReadError,
			"Input file truncated...");
			
		for(i = 0; i < cols; i++){
			line[i] = line8[i];
		}

		DGP_FREE(line8);
	}
	else{
		DGP_ERR_GENF0((in->fread((void *)line, 2, cols, in->fp) != (Dgp_UInt32)cols), cErrIO_ReadError,
			"Input file truncated...");
	}
	NORMAL_RETURN;
}


/* Initialize the buffers for each line */
ErrCode
initbuffersForEncode(	DgpJpegLsContext		*c,
						Dgp_Int32				comp)
{
    Dgp_Int32	i;
	Dgp_State	*dgp_state = c->dgp_state;
	DGP_STATE_PROLOGUE("initbuffersForEncode()", dgp_state);

    if (c->multipleInputFiles == DGP_TRUE){ 		/* The files are received independent */
		for (i = 0; i < comp; i++) {
			DGP_CHECK_ALLOC(c->c_scanl0[i] = (pixel *)DGP_CALLOC((c->c_columns[i] + LEFTMARGIN + RIGHTMARGIN), sizeof(pixel) ));
			DGP_CHECK_ALLOC(c->c_scanl1[i] = (pixel *)DGP_CALLOC((c->c_columns[i] + LEFTMARGIN + RIGHTMARGIN), sizeof(pixel) ));

		    c->c_pscanline[i] = c->c_scanl0[i] + (LEFTMARGIN - 1);
		    c->c_cscanline[i] = c->c_scanl1[i] + (LEFTMARGIN - 1);
		}
	}
    else { 			/* Only 1 file received */
	
	    DGP_CHECK_ALLOC(c->scanl0 = (pixel *)DGP_CALLOC(c->components * (c->columns + LEFTMARGIN + RIGHTMARGIN), sizeof(pixel) ));
	    DGP_CHECK_ALLOC(c->scanl1 = (pixel *)DGP_CALLOC(c->components * (c->columns + LEFTMARGIN + RIGHTMARGIN), sizeof(pixel) ));

		/*  Adjust scan line pointers taking into account the margins,
			and also the fact that indexing for scan lines starts from 1.
		*/
 	    c->pscanline = c->scanl0 + c->components * (LEFTMARGIN - 1);
	    c->cscanline = c->scanl1 + c->components * (LEFTMARGIN - 1);
    }

	/* MISSING FUNCTION */
    DGP_ERR_PASS(bitoinit(c));

	NORMAL_RETURN;
}


/* Swap the pointers to the current and previous scanlines */
static void 
swaplines(DgpJpegLsContext *c)
{
	pixel *temp;
	temp = c->pscanline;
	c->pscanline = c->cscanline;
	c->cscanline = temp;
	return;
}


/* Swap the pointers to the current and previous scanlines */
static void 
c_swaplines(DgpJpegLsContext *c, Dgp_Int32 i)
{
	pixel *temp;
	temp = c->c_pscanline[i];
	c->c_pscanline[i] = c->c_cscanline[i];
	c->c_cscanline[i] = temp;
	return;
}


/* close the line buffers */
/* Formerly "closebuffers()"... */
ErrCode
DgpJpegLsEncodeCleanup(DgpJpegLsContext		*c,
					   Dgp_Int32			*pos){
	Dgp_Int32 localPos, i;
	Dgp_State *dgp_state = c->dgp_state;
	DGP_STATE_PROLOGUE("DgpJpegLsEncodeCleanup()", dgp_state);
	DGP_ASSERT(pos);

	/* MISSING FUNCTION */
	bitoflush();

	DGP_FREE(c->qdiv0);
	c->qdiv0 = NULL;
	DGP_FREE(c->qmul0);
	c->qmul0 = NULL;

	DGP_FREE(c->vLUT);
	c->vLUT = NULL;

	DGP_FREE(c->classmap);
	c->classmap = NULL;

	DGP_FREE(c->head_frame);
	c->head_frame = NULL;

	for (i = 0; i < MAX_SCANS; i++) {
		DGP_FREE(c->head_scan[i]);
		c->head_scan[i] = NULL;
	}

	if (c->multipleInputFiles == DGP_FALSE){
		DGP_FREE(c->scanl0);
		DGP_FREE(c->scanl1);
	}
    else{
		for (i = 0; i < c->components; i++){
			DGP_FREE(c->c_scanl0[i]);
			DGP_FREE(c->c_scanl1[i]);
		}
	}

	localPos = c->out->ftell(c->out->fp);

	/* free, free */
	DGP_FREE(c->A);
	c->A = NULL;
	DGP_FREE(c->B);
	c->B = NULL;
	DGP_FREE(c->C);
	c->C = NULL;
	DGP_FREE(c->N);
	c->N = NULL;
	DGP_ERR_PASS(close_process_run(c));
	DGP_FREE(c);

	*pos = localPos;
	NORMAL_RETURN;
}


/* Input and output uds should be initialized with valid file pointers before this call, at api level. */
ErrCode
DgpJpegLsInitializeEncode(DgpJpegLsContext	*c,
						  uds				*in,
						  uds				*out,
						  Dgp_UInt32		*inputImage,
						  Dgp_Int32			width,
						  Dgp_Int32			height,
						  Dgp_Int32			longsPerRow,
						  Dgp_Int32			xOffset,
						  Dgp_Int32			xMax,
						  Dgp_Int32			yMax,
						  Dgp_Int32			bitDepth,
						  Dgp_Int32			components,
						  Dgp_Int32			*restartInterval,
						  uds				*mappingTable,
						  Dgp_Int32			*desiredReset,
						  Dgp_Int32			*color_mode,
						  Dgp_Int32			*nearLosslessError,
						  Dgp_Int32			*T1,
						  Dgp_Int32			*T2,
						  Dgp_Int32			*T3,
						  Dgp_Bool			verbose)
{
	Dgp_Int8 *color_mode_string;
	Dgp_State *dgp_state = c->dgp_state;
	Dgp_Int32	i, 
				n_c = 0, 
				common_rows, 
				common_alpha,
				min_size_rows, 
				min_size_columns,
				alpha0,
				gotinf = 0,
				gotoutf = 0;
	DGP_STATE_PROLOGUE("DgpJpegLsInitializeEncode()", dgp_state);
	DGP_ASSERT(in);
	DGP_ASSERT(out);

	c->in = in;
	c->out = out;
	c->mappingTable = mappingTable;
	c->multipleInputFiles = DGP_FALSE;

	c->color_mode = DEFAULT_COLOR_MODE;
	if(color_mode){
		c->color_mode = *color_mode;
	}
	c->need_table = DGP_FALSE;
	if(mappingTable){
		c->need_table = DGP_TRUE;
	}
	c->need_restart = DGP_FALSE;
	c->restart_interval = 0;
	if(restartInterval){
		c->need_restart = DGP_TRUE;
		c->restart_interval = *restartInterval;
	}
	c->components = components;
	c->T1 = c->T2 = c->T3 = 0;
	if(T1){
		c->need_lse = DGP_TRUE;
		c->T1 = *T1;
	}
	if(T2){
		c->need_lse = DGP_TRUE;
		c->T2 = *T2;
	}
	if(T3){
		c->need_lse = DGP_TRUE;
		c->T3 = *T3;
	}
	/* If the Tn parameters are zero at this point, they'll get set in set_thresholds(). */

	c->verbose = verbose;

/*
#ifndef FIXRESET
	RESET=DEFAULT_RESET;
#endif
*/
	c->need_lse = DGP_FALSE;
	c->theReset = DEFAULT_RESET;
	if(desiredReset){
		c->need_lse = DGP_TRUE;
		c->theReset = *desiredReset;
#ifdef FIXRESET
		DGP_ERR_GENF2(c->theReset != DEFAULT_RESET, cErrInvalidParm,
			"This version was compiled for a fixed reset parameter (%d), but I got %d.", DEFAULT_RESET, c->theReset);
#endif
	}
	
	/* Initialize nearLosslessError to zero and loss-less mode */
	c->nearLosslessError = DEF_NEAR;
	c->lossy = DGP_FALSE;
	if(nearLosslessError){
		c->lossy = DGP_TRUE;
		c->nearLosslessError = *nearLosslessError;
	}

	/* Check that color mode is valid and pick color mode string. */
	switch ( c->color_mode ) {
	    case PLANE_INT:
			color_mode_string = plane_int_string;
			c->multipleInputFiles = DGP_TRUE;
			break;
	    case LINE_INT:
			color_mode_string = line_int_string;
			if (c->components > 1) {
				c->multipleInputFiles = DGP_TRUE;
			}
			break;
	    case PIXEL_INT:
			color_mode_string = pixel_int_string;
			DGP_ERR_GENF1((c->components > 1), cErrInvalidParm,
				"You specified %d input components in pixel interleaved mode.", c->components);
			break;
	    default:
			DGP_ERR_GENF1(DGP_TRUE, cErrInvalidParm,
				"Invalid color mode (%d).", c->color_mode);
			break;
	}

#define LESS_CONTEXTS 1

	DGP_REPORTF3(verbose, "Number of contexts (non-run): %d regular + %d EOR = %d.", 
		CONTEXTS - LESS_CONTEXTS, EOR_CONTEXTS, TOT_CONTEXTS - LESS_CONTEXTS);
	
	/* Read image headers... */
	if (c->multipleInputFiles == DGP_FALSE) {
		/* Unlikely this case will be used in our implementation, but the basics are all here... / dhtesq. */
		c->columns = width;
		c->rows = height;
		alpha0 = bitDepth;
		/* A single component => PLANE_INT. */
		if ( (c->color_mode == LINE_INT || c->color_mode == PIXEL_INT) && c->components == 1) {	
			DGP_REPORTF0(verbose, "Since you only gave me a single component, I changed the color mode to plane-interleaved.");
			c->color_mode = PLANE_INT;
			color_mode_string = plane_int_string;
			c->multipleInputFiles = DGP_TRUE;
			c->c_columns[0] = c->columns;
			c->c_rows[0] = c->rows;
		}
	}
	else {
		c->columns = width;
		c->rows = height;
		alpha0 = bitDepth;
		common_rows = c->rows;
		common_alpha = alpha0;

		/* Should check for same bit depth in all components...can stick that in */
		/* if we decide to stick the poor user with passing in all that stuff. */
		for (i = 0; i < c->components; i++) {
			c->c_columns[i] = width;
			c->c_rows[i] = height;
		}
	}

#ifdef FIXALPHA
	/*alpha0++; */
	DGP_ERR_GENF2(( alpha0 != c->alpha ), cErrInvalidParm,
		"This version has been optimized for alphabet size %d, but I got %d.", c->alpha, alpha0);
#else
	c->alpha = alpha0;/* + 1;  /* number read from file header is alpha - 1 */ */
	c->ceil_half_alpha = (c->alpha + 1) / 2;
#endif

#ifdef POW2
	c->highmask = -c->alpha;
	/* check that alpha is a power of 2 */
	for ( alpha0 = c->alpha, i = -1; alpha0; alpha0 >>= 1, i++);
	DGP_ERR_GENF1(( c->alpha != (1 << i) ), cErrInvalidParm,
		"This version has been optimized for a power-of-2 alphabet size, but I got %d.", c->alpha);
#endif	

	/* Check for 16 or 8 bit mode */
	if (c->alpha <= MAXA16 && c->alpha > MAXA8){
		c->bpp16 = DGP_TRUE;
		c->lutmax = LUTMAX16;
	}
	else if (c->alpha <= MAXA8 && c->alpha >= 1){
		c->bpp16 = DGP_FALSE;
		c->lutmax = LUTMAX8;
	}
	else { 
		DGP_ERR_GENF1(DGP_TRUE, cErrInvalidParm,
			"I got %d for alpha, which I can't handle.", c->alpha);
	}

	/* print out parameters */
	DGP_REPORTF4(verbose && (c->multipleInputFiles == DGP_FALSE), "Image:  %d cols, %d rows, %d alpha, %d comp.",
		c->columns, c->rows, c->alpha, c->components);
	DGP_REPORTF2(verbose && (c->multipleInputFiles == DGP_FALSE), "Image:  color mode %d (%s).", c->color_mode, color_mode_string);
		
	if(c->verbose && c->multipleInputFiles == DGP_TRUE){
		for (i = 0; i < c->components; i++){
			DGP_REPORTF3(DGP_TRUE, "Image %d:\t%d cols, %d rows.", i, c->c_columns[i], c->c_rows[i]);
		}
		DGP_REPORTF4(DGP_TRUE, "alpha =  %d, comp = %d, color_mode = %d (%s).", alpha0, c->components, c->color_mode, color_mode_string);
		DGP_REPORTF4(DGP_TRUE, "Parameters:\tT1 = %d, T2 = %d, T3 = %d, RESET = %d.", c->T1, c->T2, c->T3, c->theReset);
	}		

	/* compute auxiliary parameters for near-lossless (globals) */
	if (c->lossy == DGP_TRUE) {
		c->quant = 2 * c->nearLosslessError + 1;
		c->qbeta = (c->alpha + 2 * c->nearLosslessError + c->quant - 1 ) / c->quant;
		c->beta = c->quant * c->qbeta;
		c->ceil_half_qbeta = (c->qbeta + 1) / 2;
		c->negNearLosslessError = -c->nearLosslessError;
		DGP_REPORTF3(verbose, "Lossy mode: nearLosslessError = %d, beta = %d, qbeta = %d.", c->nearLosslessError, c->beta, c->qbeta);
	}

	/* compute bits per sample for input symbols */
	for ( c->bpp = 1; (1L << c->bpp) < c->alpha; c->bpp++ );

	/* check if alpha is a power of 2: */
	if ( c->alpha != (1 << c->bpp) ){
	    c->need_lse = DGP_TRUE; /*	If not, MAXVAL will be non-default, and we'll need to specify it in an LSE marker. */
	}

	/* compute bits per sample for unencoded prediction errors */
	if (c->lossy == DGP_TRUE){
		for ( c->qbpp = 1; (1L << c->qbpp) < c->qbeta; c->qbpp++ );
	}
	else{
		c->qbpp = c->bpp;
	}

	if ( c->bpp < 2 ) c->bpp = 2;
	
	/* limit for unary part of Golomb code */
	if ( c->bpp < 8 ){
	    c->limit = 2 * (c->bpp + 8) - c->qbpp - 1;
	}
	else{
	    c->limit = 4 * c->bpp - c->qbpp - 1;
	}

	/* Check for smallest subsampled file and compute the sampling. */
	if ((c->components > 1) && (c->multipleInputFiles)) {
		min_size_columns = c->c_columns[c->components - 1];
		min_size_rows = c->c_rows[c->components - 1];
		for (i = 0; i< c->components - 1; i++) {
			if (c->c_columns[i] < min_size_columns) min_size_columns = c->c_columns[i];
			if (c->c_rows[i] < min_size_rows) min_size_rows = c->c_rows[i];
		}

		for (i = 0; i < components; i++) {
			c->samplingx[i] = c->c_columns[i] / min_size_columns;
			c->samplingy[i] = c->c_rows[i] / min_size_rows;
			DGP_ERR_GENF0(((c->samplingx[i] > 4) || ((c->c_columns[i] % min_size_columns) != 0)), cErrInvalidParm,
				"Image sizes not compatible.");
			DGP_ERR_GENF0(((c->samplingy[i] > 4) || ((c->c_rows[i] % min_size_rows) != 0)), cErrInvalidParm,
				"Image sizes not compatible.");
		}

		min_size_columns = c->c_columns[0];
		c->whose_max_size_columns = 0;
		min_size_rows = c->c_rows[0];
		c->whose_max_size_rows = 0;

		for (i = 1; i < c->components; i++) { 
			if (c->c_columns[i] > min_size_columns) {
				c->whose_max_size_columns = i;
				min_size_columns = c->c_columns[i];
			}
			if (c->c_rows[i] > min_size_rows) {
				c->whose_max_size_rows = i;
				min_size_rows = c->c_rows[i];
			}
		}
	}
	else {
		for (i = 0; i < c->components; i++) c->samplingx[i] = c->samplingy[i] = 1;
	}

	/* Allocate memory pools. */
	DGP_ERR_PASS(initbuffersForEncode(c, c->components));
	NORMAL_RETURN;
}


/* Main loop for decoding files */
/* Not all cases implemented:  I'm concentrating on pixel-interleaved input (for more than one component) */
/* from a single file or buffer. / dhtesq. */
/* Input and output uds should be initialized with valid file pointers before this call, at api level. */
ErrCode
DgpJpegLsEncode(Dgp_State				*dgp_state,
				uds						*in,
				uds						*out,
				Dgp_UInt32				*inputImage,
				Dgp_Int32				width,
				Dgp_Int32				height,
				Dgp_Int32				longsPerRow,
				Dgp_Int32				xOffset,
				Dgp_Int32				xMax,
				Dgp_Int32				yMax,
				Dgp_Int32				bitDepth,
				Dgp_Int32				components,
				Dgp_Int32				*restartInterval,
				uds						*mappingTable,
				Dgp_Int32				*desiredReset,
				Dgp_Int32				*color_mode,
				Dgp_Int32				*nearLosslessError,
				Dgp_Int32				*T1,
				Dgp_Int32				*T2,
				Dgp_Int32				*T3,
				Dgp_Bool				verbose,
				Dgp_Float64				*encodeTime)
{
    Dgp_Int32 n, n_c, n_r, my_i, number_of_scans, n_s, i;
	Dgp_Float64 t0, t1, get_utime();
	Dgp_Int32 tot_in = 0, tot_out = 0, pos0, pos1, temp;
	Dgp_Int32 temp_columns;
	Dgp_Int32 MCUs_counted;
	DgpJpegLsContext *c = NULL;
	pixel *local_scanl0, *local_scanl1, *local_pscanline, *local_cscanline;
	DGP_STATE_PROLOGUE("DgpJpegLsEncode()", dgp_state);
	DGP_ASSERT(in);


	DGP_CHECK_ALLOC(c = (DgpJpegLsContext *)DGP_MALLOC(sizeof(*c)));
	c->shift = 0;			/* Not implemented... */
	c->all_header = c->application_header = 0;
	c->A = c->B = c->C = c->N = NULL;

	/* Parse the parameters, initialize... */
	DGP_ERR_PASS_CLEANUP(DgpJpegLsInitializeEncode(c,
												   in,
												   out,
												   inputImage,
												   width,
												   height,
												   longsPerRow,
												   xOffset,
												   xMax,
												   yMax,
												   bitDepth,
												   components,
												   restartInterval,
												   mappingTable,
												   desiredReset,
												   color_mode,
												   nearLosslessError,
												   T1,
												   T2,
												   T3,
												   verbose));

	/* Start timer (must be AFTER initialize()) */
	t0 = get_utime();   

	/* Compute the number of scans */
	/* Multiple scans only for PLANE_INT in this implementation. */

	if (c->color_mode == PLANE_INT){
		number_of_scans = c->components;
	}
	else{
		number_of_scans = 1;
	}

	c->head_frame = NULL;
	for (n_s = 0; n_s < number_of_scans; n_s++){
		c->head_scan[n_s] = NULL;
	}

	/* Write the frame header - allocate memory for jpegls header. */
	DGP_CHECK_ALLOC_CLEANUP(c->head_frame = (jpeg_ls_header *)DGP_CALLOC(1, sizeof(jpeg_ls_header)));
	for (n_s = 0; n_s < number_of_scans; n_s++){
		DGP_CHECK_ALLOC_CLEANUP(c->head_scan[n_s] = (jpeg_ls_header *)DGP_CALLOC(1, sizeof(jpeg_ls_header)));
	}

	/* Assigns columns/rows to head_frame */
	if (c->multipleInputFiles == DGP_FALSE) {
		c->head_frame->columns = c->columns;
		c->head_frame->rows = c->rows;
	}
	else {
		c->head_frame->columns = c->c_columns[c->whose_max_size_columns];
		c->head_frame->rows = c->c_rows[c->whose_max_size_rows];
	}

	c->head_frame->alp = c->alpha;
	c->head_frame->comp = c->components;

	/* Assign component id and samplingx/samplingy. */
	for (i = 0; i < c->components; i++) {
		c->head_frame->comp_ids[i] = i + 1;
		c->head_frame->samplingx[i] = c->samplingx[i];
		c->head_frame->samplingy[i] = c->samplingy[i];
	}

	c->head_frame->nearLosslessError = c->nearLosslessError; /* Not needed, scan information */
	c->head_frame->need_lse = (Dgp_Int32)c->need_lse; /* Not needed, for completeness  */
	c->head_frame->color_mode = c->color_mode; /* Not needed, scan information */
	c->head_frame->shift = c->shift; /* Not needed, scan information */

	for (n_s = 0; n_s < number_of_scans; n_s++) {
		c->head_scan[n_s]->alp = c->alpha;
		c->head_scan[n_s]->nearLosslessError = c->nearLosslessError;
		c->head_scan[n_s]->T1 = c->T1;
		c->head_scan[n_s]->T2 = c->T2;
		c->head_scan[n_s]->T3 = c->T3;
		c->head_scan[n_s]->RES = c->theReset;
		c->head_scan[n_s]->shift = c->shift;
		c->head_scan[n_s]->color_mode = c->color_mode;
	}

	if (c->color_mode == PLANE_INT) { /* One plane per scan */
		for (n_s = 0; n_s < number_of_scans; n_s++) {
			c->head_scan[n_s]->comp = 1;
			c->head_scan[n_s]->comp_ids[0] = n_s + 1;
		}
	}
	else {
		for (n_s = 0; n_s < number_of_scans; n_s++) {
			c->head_scan[n_s]->comp = c->head_frame->comp;
			for (n_c = 0; n_c < c->head_frame->comp; n_c++){
				c->head_scan[n_s]->comp_ids[n_c] = n_c + 1;
			}
		}
	}
	
	/* Write SOI */
	c->all_header = write_marker(out, SOI);

	/* Write the frame */
	DGP_ERR_PASS_CLEANUP(write_jpegls_frame(dgp_state, out, c->head_frame, &temp));
	c->all_header += temp;

	/* End of frame header writing */

	if ((c->components > 1) && (c->multipleInputFiles == DGP_FALSE)) {
		/* Received PPM file, allocate auxiliary buffers. */

		DGP_CHECK_ALLOC_CLEANUP(local_scanl0 = (pixel *)DGP_CALLOC(c->columns + LEFTMARGIN + RIGHTMARGIN, sizeof(pixel) ));
		DGP_CHECK_ALLOC_CLEANUP(local_scanl1 = (pixel *)DGP_CALLOC(c->columns + LEFTMARGIN + RIGHTMARGIN, sizeof(pixel) ));

		local_pscanline = local_scanl0 + LEFTMARGIN - 1;
		local_cscanline = local_scanl1 + LEFTMARGIN - 1;
	}


	/* Go through each scan and process line by line */
	for (n_s = 0; n_s < number_of_scans; n_s++) {
		
		/* process scans one by one */

		if (n_s == 0) {
			/* The thresholds for the scan. Must re-do per scan is change. */
			set_thresholds(c->alpha, c->nearLosslessError, &c->T1, &c->T2, &c->T3);
			for (i = 0; i < number_of_scans; i++) {
				c->head_scan[n_s]->T1 = c->T1;
				c->head_scan[n_s]->T2 = c->T2;
				c->head_scan[n_s]->T3 = c->T3;
			}

			/* After the thresholds are set, write LSE marker if we have */
			/* non-default parameters or if we need a mapping table. */
			if ( c->need_lse == DGP_TRUE ){
				c->all_header += write_jpegls_extmarker(out, c->head_scan[n_s], LSE_PARAMS, c->mappingTable);
			}
			if ( c->need_table == DGP_TRUE ){
				c->all_header += write_jpegls_extmarker(out, c->head_scan[n_s], LSE_MAPTABLE, c->mappingTable);
			}

			/* If using restart markers, write the DRI header */
			if ( c->need_restart == DGP_TRUE ){	
				c->head_scan[n_s]->restart_interval = c->restart_interval;
				/* write_jpegls_restartmarker() was a missing function, reconstructed for this implementation. / dhtesq. */
				c->all_header += write_jpegls_restartmarker(out, c->head_scan[n_s]);
			}			

			/* Print out parameters. */
			DGP_REPORTF5(verbose, "Parameters:  T1 = %d, T2 = %d, T3 = %d, RESET = %d, limit = %d.", c->T1, c->T2, c->T3, c->theReset, c->limit);

			/* Prepare LUTs for context quantization. */
			/* Must re-do when thresholds change. */
			DGP_ERR_PASS_CLEANUP(prepareLUTs(	dgp_state,
												c->alpha,
												c->lutmax,
												c->T1,
												c->T2,
												c->T3,
												c->nearLosslessError,
												c->lossy,
												&c->classmap,
												&c->vLUT));

			if (c->lossy == DGP_TRUE){
				/* prepare div/mul tables for near-lossless quantization */
				DGP_ERR_PASS_CLEANUP(prepare_qtables(	dgp_state,
														c->alpha, 
														c->nearLosslessError,
														&c->qdiv0,
														&c->qdiv,
														&c->qmul0,
														&c->qmul));
			}

			/* Check for errors */
			DGP_ERR_PASS_CLEANUP(check_compatibility(dgp_state, c->head_frame, c->head_scan[0], 0));
		}

		/* Restart Marker is reset after every scan */
		MCUs_counted = 0;

		/* Write the scan header */
		DGP_ERR_PASS_CLEANUP(write_jpegls_scan(dgp_state, out, c->head_scan[n_s], &c->all_header));
		pos0 = out->ftell(out->fp);  /* position in output file, after header */

		/* Initializations for each scan */
		/* Start from 1st image row */
		n = 0;

		/* initialize stats arrays */
		DGP_CHECK_ALLOC_CLEANUP(c->A = (Dgp_Int32 *)DGP_MALLOC(sizeof(*c->A) * TOT_CONTEXTS));
		DGP_CHECK_ALLOC_CLEANUP(c->B = (Dgp_Int32 *)DGP_MALLOC(sizeof(*c->B) * TOT_CONTEXTS));
		DGP_CHECK_ALLOC_CLEANUP(c->C = (Dgp_Int32 *)DGP_MALLOC(sizeof(*c->C) * TOT_CONTEXTS));
		DGP_CHECK_ALLOC_CLEANUP(c->N = (Dgp_Int32 *)DGP_MALLOC(sizeof(*c->N) * TOT_CONTEXTS));
		DGP_ERR_PASS_CLEANUP(init_stats(dgp_state, c->lossy == DGP_TRUE ? c->qbeta:c->alpha, c->A, c->B, c->C, c->N));

		/* initialize run processing */
		DGP_ERR_PASS_CLEANUP(init_process_run(c, c->components, MAXRUN));

		if (c->color_mode == LINE_INT) {	/* line interleaved */
			if (c->multipleInputFiles = DGP_FALSE) { 			/* Single file received */
				/***********************************************************************/
				/*           Line interleaved mode with single file received           */
				/***********************************************************************/

				if (c->lossy == DGP_FALSE){

					/* LOSSLESS mode */
					while (++n <= c->rows) {

						DGP_ERR_PASS_CLEANUP(read_one_line(c, c->cscanline + components, c->components * c->columns));

						tot_in += c->components * c->columns;

						/* 'extend' the edges */
						for (n_c = 0; n_c < c->components; n_c++){
							c->cscanline[-c->components + n_c] = c->cscanline[n_c] = c->pscanline[c->components + n_c];
						}

						for (n_c = 0; n_c < c->components; n_c++) {
							if (c->components > 1) {
								for (my_i = 0; my_i < c->columns + LEFTMARGIN + RIGHTMARGIN; my_i++){
									local_cscanline[-1 + my_i] = c->cscanline[-c->components + my_i * c->components + n_c];
									local_pscanline[-1 + my_i] = c->pscanline[-c->components+ my_i * c->components + n_c];
								}
							}
							else {
								local_cscanline = c->cscanline;
								local_pscanline = c->pscanline;
							}

							/* process the lines */
							DGP_ERR_PASS_CLEANUP(lossless_doscanline(c, local_pscanline, local_cscanline, c->columns, n_c));
						}

						/* 'extend' the edges */
						for (n_c = 0; n_c < c->components; n_c++){
							c->cscanline[c->components * (c->columns + 1) + n_c] = c->cscanline[c->components * c->columns + n_c];
						}

						/* make the current scanline the previous one */
						swaplines(c);

						/* Insert restart markers if enabled */
						if (c->need_restart){
							/* Insert restart markers only after a restart interval. */
							if ((MCUs_counted % c->restart_interval) == 0){
								/* MISSING FUNCTION */
								bitoflush();
								temp = write_marker(out, (RSTm + ((MCUs_counted / c->restart_interval) % 8)));
							}
							MCUs_counted++;
						}
					}
				} /* end lossless clause */

				else{

					/* LOSSY mode */
					while (++n <= c->rows) {

						DGP_ERR_PASS_CLEANUP(read_one_line(c, c->cscanline + c->components, c->components * c->columns));

						tot_in += c->components * c->columns;

						/* 'extend' the edges */
						for (n_c = 0; n_c < c->components; n_c++){
							c->cscanline[-c->components + n_c] = c->cscanline[n_c] = c->pscanline[c->components + n_c];
						}

						for (n_c = 0; n_c < c->components; n_c++) {
							if (c->components > 1) {
								for (my_i = 0; my_i < c->columns + LEFTMARGIN + RIGHTMARGIN; my_i++){
									local_cscanline[-1 + my_i] = c->cscanline[-c->components + my_i * c->components + n_c];
									local_pscanline[-1 + my_i] = c->pscanline[-c->components + my_i * c->components + n_c];
								}
							}
							else {
								local_cscanline = c->cscanline;
								local_pscanline = c->pscanline;
							}

							/* process the lines */
							DGP_ERR_PASS_CLEANUP(lossy_doscanline(c, local_pscanline, local_cscanline, c->columns, n_c));

							if (c->components > 1){
								for (my_i = 0; my_i < c->columns + LEFTMARGIN + RIGHTMARGIN; my_i++){
									c->cscanline[-c->components + my_i * c->components + n_c] = local_cscanline[-1 + my_i];
								}
							}
						}

						/* 'extend' the edges */
						for (n_c = 0; n_c < c->components; n_c++){
							c->cscanline[c->components * (c->columns + 1) + n_c] = c->cscanline[c->components * c->columns + n_c];
						}

						/* make the current scanline the previous one */
						swaplines(c);

						/* Insert restart markers if enabled */
						if (c->need_restart){
							/* Insert restart markers only after a restart interval */
							if ((MCUs_counted % c->restart_interval) == 0){
								/* MISSING FUNCTION */
								bitoflush();
								temp = write_marker(out, (RSTm + ((MCUs_counted / c->restart_interval) % 8)));
							}
							MCUs_counted++;
						}
					}
				} /* end lossy clause */
			} /* end single file */
			else {	/* Multiple files */
				/* color_mode == LINE_INT and c->multipleInputFiles == DGP_TRUE  */

				/***********************************************************************/
				/*           Line interleaved mode with multiple files received        */
				/***********************************************************************/
				n++;

				if (c->lossy == DGP_FALSE){

					/* LOSSLESS mode */
					while (n <= c->c_rows[c->whose_max_size_rows]) {

						for (n_c = 0; n_c < c->components; n_c++) {
							for (n_r = 0; n_r < c->samplingy[n_c]; n_r++) {

								DGP_ERR_PASS_CLEANUP(read_one_line(c, c->c_cscanline[n_c] + 1, c->c_columns[n_c]));

								tot_in += c->c_columns[n_c];

								/* 'extend' the edges */
								c->c_cscanline[n_c][-1] = c->c_cscanline[n_c][0] = c->c_pscanline[n_c][1];

								/* process the lines */
								DGP_ERR_PASS_CLEANUP(lossless_doscanline(c, c->c_pscanline[n_c], c->c_cscanline[n_c], c->c_columns[n_c], n_c));

								/* 'extend' the edges */
								c->c_cscanline[n_c][c->c_columns[n_c] + 1] = c->c_cscanline[n_c][c->c_columns[n_c]];

								/* make the current scanline the previous one */
								c_swaplines(c, n_c);
							}
						}	/* End of loop for each component. */

						n += c->samplingy[c->whose_max_size_rows];

						/* Insert restart markers if enabled. */
						if (c->need_restart){
							/* Insert restart markers only after a restart interval */
							if ((MCUs_counted % c->restart_interval) == 0){
								/* MISSING FUNCTION */
								bitoflush();
								temp = write_marker(out, (RSTm + ((MCUs_counted / c->restart_interval) % 8)));
							}
							MCUs_counted++;
						}
					}	/* End while of LINE_INT with multiple input files */
				}  /* end lossless clause */
				else{

					/* LOSSY mode */
					while (n <= c->c_rows[c->whose_max_size_rows]) {

						for (n_c = 0; n_c < c->components; n_c++) {
							for (n_r = 0; n_r < c->samplingy[n_c]; n_r++) {
								
								DGP_ERR_PASS_CLEANUP(read_one_line(c, c->c_cscanline[n_c] + 1, c->c_columns[n_c]));

								tot_in += c->c_columns[n_c];

								/* 'extend' the edges */
								c->c_cscanline[n_c][-1] = c->c_cscanline[n_c][0] = c->c_pscanline[n_c][1];

								/* process the lines */
								DGP_ERR_PASS_CLEANUP(lossy_doscanline(c, c->c_pscanline[n_c], c->c_cscanline[n_c], c->c_columns[n_c], n_c));

								/* 'extend' the edges */
								c->c_cscanline[n_c][c->c_columns[n_c] + 1] = c->c_cscanline[n_c][c->c_columns[n_c]];

								/* make the current scanline the previous one */
								c_swaplines(c, n_c);
							}
						}	/* End of loop for each file */

						n += c->samplingy[c->whose_max_size_rows];

						/* Insert restart markers if enabled */
						if (c->need_restart){
							/* Insert restart markers only after a restart interval */
							if ((MCUs_counted % c->restart_interval) == 0){
								/* MISSING FUNCTION */
								bitoflush();
								temp = write_marker(out, (RSTm + ((MCUs_counted / c->restart_interval) % 8)));
							}
							MCUs_counted++;
						}
					}	/* End while of LINE_INT with multiple input files */
				}  /* end lossy clause */
			}	/* Closes the else, LINE_INT and c->multipleInputFiles */
		}	/* Closes part for color_mode = LINE_INT */
		else {
			if (c->color_mode == PIXEL_INT) {
				/***********************************************************************/
				/*           Pixel interleaved mode with single file received          */
				/***********************************************************************/

				if (c->lossy == DGP_FALSE){

					/* LOSSLESS mode */
					while (++n <= c->rows) {

						DGP_ERR_PASS_CLEANUP(read_one_line(c, c->cscanline + c->components, c->components * c->columns));

						tot_in += c->components * c->columns;

						/* 'extend' the edges */
						for (n_c = 0; n_c < c->components; n_c++){
							c->cscanline[-c->components + n_c] = c->cscanline[n_c] = c->pscanline[c->components + n_c];
						}

						/* process the lines */
						DGP_ERR_PASS_CLEANUP(lossless_doscanline_pixel(c, c->pscanline, c->cscanline, c->components * c->columns));

						/* 'extend' the edges */
						for (n_c = 0; n_c < c->components; n_c++){
							c->cscanline[c->components * (c->columns + 1) + n_c] = c->cscanline[c->components * c->columns + n_c];
						}

						/* make the current scanline the previous one */
						swaplines(c);

						/* Insert restart markers if enabled */
						if (c->need_restart){
							/* Insert restart markers only after a restart interval */
							if ((MCUs_counted % c->restart_interval) == 0){
								/* MISSING FUNCTION */
								bitoflush();
								temp = write_marker(out, (RSTm + ((MCUs_counted / c->restart_interval) % 8)));
							}
							MCUs_counted++;
						}
					}
				}  /* end lossless clause */

				else{

					/* LOSSY mode */
					while (++n <= c->rows) {

						DGP_ERR_PASS_CLEANUP(read_one_line(c, c->cscanline + c->components, c->components * c->columns));

						tot_in += c->components * c->columns;

						/* 'extend' the edges */
						for (n_c = 0; n_c < c->components; n_c++){
							c->cscanline[-c->components + n_c] = c->cscanline[n_c] = c->pscanline[c->components + n_c];
						}

						/* process the lines */
						DGP_ERR_PASS_CLEANUP(lossy_doscanline_pixel(c, c->pscanline, c->cscanline, c->components * c->columns));

						/* 'extend' the edges */
						for (n_c = 0; n_c < components; n_c++){
							c->cscanline[c->components * (c->columns + 1) + n_c] = c->cscanline[c->components * c->columns + n_c];
						}

						/* make the current scanline the previous one */
						swaplines(c);

						/* Insert restart markers if enabled */
						if (c->need_restart){
							/* Insert restart markers only after a restart interval */
							if ((MCUs_counted % c->restart_interval) == 0){
								/* MISSING FUNCTION */
								bitoflush();
								temp = write_marker(out, (RSTm + ((MCUs_counted / c->restart_interval) % 8)));
							}
							MCUs_counted++;
						}
					}
				}  /* end lossy clause */
			}	/* Closes if PIXEL_INT */

			else {	/* NON PIXEL_INT */
				/***********************************************************************/
				/*           Plane interleaved mode 								   */
				/***********************************************************************/

				if (c->lossy == DGP_FALSE){
				
					/* LOSSLESS mode */
					while (++n <= c->c_rows[n_s]) {
						
						temp_columns = c->c_columns[n_s];

						DGP_ERR_PASS_CLEANUP(read_one_line(c, c->c_cscanline[n_s] + 1, temp_columns));

						tot_in += temp_columns;

						/* 'extend' the edges */
						c->c_cscanline[n_s][-1] = c->c_cscanline[n_s][0] = c->c_pscanline[n_s][1];

						/* process the lines */
						DGP_ERR_PASS_CLEANUP(lossless_doscanline(c, c->c_pscanline[n_s], c->c_cscanline[n_s], temp_columns, n_s));

						/* 'extend' the edges */
						c->c_cscanline[n_s][temp_columns + 1] = c->c_cscanline[n_s][temp_columns];

						/* make the current scanline the previous one */
						c_swaplines(c, n_s);

						/* Insert restart markers if enabled */
						if (c->need_restart){
							/* Insert restart markers only after a restart interval */
							if ((MCUs_counted % c->restart_interval) == 0){
								/* MISSING FUNCTION */
								bitoflush();
								temp = write_marker(out, (RSTm + ((MCUs_counted / c->restart_interval) % 8)));
							}
							MCUs_counted++;
						}
					}
				}

				else{

					/* LOSSY mode */
					while (++n <= c->c_rows[n_s]) {

						temp_columns = c->c_columns[n_s];

						DGP_ERR_PASS_CLEANUP(read_one_line(c, c->c_cscanline[n_s] + 1, temp_columns));

						tot_in += temp_columns;

						/* 'extend' the edges */
						c->c_cscanline[n_s][-1] = c->c_cscanline[n_s][0] = c->c_pscanline[n_s][1];

						/* process the lines */
						DGP_ERR_PASS_CLEANUP(lossy_doscanline(c, c->c_pscanline[n_s], c->c_cscanline[n_s], temp_columns, n_s));

						/* 'extend' the edges */
						c->c_cscanline[n_s][temp_columns + 1] = c->c_cscanline[n_s][temp_columns];

						/* make the current scanline the previous one */
						c_swaplines(c, n_s);

						/* Insert restart markers if enabled */
						if (c->need_restart){
							/* Insert restart markers only after a restart interval */
							if ((MCUs_counted % c->restart_interval) == 0){
								/* MISSING FUNCTION */
								bitoflush();
								temp = write_marker(out, (RSTm + ((MCUs_counted / c->restart_interval) % 8)));
							}
							MCUs_counted++;
						}
					}
				}
			}	/* End for each component in PLANE_INT */
		}	/* End for non LINE_INT */
		/* MISSING FUNCTION */
		bitoflush();
	}	/* End of loop on scans */

	c->all_header += write_marker(out, EOI);

	DGP_REPORTF1(verbose && c->need_table, "I used the mapping table @fp%x.", c->mappingTable->fp);

	DGP_REPORTF1(verbose && c->need_restart, "I used restart markers with restart interval %i.", c->restart_interval);

	DGP_REPORTF1(verbose, "Marker segment bytes: %ld.", c->all_header);

	/* Close down */
	DGP_ERR_PASS_CLEANUP(DgpJpegLsEncodeCleanup(c, &pos1));
	if(local_scanl0) DGP_FREE(local_scanl0);
	if(local_scanl1) DGP_FREE(local_scanl1);
	/* total bytes out, including JPEG-LS header, but not
	   application-specific header bytes */
	/* tot_out = (pos1-all_header)*8; */
	tot_out = pos1 * 8;

	DGP_REPORTF4(verbose, "Total bits out: %ld  Symbols in: %ld  %5.3lf bps  (%5.3lf : 1).",
		tot_out, tot_in, tot_out / (Dgp_Float64)tot_in, (log((Dgp_Float64)c->alpha)/log(2.0)) * tot_in / tot_out);

	t1 = get_utime();

	DGP_REPORTF2(verbose, "Time = %1.3lf secs : %1.0lf KSymbols/sec.", 
		t1 - t0, (tot_in) / (1024 * (t1 - t0)));

	if(encodeTime) *encodeTime = t1 - t0;

    NORMAL_RETURN;
ERR_CLEANUP:
	if(c){
		if(c->qdiv0) DGP_FREE(c->qdiv0);
		if(c->qmul0) DGP_FREE(c->qmul0);
		if(c->vLUT) DGP_FREE(c->vLUT);
		if(c->classmap) DGP_FREE (c->classmap);
		if(c->head_frame) DGP_FREE(c->head_frame);
		for (n_s = 0; n_s < MAX_SCANS; n_s++) {
			if(c->head_scan[n_s]) DGP_FREE(c->head_scan[n_s]);
		}
		if(c->multipleInputFiles){
			for (n_c = 0; n_c < c->components; n_c++) {
				if(c->c_scanl0[n_c]) DGP_FREE(c->c_scanl0[n_c]);
				if(c->c_scanl1[n_c]) DGP_FREE(c->c_scanl1[n_c]);
			}
		} else {
			if(c->scanl0) DGP_FREE(c->scanl0);
			if(c->scanl1) DGP_FREE(c->scanl1);
		}
		if(c->A) DGP_FREE(c->A);
		if(c->B) DGP_FREE(c->B);
		if(c->C) DGP_FREE(c->C);
		if(c->N) DGP_FREE(c->N);
		DGP_ERR_PASS(close_process_run(c));
		DGP_FREE(c);
	}
	if(local_scanl0) DGP_FREE(local_scanl0);
	if(local_scanl1) DGP_FREE(local_scanl1);
	ERR_RETURN;
}


/* Functions that read the PGM header */
#if 0
#define HEADER_MAXLINE 256

Dgp_Int32 read_header_6(FILE *fin, Dgp_Int32 *widthp, Dgp_Int32 *heightp, Dgp_Int32 *maxvalp, Dgp_Int32 *comp)
{
	Dgp_Int8 line[HEADER_MAXLINE];
	Dgp_Int32  cols,rows,maxval;

	if ( nextline(line, fin) != 0 )
		return -10;

	if (strncmp(line,"P5",2)==0) *comp=1;
	else if (strncmp(line,"P6",2)==0) *comp=3;
	     else if (strncmp(line,"P7",2)==0) *comp=4;
		      else return -1;
	
	if ( nextline(line, fin) != 0 )
		return -10;
	
	if ( sscanf(line,"%d %d",&cols,&rows) != 2 )
		return -1;
	
	if ( nextline(line, fin) != 0 )
		return -10;

	if ( sscanf(line,"%d",&maxval) != 1 )
		return -1;
	
	*widthp = cols;
	*heightp = rows;
	*maxvalp = maxval;

	return 0;
}



Dgp_Int32 read_header_5(FILE *fin, Dgp_Int32 *widthp, Dgp_Int32 *heightp, Dgp_Int32 *maxvalp)
{
	Dgp_Int8 line[HEADER_MAXLINE];
	Dgp_Int32  cols,rows,maxval;

	if ( nextline(line, fin) != 0 )
		return -10;

	if (strncmp(line,"P5",2)!=0)
		return -1;
	
	if ( nextline(line, fin) != 0 )
		return -10;
	
	if ( sscanf(line,"%d %d",&cols,&rows) != 2 )
		return -1;
	
	if ( nextline(line, fin) != 0 )
		return -10;

	if ( sscanf(line,"%d",&maxval) != 1 )
		return -1;
	
	*widthp = cols;
	*heightp = rows;
	*maxvalp = maxval;

	return 0;
}



/* Used to read in header lines of PGM/PPM files */
Dgp_Int32 nextline(Dgp_Int8 *line, FILE *fp)
{
	Dgp_Int8 *p;

	do {
		p = fgets(line, HEADER_MAXLINE, fp);
		/*
		if ( p != NULL )
		    fprintf(stderr,"%s",line);
		*/

	} while ( p != NULL && *p == '#' );

	if ( p==NULL )
		return -1;
	
	return 0;
}

#endif /* 0 */
