/* SPMG/JPEG-LS IMPLEMENTATION V.2.1
   =====================================
   These programs are Copyright (c) University of British Columbia. All rights reserved.
   They may be freely redistributed in their entirety provided that this copyright
   notice is not removed.  THEY MAY NOT BE SOLD FOR PROFIT OR INCORPORATED IN
   COMMERCIAL PROGRAMS WITHOUT THE WRITTEN PERMISSION OF THE COPYRIGHT HOLDER.
   Each program is provided as is, without any express or implied warranty,
   without even the warranty of fitness for a particular purpose.

   =========================================================
   THIS SOFTWARE IS BASED ON HP's implementation of jpeg-ls:
   =========================================================

   LOCO-I/JPEG-LS IMPLEMENTATION V.0.90
   -------------------------------------------------------------------------------
   (c) COPYRIGHT HEWLETT-PACKARD COMPANY, 1995-1999.
        HEWLETT-PACKARD COMPANY ("HP") DOES NOT WARRANT THE ACCURACY OR
   COMPLETENESS OF THE INFORMATION GIVEN HERE.  ANY USE MADE OF, OR
   RELIANCE ON, SUCH INFORMATION IS ENTIRELY AT USER'S OWN RISK.
        BY DOWNLOADING THE LOCO-I/JPEG-LS COMPRESSORS/DECOMPRESSORS
   ("THE SOFTWARE") YOU AGREE TO BE BOUND BY THE TERMS AND CONDITIONS
   OF THIS LICENSING AGREEMENT.
        YOU MAY DOWNLOAD AND USE THE SOFTWARE FOR NON-COMMERCIAL PURPOSES
   FREE OF CHARGE OR FURTHER OBLIGATION.  YOU MAY NOT, DIRECTLY OR
   INDIRECTLY, DISTRIBUTE THE SOFTWARE FOR A FEE, INCORPORATE THIS
   SOFTWARE INTO ANY PRODUCT OFFERED FOR SALE, OR USE THE SOFTWARE
   TO PROVIDE A SERVICE FOR WHICH A FEE IS CHARGED.
        YOU MAY MAKE COPIES OF THE SOFTWARE AND DISTRIBUTE SUCH COPIES TO
   OTHER PERSONS PROVIDED THAT SUCH COPIES ARE ACCOMPANIED BY
   HEWLETT-PACKARD'S COPYRIGHT NOTICE AND THIS AGREEMENT AND THAT
   SUCH OTHER PERSONS AGREE TO BE BOUND BY THE TERMS OF THIS AGREEMENT.
        THE SOFTWARE IS NOT OF PRODUCT QUALITY AND MAY HAVE ERRORS OR DEFECTS.
   THE JPEG-LS STANDARD IS STILL UNDER DEVELOPMENT. THE SOFTWARE IS NOT A
   FINAL OR FULL IMPLEMENTATION OF THE STANDARD.  HP GIVES NO EXPRESS OR
   IMPLIED WARRANTY OF ANY KIND AND ANY IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR PURPOSE ARE DISCLAIMED.
        HP SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
   OR CONSEQUENTIAL DAMAGES ARISING OUT OF ANY USE OF THE SOFTWARE.
   -------------------------------------------------------------------------------
*/

/* global.c --- support and portability routines: error handling, safe memory
 *                              management, etc.
 *
 * Initial code by Alex Jakulin,  Aug. 1995
 *
 * Modified and optimized: Gadiel Seroussi, October 1995 - ...
 *
 * Modified and added Restart marker and input tables by:
 * David Cheng-Hsiu Chu, and Ismail R. Ismail march 1999
 *
 * 20.3.2001 / dhtesq.
 * Added to digipaper code base for evaluation purposes.
 *
 */

#include <time.h>
#include "dgpjpegls.h"

/*
 * TIMING ROUTINES
 */

Dgp_Float64 
get_utime()
{
	/*clock_t c; */
	return (Dgp_Float64)clock()/CLOCKS_PER_SEC;
}

/* Set thresholds to default unless specified by header: */
void
set_thresholds(Dgp_Int32 alfa, Dgp_Int32 nearLosslessError, Dgp_Int32 *T1p, Dgp_Int32 *T2p, Dgp_Int32 *T3p)
{
	Dgp_Int32 lambda,
	    ilambda = 256 / alfa,
	    quant = 2 * nearLosslessError + 1,
	    T1 = *T1p, 
	    T2 = *T2p, 
	    T3 = *T3p;
	
	if(alfa < 4096){
		lambda = (alfa + 127) / 256;
	} else {
		lambda = (4096 + 127) / 256;
	}

	if ( T1 <= 0 )  {
		/* compute lossless default */
		if ( lambda ) {
			T1 = lambda * (BASIC_T1 - 2) + 2;
		}
		else {  /* alphabet < 8 bits */
			T1 = BASIC_T1 / ilambda;
			if ( T1 < 2 ) T1 = 2;
		}
		/* adjust for lossy */
		T1 += 3 * nearLosslessError;

		/* check that the default threshold is in bounds */
		if ( T1 < nearLosslessError + 1 || T1 > (alfa - 1) ) {
			T1 = nearLosslessError + 1;         /* eliminates the threshold */
		}
	}
	if ( T2 <= 0 )  {
		/* compute lossless default */
		if ( lambda ) {
			T2 = lambda * (BASIC_T2 - 3) + 3;
		}
		else {
			T2 = BASIC_T2 / ilambda;
			if ( T2 < 3 ) T2 = 3;
		}
		/* adjust for lossy */
		T2 += 5 * nearLosslessError;

		/* check that the default threshold is in bounds */
		if ( T2 < T1 || T2 > (alfa - 1) ) {
		     T2 = T1;         /* eliminates the threshold */
		}
	}
	if ( T3 <= 0 )  {
		/* compute lossless default */
		if ( lambda ) {
			T3 = lambda * (BASIC_T3 - 4) + 4;
		}
		else {
			T3 = BASIC_T3 / ilambda;
			if ( T3 < 4 ) T3 = 4;
		}
		/* adjust for lossy */
		T3 += 7 * nearLosslessError;

		/* check that the default threshold is in bounds */
		if ( T3 < T2 || T3 > (alfa - 1) ) {
		     T3 = T2;         /* eliminates the threshold */
		}
	}

	*T1p = T1;
	*T2p = T2;
	*T3p = T3;
	return;
}


/* We first check compatibility with JPEG-LS, then with this implementation. */
ErrCode
check_compatibility(Dgp_State *dgp_state, jpeg_ls_header *head_frame, jpeg_ls_header *head_scan, Dgp_Int32 n_s) 
{
    Dgp_Int32 number_of_scans, i;	
    Dgp_Int32 maxreset;
	DGP_STATE_PROLOGUE("check_compatibility()", dgp_state);

	/* Check implemented color modes. */
	DGP_ERR_GENF1((head_scan->color_mode > PIXEL_INT), cErrInvalidParm,
		"Color mode %d not supported...", head_scan->color_mode);

    if (head_scan->color_mode == PLANE_INT) {
		number_of_scans = head_frame->comp;
	} else {
		number_of_scans = 1;
	}
    
	/* Test standard compatibility. */
	DGP_ERR_GENF0((head_frame->columns <= 0 || head_frame->rows <= 0), cErrInvalidParm,
		"Image dimensions must be positive...");

	DGP_ERR_GENF1((head_frame->alp < 4), cErrInvalidParm,
		"You gave me an alphabet size of %d, but I need something >= 4.", head_frame->alp);

	DGP_ERR_GENF2((head_scan->T1 > head_scan->T2 || head_scan->T2 > head_scan->T3 ||
		head_scan->T1 < head_scan->nearLosslessError + 1 || head_scan->T3 >= head_scan->alp), cErrInvalidParm,
		"Bad thresholds:  must have %d <= Ta <= Tb <= Tc <= %d", head_scan->nearLosslessError + 1, head_scan->alp - 1);

	DGP_ERR_GENF1((head_frame->comp > 255), cErrInvalidParm,
		"You gave me %d components, but I can handle at most 255.", head_frame->comp);

	DGP_ERR_GENF2((head_scan->nearLosslessError >= head_scan->alp), cErrInvalidParm,
		"Error for near-lossless mode must be smaller than alphabet (%d); I got %d.", head_scan->alp, head_scan->nearLosslessError);

    maxreset = (head_scan->alp >= 256) ? (head_scan->alp - 1):255;
	DGP_ERR_GENF2((head_scan->RES < MINRESET || head_scan->RES > maxreset), cErrInvalidParm,
		"Reset parameter must be between %d and %d.", MINRESET, head_scan->alp - 1);

    for (i = 0; i < head_frame->comp; i++){
		DGP_ERR_GENF1(head_frame->comp_ids[i] != (i+1), cErrInvalidParm,
			"Component id %d in frame not compatible with this implementation.", head_frame->comp_ids[i]);
	}

    if (number_of_scans == 1) {
		DGP_ERR_GENF0((head_frame->comp != head_scan->comp), cErrInvalidParm,
			"In this implementation, all components must be in the scan when running in single-scan mode.");

        for (i = 0; i < head_frame->comp; i++){
			DGP_ERR_GENF1(head_scan->comp_ids[i] != (i+1), cErrInvalidParm,
				"Component id %d in single scan not compatible with this implementation.", head_scan->comp_ids[i]);
		}
    }
    else {
		DGP_ERR_GENF1((head_scan->comp != 1), cErrInvalidParm,
			"Only 1 component per scan allowed for plane interleaved mode;  I got %d.", head_scan->comp);

		DGP_ERR_GENF1(head_scan->comp_ids[0] != (n_s + 1), cErrInvalidParm,
			"Component id %d in multiple scan not compatible with this implementation.", head_scan->comp_ids[0]);
    }
	NORMAL_RETURN;
}


/* Although we have a full uds input and output interface, callers */
/* will probably going to memory most of the time.  Call this before your decode */
/* to get a memory-resident uds.  So:  this is for decoding jpeg-ls into */
/* an image buffer. */
/* */
/* As it turns out the codec only uses the fwrite() in this case. */
ErrCode
DgpJpegLsFOpenImage(Dgp_State			*dgp_state,
					Dgp_Int8			*start,
					Dgp_Int32			width,
					Dgp_Int32			height,
					Dgp_Int32			lpl,
					Dgp_Int32			xOffset,
					Dgp_Int32			xMax,
					Dgp_Int32			yMax,
					Dgp_Int32			bpp,
					uds					*out)
{
	DgpJpegLsMemoryStream *stream = NULL;
	DGP_STATE_PROLOGUE("DgpJpegLsFOpenImage()", dgp_state);
	DGP_ASSERT(out);

	DGP_CHECK_ALLOC(stream = (DgpJpegLsMemoryStream *)DGP_MALLOC(sizeof(*stream)));
	stream->current = stream->start = (Dgp_Int8 *)start;
	stream->sizeInBytes = 0;		/* Unused... */
	stream->width = width;
	stream->height = height;
	stream->lpl = lpl;
	stream->currentRow = stream->currentColumn = 0;
	stream->xMax = xMax;
	stream->yMax = yMax;
	stream->xOffset = xOffset;
	stream->stopX = (MIN(xOffset + width, xMax + 1)) - xOffset;
	stream->stopY = MIN(height, yMax + 1);
	stream->bpp = bpp;
	stream->writeableBytesPerRow = bpp * (stream->stopX - xOffset) / 8;
	out->fp = (void *)stream;
	out->ffeof = NULL;
	out->fseek = NULL;
	out->ftell = NULL;
	out->fwrite = DgpJpegLsFWriteToImage;
	out->fread = NULL;
	
	NORMAL_RETURN;
}

/* Multi-byte quantities in the jls headers appear to be big-endian by construction, */
/* so a straight fread won't work.  Use this.  It only works for <= 4 bytes. */
/*
size_t
DgpJpegLsBigEndianFRead(void *buffer, size_t itemSize, size_t itemCount, uds *in)
{
	Dgp_UInt32	x = 0, i, nBytes, bytesRead = 0;
	Dgp_UInt8 ch;
	nBytes = itemSize * itemCount;
	if(nBytes > 4 && itemSize > 1){
		return 0;
	}
	for(i = 0; i < nBytes; i++){
		bytesRead += in->fread((void *)&ch, 1, 1, in->fp);
		x = (x << 8) | ch;
	}
	*(Dgp_UInt32 *)buffer = x;
	return bytesRead / itemSize;
}
*/

ErrCode
DgpJpegLsFCloseImage(Dgp_State		*dgp_state,
					 uds			*out)
{
	DgpJpegLsMemoryStream *stream = NULL;
	DGP_STATE_PROLOGUE("DgpJpegLsFCloseImage()", dgp_state);
	DGP_ASSERT(out);

	stream = (DgpJpegLsMemoryStream *)out->fp;
	DGP_FREE(stream);
	out->fp = NULL;
	NORMAL_RETURN;
}

/* N.B. This only works if fed one row of samples at a time.  Cannot write fractions of a byte. */
size_t
DgpJpegLsFWriteToImage(void *buffer, size_t itemSize, size_t itemCount, void *stream)
{
	Dgp_UInt32 *temp;
	Dgp_Int32 bytesToWrite;
	DgpJpegLsMemoryStream *fp = (DgpJpegLsMemoryStream *)stream;
	bytesToWrite = MIN((Dgp_Int32)(itemSize * itemCount), fp->writeableBytesPerRow);
	memcpy(fp->current, buffer, bytesToWrite);
	temp = (Dgp_UInt32 *)fp->current + fp->lpl;
	fp->current = (Dgp_Int8 *)temp;
	fp->currentRow++;
	return bytesToWrite / itemSize;
}

size_t
DgpJpegLs_fwriteToMemory(void *buffer, size_t itemSize, size_t itemCount, void *stream)
{
	DgpJpegLsMemoryStream *fp = (DgpJpegLsMemoryStream *)stream;
	memcpy(fp->current, buffer, (itemSize * itemCount));
	fp->current += (itemSize * itemCount);
	return itemCount;
}

size_t
DgpJpegLs_freadFromMemory(void *buffer, size_t itemSize, size_t itemCount, void *stream)
{
	DgpJpegLsMemoryStream *fp = (DgpJpegLsMemoryStream *)stream;
	memcpy(buffer, fp->current, (itemSize * itemCount));
	fp->current += (itemSize * itemCount);
	return itemCount;
}

/*typedef Dgp_Int32 (TC_CDECL *ftell_func) (void *); */
Dgp_Int32
DgpJpegLs_ftellInMemory(void *stream)
{
	DgpJpegLsMemoryStream *fp = (DgpJpegLsMemoryStream *)stream;
	if(fp){
		return((Dgp_Int32)(fp->current - fp->start));
	}
	return -1L;
}

/* Presently unused... */
Dgp_Int32
DgpJpegLs_feofInMemory(void *stream)
{
	DgpJpegLsMemoryStream *fp = (DgpJpegLsMemoryStream *)stream;
	if(fp->current - fp->start < fp->sizeInBytes){
		return 0;
	}
	return 1;
}

