/* SPMG/JPEG-LS IMPLEMENTATION V.2.1
   =====================================
   These programs are Copyright (c) University of British Columbia. All rights reserved.
   They may be freely redistributed in their entirety provided that this copyright
   notice is not removed.  THEY MAY NOT BE SOLD FOR PROFIT OR INCORPORATED IN
   COMMERCIAL PROGRAMS WITHOUT THE WRITTEN PERMISSION OF THE COPYRIGHT HOLDER.
   Each program is provided as is, without any express or implied warranty,
   without even the warranty of fitness for a particular purpose.

   =========================================================
   THIS SOFTWARE IS BASED ON HP's implementation of jpeg-ls:
   =========================================================

   LOCO-I/JPEG-LS IMPLEMENTATION V.0.90
   -------------------------------------------------------------------------------
   (c) COPYRIGHT HEWLETT-PACKARD COMPANY, 1995-1999.
        HEWLETT-PACKARD COMPANY ("HP") DOES NOT WARRANT THE ACCURACY OR
   COMPLETENESS OF THE INFORMATION GIVEN HERE.  ANY USE MADE OF, OR
   RELIANCE ON, SUCH INFORMATION IS ENTIRELY AT USER'S OWN RISK.
        BY DOWNLOADING THE LOCO-I/JPEG-LS COMPRESSORS/DECOMPRESSORS
   ("THE SOFTWARE") YOU AGREE TO BE BOUND BY THE TERMS AND CONDITIONS
   OF THIS LICENSING AGREEMENT.
        YOU MAY DOWNLOAD AND USE THE SOFTWARE FOR NON-COMMERCIAL PURPOSES
   FREE OF CHARGE OR FURTHER OBLIGATION.  YOU MAY NOT, DIRECTLY OR
   INDIRECTLY, DISTRIBUTE THE SOFTWARE FOR A FEE, INCORPORATE THIS
   SOFTWARE INTO ANY PRODUCT OFFERED FOR SALE, OR USE THE SOFTWARE
   TO PROVIDE A SERVICE FOR WHICH A FEE IS CHARGED.
        YOU MAY MAKE COPIES OF THE SOFTWARE AND DISTRIBUTE SUCH COPIES TO
   OTHER PERSONS PROVIDED THAT SUCH COPIES ARE ACCOMPANIED BY
   HEWLETT-PACKARD'S COPYRIGHT NOTICE AND THIS AGREEMENT AND THAT
   SUCH OTHER PERSONS AGREE TO BE BOUND BY THE TERMS OF THIS AGREEMENT.
        THE SOFTWARE IS NOT OF PRODUCT QUALITY AND MAY HAVE ERRORS OR DEFECTS.
   THE JPEG-LS STANDARD IS STILL UNDER DEVELOPMENT. THE SOFTWARE IS NOT A
   FINAL OR FULL IMPLEMENTATION OF THE STANDARD.  HP GIVES NO EXPRESS OR
   IMPLIED WARRANTY OF ANY KIND AND ANY IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR PURPOSE ARE DISCLAIMED.
        HP SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
   OR CONSEQUENTIAL DAMAGES ARISING OUT OF ANY USE OF THE SOFTWARE.
   -------------------------------------------------------------------------------
*/

/* initialize.c --- functions to initialize look up tables
 *                  and statistics tables            
 *
 * Initial code by Alex Jakulin,  Aug. 1995
 *
 * Modified and optimized: Gadiel Seroussi, October 1995
 *
 *
 * Modified and added Restart marker and input tables by:
 * David Cheng-Hsiu Chu, and Ismail R. Ismail march 1999
 *
 * 20.3.2001 / dhtesq.
 * Added to digipaper code base for evaluation purposes.
 *
 */

#include <stdio.h>
#include <math.h>

#include "dgpjpegls.h"
#include "bitio.h"


/* Setup look-up tables for quantized gradient merging. */
/* Space should be allocated for the passed tables before this function is called... / dhtesq. */
ErrCode
prepareLUTs(Dgp_State		*dgp_state, 
			Dgp_Int32		alpha, 
			Dgp_Int32		lutmax, 
			Dgp_Int32		T1, 
			Dgp_Int32		T2, 
			Dgp_Int32		T3,
			Dgp_Int32		nearLosslessError,
			Dgp_Bool		lossy,
			Dgp_Int32		**classmap,
			Dgp_Int32		((**vLUT)[2 * LUTMAX16]))
{
	Dgp_Int32 i, j, idx, lmax;
	Dgp_Int32 *localClassmap = NULL;
	Dgp_Int32 (*localVLUT)[2 * LUTMAX16] = NULL;
	DGP_STATE_PROLOGUE("prepareLUTs()", dgp_state);
	DGP_ASSERT(classmap);
	DGP_ASSERT(vLUT);

	lmax = min(alpha, lutmax);
	
	/* implementation limitation... */
	DGP_ERR_GENF2((T3 > lmax - 1), cErrInvalidParm,
		"Current implementation does not support threshold T3 > %d;  you gave me %d.", lmax - 1, T3);

	DGP_CHECK_ALLOC_CLEANUP(localVLUT = DGP_MALLOC(sizeof(Dgp_Int32) * 3 * 2 * LUTMAX16));

	/* Build classification tables (lossless or lossy). */
	if (lossy == DGP_FALSE) {

		for (i = -lmax + 1; i < lmax; i++) {

			if  ( i <= -T3 )        /* ...... -T3  */
				idx = 7;
			else if ( i <= -T2 )    /* -(T3-1) ... -T2 */
				idx = 5;
			else if ( i <= -T1 )    /* -(T2-1) ... -T1 */
				idx = 3;

			else if ( i <= -1 )     /* -(T1-1) ...  -1 */
				idx = 1;
			else if ( i == 0 )      /*  just 0 */
				idx = 0;

			else if ( i < T1 )      /* 1 ... T1-1 */
				idx = 2;
			else if ( i < T2 )      /* T1 ... T2-1 */
				idx = 4;
			else if ( i < T3 )      /* T2 ... T3-1 */
				idx = 6;
			else                    /* T3 ... */
				idx = 8;

			localVLUT[0][i + lutmax] = CREGIONS * CREGIONS * idx;
			localVLUT[1][i + lutmax] = CREGIONS * idx;
			localVLUT[2][i + lutmax] = idx;
		}

	} else {

		for (i = -lmax + 1; i < lmax; i++) {

			if ( nearLosslessError >= (alpha - 1) )
				idx = 0;   /* degenerate case, regardless of thresholds */
			else

				if  ( i <= -T3 )        /* ...... -T3  */
					idx = 7;
				else if ( i <= -T2 )    /* -(T3-1) ... -T2 */
					idx = 5;
				else if ( i <= -T1 )    /* -(T2-1) ... -T1 */
					idx = 3;

				else if ( i <= -nearLosslessError - 1 )     /* -(T1-1) ...  -nearLosslessError-1 */
					idx = 1;
				else if ( i <= nearLosslessError )      /*  within nearLosslessError of 0 */
					idx = 0;

				else if ( i < T1 )      /* 1 ... T1-1 */
					idx = 2;
				else if ( i < T2 )      /* T1 ... T2-1 */
					idx = 4;
				else if ( i < T3 )      /* T2 ... T3-1 */
					idx = 6;
				else                    /* T3 ... */
					idx = 8;

			localVLUT[0][i + lutmax] = CREGIONS * CREGIONS * idx;
			localVLUT[1][i + lutmax] = CREGIONS * idx;
			localVLUT[2][i + lutmax] = idx;
		}

	}

	DGP_CHECK_ALLOC_CLEANUP(localClassmap = (Dgp_Int32 *)DGP_CALLOC(1, sizeof(*localClassmap) * CONTEXTS1));

	/* prepare context mapping table (symmetric context merging) */
	localClassmap[0] = 0;
	for ( i = 1, j = 0; i < CONTEXTS1; i++) {
	    Dgp_Int32 q1, q2, q3, n1 = 0, n2 = 0, n3 = 0, ineg, sgn;

	    if(localClassmap[i]) continue;

	    q1 = i / (CREGIONS * CREGIONS);		/* first digit */
	    q2 = (i / CREGIONS) % CREGIONS;		/* second digit */
    	q3 = i % CREGIONS;		   			/* third digit */

	    if((q1 % 2) || (q1 == 0 && (q2 % 2)) || (q1 == 0 && q2 == 0 && (q3 % 2))) sgn = -1;
	    else sgn = 1;

	    /* compute negative context */
    	if(q1) n1 = q1 + ((q1 % 2) ? 1 : -1);
	    if(q2) n2 = q2 + ((q2 % 2) ? 1 : -1);
	    if(q3) n3 = q3 + ((q3 % 2) ? 1 : -1);

	    ineg = (n1 * CREGIONS + n2) * CREGIONS + n3;
	    j++ ;    /* next class number */
	    localClassmap[i] = sgn * j;
	    localClassmap[ineg] = -sgn * j;
	}
	*classmap = localClassmap;
	*vLUT = localVLUT;
	NORMAL_RETURN;
ERR_CLEANUP:
	if(localClassmap) DGP_FREE(localClassmap);
	if(localVLUT) DGP_FREE(localVLUT);
	ERR_RETURN;
}


/* prepare quantization tables for near-lossless quantization */
ErrCode
prepare_qtables(Dgp_State		*dgp_state, 
				Dgp_Int32		absize, 
				Dgp_Int32		nearLosslessError,
				Dgp_Int32		**qdiv0,
				Dgp_Int32		**qdiv,
				Dgp_Int32		**qmul0,
				Dgp_Int32		**qmul)
{
    Dgp_Int32 diff, qdiff;
    Dgp_Int32 beta, quant;
	Dgp_Int32 *qdiv0Temp = NULL, *qdivTemp = NULL, *qmul0Temp = NULL, *qmulTemp = NULL;

	DGP_STATE_PROLOGUE("prepare_qtables()", dgp_state);

	DGP_ASSERT(qdiv0);
	DGP_ASSERT(qdiv);
	DGP_ASSERT(qmul0);
	DGP_ASSERT(qmul);

    quant = 2 * nearLosslessError + 1;
    beta = absize;

	DGP_CHECK_ALLOC_CLEANUP(qdiv0Temp = (Dgp_Int32 *)DGP_CALLOC(2 * absize - 1, sizeof(Dgp_Int32)));

    qdivTemp = qdiv0Temp + absize - 1;

	DGP_CHECK_ALLOC_CLEANUP(qmul0Temp = (Dgp_Int32 *)DGP_CALLOC(2 * beta - 1, sizeof(Dgp_Int32)));

    qmulTemp = qmul0Temp + beta - 1;

    for ( diff = -(absize - 1); diff < absize; diff++ ) {
		if ( diff < 0 ){
		    qdiff = - ( (nearLosslessError - diff) / quant );
		}
		else{
		    qdiff = ( nearLosslessError + diff ) / quant;
		}
	    qdivTemp[diff] = qdiff;
    }
    for ( qdiff = -(beta - 1); qdiff < beta; qdiff++ ) {
	    diff = quant * qdiff;
	    qmulTemp[qdiff] = diff;
    }

	*qdiv0 = qdiv0Temp;
	*qdiv = qdivTemp;
	*qmul0 = qmul0Temp;
	*qmul = qmulTemp;

	NORMAL_RETURN;

ERR_CLEANUP:
	if(qdiv0Temp) DGP_FREE(qdiv0Temp);
	if(qmul0Temp) DGP_FREE(qmul0Temp);
	ERR_RETURN;
}


/* Initialize A[], B[], C[], and N[] arrays. */
/* Space should be allocated for the passed tables before this function is called... / dhtesq. */
ErrCode
init_stats(Dgp_State		*dgp_state, 
		   Dgp_Int32		absize,
		   Dgp_Int32		*A,
		   Dgp_Int32		*B,
		   Dgp_Int32		*C,
		   Dgp_Int32		*N) 
{
	Dgp_Int32 i, initabstat, slack;

	DGP_STATE_PROLOGUE("init_stats()", dgp_state);
	DGP_ASSERT(A);
	DGP_ASSERT(B);
	DGP_ASSERT(C);
	DGP_ASSERT(N);

	slack = 1 << INITABSLACK;
	initabstat = (absize + slack / 2) / slack;
	if ( initabstat < MIN_INITABSTAT ) initabstat = MIN_INITABSTAT;

	/* do the statistics initialization */
	for (i = 0; i < TOT_CONTEXTS; ++i) {
		C[i]= B[i] = 0;
		N[i] = INITNSTAT;
		A[i] = initabstat;
	}
	NORMAL_RETURN;
}



