/* SPMG/JPEG-LS IMPLEMENTATION V.2.1
   =====================================
   These programs are Copyright (c) University of British Columbia. All rights reserved.
   They may be freely redistributed in their entirety provided that this copyright
   notice is not removed.  THEY MAY NOT BE SOLD FOR PROFIT OR INCORPORATED IN
   COMMERCIAL PROGRAMS WITHOUT THE WRITTEN PERMISSION OF THE COPYRIGHT HOLDER.
   Each program is provided as is, without any express or implied warranty,
   without even the warranty of fitness for a particular purpose.

   =========================================================
   THIS SOFTWARE IS BASED ON HP's implementation of jpeg-ls:
   =========================================================

   LOCO-I/JPEG-LS IMPLEMENTATION V.0.90
   -------------------------------------------------------------------------------
   (c) COPYRIGHT HEWLETT-PACKARD COMPANY, 1995-1999.
        HEWLETT-PACKARD COMPANY ("HP") DOES NOT WARRANT THE ACCURACY OR
   COMPLETENESS OF THE INFORMATION GIVEN HERE.  ANY USE MADE OF, OR
   RELIANCE ON, SUCH INFORMATION IS ENTIRELY AT USER'S OWN RISK.
        BY DOWNLOADING THE LOCO-I/JPEG-LS COMPRESSORS/DECOMPRESSORS
   ("THE SOFTWARE") YOU AGREE TO BE BOUND BY THE TERMS AND CONDITIONS
   OF THIS LICENSING AGREEMENT.
        YOU MAY DOWNLOAD AND USE THE SOFTWARE FOR NON-COMMERCIAL PURPOSES
   FREE OF CHARGE OR FURTHER OBLIGATION.  YOU MAY NOT, DIRECTLY OR
   INDIRECTLY, DISTRIBUTE THE SOFTWARE FOR A FEE, INCORPORATE THIS
   SOFTWARE INTO ANY PRODUCT OFFERED FOR SALE, OR USE THE SOFTWARE
   TO PROVIDE A SERVICE FOR WHICH A FEE IS CHARGED.
        YOU MAY MAKE COPIES OF THE SOFTWARE AND DISTRIBUTE SUCH COPIES TO
   OTHER PERSONS PROVIDED THAT SUCH COPIES ARE ACCOMPANIED BY
   HEWLETT-PACKARD'S COPYRIGHT NOTICE AND THIS AGREEMENT AND THAT
   SUCH OTHER PERSONS AGREE TO BE BOUND BY THE TERMS OF THIS AGREEMENT.
        THE SOFTWARE IS NOT OF PRODUCT QUALITY AND MAY HAVE ERRORS OR DEFECTS.
   THE JPEG-LS STANDARD IS STILL UNDER DEVELOPMENT. THE SOFTWARE IS NOT A
   FINAL OR FULL IMPLEMENTATION OF THE STANDARD.  HP GIVES NO EXPRESS OR
   IMPLIED WARRANTY OF ANY KIND AND ANY IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR PURPOSE ARE DISCLAIMED.
        HP SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
   OR CONSEQUENTIAL DAMAGES ARISING OUT OF ANY USE OF THE SOFTWARE.
   -------------------------------------------------------------------------------
*/

/* jpegmark.c --- for JPEG markers
 *
 * Initial code by Alex Jakulin,  Aug. 1995
 *
 * Modified and optimized: Gadiel Seroussi, October 1995
 *
 * Color Enhancement: Guillermo Sapiro, August 1996
 *
 * Modified and added Restart marker and input tables by:
 * David Cheng-Hsiu Chu, and Ismail R. Ismail march 1999
 *
 * 20.3.2001 / dhtesq.
 * Added to digipaper code base for evaluation purposes.
 *
 */


#include <stdio.h>
#include "dgpjpegls.h"
#include "jpegmark.h"

#ifdef BUFINPUT
#   include "bitio.h"
#else
#error Unbuffered input not supported;  define BUFINPUT...
/*#   define	mygetc(in)	(getc(in))      */
/*#   define 	myfeof(in)	(feof(in)) */
#endif

ErrCode
check_range(Dgp_State *dgp_state, Dgp_Int32 param, Dgp_Int8 *name, Dgp_Int32 low, Dgp_Int32 high)
{
	DGP_STATE_PROLOGUE("check_range()", dgp_state);

	if( param < low || param > high ){
		Dgp_Int8 errString[255];
		sprintf(errString, "Allowed range for %s is [%d..%d];  I got %d.",	name, low, high, param);
		DGP_ERR_GENF0(DGP_TRUE, cErrInvalidParm, errString);
	}

	NORMAL_RETURN;
}

/*
 *
 *   Marker output functions
 *
 */

#if 0
Dgp_Int32
write_n_bytes(FILE *out, Dgp_Int32 value, Dgp_Int32 n)
{
	Dgp_Int32	l;

	if (n>4)  {
		fprintf(stderr,"write_n_bytes: Only 32 bits variables supported.\n");
		exit(10);
	}		

#ifdef NOT_REALLY_BIG_ENDIAN
        for (l=n-1;l>=0;l--)  {
	    if ( putc((value>>(8*l))&0x000000FF,out) == EOF )
		return EOF;
        }
#else  /* little endian */
	for (l=0;l<n;l++) {
	    if ( putc((value&0x000000FF),out) == EOF )
		return EOF;
	    value >>= 8;
	}
#endif
	return n;

}
#endif

/*
Dgp_Int32
write_2_bytes(FILE *out, Dgp_Int32 value)
{
	return write_n_bytes(out,value,2);
}
*/

/* Write a two-byte marker (just the marker identifier). */
/* This will write the two least significant bytes of the word; */
/* endianness issues? / dhtesq. */
Dgp_Int32
write_marker(uds *out, Dgp_Int32 marker)
{
	return out->fwrite((void *)&marker, 1, 2, out->fp);
}

ErrCode
write_jpegls_frame(Dgp_State *dgp_state, uds *out, jpeg_ls_header *jp, Dgp_Int32 *count)
{
    Dgp_Int32 i, marker_len, bpp, ct = 0;
	DGP_STATE_PROLOGUE("write_jpegls_frame()", dgp_state);
	DGP_ASSERT(count);
		 
    ct += write_marker(out, SOF_LS);   /* write JPEG-LS frame marker */

    DGP_ERR_PASS(check_range(dgp_state, jp->comp, "frame components", 1, 255))
		;
    marker_len = 8 + 3 * jp->comp;

    ct += out->fwrite((void *)&marker_len, 1, 2, out->fp);
		
	/* Find nearest higher power of 2 of the alphabet for the bit depth... */
    for ( bpp = 1; (1L << bpp) < jp->alp; bpp++ );

	ct += out->fwrite((void *)&bpp, 1, 1, out->fp);

    /* Current implementation only supports up to 64K samples in 
       either direction. Also, they must be specified in the frame header. */
    DGP_ERR_PASS(check_range(dgp_state, jp->rows, "rows", 1, 65535));
    DGP_ERR_PASS(check_range(dgp_state, jp->columns, "columns", 1, 65535));

	ct += out->fwrite((void *)&jp->rows, 1, 2, out->fp);
	ct += out->fwrite((void *)&jp->columns, 1, 2, out->fp);
	ct += out->fwrite((void *)&jp->comp, 1, 1, out->fp);

    /* Now write a triplet of bytes per component. */
    for ( i = 0; i < jp->comp; i++ ) {
		Dgp_Int32 sx = jp->samplingx[i], sy = jp->samplingy[i];
		Dgp_Int32 temp = (sx << 4) | sy;

		DGP_ERR_PASS(check_range(dgp_state, sx, "sampling(x)", 1, 4));
		DGP_ERR_PASS(check_range(dgp_state, sy, "sampling(y)", 1, 4));

		ct += out->fwrite((void *)&jp->comp_ids[i], 1, 1, out->fp);	/* component identifier */
		ct += out->fwrite((void *)&temp, 1, 1, out->fp);				/* sampling rates */
		temp = 0;
		ct += out->fwrite((void *)&temp, 1, 1, out->fp);				/* Tq unused */
    }
	
	*count = ct;

    NORMAL_RETURN;
}

ErrCode
write_jpegls_scan(Dgp_State *dgp_state, uds *out, jpeg_ls_header *jp, Dgp_Int32 *count)
{
	Dgp_Int32 i, marker_len, ct = 0;
	DGP_STATE_PROLOGUE("write_jpegls_scan()", dgp_state);
	DGP_ASSERT(count);
		 
    ct += write_marker(out, SOS);   /* write JPEG-LS scan marker */

    DGP_ERR_PASS(check_range(dgp_state, jp->comp, "scan components", 1, 4));

	DGP_ERR_GENF1(( jp->comp == 1 && jp->color_mode != PLANE_INT), cErrInvalidParm,
		"Interleave for 1 component must be PLANE_INT, but I got %d.", jp->color_mode);

	DGP_ERR_GENF1(( jp->comp >1 && jp->color_mode == 0 ), cErrInvalidParm,
		"Interleave for multi-component scan must be nonzero, but I got %d", jp->color_mode);

    marker_len = 6 + 2 * jp->comp;

	ct += out->fwrite((void *)&marker_len, 1, 2, out->fp);	/* Write marker length. */
	ct += out->fwrite((void *)&jp->comp, 1, 1, out->fp);		/* Number of components for the scan. */

    /* Write 2 bytes per component. */
    for ( i = 0; i < jp->comp; i++ ) {
		Dgp_Int8 temp = 0;
		ct += out->fwrite((void *)&jp->comp_ids[i], 1, 1, out->fp);		/* component identifier */
		ct += out->fwrite((void *)&temp, 1, 1, out->fp);				/* no tables in this implementation */
    }

    DGP_ERR_PASS(check_range(dgp_state, jp->nearLosslessError, "nearLosslessError", 0, 255));
	ct += out->fwrite((void *)&jp->nearLosslessError, 1, 1, out->fp);

    DGP_ERR_PASS(check_range(dgp_state, jp->color_mode, "INTERLEAVE", 0, 2));
	ct += out->fwrite((void *)&jp->color_mode, 1, 1, out->fp);

    DGP_ERR_PASS(check_range(dgp_state, jp->shift, "SHIFT", 0, 15));
	ct += out->fwrite((void *)&jp->shift, 1, 1, out->fp);

	*count = ct;

	NORMAL_RETURN;
}

/* Missing last two parameters when I arrived;  function defaulted to LSE_PARAMS and made */
/* no use of the mapping table... / dhtesq. */
Dgp_Int32
write_jpegls_extmarker(uds *out, jpeg_ls_header *jp, Dgp_Int8 markerType, uds *mappingTable)
{
    Dgp_Int32 ct = 0, temp;

	mappingTable = mappingTable;	/* What to do with this? */

    ct += write_marker(out, LSE);   /* write JPEG-LS extended marker id */

	temp = 13;
	ct += out->fwrite((void *)&temp, 1, 2, out->fp);	/* marker length */
	temp = markerType;
    ct += out->fwrite((void *)&temp, 1, 1, out->fp);	/* ext marker id */
	temp = jp->alp - 1;
	ct += out->fwrite((void *)&temp, 1, 2, out->fp);	/* MAXVAL */
    ct += out->fwrite((void *)&jp->T1, 1, 2, out->fp);
    ct += out->fwrite((void *)&jp->T2, 1, 2, out->fp);
    ct += out->fwrite((void *)&jp->T3, 1, 2, out->fp);
    ct += out->fwrite((void *)&jp->RES, 1, 2, out->fp);

    return ct;
}

/*
 *
 *   Marker input functions
 *
 */

/* Seeks a marker in the input stream.  Returns the marker head, or EOF. */
ErrCode
seek_marker(DgpJpegLsContext *context, Dgp_Int32 *mkp, Dgp_Int32 *retVal)
{
    Dgp_Int32 ct = 0;
	Dgp_UInt8 c, c2;
	DGP_STATE_PROLOGUE("seek_marker()", context->dgp_state);
	DGP_ASSERT(retVal);
	DGP_ASSERT(mkp);

	DGP_ERR_PASS(DgpJpegLsGetC(context, &c));
    while ( c != EOF ) {
		ct++;
		if ( c == 0xFF ) {
			DGP_ERR_PASS(DgpJpegLsGetC(context, &c2));
			if ( c2 == EOF ) return EOF;
			ct++;
			if ( c2 & 0x80 ) {
				*mkp = (c << 8) | c2;
				*retVal = ct;
				NORMAL_RETURN;
			}
		}
		DGP_ERR_PASS(DgpJpegLsGetC(context, &c));
    }
	*retVal = EOF;
	NORMAL_RETURN;
}

/* Reads n bytes (0 <= n <= 4 only) from the input stream. */
ErrCode
read_n_bytes(DgpJpegLsContext *c, Dgp_Int32 n, Dgp_UInt32 *m)
{
    Dgp_UInt32 mLocal = 0;
    Dgp_Int32 i;
	Dgp_UInt8 ch = 0;
	DGP_STATE_PROLOGUE("read_n_bytes()", c->dgp_state);
	DGP_ASSERT(m);
	DGP_ASSERT(n <= 4);

    for ( i = 0; i < n; i++ ){
		DGP_ERR_PASS(DgpJpegLsGetC(c, &ch));
		mLocal = (mLocal << 8) | ch;
	}

	*m = mLocal;
	NORMAL_RETURN;
}

/* Reads a marker from the next two bytes in the input stream. */
ErrCode
read_marker(DgpJpegLsContext *c, Dgp_Int32 *mkp, Dgp_Int32 *count)
{
	Dgp_UInt32 m, ct = 0;
	uds *in = c->in;
	DGP_STATE_PROLOGUE("read_marker()", c->dgp_state);
	DGP_ASSERT(count);
	DGP_ASSERT(mkp);

	DGP_ERR_PASS(read_n_bytes(c, 2, &m));
	ct += 2;

	if(in->ffeof(in->fp)){
		*count = EOF;
		NORMAL_RETURN;
	}

	DGP_ERR_GENF1((m & 0xFF00) != 0xFF00, cErrInvalidParm,
		"Expected marker, but got %04x", m);

	*mkp = m;
	*count = ct;
	NORMAL_RETURN;
}

/* Reads the JPEG-LS frame marker (not including marker head). */
ErrCode
read_jpegls_frame(DgpJpegLsContext *c, jpeg_ls_header *jp, Dgp_Int32 *count)
{
    Dgp_Int32	i,
				marker_len = 0,
				bpp = 0,
				tq,
				comp = 0,
				ct = 0;
	Dgp_State *dgp_state = c->dgp_state;
	DGP_STATE_PROLOGUE("read_jpegls_frame()", dgp_state);
	DGP_ASSERT(count);
	DGP_ASSERT(jp);

    /* Read Marker Length */
	DGP_ERR_PASS(read_n_bytes(c, 2, &marker_len));
	ct += 2;

	/* Read the bits per pixel */
	DGP_ERR_PASS(read_n_bytes(c, 1, &bpp));
	ct++;

    DGP_ERR_PASS(check_range(dgp_state, bpp, "bpp", 2, 16));
    jp->alp = 1 << bpp;

	jp->rows = 0;
	jp->columns = 0;
	/* Read the rows and columns */	
	DGP_ERR_PASS(read_n_bytes(c, 2, &jp->rows));
	ct += 2;
	DGP_ERR_PASS(read_n_bytes(c, 2, &jp->columns));
	ct += 2;
    
	/* Read component information */
	DGP_ERR_PASS(read_n_bytes(c, 1, &comp));
	ct++;

    DGP_ERR_PASS(check_range(dgp_state, comp, "COMP", 1, 255));
    jp->comp = comp;

    for ( i = 0; i < comp; i++ ) {
		Dgp_Int32 sx = 0, sy = 0, cid = 0;
		tq = 0;

		DGP_ERR_PASS(read_n_bytes(c, 1, &cid));
		ct++;
		DGP_ERR_PASS(read_n_bytes(c, 1, &sx));
		ct++;
		DGP_ERR_PASS(read_n_bytes(c, 1, &tq));
		ct++;

		DGP_ERR_PASS(check_range(dgp_state, tq, "Tq", 0, 0));

		sy = sx & 0x0f;
		sx >>= 4;
		DGP_ERR_PASS(check_range(dgp_state, sx, "sampling(x)", 1, 4));
		DGP_ERR_PASS(check_range(dgp_state, sy, "sampling(y)", 1, 4));
		jp->samplingx[i] = sx;
		jp->samplingy[i] = sy;
		jp->comp_ids[i] = cid;
    }

	/* Check for errors */
	DGP_ERR_GENF0(c->in->ffeof(c->in->fp), cErrIO_ReadError, 
		"read_jpegls_frame: EOF while reading frame marker.");

	DGP_ERR_GENF2(marker_len != 8 + 3 * comp, cErrInvalidParm,
		"read_jpegls_frame: inconsistent marker length: expected %d but got %d.", marker_len, 8 + 3 * comp);

	*count = ct;
    NORMAL_RETURN;
}


/* Reads the JPEG-LS scan marker (not including marker head). */
ErrCode
read_jpegls_scan(DgpJpegLsContext *c, jpeg_ls_header *jp, Dgp_Int32 *count)
{
    Dgp_Int32	i, 
				marker_len = 0,
				comp = 0, 
				ct = 0;
	Dgp_State *dgp_state = c->dgp_state;
	DGP_STATE_PROLOGUE("read_jpegls_scan()", dgp_state);
	DGP_ASSERT(count);
	DGP_ASSERT(jp);

	DGP_ERR_PASS(read_n_bytes(c, 2, &marker_len));
	ct += 2;

	DGP_ERR_PASS(read_n_bytes(c, 1, &comp));
	ct++;

    DGP_ERR_PASS(check_range(dgp_state, comp, "scan components", 1, 4));

    jp->comp = comp;

    /* read 2 bytes per component */
    for ( i = 0; i < comp; i++ ) {
		Dgp_Int32 cid = 0, tm = 0;

		DGP_ERR_PASS(read_n_bytes(c, 1, &cid));			/* Component identifier. */
		ct++;
		DGP_ERR_PASS(read_n_bytes(c, 1, &tm));			/* Table identifier. */
		ct++;

		DGP_ERR_GENF0(tm, cErrInvalidParm,
			"read_jpegls_scan: found nonzero table identifier, not supported.");

		jp->comp_ids[i] = cid;
    }

	DGP_ERR_PASS(read_n_bytes(c, 1, &jp->nearLosslessError));
	ct++;

    DGP_ERR_PASS(check_range(dgp_state, jp->nearLosslessError, "nearLosslessError", 0, 255));

	DGP_ERR_PASS(read_n_bytes(c, 1, &jp->color_mode));
	ct++;

    DGP_ERR_PASS(check_range(dgp_state, jp->color_mode, "INTERLEAVE", 0, 2));

    if ( jp->comp == 1 && jp->color_mode != 0 ) {
		/*
		Was commented out when I arrived... / dhtesq.
		fprintf(stderr,"Interleave for 1 component must be 0: got %d\n",
			jp->color_mode);
		*/

		/* Ignore interleave value, set to 0. */
		jp->color_mode = 0;
    }

	DGP_ERR_GENF1(( jp->comp > 1 && jp->color_mode == 0 ), cErrInvalidParm,
		"Interleave for multi-component scan must be nonzero;  got %d.", jp->color_mode);

	DGP_ERR_PASS(read_n_bytes(c, 1, &jp->shift));
	ct++;

    DGP_ERR_PASS(check_range(dgp_state, jp->shift, "SHIFT", 0, 15));

	DGP_ERR_GENF0(c->in->ffeof(c->in->fp), cErrIO_ReadError,
		"read_jpegls_scan: EOF while reading frame marker.");

	DGP_ERR_GENF2(marker_len != 6 + 2 * comp, cErrInvalidParm,
		"read_jpegls_scan: inconsistent marker length;  expected %d, got %d.", marker_len, 6 + 2 * comp);

    NORMAL_RETURN;
}


/* Reads the JPEG-LS extension marker (not including marker head). */
/* Supports non-default type (1) and mapping table type (2). */
ErrCode
read_jpegls_extmarker(DgpJpegLsContext *c, jpeg_ls_header *jp, Dgp_Int32 *count)
{
    Dgp_Int32 marker_len,		/* marker length */
			  maxval,			/* max value */
			  ct = 0;			
	Dgp_Int32 IDtype;			/* LSE type */
	Dgp_Int32 TID;				/* table ID */
	Dgp_Int32 Wt;				/* width of each table entry */
	Dgp_Int32 MAXTAB;			/* maximum table index */
	Dgp_Int32 i;
	Dgp_State *dgp_state = c->dgp_state;
	DGP_STATE_PROLOGUE("read_jpegls_extmarker()", dgp_state);
	DGP_ASSERT(count);
	DGP_ASSERT(jp);

	/* Read marker length */
	DGP_ERR_PASS(read_n_bytes(c, 2, &marker_len));
	ct += 2;

	/* Read id type */
	DGP_ERR_PASS(read_n_bytes(c, 1, &IDtype));
	ct++;

	/* For Type 1 - non default parameters. */
	if (IDtype == LSE_PARAMS){
		DGP_ERR_GENF1(marker_len != 13, cErrInvalidParm,
			"read_jpegls_extmarker: bad marker length %d.", marker_len);

		/* Read maxval. */
		DGP_ERR_PASS(read_n_bytes(c, 2, &maxval));
		ct += 2;
		jp->alp = maxval + 1;

		/* read thresholds and reset */
		DGP_ERR_PASS(read_n_bytes(c, 2, &jp->T1));
		ct += 2;
		DGP_ERR_PASS(read_n_bytes(c, 2, &jp->T2));
		ct += 2;
		DGP_ERR_PASS(read_n_bytes(c, 2, &jp->T3));
		ct += 2;
		DGP_ERR_PASS(read_n_bytes(c, 2, &jp->RES));
		ct += 2;

		DGP_ERR_GENF0(c->in->ffeof(c->in->fp), cErrIO_ReadError,
			"read_jpegls_extmarker: EOF while reading frame marker.");			

		*count = ct;
		NORMAL_RETURN;
	}
	
	/* For Type 2 - mapping table */
	if (IDtype == LSE_MAPTABLE){

		/* Indicate table used */
		jp->need_table = 1;

		/* Read table ID */
		DGP_ERR_PASS(read_n_bytes(c, 1, &TID));
		ct++;
		jp->TID = TID;

		/* Read width of table entry */
		DGP_ERR_PASS(read_n_bytes(c, 1, &Wt));
		ct++;
		DGP_ERR_GENF0((Wt <= 0 || Wt > 3), cErrInvalidParm,
			"Width of mapping table entries must be either 1,2 or 3 in this implementation. Sorry!");
		jp->Wt = Wt;

		/* Calculate value of MAXTAB */
		jp->MAXTAB = MAXTAB = ((marker_len - 5) / Wt) - 1;

		/* Get table entries */
		DGP_CHECK_ALLOC(jp->TABLE[TID] = (Dgp_UInt32 *)DGP_CALLOC((MAXTAB + 1) * sizeof(Dgp_Int32), 1));
		for (i = 0; i <= MAXTAB; i++){
			DGP_ERR_PASS(read_n_bytes(c, Wt, &jp->TABLE[TID][i]));
			ct += Wt;
		}

		*count = ct;
		NORMAL_RETURN;
	}

	/* Non supported types. */
	DGP_ERR_GENF1(DGP_TRUE, cErrInvalidParm,
		"LSE marker type %i not supported in this implementation.", IDtype);

	NORMAL_RETURN;
}


/* Read DRI restart marker */
ErrCode
read_jpegls_restartmarker(DgpJpegLsContext *c, jpeg_ls_header *jp, Dgp_Int32 *count)
{
	Dgp_Int32 ct = 0;
	Dgp_Int32 marker_len;		/* the marker length */
	Dgp_Int32 Ri;				/* the restart interval */
	Dgp_State *dgp_state = c->dgp_state;
	DGP_STATE_PROLOGUE("read_jpegls_restartmarker()", dgp_state);
	DGP_ASSERT(jp);
	DGP_ASSERT(count);

	/* Read marker length. */
	DGP_ERR_PASS(read_n_bytes(c, 2, &marker_len));
	ct += 2;

	/* Read the restart interval. */
	DGP_ERR_PASS(read_n_bytes(c, marker_len - 2, &Ri));
	ct += marker_len - 2;

	jp->restart_interval = Ri;

	*count = ct;
	NORMAL_RETURN;
}

/* This was completely missing when I arrived.  It's an educated guess based on  */
/* read_jpegls_restartmarker(). / dhtesq. */
Dgp_Int32
write_jpegls_restartmarker(uds *out, jpeg_ls_header *jp)
{
    Dgp_Int32 ct = 0, temp;

    ct += write_marker(out, DRI);   /* write DRI marker id */

	/* Write marker length, which for some reason is offset by two bytes. */
	temp = sizeof(jp->restart_interval) + 2;
	ct += out->fwrite((void *)&temp, 1, 2, out->fp);

	/* Write restart interval. */
	ct += out->fwrite((void *)&jp->restart_interval, 1, temp - 2, out->fp);

    return ct;
}
