/* SPMG/JPEG-LS IMPLEMENTATION V.2.1
   =====================================
   These programs are Copyright (c) University of British Columbia. All rights reserved.
   They may be freely redistributed in their entirety provided that this copyright
   notice is not removed.  THEY MAY NOT BE SOLD FOR PROFIT OR INCORPORATED IN
   COMMERCIAL PROGRAMS WITHOUT THE WRITTEN PERMISSION OF THE COPYRIGHT HOLDER.
   Each program is provided as is, without any express or implied warranty,
   without even the warranty of fitness for a particular purpose.

   =========================================================
   THIS SOFTWARE IS BASED ON HP's implementation of jpeg-ls:
   =========================================================

   LOCO-I/JPEG-LS IMPLEMENTATION V.0.90
   -------------------------------------------------------------------------------
   (c) COPYRIGHT HEWLETT-PACKARD COMPANY, 1995-1999.
        HEWLETT-PACKARD COMPANY ("HP") DOES NOT WARRANT THE ACCURACY OR
   COMPLETENESS OF THE INFORMATION GIVEN HERE.  ANY USE MADE OF, OR
   RELIANCE ON, SUCH INFORMATION IS ENTIRELY AT USER'S OWN RISK.
        BY DOWNLOADING THE LOCO-I/JPEG-LS COMPRESSORS/DECOMPRESSORS
   ("THE SOFTWARE") YOU AGREE TO BE BOUND BY THE TERMS AND CONDITIONS
   OF THIS LICENSING AGREEMENT.
        YOU MAY DOWNLOAD AND USE THE SOFTWARE FOR NON-COMMERCIAL PURPOSES
   FREE OF CHARGE OR FURTHER OBLIGATION.  YOU MAY NOT, DIRECTLY OR
   INDIRECTLY, DISTRIBUTE THE SOFTWARE FOR A FEE, INCORPORATE THIS
   SOFTWARE INTO ANY PRODUCT OFFERED FOR SALE, OR USE THE SOFTWARE
   TO PROVIDE A SERVICE FOR WHICH A FEE IS CHARGED.
        YOU MAY MAKE COPIES OF THE SOFTWARE AND DISTRIBUTE SUCH COPIES TO
   OTHER PERSONS PROVIDED THAT SUCH COPIES ARE ACCOMPANIED BY
   HEWLETT-PACKARD'S COPYRIGHT NOTICE AND THIS AGREEMENT AND THAT
   SUCH OTHER PERSONS AGREE TO BE BOUND BY THE TERMS OF THIS AGREEMENT.
        THE SOFTWARE IS NOT OF PRODUCT QUALITY AND MAY HAVE ERRORS OR DEFECTS.
   THE JPEG-LS STANDARD IS STILL UNDER DEVELOPMENT. THE SOFTWARE IS NOT A
   FINAL OR FULL IMPLEMENTATION OF THE STANDARD.  HP GIVES NO EXPRESS OR
   IMPLIED WARRANTY OF ANY KIND AND ANY IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR PURPOSE ARE DISCLAIMED.
        HP SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
   OR CONSEQUENTIAL DAMAGES ARISING OUT OF ANY USE OF THE SOFTWARE.
   -------------------------------------------------------------------------------
*/

/* lossless_d.c --- the main pipeline which processes a scanline by doing
 *                prediction, context computation, context quantization,
 *                and statistics gathering. (for lossless mode)
 *
 * Initial code by Alex Jakulin,  Aug. 1995
 *
 * Modified and optimized: Gadiel Seroussi, October 1995
 *
 * Modified and added Restart marker and input tables by:
 * David Cheng-Hsiu Chu, and Ismail R. Ismail march 1999
 *
 * 2.4.2001 / dhtesq.
 * Added to digipaper code base for evaluation purposes.
 *
 */

#include "dgpjpegls.h"
#include "bitio.h"

#include <stdio.h>
#include <math.h>

/* Do Golomb-Rice statistics and DECODING for LOSSLESS images */
static INLINE ErrCode
lossless_regular_mode_d(DgpJpegLsContext		*c,
						Dgp_Int32				Q, 
						Dgp_Int32				SIGN, 
						Dgp_Int32				Px,
						pixel					*cur)
{
	Dgp_Int32 At, Bt, Nt, Errval, absErrval;
	Dgp_Int32 current, k;
	Dgp_State *dgp_state = c->dgp_state;
	DGP_STATE_PROLOGUE("lossless_regular_mode_d()", dgp_state);
	DGP_ASSERT(cur);

	/* This function is called only for regular contexts. 
	   End_of_run context is treated separately. */

	Nt = c->N[Q];
	At = c->A[Q];

	{
		/* Estimate k */
	    register nst = Nt;
	    for(k = 0; nst < At; nst *= 2, k++);
	}

	/* Get the number of leading zeros */
	absErrval = 0;
	do {
		Dgp_Int32 temp;

		temp = c->zeroLUT[c->reg >> 24];
		absErrval += temp;
		if (temp != 8) {
			DGP_ERR_PASS(fillbuffer(c, temp + 1));
			break;
		}
		DGP_ERR_PASS(fillbuffer(c, 8));
	} while (1);

	if ( absErrval < c->limit ) {
		/* now add the binary part of the Rice code */
		if (k) {
			register Dgp_UInt32 temp;
			absErrval <<= k;
			GETBITS(c, temp, k);
			absErrval += temp;
		}
	}
	else {
	    /* the original unary would have been too long:
	       (mapped value) - 1 was sent verbatim */
		GETBITS(c, absErrval, c->qbpp);
		absErrval++;
	}

	/* Do the Rice mapping */
	if ( absErrval & 1 ) {        /* negative */
		absErrval = (absErrval + 1) / 2;
		Errval = -absErrval;
	} else {
		absErrval /= 2;
		Errval = absErrval;
	}

	Bt = c->B[Q];

	if ( k==0 && (2 * Bt <= -Nt) ){
		/* special case: see encoder side */
		Errval = -(Errval+1);
		absErrval = (Errval < 0) ? (-Errval):Errval;
	}

	/* center, clip if necessary, and mask final error */
	if ( SIGN == -1 ) {
	    Px -= c->C[Q];
	    CLIP(Px, c->alpha);
	    
#if defined(POW2)
	    /* this is valid if alpha is a power of 2 */
	    current = (Px - Errval) & (c->alpha - 1);
#else
	    current = Px - Errval;
#endif
	}
	else {
	    Px += c->C[Q];
	    CLIP(Px, c->alpha);
	    
#if defined(POW2)
	    /* valid if alpha is a power of 2 */
	    current = (Px + Errval) & (c->alpha - 1);
#else
	    current = Px + Errval;
#endif
	}


#if !defined(POW2)
	/* reduce mod alpha, for arbitrary alpha */
	if (current < 0)
		current += c->alpha;
	else if (current >= c->alpha)
		current -= c->alpha;
#endif

	/* update bias stats */
	c->B[Q] = (Bt += Errval);

	/* update Golomb-Rice stats */
	c->A[Q] += absErrval;


	/* check reset (joint for Rice-Golomb and bias cancelation) */
	if(Nt == c->theReset) {
		c->N[Q] = (Nt >>= 1);
		c->A[Q] >>= 1;
		c->B[Q] = (Bt >>= 1);
	}

	/* Do bias estimation for NEXT pixel */
	c->N[Q] = (++Nt);
	if ( Bt <= -Nt ) {

	    if (c->C[Q] > MIN_C) --c->C[Q];

	    Bt = (c->B[Q] += Nt);

	    if ( Bt <= -Nt ) c->B[Q] = -Nt + 1;

	} else if ( Bt > 0 ) {

	    if (c->C[Q] < MAX_C) ++c->C[Q];

	    Bt = (c->B[Q] -= Nt);

	    if ( Bt > 0 ) c->B[Q] = 0;
	}

	*cur = (Dgp_UInt16)current;
	NORMAL_RETURN;
}


/* Do end of run DECODING for LOSSLESS images. */
static INLINE ErrCode
lossless_end_of_run_d(DgpJpegLsContext			*c,
					  pixel						Ra, 
					  pixel						Rb, 
					  Dgp_Int32					RItype,
					  pixel						*returnPixel)
{
	Dgp_Int32	Ix,
				Errval,
				absErrval,
				MErrval,
				k,
				Q,
				oldmap, 
				Nt,
				At;
	DGP_STATE_PROLOGUE("lossless_end_of_run_d()", c->dgp_state);
	DGP_ASSERT(returnPixel);

	Q = EOR_0 + RItype;
	Nt = c->N[Q], 
	At = c->A[Q];

	if ( RItype ) At += Nt/2;

	/* Estimate k */
	for(k = 0; Nt < At; Nt *=2, k++);

	/* read and decode the Golomb code */
	/* Get the number of leading zeros */
	MErrval = 0;
	do {
		Dgp_Int32 temp;

		temp = c->zeroLUT[c->reg >> 24];
		MErrval += temp;
		if (temp != 8) {
			DGP_ERR_PASS(fillbuffer(c, temp + 1));
			break;
		}
		DGP_ERR_PASS(fillbuffer(c, 8));
	} while (1);

	c->eor_limit = c->limit - c->limit_reduce;

	if ( MErrval < c->eor_limit ) {
		/* now add the binary part of the Golomb code */
		if (k) {
			register Dgp_UInt32 temp;
			MErrval <<= k;
			GETBITS(c, temp, k);
			MErrval += temp;
		}
	}
	else {
	/* the original unary would have been too long: (mapped value) - 1 was sent verbatim */
		GETBITS(c, MErrval, c->qbpp);
		MErrval++;
	}

	oldmap = ( k == 0 && (RItype || MErrval) && (2 * c->B[Q] < Nt));
	/* 
	Note: the Boolean variable 'oldmap' is not 
	identical to the variable 'map' in the
	JPEG-LS draft. We have
	oldmap = (qdiff<0) ? (1-map) : map;
	*/

	MErrval += ( RItype + oldmap );

	if ( MErrval & 1 ) { /* negative */
		Errval = oldmap - (MErrval + 1) / 2;
		absErrval = -Errval - RItype;
		c->B[Q]++;
	}
	else { /* nonnegative */
		Errval = MErrval / 2;
		absErrval = Errval - RItype;
	}

	if ( Rb < Ra )
#    ifdef POW2
		Ix = ( Rb - Errval ) & (c->alpha - 1);
#    else
		Ix = Rb - Errval;
#    endif
	else   /* includes case a == b */
#    ifdef POW2
		Ix = ( Rb + Errval ) & (c->alpha - 1);
#    else
		Ix = Rb + Errval;
#    endif


#if !defined(POW2)
	/* reduce mod alpha, for arbitrary alpha */
	if (Ix < 0){
		Ix += c->alpha;
	}
	else if (Ix >= c->alpha){
		Ix -= c->alpha;
	}
#endif

	/* update stats */
	c->A[Q] += absErrval;
	if (c->N[Q] == c->theReset) {
		c->N[Q] >>= 1;
		c->A[Q] >>= 1;
		c->B[Q] >>= 1;
	}

	c->N[Q]++;  /* for next pixel */
	
	*returnPixel = (pixel)Ix;
	NORMAL_RETURN;
}


/* I don't imagine much demand for line-interleaved mode, but single-component images default */
/* to plane mode. / dhtesq. */
/* For line and plane interleaved mode in LOSSLESS mode */
/*** watch it! actual pixels in the scan line are numbered 1 to no .
     pixels with indices < 1 or > no are dummy "border" pixels  */
ErrCode
lossless_undoscanline(DgpJpegLsContext		*c,
					  pixel					*psl,			/* previous scanline */
					  pixel					*sl,			/* current scanline */
					  Dgp_Int32				no,				/* number of values in it */
					  Dgp_Int32				color)	
{
	Dgp_Int32 i, psfix;
	pixel Ra, Rb, Rc, Rd;
	Dgp_Int32 SIGN;
	Dgp_Int32 cont;
	Dgp_State *dgp_state = c->dgp_state;
	DGP_STATE_PROLOGUE("lossless_undoscanline()", dgp_state);

	psfix = 0;

	/**********************************************/
	/* Do for all pixels in the row in 8-bit mode */
	/**********************************************/
	if (c->bpp16 == DGP_FALSE){

		Rc = psl[0];
		Rb = psl[1];
		Ra = sl[0];

		i = 1;
		do {
			pixel Px;

			Rd = psl[i + 1];

			/* Quantize the gradient */
			cont =  c->vLUT[0][Rd - Rb + LUTMAX8] +
					c->vLUT[1][Rb - Rc + LUTMAX8] +
					c->vLUT[2][Rc - Ra + LUTMAX8];

			if ( cont == 0 ){
				/*********** RUN STATE **********/

				/*register*/ Dgp_Int32 n, m;

				/* get length of the run */
				/* arg is # of pixels left */
				DGP_ERR_PASS(process_run_dec(c, no - i + 1, color, &n));
				m = n;

				if ( m > 0 )  {  /* run of nonzero length, otherwise
									we go directly to the end-of-run 
									state */
					do {
						sl[i++] = Ra;
					} while(--n > 0);

					/* end of line? */
					if (i > no)	NORMAL_RETURN;

					/* update context pixels */
				    Rb = psl[i];
				    Rd = psl[i + 1];
				}

				/* Do end of run encoding for LOSSLESS images */
				DGP_ERR_PASS(lossless_end_of_run_d(c, Ra, Rb, (Ra == Rb), &Ra));
	
			}       /* Run state block */ 
			else {
				/************ REGULAR CONTEXT ************/

				PREDICT(Rb, Ra, Rc);

				/* map symmetric contexts */
				cont = c->classmap[cont];

				if (cont < 0) {
					SIGN = -1;
					cont = -cont;
				}
				else{
					SIGN = +1;
				}

				/* decode a Rice code of a given context */
				DGP_ERR_PASS(lossless_regular_mode_d(c, cont, SIGN, Px, &Ra));
			}

			sl[i] = Ra;
			Rc = Rb;
			Rb = Rd;
			++i;

		} while (i <= no);

	}
	else{
		/***********************************************/
		/* Do for all pixels in the row in 16-bit mode */
		/***********************************************/

		Rc = ENDIAN16(psl[0]);
		Rb = ENDIAN16(psl[1]);
		Ra = ENDIAN16(sl[0]);

		i = 1;
		do {
			pixel Px;

			Rd = ENDIAN16(psl[i + 1]);

			/* Quantize the gradient */
			{
				register Dgp_Int32 diff;

				/* Following segment assumes that T3 <= LUTMAX16. */
				/* This condition should have been checked when the	lookup tables were built. */
				diff = Rd - Rb;
				if (diff < 0) cont = (diff > -LUTMAX16) ? c->vLUT[0][diff + LUTMAX16] : 7 * CREGIONS  *CREGIONS;
				else cont = (diff < LUTMAX16) ? c->vLUT[0][diff + LUTMAX16] : 8 * CREGIONS * CREGIONS;

				diff = Rb - Rc;
				if (diff < 0) cont += (diff > -LUTMAX16) ? c->vLUT[1][diff + LUTMAX16] : 7 * CREGIONS;
				else cont += (diff < LUTMAX16) ? c->vLUT[1][diff + LUTMAX16] : 8 * CREGIONS;
				
				diff = Rc - Ra;
				if (diff < 0) cont += (diff > -LUTMAX16) ? c->vLUT[2][diff + LUTMAX16] : 7;
				else cont += (diff < LUTMAX16) ? c->vLUT[2][diff + LUTMAX16] : 8;
			}

			if ( cont == 0 ) {

				/********** RUN STATE **********/

				/*register*/ Dgp_Int32 n, m;

				/* get length of the run */
				/* arg is # of pixels left */
				DGP_ERR_PASS(process_run_dec(c, no - i + 1, color, &n));
				m = n;

				if ( m > 0 )  {  /* run of nonzero length, otherwise
									we go directly to the end-of-run 
									state */
					do {
						sl[i++] = ENDIAN16(Ra);
					}	while(--n > 0);

					/* end of line? */
					if (i > no) NORMAL_RETURN;

					/* update context pixels */
				    Rb = ENDIAN16(psl[i]);
				    Rd = ENDIAN16(psl[i + 1]);
				}

				/* Do end of run encoding for LOSSLESS images */
				DGP_ERR_PASS(lossless_end_of_run_d(c, Ra, Rb, (Ra == Rb), &Ra));
			}
			else {
				/********** REGULAR CONTEXT **********/
				PREDICT(Rb, Ra, Rc);

				/* map symmetric contexts */
				cont = c->classmap[cont];

				if (cont < 0) {
					SIGN = -1;
					cont = -cont;
				}
				else{
					SIGN = +1;
				}

				/* decode a Rice code of a given context */
				DGP_ERR_PASS(lossless_regular_mode_d(c, cont, SIGN, Px, &Ra));
			}

			sl[i] = ENDIAN16(Ra);
			Rc = Rb;
			Rb = Rd;
			++i;
		} while (i <= no);

	} /* End "if 8/16 bit" */

	NORMAL_RETURN;
}



/* For DEOCODING pixel interleavde mode for LOSSLESS images */
/*** watch it! actual pixels in the scan line are numbered 1 to no .
     pixels with indices < 1 or > no are dummy "border" pixels  */
ErrCode
lossless_undoscanline_pixel(DgpJpegLsContext	*c,
							pixel				*psl,		/* previous scanline */
							pixel				*sl,		/* current scanline */
							Dgp_Int32			no,			/* number of values in it */
							Dgp_Int32			*retval)	
{
	Dgp_Int32 i, psfix, n_c, color, enter_run = 0, was_in_run = 0, test_run;
	pixel Ra, Rb, Rc, Rd;
	pixel c_aa[MAX_COMPONENTS],
	      c_bb[MAX_COMPONENTS],
	      c_cc[MAX_COMPONENTS],
	      c_dd[MAX_COMPONENTS],
	      c_xx[MAX_COMPONENTS];
	Dgp_State *dgp_state = c->dgp_state;
	Dgp_Int32 SIGN;
	Dgp_Int32 cont, c_cont[MAX_COMPONENTS];
	Dgp_Int32 components = c->components;
	DGP_STATE_PROLOGUE("lossless_undoscanline_pixel()", dgp_state);
	DGP_ASSERT(psl);
	DGP_ASSERT(sl);
	DGP_ASSERT(retval);
	psfix = 0;

	/**********************************************/
	/* Do for all pixels in the row in 8-bit mode */
	/**********************************************/
	if (c->bpp16 == DGP_FALSE){

		for (n_c = 0; n_c < components; n_c++) {
			c_cc[n_c] = psl[n_c];
			c_bb[n_c] = psl[components + n_c];
			c_aa[n_c] = sl[n_c];
		}

		i = components;
		color = -1;

		do {
			pixel Px;

			if (!was_in_run) color = (color + 1) % components;
			else color = 0;

			if (color == 0)

				for (n_c = 0; n_c < components; n_c++) {
					c_dd[n_c] = psl[i + components + n_c];

					/* Quantize the gradient */		   
					c_cont[n_c] =	c->vLUT[0][c_dd[n_c] - c_bb[n_c] + LUTMAX8] +
									c->vLUT[1][c_bb[n_c] - c_cc[n_c] + LUTMAX8] +
									c->vLUT[2][c_cc[n_c] - c_aa[n_c] + LUTMAX8];
				}

			Ra = c_aa[color];
			Rb = c_bb[color];
			Rc = c_cc[color];
			Rd = c_dd[color];
			cont = c_cont[color];

			enter_run = was_in_run = test_run = 0;
		
			if (color == 0) {
				test_run = 1;
				for (n_c = 0; n_c < components; n_c++){
					if (c_cont[n_c] != 0) {
						test_run = 0;
						break;
					}
				}
			}

			if ( test_run ) {
				
				/********** RUN STATE *********/
				/* These locals were registers in original implementation, but error checking on process_run_dec() */
				/* requires that they be passed by reference. / dhtesq. */
				/*register*/ Dgp_Int32 n, m;
				
				enter_run = was_in_run = 1;
				
				/* get length of the run */
				/* arg is # of pixels left */
				DGP_ERR_PASS(process_run_dec(c, (no + components - 1 - i + 1) / components, 0, &n));
				m = n;
				
				if ( m > 0 )  {  /* run of nonzero length, otherwise
									we go directly to the end-of-run 
									state */
					do {
						for (n_c = 0; n_c < components; n_c++) {
							sl[i++] = c_aa[n_c];
						}
					} while(--n > 0);

					if (i > no + components - 1){
						/* end of line */
						*retval = 0;
						NORMAL_RETURN;
					}

					/* update context pixels */
					for (n_c = 0; n_c < components; n_c++) {
						c_bb[n_c] = psl[i + n_c];
						c_dd[n_c] = psl[i + components + n_c];
					}
				}

				/* here we handle the "end-of-run" stat */
				
				for (n_c = 0; n_c < components; n_c++) {
					/* The end of run is processed for each component */
					Ra = c_aa[n_c];
					Rb = c_bb[n_c];
					DGP_ERR_PASS(lossless_end_of_run_d(c, Ra, Rb, 0, &c_xx[n_c]));
					c_aa[n_c] = c_xx[n_c];
				}       /* Components loop */
			}       /* Run state block */
			else {

				/******* REGULAR CONTEXT *******/

				PREDICT(Rb, Ra, Rc);
		
				cont = c->classmap[cont];
				if (cont < 0) {
					SIGN = -1;
					cont = -cont;
				}
				else {
					SIGN = +1;
				}
				
				/* decode a Rice code of a given context */
				DGP_ERR_PASS(lossless_regular_mode_d(c, cont, SIGN, Px, &Ra));
				c_aa[color] = Ra;
			}
			
			if (!was_in_run) {
				sl[i] = Ra;
				c_cc[color] = Rb;
				c_bb[color] = Rd;
				i++;
			}
			else {
				for (n_c = 0; n_c < components; n_c++) {
					sl[i+n_c] = c_aa[n_c];
					c_cc[n_c] = c_bb[n_c];
					c_bb[n_c] = c_dd[n_c];
				}
				i += components;
			}
		} while (i <= (no + components - 1));
	}
	else{
	/***********************************************/
	/* Do for all pixels in the row in 16-bit mode */
	/***********************************************/

		for (n_c = 0; n_c < components; n_c++) {
			c_cc[n_c] = ENDIAN16(psl[n_c]);
			c_bb[n_c] = ENDIAN16(psl[components + n_c]);
			c_aa[n_c] = ENDIAN16(sl[n_c]);
		}

		i = components;
		color = -1;

		do {
			pixel Px;

			if (!was_in_run) color = (color + 1) % components;
			else color = 0;

			if (color == 0)
				for (n_c = 0; n_c < components; n_c++) {

					c_dd[n_c] = ENDIAN16(psl[i + components + n_c]);

					/* Quantize the gradient */
					{
						register Dgp_Int32 diff;

						/* Following segment assumes that T3 <= LUTMAX16 */
						/* This condition should have been checked when the
						   lookup tables were built. */
						diff = c_dd[n_c] - c_bb[n_c];
						if (diff < 0){
							c_cont[n_c] = (diff > -LUTMAX16) ? c->vLUT[0][diff + LUTMAX16] : 7 * CREGIONS * CREGIONS;
						}
						else {
							c_cont[n_c] = (diff < LUTMAX16) ? c->vLUT[0][diff + LUTMAX16] : 8 * CREGIONS * CREGIONS;
						}
						diff = c_bb[n_c] - c_cc[n_c];
						if (diff < 0){
							c_cont[n_c] += (diff > -LUTMAX16) ? c->vLUT[1][diff + LUTMAX16] : 7 * CREGIONS;
						}
						else {
							c_cont[n_c] += (diff < LUTMAX16) ? c->vLUT[1][diff + LUTMAX16] : 8 * CREGIONS;
						}

						diff = c_cc[n_c] - c_aa[n_c];
						if (diff < 0){
							c_cont[n_c] += (diff > -LUTMAX16) ? c->vLUT[2][diff + LUTMAX16] : 7;
						}
						else {
							c_cont[n_c] += (diff < LUTMAX16) ? c->vLUT[2][diff + LUTMAX16] : 8;
						}
					}
				}

			Ra = c_aa[color];
			Rb = c_bb[color];
			Rc = c_cc[color];
			Rd = c_dd[color];
			cont = c_cont[color];

			enter_run = was_in_run = test_run = 0;
		
			if (color == 0) {
				test_run = 1;
				for (n_c = 0; n_c < components; n_c++){
					if (c_cont[n_c] != 0) {
						test_run = 0;
						break;
					}
				}
			}

			if ( test_run ) {

				/********* RUN STATE *********/

				/* These locals were registers in original implementation, but error checking on process_run_dec() */
				/* requires that they be passed by reference. / dhtesq. */
				/*register*/ Dgp_Int32 n, m;

				enter_run = was_in_run = 1;

				/* get length of the run */
				/* arg is # of pixels left */
				DGP_ERR_PASS(process_run_dec(c, (no + components - 1 - i + 1) / components, 0, &n));
				m = n;

				if ( m > 0 )  {  /* run of nonzero length, otherwise
									we go directly to the end-of-run 
									state */
					do {
						for (n_c = 0; n_c < components; n_c++) {
							sl[i++] = ENDIAN16(c_aa[n_c]);
						}
					} while(--n > 0);

					if (i > no + components - 1){
						/* end of line */
						*retval = 0;
						NORMAL_RETURN;
					}

					/* update context pixels */
					for (n_c = 0; n_c < components; n_c++) {
			    		c_bb[n_c] = ENDIAN16(psl[i + n_c]);
			    		c_dd[n_c] = ENDIAN16(psl[i + components + n_c]);
					}
				}

				/* here we handle the "end-of-run" state */
				for (n_c = 0; n_c < components; n_c++) {
					/* The end of run is processed for each component */
					Ra = c_aa[n_c];
					Rb = c_bb[n_c];

					DGP_ERR_PASS(lossless_end_of_run_d(c, Ra, Rb, 0, &c_xx[n_c]));
					c_aa[n_c] = c_xx[n_c];
				}   /* Components loop */
			}       /* Run state block */
			else {

				/******** REGULAR CONTEXT ********/

				PREDICT(Rb, Ra, Rc);
			
				cont = c->classmap[cont];
			
				if (cont < 0) {
					SIGN = -1;
					cont = -cont;
				}
				else{
					SIGN = +1;
				}

				/* decode a Rice code of a given context */
				DGP_ERR_PASS(lossless_regular_mode_d(c, cont, SIGN, Px, &Ra));
				c_aa[color] = Ra;

			}

			if (!was_in_run) {
				sl[i] = ENDIAN16(Ra);
				c_cc[color] = Rb;
				c_bb[color] = Rd;
				i++;
			}
			else {
				for (n_c = 0; n_c < components; n_c++) {
					sl[i + n_c] = ENDIAN16(c_aa[n_c]);
					c_cc[n_c] = c_bb[n_c];
					c_bb[n_c] = c_dd[n_c];
				}
				i += components;
			}

		} while (i <= (no + components - 1));

	} /* ends "if 8/16 bit */

	*retval = 0;
	NORMAL_RETURN;
}
