/* SPMG/JPEG-LS IMPLEMENTATION V.2.1
   =====================================
   These programs are Copyright (c) University of British Columbia. All rights reserved.
   They may be freely redistributed in their entirety provided that this copyright
   notice is not removed.  They may not be sold for profit or incorporated in
   commercial programs without the written permission of the copyright holder.
   Each program is provided as is, without any express or implied warranty,
   without even the warranty of fitness for a particular purpose.

   =========================================================
   THIS SOFTWARE IS BASED ON HP's implementation of jpeg-ls:
   =========================================================

   LOCO-I/JPEG-LS IMPLEMENTATION V.0.90
   -------------------------------------------------------------------------------
   (c) COPYRIGHT HEWLETT-PACKARD COMPANY, 1995-1997.
        HEWLETT-PACKARD COMPANY ("HP") DOES NOT WARRANT THE ACCURACY OR
   COMPLETENESS OF THE INFORMATION GIVEN HERE.  ANY USE MADE OF, OR
   RELIANCE ON, SUCH INFORMATION IS ENTIRELY AT USER'S OWN RISK.
        BY DOWNLOADING THE LOCO-I/JPEG-LS COMPRESSORS/DECOMPRESSORS
   ("THE SOFTWARE") YOU AGREE TO BE BOUND BY THE TERMS AND CONDITIONS
   OF THIS LICENSING AGREEMENT.
        YOU MAY DOWNLOAD AND USE THE SOFTWARE FOR NON-COMMERCIAL PURPOSES
   FREE OF CHARGE OR FURTHER OBLIGATION.  YOU MAY NOT, DIRECTLY OR
   INDIRECTLY, DISTRIBUTE THE SOFTWARE FOR A FEE, INCORPORATE THIS
   SOFTWARE INTO ANY PRODUCT OFFERED FOR SALE, OR USE THE SOFTWARE
   TO PROVIDE A SERVICE FOR WHICH A FEE IS CHARGED.
        YOU MAY MAKE COPIES OF THE SOFTWARE AND DISTRIBUTE SUCH COPIES TO
   OTHER PERSONS PROVIDED THAT SUCH COPIES ARE ACCOMPANIED BY
   HEWLETT-PACKARD'S COPYRIGHT NOTICE AND THIS AGREEMENT AND THAT
   SUCH OTHER PERSONS AGREE TO BE BOUND BY THE TERMS OF THIS AGREEMENT.
        THE SOFTWARE IS NOT OF PRODUCT QUALITY AND MAY HAVE ERRORS OR DEFECTS.
   THE JPEG-LS STANDARD IS STILL UNDER DEVELOPMENT. THE SOFTWARE IS NOT A
   FINAL OR FULL IMPLEMENTATION OF THE STANDARD.  HP GIVES NO EXPRESS OR
   IMPLIED WARRANTY OF ANY KIND AND ANY IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR PURPOSE ARE DISCLAIMED.
        HP SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
   OR CONSEQUENTIAL DAMAGES ARISING OUT OF ANY USE OF THE SOFTWARE.
   -------------------------------------------------------------------------------
*/

/* melcode.c --- for processing in run mode
 *
 * Initial code by Alex Jakulin,  Aug. 1995
 *
 * Modified and optimized: Gadiel Seroussi, October 1995
 *
 * Color Enhancement: Guillermo Sapiro, August 1996
 *
 * Modified and added Restart marker and input tables by:
 * David Cheng-Hsiu Chu, and Ismail R. Ismail march 1999
 *
 * 20.3.2001 / dhtesq.
 * Added to digipaper code base for evaluation purposes.
 */

#include <stdio.h>
#include "dgpjpegls.h"
#include "bitio.h"

#define MELCSTATES	32	/* number of melcode states */

static J[MELCSTATES] = {
	0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3,4,4,5,5,6,6,7,
	7,8,9,10,11,12,13,14,15 
};


/*
38.3.2001 / dhtesq.
Modified to return a melcode context.
melcstate, melclen, and melcorder are no longer statics.
*/
ErrCode
init_process_run(DgpJpegLsContext *context,
				 Dgp_Int32 components,
				 Dgp_Int32 maxrun)    /* maxrun is ignored when using MELCODE; kept here for function compatibility. */
{
	Dgp_Int32	n_c;
	Dgp_State	*dgp_state = context->dgp_state;
	DGP_STATE_PROLOGUE("init_process_run()", dgp_state);
	DGP_ASSERT(context);

	maxrun = maxrun;

	context->melclen = NULL;
	context->melcorder = NULL;
	context->melcstate = NULL;

	DGP_CHECK_ALLOC_CLEANUP(context->melclen = (Dgp_Int32 *)DGP_MALLOC(sizeof(*(context->melclen)) * MAX_COMPONENTS));
	DGP_CHECK_ALLOC_CLEANUP(context->melcorder = (Dgp_Int32 *)DGP_MALLOC(sizeof(*(context->melcorder)) * MAX_COMPONENTS));
	DGP_CHECK_ALLOC_CLEANUP(context->melcstate = (Dgp_Int32 *)DGP_MALLOC(sizeof(*(context->melcstate)) * MAX_COMPONENTS));

	for (n_c = 0; n_c < components; n_c++)
	{
		context->melcstate[n_c] = 0;
		context->melclen[n_c] = J[0];
		context->melcorder[n_c] = 1 << context->melclen[n_c];
	}

	NORMAL_RETURN;

ERR_CLEANUP:
	if(context->melclen) DGP_FREE(context->melclen);
	if(context->melcorder) DGP_FREE(context->melcorder);
	if(context->melcstate) DGP_FREE(context->melcstate);
	ERR_RETURN;
}

/* Encoder function missing... */
void
process_run(Dgp_Int32			runCount, 
			Dgp_Bool			isEndOfLine, 
			Dgp_Int32			color){
	return;
}


/* decoding routine: reads bits from the input and returns a run length. */
/* argument is the number of pixels left to	end-of-line (bound on run length) */
ErrCode
process_run_dec(DgpJpegLsContext *context,
				Dgp_Int32		lineleft, 
				Dgp_Int32		color,
				Dgp_Int32		*retLineleft)
{
	Dgp_Int32 runlen = 0;
	DGP_STATE_PROLOGUE("process_run_dec()", context->dgp_state);
	DGP_ASSERT(retLineleft);
		
	do {
		register temp, hits;
		temp = context->zeroLUT[(Dgp_UInt8)(~(context->reg >> 24))];   /* Number of leading ones in the input stream, up to 8. */
		for ( hits = 1; hits <= temp; hits++ ) {
			runlen += context->melcorder[color];

			if ( runlen >= lineleft ) { /* reached end-of-line */
				if ( runlen == lineleft && context->melcstate[color] < MELCSTATES ) {
					context->melclen[color] = J[++context->melcstate[color]];
					context->melcorder[color] = (1L << context->melclen[color]);
				}
				DGP_ERR_PASS(fillbuffer(context, hits)); /* actual # of 1's consumed */
				*retLineleft = lineleft;
				NORMAL_RETURN;
			}
			if ( context->melcstate[color] < MELCSTATES ) {
				context->melclen[color] = J[++context->melcstate[color]];
				context->melcorder[color] = (1L << context->melclen[color]);
			}
		}	
		if (temp != 8) {
			DGP_ERR_PASS(fillbuffer(context, temp + 1)); /* consume the leading 0 of the remainder encoding */
			break;
        }
		DGP_ERR_PASS(fillbuffer(context, 8));
	} while ( DGP_TRUE );

	/* Read the length of the remainder. */
	if ( context->melclen[color] ) {
		register temp;
		GETBITS(context, temp, context->melclen[color]);  /*** GETBITS is a macro, not a function */
		runlen += temp;
	}
	context->limit_reduce = context->melclen[color] + 1;

	/* adjust melcoder parameters */
	if ( context->melcstate[color] ) {
		context->melclen[color] = J[--context->melcstate[color]];
		context->melcorder[color] = (1L << context->melclen[color]);
	}

	*retLineleft = runlen;
	NORMAL_RETURN;
}


/* retained for compatibility with ranked runs */
/* Also, now frees context and associated memory. / dhtesq. */
ErrCode
close_process_run(DgpJpegLsContext *context)
{
	Dgp_State *dgp_state = context->dgp_state;
	DGP_STATE_PROLOGUE("close_process_run()", dgp_state);
	DGP_ASSERT(context);

	if(context->melclen) DGP_FREE(context->melclen);
	if(context->melcorder) DGP_FREE(context->melcorder);
	if(context->melcstate) DGP_FREE(context->melcstate);

	context->melclen = NULL;
	context->melcorder = NULL;
	context->melcstate = NULL;

	NORMAL_RETURN;
}
