
#include "allocators_test.h"
#include "stdlib.h"
#include "memory.h"


Dgp_Int32 dgpmem_max_mem_usage = DGPMEM_NOMEMLIMIT;

static Dgp_Int32 dgpmem_peak_mem_usage = 0;
Dgp_Int32 dgpmem_mem_usage = 0;


static Dgp_Int32 dgpmem_num_mallocs = 0;
static Dgp_Int32 dgpmem_num_callocs = 0;
static Dgp_Int32 dgpmem_num_reallocs = 0;
static Dgp_Int32 dgpmem_num_frees = 0;

Dgp_Int32
DgpMem_getCurrentMemUsage()
{
  return dgpmem_mem_usage;
}


Dgp_Int32
DgpMem_getPeakMemUsage()
{
  return dgpmem_peak_mem_usage;
}

void
DgpMem_resetPeakMemUsage()
{
  dgpmem_peak_mem_usage = dgpmem_mem_usage;
}

void
DgpMem_getMemStats(Dgp_Int32 *num_mallocs, 
                 Dgp_Int32 *num_callocs,
                 Dgp_Int32 *num_reallocs,
                 Dgp_Int32 *num_frees)
{
  if(num_mallocs != NULL) *num_mallocs = dgpmem_num_mallocs;
  if(num_callocs != NULL) *num_callocs = dgpmem_num_callocs;
  if(num_reallocs != NULL) *num_reallocs = dgpmem_num_reallocs;
  if(num_frees != NULL) *num_frees = dgpmem_num_frees;
}

void
DgpMem_resetMemStats()
{
  dgpmem_num_mallocs = 0;
  dgpmem_num_callocs = 0;
  dgpmem_num_reallocs = 0;
  dgpmem_num_frees = 0;
}


void * 
DgpMem_mallocTest(size_t size)
{
  size_t new_size = size + sizeof(size_t);
  void *ret_ptr;

  dgpmem_num_mallocs++;

  if((dgpmem_max_mem_usage != DGPMEM_NOMEMLIMIT) && 
     ((dgpmem_mem_usage + (Dgp_Int32)size) > dgpmem_max_mem_usage)) {
    return NULL;
  }

  ret_ptr = malloc(new_size);

  if(ret_ptr == NULL) return NULL;

  *((size_t *)ret_ptr) = size;

  dgpmem_mem_usage += size;

  if(dgpmem_mem_usage > dgpmem_peak_mem_usage) dgpmem_peak_mem_usage = dgpmem_mem_usage;

  return((void *)(((char *)ret_ptr) + sizeof(size_t)));
}

void 
DgpMem_freeTest(void *ptr)
{

  dgpmem_num_frees++;

  if(ptr == NULL) return;

  ((char *)ptr) -= sizeof(size_t);

  dgpmem_mem_usage -= *((size_t *)ptr);

  free(ptr);
}

void * 
DgpMem_callocTest(size_t _nobj, size_t _size)
{
  size_t size = _nobj * _size;
  size_t new_size = size + sizeof(size_t);
  void *ret_ptr;

  dgpmem_num_callocs++;

  if((dgpmem_max_mem_usage != DGPMEM_NOMEMLIMIT) && 
     ((dgpmem_mem_usage + (Dgp_Int32)size) > dgpmem_max_mem_usage)) {
    return NULL;
  }

  ret_ptr = malloc(new_size);

  if(ret_ptr == NULL) return NULL;

  memset(ret_ptr, 0, new_size);

  *((size_t *)ret_ptr) = size;

  dgpmem_mem_usage += size;

  if(dgpmem_mem_usage > dgpmem_peak_mem_usage) dgpmem_peak_mem_usage = dgpmem_mem_usage;

  return((void *)(((char *)ret_ptr) + sizeof(size_t)));
}

void * 
DgpMem_reallocTest(void *ptr, size_t size)
{
  size_t new_size = size + sizeof(size_t);
  size_t old_size = 0;
  void *ret_ptr;

  dgpmem_num_reallocs++;

  if(ptr != NULL) {
    ((char *)ptr) -= sizeof(size_t);
    old_size = *((size_t *)ptr);
  }

  if((dgpmem_max_mem_usage != DGPMEM_NOMEMLIMIT) && 
     ((dgpmem_mem_usage + (Dgp_Int32)size - (Dgp_Int32)old_size) > 
      dgpmem_max_mem_usage)) {
    return NULL;
  }

  ret_ptr = realloc(ptr, new_size);

  if(ret_ptr == NULL) return NULL;

  *((size_t *)ret_ptr) = size;

  dgpmem_mem_usage += size - old_size;

  if(dgpmem_mem_usage > dgpmem_peak_mem_usage) dgpmem_peak_mem_usage = dgpmem_mem_usage;

  return((void *)(((char *)ret_ptr) + sizeof(size_t)));
}

void
DgpMem_setMemLimit(Dgp_Int32 max_mem_usage)
{
  if(max_mem_usage < DGPMEM_NOMEMLIMIT) max_mem_usage = DGPMEM_NOMEMLIMIT;

  dgpmem_max_mem_usage = max_mem_usage;

}


