

#include "allocators_test.h"
#include "PsgFix.h"

extern Dgp_Int32 dgpmem_max_mem_usage;
extern Dgp_Int32 dgpmem_mem_usage;

static Dgp_UInt32 dgpmem_cur_pool_num[DGP_NUM_SMALLPOOLS] =  {0, 0, 0, 0, 0, 0, 0, 0};
static Dgp_UInt32 dgpmem_peak_pool_num[DGP_NUM_SMALLPOOLS] = {0, 0, 0, 0, 0, 0, 0, 0};

ErrCode
DgpMem_mallocSmall(void *small_cookie, 
                   Dgp_Int32 pool_id, 
                   void **retptr)
{
  DGP_PROLOGUE("DgpMem_mallocSmall", NULL);
  
  DGP_ASSERT((pool_id >= 0) && (pool_id < DGP_NUM_SMALLPOOLS));

  small_cookie = small_cookie;

  if((dgpmem_max_mem_usage != DGPMEM_NOMEMLIMIT) && 
    ((Dgp_Int32)(dgpmem_mem_usage + Dgp_getPoolSize(pool_id)) > 
     dgpmem_max_mem_usage)) {
    *retptr = NULL;
    DGP_ERR(cErrNomem);
  }

  DGP_CHECK_ALLOC(*retptr = malloc(Dgp_getPoolSize(pool_id)));

  dgpmem_cur_pool_num[pool_id]++;

  if(dgpmem_cur_pool_num[pool_id] > dgpmem_peak_pool_num[pool_id]) {
    dgpmem_peak_pool_num[pool_id] = dgpmem_cur_pool_num[pool_id];
  }

  NORMAL_RETURN;
}

void
DgpMem_freeSmall(void *small_cookie,
                 Dgp_Int32 pool_id, 
                 void *ptr)
{
  DGP_VOID_PROLOGUE("DgpMem_freeSmall()", NULL);
  
  DGP_VOID_ASSERT((pool_id >= 0) && (pool_id < DGP_NUM_SMALLPOOLS));
  DGP_VOID_ASSERT(ptr != NULL);
  small_cookie = small_cookie;

  free(ptr);

  dgpmem_cur_pool_num[pool_id]--;
}

void
DgpMem_getSmPoolUsage(Dgp_Int32 *cur_num_pool_els,
                  Dgp_Int32 *peak_num_pool_els,
                  Dgp_Int32 num_pools)
{
  Dgp_Int32 i;

  num_pools = MIN(num_pools, DGP_NUM_SMALLPOOLS);
  for(i = 0; i < num_pools; i++) {
    if(cur_num_pool_els != NULL) cur_num_pool_els[i] = dgpmem_cur_pool_num[i];
    if(peak_num_pool_els != NULL) peak_num_pool_els[i] = dgpmem_peak_pool_num[i];
  }

}

void
DgpMem_resetSmPoolMUsage()
{
  Dgp_Int32 i;

  for(i = 0; i < DGP_NUM_SMALLPOOLS; i++) {
    dgpmem_peak_pool_num[i] = dgpmem_cur_pool_num[i];
  }

}
