/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

#if 0

/************************************************************************
 * iorbuf.c
 *
 *	IO data buffering/queuing for T.44 writing
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_IO, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#ifdef _MSC_VER
#pragma pack(pop, PRE_IO)
#endif

#include "iobuf.h"

IP_RCSINFO(RCSInfo, "$RCSfile: iobuf.c,v $; $Revision: 1.1 $; $Date: 2003/01/03 01:41:09 $");


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image description and fill in the ImgInfo structure.
 *
 * LongDescription:::
 *  wp_IoGetImageDesc() reads the IFD and fills in the ImgInfo structure with
 *  the basic image description as interpreted from the IFD.
 *
 * Input:::
 *  ioImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_InsertBuffer(
    IO_DOC *ioDoc,
    IOBufList *listHead,
    IOBuf *prev,
    UInt8 *buffer,
    UInt32 bufferLength
)
{
    IO_INSTANCE *instData;
    IO_FILE *ioFile;
    IOBuf *newIOBuf;
    PROLOGUE;

    IO_CheckHandle(ioDoc);
    ioFile = IO_DOCFILE(ioDoc);

    /* Check parameters */
    XERR_GEN_CLEANUP(listHead == NULL || prev == NULL, IO_BADPARAMETER);
    
    IO_CHECKMALLOC(newIOBuf = IO_MALLOC(sizeof(IOBuf)));
    newIOBuf->prev = prev;
    newIOBuf->next = prev->next;
    newIOBuf->buffer = buffer;
    newIOBuf->bufferLength = bufferLength;
    newIOBuf->bufferFilled = 0;
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}

/***
*/
ErrCode CDECL
IO_RemoveAndFreeBuffer(
    IO_DOC *ioDoc,
    IOBufList *listHead,
    IOBuf *item
)
{
    IO_INSTANCE *instData;
    IO_FILE *ioFile;
    PROLOGUE;

    IO_CheckHandle(ioDoc);
    ioFile = IO_DOCFILE(ioDoc);

    /* Check parameters */
    XERR_GEN_CLEANUP(listHead == NULL || item == NULL, IO_BADPARAMETER);

    /* Have the previous item skip this item */
    if (item->prev) {
	item->prev->next = item->next;
    }

    /* Have the next item skip this item */
    if (item->next) {
	item->next->prev = item->prev;
    }

    /* Start the list on the next item, if this is the first item */
    if (listHead->first == item) {
	listHead->first = item->next;
    }

    /* Free the data associated with this item */
    if (item->buffer != NULL && item->bufferLength > 0) {
	IO_FREE(item->buffer);
	item->buffer = NULL;	/* Avoid possible references to the buffer */
	item->bufferLength = 0;
    }

    /* Free the data associated with the structure */
    IO_FREE(item);
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}

/***
 */
ErrCode CDECL
IO_FreeBufferList(
    IO_DOC *ioDoc,
    IOBufList *listHead
)
{
    IO_INSTANCE *instData;
    IO_FILE *ioFile;
    IOBuf *item, *nextItem;
    PROLOGUE;

    IO_CheckHandle(ioDoc);
    ioFile = IO_DOCFILE(ioDoc);

    /* Check parameters */
    XERR_GEN_CLEANUP(listHead == NULL || item == NULL, IO_BADPARAMETER);

    item = listHead->first;

    while (item != NULL) {

	/* Keep track of the next item, while this one is being removed */
	nextItem = item->next;

	/* Free the data associated with this item */
	if (item->buffer != NULL && item->bufferLength > 0) {
	    IO_FREE(item->buffer);
	    item->buffer = NULL;	/* Avoid possible references to the buffer */
	    item->bufferLength = 0;
	}

	/* Free the data associated with the structure */
	IO_FREE(item);

	item = nextItem;
    }
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


#endif

/*
$Log: iobuf.c,v $
Revision 1.1  2003/01/03 01:41:09  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.4  2001/12/26 21:23:42  dabercro
Mark as released

Revision 1.3  2001/01/18 00:21:27  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.2  2000/04/12 19:06:06  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.1.1.1  2000/02/02 00:40:50  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

*/
