/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

#ifndef IODOC_H_INCLUDED
#define IODOC_H_INCLUDED

#ifndef NOIPCORE
IP_RCSINFO(iodoc_hRCSInfo, "$RCSfile: iodoc.h,v $; $Revision: 1.1 $");
/* $Date: 2003/01/03 01:41:10 $ */
#endif

#ifdef __cplusplus
extern "C"
{
#endif

/* iodoc.h
 *
 * DESCRIPTION
 *	 Declarations for IO Document level structures.
 *
 */

/*
 * INCLUDES
 */
#include "iofile.h"
#include "iomrc.h"
#include "ioinfo.h"

/*
 * CONSTANTS
 */
#if 0
/* Calling sequence status */
#define ioDoc_SEQ_GETINFO	0x0001	/* Document information gotten/set */
#define ioDoc_SEQ_READSTART	0x0010	/* Read start called */
#define ioDoc_SEQ_READFINISH	0x0000	/* Read finish called */
#define ioDoc_SEQ_READSTATUS	0x00f0	/* Read status bits */
#define ioDoc_SEQ_WRITESTART	0x0100	/* Write start called */
#define ioDoc_SEQ_W_PAGETABLE	0x0200	/* A page table was written */
#define ioDoc_SEQ_W_SYMLIBTABLE	0x0400	/* A symbol library table was written */
#define ioDoc_SEQ_WRITEFINISH	0x0000	/* Write finish called */
#define ioDoc_SEQ_WRITESTATUS	0x0f00	/* Write status bits */
#define ioDoc_SEQ_ANNOTSET	0x1000	/* Annotation set */
#define ioDoc_SEQ_REFPAGESET	0x2000	/* Reference page set */
#define ioDoc_SEQ_TOKENSTART	0x4000	/* Tokenization ongoing */
#define ioDoc_SEQ_TOKENDICTSAVED	0x8000	/* Tokenization dictionary saved */

#define READ_ALLOWED		0x1	/* File has been opened for read */
#define WRITE_ALLOWED		0x2	/* File has been opened for write */

/* Token dictionary flags */
#define TOKENDICT_WRITABLE	0x01	/* Dictionary is writable (not written
					 * yet) */
#define TOKENDICT_LOSSLESS	0x02	/* Token dictionary is for lossless
					 * compression only */

#endif

/*
 * MACROS
 */
#define IO_DOCFILE(ioDoc)	ioDoc->ioFile
#define IO_DOCINST(ioDoc)	IO_DOCFILE(ioDoc)->instance
#define IO_DOCFILETOKEN(ioDoc)	IO_FILETOKEN(IO_DOCFILE(ioDoc))

#if 0


/*
 * TYPEDEFS
 */

/* Page table header */
typedef struct PageTableHeader {
    UInt16 pageCount;		    /* Number of pages in document (total) */
    UInt16 entriesPerTable;	    /* Number of page table entries per table
				     */
    UInt16 lastOrdinal;		    /* Ordinal (1, 2, ...) of last table */
    Int32 firstTableOffset;	    /* File offset of first table */
} PageTableHeader;

/* Page table entry (describes an individual page) */
typedef struct PageTableEntry {
    UInt8 flags;		    /* Page flags (see PAGE_* above) */
    Int32 IFDoffset;		    /* File offset of the IFD */
} PageTableEntry;

/* Individual page table, linked if more than one */
typedef struct PageTable {
    struct PageTable *next;	    /* Link to next table (cached) */
    UInt16 ordinal;		    /* Ordinal (1, 2, ...) of this table */
    UInt16 validPages;		    /* Number of valid pages in table */
    struct PageTableEntry *entries; /* Allocated entry list */
    Int32 nextTableOffset;	    /* File offset of next page table */
} PageTable;
	
/* Symbol library table header */
typedef struct SymlibTableHeader {
    UInt16 symlibCount;		    /* Number of symlibs in document (total) */
    UInt16 entriesPerTable;	    /* Number of symlib table entries per table */
    UInt16 lastOrdinal;		    /* Ordinal (1, 2, ...) of last table */
    Int32 firstTableOffset;	    /* File offset of first table */
} SymlibTableHeader;

/* Symlib table entry (describes an individual symlib) */
typedef struct SymlibTableEntry {
    UInt32 bytes;		    /* Number of bytes in dictionary block */
    Int32 offset;		    /* File offset of the dictionary block */
} SymlibTableEntry;

/* Individual symlib table, linked if more than one */
typedef struct SymlibTable {
    struct SymlibTable *next;	    /* Link to next table (cached) */
    UInt16 ordinal;		    /* Ordinal (1, 2, ...) of this table */
    UInt16 validSymlibs;	    /* Number of valid symlibs in table */
    SymlibTableEntry *entries;	    /* Allocated entry list */
    Int32 nextTableOffset;	    /* File offset of next symlib table */
} SymlibTable;

/* Previously copied token dictionaries */
typedef struct CopiedDicts {
    struct CopiedDicts *next;	    /* Link to next */
    IO_DOCHANDLE dstDocHandle;	    /* Doc handle of destination */
    UInt32 dstDocID;		    /* Unique doc ID */
    UInt32 dictOffset;		    /* Offset of source dictionary */
    UInt32 newOffset;		    /* Offset of destinary dictionary */
} CopiedDicts;

/* Document status */
typedef struct IO_DOC {
    UInt32 handleID;			/* Handle ID, consistency check */
    DOCTYPE IODocType;			/* Document format: XIF, T.44, ... */
    UInt32 docID;			/* Unique document ID, from instance */
    XFile *xFile;			/* File structure */
    UInt32 seqStatus;			/* Calling sequence status */
    IO_DOCINFO docInfo;			/* Document information */
    UInt8 readWrite;			/* Read/write state (READ_ALLOWED/
					 * WRITE_ALLOWED) */
    Int32 tagOffset[DOCTAG_COUNT];	/* Offset of DocTags */
    Int32 pageTableHeaderOffset;	/* Offset to page table header */
    Int32 symlibTableHeaderOffset;	/* Offset to symlib table header */
    Int32 authorOffset;			/* Offset to author block */
    Int32 refPageOffset;		/* Offset to reference page */
    Int16 annotsPerTable;		/* User set annotations per table */
    UInt32 dictFlags;			/* Flags associated with current
					 * dictionary */
    struct PageTableHeader pageTableHeader;/* Page table header */
    struct PageTable *firstPageTable;/* Page table for first table */
    struct SymlibTableHeader symlibTableHeader;/* Symbol library table header
						*/
    struct SymlibTable *firstSymlibTable;/* Symbol library table for first */
    struct IO_PAGE *pages;		/* Page handles for pages previously
					 * visited */
    void *textCompInfo;			/* Pointer to shared text compression
					 * info */
    CopiedDicts *prevTDCopied;		/* Previous token dictionaries copied
					 * to other documents */
    UInt32 *curDictIDs;			/* List of token dictionaries
					 * currently loaded */
    UInt32 curDictCount;		/* Number of dictionaries currently
					 * loaded */
} IO_DOC;

/* Document field */
typedef struct DocField {
    UInt16 tagID;			/* Document tag */
    Int32 offset;			/* Offset of extension data */
} DocField;

#endif

/* Document status */
typedef struct IO_DOC {
    UInt32 handleID;		/* Handle ID, consistency check */
    UInt32 handleType;		/* Handle type ID */
    DOCTYPE ioDocType;		/* Document format: XIF, T.44, ... */
    UInt32 ioDocHandle;		/* Lower level document handle */
    struct IO_FILE *ioFile;	/* File structure */
    struct IO_PAGE *ioPages;	/* First in linked list of pages */
    struct IO_DOCINFO docInfo;	/* Document information */
} IO_DOC;




/*
 * PROTOTYPES
 */


#ifdef __cplusplus
}
#endif

#endif /* IODOC_H_INCLUDED */

