/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

#ifndef IOIMAGE_H_INCLUDED
#define IOIMAGE_H_INCLUDED

#ifndef NOIPCORE
IP_RCSINFO(ioimage_hRCSInfo, "$RCSfile: ioimage.h,v $; $Revision: 1.1 $");
/* $Date: 2003/01/03 01:41:10 $ */
#endif

#ifdef __cplusplus
extern "C"
{
#endif

/* ioimage.h
 *
 * DESCRIPTION
 *	 Declarations for the IO Toolkit API.
 *
 */

/*
 * INCLUDES
 */
#ifndef NOIPCORE
#ifdef _MSC_VER
#pragma pack(push, PRE_IO, 1)
#endif
#include "ipwcore.prv"
#include "geomadt.h"
#include "props.pub"
#include "rast.pub"
#include "pixr.h"
#ifdef _MSC_VER
#pragma pack(pop, PRE_IO)
#endif
#endif	/* NOIPCORE */

#include "iopage.h"
#include "ioinfo.h"

#if 0
/*
 * CONSTANTS
 */
#define MAX_EXPECTED_TAG_COUNT	22	/* Most tags an individual image
				         * is expected to have */

/* Calling sequence status */
#define ioImg_SEQ_GETINFO	0x00000001	/* Get image info called */
#define ioImg_SEQ_READSTART	0x00000010	/* Read start called */
#define ioImg_SEQ_READSTRIP	0x00000020	/* Read strip called */
#define ioImg_SEQ_READFINISH	0x00000000	/* Read finish called */
#define ioImg_SEQ_READSTATUS	0x000000f0	/* Read status bits */
#define ioImg_SEQ_WRITESTART	0x00000100	/* Write start called */
#define ioImg_SEQ_WRITESTRIP	0x00000200	/* Write strip called */
#define ioImg_SEQ_WRITEFINISH	0x00000000	/* Write finish called */
#define ioImg_SEQ_WRITESTATUS	0x00000f00	/* Write status bits */
#define ioImg_SEQ_TOKENSTART	0x00001000	/* Image is in a tokenized state
						 */
#define ioImg_SEQ_MASKADDED	0x00010000	/* Image has had a mask added */

#endif

/*
 * MACROS
 */
#define IO_IMGFILE(ioImg)		IO_PAGEFILE(ioImg->ioPage)
#define IO_IMGINST(ioImg)		IO_IMGFILE(ioImg)->instance
#define IO_IMGFILETOKEN(ioImg)		IO_FILETOKEN(IO_IMGFILE(ioImg))

#if 0

/*
 * TYPEDEFS
 */

/* A linked list of miscellaneous memory, used to store arbitrary information
 * reused between strips or pages */
typedef struct MISC_MEM {
    struct MISC_MEM *next;
    UInt32 bytes;
    void *mem;
} MISC_MEM;

typedef struct ioImg {	
    UInt32 handleID;		    /* Handle ID, consistency check */
    struct IO_IMG *prev;		    /* Previous (or last) image of this type
				     * in linked list */
    struct IO_IMG *next;		    /* Next image of this type in linked list
				     */
    IO_PAGE *ioPage;		    /* Page structure of parent */
    UInt32 seqStatus;		    /* Calling sequence status */
    Int32 ifdOffset;		    /* Offset of IFD */
    IO_IMGINFO imgInfo;	    /* Image information */
    UInt32 reservedFieldCount;	    /* Number of fields that can be written
				     * at IFD location */
    UInt32 imgFlags;		    /* As described during IO_Image(Read/
				     * Write)Start */
    UInt32 JPEGquality;		    /* JPEG quality, if JPEG compressing */
    struct IO_IMG *xMaskImg;	    /* Image handle to next (only)
				     * transparency mask image */
    struct IO_IMG *lastSubImg;	    /* Last subimage attached in SubIFDs list
				     */

    /* Field Information */
    TiffFields fields;		    /* TIFF fields */
    Int32 offsetsLocation;	    /* Strip offsets array location */
    Int32 byteCountsLocation;	    /* Strip byte counts array location */
    Int32 interchangeLocation;	    /* JPEG interchange field location */
    Int32 interchangeLengthLocation;/* Interchange length field location */
    Int32 subIFDsTagOffset;	    /* SubIFDs field value offset */
    Int32 annotationOffsetsCountOffset; /* AnnotationOffsets field count offset */
    Int32 annotationOffsetsDataOffset;/* AnnotationOffsets data array offset */
    Int32 maskIFDsTagOffset;	    /* MaskIFDs field value offset */
    Int32 imageLayerLocation;	    /* ImageLayer field value location */
    Int32 ILContentTagOffset;	    /* ImageLayerContent field value offset */
    Int32 ILContentDataOffset;	    /* ImageLayerContent data offset */
    /* Intermediate buffering */
    MISC_MEM *tmpCmpData;	    /* Temporary compression info */
    UInt32 minBytesPerRow;	    /* Physical bytes required per row */
    UInt8 *cmpBuffer;		    /* Intermediate compressed buffer */
    UInt32 cmpBufsize;		    /* Compressed buffer capacity */
    UInt32 cmpBytes;		    /* Number of compressed bytes in the
				     * buffer */
    PIXR *tmpPixr;		    /* Intermediate PIXR buffer (if needed) */
    UInt32 tmpPixrRow;		    /* Image row offset of image in tmpPixr */
    UInt32 tmpPixrRows;		    /* Number of rows currently in tmpPixr */
    UInt32 posSize;		    /* Position block size of Tundra data */
    UInt32 *dictList;		    /* List of token dictionaries used in
				     * by this image */
    UInt32 dictCount;		    /* Number of dictionaries required by
				     * this image */
    Int32 *dictSizes;		    /* Dictionary sizes (passed back to client,
				     * but preserved here for memory mgmnt) */
} ioImg;

#endif

typedef struct IO_IMG {	
    UInt32 handleID;		    /* Handle ID, consistency check */
    UInt32 handleType;		    /* Handle type ID */
    UInt32 ioImgHandle;		    /* Lower level image handle */
    IO_IMGINFO imgInfo;		    /* Image information */
    struct IO_IMG *prev;		    /* Previous (or last) image of this type
				     * in linked list */
    struct IO_IMG *next;	    /* Next image of this type in linked list
				     */
    struct IO_PAGE *ioPage;	    /* Parent page structure */
    struct IO_IMG *ioParentImg;	    /* Parent image (if mask) */
    struct IO_IMG *ioImg;	    /* First in linked list of images */
    struct IO_IMG *ioMaskImg;	    /* Image handle to first (only)
				     * transparency mask image */
    struct IO_IMG *ioHintsImgs;	    /* Image handle to first (only)
				     * rendering hints image */
} IO_IMG;





/*
 * PROTOTYPES
 */

#ifdef __cplusplus
}
#endif

#endif /* IOIMAGE_H_INCLUDED */

