/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

#ifndef IOINFO_H_INCLUDED
#define IOINFO_H_INCLUDED

#ifndef NOIPCORE
IP_RCSINFO(ioinfo_hRCSInfo, "$RCSfile: ioinfo.h,v $; $Revision: 1.1 $");
/* $Date: 2003/01/03 01:41:10 $ */
#endif

#ifdef __cplusplus
extern "C"
{
#endif

/* ioinfo.h
 *
 * DESCRIPTION
 *	 Declarations for IO information
 *
 */

/*
 * INCLUDES
 */
#include "iodoc.h"
#include "iomrc.h"

/*
 * CONSTANTS
 */

/*
 * MACROS
 */

/*
 * TYPEDEFS
 */

/*
 * IO Document information
 *
 * The following structure describes the data returned by the
 * IO_GetDocInfo() function.
 */
typedef struct IO_DOCINFO
{
    DOCTYPE docType;			/* Document format: XIF, T.44, TIFF-FX */
    UInt8 version;			/* Version of IO */
    UInt8 revision;			/* Revision of this version of IO */
    UInt32 options;			/* Document options */
    UInt16 wNPages;			/* Number of pages in file */
    Int8 *pAuthorBlock;			/* Author string (NULL if none) */
    UInt16 wAuthorBlockBytes;		/* Size of author string */
    Int32 dwReferencePageOffset;	/* Reference page offset (-1 if none) */
    IO_DOCHANDLE ioDocHandle;		/* Document handle, used to identify
					 * this document in the document level
					 * routines */
} IO_DOCINFO;


/*
 * IO Page Information
 *
 * The following structure describes the data returned by the
 * IO_GetPageInfo() function.
 */
typedef struct IO_PAGEINFO
{
    Int32 dwPageOffset;			/* Page offset */
    UInt32 *pdwImageCounts;		/* Image count for each layer */
    UInt32 dwLayersPresent;		/* Number of layers present (may
					 * include empty layers) */
    UInt16 wAnnotationCount;		/* Number of annotations on page */
    IO_IMGHANDLE xPrimImgHandle;	/* Primary image handle */
    IO_PAGEHANDLE ioPageHandle;		/* Page handle, used to identify this
					 * page in the page level routines */
} IO_PAGEINFO;


/*
 * IO Image information
 *
 * The following structure describes the data returned by the
 * IO_GetImageInfo() function.
 */
typedef struct IO_IMGINFO
{
    IO_IMGHANDLE ioImgHandle;		/* Image handle to this image */
    IO_IMGTYPE ioImgType;		/* Image type (IO_IMGTYPE_PRIMARY,
					 * IO_IMGTYPE_LAYER or
					 * IO_IMGTYPE_MASK) */
    UInt32 dwImgLayer;			/* Image layer in the MRC model */
    UInt32 dwImgOffset;			/* Image offset within the layer */
    UInt32 dwXOffset;			/* X location in page coordinates */
    UInt32 dwYOffset;			/* Y location in page coordinates */
    UInt32 dwXResolution;		/* X Resolution of image in dpm */
    UInt32 dwYResolution;		/* Y Resolution of image in dpm */
    UInt32 dwWidth;			/* Width in image's X resolution units
					 */
    UInt32 dwHeight;			/* Height in image's Y resolution units
					 */
    UInt32 dwBestStripHeight;		/* For best performance, read strips
					 * with height a multiple of this. */
    IO_COMPRESSION xCompression;	/* Compression type for src/dst */
    IO_COLORTYPE xColorType;		/* Color type (see IO_COLORTYPE) */
    UInt8 pPalette[256*3];		/* Color palette if IO_COLORTYPE is:
					 * IO_COLORTYPE_COLOR4 or
					 * IO_COLORTYPE_COLOR8.
					 * Palette is RGB triples (chunky):
					 * RGBRGB... */
    UInt32 dwMaskImages;		/* Number of mask images */
} IO_IMGINFO;


/*
 * IO Annotation information
 *
 * The following structure describes the data returned by the
 * IO_GetAnnotInfo() function.
 */
typedef struct IO_ANNOTINFO
{
    UInt16 wAnnotType;			/* Type of annotation */
    UInt32 dwXOffset;			/* Offset in X direction in page
					 * coordinates */
    UInt32 dwYOffset;			/* Offset in X direction in page
					 * coordinates */
    UInt32 dwWidth;			/* Width in page coordinates */
    UInt32 dwHeight;			/* Height in page coordinates */
    UInt32 dwFlags;			/* Annotation flags (see io.h) */
    UInt32 dwAnnotLength;		/* Length of annotation data, in bytes
					 */
} IO_ANNOTINFO;


/*
 * PROTOTYPES
 */


#ifdef __cplusplus
}
#endif

#endif /* IOINFO_H_INCLUDED */
