/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


#ifndef IOMRC_H_INCLUDED
#define IOMRC_H_INCLUDED

#ifndef NOIPCORE
IP_RCSINFO(iomrc_h_RCSInfo, "$RCSfile: iomrc.h,v $; $Revision: 1.2 $");
/* $Date: 2003/02/10 22:53:30 $ */
#endif

/* iomrc.h
 *
 * DESCRIPTION
 *	 Declarations for the MRC IO Toolkit API.
 *
 */


#ifdef __cplusplus
extern "C"
{
#endif

/*
 * INCLUDES
 */
#ifdef __GNUC__
#include <stdlib.h>
#endif
#include <stddef.h> /* added by EEP 3/27/01 */
/*
 * CONSTANTS
 */

/* Document types */
typedef enum DOCTYPE {
    DOCTYPE_UNKNOWN = 0,	/*  0. Unknow format */
    DOCTYPE_XIFF2,		/*  1. eXtended Image File Format, version 2.x */
    DOCTYPE_XIFF3,		/*  2. eXtended Image File Format, version 3.x */
    DOCTYPE_T44,		/*  3. T.44 (MRC) format, Annex A */
    DOCTYPE_T44A,		/*  4. T.44 (MRC) format, Annex A */
    DOCTYPE_T44B,		/*  5. T.44 (MRC) format, Annex B */
    DOCTYPE_TIFFFX_S,		/*  6. TIFF-FX Profile S: Standard, Base B&W */
    DOCTYPE_TIFFFX_F,		/*  7. TIFF-FX Profile F: Fax extended B&W */
    DOCTYPE_TIFFFX_J,		/*  8. TIFF-FX Profile J: Binary JBIG1 */
    DOCTYPE_TIFFFX_C,		/*  9. TIFF-FX Profile C: Base Contone */
    DOCTYPE_TIFFFX_L,		/* 10. TIFF-FX Profile L: Color JBIG1 */
    DOCTYPE_TIFFFX_M,		/* 11. TIFF-FX Profile M: Base MRC */
    DOCTYPE_TIFFFX_T,		/* 12. TIFF-FX Profile T: MRC B&W, w/JBIG2 Tokenization */
    DOCTYPE_TIFFFX_M1,		/* 13. TIFF-FX Profile M1: MRC w/JBIG2 & Color Tag */
    DOCTYPE_TIFFFX_O,		/* 14. TIFF-FX Profile M1: MRC w/JBIG2 & Color Tag */
/* Xerox standards: */
    DOCTYPE_COHTIFF_B,		/* 15. Coherent TIFF Profile B: Base B&W */
    DOCTYPE_COHTIFF_C,		/* 16. Coherent TIFF Profile C: Base Contone */
    DOCTYPE_COHTIFF_M,		/* 17. Coherent TIFF Profile M: Base MRC */
    DOCTYPE_COHTIFF_M1,		/* 18. Coherent TIFF Profile M1: Base MRC + JBIG2 */
    DOCTYPES
} DOCTYPE;

/* Annotation flags */
#define IO_ANNOT_NOTVIEWABLE	0x00000001	/* Not viewable */
#define IO_ANNOT_HYPERLINK	0x00000002	/* Annotation also hyperlinks
						 * to either another spot in
						 * the document (a page), or
						 * an URL. */

/*
 * MACROS
 */
/* Bits per pixel */
#define IO_BPP(coltype)	((coltype) == IO_COLORTYPE_BINARY? 1:	\
				 (coltype) == IO_COLORTYPE_COLOR24? 24:\
				 ((coltype) == IO_COLORTYPE_GRAY8 ||	\
				  (coltype) == IO_COLORTYPE_COLOR8)? 8:\
				 ((coltype) == IO_COLORTYPE_GRAY4 ||	\
				  (coltype) == IO_COLORTYPE_COLOR4)? 4:\
				  (coltype) == IO_COLORTYPE_COLOR12? 12: \
				  (coltype) == IO_COLORTYPE_CMYK8? 32: \
				  (coltype) == IO_COLORTYPE_CMYK4? 16: \
				  (coltype) == IO_COLORTYPE_CMYK1? 4: 0)

/* Bits per pixel element */
#define IO_BPPE(coltype)	((coltype) == IO_COLORTYPE_BINARY? 1:	\
				 (coltype) == IO_COLORTYPE_COLOR24? 8:	\
				 ((coltype) == IO_COLORTYPE_GRAY8 ||	\
				  (coltype) == IO_COLORTYPE_COLOR8)? 8:\
				 ((coltype) == IO_COLORTYPE_GRAY4 ||	\
				  (coltype) == IO_COLORTYPE_COLOR12 ||	\
				  (coltype) == IO_COLORTYPE_COLOR4)? 4: \
				  (coltype) == IO_COLORTYPE_CMYK8? 8: \
				  (coltype) == IO_COLORTYPE_CMYK4? 4: \
				  (coltype) == IO_COLORTYPE_CMYK1? 1: 0)

/* Channels per pixel */
#define IO_CHANNELS(coltype)	((coltype) == IO_COLORTYPE_BINARY? 1:	\
				 (coltype) == IO_COLORTYPE_COLOR24 ||	\
				 (coltype) == IO_COLORTYPE_COLOR12? 3:	\
				 ((coltype) == IO_COLORTYPE_GRAY8 ||	\
				  (coltype) == IO_COLORTYPE_COLOR8)? 1:\
				 ((coltype) == IO_COLORTYPE_GRAY4 ||	\
				  (coltype) == IO_COLORTYPE_COLOR4)? 1: \
				 ((coltype) == IO_COLORTYPE_CMYK8 ||	\
				  (coltype) == IO_COLORTYPE_CMYK4 ||	\
				  (coltype) == IO_COLORTYPE_CMYK1)? 1: 0)

/*
 * TYPEDEFS
 */
#if !defined(_TYPES_PUB_INCLUDED) || !defined(CDECL)
typedef unsigned char      	UInt8;
typedef unsigned short int 	UInt16;
typedef unsigned long int  	UInt32;
typedef signed char		Int8;
typedef short int		Int16;
typedef long int		Int32;
typedef unsigned long int	Bool;
typedef unsigned long int	ErrCode;
typedef float			Float32;
typedef double			Float64;
#define CDECL
#define PROLOGUE		ErrCode __error__ = IO_NOERROR
#define ERR_RETURN		return __error__
#define NORMAL_RETURN		return __error__
typedef	void (CDECL *ClientErrorProc) (
    void       *pClientError,
    char       *pErrorString);
#endif


/*
 * IO basic types
 *
 * For portability, the IO library uses platform-independent type names.
 * The definitions of the type will be in a single platform-dependent
 * section here.
 */

/*
 * IO derived types
 */

/* Handle to client's instance of library */
typedef UInt32 IO_INSTHANDLE;

/* Handle to open document */
typedef UInt32 IO_DOCHANDLE;

/* Handle to an active page */
typedef UInt32 IO_PAGEHANDLE;

/* Handle to an active image */
typedef UInt32 IO_IMGHANDLE;

/* Handle ID type */
typedef UInt32 IO_HANDLETYPE;


/* Return codes (seemingly out of sequence since they were borrowed from
 * a compression/decompression package which this library sits on top of,
 * namely IPCore). */
#define IO_NOERROR		0	/* Normal return */
#define IO_NOMEMORY		1	/* Ran out of memory */
#define IO_BADPARAMETER		2	/* Invalid parameter to IO_function */
#define IO_NOSUPPORT		6	/* Unsupported feature or condition */
#define IO_BUFFEREXCEEDED	7	/* Data buffer filled and will
					 * overflow */
#define IO_INTERNAL_ERROR	12	/* Unexpected internal state (shouldn't
					 * happen) */
#define IO_CLIENTABORT		14	/* Client requested abort */
#define IO_TIFFEMPTY		18	/* 0-page TIFF (okay for XIFF; TIFF-FX?) */
#define IO_NOTTIFF		19	/* File does not comform to TIFF */
#define IO_NOTIO		20	/* File does not comform to IO
					 * supported formats*/
#define IO_SEQUENCE		21	/* Function called out-of-sequence */
#define IO_BADFILEMODE		22	/* File not open in BINARY/READ/WRITE
					 * mode */
#define IO_BADFILEFORMAT	23	/* File format/data is corrupted */
#define IO_NOTT44		24	/* File does not comform to T.44 stream */
#define IO_READ_ERR		2000	/* Error reading */
#define IO_WRITE_ERR		2001	/* Error writing */
#define IO_SEEK_ERR		2002	/* Error seeking */
#define IO_FILELEN_ERR		2003	/* Error seeking to EOF */
#define IO_TELL_ERR		2004	/* Error telling position */

/* Ignorable errors, will always have values higher than IO_WARN_NOSUPPORT */
#define IO_WARNINGS		10000	/* First warning value, if less than
					 * this, it's a real error */
#define IO_WARN_NOSUPPORT  	10000	/* Not supported by this library,
					 * but may not be necessary */
#define IO_WARN_NOTCOMPLIANT	10001	/* Not compliant with the standards */
#define IO_WARN_BADFILEFORMAT	10002	/* Recoverable file format error */


/* Types of image */
#define IO_COLORTYPE_NONE	0	/* None specified */
#define IO_COLORTYPE_BINARY	1	/* Binary image */
#define IO_COLORTYPE_GRAY4	2	/* 4 bpp grayscale (values: 0-15) */
#define IO_COLORTYPE_GRAY8	3	/* 8 bpp grayscale (values: 0-255) */
#define IO_COLORTYPE_COLOR4	4	/* 4 bpp color (paletted indices: 0-15)
					 */
#define IO_COLORTYPE_COLOR8	5	/* 8 bpp color (paletted indices:
					 * 0-255) */
#define IO_COLORTYPE_COLOR12	6	/* 4 bpp color (3 samples x 4 bpp) */
#define IO_COLORTYPE_COLOR24	7	/* 24 bpp color (8 bpp for each of RGB)
					 */
#define IO_COLORTYPE_CMYK1	8	/* 1 bpp CMYK color */
#define IO_COLORTYPE_CMYK4	9	/* 4 bpp CMYK color */
#define IO_COLORTYPE_CMYK8	10	/* 8 bpp CMYK color */
#define IO_COLORTYPE_OTHER	11	/* Other type (12 bpp, or other color space) */

typedef unsigned long int	IO_COLORTYPE;


/* Stored color space */
#define IO_COLORSPACE_DEFAULT		0	/* Use the default */
#define IO_COLORSPACE_WHITEISZERO	1	/* 0 = white */
#define IO_COLORSPACE_BLACKISZERO	2	/* 1 = white */
#define IO_COLORSPACE_RGB		3	/* RGB */
#define IO_COLORSPACE_CMYK		4	/* CMYK */
#define IO_COLORSPACE_YCBCR		5	/* YCbCr */
#define IO_COLORSPACE_CIELAB		6	/* CIELAB */
#define IO_COLORSPACE_ICCLAB		7	/* ICCLAB */
#define IO_COLORSPACE_ITULAB		8	/* ITULAB (Gammut constrained L*a*b*) */

typedef unsigned long int	IO_COLORSPACE;


/*
 * The type of the image with respect to the image/subimage
 * hierarchy. _PRIMARY indicates text plane/page image, which
 * is binary, unless no other images are attached to this page.
 * These image types can be divided into two classes, those
 * with image content and those that modify image content.
 * For example, an IO_IMGTYPE_MASK depends on and modifes
 * the image content of a non-binary parent image.  An 
 * IO_IMGTYPE_MASK would never be displayed independently of
 * its parent.
 */
#define IO_IMGTYPE_UNKNOWN	0	/* None */
#define IO_IMGTYPE_PRIMARY	1	/* Binary text - used as a FG/BG
					 * selector, Image layer 2 */
#define IO_IMGTYPE_LAYER	2	/* Belonging to an image layer
					 * other than 2 */
#define IO_IMGTYPE_MASK		3	/* Transparency mask (XIFF only) */
#define IO_IMGTYPE_HINTS	4	/* Rendering Hints (TIFF-FX/XCTIFF only) */

typedef unsigned long int	IO_IMGTYPE;


/*
 * IO general I/O functions
 *
 * The IO client is responsible for managing the interface to the
 * local file system and must provide functions for basic file
 * I/O. To the IO library, a file is represented by set of I/O
 * functions, and a file ID. This file ID (specified by the client),
 * is sent as a parameter to each of the I/O functions when the library
 * moves about the file, or accesses the data.
 *
 * As an example, let's say I have an implementation of the functions
 * myRead(), myWrite(), mySeek(), myFilesize(), and myTell(). I setup a
 * file interface after a IO instance is established:
 *	fd = open(filename, _O_READ | _O_BINARY);
 *	IO_OpenDocumentRead(xInst, (Int32)fd, options, myRead, myWrite,
 *		mySeek, mySize, myTell, &docHandle);
 * Now the docHandle is associated with the file via the client ID (file
 * descriptor), and I/O functions.
 *
 * The following is the format of the functions:
 */
typedef Int32 (CDECL *IO_READFUNC)(UInt32 dwClientFileID,
				    Int8 *pBuf,
				    Int32 dwByteCount);
typedef Int32 (CDECL *IO_WRITEFUNC)(UInt32 dwClientFileID,
			       const Int8 *pBuf,
				     Int32 dwByteCount);
typedef Int32 (CDECL *IO_SEEKFUNC)(UInt32 dwClientFileID,
				    Int32 dwOffset);
typedef Int32 (CDECL *IO_SIZEFUNC)(UInt32 dwClientFileID);
typedef Int32 (CDECL *IO_TELLFUNC)(UInt32 dwClientFileID);


/*
 * Compression methods
 */
#define IO_COMP_UNKNOWN		0x0000	/* No valid compression */
#define IO_COMP_NOC		0x0001	/* No compression, leave as raw data */
#define IO_COMP_MH		0x0002	/* Modified Huffman (ITU-T T.?) */
#define IO_COMP_MR		0x0004	/* Modified Read (ITU-T T.?) */
#define IO_COMP_MMR		0x0008	/* Modified modified-read (CCITT
					 * group-4, ITU-T T.6) */
#define IO_COMP_JPEG		0x0010	/* JPEG */
#define IO_COMP_WAVELET		0x0020	/* Xerox Wavelet */
#define IO_COMP_TOKENIZATION	0x0040	/* Xerox Tokenization */
#define IO_COMP_LLTOKENIZATION	0x0080	/* Xerox Lossless Tokenization */
#define IO_COMP_LOSSYDITHER	0x0100	/* Xerox Lossy Dither */
#define IO_COMP_JBIG		0x0200	/* JBIG (aka JBIG1) */
#define IO_COMP_JBIG2		0x0400	/* JBIG2 (Symbol-based compression) */
#define IO_COMP_LLJBIG2		0x0800	/* Lossless JBIG2 */
#define IO_COMP_TRCE		0x1000	/* Color run-length (color JBIG2) */
#define IO_COMP_JPEG2000	0x2000	/* JPEG2000 (T.44 only) */

#define IO_COMP_OTHER		0x8000	/* Other compression type
					 * (use IO_SetCompression() or
					 * IO_GetCompression()) */

typedef unsigned long int	IO_COMPRESSION;


/*
 ******************************************************************************
 *				FUNCTION PROTOTYPES			      *
 ******************************************************************************
 */

/*
 * IO progress function. The progress function is selected by the client.
 * It is called periodically by IO functions to report progress and offer
 * the client an opportunity to cancel. The meaning of the progress function
 * depends on the function being carried out. If not otherwise documented,
 * the progress value will be -1, indicating no meaningful progress to report, 
 * or a number from 0 to 100, indicating percent complete.
 */
typedef Int32
(*IO_PROGRESSFUNC)(
	UInt32 dwClientID,
	Int32 dwProgress);

ErrCode CDECL
IO_SetProgressFunc(
	IO_INSTHANDLE ioInstHandle,
	IO_PROGRESSFUNC xProgress);


/*
 * IO memory allocation function. These function allows the client to
 * override IO's memory allocation functions. It is not required, and
 * may not be available on all platforms.
 *
 * Once the functions have been overridden, they can be restored to
 * their defaults by passing NULL for each function.
 *
 * 'hasPtrs' is non-zero when allocating an array of pointers. On some
 * systems, this indicates that the memory management must update the
 * contents of this array, should the memory pointed to by this array
 * move. It is a garbage collection feature available to the client.
 * >>> However, none of the calls made by any of the functions make use
 * >>> of this (hasPtrs always equals 0).
 */
typedef void * (CDECL *IO_MALLOCFUNC)(size_t dwBytes,
				       Bool hasPtrs,
				       void *pClientMallocID
#ifdef SHOWERR
				       , const char *file,
				       const int line
#endif
				       );
typedef void * (CDECL *IPCORE_MALLOCFUNC)(size_t dwBytes,
				       Bool hasPtrs,
				       void *pClientMallocID);
typedef void (CDECL *IO_FREEFUNC)(void *pBuf,
				   size_t dwBytes,		/* Ignored */
				   void *pClientMallocID);
typedef void * (CDECL *IO_CALLOCFUNC)(size_t nel,
				       size_t size,
				       Bool hasPtrs,
				       void *pClientMallocID
#ifdef _DEBUG
				       , const char *file,
				       const int line
#endif
				       );
typedef void * (CDECL *IPCORE_CALLOCFUNC)(size_t nel,
				       size_t size,
				       Bool hasPtrs,
				       void *pClientMallocID
				       );
typedef void * (CDECL *IO_REALLOCFUNC)(void *ptr,
					size_t dwBytes,
					Bool hasPtrs,
					void *pClientMallocID
#ifdef SHOWERR
				       , const char *file,
				       const int line
#endif
				       );
typedef void * (CDECL *IPCORE_REALLOCFUNC)(void *ptr,
					size_t dwBytes,
					Bool hasPtrs,
					void *pClientMallocID
				       );


/*
 * IO_InitInstance()
 */
ErrCode CDECL
IO_InitInstance(
	UInt32 dwClientInstID,
	IO_INSTHANDLE *pioInst
);


/*
 * IO_GetInstance()
 */
ErrCode CDECL
IO_GetInstance(
    IO_DOCHANDLE ioDocHandle,
    IO_INSTHANDLE *pioInstHandle
);


/*
 * IO_EndInstance()
 */
ErrCode CDECL
IO_EndInstance(
	IO_INSTHANDLE ioInstHandle
);


/*
 * IO_SetMallocFuncs()
 */
ErrCode CDECL
IO_SetMallocFuncs(
	IO_INSTHANDLE ioInstHandle,
	void *dwClientMallocID,
	IO_MALLOCFUNC ioMalloc,
	IO_FREEFUNC ioFree
);


/*
 * IO_Malloc()
 */
ErrCode CDECL
IO_Malloc(
    IO_INSTHANDLE xInst,
    UInt32 dwByteCount,
    void **blockPtr
#ifdef SHOWERR
    , const char *file,
    const int line
#endif
);


/*
 * IO_Free()
 */
ErrCode CDECL
IO_Free(
    IO_INSTHANDLE xInst,
    void **blockPtr
);


/*
 * IO_GetClientID()
 */
ErrCode CDECL
IO_GetClientID(
	IO_INSTHANDLE ioInstHandle,
	UInt32 *dwClientInstID
);


/*
 * IO_SetErrorProc()
 */
ErrCode IO_SetErrorProc(
	IO_INSTHANDLE ioInstHandle,
	ClientErrorProc errProc,
	void *errInfo
);


/*
 * IO_IsIo()
 */
ErrCode CDECL
IO_IsIo(
	IO_INSTHANDLE ioInstHandle,
	Int32 dwClientID,
	IO_READFUNC xReadFunc,
	IO_SEEKFUNC xSeekFunc,
	IO_SIZEFUNC xSizeFunc,
	DOCTYPE *pioDocType,
	UInt8 *pVersion,
	UInt8 *pRevision,
	UInt32 *pdwOptions
);


/*
 * IO_OpenDocumentRead()
 */
ErrCode CDECL
IO_OpenDocumentRead(
	IO_INSTHANDLE ioInstHandle,
	Int32 dwClientID,
	UInt32 options,
	IO_READFUNC xReadFunc,
	IO_WRITEFUNC xWriteFunc,
	IO_SEEKFUNC xSeekFunc,
	IO_SIZEFUNC xSizeFunc,
	IO_TELLFUNC xTellFunc,
	IO_DOCHANDLE *pioDocHandle
);


/*
 * IO_CloseDocument()
 */
ErrCode CDECL
IO_CloseDocument(
    IO_DOCHANDLE ioDocHandle
);


/*
 * IO_GetDocInfo()
 */
ErrCode CDECL
IO_GetDocInfo(
	IO_DOCHANDLE ioDocHandle,	/* Input document handle */
/* Returned values */
	UInt8 *pVersion,		/* Version of IO */
	UInt8 *pRevision,		/* Revision of this version of IO */
	UInt16 *pwNPages		/* Number of pages in file */
);


/*
 * IO_GetPageInfo()
 */
ErrCode CDECL
IO_GetPageInfo(
	IO_DOCHANDLE ioDocHandle,	/* Input document handle */
	Int32 dwPageOffset,		/* Page offset (starting with 0) */
/* Returned values */
	UInt32 **ppdwImageCounts,	/* Number of images within each layer
					 */
	UInt32 *pdwLayersPresent,	/* Number of layers present (although
					 * some layers may be missing) */
	UInt16 *pwAnnotationCount,	/* Number of annotations on page */
	IO_PAGEHANDLE *pioPageHandle	/* Page handle, used to identify this
					 * page in the page level routines */
);


/*
 * IO_GetImageInfo()
 */
ErrCode CDECL
IO_GetImageInfo(
	IO_PAGEHANDLE ioPageHandle,	/* Page image handle */
	UInt32 dwImgLayer,		/* Image layer in the IO */
	UInt32 dwImgOffset,		/* Image offset (for multiple images)
					 */
/* Returned values */
	UInt32 *pdwXOffset,		/* X location in page coordinates */
	UInt32 *pdwYOffset,		/* Y location in page coordinates */
	UInt32 *pdwXResolution,		/* X Resolution of image in dpi */
	UInt32 *pdwYResolution,		/* Y Resolution of image in dpi */
	UInt32 *pdwWidth,		/* Width in image's X resolution units
					 */
	UInt32 *pdwHeight,		/* Height in image's Y resolution units
					 */
	UInt32 *pdwBestStripHeight,	/* For best performance, read strips
					 * with height a multiple of this. */
	IO_COMPRESSION *pxCompression,	/* Compression type for src/dst */
	IO_COLORTYPE *pxColorType,	/* Color type (see IO_COLORTYPE) */
	UInt8 *pPalette,		/* Color palette if IO_COLORTYPE is:
					 * IO_COLORTYPE_COLOR4 or
					 * IO_COLORTYPE_COLOR8.
					 * Palette is RGB triples (chunky):
					 * RGBRGB... */
	UInt32 *pdwMaskImages,		/* Number of mask images */
	IO_IMGHANDLE *pioImgHandle	/* Returned image handle */
);


/*
 * IO_GetImageStats()
 */
ErrCode CDECL
IO_GetImageStats(
	IO_IMGHANDLE ioImgHandle,	/* Image handle */
/* Returned values */
	UInt32 *pdwCmpSize,		/* Compressed image byte total */
	UInt32 *pdwDictCount,		/* Number of token dictionaries */
	UInt32 **ppdwDictIDs,		/* Dictionary ID list (if tokenized) */
	Int32 **ppdwDictSizes		/* Dictionary byte list (tokenized) */
);


/*
 * IO_GetMaskImageInfo()
 */
ErrCode CDECL
IO_GetMaskImageInfo(
	IO_IMGHANDLE ioParentImgHandle,	/* Parent image handle */
	UInt32 dwImgOffset,		/* Image offset (for multiple images)
					 */
/* Returned values */
	UInt32 *pdwXOffset,		/* X location in page coordinates */
	UInt32 *pdwYOffset,		/* Y location in page coordinates */
	UInt32 *pdwXResolution,		/* X Resolution of image in dpi */
	UInt32 *pdwYResolution,		/* Y Resolution of image in dpi */
	UInt32 *pdwWidth,		/* Width in image's X resolution units
					 */
	UInt32 *pdwHeight,		/* Height in image's Y resolution units
					 */
	UInt32 *pdwBestStripHeight,	/* For best performance, read strips
					 * with height a multiple of this. */
	IO_COMPRESSION *pxCompression,	/* Compression type for src/dst */
	IO_IMGHANDLE *pioImgHandle	/* Returned image handle */
);


/*
 * IO_GetHintsImageInfo()
 */
ErrCode CDECL
IO_GetHintsImageInfo(
	IO_IMGHANDLE ioParentImgHandle,	/* Parent image handle */
	UInt32 dwImgOffset,		/* Image offset (for multiple images)
					 */
/* Returned values */
	UInt32 *pdwXOffset,		/* X location in page coordinates */
	UInt32 *pdwYOffset,		/* Y location in page coordinates */
	UInt32 *pdwXResolution,		/* X Resolution of image in dpi */
	UInt32 *pdwYResolution,		/* Y Resolution of image in dpi */
	UInt32 *pdwWidth,		/* Width in image's X resolution units
					 */
	UInt32 *pdwHeight,		/* Height in image's Y resolution units
					 */
	UInt32 *pdwBestStripHeight,	/* For best performance, read strips
					 * with height a multiple of this. */
	IO_COMPRESSION *pxCompression,	/* Compression type for src/dst */
	IO_IMGHANDLE *pioImgHandle	/* Returned image handle */
);


/*
 * IO_GetDocProfile()
 */
ErrCode CDECL
IO_GetDocProfile(
	IO_DOCHANDLE ioDocHandle,
	DOCTYPE *pdwProfile
);


/*
 * Image data representation features, some not implemented (NIMP):
 */
#define IO_IMGFLAGS_RGBOrder		(0x00)
#define IO_IMGFLAGS_BGROrder		(0x01)		/* NIMP */
#define IO_IMGFLAGS_TopToBottom		(0x00)
#define IO_IMGFLAGS_BottomToTop		(0x02)		/* NIMP */
#define IO_IMGFLAGS_WhiteIsZero		(0x00)
#define IO_IMGFLAGS_BlackIsZero		(0x04)		/* IMP for TFX only */
#define IO_IMGFLAGS_ColorPlanar		(0x00)
#define IO_IMGFLAGS_ColorChunky		(0x08)
#define IO_IMGFLAGS_ByteOrder		(0x00)		/* IMP for TFX only */
#define IO_IMGFLAGS_DwordOrder		(0x10)
#define IO_IMGFLAGS_Uncompressed	(0x00)
#define IO_IMGFLAGS_Compressed		(0x20)		/* NIMP */
#define IO_IMGFLAGS_SubSampChroma	(0x00)
#define IO_IMGFLAGS_NoSubSampChroma	(0x100)		/* IMP for TFX only */
#define IO_IMGFLAGS_JPEGTables		(0x200)		/* JPEGTables optimization,
							 * TFX only */
#define IO_IMGFLAGS_IPCoreJPEG		(0x80000000)	/* IPCore JPEG
							 * (if IJL default) */

/*
 * Image data representation feature mask:
 */
#define IO_IMGFLAGS_ColorOrder		(0x01)
#define IO_IMGFLAGS_ScanlineOrder	(0x02)
#define IO_IMGFLAGS_PhotoInterp		(0x04)
#define IO_IMGFLAGS_ColorFormat		(0x08)
#define IO_IMGFLAGS_ByteOrdering	(0x10)
#define IO_IMGFLAGS_Compression		(0x20)
#define IO_IMGFLAGS_SubSampling		(0x100)

/*
 * IO_ImageReadStart()
 */
ErrCode CDECL
IO_ImageReadStart(
	IO_IMGHANDLE ioImgHandle,
	UInt32 dwFlags
);


/*
 * IO_GetCompressedStripSize()
 */
ErrCode CDECL
IO_GetCompressedStripSize(
	IO_IMGHANDLE ioImgHandle,
	UInt32 dwImageRowOffset,
	UInt32 *pdwBytes
);


/*
 * IO_CompressedReadStrip()
 */
ErrCode CDECL
IO_CompressedReadStrip(
	IO_IMGHANDLE ioImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 **ppCmpBuffer,
	UInt32 *pdwCmpBufsize,
	UInt32 *pdwBytesFilled
);


#ifdef NOIPCORE
#define IO_ImageReadStrip(ioImgHandle, dwImgRowOffset, dwStripHeight, \
 ppImagePlanes, dwBytesPerRow, dwBufferRowOffset, pdwRowsFilled) \
	IO_NOSUPPORT
#else
/*
 * IO_ImageReadStrip()
 */
ErrCode CDECL
IO_ImageReadStrip(
	IO_IMGHANDLE ioImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 **ppImagePlanes,
	UInt32 dwBytesPerRow,
	UInt32 dwBufferRowOffset,
	UInt32 *pdwRowsFilled
);
#endif


/*
 * IO_GetDictListFromPosBlock
 */
ErrCode CDECL
IO_GetDictListFromPosBlock(
	IO_IMGHANDLE ioImgHandle,
	UInt8 *pCmpData,
	UInt32 dwCmpBytes,
	UInt16 *pdwDictCount,
	UInt32 **ppdwDictIDs,
	UInt8 **ppPositionBlock,
	UInt32 *pdwPositionBytes
);


/*
 * IO_GetStripeInfo
 */
ErrCode CDECL
IO_GetStripeInfo(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 dwRow,
	UInt32 *pdwStripeOffset,
	UInt32 *pdwStripeRow,
	UInt32 *pdwStripeHeight
);

	
/*
 * IO_GetSymbolDict
 */
ErrCode CDECL
IO_GetSymbolDict(
	IO_DOCHANDLE ioDocHandle,
	UInt32 dictID,
	UInt8 **pDictBlock,
	UInt32 *pdwDictBytes,
	UInt32 *pdwDictBlockBytes,
	UInt16 *pwLastPageUsed,
	UInt32 *pdwDecoderMemory
);


/*
 * IO_ImageReadFinish()
 */
ErrCode CDECL
IO_ImageReadFinish(
	IO_IMGHANDLE ioImgHandle
);


/*
 * IO_GetLayerInfo()
 */
ErrCode CDECL
IO_GetLayerInfo(
	IO_PAGEHANDLE ioPageHandle,
	UInt32 dwImgLayer,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	IO_COMPRESSION *pxCompression,	/* Compression type for src/dst */
	IO_COLORTYPE *pxColorType,	/* Color type (see IO_COLORTYPE) */
	UInt8 *pPalette,		/* Color palette if IO_COLORTYPE is:
					 * IO_COLORTYPE_COLOR4 or
					 * IO_COLORTYPE_COLOR8.
					 * Palette is RGB triples (chunky):
					 * RGBRGB... */
	IO_IMGHANDLE *pioImgHandle
);


#ifdef NOIPCORE
#define IO_ImageReadLayer(ioImgHandle, dwImageHeight, pImageBuffer, \
 dwBytesPerRow, pdwRowsFilled)		IO_NOSUPPORT
#else
/*
 * IO_ImageReadLayer()
 */
ErrCode CDECL
IO_ImageReadLayer(
	IO_IMGHANDLE ioImgHandle,
	UInt32 dwImageHeight,
	UInt8 **pImageBuffer,
	UInt32 dwBytesPerRow,
	UInt32 *pdwRowsFilled
);
#endif


/*
 * IO_GetAnnotationInfo()
 */
ErrCode CDECL
IO_GetAnnotationInfo(
	IO_PAGEHANDLE ioPageHandle,
	UInt16 wAnnotOffset,		    /* Offset of annotation for this
					     * page */
	UInt16 *pwAnnotType,		    /* Type of annotation */
	UInt32 *pdwXOffset,		    /* Offset in X direction in page
					     * coordinates */
	UInt32 *pdwYOffset,		    /* Offset in X direction in page
					     * coordinates */
	UInt32 *pdwWidth,		    /* Width in page coordinates */
	UInt32 *pdwHeight,		    /* Height in page coordinates */
	UInt32 *pdwFlags,		    /* Annotation flags (see ANNOT_...
					     * above */
	UInt32 *pdwAnnotLength		    /* Length of annotation data, in
					     * bytes */
);


/*
 * IO_GetAnnotationData()
 */
ErrCode CDECL
IO_GetAnnotationData(
	IO_PAGEHANDLE ioPageHandle,
	UInt16 wAnnotOffset,
	UInt32 dwBytes,
	UInt8 *pData
);


/*
 * IO_GetAnnotationOffsets
 */
ErrCode CDECL
IO_GetAnnotationOffsets(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 *pdwAnnotationOffsets,
	Bool *pbExists
);

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 
/*
 * IO_Author()
 */
ErrCode CDECL
IO_GetAuthor(
    IO_DOCHANDLE ioDocHandle,
	const Int8 **ppAuthor,
	UInt16 *pwAuthorBytes,
	Bool *pbExists
);


#if 0
/* This is not exposed */
/*
 * IO_GetBitsPerSample()
 */
ErrCode CDECL
IO_GetBitsPerSample(
    IO_IMGHANDLE ioImgHandle,
	UInt16 **ppwBitsPerSample,
	UInt16 *pwBitsPerSample,
	Bool *pbExists
);
#endif


/*
 * IO_GetChromaPositioning()
 */
ErrCode CDECL
IO_GetChromaPositioning(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwPositioning,
	Bool *pbExists
);


/*
 * IO_GetChromaSubSampling()
 */
ErrCode CDECL
IO_GetChromaSubSampling(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwChromaSubSamplingHor,
	UInt16 *pwChromaSubSamplingVer,
	Bool *pbExists
);


/*
 * IO_GetCodingMethods()
 */
ErrCode CDECL
IO_GetCodingMethods(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 *pdwCodingMethods,
	Bool *pbExists
);

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 
/*
 * IO_GetCompression()
 */
ErrCode CDECL
IO_GetCompression(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 *pwCompression,
	Bool *pbExists
);

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 
/*
 * IO_GetDateTime()
 */
ErrCode CDECL
IO_GetDateTime(
    IO_PAGEHANDLE ioPageHandle,
	Int8 **ppDateTime,
	Bool *pbExists
);


/*
 * IO_GetDecode()
 */
ErrCode CDECL
IO_GetDecode(
    IO_IMGHANDLE ioImgHandle,
	Int32 **ppwDecode,
	UInt32 *pdwDecodeValues,
	Bool *pbExists
);


/*
 * IO_GetDocumentName()
 */
ErrCode CDECL
IO_GetDocumentName(
    IO_DOCHANDLE ioDocHandle,
	Int8 **ppDocumentName,
	UInt32 *pdwDocumentNameBytes,
	Bool *pbExists
);


/*
 * IO_GetFaxProfile()
 */
ErrCode CDECL
IO_GetFaxProfile(
    IO_PAGEHANDLE ioPageHandle,
	UInt8 *pFaxProfile,
	Bool *pbExists
);


/*
 * TFX_GetFillOrder()
 */
ErrCode CDECL
IO_GetFillOrder(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwFillOrder,
	Bool *pbExists
);


/*
 * IO_GetHostComputer()
 */
ErrCode CDECL
IO_GetHostComputer(
    IO_IMGHANDLE ioImgHandle,
	Int8 **ppHostComputer,
	UInt32 *pdwHostComputer,
	Bool *pbExists
);


/*
 * IO_GetImageBaseColor()
 */
ErrCode CDECL
IO_GetIFDBaseColor(
    IO_IMGHANDLE ioImgHandle,
	Int32 **ppdwIFDBaseColor,
	UInt32 *pdwIFDBaseColorValues,
	Bool *pbExists
);


/*
 * IO_GetImageBaseColor()
 */
ErrCode CDECL
IO_GetImageBaseColor(
    IO_IMGHANDLE ioImgHandle,
	UInt16 **ppwImageBaseColor,
	UInt32 *pdwImageBaseColorValues,
	Bool *pbExists
);


/*
 * IO_GetImageDescription()
 */
ErrCode CDECL
IO_GetImageDescription(
    IO_IMGHANDLE ioImgHandle,
	Int8 **ppImageDescription,
	UInt32 *pdwImageDescriptionBytes,
	Bool *pbExists
);


/*
 * IO_GetT88Options()
 */
ErrCode CDECL
IO_GetT88Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwT88Options,
	Bool *pbExists
);


/*
 * IO_GetModeNumber()
 */
ErrCode CDECL
IO_GetModeNumber(
    IO_PAGEHANDLE ioPageHandle,
	UInt8 *pModeNumber,
	Bool *pbExists
);


/*
 * IO_GetOrientation()
 */
ErrCode CDECL
IO_GetOrientation(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwOrientation,
	Bool *pbExists
);


/*
 * IO_GetPageNumber()
 */
ErrCode CDECL
IO_GetPageNumber(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 *pwPageNumber,
	UInt16 *pwPageCount,
	Bool *pbExists
);


/*
 * IO_GetPhotometricInterpretation()
 */
ErrCode CDECL
IO_GetPhotometricInterpretation(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwPhotometricInterpretation,
	Bool *pbExists
);


/*
 * IO_GetPlanarConfiguration()
 */
ErrCode CDECL
IO_GetPlanarConfiguration(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwPlanarConfiguration,
	Bool *pbExists
);


/*
 * IO_GetProfileType()
 */
ErrCode CDECL
IO_GetProfileType(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 *pdwProfileType,
	Bool *pbExists
);


/*
 * IO_GetReferencePage()
 */
ErrCode CDECL
IO_GetReferencePage(
    IO_DOCHANDLE ioDocHandle,
	Int32 *pdwReferencePage,
	Bool *pbExists
);


/*
 * IO_GetRenderingHints()
 */
ErrCode CDECL
IO_GetRenderingHints(
    IO_IMGHANDLE ioImgHandle,
	Int8 **ppRenderingHints,
	UInt32 *pdwRenderingHintsBytes,
	Bool *pbExists
);


/*
 * IO_GetRenderingHintsImageCount()
 */
ErrCode CDECL
IO_GetRenderingHintsImageCount(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwCount
);


/*
 * IO_GetResolutionUnit()
 */
ErrCode CDECL
IO_GetResolutionUnit(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwResolutionUnit,
	Bool *pbExists
);


/*
 * IO_GetRowsPerStrip()
 */
ErrCode CDECL
IO_GetRowsPerStrip(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pwRowsPerStrip,
	Bool *pbExists
);


/*
 * IO_GetSamplesPerPixel()
 */
ErrCode CDECL
IO_GetSamplesPerPixel(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwSamplesPerPixel,
	Bool *pbExists
);


/*
 * IO_GetSoftware()
 */
ErrCode CDECL
IO_GetSoftware(
    IO_IMGHANDLE ioImgHandle,
	Int8 **ppSoftware,
	UInt32 *pdwSoftwareBytes,
	Bool *pbExists
);


/*
 * IO_GetTargetPrinter()
 */
ErrCode CDECL
IO_GetTargetPrinter(
    IO_PAGEHANDLE ioPageHandle,
	Int8 **ppTargetPrinter,
	UInt32 *pdwTargetPrinterBytes,
	Bool *pbExists
);


/*
 * IO_GetThresholding()
 */
ErrCode CDECL
IO_GetThresholding(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwThresholding,
	Bool *pbExists
);


/*
 * IO_GetT4Options()
 */
ErrCode CDECL
IO_GetT4Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwT4Options,
	Bool *pbExists
);


/*
 * IO_GetT6Options()
 */
ErrCode CDECL
IO_GetT6Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwT6Options,
	Bool *pbExists
);


/*
 * IO_GetVersionYear()
 */
ErrCode CDECL
IO_GetVersionYear(
    IO_IMGHANDLE ioImgHandle,
	UInt8 **ppVersionYear,
	UInt32 *pdwVersionYearBytes,
	Bool *pbExists
);


/*
 * IO_GetYcrCbCoefficients()
 */
ErrCode CDECL
IO_GetYCbCrCoefficients(
    IO_IMGHANDLE ioImgHandle,
	UInt32 **ppdwYcrCbCoefficients,
	UInt32 *pdwYCbCrCoefficientValues,
	Bool *pbExists
);


/*
 * IO_GetPredictor()
 */
ErrCode CDECL
IO_GetPredictor(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwPredictor,
	Bool *pbExists
);


/*
 * **********						********** 
 * **********	BEGIN WRITE API		********** 
 * **********						********** 
 */

/*
 * IO_CreateDocument()
 */
ErrCode CDECL
IO_CreateDocument(
	IO_INSTHANDLE ioInstHandle,
	DOCTYPE ioDocType,
	Int32 dwClientID,		/* Input client ID */
	UInt8 version,			/* Version of IO format to create */
	UInt32 options,			/* Format specific options */
	IO_READFUNC xReadFunc,		/* Input function */
	IO_WRITEFUNC xWriteFunc,	/* Input function */
	IO_SEEKFUNC xSeekFunc,		/* Input function */
	IO_SIZEFUNC xSizeFunc,		/* Input function */
	IO_TELLFUNC xTellFunc,		/* Input function */
	IO_DOCHANDLE *pioDocHandle	/* Returned */
);


/*
 * IO_OpenDocumentReadWrite()
 */
ErrCode CDECL
IO_OpenDocumentReadWrite(
	IO_INSTHANDLE ioInstHandle,
	DOCTYPE ioDocType,
	Int32 dwClientID,		/* Input client ID */
	UInt8 version,			/* Version of IO format to create */
	UInt32 options,			/* Format specific options */
	IO_READFUNC xReadFunc,		/* Input function */
	IO_WRITEFUNC xWriteFunc,	/* Input function */
	IO_SEEKFUNC xSeekFunc,		/* Input function */
	IO_SIZEFUNC xSizeFunc,		/* Input function */
	IO_TELLFUNC xTellFunc,		/* Input function */
	IO_DOCHANDLE *pioDocHandle	/* Returned */
);


/*
 * IO_CopyImage()
 */
ErrCode CDECL
IO_CopyImage(
	IO_IMGHANDLE ioSrcImgHandle,
	IO_IMGHANDLE ioParentImgHandle,
	IO_PAGEHANDLE ioDstPageHandle,
	UInt32 dwImgLayer,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwBestStripHeight,
	IO_IMGHANDLE *pioDstImgHandle
);


/*
 * IO_CopyPage()
 */
ErrCode CDECL
IO_CopyPage(
	IO_DOCHANDLE ioSrcDocHandle,
	IO_DOCHANDLE ioDstDocHandle,
	Int32 dwSrcPageOffset,
	Int32 dwDstPageOffset,
	Bool copyAnnotations
);


/*
 * IO_CopyDoc()
 */
ErrCode CDECL
IO_CopyDoc(
	IO_DOCHANDLE ioSrcDocHandle,
	IO_DOCHANDLE ioDstDocHandle
);


/*
 * IO_AddPageStart()
 */
ErrCode CDECL
IO_AddPageStart(
	IO_DOCHANDLE ioDocHandle,
	Int32 dwPageOffset,		    /* Page offset */
	IO_PAGEHANDLE *pioPageHandle	    /* Page handle returned */
);


/*
 * IO_AddPageFinish()
 */
ErrCode CDECL
IO_AddPageFinish(
	IO_PAGEHANDLE ioPageHandle
);


/*
 * IO_DeletePage()
 */
ErrCode CDECL
IO_DeletePage(
	IO_PAGEHANDLE ioPageHandle
);


/*
 * IO_CheckImageInfo()
 */
ErrCode CDECL
IO_CheckImageInfo(
	IO_PAGEHANDLE ioPageHandle,
	UInt32 *pdwImgLayer,
	UInt32 *pdwImgOrder,
	IO_COLORTYPE *pxColorType,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	UInt8 *pPalette,
	IO_COMPRESSION *pxCompression,
	IO_COLORSPACE *pxColorSpace
);


/*
 * IO_AddImage()
 */
ErrCode CDECL
IO_AddImage(
	IO_PAGEHANDLE ioPageHandle,	/* Page to attach to */
	UInt32 dwImgLayer,		/* Image layer in the IO */
	IO_COLORTYPE xColorType,	/* Color type (IO_COLORTYPE_BINARY
					 * ...) */
	UInt32 dwXOffset,		/* X location in page coordinates */
	UInt32 dwYOffset,		/* Y location in page coordinates */
	UInt32 dwXResolution,		/* X Resolution of image in dpi */
	UInt32 dwYResolution,		/* Y Resolution of image in dpi */
	UInt32 dwWidth,			/* Width in image's X resolution units
					 */
	UInt32 dwHeight,		/* Height in image's Y resolution units
					 */
	UInt32 dwBestStripHeight,	/* For best performance, read strips
					 * with height a multiple of this. */
	UInt8 *pPalette,		/* Color palette if IO_COLORTYPE is:
					 * IO_COLORTYPE_COLOR4 or
					 * IO_COLORTYPE_COLOR8.
					 * Palette is RGB triples (chunky):
					 * RGBRGB... */
	IO_IMGHANDLE *pioImgHandle	/* Returned image handle for image to
					 * be added */
);


/*
 * IO_DeleteImage()
 */
ErrCode CDECL
IO_DeleteImage(
	IO_IMGHANDLE ioImgHandle
);


/*
 * IO_AddMaskImage()
 */
ErrCode CDECL
IO_AddMaskImage(
	IO_IMGHANDLE ioParentImgHandle,	/* Parent image to attach to */
	UInt32 dwBestStripHeight,	/* For best performance, write strips
					 * with height a multiple of this. */
	IO_IMGHANDLE *pImgHandle	/* Returned image handle for image to
					 * be added */
);


/*
 * IO_AddHintsImage()
 */
ErrCode CDECL
IO_AddHintsImage(
	IO_IMGHANDLE ioImgHandle,	/* Parent image to attach to */
	UInt32 dwBpp,			/* Image depth, in bits/pixel */
	UInt32 dwXOffset,		/* X location in page coordinates */
	UInt32 dwYOffset,		/* Y location in page coordinates */
	UInt32 dwXResolution,		/* X Resolution of image in dpi */
	UInt32 dwYResolution,		/* Y Resolution of image in dpi */
	UInt32 dwWidth,			/* Width in image's X resolution units
					 */
	UInt32 dwHeight,		/* Height in image's Y resolution units
					 */
	UInt32 dwBestStripHeight,	/* For best performance, read strips
					 * with height a multiple of this. */
	IO_IMGHANDLE *pioImgHandle	/* Returned image handle for image to
					 * be added */
);


/*
 * IO_AddRenderingHints()
 */
ErrCode CDECL
IO_AddRenderingHints(
    IO_IMGHANDLE ioImgHandle,
	char *pHintsString,
	UInt32 dwHintLength
);


/*
 * IO_ImageWriteStart()
 */
ErrCode CDECL
IO_ImageWriteStart(
	IO_IMGHANDLE ioImgHandle,
	UInt32 dwFlags, 
	IO_COMPRESSION *pxCompression, /* IO_COMP_UNKNOWN yields default */
	UInt32 *pdwJPEGQuality,	/* 0-100:
				 *   0 - default used and passed back
				 *  75   good quality, best compression
				 *  90   better quality, good compression
				 * 100   best quality, okay compression */
	IO_COLORSPACE *pxColorSpace /* Stored color space to use */
);	  

	
/*
 * IO_CompressedWriteStrip()
 */
ErrCode CDECL
IO_CompressedWriteStrip(
	IO_IMGHANDLE ioImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 *pCmpBuffer,
	UInt32 dwCmpBytes,
	UInt32 *pdwBytesStored
);


/*
 * IO_StorePositionBlock()
 */
ErrCode CDECL
IO_StorePositionBlock(
	IO_IMGHANDLE ioImgHandle,
	UInt8 *pPosBlock,
	UInt32 dwPosSize,
	UInt32 *dwDictIDList,
	UInt16 wDictCount
);


/*
 * IO_ImageWriteStrip()
 */
ErrCode CDECL
IO_ImageWriteStrip(
	IO_IMGHANDLE ioImgHandle,
	UInt8 **ppImagePlanes,
	UInt32 dwBytesPerRow,
	UInt32 dwBufferRowOffset,
	UInt32 dwStripHeight,
	UInt32 dwImgRowOffset,
	UInt32 *pRowsFilled
);

	
/*
 * IO_StoreSymbolDictionary()
 */
ErrCode CDECL
IO_StoreSymbolDictionary(
	IO_DOCHANDLE ioDocHandle,
	UInt8 *pDictBlock,
	UInt32 dwDictBytes,
	UInt32 dwDictBlockBytes,
	UInt32 *pdwDictID
);

/*
 * IO_ShareColorTokenImage()
 */
ErrCode CDECL
IO_ShareColorTokenImage(
	IO_IMGHANDLE ioImgHandle,
	UInt32 dwBinaryImageBytes,
	IO_COLORTYPE xColorType,
	IO_IMGHANDLE *pIoImgHandle
);


/*
 * IO_ImageWriteFinish()
 */
ErrCode CDECL
IO_ImageWriteFinish(
	IO_IMGHANDLE ioImgHandle
);
	

/*
 * IO_ImageSkip()
 */
ErrCode CDECL
IO_ImageSkip(
	IO_IMGHANDLE ioImgHandle
);
	

/* Tokenization flags */	
#define TOKENIZE_OFF			(0x00)	/* No tokenization */
#define TOKENIZE_ON			(0x01)	/* Tokenization enabled */
#define TOKENIZE_LOSSLESS	    	(0x02)	/* Tokenization in lossless mode */
#define TOKENIZE_BINARYSCAN		(0x04)	/* Image was scanned with a binary
						 * scanner, indicating that halftones
						 * may be removed and lossy dither
						 * compressed before tokenization */

/*
 * IO_TokenizeImage()
 */
ErrCode CDECL
IO_TokenizeImage(
	IO_IMGHANDLE ioImgHandle,
	UInt8 *pImgData,	/* One channel only */
	UInt32 dwBytesPerRow,	/* Bytes per row */
	UInt32 dwFlags		/* Flags: Lossless Tokenization/Binary Scan (see TOKEN_FLAGS) */
);


/*
 * IO_WriteTokenImages()
 */
ErrCode CDECL
IO_WriteTokenImages(
	IO_DOCHANDLE ioDocHandle
);

/*
 * IO_WriteTokenImage()
 */
ErrCode CDECL
IO_WriteTokenImage(
	IO_IMGHANDLE ioImgHandle
);


/*
 * IO_WriteTokenDictionary()
 */
ErrCode CDECL
IO_WriteTokenDictionary(
	IO_DOCHANDLE ioDocHandle
);


/*
 * IO_SetAnnotationOffsets
 */
ErrCode CDECL
IO_SetAnnotationOffsets(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 dwAnnotationOffsets
);


/*
 * IO_SetAuthor()
 */
ErrCode CDECL
IO_SetAuthor(
	IO_DOCHANDLE ioDocHandle,
	Int8 *pAuthorBlock,
	UInt16 wAuthorBytes
);


#if 0
/* This is not exposed */
/*
 * IO_SetBitsPerSample()
 */
ErrCode CDECL
IO_SetBitsPerSample(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwBitsPerSample,
	UInt16 wBitsPerSampleValues
);
#endif


/*
 * IO_SetChromaPositioning()
 */
ErrCode CDECL
IO_SetChromaPositioning(
    IO_IMGHANDLE ioImgHandle,
	UInt16 wPositioning
);


/*
 * IO_SetChromaSubSampling()
 */
ErrCode CDECL
IO_SetChromaSubSampling(
    IO_IMGHANDLE ioImgHandle,
	UInt16 wChromaSubSamplingHor,
	UInt16 wChromaSubSamplingVer
);


/*
 * IO_SetCodingMethods()
 */
ErrCode CDECL
IO_SetCodingMethods(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 dwCodingMethods
);


/*
 * IO_SetCompression()
 */
ErrCode CDECL
IO_SetCompression(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 dwCompression
);


/*
 * IO_SetDateTime()
 */
ErrCode CDECL
IO_SetDateTime(
    IO_PAGEHANDLE ioPageHandle,
	Int8 *pDateTime
);


/*
 * IO_SetDecode()
 */
ErrCode CDECL
IO_SetDecode(
    IO_IMGHANDLE ioImgHandle,
	Int32 *pdwDecode,
	UInt32 dwDecodeValues
);


/*
 * IO_SetDocumentName()
 */
ErrCode CDECL
IO_SetDocumentName(
    IO_DOCHANDLE ioDocHandle,
	Int8 *pDocumentName,
	UInt32 dwDocumentNameBytes
);


/*
 * IO_SetFaxProfile()
 */
ErrCode CDECL
IO_SetFaxProfile(
    IO_PAGEHANDLE ioPageHandle,
	UInt8 faxProfile
);


/*
 * IO_SetFillOrder()
 */
ErrCode CDECL
IO_SetFillOrder(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 fillOrder
);


/* added by EEP 20010328 */
/*
 * IO_SetHostComputer()
 */
ErrCode CDECL
IO_SetHostComputer(
    IO_IMGHANDLE ioImgHandle,
	Int8 *pHostComputer,
	UInt32 dwHostComputerBytes
);


/*
 * IO_SetIFDBaseColor()
 */
ErrCode CDECL
IO_SetIFDBaseColor(
    IO_IMGHANDLE ioImgHandle,
	Int32 *pdwIFDBaseColor,
	UInt32 dwIFDBaseColorValues
);


/*
 * IO_SetImageBaseColor()
 */
ErrCode CDECL
IO_SetImageBaseColor(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pImageBaseColor,
	UInt32 dwImageBaseColorValues
);


/*
 * IO_SetImageDescription()
 */
ErrCode CDECL
IO_SetImageDescription(
    IO_IMGHANDLE ioImgHandle,
	Int8 *pImageDescription,
	UInt32 dwImageDescriptionBytes
);


/*
 * IO_SetT88Options()
 */
ErrCode CDECL
IO_SetT88Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwT88Options
);


/*
 * IO_SetModeNumber()
 */
ErrCode CDECL
IO_SetModeNumber(
    IO_PAGEHANDLE ioPageHandle,
	UInt8 dwModeNumber
);


/*
 * IO_SetOrientation()
 */
ErrCode CDECL
IO_SetOrientation(
    IO_IMGHANDLE ioImgHandle,
	UInt16 wOrientation
);


/*
 * IO_SetPageNumber()
 */
ErrCode CDECL
IO_SetPageNumber(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 wPageNumber,
	UInt16 wPageCount
);


/*
 * IO_SetPhotometricInterpretation()
 */
ErrCode CDECL
IO_SetPhotometricInterpretation(
    IO_IMGHANDLE ioImgHandle,
	UInt16 wPhotometricInterpretation
);


/*
 * IO_SetPlanarConfiguration()
 */
ErrCode CDECL
IO_SetPlanarConfiguration(
    IO_IMGHANDLE ioImgHandle,
	UInt16 wPlanarConfiguration
);


/*
 * IO_SetProfileType()
 */
ErrCode CDECL
IO_SetProfileType(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 dwProfileType
);


/*
 * IO_SetReferencePage()
 */
ErrCode CDECL
IO_SetReferencePage(
    IO_IMGHANDLE ioImgHandle,
	Int32 dwReferencePage
);


/*
 * IO_SetResolutionUnit()
 */
ErrCode CDECL
IO_SetResolutionUnit(
    IO_IMGHANDLE ioImgHandle,
	UInt16 wResolutionUnit
);


/*
 * IO_SetRowsPerStrip()
 */
ErrCode CDECL
IO_SetRowsPerStrip(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwRowsPerStrip
);


#if 0
/*
 * IO_SetSamplesPerPixel()
 */
ErrCode CDECL
IO_SetSamplesPerPixel(
    IO_IMGHANDLE ioImgHandle,
	UInt<32 dwSamplesPerPixel
);
#endif


/*
 * IO_SetSoftware()
 */
ErrCode CDECL
IO_SetSoftware(
    IO_IMGHANDLE ioImgHandle,
	Int8 *pSoftware,
	UInt32 dwSoftwareBytes
);


/*
 * IO_SetTargetPrinter()
 */
ErrCode CDECL
IO_SetTargetPrinter(
    IO_PAGEHANDLE ioPageHandle,
	Int8 *pTargetPrinter,
	UInt32 dwTargetPrinterBytes
);


/*
 * IO_SetThresholding()
 */
ErrCode CDECL
IO_SetThresholding(
    IO_IMGHANDLE ioImgHandle,
	UInt16 wThresholding
);


/*
 * IO_SetT4Options()
 */
ErrCode CDECL
IO_SetT4Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwT4Options
);


/*
 * IO_SetT6Options()
 */
ErrCode CDECL
IO_SetT6Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwT6Options
);


/*
 * IO_SetVersionYear()
 */
ErrCode CDECL
IO_SetVersionYear(
    IO_IMGHANDLE ioImgHandle,
	UInt8 *pVersionYear,
	UInt32 dwVersionYearBytes
);


/*
 * IO_SetYcrCbCoefficients()
 */
ErrCode CDECL
IO_SetYCbCrCoefficients(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwYcrCbCoefficients,
	UInt32 dwYCbCrCoefficientValues
);


/*
 * IO_SetPredictor()
 */
ErrCode CDECL
IO_SetPredictor(
    IO_IMGHANDLE ioImgHandle,
	UInt16 wPredictor
);


/*
 * IO_SetAnnotTableEntryCount()
 */
ErrCode CDECL
IO_SetAnnotTableEntryCount(
	IO_DOCHANDLE ioDocHandle,
	UInt16 wEntriesPerTable
);


/*
 * IO_SetPageTableEntryCount()
 */
ErrCode CDECL
IO_SetPageTableEntryCount(
	IO_DOCHANDLE ioDocHandle,
	UInt16 wEntriesPerTable
);


/*
 * IO_SetResourceTableEntryCount()
 */
ErrCode CDECL
IO_SetResourceTableEntryCount(
	IO_DOCHANDLE ioDocHandle,
	UInt16 wEntriesPerTable
);


/*
 * IO_AddAnnotation()
 */
ErrCode CDECL
IO_AddAnnotation(
	IO_PAGEHANDLE ioPageHandle,
	UInt16 wAnnotType,		    /* Type of annotation */
	UInt32 dwXOffset,		    /* Offset in X direction in page
					     * coordinates */
	UInt32 dwYOffset,		    /* Offset in X direction in page
					     * coordinates */
	UInt32 dwWidth,			    /* Width in page coordinates */
	UInt32 dwHeight,		    /* Height in page coordinates */
	UInt32 dwFlags,			    /* Annotation flags (see ANNOT_...
					     * above) */
	UInt8 *pBuf,			    /* Annotation data */
	UInt32 dwAnnotLength,		    /* Length of annotation data, in
					     * bytes */
	UInt16 *pAnnotID		    /* ID given to this annotation
					     * (offset) */
); 


/*
 * IO_DeleteAnnotation()
 */
ErrCode CDECL
IO_DeleteAnnotation(
	IO_PAGEHANDLE ioPageHandle,
	UInt16 wAnnotOffset
);


/*
 * IO_Compress()
 */
ErrCode CDECL
IO_Compress(
	IO_IMGHANDLE ioImgHandle,	/* Image handle */
	IO_COMPRESSION xCompression,	/* See IO_COMPRESSION types */
	UInt32 *pdwQuality,		/* Quality to use, actual returned */
	IO_COLORTYPE xColorType,	/* Type of input image color */
	IO_COLORSPACE ioColorSpace,	/* Color space of compressed image */
	UInt32 dwFlags,			/* IO_IMGFLAGS_... combo */
	UInt8 *pSrc[3],			/* Source buffer channels */
	UInt32 dwBytesPerRow,		/* Bytes per row in image buffer */
	UInt32 dwRowOffset,		/* Row offset in image of data to
					 * compress */
	UInt32 dwWidth,			/* Columns of COLORTYPE in src image */
	UInt32 dwHeight,		/* Rows in src image */
	UInt8 **ppDst,			/* Pointer to destination compressed
					 * data buffers */
	UInt32 *pdwBytesInDst,		/* Pointer to byte count of destination
					 * buffer */
	UInt32 *pdwBytesFilled		/* Actual bytes in destination filled
					 */
);


#ifdef NOIPCORE
#define IO_Decompress(ioImgHandle, xCompression, xColorType, dwFlags, pSrc, \
 dwBytesInSrc, dwRowOffset, dwWidth, dwHeight, pDst, dwBytesPerRow, \
 pdwRowsFilled)		IO_NOSUPPORT
#else
/*
 * IO_Decompress()
 */
ErrCode CDECL
IO_Decompress(
	IO_IMGHANDLE ioImgHandle,	/* Image handle */
	IO_COMPRESSION xCompression,	/* See IO_COMPRESSION types */
	IO_COLORTYPE xColorType,	/* Type of color */
	UInt32 dwFlags,			/* IO_IMGFLAGS_... bit combo */
	UInt8 *pSrc,			/* Source buffer */
	UInt32 dwBytesInSrc,		/* Byte count of source buffer */
	UInt32 dwRowOffset,		/* Destination row offset */
	UInt32 dwWidth,			/* Columns of COLORTYPE in dst image */
	UInt32 dwHeight,		/* Rows in dst image */
	UInt8 *pDst[3],			/* Destination image buffer channels */
	UInt32 dwBytesPerRow,		/* Bytes per row in image buffer */
	UInt32 *pdwRowsFilled		/* Number of rows successfully filled
					 */
);
#endif


/*
 * IO_UseColorStrips()
 */
ErrCode CDECL
IO_UseColorStrips(
    IO_IMGHANDLE ioImgHandle
);


/*
 * IO_UseMultiStripSubIFDs()
 */
ErrCode CDECL
IO_UseMultiStripSubIFDs(
    IO_PAGEHANDLE ioPageHandle
);


/*
 * IO_UseNoGPIFD()
 */
ErrCode CDECL
IO_UseNoGPIFD(
    IO_DOCHANDLE ioDocHandle
);


/*
 * IO_UseVariableStripHeights()
 */
ErrCode CDECL
IO_UseVariableStripHeights(
    IO_IMGHANDLE ioImgHandle
);


/*
 * IO_sRGB2ITULAB()
 */
ErrCode CDECL
IO_sRGB2ITULAB(
	      UInt8 sR8,
	      UInt8 sG8,
	      UInt8 sB8,
	      UInt16 *pwL,
	      UInt16 *pwa,
	      UInt16 *pwb,
	      Int32 *pdwGammut
);


/*
 * IO_ITULAB2sRGB()
 */
ErrCode CDECL
IO_ITULAB2sRGB(
	      UInt16 L,
	      UInt16 a,
	      UInt16 b,
	      UInt8 *psR8,
	      UInt8 *psG8,
	      UInt8 *psB8,
	      Int32 *pdwGammut
);


/*
 * IO_LinearRGB2ITULAB()
 */
ErrCode CDECL
IO_LinearRGB2ITULAB(
	      UInt8 lR8,
	      UInt8 lG8,
	      UInt8 lB8,
	      UInt16 *pwL,
	      UInt16 *pwa,
	      UInt16 *pwb,
	      Int32 *pdwGammut
);


/*
 * IO_ITULAB2LinearRGB()
 */
ErrCode CDECL
IO_ITULAB2LinearRGB(
	      UInt16 L,
	      UInt16 a,
	      UInt16 b,
	      UInt8 *plR8,
	      UInt8 *plG8,
	      UInt8 *plB8,
	      Int32 *pdwGammut
);


/*
 * IO_sRGB2CIELAB()
 */
ErrCode CDECL
IO_sRGB2CIELAB(
	      UInt8 sR8,
	      UInt8 sG8,
	      UInt8 sB8,
	      UInt16 *pwL,
	      Int16 *pwa,
	      Int16 *pwb
);


/*
 * IO_CIELAB2sRGB()
 */
ErrCode CDECL
IO_CIELAB2sRGB(
	      UInt16 L,
	      Int16 a,
	      Int16 b,
	      UInt8 *psR8,
	      UInt8 *psG8,
	      UInt8 *psB8
);


/*
 * IO_LinearRGB2CIELAB()
 */
ErrCode CDECL
IO_LinearRGB2CIELAB(
	      UInt8 lR8,
	      UInt8 lG8,
	      UInt8 lB8,
	      UInt16 *pwL,
	      UInt16 *pwa,
	      UInt16 *pwb
);


/*
 * IO_CIELAB2LinearRGB()
 */
ErrCode CDECL
IO_CIELAB2LinearRGB(
	      UInt16 L,
	      Int16 a,
	      Int16 b,
	      UInt8 *plR8,
	      UInt8 *plG8,
	      UInt8 *plB8
);


/*
 * IO_sRGB2ICCLAB()
 */
ErrCode CDECL
IO_sRGB2ICCLAB(
	      UInt8 sR8,
	      UInt8 sG8,
	      UInt8 sB8,
	      UInt16 *pwL,
	      UInt16 *pwa,
	      UInt16 *pwb
);


/*
 * IO_ICCLAB2sRGB()
 */
ErrCode CDECL
IO_ICCLAB2sRGB(
	      UInt16 L,
	      UInt16 a,
	      UInt16 b,
	      UInt8 *psR8,
	      UInt8 *psG8,
	      UInt8 *psB8
);


/*
 * IO_LinearRGB2ICCLAB()
 */
ErrCode CDECL
IO_LinearRGB2ICCLAB(
	      UInt8 lR8,
	      UInt8 lG8,
	      UInt8 lB8,
	      UInt16 *pwL,
	      UInt16 *pwa,
	      UInt16 *pwb
);


/*
 * IO_ICCLAB2LinearRGB()
 */
ErrCode CDECL
IO_ICCLAB2LinearRGB(
	      UInt16 L,
	      UInt16 a,
	      UInt16 b,
	      UInt8 *plR8,
	      UInt8 *plG8,
	      UInt8 *plB8
);


/*
 * cuberoot()
 */
float CDECL
cuberoot(float x, float error);

#ifdef __cplusplus
}
#endif

#endif /* IOMRC_H_INCLUDED */
