/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

#ifndef IOPAGE_H_INCLUDED
#define IOPAGE_H_INCLUDED

#ifndef NOIPCORE
IP_RCSINFO(iopage_hRCSInfo, "$RCSfile: iopage.h,v $; $Revision: 1.1 $");
/* $Date: 2003/01/03 01:41:10 $ */
#endif

#ifdef __cplusplus
extern "C"
{
#endif

/* iopage.h
 *
 * DESCRIPTION
 *	 Declarations for IO page level structures.
 *
 */

/*
 * INCLUDES
 */
#include "iodoc.h"
#include "ioinfo.h"

#if 0

/*
 * CONSTANTS
 */

/* Calling sequence status */
#define ioPage_SEQ_GETINFO	0x0001	/* Get image info called */
#define ioPage_SEQ_READSTART	0x0010	/* Read start called */
#define ioPage_SEQ_READFINISH	0x0000	/* Read finish called */
#define ioPage_SEQ_READSTATUS	0x00f0	/* Read status bits */
#define ioPage_SEQ_WRITESTART	0x0100	/* Write start called */
#define ioPage_SEQ_WRITEPRIMARY	0x0200	/* Primary image added */
#define ioPage_SEQ_WRITEPENDING	0x0400	/* Write finish called, waiting on
					 * tokenization */
#define ioPage_SEQ_WRITEFINISH	0x0000	/* Page is completely written */
#define ioPage_SEQ_WRITESTATUS	0x0f00	/* Write status bits */
#define ioPage_SEQ_TOKENSTART	0x1000	/* Reference page set */

#endif

/*
 * MACROS
 */
#define IO_PAGEFILE(ioPage)		IO_DOCFILE(ioPage->ioDoc)
#define IO_PAGEINST(ioPage)		IO_PAGEFILE(ioPage)->instance
#define IO_PAGEFILETOKEN(ioImg)		IO_FILETOKEN(IO_PAGEFILE(ioPage))

#if 0

/*
 * TYPEDEFS
 */

/* Annotation list table header */
typedef struct AnnotTableHeader {
    UInt16 annotCount;		    /* Number of annotations in page (total) */
    UInt16 entriesPerTable;	    /* Number of annotation table entries per
				     * table */
    UInt16 bytesPerEntry;	    /* Number of bytes per table entries */
    UInt16 lastOrdinal;		    /* Ordinal (1, 2, ...) of last table */
    UInt32 firstTableOffset;	    /* File offset of first table */
} AnnotTableHeader;

/* Annotation table entry (describes an individual annotation) */
typedef struct AnnotTableEntry {
    UInt16 id;			    /* Unique ID of this annotation (offset) */
    UInt16 type;		    /* Annotation type (see io.h) */
    UInt32 xOffset;		    /* X offset in page coordinates */
    UInt32 yOffset;		    /* Y offset in page coordinates */
    UInt32 width;		    /* Width in page coordinates */
    UInt32 height;		    /* Height in page coordinates */
    UInt32 flags;		    /* Annotation flags (see ANNOT_xxx above)
				     */
    UInt32 bytes;		    /* Number of bytes in the data */
    Int32 offset;		    /* File offset of the annotation data */
    UInt8 *data;		    /* Annotation data */
} AnnotTableEntry;

/* Individual annotation table, linked if more than one */
typedef struct AnnotTable {
    struct AnnotTable *next;	    /* Link to next table (cached) */
    UInt16 ordinal;		    /* Ordinal (1, 2, ...) of this table */
    UInt16 validAnnotations;	    /* Number of valid annotations in table */
    AnnotTableEntry *entries;	    /* Allocated entry list */
    Int32 nextTableOffset;	    /* File offset of next annotation table */
} AnnotTable;

/* Page handle
 */
typedef struct ioPage {
    UInt32 handleID;		    /* Handle ID, consistency check */
    struct IO_PAGE *prev;		    /* Previous page in linked list */
    struct IO_PAGE *next;		    /* Next page in linked list */
    IO_DOC *ioDoc;			    /* Document structure */
    UInt32 seqStatus;		    /* Calling sequence status */
    IO_PAGEINFO pageInfo;	    /* Page information */
    UInt32 *extImageCounts;	    /* Image counts exposed to client (avoids true
				     * count getting corrupted by client) */
    struct IO_IMG *ioImgs;		    /* Image list (linked) */
    Int32 annotTableHeaderOffset;   /* Offset to annotation table header
				     * (identified by the AnnotationOffsets
				     * field) */
    struct AnnotTableHeader annotTableHeader;/* Annotation list table header */
    struct AnnotTable *firstAnnotTable;	/* Annotation list table for first */
} ioPage;

#endif

typedef struct IO_PAGE {
    UInt32 handleID;		    /* Handle ID, consistency check */
    UInt32 handleType;		    /* Handle type ID */
    struct IO_PAGE *prev;	    /* Previous in list */
    struct IO_PAGE *next;	    /* Next in list */
    UInt32 ioPageHandle;	    /* Lower level page handle */
    IO_PAGEINFO pageInfo;	    /* Page information */
    struct IO_DOC *ioDoc;	    /* Parent document structure */
    struct IO_IMG *ioImgs;	    /* First in linked list of images */
} IO_PAGE;



/*
 * PROTOTYPES
 */

#ifdef __cplusplus
}
#endif

#endif /* IOPAGE_H_INCLUDED */

