/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * iormrc.c
 *
 *	IO write of IO specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	iormrc.c:	IO READ source
 *
 *	Handles read-only IO handling.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: iormrc.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:30 $");

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_IO, 1)
#endif
#include "ipwcore.prv"
#ifdef _MSC_VER
#pragma pack(pop, PRE_IO)
#endif
#endif	/* NOIPCORE */

#include "iomrc.h"
#include "tfx.h"
#include "tfximage.h"
#include "iofile.h"
#include "iodoc.h"
#include "iorpriv.h"

#ifndef NOIPCORE
#include "tfxrpriv.h"	/* For wp_TfxGetStripeInfo() */
#endif

#ifndef IO_READONLY
/*#include "mrcwmrc.h" */
#endif /* IO_READONLY */


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initialize a IO instance, used to identify the IO state.
 *
 * LongDescription:::
 *  IO_InitInstance() initializes a IO instance, used to store file I/O
 *  routines connecting the instance to a handle, for maintaining the
 *  malloc/free functions, and for retaining the default authoring description.
 *
 * Input:::
 *  dwClientInstID::	Client ID assigned to this instance
 *
 * Output:::
 *  pioInstHandle::		Pointer to the instance handle returned
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, MRC
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_InitInstance(
    UInt32 dwClientInstID,
    IO_INSTHANDLE *pioInstHandle
)
{
    IO_INSTANCE *instData = NULL;
    IO_MemoryTally *tally = NULL;
    PROLOGUE;

    /* Return nothing if failure */
    *pioInstHandle = (IO_INSTHANDLE)instData;

    /* Here's the only calls to the system malloc */
    IO_CHECKMALLOC(instData = (IO_INSTANCE *)malloc(sizeof(*instData) +
				       sizeof(*tally)));
    tally = (void *)((char *)instData + sizeof(*instData));
    
    memset(instData, 0, sizeof(*instData));
    memset(tally, 0, sizeof(*tally));
    
    /* Set defaults */
    instData->handleID = (UInt32)instData;	/* Consistency check */
    instData->handleType = IO_HANDLE_TYPE;
    instData->nextDocID = 0;	/* Doc unique IDs */
    instData->ioMagicNum = IO_MAGIC_NUM;
    
    /* Preserve the default ID for switching back and forth between the
     * client ID and the reserved default */
    instData->defaultMallocID = (void *)tally;
    
    /* Save user identifier */
    instData->clientInstID = dwClientInstID;
    
    /* Initialize the malloc functions to the defaults */
    XERR_PASS_CLEANUP(IO_SetMallocFuncs((IO_INSTHANDLE)instData,
			    instData->defaultMallocID, wp_IoMalloc,
			    wp_IoFree));
    
    /* Return what we did, obscured from the client */
    *pioInstHandle = (IO_INSTHANDLE)instData;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (instData) {
	free(instData);
    }
    if (tally) {
	free(tally);
    }
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Declare a malloc and free function to be used for memory calls.
 *
 * LongDescription:::
 *  IO_SetMallocFuncs() declares a malloc and free function to be used in
 *  all subsequent memory calls. This allows the client to provide their own
 *  memory management. Without this call, default internal routines are used
 *  instead, which may use the system malloc()/free() pair.
 *
 * Input:::
 *  xInst::			IO instance handle, returned from
 *				IO_InitInstance()
 *  dwClientMallocID::		Identifier pass back to functions when called
 *  ioMalloc::			Memory allocator function
 *  ioFree::			Memory free'r function
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, MRC
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetMallocFuncs(
    IO_INSTHANDLE xInst,
    void *dwClientMallocID,
    IO_MALLOCFUNC ioMalloc,
    IO_FREEFUNC ioFree
)
{
    IO_INSTANCE *instData = (IO_INSTANCE *)xInst;
    PROLOGUE;
    
    IO_CheckHandleValidity(xInst);	/* Must be valid, even though not used */

    if (ioMalloc && ioFree) {	/* Must both be set to use clients */
	instData->malloc = ioMalloc;
	instData->free = ioFree;
	instData->mallocID = dwClientMallocID;
    } else if (!ioMalloc || !ioFree) {
	__error__ = IO_BADPARAMETER;
	goto ERR_CLEANUP;
    } else {
	
	/* NULL functions indicate to use the defaults */
	instData->malloc = wp_IoMalloc;
	instData->free = wp_IoFree;
	instData->mallocID = instData->defaultMallocID;
    }

#if 0
    /* Use these generic functions for calloc and realloc, both
     * of which run on top of the malloc and free function(s) just installed */
    instData->calloc = wp_IoCalloc;
    instData->realloc = wp_IoRealloc;
#endif
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Malloc a block of memory, using the malloc function in the IO_ routines
 *
 * LongDescription:::
 *  IO_Malloc() allocates a block of memory, using the malloc function set up
 *  in the IO_ routines.  If a new malloc function has been declared via
 *  IO_SetMallocFuncs(), then that routine will be called.  If no function has
 *  been declared, then the default one initially set up will be called.
 *  <BR>
 *  It's IMPORTANT that the routine to free blocks of memory allocated within
 *  the TIFF-FX I/O Library match the one that allocated them.  Especially
 *  when no call to IO_SetMallocFunc() is made, causing the initial routines
 *  to be used: which aren't malloc() and free(); they're a wrapper on top
 *  of these system calls.
 *
 * Input:::
 *  xInst::			IO instance handle, returned from
 *				IO_InitInstance()
 *  dwByteCount::		Identifier pass back to functions when called
 *
 * Output:::
 *  blockPtr::			Pointer to block of data to be returned
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, MRC
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_Malloc(
    IO_INSTHANDLE xInst,
    UInt32 dwByteCount,
    void **blockPtr
#ifdef SHOWERR
    , const char *file,
    const int line
#endif
)
{
    IO_INSTANCE *instData = (IO_INSTANCE *)xInst;
    PROLOGUE;
    
    IO_CheckHandleValidity(xInst);	/* Must be valid */

    /* Check parameter */
    XERR_GEN_CLEANUP(blockPtr == NULL, IO_BADPARAMETER);

    /* Mark as free, for now */
    *blockPtr = NULL;

    /* Make sure there is a malloc routine set up */
    XERR_GEN_CLEANUP(instData->free == NULL, IO_SEQUENCE);

    /* Allocated the block */
    if ((*blockPtr = instData->malloc(dwByteCount, 0, instData->mallocID
#ifdef SHOWERR
	, file, line
#endif
	)) == NULL) {
	__error__ = IO_NOMEMORY;
	goto ERR_CLEANUP;
    }
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free a block of memory, using the free function in the IO_ routines
 *
 * LongDescription:::
 *  IO_Free() frees a block of memory, using the free function set up
 *  in the IO_ routines.  If a new free function has been declared via
 *  IO_SetMallocFuncs(), then that routine will be called.  If no function has
 *  been declared, then the default one initially set up will be called.
 *  <BR>
 *  It's IMPORTANT that the routine to free blocks of memory allocated within
 *  the TIFF-FX I/O Library match the one that allocated them.  Especially
 *  when no call to IO_SetMallocFunc() is made, causing the initial routines
 *  to be used: which aren't malloc() and free(); they're a wrapper on top
 *  of these system calls.
 *
 * Input:::
 *  xInst::			IO instance handle, returned from
 *				IO_InitInstance()
 *
 * Output:::
 *  blockPtr::			Pointer to block of data allocated (address
 *				of pointer, will be NULL'd out)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, MRC
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_Free(
    IO_INSTHANDLE xInst,
    void **blockPtr
)
{
    IO_INSTANCE *instData = (IO_INSTANCE *)xInst;
    PROLOGUE;
    
    IO_CheckHandleValidity(xInst);	/* Must be valid */

    /* Check parameter */
    XERR_GEN_CLEANUP(blockPtr == NULL, IO_BADPARAMETER);

    /* Make sure there is a malloc routine set up */
    XERR_GEN_CLEANUP(instData->free == NULL, IO_SEQUENCE);

    /* Free the block */
    IO_FREE(*blockPtr);
    
    /* NULL out the pointer, to avoid reuse */
    *blockPtr = NULL;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the error process function to call whenever an error is encountered
 *
 * LongDescription:::
 *  IO_SetErrorProc() registers an error process function and an info handle,
 *  to be called whenever an error condition arises.
 *
 * Input:::
 *  ioInstHandle::		Instance handle, returned from IO_InitInstance
 *  errorProc::			Pointer to a callback function, when an error
 *				occurs
 *  errInfo::			Handle to be used when calling the callback
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, MRC
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  mtobin::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode IO_SetErrorProc(
	IO_INSTHANDLE ioInstHandle,
	ClientErrorProc errProc,
	void *errInfo
)
{
    IO_INSTANCE *instData = (IO_INSTANCE *)ioInstHandle;
    PROLOGUE;
    
    IO_CheckHandleValidity(ioInstHandle);	/* Must be valid */

    instData->errProc = errProc;
    instData->errInfo = errInfo;

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the "IO instance" ID associated with a IO document handle
 *
 * LongDescription:::
 *  IO_GetInstance() returns the "IO instance" ID previously associated
 *  with a IO document handle.
 *
 * Input:::
 *  ioDocHandle::		Active document handle
 *  pioInstHandle::		Pointer to the IO instance handle to be
 *				returned, originally returned from
 *				IO_InitInstance(), and to be passed in to
 *				IO_EndInstance().
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, MRC
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetInstance(
    IO_DOCHANDLE ioDocHandle,
    IO_INSTHANDLE *pioInstHandle
)
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;
    
    IO_CheckHandleValidity(ioDocHandle);

    *pioInstHandle = (IO_INSTHANDLE)IO_DOCINST(ioDoc);
    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  End a previously estalished IO instance.
 *
 * LongDescription:::
 *  IO_EndInstance() ends a previously estalished IO instance. It is an
 *  error to end an instance with a current document open (a call to
 *  IO_CloseDocument() is required first).
 *
 * Input:::
 *  xInst::			IO instance handle, returned from
 *				IO_InitInstance()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, MRC
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_EndInstance(
    IO_INSTHANDLE xInst
)
{
    IO_INSTANCE *instData = (IO_INSTANCE *)xInst;
    PROLOGUE;
    
    IO_CheckHandleValidity(xInst);

#ifndef NOIPCORE    
    /* End IPCore instance */
    if (instData->pIPWState && instData->myIPWState) {
	XERR_PASS_CLEANUP(w_IPWStateTerm((IPWState **)&instData->pIPWState));
    }
#endif

    if (instData->tfxInstance) {
	TFX_EndInstance(instData->tfxInstance);
    }

    /* Free the memory associated with the IO "glue" layer */
/*    XERR_PASS_CLEANUP(wp_IoDocFree(&ioDoc)); */

    /* Check known IO memory for memory leaks */
    if (XERR_IGNORE(wp_IoMemoryCheck((IO_MemoryTally *)instData->defaultMallocID))) {
	/* Assume something wasn't called correctly */
	__error__ = IO_SEQUENCE;
	/* Contine on with dealloc */
    }
    
    /* Due to potential realloc() calls, some operating systems allow this
     * pointer to propogate. Invalidate before we free it. */
    memset((void *)instData->defaultMallocID, 0, sizeof(IO_MemoryTally));
    memset(instData, 0, sizeof(*instData));

    if (instData->defaultMallocID) {
	free((void *)instData->defaultMallocID);
    }
    free(instData);
    
    if (__error__ != IO_NOERROR) {
	goto ERR_CLEANUP;    /* This is silly, but now we reference the label */
    }
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Check a file for a header identifying it as a currenly supported IO.
 *
 * LongDescription:::
 *  IO_IsIo() reads enough of a file to identify it as a currently
 *  support IO file. The client must only provide enough input parameters
 *  to allow reading and roaming the file to make such a decision. The client
 *  ID is a parameter passed back through the I/O functions, such that any of
 *  these supporting functions could use the ID to identify a file descriptor
 *  or file pointer, or even possibly a structure with this and a file state.
 *  The IO file's version and revision are returned if the file is a IO,
 *  and are zeroed otherwise. IO_NOSUPPORT is returned if not supported.
 *  IO_BADFILEFORMAT is returned if not a IO.
 *
 * Input:::
 *  xInst::			IO instance handle, returned from
 *				IO_InitInstance()
 *  dwClientID::		Client ID, passed back through all I/O funcs
 *  ioReadFunc::			Client file read function
 *  ioSeekFunc::			Client file seek function
 *  ioSizeFunc::			Client file size function
 *
 * Output:::
 *  pVersion::			Pointer to the version
 *  pRevision::			Pointer to the revision
 *  pdwOptions::		Pointer to the options
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_IsIo(
    IO_INSTHANDLE xInst,
	Int32 dwClientID,
	IO_READFUNC ioReadFunc,
	IO_SEEKFUNC ioSeekFunc,
	IO_SIZEFUNC ioSizeFunc,
	DOCTYPE *pioDocType,
	UInt8 *pVersion,
	UInt8 *pRevision,
	UInt32 *pdwOptions
    )
{
    PROLOGUE;

    /* Default to return no valid document type */
    *pioDocType = DOCTYPE_UNKNOWN;

    /* Try TIFF-FX/Coherent TIFF */
    if (XERR_IGNORE(TFX_IsTfx(
	xInst,
	    dwClientID,
	    ioReadFunc,
	    ioSeekFunc,
	    ioSizeFunc,
	    pVersion,
	    pRevision,
	    pdwOptions
	)) == IO_NOERROR) {
	switch (*pVersion) {
	case 7:
	    switch (*pRevision /* mode */) {
	    case 0:
	    default:
		/* We need some other checks here to determine the profile */
		/* Unfortunately, this means we read at least the first IFD
		 * and parse through it's contents.  That's a little more than
		 * a simple check */
		*pioDocType = DOCTYPE_TIFFFX_M;
		NORMAL_RETURN;
		break;
#if 0
	    case 1:
		*pioDocType = DOCTYPE_TIFFFX_C;
		NORMAL_RETURN;
		break;
#endif
	    }
	}
    }

    NORMAL_RETURN;	/* Propogates __error__ from last call */
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Open a document for read only
 *
 * LongDescription:::
 *  IO_OpenDocumentRead() begins a session of reading a document. The
 *  client must already have opened the file at the operating system level,
 *  so that the read and seek methods in the file token are available.
 *  This function will perform any initialization required for reading
 *  the file, including reading enough header information to verify
 *  that the file is in a format supported by this version of
 *  the IO library.
 *  <BR>
 *  A document handle is returned through the third parameter. This handle
 *  is passed to all subsequent function calls that use the document, and
 *  will encompass the I/O functions passed in here.
 *
 * Input:::
 *  ioInstHandle::		IO instance handle, returned from
 *				IO_InitInstance()
 *  dwClientID::		Input client ID
 *  ioReadFunc::			Client file read function
 *  ioWriteFunc::		Client file write function
 *  ioSeekFunc::			Client file seek function
 *  ioSizeFunc::			Client file size function
 *  ioTellFunc::			Client file tell function
 *
 * Output:::
 *  pioDocHandle::		Document handle used to refer to this
 *				page in other functions
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_OpenDocumentRead(
    IO_INSTHANDLE ioInstHandle,
	Int32 dwClientID,
	UInt32 options,
	IO_READFUNC ioReadFunc,
	IO_WRITEFUNC ioWriteFunc,
	IO_SEEKFUNC ioSeekFunc,
	IO_SIZEFUNC ioSizeFunc,
	IO_TELLFUNC ioTellFunc,
	IO_DOCHANDLE *pioDocHandle
    )
{
    IO_INSTANCE *instData = (IO_INSTANCE *)ioInstHandle;
    IO_DOC *ioDoc = NULL;
    IO_FILE *ioFile = NULL;
    DOCTYPE ioDocType = DOCTYPE_UNKNOWN;
    UInt8 version = 0, revision = 0;
    UInt32 isOptions = 0;
    UInt32 ioDocHandle;
    PROLOGUE;

    /* Clear the results for now */
    *pioDocHandle = 0UL;
    
    IO_CheckHandleValidity(ioInstHandle);
    
    /* Check for proper calling sequence */

    
    /* We must first get the document type, then go back and open the
     * correct format.  Note this will cause the wp_...Is... functions
     * to be called multiple times, but the function layout makes that
     * necessary. */
    XERR_IGNORE(IO_IsIo(
	ioInstHandle,
	dwClientID,
	ioReadFunc,
	ioSeekFunc,
	ioSizeFunc,
	&ioDocType,
	&version,
	&revision,
	&isOptions
	));

	/* Make sure we have the right instance set up */
	if (instData->tfxInstance == 0) {
	    XERR_PASS_CLEANUP(TFX_InitInstance(0, &instData->tfxInstance));
	    XERR_PASS_CLEANUP(TFX_SetMallocFuncs(instData->tfxInstance,
		(void *)instData->mallocID, instData->malloc, instData->free));
	}

	XERR_PASS_CLEANUP(TFX_OpenDocumentRead(
	    instData->tfxInstance,
		dwClientID,
		options,
	    	ioReadFunc,
		ioWriteFunc,
		ioSeekFunc,
		ioSizeFunc,
		ioTellFunc,
		&ioDocHandle
	    ));

    /* If everything went well up to here, we can probably continue
     * on */
    
    /* Create a new document (handle) */
    IO_CHECKMALLOC(ioDoc = (IO_DOC *)IO_MALLOC(sizeof(IO_DOC)));
    memset(ioDoc, 0, sizeof(*ioDoc));
    ioDoc->handleID = (UInt32)ioDoc;	/* Self identity */
    ioDoc->handleType = IO_HANDLE_TYPE;
    
    /* Create a link back to the instance */
    IO_CHECKMALLOC(ioFile = (IO_FILE *)IO_MALLOC(sizeof(IO_FILE)));
    memset(ioFile, 0, sizeof(*ioFile));
    ioFile->handleID = (UInt32)ioFile;	/* Self identity */
    
    /* Setup the file handle */
    ioFile->instance = instData;
    IO_FILETOKEN(ioFile)->dwClientFileID = dwClientID;
    IO_FILETOKEN(ioFile)->FileRead = ioReadFunc;
    IO_FILETOKEN(ioFile)->FileWrite = ioWriteFunc;
    IO_FILETOKEN(ioFile)->FileSeek = ioSeekFunc;
    IO_FILETOKEN(ioFile)->FileSize = ioSizeFunc;
    IO_FILETOKEN(ioFile)->FileTell = ioTellFunc;
    
    /* Retain the file handle */
    ioDoc->ioFile = ioFile;
    
    /* Identify the type of document found */    
    ioDoc->ioDocType = ioDocType;
    ioDoc->ioDocHandle = ioDocHandle;

    /* Return the result */
    *pioDocHandle = ioDoc->handleID;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (ioDoc) {
	IO_FREE(ioDoc);
    }
    if (ioFile) {
	IO_FREE(ioFile);
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the basic content of the document.
 *
 * LongDescription:::
 *  IO_GetDocInfo() returns the document specific content, including: file
 *  version, file revision, number of pages, authoring description, and the
 *  reference page (if appropriate).
 *
 * Input:::
 *  ioDocHandle::		Input document handle
 *
 * Output:::
 *  pdwVersion::		Version of file format
 *  pdwRevision::		Revision of this version of file format
 *  pdwNPages::			Number of pages in file
 *  ppAuthorBlock::		Author string (NULL if none)
 *  pwAuthorBlockBytes::	Size of author string
 *  pdwReferencePageOffset::	Reference page offset (-1 if none)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, MRC
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetDocInfo(
	IO_DOCHANDLE ioDocHandle,
	UInt8 *pVersion,
	UInt8 *pRevision,
	UInt16 *pwNPages
)
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_WARNOK_CLEANUP(TFX_GetDocInfo(
	    ioDoc->ioDocHandle,
		pVersion,
		pRevision,
		pwNPages
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the basic content of the page
 *
 * LongDescription:::
 *  IO_GetPageInfo() returns the page specific content, including:
 *  image layer counts, number of layers present, annotation count, and a
 *  handle used to reference this page for other functions.
 *
 * Input:::
 *  ioDocHandle::		Document handle
 *  dwPageOffset::		Page offset within the document (-1 for the
 *				last page, including incompleted last page)
 *
 * Output:::
 *  ppdwImageCounts::		Number of images for each image layer (array,
 *				this may be NULL, in which case no array is
 *				returned)
 *  pdwLayersPresent::		Number of image layers present (indicates size
 *				of array pdwImageLayers[], may have missing
 *				layers)
 *  pdwAnnotationCount::	Number of annotations on page
 *  pioPageHandle::		Page handle, used to identify this page
 *				in the page level routines
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetPageInfo(
    IO_DOCHANDLE ioDocHandle,
	Int32 dwPageOffset,
	UInt32 **ppdwImageCounts,
	UInt32 *pdwLayersPresent,
	UInt16 *pwAnnotationCount,
	IO_PAGEHANDLE *pioPageHandle
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    IO_PAGE *ioPage = NULL;
    IO_INSTANCE *instData = NULL;
    UInt32 newPageHandle;
    PROLOGUE;

    /* Clear the result */
    *pioPageHandle = 0;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_WARNOK_CLEANUP(TFX_GetPageInfo(
	    ioDoc->ioDocHandle,
		dwPageOffset,
		ppdwImageCounts,
		pdwLayersPresent,
		pwAnnotationCount,
		&newPageHandle
	    ));

    /* Search through the image list for this new one.  We must search through
     * the entire list, since only the lower level lists are maintained in
     * page order. */
    ioPage = ioDoc->ioPages;
    while (ioPage != NULL) {
	if (ioPage->ioPageHandle == newPageHandle) {
	    break;
	}
	ioPage = ioPage->next;
    }

    if (ioPage == NULL) {

	/* Create our own copy of the container */
	instData = IO_DOCINST(ioDoc);
	IO_CHECKMALLOC(ioPage = (IO_PAGE *)IO_MALLOC(sizeof(*ioPage)));
	memset(ioPage, 0, sizeof(*ioPage));
	ioPage->handleID = (UInt32)ioPage;	/* Self identity */
	ioPage->handleType = IO_HANDLE_TYPE;
    
        /* Link back to it's parent */
        ioPage->ioDoc = ioDoc;

        /* Identify the lower level handle */
        ioPage->ioPageHandle = newPageHandle;

        if (ioDoc->ioPages == NULL) {
            /* Add the page to the doc */
            ioDoc->ioPages = ioPage;
	}
        else {
	    IO_PAGE *tempPage;

            /* Put it into the linked list at the end */
	    tempPage = ioDoc->ioPages;
	    while (tempPage->next != NULL) {
		tempPage = tempPage->next;
	    }
	    
	    tempPage->next = ioPage;
	}
    }

    /* Return the result */
    *pioPageHandle = ioPage->handleID;

    NORMAL_RETURN;
    
 ERR_CLEANUP:

    if (ioPage) {
	IO_FREE(ioPage);
    }
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image description, including IO image counts.
 *
 * LongDescription:::
 *  IO_GetImageInfo() gets the description of an image: position relative
 *  to the Primary image, resolution, size, stored strip size, stored
 *  compression type, color type, palette (if any), and mask image count
 *  (if any).
 *  <BR>
 *  Returned is an image handle to be used to reference this image in later
 *  calls. A call to IO_CloseDocument() renders this image handle invalid.
 *
 * Input:::
 *  ioPageHandle::		Page handle
 *  dwImgLayer::		Image layer (1, 2, ...)
 *  dwImgOffset::		Image offset for image within that layer (0,
 *				1, ...)
 *
 * Output:::
 *  pdwXOffset::		X location in page coordinates
 *  pdwYOffset::		Y location in page coordinates
 *  pdwXResolution::		X Resolution of image in dpm
 *  pdwYResolution::		Y Resolution of image in dpm
 *  pdwWidth::			Width in image's X resolution units
 *  pdwHeight::			Height in image's Y resolution units
 *  pdwBestStripHeight::	For best performance, pass strips with size
 *				and boundary that are multiples of this (strips
 *				are stored with this size)
 *  pioCompression::		Compression type for src/dst
 *  pioColorType::		Color type (see IO_COLORTYPE
 *  pPalette::			Color palette if IO_COLORTYPE is
 *				IO_COLORTYPE_COLOR4 or IO_COLORTYPE_COLOR8.
				Palette is RGB triples (chunky): RGBRGB...
 *  pdwMaskImages::		Number of mask images
 *  pioImgHandle::		Returned image handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetImageInfo(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 dwImgLayer,
	UInt32 dwImgOffset,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	IO_COMPRESSION *pioCompression,
	IO_COLORTYPE *pioColorType,
	UInt8 *pPalette,
	UInt32 *pdwMaskImages,
	IO_IMGHANDLE *pioImgHandle
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    IO_IMG *ioImg = NULL;
    IO_INSTANCE *instData = NULL;
    UInt32 newImgHandle;
    PROLOGUE;

    /* Clear the result */
    *pioImgHandle = 0;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	*pdwMaskImages = 0;
	XERR_WARNOK_CLEANUP(TFX_GetImageInfo(
	    ioPage->ioPageHandle,
		dwImgLayer,
		dwImgOffset,
		pdwXOffset,
		pdwYOffset,
		pdwXResolution,
		pdwYResolution,
		pdwWidth,
		pdwHeight,
		pdwBestStripHeight,
		pioCompression,
		pioColorType,
		pPalette,
		&newImgHandle
	    ));

    /* Search through the image list for this new one.  We must search through
     * the entire list, since only the lower level lists are maintained in
     * layer/offset order. */
    ioImg = ioPage->ioImgs;
    while (ioImg != NULL) {
	if (ioImg->ioImgHandle == newImgHandle) {
	    break;
	}
	ioImg = ioImg->next;
    }

    if (ioImg == NULL) {

        /* Create our own copy of the container */
        instData = IO_PAGEINST(ioPage);
        IO_CHECKMALLOC(ioImg = (IO_IMG *)IO_MALLOC(sizeof(*ioImg)));
        memset(ioImg, 0, sizeof(*ioImg));
	ioImg->handleID = (UInt32)ioImg;	/* Self identity */
	ioImg->handleType = IO_HANDLE_TYPE;
    
        /* Link back to it's parent */
        ioImg->ioPage = ioPage;

        /* Identify the lower level handle */
        ioImg->ioImgHandle = newImgHandle;

        if (ioPage->ioImgs == NULL) {
            /* Add the image to the page */
            ioPage->ioImgs = ioImg;
	}
        else {
	    IO_IMG *tempImg;

            /* Put it into the linked list at the end */
	    tempImg = ioPage->ioImgs;
	    while (tempImg->next != NULL) {
		tempImg = tempImg->next;
	    }
	    
	    tempImg->next = ioImg;
	}
    }

    /* Return the result */
    *pioImgHandle = ioImg->handleID;

    NORMAL_RETURN;
    
ERR_CLEANUP:

    if (ioImg) {
	IO_FREE(ioImg);
    }
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get some image stats (compressed size, token dictionary ID/size)
 *
 * LongDescription:::
 *  IO_GetImageStats() accumulates the total compressed image size, as
 *  stored in the file, and if appropriate, the associated token dictionary
 *  ID and size.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *
 * Output:::
 *  pdwCmpSize::		Total size of the image, compressed
 *  pdwDictCount::		If tokenized, the number of token dictionaries
 *  ppdwDictIDs::		If tokenized, the dictionary ID list
 *  ppdwDictSizes::		If tokenized, the sizes of the token
 *				dictionaries
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetImageStats(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwCmpSize,
	UInt32 *pdwDictCount,
	UInt32 **ppdwDictIDs,
	Int32 **ppdwDictSizes
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetImageStats(
	    ioImg->ioImgHandle,
		pdwCmpSize,
		pdwDictCount,
		ppdwDictIDs,
		ppdwDictSizes
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image information for a transparency mask.
 *
 * LongDescription:::
 *  IO_GetMaskImageInfo() returns the same information about an image as does
 *  IO_GetImageInfo() (see above), but does so for a transparency mask
 *  attached to a non-binary image.
 *
 * Input:::
 *  xParentImgHandle::		Parent image handle
 *  dwImgOffset::		Image offset (should always be 0 for now)
 *
 * Output:::
 *  pdwXOffset::		X location in page coordinates
 *  pdwYOffset::		Y location in page coordinates
 *  pdwXResolution::		X Resolution of image in dpm
 *  pdwYResolution::		Y Resolution of image in dpm
 *  pdwWidth::			Width in image's X resolution units
 *  pdwHeight::			Height in image's Y resolution units
 *  pdwBestStripHeight::	For best performance, pass strips with size
 *				and boundary that are multiples of this (strips
 *				are stored with this size)
 *  pioCompression::		Compression type for src/dst
 *  pioImgHandle::		Returned image handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetMaskImageInfo(
    IO_IMGHANDLE ioParentImgHandle,
	UInt32 dwImgOffset,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	IO_COMPRESSION *pioCompression,
	IO_IMGHANDLE *pioImgHandle
    )
{
    IO_IMG *ioImg = NULL;
    IO_IMG *ioParentImg = (IO_IMG *)ioParentImgHandle;
    IO_INSTANCE *instData;
    UInt32 newImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioParentImgHandle);

	XERR_GEN_CLEANUP(1, IO_NOSUPPORT);

    /* Search through the image list for this new one.  We must search through
     * the entire list, since only the lower level lists are maintained in
     * layer/offset order. */
    ioImg = ioParentImg->ioMaskImg;
    while (ioImg != NULL) {
	if (ioImg->ioImgHandle == newImgHandle) {
	    break;
	}
	ioImg = ioImg->next;
    }

    if (ioImg == NULL) {

        /* Create our own copy of the container */
        instData = IO_IMGINST(ioParentImg);
        IO_CHECKMALLOC(ioImg = (IO_IMG *)IO_MALLOC(sizeof(*ioImg)));
        memset(ioImg, 0, sizeof(*ioImg));
        ioImg->handleID = (UInt32)ioImg;	/* Self identity */
	ioImg->handleType = IO_HANDLE_TYPE;
    
        /* Link back to it's parent */
        ioImg->ioParentImg = ioParentImg;

        /* Link back to the page */
        ioImg->ioPage = ioParentImg->ioPage;

        /* Identify the lower level handle */
        ioImg->ioImgHandle = newImgHandle;

        if (ioParentImg->ioMaskImg == NULL) {
            /* Add the image to the parent */
            ioParentImg->ioMaskImg = ioImg;
	}
        else {
	    IO_IMG *tempImg;

            /* Put it into the linked list at the end */
	    tempImg = ioParentImg->ioMaskImg;
	    while (tempImg->next != NULL) {
		tempImg = tempImg->next;
	    }
	    
	    tempImg->next = ioImg;
	}
    }

    /* Return the result */
    *pioImgHandle = ioImg->handleID;

    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (ioImg) {
	wp_IoImgFree(&ioImg, 0);
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image information for a rendering hints image.
 *
 * LongDescription:::
 *  IO_GetHintsImageInfo() returns the same information about an image as does
 *  IO_GetImageInfo() (see above), but does so for a rendering hints image.
 *
 * Input:::
 *  xParentImgHandle::		Parent image handle
 *  dwImgOffset::		Image offset (should always be 0 for now)
 *
 * Output:::
 *  pdwXOffset::		X location in page coordinates
 *  pdwYOffset::		Y location in page coordinates
 *  pdwXResolution::		X Resolution of image in dpm
 *  pdwYResolution::		Y Resolution of image in dpm
 *  pdwWidth::			Width in image's X resolution units
 *  pdwHeight::			Height in image's Y resolution units
 *  pdwBestStripHeight::	For best performance, pass strips with size
 *				and boundary that are multiples of this (strips
 *				are stored with this size)
 *  pioCompression::		Compression type for src/dst
 *  pioImgHandle::		Returned image handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetHintsImageInfo(
    IO_IMGHANDLE ioParentImgHandle,
	UInt32 dwImgOffset,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	IO_COMPRESSION *pioCompression,
	IO_IMGHANDLE *pioImgHandle
    )
{
    IO_IMG *ioImg = NULL;
    IO_IMG *ioParentImg = (IO_IMG *)ioParentImgHandle;
    IO_INSTANCE *instData;
    UInt32 newImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioParentImgHandle);

	XERR_PASS_CLEANUP(TFX_GetHintsImageInfo(
	    ioParentImg->ioImgHandle,
		dwImgOffset,
		pdwXOffset,
		pdwYOffset,
		pdwXResolution,
		pdwYResolution,
		pdwWidth,
		pdwHeight,
		pdwBestStripHeight,
		pioCompression,
		&newImgHandle
	    ));


    /* Search through the image list for this new one.  We must search through
     * the entire list, since only the lower level lists are maintained in
     * layer/offset order. */
    ioImg = ioParentImg->ioHintsImgs;
    while (ioImg != NULL) {
	if (ioImg->ioImgHandle == newImgHandle) {
	    break;
	}
	ioImg = ioImg->next;
    }

    if (ioImg == NULL) {

        /* Create our own copy of the container */
        instData = IO_IMGINST(ioParentImg);
        IO_CHECKMALLOC(ioImg = (IO_IMG *)IO_MALLOC(sizeof(*ioImg)));
        memset(ioImg, 0, sizeof(*ioImg));
        ioImg->handleID = (UInt32)ioImg;	/* Self identity */
	ioImg->handleType = IO_HANDLE_TYPE;
    
        /* Link back to it's parent */
        ioImg->ioParentImg = ioParentImg;

        /* Link back to the page */
        ioImg->ioPage = ioParentImg->ioPage;

        /* Identify the lower level handle */
        ioImg->ioImgHandle = newImgHandle;

        if (ioParentImg->ioHintsImgs == NULL) {
            /* Add the image to the parent */
            ioParentImg->ioHintsImgs = ioImg;
	}
        else {
	    IO_IMG *tempImg;

            /* Put it into the linked list at the end */
	    tempImg = ioParentImg->ioHintsImgs;
	    while (tempImg->next != NULL) {
		tempImg = tempImg->next;
	    }
	    
	    tempImg->next = ioImg;
	}
    }

    /* Return the result */
    *pioImgHandle = ioImg->handleID;

    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (ioImg) {
	wp_IoImgFree(&ioImg, 0);
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read enough information to determine the profile of the TIFF-FX file
 *  (profile F, C, M, T, ...)
 *
 * LongDescription:::
 *  TFX_GetDocProfile() looks through the document content and makes a
 *  determination as to which of the TIFF-FX profiles it adheres to, then
 *  returns that as a file type.
 *
 * Input:::
 *  Get the image information for a transparency mask.
 *
 * LongDescription:::
 *  ioDocHandle::		Input document handle
 *
 * Output:::
 *  pdwProfile::		Document (TIFF-FX) profile
 *				(see iomrc.h DOCTYPE_...)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetDocProfile(
	IO_DOCHANDLE ioDocHandle,
	DOCTYPE *pdwProfile
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_GetDocProfile(
	    ioDoc->ioDocHandle,
		pdwProfile
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Specify the format of, and prepare to read image data.
 *
 * LongDescription:::
 *  IO_ImageReadStart() prepares to read image data, and is used to specify
 *  the format of the data to be returned. These values are only stored for
 *  now, but used during later calls to IO_ImageReadStrip().
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *  dwFlags::			See IO_IMGFLAGS_...
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ImageReadStart(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwFlags
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_ImageReadStart(
	    ioImg->ioImgHandle,
		dwFlags
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the number of bytes of the compressed strip
 *
 * LongDescription:::
 *  IO_GetCompressedStripSize() retrieves the size of the strip in compressed
 *  bytes.  It includes the JPEGTables, if appropriate, that would get
 *  automatically added in during TFX_CompressedReadStrip().
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *  dwImageRowOffset::		The row offset in the image proper
 *
 * Output:::
 *  pdwBytes::			The number of bytes for the compressed strip
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetCompressedStripSize(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwImageRowOffset,
	UInt32 *pdwBytes
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetCompressedStripSize(
	    ioImg->ioImgHandle,
		dwImageRowOffset,
		pdwBytes
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read an arbitrary strip in compressed form
 *
 * LongDescription:::
 *  IO_CompressedReadStrip() returns a strip of compressed data from a
 *  stored strip.  The strip size must match that stored in the file, as
 *  indicated by the dwBestStripHeight parameter passed back from
 *  IO_GetImageInfo().
 *  <BR>
 *  The JPEGTables, if appropriate, would automatically be inserted into the
 *  compressed data stream.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *  dwImgRowOffset::		Source row offset of the image proper
 *  dwStripHeight::		Number of rows to fill
 *
 * Output:::
 *  ppCmpBuffer::		Pointer to pointer to preallocated
 *				compressed data buffer (may be updated)
 *  pdwCmpBufsize::		Pointer to the number of bytes in buffer
 *  pdwBytesFilled::		Number of bytes successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data from this
 *  buffer as needed, while reading into the intermediate buffer as needed.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_CompressedReadStrip(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 **ppCmpBuffer,
	UInt32 *pdwCmpBufsize,
	UInt32 *pdwBytesFilled
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_CompressedReadStrip(
	    ioImg->ioImgHandle,
		dwImgRowOffset,
		dwStripHeight,
		ppCmpBuffer,
		pdwCmpBufsize,
		pdwBytesFilled
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}

#ifndef NOIPCORE
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read an arbitrary strip of image data.
 *
 * LongDescription:::
 *  IO_ImageReadStrip() returns a strip of image data from anywhere within
 *  the image. Appropriate buffering may be necessary to read stored strips
 *  of data required to fill the requested rows. Each stored strip is
 *  decompressed entirely before the image data is buffered out.
 *  <BR>
 *  Image data is always top-to-bottom, and left-to-right. If the client
 *  does not choose to read the entire image in as a single strip, it is
 *  advisable to use a strip height that is a multiple of "dwBestStripHeight",
 *  such that not intermediate buffering is required for partial stored strips.
 *  <BR>
 *  This routine is really a wrapper on top of w_IormrcStrip().
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *  dwImgRowOffset::		Source row offset of the image proper
 *  dwStripHeight::		Number of rows to fill
 *
 * Output:::
 *  ppImagePlanes::		Pointer to 1-3 image planes (channels)
 *  dwBytesPerRow::		Number of bytes per row of image data in each
 *				plane
 *  dwBufferRowOffset::		Destination row offset in the image planes
 *  pdwRowsFilled::		Number of rows successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data from this
 *  buffer as needed, while reading into the intermediate buffer as needed.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ImageReadStrip(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 **ppImagePlanes,
	UInt32 dwBytesPerRow,
	UInt32 dwBufferRowOffset,
	UInt32 *pdwRowsFilled
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_ImageReadStrip(
	    ioImg->ioImgHandle,
		dwImgRowOffset,
		dwStripHeight,
		ppImagePlanes,
		dwBytesPerRow,
		dwBufferRowOffset,
		pdwRowsFilled
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}
#endif


/******************************************************************************
 * OneLineDescription:::
 *  Interpret the Symbol Image
 *
 * LongDescription:::
 *  IO_GetDictListFromPosBlock() reads the header placed before the JBIG2
 *  position block. The dictionary count, dictionary ID(s), pointer to the
 *  JBIG2 position block, and JBIG2 position block byte count are returned.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle
 *  pCmpData::			Symbol compressed data stream
 *  dwCmpBytes::		Number of bytes in the data stream
 *
 * Output:::
 *  ppdwDictIDList::		Dictionary identifier list, TO BE FREED BY
 *				CLIENT
 *  pdwDictCount::		Number of dictionaries associated with the
 *				position block
 *  ppPositionBlock::		Pointer to the embedded JBIG2 position block
 *				(byte count of the header is implied by the
 *				delta from pCmpData)
 *  pdwPositionBytes::		Number of bytes for the embedded position
 *				block (note, with position block extensions,
 *				the byte count may not add up to the token
 *				stream byte count)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data from this
 *  buffer as needed, while reading into the intermediate buffer as needed.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetDictListFromPosBlock(
    IO_IMGHANDLE ioImgHandle,
	UInt8 *pCmpData,
	UInt32 dwCmpBytes,
	UInt16 *pdwDictCount,
	UInt32 **ppdwDictIDs,
	UInt8 **ppPositionBlock,
	UInt32 *pdwPositionBytes
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetDictListFromPosBlock(
	    ioImg->ioImgHandle,
		pCmpData,
		dwCmpBytes,
		pdwDictCount,
		ppdwDictIDs,
		ppPositionBlock,
		pdwPositionBytes
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the characteristics of a stripe
 *
 * LongDescription:::
 *  IO_GetStripeInfo() retrieves the information about the stripe that contains
 *  a given row offset.
 *  <BR>
 *  Note: all dimensions are in page units.
 *
 * Input:::
 *  ioPageHandle::		Page handle
 *  dwRow::			Row offset that the stripe contains
 *
 * Output:::
 *  pdwStripeOffset::		The offset of the stripe
 *  pdwStripeRow::		The offset of the row where the stripe starts
 *  pdwStripeHeight::		The number of scanlines in the stripe
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data from this
 *  buffer as needed, while reading into the intermediate buffer as needed.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetStripeInfo(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 dwRow,
	UInt32 *pdwStripeOffset,
	UInt32 *pdwStripeRow,
	UInt32 *pdwStripeHeight
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_WARNOK_CLEANUP(TFX_GetStripeInfo(
	    ioPage->ioPageHandle,
		dwRow,
		pdwStripeOffset,
		pdwStripeRow,
		pdwStripeHeight
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a token dictionary
 *
 * LongDescription:::
 *  IO_GetSymbolDict() reads the symbol dictionary into memory.
 *  <BR>
 *  If the dictionary contains extensions, then the number of bytes in the
 *  entire block will be larger than that of the back-end dictionary.
 *
 * Input:::
 *  ioDocHandle::		Document handle
 *  dictID::			Dictionary identifier
 *
 * Output:::
 *  ppDictBlock::		Pointer to dictionary memory returned
 *				(client is responsible for free-ing it)
 *  pdwDictBytes::		Number of bytes in back-end dictionary
 *  pdwDictBlockBytes::		Number of bytes in whole block (may include
 *				extensions)
 *  pwLastPageUsed::		Pointer to the last page used (ordinal: 1, 2,
 *				...), 0 means unknown
 *  pdwDecoderMemory::		Pointer to the decoder memory required to hold
 *				the dictionary, 0 means unknown
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data from this
 *  buffer as needed, while reading into the intermediate buffer as needed.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetSymbolDict(
    IO_DOCHANDLE ioDocHandle,
	UInt32 dictID,
	UInt8 **pDictBlock,
	UInt32 *pdwDictBytes,
	UInt32 *pdwDictBlockBytes,
	UInt16 *pwLastPageUsed,
	UInt32 *pdwDecoderMemory
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_GetSymbolDict(
	    ioDoc->ioDocHandle,
		dictID,
		pDictBlock,
		pdwDictBytes,
		pdwDictBlockBytes,
		pwLastPageUsed,
		pdwDecoderMemory
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do necessary cleanup for the image read sequence.
 *
 * LongDescription:::
 *  IO_ImageReadFinish() notifies the IO library that the client is
 *  finished reading an image. Every call to IO_ImageReadStart must be
 *  eventually followed by a call to this function.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ImageReadFinish(
    IO_IMGHANDLE ioImgHandle
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_ImageReadFinish(
	    ioImg->ioImgHandle
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image description for the image region of a binary "even" layer
 *
 * LongDescription:::
 *  IO_GetLayerInfo() gets the description of a single image that could
 *  represent the entire layer of a binary even-numbered layer: position
 *  relative to the Primary image, resolution, and size.
 *  <BR>
 *  Returned is an image handle to be used to reference this image in later
 *  calls to IO_ImageReadLayer(). A call to IO_CloseDocument() renders this
 *  image handle invalid.
 *
 * Input:::
 *  ioPageHandle::		Page handle
 *  dwImgLayer::		Image layer (1, 2, ...)
 *
 * Output:::
 *				or IO_COMP_UNKNOWN if not just one
 *  pdwXOffset::		Pointer to the X location in page coordinates
 *  pdwYOffset::		Pointer to the Y location in page coordinates
 *  pdwXResolution::		Pointer to the X Resolution of image in dpm
 *  pdwYResolution::		Pointer to the Y Resolution of image in dpm
 *  pdwWidth::			Pointer to the width in image's X resolution
 *				units
 *  pdwHeight::			Pointer to the height in image's Y resolution
 *				units
 *  pioCompression::		Pointer to the compression type of all images,
 *  pioColorType::		Color type (see IO_COLORTYPE
 *  pPalette::			Color palette if IO_COLORTYPE is
 *				IO_COLORTYPE_COLOR4 or IO_COLORTYPE_COLOR8.
				Palette is RGB triples (chunky): RGBRGB...
 *  pioImgHandle::		Returned image handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetLayerInfo(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 dwImgLayer,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	IO_COMPRESSION *pioCompression,	/* Compression type for src/dst */
	IO_COLORTYPE *pioColorType,	/* Color type (see IO_COLORTYPE) */
	UInt8 *pPalette,		/* Color palette if IO_COLORTYPE is:
					 * IO_COLORTYPE_COLOR4 or
					 * IO_COLORTYPE_COLOR8.
					 * Palette is RGB triples (chunky):
					 * RGBRGB... */
	IO_IMGHANDLE *pioImgHandle
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    IO_IMG *ioImg = NULL;
    IO_INSTANCE *instData;
    UInt32 newImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_WARNOK_CLEANUP(TFX_GetLayerInfo(
	    ioPage->ioPageHandle,
		dwImgLayer,
		pdwXOffset,
		pdwYOffset,
		pdwXResolution,
		pdwYResolution,
		pdwWidth,
		pdwHeight,
		pioCompression,
		pioColorType,
		pPalette,
		&newImgHandle
	    ));

    /* Search through the image list for this new one.  We must search through
     * the entire list, since only the lower level lists are maintained in
     * layer/offset order. */
    ioImg = ioPage->ioImgs;
    while (ioImg != NULL) {
	if (ioImg->ioImgHandle == newImgHandle) {
	    break;
	}
	ioImg = ioImg->next;
    }

    if (ioImg == NULL) {

        /* Create our own copy of the container */
        instData = IO_PAGEINST(ioPage);
        IO_CHECKMALLOC(ioImg = (IO_IMG *)IO_MALLOC(sizeof(*ioImg)));
        memset(ioImg, 0, sizeof(*ioImg));
        ioImg->handleID = (UInt32)ioImg;	/* Self identity */
	ioImg->handleType = IO_HANDLE_TYPE;
    
        /* Link back to it's parent */
        ioImg->ioPage = ioPage;

        /* Identify the lower level handle */
        ioImg->ioImgHandle = newImgHandle;

        if (ioPage->ioImgs == NULL) {
            /* Add the image to the page */
            ioPage->ioImgs = ioImg;
	}
        else {
	    IO_IMG *tempImg;

            /* Put it into the linked list at the end */
	    tempImg = ioPage->ioImgs;
	    while (tempImg->next != NULL) {
		tempImg = tempImg->next;
	    }
	    
	    tempImg->next = ioImg;
	}
    }

    /* Return the result */
    *pioImgHandle = ioImg->handleID;

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


#ifndef NOIPCORE
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image data used to represent one image for an even layer
 *
 * LongDescription:::
 *  IO_ImageReadLayer() returns the entire image associated with all the
 *  images belonging to a layer, as described by the image return via
 *  IO_GetLayerInfo().
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetLayerInfo()
 *  dwImageHeight::		Number of rows to fill
 *
 * Output:::
 *  ppImageBuffer::		Pointer to one ore more image plane
 *  dwBytesPerRow::		Number of bytes per row of image data
 *  pdwRowsFilled::		Number of rows successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function allows piecing together the various images separated by
 *  the tokenization and lossy dither algorithms: residual, tokenized,
 *  and lossy dithered halftones.
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ImageReadLayer(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwImageHeight,
	UInt8 **ppImageBuffer,
	UInt32 dwBytesPerRow,
	UInt32 *pdwRowsFilled
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_ImageReadLayer(
	    ioImg->ioImgHandle,
		dwImageHeight,
		ppImageBuffer,
		dwBytesPerRow,
		pdwRowsFilled
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the characteristics of an annotation
 *
 * LongDescription:::
 *  IO_GetAnnotationInfo() reads the data necessary to describe an
 *  annotation. This includes the annotation: type, position, flags, size,
 *  and number of bytes of raw data.
 *
 * Input:::
 *  ioPageHandle::		Valid page handle, as returned from
 *				IO_GetPageInfo()
 *  wAnnotOffset::		Absolute offset of this annotation on the
 *				page
 *
 * Output:::
 *  pwAnnotType::		Pointer to the annotation type
 *  pdwXOffset::		Pointer to the X offset
 *  pdwYOffset::		Pointer to the Y offset
 *  pdwWidth::			Pointer to the width
 *  pdwHeight::			Pointer to the height
 *  pdwFlags::			Pointer to the flags
 *  pdwAnnotLength::		Pointer to the number of data bytes
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetAnnotationInfo(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 wAnnotOffset,
	UInt16 *pwAnnotType,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwFlags,
	UInt32 *pdwAnnotLength
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_GetAnnotationInfo(
	    ioPage->ioPageHandle,
		wAnnotOffset,
		pwAnnotType,
		pdwXOffset,
		pdwYOffset,
		pdwWidth,
		pdwHeight,
		pdwFlags,
		pdwAnnotLength
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the characteristics of an annotation
 *
 * LongDescription:::
 *  IO_GetAnnotationData() reads the data associated with an annotation.
 *
 * Input:::
 *  ioPageHandle::		Valid page handle, as returned from
 *				IO_GetPageInfo()
 *  wAnnotOffset::		Absolute offset of this annotation on the
 *				page
 *
 * Output:::
 *  dwBytes::			Number of bytes to retrieve (shouldn't exceed
 *				the stored amount)
 *  pData::			Buffer to occupy annotation data
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetAnnotationData(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 wAnnotOffset,
	UInt32 dwBytes,
	UInt8 *pData
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_GetAnnotationData(
	    ioPage->ioPageHandle,
		wAnnotOffset,
		dwBytes,
		pData
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the AnnotationOffset characteristic for the page
 *
 * LongDescription:::
 *  IO_GetAnnotationOffsets() retrieves the AnnotationOffsets field
 *  for a page
 *
 * Input:::
 *  ioPageHandle::		Valid page handle, as returned from
 *				IO_GetPageInfo()
 *
 * Output:::
 *  pdwAnnotationOffsets::	Pointer to the AnnotationOffsets value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetAnnotationOffsets(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 *pdwAnnotationOffsets,
	Bool *pbExists
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_GetAnnotationOffsets(
	    ioPage->ioPageHandle,
		pdwAnnotationOffsets,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Author characteristic for the document
 *
 * LongDescription:::
 *  IO_GetAuthor() retrieves the Author field for a document
 *
 * Input:::
 *  ioDocHandle::		Valid document handle, as returned from
 *				IO_GetDocumentInfo()
 *
 * Output:::
 *  ppwAuthor::			Pointer to the Author array returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetAuthor(
    IO_DOCHANDLE ioDocHandle,
	const Int8 **ppAuthor,
	UInt16 *pwAuthorBytes,
	Bool *pbExists
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_GetAuthor(
	    ioDoc->ioDocHandle,
		ppAuthor,
		pwAuthorBytes,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the BitsPerSample characteristic for the image
 *
 * LongDescription:::
 *  IO_GetBitsPerSample() retrieves the BitsPerSample field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  ppwBitsPerSample::		Pointer to the BitsPerSample array returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetBitsPerSample(
    IO_IMGHANDLE ioImgHandle,
	UInt16 **ppwBitsPerSample,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetBitsPerSample(
	    ioImg->ioImgHandle,
		ppwBitsPerSample,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ChromaPositioning characteristic for the image
 *
 * LongDescription:::
 *  IO_GetChromaPositioning() retrieves the ChromaPositioning field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pwChromaPositioning::	Pointer to the ChromaPositioning value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetChromaPositioning(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwChromaPositioning,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetChromaPositioning(
	    ioImg->ioImgHandle,
		pwChromaPositioning,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ChromaSubSampling characteristic for the image
 *
 * LongDescription:::
 *  IO_GetChromaSubSampling() retrieves the ChromaSubSampling field
 *  for an image.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pwChromaSubSamplingHor::	Pointer to the horizontal ChromaSubSampling value returned
 *  pwChromaSubSamplingVer::	Pointer to the vertical ChromaSubSampling value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetChromaSubSampling(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwChromaSubSamplingHor,
	UInt16 *pwChromaSubSamplingVer,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetChromaSubSampling(
	    ioImg->ioImgHandle,
		pwChromaSubSamplingHor,
		pwChromaSubSamplingVer,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the CodingMethods characteristic for the page
 *
 * LongDescription:::
 *  IO_GetCodingMethods() retrieves the CodingMethods field
 *  for a page
 *
 * Input:::
 *  ioDocHandle::		Valid document handle, as returned from
 *				IO_GetPageInfo()
 *
 * Output:::
 *  pdwCodingMethods::		Pointer to the CodingMethods value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetCodingMethods(
    IO_PAGEHANDLE ioDocHandle,
	UInt32 *pdwCodingMethods,
	Bool *pbExists
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_GetCodingMethods(
	    ioDoc->ioDocHandle,
		pdwCodingMethods,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Compression characteristic for the image
 *
 * LongDescription:::
 *  IO_GetCompression() retrieves the Compression field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pwCompression::		Pointer to the Compression value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetCompression(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwCompression,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetCompression(
	    ioImg->ioImgHandle,
		pwCompression,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the DateTime characteristic for the page
 *
 * LongDescription:::
 *  IO_GetDateTime() retrieves the DateTime field
 *  for a page
 *
 * Input:::
 *  ioPageHandle::		Valid page handle, as returned from
 *				IO_GetPageInfo()
 *
 * Output:::
 *  ppDateTime::		Pointer to the DateTime array returned
 *  pdwDateTimeBytes::		Pointer to the DateTime length
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetDateTime(
    IO_PAGEHANDLE ioPageHandle,
	Int8 **ppDateTime,
	Bool *pbExists
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_GetDateTime(
	    ioPage->ioPageHandle,
		ppDateTime,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Decode characteristic for the image
 *
 * LongDescription:::
 *  IO_GetDecode() retrieves the Decode field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  ppdwDecode::		Pointer to the Decode array returned
 *  pdwDecodeValues::		Pointer to the Decode value count
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetDecode(
    IO_IMGHANDLE ioImgHandle,
	Int32 **ppdwDecode,
	UInt32 *pdwDecodeValues,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetDecode(
	    ioImg->ioImgHandle,
		ppdwDecode,
		pdwDecodeValues,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the DocumentName characteristic for the doc
 *
 * LongDescription:::
 *  IO_GetDocumentName() retrieves the DocumentName field
 *  for a doc
 *
 * Input:::
 *  ioDocHandle::		Valid doc handle, as returned from
 *				IO_GetDocInfo()
 *
 * Output:::
 *  ppDocumentName::		Pointer to the DocumentName array returned
 *  pdwDocumentNameBytes::	Pointer to the DocumentName length
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetDocumentName(
    IO_DOCHANDLE ioDocHandle,
	Int8 **ppDocumentName,
	UInt32 *pdwDocumentNameBytes,
	Bool *pbExists
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_GetDocumentName(
	    ioDoc->ioDocHandle,
		ppDocumentName,
		pdwDocumentNameBytes,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the FaxProfile characteristic for the page
 *
 * LongDescription:::
 *  IO_GetFaxProfile() retrieves the FaxProfile field
 *  for a page
 *
 * Input:::
 *  ioPageHandle::		Valid page handle, as returned from
 *				IO_GetPageInfo()
 *
 * Output:::
 *  pFaxProfile::		Pointer to the FaxProfile value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetFaxProfile(
    IO_PAGEHANDLE ioPageHandle,
	UInt8 *pFaxProfile,
	Bool *pbExists
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_GetFaxProfile(
	    ioPage->ioPageHandle,
		pFaxProfile,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the FillOrder characteristic for the image
 *
 * LongDescription:::
 *  IO_GetFillOrder() retrieves the FillOrder field for an image.<BR>
 *  Note that the FillOrder should only be important when passing compressed
 *  data back and forth, and will normally be important when the bit order of
 *  the data is not MSB-to-LSB.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pwFillOrder::		Pointer to the FillOrder value returned:
 *				  1 = MSB to LSB (default)
 *				  2 = LSB to MSB
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetFillOrder(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwFillOrder,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetFillOrder(
	    ioImg->ioImgHandle,
		pwFillOrder,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the HostComputer characteristic for the image
 *
 * LongDescription:::
 *  IO_GetHostComputer() retrieves the HostComputer field for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  ppdwHostComputer::		Pointer to the HostComputer array returned
 *  pdwHostComputerBytes::	Pointer to the HostComputer length
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetHostComputer(
    IO_IMGHANDLE ioImgHandle,
	Int8 **ppHostComputer,
	UInt32 *pdwHostComputerBytes,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetHostComputer(
	    ioImg->ioImgHandle,
		ppHostComputer,
		pdwHostComputerBytes,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the IFDBaseColor characteristic for the image
 *
 * LongDescription:::
 *  IO_GetIFDBaseColor() retrieves the IFDBaseColor field for an image.
 *  <BR>
 *  Note, IFDBaseColor is only used when the first extension set is being used.
 *  In this case, both the IFDBaseColor and ImageBaseColor may be used, but the
 *  ImageBaseColor field may be omitted if the color image is not interpretable
 *  by a base TIFF-FX reader (for example it has JPEGTables).
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  ppwIFDBaseColor::		Pointer to the IFDBaseColor array returned.
 *  pdwIFDBaseColorValues::	Pointer to the IFDBaseColor length:
 *				
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetIFDBaseColor(
    IO_IMGHANDLE ioImgHandle,
	Int32 **ppdwIFDBaseColor,
	UInt32 *pdwIFDBaseColorValues,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetIFDBaseColor(
	    ioImg->ioImgHandle,
		ppdwIFDBaseColor,
		pdwIFDBaseColorValues,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ImageBaseColor characteristic for the image
 *
 * LongDescription:::
 *  IO_GetImageBaseColor() retrieves the ImageBaseColor field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  ppwImageBaseColor::		Pointer to the ImageBaseColor array returned.
 *				These values are ITULAB colors
 *  pdwImageBaseColorValues::	Pointer to the ImageBaseColor length:
 *				
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetImageBaseColor(
    IO_IMGHANDLE ioImgHandle,
	UInt16 **ppwImageBaseColor,
	UInt32 *pdwImageBaseColorValues,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetImageBaseColor(
	    ioImg->ioImgHandle,
		ppwImageBaseColor,
		pdwImageBaseColorValues,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ImageDescription characteristic for the image
 *
 * LongDescription:::
 *  IO_GetImageDescription() retrieves the ImageDescription field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  ppdwImageDescription::	Pointer to the ImageDescription array returned
 *  pdwImageDescriptionBytes::	Pointer to the ImageDescription length
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetImageDescription(
    IO_IMGHANDLE ioImgHandle,
	Int8 **ppImageDescription,
	UInt32 *pdwImageDescriptionBytes,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetImageDescription(
	    ioImg->ioImgHandle,
		ppImageDescription,
		pdwImageDescriptionBytes,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the T88Options characteristic for the image
 *
 * LongDescription:::
 *  IO_GetT88Options() retrieves the T88Options field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pdwT88Options::		Pointer to the T88Options value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetT88Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwT88Options,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetT88Options(
	    ioImg->ioImgHandle,
		pdwT88Options,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ModeNumber characteristic for the page
 *
 * LongDescription:::
 *  IO_GetModeNumber() retrieves the ModeNumber field
 *  for a page
 *
 * Input:::
 *  ioPageHandle::		Valid page handle, as returned from
 *				IO_GetPageInfo()
 *
 * Output:::
 *  pModeNumber::		Pointer to the ModeNumber value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetModeNumber(
    IO_PAGEHANDLE ioPageHandle,
	UInt8 *pModeNumber,
	Bool *pbExists
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_GetModeNumber(
	    ioPage->ioPageHandle,
		pModeNumber,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Orientation characteristic for the image
 *
 * LongDescription:::
 *  IO_GetOrientation() retrieves the Orientation field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pwOrientation::		Pointer to the Orientation value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetOrientation(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwOrientation,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetOrientation(
	    ioImg->ioImgHandle,
		pwOrientation,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the PageNumber characteristic for the page
 *
 * LongDescription:::
 *  IO_GetPageNumber() retrieves the PageNumber field
 *  for a page
 *
 * Input:::
 *  ioPageHandle::		Valid page handle, as returned from
 *				IO_GetPageInfo()
 *
 * Output:::
 *  pwPageNumber::		Pointer to the PageNumber value returned
 *  pwPageCount::		Pointer to the PageCount value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetPageNumber(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 *pwPageNumber,
	UInt16 *pwPageCount,
	Bool *pbExists
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_GetPageNumber(
	    ioPage->ioPageHandle,
		pwPageNumber,
		pwPageCount,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the PhotometricInterpretation characteristic for the image
 *
 * LongDescription:::
 *  IO_GetPhotometricInterpretation() retrieves the PhotometricInterpretation field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pwPhotometricInterpretation::  Pointer to the PhotometricInterpretation
 *				value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetPhotometricInterpretation(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwPhotometricInterpretation,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetPhotometricInterpretation(
	    ioImg->ioImgHandle,
		pwPhotometricInterpretation,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the PlanarConfiguration characteristic for the image
 *
 * LongDescription:::
 *  IO_GetPlanarConfiguration() retrieves the PlanarConfiguration field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pwPlanarConfiguration::	Pointer to the PlanarConfiguration value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetPlanarConfiguration(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwPlanarConfiguration,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetPlanarConfiguration(
	    ioImg->ioImgHandle,
		pwPlanarConfiguration,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ProfileType characteristic for the page
 *
 * LongDescription:::
 *  IO_GetProfileType() retrieves the ProfileType field
 *  for a page
 *
 * Input:::
 *  ioPageHandle::		Valid page handle, as returned from
 *				IO_GetPageInfo()
 *
 * Output:::
 *  pdwProfileType::		Pointer to the ProfileType value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetProfileType(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 *pdwProfileType,
	Bool *pbExists
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_GetProfileType(
	    ioPage->ioPageHandle,
		pdwProfileType,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ReferencePage characteristic for the image
 *
 * LongDescription:::
 *  IO_GetReferencePage() retrieves the ReferencePage field
 *  for an image
 *
 * Input:::
 *  ioDocHandle::		Valid document handle, as returned from
 *				IO_GetDocInfo()
 *
 * Output:::
 *  pdwReferencePage::		Pointer to the ReferencePage value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetReferencePage(
	IO_DOCHANDLE ioDocHandle,
	Int32 *pdwReferencePageOffset,
	Bool *pbExists
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_GetReferencePage(
	    ioDoc->ioDocHandle,
	    (UInt32 *)pdwReferencePageOffset,
	    pbExists));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the RenderingHints characteristic for the image
 *
 * LongDescription:::
 *  IO_GetRenderingHints() retrieves the RenderingHints field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  ppdwRenderingHints::	Pointer to the RenderingHints array returned
 *  pdwRenderingHintsBytes::	Pointer to the RenderingHints length
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetRenderingHints(
    IO_IMGHANDLE ioImgHandle,
	Int8 **ppRenderingHints,
	UInt32 *pdwRenderingHintsBytes,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetRenderingHints(
	    ioImg->ioImgHandle,
		ppRenderingHints,
		pdwRenderingHintsBytes,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the count of RenderingHints image belonging to an image
 *
 * LongDescription:::
 *  IO_GetRenderingHintsImageCount() checks to see if a RenderingHints
 *  image exists, and returns the number of images found.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pdwCount::			Pointer to the count of images
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetRenderingHintsImageCount(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwCount
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetRenderingHintsImageCount(
	    ioImg->ioImgHandle,
		pdwCount
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ResolutionUnit characteristic for the image
 *
 * LongDescription:::
 *  IO_GetResolutionUnit() retrieves the ResolutionUnit field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pwResolutionUnit::		Pointer to the ResolutionUnit value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetResolutionUnit(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwResolutionUnit,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetResolutionUnit(
	    ioImg->ioImgHandle,
		pwResolutionUnit,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the RowsPerStrip characteristic for the image
 *
 * LongDescription:::
 *  IO_GetRowsPerStrip() retrieves the RowsPerStrip field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pdwRowsPerStrip::		Pointer to the RowsPerStrip value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetRowsPerStrip(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwRowsPerStrip,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetRowsPerStrip(
	    ioImg->ioImgHandle,
		pdwRowsPerStrip,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the SamplesPerPixel characteristic for the image
 *
 * LongDescription:::
 *  IO_GetSamplesPerPixel() retrieves the SamplesPerPixel field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pdwSamplesPerPixel::	Pointer to the SamplesPerPixel value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetSamplesPerPixel(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwSamplesPerPixel,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetSamplesPerPixel(
	    ioImg->ioImgHandle,
		pdwSamplesPerPixel,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Software characteristic for the image
 *
 * LongDescription:::
 *  IO_GetSoftware() retrieves the Software field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  ppSoftware::		Pointer to the Software array returned
 *  pdwSoftwareBytes::		Pointer to the Software bytes
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetSoftware(
    IO_IMGHANDLE ioImgHandle,
	Int8 **ppSoftware,
	UInt32 *pdwSoftwareBytes,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetSoftware(
	    ioImg->ioImgHandle,
		ppSoftware,
		pdwSoftwareBytes,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the TargetPrinter characteristic for the page
 *
 * LongDescription:::
 *  IO_GetTargetPrinter() retrieves the TargetPrinter field
 *  for a page
 *
 * Input:::
 *  ioPageHandle::		Valid page handle, as returned from
 *				IO_GetPageInfo()
 *
 * Output:::
 *  ppTargetPrinter::		Pointer to the TargetPrinter array returned
 *  pdwTargetPrinterBytes::	Pointer to the TargetPrinter bytes
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetTargetPrinter(
    IO_PAGEHANDLE ioPageHandle,
	Int8 **ppTargetPrinter,
	UInt32 *pdwTargetPrinterBytes,
	Bool *pbExists
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_GetTargetPrinter(
	    ioPage->ioPageHandle,
		ppTargetPrinter,
		pdwTargetPrinterBytes,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Thresholding characteristic for the image
 *
 * LongDescription:::
 *  IO_GetThresholding() retrieves the Thresholding field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pwThresholding::		Pointer to the Thresholding value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetThresholding(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwThresholding,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetThresholding(
	    ioImg->ioImgHandle,
		pwThresholding,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the T4Options characteristic for the image
 *
 * LongDescription:::
 *  IO_GetT4Options() retrieves the T4Options field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pdwT4Options::		Pointer to the T4Options value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetT4Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwT4Options,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetT4Options(
	    ioImg->ioImgHandle,
		pdwT4Options,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the T6Options characteristic for the image
 *
 * LongDescription:::
 *  IO_GetT6Options() retrieves the T6Options field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pdwT6Options::		Pointer to the T6Options value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetT6Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwT6Options,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetT6Options(
	    ioImg->ioImgHandle,
		pdwT6Options,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the VersionYear characteristic for the image
 *
 * LongDescription:::
 *  IO_GetVersionYear() retrieves the VersionYear field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  ppVersionYear::		Pointer to the VersionYear array returned
 *  pdwVersionYearBytes::	Pointer to the VersionYear bytes
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetVersionYear(
    IO_IMGHANDLE ioImgHandle,
	UInt8 **ppVersionYear,
	UInt32 *pdwVersionYearBytes,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetVersionYear(
	    ioImg->ioImgHandle,
		ppVersionYear,
		pdwVersionYearBytes,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the YcrCbCoefficients characteristic for the image
 *
 * LongDescription:::
 *  IO_GetYcrCbCoefficients() retrieves the YcrCbCoefficients field
 *  for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  ppdwYcrCbCoefficients::	Pointer to the YcrCbCoefficients array returned
 *  pdwYcrCbCoefficientsValues:: Pointer to the YcrCbCoefficients value count
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetYCbCrCoefficients(
    IO_IMGHANDLE ioImgHandle,
	UInt32 **ppdwYCbCrCoefficients,
	UInt32 *pdwYCbCrCoefficientsValues,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetYCbCrCoefficients(
	    ioImg->ioImgHandle,
		ppdwYCbCrCoefficients,
		pdwYCbCrCoefficientsValues,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Predictor tag value for the image
 *
 * LongDescription:::
 *  IO_GetPredictor() retrieves the Predictor field for an image
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_GetImageInfo()
 *
 * Output:::
 *  pwPredictor::		Pointer to the Predictor value returned
 *  pbExists::			Pointer to exists boolean: if the field doesn't
 *				exist, then *pbExists will be 0.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_GetPredictor(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwPredictor,
	Bool *pbExists
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_GetPredictor(
	    ioImg->ioImgHandle,
		pwPredictor,
		pbExists
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do the tasks necessary to complete communication to/from a document.
 *
 * LongDescription:::
 *  IO_CloseDocument() will, if necessary, complete the processes associated
 *  with the storage of a partially written document, including completion of
 *  the Tokenization process (see IO_WriteTokenDictionary() and
 *  IO_WriteTokenImage()). All internal data associated with an open
 *  document is freed.
 *
 * Input:::
 *  ioDocHandle::		Active document handle
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_CloseDocument(
    IO_DOCHANDLE ioDocHandle
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_CloseDocument(
	    ioDoc->ioDocHandle
	    ));

    /* Free the memory associated with the IO "glue" layer */
    XERR_PASS_CLEANUP(wp_IoDocFree(&ioDoc));

    NORMAL_RETURN;

ERR_CLEANUP:

    wp_IoDocFree(&ioDoc);

    ERR_RETURN;
}



/*****************************************************************************
 * The following routines convert a single value from RGB->LAB and LAB->RGB:
 ****************************************************************************/

#define f(x)	((x) <= 0.008856f? 7.787f * (x) + 16.0f/116.0f: cuberoot(x, 0.00001f))
#define f_inv(x)	((x) < 0.20689303f? (((x)-16.0f/116.0f)/7.787f): ((x) * (x) * (x)))
#define EIGHTBIT(x)	(unsigned char)((x) < 0.0? 0: (x) > 255.0f? 255: ((x)))
#define SIGNEDEIGHTBIT(x)	(char)((x) < -128.0? -128: (x) > 127.0f? 127: ((x)))


/* Fast computation of the cube root, computed until accurate within an error
 * tolerance.
 * Note that for normal computation, you'd use error = 0.00001; however,
 * for L*a*b* conversion, you could set error = 0.001, but due to the few
 * number of computations, I've picked 0.00001.
 */
float CDECL
cuberoot(float x, float error)
{
    float root = 0.6f, cube, diff;
    int negFlag = 0, count = 0;

    if (x < 0.0f) {
	negFlag = 1;
	x = -x;
    } else if (x == 0.0f) {
	return 0.0f;
    }
    do {
	root = (x / (root * root) + root) / 2.0f;
	count++;
	if (count > 25)
	    break;
	cube = root * root * root;
	if (cube > x)
	    diff = cube - x;
	else
	    diff = x - cube;
    } while (diff >= error);
    if (negFlag) {
	root = -root;
    }
    return root;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert from sRGB to ITULAB
 *
 * LongDescription:::
 *  IO_sRGB2ITULAB() converts 8-bit sRGB values to 8-bit ITULAB values (using
 *  16-bit containers, which are used by TIFF-FX).  If appropriate, a
 *  different gamut may be applied -- the equivalent of the Decode TIFF-FX
 *  field.
 *
 * Input:::
 *  sR8::			eight-bit sRGB Red value
 *  sG8::			eight-bit sRGB Green value
 *  sB8::			eight-bit sRGB Blue value
 *  pdwGamut::			Gamut to use (if NULL, default gamut)
 *				These are the same values used by
 *				IO_GetDecode() and IO_SetDecode(), namely
 *				signed long numerator/denominator pairs:
 *				1 pairs each for Lmin, Lmax, amin, amax,
 *				bmin, and bmax.
 *
 * Output:::
 *  pwL::			8-bit L* value returned (stored in a UInt16)
 *  pwa::			8-bit a* value returned (stored in a UInt16)
 *  pwb::			8-bit b* value returned (stored in a UInt16)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_sRGB2ITULAB(
	      UInt8 sR8,
	      UInt8 sG8,
	      UInt8 sB8,
	      UInt16 *pwL,
	      UInt16 *pwa,
	      UInt16 *pwb,
	      Int32 *pdwGamut
)
{
    float	X, Y, Z, L, a, b;
    float	fy;
    float	Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float	sR, sG, sB;
    float	sRp, sGp, sBp;
    float	Lmin = 0.0, Lmax = 100.0f;
    float	amin = -21760.0f/255.0f, amax = 21590.0f/255.0f;
    float	bmin = -19200.0f/255.0f, bmax = 31800.0f/255.0f;
    PROLOGUE;

    /* If a gamut has been passed in, use it.  If gamut is nonsense,
     * ignore it, or portions of it. */
    while (pdwGamut != NULL) {
	if (pdwGamut[1] == 0 || pdwGamut[3] == 0) break;
	Lmin = (float)pdwGamut[0] / (float)pdwGamut[1];
	Lmax = (float)pdwGamut[2] / (float)pdwGamut[3];

	if (pdwGamut[5] == 0 || pdwGamut[7] == 0) break;
	amin = (float)pdwGamut[4] / (float)pdwGamut[5];
	amax = (float)pdwGamut[6] / (float)pdwGamut[7];

	if (pdwGamut[9] == 0 || pdwGamut[11] == 0) break;
	bmin = (float)pdwGamut[8] / (float)pdwGamut[9];
	bmax = (float)pdwGamut[10] / (float)pdwGamut[11];

	break;
    }

    /* Normalize the sRGB values */
    sRp = (float)sR8 / 255.0f;
    sGp = (float)sG8 / 255.0f;
    sBp = (float)sB8 / 255.0f;

    /* Convert from sRGB to linear-RGB */
    if (sRp <= 0.04045) {
	sR = sRp / 12.92f;
    } else {
	sR = (float)pow((double)((sRp + 0.055) / 1.055), 2.4);
    }
    if (sGp <= 0.04045) {
	sG = sGp / 12.92f;
    } else {
	sG = (float)pow((double)((sGp + 0.055) / 1.055), 2.4);
    }
    if (sBp <= 0.04045) {
	sB = sBp / 12.92f;
    } else {
	sB = (float)pow((double)((sBp + 0.055) / 1.055), 2.4);
    }

    /* ------------------ Convert from linear-RGB to XYZ ------------------ */
    X =  0.4124f * sR + 0.3576f * sG + 0.1805f * sB;
    Y =  0.2126f * sR + 0.7152f * sG + 0.0722f * sB;
    Z =  0.0193f * sR + 0.1192f * sG + 0.9505f * sB;

    /* Precomputed value */
    fy = f(Y / Yn);

    /* Compute L* */
    L = 116.0f * fy - 16.0f;

    /* Converted to 8-bit range, adjusted for gamut */
    L = (L - Lmin) * 256.0f / (Lmax - Lmin);

    /* Compute a* */
    a = 500.0f * (f(X / Xn) - fy);

    /* Convert to 8-bit range, adjusted for gamut */
    a = (a - amin) * 256.0f / (amax - amin);

    /* Compute b* */
    b = 200.0f * (fy - f(Z / Zn));

    /* Convert to 8-bit range, adjusted for gamut */
    b = (b - bmin) * 256.0f / (bmax - bmin);

    /* Return the 8-bit result */
    *pwL = EIGHTBIT(L);
    *pwa = EIGHTBIT(a);
    *pwb = EIGHTBIT(b);

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert from ITULAB to sRGB
 *
 * LongDescription:::
 *  IO_ITULAB2sRGB() converts 8-bit ITULAB values (using 16-bit containers,
 *  which are used by TIFF-FX) to 8-bit sRGB values. If appropriate, a
 *  different gamut may be applied -- the equivalent of the Decode TIFF-FX
 *  field.
 *
 * Input:::
 *  L::				8-bit L* value (stored in a UInt16)
 *  a::				8-bit a* value (stored in a UInt16)
 *  b::				8-bit b* value (stored in a UInt16)
 *  pdwGamut::			Gamut to use (if NULL, default gamut)
 *				These are the same values used by
 *				IO_GetDecode() and IO_SetDecode().
 *
 * Output:::
 *  psR8::			8-bit red value returned
 *  psG8::			8-bit green value returned
 *  psB8::			8-bit blue value returned
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ITULAB2sRGB(
	      UInt16 L,
	      UInt16 a,
	      UInt16 b,
	      UInt8 *psR8,
	      UInt8 *psG8,
	      UInt8 *psB8,
	      Int32 *pdwGamut
)
{
    float	c, t;
    float	X, Y, Z;
    float	Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float	sR, sG, sB;
    float	sRp, sGp, sBp;
    float	Lmin = 0.0f, Lmax = 100.0f;
    float	amin = -21760.0f/255.0f, amax = 21590.0f/255.0f;
    float	bmin = -19200.0f/255.0f, bmax = 31800.0f/255.0f;
    PROLOGUE;
 
    while (pdwGamut != NULL) {
	if (pdwGamut[1] == 0 || pdwGamut[3] == 0) break;
	Lmin = (float)pdwGamut[0] / (float)pdwGamut[1];
	Lmax = (float)pdwGamut[2] / (float)pdwGamut[3];

	if (pdwGamut[5] == 0 || pdwGamut[7] == 0) break;
	amin = (float)pdwGamut[4] / (float)pdwGamut[5];
	amax = (float)pdwGamut[6] / (float)pdwGamut[7];

	if (pdwGamut[9] == 0 || pdwGamut[11] == 0) break;
	bmin = (float)pdwGamut[8] / (float)pdwGamut[9];
	bmax = (float)pdwGamut[10] / (float)pdwGamut[11];

	break;
    }

    /* Convert from ITULAB -> XYZ: */
    c = ((float)L / 255.0f * (Lmax - Lmin) + Lmin + 16.0f) / 116.0f;
    t = ((float)a / 255.0f * (amax - amin) + amin) / 500.0f + c;
    X = Xn * f_inv(t);
    Y = Yn * f_inv(c);
    t = c - (((float)b / 255.0f) * (bmax - bmin) + bmin) / 200.0f;
    Z = Zn * f_inv(t);

    /* Convert from XYZ -> linear-RGB: */
    sR =  3.2410f * X - 1.5374f * Y - 0.4986f * Z;
    sG = -0.9692f * X + 1.8760f * Y + 0.0416f * Z;
    sB =  0.0556f * X - 0.2040f * Y + 1.0570f * Z;

    /* Convert from linear-RGB to sRGB' */
    if (sR <= 0.00313) {
	sRp = sR * 12.92f;
    } else {
	sRp = (float)(pow((double)sR, (1.0 / 2.4)) * 1.055 - 0.055);
    }
    if (sG <= 0.00313) {
	sGp = sG * 12.92f;
    } else {
	sGp = (float)(pow((double)sG, (1.0 / 2.4)) * 1.055 - 0.055);
    }
    if (sB <= 0.00313) {
	sBp = sB * 12.92f;
    } else {
	sBp = (float)(pow((double)sB, (1.0 / 2.4)) * 1.055 - 0.055);
    }

    /* Convert from sRGB' to 8-bit sRGB */
    sRp *= 256.0;
    sGp *= 256.0;
    sBp *= 256.0;
    *psR8 = EIGHTBIT(sRp);
    *psG8 = EIGHTBIT(sGp);
    *psB8 = EIGHTBIT(sBp);

    NORMAL_RETURN;
}
	      

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert from linear-RGB to ITULAB
 *
 * LongDescription:::
 *  IO_LinearRGB2ITULAB() converts 8-bit linear-RGB values to 8-bit ITULAB
 *  values (using 16-bit containers, which are used by TIFF-FX).
 *  If appropriate, a different gamut may be applied -- the equivalent of
 *  the Decode TIFF-FX field.
 *
 * Input:::
 *  lR8::			eight-bit linear-RGB Red value
 *  lG8::			eight-bit linear-RGB Green value
 *  lB8::			eight-bit linear-RGB Blue value
 *  pdwGamut::			Gamut to use (if NULL, default gamut)
 *				These are the same values used by
 *				IO_GetDecode() and IO_SetDecode(), namely
 *				signed long numerator/denominator pairs:
 *				1 pairs each for Lmin, Lmax, amin, amax,
 *				bmin, and bmax.
 *
 * Output:::
 *  pwL::			8-bit L* value returned (stored in a UInt16)
 *  pwa::			8-bit a* value returned (stored in a UInt16)
 *  pwb::			8-bit b* value returned (stored in a UInt16)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_LinearRGB2ITULAB(
	      UInt8 lR8,
	      UInt8 lG8,
	      UInt8 lB8,
	      UInt16 *pwL,
	      UInt16 *pwa,
	      UInt16 *pwb,
	      Int32 *pdwGamut
)
{
    float	X, Y, Z, L, a, b;
    float	fy;
    float	Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float	Rp, Gp, Bp;
    float	Lmin = 0.0, Lmax = 100.0f;
    float	amin = -21760.0f/255.0f, amax = 21590.0f/255.0f;
    float	bmin = -19200.0f/255.0f, bmax = 31800.0f/255.0f;
    PROLOGUE;

    /* If a gamut has been passed in, use it.  If gamut is nonsense,
     * ignore it, or portions of it. */
    while (pdwGamut != NULL) {
	if (pdwGamut[1] == 0 || pdwGamut[3] == 0) break;
	Lmin = (float)pdwGamut[0] / (float)pdwGamut[1];
	Lmax = (float)pdwGamut[2] / (float)pdwGamut[3];

	if (pdwGamut[5] == 0 || pdwGamut[7] == 0) break;
	amin = (float)pdwGamut[4] / (float)pdwGamut[5];
	amax = (float)pdwGamut[6] / (float)pdwGamut[7];

	if (pdwGamut[9] == 0 || pdwGamut[11] == 0) break;
	bmin = (float)pdwGamut[8] / (float)pdwGamut[9];
	bmax = (float)pdwGamut[10] / (float)pdwGamut[11];

	break;
    }

    /* Normalize the 8-bit linear-RGB */
    Rp = (float)lR8 / 255.0f;
    Gp = (float)lG8 / 255.0f;
    Bp = (float)lB8 / 255.0f;

    /* ------------------ Convert from linear-RGB to XYZ ------------------ */
    X =  0.4124f * Rp + 0.3576f * Gp + 0.1805f * Bp;
    Y =  0.2126f * Rp + 0.7152f * Gp + 0.0722f * Bp;
    Z =  0.0193f * Rp + 0.1192f * Gp + 0.9505f * Bp;

    /* Precomputed value */
    fy = f(Y / Yn);

    /* Compute L* */
    L = 116.0f * fy - 16.0f;

    /* Converted to 8-bit range, adjusted for gamut */
    L = (L - Lmin) * 256.0f / (Lmax - Lmin);

    /* Compute a* */
    a = 500.0f * (f(X / Xn) - fy);

    /* Convert to 8-bit range, adjusted for gamut */
    a = (a - amin) * 256.0f / (amax - amin);

    /* Compute b* */
    b = 200.0f * (fy - f(Z / Zn));

    /* Convert to 8-bit range, adjusted for gamut */
    b = (b - bmin) * 256.0f / (bmax - bmin);

    /* Return the 8-bit result */
    *pwL = EIGHTBIT(L);
    *pwa = EIGHTBIT(a);
    *pwb = EIGHTBIT(b);

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert from ITULAB to linear-RGB
 *
 * LongDescription:::
 *  IO_ITULAB2LinearRGB() converts 8-bit ITULAB values (using 16-bit containers,
 *  which are used by TIFF-FX) to 8-bit linear-RGB values. If appropriate, a
 *  different gamut may be applied -- the equivalent of the Decode TIFF-FX
 *  field.
 *
 * Input:::
 *  L::				8-bit L* value (stored in a UInt16)
 *  a::				8-bit a* value (stored in a UInt16)
 *  b::				8-bit b* value (stored in a UInt16)
 *  pdwGamut::			Gamut to use (if NULL, default gamut)
 *				These are the same values used by
 *				IO_GetDecode() and IO_SetDecode().
 *
 * Output:::
 *  plR8::			8-bit red value returned
 *  plG8::			8-bit green value returned
 *  plB8::			8-bit blue value returned
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ITULAB2LinearRGB(
	      UInt16 L,
	      UInt16 a,
	      UInt16 b,
	      UInt8 *plR8,
	      UInt8 *plG8,
	      UInt8 *plB8,
	      Int32 *pdwGamut
)
{
    float	c, t;
    float	X, Y, Z;
    float	Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float	Rp, Gp, Bp;
    float	Lmin = 0.0f, Lmax = 100.0f;
    float	amin = -21760.0f/255.0f, amax = 21590.0f/255.0f;
    float	bmin = -19200.0f/255.0f, bmax = 31800.0f/255.0f;
    PROLOGUE;
 
    /* If a gamut has been passed in, use it.  If gamut is nonsense,
     * ignore it, or portions of it. */
    while (pdwGamut != NULL) {
	if (pdwGamut[1] == 0 || pdwGamut[3] == 0) break;
	Lmin = (float)pdwGamut[0] / (float)pdwGamut[1];
	Lmax = (float)pdwGamut[2] / (float)pdwGamut[3];

	if (pdwGamut[5] == 0 || pdwGamut[7] == 0) break;
	amin = (float)pdwGamut[4] / (float)pdwGamut[5];
	amax = (float)pdwGamut[6] / (float)pdwGamut[7];

	if (pdwGamut[9] == 0 || pdwGamut[11] == 0) break;
	bmin = (float)pdwGamut[8] / (float)pdwGamut[9];
	bmax = (float)pdwGamut[10] / (float)pdwGamut[11];

	break;
    }

    /* Convert from ITULAB -> XYZ: */
    c = ((float)L / 255.0f * (Lmax - Lmin) + Lmin + 16.0f) / 116.0f;
    t = ((float)a / 255.0f * (amax - amin) + amin) / 500.0f + c;
    X = Xn * f_inv(t);
    Y = Yn * f_inv(c);
    t = c - (((float)b / 255.0f) * (bmax - bmin) + bmin) / 200.0f;
    Z = Zn * f_inv(t);

    /* Convert from XYZ -> normalized linear-RGB: */
    Rp =  3.2410f * X - 1.5374f * Y - 0.4986f * Z;
    Gp = -0.9692f * X + 1.8760f * Y + 0.0416f * Z;
    Bp =  0.0556f * X - 0.2040f * Y + 1.0570f * Z;

    /* Convert from sRGB' to 8-bit linear-RGB */
    Rp *= 256.0;
    Gp *= 256.0;
    Bp *= 256.0;
    *plR8 = EIGHTBIT(Rp);
    *plG8 = EIGHTBIT(Gp);
    *plB8 = EIGHTBIT(Bp);

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert from sRGB to CIELAB
 *
 * LongDescription:::
 *  IO_sRGB2CIELAB() converts 8-bit sRGB values to 8-bit CIELAB values (using
 *  16-bit containers, which are used by TIFF-FX).
 *
 * Input:::
 *  sR8::			eight-bit sRGB Red value
 *  sG8::			eight-bit sRGB Green value
 *  sB8::			eight-bit sRGB Blue value
 *
 * Output:::
 *  pwL::			8-bit L* value returned ([0,255] stored in a UInt16)
 *  pwa::			8-bit a* value returned ([-128,127] stored in a Int16)
 *  pwb::			8-bit b* value returned ([-128,127] stored in a Int16)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_sRGB2CIELAB(
	      UInt8 sR8,
	      UInt8 sG8,
	      UInt8 sB8,
	      UInt16 *pwL,
	      Int16 *pwa,
	      Int16 *pwb
)
{
    float	X, Y, Z, L, a, b;
    float	fy;
    float	Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float	sR, sG, sB;
    float	sRp, sGp, sBp;
    PROLOGUE;


    /* Normalize the sRGB values */
    sRp = (float)sR8 / 255.0f;
    sGp = (float)sG8 / 255.0f;
    sBp = (float)sB8 / 255.0f;

    /* Convert from sRGB to linear-RGB */
    if (sRp <= 0.04045) {
	sR = sRp / 12.92f;
    } else {
	sR = (float)pow((double)((sRp + 0.055) / 1.055), 2.4);
    }
    if (sGp <= 0.04045) {
	sG = sGp / 12.92f;
    } else {
	sG = (float)pow((double)((sGp + 0.055) / 1.055), 2.4);
    }
    if (sBp <= 0.04045) {
	sB = sBp / 12.92f;
    } else {
	sB = (float)pow((double)((sBp + 0.055) / 1.055), 2.4);
    }

    /* ------------------ Convert from linear-RGB to XYZ ------------------ */
    X =  0.4124f * sR + 0.3576f * sG + 0.1805f * sB;
    Y =  0.2126f * sR + 0.7152f * sG + 0.0722f * sB;
    Z =  0.0193f * sR + 0.1192f * sG + 0.9505f * sB;

    /* Precomputed value */
    fy = f(Y / Yn);

    /* Compute L* */
    L = 116.0f * fy - 16.0f;

    /* Converted to 8-bit range, adjusted for gamut */
    L = L * 256.0f / 100.0f;

    /* Compute a* */
    a = 500.0f * (f(X / Xn) - fy);

    /* Compute b* */
    b = 200.0f * (fy - f(Z / Zn));

    /* Return the 8-bit result */
    *pwL = EIGHTBIT(L);
    *pwa = SIGNEDEIGHTBIT(a);
    *pwb = SIGNEDEIGHTBIT(b);

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert from CIELAB to sRGB
 *
 * LongDescription:::
 *  IO_CIELAB2sRGB() converts 8-bit ITULAB values (using 16-bit containers,
 *  which are used by TIFF-FX) to 8-bit sRGB values.
 *
 * Input:::
 *  L::				8-bit L* value ([0,255] stored in a UInt16)
 *  a::				8-bit a* value ([-128,127] stored in a Int16)
 *  b::				8-bit b* value ([-128,127] stored in a Int16)
 *
 * Output:::
 *  psR8::			8-bit red value returned
 *  psG8::			8-bit green value returned
 *  psB8::			8-bit blue value returned
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_CIELAB2sRGB(
	      UInt16 L,
	      Int16 a,
	      Int16 b,
	      UInt8 *psR8,
	      UInt8 *psG8,
	      UInt8 *psB8
)
{
    float	c, t;
    float	X, Y, Z;
    float	Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float	sR, sG, sB;
    float	sRp, sGp, sBp;
    PROLOGUE;


    /* Convert from CIELAB -> XYZ: */
    c = ((float)L * 100.0f / 255.0f + 16.0f) / 116.0f;
    t = (float)a / 500.0f + c;
    X = Xn * f_inv(t);
    Y = Yn * f_inv(c);
    t = c - (float)b / 200.0f;
    Z = Zn * f_inv(t);

    /* Convert from XYZ -> linear-RGB: */
    sR =  3.2410f * X - 1.5374f * Y - 0.4986f * Z;
    sG = -0.9692f * X + 1.8760f * Y + 0.0416f * Z;
    sB =  0.0556f * X - 0.2040f * Y + 1.0570f * Z;

    /* Convert from linear-RGB to sRGB' */
    if (sR <= 0.00313) {
	sRp = sR * 12.92f;
    } else {
	sRp = (float)(pow((double)sR, (1.0 / 2.4)) * 1.055 - 0.055);
    }
    if (sG <= 0.00313) {
	sGp = sG * 12.92f;
    } else {
	sGp = (float)(pow((double)sG, (1.0 / 2.4)) * 1.055 - 0.055);
    }
    if (sB <= 0.00313) {
	sBp = sB * 12.92f;
    } else {
	sBp = (float)(pow((double)sB, (1.0 / 2.4)) * 1.055 - 0.055);
    }

    /* Convert from sRGB' to 8-bit sRGB */
    sRp *= 256.0;
    sGp *= 256.0;
    sBp *= 256.0;
    *psR8 = EIGHTBIT(sRp);
    *psG8 = EIGHTBIT(sGp);
    *psB8 = EIGHTBIT(sBp);

    NORMAL_RETURN;
}
	      

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert from linear-RGB to CIELAB
 *
 * LongDescription:::
 *  IO_LinearRGB2CIELAB() converts 8-bit linear-RGB values to 8-bit CIELAB
 *  values (using 16-bit containers, which are used by TIFF-FX).
 *
 * Input:::
 *  lR8::			eight-bit linear-RGB Red value
 *  lG8::			eight-bit linear-RGB Green value
 *  lB8::			eight-bit linear-RGB Blue value
 *
 * Output:::
 *  pwL::			8-bit L* value returned ([0,255] stored in a UInt16)
 *  pwa::			8-bit a* value returned ([-128,127] stored in a Int16)
 *  pwb::			8-bit b* value returned ([-128,127] stored in a Int16)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_LinearRGB2CIELAB(
	      UInt8 lR8,
	      UInt8 lG8,
	      UInt8 lB8,
	      UInt16 *pwL,
	      UInt16 *pwa,
	      UInt16 *pwb
)
{
    float	X, Y, Z, L, a, b;
    float	fy;
    float	Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float	Rp, Gp, Bp;
    PROLOGUE;


    /* Normalize the 8-bit linear-RGB */
    Rp = (float)lR8 / 255.0f;
    Gp = (float)lG8 / 255.0f;
    Bp = (float)lB8 / 255.0f;

    /* ------------------ Convert from linear-RGB to XYZ ------------------ */
    X =  0.4124f * Rp + 0.3576f * Gp + 0.1805f * Bp;
    Y =  0.2126f * Rp + 0.7152f * Gp + 0.0722f * Bp;
    Z =  0.0193f * Rp + 0.1192f * Gp + 0.9505f * Bp;

    /* Precomputed value */
    fy = f(Y / Yn);

    /* Compute L* */
    L = 116.0f * fy - 16.0f;

    /* Converted to 8-bit range, adjusted for gamut */
    L = L * 256.0f / 100.0f;

    /* Compute a* */
    a = 500.0f * (f(X / Xn) - fy);

    /* Compute b* */
    b = 200.0f * (fy - f(Z / Zn));

    /* Return the 8-bit result */
    *pwL = EIGHTBIT(L);
    *pwa = SIGNEDEIGHTBIT(a);
    *pwb = SIGNEDEIGHTBIT(b);

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert from CIELAB to linear-RGB
 *
 * LongDescription:::
 *  IO_CIELAB2LinearRGB() converts 8-bit CIELAB values (using 16-bit containers,
 *  which are used by TIFF-FX) to 8-bit linear-RGB values.
 *
 * Input:::
 *  L::				8-bit L* value ([0,255] stored in a UInt16)
 *  a::				8-bit a* value ([-128,127] stored in a UInt16)
 *  b::				8-bit b* value ([-128,127] stored in a UInt16)
 *
 * Output:::
 *  plR8::			8-bit red value returned
 *  plG8::			8-bit green value returned
 *  plB8::			8-bit blue value returned
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_CIELAB2LinearRGB(
	      UInt16 L,
	      Int16 a,
	      Int16 b,
	      UInt8 *plR8,
	      UInt8 *plG8,
	      UInt8 *plB8
)
{
    float	c, t;
    float	X, Y, Z;
    float	Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float	Rp, Gp, Bp;
    PROLOGUE;
 

    /* Convert from CIELAB -> XYZ: */
    c = ((float)L * 100.0f / 255.0f + 16.0f) / 116.0f;
    t = a / 500.0f + c;
    X = Xn * f_inv(t);
    Y = Yn * f_inv(c);
    t = c - b / 200.0f;
    Z = Zn * f_inv(t);

    /* Convert from XYZ -> normalized linear-RGB: */
    Rp =  3.2410f * X - 1.5374f * Y - 0.4986f * Z;
    Gp = -0.9692f * X + 1.8760f * Y + 0.0416f * Z;
    Bp =  0.0556f * X - 0.2040f * Y + 1.0570f * Z;

    /* Convert from sRGB' to 8-bit linear-RGB */
    Rp *= 256.0;
    Gp *= 256.0;
    Bp *= 256.0;
    *plR8 = EIGHTBIT(Rp);
    *plG8 = EIGHTBIT(Gp);
    *plB8 = EIGHTBIT(Bp);

    NORMAL_RETURN;
}
	      

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert from sRGB to ICCLAB
 *
 * LongDescription:::
 *  IO_sRGB2ICCLAB() converts 8-bit sRGB values to 8-bit ICCLAB values (using
 *  16-bit containers, which are used by TIFF-FX).
 *
 * Input:::
 *  sR8::			eight-bit sRGB Red value
 *  sG8::			eight-bit sRGB Green value
 *  sB8::			eight-bit sRGB Blue value
 *
 * Output:::
 *  pwL::			8-bit L* value returned ([0,255] stored in a UInt16)
 *  pwa::			8-bit a* value returned ([0,255] stored in a UInt16)
 *  pwb::			8-bit b* value returned ([0,255] stored in a UInt16)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_sRGB2ICCLAB(
	      UInt8 sR8,
	      UInt8 sG8,
	      UInt8 sB8,
	      UInt16 *pwL,
	      UInt16 *pwa,
	      UInt16 *pwb
)
{
    float	X, Y, Z, L, a, b;
    float	fy;
    float	Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float	sR, sG, sB;
    float	sRp, sGp, sBp;
    PROLOGUE;


    /* Normalize the sRGB values */
    sRp = (float)sR8 / 255.0f;
    sGp = (float)sG8 / 255.0f;
    sBp = (float)sB8 / 255.0f;

    /* Convert from sRGB to linear-RGB */
    if (sRp <= 0.04045) {
	sR = sRp / 12.92f;
    } else {
	sR = (float)pow((double)((sRp + 0.055) / 1.055), 2.4);
    }
    if (sGp <= 0.04045) {
	sG = sGp / 12.92f;
    } else {
	sG = (float)pow((double)((sGp + 0.055) / 1.055), 2.4);
    }
    if (sBp <= 0.04045) {
	sB = sBp / 12.92f;
    } else {
	sB = (float)pow((double)((sBp + 0.055) / 1.055), 2.4);
    }

    /* ------------------ Convert from linear-RGB to XYZ ------------------ */
    X =  0.4124f * sR + 0.3576f * sG + 0.1805f * sB;
    Y =  0.2126f * sR + 0.7152f * sG + 0.0722f * sB;
    Z =  0.0193f * sR + 0.1192f * sG + 0.9505f * sB;

    /* Precomputed value */
    fy = f(Y / Yn);

    /* Compute L* */
    L = 116.0f * fy - 16.0f;

    /* Converted to 8-bit range, adjusted for gamut */
    L = L * 256.0f / 100.0f;

    /* Compute a* */
    a = 500.0f * (f(X / Xn) - fy);

    /* Convert to 8-bit range, adjusted for gamut */
    a = (a + 128.0f) * 256.0f / 255.0f;

    /* Compute b* */
    b = 200.0f * (fy - f(Z / Zn));

    /* Convert to 8-bit range, adjusted for gamut */
    b = (b + 128.0f) * 256.0f / 255.0f;

    /* Return the 8-bit result */
    *pwL = EIGHTBIT(L);
    *pwa = EIGHTBIT(a);
    *pwb = EIGHTBIT(b);

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert from ICCLAB to sRGB
 *
 * LongDescription:::
 *  IO_ICCLAB2sRGB() converts 8-bit ITULAB values (using 16-bit containers,
 *  which are used by TIFF-FX) to 8-bit sRGB values.
 *
 * Input:::
 *  L::				8-bit L* value ([0,255] stored in a UInt16)
 *  a::				8-bit a* value ([0,255] stored in a UInt16)
 *  b::				8-bit b* value ([0,255] stored in a UInt16)
 *
 * Output:::
 *  psR8::			8-bit red value returned
 *  psG8::			8-bit green value returned
 *  psB8::			8-bit blue value returned
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ICCLAB2sRGB(
	      UInt16 L,
	      UInt16 a,
	      UInt16 b,
	      UInt8 *psR8,
	      UInt8 *psG8,
	      UInt8 *psB8
)
{
    float	c, t;
    float	X, Y, Z;
    float	Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float	sR, sG, sB;
    float	sRp, sGp, sBp;
    PROLOGUE;


    /* Convert from ICCLAB -> XYZ: */
    c = ((float)L * 100.0f / 255.0f + 16.0f) / 116.0f;
    t = ((float)a - 128.0f) / 500.0f + c;
    X = Xn * f_inv(t);
    Y = Yn * f_inv(c);
    t = c - ((float)b - 128.0f) / 200.0f;
    Z = Zn * f_inv(t);

    /* Convert from XYZ -> linear-RGB: */
    sR =  3.2410f * X - 1.5374f * Y - 0.4986f * Z;
    sG = -0.9692f * X + 1.8760f * Y + 0.0416f * Z;
    sB =  0.0556f * X - 0.2040f * Y + 1.0570f * Z;

    /* Convert from linear-RGB to sRGB' */
    if (sR <= 0.00313) {
	sRp = sR * 12.92f;
    } else {
	sRp = (float)(pow((double)sR, (1.0 / 2.4)) * 1.055 - 0.055);
    }
    if (sG <= 0.00313) {
	sGp = sG * 12.92f;
    } else {
	sGp = (float)(pow((double)sG, (1.0 / 2.4)) * 1.055 - 0.055);
    }
    if (sB <= 0.00313) {
	sBp = sB * 12.92f;
    } else {
	sBp = (float)(pow((double)sB, (1.0 / 2.4)) * 1.055 - 0.055);
    }

    /* Convert from sRGB' to 8-bit sRGB */
    sRp *= 256.0;
    sGp *= 256.0;
    sBp *= 256.0;
    *psR8 = EIGHTBIT(sRp);
    *psG8 = EIGHTBIT(sGp);
    *psB8 = EIGHTBIT(sBp);

    NORMAL_RETURN;
}
	      

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert from linear-RGB to ICCLAB
 *
 * LongDescription:::
 *  IO_LinearRGB2ICCLAB() converts 8-bit linear-RGB values to 8-bit ICCLAB
 *  values (using 16-bit containers, which are used by TIFF-FX).
 *
 * Input:::
 *  lR8::			eight-bit linear-RGB Red value
 *  lG8::			eight-bit linear-RGB Green value
 *  lB8::			eight-bit linear-RGB Blue value
 *
 * Output:::
 *  pwL::			8-bit L* value returned ([0,255] stored in a UInt16)
 *  pwa::			8-bit a* value returned ([0,255] stored in a UInt16)
 *  pwb::			8-bit b* value returned ([0,255] stored in a UInt16)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_LinearRGB2ICCLAB(
	      UInt8 lR8,
	      UInt8 lG8,
	      UInt8 lB8,
	      UInt16 *pwL,
	      UInt16 *pwa,
	      UInt16 *pwb
)
{
    float	X, Y, Z, L, a, b;
    float	fy;
    float	Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float	Rp, Gp, Bp;
    PROLOGUE;


    /* Normalize the 8-bit linear-RGB */
    Rp = (float)lR8 / 255.0f;
    Gp = (float)lG8 / 255.0f;
    Bp = (float)lB8 / 255.0f;

    /* ------------------ Convert from linear-RGB to XYZ ------------------ */
    X =  0.4124f * Rp + 0.3576f * Gp + 0.1805f * Bp;
    Y =  0.2126f * Rp + 0.7152f * Gp + 0.0722f * Bp;
    Z =  0.0193f * Rp + 0.1192f * Gp + 0.9505f * Bp;

    /* Precomputed value */
    fy = f(Y / Yn);

    /* Compute L* */
    L = 116.0f * fy - 16.0f;

    /* Converted to 8-bit range, adjusted for gamut */
    L = L * 256.0f / 100.0f;

    /* Compute a* */
    a = 500.0f * (f(X / Xn) - fy);

    /* Convert to 8-bit range, adjusted for gamut */
    a = (a + 128.0f) * 256.0f / 255.0f;

    /* Compute b* */
    b = 200.0f * (fy - f(Z / Zn));

    /* Convert to 8-bit range, adjusted for gamut */
    b = (b + 128.0f) * 256.0f / 255.0f;

    /* Return the 8-bit result */
    *pwL = EIGHTBIT(L);
    *pwa = SIGNEDEIGHTBIT(a);
    *pwb = SIGNEDEIGHTBIT(b);

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert from ICCLAB to linear-RGB
 *
 * LongDescription:::
 *  IO_ICCLAB2LinearRGB() converts 8-bit ICCLAB values (using 16-bit containers,
 *  which are used by TIFF-FX) to 8-bit linear-RGB values.
 *
 * Input:::
 *  L::				8-bit L* value ([0,255] stored in a UInt16)
 *  a::				8-bit a* value ([0,255] stored in a UInt16)
 *  b::				8-bit b* value ([0,255] stored in a UInt16)
 *
 * Output:::
 *  plR8::			8-bit red value returned
 *  plG8::			8-bit green value returned
 *  plB8::			8-bit blue value returned
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ICCLAB2LinearRGB(
	      UInt16 L,
	      UInt16 a,
	      UInt16 b,
	      UInt8 *plR8,
	      UInt8 *plG8,
	      UInt8 *plB8
)
{
    float	c, t;
    float	X, Y, Z;
    float	Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;
    float	Rp, Gp, Bp;
    PROLOGUE;
 

    /* Convert from ICCLAB -> XYZ: */
    c = ((float)L * 100.0f / 255.0f + 16.0f) / 116.0f;
    t = ((float)a - 128.0f) / 500.0f + c;
    X = Xn * f_inv(t);
    Y = Yn * f_inv(c);
    t = c - ((float)b - 128.0f) / 200.0f;
    Z = Zn * f_inv(t);

    /* Convert from XYZ -> normalized linear-RGB: */
    Rp =  3.2410f * X - 1.5374f * Y - 0.4986f * Z;
    Gp = -0.9692f * X + 1.8760f * Y + 0.0416f * Z;
    Bp =  0.0556f * X - 0.2040f * Y + 1.0570f * Z;

    /* Convert from sRGB' to 8-bit linear-RGB */
    Rp *= 256.0;
    Gp *= 256.0;
    Bp *= 256.0;
    *plR8 = EIGHTBIT(Rp);
    *plG8 = EIGHTBIT(Gp);
    *plB8 = EIGHTBIT(Bp);

    NORMAL_RETURN;
}
	      

/*
$Log: iormrc.c,v $
Revision 1.2  2003/02/10 22:53:30  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:10  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.33  2001/12/26 21:23:42  dabercro
Mark as released

Revision 1.32  2001/11/07 21:43:06  dabercro
Make base color signed long, to accomodate CIELAB

Revision 1.31  2001/09/18 21:10:09  dabercro
Make work with JBIG2

Revision 1.30  2001/08/22 19:10:54  dabercro
Added IFDBaseColor

Revision 1.29  2001/07/20 23:12:27  dabercro
Update comment to reflect JPEGTable insertion.

Revision 1.28  2001/07/06 18:14:18  dabercro
Remove unused variables, for gcc

Revision 1.27  2001/06/29 22:36:08  dabercro
Move the IO_ colorspace functions here from tfxconv.c

Revision 1.26  2001/06/01 18:03:42  dabercro
Update parameters for ImageReadLayer to support reading color layers.

Revision 1.25  2001/05/01 00:43:23  dabercro
Change planar configuration to short. Added comments.

Revision 1.24  2001/04/02 18:25:11  dabercro
Add HostComputer.

Revision 1.23  2001/03/22 20:16:38  dabercro
Add support for new TIFF field: Predictor.

Revision 1.22  2001/03/22 17:33:26  dabercro
Add T.44 (no appendix) type.

Revision 1.21  2001/02/24 03:23:37  dabercro
Additions for T.44

Revision 1.20  2001/01/18 00:21:28  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.19  2000/11/18 00:53:24  dabercro
Avoid cross polenation of data handles across sublibraries
(XIF, MRC, TFX), especially for CopyImage.  Added a handle
type to structure, now checked also.

Revision 1.18  2000/11/17 23:57:34  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.17  2000/11/02 22:21:20  dabercro
Additions for Rendering Hints.

Revision 1.16  2000/10/18 20:33:33  dabercro
Use SHOWERR #define to isolate handling of the source
of memory allocations.  Define in Makefile or MSVC project if
desired.

Revision 1.15  2000/10/16 18:22:52  dabercro
Major adds to support stripes, 0-byte strips, and ImageBaseColor.

Revision 1.14  2000/09/22 21:50:21  dabercro
Variable-sized strips now handle multiple layers.

Revision 1.13  2000/09/15 22:08:37  dabercro
Continue when only a warning.

Revision 1.12  2000/08/25 23:46:19  dabercro
Additions to allow marking annotations for write.

Revision 1.11  2000/08/22 01:03:22  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.10  2000/07/31 23:50:58  dabercro
Name change from JBIG2Options to T88Options.
Additional parameters to GetLayerInfo.

Revision 1.9  2000/07/19 00:12:48  dabercro
Keep track of image/page handles returned in GetPageInfo,
GetImageInfo, and GetLayerInfo.
Added function for malloc/free; access to library versions. This
is especially useful when passing back allocated data, when the
client hasn't passed in their own malloc/free routines. Need to use
the paired up free routine!

Revision 1.8  2000/06/06 23:47:11  dabercro
Continued work, lots of stuff.

Revision 1.7  2000/05/10 20:20:00  dabercro
Comment fix

Revision 1.6  2000/05/09 01:54:07  dabercro
Added FillOrder field support.
Support for JBIG2 dictionary extensions, updated parameter to
reflect byte count for extensions.
Write fields that are explicitly set, regardless of default values.

Revision 1.5  2000/04/12 19:06:06  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.4  2000/03/13 23:26:20  dabercro
Check for XIF first, since the headers are the same for TIF.

Revision 1.3  2000/03/11 02:06:34  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.2  2000/02/08 21:25:32  dabercro
Disable JBIG2 (for now), updates to make use of GlobalParametersIFD for symbol libraries.

Revision 1.1.1.1  2000/02/02 00:40:50  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

*/
