/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * iorpriv.c
 *
 *	IO write of IO specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	iorpriv.c:	IO READ private routines
 *
 *	IO read-only private routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#ifdef __GNUC__
#include "ansiprot.h"

IP_RCSINFO(RCSInfo, "$RCSfile: iorpriv.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:30 $");

#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_IO, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#ifdef _MSC_VER
#pragma pack(pop, PRE_IO)
#endif
#endif	/* NOIPCORE */

#include "iodoc.h"
#include "iorpriv.h"


#ifdef SHOWERR
/* Here's your one shot at putting a break-point where the routines
 * are about to return an error.  This is purely for debug only. */
void io_error(
	      int error,
	      const char *file,
	      long line
)
{
    static int complaints = 0;

    switch (error) {
    case IO_WARN_NOSUPPORT:
    case IO_WARN_NOTCOMPLIANT:
    case IO_WARN_BADFILEFORMAT:
	complaints = complaints;    /* Do nothing about these errors */
	break;
    default:
	if (++complaints <= 20) {
	    printf("Error in %s, line %ld\n", file, line);
	}
    }
}
#endif




/* IO internal memory allocation adds some info to help detect leaks
 * and page faults. I place a key word before and after the
 * memory during the allocation, and check that it is still
 * in tact during the free. I also place the size of memory
 * before, so that I know where the key is after the block,
 * and to keep track of memory usage at any point.
 */
#define IO_MEMORY_KEY		0x75c38a19	/* Ya, I just made it up */
#define IO_KEYED_MEMORY	(sizeof(long) * 3)

/* Memory and I/O */
void * CDECL
wp_IoMalloc(
    size_t bytes,
    Bool hasPtrs,
    void *clientMallocID
#ifdef SHOWERR
    , const char *file,
    const int line
#endif
)
{
    IO_MemoryTally *tally = (IO_MemoryTally *)clientMallocID;
    void *memory;
    UInt32 *lmem;
	
    if (clientMallocID == NULL || bytes == 0) {
	/* reference once, to avoid compiler warnings */
	hasPtrs = 0;
#ifdef SHOWERR
	{
	    int junk;
	    junk = (int)file;
	    junk = line;
	}
#endif
	return NULL;
    }

    
    bytes = (bytes + 3UL) & ~3UL;	/* Round to next word */
    if ((memory = (void *)malloc(bytes + IO_KEYED_MEMORY)) == NULL) {
	return memory;
    }
    tally->bytesAllocated += bytes;
    if (tally->bytesAllocated > tally->peak) {
	tally->peak = tally->bytesAllocated;
    }
    
    /* Place memory key and size around buffer */
    lmem = (UInt32 *)memory;
    lmem[0] = bytes;
    lmem[1] = (UInt32)IO_MEMORY_KEY;
    lmem[((bytes + IO_KEYED_MEMORY) >> 2) - 1] = (UInt32)IO_MEMORY_KEY;
    
    return (void *)(lmem + 2);
}

/* There's a fundamental problem with making a calloc and realloc function: they both
 * require the use of the client's malloc and free, which we don't have direct access
 * to inside a hand-off function (the clientMallocID) points to the client's data, not
 * the IO_INSTANCE handle we need to know what malloc and free function to call).
 */
#if 0
void * CDECL
wp_IoCalloc(
    size_t nel,
    size_t size,
    Bool hasPtrs,
    void *clientMallocID
#ifdef SHOWERR
    , const char *file,
    const int line
#endif
)
{
    IO_INSTANCE *inst = (IO_INSTANCE *)clientMallocID;
/*    IO_MemoryTally *tally = (IO_MemoryTally *)clientMallocID; */
    char *memory;
    UInt32 bytes = nel * size;
    UInt32 *lmem;
	
    if (thing == NULL || bytes == 0) {
	/* reference once, to avoid compiler warnings */
	hasPtrs = 0;
#ifdef SHOWERR
	{
	    int junk;
	    junk = (int)file;
	    junk = line;
	}
#endif
	return NULL;
    }


    /* Get the memory from the malloc routine */
    memory = inst->malloc(bytes, hasPtrs, inst->clientInstID
#ifdef SHOWERR
    , file, line
#endif
    );

    if (memory == NULL) return memory;

    /* Now do the C part of Calloc */
    memset(memory, 0, bytes);

    return (void *)memory;
}

void * CDECL
wp_IoRealloc(
    void *ptr,
    size_t size,
    Bool hasPtrs,
    void *thing
#ifdef SHOWERR
    , const char *file,
    const int line
#endif
)
{
    UInt32 oldSize;
    IO_INSTANCE *inst = (IO_INSTANCE *)thing;
    void *buffer;
	    
    if (hasPtrs) {
      return NULL;
    }
	
    if (thing == NULL) {
	return NULL;
    }
				    
    /* This isn't right in the case of client malloc!!!!!!!!!!!!!
     *
     * We need to know the size of the old buffer, to know how to 
     * reallocate it. In the case of the memory allocation I have,
     * I keep it stored with the buffer. The client may do anything
     * they desire, so I have no way of pinging for this information.
     */
    oldSize = wp_IoAllocSize(ptr, inst->mallocID);
    buffer = inst->malloc(size, hasPtrs, inst->mallocID
#ifdef SHOWERR
	, __FILE__, __LINE__
#endif
	);
    
    if (oldSize > size) {
	memcpy(buffer, ptr, size);
    } else {
	memcpy(buffer, ptr, oldSize);
    }
    inst->free(ptr, oldSize, inst->mallocID);
    
    return buffer;
}
#endif


void CDECL
wp_IoFree(
    void *buffer,
    size_t dwBytes,
    void *clientMallocID
)
{
    IO_MemoryTally *tally = (IO_MemoryTally *)clientMallocID;
    UInt32 *lmem;
    UInt32 bytes;
	
    if (clientMallocID == NULL) {
	dwBytes = 0;	    /* Reference once, to avoid compiler warnings */
	return;
    }
    
    if ((unsigned)buffer & 3) {
	fprintf(stderr, "Invalid memory free, buffer is not long aligned\n");
	goto memoryError;
    }
    
    lmem = ((UInt32 *)buffer) - 2;
    bytes = lmem[0];		/* If seg fault here, buffer doesn't contain 4
			 * bytes -- buffer is probably invalid */
    if (bytes < 4) {
	fprintf(stderr, "Memory corruption, memory before buffer invalid\n");
	goto memoryError;
    }
    if (lmem[1] != (UInt32)IO_MEMORY_KEY) {
	fprintf(stderr, "Memory key corrupted, memory before buffer invalid\n");
	goto memoryError;
    }
    if (lmem[((bytes + IO_KEYED_MEMORY) >> 2) - 1] != (UInt32)IO_MEMORY_KEY) {
	fprintf(stderr, "Memory key corrupted, memory after buffer invalid\n");
	goto memoryError;
    }
    
    tally->bytesAllocated -= bytes;
    free((char *)lmem);
    return;		/* Normal return */
    
 memoryError:
    return;		/* Put break point here in debugger */
}

UInt32 CDECL
wp_IoMemoryCheck(
    IO_MemoryTally *tally
)
{
    
#ifdef MYDEBUG  /* This is meant to be something other than _DEBUG */
    fprintf(stderr, "%d bytes currently allocated, %d bytes used at peak.\n",
	    tally->bytesAllocated, tally->peak);
#endif /* DEBUG */
    return tally->bytesAllocated;
}


UInt32 CDECL
wp_IoAllocSize(
    void *buffer,
    void *clientID
)
{
    IO_MemoryTally *tally = (IO_MemoryTally *)clientID;
    UInt32 *lmem;
    UInt32 bytes;

    if (tally == NULL) {
	goto memoryError;
    }
    
    if ((unsigned)buffer & 3) {
	fprintf(stderr, "Invalid memory, buffer is not long aligned\n");
	goto memoryError;
    }
    
    lmem = ((UInt32 *)buffer) - 2;
    bytes = lmem[0];		/* If seg fault here, buffer doesn't contain 4
			 * bytes -- buffer is probably invalid */
    return bytes;		/* Normal return */
    
 memoryError:
    return 0UL;		/* Put break point here in debugger */
}


/***********************************************************
 *       IPCORE memory interface
 */

void * CDECL
wp_IO_Malloc(
    size_t count,
    Bool hasPtrs,
    void *thing
#ifdef SHOWERR
    , const char *file,
    const int line
#endif
)
{
    IO_INSTANCE *inst = (IO_INSTANCE *)thing;

    if (hasPtrs) {
      return NULL;
    }
	
    if (thing == NULL) {
	return NULL;
    }
    return inst->malloc(count, hasPtrs, inst->mallocID
#ifdef SHOWERR
	, file, line
#endif
	);
}



void * CDECL
wp_IO_Calloc(
    size_t nel,
    size_t size,
    Bool hasPtrs,
    void *thing
#ifdef SHOWERR
    , const char *file,
    const int line
#endif
)
{
    IO_INSTANCE *inst = (IO_INSTANCE *)thing;
    void *buffer;
	     
    if (hasPtrs) {
      return NULL;
    }
	
    if (thing == NULL) {
      return NULL;
    }
    buffer = inst->malloc(nel * size, hasPtrs, inst->mallocID
#ifdef SHOWERR
	, file, line
#endif
	);
    memset(buffer, 0, nel * size);
 
    return buffer;
}



void * CDECL
wp_IO_Realloc(
    void *ptr,
    size_t size,
    Bool hasPtrs,
    void *thing
)
{
    UInt32 oldSize;
    IO_INSTANCE *inst = (IO_INSTANCE *)thing;
    void *buffer;
	    
    if (hasPtrs) {
      return NULL;
    }
	
    if (thing == NULL) {
	return NULL;
    }
				    
    /* This isn't right in the case of client malloc!!!!!!!!!!!!!
     *
     * We need to know the size of the old buffer, to know how to 
     * reallocate it. In the case of the memory allocation I have,
     * I keep it stored with the buffer. The client may do anything
     * they desire, so I have no way of pinging for this information.
     */
    oldSize = wp_IoAllocSize(ptr, inst->mallocID);
    buffer = inst->malloc(size, hasPtrs, inst->mallocID
#ifdef SHOWERR
	, __FILE__, __LINE__
#endif
	);
    
    if (oldSize > size) {
	memcpy(buffer, ptr, size);
    } else {
	memcpy(buffer, ptr, oldSize);
    }
    inst->free(ptr, oldSize, inst->mallocID);
    
    return buffer;
}



void CDECL
wp_IO_Free(
    void *ptr,
    size_t size,
    void *thing
)
{
    IO_INSTANCE *inst = (IO_INSTANCE *)thing;

    if (thing == NULL) {
	size = 0;	/* Reference once to avoid compiler warnings */
	return;
    }
    inst->free(ptr, 0, inst->mallocID);
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free the internal data associated with an image
 *
 * LongDescription:::
 *  wp_IoImgFree() frees the data associated with an image handle, and
 *  optionally the subsequent images also. Since the images belonging to
 *  a page are stored internally in a linked list, any image freed could
 *  either have it's link removed, or all links down the chain.
 *
 * Input:::
 *  pioImg::			Data structure of valid image
 *  subsequentImagesToo::	Boolean for: free connected images also
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_IoImgFree(
    IO_IMG **pioImg,
    Int32 subsequentImagesToo
)
{
    IO_INSTANCE *instData = NULL;
    IO_IMG *ioImg, *nextImg = NULL;
    PROLOGUE;

    IO_CheckHandleValidity(*pioImg);

    ioImg = *pioImg;
    instData = IO_IMGINST(ioImg);

    while (ioImg) {
	nextImg = ioImg->next;

	if (ioImg->ioMaskImg) {
	    XERR_PASS_CLEANUP(wp_IoImgFree(&ioImg->ioMaskImg, 1));
	}

	if (ioImg->ioHintsImgs) {
	    XERR_PASS_CLEANUP(wp_IoImgFree(&ioImg->ioHintsImgs, 1));
	}

	IO_FREE(ioImg);

	if (! subsequentImagesToo) {
	    break;
	}

	ioImg = nextImg;
    }

    *pioImg = nextImg;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free the internal data associated with a page
 *
 * LongDescription:::
 *   wp_IoPageFree() frees the data associated with an page handle, and
 *  optionally the subsequent pages also. Since the pages belonging to
 *  a document are stored internally in a linked list, any page freed could
 *  either have it's link removed, or all links down the chain. All images
 *  tied to the page(s) are also freed.
 *
 * Input:::
 *  pioPage::			Pointer to data structure of valid page
 *  subsequentPagesToo::	Boolean for: free other pages along chain
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_IoPageFree(
    IO_PAGE **pioPage,
    Int32 subsequentPagesToo
)
{
    IO_INSTANCE *instData = NULL;
    IO_PAGE *ioPage, *nextPage = NULL;
    PROLOGUE;

    IO_CheckHandleValidity(*pioPage);

    ioPage = *pioPage;
    instData = IO_PAGEINST(ioPage);

    while (ioPage) {
	nextPage = ioPage->next;

	if (ioPage->ioImgs) {
	    XERR_PASS_CLEANUP(wp_IoImgFree(&ioPage->ioImgs, 1));
	}

	/* Free the page itself */
	IO_FREE(ioPage);

	if (! subsequentPagesToo) {
	    break;
	}

	ioPage = nextPage;
    }

    *pioPage = nextPage;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free all data associated with a document handle
 *
 * LongDescription:::
 *  wp_IoDocFree() free all data associated with a document handle, including
 *  all page and image data.
 *
 * Input:::
 *  pioDoc::			Pointer to data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_IoDocFree(
    IO_DOC **pioDoc
)
{
    IO_INSTANCE *instData = NULL;
    IO_DOC *ioDoc;	
    PROLOGUE;

    IO_CheckHandleValidity(*pioDoc);

    ioDoc = *pioDoc;
    instData = IO_DOCINST(ioDoc);

    if (ioDoc->ioPages) {
	XERR_PASS_CLEANUP(wp_IoPageFree(&ioDoc->ioPages, 1));
    }

    if (ioDoc->ioFile) {
	IO_FREE(ioDoc->ioFile);
    }
    IO_FREE(ioDoc);

    *pioDoc = NULL;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}

/*
$Log: iorpriv.c,v $
Revision 1.2  2003/02/10 22:53:30  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:10  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.11  2001/12/26 21:23:42  dabercro
Mark as released

Revision 1.10  2001/09/18 21:10:09  dabercro
Make work with JBIG2

Revision 1.9  2001/01/18 00:21:28  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.8  2000/11/17 23:57:34  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.7  2000/11/02 22:21:21  dabercro
Additions for Rendering Hints.

Revision 1.6  2000/10/18 20:33:33  dabercro
Use SHOWERR #define to isolate handling of the source
of memory allocations.  Define in Makefile or MSVC project if
desired.

Revision 1.5  2000/07/31 23:51:25  dabercro
Debug name change.

Revision 1.4  2000/06/06 23:47:11  dabercro
Continued work, lots of stuff.

Revision 1.3  2000/04/12 19:06:07  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.2  2000/03/11 02:06:34  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.1.1.1  2000/02/02 00:40:50  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

*/
