/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * iowmrc.c
 *
 *	IO write of IO specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/


/*	iowmrc.c:	IO WRITE source
 *
 *	IO write-only routines
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#ifdef __GNUC__
#include "ansiprot.h"

IP_RCSINFO(RCSInfo, "$RCSfile: iowmrc.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:30 $");

#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_IO, 1)
#endif
#include "ipwcore.prv"
#ifdef _MSC_VER
#pragma pack(pop, PRE_IO)
#endif
#endif	/* NOIPCORE */

#include "iomrc.h"
#include "tfx.h"
#include "iofile.h"
#include "ioimage.h"
#include "iorpriv.h"


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a new IO document
 *
 * LongDescription:::
 *  IO_CreateDocument() creates a new IO document, and prepares to write
 *  data to it. The client must have already created the file at the operating
 *  system level, so that the read, write, and seek methods are available.
 *  <BR>
 *  A handle representing the document is passed back through a parameter.
 *  This handle is passed in to all subsequent IO_  function calls, and will
 *  encompass the IO_INSTHANDLE passed in here.
 *
 * Input:::
 *  ioInstHandle::		IO instance handle, returned from
 *				IO_InitInstance()
 *  ioDocType::			File format: xif, T.44, TIFF-FX, ...
 *  dwClientID::		Client ID, passed back through all I/O funcs
 *  version::			Version of IO document to create
 *  ioReadFunc::		Client file read function
 *  ioWriteFunc::		Client file write function
 *  ioSeekFunc::		Client file seek function
 *  ioSizeFunc::		Client file size function
 *  ioTellFunc::		Client file tell function
 *
 * Output:::
 *  pioDocHandle::		Pointer to document handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_CreateDocument(
	IO_INSTHANDLE ioInstHandle,
	DOCTYPE ioDocType,
	Int32 dwClientID,
	UInt8 version,
	UInt32 options,
	IO_READFUNC ioReadFunc,
	IO_WRITEFUNC ioWriteFunc,
	IO_SEEKFUNC ioSeekFunc,
	IO_SIZEFUNC ioSizeFunc,
	IO_TELLFUNC ioTellFunc,
	IO_DOCHANDLE *pioDocHandle
)
{
    IO_INSTANCE *instData = (IO_INSTANCE *)ioInstHandle;
    IO_DOC *ioDoc = NULL;
    IO_FILE *ioFile = NULL;
    UInt32 newDocHandle;
    PROLOGUE;
    
    /* Clear the results for now */
    *pioDocHandle = 0UL;
    
    IO_CheckHandleValidity(ioInstHandle);
    
    /* Check for proper calling sequence */

    
    /* Create a new document (handle) */
    IO_CHECKMALLOC(ioDoc = (IO_DOC *)IO_MALLOC(sizeof(*ioDoc)));
    memset(ioDoc, 0, sizeof(*ioDoc));
    ioDoc->handleID = (UInt32)ioDoc;	/* Self identity */
    ioDoc->handleType = IO_HANDLE_TYPE;
    
    /* Create a file handler */
    IO_CHECKMALLOC(ioFile = (IO_FILE *)IO_MALLOC(sizeof(*ioFile)));
    memset(ioFile, 0, sizeof(*ioFile));
    ioFile->handleID = (UInt32)ioFile;	/* Self identity */
    
    /* Setup the file handle */
    ioFile->instance = instData;
    IO_FILETOKEN(ioFile)->dwClientFileID = dwClientID;
    IO_FILETOKEN(ioFile)->FileRead = ioReadFunc;
    IO_FILETOKEN(ioFile)->FileWrite = ioWriteFunc;
    IO_FILETOKEN(ioFile)->FileSeek = ioSeekFunc;
    IO_FILETOKEN(ioFile)->FileSize = ioSizeFunc;
    IO_FILETOKEN(ioFile)->FileTell = ioTellFunc;
    
    /* Retain the file handle */
    ioDoc->ioFile = ioFile;
    
    {
	    UInt32 tfxOptions;

	    switch (ioDocType) {
	    case DOCTYPE_TIFFFX_S:
		tfxOptions = (options & ~TFX_PROFILES) | TFX_PROFILE_S;
		break;
	    case DOCTYPE_TIFFFX_F:
		tfxOptions = (options & ~TFX_PROFILES) | TFX_PROFILE_F;
		break;
	    case DOCTYPE_TIFFFX_C:
		tfxOptions = (options & ~TFX_PROFILES) | TFX_PROFILE_C;
		break;
	    case DOCTYPE_TIFFFX_M:
		tfxOptions = (options & ~TFX_PROFILES) | TFX_PROFILE_M;
		break;
	    case DOCTYPE_TIFFFX_T:
		tfxOptions = (options & ~TFX_PROFILES) | TFX_PROFILE_T;
		break;
	    case DOCTYPE_TIFFFX_M1:
		tfxOptions = (options & ~TFX_PROFILES) | TFX_PROFILE_M1;
		break;
	    case DOCTYPE_COHTIFF_B:
		tfxOptions = (options & ~TFX_PROFILES) | TFX_PROFILE_COH_B;
		break;
	    case DOCTYPE_COHTIFF_C:
		tfxOptions = (options & ~TFX_PROFILES) | TFX_PROFILE_COH_C;
		break;
	    case DOCTYPE_COHTIFF_M:
		tfxOptions = (options & ~TFX_PROFILES) | TFX_PROFILE_COH_M;
		break;
	    case DOCTYPE_COHTIFF_M1:
		tfxOptions = (options & ~TFX_PROFILES) | TFX_PROFILE_COH_M1;
		break;
	    default:
		tfxOptions = 0;
		break;
	    }

	    /* Make sure we have the right instance set up */
	    if (instData->tfxInstance == 0) {
		XERR_PASS_CLEANUP(TFX_InitInstance(0, &instData->tfxInstance));
		XERR_PASS_CLEANUP(TFX_SetMallocFuncs(instData->tfxInstance,
		    (void *)instData->mallocID, instData->malloc, instData->free));
	    }

#if 0
	    /* Set the flavor of TIFF */
	    if (ioDocType == DOCTYPE_T44A) {
		version = 3;	/* T44 Annex A, SLC and N-layer */
	    } else if (ioDocType == DOCTYPE_T44B) {
		version = 4;	/* T44 Annex B, shared resources & JBIG2 */
	    }
#else
	    version = 7;
#endif
	    XERR_PASS_CLEANUP(TFX_CreateDocument(
		instData->tfxInstance,
		dwClientID,
		version,
		tfxOptions,
		ioReadFunc,
		ioWriteFunc,
		ioSeekFunc,
		ioSizeFunc,
		ioTellFunc,
		&newDocHandle
		));
	}

    /* Identify the type of document found */    
    ioDoc->ioDocType = ioDocType;

    /* Identify the lower level handle */
    ioDoc->ioDocHandle = newDocHandle;

    /* Return the result */
    *pioDocHandle = ioDoc->handleID;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (ioDoc) {
	IO_FREE(ioDoc);
    }
    if (ioFile) {
	IO_FREE(ioFile);
    }
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Open a document for read and write
 *
 * LongDescription:::
 *  IO_OpenDocumentReadWrite() begins a session of reading and writing
 *  an existing document. The client must already have opened the file at
 *  the operating system level, so that the read and seek methods in the
 *  file token are available.
 *  This function will perform any initialization required for reading
 *  the file, including reading enough header information to verify
 *  that the file is in a format supported by this version of
 *  the IO library.
 *  <BR>
 *  A document handle is returned through the third parameter. This handle
 *  is passed to all subsequent function calls that use the document, and
 *  will encompass the I/O functions passed in here.
 *
 * Input:::
 *  ioInstHandle::		IO instance handle, returned from
 *				IO_InitInstance()
 *  ioDocType::			File format: xif, T.44, TIFF-FX, ...
 *  dwClientID::		Input client ID
 *  ioReadFunc::		Client file read function
 *  ioWriteFunc::		Client file write function
 *  ioSeekFunc::		Client file seek function
 *  ioSizeFunc::		Client file size function
 *  ioTellFunc::		Client file tell function
 *
 * Output:::
 *  pioDocHandle::		Document handle used to refer to this
 *				page in other functions
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_OpenDocumentReadWrite(
	IO_INSTHANDLE ioInstHandle,
	DOCTYPE ioDocType,
	Int32 dwClientID,
	UInt8 version,
	UInt32 options,
	IO_READFUNC ioReadFunc,
	IO_WRITEFUNC ioWriteFunc,
	IO_SEEKFUNC ioSeekFunc,
	IO_SIZEFUNC ioSizeFunc,
	IO_TELLFUNC ioTellFunc,
	IO_DOCHANDLE *pioDocHandle
)
{
    IO_INSTANCE *instData = (IO_INSTANCE *)ioInstHandle;
    IO_DOC *ioDoc;
    IO_FILE *ioFile;
    Int32 fileSize = 0;
    DOCTYPE ioDocProfile;
    UInt32 newDocHandle;
    PROLOGUE;

    /* Clear results for now */
    *pioDocHandle = 0UL;

    IO_CheckHandleValidity(ioInstHandle);

    /* Create a file handler */
    IO_CHECKMALLOC(ioFile = (IO_FILE *)IO_MALLOC(sizeof(*ioFile)));
    memset(ioFile, 0, sizeof(*ioFile));
    ioFile->handleID = (UInt32)ioFile;	/* Self identity */
    
    /* Setup the file handle */
    ioFile->instance = instData;
    IO_FILETOKEN(ioFile)->dwClientFileID = dwClientID;
    IO_FILETOKEN(ioFile)->FileRead = ioReadFunc;
    IO_FILETOKEN(ioFile)->FileWrite = ioWriteFunc;
    IO_FILETOKEN(ioFile)->FileSeek = ioSeekFunc;
    IO_FILETOKEN(ioFile)->FileSize = ioSizeFunc;
    IO_FILETOKEN(ioFile)->FileTell = ioTellFunc;
    
    /* Check the file size, if it's zero, create the document instead */
    IO_FILESIZE(ioFile, fileSize);
    if (fileSize <= 8) {

	/* Don't need this any more */
	IO_FREE(ioFile);
	
	/* Create it */
	XERR_PASS_CLEANUP(IO_CreateDocument(ioInstHandle, ioDocType, dwClientID,
	    version, options, ioReadFunc, ioWriteFunc, ioSeekFunc, ioSizeFunc,
	    ioTellFunc, pioDocHandle));
	NORMAL_RETURN;
    }

    /* Get the profile type of document */

	/* Open the document read-only */
        XERR_PASS_CLEANUP(IO_OpenDocumentRead(ioInstHandle, dwClientID,
	    options, ioReadFunc, ioWriteFunc, ioSeekFunc, ioSizeFunc,
	    ioTellFunc, &newDocHandle));

	/* Specifically look for the correct profile */
	XERR_PASS_CLEANUP(IO_GetDocProfile(newDocHandle, &ioDocProfile));

	/* Fix the type, if inconsistent */
	if (ioDocProfile != ioDocType) {
	    ioDocType = ioDocProfile;
	}

	/* Close the document */
	XERR_PASS_CLEANUP(IO_CloseDocument(newDocHandle));

    /* Create a new document (handle) */
    IO_CHECKMALLOC(ioDoc = (IO_DOC *)IO_MALLOC(sizeof(*ioDoc)));
    memset(ioDoc, 0, sizeof(*ioDoc));
    ioDoc->handleID = (UInt32)ioDoc;	/* Self identity */
    ioDoc->handleType = IO_HANDLE_TYPE;
    

    /* Retain the file handle */
    ioDoc->ioFile = ioFile;

    /* Open the document, with known profile, for read and write */
	XERR_PASS_CLEANUP(TFX_OpenDocumentReadWrite(
	    instData->tfxInstance,
	    dwClientID,
	    version,
	    options,
	    ioReadFunc,
	    ioWriteFunc,
	    ioSeekFunc,
	    ioSizeFunc,
	    ioTellFunc,
	    &newDocHandle
	    ));

    /* Identify the type of document found */    
    ioDoc->ioDocType = ioDocType;

    /* Identify the lower level handle */
    ioDoc->ioDocHandle = newDocHandle;

    /* Return the result */
    *pioDocHandle = ioDoc->handleID;

    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}



/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy an individual image from one IO to another (leave compressed)
 *
 * LongDescription:::
 *  IO_CopyImage() copies an individual image from one IO to another,
 *  leaving the image data compressed during the copy.  This allows a much
 *  faster copy, since the data is not decompressed and recompressed.  It also
 *  means that the use of a lossy algorithm (like JPEG) will not produce added
 *  lossy affects by recompressing with additional or different loss.
 *  <BR>
 *  All fields from the source image are copied, including optional fields.
 *  But the annotations are NOT copied.
 *  <BR>
 *  Note: this function will NOT keep track of version inconsistencies (like
 *  the fact that a XIFF 2.0 does not support tokenization).
 *
 * Input:::
 *  ioSrcImgHandle::		Source image handle, returned from
 *				IO_GetImageInfo()
 *  ioDstParentImgHandle::	Destination image handle for parent image (if
 *				source image is a transparency mask or
 *				hints image)
 *  ioDstPageHandle::		Destination page handle, returned from
 *				IO_AddPageStart() or IO_GetPageInfo()
 *  dwImgLayer::		Image layer in MRC
 *  dwXOffset::			X pixel offset in Primary image coordinates
 *  dwYOffset::			Y pixel offset in Primary image coordinates
 *  dwBestStripHeight::		Height of strips stored
 *
 * Output:::
 *  pxDstImgHandle::		Pointer to resulting image handle, can be used
 *				to later reference this image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_CopyImage(
	IO_IMGHANDLE ioSrcImgHandle,
	IO_IMGHANDLE ioDstParentImgHandle,
	IO_PAGEHANDLE ioDstPageHandle,
	UInt32 dwImgLayer,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwBestStripHeight,
	IO_IMGHANDLE *pioDstImgHandle
)
{
    IO_INSTANCE *instData = NULL;
    IO_IMG *ioSrcImg = (IO_IMG *)ioSrcImgHandle;
    IO_IMG *ioImg = NULL;	/* New image */
    IO_IMGHANDLE dstParentImgHandle = 0;
    IO_IMGHANDLE newImgHandle;
    IO_PAGE *ioDstPage = (IO_PAGE *)ioDstPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioSrcImgHandle);
    IO_CheckHandleValidity(ioDstPageHandle);

    /* Reference the optional page handle, if it exists */
    if (ioDstParentImgHandle) {
	dstParentImgHandle = ((IO_IMG *)ioDstParentImgHandle)->ioImgHandle;
    }

	XERR_PASS_CLEANUP(TFX_CopyImage(
	    ioSrcImg->ioImgHandle,
	        dstParentImgHandle,
		((IO_PAGE *)ioDstPageHandle)->ioPageHandle,
		dwImgLayer,
		dwXOffset,
		dwYOffset,
		dwBestStripHeight,
		&newImgHandle
	    ));

    /* Create our own copy of the container */
    instData = IO_PAGEINST(ioDstPage);
    IO_CHECKMALLOC(ioImg = (IO_IMG *)IO_MALLOC(sizeof(*ioImg)));
    memset(ioImg, 0, sizeof(*ioImg));
    ioImg->handleID = (UInt32)ioImg;	/* Self identity */
    ioImg->handleType = IO_HANDLE_TYPE;
    
    /* Link back to it's parent */
    ioImg->ioPage = ioDstPage;

    /* Identify the lower level handle */
    ioImg->ioImgHandle = newImgHandle;

    /* Add the image to the page. Note: the images are in the order they
     * were added, and not the sequence dictated by image additions.  The
     * lower level routines maintain image/layer order. */
    if (ioDstPage->ioImgs == NULL) {
	/* Start the page with this image */
	ioDstPage->ioImgs = ioImg;
    }
    else {
	IO_IMG *tempImg;
	
	/* Put it into the linked list at the end */
	tempImg = ioDstPage->ioImgs;
	while (tempImg->next != NULL) {
	    tempImg = tempImg->next;
	}

	tempImg->next = ioImg;
    }

    /* Return the result */
    *pioDstImgHandle = ioImg->handleID;

    NORMAL_RETURN;

ERR_CLEANUP:

    if (ioImg) {
	IO_FREE(ioImg);
    }
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copies the entire contents of a page from one IO to another
 *
 * LongDescription:::
 *  IO_CopyPage() copies a specified page from one IO document to
 *  another. The page can be an arbitrary page in the source document,
 *  but must be the next available page in the destination document
 *  (no page insertions).  The image data is left compressed during the
 *  copy, enabling fast copies.  Version differences are NOT fixed in
 *  this copy. If the compression stream of the source image data is of
 *  an outdated format, the stream is not updated.
 *  <BR>
 *  This function can be used to facilitate page insertion, deletion and
 *  reordering; but you must place it in a loop with the appropriate page
 *  offsets to sequentially specify the new destination page order.
 *
 * Input:::
 *  ioSrcDocHandle::		Source document handle, returned from
 *				IO_OpenDocumentRead() or
 *				IO_OpenDocumentReadWrite()
 *  ioDstDocHandle::		Destination document handle, returned from
 *				IO_CreateDocument() or
 *				IO_OpenDocumentReadWrite()
 *  dwSrcPageOffset::		Source document offset of page to copy
 *  dwDstPageOffset::		Destination document offset of page to add
 *				(this should be the end of the document).
 *				-1 indicates the next available page.
 *  copyAnnotations::		Boolean for copy the page annotations
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_CopyPage(
    IO_DOCHANDLE ioSrcDocHandle,
	IO_DOCHANDLE ioDstDocHandle,
	Int32 dwSrcPageOffset,
	Int32 dwDstPageOffset,
	Bool copyAnnotations
    )
{
    IO_DOC *ioSrcDoc = (IO_DOC *)ioSrcDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioSrcDocHandle);

	XERR_PASS_CLEANUP(TFX_CopyPage(
	    ioSrcDoc->ioDocHandle,
		((IO_DOC *)ioDstDocHandle)->ioDocHandle,
		dwSrcPageOffset,
		dwDstPageOffset,
		copyAnnotations
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copies the entire document from one file to another that is
 *  stream-ordered
 *
 * LongDescription:::
 *  IO_CopyDoc() copies all known file fragments from one file to
 *  another, creating an output file that has stream-ordered data.
 *  Note, streaming order is a "recommendation" for TIFF-FX; this
 *  function is used to facilitate the rewrite of a TIFF-FX file that
 *  may not be in streamed order.
 *  <BR>
 *  All image fragments will remain compressed, but the order in which
 *  they reside in the document may change.
 *  <BR>
 *  Note, using IO_CopyPage() may not be sufficient to create streaming
 *  order.
 *
 * Input:::
 *  ioSrcDocHandle::		Source document handle, returned from
 *				IO_OpenDocumentRead() or
 *				IO_OpenDocumentReadWrite()
 *  ioDstDocHandle::		Destination document handle, returned from
 *				IO_CreateDocument() or
 *				IO_OpenDocumentReadWrite()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_CopyDoc(
	   IO_DOCHANDLE ioSrcDocHandle,
	   IO_DOCHANDLE ioDstDocHandle
)
{
    IO_DOC *ioSrcDoc = (IO_DOC *)ioSrcDocHandle;
    IO_DOC *ioDstDoc = (IO_DOC *)ioDstDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioSrcDocHandle);

	XERR_PASS_CLEANUP(TFX_CopyDoc(
	    ioSrcDoc->ioDocHandle,
		ioDstDoc->ioDocHandle
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Begin the process of adding a new page
 *
 * LongDescription:::
 *  IO_AddPageStart() adds a new page to a multipage document. Page offsets
 *  start at zero, but setting dwPageNumber < 0 causes the new page to be
 *  added at the end of the document. The operation will fail if a page
 *  is added beyond the next available page at the end of the document, i.e.
 *  {0, 1, 2, 7} is not a valid page sequence, and the attempt to add page
 *  offset 7 would fail.
 *  <BR>
 *  Note that when adding pages A, B, and C, each with offset = 0, will result
 *  in the pages having a reverse page order -- the final page offsets will be
 *  2, 1, and 0.
 *  <BR>
 *  The page handle returned from this routine is used in later calls to
 *  attach images and annotations to this page.
 *  <BR>
 *  IO_AddPageStart() may be followed by adding a Primary image using
 *  IO_AddImage(). 
 *
 * Input:::
 *  ioDocHandle::		Document handle
 *  dwPageOffset::		Offset of page to add, should be up to the
 *				existing page count, or negative, to indicate
 *				the end of the document
 *
 * Output:::
 *  pioPageHandle::		Page handle returned
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_AddPageStart(
    IO_DOCHANDLE ioDocHandle,
	Int32 dwPageOffset,
	IO_PAGEHANDLE *pioPageHandle
    )
{
    UInt32 newPageHandle = 0UL;
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    IO_PAGE *ioPage = NULL;
    IO_INSTANCE *instData = NULL;
    PROLOGUE;

    /* Clear the result */
    *pioPageHandle = 0;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_AddPageStart(
	    ioDoc->ioDocHandle,
		dwPageOffset,
		&newPageHandle
	    ));

    /* Create our own copy of the container */
    instData = IO_DOCINST(ioDoc);
    IO_CHECKMALLOC(ioPage = (IO_PAGE *)IO_MALLOC(sizeof(*ioPage)));
    memset(ioPage, 0, sizeof(*ioPage));
    ioPage->handleID = (UInt32)ioPage;	/* Self identity */
    ioPage->handleType = IO_HANDLE_TYPE;
    
    /* Link back to it's parent */
    ioPage->ioDoc = ioDoc;

    /* Identify the lower level handle */
    ioPage->ioPageHandle = newPageHandle;

    /* Add the page to the document. Note: the pages are in the order they
     * were added, and not the sequence dictated by page insertions.  The
     * lower level routines maintain page order. */
    if (ioDoc->ioPages == NULL) {
	/* Start the document with this page */
	ioDoc->ioPages = ioPage;
    }
    else {
	IO_PAGE *tempPage;

	/* Put it into the linked list at the end */
	tempPage = ioDoc->ioPages;
	while (tempPage->next != NULL) {
	    tempPage = tempPage->next;
	}
	    
	tempPage->next = ioPage;
    }

    /* Return the result */
    *pioPageHandle = ioPage->handleID;

    NORMAL_RETURN;

ERR_CLEANUP:

    if (ioPage) {
	IO_FREE(ioPage);
    }
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Complete the construction of a page, tokenization still pending
 *
 * LongDescription:::
 *  IO_AddPageFinish() completes the general construction of a page, although
 *  the completion of the tokenization process later may actually store the
 *  data for the Primary IFD, and may also add other subimages for Layer 2.
 *
 * Input:::
 *  ioPageHandle::		Page handle of page to complete
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_AddPageFinish(
    IO_PAGEHANDLE ioPageHandle
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_AddPageFinish(
	    ioPage->ioPageHandle
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Delete an entire page from the document (including all images on the page)
 *
 * LongDescription:::
 *  IO_DeletePage() deletes (removes all references to) an entire page in
 *  the file.  The image data is not physically removed (as the image is
 *  likely to be inbetween other file fragments), but the references to the
 *  page are removed.  This includes: page count in the page table header,
 *  removing the primary IFD from the page table (and shifting subsequent
 *  page references to accomodate the removal), and relinking the previous
 *  page to the subsequent page via the NextIFD IFD entry.
 *  <BR>
 *  Note that although the page is not physically removed, since the
 *  references to it are removed (and lost), there is no way to 'undo' a
 *  page deletion.
 *  <BR>
 *  The intent is to allow physically editing the file, with anticipation
 *  that later the file will be copied (see IO_CopyPage() and
 *  IO_CopyImage()), allowing the disowned file fragments to remain behind.
 *  <BR>
 *  Note also that the page offsets of all subsequent pages will automatically
 *  be shifted down one (previous page offsets will now be wrong, only for
 *  pages that followed this one).
 *
 * Input:::
 *  ioPageHandle::		Page handle of the page to be removed, as
 *				returned from IO_GetPageInfo()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_DeletePage(
    IO_PAGEHANDLE ioPageHandle
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_DeletePage(
	    ioPage->ioPageHandle
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Check and adjust (if necessary) the image info
 *
 * LongDescription:::
 *  IO_CheckImageInfo() verifies the consistency of the image info based
 *  on the IO version, and compression type. Changes are made to the
 *  parameters based on what IO_AddImage(), IO_AddMaskImage(), and/or
 *  IO_ImageWriteStart() would do, or what IO_TokenizeImage() would do.
 *
 * Input:::
 *  ioPageHandle::		Page handle
 *  pdwImgLayer::		Image layer in MRC (1, 2, ...)
 *  pdwImgOffset::		Image rank within the layer (0, 1, 2, ...)
 *  pioColorType::		Color type
 *  pdwXOffset::		X pixel offset in Primary image coordinates
 *  pdwYOffset::		Y pixel offset in Primary image coordinates
 *  pdwXResolution::		Resolution in the X direction
 *  pdwYResolution::		Resolution in the Y direction
 *  pdwWidth::			Width
 *  pdwHeight::			Height
 *  pdwBestStripHeight::	Height of strips stored
 *  pPalette::			Palette (if IO_COLORTYPE_COLOR8 or COLOR4)
 *  pioCompression::		Type of compression
 *  pioColorSpace::		Stored color space
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_CheckImageInfo(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 *pdwImgLayer,
	UInt32 *pdwImgOffset,
	IO_COLORTYPE *pioColorType,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	UInt8 *pPalette,
	IO_COMPRESSION *pioCompression,
	IO_COLORSPACE *pioColorSpace
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_CheckImageInfo(
	    ioPage->ioPageHandle,
		pdwImgLayer,
		pdwImgOffset,
		pioColorType,
		pdwXOffset,
		pdwYOffset,
		pdwXResolution,
		pdwYResolution,
		pdwWidth,
		pdwHeight,
		pdwBestStripHeight,
		pPalette,
		pioCompression,
		pioColorSpace
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add an image to a page
 *
 * LongDescription:::
 *  IO_AddImage() adds a new image to a page. Image should not be a
 *  transparancy mask, IO_AddMaskImage() is used for handling this.
 *  <BR>
 *  Note that every empty page must start with a Primary image which defines
 *  the boundaries of the page, both in width and height.  IO_BADPARAMETER
 *  will be returned if the first image added is not a primary image (image
 *  layer == 2).
 *  <BR>
 *  An image handle is returned, that is used in later calls to three
 *  functions: IO_ImageWriteStart(), IO_ImageWriteStrip(), and
 *  IO_ImageWriteFinish(); or to the two functions:
 *  IO_WriteTokenizeImage(), and IO_WriteTokenImage(); or the function:
 *  IO_WriteTokenizedImage().
 *
 * Input:::
 *  ioPageHandle::		Page to attach to
 *  dwImgLayer::		Image layer in MRC
 *  ioColorType::		Color type
 *  dwXOffset::			X pixel offset in Primary image coordinates
 *  dwYOffset::			Y pixel offset in Primary image coordinates
 *  dwXResolution::		Resolution in the X direction
 *  dwYResolution::		Resolution in the Y direction
 *  dwWidth::			Width
 *  dwHeight::			Height
 *  dwBestStripHeight::		Height of strips stored
 *  pPalette::			Palette (if IO_COLORTYPE_COLOR8 or COLOR4)
 *
 * Output:::
 *  pImgHandle::		Pointer to image handle, used to later
 *				reference this image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_AddImage(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 dwImgLayer,
	IO_COLORTYPE ioColorType,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwXResolution,
	UInt32 dwYResolution,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt32 dwBestStripHeight,
	UInt8 *pPalette,
	IO_IMGHANDLE *pioImgHandle
    )
{
    UInt32 newImgHandle = 0UL;
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    IO_IMG *ioImg = NULL;
    IO_INSTANCE *instData = NULL;
    PROLOGUE;

    /* Clear the result */
    *pioImgHandle = 0;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_AddImage(
	    ioPage->ioPageHandle,
		dwImgLayer,
		ioColorType,
		dwXOffset,
		dwYOffset,
		dwXResolution,
		dwYResolution,
		dwWidth,
		dwHeight,
		dwBestStripHeight,
		pPalette,
		&newImgHandle
	    ));

    /* Create our own copy of the container */
    instData = IO_PAGEINST(ioPage);
    IO_CHECKMALLOC(ioImg = (IO_IMG *)IO_MALLOC(sizeof(*ioImg)));
    memset(ioImg, 0, sizeof(*ioImg));
    ioImg->handleID = (UInt32)ioImg;	/* Self identity */
    ioImg->handleType = IO_HANDLE_TYPE;
    
    /* Link back to it's parent */
    ioImg->ioPage = ioPage;

    /* Identify the lower level handle */
    ioImg->ioImgHandle = newImgHandle;

    /* Add the image to the page. Note: the images are in the order they
     * were added, and not the sequence dictated by image additions.  The
     * lower level routines maintain image/layer order. */
    if (ioPage->ioImgs == NULL) {
	/* Start the page with this image */
	ioPage->ioImgs = ioImg;
    }
    else {
	IO_IMG *tempImg;
	
	/* Put it into the linked list at the end */
	tempImg = ioPage->ioImgs;
	while (tempImg->next != NULL) {
	    tempImg = tempImg->next;
	}
	    
	tempImg->next = ioImg;
    }

    /* Return the result */
    *pioImgHandle = ioImg->handleID;

    NORMAL_RETURN;

ERR_CLEANUP:

    if (ioImg) {
	IO_FREE(ioImg);
    }
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Delete an image from a page
 *
 * LongDescription:::
 *  IO_DeleteImage() deletes (removes all references to) an single image on
 *  a page.  The image data is not physically removed (as the image is
 *  likely to be inbetween other file fragments), but the references to the
 *  image are removed.  This includes: IO layer image counts and relinking
 *  the previous image to the subsequent image via the NextIFD IFD entry.
 *  <BR>
 *  Note that although the image is not physically removed, since the
 *  references to it are removed (and lost), there is no way to 'undo' an
 *  image deletion.
 *  <BR>
 *  The intent is to allow physically editing the file, with anticipation
 *  that later the file will be copied (see IO_CopyPage() and
 *  IO_CopyImage()), allowing the disowned file fragments to remain behind.
 *
 * Input:::
 *  ioImgHandle::		Image handle of the image to be removed, as
 *				returned from IO_GetImageInfo()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_DeleteImage(
    IO_IMGHANDLE ioImgHandle
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_DeleteImage(
	    ioImg->ioImgHandle
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add a transparency mask to a non-binary image
 *
 * LongDescription:::
 *  IO_AddMaskImage() is like IO_AddImage(), except that instead of adding
 *  an image to a page, we are adding a transparency mask image to a parent
 *  non-binary image.  The image layer is inherited from the parent image.
 *  This image must have the same X/Y location, width, height and resolutions
 *  as the parent image.  This image MUST also be binary, with 0 indicating
 *  the transparent  pixels, and 1 indicating the opaque pixels of the parent.
 *  <BR>
 *  An image handle is returned, that is used in later calls to three
 *  functions: IO_ImageWriteStart(), IO_ImageWriteStrip(), and
 *  IO_ImageWriteFinish().
 *
 * Input:::
 *  ioParentImgHandle::		Parent image to attach to
 *  dwBestStripHeight::		For best performance, write strips with
 *				height a multiple of this
 *
 * Output:::
 *  pImgHandle::		Returned image handle for image to be
 *				added
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_AddMaskImage(
    IO_IMGHANDLE ioParentImgHandle,
	UInt32 dwBestStripHeight,
	IO_IMGHANDLE *pioImgHandle
    )
{
    UInt32 newImgHandle = 0UL;
    IO_IMG *ioParentImg = (IO_IMG *)ioParentImgHandle;
    IO_IMG *ioImg = NULL;
    IO_INSTANCE *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioParentImgHandle);


	XERR_GEN_CLEANUP(1, IO_NOSUPPORT);

    /* Create our own copy of the container */
    instData = IO_IMGINST(ioParentImg);
    IO_CHECKMALLOC(ioImg = (IO_IMG *)IO_MALLOC(sizeof(*ioImg)));
    memset(ioImg, 0, sizeof(*ioImg));
    ioImg->handleID = (UInt32)ioImg;	/* Self identity */
    ioImg->handleType = IO_HANDLE_TYPE;
    
    /* Link back to it's parent */
    ioImg->ioParentImg = ioParentImg;

    /* Link back to it's page */
    ioImg->ioPage = ioParentImg->ioPage;

    /****************************/
    /* For now, remove the previous image, and only reference the new one */
    if (ioParentImg->ioMaskImg) {
	wp_IoImgFree(&ioParentImg->ioMaskImg, 0);
    }

    /****************************/
    /* Add the image to the page in the appropriate spot */
    ioParentImg->ioMaskImg = ioImg;

    /* Identify the lower level handle */
    ioImg->ioImgHandle = newImgHandle;

    /* Return the result */
    *pioImgHandle = ioImg->handleID;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Prepare to write image data to the IO image
 *
 * LongDescription:::
 *  IO_ImageWriteStart() initiates the write operation for an image. The
 *  image handle is that returned by IO_AddImage(). This call will be
 *  followed by one or more calls to IO_ImageWriteStrip(), and a closing
 *  call to IO_ImageWriteFinish.
 *  <BR>
 *  pCompression points to the desired compression method. If the compression
 *  type is invalid or unsupported, the actual compression to be used is
 *  returned through this pointer.
 *  <BR>
 *  If JPEG compression is used, pJPEGQuality points to the desired JPEG
 *  quality (1-100). If the JPEG quality changes from what is desired to what
 *  will be used, the actual value is passed back through the pointer. An
 *  unreasonable quality factor will yield a different output than passed in.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_AddImage()
 *  dwFlags::			Bit combination (OR'd) of (defined in iomrc.h): <UL>
 *				  <LI>IO_IMGFLAGS_WhiteIsZero <EM>(default)</EM></LI>
 *				  <LI>IO_IMGFLAGS_BlackIsZero</LI>
 *				  <LI>IO_IMGFLAGS_ColorPlanar <EM>(default)</EM></LI>
 *				  <LI>IO_IMGFLAGS_ColorChunky</LI>
 *				  <LI>IO_IMGFLAGS_SubSampChroma <EM>(default)</EM></LI>
 *				  <LI>IO_IMGFLAGS_NoSubSampChroma</LI>
 *				  <LI>IO_IMGFLAGS_JPEGTables</LI>
 *				</UL> 
 *  pioCompression::		Pointer to the requested compression type.
 *				IO_COMP_UNKNOWN yields default.
 *				The compression type used is passed back
 *				through this pointer
 *  pdwJPEGQuality::		Pointer to the requested JPEG quality (0-100)
 *				  0 - default used and passed back
 *				 75 - good quality, best compression
 *				 90 - better quality, good compression
 *				100 - best quality, okay compression
 *				The JPEG quality used is passed back
 *				through this pointer
 *  pioColorSpace::		Requested color space used to store the data.
 *				(Not to be confused with the color space of
 *				the input data, which is always RGB!)
 *				The color space used is passed back through
 *				this pointer
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ImageWriteStart(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwFlags, 
	IO_COMPRESSION *pioCompression,
	UInt32 *pdwJPEGQuality,
	IO_COLORSPACE *pioColorSpace
    )	  
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_ImageWriteStart(
	    ioImg->ioImgHandle,
		dwFlags, 
		pioCompression,
		pdwJPEGQuality,
		pioColorSpace
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}

#if 0	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy user image buffer to intermediate TIFF buffer (PIXR)
 *
 * LongDescription:::
 *  wp_CopyToTempImage() copies fragments of the user buffer necessary to buffer
 *  up a TIFF strip, such that a whole TIFF strip can be accumulated before the
 *  write. The temporary PIXR ioImg->tmpPixr is used as this intermediate buffer.
 *
 * Input:::
 *  ioImg::			Data structure of valid image
 *  ppImgChannels::		Pointer to array of 1 to 3 image channels to
 *				copy from
 *  srcBPL::			Source image bytes per line (for one channel)
 *  pdwBufferRowOffset::	Pointer to source image row offset, updated for
 *				outer loop
 *  pdwImgRowOffset::		Pointer to image proper row offset, updated for
 *				outer loop
 *  pdwRowsToWrite::		Pointer to the number of rows in outer loop to
 *				write
 *  pdwTotalRowsFilled::	Pointer to the total number of rows buffered out
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_CopyToTempImage(
    ioImg ioImg,
	UInt8 **ppImgChannels,
	Int32 srcBPL,
	UInt32 *pdwBufferRowOffset,
	UInt32 *pdwImgRowOffset,
	UInt32 *pdwRowsToWrite,
	UInt32 *pdwTotalRowsFilled
    )
{
    IO_IMG **ioImg)->ioPage->ioDoc->ioDocType) {
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(((IOHandle);

    switch (((IO_IMG *)*ioImg)->ioPage->ioDoc->ioDocType) {
    case DOCTYPE_XIFF2:
    case DOCTYPE_XIFF3:
	XERR_PASS_CLEANUP(XIF_CopyToTempImage(
	    ioImg = (((IO_IMG *))ioImg
		ppImgChannels,
		srcBPL,
		pdwBufferRowOffset,
		pdwImgRowOffset,
		pdwRowsToWrite,
		pdwTotalRowsFilled
	    ));
	break;
    case DOCTYPE_T44:
    case DOCTYPE_T44A:
    case DOCTYPE_T44B:
	XERR_PASS_CLEANUP(MRC_CopyToTempImage(
	    ioImg->ioImgHandle,
		ppImgChannels,
		srcBPL,
		pdwBufferRowOffset,
		pdwImgRowOffset,
		pdwRowsToWrite,
		pdwTotalRowsFilled
	    ));
	break;
    case DOCTYPE_TIFFFX_S:
    case DOCTYPE_TIFFFX_F:
    case DOCTYPE_TIFFFX_C:
    case DOCTYPE_TIFFFX_M:
    case DOCTYPE_TIFFFX_T:
    case DOCTYPE_TIFFFX_M1:
    case DOCTYPE_COHTIFF_B:
    case DOCTYPE_COHTIFF_C:
    case DOCTYPE_COHTIFF_M:
    case DOCTYPE_COHTIFF_M1:
	XERR_PASS_CLEANUP(TFX_CopyToTempImage(
	    ioImg->ioImgHandle,
		ppImgChannels,
		srcBPL,
		pdwBufferRowOffset,
		pdwImgRowOffset,
		pdwRowsToWrite,
		pdwTotalRowsFilled
	    ));
	break;
    default:
	XERR_GEN_CLEANUP(1, IO_BADPARAMETER);
    }

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compress and write a strip of image data
 *
 * LongDescription:::
 *  IO_ImageWriteStrip() compresses and writes a strip of image data. If
 *  necessary, the image data is broken up into TIFF strips, and stored as
 *  separate strips.  The last strip may be stored until the next block of
 *  data is added to complete a TIFF strip.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_AddImage()
 *  ppImagePlanes::		Pointer to image buffer(s)
 *  dwBytesPerRow::		Bytes per image row
 *  dwBufferRowOffset::		Row offset in image of the strip to write
 *  dwStripHeight::		Number of rows in PIXR to write
 *  dwImgRowOffset::		Row offset in image proper to write strip
 *
 * Output:::
 *  pRowsFilled::		Number of rows written
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data to this
 *  buffer as needed, and writing the intermediate buffer when filled.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ImageWriteStrip(
    IO_IMGHANDLE ioImgHandle,
	UInt8 **ppImagePlanes,
	UInt32 dwBytesPerRow,
	UInt32 dwBufferRowOffset,
	UInt32 dwStripHeight,
	UInt32 dwImgRowOffset,
	UInt32 *pRowsFilled
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_ImageWriteStrip(
	    ioImg->ioImgHandle,
		ppImagePlanes,
		dwBytesPerRow,
		dwBufferRowOffset,
		dwStripHeight,
		dwImgRowOffset,
		pRowsFilled
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a strip of compressed image data
 *
 * LongDescription:::
 *  IO_CompressedWriteStrip() writes a strip of already compressed image
 *  data. The strip must* be the strip height as given in the parameter
 *  dwBestStripHeight during the call to IO_AddImage(), except for the
 *  very last strip which may fill up to the end of the image.  The strip
 *  must* also lie on a strip boundary.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_AddImage()
 *  dwImgRowOffset::		Row offset in image proper to write strip
 *  dwStripHeight::		Number of rows in PIXR to write
 *  pCmpBuffer::		Pointer to the compressed image buffer
 *  dwCmpBytes::		Bytes of compressed data found in buffer
 *
 * Output:::
 *  pdwBytesStored::		Number of bytes written
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_CompressedWriteStrip(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 *pCmpBuffer,
	UInt32 dwCmpBytes,
	UInt32 *pdwBytesStored
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_CompressedWriteStrip(
	    ioImg->ioImgHandle,
		dwImgRowOffset,
		dwStripHeight,
		pCmpBuffer,
		dwCmpBytes,
		pdwBytesStored
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Store the JBIG2 compressed image (position block)
 *
 * LongDescription:::
 *  IO_StorePositionBlock() places the appropriate wrapper in front of a JBIG2
 *  "position block" or compressed image data, and stores the data as a single
 *  TIFF strip.
 *  <BR>
 *  The symbol dictionary or dictionaries must have been written before this
 *  call, via IO_StoreSymbolDictionary().  Each stored dictionary is given
 *  an ID, which can be used in the dictIDList[], in the order that the
 *  dictionaries are to be concatenated.
 *
 * Input:::
 *  ioImgHandle::		Handle to valid image
 *  pPosBlock::			JBIG2 (embedded) position block
 *  dwPosSize::			Bytes in posBlock[]
 *  dwDictIDList::		List of dictionary IDs associated with this
 *				position block
 *  wDictCount::		Number of dictionary IDs in dwDictIDList[]
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  The format of the header is:
 *	1 byte	- version number
 *	4 bytes	- position block byte count
 *	2 bytes	- Number of dictionaries associated with the position block
 *	4 bytes	- First dictionary ID
 *	...	- Next dictionary ID
 *      followed by the embedded position block
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_StorePositionBlock(
	IO_IMGHANDLE ioImgHandle,
	UInt8 *pPosBlock,
	UInt32 dwPosSize,
	UInt32 *dwDictIDList,
	UInt16 wDictCount
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_StorePositionBlock(
	    ioImg->ioImgHandle,
	    pPosBlock,
	    dwPosSize,
	    dwDictIDList,
	    wDictCount
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add a compressed symbol dictionary to the file
 *
 * LongDescription:::
 *  IO_StoreSymbolDictionary() adds a compressed symbol dictionary to the file,
 *  adding an entry in the shared resources.
 *  <BR>
 *  If the dictionary contains extensions, then the number of bytes in the
 *  entire block will be larger than that of the back-end dictionary.
 *
 * Input:::
 *  ioDocHandle::		Document handle
 *  pDictBlock::		JBIG2 compressed dictionary
 *  dwDictBytes::		Number of bytes in back-end dictionary
 *  dwDictBlockBytes::		Number of bytes in whole block (may include
 *				extensions)
 *
 * Output:::
 *  pdwDictID::			Returned dictionary ID
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_StoreSymbolDictionary(
	IO_DOCHANDLE ioDocHandle,
	UInt8 *pDictBlock,
	UInt32 dwDictBytes,
	UInt32 dwDictBlockBytes,
	UInt32 *pdwDictID
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_StoreSymbolDictionary(
	    ioDoc->ioDocHandle,
	    pDictBlock,
	    dwDictBytes,
	    dwDictBlockBytes,
	    pdwDictID
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Complete an image write
 *
 * LongDescription:::
 *  IO_ImageWriteFinish() cleans up the write operation state and performs
 *  consistecy checks.  Fails if correct number of rows not written.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_AddImage()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ImageWriteFinish(
    IO_IMGHANDLE ioImgHandle
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_ImageWriteFinish(
	    ioImg->ioImgHandle
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Skip the [rest of the] image
 *
 * LongDescription:::
 *  IO_ImageSkip() allows the rest of the image (or all of it) to be by-passed.
 *  In the case of MRC, the skip could replace the sequence:
 *  IO_ImageWriteStart, IO_ImageWriteStrip, IO_ImageWriteFinish.  This is
 *  especially useful when a stripe only contains contone imagery, and the
 *  binary layer must be written, but we'd prefer to not write any data.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_AddImage()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ImageSkip(
    IO_IMGHANDLE ioImgHandle
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	return IO_NOSUPPORT;

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the AnnotationOffsets for the current page
 *
 * LongDescription:::
 *  IO_SetAnnotationOffsets() sets the AnnotationOffsets field for the current
 *  PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioPageHandle::		Page handle
 *  dwAnnotationOffsets::	AnnotationOffsets value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetAnnotationOffsets(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 dwAnnotationOffsets
)
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

    /* Check the value for any coding bits */
/*    XERR_GEN_CLEANUP(dwAnnotationOffsets == 0, IO_BADPARAMETER); */

	XERR_PASS_CLEANUP(TFX_SetAnnotationOffsets(
	    ioPage->ioPageHandle,
		dwAnnotationOffsets
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set (write or rewrite) the authoring block
 *
 * LongDescription:::
 *  IO_SetAuthor() sets the authoring information block. If the authoring block
 *  already exists in the file, the block is rewritten to the new block. The
 *  data will overwrite the previous block if the old block was as big or
 *  bigger than the new block, otherwise it is appended to the file.
 *
 * Input:::
 *  ioDocHandle::		Document handle
 *  pAuthorBlock::		Pointer to the authoring block of data
 *  wAuthorBytes::		Number of bytes in authoring block
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetAuthor(
    IO_DOCHANDLE ioDocHandle,
	Int8 *pAuthorBlock,
	UInt16 wAuthorBytes
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

#if 0
	XERR_PASS_CLEANUP(TFX_SetAuthor(
	    ioDoc->ioDocHandle,
		pAuthorBlock,
		wAuthorBytes
	    ));
#endif

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the BitsPerSample for the current image
 *
 * LongDescription:::
 *  IO_SetBitsPerSample() sets the BitsPerSample field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  dwBitsPerSample::		BitsPerSample value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetBitsPerSample(
    IO_IMGHANDLE ioImgHandle,
	UInt16 dwBitsPerSample
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (dwBitsPerSample == 1) {
	NORMAL_RETURN;
    }
#endif

    switch (ioImg->ioPage->ioDoc->ioDocType) {
    case DOCTYPE_XIFF2:
    case DOCTYPE_XIFF3:
	XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	break;
    case DOCTYPE_T44:
    case DOCTYPE_T44A:
    case DOCTYPE_T44B:
	XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	break;
    case DOCTYPE_TIFFFX_S:
    case DOCTYPE_TIFFFX_F:
    case DOCTYPE_TIFFFX_C:
    case DOCTYPE_TIFFFX_M:
    case DOCTYPE_TIFFFX_T:
    case DOCTYPE_TIFFFX_M1:
    case DOCTYPE_COHTIFF_B:
    case DOCTYPE_COHTIFF_C:
    case DOCTYPE_COHTIFF_M:
    case DOCTYPE_COHTIFF_M1:
	XERR_PASS_CLEANUP(TFX_SetBitsPerSample(
	    ioImg->ioImgHandle,
		dwBitsPerSample
	    ));
	break;
    default:
	XERR_GEN_CLEANUP(1, IO_BADPARAMETER);
    }

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ChromaPositioning for the current image
 *
 * LongDescription:::
 *  IO_SetChromaPositioning() sets the ChromaPositioning field for the
 *  current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  chromaPositioning::		ChromaPositioning value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetChromaPositioning(
    IO_IMGHANDLE ioImgHandle,
	UInt16 dwChromaPositioning
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (dwChromaPositioning == 1) {
	NORMAL_RETURN;
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetChromaPositioning(
	    ioImg->ioImgHandle,
		dwChromaPositioning
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ChromaSubSampling for the current image
 *
 * LongDescription:::
 *  IO_SetChromaSubSampling() sets the ChromaSubSampling field for the
 *  current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  dwChromaSubSamplingHor::	Horizontal ChromaSubSampling value to set
 *  dwChromaSubSamplingVer::	Vertical ChromaSubSampling value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetChromaSubSampling(
    IO_IMGHANDLE ioImgHandle,
	UInt16 dwChromaSubSamplingHor,
	UInt16 dwChromaSubSamplingVer
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (dwChromaSubSamplingHor == 2 && dwChromaSubSamplingVer == 2) {
	NORMAL_RETURN;
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetChromaSubSampling(
	    ioImg->ioImgHandle,
		dwChromaSubSamplingHor,
		dwChromaSubSamplingVer
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


#if 0
See comment before TFX_SetCodingMethods().
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the CodingMethods for the current page
 *
 * LongDescription:::
 *  IO_SetCodingMethods() sets the CodingMethods field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioPageHandle::		Page handle
 *  dwCodingMethods::		CodingMethods value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetCodingMethods(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 dwCodingMethods
)
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

    /* Check the value for any coding bits */
    XERR_GEN_CLEANUP(dwCodingMethods == 0, IO_BADPARAMETER);

    switch (ioPage->ioDoc->ioDocType) {
    case DOCTYPE_XIFF2:
    case DOCTYPE_XIFF3:
	XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	break;
    case DOCTYPE_T44:
    case DOCTYPE_T44A:
    case DOCTYPE_T44B:
	XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	break;
    case DOCTYPE_TIFFFX_S:
    case DOCTYPE_TIFFFX_F:
    case DOCTYPE_TIFFFX_C:
    case DOCTYPE_TIFFFX_M:
    case DOCTYPE_TIFFFX_T:
    case DOCTYPE_TIFFFX_M1:
    case DOCTYPE_COHTIFF_B:
    case DOCTYPE_COHTIFF_C:
    case DOCTYPE_COHTIFF_M:
    case DOCTYPE_COHTIFF_M1:
	XERR_PASS_CLEANUP(TFX_SetCodingMethods(
	    ioPage->ioPageHandle,
		dwCodingMethods
	    ));
	break;
    default:
	XERR_GEN_CLEANUP(1, IO_BADPARAMETER);
    }

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the Compression for the current image
 *
 * LongDescription:::
 *  IO_SetCompression() sets the Compression field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  dwCompression::		Compression value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetCompression(
    IO_IMGHANDLE ioImgHandle,
	UInt16 dwCompression
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_SetCompression(
	    ioImg->ioImgHandle,
		dwCompression
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the DateTime string for the current page
 *
 * LongDescription:::
 *  IO_SetDateTime() sets the DateTime field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *  <BR>
 *  Note, the format for the DateTime field is fixed, and should contain
 *  no more bytes than is expected for the string.  Format is:
 *  "YYYY:MM:DD HH:MM:SS".
 *
 * Input:::
 *  ioPageHandle::		Page handle
 *  pDateTime::			DateTime string to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetDateTime(
    IO_PAGEHANDLE ioPageHandle,
	Int8 *pDateTime
)
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

    /* Check parameter */
    XERR_GEN_CLEANUP(pDateTime == NULL, IO_BADPARAMETER);

	XERR_PASS_CLEANUP(TFX_SetDateTime(
	    ioPage->ioPageHandle,
		pDateTime
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the Decode values for the current image
 *
 * LongDescription:::
 *  IO_SetDecode() sets the Decode field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  pdwDecode::			Decode array to set (num/denom pairs)
 *  dwDecodeValues::		Number of Decode values
 *				(numerator/denominator pairs, 2 for L*,
 *				6 for L*a*b*)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetDecode(
    IO_IMGHANDLE ioImgHandle,
	Int32 *pdwDecode,
	UInt32 dwDecodeValues
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* Check the parameter, should be 2 or 6 */
    XERR_GEN_CLEANUP(dwDecodeValues != 2 && dwDecodeValues != 6, IO_BADPARAMETER);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    XERR_GEN_CLEANUP(pdwDecode[1] == 0 && pdwDecode[0] != 0, IO_BADPARAMETER);
    if (pdwDecode[1] * 0 == pdwDecode[0]) {
	XERR_GEN_CLEANUP(pdwDecode[3] == 0, IO_BADPARAMETER);
	if (pdwDecode[3] * 100 == pdwDecode[2]) {
	    if (dwDecodeValues == 2) {
		NORMAL_RETURN;
	    }
	    XERR_GEN_CLEANUP(pdwDecode[5] == 0, IO_BADPARAMETER);
	    if (pdwDecode[5] * -85 == pdwDecode[4]) {
		XERR_GEN_CLEANUP(pdwDecode[7] == 0, IO_BADPARAMETER);
		if (pdwDecode[7] * 85 == pdwDecode[6]) {
		    if (pdwDecode[9] * -75 == pdwDecode[8]) {
			XERR_GEN_CLEANUP(pdwDecode[11] == 0, IO_BADPARAMETER);
			if (pdwDecode[11] * 125 == pdwDecode[10]) {
			    NORMAL_RETURN;
			}
		    }
		}
	    }
	}
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetDecode(
	    ioImg->ioImgHandle,
		pdwDecode,
		dwDecodeValues
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the DocumentName string for the current doc
 *
 * LongDescription:::
 *  IO_SetDocumentName() sets the DocumentName field for the current DOC.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioDocHandle::		Doc handle
 *  pDocumentName::		DocumentName string to set
 *  dwDocumentNameBytes::	Bytes in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetDocumentName(
    IO_DOCHANDLE ioDocHandle,
	Int8 *pDocumentName,
	UInt32 dwDocumentNameBytes
)
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_SetDocumentName(
	    ioDoc->ioDocHandle,
		pDocumentName,
		dwDocumentNameBytes
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the FaxProfile for the current page
 *
 * LongDescription:::
 *  IO_SetFaxProfile() sets the FaxProfile field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioPageHandle::		Page handle
 *  faxProfile::		FaxProfile value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetFaxProfile(
    IO_PAGEHANDLE ioPageHandle,
	UInt8 faxProfile
)
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (faxProfile == 0) {
	NORMAL_RETURN;
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetFaxProfile(
	    ioPage->ioPageHandle,
		faxProfile
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the FillOrder for the current image
 *
 * LongDescription:::
 *  IO_SetFillOrder() sets the FillOrder field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  wFillOrder::		wFillOrder value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Elizabeth Espinosa-Pardo
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetFillOrder(
    IO_IMGHANDLE ioImgHandle,
	UInt16 wFillOrder
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_SetFillOrder(
	    ioImg->ioImgHandle,
		wFillOrder
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/*****************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the HostComputer for the current image
 *
 * LongDescription:::
 *  IO_SetHostComputer() sets the HostComputer field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::			Image handle
 *  pHostComputer::			HostComputer string to set
 *  dwHostComputerBytes::		Bytes in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  EEP 20000803 based on IO_SetSoftware
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetHostComputer(
    IO_IMGHANDLE ioImgHandle,
	Int8 *pHostComputer,
	UInt32 dwHostComputerBytes
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_SetHostComputer(
	    ioImg->ioImgHandle,
		pHostComputer,
		dwHostComputerBytes
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the IFDBaseColor for the current image
 *
 * LongDescription:::
 *  IO_SetIFDBaseColor() sets the IFDBaseColor field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *  <BR>
 *  Note, IFDBaseColor is only used when the first extension set is being used.
 *  In this case, both the IFDBaseColor and ImageBaseColor may be used, but the
 *  ImageBaseColor field may be omitted if the color image is not interpretable
 *  by a base TIFF-FX reader (for example it has JPEGTables).
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  *pIFDBaseColor::		IFDBaseColor array to set.  This is a color
 *				in the color characteristics described by the
 *				IFD: PhotometricInterpretation, SamplesPerPixel,
 *				BitsPerSample, and if appropriate: Decode,
 *				Indexed and/or ICCProfiles.
 *  dwIFDBaseColorValues::	Number of color values: should be
 *				SamplesPerPixel
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetIFDBaseColor(
    IO_IMGHANDLE ioImgHandle,
	Int32 *pdwIFDBaseColor,
	UInt32 dwIFDBaseColorValues
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* We don't know the color until we're assured what Decode[] is */

	XERR_PASS_CLEANUP(TFX_SetIFDBaseColor(
	    ioImg->ioImgHandle,
		pdwIFDBaseColor,
		dwIFDBaseColorValues
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ImageBaseColor for the current image
 *
 * LongDescription:::
 *  IO_SetImageBaseColor() sets the ImageBaseColor field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  *pImageBaseColor::		ImageBaseColor array to set.  This is an
 *				ITULAB color, 3 items, for L*, a*, and b*.
 *				If the TIFF field Decode exists, then it
 *				specifies the gammut used for this ITULAB
 *				color also; otherwise, the default gammut
 *				applies: L* [0,100], b* [-85,85],
 *				and a* [-75,125].  Note, the Decode gammut
 *				applies even if the PhotometricInterpretation
 *				is NOT ITULAB.
 *  dwImageBaseColorValues::	Number of color values: 1 if L*, 3 if color.
 *				Note, this should be the same as
 *				SamplesPerPixel.
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetImageBaseColor(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pImageBaseColor,
	UInt32 dwImageBaseColorValues
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* We don't know the color until we're assured what Decode[] is */

	XERR_PASS_CLEANUP(TFX_SetImageBaseColor(
	    ioImg->ioImgHandle,
		pImageBaseColor,
		dwImageBaseColorValues
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ImageDescription string for the current image
 *
 * LongDescription:::
 *  IO_SetImageDescription() sets the ImageDescription field for the
 *  current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  pImageDescription::		ImageDescription string to set
 *  dwImageDescriptionBytes::	Bytes in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetImageDescription(
    IO_IMGHANDLE ioImgHandle,
	Int8 *pImageDescription,
	UInt32 dwImageDescriptionBytes
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_SetImageDescription(
	    ioImg->ioImgHandle,
		pImageDescription,
		dwImageDescriptionBytes
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the T88Options for the current image
 *
 * LongDescription:::
 *  IO_SetT88Options() sets the T88Options field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  T88Options::			T88Options value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetT88Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwT88Options
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (dwT88Options == 0) {
	NORMAL_RETURN;
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetT88Options(
	    ioImg->ioImgHandle,
		dwT88Options
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ModeNumber for the current page
 *
 * LongDescription:::
 *  IO_SetModeNumber() sets the ModeNumber field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioPageHandle::		Page handle
 *  modeNumber::		ModeNumber value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetModeNumber(
    IO_PAGEHANDLE ioPageHandle,
	UInt8 modeNumber
)
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */

	XERR_GEN_CLEANUP(1, IO_SEQUENCE);

	XERR_PASS_CLEANUP(TFX_SetModeNumber(
	    ioPage->ioPageHandle,
		modeNumber
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the Orientation for the current image
 *
 * LongDescription:::
 *  IO_SetOrientation() sets the Orientation field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  dwOrientation::		Orientation value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetOrientation(
    IO_IMGHANDLE ioImgHandle,
	UInt16 dwOrientation
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (dwOrientation == 1) {
	NORMAL_RETURN;
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetOrientation(
	    ioImg->ioImgHandle,
		dwOrientation
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the PageNumber for the current page
 *
 * LongDescription:::
 *  IO_SetPageNumber() sets the PageNumber field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *  <BR>
 *  Note, the TIFF PageNumber field has two values, 1: the page offset,
 *  and 2: the page count.
 *
 * Input:::
 *  ioPageHandle::		Page handle
 *  dwPageOffset::		PageOffset value to set
 *  dwPageCount::		PageCount value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetPageNumber(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 dwPageOffset,
	UInt16 dwPageCount
)
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_SetPageNumber(
	    ioPage->ioPageHandle,
		dwPageOffset,
		dwPageCount
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the PhotometricInterpretation for the current image
 *
 * LongDescription:::
 *  IO_SetPhotometricInterpretation() sets the PhotometricInterpretation
 *  field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  dwPhotometricInterpretation::	PhotometricInterpretation value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetPhotometricInterpretation(
    IO_IMGHANDLE ioImgHandle,
	UInt16 dwPhotometricInterpretation
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (dwPhotometricInterpretation == 1) {
	NORMAL_RETURN;
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetPhotometricInterpretation(
	    ioImg->ioImgHandle,
		dwPhotometricInterpretation
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the PlanarConfiguration for the current image
 *
 * LongDescription:::
 *  IO_SetPlanarConfiguration() sets the PlanarConfiguration field for the
 *  current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  wPlanarConfiguration::	PlanarConfiguration value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetPlanarConfiguration(
    IO_IMGHANDLE ioImgHandle,
	UInt16 wPlanarConfiguration
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (wPlanarConfiguration == 1) {
	NORMAL_RETURN;
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetPlanarConfiguration(
	    ioImg->ioImgHandle,
		wPlanarConfiguration
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ProfileType for the current page
 *
 * LongDescription:::
 *  IO_SetProfileType() sets the ProfileType field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioPageHandle::		Page handle
 *  dwProfileType::		ProfileType value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetProfileType(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 dwProfileType
)
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (dwProfileType == 0) {
	NORMAL_RETURN;
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetProfileType(
	    ioPage->ioPageHandle,
		dwProfileType
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set (write or rewrite) the reference page offset
 *
 * LongDescription:::
 *  IO_SetReferencePage() sets the reference page offset. If the reference
 *  page is already stored in the file, it is overwritten.
 *
 * Input:::
 *  ioDocHandle::		Document handle
 *  dwReferencePageOffset::	Reference page offset
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetReferencePage(
    IO_DOCHANDLE ioDocHandle,
	Int32 dwReferencePageOffset
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_SetReferencePage(
	    ioDoc->ioDocHandle,
		dwReferencePageOffset
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ResolutionUnit for the current image
 *
 * LongDescription:::
 *  IO_SetResolutionUnit() sets the ResolutionUnit field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  dwResolutionUnit::		ResolutionUnit value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetResolutionUnit(
    IO_IMGHANDLE ioImgHandle,
	UInt16 dwResolutionUnit
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (dwResolutionUnit == 2) {
	NORMAL_RETURN;
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetResolutionUnit(
	    ioImg->ioImgHandle,
		dwResolutionUnit
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the RowsPerStrip for the current image
 *
 * LongDescription:::
 *  IO_SetRowsPerStrip() sets the RowsPerStrip field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  dwRowsPerStrip::		RowsPerStrip value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetRowsPerStrip(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwRowsPerStrip
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_SetRowsPerStrip(
	    ioImg->ioImgHandle,
		dwRowsPerStrip
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the SamplesPerPixel for the current image
 *
 * LongDescription:::
 *  IO_SetSamplesPerPixel() sets the SamplesPerPixel field for the current
 *  image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  samplesPerPixel::		SamplesPerPixel array to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetSamplesPerPixel(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwSamplesPerPixel
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (dwSamplesPerPixel == 1) {
	NORMAL_RETURN;
    }
#endif

    switch (ioImg->ioPage->ioDoc->ioDocType) {
    case DOCTYPE_XIFF2:
    case DOCTYPE_XIFF3:
	XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	break;
    case DOCTYPE_T44:
    case DOCTYPE_T44A:
    case DOCTYPE_T44B:
	XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	break;
    case DOCTYPE_TIFFFX_S:
    case DOCTYPE_TIFFFX_F:
    case DOCTYPE_TIFFFX_C:
    case DOCTYPE_TIFFFX_M:
    case DOCTYPE_TIFFFX_T:
    case DOCTYPE_TIFFFX_M1:
    case DOCTYPE_COHTIFF_B:
    case DOCTYPE_COHTIFF_C:
    case DOCTYPE_COHTIFF_M:
    case DOCTYPE_COHTIFF_M1:
	XERR_PASS_CLEANUP(TFX_SetSamplesPerPixel(
	    ioImg->ioImgHandle,
		dwSamplesPerPixel
	    ));
	break;
    default:
	XERR_GEN_CLEANUP(1, IO_BADPARAMETER);
    }

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the Software for the current image
 *
 * LongDescription:::
 *  IO_SetSoftware() sets the Software field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  pSoftware::			Software string to set
 *  dwSoftwareBytes::		Bytes in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetSoftware(
    IO_IMGHANDLE ioImgHandle,
	Int8 *pSoftware,
	UInt32 dwSoftwareBytes
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_SetSoftware(
	    ioImg->ioImgHandle,
		pSoftware,
		dwSoftwareBytes
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the TargetPrinter for the current page
 *
 * LongDescription:::
 *  IO_SetTargetPrinter() sets the TargetPrinter field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioPageHandle::		Page handle
 *  pTargetPrinter::		TargetPrinter string to set
 *  dwTargetPrinterBytes::	Bytes in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetTargetPrinter(
    IO_PAGEHANDLE ioPageHandle,
	Int8 *pTargetPrinter,
	UInt32 dwTargetPrinterBytes
)
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_SetTargetPrinter(
	    ioPage->ioPageHandle,
		pTargetPrinter,
		dwTargetPrinterBytes
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the Thresholding string for the current image
 *
 * LongDescription:::
 *  IO_SetThresholding() sets the Thresholding field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  wThresholding::		Thresholding value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetThresholding(
    IO_IMGHANDLE ioImgHandle,
	UInt16 wThresholding
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_SetThresholding(
	    ioImg->ioImgHandle,
		wThresholding
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the T4Options for the current image
 *
 * LongDescription:::
 *  IO_SetT4Options() sets the T4Options field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  t4Options::			T4Options value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetT4Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwT4Options
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (dwT4Options == 0) {
	NORMAL_RETURN;
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetT4Options(
	    ioImg->ioImgHandle,
		dwT4Options
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the T6Options for the current image
 *
 * LongDescription:::
 *  IO_SetT6Options() sets the T6Options field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  t6Options::			T6Options value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetT6Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 dwT6Options
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (dwT6Options == 0) {
	NORMAL_RETURN;
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetT6Options(
	    ioImg->ioImgHandle,
		dwT6Options
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the VersionYear byte-string for the current image
 *
 * LongDescription:::
 *  IO_SetVersionYear() sets the VersionYear field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  pVersionYear::		VersionYear byte-string to set.  Note, this
 *				is NOT an ASCII string, but an actual
 *				byte-value sequence.
 *  dwVersionYearBytes::	Byte values in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetVersionYear(
    IO_IMGHANDLE ioImgHandle,
	UInt8 *pVersionYear,
	UInt32 dwVersionYearBytes
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_SetVersionYear(
	    ioImg->ioImgHandle,
		pVersionYear,
		dwVersionYearBytes
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the YCbCrCoefficients for the current image
 *
 * LongDescription:::
 *  IO_SetYCbCrCoefficients() sets the YCbCrCoefficients field for the
 *  current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  pdwYCbCrCoefficients::	YCbCrCoefficients value to set
 *  dwYCbCrCoefficientValues::	Number of values
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetYCbCrCoefficients(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwYCbCrCoefficients,
	UInt32 dwYCbCrCoefficientValues
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

    /* I could intercept the default value, and do nothing if found, but they
     * are probably asking for the field to be written */
#if 0
    if (dwResolutionUnit == 2) {
	NORMAL_RETURN;
    }
#endif

	XERR_PASS_CLEANUP(TFX_SetYCbCrCoefficients(
	    ioImg->ioImgHandle,
		pdwYCbCrCoefficients,
		dwYCbCrCoefficientValues
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the Predictor for the current image
 *
 * LongDescription:::
 *  IO_SetPredictor() sets the Predictor field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  wPredictor::		Predictor value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetPredictor(
    IO_IMGHANDLE ioImgHandle,
	UInt16 wPredictor
)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_SetPredictor(
	    ioImg->ioImgHandle,
		wPredictor
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the number of entries each annotation table will contain.
 *
 * LongDescription:::
 *  IO_SetAnnotTableEntryCount() sets the number of annotation table entries
 *  to be used, before the annotation table header is actually written. This
 *  routine can be called at any reasonable time, but the count passed in will
 *  only apply to new annotation table headers created, and only when an
 *  annotation is actually saved.
 *
 * Input:::
 *  ioDoc::			Data structure of open (just created) document
 *  wEntriesPerTable::		Number of entries per table to use
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetAnnotTableEntryCount(
    IO_DOCHANDLE ioDocHandle,
	UInt16 wEntriesPerTable
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_SetAnnotTableEntryCount(
	    ioDoc->ioDocHandle,
		wEntriesPerTable
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the number of entries each page table will contain.
 *
 * LongDescription:::
 *  IO_SetPageTableEntryCount() sets the number of page table entries
 *  to be used, before the page table header is actually written. It is
 *  best to call this routine just after the call to IO_CreateDocument(),
 *  although you should be able to call it any time before the first page
 *  is actually started.
 *
 * Input:::
 *  ioDoc::			Data structure of open (just created) document
 *  wEntriesPerTable::		Number of entries per table to use
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetPageTableEntryCount(
    IO_DOCHANDLE ioDocHandle,
	UInt16 wEntriesPerTable
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the number of entries each shared resource table will contain.
 *
 * LongDescription:::
 *  IO_SetResourceTableEntryCount() sets the number of shared resource table
 *  entries to be used, before the shared resource table header is actually
 *  written.
 *  It is best to call this routine just after the call to IO_CreateDocument(),
 *  although you should be able to call it any time before the first token
 *  dictionary is actually written (up to the call to IO_WriteTokenImage(),
 *  IO_WriteTokenDictionary(), IO_WriteTokenImages(), or
 *  IO_CloseDocument()).
 *
 * Input:::
 *  ioDoc::			Data structure of open (just created) document
 *  wEntriesPerTable::		Number of entries per table to use
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_SetResourceTableEntryCount(
    IO_DOCHANDLE ioDocHandle,
	UInt16 wEntriesPerTable
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_SetResourceTableEntryCount(
	    ioDoc->ioDocHandle,
		wEntriesPerTable
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add an annotation to a page
 *
 * LongDescription:::
 *  IO_AddAnnotation() adds an annotation to a page
 *
 * Input:::
 *  ioPageHandle::		Active page handle
 *  wAnnotType::		Type of annotation
 *  dwXOffset::			Offset in X direction in page coordinates
 *  dwYOffset::			Offset in Y direction in page coordinates
 *  dwWidth::			Width in page coordinates
 *  dwHeight::			Height in page coordinates
 *  pBuf::			Annotation data
 *  dwAnnotBytes::		Length of annotation data, in bytes
 *
 * Output:::
 *  pdwAnnotID::		Annotation ID assigned to this annotation
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_AddAnnotation(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 wAnnotType,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt32 dwFlags,
	UInt8 *pBuf,
	UInt32 dwAnnotBytes,
	UInt16 *pdwAnnotID
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_AddAnnotation(
	    ioPage->ioPageHandle,
		wAnnotType,
		dwXOffset,
		dwYOffset,
		dwWidth,
		dwHeight,
		dwFlags,
		pBuf,
		dwAnnotBytes,
		pdwAnnotID
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Deletes an annotation from a page
 *
 * LongDescription:::
 *  IO_DeleteAnnotation() deletes an annotation from a page. Note that the
 *  remaining subsequent annotations will automatically shift down one
 *  position.
 *
 * Input:::
 *  ioPageHandle::		Page handle, returned from IO_GetPageInfo()
 *  dwAnnotOffset::		Offset (or ID) of annotation to delete (0, 1, ...)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_DeleteAnnotation(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 dwAnnotOffset
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_DeleteAnnotation(
	    ioPage->ioPageHandle,
		dwAnnotOffset
	    ));

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Use color strips, instead of compressed strips
 *
 * LongDescription:::
 *  IO_UseColorStrips() sets the output image into a mode where strips of a
 *  solid color are marked with a single color, instead of being compressed.
 *  In such cases, if ImageBaseColor of IFDBaseColor has been set, then only
 *  strips with the solid color that matches IFDBaseColor (or ImageBaseColor)
 *  will be left uncompressed (0 byte strips).  Otherwise, the first strip
 *  with a solid color will set the precedence for the color marked for the
 *  entire image.
 *  <BR>
 *  Note, if neither IFDBaseColor nor ImageBaseColor have been set before
 *  TFX_ImageWriteStart() is called, the IFD will be rewritten when the
 *  document is closed; so that the new field will be added.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_AddImage()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data to this
 *  buffer as needed, and writing the intermediate buffer when filled.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_UseColorStrips(
    IO_IMGHANDLE ioImgHandle
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_UseColorStrips(
	    ioImg->ioImgHandle
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Use multi-strip SubIFD feature
 *
 * LongDescription:::
 *  IO_UseMultiStripSubIFDs() requests that the document use single multi-strip
 *  SubIFDs, when adding images that span across stripe boundaries.
 *  If images in the page have already been added, then they are not affected.
 *  Only images added to the page after this call will be affected.<BR>
 *  There is currently no way of turning this feature off for a page, once it
 *  has been turned on.
 *
 * Input:::
 *  ioPageHandle::		Valid page handle, as returned from
 *				IO_AddPage()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_UseMultiStripSubIFDs(
    IO_PAGEHANDLE ioPageHandle
    )
{
    IO_PAGE *ioPage = (IO_PAGE *)ioPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioPageHandle);

	XERR_PASS_CLEANUP(TFX_UseMultiStripSubIFDs(
	    ioPage->ioPageHandle
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do not write the GPIFD field
 *
 * LongDescription:::
 *  IO_UseNoGPIFD() requests that the document NOT have the GPIFD written.
 *  If the first IFD has yet to be written, and there's no need for a GPIFD,
 *  then it won't be written.  If later the GPIFD is required, the IFD of the
 *  first page will be rewritten with the GPIFD included.<BR>
 *  If the first IFD is already written, then the GPIFD is NOT removed from it.
 *
 * Input:::
 *  ioDocHandle::		Valid document handle, as returned from
 *				IO_CreateDocument() or
 *				IO_OpenDocumentReadWrite()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data to this
 *  buffer as needed, and writing the intermediate buffer when filled.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_UseNoGPIFD(
    IO_DOCHANDLE ioDocHandle
    )
{
    IO_DOC *ioDoc = (IO_DOC *)ioDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioDocHandle);

	XERR_PASS_CLEANUP(TFX_UseNoGPIFD(
	    ioDoc->ioDocHandle
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Use variable strip heights
 *
 * LongDescription:::
 *  IO_UseVariableStripHeights() signals an image to be stored using variable
 *  strip heights (StripRowCounts), instead of fixed strip heights.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_AddImage()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data to this
 *  buffer as needed, and writing the intermediate buffer when filled.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_UseVariableStripHeights(
    IO_IMGHANDLE ioImgHandle
    )
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_UseVariableStripHeights(
	    ioImg->ioImgHandle
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add Rendering Hints to an image
 *
 * LongDescription:::
 *  IO_AddRenderingHints() adds the rendering hints dictionary (an ASCII string)
 *  to an image.  The format of the dictionary is spelled out in the XCTIFF
 *  specification, or some related documentation.
 *
 * Input:::
 *  ioImgHandle::		Valid image handle, as returned from
 *				IO_AddImage()
 *  pHintsString::		Hint dictionary, as an ASCII string
 *  dwHintLength::		Length of pHintString
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data to this
 *  buffer as needed, and writing the intermediate buffer when filled.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_AddRenderingHints(
    IO_IMGHANDLE ioImgHandle,
	char *pHintsString,
	UInt32 dwHintLength
	)
{
    IO_IMG *ioImg = (IO_IMG *)ioImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_AddRenderingHints(
	    ioImg->ioImgHandle,
	    pHintsString,
	    dwHintLength
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add a hints image to an image
 *
 * LongDescription:::
 *  IO_AddHintsImage() adds a new image to a parent image, to be added as the
 *  rendering hints IFD for the image layer of the parent.
 *  <BR>
 *  Note that the pixel values stored correspond to a hint defined in the hints
 *  dictionary -- added from IO_AddRenderingHints().
 *  <BR>
 *  An image handle is returned, that is used in later calls to three
 *  functions: IO_ImageWriteStart(), IO_ImageWriteStrip() or
 *  IO_CompressedWriteStrip(), and IO_ImageWriteFinish().
 *
 * Input:::
 *  ioImgHandle::		Parent image to attach to
 *  dwBpp::			Bits per pixel
 *  dwXOffset::			X pixel offset in Primary image coordinates.
 *				This should be the same as the parent.
 *  dwYOffset::			Y pixel offset in Primary image coordinates.
 *				This should be the same as the parent.
 *  dwXResolution::		Resolution in the X direction.
 *				This should be the same as the parent.
 *  dwYResolution::		Resolution in the Y direction.
 *				This should be the same as the parent.
 *  dwWidth::			Width.
 *				This should be the same as the parent.
 *  dwHeight::			Height.
 *				This should be the same as the parent.
 *  dwBestStripHeight::		Height of strips stored
 *
 * Output:::
 *  pImgHandle::		Pointer to image handle, used to later
 *				reference this image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_AddHintsImage(
	IO_IMGHANDLE ioImgHandle,	/* Parent image to attach to */
	UInt32 dwBpp,			/* Image depth, in bits/pixel */
	UInt32 dwXOffset,		/* X location in page coordinates */
	UInt32 dwYOffset,		/* Y location in page coordinates */
	UInt32 dwXResolution,		/* X Resolution of image in dpi */
	UInt32 dwYResolution,		/* Y Resolution of image in dpi */
	UInt32 dwWidth,			/* Width in image's X resolution units
					 */
	UInt32 dwHeight,		/* Height in image's Y resolution units
					 */
	UInt32 dwBestStripHeight,	/* Height of strips stored */
	IO_IMGHANDLE *pioImgHandle	/* Returned image handle for image to
					 * be added */
)
{
    UInt32 newImgHandle = 0UL;
    IO_IMG *ioParentImg = (IO_IMG *)ioImgHandle;
    IO_IMG *ioImg = NULL;
    IO_INSTANCE *instData = NULL;
    PROLOGUE;

    /* Clear the result */
    *pioImgHandle = 0;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioImgHandle);

	XERR_PASS_CLEANUP(TFX_AddHintsImage(
	    ioParentImg->ioImgHandle,
		dwBpp,
		dwXOffset,
		dwYOffset,
		dwXResolution,
		dwYResolution,
		dwWidth,
		dwHeight,
		dwBestStripHeight,
		&newImgHandle
	    ));

    /* Create our own copy of the container */
    instData = IO_IMGINST(ioParentImg);
    IO_CHECKMALLOC(ioImg = (IO_IMG *)IO_MALLOC(sizeof(*ioImg)));
    memset(ioImg, 0, sizeof(*ioImg));
    ioImg->handleID = (UInt32)ioImg;	/* Self identity */
    ioImg->handleType = IO_HANDLE_TYPE;
    
    /* Link back to it's parent */
    ioImg->ioPage = ioParentImg->ioPage;

    /* Identify the lower level handle */
    ioImg->ioImgHandle = newImgHandle;

    /* Add the image to the page. Note: the images are in the order they
     * were added, and not the sequence dictated by image additions.  The
     * lower level routines maintain image/layer order. */
    if (ioParentImg->ioPage->ioImgs == NULL) {
	/* Start the page with this image */
	ioParentImg->ioPage->ioImgs = ioImg;
    }
    else {
	IO_IMG *tempImg;
	
	/* Put it into the linked list at the end */
	tempImg = ioParentImg->ioPage->ioImgs;
	while (tempImg->next != NULL) {
	    tempImg = tempImg->next;
	}
	    
	tempImg->next = ioImg;
    }

    /* Return the result */
    *pioImgHandle = ioImg->handleID;

    NORMAL_RETURN;

ERR_CLEANUP:

    if (ioImg) {
	IO_FREE(ioImg);
    }
    ERR_RETURN;
}

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reuse the color tokenized image data for both the mask and color layers.
 *
 * LongDescription:::
 *  IO_ShareColorTokenImage() copies the token image IFD for an image in an
 *  even layer (mask layer, already stored) to a matching image in the layer
 *  just above (foreground layer) image, and shares the compressed image data
 *  (both size and location); then resets the byte count of the compressed
 *  image in the even layer image to the binary subset, described by the byte
 *  count passed in (second parameter).
 *  <BR>
 *  Note that the simple case is layers 2 and 3, where the color tokenized
 *  stream is first stored in layer 2 as a binary image, but with the color
 *  section included. Then the correct size of the binary portion is passed
 *  in to IO_ShareColorTokenImage(), which then clones the IFD to make a
 *  layer 3 IFD (modifying the color channel count and photometric), points
 *  to the exact same data in it's entirety, then modifies the layer 2 byte
 *  count.
 *  <BR>
 *  Ya, a bit hokey, but it lets you keep the previous way of doing things,
 *  with a minimal number of extra calls.
 *
 * Input:::
 *  ioBinaryImgHandle::		Valid image handle of the completed binary
 *				image.  This image should have the image data
 *				already written, as a <B>single strip</B>.
 *  dwBinaryImageBytes::	Byte count of compressed image data (already
 *				stored in IFD of even layer) of the binary
 *				subsection. The current byte count will
 *				be relayed to the IFD of the next layer,
 *				and this value will be written over the
 *				current value.
 *  xColorType::		Color type of color tokens store in odd
 *				layer IFD
 *  pIoImgHandle::		Resulting image handle
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
IO_ShareColorTokenImage(
	IO_IMGHANDLE ioBinaryImgHandle,
	UInt32 dwBinaryImageBytes,
	IO_COLORTYPE xColorType,
	IO_IMGHANDLE *pIoImgHandle
)
{
    IO_IMG *ioImg = (IO_IMG *)ioBinaryImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    IO_CheckHandleValidity(ioBinaryImgHandle);

	XERR_PASS_CLEANUP(TFX_ShareColorTokenImage(
	    ioImg->ioImgHandle,
		dwBinaryImageBytes,
		xColorType,
		pIoImgHandle
	    ));

    NORMAL_RETURN;
    
ERR_CLEANUP:

    ERR_RETURN;
}



/*
$Log: iowmrc.c,v $
Revision 1.2  2003/02/10 22:53:30  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:10  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.32  2001/12/26 21:23:42  dabercro
Mark as released

Revision 1.31  2001/11/07 21:43:06  dabercro
Make base color signed long, to accomodate CIELAB

Revision 1.30  2001/10/30 20:06:02  dabercro
Resolve gcc complaint

Revision 1.29  2001/09/18 21:10:09  dabercro
Make work with JBIG2

Revision 1.28  2001/08/22 19:10:54  dabercro
Added IFDBaseColor

Revision 1.27  2001/05/03 16:39:08  dabercro
Added a new function UseMultiStripSubIFDs.

Revision 1.26  2001/05/02 20:34:31  dabercro
Comment fixes

Revision 1.25  2001/05/01 00:42:57  dabercro
Add function IO_UseNoGPIFD.  Change planar configuration to short.
Added comments.

Revision 1.24  2001/04/02 18:25:11  dabercro
Add HostComputer.

Revision 1.23  2001/03/22 20:16:39  dabercro
Add support for new TIFF field: Predictor.

Revision 1.22  2001/03/22 17:33:52  dabercro
Add T.44 (no appendix) type.

Revision 1.21  2001/02/24 03:23:49  dabercro
Use char* for string to rendering hints.

Revision 1.20  2001/01/18 00:21:28  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.19  2000/11/18 00:53:25  dabercro
Avoid cross polenation of data handles across sublibraries
(XIF, MRC, TFX), especially for CopyImage.  Added a handle
type to structure, now checked also.

Revision 1.18  2000/11/17 23:57:34  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.17  2000/11/02 22:21:21  dabercro
Additions for Rendering Hints.

Revision 1.16  2000/10/19 17:59:10  dabercro
Return the FG image handle when adding an image for the
color JBIG2 component.

Revision 1.15  2000/10/19 17:12:28  dabercro
Finished support for color JBIG2 component.

Revision 1.14  2000/10/16 18:22:53  dabercro
Major adds to support stripes, 0-byte strips, and ImageBaseColor.

Revision 1.13  2000/09/27 21:10:35  dabercro
Added function to force variable strip heights.

Revision 1.12  2000/08/29 22:00:17  dabercro
Use right instance.

Revision 1.11  2000/08/25 23:46:19  dabercro
Additions to allow marking annotations for write.

Revision 1.10  2000/08/22 01:03:23  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.9  2000/08/04 21:54:10  dabercro
Make provisions for read/write mode.  Had roughed it in with
read-only code.

Revision 1.8  2000/08/02 00:19:59  dabercro
Keep handles around in a linked-list.

Revision 1.7  2000/07/31 23:51:43  dabercro
Name change from JBIG2Options to T88Options.

Revision 1.6  2000/06/06 23:47:11  dabercro
Continued work, lots of stuff.

Revision 1.5  2000/05/09 01:54:07  dabercro
Added FillOrder field support.
Support for JBIG2 dictionary extensions, updated parameter to
reflect byte count for extensions.
Write fields that are explicitly set, regardless of default values.

Revision 1.4  2000/04/12 19:06:07  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.3  2000/03/11 02:06:34  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.2  2000/02/08 21:25:32  dabercro
Disable JBIG2 (for now), updates to make use of GlobalParametersIFD for symbol libraries.

Revision 1.1.1.1  2000/02/02 00:40:50  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

*/
