/*
 * Trade secret of Xerox Corporation.
 * Copyright (c) 1991-2001, Xerox Corporation.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */


#ifndef TFX_H_INCLUDED
#define TFX_H_INCLUDED

#ifndef NOIPCORE
#include "ansiprot.h"

IP_RCSINFO(tfx_h_RCSInfo, "$RCSfile: tfx.h,v $; $Revision: 1.1 $");
/* $Date: 2003/01/03 01:41:10 $ */
#endif

/* tfx.h
 *
 * DESCRIPTION
 *	 Declarations for the TIFF-FX Toolkit API.
 *
 */


#ifdef __cplusplus
extern "C"
{
#endif

#ifdef _MSC_VER
#pragma pack(push, tfx_h, 1)
#endif

/*
 * INCLUDES
 */
#include "iomrc.h"

/*
 * CONSTANTS
 */

/* Current TIFF-FX version: 7.0
 */
#define CURRENT_TFX_VERSION	7	/* Major version number */
#define CURRENT_TFX_REVISION	0	/* Minor version number */

/* Coherent TIFF/TIFF-FX profiles, used in "options"  */
#define TFX_PROFILE_UNKNOWN	0	/* Profile S: standard B&W */
#define TFX_PROFILE_S		1	/* Profile S: standard B&W */
#define TFX_PROFILE_F		2	/* Profile F: Fax B&W */
#define TFX_PROFILE_J		3	/* Profile J: Lossless JBIG (JBIG1) */
#define TFX_PROFILE_C		4	/* Profile C: Contone */
#define TFX_PROFILE_L		5	/* Profile L: Lossless color (JBIG1) */
#define TFX_PROFILE_M		6	/* Profile M: MRC */
#define TFX_PROFILE_T		7	/* Profile T: Tokenization (JBIG2) */
#define TFX_PROFILE_M1		8	/* Profile M1: MRC + JBIG2 + color tags? */
#define TFX_PROFILE_COH_B	10	/* Coherent TIFF Profile B: Binary */
#define TFX_PROFILE_COH_C	11	/* Coherent TIFF Profile C: Contone */
#define TFX_PROFILE_COH_M	12	/* Coherent TIFF Profile M: MRC */
#define TFX_PROFILE_COH_M1	13	/* Coherent TIFF Profile M1: MRC + JBIG2 ??? */
#define TFX_PROFILE_MULTIPLE	14	/* Yet to be implemented, each page may have
					 * it's own profile */
#define TFX_PROFILES		0xff	/* Profile mask, must be UInt8 for
					 * docType->docInfo.profile */

/* MultiProfile bits */
#define TFX_MULTIPROFILE_S	0x00000001	/* Bit 0: minimal black & white lossless, Profile S */
#define TFX_MULTIPROFILE_F	0x00000002	/* Bit 1: extended black & white lossless, Profile F */
#define TFX_MULTIPROFILE_J	0x00000004	/* Bit 2: lossless JBIG black & white, Profile J */
#define TFX_MULTIPROFILE_C	0x00000008	/* Bit 3: lossy color and grayscale, Profile C */
#define TFX_MULTIPROFILE_L	0x00000010	/* Bit 4: lossless color and grayscale, Profile L*/
#define TFX_MULTIPROFILE_M	0x00000020	/* Bit 5: Mixed Raster Content, Profile M */
#define TFX_MULTIPROFILE_T	0x00000040	/* Bit 6: lossy and lossless JBIG2 black & white, Profile T */
#define TFX_MULTIPROFILE_E1	0x00000080	/* Bit 7: Extension 1, resolution and imagewidth extensions */
#define TFX_MULTIPROFILE_E2	0x00000100	/* Bit 8: Extension 2, N-Layer Profile M extension */
#define TFX_MULTIPROFILE_E3	0x00000200	/* Bit 9: Extension 3, shared data extension */
#define TFX_MULTIPROFILE_E5	0x00000400	/* Bit 10: Extension 5, JBIG2 extension of Profile M */


#if 0
/* Doc tags (TIFF-FX extension tags)
 */
#define DOCTAG_NULL		0	/* Placeholder for future tag */
#define DOCTAG_PAGETABLE	1	/* First page table location */
#define DOCTAG_AUTHOR		2	/* Author block */
#define DOCTAG_REFERENCEPAGE	3	/* Reference page */
#define DOCTAG_NEXTEXT		4	/* Next set of document tags */
#define DOCTAG_SYMLIBTABLE	5	/* First symbol library */
#define DOCTAG_COUNT		6	/* One more than last one */

typedef unsigned long int	DocTags;
#endif

/* Annotation flags */
#define ANNOT_NOTVIEWABLE	0x00000001	/* Not viewable */
#define ANNOT_HYPERLINK		0x00000002	/* Annotation also hyperlinks
						 * to either another spot in
						 * the document (a page), or
						 * an URL. */


/*
 * MACROS
 */

/*
 * TYPEDEFS
 */
#if !defined(_TYPES_PUB_INCLUDED) && !defined(CDECL)
typedef unsigned char      	UInt8;
typedef unsigned short int 	UInt16;
typedef unsigned long int  	UInt32;
typedef signed char		Int8;
typedef short int		Int16;
typedef long int		Int32;
typedef unsigned long int	Bool;
typedef unsigned long int	ErrCode;
typedef float			Float32;
typedef double			Float64;
#define CDECL
#define PROLOGUE		ErrCode __error__ = IO_NOERROR
#endif


/*
 * TIFF-FX basic types
 *
 * For portability, the TIFF-FX library uses platform-independent type names.
 * The definitions of the type will be in a single platform-dependent
 * section here.
 */
#if defined(sparc)
#define signed
#endif


/*
 ******************************************************************************
 *				FUNCTION PROTOTYPES			      *
 ******************************************************************************
 */

/*
 * TIFF-FX progress function. The progress function is selected by the client.
 * It is called periodically by TIFF-FX functions to report progress and offer
 * the client an opportunity to cancel. The meaning of the progress function
 * depends on the function being carried out. If not otherwise documented,
 * the progress value will be -1, indicating no meaningful progress to report, 
 * or a number from 0 to 100, indicating percent complete.
 */
ErrCode CDECL
TFX_SetProgressFunc(
	IO_INSTHANDLE xInstHandle,
	IO_PROGRESSFUNC xProgress);


/*
 * TIFF-FX memory allocation function. These function allows the client to
 * override TIFF-FX's memory allocation functions. It is not required, and
 * may not be available on all platforms.
 *
 * Once the functions have been overridden, they can be restored to
 * their defaults by passing NULL for each function.
 *
 * 'hasPtrs' is non-zero when allocating an array of pointers. On some
 * systems, this indicates that the memory management must update the
 * contents of this array, should the memory pointed to by this array
 * move. It is a garbage collection feature available to the client.
 * >>> However, none of the calls made by any of the functions make use
 * >>> of this (hasPtrs always equals 0).
 */
/*
 * TFX_SetMallocFuncs
 */
ErrCode CDECL
TFX_SetMallocFuncs(
	IO_INSTHANDLE xInstHandle,
	void *dwClientMallocID,
	IO_MALLOCFUNC xMalloc,
	IO_FREEFUNC xFree
);


/*
 * TFX_InitInstance
 */
ErrCode CDECL
TFX_InitInstance(
	UInt32 dwClientInstID,
	IO_INSTHANDLE *pxInst
);


/*
 * TFX_GetInstance
 */
ErrCode CDECL
TFX_GetInstance(
    IO_DOCHANDLE tfxDocHandle,
    IO_INSTHANDLE *pxInstHandle
);


/*
 * TFX_EndInstance
 */
ErrCode CDECL
TFX_EndInstance(
	IO_INSTHANDLE xInstHandle
);


/*
 * TFX_GetClientID
 */
ErrCode CDECL
TFX_GetClientID(
	IO_INSTHANDLE xInstHandle,
	UInt32 *dwClientInstID
);


/*
 * TFX_IsTfx
 */
ErrCode CDECL
TFX_IsTfx(
	IO_INSTHANDLE xInstHandle,
	Int32 dwClientID,
	IO_READFUNC xReadFunc,
	IO_SEEKFUNC xSeekFunc,
	IO_SIZEFUNC xSizeFunc,
	UInt8 *pVersion,
	UInt8 *pRevision,
	UInt32 *pdwOptions
);


/*
 * TFX_OpenDocumentRead
 */
ErrCode CDECL
TFX_OpenDocumentRead(
	IO_INSTHANDLE xInstHandle,
	Int32 dwClientID,
	UInt32 options,
	IO_READFUNC xReadFunc,
	IO_WRITEFUNC xWriteFunc,
	IO_SEEKFUNC xSeekFunc,
	IO_SIZEFUNC xSizeFunc,
	IO_SIZEFUNC xTellFunc,
	IO_DOCHANDLE *pTfxDocHandle
);


/*
 * TFX_CloseDocument
 */
ErrCode CDECL
TFX_CloseDocument(
    IO_DOCHANDLE tfxDocHandle
);


/*
 * TFX_GetDocInfo
 */
ErrCode CDECL
TFX_GetDocInfo(
	IO_DOCHANDLE tfxDocHandle,	/* Input document handle */
/* Returned values */
	UInt8 *pVersion,		/* Version of TIFF-FX */
	UInt8 *pRevision,		/* Revision of this version of TIFF-FX */
	UInt16 *pwNPages		/* Number of pages in file */
);


/*
 * TFX_GetPageInfo
 */
ErrCode CDECL
TFX_GetPageInfo(
	IO_DOCHANDLE tfxDocHandle,	/* Input document handle */
	Int32 dwPageOffset,		/* Page offset (starting with 0) */
/* Returned values */
	UInt32 **ppdwImageCounts,	/* Number of images within each layer
					 */
	UInt32 *pdwLayersPresent,	/* Number of layers present (although
					 * some layers may be missing) */
	UInt16 *pwAnnotationCount,	/* Number of annotations on page */
	IO_PAGEHANDLE *pTfxPageHandle	/* Page handle, used to identify this
					 * page in the page level routines */
);


/*
 * TFX_GetImageInfo
 */
ErrCode CDECL
TFX_GetImageInfo(
	IO_PAGEHANDLE tfxPageHandle,	/* Page image handle */
	UInt32 dwImgLayer,		/* Image layer in the MRC */
	UInt32 dwImgOffset,		/* Image offset (for multiple images)
					 */
/* Returned values */
	UInt32 *pdwXOffset,		/* X location in page coordinates */
	UInt32 *pdwYOffset,		/* Y location in page coordinates */
	UInt32 *pdwXResolution,		/* X Resolution of image in dpm */
	UInt32 *pdwYResolution,		/* Y Resolution of image in dpm */
	UInt32 *pdwWidth,		/* Width in image's X resolution units
					 */
	UInt32 *pdwHeight,		/* Height in image's Y resolution units
					 */
	UInt32 *pdwBestStripHeight,	/* For best performance, read strips
					 * with height a multiple of this. */
	IO_COMPRESSION *pxCompression,	/* Compression type for src/dst */
	IO_COLORTYPE *pxColorType,	/* Color type (see IO_COLORTYPE) */
	UInt8 *pPalette,		/* Color palette if IO_COLORTYPE is:
					 * IO_COLORTYPE_COLOR4 or
					 * IO_COLORTYPE_COLOR8.
					 * Palette is RGB triples (chunky):
					 * RGBRGB... */
	IO_IMGHANDLE *pTftfxImgHandle	/* Returned image handle */
);


/*
 * TFX_GetImageStats
 */
ErrCode CDECL
TFX_GetImageStats(
	IO_IMGHANDLE tfxImgHandle,	/* Image handle */
/* Returned values */
	UInt32 *pdwCmpSize,		/* Compressed image byte total */
	UInt32 *pdwDictCount,		/* Number of token dictionaries */
	UInt32 **ppdwDictIDs,		/* Dictionary ID list (if tokenized) */
	Int32 **ppdwDictSizes		/* Dictionary byte list (tokenized) */
);


/*
 * TFX_GetHintsImageInfo
 */
ErrCode CDECL
TFX_GetHintsImageInfo(
	IO_IMGHANDLE tfxParentImgHandle,
	UInt32 dwImgOffset,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	IO_COMPRESSION *pxCompression,
	IO_IMGHANDLE *pTfxImgHandle
);


/*
 * TFX_GetDocProfile
 */
ErrCode CDECL
TFX_GetDocProfile(
	IO_DOCHANDLE tfxDocHandle,
	DOCTYPE *pdwProfile
);


/*
 * TFX_ImageReadStart
 */
ErrCode CDECL
TFX_ImageReadStart(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwFlags
);


/*
 * TFX_GetCompressedStripSize
 */
ErrCode CDECL
TFX_GetCompressedStripSize(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwImageRowOffset,
	UInt32 *pdwBytes
);


/*
 * TFX_CompressedReadStrip
 */
ErrCode CDECL
TFX_CompressedReadStrip(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 **ppCmpBuffer,
	UInt32 *pdwCmpBufsize,
	UInt32 *pdwBytesFilled
);


/*
 * TFX_ImageReadStrip
 */
#ifndef NOIPCORE
ErrCode CDECL
TFX_ImageReadStrip(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 **ppImagePlanes,
	UInt32 dwBytesPerRow,
	UInt32 dwBufferRowOffset,
	UInt32 *pdwRowsFilled
);
#else
#define TFX_ImageReadStrip(tfxImgHandle, dwImgRowOffset, dwStripHeight, \
	ppImagePlanes, dwBytesPerRow, dwBufferRowOffset, pdwRowsFilled) \
IO_NOSUPPORT
#endif


/*
 * TFX_GetDictListFromPosBlock
 */
ErrCode CDECL
TFX_GetDictListFromPosBlock(
	IO_IMGHANDLE tfxImgHandle,
	UInt8 *pCmpData,
	UInt32 dwCmpBytes,
	UInt16 *pdwDictCount,
	UInt32 **ppdwDictIDs,
	UInt8 **ppPositionBlock,
	UInt32 *pdwPositionBytes
);


/*
 * TFX_GetSymbolDict
 */
ErrCode CDECL
TFX_GetSymbolDict(
	IO_DOCHANDLE tfxDocHandle,
	UInt32 dictID,
	UInt8 **pDictBlock,
	UInt32 *pdwDictBytes,
	UInt32 *pdwDictBlockBytes,
	UInt16 *pwLastPageUsed,
	UInt32 *pdwDecoderMemory
);


/*
 * TFX_ImageReadFinish
 */
ErrCode CDECL
TFX_ImageReadFinish(
	IO_IMGHANDLE tfxImgHandle
);


/*
 * TFX_GetLayerInfo
 */
ErrCode CDECL
TFX_GetLayerInfo(
    IO_PAGEHANDLE tfxPageHandle,
    UInt32 dwImgLayer,
    UInt32 *pdwXOffset,
    UInt32 *pdwYOffset,
    UInt32 *pdwXResolution,
    UInt32 *pdwYResolution,
    UInt32 *pdwWidth,
    UInt32 *pdwHeight,
    IO_COMPRESSION *pxCompression,
    IO_COLORTYPE *pxColorType,
    UInt8 *pPalette,
    IO_IMGHANDLE *pTftfxImgHandle
);


/*
 * TFX_ImageReadLayer
 */
#ifndef NOIPCORE
ErrCode CDECL
TFX_ImageReadLayer(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwImageHeight,
	UInt8 **pImageBuffer,
	UInt32 dwBytesPerRow,
	UInt32 *pdwRowsFilled
);
#else
#define TFX_ImageReadLayer(tfxImgHandle, dwImageHeight, pImageBuffer, \
	dwBytesPerRow, pdwRowsFilled)		IO_NOSUPPORT
#endif


/*
 * TFX_GetAnnotationInfo
 */
ErrCode CDECL
TFX_GetAnnotationInfo(
	IO_PAGEHANDLE tfxPageHandle,
	UInt16 wAnnotOffset,		    /* Offset of annotation for this
					     * page */
	UInt16 *pwAnnotType,		    /* Type of annotation */
	UInt32 *pdwXOffset,		    /* Offset in X direction in page
					     * coordinates */
	UInt32 *pdwYOffset,		    /* Offset in X direction in page
					     * coordinates */
	UInt32 *pdwWidth,		    /* Width in page coordinates */
	UInt32 *pdwHeight,		    /* Height in page coordinates */
	UInt32 *pdwFlags,		    /* Annotation flags (see ANNOT_xxx
					     * above */
	UInt32 *pdwAnnotLength		    /* Length of annotation data, in
					     * bytes */
);


/*
 * TFX_GetAnnotationData
 */
ErrCode CDECL
TFX_GetAnnotationData(
	IO_PAGEHANDLE tfxPageHandle,
	UInt16 wAnnotOffset,
	UInt32 dwBytes,
	UInt8 *pData
);


/*
 * TFX_AnnotationOffsets
 */
ErrCode CDECL
TFX_GetAnnotationOffsets(
    IO_PAGEHANDLE tfxPageHandle,
	UInt32 *pdwAnnotationOffsets,
	Bool *pbExists
);


/*
 * TFX_Author
 */
ErrCode CDECL
TFX_GetAuthor(
    IO_DOCHANDLE tfxDocHandle,
	const Int8 **ppAuthor,
	UInt16 *pwAuthorBytes,
	Bool *pbExists
);


/*
 * TFX_GetBitsPerSample
 */
ErrCode CDECL
TFX_GetBitsPerSample(
    IO_IMGHANDLE ioImgHandle,
	UInt16 **ppwBitsPerSample,
	Bool *pbExists
);


/*
 * TFX_GetChromaPositioning
 */
ErrCode CDECL
TFX_GetChromaPositioning(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwChromaPositioning,
	Bool *pbExists
);


/*
 * TFX_GetChromaSubSampling
 */
ErrCode CDECL
TFX_GetChromaSubSampling(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwChromaSubSamplingHor,
	UInt16 *pwChromaSubSamplingVer,
	Bool *pbExists
);


/*
 * TFX_GetCodingMethods
 */
ErrCode CDECL
TFX_GetCodingMethods(
    IO_DOCHANDLE ioDocHandle,
	UInt32 *pdwCodingMethods,
	Bool *pbExists
);


/*
 * TFX_GetCompression
 */
ErrCode CDECL
TFX_GetCompression(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 *pwCompression,
	Bool *pbExists
);


/*
 * TFX_GetDateTime
 */
ErrCode CDECL
TFX_GetDateTime(
    IO_PAGEHANDLE ioPageHandle,
	Int8 **ppDateTime,
	Bool *pbExists
);


/*
 * TFX_GetDecode
 */
ErrCode CDECL
TFX_GetDecode(
    IO_IMGHANDLE ioImgHandle,
	Int32 **ppdwDecode,
	UInt32 *pdwDecodeValues,
	Bool *pbExists
);


/*
 * TFX_GetDocumentName
 */
ErrCode CDECL
TFX_GetDocumentName(
    IO_DOCHANDLE ioDocHandle,
	Int8 **ppDocumentName,
	UInt32 *pDocumentNameBytes,
	Bool *pbExists
);


/*
 * TFX_GetFaxProfile
 */
ErrCode CDECL
TFX_GetFaxProfile(
    IO_PAGEHANDLE ioPageHandle,
	UInt8 *pFaxProfile,
	Bool *pbExists
);


/*
 * TFX_GetFillOrder
 */
ErrCode CDECL
TFX_GetFillOrder(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwFillOrder,
	Bool *pbExists
);


/*
 * TFX_GetImageDescription
 */
ErrCode CDECL
TFX_GetHostComputer(
    IO_IMGHANDLE ioImgHandle,
	Int8 **ppHostComputer,
	UInt32 *pdwHostComputerBytes,
	Bool *pbExists
);


/*
 * TFX_GetImageBaseColor
 */
ErrCode CDECL
TFX_GetIFDBaseColor(
    IO_IMGHANDLE ioImgHandle,
	Int32 **ppdwIFDBaseColor,
	UInt32 *pdwIFDBaseColorValues,
	Bool *pbExists
);


/*
 * TFX_GetImageBaseColor
 */
ErrCode CDECL
TFX_GetImageBaseColor(
    IO_IMGHANDLE ioImgHandle,
	UInt16 **ppImageBaseColor,
	UInt32 *pdwImageBaseColorValues,
	Bool *pbExists
);


/*
 * TFX_GetImageDescription
 */
ErrCode CDECL
TFX_GetImageDescription(
    IO_IMGHANDLE ioImgHandle,
	Int8 **ppImageDescription,
	UInt32 *pdwImageDescriptionBytes,
	Bool *pbExists
);


/*
 * TFX_GetT88Options
 */
ErrCode CDECL
TFX_GetT88Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwT88Options,
	Bool *pbExists
);


/*
 * TFX_GetModeNumber
 */
ErrCode CDECL
TFX_GetModeNumber(
    IO_PAGEHANDLE ioPageHandle,
	UInt8 *pModeNumber,
	Bool *pbExists
);


/*
 * TFX_GetOrientation
 */
ErrCode CDECL
TFX_GetOrientation(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwOrientation,
	Bool *pbExists
);


/*
 * TFX_GetPageNumber
 */
ErrCode CDECL
TFX_GetPageNumber(
    IO_PAGEHANDLE ioPageHandle,
	UInt16 *pwPageNumber,
	UInt16 *pwPageCount,
	Bool *pbExists
);


/*
 * TFX_GetPhotometricInterpretation
 */
ErrCode CDECL
TFX_GetPhotometricInterpretation(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwPhotometricInterpretation,
	Bool *pbExists
);


/*
 * TFX_GetPlanarConfiguration
 */
ErrCode CDECL
TFX_GetPlanarConfiguration(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwPlanarConfiguration,
	Bool *pbExists
);


/*
 * TFX_GetProfileType
 */
ErrCode CDECL
TFX_GetProfileType(
    IO_PAGEHANDLE ioPageHandle,
	UInt32 *pdwProfileType,
	Bool *pbExists
);


/*
 * TFX_GetReferencePage
 */
ErrCode CDECL
TFX_GetReferencePage(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 *pdwReferencePage,
	Bool *pbExists
);


/*
 * TFX_GetRenderingHints
 */
ErrCode CDECL
TFX_GetRenderingHints(
    IO_IMGHANDLE ioImgHandle,
	Int8 **ppRenderingHints,
	UInt32 *pdwRenderingHintsBytes,
	Bool *pbExists
);


/*
 * TFX_GetRenderingHintsImageCount
 */
ErrCode CDECL
TFX_GetRenderingHintsImageCount(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 *pdwCount
);


/*
 * TFX_GetResolutionUnit
 */
ErrCode CDECL
TFX_GetResolutionUnit(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwResolutionUnit,
	Bool *pbExists
);


/*
 * TFX_GetRowsPerStrip
 */
ErrCode CDECL
TFX_GetRowsPerStrip(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwRowsPerStrip,
	Bool *pbExists
);


/*
 * TFX_GetSamplesPerPixel
 */
ErrCode CDECL
TFX_GetSamplesPerPixel(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwSamplesPerPixel,
	Bool *pbExists
);


/*
 * TFX_GetSoftware
 */
ErrCode CDECL
TFX_GetSoftware(
    IO_IMGHANDLE ioImgHandle,
	Int8 **ppSoftware,
	UInt32 *pdwSoftwareBytes,
	Bool *pbExists
);


/*
 * TFX_GetTargetPrinter
 */
ErrCode CDECL
TFX_GetTargetPrinter(
    IO_PAGEHANDLE ioPageHandle,
	Int8 **ppTargetPrinter,
	UInt32 *pdwTargetPrinterBytes,
	Bool *pbExists
);


/*
 * TFX_GetThresholding
 */
ErrCode CDECL
TFX_GetThresholding(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwThresholding,
	Bool *pbExists
);


/*
 * TFX_GetT4Options
 */
ErrCode CDECL
TFX_GetT4Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwT4Options,
	Bool *pbExists
);


/*
 * TFX_GetT6Options
 */
ErrCode CDECL
TFX_GetT6Options(
    IO_IMGHANDLE ioImgHandle,
	UInt32 *pdwT6Options,
	Bool *pbExists
);


/*
 * TFX_GetVersionYear
 */
ErrCode CDECL
TFX_GetVersionYear(
    IO_IMGHANDLE ioImgHandle,
	UInt8 **ppVersionYear,
	UInt32 *pdwVersionYearBytes,
	Bool *pbExists
);


/*
 * TFX_GetYcrCbCoefficients
 */
ErrCode CDECL
TFX_GetYCbCrCoefficients(
    IO_IMGHANDLE ioImgHandle,
	UInt32 **ppdwYCbCrCoefficients,
	UInt32 *pdwYCbCrCoefficientsValues,
	Bool *pbExists
);


/*
 * TFX_GetPredictor
 */
ErrCode CDECL
TFX_GetPredictor(
    IO_IMGHANDLE ioImgHandle,
	UInt16 *pwPredictor,
	Bool *pbExists
);

/*
 * **********						********** 
 * **********	BEGIN WRITE API		********** 
 * **********						********** 
 */

/*
 * TFX_CreateDocument
 */
ErrCode CDECL
TFX_CreateDocument(
	IO_INSTHANDLE xInstHandle,
	Int32 dwClientID,		/* Input client ID */
	UInt8 version,			/* Version of TIFF-FX format to create */
	UInt32 options,			/* Format specific options */
	IO_READFUNC xReadFunc,		/* Input function */
	IO_WRITEFUNC xWriteFunc,	/* Input function */
	IO_SEEKFUNC xSeekFunc,		/* Input function */
	IO_SIZEFUNC xSizeFunc,		/* Input function */
	IO_SIZEFUNC xTellFunc,		/* Input function */
	IO_DOCHANDLE *pTfxDocHandle	/* Returned */
);


/*
 * TFX_OpenDocumentReadWrite
 */
ErrCode CDECL
TFX_OpenDocumentReadWrite(
	IO_INSTHANDLE xInstHandle,
	Int32 dwClientID,		/* Input client ID */
	UInt8 version,			/* Version of TIFF-FX format to create */
	UInt32 options,			/* Format specific options */
	IO_READFUNC xReadFunc,		/* Input function */
	IO_WRITEFUNC xWriteFunc,	/* Input function */
	IO_SEEKFUNC xSeekFunc,		/* Input function */
	IO_SIZEFUNC xSizeFunc,		/* Input function */
	IO_SIZEFUNC xTellFunc,		/* Input function */
	IO_DOCHANDLE *pTfxDocHandle	/* Returned */
);


/*
 * TFX_CopyImage
 */
ErrCode CDECL
TFX_CopyImage(
	IO_IMGHANDLE xSrcImgHandle,
	IO_IMGHANDLE xSrcParentImgHandle,
	IO_PAGEHANDLE xDstPageHandle,
	UInt32 dwImgLayer,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwBestStripHeight,
	IO_IMGHANDLE *pxDstImgHandle
);


/*
 * TFX_CopyPage
 */
ErrCode CDECL
TFX_CopyPage(
	IO_DOCHANDLE xSrcDocHandle,
	IO_DOCHANDLE xDstDocHandle,
	Int32 dwSrcPageOffset,
	Int32 dwDstPageOffset,
	Bool copyAnnotations
);


/*
 * TFX_CopyDoc
 */
ErrCode CDECL
TFX_CopyDoc(
	    IO_DOCHANDLE xSrcDocHandle,
	    IO_DOCHANDLE xDstDocHandle
);


/*
 * TFX_AddPageStart
 */
ErrCode CDECL
TFX_AddPageStart(
	IO_DOCHANDLE tfxDocHandle,
	Int32 dwPageOffset,		    /* Page offset */
	IO_PAGEHANDLE *pTfxPageHandle	    /* Page handle returned */
);


/*
 * TFX_AddPageFinish
 */
ErrCode CDECL
TFX_AddPageFinish(
	IO_PAGEHANDLE tfxPageHandle
);


/*
 * TFX_DeletePage
 */
ErrCode CDECL
TFX_DeletePage(
	IO_PAGEHANDLE tfxPageHandle
);


/*
 * TFX_CheckImageInfo
 */
ErrCode CDECL
TFX_CheckImageInfo(
	IO_PAGEHANDLE tfxPageHandle,
	UInt32 *pdwImgLayer,
	UInt32 *pdwImgOrder,
	IO_COLORTYPE *pxColorType,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	UInt8 *pPalette,
	IO_COMPRESSION *pxCompression,
	IO_COLORSPACE *pxColorSpace
);


/*
 * TFX_AddImage
 */
ErrCode CDECL
TFX_AddImage(
	IO_PAGEHANDLE tfxPageHandle,	/* Page to attach to */
	UInt32 dwImgLayer,		/* Image layer in the MRC */
	IO_COLORTYPE xColorType,	/* Color type (IO_COLORTYPE_BINARY
					 * ...) */
	UInt32 dwXOffset,		/* X location in page coordinates */
	UInt32 dwYOffset,		/* Y location in page coordinates */
	UInt32 dwXResolution,		/* X Resolution of image in dpm */
	UInt32 dwYResolution,		/* Y Resolution of image in dpm */
	UInt32 dwWidth,			/* Width in image's X resolution units
					 */
	UInt32 dwHeight,		/* Height in image's Y resolution units
					 */
	UInt32 dwBestStripHeight,	/* For best performance, read strips
					 * with height a multiple of this. */
	UInt8 *pPalette,		/* Color palette if IO_COLORTYPE is:
					 * IO_COLORTYPE_COLOR4 or
					 * IO_COLORTYPE_COLOR8.
					 * Palette is RGB triples (chunky):
					 * RGBRGB... */
	IO_IMGHANDLE *pTftfxImgHandle	/* Returned image handle for image to
					 * be added */
);


/*
 * TFX_AddRenderingHints
 */
ErrCode CDECL
TFX_AddRenderingHints(
    IO_IMGHANDLE tfxImgHandle,
	char *pRenderingHintsString,
	UInt32 dwRenderingHintsBytes
	);


/*
 * TFX_AddHintsImage
 */
ErrCode CDECL
TFX_AddHintsImage(
	IO_IMGHANDLE tfxParentImgHandle,
	UInt32 dwBpp,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwXResolution,
	UInt32 dwYResolution,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt32 dwBestStripHeight,
	IO_IMGHANDLE *pImgHandle
);


/*
 * TFX_DeleteImage
 */
ErrCode CDECL
TFX_DeleteImage(
	IO_IMGHANDLE tfxImgHandle
);

#if 0
/*
 * TFX_AddMaskImage
 */
ErrCode CDECL
TFX_AddMaskImage(
	IO_IMGHANDLE tfxParentImgHandle,
	UInt32 dwBestStripHeight,
	IO_IMGHANDLE *pImgHandle
);
#endif


/*
 * TFX_ImageWriteStart
 *
 * Initiate an image write, to be followed with a sequence of one or more calls
 * to TFX_ImageWriteStrip() or TFX_CompressedWriteStrip() or
 * TFX_TokenizeImage().
 * In the case of TFX_TokenizeImage(), if using color tokenization, then
 * the color image must have been previously passed, and marked for compression
 * using compression type: TFX_COLORFORTOKENS.
 */
ErrCode CDECL
TFX_ImageWriteStart(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwFlags, 
	IO_COMPRESSION *pxCompression, /* IO_COMP_UNKNOWN yields default */
	UInt32 *pdwJPEGQuality,	/* 0-100:
				 *   0 - default used and passed back
				 *  75   good quality, best compression
				 *  90   better quality, good compression
				 * 100   best quality, okay compression */
	IO_COLORSPACE *pxColorSpace	/* Stored color space to use */
 );	  

	
/*
 * TFX_CompressedWriteStrip
 */
ErrCode CDECL
TFX_CompressedWriteStrip(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 *pCmpBuffer,
	UInt32 dwCmpBytes,
	UInt32 *pdwBytesStored
);

	
/*
 * TFX_ImageWriteStrip
 */
#ifndef NOIPCORE
ErrCode CDECL
TFX_ImageWriteStrip(
	IO_IMGHANDLE tfxImgHandle,
	UInt8 **ppImagePlanes,
	UInt32 dwBytesPerRow,
	UInt32 dwBufferRowOffset,
	UInt32 dwStripHeight,
	UInt32 dwImgRowOffset,
	UInt32 *pRowsFilled
);
#else
#define TFX_ImageWriteStrip(tfxImgHandle, ppImagePlanes, dwBytesPerRow, \
	dwBufferRowOffset, dwStripHeight, dwImgRowOffset, pRowsFilled) \
IO_NOSUPPORT
#endif
	
/*
 * TFX_ImageWriteFinish
 */
ErrCode CDECL
TFX_ImageWriteFinish(
	IO_IMGHANDLE tfxImgHandle
);
	

/* Tokenization flags */	
#define TOKENIZE_OFF			(0x00)	/* No tokenization */
#define TOKENIZE_ON			(0x01)	/* Tokenization enabled */
#define TOKENIZE_LOSSLESS	    	(0x02)	/* Tokenization in lossless
						 * mode */
#define TOKENIZE_BINARYSCAN		(0x04)	/* Image was scanned with a
						 * binary scanner, indicating
						 * that halftones may be
						 * removed and lossy dither
						 * compressed before
						 * tokenization */
#define TOKENIZE_COLORTOO	    	(0x08)	/* Color tokenize: use
						 * previously added FG image
						 * to extract colors from */

/*
 * TFX_TokenizeImage()
 */
#if !defined(NOIPCORE) && !defined(NODGP)
ErrCode CDECL
TFX_TokenizeImage(
	IO_IMGHANDLE tfxImgHandle,
	UInt8 *pImgData,	/* One channel only */
	UInt32 dwBytesPerRow,	/* Bytes per row */
	UInt32 dwFlags		/* Flags: Lossless Tokenization/Binary Scan
				 * (see TOKEN_FLAGS) */
);
#else
#define TFX_TokenizeImage(tfxImgHandle, pImgData, dwBytesPerRow, dwFlags) \
IO_NOSUPPORT
#endif


/*
 * TFX_WriteTokenImages()
 */
#if !defined(NOIPCORE) && !defined(NODGP)
ErrCode CDECL
TFX_WriteTokenImages(
	IO_DOCHANDLE tfxDocHandle
);
#else
#define TFX_WriteTokenImages(tfxDocHandle)	IO_NOSUPPORT
#endif


/*
 * TFX_WriteTokenImage()
 */
#if !defined(NOIPCORE) && !defined(NODGP)
ErrCode CDECL
TFX_WriteTokenImage(
	IO_IMGHANDLE tfxImgHandle
);
#else
#define TFX_WriteTokenImage(tfxImgHandle)	IO_NOSUPPORT
#endif


/*
 * TFX_WriteTokenDictionary()
 */
#if !defined(NOIPCORE) && !defined(NODGP)
ErrCode CDECL
TFX_WriteTokenDictionary(
	IO_DOCHANDLE tfxDocHandle
);
#else
#define TFX_WriteTokenDictionary(tfxDocHandle)	IO_NOSUPPORT
#endif


/*
 * TFX_StoreJBIG2Dictionary()
 */
ErrCode CDECL
TFX_StoreSymbolDictionary(
	IO_DOCHANDLE tfxDocHandle,
	UInt8 *dictBlock,
	UInt32 dictBytes,
	UInt32 dictBlockBytes,
	UInt32 *dictID
);


/*
 * TFX_StoreJBIG2Image()
 */
ErrCode CDECL
TFX_StorePositionBlock(
	IO_IMGHANDLE tfxImgHandle,
	UInt8 *pPosBlock,
	UInt32 dwPosBytes,
	UInt32 *pdwDictIDList,
	UInt16 wDictCount
);


/*
 * TFX_ShareColorTokenImage()
 */
ErrCode CDECL
TFX_ShareColorTokenImage(
	IO_IMGHANDLE tfxBinaryImgHandle,
	UInt32 dwBinaryImageBytes,	/* Bytes shared by binary image */
	IO_COLORTYPE xColorType,	/* Color image color type */
	IO_IMGHANDLE *pTfxImgHandle
);


/*
 * TFX_SetAuthor
 */
ErrCode CDECL
TFX_SetAuthor(
	IO_DOCHANDLE tfxDocHandle,
	Int8 *pAuthorBlock,
	UInt16 wAuthorBytes
);


/*
 * TFX_SetReferencePage
 */
ErrCode CDECL
TFX_SetReferencePage(
	IO_DOCHANDLE tfxDocHandle,
	Int32 dwReferencePageOffset
);


/*
 * TFX_SetAnnotTableEntryCount
 */
ErrCode CDECL
TFX_SetAnnotTableEntryCount(
	IO_DOCHANDLE tfxDocHandle,
	UInt16 wEntriesPerTable
);


/*
 * TFX_SetPageTableEntryCount
 */
ErrCode CDECL
TFX_SetPageTableEntryCount(
	IO_DOCHANDLE tfxDocHandle,
	UInt16 wEntriesPerTable
);


/*
 * TFX_SetResourceTableEntryCount
 */
ErrCode CDECL
TFX_SetResourceTableEntryCount(
	IO_DOCHANDLE tfxDocHandle,
	UInt16 wEntriesPerTable
);


/*
 * TFX_AddAnnotation
 */
ErrCode CDECL
TFX_AddAnnotation(
	IO_PAGEHANDLE tfxPageHandle,
	UInt16 wAnnotType,		    /* Type of annotation */
	UInt32 dwXOffset,		    /* Offset in X direction in page
					     * coordinates */
	UInt32 dwYOffset,		    /* Offset in X direction in page
					     * coordinates */
	UInt32 dwWidth,			    /* Width in page coordinates */
	UInt32 dwHeight,		    /* Height in page coordinates */
	UInt32 dwFlags,			    /* Annotation flags (see ANNOT_xxx
					     * above) */
	UInt8 *pBuf,			    /* Annotation data */
	UInt32 dwAnnotLength,		    /* Length of annotation data, in
					     * bytes */
	UInt16 *pTfxAnnotID		    /* ID given to this annotation
					     * (offset) */
); 


/*
 * TFX_DeleteAnnotation
 */
ErrCode CDECL
TFX_DeleteAnnotation(
	IO_PAGEHANDLE tfxPageHandle,
	UInt16 wAnnotOffset
);


/*
 * TFX_Compress
 */
#ifndef NOIPCORE
ErrCode CDECL
TFX_Compress(
	IO_IMGHANDLE tfxImgHandle,	/* Image handle */
	IO_COMPRESSION xCompression,	/* See IO_COMPRESSION types */
	UInt32 *pdwQuality,		/* Quality to use, actual returned */
	IO_COLORTYPE xColorType,	/* Type of input image color */
	IO_COLORSPACE xColorSpace,	/* Color space of compressed image */
	UInt32 dwFlags,			/* TFX_IMGFLAGS_xxx combo */
	UInt8 *pSrc[3],			/* Source buffer channels */
	UInt32 dwBytesPerRow,		/* Bytes per row in image buffer */
	UInt32 dwRowOffset,		/* Row offset in image of data to
					 * compress */
	UInt32 dwWidth,			/* Columns of COLORTYPE in src image */
	UInt32 dwHeight,		/* Rows in src image */
	UInt8 **ppDst,			/* Pointer to destination compressed
					 * data buffers */
	UInt32 *pdwBytesInDst,		/* Pointer to byte count of destination
					 * buffer */
	UInt32 *pdwBytesFilled		/* Actual bytes in destination filled
					 */
);
#else
#define TFX_Compress(tfxImgHandle, xCompression, pdwQuality, xColorType, \
	xColorSpace, dwFlags, pSrc, dwBytesPerRow, dwRowOffset, dwWidth, \
	dwHeight, ppDst, pdwBytesInDst, pdwBytesFilled)		IO_NOSUPPORT
#endif


/*
 * TFX_Decompress
 */
#ifndef NOIPCORE
ErrCode CDECL
TFX_Decompress(
	IO_IMGHANDLE tfxImgHandle,	/* Image handle */
	IO_COMPRESSION xCompression,	/* See IO_COMPRESSION types */
	IO_COLORTYPE xColorType,	/* Type of color */
	UInt32 dwFlags,			/* TFX_IMGFLAGS_xxx bit combo */
	UInt8 *pSrc,			/* Source buffer */
	UInt32 dwBytesInSrc,		/* Byte count of source buffer */
	UInt32 dwRowOffset,		/* Destination row offset */
	UInt32 dwWidth,			/* Columns of COLORTYPE in dst image */
	UInt32 dwHeight,		/* Rows in dst image */
	UInt8 *pDst[3],			/* Destination image buffer channels */
	UInt32 dwBytesPerRow,		/* Bytes per row in image buffer */
	UInt32 *pdwRowsFilled		/* Number of rows successfully filled
					 */
);
#else
#define TFX_Decompress(tfxImgHandle, xCompression, xColorType, dwFlags, \
	pSrc, dwBytesInSrc, dwRowOffset, dwWidth, dwHeight, pDst, \
	dwBytesPerRow, pdwRowsFilled)		IO_NOSUPPORT
#endif


/*
 * TFX_SetAnnotationOffsets
 */
ErrCode CDECL
TFX_SetAnnotationOffsets(
    IO_PAGEHANDLE tfxPageHandle,
	UInt32 dwAnnotationOffsets
);


/*
 * TFX_SetBitsPerSample
 */
ErrCode CDECL
TFX_SetBitsPerSample(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wBitsPerSample,
	UInt32 dwBitsPerSampleValues
);


/*
 * TFX_SetChromaPositioning
 */
ErrCode CDECL
TFX_SetChromaPositioning(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wPositioning
);


/*
 * TFX_SetChromaSubSampling
 */
ErrCode CDECL
TFX_SetChromaSubSampling(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wChromaSubSamplingHor,
	UInt16 wChromaSubSamplingVer
    );


/*
 * TFX_SetCodingMethods
 */
ErrCode CDECL
TFX_SetCodingMethods(
    IO_PAGEHANDLE tfxPageHandle,
	UInt32 dwCodingMethods
);


/*
 * TFX_SetCompression
 */
ErrCode CDECL
TFX_SetCompression(
    IO_PAGEHANDLE tfxPageHandle,
	UInt16 wCompression
);


/*
 * TFX_SetDateTime
 */
ErrCode CDECL
TFX_SetDateTime(
    IO_PAGEHANDLE tfxPageHandle,
	Int8 *pDateTime
);


/*
 * TFX_SetDecode
 */
ErrCode CDECL
TFX_SetDecode(
    IO_IMGHANDLE tfxImgHandle,
	Int32 *pdwDecode,
	UInt32 dwDecodeValues
    );


/*
 * TFX_SetDocumentName
 */
ErrCode CDECL
TFX_SetDocumentName(
    IO_DOCHANDLE tfxDocHandle,
	Int8 *pDocumentName,
	UInt32 dwDocumentNameBytes
    );


/*
 * TFX_SetFaxProfile
 */
ErrCode CDECL
TFX_SetFaxProfile(
    IO_PAGEHANDLE tfxPageHandle,
	UInt8 faxProfile
);


/*
 * TFX_SetFillOrder
 */
ErrCode CDECL
TFX_SetFillOrder(
	IO_PAGEHANDLE tfxPageHandle,
	UInt16 fillOrder
);


/* EEP 20000803 
 * TFX_SetHostComputer()
 */
ErrCode CDECL
TFX_SetHostComputer(
    IO_IMGHANDLE tfxImgHandle,
	Int8 *pHostComputer,
	UInt32 dwHostComputerBytes
);


/*
 * TFX_SetIFDBaseColor
 */
ErrCode CDECL
TFX_SetIFDBaseColor(
    IO_IMGHANDLE tfxImgHandle,
	Int32 *pdwIFDBaseColor,
	UInt32 dwIFDBaseColorValues
    );


/*
 * TFX_SetImageBaseColor
 */
ErrCode CDECL
TFX_SetImageBaseColor(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 *pImageBaseColor,
	UInt32 dwImageBaseColorValues
    );


/*
 * TFX_SetImageDescription
 */
ErrCode CDECL
TFX_SetImageDescription(
    IO_IMGHANDLE tfxImgHandle,
	Int8 *pImageDescription,
	UInt32 dwImageDescriptionBytes
    );


/*
 * TFX_SetT88Options
 */
ErrCode CDECL
TFX_SetT88Options(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 dwT88Options
);


/*
 * TFX_SetModeNumber
 */
ErrCode CDECL
TFX_SetModeNumber(
    IO_PAGEHANDLE tfxPageHandle,
	UInt8 modeNumber
);


/*
 * TFX_SetOrientation
 */
ErrCode CDECL
TFX_SetOrientation(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wOrientation
);


/*
 * TFX_SetPageNumber
 */
ErrCode CDECL
TFX_SetPageNumber(
    IO_PAGEHANDLE tfxPageHandle,
	UInt16 wPageOffset,
	UInt16 wPageCount
    );


/*
 * TFX_SetPhotometricInterpretation
 */
ErrCode CDECL
TFX_SetPhotometricInterpretation(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wPhotometricInterpretation
);


/*
 * TFX_SetPlanarConfiguration
 */
ErrCode CDECL
TFX_SetPlanarConfiguration(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wPlanarConfiguration
);


/*
 * TFX_SetProfileType
 */
ErrCode CDECL
TFX_SetProfileType(
    IO_PAGEHANDLE tfxPageHandle,
	UInt32 dwProfileType
);


#if 0
/*
 * TFX_SetReferencePage
 */
ErrCode CDECL
TFX_SetReferencePage(
    IO_DOCHANDLE tfxDocHandle,
	Int32 dwReferencePageOffset
    );
#endif


/*
 * TFX_SetResolutionUnit
 */
ErrCode CDECL
TFX_SetResolutionUnit(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wResolutionUnit
);


/*
 * TFX_SetRowsPerStrip
 */
ErrCode CDECL
TFX_SetRowsPerStrip(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 dwRowsPerStrip
);


/*
 * TFX_SetSamplesPerPixel
 */
ErrCode CDECL
TFX_SetSamplesPerPixel(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wSamplesPerPixel
);


/*
 * TFX_SetSoftware
 */
ErrCode CDECL
TFX_SetSoftware(
    IO_IMGHANDLE tfxImgHandle,
	Int8 *pSoftware,
	UInt32 dwSoftwareBytes
    );


/*
 * TFX_SetTargetPrinter
 */
ErrCode CDECL
TFX_SetTargetPrinter(
    IO_PAGEHANDLE tfxPageHandle,
	Int8 *pTargetPrinter,
	UInt32 dwTargetPrinterBytes
    );


/*
 * TFX_SetTargetPrinter
 */
ErrCode CDECL
TFX_SetTargetPrinter(
    IO_PAGEHANDLE tfxPageHandle,
	Int8 *pTargetPrinter,
	UInt32 dwTargetPrinterBytes
    );


/*
 * TFX_SetThresholding
 */
ErrCode CDECL
TFX_SetThresholding(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wThresholding
    );


/*
 * TFX_SetT4Options
 */
ErrCode CDECL
TFX_SetT4Options(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 dwT4Options
);


/*
 * TFX_SetT6Options
 */
ErrCode CDECL
TFX_SetT6Options(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 dwT6Options
);


/*
 * TFX_SetVersionYear
 */
ErrCode CDECL
TFX_SetVersionYear(
    IO_IMGHANDLE tfxImgHandle,
	UInt8 *pVersionYear,
	UInt32 dwVersionYearBytes
    );


/*
 * TFX_SetYCbCrCoefficients
 */
ErrCode CDECL
TFX_SetYCbCrCoefficients(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 *pdwYCbCrCoefficients,
	UInt32 dwYCbCrCoefficientValues
    );


/*
 * TFX_SetPredictor
 */
ErrCode CDECL
TFX_SetPredictor(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wPredictor
    );


/*
 * TFX_UseMultiStripSubIFDs
 */
ErrCode CDECL
TFX_UseMultiStripSubIFDs(IO_PAGEHANDLE tfxPageHandle);


/*
 * TFX_UseNoGPIFD
 */
ErrCode CDECL
TFX_UseNoGPIFD(IO_DOCHANDLE tfxDocHandle);


/*
 * TFX_UseVariableStripHeights
 */
ErrCode CDECL
TFX_UseVariableStripHeights(IO_IMGHANDLE tfxImgHandle);


/*
 * TFX_UseColorStrips
 */
ErrCode CDECL
TFX_UseColorStrips(IO_IMGHANDLE tfxImgHandle);


/*
 * TFX_GetStripInfo
 */
ErrCode CDECL
TFX_GetStripeInfo(IO_PAGEHANDLE ioPageHandle, UInt32 rowOffset,
	UInt32 *pdwStripeOffset, UInt32 *pdwStripeRow, UInt32 *pdwStripeHeight);


#ifdef _MSC_VER
#pragma pack(pop, tfx_h)
#endif

#ifdef __cplusplus
}
#endif

#endif /* TFX_H_INCLUDED */
