/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxconv.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxconv.c:	TIFF-FX Image conversion routines
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tfxconv.c,v $; $Revision: 1.1 $; $Date: 2003/01/03 01:41:10 $");

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "imageref.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif
#endif	/* NOIPCORE */

#include "tfx.h"
#include "tfximage.h"
#include "tfxconv.h"
#include "tfxrpriv.h"


#ifndef NOIPCORE

#define CIELAB		1
#define ICCLAB		2
#define ITULAB		3

/* Prototypes */
static ErrCode CDECL
wp_LAB2sRGB(
	TFX_IMG *tfxImg,
	PIXR *pPixr,
	Int32 cspace,
	Int32 *gamut
);
static ErrCode CDECL
wp_sRGB2LAB(
	TFX_IMG *tfxImg,
	PIXR *pPixr,
	Int32 cspace,
	Int32 *gamut
);
static ErrCode CDECL 
gray2ITUL(
    IPWState    *pIPWState,
    UInt8	*image,
    UInt32	w,
    UInt32	h,
    UInt32	bpl,
    Int32	*gamut
    );
static ErrCode CDECL 
ITUL2Gray(
    IPWState    *pIPWState,
    UInt8	*image,
    UInt32	w,
    UInt32	h,
    UInt32	bpl,
    Int32	*gamut
    );




ErrCode CDECL
wp_reverseLongs(
		UInt8 *buffer,
		UInt32 w
)
{
    UInt32 *lptr, l, newl;
    UInt32 word, words = (w + 3) / 4;

    lptr = (void *)buffer;
    for (word = 0; word < words; word++) {
	l = *lptr;
	newl = l, newl >>= 8, l <<= 8;
	newl |= (l & 0xff000000), newl >>= 8, l <<= 8;
	newl |= (l & 0xff000000), newl >>= 8, l <<= 8;
	newl |= (l & 0xff000000);
	*lptr++ = newl;
    }

    return 0;
}



/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do internal image processing to convert from TIFF format to PIXR
 *
 * LongDescription:::
 *  w_TfxConvertImage() does minimal image processing required to put
 *  image data into an expected PIXR state (binary: 0 = white, non-paletted:
 *  0 = black).
 *
 * Input:::
 *  tfxImg::			Valid image
 *  pPixr::			Image data structure
 *  dwFlags::			Processes (see tfxconv.h TFX_CONV_xxx)
 *  rowOffset::			Offset of first row to process
 *				(ignore, placeholders for later implementation)
 *  rowCount::			Number of rows to process
 *				(ignore, placeholders for later implementation)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_TfxConvertImage(
	TFX_IMG *tfxImg,
	PIXR *pPixr,
	UInt32 dwFlags,
	UInt32 rowOffset,
	UInt32 rowCount
)
{
    TFX_INSTANCE *instData = TFX_IMGINST(tfxImg);
    TFX_IMGINFO *pTftfxImgInfo;
    IPWState *pIPWState;
    PIXR *newPixr = NULL;
    UInt8 *rgbBuffer = NULL;
    PROLOGUE;

    /* Reference the image dimensions */
    pTftfxImgInfo = &tfxImg->imgInfo;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    if (dwFlags & TFX_CONV_INVERT) {
	XERR_PASS_CLEANUP(w_invert(pIPWState, pPixr, &pPixr));
    }

    if (dwFlags & TFX_CONV_UNCHUNK) {
	UInt8 *r, *g, *b, *rgb, *rRow, *gRow, *bRow;
	UInt32 row, col, w, h;
	UInt32 byteCount;
	UInt32 channel, bpl;
	PIXR *pSubPixr;

	/* There's no easy way of shuffling the data around.  We're gunna
	 * have to copy the data, then stuff it back in in the right order.
	 */
	XERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, pPixr, (Int32 *)&w));
	XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, (Int32 *)&h));
	byteCount = w * h * 3;
	IO_CHECKMALLOC(rgbBuffer = IO_MALLOC(byteCount));

	XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, (Int32 *)&bpl));

	/* Copy the chunk data to the buffer */
	rgb = rgbBuffer;
	for (channel = 0; channel < 3; channel++) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
					  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &rRow));

	    for (row = 0; row < h; row++) {
		wp_reverseLongs(rRow, w);
		memcpy(rgb, rRow, w);
		rRow += bpl;
		rgb += w;
	    }
	}

	/* Get the Red channel */
	pSubPixr = NULL;
	XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, 0, &pSubPixr));
	XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &rRow));

	/* Get the Green channel */
	pSubPixr = NULL;
	XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, 1, &pSubPixr));
	XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &gRow));

	/* Get the Blue channel */
	pSubPixr = NULL;
	XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, 2, &pSubPixr));
	XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &bRow));

	/* Copy the data back, from chunky to planar */
	rgb = rgbBuffer;
	for (row = 0; row < h; row++) {
	    r = rRow, g = gRow, b = bRow;
	    for (col = 0; col < w; col++) {
		*r++ = *rgb++;
		*g++ = *rgb++;
		*b++ = *rgb++;
	    }

	    /* Reverse the order of the byte-packed longs */
	    if ((dwFlags & TFX_CONV_DWORD2BYTEORDER) == 0) {
		wp_reverseLongs(rRow, w);
		wp_reverseLongs(gRow, w);
		wp_reverseLongs(bRow, w);
	    }

	    /* Go to the next row of data */
	    rRow += bpl;
	    gRow += bpl;
	    bRow += bpl;
	}

	/* Cleanup */
	IO_FREE(rgbBuffer);
	rgbBuffer = NULL;
	if (dwFlags & TFX_CONV_DWORD2BYTEORDER) {
	    dwFlags ^= TFX_CONV_DWORD2BYTEORDER;
	}
    }

    switch (dwFlags & TFX_CONV_COLORTRANSFORM) {
    case TFX_CONV_CIELAB2SRGB:
	XERR_PASS_CLEANUP(wp_LAB2sRGB(tfxImg, pPixr, CIELAB, NULL));
	break;
    case TFX_CONV_ICCLAB2SRGB:
	XERR_PASS_CLEANUP(wp_LAB2sRGB(tfxImg, pPixr, ICCLAB, NULL));
	break;
    case TFX_CONV_ITULAB2SRGB:
	XERR_PASS_CLEANUP(wp_LAB2sRGB(tfxImg, pPixr, ITULAB, tfxImg->fields.Decode));
	break;
    case TFX_CONV_SRGB2CIELAB:
	XERR_PASS_CLEANUP(wp_sRGB2LAB(tfxImg, pPixr, CIELAB, NULL));
	break;
    case TFX_CONV_SRGB2ICCLAB:
	XERR_PASS_CLEANUP(wp_sRGB2LAB(tfxImg, pPixr, ICCLAB, NULL));
	break;
    case TFX_CONV_SRGB2ITULAB:
	XERR_PASS_CLEANUP(wp_sRGB2LAB(tfxImg, pPixr, ITULAB, tfxImg->fields.Decode));
	break;
    }

    /* For conversion from planar to chunky, then writing the output, it
     * makes more sense to use a line buffer, and chunk to it intermediately.
     * For this to happen, we'd have to do it when writing out.  Since this routine
     * is used to make all the necessary preparations for the raw data stored in
     * a TIFF file, we have to do it here.
     */
    if (dwFlags & TFX_CONV_CHUNK) {
	UInt8 *rgb, *rows[4], *rp[4], *pd;
	UInt32 row, col, w, h;
	UInt32 byteCount;
	UInt32 channel, channels, bpl;
	PIXR *pSubPixr;

	/* There's no easy way of shuffling the data around.  We're gunna
	 * have to copy the data, then stuff it back in in the right order.
	 */
	XERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, pPixr, (Int32 *)&w));
	XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, (Int32 *)&h));
	byteCount = w * h * 3;
	IO_CHECKMALLOC(rgbBuffer = IO_MALLOC(byteCount));
	channels = IO_CHANNELS(tfxImg->imgInfo.xColorType);

	XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, (Int32 *)&bpl));

	/* Get the image channels */
	for (channel = 0; channel < channels; channel++) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel, &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &rows[channel]));
	}

	/* Copy the planar data to the buffer, and interleave the channels as copied */
	rgb = rgbBuffer;
	for (row = 0; row < h; row++) {
	    /* Start at the beginning of the row */
	    for (channel = 0; channel < channels; channel++) {
		rp[channel] = rows[channel];
	    }
	    /* Copy the data and interleave the channels */
	    for (col = 0; col < w; col++) {
		for (channel = 0; channel < channels; channel++) {
		    *rgb++ = UCHAR_ACCESS(rp[channel]);
		    rp[channel]++;
		}
	    }
	    /* Advance to the next row */
	    for (channel = 0; channel < channels; channel++) {
		rows[channel] += bpl;
	    }
	}

	/* Get the first row of the image channels again */
	for (channel = 0; channel < channels; channel++) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel, &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &rows[channel]));
	}
	/* Copy the interleaved data back to the PIXRs */
	rgb = rgbBuffer;
	for (channel = 0; channel < channels; channel++) {
	    pd = rows[channel];
	    for (row = 0; row < h; row++) {
		memcpy(pd, rgb, w);
		if ((dwFlags & TFX_CONV_DWORD2BYTEORDER) == 0) {
		    wp_reverseLongs(pd, w);
		}
		pd += bpl;
		rgb += w;
	    }
	}

	/* Cleanup */
	IO_FREE(rgbBuffer);
	rgbBuffer = NULL;
	if (dwFlags & TFX_CONV_DWORD2BYTEORDER) {
	    dwFlags ^= TFX_CONV_DWORD2BYTEORDER;
	}
    }

    if (dwFlags & (TFX_CONV_DWORD2BYTEORDER | TFX_CONV_BYTE2DWORDORDER)) {
	UInt8 *image;
	UInt32 row, w, h;
	UInt32 channel, channels, bpl;
	PIXR *pSubPixr;

	/* Convert each row */
	XERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, pPixr, (Int32 *)&w));
	XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, (Int32 *)&h));
	XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, (Int32 *)&channels));

	XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, (Int32 *)&bpl));

	/* Grab the image buffer(s) */
	for (channel = 0; channel < channels; channel++) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
					  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &image));

	    /* Reverse the longs in each row */
	    for (row = 0; row < h; row++) {
		wp_reverseLongs(image, w);
		image += bpl;
	    }
	}

    }

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (newPixr) {
	w_destroyPixr(pIPWState, &newPixr);
    }
    if (rgbBuffer) {
	IO_FREE(rgbBuffer);
    }

    /* Avoid compiler warning */
    rowOffset = rowOffset;
    rowCount = rowCount;

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert the byte ordering of arbitrary imagery to PIXR order.
 *
 * LongDescription:::
 *  wp_TfxPcToPixr() converts the byte ordering of arbitrary imagery to PIXR order.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  inImg::			Image to convert
 *  inBPL::			Bytes per image line in inImg
 *
 * Output:::
 *  outImg::			Image to receive the converted data
 *  outBPL::			Bytes per image line in outImg
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxPcToPixr(
	 TFX_IMG *tfxImg,
	 UInt8 **inImg,
	 UInt32 inBPL,
	 UInt8 **outImg,
	 UInt32 outBPL
)
{
    UInt32 row, rows;
    UInt32 quad, quads;
    UInt8 *in;
    UInt32 *out;
    PROLOGUE;

    rows = tfxImg->imgInfo.dwHeight;
    quads = (inBPL + 3) / 4;

    for (row = 0; row < rows; row++) {
	in = *inImg + inBPL * row;
        out = (void *)(*outImg + outBPL * row);
	for (quad = 0; quad < quads; quad++) {
	    *out++ = (in[0] << 24) | (in[1] << 16) | (in[2] << 8) | in[3];
	    in += 4;
	}
    }

    NORMAL_RETURN;
}



static UInt8 _bitFlipLUT[256] = {
      0, 128,  64, 192,  32, 160,  96, 224,
     16, 144,  80, 208,  48, 176, 112, 240,
      8, 136,  72, 200,  40, 168, 104, 232,
     24, 152,  88, 216,  56, 184, 120, 248,
      4, 132,  68, 196,  36, 164, 100, 228,
     20, 148,  84, 212,  52, 180, 116, 244,
     12, 140,  76, 204,  44, 172, 108, 236,
     28, 156,  92, 220,  60, 188, 124, 252,
      2, 130,  66, 194,  34, 162,  98, 226,
     18, 146,  82, 210,  50, 178, 114, 242,
     10, 138,  74, 202,  42, 170, 106, 234,
     26, 154,  90, 218,  58, 186, 122, 250,
      6, 134,  70, 198,  38, 166, 102, 230,
     22, 150,  86, 214,  54, 182, 118, 246,
     14, 142,  78, 206,  46, 174, 110, 238,
     30, 158,  94, 222,  62, 190, 126, 254,
      1, 129,  65, 193,  33, 161,  97, 225,
     17, 145,  81, 209,  49, 177, 113, 241,
      9, 137,  73, 201,  41, 169, 105, 233,
     25, 153,  89, 217,  57, 185, 121, 249,
      5, 133,  69, 197,  37, 165, 101, 229,
     21, 149,  85, 213,  53, 181, 117, 245,
     13, 141,  77, 205,  45, 173, 109, 237,
     29, 157,  93, 221,  61, 189, 125, 253,
      3, 131,  67, 195,  35, 163,  99, 227,
     19, 147,  83, 211,  51, 179, 115, 243,
     11, 139,  75, 203,  43, 171, 107, 235,
     27, 155,  91, 219,  59, 187, 123, 251,
      7, 135,  71, 199,  39, 167, 103, 231,
     23, 151,  87, 215,  55, 183, 119, 247,
     15, 143,  79, 207,  47, 175, 111, 239,
     31, 159,  95, 223,  63, 191, 127, 255,
};

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reverse the order of the bits in each byte.
 *
 * LongDescription:::
 *  wp_TfxBitFlip() reverses the order of the bits in each byte of a buffer.
 *  This is the function required by TIFF field FillOrder set to 2.
 *
 * Input:::
 *  buffer::			Data buffer
 *  bufSize::			Size of data buffer, in bytes
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxBitFlip(
	      UInt8 *buffer,
	      UInt32 bufSize
)
{
    UInt8 *ptr = buffer;
#if 0
    static char computed = 0;
#endif
    PROLOGUE;

#if 0
    if (! computed) {
	Int32 i, mask;
	FILE *fp = fopen("bitflip.lut", "wb");
	UInt8 flip;

	for (i=0; i < 256; i++) {
	    mask = 0x80, flip = 0;
	    while (mask) {
		flip >>= 1;
		if (i & mask) {
		    flip |= 0x80;
		}
		mask >>= 1;
	    }
	    fprintf(fp, "%3d\n", flip);
	    _bitFlipLUT[i] = flip;
	}
	fclose(fp);
	computed = 1;
    }
#endif

    while (bufSize--) {
	*ptr = _bitFlipLUT[*ptr];
	ptr++;
    }

    NORMAL_RETURN;
}

#endif /* NOIPCORE */


/*****************************************************************************
 * The following is the RGB->LAB and LAB->RGB code:
 ****************************************************************************/

#ifndef NOIPCORE

#ifndef USELINEARRGB

/*****************************************************************************
 * Used for the RGB->LAB conversion:
 ****************************************************************************/
#define N_BITS_F	12	/* Works best at 14, but costs extra time */
#define N_F		(1 << N_BITS_F)
#define KZERO_BITS	12	/* If you change, see srgbLUT[] */
#define KONE_BITS	14
#define KTWO_BITS	3
#define KZERO		(1 << KZERO_BITS)
#define KONE		(1 << KONE_BITS)
#define KTWO		(1 << KTWO_BITS)
#define EIGHTBITFROMINT(x)	(unsigned char)((x) < 0? 0: (x) > 255? 255: ((x)))
#define SIGNEDEIGHTBITFROMINT(x) (unsigned char)((x) < 0? (((x) < -128)? 128: 256 + (x)): (x) > 127? 127: ((x)))
#define INTLAB(x)	((x) >> (KONE_BITS + KTWO_BITS))
#define KONMO		(1 << (KZERO_BITS + KONE_BITS - N_BITS_F - 1))
#define f(x)		(ctab[((x) + KONMO) >> (KZERO_BITS + KONE_BITS - N_BITS_F)])

/*****************************************************************************
 * Used for the LAB->RGB conversion:
 ****************************************************************************/
#define N_BITS_F_INV	10
#define N_F_INV		(1 << N_BITS_F_INV)
#define KTHREE_BITS	22
#define KFOUR_BITS	14U
#define KFIVE_BITS	9U
#define KSIX_BITS	10  /* works best at 14, but costs extra time */
#define KTHREE		(1 << KTHREE_BITS)
#define KFOUR		(1 << KFOUR_BITS)
#define KFIVE		(1 << KFIVE_BITS)
#define KSIX		(1 << KSIX_BITS)

#define EIGHTBITRGB(x)	(unsigned char)((x) < 0? 0: (x) > (KFOUR * KFIVE * 255)? 255: \
			(x) >> (KFOUR_BITS + KFIVE_BITS))
/* If negative, use the first array element.
 * If greater than the maximum value less the added rounding factor, use the
 * last array element.
 * Otherwise, round, shift to KSIX elements, and use the LUT.
 */
#define ARRAY_MASK_INV	~(0xffffffff << (KTHREE_BITS))
#define f_inv(x)	(((x) < 0)? \
	ctab[0]: \
	((unsigned)(x) > (ARRAY_MASK_INV)? ctab[N_F_INV]: \
	ctab[(x) >> (KTHREE_BITS - N_BITS_F_INV)]))

/* If negative, use the first array element.
 * If greater than the maximum value, less the added rounding factor, use the
 * last array element.
 * Otherwise, round, and shift to KSIX elements, and use the LUT.
 */
#define KSMO		(1 << (KSIX_BITS - 1))
#define ARRAY_MASK_SRGB	~(0xffffffff << (KFOUR_BITS + KFIVE_BITS))
#define sRGBp2sRGB8(x)	(((x) < 0)? ptab[0]: \
	((unsigned)(x) > (ARRAY_MASK_SRGB - KSMO)? ptab[KSIX - 1]: \
	ptab[((x) + KSMO) >> (KFOUR_BITS + KFIVE_BITS - KSIX_BITS)]))


#else	/* Here's the linear-RGB values: (might need optimization, see comments
	 * above) */

/*****************************************************************************
 * Used for the RGB->LAB conversion:
 ****************************************************************************/
#define N_BITS_F	10
#define N_F		(1 << N_BITS_F)
#define KONE_BITS	21
#define KTWO_BITS	1
#define KONE		(1 << KONE_BITS)
#define KTWO		(1 << KTWO_BITS)
#define EIGHTBITLAB(x)	(unsigned char)((x) < 0? 0: (x) > (KONE * KTWO * 255)? 255: \
			(x) >> (KONE_BITS + KTWO_BITS))
#define KONMO		(1 << (KONE_BITS - N_BITS_F - 1))
#define f(x)		(ctab[((x) + KONMO) >> (KONE_BITS - N_BITS_F)])

/*****************************************************************************
 * Used for the LAB->RGB conversion:
 ****************************************************************************/
#define N_BITS_F_INV	10
#define N_F_INV		(1 << N_BITS_F_INV)
#define KTHREE_BITS	22
#define KFOUR_BITS	15
#define KFIVE_BITS	4
#define KTHREE		(1 << KTHREE_BITS)
#define KFOUR		(1 << KFOUR_BITS)
#define KFIVE		(1 << KFIVE_BITS)

#define EIGHTBITRGB(x)	(unsigned char)((x) < 0? 0: (x) > (KFOUR * KFIVE * 255)? 255: \
			(x) >> (KFOUR_BITS + KFIVE_BITS))
#define KTNMO		(1 << (KTHREE_BITS - N_BITS_F_INV - 1))
#define ARRAY_MASK_INV	(0xffffffffUL << (KTHREE_BITS))
#define f_inv(x)	(((unsigned)(x) & ARRAY_MASK_INV)? \
	(((x) < 0)? ctab[0]: ctab[N_F_INV]): \
	ctab[((x) + KTNMO) >> (KTHREE_BITS - N_BITS_F_INV)])

#endif	/* USELINEARRGB */


Int32 *
srgb_lab_table(void)
{
    Int32 i;
    float f;
    static Int32 rgb_lab_tab[N_F+1];
    static int initialized = 0;

    if (! initialized) {
	for (i = 0; i < N_F+1; i++) {
	    f = (float)i / (float)(N_F - 1);
	    if (f <= 0.008856f) {
		f = 7.787f * f + 16.0f / 116.0f;
	    } else {
		f = cuberoot(f, 0.00001f);
	    }
	    rgb_lab_tab[i] = (Int32)(f * (float)KONE);
	}
	initialized = 1;
    }

    return rgb_lab_tab;
}

typedef struct LabRGBTables {
    Int32 *lab_rgb_tab;
    UInt8 *srp_sr8_tab;
} LabRGBTables;


LabRGBTables *
lab_rgb_tables(void)
{
    Int32 i;
    float f;
    static Int32 lab_rgb_tab[N_F_INV+1];
    static UInt8 srp_sr8_tab[KSIX];
    static LabRGBTables tables = {lab_rgb_tab, srp_sr8_tab};
    static int initialized = 0;

    if (! initialized) {
	for (i = 0; i < N_F_INV+1; i++) {
	    f = (float)i / (float)(N_F_INV - 1);
	    if (f < 0.20689303f) {
		f = ((f) - 16.0f / 116.0f) / 7.787f;
	    } else {
		f = f * f * f;
	    }
	    lab_rgb_tab[i] = (Int32)(f * (float)KFOUR);
	}

	for (i = 0; i < KSIX; i++) {
	    f = (float)i / (float)(KSIX - 1);
	    if (f <= 0.00313) {
		f = f * 12.92f * 256.0f;
	    } else {
		f = ((float)pow((double)f, 1.0 / 2.4) * 1.055f - 0.055f) * 256.0f;
	    }
	    if (f > 255.0f) {
		srp_sr8_tab[i] = 255;
	    } else if (f < 0.0f) {
		srp_sr8_tab[i] = 0;
	    } else {
		srp_sr8_tab[i] = (UInt8)f;
	    }
	}
	initialized = 1;
    }

    return &tables;
}

ErrCode CDECL 
sRGB2Lab(
    IPWState    *pIPWState,
    UInt8	*image[3],
    UInt32	w,
    UInt32	h,
    UInt32	bpl,
    Int32	cspace,
    Int32	*gamut
    )
{
    UInt8 *rRow = image[0];
    UInt8 *gRow = image[1];
    UInt8 *bRow = image[2];
    UInt8 *rp, *gp, *bp;
    UInt32 row, col;
#ifdef USEFLT
    Int16 L, a, b;
    UInt8 R, G, B;
#else
    Int32 X, Y, Z, L, a, b;
    Int32 R, G, B;
    Int32 ebL, eba, ebb;	/* Eight-Bit L, a, b */
    Int32 fy;
#endif
    Int32 x1, x2, x3, y1, y2, y3, z1, z2, z3;
    float Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;	/* White point */
    float Lmin = 0.0f, Lmax = 100.0f;
    float amin = -128.0f, amax = 127.0f;
    float bmin = -128.0f, bmax = 127.0f;
    Int32 const1, const2, const3, const4, const5, const6;
    /* RGB to Lab power function */
    Int32 *ctab;
    /* sRGB8 to sRGB', precomputed for KZERO_BITS = 12 */
    Int32 srgbLUT[256] = {
	0,     1,     2,     4,     5,     6,     7,     9,
	10,    11,    12,    14,    15,    16,    18,    20,
	21,    23,    25,    27,    29,    31,    33,    35,
	37,    40,    42,    45,    48,    50,    53,    56,
	59,    62,    66,    69,    72,    76,    79,    83,
	87,    91,    95,    99,   103,   107,   112,   116,
	121,   126,   131,   136,   141,   146,   151,   156,
	162,   168,   173,   179,   185,   191,   197,   204,
	210,   217,   223,   230,   237,   244,   251,   258,
	265,   273,   280,   288,   296,   304,   312,   320,
	329,   337,   346,   354,   363,   372,   381,   390,
	400,   409,   419,   429,   438,   448,   458,   469,
	479,   490,   500,   511,   522,   533,   544,   556,
	567,   579,   590,   602,   614,   626,   639,   651,
	664,   676,   689,   702,   715,   729,   742,   756,
	769,   783,   797,   811,   826,   840,   855,   869,
	884,   899,   914,   930,   945,   961,   976,   992,
	1008,  1025,  1041,  1058,  1074,  1091,  1108,  1125,
	1142,  1160,  1177,  1195,  1213,  1231,  1249,  1268,
	1286,  1305,  1324,  1343,  1362,  1381,  1400,  1420,
	1440,  1460,  1480,  1500,  1521,  1541,  1562,  1583,
	1604,  1625,  1647,  1668,  1690,  1712,  1734,  1756,
	1778,  1801,  1824,  1846,  1869,  1893,  1916,  1940,
	1963,  1987,  2011,  2035,  2060,  2084,  2109,  2134,
	2159,  2184,  2210,  2235,  2261,  2287,  2313,  2339,
	2366,  2392,  2419,  2446,  2473,  2501,  2528,  2556,
	2584,  2612,  2640,  2668,  2697,  2725,  2754,  2783,
	2813,  2842,  2872,  2902,  2931,  2962,  2992,  3022,
	3053,  3084,  3115,  3146,  3178,  3209,  3241,  3273,
	3305,  3338,  3370,  3403,  3436,  3469,  3502,  3535,
	3569,  3603,  3637,  3671,  3705,  3740,  3775,  3810,
	3845,  3880,  3916,  3951,  3987,  4023,  4060,  4096
    };	
    PROLOGUE;

#if 0
    /* Create a LUT to go from 8-bit sRGB to sRGB (srgbLUT[], found above) */
    {
	FILE *fp = fopen("srgb2srgbp.h", "w");
	Int32 i;
	double tmp;
	for (i = 0; i < 256; i++) {
	    /* Convert from 8-bit sRGB to sRGB' to sRGB */
	    if (i <= 10) {
		srgbLUT[i] = (Int32)((float)i / 255.0f / 12.92f
		    * (float)KZERO + 0.5f);
	    } else {
		tmp = (double)((((float)i / 255.0f) + 0.055f) / 1.055f);
		srgbLUT[i] = (Int32)((float)pow(tmp, 2.4)
		    * (float)KZERO + 0.5f);
	    }
	    fprintf(fp, "%5d, ", srgbLUT[i]);
	}
	fclose(fp);
    }
#endif

    /* If a different gamut exists, grab it */
    if (gamut && gamut[1] && gamut[3]) {
	float f1, f2;

	f1 = (float)gamut[0] / (float)gamut[1];
	f2 = (float)gamut[2] / (float)gamut[3];
	if (f1 < f2) {
	    Lmin = f1;
	    Lmax = f2;
	} else {
	    XERR_GEN_CLEANUP(f1 > f2, IO_BADPARAMETER);
	}
	if (gamut[5] && gamut[7]) {
	    f1 = (float)gamut[4] / (float)gamut[5];
	    f2 = (float)gamut[6] / (float)gamut[7];
	}
	if (f1 < f2) {
	    amin = f1;
	    amax = f2;
	} else {
	    XERR_GEN_CLEANUP(f1 > f2, IO_BADPARAMETER);
	}
	if (gamut[9] && gamut[11]) {
	    f1 = (float)gamut[8] / (float)gamut[9];
	    f2 = (float)gamut[10] / (float)gamut[11];
	}
	if (f1 < f2) {
	    bmin = f1;
	    bmax = f2;
	} else {
	    XERR_GEN_CLEANUP(f1 > f2, IO_BADPARAMETER);
	}

    } else {
	/* Pick min/max for L*, a* and b* based on the desired colorspace */
	switch (cspace) {
	case ITULAB:
	    amin = -21760.0f/255.0f, amax = 21590.0f/255.0f;
	    bmin = -19200.0f/255.0f, bmax = 31800.0f/255.0f;
	    /* Note, here is the computed full spectrum sRGB gamut (softcopy):
	     * amin = -88.0f, amax = 99.0f;
	     * bmin = -108.8f, bmax = 95.2f;
	     */
	    break;
	case CIELAB:
	case ICCLAB:
	    /* Use the defaults, set above */
	    break;
	default:
	    goto ERR_CLEANUP;
	}
    }

    /* Precompute the power function */
    ctab = srgb_lab_table();

    /* Precompute the constants */
    const1 = (Int32)((116.0f * 256.0f / (Lmax - Lmin)) * KTWO);
    const2 = (Int32)((-(16.0f + Lmin) * 256.0f / (Lmax - Lmin)) * KONE * KTWO);
    const3 = (Int32)((500.0f * 256.0f / (amax - amin)) * KTWO);
    const4 = (Int32)((-amin * 256.0f / (amax - amin)) * KONE * KTWO);
    const5 = (Int32)((200.0f * 256.0f / (bmax - bmin)) * KTWO);
    const6 = (Int32)((-bmin * 256.0f / (bmax - bmin)) * KONE * KTWO);
    x1 = (UInt32)((0.4124f / Xn) * KONE);
    x2 = (UInt32)((0.3576f / Xn) * KONE);
    x3 = (UInt32)((0.1805f / Xn) * KONE);
    y1 = (UInt32)((0.2126f / Yn) * KONE);
    y2 = (UInt32)((0.7152f / Yn) * KONE);
    y3 = (UInt32)((0.0722f / Yn) * KONE);
    z1 = (UInt32)((0.0193f / Zn) * KONE);
    z2 = (UInt32)((0.1192f / Zn) * KONE);
    z3 = (UInt32)((0.9505f / Zn) * KONE);


    /* Convert the image, a pixel at a time */
    for (row = 0; row < h; row++) {
	rp = rRow, gp = gRow, bp = bRow;
	for (col = 0; col < w; col++) {
	    R = UCHAR_ACCESS(rp);
	    G = UCHAR_ACCESS(gp);
	    B = UCHAR_ACCESS(bp);

#ifdef USEFLT
	    switch (cspace) {
	    case CIELAB:
		IO_sRGB2CIELAB(R, G, B, &L, &a, &b);
		UCHAR_ACCESS(rp) = (UInt8)L;
		SCHAR_ACCESS(gp) = (Int8)a;
		SCHAR_ACCESS(bp) = (Int8)b;
		break;
	    case ICCLAB:
		IO_sRGB2ICCLAB(R, G, B, &L, &a, &b);
		UCHAR_ACCESS(rp) = (UInt8)L;
		UCHAR_ACCESS(gp) = (UInt8)a;
		UCHAR_ACCESS(bp) = (UInt8)b;
		break;
	    case ITULAB:
		IO_sRGB2ITULAB(R, G, B, &L, &a, &b, gamut);
		UCHAR_ACCESS(rp) = (UInt8)L;
		UCHAR_ACCESS(gp) = (UInt8)a;
		UCHAR_ACCESS(bp) = (UInt8)b;
		break;
	    default:
		goto ERR_CLEANUP;
	    }
#else
	    /* Convert from RGB -> XYZ -> L*a*b* */
	    X = srgbLUT[R] * x1 + srgbLUT[G] * x2 + srgbLUT[B] * x3;
	    Y = srgbLUT[R] * y1 + srgbLUT[G] * y2 + srgbLUT[B] * y3;
	    Z = srgbLUT[R] * z1 + srgbLUT[G] * z2 + srgbLUT[B] * z3;

	    fy = f(Y);
	    L = const1 * fy + const2;
	    a = const3 * (f(X) - fy) + const4;
	    b = const5 * (fy - f(Z)) + const6;

	    /* Compute the integer value */
	    ebL = INTLAB(L);
	    eba = INTLAB(a);
	    ebb = INTLAB(b);

	    /* If mapping back to gamut, do here */

	    /* Output the 8-bit result */
	    if (cspace == CIELAB) {
		UCHAR_ACCESS(rp) = EIGHTBITFROMINT(ebL);
		eba -= 128;
		SCHAR_ACCESS(gp) = SIGNEDEIGHTBITFROMINT(eba);
		ebb -= 128;
		SCHAR_ACCESS(bp) = SIGNEDEIGHTBITFROMINT(ebb);
	    } else {
		UCHAR_ACCESS(rp) = EIGHTBITFROMINT(ebL);
		UCHAR_ACCESS(gp) = EIGHTBITFROMINT(eba);
		UCHAR_ACCESS(bp) = EIGHTBITFROMINT(ebb);
	    }
#endif
	    rp++, gp++, bp++;
	}

	/* Go to the next row of data */
	rRow += bpl;
	gRow += bpl;
	bRow += bpl;

    }

    NORMAL_RETURN;

 ERR_CLEANUP:
    pIPWState = pIPWState;	/* Avoid compiler warning */
    ERR_RETURN;
}



ErrCode CDECL 
Lab2sRGB(
    IPWState    *pIPWState,
    UInt8	*image[3],
    UInt32	w,
    UInt32	h,
    UInt32	bpl,
    Int32	cspace,
    Int32	*gamut
    )
{
    UInt8 *rRow = image[0];
    UInt8 *gRow = image[1];
    UInt8 *bRow = image[2];
    UInt8 *rp, *gp, *bp;
    UInt32 row, col;
#ifdef USEFLT
    Int16 L, a, b;
    UInt8 R, G, B;
#else
    Int32 L, a, b;
    Int32 X, Y, Z, R, G, B;
    Int32 c, t;
#endif
    Int32 r1, r2, r3, g1, g2, g3, b1, b2, b3;
    float Xn = 0.9505f, Yn = 1.000f, Zn = 1.089f;	/* White point */
    float Lmin = 0.0f, Lmax = 100.0f;
    float amin = -128.0f, amax = 127.0f;
    float bmin = -128.0f, bmax = 127.0f;
    Int32 const1, const2, const3, const4, const5, const6;
    LabRGBTables *tables;
    /* f_inv() LUT */
    Int32 *ctab;
    /* sRGB to 8-bit sRGB power function LUT */
    UInt8 *ptab;
    PROLOGUE;


    /* If a different gamut exists, grab it */
    if (gamut && gamut[1] && gamut[3]) {
	float f1, f2;

	f1 = (float)gamut[0] / (float)gamut[1];
	f2 = (float)gamut[2] / (float)gamut[3];
	if (f1 < f2) {
	    Lmin = f1;
	    Lmax = f2;
	} else {
	    XERR_GEN_CLEANUP(f1 > f2, IO_BADPARAMETER);
	}
	if (gamut[5] && gamut[7]) {
	    f1 = (float)gamut[4] / (float)gamut[5];
	    f2 = (float)gamut[6] / (float)gamut[7];
	}
	if (f1 < f2) {
	    amin = f1;
	    amax = f2;
	} else {
	    XERR_GEN_CLEANUP(f1 > f2, IO_BADPARAMETER);
	}
	if (gamut[9] && gamut[11]) {
	    f1 = (float)gamut[8] / (float)gamut[9];
	    f2 = (float)gamut[10] / (float)gamut[11];
	}
	if (f1 < f2) {
	    bmin = f1;
	    bmax = f2;
	} else {
	    XERR_GEN_CLEANUP(f1 > f2, IO_BADPARAMETER);
	}
    } else {

	/* Pick min/max for L*, a* and b* based on the desired colorspace */
	switch (cspace) {
	case ITULAB:
	    amin = -21760.0f/255.0f, amax = 21590.0f/255.0f;
	    bmin = -19200.0f/255.0f, bmax = 31800.0f/255.0f;
	    /* Note, here is the computed full spectrum sRGB gamut (softcopy):
	     * amin = -88.0f, amax = 99.0f;
	     * bmin = -108.8f, bmax = 95.2f;
	     */
	    break;
	case CIELAB:
	case ICCLAB:
	    /* Use the defaults, set above */
	    break;
	default:
	    goto ERR_CLEANUP;
	}
    }

    /* Precompute the power function */
    tables = lab_rgb_tables();
    ctab = tables->lab_rgb_tab;
    ptab = tables->srp_sr8_tab;

    /* Precompute the constants */
    const1 = (Int32)(((Lmax - Lmin) / (255.0f * 116.0f)) * KTHREE);
    const2 = (Int32)(((Lmin + 16.0f) / 116.0f) * KTHREE);
    const3 = (Int32)(((amax - amin) / (255.0f * 500.0f)) * KTHREE);
    const4 = (Int32)((amin / 500.0f) * KTHREE);
    const5 = (Int32)(((bmax - bmin) / (255.0f * 200.0f)) * KTHREE);
    const6 = (Int32)((-bmin / 200.0f) * KTHREE);

    r1 = (Int32)(( 3.2410f * Xn) * KFIVE);
    r2 = (Int32)((-1.5374f * Yn) * KFIVE);
    r3 = (Int32)((-0.4986f * Zn) * KFIVE);
    g1 = (Int32)((-0.9692f * Xn) * KFIVE);
    g2 = (Int32)(( 1.8760f * Yn) * KFIVE);
    g3 = (Int32)(( 0.0416f * Zn) * KFIVE);
    b1 = (Int32)(( 0.0556f * Xn) * KFIVE);
    b2 = (Int32)((-0.2040f * Yn) * KFIVE);
    b3 = (Int32)(( 1.0570f * Zn) * KFIVE);

    /* Convert the image, a pixel at a time */
    for (row = 0; row < h; row++) {
	rp = rRow, gp = gRow, bp = bRow;
	for (col = 0; col < w; col++) {
	    if (cspace == CIELAB) {
		L = UCHAR_ACCESS(rp);
		a = SCHAR_ACCESS(gp) + 128;	/* The computation worked out assumes unsigned */
		b = SCHAR_ACCESS(bp) + 128;	/* The computation worked out assumes unsigned */
	    } else {
		L = UCHAR_ACCESS(rp);
		a = UCHAR_ACCESS(gp);
		b = UCHAR_ACCESS(bp);
	    }

#ifdef USEFLT
	    switch (cspace) {
	    case CIELAB:
		/* Adjust back to signed */
		a -= 128, b -= 128;
		IO_CIELAB2sRGB(L, a, b, &R, &G, &B);
		break;
	    case ICCLAB:
		IO_ICCLAB2sRGB(L, a, b, &R, &G, &B);
		break;
	    case ITULAB:
		IO_ITULAB2sRGB(L, a, b, &R, &G, &B, gamut);
		break;
	    default:
		goto ERR_CLEANUP;
	    }

	    UCHAR_ACCESS(rp) = R;
	    UCHAR_ACCESS(gp) = G;
	    UCHAR_ACCESS(bp) = B;
#else
	    /* Convert from ITULAB -> XYZ: */
	    c = const1 * L + const2;
	    t = const3 * a + const4 + c;
	    X = f_inv(t);
	    Y = f_inv(c);
	    t = c - b * const5 + const6;
	    Z = f_inv(t);

	    /* Note: real XYZ values are these divided by KFOUR (=2^14 =16384) */

	    /* Convert from XYZ to RGB */
	    R = r1 * X + r2 * Y + r3 * Z;
	    G = g1 * X + g2 * Y + g3 * Z;
	    B = b1 * X + b2 * Y + b3 * Z;

	    /* Note: realy RGB values are these divided by KFOUR*KFIVE (=2^23 =8388608) */

	    /* Convert from linear-RGB to sRGB' to 8-bit sRGB */
	    /* Rail and integerize to 8-bit */
	    UCHAR_ACCESS(rp) = sRGBp2sRGB8(R);
	    UCHAR_ACCESS(gp) = sRGBp2sRGB8(G);
	    UCHAR_ACCESS(bp) = sRGBp2sRGB8(B);
#endif
	    rp++, gp++, bp++;
	}

	/* Go to the next row of data */
	rRow += bpl;
	gRow += bpl;
	bRow += bpl;

    }

    NORMAL_RETURN;

 ERR_CLEANUP:
    pIPWState = pIPWState;	/* Avoid compiler warning */
    ERR_RETURN;
}


/* Convert from gray to ITULAB L* (with gamut limits) */
static ErrCode CDECL 
gray2ITUL(
    IPWState    *pIPWState,
    UInt8	*image,
    UInt32	w,
    UInt32	h,
    UInt32	bpl,
    Int32	*gamut
    )
{
    UInt8 *lRow = image;
    UInt8 *lp;
    UInt32 row, col;
    UInt8 gray_itulab_tab[256];
    Int32 i;
    float f;
    UInt8 gray;
    float Lmin = 0.0f, Lmax = 100.0f;
    PROLOGUE;

    /* If a different gamut exists, grab it */
    if (gamut && gamut[1] && gamut[3]) {
	float f1, f2;

	f1 = (float)gamut[0] / (float)gamut[1];
	f2 = (float)gamut[2] / (float)gamut[3];
	if (f1 < f2) {
	    Lmin = f1;
	    Lmax = f2;
	} else {
	    XERR_GEN_CLEANUP(f1 > f2, IO_BADPARAMETER);
	}
    }

    if (Lmax - Lmin <= 0.0f || (Lmin == 0.0f && Lmax == 100.0f)) {
	NORMAL_RETURN;
    }

    /* Precompute the mapping for all 256 values */
    for (i = 0; i < 256; i++) {
	f = (float)i / 255.0f * 100.0f;
	if (f < Lmin) {
	    f = Lmin;
	} else if (f > Lmax) {
	    f = Lmax;
	}
	f -= Lmin;
	f *= 256.0f / (Lmax - Lmin);
	gray_itulab_tab[i] = f < 0.0f? 0: f > 255.0f? 255: (UInt8)f;
    }

    /* Convert the image, a pixel at a time */
    for (row = 0; row < h; row++) {
	lp = lRow;
	for (col = 0; col < w; col++) {
	    gray = UCHAR_ACCESS(lp);

	    /* Convert from gray to ITULAB L*, store the result */
	    UCHAR_ACCESS(lp) = gray_itulab_tab[gray];
	    lp++;
	}

	/* Go to the next row of data */
	lRow += bpl;
    }

    NORMAL_RETURN;

 ERR_CLEANUP:
    pIPWState = pIPWState;	/* Avoid compiler warning */
    ERR_RETURN;
}


/* Convert from ITULAB L* (with gamut limits) to gray */
static ErrCode CDECL 
ITUL2Gray(
    IPWState    *pIPWState,
    UInt8	*image,
    UInt32	w,
    UInt32	h,
    UInt32	bpl,
    Int32	*gamut
    )
{
    UInt8 *lRow = image;
    UInt8 *lp;
    UInt32 row, col;
    UInt8 itulab_gray_tab[256];
    Int32 i;
    float f;
    UInt8 gray;
    float Lmin = 0.0f, Lmax = 100.0f;
    PROLOGUE;

    /* If a different gamut exists, grab it */
    if (gamut && gamut[1] && gamut[3]) {
	float f1, f2;

	f1 = (float)gamut[0] / (float)gamut[1];
	f2 = (float)gamut[2] / (float)gamut[3];
	if (f1 < f2) {
	    Lmin = f1;
	    Lmax = f2;
	} else {
	    XERR_GEN_CLEANUP(f1 > f2, IO_BADPARAMETER);
	}
	
    }

    if (Lmax - Lmin <= 0.0f || (Lmin == 0.0f && Lmax == 100.0f)) {
	NORMAL_RETURN;
    }

    /* Precompute the mapping for all 256 values */
    for (i = 0; i < 256; i++) {
	f = Lmin + (float)i / 255.0f * (float)(Lmax - Lmin);
	itulab_gray_tab[i] = f < 0.0f? 0: f > 255.0f? 255: (UInt8)f;
    }

    /* Convert the image, a pixel at a time */
    for (row = 0; row < h; row++) {
	lp = lRow;
	for (col = 0; col < w; col++) {
	    gray = UCHAR_ACCESS(lp);

	    /* Convert from gray to ITULAB L*, store the result */
	    UCHAR_ACCESS(lp) = itulab_gray_tab[gray];
	    lp++;
	}

	/* Go to the next row of data */
	lRow += bpl;
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    pIPWState = pIPWState;	/* Avoid compiler warning */
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert an image of one of the three LAB flavors to sRGB
 *
 * LongDescription:::
 *  wp_LAB2sRGB() converts the image data to sRGB from one of: CIELAB, ICCLAB,
 *  or ITULAB.
 *
 * Input:::
 *  imagePlanes::		Image planes to convert (in place)
 *  width::			Width of image, in pixels
 *  bpl::			Bytes/line
 *  height::			Height of image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_LAB2sRGB(
	TFX_IMG *tfxImg,
	PIXR *pPixr,
	Int32 cspace,
	Int32 *gamut
)
{
    UInt8 *imgChannels[3];
    UInt32 channels, w, h;
    UInt32 bpl;
    PIXR *pSubPixr;
    IPWState *pIPWState = NULL;
    PROLOGUE;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    /* Get the image characteristics, and data pointers */
    XERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, pPixr, (Int32 *)&w));
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, (Int32 *)&h));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, (Int32 *)&bpl));
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, (Int32 *)&channels));

    switch (channels) {
    case 1:
	XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr, &imgChannels[0]));

	/* Do the conversion */
	ITUL2Gray(pIPWState, imgChannels[0], w, h, bpl, gamut);
	break;
    case 3:

	/* Get the Red (L) channel */
	pSubPixr = NULL;
	XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, 0, &pSubPixr));
	XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &imgChannels[0]));
    
	/* Get the Green (a) channel */
	pSubPixr = NULL;
	XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, 1, &pSubPixr));
	XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &imgChannels[1]));
    
	/* Get the Blue (b) channel */
	pSubPixr = NULL;
	XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, 2, &pSubPixr));
	XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &imgChannels[2]));

	/* Do the conversion */
	Lab2sRGB(pIPWState, &imgChannels[0], w, h, bpl, cspace, gamut);
    	break;
    default:
	XERR_GEN_CLEANUP(1, IO_NOSUPPORT);
    }

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert an sRGB image to any of the three LAB flavors
 *
 * LongDescription:::
 *  wp_sRGB2LAB() converts the image data from sRGB to CIELAB, ICCLAB,
 *  or ITULAB
 *
 * Input:::
 *  imagePlanes::		Image planes to convert (in place)
 *  width::			Width of image, in pixels
 *  bpl::			Bytes/line
 *  height::			Height of image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_sRGB2LAB(
	TFX_IMG *tfxImg,
	PIXR *pPixr,
	Int32 cspace,
	Int32 *gamut
)
{
    UInt8 *imgChannels[3];
    UInt32 w, h;
    UInt32 bpl, channels;
    PIXR *pSubPixr;
    IPWState *pIPWState = NULL;
    PROLOGUE;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    /* Get the image characteristics, and data pointers */
    XERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, pPixr, (Int32 *)&w));
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, (Int32 *)&h));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, (Int32 *)&bpl));
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, (Int32 *)&channels));

    switch (channels) {
    case 1:
	XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr, &imgChannels[0]));

	/* Do the conversion */
	gray2ITUL(pIPWState, imgChannels[0], w, h, bpl, gamut);
	break;
    case 3:

	/* Get the Red channel */
	pSubPixr = NULL;
	XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, 0, &pSubPixr));
	XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &imgChannels[0]));
    
	/* Get the Green channel */
	pSubPixr = NULL;
	XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, 1, &pSubPixr));
	XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &imgChannels[1]));
    
	/* Get the Blue channel */
	pSubPixr = NULL;
	XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, 2, &pSubPixr));
	XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &imgChannels[2]));

	sRGB2Lab(pIPWState, &imgChannels[0], w, h, bpl, cspace, gamut);
	break;
    default:
	XERR_GEN_CLEANUP(1, IO_NOSUPPORT);
    }
	
    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


#endif /* NOIPCORE */

/*
$Log: tfxconv.c,v $
Revision 1.1  2003/01/03 01:41:10  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.15  2001/12/26 21:23:44  dabercro
Mark as released

Revision 1.14  2001/07/06 21:38:26  dabercro
Use a struct to pass back the arrays, since gcc complains on Sparc

Revision 1.13  2001/07/06 18:16:05  dabercro
Remove unused variables, consistent use of static, for gcc

Revision 1.12  2001/07/02 20:53:00  dabercro
Added conversion for L* only ITULAB

Revision 1.11  2001/06/29 22:35:33  dabercro
Move the IO_ functions to an io .c  Keep chunky conversion here (needs to
for various reasons).  Propogate gamut.

Revision 1.10  2001/06/29 17:51:34  dabercro
Some cleanup

Revision 1.9  2001/06/27 18:58:51  dabercro
Moved colorspace conversion here, since it is needed for uncompressed
images, and it isn't appropriate to do in JPEG anyways.
Extend conversions to include chunky/planar, and byte order (vs. PIXR).

Revision 1.8  2001/03/13 02:59:12  dabercro
Remove lossy-dither and wavelet as valid compression types.

Revision 1.7  2001/01/18 00:21:30  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.6  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.5  2000/08/22 01:03:23  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.4  2000/04/12 19:06:08  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.3  2000/03/13 23:29:14  dabercro
Use IPCore invert.

Revision 1.2  2000/03/11 02:06:34  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.4  1997/06/26 22:24:34  aber
aber on Thu Jun 26 15:23:49 PDT 1997

tfxconv.c -> /product/ipcore/ipwrap/src/RCS/tfxconv.c,v

Add conversion routine for inverting "white-is-zero" images (TIFF-FX 2.0
reading only).

Revision 1.3  1997/05/30 00:50:56  aber
aber on Thu May 29 17:50:56 PDT 1997

tfxconv.c -> /product/ipcore/ipwrap/src/RCS/tfxconv.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.2  1997/05/20 22:29:00  aber
aber on Tue May 20 15:29:00 PDT 1997

tfxconv.c -> /product/ipcore/ipwrap/src/RCS/tfxconv.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.1  1997/05/17 00:46:33  aber
Initial revision

*/
