/*
 * Trade secret of Xerox Corporation.
 * Copyright (c) 1991-2001, Xerox Corporation.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef TFXDOC_H_INCLUDED
#define TFXDOC_H_INCLUDED

#ifndef NOIPCORE
IP_RCSINFO(tfxdoc_hRCSInfo, "$RCSfile: tfxdoc.h,v $; $Revision: 1.3 $");
/* $Date: 2003/02/11 01:31:42 $ */
#endif

#ifdef __cplusplus
extern "C"
{
#endif

/* tfxdoc.h
 *
 * DESCRIPTION
 *	 Declarations for TIFF-FX Document level structures.
 *
 */

/*
 * INCLUDES
 */
#include "tfxfile.h"
#include "tfx.h"
#include "tfxinfo.h"

#ifdef _MSC_VER
#pragma pack(push, tfxdoc_h, 1)
#endif

/*
 * CONSTANTS
 */

/* Calling sequence status */
#define TFX_DOC_SEQ_GETINFO	0x0001	/* Document information gotten/set */
#define TFX_DOC_SEQ_READSTART	0x0010	/* Read start called */
#define TFX_DOC_SEQ_READFINISH	0x0000	/* Read finish called */
#define TFX_DOC_SEQ_READSTATUS	0x00f0	/* Read status bits */
#define TFX_DOC_SEQ_WRITESTART	0x0100	/* Write start called */
/*#define TFX_DOC_SEQ_W_PAGETABLE	0x0200	// A page table was written */
#define TFX_DOC_SEQ_W_RESOURCETABLE	0x0400	/* A shared resource table was written */
#define TFX_DOC_SEQ_WRITEFINISH	0x0000	/* Write finish called */
#define TFX_DOC_SEQ_WRITESTATUS	0x0f00	/* Write status bits */
#define TFX_DOC_SEQ_ANNOTSET	0x1000	/* Annotation set */
#define TFX_DOC_SEQ_REFPAGESET	0x2000	/* Reference page set */
#define TFX_DOC_SEQ_TOKENSTART	0x4000	/* Tokenization ongoing */
#define TFX_DOC_SEQ_TOKENDICTSAVED	0x8000	/* Tokenization dictionary saved */
#define TFX_DOC_SEQ_MULTIDICTS	0x10000	/* Storing multiple dictionaries
					 * per page */
#define TFX_DOC_SEQ_SHAREDDICT	0x20000	/* Shared dictionary used */
#define TFX_DOC_SEQ_GPEXISTS	0x40000	/* GPIFD is there */
#define TFX_DOC_SEQ_NOGPIFDREQ	0x80000	/* Request not to write GPIFD */

#define READ_ALLOWED		0x1	/* File has been opened for read */
#define WRITE_ALLOWED		0x2	/* File has been opened for write */

/* Token dictionary flags */
#define TOKENDICT_WRITABLE	0x01	/* Dictionary is writable (not written
					 * yet) */
#define TOKENDICT_LOSSLESS	0x02	/* Token dictionary is for lossless
					 * compression only */

/* CodingMethods bits */
typedef enum CodingMethodsBits {
    /* Bit 0: unspecified compression */
    CODINGMETHOD_UNSPECIFIED = 0x00000001,
    /* Bit 1: 1-dimensional coding, ITU-T Rec. T.4 (MH - Modified Huffman) */
    CODINGMETHOD_MH = 0x00000002,
    /* Bit 2: 2-dimensional coding, ITU-T Rec. T.4 (MR - Modified Read) */
    CODINGMETHOD_MR = 0x00000004,
    /* Bit 3: 2-dimensional coding, ITU-T Rec. T.6 (MMR - Modified MR) */
    CODINGMETHOD_MMR = 0x00000008,
    /* Bit 4: ITU-T Rec. T.82 coding, using ITU-T Rec. T.85 (JBIG) */
    CODINGMETHOD_JBIG = 0x00000010,
    /* Bit 5: ITU-T Rec. T.81 (Baseline JPEG) */
    CODINGMETHOD_JPEG = 0x00000020,
    /* Bit 6: ITU-T Rec. T.82 coding, using ITU-T Rec. T.43 (JBIG color) */
    CODINGMETHOD_CJBIG = 0x00000040,
    /* Bit 7: ITU-T Rec. T.88 coding, using ITU-T Rec. T.89 (JBIG2) */
    CODINGMETHOD_JBIG2 = 0x00000080
} CodingMethodsBits;

/*
 * MACROS
 */
#define TFX_DOCFILE(tfxDoc)	tfxDoc->tfxFile
#define TFX_DOCINST(tfxDoc)	TFX_DOCFILE(tfxDoc)->instance
#define TFX_DOCFILETOKEN(tfxDoc)	TFX_FILETOKEN(TFX_DOCFILE(tfxDoc))


/*
 * TYPEDEFS
 */

/* Page IFD offset list item */
typedef struct TFX_IFDOFFSET_ITEM {
    struct TFX_IFDOFFSET_ITEM *next;
    UInt32 ifdOffset;
} TFX_IFDOFFSET_ITEM;

#if 0
/* Page table header */
typedef struct TFX_PAGETABLEHEADER {
    UInt16 pageCount;		    /* Number of pages in document (total) */
    UInt16 entriesPerTable;	    /* Number of page table entries per table
				     */
    UInt16 lastOrdinal;		    /* Ordinal (1, 2, ...) of last table */
    Int32 firstTableOffset;	    /* File offset of first table */
} TFX_PAGETABLEHEADER;

/* Page table entry (describes an individual page) */
typedef struct TFX_PAGETABLEENTRY {
    UInt8 flags;		    /* Page flags (see PAGE_* above) */
    Int32 IFDoffset;		    /* File offset of the IFD */
} TFX_PAGETABLEENTRY;

/* Individual page table, linked if more than one */
typedef struct TFX_PAGETABLE {
    struct TFX_PAGETABLE *next;	    /* Link to next table (cached) */
    UInt16 ordinal;		    /* Ordinal (1, 2, ...) of this table */
    UInt16 validPages;		    /* Number of valid pages in table */
    struct TFX_PAGETABLEENTRY *entries; /* Allocated entry list */
    Int32 nextTableOffset;	    /* File offset of next page table */
} TFX_PAGETABLE;
#endif
	
/* Resource table entry (describes an individual resource) */
typedef struct TFX_RESOURCETABLEENTRY {
    UInt32 bytes;		    /* Number of bytes in shared resource */
    Int32 offset;		    /* File offset of the resource block */
    Int16 resourceType;		    /* Resource type, see TFX_RESOURCE_...
				     * in tfxtiff.h */
    UInt16 lastPage;		    /* Ordinal of last page to use the resource,
				     * 0 if not known */
    UInt32 decMem;		    /* Decoder memory required to hold the
				     * resource, 0 if not known */
} TFX_RESOURCETABLEENTRY;

/* Individual resource table, linked if more than one */
typedef struct TFX_RESOURCETABLE {
    struct TFX_RESOURCETABLE *next; /* Link to next table (cached) */
    UInt16 firstID;		    /* ID of the first physical entry */
    UInt16 physicalEntries;	    /* Physical entries in this table */
    TFX_RESOURCETABLEENTRY *entries;/* Allocated entry list */
    Int32 nextTableOffset;	    /* File offset of next resource table */
} TFX_RESOURCETABLE;

/* Previously copied token dictionaries */
typedef struct TFX_COPIEDDICTS {
    struct TFX_COPIEDDICTS *next;   /* Link to next */
    IO_DOCHANDLE dstDocHandle;	    /* Doc handle of destination */
    UInt32 dstDocID;		    /* Unique doc ID */
    UInt32 dictOffset;		    /* Offset of source dictionary */
    UInt32 newOffset;		    /* Offset of destinary dictionary */
} TFX_COPIEDDICTS;

#ifndef NO_DGP
/* JBIG2 decoded dictionary handles */
typedef struct TFX_DICTHANDLE {
    struct TFX_DICTHANDLE *prev;
    struct TFX_DICTHANDLE *next;
    UInt32 handle;
    UInt32 id;
    UInt8 unused;
} TFX_DICTHANDLE;
#endif

/* Document status */
typedef struct TFX_DOC {
    UInt32 handleID;			/* Handle ID, consistency check */
    UInt32 handleType;		    /* Handle type ID */
    UInt32 docID;			/* Unique document ID, from instance */
    TFX_FILE *tfxFile;			/* File structure */
    UInt32 seqStatus;			/* Calling sequence status */
    TFX_DOCINFO docInfo;		/* Document information */
    UInt8 readWrite;			/* Read/write state (READ_ALLOWED/ */
					/* WRITE_ALLOWED) */
    UInt32 firstResourceTableLocation;	/* Location of first shared resource table */
    Int32 authorOffset;			/* Offset to author block */
    Int16 annotsPerTable;		/* User set annotations per table */
    UInt16 resourcesPerTable;		/* User set resources per table */
    UInt32 dictFlags;			/* Flags associated with current
					 * dictionary */
    struct TFX_IFDOFFSET_ITEM *ifdList;	/* List of IFD offsets for primary IFDs
					*/
    struct TFX_RESOURCETABLE *firstResourceTable;/* Shared resource table for first */
    struct TFX_PAGE *pages;		/* Page handles for pages previously
					 * visited */
#ifndef NOIPCORE
    void *textCompInfo;			/* Pointer to shared text compression
					 * info */
#endif
#ifndef NODGP
    void *tokEncodeHandle;		/* Pointer to tokenization encode handle */
    void *tokDecodeHandle;		/* Pointer to tokenization decode handle */
#endif
    void *dictSection;			/* Dictionary section reference */
    void *outCookie;			/* Output JBIG2 cookie */
    TFX_COPIEDDICTS *prevTDCopied;	/* Previous token dictionaries copied
					 * to other documents */
    UInt32 *curDictIDs;			/* List of token dictionaries
					 * currently loaded */
    UInt32 curDictCount;		/* Number of dictionaries currently
					 * loaded */
    UInt32 nextPageTokenID;		/* Unique page ID for next page */
    UInt32 nextDocTokenID;		/* Unique doc ID for next page */
    UInt32 gpifdOffset;			/* Offset of Global Parameters IFD */
    UInt16 reservedGPFieldCount;	/* Number of GPIFD fields */
    UInt32 tifffxExtensionsLocation;	/* Location of TIFF-FXExtensions field data */
/* Yet to implement (7/24/00): */
    UInt32 multiProfiles;		/* Accumulated MultiProfiles for all pages */
    UInt32 multiProfilesLocation;	/* Location of MultiProfiles field data */

    UInt32 sharedResourcesLocation;	/* Location of SharedResources field data */
    UInt32 refPageOffsetLocation;	/* Location of reference page offset */
    UInt32 modeNumberLocation;		/* Location of ModeNumber field value */
    UInt32 codingMethodsLocation;	/* Location of CodingMethods field value */
    CodingMethodsBits codingMethods;	/* Bit indication of coding methods */

/* TIFF fields propogated to the document */
    Int8 *DocumentName;			/* Name of the document */
    UInt32 DocumentNameBytes;		/* Size of the name */
    Int8 *Author;			/* Name of the author */
    UInt16 AuthorBytes;			/* Size of the author block */
#ifndef NO_DGP
    TFX_DICTHANDLE *dictHandles;	/* Decoded dictionary handles */
    UInt32 dictList;			/* Handle to the JBIG2 dictionaries list */
#endif
} TFX_DOC;

/* Document field */
typedef struct TFX_DOCFIELD {
    UInt16 tagID;			/* Document tag */
    Int32 offset;			/* Offset of extension data */
} TFX_DOCFIELD;


/*
 * PROTOTYPES
 */


#ifdef _MSC_VER
#pragma pack(pop, tfxdoc_h)
#endif

#ifdef __cplusplus
}
#endif

#endif /* TFXDOC_H_INCLUDED */

