/*
 * Trade secret of Xerox Corporation.
 * Copyright (c) 1991-2001, Xerox Corporation.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef TFXFILE_H_INCLUDED
#define TFXFILE_H_INCLUDED

#ifndef NOIPCORE
IP_RCSINFO(tfxfile_hRCSInfo, "$RCSfile: tfxfile.h,v $; $Revision: 1.1 $");
/* $Date: 2003/01/03 01:41:11 $ */
#endif

#ifdef __cplusplus
extern "C"
{
#endif

/* tfxfile.h
 *
 * DESCRIPTION
 *	 Declarations for TIFF-FX file level structures (and above, like instance
 *	 handles).
 *
 */

/*
 * INCLUDES
 */
#include "tfx.h"
#include "iofile.h"


#ifdef _MSC_VER
#pragma pack(push, tfxfile_h, 1)
#endif

/*
 * CONSTANTS
 */
#define TFX_MAGIC_NUM	0x123abc02		/* TFX_INSTANCE identifier */
#define TFX_HANDLE_TYPE	(0x01000000 * 'T' + 0x00010000 * 'F' + 0x00000100 * 'X')

/*
 * MACROS
 */

#define TFX_FILETOKEN(tfxFile)			(&(tfxFile)->fileToken)
#define TFX_READ(tfxFile, buffer, bytes)		\
	if (TFX_FILETOKEN(tfxFile)->FileRead(TFX_FILETOKEN(tfxFile)->dwClientFileID, \
	    (Int8 *)(buffer), bytes) != (bytes)) {	\
	    __error__ = IO_READ_ERR;		\
	    goto ERR_CLEANUP;			\
	}
#define TFX_WRITE(tfxFile, buffer, bytes)		\
	XERR_GEN_CLEANUP(tfxFile->bQueuing, IO_INTERNAL_ERROR);    \
	if (TFX_FILETOKEN(tfxFile)->FileWrite(TFX_FILETOKEN(tfxFile)->dwClientFileID, \
	    (const Int8 *)(buffer), bytes) != (bytes)) {	\
	    __error__ = IO_WRITE_ERR;		\
	    goto ERR_CLEANUP;			\
	}
#define TFX_SEEK(tfxFile, offset)			\
	if (TFX_FILETOKEN(tfxFile)->FileSeek(TFX_FILETOKEN(tfxFile)->dwClientFileID, \
	    offset) != (Int32)(offset)) {	\
	    __error__ = IO_SEEK_ERR;		\
	    goto ERR_CLEANUP;			\
	}
#define TFX_FILESIZE(tfxFile, size)		\
	if (((size) = TFX_FILETOKEN(tfxFile)->FileSize(TFX_FILETOKEN(tfxFile)->\
		dwClientFileID)) < 0) {		\
	    __error__ = IO_FILELEN_ERR;		\
	    goto ERR_CLEANUP;			\
	}

#define TFX_FILETELL(tfxFile)		\
	TFX_FILETOKEN(tfxFile)->FileTell(TFX_FILETOKEN(tfxFile)->dwClientFileID)

/* Malloc and free, must have a local variable *instData which is active. */
#ifdef SHOWERR
#define TFX_MALLOC(bytes)			\
	instData->malloc(bytes, 0, instData->mallocID	\
	, __FILE__, __LINE__	\
)
#else
#define TFX_MALLOC(bytes)			\
	instData->malloc(bytes, 0, instData->mallocID	\
)
#endif

    /* These two aren't a part of TFX_INSTANCE (shouldn't they be) */
#ifdef SHOWERR
#define TFX_CALLOC(nels, count)			\
	instData->calloc(nels, count, 0, instData->mallocID	\
	, __FILE__, __LINE__	\
)
#else
#define TFX_CALLOC(nels, count)			\
	instData->calloc(nels, count, 0, instData->mallocID	\
)
#endif

#ifdef SHOWERR
#define TFX_REALLOC(buffer, bytes)			\
	instData->realloc(buffer, bytes, 0, instData->mallocID	\
	, __FILE__, __LINE__	\
)
#else
#define TFX_REALLOC(buffer, bytes)			\
	instData->realloc(buffer, bytes, 0, instData->mallocID	\
)
#endif

#define TFX_FREE(buffer)			\
	instData->free(buffer, 0, instData->mallocID)

#define IO_CHECKMALLOC(call)			\
	if ((call) == NULL) {			\
	    __error__ = IO_NOMEMORY;		\
	    goto ERR_CLEANUP;			\
	}


#define TFX_CheckHandle(structPtr)		\
	if (structPtr == 0 ||			\
		((TFX_INSTANCE *)structPtr)->handleID != (UInt32)structPtr) {  \
		__error__ = IO_BADPARAMETER;	\
	} else if (((TFX_INSTANCE *)structPtr)->handleType != (UInt32)TFX_HANDLE_TYPE) {  \
		__error__ = IO_WARN_NOSUPPORT;	\
		ERR_RETURN;			\
	}
#define TFX_CheckStructureSize(structPtr)		\
	if (structPtr == NULL ||		\
		structPtr->dwSize != sizeof(*structPtr)) {	\
		__error__ = IO_BADPARAMETER;	\
		ERR_RETURN;			\
	}

#define TFX_SWAPS(p)		wp_TfxSwaps((Int8 *)p)
#define TFX_SWAPL(p)		wp_TfxSwapl((Int8 *)p)
#define TFX_SWAPD(p)		wp_TfxSwapd((Int8 *)p)

/* Conditional swap, depends on file token local variable: tfxFile */
#define TFX_SWAPSHORT(tfxFile, p)	\
	if (TFX_FILETOKEN(tfxFile)->bSwapBytes) { \
		TFX_SWAPS(p); \
	}
#define TFX_SWAPLONG(tfxFile, p)	\
	if (TFX_FILETOKEN(tfxFile)->bSwapBytes) { \
		TFX_SWAPL(p); \
	}
#define TFX_SWAPDOUBLE(tfxFile, p)	\
	if (TFX_FILETOKEN(tfxFile)->bSwapBytes) { \
		TFX_SWAPD(p); \
	}

/*
 * TYPEDEFS
 */

/* Document token
 */
typedef struct TFX_TOKEN
{
    UInt32 dwSize;		/* Size of this structure */
    Int32 dwClientFileID;	/* Client's file identifier (fd, fp, structure, etc.) */
    IO_READFUNC FileRead;	/* Read function, set by client */
    IO_WRITEFUNC FileWrite;	/* Write function, set by client */
    IO_SEEKFUNC FileSeek;	/* Seek function, set by client */
    IO_SIZEFUNC FileSize;	/* File size function, set by client */
    IO_TELLFUNC FileTell;	/* File tell function, set by client */
    UInt32 dwCurPos;		/* Current file position */
    Bool bSwapBytes;		/* File must be byte swapped for endian differences */
} TFX_TOKEN;


typedef struct TFX_FILE {
    UInt32 handleID;			/* Handle ID, consistency check */
    TFX_TOKEN fileToken;		/* File token (see tfx.prv) */
    struct TFX_INSTANCE *instance;	/* TIFF-FX instance */
    Bool bQueuing;			/* Queuing data pending IFD construction */
} TFX_FILE;

typedef struct TFX_INSTANCE {
    UInt32 handleID;		    /* Handle ID, consistency check */
    UInt32 handleType;		    /* Handle type ID */
    UInt32 nextDocID;		    /* Unique ID assigned to each doc */
    UInt32 tfxMagicNum;		    /* TIFF-FX instance identifier */
    UInt32 clientInstID;	    /* Client passed instance identifier */
    IO_MALLOCFUNC malloc;	    /* Client malloc function */
    IO_CALLOCFUNC calloc;	    /* Client calloc function */
    IO_REALLOCFUNC realloc;	    /* Client realloc function */
    IO_FREEFUNC free;		    /* Client free function */
    void *mallocID;		    /* Malloc ID sent to malloc/free */
    void *defaultMallocID;	    /* Reserved default Malloc ID */
    UInt32 pIPWState;		    /* Handle for IPCore IPWrap layer */
    UInt32 myIPWState;		    /* Boolean for private (vs. client)
				     * IPWrap handle */
} TFX_INSTANCE;

/*
 * PROTOTYPES
 */
ErrCode CDECL
wp_TfxFileLock(TFX_TOKEN *fileToken);
ErrCode CDECL
wp_TfxFileUnLock(TFX_TOKEN *fileToken);

#if 0
#ifdef SHOWERR
void tfx_error(const char *file, long line);
#endif
#endif

void wp_TfxSwaps(Int8 *x);
void wp_TfxSwapl(Int8 *x);
void wp_TfxSwapd(Int8 *x);


#ifdef _MSC_VER
#pragma pack(pop, tfxfile_h)
#endif

#ifdef __cplusplus
}
#endif

#endif /* TFXFILE_H_INCLUDED */

