/*
 * Trade secret of Xerox Corporation.
 * Copyright (c) 1991-2001, Xerox Corporation.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef TFXIMAGE_H_INCLUDED
#define TFXIMAGE_H_INCLUDED

#ifndef NOIPCORE
IP_RCSINFO(tfximage_hRCSInfo, "$RCSfile: tfximage.h,v $; $Revision: 1.3 $");
/* $Date: 2003/02/11 01:31:42 $ */
#endif

#ifdef __cplusplus
extern "C"
{
#endif

/* tfximage.h
 *
 * DESCRIPTION
 *	 Declarations for the TIFF-FX Toolkit API.
 *
 */

/*
 * INCLUDES
 */
#ifndef NOIPCORE
#include "ipwcore.prv"
#include "geomadt.h"
#include "props.pub"
#include "rast.pub"
#include "pixr.h"
#endif

#include "tfxpage.h"
#include "tfxinfo.h"
#include "tfxtiff.h"

#ifdef _MSC_VER
#pragma pack(push, tfximage_h, 1)
#endif

/*
 * CONSTANTS
 */
#define MAX_EXPECTED_TAG_COUNT	22	/* Most tags an individual image
				         * is expected to have */

/* Calling sequence status */
#define TFX_IMG_SEQ_GETINFO	0x00000001	/* Get image info called */
#define TFX_IMG_SEQ_READSTART	0x00000010	/* Read start called */
#define TFX_IMG_SEQ_READSTRIP	0x00000020	/* Read strip called */
#define TFX_IMG_SEQ_READFINISH	0x00000000	/* Read finish called */
#define TFX_IMG_SEQ_READSTATUS	0x000000f0	/* Read status bits */
#define TFX_IMG_SEQ_WRITESTART	0x00000100	/* Write start called */
#define TFX_IMG_SEQ_WRITESTRIP	0x00000200	/* Write strip called */
#define TFX_IMG_SEQ_WRITEFINISH	0x00000000	/* Write finish called */
#define TFX_IMG_SEQ_WRITESTATUS	0x00000f00	/* Write status bits */
#define TFX_IMG_SEQ_TOKENSTART	0x00001000	/* Image is in a tokenized state
						 */
#define TFX_IMG_SEQ_MASKADDED		0x00010000	/* Image has had a mask added */
#define TFX_IMG_SEQ_USECOLORSTRIPS	0x00020000	/* Use variable-sized strips */
#define TFX_IMG_SEQ_USEVSS		0x00040000	/* Use variable-sized strips */



/*
 * MACROS
 */
#define TFX_IMGFILE(tfxImg)		TFX_PAGEFILE(tfxImg->tfxPage)
#define TFX_IMGINST(tfxImg)		TFX_IMGFILE(tfxImg)->instance
#define TFX_IMGFILETOKEN(tfxImg)		TFX_FILETOKEN(TFX_IMGFILE(tfxImg))

/*
 * TYPEDEFS
 */

/* A linked list of miscellaneous memory, used to store arbitrary information
 * reused between strips or pages */
typedef struct TFX_MISCMEM {
    struct TFX_MISCMEM *next;
    UInt32 bytes;
    void *mem;
} TFX_MISCMEM;

typedef struct TFX_STRIP {
    struct TFX_STRIP *next;
    UInt32 offset;		/* File offset of strip data */
    UInt32 bytes;		/* Byte count of strip data */
    UInt32 rows;		/* Number of rows in strip */
} TFX_STRIP;

typedef struct TFX_STRIPS {
    struct TFX_STRIP *firstStrip;
    struct TFX_STRIP *lastStrip;
    UInt32 rows;
    UInt32 stripCount;
} TFX_STRIPS;

typedef struct TFX_BASECOLOR {
    int color_set:1;	/* Is this (rgb) color active or not */
    int IFD_color_set:1;/* Is this IFDColor active or not */
    UInt8 rgb[3];	/* Color in sRGB */
    Int32 IFDColor[4];	/* Color in the colorspace and parameters of the IFD.
		         * Note: only if IFD_color_set */
} TFX_BASECOLOR;

typedef struct TFX_IMG {	
    UInt32 handleID;		    /* Handle ID, consistency check */
    UInt32 handleType;		    /* Handle type ID */
    struct TFX_IMG *prev;	    /* Previous (or last) image of this type
				     * in linked list */
    struct TFX_IMG *next;	    /* Next image of this type in linked list
				     */
    struct TFX_IMG *parentImg;	    /* Parent image, if this is a hints img */
    TFX_PAGE *tfxPage;		    /* Page structure of parent */
    UInt32 seqStatus;		    /* Calling sequence status */
    Int32 ifdOffset;		    /* Offset of IFD */
    TFX_IMGINFO imgInfo;	    /* Image information */
    UInt32 reservedFieldCount;	    /* Number of fields that can be written
				     * at IFD location */
    UInt32 imgFlags;		    /* As described during TFX_Image(Read/
				     * Write)Start */
    UInt32 JPEGquality;		    /* JPEG quality, if JPEG compressing */
    UInt8 *qtableBuf;		    /* Extraneous Q-table for JPEG */
    Int32 qtableBufsize;	    /* Bytes in qtableBuf */
    struct TFX_IMG *xHintsImgs;	    /* Image handle to hints image list */
/*    struct TFX_IMG *xMaskImg;	    // Image handle to next (only) */
  //				       transparency mask image */
    struct TFX_IMG *lastSubImg;	    /* Last subimage attached in SubIFDs list
				     */
    /* Field Information */
    TFX_TIFFIELDS fields;	    /* TIFF fields */
    TFX_FIELD_LOCATION fieldLocation;	/* Location of field itself */
    TFX_FIELD_DATA_LOCATION fieldDataLocation;	/* Location of field data */

    /* Intermediate buffering */
    TFX_MISCMEM *tmpCmpData;	    /* Temporary compression info */
    UInt32 minBytesPerRow;	    /* Physical bytes required per row */
    UInt8 *cmpBuffer;		    /* Intermediate compressed buffer */
    UInt32 cmpBufsize;		    /* Compressed buffer capacity */
    UInt32 cmpBytes;		    /* Number of compressed bytes in the
				     * buffer */
#ifndef NOIPCORE
    PIXR *tmpPixr;		    /* Intermediate PIXR buffer (if needed) */
#endif
    /* Note that these are used to keep track of compressed image position
     * and size (not just a real PIXR) */
    UInt32 tmpPixrRow;		    /* Image row offset of image in tmpPixr */
    UInt32 tmpPixrRows;		    /* Number of rows currently in tmpPixr */
    UInt32 posSize;		    /* Position block size of Tundra data */
    UInt32 *dictList;		    /* List of token dictionaries used in
				     * by this image */
    UInt32 dictCount;		    /* Number of dictionaries required by
				     * this image */
    Int32 *dictSizes;		    /* Dictionary sizes (passed back to client,
				     * but preserved here for memory mgmnt) */
    UInt32 headerBytes;		    /* Preserved byte count of JBIG2 header */
    TFX_STRIPS strips;		    /* Strips, when RowStripCounts is used */
    TFX_BASECOLOR baseColor;	    /* Base color, in sRGB */
} TFX_IMG;


/*
 * PROTOTYPES
 */

#ifdef _MSC_VER
#pragma pack(pop, tfximage_h)
#endif

#ifdef __cplusplus
}
#endif

#endif /* TFXIMAGE_H_INCLUDED */

