/*
 * Trade secret of Xerox Corporation.
 * Copyright (c) 1991-2001, Xerox Corporation.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef TFXINFO_H_INCLUDED
#define TFXINFO_H_INCLUDED

#ifndef NOIPCORE
IP_RCSINFO(tfxinfo_hRCSInfo, "$RCSfile: tfxinfo.h,v $; $Revision: 1.3 $");
/* $Date: 2003/02/11 01:31:42 $ */
#endif

#ifdef __cplusplus
extern "C"
{
#endif

#ifdef _MSC_VER
#pragma pack(push, tfxinfo_h, 1)
#endif

/* tfxinfo.h
 *
 * DESCRIPTION
 *	 Declarations for TIFF-FX information
 *
 */

/*
 * INCLUDES
 */

/*
 * CONSTANTS
 */

/*
 * MACROS
 */

/*
 * TYPEDEFS
 */

/*
 * TIFF-FX Document information
 *
 * The following structure describes the data returned by the
 * TFX_GetDocInfo() function.
 */
typedef struct TFX_DOCINFO
{
    UInt8 version;			/* Version of TIFF-FX */
    UInt8 revision;			/* Revision of this version of TIFF-FX */
    UInt8 profile;			/* TIFF-FX profile (see tfx.h) */
    UInt16 wNPages;			/* Number of pages in file */
    Int8 *pAuthorBlock;			/* Author string (NULL if none) */
    UInt16 wAuthorBlockBytes;		/* Size of author string */
    Int32 dwReferencePageOffset;	/* Reference page offset (-1 if none) */
    IO_DOCHANDLE tfxDocHandle;		/* Document handle, used to identify
					 * this document in the document level
					 * routines */
} TFX_DOCINFO;


/*
 * TIFF-FX Page Information
 *
 * The following structure describes the data returned by the
 * TFX_GetPageInfo() function.
 */
typedef struct TFX_PAGEINFO
{
    Int32 dwPageOffset;			/* Page offset */
    UInt32 *pdwImageCounts;		/* Image count for each layer */
    UInt32 dwLayersPresent;		/* Number of layers present (may
					 * include empty layers) */
    UInt16 wAnnotationCount;		/* Number of annotations on page */
    IO_IMGHANDLE tfxPrimImgHandle;	/* Primary image handle */
    IO_PAGEHANDLE tfxPageHandle;		/* Page handle, used to identify this
					 * page in the page level routines */
} TFX_PAGEINFO;


/*
 * TIFF-FX Image information
 *
 * The following structure describes the data returned by the
 * TFX_GetImageInfo() function.
 */
typedef struct TFX_IMGINFO
{
    IO_IMGHANDLE tfxImgHandle;		/* Image handle to this image */
    IO_IMGTYPE tfxImgType;		/* Image type (IO_IMGTYPE_PRIMARY,
					 * IO_IMGTYPE_LAYER or
					 * IO_IMGTYPE_MASK) */
    UInt32 dwImgLayer;			/* Image layer in the MRC model */
    UInt32 dwImgOffset;			/* Image offset within the layer */
    UInt32 dwXOffset;			/* X location in page coordinates */
    UInt32 dwYOffset;			/* Y location in page coordinates */
    UInt32 dwXResolution;		/* X Resolution of image in dpm */
    UInt32 dwYResolution;		/* Y Resolution of image in dpm */
    UInt32 dwWidth;			/* Width in image's X resolution units
					 */
    UInt32 dwHeight;			/* Height in image's Y resolution units
					 */
    UInt32 dwBestStripHeight;		/* For best performance, read strips
					 * with height a multiple of this. */
    IO_COMPRESSION xCompression;	/* Compression type for src/dst */
    IO_COLORTYPE xColorType;		/* Color type (see IO_COLORTYPE) */
    UInt8 pPalette[256*3];		/* Color palette if IO_COLORTYPE is:
					 * IO_COLORTYPE_COLOR4 or
					 * IO_COLORTYPE_COLOR8.
					 * Palette is RGB triples (chunky):
					 * RGBRGB... */
/*    UInt32 dwMaskImages;		// Number of mask images */
    UInt32 dwHintsImages;		/* Number of hints images */
    IO_COLORSPACE xColorSpace;		/* Stored colorspace */
} TFX_IMGINFO;


/*
 * TIFF-FX Annotation information
 *
 * The following structure describes the data returned by the
 * TFX_GetAnnotInfo() function.
 */
typedef struct TFX_ANNOTINFO
{
    UInt16 wAnnotType;			/* Type of annotation */
    UInt32 dwXOffset;			/* Offset in X direction in page
					 * coordinates */
    UInt32 dwYOffset;			/* Offset in X direction in page
					 * coordinates */
    UInt32 dwWidth;			/* Width in page coordinates */
    UInt32 dwHeight;			/* Height in page coordinates */
    UInt32 dwFlags;			/* Annotation flags (see tfx.h) */
    UInt32 dwAnnotLength;		/* Length of annotation data, in bytes
					 */
} TFX_ANNOTINFO;


/*
 * PROTOTYPES
 */


#ifdef _MSC_VER
#pragma pack(pop, tfxinfo_h)
#endif

#ifdef __cplusplus
}
#endif

#endif /* TFXINFO_H_INCLUDED */

