/*
 * Trade secret of Xerox Corporation.
 * Copyright (c) 1991-2001, Xerox Corporation.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef TFXPAGE_H_INCLUDED
#define TFXPAGE_H_INCLUDED

#ifndef NOIPCORE
IP_RCSINFO(tfxpage_hRCSInfo, "$RCSfile: tfxpage.h,v $; $Revision: 1.1 $");
/* $Date: 2003/01/03 01:41:11 $ */
#endif

#ifdef __cplusplus
extern "C"
{
#endif

/* tfxpage.h
 *
 * DESCRIPTION
 *	 Declarations for TIFF-FX page level structures.
 *
 */

/*
 * INCLUDES
 */
#include "tfxdoc.h"
#include "tfxinfo.h"

#ifdef _MSC_VER
#pragma pack(push, tfxpage_h, 1)
#endif

/*
 * CONSTANTS
 */

/* Calling sequence status */
#define TFX_PAGE_SEQ_GETINFO		0x0001	/* Get image info called */
#define TFX_PAGE_SEQ_READSTART		0x0010	/* Read start called */
#define TFX_PAGE_SEQ_READFINISH		0x0000	/* Read finish called */
#define TFX_PAGE_SEQ_READSTATUS		0x00f0	/* Read status bits */
#define TFX_PAGE_SEQ_WRITESTART		0x0100	/* Write start called */
#define TFX_PAGE_SEQ_WRITEPRIMARY	0x0200	/* Primary image added */
#define TFX_PAGE_SEQ_WRITEPENDING	0x0400	/* Write finish called, waiting
						 * on tokenization */
#define TFX_PAGE_SEQ_WRITEFINISH	0x0000	/* Page is completely written */
#define TFX_PAGE_SEQ_WRITESTATUS	0x0f00	/* Write status bits */
#define TFX_PAGE_SEQ_TOKENSTART		0x1000	/* Reference page set */
#define TFX_PAGE_SEQ_MULTIDICTS		0x2000	/* Multiple token dictionaries */
#define TFX_PAGE_SEQ_TKCOLORADDED	0x4000	/* Color for tokenization added */
#define TFX_PAGE_SEQ_TOKENDICTSAVED	0x8000	/* Token dictionary saved */
#define TFX_PAGE_SEQ_MULTISTRIPSUBIFDS	0x10000	/* Use multi-strip SubIFDs */

/*
 * MACROS
 */
#define TFX_PAGEFILE(tfxPage)		TFX_DOCFILE(tfxPage->tfxDoc)
#define TFX_PAGEINST(tfxPage)		TFX_PAGEFILE(tfxPage)->instance
#define TFX_PAGEFILETOKEN(tfxImg)		TFX_FILETOKEN(TFX_PAGEFILE(tfxPage))

/*
 * TYPEDEFS
 */

/* Annotation list table header */
typedef struct TFX_ANNOTTABLEHEADER {
    UInt16 annotCount;		    /* Number of annotations in page (total) */
    UInt16 entriesPerTable;	    /* Number of annotation table entries per
				     * table */
    UInt16 bytesPerEntry;	    /* Number of bytes per table entries */
    UInt16 lastOrdinal;		    /* Ordinal (1, 2, ...) of last table */
    UInt32 firstTableOffset;	    /* File offset of first table */
} TFX_ANNOTTABLEHEADER;

/* Annotation table entry (describes an individual annotation) */
typedef struct TFX_ANNOTTABLEENTRY {
    UInt16 id;			    /* Unique ID of this annotation (offset) */
    UInt16 type;		    /* Annotation type (see tfx.h) */
    UInt32 xOffset;		    /* X offset in page coordinates */
    UInt32 yOffset;		    /* Y offset in page coordinates */
    UInt32 width;		    /* Width in page coordinates */
    UInt32 height;		    /* Height in page coordinates */
    UInt32 flags;		    /* Annotation flags (see ANNOT_xxx above)
				     */
    UInt32 bytes;		    /* Number of bytes in the data */
    Int32 offset;		    /* File offset of the annotation data */
    UInt8 *data;		    /* Annotation data */
} TFX_ANNOTTABLEENTRY;

/* Individual annotation table, linked if more than one */
typedef struct TFX_ANNOTTABLE {
    struct TFX_ANNOTTABLE *next;    /* Link to next table (cached) */
    UInt16 ordinal;		    /* Ordinal (1, 2, ...) of this table */
    UInt16 validAnnotations;	    /* Number of valid annotations in table */
    TFX_ANNOTTABLEENTRY *entries;   /* Allocated entry list */
    Int32 nextTableOffset;	    /* File offset of next annotation table */
} TFX_ANNOTTABLE;

/* Page handle
 */
typedef struct TFX_PAGE {
    UInt32 handleID;		    /* Handle ID, consistency check */
    UInt32 handleType;		    /* Handle type ID */
    struct TFX_PAGE *prev;	    /* Previous page in linked list */
    struct TFX_PAGE *next;	    /* Next page in linked list */
    TFX_DOC *tfxDoc;		    /* Document structure */
    UInt32 seqStatus;		    /* Calling sequence status */
    TFX_PAGEINFO pageInfo;	    /* Page information */
    UInt32 *extImageCounts;	    /* Image counts exposed to client (avoids
				     * true count getting corrupted by client)
				     */
    struct TFX_IMG *tfxImgs;		    /* Image list (linked) */
    Int32 annotTableHeaderOffset;   /* Offset to annotation table header
				     * (identified by the AnnotationOffsets
				     * field) */
    struct TFX_ANNOTTABLEHEADER annotTableHeader;/* Annotation list table header */
    struct TFX_ANNOTTABLE *firstAnnotTable;	/* Annotation list table for first */
    struct TFX_IMG *fgImg;	    /* Foreground image when attaching color
				     * to tokenization */
    UInt32 tokenPageID;		    /* Page ID assigned during tokenization */
    UInt32 tokenDictID;		    /* Symbol library dictionary ID of page
				     * specific symbol dictionary */

/* TIFF fields propogated to the page */
    UInt16 CodingMethods;	    /* Compression algorithms used */
    Int8 *DateTime;		    /* ASCII date/time: YYYY:MM:DD HH:MM::SS */
    UInt32 DateTimeBytes;	    /* Length of DateTime */
    UInt8 FaxProfile;		    /* ITU-compatible fax profile */
    UInt8 ModeNumber;		    /* Version of the T.44 standard */
    UInt16 PageNumber;		    /* Page number: 0,1,2,... */
    UInt16 PageCount;		    /* Page count: 1,2,... */
    UInt32 ProfileType;		    /* Type of data stored ??? */
    Int8 *TargetPrinter;	    /* ASCII description of the intended
				     * printer */
    UInt32 TargetPrinterLength;	    /* Length of TargetPrinter string */
} TFX_PAGE;



/*
 * PROTOTYPES
 */

#ifdef _MSC_VER
#pragma pack(pop, tfxpage_h)
#endif

#ifdef __cplusplus
}
#endif

#endif /* TFXPAGE_H_INCLUDED */

