/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxrjb2.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxrjb2.c:	TIFF-FX READ JBIG2 source
 *
 *	TIFF-FX read-only routines for JBIG2.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "shrcmprs.pub"
#include "iaerror.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif

#if !defined(NODGP) && !defined(NOIPCORE)
#define _IPWSTATE_DEFINED_
#include "decode_api.h"	/* For decode */
#include "jtokapi.h"	/* For encode */
#endif

#include "tfximage.h"
#include "tfxrtiff.h"
#include "tfxrjb2.h"
#include "tfxrpriv.h"
#include "tfxrtfx.h"

IP_RCSINFO(RCSInfo, "$RCSfile: tfxrjb2.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:31 $");


#ifndef NODGP
/* Local memory management for JBIG2 library,
 * Since they may call realloc, and I need to maintain both a generic interface 
 * to memory management, and the client version, I must have my own realloc function
 * without the aid of a client implementation of realloc.  To do this, I must keep the
 * size around with each parcel of memory, and use it when realloc-ing, to know the
 * size of the old buffer.
 * The local implementations of calloc and realloc simply call malloc or/and free.
 */
static TFX_INSTANCE *globalInstData = NULL;
static void * CDECL
wp_TfxTextMalloc(
	   size_t bytes
)
{
    TFX_INSTANCE *instData = globalInstData;
    void *buf = NULL;
    UInt32 *lmem;
    PROLOGUE;

    IO_CHECKMALLOC(lmem = TFX_MALLOC(bytes + sizeof(*lmem)));
    buf = (void *)&lmem[1];
    lmem[0] = bytes;

ERR_CLEANUP:

    return (void *)buf;
}

static void * CDECL
wp_TfxTextCalloc(
	   size_t nels,
	   size_t count
)
{
/*    TFX_INSTANCE *instData = globalInstData; */
    void *buf = NULL;
    UInt32 bytes = nels * count;
    PROLOGUE;

    IO_CHECKMALLOC(buf = wp_TfxTextMalloc(bytes));
    memset(buf, 0, bytes);

ERR_CLEANUP:

    return (void *)buf;
}


static void CDECL
wp_TfxTextFree(
	 void *buf
)
{
    TFX_INSTANCE *instData = globalInstData;
    UInt32 *lmem = ((UInt32 *)buf) - 1;

    TFX_FREE(lmem);

    return;
}


static void * CDECL
wp_TfxTextRealloc(
    void *ptr,
    size_t size
)
{
/*    TFX_INSTANCE *instData = globalInstData; */
    UInt32 oldSize;
    UInt32 *lmem = (UInt32 *)ptr - 1;
    void *buffer;
    PROLOGUE;

    /* We need to know the size of the old buffer, to know how to
     * reallocate it. In the case of the memory allocation I have,
     * I keep it stored with the buffer. The client may do anything
     * they desire, so I have no way of pinging for this information.
     */
    oldSize = lmem[0];		/* If seg fault here, buffer doesn't contain 4
				 * bytes -- buffer is probably invalid */
    IO_CHECKMALLOC(buffer = wp_TfxTextMalloc(size));

    if (oldSize > size) {
	memcpy(buffer, ptr, size);
    } else {
	memcpy(buffer, ptr, oldSize);
    }

ERR_CLEANUP:
    wp_TfxTextFree(ptr);

    return buffer;
}

/* Small allocators (don't do anything but call the big ones */
static ErrCode CDECL
wp_TfxTextSmallMalloc(
	void *cookie,
	Int32 id,
	void **ptr
)
{
    void *buf = NULL;
    UInt32 bytes = 0;

    /* Don't return anything (yet) */
    *ptr = NULL;

    switch (id) {
    case DGP_SMALLPOOL_MIN8:
	bytes = 8;
	break;
    case DGP_SMALLPOOL_MIN16:
	bytes = 16;
	break;
    case DGP_SMALLPOOL_MIN20:
	bytes = 20;
	break;
    case DGP_SMALLPOOL_MIN24:
	bytes = 24;
	break;
    case DGP_SMALLPOOL_MIN28:
	bytes = 28;
	break;
    case DGP_SMALLPOOL_MIN32:
	bytes = 32;
	break;
    case DGP_SMALLPOOL_MIN44:
	bytes = 44;
	break;
    case DGP_SMALLPOOL_MIN128:
	bytes = 128;
	break;
    default:
	return -1;
    }

    if ((buf = wp_TfxTextMalloc(bytes)) != NULL) {
	*ptr = buf;
	return 0;
    }
    (void)cookie;
    (void)id;

    return -1;
}

static void CDECL
wp_TfxTextSmallFree(
	void *cookie,
	Int32 id,
	void *buf
)
{

    wp_TfxTextFree(buf);
    (void)cookie;
    (void)id;
}



/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Converts the compressed JBIG2 data into a binary image
 *
 * LongDescription:::
 *  wp_TfxDetokenizeImage() converts a JBIG-2 position/ID block to
 *  positions and IDs, used by the IPShared routines, and "stamps" them
 *  back onto a blank image. The data is assumed to be found in the
 *  cmpBuffer/cmpBytes structure elements of the tfxImg pointer passed in.
 *  The pointer to the image (pImgData) must point to the entire binary image.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  pCmpData::			Pointer to compressed data
 *  dwCmpBytes::		Number of compressed bytes
 *  pImgData::			Pointer to image data
 *  dwBytesPerRow::		Bytes per image row of image data
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxDetokenizeImage(
    TFX_IMG *tfxImg,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    UInt8 *pImgData,
    UInt32 dwBytesPerRow
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_DOC *tfxDoc;
    TFX_IMGINFO *pTftfxImgInfo;
    void *textInfo;
    UInt8 *JBIG2Data;
    UInt32 JBIG2Bytes;
    UInt32 w, h;
    UInt16 dictNum, dictCount;
    UInt32 *dictIDs = NULL;
    Bool reload_dictionaries;
    DecodeHandle decHandle;
    TFX_DICTHANDLE *dl;
    PROLOGUE;

    TFX_CheckHandle(tfxImg);
    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    tfxDoc = tfxImg->tfxPage->tfxDoc;
    pTftfxImgInfo = &tfxImg->imgInfo;

    /* Get the header information */
    XERR_PASS_CLEANUP(TFX_GetDictListFromPosBlock((IO_IMGHANDLE)tfxImg,
	pCmpData, dwCmpBytes, &dictCount, &dictIDs, &JBIG2Data, &JBIG2Bytes));

    if ((tfxImg->dictCount = dictCount) == 0) {
	/* No token dictionary (either a blank image, or just singletons
	 * placed in the dictionary inside the position block) */
	if (dwCmpBytes == 0) {
	    NORMAL_RETURN;
	}
    }

    if (tfxImg->imgInfo.xColorType != IO_COLORTYPE_BINARY) {
	/* We could be nasty and return a real error. We'll be nice and
	 * consider that the image is probably an up-and-coming color
	 * tokenized foreground plane, and just return (the image has
	 * already been set to zero (black), so the image will look
	 * correct when rendered with this image.
	 * TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
	 * TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
	 * TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
	 * TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
	 * Fix this later. */
	if (dictIDs) {
	    TFX_FREE(dictIDs);
	}
	NORMAL_RETURN;
    }

    /* If the loaded dictionary list is different than the current
     * required list, and in the same order, reload all dictionaries */
    textInfo = tfxDoc->tokDecodeHandle;
    reload_dictionaries = 0;
    if (textInfo == NULL) {
	reload_dictionaries = 1;
    } else if (tfxDoc->curDictCount > 0 &&
	tfxDoc->curDictCount != dictCount) {
	reload_dictionaries = 1;
    } else {

	/* Check to see that the lists match (same ones in same order) */
	for (dictNum = 0; dictNum < dictCount; dictNum++) {
	    if (dictIDs[dictNum] != tfxDoc->curDictIDs[dictNum]) {
		reload_dictionaries = 1;
		break;
	    }
	}
    }

    /* Initiate a new tokenization process */
    if (textInfo == NULL) {
	Dgp_Allocators a;
	Dgp_SmallAllocators sm_a;

	/* Work around a client ID problem (for now) */
	globalInstData = TFX_IMGINST(tfxImg);

	/* Propogate the allocation routines */
	memset(&a, 0, sizeof(a));
	memset(&sm_a, 0, sizeof(sm_a));
	a.malloc = wp_TfxTextMalloc;
	a.calloc = wp_TfxTextCalloc;
	a.realloc = wp_TfxTextRealloc;
	a.free = wp_TfxTextFree;
	sm_a.mallocSmall = wp_TfxTextSmallMalloc;
	sm_a.freeSmall = wp_TfxTextSmallFree;

	XERR_PASS_CLEANUP(DgpDec_init(&decHandle, &a, &sm_a));
	tfxDoc->tokDecodeHandle = (UInt8 *)decHandle;
    } else {
	decHandle = (DecodeHandle)tfxDoc->tokDecodeHandle;
    }

    /* If appropriate, read the dictionary used for this image */
    if (reload_dictionaries) {

	/* Free the old "current" list */
	if (tfxDoc->curDictIDs) {
	    TFX_FREE(tfxDoc->curDictIDs);
	}
	tfxDoc->curDictCount = 0;

	/* Expel the old JBIG2 list */
	XERR_PASS_CLEANUP(DgpDec_destroyDictList(decHandle, &tfxDoc->dictList));

	/* Reset the list of decoded dictionaries to a state of non-use,
	 * then if they are reused in the new list, they'll be marked for
	 * retention.
	 */
	dl = tfxDoc->dictHandles;
	while (dl) {
	    dl->unused = 1;
	    if ((dl = dl->next) == tfxDoc->dictHandles) {
		break;
	    }
	}

	if (dictCount) {

	    /* Allocate a new list */
	    IO_CHECKMALLOC(tfxDoc->curDictIDs = TFX_MALLOC(dictCount * sizeof(UInt32)));

	    /* Load each of them */
	    for (dictNum = 0; dictNum < dictCount; dictNum++) {

		/* See if it's already decoded */
		{
		    UInt8 decoded = 0;
		    
		    dl = tfxDoc->dictHandles;
		    while (dl) {
			if (dl->id == dictIDs[dictNum]) {
			    decoded = 1;
			    break;
			}
			if ((dl = dl->next) == tfxDoc->dictHandles) {
			    break;
			}
		    }

		    if (! decoded) {
			
			/* Decode the dictionary */
			XERR_PASS_CLEANUP(wp_TfxReadTokenDictionary(tfxDoc, dictIDs[dictNum]));

			/* Should have gone to the end of the list */
			dl = tfxDoc->dictHandles->prev;
			XERR_GEN_CLEANUP(dl->id != dictIDs[dictNum], IO_INTERNAL_ERROR);
		    }

		    /* Mark it as still used */
		    dl->unused = 0;
		}
		
		/* Add it to the list */
		XERR_PASS_CLEANUP(DgpDec_buildDictList(decHandle, &tfxDoc->dictList, dl->handle));
		tfxDoc->curDictIDs[dictNum] = dictIDs[dictNum];
		tfxDoc->curDictCount++;
	    }
	} else {
	    /* No dictionaries */
	    XERR_PASS_CLEANUP(DgpDec_buildDictList(decHandle, &tfxDoc->dictList, 0));
	}


	/* Go through the list, and free up any unused dictionaries */
	dl = tfxDoc->dictHandles;
	while (dl) {
	    if (dl->unused) {

		/* Decode the dictionary */
		XERR_PASS_CLEANUP(DgpDec_destroyDict(decHandle, &dl->handle));

		/* Remove it from the list */
		if (dl == tfxDoc->dictHandles) {

		    /* It's the first one on the list */
		    if (dl->next == tfxDoc->dictHandles) {

			/* It's the only one, free it and stop */
			IO_FREE(tfxDoc->dictHandles);
			tfxDoc->dictHandles = NULL;
			break;
		    }
		
		    /* There's another one after, start the list there */
		    tfxDoc->dictHandles = dl->next;
		    tfxDoc->dictHandles->prev = dl->prev;
		    tfxDoc->dictHandles->next = dl->next;
		    IO_FREE(dl);
		    dl = tfxDoc->dictHandles;	/* Continue on with the new first in the list */
		    continue;	    /* We must continue here, to avoid the loop-around check below */
		} else {

		    /* It's not the first one on the list, remove it */
		    TFX_DICTHANDLE *nextdl = dl->next;
		    dl->prev->next = nextdl;
		    nextdl->prev = dl->prev;
		    IO_FREE(dl);
		    dl = nextdl;	/* Continue on with the next one */
		}
	    }
	    if (dl == tfxDoc->dictHandles) {
		break;
	    }
	}
    }

    /* Decode the position blocks, and render it to the image */
    w = pTftfxImgInfo->dwWidth + 64;
    h = pTftfxImgInfo->dwHeight + 64;
    XERR_PASS_CLEANUP(DgpDec_renderBinaryImage(decHandle, tfxDoc->dictList,
	JBIG2Data, JBIG2Bytes, pImgData, w, h, dwBytesPerRow / 4));

ERR_CLEANUP:

    if (dictIDs) {
	TFX_FREE(dictIDs);
    }
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a token dictionary
 *
 * LongDescription:::
 *  wp_TfxReadTokenDictionary() reads the JBIG2 dictionary format for the token
 *  dictionary, and loads it as the single internal dictionary. Client is
 *  responsible for adding this ID to the list of "current" IDs.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  dictID::			Dictionary identifier
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxReadTokenDictionary(
	TFX_DOC *tfxDoc,
	UInt32 dictID
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    void *textInfo;
    UInt32 dictBytes;
    UInt32 dictBlockBytes;
    UInt8 *dictBlock = NULL;
    UInt16 lastPage;
    UInt32 decMem;
    DecodeHandle decHandle = 0UL;
    UInt32 dictHandle;
    TFX_DICTHANDLE *newDictHandle = NULL;
    PROLOGUE;

    TFX_CheckHandle(tfxDoc);

    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);

    XERR_PASS_CLEANUP(TFX_GetSymbolDict((IO_DOCHANDLE)tfxDoc, dictID,
	&dictBlock, &dictBytes, &dictBlockBytes, &lastPage, &decMem));

    /* Load it */
    textInfo = tfxDoc->tokDecodeHandle;
    XERR_GEN_CLEANUP((textInfo == NULL), IO_INTERNAL_ERROR);

    /* For us, we no nothing about dictionary extensions */
    XERR_GEN_CLEANUP(dictBytes != dictBlockBytes, IO_NOSUPPORT);

    /* Load it */
    decHandle = (DecodeHandle)tfxDoc->tokDecodeHandle;
    XERR_GEN_CLEANUP(decHandle == 0, IO_INTERNAL_ERROR);

    /* Have the JBIG2 library decompress the dictionary block into
     * one or more decompressed segments */
    XERR_PASS_CLEANUP(DgpDec_decodeDict(decHandle, dictBlock, dictBytes, &dictHandle));

    /* Create a new item for the list of dict handles */
    IO_CHECKMALLOC(newDictHandle = TFX_MALLOC(sizeof(*newDictHandle)));
    memset(newDictHandle, 0, sizeof(*newDictHandle));
    newDictHandle->handle = dictHandle;
    newDictHandle->id = dictID;

    /* Add the dictionary to the linked-list of decoded dictionaries handles */
    if (tfxDoc->dictHandles == NULL) {
	/* Initialize the list with this one */
	tfxDoc->dictHandles = newDictHandle->prev = newDictHandle->next = newDictHandle;
    } else {
	/* Link it in at the end */
	newDictHandle->prev = tfxDoc->dictHandles->prev;
	newDictHandle->next = tfxDoc->dictHandles;
	tfxDoc->dictHandles->prev->next = newDictHandle;
	tfxDoc->dictHandles->prev = newDictHandle;
    }
    newDictHandle = NULL;   /* Memory location transferred, don't clobber it */
	
    /* Fall through to clean up temporary memory */

 ERR_CLEANUP:
    if (dictBlock) {
	TFX_FREE(dictBlock);
	dictBlock = NULL;
    }
    if (newDictHandle) {
	IO_FREE(newDictHandle);
    }
    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free memory associated with tokenization process(es)
 *
 * LongDescription:::
 *  wp_TfxFreeCloseTokenization() frees all memory associated with the token
 *  read and write processes.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxCloseTokenization(
	TFX_DOC *tfxDoc
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_DICTHANDLE *dl;
    PROLOGUE;

    TFX_CheckHandle(tfxDoc);
    instData = TFX_DOCINST(tfxDoc);

    /* Free the dictionary handles */
    while (tfxDoc->dictHandles) {
	dl = tfxDoc->dictHandles->next;
	dl->prev = tfxDoc->dictHandles->prev;
	tfxDoc->dictHandles->prev->next = dl;
	TFX_FREE(tfxDoc->dictHandles);
	if (dl == tfxDoc->dictHandles) {
	    break;
	}
	tfxDoc->dictHandles = dl;
    }

    if (tfxDoc->tokDecodeHandle) {

	/* Free the list of dictionaries */
	DgpDec_destroyDictList((UInt32)tfxDoc->tokDecodeHandle, &tfxDoc->dictList);

	/* Close the decoder */
	XERR_PASS_CLEANUP(DgpDec_close((UInt32 *)&tfxDoc->tokDecodeHandle));
    }

    if (tfxDoc->tokEncodeHandle) {

	/* Free data associated with the read handle */
	DgpApi_cleanSection(&tfxDoc->tokEncodeHandle);
    }

    /* Disable the dictionary list handle */
    tfxDoc->dictHandles = 0UL;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}
#endif /* NODGP */


/*
$Log: tfxrjb2.c,v $
Revision 1.2  2003/02/10 22:53:31  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:11  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.3  2001/12/26 21:23:45  dabercro
Mark as released

Revision 1.2  2001/11/07 21:44:36  dabercro
Refined NODGP code carve-out.

Revision 1.1  2001/10/31 00:11:58  dabercro
Initial addition of JBIG2 from PARC

*/
