/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxrjpeg.c
 *
 *	TIFF-FX read of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxrjpeg.c:	TIFF-FX READ JPEG source
 *
 *	TIFF-FX read-only JPEG routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

#ifndef IJL

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "jpeglib.h"
#include "iw_jpeg.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif

#else	/* IJL */

/* IJL */
#include "ijl.h"

#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "shrpixr.pub"
#include "pcconv.pub"	/* Conversion to/from DIB */
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif

#endif	/* IJL */

#include "tfximage.h"
#include "tfxrtiff.h"
#include "tfxrjpeg.h"
#include "tfxrpriv.h"

IP_RCSINFO(RCSInfo, "$RCSfile: tfxrjpeg.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:31 $");


#ifndef IJL	/* IJL */

typedef struct ReadData {
    UInt8 *buffer;
    UInt8 *next;
    UInt32 bufBytes;
    UInt32 bytesLeft;
} ReadData;

static ErrCode CDECL
IO_JPEG_ReadCallback(
    IPWState *pIPWState,
    void *userData,
    UInt8 *ptr,
    UInt32 bytes,
    UInt32 *pdwBytesRead
)
{
    ReadData *readData = (ReadData *)userData;
    pIPWState = pIPWState;
    
    if (readData->bytesLeft < bytes) {
	bytes = readData->bytesLeft;
    }
    memcpy(ptr, readData->next, bytes);
    readData->next += bytes;
    readData->bytesLeft -= bytes;
    
    *pdwBytesRead = bytes;

    return cSuccessful;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompresses a strip of JPEG data.
 *
 * LongDescription:::
 *  wp_TfxDecompressStripJPEG() decompresses a strip of JPEG data. Temporary
 *  PIXR found in tfxImg is used to store the image data decompressed.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  pCmpData::			Pointer to compressed data
 *  dwCmpBytes::		Number of compressed bytes
 *  pPixr::			Pointer to image PIXR
 *  dwStripHeight::		Number of rows to decompress
 *  dwRowOffset::		Destination row offset
 *
 * Output:::
 *  pdwRowsFilled::		Number of rows decompressed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxDecompressStripJPEG(
    TFX_IMG *tfxImg,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    PIXR *pPixr,
    UInt32 dwStripHeight,
    UInt32 dwRowOffset,
    UInt32 *pdwRowsFilled
)
{
    TFX_IMGINFO *pTfxImgInfo;
    IPWState *pIPWState;
    Int32 rowsToRead = (Int32)dwStripHeight;
    UInt32 totalRowsFilled = 0UL;
    DecompressInfo      cinfo_struct;
    Decompress_info_ptr cinfo=&cinfo_struct;
    UInt8 *image[3];
    Int32 w, h, bpl, ncomps;
    Int32 row, line;
    Int32 nlines;
    Int32 channel, channels;
    UInt32 memorg = 0UL;
    JSAMPIMAGE userBuf = NULL;
    ReadData readData;
    UInt32 options = 0UL;
    PROLOGUE;
    
    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));
    XERR_PASS_CLEANUP( w_pixrGetBpl(pIPWState, pPixr, &bpl));

    pTfxImgInfo = &tfxImg->imgInfo;
    
    /* Initialize */
    readData.buffer = pCmpData;
    readData.next = readData.buffer;
    readData.bufBytes = dwCmpBytes;
    readData.bytesLeft = readData.bufBytes;
    switch (IO_CHANNELS(tfxImg->imgInfo.xColorType)) {
    case 3:
	memorg = JMEMORG_BSQ_PIXR;
	break;
    case 1:
	switch (IO_BPPE(tfxImg->imgInfo.xColorType)) {
	case 4:
	    memorg = JMEMORG_GRAY4_PIXR;
	    break;
	case 8:
	    memorg = JMEMORG_GRAY8_PIXR;
	    break;
	}
	break;
    }

    switch (tfxImg->fields.PhotometricInterpretation) {
    case 8:
    case 9:	/* Works out that they appear the same */
	options = JD_CIELAB;
	break;
    case 10:	/* ITULAB */
	options = JD_ITULAB;
	break;
    }

    /* Initialize the JPEG process */
    memset(cinfo, 0, sizeof(*cinfo));
    XERR_PASS_CLEANUP(w_jdInit(pIPWState, cinfo, IO_JPEG_ReadCallback,
		  (void *)&readData, NULL, memorg, 1.0f,
		  &w, &h, &ncomps, options));

    XERR_PASS_CLEANUP(w_jdSetProgressiveDisplay(pIPWState, cinfo,
						0));

    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, &channels));
    /*//////////////////////////  THIS IS REALLY A WAY TO TRY TO READ SOMETHING SOMEBODY SHOULDN'T HAVE WRITTEN */
    /*//////////////////////////  THIS IS REALLY A WAY TO TRY TO READ SOMETHING SOMEBODY SHOULDN'T HAVE WRITTEN */
    /*//////////////////////////  THIS IS REALLY A WAY TO TRY TO READ SOMETHING SOMEBODY SHOULDN'T HAVE WRITTEN */
    /*//////////////////////////  THIS IS REALLY A WAY TO TRY TO READ SOMETHING SOMEBODY SHOULDN'T HAVE WRITTEN */
    /*//////////////////////////////// */
    /* What happens if we decode only the smaller of the two? */
/*    XERR_GEN_CLEANUP(w > (Int32)((tfxImg->imgInfo.dwWidth + 31) & ~31), */
/*	IO_BADPARAMETER); */
/*    XERR_GEN_CLEANUP(h < (Int32)dwStripHeight, IO_BADPARAMETER); */
    if (w < (Int32)tfxImg->imgInfo.dwWidth) {
	w = tfxImg->imgInfo.dwWidth;
    }
    if (h < (Int32)dwStripHeight) {
	h = dwStripHeight;
    }
    XERR_GEN_CLEANUP(ncomps != channels, IO_BADPARAMETER);

    XERR_GEN_CLEANUP(h < rowsToRead, IO_BADPARAMETER);
    
    /* Get the location of the image and the line pitch */
    {
	PIXR  *pSubPixr;
	for (channel = 0; channel < channels; channel++) {
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
				  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
				&(image[channel])));
	    image[channel] += dwRowOffset * bpl;
	}
    }
    
    XERR_PASS_CLEANUP(w_jdAllocMCURowBuffer(pIPWState, cinfo, 0, 
		                     	       &userBuf));
    XERR_GEN_CLEANUP(userBuf == NULL, IO_INTERNAL_ERROR);
    
    /* Decompress the image an MCU row at a time */
    XERR_PASS_CLEANUP(w_jdGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    line =0;
    for (row = 0; row < rowsToRead; ++row) {
	for (channel = 0; channel < channels; channel++) {
	    userBuf[channel][line] = image[channel];
	    image[channel] += bpl;
	}
	++line;
	if (line >= nlines || row == ((Int32)dwStripHeight-1)) {    
	    /* Decompress an MCU row into userBuf */
	    XERR_PASS_CLEANUP( w_jdMCURow(pIPWState, cinfo, userBuf));
	    line =0;
	}
    }
    
    XERR_PASS_CLEANUP(w_jdMCURowFree(pIPWState, cinfo, 0, userBuf));
    userBuf = NULL;

    /*//////////////////////////  THIS IS REALLY A WAY TO TRY TO READ SOMETHING SOMEBODY SHOULDN'T HAVE WRITTEN */
    /*//////////////////////////  THIS IS REALLY A WAY TO TRY TO READ SOMETHING SOMEBODY SHOULDN'T HAVE WRITTEN */
    /*//////////////////////////  THIS IS REALLY A WAY TO TRY TO READ SOMETHING SOMEBODY SHOULDN'T HAVE WRITTEN */
    /*//////////////////////////  THIS IS REALLY A WAY TO TRY TO READ SOMETHING SOMEBODY SHOULDN'T HAVE WRITTEN */
    /*//////////////////////////  THIS IS REALLY A WAY TO TRY TO READ SOMETHING SOMEBODY SHOULDN'T HAVE WRITTEN */
    /* Can we ignore the term, if we think we have some data? */
/*    XERR_PASS_CLEANUP( w_jdTerm(pIPWState, cinfo)); */
    XERR_IGNORE(__error__ = w_jdTerm(pIPWState, cinfo));

    totalRowsFilled = row * channels;
    *pdwRowsFilled = totalRowsFilled;
    
    if (__error__ == IO_NOERROR) {
	NORMAL_RETURN;
    }
    
 ERR_CLEANUP:
    if (userBuf != NULL)
	w_jdMCURowFree(pIPWState, cinfo, 0, userBuf);
    
    jpeg_destroy_decompress(pIPWState, &cinfo->j_dinfo);
    
    w_jdFreeErrorStr(pIPWState, cinfo);

    /* Reset the error (ignore it) */
    __error__ = IO_NOERROR;
    
    ERR_RETURN;
}

#else	/* IJL */

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image characteristics from a JPEG stream
 *
 * LongDescription:::
 *  wp_TfxGetJPEGInfo() reads enough of a JPEG (JFIF) stream to determine the
 *  image size.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  pCmpData::			Pointer to compressed data
 *  dwCmpBytes::		Number of compressed bytes
 *
 * Output:::
 *  pdwWidth::			The image width
 *  pdwHeight::			The image height
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxGetJPEGInfo(
    IPWState *pIPWState,
    IO_COLORTYPE xColorType,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    UInt32 *pdwWidth,
    UInt32 *pdwHeight
)
{
    DecompressInfo      cinfo_struct;
    Decompress_info_ptr cinfo=&cinfo_struct;
    Int32 w, h, ncomps;
    UInt32 memorg = 0UL;
    ReadData readData;
    UInt32 options = 0UL;
    PROLOGUE;
    
    /* Default to return nothing */
    *pdwWidth = 0UL;
    *pdwHeight = 0UL;


    /* Initialize */
    readData.buffer = pCmpData;
    readData.next = readData.buffer;
    readData.bufBytes = dwCmpBytes;
    readData.bytesLeft = readData.bufBytes;
    switch (IO_CHANNELS(xColorType)) {
    case 3:
	memorg = JMEMORG_BSQ_PIXR;
	break;
    case 1:
	switch (IO_BPPE(xColorType)) {
	case 4:
	    memorg = JMEMORG_GRAY4_PIXR;
	    break;
	case 8:
	    memorg = JMEMORG_GRAY8_PIXR;
	    break;
	}
	break;
    }

    memset(cinfo, 0, sizeof(*cinfo));
    XERR_PASS_CLEANUP(w_jdSetProgressiveDisplay(pIPWState, cinfo,
						0));
    XERR_PASS_CLEANUP(w_jdInit(pIPWState, cinfo, IO_JPEG_ReadCallback,
		  (void *)&readData, NULL, memorg, 1.0f,
		  &w, &h, &ncomps, options));
    XERR_IGNORE(__error__ = w_jdTerm(pIPWState, cinfo));

    /* Return what we've found */
    *pdwWidth = w;
    *pdwHeight = h;
    
    if (__error__ == IO_NOERROR) {
	NORMAL_RETURN;
    }
    
 ERR_CLEANUP:
    
    jpeg_destroy_decompress(pIPWState, &cinfo->j_dinfo);
    
    w_jdFreeErrorStr(pIPWState, cinfo);

    /* Reset the error (ignore it) */
    __error__ = IO_NOERROR;
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompresses a strip of JPEG data.
 *
 * LongDescription:::
 *  wp_TfxDecompressStripJPEG_IJL() decompresses a strip of JPEG data. Temporary
 *  PIXR found in tfxImg is used to store the image data decompressed.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  pCmpData::			Pointer to compressed data
 *  dwCmpBytes::		Number of compressed bytes
 *  pPixr::			Pointer to image PIXR
 *  dwStripHeight::		Number of rows to decompress
 *  dwRowOffset::		Destination row offset
 *
 * Output:::
 *  pdwRowsFilled::		Number of rows decompressed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxDecompressStripJPEG_IJL(
    TFX_IMG *tfxImg,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    PIXR *pPixr,
    UInt32 dwStripHeight,
    UInt32 dwRowOffset,
    UInt32 *pdwRowsFilled
)
{
    TFX_INSTANCE *instData = TFX_IMGINST(tfxImg);
    TFX_IMGINFO *pTfxImgInfo;
    IPWState *pIPWState;
    UInt32 totalRowsFilled = 0UL;
    Int32 channels;
    JPEG_CORE_PROPERTIES jcprops;
    IJLERR jerr;
    Bool bres = TRUE;
    UInt8 *dibImage = NULL;
    UInt32 byteCount = 0UL;
    UInt32 dibBPL;
    PROLOGUE;
    
    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));
    pTfxImgInfo = &tfxImg->imgInfo;
    channels = IO_CHANNELS(pTfxImgInfo->xColorType);
    
    /* Initialize */
    __try
    {
	/* Initialize the IJL */
	jerr = ijlInit(&jcprops);
	if (jerr < IJL_OK) {
	    bres = FALSE;
	    __leave;
	}
#if 0
	switch (tfxImg->fields.PhotometricInterpretation) {
	case 8:
	case 9:	/* Works out that they appear the same */
	    /* Fall through and do the same thing as ITULAB */
	case 10:	/* ITULAB */
	    /* Force using both quantization tables */
	    jcprops.jprops.maxquantindex = 2;	/* Luminance and chrominance */
	    jcprops.jprops.nqtables = 2;	/* Use both */

	    /* Use chroma tables for chroma components */
	    jcprops.jprops.jframe.comps[0].quant_sel = 0;
	    jcprops.jprops.jframe.comps[1].quant_sel = 1;
	    jcprops.jprops.jframe.comps[2].quant_sel = 1;

	    /* Force chroma subsampling */
	    jcprops.jprops.JPGSubsampling = IJL_411;

	    /* Use "other" colorspace */
	    jcprops.jprops.JPGColor = IJL_OTHER;
	    break;
	}
#endif

	/* Use buffer mode */
	jcprops.JPGFile = NULL;
	jcprops.JPGBytes = pCmpData;
	jcprops.JPGSizeBytes = dwCmpBytes;

	jerr = ijlRead(&jcprops, IJL_JBUFF_READPARAMS);
	if (jerr < IJL_OK) {
	    bres = FALSE;
	    __leave;
	}

	/* Fix the reported JPEG color type */
	switch (tfxImg->fields.PhotometricInterpretation) {
	case 8:
	case 9:	/* Works out that they appear the same */
	    /* Fall through and do the same thing as ITULAB */
	case 10:	/* ITULAB */
	    jcprops.JPGColor = IJL_OTHER;
	    break;
	}
/*jcprops.JPGWidth = pTfxImgInfo->dwWidth; */
/*jcprops.JPGHeight = dwStripHeight; */


	/* Allocate a DIB image buffer */
	byteCount = ((jcprops.JPGWidth * IO_BPPE(pTfxImgInfo->xColorType) + 31)
	    / 4) * jcprops.JPGHeight * jcprops.JPGChannels;
	IO_CHECKMALLOC(dibImage = IO_MALLOC(byteCount));

	/* Make sure it's consistent with what we think we have */
	XERR_GEN_CLEANUP(jcprops.JPGChannels != channels, IO_BADPARAMETER);

	/* Use it as the output */
	jcprops.DIBWidth = jcprops.JPGWidth;
	jcprops.DIBHeight = -jcprops.JPGHeight;
	jcprops.DIBChannels = jcprops.JPGChannels;
	jcprops.DIBColor = IJL_OTHER;
	dibBPL = (((jcprops.JPGWidth * IO_BPP(pTfxImgInfo->xColorType))
	    + 31) / 32) * 4;
	jcprops.DIBPadBytes = dibBPL -
	    ((jcprops.JPGWidth * IO_BPP(pTfxImgInfo->xColorType) + 7) / 8);
	jcprops.DIBSubsampling = 0;
	jcprops.DIBBytes = dibImage;

	/* Disable the color conversion */
	jcprops.cconversion_reqd = 0;

	jerr = ijlRead(&jcprops, IJL_JBUFF_READWHOLEIMAGE);
	if (jerr < IJL_OK) {
	    bres = FALSE;
	    __leave;
	}
    } /* __try */

    __finally {
	if (dibImage) {
	    totalRowsFilled = MIN(jcprops.DIBHeight, dwStripHeight);

	    /* Create a proper DIB from the DIB components provided by IJL */

	    /* Convert the DIB to PIXR */
	    XERR_PASS_CLEANUP(w_BMPToRGBPixrs(pIPWState, dibImage,
		0, totalRowsFilled, pPixr));

	    /* Free the temporary DIB image */
	    IO_FREE(dibImage), dibImage = NULL;

	    /* Swap the R and B channels (DIB created in BGR order) */
	    {
		void *r = pPixr->pixrData.childPlanes[0];

		pPixr->pixrData.childPlanes[0] = pPixr->pixrData.childPlanes[2];
		pPixr->pixrData.childPlanes[2] = r;
	    }
	}

	/* Clean up the IJL */
	ijlFree(&jcprops);
    }

    *pdwRowsFilled = totalRowsFilled * channels;
    
    if (__error__ == IO_NOERROR && bres != FALSE) {
	NORMAL_RETURN;
    }
    
 ERR_CLEANUP:
    if (dibImage) {
	IO_FREE(dibImage), dibImage = NULL;
    }

    /* Reset the error (ignore it) */
    __error__ = IO_NOERROR;
    
    ERR_RETURN;
}

#endif	/* IJL */

/*
$Log: tfxrjpeg.c,v $
Revision 1.2  2003/02/10 22:53:31  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:11  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.16  2001/12/26 21:13:59  dabercro
Fix DIB padding

Revision 1.15  2001/12/12 00:11:21  dabercro
Add support for Intel's JPEG Library

Revision 1.14  2001/11/07 21:45:28  dabercro
Sign change forgcc  compiler complaint

Revision 1.13  2001/07/23 20:50:48  dabercro
Bug fix: if the JFIF stream claims to be higher than the TIFF claims the image
is, and there's not a multiple of the MCU, then I was failing to decode the last
MCU.

Revision 1.12  2001/06/27 18:54:45  dabercro
Remove colorspace calls to JPEG, moved to image conversion in tfxconv.c.

Revision 1.11  2001/01/18 23:31:25  dabercro
Must clear the outer structure, since only the inner IPShared
structure gets cleared.

Revision 1.10  2001/01/18 00:21:31  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.9  2000/11/17 23:57:35  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.8  2000/10/16 18:22:53  dabercro
Major adds to support stripes, 0-byte strips, and ImageBaseColor.

Revision 1.7  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.6  2000/08/01 18:19:12  dabercro
Update read/write callbacks to reflect change in IPCore.

Revision 1.5  2000/06/07 00:00:17  dabercro
Continued work, lots of stuff.

Revision 1.4  2000/04/12 19:06:09  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.3  2000/03/11 02:06:34  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.2  2000/02/29 00:19:42  dabercro
Incremental improvements, profile distinctions,etc.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.13  1998/05/08 21:51:41  aber
aber on Fri May 8 14:50:35 PDT 1998

tfxrjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxrjpeg.c,v

To avoid purify warning, set the progressive display option explicitly;
although it is never being used, at least this avoids a complaint about
an UMR.

Revision 1.12  1998/01/14 00:16:29  ddavies
tfxrjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxrjpeg.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.11  1997/09/02 20:02:53  aber
tfxrjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxrjpeg.c,v

CommonCheck: Rename functions to private names (wp_).

Revision 1.10  1997/07/14 23:21:28  aber
aber on Mon Jul 14 16:20:42 PDT 1997

tfxrjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxrjpeg.c,v

Update error checking. Avoid duplicate free when error occurs.

Revision 1.9  1997/07/10 21:52:04  aber
aber on Thu Jul 10 14:51:26 PDT 1997

tfxrjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxrjpeg.c,v

Use GEN_CLEANUP instead of PASS_CLEANUP, to indicate the
problem instead of just returning an error.

Revision 1.8  1997/06/26 22:25:12  aber
tfxrjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxrjpeg.c,v

Allow reading JPEG images with multiple tiles.

Revision 1.7  1997/06/10 17:59:10  aber
aber on Tue Jun 10 10:58:58 PDT 1997

tfxrjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxrjpeg.c,v

Remove commented out section of code.

Revision 1.6  1997/06/10 01:26:39  aber
tfxrjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxrjpeg.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:00  aber
tfxrjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxrjpeg.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:04  aber
tfxrjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxrjpeg.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:11  aber
tfxrjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxrjpeg.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:35  aber
Initial revision

*/
