/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1997-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxronly.c
 *
 *	TIFF-FX read-only stubs (mainly for stubbing pieces of functions that
 *		are shared by read and write calls)
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxronly.c:	TIFF-FX READ-ONLY source
 *
 *	Handles read-only stubs.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#ifdef __GNUC__
#include "ansiprot.h"

IP_RCSINFO(RCSInfo, "$RCSfile: tfxronly.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:31 $");
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif
#endif	/* NOIPCORE */

#include "tfx.h"
#include "tfxfile.h"
#include "tfximage.h"
#include "tfxwtfx.h"


#if 0 /*defined(NOIPCORE) || defined(NODGP) */
/* Stub for TFX_WriteTokenImages()
 *****************************************************************************/
ErrCode CDECL
TFX_WriteTokenImages(
	IO_DOCHANDLE tfxDocHandle
)
{
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);

    NORMAL_RETURN;
}


/* Stub for TFX_WriteTokenImage()
 *****************************************************************************/
ErrCode CDECL
TFX_WriteTokenImage(
	IO_IMGHANDLE tfxImgHandle
)
{
    PROLOGUE;

    TFX_CheckHandle(tfxImgHandle);

    NORMAL_RETURN;
}


/* Stub for TFX_WriteTokenDictionary()
 *****************************************************************************/
ErrCode CDECL
TFX_WriteTokenDictionary(
	IO_DOCHANDLE tfxDocHandle
)
{
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);

    NORMAL_RETURN;
}
#endif	/* !NOIPCORE && !NODGP */


/* Stub for TFX_SetAuthor()
 *****************************************************************************/
ErrCode CDECL
TFX_SetAuthor(
	IO_DOCHANDLE tfxDocHandle,
	Int8 *pAuthorBlock,
	UInt16 wAuthorBytes
)
{
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);

    /* Avoid compiler complaints about unused variables */
    if (pAuthorBlock == NULL || wAuthorBytes == 0) {
      NORMAL_RETURN;
    }

    NORMAL_RETURN;
} 


/* Stub for TFX_SetReferencePage()
 *****************************************************************************/
ErrCode CDECL
TFX_SetReferencePage(
	IO_DOCHANDLE tfxDocHandle,
	Int32 dwReferencePageOffset
)
{
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);

    /* Avoid compiler complaints about unused variables */
    dwReferencePageOffset = 0;

    NORMAL_RETURN;
} 


/* Stub for TFX_SetAnnotTableEntryCount()
 *****************************************************************************/
ErrCode CDECL
TFX_SetAnnotTableEntryCount(
	IO_DOCHANDLE tfxDocHandle,
	UInt16 wEntriesPerTable
)
{
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);
    
    /* Avoid compiler complaints about unused variables */
    wEntriesPerTable = 0;

    NORMAL_RETURN;
}


/* Stub for TFX_SetPageTableEntryCount()
 *****************************************************************************/
ErrCode CDECL
TFX_SetPageTableEntryCount(
	IO_DOCHANDLE tfxDocHandle,
	UInt16 wEntriesPerTable
)
{
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);

    /* Avoid compiler complaints about unused variables */
    wEntriesPerTable = 0;
    
    NORMAL_RETURN;
}


/* Stub for wp_TfxSetPageIFD()
 *****************************************************************************/
ErrCode CDECL
wp_TfxSetPageIFD(
	      TFX_DOC *tfxDoc,
	      UInt32 pageOffset,
	      UInt8 flags,
	      UInt32 ifdOffset,
	      TFX_PAGE *tfxPage
)
{
    PROLOGUE;
    
    /* Check parameters / mode */
    TFX_CheckHandle(tfxDoc);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxDoc->seqStatus & TFX_DOC_SEQ_WRITESTART) == 0),
		     IO_SEQUENCE);
    XERR_GEN_CLEANUP(ifdOffset == 0, IO_BADPARAMETER);

    /* We can only add the next page, not allow a gap in the pages */
    XERR_GEN_CLEANUP(((Int32)pageOffset > tfxDoc->docInfo.wNPages - 1),
		     IO_BADPARAMETER);
    XERR_GEN_CLEANUP((tfxDoc->readWrite & WRITE_ALLOWED) == 0,
		     IO_INTERNAL_ERROR);

    /* Since this is a stub, it should never be called anyway. Check these
     * values, to keep the compiler from complaining */
    TFX_CheckHandle(tfxPage);
    XERR_GEN_CLEANUP(flags != 0, IO_BADPARAMETER);

    NORMAL_RETURN;
    
 ERR_CLEANUP:

    ERR_RETURN;
}


/*#ifndef NODGP */
/* Stub for Digipaper
 *****************************************************************************/
ErrCode CDECL
DgpApi_close(
	void *handle
)
{
    handle = handle;
    return 0;
}

ErrCode CDECL
DgpApi_cleanSection(
	void *handle
)
{    
    handle = handle;
    return 0;
}
/*#endif */


/* Stub for wp_TfxTestCodingMethods()
 *****************************************************************************/
ErrCode CDECL
wp_TfxTestCodingMethods(
	IO_DOCHANDLE tfxDocHandle
)
{
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);

    NORMAL_RETURN;
}


/* Stub for wp_TfxRewriteIFDs()
 *****************************************************************************/
ErrCode CDECL
wp_TfxRewriteIFDs(
	IO_DOCHANDLE tfxDocHandle
)
{
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);

    NORMAL_RETURN;
}


/*
$Log: tfxronly.c,v $
Revision 1.2  2003/02/10 22:53:31  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:11  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.7  2001/12/26 21:23:45  dabercro
Mark as released

Revision 1.6  2001/11/07 21:45:40  dabercro
Refined NODGP code carve-out.

Revision 1.5  2001/09/19 00:34:59  dabercro
Add stub for Digipaper function

Revision 1.4  2001/01/19 23:55:23  dabercro
Don't define stubs when function definitions are overridden (as
is the case when NOIPCORE).

Revision 1.3  2001/01/18 00:21:31  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.2  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.2  1998/02/12 20:09:16  aber
aber on Thu Feb 12 12:08:46 PST 1998

tfxronly.c -> /product/ipcore/ipwrap/src/RCS/tfxronly.c,v

Check all parameters, to keep Watcom happy.

Revision 1.1  1998/02/11 23:30:38  aber
Initial revision

*/
