/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxrpixr.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxrpixr.c:	TIFF-FX READ PIXR source
 *
 *	Handles read-only TIFF-FX PIXR
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
#include "ansiprot.h"
#endif

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "props.pub"
#include "rast.pub"
#include "frames.pub"
#include "ccitt.pub"
#include "mmr.pub"
#include "shrcmprs.prv"
#include "imageref.pub"
#include "shrpixr.pub"
#include "shrpixr.prv"
#include "shrcmprs.pub"
#include "color.pub"

#include "ipstate.pub"
#include "morphg.h"
#include "streldef.h"
#include "strel.pub"
#include "morph.pub"
#include "logicops.h"
#include "logicops.pub"

#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif

#include "tfx.pub"
#include "tfximage.h"
#include "tfxrtiff.h"
#include "tfxrtfx.h"
#include "tfxrpriv.h"
#include "tfxrjpeg.h"
#include "tfxrjb2.h"
#include "tfxconv.h"

IP_RCSINFO(RCSInfo, "$RCSfile: tfxrpixr.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:31 $");



#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initialize a TIFF-FX instance, used to reference files and memory.
 *
 * LongDescription:::
 *  w_TfxInitInstance()
 *
 * Input:::
 *  pIPWState::			IPW state
 *
 * Output:::
 *  pxInst::			TIFF-FX instance returned (passed to
 *				w_TfxEndInstance() when done)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_TfxInitInstance(
	IPWState *pIPWState,
	UInt32 dwClientInstID,
	IO_INSTHANDLE *pxInst
)
{
    TFX_INSTANCE *instData = NULL;
    MemoryTally *tally = NULL;
    PROLOGUE;

    /* Here's the only calls to the system malloc */
    IO_CHECKMALLOC(instData = (TFX_INSTANCE *)malloc(sizeof(*instData) +
				       sizeof(*tally)));
    tally = (MemoryTally *)((char *)instData + sizeof(*instData));

    memset(instData, 0, sizeof(*instData));
    memset(tally, 0, sizeof(*tally));

    /* Set defaults */
    instData->handleID = (UInt32)instData;	/* Consistency check */
    instData->handleType = TFX_HANDLE_TYPE;
    instData->tfxMagicNum = TFX_MAGIC_NUM;

    /* Preserve the default ID for switching back and forth between the
     * client ID and the reserved default */
    instData->defaultMallocID = (UInt32)tally;

    /* Save user identifier */
    instData->clientInstID = dwClientInstID;

    /* Initialize the malloc functions to the defaults */
    if (xMalloc && xFree) {	/* Must both be set to use clients */
	instData->malloc = xMalloc;
	instData->free = xFree;
	instData->mallocID = dwClientMallocID;
    } else if (!xMalloc || !xFree) {
	XERR_PASS_CLEANUP(IO_BADPARAMETER);
    } else {

	/* NULL functions indicate to use the defaults */
	instData->malloc = TfxMalloc;
	instData->free = TfxFree;
	instData->mallocID = instData->defaultMallocID;
    }

    /* Return what we did, obscured from the client */
    *pxInst = (IO_INSTHANDLE)instData;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (instData) {
	free(instData);
    }
    if (tally) {
	free(tally);
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  End a TIFF-FX instance.
 *
 * LongDescription:::
 *   w_TfxEndInstance() ends a previously established TIFF-FX instance.
 *
 * Input:::
 *  pIPWState::			IPW state
 *  xInst::			TIFF-FX instance
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_TfxEndInstance(
	IPWState *pIPWState,
	IO_INSTHANDLE xInst
)
{
    TFX_INSTANCE *instData = (TFX_INSTANCE *)xInst;
    PROLOGUE;

    TFX_CheckHandle(xInst);

    /* End IPCore instance */
    if (instData->pIPWState) {
	w_IPWStateTerm((IPWState **)&instData->pIPWState);
    }

    /* Check known TIFF-FX memory for memory leaks */
    if (TfxMemoryCheck((MemoryTally *)instData->defaultMallocID)) {
	/* Assume something wasn't called correctly */
	__error__ = IO_SEQUENCE;
	/* Contine on with dealloc */
    }

    /* Due to potential realloc() calls, some operating systems allow this
     * pointer to propogate. Invalidate before we free it. */
    memset((void *)instData->defaultMallocID, 0, sizeof(MemoryTally));
    memset(instData, 0, sizeof(*instData));

    if (instData->defaultMallocID) {
	free((void *)instData->defaultMallocID);
    }
    free(instData);

    if (__error__ != IO_NOERROR) {
	goto ERR_CLEANUP;    /* This is silly, but now we reference the label */
    }
    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Share an IPWrap state structure pointer with the TIFF-FX layer
 *
 * LongDescription:::
 *  w_TfxShareIPWState() propogates an IPWrap state structure to the TIFF-FX layer.
 *  This can be necessary when using a mixture of TFX_ functions with the w_
 *  functions. Noteably, when two underlying memory allocators are used, one
 *  as passed to w_IPWStateInit() during the creation of the pIPWState passed
 *  to the w_ functions, and a different memory allocator passed in to
 *  TFX_SetMallocFuncs().
 *
 * Input:::
 *  instHandle::		TFX_INSTANCE handle, returned from
 *				TFX_InitInstance()
 *  pIPWState::			Pointer to the IPWrap state to be shared
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_TfxShareIPWState(
	IO_INSTHANDLE instHandle,
	IPWState *pIPWState
)
{
    TFX_INSTANCE *instance = (TFX_INSTANCE *)instHandle;
    PROLOGUE;

    TFX_CheckHandle(instHandle);
    XERR_GEN_CLEANUP(pIPWState == NULL, IO_BADPARAMETER);

    instance->pIPWState = (UInt32)pIPWState;	/* A one-liner does all the work */

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Determine the base color, in sRGB
 *   
 *
 * LongDescription:::
 *  wp_TfxGetBaseColor() determines from ImageBaseColor and IFDBaseColor fields
 *  what the base color in sRGB is.  If neither of these is found, then the
 *  default color for the layer is returned.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *
 * Output:::
 *  pRGB::			The sRGB color found
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxGetBaseColor(
    TFX_IMG *tfxImg,
    UInt8 *pRGB
)
{
    PROLOGUE;

    TFX_CheckHandle(tfxImg);
    XERR_GEN_CLEANUP(pRGB == NULL, IO_BADPARAMETER);

    /* For now, pick what we already have (which may be zero's) */
    memcpy(pRGB, tfxImg->baseColor.rgb, 3);

    /* If we've already figured out what it should be, we're done; otherwise,
     * continue on... */
    if (tfxImg->baseColor.color_set == 0) {

	/* If we have the IFDBaseColor field, it takes precedence */
	if (tfxImg->fields.found.IFDBaseColor) {

	    /* Key off the IFD colorspace, then convert to sRGB */
	    switch (tfxImg->imgInfo.xColorType) {
	    case IO_COLORTYPE_BINARY:
		if ((tfxImg->fields.IFDBaseColor[0] & 1) == 0) {
		    tfxImg->baseColor.rgb[0] = 0;
		} else {
		    tfxImg->baseColor.rgb[0] = 1;
		}
		break;
	    case IO_COLORTYPE_COLOR4:
	    case IO_COLORTYPE_COLOR8:
		/* Paletted, get the index for the appropriate color */
		{
		    UInt8 index = (UInt8)tfxImg->fields.IFDBaseColor[0];

		    /* Make sure there's a palette */
		    XERR_GEN_CLEANUP(tfxImg->imgInfo.pPalette == NULL, IO_BADFILEFORMAT);

		    /* Make sure the index is within range */
		    if (IO_BPP(tfxImg->imgInfo.xColorType) == 4 && index >= 1 << 4) {
			index = (1 << 4) - 1;
		    }

		    /* Look up the palette value */
		    tfxImg->baseColor.rgb[0] = tfxImg->imgInfo.pPalette[index * 3];
		    tfxImg->baseColor.rgb[1] = tfxImg->imgInfo.pPalette[index * 3 + 1];
		    tfxImg->baseColor.rgb[2] = tfxImg->imgInfo.pPalette[index * 3 + 2];
		}
		break;
	    case IO_COLORTYPE_GRAY4:
		/* Grab the least significant 4 bits */
		tfxImg->baseColor.rgb[0] = tfxImg->baseColor.rgb[1] =
		    tfxImg->baseColor.rgb[2] = 
		    tfxImg->fields.IFDBaseColor[0] & 0xf;
		break;
	    case IO_COLORTYPE_GRAY8:
		/* Grab the least significant 8 bits */
		tfxImg->baseColor.rgb[0] = tfxImg->baseColor.rgb[1] =
		    tfxImg->baseColor.rgb[2] = 
		    tfxImg->fields.IFDBaseColor[0] & 0xff;
		break;
	    case IO_COLORTYPE_COLOR24:
		/* Grab the least significant 8 bits */
		tfxImg->baseColor.rgb[0] = tfxImg->fields.IFDBaseColor[0] & 0xff;
		tfxImg->baseColor.rgb[1] = tfxImg->fields.IFDBaseColor[1] & 0xff;
		tfxImg->baseColor.rgb[2] = tfxImg->fields.IFDBaseColor[2] & 0xff;
		break;
	    }

	    /* Convert the colorspace, if needed */
	    switch (tfxImg->fields.PhotometricInterpretation) {
	    case TFX_TIFFPHOTOMETRIC_RGB:
		/* Already where we want to be */
		tfxImg->baseColor.rgb[0] = (UInt8)tfxImg->fields.IFDBaseColor[0];
		tfxImg->baseColor.rgb[1] = (UInt8)tfxImg->fields.IFDBaseColor[1];
		tfxImg->baseColor.rgb[2] = (UInt8)tfxImg->fields.IFDBaseColor[2];
		break;
	    case TFX_TIFFPHOTOMETRIC_CIELAB:
		XERR_PASS_CLEANUP(IO_CIELAB2sRGB((UInt16)tfxImg->fields.IFDBaseColor[0],
		    (Int16)tfxImg->fields.IFDBaseColor[1], (Int16)tfxImg->fields.IFDBaseColor[2],
		    &tfxImg->baseColor.rgb[0], &tfxImg->baseColor.rgb[1],
		    &tfxImg->baseColor.rgb[2]));
		break;
	    case TFX_TIFFPHOTOMETRIC_ICCLAB:
		XERR_PASS_CLEANUP(IO_ICCLAB2sRGB((UInt16)tfxImg->fields.IFDBaseColor[0],
		    (UInt16)tfxImg->fields.IFDBaseColor[1],
		    (UInt16)tfxImg->fields.IFDBaseColor[2],
		    &tfxImg->baseColor.rgb[0], &tfxImg->baseColor.rgb[1],
		    &tfxImg->baseColor.rgb[2]));
		break;
	    case TFX_TIFFPHOTOMETRIC_ITULAB:
		XERR_PASS_CLEANUP(IO_ITULAB2sRGB((UInt16)tfxImg->fields.IFDBaseColor[0],
		    (UInt16)tfxImg->fields.IFDBaseColor[1],
		    (UInt16)tfxImg->fields.IFDBaseColor[2],
		    &tfxImg->baseColor.rgb[0], &tfxImg->baseColor.rgb[1],
		    &tfxImg->baseColor.rgb[2], tfxImg->fields.Decode));
		break;
	    case TFX_TIFFPHOTOMETRIC_CMYK:
	    default:
		XERR_PASS_CLEANUP(IO_NOSUPPORT);
		break;
	    }
	} else if (tfxImg->fields.found.ImageBaseColor) {
	    /* If no IFDBaseColor, then the ImageBaseColor prevails */
	    if (IO_BPP(tfxImg->imgInfo.xColorType) == 1) {
		if (tfxImg->fields.ImageBaseColor[0] == 0) {
		    tfxImg->baseColor.rgb[0] = 0;
		} else {
		    tfxImg->baseColor.rgb[0] = 1;
		}
	    } else {
		XERR_PASS_CLEANUP(IO_ITULAB2sRGB(tfxImg->fields.ImageBaseColor[0],
			tfxImg->fields.ImageBaseColor[1], tfxImg->fields.ImageBaseColor[2],
			&tfxImg->baseColor.rgb[0], &tfxImg->baseColor.rgb[1],
			&tfxImg->baseColor.rgb[2], NULL));
	    }
	} else {

	    /* Note, this is an undefined situation.  Yet the intent of
	     * a 0-byte strip is to either:
	     *	1. have an all-zero strip, when uncompressed, which would
	     *	   be interpreted by the PhotometricInterpretation (an other
	     *     fields) to be a specific color
	     *  or
	     *  2. have a strip with default color, as determined by the
	     *     image layer, as if the strip weren't there
	     * Below is how to do both, with one #ifdef'd out:
	     */
#if 0	/* Use #if 0 if doing #1 above, and #if 1 if doing #2 above */

	    /* Use the color described by an all-zero decoded image */
	    switch (tfxImg->fields.PhotometricInterpretation) {
	    case TFX_TIFFPHOTOMETRIC_WHITEISZERO:
		if (tfxImg->imgInfo.dwImgLayer & 1) {
		    tfxImg->baseColor.rgb[0] = tfxImg->baseColor.rgb[1] =
			    tfxImg->baseColor.rgb[2] = 255;	/* White */
		} else {
		    tfxImg->baseColor.rgb[0] = 0;		/* White */
		}
		break;
	    case TFX_TIFFPHOTOMETRIC_BLACKISZERO:
		if (tfxImg->imgInfo.dwImgLayer & 1) {
		    tfxImg->baseColor.rgb[0] = tfxImg->baseColor.rgb[1] =
			    tfxImg->baseColor.rgb[2] = 0;	/* Black */
		} else {
		    tfxImg->baseColor.rgb[0] = 1;		/* Black */
		}
		break;
	    case TFX_TIFFPHOTOMETRIC_RGB:
		tfxImg->baseColor.rgb[0] = tfxImg->baseColor.rgb[1] =
		    tfxImg->baseColor.rgb[2] = 0;	/* Black */
		break;
	    case TFX_TIFFPHOTOMETRIC_PALETTE:
		/* Paletted, get the color at index 0 */
		{
		    UInt8 index = 0;

		    /* Make sure there's a palette */
		    XERR_GEN_CLEANUP(tfxImg->imgInfo.pPalette == NULL, IO_BADFILEFORMAT);

		    /* Look up the palette value */
		    tfxImg->baseColor.rgb[0] = tfxImg->imgInfo.pPalette[index * 3];
		    tfxImg->baseColor.rgb[1] = tfxImg->imgInfo.pPalette[index * 3 + 1];
		    tfxImg->baseColor.rgb[2] = tfxImg->imgInfo.pPalette[index * 3 + 2];
		}
		break;
	    case TFX_TIFFPHOTOMETRIC_CIELAB:
		XERR_PASS_CLEANUP(IO_CIELAB2sRGB(0, 0, 0,
		    &tfxImg->baseColor.rgb[0], &tfxImg->baseColor.rgb[1],
		    &tfxImg->baseColor.rgb[2]));
		break;
	    case TFX_TIFFPHOTOMETRIC_ICCLAB:
		XERR_PASS_CLEANUP(IO_ICCLAB2sRGB(0, 0, 0,
		    &tfxImg->baseColor.rgb[0], &tfxImg->baseColor.rgb[1],
		    &tfxImg->baseColor.rgb[2]));
		break;
	    case TFX_TIFFPHOTOMETRIC_ITULAB:
		XERR_PASS_CLEANUP(IO_ITULAB2sRGB(0, 0, 0,
		    &tfxImg->baseColor.rgb[0], &tfxImg->baseColor.rgb[1],
		    &tfxImg->baseColor.rgb[2], tfxImg->fields.Decode));
		break;
	    case TFX_TIFFPHOTOMETRIC_CMYK:
	    default:
		XERR_PASS_CLEANUP(IO_NOSUPPORT);
	    }

#else	/* #1 or #2 */
	    /* Use the default layer color */
	    if (tfxImg->imgInfo.dwImgLayer & 1) {
		if (tfxImg->imgInfo.dwImgLayer == 1) {
		    tfxImg->baseColor.rgb[0] = tfxImg->baseColor.rgb[1] =
			    tfxImg->baseColor.rgb[2] = 255;
		} else {
		    tfxImg->baseColor.rgb[0] = tfxImg->baseColor.rgb[1] =
			tfxImg->baseColor.rgb[2] = 0;
		}
	    } else {
		tfxImg->baseColor.rgb[0] = 0;
	    }
#endif	/* #1 or #2 */
	}

	/* Mark as determined */
	tfxImg->baseColor.color_set = 1;

	/* Return the sRGB color determined */
	memcpy(pRGB, tfxImg->baseColor.rgb, 3);
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy image data stored in PIXR buffer to output image buffer.
 *
 * LongDescription:::
 *  wp_CopyFromTempPixr() copies image data temporarily stored in the intermediate
 *  PIXR, to the output buffer. Image position and rows left maintained in an
 *  outer loop are updated.
 *  <BR>
 *  If no intermediate buffer is used, there's no reason to copy.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  dstPixr::			Destination PIXR
 *  pdwDstPixrRowOffset::	Pointer to the destination PIXR row offset
 *  pdwImgRowOffset::		Pointer to the image proper row offset
 *  pdwTotalRowsFilled::	Pointer to the number of rows filled in outer
 *				loop
 *  pdwRowsToRead::		Pointer to the number of rows to fill in
 *				outer loop
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_CopyFromTempPixr(
    TFX_IMG *tfxImg,
    PIXR *dstPixr,
    UInt32 *pdwDstPixrRowOffset,
    UInt32 *pdwImgRowOffset,
    UInt32 *pdwTotalRowsFilled,
    UInt32 *pdwRowsToRead
)
{
    UInt8 *srcPtr, *dstPtr;
    Int32 byteCount;
    IPWState *pIPWState;
    Int32 channel, channels;
    PIXR *srcPixr, *pSubPixr;
    Int32 srcBPL, dstBPL = 0L;
    UInt32 rowsFilled = 0UL;
    UInt32 rowsAvailable;
    PROLOGUE;

    if (tfxImg->tmpPixr == NULL || *pdwImgRowOffset < tfxImg->tmpPixrRow ||
	*pdwImgRowOffset >= tfxImg->tmpPixrRow + tfxImg->tmpPixrRows) {
	NORMAL_RETURN;
    }

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    srcPixr = tfxImg->tmpPixr;
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, dstPixr, &channels));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, srcPixr, &srcBPL));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, dstPixr, &dstBPL));
    byteCount = (tfxImg->minBytesPerRow + 3) & ~3;
    if (byteCount > dstBPL) {
	byteCount = dstBPL;		/* This might be a problem */
    }
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, srcPixr, channel,
				  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &srcPtr));
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, dstPixr, channel,
				  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &dstPtr));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, srcPixr, &srcPtr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, dstPixr, &dstPtr));
	}
	srcPtr += (*pdwImgRowOffset - tfxImg->tmpPixrRow) * srcBPL;
	dstPtr += *pdwDstPixrRowOffset * dstBPL;
	rowsAvailable = tfxImg->tmpPixrRows - (*pdwImgRowOffset - tfxImg->tmpPixrRow);
	if (rowsAvailable > *pdwRowsToRead) {
	    rowsAvailable = *pdwRowsToRead;
	}
	while (rowsAvailable--) {
	    memcpy(dstPtr, srcPtr, byteCount);
	    srcPtr += srcBPL;
	    dstPtr += dstBPL;
	    rowsFilled++;
	}
    } /* for each channel */
    rowsFilled /= channels;
    (*pdwTotalRowsFilled) += rowsFilled;
    *pdwImgRowOffset += rowsFilled;
    *pdwDstPixrRowOffset += rowsFilled;
    *pdwRowsToRead -= rowsFilled;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read an arbitrary stip image data, and decompress if necessary
 *
 * LongDescription:::
 *  w_TfxReadStrip() reads the next strip of image data from an image that was
 *  requested by TFX_ImageReadStart. Image data is read from the top of the
 *  image to the bottom. If the client does not choose to read the entire
 *  image in as a single strip, it is advisable to use a strip height that is
 *  a multiple of the "dwBestStripHeight" (from the TFX_IMGINFO structure)
 *  for best performance.
 *  <BR>
 *  If the BottomToTop flag was selected, this pointer
 *  refer to the beginning of the last row in each buffer.
 *  <BR>
 *  In the case of Uncompressed data returned, the number of bytes filled
 *  should be equal to the bytes/row times either the strip height, or the
 *  number of rows successfully read.
 *  In the case of compressed data returned, the number of bytes filled will
 *  be the number of compressed bytes filled in. For compressed data, the
 *  dwBytesPerRow specifies the size of the buffer. If the buffer is not big
 *  enough to accomodate the compressed byte stream needed for the strip
 *  size, IO_BUFFEREXCEEDED is returned.
 *
 * Input:::
 *  pIPWState::			IPW state
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_AddImage()
 *  dwImgRowOffset::		Row offset of image proper
 *  dwStripHeight::		Number of lines to be retrieved
 *  pPixr::			Pointer to client's image PIXR to retrieve into
 *  dwPixrRowOffset::		PIXR row offset to place the strip
 *
 * Output:::
 *  pdwRowsFilled::		Pointer to the number of rows filled in
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_TfxReadStrip(
	IPWState *pIPWState,
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	PIXR *pPixr,
	UInt32 dwPixrRowOffset,
	UInt32 *pdwRowsFilled
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;	/* Image structure */
    IPWState *internalIPWState = NULL;	/* IPW state used by TIFF-FX API */
    TFX_IMGINFO *pTftfxImgInfo;		/* Image info structure */
    UInt32 clientRowsLeft;		/* Number of client rows to fill */
    Int32 bpl = 0L;			/* Bytes per line */
    UInt32 rowsFilled;			/* Rows filled in lower function */
    UInt32 totalRowsFilled = 0UL;	/* Total rows filled */
    ColorType ctype;			/* Color type of intermediate PIXR */
    UInt32 tiffRowsToRead;		/* Rows to read in TIFF strip */
    Int32 framedWidth, framedBytes;	/* Minimum framed PIXR size */
    Int32 pixrHeight;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    TFX_CheckHandle(tfxImgHandle);

    pTftfxImgInfo = &tfxImg->imgInfo;

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxImg->seqStatus & TFX_IMG_SEQ_READSTATUS) !=
		     TFX_IMG_SEQ_READSTART), IO_SEQUENCE);

    /* If it's the first strip, populate the PIXR colormap */
    if (dwImgRowOffset == 0 &&
	(pTftfxImgInfo->xColorType == IO_COLORTYPE_COLOR4 ||
	 pTftfxImgInfo->xColorType == IO_COLORTYPE_COLOR8)) {
	ColorMap *cMap = NULL;

	XERR_PASS_CLEANUP(w_pixrGetColorMap(pIPWState, pPixr, &cMap));
	if (cMap == NULL) {

	    ColorRecord *cRec;
	    UInt8 *color, *srcColor;
	    UInt32 index, indices = 0UL;

	    indices = 1 << IO_BPPE(pTftfxImgInfo->xColorType);
	    XERR_PASS_CLEANUP(w_createColorMap(pIPWState, indices, &cMap));
	    XERR_PASS_CLEANUP(w_pixrSetColorMap(pIPWState, pPixr, cMap));
	    XERR_GEN_CLEANUP(cMap == NULL, IO_INTERNAL_ERROR);
	    cRec = cMap->map;
	    srcColor = pTftfxImgInfo->pPalette;
	    for (index = 0; index < indices; index++) {
		color = (UInt8 *)cRec++;
		color[cRedPlane] = *srcColor++;
		color[cGreenPlane] = *srcColor++;
		color[cBluePlane] = *srcColor++;
	    }
	}
    }

    /* Save the internally used IPWState, and use the passed in one.
     * Restore when finished. */
    if (tfxImg->tfxPage->tfxDoc->tfxFile->instance->pIPWState != 0) {
        internalIPWState = (IPWState *)tfxImg->tfxPage->tfxDoc->tfxFile->instance->pIPWState;
    }
    tfxImg->tfxPage->tfxDoc->tfxFile->instance->pIPWState = (UInt32)pIPWState;

    /* Copy over any left over rows from previous calls */
    clientRowsLeft = dwStripHeight;
    XERR_PASS_CLEANUP(wp_CopyFromTempPixr(tfxImg, pPixr, &dwPixrRowOffset,
					  &dwImgRowOffset, &totalRowsFilled,
					  &clientRowsLeft));

    /* Keep track of user PIXR status */
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &bpl));

    /* Get the size of the PIXR passed in */
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &pixrHeight));
    
    /* Check the size of the PIXR if framed properly */
    XERR_PASS_CLEANUP(w_pixrFramedWidth(pIPWState, pPixr, &framedWidth));
    framedBytes = (((framedWidth * IO_BPPE(pTftfxImgInfo->xColorType)) + 31) / 8) & ~3;

    /* Read strips from the file up to the last whole file strip that won't
     * overfill the client strip */
    while (clientRowsLeft > 0) {

	/* Compute the number of physical rows to read */
	if (pTftfxImgInfo->dwBestStripHeight <= 0) {
	    UInt32 stripOffset = 0;
	    UInt32 stripRow = 0;

	    /* Look for the next variable-sized strip */
	    while (stripRow < dwImgRowOffset &&
		stripOffset < tfxImg->fields.StripsPerImage) {
		stripRow += tfxImg->fields.StripRowCounts[stripOffset];
		stripOffset++;
	    }
	    XERR_GEN_CLEANUP(stripRow != dwImgRowOffset, IO_SEQUENCE);
	    XERR_GEN_CLEANUP(stripOffset >= tfxImg->fields.StripsPerImage,
		IO_BADFILEFORMAT);
	    tiffRowsToRead = tfxImg->fields.StripRowCounts[stripOffset];
	} else {
	    tiffRowsToRead = pTftfxImgInfo->dwBestStripHeight;
	}
	if (dwImgRowOffset + tiffRowsToRead > pTftfxImgInfo->dwHeight) {
	    tiffRowsToRead = pTftfxImgInfo->dwHeight - dwImgRowOffset;
	}

	if (tiffRowsToRead == 0UL) {
	    break;
	}

	/* Determine if the data needs to be buffered:
	 * 1. If the client has requested something in the middle of a
	 *    TIFF strip.
	 * 2. The strip given to us by the client doesn't fill a full
	 *    TIFF strip.
	 */
	if ((pTftfxImgInfo->dwBestStripHeight &&
	    (dwImgRowOffset % pTftfxImgInfo->dwBestStripHeight) != 0) ||
	    (clientRowsLeft < tiffRowsToRead) ||
	    (framedBytes < bpl)) {

	    /* Use intermediate buffer */
	    if (tfxImg->tmpPixr == NULL) {
		ScaledInt1616 xres, yres;

		/* Prepare a temporary PIXR */
		xres = Int32_TO_ScaledInt1616(pTftfxImgInfo->dwXResolution);
		yres = Int32_TO_ScaledInt1616(pTftfxImgInfo->dwYResolution);
		switch ((int)pTftfxImgInfo->xColorType) {
		case IO_COLORTYPE_COLOR24:
		    ctype = cColorRGB;
		    break;
		case IO_COLORTYPE_BINARY:
		case IO_COLORTYPE_GRAY4:
		case IO_COLORTYPE_COLOR4:
		case IO_COLORTYPE_GRAY8:
		case IO_COLORTYPE_COLOR8:
		    ctype = cColorGray;
		    break;
		default:
		    ctype = cColorNone;
		}
		tfxImg->tmpPixrRow = tfxImg->tmpPixrRows = 0UL;
		XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState,
		    pTftfxImgInfo->dwWidth,
		    tiffRowsToRead, IO_BPPE(pTftfxImgInfo->xColorType),
		    xres, yres, 32, IO_CHANNELS(pTftfxImgInfo->xColorType),
		    ctype, NULL, &tfxImg->tmpPixr));
	    }

	    /* Fill intermediate PIXR */
	    XERR_PASS_CLEANUP(wp_TfxReadStrip(tfxImg, tfxImg->tmpPixr, 0UL,
		dwImgRowOffset, &tiffRowsToRead, &tfxImg->tmpPixrRows));
	    tfxImg->tmpPixrRow = dwImgRowOffset;

	} else {

	    /* Use client PIXR directly */
	    rowsFilled = 0;
	    XERR_PASS_CLEANUP(wp_TfxReadStrip(tfxImg, pPixr, dwPixrRowOffset,
		dwImgRowOffset, &tiffRowsToRead, &rowsFilled));
	    clientRowsLeft -= rowsFilled;
	    dwPixrRowOffset += rowsFilled;
	    dwImgRowOffset += rowsFilled;
	    totalRowsFilled += rowsFilled;
	}

	/* Copy any lines needed to fill the client strip */
	XERR_PASS_CLEANUP(wp_CopyFromTempPixr(tfxImg, pPixr, &dwPixrRowOffset,
	    &dwImgRowOffset, &totalRowsFilled, &clientRowsLeft));
    }

    /* Copy any lines needed to fill the client strip */
    XERR_PASS_CLEANUP(wp_CopyFromTempPixr(tfxImg, pPixr, &dwPixrRowOffset,
	&dwImgRowOffset, &totalRowsFilled, &clientRowsLeft));

    *pdwRowsFilled = totalRowsFilled;

    if (internalIPWState != NULL) {
	tfxImg->tfxPage->tfxDoc->tfxFile->instance->pIPWState = (UInt32)internalIPWState;
    }
    NORMAL_RETURN;

 ERR_CLEANUP:

    if (internalIPWState != NULL) {
	tfxImg->tfxPage->tfxDoc->tfxFile->instance->pIPWState = (UInt32)internalIPWState;
    }
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Clears a rectangular image region in a PIXR
 *
 * LongDescription:::
 *  wp_TfxClearPixrRegion() clears (0's) a rectangular image region in a PIXR
 *  <BR>
 *  Note: does not actually clear an arbitrary region (yet). Needs modification
 *  to actually clear in PIXR format, just clears (long)s.
 *
 * Input:::
 *  pIPWState::			IPW state
 *  pPixr::			PIXR to clear
 *  dstCol::			Destination column, must be 0 (for now)
 *  dstRow::			Destination row, must be 0 (for now)
 *  dstWidth::			Region width
 *  dstHeight::			Region height
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  Even though this was parameterized for a region, assumes the entire
 *  image is to be cleared.
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxClearPixrRegion(
    IPWState *pIPWState,
    PIXR *pPixr,
    Int32 dstCol,
    Int32 dstRow,
    Int32 dstWidth,
    Int32 dstHeight
)
{
    UInt32 *imgPtr, *rowPtr;
    Int32 channels;
    Int32 row, rows, lastRow;
    Int32 col, cols, lastCol;
    Int32 bpl;
    PROLOGUE;

    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, &channels));
    if (channels > 1) {
	PIXR *pSubPixr;
	Int32 channel;

	for (channel = 0; channel < channels; channel++) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
		&pSubPixr));
	    XERR_PASS_CLEANUP(wp_TfxClearPixrRegion(pIPWState, pSubPixr, dstCol,
		dstRow, dstWidth, dstHeight));
	}
	NORMAL_RETURN;
    }
    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr, (UInt8 **)&imgPtr));
    XERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, pPixr, &cols));
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &rows));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &bpl));

    /* Cleared region will include fractional longs on either side of
     * rectangle */
    dstWidth += (dstCol & 3) + 3;
    dstCol &= ~3;		/* Must be divisible by 4 */
    imgPtr += (dstRow * bpl + dstCol) / 4;
    if (dstCol + dstWidth > cols) {
	dstWidth = cols - dstCol;
    }
    if (dstRow + dstHeight > rows) {
	dstHeight = rows - dstRow;
    }
    lastCol = dstCol + dstWidth;
    lastRow = dstRow + dstHeight;
    for (row = dstRow; row < lastRow; row++) {
	rowPtr = imgPtr;
	for (col = 0; col < lastCol; col += 4) {
	    *rowPtr++ = 0UL;
	}
	imgPtr += bpl / 4;
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress image strips as given from the TIFF-FX read routines.
 *
 * LongDescription:::
 *  w_TfxDecompress() decompresses image strips as given from the TIFF-FX read
 *  routines. The image handle is used for extracting memory functions.
 *
 * Source buffer is decompressed into the destination image.
 *
 * Input:::
 *  pIPWState::			IPW state
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_AddImage()
 *  xCompression::		See IO_COMPRESSION types
 *  xColorType::		Type of color
 *  dwFlags::			IO_IMGFLAGS_xxx bit combo
 *  pSrc::			Source buffer
 *  dwBytesInSrc::		Byte count of source buffer
 *  dwRowOffset::		Destination row offset
 *  dwWidth::			Columns of COLORTYPE in dst image
 *  dwStripHeight::		Rows to decompress
 *
 * Output:::
 *  pPixr::			Destination PIXR
 *  pdwRowsFilled::		Number of rows successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_TfxDecompress(
	IPWState *pIPWState,
	IO_IMGHANDLE tfxImgHandle,
	IO_COMPRESSION xCompression,
	IO_COLORTYPE xColorType,
	UInt32 dwFlags,
	UInt8 *pSrc,
	UInt32 dwBytesInSrc,
	UInt32 dwRowOffset,
	UInt32 dwWidth,
	UInt32 dwStripHeight,
	PIXR *pPixr,
	UInt32 *pdwRowsFilled
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_IMGINFO *pTftfxImgInfo;
    Int32 channel, channels;
    UInt8 *imageChannels[3];
    PIXR *pSubPixr;
    UInt8 *pDst;
    UInt8 *iPtr, *cPtr;
    UInt32 rowsLeft, rowsFilled;
    UInt32 bytesLeft;
    Int32 srcBPL, dstBPL, bytesPerRow;
    Int32 dstHeight;
    UInt32 bytesUsed;
    UInt32 conversion = 0;
    PROLOGUE;

    *pdwRowsFilled = 0UL;
    TFX_CheckHandle(tfxImgHandle);
    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    pTftfxImgInfo = &tfxImg->imgInfo;

    /* Check buffers and sizes */
    XERR_GEN_CLEANUP((pSrc == NULL || pPixr == NULL), IO_BADPARAMETER);
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &dstHeight));
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, &channels));
    XERR_GEN_CLEANUP(IO_CHANNELS(xColorType) != channels, IO_BADPARAMETER);
    XERR_GEN_CLEANUP((Int32)dwRowOffset >= dstHeight, IO_BADPARAMETER);
    if (dwRowOffset + dwStripHeight > (UInt32)dstHeight) {
	dwStripHeight = dstHeight - dwRowOffset;
    }

    /* Check the input PIXR */
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &dstBPL));
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
				       &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
				     &imageChannels[channel]));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr,
				     &imageChannels[channel]));
	}
    }
    pDst = imageChannels[0];

    /* Clear the image */
    XERR_PASS_CLEANUP(wp_TfxClearPixrRegion(pIPWState, pPixr, 0, dwRowOffset,
	dwWidth, dwStripHeight));

    cPtr = pSrc;
    iPtr = pDst + (dwRowOffset * dstBPL) + 0;
    rowsLeft = dwStripHeight;
    bytesLeft = dwBytesInSrc;
    srcBPL = tfxImg->minBytesPerRow;
    bytesPerRow = srcBPL;
    if (bytesPerRow > dstBPL) {
	bytesPerRow = dstBPL;
    }
    switch ((int)xCompression) {
    case IO_COMP_NOC:
        XERR_PASS(wp_TfxUnpackImage(pIPWState, cPtr, srcBPL, rowsLeft, channels,
	    imageChannels, dstBPL, dwRowOffset, &rowsFilled));
	*pdwRowsFilled += rowsFilled;
	break;
    case IO_COMP_MH:
	XERR_PASS_CLEANUP(wp_TfxDecompressStripMH(tfxImg, cPtr, bytesLeft,
	    iPtr, dstBPL, 0, rowsLeft, &rowsFilled));
	*pdwRowsFilled += rowsFilled;
	break;
    case IO_COMP_MR:
	XERR_PASS_CLEANUP(wp_TfxDecompressStripMR(tfxImg, cPtr, bytesLeft,
	    iPtr, dstBPL, 0, rowsLeft, &rowsFilled));
	*pdwRowsFilled += rowsFilled;
	break;
    case IO_COMP_MMR:
	XERR_PASS_CLEANUP(wp_TfxDecompressStripMMR(tfxImg, cPtr, bytesLeft,
	    iPtr, dstBPL, 0, rowsLeft, &rowsFilled));
	*pdwRowsFilled += rowsFilled;
	break;
    case IO_COMP_JPEG:
	if ((tfxImg->imgFlags & IO_IMGFLAGS_IPCoreJPEG) ||
	    IO_BPP(tfxImg->imgInfo.xColorType) != 24) {
	    XERR_IGNORE(wp_TfxDecompressStripJPEG(tfxImg, cPtr,
		bytesLeft, pPixr, rowsLeft, dwRowOffset, &bytesUsed));
	} else {
	    XERR_IGNORE(wp_TfxDecompressStripJPEG_IJL(tfxImg, cPtr,
		bytesLeft, pPixr, rowsLeft, dwRowOffset, &bytesUsed));
	}
	*pdwRowsFilled += dwStripHeight;
	XERR_PASS_CLEANUP(__error__);
	break;
#ifndef NODGP
    case IO_COMP_TOKENIZATION:
    case IO_COMP_LLTOKENIZATION:
    case IO_COMP_JBIG2:
    case IO_COMP_LLJBIG2:
	XERR_PASS_CLEANUP(wp_TfxDetokenizeImage(tfxImg, cPtr, bytesLeft, iPtr,
	    dstBPL));
	*pdwRowsFilled += rowsLeft;
	break;
#endif

	/* Can't decompress these */
    case IO_COMP_OTHER:
    case IO_COMP_UNKNOWN:
    default:
	XERR_PASS_CLEANUP(IO_NOSUPPORT);
    }

    /* Check existing format, and desired format, and adjust when mismatched
     */
    /* Deal with planar configuration of color images */
    if (tfxImg->imgInfo.xColorType == IO_COLORTYPE_COLOR24) {
	switch (tfxImg->fields.PlanarConfiguration) {
	case 2:	/* Planar:  RRRR  GGGG  BBBB */
	    if ((dwFlags & IO_IMGFLAGS_ColorFormat) == IO_IMGFLAGS_ColorChunky) {
		/* We have planar, they want chunky */
		switch (tfxImg->imgInfo.xCompression) {
		case IO_COMP_NOC:
		case IO_COMP_JPEG:	/* JPEG is always decompressed to planar */
/*//////  Fix this one day.  We shouldn't have anyone request chunky */
		    XERR_PASS_CLEANUP(IO_NOSUPPORT);
		}
	    }
	    break;
	case 1:	/* Chunky: RGBRGBRGB... */
	default:
	    if ((dwFlags & IO_IMGFLAGS_ColorFormat) == IO_IMGFLAGS_ColorPlanar) {
		/* We have chunky, they want planar */
		switch (tfxImg->imgInfo.xCompression) {
		case IO_COMP_JPEG:	/* JPEG gets decompressed to planar */
		    break;
		case IO_COMP_NOC:
		    conversion |= TFX_CONV_UNCHUNK;
		    break;
		default:
		    XERR_GEN_CLEANUP(1, IO_NOSUPPORT);
		}
	    }
	}
    }

    /* Mark as needing inverted, if photometric interpretation indicates opposite */
    switch ((int)tfxImg->imgInfo.xColorType) {
    case IO_COLORTYPE_BINARY:
    case IO_COLORTYPE_GRAY4:
    case IO_COLORTYPE_GRAY8:
	switch (tfxImg->fields.PhotometricInterpretation) {
	case TFX_TIFFPHOTOMETRIC_BLACKISZERO:
	case TFX_TIFFPHOTOMETRIC_YCBCR:
	case TFX_TIFFPHOTOMETRIC_CIELAB:
	case TFX_TIFFPHOTOMETRIC_ICCLAB:
	case TFX_TIFFPHOTOMETRIC_ITULAB:
	    if ((dwFlags & IO_IMGFLAGS_PhotoInterp) ==
		IO_IMGFLAGS_WhiteIsZero) {
		/* We have black-is-zero, they want white-is-zero */
		conversion |= TFX_CONV_INVERT;
	    }
	    break;
	case TFX_TIFFPHOTOMETRIC_WHITEISZERO:
	    if ((dwFlags & IO_IMGFLAGS_PhotoInterp) ==
		IO_IMGFLAGS_BlackIsZero) {
		/* We have white-is-zero, they want black-is-zero */
		conversion |= TFX_CONV_INVERT;
	    }
	    break;
	}
	break;
    case IO_COLORTYPE_COLOR24:
	break;
    default:
	break;
    } /* switch color type */

    /* Convert the colorspace, if needed */
    switch ((int)tfxImg->imgInfo.xColorType) {
    case IO_COLORTYPE_BINARY:
	break;
    case IO_COLORTYPE_GRAY4:
	break;
    case IO_COLORTYPE_GRAY8:
	switch (tfxImg->fields.Compression) {
	case TFX_TIFFCOMP_NOCOMPRESSBYTE:
	case TFX_TIFFCOMP_JPEG:
	    switch (tfxImg->fields.PhotometricInterpretation) {
	    case TFX_TIFFPHOTOMETRIC_ITULAB:
		conversion |= TFX_CONV_ITULAB2SRGB;
		break;
	    default:
		break;
	    }
	    break;
	}
	break;
    case IO_COLORTYPE_COLOR24:
	switch (tfxImg->fields.Compression) {
	case TFX_TIFFCOMP_NOCOMPRESSBYTE:
	case TFX_TIFFCOMP_JPEG:
	    switch (tfxImg->fields.PhotometricInterpretation) {
	    case TFX_TIFFPHOTOMETRIC_CIELAB:
		conversion |= TFX_CONV_CIELAB2SRGB;
		break;
	    case TFX_TIFFPHOTOMETRIC_ICCLAB:
		conversion |= TFX_CONV_ICCLAB2SRGB;
		break;
	    case TFX_TIFFPHOTOMETRIC_ITULAB:
		conversion |= TFX_CONV_ITULAB2SRGB;
		break;
	    default:
		break;
	    }
	    break;
	}
	break;
    default:
	break;
    } /* switch color type */

    /* If output in byte-order is desired, convert from PIXR long-packed bytes */
    if (tfxImg->imgFlags & IO_IMGFLAGS_ByteOrder) {
	conversion |= TFX_CONV_DWORD2BYTEORDER;
    }

    /* Do all the conversions */
    if (conversion) {
	XERR_PASS_CLEANUP(w_TfxConvertImage(tfxImg, pPixr, conversion, dwRowOffset,
	    *pdwRowsFilled));
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}



/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy image data to a PIXR from a byte-packed image stream
 *
 * LongDescription:::
 *  wp_TfxUnpackImage() copies image data to a PIXR from a byte-packed image
 *  buffer. This is used to retrieve the "uncompressed" form of data,
 *  independent of the platform. If the source image is multi-channel,
 *  then the source image is one channel after another.
 *  <BR>
 *  The source image buffer is left intact.
 *
 * Input:::
 *  pIPWState::			Ptr to IPCore state struct.
 *  src::			Input image stream
 *  srcBPL::			Bytes per line of packed data
 *  height::			Number of lines (each channel) to transfer
 *  channels::			Number of channels to transfer
 *
 * Output:::
 *  dstChannels::		Destination image buffer(s)
 *  dstBPL::			Destination image bytes per line
 *  dstRowOffset::		Destination row offset
 *  pRowsFilled::		Rows of image data sucessfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxUnpackImage(
	       IPWState *pIPWState,
	       UInt8 *src,
	       Int32 srcBPL,
	       Int32 height,
	       Int32 channels,
	       UInt8 **dstChannels,
	       Int32 dstBPL,
	       UInt32 dstRowOffset,
	       UInt32 *pRowsFilled
)
{
    GraySwapNearProc swapLineProc;
    UInt8 *dst;
    Int32 channel;
    Int32 row, lineBPL;
    UInt8 *line = NULL;
    PROLOGUE;

    /* Start off with nothing filled */
    *pRowsFilled = 0UL;

    XERR_GEN_CLEANUP(srcBPL > dstBPL, IO_BADPARAMETER);

    ip_getByteSwapProc (pIPWState, 0, cPixrToPixr, NULL, NULL, &swapLineProc);

    /* Compute the intermediate line BPL */
    lineBPL = (srcBPL + 3) & ~3;

    /* Allocate an intermediate line buffer (for swapping) */
    XERR_GEN_CLEANUP(((line = (UInt8 *)malloc(lineBPL)) == NULL),
		      IO_NOMEMORY);

    for (channel = 0; channel < channels; channel++) {
      dst = dstChannels[channel] + dstRowOffset * dstBPL;
      for (row = 0; row < height; row++) {

	/* Copy the byte packed image line to the intermediate line */
	memcpy(line, src, srcBPL);
	src += srcBPL;

	/* Swap the intermediate line to the destination image */
	(*swapLineProc)(pIPWState, (void *)line, (void *)dst, lineBPL / 4);
	dst += dstBPL;
      }
    }

    free(line);
    *pRowsFilled = height;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (line) {
      free(line);
    }

     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image data used to represent one image for an even layer
 *
 * LongDescription:::
 *  w_TfxImageReadLayer() returns the entire image associated with all the
 *  images belonging to an even numbered layer, as described by the image
 *  return via TFX_GetLayerInfo().
 *
 * Input:::
 *  pIPWState::			IPW state
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetLayerInfo()
 *
 * Output:::
 *  pPixr::			Image data structure
 *  pdwRowsFilled::		Number of rows successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function allows piecing together the various images separated by
 *  into overlapping or abutting same-width images (it does NOT take care
 *  of creating a mask for other conditions).
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_TfxImageReadLayer(
	IPWState *pIPWState,
	IO_IMGHANDLE tfxImgHandle,
	PIXR *pPixr,
	UInt32 *pdwRowsFilled
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_IMGINFO *pTftfxImgInfo;
    Int32 pixrHeight;
    UInt32 rowsFilled = 0UL;
    PIXR *newPixr = NULL, *openedPixr = NULL;
    IO_COMPRESSION layerCompression;
    UInt32 layerXOffset, layerYOffset, layerXRes, layerYRes;
    UInt32 layerWidth, layerHeight;
    UInt32 imgXOffset, imgYOffset, imgXRes, imgYRes;
    UInt32 imgWidth, imgHeight, imgBestStripHeight;
    IO_COMPRESSION imgComp;
    IO_COLORTYPE imgColorType, layerColorType;
    UInt8 layerPalette[768];
    IO_IMGHANDLE imgHandle;
    UInt32 compNum;
    IO_COMPRESSION compType[10] = {IO_COMP_MH, IO_COMP_MR, IO_COMP_MMR, IO_COMP_NOC,
				  IO_COMP_TOKENIZATION, IO_COMP_LLTOKENIZATION,
				  IO_COMP_JBIG2, IO_COMP_LLJBIG2,
				  IO_COMP_JPEG};
    UInt32 imgNum, layer;
    ScaledInt1616 xres, yres;
/*    STREL *pStrel3 = NULL; */
    TFX_IMG *primImg;
    ColorType ctype = cColorGray;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    TFX_CheckHandle(tfxImgHandle);

    pTftfxImgInfo = &tfxImg->imgInfo;

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((tfxImg->tfxPage->seqStatus & TFX_PAGE_SEQ_GETINFO) == 0,
	IO_SEQUENCE);
    layer = tfxImg->imgInfo.dwImgLayer;
/*    XERR_GEN_CLEANUP(layer & 1, IO_BADPARAMETER);  /* Should be even layer */ */

    /* If there is only one image, return it instead, since we don't
     * have to mess around with going through the binary pieces */
    if (tfxImg->tfxPage->pageInfo.pdwImageCounts[layer-1] == 1) {

	/* Get the characteristics of the image */
	XERR_WARNOK_CLEANUP(TFX_GetImageInfo((IO_PAGEHANDLE)tfxImg->tfxPage,
	    layer, 0, &imgXOffset, &imgYOffset,
	    &imgXRes, &imgYRes, &imgWidth, &imgHeight, &imgBestStripHeight,
	    &imgComp, &imgColorType, NULL, &imgHandle));

	/* Read it directly */
	XERR_PASS_CLEANUP(TFX_ImageReadStart(imgHandle, 0UL));
	XERR_PASS_CLEANUP(w_TfxReadStrip(pIPWState, imgHandle,
	    0, imgHeight, pPixr, 0, pdwRowsFilled));
	XERR_PASS_CLEANUP(TFX_ImageReadFinish(imgHandle));

	/* Return */
	NORMAL_RETURN;
    }

    /* Get the characteristics of the layer image */
    XERR_PASS_CLEANUP(TFX_GetLayerInfo((IO_PAGEHANDLE)tfxImg->tfxPage,
	layer, &layerXOffset, &layerYOffset, &layerXRes, &layerYRes,
	&layerWidth, &layerHeight, &layerCompression, &layerColorType,
	&layerPalette[0], &tfxImgHandle));

    /* Examine the PIXR */
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &pixrHeight));

    /* There should be some consistency */
    XERR_GEN_CLEANUP(tfxImgHandle != (IO_IMGHANDLE)tfxImg, IO_INTERNAL_ERROR);
    XERR_GEN_CLEANUP((Int32)layerHeight != pixrHeight, IO_BADPARAMETER);

    /* Get the page reference image */
    primImg = (TFX_IMG *)tfxImg->tfxPage->pageInfo.tfxPrimImgHandle;

    /* Loop through the compression methods */
    for (compNum = 0; compNum < sizeof(compType) / sizeof(*compType);
	 compNum++) {

	/* Loop through each image within the layer */
	for (imgNum = 0;
	     imgNum < tfxImg->tfxPage->pageInfo.pdwImageCounts[layer - 1];
	     imgNum++) {

	    /* Get the characteristics of the image */
	    XERR_WARNOK_CLEANUP(TFX_GetImageInfo((IO_PAGEHANDLE)tfxImg->tfxPage,
		pTftfxImgInfo->dwImgLayer, imgNum, &imgXOffset, &imgYOffset,
		&imgXRes, &imgYRes, &imgWidth, &imgHeight, &imgBestStripHeight,
		&imgComp, &imgColorType, NULL, &imgHandle));

	    if (imgComp != compType[compNum]) {
		continue;
	    }

	    /* Make sure this image really fits */
	    XERR_GEN_CLEANUP(imgXOffset < layerXOffset, IO_INTERNAL_ERROR);
	    XERR_GEN_CLEANUP(imgYOffset < layerYOffset, IO_INTERNAL_ERROR);

	    /* Adjust for the layer position */
	    imgXOffset -= layerXOffset;
	    imgYOffset -= layerYOffset;

	    /* Adjust the position for the page resolution difference */
	    imgXOffset = imgXOffset * imgXRes / primImg->imgInfo.dwXResolution;
	    imgYOffset = imgYOffset * imgYRes / primImg->imgInfo.dwYResolution;

	    /* Make sure this image really fits */
	    XERR_GEN_CLEANUP(imgXOffset + imgWidth > layerWidth, IO_INTERNAL_ERROR);
	    XERR_GEN_CLEANUP(imgYOffset + imgHeight > layerHeight, IO_INTERNAL_ERROR);

	    /* Allocate a PIXR for the entire image */
	    newPixr = NULL;
	    xres = Int32_TO_ScaledInt1616(imgXRes);
	    yres = Int32_TO_ScaledInt1616(imgYRes);
	    if (imgColorType == IO_COLORTYPE_COLOR24) {
		ctype = cColorRGB;
	    }
	    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, imgWidth,
		imgHeight, IO_BPPE(imgColorType), xres, yres, 32,
		IO_CHANNELS(imgColorType), ctype, NULL, &newPixr));

	    /* Clear the image */
	    XERR_PASS_CLEANUP(wp_TfxClearPixrRegion(pIPWState, newPixr, 0, 0,
		imgWidth, imgHeight));

	    /* Get the entire image */
	    XERR_PASS_CLEANUP(TFX_ImageReadStart(imgHandle, 0UL));
	    XERR_PASS_CLEANUP(w_TfxReadStrip(pIPWState, imgHandle,
		0, imgHeight, newPixr, 0, &rowsFilled));
	    XERR_PASS_CLEANUP(TFX_ImageReadFinish(imgHandle));

	    /* Add it in to the combined image */
	    switch ((int)imgComp) {
	    case IO_COMP_NOC:
	    case IO_COMP_MH:
	    case IO_COMP_MR:
	    case IO_COMP_MMR:
	    case IO_COMP_TOKENIZATION:
	    case IO_COMP_LLTOKENIZATION:
	    case IO_COMP_JBIG2:
	    case IO_COMP_LLJBIG2:

		/* Use a boolean OR to recombine */
		XERR_PASS_CLEANUP(w_rasterop(pIPWState, 0, 0, newPixr,
		    PIX_SRC | PIX_DST, imgXOffset, imgYOffset, imgWidth,
		    imgHeight, pPixr));
		break;
	    case IO_COMP_JPEG:

		/* Use a copy */
		XERR_PASS_CLEANUP(w_rasterop(pIPWState, 0, 0, newPixr,
		    PIX_SRC, imgXOffset, imgYOffset, imgWidth,
		    imgHeight, pPixr));
		break;

	    default:
		break;
	    } /* switch (compression type) */

	    /* Clean up */
	    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &newPixr));
	    if (openedPixr) {
		XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &openedPixr));
	    }

	}  /* For each image in layer */
    }  /* For each compression type in order */

    /* Output the total number of rows in user's image filled */
    *pdwRowsFilled = layerHeight;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (newPixr) {
	w_destroyPixr(pIPWState, &newPixr);
    }
    if (openedPixr) {
	w_destroyPixr(pIPWState, &openedPixr);
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initiate a common IPWrap connection
 *
 * LongDescription:::
 *  wp_TfxIPWState() initiates an IPWrap state, to be used throughout the
 *  TIFF-FX API.  If the state has been initated previously, as attached through
 *  the document handle, it is simply returned.
 *
 * Input:::
 *  tfxDoc::			Valid open document structure
 *
 * Output:::
 *  ppIPWState::		Pointer to the resulting IPWrap state
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxIPWState(
	TFX_DOC *tfxDoc,
	IPWState **ppIPWState
)
{
    PROLOGUE;

    /* Default to return nothing */
    *ppIPWState = NULL;

    TFX_CheckHandle(tfxDoc);
    XERR_GEN_CLEANUP(ppIPWState == NULL, IO_BADPARAMETER);

    *ppIPWState = (IPWState *)tfxDoc->tfxFile->instance->pIPWState;
    if (*ppIPWState == NULL) {

	/* Initiate an IPCore interface
	 */
	XERR_PASS_CLEANUP(w_IPWStateInit(wp_TfxIPCMalloc, wp_TfxIPCCalloc,
		wp_TfxIPCRealloc, wp_TfxIPCFree, (void *)tfxDoc->tfxFile->instance,
		wp_TfxIPCMalloc, wp_TfxIPCCalloc, wp_TfxIPCRealloc, wp_TfxIPCFree,
		(void *)tfxDoc->tfxFile->instance, NULL, stderr, 0, NULL, NULL,
		ppIPWState));
	tfxDoc->tfxFile->instance->pIPWState = (UInt32)*ppIPWState;
	tfxDoc->tfxFile->instance->myIPWState = 1;
    }
    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress a CCITT G3 image (MH)
 *
 * LongDescription:::
 *  wp_DecompressStripMH() decompresses a strip of CCITT G3 (MH = T.4)
 *  data. The temporary PIXR found in tfxImg is used to store the image data
 *  decompressed.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  pCmpData::			Pointer to compressed data
 *  dwCmpBytes::		Number of compressed bytes
 *
 * Output:::
 *  pImgData::			Pointer to the first pixel of image data
 *  dwBytesPerRow::		Bytes per row of image data
 *  dwRowOffset::		Row offset in image data to place output
 *  dwStripHeight::		Number of rows to decompress
 *  pdwRowsFilled::		Number of rows decompressed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxDecompressStripMH(
    TFX_IMG *tfxImg,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    UInt8 *pImgData,
    UInt32 dwBytesPerRow,
    UInt32 dwRowOffset,
    UInt32 dwStripHeight,
    UInt32 *pdwRowsFilled
)
{
    TFX_INSTANCE *instData;
    TFX_IMGINFO *pTftfxImgInfo;
    IPWState *pIPWState;
    Int32 w;
    void *compInfo;
    UInt32 imgExh, cmpExh;
    Int32 stateBlockSize;
    UInt32 padType = cNoPad;
    UInt8 *iPtr, *pDst;
    UInt8 *cPtr;
    Int32 bytesLeft, bytesUsed, rowsLeft, pct;
    Bool contImg, contCmp;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));
    instData = TFX_IMGINST(tfxImg);

    pTftfxImgInfo = &tfxImg->imgInfo;
    w = pTftfxImgInfo->dwWidth;

    if (tfxImg->tmpCmpData == NULL) {	/* First strip */

	/* Prepare for pre-decode */
	XERR_PASS_CLEANUP(i_getSizeOfCCITTDecompressionInfo(
	    pIPWState,
	    w + 2*cFrameBits, &stateBlockSize));

	/* Allocate temporary compression data, attach to image */
	IO_CHECKMALLOC(tfxImg->tmpCmpData =
	    (TFX_MISCMEM *)TFX_MALLOC(sizeof(TFX_MISCMEM)));
	IO_CHECKMALLOC(tfxImg->tmpCmpData->mem =
	    (UInt8 *)TFX_MALLOC(stateBlockSize));
	tfxImg->tmpCmpData->next = NULL;
	compInfo = tfxImg->tmpCmpData->mem;

	/* Use TIFF option for padding */
	if ((tfxImg->fields.T4Options & 4) ||
	    (tfxImg->fields.Compression == TFX_TIFFCOMP_CCITT3)) {
	    padType = cPadToByte;		/* Fill bits for byte alignment */
	}

	/* Initialize the decompressor */
	XERR_PASS_CLEANUP(i_initializeCCITTDecompressionInfo(
	    pIPWState,
	    compInfo, w + 2*cFrameBits, 1,
	    padType, stateBlockSize));
    } else {
	compInfo = tfxImg->tmpCmpData->mem;
    }

    pDst = pImgData;
    iPtr = pDst - (cFrameBits * IO_BPPE(pTftfxImgInfo->xColorType)) / 8;
    cPtr = pCmpData;
    rowsLeft = dwStripHeight;
    bytesLeft = dwCmpBytes;
    contImg = 0;
    contCmp = 0;

    do {
	XERR_PASS_CLEANUP(i_decompressRowsMH (pIPWState, compInfo,
	    iPtr, dwRowOffset,
	    rowsLeft, dwBytesPerRow, 0, cPtr, bytesLeft, contImg,
	    contCmp, 1, &imgExh, &cmpExh, &pct, &bytesUsed));

	bytesLeft -= bytesUsed;

	if (imgExh == 1 && cmpExh == 0) {

	    /* Okay, we can live with it, but it's wrong */

	} else if (cmpExh == 1 && imgExh == 0) {

	    /* Can't decompress all of it, bad! */
	    XERR_PASS_CLEANUP(IO_BADFILEFORMAT);
	}
	contImg = 1;
	contCmp = 1;
    } while (imgExh == 0 && cmpExh == 0 && bytesLeft > 0);
    rowsLeft = 0;

    *pdwRowsFilled = dwStripHeight - rowsLeft;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress a CCITT G4 ?-D image (MR)
 *
 * LongDescription:::
 *  wp_DecompressStripMR() decompresses a strip of CCITT G4 ?-D (MR = T.6?)
 *  data. The temporary PIXR found in tfxImg is used to store the image data
 *  decompressed.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  pCmpData::			Pointer to compressed data
 *  dwCmpBytes::		Number of compressed bytes
 *
 * Output:::
 *  pImgData::			Pointer to the first pixel of image data
 *  dwBytesPerRow::		Bytes per row of image data
 *  dwRowOffset::		Row offset in image data to place output
 *  dwStripHeight::		Number of rows to decompress
 *  pdwRowsFilled::		Number of rows decompressed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxDecompressStripMR(
    TFX_IMG *tfxImg,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    UInt8 *pImgData,
    UInt32 dwBytesPerRow,
    UInt32 dwRowOffset,
    UInt32 dwStripHeight,
    UInt32 *pdwRowsFilled
)
{
    TFX_INSTANCE *instData;
    TFX_IMGINFO *pTftfxImgInfo;
    IPWState *pIPWState;
    Int32 w;
    void *compInfo;
    UInt32 imgExh, cmpExh;
    Int32 stateBlockSize;
    UInt32 padType = cNoPad;
    UInt8 *iPtr, *pDst;
    UInt8 *cPtr;
    Int32 bytesLeft, bytesUsed, rowsLeft, pct;
    Bool contImg, contCmp, is2D;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));
    instData = TFX_IMGINST(tfxImg);

    pTftfxImgInfo = &tfxImg->imgInfo;
    w = pTftfxImgInfo->dwWidth;

    if (tfxImg->tmpCmpData == NULL) {	/* First strip */

	/* Prepare for pre-decode */
	XERR_PASS_CLEANUP(i_getSizeOfCCITTDecompressionInfo(
	    pIPWState,
	    w + 2*cFrameBits, &stateBlockSize));

	/* Allocate temporary compression data, attach to image */
	IO_CHECKMALLOC(tfxImg->tmpCmpData =
	    (TFX_MISCMEM *)TFX_MALLOC(sizeof(TFX_MISCMEM)));
	IO_CHECKMALLOC(tfxImg->tmpCmpData->mem =
	    (UInt8 *)TFX_MALLOC(stateBlockSize));
	tfxImg->tmpCmpData->next = NULL;
	compInfo = tfxImg->tmpCmpData->mem;

	/* Use TIFF options for 2-D encoding, and bit padding */
	if (tfxImg->fields.T4Options & 1) {		/* 2-D encoding */
	    is2D = 1;
	} else {
	    is2D = 0;
	}
	if (tfxImg->fields.T4Options & 4) {		/* Fill bits for byte alignment */
	    padType = cPadToByte;
	}

	/* Initialize the decompressor */
	XERR_PASS_CLEANUP(i_initializeCCITTDecompressionInfo(
	    pIPWState,
	    compInfo, w + 2*cFrameBits, is2D,
	    padType, stateBlockSize));
    } else {
	compInfo = tfxImg->tmpCmpData->mem;
    }

    pDst = pImgData;
    iPtr = pDst - (cFrameBits * IO_BPPE(pTftfxImgInfo->xColorType)) / 8;
    cPtr = pCmpData;
    rowsLeft = dwStripHeight;
    bytesLeft = dwCmpBytes;
    contImg = 0;
    contCmp = 0;

    do {
	XERR_PASS_CLEANUP(i_decompressRowsMR (pIPWState, compInfo,
	    iPtr, dwRowOffset,
	    rowsLeft, dwBytesPerRow, 0, cPtr, bytesLeft, contImg,
	    contCmp, 1, &imgExh, &cmpExh, &pct, &bytesUsed));

	bytesLeft -= bytesUsed;

	if (imgExh == 1 && cmpExh == 0) {

	    /* Okay, we can live with it, but it's wrong */

	} else if (cmpExh == 1 && imgExh == 0) {

	    /* Can't decompress all of it, bad! */
	    XERR_PASS_CLEANUP(IO_BADFILEFORMAT);
	}
	contImg = 1;
	contCmp = 1;
    } while (imgExh == 0 && cmpExh == 0 && bytesLeft > 0);
    rowsLeft = 0;

    *pdwRowsFilled = dwStripHeight - rowsLeft;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress a CCITT G4 image (MMR)
 *
 * LongDescription:::
 *  wp_DecompressStripMMR() decompresses a strip of CCITT G4 (MMR = T.6)
 *  data. The temporary PIXR found in tfxImg is used to store the image data
 *  decompressed.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  pCmpData::			Pointer to compressed data
 *  dwCmpBytes::		Number of compressed bytes
 *
 * Output:::
 *  pImgData::			Pointer to the first pixel of image data
 *  dwBytesPerRow::		Bytes per row of image data
 *  dwRowOffset::		Row offset in image data to place output
 *  dwStripHeight::		Number of rows to decompress
 *  pdwRowsFilled::		Number of rows decompressed
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxDecompressStripMMR(
    TFX_IMG *tfxImg,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    UInt8 *pImgData,
    UInt32 dwBytesPerRow,
    UInt32 dwRowOffset,
    UInt32 dwStripHeight,
    UInt32 *pdwRowsFilled
)
{
    TFX_INSTANCE *instData;
    TFX_IMGINFO *pTftfxImgInfo;
    IPWState *pIPWState;
    Int32 w;
    void *compInfo;
    UInt32 imgExh, cmpExh;
    Int32 stateBlockSize;
    UInt32 padType = cNoPad;
    UInt8 *iPtr, *pDst;
    UInt8 *cPtr;
    Int32 bytesLeft, bytesUsed, rowsLeft, pct;
    Bool contImg, contCmp;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));
    instData = TFX_IMGINST(tfxImg);

    pTftfxImgInfo = &tfxImg->imgInfo;
    w = pTftfxImgInfo->dwWidth;

    if (tfxImg->tmpCmpData == NULL) {	/* First strip */

	/* Prepare for pre-decode */
	XERR_PASS_CLEANUP(i_getSizeOfCCITTDecompressionInfo(
	    pIPWState,
	    w + 2*cFrameBits, &stateBlockSize));

	/* Allocate temporary compression data, attach to image */
	IO_CHECKMALLOC(tfxImg->tmpCmpData =
	    (TFX_MISCMEM *)TFX_MALLOC(sizeof(TFX_MISCMEM)));
	IO_CHECKMALLOC(tfxImg->tmpCmpData->mem =
	    (UInt8 *)TFX_MALLOC(stateBlockSize));
	tfxImg->tmpCmpData->next = NULL;
	compInfo = tfxImg->tmpCmpData->mem;

	XERR_PASS_CLEANUP(i_initializeCCITTDecompressionInfo(
	    pIPWState,
	    compInfo, w + 2*cFrameBits, 1,
	    padType, stateBlockSize));
    } else {
	compInfo = tfxImg->tmpCmpData->mem;
    }

    pDst = pImgData;
    iPtr = pDst - (cFrameBits * IO_BPPE(pTftfxImgInfo->xColorType)) / 8;
    cPtr = pCmpData;
    rowsLeft = dwStripHeight;
    bytesLeft = dwCmpBytes;
    contImg = 0;
    contCmp = 0;

    do {
	XERR_PASS_CLEANUP(i_decompressRowsMMR (pIPWState, compInfo,
	    iPtr, dwRowOffset,
	    rowsLeft, dwBytesPerRow, 0, cPtr, bytesLeft, contImg,
	    contCmp, 1, &imgExh, &cmpExh, &pct, &bytesUsed));

	bytesLeft -= bytesUsed;

	if (imgExh == 1 && cmpExh == 0) {

	    /* Okay, we can live with it, but it's wrong */

	} else if (cmpExh == 1 && imgExh == 0) {

	    /* Can't decompress all of it, bad! */
	    XERR_PASS_CLEANUP(IO_BADFILEFORMAT);
	}
	contImg = 1;
	contCmp = 1;
    } while (imgExh == 0 && cmpExh == 0 && bytesLeft > 0);
    rowsLeft = 0;

    *pdwRowsFilled = dwStripHeight - rowsLeft;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*
$Log: tfxrpixr.c,v $
Revision 1.2  2003/02/10 22:53:31  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:11  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.25  2001/12/26 21:14:44  dabercro
Only use IJL if 24-bit images (doesn't work for 8- or 4-bit yet).

Revision 1.24  2001/12/12 00:10:23  dabercro
Add option to choose which JPEG codec to use

Revision 1.23  2001/11/07 21:46:18  dabercro
Make base color signed long, to accomodate CIELAB.
Refined NODGP code carve-out.

Revision 1.22  2001/08/22 19:49:34  dabercro
Added support for new field: IFDBaseColor

Revision 1.21  2001/07/20 23:14:13  dabercro
Remove insertion of JPEGTables just before decompression.  Moved to
being included in the compressed data stream earlier on.

Revision 1.20  2001/06/29 17:50:37  dabercro
Concatenate the tables and the JPEG stream together, before decode.

Revision 1.19  2001/06/27 18:56:35  dabercro
Setup and do image conversions colorspace, chunk/planar, invert, etc.

Revision 1.18  2001/06/01 18:07:41  dabercro
Comment fix.

Revision 1.17  2001/04/02 18:24:07  dabercro
Explicitly mention other compression type in switch.

Revision 1.16  2001/03/13 02:59:12  dabercro
Remove lossy-dither and wavelet as valid compression types.

Revision 1.15  2001/01/18 00:21:31  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.14  2000/11/18 00:53:26  dabercro
Avoid cross polenation of data handles across sublibraries
(XIF, MRC, TFX), especially for CopyImage.  Added a handle
type to structure, now checked also.

Revision 1.13  2000/10/18 20:27:55  dabercro
Add other Photometric types for invert check.

Revision 1.12  2000/10/16 18:22:53  dabercro
Major adds to support stripes, 0-byte strips, and ImageBaseColor.

Revision 1.11  2000/09/27 21:54:30  dabercro
A couple of check to make sure we don't seg fault.

Revision 1.10  2000/09/21 21:19:14  dabercro
Add variable-sized strips: StripRowCounts.

Revision 1.9  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.8  2000/08/30 23:29:07  dabercro
Check for both JBIG2 and Tokenization.

Revision 1.7  2000/08/25 23:50:57  dabercro
Fix from Mark Tobin for inverting gray images.
Fix to allow decoding of MH (comp = 2) images (unsupported,
but what the hey).

Revision 1.6  2000/08/22 01:03:23  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.5  2000/08/01 00:02:02  dabercro
Update the clip regions when not at the origin (0,0).

Revision 1.4  2000/06/07 00:00:17  dabercro
Continued work, lots of stuff.

Revision 1.3  2000/04/12 19:06:09  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.2  2000/03/11 02:06:34  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.25  1998/07/27 20:54:11  aber
aber on Mon Jul 27 13:53:43 PDT 1998

tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

Added a new function to specifically propogate the IPWState structure.

Revision 1.24  1998/05/05 07:15:24  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

CommonCheck: Additions necessary to fill in the never completed functions
TFX_Compress and TFX_Decompress.

Revision 1.23  1998/03/16 21:01:53  aber
aber on Mon Mar 16 13:00:48 PST 1998

tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

In read-layer function, allow uncompressed images to be added.

Revision 1.22  1998/02/09 20:21:25  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

CommonCheck: Signed/unsigned casting to make MSVC happy.

Revision 1.21  1998/01/20 03:21:55  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

Fix status check to work in read/write mode.

Revision 1.20  1998/01/14 00:16:30  ddavies
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.19  1997/11/01 20:51:18  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

Fix color palette handling (don't add palette to PIXR if already there).
Add exact strip buffering for wavelet.

Revision 1.18  1997/07/15 23:42:27  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

Buffering fixes.

Revision 1.17  1997/07/14 23:22:59  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

Review image buffering during read. Rename variables to be
more intuitive. Allow non-recommended strip sizes.

Revision 1.16  1997/06/26 23:29:13  aber
aber on Thu Jun 26 16:28:26 PDT 1997

tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

Make potential ansification fixes that the Mac might complain about.

Revision 1.15  1997/06/26 22:25:37  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

Convert "white-is-zero" TIFF-FX 2.0 image (invert).

Revision 1.14  1997/06/26 20:05:01  ddavies
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

CommonCheck: Add inclusion of new file rast.pub.

Revision 1.13  1997/06/23 22:59:03  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

CommonCheck: Complete function calls for paletted image support.

Revision 1.12  1997/06/20 17:44:38  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

When asked for a layer, return the single image (non-binary
primary image case).

Revision 1.11  1997/06/18 00:41:26  aber
aber on Tue Jun 17 17:41:26 PDT 1997

tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

CommonCheck: Updated prototype for TFX_GetLayerInfo.

Revision 1.10  1997/06/17 23:45:06  aber
aber on Tue Jun 17 16:44:44 PDT 1997

tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

Update return value before returning.

Revision 1.9  1997/06/17 22:35:18  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

Clean up if completed the image read -- was simply returning.

Revision 1.7  1997/06/12 00:50:21  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

Enable lossless tokenization. If TIFF-FX has unsupported compression type,
return no support, vs. bad file format.

Revision 1.6  1997/06/10 23:34:01  aber
aber on Tue Jun 10 16:33:31 PDT 1997

tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

Fix documentation (nag nag).
Clean up memory in proper loop, fixes memory leak.

Revision 1.5  1997/06/10 01:26:40  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.4  1997/05/30 00:51:02  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.3  1997/05/20 22:29:06  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.2  1997/05/20 00:54:13  aber
tfxrpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxrpixr.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:37  aber
Initial revision

*/
