/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxrpriv.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxrpriv.c:	TIFF-FX READ private routines
 *
 *	TIFF-FX read-only private routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tfxrpriv.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:31 $");

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif
#endif

#include "tfximage.h"
#include "tfxrtiff.h"
#include "tfxrtfx.h"
#include "tfxrpriv.h"

/* TIFF-FX internal memory allocation adds some info to help detect leaks
 * and page faults. I place a key word before and after the
 * memory during the allocation, and check that it is still
 * in tact during the free. I also place the size of memory
 * before, so that I know where the key is after the block,
 * and to keep track of memory usage at any point.
 */
#define MEMORY_KEY		0x75c38a19	/* Ya, I just made it up */
#define KEYED_MEMORY	(sizeof(long) * 3)

/* Memory and I/O */
void * CDECL
wp_TfxMalloc(
    size_t bytes,
    Bool hasPtrs,
    void *clientMallocID
#ifdef SHOWERR
    , const char *file,
    const int line
#endif
)
{
    TfxMemoryTally *tally = (TfxMemoryTally *)clientMallocID;
    void *memory;
    UInt32 *lmem;
	
    if (clientMallocID == NULL || bytes == 0) {
	/* reference once, to avoid compiler warnings */
	hasPtrs = 0;
#ifdef SHOWERR
	{
	    int junk;
	    junk = (int)file;
	    junk = line;
	}
#endif
	return NULL;
    }

    
    bytes = (bytes + 3UL) & ~3UL;	/* Round to next word */
    if ((memory = (void *)malloc(bytes + KEYED_MEMORY)) == NULL) {
	return memory;
    }
    tally->bytesAllocated += bytes;
    if (tally->bytesAllocated > tally->peak) {
	tally->peak = tally->bytesAllocated;
    }
    
    /* Place memory key and size around buffer */
    lmem = (UInt32 *)memory;
    lmem[0] = bytes;
    lmem[1] = (UInt32)MEMORY_KEY;
    lmem[((bytes + KEYED_MEMORY) >> 2) - 1] = (UInt32)MEMORY_KEY;
    
    return (void *)(lmem + 2);
}

void * CDECL
wp_TfxCalloc(
    size_t nel,
    size_t size,
    Bool hasPtrs,
/*    void *clientMallocID */
    void *thing
#ifdef SHOWERR
    , const char *file,
    const int line
#endif
)
{
    IO_INSTANCE *inst = (IO_INSTANCE *)thing;
/*    IO_MemoryTally *tally = (IO_MemoryTally *)clientMallocID; */
    char *memory;
    UInt32 bytes = nel * size;
	
    if (thing == NULL || bytes == 0) {
	/* reference once, to avoid compiler warnings */
	hasPtrs = 0;
	return NULL;
    }


    /* Get the memory from the malloc routine */
    memory = inst->malloc(bytes, hasPtrs, (void *)inst->mallocID
#ifdef SHOWERR
    , file, line
#endif
    );

    if (memory == NULL) return memory;

    /* Now do the C part of Calloc */
    memset(memory, 0, bytes);

    return (void *)memory;
}

void * CDECL
wp_TfxRealloc(
    void *ptr,
    size_t size,
    Bool hasPtrs,
    void *thing
#ifdef SHOWERR
    , const char *file,
    const int line
#endif
)
{
    UInt32 oldSize;
    IO_INSTANCE *inst = (IO_INSTANCE *)thing;
    void *buffer;
	    
    if (hasPtrs) {
      return NULL;
    }
	
    if (thing == NULL) {
	return NULL;
    }
				    
    /* This isn't right in the case of client malloc!!!!!!!!!!!!!
     *
     * We need to know the size of the old buffer, to know how to 
     * reallocate it. In the case of the memory allocation I have,
     * I keep it stored with the buffer. The client may do anything
     * they desire, so I have no way of pinging for this information.
     */
    oldSize = wp_TfxAllocSize(ptr, inst->mallocID);
    buffer = inst->malloc(size, hasPtrs, inst->mallocID
#ifdef SHOWERR
	, file, line
#endif
	);
    
    if (oldSize > size) {
	memcpy(buffer, ptr, size);
    } else {
	memcpy(buffer, ptr, oldSize);
    }
    inst->free(ptr, oldSize, inst->mallocID);
    
    return buffer;
}



void CDECL
wp_TfxFree(
    void *buffer,
    size_t dwBytes,
    void *clientMallocID
)
{
    TfxMemoryTally *tally = (TfxMemoryTally *)clientMallocID;
    UInt32 *lmem;
    UInt32 bytes;
	
    if (clientMallocID == NULL) {
	dwBytes = 0;	    /* Reference once, to avoid compiler warnings */
	return;
    }
    
    if ((unsigned)buffer & 3) {
	fprintf(stderr, "Invalid memory free, buffer is not long aligned\n");
	goto memoryError;
    }
    
    lmem = ((UInt32 *)buffer) - 2;
    bytes = lmem[0];		/* If seg fault here, buffer doesn't contain 4
			 * bytes -- buffer is probably invalid */
    if (bytes < 4) {
	fprintf(stderr, "Memory corruption, memory before buffer invalid\n");
	goto memoryError;
    }
    if (lmem[1] != (UInt32)MEMORY_KEY) {
	fprintf(stderr, "Memory key corrupted, memory before buffer invalid\n");
	goto memoryError;
    }
    if (lmem[((bytes + KEYED_MEMORY) >> 2) - 1] != (UInt32)MEMORY_KEY) {
	fprintf(stderr, "Memory key corrupted, memory after buffer invalid\n");
	goto memoryError;
    }
    
    tally->bytesAllocated -= bytes;
    free((char *)lmem);
    return;		/* Normal return */
    
 memoryError:
    return;		/* Put break point here in debugger */
}

UInt32 CDECL
wp_TfxMemoryCheck(
    TfxMemoryTally *tally
)
{
    
#ifdef MYDEBUG  /* This is meant to be something other than _DEBUG */
    fprintf(stderr, "%d bytes currently allocated, %d bytes used at peak.\n",
	    tally->bytesAllocated, tally->peak);
#endif /* DEBUG */
    return tally->bytesAllocated;
}


UInt32 CDECL
wp_TfxAllocSize(
    void *buffer,
    void *clientID
)
{
    TfxMemoryTally *tally = (TfxMemoryTally *)clientID;
    UInt32 *lmem;
    UInt32 bytes;

    if (tally == NULL) {
	goto memoryError;
    }
    
    if ((unsigned)buffer & 3) {
	fprintf(stderr, "Invalid memory, buffer is not long aligned\n");
	goto memoryError;
    }
    
    lmem = ((UInt32 *)buffer) - 2;
    bytes = lmem[0];		/* If seg fault here, buffer doesn't contain 4
			 * bytes -- buffer is probably invalid */
    return bytes;		/* Normal return */
    
 memoryError:
    return 0UL;		/* Put break point here in debugger */
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Lock the file -- mark as unavailable by other threads.
 *
 * LongDescription:::
 *  wp_TfxFileLock() locks the file for full ownership by the calling thread,
 *  disallowing another thread to request ownership until this thread has
 *  released the file using TFX_FileUnLock().
 *
 * Input:::
 *  fileToken::			File token referencing file to lock
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxFileLock(
    TFX_TOKEN *fileToken
)
{
    PROLOGUE;

    XERR_GEN_CLEANUP(fileToken == NULL, IO_INTERNAL_ERROR);

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Unlock the file -- mark as free for the next thread to take control.
 *
 * LongDescription:::
 *  wp_TfxFileUnLock() unlocks the file from the current thread, allowing
 *  another thread to request full ownership of the file.
 *
 * Input:::
 *  fileToken::		File token referencing file to unlock
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxFileUnLock(
    TFX_TOKEN *fileToken
)
{
    PROLOGUE;

    XERR_GEN_CLEANUP(fileToken == NULL, IO_INTERNAL_ERROR);

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}



/***********************************************************
 *       IPCORE memory interface
 */

void * CDECL
wp_TfxIPCMalloc(
    size_t count,
    Bool hasPtrs,
    void *thing
)
{
    TFX_INSTANCE *inst = (TFX_INSTANCE *)thing;

    if (hasPtrs) {
      return NULL;
    }
	
    if (thing == NULL) {
	return NULL;
    }
    return inst->malloc(count, hasPtrs, inst->mallocID
#ifdef SHOWERR
	, __FILE__, __LINE__
#endif
	);
}



void * CDECL
wp_TfxIPCCalloc(
    size_t nel,
    size_t size,
    Bool hasPtrs,
    void *thing
)
{
    TFX_INSTANCE *inst = (TFX_INSTANCE *)thing;
    void *buffer;
	     
    if (hasPtrs) {
      return NULL;
    }
	
    if (thing == NULL) {
      return NULL;
    }
    buffer = inst->malloc(nel * size, hasPtrs, inst->mallocID
#ifdef SHOWERR
	, __FILE__, __LINE__
#endif
	);

    if (buffer != NULL) {
	memset(buffer, 0, nel * size);
    }
 
    return buffer;
}



void * CDECL
wp_TfxIPCRealloc(
    void *ptr,
    size_t size,
    Bool hasPtrs,
    void *thing
)
{
    UInt32 oldSize;
    TFX_INSTANCE *inst = (TFX_INSTANCE *)thing;
    void *buffer;
	    
    if (hasPtrs) {
      return NULL;
    }
	
    if (thing == NULL) {
	return NULL;
    }
				    
    /* This isn't right in the case of client malloc!!!!!!!!!!!!!
     *
     * We need to know the size of the old buffer, to know how to 
     * reallocate it. In the case of the memory allocation I have,
     * I keep it stored with the buffer. The client may do anything
     * they desire, so I have no way of pinging for this information.
     */
    oldSize = wp_TfxAllocSize(ptr, inst->mallocID);
    buffer = inst->malloc(size, hasPtrs, inst->mallocID
#ifdef SHOWERR
	, __FILE__, __LINE__
#endif
	);
    if (buffer != NULL) {
	if (oldSize > size) {
	    memcpy(buffer, ptr, size);
	} else {
	    memcpy(buffer, ptr, oldSize);
	}
	inst->free(ptr, oldSize, inst->mallocID);
	return buffer;
    }

    /* realloc failed, leave ptr alone and return NULL */
    return NULL;
}



void CDECL
wp_TfxIPCFree(
    void *ptr,
    size_t size,
    void *thing
)
{
    TFX_INSTANCE *inst = (TFX_INSTANCE *)thing;

    if (thing == NULL) {
	size = 0;	/* Reference once to avoid compiler warnings */
	return;
    }
    inst->free(ptr, 0, inst->mallocID);
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Allocate a new image handle, and assign it to its owner.
 *   
 *
 * LongDescription:::
 *  wp_TfxNewImg() allocates a new image handle, and assign it to its owner.
 *
 * Input:::
 *  tfxPage::			Data structure of valid page
 *  owner::			Pointer to pointer to image (points to new
 *				image's owner)
 *  ifdOffset::			IFD offset of the new image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxNewImg(
    TFX_PAGE *tfxPage,
    TFX_IMG **owner,
    Int32 ifdOffset
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_IMG *tfxImg = NULL;
    PROLOGUE;
    
    instData = TFX_PAGEINST(tfxPage);
    tfxFile = TFX_PAGEFILE(tfxPage);
    
    /* Begin construction of the primary image data */
    IO_CHECKMALLOC(tfxImg = (TFX_IMG *)TFX_MALLOC(sizeof(TFX_IMG)));
    memset(tfxImg, 0, sizeof(TFX_IMG));
    tfxImg->handleID = (UInt32)tfxImg;		/* self identity */
    tfxImg->handleType = TFX_HANDLE_TYPE;
    tfxImg->tfxPage = tfxPage;
    tfxImg->imgInfo.tfxImgHandle = (UInt32)tfxImg;
    
    tfxImg->ifdOffset = ifdOffset;
    
    /* Attach it to the owner */
    *owner = tfxImg;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Increment the number of images within a layer
 *
 * LongDescription:::
 *  wp_TfxIncrementImageCounts() makes the appropriate steps to increment the
 *  count of images within a layer, which may include reallocating the image
 *  counts array stored in the page info.
 *
 * Input:::
 *  tfxPage::			Data structure of valid page
 *  layer::			Layer to increment the count
 *  inc::			Number of times to increment
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxIncrementImageCounts(
	TFX_PAGE *tfxPage,
	UInt32 layer,
	Int32 inc
)
{
    TFX_INSTANCE *instData = TFX_PAGEINST(tfxPage);
    UInt32 *newImageCounts = NULL;
    PROLOGUE;

    if (inc == 0) {
	NORMAL_RETURN;
    }

    /* If decrementing, see if this collapses the number of layers required */
    if (inc < 0) {
	XERR_GEN_CLEANUP(tfxPage->pageInfo.dwLayersPresent < layer,
			 IO_INTERNAL_ERROR);
	if (layer == tfxPage->pageInfo.dwLayersPresent &&
	    tfxPage->pageInfo.pdwImageCounts[layer - 1] == (UInt32)-inc) {

	    IO_CHECKMALLOC(newImageCounts = (UInt32 *)TFX_MALLOC(sizeof(UInt32) *
							       layer - 1));
	    memset(newImageCounts, 0, sizeof(UInt32) * layer - 1);
	    memcpy(newImageCounts, tfxPage->pageInfo.pdwImageCounts,
		   sizeof(UInt32) * layer - 1);
	    TFX_FREE(tfxPage->pageInfo.pdwImageCounts);

	    tfxPage->pageInfo.pdwImageCounts = newImageCounts;
	    newImageCounts = NULL;
	    tfxPage->pageInfo.dwLayersPresent = layer - 1;
	} else {

	    /* Increment (decrement) the count for this layer */
	    tfxPage->pageInfo.pdwImageCounts[layer - 1] += inc;
	}

    } else {

	/* See if the array needs to be expanded */
	if (tfxPage->pageInfo.dwLayersPresent < layer) {

	    IO_CHECKMALLOC(newImageCounts = (UInt32 *)TFX_MALLOC(sizeof(UInt32) *
							       layer));
	    memset(newImageCounts, 0, sizeof(UInt32) * layer);
	    if (tfxPage->pageInfo.pdwImageCounts != NULL) {
		memcpy(newImageCounts, tfxPage->pageInfo.pdwImageCounts,
		       sizeof(UInt32) * tfxPage->pageInfo.dwLayersPresent);
		TFX_FREE(tfxPage->pageInfo.pdwImageCounts);
	    }
	    tfxPage->pageInfo.pdwImageCounts = newImageCounts;
	    newImageCounts = NULL;
	    tfxPage->pageInfo.dwLayersPresent = layer;
	}

	/* Increment the count for this layer */
	tfxPage->pageInfo.pdwImageCounts[layer - 1] += inc;
    }
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (newImageCounts != NULL) {
	TFX_FREE(newImageCounts);
    }
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image description and fill in the ImgInfo structure.
 *
 * LongDescription:::
 *  wp_TfxGetImageDesc() reads the IFD and fills in the ImgInfo structure with
 *  the basic image description as interpreted from the IFD.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxGetImageDesc(
    TFX_IMG *tfxImg
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_PAGE *tfxPage = NULL;
    TFX_DOC *tfxDoc = NULL;
    TFX_IMG *childImg = NULL;
    TFX_IMG *siblingImg = NULL, *nextImg;
    UInt32 nextIFD;
    UInt32 imgNum;
    UInt32 imgLayer = 0;
    UInt32 layer, imgOffset;
    Bool imgCountsUpdated = 0;
    PROLOGUE;
    
    /* We may have previously gotten the image info... */
    if ((tfxImg->seqStatus & TFX_IMG_SEQ_GETINFO) == 0) {
	
	tfxPage = tfxImg->tfxPage;
	tfxDoc = tfxPage->tfxDoc;
	instData = TFX_DOCINST(tfxDoc);
	tfxFile = TFX_DOCFILE(tfxDoc);
	
	/* Call the TIFF routine to read the tags */
	XERR_WARNOK_CLEANUP(wp_TifFXReadIFD(tfxImg));

	/* Check the Image Layer */
	if (tfxPage->pageInfo.dwLayersPresent == 0) {	/* No layers added, this is the first (Primary) */
	    if ((tfxImg->imgInfo.dwImgLayer % 2) == 1
		&& tfxImg->imgInfo.xColorType == IO_COLORTYPE_BINARY) {
		/* The primary image should be on layer 2, move it there */
		tfxImg->imgInfo.dwImgLayer = 2;
	    }

	    XERR_GEN_CLEANUP(tfxImg->imgInfo.dwImgLayer == 0 ||
		tfxImg->imgInfo.dwImgLayer > 3, IO_BADFILEFORMAT);
	}

	/* Check the layer against it's content (if not binary, it should be an odd layer) */
	XERR_GEN_CLEANUP(IO_BPP(tfxImg->imgInfo.xColorType) > 1
	    && (tfxImg->imgInfo.dwImgLayer % 2) != 1,
	    IO_BADFILEFORMAT);

	/* Update the layer count */
	if (tfxImg->imgInfo.tfxImgType != IO_IMGTYPE_HINTS) {
	    XERR_PASS_CLEANUP(wp_TfxIncrementImageCounts(tfxPage,
		tfxImg->imgInfo.dwImgLayer, 1));
	    imgCountsUpdated = 1;
	}
	imgLayer = tfxImg->imgInfo.dwImgLayer;

	/* Mark as gotten now, otherwise we'll appear to need it again during a
	 * child image call */
	tfxImg->seqStatus |= TFX_IMG_SEQ_GETINFO;
	tfxImg->imgInfo.tfxImgHandle = (UInt32)tfxImg;
	
	/* Retrieve information about the layout of the page */
	if (tfxImg->imgInfo.tfxImgType == IO_IMGTYPE_PRIMARY) {
	    UInt32 chain;

	    /* We should have the primary image as the first in the list */

	    /* Parse through the entire list of images to extract all
	     * the info we desire */
	    for (chain = 0; chain < tfxImg->fields.SubIFDCount; chain++) {
		XERR_GEN_CLEANUP(tfxImg->fields.SubIFDs == NULL, IO_INTERNAL_ERROR);
		nextIFD = tfxImg->fields.SubIFDs[chain];
		imgNum = 0;	/* Reset the image offset for this layer */
		while (nextIFD) {
		    
		    XERR_PASS_CLEANUP(wp_TfxNewImg(tfxPage, &childImg, nextIFD));

		    /* Get the image layer information from the sub-image */
		    childImg->imgInfo.tfxImgType = IO_IMGTYPE_LAYER;
		    XERR_IGNORE(__error__ = wp_TfxGetImageDesc(childImg));

		    if (__error__) {
			/* This image can't be handled, whether unsupported compression,
			 * or any other reason.
			 * Mention it to the client, then remove it from the list.
			 */
			/* IO_REPORT_PROBLEM(__error__); */

			/* Attempt to get the next image in the chain */
			XERR_IGNORE(__error__ = wp_TifFXGetNextIFD(childImg));
			if (__error__ == 0) {
			    nextIFD = childImg->fields.nextIFD;
			} else {
			    nextIFD = 0;
			}

			/* Get rid of this image */
			wp_TfxImgFree(&childImg, 0);

		    } else {

			/* Does the IFD order take precedence over the claimed image order???
			 * Frankly, the recommended order is:
			 *    1. Stripes in order (top-to-bottom, although this
			 *	is different when Orientation is not the default)
			 *    2. Layers within the stripe are in T.44 order: 2, 1, 3, 4, 5, ...
			 */
			if (childImg->imgInfo.dwImgLayer == 0) {
			    /* If no designation, assume that chain it was found on is correct
			     * (which is pretty silly, since that's more sophisticated than
			     * simply attaching the right layer number to the image, but hey,
			     * it's better than nothing). */
			    if (chain > 0) {
				/* Chains are for layers 1, 3, 4, 5, ... */
				childImg->imgInfo.dwImgLayer = chain + 2;
			    } else {
				childImg->imgInfo.dwImgLayer = chain + 1;
			    }
			} else if (childImg->imgInfo.dwImgLayer <= tfxPage->pageInfo.dwLayersPresent
			    && tfxPage->tfxImgs != NULL) {
			    /* If it's within the layers that already exist, make sure it doesn't
			     * coexist with another image (if so, bump this one to the end */

			    /* Search through the existing list for other image on the same layer */
			    {
				Bool duplicateFound = 0;
				UInt32 lastInLayer = 0;

				siblingImg = tfxPage->tfxImgs;
				while (siblingImg) {
				    if (siblingImg->imgInfo.dwImgLayer == childImg->imgInfo.dwImgLayer) {
					if (siblingImg->imgInfo.dwImgOffset > lastInLayer) {
					    lastInLayer = siblingImg->imgInfo.dwImgOffset;
					}
					if (siblingImg->imgInfo.dwImgOffset == childImg->imgInfo.dwImgOffset) {
					    duplicateFound = 1;
					    /* Keep searching for the last in this layer,
					     * then assign this one to the end */
					}
				    } else if (siblingImg->imgInfo.dwImgLayer > childImg->imgInfo.dwImgLayer) {
					break;
				    }
				    siblingImg = siblingImg->next;
				}

				/* If a duplicate found, assign this one to the end of the list */
				if (duplicateFound) {
				    childImg->imgInfo.dwImgOffset = lastInLayer + 1;
				}
			    }
			}

			/* Search for the appropriate place to insert this image */
			siblingImg = tfxPage->tfxImgs;
			layer = childImg->imgInfo.dwImgLayer;
			imgOffset = childImg->imgInfo.dwImgOffset;

			/* If this is the first, just insert it */
			if (siblingImg == NULL) {
			    tfxPage->tfxImgs = childImg;
			} else if (siblingImg->imgInfo.dwImgLayer > layer ||
			    (siblingImg->imgInfo.dwImgLayer == layer &&
			    siblingImg->imgInfo.dwImgOffset > (UInt32)imgOffset)) {

			    /* This one should really be first. Insert as such. */
			    childImg->next = siblingImg;
			    siblingImg->prev = childImg;
			    tfxPage->tfxImgs = childImg;
			} else {

			    /* Search for the proper point to insert *after* */
			    while (((nextImg = siblingImg->next) != NULL) &&
				(nextImg->imgInfo.dwImgLayer < layer ||
				(nextImg->imgInfo.dwImgLayer == layer &&
				nextImg->imgInfo.dwImgOffset < (UInt32)imgOffset))) {
				siblingImg = nextImg;
			    }

			    /* Link this image to the one that was next */
			    childImg->next = nextImg;
			    if (nextImg) {
				nextImg->prev = childImg;
			    }

			    /* Link prev image to this one */
			    siblingImg->next = childImg;
			    childImg->prev = siblingImg;
			}

/*			XERR_GEN_CLEANUP(childImg->imgInfo.dwImgLayer != chain + 1, IO_BADFILEFORMAT); */
/*			XERR_GEN_CLEANUP(childImg->imgInfo.dwImgOffset != imgNum, IO_BADFILEFORMAT); */
/*			childImg->imgInfo.dwImgOffset = imgNum; */
			imgNum++;
		    
			XERR_IGNORE(wp_TifFXGetNextIFD(childImg));

			nextIFD = childImg->fields.nextIFD;
		    }
		} /* For each IFD in chain */
	    } /* For each chain */

	    /* Fix the order, if the file isn't quite right */
	    {
		UInt32 layer = 0, offset = 0;
		Bool newLayer = 1;
		TFX_IMG *thisImg, *primImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

		thisImg = tfxPage->tfxImgs;
		while (thisImg) {
		    XERR_GEN_CLEANUP(thisImg->imgInfo.dwImgLayer < 1,
			IO_BADFILEFORMAT);
		    if (thisImg->imgInfo.dwImgLayer != layer) {
			newLayer = 1;
		    }
		    if (newLayer) {
			layer = thisImg->imgInfo.dwImgLayer;
			if (layer == primImg->imgInfo.dwImgLayer) {
			    offset = 1;	/* The primary image is kept separate */
			} else {
			    offset = 0;
			}
			newLayer = 0;
		    } else {
			offset++;
		    }

		    /* Make sure the offsets are sequential */
		    if (thisImg->imgInfo.dwImgOffset != offset) {
/*			XERR_GEN_CLEANUP(tfxImg->imgInfo.dwImgOffset != offset, */
/*			    IO_BADFILEFORMAT); */

			/* Fix the problem */
			thisImg->imgInfo.dwImgOffset = offset;
		    }

		    /* Go to the next image */
		    thisImg = thisImg->next;
		}
	    }
	} /* Get page layout */

	/* Check for a correct "best strip height" value */
	if (tfxImg->imgInfo.dwBestStripHeight > tfxImg->imgInfo.dwHeight) {
	    tfxImg->imgInfo.dwBestStripHeight = tfxImg->imgInfo.dwHeight;
	}
	
    } /* Info not already gotten */
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:

    /* Decrement the image layer count, if the it was incremented */
    if (imgCountsUpdated) {
	XERR_IGNORE(wp_TfxIncrementImageCounts(tfxPage,
	    imgLayer, -1));
    }
    
    ERR_RETURN;
}


/*
$Log: tfxrpriv.c,v $
Revision 1.2  2003/02/10 22:53:31  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:11  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.14  2001/12/26 21:23:45  dabercro
Mark as released

Revision 1.13  2001/10/30 20:04:26  dabercro
Bug fix: don't increment image counts when examining a hints image.

Revision 1.12  2001/09/18 21:10:11  dabercro
Make work with JBIG2

Revision 1.11  2001/01/19 23:53:57  dabercro
Return NULL on failure.

Revision 1.10  2001/01/18 00:21:31  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.9  2000/11/18 00:53:26  dabercro
Avoid cross polenation of data handles across sublibraries
(XIF, MRC, TFX), especially for CopyImage.  Added a handle
type to structure, now checked also.

Revision 1.8  2000/10/18 20:33:34  dabercro
Use SHOWERR #define to isolate handling of the source
of memory allocations.  Define in Makefile or MSVC project if
desired.

Revision 1.7  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.6  2000/08/22 01:03:23  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.5  2000/06/07 00:53:30  dabercro
Initialize variable.

Revision 1.4  2000/06/07 00:00:17  dabercro
Continued work, lots of stuff.

Revision 1.3  2000/04/12 19:06:09  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.2  2000/03/11 02:06:35  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.13  1998/07/27 17:36:23  aber
tfxrpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxrpriv.c,v

Use private memory tally structure.

Revision 1.12  1998/02/09 20:21:26  aber
tfxrpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxrpriv.c,v

CommonCheck: Signed/unsigned casting to make MSVC happy.

Revision 1.11  1997/12/18 00:11:13  aber
tfxrpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxrpriv.c,v

Allow negative increments in image count, handle shrinking image counts
in MixedRasterContent field.

Revision 1.10  1997/08/29 17:28:02  aber
tfxrpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxrpriv.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.9  1997/07/10 21:52:30  aber
tfxrpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxrpriv.c,v

Use size_t as the memory allocation size (instead of UInt32).

Revision 1.8  1997/07/02 01:05:15  aber
tfxrpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxrpriv.c,v

Convert memory functions to IPWrap functions.

Revision 1.7  1997/06/20 17:46:19  aber
tfxrpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxrpriv.c,v

Test for sub-images before referencing the list, NULL-pointer reference
otherwise.

Revision 1.6  1997/06/10 01:26:42  aber
tfxrpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxrpriv.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:03  aber
tfxrpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxrpriv.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:07  aber
tfxrpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxrpriv.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:56:25  aber
aber on Mon May 19 17:55:49 PDT 1997

tfxrpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxrpriv.c,v

Fixed CDECL in prototype, and potentially unused variables.

Revision 1.1  1997/05/17 00:46:38  aber
Initial revision

*/
