/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxrtfx.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxrtfx.c:	READ TIFF-FX source
 *
 *	TIFF-FX read-only routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tfxrtfx.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:31 $");

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "geomadt.h"
#include "props.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif

/* The idea is to remove this one day: */
#include "tfx.pub"

#endif	/* NOIPCORE */


#include "tfximage.h"
#include "tfxrtiff.h"
#include "tfxrtfx.h"
#include "tfxrpriv.h"
#include "tfxconv.h"

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free the internal data associated with an image
 *
 * LongDescription:::
 *  wp_TfxImgFree() frees the data associated with an image handle, and
 *  optionally the subsequent images also. Since the images belonging to
 *  a page are stored internally in a linked list, any image freed could
 *  either have it's link removed, or all links down the chain.
 *
 * Input:::
 *  pTftfxImg::			Data structure of valid image
 *  subsequentImagesToo::	Boolean for: free connected images also
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxImgFree(
    TFX_IMG **pTftfxImg,
    Int32 subsequentImagesToo
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_IMG *tfxImg, *nextImg = NULL;
    TFX_MISCMEM *mmptr, *next_mm;
    TFX_STRIP *strip, *nextStrip;
    PROLOGUE;

    TFX_CheckHandle(*pTftfxImg);

    tfxImg = *pTftfxImg;
    instData = TFX_IMGINST(tfxImg);

    while (tfxImg) {
	nextImg = tfxImg->next;

	if (tfxImg->xHintsImgs) {
	    XERR_PASS_CLEANUP(wp_TfxImgFree(&tfxImg->xHintsImgs, 1));
	}

	XERR_PASS_CLEANUP(wp_TifFXFieldCleanup(tfxImg, &tfxImg->fields));
	mmptr = tfxImg->tmpCmpData;
	while (mmptr) {
	    if (mmptr->mem) {
		TFX_FREE(mmptr->mem);
	    }
	    next_mm = mmptr->next;
	    TFX_FREE(mmptr);
	    mmptr = next_mm;
	}
	strip = tfxImg->strips.firstStrip;
	while (strip) {
	    nextStrip = strip->next;
	    TFX_FREE(strip);
	    strip = nextStrip;
	}
	if (tfxImg->cmpBuffer) {
	    TFX_FREE(tfxImg->cmpBuffer);
	    tfxImg->cmpBuffer = NULL, tfxImg->cmpBytes = 0UL;
	}
#ifndef NOIPCORE
	if (tfxImg->tmpPixr) {
	    IPWState *pIPWState;
	    
	    /* Reference the IPWrap state */
	    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

	    ERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &tfxImg->tmpPixr));
	}
#endif
	if (tfxImg->dictList) {
	    TFX_FREE(tfxImg->dictList);
	}
	if (tfxImg->dictSizes) {
	    TFX_FREE(tfxImg->dictSizes);
	}
	if (tfxImg->qtableBuf) {
	    TFX_FREE(tfxImg->qtableBuf);
	}

	TFX_FREE(tfxImg);

	if (! subsequentImagesToo) {
	    break;
	}

	tfxImg = nextImg;
    }

    *pTftfxImg = nextImg;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free the internal data associated with a page
 *
 * LongDescription:::
 *   wp_TfxPageFree() frees the data associated with an page handle, and
 *  optionally the subsequent pages also. Since the pages belonging to
 *  a document are stored internally in a linked list, any page freed could
 *  either have it's link removed, or all links down the chain. All images
 *  tied to the page(s) are also freed.
 *
 * Input:::
 *  pTfxPage::			Pointer to data structure of valid page
 *  subsequentPagesToo::	Boolean for: free other pages along chain
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxPageFree(
    TFX_PAGE **pTfxPage,
    Int32 subsequentPagesToo
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_PAGE *tfxPage, *nextPage = NULL;
    TFX_ANNOTTABLE *at, *nextAt;
    PROLOGUE;

    TFX_CheckHandle(*pTfxPage);

    tfxPage = *pTfxPage;
    instData = TFX_PAGEINST(tfxPage);

    while (tfxPage) {
	nextPage = tfxPage->next;

	/* Free the images associated with this page */
	if (tfxPage->pageInfo.tfxPrimImgHandle) {
	    XERR_PASS_CLEANUP(wp_TfxImgFree((TFX_IMG **)
					 &tfxPage->pageInfo.tfxPrimImgHandle, 1));
	}
	if (tfxPage->pageInfo.pdwImageCounts) {
	  TFX_FREE(tfxPage->pageInfo.pdwImageCounts);
	}
	if (tfxPage->extImageCounts) {
	  TFX_FREE(tfxPage->extImageCounts);
	}
	if (tfxPage->tfxImgs) {
	    XERR_PASS_CLEANUP(wp_TfxImgFree(&tfxPage->tfxImgs, 1));
	}
	at = tfxPage->firstAnnotTable;
	while (at) {
	    nextAt = at->next;
	    if (at->entries) {
		TFX_FREE(at->entries);
	    }
	    TFX_FREE(at);
	    at = nextAt;
	}

	/* Free the page itself */
	TFX_FREE(tfxPage);

	if (! subsequentPagesToo) {
	    break;
	}

	tfxPage = nextPage;
    }

    *pTfxPage = nextPage;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free all data associated with a document handle
 *
 * LongDescription:::
 *  wp_TfxDocFree() free all data associated with a document handle, including
 *  all page and image data.
 *
 * Input:::
 *  pTfxDoc::			Pointer to data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxDocFree(
    TFX_DOC **pTfxDoc
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_DOC *tfxDoc;	
/*    TFX_PAGETABLE *pt, *nextPt; */
    TFX_IFDOFFSET_ITEM *ifdp, *nextIFDp;
    TFX_RESOURCETABLE *rt, *nextRt;
    TFX_COPIEDDICTS *pd, *nextPd;
    PROLOGUE;

    TFX_CheckHandle(*pTfxDoc);

    tfxDoc = *pTfxDoc;
    instData = TFX_DOCINST(tfxDoc);

    if (tfxDoc->pages) {
	XERR_PASS_CLEANUP(wp_TfxPageFree(&tfxDoc->pages, 1));
    }

    if (tfxDoc->tfxFile) {
	TFX_FREE(tfxDoc->tfxFile);
    }
    if (tfxDoc->docInfo.pAuthorBlock) {
	TFX_FREE(tfxDoc->docInfo.pAuthorBlock);
    }
    ifdp = tfxDoc->ifdList;
    while (ifdp) {
	nextIFDp = ifdp->next;
	TFX_FREE(ifdp);
	ifdp = nextIFDp;
    }
#if 0
    pt = tfxDoc->firstPageTable;
    while (pt) {
	nextPt = pt->next;
	if (pt->entries) {
	    TFX_FREE(pt->entries);
	}
	TFX_FREE(pt);
	pt = nextPt;
    }
#endif

    rt = tfxDoc->firstResourceTable;
    while (rt) {
	nextRt = rt->next;
	if (rt->entries) {
	    TFX_FREE(rt->entries);
	}
	TFX_FREE(rt);
	rt = nextRt;
    }

    pd = tfxDoc->prevTDCopied;
    while (pd) {
	nextPd = pd->next;
	TFX_FREE(pd);
	pd = nextPd;
    }

    if (tfxDoc->curDictIDs) {
	TFX_FREE(tfxDoc->curDictIDs);
    }

    TFX_FREE(tfxDoc);

    *pTfxDoc = NULL;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Check to see if the file is a valid TIFF-FX, and get the version number
 *
 * LongDescription:::
 *  wp_TfxIsTfx() checks for the necessary header information to identify the
 *  file as a TIFF-FX file. The version and revision are returned if it has
 *  the TIFF-FX features.
 *
 * Input:::
 *  tfxFile::			File handle structure
 *
 * Output:::
 *  pVersion::			Version number
 *  pRevision::			Revision number
 *  pFirstIFD::			Offset of first IFD
 *
 * Returns:::
 *  IO_NOTTIFF::		File is not even a TIF
 *  TFX_NOTTIFFFX::		File is not ag TIFF-FX
 *  IO_NOSUPPORT::		File appears to be a unsupported version of TIFF-FX
 *  IO_NOERROR::		File is a supported TIFF-FX
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxIsTfx(
	TFX_FILE *tfxFile,
	UInt8 *pVersion,
	UInt8 *pRevision,
	UInt32 *pFirstIFD
)
{
    PROLOGUE;

    /* Return nothing until something found */
    *pVersion = 0UL;
    *pRevision = 0UL;

    /* Wait until I have full control over the file */
    XERR_PASS_CLEANUP(wp_TfxFileLock(TFX_FILETOKEN(tfxFile)));

    /* Check the TIFF header */
    if ((__error__ = wp_TifFXIsTiff(tfxFile, pFirstIFD)) != IO_NOERROR) {
	return __error__;
    }

    /* This should be the case for wp_TifFXIsTiff to pass */
    *pVersion = 7;
    *pRevision = 0;

    NORMAL_RETURN;

 ERR_CLEANUP:


    wp_TfxFileUnLock(TFX_FILETOKEN(tfxFile));
    NORMAL_RETURN;
}


#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read the data associated with the TIFF-FX fields
 *
 * LongDescription:::
 *  wp_TfxReadTagData() reads the data associated with the TIFF-FX fields. The
 *  location is that derived from the previously read fields.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxReadTagData(
    TFX_DOC *tfxDoc
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt16 sval;
    PROLOGUE;

    TFX_CheckHandle(tfxDoc);
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);

    /* Check for proper calling sequence */


#if 0
    /* Read the page table header */
    if (tfxDoc->pageTableHeaderOffset != 0UL) {
	TFX_SEEK(tfxFile, tfxDoc->pageTableHeaderOffset);
	TFX_READ(tfxFile, &tfxDoc->pageTableHeader.pageCount,
	    sizeof(tfxDoc->pageTableHeader.pageCount));
	TFX_SWAPSHORT(tfxFile, &tfxDoc->pageTableHeader.pageCount);
	TFX_READ(tfxFile, &tfxDoc->pageTableHeader.entriesPerTable,
	    sizeof(tfxDoc->pageTableHeader.entriesPerTable));
	TFX_SWAPSHORT(tfxFile, &tfxDoc->pageTableHeader.entriesPerTable);
	TFX_READ(tfxFile, &tfxDoc->pageTableHeader.lastOrdinal,
	    sizeof(tfxDoc->pageTableHeader.lastOrdinal));
	TFX_SWAPSHORT(tfxFile, &tfxDoc->pageTableHeader.lastOrdinal);
	TFX_READ(tfxFile, &tfxDoc->pageTableHeader.firstTableOffset,
	    sizeof(tfxDoc->pageTableHeader.firstTableOffset));
	TFX_SWAPLONG(tfxFile, &tfxDoc->pageTableHeader.firstTableOffset);

	/* Read the page table, handled separately, due to multiple table entries */
	if (tfxDoc->pageTableHeader.pageCount > 0) {
	    XERR_PASS_CLEANUP(wp_TfxReadPageTable(tfxDoc, 0UL, NULL));
	}

	/* Fill in the page count */
	tfxDoc->docInfo.wNPages = tfxDoc->pageTableHeader.pageCount;
    } else {
	tfxDoc->docInfo.wNPages = 0;
    }
#else
    tfxDoc->docInfo.wNPages = 0;
#endif

    /* Read the author block */
    if (tfxDoc->authorOffset) {
	TFX_SEEK(tfxFile, tfxDoc->authorOffset);
	TFX_READ(tfxFile, &sval, sizeof(sval));
	TFX_SWAPSHORT(tfxFile, &sval);
	tfxDoc->docInfo.wAuthorBlockBytes = sval;
	if (tfxDoc->docInfo.wAuthorBlockBytes > 0) {
	    IO_CHECKMALLOC(tfxDoc->docInfo.pAuthorBlock = (Int8 *)
			 TFX_MALLOC(tfxDoc->docInfo.wAuthorBlockBytes));
	    TFX_READ(tfxFile, tfxDoc->docInfo.pAuthorBlock,
		     (Int32)tfxDoc->docInfo.wAuthorBlockBytes);
	} else {
	    tfxDoc->docInfo.pAuthorBlock = NULL;
	}
    }

    /* Read the reference page offset */
    if (tfxDoc->docInfo.dwReferencePageOffset) {
	TFX_SEEK(tfxFile, tfxDoc->refPageOffset);
	TFX_READ(tfxFile, &tfxDoc->docInfo.dwReferencePageOffset,
		 sizeof(tfxDoc->docInfo.dwReferencePageOffset));
	TFX_SWAPLONG(tfxFile, &tfxDoc->docInfo.dwReferencePageOffset);

	/* The stored value is actually the ordinal, convert
	 * to offset */
	tfxDoc->docInfo.dwReferencePageOffset--;
    }

    /* Read the resource table header */
    if (tfxDoc->resourceTableHeaderOffset != 0UL) {
	TFX_SEEK(tfxFile, tfxDoc->resourceTableHeaderOffset);
	TFX_READ(tfxFile, &tfxDoc->resourceTableHeader.resourceCount,
	    sizeof(tfxDoc->resourceTableHeader.resourceCount));
	TFX_SWAPSHORT(tfxFile, &tfxDoc->resourceTableHeader.resourceCount);
	TFX_READ(tfxFile, &tfxDoc->resourceTableHeader.entriesPerTable,
	    sizeof(tfxDoc->resourceTableHeader.entriesPerTable));
	TFX_SWAPSHORT(tfxFile, &tfxDoc->resourceTableHeader.entriesPerTable);
	TFX_READ(tfxFile, &tfxDoc->resourceTableHeader.lastOrdinal,
	    sizeof(tfxDoc->resourceTableHeader.lastOrdinal));
	TFX_SWAPSHORT(tfxFile, &tfxDoc->resourceTableHeader.lastOrdinal);
	TFX_READ(tfxFile, &tfxDoc->resourceTableHeader.firstTableOffset,
	    sizeof(tfxDoc->resourceTableHeader.firstTableOffset));
	TFX_SWAPLONG(tfxFile, &tfxDoc->resourceTableHeader.firstTableOffset);

	/* Read the shared resource header, handled separately */
	XERR_PASS_CLEANUP(wp_TfxReadResourceTable(tfxDoc, 0UL, NULL));
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read all TIFF-FX fields
 *
 * LongDescription:::
 *  wp_TfxReadExtensions() reads all TIFF-FX fields stored in the file. These
 *  fields may be store in one or more list.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxReadExtensions(
    TFX_DOC *tfxDoc
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    PROLOGUE;

    TFX_CheckHandle(tfxDoc);
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);

    /* Check for proper calling sequence */

    /* Wait until I have full control over the file */
    XERR_PASS_CLEANUP(wp_TfxFileLock(TFX_DOCFILETOKEN(tfxDoc)));

    /* Seek to base extension location */
    TFX_SEEK(tfxFile, 8);

    /* Recursively read the extension list */
    XERR_PASS_CLEANUP(wp_TfxReadExtensionList(tfxDoc, 0, 8));

    NORMAL_RETURN;

 ERR_CLEANUP:

    wp_TfxFileUnLock(TFX_DOCFILETOKEN(tfxDoc));
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a single page table
 *
 * LongDescription:::
 *  wp_TfxReadPageTable() reads a single page table, and inserts it into a
 *  linked list of page tables.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  tableOffset::		Offset of page table to read
 *
 * Output:::
 *  ppPageTable::		If NULL, the table is read and nothing
 *				returned. Otherwise the pointer will point to
 *				the resulting table.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxReadPageTable(
    TFX_DOC *tfxDoc,
    UInt32 tableOffset,
    TFX_PAGETABLE **ppPageTable
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    TFX_PAGETABLE *pt, *newpt = NULL;		/* Linked page table */
    TFX_PAGETABLEENTRY *ep;
    UInt32 mallocbytes;
    UInt32 offset;
    UInt16 entriesLeft;
    PROLOGUE;

    /* Default to return nothing */
    if (ppPageTable) {
	*ppPageTable = NULL;
    }

    TFX_CheckHandle(tfxDoc);
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);
    pt = tfxDoc->firstPageTable;
    offset = tfxDoc->pageTableHeader.firstTableOffset;

    /* Check for proper calling sequence */


    /* If initial page table is not in place, we cannot link
     * another */
    XERR_GEN_CLEANUP((tableOffset > 0 && pt == NULL), IO_SEQUENCE);

    /* If we're installing the first, don't look for cached tables */
    if (pt) {

	/* Since the page tables are always read in order, and cached
	 * when read, go through the cached list until we come to the
	 * last one cached */
	while (pt->ordinal <= tableOffset) {
	    /* There is no such page table */
	    XERR_GEN_CLEANUP(((offset = pt->nextTableOffset) == 0),
		     IO_BADPARAMETER);
	    if (pt->next == NULL) {
		/* Table exists, but hasn't been read yet */
		break;
	    }
	    pt = pt->next;
	}

	/* If we're at the right page table, and it's already read, bail now */
	if (pt->ordinal == tableOffset + 1) {
	    goto found;
	}
    }

    /* Read as many page tables as necessary to retrieve the one requested */
    do {
	/* Seek to the next page table offset.
	 * If there aren't the number of page tables that the page
	 * table header claims, the file is corrupt */
	XERR_GEN_CLEANUP((! offset), IO_BADFILEFORMAT);
	TFX_SEEK(tfxFile, offset);

	IO_CHECKMALLOC(newpt = (TFX_PAGETABLE *)TFX_MALLOC(sizeof(TFX_PAGETABLE)));
	memset(newpt, 0, sizeof(TFX_PAGETABLE));
	mallocbytes = sizeof(TFX_PAGETABLEENTRY) * 
	    tfxDoc->pageTableHeader.entriesPerTable;
	IO_CHECKMALLOC(newpt->entries =
		     (TFX_PAGETABLEENTRY *)TFX_MALLOC(mallocbytes));
	memset(newpt->entries, 0, mallocbytes);

	if (pt) {
	    /* Attach to end of linked list */
	    pt->next = newpt;
	    pt = newpt;
	} else {
	    /* Begin list here */
	    pt = tfxDoc->firstPageTable = newpt;
	}
	newpt = NULL;	/* Avoid freeing if later failure */

	/* Read the table */
	TFX_READ(tfxFile, &pt->ordinal, sizeof(pt->ordinal));
	TFX_SWAPSHORT(tfxFile, &pt->ordinal);
	ep = pt->entries;
	entriesLeft = tfxDoc->pageTableHeader.entriesPerTable;
	while (entriesLeft--) {
	    TFX_READ(tfxFile, &ep->flags, sizeof(ep->flags));
	    TFX_READ(tfxFile, &ep->IFDoffset, sizeof(ep->IFDoffset));
	    TFX_SWAPLONG(tfxFile, &ep->IFDoffset);
	    ep++;
	}

	/* Read the next page table offset */
	TFX_READ(tfxFile, &pt->nextTableOffset, sizeof(pt->nextTableOffset));
	TFX_SWAPLONG(tfxFile, &pt->nextTableOffset);
	offset = pt->nextTableOffset;

    } while (pt->ordinal != tableOffset + 1);

 found:

    /* Return the one found */
    if (ppPageTable) {
	*ppPageTable = pt;
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    if (newpt) {
	TFX_FREE(newpt);
	if (newpt->entries) {
	    TFX_FREE(newpt->entries);
	}
    }

    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the page IFD offset
 *
 * LongDescription:::
 *  wp_TfxGetPageIFD() returns the page IFD offset, reading the appropriate
 *  page table, if necessary.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  pageOffset::		Offset of the desired page (0, 1, ...)
 *
 * Output:::
 *  pFlags::			Pointer to the flags stored for the page
 *				(see xxxxxxxxxxxxxxxxxx)
 *  pIfdOffset::		Pointer to the IFD offset
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxGetPageIFD(
    TFX_DOC *tfxDoc,
    UInt32 pageOffset,
    UInt8 *pFlags,
    Int32 *pIfdOffset
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    TFX_IFDOFFSET_ITEM *ifdp;
    UInt32 offset;
/*    UInt32 tableOffset; */
/*    UInt32 entryOffset; */
/*    TFX_PAGETABLE *pageTable; */
    PROLOGUE;

    /* Default to no return value */
    *pFlags = 0;
    *pIfdOffset = 0UL;

    TFX_CheckHandle(tfxDoc);
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((tfxDoc->seqStatus & TFX_DOC_SEQ_GETINFO) == 0,
	IO_SEQUENCE);

#if 0
    /* Read the appropriate page table */
    if (tfxDoc->pageTableHeader.entriesPerTable == 0) {
	tfxDoc->pageTableHeader.entriesPerTable = 32;
    }
    tableOffset = pageOffset / tfxDoc->pageTableHeader.entriesPerTable;
    entryOffset = pageOffset % tfxDoc->pageTableHeader.entriesPerTable;
    XERR_PASS_CLEANUP(wp_TfxReadPageTable(tfxDoc, tableOffset, &pageTable));
#endif

    ifdp = tfxDoc->ifdList;
    offset = 0;
    while (offset < pageOffset && ifdp != NULL) {
	if (ifdp->next == NULL) {
	    /* Make sure we've read that far down the IFD chain */
	    XERR_PASS_CLEANUP(wp_TfxTraversePrimaryImages(tfxDoc, pageOffset + 1));
	}
	ifdp = ifdp->next;
	offset++;
    }

    /* If there's no list item, then there's probably no pages (which is okay) */
    if (ifdp != NULL) {

	XERR_GEN_CLEANUP(offset != pageOffset, IO_BADPARAMETER);

        /* Return the IFD offset for the requested page */
/*    *pFlags = pageTable->entries[entryOffset].flags; */
/*    *pIfdOffset = pageTable->entries[entryOffset].IFDoffset; */
        *pIfdOffset = ifdp->ifdOffset;
    } else {

	/* If there was a list, and we ran off the end of it, then we probably
	 * haven't kept track of the IFD list very well */
	XERR_GEN_CLEANUP(tfxDoc->ifdList, IO_INTERNAL_ERROR);
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a single shared resource table
 *   
 *
 * LongDescription:::
 *  wp_TfxReadResourceTable() reads a single shared resource table, and inserts
 *  it into a linked list of resource tables.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  resourceOffset::		Resource offset (ID)
 *
 * Output:::
 *  ppResourceTable::		If NULL, the table is read and nothing
 *				returned. Otherwise the pointer will point to
 *				the resulting table.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxReadResourceTable(
    TFX_DOC *tfxDoc,
    UInt32 resourceOffset,
    TFX_RESOURCETABLE **ppResourceTable
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    TFX_RESOURCETABLE *rt, *newRt = NULL;		/* Linked resource table */
    TFX_RESOURCETABLEENTRY *ep;
    UInt32 mallocbytes;
    Int32 offset;
    UInt16 entriesLeft;
    PROLOGUE;

    /* Default to return nothing */
    if (ppResourceTable) {
	*ppResourceTable = NULL;
    }

    TFX_CheckHandle(tfxDoc);
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);
    rt = tfxDoc->firstResourceTable;
    offset = tfxDoc->firstResourceTableLocation;

    /* Check for proper calling sequence */


    /* If initial resource table is not known, we have nothing to start from */
    XERR_GEN_CLEANUP(offset == 0, IO_SEQUENCE);

    /* If we've read any of the tables, look to see if it's already cached */
    if (rt) {

	/* Since the resource tables are always read in order, and cached
	 * when read, go through the cached list until we come to the
	 * last one cached */
	while ((Int32)resourceOffset >= rt->firstID + rt->physicalEntries) {

	    /* There is no such resource table */
	    XERR_GEN_CLEANUP(((offset = rt->nextTableOffset) == 0),
		     IO_BADFILEFORMAT);
	    if (rt->next == NULL) {
		/* Table exists but hasn't been read yet since 'offset != 0' */
		break;
	    }
	    rt = rt->next;
	}

	/* If we're at the right resource table, and it's already read,
	 * bail now */
	if (rt->firstID <= (Int32)resourceOffset &&
	    (Int32)resourceOffset < rt->firstID + rt->physicalEntries) {
	    goto found;
	}
    }

    /* Read as many resource tables as necessary to retrieve the one requested */
    do {
	/* Seek to the next resource table offset. */
	XERR_GEN_CLEANUP((! offset), IO_BADFILEFORMAT);
	TFX_SEEK(tfxFile, offset);

	IO_CHECKMALLOC(newRt = (TFX_RESOURCETABLE *)TFX_MALLOC(sizeof(TFX_RESOURCETABLE)));
	memset(newRt, 0, sizeof(TFX_RESOURCETABLE));

	if (rt) {

	    /* Update the first ID of the new list */
	    newRt->firstID = rt->firstID + rt->physicalEntries;

	    /* Attach to end of linked list */
	    rt->next = newRt;
	    rt = newRt;
	} else {
	    /* Begin list here */
	    newRt->firstID = 0;
	    rt = tfxDoc->firstResourceTable = newRt;
	}
	newRt = NULL;	/* Avoid freeing if later failure */

	/* Read the first of the table */
	TFX_READ(tfxFile, &rt->physicalEntries, sizeof(rt->physicalEntries));
	TFX_SWAPSHORT(tfxFile, &rt->physicalEntries);

	mallocbytes = sizeof(TFX_RESOURCETABLEENTRY) * rt->physicalEntries;
	IO_CHECKMALLOC(rt->entries =
		     (TFX_RESOURCETABLEENTRY *)TFX_MALLOC(mallocbytes));
	memset(rt->entries, 0, mallocbytes);

	/* Read the entries */
	ep = rt->entries;
	entriesLeft = rt->physicalEntries;
	while (entriesLeft--) {
	    TFX_READ(tfxFile, &ep->bytes, sizeof(ep->bytes));
	    TFX_SWAPLONG(tfxFile, &ep->bytes);
	    TFX_READ(tfxFile, &ep->offset, sizeof(ep->offset));
	    TFX_SWAPLONG(tfxFile, &ep->offset);
	    TFX_READ(tfxFile, &ep->resourceType, sizeof(ep->resourceType));
	    TFX_SWAPSHORT(tfxFile, &ep->resourceType);
	    TFX_READ(tfxFile, &ep->lastPage, sizeof(ep->lastPage));
	    TFX_SWAPSHORT(tfxFile, &ep->lastPage);
	    TFX_READ(tfxFile, &ep->decMem, sizeof(ep->decMem));
	    TFX_SWAPLONG(tfxFile, &ep->decMem);
	    ep++;
	}

	/* Read the next resource table offset */
	TFX_READ(tfxFile, &rt->nextTableOffset, sizeof(rt->nextTableOffset));
	TFX_SWAPLONG(tfxFile, &rt->nextTableOffset);
	offset = rt->nextTableOffset;

    } while (offset && rt->firstID + rt->physicalEntries <=
	(Int32)resourceOffset);

 found:

    /* Return the one found */
    if (ppResourceTable) {
	*ppResourceTable = rt;
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    if (newRt) {
	TFX_FREE(newRt);
	if (newRt->entries) {
	    TFX_FREE(newRt->entries);
	}
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the size and offset of a token dictionary
 *
 * LongDescription:::
 *  wp_TfxGetResourceOffset() returns the size and file offset of a shared
 *  resource found within the shared resource set.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  dictID::			Unique ID of the token dictionary
 *
 * Output:::
 *  pdwOffset::			Pointer to the data offset
 *  pdwBytes::			Pointer to the number of token dictionary bytes
 *  pwResourceType::		Type of resource, see TFX_RESOURCE... in tfxtiff.h
 *  pwLastPageUsed::		Pointer to the last page used (ordinal: 1, 2,
 *				...), 0 means unknown
 *  pdwDecoderMemory::		Pointer to the decoder memory required to hold
 *				the dictionary, 0 means unknown
 *
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxGetResourceOffset(
    TFX_DOC *tfxDoc,
    UInt32 dictID,
    Int32 *pdwOffset,
    Int32 *pdwBytes,
    UInt16 *pwResourceType,
    UInt16 *pwLastPageUsed,
    UInt32 *pdwDecoderMemory
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt32 entryOffset;
    TFX_RESOURCETABLE *resourceTable;
    PROLOGUE;

    /* Default to no return value */
    *pdwOffset = 0L;
    *pdwBytes = 0L;
    *pwResourceType = 0;
    *pwLastPageUsed = 0;
    *pdwDecoderMemory = 0UL;

    TFX_CheckHandle(tfxDoc);
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);

    /* Check for proper calling sequence */

    /* If we've not read any resources, make sure that we've found the appropriate
     * field in the GlobalParametersIFD */
    if (tfxDoc->firstResourceTable == NULL) {

	/* There should have been a SHAREDRESOURCES field read */
	XERR_GEN_CLEANUP(tfxDoc->sharedResourcesLocation == 0UL, IO_BADFILEFORMAT);
    }

    /* Read the appropriate table */
    XERR_PASS_CLEANUP(wp_TfxReadResourceTable(tfxDoc, dictID, &resourceTable));

    /* Determine the entry offset */
    entryOffset = dictID - resourceTable->firstID;

    XERR_GEN_CLEANUP(entryOffset > resourceTable->physicalEntries,
	IO_INTERNAL_ERROR);

    /* Return the offset and size for the requested resource */
    *pdwOffset = resourceTable->entries[entryOffset].offset;
    *pdwBytes = resourceTable->entries[entryOffset].bytes;
    *pwResourceType = resourceTable->entries[entryOffset].resourceType;
    *pwLastPageUsed = resourceTable->entries[entryOffset].lastPage;
    *pdwDecoderMemory = resourceTable->entries[entryOffset].decMem;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read one strip of image data, leave compressed
 *
 * LongDescription:::
 *  wp_TfxCompressedReadStrip() reads a single strip of image data from a TIFF
 *  (TIFF-FX), returning the compressed data.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  dwImgRowOffset::		Row offset in image proper, to read
 *
 * Output:::
 *  ppCmpBuffer::		Pointer to pointer to compressed buffer
 *				to store data (may be updated to accomodate
 *				the compressed strip)
 *  pdwCmpBufsize::		Pointer to size of compressed buffer (may be
 *				updated)
 *  pdwBytesFilled::		Pointer to the number of bytes filled in
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxCompressedReadStrip(
    TFX_IMG *tfxImg,
    UInt32 dwImgRowOffset,
    UInt8 **ppCmpBuffer,
    UInt32 *pdwCmpBufsize,
    UInt32 *pdwBytesFilled
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt32 stripOffset;
    UInt32 stripBytes;
    UInt8 *newBuffer = NULL;
    PROLOGUE;

    /* Default to return nothing */
    *pdwBytesFilled = 0UL;

    TFX_CheckHandle(tfxImg);
    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);

    /* Get the offset and size of the compressed data */
    XERR_PASS_CLEANUP(wp_TfxGetCompressedSize(tfxImg, dwImgRowOffset,
	&stripOffset, &stripBytes));

    /* Make sure we have allocated the right number of bytes */
    if (*pdwCmpBufsize < (UInt32)stripBytes || *ppCmpBuffer == NULL) {
	if (*pdwCmpBufsize > 0 && *ppCmpBuffer != NULL) {
	    TFX_FREE(*ppCmpBuffer);
	}
	IO_CHECKMALLOC(*ppCmpBuffer = TFX_MALLOC(stripBytes));
	*pdwCmpBufsize = stripBytes;
    }

    /* Read the "compressed" data */
    TFX_SEEK(tfxFile, stripOffset);
    TFX_READ(tfxFile, (char *)*ppCmpBuffer, (Int32)stripBytes);

    /* If the wrong fill order, adjust the data */
    switch (tfxImg->fields.FillOrder) {
    case 1:
	break;
    case 2:
#ifndef NOIPCORE
	XERR_IGNORE(wp_TfxBitFlip(*ppCmpBuffer, stripBytes));
#else
	XERR_PASS_CLEANUP(IO_NOSUPPORT);
#endif
	break;
    default:
	XERR_PASS_CLEANUP(IO_NOSUPPORT);
    }

    /* If it's JPEG, and the JPEGTables field was found, insert the tables
     * into the stream */
    if (tfxImg->qtableBufsize) {
	switch (tfxImg->imgInfo.xCompression) {
	case IO_COMP_JPEG:
	    {
		/* Insert the JPEGTables data, minus the SOI/EOI pair */
		UInt8 *np;			/* New stream pointer */
		UInt8 *jp = *ppCmpBuffer;	/* JPEG data stream pointer */
		Int32 newBufsize = stripBytes + tfxImg->qtableBufsize - 4;
		UInt8 *eoipos;			/* Where EOI should occur */
		UInt32 bytes;			/* Bytes in segment */

		/* Allocate a buffer that includes both the tables and the
		 * stream */
		IO_CHECKMALLOC(newBuffer = TFX_MALLOC(newBufsize));
		np = newBuffer;

		/* Copy over, and make sure the stream starts with an SOI */
		XERR_GEN_CLEANUP(*jp != 0xff, IO_BADFILEFORMAT);
		jp++;
		XERR_GEN_CLEANUP(*jp != 216 /* SOI */, IO_BADFILEFORMAT);
		jp++;

		/* Keep track of where the EOI should occur */
		eoipos = *ppCmpBuffer + stripBytes - 2;

		/* Keep looking through the data until we come to a marker that
		 * isn't an APP marker, or a comment */
		while (jp + 1 < eoipos) {
		    /* Is it a marker? */
		    XERR_GEN_CLEANUP(*jp != 0xff, IO_BADFILEFORMAT);

		    /* It's a marker. */
		    if (jp[1] < 0xe0 || (jp[1] > 0xef /* APP */
			&& jp[1] != 0xfe /* comment */)) {
			break;	/* It's something other than APP or Comment */
		    }
		    jp += 2;

		    /* It's an APP marker or Comment.  Get the number of bytes.
		     */
		    bytes = (jp[0] << 8) | jp[1];
		    XERR_GEN_CLEANUP(bytes < 2, IO_BADFILEFORMAT);
		    XERR_GEN_CLEANUP(jp + bytes > eoipos, IO_BADFILEFORMAT);
		    jp += bytes;
		}

		/* Did we parse to the EOI? */
		XERR_GEN_CLEANUP(jp >= eoipos, IO_BADFILEFORMAT);

		/* Copy the JPEG stream up to the non-APP/Comment */
		memcpy(np, *ppCmpBuffer, jp - *ppCmpBuffer);
		np += jp - *ppCmpBuffer;

		/* Insert the tables, minus SOI/EOI */
		memcpy(np, tfxImg->qtableBuf + 2, tfxImg->qtableBufsize - 4);
		np += tfxImg->qtableBufsize - 4;

		/* Copy over the rest of the JPEG stream */
		memcpy(np, jp, stripBytes - (jp - *ppCmpBuffer));

		/* If it fits, reuse the previously allocated buffer */
		if (newBufsize <= (Int32)*pdwCmpBufsize) {

		    /* Copy it over */
		    memcpy(*ppCmpBuffer, newBuffer, newBufsize);

		    /* Free the new buffer */
		    TFX_FREE(newBuffer);
		    newBuffer = NULL;

		} else {
		    /* Otherwise, use this new buffer, and free the old */
		    TFX_FREE(*ppCmpBuffer);
		    *ppCmpBuffer = newBuffer;
		    *pdwCmpBufsize = newBufsize;
		    newBuffer = NULL;
		}

		/* Adjust the number of bytes */
		stripBytes = newBufsize;
	    }
	    break;
	default:
	    break;
	}
    }


    /* Return the number of bytes stored */
    *pdwBytesFilled = stripBytes;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (newBuffer) {
	TFX_FREE(newBuffer);
	newBuffer = NULL;
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the compressed strip data offset and byte count
 *
 * LongDescription:::
 *  wp_TfxGetCompressedSize() returns the appropriate file offset and byte
 *  count for the strip data, as previously retrieve from the TIFF fields.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  dwImgRowOffset::		Row offset in image proper, to read
 *
 * Output:::
 *  pdwOffset::			Pointer to the file offset
 *  pdwBytes::			Pointer to the byte count
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxGetCompressedSize(
    TFX_IMG *tfxImg,
    UInt32 dwImgRowOffset,
    UInt32 *pdwOffset,
    UInt32 *pdwBytes
)
{
    UInt32 stripOffset = 0;
    Int32 stripFileOffset;
    UInt32 stripBytes;
    UInt32 stripRows;
    PROLOGUE;

    /* Default to return nothing */
    *pdwOffset = 0L;
    *pdwBytes = 0UL;

    TFX_CheckHandle(tfxImg);

    if (tfxImg->imgInfo.dwBestStripHeight == 0) {
	UInt32 stripRow = 0;

	/* Look for the next variable-sized strip */
	while (stripRow < dwImgRowOffset &&
	    stripOffset < tfxImg->fields.StripsPerImage) {
	    stripRow += tfxImg->fields.StripRowCounts[stripOffset];
	    stripOffset++;
	}
	XERR_GEN_CLEANUP(stripRow != dwImgRowOffset, IO_BADPARAMETER);
	XERR_GEN_CLEANUP(stripOffset >= tfxImg->fields.StripsPerImage,
	    IO_BADFILEFORMAT);
    } else {

	/* Check the strip offset */
	stripRows = tfxImg->imgInfo.dwBestStripHeight;
	stripOffset = dwImgRowOffset / stripRows;
	while (stripOffset > tfxImg->fields.StripsPerImage) {
	    if (tfxImg->fields.StripsPerImage == 0) {
		if (tfxImg->imgInfo.dwHeight == 0 || stripRows == 0) {
		    NORMAL_RETURN;
		}
		tfxImg->fields.StripsPerImage = (tfxImg->imgInfo.dwHeight + stripRows - 1)
		    / stripRows;
		XERR_GEN_CLEANUP((tfxImg->fields.StripsPerImage == 0),
		    IO_BADFILEFORMAT);
	    } else {
		__error__ = IO_BADPARAMETER;
		goto ERR_CLEANUP;
	    }
	}
    }

    /* Get the offset and size of the compressed data */
    if (tfxImg->fields.StripOffsets == NULL || tfxImg->fields.StripByteCounts == NULL) {
	if ((tfxImg->fields.TileOffsets == NULL ||
	    tfxImg->fields.TileByteCounts == NULL)) {
	    /* This must be a profile M variant, or it's invalid
	     * to have a 0-byte strip */
	    switch (tfxImg->tfxPage->tfxDoc->docInfo.profile) {
	    case TFX_PROFILE_M:
	    case TFX_PROFILE_M1:
	    case TFX_PROFILE_COH_M:
	    case TFX_PROFILE_COH_M1:
		break;
	    default:
		/* It's invalid to have a 0-byte strip if not a profile M variant */
		XERR_GEN_CLEANUP(1, IO_BADFILEFORMAT);
	    }
	    stripFileOffset = stripBytes = 0;
	} else {
	    stripFileOffset = tfxImg->fields.TileOffsets[stripOffset];
	    stripBytes = tfxImg->fields.TileByteCounts[stripOffset];
	}
    } else {
	stripFileOffset = tfxImg->fields.StripOffsets[stripOffset];
	stripBytes = tfxImg->fields.StripByteCounts[stripOffset];
    }

    /* Return the desired results */
    *pdwOffset = stripFileOffset;
    *pdwBytes = stripBytes;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


#ifndef NOIPCORE
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read and decompress one strip of image data
 *
 * LongDescription:::
 *  wp_TfxReadStrip() reads and decompresses a single strip of image data
 *  from a TIFF (TIFF-FX).
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  pPixr::			Pointer to image PIXR
 *  dwPixrRowOffset::		Row offset in PIXR to store image strip
 *  dwImgRowOffset::		Row offset in image proper, to read
 *
 * Output:::
 *  pdwRowsToRead::		Rows of image to read, in outer loop
 *  pdwRowsFilled::		Pointer to the number of rows filled in
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxReadStrip(
    TFX_IMG *tfxImg,
    PIXR *pPixr,
    UInt32 dwPixrRowOffset,
    UInt32 dwImgRowOffset,
    UInt32 *pdwRowsToRead,
    UInt32 *pdwRowsFilled
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt32 stripOffset;
    UInt32 stripBytes;
    UInt32 rowsFilled;
    IPWState *pIPWState;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    TFX_CheckHandle(tfxImg);

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);

    /* Check the strip size & offset:
     *  1. There must be a TIFF "strip" size set.
     *  2. The requested strip size must align with the TIFF strip size.
     */
    if (tfxImg->imgInfo.dwBestStripHeight == 0) {
	UInt32 stripOffset = 0;
	UInt32 stripRow = 0;

	/* Look for the next variable-sized strip */
	while (stripRow < dwImgRowOffset &&
	    stripOffset < tfxImg->fields.StripsPerImage) {
	    stripRow += tfxImg->fields.StripRowCounts[stripOffset];
	    stripOffset++;
	}
	XERR_GEN_CLEANUP(stripRow != dwImgRowOffset, IO_SEQUENCE);
	XERR_GEN_CLEANUP(stripOffset >= tfxImg->fields.StripsPerImage,
	    IO_BADFILEFORMAT);
	XERR_GEN_CLEANUP(*pdwRowsToRead != tfxImg->fields.StripRowCounts[stripOffset],
	    IO_BADPARAMETER);
    } else {
	XERR_GEN_CLEANUP(dwImgRowOffset % tfxImg->imgInfo.dwBestStripHeight != 0,
	    IO_BADPARAMETER);
	XERR_GEN_CLEANUP(*pdwRowsToRead != tfxImg->imgInfo.dwBestStripHeight &&
	    dwImgRowOffset + *pdwRowsToRead != tfxImg->imgInfo.dwHeight,
	    IO_BADPARAMETER);
    }

    /* Get the strip location and size */
    XERR_PASS_CLEANUP(wp_TfxGetCompressedSize(tfxImg, dwImgRowOffset,
	&stripOffset, &stripBytes));

    /* If there's no data, then there's no need to decompress anything */
    if (stripBytes == 0) {
	UInt8 RGBcolor[3];

	/* Fill the strip in with the base color (IFDBaseColor or ImageBaseColor),
	 * if there is one, or the default color */
	XERR_PASS_CLEANUP(wp_TfxGetBaseColor(tfxImg, &RGBcolor[0]));

	/* Fill in the strip with the color */
	XERR_PASS_CLEANUP(w_paint(pIPWState, pPixr, NULL, RGBcolor));

	/* Update the row count */
	rowsFilled = *pdwRowsToRead;
	*pdwRowsToRead -= rowsFilled;
        *pdwRowsFilled = rowsFilled;

	NORMAL_RETURN;
    }

    XERR_PASS_CLEANUP(wp_TfxCompressedReadStrip(tfxImg, dwImgRowOffset,
	&tfxImg->cmpBuffer, &tfxImg->cmpBytes, &stripBytes));

    /* Decompress into output image */
    XERR_PASS_CLEANUP(w_TfxDecompress(pIPWState, (UInt32)tfxImg,
	tfxImg->imgInfo.xCompression, tfxImg->imgInfo.xColorType,
	tfxImg->imgFlags, tfxImg->cmpBuffer, stripBytes,
	dwPixrRowOffset, tfxImg->imgInfo.dwWidth, *pdwRowsToRead,
	pPixr, &rowsFilled));
    XERR_GEN_CLEANUP((rowsFilled < 1), IO_BADFILEFORMAT);
    XERR_GEN_CLEANUP(rowsFilled > *pdwRowsToRead, IO_INTERNAL_ERROR);
    *pdwRowsToRead -= rowsFilled;
    *pdwRowsFilled = rowsFilled;

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read through the Primary IFDs list(s) and store the IFD information
 *
 * LongDescription:::
 *  wp_TfxTraversePrimaryImages() reads through each of the Primary IFDs
 *  chains and stores the information contained within the IFDs.
 *
 * Input:::
 *  docHandle::			Document handle
 *  dwPageOrdinal::		Ordinal (1, 2, ...) of page to search for
 *				(0 for all)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxTraversePrimaryImages(
    TFX_DOC *tfxDoc,
    UInt32 dwPageOrdinal
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_IFDOFFSET_ITEM *ifdp;
    UInt16 fieldCount;
    UInt32 nextIFDOffset;
    UInt16 pageCount = 0UL;
    PROLOGUE;
#if 1
    TFX_CheckHandle(tfxDoc);

    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);

    /* Go through each IFD, and keep it on a linked list */
    ifdp = tfxDoc->ifdList;

    /* Traverse the list as far as it was already created */
    while (ifdp && ifdp->ifdOffset) {
	if (ifdp->next) {
	    if (++pageCount == dwPageOrdinal) {
		break;
	    }
	    nextIFDOffset = ifdp->ifdOffset;
	    ifdp = ifdp->next;
	    continue;
	}	

	/* Go to the start of the IFD */
	TFX_SEEK(tfxFile, ifdp->ifdOffset);

	/* Read the field count */
	TFX_READ(tfxFile, &fieldCount, sizeof(fieldCount));
	TFX_SWAPSHORT(tfxFile, &fieldCount);

	/* Go to the end of the IFD, where the nextIFD is located */
	TFX_SEEK(tfxFile, ifdp->ifdOffset + sizeof(fieldCount) + fieldCount * TFX_TIFFFIELDSIZE);

	/* Read the nextIFD offset */
	TFX_READ(tfxFile, &nextIFDOffset, sizeof(nextIFDOffset));
	TFX_SWAPLONG(tfxFile, &nextIFDOffset);

	/* Update the page count */
	pageCount++;

	/* If there's another one, allocate a new linked list item for it */
	if (nextIFDOffset) {
	    IO_CHECKMALLOC(ifdp->next = (TFX_IFDOFFSET_ITEM *)
		TFX_MALLOC(sizeof(*ifdp->next)));
	    memset(ifdp->next, 0, sizeof(*ifdp->next));
	    ifdp = ifdp->next;
	    ifdp->ifdOffset = nextIFDOffset;
	} else {
	    break;
	}

	/* If this is the last page we were looking for, stop looking further */
	if (dwPageOrdinal == pageCount) {
	    break;
	}
    }

    if (dwPageOrdinal == 0) {
	/* Update the page count */
	tfxDoc->docInfo.wNPages = pageCount;
    }

#else
    TFX_PAGE *tfxPage = parentImg->tfxPage;
    TFX_IMG *childImg = NULL;	/* Image found on subimage path (unlinked) */
    TFX_IMG *curImg;		/* Currently linked image */
    TFX_IMG *siblingImg;		/* Sibling already linked */
    TFX_IMG *nextImg;		/* Next sibling in links */
    UInt32 nextIFD;
    UInt32 maskOffset = 0;
    Int32 imgOffset = 0L;
    UInt32 layer = 0UL;
    TFX_IMG **listPtr = NULL;
    UInt32 subPath;
    IO_IMGTYPE tfxImgType;
    PROLOGUE;

    TFX_CheckHandle(parentImg);

    /* Search through the images starting with the first subimage. There may
     * exist more than one starting node. */
    for (subPath = 0; subPath < parentImg->fields.SubIFDCount; subPath++) {
	nextIFD = parentImg->fields.SubIFDs[subPath];

	/* Loop while there's one attached */
	while (nextIFD) {

	    /* Add the next in the sequence */
	    XERR_PASS_CLEANUP(wp_TfxNewImg(tfxPage, &childImg, nextIFD));

	    /* Get the scoop on this image (esp. ImageLayer) */
	    XERR_PASS_CLEANUP(wp_TifFXReadFields(childImg));

	    /* Check to see if it was a mask image */
	    tfxImgType = IO_IMGTYPE_UNKNOWN;
	    if (tfxPage->tfxDoc->docInfo.version > 2) {
		if (childImg->fields.NewSubFileType & TFX_TIFFSUBFILE_MASK) {
		    tfxImgType = IO_IMGTYPE_MASK;
		    layer = 0;	    /* Not part of the MRC */
		    imgOffset = maskOffset++;
		    listPtr = &childImg->xMaskImg;
		}
	    }

	    if (tfxImgType == IO_IMGTYPE_UNKNOWN) {

		/* Determine which type of image it is */
		switch ((int)childImg->fields.ImageLayer[0]) {
		case 0:
		    if (tfxPage->tfxDoc->docInfo.version == 2) {
			XERR_PASS_CLEANUP(wp_TfxIncrementImageCounts(tfxPage,
			    layer = 5, 1));
			tfxImgType = IO_IMGTYPE_LAYER;
			imgOffset = (Int32)tfxPage->pageInfo.pdwImageCounts[5-1]
			    - 1;
			listPtr = &tfxPage->tfxImgs;
		    } else if ((tfxPage->tfxDoc->docInfo.version > 2) &&
			(childImg->fields.PhotometricInterpretation == 4) &&
			((childImg->fields.NewSubFileType & TFX_TIFFSUBFILE_MASK) != 0)) {
			tfxImgType = IO_IMGTYPE_MASK;
			listPtr = &parentImg->xMaskImg;
			imgOffset = maskOffset++;
		    } else {
			__error__ = IO_BADFILEFORMAT;
			goto ERR_CLEANUP;
		    }
		    break;
		default:
		    XERR_PASS_CLEANUP(wp_TfxIncrementImageCounts(tfxPage,
			layer = childImg->fields.ImageLayer[0], 1));
		    tfxImgType = IO_IMGTYPE_LAYER;
		    imgOffset = (Int32)childImg->fields.ImageLayer[1] - 1;
		    if (imgOffset < 0) {
			imgOffset = tfxPage->pageInfo.pdwImageCounts[layer - 1];
		    }
		    listPtr = &tfxPage->tfxImgs;
		    break;
		}
	    }

	    /* Retain it's identity */
	    childImg->imgInfo.tfxImgType = tfxImgType;
	    childImg->imgInfo.dwImgLayer = layer;
	    childImg->imgInfo.dwImgOffset = (UInt32)imgOffset;

	    /* Reference the appropriate linked list */
	    siblingImg = *listPtr;

	    /* If this is the first, just insert it */
	    if (siblingImg == NULL) {
		*listPtr = curImg = childImg;
		childImg = NULL;
	    } else if (siblingImg->imgInfo.dwImgLayer > layer ||
		(siblingImg->imgInfo.dwImgLayer == layer &&
		siblingImg->imgInfo.dwImgOffset > (UInt32)imgOffset)) {

		/* This one should really be first. Insert as such. */
		childImg->next = siblingImg;
		siblingImg->prev = childImg;
		*listPtr = curImg = childImg;
		childImg = NULL;
	    } else {

		/* Search for the proper point to insert *after* */
		while (((nextImg = siblingImg->next) != NULL) &&
		    (nextImg->imgInfo.dwImgLayer < layer ||
		    (nextImg->imgInfo.dwImgLayer == layer &&
		    nextImg->imgInfo.dwImgOffset < (UInt32)imgOffset))) {
		    siblingImg = nextImg;
		}

		/* Link this image to the one that was next */
		childImg->next = nextImg;
		if (nextImg) {
		    nextImg->prev = childImg;
		}

		/* Link prev image to this one */
		siblingImg->next = curImg = childImg;
		childImg->prev = siblingImg;
		childImg = NULL;
	    }

	    /* Keep the loop going if there's another IFD attached */
	    nextIFD = curImg->fields.nextIFD;

	    if (tfxPage->tfxDoc->docInfo.version == 2) {
		/* Well, it can't be a mask, since they're attached via
		 * the MaskIFDs field. Since it's not a mask, look for one. */
		XERR_PASS_CLEANUP(wp_TfxTraverseMaskImages(curImg));
	    }
	} /* For each subimage in sub path */
    } /* For each sub path */

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (childImg) {
	TFX_INSTANCE *instData = TFX_IMGINST(parentImg);
	TFX_FREE(childImg);
    }
#endif

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read through the SubIFDs list(s) and store the IFD information
 *
 * LongDescription:::
 *  wp_TfxTraverseSubImages() reads through each of the SubIFDs chains and
 *  stores the information contained within the IFDs.
 *
 * Input:::
 *  parentImg::			Image handle of parent image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxTraverseSubImages(
    TFX_IMG *parentImg
)
{
    TFX_PAGE *tfxPage = parentImg->tfxPage;
    TFX_IMG *childImg = NULL;	/* Image found on subimage path (unlinked) */
    TFX_IMG *curImg;		/* Currently linked image */
    TFX_IMG *siblingImg;		/* Sibling already linked */
    TFX_IMG *nextImg;		/* Next sibling in links */
    UInt32 nextIFD;
    Int32 imgOffset = 0L;
    UInt32 layer = 0UL;
    TFX_IMG **listPtr = NULL;
    UInt32 subPath;
    IO_IMGTYPE tfxImgType;
    PROLOGUE;

    TFX_CheckHandle(parentImg);

    /* Search through the images starting with the first subimage. There may
     * exist more than one starting node. */
    for (subPath = 0; subPath < parentImg->fields.SubIFDCount; subPath++) {
	nextIFD = parentImg->fields.SubIFDs[subPath];

	/* Loop while there's one attached */
	while (nextIFD) {

	    /* Add the next in the sequence */
	    XERR_PASS_CLEANUP(wp_TfxNewImg(tfxPage, &childImg, nextIFD));

	    /* Get the scoop on this image (esp. ImageLayer) */
	    XERR_PASS_CLEANUP(wp_TifFXReadIFD(childImg));

	    /* Check to see if it was a mask image */
	    tfxImgType = IO_IMGTYPE_UNKNOWN;

	    if (tfxImgType == IO_IMGTYPE_UNKNOWN) {

		/* Determine which type of image it is */
		switch ((int)childImg->fields.ImageLayer[0]) {
		case 0:
		    if (tfxPage->tfxDoc->docInfo.version == 2) {
			XERR_PASS_CLEANUP(wp_TfxIncrementImageCounts(tfxPage,
			    layer = 5, 1));
			tfxImgType = IO_IMGTYPE_LAYER;
			imgOffset = (Int32)tfxPage->pageInfo.pdwImageCounts[5-1]
			    - 1;
			listPtr = &tfxPage->tfxImgs;
		    } else {
			__error__ = IO_BADFILEFORMAT;
			goto ERR_CLEANUP;
		    }
		    break;
		default:
		    XERR_PASS_CLEANUP(wp_TfxIncrementImageCounts(tfxPage,
			layer = childImg->fields.ImageLayer[0], 1));
		    tfxImgType = IO_IMGTYPE_LAYER;
		    imgOffset = (Int32)childImg->fields.ImageLayer[1] - 1;
		    if (imgOffset < 0) {
			imgOffset = tfxPage->pageInfo.pdwImageCounts[layer - 1];
		    }
		    listPtr = &tfxPage->tfxImgs;
		    break;
		}
	    }

	    /* Retain it's identity */
	    childImg->imgInfo.tfxImgType = tfxImgType;
	    childImg->imgInfo.dwImgLayer = layer;
	    childImg->imgInfo.dwImgOffset = (UInt32)imgOffset;

	    /* Reference the appropriate linked list */
	    siblingImg = *listPtr;

	    /* If this is the first, just insert it */
	    if (siblingImg == NULL) {
		*listPtr = curImg = childImg;
		childImg = NULL;
	    } else if (siblingImg->imgInfo.dwImgLayer > layer ||
		(siblingImg->imgInfo.dwImgLayer == layer &&
		siblingImg->imgInfo.dwImgOffset > (UInt32)imgOffset)) {

		/* This one should really be first. Insert as such. */
		childImg->next = siblingImg;
		siblingImg->prev = childImg;
		*listPtr = curImg = childImg;
		childImg = NULL;
	    } else {

		/* Search for the proper point to insert *after* */
		while (((nextImg = siblingImg->next) != NULL) &&
		    (nextImg->imgInfo.dwImgLayer < layer ||
		    (nextImg->imgInfo.dwImgLayer == layer &&
		    nextImg->imgInfo.dwImgOffset < (UInt32)imgOffset))) {
		    siblingImg = nextImg;
		}

		/* Link this image to the one that was next */
		childImg->next = nextImg;
		if (nextImg) {
		    nextImg->prev = childImg;
		}

		/* Link prev image to this one */
		siblingImg->next = curImg = childImg;
		childImg->prev = siblingImg;
		childImg = NULL;
	    }

	    /* Keep the loop going if there's another IFD attached */
	    nextIFD = curImg->fields.nextIFD;

	} /* For each subimage in sub path */
    } /* For each sub path */

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (childImg) {
	TFX_INSTANCE *instData = TFX_IMGINST(parentImg);
	TFX_FREE(childImg);
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read the HintsIFDs chain and store the IFD information
 *
 * LongDescription:::
 *  wp_TfxTraverseHintsImages() reads through the HintsIFDs chain and
 *  stores the information contained within the IFDs as transparency hints
 *  information.
 *
 * Input:::
 *  parentImg::			Image handle of parent image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TFX
 *
 * Keywords:::
 *  file io, MRC, TFX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxTraverseHintsImages(
    TFX_IMG *parentImg
)
{
    TFX_PAGE *tfxPage = parentImg->tfxPage;
    TFX_DOC *tfxDoc;
    TFX_IMG *childImg = NULL;
    UInt32 nextIFD = 0UL;
    UInt32 imgOffset = 0UL;
    TFX_IMG **listPtr;
    PROLOGUE;

    TFX_CheckHandle(parentImg);

    if ((tfxPage->tfxDoc->readWrite & READ_ALLOWED) == 0) {
	NORMAL_RETURN;
    }

    /* Search through the images starting with the first subimage.
     */
    tfxDoc = tfxPage->tfxDoc;
    nextIFD = parentImg->fields.RenderingHintsIFD;

    /* Loop while there's one attached */
    while (nextIFD) {
#if 1
	/* Add the next in the sequence */
	XERR_PASS_CLEANUP(wp_TfxNewImg(tfxPage, &childImg, nextIFD));

	/* Get the scoop on this image */
	XERR_PASS_CLEANUP(wp_TifFXReadIFD(childImg));

#else

	/* Add the next in the sequence */
	XERR_PASS_CLEANUP(wp_TfxNewImg(tfxPage, &childImg, nextIFD));

	/* Get the scoop on this image */
	XERR_PASS_CLEANUP(wp_TifFXReadFields(childImg, nextIFD, &nextIFD));

#endif

	/* Retain it's identity */
	childImg->imgInfo.tfxImgType = IO_IMGTYPE_HINTS;
	childImg->imgInfo.dwImgOffset = imgOffset++;

	/* Place it in the list of the appropriate type */
	listPtr = &parentImg->xHintsImgs;
	while (*listPtr && (*listPtr)->imgInfo.dwImgOffset <= imgOffset) {
	    listPtr = &((*listPtr)->next);
	}
	if (*listPtr) {
	    childImg->next = *listPtr;  /* Relink the next one here */
	}
	*listPtr = childImg;	    /* Insert this one */

	/* Go to the next image */
	nextIFD = childImg->fields.nextIFD;

	/* Keep from freeing this one */
	childImg = NULL;

    } /* For each subimage in sub path */

    /* Mark hints count in parent */
    parentImg->imgInfo.dwHintsImages = imgOffset;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (childImg) {
	TFX_INSTANCE *instData = TFX_IMGINST(parentImg);
	TFX_FREE(childImg);
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read or retrieve the annotation table header
 *
 * LongDescription:::
 *  wp_TfxGetAnnotTableHeader() reads the annotation table header. If it has
 *  previously been gotten, nothing is done. The annotation table header is
 *  stored in the tfxPage structure.
 *
 * Input:::
 *  tfxPage::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxGetAnnotTableHeader(
    TFX_PAGE *tfxPage
)
{
    TFX_FILE *tfxFile;
    TFX_IMG *primImg = NULL;
    Int32 fileOffset;
    UInt16 sval; UInt32 lval;
    PROLOGUE;

    TFX_CheckHandle(tfxPage);

    tfxFile = TFX_PAGEFILE(tfxPage);


    /* Refer to primary image for annotation field */
    primImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    if (tfxPage->annotTableHeader.entriesPerTable > 0) {
	NORMAL_RETURN;
    }

    if (tfxPage->tfxDoc->docInfo.version == 2) {

    /* Since no real "table" exists, call each entry in the array
	* of offsets a "table entry" */
	tfxPage->annotTableHeader.entriesPerTable = 1;

	/* Make sure we don't "think" there are annotations when there aren't */
	if (primImg->fields.AnnotationOffsets == NULL) {
	    tfxPage->annotTableHeader.annotCount = 0;
	} else {
	    tfxPage->annotTableHeader.annotCount = (UInt16)primImg->fields.AnnotationCount;
	}
    } else if (tfxPage->tfxDoc->docInfo.version > 2) {

	/* Check to see if any exist */
	if (primImg->fields.AnnotationOffsets == NULL) {
	    tfxPage->annotTableHeader.annotCount = 0;
	} else {

	    /* The count should always be 1 in a 3.x file */
	    XERR_GEN_CLEANUP(primImg->fields.AnnotationCount != 1, IO_BADFILEFORMAT);

	    /* Read the header */
	    tfxPage->annotTableHeaderOffset = fileOffset =
		primImg->fields.AnnotationOffsets[0];

	    /* If the offset is zero, there is no annotation table header */
	    if (fileOffset == 0) {
		tfxPage->annotTableHeader.annotCount = 0;
		tfxPage->annotTableHeader.entriesPerTable = 20;
		tfxPage->annotTableHeader.lastOrdinal = 0;
		tfxPage->annotTableHeader.firstTableOffset = 0;
		NORMAL_RETURN;
	    }
	    XERR_GEN_CLEANUP(fileOffset == 0, IO_BADFILEFORMAT);
	    TFX_SEEK(tfxFile, fileOffset);

	    TFX_READ(tfxFile, &sval, sizeof(sval));
	    TFX_SWAPSHORT(tfxFile, &sval);
	    tfxPage->annotTableHeader.annotCount = sval;
	    TFX_READ(tfxFile, &sval, sizeof(sval));
	    TFX_SWAPSHORT(tfxFile, &sval);
	    tfxPage->annotTableHeader.entriesPerTable = sval;
	    TFX_READ(tfxFile, &sval, sizeof(sval));
	    TFX_SWAPSHORT(tfxFile, &sval);
	    tfxPage->annotTableHeader.bytesPerEntry = sval;
#if 1 	/* To be taken out when those reading non-updated TIFF-FX 3.0 files
	 * have updated their files */
	    {
		TFX_ANNOTTABLEENTRY *ep = NULL;

		if (tfxPage->annotTableHeader.bytesPerEntry !=
		    sizeof(ep->id) + sizeof(ep->type) +
		    sizeof(ep->xOffset) + sizeof(ep->yOffset) +
		    sizeof(ep->width) + sizeof(ep->height) +
		    sizeof(ep->flags) +
		    sizeof(ep->bytes) + sizeof(ep->offset)) {
		    tfxPage->annotTableHeader.lastOrdinal =
			tfxPage->annotTableHeader.bytesPerEntry;

		    /* Must be missing the new flags element */
		    tfxPage->annotTableHeader.bytesPerEntry = 28;
		} else {
		    TFX_READ(tfxFile, &sval, sizeof(sval));
		    TFX_SWAPSHORT(tfxFile, &sval);
		    tfxPage->annotTableHeader.lastOrdinal = sval;
		}
	    }
#else
	    TFX_READ(tfxFile, &sval, sizeof(sval));
	    TFX_SWAPSHORT(tfxFile, &sval);
	    tfxPage->annotTableHeader.lastOrdinal = sval;
#endif
	    TFX_READ(tfxFile, &lval, sizeof(lval));
	    TFX_SWAPLONG(tfxFile, &lval);
	    tfxPage->annotTableHeader.firstTableOffset = lval;
	} /* AnnotationOffset tag exists */
    } /* version 2 else > 2 */

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a single annotation table
 *   
 *
 * LongDescription:::
 *  wp_TfxGetAnnotTable() reads a single annotation table, and inserts it into
 *  a linked list of annotation tables. If the data has previously been read
 *  or added, the data is found and returned.*
 *
 * Input:::
 *  tfxPage::			Data structure of valid page
 *  annotOffset::		Annotation offset seeking for
 *
 * Output:::
 *  ppAnnotPageTable::		If NULL, the table is read and nothing
 *				returned. Otherwise the pointer will point to
 *				the resulting table.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxGetAnnotTable(
    TFX_PAGE *tfxPage,
    UInt16 annotOffset,
    TFX_ANNOTTABLE **ppAnnotPageTable
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    TFX_IMG *primImg = NULL;
    TFX_ANNOTTABLE *at, *newAt = NULL;		/* Linked annotation table */
    TFX_ANNOTTABLEENTRY *ep;
    UInt32 mallocbytes;
    Int32 fileOffset;
    UInt16 entriesLeft;
    UInt16 ordinal = 0;		    /* Table ordinal */
    UInt32 tableOffset;		    /* Offset of all tables that annotation is
				     * found */
    UInt32 entryOffset;		    /* Entry within table of annotation */
    UInt32 curAnnotOffset;	    /* Annotation offset within loop */
    UInt8 *annotEntryData = NULL;   /* Annotation table intermediate buffer */
    UInt8 *entryStart, *entry;	    /* annotEntryData pointers */
    UInt32 bytesLeft;	 	    /* Bytes left in entry data */
    UInt32 twoHeaderSize = 0;	    /* If v 2.0, the bytes in the header
				     * portion */
    PROLOGUE;

    /* Default to return nothing */
    if (ppAnnotPageTable) {
	*ppAnnotPageTable = NULL;
    }

    TFX_CheckHandle(tfxPage);
    instData = TFX_PAGEINST(tfxPage);
    tfxFile = TFX_PAGEFILE(tfxPage);
    at = tfxPage->firstAnnotTable;

    /* Check for proper calling sequence */


    /* Make sure we've read the annotation table header (if any) */
    if (tfxPage->annotTableHeader.entriesPerTable == 0UL) {
	XERR_GEN_CLEANUP(tfxPage->annotTableHeader.annotCount > 0,
			 IO_INTERNAL_ERROR);

	XERR_PASS_CLEANUP(wp_TfxGetAnnotTableHeader(tfxPage));
    }

    /* Check parameters */
    XERR_GEN_CLEANUP(annotOffset >= tfxPage->annotTableHeader.annotCount,
		     IO_BADPARAMETER);

    /* Refer to primary image for annotation field */
    primImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    if (tfxPage->tfxDoc->readWrite & READ_ALLOWED) {

        /* For TIFF-FX 2.0, the annotation information is at the same place as
	 * the data itself. Get the file offset of the header.
	 * For TIFF-FX 3.0, the table offset is contained in the table header. */
	if (tfxPage->tfxDoc->docInfo.version == 2 &&
	    tfxPage->annotTableHeader.firstTableOffset == 0) {
	    XERR_GEN_CLEANUP(primImg->fields.AnnotationOffsets == NULL,
		IO_INTERNAL_ERROR);
	    fileOffset = primImg->fields.AnnotationOffsets[annotOffset];
	} else {
	    fileOffset = tfxPage->annotTableHeader.firstTableOffset;
	}
    } else {
	fileOffset = 0;
    }

    /* Compute the table and entry */
    tableOffset = annotOffset / tfxPage->annotTableHeader.entriesPerTable;
    entryOffset = annotOffset % tfxPage->annotTableHeader.entriesPerTable;

    /* If we're installing the first, don't look for cached tables */
    if (at) {

	/* Since the annotation tables are always read in order, and cached
	 * when read, go through the cached list until we come to the
	 * last one cached */
	while (at->ordinal <= tableOffset) {

	    /* Should be in sequence */
	    XERR_GEN_CLEANUP(++ordinal != at->ordinal, IO_INTERNAL_ERROR);

	    /* There is no such annotation table */
	    if (tfxPage->tfxDoc->docInfo.version >= 3 &&
		(tfxPage->tfxDoc->readWrite & READ_ALLOWED)) {
		XERR_GEN_CLEANUP(((fileOffset = at->nextTableOffset) == 0),
		    IO_BADPARAMETER);
	    }
	    if (at->next == NULL) {

		/* Table may exist, but hasn't been read yet */
		break;
	    }
	    at = at->next;
	}

	/* If we're at the right annotation table, and it's already read,
	 * bail now */
	if (at->ordinal == tableOffset + 1) {
	    goto found;
	}
    }

    /* We shouldn't get here if we're not reading */
    XERR_GEN_CLEANUP((tfxPage->tfxDoc->readWrite & READ_ALLOWED) == 0,
	IO_INTERNAL_ERROR);

    /* Prepare to read the entire table (3.0) or entry (2.0) into a single
     * buffer in a single read */
    if (tfxPage->tfxDoc->docInfo.version == 2) {
	mallocbytes = twoHeaderSize = sizeof(ep->bytes) + sizeof(ep->type) +
	    sizeof(ep->xOffset) + sizeof(ep->yOffset) +
	    sizeof(ep->width) + sizeof(ep->height);
    } else {
	mallocbytes = sizeof(short) + tfxPage->annotTableHeader.entriesPerTable *
	    tfxPage->annotTableHeader.bytesPerEntry;
    }
    IO_CHECKMALLOC(annotEntryData = (UInt8 *)TFX_MALLOC(mallocbytes));

    /* Read as many annotation tables as necessary to retrieve the one
     * requested */
    do {
	/* Seek to the next annotation table offset.
	 * If there aren't the number of annotation tables that the annotation
	 * table header claims, the file is corrupt */
	XERR_GEN_CLEANUP(fileOffset < 8, IO_BADFILEFORMAT);
	TFX_SEEK(tfxFile, fileOffset);

	IO_CHECKMALLOC(newAt = (TFX_ANNOTTABLE *)TFX_MALLOC(sizeof(TFX_ANNOTTABLE)));
	memset(newAt, 0, sizeof(TFX_ANNOTTABLE));
	mallocbytes = sizeof(TFX_ANNOTTABLEENTRY) * 
	    tfxPage->annotTableHeader.entriesPerTable;
	IO_CHECKMALLOC(newAt->entries =
		     (TFX_ANNOTTABLEENTRY *)TFX_MALLOC(mallocbytes));
	memset(newAt->entries, 0, mallocbytes);

	if (at) {
	    /* Attach to end of linked list */
	    at->next = newAt;
	    at = newAt;
	} else {
	    /* Begin list here */
	    at = tfxPage->firstAnnotTable = newAt;
	}
	newAt = NULL;	/* Avoid freeing if later failure */

	/* Read the table */
	ep = at->entries;
	if (tfxPage->tfxDoc->docInfo.version == 2) {
	    at->ordinal = annotOffset + 1;
	    ep->id = annotOffset;
	    ep->offset = primImg->fields.AnnotationOffsets[annotOffset] +
		twoHeaderSize;

	    /* Read the entire header portion of the data (expect that the
	     * byte count should at least include the header portion). */
	    TFX_READ(tfxFile, annotEntryData, (Int32)twoHeaderSize);
	    entry = annotEntryData;

	    memcpy(&ep->bytes, entry, sizeof(long));
	    TFX_SWAPLONG(tfxFile, &ep->bytes);
	    entry += sizeof(long);

	    XERR_GEN_CLEANUP(ep->bytes < sizeof(ep->type) + sizeof(ep->xOffset)
		+ sizeof(ep->yOffset) + sizeof(ep->width) + sizeof(ep->height),
		IO_BADFILEFORMAT);
	    ep->bytes -= sizeof(ep->type) + sizeof(ep->xOffset) +
		sizeof(ep->yOffset) + sizeof(ep->width) + sizeof(ep->height);

	    memcpy(&ep->type, entry, sizeof(short));
	    TFX_SWAPSHORT(tfxFile, &ep->type);
	    entry += sizeof(short);
	    memcpy(&ep->xOffset, entry, sizeof(long));
	    TFX_SWAPLONG(tfxFile, &ep->xOffset);
	    entry += sizeof(long);
	    memcpy(&ep->yOffset, entry, sizeof(long));
	    TFX_SWAPLONG(tfxFile, &ep->yOffset);
	    entry += sizeof(long);
	    memcpy(&ep->width, entry, sizeof(long));
	    TFX_SWAPLONG(tfxFile, &ep->height);
	    entry += sizeof(long);
	    memcpy(&ep->height, entry, sizeof(long));
	    TFX_SWAPLONG(tfxFile, &ep->height);
	    entry += sizeof(long);
	    if (annotOffset + 1 < primImg->fields.AnnotationCount) {
		at->nextTableOffset =
		  primImg->fields.AnnotationOffsets[annotOffset+1];
	    }
	} else if (tfxPage->tfxDoc->docInfo.version > 2) {

	    /* Read the ID and whole list of table entries */
	    TFX_READ(tfxFile, annotEntryData, (Int32)(sizeof(short) +
		     tfxPage->annotTableHeader.entriesPerTable *
		     tfxPage->annotTableHeader.bytesPerEntry));

	    memcpy(&at->ordinal, annotEntryData, sizeof(short));
	    TFX_SWAPSHORT(tfxFile, &at->ordinal);
	    entryStart = annotEntryData + sizeof(short);

	    entriesLeft = tfxPage->annotTableHeader.entriesPerTable;
	    curAnnotOffset = tableOffset *
	      tfxPage->annotTableHeader.entriesPerTable;

	    /* Now parse them out: pick out the number of bytes that
	     * the annotation table header says are in this version,
	     * and stuff them as they fit into the structure known
	     * at the time this implementation was made.  In future
	     * versions of TIFF-FX, there will be more annotation data
	     * than we know what to do with, so we'll just take what
	     * we know about. And in older versions of TIFF-FX, we will
	     * have less data than our structure, so we won't fill in
	     * the whole structure. */
	    while (entriesLeft--) {
		entry = entryStart;
		bytesLeft = tfxPage->annotTableHeader.bytesPerEntry;
		if (bytesLeft >= sizeof(short)) {
		    memcpy(&ep->id, entry, sizeof(short));
		    TFX_SWAPSHORT(tfxFile, &ep->id);
		    entry += sizeof(short), bytesLeft -= sizeof(short);
		    XERR_GEN_CLEANUP(ep->id != curAnnotOffset &&
			curAnnotOffset < tfxPage->annotTableHeader.annotCount,
			IO_BADFILEFORMAT);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(short)) {
		    memcpy(&ep->type, entry, sizeof(short));
		    TFX_SWAPSHORT(tfxFile, &ep->type);
		    entry += sizeof(short), bytesLeft -= sizeof(short);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->xOffset, entry, sizeof(long));
		    TFX_SWAPLONG(tfxFile, &ep->xOffset);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->yOffset, entry, sizeof(long));
		    TFX_SWAPLONG(tfxFile, &ep->yOffset);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->width, entry, sizeof(long));
		    TFX_SWAPLONG(tfxFile, &ep->width);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->height, entry, sizeof(long));
		    TFX_SWAPLONG(tfxFile, &ep->height);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->bytes, entry, sizeof(long));
		    TFX_SWAPLONG(tfxFile, &ep->bytes);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->offset, entry, sizeof(long));
		    TFX_SWAPLONG(tfxFile, &ep->offset);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}
		if (bytesLeft >= sizeof(long)) {
		    memcpy(&ep->flags, entry, sizeof(long));
		    TFX_SWAPLONG(tfxFile, &ep->flags);
		    entry += sizeof(long), bytesLeft -= sizeof(long);
		} else {
		    goto nextEntry;
		}

		/* Always add new entry fields to the end of the list
		 * for backwards compatibility */

	    nextEntry:
		entryStart += tfxPage->annotTableHeader.bytesPerEntry;
		ep++;
		curAnnotOffset++;
	    } /* While entries left */

	    /* Read the next annotation table offset */
	    TFX_READ(tfxFile, &at->nextTableOffset, sizeof(at->nextTableOffset));
	    TFX_SWAPLONG(tfxFile, &at->nextTableOffset);
	    fileOffset = at->nextTableOffset;
	} /* version > 2 */

	/* Expect to find the tables in sequence */
	ordinal++;

    } while (at->ordinal != tableOffset + 1);

    /* Go to the entry offset */
    ep = &at->entries[entryOffset];

    /* Expect not to find an annotation with an offset of zero */
    XERR_GEN_CLEANUP(ep->offset == 0, IO_BADFILEFORMAT);

    /* Read the header portion of the pertinent entry if version 2 and it
     * hasn't already been read */
    if (tfxPage->tfxDoc->docInfo.version == 2 && ep->bytes == 0) {

	/* Seek to the data (already there) */

	/* Read the header */
	TFX_READ(tfxFile, &ep->bytes, sizeof(ep->bytes));
	TFX_SWAPLONG(tfxFile, &ep->bytes);

	/* Remove the header from the byte count and offset */
	ep->bytes -= sizeof(UInt32) * 5 + sizeof(UInt16);
	ep->offset += sizeof(UInt32) * 5 + sizeof(UInt16);

	/* Continue reading */
	TFX_READ(tfxFile, &ep->type, sizeof(ep->type));
	TFX_SWAPSHORT(tfxFile, &ep->type);
	TFX_READ(tfxFile, &ep->xOffset, sizeof(ep->xOffset));
	TFX_SWAPLONG(tfxFile, &ep->xOffset);
	TFX_READ(tfxFile, &ep->yOffset, sizeof(ep->yOffset));
	TFX_SWAPLONG(tfxFile, &ep->yOffset);
	TFX_READ(tfxFile, &ep->width, sizeof(ep->width));
	TFX_SWAPLONG(tfxFile, &ep->width);
	TFX_READ(tfxFile, &ep->height, sizeof(ep->height));
	TFX_SWAPLONG(tfxFile, &ep->height);
    } /* version */

 found:

    if (annotEntryData) {
	TFX_FREE(annotEntryData);
	annotEntryData = NULL;
    }

    /* Return the one found */
    if (ppAnnotPageTable) {
	*ppAnnotPageTable = at;
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    if (annotEntryData) {
	TFX_FREE(annotEntryData);
	annotEntryData = NULL;
    }

    if (newAt) {
	TFX_FREE(newAt);
	if (newAt->entries) {
	    TFX_FREE(newAt->entries);
	}
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the data associated with an annotation
 *
 * LongDescription:::
 *  wp_TfxGetAnnotData() reads the data associated with an annotation. The data
 *  is disjoint from the annotation table, which contains a reference to the
 *  data.
 *
 * Input:::
 *  tfxPage::			Data structure of valid document
 *  annotOffset::		Annotation offset
 *  bytes::			Number of data bytes (excludes the TIFF-FX 2.0
 *				annotation description)
 *
 * Output:::
 *  pData::			Pointer to a preallocated data buffer with
 *				the same number of bytes
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxGetAnnotData(
    TFX_PAGE *tfxPage,
    UInt16 annotOffset,
    UInt32 bytes,
    UInt8 *pData
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt16 tableOffset;
    UInt16 entryOffset;
    TFX_ANNOTTABLE *annotTable;
    TFX_ANNOTTABLEENTRY *ep;
    PROLOGUE;

    TFX_CheckHandle(tfxPage);
    instData = TFX_PAGEINST(tfxPage);
    tfxFile = TFX_PAGEFILE(tfxPage);

    /* Check for proper calling sequence */


    /* Read the appropriate annot table */
    tableOffset = annotOffset / tfxPage->annotTableHeader.entriesPerTable;
    entryOffset = annotOffset % tfxPage->annotTableHeader.entriesPerTable;
    XERR_PASS_CLEANUP(wp_TfxGetAnnotTable(tfxPage,
	(UInt16)(tableOffset * tfxPage->annotTableHeader.entriesPerTable),
	&annotTable));

    /* Check to make sure the annotation information has already been requested */
    ep = &annotTable->entries[entryOffset];
    XERR_GEN_CLEANUP(ep->offset == 0, IO_SEQUENCE);
    XERR_GEN_CLEANUP(ep->bytes != bytes, IO_BADPARAMETER);

    /* Go to the file offset of the data */
    TFX_SEEK(tfxFile, ep->offset);

    /* Read the data */
    TFX_READ(tfxFile, pData, (Int32)bytes);

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}

/*
$Log: tfxrtfx.c,v $
Revision 1.2  2003/02/10 22:53:31  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:11  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.28  2002/02/06 19:28:45  dabercro
Use the page count from the PageNumber in the first IFD, if non-zero.
Otherwise, traverse the IFD chain to obtain the page count.

Revision 1.27  2001/12/26 21:23:45  dabercro
Mark as released

Revision 1.26  2001/10/30 20:03:37  dabercro
Remove unused variables.

Revision 1.25  2001/08/22 19:49:34  dabercro
Added support for new field: IFDBaseColor

Revision 1.24  2001/07/20 23:20:08  dabercro
Automatically insert the JPEGTables.  Consolidate the read of the
compressed data to a single function.

Revision 1.23  2001/06/29 17:50:11  dabercro
Free the tables from the JPEGTABLES field.

Revision 1.22  2001/06/27 18:56:10  dabercro
Rename RGB to sRGB

Revision 1.21  2001/06/20 19:56:01  dabercro
Fix error condition to check only for non-existing resources, not whether the
first resource table was previously read.

Revision 1.20  2001/05/08 23:42:39  dabercro
comment changes

Revision 1.19  2001/03/22 17:50:44  dabercro
Allow 0-byte strips in profile M and it's variants ONLY.

Revision 1.18  2001/02/24 03:26:15  dabercro
Avoid complaining if it's just not this format.

Revision 1.17  2001/01/18 00:21:31  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.16  2000/11/17 23:57:35  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.15  2000/11/02 22:21:21  dabercro
Additions for Rendering Hints.

Revision 1.14  2000/10/18 20:29:53  dabercro
Interpret a 0-byte strip as an all zero image.

Revision 1.13  2000/10/16 18:22:53  dabercro
Major adds to support stripes, 0-byte strips, and ImageBaseColor.

Revision 1.12  2000/09/27 21:56:44  dabercro
Variable strips and stripe work.

Revision 1.11  2000/09/21 21:19:14  dabercro
Add variable-sized strips: StripRowCounts.

Revision 1.10  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.9  2000/08/22 01:03:23  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.8  2000/08/04 21:54:55  dabercro
Adjusted list check.

Revision 1.7  2000/08/01 00:02:57  dabercro
Significant changes to the handling of shared resources, as per
the draft of TIFF-FX Extension 1.

Revision 1.6  2000/06/07 00:00:17  dabercro
Continued work, lots of stuff.

Revision 1.5  2000/05/09 01:52:58  dabercro
Added FillOrder field support.
Updated JBIG2 dictionary front-end.
Support for JBIG2 dictionary extensions, updated parameter to
reflect byte count for extensions.
Write fields that are explicitly set, regardless of default values.
Check if document is single-page when closing, reflect in
NewSubfileType.

Revision 1.4  2000/04/12 19:06:09  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.3  2000/03/11 02:06:35  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.2  2000/02/08 21:25:33  dabercro
Disable JBIG2 (for now), updates to make use of GlobalParametersIFD for symbol libraries.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.35  1998/09/25 22:14:09  aber
aber on Fri Sep 25 15:13:04 PDT 1998

tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

If reading an image that was supposedly compressed to zero bytes,
handle it gracefully. It can't really have zero bytes, since no compression
algorithm we have will compress to nothing.

Revision 1.34  1998/08/04 00:21:34  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Support multiple dictionaries when token compressing, used to retain
single-use tokens in the same dictionary -- to be used when color
tokenizing.

Revision 1.33  1998/07/30 22:39:40  aber
aber on Thu Jul 30 15:39:40 PDT 1998

tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Prototype updates for signed-ness.

Revision 1.32  1998/04/14 01:34:57  aber
aber on Mon Apr 13 18:33:22 PDT 1998

tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Bug fix when read-writing a TIFF-FX 2.0 file and storing multiple annotations
in the file.

Revision 1.31  1998/02/09 20:21:27  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Signed/unsigned casting to make MSVC happy.

Revision 1.30  1998/01/22 00:23:52  aber
aber on Wed Jan 21 16:23:52 PST 1998

tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Allow a 0-byte author block.

Revision 1.29  1998/01/16 00:41:20  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Retain annotation table header file location, for rewrite.

Revision 1.28  1997/11/19 22:30:44  aber
aber on Wed Nov 19 14:30:43 PST 1997

tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Used right error code.

Revision 1.27  1997/11/19 21:42:57  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Allow deleting all annotations, but retaining the annotation
table header.

Revision 1.26  1997/11/19 17:54:31  aber
aber on Wed Nov 19 09:53:58 PST 1997

tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Avoid compiler complaint, it's smart, but not TOO smart -- the
complaint is unwarranted.

Revision 1.25  1997/11/18 23:43:39  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Add bytesPerEntry field in annotation header. For now there's some
code to accept the absense of this entry, but once Pagis ships we'll
take it back out.
Streamline the reading of the annotation table -- read into a single
array, then remove from the array as needed. Allows for using the
bytesPerEntry header field, and only extracts the bytes that are
supposed to be in the file version. This should make it both backwards
compatible and forward extendible.
Fixed a byte swapping problem when reading annotation type (TIFF-FX 2.0
only), that doesn't really appear since byte swapping hasn't been
required yet.

Revision 1.24  1997/11/14 02:09:00  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Use signed value for reference page; clarify that it's a page offset,
not an ordinal. Store as ordinal (was only reading as ordinal).

Revision 1.23  1997/11/01 20:57:58  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Add strip buffering for wavelet compression.

Revision 1.22  1997/10/24 21:38:56  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: More exhaustive checking on calling sequence.

Revision 1.21  1997/10/16 16:15:50  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Remove carriage-returns (left over from PC).

Revision 1.20  1997/10/15 22:15:41  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Allow reading the page table header without reading the page table itself.

Revision 1.19  1997/08/29 17:36:57  aber
aber on Fri Aug 29 10:36:53 PDT 1997

tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Fix documentation.

Revision 1.18  1997/08/29 17:28:05  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.17  1997/08/15 23:48:01  aber
aber on Fri Aug 15 16:48:00 PDT 1997

tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Collective decision by Dan Huttenlocher, William Rucklidge, and I
to remove the token dictionary ID from the symbol library table
entry, since the offset is typically used, and the ID does nothing
for us.

Revision 1.16  1997/08/15 00:10:05  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Slight code cleanup.

Revision 1.15  1997/08/07 21:27:47  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Make provisions for a 4-byte dictionary block size, instead of 2-bytes
spelled out in Williams table. Also make provisions for multiple
dictionaries per page.

Revision 1.14  1997/07/24 00:38:28  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

When reading a mask image, mark parent as having one.
Fix logic to keep track of subimages in order.

Revision 1.13  1997/07/14 23:23:30  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Enforce TIFF strip size during lower level image reads.

Revision 1.12  1997/07/10 22:01:26  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Support the ID within TIFF-FX 3.0, fix multiple-table reads.

Revision 1.11  1997/07/02 01:06:03  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Fix return values for no support vs. not a TIFF-FX.

Revision 1.10  1997/06/26 22:28:01  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

Return a new error parameter if the file is not a TIFF-FX.

Revision 1.9  1997/06/17 01:29:28  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: TIFF-FX 2.0 support. Annotation bug fix.

Revision 1.8  1997/06/10 01:26:50  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:10  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:14  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:20  aber
tfxrtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxrtfx.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:42  aber
Initial revision

*/
