/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxrtiff.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxrtiff.c:	READ TIFF source
 *
 *	TIFF read-only routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <string.h>

#ifndef NOIPCORE
#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tfxrtiff.c,v $; $Revision: 1.3 $; $Date: 2003/02/11 01:31:42 $");

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif
#endif	/* NOIPCORE */

#include "tfximage.h"
#include "tfx.h"
#include "tfxrtiff.h"
#include "tfxtiff.h"
#include "tfxrpriv.h"

/* Data type sizes (in bytes) */
char TfxFieldTypeSize[] = {
    0,		/*  0 - invalid type */
	1,	/*  1 - byte */
	1,	/*  2 - ASCII */
	2,	/*  3 - short (unsigned short) */
	4,	/*  4 - long (unsigned long) */
	8,	/*  5 - rational (unsigned long / unsigned long) */
	1,	/*  6 - sbyte (char) */
	1,	/*  7 - undefined */
	2,	/*  8 - sshort (short) */
	4,	/*  9 - slong (long) */
	8,	/* 10 - srational (long / long) */
	4,	/* 11 - float */
	8,	/* 12 - double */
	4	/* 13 - IFD (long file offset) */
};


/* Data type sub-element sizes (in bytes) */
char TfxFieldTypeElementSize[] = {
    0,		/*  0 - invalid type */
	1,	/*  1 - byte */
	1,	/*  2 - ASCII */
	2,	/*  3 - short (unsigned short) */
	4,	/*  4 - long (unsigned long) */
	4,	/*  5 - rational (unsigned long / unsigned long) */
	1,	/*  6 - sbyte (char) */
	1,	/*  7 - undefined */
	2,	/*  8 - sshort (short) */
	4,	/*  9 - slong (long) */
	4,	/* 10 - srational (long / long) */
	4,	/* 11 - float */
	8,	/* 12 - double */
	4	/* 13 - IFD (long file offset) */
};


/* Short swap: BA -> AB */
void
wp_TfxSwaps(Int8 *x)
{
    Int8   tmp;
    Int8   *a = x;

    tmp = a[1];
    a[1]= a[0];
    a[0]= tmp;
}


/* Long swap: DCBA -> ABCD */
void
wp_TfxSwapl(Int8 *x)
{
    Int8  tmp;
    Int8   *a = (Int8 *)x;
    
    tmp = a[0];
    a[0]= a[3];
    a[3]= tmp;
    
    tmp = a[1];
    a[1]= a[2];
    a[2]= tmp;
}


/* Double long swap: HGFE DCBA -> ABCD EFGH */
void
wp_TfxSwapd(Int8 *x)
{
    Int8  tmp;
    Int8   *a = (Int8 *)x;
    
    tmp = a[0];
    a[0]= a[7];
    a[7]= tmp;
    
    tmp = a[1];
    a[1]= a[6];
    a[6]= tmp;
    
    tmp = a[2];
    a[2]= a[5];
    a[5]= tmp;
    
    tmp = a[3];
    a[3]= a[4];
    a[4]= tmp;
}


/* Copy the word in 'src' to 'dst', swapping if necessary */
ErrCode CDECL
wp_TifFXConvcopy(
	 void *dst,
	 const void *src,
	 UInt16 type,   /* See "Data Types" in tifffields.h */
	 UInt32 count,
	 UInt8 swap
)
{
    UInt32 bytes = TfxFieldTypeSize[type] * count;
    UInt32 index;
    UInt16 *sptr;
    UInt32 *lptr;
    double *dptr;
    PROLOGUE;

    /* If there's nothing to do, do exactly that */
    if (count == 0 || src == NULL) {
	NORMAL_RETURN;
    }
    
    /* Copy the data */
    memcpy(dst, src, bytes);

    if (swap) {

	/* Adjust the count for the different in the type size, and the type element size */
	count *= TfxFieldTypeSize[type] / TfxFieldTypeElementSize[type];
	switch (TfxFieldTypeElementSize[type]) {
	case 2:
	    sptr = (UInt16 *)dst;
	    for (index = 0; index < count; index++) {
		TFX_SWAPS(sptr);
		sptr++;
	    }
	    break;
	case 4:
	    lptr = (UInt32 *)dst;
	    for (index = 0; index < count; index++) {
		TFX_SWAPL(lptr);
		lptr++;
	    }
	    break;
	case 8:
	    dptr = (double *)dst;
	    for (index = 0; index < count; index++) {
		TFX_SWAPD(dptr);
		dptr++;
	    }
	    break;
	}
    }
    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Check to see if the file is a valid TIFF
 *
 * LongDescription:::
 *  wp_TifFXIsTiff() checks for the necessary header information to identify the
 *  file as a TIFF file.
 *
 * Input:::
 *  pTfxFile::			Pointer to file handle
 *
 * Output:::
 *  pdwFirstIfd::		Pointer to first IFD offset
 *
 * Returns:::
 *  IO_NOTTIFF::		File is not a TIF
 *  IO_NOERROR::		File is a supported TIF
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXIsTiff(
    TFX_FILE *pTfxFile,
    UInt32 *pdwFirstIfd
)
{
    TFX_TIFFHEADER header;
    PROLOGUE;

    /* Initialize */
    *pdwFirstIfd = 0;

    /* Check the TIFF header */
    TFX_SEEK(pTfxFile, 0UL);
    TFX_READ(pTfxFile, &header.byteOrder, sizeof(header.byteOrder));
    TFX_READ(pTfxFile, &header.version, sizeof(header.version));
    TFX_READ(pTfxFile, &header.firstIfd, sizeof(header.firstIfd));
    
    /* Check for a supported version */
    if (header.byteOrder != MM && header.byteOrder != II) {
	return IO_NOTTIFF;
    }
    TFX_FILETOKEN(pTfxFile)->bSwapBytes = header.byteOrder != TIF_NATIVEORDER;
    TFX_SWAPSHORT(pTfxFile, &header.version);
    TFX_SWAPLONG(pTfxFile, &header.firstIfd);
    if (header.version != 42) {
	return IO_NOTTIFF;
    }

    /* TIFF defines: a first page must exist. However, TIFF-FX must provide
     * 0-page files. If we were to be clear about this, we would do
     * the following: */
    /* XERR_GEN_CLEANUP((header.firstIfd < 8), IO_TIFFEMPTY); */

    *pdwFirstIfd = header.firstIfd;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Free the IFD information allocated
 *
 * LongDescription:::
 *  wp_TifFXFieldCleanup() frees any miscellaneous data allocated for an IFD.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  fieldList::			TIFF field structure to clean
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXFieldCleanup(
    TFX_IMG *tfxImg,
    TFX_TIFFIELDS *fieldList
)
{
    TFX_INSTANCE *instData;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);
    
    instData = TFX_IMGINST(tfxImg);

    /* Check for memory blocks */
    if (fieldList->BitsPerSample) {
	TFX_FREE(fieldList->BitsPerSample);
    }
    if (fieldList->DateTime) {
	TFX_FREE(fieldList->DateTime);
    }
    if (fieldList->DocumentName) {
	TFX_FREE(fieldList->DocumentName);
    }
    if (fieldList->HostComputer) {
	TFX_FREE(fieldList->HostComputer);
    }
    if (fieldList->ImageDescription) {
	TFX_FREE(fieldList->ImageDescription);
    }
    if (fieldList->Software) {
	TFX_FREE(fieldList->Software);
    }
    if (fieldList->StripOffsets) {
	TFX_FREE(fieldList->StripOffsets);
    }
    if (fieldList->StripByteCounts) {
	TFX_FREE(fieldList->StripByteCounts);
    }
    if (fieldList->StripRowCounts) {
	TFX_FREE(fieldList->StripRowCounts);
    }
    if (fieldList->TargetPrinter) {
	TFX_FREE(fieldList->TargetPrinter);
    }
    if (fieldList->TileOffsets) {
	TFX_FREE(fieldList->TileOffsets);
    }
    if (fieldList->TileByteCounts) {
	TFX_FREE(fieldList->TileByteCounts);
    }
    if (fieldList->YCbCrCoefficients) {
	TFX_FREE(fieldList->YCbCrCoefficients);
    }
    if (fieldList->AnnotationOffsets) {
	TFX_FREE(fieldList->AnnotationOffsets);
    }
    if (fieldList->Colormap) {
	TFX_FREE(fieldList->Colormap);
    }
    if (fieldList->SubIFDs) {
	TFX_FREE(fieldList->SubIFDs);
    }
    if (fieldList->ImageLayerContent.imageCounts) {
	TFX_FREE(fieldList->ImageLayerContent.imageCounts);
    }
    if (fieldList->RenderingHints) {
	TFX_FREE(fieldList->RenderingHints);
    }
    
    NORMAL_RETURN;
}


#if 0
static long
FieldLong(
    TFX_TIFFFIELDENTRY *fieldEntry,
    int swap
)
{
    long lval = 0L;
    
    switch (fieldEntry->type) {
    case IFDBYTE:
    case IFDASCII:
    case IFDSHORT:
	break;
    case IFDLONG:
	if (fieldEntry->len == 1) {
	    lval = fieldEntry->offset;
	}
	break;
    case IFDRATIONAL:
    case IFDSBYTE:
    case IFDUNDEFINED:
    case IFDSSHORT:
    case IFDSLONG:
    case IFDSRATIONAL:
    case IFDFLOAT:
    case IFDDOUBLE:
	break;
    }
    
    if (swap) {
	TFX_SWAPL(&lval);
    }
    return lval;
}
#endif


/* Stuff the field data into it's destination.  If the data is in the
 * field offset, grab it.  Do appropriate byte swapping.  Also adjust
 * the data for type content (i.e. byte data but stuffed into longs,
 * etc.).
 */
static int
wp_TifFXReadFieldData(
    TFX_IMG *tfxImg,
    TFX_TIFFFIELDENTRY *fieldEntry,
    void *data,
    UInt32 dstTypeBytes,
    int swap
)
{
    TFX_FILE *tfxFile = TFX_IMGFILE(tfxImg);
    long readBytes = TfxFieldTypeSize[fieldEntry->type] * fieldEntry->len;
    long elementBytes = TfxFieldTypeElementSize[fieldEntry->type];
    long numElements = readBytes / elementBytes;
    int elementsLeft;
    PROLOGUE;

    /* Handle the extraction of data in three sections:
     * 1. Get the data, store in output:
     *    a. If out is a single 4-byte type, copy directly from the field.
     *    b. If out is otherwise 4 bytes or less:
     *       - undo the swap previously done to the 4 bytes
     *       - copy to the new output
     *    c. Read from file directly.
     * 2. Swap, if necessary.
     * 3. If output type size is different than input type size, adjust
     *    for the size difference.
     */

    if (readBytes == 4 && elementBytes == 4) {
	memcpy(data, &fieldEntry->offset, readBytes);
    } else if (readBytes <= 4) {
	
	/* Data found in offset entry element */
	if (swap) {
	    TFX_SWAPL(&fieldEntry->offset);	/* swap back */
	}

	/* Copy from field entry */
	memcpy(data, &fieldEntry->offset, readBytes);

	/* Swap if necessary */
	if (swap) {
	    switch (elementBytes) {
	    case 1:
		break;
	    case 2:
		{
		    UInt16 *sptr = (UInt16 *)data;
		    elementsLeft = numElements;
		    while (elementsLeft--) {
			TFX_SWAPS(sptr);
			sptr++;
		    }
		}
		break;
	    case 4:
		TFX_SWAPL(data);
		break;
	    }
	}
	/* Data already swapped when placed in fieldEntry->offset */
	
    } else {	/* bytes > 4 */
	
	/* Data is still in the file, extract */
	TFX_SEEK(tfxFile, fieldEntry->offset);
	TFX_READ(tfxFile, data, readBytes);

    	if (swap) {
	    switch (elementBytes) {
	    case 1:	/* No swapping required */
		break;
		
	    case 2:
		{
		    UInt16 *sptr = (UInt16 *)data;

		    elementsLeft = numElements;
		    while (elementsLeft--) {
			TFX_SWAPS(sptr++);
		    }
		}
		break;
		
	    case 4:
		{
		    UInt32 *lptr = (UInt32 *)data;

		    elementsLeft = numElements;
		    while (elementsLeft--) {
			TFX_SWAPL(lptr++);
		    }
		}
		break;

	    case IFDDOUBLE:
		{
		    double *dptr = (double *)data;
		
		    elementsLeft = numElements;
		    while (elementsLeft--) {
			TFX_SWAPD(dptr++);
		    }
		}
		break;

	    default:
		/* This is probably bad */
		XERR_GEN_CLEANUP(1, IO_NOSUPPORT);
	    }
	}
    }
    
    /* Handle mismatch in the size of the data types.
     * Note: what is done here assumes that sign-extension will take place
     * (assumes signed values instead of unsigned values).
     * THIS SHOULD BE FIXED!!!!!!!!!!
     */
    switch (elementBytes) {
    case 1:	/* Source is 1-byte elements */
	/* Distinguish signed from unsigned, so that the sign bit extends correctly */
	switch (fieldEntry->type) {
	case IFDASCII:
	case IFDBYTE:
	    switch ((int)dstTypeBytes) {
	    case 1:
		break;
	    case 2:		/* Destination is 2-byte elements */
		{
		    UInt8 *inptr = (UInt8 *)data + numElements - 1;
		    Int16 *outptr = (Int16 *)data + numElements - 1;
		    
		    elementsLeft = numElements;
		    while (elementsLeft--) {
			*outptr-- = (Int16)*inptr--;
		    }
		}
		break;
	    case 4:		/* Destination is 4-byte elements (should be long) */
		{
		    UInt8 *inptr = (UInt8 *)data + numElements - 1;
		    Int32 *outptr = (Int32 *)data + numElements - 1;
		    
		    elementsLeft = numElements;
		    while (elementsLeft--) {
			*outptr-- = (Int32)*inptr--;
		    }
		}
		break;
	    case 8:		/* Destination is 8-byte elements (doubles) */
		{
		    UInt8 *inptr = (UInt8 *)data + numElements - 1;
		    double *outptr = (double *)data + numElements - 1;
		    
		    elementsLeft = numElements;
		    while (elementsLeft--) {
			*outptr-- = (double)*inptr--;
		    }
		}
		break;
	    default:
		XERR_GEN_CLEANUP(1, IO_NOSUPPORT);	/* This is probably weird */
	    }
	    break;
	case IFDSBYTE:
	    switch ((int)dstTypeBytes) {
	    case 1:
		break;
	    case 2:		/* Destination is 2-byte elements */
		{
		    Int8 *inptr = (Int8 *)data + numElements - 1;
		    Int16 *outptr = (Int16 *)data + numElements - 1;
		    
		    elementsLeft = numElements;
		    while (elementsLeft--) {
			*outptr-- = (Int16)*inptr--;
		    }
		}
		break;
	    case 4:		/* Destination is 4-byte elements (should be long) */
		{
		    Int8 *inptr = (Int8 *)data + numElements - 1;
		    Int32 *outptr = (Int32 *)data + numElements - 1;
		    
		    elementsLeft = numElements;
		    while (elementsLeft--) {
			*outptr-- = (Int32)*inptr--;
		    }
		}
		break;
	    case 8:		/* Destination is 8-byte elements (doubles) */
		{
		    Int8 *inptr = (Int8 *)data + numElements - 1;
		    double *outptr = (double *)data + numElements - 1;
		    
		    elementsLeft = numElements;
		    while (elementsLeft--) {
			*outptr-- = (double)*inptr--;
		    }
		}
		break;
	    default:
		XERR_GEN_CLEANUP(1, IO_NOSUPPORT);	/* This is probably weird */
	    }
	    break;
	}
	break;
    case 2:	/* Source is 2-byte elements */
	/* Distinguish signed from unsigned, so that the sign bit extends correctly */
	switch (fieldEntry->type) {
	case IFDSHORT:
	    switch ((int)dstTypeBytes) {
	    case 2:
		break;
	    case 4:		/* Destination is 4-byte elements (should be long) */
	    case 8:	    	/* RATIONAL */
		{
		    UInt16 *inptr = (UInt16 *)data + numElements - 1;
		    UInt32 *outptr = (UInt32 *)data + numElements - 1;
		    
		    elementsLeft = numElements;
		    while (elementsLeft--) {
			*outptr-- = (UInt32)*inptr--;
		    }
		}
		break;
	    default:
		XERR_GEN_CLEANUP(1, IO_NOSUPPORT);	/* This is probably weird */
	    }
	    break;
	case IFDSSHORT:
	    switch ((int)dstTypeBytes) {
	    case 2:
		break;
	    case 4:		/* Destination is 4-byte elements (should be long) */
	    case 8:		/* RATIONAL */
		{
		    Int16 *inptr = (Int16 *)data + numElements - 1;
		    Int32 *outptr = (Int32 *)data + numElements - 1;

		    elementsLeft = numElements;
		    while (elementsLeft--) {
			*outptr-- = (Int32)*inptr--;
		    }
		}
		break;
	    default:
		XERR_GEN_CLEANUP(1, IO_NOSUPPORT);	/* This is probably weird */
	    }
	}
	break;
    case 4:	/* Source is 4-byte elements */
	switch ((int)dstTypeBytes) {
	case 4:
	case 8:
	    break;
#if 0
	case 8:		/* Destination is 2-byte elements (doubles) */
	    /* For now, no distinction is made between the real destination
	     * type of RATIONAL and DOUBLE.  If doubles get used eventually,
	     * then this is how it would be implemented:
	     */
	    {
		/* Assume this is floats in, but doubles out */
		float *inptr = (float *)data + numElements - 1;
		double *outptr = (double *)data + numElements - 1;

		elementsLeft = numElements;
		while (elementsLeft--) {
		    *outptr-- = (double)*inptr--;
		}
	    }
#endif
	default:
	    XERR_GEN_CLEANUP(1, IO_NOSUPPORT);	/* This is probably bad */
	}
	break;
    }
    
    NORMAL_RETURN;

 ERR_CLEANUP:
    
    ERR_RETURN;
}


/* Special macro to automatically pass in the size of the destination
 */
#define ReadFieldData(field) \
    wp_TifFXReadFieldData(tfxImg, fieldEntry, &(field), sizeof(field), swap);



static int CDECL
wp_TifFXParseField(
    TFX_IMG *tfxImg,
    TFX_TIFFFIELDENTRY *fieldEntry,
    TFX_TIFFIELDS *fieldList,
    UInt32 fieldOffset,
    int swap
)
{
    TFX_INSTANCE *instData = TFX_IMGINST(tfxImg);
    TFX_DOC *tfxDoc = tfxImg->tfxPage->tfxDoc;
    TFX_TIFGPFIELDS *gpFieldList;
    PROLOGUE;
    
    gpFieldList = &fieldList->gpIFD;

    switch (fieldEntry->tagID) {
    case NEWSUBFILETYPE:
	ReadFieldData(fieldList->NewSubfileType);
	break;
    case SUBFILETYPE:
	break;
    case IMAGEWIDTH:
	ReadFieldData(fieldList->ImageWidth);
	break;
    case IMAGELENGTH:
	ReadFieldData(fieldList->ImageLength);
	break;
    case BITSPERSAMPLE:
	if (fieldEntry->len > 0) {
	    IO_CHECKMALLOC(fieldList->BitsPerSample =
		TFX_MALLOC(fieldEntry->len * TfxFieldTypeSize[fieldEntry->type]));
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->BitsPerSample,
		sizeof(*fieldList->BitsPerSample), swap);

	    /* Mark as information found */
	    fieldList->found.BitsPerSample = 1;
	}
	break;
    case COMPRESSION:
	ReadFieldData(fieldList->Compression);

	/* Mark as information found */
	fieldList->found.Compression = 1;
	break;
    case PHOTOMETRICINTERPRETATION:
	ReadFieldData(fieldList->PhotometricInterpretation);

	/* Mark as information found */
	fieldList->found.PhotometricInterpretation = 1;
	break;
    case THRESHHOLDING:
	ReadFieldData(fieldList->Thresholding);

	/* Mark as information found */
	fieldList->found.Thresholding = 1;
	break;
    case CELLWIDTH:
    case CELLLENGTH:
	break;
    case FILLORDER:
	ReadFieldData(fieldList->FillOrder);
	XERR_GEN_CLEANUP((fieldList->FillOrder != 1) && (fieldList->FillOrder != 2),
	    IO_NOSUPPORT);

	/* Mark as information found */
	fieldList->found.FillOrder = 1;
	break;
    case DOCUMENTNAME:
	if (fieldEntry->len > 0) {
	    IO_CHECKMALLOC(fieldList->DocumentName =
		TFX_MALLOC(fieldEntry->len + 1));
	    fieldList->DocumentName[fieldEntry->len] = '\0';
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->DocumentName,
		sizeof(*fieldList->DocumentName), swap);

	    /* Keep the size of it */
	    fieldList->DocumentNameLength = fieldEntry->len;

	    /* Mark as information found */
	    fieldList->found.DocumentName = 1;
	}
	break;
    case IMAGEDESCRIPTION:
	if (fieldEntry->len > 0) {
	    IO_CHECKMALLOC(fieldList->ImageDescription =
		TFX_MALLOC(fieldEntry->len + 1));
	    fieldList->ImageDescription[fieldEntry->len] = '\0';
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->ImageDescription,
		1, swap);

	    /* Keep the size of it */
	    fieldList->ImageDescriptionLength = fieldEntry->len;

	    /* Mark as information found */
	    fieldList->found.ImageDescription = 1;
	}
	break;
    case MAKE:
    case MODEL:
	break;
    case STRIPOFFSETS:
	if (fieldList->StripsPerImage == 0) {
	    fieldList->StripsPerImage = fieldEntry->len;
	} else {
	    XERR_GEN_CLEANUP(fieldList->StripsPerImage != fieldEntry->len,
		IO_BADFILEFORMAT);	    
	}
	if (fieldList->StripsPerImage > 0) {
	    IO_CHECKMALLOC(fieldList->StripOffsets =
		TFX_MALLOC(fieldList->StripsPerImage *
		sizeof(*fieldList->StripOffsets)));
	}
	wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->StripOffsets,
		       sizeof(*fieldList->StripOffsets), swap);
	break;
    case ORIENTATION:
	ReadFieldData(fieldList->Orientation);

	/* Mark as information found */
	fieldList->found.Orientation = 1;
	break;
    case SAMPLESPERPIXEL:
	ReadFieldData(fieldList->SamplesPerPixel);

	/* Mark as information found */
	fieldList->found.SamplesPerPixel = 1;
	break;
    case ROWSPERSTRIP:
	ReadFieldData(fieldList->RowsPerStrip);

	/* Mark as information found */
	fieldList->found.RowsPerStrip = 1;
	break;
    case STRIPBYTECOUNTS:
	if (fieldList->StripsPerImage == 0) {
	    fieldList->StripsPerImage = fieldEntry->len;
	} else {
	    XERR_GEN_CLEANUP(fieldList->StripsPerImage != fieldEntry->len,
		IO_BADFILEFORMAT);	    
	}
	if (fieldList->StripsPerImage > 0) {
	    IO_CHECKMALLOC(fieldList->StripByteCounts =
		TFX_MALLOC(fieldList->StripsPerImage *
		sizeof(*fieldList->StripByteCounts)));
	}
	wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->StripByteCounts,
		       sizeof(*fieldList->StripByteCounts), swap);
	break;
    case MINSAMPLEVALUE:
    case MAXSAMPLEVALUE:
	break;
    case XRESOLUTION:
	ReadFieldData(fieldList->XResolution);
	break;
    case YRESOLUTION:
	ReadFieldData(fieldList->YResolution);
	break;
    case PLANARCONFIGURATION:
	ReadFieldData(fieldList->PlanarConfiguration);

	/* Mark as information found */
	fieldList->found.PlanarConfiguration = 1;
	break;
    case PAGENAME:
	break;
    case XPOSITION:
	ReadFieldData(fieldList->XPosition);
	break;
    case YPOSITION:
	ReadFieldData(fieldList->YPosition);
	break;
    case FREEOFFSETS:
    case FREEBYTECOUNTS:
    case GRAYRESPONSEUNIT:
    case GRAYRESPONSECURVE:
	break;
    case T4OPTIONS:
	ReadFieldData(fieldList->T4Options);

	/* Mark as information found */
	fieldList->found.T4Options = 1;
	break;
    case T6OPTIONS:
	ReadFieldData(fieldList->T6Options);

	/* Mark as information found */
	fieldList->found.T6Options = 1;
	break;
    case RESOLUTIONUNIT:
	ReadFieldData(fieldList->ResolutionUnit);

	/* Mark as information found */
	fieldList->found.ResolutionUnit = 1;
	break;
    case PAGENUMBER:
	wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->PageNumber,
		       sizeof(*fieldList->PageNumber), swap);

	/* Mark as information found */
	fieldList->found.PageNumber = 1;
	break;
    case COLORRESPONSEUNIT:
    case TRANSFERFUNCTION:
	break;
    case SOFTWARE:
	if (fieldEntry->len > 0) {
	    IO_CHECKMALLOC(fieldList->Software =
		TFX_MALLOC(fieldEntry->len + 1));
	    fieldList->Software[fieldEntry->len] = '\0';
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->Software,
		sizeof(*fieldList->Software), swap);

	    /* Keep the size of it */
	    fieldList->SoftwareLength = fieldEntry->len;

	    /* Mark as information found */
	    fieldList->found.Software = 1;
	}
	break;
    case DATETIME:
	if (fieldEntry->len > 0) {
	    IO_CHECKMALLOC(fieldList->DateTime =
		TFX_MALLOC(fieldEntry->len + 1));
	    fieldList->DateTime[fieldEntry->len] = '\0';
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->DateTime,
		sizeof(*fieldList->DateTime), swap);

	    /* This is supposed to be 19, for a strict format:
	     * "YYYY:MM:DD HH:MM:SS" */
	    fieldList->DateTimeLength = fieldEntry->len;

	    /* Mark as information found */
	    fieldList->found.DateTime = 1;
	}
	break;
    case ARTIST:
    case HOSTCOMPUTER:
	if (fieldEntry->len > 0) {
	    IO_CHECKMALLOC(fieldList->HostComputer =
		TFX_MALLOC(fieldEntry->len + 1));
	    fieldList->HostComputer[fieldEntry->len] = '\0';
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->HostComputer,
		1, swap);

	    /* Keep the size of it */
	    fieldList->HostComputerLength = fieldEntry->len;

	    /* Mark as information found */
	    fieldList->found.HostComputer = 1;
	}
	break;
    case PREDICTOR:
	ReadFieldData(fieldList->Predictor);
	/* Mark as information found */
	fieldList->found.Predictor = 1;
	break;
    case WHITEPOINT:
    case PRIMARYCHROMATICITIES:
	break;
    case COLORMAP:
	if (fieldEntry->len > 0) {
	    XERR_GEN_CLEANUP((TfxFieldTypeSize[fieldEntry->type] != 2), IO_BADFILEFORMAT);
	    IO_CHECKMALLOC(fieldList->Colormap =
		TFX_MALLOC(fieldEntry->len * sizeof(*fieldList->Colormap)));
	    memset(fieldList->Colormap, 0, fieldEntry->len * sizeof(*fieldList->Colormap));
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->Colormap,
		sizeof(*fieldList->Colormap), swap);
	}
	break;
    case HALFTONEHINTS:
	break;
    case TILEWIDTH:
	ReadFieldData(fieldList->TileWidth);
	break;
    case TILELENGTH:
	ReadFieldData(fieldList->TileLength);
	break;
    case TILEOFFSETS:
	if (fieldList->TileCount == 0) {
	    fieldList->StripsPerImage = fieldList->TileCount = fieldEntry->len;
	} else if (fieldList->TileCount != fieldEntry->len) {
	    __error__ = IO_BADFILEFORMAT;
	    goto ERR_CLEANUP;
	}
	IO_CHECKMALLOC(fieldList->TileOffsets =
		     TFX_MALLOC(fieldEntry->len * TfxFieldTypeSize[fieldEntry->type]));
	wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->TileOffsets,
		       sizeof(*fieldList->TileOffsets), swap);
	break;
    case TILEBYTECOUNTS:
	if (fieldList->TileCount == 0) {
	    fieldList->StripsPerImage = fieldList->TileCount = fieldEntry->len;
	} else if (fieldList->TileCount != fieldEntry->len) {
	    __error__ = IO_BADFILEFORMAT;
	    goto ERR_CLEANUP;
	}
	IO_CHECKMALLOC(fieldList->TileByteCounts =
		     TFX_MALLOC(fieldEntry->len * TfxFieldTypeSize[fieldEntry->type]));
	wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->TileByteCounts,
		       sizeof(*fieldList->TileByteCounts), swap);
	break;
    case SUBIFDS:
	if (fieldEntry->len > 0) {
	    fieldList->SubIFDCount = fieldEntry->len;
	    IO_CHECKMALLOC(fieldList->SubIFDs =
		TFX_MALLOC(fieldEntry->len * TfxFieldTypeSize[fieldEntry->type]));
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->SubIFDs,
		sizeof(*fieldList->SubIFDs), swap);
	    if (fieldEntry->len <= 1) {
		tfxImg->fieldDataLocation.SubIFDs = fieldOffset + 8;
	    } else {
		tfxImg->fieldDataLocation.SubIFDs = fieldEntry->offset;
	    }
	}
	break;
    case INKSET:
    case INKNAMES:
    case NUMBEROFINKS:
    case DOTRANGE:
	break;
    case TARGETPRINTER:
	if (fieldEntry->len > 0) {
	    IO_CHECKMALLOC(fieldList->TargetPrinter =
		TFX_MALLOC(fieldEntry->len + 1));
	    fieldList->TargetPrinter[fieldEntry->len] = '\0';
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->TargetPrinter,
		sizeof(*fieldList->TargetPrinter), swap);

	    /* Keep the size of it */
	    fieldList->TargetPrinterLength = fieldEntry->len;

	    /* Mark as information found */
	    fieldList->found.TargetPrinter = 1;
	}
	break;
    case EXTRASAMPLES:
    case SAMPLEFORMAT:
    case SMINSAMPLEVALUE:
    case SMAXSAMPLEVALUE:
    case TRANSFERRANGE:
    case JPEGPROC:
    case JPEGINTERCHANGEFORMAT:
    case JPEGINTERCHANGEFORMATLENGTH:
    case JPEGRESTARTINTERVAL:
    case JPEGLOSSLESSPREDICTORS:
    case JPEGPOINTTRANSFORMS:
    case JPEGQTABLES:
    case JPEGDCTABLES:
    case JPEGACTABLES:
	break;
    case YCBCRCOEFFICIENTS:
	if (fieldEntry->len > 0) {
	    IO_CHECKMALLOC(fieldList->YCbCrCoefficients =
		TFX_MALLOC(fieldEntry->len * TfxFieldTypeSize[fieldEntry->type]));
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->YCbCrCoefficients,
		sizeof(*fieldList->YCbCrCoefficients), swap);

	    /* Keep the size of it */
	    fieldList->YCbCrCoefficientsValues = fieldEntry->len;

	    /* Mark as information found */
	    fieldList->found.YCbCrCoefficients = 1;
	}
	break;
    case CHROMASUBSAMPLING:
	{
	    UInt32 sampling[2];
	    ReadFieldData(sampling);
/*	    XERR_GEN_CLEANUP(sampling[0] != 2, IO_NOSUPPORT); */
/*	    XERR_GEN_CLEANUP(sampling[1] != 2, IO_NOSUPPORT); */

	    /* Mark as information found */
	    fieldList->found.ChromaSubSampling = 1;
	}
	break;
    case CHROMAPOSITIONING:
	{
	    ReadFieldData(fieldList->ChromaPositioning);
	    if (fieldList->ChromaPositioning != 1) {
		__error__ = IO_WARN_NOSUPPORT;
	    }

	    /* Mark as information found */
	    fieldList->found.ChromaPositioning = 1;	
	}
	break;
    case REFERENCEBLACKWHITE:
	break;
    case STRIPROWCOUNTS:
	if (fieldEntry->len > 0) {
	    IO_CHECKMALLOC(fieldList->StripRowCounts =
		TFX_MALLOC(fieldEntry->len * TfxFieldTypeSize[fieldEntry->type]));
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->StripRowCounts,
		sizeof(*fieldList->StripRowCounts), swap);

	    /* Keep the number of strips around */
	    tfxImg->fields.StripsPerImage = fieldEntry->len;

	    /* Store the location of the field */
	    tfxImg->fieldLocation.StripRowCounts = fieldOffset + 8;

	    /* Retain the position of this field value */
	    if (fieldEntry->len <= 1) {
		tfxImg->fieldDataLocation.StripRowCounts = fieldOffset + 8;
	    } else {
		tfxImg->fieldDataLocation.StripRowCounts = fieldEntry->offset;
	    }
	}
	break;
    case COPYRIGHT:
	break;
    case JPEGTABLES:
	if (fieldEntry->len > 4) {  /* Must be at least 4 bytes */
	    UInt8 *tp;			/* Tables stream pointer */
	    UInt8 *eoipos;		/* Where EOI should occur */
	    UInt32 bytes;		/* Bytes in segment */
	    
	    IO_CHECKMALLOC(tfxImg->qtableBuf = TFX_MALLOC(fieldEntry->len));
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, tfxImg->qtableBuf, 1, swap);

	    /* Keep the size of it */
	    tfxImg->qtableBufsize = fieldEntry->len;

	    /* Make sure the tables starts with an SOI */
	    tp = tfxImg->qtableBuf;
	    XERR_GEN_CLEANUP(*tp != 0xff, IO_BADFILEFORMAT);
	    tp++;
	    XERR_GEN_CLEANUP(*tp != 216 /* SOI */, IO_BADFILEFORMAT);
	    tp++;

	    /* Find out where the EOI is in the tables stream (might not be
	     * the last byte in the stream) */
	    eoipos = tfxImg->qtableBuf + tfxImg->qtableBufsize - 2;	/* Expected position */
	    while (tp <= eoipos) {
		/* Is it a marker? */
		XERR_GEN_CLEANUP(*tp != 0xff, IO_BADFILEFORMAT);
		
		/* Is it the EOI */
		if (tp[1] == 0xD9) break;

		tp += 2;	/* Skip just the marker */
		
		/* Look for markers that don't have segments */
		switch (tp[1]) {
		case 0x01:	/* TEM */
		    break;
		default:
		    bytes = (tp[0] << 8) | tp[1];
		    XERR_GEN_CLEANUP(bytes < 2, IO_BADFILEFORMAT);
		    XERR_GEN_CLEANUP(tp + bytes > eoipos, IO_BADFILEFORMAT);
		    tp += bytes;
		    break;
		}
	    }
	    
	    /* Adjust the byte count for where the EOI was found */
	    tfxImg->qtableBufsize = tp + 2 - tfxImg->qtableBuf;
	    
	}
	break;
    case GLOBALPARAMETERSIFD:
	/* Read the GPIFD */
	{
	    UInt32 nextGPIFD = fieldEntry->offset;
	    UInt16 numFields = fieldList->fieldCount;

	    while (nextGPIFD) {
		XERR_PASS_CLEANUP(wp_TifFXReadFields(tfxImg, nextGPIFD, &nextGPIFD));
	    }

	    /* Keep the field count of the IFD and the GPIFD separate */
	    fieldList->gpIFD.fieldCount = fieldList->fieldCount;
	    fieldList->fieldCount = numFields;

	    /* Reflect that the GPIFD is there */
	    tfxImg->tfxPage->tfxDoc->seqStatus |= TFX_DOC_SEQ_GPEXISTS;
	}
	break;
    case PROFILETYPE:
	ReadFieldData(gpFieldList->ProfileType);

	/* Mark as information found */
	fieldList->found.ProfileType = 1;
	break;
    case FAXPROFILE:
	ReadFieldData(gpFieldList->FaxProfile);

	/* Mark as information found */
	fieldList->found.FaxProfile = 1;
	break;
    case CODINGMETHODS:
	ReadFieldData(gpFieldList->CodingMethods);

	/* Remember the location */
	tfxDoc->codingMethodsLocation = fieldOffset + 8;
	break;
    case VERSIONYEAR:
	ReadFieldData(gpFieldList->VersionYear);
	break;
    case MODENUMBER:
	ReadFieldData(gpFieldList->ModeNumber);

	/* Remember the location */
	tfxDoc->modeNumberLocation = fieldOffset + 8;
	break;
    case DECODE:
	{
	    switch (fieldEntry->type) {
	    case IFDSRATIONAL:
		XERR_GEN_CLEANUP(fieldEntry->len != 2 && fieldEntry->len != 6,
		    IO_BADFILEFORMAT);

		wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->Decode,
		    sizeof(*fieldList->Decode), swap);
		fieldList->DecodeValues = fieldEntry->len;
		break;
	    default:
		XERR_GEN_CLEANUP(1, IO_BADFILEFORMAT);	    /* We don't understand this, and could just ignore it */
	    }

	    /* Mark as information found */
	    fieldList->found.Decode = 1;
	}
	break;
    case IMAGEBASECOLOR:
	{
	    switch (fieldEntry->type) {
	    case IFDSHORT:
		XERR_GEN_CLEANUP(fieldEntry->len != 1 && fieldEntry->len != 3
		    && fieldEntry->len != 4,
		    IO_BADFILEFORMAT);

		wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->ImageBaseColor,
		    sizeof(*fieldList->ImageBaseColor), swap);
		fieldList->ImageBaseColorValues = fieldEntry->len;
		break;
	    default:
		XERR_GEN_CLEANUP(1, IO_BADFILEFORMAT);	    /* We don't understand this, and could just ignore it */
	    }

	    /* Mark as information found */
	    fieldList->found.ImageBaseColor = 1;
	}
	break;
    case T82OPTIONS:
	ReadFieldData(fieldList->T82Options);
	break;
    case MULTIPROFILES_UO:
    case MULTIPROFILES:
	ReadFieldData(gpFieldList->MultiProfiles);

	/* Propogate to the document */
	tfxDoc->multiProfiles = gpFieldList->MultiProfiles;

	/* Store the location of the MultiProfiles field data offset */
	tfxDoc->multiProfilesLocation = fieldOffset + 8;
	break;
    case T88OPTIONS_UO:
    case T88OPTIONS:
	ReadFieldData(fieldList->T88Options);

	/* Mark as information found */
	fieldList->found.T88Options = 1;

	/* Store the location of the T88Options field data offset */
	tfxImg->fieldDataLocation.T88Options = fieldOffset + 8;
	break;
    case SHAREDRESOURCES_UO:
    case SHAREDRESOURCES:
	/* This should appear inside a GPIFD.
	 * For now, just keep the location of the field, later we'll actually
	 * read the shared resource table. */
	tfxImg->tfxPage->tfxDoc->sharedResourcesLocation = fieldOffset;
	tfxImg->tfxPage->tfxDoc->firstResourceTableLocation = fieldEntry->offset;
	break;
    case TIFFFXEXTENSIONS:
	ReadFieldData(gpFieldList->TiffFXExtensions);

	/* Store the location of the TIFF-FX Extensions data offset */
	tfxDoc->tifffxExtensionsLocation = fieldOffset + 8;
	break;
    case RENDERINGHINTS:
	if (fieldEntry->len > 0) {
	    IO_CHECKMALLOC(fieldList->RenderingHints =
		TFX_MALLOC(fieldEntry->len + 1));
	    fieldList->RenderingHints[fieldEntry->len] = '\0';
	    wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->RenderingHints,
		1, swap);

	    /* Keep the size of it */
	    fieldList->RenderingHintsLength = fieldEntry->len;

	    /* Mark as information found */
	    fieldList->found.RenderingHints = 1;
	}
	break;
    case RENDERINGHINTSIFD:
	ReadFieldData(fieldList->RenderingHintsIFD);

	/* Mark as information found */
	if (fieldList->RenderingHintsIFD >= 8) {
	    fieldList->found.RenderingHintsIFD = 1;
	    
	    /* There's at least one hints image.  We should later traverse the
	     * NextIFD of this image. */
	    tfxImg->imgInfo.dwHintsImages = 1;
	}
	break;
    case IFDBASECOLOR:
	{
	    switch (fieldEntry->type) {
	    case IFDBYTE:
	    case IFDSHORT:
	    case IFDSLONG:
		fieldList->IFDBaseColorType = fieldEntry->type;
		XERR_GEN_CLEANUP(fieldEntry->len != fieldList->SamplesPerPixel,
		    IO_BADFILEFORMAT);

		wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->IFDBaseColor,
		    sizeof(*fieldList->IFDBaseColor), swap);
		fieldList->IFDBaseColorValues = fieldEntry->len;
		break;
	    default:
		XERR_GEN_CLEANUP(1, IO_BADFILEFORMAT);	    /* We don't understand this, and could just ignore it */
	    }

	    /* Mark as information found */
	    fieldList->found.IFDBaseColor = 1;
	}
	break;
    case ANNOTATIONOFFSETS:
	if (fieldEntry->len > 0) {
	    fieldList->AnnotationCount = (UInt16)fieldEntry->len;
	    if (fieldList->AnnotationCount > 0) {
		IO_CHECKMALLOC(fieldList->AnnotationOffsets =
		    TFX_MALLOC(fieldList->AnnotationCount *
		    sizeof(*fieldList->AnnotationOffsets)));
		wp_TifFXReadFieldData(tfxImg, fieldEntry, fieldList->AnnotationOffsets,
		    sizeof(*fieldList->AnnotationOffsets), swap);
		tfxImg->fieldLocation.AnnotationOffsets = fieldOffset + 4;
		if (fieldEntry->len <= 1) {
		    tfxImg->fieldDataLocation.AnnotationOffsets = fieldOffset + 8;
		} else {
		    tfxImg->fieldDataLocation.AnnotationOffsets = fieldEntry->offset;
		}
	    } /* Annotations attached */

	    /* Mark as information found */
	    fieldList->found.AnnotationOffsets = 1;
	}
	break;
    case MASKIFDS:
	ReadFieldData(fieldList->MaskIFDs);
	tfxImg->fieldLocation.MaskIFDs = fieldOffset + 8;
	break;
    case IMAGELAYER:
	ReadFieldData(fieldList->ImageLayer);

	/* Remember the field data location */
	tfxImg->fieldDataLocation.ImageLayer = fieldEntry->offset;

	/* Check the validity of the content */
	XERR_GEN_CLEANUP(fieldEntry->len != 2, IO_BADFILEFORMAT);
	XERR_GEN_CLEANUP(fieldList->ImageLayer[0] < 1 ||
	    fieldList->ImageLayer[1] < 1, IO_BADFILEFORMAT);
	break;
#if 0
    case IMAGELAYERCONTENT:
	if (fieldEntry->len > 0) {
	    fieldList->ImageLayerContent.layersPresent = fieldEntry->len;
	    if (fieldList->ImageLayerContent.imageCounts) {
		TFX_FREE(fieldList->ImageLayerContent.imageCounts);
	    }
	    if (fieldList->ImageLayerContent.layersPresent > 0) {
		IO_CHECKMALLOC(fieldList->ImageLayerContent.imageCounts = (UInt32 *)
		    TFX_MALLOC(fieldList->ImageLayerContent.layersPresent * sizeof(UInt32)));
		ReadFieldData(fieldList->ImageLayerContent.imageCounts[0]);
	    }
	    tfxImg->fieldLocation.ImgLayerContent = fieldOffset + 8;
	    if (fieldEntry->len <= 1) {
		tfxImg->imgLayerContentDataOffset = fieldOffset + 8;
	    } else {
		tfxImg->imgLayerContentDataOffset = fieldEntry->offset;
	    }
	}
	break;
#endif
    case REFERENCEPAGE:
	ReadFieldData(gpFieldList->ReferencePage);

	/* Mark as information found */
	fieldList->found.ReferencePage = 1;
	    
	/* Store the location of the shared resource header offset */
	tfxDoc->refPageOffsetLocation = fieldOffset + 8;

	/* Reflect the write in the status */
	tfxDoc->seqStatus |= TFX_DOC_SEQ_REFPAGESET;
	break;
    } /* which field ID */
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


static int CDECL
wp_TifFXFieldList2Info(
    TFX_IMG *tfxImg,
    TFX_IMGINFO *imgInfo
)
{
    Int32 baseXres;
    Int32 baseYres;
    TFX_TIFFIELDS *fieldList = &tfxImg->fields;
/*    TFX_PAGE *tfxPage; */
    PROLOGUE;

    switch (fieldList->ResolutionUnit) {
    case CENTIMETERS:
	imgInfo->dwXResolution = (long)(TFX_RATIONAL2Float(fieldList->XResolution) / 2.53995f + 0.5f);
	imgInfo->dwYResolution = (long)(TFX_RATIONAL2Float(fieldList->YResolution) / 2.53995f + 0.5f);
	break;
    default: /* expect inches */
	imgInfo->dwXResolution = TFX_RATIONAL2Long(fieldList->XResolution);
	imgInfo->dwYResolution = TFX_RATIONAL2Long(fieldList->YResolution);
    }
    baseXres = ((TFX_IMG *)tfxImg->tfxPage->pageInfo.tfxPrimImgHandle)->
	imgInfo.dwXResolution;
    baseYres = ((TFX_IMG *)tfxImg->tfxPage->pageInfo.tfxPrimImgHandle)->
	imgInfo.dwYResolution;
    imgInfo->dwXOffset = (UInt32)(TFX_RATIONAL2Float(fieldList->XPosition) *
			      (float)baseXres + 0.5);
    imgInfo->dwYOffset = (UInt32)(TFX_RATIONAL2Float(fieldList->YPosition) *
			      (float)baseYres + 0.5);
    imgInfo->dwWidth = fieldList->ImageWidth;
    imgInfo->dwHeight = fieldList->ImageLength;
#if 0
    if (imgInfo->tfxImgType == IO_IMGTYPE_PRIMARY) {
	Int32 layer;

	tfxPage = tfxImg->tfxPage;
	for (layer = fieldList->ImageLayerContent.layersPresent; layer > 0; layer--) {
	    XERR_PASS_CLEANUP(wp_TfxIncrementImageCounts(tfxPage, layer,
		fieldList->ImageLayerContent.imageCounts[layer-1]));
	}
	if (tfxPage->tfxDoc->docInfo.version == 2) {
	    tfxPage->pageInfo.wAnnotationCount = fieldList->AnnotationCount;
	}
    }
#else
#if 1
    imgInfo->dwImgLayer = 1;
    imgInfo->dwImgOffset = 0;
    if ((fieldList->NewSubfileType & TFX_TIFFSUBFILE_MRC)
	&& fieldList->ImageLayer[0] != 0) {
	imgInfo->dwImgLayer = fieldList->ImageLayer[0];
	if (fieldList->ImageLayer[1] >= 1) {
	    imgInfo->dwImgOffset = fieldList->ImageLayer[1] - 1;
	}
    }
#else
    switch (tfxImg->tfxPage->tfxDoc->docInfo.profile) {
    case TFX_PROFILE_M:
    case TFX_PROFILE_M1:
    case TFX_PROFILE_COH_M:
    case TFX_PROFILE_COH_J:
	XERR_GEN_CLEANUP(fieldList->ImageLayer[1] < 1, IO_BADFILEFORMAT);
	imgInfo->dwImgLayer = fieldList->ImageLayer[0];
	imgInfo->dwImgOffset = fieldList->ImageLayer[1] - 1;
	break;
    default:
	imgInfo->dwImgLayer = 1;
	imgInfo->dwImgOffset = 0;
    }
#endif
#endif
/*    imgInfo->dwMaskImages = 0; */
    imgInfo->dwHintsImages = 0;
    if (fieldList->RowsPerStrip) {
	imgInfo->dwBestStripHeight = fieldList->RowsPerStrip;
    } else if (fieldList->TileCount) {
	if (fieldList->TileLength) {
	    imgInfo->dwBestStripHeight = fieldList->TileLength;
	} else {
	    imgInfo->dwBestStripHeight = 0;	/* ??? */
	}
    } else if (fieldList->StripRowCounts) {
	imgInfo->dwBestStripHeight = 0;
    } else {
	imgInfo->dwBestStripHeight = fieldList->ImageLength;
    }
    
    switch (fieldList->Compression) {
    case TFX_TIFFCOMP_NOCOMPRESSBYTE:
	imgInfo->xCompression = IO_COMP_NOC;
	break;
    case TFX_TIFFCOMP_CCITT3:
	/* comp=2: technically, this is not compliant with TIFF-FX, but it looks
	 * like is comp=3 with T4Options = 4 (byte padded), but with no EOLs. */
	imgInfo->xCompression = IO_COMP_MH;

	/* Fake the user into thinking that the T4Options field exists */
	fieldList->T4Options |= 4;
/* This wouldn't be accurate:	fieldList->found.T4Options = 1; */
	break;
    case TFX_TIFFCOMP_FAXCCITT3:
	/* Note, even if MH is used in TIFF-FX, it is still deemed MR with the 2-D
	 * option turned off. Compression type 2 (MH) doesn't use EOLs. */
	/* Bit 0: 2-D encoding: 0 = MH, 1 = MR
	 * Bit 1: must be 0
	 * Bit 2: 0 = EOLs not byte aligned, 1 = EOLs are byte aligned
	 */
	if (fieldList->T4Options & 1) {
	    imgInfo->xCompression = IO_COMP_MR;
	} else {
	    imgInfo->xCompression = IO_COMP_MH;
	}
	break;
    case TFX_TIFFCOMP_FAXCCITT4:
	imgInfo->xCompression = IO_COMP_MMR;
	/* Bit 0: always 0
	 * Bit 1: uncompressed mode allowed */
	if (fieldList->T6Options & 2) {		/* Uncompressed */
	    __error__ = IO_WARN_NOSUPPORT;
	}
	break;
    case TFX_TIFFCOMP_JPEG:
	imgInfo->xCompression = IO_COMP_JPEG;
	break;
    case TFX_TIFFCOMP_JBIG2:
    case TFX_TIFFCOMP_XEROX_JBIG2:
	/* Determine whether the lossless mode is turned on, from T88Options */
	if (fieldList->T88Options & T88OPTION_LOSSLESS) {
	    imgInfo->xCompression = IO_COMP_LLJBIG2;
	} else {
	    imgInfo->xCompression = IO_COMP_JBIG2;
	}
	break;
    case TFX_TIFFCOMP_XEROX_TOKENIZATION:
	imgInfo->xCompression = IO_COMP_TOKENIZATION;
	break;
    case TFX_TIFFCOMP_XEROX_LLTOKENIZATION:
	imgInfo->xCompression = IO_COMP_LLTOKENIZATION;
	break;
    default:
	imgInfo->xCompression = IO_COMP_OTHER;
    }

    /* Check bit depth and color planes */
    switch (fieldList->SamplesPerPixel) {
    case 1:
	switch (fieldList->BitsPerSample? fieldList->BitsPerSample[0]: 1) {
	case 1:
	    imgInfo->xColorType = IO_COLORTYPE_BINARY;
	    break;
	case 4:
	    if (fieldList->Colormap) {
		imgInfo->xColorType = IO_COLORTYPE_COLOR4;
	    } else {
		imgInfo->xColorType = IO_COLORTYPE_GRAY4;
	    }
	    break;
	case 8:
	    if (fieldList->Colormap) {
		imgInfo->xColorType = IO_COLORTYPE_COLOR8;
	    } else {
		imgInfo->xColorType = IO_COLORTYPE_GRAY8;
	    }
	    break;
	default:
	    __error__ = IO_WARN_NOSUPPORT;
	}
	break;
    case 3:
	switch (fieldList->BitsPerSample? fieldList->BitsPerSample[0]: 1) {
	case 4:
	    imgInfo->xColorType = IO_COLORTYPE_COLOR12;
	    break;
	case 8:
	    imgInfo->xColorType = IO_COLORTYPE_COLOR24;
	    break;
	default:
	    __error__ = IO_WARN_NOSUPPORT;
	}
	break;
    case 4:
	switch (fieldList->BitsPerSample? fieldList->BitsPerSample[0]: 1) {
	case 1:
	    imgInfo->xColorType = IO_COLORTYPE_CMYK1;
	    break;
	case 4:
	    imgInfo->xColorType = IO_COLORTYPE_CMYK4;
	    break;
	case 8:
	    imgInfo->xColorType = IO_COLORTYPE_CMYK8;
	    break;
	default:
	    __error__ = IO_WARN_NOSUPPORT;
	}
	break;
    default:
	__error__ = IO_WARN_NOSUPPORT;
    } /* Samples/pixel */
    
    /* Convert colormap */
    if (imgInfo->xColorType == IO_COLORTYPE_COLOR4 ||
	imgInfo->xColorType == IO_COLORTYPE_COLOR8) {
	{
	    UInt32 component;
	    UInt32 components = 1 << fieldList->BitsPerSample[0];
	    UInt16 *red, *green, *blue;
	    UInt8 *dstCPtr = imgInfo->pPalette;
	    Bool msb = 0;
	    
	    /* Missing colormap */
	    XERR_GEN_CLEANUP((fieldList->Colormap == NULL), IO_BADFILEFORMAT);
	    XERR_GEN_CLEANUP((components !=
		      (1UL << IO_BPPE(imgInfo->xColorType))),
		     IO_BADFILEFORMAT);

	    /* Make one pass through the colormap, and decide whether they
	     * store 16-bit values, or 8-bit values (in the 16-bit words) */
	    red = fieldList->Colormap;
	    green = red + components;
	    blue = green + components;
	    for (component = 0; component < components; component++) {
		if (*red++ > 0xff) {
		    msb = 1; break;
		}
		if (*green++ > 0xff) {
		    msb = 1; break;
		}
		if (*blue++ > 0xff) {
		    msb = 1; break;
		}
	    }

	    /* Copy the data to an 8-bit palette */
	    red = fieldList->Colormap;
	    green = red + components;
	    blue = green + components;
	    if (msb) {
		for (component = 0; component < components; component++) {
		    *dstCPtr++ = (UInt8)((*red++) >> 8);
		    *dstCPtr++ = (UInt8)((*green++) >> 8);
		    *dstCPtr++ = (UInt8)((*blue++) >> 8);
		}
	    } else {
		for (component = 0; component < components; component++) {
		    *dstCPtr++ = (UInt8)((*red++) & 0xff);
		    *dstCPtr++ = (UInt8)((*green++) & 0xff);
		    *dstCPtr++ = (UInt8)((*blue++) & 0xff);
		}
	    }
	} /* Convert colormap */
    }

    /* Transfer the color space */
    switch (fieldList->PhotometricInterpretation) {
    case TFX_TIFFPHOTOMETRIC_WHITEISZERO:
	imgInfo->xColorSpace = IO_COLORSPACE_WHITEISZERO;
	break;
    case TFX_TIFFPHOTOMETRIC_BLACKISZERO:
	imgInfo->xColorSpace = IO_COLORSPACE_BLACKISZERO;
	break;
    case TFX_TIFFPHOTOMETRIC_RGB:
	imgInfo->xColorSpace = IO_COLORSPACE_RGB;
	break;
    case TFX_TIFFPHOTOMETRIC_PALETTE:
	imgInfo->xColorSpace = IO_COLORSPACE_DEFAULT;
	break;
    case TFX_TIFFPHOTOMETRIC_TRANSPARENCY:
	imgInfo->xColorSpace = IO_COLORSPACE_DEFAULT;
	break;
    case TFX_TIFFPHOTOMETRIC_CMYK:
	imgInfo->xColorSpace = IO_COLORSPACE_CMYK;
	break;
    case TFX_TIFFPHOTOMETRIC_YCBCR:
	imgInfo->xColorSpace = IO_COLORSPACE_YCBCR;
	break;
    case TFX_TIFFPHOTOMETRIC_CIELAB:
	imgInfo->xColorSpace = IO_COLORSPACE_CIELAB;
	break;
    case TFX_TIFFPHOTOMETRIC_ICCLAB:
	imgInfo->xColorSpace = IO_COLORSPACE_ICCLAB;
	break;
    case TFX_TIFFPHOTOMETRIC_ITULAB:
	imgInfo->xColorSpace = IO_COLORSPACE_ITULAB;
	break;
    default:
	imgInfo->xColorSpace = IO_COLORSPACE_DEFAULT;
    }
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the nextIFD offset
 *
 * LongDescription:::
 *  wp_TifFXGetNextIFD() reads the nextIFD stored as part of the IFD for the
 *  given image.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXGetNextIFD(
    TFX_IMG *tfxImg
)
{
    TFX_FILE *tfxFile;
    UInt16 fieldCount;
    int swap = 0;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    swap = TFX_FILETOKEN(tfxFile)->bSwapBytes;
    
    /* Go to the beginning of the IFD */
    TFX_SEEK(tfxFile, tfxImg->ifdOffset);
    
    /* Read the field count */
    TFX_READ(tfxFile, &fieldCount, sizeof(fieldCount));
    TFX_SWAPSHORT(tfxFile, &fieldCount);
    tfxImg->fields.fieldCount = fieldCount;
    
    /* Go past the fields */
    TFX_SEEK(tfxFile, tfxImg->ifdOffset + sizeof(fieldCount) + fieldCount * TFX_TIFFFIELDSIZE);
    
    /* Read the next IFD */
    TFX_READ(tfxFile, &tfxImg->fields.nextIFD, sizeof(tfxImg->fields.nextIFD));
    TFX_SWAPLONG(tfxFile, &tfxImg->fields.nextIFD);
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read the TIFF fields
 *
 * LongDescription:::
 *  wp_TifFXReadFields() reads the IFD fields, and copies over the pertinent
 *  information regarding the image.  Will also read the GlobalParametersIFD.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  ifdOffset::			File offset of IFD (normal or Global
 *				Parameters)
 *
 * Output:::
 *  pNextIFD::			Pointer to the next IFD location
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXReadFields(
    TFX_IMG *tfxImg,
    UInt32 ifdOffset,
    UInt32 *pNextIFD
)
{
    TFX_FILE *tfxFile;
    UInt16 fieldCount;
    UInt16 fieldNum;
    TFX_TIFFFIELDENTRY fieldEntry;
    UInt32 fieldSize;
    int swap = 0;
    Int32 fileOffset;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    swap = TFX_FILETOKEN(tfxFile)->bSwapBytes;
    
    /* Go to the beginning of the IFD */
    TFX_SEEK(tfxFile, ifdOffset);
    
    /* Read the field count */
    TFX_READ(tfxFile, &fieldCount, sizeof(fieldCount));
    TFX_SWAPSHORT(tfxFile, &fieldCount);

    /* Save the count (allows us to know how to skip past the fields, when
     * writing the nextIFD */
    tfxImg->fields.fieldCount = fieldCount;
    
    fieldSize = sizeof(fieldEntry.tagID) + sizeof(fieldEntry.type) +
	sizeof(fieldEntry.len) + sizeof(fieldEntry.offset);
    
#ifdef DEBUG
    printf("\tField\tID\tType\tLength\tOffset\n");
#endif
    
    /* Loop through the fields */
    for (fieldNum=0; fieldNum < fieldCount; fieldNum++) {
	
	/* Go to the beginning of the field */
	fileOffset = ifdOffset + sizeof(fieldCount) + fieldNum * fieldSize;
	TFX_SEEK(tfxFile, fileOffset);
	
	/* Read a field */
	TFX_READ(tfxFile, &fieldEntry.tagID, sizeof(fieldEntry.tagID));
	TFX_SWAPSHORT(tfxFile, &fieldEntry.tagID);
	TFX_READ(tfxFile, &fieldEntry.type, sizeof(fieldEntry.type));
	TFX_SWAPSHORT(tfxFile, &fieldEntry.type);
	TFX_READ(tfxFile, &fieldEntry.len, sizeof(fieldEntry.len));
	TFX_SWAPLONG(tfxFile, &fieldEntry.len);
	TFX_READ(tfxFile, &fieldEntry.offset, sizeof(fieldEntry.offset));
	TFX_SWAPLONG(tfxFile, &fieldEntry.offset);
	
#ifdef DEBUG
	printf("\t%2d:\t%2d\t%2d%8d%8d\n", fieldNum, fieldEntry.tagID,
	       fieldEntry.type, fieldEntry.len, fieldEntry.offset);
#endif
	
	XERR_PASS_CLEANUP(wp_TifFXParseField(tfxImg, &fieldEntry, &tfxImg->fields, fileOffset, swap));
    }
    
    /* Go past the fields */
    TFX_SEEK(tfxFile, ifdOffset + sizeof(fieldCount) + fieldCount * fieldSize);
    
    /* Read the next IFD */
    TFX_READ(tfxFile, pNextIFD, sizeof(*pNextIFD));
    TFX_SWAPLONG(tfxFile, pNextIFD);

    tfxImg->fields.dataTransferred = 1 /* 1 */;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read the entire IFD
 *
 * LongDescription:::
 *  wp_TifFXReadIFD() reads the entire IFD and stores the pertinent data.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXReadIFD(
    TFX_IMG *tfxImg
)
{
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);

    if (tfxImg->fields.dataTransferred == 0) {

	/* Use default values */
	memset(&tfxImg->fields, 0, sizeof(tfxImg->fields));	/* 0 for everything */
	tfxImg->fields.SamplesPerPixel = 1;
	tfxImg->fields.FillOrder = 1;
	tfxImg->fields.XResolution.numerator =
	    tfxImg->fields.YResolution.numerator = 300;
	tfxImg->fields.XResolution.denominator =
	    tfxImg->fields.YResolution.denominator = 1;

	/* Read the IFD count, fields, and nextIFD */
	XERR_PASS_CLEANUP(wp_TifFXReadFields(tfxImg, tfxImg->ifdOffset,
	    &tfxImg->fields.nextIFD));

	/* Transfer the field data to the structures used */
	XERR_WARNOK_CLEANUP(wp_TifFXFieldList2Info(tfxImg, &tfxImg->imgInfo));

	/* Compute the bytes per row */
	tfxImg->minBytesPerRow = (tfxImg->imgInfo.dwWidth *
	    IO_BPPE(tfxImg->imgInfo.xColorType) + 7) / 8;
    
    }
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/*
$Log: tfxrtiff.c,v $
Revision 1.3  2003/02/11 01:31:42  mkrocken
fixed problems with comments caused by previous commit

Revision 1.2  2003/02/10 22:53:32  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:13  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.40  2001/12/26 21:15:31  dabercro
Fix logic for identifying MR

Revision 1.39  2001/12/12 00:10:07  dabercro
Tried support for old JPEG, didn't work.

Revision 1.38  2001/11/14 00:07:25  dabercro
Support for 16-bit CIELAB requires the 16th bit for the MSB on L* component,
we must change support for CIELAB to SLONG, not SSHORT.

Revision 1.37  2001/11/07 21:47:16  dabercro
Allow IFDBaseColor to be SSHORT.  Keep track of which type it is.

Revision 1.36  2001/08/22 19:50:41  dabercro
Added support for new field: IFDBaseColor.
Some code cleanup.

Revision 1.35  2001/07/20 23:20:41  dabercro
Parse through the JPEGTables data to make sure the EOI is the last byte.

Revision 1.34  2001/07/05 21:45:43  dabercro
Avoid conversion if field count is zero.

Revision 1.33  2001/06/29 17:49:27  dabercro
Add JPEGTABLES to the list of supported fields.

Revision 1.32  2001/06/01 21:03:28  dabercro
Support centimeters in ResolutionUnit.

Revision 1.31  2001/05/08 23:42:17  dabercro
Fix the byte-swapping on multi-entry fields.

Revision 1.30  2001/05/03 16:40:38  dabercro
Fix FaxProfile "found".

Revision 1.29  2001/04/02 18:25:03  dabercro
Add HostComputer.

Revision 1.28  2001/03/22 20:17:39  dabercro
Add support for new TIFF field: Predictor.

Revision 1.27  2001/03/13 02:59:13  dabercro
Remove lossy-dither and wavelet as valid compression types.

Revision 1.26  2001/02/24 03:26:32  dabercro
Avoid complaining if it's just not this format.

Revision 1.25  2001/01/19 23:56:49  dabercro
Correct structure item name (as per Mark Tobin).

Revision 1.24  2001/01/18 00:21:31  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.23  2000/11/17 23:57:35  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.22  2000/11/02 22:21:21  dabercro
Additions for Rendering Hints.

Revision 1.21  2000/10/19 17:09:55  dabercro
Keep track of T88Options field value location.

Revision 1.20  2000/10/18 20:29:22  dabercro
Keep the GPIFD field count separate.

Revision 1.19  2000/10/16 18:22:54  dabercro
Major adds to support stripes, 0-byte strips, and ImageBaseColor.

Revision 1.18  2000/09/27 21:55:49  dabercro
Look for whether the upper bits in a palette are used, or just the
lower ones.  Copy over which ones appear significant.

Revision 1.17  2000/09/21 21:19:14  dabercro
Add variable-sized strips: StripRowCounts.

Revision 1.16  2000/09/15 22:12:32  dabercro
Use consistent definitions, from iomrc.h. Other bug fixes.

Revision 1.15  2000/08/30 23:29:07  dabercro
Check for both JBIG2 and Tokenization.

Revision 1.14  2000/08/29 22:35:45  dabercro
Allow other compression types.

Revision 1.13  2000/08/25 23:51:28  dabercro
Fix to allow decoding of MH (comp = 2) images (unsupported,
but what the hey).
Mark AnnotationOffsets as read.

Revision 1.12  2000/08/22 01:03:23  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.11  2000/08/04 21:55:35  dabercro
Keep track of the field count, so that we can write the nextIFD
when inserting pages.

Revision 1.10  2000/08/01 00:05:04  dabercro
Correctly handle the sign when reading TIFF fields.
Compression name expansion (_COMP_).

Revision 1.9  2000/06/07 00:53:44  dabercro
Fix merging problem.

Revision 1.8  2000/06/07 00:00:17  dabercro
Continued work, lots of stuff.

Revision 1.7  2000/05/11 01:01:44  mkrocken
Added JBIG2 as possible compression method when reading TIFF-FX image headers

Revision 1.6  2000/05/09 01:52:58  dabercro
Added FillOrder field support.
Updated JBIG2 dictionary front-end.
Support for JBIG2 dictionary extensions, updated parameter to
reflect byte count for extensions.
Write fields that are explicitly set, regardless of default values.
Check if document is single-page when closing, reflect in
NewSubfileType.

Revision 1.5  2000/04/12 19:06:09  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.4  2000/03/11 02:06:35  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.3  2000/02/29 00:19:42  dabercro
Incremental improvements, profile distinctions,etc.

Revision 1.2  2000/02/08 21:25:33  dabercro
Disable JBIG2 (for now), updates to make use of GlobalParametersIFD for symbol libraries.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.19  1998/01/20 03:22:32  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

Rename MixedRasterContent field to ImageLayerContent, per Lloyd's suggestion.

Revision 1.18  1997/12/18 00:12:39  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

Remember the ImageLayer TIFF field data location, in case the image
ordering has to be changed (as in the case of a deleted image).

Revision 1.17  1997/10/15 21:35:44  aber
aber on Wed Oct 15 14:30:51 PDT 1997

tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

Modify return parameters from IsTIFF, allow no first IFD.

Revision 1.16  1997/09/02 20:02:52  aber
aber on Tue Sep 2 13:02:52 PDT 1997

tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

CommonCheck: Rename functions to private names (wp_).

Revision 1.15  1997/08/29 17:28:03  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.14  1997/07/10 21:59:11  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

Use the field count offset to update the count also.

Revision 1.13  1997/07/02 17:03:37  aber
aber on Wed Jul 2 10:02:25 PDT 1997

tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

Have type conversion routine return a value (can't use void, since
a call to this function MUST have a return value checked, otherwise
coreCheckC complains that the return value isn't used).

Revision 1.12  1997/07/02 01:05:32  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

Use private names for swap functions.

Revision 1.11  1997/06/26 22:26:12  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

Change return values to indicate "not TIFF-FX".

Revision 1.10  1997/06/23 22:59:05  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

CommonCheck: Complete function calls for paletted image support.

Revision 1.9  1997/06/20 17:47:15  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

Round the X/Y position vs. truncation.

Revision 1.8  1997/06/17 22:34:05  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

Annotation count type cast.

Revision 1.7  1997/06/12 00:50:53  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

Revision 1.6  1997/06/10 01:26:45  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:05  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:09  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:14  aber
tfxrtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxrtiff.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:39  aber
Initial revision

*/
