/*
 * Trade secret of Xerox Corporation.
 * Copyright (c) 1991-2001, Xerox Corporation.  All rights reserved.
 * Copyright protection claimed includes all forms and matters of 
 * copyrightable material and information now allowed by statutory or 
 * judicial law or hereinafter granted; including, without limitation,
 * material generated from the software programs which are displayed on 
 * the screen such as icons, screen display looks, etc.
 */

#ifndef TFXTFX_TIFFH_INCLUDED
#define TFXTFX_TIFFH_INCLUDED

#ifndef NOIPCORE
IP_RCSINFO(tfxTFX_TIFFhRCSInfo, "$RCSfile: tfxtiff.h,v $; $Revision: 1.3 $");
/* $Date: 2003/02/11 01:31:43 $ */
#endif

#ifdef __cplusplus
extern "C"
{
#endif

#ifdef _MSC_VER
#pragma pack(push, tfxtiff_h, 1)
#endif

/* tfxtiff.h
 *
 */

/*
 * INCLUDES
 */


/*
 * CONSTANTS
 */

/* TIFF equates for Tag names and Field types */

/* Byte orders */
#define II      0x4949
#define MM      0x4D4D

#if defined (__WATCOMC__) || defined (_WIN32) || defined(WIN32) || defined(PC_ENV)
#define TIF_NATIVEORDER II
#else
#define TIF_NATIVEORDER MM
#endif
 
/* Public Tags */
#define NEWSUBFILETYPE			254
#define SUBFILETYPE			255
#define IMAGEWIDTH			256 
#define IMAGELENGTH			257
#define BITSPERSAMPLE			258
#define COMPRESSION			259
#define PHOTOMETRICINTERPRETATION	262
#define THRESHHOLDING			263
#define CELLWIDTH			264
#define CELLLENGTH			265
#define FILLORDER			266
#define DOCUMENTNAME			269
#define IMAGEDESCRIPTION		270
#define MAKE				271
#define MODEL				272
#define STRIPOFFSETS			273
#define ORIENTATION			274
#define SAMPLESPERPIXEL			277
#define ROWSPERSTRIP			278
#define STRIPBYTECOUNTS			279
#define MINSAMPLEVALUE			280
#define MAXSAMPLEVALUE			281
#define XRESOLUTION			282
#define YRESOLUTION			283
#define PLANARCONFIGURATION		284
#define PAGENAME			285
#define XPOSITION			286
#define YPOSITION			287
#define FREEOFFSETS			288
#define FREEBYTECOUNTS			289
#define GRAYRESPONSEUNIT		290
#define GRAYRESPONSECURVE		291
#define T4OPTIONS			292
#define T6OPTIONS			293
#define RESOLUTIONUNIT			296
#define PAGENUMBER			297
#define COLORRESPONSEUNIT		300
#define TRANSFERFUNCTION		301
#define SOFTWARE			305
#define DATETIME			306
#define ARTIST				315
#define HOSTCOMPUTER			316
#define PREDICTOR			317
#define WHITEPOINT			318
#define PRIMARYCHROMATICITIES		319
#define COLORMAP			320
#define HALFTONEHINTS			321
#define TILEWIDTH			322
#define TILELENGTH			323
#define TILEOFFSETS			324
#define TILEBYTECOUNTS			325
#define SUBIFDS				330
#define INKSET				332
#define INKNAMES			333
#define NUMBEROFINKS			334
#define DOTRANGE			336
#define TARGETPRINTER			337
#define EXTRASAMPLES			338
#define SAMPLEFORMAT			339
#define SMINSAMPLEVALUE			340
#define SMAXSAMPLEVALUE			341
#define TRANSFERRANGE			342
#define INDEXED				346
#define JPEGTABLES			347

#define GLOBALPARAMETERSIFD		400
#define PROFILETYPE			401
#define FAXPROFILE			402
#define CODINGMETHODS			403
#define VERSIONYEAR			404
#define MODENUMBER			405

#define DECODE				433
#define IMAGEBASECOLOR			434	/* MRC layer default color? */
#define T82OPTIONS			435	/* Progressive bi-level options */

/* The following 4 fields are current requests of Adobe: */ 
#define MULTIPROFILES_UO		406   /* Multi (fax) profiles (UnOfficial) */
#define TIFFFXEXTENSIONS_UO		407   /* TIFF-FX Extensions (UnOfficial) */
#define T88OPTIONS_UO			436   /* T.88 options (UnOfficial) */
#define SHAREDRESOURCES_UO		437   /* Shared resources (UnOfficial) */

#define JPEGPROC			512
#define JPEGINTERCHANGEFORMAT		513
#define JPEGINTERCHANGEFORMATLENGTH	514
#define JPEGRESTARTINTERVAL		515
#define JPEGLOSSLESSPREDICTORS		517
#define JPEGPOINTTRANSFORMS		518
#define JPEGQTABLES			519
#define JPEGDCTABLES			520
#define JPEGACTABLES			521
#define YCBCRCOEFFICIENTS		529
#define YCBCRSUBSAMPLING		530
#define CHROMASUBSAMPLING		530
#define YCBCRPOSITIONING		531
#define CHROMAPOSITIONING		531
#define REFERENCEBLACKWHITE		532

#define STRIPROWCOUNTS			559

#define COPYRIGHT			33432

#define TIFFFXEXTENSIONS		34687   /* private Xerox tag: TIFF-FX Extensions */
#define MULTIPROFILES			34688   /* private Xerox tag: Multi (Fax) Profiles */
#define SHAREDRESOURCES			34689   /* private Xerox tag: Shared resources */
#define T88OPTIONS			34690   /* private Xerox tag: T.88 Options */
#define PAGEWHITEPOINT			34691   /* private Xerox tag: Page white-point */
#define PAGEBLACKPOINT			34692   /* private Xerox tag: Page black-point */
#define RENDERINGHINTS			34693   /* private Xerox tag: Rendering hints (list) */
#define RENDERINGHINTSIFD		34694   /* private Xerox tag: Rendering hints IFD */
#define IMAGEHISTORY			34695   /* private Xerox tag: Image history */
#define IMAGEPROPERTIES			34696   /* private Xerox tag: Image properties */
#define RESERVEDTAG4			34697   /* private Xerox tag */
#define RESERVEDTAG3			34698   /* private Xerox tag */
#define RESERVEDTAG2			34699   /* private Xerox tag */
#define IFDBASECOLOR			34700   /* private Xerox tag: IFD Base Color */
#define RESERVEDTAG1			34701   /* private Xerox tag */

#define ANNOTATIONOFFSETS		34730   /* private Xerox tag */
#define MASKIFDS			34731   /* private Xerox tag (XIFF 2.0 only) */
#define IMAGELAYER			34732   /* private Xerox tag, made public */
#define IMAGELAYERCONTENT		34733   /* private Xerox tag */
#define REFERENCEPAGE			34734   /* private Xerox tag */


/* Data Types */
#define IFDBYTE				1
#define IFDASCII			2
#define IFDSHORT			3
#define IFDLONG				4
#define IFDRATIONAL			5
#define IFDSBYTE			6
#define IFDUNDEFINED			7
#define IFDSSHORT			8
#define IFDSLONG			9
#define IFDSRATIONAL			10
#define IFDFLOAT			11
#define IFDDOUBLE			12
#define IFDIFD				13  /* Defined in PageMaker 6.0 TIFF
					     * Technical Notes, TTN1 for SubIFDs */

/* Resolution Units */
#define NONE				1
#define INCHES				2
#define CENTIMETERS			3
 
/* compression constants */
#define TFX_TIFFCOMP_NOCOMPRESSBYTE	1       /* No compression, byte aligned */
#define TFX_TIFFCOMP_CCITT3		2       /* CCITT Group 3 1-Dimensional (MH?) */
#define TFX_TIFFCOMP_FAXCCITT3		3       /* Fax Compatible CCITT Group 3 (G3 = MR?) */
#define TFX_TIFFCOMP_FAXCCITT4		4       /* Fax Compatible CCITT Group 4 (G4 = MMR) */
#define TFX_TIFFCOMP_LZW		5       /* LZW compression */
#define TFX_TIFFCOMP_OJPEG		6       /* Old JPEG compression (obsolete) */
#define TFX_TIFFCOMP_JPEG		7       /* JPEG compression */

#define TFX_TIFFCOMP_JBIG_BINARY	9	/* JBIG binary (JBIG1) */
#define TFX_TIFFCOMP_JBIG_COLOR		10	/* JBIG color (JBIG1) */
#define TFX_TIFFCOMP_JBIG2		12	/* JBIG2 (??) */
#define TFX_TIFFCOMP_NOCOMPRESSWORD	32771   /* No compression, word aligned */
#define TFX_TIFFCOMP_PACKBITS		32773   /* Packbits compression */

#define TFX_TIFFCOMP_XEROX_TOKENIZATION	34667   /* Tokenization (symbols only) */
#define TFX_TIFFCOMP_XEROX_WAVELET	34668   /* Wavelet */
#define TFX_TIFFCOMP_XEROX_HVQ		34669   /* Hierarchical vector quantization */
#define TFX_TIFFCOMP_XEROX_LZ77		34670	/* Xerox's Lempel-Ziv 77 */
#define TFX_TIFFCOMP_XEROX_LOSSYDITHER	34672   /* Lossy Dither
					 * Note, these images require special
					 * image processing to reapply them to
					 * the image they were extracted from
					 * (base image):
					 * 1. Morphological "open" with a 3x3
					 *   strel, keep opened and original
					 *   around.
					 * 2. XOR the opened with the base
					 *   image.
					 * 3. SUBTRACT (and with the inverse)
					 *   the opened image from the original
					 *   image.
					 * 4. OR the subtracted image with the
					 *   base image.
					 */
#define TFX_TIFFCOMP_XEROX_LLTOKENIZATION 34673	/* Lossless Tokenization
						 * (symbols only) */
#define TFX_TIFFCOMP_XEROX_DEFLATE	34714	/* Deflate (LZ variant) */
#define TFX_TIFFCOMP_XEROX_JBIG2	34715	/* JBIG2 compression (binary) */
#define TFX_TIFFCOMP_XEROX_JPEG2000	34716	/* JPEG2000 compression stream
						 */
#define TFX_TIFFCOMP_XEROX_LZA		34820	/* LZA Lempel-Ziv via Apollo
						 * chip, owned by Dan Davies */
#define TFX_TIFFCOMP_XEROX_TBD3		34821	/* TBD */
#define TFX_TIFFCOMP_XEROX_TBD2		34822	/* TBD */
#define TFX_TIFFCOMP_XEROX_TBD1		34823	/* TBD */


/* Photometric interpretation values */
#define TFX_TIFFPHOTOMETRIC_WHITEISZERO	0   /* 0 = white */
#define TFX_TIFFPHOTOMETRIC_BLACKISZERO	1   /* 0 = black */
#define TFX_TIFFPHOTOMETRIC_RGB		2   /* Values are RG&B,
					     * black = (0,0,0) */
#define TFX_TIFFPHOTOMETRIC_PALETTE	3   /* Values are indices to a color
					     * palette */
#define TFX_TIFFPHOTOMETRIC_TRANSPARENCY 4  /* 0 = transparent, 1 = opaque */
#define TFX_TIFFPHOTOMETRIC_CMYK	5   /* CMYK photometry */
#define TFX_TIFFPHOTOMETRIC_YCBCR	6   /* YCbCr photometry */
#define TFX_TIFFPHOTOMETRIC_CIELAB	8   /* CIELab photometry */
#define TFX_TIFFPHOTOMETRIC_ICCLAB	9   /* ICCLab photometry */
#define TFX_TIFFPHOTOMETRIC_ITULAB	10  /* ITULAB/FaxLab photometry */

/* Planar configurations */
#define TFX_TIFFPLANAR_CHUNKY		1   /* RGBRGBRGB... */
#define TFX_TIFFPLANAR_PLANES		2   /* RRR...RGGG...GBBB...B */

/* NewSubFileTypes */
#define TFX_TIFFSUBFILE_LOWERRES	0x01   /* Is a reduced resolution
						* version of another */
#define TFX_TIFFSUBFILE_MULTIPAGE	0x02   /* Is a single page of a multi
						* page document */
#define TFX_TIFFSUBFILE_MASK		0x04   /* Is a transparency mask */
#define TFX_TIFFSUBFILE_MRC		0x10   /* Is a part of Mixed Raster
						* Content page */
#define TFX_TIFFSUBFILE_HINTS		0x20   /* Image contains hints (isn't
						* viewable) */

						/* ProfileType bits */
#define TFX_TIFFPROFILETYPE_UNSPECIFIED	0x00000000  /* Unspecified */
#define TFX_TIFFPROFILETYPE_G3FAX	0x00000001  /* G3 Fax */

/* FaxProfile bits */
#define TFX_TIFFFAXPROFILE_NONE		0x00000000  /* Does not conform to a specific profile */
#define TFX_TIFFFAXPROFILE_S		0x00000001  /* Profile S - simple, b&w lossless */
#define TFX_TIFFFAXPROFILE_F		0x00000002  /* Profile F - fax, extended b&w lossless */
#define TFX_TIFFFAXPROFILE_J		0x00000003  /* Profile J - JBIG, b&w lossless */
#define TFX_TIFFFAXPROFILE_C		0x00000004  /* Profile C - color, lossy color & gray */
#define TFX_TIFFFAXPROFILE_L		0x00000005  /* Profile L - lossless color, and gray */
#define TFX_TIFFFAXPROFILE_M		0x00000006  /* Profile M - MRC */

/* CodingMethod bits */
#define TFX_CODINGMETHOD_UNKNOWN	0x00000000  /* Unspecified */
#define TFX_IO_COMP_MH			0x00000001  /* Modified Huffman - T.4 1-D no EOLs */
#define TFX_IO_COMP_MR			0x00000002  /* Modified Read - T.4 1/2-D */
#define TFX_IO_COMP_MMR			0x00000003  /* Modified Modified Read - T.6 2-D */
#define TFX_CODINGMETHOD_JBIG		0x00000004  /* JBIG (binary) */
#define TFX_IO_COMP_JPEG		0x00000005  /* Baseline JPEG */
#define TFX_CODINGMETHOD_JBIGCOLOR	0x00000006  /* JBIG Color */
#define TFX_CODINGMETHOD_JBIG2		0x00000007  /* JBIG2 */

/* Resource types */
#define TFX_RESOURCE_JBIG2		1

/* TIFF-FXExtensions types */
#define TIFFFXEXT_EXT1		0x00000001	    /* Resolution and widths */
#define TIFFFXEXT_EXT2		0x00000002	    /* N-layer MRC */
#define TIFFFXEXT_EXT3		0x00000004	    /* Shared Resources */
#define TIFFFXEXT_EXT4		0x00000008	    /* Binary JBIG2 (Profile T) */
#define TIFFFXEXT_EXT5		0x00000010	    /* JBIG2 for Profile M */

/* T88Options */
#define T88OPTION_NOHUFFMAN	0x00000001	    /* Huffman encoding NOT present */
#define T88OPTION_ARITHMETIC	0x00000002	    /* Arithmetic coding IS present */
#define T88OPTION_LOSSLESS	0x00000004	    /* Data is marked as losslessly
						     * encoded */
#define T88OPTION_SINGLESTRIPED	0x00000008	    /* Each TIFF "strip" has a single
						     * JBIG2 "stripe" */
#define T88OPTION_TEXTSTRIPESMIXED 0x00000010	    /* Text stripes may have other
						     * stripe types (like generic) */
#define T88OPTION_COLOURTAGSFOLLOW 0x00000020	    /* Colour Tags follow, in physical
						     * location, but identified in
						     * another IFD */
#define T88OPTION_COLOURTAGS	0x00000040	    /* Colour Tags present (in this
						     * IFD -- this is the color tag
						     * IFD that followed) */
#define T88OPTION_HALFTONEPRESENT 0x00000080	    /* Halftone regions present */


/*
 * MACROS
 */
#define TFX_RATIONAL2Long(rat)	\
	(rat.denominator? (long)(((long)rat.numerator + \
				  ((long)rat.denominator >> 1)) / \
				 (long)rat.denominator): 0L)
#define TFX_RATIONAL2Float(rat)	\
	(rat.denominator != 0.0? (float)rat.numerator / \
	 (float)rat.denominator: 0.0)

/*
 * TYPEDEFS
 */

typedef struct TFX_TIFFHEADER {
    UInt16 byteOrder;
    UInt16 version;
    UInt32 firstIfd;
} TFX_TIFFHEADER;

typedef struct TFX_TIFFFILE {
    void* file;
    UInt16 byteOrder;
    Int32 firstIfd;
    Int32 pageIFD;
} TFX_TIFFFILE;

typedef struct TFX_TIFFFIELDENTRY {  /* structure of a TIFF directory entry */
    UInt16 tagID;
    UInt16 type;
    UInt32 len;
    UInt32 offset;
    void* data;
} TFX_TIFFFIELDENTRY;
#define TFX_TIFFFIELDSIZE	    12	    /* 12 bytes per field entry */

typedef struct {
    UInt32 numerator;
    UInt32 denominator;
} TFX_RATIONAL;

typedef struct {
    Int32 numerator;
    Int32 denominator;
} TFX_SIGNEDRATIONAL;

/* Image Layer Content, found in primary IFD */
typedef struct TFX_IMGLAYERCONTENT {
    UInt32 layersPresent;	    /* Number of layers mentioned */
    UInt32 *imageCounts;	    /* Number of image each layer. Note:
				     * layer 2 count does not include the
				     * primary IFD. */
} TFX_IMGLAYERCONTENT;

/*
typedef float	Float32;
typedef double	Float64;
*/


/* The GlobalParametersIFD fields */
typedef struct TFX_TIFGPFIELDS {
    UInt32 fieldCount;
    UInt32 TiffFXExtensions;		/* TIFF-FX Extensions (bits) */
    UInt32 MultiProfiles;		/* MultiProfiles (bits) */
    UInt32 SharedResourcesOffset;	/* Offset to SharedResources */
    UInt32 ProfileType;			/* Profile type */
    UInt32 FaxProfile;			/* Profile type */
    UInt32 CodingMethods;		/* Which coders are used */
    UInt8 VersionYear[4];		/* VERSIONYEAR */
    UInt32 VersionYearLength;		/* Length of VersionYear string */
    UInt32 ModeNumber;			/* Mode of the faxProfile standard */
    UInt32 ReferencePage;		/* Reference Page (ordinal) */
} TFX_TIFGPFIELDS;

/* Information fields found */
typedef struct TFX_FIELDSFOUND {
    int AnnotationOffsets:1;
    int Author:1;
    int BitsPerSample:1;
    int ChromaPositioning:1;
    int ChromaSubSampling:1;
    int CodingMethods:1;
    int Compression:1;
    int CopyRight:1;
    int DateTime:1;
    int Decode:1;
    int DocumentName:1;
    int FaxProfile:1;
    int FillOrder:1;
    int HostComputer:1;
    int IFDBaseColor:1;
    int ImageBaseColor:1;
    int ImageDescription:1;
    int ModeNumber:1;
    int Orientation:1;
    int PageNumber:1;
    int PhotometricInterpretation:1;
    int Predictor:1;
    int ProfileType:1;
    int PlanarConfiguration:1;
    int ReferencePage:1;
    int RenderingHints:1;
    int RenderingHintsIFD:1;
    int ResolutionUnit:1;
    int RowsPerStrip:1;
    int SamplesPerPixel:1;
    int Software:1;
    int T4Options:1;
    int T6Options:1;
    int T88Options:1;
    int TargetPrinter:1;
    int Thresholding:1;
    int TiffFXExtensions:1;
    int VersionYear:1;
    int YCbCrCoefficients:1;
} TFX_FIELDSFOUND;

/* Field locations (pointer to the offsets/value item within the field) */
typedef struct TFX_FIELD_LOCATION {
    Int32 AnnotationOffsets;
    Int32 JPEGTables;
    Int32 MaskIFDs;
    Int32 StripOffsets;
    Int32 StripByteCounts;
    Int32 StripRowCounts;
    Int32 SubIFDs;
    Int32 RenderingHints;
    Int32 RenderingHintsIFD;
} TFX_FIELD_LOCATION;


/* Field data locations */
typedef struct TFX_FIELD_DATA_LOCATION {
    Int32 AnnotationOffsets;
    Int32 CodingMethods;
    Int32 ImageLayer;
    Int32 IFDBaseColor;
    Int32 ImageBaseColor;
    Int32 JPEGTables;
    Int32 NewSubfileType;
    Int32 PageNumber;
    Int32 StripOffsets;
    Int32 StripByteCounts;
    Int32 StripRowCounts;
    Int32 SubIFDs;
    Int32 T88Options;
} TFX_FIELD_DATA_LOCATION;


/* The fields that should be used by normal users of this library are
 *   ImageWidth - the width of the image in pixels
 *   ImageLength - the height of the image in pixels
 *   type - TFX_TIFFBINARY: 1 bit image 
 *          TFX_TIFFGRAY16: 4 bit grayscale image
 *          TFX_TIFFGRAY256: 8 bit grayscale image
 *          TFX_TIFFPALETTE256: 8 bit image with a colormap
 *          TFX_TIFFFULLCOLOR: 24 bit image
 *          TFX_TIFFOTHER: some other type of image that can only be
 *                      determined by looking at other variables
 *   Xdpi, Ydpi - the resolution in both dimensions
 *   Xloc, Yloc - the location of the upper left corner in DPM
 *   RowsPerStrip - how much to break up the image, things are likely to
 *                  to go most efficiently if the data is read or written
 *                  in chunks of this many scanlines. It is also better
 *                  if this number is a multiple of 16.
 *
 */

typedef struct TFX_TIFFIELDS {
    TFX_FIELDSFOUND found;		/* Fields found and read */
    UInt8 dataTransferred;		/* the fields have been read/written */
    UInt16 fieldCount;			/* number of fields in the IFD */
    UInt32 nextIFD;			/* file offset of the next IFD */
    UInt8 IFDrewriteNeeded;

    Float64 Xdpi;                       /* derived value */
    Float64 Ydpi;                       /* derived value */
    Float64 Xloc;                       /* derived value, in inches */
    Float64 Yloc;                       /* derived value, in inches */
    Int16 type;				/* derived value */

    UInt16 Compression;			/* COMPRESSION */
    UInt16 PhotometricInterpretation;	/* PHOTOMETRICINTERPRETATION */
    UInt16 Orientation;			/* ORIENTATION */
    UInt16* GrayResponseCurve;		/* GRAYRESPONSECURVE */
    UInt16* HalftoneHints;		/* HALFTONEHINTS */
    UInt16 GrayResponseUnit;		/* GRAYRESPONSEUNIT */
    UInt16 SubFileType;			/* SUBFILETYPE */
    UInt32 NewSubfileType;		/* NEWSUBFILETYPE */
    UInt16 Threshholding;		/* THRESHHOLDING */
    UInt16 CellWidth;			/* CELLWIDTH */
    UInt16 CellLength;			/* CELLLENGTH */
    UInt16* MinSampleValue;		/* MINSAMPLEVALUE */
    UInt16* MaxSampleValue;		/* MAXSAMPLEVALUE */

    UInt16 PageNumber[2];		/* PAGENUMBER [0] */

    /* layout of the uncompressed data */
    UInt16* BitsPerSample;		/* BITSPERSAMPLE */
    UInt16* SampleFormat;		/* SAMPLEFORMAT */
    UInt16 PlanarConfiguration;		/* PLANARCONFIGURATION */
    UInt16 SamplesPerPixel;		/* SAMPLESPERPIXEL */
    UInt16 FillOrder;			/* FILLORDER */
    Int8* DocumentName;			/* DOCUMENTNAME */
    UInt32 DocumentNameLength;		/* Length of DocumentName string */
    
    /* The geometry of the image */
    UInt32 ImageWidth;			/* IMAGEWIDTH */
    UInt32 ImageLength;			/* IMAGELENGTH */
    TFX_RATIONAL XPosition;             /* XPOSITION */
    TFX_RATIONAL YPosition;             /* YPOSITION */
    TFX_RATIONAL XResolution;           /* XRESOLUTION */
    TFX_RATIONAL YResolution;           /* YRESOLUTION */
    UInt16 ResolutionUnit;		/* RESOLUTIONUNIT */

    /* Alpha information */
    UInt16 ExtraSamplesCount;		/* EXTRASAMPLES */
    UInt16* ExtraSamples;

    /* Palette color information */
    UInt16* Colormap;			/* COLORMAP */

    /* CCITT G3 & G4 stuff */
    UInt32 T4Options;			/* T4OPTIONS */
    UInt32 T6Options;			/* T6OPTIONS */

    /* LZW stuff */
    UInt16 Predictor;			/* PREDICTOR */

    /* JPEG Stuff */
    UInt16 JPEGProc;			/* JPEGPROC */
    UInt16 JPEGRestartInterval;		/* JPEGRESTARTINTERVAL */
    UInt32 JPEGInterchangeFormat;	/* JPEGINTERCHANGEFORMAT */
    UInt32 JPEGInterchangeFormatLength;	/* JPEGINTERCHANGEFORMATLENGTH */
    UInt16* JPEGLosslessPredictors;	/* JPEGLOSSLESSPREDICTORS */
    UInt16* JPEGPointTransforms;	/* JPEGPOINTTRANSFORMS */
    UInt32* JPEGQTableOffsets;		/* JPEGQTABLES */
    UInt32* JPEGDCTableOffsets;		/* JPEGDCTABLES */
    UInt32* JPEGACTableOffsets;		/* JPEGACTABLES */
    UInt32 JPEGQuality;			/* JPEG Quality factor (1-100) */

    /* Tiling information */
    UInt32* TileOffsets;		/* TILEOFFSETS */
    UInt32* TileByteCounts;		/* TILEBYTECOUNTS */
    UInt32 TileWidth;			/* TILEWIDTH */
    UInt32 TileLength;			/* TILELENGTH */
    UInt32 TilesAcross;			/* derived value */
    UInt32 TilesDown;			/* derived value */
    UInt32 TileCount;			/* derived value */

    /* strip information */
    UInt32* StripOffsets;		/* STRIPOFFSETS */
    UInt32* StripByteCounts;		/* STRIPBYTECOUNTS */
    UInt32 RowsPerStrip;		/* ROWSPERSTRIP */
    UInt32 StripsPerImage;		/* derived value */

#define TFX_TIFFYCBCR
#ifdef TFX_TIFFYCBCR
    TFX_RATIONAL* YCbCrCoefficients;	/* YCBCRCOEFFICIENTS */
    UInt32 YCbCrCoefficientsValues;	/* Length of YCbCrCoefficients array */
    UInt16 ChromaSubSamplingHor;	/* YCBCR/CHROMASUBSAMPLING[0] */
    UInt16 ChromaSubSamplingVer;	/* YCBCR/CHROMASUBSAMPLING[1] */
    UInt16 ChromaPositioning;		/* YCBCR/CHROMAPOSITIONING */
#endif
#ifdef TFX_TIFFCOLORIMETRY_INFO
    TFX_RATIONAL* WhitePoint;           /* WHITEPOINT */
    TFX_RATIONAL* PrimaryChromaticities;/* PRIMARYCHROMATICITIES */
    UInt16* TransferFunction;		/* TRANSFERFUNCTION */
    UInt16 TransferFunctionSize;	/* the size of TRANSFERFUNCTION */
    UInt16* TransferRange;		/* TRANSFERRANGE */
    TFX_RATIONAL* ReferenceBlackWhite;  /* REFERENCEBLACKWHITE */
#endif
#define TFX_TIFFSEP_INFO
#ifdef TFX_TIFFSEP_INFO
    /* Information about separated images */
    UInt16 InkSet;			/* INKSET */
    UInt16 NumberOfInks;		/* NUMBEROFINKS */
    Int8* InkNames;			/* INKNAMES */
    UInt16 InkNameLength;		/* INKNAMELENGTH */
    UInt16* DotRange;			/* DOTRANGE */
    UInt16 DotRangeLength;		/* DOTRANGELENGTH */
    Int8* TargetPrinter;		/* TARGETPRINTER */
    UInt32 TargetPrinterLength;		/* Length of TargetPrinter string */
    UInt16 Thresholding;		/* THRESHOLDING */
#endif
#define TFX_TIFFTEXT_INFO
#ifdef TFX_TIFFTEXT_INFO
    Int8* ImageDescription;		/* IMAGEDESCRIPTION */
    UInt32 ImageDescriptionLength;	/* Length of ImageDescription string */
    Int8* Software;		        /* SOFTWARE */
    UInt32 SoftwareLength;		/* Length of Software string */
    Int8* DateTime;		        /* DATETIME */
    UInt32 DateTimeLength;		/* Length of DateTime string */
    Int8* HostComputer;			/* HOSTCOMPUTER */
    UInt32 HostComputerLength;		/* Length of HostComputer string */
#endif

    UInt32 *SubIFDs;			/* SUBIFDS */
    UInt32 SubIFDCount;			/* derived value */

    TFX_TIFGPFIELDS gpIFD;		/* GlobalParametersIFD fields */

    Int32 Decode[12];			/* Decode field values (6 rationals
					 * for L*a*b ranges) */
    UInt32 DecodeValues;		/* Number of Decode values found */
    UInt32 T82Options;			/* T82Options */
    Int32 IFDBaseColor[4];		/* IFD base color for the layer,
					 * in the IFD color format.
					 * Note, this color is what is used
					 * in the absense of image data. */
    UInt32 IFDBaseColorValues;		/* Number of values in IFDBaseColor */
    UInt32 IFDBaseColorType;		/* Type used for IFDBaseColor */
    UInt16 ImageBaseColor[4];		/* Image base color for the layer,
					 * in ITULAB L*a*b* (default gamut).
					 * Note, this color is what is used
					 * in the absense of image data. */
    UInt32 ImageBaseColorValues;	/* Number of values in ImageBaseColor */

    Int8 *CopyRight;			/* COPYRIGHT */
    UInt32 CopyRightLength;		/* Length of Copyright string */
    UInt32 *StripRowCounts;		/* Rows per strip (note: count is
					 * StripsPerImage) */

    /* Xerox extensions to TIFF */
    Int32 MaskIFDs;			/* MASKIFDS */
    UInt32 ImageLayer[2];		/* Image layer and order within layer (1, 2, ... ) */
    UInt32 *AnnotationOffsets;		/* ANNOTATIONS (TIFF-FX 2.0 only) */
    UInt16 AnnotationCount;		/* derived value (TIFF-FX 2.0 only) */
    TFX_IMGLAYERCONTENT ImageLayerContent;	/* IMAGE LAYER CONTENT */

    /* Doesn't exist (yet?) */
    Int8 *Author;			/* AUTHOR */
    UInt16 AuthorLength;		/* Length of Author string */

    /* TIFF-FX Extensions 1 */
    UInt32 T88Options;			/* T88OPTIONS */

    Int8 *RenderingHints;		/* RENDERINGHINTS */
    UInt32 RenderingHintsLength;	/* Length of RenderingHints string */
    Int32 RenderingHintsIFD;		/* RENDERINGHINTSIFD */

} TFX_TIFFIELDS;


/*
 * ENUMS
 */

/*
 * GLOBAL VARIABLE DECLARATIONS
 */

/*
 * FUNCTION PROTOTYPES
 */

#ifdef _MSC_VER
#pragma pack(pop, tfxtiff_h)
#endif

#ifdef __cplusplus
}
#endif

#endif /* TFXTFX_TIFFH_INCLUDED */
