/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/* NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE
 * There are two constraints that MUST be accomodated when doing JBIG2
 * encoding:
 *	1. There MUST be at least 2 JBIG2 stripes in every "page".  Lloyd
 *	says that that means one layer (he says I can't have a 1-stripe
 *	JBIG2 image for layer 2, and a 1-stripe JBIG2 image for layer 4).
 *	2. If using "profile 1", then a JBIG2 stripe may have either a
 *	text region ... MUST NOT use color JBIG2
 * 	(see T.89 for completion of this list)
 */

/************************************************************************
 * tfxwjb2.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxwjb2.c:	TIFF-FX WRITE JBIG2 source
 *
 *	TIFF-FX write-only routines for JBIG2
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tfxwjb2.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:32 $");

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif

#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif

#endif	/* NOIPCORE */

#include "tfximage.h"
#include "tfxwtiff.h"
#include "tfxrjb2.h"
#include "tfxwpriv.h"
#include "tfxwtfx.h"
#include "tfxrpriv.h"

#ifndef NODGP
#define _IPWSTATE_DEFINED_  /* Avoid duplicate definition */
/*#include "allocators.h" */
#include "jtokapi.h"
#endif	/* NODGP */
#ifndef NOIPCORE
#include "tfxwjb2.h"
#endif


#ifndef NODGP
#ifdef _WIN32
/*****************************************************************************
 * BeginDocumentation::: "strdup"
 *
 * OneLineDescription:::
 *   Implements strdup for the MS
 *
 * Input:::
 *  s::              Ptr to string.
 *
 * Returns:::
 *  Ptr to duplicate copy of input string.
 *
 * Notes:::
 *  Non-standard error return codes!:: 
 *   This is just Mac-OS based silliness.
 *
 * FunctionCategory:::
 *  client interface
 *
 * Keywords:::
 *  client interface
 *
 * Level:::
 *  deterministic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  dferrell::
 *
 * EndDocumentation::: <A NAME = "strdup"></A>
 *****************************************************************************/
char * CDECL
strdup (
    const char               *s)
{
    char    *rs;
    long int l;

    l = strlen (s);

    rs = (char *)malloc (l);
    if (rs != NULL)
        memcpy (rs, s, l);

    return (rs);

}  /* strdup () */


#endif  /* _WIN32 */


#define DGP_MEM_FIXED
#ifdef DGP_MEM_FIXED

/* Local memory management for JBIG2 library,
 * Since they may call realloc, and I need to maintain both a generic interface 
 * to memory management, and the client version, I must have my own realloc function
 * without the aid of a client implementation of realloc.  To do this, I must keep the
 * size around with each parcel of memory, and use it when realloc-ing, to know the
 * size of the old buffer.
 * The local implementations of calloc and realloc simply call malloc or/and free.
 */
static TFX_INSTANCE *globalInstData = NULL;
static void * CDECL
wp_TfxTextMalloc(
	   size_t bytes
)
{
    TFX_INSTANCE *instData = globalInstData;
    void *buf = NULL;
    UInt32 *lmem;
    PROLOGUE;

    IO_CHECKMALLOC(lmem = TFX_MALLOC(bytes + sizeof(*lmem)));
    buf = (void *)&lmem[1];
    lmem[0] = bytes;

ERR_CLEANUP:

    return (void *)buf;
}

static void * CDECL
wp_TfxTextCalloc(
	   size_t nels,
	   size_t count
)
{
/*    TFX_INSTANCE *instData = globalInstData; */
    void *buf = NULL;
    UInt32 bytes = nels * count;
    PROLOGUE;

    IO_CHECKMALLOC(buf = wp_TfxTextMalloc(bytes));
    memset(buf, 0, bytes);

ERR_CLEANUP:

    return (void *)buf;
}


static void CDECL
wp_TfxTextFree(
	 void *buf
)
{
    TFX_INSTANCE *instData = globalInstData;
    UInt32 *lmem = ((UInt32 *)buf) - 1;

    TFX_FREE(lmem);

    return;
}


static void * CDECL
wp_TfxTextRealloc(
    void *ptr,
    size_t size
)
{
/*    TFX_INSTANCE *instData = globalInstData; */
    UInt32 oldSize;
    UInt32 *lmem = (UInt32 *)ptr - 1;
    void *buffer;
    PROLOGUE;

    /* We need to know the size of the old buffer, to know how to
     * reallocate it. In the case of the memory allocation I have,
     * I keep it stored with the buffer. The client may do anything
     * they desire, so I have no way of pinging for this information.
     */
    oldSize = lmem[0];		/* If seg fault here, buffer doesn't contain 4
				 * bytes -- buffer is probably invalid */
    IO_CHECKMALLOC(buffer = wp_TfxTextMalloc(size));

    if (oldSize > size) {
	memcpy(buffer, ptr, size);
    } else {
	memcpy(buffer, ptr, oldSize);
    }

ERR_CLEANUP:
    wp_TfxTextFree(ptr);

    return buffer;
}

/* Small allocators (don't do anything but call the big ones */
static ErrCode CDECL
wp_TfxTextSmallMalloc(
	void *cookie,
	Int32 id,
	void **ptr
)
{
    void *buf = NULL;
    UInt32 bytes = 0;

    /* Don't return anything (yet) */
    *ptr = NULL;

    switch (id) {
    case DGP_SMALLPOOL_MIN8:
	bytes = 8;
	break;
    case DGP_SMALLPOOL_MIN16:
	bytes = 16;
	break;
    case DGP_SMALLPOOL_MIN20:
	bytes = 20;
	break;
    case DGP_SMALLPOOL_MIN24:
	bytes = 24;
	break;
    case DGP_SMALLPOOL_MIN28:
	bytes = 28;
	break;
    case DGP_SMALLPOOL_MIN32:
	bytes = 32;
	break;
    case DGP_SMALLPOOL_MIN44:
	bytes = 44;
	break;
    case DGP_SMALLPOOL_MIN128:
	bytes = 128;
	break;
    default:
	return -1;
    }

    if ((buf = wp_TfxTextMalloc(bytes)) != NULL) {
	*ptr = buf;
	return 0;
    }
    (void)cookie;
    (void)id;

    return -1;
}

static void CDECL
wp_TfxTextSmallFree(
	void *cookie,
	Int32 id,
	void *buf
)
{

    wp_TfxTextFree(buf);
    (void)cookie;
    (void)id;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert a binary image to token positions and IDs, to be stored later.
 *
 * LongDescription:::
 *  wp_TfxTokenizeImage() converts a binary image from raster form to token
 *  positions and IDs. This is the first of several processes in the
 *  tokenization list. Later, when all pages have been tokenized, the
 *  symbols in the dictionary are consolidated and all position information
 *  is remapped to the cosolidated dictionary. Then all data is converted
 *  to JBIG-2 format, which is more compressed.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  pImgData::			Pointer to first pixel of image data
 *  dwBytesPerRow::		Bytes per row of image data
 *  dwFlags::			Tokenization flags (see TOKEN_... in tfx.h)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxTokenizeImage(
    TFX_IMG *tfxImg,
    UInt8 *pImgData,
    Int32 dwBytesPerRow,
    UInt32 dwFlags
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_PAGE *tfxPage;
    TFX_DOC *tfxDoc;
    TFX_FILE *tfxFile;
#ifndef NOIPCORE
    IPWState *pIPWState = NULL;
    PIXR *imgPixr = NULL;
#endif
    TFX_IMGINFO *imgInfo;
#if 0
    /* Fix when color actually added */
    UInt8 *fgImgPlanes[3];
    Int32 fgImgBpl;
#endif
    Int8 flags = 0;
    DgpApi_Handle dgpHandle = 0;
    DgpApi_PageData pageDesc;
    UInt8 overflow = 0;
#ifdef DGP_MEM_FIXED
    Dgp_Allocators a;
    Dgp_SmallAllocators sm_a;
#endif
    PROLOGUE;

    TFX_CheckHandle(tfxImg);
    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    imgInfo = &tfxImg->imgInfo;
    tfxPage = tfxImg->tfxPage;
    tfxDoc = tfxPage->tfxDoc;


    /* If we're going to queue up a JBIG2 compressed image, we need to write the IFD.
     */
    XERR_PASS_CLEANUP(wp_TifFXWriteIFD(tfxImg));
#if 1
    /* If this isn't the Primary image, link it to the SubIFD chain of the parent */
    {
	TFX_IMG *tfxPrimImg = (TFX_IMG *)tfxImg->tfxPage->pageInfo.tfxPrimImgHandle;

	if (tfxPrimImg != tfxImg) {

	    /* Attach this image to its parent */
	    XERR_PASS_CLEANUP(wp_TifFXAttachChildIFD(tfxPrimImg, tfxImg, tfxImg->ifdOffset));
	}
    }
#endif

    XERR_PASS_CLEANUP(wp_TfxSetPageIFD(tfxPage->tfxDoc, tfxPage->pageInfo.dwPageOffset,
	flags, tfxImg->ifdOffset, tfxPage));

    /* If in a color tokenization state, make sure we pass the color
     * information downward */

#if 0
    /* Fix when color actually added */

    if (tfxPage->seqStatus & TFX_PAGE_SEQ_TKCOLORADDED) {
	TFX_IMG *fgImg = tfxPage->fgImg;
	Int32 channel, channels;
	PIXR *pSubPixr;

	XERR_GEN_CLEANUP(fgImg == NULL, IO_INTERNAL_ERROR);
	XERR_GEN_CLEANUP(fgImg->tmpPixr == NULL, IO_INTERNAL_ERROR);
	XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, fgImg->tmpPixr,
						&channels));
	XERR_GEN_CLEANUP(channels != 3, IO_INTERNAL_ERROR);
	for (channel = 0; channel < channels; channel++) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, fgImg->tmpPixr,
					       channel, &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
				     &fgImgPlanes[channel]));
	}
	XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, fgImg->tmpPixr, &fgImgBpl));
    }
    /* Fix when color actually added */
#endif

#if 0
    /* TIFF-FX and JBIG2 can't handle overlapping layers/images, however,
     * a JBIG2 image might... */
    /* If scanned in binary, halftones may exist which will cause the
     * tokenization process to spend a lot of effort on tiny pixel clumps.
     * Attempt to remove the clumps */
    if (dwFlags & TOKENIZE_BINARYSCAN) {

#if 1
 	XERR_GEN_CLEANUP(imgInfo->dwXResolution <= 0 ||
			 imgInfo->dwYResolution <= 0, IO_BADFILEFORMAT);
#else
	/* Make sure the resolution isn't totally bogus, since it's
	 * important here */
	if (imgInfo->dwXResolution < 50 || imgInfo->dwXResolution > 2000) {
	    if (imgInfo->dwYResolution < 50 || imgInfo->dwYResolution > 2000) {
		imgInfo->dwXResolution = imgInfo->dwYResolution = 300;
	    } else {
		imgInfo->dwXResolution = imgInfo->dwYResolution;
	    }
	}
	if (imgInfo->dwYResolution < 50 || imgInfo->dwYResolution > 2000) {
	    imgInfo->dwYResolution = imgInfo->dwXResolution;
	}
#endif


/* We can't do this in T.44 */
#if 0
	{
	    TFX_INSTANCE *instData = NULL;
	    TFX_PAGE *tfxPage;
	    TFX_DOC *tfxDoc;
	    TFX_FILE *tfxFile;
	    IPWState *pIPWState = NULL;
	    TFX_IMGINFO *imgInfo;
	    UInt8 *positionBuffer = NULL;
	    UInt32 positionBufferSize;
	    UInt32 positionBytes;
	    void *textInfo;
	    ScaledInt1616 xres, yres;
	    Int32 h, bpl, dpi;
	    UInt8 *srcptr, *dstptr;
	    UInt32 row;
	    PixrRef *stipplePixrs = NULL;
	    Int32 numStippleImgs = 0, stippleNum;
	    Int32 stippleWidth, stippleHeight;
	    UInt32 stippleXOffset = 0, stippleYOffset = 0;
	    UInt8 *imgData, *cmpBuffer = NULL, *imgPlanes[3];
	    UInt32 cmpBufsize, cmpBytes;
	    IO_IMGHANDLE stippleImgHandle;
	    TFX_IMG *stippleImg;
	    UInt32 bytesPerRow;
	    IO_COMPRESSION compression;
	    UInt32 quality;
	    UInt32 bytesWritten;
	    Int32 pctIncrease;
	    UInt8 *pFramedImgData;
	    Int32 framedWidth, framedHeight;

	    /* Copy the image data to a PIXR that can be edited */
	    xres = Int32_TO_ScaledInt1616(imgInfo->dwXResolution);
	    yres = Int32_TO_ScaledInt1616(imgInfo->dwYResolution);
	    dpi = MIN(imgInfo->dwXResolution, imgInfo->dwYResolution);
	    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, imgInfo->dwWidth,
		imgInfo->dwHeight, IO_BPPE(imgInfo->xColorType),
		xres, yres, 32, 1, cColorGray, NULL, &imgPixr));
	    
	    h = imgInfo->dwHeight;
	    srcptr = pImgData;
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, imgPixr, &dstptr));
	    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, imgPixr, &bpl));
	    bytesPerRow = MIN(dwBytesPerRow, bpl);
	    for (row = 0; row < h; row++) {
		memcpy(dstptr, srcptr, bytesPerRow);
		srcptr += dwBytesPerRow;
		dstptr += bpl;
	    }
	    
	    /* Initialize strels */
	    XERR_PASS_CLEANUP(w_strel1Inits(pIPWState));
	    XERR_PASS_CLEANUP(w_strel11Inits(pIPWState));
	    
	    /* Remove the stipple from the image */
	    if (dpi >= 300) {
		XERR_PASS_CLEANUP(w_segmentBinaryPagePixrs(pIPWState, imgPixr,
		    4, 8, 1, 99, 15, &stipplePixrs, &numStippleImgs));
	    } else if (dpi >= 200) {
		XERR_PASS_CLEANUP(w_segmentBinaryPagePixrs(pIPWState, imgPixr,
		    2, 4, 1, 66, 10, &stipplePixrs, &numStippleImgs));
	    } else {
		XERR_PASS_CLEANUP(w_segmentBinaryPagePixrs(pIPWState, imgPixr,
		    1, 2, 1, 33, 5, &stipplePixrs, &numStippleImgs));
	    }
	    
	    /* If there's no stipple image(s), continue as normal */
	    if (stipplePixrs == NULL) {
		XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &imgPixr));
	    } else {
		
		/* A stipple was found, loop through them and remove each.
		 * Use lossy dither algorithm to compress */
		for (stippleNum = 0; stippleNum < numStippleImgs; stippleNum++) {
		    
		    /* Get the size */
		    XERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState,
			stipplePixrs[stippleNum].pPixr, &stippleWidth));
		    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState,
			stipplePixrs[stippleNum].pPixr, &stippleHeight));
		    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState,
			stipplePixrs[stippleNum].pPixr, &imgData));
		    stippleXOffset = stipplePixrs[stippleNum].xOffset;
		    stippleYOffset = stipplePixrs[stippleNum].yOffset;
		    
		    /* Store the image */
		    XERR_PASS_CLEANUP(TFX_AddImage((IO_PAGEHANDLE)tfxImg->tfxPage,
			imgInfo->dwImgLayer, IO_COLORTYPE_BINARY, stippleXOffset,
			stippleYOffset, imgInfo->dwXResolution,
			imgInfo->dwYResolution, stippleWidth, stippleHeight,
			stippleHeight, NULL, &stippleImgHandle));
		    stippleImg = (TFX_IMG *)stippleImgHandle;
		    
		    /* Initiate the write process */
		    compression = IO_COMP_LOSSYDITHER;
		    quality = 0;
		    XERR_PASS_CLEANUP(TFX_ImageWriteStart(stippleImgHandle,
			IO_IMGFLAGS_Compressed, &compression, &quality));
		    XERR_GEN_CLEANUP(compression != IO_COMP_LOSSYDITHER, IO_NOSUPPORT);
		    
		    /* Create an image for the lossy dithered data */
		    cmpBufsize = (stippleWidth * stippleHeight / 4) / 8;
		    IO_CHECKMALLOC(cmpBuffer = TFX_MALLOC(cmpBufsize));
		    
		    /* Lossy dither the image */
		    imgPlanes[0] = imgData;
		    XERR_PASS_CLEANUP(w_TfxCompress(pIPWState, stippleImgHandle,
			IO_COMP_LOSSYDITHER, IO_COLORTYPE_BINARY, 0,
			stippleImg->imgInfo.xColorSpace, stipplePixrs[stippleNum].pPixr,
			0, stippleWidth, stippleHeight, &cmpBuffer, &cmpBufsize,
			&cmpBytes));
		    
		    XERR_PASS_CLEANUP(TFX_CompressedWriteStrip(stippleImgHandle,
			0, stippleHeight, cmpBuffer, cmpBytes, &bytesWritten));
		    
		    XERR_PASS_CLEANUP(TFX_ImageWriteFinish(stippleImgHandle));
		    
		    /* Remove it from the original */
		    XERR_PASS_CLEANUP(w_rasterop(pIPWState, 0, 0,
			stipplePixrs[stippleNum].pPixr, PIX_SRC ^ PIX_DST,
			stipplePixrs[stippleNum].xOffset,
			stipplePixrs[stippleNum].yOffset, stippleWidth,
			stippleHeight, imgPixr));
		    
		    /* Delete PIXR */
		    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState,
			&stipplePixrs[stippleNum].pPixr));
		    
		    /* Free the compressed data buffer */
		    TFX_FREE(cmpBuffer), cmpBuffer = NULL;
		    
		} /* For each stipple image */
		
		/* Have free the same way we allocated */
		IPW_FREE(stipplePixrs, cAllocatorData);
		stipplePixrs = NULL, numStippleImgs = 0;

		/* Now repoint the input image to the image with the stipples
		 * removed */
		XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, imgPixr, &pImgData));
		XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, imgPixr, &dwBytesPerRow));
	    }
	}
#endif
    }   /* If a binary scan */
#endif

    /* Setup */
    dgpHandle = tfxDoc->tokEncodeHandle;
    if (dgpHandle == NULL) {
	DgpApi_Options options;
/*	tokOptions tokOptions; */

	/* *********** Work around a client ID problem (for now) *********** */
#ifdef DGP_MEM_FIXED
	globalInstData = TFX_IMGINST(tfxImg);

	/* Note: the Digipaper code does not always use the client memory
	 * allocation routines.  Until they're fixed, let them use the
	 * system malloc and free */
	a.malloc = wp_TfxTextMalloc;
	a.calloc = wp_TfxTextCalloc;
	a.realloc = wp_TfxTextRealloc;
	a.free = wp_TfxTextFree;
	sm_a.mallocSmall = wp_TfxTextSmallMalloc;
	sm_a.freeSmall = wp_TfxTextSmallFree;
#endif

	/* Clear the parameter lists passed in */
	memset(&options, 0, sizeof(options));
/*	memset(&tokOptions, 0, sizeof(tokOptions)); */

	/* Set default parameters */
	options.sizeThresh = 1;
	options.classThresh = 10;
	options.skewAccuracy = 3;
	options.baselineOrient = BO_DOWN;
	options.rotation = 0;
	options.halftone = DGP_FALSE;
	options.forceRes = 300;
	options.writeclasses = DGP_FALSE;
	options.instrument = DGP_FALSE;
	options.deskew = DGP_FALSE;
	options.g4_toks = DGP_FALSE;
	options.jbig2Profile = 2;
	options.bitmapSizeLimit = DGP_TRUE;
	options.membond = -1;
	options.verbose = DGP_FALSE;

	/*
	 * Set optional parameters
	 */

	/* Decoder memory limit: -1 = unlimited, 1 = 1M, 2 = 2M, 3 = 32M, other = 32M */ 
	options.membond = -1;

	/* T.89 Profile: 1 = Profile 1, 2 = Profile 2, 3 = Profile 3 */
	options.jbig2Profile = 2;

	/* Use Halftone detection */
	options.halftone = 0;

	
#ifndef DGP_MEM_FIXED
	XERR_PASS_CLEANUP(DgpApi_init(&dgpHandle, NULL, NULL,
	    &options, NULL, NULL, NULL));
#else
	XERR_PASS_CLEANUP(DgpApi_init(&dgpHandle, &a, &sm_a,
	    &options, NULL, NULL, NULL));
#endif
	tfxDoc->tokEncodeHandle = (UInt8 *)dgpHandle;

	/* Start a "section" */
	XERR_PASS_CLEANUP(DgpApi_startSection(dgpHandle));
    }

    /* If we have no dictionary, or the one we have is inconsistent with the
     * one we need, start with a new dictionary */
    if (((tfxDoc->dictFlags & TOKENDICT_WRITABLE) == 0) ||
	(((tfxDoc->dictFlags & TOKENDICT_LOSSLESS) == TOKENDICT_LOSSLESS) ^
	((dwFlags & TOKENIZE_LOSSLESS) == TOKENIZE_LOSSLESS))) {

	/* If we have a token dictionary already, flush it */
	if (((tfxDoc->seqStatus & (TFX_DOC_SEQ_TOKENSTART)) == 1) &&
	    ((tfxDoc->dictFlags & TOKENDICT_WRITABLE) == 0)) {
	    XERR_PASS_CLEANUP(TFX_WriteTokenDictionary((IO_DOCHANDLE)tfxDoc));
	}

	/* Start a new list of "current" dictionary IDs */
	if (tfxDoc->curDictIDs) {
	    TFX_FREE(tfxDoc->curDictIDs);
	}
	tfxDoc->curDictCount = 1;
	IO_CHECKMALLOC(tfxDoc->curDictIDs =
	    TFX_MALLOC(tfxDoc->curDictCount * sizeof(UInt32)));

	/* Set the dictionary number (must start at 0 for IPShared routines)
	 * Note, the global dictionary is ALWAYS ID 0.  Once a new global
	 * dictionary is started, the remaining IDs for the pages start at 1.
	 */
/*	tfxDoc->curDictIDs[0] = tfxDoc->resourceTableHeader.resourceCount; */
/*	XERR_PASS_CLEANUP(i_setDictionaryNumber (pIPWState, */
/*	    textInfo, 0)); */

	/* Fix the dictionary ID now, since compressed dictionaries could
	 * later be copied before this dictionary is resolved */
	XERR_PASS_CLEANUP(wp_TfxSetResourceOffset(tfxDoc, 0, 0, TFX_TIFFCOMP_JBIG2,
	    0, 0, &tfxDoc->curDictIDs[0]));

	tfxDoc->dictFlags = TOKENDICT_WRITABLE;
	if (dwFlags & TOKENIZE_LOSSLESS) {
	    tfxDoc->dictFlags |= TOKENDICT_LOSSLESS;
	}
    }

    /* If we're gunna use a previously created dictionary, load now */
#ifdef MERGING_TO_ANOTHER_TOKEN_DICTIONARY
    /* This won't happen, but here's the call we'd use */
    XERR_PASS_CLEANUP(i_loadDictionaryBlock (pIPWState));
#endif

#if 0
    /* This is what was started for IPCore, using a color callback function */
    if (tfxPage->seqStatus & TFX_PAGE_SEQ_TKCOLORADDED) {
	callbackData.img = tfxPage->fgImg;

	/* Add a function to the text compression handle, to make use of the
	 * color information */
	XERR_PASS_CLEANUP(i_initializeTextColorCallback(pIPWState,
	    textInfo, wp_TfxAssignTokenColor, (void *)&callbackData));
    }
#endif

    /* Assign a page ID */
    tfxPage->tokenPageID = ++tfxDoc->nextPageTokenID;

    /* Describe the page */
    memset(&pageDesc, 0, sizeof(pageDesc));
    pageDesc.image = pImgData - (32 * dwBytesPerRow + 4);
    pageDesc.width = imgInfo->dwWidth;
    pageDesc.height = imgInfo->dwHeight;
    pageDesc.bpl = dwBytesPerRow;
    pageDesc.xRes = imgInfo->dwXResolution;
    pageDesc.yRes = imgInfo->dwYResolution;

    /* Store the image as symbols, internally */
    XERR_PASS_CLEANUP(DgpApi_addMonoPage(dgpHandle, &pageDesc, &overflow));
    XERR_GEN_CLEANUP(overflow, IO_BUFFEREXCEEDED);

    /* Reflect tokenization environment (for later info queries) */
    tfxImg->imgInfo.dwBestStripHeight = tfxImg->imgInfo.dwHeight;

    /* Indicate tokenization in progress */
    tfxImg->seqStatus |= TFX_IMG_SEQ_TOKENSTART;
    tfxImg->tfxPage->seqStatus |= TFX_PAGE_SEQ_TOKENSTART;
    tfxImg->tfxPage->seqStatus |= TFX_PAGE_SEQ_MULTIDICTS;
    tfxDoc->seqStatus |= TFX_DOC_SEQ_TOKENSTART;
    tfxDoc->seqStatus |= TFX_DOC_SEQ_MULTIDICTS;

    /* Fall through to free up any temporary memory */
    
 ERR_CLEANUP:

/*    if (positionBuffer) { */
/*	TFX_FREE(positionBuffer); */
/*	positionBuffer = NULL; */
/*    } */

    if (imgPixr) {
	w_destroyPixr(pIPWState, &imgPixr);
    }

#if 0
    if (cmpBuffer) {
	TFX_FREE(cmpBuffer);
    }
    if (numStippleImgs) {
	for (stippleNum = 0; stippleNum < numStippleImgs; stippleNum++) {
	    if (stipplePixrs[stippleNum].pPixr != NULL) {
		w_destroyPixr(pIPWState, &stipplePixrs[stippleNum].pPixr);
	    }
	}
    }
#endif

    ERR_RETURN;
}

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Convert the token dictionary to JBIG2 format, and store in the resource.
 *
 * LongDescription:::
 *  wp_TfxWriteTokenDictionary() converts the current token dictionary to
 *  JBIG2 format and attach it to the list of token dictionaries in the
 *  shared resources.
 *  <BR>
 *  If the multi-page (or global) dictionary hasn't been written yet, then it
 *  is resolved and stored first. If a specific page is passed in, then the
 *  residual (non-matching symbols) are stored in a separate dictionary.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  tfxPage::			Data structure of individual page to store
 *				dictionary of residual
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxWriteTokenDictionary(
	TFX_DOC *tfxDoc,
	TFX_PAGE *tfxPage
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    void *compInfo;
    UInt32 pageRequest = 0;
    UInt32 dictID;
    DgpApi_Jbig2Data dictionary = {NULL, 0UL, 0UL};
    PROLOGUE;

    TFX_CheckHandle(tfxDoc);

    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);

    /* Check sequence */
    if ((tfxDoc->dictFlags & TOKENDICT_WRITABLE) == 0) {
	if (! tfxPage || (tfxPage->seqStatus & TFX_PAGE_SEQ_TOKENDICTSAVED) != 0) {
	    NORMAL_RETURN;
	}
    } else if (tfxPage) {
	/* Make sure we've written the multi-page dictionary first */
	XERR_PASS_CLEANUP(wp_TfxWriteTokenDictionary(tfxDoc, NULL));
    }

    /* Extract the text compression info structure from the document */
    compInfo = tfxDoc->tokEncodeHandle;
    XERR_GEN_CLEANUP(compInfo == NULL, IO_INTERNAL_ERROR);

    /* Determine the set of dictionary symbols that will be written */
    if (tfxPage && (tfxPage->seqStatus & TFX_PAGE_SEQ_TOKENSTART)) {
	pageRequest = tfxPage->tokenPageID;

	/*
	 * Get the per-page dictionary
	 */
/*	XERR_PASS_CLEANUP(DgpApi_endSectionGetDict(compInfo, DGP_INCR_DICT, &dictionary)); */

	/* If no symbols exist, there's no dictionary to write */
	if (dictionary.data_size == 0UL) {
	    tfxPage->seqStatus |= TFX_PAGE_SEQ_TOKENDICTSAVED;

	    /* Reset the ID so that we won't try to reference it in the list
	     * of shared resources */
	    tfxPage->tokenDictID = 0;

	    NORMAL_RETURN;
	}

    } else {
	pageRequest = 0;

	/*
	 * Get the global dictionary
	 */
	XERR_PASS_CLEANUP(DgpApi_endSectionGetDict(compInfo, DGP_INCR_DICT,
	    &dictionary));

	/* If no symbols exist, there's nothing to write */
	if (dictionary.data_size == 0UL) {
	    /* Indicate that no shared dictionary even exists */
	    tfxDoc->dictFlags &= ~TOKENDICT_WRITABLE;
	    tfxDoc->seqStatus &= ~TFX_DOC_SEQ_SHAREDDICT;

	    /* Make sure to clean up the data pointer */
	    goto ERR_CLEANUP;
	}
    }

    XERR_PASS_CLEANUP(TFX_StoreSymbolDictionary((IO_DOCHANDLE)tfxDoc,
	(UInt8 *)dictionary.data, dictionary.data_size, dictionary.data_size, &dictID));

    /* Mark this dictionary as unwritable (already written) */
    if (pageRequest) {
	tfxPage->tokenDictID = dictID;
	tfxPage->seqStatus |= TFX_PAGE_SEQ_TOKENDICTSAVED;
    } else {
	tfxDoc->curDictIDs[0] = dictID;
	tfxDoc->dictFlags &= ~TOKENDICT_WRITABLE;
	tfxDoc->seqStatus |= TFX_DOC_SEQ_SHAREDDICT;
    }

    /* Fall through to free up any temporary memory */

 ERR_CLEANUP:

    /* Note: an empty dictionary may still have a pointer allocated */
    if (dictionary.data) {
#ifndef DGP_MEM_FIXED
	free(dictionary.data);
#else
	wp_TfxTextFree(dictionary.data);
#endif
    }
    NORMAL_RETURN;
} 


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compress the symbols as a JBIG2 "position block", and store.
 *
 * LongDescription:::
 *  wp_TfxWriteTokenImage() compresses the symbols to a JBIG2 position block,
 *  and stores it.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  The format of the pre-Position Block header is: (IS THIS STILL TRUE????????? Code is current!)
 *	1 byte	version number
 *	4 bytes	position block byte count
 *	2 bytes	Number of dictionaries associated with the position block
 *	4 bytes	First dictionary ID
 *	...	Next dictionary ID
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxWriteTokenImage(
	TFX_IMG *tfxImg
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    TFX_DOC *tfxDoc;
    UInt16 dictCount = 0;
    UInt32 binaryCompBytes = 0;
    UInt32 dictIDList[2];   /* List of dictionary IDs to use */
    DgpApi_Jbig2Data posBlock = {(char *)NULL, 0UL, 0UL};
    PROLOGUE;

    TFX_CheckHandle(tfxImg);
    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    tfxDoc = tfxImg->tfxPage->tfxDoc;

    /* Need to make sure the next position block is the one we're asking for */
    /* Need to make sure the next position block is the one we're asking for */
    /* Need to make sure the next position block is the one we're asking for */
    /* Need to make sure the next position block is the one we're asking for */
    /* Need to make sure the next position block is the one we're asking for */
    /* Need to make sure the next position block is the one we're asking for */
    /* Need to make sure the next position block is the one we're asking for */
    /* Need to make sure the next position block is the one we're asking for */
    /* Need to make sure the next position block is the one we're asking for */
    /* Need to make sure the next position block is the one we're asking for */
    /* Need to make sure the next position block is the one we're asking for */
    XERR_PASS_CLEANUP(DgpApi_getNextPosBlock(tfxDoc->tokEncodeHandle, &posBlock));

    /* Write the position block out */
    if (posBlock.data_size) {

	if ((tfxImg->tfxPage->tfxDoc->seqStatus & TFX_DOC_SEQ_SHAREDDICT) != 0) {
	    dictIDList[dictCount++] = tfxDoc->curDictIDs[0];    /* Shared dictionary */
	}

	/* If we stored a second dictionary, include it also */
	if ((tfxImg->tfxPage->tokenDictID > 0) && tfxImg->posSize > 0) {
	    dictIDList[dictCount++] = tfxImg->tfxPage->tokenDictID;
/*	    pageRequest = tfxImg->tfxPage->tokenPageID; */
	}

	XERR_PASS_CLEANUP(TFX_StorePositionBlock((IO_IMGHANDLE)tfxImg,
	    (UInt8 *)posBlock.data, posBlock.data_size, &dictIDList[0], dictCount));
    }

    /* Clean up the memory */
    if (posBlock.data) {
#ifndef DGP_MEM_FIXED
	free(posBlock.data);
#else
	wp_TfxTextFree(posBlock.data);
#endif
	posBlock.data = NULL;
    }

    /* Attach the FG image using the same compressed image stream, and
     * removed the color section from the size of the stream just added */
    if (binaryCompBytes) {
	IO_IMGHANDLE fgImg;

	XERR_PASS_CLEANUP(TFX_ShareColorTokenImage((IO_IMGHANDLE)tfxImg,
	    binaryCompBytes, IO_COLORTYPE_COLOR24, &fgImg));
    }


    /* Fall through to free up any temporary buffers */

 ERR_CLEANUP:

    if (posBlock.data) {
#ifndef DGP_MEM_FIXED
	free(posBlock.data);
#else
	wp_TfxTextFree(posBlock.data);
#endif
	posBlock.data = NULL;
    }
    NORMAL_RETURN;
}
#endif	/* NODGP */


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reuse the color tokenized image data for both the mask and color layers.
 *
 * LongDescription:::
 *  TFX_ShareColorTokenImage() copies the token image IFD for an image in an
 *  even layer image (mask layer, already stored) to a matching image in the
 *  layer just above (foreground layer), and shares the compressed image data
 *  (both size and location); then resets the byte count of the compressed
 *  image in the even layer image to the binary subset, described by the byte
 *  count passed in (second parameter).
 *  <BR>
 *  Note that the simple case is layers 2 and 3, where the color tokenized
 *  stream is first stored in layer 2 as a binary image, but with the color
 *  section included. Then the correct size of the binary portion is passed
 *  in to TFX_ShareColorTokenImage(), which then clones the IFD to make a
 *  layer 3 IFD (modifying the color channel count and photometric), points
 *  to the exact same data in it's entirety, then modifies the layer 2 byte
 *  count.
 *  <BR>
 *  Ya, a bit hokey, but it lets you keep the previous way of doing things,
 *  with a minimal number of extra calls.
 *  <BR>
 *  Note, if the data passed in to store the token image was written to the
 *  file on another computer that has a different byte-order (endian), then
 *  you're basically hosed.  Each extension written to the file takes the
 *  form (2-byte extension type, 4-byte count of bytes, N bytes).  The
 *  2- and 4-byte values for each extension would ideally need to be byte-
 *  swapped when we know that we're writing to a file with a different endian
 *  designation.
 *  <BR>
 *  Here's some scenarios:<BR>
 *  <OL>
 *  <LI>
 *  The binary-color stream was written little-endian, the file was closed,
 *  the file was opened on a big-endian host, then the color extension was
 *  identified.  Here we don't know whether we just wrote the binary-color
 *  stream, or if we're modifying a stream that was already there.  In the
 *  first case, the byte-order is already correct; but in the second, it
 *  needs to be byte-swapped.</LI>
 *  <LI>
 *  The file was created, the binary-color stream was written (endian doesn't
 *  matter), then the color extension was identified.  Here the extension
 *  is already in the correct byte order.  No swapping required.  This would
 *  be the normal case.</LI>
 *  <LI
 *  An existing file is opened (endian could be either), then a page is added
 *  with a binary-color stream, then the color extension is identified.  Now
 *  we have a case where only we know if the endianness of the host matches
 *  the file.  This could reasonably happen occasionally.  We should swap
 *  the bytes.</LI>
 *  </OL>
 *  Because #1 shouldn't really happen, #2 should happen most of the time, but
 *  doesn't require any work, and #3 might happen occasionally, we'll do the
 *  right thing and swap the bytes when the endianness of the file isn't that
 *  of the host (we know that we've just written the binary-color pair due
 *  to a specific variable for this: tfxImg->headerBytes).
 *
 * Input:::
 *  tfxBinaryImgHandle::	Valid image handle of the completed binary
 *				image.  This image should have the image data
 *				already written, as a <B>single strip</B>.
 *  dwBinaryImageBytes::	Byte count of compressed image data (already
 *				stored in IFD of even layer) of the binary
 *				subsection. The current byte count will
 *				be relayed to the IFD of the next layer,
 *				and this value will be written over the
 *				current value.<BR>
 *				Note that the byte count for this block of
 *				data will be adjusted to accomodate the
 *				addition of the header (shared resource list).
 *  xColorType::		Color type of color tokens store in odd
 *				layer IFD
 *  pTfxImgHandle::		Resulting image handle
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_ShareColorTokenImage(
	IO_IMGHANDLE tfxBinaryImgHandle,
	UInt32 dwBinaryImageBytes,
	IO_COLORTYPE xColorType,
	IO_IMGHANDLE *pTfxImgHandle
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    /* Token image (even layer) */
    TFX_IMG *tfxImg = (TFX_IMG *)tfxBinaryImgHandle;
    UInt32 colorTokenBytesOffset = 0;	/* File offset of even layer image
				         * byte count */
    UInt32 colorTokenBytes = 0;	/* Original color token byte count */
    UInt32 headerBytes = 0;	/* Header bytes for the resource list */
    IO_IMGHANDLE fgImgHandle;	/* New FG image handle */
    TFX_IMG *fgImg = NULL;	/* FG image being added */
    UInt32 lval;		/* Temporary long value, used for byte
				 * swapping */
    TFX_IMG *tfxPrimImg;
    PROLOGUE;

    TFX_CheckHandle(tfxImg);
    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);

    /* Make sure it's token compressed */
    XERR_GEN_CLEANUP(tfxImg->imgInfo.xCompression != IO_COMP_TOKENIZATION
	&& tfxImg->imgInfo.xCompression != IO_COMP_LLTOKENIZATION
	&& tfxImg->imgInfo.xCompression != IO_COMP_JBIG2
	&& tfxImg->imgInfo.xCompression != IO_COMP_LLJBIG2,
	IO_BADPARAMETER);

    /* This can only be called when the binary image has a header
     * (that was just written) */
    XERR_GEN_CLEANUP(tfxImg->headerBytes == 0, IO_SEQUENCE);

    /* Get the previously stored token compression stream size */
    XERR_GEN_CLEANUP((tfxImg->fields.StripOffsets == NULL ||
	tfxImg->fields.StripByteCounts == NULL), IO_SEQUENCE);
    colorTokenBytesOffset = tfxImg->fields.StripOffsets[0];
    colorTokenBytes = tfxImg->fields.StripByteCounts[0];
    headerBytes = tfxImg->headerBytes;
    XERR_GEN_CLEANUP(colorTokenBytesOffset == 0, IO_INTERNAL_ERROR);
    XERR_GEN_CLEANUP(colorTokenBytes == 0, IO_INTERNAL_ERROR);
    XERR_GEN_CLEANUP(colorTokenBytes < dwBinaryImageBytes, IO_BADPARAMETER);

    /* Note, we only deal with a single strip (although we don't have to) */
    XERR_GEN_CLEANUP(tfxImg->fields.StripsPerImage != 1, IO_NOSUPPORT);

    /* Create a new image for the color tokenized foreground */
    XERR_PASS_CLEANUP(TFX_AddImage((IO_PAGEHANDLE)tfxImg->tfxPage,
	tfxImg->imgInfo.dwImgLayer + 1, xColorType,
	tfxImg->imgInfo.dwXOffset, tfxImg->imgInfo.dwYOffset,
	tfxImg->imgInfo.dwXResolution, tfxImg->imgInfo.dwYResolution,
	tfxImg->imgInfo.dwWidth, tfxImg->imgInfo.dwHeight,
	 tfxImg->imgInfo.dwHeight, NULL, &fgImgHandle));
    fgImg = (TFX_IMG *)fgImgHandle;

    /* Mark it with the same flavor of token compressed */
    fgImg->imgInfo.xCompression = tfxImg->imgInfo.xCompression;

    /* Write the new IFD */
    XERR_PASS_CLEANUP(wp_TifFXWriteIFD(fgImg));

    /* Link it to the SubIFD chain of the parent */
    tfxPrimImg = (TFX_IMG *)tfxImg->tfxPage->pageInfo.tfxPrimImgHandle;
    XERR_PASS_CLEANUP(wp_TifFXAttachChildIFD(tfxPrimImg, fgImg,
	fgImg->ifdOffset));

    /* Point the strip data to the existing token stream */
    XERR_GEN_CLEANUP((fgImg->fieldDataLocation.StripOffsets == 0 ||
	fgImg->fieldDataLocation.StripByteCounts == 0), IO_SEQUENCE);
    TFX_SEEK(tfxFile, fgImg->fieldDataLocation.StripOffsets);
    lval = colorTokenBytesOffset;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));
    TFX_SEEK(tfxFile, fgImg->fieldDataLocation.StripByteCounts);
    lval = colorTokenBytes;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));

    /* Now go back to the original image, and revise the byte count
     * of both the TIFF field, and the stream header.
     * NOTE that we adjust the byte count to include the header. */
    TFX_SEEK(tfxFile, tfxImg->fieldDataLocation.StripByteCounts);
    lval = dwBinaryImageBytes + headerBytes;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));

    /* Now the stream */
    TFX_SEEK(tfxFile, tfxImg->fields.StripOffsets[0] + headerBytes - sizeof(lval));
    lval = dwBinaryImageBytes;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));

    /* Mark the binary IFD as having color tags that follow */
    XERR_GEN_CLEANUP(tfxImg->fieldDataLocation.T88Options <= 8, IO_SEQUENCE);
    TFX_SEEK(tfxFile, tfxImg->fieldDataLocation.T88Options);
    tfxImg->fields.T88Options |= T88OPTION_COLOURTAGSFOLLOW;
    lval = tfxImg->fields.T88Options;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));

    /* If the endian-ness of our host doesn't match that of the file
     * (most likely happens when adding new image/pages to an existing file),
     * then we'll have to go through the extensions and swap the bytes */
    if (TFX_FILETOKEN(tfxFile)->bSwapBytes) {
	Int32 seekOffset;
	Int32 bytesLeft;
	UInt16 type, sval;
	Int32 bytes;

	/* Keep track of the extension position and bytes left in the
	 * extension */
	seekOffset = colorTokenBytesOffset;
	bytesLeft = colorTokenBytes - dwBinaryImageBytes;

	/* Loop through each extension */
	while (bytesLeft > 0) {

	    /* Go to the location where the next(first) extension resides */
	    TFX_SEEK(tfxFile, seekOffset);

	    /* Read in the 2-byte type, and 4-byte byteCount */
	    TFX_READ(tfxFile, &type, sizeof(type));
	    TFX_READ(tfxFile, &bytes, sizeof(bytes));

	    /* Check the byte count, and make sure that it jives with the
	     * total in all extensions */
	    XERR_GEN_CLEANUP(bytes > bytesLeft, IO_BADPARAMETER);
	    /* Note, if erroring here, it means that the byte count in the
	     * data stream is not consistent with the data left in the
	     * extension.  It may be that the client has tried to byte-swap
	     * themselves, or that the extension isn't located at the
	     * offset dictated by the binary image size */

	    /* Swap them */
	    sval = type, lval = bytes;
	    TFX_SWAPSHORT(tfxFile, &sval);
	    TFX_SWAPLONG(tfxFile, &lval);

	    /* Seek back to the beginning of the extension */
	    TFX_SEEK(tfxFile, seekOffset);
	    
	    /* Write the byte-swapped values back out */
	    TFX_WRITE(tfxFile, &sval, sizeof(sval));
	    TFX_WRITE(tfxFile, &lval, sizeof(lval));

	    /* Subtract the extension from the total */
	    seekOffset += bytes;
	    bytesLeft -= bytes;

	    /* Go on to the next extension (if any) */
	}
    }

    /* Return the resulting image handle */
    *pTfxImgHandle = fgImg->handleID;

    NORMAL_RETURN;

 ERR_CLEANUP:
    ERR_RETURN;
}


#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Assign the color to a connected component, found during tokenization
 *
 * LongDescription:::
 *  wp_TfxAssignTokenColor() determines the color associated with a binary
 *  connected component, found during tokenization. If the color is fairly
 *  constant, then the color will be placed in a list, and the color ID will
 *  be attached to the token symbol. If the color is not fairly consistent,
 *  then a miniature image is kept with the token dictionary, and the color
 *  ID for the image is attached to the token symbol.
 *
 * Input:::
 *  pIPWState::			Pointer to IPWrap state structure
 *  textInfo::			Text info structure used in symbol compression
 *  callbackData::		Callback function used to assign color to
 *				symbol
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxAssignTokenColor(
	IPWState *pIPWState,
	void *textInfo,
	void *callbackData
)
{
    TCData *tcData = (TCData *)callbackData;
    PROLOGUE;

    pIPWState = pIPWState;
    textInfo = textInfo;
    tcData = tcData;

    NORMAL_RETURN;
#if 0    
 ERR_CLEANUP:
    ERR_RETURN;
#endif
}
#endif	/* if 0 */


/*
$Log: tfxwjb2.c,v $
Revision 1.2  2003/02/10 22:53:32  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:14  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.14  2001/12/26 21:23:45  dabercro
Mark as released

Revision 1.13  2001/11/07 21:50:25  dabercro
Compiler warning fixes.

Revision 1.12  2001/09/18 21:10:11  dabercro
Make work with JBIG2

Revision 1.11  2001/06/27 18:55:45  dabercro
Borrow flag names from IO_ layer, expand list of ones implemented.

Revision 1.10  2001/03/13 02:59:13  dabercro
Remove lossy-dither and wavelet as valid compression types.

Revision 1.9  2001/01/18 00:21:31  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.8  2000/11/17 23:57:35  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.7  2000/09/21 21:19:14  dabercro
Add variable-sized strips: StripRowCounts.

Revision 1.6  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.5  2000/08/30 23:29:07  dabercro
Check for both JBIG2 and Tokenization.

Revision 1.4  2000/06/07 00:00:17  dabercro
Continued work, lots of stuff.

Revision 1.3  2000/04/12 19:06:09  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.2  2000/03/11 02:06:35  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.35  1998/08/06 16:43:20  aber
aber on Thu Aug 6 09:43:05 PDT 1998

tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Use unused variables, to avoid Watcom complaints.

Revision 1.34  1998/08/05 23:31:11  aber
aber on Wed Aug 5 16:30:58 PDT 1998

tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Fixed coreCheckC complaints.

Revision 1.33  1998/08/04 00:21:39  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

CommonCheck: Support multiple dictionaries when token compressing, used to retain
single-use tokens in the same dictionary -- to be used when color
tokenizing.

Revision 1.32  1998/08/01 19:53:43  aber
aber on Sat Aug 1 12:53:26 PDT 1998

tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Correct the name of the parameter in the comment.

Revision 1.31  1998/07/30 22:41:55  aber
aber on Thu Jul 30 15:41:55 PDT 1998

tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

CommonCheck: New function TFX_ShareColorTokenImage(), used to add a FG image IFD, while
reusing the existing token data, but then overwriting the byte count of
the binary image to reflect the non-color subset of the token stream.

Revision 1.30  1998/05/21 06:14:55  aber
aber on Wed May 20 23:08:24 PDT 1998

tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Remove temporary fix intended to be "nice". If the resolution is set too
low (like 0), I automatically bumped the resolution to what we normally
see (300). However, I have one case where I've forced the text plane
to be very low (10 dpi), and this "nice" fix causes it to crash.
So for now, we'll force the file to have the correct resolution.

Revision 1.29  1998/04/03 00:00:55  aber
aber on Thu Apr 2 15:58:49 PST 1998

tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Fix a bug when tokenizing a page (first dictionary, but not yet saved),
then copying a second token dictionary from another file (now the first),
then saving the thought-to-be first token dictionary; which is now the
second.

Revision 1.28  1998/02/19 21:28:14  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Rename local text allocation routines to use wp_ names.

Revision 1.27  1998/01/20 18:06:11  aber
aber on Tue Jan 20 10:05:57 PST 1998

tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Compiler signed/unsigned complaint fix.

Revision 1.26  1998/01/20 03:25:15  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Reset the current dictionary status, when writing and then reading back
a page that has been tokenized.

Revision 1.25  1998/01/14 00:16:44  ddavies
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.24  1997/11/26 23:28:06  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Added a check to make sure that the tokenized image (supposedly the
primary image) does indeed have the tokenized data attached to the
image structure.

Revision 1.23  1997/10/01 23:44:22  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Loop with more memory until the dictionary compresses. Don't add
the dictionary offset to the header when the tokenization stream
is empty.

Revision 1.22  1997/08/31 17:09:31  aber
aber on Sun Aug 31 10:07:50 PDT 1997

tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Handle case where residual buffer doesn't increase (Andrew found bug).

Revision 1.21  1997/08/27 23:40:05  aber
aber on Wed Aug 27 16:39:41 PDT 1997

tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Fix the token dictionary ID, start at 0, not -1.

Revision 1.20  1997/08/27 00:17:40  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Avoid repeated file seeks.

Revision 1.19  1997/08/15 23:48:03  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

CommonCheck: Collective decision by Dan Huttenlocher, William Rucklidge, and I
to remove the token dictionary ID from the symbol library table
entry, since the offset is typically used, and the ID does nothing
for us.

Revision 1.18  1997/08/15 01:54:19  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

CommonCheck: Modify token dictionary header to include the Tunrda dictionary
byte count, allows for extensible dictionary data.

Revision 1.17  1997/08/15 00:11:41  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Use predefined offsets for pre-Tundra header.

Revision 1.16  1997/08/07 21:27:50  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

CommonCheck: Make provisions for a 4-byte dictionary block size, instead of 2-bytes
spelled out in Williams table. Also make provisions for multiple
dictionaries per page.

Revision 1.15  1997/07/02 01:09:04  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Use the same unframed parameters for both stipple and non-stipple modes.

Revision 1.14  1997/06/26 23:29:22  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Make potential ansification fixes that the Mac might complain about.

Revision 1.13  1997/06/26 22:34:10  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Check for no symbols, don't output the token dictionary if such,
but store the minimum position information with a dictionary count of 0.

Revision 1.12  1997/06/23 22:48:10  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

CommonCheck: Sync up image pointers so that lossy dither works right.

Revision 1.11  1997/06/17 22:35:53  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Define renames.

Revision 1.9  1997/06/12 00:52:42  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Enable lossless tokenization.

Revision 1.7  1997/06/10 23:35:00  aber
aber on Tue Jun 10 16:34:20 PDT 1997

tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

Clean up temporary memory, fixes memory leak. Fix documentation.

Revision 1.6  1997/06/10 01:27:04  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:21  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:25  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:35  aber
tfxwtok.c -> /product/ipcore/ipwrap/src/RCS/tfxwtok.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:50  aber
Initial revision

*/
