/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxwjpeg.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxwjpeg.c:	TIFF-FX WRITE JPEG source
 *
 *	Compression (write) for JPEG.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#ifdef __GNUC__
#include "ansiprot.h"
#endif
#endif	/* NOIPCORE */

/*#ifndef IJL */

/* In the no-IPCore version, only the JPEG table stripping function is
 * enabled */
#ifndef NOIPCORE

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "jpeglib.h"
#include "iw_jpeg.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif

/*#else	/* IJL */ */

/* IJL */
#include "ijl.h"

#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "shrpixr.pub"
#include "pcconv.pub"	/* Conversion to/from DIB */
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif

#endif	/* NOIPCORE */
/*#endif	/* IJL */ */

#include "tfximage.h"
#include "tfxrtiff.h"
#include "tfxwjpeg.h"
#include "tfxrpriv.h"
#include "tfxwpriv.h"


#ifndef NOIPCORE
IP_RCSINFO(RCSInfo, "$RCSfile: tfxwjpeg.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:32 $");


/*#ifndef IJL	/* IJL */ */

typedef struct WriteData {
    TFX_INSTANCE *instData;
    UInt8 **ppBuffer;
    UInt8 *next;
    UInt32 *pBufBytes;
    UInt32 bytesLeft;
} WriteData;

static ErrCode CDECL
IO_JPEG_WriteCallback(
    IPWState *pIPWState,
    void *userData,
    UInt8 *ptr,
    UInt32 bytes,
    UInt32 *pdwBytesWritten
)
{
    TFX_INSTANCE *instData;
    WriteData *writeData = (WriteData *)userData;
    UInt32 newbufsize;
    UInt8 *newbuf = NULL;
    PROLOGUE;

    pIPWState = pIPWState;
    
    if (writeData->bytesLeft < bytes) {
	instData = writeData->instData;

	/* Must reallocate the buffer to accomodate the new amount,
	 * increase by 20% more than what we know we need now */
	newbufsize = (*writeData->pBufBytes - writeData->bytesLeft + bytes)
	    * 120 / 100;
	IO_CHECKMALLOC(newbuf = TFX_MALLOC(newbufsize));

	/* Copy the data to the new buffer */
	memcpy(newbuf, *writeData->ppBuffer,
	    writeData->next - *writeData->ppBuffer);
	writeData->next = newbuf +
	    (*writeData->pBufBytes - writeData->bytesLeft);
	writeData->bytesLeft = newbufsize -
	    (*writeData->pBufBytes - writeData->bytesLeft);

	/* Update the external buffer */
	TFX_FREE(*writeData->ppBuffer);
	*writeData->ppBuffer = newbuf;
	*writeData->pBufBytes = newbufsize;

	/* Dereference the now stored buffer */
	newbuf = NULL;
    }
    memcpy(writeData->next, ptr, bytes);
    writeData->next += bytes;
    writeData->bytesLeft -= bytes;
    *pdwBytesWritten = bytes;
    NORMAL_RETURN;

ERR_CLEANUP:
    if (newbuf) {
	instData = writeData->instData;
	TFX_FREE(newbuf);
    }

    return cErrCallbackError;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  JPEG compress a strip of image data
 *
 * LongDescription:::
 *  wp_TfxCompressStripJPEG() JPEG compresses a strip of image data.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  pPixr::			Pointer to image PIXR
 *  dwRowOffset::		Row offset of image strip to compress
 *  dwStripHeight::		Number of rows to compress
 *  xColorSpace::		Color space to convert to during compression
 *				IO_COLORSPACE_<c> where <c> is one of:
 *				    CMYK, YCBCR, CIELAB, ICCLAB, or ITULAB
 *
 * Output:::
 *  ppCmpData::			Pointer to pointer to the compression buffer,
 *				may be reallocated if not enough space
 *  pdwCmpBytes::		Number of bytes in compression buffer, may
 *				be modified if not enough space
 *  pdwBytesFilled::		Number of bytes filled, in compression buffer
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxCompressStripJPEG(
    TFX_IMG *tfxImg,
    PIXR *pPixr,
    UInt32 dwRowOffset,
    UInt32 dwStripHeight,
    IO_COLORSPACE xColorSpace,
    UInt8 **ppCmpData,
    UInt32 *pdwCmpBytes,
    UInt32 *pdwBytesFilled
)
{
    TFX_IMGINFO *pTfxImgInfo;
    IPWState *pIPWState = NULL;
    UInt32 rowsToWrite = dwStripHeight;
    CompressInfo      cinfo_struct;
    Compress_info_ptr cinfo=&cinfo_struct;
    UInt8 *image[3] = {NULL, NULL, NULL};
    Int32 h, bpl;
    Int32 row, line;
    Int32 nlines;
    Int32 channel, channels;
    UInt32 memorg = 0UL;
    JSAMPIMAGE userBuf = NULL;
    WriteData writeData;
    Int32 quality = 0;
    UInt32 options;
    PROLOGUE;
    
    /* Default to return nothing */
    *pdwBytesFilled = 0;

    /* Can't offset into PIXR for now */
    pTfxImgInfo = &tfxImg->imgInfo;
    XERR_GEN_CLEANUP(dwRowOffset + dwStripHeight > pTfxImgInfo->dwHeight,
	IO_BADPARAMETER);

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &h));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &bpl));

    /* Initialize */
    writeData.instData = TFX_IMGINST(tfxImg);
    writeData.ppBuffer = ppCmpData;
    writeData.next = *writeData.ppBuffer;
    writeData.pBufBytes = pdwCmpBytes;
    writeData.bytesLeft = *writeData.pBufBytes;
    switch (IO_CHANNELS(tfxImg->imgInfo.xColorType)) {
    case 3:
	memorg = JMEMORG_BSQ_PIXR;
	break;
    case 1:
	switch (IO_BPPE(tfxImg->imgInfo.xColorType)) {
	case 4:
	    memorg = JMEMORG_GRAY4_PIXR;
	    break;
	case 8:
	    memorg = JMEMORG_GRAY8_PIXR;
	    break;
	}
	break;
    }
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, &channels));

    options = JC_DEFAULT;
    switch (xColorSpace) {
    case IO_COLORSPACE_CIELAB:
    case IO_COLORSPACE_ICCLAB:
	/* Adjust the quality for CIELAB */
	quality = tfxImg->JPEGquality;
	options |= JC_CIELAB;
	break;
    case IO_COLORSPACE_ITULAB:
	/* Adjust the quality for ITULAB */
	quality = tfxImg->JPEGquality;
	options |= JC_ITULAB;
	break;
    case IO_COLORSPACE_CMYK:
    case IO_COLORSPACE_YCBCR:
    default:
	quality = tfxImg->JPEGquality;
    }

    /* set chrominance subsample (off) */
    if ((tfxImg->imgFlags & IO_IMGFLAGS_SubSampling) ==
	IO_IMGFLAGS_NoSubSampChroma) {
	options |= JC_NO_CHROMA_SS;
    }

    /* use one set of Q/Huffman tables: place in JPEGTables field,
     * not embedded in each strip of data */
/*    if (tfxImg->imgFlags & IO_IMGFLAGS_JPEGTables) { */
/*	options |= JC_JPEGTABLES; */
/*    } */

    /* Initialize the JPEG process */
    memset(cinfo, 0, sizeof(*cinfo));
    XERR_PASS_CLEANUP(w_jcInit(pIPWState, cinfo, tfxImg->imgInfo.dwWidth,
	dwStripHeight, channels, quality, options,
	tfxImg->imgInfo.dwXResolution, tfxImg->imgInfo.dwYResolution,
	IO_JPEG_WriteCallback, (void *)&writeData, NULL, memorg));
    
    /* Get the location of the image and the line pitch */
    {
	PIXR  *pSubPixr;
	for (channel = 0; channel < channels; channel++) {
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
				  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
				&(image[channel])));
	    image[channel] += dwRowOffset * bpl;
	}
    }
    if (h < (Int32)rowsToWrite) {
	rowsToWrite = h;
    }
    
    XERR_PASS_CLEANUP(w_jcAllocMCURowBuffer(pIPWState, cinfo, 0, 
		                      	       &userBuf));
    XERR_PASS_CLEANUP(userBuf == NULL);
    
    /* Decompress the image an MCU row at a time */
    XERR_PASS_CLEANUP(w_jcGetImageLinesPerMCU(pIPWState, cinfo, &nlines));
    line =0;
    for (row = 0; row < (Int32)rowsToWrite; ++row) {
	for (channel = 0; channel < channels; channel++) {
	    userBuf[channel][line] = image[channel];
	    image[channel] += bpl;
	}
	++line;
	if (line >= nlines || row == ((Int32)dwStripHeight - 1)) {    
	    /* Compress an MCU row from userBuf */
	    XERR_PASS_CLEANUP( w_jcMCURow(pIPWState, cinfo, userBuf));
	    line =0;
	}
    }
    
    XERR_PASS_CLEANUP(w_jcMCURowFree(pIPWState, cinfo, 0, userBuf));
    userBuf = NULL;
    
/*    XERR_PASS_CLEANUP( w_jcTerm(pIPWState, cinfo)); */
    XERR_IGNORE( w_jcTerm(pIPWState, cinfo));

    *pdwBytesFilled = *writeData.pBufBytes - writeData.bytesLeft;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (userBuf != NULL)
	w_jcMCURowFree(pIPWState, cinfo, 0, userBuf);
    
    jpeg_destroy_compress(pIPWState, &cinfo->j_cinfo);
    
    w_jcFreeErrorStr(pIPWState, cinfo);
    
    ERR_RETURN;
}

/*#else	/* IJL */ */

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  JPEG compress a strip of image data
 *
 * LongDescription:::
 *  wp_TfxCompressStripJPEG_IJL() JPEG compresses a strip of image data using
 *  Intel's JPEG library.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  pPixr::			Pointer to image PIXR
 *  dwRowOffset::		Row offset of image strip to compress
 *  dwStripHeight::		Number of rows to compress
 *  xColorSpace::		Color space to convert to during compression
 *				IO_COLORSPACE_<c> where <c> is one of:
 *				    CMYK, YCBCR, CIELAB, ICCLAB, or ITULAB
 *
 * Output:::
 *  ppCmpData::			Pointer to pointer to the compression buffer,
 *				may be reallocated if not enough space
 *  pdwCmpBytes::		Number of bytes in compression buffer, may
 *				be modified if not enough space
 *  pdwBytesFilled::		Number of bytes filled, in compression buffer
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxCompressStripJPEG_IJL(
    TFX_IMG *tfxImg,
    PIXR *pPixr,
    UInt32 dwRowOffset,
    UInt32 dwStripHeight,
    IO_COLORSPACE xColorSpace,
    UInt8 **ppCmpData,
    UInt32 *pdwCmpBytes,
    UInt32 *pdwBytesFilled
)
{
    TFX_INSTANCE *instData = TFX_IMGINST(tfxImg);
    TFX_IMGINFO *pTfxImgInfo;
    IPWState *pIPWState = NULL;
    UInt32 rowsToWrite = dwStripHeight;
    UInt8 *image[3] = {NULL, NULL, NULL};
    Int32 h, bpl;
    Int32 channels;
    JPEG_CORE_PROPERTIES jcprops;
    IJLERR jerr;
    Bool bres = TRUE;
    UInt8 *dibImage = NULL;
    UInt32 byteCount = 0UL;
    UInt32 dibBPL;
    PROLOGUE;
    
    /* Default to return nothing */
    *pdwBytesFilled = 0;

    /* Can't offset into PIXR for now */
    pTfxImgInfo = &tfxImg->imgInfo;
    XERR_GEN_CLEANUP(dwRowOffset + dwStripHeight > pTfxImgInfo->dwHeight,
	IO_BADPARAMETER);

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &h));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &bpl));

    /* Initialize */
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, &channels));

    /* Initialize */
    __try
    {
	/* Initialize the IJL */
	jerr = ijlInit(&jcprops);
	if (jerr < IJL_OK) {
	    bres = FALSE;
	    __leave;
	}

	switch (tfxImg->fields.PhotometricInterpretation) {
	case 8:
	case 9:	/* Works out that they appear the same */
	    /* Fall through and do the same thing as ITULAB */
	case 10:	/* ITULAB */
	    /* Force using both quantization tables */
	    jcprops.jprops.maxquantindex = 2;	/* Luminance and chrominance */
	    jcprops.jprops.nqtables = 2;	/* Use both */

	    /* Use chroma tables for chroma components */
	    jcprops.jprops.jframe.comps[0].quant_sel = 0;
	    jcprops.jprops.jframe.comps[1].quant_sel = 1;
	    jcprops.jprops.jframe.comps[2].quant_sel = 1;

	    /* Force chroma subsampling */
	    jcprops.jprops.JPGSubsampling = IJL_411;

	    /* Use "other" colorspace */
	    jcprops.jprops.JPGColor = IJL_OTHER;
	    jcprops.JPGColor = IJL_OTHER;
	    jcprops.cconversion_reqd = 0;
	    break;
	}

	/* set chrominance subsample (off) */
	if ((tfxImg->imgFlags & IO_IMGFLAGS_SubSampling) ==
	    IO_IMGFLAGS_NoSubSampChroma) {
	    jcprops.jprops.JPGSubsampling = IJL_NONE;
	}

	/* Set the image parameters */
	jcprops.JPGWidth = pTfxImgInfo->dwWidth;
	jcprops.JPGHeight = dwStripHeight;
	jcprops.JPGChannels = channels;
	jcprops.jquality = tfxImg->JPEGquality;
	jcprops.DIBWidth = jcprops.JPGWidth;
	jcprops.DIBHeight = -jcprops.JPGHeight;
	jcprops.DIBChannels = jcprops.JPGChannels;
	jcprops.DIBColor = IJL_OTHER;
	dibBPL = (((jcprops.JPGWidth * IO_BPP(pTfxImgInfo->xColorType))
	    + 31) / 32) * 4;
	jcprops.DIBPadBytes = dibBPL -
	    ((jcprops.JPGWidth * IO_BPP(pTfxImgInfo->xColorType) + 7) / 8);
	jcprops.DIBSubsampling = 0;
	jcprops.upsampling_reqd = 0;

	/* Allocate the DIB image */
	byteCount = ((jcprops.JPGWidth * IO_BPPE(pTfxImgInfo->xColorType) + 31)
	    / 4) * jcprops.JPGHeight * channels;
	IO_CHECKMALLOC(dibImage = IO_MALLOC(byteCount));

	/* Swap the R and B channels (DIB created in BGR order) */
	{
	    void *r = pPixr->pixrData.childPlanes[0];

	    pPixr->pixrData.childPlanes[0] = pPixr->pixrData.childPlanes[2];
	    pPixr->pixrData.childPlanes[2] = r;
	}

	/* Convert the PIXR image data to DIB */
	XERR_PASS_CLEANUP(w_RGBPixrsToBMP(pIPWState, pPixr, 0,
	    jcprops.JPGHeight, dibImage));

	/* Free the temporary PIXR image */
/*	w_PixrFree(&pPixr); */


	/* Use it as the input */
	jcprops.DIBBytes = dibImage;

	/* Allocate a buffer for the compressed data */
	if (*ppCmpData == NULL || *pdwCmpBytes == 0) {
	    byteCount = 100 + byteCount / (101 - tfxImg->JPEGquality);
	    IO_CHECKMALLOC(jcprops.JPGBytes = IO_MALLOC(byteCount));
	    jcprops.JPGSizeBytes = byteCount;
	} else {
	    jcprops.JPGBytes = *ppCmpData;
	    jcprops.JPGSizeBytes = byteCount = *pdwCmpBytes;
	}

	/* Tell IJL there are fewer bytes than there actually are (it
	 * seems to overwrite the end of the buffer before failing) */
	jcprops.JPGSizeBytes -= 10;

	/* Use buffer mode */
	jcprops.JPGFile = NULL;

	/* Compress the image */
	while ((jerr = ijlWrite(&jcprops, IJL_JBUFF_WRITEWHOLEIMAGE)) == IJL_BUFFER_TOO_SMALL) {
	    /* Free the buffer */
	    IO_FREE(*ppCmpData);

	    /* Increase the buffer 50% */
	    byteCount = byteCount * 150 / 100;
	    IO_CHECKMALLOC(*ppCmpData = IO_MALLOC(byteCount));
	    jcprops.JPGBytes = *ppCmpData;
	    jcprops.JPGSizeBytes = *pdwCmpBytes = byteCount;

	    /* Tell IJL there are fewer bytes than there actually are (it
	     * seems to overwrite the end of the buffer before failing) */
	    jcprops.JPGSizeBytes -= 10;
	}
	if (jerr < IJL_OK) {
	    bres = FALSE;
	    __leave;
	}

	/* Grab the output buffer */
	*ppCmpData = jcprops.JPGBytes;
	*pdwBytesFilled = jcprops.JPGSizeBytes;

	/* Free the image */
	IO_FREE(dibImage), dibImage = NULL;

	/* Clean up the IJL */
	ijlFree(&jcprops);
    } /* __try */

    __finally {
	/* Clean up the IJL */
	ijlFree(&jcprops);
   }

    if (__error__ == IO_NOERROR && bres != FALSE) {
	NORMAL_RETURN;
    }
    
 ERR_CLEANUP:
    if (dibImage) {
	IO_FREE(dibImage), dibImage = NULL;
    }

    /* Reset the error (ignore it) */
    __error__ = IO_NOERROR;
    
    ERR_RETURN;
}

/*#endif	/* IJL */ */
#endif	/* NOIPCORE */


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Strip the tables from a JPEG (JFIF) stream, writing them if the first strip
 *
 * LongDescription:::
 *  wp_TfxStripJPEGTables() strips the tables from the JFIF stream if
 *  JPEGTables mode is on. After stripping, the tables are written to the
 *  JPEGTables field if they haven't been already written.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  pCmpData::			Pointer to pointer to the compressed image.
 *  pdwCmpBytes::		Number of bytes in compression buffer, to be
 *				modified with the size once tables are stripped.
 *
 * Output:::
 *  pdwCmpBytes::		Number of bytes left in the JFIF stream, once
 *				the tables have been stripped.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxStripJPEGTables(
    TFX_IMG *tfxImg,
    UInt8 **ppCmpData,
    UInt32 *pdwCmpBytes
)
{
    TFX_INSTANCE *instData = TFX_IMGINST(tfxImg);
    UInt8 *tp, *tables = NULL;
    UInt8 *sp, *eoipos;
    UInt32 bytes;
    PROLOGUE;

    XERR_GEN_CLEANUP(ppCmpData == NULL, IO_BADPARAMETER);

    if (*pdwCmpBytes == 0)
	NORMAL_RETURN;

    /* Create a buffer to hold the tables (worst case is that it's as big
     * as the entire stream, but add enough for an SOI/EOI pair) */
    IO_CHECKMALLOC(tables = IO_MALLOC(*pdwCmpBytes + 4));
    tp = tables + 2;

    /* Make sure the tables starts with an SOI */
    sp = *ppCmpData;
    XERR_GEN_CLEANUP(*sp != 0xff, IO_BADFILEFORMAT);
    sp++;
    XERR_GEN_CLEANUP(*sp != 216 /* SOI */, IO_BADFILEFORMAT);
    sp++;

    /* Loop through the markers until an EOI is found (might not be
     * the last byte in the stream) */
    eoipos = *ppCmpData + *pdwCmpBytes - 2;	/* Expected position */
    while (sp <= eoipos) {
	/* Is it a marker? */
	XERR_GEN_CLEANUP(*sp != 0xff, IO_BADFILEFORMAT);
		
	if (sp[1] == 0xDA) {	/* Start of scan */
	    /* We could assume there's no more table, since they must precede the data */
	    sp = eoipos;
	    break;
	}

	/* Look for markers that don't have segments */
	switch (sp[1]) {
	case 0x01:	/* TEM */
	case 0xD8:	/* SOI */
	case 0xD9:	/* EOI */
	    break;
	default:
	    bytes = (sp[2] << 8) | sp[3];
	    XERR_GEN_CLEANUP(bytes < 2, IO_BADFILEFORMAT);
	    XERR_GEN_CLEANUP(sp + bytes > eoipos, IO_BADFILEFORMAT);
	    break;
	}

	switch (sp[1]) {
	case 0x01:	/* TEM */
	case 0xD8:	/* SOI */
	case 0xD9:	/* EOI */
	    sp += 2;
	    /* Ignore */
	    break;
	case 0xDB:	/* DQT - Define Quantization Table */
	case 0xC4:	/* DHT - Define Huffman Table */
	    /* Is it a marker? */
	    XERR_GEN_CLEANUP(*sp != 0xff, IO_BADFILEFORMAT);

	    /* Copy this to the tables stream. Note: bytes includes byte count,
	     * but not marker itself. */
	    memcpy(tp, sp, bytes + 2);
	    tp += bytes + 2;

	    /* Copy the rest of the stream to where the table is
	     * (remove the table) */
	    memmove(sp, sp + 2 + bytes, eoipos + 2 - (sp + 2 + bytes));

	    /* Adjust the new EOI position */
	    eoipos -= bytes + 2;
	    break;
	default:
	    /* Skip the marker, byte count and data portions of all the other markers */
	    sp += 2 + bytes;
	}
    }

    /* Did we find the EOI? */
    XERR_GEN_CLEANUP(sp != eoipos, IO_BADFILEFORMAT);
    sp = eoipos + 2;

    /* If we're storing JPEGTables, use this set of tables */
    if (tfxImg->imgFlags & IO_IMGFLAGS_JPEGTables) {

	if (tfxImg->qtableBufsize == 0) {
	    TFX_FILE *tfxFile = TFX_IMGFILE(tfxImg);
	    Int32 seekOffset, offset;
	    UInt32 lval;

	    /* Add in the SOI/EOI pair */
	    tables[0] = 0xFF;
	    tables[1] = 0xD8;
	    tp[0] = 0xFF;
	    tp[1] = 0xD9;
	    tp += 2;

	    /* Copy to the image JPEGTables stream */
	    tfxImg->qtableBufsize = tp - tables;
	    XERR_GEN_CLEANUP(tfxImg->qtableBufsize <= 0, IO_INTERNAL_ERROR);
	    IO_CHECKMALLOC(tfxImg->qtableBuf =
		TFX_MALLOC(tfxImg->qtableBufsize));
	    memcpy(tfxImg->qtableBuf, tables, tfxImg->qtableBufsize);
	    IO_FREE(tables), tables = NULL;

	    /* Write the JPEGTables data */
	    if (tfxImg->qtableBufsize <= 4) {
		seekOffset = tfxImg->fieldLocation.JPEGTables;
		XERR_GEN_CLEANUP(seekOffset == 0, IO_INTERNAL_ERROR);
		TFX_SEEK(tfxFile, seekOffset);
	    } else {
		/* Put it at the end of the file */
		XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &offset));
		tfxImg->fieldDataLocation.JPEGTables = offset;
	    }
	    TFX_WRITE(tfxFile, tfxImg->qtableBuf, tfxImg->qtableBufsize);

	    /* Fill in the data location */
	    seekOffset = tfxImg->fieldLocation.JPEGTables;
	    XERR_GEN_CLEANUP(seekOffset <= 12, IO_INTERNAL_ERROR);
	    seekOffset -= 4;	/* Go back and fill the count in too */
	    TFX_SEEK(tfxFile, seekOffset);
	    lval = tfxImg->qtableBufsize;
	    TFX_SWAPLONG(tfxFile, &lval);
	    TFX_WRITE(tfxFile, &lval, sizeof(lval));
	    if (tfxImg->qtableBufsize > 4) {
		/* Location of the data, if not placed in the value/offset
		 * location already */
		lval = tfxImg->fieldDataLocation.JPEGTables;
		TFX_SWAPLONG(tfxFile, &lval);
		TFX_WRITE(tfxFile, &lval, sizeof(lval));
	    }
	}
    }

    if (tables) {
	IO_FREE(tables), tables = NULL;
    }

    /* Return the new stream size */
    *pdwCmpBytes = sp - *ppCmpData;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (tables) {
	IO_FREE(tables), tables = NULL;
    }
    ERR_RETURN;
}


/*
$Log: tfxwjpeg.c,v $
Revision 1.2  2003/02/10 22:53:32  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:14  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.17  2001/12/26 21:18:13  dabercro
Fix DIB padding.  Hack for IJL: tell IJL there are fewer bytes in compression
stream, since it overwrites data after the buffer before checking for the end
of the stream.

Revision 1.16  2001/12/12 00:07:18  dabercro
Add support for Intel's JPEG Library
Separate function to remove JPEG tables.

Revision 1.15  2001/11/14 00:08:19  dabercro
Comment fix.

Revision 1.14  2001/07/20 23:13:22  dabercro
Comments.

Revision 1.13  2001/07/05 21:46:06  dabercro
Support for the JPEGTables field

Revision 1.12  2001/06/27 18:54:45  dabercro
Remove colorspace calls to JPEG, moved to image conversion in tfxconv.c.

Revision 1.11  2001/03/13 02:19:46  dabercro
Remove quality fudge, now that sRGB is assumed.

Revision 1.10  2001/01/18 23:31:34  dabercro
Must clear the outer structure, since only the inner IPShared
structure gets cleared.

Revision 1.9  2001/01/18 00:21:31  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.8  2000/11/17 23:57:35  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.7  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.6  2000/08/22 01:03:24  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.5  2000/08/01 18:19:12  dabercro
Update read/write callbacks to reflect change in IPCore.

Revision 1.4  2000/06/07 00:00:17  dabercro
Continued work, lots of stuff.

Revision 1.3  2000/04/12 19:06:09  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.2  2000/03/11 02:06:35  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.14  1998/07/23 21:06:58  xzhu
xzhu on Thu Jul 23 14:06:41 PDT 1998

tfxwjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxwjpeg.c,v

Follow the change of w_jcInit.

Revision 1.13  1998/01/14 00:16:36  ddavies
tfxwjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxwjpeg.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.12  1997/11/20 00:16:53  aber
aber on Wed Nov 19 16:16:22 PST 1997

tfxwjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxwjpeg.c,v

Add new parameters to set JPEG stream X/Y resolution.

Revision 1.11  1997/09/02 20:02:55  aber
tfxwjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxwjpeg.c,v

CommonCheck: Rename functions to private names (wp_).

Revision 1.10  1997/06/26 22:28:38  aber
tfxwjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxwjpeg.c,v

Allow multiple Wavelet tiles per image, untested.

Revision 1.9  1997/06/17 01:29:30  aber
tfxwjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxwjpeg.c,v

CommonCheck: TIFF-FX 2.0 support. Annotation bug fix.

Revision 1.8  1997/06/10 21:16:29  xzhu
xzhu on Tue Jun 10 14:15:08 PDT 1997

tfxwjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxwjpeg.c,v

Followed the w_jcInit prototype changes.

Revision 1.7  1997/06/10 17:59:22  aber
tfxwjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxwjpeg.c,v

Remove commented out section of code.

Revision 1.6  1997/06/10 01:26:52  aber
tfxwjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxwjpeg.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:12  aber
tfxwjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxwjpeg.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:16  aber
tfxwjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxwjpeg.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:22  aber
tfxwjpeg.c -> /product/ipcore/ipwrap/src/RCS/tfxwjpeg.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:44  aber
Initial revision

*/
