/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxpixrw.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxpixrw.c:	TIFF-FX IPWrap WRITE source
 *
 *	TIFF-FX write-only IPWrap level routines. Calls the "Generic TIFF-FX"
 *  routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tfxwpixr.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:32 $");

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "props.pub"
#include "frames.pub"
#include "ccitt.pub"
#include "mmr.pub"
#include "shrcmprs.prv"
#include "imageref.pub"
#include "shrpixr.pub"
#include "shrpixr.prv"
#include "intops.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif

#include "tfximage.h"
#include "tfxwtiff.h"
#include "tfxrtfx.h"
#include "tfxwtfx.h"
#include "tfxwpriv.h"
#include "tfxrpriv.h"
#include "tfxwjpeg.h"
#include "tfxwjb2.h"
#include "tfx.pub"
#include "tfxconv.h"

#include "iomrc.h"


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the base color for an image
 *
 * LongDescription:::
 *  wp_TfxSetBaseColor() sets the base color for an image, if it hasn't already
 *  been set via the IFDBaseColor, ImageBaseColor, or previously using this
 *  routine.<BR>
 *  If appropriate, trigger the IFDBaseColor and/or ImageBaseColor field(s)
 *  to be written.
 *
 * Input:::
 *  tfxImg::			Valid image
 *  pRGB::			Pointer to an sRGB color: r = pRGB[0],
 *				g = pRGB[1], b = pRGB[2]
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxSetBaseColor(
	TFX_IMG *tfxImg,
	UInt8 *pRGB
)
{
    UInt32 channel, channels;
    Bool sameColor = 0;
    PROLOGUE;

    TFX_CheckHandle(tfxImg);

    /* Compute the number of channels */
    channels = IO_CHANNELS(tfxImg->imgInfo.xColorType);

    /* Determine the color that has been previously set, if any */
    if (! tfxImg->baseColor.color_set) {
	switch (tfxImg->tfxPage->tfxDoc->docInfo.profile) {
	case TFX_PROFILE_M1:
	case TFX_PROFILE_COH_M1:

            /* A TIFF-FX extension is being used, expect that the
	     * IFDBaseColor field indicates the prevailing color */
	    if (tfxImg->fields.found.IFDBaseColor) {
		UInt8 colorComponents[4];
	    
		/* Convert that color to sRGB.
		 * Note: this routine fills in the tfxImg->baseColor
		 * structure items */
		XERR_PASS_CLEANUP(wp_TfxGetBaseColor(tfxImg, &colorComponents[0]));

		tfxImg->baseColor.color_set = 1;
	    }
	    break;
	default:
	    /* Not an extension, the ImageBaseColor would indicate
	     * the base color */
	    if (tfxImg->fields.found.ImageBaseColor) {
	    
		/* Convert the color from LAB to sRGB */
		XERR_PASS_CLEANUP(IO_ITULAB2sRGB(
		    tfxImg->fields.ImageBaseColor[0],
		    tfxImg->fields.ImageBaseColor[1],
		    tfxImg->fields.ImageBaseColor[2],
		    &tfxImg->baseColor.rgb[0],
		    &tfxImg->baseColor.rgb[1],
		    &tfxImg->baseColor.rgb[2],
		    tfxImg->fields.Decode));

		tfxImg->baseColor.color_set = 1;
	    }
	}
    }	/* If no base color set */

    /* If previously set, there might be a problem in that a previous
     * image strip may have expected that the color be what it is now.
     */
    if (tfxImg->baseColor.color_set) {
	/* If it's the same color, we're okay */
	for (channel = 0; channel < channels; channel++) {
	    if (pRGB[channel] != tfxImg->baseColor.rgb[channel]) {
		sameColor = 0;
		break;
	    }
	}

	/* If it isn't the same color, then this routine was called twice
	 * in the same image, but with difference colors */
	XERR_GEN_CLEANUP(sameColor, IO_SEQUENCE);

	/* Otherwise, it's okay (it's the same color) */
	NORMAL_RETURN;
    } else {

	/* Okay, we can use the color that was passed in to this routine */
	tfxImg->baseColor.rgb[0] = pRGB[0];
	tfxImg->baseColor.rgb[1] = pRGB[1];
	tfxImg->baseColor.rgb[2] = pRGB[2];
	tfxImg->baseColor.color_set = 1;
    }

    /* Trigger the IFDBaseColor and/or ImageBaseColor field to be written,
     * if appropriate */
    switch (tfxImg->tfxPage->tfxDoc->docInfo.profile) {
    case TFX_PROFILE_M1:
    case TFX_PROFILE_COH_M1:

        /* A TIFF-FX extension is being used, we are required to set
       	 * the IFDBaseColor field. */

	/* If there is no current ImageBaseColor, use this one */
	if (! tfxImg->fields.found.IFDBaseColor) {
	    Int32 colorComponents[4];
	    UInt16 sval[4];
	    Int16 ssval[4];

	    /* Convert the sRGB color to the color value associated
	     * with the color description indicated by the IFD parameters:
	     * PhotometricInterpretation, SamplesPerPixel, BitsPerSample,
	     * and if appropriate: Decode.
	     */
	    switch (tfxImg->fields.PhotometricInterpretation) {
	    case TFX_TIFFPHOTOMETRIC_RGB:
		/* Already where we want to be */
		colorComponents[0] = tfxImg->baseColor.rgb[0];
		colorComponents[1] = tfxImg->baseColor.rgb[1];
		colorComponents[2] = tfxImg->baseColor.rgb[2];
		break;
	    case TFX_TIFFPHOTOMETRIC_CIELAB:
		XERR_PASS_CLEANUP(IO_sRGB2CIELAB(tfxImg->baseColor.rgb[0],
		    tfxImg->baseColor.rgb[1], tfxImg->baseColor.rgb[2],
		    &sval[0], &ssval[1], &ssval[2]));
		colorComponents[0] = sval[0];
		colorComponents[1] = ssval[1];
		colorComponents[2] = ssval[2];
		break;
	    case TFX_TIFFPHOTOMETRIC_ICCLAB:
		XERR_PASS_CLEANUP(IO_sRGB2ICCLAB(tfxImg->baseColor.rgb[0],
		    tfxImg->baseColor.rgb[1], tfxImg->baseColor.rgb[2],
		    &sval[0], &sval[1], &sval[2]));
		colorComponents[0] = sval[0];
		colorComponents[1] = sval[1];
		colorComponents[2] = sval[2];
		break;
	    case TFX_TIFFPHOTOMETRIC_ITULAB:
		XERR_PASS_CLEANUP(IO_sRGB2ITULAB(tfxImg->baseColor.rgb[0],
		    tfxImg->baseColor.rgb[1], tfxImg->baseColor.rgb[2],
		    &sval[0], &sval[1], &sval[2], tfxImg->fields.Decode));
		colorComponents[0] = sval[0];
		colorComponents[1] = sval[1];
		colorComponents[2] = sval[2];
		break;
	    case TFX_TIFFPHOTOMETRIC_CMYK:
	    default:
		/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */
		/* Not implemented yet */
		/* xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */
		XERR_PASS_CLEANUP(IO_NOSUPPORT);
	    }

	    /* Set the IFDBaseColor field */
	    XERR_PASS_CLEANUP(TFX_SetIFDBaseColor((IO_IMGHANDLE)tfxImg,
		&colorComponents[0], IO_CHANNELS(tfxImg->imgInfo.xColorType)));
			    
	    /* Write the ImageBaseColor if:
	     * 1. there was already an ImageBaseColor field (explicitly set
	     *    by the user
	     * or
	     * 2. we don't have a color image that requires TIFF-FX extension set 1 to 
	     *    be read anyways (i.e. JPEGTables is present).
	     */
	    if (tfxImg->fields.found.ImageBaseColor ||
		(tfxImg->imgFlags & IO_IMGFLAGS_JPEGTables) == 0) {
		UInt16 sval[4];

		XERR_PASS_CLEANUP(IO_sRGB2ITULAB(
		    tfxImg->baseColor.rgb[0],
		    tfxImg->baseColor.rgb[1],
		    tfxImg->baseColor.rgb[2],
		    &sval[0], &sval[1], &sval[2],
		    &tfxImg->fields.Decode[0]));
		XERR_PASS_CLEANUP(TFX_SetImageBaseColor((IO_IMGHANDLE)tfxImg,
		    &sval[0], IO_CHANNELS(tfxImg->imgInfo.xColorType)));
	    }
	    break;
	default:
	    if (! tfxImg->fields.found.ImageBaseColor) {
		UInt16 sval[4];

		XERR_PASS_CLEANUP(IO_sRGB2ITULAB(
		    tfxImg->baseColor.rgb[0],
		    tfxImg->baseColor.rgb[1],
		    tfxImg->baseColor.rgb[2],
		    &sval[0], &sval[1], &sval[2],
		    tfxImg->fields.Decode));
		XERR_PASS_CLEANUP(TFX_SetImageBaseColor((IO_IMGHANDLE)tfxImg,
		    &sval[0], IO_CHANNELS(tfxImg->imgInfo.xColorType)));
	    }
	}
    }

ERR_CLEANUP:
    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compress and write a strip of image data
 *
 * LongDescription:::
 *  w_TfxWriteStrip() compresses and writes a strip of image data. If necessary,
 *  the image data is broken up into TIFF strips, and stored as separate strips.
 *  The last strip may be stored until the next block of data is added to
 *  complete a TIFF strip.
 *
 * Input:::
 *  pIPWState::			IPW state
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_AddImage()
 *  pPixr::			Pointer to the image PIXR
 *  dwPixrRowOffset::		Row offset in PIXR of strip to write
 *  dwStripHeight::		Number of rows in PIXR to write
 *  dwImgRowOffset::		Row offset in image to write strip
 *
 * Output:::
 *  pdwRowsFilled::		Number of rows written
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_TfxWriteStrip(
	IPWState *pIPWState,
	IO_IMGHANDLE tfxImgHandle,
	PIXR *pPixr,
	UInt32 dwPixrRowOffset,
	UInt32 dwStripHeight,
	UInt32 dwImgRowOffset,
	UInt32 *pdwRowsFilled
)
{
    TFX_INSTANCE *instData;
    IPWState *internalIPWState = NULL;
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_IMGINFO *pTftfxImgInfo;
    UInt32 mallocBytes;
    UInt32 stripOffset;
    UInt32 channel, channels;
    UInt32 clientRowsLeft;
    UInt32 tiffRows;
    UInt32 rowsFilled;			    /* Rows filled in intermediate buffer */
    Int32 srcBPL, dstBPL;
    UInt8 *srcPtr, *dstPtr;		    /* Image data pointers */
    UInt32 totalRowsStored = 0UL;	    /* Number of rows stored (buffered/written) */
    UInt32 bytesWritten;
    Int32 pixrHeight;
    Bool onevalue = 0;
    Bool sameColor = 0;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;

    TFX_CheckHandle(tfxImgHandle);

    instData = TFX_IMGINST(tfxImg);    
    pTftfxImgInfo = &tfxImg->imgInfo;
    channels = IO_CHANNELS(pTftfxImgInfo->xColorType);
    
    /* Check for proper calling sequence */
    if ((tfxImg->seqStatus & TFX_IMG_SEQ_WRITESTATUS) != TFX_IMG_SEQ_WRITESTART) {
	__error__ = IO_SEQUENCE;
    }
    XERR_GEN_CLEANUP(dwImgRowOffset >= pTftfxImgInfo->dwHeight,
	IO_BADPARAMETER);

    /* Be kind, and ignore any extra that was given */
    if (dwImgRowOffset + dwStripHeight > pTftfxImgInfo->dwHeight) {
	dwStripHeight = pTftfxImgInfo->dwHeight - dwImgRowOffset;
    }
    clientRowsLeft = dwStripHeight;

    /* Save the internally used IPWState, and use the passed in one.
     * Restore when finished. */
    if (tfxImg->tfxPage->tfxDoc->tfxFile->instance->pIPWState != 0) {
        internalIPWState =
	    (IPWState *)tfxImg->tfxPage->tfxDoc->tfxFile->instance->pIPWState;
    }
    tfxImg->tfxPage->tfxDoc->tfxFile->instance->pIPWState = (UInt32)pIPWState;

    /* Create a compressed buffer, if we haven't already */
    if ((tfxImg->cmpBuffer == NULL) &&
	(pTftfxImgInfo->xCompression != IO_COMP_TRCE)) {

	/* Expect reasonable compression, unless storing uncompressed */
	if (pTftfxImgInfo->xCompression == IO_COMP_NOC) {
	    mallocBytes = tfxImg->minBytesPerRow * dwStripHeight *
		IO_CHANNELS(pTftfxImgInfo->xColorType);
	} else {
	    mallocBytes = (tfxImg->minBytesPerRow * dwStripHeight * 50) / 100;
	}
	if (mallocBytes < 1000) {
	    mallocBytes = 1000;
	}
	IO_CHECKMALLOC(tfxImg->cmpBuffer = TFX_MALLOC(mallocBytes));
	tfxImg->cmpBufsize = mallocBytes;
    }

    /* Get the size of the input PIXR */
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &pixrHeight));
    
    /* Continue filling intermediate buffer (if buffering) */
    /*
    CopyToTempImage(tfxImg, &pBuffer, &clientRowsLeft, dwBytesPerRow);
    */
    
    /* Loop over input data until it's exhausted */
    while (clientRowsLeft > 0) {

	/* Compute the TIFF strip size */
	if (pTftfxImgInfo->dwBestStripHeight > 0) {
	    stripOffset = dwImgRowOffset / pTftfxImgInfo->dwBestStripHeight;
	    if (stripOffset * pTftfxImgInfo->dwBestStripHeight +
		pTftfxImgInfo->dwBestStripHeight > pTftfxImgInfo->dwHeight) {
		tiffRows = pTftfxImgInfo->dwHeight -
		    stripOffset * pTftfxImgInfo->dwBestStripHeight;
	    } else {
		tiffRows = pTftfxImgInfo->dwBestStripHeight;
	    }
	} else {
	    tiffRows = dwStripHeight;
	}

	/* Determine if the data needs to be buffered:
	 * 1. If the number of rows doesn't fill a TIFF strip.
	 * 2. I have previously buffered data in an unwritten strip.
	 */
	if ((tfxImg->tmpPixrRows > 0 && tfxImg->tmpPixrRows < tiffRows)
	    || clientRowsLeft < tiffRows) {

	    /* If no intermediate buffer in place, create it */
	    if (tfxImg->tmpPixr == NULL) {
		UInt32 stripeOffset, stripeRow, stripeHeight;

		/* If this is the first buffered row, it should fall on
		 * a TIFF strip boundary */
		XERR_PASS_CLEANUP(TFX_GetStripeInfo((IO_PAGEHANDLE)tfxImg->tfxPage,
		    dwImgRowOffset, &stripeOffset, &stripeRow, &stripeHeight));
		XERR_GEN_CLEANUP(dwImgRowOffset % stripeRow, IO_SEQUENCE);

		/* Prepare to allocate the data */
		{
		    ScaledInt1616 xres, yres;
		    ColorType ctype = cColorNone;
		
		    /* Prepare a temporary PIXR */
		    xres = Int32_TO_ScaledInt1616(pTftfxImgInfo->dwXResolution);
		    yres = Int32_TO_ScaledInt1616(pTftfxImgInfo->dwYResolution);
		    switch ((int)pTftfxImgInfo->xColorType) {
		    case IO_COLORTYPE_COLOR24:
			ctype = cColorRGB;
			break;
		    case IO_COLORTYPE_BINARY:
		    case IO_COLORTYPE_GRAY4:
		    case IO_COLORTYPE_COLOR4:
		    case IO_COLORTYPE_GRAY8:
		    case IO_COLORTYPE_COLOR8:
			ctype = cColorGray;
			break;
		    default:
			goto ERR_CLEANUP;
		    }   
		    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, pTftfxImgInfo->dwWidth,
			tiffRows,
			IO_BPPE(pTftfxImgInfo->xColorType),
			xres, yres, 32,
			IO_CHANNELS(pTftfxImgInfo->xColorType),
			ctype, NULL, &tfxImg->tmpPixr));
		}
		tfxImg->tmpPixrRow = dwImgRowOffset;
		tfxImg->tmpPixrRows = 0UL;
	    } /* If creating an intermediate PIXR */

	    /* Make sure there are no gaps in the image data (adjacent calls
	     * didn't specify adjacent image data) */
	    XERR_GEN_CLEANUP(tfxImg->tmpPixrRow + tfxImg->tmpPixrRows !=
		dwImgRowOffset, IO_SEQUENCE);

	    /* Copy the rows relevant to this TIFF strip */
	    rowsFilled = 0;
	    for (channel = 0; channel < channels; channel++) {
		PIXR *pSubPixr;
		UInt32 rowsAvailable, byteCount;

		XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &srcBPL));
		XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, tfxImg->tmpPixr,
					       &dstBPL));
		if (channels > 1) {
		    pSubPixr = NULL;
		    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr,
						       channel,
			&pSubPixr));
		    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
						     &srcPtr));
		    pSubPixr = NULL;
		    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState,
						       tfxImg->tmpPixr, channel,
			&pSubPixr));
		    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
						     &dstPtr));
		} else {
		    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr,
						     &srcPtr));
		    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, tfxImg->tmpPixr,
						     &dstPtr));
		}
		srcPtr += dwPixrRowOffset * srcBPL;
		dstPtr += (dwImgRowOffset - tfxImg->tmpPixrRow) * dstBPL;
		rowsAvailable = tiffRows - (dwImgRowOffset - tfxImg->tmpPixrRow);
		if (rowsAvailable > clientRowsLeft) {
		    rowsAvailable = clientRowsLeft;
		}
		byteCount = MIN(srcBPL, dstBPL);

		/* Copy the PIXR image to the user's image */
		while (rowsAvailable--) {
		    memcpy(dstPtr, srcPtr, byteCount);
		    srcPtr += srcBPL;
		    dstPtr += dstBPL;
		    rowsFilled++;
		}
	    } /* for each channel */
	    rowsFilled /= channels;
	    tfxImg->tmpPixrRows += rowsFilled;
	    dwPixrRowOffset += rowsFilled;
	    dwImgRowOffset += rowsFilled;
	    clientRowsLeft -= rowsFilled;
	    totalRowsStored += rowsFilled;

	    if (pTftfxImgInfo->xCompression != IO_COMP_TRCE) {
		UInt8 RGBcolor[3], *rgb = &RGBcolor[0];

		onevalue = 0;

		if (IO_BPP(tfxImg->imgInfo.xColorType) > 1 &&
		    (tfxImg->seqStatus & TFX_IMG_SEQ_USECOLORSTRIPS)) {
		
		    /* Check to see if the image is all one value */
		    XERR_PASS_CLEANUP(w_isAllOneColor(pIPWState, pPixr,
			&onevalue, &rgb));
		}

		if (onevalue) {
		    UInt32 channel, channels = IO_CHANNELS(tfxImg->imgInfo.xColorType);

		    /* If it's not the default color already, store it as a new one */
		    if (! tfxImg->baseColor.color_set) {
			XERR_PASS_CLEANUP(wp_TfxSetBaseColor(tfxImg, rgb));
		    }

		    /* Compare the colors, until a mismatch is found */
		    sameColor = 1;
		    for (channel = 0; channel < channels; channel++) {
			if (rgb[channel] != tfxImg->baseColor.rgb[channel]) {
			    sameColor = 0;
			    break;
			}
		    }
		}

		if (onevalue && sameColor) {

		    /* Store a 0-byte strip */
		    tfxImg->cmpBufsize = 0;
		    tfxImg->cmpBytes = 0;
		    XERR_PASS_CLEANUP(wp_TifFXWriteStrip(tfxImg));
		} else {

		    /* Compress intermediate buffer if filled */
		    if (tfxImg->tmpPixrRows == tiffRows) {
			XERR_PASS_CLEANUP(w_TfxCompress(pIPWState, (UInt32)tfxImg,
			    pTftfxImgInfo->xCompression, pTftfxImgInfo->xColorType,
			    pTftfxImgInfo->xColorSpace, tfxImg->imgFlags,
			    tfxImg->tmpPixr, 0, pTftfxImgInfo->dwWidth,
			    tfxImg->tmpPixrRows, &tfxImg->cmpBuffer,
			    &tfxImg->cmpBufsize, &tfxImg->cmpBytes));
		    }
		}
	    }
	} else { /* Not using an intermediate buffer: */
	    UInt8 RGBcolor[3], *rgb = &RGBcolor[0];

	    onevalue = 0;

	    if (IO_BPP(tfxImg->imgInfo.xColorType) > 1 &&
		(tfxImg->seqStatus & TFX_IMG_SEQ_USECOLORSTRIPS)) {
		
		/* Check to see if the image is all one value */
		XERR_PASS_CLEANUP(w_isAllOneColor(pIPWState, pPixr,
		    &onevalue, &rgb));
	    }

	    if (onevalue) {

		/* If it's not the default color already, store it as a new one */
		XERR_PASS_CLEANUP(wp_TfxSetBaseColor(tfxImg, &RGBcolor[0]));

		/* Is this the same color as the base color? */
		{
		    Int32 colors = IO_CHANNELS(tfxImg->imgInfo.xColorType);
		    Int32 colorIndex;
	
		    sameColor = 1;
		    
		    /* Compare the colors, until a mismatch is found */
		    for (colorIndex = 0; colorIndex < colors; colorIndex++) {
			if (RGBcolor[colorIndex] != tfxImg->baseColor.rgb[colorIndex]) {
			    sameColor = 0;
			    break;
			}
		    }

		}
	    }

	    if (onevalue && sameColor) {

		/* Store a 0-byte strip */
		tfxImg->cmpBufsize = 0;
		tfxImg->cmpBytes = 0;
		XERR_PASS_CLEANUP(wp_TifFXWriteStrip(tfxImg));
	    } else {
		/* Compress from client image directly */
		XERR_PASS_CLEANUP(w_TfxCompress(pIPWState, (UInt32)tfxImg,
		    pTftfxImgInfo->xCompression, pTftfxImgInfo->xColorType,
		    pTftfxImgInfo->xColorSpace, tfxImg->imgFlags, pPixr,
		    dwPixrRowOffset, pTftfxImgInfo->dwWidth,
		    tiffRows, &tfxImg->cmpBuffer, &tfxImg->cmpBufsize,
		    &tfxImg->cmpBytes));
	    }
	    dwPixrRowOffset += tiffRows;
	    dwImgRowOffset += tiffRows;
	    clientRowsLeft -= tiffRows;
	    totalRowsStored += tiffRows;
	} /* Not using an intermediate buffer */

	/* Write the strip */
	if (tfxImg->cmpBytes) {
	    /* Enable writing compressed data */
	    tfxImg->imgFlags |= IO_IMGFLAGS_Compressed;

	    /* Write the compressed data */
	    XERR_PASS_CLEANUP(TFX_CompressedWriteStrip(tfxImgHandle,
		tfxImg->tmpPixrRow, tiffRows, tfxImg->cmpBuffer,
		tfxImg->cmpBytes, &bytesWritten));
	    	
	    /* Disable writing compressed data */
	    tfxImg->imgFlags &= ~IO_IMGFLAGS_Compressed;

	    /* Reset for next TIFF strip */
	    tfxImg->tmpPixrRow = dwImgRowOffset;
	    tfxImg->tmpPixrRows = 0;
	    tfxImg->cmpBytes = 0UL;
	}
    } /* While rows to write */

    /* Restore the state */
    if (internalIPWState != NULL) {
	tfxImg->tfxPage->tfxDoc->tfxFile->instance->pIPWState =
	    (UInt32)internalIPWState;
    }

    /* Return the number of rows buffered/written */
    *pdwRowsFilled += totalRowsStored;
    NORMAL_RETURN;
    
 ERR_CLEANUP:

    if (internalIPWState != NULL) {
	tfxImg->tfxPage->tfxDoc->tfxFile->instance->pIPWState =
	    (UInt32)internalIPWState;
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compress a strip of image data
 *
 * LongDescription:::
 *  w_TfxCompress() compresses a single image strip. Used as a standalone
 *  compressor, allows an application to retrieve the image data from file
 *  in raw compressed form, and decode individual strips on the fly as needed.
 *  <BR>
 *  Source image is compressed into the destination buffer. If the destination
 *  buffer is filled and would overflow, IO_BUFFEREXCEEDED is returned.
 *
 * Input:::
 *  pIPWState::			IPW state
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_AddImage()
 *  xCompression::		See IO_COMPRESSION types
 *  xColorType::		Type of color
 *  xColorSpace::		Color space to convert to during compression
 *				IO_COLORSPACE_<c> where <c> is one of:
 *				    CMYK, YCBCR, CIELAB, ICCLAB, or ITULAB
 *  pPixr::			Pointer to the source PIXR
 *  dwFlags::			IO_IMGFLAGS_xxx combo
 *  dwRowOffset::		Row offset in image of data to compress
 *  dwWidth::			Columns of COLORTYPE in dst image
 *  dwHeight::			Rows in dst image
 *
 * Output:::
 *  ppDst::			Pointer to the destination compressed data buffer,
 *				to be modified if compression requires more
 *  pdwBytesInDst::		Pointer to the byte count of destination buffer,
 *				to be modified if compression requires more
 *  pdwBytesFilled::		Actual bytes in destination filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
w_TfxCompress(
	IPWState *pIPWState,
	IO_IMGHANDLE tfxImgHandle,
	IO_COMPRESSION xCompression,
	IO_COLORTYPE xColorType,
	IO_COLORSPACE xColorSpace,
	UInt32 dwFlags,
	PIXR *pPixr,
	UInt32 dwRowOffset,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt8 **ppDst,
	UInt32 *pdwBytesInDst,
	UInt32 *pdwBytesFilled
)
{
    TFX_INSTANCE *instData;
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    Int32 channel, channels;
    UInt32 rowsLeft;
    UInt32 bytesLeft;
    Int32 pixrWidth;
    Int32 pixrHeight;
    Int32 srcBPL, dstBPL, bytesPerRow;
    UInt32 bytesUsed = 0UL;
    Int32 bytesFilled;
    Int32 colorType;
    Int32 colorDepth;
    PIXR *pSubPixr;
    UInt8 *imageChannels[3];
    UInt8 *src, *dst;
    UInt8 continued = 0;
    UInt8 stage;
    UInt8 *newbuf;
    UInt32 newbufSize;
    UInt32 conversion = 0;
    PROLOGUE;

    /* Initialize the return value */
    *pdwBytesFilled = 0UL;

    TFX_CheckHandle(tfxImgHandle);
    instData = TFX_IMGINST(tfxImg);

    /* Check buffers and sizes */
#if 0
/*    XERR_GEN_CLEANUP((pPixr == NULL || ppDst == NULL || *ppDst == NULL), */
/*		     IO_BADPARAMETER); */
#endif
    XERR_GEN_CLEANUP(pPixr == NULL || ppDst == NULL,
		     IO_BADPARAMETER);

    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, pPixr, &channels));
    XERR_PASS_CLEANUP(w_pixrGetType(pIPWState, pPixr, &colorType));
    XERR_PASS_CLEANUP(w_pixrGetWidth(pIPWState, pPixr, &pixrWidth));
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, pPixr, &pixrHeight));
    XERR_PASS_CLEANUP(w_pixrGetDepth(pIPWState, pPixr, &colorDepth));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &srcBPL));
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
				       &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
				     &imageChannels[channel]));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr,
				     &imageChannels[channel]));
	}
    }

    /* If there's no memory preallocated, initialize to something */
    if (*pdwBytesInDst == 0 || *ppDst == NULL) {
	if ((int)xCompression == IO_COMP_NOC) {
	    newbufSize = dwWidth * dwHeight * IO_BPP(xColorType) / 8;
	} else {
	    newbufSize = dwWidth * dwHeight * IO_BPP(xColorType) / 8 / 100;    /* 100:1 compression */
	}
	IO_CHECKMALLOC(newbuf = TFX_MALLOC(newbufSize));

	/* Replace old buffer */
	if (*ppDst) {
	    TFX_FREE(*ppDst);
	}
	*ppDst = newbuf;
	*pdwBytesInDst = newbufSize;
    }
    
    rowsLeft = dwHeight;
    bytesLeft = *pdwBytesInDst;
    dstBPL = tfxImg->minBytesPerRow;
    bytesPerRow = dstBPL;
    if (bytesPerRow > srcBPL) {
	bytesPerRow = srcBPL;
    }
    dst = *ppDst;


    /* Check existing format, and desired format, and adjust when mismatched
     */
    /* Deal with planar configuration of color images */
    if (xColorType == IO_COLORTYPE_COLOR24) {
	switch (tfxImg->fields.PlanarConfiguration) {
	case 2:	/* Planar:  RRRR  GGGG  BBBB */
	    if ((dwFlags & IO_IMGFLAGS_ColorFormat) == IO_IMGFLAGS_ColorChunky) {
		/* We're writing planar, they have chunky */
		switch (tfxImg->imgInfo.xCompression) {
		case IO_COMP_NOC:
		case IO_COMP_JPEG:	/* JPEG is always compressed from planar */
/*//////  Fix this one day.  We shouldn't have anyone request chunky */
		    XERR_PASS_CLEANUP(IO_NOSUPPORT);
		}
	    }
	    break;
	case 1:	/* Chunky: RGBRGBRGB... */
	default:
	    if ((dwFlags & IO_IMGFLAGS_ColorFormat) == IO_IMGFLAGS_ColorPlanar) {
		/* We're writing chunky, they have planar */
		switch (tfxImg->imgInfo.xCompression) {
		case IO_COMP_JPEG:	/* JPEG gets compressed from planar */
		    break;
		case IO_COMP_NOC:
		    conversion |= TFX_CONV_CHUNK;
		    break;
		default:
		    XERR_GEN_CLEANUP(1, IO_NOSUPPORT);
		}
	    }
	}
    }

    /* Mark as needing inverted, if photometric interpretation indicates opposite */
    switch (xColorType) {
    case IO_COLORTYPE_BINARY:
    case IO_COLORTYPE_GRAY4:
    case IO_COLORTYPE_GRAY8:
	switch (tfxImg->fields.PhotometricInterpretation) {
	case TFX_TIFFPHOTOMETRIC_BLACKISZERO:
	case TFX_TIFFPHOTOMETRIC_YCBCR:
	case TFX_TIFFPHOTOMETRIC_CIELAB:
	case TFX_TIFFPHOTOMETRIC_ICCLAB:
	case TFX_TIFFPHOTOMETRIC_ITULAB:
	    if ((dwFlags & IO_IMGFLAGS_PhotoInterp) ==
		IO_IMGFLAGS_WhiteIsZero) {
		/* They have white-is-zero, we're writing black-is-zero */
		conversion |= TFX_CONV_INVERT;
	    }
	    break;
	case TFX_TIFFPHOTOMETRIC_WHITEISZERO:
	    if ((dwFlags & IO_IMGFLAGS_PhotoInterp) ==
		IO_IMGFLAGS_BlackIsZero) {
		/* They have black-is-zero, we're writing white-is-zero */
		conversion |= TFX_CONV_INVERT;
	    }
	    break;
	}
	break;
    case IO_COLORTYPE_COLOR24:
	break;
    default:
	break;
    } /* switch color type */

    /* Convert the colorspace, if needed */
    switch (xColorType) {
    case IO_COLORTYPE_BINARY:
	break;
    case IO_COLORTYPE_GRAY4:
	break;
    case IO_COLORTYPE_GRAY8:
	switch (xCompression) {
	case TFX_TIFFCOMP_NOCOMPRESSBYTE:
	case TFX_TIFFCOMP_JPEG:
	    switch (tfxImg->fields.PhotometricInterpretation) {
	    case TFX_TIFFPHOTOMETRIC_ITULAB:
		conversion |= TFX_CONV_SRGB2ITULAB;
		break;
	    default:
		break;
	    }
	    break;
	}
	break;
    case IO_COLORTYPE_COLOR24:
	switch (xCompression) {
	case IO_COMP_NOC:
	case IO_COMP_JPEG:
	    switch (tfxImg->fields.PhotometricInterpretation) {
	    case TFX_TIFFPHOTOMETRIC_CIELAB:
		conversion |= TFX_CONV_SRGB2CIELAB;
		break;
	    case TFX_TIFFPHOTOMETRIC_ICCLAB:
		conversion |= TFX_CONV_SRGB2ICCLAB;
		break;
	    case TFX_TIFFPHOTOMETRIC_ITULAB:
		conversion |= TFX_CONV_SRGB2ITULAB;
		break;
	    default:
		break;
	    }
	    break;
	}
	break;
    default:
	break;
    } /* switch color type */

    /* If input is byte-order, convert to PIXR long-packed bytes */
    if (tfxImg->imgFlags & IO_IMGFLAGS_ByteOrder) {
	conversion |= TFX_CONV_BYTE2DWORDORDER;
    }

    /* Do all the conversions */
    if (conversion) {
	XERR_PASS_CLEANUP(w_TfxConvertImage(tfxImg, pPixr, conversion, dwRowOffset,
	    dwHeight));
    }

    switch ((int)xCompression) {
    case IO_COMP_NOC:

	XERR_GEN_CLEANUP((bytesPerRow * dwHeight > bytesLeft),
		 IO_BUFFEREXCEEDED);
	XERR_PASS(wp_TfxPackImage(pIPWState, imageChannels, srcBPL, dwRowOffset,
	    dwHeight, channels, dst, dstBPL, (UInt32 *)&bytesFilled));
	bytesUsed += bytesFilled;
	break;
    case IO_COMP_MH:
	{
	    void *compInfo;
	    UInt32 imgExh, cmpExh;
	    Int32 pct = 0;
	    
	    if (tfxImg->tmpCmpData == NULL) {	/* First strip */
		Int32 stateBlockSize;
		UInt32 padType = cPadToByte;
		
		/* Prepare for pre-decode */
		XERR_PASS_CLEANUP(i_getSizeOfCCITTCompressionInfo(
		    pIPWState, dwWidth + 2*cFrameBits, &stateBlockSize));
		
		/* Allocate temporary compression data, attach to image */
		IO_CHECKMALLOC(tfxImg->tmpCmpData =
		     (TFX_MISCMEM *)TFX_MALLOC(sizeof(TFX_MISCMEM) +
				    stateBlockSize));
		tfxImg->tmpCmpData->next = NULL;
		
		/* Data follows immediately */
		tfxImg->tmpCmpData->mem = compInfo = (void *)(tfxImg->tmpCmpData + 1);
		
		XERR_PASS_CLEANUP(i_initializeCCITTCompressionInfo(
		    pIPWState, compInfo, dwWidth + 2*cFrameBits, 0,
		    0, padType,
		    stateBlockSize));
	    } else {
		compInfo = tfxImg->tmpCmpData->mem;
	    }

	    src = imageChannels[0] - (cFrameBits * IO_BPPE(xColorType)) / 8;
	    continued = 0;
	    bytesUsed = 0UL;
	    stage = 1;
	    do {
		bytesFilled = 0;
		switch (stage) {
		case 1:
		    /* Compress using MMR, don't flush yet */
		    XERR_PASS_CLEANUP(i_compressRowsMH (pIPWState,
		        compInfo, src,
			dwWidth + 2 * cFrameBits, dwRowOffset, rowsLeft,
			srcBPL, dst, bytesLeft, continued, continued,
			0, &imgExh, &cmpExh, &pct, &bytesFilled));

		    if (imgExh == 0) {
			XERR_GEN_CLEANUP((cmpExh == 0), IO_INTERNAL_ERROR);
		    } else {
			stage++;
		    }

		    break;
		case 2:
		    /* Flush the rest of the data */
		    XERR_PASS_CLEANUP(i_flushCCITT (pIPWState,
			compInfo, dst, bytesLeft,
			&bytesFilled));
		    stage++;
		    break;
		}
		bytesUsed += bytesFilled;
		bytesLeft -= bytesFilled;
		dst += bytesFilled;

		/* Increase the compressed buffer size, if needed */
		if ((imgExh == 0 && cmpExh == 1) || bytesLeft < 10) {

		    XERR_GEN_CLEANUP(imgExh && cmpExh == 0 && bytesLeft >= 10,
				     IO_INTERNAL_ERROR);

		    /* Filled the compressed buffer, reallocate the compressed
		     * data buffer */

		    /* Anticipate the compression due to the percentage, plus a
		     * fudge */
		    if (pct > 0 && pct <= 100) {
			newbufSize = *pdwBytesInDst * (100 + 10) / pct;
		    } else {
			newbufSize = *pdwBytesInDst * (200 - pct + 10) / 100;
		    }
		    IO_CHECKMALLOC(newbuf = TFX_MALLOC(newbufSize));
		    memcpy(newbuf, *ppDst, bytesUsed);

		    /* Replace old buffer */
		    TFX_FREE(*ppDst);
		    *ppDst = newbuf;
		    *pdwBytesInDst = newbufSize;
		    dst = newbuf + bytesUsed;
		    bytesLeft = newbufSize - bytesUsed;
		    continued = 1;
		}
	    } while (imgExh == 0 || stage < 3);

	    /* Can't compress given the buffer size, reallocate and go back
	     * in */
	    XERR_GEN_CLEANUP((cmpExh && imgExh == 0), IO_BUFFEREXCEEDED);
	}	/* G4 write */
	break;
    case IO_COMP_MR:
	{
	    void *compInfo;
	    UInt32 imgExh, cmpExh;
	    Int32 pct = 0;
	    
	    if (tfxImg->tmpCmpData == NULL) {	/* First strip */
		Int32 stateBlockSize;
		UInt32 padType = cNoPad;
		
		/* Prepare for pre-decode */
		XERR_PASS_CLEANUP(i_getSizeOfCCITTCompressionInfo(
		    pIPWState, dwWidth + 2*cFrameBits, &stateBlockSize));
		
		/* Allocate temporary compression data, attach to image */
		IO_CHECKMALLOC(tfxImg->tmpCmpData =
		     (TFX_MISCMEM *)TFX_MALLOC(sizeof(TFX_MISCMEM) +
				    stateBlockSize));
		tfxImg->tmpCmpData->next = NULL;
		
		/* Data follows immediately */
		tfxImg->tmpCmpData->mem = compInfo = (void *)(tfxImg->tmpCmpData + 1);
		
		XERR_PASS_CLEANUP(i_initializeCCITTCompressionInfo(
		    pIPWState, compInfo, dwWidth + 2*cFrameBits, 1,
		    (tfxImg->imgInfo.dwXResolution > 150 ? 4 : 2), padType,
		    stateBlockSize));
	    } else {
		compInfo = tfxImg->tmpCmpData->mem;
	    }

	    src = imageChannels[0] - (cFrameBits * IO_BPPE(xColorType)) / 8;
	    continued = 0;
	    bytesUsed = 0UL;
	    stage = 1;
	    do {
		bytesFilled = 0;
		switch (stage) {
		case 1:
		    /* Compress using MMR, don't flush yet */
		    XERR_PASS_CLEANUP(i_compressRowsMR (pIPWState,
		        compInfo, src,
			dwWidth + 2 * cFrameBits, dwRowOffset, rowsLeft,
			srcBPL, dst, bytesLeft, continued, continued,
			&imgExh, &cmpExh, &pct, &bytesFilled));

		    if (imgExh == 0) {
			XERR_GEN_CLEANUP((cmpExh == 0), IO_INTERNAL_ERROR);
		    } else {
			stage++;
		    }

		    break;
		case 2:
		    /* Add the two EOL markers required by TIFF G4 */
#if 0
		    XERR_PASS_CLEANUP(i_writeCCITTEOL (pIPWState,
			compInfo, cNoPad,
			cPlainEOL, dst, bytesLeft, &bytesFilled));
#endif
		    stage++;
		    break;
		case 3:
		    /* Flush the rest of the data */
		    XERR_PASS_CLEANUP(i_flushCCITT (pIPWState,
			compInfo, dst, bytesLeft,
			&bytesFilled));
		    stage++;
		    break;
		}
		bytesUsed += bytesFilled;
		bytesLeft -= bytesFilled;
		dst += bytesFilled;

		/* Increase the compressed buffer size, if needed */
		if ((imgExh == 0 && cmpExh == 1) || bytesLeft < 10) {

		    XERR_GEN_CLEANUP(imgExh && cmpExh == 0 && bytesLeft >= 10,
				     IO_INTERNAL_ERROR);

		    /* Filled the compressed buffer, reallocate the compressed
		     * data buffer */

		    /* Anticipate the compression due to the percentage, plus a
		     * fudge */
		    if (pct > 0 && pct <= 100) {
			newbufSize = *pdwBytesInDst * (100 + 10) / pct;
		    } else {
			newbufSize = *pdwBytesInDst * (200 - pct + 10) / 100;
		    }
		    IO_CHECKMALLOC(newbuf = TFX_MALLOC(newbufSize));
		    memcpy(newbuf, *ppDst, bytesUsed);

		    /* Replace old buffer */
		    TFX_FREE(*ppDst);
		    *ppDst = newbuf;
		    *pdwBytesInDst = newbufSize;
		    dst = newbuf + bytesUsed;
		    bytesLeft = newbufSize - bytesUsed;
		    continued = 1;
		}
	    } while (imgExh == 0 || stage < 4);

	    /* Can't compress given the buffer size, reallocate and go back
	     * in */
	    XERR_GEN_CLEANUP((cmpExh && imgExh == 0), IO_BUFFEREXCEEDED);
	}	/* G4 write */
	break;
    case IO_COMP_MMR:
	{
	    void *compInfo;
	    UInt32 imgExh, cmpExh;
	    Int32 pct = 0;
	    
	    if (tfxImg->tmpCmpData == NULL) {	/* First strip */
		Int32 stateBlockSize;
		UInt32 padType = cNoPad;
		
		/* Prepare for pre-decode */
		XERR_PASS_CLEANUP(i_getSizeOfCCITTCompressionInfo(
		    pIPWState, dwWidth + 2*cFrameBits, &stateBlockSize));
		
		/* Allocate temporary compression data, attach to image */
		IO_CHECKMALLOC(tfxImg->tmpCmpData =
		     (TFX_MISCMEM *)TFX_MALLOC(sizeof(TFX_MISCMEM) +
				    stateBlockSize));
		tfxImg->tmpCmpData->next = NULL;
		
		/* Data follows immediately */
		tfxImg->tmpCmpData->mem = compInfo = (void *)(tfxImg->tmpCmpData + 1);
		
		XERR_PASS_CLEANUP(i_initializeCCITTCompressionInfo(
		    pIPWState, compInfo, dwWidth + 2*cFrameBits, 1,
		    (tfxImg->imgInfo.dwXResolution > 150 ? 4 : 2), padType,
		    stateBlockSize));
	    } else {
		compInfo = tfxImg->tmpCmpData->mem;
	    }

	    src = imageChannels[0] - (cFrameBits * IO_BPPE(xColorType)) / 8;
	    continued = 0;
	    bytesUsed = 0UL;
	    stage = 1;
	    do {
		bytesFilled = 0;
		switch (stage) {
		case 1:
		    /* Compress using MMR, don't flush yet */
		    XERR_PASS_CLEANUP(i_compressRowsMMR (pIPWState,
		        compInfo, src,
			dwWidth + 2 * cFrameBits, dwRowOffset, rowsLeft,
			srcBPL, dst, bytesLeft, continued, continued,
			&imgExh, &cmpExh, &pct, &bytesFilled));

		    if (imgExh == 0) {
			XERR_GEN_CLEANUP((cmpExh == 0), IO_INTERNAL_ERROR);
		    } else {
			stage++;
		    }

		    break;
		case 2:
		    /* Add the two EOL markers required by TIFF G4 */
		    XERR_PASS_CLEANUP(i_writeCCITTEOL (pIPWState,
			compInfo, cNoPad,
			cPlainEOL, dst, bytesLeft, &bytesFilled));
		    stage++;
		    break;
		case 3:
		    /* Flush the rest of the data */
		    XERR_PASS_CLEANUP(i_flushCCITT (pIPWState,
			compInfo, dst, bytesLeft,
			&bytesFilled));
		    stage++;
		    break;
		}
		bytesUsed += bytesFilled;
		bytesLeft -= bytesFilled;
		dst += bytesFilled;

		/* Increase the compressed buffer size, if needed */
		if ((imgExh == 0 && cmpExh == 1) || bytesLeft < 10) {

		    XERR_GEN_CLEANUP(imgExh && cmpExh == 0 && bytesLeft >= 10,
				     IO_INTERNAL_ERROR);

		    /* Filled the compressed buffer, reallocate the compressed
		     * data buffer */

		    /* Anticipate the compression due to the percentage, plus a
		     * fudge */
		    if (pct > 0 && pct <= 100) {
			newbufSize = *pdwBytesInDst * (100 + 10) / pct;
		    } else {
			newbufSize = *pdwBytesInDst * (200 - pct + 10) / 100;
		    }
		    IO_CHECKMALLOC(newbuf = TFX_MALLOC(newbufSize));
		    memcpy(newbuf, *ppDst, bytesUsed);

		    /* Replace old buffer */
		    TFX_FREE(*ppDst);
		    *ppDst = newbuf;
		    *pdwBytesInDst = newbufSize;
		    dst = newbuf + bytesUsed;
		    bytesLeft = newbufSize - bytesUsed;
		    continued = 1;
		}
	    } while (imgExh == 0 || stage < 4);

	    /* Can't compress given the buffer size, reallocate and go back
	     * in */
	    XERR_GEN_CLEANUP((cmpExh && imgExh == 0), IO_BUFFEREXCEEDED);
	}	/* G4 write */
	break;
    case IO_COMP_JPEG:
	if ((tfxImg->imgFlags & IO_IMGFLAGS_IPCoreJPEG) ||
	    IO_BPP(tfxImg->imgInfo.xColorType) != 24) {
	    XERR_PASS_CLEANUP(wp_TfxCompressStripJPEG(tfxImg, pPixr, dwRowOffset,
		rowsLeft, xColorSpace, ppDst, pdwBytesInDst, (UInt32 *)&bytesFilled));
	} else {
	    XERR_PASS_CLEANUP(wp_TfxCompressStripJPEG_IJL(tfxImg, pPixr, dwRowOffset,
		rowsLeft, xColorSpace, ppDst, pdwBytesInDst, (UInt32 *)&bytesFilled));
	}

	bytesUsed += bytesFilled;
	break;
    case IO_COMP_LLTOKENIZATION:
    case IO_COMP_TOKENIZATION:
    case IO_COMP_JBIG2:
    case IO_COMP_LLJBIG2:
	XERR_PASS_CLEANUP(IO_NOSUPPORT);
	break;
    case IO_COMP_OTHER:
	/* Note: if you've specified your own compression, then you should be calling
	 * ..._CompressedWriteStrip() */
	XERR_PASS_CLEANUP(IO_NOSUPPORT);
	break;
    default:
	XERR_PASS_CLEANUP(IO_NOSUPPORT);
	break;
    }

    /* Return the compressed buffer size */
    *pdwBytesFilled = bytesUsed;

    /* Set the appropriate mode */
#if 0
/*    tfxImg->imgFlags |= IO_IMGFLAGS_Compressed; */
#endif


    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy image data in PIXR order to a byte-packed image stream
 *
 * LongDescription:::
 *  wp_TfxPackImage() copies image data from a PIXR to a byte-packed image
 *  buffer. This is used to store the "uncompressed" form of data, independent
 *  of platform. If the source image is multi-channel, then the destination
 *  image is one channel after another.
 *  <BR>
 *  The source image buffers are left intact.
 *
 * Input:::
 *  pIPWState::			Pointer to shared IPCore state
 *  srcChannels::		Input image channels (1-3)
 *  srcBPL::			Bytes per line of data (each channel)
 *  srcRowOffset::		Source row offset
 *  height::			Number of lines to transfer
 *  channels::			Number of channels to transfer
 *
 * Output:::
 *  dst::			Destination image buffer
 *  dstBPL::			Destination image bytes per line
 *  pBytesFilled::		Bytes filled in destination image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxPackImage(
	     IPWState *pIPWState,
	     UInt8 **srcChannels,
	     Int32 srcBPL,
	     UInt32 srcRowOffset,
	     Int32 height,
	     Int32 channels,
	     UInt8 *dst,
	     Int32 dstBPL,
	     UInt32 *pBytesFilled
)
{
    GraySwapNearProc swapLineProc;
    UInt8 *src;
    Int32 channel;
    Int32 row, lineBPL;
    UInt8 *line = NULL;
    PROLOGUE;

    XERR_GEN_CLEANUP(srcBPL < dstBPL, IO_BADPARAMETER);

    ip_getByteSwapProc (pIPWState, 0, cPixrToPixr,
			NULL, NULL, &swapLineProc);

    /* Compute the intermediate line BPL */
    lineBPL = (dstBPL + 3) & ~3;

    /* Allocate an intermediate line buffer (for swapping) */
    XERR_GEN_CLEANUP(((line = (UInt8 *)malloc(lineBPL)) == NULL),
		      IO_NOMEMORY);

    for (channel = 0; channel < channels; channel++) {
      src = srcChannels[channel] + srcRowOffset * srcBPL;
      for (row = 0; row < height; row++) {

	/* Swap the source line to the intermediate one */
	(*swapLineProc)(pIPWState, (void *)src, (void *)line, lineBPL / 4);
	src += srcBPL;

	/* Copy the intermediate line to the destination, byte packing */
	memcpy(dst, line, dstBPL);
	dst += dstBPL;
      }
    }

    free(line);

    *pBytesFilled = height * dstBPL * channels;
    NORMAL_RETURN;

 ERR_CLEANUP:
    if (line) {
      free(line);
    }

     ERR_RETURN;
}

/*
$Log: tfxwpixr.c,v $
Revision 1.2  2003/02/10 22:53:32  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:14  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.23  2001/12/26 21:19:19  dabercro
Only use IJL if 24-bit images (doesn't work for 8- or 4-bit yet).

Revision 1.22  2001/12/12 00:06:15  dabercro
Add option to choose which JPEG codec to use

Revision 1.21  2001/11/07 21:48:15  dabercro
Make base color signed long, to accomodate CIELAB

Revision 1.20  2001/10/30 20:05:07  dabercro
Add in writing of MH and MR images.

Revision 1.19  2001/08/22 19:49:34  dabercro
Added support for new field: IFDBaseColor

Revision 1.18  2001/06/27 18:53:39  dabercro
Setup and do image conversions colorspace, chunk/planar, invert, etc.

Revision 1.17  2001/03/13 03:02:16  dabercro
Remove wavelet header.

Revision 1.16  2001/03/13 02:59:13  dabercro
Remove lossy-dither and wavelet as valid compression types.

Revision 1.15  2001/02/24 03:27:06  dabercro
Remove wavelets altogether

Revision 1.14  2001/01/18 00:21:31  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.13  2000/11/17 23:57:35  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.12  2000/10/18 20:28:50  dabercro
Change logic for 0-byte strips, to equate to an all zero image.

Revision 1.11  2000/10/16 18:22:54  dabercro
Major adds to support stripes, 0-byte strips, and ImageBaseColor.

Revision 1.10  2000/09/27 21:56:45  dabercro
Variable strips and stripe work.

Revision 1.9  2000/09/22 21:50:21  dabercro
Variable-sized strips now handle multiple layers.

Revision 1.8  2000/09/21 21:19:14  dabercro
Add variable-sized strips: StripRowCounts.

Revision 1.7  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.6  2000/08/30 23:29:07  dabercro
Check for both JBIG2 and Tokenization.

Revision 1.5  2000/08/29 22:35:59  dabercro
Allow other compression types.

Revision 1.4  2000/08/22 01:03:24  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.3  2000/06/07 00:00:17  dabercro
Continued work, lots of stuff.

Revision 1.2  2000/04/12 19:06:09  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.24  1998/08/04 00:21:35  aber
tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

CommonCheck: Support multiple dictionaries when token compressing, used to retain
single-use tokens in the same dictionary -- to be used when color
tokenizing.

Revision 1.23  1998/05/08 05:59:35  aber
aber on Thu May 7 22:58:20 PDT 1998

tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

Limit how small the estimated compression buffer will be to be at least
1000 bytes. Found a case where a 1-line strip estimates to 0 bytes, which
returns NULL from malloc, which signals an error condition.

Revision 1.22  1998/05/05 10:00:37  aber
aber on Tue May 5 02:59:24 PDT 1998

tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

One more G4 cleanup. Added special case for output compression size
when leaving uncompressed.

Revision 1.21  1998/05/05 08:45:09  aber
aber on Tue May 5 01:44:35 PDT 1998

tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

Logic fix: avoid generating an error when just adding a small
cushion for G4 compression.

Revision 1.20  1998/05/05 07:52:53  aber
aber on Tue May 5 00:52:01 PDT 1998

tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

Changed logic in G4 compress to add a cushion when adding the
EOL marker. Found a case where it could fail.

Revision 1.19  1998/05/05 07:38:53  aber
aber on Tue May 5 00:37:28 PDT 1998

tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

Modified compression expansion equation when G4 compressed buffer
estimate was off, improves efficiency on a test case I looked at.

Revision 1.18  1998/05/05 07:15:28  aber
tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

CommonCheck: Additions necessary to fill in the never completed functions
TFX_Compress and TFX_Decompress.

Revision 1.17  1998/04/06 22:44:50  aber
aber on Mon Apr 6 15:44:02 PDT 1998

tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

To make things work potentially a little smoother, allow the client to
pass in more data than specified in the image, but ignore the extra.

Revision 1.16  1998/03/16 21:03:24  aber
tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

Fix to uncompressed image write (recompute byte count when storing
3-channel imagery). Normally RGB images are compressed, but we should
still handle uncompressed mode.

Revision 1.15  1998/01/14 00:16:38  ddavies
tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.14  1997/11/01 20:59:22  aber
tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

Add strip buffering for wavelet compression.

Revision 1.13  1997/07/29 23:52:05  aber
tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

Fix bug where G4 data < 10 bytes, but failure to read.

Revision 1.12  1997/07/15 23:43:20  aber
tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

Redo the buffering logic (finish it).

Revision 1.11  1997/06/24 17:44:21  aber
aber on Tue Jun 24 10:43:49 PDT 1997

tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

Less-than should be greater-than.

Revision 1.10  1997/06/23 22:59:01  aber
aber on Mon Jun 23 15:59:00 PDT 1997

tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

CommonCheck: Complete function calls for paletted image support.

Revision 1.9  1997/06/23 16:54:58  aber
aber on Mon Jun 23 09:54:32 PDT 1997

tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

Reference previously "unused variable".

Revision 1.8  1997/06/20 17:47:32  aber
tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

Added parameter check.

Revision 1.7  1997/06/17 22:25:27  aber
aber on Tue Jun 17 15:24:31 PDT 1997

tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

Use "compression buffer continued" = 1 for CCITT compression. Bug fix.

Revision 1.5  1997/06/10 01:26:54  aber
tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.4  1997/05/30 00:51:13  aber
tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.3  1997/05/20 22:29:17  aber
tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.2  1997/05/20 00:54:24  aber
tfxwpixr.c -> /product/ipcore/ipwrap/src/RCS/tfxwpixr.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:45  aber
Initial revision

*/
