/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxwpriv.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxwpriv.c:	TIFF-FX WRITE private routines
 *
 *	TIFF-FX private miscellaneous routines
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tfxwpriv.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:32 $");

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif
#endif	/* NOIPCORE */

#include "tfximage.h"
#include "tfxwtiff.h"
#include "tfxwtfx.h"
#include "tfxwpriv.h"

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Align a file to a "long" byte count
 *
 * LongDescription:::
 *  wp_MrcFileLongAlign() adjusts the size of the file such that the total byte
 *  count is divisible by 4. Empty bytes are used to pad the end of the file
 *  to adjust the file size.
 *
 * Input:::
 *  tfxFile::			Active file token
 *
 * Output:::
 *  offset::			Offset of the resulting file size
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxFileLongAlign(
	      TFX_FILE *tfxFile,
	      Int32 *offset
)
{
    UInt32 zero = 0UL;
    PROLOGUE;

    /* Make sure we aren't in "queue" mode (queuing up the image data, pending
     * a final IFD construction) */
/*    XERR_GEN_CLEANUP(tfxFile->bQueuing, IO_INTERNAL_ERROR); */

    TFX_FILESIZE(tfxFile, *offset);
    TFX_SEEK(tfxFile, *offset);
    if (*offset & 3) {
	TFX_WRITE(tfxFile, &zero, 4 - (*offset & 3));
	*offset = (*offset + 3) & ~3;
    }

    NORMAL_RETURN;
ERR_CLEANUP:

    ERR_RETURN;
}

/*
$Log: tfxwpriv.c,v $
Revision 1.2  2003/02/10 22:53:32  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:14  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.5  2001/12/26 21:23:45  dabercro
Mark as released

Revision 1.4  2001/01/18 00:21:31  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.3  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.2  2000/04/12 19:06:09  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.6  1997/06/12 00:52:13  aber
tfxwpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxwpriv.c,v

Actually seek to the end of the file when long aligning, instead
of assuming that the file size routine seeks there.

Revision 1.5  1997/05/30 00:51:15  aber
tfxwpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxwpriv.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:19  aber
tfxwpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxwpriv.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:27  aber
tfxwpriv.c -> /product/ipcore/ipwrap/src/RCS/tfxwpriv.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:46  aber
Initial revision

*/
