/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxwpub.c
 *
 *	TIFF-FX write public "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tfxwpub.c,v $; $Revision: 1.2 $; $Date: 2003/02/10 22:53:32 $");

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "props.pub"
#include "frames.pub"
#include "ccitt.pub"
#include "mmr.pub"
#include "shrcmprs.prv"
#include "imageref.pub"
#include "shrpixr.pub"
#include "shrpixr.prv"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif
#endif	/* NOIPCORE */

#include "tfximage.h"
#include "tfxwtiff.h"
#include "tfxwtfx.h"
#include "tfxwpriv.h"
#include "tfxrtfx.h"
#include "tfxwpub.h"
#include "tfxrpriv.h"
#include "tfxrjb2.h"
#include "tfxconv.h"
#ifndef NOIPCORE
#include "tfxwjb2.h"
#include "tfx.pub"
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy token dictionaries associated with a position block to a new document
 *
 * LongDescription:::
 *  wp_TfxCopyTokenDictionary() copies the token dictionary/ies, associated
 *  with a position block, from one TIFF-FX document to another; and modifies
 *  the token dictionary IDs found in the Tundra position/ID stream.  The
 *  token dictionary is left in Tundra format during the copy, although the
 *  dictionary IDs associated with the Tundra dictionaries may change during
 *  the transfer.  This routine is used for image/page coping, and allows
 *  much faster copies.
 *  <BR>
 *  Since token dictionaries span more than one page/image, there is some
 *  internal states that are maintained to keep track of which token
 *  dictionaries have already been copied.  These are keyed off the ID of
 *  the document handle to which the dictionary is copied to, and
 *  maintained in the document handle of the originating dictionary.
 *  <BR>
 *  If the dictionary has previously been copied, the copy is not repeated.
 *  The new dictionary ID is returned, and replaces the ID stored in the
 *  Tundra position/ID header of the input stream.
 *
 * Input:::
 *  srcImg::			Source image handle, with Tundra
 *				position/ID block
 *  imgData::			Shared data position/ID block (header
 *				containing old dictionary ID will be
 *				updated)
 *  dataBytes::			Number of bytes in imgData
 *
 * Output:::
 *  dstDoc::			Destination document
 *  ppDictIDList::		Pointer to the new dictionay ID list (memory
 *				is attached to the image, and should not
 *				be freed by the client)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxCopyTokenDictionary(
	     TFX_IMG *srcImg,
	     UInt8 *imgData,
	     UInt32 dataBytes,
	     TFX_DOC *dstDoc,
	     UInt32 **ppDictIDList
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *srcFile, *dstFile;
    TFX_DOC *srcDoc;
    TFX_IMGINFO *inImgInfo;
    UInt8 *dictData = NULL;
    Int32 dictBytes;
    TFX_COPIEDDICTS *pd;		/* Previously copied dictionaries */
    UInt32 srcDictOffset, newOffset;
    Int32 srcOffset, dstOffset;	/* File offsets */
    UInt16 dictNum, dictCount;
    UInt32 tempLong;
    UInt16 tempShort;
    UInt32 *newIDs = NULL;
    UInt16 resourceType;
    UInt16 lastPage;
    UInt32 decMem;
    UInt8 *posBlock;
    UInt32 posBytes;
    PROLOGUE;

    /* Reference the image information */
    instData = TFX_IMGINST(srcImg);
    inImgInfo = &srcImg->imgInfo;
    XERR_GEN_CLEANUP(inImgInfo->xCompression != IO_COMP_TOKENIZATION
	&& inImgInfo->xCompression != IO_COMP_LLTOKENIZATION
	&& inImgInfo->xCompression != IO_COMP_JBIG2
	&& inImgInfo->xCompression != IO_COMP_LLJBIG2,
	IO_INTERNAL_ERROR);

    /* The header should exist and be at least 6 bytes */
    XERR_GEN_CLEANUP(dataBytes < 6, IO_BADPARAMETER);

    /* Reference the source document, and document files */
    srcDoc = srcImg->tfxPage->tfxDoc;
    srcFile = TFX_IMGFILE(srcImg);
    dstFile = TFX_DOCFILE(dstDoc);

    /* Get the token dictionary ID from the TIFF stream (header to JBIG2 position block)
     */
    XERR_PASS_CLEANUP(TFX_GetDictListFromPosBlock((IO_IMGHANDLE)srcImg, imgData,
	dataBytes, &dictCount, &newIDs, &posBlock, &posBytes));
    /* Point now to the array of dictionary IDs */
    for (dictNum = 0; dictNum < dictCount; dictNum++) {
	srcDictOffset = newIDs[dictNum];

	/* Check the list of previously copied token dictionary IDs */
	pd = srcDoc->prevTDCopied;
	while (pd && (pd->dstDocHandle != (IO_DOCHANDLE)dstDoc ||
		      pd->dstDocID != dstDoc->docID ||
		      pd->dictOffset != srcDictOffset)) {
	    pd = pd->next;
	}

	/* If we found it, use it */
	if (pd) {
	    newOffset = pd->newOffset;
	}

	/*
	 * Copy the token dictionary if needed
	 */
	if (! pd) {		/* Not found in list of previous copies */
	    
	    /* Get the size and offset of the token dictionary */
	    XERR_PASS_CLEANUP(wp_TfxGetResourceOffset(srcDoc, srcDictOffset,
		&srcOffset, &dictBytes, &resourceType, &lastPage, &decMem));

	    /* Check for the right resource type */
	    switch (resourceType) {
	    case TFX_RESOURCE_JBIG2:
		break;
	    default:
		XERR_PASS_CLEANUP(IO_BADFILEFORMAT);
	    }

	    /* Check for a valid dictionary file position */
	    XERR_GEN_CLEANUP(srcOffset == 0, IO_BADFILEFORMAT);

	    /* Allocate a dictionary buffer */
	    IO_CHECKMALLOC(dictData = (UInt8 *)TFX_MALLOC(dictBytes));

	    /* Read the dictionary */
	    TFX_SEEK(srcFile, srcOffset);
	    TFX_READ(srcFile, dictData, dictBytes);

	    /* At this point we need to reswap the endian dependent
	     * items */
	    if (TFX_FILETOKEN(srcFile)->bSwapBytes !=
		TFX_FILETOKEN(dstFile)->bSwapBytes) {
		UInt32 dictSize;

		memcpy(&dictSize, dictData + cTfxTCPosDictSize,
		    sizeof(dictSize));
		TFX_SWAPL(&dictSize);
		memcpy(dictData + cTfxTCPosDictSize, &dictSize,
		    sizeof(dictSize));
	    }

	    /* Store the dictionary */
	    XERR_PASS_CLEANUP(wp_TfxFileLongAlign(dstFile, &dstOffset));
	    TFX_WRITE(dstFile, dictData, dictBytes);

	    /* Deallocate compressed data buffer */
	    TFX_FREE(dictData);
	    dictData = NULL;

	    /* Update the resource table */
	    XERR_PASS_CLEANUP(wp_TfxSetResourceOffset(dstDoc,
		dictBytes, dstOffset, TFX_TIFFCOMP_JBIG2, 0, 0, &newOffset));

	    /* Remember that this dictionary was transferred */
	    IO_CHECKMALLOC(pd = (TFX_COPIEDDICTS *)TFX_MALLOC(sizeof(TFX_COPIEDDICTS)));
	    memset(pd, 0, sizeof(*pd));
	    pd->dstDocHandle = (IO_DOCHANDLE)dstDoc;
	    pd->dstDocID = dstDoc->docID;
	    pd->dictOffset = srcDictOffset;
	    pd->newOffset = newOffset;

	    if (srcDoc->prevTDCopied) {
		pd->next = srcDoc->prevTDCopied;
	    }
	    srcDoc->prevTDCopied = pd;
	} /* Not previously copied */

	/* Replace the ID in the input position block Tundra stream */
	/* Note, because of a change in the order of the header, we're
	 * gunna force reinsertion of the ID, whether it changed or not,
	 * so that we create a new header regardless. */
	tempLong = newOffset;
	if (srcDictOffset != newOffset ||
	    TFX_FILETOKEN(srcFile)->bSwapBytes !=
	    TFX_FILETOKEN(dstFile)->bSwapBytes) {
	    TFX_SWAPLONG(dstFile, &tempLong);
	}
	memcpy(imgData + cTfxTCPosPosBlockDictIDs + dictNum *
	    sizeof(tempLong), &tempLong, sizeof(tempLong));

	/* Add it to the list of new IDs */
	newIDs[dictNum] = newOffset;
    } /* For each dictionary referenced */

    /* Note, because of a change in the order of the header, we're
     * gunna force reinsertion of the count and size, whether it changed or not,
     * so that we create a new header regardless. If we're converting the
     * endian-ness, we need to do this anyways. */
    tempShort = dictCount;
    TFX_SWAPSHORT(dstFile, &tempShort);
    memcpy(imgData + cTfxTCPosPosBlockDictCount, &tempShort, sizeof(tempShort));
    tempLong = posBytes;
    TFX_SWAPLONG(dstFile, &tempLong);
    memcpy(imgData + cTfxTCPosPosBlockSize(dictCount), &tempLong, sizeof(tempLong));

    /* If we're copying over extensions to JBIG2, and the endian-ness is different,
     * we need to go through and convert the extension tag, and byte count of each
     * extension encountered. */
    if (TFX_FILETOKEN(srcFile)->bSwapBytes != TFX_FILETOKEN(dstFile)->bSwapBytes) {
	UInt32 pos, extSize;

        pos = cTfxTCPosBlockHeaderSize(dictCount) + posBytes;
	while (pos + 6 < dataBytes) {

	    /* Grab the extension type */
	    memcpy(&tempShort, imgData + pos, sizeof(tempShort));
	    TFX_SWAPSHORT(srcFile, &tempShort);

	    /* Grab the extension byte count */
	    memcpy(&tempLong, imgData + pos + sizeof(tempShort), sizeof(tempLong));
	    TFX_SWAPLONG(srcFile, &tempLong);
	    extSize = tempLong;

	    /* Now put them back with the new byte ordering */
	    TFX_SWAPSHORT(srcFile, &tempShort);
	    memcpy(imgData + pos, &tempShort, sizeof(tempShort));
	    TFX_SWAPLONG(srcFile, &tempLong);
	    memcpy(imgData + pos + sizeof(tempShort), &tempLong, sizeof(tempLong));

	    /* Go to the next extension */
	    pos += 6 + extSize;
	}
    }

    /* Attach the list of dictionary IDs to the document, for
     * memory management (freed when document is closed). */
    if (srcDoc->curDictIDs) {
	TFX_FREE(srcDoc->curDictIDs);
    }
    srcDoc->curDictIDs = newIDs;
	
    /* Return the new ID */
    *ppDictIDList = newIDs;
    newIDs = NULL;
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (dictData) {
	TFX_FREE(dictData);
	dictData = NULL;
    }
    if (newIDs) {
	TFX_FREE(newIDs);
	newIDs = NULL;
    }
    ERR_RETURN;
}

/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy an individual image from one TIFF-FX to another
 *
 * LongDescription:::
 *  wp_TfxCopyImage() copies an individual image from one TIFF-FX to another,
 *  leaving the image data compressed during the copy.  This allows a much
 *  faster copy, since the data is not decompressed and recompressed.  It also
 *  means that the use of a lossy algorithm (like JPEG) will not produce added
 *  lossy affects by recompressing with additional or different loss.
 *  <BR>
 *  All fields from the source image are copied, including optional fields.
 *  But the annotations are NOT copied.
 *
 * Input:::
 *  srcImg::			Source image handle
 *  dstParentImg::		Destination image handle for parent image (if
 *				source image is a rendering hints image)
 *  firstStripOnly::		Boolean to copy only the image data from the
 *				first strip, and not finish the image write.
 *  useFGstripSize::		Boolean to use strip size from foreground image
 *				(used when color JBIG2 extension coupled to
 *				 the binary image)
 *
 * Output:::
 *  dstPage::			Page to attach copy of image to
 *  pOutImgInfo::		Resulting output image information. Image
 *				characteristics will be adopted from the srcImg
 *				except for the following structure elements of
 *				pOutImgInfo: layer, resolution, position,
 *				palette.
 *				Resulting tfxImgHandle will be returned via
 *				this structure also.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxCopyImage(
	     TFX_IMG *srcImg,
	     TFX_IMG *dstParentImg,
	     TFX_PAGE *dstPage,
	     TFX_IMGINFO *pOutImgInfo,
	     Bool firstStripOnly,
	     Bool useFGstripSize
)
{
    TFX_INSTANCE *instData;
    UInt32 row, rowsPerStrip;
    TFX_IMG *outImg = NULL;
    TFX_IMGINFO *inImgInfo;
    IO_COMPRESSION compressionType;
    UInt32 quality = 0UL;
    UInt8 *cmpBuffer = NULL;
    UInt32 cmpBufsize = 0UL;
    UInt32 cmpBytes = 0UL;
    UInt32 bytesFilled;
    UInt32 *dictIDs;
    UInt8 cval;
    UInt16 sval, sval1, sval2, *svals;
    Int32 *slvals;
    UInt32 lval, *lvals;
    Int8 *string;
    UInt32 stringlen;
    Bool exists;
    IO_DOCHANDLE srcDocHandle = (IO_DOCHANDLE)srcImg->tfxPage->tfxDoc;
    IO_PAGEHANDLE srcPageHandle = (IO_PAGEHANDLE)srcImg->tfxPage;
    IO_IMGHANDLE srcImgHandle = (IO_IMGHANDLE)srcImg;
    IO_DOCHANDLE dstDocHandle = (IO_DOCHANDLE)dstPage->tfxDoc;
    IO_PAGEHANDLE dstPageHandle = (IO_PAGEHANDLE)dstPage;
    IO_IMGHANDLE dstImgHandle, hintsHandle;
    IO_COLORSPACE colorSpace = IO_COLORSPACE_DEFAULT;
    UInt32 stripOffset, stripRow;
    UInt32 hintImages = 0;
    PROLOGUE;

    /* Reference the image information */
    inImgInfo = &srcImg->imgInfo;
    instData = TFX_IMGINST(srcImg);


    /* Begin the image read process */
    XERR_PASS_CLEANUP(TFX_ImageReadStart(inImgInfo->tfxImgHandle, 0UL));

    /* Copy the input parameters for the output image */
    memcpy(pOutImgInfo, inImgInfo, sizeof(*inImgInfo));

    /* Check the parameters */
#if 0
    XERR_PASS_CLEANUP(TFX_CheckImageInfo(dstPage->handleID, &pOutImgInfo->dwImgLayer,
	&pOutImgInfo->dwImgOffset, &pOutImgInfo->xColorType, &pOutImgInfo->dwXOffset,
	&pOutImgInfo->dwYOffset, &pOutImgInfo->dwXResolution,
	&pOutImgInfo->dwYResolution, &pOutImgInfo->dwWidth, &pOutImgInfo->dwHeight,
	&pOutImgInfo->dwBestStripHeight, pOutImgInfo->pPalette,
	&pOutImgInfo->xCompression, &pOutImgInfo->xColorSpace));
#endif

    /* Begin the image write process */
#if 0
    if (inImgInfo->tfxImgType == IO_IMGTYPE_MASK) {
	XERR_GEN_CLEANUP(dstParentImg == NULL, IO_BADPARAMETER);
	XERR_PASS_CLEANUP(TFX_AddMaskImage((IO_IMGHANDLE)dstParentImg,
	   inImgInfo->dwBestStripHeight, (IO_IMGHANDLE *)&outImg));
    } else {
#endif
    if (inImgInfo->tfxImgType == IO_IMGTYPE_HINTS) {
	XERR_GEN_CLEANUP(dstParentImg == NULL, IO_BADPARAMETER);
	XERR_PASS_CLEANUP(TFX_AddHintsImage((IO_IMGHANDLE)dstParentImg,
	   IO_BPP(inImgInfo->xColorType), inImgInfo->dwXOffset,
	   inImgInfo->dwYOffset, inImgInfo->dwXResolution,
	   inImgInfo->dwYResolution, inImgInfo->dwWidth, inImgInfo->dwHeight,
	   inImgInfo->dwBestStripHeight, (IO_IMGHANDLE *)&outImg));
    } else {
	XERR_PASS_CLEANUP(TFX_AddImage((IO_PAGEHANDLE)dstPage,
	    pOutImgInfo->dwImgLayer, inImgInfo->xColorType,
	    pOutImgInfo->dwXOffset, pOutImgInfo->dwYOffset,
	    pOutImgInfo->dwXResolution, pOutImgInfo->dwYResolution,
	    pOutImgInfo->dwWidth, pOutImgInfo->dwHeight,
	    pOutImgInfo->dwBestStripHeight, pOutImgInfo->pPalette,
	    (IO_IMGHANDLE *)&outImg));
    }

    dstImgHandle = (IO_IMGHANDLE)outImg;

    /* If it's the primary image, preset the layer count (enables one
     * SubIFD chain per layer).
     * Also preset the page count. */
    if (pOutImgInfo->tfxImgType == IO_IMGTYPE_PRIMARY) {

	switch (outImg->tfxPage->tfxDoc->docInfo.profile) {
	case TFX_PROFILE_M:		/* Profile M: MRC */
	case TFX_PROFILE_M1:		/* Profile M1: MRC + JBIG2 */
	case TFX_PROFILE_COH_M:		/* Coherent TIFF Profile M: MRC */
	case TFX_PROFILE_COH_M1:	/* Coherent TIFF Profile M1: MRC + JBIG2 */

	    /* Reserve a place for a SubIFD chain for each layer, except layer 2 */
	    outImg->fields.SubIFDCount = srcImg->tfxPage->pageInfo.dwLayersPresent;
	    if (outImg->fields.SubIFDCount > 1) {
		outImg->fields.SubIFDCount--;	/* Discount layer 2 */
	    }
	    break;
	default:
	    break;
	}

	/* Preset the page number (page count) */
	XERR_PASS_CLEANUP(TFX_SetPageNumber((IO_PAGEHANDLE)dstPage,
	    (UInt16)(srcImg->tfxPage->pageInfo.dwPageOffset),
	    srcImg->tfxPage->tfxDoc->docInfo.wNPages));

	/* If there's annotations already (or at least the field) allow
	 * annotations to be written */
/*	if (srcImg->fields.found.annotationOffsets) { */
/*	    XERR_PASS_CLEANUP(TFX_AllowAnnotations(outImg->tfxPage)); */
/*	} */
    }

    /*
     * Transfer optional fields over
     */

    /* The document level stuff */
    if ( srcImg->imgInfo.tfxImgType == IO_IMGTYPE_PRIMARY
	&& srcImg->tfxPage->pageInfo.dwPageOffset == 0) {
/* Why is this tied to the page, and not the document? */
/*	if (TFX_GetCodingMethods(srcDocHandle, &lval, &exists) == IO_NOERROR */
/*	    && exists) { */
/*	    XERR_PASS_CLEANUP(TFX_SetCodingMethods(dstDocHandle, lval)); */
/*	} */
	if (TFX_GetDocumentName(srcDocHandle, &string, &stringlen, &exists) ==
	    IO_NOERROR && exists) {
	    XERR_PASS_CLEANUP(TFX_SetDocumentName(dstDocHandle, string,
		stringlen));
	}
/* Note: this will have to move when multi-profiles is implemented??? */
	if (TFX_GetFaxProfile(srcPageHandle, &cval, &exists) == IO_NOERROR
	    && exists) {
	    XERR_PASS_CLEANUP(TFX_SetFaxProfile(dstPageHandle, cval));
	}
	if (TFX_GetModeNumber(srcPageHandle, &cval, &exists) == IO_NOERROR
	    && exists) {
	    XERR_PASS_CLEANUP(TFX_SetModeNumber(dstPageHandle, cval));
	}
/* Note: this will have to move when multi-profiles is implemented??? */
	if (TFX_GetProfileType(srcPageHandle, &lval, &exists) == IO_NOERROR
	    && exists) {
	    XERR_PASS_CLEANUP(TFX_SetProfileType(dstPageHandle, lval));
	}
	if (TFX_GetReferencePage(srcDocHandle, &lval, &exists) == IO_NOERROR
	    && exists) {
	    XERR_PASS_CLEANUP(TFX_SetReferencePage(dstDocHandle, lval));
	}
	if (TFX_GetVersionYear(srcImgHandle, (UInt8 **)&string, &stringlen,
	    &exists) == IO_NOERROR && exists) {
	    XERR_PASS_CLEANUP(TFX_SetVersionYear(dstImgHandle,
		(UInt8 *)&string[0], stringlen));
	}
    }

    /* The page level stuff */
    if (srcImg->imgInfo.tfxImgType == IO_IMGTYPE_PRIMARY) {
	if (TFX_GetAnnotationOffsets(srcPageHandle, &lval, &exists) == IO_NOERROR
	    && exists) {
	    XERR_PASS_CLEANUP(TFX_SetAnnotationOffsets(dstPageHandle, 0));
	}
	if (TFX_GetDateTime(srcPageHandle, &string, &exists) == IO_NOERROR
	    && exists) {
	    XERR_PASS_CLEANUP(TFX_SetDateTime(dstPageHandle, string));
	}
	if (TFX_GetPageNumber(dstPageHandle, &sval1, &sval2, &exists) == IO_NOERROR
	    && exists) {
	    XERR_PASS_CLEANUP(TFX_SetPageNumber(dstPageHandle, sval1, sval2));
	}
	if (TFX_GetTargetPrinter(srcPageHandle, &string, &stringlen, &exists) == IO_NOERROR
	    && exists) {
	    XERR_PASS_CLEANUP(TFX_SetTargetPrinter(dstPageHandle, string, stringlen));
	}
    }


    /* The image level stuff */
#if 0
    /* This should be automatic, based on the color type */
    if (TFX_GetBitsPerSample(srcImgHandle, &lval1, &lval2, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetBitsPerSample(dstImgHandle, lval1, lval2));
    }
#endif
    if (TFX_GetChromaPositioning(srcImgHandle, &sval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetChromaPositioning(dstImgHandle, sval));
    }
    if (TFX_GetChromaSubSampling(srcImgHandle, &sval1, &sval2, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetChromaSubSampling(dstImgHandle, sval1, sval2));
    }
    if (TFX_GetCompression(srcImgHandle, &sval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetCompression(dstImgHandle, sval));
    }
    if (TFX_GetDecode(srcImgHandle, (Int32 **)&lvals, &lval, &exists) ==
	IO_NOERROR && exists) {
	XERR_PASS_CLEANUP(TFX_SetDecode(dstImgHandle, (Int32 *)lvals, lval));
    }
    if (TFX_GetFillOrder(srcImgHandle, &sval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetFillOrder(dstImgHandle, sval));
    }
    if (TFX_GetHostComputer(srcImgHandle, &string, &stringlen, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetHostComputer(dstImgHandle, string, stringlen));
    }
    if (TFX_GetIFDBaseColor(srcImgHandle, &slvals, &lval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetIFDBaseColor(dstImgHandle, slvals, lval));
    }
    if (TFX_GetImageBaseColor(srcImgHandle, &svals, &lval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetImageBaseColor(dstImgHandle, svals, lval));
    }
    if (TFX_GetImageDescription(srcImgHandle, &string, &stringlen, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetImageDescription(dstImgHandle, string, stringlen));
    }
    if (TFX_GetOrientation(srcImgHandle, &sval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetOrientation(dstImgHandle, sval));
    }
    if (TFX_GetPhotometricInterpretation(srcImgHandle, &sval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetPhotometricInterpretation(dstImgHandle, sval));
    }
    if (TFX_GetPlanarConfiguration(srcImgHandle, &sval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetPlanarConfiguration(dstImgHandle, sval));
    }
    if (TFX_GetPredictor(srcImgHandle, &sval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetPredictor(dstImgHandle, sval));
    }
    if (TFX_GetRenderingHints(srcImgHandle, &string, &stringlen, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_AddRenderingHints(dstImgHandle, (char *)string, stringlen));
    }
    if (TFX_GetRenderingHintsImageCount(srcImgHandle, &hintImages) == IO_NOERROR
	&& hintImages) {
	UInt32 hintOffset = 0;
	IO_IMGHANDLE outHintsImg = 0;
	TFX_IMGINFO hintsInfo;

	while (hintOffset < hintImages) {
	    XERR_PASS_CLEANUP(TFX_GetHintsImageInfo(srcImgHandle, hintOffset,
		&hintsInfo.dwXOffset, &hintsInfo.dwYOffset,
		&hintsInfo.dwXResolution,
		&hintsInfo.dwYResolution,
		&hintsInfo.dwWidth, &hintsInfo.dwHeight,
		&hintsInfo.dwBestStripHeight,
		&hintsInfo.xCompression,
		&hintsHandle));
	    XERR_IGNORE(TFX_CopyImage(hintsHandle, dstImgHandle, dstPageHandle,
		hintOffset, hintsInfo.dwXOffset, hintsInfo.dwYOffset,
		hintsInfo.dwHeight, &outHintsImg));
	    hintOffset++;
	}
    }
    if (TFX_GetResolutionUnit(srcImgHandle, &sval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetResolutionUnit(dstImgHandle, sval));
    }
    if (TFX_GetRowsPerStrip(srcImgHandle, &lval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetRowsPerStrip(dstImgHandle, lval));
    }
#if 0
    /* This should be automatic, based on the color type */
    if (TFX_GetSamplesPerPixel(srcImgHandle, lvals, &lval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetSamplesPerPixel(dstImgHandle, lvals, lval));
    }
#endif
    if (TFX_GetSoftware(srcImgHandle, &string, &stringlen, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetSoftware(dstImgHandle, string, stringlen));
    }
    if (TFX_GetThresholding(srcImgHandle, &sval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetThresholding(dstImgHandle, sval));
    }
    if (TFX_GetT4Options(srcImgHandle, &lval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetT4Options(dstImgHandle, lval));
    }
    if (TFX_GetT6Options(srcImgHandle, &lval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetT6Options(dstImgHandle, lval));
    }
    if (TFX_GetT88Options(srcImgHandle, &lval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetT88Options(dstImgHandle, lval));
    }
    if (TFX_GetYCbCrCoefficients(srcImgHandle, &lvals, &lval, &exists) == IO_NOERROR
	&& exists) {
	XERR_PASS_CLEANUP(TFX_SetYCbCrCoefficients(dstImgHandle, lvals, lval));
    }

    /* If there are variable-sized strips, we must use their sizes */
    if (inImgInfo->dwBestStripHeight == 0) {
	XERR_PASS_CLEANUP(TFX_UseVariableStripHeights((IO_IMGHANDLE)outImg));
    } else {

        /* If there is a recommendation for a new strip size, ignore the
	 * recommendation (we must use the same strip sizes), especially
	 * for variable-sized strips (StripRowCounts). */
	if (outImg->imgInfo.dwBestStripHeight != inImgInfo->dwBestStripHeight) {
	    pOutImgInfo->dwBestStripHeight = outImg->imgInfo.dwBestStripHeight =
		inImgInfo->dwBestStripHeight;
	}
    }

    /* Write the IFD structure */    
    compressionType = inImgInfo->xCompression;
    XERR_PASS_CLEANUP(TFX_ImageWriteStart((IO_IMGHANDLE)outImg,
	IO_IMGFLAGS_Compressed, &compressionType, &quality, &colorSpace));

    /* Again, if there is a recommendation for a new strip size, ignore the
     * recommendation (we must use the same strip sizes), especially
     * for variable-sized strips (StripRowCounts). */
    if (outImg->imgInfo.dwBestStripHeight != inImgInfo->dwBestStripHeight) {
	pOutImgInfo->dwBestStripHeight = outImg->imgInfo.dwBestStripHeight =
	    inImgInfo->dwBestStripHeight;
    }

    /* Loop through the image strips, copying them compressed one by one */
    rowsPerStrip = inImgInfo->dwBestStripHeight;
    for (row = 0; row < inImgInfo->dwHeight; row += rowsPerStrip) {

	/* Get the number of rows in this strip */
	if (inImgInfo->dwBestStripHeight == 0) {
	    XERR_PASS_CLEANUP(wp_TfxGetStripInfo(srcImg, row,
		&stripOffset, &stripRow, &rowsPerStrip));
	    XERR_GEN_CLEANUP(stripRow != row, IO_INTERNAL_ERROR);
	}
	if (row + rowsPerStrip > inImgInfo->dwHeight) {
	    rowsPerStrip = inImgInfo->dwHeight - row;
	}
	if (useFGstripSize) {
	    TFX_IMG *fgImg;
	    UInt32 imgOffset = 0, x, y, xres, yres, w, h, stripHeight;
	    IO_COMPRESSION compression;
	    IO_COLORTYPE colorType;
	    UInt8 palette[768];

	    while (1) {
		XERR_IGNORE(__error__ = TFX_GetImageInfo(srcImg->tfxPage->handleID,
		    inImgInfo->dwImgLayer + 1, imgOffset, &x, &y, &xres, &yres,
		    &w, &h, &stripHeight, &compression, &colorType, &palette[0],
		    (IO_IMGHANDLE *)&fgImg));
		if (__error__ < IO_WARNINGS && __error__ != IO_NOERROR) {
		    
		    /* Foreget the foreground image */
		    XERR_PASS_CLEANUP(TFX_GetCompressedStripSize(inImgInfo->tfxImgHandle,
			row, &cmpBytes));
		    break;
		}
		if (x == inImgInfo->dwXOffset && y == inImgInfo->dwYOffset &&
		    xres == inImgInfo->dwXResolution && yres == inImgInfo->dwYResolution &&
		    w == inImgInfo->dwWidth && h == inImgInfo->dwHeight) {
		    break;  /* Okay, we found the matching foreground image */
		}

		/* Haven't found the matching foreground, try the next */
		imgOffset++;
	    }

	    XERR_PASS_CLEANUP(TFX_GetCompressedStripSize(fgImg->handleID,
		row, &cmpBytes));
	} else {
	    XERR_PASS_CLEANUP(TFX_GetCompressedStripSize(inImgInfo->tfxImgHandle,
		row, &cmpBytes));
	}

	if (cmpBytes == 0) {	/* Don't have to do much if it's 0 bytes */
	    /* Add the strip to the strip list */
	    XERR_PASS_CLEANUP(wp_TifFXAddStripInfo(outImg, row, rowsPerStrip, 0));

	    /* Update the internal row offset */
/*	    XERR_PASS_CLEANUP(TFX_CompressedWriteStrip((IO_IMGHANDLE)outImg, */
/*		row, rowsPerStrip, cmpBuffer, cmpBytes, &bytesFilled)); */
	} else {	/* Has compressed data */
	    if (cmpBufsize < cmpBytes) {
		if (cmpBufsize && cmpBuffer) {
		    TFX_FREE(cmpBuffer);
		}
		IO_CHECKMALLOC(cmpBuffer = (UInt8 *)TFX_MALLOC(cmpBytes));
		cmpBufsize = cmpBytes;
	    }

	    /* Read the compressed strip */
	    XERR_PASS_CLEANUP(TFX_CompressedReadStrip(inImgInfo->tfxImgHandle,
		row, rowsPerStrip, &cmpBuffer, &cmpBufsize, &bytesFilled));
	    cmpBytes = bytesFilled;

	    /* If it's a tokenized image, copy the token dictionary first */
	    if (row == 0) {
		switch ((int)inImgInfo->xCompression) {
		case IO_COMP_TOKENIZATION:
		case IO_COMP_LLTOKENIZATION:
		case IO_COMP_JBIG2:
		case IO_COMP_LLJBIG2:
		    XERR_PASS_CLEANUP(wp_TfxCopyTokenDictionary(srcImg,
			cmpBuffer, cmpBytes, dstPage->tfxDoc, &dictIDs));
		    break;
		default:
		    break;
		}
	    }

	    /* Write the compressed strip */
	    XERR_PASS_CLEANUP(TFX_CompressedWriteStrip((IO_IMGHANDLE)outImg,
		row, rowsPerStrip, cmpBuffer, cmpBytes, &bytesFilled));
	    XERR_GEN_CLEANUP(bytesFilled < cmpBytes, IO_INTERNAL_ERROR);
	}

	if (firstStripOnly) {
	    rowsPerStrip = inImgInfo->dwHeight;
	}
    }

    /* Deallocate compressed data buffer */
    if (cmpBuffer) {
	TFX_FREE(cmpBuffer);
    }
    cmpBuffer = NULL, cmpBufsize = 0UL;

    /* Finish up the read and write processes */
    if (! firstStripOnly) {
	XERR_PASS_CLEANUP(TFX_ImageReadFinish(inImgInfo->tfxImgHandle));
	XERR_PASS_CLEANUP(TFX_ImageWriteFinish((IO_IMGHANDLE)outImg));
    }

    /* Return the resulting image info */
    memcpy(pOutImgInfo, &outImg->imgInfo, sizeof(*pOutImgInfo));
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if (cmpBuffer) {
	TFX_FREE(cmpBuffer);
	cmpBuffer = NULL, cmpBufsize = 0UL;
    }
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copies the entire contents of a page from one TIFF-FX to another
 *
 * LongDescription:::
 *  wp_TfxCopyPage() copies a specified page from one TIFF-FX document to
 *  another. The page can be an arbitrary page in the source document,
 *  but must be the next available page in the destination document
 *  (no page insertions).  The image data is left compressed during the
 *  copy, enabling fast copies.  Version differences are accounted for
 *  by decompressing and recompressing (using the default compression
 *  method).
 *  <BR>
 *  The order of the data in the destination will be according to the
 *  internet fax recommendation: stripe ordering, top-to-bottom, layers in
 *  the order 2, 1, 3, 4, 5, ...  Note, no care is taken to assure that
 *  stripe boundaries are not violated; it's assumed that the page has
 *  already been broken into stripes (if any).  Keep in mind that only the
 *  primary IFD dictates stripe boundaries (any strips in the primary IFD
 *  become the stripes for all layers).
 *  <BR>
 *  This function can be used to facilitate page insertion, deletion and
 *  reordering; but you must place it in a loop with the appropriate page
 *  offsets to sequentially specify the new destination page order.
 *
 * Input:::
 *  xSrcDocHandle::		Source document handle, returned from
 *				TFX_OpenDocumentRead() or
 *				TFX_OpenDocumentReadWrite()
 *  xDstDocHandle::		Destination document handle, returned from
 *				TFX_CreateDocument() or
 *				TFX_OpenDocumentReadWrite()
 *  dwSrcPageOffset::		Source document offset of page to copy
 *  dwDstPageOffset::		Destination document offset of page to add
 *				(this should be the end of the document)
 *  copyAnnotations::		Boolean for copy the page annotations
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxCopyPage(
	TFX_DOC *xSrcDoc,
	Int32 dwSrcPageOffset,
	TFX_DOC *xDstDoc,
	Int32 dwDstPageOffset,
	Bool copyAnnotations
)
{
    TFX_INSTANCE *instData;
    UInt32 imgNum;
    UInt16 annotNum;
    IO_DOCHANDLE inDoc = 0, outDoc = 0;
    IO_PAGEHANDLE inPage, outPage;
    TFX_IMG *inPrimImg, *outPrimImg = NULL, *inImg;
    TFX_PAGEINFO inPageInfo;
    TFX_IMGINFO inPrimImgInfo, outPrimImgInfo, inImgInfo, outImgInfo;
    TFX_ANNOTINFO annot;
    UInt32 layer, primaryLayer = 2;
    UInt8 *annotData = NULL;
    UInt16 annotID;
#ifndef NOIPCORE
    IPWState *pIPWState = NULL;
    PIXR *pPixr = NULL;
    UInt8 *imageChannels[3];
    Int32 bpl;
    ScaledInt1616 xres, yres;
    ColorType ctype = cColorNone;
    Int32 channel, channels;
    PIXR *pSubPixr;
    UInt32 quality;
    IO_COLORSPACE colorSpace = IO_COLORSPACE_DEFAULT;
    IO_COMPRESSION compression;
    UInt32 rowsFilled, rowsWritten;
    IO_IMGHANDLE outImgHandle = 0;
#if 0
    TFX_IMGINFO inLayerInfo;
#endif
#endif
    Bool recompress;
    Int32 imageHeight, scale;
    UInt32 primaryRow, stripeOffset, stripeRowOffset, stripeRows;
    UInt32 cmpBytes, cmpBufsize = 0, bytesFilled;
    UInt8 *cmpBuffer = NULL;
    UInt32 *dictIDs;
    Bool useColorJBIG2;
    PROLOGUE;

    TFX_CheckHandle(xSrcDoc);
    TFX_CheckHandle(xDstDoc);

    /* Check for proper calling sequence */
    XERR_GEN((xSrcDoc->seqStatus & TFX_DOC_SEQ_GETINFO) == 0,
	IO_SEQUENCE);
    
    /* Reference the handles/data */
    inDoc = (IO_DOCHANDLE)xSrcDoc;
    outDoc = (IO_DOCHANDLE)xDstDoc;
    instData = TFX_DOCINST(xSrcDoc);

    /* Make sure that any unwritten pages for the input document are flushed,
     * before we get the image counts for the layers */
    if ((xSrcDoc->readWrite & WRITE_ALLOWED) &&
	xSrcDoc->seqStatus & (TFX_DOC_SEQ_TOKENSTART)) {
	XERR_PASS_CLEANUP(TFX_WriteTokenImages((IO_DOCHANDLE)xSrcDoc));
    }

    /* Make sure that any unwritten pages for the output document are flushed,
     * before we potentially start a new token dictionary */
    if ((xDstDoc->readWrite & WRITE_ALLOWED) &&
	xDstDoc->seqStatus & (TFX_DOC_SEQ_TOKENSTART)) {
	XERR_PASS_CLEANUP(TFX_WriteTokenImages((IO_DOCHANDLE)xDstDoc));
    }

    /* Get the page info */
    inPageInfo.dwPageOffset = dwSrcPageOffset;
    XERR_WARNOK_CLEANUP(TFX_GetPageInfo(inDoc, dwSrcPageOffset,
	&inPageInfo.pdwImageCounts, &inPageInfo.dwLayersPresent,
	&inPageInfo.wAnnotationCount, &inPage));

    /* Begin addition of a new page */
    XERR_PASS_CLEANUP(TFX_AddPageStart(outDoc,
	dwDstPageOffset, &outPage));
	    
    /* Start with some basic assumptions about all images */
    memset(&inPrimImgInfo, 0, sizeof(inImgInfo));

    /* Determine the primary image layer */
    inPrimImg = (TFX_IMG *)(((TFX_PAGE*)inPage)->pageInfo.tfxPrimImgHandle);
    primaryLayer = inImgInfo.dwImgLayer = inPrimImg->imgInfo.dwImgLayer;
    inPrimImgInfo.dwImgLayer = inPrimImg->imgInfo.dwImgLayer;
    inPrimImgInfo.dwImgOffset = inPrimImg->imgInfo.dwImgOffset;
	    
    /* Get the Primary image info, keep track of it for the page */
    inPrimImgInfo.tfxImgType = IO_IMGTYPE_PRIMARY;
    XERR_WARNOK_CLEANUP(TFX_GetImageInfo(inPage,
	inPrimImgInfo.dwImgLayer, inPrimImgInfo.dwImgOffset,
	&inPrimImgInfo.dwXOffset, &inPrimImgInfo.dwYOffset,
	&inPrimImgInfo.dwXResolution, &inPrimImgInfo.dwYResolution,
	&inPrimImgInfo.dwWidth, &inPrimImgInfo.dwHeight,
	&inPrimImgInfo.dwBestStripHeight, &inPrimImgInfo.xCompression,
	&inPrimImgInfo.xColorType, (UInt8 *)&inPrimImgInfo.pPalette[0],
	&inPrimImgInfo.tfxImgHandle));
    
    /* Borrow the image characteristics for the output image */
    memcpy(&outPrimImgInfo, &inPrimImgInfo, sizeof(outPrimImgInfo));
    inPrimImg = (TFX_IMG *)inPrimImgInfo.tfxImgHandle;

    /* Keep track of the stripe */
    imageHeight = inPrimImgInfo.dwHeight;
/*    scale = inPrimImg->page->pageYResolution / imgInfo->dwYResolution; */
    primaryRow = 0;

    /* The page resolution should be the highest resolution */
/*    XERR_GEN_CLEANUP(scale == 0, IO_BADFILEFORMAT); */

    /* Loop over the stripes, until we've written them all */
    while (imageHeight > 0) {

	/* Compute the end of this stripe */
	XERR_PASS_CLEANUP(TFX_GetStripeInfo((IO_PAGEHANDLE)inPage,
	    primaryRow, &stripeOffset, &stripeRowOffset, &stripeRows));

	/* Copy the primary image first (always) */
        layer = primaryLayer;

	/* Decide whether we need to recompress this strip */
	recompress = 0;

	/* Decide whether we need to grab the color-JBIG2 strip, instead of the
	 * binary strip */
	useColorJBIG2 = 0;
	switch ((int)inPrimImgInfo.xCompression) {
	case IO_COMP_TOKENIZATION:
	case IO_COMP_LLTOKENIZATION:
	case IO_COMP_JBIG2:
	case IO_COMP_LLJBIG2:
	    {
		UInt32 T88Options = 0;
		Bool exists;

		/* Check to see if the T88Options claims a color component is attached */
		XERR_PASS_CLEANUP(TFX_GetT88Options(inPrimImg->handleID, &T88Options, &exists));

		if (exists && (T88Options & T88OPTION_COLOURTAGSFOLLOW)) {
		    useColorJBIG2 = 1;
		}
	    }
	    break;
	}

#if 0
	/* Copy the image */
        if (recompress) {

#ifdef NOIPCORE
	    XERR_PASS_CLEANUP(IO_NOSUPPORT);
#else
	    /* This image is always the entire page, based on the primary IFD.
	     * Read the entire layer. */
	    XERR_PASS_CLEANUP(TFX_GetLayerInfo(inPage,
		inLayerInfo.dwImgLayer = layer,
		&inLayerInfo.dwXOffset, &inLayerInfo.dwYOffset,
		&inLayerInfo.dwXResolution, &inLayerInfo.dwYResolution,
		&inLayerInfo.dwWidth, &inLayerInfo.dwHeight,
		&inLayerInfo.xCompression, &inLayerInfo.xColorType,
		&inLayerInfo.pPalette[0],
		&inLayerInfo.tfxImgHandle));

	    inLayerInfo.dwImgOffset = 0;
	    /*	inLayerInfo.xColorType = inImgInfo.xColorType; */

	    /* Will be default, set elsewhere */
	    inLayerInfo.dwBestStripHeight = 0;

	    /* Borrow the image characteristics for the output image */
	    memcpy(&outPrimImgInfo, &inLayerInfo, sizeof(outPrimImgInfo));
	    outPrimImgInfo.tfxImgType = IO_IMGTYPE_PRIMARY;

	    /* Access the common IPWState */
	    XERR_PASS_CLEANUP(wp_TfxIPWState(((TFX_PAGE *)outPage)->tfxDoc,
		&pIPWState));

	    /* Create a PIXR image to hold the image */
	    xres = Int32_TO_ScaledInt1616(inLayerInfo.dwXResolution);
	    yres = Int32_TO_ScaledInt1616(inLayerInfo.dwYResolution);
	    switch (inLayerInfo.xColorType) {
	    case IO_COLORTYPE_COLOR24:
		ctype = cColorRGB;
		break;
	    case IO_COLORTYPE_BINARY:
	    case IO_COLORTYPE_GRAY4:
	    case IO_COLORTYPE_COLOR4:
	    case IO_COLORTYPE_GRAY8:
	    case IO_COLORTYPE_COLOR8:
		ctype = cColorGray;
		break;
	    default:
		break;
	    }
	    channels = IO_CHANNELS(inLayerInfo.xColorType);
	    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, outPrimImgInfo.dwWidth,
		outPrimImgInfo.dwHeight, IO_BPPE(outPrimImgInfo.xColorType),
		xres, yres, 32, channels, ctype, NULL, &pPixr));

	    /* Get some characteristics of the PIXR, retain the image data
	     * location for each channel(s) */
	    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &bpl));
	    for (channel = 0; channel < channels; channel++) {
		if (channels > 1) {
		    pSubPixr = NULL;
		    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
			&pSubPixr));
		    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
			&imageChannels[channel]));
		} else {
		    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr,
			&imageChannels[channel]));
		}
	    }
	
	    /* Read the entire layer */
	    rowsFilled = 0;
	    XERR_PASS_CLEANUP(w_TfxImageReadLayer(pIPWState,
		inLayerInfo.tfxImgHandle, pPixr, &rowsFilled));

	    XERR_PASS_CLEANUP(TFX_AddImage(outPage, outPrimImgInfo.dwImgLayer,
		outPrimImgInfo.xColorType, outPrimImgInfo.dwXOffset,
		outPrimImgInfo.dwYOffset, outPrimImgInfo.dwXResolution,
		outPrimImgInfo.dwYResolution, outPrimImgInfo.dwWidth,
		outPrimImgInfo.dwHeight, outPrimImgInfo.dwBestStripHeight,
		outPrimImgInfo.pPalette, &outPrimImgInfo.tfxImgHandle));
	    outPrimImg = (TFX_IMG *)outPrimImgInfo.tfxImgHandle;

	    compression = IO_COMP_UNKNOWN;	    /* Let it pick up the default */
	    quality = 0;			    /* Let it pick up the default */

	    XERR_PASS_CLEANUP(TFX_ImageWriteStart(outPrimImg->handleID, 0, &compression,
		&quality, &colorSpace));

	    /* Reserve a place for a SubIFD chain for each layer, except layer 2 */
	    outPrimImg->fields.SubIFDCount = inPageInfo.dwLayersPresent;
	    if (outPrimImg->fields.SubIFDCount > 1) {
		outPrimImg->fields.SubIFDCount--;	/* Discount layer 2 */
	    }

	    /* Write the entire image (as one strip). Let this function parcel it
	     * into strips. */
	    XERR_PASS_CLEANUP(w_TfxWriteStrip(pIPWState, outPrimImg->handleID, pPixr, 0,
		rowsFilled, 0, &rowsWritten));

	    /* Finish the write */
	    XERR_PASS_CLEANUP(TFX_ImageWriteFinish(outPrimImg->handleID));

	    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &pPixr));
#endif	/* NOIPCORE */
	} else {	/* Not recompressing */
#endif	/* #if 0 */

	    /* If this is the first stripe, start the primary IFD */
	    if (stripeOffset == 0) {

		/* Copy the image, but only copy the first strip */
		XERR_PASS_CLEANUP(wp_TfxCopyImage(inPrimImg,
		    NULL, (TFX_PAGE *)outPage, &outPrimImgInfo, 1, useColorJBIG2));
		outPrimImg = (TFX_IMG *)outPrimImgInfo.tfxImgHandle;
	    } else {

		XERR_GEN_CLEANUP(outPrimImg == NULL, IO_INTERNAL_ERROR);

		XERR_PASS_CLEANUP(TFX_GetCompressedStripSize(inPrimImgInfo.tfxImgHandle,
		    primaryRow, &cmpBytes));

#if 0
		compression = inPrimImgInfo.xCompression;
		quality = 0;
		colorSpace = inPrimImgInfo.xColorSpace;
		XERR_PASS_CLEANUP(TFX_ImageWriteStart(outPrimImg->handleID, 0, &compression,
		    &quality, &colorSpace));
#endif

		if (cmpBytes > 0) {
		    if (cmpBufsize < cmpBytes) {
			if (cmpBufsize && cmpBuffer) {
			    TFX_FREE(cmpBuffer);
			}
			IO_CHECKMALLOC(cmpBuffer = (UInt8 *)TFX_MALLOC(cmpBytes));
			cmpBufsize = cmpBytes;
		    }
		    
		    /* Read the compressed strip */
		    XERR_PASS_CLEANUP(TFX_CompressedReadStrip(inPrimImgInfo.tfxImgHandle,
			primaryRow, stripeRows, &cmpBuffer, &cmpBufsize, &bytesFilled));
		    cmpBytes = bytesFilled;
		    
		    /* If it's a tokenized image, copy the token dictionary first */
		    if (primaryRow == 0) {
			switch ((int)inPrimImgInfo.xCompression) {
			case IO_COMP_TOKENIZATION:
			case IO_COMP_LLTOKENIZATION:
			case IO_COMP_JBIG2:
			case IO_COMP_LLJBIG2:
			    XERR_PASS_CLEANUP(wp_TfxCopyTokenDictionary(inPrimImg,
				cmpBuffer, cmpBytes, outPrimImg->tfxPage->tfxDoc, &dictIDs));
			    break;
			default:
			    break;
			}
		    }
		    
		    /* Write the compressed strip */
		    XERR_PASS_CLEANUP(TFX_CompressedWriteStrip(outPrimImg->handleID,
			primaryRow - outPrimImg->imgInfo.dwYOffset, stripeRows,
			cmpBuffer, cmpBytes, &bytesFilled));
		    XERR_GEN_CLEANUP(bytesFilled < cmpBytes, IO_INTERNAL_ERROR);
		}

		/* If this is the last strip, finish the image */
		if (imageHeight == (Int32)stripeRows) {
		    XERR_PASS_CLEANUP(TFX_ImageReadFinish(inPrimImgInfo.tfxImgHandle));
		    XERR_PASS_CLEANUP(TFX_ImageWriteFinish(outPrimImg->handleID));
		}
	    }
#if 0
	} /* recompress */
#endif

	/* Done with the primary layer, now loop through each other layer present */
	for (layer = 1; layer <= inPageInfo.dwLayersPresent; layer++) {

	    /* Skip the primary image (already stored above) */
	    if (layer == primaryLayer) {
		imgNum = 1;	/* Can only be one primary layer */
	    } else {
		imgNum = 0;
	    }
	    for ( ; imgNum < inPageInfo.pdwImageCounts[layer - 1]; imgNum++) {
		
		inImgInfo.tfxImgType = IO_IMGTYPE_LAYER;
		XERR_WARNOK_CLEANUP(TFX_GetImageInfo(inPage,
		    inImgInfo.dwImgLayer = layer,
		    inImgInfo.dwImgOffset = imgNum,
		    &inImgInfo.dwXOffset, &inImgInfo.dwYOffset,
		    &inImgInfo.dwXResolution, &inImgInfo.dwYResolution,
		    &inImgInfo.dwWidth, &inImgInfo.dwHeight,
		    &inImgInfo.dwBestStripHeight, &inImgInfo.xCompression,
		    &inImgInfo.xColorType, &inImgInfo.pPalette[0],
		    &inImgInfo.tfxImgHandle));

#if 0
		printImgInfo(2, &inImgInfo);
#endif

		/* Make sure the resolution is an integer multiple of the page
		 * resolution */
		XERR_GEN_CLEANUP(inImgInfo.dwYResolution == 0, IO_BADFILEFORMAT);
		XERR_GEN_CLEANUP(inPrimImgInfo.dwYResolution % inImgInfo.dwYResolution,
		    IO_BADFILEFORMAT);

		/* Account for scale difference */
		scale = inPrimImgInfo.dwYResolution / inImgInfo.dwYResolution;

		/* Use this image, only if the image region lies completely inside
		 * the current stripe.
		 * Skip if there is a scanline completely outside the stripe,
		 * but error if a scanline is completely inside also.
		 * Check here for one completely outside. */
		if (inImgInfo.dwYOffset + scale - 1 < primaryRow ||
		    inImgInfo.dwYOffset + ((inImgInfo.dwHeight - 1) * scale) >=
		    primaryRow + stripeRows) {

		    /* It has at least one entire scanline that is outside of
		     * this stripe.  If the image has any scanline completely
		     * within this stripe also, then the file must support multi-strip
		     * SubIFDs (XCTIFF M profiles, or TIFF-FX Ext 2). */
		    if (inImgInfo.dwYOffset + scale - 1 <
			primaryRow + stripeRows &&
			inImgInfo.dwYOffset + ((inImgInfo.dwHeight - 1) * scale) >=
			primaryRow) {
			switch (xSrcDoc->docInfo.profile) {
			case TFX_PROFILE_M:
			    /* Note this profile may have been identified by mistake,
			     * when in fact is it COH_M */
			case TFX_PROFILE_M1:
			case TFX_PROFILE_COH_M:
			case TFX_PROFILE_COH_M1:
			    /* Can't do this yet */
			    XERR_GEN_CLEANUP(1, IO_NOSUPPORT);
			    /* Make sure the page is enabled to use multi-strip
			     * SubIFDs */
			    XERR_PASS_CLEANUP(TFX_UseMultiStripSubIFDs(outPage));
			    break;
			default:
			    /* It's wrong */
			    XERR_GEN_CLEANUP(1, IO_BADFILEFORMAT);
			}
		    } else {
			/* Go to the next image */
			continue;
		    }
		}

		recompress = 0;
		if (((TFX_PAGE *)outPage)->tfxDoc->docInfo.version < 3) {

		    /* For TIFF-FX 2 files, if the file is stored in an unsupported
		     * compression type, decompress and store off in default
		     * compression type. */
		    switch (inImgInfo.xCompression) {
		    case IO_COMP_NOC:
		    case IO_COMP_MMR:
		    case IO_COMP_JPEG:
			break;
		    default:
			recompress = 1;
		    }

		    /* If PhotometricInterpretation is not right, force fixing it
		     */
		    inImg = (TFX_IMG *)inImgInfo.tfxImgHandle;
		    switch (inImgInfo.xColorType) {
		    case IO_COLORTYPE_BINARY:
			if (inImg->fields.PhotometricInterpretation !=
			    TFX_TIFFPHOTOMETRIC_WHITEISZERO) {
			    recompress = 1;
			}
			break;
		    case IO_COLORTYPE_GRAY4:
		    case IO_COLORTYPE_GRAY8:
			if (inImg->fields.PhotometricInterpretation !=
			    TFX_TIFFPHOTOMETRIC_BLACKISZERO) {
			    recompress = 1;
			}
			break;
		    case IO_COLORTYPE_COLOR4:
		    case IO_COLORTYPE_COLOR8:
			if (inImg->fields.PhotometricInterpretation !=
			    TFX_TIFFPHOTOMETRIC_PALETTE) {
			    recompress = 1;
			}
			break;
		    case IO_COLORTYPE_COLOR24:
			break;
		    }
		}
	    
		/* Borrow the image characteristics for the output image */
		memcpy(&outImgInfo, &inImgInfo, sizeof(outImgInfo));

		/* Test the validity of the parameters */
		XERR_IGNORE(__error__ = TFX_CheckImageInfo(outPage,
		    &outImgInfo.dwImgLayer, &outImgInfo.dwImgOffset,
		    &outImgInfo.xColorType,
		    &outImgInfo.dwXOffset, &outImgInfo.dwYOffset,
		    &outImgInfo.dwXResolution, &outImgInfo.dwYResolution,
		    &outImgInfo.dwWidth, &outImgInfo.dwHeight,
		    &outImgInfo.dwBestStripHeight, outImgInfo.pPalette,
		    &outImgInfo.xCompression, &outImgInfo.xColorSpace));
		XERR_GEN_CLEANUP(__error__ && __error__ < IO_WARNINGS, __error__);

		/* If the strip height will change, reform the strips */
/*		if (inImgInfo.dwBestStripHeight != outImgInfo.dwBestStripHeight) { */
/*		    recompress = 1; */
/*		} */

		/* Copy the image */
		if (recompress) {
#ifdef NOIPCORE
		    XERR_PASS_CLEANUP(IO_NOSUPPORT);
#else
		    TFX_IMG *outImg = NULL;

		    /* Will be default, set elsewhere */
		    inImgInfo.dwBestStripHeight = 0;

		    /* Access the common IPWState */
		    XERR_PASS_CLEANUP(wp_TfxIPWState(((TFX_PAGE *)outPage)->tfxDoc,
			&pIPWState));

		    /* Create a PIXR image to hold the image */
		    xres = Int32_TO_ScaledInt1616(inImgInfo.dwXResolution);
		    yres = Int32_TO_ScaledInt1616(inImgInfo.dwYResolution);
		    switch (inImgInfo.xColorType) {
		    case IO_COLORTYPE_COLOR24:
			ctype = cColorRGB;
			break;
		    case IO_COLORTYPE_BINARY:
		    case IO_COLORTYPE_GRAY4:
		    case IO_COLORTYPE_COLOR4:
		    case IO_COLORTYPE_GRAY8:
		    case IO_COLORTYPE_COLOR8:
			ctype = cColorGray;
			break;
		    default:
			break;
		    }
		    channels = IO_CHANNELS(inImgInfo.xColorType);
		    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState,
			outImgInfo.dwWidth, outImgInfo.dwHeight,
			IO_BPPE(outImgInfo.xColorType), xres, yres,
			32, channels, ctype, NULL, &pPixr));
		    
		    /* Get some characteristics of the PIXR, retain the image
		     * data location for each channel(s) */
		    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &bpl));
		    for (channel = 0; channel < channels; channel++) {
			if (channels > 1) {
			    pSubPixr = NULL;
			    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr,
				channel, &pSubPixr));
			    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr,
				&imageChannels[channel]));
			} else {
			    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr,
				&imageChannels[channel]));
			}
		    }
		    
		    /* Read the entire image */
		    XERR_PASS_CLEANUP(TFX_ImageReadStart(inImgInfo.tfxImgHandle, 0));
		    
		    rowsFilled = 0;
		    XERR_PASS_CLEANUP(w_TfxReadStrip(pIPWState,
			inImgInfo.tfxImgHandle, 0, inImgInfo.dwHeight, pPixr,
			0, &rowsFilled));
		    
		    XERR_PASS_CLEANUP(TFX_ImageReadFinish(inImgInfo.tfxImgHandle));
		    
		    /* Add the image to the output document */
		    XERR_PASS_CLEANUP(TFX_AddImage(outPage, outImgInfo.dwImgLayer,
			outImgInfo.xColorType, outImgInfo.dwXOffset,
			outImgInfo.dwYOffset, outImgInfo.dwXResolution,
			outImgInfo.dwYResolution, outImgInfo.dwWidth,
			outImgInfo.dwHeight, outImgInfo.dwBestStripHeight,
			outImgInfo.pPalette, &outImgInfo.tfxImgHandle));
		    outImgHandle = outImgInfo.tfxImgHandle;
		    outImg = (TFX_IMG *)outImgHandle;
		    
		    /* Let it pick up the defaults: */
		    compression = IO_COMP_UNKNOWN;
		    quality = 0;
		    
		    /* If there are variable-sized strips, we must use their
		     * sizes */
		    if (inImgInfo.dwBestStripHeight == 0) {
			XERR_PASS_CLEANUP(TFX_UseVariableStripHeights(
			    outImgHandle));
		    } else {

			/* If there is a recommendation for a new strip size,
			 * ignore the recommendation (we must use the same
			 * strip sizes), especially for variable-sized strips
			 * (StripRowCounts). */
			if (outImgInfo.dwBestStripHeight !=
			    inImgInfo.dwBestStripHeight) {
			    outImgInfo.dwBestStripHeight =
				outImg->imgInfo.dwBestStripHeight =
				inImgInfo.dwBestStripHeight;
			}
		    }

		    XERR_PASS_CLEANUP(TFX_ImageWriteStart(outImgHandle, 0, &compression,
			&quality, &colorSpace));
		    
		    /* Write the entire image (as one strip). Let this function
		     * parcel it into strips. */
		    XERR_PASS_CLEANUP(w_TfxWriteStrip(pIPWState, outImgHandle, pPixr, 0,
			rowsFilled, 0, &rowsWritten));

		    /* Finish the write */
		    XERR_PASS_CLEANUP(TFX_ImageWriteFinish(outImgHandle));

		    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &pPixr));
#endif /* NOIPCORE */
		} else {
		    XERR_PASS_CLEANUP(wp_TfxCopyImage((TFX_IMG *)inImgInfo.tfxImgHandle,
			NULL, (TFX_PAGE *)outPage, &outImgInfo, 0, 0));
		} /* recompress */
	    } /* For each image */
        } /* For each layer */

	/* Go to the next stripe */
	primaryRow += stripeRows;
	imageHeight -= stripeRows;

    } /* For each stripe */

    /* Free the temporary buffer */
    if (cmpBuffer) {
	TFX_FREE(cmpBuffer);
	cmpBuffer = NULL, cmpBufsize = 0;
    }

    if (copyAnnotations) {
	/* Loop through the annotations */
	for (annotNum = 0; annotNum < inPageInfo.wAnnotationCount;
	     annotNum++) {
	    XERR_PASS_CLEANUP(TFX_GetAnnotationInfo(inPage, annotNum,
		&annot.wAnnotType, &annot.dwXOffset, &annot.dwYOffset,
		&annot.dwWidth, &annot.dwHeight, &annot.dwFlags,
		&annot.dwAnnotLength));
	
	    IO_CHECKMALLOC(annotData = (UInt8 *)TFX_MALLOC(annot.dwAnnotLength));
	    XERR_PASS_CLEANUP(TFX_GetAnnotationData(inPage, annotNum,
		annot.dwAnnotLength, annotData));
	    XERR_PASS_CLEANUP(TFX_AddAnnotation(outPage, annot.wAnnotType,
		annot.dwXOffset, annot.dwYOffset, annot.dwWidth,
		annot.dwHeight, annot.dwFlags, annotData, annot.dwAnnotLength,
		&annotID));
	    TFX_FREE(annotData), annotData = NULL;
	} /* For each annotation */
    } /* If copying annotations */
    
    XERR_PASS_CLEANUP(TFX_AddPageFinish(outPage));
    
    NORMAL_RETURN;
	
ERR_CLEANUP:
    if (annotData) {
	TFX_FREE(annotData);
	annotData = NULL;
    }
#ifndef NOIPCORE
    if (pPixr) {
	w_destroyPixr(pIPWState, &pPixr);
    }
#endif

    ERR_RETURN;
}

/*
$Log: tfxwpub.c,v $
Revision 1.2  2003/02/10 22:53:32  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:14  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.32  2001/12/26 21:23:45  dabercro
Mark as released

Revision 1.31  2001/11/07 21:49:58  dabercro
Make base color signed long, to accomodate CIELAB

Revision 1.30  2001/08/22 19:49:34  dabercro
Added support for new field: IFDBaseColor

Revision 1.29  2001/06/27 18:44:11  dabercro
Borrow flag names from IO_ layer

Revision 1.28  2001/05/09 19:43:01  dabercro
Byte swap the extension to JBIG2, if needed, during a dictionary copy.

Revision 1.27  2001/05/09 18:36:29  dabercro
Reinsert the pos block size in the header also.

Revision 1.26  2001/05/08 23:40:52  dabercro
More old-header fix.  Check the byte count, make sure it's reasonable. The
dict count wasn't always enough.  If writing, make sure the new header
is always the "new" order, especially if rewriting an old one.

Revision 1.25  2001/05/03 16:43:10  dabercro
Added a new function UseMultiStripSubIFDs.  Rethink logic about
automatically generating multi-strip SubIFDs.

Revision 1.24  2001/05/01 00:57:18  dabercro
Make planar configuration a short.
Added fill order and reference page to copied fields.

Revision 1.23  2001/04/02 18:25:04  dabercro
Add HostComputer.

Revision 1.22  2001/03/22 20:17:39  dabercro
Add support for new TIFF field: Predictor.

Revision 1.21  2001/03/22 17:52:39  dabercro
Bug fix: some cases where images close to the stripe boundary
were being stored with the wrong stripe.

Revision 1.20  2001/02/24 03:27:19  dabercro
Use char* for string to rendering hints.

Revision 1.19  2001/01/18 00:21:32  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.18  2000/12/11 22:23:24  dabercro
Modifications to quick-copy in recommended TIFF-FX file order.

Revision 1.17  2000/11/17 23:57:35  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.16  2000/11/02 22:21:21  dabercro
Additions for Rendering Hints.

Revision 1.15  2000/10/16 18:22:54  dabercro
Major adds to support stripes, 0-byte strips, and ImageBaseColor.

Revision 1.14  2000/09/27 21:56:45  dabercro
Variable strips and stripe work.

Revision 1.13  2000/09/15 22:12:32  dabercro
Use consistent definitions, from iomrc.h. Other bug fixes.

Revision 1.12  2000/08/30 23:29:07  dabercro
Check for both JBIG2 and Tokenization.

Revision 1.11  2000/08/30 19:37:57  dabercro
We can't assume the primary IFD is layer 2.

Revision 1.10  2000/08/25 23:52:57  dabercro
Fixes for "copy compressed", and page numbers.

Revision 1.9  2000/08/22 01:03:24  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.8  2000/08/02 23:22:33  dabercro
Ignore warnings from TFX_CheckImageInfo().

Revision 1.7  2000/08/01 00:08:57  dabercro
Shared resource changes.

Revision 1.6  2000/06/07 00:00:17  dabercro
Continued work, lots of stuff.

Revision 1.5  2000/05/09 01:52:58  dabercro
Added FillOrder field support.
Updated JBIG2 dictionary front-end.
Support for JBIG2 dictionary extensions, updated parameter to
reflect byte count for extensions.
Write fields that are explicitly set, regardless of default values.
Check if document is single-page when closing, reflect in
NewSubfileType.

Revision 1.4  2000/04/12 19:06:09  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.3  2000/03/11 02:06:35  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.2  2000/02/08 21:25:34  dabercro
Disable JBIG2 (for now), updates to make use of GlobalParametersIFD for symbol libraries.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.28  1998/07/30 22:39:44  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

CommonCheck: Prototype updates for signed-ness.

Revision 1.27  1998/03/25 23:25:15  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

CommonCheck: Use a unique document ID for each document handle. This allows tagging
each document so that during "quick copy", the token dictionaries are
copy only once, and only once (in the case of two destination documents
somehow ending up with the same memory location for the structure data).

Revision 1.26  1998/02/09 20:20:16  aber
aber on Mon Feb 9 12:17:48 PST 1998

tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Flush token dictionary for both source and destination when preparing
to copy a page, instead of just the source. This was causing the destination
token dictionary to get trampled on, in the case where the destination
file had both new pages and copied "old" pages.

Revision 1.25  1998/01/20 03:22:40  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Before reading a tokenized image, make sure it's flushed to disk.

Revision 1.24  1998/01/16 00:41:38  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Allow page inserts.

Revision 1.23  1997/11/24 21:08:11  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Check strip height before deciding whether to recompress during
a page copy in compressed mode.

Revision 1.22  1997/11/20 02:54:05  aber
aber on Wed Nov 19 18:53:36 PST 1997

tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

If copying from and to images with different strip sizes, recompress.

Revision 1.21  1997/11/18 23:44:55  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Add new parameter for disabling annotation copy during page copy.
Add annotation flags argument to annotation calls.

Revision 1.20  1997/11/17 19:26:01  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

CommonCheck: Allow negative page offsets to automatically append the page to
the end of the document.

Revision 1.19  1997/11/14 23:04:42  aber
aber on Fri Nov 14 15:03:43 PST 1997

tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Fix bug in copy page, the layer wasn't being set correctly when
not recompressing some images.

Revision 1.18  1997/11/11 22:44:27  aber
aber on Tue Nov 11 14:43:41 PST 1997

tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Allow new image parameters to change when using TFX_CopyImage() in lower
level function wp_TfxCopyImage().

Revision 1.17  1997/11/06 22:49:53  aber
aber on Thu Nov 6 14:49:48 PST 1997

tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Cosmetic changes.

Revision 1.16  1997/10/27 18:13:20  aber
aber on Mon Oct 27 10:12:18 PST 1997

tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Change handling of error check, to avoid condition compiler "warns" about.

Revision 1.15  1997/10/24 21:38:59  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

CommonCheck: More exhaustive checking on calling sequence.

Revision 1.14  1997/10/16 19:04:22  aber
aber on Thu Oct 16 12:01:55 PDT 1997

tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Improve the functionality of the CopyPage function (quick copy). Now
the images are checked for consistency across version change copies.
If the output version has different support, the image is decompressed
and recompressed automatically. Also, the multiple images in layer 2
are recombined when the output is TIFF-FX 2.0.

Revision 1.13  1997/10/02 00:51:36  aber
aber on Wed Oct 1 17:50:33 PDT 1997

tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

If no token dictionaries referenced by the tokenization stream
(blank image), don't allocate a 0-byte array for the dictionary
references when copying in compressed mode.

Revision 1.12  1997/08/29 17:28:07  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.11  1997/08/15 23:48:02  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

CommonCheck: Collective decision by Dan Huttenlocher, William Rucklidge, and I
to remove the token dictionary ID from the symbol library table
entry, since the offset is typically used, and the ID does nothing
for us.

Revision 1.10  1997/08/15 01:54:20  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

CommonCheck: Modify token dictionary header to include the Tunrda dictionary
byte count, allows for extensible dictionary data.

Revision 1.9  1997/08/15 00:10:46  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Use predefined pre-Tundra header offsets for dictionary copies.

Revision 1.8  1997/08/07 21:27:49  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

CommonCheck: Make provisions for a 4-byte dictionary block size, instead of 2-bytes
spelled out in Williams table. Also make provisions for multiple
dictionaries per page.

Revision 1.7  1997/07/10 22:01:44  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Use 16-bit value for the annotation ID.

Revision 1.6  1997/06/17 01:29:36  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

CommonCheck: TIFF-FX 2.0 support. Annotation bug fix.

Revision 1.5  1997/06/10 01:26:58  aber
tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.3  1997/05/30 22:36:52  aber
aber on Fri May 30 15:36:20 PDT 1997

tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Use ERR_PASS vs. ERR_PASS_CLEANUP to avoid possible use of uninitialized
variable.

Revision 1.2  1997/05/30 17:14:31  aber
aber on Fri May 30 10:14:13 PDT 1997

tfxwpub.c -> /product/ipcore/ipwrap/src/RCS/tfxwpub.c,v

Check variable ranges for validity, avoids Watcom complaint.

Revision 1.1  1997/05/30 00:51:16  aber
Initial revision

*/
