/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxwtfx.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tfxwtfx.c,v $; $Revision: 1.3 $; $Date: 2003/02/11 01:31:43 $");

#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "geomadt.h"
#include "props.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif
#endif	/* NOIPCORE */

#include "tfx.h"
#include "tfximage.h"
#include "tfxwtiff.h"
#include "tfxwtfx.h"
#include "tfxwpriv.h"
#include "tfxrtfx.h"
#include "tfxrpriv.h"

#ifndef NOIPCORE
/* The idea is to remove this one day: */
#include "tfx.pub"
#endif


#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write an individual TIFF-FX extension (at current file location)
 *   
 *
 * LongDescription:::
 *   wp_TfxWriteExtension() writes an individual TIFF-FX extension
 *   (TFX_DOCFIELD) at the current file location.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  docField::			TIFF-FX fields to write
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxWriteExtension(
    TFX_DOC *tfxDoc,
    TFX_DOCFIELD *docField
)
{
    TFX_FILE *tfxFile = TFX_DOCFILE(tfxDoc);
    PROLOGUE;

    TFX_WRITE(tfxFile, &docField->tagID, sizeof(docField->tagID));
    TFX_WRITE(tfxFile, &docField->offset, sizeof(docField->offset));

    NORMAL_RETURN;
ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write TIFF-FX extension fields data, derived from offsets
 *
 * LongDescription:::
 *  wp_TfxWriteFieldData() writes the data associated with a TIFF-FX fields
 *  (TFX_DOCFIELDs).
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  pTfxDocInfo::			Pointer to document info, internal data copied
 *				here to expose to client
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released(incomplete)
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxWriteFieldData(
    TFX_DOC *tfxDoc,
    TFX_DOCINFO *pTfxDocInfo
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
/*    UInt16 sval; UInt32 lval; */
    PROLOGUE;
    
    TFX_CheckHandle(tfxDoc);
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);
    
    /* Check for proper calling sequence */
    
#if 0   
    /* Write the page table header */
    TFX_SEEK(tfxFile, tfxDoc->pageTableHeaderOffset);
    sval = tfxDoc->pageTableHeader.pageCount;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    sval = tfxDoc->pageTableHeader.entriesPerTable;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    sval = tfxDoc->pageTableHeader.lastOrdinal;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    lval = tfxDoc->pageTableHeader.firstTableOffset;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));

    /* Fill in the page count */
    tfxDoc->docInfo.wNPages = tfxDoc->pageTableHeader.pageCount;
    pTfxDocInfo->wNPages = tfxDoc->pageTableHeader.pageCount;
#endif
    
    NORMAL_RETURN;
    
/* ERR_CLEANUP: */
    
/*    ERR_RETURN; */
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write TIFF-FX header and necessary components to initiate a TIFF-FX file.
 *
 * LongDescription:::
 *  wp_TfxWriteHeader() writes the TIFF and TIFF-FX headers, and the necessary
 *  components to initiate a TIFF-FX file.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  pVersion::			Pointer to the TIFF-FX version
 *  pRevision::			Pointer to the TIFF-FX revision
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxWriteHeader(
    TFX_DOC *tfxDoc,
    UInt8 *pVersion,
    UInt8 *pRevision
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    PROLOGUE;
    
    TFX_CheckHandle(tfxDoc);
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);
    
    /* Check for proper calling sequence */


    /* Begin a write sequence */
    if ((tfxDoc->seqStatus & TFX_DOC_SEQ_WRITESTART) == 0) {
	tfxDoc->seqStatus |= TFX_DOC_SEQ_WRITESTART;
    }
    
    /* Write a TIFF header first */
    XERR_PASS_CLEANUP(wp_TifFXWriteHeader(tfxFile, 0UL));

    /* Keep the version/revision */
    tfxDoc->docInfo.version = *pVersion;
    tfxDoc->docInfo.revision = *pRevision;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write or rewrite the author block
 *
 * LongDescription:::
 *  wp_TfxSetAuthor() writes the authoring block of data and attaches it to the
 *  TIFF-FX. If the authoring block previously existed, it is overwritten (if it
 *  fits), or appended to the file.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  pAuthorBlock::		Pointer to the authoring block of data
 *  wAuthorBytes::		Number of bytes in authoring block
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxSetAuthor(
	TFX_DOC *tfxDoc,
    	Int8 *pAuthorBlock,
	UInt16 wAuthorBytes
)
{
#if 0
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile = TFX_DOCFILE(tfxDoc);
    Int32 fileOffset = 0UL;
    TFX_DOCFIELD docField;
    UInt16 sval;
    PROLOGUE;
    /* LATER: see about writing to GPIFD */

    instData = TFX_DOCINST(tfxDoc);

    /* Make sure we're allowed to write */
    XERR_GEN_CLEANUP((tfxDoc->readWrite & WRITE_ALLOWED) == 0, IO_BADFILEMODE);

    /* If the file is read/write, check to see if we can write over the previous
     * authoring block */
    if (tfxDoc->readWrite & READ_ALLOWED) {

	/* Make sure we've read the current authoring block first */
	XERR_GEN_CLEANUP((tfxDoc->seqStatus & TFX_DOC_SEQ_GETINFO) == 0,
	    IO_INTERNAL_ERROR);

	if (tfxDoc->docInfo.wAuthorBlockBytes >= wAuthorBytes) {
	    fileOffset = tfxDoc->authorOffset;
	    XERR_GEN_CLEANUP(fileOffset < 8, IO_BADFILEFORMAT);

	    /* Seek to the authoring block location */
	    TFX_SEEK(tfxFile, fileOffset);
	}
    }
    
    /* If we haven't found a spot, either it didn't fit, or there wasn't one */
    if (fileOffset == 0L) {

	/* Place it at the end of the file */
	XERR_PASS(wp_TfxFileLongAlign(tfxFile, &fileOffset));
	tfxDoc->authorOffset = fileOffset;
    }

    /* Write the authoring block out */
    sval = wAuthorBytes;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    if (wAuthorBytes > 0) {
	XERR_GEN_CLEANUP(pAuthorBlock == NULL, IO_BADPARAMETER);
	TFX_WRITE(tfxFile, pAuthorBlock, wAuthorBytes);
	IO_CHECKMALLOC(tfxDoc->docInfo.pAuthorBlock = TFX_MALLOC(wAuthorBytes));
	memcpy(tfxDoc->docInfo.pAuthorBlock, pAuthorBlock, wAuthorBytes);
    } else {
	tfxDoc->docInfo.pAuthorBlock = NULL;
    }
    tfxDoc->docInfo.wAuthorBlockBytes = wAuthorBytes;

    /* Update the TIFF-FX authoring block fields
     */
    docField.tagID = DOCTAG_AUTHOR;
    docField.offset = tfxDoc->authorOffset;
    fileOffset = tfxDoc->tagOffset[docField.tagID];
    XERR_GEN_CLEANUP(fileOffset < 8, IO_INTERNAL_ERROR);

    /* Write the authoring block fields at the correct location */
    TFX_SEEK(tfxFile, fileOffset);
    XERR_PASS_CLEANUP(wp_TfxWriteExtension(tfxDoc, &docField));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
#else
    XERR_GEN(1, IO_NOSUPPORT);

    /* Avoid compiler warnings */
    pAuthorBlock = pAuthorBlock;
    wAuthorBytes = wAuthorBytes;
    tfxDoc = tfxDoc;
#endif
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write or rewrite the reference page offset
 *
 * LongDescription:::
 *  wp_TfxSetReferencePage() writes the reference page offset to the TIFF-FX.
 *  The reference page field is attached to the GlobalParametersIFD, if it
 *  exists; otherwise, the GlobalParametersIFD is created.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  dwReferencePageOffset::	Reference page offset
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxSetReferencePage(
	TFX_DOC *tfxDoc,
	Int32 dwReferencePageOffset
)
{
    TFX_FILE *tfxFile = TFX_DOCFILE(tfxDoc);
    Int32 fileOffset = 0UL;
    Int32 lval;
    PROLOGUE;

    /* Retain the value (especially if we really do the writing later) */
    tfxDoc->docInfo.dwReferencePageOffset = dwReferencePageOffset;

    /* Reflect the write in the status */
    tfxDoc->seqStatus |= TFX_DOC_SEQ_REFPAGESET;

    /* Make sure we're allowed to write */
    XERR_GEN_CLEANUP((tfxDoc->readWrite & WRITE_ALLOWED) == 0,
	IO_BADFILEMODE);

    /* If the file is read/write, check to see if we can write over the
     * previous one */
    if (tfxDoc->readWrite & READ_ALLOWED) {

	/* Make sure we've at least read the first page */
	XERR_GEN_CLEANUP((tfxDoc->seqStatus & TFX_DOC_SEQ_GETINFO) == 0,
	    IO_INTERNAL_ERROR);
    }

    if ((fileOffset = tfxDoc->refPageOffsetLocation) > 0) {
	XERR_GEN_CLEANUP(fileOffset < 8, IO_BADFILEFORMAT);

	/* Seek to the existing reference page data location */
	TFX_SEEK(tfxFile, fileOffset);

    } else {
    
	/* If we haven't found a spot, either it didn't fit, or there wasn't one */

	/* Check to see if the GlobalParametersIFD was written (or read) */
	if ((tfxDoc->seqStatus & TFX_DOC_SEQ_GPEXISTS) == 0) {

	    /* Nope, it will get written when the GlobalParametersIFD gets written */
	    NORMAL_RETURN;
	}

	/* Yep, the GlobalParametersIFD was written, yet there was no reference page.
	 * This function should have been called first. */
	XERR_PASS_CLEANUP(IO_SEQUENCE);
    }

    /* Stored value is page number (ordinal), not offset */
    lval = tfxDoc->docInfo.dwReferencePageOffset + 1;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a page table entry.
 *
 * LongDescription:::
 *   wp_TfxWritePageTableEntry() writes an individual page table entry.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  fileOffset::		Where to write the table entry
 *  pageTableEntry::		Table entry to write
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxWritePageTableEntry(
    TFX_DOC *tfxDoc,
    Int32 fileOffset,
    TFX_PAGETABLEENTRY *pageTableEntry
)
{
    TFX_FILE *tfxFile = TFX_DOCFILE(tfxDoc);
    TFX_PAGETABLEENTRY tempEntry;
    PROLOGUE;

    /* Seek to the entry location */
    TFX_SEEK(tfxFile, fileOffset);

    /* Copy to a temporary entry (for swapping) */
    memcpy(&tempEntry, pageTableEntry, sizeof(tempEntry));

    /* Swap if necessary */
    TFX_SWAPLONG(tfxFile, &tempEntry.IFDoffset);

    /* Write this puppy out */
    TFX_WRITE(tfxFile, &tempEntry.flags, sizeof(tempEntry.flags));
    TFX_WRITE(tfxFile, &tempEntry.IFDoffset, sizeof(tempEntry.IFDoffset));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rewrites the page table entries, used to accomodate a page insert/delete
 *
 * LongDescription:::
 *  wp_TfxRewritetPageTableEntries() starts with a given page, and rewrites
 *  the page table entries. The TIFF primary IFD nextIFD entries are also
 *  checked for consistency, and adjusted if necessary (as is the case AFTER
 *  the insertion or deletion of a page from the document).
 *  <BR>
 *  It is assumed that after the insertion or deletion of a page (or pages),
 *  that the internal linked list of pages is now correct, and that the page
 *  table needs to be adjusted to match the insertion or removal of a link or
 *  links.
 *  <BR>
 *  All shifted data is taken from that of the internal data structures --
 *  the entire page table MUST be read (or written) prior to this function.
 *
 * Input:::
 *  tfxDoc::			Document handle, required when last page was
 *				just deleted.
 *  tfxPage::			Page where shift will begin (inclusive). This
 *				is only a hint, the very first page could be
 *				used, but only the affected pages will result
 *				in change.
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxRewritePageTableEntries(
    TFX_DOC *tfxDoc,
    TFX_PAGE *tfxPage
)
{
    Int32 pages;
#if 0
    Int32 startOffset;
    Int32 shiftsLeft;			/* Shifts left to apply */
    Int32 srcEntryOffset, dstEntryOffset;
    TFX_PAGE *dstPage;
#endif
    Int32 pageOffset;
    UInt8 flags;
    Int32 ifdOffset;
    TFX_IMG *tfxImg;
    PROLOGUE;

    TFX_CheckHandle(tfxDoc);

    /* Allow the case where the page handle is NULL (no pages = last page
     * was just deleted) */
    if (tfxPage) {
	TFX_CheckHandle(tfxPage);
	pages = tfxDoc->docInfo.wNPages;
	pageOffset = tfxPage->pageInfo.dwPageOffset;  /* This must be correct */
    } else {
	pageOffset = 0;

	/* No pages remain: special case, reset the TIFF header to have no first
	 * page */
	XERR_PASS_CLEANUP(wp_TfxSetPageIFD(tfxDoc, pageOffset, 0, 0, NULL));
    }

    /* Compute the last entry positions to be copied (source and destination)
     */
    while (tfxPage) {

	/* Make sure things are consistent */
	XERR_GEN_CLEANUP(pageOffset != tfxPage->pageInfo.dwPageOffset,
	    IO_INTERNAL_ERROR);

	/* Get the page table information */
	XERR_PASS_CLEANUP(wp_TfxGetPageIFD(tfxDoc, pageOffset, &flags,
	    &ifdOffset));

	/* Update the IFD pointers info if not correct */
	tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;
	if (tfxImg && ifdOffset != tfxImg->ifdOffset) {

	    /* Write the source information to the new destination position */
	    XERR_PASS_CLEANUP(wp_TfxGeneratePageFlags(tfxPage, &flags));
	    XERR_PASS_CLEANUP(wp_TfxSetPageIFD(tfxDoc, pageOffset, flags,
		tfxImg->ifdOffset, tfxPage));
	}
      
	/* Go to the next page */
	tfxPage = tfxPage->next;
	pageOffset++;
    }

#if 0


    if (shifts > 0) {	/* Insertion case */

	srcEntryOffset = pages - 1 - shifts;	/* New last page position */
	dstEntryOffset = pages - 1;		/* Last page */
	shiftsLeft = srcEntryOffset - startOffset + 1;

	/* Get the new destination page */
	pageOffset = 0;
	dstPage = tfxDoc->pages;
	while (pageOffset++ < dstEntryOffset) {
	    dstPage = dstPage->next;
	    XERR_GEN_CLEANUP(dstPage == NULL, IO_INTERNAL_ERROR);
	}

	/*
         * Regenerate the destination entries, one at a time, and place them in
	 * their new (shifted) positions.
	 */
	while (shiftsLeft > 0) {
	    /* Get the source information */
	    XERR_PASS_CLEANUP(wp_TfxGetPageIFD(tfxDoc, srcEntryOffset, &flags,
		&ifdOffset));

		/* Update the internal structure so that the nextIFD doesn't change,
		 * which would trigger a rewrite of the nextIFD field during the call
		 * to wp_TfxSetPageIFD() */

	    /* Write the source information to the new destination position */
	    XERR_PASS_CLEANUP(wp_TfxSetPageIFD(tfxDoc, dstEntryOffset, flags,
		ifdOffset, dstPage));
      
	    /* Go to the next page (previous) */
	    srcEntryOffset--, dstEntryOffset--, shiftsLeft--;
	    dstPage = dstPage->prev;
	    XERR_GEN_CLEANUP(dstPage == NULL, IO_INTERNAL_ERROR);
	}
    } else {	/* Deletion case */

	srcEntryOffset = startOffset - shifts;		/* New "next" page */
	dstEntryOffset = startOffset;			/* First "clobbered" page */
	shiftsLeft = pages - srcEntryOffset;
	dstPage = tfxPage;

	/*
	 * Regenerate the destination entries, one at a time, and place them in
	 * their new (shifted) positions.
	 */
	while (shiftsLeft > 0) {
	    /* Get the source information */
	    XERR_PASS_CLEANUP(wp_TfxGetPageIFD(tfxDoc, srcEntryOffset, &flags,
		&ifdOffset));

	    /* Update the internal structure so that the nextIFD doesn't change,
	     * which would trigger a rewrite of the nextIFD field during the call
	     * to wp_TfxSetPageIFD() */
	    
	    /* Write the source information to the new destination position */
	    XERR_PASS_CLEANUP(wp_TfxSetPageIFD(tfxDoc, dstEntryOffset, flags,
		ifdOffset, dstPage));
      
	    /* Go to the next page */
	    srcEntryOffset++, dstEntryOffset++, shiftsLeft--;
	    dstPage = dstPage->next;
	    XERR_GEN_CLEANUP(dstPage == NULL, IO_INTERNAL_ERROR);
	}
    }

    /* If there's nothing that needs shifting, we don't need to continue */
    if (startOffset >= pages || shiftsLeft <= 0) {
	NORMAL_RETURN;
    }

#endif


    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add an annotation to a page
 *   
 * LongDescription:::
 *  wp_TfxAddAnnotation() adds annotation data and an updated annotation
 *  offsets list of a TIFF-FX 2.0 file, or the annotation table entry (and table)
 *  of a TIFF-FX 3.0 and later file.
 *
 * Input:::
 *  tfxPage::			Data structure of valid page
 *  type::			Type of annotation
 *  xOffset::			X offset in page coordinates
 *  yOffset::			Y offset in page coordinates
 *  width::			Width in page coordinates
 *  height::			Height in page coordinates
 *  flags::			Annotation flags (see ANNOT_... in tfxpags.h)
 *  pData::			Annotation data
 *  bytes::			Number of bytes in annotation data
 *
 * Output:::
 *  pID::			ID assigned to this annotation
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxAddAnnotation(
	      TFX_PAGE *tfxPage,
	      UInt16 type,
	      UInt32 xOffset,
	      UInt32 yOffset,
	      UInt32 width,
	      UInt32 height,
	      UInt32 flags,
	      UInt8 *pData,
	      UInt32 bytes,
	      UInt16 *pID
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt16 annotationOffset;
    UInt32 tableOffset;
    UInt32 entryOffset;
    TFX_ANNOTTABLE *prevTbl, *tblPtr;
    TFX_ANNOTTABLE *newTblPtr = NULL;
    TFX_ANNOTTABLEENTRY *ep;
    Int32 tblOffset, prevTblOffset;
    UInt32 mallocbytes;
    Int32 fileOffset;
    UInt16 sval; UInt32 lval;
    Int32 annotationFileOffset = 0UL;
    PROLOGUE;

    /* Check parameters / mode */
    TFX_CheckHandle(tfxPage);
    instData = TFX_PAGEINST(tfxPage);
    tfxFile = TFX_PAGEFILE(tfxPage);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxPage->seqStatus &
	(TFX_PAGE_SEQ_READSTART | TFX_PAGE_SEQ_WRITESTART)) == 0),
	IO_SEQUENCE);
    XERR_GEN_CLEANUP((tfxPage->tfxDoc->readWrite & WRITE_ALLOWED) == 0,
	IO_INTERNAL_ERROR);

    /* For now, we must have a primary image attached */
    XERR_GEN_CLEANUP(tfxPage->pageInfo.tfxPrimImgHandle == 0,
		     IO_SEQUENCE);

    /* Write the annotation table header if none is attached */
    if (tfxPage->annotTableHeader.entriesPerTable == 0UL) {
	XERR_GEN_CLEANUP(tfxPage->annotTableHeader.annotCount > 0,
			 IO_INTERNAL_ERROR);
	if (tfxPage->tfxDoc->annotsPerTable == 0) {
	    tfxPage->tfxDoc->annotsPerTable  = 4;
	}
	tfxPage->annotTableHeader.entriesPerTable =
	    tfxPage->tfxDoc->annotsPerTable;

	/* Figure out the byte count of a table entry */
	tfxPage->annotTableHeader.bytesPerEntry =
	    sizeof(ep->id) + sizeof(ep->type) +
	    sizeof(ep->xOffset) + sizeof(ep->yOffset) +
	    sizeof(ep->width) + sizeof(ep->height) +
	    sizeof(ep->flags) + sizeof(ep->bytes) + sizeof(ep->offset);

	XERR_PASS_CLEANUP(wp_TfxWriteAnnotTableHeader(tfxPage));
    }

    /* We can only add the next annotation, not allow a gap in the list */
    annotationOffset = tfxPage->annotTableHeader.annotCount;
    
    tableOffset = annotationOffset / tfxPage->annotTableHeader.entriesPerTable;
    entryOffset = annotationOffset % tfxPage->annotTableHeader.entriesPerTable;

    /* Make sure we've read all the annotation tables if the destination is
     * a read-write file */
    if ((tfxPage->tfxDoc->readWrite & READ_ALLOWED) &&
	tfxPage->annotTableHeader.annotCount > 0) {
	UInt32 tmpTfxAnnotationOffset;
	UInt16 tmpTableOffset;
	TFX_ANNOTTABLE *pAnnotPageTable = NULL;

	tmpTfxAnnotationOffset = tfxPage->annotTableHeader.annotCount - 1;
	tmpTableOffset = (UInt16)(tmpTfxAnnotationOffset /
	    tfxPage->annotTableHeader.entriesPerTable);

	XERR_PASS_CLEANUP(wp_TfxGetAnnotTable(tfxPage, tmpTableOffset,
					      &pAnnotPageTable));
    }

    /* Search for the table itself */
    prevTbl = tblPtr = tfxPage->firstAnnotTable;
    prevTblOffset = 0UL;
    tblOffset = tfxPage->annotTableHeader.firstTableOffset;
    while (tblPtr) {
	if (tblPtr->ordinal == tableOffset + 1) {
	    break;
	}
	prevTblOffset = tblOffset;
	tblOffset = tblPtr->nextTableOffset;	
	prevTbl = tblPtr;
	tblPtr = tblPtr->next;
    }

    /* Create a new table if the one desired doesn't exist yet */
    if (tblPtr == NULL) {

	IO_CHECKMALLOC(newTblPtr = (TFX_ANNOTTABLE *)TFX_MALLOC(sizeof(TFX_ANNOTTABLE)));
	memset(newTblPtr, 0, sizeof(TFX_ANNOTTABLE));
	mallocbytes = sizeof(TFX_ANNOTTABLEENTRY) * 
	    tfxPage->annotTableHeader.entriesPerTable;
	IO_CHECKMALLOC(newTblPtr->entries =
		     (TFX_ANNOTTABLEENTRY *)TFX_MALLOC(mallocbytes));
	memset(newTblPtr->entries, 0, mallocbytes);
	newTblPtr->ordinal = (UInt16)tableOffset + 1;

	/* Add to the linked list */
	if (prevTbl) {

	    /* Attach to end of linked list */
	    prevTbl->next = newTblPtr;

	} else {

	    /* Begin list here */
	    tfxPage->firstAnnotTable = newTblPtr;
	    tfxPage->annotTableHeader.firstTableOffset = tblOffset;
	}

	/* Write the annotation table */
	XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &tblOffset));
	sval = newTblPtr->ordinal;
	TFX_SWAPSHORT(tfxFile, &sval);
	TFX_WRITE(tfxFile, &sval, sizeof(sval));

	/* Regardless of structure alignment issues, the structure entry
	 * data is at least the size of the file entries, and is zero'd
	 * out. Just write out the number of 0-bytes to fill the file
	 * entries */
	TFX_WRITE(tfxFile, newTblPtr->entries,
	    tfxPage->annotTableHeader.entriesPerTable *
	    tfxPage->annotTableHeader.bytesPerEntry);
	lval = newTblPtr->nextTableOffset;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));

	if (prevTbl) {

	    /* Look for the annotation table 'nextTableOffset' location
	     * of the table were attaching this one to */
	    fileOffset = prevTblOffset +
		sizeof(tfxPage->firstAnnotTable->ordinal) +
		tfxPage->annotTableHeader.entriesPerTable *
		tfxPage->annotTableHeader.bytesPerEntry;
	    XERR_GEN_CLEANUP(fileOffset < 8, IO_INTERNAL_ERROR);

	    /* Link the new table to the previous one, in the file */
	    TFX_SEEK(tfxFile, fileOffset);
	    lval = tblOffset;
	    TFX_SWAPLONG(tfxFile, &lval);
	    TFX_WRITE(tfxFile, &lval, sizeof(lval));
	    prevTbl->nextTableOffset = tblOffset;
	} else {
	    tfxPage->annotTableHeader.firstTableOffset = tblOffset;
	}
	tblPtr = newTblPtr;
	newTblPtr = NULL;	/* Avoid freeing if later failure */

	/* Update the header */
	tfxPage->annotTableHeader.lastOrdinal++;
    } /* Adding a new table */

    /* Prepare to write the data at the end of the file */
    XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &annotationFileOffset));

    /* Write the annotation data */
    TFX_WRITE(tfxFile, pData, (Int32)bytes);

    /* Write the new entry in the annotation table */
    XERR_GEN_CLEANUP(tblOffset == 0, IO_INTERNAL_ERROR);
    fileOffset = tblOffset + sizeof(tfxPage->firstAnnotTable->ordinal) +
	entryOffset * tfxPage->annotTableHeader.bytesPerEntry;
    XERR_GEN_CLEANUP(fileOffset < 8, IO_INTERNAL_ERROR);
    ep = &tblPtr->entries[entryOffset];
    ep->id = annotationOffset;
    ep->type = (UInt16)type;
    ep->xOffset = xOffset;
    ep->yOffset = yOffset;
    ep->width = width;
    ep->height = height;
    ep->flags = flags;
    ep->bytes = bytes;
    ep->offset = annotationFileOffset;
    XERR_PASS_CLEANUP(wp_TfxWriteAnnotTableEntry(tfxPage, fileOffset, ep));

    /* Increment the total count */
    tfxPage->annotTableHeader.annotCount++;    
    tfxPage->pageInfo.wAnnotationCount++;    
    XERR_PASS_CLEANUP(wp_TfxWriteAnnotTableHeader(tfxPage));

    /* Return the annotation ID (offset) */
    *pID = annotationOffset;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (newTblPtr) {
	if (newTblPtr->entries) {
	    TFX_FREE(newTblPtr->entries);
	}
	TFX_FREE(newTblPtr);
    }
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write/rewrite the annotation table header
 *   
 *
 * LongDescription:::
 *  wp_TfxWriteAnnotTableHeader() writes or rewrites the annotation table header
 *  to take account of an update in the information (usually the number of
 *  annotations).
 *
 * Input:::
 *  tfxPage::			Data structure of valid page
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxWriteAnnotTableHeader(
    TFX_PAGE *tfxPage
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt16 sval; UInt32 lval;
    PROLOGUE;
    
    TFX_CheckHandle(tfxPage);
    instData = TFX_PAGEINST(tfxPage);
    tfxFile = TFX_PAGEFILE(tfxPage);

    /* Check for proper calling sequence */


    /* If this is the first time it is written, place at the end of file */
    if (tfxPage->annotTableHeaderOffset == 0UL) {
	TFX_IMG *primImg;

	/* Make sure we have the TIFF fields written */
	primImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;
	XERR_GEN_CLEANUP(primImg == NULL, IO_SEQUENCE);

	/* Place the annotation table header at the end of the file */
	XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile,
	    &tfxPage->annotTableHeaderOffset));

	/* Update the TIFF fields */
	if (primImg->fields.AnnotationOffsets == NULL) {
	    IO_CHECKMALLOC(primImg->fields.AnnotationOffsets =
		TFX_MALLOC(1 * sizeof(*primImg->fields.AnnotationOffsets)));
	    primImg->fields.AnnotationCount = 1;
	}
	primImg->fields.AnnotationOffsets[0] = tfxPage->annotTableHeaderOffset;
	if (primImg->fieldLocation.AnnotationOffsets != 0) {
	    Int32 lval;

	    /* Actually attach it */
	    TFX_SEEK(tfxFile, primImg->fieldLocation.AnnotationOffsets);
	    lval = tfxPage->annotTableHeaderOffset;
	    TFX_SWAPLONG(tfxFile, &lval);
	    TFX_WRITE(tfxFile, &lval, sizeof(lval));
	} else if (primImg->fields.dataTransferred) {
	    primImg->fields.IFDrewriteNeeded = 1;
	}
    }

    /* If we have written the primary image, attach the annotations */

    /* Seek to the header */
    TFX_SEEK(tfxFile, tfxPage->annotTableHeaderOffset);

    /* Write out the header */
    sval = tfxPage->annotTableHeader.annotCount;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    sval = tfxPage->annotTableHeader.entriesPerTable;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    sval = tfxPage->annotTableHeader.bytesPerEntry;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    sval = tfxPage->annotTableHeader.lastOrdinal;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    lval = tfxPage->annotTableHeader.firstTableOffset;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write an annotation table entry.
 *   
 * LongDescription:::
 *  wp_TfxWriteAnnotTableEntry() stores a single annotation entry.
 *
 * Input:::
 *  tfxPage::			Data structure of valid page
 *  fileOffset::		Where to write the table entry
 *  annotTableEntry::		Table entry to write
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxWriteAnnotTableEntry(
    TFX_PAGE *tfxPage,
    Int32 fileOffset,
    TFX_ANNOTTABLEENTRY *annotTableEntry
)
{
    TFX_FILE *tfxFile = TFX_PAGEFILE(tfxPage);
    TFX_ANNOTTABLEENTRY tempEntry;
    UInt32 bytesLeft;
    PROLOGUE;

    /* Seek to the entry location */
    TFX_SEEK(tfxFile, fileOffset);

    /* Copy to a temporary entry (for swapping) */
    memcpy(&tempEntry, annotTableEntry, sizeof(tempEntry));

    /* Swap if necessary */
    TFX_SWAPSHORT(tfxFile, &tempEntry.id);
    TFX_SWAPSHORT(tfxFile, &tempEntry.type);
    TFX_SWAPLONG(tfxFile, &tempEntry.xOffset);
    TFX_SWAPLONG(tfxFile, &tempEntry.yOffset);
    TFX_SWAPLONG(tfxFile, &tempEntry.width);
    TFX_SWAPLONG(tfxFile, &tempEntry.height);
    TFX_SWAPLONG(tfxFile, &tempEntry.flags);
    TFX_SWAPLONG(tfxFile, &tempEntry.bytes);
    TFX_SWAPLONG(tfxFile, &tempEntry.offset);

    /* Keep track of how many bytes we're allowed to write */
    bytesLeft = tfxPage->annotTableHeader.bytesPerEntry;

    /* Write this puppy out */
    if (bytesLeft >= sizeof(tempEntry.id)) {
	TFX_WRITE(tfxFile, &tempEntry.id, sizeof(tempEntry.id));
	bytesLeft -= sizeof(tempEntry.id);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.type)) {
	TFX_WRITE(tfxFile, &tempEntry.type, sizeof(tempEntry.type));
	bytesLeft -= sizeof(tempEntry.type);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.xOffset)) {
	TFX_WRITE(tfxFile, &tempEntry.xOffset, sizeof(tempEntry.xOffset));
	bytesLeft -= sizeof(tempEntry.xOffset);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.yOffset)) {
	TFX_WRITE(tfxFile, &tempEntry.yOffset, sizeof(tempEntry.yOffset));
	bytesLeft -= sizeof(tempEntry.yOffset);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.width)) {
	TFX_WRITE(tfxFile, &tempEntry.width, sizeof(tempEntry.width));
	bytesLeft -= sizeof(tempEntry.width);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.height)) {
	TFX_WRITE(tfxFile, &tempEntry.height, sizeof(tempEntry.height));
	bytesLeft -= sizeof(tempEntry.height);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.bytes)) {
	TFX_WRITE(tfxFile, &tempEntry.bytes, sizeof(tempEntry.bytes));
	bytesLeft -= sizeof(tempEntry.bytes);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.offset)) {
	TFX_WRITE(tfxFile, &tempEntry.offset, sizeof(tempEntry.offset));
	bytesLeft -= sizeof(tempEntry.offset);
    } else {
	goto EntryDone;
    }
    if (bytesLeft >= sizeof(tempEntry.flags)) {
	TFX_WRITE(tfxFile, &tempEntry.flags, sizeof(tempEntry.flags));
	bytesLeft -= sizeof(tempEntry.flags);
    } else {
	goto EntryDone;
    }

 EntryDone:
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Delete an annotation from a page
 *   
 * LongDescription:::
 *  wp_TfxDeleteAnnotation() deletes an annotation from a page. For 2.0 files,
 *  the annotation offsets list is rewritten, with the deleted annotation
 *  removed from the list. For 3.0 and later versions of TIFF-FX, the table
 *  entries are updated to reflect one less annotation; if appropriate, the
 *  entries are shifted by one, including any shifts from later tables. For
 *  this reason, it might be most efficient to delete the latest entry
 *  first.
 *
 * Input:::
 *  tfxPage::			Data structure of valid page
 *  offset::			Annotation offset
 *
 * Output:::
 *  *pdwAnnotationCount::	New annotation count
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  IDs returned to any of the annotations on this page will be invalid
 *  after an annotation is deleted. This is due to the fact that the ID
 *  is the offset of the annotation in an absolute list (0, 1, 2, ...),
 *  which may change depending on the position of the deleted annotation.
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxDeleteAnnotation(
	      TFX_PAGE *tfxPage,
	      UInt16 offset,
	      UInt16 *pdwAnnotationCount
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt16 entryOffset, tableOffset;
    TFX_ANNOTTABLE *prevTbl = NULL, *al = NULL, *annotTable;
    TFX_ANNOTTABLEENTRY *srcEp, *dstEp, *prevEp, *storeEp = NULL;
    UInt16 annotOffset, annotCount;
    UInt8 *entryData = NULL, *bp;
    UInt16 shiftEntryCount;
    UInt16 storeEntryCount;
    UInt32 storeEntryOffset = 0UL;
    UInt16 sval; UInt32 lval;
    UInt32 prevTableFileOffset = 0UL;	/* File offset of previous table */
    UInt32 thisTableFileOffset = 0UL;	/* File offset of this table */
    PROLOGUE;

    /* Check parameters / mode */
    TFX_CheckHandle(tfxPage);
    instData = TFX_PAGEINST(tfxPage);
    tfxFile = TFX_PAGEFILE(tfxPage);

    annotCount = tfxPage->annotTableHeader.annotCount;
    XERR_GEN_CLEANUP(offset >= annotCount, IO_BADPARAMETER);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxPage->seqStatus & TFX_PAGE_SEQ_GETINFO) == 0),
		     IO_SEQUENCE);
    XERR_GEN_CLEANUP((tfxPage->tfxDoc->readWrite & WRITE_ALLOWED) == 0,
		     IO_INTERNAL_ERROR);

    /* Make sure we have an accurate accouting of all the annotations */
    XERR_PASS_CLEANUP(wp_TfxGetAnnotTable(tfxPage, (UInt16)(annotCount - 1),
	&annotTable));

    /* The table entries must be modified
     */

    /* If it's already the last item, no update is necessary, go straight
     * to updating the count of annotations in the header; otherwise do
     * the shifting. */
    if (offset < annotCount - 1) {

	/* If the items do need shifting, we have to make sure we can
         * handle shifting data from later tables to earlier tables.
	 */

	/* Allocate an array that will contain an entire table of entries
	 */
	IO_CHECKMALLOC(entryData = (UInt8 *)TFX_MALLOC(
	    tfxPage->annotTableHeader.entriesPerTable *
	    tfxPage->annotTableHeader.bytesPerEntry)); 
	bp = entryData;
	annotOffset = 0;

	/* Loop through the structures stored in memory, reconstructing
	 * the pieces of the tables that would have moved in the file.
	 * Store the pieces whenever the "rest of" a table is completed.
	 */

	/* Find the first affected table */
	tableOffset = offset / tfxPage->annotTableHeader.entriesPerTable;
	al = tfxPage->firstAnnotTable;
	prevTableFileOffset = 0;
	thisTableFileOffset = tfxPage->annotTableHeader.firstTableOffset;
	annotOffset = 0;
	while (al) {
	    if (al->ordinal == tableOffset + 1) {
		break;
	    }
	    prevTableFileOffset = thisTableFileOffset;
	    thisTableFileOffset = al->nextTableOffset;
	    al = al->next;
	    annotOffset += tfxPage->annotTableHeader.entriesPerTable;
	}
	annotOffset = tableOffset * tfxPage->annotTableHeader.entriesPerTable;

	/* Compute the number of items to shift within this table */
	shiftEntryCount = tfxPage->annotTableHeader.entriesPerTable - (offset
	    - annotOffset) - 1;

	/* If this is the last table, only shift the rest of them (not the
	 * entire rest within the table */
	if (tableOffset == annotCount /
	    tfxPage->annotTableHeader.entriesPerTable) {
	    shiftEntryCount = annotCount - (offset - annotOffset) - 1;
	}
	entryOffset = offset - annotOffset;
	dstEp = al->entries + entryOffset;
	srcEp = dstEp + 1;

	/* Loop through the remaining tables, shifting the entries
	* in both the internal structures, and the tables in file */
	prevTbl = NULL;
	prevEp = NULL;
	storeEntryCount = 0;
	while (al || prevTbl) {

	    /* Finish up the copy: move first item of this table to
	     * the tail of the previous one, or simply flush the previous
	     * one
	     */
	    if (prevTbl != NULL) {

		/* Copy the first from this table to the tail of the
		 * previous table */
		if (al) {
		    srcEp = al->entries;
		    memcpy(dstEp, srcEp, sizeof(TFX_ANNOTTABLEENTRY));
		    storeEntryCount++;
		}

		/* Loop through the entries that shifted, and copy them to
		 * a buffer to be stored in the file */
		while (storeEntryCount--) {
		    sval = --storeEp->id;    /* Shift the IDs down one */
		    TFX_SWAPSHORT(tfxFile, &sval);
		    memcpy(bp, &sval, sizeof(sval)), bp += sizeof(sval);
		    sval = storeEp->type;
		    TFX_SWAPSHORT(tfxFile, &sval);
		    memcpy(bp, &sval, sizeof(sval)), bp += sizeof(sval);
		    lval = storeEp->xOffset;
		    TFX_SWAPLONG(tfxFile, &lval);
		    memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
		    lval = storeEp->yOffset;
		    TFX_SWAPLONG(tfxFile, &lval);
		    memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
		    lval = storeEp->width;
		    TFX_SWAPLONG(tfxFile, &lval);
		    memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
		    lval = storeEp->height;
		    TFX_SWAPLONG(tfxFile, &lval);
		    memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
		    lval = storeEp->flags;
		    TFX_SWAPLONG(tfxFile, &lval);
		    memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
		    lval = storeEp->bytes;
		    TFX_SWAPLONG(tfxFile, &lval);
		    memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
		    lval = storeEp->offset;
		    TFX_SWAPLONG(tfxFile, &lval);
		    memcpy(bp, &lval, sizeof(lval)), bp += sizeof(lval);
		    
		    storeEp++;
		}

		/* Store the data */
		XERR_GEN_CLEANUP(prevTableFileOffset == 0,
		    IO_INTERNAL_ERROR);
		XERR_GEN_CLEANUP((Int32)((bp - entryData) /
		    tfxPage->annotTableHeader.bytesPerEntry) >
		    (Int32)(tfxPage->annotTableHeader.entriesPerTable -
		    (storeEntryOffset % tfxPage->annotTableHeader.entriesPerTable)),
		    IO_INTERNAL_ERROR);
		TFX_SEEK(tfxFile, prevTableFileOffset + sizeof(short) +
		    (storeEntryOffset % tfxPage->annotTableHeader.entriesPerTable)
		    * tfxPage->annotTableHeader.bytesPerEntry);
		TFX_WRITE(tfxFile, entryData, bp - entryData);
		
		/* If no more tables follow, break out of loop */
		if (al == NULL) {
		    break;
		}
		/* Prepare for next table */
		shiftEntryCount =
		    tfxPage->annotTableHeader.entriesPerTable - 1;
		dstEp = al->entries;
		srcEp = dstEp + 1;
		bp = entryData;
	    }
	    
	    /* Make sure we're gunna store no more than we have */
	    if (shiftEntryCount > annotCount - annotOffset) {
		if (annotOffset > annotCount) {
		    shiftEntryCount = 0;
		} else {
		    shiftEntryCount = annotCount - annotOffset;
		}
	    }
	    
	    /* Shift the entries */
	    memcpy(dstEp, srcEp, shiftEntryCount * sizeof(TFX_ANNOTTABLEENTRY));
	    
	    /* Remember the location of the first item that shifted */
	    storeEp = dstEp;
	    storeEntryCount = shiftEntryCount;
	    storeEntryOffset = annotOffset + entryOffset;
	    
	    /* Rember where to store the first entry of the next table */
	    dstEp += shiftEntryCount;
	    
	    /* Remember this as previous table, then go to next */
	    prevTableFileOffset = thisTableFileOffset;
	    thisTableFileOffset = al->nextTableOffset;
	    prevTbl = al;
	    al = al->next;
	    annotOffset += tfxPage->annotTableHeader.entriesPerTable;
	    entryOffset = 0;
	    if (al) {
		srcEp = al->entries;
	    }
        } /* for each table */
	    
	/* Free the temporary buffer */
	TFX_FREE(entryData);
	entryData = NULL;
    }

    /* Update the total */
    tfxPage->annotTableHeader.annotCount--;
    tfxPage->pageInfo.wAnnotationCount--;
    XERR_PASS_CLEANUP(wp_TfxWriteAnnotTableHeader(tfxPage));

    /* Return the new count */
    *pdwAnnotationCount = tfxPage->annotTableHeader.annotCount;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (entryData) {
	TFX_FREE(entryData);
	entryData = NULL;
    }
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Update the nextIFD and page table entry necessary to attach a new IFD.
 *
 * LongDescription:::
 *  wp_TfxSetPageIFD() makes the updates necessary to place a new IFD in the
 *  primary chains of IFDs for a new page. This includes updating the page
 *  table entry, and adding a new page table if necessary; and also, if
 *  necessary, relinking the "nextIFD" fields of the previous IFD, or updating
 *  the "nextIFD" in the TIFF header.
 *  <BR>
 *  Note that "if necessary" means that the update occurs only if the data
 *  has changed.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  pageOffset::		Offset of this page
 *  flags::			Page table flags (currently unused)
 *  ifdOffset::			Offset of primary IFD
 *  tfxPage::			Page being added (only used if page is
 *				attached to a new location, which includes
 *				actually being added)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *   If just attempting to add a page table during the initial write, the last
 *   four parameters can all be zero.
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxSetPageIFD(
	      TFX_DOC *tfxDoc,
	      UInt32 pageOffset,
	      UInt8 flags,
	      UInt32 ifdOffset,
	      TFX_PAGE *tfxPage
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
/*    UInt32 tableOffset; */
/*    UInt32 entryOffset; */
/*    TFX_PAGETABLE *prevPt, *pt; */
/*    TFX_PAGETABLE *newPt = NULL; */
/*    TFX_PAGETABLEENTRY *tableEntry; */
/*    Int32 ptOffset, prevPtOffset; */
/*    UInt32 mallocbytes; */
/*    Int32 fileOffset; */
#if 0
    UInt16 sval; UInt32 lval;
    UInt8 rewriteHeader = 0;
#endif
/*    UInt8 prevPageFlags; */
    Int32 prevPageIFDOffset = 0;
    TFX_IMG *tfxImg = NULL;
    TFX_IFDOFFSET_ITEM *ifdp = NULL, *newIFDp = NULL;
    PROLOGUE;
    
    /* Check parameters / mode */
    TFX_CheckHandle(tfxDoc);
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxDoc->seqStatus & TFX_DOC_SEQ_WRITESTART) == 0),
		     IO_SEQUENCE);

    /* We can only add the next page, not allow a gap in the pages */
    XERR_GEN_CLEANUP((Int32)pageOffset > tfxDoc->docInfo.wNPages - 1,
		     IO_BADPARAMETER);
    XERR_GEN_CLEANUP((tfxDoc->readWrite & WRITE_ALLOWED) == 0,
		     IO_INTERNAL_ERROR);

#if 1
/*    if (tfxDoc->readWrite & READ_ALLOWED) { */
    if (tfxPage && tfxPage->pageInfo.dwPageOffset > 0) {

        /* Make sure we've read all the page tables if the destination is
	 * a read-write file */
	UInt8 flags;

	/* Get the IFD of the previous page (offset = 1 less than this page,
	 * which has page offset wNPages - 1) */
	XERR_PASS_CLEANUP(wp_TfxGetPageIFD(tfxDoc,

	    tfxPage->pageInfo.dwPageOffset - 1, &flags, &prevPageIFDOffset));

	/* Well here's an interesting one.  What if we call AddPageStart twice
	 * in a row, without having added the IFD for the first one?
	 * The one time I encountered this, was when writing a binary profile,
	 * and there was a contone-only image in the middle of other pages.
	 * The correct thing to do is: catch the fact that there's a
	 * contone-only page BEFORE calling AddPageStart for that page. */
	XERR_GEN_CLEANUP(tfxDoc->docInfo.wNPages > 0 && prevPageIFDOffset == 0,
	    IO_SEQUENCE);
    }

    if (tfxPage != NULL) {

        /* Now we must have the page handle, so we can attach the nextIFD
	 * field of the previous page */
	TFX_CheckHandle(tfxPage);
	tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;
	XERR_PASS_CLEANUP(wp_TifFXWriteNextIFD(tfxImg, NULL, prevPageIFDOffset,
	    ifdOffset));

	/* Allocate a new item */
	IO_CHECKMALLOC(newIFDp = (TFX_IFDOFFSET_ITEM *)
	    TFX_MALLOC(sizeof(*ifdp->next)));
	memset(newIFDp, 0, sizeof(*newIFDp));

	/* Populate it */
	newIFDp->ifdOffset = ifdOffset;

	/* Find the spot in the list where it goes *after* */
	{
	    UInt32 listOffset = 0UL;

	    ifdp = tfxDoc->ifdList;
	    listOffset = 0;
	    while (ifdp && pageOffset && listOffset < pageOffset - 1) {
		ifdp = ifdp->next;
		listOffset++;
	    }
	}

	/* Add the new item to the list */
	if (ifdp) {
	    if (ifdp == tfxDoc->ifdList) {
		newIFDp->next = tfxDoc->ifdList->next;
		tfxDoc->ifdList->next = newIFDp;
	    } else {
		newIFDp->next = ifdp->next;
		ifdp->next = newIFDp;
	    }
	} else {
	    XERR_GEN_CLEANUP(tfxDoc->ifdList != NULL, IO_INTERNAL_ERROR);
	    tfxDoc->ifdList = newIFDp;
	}
	ifdp = newIFDp;
	newIFDp = NULL;	/* Avoid freeing */

    } else {
	/* Special case: last page was just deleted, remove the first IFD
	 * offset from the TIFF header */
	XERR_PASS_CLEANUP(wp_TifFXWriteNextIFD(tfxImg, NULL, prevPageIFDOffset,
	    ifdOffset));
    }
#else
    /* Update the appropriate page table */
    if (tfxDoc->pageTableHeader.entriesPerTable == 0UL) {
	tfxDoc->pageTableHeader.entriesPerTable = 32;
    }
    XERR_GEN_CLEANUP(tfxDoc->pageTableHeader.entriesPerTable == 0,
	IO_INTERNAL_ERROR);
    tableOffset = pageOffset / tfxDoc->pageTableHeader.entriesPerTable;
    entryOffset = pageOffset % tfxDoc->pageTableHeader.entriesPerTable;

    /* Search for the table itself */
    prevPt = pt = tfxDoc->firstPageTable;
    prevPtOffset = 0UL;
    ptOffset = tfxDoc->pageTableHeader.firstTableOffset;
    while (pt) {
	if (pt->ordinal == tableOffset + 1) {
	    break;
	}
	prevPtOffset = ptOffset;
	ptOffset = pt->nextTableOffset;	
	prevPt = pt;
	pt = pt->next;
    }

    /* Create a new table if the one desired doesn't exist yet */
    if (pt == NULL) {

	IO_CHECKMALLOC(newPt = (TFX_PAGETABLE *)TFX_MALLOC(sizeof(TFX_PAGETABLE)));
	memset(newPt, 0, sizeof(TFX_PAGETABLE));
	mallocbytes = sizeof(TFX_PAGETABLEENTRY) * 
	    tfxDoc->pageTableHeader.entriesPerTable;
	IO_CHECKMALLOC(newPt->entries =
		     (TFX_PAGETABLEENTRY *)TFX_MALLOC(mallocbytes));
	memset(newPt->entries, 0, mallocbytes);
	newPt->ordinal = (UInt16)tableOffset + 1;

	/* Write the new page table */
	XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &ptOffset));
	sval = newPt->ordinal;
	TFX_SWAPSHORT(tfxFile, &sval);
	TFX_WRITE(tfxFile, &sval, sizeof(sval));
	TFX_WRITE(tfxFile, newPt->entries, tfxDoc->pageTableHeader.entriesPerTable
	    * (sizeof(tableEntry->flags) + sizeof(tableEntry->IFDoffset)));
	lval = newPt->nextTableOffset;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
	
	/* Update the status to reflect the first page table written */
	tfxDoc->seqStatus |= TFX_DOC_SEQ_W_PAGETABLE;

	if (prevPt) {
	    /* Attach to end of linked list */
	    prevPt->next = newPt;

	    /* Look for the page table 'nextTableOffset' location of the table
	     * were attaching this one to */
	    fileOffset = prevPtOffset + sizeof(tfxDoc->firstPageTable->ordinal) +
		tfxDoc->pageTableHeader.entriesPerTable *
		(sizeof(tableEntry->flags) + sizeof(tableEntry->IFDoffset));
	    XERR_GEN_CLEANUP(fileOffset < 8, IO_INTERNAL_ERROR);

	    /* Link the new table to the previous one, in the file */
	    TFX_SEEK(tfxFile, fileOffset);
	    lval = ptOffset;
	    TFX_SWAPLONG(tfxFile, &lval);
	    TFX_WRITE(tfxFile, &lval, sizeof(lval));
	    prevPt->nextTableOffset = ptOffset;
	} else {
	    /* Begin list here */
	    tfxDoc->firstPageTable = newPt;
	    tfxDoc->pageTableHeader.firstTableOffset = ptOffset;
	}
	pt = newPt;
	newPt = NULL;	/* Avoid freeing if later failure */

	/* Update the header */
	tfxDoc->pageTableHeader.lastOrdinal++;
	rewriteHeader = 1;
    } /* Adding a new table */

    /* Point to the right entry */
    tableEntry = &pt->entries[entryOffset];

    if (pageOffset == tfxDoc->pageTableHeader.pageCount && ifdOffset) {
	tfxDoc->pageTableHeader.pageCount = (UInt16)(pageOffset + 1);
	rewriteHeader = 1;
    }

    if (rewriteHeader) {
	XERR_PASS_CLEANUP(wp_TfxWritePageTableHeader(tfxDoc));
    }

    /* Write the new IFD offset to the page table entry, and if appropriate,
     * to the previous IFD (or TIFF header). */
    if (ifdOffset != pt->entries[entryOffset].IFDoffset) {

	/* Write to the previous IFD */
	if (pageOffset > 0) {

	    /* Get the IFD of the previous page */
	    XERR_PASS_CLEANUP(wp_TfxGetPageIFD(tfxDoc, pageOffset - 1,
		&prevPageFlags, &prevPageIFDOffset));
	} else {
	    prevPageIFDOffset = 0UL;
	}

	if (tfxPage != NULL) {

	    /* Now we must have the page handle, so we can attach the nextIFD
	     * field of the previous page */
	    TFX_CheckHandle(tfxPage);
	    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;
	    XERR_PASS_CLEANUP(wp_TifFXWriteNextIFD(tfxImg, NULL, prevPageIFDOffset,
		    ifdOffset));

	    /* Now write the page table entry */
	    XERR_GEN_CLEANUP(ptOffset == 0, IO_INTERNAL_ERROR);
	    fileOffset = ptOffset + sizeof(tfxDoc->firstPageTable->ordinal) +
		entryOffset * (sizeof(tableEntry->flags) +
		sizeof(tableEntry->IFDoffset));
	    XERR_GEN_CLEANUP(fileOffset < 8, IO_INTERNAL_ERROR);
	    tableEntry->flags = flags;
	    tableEntry->IFDoffset = ifdOffset;
	    XERR_PASS_CLEANUP(wp_TfxWritePageTableEntry(tfxDoc, fileOffset,
		tableEntry));
	} else {
	    /* Special case: last page was just deleted, remove the first IFD
	     * offset from the TIFF header */
	    XERR_PASS_CLEANUP(wp_TifFXWriteNextIFD(tfxImg, NULL, prevPageIFDOffset,
		    ifdOffset));
	}
    }
#endif
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (newIFDp) {
	TFX_FREE(newIFDp);
    }
#if 0
    if (newPt) {
	if (newPt->entries) {
	    TFX_FREE(newPt->entries);
	}
	TFX_FREE(newPt);
    }
#endif
/* Avoid compiler warning */
    flags = flags;
    
    ERR_RETURN;
}


#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Generate the compatible page table flags for a page
 *
 * LongDescription:::
 *  wp_TfxGeneratePageFlags() generates the page table flags associated with
 *  a page.  Currently, the flags are not used, but maybe some day?
 *
 * Input:::
 *  tfxPage::			Data structure of valid page
 *
 * Output:::
 *  pFlags::			Pointer to 8-bit flags
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxGeneratePageFlags(
		TFX_PAGE *tfxPage,
		UInt8 *pFlags
)
{
    PROLOGUE;
    
    TFX_CheckHandle(tfxPage);

    /* Currently, the flags aren't used (yet) */
    *pFlags = 0;
    
    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write or rewrite the page table header
 *
 * LongDescription:::
 *  wp_TfxWritePageTableHeader() writes or rewrites the page table header,
 *  usually to accomodate the addition of a page to update the page count.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxWritePageTableHeader(
		TFX_DOC *tfxDoc
)
{
    TFX_FILE *tfxFile;
    UInt16 sval; UInt32 lval;
    PROLOGUE;
    
    TFX_CheckHandle(tfxDoc);
    
    /* Check for proper calling sequence */

    tfxFile = TFX_DOCFILE(tfxDoc);

    /* If this is the first time it is written, place at the end of file */
    if (tfxDoc->pageTableHeaderOffset == 0UL) {
	TFX_DOCFIELD docField;

	XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile,
	    &tfxDoc->pageTableHeaderOffset));
	docField.tagID = DOCTAG_PAGETABLE;
	docField.offset = tfxDoc->pageTableHeaderOffset;
	TFX_SEEK(tfxFile, tfxDoc->tagOffset[DOCTAG_PAGETABLE]);
	XERR_GEN_CLEANUP(tfxDoc->tagOffset[DOCTAG_PAGETABLE] == 0,
	    IO_INTERNAL_ERROR);
	XERR_PASS_CLEANUP(wp_TfxWriteExtension(tfxDoc, &docField));
    }
    TFX_SEEK(tfxFile, tfxDoc->pageTableHeaderOffset);
    sval = tfxDoc->pageTableHeader.pageCount;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    sval = tfxDoc->pageTableHeader.entriesPerTable;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    sval = tfxDoc->pageTableHeader.lastOrdinal;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    lval = tfxDoc->pageTableHeader.firstTableOffset;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the shared resource characteristics for an individual dictionary.
 *   
 * LongDescription:::
 *  wp_TfxSetResourceOffset() sets the characteristics for an individual
 *  resource (token dictionary) in the list of resources.  Then returns the ID
 *  assigned to that resource.
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  bytes::			Number of bytes stored
 *  offset::			Offset in the file of the dictionary
 *  compression::		Compression type used for the data
 *  lastPage::			Last page (ordinal) that uses this resource
 *				(0 if unknown)
 *  decMem::			Decoder memory measure to hold resource
 *				in some uncompressed form
 *
 * Output:::
 *  pdwResourceOffset::		Offset of new resource
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxSetResourceOffset(
	      TFX_DOC *tfxDoc,
	      UInt32 bytes,
	      UInt32 offset,
	      UInt16 compression,
	      UInt16 lastPageUsed,
	      UInt32 decMem,
	      UInt32 *pdwResourceOffset
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt32 entryOffset = 0UL;
    TFX_RESOURCETABLE *prevRt, *rt;
    TFX_RESOURCETABLE *newRt = NULL;
    TFX_RESOURCETABLEENTRY *tableEntry, *ep;
    Int32 rtOffset, prevRtOffset;
    UInt32 mallocbytes;
    Int32 fileOffset;
    UInt16 sval; UInt32 lval;
    UInt32 resourceOffset = 0;
    PROLOGUE;
    
    /* Check parameters / mode */
    TFX_CheckHandle(tfxDoc);
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxDoc->seqStatus & TFX_DOC_SEQ_WRITESTART) == 0),
		     IO_SEQUENCE);

    /* We should be able to write */
    XERR_GEN_CLEANUP((tfxDoc->readWrite & WRITE_ALLOWED) == 0,
		     IO_INTERNAL_ERROR);
    
    /* If we haven't yet read the last resource table, make sure we read it */
    if (tfxDoc->readWrite & READ_ALLOWED) {
	Int32 offset, bytes;
	
	/* Pick a large number, and ask for that resource (force reading until
	 * we realize the resource doesn't exist */
	XERR_IGNORE(wp_TfxGetResourceOffset(tfxDoc, 65535, &offset, &bytes,
	    &compression, &lastPageUsed, &decMem));
    }

    /* Search for the table with a free resource entry (0-bytes) */
    prevRt = rt = tfxDoc->firstResourceTable;
    prevRtOffset = 0UL;
    rtOffset = tfxDoc->firstResourceTableLocation;
    while (rt) {
	for (ep = rt->entries, entryOffset = 0; entryOffset < rt->physicalEntries; ep++, entryOffset++) {
	    if (ep->bytes == 0) {
		/* Found an available entry */
		resourceOffset = entryOffset + rt->firstID;
		break;
	    }
	}
	if (entryOffset < rt->physicalEntries) {
	    /* Break out of while loop also */
	    break;
	}
	prevRtOffset = rtOffset;
	rtOffset = rt->nextTableOffset;	
	prevRt = rt;
	rt = rt->next;
    }

    /* Create a new table if the one desired doesn't exist yet */
    if (rt == NULL) {

	/* Allocate the table */
	IO_CHECKMALLOC(newRt = (TFX_RESOURCETABLE *)TFX_MALLOC(sizeof(TFX_RESOURCETABLE)));
	memset(newRt, 0, sizeof(TFX_RESOURCETABLE));

	if (prevRt) {
	    resourceOffset = newRt->firstID = prevRt->firstID + prevRt->physicalEntries;
	} else {
	    newRt->firstID = 0;
	}

	/* Make sure we have a reasonable minimum */
	if (tfxDoc->resourcesPerTable < 1) {
	    tfxDoc->resourcesPerTable = 1;
	}
	newRt->physicalEntries = tfxDoc->resourcesPerTable;

	/* Allocate the entries for the table */
	mallocbytes = sizeof(TFX_RESOURCETABLEENTRY) * newRt->physicalEntries;
	IO_CHECKMALLOC(newRt->entries =
		     (TFX_RESOURCETABLEENTRY *)TFX_MALLOC(mallocbytes));
	memset(newRt->entries, 0, mallocbytes);

	/* Write the new resource table */
	XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &rtOffset));
	sval = newRt->physicalEntries;
	TFX_SWAPSHORT(tfxFile, &sval);
	TFX_WRITE(tfxFile, &sval, sizeof(sval));
	TFX_WRITE(tfxFile, newRt->entries,
	    (Int32)(newRt->physicalEntries *
	    (sizeof(tableEntry->bytes) + sizeof(tableEntry->offset)
	    + sizeof(tableEntry->resourceType)
	    + sizeof(tableEntry->lastPage) + sizeof(tableEntry->decMem))));
	lval = newRt->nextTableOffset;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));

	/* Update the status to reflect the page table header written */
	tfxDoc->seqStatus |= TFX_DOC_SEQ_W_RESOURCETABLE;

	if (prevRt) {
	    /* Attach to end of linked list */
	    prevRt->next = newRt;

	    /* Look for the resource table 'nextTableOffset' location of the
	     * table were attaching this one to */
	    fileOffset = prevRtOffset + sizeof(prevRt->physicalEntries)
		+ prevRt->physicalEntries *
		(sizeof(tableEntry->bytes) + sizeof(tableEntry->offset)
		 + sizeof(tableEntry->resourceType)
		 + sizeof(tableEntry->lastPage) + sizeof(tableEntry->decMem));
	    XERR_GEN_CLEANUP(fileOffset < 8, IO_INTERNAL_ERROR);

	    /* Link the new table to the previous one, in the file */
	    TFX_SEEK(tfxFile, fileOffset);
	    lval = rtOffset;
	    TFX_SWAPLONG(tfxFile, &lval);
	    TFX_WRITE(tfxFile, &lval, sizeof(lval));
	    prevRt->nextTableOffset = rtOffset;
	} else {
	    /* Begin list here */
	    tfxDoc->firstResourceTable = newRt;
	    tfxDoc->firstResourceTableLocation = rtOffset;

	    /* Write the location to the SharedResources field of the
	     * GlobalParametersIFD */
	    XERR_GEN_CLEANUP(tfxDoc->sharedResourcesLocation < 8, IO_INTERNAL_ERROR);
	    TFX_SEEK(tfxFile, tfxDoc->sharedResourcesLocation);
	    lval = rtOffset;
	    TFX_SWAPLONG(tfxFile, &lval);
	    TFX_WRITE(tfxFile, &lval, sizeof(lval));
	}
	rt = newRt;
	newRt = NULL;	/* Avoid freeing if later failure */

	/* Use first entry in new table, for this one */
	entryOffset = 0;
    } /* Adding a new table */

    /* Point to the entry */
    tableEntry = &rt->entries[entryOffset];

/*    if (resourceOffset == tfxDoc->resourceTableHeader.resourceCount) { */
/*	tfxDoc->resourceTableHeader.resourceCount = (UInt16)(resourceOffset + 1); */
/*    } */

    /* Write the new entry in the resource table */
    if (offset) {
	XERR_GEN_CLEANUP(rtOffset == 0, IO_INTERNAL_ERROR);
	fileOffset = rtOffset + sizeof(sval)
	    + entryOffset * (sizeof(tableEntry->bytes)
	    + sizeof(tableEntry->offset)
	    + sizeof(tableEntry->resourceType)
	    + sizeof(tableEntry->lastPage)
	    + sizeof(tableEntry->decMem));
	XERR_GEN_CLEANUP(fileOffset < 8, IO_INTERNAL_ERROR);
	tableEntry->bytes = bytes;
	tableEntry->offset = offset;
	tableEntry->resourceType = TFX_RESOURCE_JBIG2;
	tableEntry->lastPage = 0;   /* Set to 0 for now, until document closed */
/* XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX Need to work this in */
	tableEntry->decMem = 0;	    /* Decoder memory required to hold the dictionary */
	XERR_PASS_CLEANUP(wp_TfxWriteResourceTableEntry(tfxDoc, fileOffset,
	    tableEntry));
    }

    /* Return the resulting ID */
    *pdwResourceOffset = resourceOffset;

    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (newRt) {
	if (newRt->entries) {
	    TFX_FREE(newRt->entries);
	}
	TFX_FREE(newRt);
    }
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a resource table entry.
 *   
 * LongDescription:::
 *  wp_TfxWriteResourceTableEntry() stores a single resource entry for a
 *  resource (token dictionary).
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *  fileOffset::		Where to write the table entry
 *  resourceTableEntry::	Table entry to write
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxWriteResourceTableEntry(
    TFX_DOC *tfxDoc,
    Int32 fileOffset,
    TFX_RESOURCETABLEENTRY *resourceTableEntry
)
{
    TFX_FILE *tfxFile = TFX_DOCFILE(tfxDoc);
    TFX_RESOURCETABLEENTRY tempEntry;
    PROLOGUE;

    /* Seek to the entry location */
    TFX_SEEK(tfxFile, fileOffset);

    /* Copy to a temporary entry (for swapping) */
    memcpy(&tempEntry, resourceTableEntry, sizeof(tempEntry));

    /* Swap if necessary */
    TFX_SWAPLONG(tfxFile, &tempEntry.bytes);
    TFX_SWAPLONG(tfxFile, &tempEntry.offset);
    TFX_SWAPSHORT(tfxFile, &tempEntry.resourceType);
    TFX_SWAPSHORT(tfxFile, &tempEntry.lastPage);
    TFX_SWAPLONG(tfxFile, &tempEntry.decMem);

    /* Write this puppy out */
    TFX_WRITE(tfxFile, &tempEntry.bytes, sizeof(tempEntry.bytes));
    TFX_WRITE(tfxFile, &tempEntry.offset, sizeof(tempEntry.offset));
    TFX_WRITE(tfxFile, &tempEntry.resourceType, sizeof(tempEntry.resourceType));
    TFX_WRITE(tfxFile, &tempEntry.lastPage, sizeof(tempEntry.lastPage));
    TFX_WRITE(tfxFile, &tempEntry.decMem, sizeof(tempEntry.decMem));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


#if 0	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write/rewrite the resource table header
 *   
 * LongDescription:::
 *  wp_TfxWriteResourceTableHeader() writes or rewrites the shared resource
 *  header to take account of an update in the information (usually the
 *  resource count).
 *
 * Input:::
 *  tfxDoc::			Data structure of valid document
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxWriteResourceTableHeader(
    TFX_DOC *tfxDoc
)
{
    TFX_FILE *tfxFile;
    UInt16 sval; UInt32 lval;
    PROLOGUE;
    
    TFX_CheckHandle(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(tfxDoc->firstSharedResourceLocation == 0, IO_SEQUENCE);

    /* If this is the first time it is written, place at the end of file */
    if (tfxDoc->resourceTableHeaderOffset == 0UL) {
	XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile,
	    &tfxDoc->resourceTableHeaderOffset));

	/* Write the offset to the location of the offset */
	TFX_SEEK(tfxFile, tfxDoc->firstSharedResourceLocation);
	lval = tfxDoc->resourceTableHeaderOffset;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
    }
    TFX_SEEK(tfxFile, tfxDoc->resourceTableHeaderOffset);
    sval = tfxDoc->resourceTableHeader.resourceCount;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    sval = tfxDoc->resourceTableHeader.entriesPerTable;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    sval = tfxDoc->resourceTableHeader.lastOrdinal;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    lval = tfxDoc->resourceTableHeader.firstTableOffset;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}
#endif


#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a temporary IFD for filling in later
 *
 * LongDescription:::
 *  wp_TfxQueueIFD() begins the mode for constructing an IFD.  The IFD is
 *  not resolved until the image compression and construct is known, which
 *  doesn't take place until either: the first strip of data is added, or
 *  the last strip of data is complete (in the case of variable-sized strips
 *  -- StripRowCounts is used).
 *  While in queue mode, any non-IFD data that is written, is instead placed
 *  in a queue, and later written when the IFD is complete.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxQueueIFD(
		TFX_IMG *tfxImg
)
{
    TFX_FILE *tfxFile;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(tfxFile->bQueuing != 0, IO_INTERNAL_ERROR);

    /* Begin queuing */
    tfxFile->bQueuing = 1;
    
    /* Fall through to free up any temporary memory */

 ERR_CLEANUP:
    
    NORMAL_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a temporary IFD for filling in later
 *
 * LongDescription:::
 *  wp_TfxWriteMockIFD() makes space for the IFD fields.  Since the IFD is
 *  constructed of both fields and field data, this allows us to place the
 *  IFD fields before the field data.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  fieldCount::		Number of fields to make room for
 *  fileOffset::		File offset where IFD was written
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxWriteMockIFD(
		TFX_IMG *tfxImg,
		UInt32 fieldCount,
		UInt32 *fileOffset
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt8 *zeros = NULL;
    UInt32 mallocSize;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);
    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);

    /* Return a bad file offset for now */
    *fileOffset = 0UL;

    /* Use a temporary buffer with zeros as a place-holder for the IFD */
    mallocSize = 2 + fieldCount * 12 /* tag size */ + 4;
    IO_CHECKMALLOC(zeros = TFX_MALLOC(mallocSize));
    memset(zeros, 0, mallocSize);

    /* Place at the end of the file */
    XERR_PASS(wp_TfxFileLongAlign(tfxFile, (Int32 *)fileOffset));
    TFX_WRITE(tfxFile, zeros, (Int32)mallocSize);

    /* Fall through to free up any temporary memory */

 ERR_CLEANUP:
    
    if (zeros) {
	TFX_FREE(zeros);
    }
    
    NORMAL_RETURN;
}

/*
$Log: tfxwtfx.c,v $
Revision 1.3  2003/02/11 01:31:43  mkrocken
fixed problems with comments caused by previous commit

Revision 1.2  2003/02/10 22:53:33  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:14  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.17  2001/12/26 21:23:46  dabercro
Mark as released

Revision 1.16  2001/10/08 20:59:57  dabercro
Rewrite Annotations field, if annotation added, and no field exists.

Revision 1.15  2001/03/13 02:20:06  dabercro
Avoid compiler warning.

Revision 1.14  2001/01/18 00:21:32  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.13  2000/11/17 23:57:36  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.12  2000/09/21 21:19:15  dabercro
Add variable-sized strips: StripRowCounts.

Revision 1.11  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.10  2000/08/25 23:55:38  dabercro
Logic fixes for IFD tracking, and page numbers/offsets.

Revision 1.9  2000/08/22 01:03:24  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.8  2000/08/04 22:53:08  dabercro
Only check preexisting nextIFD of read/write.

Revision 1.7  2000/08/04 21:57:09  dabercro
Improve IFD linked list, since we aren't using page tables.

Revision 1.6  2000/08/02 22:43:48  dabercro
Check the calling sequence to make sure the previous page
has been started, before we expect the "nextIFD" structure to
be in place for a new page.

Revision 1.5  2000/08/01 00:11:49  dabercro
Updates to the handling of shared resources.

Revision 1.4  2000/06/07 00:00:17  dabercro
Continued work, lots of stuff.

Revision 1.3  2000/05/09 01:52:58  dabercro
Added FillOrder field support.
Updated JBIG2 dictionary front-end.
Support for JBIG2 dictionary extensions, updated parameter to
reflect byte count for extensions.
Write fields that are explicitly set, regardless of default values.
Check if document is single-page when closing, reflect in
NewSubfileType.

Revision 1.2  2000/04/12 19:06:10  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.1.1.1  2000/02/02 00:40:52  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.36  1998/05/07 20:01:41  aber
aber on Thu May 7 13:00:28 PDT 1998

tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Just to be real sure, I'm gunna do the same thing for the annotation
tables also; although it looks like it doesn't impact Pagis.

Revision 1.35  1998/05/07 19:36:01  aber
aber on Thu May 7 12:35:34 PDT 1998

tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

I found the same problem with the page tables.

Revision 1.34  1998/05/07 17:39:27  aber
aber on Thu May 7 10:38:29 PDT 1998

tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

If checking to read all symlib tables on a read-write file, make sure
there even is any. This causes a crash if there isn't any.

Revision 1.33  1998/05/06 23:10:48  aber
aber on Wed May 6 16:09:58 PDT 1998

tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Make sure we've read the N-th symbol library table, before we append
the N+1th symbol dictionary.

Revision 1.32  1998/04/30 21:08:59  aber
aber on Thu Apr 30 14:07:19 PDT 1998

tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Fixed write to AnnotationOffsets field when storing a single annotation
to a TIFF-FX 2.0 file. This hasn't causes a problem before (and won't in the
future) because Pagis has always written 2 annotations. However, this
is the right thing to do.

Revision 1.31  1998/03/28 01:25:22  aber
aber on Fri Mar 27 17:24:49 PST 1998

tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

I checked in the wrong file last time. This one errors if adding an
annotation to a page with no primary image.

Revision 1.30  1998/01/22 00:23:55  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

CommonCheck: Allow a 0-byte author block.

Revision 1.29  1998/01/20 03:44:03  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Prototype update to allow deleting last page of a document.

Revision 1.28  1998/01/20 03:27:44  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Allow deleting the last page of a single-page document.
Update the symbol library location and size during write, so that we
can read the data without having to read the table header.

Revision 1.27  1998/01/16 00:45:00  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Allow page inserts.
New private function used to generate the page table flage (in one
common place). Currently no page table flags are used.

Revision 1.26  1997/12/18 00:25:28  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Remove unreferenced function.

Revision 1.25  1997/11/19 22:30:47  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

CommonCheck: Used right error code.

Revision 1.24  1997/11/19 21:43:45  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Allow deleting all annotations, but retaining the annotation
table header.

Revision 1.23  1997/11/19 20:25:46  aber
aber on Wed Nov 19 12:25:21 PST 1997

tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Allow annotation deletion with either a call to TFX_AddPageStart(),
or TFX_GetPageInfo().

Revision 1.22  1997/11/18 23:48:27  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Add annotation flags entry to annotation functions.
Add bytesPerEntry field to annotation header.
Store only the annotation table entries that match the existing
annotation table header byte count (ignore the rest).

Revision 1.21  1997/11/14 02:09:07  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

CommonCheck: Use signed value for reference page; clarify that it's a page offset,
not an ordinal. Store as ordinal (was only reading as ordinal).

Revision 1.20  1997/10/15 21:37:53  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Update SetPageIFD function to handle adding a page table without
actually adding a page.

Revision 1.19  1997/09/19 00:19:15  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Update page info for annotation count when adding or deleting one.

Revision 1.18  1997/08/29 17:28:09  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.17  1997/08/27 21:55:33  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

When attaching another page table or symlib table, keep track of it,
otherwise we start clobbering data elsewhere in the file.

Revision 1.16  1997/08/27 00:18:21  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Correct the symlib count check.

Revision 1.15  1997/08/16 00:16:33  aber
aber on Fri Aug 15 17:16:23 PDT 1997

tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Update documentation to match ID removal.

Revision 1.14  1997/08/15 23:48:04  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

CommonCheck: Collective decision by Dan Huttenlocher, William Rucklidge, and I
to remove the token dictionary ID from the symbol library table
entry, since the offset is typically used, and the ID does nothing
for us.

Revision 1.13  1997/08/15 00:11:56  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Slight code cleanup.

Revision 1.12  1997/07/14 23:24:20  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Complete annotation deletion.

Revision 1.11  1997/07/10 22:10:43  aber
aber on Thu Jul 10 15:09:25 PDT 1997

tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Remove unused variable. Initial 2 variables to NULL, since the compiler
complains they might be used before set, although it isn't really smart
enough to realize that they can't be used the first time through the
loop. Isn't it just only so clever!

Revision 1.10  1997/07/10 22:03:56  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Initial annotation deletion code. Use 16-bit ID.

Revision 1.9  1997/06/17 22:33:45  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

Fix TIFF-FX 3.0 annotation writes, was clobbering the TIFF-FX extension fields.

Revision 1.7  1997/06/10 01:27:07  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.6  1997/05/30 00:51:25  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.5  1997/05/20 22:29:30  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.4  1997/05/20 00:54:39  aber
tfxwtfx.c -> /product/ipcore/ipwrap/src/RCS/tfxwtfx.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:52  aber
Initial revision

*/
