/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxwtiff.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxwtiff.c:	WRITE TIFF source
 *
 *	TIFF write-only routines.
 *
 */


#include <stdio.h>
#include <math.h>
#include <string.h>

#ifndef NOIPCORE
#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tfxwtiff.c,v $; $Revision: 1.3 $; $Date: 2003/02/11 01:31:43 $");

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif
#endif	/* NOIPCORE */

#include "tfximage.h"
#include "tfx.h"
#include "tfxwpriv.h"
#include "tfxwtiff.h"
#include "tfxwtfx.h"
#include "tfxwjpeg.h"
#include "tfxtiff.h"
#include "tfxrtiff.h"
#include "tfxrtfx.h"
#include "tfxrpriv.h"


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write the TIFF header.
 *
 * LongDescription:::
 *  wp_TifFXWriteHeader() writes the TIFF header.
 *
 * Input:::
 *  pTfxFile::			Pointer to TIFF-FX file handle
 *  dwFirstIfd::		Offset of first IFD
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXWriteHeader(
    TFX_FILE *pTfxFile,
    UInt32 dwFirstIfd
)
{
    TFX_TIFFHEADER header;
#ifdef _DEBUG
    UInt8 cr[1] = {10};
    Int32 fileSize;
#endif
    PROLOGUE;

#ifdef _DEBUG
    /* We're gunna have some fun first. Write out a carriage-return, and
     * see if only one byte was written */
    TFX_FILESIZE(pTfxFile, fileSize);
    XERR_GEN_CLEANUP(fileSize != 0L, IO_BADFILEFORMAT);    /* File not empty */
    TFX_SEEK(pTfxFile, 0UL);
    TFX_WRITE(pTfxFile, cr, 1);
    TFX_FILESIZE(pTfxFile, fileSize);
    XERR_GEN_CLEANUP(fileSize == 0L, IO_WRITE_ERR);
    XERR_GEN_CLEANUP(fileSize == 2L, IO_BADFILEMODE);
#endif

    /* Create the header */
    header.byteOrder = TIF_NATIVEORDER;
    header.version = 42;
    header.firstIfd = dwFirstIfd;
    
    /* Write the TIFF header */
    TFX_SEEK(pTfxFile, 0UL);
    TFX_WRITE(pTfxFile, &header.byteOrder, sizeof(header.byteOrder));
    TFX_WRITE(pTfxFile, &header.version, sizeof(header.version));
    TFX_WRITE(pTfxFile, &header.firstIfd, sizeof(header.firstIfd));
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write the GlobalParametersIFD fields
 *
 * LongDescription:::
 *  wp_TifFXWriteBPFields() writes the TIFF fields associated with the
 *  GlobalParametersIFD.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  writing::			Writing the fields, or just counting them?
 *
 * Output:::
 *  pFieldCount::		Pointer to the number of fields written
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXWriteGPFields(
    TFX_IMG *tfxImg,
    UInt16 *pFieldCount,
    Bool writing
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_IMGINFO *pTfxImgInfo = NULL;
    TFX_DOC *tfxDoc;
    UInt32 ucval;
    UInt32 lval;
    Int32 dataOffset;
    TFX_TIFFIELDS *fieldList;
    TFX_TIFGPFIELDS *gpFieldList;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);
    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    pTfxImgInfo = &tfxImg->imgInfo;
    tfxDoc = tfxImg->tfxPage->tfxDoc;
    fieldList = &tfxImg->fields;
    gpFieldList = &fieldList->gpIFD;

    /* Start with no fields */
    *pFieldCount = 0;

    /* Must be the primary image (should also be the first page) */
    XERR_GEN_CLEANUP(pTfxImgInfo->tfxImgType != IO_IMGTYPE_PRIMARY, IO_INTERNAL_ERROR);
    
#ifdef DEBUG
    printf("\tField\tID\tType\tLength\tOffset\n");
#endif

    /* NOTE:  ONCE ADOBE SANCTIONS NEW PUBLIC VALUES FOR: TIFF-FXExtensions,
     *	MultiProfiles, SharedResources, and T88Options; THESE WRITES MAY HAVE
     *  TO BE REARRANGED TO THE SAME ORDER AS THE VALUES WARRANT (TIFF VALUES
     *  MUST BE IN INCREASING ORDER).
     */

    /* Write the ProfileType field, if appropriate */
    switch (tfxDoc->docInfo.profile) {
    case TFX_PROFILE_S:
    case TFX_PROFILE_F:
    case TFX_PROFILE_M:
    case TFX_PROFILE_M1:
    case TFX_PROFILE_COH_B:
    case TFX_PROFILE_COH_M:
    case TFX_PROFILE_COH_M1:
	/* Note, the values for ProfileType are:
	 * 0 - unspecified
	 * 1 - G3 Fax
	 */
	if (writing) {
	    lval = gpFieldList->ProfileType;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, PROFILETYPE,
		IFDLONG, 1, &dataOffset, &lval, pFieldCount, 1));
	} else {
	    (*pFieldCount)++;
	}
	break;
    default:
	break;
    }

    /* Write the FaxProfile field, if appropriate */
    switch (tfxDoc->docInfo.profile) {
    case TFX_PROFILE_S:
    case TFX_PROFILE_F:
    case TFX_PROFILE_J:
    case TFX_PROFILE_C:
    case TFX_PROFILE_L:
    case TFX_PROFILE_M:
    case TFX_PROFILE_T:
    case TFX_PROFILE_M1:
    case TFX_PROFILE_COH_C:
    case TFX_PROFILE_COH_M:
    case TFX_PROFILE_COH_M1:
	if (writing) {
	    switch (tfxDoc->docInfo.profile) {
	    case TFX_PROFILE_S:
		gpFieldList->FaxProfile = 1;
		break;
	    case TFX_PROFILE_F:
		gpFieldList->FaxProfile = 2;
		break;
	    case TFX_PROFILE_J:
		gpFieldList->FaxProfile = 3;
		break;
	    case TFX_PROFILE_C:
	    case TFX_PROFILE_COH_C:
		gpFieldList->FaxProfile = 4;
		break;
	    case TFX_PROFILE_L:
		gpFieldList->FaxProfile = 5;
		break;
	    case TFX_PROFILE_M:
	    case TFX_PROFILE_COH_M:
		gpFieldList->FaxProfile = 6;
		break;
	    case TFX_PROFILE_T:
		gpFieldList->FaxProfile = 7;
		break;
	    case TFX_PROFILE_M1:
	    case TFX_PROFILE_COH_M1:
	    case TFX_PROFILE_MULTIPLE:
		gpFieldList->FaxProfile = 255;
		break;
	    default:
		gpFieldList->FaxProfile = 0;	/* Does not conform to a profile */
	    }
	    ucval = gpFieldList->FaxProfile;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, FAXPROFILE,
		IFDBYTE, 1, &dataOffset, &ucval, pFieldCount, 1));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the CodingMethods field, if appropriate */
    switch (tfxDoc->docInfo.profile) {
    case TFX_PROFILE_C:
    case TFX_PROFILE_M:
	if (writing) {
	    lval = tfxImg->fields.gpIFD.CodingMethods = tfxDoc->codingMethods;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, CODINGMETHODS,
		IFDLONG, 1, &dataOffset, &lval, pFieldCount, 1));

	    /* Remember the location */
	    tfxDoc->codingMethodsLocation = dataOffset;
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the VersionYear field, if appropriate */
    switch (tfxDoc->docInfo.profile) {
    case TFX_PROFILE_C:
    case TFX_PROFILE_L:
	if (writing) {
	    /* Current version year is 1992 for JPEG, and 1997 for JBIG */
	    gpFieldList->VersionYear[0] = 1;
	    gpFieldList->VersionYear[1] = 9;
	    gpFieldList->VersionYear[2] = 9;
	    switch (tfxDoc->docInfo.profile) {
	    case TFX_PROFILE_C:		/* JPEG: 1992 */
		gpFieldList->VersionYear[3] = 2;
		break;
	    case TFX_PROFILE_L:		/* JBIG: 1997 */
		gpFieldList->VersionYear[3] = 7;
		break;
	    }
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, VERSIONYEAR,
		IFDBYTE, 4, &dataOffset, &gpFieldList->VersionYear[0],
		pFieldCount, 1));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the ModeNumber field, if appropriate */
    switch (tfxDoc->docInfo.profile) {
    case TFX_PROFILE_M:
    case TFX_PROFILE_M1:
    case TFX_PROFILE_COH_M:
    case TFX_PROFILE_COH_M1:
	if (writing) {

	    /* Set the field value unless a valid value was set by the client
	     */
	    if (! fieldList->found.ModeNumber
		|| fieldList->gpIFD.ModeNumber == TFX_CODINGMETHOD_UNKNOWN) {

		/* 0: mode 1
		 * 1: mode 2
		 * 2: mode 3 - N layers?
		 * 3: mode 4 - shared resources, JBIG2, and colour tags
		 */
		if (fieldList->gpIFD.CodingMethods & TFX_CODINGMETHOD_JBIG2) {
		    fieldList->gpIFD.ModeNumber = 3;	/* mode 4 */
		} else {
		    fieldList->gpIFD.ModeNumber = 2;	/* mode 3 */
		}
	    }
	    ucval = fieldList->gpIFD.ModeNumber;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, MODENUMBER,
		IFDBYTE, 1, &dataOffset, &ucval, pFieldCount, 1));

	    /* Remember the location */
	    tfxDoc->modeNumberLocation = dataOffset;
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the TIFF-FXExtensions field, if appropriate */
    switch (tfxDoc->docInfo.profile) {
    case TFX_PROFILE_T:
    case TFX_PROFILE_M1:
    case TFX_PROFILE_COH_M1:
	if (writing) {
	    /* Write the field */
	    lval = 0;
	    if (tfxDoc->docInfo.profile == TFX_PROFILE_T) {
		lval |= TIFFFXEXT_EXT3;	    /* Shared Resources */
		lval |= TIFFFXEXT_EXT4;	    /* Binary JBIG2 (Profile T) */
	    }
	    if (tfxDoc->docInfo.profile == TFX_PROFILE_M1 ||
		tfxDoc->docInfo.profile == TFX_PROFILE_COH_M1) {
		lval |= TIFFFXEXT_EXT3;	    /* Shared Resources */
		lval |= TIFFFXEXT_EXT5;	    /* JBIG2 for Profile M */
	    }
	    gpFieldList->TiffFXExtensions = lval;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, TIFFFXEXTENSIONS,
		IFDLONG, 1, &dataOffset, &lval, pFieldCount, 1));

	    /* Store the location of the TIFF-FX Extensions data offset */
	    tfxDoc->tifffxExtensionsLocation = dataOffset;
	} else {
	    (*pFieldCount)++;
	}

	break;
    }

    /* Write the MultiProfile, if appropriate */
    switch (tfxDoc->docInfo.profile) {
    case TFX_PROFILE_T:
    case TFX_PROFILE_M1:
    case TFX_PROFILE_COH_M1:
    case TFX_PROFILE_MULTIPLE:
	if (writing) {
	    /* If not specifically set, assume certain behavior */
	    if (tfxDoc->multiProfiles == 0) {
		switch (tfxDoc->docInfo.profile) {
		case TFX_PROFILE_T:
		    tfxDoc->multiProfiles |= TFX_MULTIPROFILE_T;
		    tfxDoc->multiProfiles |= TFX_MULTIPROFILE_E3;	/* Shared data */
		    tfxDoc->multiProfiles |= TFX_MULTIPROFILE_E5;	/* JBIG2 */
		    break;
		case TFX_PROFILE_M1:
		case TFX_PROFILE_COH_M1:
		    tfxDoc->multiProfiles |= TFX_MULTIPROFILE_M;
		    tfxDoc->multiProfiles |= TFX_MULTIPROFILE_E1;	/* Other resolutions */
		    tfxDoc->multiProfiles |= TFX_MULTIPROFILE_E2;	/* N layers */
		    tfxDoc->multiProfiles |= TFX_MULTIPROFILE_E3;	/* Shared data */
		    tfxDoc->multiProfiles |= TFX_MULTIPROFILE_E5;	/* JBIG2 */
		    break;
		}
	    }

	    /* Write the field */
	    lval = gpFieldList->MultiProfiles = tfxDoc->multiProfiles;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, MULTIPROFILES,
		IFDLONG, 1, &dataOffset, &lval, pFieldCount, 1));

	    /* Store the location of the MultiProfiles field data offset */
	    tfxDoc->multiProfilesLocation = dataOffset;
	} else {
	    (*pFieldCount)++;
	}
	break;
    default:
	break;
    }

    /* Write the SharedResources field, if appropriate */
    switch (tfxDoc->docInfo.profile) {
    case TFX_PROFILE_T:
    case TFX_PROFILE_M1:
    case TFX_PROFILE_COH_M1:
	if (writing) {
	    /* Write the field */
	    lval = gpFieldList->SharedResourcesOffset =
		tfxDoc->firstResourceTableLocation;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, SHAREDRESOURCES,
		IFDLONG, 1, &dataOffset, &lval, pFieldCount, 1));

	    /* Store the location of the shared resource header offset */
	    tfxDoc->sharedResourcesLocation = dataOffset;
	} else {
	    (*pFieldCount)++;
	}
	break;
    default:
	break;
    }

    /* Write the ReferencePage field, if appropriate */
    if (tfxDoc->seqStatus & TFX_DOC_SEQ_REFPAGESET) {
	if (writing) {
	    /* Write the field */
	    /* Stored value is really the page number */
	    lval = tfxDoc->docInfo.dwReferencePageOffset + 1;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, REFERENCEPAGE,
		IFDLONG, 1, &dataOffset, &lval, pFieldCount, 1));

	    /* Mark this as found */
	    fieldList->found.ReferencePage = 1;

	    /* Store the location of the shared resource header offset */
	    tfxDoc->refPageOffsetLocation = dataOffset;

	    /* Reflect the write in the status */
	    tfxDoc->seqStatus |= TFX_DOC_SEQ_REFPAGESET;
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Reflect that the GPIFD is there */
    tfxDoc->seqStatus |= TFX_DOC_SEQ_GPEXISTS;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write the TIFF fields
 *
 * LongDescription:::
 *  wp_TifFXWriteFields() write the TIFF fields necessary to describe the TIFF-FX file.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  writing::			Doing the write, or just counting the fields?
 *
 * Output:::
 *  pFieldCount::		Pointer to the number of fields written
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXWriteFields(
    TFX_IMG *tfxImg,
    UInt16 *pFieldCount,
    Bool writing
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_IMGINFO *pTfxImgInfo = NULL;
    UInt8 paletted = 0;
    UInt8 bpp;
    UInt8 bppe;
    UInt8 channels = 1;
    IO_COMPRESSION compression;
    UInt16 sval;
    UInt16 svals[2];
    UInt32 lval;
    UInt32 lvals[4];
    UInt32 indices;
    Int32 dataOffset;
    TFX_TIFFIELDS *fieldList;
    TFX_RATIONAL rval;
    UInt32 version;

    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);

    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    pTfxImgInfo = &tfxImg->imgInfo;
    fieldList = &tfxImg->fields;
    version = tfxImg->tfxPage->tfxDoc->docInfo.version;

    XERR_GEN_CLEANUP(pTfxImgInfo->xColorType == IO_COLORTYPE_NONE, IO_NOSUPPORT);

    /* Start with no fields */
    *pFieldCount = 0;

    /* Determine compression and color mode */
    compression = pTfxImgInfo->xCompression;
    if (pTfxImgInfo->xColorType == IO_COLORTYPE_OTHER) {
	/* If samples/pixel hasn't been explicitly set, it must be 1 */
	if (fieldList->found.SamplesPerPixel == 0) {
	    fieldList->SamplesPerPixel = 1;
	}
	XERR_GEN_CLEANUP(fieldList->found.BitsPerSample == 0, IO_SEQUENCE);
	XERR_GEN_CLEANUP(fieldList->BitsPerSample == NULL, IO_SEQUENCE);
	bppe = (UInt8)fieldList->BitsPerSample[0];
	bpp = bppe * fieldList->SamplesPerPixel;
    } else {
	bpp = IO_BPP(pTfxImgInfo->xColorType);
	bppe = IO_BPPE(pTfxImgInfo->xColorType);
    }
    XERR_GEN_CLEANUP(bppe < 1, IO_NOSUPPORT);
    channels = bpp / bppe;
    if (pTfxImgInfo->xColorType == IO_COLORTYPE_COLOR4 ||
	pTfxImgInfo->xColorType == IO_COLORTYPE_COLOR8) {
	paletted = 1;
    }
    
    /* Use default values */
    fieldList->SamplesPerPixel = 1;
    fieldList->FillOrder = 1;
    tfxImg->minBytesPerRow = (tfxImg->imgInfo.dwWidth * bppe + 7) / 8;

#ifdef DEBUG
    printf("\tField\tID\tType\tLength\tOffset\n");
#endif

    /* Write the NewSubfileType field, always */
    fieldList->NewSubfileType = 0;
    if (pTfxImgInfo->tfxImgType == IO_IMGTYPE_PRIMARY) {
	/* Assume, for now, that more than one page will be generated.
	 * If the page count is only 1 when the file is closed, go back and
	 * remove this bit. */
	/* or maybe not */
	fieldList->NewSubfileType |= TFX_TIFFSUBFILE_MULTIPAGE;
    }
    switch (tfxImg->tfxPage->tfxDoc->docInfo.profile) {
    case TFX_PROFILE_M:	/* Profile M: MRC */
    case TFX_PROFILE_M1:	/* Profile M1: MRC + JBIG2 */
    case TFX_PROFILE_COH_M:	/* Coherent TIFF Profile M: MRC */
    case TFX_PROFILE_COH_M1:	/* Coherent TIFF Profile M1: MRC + JBIG2 */
	fieldList->NewSubfileType |= TFX_TIFFSUBFILE_MRC;
	break;
    }
    if (tfxImg->imgInfo.tfxImgType == IO_IMGTYPE_HINTS) {
	fieldList->NewSubfileType |= TFX_TIFFSUBFILE_HINTS;
    }

    if (writing) {
	lval = (UInt32)fieldList->NewSubfileType;
	XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, NEWSUBFILETYPE,
	    IFDLONG, 1, &dataOffset, &lval, pFieldCount, 0));

	/* Store the location of the NewSubfileType */
	tfxImg->fieldDataLocation.NewSubfileType = dataOffset;
    } else {
	(*pFieldCount)++;
    }

    /* Write the ImageWidth field, always */
    if (writing) {
	lval = fieldList->ImageWidth = pTfxImgInfo->dwWidth;
        XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, IMAGEWIDTH, IFDLONG,
	    1, &dataOffset, &lval, pFieldCount, 0));
    } else {
	(*pFieldCount)++;
    }

    /* Write the ImageLength field, always */
    if (writing) {
	lval = fieldList->ImageLength = pTfxImgInfo->dwHeight;
        XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, IMAGELENGTH, IFDLONG,
	    1, &dataOffset, &lval, pFieldCount, 0));
    } else {
	(*pFieldCount)++;
    }

    /* Write the BitsPerSample field, always */
    if (writing) {
	if (fieldList->BitsPerSample == NULL) {
	    IO_CHECKMALLOC(fieldList->BitsPerSample =
		TFX_MALLOC(channels * TfxFieldTypeSize[IFDSHORT]));
	}
	for (sval = 0; sval < channels; sval++) {
	    fieldList->BitsPerSample[sval] = bppe;
	}
	XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, BITSPERSAMPLE, IFDSHORT,
	    channels, &dataOffset, fieldList->BitsPerSample, pFieldCount, 0));
    } else {
	(*pFieldCount)++;
    }

    /* Write the Compression field, always */
    if (writing) {
	sval = fieldList->Compression;
	XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, COMPRESSION, IFDSHORT,
	    1, &dataOffset, &sval, pFieldCount, 0));
    } else {

	/* Set the actual value here, since it may trigger the use of JPEGTables
	 * later */
	switch ((int)pTfxImgInfo->xCompression) {
	case IO_COMP_NOC:
	    fieldList->Compression = TFX_TIFFCOMP_NOCOMPRESSBYTE;
	    break;
	case IO_COMP_MH:
	    fieldList->Compression = TFX_TIFFCOMP_FAXCCITT3;
	    /* Note: TFX_TIFFCOMP_CCITT3 is obsolete (comp = 2) */
	    break;
	case IO_COMP_MR:
	    fieldList->Compression = TFX_TIFFCOMP_FAXCCITT3;
	    break;
	case IO_COMP_MMR:
	    fieldList->Compression = TFX_TIFFCOMP_FAXCCITT4;
	    break;
	case IO_COMP_JPEG:
	    fieldList->Compression = TFX_TIFFCOMP_JPEG;
	    break;
	case IO_COMP_JBIG2:
	    fieldList->Compression = TFX_TIFFCOMP_XEROX_JBIG2;
/* Later, use this sanctioned value (if ever) */
/*	    fieldList->Compression = TFX_TIFFCOMP_JBIG2; */
	    break;
	case IO_COMP_TOKENIZATION:
	    fieldList->Compression = TFX_TIFFCOMP_XEROX_TOKENIZATION;
	    break;
	case IO_COMP_LLTOKENIZATION:
	    fieldList->Compression = TFX_TIFFCOMP_XEROX_LLTOKENIZATION;
	    break;
	case IO_COMP_OTHER:
	    XERR_GEN_CLEANUP(fieldList->found.Compression == 0, IO_SEQUENCE);
	    /* fieldList->Compression set in TFX_SetCompression() */
	    break;
	default:
	    __error__ = IO_BADPARAMETER;
	    goto ERR_CLEANUP;
	}

	(*pFieldCount)++;
    }

    /* Write the Photometric field, always */
    if (writing) {
	if (! tfxImg->fields.found.PhotometricInterpretation) {
	    if (paletted) {
		fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_PALETTE;
	    } else if (bppe == 1) {
		fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_WHITEISZERO;
	    } else if (pTfxImgInfo->xColorType == IO_COLORTYPE_COLOR24) {
		if (pTfxImgInfo->xCompression == IO_COMP_JPEG
		    || pTfxImgInfo->xCompression == IO_COMP_NOC) {
		    switch (pTfxImgInfo->xColorSpace) {
		    case IO_COLORSPACE_RGB:
			fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_RGB;
			break;
		    case IO_COLORSPACE_CMYK:
			fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_CMYK;
			break;
		    case IO_COLORSPACE_CIELAB:
			fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_CIELAB;
			break;
		    case IO_COLORSPACE_ICCLAB:
			fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_ICCLAB;
			break;
		    case IO_COLORSPACE_ITULAB:
			fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_ITULAB;
			break;
		    default:
			fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_YCBCR;
		    }
		} else {
		    fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_RGB;
		}
	    } else {
		switch (pTfxImgInfo->xColorSpace) {
		case IO_COLORSPACE_CIELAB:
		    fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_CIELAB;
		    break;
		case IO_COLORSPACE_ICCLAB:
		    fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_ICCLAB;
		    break;
		case IO_COLORSPACE_ITULAB:
		    fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_ITULAB;
		    break;
		default:
		    fieldList->PhotometricInterpretation = TFX_TIFFPHOTOMETRIC_BLACKISZERO;
		}
	    }
	}
	sval = fieldList->PhotometricInterpretation;
	XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, PHOTOMETRICINTERPRETATION, IFDSHORT,
	    1, &dataOffset, &sval, pFieldCount, 0));
    } else {
	(*pFieldCount)++;
    }

    /* Write the FillOrder field value, if it's been set specifically */
    if (fieldList->found.FillOrder) {
	if (writing) {
	    sval = fieldList->FillOrder;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, FILLORDER,
		IFDSHORT, 1, &dataOffset, &sval, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the DocumentName field string, if it's been set specifically */
    if (fieldList->found.DocumentName) {
	if (writing) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, DOCUMENTNAME,
		IFDASCII, fieldList->DocumentNameLength, &dataOffset,
		fieldList->DocumentName, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the ImageDescription field string, if it's been set specifically
     */
    if (fieldList->found.ImageDescription) {
	if (writing) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, IMAGEDESCRIPTION,
		IFDASCII, fieldList->ImageDescriptionLength, &dataOffset,
		fieldList->ImageDescription, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the StripOffsets field */
    if (pTfxImgInfo->xCompression != IO_COMP_UNKNOWN) {

	/* Compute the number of strips */
	if (writing) {
	    if (pTfxImgInfo->dwBestStripHeight > 0 &&
		fieldList->StripOffsets == NULL) {
		fieldList->StripsPerImage = (pTfxImgInfo->dwHeight +
		    pTfxImgInfo->dwBestStripHeight - 1) /
		    pTfxImgInfo->dwBestStripHeight;
		IO_CHECKMALLOC(fieldList->StripOffsets =
		    TFX_MALLOC(fieldList->StripsPerImage *
		    TfxFieldTypeSize[IFDLONG]));
		memset(fieldList->StripOffsets, 0, TfxFieldTypeSize[IFDLONG] *
		    fieldList->StripsPerImage);
	    }
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, STRIPOFFSETS,
		IFDLONG, fieldList->StripsPerImage, &dataOffset,
		fieldList->StripOffsets, pFieldCount, 0));

	    /* Store the location of the field */
	    tfxImg->fieldLocation.StripOffsets = tfxImg->ifdOffset + 2 +
		(*pFieldCount * 12) - 4;

	    /* Store the location of the strip offsets array */
	    tfxImg->fieldDataLocation.StripOffsets = dataOffset;
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the Orientation field value, if it's been set specifically */
    if (fieldList->found.Orientation) {
	if (writing) {
	    sval = fieldList->Orientation;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, ORIENTATION,
		IFDSHORT, 1, &dataOffset, &sval, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /*
     * Write the SamplesPerPixel field if it is not the default (1)
       or if the doc is an RFC2304 profile */

    if ((channels > 1)
/*	|| (tfxImg->tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_S) */
/*	|| (tfxImg->tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_F) */
/*	|| (tfxImg->tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_C) */
/*	|| (tfxImg->tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_M) */
/*	|| (tfxImg->tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_T) */
/*	|| (tfxImg->tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_M1) */
) {
	if (writing) {
	    sval = fieldList->SamplesPerPixel = channels;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, SAMPLESPERPIXEL,
		IFDSHORT, 1, &dataOffset, &sval, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the RowsPerStrip field */
    if (pTfxImgInfo->dwBestStripHeight > 0 &&
	pTfxImgInfo->xCompression != IO_COMP_UNKNOWN) {
	if (writing) {
	    sval = (UInt16)(fieldList->RowsPerStrip =
		pTfxImgInfo->dwBestStripHeight);
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, ROWSPERSTRIP,
		IFDSHORT, 1, &dataOffset, &sval, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the StripByteCounts field */
    if (pTfxImgInfo->xCompression != IO_COMP_UNKNOWN) {
	if (writing) {
	    if (fieldList->StripsPerImage > 0 &&
		fieldList->StripByteCounts == NULL) {
		IO_CHECKMALLOC(fieldList->StripByteCounts =
		    TFX_MALLOC(fieldList->StripsPerImage *
			TfxFieldTypeSize[IFDLONG]));
		memset(fieldList->StripByteCounts, 0, TfxFieldTypeSize[IFDLONG] *
		    fieldList->StripsPerImage);
	    }
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, STRIPBYTECOUNTS,
		    IFDLONG, fieldList->StripsPerImage, &dataOffset,
		fieldList->StripByteCounts, pFieldCount, 0));

	    /* Store the location of the field */
	    tfxImg->fieldLocation.StripByteCounts = tfxImg->ifdOffset + 2 +
		(*pFieldCount * 12) - 4;

	    /* Store the location of the strip byte count */
	    tfxImg->fieldDataLocation.StripByteCounts = dataOffset;
	} else {
	    (*pFieldCount)++;
	}
    }

    /* If there's not been a specific mention of the resolution unit,
     * use inches */
    if (! fieldList->found.ResolutionUnit) {
	fieldList->ResolutionUnit = INCHES;
    }

    /* Write the XResolution field, always */
    if (writing) {
	switch (fieldList->ResolutionUnit) {
	case CENTIMETERS:
	    rval.numerator = fieldList->XResolution.numerator =
		(long)(pTfxImgInfo->dwXResolution * 2.53995f);
	    rval.denominator = fieldList->XResolution.denominator = 1;
	    break;
	default: /* expect inches */
	    rval.numerator = fieldList->XResolution.numerator =
		pTfxImgInfo->dwXResolution;
	    rval.denominator = fieldList->XResolution.denominator = 1;
	}
	XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, XRESOLUTION,
	    IFDRATIONAL, 1, &dataOffset, &rval, pFieldCount, 0));
    } else {
	(*pFieldCount)++;
    }

    /* Write the YResolution field, always */
    if (writing) {
	rval.numerator = fieldList->YResolution.numerator =
	    pTfxImgInfo->dwYResolution;
	rval.denominator = fieldList->YResolution.denominator = 1;
	XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, YRESOLUTION,
	    IFDRATIONAL, 1, &dataOffset, &rval, pFieldCount, 0));
    } else {
	(*pFieldCount)++;
    }

    /* Write the PlanarConfiguration if it's JPEG */
    if (pTfxImgInfo->xCompression == IO_COMP_JPEG) {
	if (writing) {
	    sval = fieldList->PlanarConfiguration = TFX_TIFFPLANAR_CHUNKY;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg,
		PLANARCONFIGURATION, IFDSHORT, 1, &dataOffset, &sval,
		pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the XPosition field, if it's not the primary IFD */
    if (pTfxImgInfo->tfxImgType != IO_IMGTYPE_PRIMARY) {
	if (writing) {
	    rval.numerator = fieldList->XPosition.numerator =
		pTfxImgInfo->dwXOffset;
	    rval.denominator = fieldList->XPosition.denominator =
		((TFX_IMG *)(tfxImg->tfxPage->pageInfo.tfxPrimImgHandle))->imgInfo.dwXResolution;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, XPOSITION,
		IFDRATIONAL, 1, &dataOffset, &rval, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the YPosition field, if it's not the primary IFD */
    if (pTfxImgInfo->tfxImgType != IO_IMGTYPE_PRIMARY) {
	if (writing) {
	    rval.numerator = fieldList->YPosition.numerator =
		pTfxImgInfo->dwYOffset;
	    rval.denominator = fieldList->YPosition.denominator =
		((TFX_IMG *)(tfxImg->tfxPage->pageInfo.tfxPrimImgHandle))->
		imgInfo.dwYResolution;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, YPOSITION,
		IFDRATIONAL, 1, &dataOffset, &rval, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the ResolutionUnit field, always */
    if (writing) {
	sval = (UInt16)fieldList->ResolutionUnit;
	XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, RESOLUTIONUNIT,
	    IFDSHORT, 1, &dataOffset, &sval, pFieldCount, 0));
    } else {
	(*pFieldCount)++;
    }

    /* Write the PageNumber field if it's the primary IFD */
    if (pTfxImgInfo->tfxImgType == IO_IMGTYPE_PRIMARY) {
	if (writing) {
	    if (fieldList->found.PageNumber) {
		/* If the values have been specifically set, use them.
		 * (Do we trust them?) */
		svals[0] = fieldList->PageNumber[0];
		svals[1] = fieldList->PageNumber[1];
	    } else {
		svals[0] = (UInt16)tfxImg->tfxPage->pageInfo.dwPageOffset;
		svals[1] = 0;	/* Don't know the page count yet */
	    }
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, PAGENUMBER,
		IFDSHORT, 2, &dataOffset, svals, pFieldCount, 0));

	    /* Store the location of the page number values.
	     * Later, when the document is finalized, write the real page
	     * count. */
	    tfxImg->fieldDataLocation.PageNumber = dataOffset;
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the Software field string, if it's been set specifically */
    if (fieldList->found.Software) {
	if (writing) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, SOFTWARE,
		IFDASCII, fieldList->SoftwareLength, &dataOffset,
		fieldList->Software, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the HostComputer field string, if it's been set specifically */
    if (fieldList->found.HostComputer) {
	if (writing) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, HOSTCOMPUTER,
		IFDASCII, fieldList->HostComputerLength, &dataOffset,
		fieldList->HostComputer, pFieldCount, 0)); 
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the DateTime field string, if it's been set specifically */
    if (fieldList->found.DateTime) {
	if (writing) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, DATETIME,
		IFDASCII, fieldList->DateTimeLength, &dataOffset,
		fieldList->DateTime, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the Colormap field and data, if it's paletted */
    if (paletted) {
	if (writing) {
	    UInt32 index;
	    UInt32 indices = 1 << bppe;
	    UInt16 *red, *green, *blue;
	    UInt8 *cmap;

	    if (fieldList->Colormap == NULL) {
		IO_CHECKMALLOC(fieldList->Colormap =
		    TFX_MALLOC(indices * 3 *sizeof(*fieldList->Colormap)));
		memset(fieldList->Colormap, 0,
		    indices * sizeof(*fieldList->Colormap));
	    }

	    /* Convert from the internal colormap, to a TIFF format */
	    red = fieldList->Colormap;
	    green = red + indices;
	    blue = green + indices;
	    cmap = pTfxImgInfo->pPalette;
	    for (index = 0; index < indices; index++) {
		*red++ = ((UInt16)*cmap << 8) | *cmap;
		cmap++;
		*green++ = ((UInt16)*cmap << 8) | *cmap;
		cmap++;
		*blue++ = ((UInt16)*cmap << 8) | *cmap;
		cmap++;
	    }

	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, COLORMAP,
		IFDSHORT, indices * 3, &dataOffset, fieldList->Colormap,
		pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the T4Options field options, if it's MH/MR compressed */
    if (fieldList->Compression == TFX_TIFFCOMP_FAXCCITT3) {
	/* Bit 0: 2-D encoding: 0 = MH, 1 = MR
	 * Bit 1: must be 0
	 * Bit 2: 0 = EOLs not byte aligned, 1 = EOLs are byte aligned
	 */
	if (writing) {
	    if (fieldList->found.T4Options) {
		lval = fieldList->T4Options;
	    } else if (pTfxImgInfo->xCompression == IO_COMP_MH) {
		lval = 4;	/* Our compressor does this */
	    } else if (pTfxImgInfo->xCompression == IO_COMP_MR) {
		lval = 1;	/* Our compressor does this */
	    } else {
		lval = 0;	/* Should have been one of the above */
	    }
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, T4OPTIONS, IFDLONG,
		1, &dataOffset, &lval, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the T6Options field options, if it's G4 compressed AND it's
     * been set specifically */
    if (fieldList->Compression == TFX_TIFFCOMP_FAXCCITT4
	&& fieldList->found.T6Options) {
	/* Bit 0: always 0
	 * Bit 1: uncompressed mode allowed */
	if (writing) {
	    lval = fieldList->T6Options;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, T6OPTIONS, IFDLONG,
		1, &dataOffset, &lval, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

#if 0
    /* Write the TileWidth field if it's .... compressed */
    if (pTfxImgInfo->xCompression == IO_COMP_JPEG) {
	if (writing) {
	    lval = fieldList->TileWidth = pTfxImgInfo->dwWidth;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, TILEWIDTH, IFDLONG,
		1, &dataOffset, &lval, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the TileLength field if it's .... compressed */
    if (pTfxImgInfo->xCompression == IO_COMP_JPEG) {
	if (writing) {
	    lval = fieldList->TileLength = pTfxImgInfo->dwBestStripHeight;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, TILELENGTH, IFDLONG,
		1, &dataOffset, &lval, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the TileOffsets field if it's .... compressed */
    if (pTfxImgInfo->xCompression == IO_COMP_JPEG) {
	if (writing) {
	    fieldList->TileCount = (pTfxImgInfo->dwHeight +
		pTfxImgInfo->dwBestStripHeight - 1) / pTfxImgInfo->dwBestStripHeight;
	    IO_CHECKMALLOC(fieldList->TileOffsets =
		TFX_MALLOC(fieldList->TileCount * TfxFieldTypeSize[IFDLONG]));
	    memset(fieldList->TileOffsets, 0, TfxFieldTypeSize[IFDLONG] *
		fieldList->TileCount);
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, TILEOFFSETS, IFDLONG,
		fieldList->TileCount, &dataOffset, fieldList->TileOffsets,
		pFieldCount, 0));

	    /* Store the location of the tile offsets */
	    tfxImg->fieldDataLocation.StripOffsets = dataOffset;
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the TileByteCounts field if it's .... compressed */
    if (pTfxImgInfo->xCompression == IO_COMP_JPEG) {
	if (writing) {
	    IO_CHECKMALLOC(fieldList->TileByteCounts =
		TFX_MALLOC(fieldList->TileCount * TfxFieldTypeSize[IFDLONG]));
	    memset(fieldList->TileByteCounts, 0, TfxFieldTypeSize[IFDLONG] *
		fieldList->TileCount);
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, TILEBYTECOUNTS, IFDLONG,
		fieldList->TileCount, &dataOffset, fieldList->TileByteCounts,
		pFieldCount, 0));

	    /* Store the location of the tile offsets */
	    tfxImg->fieldDataLocation.StripByteCounts = dataOffset;
	} else {
	    (*pFieldCount)++;
	}
    }
#endif

    /* Write the SubIFDs field, if it could have child images:
     *	- primary image and binary
     *  - special exception: primary image may now have a transparency mask
     *    (color only?)
     *  - MRC profile
     */
#if 0
    /* Why did I do it this way before, see "huh?" above. */
    if ((pTfxImgInfo->tfxImgType == IO_IMGTYPE_PRIMARY) ||
	(pTfxImgInfo->dwImgLayer & 1)) {
    }
#endif
    if (pTfxImgInfo->tfxImgType == IO_IMGTYPE_PRIMARY) {
	switch (tfxImg->tfxPage->tfxDoc->docInfo.profile) {
	case TFX_PROFILE_M:	/* Profile M: MRC */
	case TFX_PROFILE_M1:	/* Profile M1: MRC + JBIG2 */
	case TFX_PROFILE_COH_M:	/* Coherent TIFF Profile M: MRC */
	case TFX_PROFILE_COH_M1:/* Coherent TIFF Profile M1: JBIG2 */

	    if (writing) {
		/* Allocate the list of SubIFDs if not already done (may have
		 * been done when attaching child images to a tokenized parent
		 * that hasn't been actually written).
		 * Note that the recommendation is that there be one IFD chaing
		 * for each layer.  Since layer 2 is normally a single IFD,
		 * there's no need for a chain for that layer.  As such, I
		 * will create one IFD chain for all layers up to the highest
		 * layer number, minus one.
		 * During a streaming copy, set fieldList->SubIFDCount to the
		 * maxiumum layer, minus one.
		 */
		if (fieldList->SubIFDCount == 0) {
		    /* Extend up to 1 without having to rewrite when extending */
		    fieldList->SubIFDCount = 1;
		}
		if (fieldList->SubIFDs == NULL) {
		    IO_CHECKMALLOC(fieldList->SubIFDs =
			TFX_MALLOC(fieldList->SubIFDCount * TfxFieldTypeSize[IFDIFD]));
		    memset(fieldList->SubIFDs, 0, fieldList->SubIFDCount *
			TfxFieldTypeSize[IFDIFD]);
		}

		XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, SUBIFDS, IFDIFD,
		    fieldList->SubIFDCount, &dataOffset, fieldList->SubIFDs,
		    pFieldCount, 0));

		/* Store the location of the field */
		tfxImg->fieldLocation.SubIFDs = tfxImg->ifdOffset + 2 +
		    (*pFieldCount * 12) - 4;

		/* Retain the location of the field data */
		tfxImg->fieldDataLocation.SubIFDs = dataOffset;
	    } else {
		(*pFieldCount)++;
	    }
	}
    }

    /* Write the TargetPrinter field string, if it's been set specifically */
    if (fieldList->found.TargetPrinter) {
	if (writing) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, TARGETPRINTER,
		IFDASCII, fieldList->TargetPrinterLength, &dataOffset,
		fieldList->TargetPrinter, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the JPEGTables field, if using JPEGTables mode, and the
     * image was JPEG compressed */
    if ((tfxImg->imgFlags & IO_IMGFLAGS_JPEGTables) &&
	(fieldList->Compression == 6 || fieldList->Compression == 7)) {
	if (writing) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, JPEGTABLES,
		IFDUNDEFINED, tfxImg->qtableBufsize, &dataOffset,
		tfxImg->qtableBuf, pFieldCount, 0));

	    /* Store the location of the field */
	    tfxImg->fieldLocation.JPEGTables = tfxImg->ifdOffset + 2 +
		(*pFieldCount * 12) - 4;

	    /* Retain the position of this field value */
	    tfxImg->fieldDataLocation.JPEGTables = dataOffset;
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the GlobalParameters IFD on the first primary images, if not opted
     * to be left off, and if we have a JBIG2 Profile (only).
     */
    switch (tfxImg->tfxPage->tfxDoc->docInfo.profile) {
    case TFX_PROFILE_M:		/* Profile M: MRC */
    case TFX_PROFILE_T:		/* Profile T: Tokenization (JBIG2) */
    case TFX_PROFILE_M1:	/* Profile M1: MRC + JBIG2 */
    case TFX_PROFILE_COH_M:	/* Coherent TIFF Profile M: MRC */
    case TFX_PROFILE_COH_M1:	/* Coherent TIFF Profile M1: JBIG2 */
	if (pTfxImgInfo->tfxImgType == IO_IMGTYPE_PRIMARY
	    && tfxImg->tfxPage->pageInfo.dwPageOffset == 0
	    && (tfxImg->tfxPage->tfxDoc->seqStatus & TFX_DOC_SEQ_NOGPIFDREQ) == 0) {
	    if (writing) {

		/* Write the GlobalParametersIFD */
		XERR_PASS_CLEANUP(wp_TifFXWriteGPIFD(tfxImg));

		/* Write the location of the IFD */
		lval = tfxImg->tfxPage->tfxDoc->gpifdOffset;
		XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg,
		    GLOBALPARAMETERSIFD, IFDIFD, 1, &dataOffset, &lval,
		    pFieldCount, 0));
	    } else {
		(*pFieldCount)++;
	    }
	}
	break;
    default:
	break;
    }

    /* Write the Decode field and data, if it's been set specifically */
    if (fieldList->found.Decode &&
	(pTfxImgInfo->xColorType == IO_COLORTYPE_GRAY4
	|| pTfxImgInfo->xColorType == IO_COLORTYPE_GRAY8
	|| pTfxImgInfo->xColorType == IO_COLORTYPE_COLOR24)) {
	if (writing) {
	    switch (pTfxImgInfo->xColorType) {
	    case IO_COLORTYPE_GRAY4:
	    case IO_COLORTYPE_GRAY8:
		indices = 2;
		break;
	    case IO_COLORTYPE_COLOR24:
	    default:
		indices = 6;
		break;
	    }
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, DECODE,
		IFDSRATIONAL, indices, &dataOffset, fieldList->Decode,
		pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the ImageBaseColor field and data, if it's been set specifically
     */
    if (fieldList->found.ImageBaseColor &&
	(tfxImg->tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_M ||
	 tfxImg->tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_T ||
	 tfxImg->tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_COH_M ||
	 tfxImg->tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_COH_M1)) {
	
	if (writing) {
	    switch (pTfxImgInfo->xColorType) {
	    case IO_COLORTYPE_BINARY:
	    case IO_COLORTYPE_GRAY4:
	    case IO_COLORTYPE_GRAY8:
		indices = 1;
		break;
	    case IO_COLORTYPE_CMYK1:
	    case IO_COLORTYPE_CMYK4:
	    case IO_COLORTYPE_CMYK8:
		indices = 4;
		break;
	    case IO_COLORTYPE_COLOR24:
	    default:
		if (paletted) {
		    indices = 1;
		} else {
		    indices = 3;
		}
		break;
	    }
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, IMAGEBASECOLOR,
		IFDSHORT, indices, &dataOffset, fieldList->ImageBaseColor,
		pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the ChromaSubSampling field options, if it's been explicitly set
     */
    if (fieldList->found.ChromaSubSampling && bppe > 1) {
	if (writing) {
	    svals[0] = fieldList->ChromaSubSamplingHor;
	    svals[1] = fieldList->ChromaSubSamplingVer;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg,
		CHROMASUBSAMPLING, IFDSHORT, 2, &dataOffset, &svals[0],
		pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the ChromaPositioning field options, if it's been explicitly set
     */
    if (fieldList->found.ChromaPositioning && bppe > 1) {
	if (writing) {
	    sval = fieldList->ChromaPositioning;
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg,
		CHROMAPOSITIONING, IFDSHORT, 1, &dataOffset, &sval,
		pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the StripRowsCounts field, only if we don't have a fixed strip size */
    if (pTfxImgInfo->dwBestStripHeight == 0) {
	if (writing) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, STRIPROWCOUNTS,
		IFDLONG, fieldList->StripsPerImage, &dataOffset, fieldList->StripRowCounts, pFieldCount, 0));

	    /* Store the location of the field */
	    tfxImg->fieldLocation.StripRowCounts = tfxImg->ifdOffset + 2 +
		(*pFieldCount * 12) - 4;

	    /* Retain the position of this field value */
	    tfxImg->fieldDataLocation.StripRowCounts = dataOffset;
	} else {
	    (*pFieldCount)++;
	}
    }


    /* Write the Copyright field string, if it's been set specifically */
    if (fieldList->found.CopyRight) {
	if (writing) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, COPYRIGHT,
		IFDASCII, fieldList->CopyRightLength, &dataOffset,
		fieldList->CopyRight, pFieldCount, 0));
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the AnnotationOffsets field, if it's the primary image,
     * and it's be explicitly set */
    if (pTfxImgInfo->tfxImgType == IO_IMGTYPE_PRIMARY &&
	fieldList->found.AnnotationOffsets) {
	if (writing) {
	    if (fieldList->AnnotationOffsets == NULL) {
		IO_CHECKMALLOC(fieldList->AnnotationOffsets =
		    TFX_MALLOC(1 * TfxFieldTypeSize[IFDLONG]));  /* 1 empty entry */
		fieldList->AnnotationOffsets[0] = tfxImg->tfxPage->annotTableHeaderOffset;
	    }
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, ANNOTATIONOFFSETS, IFDLONG,
		1, &dataOffset, fieldList->AnnotationOffsets, pFieldCount, 0));
	    
	    /* Store the location of the field */
	    tfxImg->fieldLocation.AnnotationOffsets = tfxImg->ifdOffset + 2 +
		(*pFieldCount * 12) - 4;
	    
	    /* Preserve the field location */
	    tfxImg->fieldDataLocation.AnnotationOffsets = dataOffset;
	    
	} else {
	    (*pFieldCount)++;
	}
    }

#if 0
    /* Write the MaskIFDs field, if it could have a transparency mask
     * (TIFF-FX 2.0 only)
     */
    if (version < 3 && pTfxImgInfo->tfxImgType != IO_IMGTYPE_PRIMARY && bpp != 1) {
	if (writing){
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, MASKIFDS, IFDLONG,
	        1, &dataOffset, &fieldList->MaskIFDs, pFieldCount, 0));

	    /* Preserve the field location */
	    tfxImg->fieldLocation.MaskIFDs = tfxImg->ifdOffset + 2 +
		(*pFieldCount * 12) - 4;
	} else {
	    (*pFieldCount)++;
	}
    }
#endif

    /* Write the T88Options field options, if it's JBIG2 compressed */
    switch (tfxImg->tfxPage->tfxDoc->docInfo.profile) {
    case TFX_PROFILE_T:
    case TFX_PROFILE_M1:
    case TFX_PROFILE_COH_M1:
	switch (tfxImg->imgInfo.xCompression) {
	case IO_COMP_TOKENIZATION:
	case IO_COMP_LLTOKENIZATION:
	case IO_COMP_JBIG2:
	case IO_COMP_LLJBIG2:
	    if (writing) {

		/* I guess there's no other good opportunity to propogate the
		 * lossless characteristic.  Let's do it here: */
		switch (tfxImg->imgInfo.xCompression) {
		case IO_COMP_LLTOKENIZATION:
		case IO_COMP_LLJBIG2:
		    fieldList->T88Options |= T88OPTION_LOSSLESS;
		    break;
		}

		/* If the image is a contone layer, then the binary image
		 * MUST have been placed in a previous layer */
		if (tfxImg->imgInfo.dwImgLayer & 1) {
		    fieldList->T88Options |= T88OPTION_COLOURTAGS;
		}

		lval = fieldList->T88Options;
		XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, T88OPTIONS,
		    IFDLONG, 1, &dataOffset, &lval, pFieldCount, 0));

		/* Remember the field data location */
		tfxImg->fieldDataLocation.T88Options = dataOffset;
	    } else {
		(*pFieldCount)++;
	    }
	}
    }

    /* Write the RenderingHints field string, if it's been set specifically,
     * or if the IFD is a RenderingHintsIFD (it sounds reasonable that ALL
     * RenderingHintsIFDs should have a hints associated with the image).
     */
    if ((fieldList->found.RenderingHints && fieldList->RenderingHints) ||
	tfxImg->imgInfo.tfxImgType == IO_IMGTYPE_HINTS) {
	if (writing) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, RENDERINGHINTS,
		IFDASCII, fieldList->RenderingHintsLength, &dataOffset,
		fieldList->RenderingHints, pFieldCount, 0));

	    /* Store the location of the field */
	    tfxImg->fieldLocation.RenderingHints = tfxImg->ifdOffset + 2 +
		(*pFieldCount * 12) - 4;
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the RenderingHintsIFD field, if it's been set specifically
     */
    if (fieldList->found.RenderingHintsIFD && tfxImg->xHintsImgs) {
	if (writing) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, RENDERINGHINTSIFD, IFDLONG,
		1, &dataOffset, &fieldList->RenderingHintsIFD, pFieldCount, 0));

	    /* Store the location of the field */
	    tfxImg->fieldLocation.RenderingHintsIFD = tfxImg->ifdOffset + 2 +
		(*pFieldCount * 12) - 4;
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the IFDBaseColor field and data, if an MRC profile, and an extension,
     * and if it's been set specifically
     */
    if (fieldList->found.IFDBaseColor &&
	(tfxImg->tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_M1 ||
	 tfxImg->tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_COH_M1)) {
	
	if (writing) {

	    switch (pTfxImgInfo->xColorType) {
	    case IO_COLORTYPE_BINARY:
	    case IO_COLORTYPE_GRAY4:
	    case IO_COLORTYPE_GRAY8:
		indices = 1;
		break;
	    case IO_COLORTYPE_CMYK1:
	    case IO_COLORTYPE_CMYK4:
	    case IO_COLORTYPE_CMYK8:
		indices = 4;
		break;
	    case IO_COLORTYPE_COLOR24:
	    default:
		if (paletted) {
		    indices = 1;
		} else {
		    indices = 3;
		}
		break;
	    }

	    if (IO_BPPE(tfxImg->imgInfo.xColorType) <= 8) {
		UInt8 byteArray[4];
		UInt32 index = 0;

		/* Copy over the data to a byte array */
		for (index = 0; index < indices; index++) {
		    byteArray[index] = (UInt8)fieldList->IFDBaseColor[index];
		}

		XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, IFDBASECOLOR,
		    IFDBYTE, indices, &dataOffset, byteArray,
			pFieldCount, 0));
		fieldList->IFDBaseColorType = IFDBYTE;
	    } else {
		if (pTfxImgInfo->xColorSpace == IO_COLORSPACE_CIELAB) {
		    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, IFDBASECOLOR,
			IFDSLONG, indices, &dataOffset, fieldList->IFDBaseColor,
			pFieldCount, 0));
		    fieldList->IFDBaseColorType = IFDSLONG;
		} else {
		    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, IFDBASECOLOR,
			IFDSHORT, indices, &dataOffset, fieldList->IFDBaseColor,
			pFieldCount, 0));
		    fieldList->IFDBaseColorType = IFDSHORT;
		}
	    }
	} else {
	    (*pFieldCount)++;
	}
    }

    /* Write the ImageLayer field if it's an image within the normal model,
     * and is an MRC profile.
     * Note, only XIFF has MRC images that could not have ImageLayer field,
     * namely mask images */
    switch (tfxImg->tfxPage->tfxDoc->docInfo.profile) {
    case TFX_PROFILE_M:	/* Profile M: MRC */
    case TFX_PROFILE_M1:	/* Profile M1: MRC + JBIG2 */
    case TFX_PROFILE_COH_M:	/* Coherent TIFF Profile M: MRC */
    case TFX_PROFILE_COH_M1:/* Coherent TIFF Profile M1: MRC + JBIG2 */

	switch (tfxImg->imgInfo.tfxImgType) {
	case IO_IMGTYPE_PRIMARY:
	case IO_IMGTYPE_LAYER:
	    if (writing) {
		lvals[0] = fieldList->ImageLayer[0] = pTfxImgInfo->dwImgLayer;
		lvals[1] = fieldList->ImageLayer[1] = pTfxImgInfo->dwImgOffset + 1;
		XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, IMAGELAYER, IFDLONG,
		    2, &dataOffset, &lvals, pFieldCount, 0));
		
		/* Remember the field data location */
		tfxImg->fieldDataLocation.ImageLayer = dataOffset;
	    } else {
		(*pFieldCount)++;
	    }
	    break;
	default:
	    break;
	}
    }

#if 0
    /* Write the ImageLayerContent field, if it's the primary image */
    if (pTfxImgInfo->tfxImgType == IO_IMGTYPE_PRIMARY && version > 2) {
	if (writing) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteFieldEntry(tfxImg, IMAGELAYERCONTENT,
		IFDLONG, 0, &dataOffset, NULL, pFieldCount, 0));

	    /* Will be the tag offset because < 4 bytes was written . Something
	     * key to note here, is that the TIFF count is what tells the maximum
	     * MRC layer, and will be located 4 bytes before this offset. */
	    tfxImg->fieldLocation.ImgLayerContent = tfxImg->ifdOffset + 2 +
		(*pFieldCount * 12) - 4;

	    /* Now actually write the data */
	    XERR_PASS_CLEANUP(wp_TifFXUpdateILContent(tfxImg));

	    /* Leave the file pointer positioned at the end of the field */
	    TFX_SEEK(tfxFile, tfxImg->ifdOffset + sizeof(*pFieldCount) +
		 *pFieldCount * TFX_TIFFFIELDSIZE);
	} else {
	    (*pFieldCount)++;
	}
    }
#endif
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


#define TEMP_DATA_BUFFER_SIZE	1024
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write an individual TIFF field
 *
 * LongDescription:::
 *  wp_TifFXWriteFieldEntry() writes a single field entry, including the data
 *  associated with it. The field entry must have the fields fieldID, type,
 *  and len filled in, and then the associated data pointer passed in
 *  through the *fieldData parameter. This function will fill in the
 *  appropriate value for the offset field. This function will also do
 *  the necessary byte swapping, as appropriate for the endian of the
 *  file, and the type of data.
 *  <BR>
 *  Some of the field 'offset' values are retained for later update. This
 *  is a somewhat obscure thing to do, but it avoids rereading the IFD to
 *  relocate the data, which can't be done anyway if the file was opened
 *  write only (no read).
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  tagID::			TIFF field tag ID
 *  type::			TIFF field data type
 *  len::			TIFF field data length
 *  pFieldData::		Data to store
 *  gpifd::			Is this the GPIFD (or just the normal IFD)?
 *
 * Output:::
 *  pOffset::			Offset of where the data was stored
 *  pFieldCount::		Pointer to accumulating field count
 *
 * Returns:::
 *  error status, sets the appropriate
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXWriteFieldEntry(
    TFX_IMG *tfxImg,
    UInt16 tagID,
    UInt16 type,
    UInt32 len,
    Int32 *pOffset,
    const void *pFieldData,
    UInt16 *pFieldCount,
    Bool gpifd
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt16 sval;
    UInt32 lval;
    UInt8 tempBuffer[TEMP_DATA_BUFFER_SIZE];
    UInt8 *tempBuf = NULL;
    UInt8 swap;
    UInt32 bytes;
    UInt32 nullBytes = 0;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);

    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    swap = (UInt8)TFX_FILETOKEN(tfxFile)->bSwapBytes;

    XERR_GEN_CLEANUP(pOffset == 0, IO_INTERNAL_ERROR);

    /* If we haven't made room for the field, we can't fit it in */
    if (gpifd) {
	XERR_GEN_CLEANUP((tfxImg->tfxPage->tfxDoc->reservedGPFieldCount <= *pFieldCount),
	    IO_INTERNAL_ERROR);
    } else {
	XERR_GEN_CLEANUP((tfxImg->reservedFieldCount <= *pFieldCount),
	    IO_INTERNAL_ERROR);
    }

    /* If the type is ASCII, make sure it's NULL terminated */
    if (type == IFDASCII) {
	UInt8 *byteArray = (UInt8 *)pFieldData;

	if (byteArray[len - 1] != '\0') {
	    nullBytes = 1;
	}
    }

    /* Check to see if the data will fit into the offset value */
    bytes = TfxFieldTypeSize[type] * (len + nullBytes);
    if (bytes <= 4) {
	XERR_PASS_CLEANUP(wp_TifFXConvcopy(pOffset, pFieldData, type, len, swap));

    } else {	/* Data won't stuff into the offset value */

	/* Place at the end of the file */
	XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, pOffset));

	if (swap) {

	    /* If data buffer is small, use local stack */
	    if (bytes <= TEMP_DATA_BUFFER_SIZE) {
		tempBuf = tempBuffer;
	    } else {	/* Otherwise use the client memory */
		IO_CHECKMALLOC(tempBuf = TFX_MALLOC(bytes - nullBytes));
	    }

	    /* Copy and swap the data */
	    XERR_PASS_CLEANUP(wp_TifFXConvcopy(tempBuf, pOffset, type, len,
		swap));

	    /* Write the field data out */
	    TFX_WRITE(tfxFile, tempBuf, (Int32)(bytes - nullBytes));
	} else {
	    /* Write the field data out */
	    TFX_WRITE(tfxFile, pFieldData, (Int32)(bytes - nullBytes));
	}

	/* If adding the ASCII NULL, write it separately (wasn't in the
	 * buffer passed in) */
	if (nullBytes) {
	    UInt8 null = 0;
	    TFX_WRITE(tfxFile, &null, (Int32)nullBytes);
	}

    }

    /* Go to the beginning of the field */
    if (gpifd) {
	TFX_SEEK(tfxFile, tfxImg->tfxPage->tfxDoc->gpifdOffset + sizeof(*pFieldCount) +
	    *pFieldCount * TFX_TIFFFIELDSIZE);
    } else {
	TFX_SEEK(tfxFile, tfxImg->ifdOffset + sizeof(*pFieldCount) + *pFieldCount *
	    TFX_TIFFFIELDSIZE);
    }
	
    /* Write the field. Keep in mind, if we're writing to an existing file, we
     * may need to swap if the endian mismatches */
    sval = tagID;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    sval = type;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    lval = len;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));
    lval = *pOffset;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));
	
    /* Retain the data location (used for some fields) */
    if (bytes <= 4) {
	if (gpifd) {
	    *pOffset = tfxImg->tfxPage->tfxDoc->gpifdOffset + sizeof(*pFieldCount) +
		*pFieldCount * TFX_TIFFFIELDSIZE + sizeof(sval) * 2 + sizeof(lval);
	} else {
	    *pOffset = tfxImg->ifdOffset + sizeof(*pFieldCount) + *pFieldCount *
		TFX_TIFFFIELDSIZE + sizeof(sval) * 2 + sizeof(lval);
	}
    }

#ifdef DEBUG
    printf("\t%2d:\t%2d\t%2d%8d%8d\n", *pFieldCount, fieldID,
	type, len, pOffset);
#endif

    /* Update the field count */
    (*pFieldCount)++;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    if (tempBuf && (tempBuf != tempBuffer)) {
	TFX_FREE(tempBuf);
    }
    
    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Update the "nextIFD" part of either the previous IFD or TIFF header.
 *
 * LongDescription:::
 *  wp_TifFXWriteNextIFD() writes the "nextIFD" to the IFD of the previous IFD,
 *  or the TIFF header itself. If the prevImg is passed in, it will be used
 *  to identify the location of the nextIFD location to set. Otherwise, the
 *  nextIFD of the IFD located at prevIFDOffset is set. If both are 0 (or
 *  NULL), then the nextIFD in the TIFF header is set.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  prevImg::			Previous IFD in chain
 *  prevIFDOffset::		If no prevImg, this IFD is modified
 *  newIFDOffset::		IFD offset of the IFD being added
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXWriteNextIFD(
		TFX_IMG *tfxImg,
		TFX_IMG *prevImg,
		UInt32 prevIFDOffset,
		UInt32 newIFDOffset
)
{
    TFX_FILE *tfxFile;
    UInt16 fieldCount;
    UInt32 lval;
    Int32 seekOffset = 0L;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);

    tfxFile = TFX_IMGFILE(tfxImg);

    if (prevImg != NULL && prevIFDOffset == 0UL) {
	prevIFDOffset = prevImg->ifdOffset;
    }

    if (prevIFDOffset == 0UL) {
	TFX_TIFFHEADER tiffHeader;	/* Use only for sizeof macros */

	/* Not used, avoids compiler warnings */
	tiffHeader.firstIfd = newIFDOffset;
	seekOffset = sizeof(tiffHeader.byteOrder) + sizeof(tiffHeader.version);

	/* Make sure we're not referencing something that doesn't exist.
	 * If we have a previous image, but we think we're pointing to the
	 * TIFF header, something is really wrong */
	XERR_GEN_CLEANUP(prevImg != NULL, IO_BADPARAMETER);
    } else {

	TFX_CheckHandle(tfxImg);

	if (tfxImg->imgInfo.tfxImgType == IO_IMGTYPE_PRIMARY &&
	       tfxImg->tfxPage->prev) {

	    /* Use the previous primary IFD from memory */

	    TFX_PAGE *prevPage;

	    /* Look for the previous page */
	    prevPage = tfxImg->tfxPage->prev;
	    prevImg = (TFX_IMG *)prevPage->pageInfo.tfxPrimImgHandle;

	    /* If the image hasn't been stored yet, or it already
	     * points to this IFD offset, no need to continue */
	    if (prevImg == NULL || (prevImg->fields.dataTransferred &&
		prevImg->fields.nextIFD == newIFDOffset)) {
		NORMAL_RETURN;
	    }
	    
	    /* Store the nextIFD on the previous image (may be written later)
	     */
	    prevImg->fields.nextIFD = newIFDOffset;

	    /* If we haven't written the data yet, save it for later */
	    if (prevImg->fields.dataTransferred == 0) {
		NORMAL_RETURN;
	    }
	}
    }

    if (prevImg) {

	/* Make sure the IFD offset is saved internally */
	prevImg->fields.nextIFD = newIFDOffset;
    }

    /* If we have found the previous image, see whether the IFD has been
     * written */
    if (prevImg && prevImg->fields.fieldCount) {

	/* Go to the nextIFD location, past the fields */
	seekOffset = prevIFDOffset + 2 +
	    prevImg->fields.fieldCount * TFX_TIFFFIELDSIZE;
    } else if (prevImg) {

	/* The nextIFD value (now stored in the fields structure) will be
	 * written to the file later (this happens when the image has been
	 * tokenized, but not written until the dictionary is resolved).
	 */
	NORMAL_RETURN;

    } else if (seekOffset == 0 && tfxImg &&
	(tfxImg->tfxPage->tfxDoc->readWrite & READ_ALLOWED)) {

	/* Go to the beginning of the IFD */
	TFX_SEEK(tfxFile, prevIFDOffset);

	/* Read the field count */
	TFX_READ(tfxFile, &fieldCount, sizeof(fieldCount));
	TFX_SWAPSHORT(tfxFile, &fieldCount);

	/* Go to the nextIFD location, past the fields */
	seekOffset = prevIFDOffset + 2 + fieldCount * TFX_TIFFFIELDSIZE;
    }

    /* Make sure we've found what we're looking for */
    XERR_GEN_CLEANUP(seekOffset < 4, IO_INTERNAL_ERROR);

    /* Seek to the nextIFD location */
    TFX_SEEK(tfxFile, seekOffset);

    /* Write the nextIFD */
    lval = newIFDOffset;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}



/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write an entire GlobalParametersIFD
 *
 * LongDescription:::
 *  wp_TifFXWriteGPIFD() writes the various components of a GlobalParametersIFD.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXWriteGPIFD(
    TFX_IMG *tfxImg
)
{
    TFX_FILE *tfxFile;
    TFX_DOC *tfxDoc;
    UInt16 fieldCount = 0;
    TFX_IMGINFO *pTfxImgInfo = NULL;
    UInt16 sval;
    UInt32 lval;
    UInt32 offset = 0UL;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    tfxDoc = tfxImg->tfxPage->tfxDoc;
    pTfxImgInfo = &tfxImg->imgInfo;

    /* Write an IFD placeholder */
    fieldCount = tfxDoc->reservedGPFieldCount;
    if (fieldCount == 0) {
	XERR_PASS_CLEANUP(wp_TifFXWriteGPFields(tfxImg, &fieldCount, 0));

	/* Make a place-holder for the IFD */
	XERR_PASS_CLEANUP(wp_TfxWriteMockIFD(tfxImg, fieldCount, &offset));

	/* Retain the file offset and field count */
	tfxDoc->gpifdOffset = offset;
	tfxDoc->reservedGPFieldCount = fieldCount;
    }

    /* Go to the beginning of the IFD */
    XERR_GEN_CLEANUP((tfxDoc->gpifdOffset == 0UL),
		     IO_INTERNAL_ERROR);
    TFX_SEEK(tfxFile, tfxDoc->gpifdOffset);
    
    /* Write the field count */
    sval = fieldCount;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    tfxImg->fields.gpIFD.fieldCount = fieldCount;

    /* Write the fields */
    XERR_PASS_CLEANUP(wp_TifFXWriteGPFields(tfxImg, &fieldCount, 1));

    /* Write the next IFD (must be positioned just after the fields) */
/*    lval = tfxDoc->gpifdNextIFD; */
    lval = 0;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write an entire IFD
 *
 * LongDescription:::
 *  wp_TifFXWriteIFD() writes the various components of a TIFF-FX to complete a
 *  separate IFD (to be attached outside of this function).
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXWriteIFD(
    TFX_IMG *tfxImg
)
{
    TFX_FILE *tfxFile;
    UInt16 fieldCount = 0;
    TFX_IMGINFO *pTfxImgInfo = NULL;
    UInt16 sval;
    UInt32 lval;
    UInt32 offset = 0UL;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    pTfxImgInfo = &tfxImg->imgInfo;

    /* Write an IFD placeholder */
    if (tfxImg->reservedFieldCount == 0) {
	XERR_PASS_CLEANUP(wp_TifFXWriteFields(tfxImg, &fieldCount, 0));

/*	XERR_PASS_CLEANUP(wp_TfxQueueIFD(tfxImg)); */

	/* End queuing */
/*	tfxFile->bQueuing = 0; */

	/* Make a place-holder for the IFD */
	XERR_PASS_CLEANUP(wp_TfxWriteMockIFD(tfxImg, fieldCount, &offset));

	/* Retain the file offset and field count */
	tfxImg->ifdOffset = offset;
	tfxImg->reservedFieldCount = fieldCount;
#if 0
	/* Check to see if this is the primary image or a child */
	{
	    TFX_IMG *tfxPrimImg = (TFX_IMG *)tfxImg->tfxPage->pageInfo.tfxPrimImgHandle;

	    if (tfxPrimImg != tfxImg) {

		/* Attach this image to its parent */
		XERR_PASS_CLEANUP(wp_TifFXAttachChildIFD(tfxPrimImg, tfxImg, tfxImg->ifdOffset));
	    }
	}
#endif
    }

    /* Use default values */
    tfxImg->fields.SamplesPerPixel = 1;
    tfxImg->fields.FillOrder = 1;

    /* Populate the field list */

    /* Go to the beginning of the IFD */
    XERR_GEN_CLEANUP((tfxImg->ifdOffset == 0UL),
		     IO_INTERNAL_ERROR);
    TFX_SEEK(tfxFile, tfxImg->ifdOffset);
    
    /* Write the field count */
    sval = fieldCount;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    tfxImg->fields.fieldCount = fieldCount;

    /* Signal data as being written so that all writes take place */
    tfxImg->fields.dataTransferred = 1 /* 1 */;

    /* Write the fields */
    XERR_PASS_CLEANUP(wp_TifFXWriteFields(tfxImg, &fieldCount, 1));

    /* If we don't have the IFD (as in the case of a page insertion), get it
     * from the page table */
    if (tfxImg->fields.nextIFD == 0
	&& tfxImg->imgInfo.tfxImgType == IO_IMGTYPE_PRIMARY &&
	tfxImg->tfxPage->tfxDoc->docInfo.wNPages >
	tfxImg->tfxPage->pageInfo.dwPageOffset + 1) {
	UInt8 nextPageFlags;
	Int32 nextPageIFDOffset;

	XERR_PASS_CLEANUP(wp_TfxGetPageIFD(tfxImg->tfxPage->tfxDoc,
	    tfxImg->tfxPage->pageInfo.dwPageOffset,
	    &nextPageFlags, &nextPageIFDOffset));
	tfxImg->fields.nextIFD = nextPageIFDOffset;
    }

    /* Write the next IFD (must be positioned just after the fields) */
    lval = tfxImg->fields.nextIFD;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));
#if 0
    /* Now go back and write the real field count */
    TFX_SEEK(tfxFile, tfxImg->ifdOffset);
    sval = fieldCount;
    TFX_SWAPSHORT(tfxFile, &sval);
    TFX_WRITE(tfxFile, &sval, sizeof(sval));
    tfxImg->fields.fieldCount = fieldCount;
#endif

    /* Keep track of which compression types were used */
    switch (tfxImg->imgInfo.xCompression) {
    case IO_COMP_NOC:	    /* No compression, leave as raw data */
	tfxImg->tfxPage->tfxDoc->codingMethods |= CODINGMETHOD_UNSPECIFIED;
	break;
    case IO_COMP_MH:	    /* Modified Huffman (ITU-T T.?) */
	tfxImg->tfxPage->tfxDoc->codingMethods |= CODINGMETHOD_MH;
	break;
    case IO_COMP_MR:	    /* Modified Read (ITU-T T.?) */
	tfxImg->tfxPage->tfxDoc->codingMethods |= CODINGMETHOD_MR;
	break;
    case IO_COMP_MMR:	    /* Modified modified-read (CCITT
			     * group-4, ITU-T T.6) */
	tfxImg->tfxPage->tfxDoc->codingMethods |= CODINGMETHOD_MMR;
	break;
    case IO_COMP_JPEG:	    /* JPEG */
	tfxImg->tfxPage->tfxDoc->codingMethods |= CODINGMETHOD_JPEG;
	break;
    case IO_COMP_JBIG:	    /* JBIG (aka JBIG1) */
	tfxImg->tfxPage->tfxDoc->codingMethods |= CODINGMETHOD_JBIG;
	break;
    case IO_COMP_JBIG2:	    /* JBIG2 (Symbol-based compression) */
    case IO_COMP_LLJBIG2:   /* Lossless JBIG2 */
	tfxImg->tfxPage->tfxDoc->codingMethods |= CODINGMETHOD_JBIG2;
	break;
    default:
	tfxImg->tfxPage->tfxDoc->codingMethods |= CODINGMETHOD_UNSPECIFIED;
	break;
    }

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Attach a child image to a parent, along the SubIFDs list
 *
 * LongDescription:::
 *  wp_TifFXAttachChildIFD() writes the "nextIFD" to the IFD of the previous IFD
 *  in the SubIFDs list, or the SubIFDs field itself (if the first child).
 *
 * Input:::
 *  parentImg::			Parent image to attach to
 *  tfxImg::			Data structure of valid image
 *  childIFDOffset::		Offset of child IFD
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXAttachChildIFD(
		TFX_IMG *parentImg,
		TFX_IMG *tfxImg,
		UInt32 childIFDOffset
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt32 lval;
    UInt32 offset;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);

    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);

    /* The image must not be the primary image if it has a parent. All other
     * images are children of another image. */
    XERR_GEN_CLEANUP((tfxImg->imgInfo.tfxImgType == IO_IMGTYPE_PRIMARY ||
	(parentImg->imgInfo.tfxImgType != IO_IMGTYPE_PRIMARY)),
		     IO_BADPARAMETER);

    /* The IFD of the parent images must be known */
    XERR_GEN_CLEANUP((parentImg->ifdOffset == 0UL),
		     IO_INTERNAL_ERROR);

    /* If we have a separate IFD chain for each layer, then grab the last
     * subimage along the appropriate chain */
    if (parentImg->fields.SubIFDCount > 0) {
	UInt32 highestRank = 0;
	TFX_IMG *lastImg = NULL, *img;

	/* Search for the last image on that chain (layer) */
	img = parentImg->tfxPage->tfxImgs;
	while (img) {
	    if (img->imgInfo.dwImgLayer == tfxImg->imgInfo.dwImgLayer
		&& img->imgInfo.dwImgOffset >= highestRank
		&& img->fields.nextIFD == 0
		&& img != tfxImg) {
		highestRank = img->imgInfo.dwImgOffset;
		lastImg = img;
	    }
	    img = img->next;
	}

	if (lastImg == NULL) {
	    UInt32 layerOffset;

	    /* None on this chain yet, begin this chain */

	    /* Pick the right chain */
	    layerOffset = tfxImg->imgInfo.dwImgLayer - 1;
	    if (layerOffset >= 1) layerOffset--;	    /* Discount layer 2 */

	    /* If we've gone past the number of anticipated layers, abandon the
	     * SubIFD chain list, and create a new larger one */
	    if (layerOffset >= parentImg->fields.SubIFDCount) {
		UInt32 subImg = 0, newSubIFDCount;
		UInt32 *newSubIFDs;

		/* Increase the count (account for a pair, including the
		 * current addition, but excluding layer 2) */
		newSubIFDCount = tfxImg->imgInfo.dwImgLayer & ~1;

		/* Allocate the updated list, and copy the old items to it. */
		IO_CHECKMALLOC(newSubIFDs = TFX_MALLOC(newSubIFDCount * sizeof(*newSubIFDs)));
		memset(newSubIFDs, 0, newSubIFDCount * sizeof(*newSubIFDs));
		memcpy(newSubIFDs, parentImg->fields.SubIFDs,
		    parentImg->fields.SubIFDCount * sizeof(*newSubIFDs));

		/* Use the new list, discard the old */
		TFX_FREE(parentImg->fields.SubIFDs);
		parentImg->fields.SubIFDs = newSubIFDs;
		newSubIFDs = NULL;
		parentImg->fields.SubIFDCount = newSubIFDCount;

		/* Copy the new list to the end of the file */
		XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile,
		    &parentImg->fieldDataLocation.SubIFDs));
		while (subImg < parentImg->fields.SubIFDCount) {
		    lval = parentImg->fields.SubIFDs[subImg++];
		    TFX_SWAPLONG(tfxFile, &lval);
		    TFX_WRITE(tfxFile, &lval, sizeof(lval));
		}

		/* Fill in the SubIFDs field with the new count and array location */
		XERR_GEN_CLEANUP(parentImg->fieldLocation.SubIFDs == 0,
		    IO_INTERNAL_ERROR);
		TFX_SEEK(tfxFile, parentImg->fieldLocation.SubIFDs - sizeof(lval));
		lval = parentImg->fields.SubIFDCount;
		TFX_SWAPLONG(tfxFile, &lval);
		TFX_WRITE(tfxFile, &lval, sizeof(lval));
		lval = parentImg->fieldDataLocation.SubIFDs;
		TFX_SWAPLONG(tfxFile, &lval);
		TFX_WRITE(tfxFile, &lval, sizeof(lval));
	    }

	    /* There's already room for this one.  Write it to the appropriate place */
	    offset = parentImg->fieldDataLocation.SubIFDs + layerOffset * sizeof(lval);
	    parentImg->fields.SubIFDs[layerOffset] = lval = childIFDOffset;
	    TFX_SEEK(tfxFile, offset);
	    TFX_SWAPLONG(tfxFile, &lval);
	    TFX_WRITE(tfxFile, &lval, sizeof(lval));
	} else {
	    /* Found the last one on the chain, continue the chain */

	    /* Make sure it's really the last on that chain */
	    XERR_GEN_CLEANUP(lastImg->fields.nextIFD != 0, IO_BADFILEFORMAT);

	    /* Attach to the last SubIFD in the list via the nextIFD */
	    XERR_PASS_CLEANUP(wp_TifFXWriteNextIFD(tfxImg,
		lastImg, 0UL, childIFDOffset));
	}

    } else if (parentImg->lastSubImg != NULL){
	TFX_IMG *prevImg;

        /* There is already a SubIFD chain hanging off the parent. If we
	 * know which one to attach this one to, attach it there.
	 */
	prevImg = tfxImg->prev;

	/* If we don't know what the previous image is, attach to the end of
	 * the chain */
	if (prevImg == NULL) {
	    /* If this image is an insertion (lower layer than exists, or previous
	     * one in this layer), we should revise the SubIFDs list instead of placing
	     * this at the end.  This should work though. */
	    prevImg = parentImg->lastSubImg;
	}

	/* Attach to the last SubIFD in the list via the nextIFD */
	XERR_PASS_CLEANUP(wp_TifFXWriteNextIFD(tfxImg,
	    prevImg, 0UL, childIFDOffset));

    } else if (parentImg->fields.dataTransferred &&
	parentImg->fields.SubIFDCount > 0UL &&
	parentImg->fieldDataLocation.SubIFDs) {

	/* We have an active parent image, with no SubIFDs (yet).
	 * Attach to the SubIFDs field of the parent */
	offset = parentImg->fieldDataLocation.SubIFDs;
	parentImg->fields.SubIFDs[0] = lval = childIFDOffset;
	TFX_SEEK(tfxFile, offset);
 	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
    } else {

	/* The only time we should wait to write later is when we're
	 * tokenizing an image and waiting for other pages before actually
	 * writing out the data. */
	XERR_GEN_CLEANUP((parentImg->fields.SubIFDCount > 0UL),
		 IO_INTERNAL_ERROR);

	/* Keep it around as the first one for now (write later) */
	parentImg->fields.SubIFDCount = 1;
	IO_CHECKMALLOC(parentImg->fields.SubIFDs =
	    TFX_MALLOC(parentImg->fields.SubIFDCount *
	    TfxFieldTypeSize[IFDLONG]));
	memset(parentImg->fields.SubIFDs, 0,
	    parentImg->fields.SubIFDCount * TfxFieldTypeSize[IFDLONG]);

	parentImg->fields.SubIFDs[0] = childIFDOffset;
    }

    /* Retain the position of the last in the string */
    parentImg->lastSubImg = tfxImg;


    /* Update the parent image */
    /* XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX */


    /* Update the page statistics */
    /* XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX */


    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}



/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Reattach a rewritten IFD
 *
 * LongDescription:::
 *  wp_TifFXReattachIFD() writes the "nextIFD" or SubIFD entry to accomodate the move
 *  of an IFD to a new location.
 *
 * Input:::
 *  tfxImg::			Data structure of image that moved
 *  oldIFDOffset::		Offset of IFD before move
 *  newIFDOffset::		Offset of IFD after move
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXReattachIFD(
		TFX_IMG *tfxImg,
		UInt32 oldIFDOffset,
		UInt32 newIFDOffset
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt32 lval;
    UInt32 addressToPrevNextIFD = 0;	/* nextIFD (or SubIFD) array location */
    UInt32 method = 0;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImg);

    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);


    /* Keep looking until we have a location to link from */
    while (addressToPrevNextIFD == 0) {

	/* Try each approach, until we've found what we're looking for */
	switch (method++) {
	case 0:
	    /* Look first at the previous image */
	    if (tfxImg->prev != NULL) {
		if (tfxImg->prev->fields.nextIFD == oldIFDOffset) {
		    addressToPrevNextIFD = tfxImg->prev->ifdOffset + 2 +
			tfxImg->prev->fields.fieldCount * TFX_TIFFFIELDSIZE;

		    /* Store the new value */
		    tfxImg->prev->fields.nextIFD = newIFDOffset;
		    break;
		}
	    }
	    break;
	case 1:
	    /* Next try all the subimages of the parent */
	    if (tfxImg->imgInfo.tfxImgType != IO_IMGTYPE_PRIMARY) {
		TFX_IMG *childImg = tfxImg->tfxPage->tfxImgs;

		while (childImg) {
		    if (childImg->fields.nextIFD == oldIFDOffset) {
			XERR_GEN_CLEANUP(childImg->fields.nextIFD == 0,
			    IO_INTERNAL_ERROR);
			addressToPrevNextIFD = childImg->ifdOffset + 2 +
			    childImg->fields.fieldCount * TFX_TIFFFIELDSIZE;

			/* Store the new value */
			childImg->fields.nextIFD = newIFDOffset;
			break;
		    }
		    childImg = childImg->next;
		}
	    }
	    break;
	case 2:
	    /* Next try the SubIFDs array itself */
	    if (tfxImg->imgInfo.tfxImgType != IO_IMGTYPE_PRIMARY) {
		TFX_IMG *parentImg = (TFX_IMG *)tfxImg->tfxPage->pageInfo.tfxPrimImgHandle;
		UInt32 chainNum;

		for (chainNum = 0; chainNum < parentImg->fields.SubIFDCount; chainNum++) {
		    if (parentImg->fields.SubIFDs[chainNum] == oldIFDOffset) {
			XERR_GEN_CLEANUP(parentImg->fieldDataLocation.SubIFDs == 0,
			    IO_INTERNAL_ERROR);
			addressToPrevNextIFD = parentImg->fieldDataLocation.SubIFDs + chainNum * sizeof(long);

			/* Store the new value */
			parentImg->fields.SubIFDs[chainNum] = newIFDOffset;
			break;
		    }
		}
	    }
	    break;
	case 3:
	    /* Then try the primary IFD chain */
	    if (tfxImg->imgInfo.tfxImgType == IO_IMGTYPE_PRIMARY) {
		TFX_IMG *prevImg;
		TFX_IFDOFFSET_ITEM *ifdp = NULL;

		if (tfxImg->tfxPage->prev) {
		    prevImg = (TFX_IMG *)tfxImg->tfxPage->prev->pageInfo.tfxPrimImgHandle;
		    if (prevImg->fields.nextIFD == oldIFDOffset) {
			addressToPrevNextIFD = prevImg->ifdOffset + 2 +
			    prevImg->fields.fieldCount * TFX_TIFFFIELDSIZE;

			/* Find this item in the list */
			ifdp = tfxImg->tfxPage->tfxDoc->ifdList;
			while (ifdp && ifdp->ifdOffset != oldIFDOffset) {
			    ifdp = ifdp->next;
			}

			/* Make sure we found it */
			XERR_GEN_CLEANUP(ifdp == NULL || ifdp->ifdOffset != oldIFDOffset,
			    IO_INTERNAL_ERROR);

			/* Store the new value */
			ifdp->ifdOffset = newIFDOffset;
			prevImg->fields.nextIFD = newIFDOffset;
			break;
		    }
		}
	    }
	    break;
	case 4:
	    /* Lastly, it must be the first primary IFD */
	    XERR_GEN_CLEANUP(tfxImg->imgInfo.tfxImgType != IO_IMGTYPE_PRIMARY,
		IO_INTERNAL_ERROR);
	    addressToPrevNextIFD = 4;
	    break;
	default:
	    addressToPrevNextIFD = 1;
	}
    }

    /* We have an active parent image, with no SubIFDs (yet).
     * Attach to the SubIFDs field of the parent */
    lval = newIFDOffset;
    TFX_SEEK(tfxFile, addressToPrevNextIFD);
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}



/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add a strip to the strip structure
 *
 * LongDescription:::
 *  wp_TifFXAddStripInfo() appends a new strip to the end of the strip
 *  structure, and increments the strip count.  This structure maintains
 *  information about the strips, when variable-sized strips are used.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  stripRowOffset::		Offset from the top of the image where
 *				the row begins (0-based)
 *  stripRowCount::		Number of rows in the new strip
 *  stripBytes::		Number of bytes in the compressed image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXAddStripInfo(
	TFX_IMG *tfxImg,
	UInt32 stripRowOffset,
	UInt32 stripRowCount,
	UInt32 stripBytes
)
{
    TFX_INSTANCE *instData;
    TFX_STRIP *newStrip;
    PROLOGUE;

    TFX_CheckHandle(tfxImg);
    instData = TFX_IMGINST(tfxImg);

    /* We must be using the StripRowCounts field.
     * Augment the list of strip rows (we can't write them
     * until we know how many there are, and we don't know this
     * until the last scanline is written)
     */
    IO_CHECKMALLOC(newStrip = IO_MALLOC(sizeof(*newStrip)));
    memset(newStrip, 0, sizeof(newStrip));
    newStrip->offset = stripRowOffset;
    newStrip->bytes = stripBytes;
    newStrip->rows = stripRowCount;

    /* Link it in at the end */
    if (tfxImg->strips.lastStrip) {
	tfxImg->strips.lastStrip->next = newStrip;
	tfxImg->strips.lastStrip = newStrip;
    } else {
	tfxImg->strips.firstStrip = tfxImg->strips.lastStrip = newStrip;
    }

    /* Update the tally for all the strips */
    tfxImg->strips.rows += stripRowCount;
    tfxImg->strips.stripCount++;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a strip of data and update the offset/byteCount field entries
 *
 * LongDescription:::
 *  wp_TifFXWriteStrip() write a strip of (usually compressed) data, and updates
 *  the StripOffsets and ByteCounts fields.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXWriteStrip(
	      TFX_IMG *tfxImg
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    Int32 offset = 0;
    UInt32 stripNum;
    UInt32 seekOffset;
    UInt32 lval;
    PROLOGUE;

    TFX_CheckHandle(tfxImg);
    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);

    /* Make sure it's a non-zero byte strip */
    if (tfxImg->cmpBytes > 0) {

	/* Align the file to a long boundary */
	XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &offset));

	/* Write the data */
	TFX_WRITE(tfxFile, tfxImg->cmpBuffer, (Int32)tfxImg->cmpBytes);
    }

    XERR_GEN_CLEANUP(tfxImg->imgInfo.dwBestStripHeight != tfxImg->fields.RowsPerStrip,
	IO_INTERNAL_ERROR);

    if (tfxImg->fields.RowsPerStrip == 0) {
	UInt32 rowCount;

	if (tfxImg->tmpPixrRows > 0) {
	    rowCount = tfxImg->tmpPixrRows;
	} else {
	    UInt32 stripOffset, stripRow, stripHeight;

	    /* We have to go figure out how many it was */
	    XERR_PASS_CLEANUP(wp_TfxGetStripInfo(tfxImg, tfxImg->strips.rows,
		&stripOffset, &stripRow, &stripHeight));

	    /* Make sure it was a strip boundary */
	    XERR_GEN_CLEANUP(stripRow != tfxImg->strips.rows, IO_INTERNAL_ERROR);

	    /* Must be the entire image */
	    rowCount = tfxImg->imgInfo.dwHeight;
	}

	/* Add the strip to the strip list */
	XERR_PASS_CLEANUP(wp_TifFXAddStripInfo(tfxImg, offset, rowCount, tfxImg->cmpBytes));

	/* If we just wrote the last strip, fill in the StripRowCounts field.
	 * Note, this should get called in TFX_ImageWriteFinish() also, but
	 * let's do it now anyways. */
	if (tfxImg->strips.rows >= tfxImg->imgInfo.dwHeight) {
	    XERR_PASS_CLEANUP(wp_TifFXWriteStripSizes(tfxImg));
	}

    } else {
	/* Write the strip offset and byte count to the arrays */
	stripNum = tfxImg->tmpPixrRow / tfxImg->imgInfo.dwBestStripHeight;
	seekOffset = tfxImg->fieldDataLocation.StripOffsets + stripNum * sizeof(long);
	XERR_GEN_CLEANUP(seekOffset == 0, IO_INTERNAL_ERROR);
	TFX_SEEK(tfxFile, seekOffset);
	lval = offset;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
	seekOffset = tfxImg->fieldDataLocation.StripByteCounts + stripNum * sizeof(long);
	XERR_GEN_CLEANUP(seekOffset == 0, IO_INTERNAL_ERROR);
	TFX_SEEK(tfxFile, seekOffset);
	lval = tfxImg->cmpBytes;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
	
	/* Update the field values */
	if (tfxImg->fields.StripOffsets == NULL ||
	    tfxImg->fields.StripByteCounts == NULL) {
	    XERR_GEN_CLEANUP((tfxImg->fields.TileOffsets == NULL ||
		tfxImg->fields.TileByteCounts == NULL), IO_INTERNAL_ERROR);
	    
	    tfxImg->fields.TileOffsets[stripNum] = offset;
	    tfxImg->fields.TileByteCounts[stripNum] = tfxImg->cmpBytes;
	} else {
	    tfxImg->fields.StripOffsets[stripNum] = offset;
	    tfxImg->fields.StripByteCounts[stripNum] = tfxImg->cmpBytes;
	}
    }

#if 0
    /* Also update the JPEG Interchange values */
    if (stripNum == 0 && tfxImg->imgInfo.xCompression == IO_COMP_JPEG) {
	seekOffset = tfxImg->interchangeLocation;
	XERR_GEN_CLEANUP(seekOffset == 0, IO_INTERNAL_ERROR);
        TFX_SEEK(tfxFile, seekOffset);
        lval = offset;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
	seekOffset = tfxImg->interchangeLengthLocation;
	XERR_GEN_CLEANUP(seekOffset == 0, IO_INTERNAL_ERROR);
	TFX_SEEK(tfxFile, seekOffset);
	lval = tfxImg->cmpBytes;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
    }
#endif
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Finish up the variable-sized strips
 *
 * LongDescription:::
 *  wp_TifFXWriteStripSizes() writes the final information about the strips,
 *  when the strip count is finally known.  This include: StripOffsets,
 *  StripByteCounts, and StripRowCounts.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXWriteStripSizes(
	      TFX_IMG *tfxImg
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    Int32 offset;
    UInt32 stripNum;
    UInt32 seekOffset;
    UInt32 lval;
    TFX_STRIP *strip;
    UInt32 stripRowCountsLocation = 0;
    PROLOGUE;

    TFX_CheckHandle(tfxImg);
    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);

    if (tfxImg->fields.RowsPerStrip > 0) {
	NORMAL_RETURN;
    }

    /* We should have enough strips to fill the image */
    XERR_GEN_CLEANUP(tfxImg->strips.rows != tfxImg->imgInfo.dwHeight, IO_INTERNAL_ERROR);

    /* Transfer the strip count */
    tfxImg->fields.StripsPerImage = tfxImg->strips.stripCount;

    /* Allocate an array for the StripOffsets */
    IO_CHECKMALLOC(tfxImg->fields.StripOffsets =
	TFX_MALLOC(tfxImg->fields.StripsPerImage *
	TfxFieldTypeSize[IFDLONG]));
    memset(tfxImg->fields.StripOffsets, 0, TfxFieldTypeSize[IFDLONG] *
	tfxImg->fields.StripsPerImage);
	
    /* Write the StripOffsets array */
    if (tfxImg->strips.stripCount == 1) {
	seekOffset = tfxImg->fieldDataLocation.StripOffsets;
	XERR_GEN_CLEANUP(seekOffset == 0, IO_INTERNAL_ERROR);
	TFX_SEEK(tfxFile, seekOffset);
    } else {
	/* Put it at the end of the file */
	XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &offset));
	tfxImg->fieldDataLocation.StripOffsets = offset;
    }
    strip = tfxImg->strips.firstStrip;
    for (stripNum = 0; stripNum < tfxImg->strips.stripCount;
    stripNum++, strip = strip->next) {
	lval = strip->offset;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
	tfxImg->fields.StripOffsets[stripNum] = strip->offset;
    }
    
    /* Allocate an array for the StripByteCounts */
    IO_CHECKMALLOC(tfxImg->fields.StripByteCounts =
	TFX_MALLOC(tfxImg->fields.StripsPerImage *
	TfxFieldTypeSize[IFDLONG]));
    memset(tfxImg->fields.StripByteCounts, 0,
	TfxFieldTypeSize[IFDLONG] * tfxImg->fields.StripsPerImage);
    
    /* Write the StripByteCounts array */
    if (tfxImg->strips.stripCount == 1) {
	seekOffset = tfxImg->fieldDataLocation.StripByteCounts;
	XERR_GEN_CLEANUP(seekOffset == 0, IO_INTERNAL_ERROR);
	TFX_SEEK(tfxFile, seekOffset);
    } else {
	/* Put it at the end of the file */
	XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &offset));
	tfxImg->fieldDataLocation.StripByteCounts = offset;
    }
    strip = tfxImg->strips.firstStrip;
    for (stripNum = 0; stripNum < tfxImg->strips.stripCount;
    stripNum++, strip = strip->next) {
	lval = strip->bytes;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
	tfxImg->fields.StripByteCounts[stripNum] = strip->bytes;
    }
    
    /* Allocate an array for the StripRowCounts */
    IO_CHECKMALLOC(tfxImg->fields.StripRowCounts =
	TFX_MALLOC(tfxImg->fields.StripsPerImage *
	TfxFieldTypeSize[IFDLONG]));
    memset(tfxImg->fields.StripRowCounts, 0, TfxFieldTypeSize[IFDLONG] *
	tfxImg->fields.StripsPerImage);
    
    /* Write the StripRowCounts array */
    if (tfxImg->strips.stripCount == 1) {
	seekOffset = tfxImg->fieldLocation.StripRowCounts;
	XERR_GEN_CLEANUP(seekOffset == 0, IO_INTERNAL_ERROR);
	TFX_SEEK(tfxFile, seekOffset);
    } else {
	/* Put it at the end of the file */
	XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &offset));
	stripRowCountsLocation = offset;
    }
    strip = tfxImg->strips.firstStrip;
    for (stripNum = 0; stripNum < tfxImg->strips.stripCount;
    stripNum++, strip = strip->next) {
	lval = strip->rows;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
	tfxImg->fields.StripRowCounts[stripNum] = strip->rows;
    }
    
    /* Fill in the number of strips, and the array location,
     * in the StripOffsets field */
    seekOffset = tfxImg->fieldLocation.StripOffsets;
    XERR_GEN_CLEANUP(seekOffset == 0, IO_INTERNAL_ERROR);
    seekOffset -= 4;	/* Go back and fill the count in too */
    TFX_SEEK(tfxFile, seekOffset);
    lval = tfxImg->strips.stripCount;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));
    if (tfxImg->strips.stripCount > 1) {
        /* Location of the array (if array of one, the offset field
	 * has the single value, filled in above) */
	lval = tfxImg->fieldDataLocation.StripOffsets;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
    }
    
    /* Fill in the number of strips, and the array location,
     * in the StripByteCounts field */
    seekOffset = tfxImg->fieldLocation.StripByteCounts;
    XERR_GEN_CLEANUP(seekOffset == 0, IO_INTERNAL_ERROR);
    seekOffset -= 4;	/* Go back and fill the count in too */
    TFX_SEEK(tfxFile, seekOffset);
    lval = tfxImg->strips.stripCount;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));
    if (tfxImg->strips.stripCount > 1) {
       /* Location of the array (if array of one, the offset field
	* has the single value, filled in above) */
	lval = tfxImg->fieldDataLocation.StripByteCounts;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
    }
    
    /* Fill in the number of strips, and the array location,
     * in the StripRowCounts field */
    seekOffset = tfxImg->fieldLocation.StripRowCounts;
    XERR_GEN_CLEANUP(seekOffset == 0, IO_INTERNAL_ERROR);
    seekOffset -= 4;	/* Go back and fill the count in too */
    TFX_SEEK(tfxFile, seekOffset);
    lval = tfxImg->strips.stripCount;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));
    if (tfxImg->strips.stripCount > 1) {
        /* Location of the array (if array of one, the offset field
	 * has the single value, filled in above) */
	lval = stripRowCountsLocation;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
    }
    
    /* Free the list of strip sizes */
    {
	TFX_STRIP *strip, *nextStrip;
	
	strip = tfxImg->strips.firstStrip;
	while (strip) {
	    nextStrip = strip->next;
	    IO_FREE(strip);
	    strip = nextStrip;
	}
	memset(&tfxImg->strips, 0, sizeof(tfxImg->strips));
    }

    NORMAL_RETURN;
    
 ERR_CLEANUP:

    ERR_RETURN;
}
	
#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Update the ImageLayerContent TIFF field
 *
 * LongDescription:::
 *  wp_TifFXUpdateILContent() update the contents of the ImageLayerContent TIFF
 *  field.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXUpdateILContent(
		   TFX_IMG *tfxImg
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_PAGE *tfxPage = tfxImg->tfxPage;
    UInt32 lval, *lvals = NULL;
    UInt32 layer;
    TFX_TIFFIELDS *fieldList;
    PROLOGUE;

    fieldList = &tfxImg->fields;

    if (tfxImg->fields.dataTransferred == 0 ||
	tfxImg->tfxPage->tfxDoc->docInfo.version < 3) {

	/* The field hasn't been written yet, no need to update now */
	NORMAL_RETURN;
    }

    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);

    /* If the number of stored image counts has increased, we need to
     * rewrite the data to a new location */
    if (tfxPage->pageInfo.dwLayersPresent > fieldList->ImageLayerContent.layersPresent) {
	XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &tfxImg->imgLayerContentDataOffset));
	TFX_SEEK(tfxFile, tfxImg->fieldLocation.ImgLayerContent - sizeof(UInt32));
	lval = tfxPage->pageInfo.dwLayersPresent;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
	lval = tfxImg->imgLayerContentDataOffset;
	TFX_SWAPLONG(tfxFile, &lval);
	TFX_WRITE(tfxFile, &lval, sizeof(lval));
	fieldList->ImageLayerContent.layersPresent = tfxPage->pageInfo.dwLayersPresent;
    }

    IO_CHECKMALLOC(lvals = (UInt32 *)TFX_MALLOC(
	fieldList->ImageLayerContent.layersPresent * sizeof(UInt32)));
    for (layer = 0; layer < fieldList->ImageLayerContent.layersPresent; layer++) {
	lvals[layer] = tfxPage->pageInfo.pdwImageCounts[layer];
    }

    if ((UInt8)TFX_FILETOKEN(tfxFile)->bSwapBytes) {
	XERR_PASS_CLEANUP(wp_TifFXConvcopy(lvals, lvals, IFDLONG,
	    fieldList->ImageLayerContent.layersPresent,
	    (UInt8)TFX_FILETOKEN(tfxFile)->bSwapBytes));
    }

    TFX_SEEK(tfxFile, tfxImg->imgLayerContentDataOffset);
    TFX_WRITE(tfxFile, lvals, sizeof(lvals) * fieldList->ImageLayerContent.layersPresent);
    TFX_FREE(lvals);
    lvals = NULL;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (lvals) {
	TFX_FREE(lvals);
	lvals = NULL;
    }
    
    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Delete an image from a page
 *
 * LongDescription:::
 *  wp_TifFXDeleteImage() deletes (removes all references to) an single image
 *  on a page.  The image data is not physically removed (as the image is
 *  likely to be inbetween other file fragments), but the references to the
 *  image are removed.  This includes: MRC layer image counts and relinking
 *  the previous image to the subsequent image via the NextIFD IFD entry.
 *  <BR>
 *  Note that although the image is not physically removed, since the
 *  references to it are removed (and lost), there is no way to 'undo' an
 *  image deletion.
 *  <BR>
 *  The intent is to allow physically editing the file, with anticipation
 *  that later the file will be copied (see TFX_CopyPage() and
 *  TFX_CopyImage()), allowing the disowned file fragments to remain behind.
 *
 * Input:::
 *  tfxImg::			Image to be removed
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TifFXDeleteImage(
	TFX_IMG *tfxImg
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_IMG *tfxParentImg = NULL;		/* Parent image */
    TFX_IMG *childImg;
    TFX_PAGE *tfxPage = tfxImg->tfxPage;
    TFX_DOC *tfxDoc = tfxPage->tfxDoc;
    TFX_IMGINFO *pTfxImgInfo;
    UInt32 lval;
    PROLOGUE;

    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    pTfxImgInfo = &tfxImg->imgInfo;

    /* Deleting the primary image from a page is invalid, since the page
     * has the primary image as it's foundation (without it, there is no
     * page). What is probably meant here is a call to TFX_DeletePage().
     */
    XERR_GEN(pTfxImgInfo->tfxImgType == IO_IMGTYPE_PRIMARY, IO_BADPARAMETER);
    XERR_GEN(tfxImg == tfxPage->tfxImgs, IO_BADPARAMETER);

    /* Either this is the primary image, or there should be one. Primary
     * image must be called layer 1, 2 or 3. */
    XERR_GEN(tfxPage->pageInfo.tfxPrimImgHandle == 0 &&
	     (pTfxImgInfo->dwImgLayer != 1 || pTfxImgInfo->dwImgLayer != 2 ||
	      pTfxImgInfo->dwImgLayer != 3) && pTfxImgInfo->dwImgOffset != 0,
	     IO_SEQUENCE);
    tfxParentImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    /* Check for proper calling sequence */
    XERR_GEN((tfxDoc->readWrite & WRITE_ALLOWED) == 0, IO_SEQUENCE);

    /* Unlink it from the image(s) that follows */
    if (tfxImg->next) {

	/* Traverse the following images, and decrement the image offset
	 * of any that reside on the same level, and have a subsequent offset
	 */
	childImg = tfxImg;  /* Start here, but immediately go to the next */
	while ((childImg = childImg->next) != NULL) {
	    if (childImg->imgInfo.dwImgLayer == pTfxImgInfo->dwImgLayer &&
		childImg->imgInfo.dwImgOffset > pTfxImgInfo->dwImgOffset) {
		childImg->imgInfo.dwImgOffset--;

		/* Rewrite the ImageLayer field item: image offset */
		TFX_SEEK(tfxFile, childImg->fieldDataLocation.ImageLayer + sizeof(long));
		lval = childImg->imgInfo.dwImgOffset + 1;	/* 1-based */
		TFX_SWAPLONG(tfxFile, &lval);
		TFX_WRITE(tfxFile, &lval, sizeof(long));
	    }
	}
    }

    /* Point the previous image to the next one, via the NextIFD IFD element */
    if (tfxImg->prev) {

	/* Rewrite the NextIFD for the elder sibling (or parent) */
	XERR_PASS_CLEANUP(wp_TifFXWriteNextIFD(tfxImg->prev, tfxImg->prev->prev,
					     tfxImg->prev->ifdOffset,
					     tfxImg->fields.nextIFD));
	
    } else {

	/* There must only be one if there's no backwards link, which
	 * is in error, since we checked above for deleting the primary
	 * image. */
	tfxPage->tfxImgs = NULL;
	XERR_GEN_CLEANUP(tfxPage->tfxImgs == tfxImg, IO_INTERNAL_ERROR);
    }

    /* Decrement the ImageLayerContent field value for the layer */
    XERR_PASS_CLEANUP(wp_TfxIncrementImageCounts(tfxPage, pTfxImgInfo->dwImgLayer,
						 -1));

/*    XERR_PASS_CLEANUP(wp_TifFXUpdateILContent(tfxParentImg)); */

    /* Unlink it from the page list */
    if (tfxImg->prev) {
	tfxImg->prev->next = tfxImg->next;
	tfxImg->next->prev = tfxImg->prev;
    } else {

	/* There must only be one if there's no backwards link, which
	 * is in error, since we checked above for deleting the primary
	 * image. */
	tfxPage->tfxImgs = NULL;
	XERR_GEN_CLEANUP(tfxPage->tfxImgs == tfxImg, IO_INTERNAL_ERROR);
    }

    /* Free the image structure */
    XERR_PASS_CLEANUP(wp_TfxImgFree(&tfxImg, 0));

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


#if 0
/* It's been talked about that this bit should remain on, regardless of the
 * number of pages.
 */
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  If the page count is only one, remove the multi-page bit from the NewSubfileType
 *
 * LongDescription:::
 *  wp_TfxTestMultipageBit() sets or clears the multi-page bit in NewSubfileType,
 *  according to whether only one page survives the file.
 *
 * Input:::
 *  tfxDocHandle::		Document handle
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxTestMultipageBit(
	IO_DOCHANDLE tfxDochandle
)
{
    TFX_FILE *tfxFile;
    TFX_IMG *tfxImg = NULL;		/* Parent image */
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDochandle;
    UInt32 lval;
    PROLOGUE;

    TFX_CheckHandle(tfxDoc);

    tfxFile = TFX_DOCFILE(tfxDoc);

    if (tfxDoc->docInfo.wNPages < 1)
	NORMAL_RETURN;

    /* Get the primary image of the first page */
    tfxImg = (TFX_IMG *)tfxDoc->pages->pageInfo.tfxPrimImgHandle;

    /* Make sure there even is a NewSubfileType, and that we're able to write */
    if (tfxImg == NULL || tfxImg->fieldDataLocaion.NewSubfileType == 0
	|| (tfxImg->tfxPage->tfxDoc->readWrite & WRITE_ALLOWED) == 0) {
	NORMAL_RETURN;
    }

    /* Either this is the primary image, or there should be one. Primary
     * image must be called layer 1, 2 or 3. */
    if (tfxDoc->docInfo.wNPages == 1) {
	/* Bit 2 should be cleared */
	if (((tfxImg->fields.NewSubfileType & TFX_TIFFSUBFILE_MULTIPAGE)
	    != 0)) {
	    /* Bit is set, and should be (it's really a 1 page file, and some
	     * have expressed that a 1-page TIFF shouldn't have the multi-page
	     * bit set. */
	    tfxImg->fields.NewSubfileType &= ~TFX_TIFFSUBFILE_MULTIPAGE;
	} else {
	    NORMAL_RETURN;
	}
    } else {
	/* Bit 2 should be set */
	if ((tfxImg->fields.NewSubfileType & TFX_TIFFSUBFILE_MULTIPAGE) == 0) {
	    tfxImg->fields.NewSubfileType |= TFX_TIFFSUBFILE_MULTIPAGE;
	} else {
	    NORMAL_RETURN;
	}
    }

    /* Write the new value */
    lval = tfxImg->fields.NewSubfileType;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_SEEK(tfxFile, tfxImg->fieldDataLocation.NewSubfileType);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Update the CodingMethods if out-of-date
 *
 * LongDescription:::
 *  wp_TfxTestCodingMethods() checks to see if the CodingMethods field in the GPIFD
 *  is up-to-date, and rewrites it if it isn't.
 *
 * Input:::
 *  tfxDocHandle::		Document handle
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxTestCodingMethods(
	IO_DOCHANDLE tfxDochandle
)
{
    TFX_FILE *tfxFile;
    TFX_IMG *tfxImg = NULL;		/* Parent image */
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDochandle;
    UInt32 lval;
    PROLOGUE;

    TFX_CheckHandle(tfxDoc);

    tfxFile = TFX_DOCFILE(tfxDoc);

    if (tfxDoc->docInfo.wNPages < 1)
	NORMAL_RETURN;

    /* Get the primary image of the first page */
    tfxImg = (TFX_IMG *)tfxDoc->pages->pageInfo.tfxPrimImgHandle;

    /* Make sure there even is a GPIFD with a CodingMethods field,
     * and that we're able to write */
    if (tfxImg == NULL || tfxDoc->codingMethodsLocation == 0
	|| (tfxImg->tfxPage->tfxDoc->readWrite & WRITE_ALLOWED) == 0) {
	NORMAL_RETURN;
    }

    /* If the field is up-to-date, then we don't need to do anything. */
    if ((UInt32)tfxDoc->codingMethods == tfxImg->fields.gpIFD.CodingMethods) {
	NORMAL_RETURN;
    }

    /* Write the new value */
    lval = tfxImg->fields.gpIFD.CodingMethods = tfxDoc->codingMethods;
    TFX_SWAPLONG(tfxFile, &lval);
    TFX_SEEK(tfxFile, tfxDoc->codingMethodsLocation);
    TFX_WRITE(tfxFile, &lval, sizeof(lval));

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Rewrite any IFDs that have been added to
 *
 * LongDescription:::
 *  wp_TfxRewriteIFDs() writes new IFDs for any images that have had new fields
 *  added to them since they were first written.
 *
 * Input:::
 *  tfxDocHandle::		Document handle
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxRewriteIFDs(
	IO_DOCHANDLE tfxDochandle
)
{
    TFX_FILE *tfxFile;
    TFX_IMG *tfxImg = NULL;
    TFX_PAGE *tfxPage = NULL;
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDochandle;
    UInt32 prevIFDOffset = 0;
    PROLOGUE;

    TFX_CheckHandle(tfxDoc);

    tfxFile = TFX_DOCFILE(tfxDoc);

    if (tfxDoc->docInfo.wNPages < 1) {
	NORMAL_RETURN;
    }

    /* Loop through the pages */
    for (tfxPage = tfxDoc->pages; tfxPage; tfxPage = tfxPage->next) {

	/* Start with the primary IFD, then go through the list */
	tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

	/* Loop through the images */
	while (tfxImg) {
	    if (tfxImg->fields.IFDrewriteNeeded) {

		/* Keep track of what the IFD location was */
		prevIFDOffset = tfxImg->ifdOffset;

		/* Reset the number of fields actually written */
		tfxImg->reservedFieldCount = 0;

		/* Rewrite the IFD */
		XERR_PASS_CLEANUP(wp_TifFXWriteIFD(tfxImg));

		/* Reattach to the previous IFD */
		XERR_PASS_CLEANUP(wp_TifFXReattachIFD(tfxImg, prevIFDOffset, tfxImg->ifdOffset));

	    }
	    /* Go to the next image, or the beginning of the list */
	    if (tfxImg->handleID == tfxPage->pageInfo.tfxPrimImgHandle) {
		tfxImg = tfxPage->tfxImgs;
	    } else {
		tfxImg = tfxImg->next;
	    }
	}
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/*
$Log: tfxwtiff.c,v $
Revision 1.3  2003/02/11 01:31:43  mkrocken
fixed problems with comments caused by previous commit

Revision 1.2  2003/02/10 22:53:33  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:15  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.47  2001/12/26 21:21:07  dabercro
Add a NULL byte in ASCII types automatically.
Bug fix: Only strip JPEGTables from JPEG streams (was stripping off all
streams in JPEGTables mode)

Revision 1.46  2001/12/12 00:05:57  dabercro
Special case the removal of JPEGTables from compressed data

Revision 1.45  2001/11/14 00:07:15  dabercro
Support for 16-bit CIELAB requires the 16th bit for the MSB on L* component,
we must change support for CIELAB to SLONG, not SSHORT.

Revision 1.44  2001/11/07 21:48:41  dabercro
Allow IFDBaseColor to be SSHORT.  Keep track of which type it is.

Revision 1.43  2001/10/30 20:05:07  dabercro
Add in writing of MH and MR images.

Revision 1.42  2001/09/18 21:10:12  dabercro
Make work with JBIG2

Revision 1.41  2001/08/22 19:50:16  dabercro
Added support for new field: IFDBaseColor.
Output JPEGTables only when JPEG used

Revision 1.40  2001/07/05 21:46:06  dabercro
Support for the JPEGTables field

Revision 1.39  2001/06/29 22:34:36  dabercro
Fix the PhotometricInterpretation assignment logic when using NOC

Revision 1.38  2001/06/27 18:52:17  dabercro
Minor fixes, add uncompressed support.

Revision 1.37  2001/06/01 21:03:28  dabercro
Support centimeters in ResolutionUnit.

Revision 1.36  2001/05/22 18:55:21  dabercro
Reorder ImageLayer write to follow TIFF convention.

Revision 1.35  2001/05/08 23:38:42  dabercro
Fix the parameters in a seldom-called section of code.

Revision 1.34  2001/05/03 16:43:10  dabercro
Added a new function UseMultiStripSubIFDs.  Rethink logic about
automatically generating multi-strip SubIFDs.

Revision 1.33  2001/05/01 00:59:57  dabercro
Add UseNoGPIFD.

Revision 1.32  2001/04/02 18:25:04  dabercro
Add HostComputer.

Revision 1.31  2001/03/13 02:59:13  dabercro
Remove lossy-dither and wavelet as valid compression types.

Revision 1.30  2001/03/13 02:20:37  dabercro
Remove extra field write for ChromaPositioning.

Revision 1.29  2001/02/26 19:15:37  dabercro
Typecast for compare (don't know what typedef is casting to)

Revision 1.28  2001/01/18 00:21:32  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.27  2000/12/11 22:23:24  dabercro
Modifications to quick-copy in recommended TIFF-FX file order.

Revision 1.26  2000/11/17 23:57:36  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.25  2000/11/02 22:21:22  dabercro
Additions for Rendering Hints.

Revision 1.24  2000/10/19 17:12:29  dabercro
Finished support for color JBIG2 component.

Revision 1.23  2000/10/18 20:30:45  dabercro
Use the PhotometricInterpretation specifically set by the user.
Also avoid reallocating the color palette.

Revision 1.22  2000/10/16 18:22:54  dabercro
Major adds to support stripes, 0-byte strips, and ImageBaseColor.

Revision 1.21  2000/09/27 21:56:45  dabercro
Variable strips and stripe work.

Revision 1.20  2000/09/22 21:50:21  dabercro
Variable-sized strips now handle multiple layers.

Revision 1.19  2000/09/21 21:19:15  dabercro
Add variable-sized strips: StripRowCounts.

Revision 1.18  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.17  2000/08/30 23:29:08  dabercro
Check for both JBIG2 and Tokenization.

Revision 1.16  2000/08/29 22:51:25  dabercro
Allow other compression types.
Leave multi-page bit on.

Revision 1.15  2000/08/25 23:56:17  dabercro
Fixes for "copy compressed".

Revision 1.14  2000/08/22 01:03:24  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.13  2000/08/08 00:26:34  dabercro
Pass the right pointers for ASCII fields.

Revision 1.12  2000/08/04 21:57:55  dabercro
Use page offset of previous occupant, when inserting a page.

Revision 1.11  2000/08/02 22:44:27  dabercro
Make sure the file is writable before writing to it.

Revision 1.10  2000/08/02 00:23:16  dabercro
Fix fillorder field.  Place-holder for requests from Mark Tobin.

Revision 1.9  2000/08/01 00:12:58  dabercro
Updates to the handling of shared resources.
Additions to support JBIG profiles.
Additions for TIFF-FX Extension 1.

Revision 1.8  2000/06/07 00:00:18  dabercro
Continued work, lots of stuff.

Revision 1.7  2000/05/10 22:49:02  dabercro
Add JBIG2 as a real compression type (in addition to Xerox tokenization)

Revision 1.6  2000/05/09 01:52:58  dabercro
Added FillOrder field support.
Updated JBIG2 dictionary front-end.
Support for JBIG2 dictionary extensions, updated parameter to
reflect byte count for extensions.
Write fields that are explicitly set, regardless of default values.
Check if document is single-page when closing, reflect in
NewSubfileType.

Revision 1.5  2000/04/12 19:06:10  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.4  2000/03/11 02:06:35  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.3  2000/02/29 00:19:42  dabercro
Incremental improvements, profile distinctions,etc.

Revision 1.2  2000/02/08 21:25:34  dabercro
Disable JBIG2 (for now), updates to make use of GlobalParametersIFD for symbol libraries.

Revision 1.1.1.1  2000/02/02 00:40:52  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.21  1998/07/30 22:42:53  aber
aber on Thu Jul 30 15:42:24 PDT 1998

tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

Allow color tokenized images, call them RGB.

Revision 1.20  1998/04/02 00:09:55  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

CommonCheck: Allow attaching a transparency mask to the primary image, special request.

Revision 1.19  1998/01/21 22:56:10  aber
aber on Wed Jan 21 14:54:57 PST 1998

tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

Change logic to allow a binary image to have a mask, in an odd-numbered
layer only. Note that this formally breaks MRC.

Revision 1.18  1998/01/20 03:43:53  aber
aber on Mon Jan 19 19:43:22 PST 1998

tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

Prototype update to allow deleting last page of a document.

Revision 1.17  1998/01/20 03:24:17  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

Rename MixedRasterContent field to ImageLayerContent, per Lloyd's suggestion.

Revision 1.16  1998/01/16 00:43:57  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

Allow page inserts.

Revision 1.15  1997/12/18 00:24:50  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

Keep track of the ImageLayer field data location.
Make sure the nextIFD value is kept internally.
Keep track of the strip sizes and locations (required when in read/write)

Revision 1.14  1997/11/24 21:14:07  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

Return the file pointer to the end of the field list, so that when
writing the NextIFD, it written to the right place. This only happens
when writing the MixedRasterContent field, when the field is updated
after the subimages have been attached (tokenized with subimages).

Revision 1.13  1997/08/29 17:28:08  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.12  1997/07/10 22:02:56  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

Keep track of the annotation count within the TIFF field.

Revision 1.11  1997/07/02 17:04:58  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

Make the right return value check.

Revision 1.10  1997/07/02 01:07:24  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

Use private function names and check return values for type
conversion.

Revision 1.9  1997/06/26 22:33:09  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

Return a new parameter if the file is not a TIF.
Allow multiple tiles (JPEG/Wavelet).

Revision 1.8  1997/06/17 22:37:15  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

Fixed bug (editing error) in last checkin.

Revision 1.6  1997/06/10 01:27:02  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.5  1997/05/30 00:51:20  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:23  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:32  aber
tfxwtiff.c -> /product/ipcore/ipwrap/src/RCS/tfxwtiff.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:49  aber
Initial revision

*/
