#ifndef REORIENT_H
#define REORIENT_H

#include "dgp_state.h"
#include "orientation_t.h"

#ifdef __cplusplus
extern "C"
{
#endif


/* reorient.h
   higher level primitives that use the rotate primitives; presents an 
   interface for converting images from one DgpCoord_Orientation to another.
*/


/* reorientBinaryNeeded()
   If I'm planning on reorienting a binary image, what kind of output do I 
   need to allocate.
   in - the DgpCoord_Orientation of the input image
   out - the desired output DgpCoord_Orientation.
   srcW - the image's fast width in it's input DgpCoord_Orientation.
   srcH - the image's slow width in it's input DgpCoord_Orientation.
   srcBpl - bytes per line in the source.
   will be set if valid pointers:
   dstW - the fast width of the output image 
   dstH - the slow width of the output image
   dstBpl - bytes per line in the dest.
   inplace - can this transformation be done in place with the rotation
     primitives?
*/
ErrCode
DgpRot_reorientBinaryNeeded(DgpCoord_Orientation in, DgpCoord_Orientation out, 
                     Dgp_Int32 srcW, Dgp_Int32 srcH, Dgp_Int32 srcBpl,
                     Dgp_Int32 * dstW, Dgp_Int32 * dstH, Dgp_Int32 * dstBpl,
                     Dgp_Bool * inplace, Dgp_State * dgp_state);

/* reorientBinary()
   Reorient the following binary image into the given output image.

   in - the DgpCoord_Orientation of the input image.
   inp - a pointer to (0,0) of the input image.
   srcW - the image's fast width in it's input DgpCoord_Orientation.
   srcH - the image's slow width in it's input DgpCoord_Orientation.
   srcBpl - bytes per line in the source.
   dstW - the fast width of the output image 
   dstH - the slow width of the output image
   dstBpl - bytes per line in the dest.

   If reorientBinaryNeeded() told you that this could be done in place, feel
   free to pass in the same pointers.
*/

ErrCode
DgpRot_reorientBinary(DgpCoord_Orientation in, Dgp_UInt32 * inp,
               Dgp_Int32 srcW, Dgp_Int32 srcH, Dgp_Int32 srcBpl,
               DgpCoord_Orientation out, Dgp_UInt32 * outp,
               Dgp_Int32 dstW, Dgp_Int32 dstH, Dgp_Int32 dstBpl, 
               Dgp_State * dgp_state);
              

#ifdef __cplusplus
}
#endif


#endif


