#include "dgp_state.h"
#include "rot_binary.h"
#include "endianness.h"
#include "util.h"
#include "rot_binary_tables.h"
#include <math.h>

/* 30.10.2000 / dhtesq. */
/* Moved from rot_binary_tables.h... */
static INLINE Dgp_UInt32 flipInt32(Dgp_UInt32 input)
{
  Dgp_UInt32 out;

  out = flippedByte[(input & 0x000000FF)];
  out <<= 8;
  out |= flippedByte[(input & 0x0000FF00) >> 8];
  out <<= 8;
  out |= flippedByte[(input & 0x00FF0000) >> 16];
  out <<= 8;
  out |= flippedByte[input >> 24];
  return out;
}

static INLINE void
rotateBinaryChunk(Dgp_UInt8 * src, Dgp_Int32 srcOff, Dgp_Int32 srcPitch, 
		  Dgp_UInt8 * dst, Dgp_Int32 dstPitch)
{
  /* don't initialize these guys to make compiler warnings go away, the */
  /* compiler might not realize they'll be set, and it'll put those */
  /* initializations in. This code needs to be fast. */
  Dgp_UInt32 w;
  Dgp_UInt32 w0;
  Dgp_UInt32 w1;
  Dgp_UInt32 w2;
  Dgp_UInt32 w3;
  Dgp_UInt32 w4;
  Dgp_UInt32 w5;
  Dgp_UInt32 w6;
  Dgp_UInt32 w7;
  Dgp_UInt32 mask;
  Dgp_UInt32 off;

  /* RotateChunk: */ 
  off = srcOff;
  w0 = *(src+off);
  off = off + srcPitch;
  w1 = *(src+off);
  off = off + srcPitch;
  w2 = *(src+off);
  off = off + srcPitch;
  w3 = *(src+off);
  off = off + srcPitch;
  w4 = *(src+off);
  off = off + srcPitch;
  w5 = *(src+off);
  off = off + srcPitch;
  w6 = *(src+off);
  off = off + srcPitch;
  w7 = *(src+off);
  off = off + srcPitch;
  w = *(src+off);
  w0 = w0 + ( w << 8 );
  off = off + srcPitch;
  w = *(src+off);
  w1 = w1 + ( w << 8 );
  off = off + srcPitch;
  w = *(src+off);
  w2 = w2 + ( w << 8 );
  off = off + srcPitch;
  w = *(src+off);
  w3 = w3 + ( w << 8 );
  off = off + srcPitch;
  w = *(src+off);
  w4 = w4 + ( w << 8 );
  off = off + srcPitch;
  w = *(src+off);
  w5 = w5 + ( w << 8 );
  off = off + srcPitch;
  w = *(src+off);
  w6 = w6 + ( w << 8 );
  off = off + srcPitch;
  w = *(src+off);
  w7 = w7 + ( w << 8 );
  off = off + srcPitch;
  w = *(src+off);
  w0 = w0 + ( w << 16 );
  off = off + srcPitch;
  w = *(src+off);
  w1 = w1 + ( w << 16 );
  off = off + srcPitch;
  w = *(src+off);
  w2 = w2 + ( w << 16 );
  off = off + srcPitch;
  w = *(src+off);
  w3 = w3 + ( w << 16 );
  off = off + srcPitch;
  w = *(src+off);
  w4 = w4 + ( w << 16 );
  off = off + srcPitch;
  w = *(src+off);
  w5 = w5 + ( w << 16 );
  off = off + srcPitch;
  w = *(src+off);
  w6 = w6 + ( w << 16 );
  off = off + srcPitch;
  w = *(src+off);
  w7 = w7 + ( w << 16 );
  off = off + srcPitch;
  w = *(src+off);
  w0 = w0 + ( w << 24 );
  off = off + srcPitch;
  w = *(src+off);
  w1 = w1 + ( w << 24 );
  off = off + srcPitch;
  w = *(src+off);
  w2 = w2 + ( w << 24 );
  off = off + srcPitch;
  w = *(src+off);
  w3 = w3 + ( w << 24 );
  off = off + srcPitch;
  w = *(src+off);
  w4 = w4 + ( w << 24 );
  off = off + srcPitch;
  w = *(src+off);
  w5 = w5 + ( w << 24 );
  off = off + srcPitch;
  w = *(src+off);
  w6 = w6 + ( w << 24 );
  off = off + srcPitch;
  w = *(src+off);
  w7 = w7 + ( w << 24 );
   
  w = w0 | w1 | w2 | w3 | w4 | w5 | w6 | w7;
   
  if (w != 0) {
    /* quick kill for all zeros is often effective */
    mask = 0x55555555;
   
    w = w0 & mask;
    w0 = mask & (w0 >> 1);
    w0 = w0 + ((~mask) & w1);
    w1 = w + ((mask & w1) << 1);
   
    w = w2 & mask;
    w2 = mask & (w2 >> 1);
    w2 = w2 + ((~mask) & w3);
    w3 = w + ((mask & w3) << 1);
   
    w = w4 & mask;
    w4 = mask & (w4 >> 1);
    w4 = w4 + ((~mask) & w5);
    w5 = w + ((mask & w5) << 1);
   
    w = w6 & mask;
    w6 = mask & (w6 >> 1);
    w6 = w6 + ((~mask) & w7);
    w7 = w + ((mask & w7) << 1);
   
    mask = 0x33333333;
   
    w = w0 & mask;
    w0 = mask & (w0 >> 2);
    w0 = w0 + ((~mask) & w2);
    w2 = w + ((mask & w2) << 2);
   
    w = w1 & mask;
    w1 = mask & (w1 >> 2);
    w1 = w1 + ((~mask) & w3);
    w3 = w + ((mask & w3) << 2);
   
    w = w4 & mask;
    w4 = mask & (w4 >> 2);
    w4 = w4 + ((~mask) & w6);
    w6 = w + ((mask & w6) << 2);
   
    w = w5 & mask;
    w5 = mask & (w5 >> 2);
    w5 = w5 + ((~mask) & w7);
    w7 = w + ((mask & w7) << 2);
   
    mask = 0x0f0f0f0f;
   
    w = w0 & mask;
    w0 = mask & (w0 >> 4);
    w0 = w0 + ((~mask) & w4);
    w4 = w + ((mask & w4) << 4);
   
    w = w1 & mask;
    w1 = mask & (w1 >> 4);
    w1 = w1 + ((~mask) & w5);
    w5 = w + ((mask & w5) << 4);
   
    w = w2 & mask;
    w2 = mask & (w2 >> 4);
    w2 = w2 + ((~mask) & w6);
    w6 = w + ((mask & w6) << 4);
   
    w = w3 & mask;
    w3 = mask & (w3 >> 4);
    w3 = w3 + ((~mask) & w7);
    w7 = w + ((mask & w7) << 4);
  };
   
  /* store the 8 registers */
  off = dstPitch*2;
  (*  (Dgp_UInt32 *) dst ) = w0;
  (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w1;
  dst = dst + off;
  (*  (Dgp_UInt32 *) dst ) = w2;
  (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w3;
  dst = dst + off;
  (*  (Dgp_UInt32 *) dst ) = w4;
  (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w5;
  dst = dst + off;
  (*  (Dgp_UInt32 *) dst ) = w6;
  (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w7;
}

static INLINE void
rotateBinaryChunkHallucinate(Dgp_UInt8 * src, Dgp_Int32 srcOff, 
			     Dgp_Int32 skipRows, Dgp_Int32 srcPitch, 
			     Dgp_UInt8 * dst, Dgp_Int32 numWriteRows,
			     Dgp_Int32 dstPitch)
{
  Dgp_UInt32 w;
  Dgp_UInt32 w0 = 0;
  Dgp_UInt32 w1 = 0;
  Dgp_UInt32 w2 = 0;
  Dgp_UInt32 w3 = 0;
  Dgp_UInt32 w4 = 0;
  Dgp_UInt32 w5 = 0;
  Dgp_UInt32 w6 = 0;
  Dgp_UInt32 w7 = 0;
  Dgp_UInt32 mask;
  Dgp_UInt32 off;

  off=srcOff;
  /* RotateChunk: */ 
  switch(skipRows)
    {
    case 0:
      w0 = *(src+off);
      off = off + srcPitch;
    case 1:
      w1 = *(src+off);
      off = off + srcPitch;
    case 2:
      w2 = *(src+off);
      off = off + srcPitch;
    case 3:
      w3 = *(src+off);
      off = off + srcPitch;
    case 4:
      w4 = *(src+off);
      off = off + srcPitch;
    case 5:
      w5 = *(src+off);
      off = off + srcPitch;
    case 6:
      w6 = *(src+off);
      off = off + srcPitch;
    case 7:
      w7 = *(src+off);
      off = off + srcPitch;
    case 8:
      w = *(src+off);
      w0 = w0 + ( w << 8 );
      off = off + srcPitch;
    case 9:
      w = *(src+off);
      w1 = w1 + ( w << 8 );
      off = off + srcPitch;
    case 10:
      w = *(src+off);
      w2 = w2 + ( w << 8 );
      off = off + srcPitch;
    case 11:
      w = *(src+off);
      w3 = w3 + ( w << 8 );
      off = off + srcPitch;
    case 12:
      w = *(src+off);
      w4 = w4 + ( w << 8 );
      off = off + srcPitch;
    case 13:
      w = *(src+off);
      w5 = w5 + ( w << 8 );
      off = off + srcPitch;
    case 14:
      w = *(src+off);
      w6 = w6 + ( w << 8 );
      off = off + srcPitch;
    case 15:
      w = *(src+off);
      w7 = w7 + ( w << 8 );
      off = off + srcPitch;
    case 16:
      w = *(src+off);
      w0 = w0 + ( w << 16 );
      off = off + srcPitch;
    case 17:
      w = *(src+off);
      w1 = w1 + ( w << 16 );
      off = off + srcPitch;
    case 18:
      w = *(src+off);
      w2 = w2 + ( w << 16 );
      off = off + srcPitch;
    case 19:
      w = *(src+off);
      w3 = w3 + ( w << 16 );
      off = off + srcPitch;
    case 20:
      w = *(src+off);
      w4 = w4 + ( w << 16 );
      off = off + srcPitch;
    case 21:
      w = *(src+off);
      w5 = w5 + ( w << 16 );
      off = off + srcPitch;
    case 22:
      w = *(src+off);
      w6 = w6 + ( w << 16 );
      off = off + srcPitch;
    case 23:
      w = *(src+off);
      w7 = w7 + ( w << 16 );
      off = off + srcPitch;
    case 24:
      w = *(src+off);
      w0 = w0 + ( w << 24 );
      off = off + srcPitch;
    case 25:
      w = *(src+off);
      w1 = w1 + ( w << 24 );
      off = off + srcPitch;
    case 26:
      w = *(src+off);
      w2 = w2 + ( w << 24 );
      off = off + srcPitch;
    case 27:
      w = *(src+off);
      w3 = w3 + ( w << 24 );
      off = off + srcPitch;
    case 28:
      w = *(src+off);
      w4 = w4 + ( w << 24 );
      off = off + srcPitch;
    case 29:
      w = *(src+off);
      w5 = w5 + ( w << 24 );
      off = off + srcPitch;
    case 30:
      w = *(src+off);
      w6 = w6 + ( w << 24 );
      off = off + srcPitch;
    case 31:
      w = *(src+off);
      w7 = w7 + ( w << 24 );
    }   
  w = w0 | w1 | w2 | w3 | w4 | w5 | w6 | w7;
   
  if (w != 0)
    {
      /* quick kill for all zeros is often effective */
      mask = 0x55555555;
   
      w = w0 & mask;
      w0 = mask & (w0 >> 1);
      w0 = w0 + ((~mask) & w1);
      w1 = w + ((mask & w1) << 1);
      
      w = w2 & mask;
      w2 = mask & (w2 >> 1);
      w2 = w2 + ((~mask) & w3);
      w3 = w + ((mask & w3) << 1);
   
      w = w4 & mask;
      w4 = mask & (w4 >> 1);
      w4 = w4 + ((~mask) & w5);
      w5 = w + ((mask & w5) << 1);
   
      w = w6 & mask;
      w6 = mask & (w6 >> 1);
      w6 = w6 + ((~mask) & w7);
      w7 = w + ((mask & w7) << 1);
   
      mask = 0x33333333;
   
      w = w0 & mask;
      w0 = mask & (w0 >> 2);
      w0 = w0 + ((~mask) & w2);
      w2 = w + ((mask & w2) << 2);
   
      w = w1 & mask;
      w1 = mask & (w1 >> 2);
      w1 = w1 + ((~mask) & w3);
      w3 = w + ((mask & w3) << 2);
   
      w = w4 & mask;
      w4 = mask & (w4 >> 2);
      w4 = w4 + ((~mask) & w6);
      w6 = w + ((mask & w6) << 2);
   
      w = w5 & mask;
      w5 = mask & (w5 >> 2);
      w5 = w5 + ((~mask) & w7);
      w7 = w + ((mask & w7) << 2);
   
      mask = 0x0f0f0f0f;
   
      w = w0 & mask;
      w0 = mask & (w0 >> 4);
      w0 = w0 + ((~mask) & w4);
      w4 = w + ((mask & w4) << 4);
   
      w = w1 & mask;
      w1 = mask & (w1 >> 4);
      w1 = w1 + ((~mask) & w5);
      w5 = w + ((mask & w5) << 4);
   
      w = w2 & mask;
      w2 = mask & (w2 >> 4);
      w2 = w2 + ((~mask) & w6);
      w6 = w + ((mask & w6) << 4);
   
      w = w3 & mask;
      w3 = mask & (w3 >> 4);
      w3 = w3 + ((~mask) & w7);
      w7 = w + ((mask & w7) << 4);
    };
   
  /* store the numWriteRows registers */
  off = dstPitch*2;
  switch(numWriteRows)
    {
    case 0:
      return;
    case 1:
      (*  (Dgp_UInt32 *) dst ) = w0;
      return;
    case 2:
      (*  (Dgp_UInt32 *) dst ) = w0;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w1;
      return;
    case 3:
      (*  (Dgp_UInt32 *) dst ) = w0;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w1;
      dst = dst + off;
      (*  (Dgp_UInt32 *) dst ) = w2;
      return;
    case 4:
      (*  (Dgp_UInt32 *) dst ) = w0;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w1;
      dst = dst + off;
      (*  (Dgp_UInt32 *) dst ) = w2;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w3;
      return;
    case 5:
      (*  (Dgp_UInt32 *) dst ) = w0;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w1;
      dst = dst + off;
      (*  (Dgp_UInt32 *) dst ) = w2;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w3;
      dst = dst + off;
      (*  (Dgp_UInt32 *) dst ) = w4;
      return;

    case 6:
      (*  (Dgp_UInt32 *) dst ) = w0;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w1;
      dst = dst + off;
      (*  (Dgp_UInt32 *) dst ) = w2;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w3;
      dst = dst + off;
      (*  (Dgp_UInt32 *) dst ) = w4;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w5;
      return;
    case 7:
      (*  (Dgp_UInt32 *) dst ) = w0;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w1;
      dst = dst + off;
      (*  (Dgp_UInt32 *) dst ) = w2;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w3;
      dst = dst + off;
      (*  (Dgp_UInt32 *) dst ) = w4;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w5;
      dst = dst + off;
      (*  (Dgp_UInt32 *) dst ) = w6;
      return;
    case 8:
      (*  (Dgp_UInt32 *) dst ) = w0;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w1;
      dst = dst + off;
      (*  (Dgp_UInt32 *) dst ) = w2;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w3;
      dst = dst + off;
      (*  (Dgp_UInt32 *) dst ) = w4;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w5;
      dst = dst + off;
      (*  (Dgp_UInt32 *) dst ) = w6;
      (*  (Dgp_UInt32 *) (dst + dstPitch) ) = w7;
      return;
    }

}

#ifdef _BIG_ENDIAN
#define CURSRCBYTE ((curXsrc & 31) / 8)
#endif

#ifdef _LITTLE_ENDIAN
#define CURSRCBYTE (3 - ((curXsrc & 31) / 8))
#endif

/*
  rotates a block of size (srcW,srcH) at src 90 degrees clockwise into dst
  (srcH, srcW) assumes no needed padding (although the end of a line
  must be padded to 32-bits, but that's nothing unusual.)
  But no frame is needed. Assumes 32-bit packed.
*/
void
DgpRot_rotateBinary90CW(Dgp_UInt32 * src, Dgp_Int32 srcW, Dgp_Int32 srcH,
			Dgp_Int32 srcBpl,
			Dgp_UInt32 * dst, Dgp_Int32 dstBpl)
{
  /* so here's the high level plan:
     We're going to rotate 8x32 blocks of the source into 32x8 blocks in the
     destination as long as we can. Then we'll have to pull a trick where we
     hallucinate pixels on the border cases so that it all works out.
     we'll travel in the source from [loop orders] left to right, bottom to 
     top and in the destination top to bottom, left to right and then have 
     to handle boundary cases when the we're at a height in the source where 
     we have less than 32 rows.
  */
  Dgp_UInt32 * pCurDst;
  Dgp_UInt32 * pCurSrc;
  
  Dgp_UInt32 curSrcByte;
  Dgp_Int32 curYdst = 0;
  Dgp_Int32 curXdst = 0;
  Dgp_Int32 curYsrc = 0;
  Dgp_Int32 curXsrc = 0;
  
  Dgp_Int32 stop=(srcW - (srcW % 8));
	
  for( curYdst = 0,curXsrc = 0,curSrcByte = CURSRCBYTE; curXsrc < stop; 
       curYdst += 8, curXsrc += 8,curSrcByte = CURSRCBYTE){
	    
      for(curYsrc = srcH - 32, curXdst = 0; curYsrc >= 0; 
	  curYsrc -= 32, curXdst += 32){
	  pCurSrc = src +(((curYsrc * srcBpl) / 4) + (curXsrc / 32));
	  pCurDst = dst +(((curYdst * dstBpl) / 4) + (curXdst / 32));
		
	  rotateBinaryChunk((Dgp_UInt8 *)pCurSrc, curSrcByte, srcBpl,
			    (Dgp_UInt8 *)pCurDst, dstBpl);
      }
  }
  
  /* Now, we've filled in all the destination except the last byte
     in each row, and upto 7 rows at the bottom of the destination. We 
     know that the destination width is padded to 32 bits, so
     we don't have any problem writing off the end of the image 
     (but we must keep the padding = 0), but we don't necessarily have
     32 rows left over at the top of the source. And people get 
     pissed when you read memory that's not yours. 
     So we'll hallucinate those extra lines inside the modified 
     rotate chunk function.
     So the first condition handles the last byte in each row of
     the destination-- note we have to special case the last one
     in case our srcW is not a multiple of 8. What joy.
     
     The second condition handles the bottom (up to) 7 rows that
     we haven't put down yet if the srcW is not a multiple of 8.
  */
  if(srcH % 32){
    Dgp_Int32 skipNum = 32 - (srcH % 32);
    curXdst = (srcH / 32) * 32;
	    
    for( curYdst = 0,curXsrc = 0,curSrcByte = CURSRCBYTE; curXsrc < stop; 
	 curYdst += 8, curXsrc += 8,curSrcByte = CURSRCBYTE){
      pCurSrc = src + (curXsrc / 32);
      pCurDst = dst + (((curYdst * dstBpl) / 4) + (curXdst / 32));	
      rotateBinaryChunkHallucinate((Dgp_UInt8 *)pCurSrc, curSrcByte,
				   skipNum, srcBpl,
				   (Dgp_UInt8 *)pCurDst, 8, dstBpl);
    }
    if(srcW % 8){
      curYsrc = 0;
      curXsrc = stop;
      curYdst = stop;
      curXdst = srcH - (srcH % 8);
      
      pCurSrc = src + (((curYsrc * srcBpl) / 4) + (curXsrc / 32));
      pCurDst = dst + (((curYdst * dstBpl) / 4) + (curXdst / 32));
      rotateBinaryChunkHallucinate((Dgp_UInt8 *)pCurSrc, curSrcByte,
				   skipNum, srcBpl,
				   (Dgp_UInt8 *)pCurDst,
				   srcW % 8, dstBpl);
    }
  }
  
  
  if(srcW % 8){
    Dgp_Int32 stopLoop = srcH - (srcH % 32);
    Dgp_UInt32 numWrite = srcW % 8;
    curXsrc = stop;
    curYdst = stop;
	    
    for(curYsrc = srcH - 32, curXdst = 0; curXdst < stopLoop;
	curXdst += 32,curYsrc -= 32){
      pCurSrc = src + (((curYsrc * srcBpl) / 4) + (curXsrc / 32));
      pCurDst = dst + (((curYdst * dstBpl) / 4) + (curXdst / 32));
      rotateBinaryChunkHallucinate((Dgp_UInt8 *)pCurSrc, curSrcByte,
				   0, srcBpl, (Dgp_UInt8 *)pCurDst,
				   numWrite, dstBpl);
    }
  }
}

void
DgpRot_rotateBinary90CCW(Dgp_UInt32 * src, Dgp_Int32 srcW, Dgp_Int32 srcH,
			 Dgp_Int32 srcBpl,
			 Dgp_UInt32 * dst, Dgp_Int32 dstBpl)
     
{
  Dgp_UInt32 * pCurDst;
  Dgp_UInt32 * pCurSrc;
  
  Dgp_UInt32 curSrcByte;
  Dgp_Int32 curYdst = 0;
  Dgp_Int32 curXdst = 0;
  Dgp_Int32 curYsrc = 0;
  Dgp_Int32 curXsrc = 0;
  
  for(curYsrc = 31, curXdst = 0; 
      curYsrc < srcH;
      curYsrc += 32, curXdst +=32){

      for(curXsrc = 0, curYdst = (srcW - 1), curSrcByte = CURSRCBYTE;
	  curYdst >= 7; curXsrc += 8, curYdst -= 8, curSrcByte = CURSRCBYTE){
	pCurSrc = src + (((curYsrc * srcBpl) / 4) + (curXsrc / 32));
	pCurDst = dst + (((curYdst * dstBpl) / 4) + (curXdst / 32));
	
	rotateBinaryChunk((Dgp_UInt8 *)pCurSrc, curSrcByte, -srcBpl,
			  (Dgp_UInt8 *)pCurDst, -dstBpl);
      }
  }
  
  /* so because of the annoyances of all this, we actually have 2 conditions
     to deal with:
     1) we have less than 8 rows at the top of the destination to fill
     in from the the last word in each line of the source.
     2) the last word in each line of the destination is not filled in, 
     because we have less than 32 lines to draw from the bottom of 
     the source.
     
     so we'll check for both conditions, and handle them.
  */
  
  /* if our source height is not a multiple of 32, then we have a partial  */
  /*	word to blat at the end of each row in the destination. */
  if(srcH % 32){
    /* so this loop is going to run from 0 to srcW of the source, */
    /* of the remaining < 32 lines */
    /* and from srcW to 0 of the destination writing the last word. */
    Dgp_Int32 skipNum = 32 - (srcH % 32);
    curYsrc = srcH - 1;
    curXdst = srcH - (srcH % 32);
    
    for( curYdst = (srcW - 1), curXsrc = 0, curSrcByte = CURSRCBYTE;
	 curYdst > 7; 
	 curYdst -= 8, curXsrc += 8,curSrcByte = CURSRCBYTE){

      pCurSrc = src + (((curYsrc * srcBpl) / 4) + (curXsrc / 32));
      pCurDst = dst + (((curYdst * dstBpl) / 4) + (curXdst / 32));
      rotateBinaryChunkHallucinate((Dgp_UInt8 *)pCurSrc, curSrcByte, 
				   skipNum, -srcBpl,
				   (Dgp_UInt8 *)pCurDst, 8, -dstBpl);
    }
    if(curYdst >= 0){
      pCurSrc = src + (((curYsrc * srcBpl) / 4) + (curXsrc / 32));
      pCurDst = dst + (((curYdst * dstBpl) / 4) + (curXdst / 32));
      rotateBinaryChunkHallucinate((Dgp_UInt8 *)pCurSrc,
				   curSrcByte, skipNum,
				   -srcBpl,(Dgp_UInt8 *)pCurDst,
				   curYdst + 1, -dstBpl);
    }
  }	
  if(srcW % 8){
    Dgp_Int32 stop = srcH - (srcH % 32);
    curXsrc = srcW - (srcW % 8);
    curSrcByte = CURSRCBYTE;
    curYdst = (srcW - 1) % 8;
    
    for(curYsrc = 31, curXdst = 0;
	curXdst < stop;
	curXdst += 32, curYsrc += 32){
      pCurSrc = src + (((curYsrc * srcBpl) / 4) + (curXsrc / 32));
      pCurDst = dst + (((curYdst * dstBpl) / 4) + (curXdst / 32));
      rotateBinaryChunkHallucinate((Dgp_UInt8 *)pCurSrc, curSrcByte, 0,
				   -srcBpl, (Dgp_UInt8 *)pCurDst,
				   curYdst + 1, -dstBpl);
    }
  }
}

#undef CURSRCBYTE



/* 
   rotates a block of size (srcW,srcH) of src 180 degrees in place.
   Requires no temporary space, and can be done in place.
   Not the fastest (could be blocked), but gets the job done.
   Some serious 32bit assumptions here.
*/
#define PACK_SIZE 32
void 
DgpRot_rotateBinary180InPlace(Dgp_UInt32 * src, 
			      Dgp_Int32 srcW, Dgp_Int32 srcH,
			      Dgp_Int32 srcBpl)
{
  /* We have two cases, one where we can just blast away a word at a time,
     and one where each destination word comes from 2 words of the source
     (because the source width is not a multiple of 32)
  */
  if((srcW % PACK_SIZE) == 0) {
    UInt32 srclpl, swp, swapsPerLine;
    UInt32 *p1 = NULL, *p2 = NULL, *pEndLine = NULL, *p1End = NULL;
	
    srclpl = srcBpl / sizeof(UInt32);

    /*
      how many swaps (swapping 32-bit dwords) will we have to do each row?
      make sure to ceil, because you want to include all of the last partial
      dword at the end of the row.
    */
    swapsPerLine = (UInt32)ceil(srcW/(8.0*sizeof(UInt32)));
	
    /* p2 is the last word of the last src line */
    p2 = src + ((srcH - 1) * srclpl) + (swapsPerLine - 1);
    /* p2 is the first word of the src line */
    p1 = src;

    /* now, we'll just actually do a swap of the words. */
    /* and therefore only need to go from 0...srcH/2, otherwise, it's the full */
    /* thing. */
    pEndLine = src + ((srcH / 2) * srclpl);
    
    /* so this is nice, we just grab a dword from the front, flip it, */
    /* and swap it with a dword from the end that's been flipped as well. */
    while(p1 < pEndLine){
      for(p1End = p1 + (swapsPerLine - 1); p1 <= p1End; p1++, p2--){
        swp = flipInt32(*p1);
        *p1 = flipInt32(*p2);
        *p2 = swp;
      }
        
      p1 += (srclpl - swapsPerLine);
      p2 -= (srclpl - swapsPerLine);
    }
    
    /* If we have an even number of rows, we're done. If there's an odd 
       number, we still have one. And we only go halfway.
    */
    if(srcH % 2) {
      for( ; p1 < p2 ; p1++, p2--){
        swp = flipInt32(*p1);
        *p1 = flipInt32(*p2);
        *p2 = swp;
      }
      if (swapsPerLine % 2){
        *p1 = flipInt32(*p1);
      }
    }
  }
  else {
    /* since the destination width is not a multiple of 32, we're going to have
       to take the last word plus some bits of the word before it, and swap
       that with the first word. We then continue with this pattern ad
       nauseum.
    */

    /* p1 will be the left side of the src row, */
    /* and p2 will be the right side of the dst row. */
    UInt32 *p1=NULL,*p2=NULL,*p1EndImg=NULL,*p1EndLine=NULL;
    UInt32 lpl=0;		
	
    UInt32 endLastDwordBit;
    UInt32 offsetLastDword;
	
    UInt32 flippedD=0,flippedA=0;
	
    UInt32 beginDwordBit=0,endDwordBit=0;
    UInt32 endSwap=0,beginSwap=0;
	
    UInt32 swp;

    offsetLastDword = (UInt32)ceil(srcW/(8.0*sizeof(UInt32)))-1;

    endLastDwordBit= 32 - (srcW % (8*sizeof(UInt32)));
	
    lpl = srcBpl /= sizeof(UInt32);
	
    /* set p1 up to be first word of upper left corner. */
    p1 = src;
    /* set p2 to be last word of lower left corner. */
    p2 = src + (srcH - 1) * lpl;
    p2 += offsetLastDword;
	
    p1EndImg = src + (int)(srcH / 2) * lpl;
	
    while(p1 < p1EndImg){
      p1EndLine = p1 + offsetLastDword;
      /* do a line. */
      while(p1 <= p1EndLine){
        /* now what is flippedD (our 'assembled' dword from the last */
        /* 2 dwords), we'll it's what we haven't */
        /* used out of p2 in the lower bits, and the remaining */
        /* 32 - those bits in the higher bits from p2-1. */
        /* and then we flip it horizontally. */
        flippedD = 
          ((*(p2-1)) << (PACK_SIZE - endLastDwordBit))
          | ((*p2) >> endLastDwordBit);
        flippedD = flipInt32(flippedD);
	    
        /* flippedA is mearly the Dgp_Int32 at p1, flipped horizontally. */
        flippedA = flipInt32(*p1);
	    
        /* now put D in A. Easy. */
        (*p1) = flippedD;
	    
        /* now put D in A. Not so easy. */
        /* so we have break up flippedA into 2 portions,  */
        /* and put the lower bits of it in (*p2) */
        /* and the higher bits of it in (*p2-1); */
        (*p2) &= gen1s(endLastDwordBit);
        (*(p2-1)) &= (~gen1s(endLastDwordBit));
	    
        (*p2) |= (flippedA << endLastDwordBit);
        (*(p2-1)) |= (flippedA >> (PACK_SIZE-endLastDwordBit));
	    
        /* advance both of our pointers towards the center. */
        p1++;
        p2--;
      }
	  
      /* so what turns out to be nice is that we don't need to */
      /* special case the bits at the end of the line,  */
      /* because they will just be blatting into the padding */
      /* at the end of the dword. */
	  
      /* so all's said and done and a row is all over. */
      /* Now we need to get p1 back to the beginning of the next */
      /* row so everything'll work out. */
      /* so at this point, p1 has traveled from the beginning */
      /* of the line to the center of the line  */
      /* (the centermost long), so to get it to the */
      /* beginning of the next line, we add the longs per line to it */
      /* - 1/2 of the row width  */
      p1 += (lpl-(offsetLastDword+1));
      p2 -= (lpl-(offsetLastDword+1));
    }

    /* Now the only special case is the middle row on an odd sized image. */
    if(srcH % 2){
      p1 = src + ((int)(srcH / 2) * lpl);
      p2 = src + ((int)(srcH / 2) * lpl);
      p2 += offsetLastDword;
	  
      /* we want to run until our two pointers are an int apart */
      /* (that they're on the two sides of the center line) */
      while(p1 < (p2-1)){
        flippedD=
          ((*(p2-1)) << (PACK_SIZE - endLastDwordBit)) 
          | ((*p2) >> endLastDwordBit);
        flippedD = flipInt32(flippedD);
	    
        /* flippedA is mearly the dword at p1, flipped horizontally. */
        flippedA = flipInt32(*p1);
	    
        /* now put D in A. Easy. */
        (*p1) = flippedD;
	    
        /* now put D in A. Not so easy. */
        /* so we have break up flippedA into 2 portions, */
        /* and put the lower bits of it in (*p2) */
        /* and the higher bits of it in (*p2-1); */
        (*p2) &= gen1s(endLastDwordBit);
        (*(p2-1)) &= (~gen1s(endLastDwordBit));
	    
        (*p2) |= (flippedA<<endLastDwordBit);
        (*(p2-1)) |= (flippedA>>(PACK_SIZE-endLastDwordBit));
	    
        /* advance both of our pointers towards the center. */
        p1++;
        p2--;
      }
      /* we can't pull this trick forever. one can only do these
         32 bit swaps for a while. eventually we have less than 64 bits,
         and we can't do this. So when we have less than
         64 bits to swap, we'll do it bit by bit.
         now handle the 32+(32-endLastDwordBit) bits that are left over.
      */
      beginDwordBit = PACK_SIZE - 1;
      beginSwap=1ul << (PACK_SIZE - 1);
      
      endDwordBit = endLastDwordBit;
      endSwap=1 << endDwordBit;
      /* our stopping condition is when the left side gets equal to */
      /* or ahead of the right side. */
      while(p1 < p2 || beginDwordBit > endDwordBit){
        /* get the bit at x,y */
        swp = (*p1) & beginSwap;
        /* set the bit at beginSwap; clear the bit we're writing */
        /* ((*p1) & ~beginSwap) and then  */
        /* or it with the bit at endSwap */
        (*p1) = ((*p1) & ~beginSwap) |
          (((*p2 & endSwap) >> endDwordBit) << beginDwordBit);
        /* now set the bit at endswap */
        (*p2) = ((*p2) & ~endSwap) | ((swp >> beginDwordBit) << endDwordBit);
        
        /* advance both counters. Notice we don't have to do a check */
        /* for beginSwap becoming zero, */
        /* we only have less than 31 cycles of this loop and */
        /*beginSwap starts at 1<<31. */
        /* They'll both advance at the same time,  */
        /*and it'll never have the chance to get down to zero. */
        beginSwap >>= 1;
        beginDwordBit--;
        
        endSwap <<= 1;
        endDwordBit++;
        if(!endSwap){
          p2--;
          endDwordBit = 0;
          endSwap = 1;
        }
      }
    }
  }
}










