/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libsamptok/tok.c,v 1.17 2003/02/10 22:53:33 mkrocken Exp $
 *
 */
#include "samptok.h"
#include "tokprivate.h"
#ifndef VXWORKS
#include <memory.h>
#endif /* (VXWORKS) */
#include <math.h>
#include "sm_image.h"
#include <image.h>
#include "image_defs.h"
#include "dgp_allocators.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsamptok/tok.c,v 1.17 2003/02/10 22:53:33 mkrocken Exp $");

static ErrCode 
checkTokToBin(ClassTable classes, 
              DgpImage_ColorRGB oursampIm, 
              samptokClass *samptok);

static ErrCode 
tokToClass(samptokClass *newClass, 
           DgpImage_ColorRGB oursampIm, 
           Dgp_State *dgp_state);

static ErrCode 
addToTable(ClassTable classes, 
           samptokClass c, int *num, 
           Dgp_State *dgp_state);

static void 
freeState(intSampTokState *state);

static void 
maybeMergeTok(samptokClass c, 
              DgpImage_ColorRGB oursampIm,
              int *samptoknum);

static Dgp_Bool 
exactMatch(DgpImage_ColorRGB model1, 
           DgpImage_ColorRGB model2);

static void 
mergeTok(samptokClass c, 
         DgpImage_ColorRGB oursampIm);


#define freeSampTokClass(_samptok) \
do { \
  if ((_samptok)->im != (DgpImage_ColorRGB)NULL) { \
    DgpImage_free((_samptok)->im); \
    (_samptok)->im = NULL; \
  } \
  SAMPTOKCLASS_FREE((_samptok)); \
} while(0)



ErrCode
samptokInit(SampTokState *_state, samptokOptions *options)
{
  DGP_STATE_PROLOGUE2("samptokInit()");
  intSampTokState *newSampTokState = NULL;
  Dgp_State *dgp_state;


  /* check params */
  DGP_CHECK_NULL(_state);
  DGP_CHECK_NULL(options);
  dgp_state = options->dgp_state;
  DGP_ERR_BEGIN(dgp_state);

  *_state = NULL;

  /* check our small pool usage */
  DGP_SMALLPOOL_VALIDATE(struct _samptokClass, SAMPTOKCLASS_POOL_ID);

  /* Now we can start allocating memory. */
  DGP_CHECK_ALLOC_CLEANUP(newSampTokState = (DGP_MALLOC)(sizeof(intSampTokState)));
  newSampTokState->dgp_state = dgp_state;

  /* Copy over the options */
  newSampTokState->flags = options->flags;

  /* Initialize the class table to empty */
  newSampTokState->classes.elts = (DgpAdt_SMImage)NULL;
  newSampTokState->classes.count = 0;
  newSampTokState->returned_classes = DGP_FALSE;
  newSampTokState->memUsage = 0;

  *_state = newSampTokState;
  newSampTokState = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(newSampTokState != NULL) {
    (options->DGP_FREE)(newSampTokState);
  }

  ERR_RETURN;
}

Dgp_Int32
samptokMemUsage(SampTokState state)
{
  intSampTokState *intstate = (intSampTokState *)state;

  if(intstate != NULL) return intstate->memUsage;

  return 0;
}

/* samptokAddToken
   
   Args:
   SampTokState state -- state cookie generated by samptokInit()

   DgpImage_ColorRGB sampledIm -- the DgpImage_ColorRGB to tokenize.

   Return value: 
   This function returns a samptoken number, which all and only instances of
   this same samptoken will have.  On error, it returns -1.

   IMPORTANT:
   If you didn't tell samptokInit that you wanted samptokAddToken to make
   copies of the samptokens you hand it, the DgpImage_ColorRGBs you pass in may
   be invalid after this function returns.
*/
ErrCode
samptokAddToken(SampTokState state, DgpImage_ColorRGB sampledIm, samptokClass *_newClass)
{
  intSampTokState *intstate = (intSampTokState *)state;
  DGP_STATE_PROLOGUE2("samptokAddToken()");
  ClassTable classes;
  samptokClass newClass = (samptokClass)NULL;
  DgpImage_ColorRGB oursampIm; /* either a copy of sampledIm or sampledIm itself */
  int samptoknum;
  Dgp_State *dgp_state;


  /* validate params */
  DGP_CHECK_NULL(_newClass);
  DGP_CHECK_NULL(sampledIm);
  DGP_CHECK_NULL(intstate);
  dgp_state = intstate->dgp_state;
  DGP_ERR_BEGIN(dgp_state);

  /*
   * Convert it to our internal representation.  It either copies it or
   * takes ownership, according to options->flags.
   */
  if (intstate->flags & SAMPTOK_OPT_COPYTOKENS) {
    /* we should copy the image */
    DGP_ERR_PASS_CLEANUP(DgpImage_dup(sampledIm, DGP_FALSE, (void **) &oursampIm));
  } else {
    oursampIm = sampledIm;
  }
  
  classes = &(intstate->classes);

  DGP_ERR_PASS_CLEANUP(checkTokToBin(classes, oursampIm, &newClass));

  if (newClass != NULL) { 
    /* the samptoken we were asked to add matched a samptoken we already have
       in our table free the images and return the matching class
    */
    *_newClass = newClass;
    newClass = NULL;

    NORMAL_RETURN;
  }

  /* It matched nothing - make a new class. */
  DGP_ERR_PASS_CLEANUP(tokToClass(&newClass, oursampIm, dgp_state));
  
  /* and put it in the table */
  DGP_ERR_PASS_CLEANUP(addToTable(classes, newClass, &samptoknum, dgp_state));

  /* All done - it's added. */

  intstate->memUsage += sizeof(*newClass) + 
    DGPIMG_WIDTH(oursampIm) * DGPIMG_HEIGHT(oursampIm) * sizeof(Dgp_Int32);


  *_newClass = newClass;
  newClass = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  /* Free the stuff we've allocated so far */

  if (newClass != (samptokClass)NULL) {
    freeSampTokClass(newClass);
  }

  NORMAL_RETURN;
}

/* return the classes of sampled images in an array.  Free the SampTokState.
   The number of classes will be returned in *num_classes. *num_classes will
   be set to -1 on error. */
ErrCode
samptokFinal(SampTokState *state, Dgp_Int32 *num_classes, samptokClass **_newClasses)
{
  intSampTokState *intstate = NULL;
  DGP_STATE_PROLOGUE2("samptokFinal()");
  ClassTable classes;
  samptokClass *newClasses = NULL;
  samptokClass c;
  DgpAdt_ListNode classln;
  DgpAdt_SMImage elts;
  DgpAdt_List bin;
  int w, h;


  if(state == NULL) NORMAL_RETURN;

  intstate = *state;
  *state = NULL;

  if(intstate == NULL) NORMAL_RETURN;

  DGP_ERR_BEGIN(intstate->dgp_state);
  
  /* check other params */
  DGP_CHECK_NULL(_newClasses);
  DGP_CHECK_NULL(num_classes);

  classes = &(intstate->classes);

  if (classes->elts == (DgpAdt_SMImage)NULL) {
    /* They never gave us any bitmaps at all... */
    DGP_ASSERT(classes->count == 0);
    *num_classes = 0;

    /* Get rid of the state */
    intstate->returned_classes = DGP_TRUE;
    freeState(intstate);

    *_newClasses = NULL;

    NORMAL_RETURN;
  }

  /* Generate the final output */
  *num_classes = classes->count;
  DGP_CHECK_ALLOC_CLEANUP(newClasses = (samptokClass *)
       (intstate->DGP_MALLOC)(sizeof(*newClasses) * classes->count));

  elts = intstate->classes.elts;

  DGPADT_SM_FOREACH(elts,(Dgp_UInt16)w,(Dgp_UInt16)h) {
      
      /* Don't call smGet() unless the memory for the point we want
   is allocated (because otherwise, smGet() will allocate it
   for us on the fly -- not a good way to cut down on the
   amount of memory we're using). */
    if (DgpAdt_smIsAllocated(elts,(Dgp_UInt16)w,(Dgp_UInt16)h)) {
        DGP_ERR_PASS_CLEANUP(DgpAdt_smGet(elts,(Dgp_UInt16)w,(Dgp_UInt16)h,(void **)&bin));
    } else {
        bin = DGP_NULLLIST;
    }

      if (bin != DGP_NULLLIST) {
        DGPADT_FOREACH (classln, bin) {
          /* For each class in the bin */
          c = ClassData(classln);
          DGP_ASSERT(c != (samptokClass)NULL);
          newClasses[c->samptoknum] = c;
        }
      }
  }
  
  intstate->returned_classes = DGP_TRUE;
  /* Get rid of the state */
  freeState(intstate);

  *_newClasses = newClasses;
  newClasses = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  /* Free any partially constructed things */

  if (newClasses != NULL) {
    (intstate->DGP_FREE)((void *)newClasses);
  }

  if(intstate != NULL) {
    freeState(intstate);
  }

  ERR_RETURN;
}

void
samptokFreeOutput(samptokClass *sampoutput, Dgp_Int32 num_classes, Dgp_State *dgp_state)
{
  DGP_VOID_PROLOGUE("samptokFreeOutput()", NULL);
  int i;
  samptokClass stClass;

  if((sampoutput == NULL) || (dgp_state == NULL)) return;

  for (i = 0; i < num_classes; i ++) {
    stClass = sampoutput[i];
    DGP_VOID_ASSERT(stClass != (samptokClass)NULL);

    freeSampTokClass(stClass); /* freeSampTokClass() frees the DgpImage_ColorRGB */
  }
  (DGP_FREE)((void *)sampoutput);
}

void
samptokAbort(SampTokState *state)
{
  intSampTokState *intstate;

  if(state == NULL) return;

  intstate = *state;
  *state = NULL;

  if(intstate == NULL) return;

  /* Free up all the stuff hung off the state */
  freeState(intstate);
}


/*
 * Take a token and make it into its own class.
 */
static ErrCode
tokToClass(samptokClass *_new, DgpImage_ColorRGB oursampIm, Dgp_State *dgp_state)
{
  DGP_PROLOGUE("tokToClass()", DGP_PERRSTATE);
  samptokClass stClass = (samptokClass)NULL;
  DGP_CHECK_NULL (oursampIm);

  SAMPTOKCLASS_SEC_MALLOC(&stClass);

  stClass->im = oursampIm;
  stClass->width = DGPIMG_WIDTH(oursampIm);
  stClass->height = DGPIMG_HEIGHT(oursampIm);
  stClass->num_instances = 1;
  stClass->samptoknum = -1; /* this will be filled in by addToTable() */
  
  *_new = stClass;
  stClass = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if (stClass != (samptokClass)NULL) {
    SAMPTOKCLASS_FREE(stClass);
  }

  ERR_RETURN;
}


/*
 * Check the given box against the given bin in the class table. Set
 * *samptoknum to the appropriate number if the box matches some class in
 * the bin, -1 if it doesn't.
 *
 * If it does match, merge it into that class.
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure.
 */
static ErrCode
checkTokToBin(ClassTable classes, DgpImage_ColorRGB oursampIm, samptokClass *samptok)
{
  DGP_PROLOGUE("checkTokToBin()", NULL);
  DgpAdt_List bin;
  DgpAdt_ListNode ln;
  samptokClass c;
  int samptoknum ;
  int bin_w = DGPIMG_WIDTH(oursampIm), 
    bin_h = DGPIMG_HEIGHT(oursampIm);

  *samptok = NULL;

  if ((classes->elts == (DgpAdt_SMImage)NULL) ||
      (bin_h < 0) || (bin_h > DGPADT_SMGET_YMAX(classes->elts)) ||  
      (bin_w < 0) || (bin_w > DGPADT_SMGET_XMAX(classes->elts,bin_h))) {
    NORMAL_RETURN;
  }
    
  if (DgpAdt_smIsAllocated(classes->elts,(Dgp_UInt16)bin_w,(Dgp_UInt16)bin_h))
    DGP_ERR_PASS_CLEANUP(DgpAdt_smGet(classes->elts,(Dgp_UInt16)bin_w,(Dgp_UInt16)bin_h,(void **)&bin));
  else
    bin = DGP_NULLLIST;
  if (bin == DGP_NULLLIST) {
    NORMAL_RETURN;
  }

  samptoknum = -1;
  /* Found the bin - start checking. */
  DGPADT_FOREACH (ln, bin) {
    c = ClassData(ln);
    maybeMergeTok(c, oursampIm, &samptoknum);

    if (samptoknum != -1) {
      /* Done. */
      *samptok = c;
      break;
    }
  }

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;

}

/*
 * This is like checkTokToBin, except it compares a token and a single class
 * (and assumes that the class has the same width and height as the token)
 *
 * It has the same caveats about failure.
 */
static void
maybeMergeTok(samptokClass c, DgpImage_ColorRGB oursampIm, int *samptoknum)
{
  if (exactMatch(c->im, oursampIm)) {
    mergeTok(c, oursampIm);
  }
  else {
    *samptoknum = -1;
    return;
  }

  *samptoknum = c->samptoknum;
}

/*
 * Merge the given token into the given class.
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure. On failure, the token has not
 * been merged into the class.
 */
static void
mergeTok(samptokClass c, DgpImage_ColorRGB oursampIm)
{
  c->num_instances++;
  /* Delete the runs.  If the user wanted us to preserve his/her memory,
     box->runs will be a copy.  If the user wanted us to take responsibility
     for his/her memory, it's our perogative to free this now if we want to.*/
  DgpImage_free(oursampIm);
}

/*
 * Add the given class to the table.
 */
static ErrCode
addToTable(ClassTable classes, samptokClass c, int *num, Dgp_State *dgp_state)
{
  DGP_PROLOGUE("addToTable()", DGP_PERRSTATE);
  DgpAdt_List newBin = DGP_NULLLIST;
  DgpAdt_ListNode retnode;
  void *e;

  *num = -1;

  if (classes->elts == (DgpAdt_SMImage)NULL) {
    DGP_ERR_PASS_CLEANUP(DgpAdt_smNew(&(classes->elts), DGP_TRUE, 0, dgp_state));
  }

  DGP_ERR_PASS_CLEANUP(DgpAdt_smGet(classes->elts, (Dgp_UInt16)(c->width), (Dgp_UInt16)(c->height), &e));

  if (e == NULL) {
    /* No bin there to hold c */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&newBin, dgp_state));
    e = newBin;

    DGP_ERR_PASS_CLEANUP(DgpAdt_smSet(classes->elts, (Dgp_UInt16)(c->width), (Dgp_UInt16)(c->height), (void *)newBin));
    /* We no Dgp_Int32er own newBin - don't free it on cleanup */
    newBin = DGP_NULLLIST;
  }

  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd((DgpAdt_List)e, (void *)c, &retnode));

  c->samptoknum = classes->count;
  classes->count++;

  *num = c->samptoknum;
  NORMAL_RETURN;

ERR_CLEANUP:

  if (newBin != DGP_NULLLIST) {
    DgpAdt_liFree(newBin);
  }

  ERR_RETURN;
}


/*
 * Free up everything that's hooked into the state. Watch out for NULL
 * pointers - they're valid in most fields.
 */
static void
freeState(intSampTokState *state)
{
  DGP_VOID_PROLOGUE("freeState)", NULL);
  Dgp_State *dgp_state;
  int w, h;
  DgpAdt_SMImage elts;
  DgpAdt_List bin;
  DgpAdt_ListNode classln;
  samptokClass c;

  if(state == NULL) return;

  dgp_state = state->dgp_state;

  elts = state->classes.elts;
  if (elts != (DgpAdt_SMImage)NULL) {
    DGPADT_SM_FOREACH(elts,(Dgp_UInt16)w,(Dgp_UInt16)h) {

  /* Don't call smGet() unless the memory for the point we want
     is allocated (because otherwise, smGet() will allocate it
     for us on the fly -- not a good way to cut down on the
     amount of memory we're using). */
      if (DgpAdt_smIsAllocated(elts,(Dgp_UInt16)w,(Dgp_UInt16)h)) {
        DgpAdt_smGet(elts,(Dgp_UInt16)w,(Dgp_UInt16)h,(void **)&bin);
      } else {
        bin = DGP_NULLLIST;
      }

      if (bin != DGP_NULLLIST) {
        if (!state->returned_classes) {
          /* we still have responsibility for these classes, so we
             need to free them (if we've passed them back to the user
             via samptokFinal(), it's his/her responsibility to clean them
             up). */
          DGPADT_FOREACH (classln, bin) {
            c = (samptokClass)DGPADT_LIGET(classln);
            DGP_VOID_ASSERT(c != (samptokClass)NULL);
            freeSampTokClass(c);
          }
        }

        DgpAdt_liFree(bin);
      }
    }

    DgpAdt_smFree(elts);
  }


  (DGP_FREE)((void *)state);
}

/*
 * Do two DgpImage_ColorRGBs match exactly? (Don't check width or height
 */
static Dgp_Bool
exactMatch(DgpImage_ColorRGB model1, DgpImage_ColorRGB model2)
{
  DgpImage_RGB *r1, *r1end, *r2;

  r1 = DGPIMG_PIXPTR(model1,DGPIMG_XBASE(model1),DGPIMG_YBASE(model1));
  r2 = DGPIMG_PIXPTR(model2,DGPIMG_XBASE(model2),DGPIMG_YBASE(model2));
  
  for (r1end = r1 + DGPIMG_WIDTH(model1) * DGPIMG_HEIGHT(model1);
       r1 < r1end;
       r1++, r2++)
    if ((r1->r != r2->r) || (r1->g != r2->g) || (r1->b != r2->b))
      return DGP_FALSE;

  return DGP_TRUE;
}
