#ifndef DOC_HELPERS_H
#define DOC_HELPERS_H

#include "santa_t.h"
#include "dgp_state.h"

/* doc_helpers.h 
   santaDoc helper functions.
   wbell 10/16/99
 */

#ifdef __cplusplus
extern "C"
{
#endif


/* docOpen()
   Given a previously allocated and cleared santaDoc, docOpen will
   fill it in and open the necessary document wide data structures.
   also initializes the document memory manager, which controls overall
   memory consumption for rendering processes on the document.
   
   pdoc - previously created document structure.
   mem_limit - the memory limit for rendering in this document, not counting
    santaRenderJobs and the memory associated with them (which comes from
    the caller)
   mem_used - how much memory has already been used (because presumably, this
    isn't the first thing you've allocated.)
   opts - the document rendering options.
   docuds - a uds structure that has an opened file pointer and
    will be shallowly copied
*/
ErrCode 
docOpen(santaDoc * pdoc, uds * docuds, Dgp_Int32 mem_limit, Dgp_Int32 mem_used, 
	santaDocOptions opts,
	Dgp_State * dgp_state);

/* docClose()
   Given a document opened with docOpen, this'll clean up all the state
   that's document wide. Yes, all pages should be closed before doing this.
   
   pdoc - the opened doc
*/
ErrCode
docClose(santaDoc * pdoc, Dgp_State * dgp_state);

#ifdef __cplusplus
}
#endif
#endif
