/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  wbell 7/10/98
**
**  File:  im_utils_tables.h
**
**  Stub declaration for external tables and inline functions used in im_utils.c.
**
**  So in im_utils.c, we do stuff like flipping a binary packed image horizontally
**  and to do so faster, when we want to flip a dword (32-bits), we just want to
**  do table lookup rather than actually figure out the flipping, so we have a
**  16-bit table flippedWord[] which is generated by generate_tables.c that if
**  everything is right should be in the makefile and life should be happy. 
**  But flippedWord[i] returns you the horizontally flipped version of the word
**  i (16-bits). We have an inline helper function for the dword case, 
**  which is just 2 16-bit flips.
**  Another table we need is the ones mask-- for some Dgp_UInt32 i< 32, we
**  want to generate all 1's in the lower i bits of the dword; so for example 
**  i = 3, it returns
**  0x00000007 (I think that's the right hex for the bottom 3 bits being 1.)
**  We also wrap that with an inline function.
**
**
**	Modification history:
**	    Dillon H. Tracy esq. / March 2000
**		o  dgp types namespace issues.
**      o  DGP_LIBEXPORTed onesMask[] and flippedWord[];  correct?
**		o  Search on 'dhtesq' for some of this coder's comments & changes.
**  
**************************************************************************/

#ifndef __IM_UTILS_TABLES_H__
#define __IM_UTILS_TABLES_H__

#include "dgp_types.h"
#include "util.h"


/*
  These two will be generated from the output of generate_tables.c. 
  Please link that in.
*/
extern Dgp_UInt32 onesMask[];
extern Dgp_UInt16 flippedWord[];


/*************************************************************************
**
**  Function:  static INLINE Dgp_UInt32 flipDWord(...)
**
**
**	Arguments:
**      o  input...
**
**	Return value:
**		o  out...
**
**	Modification history:
**      o  None.
**  
**/

static INLINE Dgp_UInt32 flipDWord(Dgp_UInt32 input){
  Dgp_UInt32 out = 0;
  
  out = flippedWord[(input & 0x0000FFFF)];
  out <<= 16;
  out |= flippedWord[(input >> 16)];
  
  return out;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static INLINE Dgp_UInt32 gen1s(...)
**
**
**	Arguments:
**      o  input...
**
**	Return value:
**		o  out...
**
**	Modification history:
**      o  None.
**  
**/
static INLINE Dgp_UInt32 gen1s(Dgp_UInt32 input)
{
  return onesMask[input];
}
/*************************************************************************/


#endif /*__IM_UTILS_TABLES_H__ */
