/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  wbell 6/15/98
**
**  File:  libxiff3_ext.h
**
**  My personal extensions to libXiff.
**
**	Modification history:
**	    Dillon H. Tracy esq. / March 2000
**		o  dgp types namespace issues.
**		o  Search on 'dhtesq' for some of this coder's comments & changes.
**      June 2000
**      o  Reimported into new santa;  need color extensions.
**  
**************************************************************************/

#ifndef __LIBXIFF3_EXT_H__
#define __LIBXIFF3_EXT_H__

/*#include "readXiff3.h" */
#include "color-extension.h"
#include "resolution_t.h"
#include "image.h"
#include "mmcompress.h"
#include "libsupercache.h"
#include "colortoken_t.h"
#include "dgp_types.h"

/* So because certain image types (wavelet,etc) cannot easily be 
   read in small strips; we have a FullImageLoadedStatus, which 
   tells us if we've decided to load one of those images in its 
   entirety and cache it until we're done with the page.
   so we have an enumerated type here that tells you which
   structure of the union in a XiffImageInfo to look for the 
   loaded image (kind of redundant, since you prolly could tell
   which type it should be from the XIF_COLORTYPE tag)
	Note; if we decide to fully load a binary image, you'll note 
		we have 2 choices-- packed or binary pixr.
 */
typedef Dgp_UInt32          FullImageLoadedStatus;
#define FullImageNotLoaded	0
#define FullImageMono		1
#define FullImageRGB		2
#define FullImagePixrRGB	3
#define	FullImagePixrMono	4

/* Since different image types also need to carry around with
	them different information, we made an 'image addition'
	extension (basically a void pointer) that you can use to 
	add on any type of extra information you need about an image,
	we currently just use it for tokenized layers. We declare
	an enum as well to tell what kind of addition is there, if
	any.
*/
typedef Dgp_UInt32                    ImageAdditionStatus;
#define	ImageAdditionNone             0
#define	ImageAdditionTokenized        1
#define ImageAdditionColorTokenized   2
#define ImageAdditionUnspecified      3


typedef struct
{
  /* This is a pointer to the cache entry that contains information */
  /* about the token positions. */
  suCacheEntry pPosEntry;
  
  /* The regular b&w position block. */
  Dgp_Int32 numBoxes;
  jb2SymbolInstance *boxes;
  
  /* The color tags, the color dictionaries, and cache entries. */
  suCacheEntry pColorTagEntry;
  color_tag *color_tags;
  
  suCacheEntry pColorDictEntry;
  color_token **color_dict;
  
  /* We'll also keep around a list of the sorted jb2SymbolInstance. */
  jb2SymbolInstance **sorted_positions;
  
}ColorTokenizedImageAddition;

typedef struct
{
  /* This is a pointer to the cache entry that contains information */
  /* about the token positions. */
  suCacheEntry pPosEntry;
  
  /* The position block stuff for a tokenized layer. */
  Dgp_Int32 numBoxes;
  jb2SymbolInstance *boxes;
  
  /* ...extensions to the position block. */
  Dgp_Int32 numExts;
  Dgp_Int32 *extTags;
  
}TokenizedImageAddition;

#if 0
/* The information for a transparency mask of an image. */
typedef struct
{
  UInt32 width,height;
  UInt32 xRes,yRes;
  Dgp_Int32 xPos,yPos;
  UInt32 stripReadHeight;
  XIF_COMPRESSION comp;
  
  FullImageLoadedStatus fullImgState;
  union 
  {
    DgpImage_Packed packedMonoIm;
    PIXR *pixrMonoIm;
  }fullImg;
  
  XIF_IMGHANDLE handle;
}XiffTransparencyMaskInfo;

/* An image. */
typedef struct
{
  UInt32 width,height;
  UInt32 xRes,yRes;
  Dgp_Int32 xPos,yPos;
  
  UInt32 stripReadHeight;
  XIF_COMPRESSION comp;
  
  UInt32 comp_size;
  XIF_COLORTYPE color;
  
  XIF_IMGHANDLE handle;
  
  /* transparency masks */
  XiffTransparencyMaskInfo *pMask;
  
  /* did we decide to load the full image  */
  /* at one time? */
  FullImageLoadedStatus fullImgState;
  union 
  {
    DgpImage_ColorRGB rgbIm;
    PIXR *pixrRgbIm;
    DgpImage_Packed packedMonoIm;
    PIXR *pixrMonoIm;
  }fullImg;
  
	ImageAdditionStatus imgAdditionState;
  union
  {
    ColorTokenizedImageAddition *colorTokImgExt;
    TokenizedImageAddition *tokImgExt;
    void * unspecifiedExt;
  }imgAddition;
  
} XiffImageInfo;
#endif /* 0 */
#endif /*__LIBXIFF3_EXT_H__ */



