#ifndef MEMORY_SENTRY_H
#define MEMORY_SENTRY_H

#include "memory_sentry_t.h"
#include "dgp_state.h"
#include "santa_t.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* memory_sentry.h
   wbell 10/12/99 
   the memory manager datatype manipulation functions.
   Note: the memory sentry mearly presents an interface for keeping track
   of memory that diff'rent parts of the decompresser are using at any point
   in time. It also provides a policy for purging memory from the santaLayers
   when memory is needed to continue.

   It *only* deals with the supercache in the document, and for each
   open page of the document, deals with the santaLayers for each layer
   in each open page. There are special hooks in santaLayer and supercache
   to facilitate memory querying 
*/

/* here's our specific. */
typedef struct 
{
  Dgp_Int32 mem_limit;
  Dgp_Int32 mem_used;
  santaDoc * pdoc;
} memorySentrySpecific;

/* memorySentryCreate()
   create a memorySentry, which keeps track of large memory allocations
   in santa. Any requests for temporary space for rendering or 
   permanent save space must go through the memory sentry for the job.
   It also has the hooks to purge cached memory when times get rough
   and we really need it.

   pdoc - the current doc.
   ppage - the page in question.
   mem_limit - the limit for large block allocation, in bytes.
   mem_used - how much memory has already been used (because presumably, this
    isn't the first thing you've allocated.)
*/
ErrCode memorySentryCreate(santaDoc * pdoc, Dgp_Int32 mem_limit, Dgp_Int32 mem_used,
			   memorySentry ** ppsentry,
			   Dgp_State * dgp_state);



/* memorySentryDestroy()
   destroy a memory sentry.

   ppsentry - a sentry object, will set *ppsentry = NULL;
*/
ErrCode memorySentryDestroy(memorySentry ** ppsentry, 
			    Dgp_State * dgp_state);


#ifdef __cplusplus
}
#endif
#endif

