#ifndef __PAGE_STRIP_H__
#define __PAGE_STRIP_H__

#include "dgp_state.h"
#include "page_strip_t.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* File page_strip.h
   wbell 6/8/98
   page strip creation/destruction routines.

   Revised for the millenium by wbell on 10/3/99
 */


/*	
  pageStripCreate()
	
  creates a new pageStrip that represents part of a rendered page
	
  pimg - a rendered image in which we'll use for the image to blat
  into. It must cover the desired portion of the page, and be
  at page resolution. We're going to stash it away, so don't 
  free it until after you call pageStripDestroy

  cleared - is this memory cleared?

  ppstrip - a pointer to the created page strip is placed here.
*/
ErrCode
pageStripCreate(renderedImage * pimg,
		Dgp_Bool cleared,
		pageStrip ** ppstrip,
		Dgp_State * dgp_state);

/*	
  pageStripDestroy()
  
  destroys a pageStrip created with pageStripCreate
  
  pppageStrip - the strip to destroy. We'll set the pointer to it
  to be NULL when we're done.
  
  Note: if you gave us a renderedImage when you called pageStripCreate.
  It's still yours. Don't forget to free it.
*/
ErrCode
pageStripDestroy(pageStrip ** ppstrip, Dgp_State * dgp_state);

/*
  pageStripGetCoords()
   returns the coordinates of this page strip in the page coordinate
   system (native orientation)
   
   pstrip - the strip in question.
   px, py, pw, ph - the span of the strip in page coordinates.
    if NULL, will not be set.
*/
ErrCode
pageStripGetCoords(pageStrip * pstrip, Dgp_Int32 * px, Dgp_Int32 * py,
		   Dgp_Int32 * pw, Dgp_Int32 * ph, Dgp_State * dgp_stateb);

/*	
  pageStripDoesOverlap()
    determines if these coordinates (page coordinates forming a rectangle)
    have any points on the given pageStrip.
    
    returns DGP_TRUE if part of this rectangle overlaps the strip
    DGP_FALSE if not
	
    pstrip - the strip of the page in question.
    x,y,w,h - coordinates at page resolution to check if they're 
      on the pageStrip.
*/
Dgp_Bool 
pageStripDoesOverlap(pageStrip * pstrip,
		     Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h);

#ifdef __cplusplus
}
#endif

#endif
