#ifndef __PAGE_STRIP_ORDER_T_H__
#define __PAGE_STRIP_ORDER_T_H__

#include "orientation_t.h"
#include "resolution_t.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* File page_strip_order_t.h
   wbell 6/8/98
   primitive for deciding how PageStrips will be generated for a particular 
   job, either by progressing down the page, or progressing through the color
   space (single/multiple Separations of CMYK) and then down the page, or
   the reverse. Note, going down the page means nothing about the orientation
   (it just means that progressive strips represent a different physical area
    of the page.)

   Revised for the millenium by wbell on 10/3/99
 */


typedef Dgp_Int32 stripAdvanceOrderType;
/* go down the page, and once through with the desired region of the page */
/* and then go through the color space (if applicable.) */
#define saoPageAreaProgression   (stripAdvanceOrderType)1
/* first go through the color space, and then down the desired region of */
/* the page. */
#define saoColorSpaceProgression (stripAdvanceOrderType)2

/* one can define how one would handle doing multiple cmyk color planes at once */
/* there's no inherent limitation, just we haven't implemented it, because  */
/* we frankly have no idea why someone would want to do it. */

typedef Dgp_Int32 orderInputType;
/* what input type are we taking in? single Separations, packed mono? */
/* cmyk single Separations or myck single Separations? */
#define oiMonoPacked            (orderInputType)1
#define oiRGBInterleaved        (orderInputType)2
#define oiSingleSeparationsCMYK (orderInputType)4
#define oiAllSeparationsCMYK    (orderInputType)8

typedef Dgp_Int32 pageAreaAdvanceOrderType;
/* how do we advance strips across the page? Increasing in the slow direction. */
/* If we cared, we could build this to be as general as we cared. I see no need */
/* for further generalization. */
#define paoAdvanceSlow (pageAreaAdvanceOrderType)1
  

typedef struct{
  /* total number of strips in this job. */
  Dgp_Int32 numStrips;
  /* which one are we currently on? */
  Dgp_Int32 curStrip;

  stripAdvanceOrderType advOrder;
  pageAreaAdvanceOrderType pageAdvOrder;
  orderInputType inputType;

  DgpCoord_Orientation o;
  DgpCoord_Resolution res;
  /* the coordinates of the page that we're going to */
  /* generate, in page DgpCoord_Orientation (RD) at page res. */
  Dgp_Int32 x,y,w,h;
  
  /* pageStrip size. */
  Dgp_Int32 fastWidth, slowWidth;

} pageStripOrder;


#ifdef __cplusplus
}
#endif
#endif
