#ifndef RENDERED_IMAGE_T_H
#define RENDERED_IMAGE_T_H

#include "image.h"

#include "resolution_t.h"
#include "orientation_t.h"

#ifdef __cplusplus
extern "C"
{
#endif


/* File rendered_image_t.h
   wbell 10/3/99
   rendered image primatives. rendered images are used as our 'generic'
   image data type for keeping around pieces of images and to represent
   page images.
*/

typedef Dgp_Int32 renderedImageType;
#define riMonoPacked 		(renderedImageType)1
#define riRGBInterleaved	(renderedImageType)2
/* for CMYK seperations, it's an ored combination of CMYKSeperations */
/* plus which ones are valid. For any combination of c,m,y,k, the  */
/* CMYKSeperations will be set. */
#define riCMYKSeparations	  (renderedImageType)4
#define riCSeparation             (renderedImageType)12
#define riMSeparation             (renderedImageType)20
#define riYSeparation             (renderedImageType)36
#define riKSeparation             (renderedImageType)70

/* the index into the cmyk array for the diff'rent seperations. */
#define riCSepIdx 0
#define riMSepIdx 1
#define riYSepIdx 2
#define riKSepIdx 3

typedef struct {	
  Dgp_Int32 x,y;
  Dgp_Int32 w,h;
  /* what DgpCoord_Resolution is this image at? */
  DgpCoord_Resolution src_res;
  /* 11.8.2000 / dhtesq. */
  /* Adding scaling support. */
  DgpCoord_Resolution dst_res;

  /* DgpCoord_Orientation? */
  DgpCoord_Orientation orient;
  
  /* how many bytes per line is the image data  */
  /* (note that this is independent of the width of the image, because */
  /*  of packing / padding reasons.) */
  Dgp_Int32 bpl;
  
  /* 7.8.2000 / dhtesq. */
  /* Adding transparency mask support. */
  void *trans_src;
  Dgp_Int32 trans_bpl;

  /* the data for this image is in some format, this tag tells us which. */
  /* for cmyk seperations, it's 4 pointers where one or more of them */
  /* is non-null. The order of the pointers is c,m,y,k */
  renderedImageType type;
  union {
    Dgp_UInt32 * packed;
    Dgp_UInt8 * cmyk[4];
    DgpImage_RGB * rgb;
  } buffer;
  Dgp_UInt8 freeMem;
} renderedImage;

#ifdef __cplusplus
}
#endif
#endif



