#ifndef SANTA_IMAGE_T_H
#define SANTA_IMAGE_T_H

#include "zephimage_t.h"
#include "rendered_image_t.h"
#include "sucache_t.h"
#include "sutokposblock_t.h"
#include "mmcompress.h"
#ifdef __cplusplus
extern "C"
{
#endif

/* santa_image_t.h
   wbell 11.27.99
   the santaImage datatype, which is a thin coating over the
   zephImageInfo of libzephyr. The santaImage datatype is maintained
   by the functions in santa_image.h, which keep it's state in a consistent
   state, but the layer functions involved in blatting freely read
   values from inside the santaImage and the underlying zephImage.

*/
 
typedef struct
{
  /* the libzephyr image which contains all the file related stuff like */
  /* its resolution and position on the page. */
  zephImageInfo zephImage;

  /* we could potentially have a portion of this image stored around already */
  /* partially rendered. If this pointer is non-null, this is a portion of */
  /* the image already decompressed. */
  renderedImage * prendered;

  /* same for this image's transparency mask. */
  renderedImage * pmask;

  /* we have a series of pins which allow us to lock memory with a fine */
  /* granularity so that the memory sentry will not purge things when we're */
  /* in the middle of an operation. We have the ability to lock down  */
  /* the actual temporary memory inside the zephimage, the rendered image */
  /* that contains a portion of the image and the one for the mask seperately. */
  /* if these counts are non-zero, someone cares about this memory.  */
  Dgp_UInt32 pin_image;
  Dgp_UInt32 pin_rendered;
  Dgp_UInt32 pin_mask;

  /* if this image is tokenized, we keep the position block and the */
  /* supercache state here. */
  /* this stuff is locked down by the pin_image pin. */
  suCacheEntry tokimg_entry;
  suTokenPositionBlock pos_block;
} santaImage;


#ifdef __cplusplus
}
#endif
#endif
