#ifndef SANTA_LAYER_BLAT_H
#define SANTA_LAYER_BLAT_H

#include "dgp_state.h"
#include "santa_t.h"
#include "santa_layer_t.h"
#include "memory_sentry_t.h"
#include "page_strip_order_t.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* santa_layer_blat.h
   wbell 11.6.99
   blatting an image in a santa page layer onto a renderd image.
*/

/* santaLayerBlatImage()
   blats an image in a layer on a renderedImage. Handles dispatching
   on image type / output type / scale / transparency / rotation / cropping.
   takes care of memory allocation with the memorySentry, and uses the
   pageStripOrder to make intelligent decisions what to keep (if any)
   of partial images.

   pdoc - an opened doc. used for the supercache if image is tokenized
   ppage - the page in question. used for general page properties like
    width / height.
   player - the layer to which the image belongs. uses the layer as a place
    to cache information about partially decoded images when we're all done.
   img_idx - which image in that layer you want to put down, starting
    from 0
   pdst - the destination buffer, already cleared / prepared suitably.
   porder - gives information about what should / should not be cached
    for later use.
   psentry - use in order to claim and release memory.
*/

ErrCode
santaLayerBlatImage(santaDoc * pdoc, santaPage * ppage,
		    santaLayer * player,
		    Dgp_Int32 img_idx, renderedImage * pdst, 
		    pageStripOrder * porder, memorySentry * psentry,
            renderedImage *selector,
			santaImage *selectorImage,
		    Dgp_State * dgp_state);


/* santaLayerBlatSingleImage()
   blats an image in a layer on a renderedImage, ignores transparency 
   masks. Handles dispatching on input type and deciding what to cache
   while decoding.

   pdoc - an opened doc. used for the supercache if image is tokenized
   ppage - the page in question. used for general page properties like
    width / height.
   player - the layer to which the image belongs. uses the layer as a place
    to cache information about partially decoded images when we're all done.
   img_idx - which image in that layer you want to put down, starting
    from 0
   pdst - the destination buffer, already cleared / prepared suitably.
   porder - gives information about what should / should not be cached
    for later use.
   psentry - use in order to claim and release memory.
*/

ErrCode
santaLayerBlatSingleImage(santaDoc * pdoc, santaPage * ppage,
			  santaLayer * player,
			  Dgp_Int32 img_idx, renderedImage * pdst, 
			  pageStripOrder * porder,
			  memorySentry * psentry,
              renderedImage *selector,
			  Dgp_State * dgp_state);

/* 4.8.2000 / dhtesq.

  santaLayerBlatTransparencyMask().
    Actually this just tries to render the zephImage pointed to by
    pimg->zephImage->maskInfo.  Typically it will be g4 encoded,
    but given the generality of the underlying interface, we don't care.

   pdoc - an opened doc. used for the supercache if image is tokenized
   ppage - the page in question. used for general page properties like
    width / height.
   pimg -- the image whose transparency mask we will try to render.
   pdst - the destination buffer, already cleared / prepared suitably.
   porder - gives information about what should / should not be cached
    for later use.
   psentry - use in order to claim and release memory.
*/

ErrCode
santaLayerBlatTransparencyMask(santaDoc * pdoc, 
                               santaPage * ppage,
                               santaImage * pimg,
                               renderedImage * pdst,
                               pageStripOrder * porder, 
                               memorySentry * psentry,
                               renderedImage *selector,
					           santaImage *selectorImage,
                               Dgp_State * dgp_state);
  
#ifdef DGP_DBUG
/*static */
ErrCode writeOutTiff(	Dgp_Int8                    *outName, 
                        Dgp_Int32                   pgXRes, 
                        Dgp_Int32                   pgYRes,
                        Dgp_UInt32		            *pageIm, 
						Dgp_Int32					bpl,
						Dgp_Int32					w,
						Dgp_Int32					h,
                        Dgp_State                   *dgp_state);
#endif /* DGP_DBUG */

#ifdef __cplusplus
}
#endif
#endif
