#ifndef SANTA_PAGE_H
#define SANTA_PAGE_H

#include "dgp_state.h"
#include "santa_t.h"
#include "orientation_t.h"
/* for Dgp_Bool  */
#include "util.h"


#ifdef __cplusplus
extern "C"
{
#endif

/* santa_page.h 
   easier primitives for generating a page.
   10.21.99
 */

/* santaRenderInStrips()
   Given an open document and an open page, render this portion of the
   page in this buffer in this orientation, given the entire destination
   buffer, but do it in bandsize strips to save decompression memory.

   pdoc - the document for ppage
   ppage - the page in question
   o - what orientation do you want to render this in. You pass in
    a pointer to data that represents the point in the rendering area that
    is closest to the origin of the full page in that orientation.
   x, y, w, h - what region of the page (in native coordinates) do you want
    to render.
   opts - any job specific options. If unsure, pass santaJobOptsDefault
   inputtype - what is the inputtype of this data (how is the dst pointer
    to be interpreted?)
   dstBpl - what is the number of bytes per line in the source?
   dst - the destination pointer.
   cleared - has this memory been cleared?

  17.7.2000 / dhtesq.
   out -- the image structure;  typically will be wrapped around dst.  Need this for DgpImage_() fns.
   force_stripwise -- i.e., has the user opted for true decompression in bands.  If so, we'll output each
    complete strip to stdout.  Currently only pnm is supported.
  14.9.2000 / dhtesq.
   jbig2_only -- caller only wants the tokenized layer.
   outgoing -- array of 4 file pointers used for stripwise cmyk decode.

  4.5.2001 / dhtesq.
   layerOptions -- contains bit settings which direct the decoder to ignore certain layers
   when rendering.  Replaces 'jbig2_only' flag.

   santaInputCMYKSingle is not supported for this inputtype, because that
   would imply that there is only 1 pointer in dst. We don't currently
   support doing one seperation of a CMYK page without doing the others.
*/
ErrCode santaRenderInStrips(santaDoc * pdoc,
			    santaPage * ppage,
			    DgpCoord_Orientation o, Dgp_Int32 bandsize,
			    Dgp_Int32 x, Dgp_Int32 y, Dgp_Int32 w, Dgp_Int32 h,
			    santaJobOptions opts,
			    santaInputType inputtype,
			    Dgp_Int32 dstBpl, void * dst,
			    void *out,
			    Dgp_Bool cleared,
			    Dgp_State * dgp_state,
			    FILE **outgoing,
                Dgp_Bool force_stripwise,
                Dgp_Int32 layerOptions);

#ifdef __cplusplus
}
#endif

#endif
