#ifndef SANTA_T_H
#define SANTA_T_H

#include "resolution_t.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* santa_t.h
   wbell 10/12/99
   reincarnated core santa datatypes.
   visible to the outside world.
*/

typedef Dgp_Int32 santaPageType;
#define santaPageMono  (santaPageType)1
#define santaPageColor (santaPageType)2

typedef struct _santaPrivateDoc * santaPrivateDoc;
typedef struct _santaPrivatePage * santaPrivatePage;
typedef struct _santaPrivateRenderJob * santaPrivateRenderJob;

typedef struct { 
  /* user visible stuff: */
  /* number of pages in the document. */
  Dgp_Int32 npages;
  
  santaPrivateDoc privateInfo;
} santaDoc;

typedef struct {
  /* what page is this? */
  Dgp_Int32 page_num;
  santaPageType page_type;
  /* Added field for forced color, mono output.  5.3.2001 / dhtesq. */
  santaPageType outputType;
  Dgp_UInt32 width, height;
  DgpCoord_Resolution res;
  DgpCoord_Resolution dst_res;

  santaPrivatePage privateInfo;

  Dgp_Bool scale_me;
  Dgp_Float32 scale_factor;
} santaPage;

typedef struct {
  santaPage * ppage;
  
  santaPrivateRenderJob privateInfo;
} santaRenderJob;

/* When opening a document, you're allowed to pass a santaDocOptions,
   which specifies general policies you want when this document is rendered.
*/
typedef Dgp_Int32 santaDocOptions;
#define santaDocOptsDefault (santaDocOptions)0
/* if there is any problem while rendering a job, bail; ie, don't try and  */
/* recover from errors and continue rendering. */
#define santaDocOptsSafe    (santaDocOptions)1

/* You can pass a similar structure in for the job, which sets policies for
   rendering a job
*/
typedef Dgp_Int32 santaJobOptions;
#define santaJobOptsDefault (santaJobOptions)0
/* we don't quite render to the xiff 3 specification at the moment. It's a  */
/* little broken. This will force xiff 3 compatibility, but will be slower. */
#define santaJobOptsXiff3   (santaJobOptions)1


/* Here's the possible input types for raster data which is
   a void pointer in the do band call. Note this is independent
   of the page type because we can generate color pages in binary
   and vice versa.
*/
typedef Dgp_Int32 santaInputType;
/* the pointer will be a word aligned ulong pointer to a packed binary */
/* buffer. */
#define santaInputMonoPacked (santaInputType) 1
/* the pointer will be a pointer to an RGB structure which is the */
/* three planes interleaved. */
#define santaInputRGB	     (santaInputType) 2
/* the pointer will be a Dgp_UInt8 pointer which corresponds to */
/* a section of a plane of the output image. */
#define santaInputCMYKSingle (santaInputType) 4
/* the pointer will be an array of 4 Dgp_UInt8 pointers, which */
/* will be copied, and all have the same number of bytes per line. */
/* the order of the pointers in this array is c,m,y,k. */
#define santaInputCMYKAll    (santaInputType) 8


/* this typedef states how we're going to progress through bands
 in a rendering job. For most input types (all but CMYKSingle)
 the only thing that makes sense is SantaJobAdvanceSlow, which
 means the next band after the current band will be the one next
 in the slow direction of the rendered orientation.
 
 For cmyk single, we can either go in the slow direction across the
 page, and then into the next color plane [SantaJobAdvanceCMYKSlow],
 or do a specific portion of the page in all color planes sequentially,
 and then move down in the slow direction [SantaJobAdvanceSlowCMYK]. 

*/

typedef Dgp_Int32 santaJobAdvanceOrder;
#define santaJobAdvanceSlow      (santaJobAdvanceOrder) 1 
#define santaJobAdvanceCMYKSlow  (santaJobAdvanceOrder) 2
#define santaJobAdvanceSlowCMYK  (santaJobAdvanceOrder) 4 


#ifdef __cplusplus
}
#endif
#endif



