/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  wbell 6/15/98
**
**  File:  selector_blat.c
**
**  Routines for generating selector layer packed images.
**
**	Modification history:
**	    Dillon H. Tracy esq. / March 2000
**		o  dgp types namespace issues.
**      o  liberr compliance issues.
**		o  Search on 'dhtesq' for some of this coder's comments & changes.
**  
**************************************************************************/

#include "selector_blat.h"
#include "santa_layer_blat.h"
#include "santa_layer_t.h"

#include "resolution_t.h"
#include "santa_private_t.h"
#include "orientation.h"
#include "rendered_image.h"
#include "im_utils_tables.h"
#include "image.h"
#include "image_defs.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsanta/selector_blat.c,v 1.41 2003/02/10 22:53:35 mkrocken Exp $");

#if 0
/*************************************************************************
**
**  Function:  ErrCode selectorBlatImageOnto(...)
**
**
**  Given a selector image, decodes a binary image and blats it onto the
**  selector (or'ing). If the image is the resolution of the selector, and the
**  compression types permit, it can decode directly onto the selector and
**  save a copy.
**	
**	Arguments:
**      o  image - a previously opened image info structure about the image to be blatted.
**      o  pStrip - the coordinates of the selector wrt the page.
**      o  selector	- the selector itself.
**      o  dgp_state references a prepared Dgp_State{}.
**      o  ciffState...
**
**	Return value:
**		o  ErrCode.
**
**	Modification history:
**      Dillon H. Tracy esq. / March 2000
**	    o  Implemented liberr exception handling;  return type changed to ErrCode from Dgp_Bool.
**  
**/


ErrCode
selectorBlatImageOnto(zephImageInfo             image,
                      pageStrip                 *pStrip,
                      DgpImage_Packed           selector,
                      Dgp_State                 *dgp_state, 
                      santaPage                 *ppage){
    DGP_PROLOGUE("selectorBlatImageOnto()", DGP_PERRSTATE);

    DGP_ASSERT(image);
    DGP_ASSERT(pStrip);
    DGP_ASSERT(selector);
    DGP_ASSERT(ppage);

    /*
    This sucker better be binary, or we're gonna have a lot of problems.
    */
    DGP_ERR_GEN_CLEANUPF0((image->colorType != XIF_COLORTYPE_BINARY),
                          cErrInvalidParm,
                          "BlatImageOntoSelector: this doesn't even make sense, this selector image isn't binary.\n");

    /*
    So before we even go into this whole mess of seeing what to read, let's
    check the image and see if we've loaded the whole thing already; because
    if we have, life is good.
    */
    switch(image->fullImgState){
        /*
        In the case where we have the full image around, we just fudge the
        stripReadImInfo and do the blat.
        */
        case FullImageMono:
            read.ULImage.x = read.ULImage.y = 0;
            read.BRImage.x = image->width;
            read.BRImage.y = image->height;
            
            DGP_ERR_PASS_CLEANUPF0(stripBlatPackedOnPacked( dgp_state,
                                                            pStrip,
                                                            selector,
                                                            image,
                                                            &read,
                                                            image->fullImg.packedMonoIm,
                                                            &blat),
                                   "BlatImageOntoSelector: couldn't blat fully decompressed packed mono im.\n");
            break;
        case FullImagePixrMono:
            read.ULImage.x = read.ULImage.y = 0;
            read.BRImage.x = image->width;
            read.BRImage.y = image->height;
            
            DGP_ERR_PASS_CLEANUPF0(stripBlatPIXROnPacked( pStrip,
                                                        selector,
                                                        image,
                                                        &read,
                                                        image->fullImg.pixrMonoIm,
                                                        &blat,
                                                        dgp_state),
                                 "BlatImageOntoSelector: couldn't blat fully decompressed pixr mono im.\n");
            break;
        case FullImageRGB:
            /*  Fall through... */
        case FullImagePixrRGB:
            DGP_ERR_GEN_CLEANUPF0(DGP_TRUE, 
                                  cErrInvalidParm, 
                                  "BlatImageOntoSelector: full image loaded is DGPIMAGE_RGB? Not on the selector.\n");
            break;
        case FullImageNotLoaded:
        /*
        So we do something different depending on the types of the image:
        for types that we can blat straight onto the selector (tokenized),
        we do just that, blatting with or.
        For all other types (g4) we decompress into a PIXR, and then blat
        that image.
        Note that if your tokenized image isn't at the right resolution,
        we're going to blat it into a new image, then we'll scale it into the
        selector. You lose.
        */
            switch(image->comp){
                case XIF_TOKENIZATION:
                case XIF_LLTOKENIZATION:
                    /*
                    Are we at the resolution of the selector?
                    If so, we blat right on the selector, else,
                    we generate an image
                    and scale that puppy onto the selector.
                    */
                    if((blat.ResRatio.x.den == blat.ResRatio.x.num) && (blat.ResRatio.y.den==blat.ResRatio.y.num)){
                        DGP_ERR_PASS_CLEANUPF0(stripBlatTokenizeOnPacked( pStrip,
                                                                          &selector,
                                                                          &read,
                                                                          image,
                                                                          &blat,
                                                                          dgp_state),
                                               "BlatImageOntoSelector: Error with stripBlatTokenizeOnPacked\n");
                    } else {
                        DgpImage_Packed unscaledIm = NULL;
                        DGP_ERR_PASS_CLEANUPF0(stripBlatTokenizeOnPacked(   pStrip,
                                                                            &unscaledIm,
                                                                            &read,
                                                                            image,
                                                                            &blat,
                                                                            dgp_state),
                            "BlatImageOntoSelector: Error with stripBlatTokenizeOnPacked for unscaled selector.\n");
                        DGP_ERR_PASS_CLEANUPF0(stripBlatPackedOnPacked( dgp_state,
                                                                        pStrip,
                                                                        selector,
                                                                        image,
                                                                        &read,
                                                                        unscaledIm,
                                                                        &blat),
                                               "BlatImageOntoSelector: couldn't scale blat tokenized im.\n");
                    }
                    NORMAL_RETURN;
                    break;
                default:
                    {
                        PIXR *decodedIm = NULL;
                        /* In the vertical case; we're going to load the whole thing and
                        keep it around as a full image.
                        */
                        DGP_ERR_PASS_CLEANUPF0(stripReadPIXRImageSegment( &decodedIm,
                                                                        &read,
                                                                        image,
                                                                        isVerticalStrip(pStrip),
                                                                        dgp_state,ciffState),
                                             "BlatImageOntoSelector: Error with readPIXRImageSegment\n");
                        DGP_ERR_PASS_CLEANUPF0(stripBlatPIXROnPacked( pStrip,
                                                                    selector,
                                                                    image,
                                                                    &read,
                                                                    decodedIm,
                                                                    &blat,
                                                                    dgp_state),
                                               "BlatImageOntoSelector: couldn't blat pixr mono im.\n");
                        if(!isVerticalStrip(pStrip)){
                            w_destroyPixr(dgp_state->ipwstate, &decodedIm);
                        }
                    }
                    break;
            }
    }
    NORMAL_RETURN;
ERR_CLEANUP:
    ERR_RETURN;
}
/*************************************************************************/
#endif /* 0 */

/*************************************************************************
**
**  Function:  ErrCode selectorRemoveImageAreaFrom(...)
**
**
**  Clears out the area where an image will be blatted on a selector so that
**  the or'ing works correctly for unspecified regions of the selector.
**	
**	Arguments:
**      o  dgp_state references a prepared Dgp_State{}.
**      o  pStrip - the coordinates of the selector wrt the page.
**      o  selector	- the selector itself.
**      o  image		- the image to clear space off the selector for.
**
**	Return value:
**		o  ErrCode.
**
**	Modification history:
**      Dillon H. Tracy esq. / March 2000
**	    o  Implemented liberr exception handling;  return type changed to ErrCode from Dgp_Bool.
**  
**/

ErrCode
selectorRemoveImageAreaFrom(Dgp_State       *dgp_state,
                            pageStrip       *pStrip,
                            renderedImage   *selector,
                            zephImageInfo   image,
                            pageStripOrder  *porder){
    DGP_PROLOGUE("selectorRemoveImageAreaFrom()", DGP_PERRSTATE);
    DGP_STOP_WHINE; 
    DGP_ASSERT(image);
    DGP_ASSERT(pStrip);

    porder = porder;

    /* Notice on this step, we don't have to do anything about scaling,
    because we know the region we want to black out
    in terms of the scaled coordinate system (the strip),
    and don't have to do anything special. */
    if(!DgpCoord_isRotated(ORIENT_RD, selector->orient)){
        /* Set all the bits to off. */
        packedSetAreaTo(selector->buffer.packed,
                        selector->bpl / sizeof(Dgp_Int32),
                        image->xPos - pStrip->pimg->x,
                        image->yPos - pStrip->pimg->y,
                        image->width,
                        image->height,
                        0ul);
    } else {
        /* Set all the bits to off. */
        packedSetAreaTo(selector->buffer.packed,
                        selector->bpl / sizeof(Dgp_Int32),
                        image->yPos - pStrip->pimg->y,
                        image->xPos - pStrip->pimg->x,
                        image->height,
                        image->width,
                        0ul);
    }
    NORMAL_RETURN;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  void packedSetAreaTo(...)
**
**
**	?...
**	
**	Arguments:
**      o  pSrc...
**      o  lpl...
**      o  clearX, clearY...
**      o  clearWidth, clearHeight...
**      o  value...
**
**	Return value:
**		o  None.
**
**	Modification history:
**	    Dillon H. Tracy esq. / March 2000
**      o  asserts -> DGP_VOID_ASSERT's.
**      June 2000
**      o  This used to be in packed_im_utils.c in the old santa.  Imported to new santa to get
**         selector layer decoding to work.
**  
**/

void 
packedSetAreaTo(Dgp_UInt32          *pSrc, 
                Dgp_Int32           lpl,
                Dgp_Int32           clearX,
                Dgp_Int32           clearY,
                Dgp_Int32           clearWidth,
                Dgp_Int32           clearHeight,
                Dgp_UInt32          value){
    DGP_VOID_PROLOGUE("packedSetAreaTo()", NULL);
    Dgp_UInt32        *pCur;
    Dgp_Int32         fullWords,yPos;
  /* 
     So here's the plan, for each line to clear, we've possibly got a 
     boundary case on both sides if we're not quite word aligned.
     so, we'll clear what we need of the first word by hand, and
     then just do a memset for all the full words we need to clear 
     on the line, and then we'll special case the final word as well.
  */
  DGP_VOID_ASSERT(pSrc);
  if(!clearWidth || !clearHeight) return;
  
  /* How many full words per line can we clear?  */
  /* That's the width - the number of pixels we can't do on the left */
  /* side, and the ones we have on the right after the last number-- */
  /* all that over the PACK_SIZE. */
  fullWords = (clearWidth-((clearX % PACK_SIZE)+
			  ((clearX+clearWidth) % PACK_SIZE)))/PACK_SIZE;
  
  for(yPos = clearY; yPos < clearHeight; yPos++){
      /* Where is the beginning of the line? */
      pCur = pSrc + (yPos * lpl);
      pCur += (Dgp_Int32)((Dgp_Float32)clearX / PACK_SIZE);
      
      /* Is there any part of the left side of the rect we need to clear */
      /* by hand? */
      if(clearX % PACK_SIZE){
          (*pCur) &= ~gen1s(PACK_SIZE - (clearX % PACK_SIZE));
          (*pCur) |= (gen1s(PACK_SIZE - (clearX % PACK_SIZE)) | value);
          pCur++;
      }
      /* So now we can clear the number of full words per line with */
      /* memset. */
      memset(pCur, value, fullWords * sizeof(Dgp_UInt32));
      pCur += fullWords;
            
      /* What do we need to clear off the right side? */
      if((clearX + clearWidth) % PACK_SIZE){
          (*pCur) &= gen1s(PACK_SIZE - ((clearX+clearWidth) % PACK_SIZE));
          (*pCur) |= (~gen1s(PACK_SIZE-((clearX+clearWidth) % PACK_SIZE)) | value);
          pCur++;
      }
  }
  return;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  ErrCode stripDecodeSelectorLayer(...)
**
**
**  Decodes a selector layer into a DgpImage_Packed.
**	
**	Arguments:
**      o  layerNum	- the layer of the selector to decode (must be even).
**      o  ppage - a previously opened page information structure.
**      o  pStrip - the necessary coordinates of the selector to generate.
**      o  output - a DgpImage_Packed to fill in, or created and returned,
**         that corresponds to the portion of the page specified by readSel.
**      o  dgp_state references a prepared Dgp_State{}.
**      o  ciffState...
**
**	Return value:
**		o  ErrCode.
**
**	Modification history:
**      Dillon H. Tracy esq. / March 2000
**	    o  Implemented liberr exception handling;  return type changed to ErrCode from Dgp_Bool.
**  
**/

ErrCode
stripDecodeSelectorLayer(Dgp_Int32          layerNum, 
                         santaPage          *ppage,
                         pageStrip          *pstrip,
                         renderedImage      **output,
                         pageStripOrder     *porder,
                         memorySentry       *psentry,
                         santaDoc           *pdoc,
                         Dgp_State          *dgp_state){
    DGP_PROLOGUE("stripDecodeSelectorLayer()", DGP_PERRSTATE);
    Dgp_Int32      imNum;

    Dgp_Bool        fullUseSelector = DGP_FALSE;
    /*DgpImage_Packed renderInto = NULL; */
    renderedImage   *renderInto = NULL;
    santaLayer      *player_sel, *player_for;
	renderedImageType	theReadType;

    DGP_ASSERT(ppage);
    DGP_ASSERT(output);
    DGP_ASSERT(pstrip);
    DGP_ASSERT(dgp_state);

  /* In order to gain an optimization for the color
     tokenized layer case, we don't want to generate
     a selector layer; so right now, we check for one
     image in the selector that's tokenized, and one
     image in the foreground that's tokenized as
     well (color) we just return.
  */
  player_sel = &page_private(ppage)->players[layerNum];


  /* If there are more layers, check if the next layer is a color token layer. */
  if(page_private(ppage)->nlayers >= layerNum + 1)
  {
    player_for = &page_private(ppage)->players[layerNum + 1];
    
    if (player_sel->nimages == 1 && player_for->nimages == 1ul && 
        (player_sel->pimgs[0].zephImage->compType == IO_COMP_LLTOKENIZATION ||
         player_sel->pimgs[0].zephImage->compType == IO_COMP_JBIG2) && 
        player_for->pimgs[0].zephImage->compType == IO_COMP_JBIG2)
    {
     output = NULL;
     NORMAL_RETURN;
    }
  }

  /* If there are no more layers and the current layer is a token layer, return. */
  if (page_private(ppage)->nlayers > layerNum && 
       (player_sel->pimgs[0].zephImage->compType == IO_COMP_LLTOKENIZATION ||
        player_sel->pimgs[0].zephImage->compType == IO_COMP_JBIG2))
  {
     output = NULL;
     NORMAL_RETURN;
  }

  /* I don't know if it's tfx-legal to pass in a >1 bpp selector here, but let's */
  /* cover the bases. */
  theReadType = (player_sel->pimgs[0].zephImage->compType == IO_COMP_JPEG ||
	  player_sel->pimgs[0].zephImage->compType == IO_COMP_JPEG2000) ? riRGBInterleaved:riMonoPacked;
  /*
   * Create the output selector renderedImage;  note it has to have the same orientation
   * as the strip.
   * 
   * The new santa's image blatting interfaces all use a rendered image structure, so it does
   * not suffice to create the selector just using DgpImage_new().
   *
   */
  if(!(*output)){
    DGP_ERR_PASS_CLEANUP(renderedImageCreate(pstrip->pimg->x,
                                             pstrip->pimg->y,
                                             pstrip->pimg->w,
                                             pstrip->pimg->h,
                                             &pstrip->pimg->src_res,
                                             &ppage->dst_res,
                                             pstrip->pimg->orient,
                                             0,                 /* Let the function calculate this... */
											 /* Zephyr is choking on this downstream;  it thinks it has to */
											 /* (e.g.) blat a mono g4 image into a color buffer... / dhtesq. */
                                             /*porder->inputType, // new now handle color too (jlin) */
											 theReadType,
                                             NULL,
                                             &renderInto,
                                             dgp_state));
  } else {
      renderInto = *output;
  }

  /*
     So the xiff spec says that by default, a selector is entirely 'selected',
     as weird as that may be.  So we'll go with that.  We're going to be smart,
     and select the whole selector first, and then remove out regions that
     we'll place images in the selector so everything works out right.  Now
     being that we're so smart, if there's an image that covers the whole
     area of the page, why even bother to select the entire selector and
     then deselect it?  That would be silly.  And there's no silliness in
     santa.  None at all.  You can quote me on that.

     So run through the images, and see if any of them covers the entire
     selector.
  */
  for(imNum = 0; imNum < player_sel->nimages; imNum++){
      zephImageInfo image;
      image = player_sel->pimgs[imNum].zephImage;

      if(image->xPos == 0 && image->yPos == 0 && image->width == (Dgp_UInt32)pstrip->pimg->w && 
         image->height == (Dgp_UInt32)pstrip->pimg->h){
          fullUseSelector = DGP_TRUE;
      }
  }

  /*
     So if fullUseSelector is set, we don't have to remove anything from
     the selector, we can just blat.
  */

  if(!fullUseSelector){
      /* Set all the bits to on. */
      if(DgpCoord_isRotated(ORIENT_RD, renderInto->orient)){
          packedSetAreaTo(  renderInto->buffer.packed,
                            renderInto->bpl / sizeof(Dgp_Int32),
                            0,
                            0,
                            renderInto->h,
                            renderInto->w,
                            0xFFFFFFFF);
      } else {
          packedSetAreaTo(  renderInto->buffer.packed,
                            renderInto->bpl / sizeof(Dgp_Int32),
                            0,
                            0,
                            renderInto->w,
                            renderInto->h,
                            0xFFFFFFFF);
      }

      /* Zero out anywhere an image will be. */
      for(imNum = 0; imNum < player_sel->nimages; imNum++){
          DGP_ERR_PASS_CLEANUPF2(selectorRemoveImageAreaFrom(dgp_state, 
                                                             pstrip,
                                                             renderInto,
                                                             player_sel->pimgs[imNum].zephImage,
                                                             porder),
                                 "StripDecodeSelectorLayer: error removing image %d from layer %d\n", imNum, layerNum);
      }
  }

  /* Now fill in this selector with all the images. */
  for(imNum = 0; imNum < player_sel->nimages; imNum++){
      zephImageInfo image;
      image = player_sel->pimgs[imNum].zephImage;

      DGP_ASSERT(image->colorType == IO_COLORTYPE_BINARY);

      DGP_ERR_PASS_CLEANUPF2(santaLayerBlatImage(   pdoc,
                                                    ppage,
                                                    player_sel,
                                                    imNum,
                                                    renderInto,
                                                    porder,
                                                    psentry,
                                                    NULL,
													NULL,
                                                    dgp_state),
                             "Error blatting image %d onto layer %d...", imNum, layerNum);
  }
/*
		DGP_ERR_PASS_CLEANUP(writeOutTiff(	"c:\\testbed\\tif\\g4.tif", 
											DgpCoord_floor(renderInto->dst_res.xRes),
											DgpCoord_floor(renderInto->dst_res.yRes),
											renderInto->buffer.packed,
											renderInto->bpl,
											renderInto->w,
											renderInto->h,
											dgp_state));*/

  *output = renderInto;
  NORMAL_RETURN;
ERR_CLEANUP:
  if(renderInto != *output){
      DGP_ERR_PASS(renderedImageDestroy(&renderInto, dgp_state));
  }
  ERR_RETURN;
}
/*************************************************************************/

