/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  wbell 6/15/98
**
**  File:  selector_blat.h
**
**  Routines for generating selector layers or portions of a selector.
**
**	Modification history:
**	    Dillon H. Tracy esq. / March 2000
**		o  dgp types namespace issues.
**		o  Search on 'dhtesq' for some of this coder's comments & changes.
**  
**************************************************************************/

#ifndef __SELECTORBLAT_H__
#define __SELECTORBLAT_H__

#include "util.h"
#include "santa.h"
#include "page_strip_t.h"
#include "page_strip_order_t.h"
#include "zephimage_t.h"
#include "memory_sentry_t.h"

/*
This is used by packedSetAreaTo().  In the old santa, it was in memory_def.h,
which no longer exists. / dhtesq.
*/
#define PACK_SIZE 32

/* stripDecodeSelectorLayer()
   decodes a selector layer into a DgpImage_Packed 
   
   layerNum - the layer of the selector to decode (must be odd)
   pInfo - a previously opened page information structure
   pStrip - the necessary coordinates of the selector to generate
   output - a DgpImage_Packed to fill in, or created and returned, that
     corresponds to the portion of the page specified by
     readSel.
*/

ErrCode
selectorBlatImageOnto(zephImageInfo             image,
                      pageStrip                 *pStrip,
                      DgpImage_Packed           selector,
                      Dgp_State                 *dgp_state, 
                      santaPage                 *ppage);

ErrCode
selectorRemoveImageAreaFrom(Dgp_State       *dgp_state,
                            pageStrip       *pStrip,
                            renderedImage   *selector,
                            zephImageInfo   image,
                            pageStripOrder  *porder);

ErrCode
stripDecodeSelectorLayer(Dgp_Int32          layerNum, 
                         santaPage          *ppage,
                         pageStrip          *pstrip,
                         renderedImage      **output,
                         pageStripOrder     *porder,
                         memorySentry       *psentry,
                         santaDoc           *pdoc,
                         Dgp_State          *dgp_state);

void 
packedSetAreaTo(Dgp_UInt32      *pSrc, 
                Dgp_Int32       lpl,
                Dgp_Int32       clearX, 
                Dgp_Int32       clearY,
                Dgp_Int32       clearWidth, 
                Dgp_Int32       clearHeight,
                Dgp_UInt32      value);


#endif /*__SELECTORBLAT_H__ */
