#ifndef TOKENIZED_BLAT_MONO_H
#define TOKENIZED_BLAT_MONO_H

#include "orientation_t.h"
#include "memory_sentry_t.h"
#include "dgp_state.h"
#include "symbol.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* tokenized_blat_mono.h
   wbell 11.29.99
   blatting a set of tokens down into a mono packed image; this level
   no longer deals with data structures, just a simple representation
   of the output buffer and the tokens.
*/

/* tokenizeBlatMono()
     given some jb2SymbolInstance and an output buffer at some orientation, blats
     these tokens onto the output buffer with no scaling, cropped to 
     a region inside the buffer. jb2SymbolInstance are considered to be in the
     page coordinate space, and will be cropped as such.

     pdst - a pointer to (x,y) in the destination buffer.
     bpl - number of bytes per line in one row of the dest buffer
     x, y - in the page coordinate space, what point does pdst represent
     o - what orientation is this buffer in?
     cropx, cropy, cropw, croph - in the page coordinate space, 
      what rectangle inside the destination do you want to blat with
      tokens (gives you the ability to crop)
     nboxes - how many tokens do we want to blat?
     pboxes - an array of jb2SymbolInstance, where the required dictionaries have already
      been fetched. Does not assume tokens are in the right orientation;
      will rotate to suit. 
     psentry - a memory sentry object to keep track of changes in usage
      of memory.
     pmod - did we modify any tokens?
*/
ErrCode
tokenizeBlatMono(Dgp_UInt32 * pdst, Dgp_Int32 bpl, Dgp_Int32 x, Dgp_Int32 y, DgpCoord_Orientation o, 
		 Dgp_Int32 cropx, Dgp_Int32 cropy, Dgp_Int32 cropw, Dgp_Int32 croph,
		 Dgp_Int32 nboxes, jb2SymbolInstance * pboxes, memorySentry * psentry,
                 Dgp_Bool * pmod, Dgp_State * dgp_state);

ErrCode
tokenizeBlatDecompressG4(jb2Symbol * ptok, Dgp_Int32 * mem_diff, 
                         Dgp_State *dgp_state);

ErrCode
tokenizeBlatRotateBinaryToken(jb2Symbol * ptok, DgpCoord_Orientation o, Dgp_Int32 * mem_diff,
                        Dgp_State * dgp_state);
#ifdef __cplusplus
}
#endif
#endif
