/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Author:  ?, ?
**
**  Filename:  santa_uncomp_all.c
**
**  Front end for decoder.
**
**
**	Modification history:
**	    Dillon H. Tracy esq. / March 2000
**		o  Auto-implemented dgp types.
**		o  Search on 'dhtesq' for this coder's comments & changes.
**  
**************************************************************************/

#if 0
#ifdef _WIN32
/* libipw defines cPlanes which is used in winsock2
 * so if a user include readXiff3.h and then winsock2.h they
 * get a really nasty error.  So we'll just include it first 
 * so this doesn't happen.
 */
#include <winsock2.h>
#endif
#endif /* 0 */

#include <libiow.pub>
#include <libpixrw.pub>

#include <math.h>
#include "dgp_err_control.h"
#include "dgp_err_printers.h"
#include "santa_uncomp_all.h"
#include "santa_halftone.h"


/*************************************************************************
**
**  Function:  Dgp_Bool santaUncompGetType(Dgp_Int8 *pType, Dgp_Int32 *pOutType, Dgp_Int8 *outExt)
**
**
**  Converts character-based orientation instructions to internal santa orientation.
**
**	Arguments:
**      o  pType is output type character array.
**      o  pOutType is the output type, internal representation.
**      o  outExt is the output extension character array, which gets filled on successful completion.
**         There should be enough space in it to hold four characters.
**
**	Return value:
**		o  Boolean true iff one of the valid output extensions has been passed:  pnm, tif, or bmp.
**
**	Modification history:
**      o  None.
**  
**/

Dgp_Bool santaUncompGetType(Dgp_Int8 *pType, Dgp_Int32 *pOutType, Dgp_Int8 *outExt) {
  if(!strcmp(pType, "PNM") || !strcmp(pType, "pnm")) {
      *pOutType = PNM;
      strcpy(outExt, "pnm");
      return DGP_TRUE;
  } else if(!strcmp(pType, "TIF") || !strcmp(pType, "tif")) {
      *pOutType = TIFF_G4;
      strcpy(outExt, "tif");
      return DGP_TRUE;
  } else if(!strcmp(pType, "BMP") || !strcmp(pType, "bmp")) {
      *pOutType = FF_BMP;
      strcpy(outExt, "bmp");
      return DGP_TRUE;
  } else {
    return DGP_FALSE;
  }
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  void santaUncompInitState(santaUncompAllState *pState)
**
**
**  Sets defaults for the santa cookie.
**
**	Arguments:
**      o  pState references the prepared santa cookie.
**
**	Return value:
**		o  None.
**
**	Modification history:
**      o  None.
**  
**/

void santaUncompInitState(santaUncompAllState *pState){
 if(!pState) return;

 pState->force = DGP_FALSE;
 pState->stripSize = 1 << (DGP_BITS_UINT32 - 2); /* FIX: should be max Dgp_Int32 */
 pState->o = ORIENT_RD;
 pState->start = -1;
 pState->stop = -1;
 pState->outputFiles = DGP_TRUE;
 pState->verbose = DGP_FALSE;
 pState->outtype = PNM;
 pState->stripwise = DGP_FALSE;
 pState->output_fname = NULL;
 pState->to_stdout = DGP_FALSE;
 pState->do_cmyk = DGP_FALSE;
 pState->jbig2_only = DGP_FALSE;
 pState->images_only = DGP_FALSE;
 pState->output_res.xRes = DgpCoord_eq(0,1);
 pState->output_res.yRes = DgpCoord_eq(0,1);
 pState->output_fname = NULL;
 strcpy(pState->ext, "pnm");
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode handleMonoPage( santaUncompAllState         *pState, 
**                                     Dgp_Int8                    *outName, 
**                                     Dgp_Int32                   pgXRes, 
**                                     Dgp_Int32                   pgYRes,
**                                     DgpImage_Packed             pageIm, 
**                                     Dgp_State                   *dgp_state
**                                     TIFF                        **tiff
**                                     Bool                        firstPage,
**                                     Bool                        lastPage)
**
**
**  Do the switch on type.  We just save it if it's a pnm output,
**  else, make a pixr, copy the packedimage, and ask ipcore to save it.
**
**  Writes/saves the passed page image, creating pixr representation as necessary. / dhtesq.
**
**	Arguments:
**      o  pState references the santa cookie.
**      o  outName is the output file name.
**      o  pgXRes, pgYRes in dpi.
**      o  pageIm is the packed image to be handled.
**      o  dgp_state points to a prepared Dgp_State{}.
**
**	Return value:
**		o  ErrCode;  can fail on bad save/write or pixr creation.
**
**	Modification history:
**      Dillon H. Tracy esq. / March 2000
**      o  Return type changed to ErrCode.  No longer outputs to console necessarily.
**  
**/

ErrCode handleMonoPage( santaUncompAllState         *pState, 
                        Dgp_Int8                    *outName, 
                        Dgp_Int32                   pgXRes, 
                        Dgp_Int32                   pgYRes,
                        DgpImage_Packed             pageIm, 
                        Dgp_State                   *dgp_state,
                        TIFF                        **tiff,
                        Bool                        firstPage,
                        Bool                        lastPage)
{
    DGP_PROLOGUE("handleMonoPage", DGP_PERRSTATE);
    PIXR *outIm = NULL;

  if(pState->outtype == PNM){
      DGP_ERR_PASS_CLEANUPF1(DgpImage_save(pageIm, outName), 
                     "santa-uncomp-all:  I couldn't save the pnm file '%s'.\n", outName);
  } else {
      Dgp_Int32 y;
      UInt32 *pPixrRow = NULL;
      ScaledInt1616 xdpi, ydpi;
      
      xdpi = Int32_TO_ScaledInt1616((Int32)pgXRes);
      ydpi = Int32_TO_ScaledInt1616((Int32)pgYRes);

      DGP_ERR_PASS_CLEANUPF1(w_createPixrExtended(DGP_PIPWSTATE, 
                                        DGPIMG_WIDTH(pageIm), 
                                        DGPIMG_HEIGHT(pageIm),
                                        1, 
                                        xdpi, 
                                        ydpi, 
                                        cFrameBits,
                                        1, 
                                        1, 
                                        NULL, 
                                        &outIm),
                     "santa-uncomp-all:  I couldn't create the pixr for '%s'.\n", outName);

      w_pixrSetType(DGP_PIPWSTATE, outIm, cColorNone);
      
      for(y = 0; y < DGPIMG_HEIGHT(pageIm); y++){
          w_getLine(DGP_PIPWSTATE, outIm, y,(UInt8 **)&pPixrRow);
          memcpy(pPixrRow, DGPIMG_GETROW(pageIm, y), DGPIMG_GETROWWIDTH(pageIm));
      }

      DGP_ERR_PASS_CLEANUPF1(w_writeEx(DGP_PIPWSTATE, 
                                       outName,
                                       outIm, 
                                       pState->outtype,
                                       tiff, 
                                       firstPage,
                                       lastPage),
                             "santa-uncomp-all:  I couldn't write the pixr for '%s'.\n", outName);
      
      w_destroyPixr(DGP_PIPWSTATE, &outIm);
  }
  NORMAL_RETURN;
ERR_CLEANUP:
  w_destroyPixr(DGP_PIPWSTATE, &outIm);
  ERR_RETURN;
}
/*************************************************************************/



/*************************************************************************
**
**  Function:  ErrCode handleColorPage( santaUncompAllState        *pState, 
**                                      Dgp_Int8                   *outName, 
**                                      Dgp_Int32                  pgXRes, 
**                                      Dgp_Int32                  pgYRes, 
**                                      RGBImage                   pageIm,
**                                      Dgp_State                  *dgp_state
**                                     TIFF                        ** tiff,
**                                     Bool                        firstPage,
**                                     Bool                        lastPage)
**
**
**  Writes/saves the passed page image, creating pixr representation as 
**  necessary.  Very similar to
**  preceding handleMonoPage().
**
**	Arguments:
**      o  pState references the santa cookie.
**      o  outName is the output file name, good to go.
**      o  pgXRes, pgYRes in dpi.
**      o  pageIm is the RGBImage to be handled.
**      o  dgp_state points to a prepared Dgp_State{}.
**
**	Return value:
**		o  ErrCode;  can fail on bad save/write or pixr creation.
**
**	Modification history:
**      Dillon H. Tracy esq. / March 2000
**      o  Return type changed to ErrCode.  
**        No longer outputs to console necessarily.
**  
**/
ErrCode 
handleColorPage(         santaUncompAllState        *pState, 
                         Dgp_Int8                   *outName, 
                         Dgp_Int32                  pgXRes, 
                         Dgp_Int32                  pgYRes, 
                         DgpImage_ColorRGB          pageIm,
                         Dgp_State                  *dgp_state,
                         TIFF                       **tiff,
                         Bool                       firstPage,
                         Bool                       lastPage)
{
    DGP_PROLOGUE("handleColorPage", DGP_PERRSTATE);
    PIXR *outIm = NULL;

    if(pState->outtype == PNM){
      DGP_ERR_PASS_CLEANUPF1(DgpImage_save(pageIm, outName),
                             "santa-uncomp-all:  I couldn't save the pnm file '%s'.\n", outName);
    } else {
        Dgp_Int32 y, x, width, height;
        
        PIXR *red, *green, *blue;
        UInt32 red4, blue4, green4;
        DgpImage_ColorRGBPixPtr pagePix;
        DgpImage_RGB r;
        Dgp_UInt32 *red_ptr, *green_ptr, *blue_ptr;
        ScaledInt1616 xdpi, ydpi;
        
        width = DGPIMG_WIDTH(pageIm);
        height = DGPIMG_HEIGHT(pageIm);
    
        xdpi = Int32_TO_ScaledInt1616((Int32)pgXRes);
        ydpi = Int32_TO_ScaledInt1616((Int32)pgYRes);
        DGP_ERR_PASS_CLEANUPF1(w_createPixrExtended(DGP_PIPWSTATE, 
                                           width, 
                                           height,
                                           8, 
                                           xdpi, 
                                           ydpi,
                                           cFrameBits,
                                           3, 
                                           1, 
                                           NULL, 
                                           &outIm),
                                "santa-uncomp-all:  I couldn't create the pixr for '%s'.\n", outName);
        
        w_pixrGetChannel(DGP_PIPWSTATE, 
                         outIm,
                         (pState->outtype == TIFF_G4 ? 0 : 2),
                         &red);
        w_pixrGetChannel(DGP_PIPWSTATE, 
                         outIm, 
                         1, 
                         &green);
        w_pixrGetChannel(DGP_PIPWSTATE, 
                         outIm, 
                         (pState->outtype == TIFF_G4 ? 2 : 0), 
                         &blue);
    
        for(y=0; y<height; y++){
            w_getLine(DGP_PIPWSTATE, red, y, (UInt8 **)&red_ptr);
            w_getLine(DGP_PIPWSTATE, green, y, (UInt8 **)&green_ptr);
            w_getLine(DGP_PIPWSTATE, blue, y, (UInt8 **)&blue_ptr);
            pagePix = DGPIMG_PIXPTR(pageIm, 0, y);
            
            for(x=0; x<width; red_ptr++, green_ptr++, blue_ptr++){
                r = *(pagePix++);
                x++;
                
                red4 = r.r << 24;
                green4 = r.g << 24;
                blue4 = r.b << 24;
                
                if(x >= width){
                    *red_ptr = green4;
                    *green_ptr = blue4;
                    *blue_ptr = red4;
                    break;
                }
                
                r=*(pagePix++);
                x++;
                
                red4 |= r.r << 16;
                green4 |= r.g << 16;
                blue4 |= r.b << 16;
                
                if(x>=width){
                    *red_ptr = green4;
                    *green_ptr = blue4;
                    *blue_ptr = red4;
                    break;
                }
                
                r = *(pagePix++);
                x++;
                
                red4 |= r.r << 8;
                green4 |= r.g << 8;
                blue4 |= r.b << 8;
                
                if(x >= width){
                    *red_ptr = green4;
                    *green_ptr = blue4;
                    *blue_ptr = red4;
                    break;
                }
                
                r = *(pagePix++);
                x++;
                
                red4 |= r.r;
                green4 |= r.g;
                blue4 |= r.b;
                
                *red_ptr = red4;
                *green_ptr = green4;
                *blue_ptr = blue4;
            }
        }
        DGP_ERR_PASS_CLEANUPF1(w_writeEx(DGP_PIPWSTATE, 
                                       outName,
                                       outIm, 
                                       pState->outtype,
                                       tiff,
                                       firstPage,
                                       lastPage),
       "santa-uncomp-all:  I couldn't write the pixr for '%s'.\n", outName);

        
        w_destroyPixr(DGP_PIPWSTATE, &outIm);
    }
    NORMAL_RETURN;
ERR_CLEANUP:
    w_destroyPixr(DGP_PIPWSTATE, &outIm);
    ERR_RETURN;
}
/*************************************************************************/

/*************************************************************************/
ErrCode 
handleCMYKPage(          santaUncompAllState        *pState, 
                         Dgp_Int8                   *outName, 
                         Dgp_Int32                  pgXRes, 
                         Dgp_Int32                  pgYRes, 
                         void                       **pageIm,
                         Dgp_State                  *dgp_state,
                         TIFF                       **tiff,
                         Bool                       firstPage,
                         Bool                       lastPage)
{
    DGP_PROLOGUE("handleColorPage", DGP_PERRSTATE);
    Dgp_Int8 chars_cmyk[5] = "cmyk";
    Dgp_Int8 fnm[256];
    Dgp_Int32 i;

    pgXRes = pgXRes;
    pgYRes = pgYRes;
    tiff = tiff;
    firstPage = firstPage;
    lastPage = lastPage;

    if(pState->outtype == PNM){
        for(i = 0; i < 4; i++){
            sprintf(fnm, "%s%c.%s", outName, chars_cmyk[i], pState->ext);
            DGP_ERR_PASS_CLEANUPF2(DgpImage_save((DgpImage_Gray)pageIm[i], fnm),
                "I couldn't save the %s file '%s'.\n", pState->ext, fnm);
        }
    }

    NORMAL_RETURN;
ERR_CLEANUP:
    ERR_RETURN;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  ErrCode santaUncompProcessDocument(...)
**
**
**  Unscrambles state options and invokes santa.  Most option logic concentrated here.
**
**	Arguments:
**      o  pState references the santa cookie.
**      o  docName is the document's destination file name, less the extension.
**      o  docSource is the document's original name, including path.
**      o  docUds...
**      o  dgp_state points to a prepared Dgp_State{}.
**
**	Return value:
**		o  ErrCode...
**
**	Modification history:
**      Dillon H. Tracy esq. / March 2000
**      o  Return type changed to ErrCode.  No longer outputs to console necessarily.
**      o  Modified call to santaGeneratePageInStrips() to use DGP_ERR_PRINTER only.
**      o  Should errors be thrown on auto/unknown page types?
**      August 2000
**      o  Adding support for semi-arbitrary destination resolutions.
**      o  Adding cmyk support.
**      o  For FX release, many of the color-related functions (scaling, cmyk...) have been disabled.
**
**/

ErrCode santaUncompProcessDocument(santaUncompAllState         *pState, 
                                   Dgp_Int8                    *docName,
                                   Dgp_Int8                    *docSource,
                                   uds                         *docUds,
                                   Dgp_State                   *dgp_state){
	DGP_PROLOGUE("santaUncompProcessDocument()", DGP_PERRSTATE);

	Dgp_Int32       pgNum;
    TIFF            *tiff = NULL;

	void * out = NULL, *binaryOut = NULL;
	void * dst = NULL;
    void *cmyk_img[4];
    void *cmyk_dst[4];
	Dgp_Int32 dstBpl = -1;
	Dgp_Int32 layerOptions = DGP_SHOWALL_MASK;

	Dgp_Int8        outName[255];
	Dgp_Int32       numDig;
	Dgp_Int8        *prefix = NULL;
	santaDoc * pdoc = NULL;
	santaPage * ppage = NULL;
	Dgp_Int32 x = 0, y = 0, w = 0, h = 0, i;
	Dgp_Int32 nw = 0, nh = 0;
	Dgp_UInt16 rgbImageOpts;
    Dgp_Float32 scale_factor;
	Dgp_Int32 renderedWidth, renderedHeight;

	santaInputType inputtype = -1;
    FILE            *outgoing[4];

	PIXR		*binaryPixr = NULL;

    for(i = 0; i < (Dgp_Int32)(sizeof(cmyk_img)/sizeof(cmyk_img[0])); i++){
        cmyk_img[i] = NULL;
        cmyk_dst[i] = NULL;
        outgoing[i] = NULL;
    }

	DGP_ERR_PASS_CLEANUPF0(santaOpenDocumentEx(docUds, 1 << (DGP_BITS_UINT32 - 2), 
						   santaDocOptsDefault,
						   dgp_state,
						   &pdoc),
 			       "santaUncompAll: Cannot open document\n");
    
    if(pState->verbose){
        fprintf(stderr, "%s, %ld pages\n", docSource, pdoc->npages);
    }

	numDig = (Dgp_Int32)ceil(log10(pdoc->npages));

    /* if output_fname is NULL or '\0', prefix will be docName */
    if (!pState->output_fname) 
        prefix = docName;
    else if( !pState->output_fname[0]) 
        prefix = docName;
    else 
        prefix = pState->output_fname;


	pState->start = pState->start > 0 ? pState->start : 0; 
    pState->stop = pState->stop > 0 ? pState->stop : pdoc->npages; 

    DGP_ERR_GENF0(!DgpCoord_equal(pState->output_res.xRes, pState->output_res.yRes),
        cErrInvalidParm,
        "Non-square output resolution specified...");

    for(pgNum = pState->start; pgNum < pState->stop; pgNum++){
        DGP_ERR_PASS_CLEANUPF1(santaOpenPage(pdoc, pgNum, dgp_state, &ppage),
            "I couldn't open page %d", pgNum);

        if(DgpCoord_g(pState->output_res.xRes, ppage->res.xRes)){
            ppage->dst_res = ppage->res;
            pState->output_res = ppage->res;
            ppage->scale_factor = 1.0;
            ppage->scale_me = DGP_FALSE;
            pState->scale_me = DGP_FALSE;
			DGP_WARNF2("Page %d is only at %d dpi;  I can't upscale.", pgNum, (Dgp_Int32)DgpCoord_float(ppage->res.xRes));
		} else if(DgpCoord_equal(pState->output_res.xRes, ppage->res.xRes)){
            ppage->dst_res = ppage->res;
            pState->output_res = ppage->res;
            ppage->scale_factor = 1.0;
            ppage->scale_me = DGP_FALSE;
            pState->scale_me = DGP_FALSE;
			/*DGP_WARNF1("Requested resolution (%d dpi) is page resolution;  not scaling...", (Dgp_Int32)DgpCoord_float(ppage->res.xRes)); */
        } else {
            if((DgpCoord_float(pState->output_res.xRes) <= 0) || (DgpCoord_float(pState->output_res.xRes) <= 0)){
                ppage->dst_res = ppage->res;
                pState->output_res = ppage->res;
                ppage->scale_factor = 1.0;
                ppage->scale_me = DGP_FALSE;
                pState->scale_me = DGP_FALSE;
            } else {
                ppage->dst_res = pState->output_res;
                ppage->scale_factor = DgpCoord_float(DgpCoord_div(ppage->dst_res.xRes, ppage->res.xRes));
                ppage->scale_me = DGP_TRUE;
                pState->scale_me = DGP_TRUE;
            }
        }

        if(pState->outtype == TIFF_G4){
            /* Multi-page output... */
            sprintf(outName, "%s.%s", prefix, pState->ext);
        } else if(pState->do_cmyk){
            /* Form will be foo.xif.n.[c | m | y | k].pnm ...but hold off on the stuff after the page number. */
            sprintf(outName, "%s.%0*ld.", prefix, (int)numDig, pgNum);
        } else {
            sprintf(outName, "%s.%0*ld.%s", prefix, (int)numDig, pgNum, pState->ext);
        }
        x = 0;
        y = 0;
        w = ppage->width;
        /* Note as currently constructed, santa_uncomp_all has to allocate a full page buffer to take advantage */
        /* of ipcore's file output functions.   When outputting tiff, setting stripSize will have no first-order  */
        /* effect on memory usage */
        /* and will certainly hurt performance. / dhtesq. */
        /*h = pState->stripSize; */
        h = ppage->height;

        if(pState->stripwise == DGP_FALSE){
            nw = (DgpCoord_isRotated(pState->o, ORIENT_RD) ? h : w);
            nh = (DgpCoord_isRotated(pState->o, ORIENT_RD) ? w : h);
        } else {
            nw = (DgpCoord_isRotated(pState->o, ORIENT_RD) ? h : w);
            nh = (DgpCoord_isRotated(pState->o, ORIENT_RD) ? w : h);
        }

        if((pState->stripSize <= 0) || (pState->stripSize > nh)){
            pState->stripSize = nh;
        }
        if(pState->stripwise){
            nh = pState->stripSize;
        }

        scale_factor = DgpCoord_float(DgpCoord_div(pState->output_res.xRes, ppage->res.xRes));
        nw = (Dgp_Int32)(nw * scale_factor);
        nh = (Dgp_Int32)(nh * scale_factor);

        if (pState->force){
            if (pState->forceType == santaPageColor){
                inputtype = santaInputRGB;
            }
            else if (pState->forceType == santaPageMono){
                inputtype = santaInputMonoPacked;
            }
        } else if(pState->do_cmyk && (ppage->page_type == santaPageColor)){
            inputtype = santaInputCMYKAll;
        } else { 
            if (ppage->page_type == santaPageColor){
                inputtype = santaInputRGB;
            }
            else if (ppage->page_type == santaPageMono){
                inputtype = santaInputMonoPacked;
            }
        }

        if(inputtype == santaInputRGB){
			rgbImageOpts = DGPIMAGE_RGB | DGPIMAGE_DONTCLEAR;
            DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, rgbImageOpts, nw, nh, &out));
            
            dst = (void *)DGPIMG_PIXPTR((DgpImage_ColorRGB)out, 0, 0);
            if (nh != 0){
                dstBpl = (Dgp_Int8 *)DGPIMG_PIXPTR((DgpImage_ColorRGB)out, 0, 1) - 
                    (Dgp_Int8 *)DGPIMG_PIXPTR((DgpImage_ColorRGB)out, 0, 0);
            }
            else {
                dstBpl = 0;
            }
        }
        else if(inputtype == santaInputMonoPacked){
            DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_PACKED, nw, nh, &out));
            dst = (void *)DGPIMG_GETROW((DgpImage_Packed)out, 0);
            dstBpl = DGPIMG_GETROWWIDTH((DgpImage_Packed)out);
        }
        else if(inputtype == santaInputCMYKAll){
            DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_GRAY, nw, nh, &cmyk_img[0]));
            DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_GRAY, nw, nh, &cmyk_img[1]));
            DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_GRAY, nw, nh, &cmyk_img[2]));
            DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_GRAY, nw, nh, &cmyk_img[3]));
            cmyk_dst[0] = (void *)DGPIMG_GETROW((DgpImage_Gray)cmyk_img[0], 0);
            cmyk_dst[1] = (void *)DGPIMG_GETROW((DgpImage_Gray)cmyk_img[1], 0);
            cmyk_dst[2] = (void *)DGPIMG_GETROW((DgpImage_Gray)cmyk_img[2], 0);
            cmyk_dst[3] = (void *)DGPIMG_GETROW((DgpImage_Gray)cmyk_img[3], 0);
            dst = cmyk_dst;
            out = cmyk_img;
            if (nh != 0){
                dstBpl = (Dgp_Int8 *)DGPIMG_PIXPTR((DgpImage_ColorRGB)cmyk_img[0], 0, 1) - 
                    (Dgp_Int8 *)DGPIMG_PIXPTR((DgpImage_ColorRGB)cmyk_img[0], 0, 0);
            }
            else {
                dstBpl = 0;
            }
        }

        /* For stripwise decodes, we're going to open the outgoing files here and let santa write to them */
        /* as it fills the strips. */
        if(pState->stripwise && pState->outputFiles){
            if(inputtype == santaInputCMYKAll){
                Dgp_Int8 fnm[256];
                Dgp_Int8 chars_cmyk[5] = "cmyk";
                for(i = 0; i < 4; i++){
                    sprintf(fnm, "%s%c.%s", outName, chars_cmyk[i], pState->ext);
                    DGP_CHECK_NULL_CLEANUP(outgoing[i] = pState->to_stdout ? stdout:fopen(fnm, "wb"));
                }
            } else {
                DGP_CHECK_NULL_CLEANUP(outgoing[0] = pState->to_stdout ? stdout:fopen(outName, "wb"));
            }
        }

		/* layerOptions... */
		if(pState->jbig2_only ^ pState->images_only){
			layerOptions = (!pState->jbig2_only << DGP_SHOWIMAGES_BITSHIFT) | (!pState->images_only << DGP_SHOWJBIG2_BITSHIFT);
		}

        DGP_ERR_PASS_CLEANUP(santaRenderInStrips(   pdoc, 
                                                    ppage, 
                                                    pState->o, 
                                                    pState->stripSize, 
                                                    x, y, 
                                                    w, h,
                                                    santaJobOptsDefault, 
                                                    inputtype,
                                                    dstBpl, 
                                                    dst, 
                                                    out, 
                                                    DGP_FALSE,		/* Is buffer cleared? */
                                                    dgp_state, 
                                                    outgoing,
                                                    (Dgp_Bool)(pState->stripwise || (Dgp_Bool)((!pState->outputFiles) && (pState->stripSize < nh))),
                                                    layerOptions));
        
        if(pState->outputFiles){
            /* Output image is only monochrome if we haven't been asked to scale. */
            if(inputtype == santaInputMonoPacked){
				/* If forcing mono output on a color document, or if scaling to a new resolution, */
				/* we have to binarize santa's output. */
				binaryOut = out;

                handleMonoPage( pState, 
                                outName, 
                                (Dgp_Int32)DgpCoord_float(ppage->res.xRes),
                                (Dgp_Int32)DgpCoord_float(ppage->res.yRes),  
                                (DgpImage_Packed)binaryOut, 
                                dgp_state,
                                &tiff,
                                pgNum == pState->start? cTrue : cFalse,
                                pgNum == (pState->stop -1)? cTrue : cFalse);
				renderedWidth = DGPIMG_WIDTH((DgpImage_Packed)binaryOut);
				renderedHeight = DGPIMG_HEIGHT((DgpImage_Packed)binaryOut);

				DGPIMAGE_FREE((DgpImage_Packed)binaryOut);
            }
            else if(inputtype == santaInputRGB){
                handleColorPage(pState, 
                                outName, 
                                (Dgp_Int32)DgpCoord_float(ppage->dst_res.xRes),
                                (Dgp_Int32)DgpCoord_float(ppage->dst_res.yRes),  
                                (DgpImage_ColorRGB)out, 
                                dgp_state,
                                &tiff,
                                pgNum == pState->start ? cTrue : cFalse, 
                                pgNum == (pState->stop - 1) ? cTrue : cFalse);
				renderedWidth = DGPIMG_WIDTH((DgpImage_ColorRGB)out);
				renderedHeight = DGPIMG_HEIGHT((DgpImage_ColorRGB)out);
                DGPIMAGE_FREE((DgpImage_ColorRGB)out);
            }
            else if(inputtype == santaInputCMYKAll){
                handleCMYKPage( pState, 
                                outName, 
                                (Dgp_Int32)DgpCoord_float(ppage->res.xRes),
                                (Dgp_Int32)DgpCoord_float(ppage->res.yRes),  
                                cmyk_img,
                                dgp_state,
                                &tiff,
                                pgNum == pState->start? cTrue : cFalse, 
                                pgNum == (pState->stop - 1) ? cTrue : cFalse);
                DGPIMAGE_FREE((DgpImage_Gray)cmyk_img[0]);
                DGPIMAGE_FREE((DgpImage_Gray)cmyk_img[1]);
                DGPIMAGE_FREE((DgpImage_Gray)cmyk_img[2]);
                DGPIMAGE_FREE((DgpImage_Gray)cmyk_img[3]);
                sprintf(outName, "%s.%0*ld.[c|m|y|k].%s", prefix, (int)numDig, pgNum, pState->ext);
            }

        }

		DGP_REPORTF5(pState->verbose && pState->outputFiles,
			"page %d, %d by %d, %d xspi, %d yspi", 
			(int)pgNum, (int)renderedWidth, (int)renderedHeight, 
			(int)DgpCoord_float(pState->output_res.xRes),
			(int)DgpCoord_float(pState->output_res.yRes));
		DGP_REPORTF1(pState->verbose && pState->outputFiles,
			"%s written.", outName);
		DGP_REPORTF5(pState->verbose && !pState->outputFiles,
			"page %d, %d by %d, %d xspi, %d yspi",
			(int)pgNum, (int)ppage->width, (int)ppage->height, 
			(int)DgpCoord_float(ppage->res.xRes),
			(int)DgpCoord_float(ppage->res.yRes));

        DGP_ERR_PASS_CLEANUP(santaClosePage(pdoc, &ppage, dgp_state));
    }
	DGP_ERR_PASS_CLEANUP(santaCloseDocument(&pdoc, dgp_state));
    for(i = 0; i < 4; i++){
        if(outgoing[i]) fclose(outgoing[i]);
    }
	NORMAL_RETURN;

ERR_CLEANUP:
    if(tiff != NULL) w_cleanupTiff(tiff);
    if(out) DGP_FREE(out);
    if(cmyk_img[0]) DGP_FREE(cmyk_img[0]);
    if(cmyk_img[1]) DGP_FREE(cmyk_img[1]);
    if(cmyk_img[2]) DGP_FREE(cmyk_img[2]);
    if(cmyk_img[3]) DGP_FREE(cmyk_img[3]);
	ERR_RETURN;
}
/*************************************************************************/




