/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Author:  ?, ?
**
**  Filename:  santa_uncomp_all.h
**
**  Front end header for xiff decompressor.
**
**
**	Modification history:
**	    Dillon H. Tracy esq. / March 2000
**		o  Implemented dgp types.
**		o  Search on 'dhtesq' for some of this coder's comments & changes.
**  
**************************************************************************/

#ifndef _SANTA_UNCOMP_ALL_H
#define _SANTA_UNCOMP_ALL_H

#ifdef __cplusplus
extern "C"
{
#endif

#include <math.h>
#include "dgp_err_control.h"
#include "util.h"
#include "orientation_t.h"
#include "santa.h"
#include "santa_page.h"
#include "image.h"
#include "orientation.h"
#include "image_defs.h"
#include "simple_utils.h"
#include <binarize.pub>
#include <color.pub>

#include "tiffio.h"
/* for the error outputs */

typedef struct
{
	Dgp_Bool force;
	santaPageType forceType;

	Dgp_Int32 stripSize;
    Dgp_Bool stripwise;     /* Added for true stripwise decompositions. / dhtesq. */
	
	DgpCoord_Orientation o;

	Dgp_Int32 start;
	Dgp_Int32 stop;

	Dgp_Int32 outtype;
	Dgp_Int8 ext[32];

	Dgp_Bool outputFiles;
	Dgp_Bool verbose;

    Dgp_Int8 *output_fname;

    DgpCoord_Resolution output_res;
    Dgp_Bool scale_me;
    Dgp_Bool do_cmyk;       /* Added 24.8.2000 for cmyk (all separations at once) decodes. */
    Dgp_Bool jbig2_only;    /* Added 14.9.2000 for looking at token data only. */
	Dgp_Bool images_only;	/* Added 3.5.2001 for looking at image data only. */

    Dgp_Bool to_stdout;     /* Added 7.9.2000. */
} santaUncompAllState;

Dgp_Bool santaUncompGetType(Dgp_Int8 *pType, Dgp_Int32 *pOutType, 
			    Dgp_Int8 *outExt);

void santaUncompInitState(santaUncompAllState *pState);

ErrCode santaUncompProcessDocument(santaUncompAllState         *pState,
                                   Dgp_Int8                    *docName,
                                   Dgp_Int8                    *docSource,
                                   uds                         *docUds,
                                   Dgp_State                   *dgp_state);

ErrCode handleMonoPage( santaUncompAllState         *pState, 
                        Dgp_Int8                    *outName, 
                        Dgp_Int32                   pgXRes, 
                        Dgp_Int32                   pgYRes,
                        DgpImage_Packed             pageIm, 
                        Dgp_State                   *dgp_state,
                        TIFF                        **tiff,
                        Bool                        firstPage,
                        Bool                        lastPage);

ErrCode 
handleColorPage(         santaUncompAllState        *pState, 
                         Dgp_Int8                   *outName, 
                         Dgp_Int32                  pgXRes, 
                         Dgp_Int32                  pgYRes, 
                         DgpImage_ColorRGB          pageIm,
                         Dgp_State                  *dgp_state,
                         TIFF                       **tiff,
                         Bool                       firstPage,
                         Bool                       lastPage);


#ifdef __cplusplus
}
#endif

#endif /* _SANTA_UNCOMP_ALL_H */
