/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 25.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename: ccs.c.
**
**  Utilities for counting, blatting, marking, and examining connected components.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#ifndef VXWORKS
#include <memory.h>
#endif /* !(VXWORKS) */
#include <limits.h>
#include <math.h>
#include "image.h"
#include "image_defs.h"
#include "list.h"
#include "dgp_state.h"		/* For liberr exception handling. */
#include "util.h"			/* For INLINE / dhtesq. */
/*#include "stack.h"		// Used? / dhtesq. */
#include "findCC.h"
#include "ccs.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libscancolor/ccs.c,v 1.11 2003/02/10 22:53:36 mkrocken Exp $");

/* Initial stack size for findCC. */
#define STACK_SIZE 100000					/*--? appears nowhere else in file / dhtesq. */
/* define for 4 connectivity in findCC		--? appears nowhere else in file / dhtesq. */
#undef FOUR_CONNECTED


/*************************************************************************
**
**  Function: ErrCode blatComponents(	Dgp_State		*dgp_state,
**										DgpImage_Long		compMap,
**										ComponentInfo	*components,
**										Dgp_Int32		cc_blat,
**										DgpImage_ColorRGB		*_out)
**
**  Blats a pile of components onto a color image using their ComponentInfo{} information.
**	Will only blat cc's with a 'text' field >= cc_blat.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State.
**		o  compMap is the map containing components to be blatted.
**		o  components is an array of the ComponentInfo{} of all the cc's in the component map.
**		o  cc_blat is a threshold;  components having a 'text' field >= cc_blat will be blatted.
**		o  _out is the image to be blatted on.  If set to NULL for a call, I will make the necessary
**		   allocation -- if not, I will assume the storage is safe.  Careful!
**
**	Return value:
**		o  An ErrCode;  will fail on bad allocation of image storage.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling.
**		o  Context modified to take Dgp_State* for exception handling.
**  
**/

ErrCode
blatComponents(Dgp_State *dgp_state, DgpImage_Long compMap, ComponentInfo *components, Dgp_Int32 cc_blat, DgpImage_ColorRGB *_out) {
	DGP_PROLOGUE("blatComponents()", DGP_PERRSTATE);

	Dgp_Int32		cc;
	Dgp_Int32		x, y;
	Dgp_Int32		width, height;
	DgpImage_ColorRGB		out;
	DgpImage_RGB	not_assigned;

	width = DGPIMG_WIDTH(compMap) - 2;		/* compMap has border of 1 pixel... */
	height = DGPIMG_HEIGHT(compMap) - 2;

	/* Allocate output if necessary... */
	if (*_out == NULL) {
		DGP_ERR_PASS(DgpImage_new(dgp_state, DGPIMAGE_RGB, width, height, (void **)&out));
	} else {
		/*	Caller has already done it... */
		out = *_out;
	}

	/* "Not assigned" color... */
	not_assigned.r = 255;
	not_assigned.g = 255;
	not_assigned.b = 255;

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			cc = DGPIMG_REF(compMap, x, y);
			if (cc == NOT_ASSIGNED) {
				DGPIMG_REF(out, x, y) = not_assigned;
			} else {
				if ((Dgp_Int32)components[cc].text >= cc_blat) {
					DGPIMG_REF(out, x, y).r = (Dgp_UInt8)components[cc].color.r;
					DGPIMG_REF(out, x, y).g = (Dgp_UInt8)components[cc].color.g;
					DGPIMG_REF(out, x, y).b = (Dgp_UInt8)components[cc].color.b;
				} else {
					DGPIMG_REF(out, x, y) = not_assigned;
				}
			}
		}
	}

	NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function: void blatComponent(	DgpImage_ColorRGB		out,
**									DgpImage_Long		compMap,
**									ComponentInfo	*components,
**									Dgp_Int32		bcc)
**
**  Blats a single component onto a color image using its existing ComponentInfo{} information.
**	The component to be blatted has index bcc.
**
**	Arguments:
**		o  out is the image to be written to.
**		o  compMap is the map containing component to be blatted.
**		o  components is an array of the ComponentInfo{} of all the cc's in the component map.
**		o  bcc is the component map index of the component to be blatted.
**
**	Return value:
**		o  None.  No outgoing image allocation takes place here.
**
**	Modification history:
**		o  None.
**  
**/

void
blatComponent(DgpImage_ColorRGB out, DgpImage_Long compMap, ComponentInfo *components, Dgp_Int32 bcc) {

	Dgp_Int32 cc;
	Dgp_Int32 x, y;
	Dgp_Int32 width, height;

	width = DGPIMG_WIDTH(out);
	height = DGPIMG_HEIGHT(out);

	/*	Given that we have the component's center and extent in components[bcc], why */
	/*	are we looping over the entire component map?  Why not loop from minX->maxX, minY->maxY? */
	/*	/ dhtesq. */
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			cc = DGPIMG_REF(compMap, x, y);
			if (cc == bcc) {
				DGPIMG_REF(out, x, y).r = (Dgp_UInt8)components[cc].color.r;
				DGPIMG_REF(out, x, y).g = (Dgp_UInt8)components[cc].color.g;
				DGPIMG_REF(out, x, y).b = (Dgp_UInt8)components[cc].color.b;
			}
		}
	}
}
/*************************************************************************/


/*************************************************************************
**
**  Function: Dgp_Int32 countMarked(ComponentInfo *components, Dgp_Int32 numccs)
**
**
**  Counts number of connected components in the passed array which have a userdata field != NULL.
**
**	In practice, the userdata member is set to (void *)1 in getSimilarComponents() in edges.c
**	for each considered component ("marked").
**
**	Arguments:
**		o  components is an array of the ComponentInfo{} of all the cc's in the component map.
**		o  numccs is length(components).
**
**	Return value:
**		o  Number of marked components, 0 <= n <= numccs.
**
**	Modification history:
**		o  None.
**  
**/

Dgp_Int32
countMarked(ComponentInfo *components, Dgp_Int32 numccs) {
	Dgp_Int32 cc;
	Dgp_Int32 num = 0;

	for (cc = 0; cc < numccs; cc++) {
		if (components[cc].userdata != NULL){
			num++;
		}
	}
	return num;
}
/*************************************************************************/


/*************************************************************************
**
**  Function: ErrCode getProperties(Dgp_State		*dgp_state,
**									DgpImage_ColorRGB		colors,
**									DgpImage_Long		compMap,
**									ComponentInfo	*components,
**									Dgp_Int32		numccs)
**
**
**  Calculates or gets someone to calculate most of the entries in ComponentInfo{}
**	for a given image, component map, and component array.  The complexity measure is handed off
**	to the complexity.c module.
**
**
**	Arguments:
**		o  dgp_state is a valid Dgp_State* used for error handling.
**		o  colors is the image the cc's were taken from.
**		o  compMap is the component map for colors.
**		o  components is the array of cc's to be filled in.
**		o  numccs is length(components).
**
**	Return value:
**		o  An ErrCode.  Can fail on allocation of storage for running sum of cc colors.
**
**	Modification history:
**		Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling.
**		o  Context modified to take Dgp_State* for exception handling.
**		o  Changed doubles to single-precision.  This whole function should be kicked up
**		to the cc finder.  Surprisingly expensive -- ~10% of encoding time.
**  
**/

ErrCode
getProperties(Dgp_State *dgp_state, DgpImage_ColorRGB colors, DgpImage_Long compMap, ComponentInfo *components, Dgp_Int32 numccs) {
	DGP_PROLOGUE("getProperties()", DGP_PERRSTATE);

	Dgp_Int32				x, y, num_pts, ccX, ccY;
	Dgp_Int32				cc, cc1, cc2, cc3, cc4, cc5;
	Dgp_Int32				width = DGPIMG_WIDTH(compMap) - 2;		/* compMap has border of 1 pixel... */
	Dgp_Int32				height = DGPIMG_HEIGHT(compMap) - 2;
	Dgp_Float32			    *r, *g, *b;

	/* This sucks... I have to use doubles because 32 bits could overflow. */
	/*_____________________________________________________________________ */
	/*	These will be used to hold the average color of a given component, */
	/*	but in the interim must hold a large running sum.  Storage in components */
	/*	is already typed for Dgp_Int32. */
	/* */
	/*	I think colors in a given component will only deviate from the mean by some specified parameter epsilon, */
	/*	so perhaps we could operate on (r - r0) instead of r? / dhtesq. */
	r = g = b = NULL;
	DGP_CHECK_ALLOC_CLEANUP(r = (Dgp_Float32 *)DGP_CALLOC(1, numccs * sizeof(Dgp_Float32)));
	DGP_CHECK_ALLOC_CLEANUP(g = (Dgp_Float32 *)DGP_CALLOC(1, numccs * sizeof(Dgp_Float32)));
	DGP_CHECK_ALLOC_CLEANUP(b = (Dgp_Float32 *)DGP_CALLOC(1, numccs * sizeof(Dgp_Float32)));

	/* Something wrong with calloc()? / dhtesq. */
	/*
	memset(r, 0, numccs*sizeof(Dgp_Float64));
	memset(g, 0, numccs*sizeof(Dgp_Float64));
	memset(b, 0, numccs*sizeof(Dgp_Float64));
	*/

	for (cc = 0; cc < numccs; cc++) {
		components[cc].minX = INT_MAX;
		components[cc].minY = INT_MAX;
		components[cc].maxY = 0;
		components[cc].maxX = 0;
		components[cc].num_pts = 0;
		components[cc].ccnum = cc;
	}

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			cc = DGPIMG_REF(compMap, x, y);
			if (cc != NOT_ASSIGNED) {
				components[cc].minX = MIN(x, components[cc].minX);
				components[cc].minY = MIN(y, components[cc].minY);
				components[cc].maxX = MAX(x, components[cc].maxX);
				components[cc].maxY = MAX(y, components[cc].maxY);
				components[cc].num_pts++;
				r[cc] += (Dgp_Float32)DGPIMG_REF(colors, x, y).r;
				g[cc] += (Dgp_Float32)DGPIMG_REF(colors, x, y).g;
				b[cc] += (Dgp_Float32)DGPIMG_REF(colors, x, y).b;	
			}
		}
	}

	for (cc = 0; cc < numccs; cc++) {
		components[cc].width = (components[cc].maxX - components[cc].minX) + 1;
		components[cc].height = (components[cc].maxY - components[cc].minY) + 1;
		components[cc].size = MAX(components[cc].width, components[cc].height);
		components[cc].centerX = components[cc].minX + components[cc].width/2;
		components[cc].centerY = components[cc].minY + components[cc].height/2;
		components[cc].color.r = (Dgp_UInt8)(r[cc]/components[cc].num_pts);
		components[cc].color.g = (Dgp_UInt8)(g[cc]/components[cc].num_pts);
		components[cc].color.b = (Dgp_UInt8)(b[cc]/components[cc].num_pts);
        r[cc] = 0;
        g[cc] = 0;
        b[cc] = 0;
	}

    /* redefine color */
	for (cc = 0; cc < numccs; cc++) {
        if(components[cc].width > 2 && components[cc].height > 2){
            num_pts = 0;
	        for (y = 1; y < components[cc].height-1; y++) {
                for( x = 1; x < components[cc].width-1; x++){
                    ccX = components[cc].minX + x;
                    ccY = components[cc].minY + y;
			        cc1  = DGPIMG_REF(compMap, ccX, ccY);
                    cc2 = DGPIMG_REF(compMap,  ccX -1, ccY);
                    cc3 = DGPIMG_REF(compMap, ccX , ccY -1);
                    cc4 = DGPIMG_REF(compMap, ccX + 1 , ccY);
                    cc5 = DGPIMG_REF(compMap, ccX , ccY + 1);
                    if (cc1 == cc2 && cc2 == cc3 && cc3 == cc4 && cc4 == cc5 && cc5 == cc) {
				        r[cc] += (Dgp_Float32)DGPIMG_REF(colors, ccX, ccY).r *2;
				        g[cc] += (Dgp_Float32)DGPIMG_REF(colors, ccX, ccY).g *2;
				        b[cc] += (Dgp_Float32)DGPIMG_REF(colors, ccX, ccY).b *2;
                        num_pts += 2 ;
                    }else if(cc1 ==cc){
				        r[cc] += (Dgp_Float32)DGPIMG_REF(colors, ccX, ccY).r;
				        g[cc] += (Dgp_Float32)DGPIMG_REF(colors, ccX, ccY).g;
				        b[cc] += (Dgp_Float32)DGPIMG_REF(colors, ccX, ccY).b;
                        num_pts++;
                    }
                }
            }
            if(num_pts >0){
		        components[cc].color.r = (Dgp_UInt8)(r[cc]/num_pts);
		        components[cc].color.g = (Dgp_UInt8)(g[cc]/num_pts);
		        components[cc].color.b = (Dgp_UInt8)(b[cc]/num_pts);
            }
        }
    }
	DGP_FREE(r);
	DGP_FREE(g);
	DGP_FREE(b);
	NORMAL_RETURN;

ERR_CLEANUP:
	if(r != NULL){
		DGP_FREE(r);
	}
	if(g != NULL){
		DGP_FREE(g);
	}
	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function: ErrCode getComponents(Dgp_State	    	*dgp_state,
**									DgpImage_ColorRGB	colorIm,
**									DgpImage_Long		compMap,
**									Dgp_Int32		    numccs,
**									ComponentInfo	    **_components)
**
**
**  Allocates storage for connected component descriptions and calls
**	getProperties() to fill in the ComponentInfo{}'s.
**
**
**	Arguments:
**		o  dgp_state is a valid Dgp_State* used for error handling.
**		o  colorIm is the image the cc's were taken from.
**		o  compMap is the component map for colorIm.
**		o  numccs is the number of cc's found in colorIm.
**		o  _components points to an array of ComponentInfo{}'s on successful completion.
**		
**
**	Return value:
**		o  An ErrCode.  Can fail on allocation or on fill.
**
**	Modification history:
**		Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling;  context modified to take Dgp_State*.
**		o  _components added to argument list.
**  
**/

ErrCode
getComponents(Dgp_State         *dgp_state, 
              DgpImage_ColorRGB colorIm, 
              DgpImage_Long     compMap, 
              Dgp_Int32         numccs, 
              ComponentInfo     **_components) {
	DGP_PROLOGUE("getComponents()", DGP_PERRSTATE);
	
	ComponentInfo *components = NULL;

	/* Allocate for each ComponentInfo{}. */
	DGP_CHECK_ALLOC(components = DGP_MALLOC(numccs * sizeof(ComponentInfo)));

	/* Fill the allocated structures. */
	DGP_ERR_PASS_CLEANUP(getProperties(dgp_state, colorIm, compMap, components, numccs));
	
	*_components = components;
	NORMAL_RETURN;

ERR_CLEANUP:
	if(components != NULL){
		DGP_FREE(components);
	}
	ERR_RETURN;
}

Dgp_Float32
ccDistance(ComponentInfo *cc1, ComponentInfo *cc2) {
	Dgp_Int32 dx, dy;

	dx = cc2->centerX - cc1->centerX;
	dy = cc2->centerY - cc1->centerY;

	return (Dgp_Float32)sqrt(dx*dx + dy*dy);
}


/*************************************************************************/
