/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 31.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename:  chains.h.
**
**  Types & prototypes for chain construction module.  No exports aside from
**	debugging functions.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**		o  Added Dgp_State *dgp_state to chain{}, since chain{} functions as the cookie for
**		   most of this module.  This will simplify calling conventions.
**
**  
**************************************************************************/


#ifndef CHAINS_H
#define CHAINS_H

#include "dgp_state.h"

typedef struct {
	Dgp_Float32		alignEpsilon;
	Dgp_Float32		alignInlierEpsilon;
	Dgp_Int32		minChainSize;
	Dgp_Float32		baseDot;
	Dgp_Int32		minAvgPts;
	Dgp_Int32		maxAvgPts;
	Dgp_Float32		sizeEpsilon;
	Dgp_Int32		colorEpsilon;
} chainOptions;

typedef struct {
	Dgp_Int32		minred, maxred, mingreen, maxgreen, minblue, maxblue;
	Dgp_Int32		minSize, maxSize;
	DgpAdt_List		ccs;
	Dgp_Float32		maxBaseDist;
	Dgp_Float32		maxInlierBaseDist;
	Dgp_Float64		dr_cog_x, dr_cog_y, ul_cog_x, ul_cog_y;
	Dgp_Float64		dr_base_x, dr_base_y, ul_base_x, ul_base_y;
	DgpAdt_List		ul, dr;
	Dgp_Bool		horizontal;
	Dgp_State	    *dgp_state;			/* Added for liberr compliance / dhtesq. */
} chain;

void
defaultChainOptions(chainOptions *p, Dgp_Int32 xres, Dgp_Int32 yres);

ErrCode
getChains(Dgp_State *dgp_state, ComponentInfo *components, Dgp_Int32 numccs, DgpAdt_List *_chains, chainOptions *p);

ErrCode
freeChains(Dgp_State *dgp_state, DgpAdt_List chains);

void
drawChains(DgpImage_Binary out, DgpAdt_List chains);

void
blatChain(DgpImage_ColorRGB out, DgpImage_Long compMap, chain *pchain, ComponentInfo *components);

void
debugChain(Dgp_Int32 dcc, DgpAdt_List chains, ComponentInfo *components, DgpImage_Long compMap, DgpImage_ColorRGB tmp);

ErrCode
saveChains(Dgp_State *dgp_state, DgpImage_Long compMap, DgpAdt_List chains, ComponentInfo *components);

#endif
