/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 26.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename:  findCC.c
**
**
**  Your typical find connected components code.  Currently we have an implementation
**	only for packed binary images.  "Connected components" are, for these purposes,
**	any set of similarly colored points sharing 8-connectivity.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/

#include "dgp_state.h"
#include "image_defs.h"
#include "stack.h"
#include "imutil.h"
#include "findCC.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libscancolor/findCC.c,v 1.8 2003/02/11 01:31:43 mkrocken Exp $");


/* How big should we start the stack used for the "recursive" find cc? */
#define STACK_SIZE 100000

/* Define this for 4-connected components (instead of 8-connected -- dhtesq.) */
#undef FOUR_CONNECTED


/*	Macro for getCC_pb() only. / dhtesq. */
#define pushPoint(x,y) \
if ((DGPIMG_REF(compMap, x, y) == NOT_ASSIGNED) && ((DGPIMG_GETBIT(image, x, y) == 0) == v)) { \
	if (!DgpAdt_staPush(xstack,(void *)(x))) { \
		DGP_ERR_PASS(DgpAdt_staRealloc(dgp_state, xstack, 2*DgpAdt_staMaxNumNodes(xstack))); \
		DgpAdt_staPush(xstack,(void *)(x)); \
		DGP_ERR_PASS(DgpAdt_staRealloc(dgp_state, ystack, 2*DgpAdt_staMaxNumNodes(ystack))); \
        DgpAdt_staPush(ystack,(void *)(y)); \
	} else \
        DgpAdt_staPush(ystack,(void *)(y)); \
	DGPIMG_REF(compMap, x, y) = ccnum; }


/*************************************************************************
**
**  Function: ErrCode getCC_pb(	Dgp_State		*dgp_state,
**								DgpImage_Packed		image,
**								Dgp_Int32				sx,
**								Dgp_Int32				sy,
**								DgpImage_Long		compMap,
**								Dgp_Int32				ccnum,
**								DgpAdt_Stack			xstack,
**								DgpAdt_Stack			ystack,
**								Dgp_Int32				*_size);
**
**
**  "_pb" == "packed binary".  Finds the size of the cc at the passed point in the passed image
**	and enters it on the passed component map.  (Packed images have data type Dgp_UInt32.)
**
**
**	Arguments:
**		o  dgp_state is a valid Dgp_State* used for error handling.
**		o  image is the packed image containing the cc.
**		o  (sx, sy) are the coordinates on image of a point on the cc.
**		o  compMap is the component map to be updated;  if the cc is later found for whatever reason to be
**		   unsatisfactory, this map must be manually revised.
**		o  ccnum is the index with which to label the found component.
**		o  xstack and ystack are prepared stacks used for exploring the connectivity of the passed point.
**		o  _size is the location where the cc's size in pixels will be placed on successful return.
**
**
**	Return value:
**		o  An ErrCode.  Passes up stack reallocation errors.
**
**	Modification history:
**		Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling.
**		o  Context modified to take Dgp_State* for exception handling.
**  
**/

ErrCode
getCC_pb(	Dgp_State	*dgp_state,
			DgpImage_Packed image,
			Dgp_Int32			sx,
			Dgp_Int32			sy,
			DgpImage_Long	compMap,
			Dgp_Int32			ccnum,
			DgpAdt_Stack		xstack,
			DgpAdt_Stack		ystack,
			Dgp_Int32			*_size) {
	DGP_PROLOGUE("getCC_pb()", DGP_PERRSTATE);
	Dgp_Int32			x, y;
	Dgp_Int32			size = 0;

	/*	To connect with point (sx, sy), a given neighboring point must have the value v: */
    /* Why not just use the bit's value?  Must be lhs? / dhtesq. */
	Dgp_Int32			v = (DGPIMG_GETBIT(image, sx, sy) == 0);	

	/* Add the point to our stack. */
	pushPoint(sx, sy);

	while (!DgpAdt_staEmpty(xstack)) {
		/* Pop the stack to get the current point. */
		x = (Dgp_Int32)DgpAdt_staPop(xstack);
		y = (Dgp_Int32)DgpAdt_staPop(ystack);
		size++;

		/* Push neighbors. */
		pushPoint(x-1, y);
		pushPoint(x+1, y);
		pushPoint(x, y+1);
		pushPoint(x, y-1);
#ifndef FOUR_CONNECTED
		pushPoint(x-1, y-1);
		pushPoint(x+1, y+1);
		pushPoint(x-1, y+1);
		pushPoint(x+1, y-1);
#endif
    }

	*_size = size;
	NORMAL_RETURN;
}
/*************************************************************************/

#undef pushPoint

/*
14.5.2001 / dhtesq.
As an experiment, try using the runs-based, callback-based ipcore cc finder
instead of the stack-based approach, which profiling shows to be very slow.

We still mark the component map as usual.
*/

/*	Macro for updateCompMap() only. / dhtesq. */
static ErrCode
updateCompMap(IPState		*ipwstate, 
			  void			*data, 
			  Int32			x, 
			  Int32			y,
			  Int32			w, 
			  Int32			h, 
			  Int32			m, 
			  Int32			nRuns,
			  TCScanLineRun *pRuns)
{
  CCContext *ccContext = (CCContext *)data;
  Dgp_State *dgp_state = ccContext->dgp_state;
  DGP_PROLOGUE("updateCompMap()", DGP_PERRSTATE);
  DgpImage_Long compMap = *(ccContext->compMap);
  Dgp_Int32 yIndex = y + h, xIndex, lastRunLeft = 0;
  
  DGP_STOP_WHINE;
  ipwstate = ipwstate;
  x = x;
  w = w;
  m = m;
  nRuns = nRuns;

  /*
   * Read out the runs. They were generated bottom-to-top...
   */
  while (pRuns) {
    if (lastRunLeft <= pRuns->runRight + 1) {
      yIndex--;
    }

	for(xIndex = pRuns->runLeft; xIndex <= pRuns->runRight; xIndex++){
		/*DGP_ASSERT(DGPIMG_REF(compMap, xIndex, yIndex) == NOT_ASSIGNED); */
		DGPIMG_REF(compMap, xIndex - 32, yIndex - 32) = ccContext->numccs;
	}

    lastRunLeft = pRuns->runLeft;      
    pRuns = pRuns->pNext;
  }

  /* Having reached here, we are guaranteed the cc exceeds the size threshold. */
  ccContext->numccs++;

  NORMAL_RETURN;
}


/*************************************************************************
**
**  Function: ErrCode findCC_pb(	Dgp_State		*dgp_state,
**									DgpImage_Packed		image,
**									DgpImage_Long		*_compMap,
**									Dgp_Int32				*_numccs)
**
**
**  Finds and counts the connected components in the passed image.  Fills in a component map,
**	allocating storage if necessary.  The component map has a one pixel BORDER, but the coordinates
**	are arranged such that (0, 0) is real data;  each component has a unique label in the range
**	0...numccs - 1.
**
**	Currently cc's of size one are discarded.
**
**
**	Arguments:
**		o  dgp_state is a valid Dgp_State* used for error handling.
**		o  image is the packed image to be torn apart for cc's.
**		o  _compMap will point to a filled component map on successful return.  If passed as NULL,
**		   this function will allocate for the map.
**		o  _numccs will point to the number of found cc's on successful completion.
**
**	Return value:
**		o  An ErrCode;  fails on stack or map creation, or if stack reallocation goofs in getCC_pb().  On an error, this
**		   function will free the component map iff the passed _compMap is NULL.  Otherwise the caller is responsible
**		   for it.
**
**	Modification history:
**		Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling.
**		o  Context modified to take Dgp_State* for exception handling.
**  
**/

ErrCode
findCC_pb(Dgp_State			*dgp_state, 
		  DgpImage_Packed	image, 
		  PIXR				*binaryPixr,
		  DgpImage_Long		*_compMap, 
		  Dgp_Int32			sizeThresh,
		  Dgp_Int32			*_numccs) {
	DGP_PROLOGUE("findCC_pb()", DGP_PERRSTATE);
	Dgp_Int32			width = DGPIMG_WIDTH(image);
	Dgp_Int32			height = DGPIMG_HEIGHT(image);
	Dgp_Int32			numccs = *_numccs;
	DgpImage_Long		compMap;
	Dgp_Int32			x, y;/* size; */
	DgpAdt_Stack		xstack = NULL, ystack = NULL;
	CCContext			ccContext;
	IPWState			*ipwstate = DGP_PIPWSTATE;
	UInt8				*w_image = NULL;
	Dgp_Int32			bpl =  DGPIMG_GETROWWIDTH(image);
	Int32				returnValue;


	/* Get some memory for the blob map if input argument is unassigned. */
	if (*_compMap == NULL) {
		DGP_ERR_PASS(DgpImage_newOffset(dgp_state, 
                                        DGPIMAGE_LONG | DGPIMAGE_DONTCLEAR, 
                                        width + 2, 
                                        height + 2, 
                                        -1, 
                                        -1, 
                                        (void **)&compMap));
		imSetL(compMap, NOT_ASSIGNED);
	} else {
		compMap = *_compMap;
	}

	/*	Blob map has 1 pixel border with special value to allow faster implementation of findCC (?;  getCC_pb()? / dhtesq.). */
	/* */
	/*	Remember, an array index of -1 will be interpreted as the address preceding the nominal origin */
	/*	of the image, which has been allocated with a 1 pixel safe border.  Makes edge cases a little easier */
	/*	to deal with.  / dhtesq. */
	for (x = -1; x <= width; x++) {
		DGPIMG_REF(compMap, x, -1) = BORDER;
		DGPIMG_REF(compMap, x, height) = BORDER;
	}
	for (y = -1; y <= height; y++) {
		DGPIMG_REF(compMap, -1, y) = BORDER;
		DGPIMG_REF(compMap, width, y) = BORDER;
	}

	/* DgpAdt_Stack for findCC (?;  getCC_pb? / dhtesq.). */
	/*DGP_ERR_PASS_CLEANUP(DgpAdt_staNew(dgp_state, STACK_SIZE, &xstack)); */
	/*DGP_ERR_PASS_CLEANUP(DgpAdt_staNew(dgp_state, STACK_SIZE, &ystack)); */

	DGP_ERR_PASS_CLEANUP(w_pixrGetImageAll(ipwstate, binaryPixr, &w_image));

	ccContext.numccs = numccs;
	ccContext.compMap = &compMap;
	ccContext.image = &image;
	ccContext.dgp_state = dgp_state;
	ccContext.connectedness = 1;	/* 1 for 8-connected, 0 for 4-connected. */

	DGP_ERR_PASS_CLEANUP(i_findConnectedComponentRuns(	
                                                    ipwstate, 
							                        w_image, 
							                        (unsigned)bpl, 
							                        0, 
							                        0,
						                            width, 
							                        height,
						                            (unsigned)sizeThresh, 
							                        (unsigned)sizeThresh,
						                            ccContext.connectedness, 
						                            (void *)&ccContext,
						                            updateCompMap, 
						                            &returnValue));

	/* Thank purify. */
	/*DgpAdt_staFree(dgp_state, xstack); */
	/*DgpAdt_staFree(dgp_state, ystack); */

	*_compMap = compMap;
	*_numccs = ccContext.numccs;
	NORMAL_RETURN;
	
ERR_CLEANUP:
	if(xstack != NULL){
		DgpAdt_staFree(dgp_state, xstack);
	}
	if(ystack != NULL){
		DgpAdt_staFree(dgp_state, ystack);
	}
	/*	If I have allocated for the component map, I will free it here.  Otherwise leave it alone. */
	if(*_compMap == NULL){
		DGP_ERR_PASS(DgpImage_free(compMap));
	}
	ERR_RETURN;
}
/*************************************************************************/
