/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 21.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename: imutil.c.
**
**  "Random image utilities."  Setting, clearing, and drawing to various types of images.
**	One function for converting PIXR to DgpImage_ColorRGB.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling;  not much affected.
**		o  static SwapInt() introduced to simplify the drawLineX() functions.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/

#include <libutilw.pub>
#include <libpixrw.pub>
#include <libiow.pub>
#include "dgp_state.h"
#include "image.h"
#include "image_defs.h"
/*#include "pixr.h" */
#include "imutil.h"
#include "endianness.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libscancolor/imutil.c,v 1.12 2003/02/10 22:53:37 mkrocken Exp $");

/*************************************************************************
**
**  Function:  void drawLineB(DgpImage_Binary out, Dgp_Int32 x1, Dgp_Int32 y1, Dgp_Int32 x2, Dgp_Int32 y2)
**
**
**  Draws a line segment on the passed image joining the two passed points.  There
**	is no bounds checking on the two passed points.
**
**	Arguments:
**		o  out is the prepared DgpImage_Binary{} to be drawn to.
**		o  (x1, y1) and (x2, y2) are the two points on the image to be joined.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None;  might be wise to trap for xrange == yrange == 0.  Seems function is primarily used
**		   for debugging.
**
**  
**/

void
drawLineB(DgpImage_Binary out, Dgp_Int32 x1, Dgp_Int32 y1, Dgp_Int32 x2, Dgp_Int32 y2) {
	Dgp_Int32 i;
	Dgp_Int32 xrange, yrange;
	
	xrange = ABS(x2 - x1);
	yrange = ABS(y2 - y1);

	/*	Consider cases (xrange > yrange), (yrange > xrange) independently. */
	/*	This ensures stepping is done along the longer axis of the segment. / dhtesq. */
	if (xrange > yrange) {
		/* Line is nominally horizontal. */
		/* Ensure small x comes first, swapping arguments if necessary... */
		if (x2 < x1) {
			swapInt(&x1, &x2);
			swapInt(&y1, &y2);
		}
		/* Need signed yrange for slope calculation. */
		yrange = y2 - y1;
		for (i = 0; i <= xrange; i++) {
			DGPIMG_REF(out, x1 + i, y1 + ((i*yrange)/xrange)) = 1;
		}
	} else {
		/* Line is nominally vertical. */
		/* Ensure small y comes first, swapping arguments if necessary... */
		if (y2 < y1) {
			swapInt(&x1, &x2);
			swapInt(&y1, &y2);
		}
		/* Need signed xrange for slope calculation. */
		xrange = x2 - x1;
		for (i = 0; i <= yrange; i++) {
			/* Possible division by zero here / dhtesq. */
			DGPIMG_REF(out, x1 + ((i*xrange)/yrange), y1 + i) = 1;
		}
	}
	return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  void drawLineRGB(DgpImage_ColorRGB      out, 
**                              DgpImage_RGB           *color, 
**                              Dgp_Int32              x1, 
**                              Dgp_Int32              y1, 
**                              Dgp_Int32              x2, 
**                              Dgp_Int32              y2)
**
**
**  Draws a line segment on the passed image joining the two passed points.  There
**	is no bounds checking on the two passed points.  The passed color is used for the line.
**
**	Arguments:
**		o  out is the prepared DgpImage_ColorRGBImage{} to be drawn to.
**		o  (x1, y1) and (x2, y2) are the two points on the image to be joined.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None;  might be wise to trap for xrange == yrange == 0.  Seems function is primarily used
**		   for debugging.
**
**  
**/

void
drawLineRGB(DgpImage_ColorRGB       out, 
            DgpImage_RGB            *color, 
            Dgp_Int32               x1, 
            Dgp_Int32               y1, 
            Dgp_Int32               x2, 
            Dgp_Int32               y2) {
	Dgp_Int32 i;
	Dgp_Int32 xrange, yrange;
	
	xrange = ABS(x2 - x1);
	yrange = ABS(y2 - y1);

	/*	Consider cases (xrange > yrange), (yrange > xrange) independently. */
	/*	This ensures stepping is done along the longer axis of the segment. / dhtesq. */
	if (xrange > yrange) {
		/* Line is nominally horizontal. */
		/* Ensure small x comes first, swapping arguments if necessary... */
		if (x2 < x1) {
			swapInt(&x1, &x2);
			swapInt(&y1, &y2);
		}
		/* Need signed yrange for slope calculation. */
		yrange = y2 - y1;
		for (i = 0; i <= xrange; i++) {
			DGPIMG_REF(out, x1 + i, y1 + ((i*yrange)/xrange)) = *color;
		}
	} else {
		/* Line is nominally vertical. */
		/* Ensure small y comes first, swapping arguments if necessary... */
		if (y2 < y1) {
			swapInt(&x1, &x2);
			swapInt(&y1, &y2);
		}
		/* Need signed xrange for slope calculation. */
		xrange = x2 - x1;
		for (i = 0; i <= yrange; i++) {
			DGPIMG_REF(out, x1 + ((i*xrange)/yrange), y1 + i) = *color;
		}
	}
	return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  void imSetL(DgpImage_Long image, Dgp_Int32 value)
**
**
**  Sets all points in the passed Dgp_Int32 image to the passed value.
**
**	Arguments:
**		o  image is the prepared DgpImage_Long{} to be set.
**		o  value will paint the passed image.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void
imSetL(DgpImage_Long image, Dgp_Int32 value) {
	Dgp_Int32 x, y;
	Dgp_Int32 xmin = DGPIMG_XBASE(image);
	Dgp_Int32 xmax = DGPIMG_XMAX(image);
	Dgp_Int32 ymin = DGPIMG_YBASE(image);
	Dgp_Int32 ymax = DGPIMG_YMAX(image);

	for (y = ymin; y <= ymax; y++) {
		for (x = xmin; x <= xmax; x++) {
			DGPIMG_REF(image, x, y) = value;
		}
	}
	return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  void imSetF(DgpImage_Float image, Dgp_Float32 value)
**
**
**  Sets all points in the passed Dgp_Float32 image to the passed value.
**
**	Arguments:
**		o  image is the prepared DgpImage_Long{} to be set.
**		o  value will paint the passed image.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void
imSetF(DgpImage_Float image, Dgp_Float32 value) {
	Dgp_Int32 x, y;
	Dgp_Int32 xmin = DGPIMG_XBASE(image);
	Dgp_Int32 xmax = DGPIMG_XMAX(image);
	Dgp_Int32 ymin = DGPIMG_YBASE(image);
	Dgp_Int32 ymax = DGPIMG_YMAX(image);

	for (y = ymin; y <= ymax; y++) {
		for (x = xmin; x <= xmax; x++) {
			DGPIMG_REF(image, x, y) = value;
		}
	}
	return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  void imClearRGB(DgpImage_ColorRGB image)
**
**
**  Zeros an DgpImage_ColorRGB image.
**
**	Arguments:
**		o  image is the existing DgpImage_ColorRGBImage{} to be cleared.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void
imClearRGB(DgpImage_ColorRGB image) {
	Dgp_Int32           x, y;
	Dgp_Int32           xmin = DGPIMG_XBASE(image);
	Dgp_Int32           xmax = DGPIMG_XMAX(image);
	Dgp_Int32           ymin = DGPIMG_YBASE(image);
	Dgp_Int32           ymax = DGPIMG_YMAX(image);
	DgpImage_RGB        blank;

	blank.r = 0;
	blank.g = 0;
	blank.b = 0;

	for (y = ymin; y <= ymax; y++) {
		for (x = xmin; x <= xmax; x++) {
			DGPIMG_REF(image, x, y) = blank;
		}
	}
	return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode pixr_to_rgb(Dgp_State *dgp_state, PIXR *pixr, DgpImage_ColorRGB *_im)
**
**
**  Convert PIXR representation to an DgpImage_ColorRGB{}.  This function is responsible for allocation
**	for the destination image.
**
**	Arguments:
**		o  dgp_state references a valid Dgp_State{}.
**		o  pixr is the PIXR to be converted.
**		o  _im will point to the new DgpImage_ColorRGB{} on successful completion.
**
**	Return value:
**		o  ErrCode;  can fail on allocation of DgpImage_ColorRGB image.
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Retrofit for liberr exception handling.
**  
**/

ErrCode
pixr_to_rgb(Dgp_State *dgp_state, PIXR *pixr, DgpImage_ColorRGB *_im) {
	DGP_PROLOGUE("pixr_to_rgb()", DGP_PERRSTATE);
  PIXR			        *red, *green, *blue;
  UInt8			        *rptr, *gptr, *bptr;
  DgpImage_ColorRGBPixPtr		dptr;								/* DgpImage_ColorRGBPixPtr == DgpImage_ColorRGB* (image.h). / dhtesq. */
  Int32			        width, height;
  Dgp_Int32			    i, j;
  IPWState		        *ipstate = dgp_state->ipwstate;
  DgpImage_ColorRGB		im;

  w_pixrGetWidth(ipstate, pixr, &width);
  w_pixrGetHeight(ipstate, pixr, &height);
  
  DGP_ERR_PASS(DgpImage_new(dgp_state, DGPIMAGE_RGB, width, height, (void **)&im));

  w_pixrGetChannel(ipstate, pixr, 0, &red);
  w_pixrGetChannel(ipstate, pixr, 1, &green);
  w_pixrGetChannel(ipstate, pixr, 2, &blue);

  /* dptr is here assigned the address of the data field (why not do this directly? -- OOC? / dhtesq). */
  dptr = DGPIMG_PIXPTR(im, 0, 0);

  for(j = 0; j < height; j++) {
    w_getLine(ipstate, red, j, &rptr);
    w_getLine(ipstate, green, j, &gptr);
    w_getLine(ipstate, blue, j, &bptr);
    for(i = 0; i < width; i++, rptr++, gptr++, bptr++, dptr++) {
/* Clever endianness-switcher.  But don't assume _WIN32 is the only little-endian platform... / dhtesq */
/*#ifdef _WIN32 */
#ifdef _LITTLE_ENDIAN
	  dptr->r = *(UInt8 *)((Dgp_UInt32)rptr ^ 0x3);
      dptr->g = *(UInt8 *)((Dgp_UInt32)gptr ^ 0x3);
      dptr->b = *(UInt8 *)((Dgp_UInt32)bptr ^ 0x3);
#else
	  dptr->r = *(rptr);
      dptr->g = *(gptr);
      dptr->b = *(bptr);
#endif
    }
  }

	*_im = im;
	NORMAL_RETURN;
}
/*************************************************************************/


ErrCode
pixr_to_rgb_gray(Dgp_State *dgp_state, PIXR *pixr, DgpImage_ColorRGB *_im) {
	DGP_PROLOGUE("pixr_to_rgb_gray()", DGP_PERRSTATE);
  PIXR			        *gray;
  UInt8			        *gptr;
  DgpImage_ColorRGBPixPtr		dptr;								/* DgpImage_ColorRGBPixPtr == DgpImage_ColorRGB* (image.h). / dhtesq. */
  Int32			        width, height;
  Dgp_Int32			    i, j;
  IPWState		        *ipstate = dgp_state->ipwstate;
  DgpImage_ColorRGB		im;

  w_pixrGetWidth(ipstate, pixr, &width);
  w_pixrGetHeight(ipstate, pixr, &height);
  
  DGP_ERR_PASS(DgpImage_new(dgp_state, DGPIMAGE_RGB, width, height, (void **)&im));

  gray = pixr;

  /* dptr is here assigned the address of the data field (why not do this directly? -- OOC? / dhtesq). */
  dptr = DGPIMG_PIXPTR(im, 0, 0);

  for(j = 0; j < height; j++) {
    w_getLine(ipstate, gray, j, &gptr);
    for(i = 0; i < width; i++, gptr++, dptr++) {
/* Clever endianness-switcher.  But don't assume _WIN32 is the only little-endian platform... / dhtesq */
/*#ifdef _WIN32 */
#ifdef _LITTLE_ENDIAN
	  dptr->r = *(UInt8 *)((Dgp_UInt32)gptr ^ 0x3);
      dptr->g = *(UInt8 *)((Dgp_UInt32)gptr ^ 0x3);
      dptr->b = *(UInt8 *)((Dgp_UInt32)gptr ^ 0x3);
#else
	  dptr->r = *(gptr);
      dptr->g = *(gptr);
      dptr->b = *(gptr);
#endif
    }
  }

	*_im = im;
	NORMAL_RETURN;
}
/*************************************************************************/

/* Added 23.3.2001 / dhtesq. */
ErrCode
pixr_to_gray(Dgp_State *dgp_state, PIXR *pixr, DgpImage_Gray *_im) {
	DGP_PROLOGUE("pixr_to_gray()", DGP_PERRSTATE);
  PIXR			        *gray;
  UInt8			        *gptr;
  DgpImage_GrayPixPtr	dptr;
  Int32			        width, height;
  Dgp_Int32			    i, j;
  IPWState		        *ipstate = dgp_state->ipwstate;
  DgpImage_Gray			im;

  w_pixrGetWidth(ipstate, pixr, &width);
  w_pixrGetHeight(ipstate, pixr, &height);
  
  DGP_ERR_PASS(DgpImage_new(dgp_state, DGPIMAGE_GRAY, width, height, (void **)&im));

  gray = pixr;

  dptr = DGPIMG_PIXPTR(im, 0, 0);

  if(pixr->pixrData.plane.d == 4){
	  for(j = 0; j < height; j++) {
		  w_getLine(ipstate, gray, j, &gptr);
		  for(i = 0; i < width; i++) {
/* Clever endianness-switcher.  But don't assume _WIN32 is the only little-endian platform... / dhtesq */
/*#ifdef _WIN32 */
#ifdef _LITTLE_ENDIAN
			  if(((i + 1) % 2)){
				  *dptr++ = *(UInt8 *)((Dgp_UInt32)gptr ^ 0x3) & 0xf;
			  } else {
				  *dptr++ = *(UInt8 *)((Dgp_UInt32)gptr ^ 0x3) >> 4;
				  gptr++;
			  }
#else
			  if(((i + 1) % 2)){
				  *dptr++ = *gptr & 0xf;
			  } else {
				  *dptr++ = *gptr++ >> 4;
			  }
#endif
		  }
	  }
  } else {
	  for(j = 0; j < height; j++) {
		  w_getLine(ipstate, gray, j, &gptr);
		  for(i = 0; i < width; i++) {
/* Clever endianness-switcher.  But don't assume _WIN32 is the only little-endian platform... / dhtesq */
/*#ifdef _WIN32 */
#ifdef _LITTLE_ENDIAN
			  *dptr++ = *(UInt8 *)((Dgp_UInt32)gptr ^ 0x3);
			  gptr++;
#else
			  *dptr++ = *gptr++;
#endif
		  }
	  }
  }

	*_im = im;
	NORMAL_RETURN;
}


INLINE void
swapInt(Dgp_Int32 *apple, Dgp_Int32 *orange) {
        Dgp_Int32 temp;

        temp = *apple;
        *apple = *orange;
        *orange = temp;
        return;
}
/*************************************************************************/
