/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 25.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename: propagate.c
**
**  Functions for "growing" components in the various types of images.  Typically called with two images:
**	one component map, which will be altered, and one 'distance' image, which is typically discarded after use.
**	Module is applied both for finding Voronoi regions in edges.c and for distinguishing text from holes
**	in sc_filterText().
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling;  very few changes.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/


#include <limits.h>
#include "dgp_state.h"
#include "image.h"
#include "image_defs.h"
#include "propagate.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libscancolor/propagate.c,v 1.5 2003/02/10 22:53:37 mkrocken Exp $");


/*************************************************************************
**
**  Function: void propagate_g(DgpImage_Gray image, DgpImage_Long dist)
**
**
**  A gray image has an Dgp_UInt8 data type, hence 8-bit grayscale.
**
**	Arguments:
**		o  image is the gray image to be propagated.
**		o  dist is the propagation record.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void
propagate_g(DgpImage_Gray image, DgpImage_Long dist) {
	Dgp_Int32			width, height, xmax, ymax;
	Dgp_Int32			x, y;
	Dgp_Int32		current, up, left, down, right;

	width = DGPIMG_WIDTH(dist);
	xmax = width - 1;
	height = DGPIMG_HEIGHT(dist);
	ymax = height - 1;
	
	/* First, the down right pass. */
	
	/* First line. */
	current = DGPIMG_REF(dist, 0, 0);
	for (x = 1; x < width; x++) {
		left = current + 1;
		current = DGPIMG_REF(dist, x, 0);
		if (left < current) {
			current = DGPIMG_REF(dist, x, 0) = left;
			DGPIMG_REF(image, x, 0) = DGPIMG_REF(image, x-1, 0);
		}
	}
	/* Internal lines. */
	for (y = 1; y < height; y++) {
		/* Leftmost column. */
		up = DGPIMG_REF(dist, 0, y-1) + 1;
		current = DGPIMG_REF(dist, 0, y);
		if (up < current) {
			current = DGPIMG_REF(dist, 0, y) = up;
			DGPIMG_REF(image, 0, y) = DGPIMG_REF(image, 0, y-1);
		}
		/* Internal columns. */
		for (x = 1; x < width; x++) {
			left = current + 1;
			up = DGPIMG_REF(dist, x, y-1) + 1;
			current = DGPIMG_REF(dist, x, y);
			if (left < current) {
				current = DGPIMG_REF(dist, x, y) = left;
				DGPIMG_REF(image, x, y) = DGPIMG_REF(image, x-1, y);
			}
			if (up < current) {
				current = DGPIMG_REF(dist, x, y) = up;
				DGPIMG_REF(image, x, y) = DGPIMG_REF(image, x, y-1);
			}
		}
	}

	/* Upper left pass. */
	
	/* First line. */
	current = DGPIMG_REF(dist, xmax, ymax);
	for (x = xmax-1; x >= 0; x--) {
		right = current + 1;
		current = DGPIMG_REF(dist, x, ymax);
		if (right < current) {
			current = DGPIMG_REF(dist, x, ymax) = right;
			DGPIMG_REF(image, x, ymax) = DGPIMG_REF(image, x+1, ymax);
		}
	}
	/* Internal lines. */
	for (y = ymax-1; y >= 0; y--) {
		/* Rightmost column. */
		down = DGPIMG_REF(dist, xmax, y+1) + 1;
		current = DGPIMG_REF(dist, xmax, y);
		if (down < current) {
			current = DGPIMG_REF(dist, xmax, y) = down;
			DGPIMG_REF(image, xmax, y) = DGPIMG_REF(image, xmax, y+1);
		}
		/* Internal columns. */
		for (x = xmax-1; x >= 0; x--) {
			right = current + 1;
			down = DGPIMG_REF(dist, x, y+1) + 1;
			current = DGPIMG_REF(dist, x, y);
			if (right < current) {
				current = DGPIMG_REF(dist, x, y) = right;
				DGPIMG_REF(image, x, y) = DGPIMG_REF(image, x+1, y);
			}
			if (down < current) {
				current = DGPIMG_REF(dist, x, y) = down;
				DGPIMG_REF(image, x, y) = DGPIMG_REF(image, x, y+1);
			}
		}
	}
}
/*************************************************************************/


/*************************************************************************
**
**  Function: void propagate_rgb(DgpImage_ColorRGB image, DgpImage_Long dist)
**
**
**  As above, for color images.  As called in sc_filterText(), the color image contains the image being parsed
**	and the 'dist' matrix contains (LONG_MAX - 1) at every point where 1) a cc has been assigned and 
**	2) the cc has been associated with a chain, and	therefore may be considered as text.  This means
**	the background will propagate into the connected component rather than vice versa, as it's done in edges.c.
**
**	Arguments:
**		o  image is the color image to be propagated.
**		o  dist is the propagation record.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void
propagate_rgb(DgpImage_ColorRGB image, DgpImage_Long dist) {
	Dgp_Int32			width, height, xmax, ymax;
	Dgp_Int32			x, y;
	Dgp_Int32		current, up, left, down, right;

	width = DGPIMG_WIDTH(dist);
	xmax = width - 1;
	height = DGPIMG_HEIGHT(dist);
	ymax = height - 1;
	
	/* Down right pass. */
	
	/* First line. */
	current = DGPIMG_REF(dist, 0, 0);
	for (x = 1; x < width; x++) {
		left = current + 1;
		current = DGPIMG_REF(dist, x, 0);
		if (left < current) {
			current = DGPIMG_REF(dist, x, 0) = left;
			DGPIMG_REF(image, x, 0) = DGPIMG_REF(image, x - 1, 0);
		}
	}
	/* Internal lines. */
	for (y = 1; y < height; y++) {
		/* Leftmost column. */
		up = DGPIMG_REF(dist, 0, y - 1) + 1;
		current = DGPIMG_REF(dist, 0, y);
		if (up < current) {
			current = DGPIMG_REF(dist, 0, y) = up;
			DGPIMG_REF(image, 0, y) = DGPIMG_REF(image, 0, y - 1);
		}
		/* Internal columns. */
		for (x = 1; x < width; x++) {
			left = current + 1;
			up = DGPIMG_REF(dist, x, y - 1) + 1;
			current = DGPIMG_REF(dist, x, y);
			if (left < current) {
				current = DGPIMG_REF(dist, x, y) = left;
				DGPIMG_REF(image, x, y) = DGPIMG_REF(image, x-1, y);
			}
			if (up < current) {
				current = DGPIMG_REF(dist, x, y) = up;
				DGPIMG_REF(image, x, y) = DGPIMG_REF(image, x, y-1);
			}
		}
	}

	/* Upper left pass. */
	
	/* First line. */
	current = DGPIMG_REF(dist, xmax, ymax);
	for (x = xmax - 1; x >= 0; x--) {
		right = current + 1;
		current = DGPIMG_REF(dist, x, ymax);
		if (right < current) {
			current = DGPIMG_REF(dist, x, ymax) = right;
			DGPIMG_REF(image, x, ymax) = DGPIMG_REF(image, x + 1, ymax);
		}
	}
	/* Internal lines. */
	for (y = ymax - 1; y >= 0; y--) {
		/* Rightmost column. */
		down = DGPIMG_REF(dist, xmax, y + 1) + 1;
		current = DGPIMG_REF(dist, xmax, y);
		if (down < current) {
			current = DGPIMG_REF(dist, xmax, y) = down;
			DGPIMG_REF(image, xmax, y) = DGPIMG_REF(image, xmax, y + 1);
		}
		/* Internal columns. */
		for (x = xmax - 1; x >= 0; x--) {
			right = current + 1;
			down = DGPIMG_REF(dist, x, y + 1) + 1;
			current = DGPIMG_REF(dist, x, y);
			if (right < current) {
				current = DGPIMG_REF(dist, x, y) = right;
				DGPIMG_REF(image, x, y) = DGPIMG_REF(image, x + 1, y);
			}
			if (down < current) {
				current = DGPIMG_REF(dist, x, y) = down;
				DGPIMG_REF(image, x, y) = DGPIMG_REF(image, x, y + 1);
			}
		}
	}
}
/*************************************************************************/

/*	As used in edges.c, the 'dist' matrix contains the value (maxDist + 1) in every location where the matrix */
/*	'image' is NOT_ASSIGNED.  I guess it's zero everywhere else. */
/* */

/*************************************************************************
**
**  Function: void propagate_l(DgpImage_Long image, DgpImage_Long dist)
**
**
**  As above, for Dgp_Int32 images.  
**
**	As called in edges.c, the 'dist' matrix contains the value (maxDist + 1) in every location where the matrix
**	'image' is NOT_ASSIGNED.  I guess it's zero everywhere else.  This means the cc's propagate into the background
**	rather than vice versa.  "Distance" appears to be defined from the edge of the cc, not the cog, and effectively
**	only along one of the two principal axes.  Or? / dhtesq.
**
**
**	Arguments:
**		o  image is the Dgp_Int32 image to be propagated.
**		o  dist is the propagation record.
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void
propagate_l(DgpImage_Long image, DgpImage_Long dist) {
	Dgp_Int32			width, height, xmax, ymax;
	Dgp_Int32			x, y;
	Dgp_Int32		current, up, left, down, right;

	width = DGPIMG_WIDTH(dist);
	xmax = width - 1;
	height = DGPIMG_HEIGHT(dist);
	ymax = height - 1;
	
	/* Down right pass. */
	
	/* First line. */
	current = DGPIMG_REF(dist, 0, 0);
	for (x = 1; x < width; x++) {
		left = current + 1;
		current = DGPIMG_REF(dist, x, 0);
		if (left < current) {
			current = DGPIMG_REF(dist, x, 0) = left;
			DGPIMG_REF(image, x, 0) = DGPIMG_REF(image, x-1, 0);
		}
	}
	/* Internal lines. */
	for (y = 1; y < height; y++) {
		/* Leftmost column. */
		up = DGPIMG_REF(dist, 0, y-1) + 1;
		current = DGPIMG_REF(dist, 0, y);
		if (up < current) {
			current = DGPIMG_REF(dist, 0, y) = up;
			DGPIMG_REF(image, 0, y) = DGPIMG_REF(image, 0, y-1);
		}
		/* Internal columns. */
		for (x = 1; x < width; x++) {
			left = current + 1;
			up = DGPIMG_REF(dist, x, y-1) + 1;
			current = DGPIMG_REF(dist, x, y);
			if (left < current) {
				current = DGPIMG_REF(dist, x, y) = left;
				DGPIMG_REF(image, x, y) = DGPIMG_REF(image, x-1, y);
			}
			if (up < current) {
				current = DGPIMG_REF(dist, x, y) = up;
				DGPIMG_REF(image, x, y) = DGPIMG_REF(image, x, y-1);
			}
		}
	}

	/* Upper left pass. */
	
	/* First line. */
	current = DGPIMG_REF(dist, xmax, ymax);
	for (x = xmax-1; x >= 0; x--) {
		right = current+1;
		current = DGPIMG_REF(dist, x, ymax);
		if (right < current) {
			current = DGPIMG_REF(dist, x, ymax) = right;
			DGPIMG_REF(image, x, ymax) = DGPIMG_REF(image, x+1, ymax);
		}
	}
	/* Internal lines. */
	for (y = ymax-1; y >= 0; y--) {
		/* Rightmost column. */
		down = DGPIMG_REF(dist, xmax, y+1)+1;
		current = DGPIMG_REF(dist, xmax, y);
		if (down < current) {
			current = DGPIMG_REF(dist, xmax, y) = down;
			DGPIMG_REF(image, xmax, y) = DGPIMG_REF(image, xmax, y+1);
		}
		/* Internal columns. */
		for (x = xmax-1; x >= 0; x--) {
			right = current+1;
			down = DGPIMG_REF(dist, x, y+1)+1;
			current = DGPIMG_REF(dist, x, y);
			if (right < current) {
				current = DGPIMG_REF(dist, x, y) = right;
				DGPIMG_REF(image, x, y) = DGPIMG_REF(image, x+1, y);
			}
			if (down < current) {
				current = DGPIMG_REF(dist, x, y) = down;
				DGPIMG_REF(image, x, y) = DGPIMG_REF(image, x, y+1);
			}
		}
	}
}
/*************************************************************************/

/*************************************************************************
**
**  Function: ErrCode propagate_seed_l(Dgp_State *dgp_state, DgpImage_Long image, DgpImage_Binary seeds)
**
**
**  Propagates the passed image, using the passed seed matrix to create the distance matrix.
**
**	I think this is a debugging function. / dhtesq.
**
**
**	Arguments:
**		o  image is the Dgp_Int32 image to be propagated.
**		o  seeds will give the dist matrix the value LONG_MAX - 1 at every location where it is non-zero.
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling.
**  
**/

ErrCode
propagate_seed_l(Dgp_State *dgp_state, DgpImage_Long image, DgpImage_Binary seeds) {
	DGP_PROLOGUE("propagate_seed_l()", DGP_PERRSTATE);
	DgpImage_Long		dist;
	Dgp_Int32				x, y;
	Dgp_Int32				width = DGPIMG_WIDTH(seeds);
	Dgp_Int32				height = DGPIMG_HEIGHT(seeds);

	DGP_ERR_PASS(DgpImage_new(dgp_state, DGPIMAGE_LONG, width, height, (void **)&dist));

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			if (!DGPIMG_REF(seeds, x, y))
				DGPIMG_REF(dist, x, y) = LONG_MAX - 1;
		}
	}
	propagate_l(image, dist);
	DGP_ERR_PASS(DgpImage_free(dist));
	NORMAL_RETURN;
}
/*************************************************************************/
