/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 26.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename: scancolor.c
**
**  Main pipeline for the color segmentation machinery.  Principal entry points
**	to this library occur in this file.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling.
**  
**************************************************************************/

#include <stdlib.h>
#include <limits.h>
/*#include <tok-lossy.h> */
#include <libutilw.pub>
#include <libpixrw.pub>
#include <libiow.pub>
#include <libscalw.pub>
#include <math.h>

#include "dgp_state.h"				/* For liberr exception handling. */
#include "image.h"
#include "image_defs.h"
#include "ccs.h"
#include "seg.h"
#include "edges.h"
#include "chains.h"
#include "scancolor.h"
#include "complexity.h"
#include "propagate.h"
#include "findCC.h"
#include "imutil.h"
#include "PsgFix.h"
/*#include "simple_utils.h"			Used? */




DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libscancolor/scancolor.c,v 1.15 2003/02/10 22:53:37 mkrocken Exp $");

/* We can be in one of many stages... */
#define STAGE_INIT 0
#define STAGE_FINDTEXT 1

/* This is the resolution we will use if the image has no resolution information. */
/*	400 preferable for color images? / dhtesq. */
#define DEFAULT_DPI 300


/*************************************************************************
**
**  Function:  DGP_LIBEXPORT ErrCode DgpSC_init(Dgp_State *dgp_state, Dgpsc_cookie **_cookie)
**
**  Cookie initialization for scan color pipeline.  This function must be called
**	before invoking any other function in the pipeline.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State.
**		o  _cookie points to the location where the cookie pointer will be stored.
**
**	Return value:
**		o  ErrCode;  will fail on bad allocation of cookie storage.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling.
**		o  Context modified to take Dgp_State* rather than IPWState* (to use liberr).
**			Use cookie->dgp_state->ipwstate to recover IPWState*.		
**  
**/

DGP_LIBEXPORT ErrCode
DgpSC_init(Dgp_State *dgp_state, Dgpsc_cookie **_cookie) {
	/* Set the error state for this function to the passed error state. */
	DGP_PROLOGUE("DgpSC_init()", DGP_PERRSTATE);

	Dgpsc_cookie *cookie;

	/*	This will throw cErrNomem if the cookie is NULL... */
	DGP_CHECK_ALLOC(cookie = (Dgpsc_cookie *)DGP_MALLOC(sizeof(Dgpsc_cookie)));
	
	/* Set initial stage. */
	cookie->stage = STAGE_INIT;	

	/* Track dgp_state... */
	cookie->dgp_state = dgp_state;

    /* Null pointers in case of intermediate error: */
    cookie->components = NULL;
    cookie->colorIm = NULL;
    cookie->compMap = NULL;
	*_cookie = cookie;

	NORMAL_RETURN;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  DGP_LIBEXPORT ErrCode DgpSC_defaultOptions(Dgpsc_cookie *cookie, Dgp_Int32 xres, Dgp_Int32 yres)
**
**  Sets default parameters for all heuristics.  Must be called while in the initialization
**	stage (cookie->stage == STAGE_INIT).  Some parameters will depend on MIN(xres, yres).
**
**	Arguments:
**		o  cookie is an Dgpsc_cookie* previously initialized via a call to DgpSC_init().
**		o  xres is the horizontal resolution of the image in DPI (?).
**		o  yres is the vertical resolution of the image in DPI (?).
**
**	Return value:
**		o  None.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling;  return type changed from Dgp_Int32 to ErrCode.
**  
**/

ErrCode
DgpSC_defaultOptions(Dgpsc_cookie *cookie, Dgp_Int32 xres, Dgp_Int32 yres) {
	Dgp_State		*dgp_state = cookie->dgp_state;
	DGP_PROLOGUE("DgpSC_defaultOptions", DGP_PERRSTATE);

	Dgp_Int32 res = MIN(xres, yres);
	DGP_ERR_GEN((cookie->stage != STAGE_INIT), cSCNotInStageInit);

	defaultEdgeOptions(&(cookie->edgeOpt), xres, yres);
	defaultChainOptions(&(cookie->chainOpt), xres, yres);
	defaultSegmentOptions(&(cookie->segOpt), xres, yres);
	cookie->complexityThresh = 10;			/* Threshold for component complexity. */
	cookie->extraScale = 60.0F/res;			/* Scale on which to look for i dots, commas, periods, etc. */
	cookie->colorEpsilon = 100;				/* Delta inside of which colors are to be considered similar. */
	cookie->extraSizeEpsilon = 0.5F;		/* Maximum size of i dots, commas, etc. relative to text size. */
	cookie->extraDistEpsilon = 1.0F;		/* Maximum distance of i dots, etc relative to text size. */

	NORMAL_RETURN;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  ErrCode sc_findTextPass1(Dgpsc_cookie *cookie)
**
**  Looks for text candidates:  assembles connected components, assesses their complexity, finds their edges,
**	and strings them into chains.  In the pipeline, this function is called from sc_findText().
**
**	Arguments:
**		o  cookie is an Dgpsc_cookie* previously initialized via a call to DgpSC_init().
**
**	Return value:
**		o  An ErrCode passed up to the invoking function.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling;  return type changed from Dgp_Int32 to ErrCode.
**  
**/

ErrCode
sc_findTextPass1(Dgpsc_cookie *cookie) {
	Dgp_State		*dgp_state = cookie->dgp_state;
	DGP_PROLOGUE("sc_findTextPass1", DGP_PERRSTATE);

	ComponentInfo	*components;
	Dgp_Int32		complex;
	DgpAdt_List		chains;

	/* Get component info structs. */
	DGP_ERR_PASS(getComponents(dgp_state, cookie->colorIm, cookie->compMap, cookie->numccs, &components));
	
	/* Calculate complexity measure. */
	DGP_ERR_PASS(getComplexity(dgp_state, cookie->compMap, components, cookie->numccs));

	/* Mark components that are too complex by setting their size to zero. */
	complex = remComplex(components, cookie->numccs, cookie->complexityThresh);

#ifdef DGP_DBUG
	DGP_REPORTF2(1, "I removed %d ccs exceeding complexity threshold %d.", complex, cookie->complexityThresh);
#endif

	/* Get edges between similar & close components. */
	DGP_ERR_PASS(getEdges(dgp_state, cookie->compMap, components, cookie->numccs, &(cookie->edgeOpt)));

#ifdef DGP_DBUG
	DGP_REPORTF2(1, "I found %d/%d ccs connected by edges.", countMarked(components, cookie->numccs), cookie->numccs);
#endif

	/* Get chains. */
	DGP_ERR_PASS(getChains(dgp_state, components, cookie->numccs, &chains, &(cookie->chainOpt)));

#ifdef DGP_DBUG
	DGP_REPORTF1(1, "I found %d chains.", DGPADT_LILEN(chains));
#endif

/*********************************/
#if 0
	do {
		DgpImage_ColorRGB tmp;
		DGP_ERR_PASS(DgpImage_new(dgp_state, 
                                  DGPIMAGE_RGB, 
                                  DGPIMG_WIDTH(cookie->colorIm), 
                                  DGPIMG_HEIGHT(cookie->colorIm), 
                                  &tmp));

		/* Save chains... to debug. */
		Dgp_Int32 i = 0;
		do {
			ListNode ln;
			chain *c;
			Dgp_Int8 tmp_str[256];

			DGPADT_FOREACH (ln, chains) {
				c = DGPADT_LIGET(ln);
				imClearRGB(tmp);
				printf("saving chain %d\n", i);
				blatChain(tmp, cookie->compMap, c, components);
				sprintf(tmp_str,"chain%d.ppm",i);
				imSave(tmp, tmp_str);
				i++;
			}
		} while (0);
		DGP_ERR_PASS(DgpImage_free(tmp));
	} while (0);
#endif
/*********************************/

	/* Dont need edges anymore;  free. */
	freeEdges(dgp_state, components, cookie->numccs);	

	/* Store some of our findings in the cookie. */
	cookie->components = components;
	cookie->chains = chains;

	/* Return with no errors. */
	NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode sc_getExtraText(Dgpsc_cookie *cookie)
**
**  Looks for ancillary components that nominally associated with text -- like 'i' dots,
**	commas, radicals, etc. -- that might not have been placed in chains because of distance
**	from the baseline (I think).  The function looks for non-text cc's which fall within 
**	some specified range in color, distance, and size from their nearest neighbor text cc.
**
**	Arguments:
**		o  cookie is an Dgpsc_cookie* previously initialized via a call to DgpSC_init().
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling;  return type changed from Dgp_Int32 to ErrCode.
**  
**/

ErrCode
sc_getExtraText(Dgpsc_cookie *cookie) {
	Dgp_State			*dgp_state = cookie->dgp_state;
	DGP_PROLOGUE("sc_getExtraText", DGP_PERRSTATE);

	DgpImage_Long			compMap = cookie->compMap;
	DgpImage_Long			dist;
	DgpImage_Long			out;
	Dgp_Int32				width = DGPIMG_WIDTH(compMap) - 2;		/* compMap has border of 1 pixel. */
	Dgp_Int32				height = DGPIMG_HEIGHT(compMap) - 2;
	Dgp_Float32				scale = cookie->extraScale;
	Dgp_Int32				numccs = cookie->numccs;
	ComponentInfo   		*components = cookie->components;
	Dgp_Int32				i, j;
	Dgp_Int32				colorEpsilon = cookie->colorEpsilon;
	Dgp_Float32				sizeEpsilon = cookie->extraSizeEpsilon;
	Dgp_Float32				distEpsilon = cookie->extraDistEpsilon;
	Dgp_Float32				maxDist;
	Dgp_Int32				recoveredCcs = 0;
  Dgp_Bool        haveText = DGP_FALSE;

	/* Get an image with text seeds & propagate it to find the closest text character for every component. */
	/* Do this at a smaller resolution for speed. */

	width = (Dgp_Int32)((width - 1) * scale) + 1;
	height = (Dgp_Int32)((height - 1) * scale) + 1;

	DGP_ERR_PASS(DgpImage_new(dgp_state, DGPIMAGE_LONG | DGPIMAGE_DONTCLEAR, width, height, (void **)&out));
	imSetL(out, NOT_ASSIGNED);	
	
	DGP_ERR_PASS(DgpImage_new(dgp_state, DGPIMAGE_LONG | DGPIMAGE_DONTCLEAR, width, height, (void **)&dist));
	imSetL(dist, LONG_MAX - 1);

	for (i = 0; i < numccs; i++) {
		/* Is this a text component? */
		if (components[i].text == TEXT_1) {			
			/* This may overwrite some other component.. but it shouldn't happen very often (ugh). */
			/* */
			/*	-- due to (Dgp_Int32) type cast?  Does this need fixing? / dhtesq. */
			DGPIMG_REF(out, (Dgp_Int32)(components[i].centerX*scale), (Dgp_Int32)(components[i].centerY*scale)) = i;
			DGPIMG_REF(dist, (Dgp_Int32)(components[i].centerX*scale), (Dgp_Int32)(components[i].centerY*scale)) = 0;
      haveText = DGP_TRUE;
		}
	}
	

  if (haveText)
  {
  /* Propagate. */
	/*	In this case the 'dist' matrix is non-zero where cc's do _not_ exist, so the components will be propagated into */
	/*	the background. / dhtesq. */
	propagate_l(out, dist);
	DGP_ERR_PASS(DgpImage_free(dist));

	/* Check each component to see if it could be a character modifier such as an i dot, comma, or whatever. */
	for (i = 0; i < numccs; i++) {
		/* Is it nominally not text? */
		/*	It was never recorded as TEXT_1 because it fell too far out of the baseline.  Or? / dhtesq. */
		if (components[i].text != TEXT_1) {
			/* What is its closest text character? */
			/*	i.e., whose Voronoi region does it lie in? / dhtesq. */
			j = DGPIMG_REF(out, (Dgp_Int32)(components[i].centerX*scale), (Dgp_Int32)(components[i].centerY*scale));
			/* Check distance, color distance, & size range (remember size 0 indicates a bad component). */
			maxDist = components[j].size*distEpsilon;
			if ((components[i].size > 0) &&
				(ccDistance(&components[i], &components[j]) <= maxDist) &&
				(ABS(components[i].color.r - components[j].color.r) <= colorEpsilon) &&
				(ABS(components[i].color.g - components[j].color.g) <= colorEpsilon) &&
				(ABS(components[i].color.b - components[j].color.b) <= colorEpsilon) &&
				(((Dgp_Float32)components[i].size/(Dgp_Float32)components[j].size) <= sizeEpsilon)) {
				/* Mark as text! */
				components[i].text = TEXT_1;
				/* Get the text color from the character, so they match. */
				components[i].color = components[j].color;
				recoveredCcs++;
			}
		}
	}
  }
  else
  {
	DGP_ERR_PASS(DgpImage_free(dist));
  }
  
  DGP_ERR_PASS(DgpImage_free(out));

#ifdef DGP_DBUG
	DGP_REPORTF1(1, "I recovered %d ccs as possible modifiers.", recoveredCcs);
#endif

	NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode sc_filterText(Dgpsc_cookie *cookie)
**
**  Connected components may have the same color as the background.  Closed background curves
**	like the insides of o's, a's, and e's will tend to align like text;  here we
**	get rid of them.
**
**  Pedro grows the background color into the components at all cc boundaries.  He then
**	checks the component's new color against its old one.  If they match, the text candidate
**	is likely to be background and he eliminates it.
**
**	Arguments:
**		o  cookie is an Dgpsc_cookie* previously initialized via a call to DgpSC_init().
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling;  return type changed from Dgp_Int32 to ErrCode.
**		o  15.5.2001 / double->single precision.
**  
**/

ErrCode
sc_filterText(Dgpsc_cookie *cookie) {
	Dgp_State			*dgp_state = cookie->dgp_state;
	DGP_PROLOGUE("sc_filterText", DGP_PERRSTATE);
	
	DgpImage_Long			    dist = NULL;
	DgpImage_ColorRGB			colors = NULL;
	Dgp_Int32					width = DGPIMG_WIDTH(cookie->colorIm);
	Dgp_Int32					height = DGPIMG_HEIGHT(cookie->colorIm);
	DgpImage_Long			    compMap = cookie->compMap;
	ComponentInfo		        *components = cookie->components;
	Dgp_Int32					x, y, xmax, ymax;
	Dgp_Int32					i;
	Dgp_Float32				    *r, *g, *b;
	Dgp_Int32					numccs = cookie->numccs;
	Dgp_Int32					colorEpsilon = cookie->colorEpsilon;
	Dgp_Int32					tmp_r, tmp_g, tmp_b;
	Dgp_Int32					rejectedComponents = 0;

	/* Null all pointers that are going to get dynamic allocations: */
	r = g = b = NULL;

	/* Copy colors. */
	DGP_ERR_PASS(DgpImage_new(dgp_state, DGPIMAGE_RGB, width, height, (void **)&colors));
	memcpy(DGPIMG_PIXPTR(colors, 0, 0), DGPIMG_PIXPTR(cookie->colorIm, 0, 0), width*height*sizeof(DgpImage_RGB));

	/* Make distance image to grow background on top of text. */
	DGP_ERR_PASS_CLEANUP(DgpImage_new(dgp_state, DGPIMAGE_LONG, width, height, (void **)&dist));

	ymax = height - 1;
	xmax = width - 1;
	for (y = 1; y < ymax; y++) {
		for (x = 1; x < xmax; x++) {
			i = DGPIMG_REF(compMap, x, y);
			if ((i != NOT_ASSIGNED) && (components[i].text == TEXT_1)) {
				DGPIMG_REF(dist, x, y) = LONG_MAX-1;
				DGPIMG_REF(dist, x+1, y) = LONG_MAX-1;
				DGPIMG_REF(dist, x-1, y) = LONG_MAX-1;
				DGPIMG_REF(dist, x, y+1) = LONG_MAX-1;
				DGPIMG_REF(dist, x, y-1) = LONG_MAX-1;
				DGPIMG_REF(dist, x+1, y+1) = LONG_MAX-1;
				DGPIMG_REF(dist, x-1, y+1) = LONG_MAX-1;
				DGPIMG_REF(dist, x+1, y-1) = LONG_MAX-1;
				DGPIMG_REF(dist, x-1, y-1) = LONG_MAX-1;
			}
		}
	}
	/* Propagate background. */
	/*	Note that this is inverted usage wrt edges.c, where the 'dist' matrix was valued at image points that carried the */
	/*	value NOT_ASSIGNED.  In this case we want the background to propagate into the image, not the image into the background. */
	/*	/ dhtesq. */
	propagate_rgb(colors, dist);
	DGP_ERR_PASS_CLEANUP(DgpImage_free(dist));

	/* Now I want to get the color of each token in this new image.  If it hasn't changed much from the */
	/* color on the original color image it must be that the token looks like the background. */
	DGP_CHECK_ALLOC_CLEANUP(r = (Dgp_Float32 *)DGP_CALLOC(1, numccs * sizeof(Dgp_Float32)));
	DGP_CHECK_ALLOC_CLEANUP(g = (Dgp_Float32 *)DGP_CALLOC(1, numccs * sizeof(Dgp_Float32)));
	DGP_CHECK_ALLOC_CLEANUP(b = (Dgp_Float32 *)DGP_CALLOC(1, numccs * sizeof(Dgp_Float32)));

	/*	calloc()? / dhtesq. */
	/*
	memset(r, 0, numccs*sizeof(Dgp_Float64));
	memset(g, 0, numccs*sizeof(Dgp_Float64));
	memset(b, 0, numccs*sizeof(Dgp_Float64));
	*/
	
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			i = DGPIMG_REF(compMap, x, y);
			if ((i != NOT_ASSIGNED) && (components[i].text == TEXT_1)) {
				r[i] += (Dgp_Float32)DGPIMG_REF(colors, x, y).r;
				g[i] += (Dgp_Float32)DGPIMG_REF(colors, x, y).g;
				b[i] += (Dgp_Float32)DGPIMG_REF(colors, x, y).b;	
			}
		}
	}
	DGP_ERR_PASS_CLEANUP(DgpImage_free(colors));

	for (i = 0; i < numccs; i++) {
		if (components[i].text == TEXT_1) {
			tmp_r = (Dgp_Int32)r[i]/components[i].num_pts;
			tmp_g = (Dgp_Int32)g[i]/components[i].num_pts;
			tmp_b = (Dgp_Int32)b[i]/components[i].num_pts;
			if ((ABS(tmp_r - components[i].color.r) <= colorEpsilon) &&
				(ABS(tmp_g - components[i].color.g) <= colorEpsilon) &&
				(ABS(tmp_b - components[i].color.b) <= colorEpsilon)) {
				/* This can't be text -- it's the same color as the background! */
				components[i].text = 0;
				rejectedComponents++;
			}	
		}
	}

#ifdef DGP_DBUG
	DGP_REPORTF1(1, "I rejected %d ccs that looked like background.", rejectedComponents);
#endif

	DGP_FREE(r);
	DGP_FREE(g);
	DGP_FREE(b);

	NORMAL_RETURN;

ERR_CLEANUP:
	if(r != NULL) DGP_FREE(r);
	if(g != NULL) DGP_FREE(g);
	if(b != NULL) DGP_FREE(b);
	if(colors != NULL) DGP_ERR_PASS(DgpImage_free(colors));
	if(dist != NULL) DGP_ERR_PASS(DgpImage_free(dist));
	ERR_RETURN;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  ErrCode sc_findText(Dgpsc_cookie         *cookie, 
**                                 DgpImage_ColorRGB    colorIm, 
**                                 Dgp_Int32            xres, 
**                                 Dgp_Int32            yres, 
**                                 DgpImage_Long        compMap, 
**                                 Dgp_Int32            numccs)
**
**  A "foreman function".  Orchestrates the location and filtering of text chains;  calls sc_filterText()
**	to collect i-dots, punctuation, etc.  sc_findText() must be called during STAGE_INIT, or it will throw
**	an error.
**
**	Arguments:
**		o  cookie is an Dgpsc_cookie* previously initialized via a call to DgpSC_init().
**		o  colorIm is the DgpImage_ColorRGB to be examined.
**		o  xres, yres of the image in dpi (?).
**		o  compMap is a valid component map obtained from previous segmentation.
**		o  numccs is the number of connected components in the compMap.
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling;  return type changed from Dgp_Int32 to ErrCode.
**  
**/

ErrCode
sc_findText(Dgpsc_cookie        *cookie, 
            DgpImage_ColorRGB   colorIm, 
            Dgp_Int32           xres, 
            Dgp_Int32           yres, 
            DgpImage_Long       compMap, 
            Dgp_Int32           numccs) {
	Dgp_State			*dgp_state = cookie->dgp_state;
	DGP_PROLOGUE("sc_findText", DGP_PERRSTATE);

	/* We have to be in the initialization stage at this point. */
	DGP_ERR_GEN((cookie->stage != STAGE_INIT), cSCNotInStageInit);
	
	cookie->colorIm = colorIm;
	cookie->compMap = compMap;
	cookie->numccs = numccs;
	cookie->xres = xres;
	cookie->yres = yres;

	/* Get text candidates. */
	/*	...as identified by their presence in chains...? / dhtesq. */
	DGP_ERR_PASS(sc_findTextPass1(cookie));

	/* We don't need the chains anymore. */
	/*	...now that the text has been nominally identified. / dhtesq. */
	freeChains(dgp_state, cookie->chains);
	cookie->chains = NULL;

	/* Filter out tokens that are the same color as the background such as the inside of o's, a's, p's, etc. */
	/*	...the components are not deleted;  their 'text' member is zeroed. / dhtesq. */
	/* */
	/* 15.5.2001 / dhtesq. */
	/* Try skipping this step -- it's quite expensive, and is useful only if you think you have cc's around */
	/* that are the same color as the background.  But binarization, by definition, should have taken care of that. */
	/* Tests only show 0-2% rejection anyway. */
	/* ...I'm putting it back in to test my 2-pass cc finder:  once on binarized image, once on inverted binarization. */
	/* This will definitely create "innies." */
	DGP_ERR_PASS(sc_filterText(cookie));

	/* Get i dots, commas, & etc. */
	DGP_ERR_PASS(sc_getExtraText(cookie));

	cookie->stage = STAGE_FINDTEXT;

	NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  DGP_LIBEXPORT ErrCode DgpSC_getTextMask(Dgpsc_cookie *cookie, PIXR **_pixr)
**
**  Places in *_pixr a mask whose bits are set if they lie on assigned & valid TEXT_1 components.
**
**	If not called during STAGE_FINDTEXT, an error will be thrown.
**
**	Arguments:
**		o  cookie is an Dgpsc_cookie* previously initialized via a call to DgpSC_init().
**		o  _pixr is an internal image format I think will be outmoded soon (Pixellated Representation?).  Or?
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling;  return type changed from Dgp_Int32 to ErrCode.
**  
**/

DGP_LIBEXPORT ErrCode
DgpSC_getTextMask(Dgpsc_cookie *cookie, PIXR **_pixr) {
	Dgp_State			*dgp_state = cookie->dgp_state;
	DGP_PROLOGUE("DgpSC_getTextMask", DGP_PERRSTATE);

	DgpImage_Packed			text;
	Dgp_Int32				width = DGPIMG_WIDTH(cookie->colorIm);
	Dgp_Int32				height = DGPIMG_HEIGHT(cookie->colorIm);
	DgpImage_Long			compMap = cookie->compMap;
	ComponentInfo			*components = cookie->components;
	PIXR					*pixr = NULL;
	IPWState				*ipstate = dgp_state->ipwstate;
	UInt8					*ptr;
	Int32					longsperline;
	Dgp_Int32			    i, x, y;
	Dgp_UInt32			    *retval;		/* For DgpImage_unwrap... / dhtesq. */

	/* You better have called findText by now... */
	DGP_ERR_GEN((cookie->stage != STAGE_FINDTEXT), cSCNotInStageFindText);
	
	/* Create a pixr... */
	DGP_ERR_PASS(w_createPixr(	ipstate,
								width,
								height,
								1, 
								Int32_TO_ScaledInt1616(cookie->xres), 
								Int32_TO_ScaledInt1616(cookie->yres), 
								cFrameBits,
								&pixr));

	/* Wrap with a wjr image for ease of use. */
	w_pixrGetImage(ipstate, pixr, &ptr);
	w_pixrGetBpl(ipstate, pixr, &longsperline);
	longsperline /= sizeof(Dgp_Int32);
	DGP_ERR_PASS_CLEANUP(DgpImage_wrap(dgp_state, (Dgp_UInt32 *)ptr, width, height, longsperline, &text));

	/* Blat text components... */
	for (y = 1; y < height; y++) {
		for (x = 1; x < width; x++) {
			i = DGPIMG_REF(compMap, x, y);
			if ((i != NOT_ASSIGNED) && (components[i].text == TEXT_1)) {
				DGPIMG_SETBIT(text, x, y);		/* Again with the pimX macros.  What are the operations on the x coordinate? / dhtesq. */
			}
		}
	}

	/* Done with wjr wrap. */
	DGP_ERR_PASS_CLEANUP(DgpImage_unwrap(text, &retval));

	*_pixr = pixr;

	NORMAL_RETURN;

ERR_CLEANUP:
	if(pixr != NULL){
		DGP_ERR_PASS(w_destroyPixr(ipstate, &pixr));
	}

	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  DGP_LIBEXPORT ErrCode DgpSC_getTextComponents(Dgpsc_cookie *cookie, sc_cc **_ccs, Dgp_Int32 *_num_text)
**
**  Outputs an array of all text components and the length of this array.  Storage for the array
**	is allocated here, but not freed;  caller must free when done.  This is also true of the packed
**	images of each component.
**
**  N.B.  On an allocation error, this function will attempt to compose a partial array of components.
**
**	Arguments:
**		o  cookie is an Dgpsc_cookie* previously initialized via a call to DgpSC_init().
**		o  _ccs is the address of the pointer to the sc_cc{} which represents this component array.
**		o  _num_text points to the location where the length of this array (in members) will be written.
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling;  return type changed from Dgp_Int32 to ErrCode.  _num_text added
**		since ErrCode ate the return value.
**		
**  
**/

DGP_LIBEXPORT ErrCode
DgpSC_getTextComponents(Dgpsc_cookie *cookie, sc_cc **_ccs, Dgp_Int32 *_num_text) {
	Dgp_State			*dgp_state = cookie->dgp_state;
	DGP_PROLOGUE("DgpSC_getTextComponents", DGP_PERRSTATE);

	sc_cc				*ccs = NULL;
	Dgp_Int32			i, x, y;
	DgpImage_Long		compMap = cookie->compMap;
	ComponentInfo		*components = cookie->components;
	Dgp_Int32			width = DGPIMG_WIDTH(compMap) - 2;		/* compMap has border of 1 pixel. */
	Dgp_Int32			height = DGPIMG_HEIGHT(compMap) - 2;
	Dgp_Int32			numccs = cookie->numccs;
	Dgp_Int32			num_text;
	sc_cc				*pcc;
	
	/* Count the number of text components. */
	/*	...no check for NOT_ASSIGNED;  presumably taken care of in calculation of numccs? / dhtesq. */
	for (num_text = 0, i = 0; i < numccs; i++) {
		if (components[i].text == TEXT_1) {
			num_text++;

			/*	Null the userdata field in case allocation fails. */
			components[i].userdata = NULL;
		}
	}

	/* Get one cc struct for each text cc. */
	DGP_CHECK_ALLOC(ccs = DGP_MALLOC(sizeof(sc_cc)*num_text));

	/* Set up and fill the component's struct. */
	for (num_text = 0, i = 0; i < numccs; i++) {
		if (components[i].text == TEXT_1) {
			ccs[num_text].x = components[i].minX;
			ccs[num_text].y = components[i].minY;
			ccs[num_text].color = components[i].color;

			DGP_ERR_PASS_CLEANUP(DgpImage_new(	dgp_state,
												DGPIMAGE_PACKED,
												components[i].width,
												components[i].height,
												(void **)&ccs[num_text].image));

			/* We set the userdata member in each component struct to point to a new cc box struct */
			/*	containing the text. */

			components[i].userdata = &(ccs[num_text]);
			num_text++;
		}
	}

	/* Now we need to fill in the images. */
	/*	Every assigned TEXT_1 component in the image has allocated storage at this point. */
	/*	Also, note how neighboring cc's do _not_ appear in the image, even if they fall nominally */
	/*	within the (width x height) box around the component. */
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			i = DGPIMG_REF(compMap, x, y);
			if ((i != NOT_ASSIGNED) && (components[i].text == TEXT_1)) {
				pcc = (sc_cc *)components[i].userdata;
				DGPIMG_SETBIT(pcc->image, x - pcc->x, y - pcc->y);
			}
		}
	}
	*_ccs = ccs;
	*_num_text = num_text;
	NORMAL_RETURN;

ERR_CLEANUP:
	/*	If one component allocation fails, we will try to fill in the ones we did get. */
	/*	ErrCode will still be returned, so caller can choose to ignore these data. / dhtesq. */
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			i = DGPIMG_REF(compMap, x, y);
			if ((i != NOT_ASSIGNED) && (components[i].text == TEXT_1) && (components[i].userdata != NULL)) {
				pcc = (sc_cc *)components[i].userdata;
				DGPIMG_SETBIT(pcc->image, x - pcc->x, y - pcc->y);
			}
		}
	}
	*_ccs = ccs;
	*_num_text = num_text;
	
	/*	We're returning a partial array, so nothing to free here...but note ccs is still allocated */
	/*	for the full number of text components. */

	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode DgpSC_getBackground(Dgpsc_cookie *cookie, Dgp_Int32 bgRes, PIXR **_pixr)
**
**  Dilates the text mask, then obtains the background image by propagating the background over the text.
**  If necessary, we scale down the background image.  An error will be thrown if this function is called
**	when not in the STAGE_FINDTEXT pipeline sequence.
**
**	If the xres and yres in the cookie are different, this function will throw an error of type
**	sLSCCantScaleBackground.
**
**	Arguments:
**		o  cookie is an Dgpsc_cookie* previously initialized via a call to DgpSC_init().
**		o  bgRes is the background resolution in dpi (?).
**		o  _pixr is the address of a PIXR representation where the background image will be stored.
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling;  return type changed from Dgp_Int32 to ErrCode.
**		
**  
**/

DGP_LIBEXPORT ErrCode
DgpSC_getBackground(Dgpsc_cookie *cookie, Dgp_Int32 bgRes, PIXR **_pixr) {
  Dgp_State				*dgp_state = cookie->dgp_state;
  DGP_PROLOGUE("DgpSC_getBackground()", DGP_PERRSTATE);
  
  DgpImage_Long			dist;
  Dgp_Int32			width = DGPIMG_WIDTH(cookie->colorIm);
  Dgp_Int32			height = DGPIMG_HEIGHT(cookie->colorIm);
  DgpImage_Long   		compMap = cookie->compMap;
  ComponentInfo			*components = cookie->components;
  Dgp_Int32			x, y, xmax, ymax;
  PIXR		*pixr = NULL, *pixr_r = NULL, *pixr_g = NULL, *pixr_b = NULL;
  DgpImage_RGB			*src, *src_end;
  Dgp_UInt8				*dst_r, *dst_g, *dst_b;
  Dgp_Int32				i;
  PIXR					*scaled_pixr = NULL;
  IPWState				*ipstate = dgp_state->ipwstate;

  /* You better have called findText() by now... */
  DGP_ERR_GEN((cookie->stage != STAGE_FINDTEXT), cSCNotInStageFindText);
  
/*  DGP_ERR_GEN((cookie->xres != cookie->yres), cSCCantScaleBackground); */
  
  /* Make distance image to grow background on top of text. */
  DGP_ERR_PASS(DgpImage_new(dgp_state, DGPIMAGE_LONG, width, height, (void **)&dist));
  
  ymax = height - 1;
  xmax = width - 1;
  for (y = 1; y < ymax; y++) {
    for (x = 1; x < xmax; x++) {
      i = DGPIMG_REF(compMap, x, y);
      if ((i != NOT_ASSIGNED) && (components[i].text == TEXT_1)) {
		  /* We blat the text pixel plus its 8 neighbors to 'dilate' the text mask */
		  /* before growing on top of it. */
		  DGPIMG_REF(dist, x, y) = LONG_MAX-1;
		  DGPIMG_REF(dist, x+1, y) = LONG_MAX-1;
		  DGPIMG_REF(dist, x-1, y) = LONG_MAX-1;
		  DGPIMG_REF(dist, x, y+1) = LONG_MAX-1;
		  DGPIMG_REF(dist, x, y-1) = LONG_MAX-1;
		  DGPIMG_REF(dist, x+1, y+1) = LONG_MAX-1;
		  DGPIMG_REF(dist, x-1, y+1) = LONG_MAX-1;
		  DGPIMG_REF(dist, x+1, y-1) = LONG_MAX-1;
		  DGPIMG_REF(dist, x-1, y-1) = LONG_MAX-1;
      }
    }
  }
  /* Propagate background. */
  /*	'dist' matrix is non-zero where components are assigned text, so background will grow over text. */
  propagate_rgb(cookie->colorIm, dist);
  DGP_ERR_PASS(DgpImage_free(dist));
  
  /* Create a pixr. */
  DGP_ERR_PASS(w_createPixrExtended(	ipstate,
					width,
					height,
					8,
					Int32_TO_ScaledInt1616(cookie->xres),
					Int32_TO_ScaledInt1616(cookie->yres), 
					cFrameBits,
					3,
					cColorRGB,
					NULL,
					&pixr));
  
  
  DGP_ERR_PASS_CLEANUP(w_pixrGetChannel(ipstate, pixr, 0, &pixr_r));
  DGP_ERR_PASS_CLEANUP(w_pixrGetChannel(ipstate, pixr, 1, &pixr_g));
  DGP_ERR_PASS_CLEANUP(w_pixrGetChannel(ipstate, pixr, 2, &pixr_b));
  
  
  /* Copy into the new pixr. */
  for (y = 0; y < height; y++) {
    src = DGPIMG_PIXPTR(cookie->colorIm, 0, y);
    src_end = DGPIMG_PIXPTR(cookie->colorIm, width, y);
    w_getLine(ipstate, pixr_r, y, &dst_r);
    w_getLine(ipstate, pixr_g, y, &dst_g);
    w_getLine(ipstate, pixr_b, y, &dst_b);
    for (; src < src_end; src++, dst_r++, dst_g++, dst_b++) {
      *dst_r = src->r;
      *dst_g = src->g;
      *dst_b = src->b;
    }
  }
  /* Byte swap for little-endian platforms. */
  DGP_ERR_PASS_CLEANUP(w_flip(ipstate, pixr));
  
  /* Don't scale if not necessary. */
  if (bgRes == cookie->xres) {
    *_pixr = pixr;
    NORMAL_RETURN;
  }
  
  /* Scale using resConv -- it looks nice. */
  DGP_ERR_PASS_CLEANUP(w_resConv(	ipstate,
					pixr,
					cookie->xres,
					bgRes,
					&scaled_pixr));
  
  DGP_ERR_PASS(w_destroyPixr(ipstate, &pixr));
  
  *_pixr = scaled_pixr;
  NORMAL_RETURN;
  
 ERR_CLEANUP:
  if(pixr != NULL){
    DGP_ERR_PASS(w_destroyPixr(ipstate, &pixr));
  }
  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
 **
 **  Function:  DGP_LIBEXPORT ErrCode DgpSC_process(Dgpsc_cookie *cookie, PIXR *image)
 **
 **  Principal entry point to the library and scancolor in particular.  Manages the pipeline
 **  for processing of a single page in PIXR representation.  An error will be thrown if this function is called
 **	when not in the STAGE_FINDTEXT pipeline sequence.
 **
 **	Will throw error sLSCNotInStageInit if called out of sequence.
 **
 **	Arguments:
 **		o  cookie is an Dgpsc_cookie* previously initialized via a call to DgpSC_init().
**		o  image is the PIXR representation of the page to be processed (?).
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling;  return type changed from Dgp_Int32 to ErrCode.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**		
**  
**/

DGP_LIBEXPORT ErrCode
DgpSC_process(Dgpsc_cookie *cookie, PIXR *image) {
	Dgp_State				*dgp_state = cookie->dgp_state;
	DGP_PROLOGUE("DgpSC_process", DGP_PERRSTATE);
	
	DgpImage_Long			compMap = NULL;
	Dgp_Int32				numccs;
	ScaledInt1616		    tmp_xres, tmp_yres;
	Dgp_Int32				xres, yres;
	DgpImage_ColorRGB		colorIm;
	IPWState			    *ipstate = dgp_state->ipwstate;

	/* You better not have called findText() yet... */
	DGP_ERR_GEN((cookie->stage != STAGE_INIT), cSCNotInStageInit);

	DGP_ERR_PASS(segment(dgp_state, image, &(cookie->segOpt), &compMap, &numccs));

#ifdef DGP_DBUG
	DGP_REPORTF1(1, "DgpSC_getTextMask() / segmenter returned %d cc's.", numccs);
#endif

	DGP_ERR_PASS(w_pixrGetResolutionDPI(ipstate, image, &tmp_xres, &tmp_yres));
	xres = ScaledInt1616_TO_Int32(tmp_xres);
	yres = ScaledInt1616_TO_Int32(tmp_yres);	
	if (xres == 0){
		xres = DEFAULT_DPI;
	}
	if (yres == 0){
		yres = DEFAULT_DPI;
	}
#ifdef DGP_DBUG
	DGP_REPORTF2(1, "DgpSC_getTextMask() / xres:  %d\tyres:  %d", xres, yres);
#endif

	/* The rest of the code likes wjr images instead of pixrs: */
  if (image->channels == 3)
  {
  DGP_ERR_PASS(pixr_to_rgb(dgp_state, image, &colorIm)); 
  }
  else /* gray scale image, reset colorEpsilon */
  {
      cookie->colorEpsilon /= 3;
      cookie->chainOpt.colorEpsilon /= 3;
      cookie->edgeOpt.colorEpsilon /= 3;
      DGP_ERR_PASS(pixr_to_rgb_gray(dgp_state, image, &colorIm)); 
  }

	DGP_ERR_PASS(w_destroyPixr(ipstate, &image));

	DGP_ERR_PASS(sc_findText(cookie, colorIm, xres, yres, compMap, numccs));

	NORMAL_RETURN;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  ErrCode DgpSC_endPage(Dgpsc_cookie *cookie)
**
**  Frees the colorIm, compMap, and components members of the cookie and resets the stage to STAGE_INIT.
**
**	Arguments:
**		o  cookie is an Dgpsc_cookie* previously initialized via a call to DgpSC_init().
**
**	Return value:
**		o  An ErrCode.  An error will be thrown if DgpImage_free cannot free either the colorIm or the compMap.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling;  return type changed from void to ErrCode.
**		
**  
**/

DGP_LIBEXPORT ErrCode
DgpSC_endPage(Dgpsc_cookie *cookie) {
	Dgp_State *dgp_state = cookie->dgp_state;
	DGP_PROLOGUE("DgpSC_endPage", DGP_PERRSTATE);

	if (cookie->components != NULL) {
		DGP_FREE(cookie->components);
		cookie->components = NULL;
	}
	if (cookie->colorIm != NULL) {
		DGP_ERR_PASS_CLEANUP(DgpImage_free(cookie->colorIm));
		cookie->colorIm = NULL;
	}
	if (cookie->compMap != NULL) {
		DGP_ERR_PASS(DgpImage_free(cookie->compMap));
		cookie->compMap = NULL;
	}
	cookie->stage = STAGE_INIT;
	NORMAL_RETURN;

ERR_CLEANUP:
	if (cookie->compMap != NULL) {
		DGP_ERR_PASS(DgpImage_free(cookie->compMap));
		cookie->compMap = NULL;
	}
	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  DGP_LIBEXPORT ErrCode DgpSC_done(Dgpsc_cookie *cookie)
**
**  Calls sc_endPage() to free the cookie members, then frees the cookie.
**
**	Arguments:
**		o  cookie is an Dgpsc_cookie* previously initialized via a call to DgpSC_init().
**
**	Return value:
**		o  An ErrCode.  Passes errors thrown by sc_endPage.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling;  return type changed from void to ErrCode.
**		
**  
**/

DGP_LIBEXPORT ErrCode
DgpSC_done(Dgpsc_cookie *cookie) {
	Dgp_State *dgp_state = cookie->dgp_state;
	DGP_PROLOGUE("DgpSC_done", DGP_PERRSTATE);

	DGP_ERR_PASS(DgpSC_endPage(cookie));
	DGP_FREE(cookie);
	NORMAL_RETURN;
}

/*************************************************************************/
/** Get the gray chanel
 **/
DGP_LIBEXPORT ErrCode
DgpSC_getGray(Dgpsc_cookie *cookie, DgpImage_Gray *grayImage) 
{
  Dgp_State				*dgp_state = cookie->dgp_state;
  DGP_PROLOGUE("DgpSC_getGray()", DGP_PERRSTATE);
  Dgp_Int32			width = DGPIMG_WIDTH(cookie->colorIm);
  Dgp_Int32			height = DGPIMG_HEIGHT(cookie->colorIm);
  Dgp_Int32				i,j;

  /* You better have called findText() by now... */
  DGP_ERR_GEN(grayImage == NULL, cErrInvalidParm);
  
  /* Make distance image to grow background on top of text. */
  DGP_ERR_PASS_CLEANUP(DgpImage_new(dgp_state, DGPIMAGE_GRAY, width, height, (void **)grayImage));
  
  for(j=0; j < height; j++){
      for(i=0; i < width; i++){
            DGPIMG_REF(*grayImage, i,j) = DGPIMG_REF(cookie->colorIm,i,j).g;
      }
  }
	NORMAL_RETURN;

ERR_CLEANUP:
	if (*grayImage != NULL) {
		DGPIMAGE_FREE(*grayImage);
		*grayImage = NULL;
	}
	ERR_RETURN;

}
