/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 31.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename: scancolor.h
**
**  Types & prototypes for scancolor.c.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  To accommodate liberr, Dgpsc_cookie{} now holds a Dgp_State* rather than an IPWState*.  If you need it,
**		   the IPWState* is obtainable via Dgp_State *dgp_state->ipwstate.
**  
**************************************************************************/

#ifndef SCAN_COLOR_H
#define SCAN_COLOR_H

#include "dgp_err_control.h"
#include "image.h"
#include "ccs.h"
#include "seg.h"
#include "edges.h"
#include "chains.h"
#include "image.h"

/*	Struct for scancolor connected components management, e.g. */
/*	constructing arrays of text components. */
typedef struct {
    DgpImage_Packed		image;		
    Dgp_Int32   		x, y;
    DgpImage_RGB       	color;
} sc_cc;

/*	Context for most scancolor functions;  exported, but contents   */
/*	not modified at the API level unless the user wants other than  */
/*	default options. */
typedef struct {
	Dgp_Int32			stage;		/*	Timing context to avoid out-of-sequence calls. */
	DgpImage_ColorRGB	colorIm;		/*	Image under analysis. */
	DgpImage_Long	    compMap;			/*	Component map for colorIm. */
	ComponentInfo		*components;		/*	Array of connected component context. */
	Dgp_Int32			numccs;		/*	Length(components). */
	DgpAdt_List 		chains;			/*	ll of neighboring cc's baseline-aligned;  hopefully text. */
	segOptions	        segOpt;			/*	| */
	edgeOptions			edgeOpt;	/*	|	Options structs for segmentation, edge-finding, & chain-finding. */
	chainOptions        chainOpt;			/*	| */
	Dgp_Int32			complexityThresh;	/*	cc's with complexity greater than this discarded as text. */
	Dgp_Int32			xres, yres;		/*	In dpi (?). */
	Dgp_Int32			text_tag;		/*	Used? */
	Dgp_Float32			extraScale;		/*	Absolute distance on which to look for "extra" text -- dots, commas, etc. */
	Dgp_Int32			colorEpsilon;		/*	Allowed color variance in a single chain;  24-bit color assumed. */
	Dgp_Float32			extraSizeEpsilon;	/*	cc size over extra text size may not exceed this ratio. */
	Dgp_Float32			extraDistEpsilon;	/*	When multiplied by the cc size, this gives the maximum distance from */
								/*	a given cc that "extra" text is allowed to lie. */
	Dgp_State			*dgp_state;		/*	Changed from IPWState* ipstate to accommodate liberr. */
} Dgpsc_cookie;

/*	Prototypes.  Unclear as yet which can be static vs. external. */
/*	See scancolor.c for full descriptions. */
DGP_LIBEXPORT ErrCode
DgpSC_init(Dgp_State *dgp_state, Dgpsc_cookie **_cookie);

DGP_LIBEXPORT ErrCode
DgpSC_defaultOptions(Dgpsc_cookie *cookie, Dgp_Int32 xres, Dgp_Int32 yres);

DGP_LIBEXPORT ErrCode
DgpSC_process(Dgpsc_cookie *cookie, PIXR *image);

DGP_LIBEXPORT ErrCode
DgpSC_getTextMask(Dgpsc_cookie *cookie, PIXR **_pixr);

DGP_LIBEXPORT ErrCode
DgpSC_getTextComponents(Dgpsc_cookie *cookie, sc_cc **_ccs, Dgp_Int32 *_num_text);

DGP_LIBEXPORT ErrCode
DgpSC_getBackground(Dgpsc_cookie *cookie, Dgp_Int32 bgRes, PIXR **_pixr);

DGP_LIBEXPORT ErrCode
DgpSC_done(Dgpsc_cookie *cookie);

DGP_LIBEXPORT ErrCode
DgpSC_endPage(Dgpsc_cookie *cookie);

DGP_LIBEXPORT ErrCode
DgpSC_getGray(Dgpsc_cookie *cookie, DgpImage_Gray *grayImage) ;
#endif
