/*
* Test the scan color pipeline...
*/
/*
25.2.2000 / dhtesq.
This is ancillary to libscancolor, and may or may not be well maintained.

*/

/*#include <libutilw.pub>		//Used? */
#include <libpixrw.pub>
#include <prxform.pub>
#include <libiow.pub>			/*Used? */
#include <thrshold.pub>
#include <logicops.pub>
#include <thrshold.h>
#include <logicops.h>
#include <iw_scale.pub>
#include <appsinc.h>
#include <halftone.pub>
#include <fill.pub>
#include <rast.pub>

/* morphology */
#include <streldef.h>
#include <shrmorph.pub>
#include <morph.h>
#include <morphg.h>
#include <strel.pub>
#include <morph.pub>

/* conected components */
#include <ccs.pub>


#include "image.h"
#include "list.h"
#include "dgp_state.h"
#include "dgp_err_control.h"
#include "dgp_err_printers.h"
#include "pixr.h"
#include "edges.h"
#include "chains.h"
#include "seg.h"
#include "scancolor.h"
#include "tfx_io.h"

#define DEFAULT_DPI 300
#define IMAGE_DPI   300
#define BIN_SIZE    4
#define IMAGE_DEPTH 256
#define HIST_SIZE   64 /* IMAGE_DEPTH / BIN_SIZE */

typedef struct _hist_entr {
  Dgp_Int32 bin_num;
  Dgp_Int32 count;
} hist_entry;

/* ***  Generate HEAPSORT  *** */
#define GENSORT_NAME                 heapsort
#define GENSORT_TYPE                 hist_entry
#define GENSORT_KEYTYPE              Dgp_Int32
#define GENSORT_INTTYPE              Dgp_Int32
#define GENSORT_GETKEY(a)            a.count
#define GENSORT_COMPAREKEYS(k1,k2)   k1 > k2
#define GENSORT_HEAPSORT
#define GENSORT_USEPOINTERS
#include "gensort.h"

/* ***  Generate HEAPSORT  *** */
#define GENSORT_NAME                 heapsort2
#define GENSORT_TYPE                 hist_entry
#define GENSORT_KEYTYPE              Dgp_Int32
#define GENSORT_INTTYPE              Dgp_Int32
#define GENSORT_GETKEY(a)            a.bin_num
#define GENSORT_COMPAREKEYS(k1,k2)   k1 < k2
#define GENSORT_HEAPSORT
#define GENSORT_USEPOINTERS
#include "gensort.h"


/* function that returns a single channel image mask */
ErrCode
DgpSeg_getBinaryBackgroundMask (Dgp_State *dgp_state, PIXR *gray_image, 
                                PIXR **mask); 


main(Dgp_Int32 argc, Dgp_Int8 *argv[]) {
  DGP_MAIN_PROLOGUE("main()", stderr, Dgp_fileErrPrinter);
  Dgp_State	        *dgp_state = NULL;
  PIXR		        *read = NULL, *background = NULL, *prTmp1 = NULL;
  PIXR		        *text = NULL, *mask = NULL, *prTmp = NULL;
  PIXR		        *token = NULL, *jpeg = NULL;
  PIXR		        *pixr_g = NULL, *pixr_r = NULL, *pixr_b = NULL;
  PIXR		        *mask_g = NULL, *mask_r = NULL, *mask_b = NULL;
  void		        *ipstate;
  Dgpsc_cookie	        *scCookie;
  Int32		        xres, yres, *histo = NULL, i;
  ScaledInt1616           tmp_xres, tmp_yres;
  Dgp_Int32		my_error_cookie = 1;
  Dgp_Int8                outfile_name[256];
  
  /*	printf("hello sctest...\n"); */
  
  /* Get command line options. */
  if(argc != 3){
    printf("Usage:\t%s [input] [output]\n", argv[0]);
    exit(1);
  }
  
  /* init ipcore */
  /*  Let Dgp_InitDgpState() do this... / dhtesq. */
  /* DGP_CHECK_NULL(ipstate = init_ipcore()); */
  
  DGP_ERR_PASS(Dgp_initDgpState(&dgp_state, NULL, NULL, stderr, Dgp_fileErrPrinter, 1, 1));
  /*  We set flags for creating an ipstate and an ipwstate, so no need to read the pointer in here. */
  /* dgp_state->ipstate = ipstate; */
  
  /*
  ErrCode
  Dgp_InitDgpState(Dgp_State **_dgp_state, Dgp_Allocators *a, void *userErrCookie, 
  dgpErrPrinterType errPrinter, Dgp_Bool createIPWState, Dgp_Bool createIPState);
  
  */
  ipstate = dgp_state->ipwstate;
  
  DGP_SET_ERR_STATE(DGP_PERRSTATE);
  
  /* read input image */
  DGP_ERR_PASS_CLEANUP(w_read(ipstate, argv[1], 1, &read));
  
  /* init scan color pipeline */
  DGP_ERR_PASS_CLEANUP(DgpSC_init(dgp_state, &scCookie));
  
  /* set scan color parameters based on input image resolution */
  DGP_ERR_PASS_CLEANUP(w_pixrGetResolutionDPI(ipstate, read, &tmp_xres, &tmp_yres));
  xres = ScaledInt1616_TO_Int32(tmp_xres);
  yres = ScaledInt1616_TO_Int32(tmp_yres);	
  if (xres == 0)
    xres = DEFAULT_DPI;
  if (yres == 0)
    yres = DEFAULT_DPI;
  DGP_ERR_PASS_CLEANUP(DgpSC_defaultOptions(scCookie, xres, yres));
  
  /* process this image */
  DGP_ERR_PASS_CLEANUP(DgpSC_process(scCookie, read));
  
  /* get a text mask */
  DGP_ERR_PASS_CLEANUP(DgpSC_getTextMask(scCookie, &text));
  
  /* save text mask */
  strcpy (outfile_name, argv[2]);
  strcat (outfile_name, "_text.pnm");
  DGP_ERR_PASS_CLEANUP(w_write(ipstate, outfile_name, text, PNM));
  /*
  if (w_write(ipstate, argv[2], text, PNM) != cSuccessful) {
		fprintf(stderr, "error saving %s\n", argv[2]);
    goto bailout;
    }
  */
  /* create a low res background image */
  DGP_ERR_PASS_CLEANUP(DgpSC_getBackground(scCookie, IMAGE_DPI, 
    &background));
  
  /* save background */
  strcpy (outfile_name, argv[2]);
  strcat (outfile_name, "_bg.pnm");
  DGP_ERR_PASS_CLEANUP(w_write(ipstate, outfile_name, background, PNM));

  /* save text & background in tfx file */
  strcpy (outfile_name, argv[2]);
  strcat (outfile_name, ".tfx");
  saveAsTfx (outfile_name, text, background, dgp_state);
  
  /* create image mask for red channel */
  DGP_ERR_PASS_CLEANUP(w_pixrGetChannel(ipstate, background, 0, 
    &pixr_r));
  DGP_ERR_PASS_CLEANUP(DgpSeg_getBinaryBackgroundMask (
    dgp_state, pixr_r, &mask_r));
  if (mask_r)
  {
    strcpy (outfile_name, argv[2]);
    strcat (outfile_name, "_mask_r.pnm");
    DGP_ERR_PASS_CLEANUP(w_write(ipstate, outfile_name, mask_r, PNM));
  }
  
  /* create image mask for green channel */
  DGP_ERR_PASS_CLEANUP(w_pixrGetChannel(ipstate, background, 1, 
    &pixr_g));
  DGP_ERR_PASS_CLEANUP(DgpSeg_getBinaryBackgroundMask (
    dgp_state, pixr_g, &mask_g));
  if (mask_g)
  {
    strcpy (outfile_name, argv[2]);
    strcat (outfile_name, "_mask_g.pnm");
    DGP_ERR_PASS_CLEANUP(w_write(ipstate, outfile_name, mask_g, PNM));
  }
  
  /* create image mask for blue channel */
  DGP_ERR_PASS_CLEANUP(w_pixrGetChannel(ipstate, background, 2, 
    &pixr_b));
  DGP_ERR_PASS_CLEANUP(DgpSeg_getBinaryBackgroundMask (
    dgp_state, pixr_b, &mask_b));
  if (mask_b)
  {
    strcpy (outfile_name, argv[2]);
    strcat (outfile_name, "_mask_b.pnm");
    DGP_ERR_PASS_CLEANUP(w_write(ipstate, outfile_name, mask_b, PNM));
  }
  
  if (mask_r && mask_g && mask_b)
  {
    
    STREL *pStrel50v = NULL, *pStrel50h = NULL, *pStrel2 = NULL, *pStrel5 = NULL;
    ConnectedComponent *p = NULL, *pCCList = NULL;
    
    /* combine the image masks */
    DGP_ERR_PASS_CLEANUP(w_copy(ipstate, mask_r, &mask));
    DGP_ERR_PASS_CLEANUP(w_logicOp(ipstate, mask_g, OR, mask));
    DGP_ERR_PASS_CLEANUP(w_logicOp(ipstate, mask_b, OR, mask));
    
      strcpy (outfile_name, argv[2]);
      strcat (outfile_name, "_bg_mask.pnm");
      DGP_ERR_PASS_CLEANUP(w_write(ipstate, outfile_name, mask, PNM));

    /*
    /* create low res text mask, subtract from image mask_array */
    DGP_ERR_PASS_CLEANUP(w_scale(ipstate, text, mask->pixrData.plane.w, 
    mask->pixrData.plane.h, 
    cFalse, IP_LO_QUALITY, &text_small));
    DGP_ERR_PASS_CLEANUP(w_logicOp(ipstate, text_small, OR, mask));
    DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipstate, &text_small));
    */
    
    /* Fill the holes
    DGP_ERR_PASS_CLEANUP (w_edgeFillSweep4 (ipstate, mask, 1, &prTmp));
    DGP_ERR_PASS_CLEANUP (w_invert (ipstate, prTmp, &prTmp));
    DGP_ERR_PASS_CLEANUP (w_logicOp (ipstate, prTmp, OR, mask));
    DGP_ERR_PASS_CLEANUP (w_destroyPixr (ipstate, &prTmp)); */
    
    DGP_ERR_PASS (w_initStrelByName (ipstate, strel50v));
    DGP_ERR_PASS (w_initStrelByName (ipstate, strel50h));
    DGP_ERR_PASS (w_initStrelByName (ipstate, strel2));
    DGP_ERR_PASS (w_initStrelByName (ipstate, strel5));
    
    DGP_ERR_PASS (w_getStrel (ipstate, strel50v, &pStrel50v));
    DGP_ERR_PASS (w_getStrel (ipstate, strel50h, &pStrel50h));
    DGP_ERR_PASS (w_getStrel (ipstate, strel2, &pStrel2));
    DGP_ERR_PASS (w_getStrel (ipstate, strel5, &pStrel5));
    
    /* dilate big text hole 
    prTmp = NULL;
    DGP_ERR_PASS_CLEANUP(w_scale(ipstate, text, mask->pixrData.plane.w, 
    mask->pixrData.plane.h, 
    cFalse, IP_LO_QUALITY, &text_small));
    DGP_ERR_PASS_CLEANUP (w_edgeFillSweep4 (ipstate, mask, 10, &prTmp));
    DGP_ERR_PASS_CLEANUP (w_invert (ipstate, prTmp, &prTmp));
    DGP_ERR_PASS_CLEANUP (w_logicOp (ipstate, text_small, AND, prTmp)); 
    
      DGP_ERR_PASS_CLEANUP (w_open (ipstate, prTmp, pStrel20v, &prTmp));
      DGP_ERR_PASS_CLEANUP (w_open (ipstate, prTmp, pStrel50h, &prTmp)); 
      DGP_ERR_PASS_CLEANUP (w_dilate (ipstate, prTmp, pStrel50h, &prTmp));
      DGP_ERR_PASS_CLEANUP (w_dilate (ipstate, prTmp, pStrel20v, &prTmp));
      DGP_ERR_PASS_CLEANUP (w_invert (ipstate, prTmp, &prTmp));
      DGP_ERR_PASS_CLEANUP (w_logicOp (ipstate, prTmp, AND, mask));
    DGP_ERR_PASS_CLEANUP (w_destroyPixr (ipstate, &prTmp)); */
    
/*    DGP_ERR_PASS_CLEANUP (w_invert(ipstate, mask, &mask)); */
    DGP_ERR_PASS_CLEANUP (w_clonePixr(ipstate, mask, &jpeg));
    DGP_ERR_PASS_CLEANUP (w_clear(ipstate, jpeg));
    DGP_ERR_PASS_CLEANUP (w_clonePixr(ipstate, mask, &token));
    DGP_ERR_PASS_CLEANUP (w_clear(ipstate, token));
    
    /* get rid of CC smaller than 2x2 */
    /*DGP_ERR_PASS_CLEANUP (w_open (ipstate, mask, pStrel2, &mask)); */
    
    DGP_ERR_PASS_CLEANUP (w_findConnectedComponents (ipstate,
      mask, 1, 1,
      0, &pCCList));
    
    for (p = pCCList; p != NULL ;  p = p->next)
    { 
      /* require a certain minimum size, */
      /* otherwise just leave as background */
      if (p->width * p->height > 50)
      {
        
        /* very narrow stuff should be tokenized */
        if (p->width < 10 || p->height < 10)
        {
          DGP_ERR_PASS_CLEANUP (w_CCSet (ipstate, p, 0, 0, token));
        }
        else
        {
          DGP_ERR_PASS_CLEANUP (w_CCSet (ipstate, p, 0, 0, jpeg));
         	
          /* get histogram of pixels under mask
          DGP_ERR_PASS_CLEANUP (w_histoMask (ipstate, pixr_g, prTmp, NULL, &histo)); */
          
          /* close the mask, horizontally and vertically
            DGP_ERR_PASS_CLEANUP (w_close (ipstate, prTmp, pStrel20v, &prTmp));
          DGP_ERR_PASS_CLEANUP (w_close (ipstate, prTmp, pStrel20h, &prTmp)); */
          
          /* fill holes */
          /*DGP_ERR_PASS_CLEANUP (w_edgeFillSweep4 (ipstate, prTmp, 1, &prTmp)); */
          
          /* copy mask into jpeg mask */
          /* DGP_ERR_PASS_CLEANUP(w_logicOp(ipstate, prTmp, OR, jpeg)); */
          
/*          DGP_FREE (histo);
          histo = NULL;
          DGP_ERR_PASS_CLEANUP (w_clear(ipstate, prTmp)); */
          
          
        }
      }
    }
    DGP_ERR_PASS_CLEANUP (w_destroyConnectedComponents (ipstate, &pCCList));
    
    /*    DGP_ERR_PASS_CLEANUP (w_clear(ipstate, mask));
    DGP_ERR_PASS_CLEANUP (w_logicOp (ipstate, prTmp, OR, mask));
    DGP_ERR_PASS_CLEANUP (w_rasterop (ipstate, 0, 0, NULL, PIX_CLR, 
    (p->x - cFrameBits), (p->y - cFrameBits),
    p->width, p->height, prTmp));
    } */

    /* 
     * move long horizontal and vertical lines from jpeg to token
     */
    DGP_ERR_PASS_CLEANUP (w_copy(ipstate, jpeg, &prTmp));
    DGP_ERR_PASS_CLEANUP (w_copy(ipstate, jpeg, &prTmp1));

    /* only keep small, long and thin components */
    DGP_ERR_PASS_CLEANUP (w_open (ipstate, prTmp, pStrel5, &prTmp));
    DGP_ERR_PASS_CLEANUP (w_logicOp (ipstate, prTmp, SUBTRACT, prTmp1));
    DGP_ERR_PASS_CLEANUP (w_destroyPixr (ipstate, &prTmp));
    prTmp = NULL;

      strcpy (outfile_name, argv[2]);
      strcat (outfile_name, "_mask_jpeg_open.pnm");
      DGP_ERR_PASS_CLEANUP(w_write(ipstate, outfile_name, prTmp1, PNM));

    /* find all lines, horizontal and vertical */
    DGP_ERR_PASS_CLEANUP (w_open (ipstate, prTmp1, pStrel50v, &prTmp));
    DGP_ERR_PASS_CLEANUP (w_open (ipstate, prTmp1, pStrel50h, &prTmp1));
    DGP_ERR_PASS_CLEANUP (w_logicOp (ipstate, prTmp, OR, prTmp1));

      strcpy (outfile_name, argv[2]);
      strcat (outfile_name, "_mask_jpeg_open2.pnm");
      DGP_ERR_PASS_CLEANUP(w_write(ipstate, outfile_name, prTmp1, PNM));


    /* subtract lines form jpeg, add to token */
    DGP_ERR_PASS_CLEANUP (w_logicOp (ipstate, prTmp1, SUBTRACT, jpeg));
    DGP_ERR_PASS_CLEANUP (w_logicOp (ipstate, prTmp1, OR, token));

    
    /* clean up jpeg mask */
    /* DGP_ERR_PASS_CLEANUP (w_open (ipstate, jpeg, pStrel5, &jpeg)); */

    /* dilate jpeg mask to compensate for previous opening */
    /* DGP_ERR_PASS_CLEANUP (w_dilate (ipstate, jpeg, pStrel3, &jpeg)); */


    if (jpeg)
    {
      strcpy (outfile_name, argv[2]);
      strcat (outfile_name, "_mask_jpeg.pnm");
      DGP_ERR_PASS_CLEANUP(w_write(ipstate, outfile_name, jpeg, PNM));
    }
    if (token)
    {
      strcpy (outfile_name, argv[2]);
      strcat (outfile_name, "_mask_token.pnm");
      DGP_ERR_PASS_CLEANUP(w_write(ipstate, outfile_name, token, PNM));
    }
    
  }
  
  /* clean up the memory */
  DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipstate, &pixr_r));
  DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipstate, &pixr_b));
  DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipstate, &pixr_g));
  DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipstate, &mask_r));
  DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipstate, &mask_b));
  DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipstate, &mask_g));
  DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipstate, &mask));
  DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipstate, &text));
  
  DGP_ERR_PASS_CLEANUP(DgpSC_done(scCookie));
  
  /*	printf("tchau sctest...\n"); */
  
  Dgp_cleanupDgpState(&dgp_state);
  MAIN_NORMAL_RETURN;
  
ERR_CLEANUP:
  
  Dgp_cleanupDgpState(&dgp_state);	
  MAIN_ERR_RETURN;
}

ErrCode
DgpSeg_getBinaryBackgroundMask (Dgp_State *dgp_state, PIXR *gray_image, 
                                PIXR **mask)
{
  
  Dgp_Int32      start, end, sum, cut_off;
  Dgp_Int32      min, max, median, n_pixels = 0;
  hist_entry     down_sampled[HIST_SIZE], down_sampled2[HIST_SIZE];
  Dgp_Int32      org_bin, new_bin;
  Dgp_Int32      l_lim, u_lim, *histogram = NULL;
  DGP_PROLOGUE ("DgpSeg_getBinaryBackgroundMask", DGP_PERRSTATE);
  IPWState       *pIPWState = dgp_state->ipwstate;
  
  const Dgp_Int32 n_highest = 6;
  const Dgp_Int32 req_peak_div = 3;
  const Dgp_Int32 median_limit = 3;
  
  /* set default return value */
  *mask = NULL;
  
  /* create histogram of pixel values */
  DGP_ERR_PASS_CLEANUP(w_histo(pIPWState, gray_image, NULL, &histogram));
  
  /* downsample histogram by BIN_SIZE */
  for (new_bin = 0; new_bin < HIST_SIZE; new_bin ++)
  {
    start = new_bin * BIN_SIZE;
    end = start + BIN_SIZE;
    sum = 0;
    for (org_bin = start; org_bin < end; org_bin++)
    {
      sum += histogram[org_bin];
    }
    down_sampled[new_bin].count = sum;
    down_sampled2[new_bin].count = sum;
    n_pixels += sum;
    down_sampled[new_bin].bin_num = new_bin;
    down_sampled2[new_bin].bin_num = new_bin;
  }
  DGP_FREE (histogram);
  histogram = NULL;
  for (new_bin = 0; new_bin < HIST_SIZE; new_bin ++)
    printf ("%3d  %3d  %6d\n", down_sampled[new_bin].bin_num, new_bin * 4 + 2, 
    down_sampled[new_bin].count);
  
  /* sort the histogram counts */
  heapsort (down_sampled, HIST_SIZE);
  median = down_sampled[HIST_SIZE / 2].count;
  
  /* get n_highest highest bins and sum neighboring bins */
  heapsort2 (down_sampled, n_highest);
  for (new_bin = 0; new_bin < n_highest - 1; new_bin++)
  {
    if (down_sampled[new_bin + 1].bin_num == 
      down_sampled[new_bin].bin_num + 1)
    {
      down_sampled[new_bin + 1].count += down_sampled[new_bin].count;
      down_sampled[new_bin].count = 0;
    }
  }
  for (new_bin = 0; new_bin < n_highest; new_bin ++)
    printf ("%3d  %3d  %6d\n", new_bin, down_sampled[new_bin].bin_num,
    down_sampled[new_bin].count);
  
  /* get highest peak */
  max = 0;
  for (new_bin = 1; new_bin < n_highest; new_bin ++)
  {
    if (down_sampled[new_bin].count > down_sampled[max].count)
    {
      max = new_bin;
    }
  }
  
  /* check if peak is significant  */
  if (down_sampled[max].count * req_peak_div < n_pixels)
  {
    NORMAL_RETURN ;
  }
  
  /* find appropriate limits */
  cut_off = median * median_limit;
  
  start = down_sampled[max].bin_num - n_highest;
  if (start < 0) start = 0;
  end = down_sampled[max].bin_num + n_highest;
  if (end > HIST_SIZE) end = HIST_SIZE;
  
  max = start;
  for (new_bin = start; new_bin < end; new_bin++)
  {
    if (down_sampled2[max].count < down_sampled2[new_bin].count)
    {
      max = new_bin;
    }
  }
  printf ("looking between %d and %d, decided on %d\n", start, end, max);
  
  min = max;
  while (min > 0 && down_sampled2[min].count > cut_off) min--;
  if (min != 0) min++;
  l_lim = min * BIN_SIZE;
  
  while (max < HIST_SIZE && down_sampled2[max].count > cut_off) max++;
  u_lim = max * BIN_SIZE;
  if (u_lim == IMAGE_DEPTH) u_lim--;
  printf ("Limits set to %d, %d   median is %d\n", l_lim, u_lim, median);
  
  /* binarize by thresholding the gray image */
  DGP_ERR_PASS_CLEANUP(w_threshold(pIPWState, gray_image, l_lim, u_lim, 
    THRESH_INVERSE, mask));
  
  printf ("\nEnd of BinaryBackground routine\n\n\n");
  NORMAL_RETURN;
  
ERR_CLEANUP:
  
  if (histogram != NULL) 
    DGP_FREE (histogram);
  ERR_RETURN;
}

