/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/
/* Trade secret boilerplate added 26.1.2000 / dhtesq. */

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb (sp?), 19nn - 1998?
**
**  Filename: seg.c.
**
**  Segmentation flow control.  In normal pipeline usage, segment() is called from
**	DgpSC_process() in scancolor.c.  Uses IPCore's w_binarize2(), which as I understand
**	is somewhat flaky and may be implemented differently in the future.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Jan. 2000
**		o  Implemented liberr exception handling.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <image.h>
/*#include <memory.h>			//Used? */
/*#include <libutilw.pub>		//Used? */
#include <libpixrw.pub>
/*#include <libiow.pub>			//Used? */
#include <binarize.pub>
#include <color.pub>
#include "dgp_state.h"
#include "findCC.h"
#include "seg.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libscancolor/seg.c,v 1.12 2003/02/10 22:53:37 mkrocken Exp $");

/* Is this used?  dhtesq. 21.1.2000 */
/*#define WIDTH 4 */


/*************************************************************************
**
**  Function: ErrCode segment(Dgp_State *dgp_state, PIXR *input, segOptions *p, DgpImage_Long *_compMap, Dgp_Int32 *_numccs)
**
**	This function segments a color image into components by using w_binarize2() and finding
**	the black & white connected components in the binary image.
**	The output is a DgpImage_Long component map indicating the component each pixel belongs to
**	(each component gets a unique label in the range 0 ... numccs - 1).
**	The image has a 1 pixel border labeled BORDER but the coordinates are set up such that 
**	0,0 is real data.
**
**	Arguments:
**		o  dgp_state points to a valid Dgp_State{}.
**		o  input references the PIXR representation of the page to be segmented.
**		o  p points to previously initialized segmentation options.
**		o  _compMap will point to the component map on successful return.
**		o  _numccs will point to the number of connected components found on successful return.
**
**	Return value:
**		o  An ErrCode.  If an error is thrown, *_compMap and *_numccs will not be modified.
**
**	Modification history:
**
**	Dillon H. Tracy esq. / January 2000
**		o  Implemented liberr exception handling.
**		o  Context modified to take Dgp_State* rather than IPWState* (to use liberr).
**		   Use dgp_state->ipwstate to recover IPWState*.
**		o  ErrCode ate the return value, so _numccs introduced as input argument.
**  
**/

ErrCode
segment(Dgp_State *dgp_state, PIXR *input, segOptions *p, DgpImage_Long *_compMap, Dgp_Int32 *_numccs) {
	DGP_PROLOGUE("segment()", DGP_PERRSTATE);

	Dgp_Int32			numccs = 0;
	DgpImage_Long		compMap = NULL;
	PIXR    			*gray = NULL;
	PIXR	    		*binary = NULL;
	DgpImage_Packed		pbinary;
	Dgp_Int32			width, height;
	Dgp_Int32			longsperline;
    Dgp_UInt32          *data_area;         /* Dummy for DgpImage_unwrap(). */
	Dgp_UInt8			*ptr;
	IPWState    		*ipstate = dgp_state->ipwstate;

#ifdef DGP_DBUG
	DGP_REPORTF0(1, "Segmenting...");
#endif

    /* Convert to gray (should be RGB). */
    DGP_CHECK_NULL(ipstate);
    DGP_CHECK_NULL(input);

    if (input->channels == 3){
		DGP_ERR_PASS(w_RGBToY(ipstate, input, &gray));
    }
    else{
		gray = input;
    }

	/* Binarize... */
	/*	c.f. James' comment on Module Summary;  is w_binarize2 unstable? / dhtesq. */
	DGP_ERR_PASS_CLEANUP(w_binarize2(ipstate, gray, 0, p->thresh, &binary));
	/*DGP_ERR_PASS_CLEANUP(w_binarize3(ipstate, gray, p->thresh, &binary)); */

	/*	Having used the grayscale image to get the binary one, can now dispose the gray one. */
    if (input->channels == 3){
		DGP_ERR_PASS(w_destroyPixr(ipstate, &gray));
    }

	/* Wrap a wjr packed image around the ipcore one. */
	DGP_ERR_PASS_CLEANUP(w_pixrGetWidth(ipstate, binary, &width));
	DGP_ERR_PASS_CLEANUP(w_pixrGetHeight(ipstate, binary, &height));
	DGP_ERR_PASS_CLEANUP(w_pixrGetImage(ipstate, binary, &ptr));
	DGP_ERR_PASS_CLEANUP(w_pixrGetBpl(ipstate, binary, &longsperline));
	longsperline /= sizeof(Dgp_Int32);

	DGP_ERR_PASS_CLEANUP(DgpImage_wrap(dgp_state, (Dgp_UInt32 *)ptr, width, height, longsperline, &pbinary));

#if defined(DGP_DBUG) && defined(SAVE_INTERMEDIATE_IMAGES)
	DGP_ERR_PASS_CLEANUP(DgpImage_save(pbinary, "seg.pbm")); 
#endif
	/*DGP_ERR_PASS_CLEANUP(DgpImage_save(pbinary, "c:\\testbed\\pnm\\a.pbm"));  */

	DGP_ERR_PASS_CLEANUP(findCC_pb(dgp_state, pbinary, binary, &compMap, p->sizeThresh, &numccs));

	/* 15.5.2001 / dhtesq. */
	/* Old cc finder didn't care about the "polarity" of the component, but this one does. */
	/* Invert the image and feed it to findCC_pb() again. */
	/* Additional ccs will just be added to the compMap. */
	DGP_ERR_PASS_CLEANUP(w_invert(ipstate, binary, &binary));
	DGP_ERR_PASS_CLEANUP(findCC_pb(dgp_state, pbinary, binary, &compMap, p->sizeThresh, &numccs));

	DGP_ERR_PASS_CLEANUP(DgpImage_unwrap(pbinary, &data_area));
	DGP_ERR_PASS(w_destroyPixr(ipstate, &binary));
	

#if defined(DGP_DBUG) && defined(SAVE_INTERMEDIATE_IMAGES)
	DGP_ERR_PASS(DGPIMAGE_SAVE(compMap, "seg.plm")); 
#endif

	*_compMap = compMap;
	*_numccs = numccs;

	NORMAL_RETURN;

ERR_CLEANUP:
	if(gray != NULL){
		DGP_ERR_PASS(w_destroyPixr(ipstate, &gray));
	}
	if(binary != NULL){
		DGP_ERR_PASS(w_destroyPixr(ipstate, &binary));
	}

	ERR_RETURN;
}
/*************************************************************************/

/*************************************************************************
**
**  Function:	void defaultSegmentOptions(segOptions *p, Dgp_Int32 xres, Dgp_Int32 yres)
**
**	Sets default heuristic options for the segmenter.  Currently this has only 
**	one parameter.
**
**	Arguments:
**		o  p points to options to be modified.
**		o  xres is horizontal resolution in dpi (?).
**		o  yres is vertical resolution in dpi (?).
**
**	Return value:
**		o  None.
**
**	Modification history:
**		o  None.
**  
**/

void
defaultSegmentOptions(segOptions *p, Dgp_Int32 xres, Dgp_Int32 yres) {
	p->thresh = 32;
	xres = xres;
	yres = yres;
}
/*************************************************************************/
