

#include "core.h"
#include "sucache.h"                /* suReleaseReference */
#include "binary_huffman_codes.h"
#include "binary_tokens.h"

#include "zephdictionary.h"
#include "zephtokenize.h"

#include "sutokposblock.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercache/binary_huffman_codes.c,v 1.21 2000/11/08 17:25:31 mkrocken Exp $");

_suEntryOps binaryHuffmanOps = { compareBinaryHuffmanCodes, 
                                 getBinaryHuffmanCodes, 
                                 destroyBinaryHuffmanCodes,
                                 calculateBinaryHuffmanCodes};

Dgp_Bool
compareBinaryHuffmanCodes( void *identifier, suCacheEntry entry) {
  suBinaryHuffmanCodes *suHuff = 
           (suBinaryHuffmanCodes *)((_suCacheEntry *)entry)->data;
  suBinaryHuffmanIdentifier *suId = (suBinaryHuffmanIdentifier *)identifier;

  SU_DEBUG2( "compareBinaryHuffmanCodes(): #%d ?= #%d\n", 
             suId->libCount, suHuff->identifier.libCount);

  if( suId->libCount != suHuff->identifier.libCount) {
    return DGP_FALSE;
  } else {
    Dgp_Int32 i, end;
    for( i = 0, end = suId->libCount; i <= end; i++) {
      if( i == end) {
        return DGP_TRUE;
      } else {
        if( suId->libIds[ i] != suHuff->identifier.libIds[ i]) {
          return DGP_FALSE;
        }
      }
    }
  }
  
  /* Never reached. */
  return DGP_FALSE;
}


ErrCode
getBinaryHuffmanCodes( suXiffCache xc, void *identifier, Dgp_Int32 currentPage, 
                       suCacheEntry *entry)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)NULL,
    *_otherEntry = (_suCacheEntry *)NULL;

  suBinaryHuffmanIdentifier *suId = (suBinaryHuffmanIdentifier *)identifier;
  suBinaryHuffmanCodes *suHuff = (suBinaryHuffmanCodes *)NULL;
  
  /* Needed to get info on each individual symlib. */
  suBinaryTokens *tokData = (suBinaryTokens *)NULL;
  suBinaryTokensIdentifier *tokId = (suBinaryTokensIdentifier *)NULL;
  zephDictionaryInfo zephTempDict = (zephDictionaryInfo)NULL;
  
  DgpAdt_ListNode notused = NULL;
  /* Keep track of token data for _all_ symlibs in this dictionary. */
  Dgp_Int32 *ntok = (Dgp_Int32 *)NULL;
  token **toks = (token **)NULL;
  
  Dgp_UInt32 usage, estUsage = 0;
  Dgp_UInt32 i;
  Dgp_Int32 j, end;

SU_PROLOGUE_XC( "getBinaryHuffmanCodes()", _xc);

  SU_DEBUG( "Entering getBinaryHuffmanCodes()...\n");

DGP_ERR_PASS_CLEANUP(
  suInitEntryInternal( xc, BINARY_HUFFMAN_CODES, (suCacheEntry *)&_entry ));

  SU_CLAIM_MEMORY( sizeof( *suHuff));
DGP_CHECK_ALLOC_CLEANUP(
  suHuff = (suBinaryHuffmanCodes *)(DGP_MALLOC)( sizeof( *suHuff)));

  /* Predictable state... */
  suHuff->identifier.libIds  = (Dgp_UInt32 *)NULL;
  suHuff->identifier.libCount = 0;
  suHuff->tokDict.zephDictState = (zephDictionaryState)NULL;
  suHuff->zephDictStateSize = 0;

  /* Copy over unique Ids to the new DictBlock. */
  suHuff->identifier.libCount = (Dgp_Int16)suId->libCount;
  SU_CLAIM_MEMORY( suId->libCount * sizeof( Dgp_UInt32));
DGP_CHECK_ALLOC_CLEANUP( 
  suHuff->identifier.libIds = (Dgp_UInt32 *)(DGP_MALLOC)( 
                                suId->libCount * sizeof( Dgp_UInt32)));
  _entry->memUsage += suId->libCount * sizeof( Dgp_UInt32);

  SU_DEBUG( "Request for dictionaries:\n");

  for( i = 0; i < suId->libCount; i++) {
    SU_DEBUG1( "\t%d\n", suId->libIds[ i]);
    suHuff->identifier.libIds[ i] = suId->libIds[ i];
  }

  /* Allocate space for the individual sets of tokens. */
  SU_CLAIM_MEMORY( suId->libCount * (sizeof( Dgp_Int32) + sizeof( token *)));
DGP_CHECK_ALLOC_CLEANUP(
  ntok = (Dgp_Int32 *)DGP_MALLOC( suId->libCount * sizeof( Dgp_Int32)));
DGP_CHECK_ALLOC_CLEANUP(
  toks = (token **)DGP_MALLOC( suId->libCount * sizeof( token *)));

  /* Set up the identifier for the individual sets of tokens. */
  SU_CLAIM_MEMORY( sizeof( *tokId)); 
DGP_CHECK_ALLOC_CLEANUP( 
  tokId = (suBinaryTokensIdentifier *)DGP_MALLOC( sizeof( *tokId)));

  /* Let's do it! */
  for( i = 0; i < suId->libCount; i++) {
    _otherEntry = (_suCacheEntry *)NULL;

DGP_ERR_PASS_CLEANUP(
    zephSplitDictionary( suId->zephDict, i, &zephTempDict));
    SU_CLAIM_MEMORY( zephGetDictionaryMemoryUsage( zephTempDict));

    tokId->libId = zephTempDict->libIds[ 0];
    tokId->zephDict = zephTempDict;

DGP_ERR_PASS_CLEANUP(
    suGetEntryInternal( xc, BINARY_TOKENS, tokId,
                        currentPage, (suCacheEntry *)&_otherEntry));

    /* This usage will include any memory used for temporary storage of the
     * embedded dictionary. */ 
    usage = zephGetDictionaryMemoryUsage( zephTempDict);
DGP_ERR_PASS_CLEANUP(
    zephCloseDictionary( &zephTempDict));
    SU_RELEASE_MEMORY( usage);

    /* Add the new entry to out list. */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd( _entry->relations, _otherEntry, &notused));


    /* Add ourselves to the relations list of this entry. */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd( _otherEntry->relations, _entry, &notused));
    /* Make sure that none of the tokens are "committed".  */
    tokData = (suBinaryTokens *)_otherEntry->data;
    for( j = 0, end = tokData->ntok; j < end; j++) {
      tokData->toks[ j].committed = DGP_FALSE;
    }

    /* Save pointer to the token data. */
    ntok[ i] = tokData->ntok;
    toks[ i] = tokData->toks;
  }
    
  DGP_FREE( (void *)tokId);
  tokId = (suBinaryTokensIdentifier *)NULL;
  SU_RELEASE_MEMORY( sizeof( *tokId));

  _otherEntry = (_suCacheEntry *)NULL;

  /* Now the dirty work... build the codes. */
  usage = 0;

DGP_ERR_PASS_CLEANUP(
  zephInitializeDictionaryState( suId->zephDict,
                                 &suHuff->tokDict.zephDictState,
                                 &usage));
  SU_CLAIM_MEMORY( usage);
  _entry->memUsage += usage;
  suHuff->zephDictStateSize = usage;
  usage = 0;

/*DGP_ERR_PASS_CLEANUP( 
  zephEstimateDecodedDictionaryStateSize( suHuff->tokDict.zephDictState,
                                          (zephDictionaryState)NULL,
                                          &estUsage));
  SU_CLAIM_MEMORY( estUsage); */

DGP_ERR_PASS_CLEANUP(
  zephDecodeBinaryDictionaryState( suHuff->tokDict.zephDictState,
                                   (zephDictionaryState)NULL,
                                   ntok,
                                   toks,
                                   &usage));
/*  if( usage > estUsage)
  { 
    SU_CLAIM_MEMORY( usage - estUsage); 
  }
  else if( estUsage > usage)
  {
    SU_RELEASE_MEMORY( estUsage - usage);
  } */
  SU_CLAIM_MEMORY(usage);
  _entry->memUsage += usage;
  suHuff->zephDictStateSize += usage;

  /* Take out the trash. */
  DGP_FREE( (void *)ntok);
  ntok = (Dgp_Int32 *)NULL;
  SU_RELEASE_MEMORY( suId->libCount * sizeof( Dgp_Int32));
  DGP_FREE( (void *)toks);
  toks = (token **)NULL;
  SU_RELEASE_MEMORY( suId->libCount * sizeof( token *));
 
  /* Make sure to set the usage before called suAddEntryInternal. */
  _entry->memUsage += sizeof( *suHuff);
  _entry->data = (void *)suHuff;
  
  /* Tidy up the cache state. */
DGP_ERR_PASS_CLEANUP(
  suAddEntryInternal( xc, (suCacheEntry)_entry));

  /* And copy back the result. */
  *entry = (suCacheEntry)_entry;

  NORMAL_RETURN;
    
ERR_CLEANUP:
  if( suHuff != (suBinaryHuffmanCodes *)NULL) {
    if( suHuff->identifier.libIds != (Dgp_UInt32 *)NULL) {
      DGP_FREE( (void *)suHuff->identifier.libIds);
      SU_RELEASE_MEMORY_VOID( suId->libCount * sizeof( Dgp_UInt32));
      _entry->memUsage -= suId->libCount * sizeof( Dgp_UInt32);
    }
    if( suHuff->tokDict.zephDictState != (void *)NULL) {
      zephFreeDictionaryState( &suHuff->tokDict.zephDictState);
      SU_RELEASE_MEMORY_VOID( suHuff->zephDictStateSize);
      _entry->memUsage -= suHuff->zephDictStateSize;
    }

    DGP_FREE( (void *)suHuff);
    SU_RELEASE_MEMORY_VOID( sizeof( *suHuff));
    if( _entry->data != (suBinaryHuffmanCodes *)NULL)
    {
      _entry->memUsage -= sizeof( *suHuff);
    }
    _entry->data = (void *)NULL;
  }

  if( ntok != (Dgp_Int32 *)NULL) {
    DGP_FREE( (void *)ntok);
    SU_RELEASE_MEMORY_VOID( suId->libCount * sizeof( Dgp_Int32));
  }
  if( toks != (token **)NULL) {
    DGP_FREE( (void *)toks);
    SU_RELEASE_MEMORY_VOID( suId->libCount * sizeof( token*));
  }
  
  if( tokId != (suBinaryTokensIdentifier *)NULL) {
    DGP_FREE( (void *)tokId);
    SU_RELEASE_MEMORY_VOID( sizeof( tokId));
  }

  /* We know that the entry is NOT in the cache. */
  if( _entry != (_suCacheEntry *)NULL) {
    suFreeEntryInternal( xc, (suCacheEntry)_entry, DGP_FALSE);
  }
  
  ERR_RETURN;
}


void
destroyBinaryHuffmanCodes( suXiffCache xc, suCacheEntry entry) {
    _suXiffCache *_xc = (_suXiffCache *)xc;
    _suCacheEntry *_entry = (_suCacheEntry *)entry, *_relEntry;
    Dgp_State *dgp_state = _xc->options.dgp_state;
    suBinaryHuffmanCodes *suHuff;
    DgpAdt_ListNode thisLn, relLn;
    Dgp_UInt32 i;

DGP_VOID_PROLOGUE( "destroyBinaryHuffmanCodes()", DGP_PERRSTATE);

  /* First go through each of the relations and references to
   * the current entry. We assume that all these entries are
   * binary token entries. (Only make this pass during normal
   * cleanup, not cache destruction.) */
  if( _entry->uses >= 0) {
    for( relLn = DGPADT_LIFIRST( _entry->relations); relLn != DGP_NULLLISTNODE;
           DgpAdt_liRemAndNext( relLn, &relLn)) {
      _relEntry = (_suCacheEntry *)DGPADT_LIGET( relLn);
      if( !isBinaryTokenEntry( _relEntry)) {
        DGP_WARNF0( "relative entry is not a binary token entry!");
      }
                
      DgpAdt_liIsIn( _relEntry->relations, entry, &thisLn);
      if( thisLn == DGP_NULLLISTNODE) {
        DGP_WARNF0( "this entry is not cross referenced by relation.");
      }
                
      /* Remove the reference from the other entry. */
      DgpAdt_liRem( thisLn);

      /* Decrement the other entry's reference count. */
      suReleaseReference( (suCacheEntry)_relEntry);
    }
  }

  if( _entry->data != (void *)NULL) {
    suHuff = (suBinaryHuffmanCodes *)_entry->data;

    SU_DEBUG( "Freeing dictionaries:\n");
    for( i = 0; i < suHuff->identifier.libCount; i++) {
      SU_DEBUG1( "\t%d\n", suHuff->identifier.libIds[ i]);
    }

    /* Now free the memory associated with this entry. */
    zephFreeDictionaryState( &suHuff->tokDict.zephDictState);
    SU_RELEASE_MEMORY_VOID( suHuff->zephDictStateSize);
    _entry->memUsage -= suHuff->zephDictStateSize;

    DGP_FREE( (void *)suHuff->identifier.libIds);
    SU_RELEASE_MEMORY_VOID( suHuff->identifier.libCount 
                            * sizeof( Dgp_UInt32));
    _entry->memUsage -= suHuff->identifier.libCount * sizeof( Dgp_UInt32);

    DGP_FREE( (void *)_entry->data);
    SU_RELEASE_MEMORY_VOID( sizeof( *suHuff));
    _entry->memUsage -= sizeof( *suHuff);
    _entry->data = (void *)NULL;
  }
  
  VOID_RETURN;
}

void
calculateBinaryHuffmanCodes( suXiffCache xc, suCacheEntry entry)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  Dgp_State *dgp_state = _xc->options.dgp_state;
  _suCacheEntry *_entry = (_suCacheEntry *)entry, *_relEntry;
  DgpAdt_ListNode ln;
  
DGP_VOID_PROLOGUE( "calculateBinaryHuffmanCodes()", DGP_PERRSTATE);

  /* There is no data that can change size in the dictionary state, but the 
   * bitmaps in the token structures can change. So call calculate on each
   * of the libraries of binary tokens that this dictionary references. */

  DGPADT_FOREACH( ln, _entry->relations) {
    _relEntry = (_suCacheEntry *)DGPADT_LIGET( ln);

    if( !isBinaryTokenEntry( _relEntry)) {
      DGP_WARNF0( "relative entry is not a binary token entry!");
    }

    suRecalculateUsageInternal( xc, (suCacheEntry)_relEntry);
  }
}
