
#include "core.h"
#include "binary_tokens.h"

#include "zephtokenize.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercache/binary_tokens.c,v 1.19 2003/02/10 22:53:37 mkrocken Exp $");


_suEntryOps binaryOps = { compareBinaryTokens, 
                          getBinaryTokens, 
                          destroyBinaryTokens,
                          calculateBinaryTokens};

Dgp_Bool
compareBinaryTokens( void *identifier, suCacheEntry entry) 
{
  suBinaryTokens *tokData = (suBinaryTokens *)((_suCacheEntry *)entry)->data;
  suBinaryTokensIdentifier *tokId = (suBinaryTokensIdentifier *)identifier;

  SU_DEBUG2( "compareBinaryTokens(): %d ?= %d\n", tokId->dictId, 
             tokData->identifier.dictId);

  if( tokData->identifier.dictId == tokId->dictId) {
    return DGP_TRUE;
  } else {
    return DGP_FALSE;
  }
}


ErrCode
getBinaryTokens( suXiffCache xc, void *identifier, Dgp_Int32 currentPage, 
                 suCacheEntry *entry) 
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)NULL;
  suBinaryTokens *tokData = (suBinaryTokens *)NULL;
  suBinaryTokensIdentifier *binTokId;
  Dgp_UInt32 usage;


SU_PROLOGUE_XC( "getBinaryTokens()", _xc);

  currentPage = currentPage;
  binTokId = (suBinaryTokensIdentifier *)identifier;

SU_DEBUG( "Entering getBinaryTokens()...\n");
SU_DEBUG1( "Request for tokens with Id: %d\n", binTokId->dictId);

DGP_ERR_PASS_CLEANUP(
  suInitEntryInternal( xc, BINARY_TOKENS, (suCacheEntry *)&_entry ));

  SU_CLAIM_MEMORY( sizeof( *tokData)); 
DGP_CHECK_ALLOC_CLEANUP(
  tokData = (suBinaryTokens *)(DGP_MALLOC)( sizeof( *tokData)));

  /* Predictable state... */
  tokData->identifier.dictId = binTokId->dictId;
  tokData->identifier.imageInfo = binTokId->imageInfo;
  tokData->segment = (jb2Segment *)NULL;
    
  /* Now as libzephyr for the data! */
DGP_ERR_PASS_CLEANUP(
  zephReadBinaryDictionary ( binTokId->imageInfo, binTokId->dictId, 
			     &tokData->segment, &usage));

 SU_CLAIM_MEMORY( usage );
  tokData->mem_usage = usage;

  _entry->memUsage += usage;
  _entry->memUsage += sizeof( *tokData); 
  _entry->data = (void *) tokData;
  
  /* Tidy up the cache state. */
DGP_ERR_PASS_CLEANUP(
  suAddEntryInternal( xc, (suCacheEntry)_entry));

  /* Copy over the results. */
  *entry = (suCacheEntry)_entry;

  NORMAL_RETURN;

 ERR_CLEANUP:
    /* FIXME extensions */

  /* We know that the entry was NOT successfully added to the cache. */
  if( _entry != (_suCacheEntry *)NULL) {
    suFreeEntryInternal( xc, (suCacheEntry)_entry, DGP_FALSE);
  }
  
  ERR_RETURN;
}


void
destroyBinaryTokens( suXiffCache xc, suCacheEntry entry) {
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)entry;
  Dgp_State *dgp_state = _xc->options.dgp_state;
  
  suBinaryTokens *tokData = (suBinaryTokens *)_entry->data;

  if( tokData != (suBinaryTokens *)NULL) {

    SU_DEBUG1( "Freeing binary tokens with Id: %d\n", 
               tokData->identifier.libId);

    /* free the dictionary segment */
    jb2FreeSegmentInfo(tokData->segment, dgp_state); 
    DGP_FREE (tokData->segment);

    _entry->memUsage -= tokData->mem_usage;
    DGP_FREE( (void *)tokData);
    SU_RELEASE_MEMORY_VOID( sizeof( *tokData));
    _entry->memUsage -= sizeof( *tokData);
    _entry->data = (void *)NULL;
  }
  return;
}


void
calculateBinaryTokens( suXiffCache xc, suCacheEntry entry)
{
  _suCacheEntry *_entry = (_suCacheEntry *)entry;
  _suXiffCache *_xc = (_suXiffCache *)xc;
  suBinaryTokens *tokData = (suBinaryTokens *)_entry->data;
  Dgp_UInt32 newBitmapUsage = 0;

  newBitmapUsage = newBitmapUsage;
  tokData = tokData;
  _xc = _xc;
}
