/*

  19.6.2000 / dhtesq.

  Reconditioning color parts of libsupercache following WB's revision.
  This file is not required for the black-and-white case.

*/

#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif


#include "libsupercache.h"
#include "core.h"
#include "color_dictionary.h"
#include "color_tokens.h"
#include "colortoken_t.h"

#include "list.h"		    /* DgpAdt_List */
#include "dgp_types.h"		/* Dgp_Bool */
#include "resolution_t.h"


_suEntryOps colorDictionaryOps = { compareColorDictionary, getColorDictionary, destroyColorDictionary, NULL};

Dgp_Bool
compareColorDictionary( void *identifier, suCacheEntry entry) {
    suColorDictionary *dictData = (suColorDictionary *)((_suCacheEntry *)entry)->data;
    colorDictInfo *dictId = (colorDictInfo *)identifier;

    SU_DEBUG2( "compareColorDictionary(): #%d ?= #%d\n", dictId->num_dict_ids,
    			dictData->identifier.num_dict_ids);

    if( dictId->num_dict_ids != dictData->identifier.num_dict_ids) {
        return DGP_FALSE;
    } else {
        Dgp_Int32 i, end;
    	for( i = 0, end = dictId->num_dict_ids; i <= end; i++) {
    	    if( i == end) {
    	        return DGP_TRUE;
    	    } else {
    	        if( dictId->dict_ids[ i] != dictData->identifier.dict_ids[ i]) {
        		    return DGP_FALSE;
       	        }
    	    }
    	}
    }

    /* Never reached. */
    return DGP_FALSE;
}


ErrCode
getColorDictionary( suXiffCache xc, void *identifier, Dgp_Int32 currentPage, suCacheEntry *entry) {
    _suXiffCache *_xc = (_suXiffCache *)xc;
    _suCacheEntry *_entry = (_suCacheEntry *)NULL,
    		*_otherEntry = (_suCacheEntry *)NULL;
    colorDictInfo *dictId = (colorDictInfo *)identifier;
    suColorDictionary *dictData = (suColorDictionary *)NULL;
	DgpAdt_ListNode retnode;

    suColorTokens *tokData = (suColorTokens *)NULL;
    DgpAdt_ListNode ln;
    Dgp_Int32 i, j, end;

SU_PROLOGUE_XC( "getColorDictionary()", _xc);

    SU_DEBUG( "Entering getColorDictionary()...\n");

DGP_ERR_PASS_CLEANUP(
    suInitEntryInternal( xc, COLOR_DICTIONARY, (suCacheEntry *)&_entry ));

DGP_CHECK_ALLOC_CLEANUP(
    dictData = (suColorDictionary *)DGP_MALLOC( sizeof( *dictData)));

    /* Predictable state... */
    dictData->identifier.dict_ids = (Dgp_UInt32 *)NULL;
    dictData->identifier.num_dict_ids = 0;
    dictData->toks = (color_token **)NULL;
    dictData->ntok = 0;

    /* Copy over unique Ids to the new suColorDictionary. */
    dictData->identifier.num_dict_ids = dictId->num_dict_ids;
    if( dictData->identifier.num_dict_ids > 0) {
DGP_CHECK_ALLOC_CLEANUP(
        dictData->identifier.dict_ids = (Dgp_UInt32 *)DGP_MALLOC(
			dictId->num_dict_ids * sizeof( Dgp_UInt32)));
    }

    SU_DEBUG( "Request for color dictionaries:\n");

    for( i = 0; i < dictId->num_dict_ids; i++) {
        SU_DEBUG1( "\t%d\n", dictData->identifier.dict_ids[ i]);

        dictData->identifier.dict_ids[ i] = dictId->dict_ids[ i];
    }

    /* Let's do it! */
    for( i = 0; i < dictId->num_dict_ids; i++) {
        _otherEntry = (_suCacheEntry *)NULL;

DGP_ERR_PASS_CLEANUP(
        suGetEntryInternal( xc, COLOR_TOKENS, &dictId->dict_ids[i], currentPage,
    		(suCacheEntry *)&_otherEntry));

        /* Add the new entry to out list. */
	    DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd( _entry->relations, _otherEntry, &retnode));

        /* Add ourselves to the relations list of this entry. */
	    DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd( _otherEntry->relations, _entry, &retnode));

	    dictData->ntok += ((suColorTokens *)_otherEntry->data)->ntok;
    }

    if( dictData->ntok > 0) {

DGP_CHECK_ALLOC_CLEANUP(
        dictData->toks = (color_token **)DGP_MALLOC( dictData->ntok *
        		sizeof( color_token *)));

DGP_ASSERT( DGPADT_LILEN( _entry->relations) > 0);
        ln = DGPADT_LIFIRST( _entry->relations);
        _otherEntry = (_suCacheEntry *)DGPADT_LIGET( ln);
        tokData = (suColorTokens *)_otherEntry->data;
        for( i = 0, end = dictData->ntok, j = 0; i < end; i++, j++)
        {
            if( j == tokData->ntok)
		    {
		        ln = DGPADT_LINEXT( ln);
		        _otherEntry = (_suCacheEntry *)DGPADT_LIGET( ln);
		        tokData = (suColorTokens *)_otherEntry->data;
		        j = 0;
		    }

		    dictData->toks[i] = &tokData->toks[j];
        }
    }

    _otherEntry = (_suCacheEntry *)NULL;

    _entry->memUsage = sizeof( *_entry);
    _entry->memUsage += sizeof( *dictData);
    _entry->memUsage += (dictData->ntok * sizeof( color_token *));

    _entry->data = (void *)dictData;

    /* Tidy up the cache state. */
DGP_ERR_PASS_CLEANUP(
    suAddEntryInternal( xc, (suCacheEntry)_entry));

    /* And copy back the result. */
    *entry = (suCacheEntry)_entry;

    NORMAL_RETURN;

ERR_CLEANUP:
    if( dictData != (suColorDictionary *)NULL) {
        if( dictData->identifier.dict_ids != (Dgp_UInt32 *)NULL) {
            DGP_FREE( (void *)dictData->identifier.dict_ids);
    	    dictData->identifier.dict_ids = (Dgp_UInt32 *)NULL;
    	}
        DGP_FREE( (void *)dictData);
    	_entry->data = (void *)NULL;
    }

    if( _entry != (_suCacheEntry *)NULL) {
        /* This now takes an additional argument of Dgp_Bool inCache. */
        /* The entry will only be destroyed if this is true...?  Looks  */
        /* like DGP_FALSE is used mostly in cleanup cases. / dhtesq. */
        /*suFreeEntryInternal( xc, (suCacheEntry)_entry); */
        suFreeEntryInternal( xc, (suCacheEntry)_entry, DGP_FALSE);
    }
    ERR_RETURN;
}


void
destroyColorDictionary( suXiffCache xc, suCacheEntry entry) {
    _suXiffCache *_xc = (_suXiffCache *)xc;
    _suCacheEntry *_entry = (_suCacheEntry *)entry, *_relEntry;
    Dgp_State *dgp_state = _xc->options.dgp_state;
    suColorDictionary *dictData = (suColorDictionary *)_entry->data;
    DgpAdt_ListNode thisLn, relLn;
    Dgp_Int32 i;

DGP_VOID_PROLOGUE( "destroyColorDictionary()", DGP_PERRSTATE);

    /* First go through each of the relations and references to
     * the current entry. We assume that all these entries are
     * color token entries. (Only make this pass during normal
     * cleanup, not cache destruction.) */
    if( _entry->uses >= 0) {
        for( relLn = DGPADT_LIFIRST( _entry->relations); relLn != DGP_NULLLISTNODE;
        		DgpAdt_liRemAndNext( relLn, &relLn)) {
            _relEntry = (_suCacheEntry *)DGPADT_LIGET( relLn);

            if( !isColorTokenEntry( _relEntry)) {
                DGP_WARNF0( "relative entry is not a color token entry!");
            }

       	    DgpAdt_liIsIn( _relEntry->relations, entry, &thisLn);
    	    if( thisLn == DGP_NULLLISTNODE) {
                DGP_WARNF0( "this entry is not cross referenced by relation.");
            }

       	    /* Remove the reference from the other entry. */
    	    DgpAdt_liRem( thisLn);

       	    /* Decrement the other entry's reference count. */
    	    suReleaseReference( (suCacheEntry)_relEntry);
        }
    }

    if( _entry->data != (suColorDictionary *)NULL) {

        SU_DEBUG( "Freeing color dictionary:\n");
        for( i = 0; i < dictData->identifier.num_dict_ids; i++) {
            SU_DEBUG1( "\t%d\n", dictData->identifier.dict_ids[ i]);
        }

        DGP_FREE( (void *)dictData->identifier.dict_ids);
        DGP_FREE( (void *)dictData->toks);
        DGP_FREE( (void *)_entry->data);
        _entry->data = (void *)NULL;
    }

    VOID_RETURN;
}
