/*

  19.6.2000 / dhtesq.

  Reconditioning color parts of libsupercache following WB's revision.
  This file is not required for the black-and-white case.

*/

#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif


#include "libsupercache.h"
#include "core.h"
#include "color_position_extension.h"
#include "raw_position_extension.h"
#include "token_positions.h"

#include "list.h"		/* DgpAdt_List */
#include "util.h"		/* Dgp_Bool */
#include "extension_tags.h"
#include "zephtokenize.h"

_suEntryOps colorPositionExtensionOps = { compareColorPositionExtension,
	getColorPositionExtension, destroyColorPositionExtension, NULL};

Dgp_Bool
compareColorPositionExtension( void *identifier, suCacheEntry entry) {
    suColorPositionExtension *extData =
    	(suColorPositionExtension *)((_suCacheEntry *)entry)->data;
    suPositionExtensionIdentifier *extId =
        (suPositionExtensionIdentifier *)identifier;

    SU_DEBUG2( "compareColorPositionExtension(): %d ?= %d\n",
    			extId->page, extData->identifier.page);
    SU_DEBUG2( "                                  %d ?= %d\n",
    			extId->layer, extData->identifier.layer);
    SU_DEBUG2( "                                  %d ?= %d\n",
    			extId->image, extData->identifier.image);

    if( (extData->identifier.page == extId->page) &&
        	(extData->identifier.layer == extId->layer) &&
    		(extData->identifier.image == extId->image)) {
    	return DGP_TRUE;
    } else {
        return DGP_FALSE;
    }
}


ErrCode
getColorPositionExtension( suXiffCache xc, void *identifier, Dgp_Int32 currentPage,
		suCacheEntry *entry) {
    _suXiffCache *_xc = (_suXiffCache *)xc;
    _suCacheEntry *_entry = (_suCacheEntry *)NULL, *_otherEntry = (_suCacheEntry *)NULL;
    suPositionExtensionIdentifier *extId = (suPositionExtensionIdentifier *)identifier;
    suColorPositionExtension *extData = (suColorPositionExtension *)NULL;
    suPositionIdentifier *posId = (suPositionIdentifier *)NULL;
    suTokenPositions *posData = (suTokenPositions *)NULL;

    Extension *colorExt;
    Dgp_Int32 i;
    Dgp_UInt32 mem_usage = 0;

SU_PROLOGUE_XC( "getColorPositionExtension()", _xc);

    SU_DEBUG( "Entering getColorPositionExtension()...\n");
    SU_DEBUG1( "Request for color extension on page: %d\n", extId->page);

DGP_ERR_PASS_CLEANUP(
    suInitEntryInternal( xc, COLOR_POSITION_EXTENSION,
    		(suCacheEntry *)&_entry ));

DGP_CHECK_ALLOC_CLEANUP(
    extData = (suColorPositionExtension *)DGP_MALLOC( sizeof( *extData)));

    /* Predictable state... */
    extData->identifier.page = extId->page;
    extData->identifier.layer = extId->layer;
    extData->identifier.image = extId->image;
    extData->identifier.extTag = extId->extTag;
    extData->identifier.zephImage = extId->zephImage;

    extData->color_tags = (color_tag *)NULL; 
    extData->color_dictinfo = NULL;

    /* Get the position block from the cache.  */
DGP_CHECK_ALLOC_CLEANUP(
      posId = (suPositionIdentifier *)DGP_MALLOC( sizeof( *posId)));
    
    posId->page = extId->page;
    posId->layer = extId->layer;
    posId->image = extId->image;
    posId->extDecoded = ~POS_EXTFLAG_COLOR;
    
    posId->zephImage = extId->zephImage;        /* dhtesq. */
    /*posId->imgHandle = extId->imgHandle; */
    /*posId->imWidth = extId->imWidth; */
    /*posId->imHeight = extId->imHeight; */
    
DGP_ERR_PASS_CLEANUP(
      suGetEntryInternal( xc, TOKEN_POSITIONS, (void *)posId, currentPage,
      (suCacheEntry *)&_otherEntry));
    
    DGP_FREE( (void *)posId);
    posId = (suPositionIdentifier *)NULL;
    posData = (suTokenPositions *)_otherEntry->data;
    
    if (!posData->posBlock.colorExt)
    {
      /* There is no color extension, set them all to RGB = 0,0,0 = black */
      extData->ntag = posData->posBlock.posCount;
DGP_CHECK_ALLOC_CLEANUP(
        extData->color_tags = DGP_CALLOC (extData->ntag * sizeof (color_tag), 1));
      mem_usage = extData->ntag * sizeof (color_tag);
      colorExt = NULL;
      for (i = 0; i < extData->ntag; i++)
      {
        extData->color_tags[i].color_type = RGB_COLOR;
      }
    }
    else
    {
      /* decode the color extensions */
      colorExt = posData->exts;
      
      /* This extension should really still be there. */
      DGP_ASSERT( colorExt->tag == COLOR_EXT_TAG);
      DGP_ASSERT( colorExt->extension != (void *)NULL);
      
      /* Now ask libxiff to decode them for us. Isn't that nice. */
      /* Make that libzephyr... / dhtesq. */
      DGP_ERR_PASS_CLEANUP(zephDecodeColorPositionExtension(  extData->identifier.zephImage,
        colorExt,
        &extData->ntag,
        &extData->color_tags,
        &mem_usage));      
      
    }
    
    
    
    /* Check to see if it has the right number of tags. */
    if( extData->ntag != posData->posBlock.posCount /*nbox*/) {     /* dhtesq. */
      DGP_ERR_CLEANUPF0( cErrDecodeFailure,
        "number of position color tags doesn't match number of positions.");
    }
    
    /* Clean up and make sure to release the reference to the position block. */
    posData->identifier.extDecoded |= POS_EXTFLAG_COLOR;
    
    
    suReleaseReference( (suCacheEntry)_otherEntry);
    _otherEntry = (_suCacheEntry *)NULL;
    
    /*_entry->memUsage = (extData->ntag * sizeof( color_tag)); */
    _entry->memUsage += mem_usage;
    /*_entry->memUsage += (extData->identifier.zephImage->libCount * sizeof( Dgp_UInt32)); */
    _entry->memUsage += sizeof( *extData);
    /*_entry->memUsage += sizeof( *_entry); */
    
    _entry->data = (void *)extData;
    
    /* Tidy up the cache state. */
    DGP_ERR_PASS_CLEANUP(
      suAddEntryInternal( xc, (suCacheEntry)_entry));
    
    /* Copy over the results. */
    *entry = (suCacheEntry)_entry;
    
    NORMAL_RETURN;

ERR_CLEANUP:
    if( posId != (suPositionIdentifier *)NULL) {
        DGP_FREE( (void *)posId);
    }

    if( extData != (suColorPositionExtension *)NULL) {
	    if( extData->color_tags != (color_tag *)NULL) {
            DGP_FREE( (void *)extData->color_tags);
	    }
        DGP_FREE( (void *)extData);
	    _entry->data = (suColorPositionExtension *)NULL;
    }

    if( posData != (suTokenPositions *)NULL) {
        DGP_FREE( (void *)posData);
    }
    if( _otherEntry != (_suCacheEntry *)NULL) {
        suReleaseReference( (suCacheEntry)_otherEntry);
    }
    if( _entry != (_suCacheEntry *)NULL) {
        suFreeEntryInternal( xc, (suCacheEntry)_entry, DGP_FALSE);
    }

    ERR_RETURN;
}


void
destroyColorPositionExtension( suXiffCache xc, suCacheEntry entry) {
    _suXiffCache *_xc = (_suXiffCache *)xc;
    _suCacheEntry *_entry = (_suCacheEntry *)entry;
    Dgp_State *dgp_state = _xc->options.dgp_state;
    suColorPositionExtension *extData =
    	(suColorPositionExtension *)_entry->data;

    if( extData != (suColorPositionExtension *)NULL) {

        SU_DEBUG1( "Freeing color extension for page: %d\n", extData->identifier.page);

        if( extData->color_tags != (color_tag *)NULL) {
            DGP_FREE( (void *)extData->color_tags);
            _entry->memUsage -= extData->ntag * sizeof(color_tag);
        }

        DGP_FREE((void *)extData);
        _entry->memUsage -= sizeof(*extData);

        _entry->data = (void *)NULL;
    }
}
