/*

  19.6.2000 / dhtesq.

  Reconditioning color parts of libsupercache following WB's revision.
  Apparently this file is not required for the black-and-white case.

*/
/*#if 0 */

#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif


#include "libsupercache.h"
#include "core.h"
#include "color_tokens.h"
#include "libpixrw.pub"         /* Added for PIXR / dhtesq. */
#include "props.pub"

_suEntryOps colorOps = { compareColorTokens, getColorTokens, destroyColorTokens, NULL };


Dgp_Bool
compareColorTokens( void *identifier, suCacheEntry entry) {
    suColorTokens *tokData = (suColorTokens *)((_suCacheEntry *)entry)->data;

    SU_DEBUG2( "compareColorTokens(): %d ?= %d\n", *(Dgp_UInt32 *)identifier, tokData->uniqueId);

    if( tokData->uniqueId == *(Dgp_UInt32 *)identifier) {
        return DGP_TRUE;
    } else {
        return DGP_FALSE;
    }
}

/*
This function still reads the color tokens into PIXRs.

We'll try to move over to libzephyr and then come back and change this.
dhtesq.
*/
ErrCode
getColorTokens( suXiffCache xc, void *identifier, Dgp_Int32 currentPage, suCacheEntry *entry) {
    _suXiffCache *_xc = (_suXiffCache *)xc;
    _suCacheEntry *_entry = (_suCacheEntry *)NULL;
    suColorTokens *tokData = (suColorTokens *)NULL;
    DgpAdt_List images = DGP_NULLLIST;
    DgpAdt_ListNode ln;
    PIXR *pix;
    Dgp_Int32 numBytes;
    Dgp_Int32 xres = 0, yres = 0, i;

SU_PROLOGUE_XC( "getColorTokens()", _xc);

    currentPage = currentPage;

    SU_DEBUG( "Entering getColorTokens()...\n");
    SU_DEBUG1( "Request for tokens with Id: %d\n", *(Dgp_Int32 *)identifier);

DGP_ERR_PASS_CLEANUP(
    suInitEntryInternal( xc, COLOR_TOKENS, (suCacheEntry *)&_entry ));

DGP_CHECK_ALLOC_CLEANUP(
    tokData = (suColorTokens *)DGP_MALLOC( sizeof( *tokData)));

    /* Predictable state... */
    tokData->uniqueId = *(Dgp_Int32 *)identifier;
    tokData->toks = (color_token *)NULL;

    /* Now as libxiff for the data! */
    /* Make that libzephyr... / dhtesq. */

    /*
DGP_ERR_PASS_CLEANUP(
    readXiffColorDictionary( *(Dgp_Int32 *)identifier, &images, &xres, &yres,
            _xc->zephDoc));
    */

    _entry->memUsage = sizeof( *_entry);
    _entry->memUsage += sizeof( *tokData);
    _entry->memUsage += (sizeof( color_token) * DGPADT_LILEN( images));

    tokData->ntok = DGPADT_LILEN( images);
DGP_CHECK_ALLOC_CLEANUP(
    tokData->toks = (color_token *)DGP_MALLOC( tokData->ntok
            * sizeof( color_token)));

    for( i = 0; i < tokData->ntok; i++)
    {
    	tokData->toks[i].res.xRes = DgpCoord_eqi( xres);
    	tokData->toks[i].res.yRes = DgpCoord_eqi( yres);
    	tokData->toks[i].type = SU_COLOR_TOKEN_IMAGETYPE_PIXR;
    	tokData->toks[i].userFlags = 0;
        /* image field is now void *.... */
        tokData->toks[i].image = NULL;
    	/*tokData->toks[i].image.pixr = (PIXR *)NULL;  //dhtesq. */
    }

    for( i = 0, ln = DGPADT_LIFIRST( images); i < tokData->ntok; i++, ln = DGPADT_LINEXT( ln))
    {
        pix = (PIXR *)DGPADT_LIGET( ln);
    	tokData->toks[i].image = pix;
DGP_ERR_CHECK_NO_MSG(
            w_pixrImageBytes( DGP_PIPWSTATE, pix, &numBytes))
        {
		    DGP_WARNF0( "unable to get pixr size.");
		}
	    _entry->memUsage += numBytes;
    }

    DgpAdt_liFree( images);
    images = DGP_NULLLIST;
    _entry->data = (void *)tokData;

    /* Tidy up the cache state. */
DGP_ERR_PASS_CLEANUP(
    suAddEntryInternal( xc, (suCacheEntry)_entry));

    /* Copy over the results. */
    *entry = (suCacheEntry)_entry;

    NORMAL_RETURN;

ERR_CLEANUP:
    if( images != DGP_NULLLIST) {
        DgpAdt_liFree( images);
    }
    if( tokData != (suColorTokens *)NULL) {
        if( tokData->toks != (color_token *)NULL) {
            for( i = 0; i < tokData->ntok; i++)
    	    {
                if( tokData->toks[i].image != NULL) {
                    DGP_FREE(tokData->toks[i].image);
                }
    	    }
            DGP_FREE( (void *)tokData->toks);
    	}
        DGP_FREE( (void *)tokData);
    	_entry->data = (void *)NULL;
    }

    if( _entry != (_suCacheEntry *)NULL) {
        suFreeEntryInternal( xc, (suCacheEntry)_entry, DGP_FALSE);
    }

    ERR_RETURN;
}


void
destroyColorTokens( suXiffCache xc, suCacheEntry entry) {
    _suXiffCache *_xc = (_suXiffCache *)xc;
    _suCacheEntry *_entry = (_suCacheEntry *)entry;
    Dgp_State *dgp_state = _xc->options.dgp_state;
    suColorTokens *tokData = (suColorTokens *)_entry->data;
    Dgp_Int32 i;

    DGP_VOID_PROLOGUE("destroyColorTokens()", DGP_PERRSTATE);
    SU_DEBUG1( "Freeing color tokens with Id: %d\n", tokData->uniqueId);

    if( tokData->toks != (color_token *)NULL) {
        for( i = 0; i < tokData->ntok; i++)
        {
            switch( tokData->toks[i].type)
            {
    	    case SU_COLOR_TOKEN_IMAGETYPE_PIXR:
                DGP_WARNF0("destroyColorTokens(), warning:  there shouldn't be any PIXR tokens around to free...");
                break;
            case SU_COLOR_TOKEN_IMAGETYPE_RGBIMAGE:
                if( tokData->toks[i].image != NULL) {
                    DGP_FREE(tokData->toks[i].image);
        		}
                break;
    	    }
    	}
        DGP_FREE( (void *)tokData->toks);
    }
    DGP_FREE( (void *)tokData);
    _entry->data = (void *)NULL;
}
/*#endif // 0 */
