
#include "core.h"

#include "sutokdict.h"
#include "sutokposblock.h"

#include "binary_tokens.h"
#include "binary_huffman_codes.h"
#include "token_positions.h"

#include "raw_position_extension.h"
#include "color_position_extension.h"
#include "color_tokens.h"
#include "color_dictionary.h"

#include "zephpage.h"
#include "zephdictionary.h"
#include "zephimage.h"

#include <limits.h>

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercache/periphery.c,v 1.25 2003/02/10 22:53:38 mkrocken Exp $");

/* Value Constants. */
#define AGE_WEIGHT      -10
#define USES_WEIGHT     100


/* Static functions. */
static Dgp_Int32
suGetValue( suCacheEntry entry, Dgp_Int32 currentPage);


/* Declared in periphery.h and used by core functions. */

/*
suEntryOps su_entry_ops[4] = {NULL, &binaryOps, &binaryHuffmanOps,
                                &tokenPositionOps};
*/

suEntryOps su_entry_ops[4] = { NULL, &binaryOps, 
&tokenPositionOps, &colorPositionExtensionOps};

suValuator su_valuator = suGetValue;


/* The Periphery Functions */

ErrCode suGetTokenDictionary( suXiffCache xc, Dgp_Int32 currentPage,
                Dgp_UInt32 libCount, Dgp_UInt32 *libIds,
                suCacheEntry *entry)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  zephDictionaryInfo zephDict = (zephDictionaryInfo)NULL;
  Dgp_UInt32 usage;

SU_PROLOGUE_XC( "suGetTokenDictionary()", _xc);
 
DGP_ERR_PASS_CLEANUP(
  zephOpenDictionary( _xc->zephDoc, libCount, libIds, &zephDict));
  SU_CLAIM_MEMORY( zephGetDictionaryMemoryUsage( zephDict)); 

DGP_ERR_PASS_CLEANUP(
  suGetTokenDictionaryH( xc, currentPage, zephDict, entry));

  usage = zephGetDictionaryMemoryUsage( zephDict); 
DGP_ERR_PASS_CLEANUP(
  zephCloseDictionary( &zephDict));
  SU_RELEASE_MEMORY( usage);

  NORMAL_RETURN;

ERR_CLEANUP:

  if( zephDict != (zephDictionaryInfo)NULL) {
    usage = zephGetDictionaryMemoryUsage( zephDict);
    zephCloseDictionary( &zephDict);
    SU_RELEASE_MEMORY_VOID( usage);
  }

  ERR_RETURN;
}

ErrCode suGetTokenDictionaryH(  suXiffCache xc, Dgp_Int32 currentPage,
                zephDictionaryInfo zephDict, suCacheEntry *entry)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)NULL;
  suBinaryHuffmanIdentifier *suId = (suBinaryHuffmanIdentifier *)NULL;

SU_PROLOGUE_XC( "suGetTokenDictionaryH()", _xc);
DGP_ASSERT( currentPage >= 0);
DGP_ASSERT( zephDict != (zephDictionaryInfo)NULL);
DGP_ASSERT( (zephDict->libIds != (Dgp_UInt32 *)NULL) 
               || (zephDict->libCount == 0));

  SU_CLAIM_MEMORY( sizeof( *suId)); 

  /* Create the identifier. */
DGP_CHECK_ALLOC_CLEANUP(
  suId = (suBinaryHuffmanIdentifier *)DGP_MALLOC( sizeof( *suId)));

  suId->libCount = zephDict->libCount;
  suId->libIds = zephDict->libIds;
  suId->zephDict = zephDict;

DGP_ERR_PASS_CLEANUP(
  suGetEntryInternal( xc, BINARY_HUFFMAN_CODES, suId, currentPage,
    		(suCacheEntry *)&_entry));

  DGP_FREE( (void *)suId);
  suId = (suBinaryHuffmanIdentifier *)NULL;
  SU_RELEASE_MEMORY( sizeof( *suId));

  _xc->lastUsedEntry = _entry;

  /* Finally, copy over the results. */
  *entry = (suCacheEntry)_entry;

  NORMAL_RETURN;

ERR_CLEANUP:
  if( suId != (suBinaryHuffmanIdentifier *)NULL) {
    DGP_FREE( (void *)suId);
    SU_RELEASE_MEMORY_VOID( sizeof( *suId));
  }
  ERR_RETURN;
}

ErrCode suExtractTokenDictionary( suCacheEntry entry,
                                  jb2Segment **segment) 
{
  _suCacheEntry *_entry = (_suCacheEntry *)entry;

SU_PROLOGUE_ENTRY( "suExtractTokenDictionary()", _entry);


 *segment = ((suBinaryTokens *)_entry->data)->segment;
                                  
 NORMAL_RETURN;
}

ErrCode
suGetTokenPositionBlock( suXiffCache xc, Dgp_Int32 currentPage,
                         Dgp_Int32 layer, Dgp_Int32 image, 
                         suCacheEntry *entry)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  zephPageInfo zephPage;
  zephImageInfo zephImage; 
  Dgp_UInt32 usage;

SU_PROLOGUE_XC( "suGetTokenPositionBlock()", _xc);

DGP_ERR_PASS_CLEANUP(
  zephOpenPage( _xc->zephDoc, (Dgp_UInt16)currentPage, &zephPage));
  SU_CLAIM_MEMORY( zephGetPageMemoryUsage( zephPage));

DGP_ERR_PASS_CLEANUP( 
  zephOpenImage( zephPage, layer, image, &zephImage));
  SU_CLAIM_MEMORY( zephGetImageMemoryUsage( zephImage));

DGP_ERR_PASS_CLEANUP(
  suGetTokenPositionBlockH( xc, currentPage, zephImage,
    		entry));

  usage = zephGetImageMemoryUsage( zephImage);
DGP_ERR_PASS(
  zephCloseImage( &zephImage));
  SU_RELEASE_MEMORY( usage); 

  usage = zephGetPageMemoryUsage( zephPage);
DGP_ERR_PASS(
  zephClosePage( &zephPage));
  SU_RELEASE_MEMORY( usage); 

  NORMAL_RETURN;

ERR_CLEANUP:
  if( zephImage != (zephImageInfo)NULL) {
    usage = zephGetImageMemoryUsage( zephImage);
    zephCloseImage( &zephImage);
    SU_RELEASE_MEMORY_VOID( usage);
  }
  if( zephPage != (zephPageInfo)NULL) {
    usage = zephGetPageMemoryUsage( zephPage);
    zephClosePage( &zephPage);
    SU_RELEASE_MEMORY_VOID( usage);
  }
  ERR_RETURN;
}

ErrCode suGetTokenPositionBlockH( suXiffCache xc, Dgp_Int32 currentPage,
		zephImageInfo zephImage,
		suCacheEntry *entry)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)NULL;
  suPositionIdentifier *posId = NULL;

SU_PROLOGUE_XC( "suGetTokenPositionBlockH()", _xc);
DGP_ASSERT( currentPage >= 0);

  SU_CLAIM_MEMORY( sizeof( *posId)); 

  /* Create the identifier. */
DGP_CHECK_ALLOC_CLEANUP(
  posId = (suPositionIdentifier *)DGP_MALLOC( sizeof( *posId)));

  posId->page = currentPage; 
  posId->layer = zephImage->layerNum;
  posId->image = zephImage->imageNum;
  /* We don't care, at this point, about any of the extensions. */
  posId->extDecoded = 0xFFFFFFFF;

  posId->zephImage = zephImage;

DGP_ERR_PASS_CLEANUP(
  suGetEntryInternal( xc, TOKEN_POSITIONS, posId, currentPage,
    		(suCacheEntry *)&_entry));

  DGP_FREE( (void *)posId);
  posId = (suPositionIdentifier *)NULL;
  SU_RELEASE_MEMORY( sizeof( *posId));

  _xc->lastUsedEntry = _entry;
  
  /* Finally, copy over the results. */
  *entry = (suCacheEntry)_entry;

  NORMAL_RETURN;
    
ERR_CLEANUP:
  if( posId != (suPositionIdentifier *)NULL) {
    DGP_FREE( (void *)posId);
    SU_RELEASE_MEMORY_VOID( sizeof( *posId));
  }
  ERR_RETURN;
}

ErrCode suExtractTokenPositionBlock( suCacheEntry entry, 
                                     suTokenPositionBlock *pPosBlock)
{
  _suCacheEntry *_entry = (_suCacheEntry *)entry;
  suTokenPositions *pos = (suTokenPositions *)NULL;

SU_PROLOGUE_ENTRY( "suExtractTokenData()", _entry);

  if( !isTokenPositionEntry( _entry)) {
    DGP_ERRF0( cErrInvalidParm,
               "can only extract position data from token position entry.");
  }

  pos = (suTokenPositions *)_entry->data;
  
  *pPosBlock = (suTokenPositionBlock)&pos->posBlock;

  NORMAL_RETURN;
}

/* 16.6.2000 / dhtesq. */
/* I found this part of supercache pre-processed out;  attempting to resurrect because */
/* I need the color stuff. */
/*#if 0 */

#if 0
/*
I changed the interface here to match what I found in the header file.
dhtesq.
*/
ErrCode 
suClaimTokenPositionBlock(suCacheEntry entry,
                          suTokenPositionBlock *pPosBlock)
/*
suClaimTokenPositionBlock( suCacheEntry entry, Dgp_Int32 *numBoxes,
		jb2SymbolInstance **boxes, Dgp_Int32 *numExts, Dgp_Int32 **extTags)
*/
{
    _suCacheEntry *_entry = (_suCacheEntry *)entry;
    _suXiffCache *_xc = (_suXiffCache *)_entry->xc;
    Dgp_Int32 _numBoxes;
    jb2SymbolInstance *_boxes;
    Dgp_Int32 _numExts;
    Dgp_Int32 *_extTags;

SU_PROLOGUE_XC( "suClaimColorPositionBlock()", _xc);

DGP_ERR_PASS_CLEANUP(
    suExtractTokenPositionBlock( entry, &_numBoxes, &_boxes, &_numExts,
    		&_extTags));

    if( _entry->refCount == 1) {
        suTokenPositions *pos;

	    /* No need to deep copy. */
        *numBoxes = _numBoxes;
        *boxes = _boxes;
	    *numExts = _numExts;
	    *extTags = _extTags;

        /* Here we actually destroy the entry since we can't give it to
	     * anyone else. */
	    _entry->refCount = 0;

	    /* But don't free the stuff we just gave away. */
	    pos = (suTokenPositions *)_entry->data;
	    pos->boxes = (jb2SymbolInstance *)NULL;
	    pos->extTags = (Dgp_Int32 *)NULL;

DGP_ERR_CHECK_NO_MSG(
            suRemoveEntryInternal( (suXiffCache)_xc, (suCacheEntry)_entry, NULL))
        {
            DGP_WARNF0( "unable to remove dirty cache entry. cache may be inconsistent.");
        }
    }
    else
    {
	    *boxes = (jb2SymbolInstance *)NULL;
	    *extTags = (Dgp_Int32 *)NULL;

        /* Copy the data. */
	    *numBoxes = _numBoxes;
DGP_CHECK_ALLOC_CLEANUP(
        *boxes = (jb2SymbolInstance *)DGP_MALLOC( _numBoxes * sizeof( jb2SymbolInstance)));
	    memcpy( (Dgp_Int8 *)*boxes, (Dgp_Int8 *)_boxes, (_numBoxes * sizeof( jb2SymbolInstance)));

	    *numExts = _numExts;
DGP_CHECK_ALLOC_CLEANUP(
        *extTags = (Dgp_Int32 *)DGP_MALLOC( _numExts * sizeof( Dgp_Int32)));
        memcpy( (Dgp_Int8 *)*extTags, (Dgp_Int8 *)_extTags, (_numExts * sizeof( Dgp_Int32)));

    	/* Release the current reference to the entry. */
    	suReleaseReference( entry);
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if( *boxes != (jb2SymbolInstance *)NULL) {
        DGP_FREE( (void *)*boxes);
    	*boxes = (jb2SymbolInstance *)NULL;
    }
    if( *extTags != (Dgp_Int32 *)NULL) {
        DGP_FREE( (void *)*extTags);
    	*extTags = (Dgp_Int32 *)NULL;
    }
    ERR_RETURN;
}


ErrCode suGetRawTokenPositionExtension( suXiffCache xc, Dgp_Int32 currentPage,
		Dgp_Int32 layer, Dgp_Int32 image, Dgp_Int32 extTag, suCacheEntry *entry)
{
SU_PROLOGUE_XC( "suGetRawTokenPositionExtension()", (_suXiffCache *)xc);

DGP_ERR_PASS(
    suGetRawTokenPositionExtensionH( xc, currentPage, layer, image,
    		(XIF_IMGHANDLE)NULL, 0, 0, extTag, entry));

    NORMAL_RETURN;
}

ErrCode suGetRawTokenPositionExtensionH( suXiffCache xc, Dgp_Int32 currentPage,
		Dgp_Int32 layer, Dgp_Int32 image, XIF_IMGHANDLE imgHandle,
		Dgp_UInt32 imWidth, Dgp_UInt32 imHeight,
		Dgp_Int32 extTag, suCacheEntry *entry)
{
    _suXiffCache *_xc = (_suXiffCache *)xc;
    _suCacheEntry *_entry = (_suCacheEntry *)NULL;
    suPositionExtensionIdentifier *extId;

SU_PROLOGUE_XC( "suGetRawTokenPositionExtension()", _xc);
DGP_ASSERT( currentPage >= 0);

    /* Create the identifier. */
DGP_CHECK_ALLOC_CLEANUP(
    extId = (suPositionExtensionIdentifier *)DGP_MALLOC(
    		sizeof( *extId)));

    extId->page = currentPage;
    extId->layer = layer;
    extId->image = image;
    extId->extTag = extTag;
    extId->imgHandle = imgHandle;
    extId->imWidth = imWidth;
    extId->imHeight = imHeight;

DGP_ERR_PASS_CLEANUP(
    suGetEntryInternal( xc, RAW_POSITION_EXTENSION, extId, currentPage,
    		(suCacheEntry *)&_entry));

    DGP_FREE( (void *)extId);
    extId = (suPositionExtensionIdentifier *)NULL;

    _xc->lastUsedEntry = _entry;

    /* Finally, copy over the results. */
    *entry = (suCacheEntry)_entry;

    NORMAL_RETURN;

ERR_CLEANUP:
    if( extId != (suPositionExtensionIdentifier *)NULL) {
        DGP_FREE( (void *)extId);
    }
    ERR_RETURN;
}


ErrCode suExtractRawTokenPositionExtension( suCacheEntry entry,
		Extension **ext)
{
    _suCacheEntry *_entry = (_suCacheEntry *)entry;
    suRawPositionExtension *extData = (suRawPositionExtension *)NULL;

SU_PROLOGUE_ENTRY( "suExtractRawTokenPositionExtension()", _entry);

    if( !isRawPositionExtensionEntry( _entry)) {
         DGP_ERRF0( cErrInvalidParm,
            "can only extract extension data from extension entry.");
    }

    extData = (suRawPositionExtension *)_entry->data;

    *ext = &extData->ext;

    NORMAL_RETURN;
}

ErrCode suClaimRawTokenPositionExtension( suCacheEntry entry,
		Extension **ext)
{
    _suCacheEntry *_entry = (_suCacheEntry *)entry;
    _suXiffCache *_xc = (_suXiffCache *)_entry->xc;
    suRawPositionExtension *extData = (suRawPositionExtension *)NULL;
    Extension *_ext = (Extension *)NULL;

SU_PROLOGUE_XC( "suClaimRawTokenPositionExtension()", _xc);
DGP_ASSERT( _entry != (_suCacheEntry *)NULL);

    if( !isRawPositionExtensionEntry( _entry)) {
         DGP_ERRF0( cErrInvalidParm,
            "can only extract extension data from extension entry.");
    }

    extData = (suRawPositionExtension *)_entry->data;

DGP_CHECK_ALLOC_CLEANUP(
    _ext = (Extension *)DGP_MALLOC( sizeof( *ext)));

    if( _entry->refCount == 1)
    {
        /* No need to deep copy. */
	    _ext->tag = extData->ext.tag;
	    _ext->length = extData->ext.length;
	    _ext->extension = extData->ext.extension;

        _entry->refCount = 0;

	    /* Remember not to free the data we are giving away. */
	    extData->ext.extension = (void *)NULL;

DGP_ERR_CHECK_NO_MSG(
            suRemoveEntryInternal( (suXiffCache)_xc, (suCacheEntry)_entry, NULL))
        {
            DGP_WARNF0( "unable to remove dirty cache entry. cache may be inconsistent.");
        }
    }
    else
    {
        _ext->tag = extData->ext.tag;
	    _ext->length = extData->ext.length;
DGP_CHECK_ALLOC_CLEANUP(
        _ext->extension = DGP_MALLOC( _ext->length));

        memcpy( (Dgp_Int8 *)_ext->extension, (Dgp_Int8 *)extData->ext.extension,
			_ext->length);

	    /* Release the current reference. */
	    suReleaseReference( (suCacheEntry)_entry);
    }

    *ext = _ext;
    NORMAL_RETURN;

ERR_CLEANUP:

    if( _ext != (Extension *)NULL) {
        if( _ext->extension != (void *)NULL) {
            DGP_FREE( _ext->extension);
	    }
        DGP_FREE( (void *)_ext);
    }
    ERR_RETURN;
}
#endif /* 0 */

ErrCode suGetColorTokenPositionExtension( suXiffCache xc, Dgp_Int32 currentPage,
		Dgp_Int32 layer, Dgp_Int32 image, suCacheEntry *entry)
{
SU_PROLOGUE_XC( "suGetColorTokenPositionExtension()", (_suXiffCache *)xc);

DGP_ERR_PASS(
    suGetColorTokenPositionExtensionH( xc, currentPage, layer, image,
    		(zephImageInfo)NULL, entry));

    NORMAL_RETURN;
}

ErrCode suGetColorTokenPositionExtensionH( suXiffCache xc, Dgp_Int32 currentPage,
		Dgp_Int32 layer, Dgp_Int32 image, zephImageInfo zephImage,
		suCacheEntry *entry)
{
    _suXiffCache *_xc = (_suXiffCache *)xc;
    _suCacheEntry *_entry = (_suCacheEntry *)NULL;
    suPositionExtensionIdentifier *extId;

SU_PROLOGUE_XC( "suGetColorTokenPositionExtensionH()", _xc);
DGP_ASSERT( currentPage >= 0);

    /* Create the identifier. */
DGP_CHECK_ALLOC_CLEANUP(
    extId = (suPositionExtensionIdentifier *)DGP_MALLOC(
    		sizeof( *extId)));

    extId->page = currentPage;
    extId->layer = layer;
    extId->image = image;
    extId->extTag = COLOR_EXT_TAG;

    extId->zephImage = zephImage;
    /*extId->imWidth = imWidth; */
    /*extId->imHeight = imHeight; */

DGP_ERR_PASS_CLEANUP(
    suGetEntryInternal( xc, COLOR_POSITION_EXTENSION, extId, currentPage,
    		(suCacheEntry *)&_entry));

    DGP_FREE( (void *)extId);
    extId = (suPositionExtensionIdentifier *)NULL;

    _xc->lastUsedEntry = _entry;

    /* Finally, copy over the results. */
    *entry = (suCacheEntry)_entry;

    NORMAL_RETURN;

ERR_CLEANUP:
    if( extId != (suPositionExtensionIdentifier *)NULL) {
        DGP_FREE( (void *)extId);
    }
    ERR_RETURN;
}


ErrCode suExtractColorTokenPositionExtension( suCacheEntry entry,
		color_tag **color_tags, colorDictInfo **color_dictinfo)
{
    _suCacheEntry *_entry = (_suCacheEntry *)entry;
    suColorPositionExtension *extData = (suColorPositionExtension *)NULL;

SU_PROLOGUE_ENTRY( "suExtractColorTokenPositionExtension()", _entry);

    if( !isColorPositionExtensionEntry( _entry)) {
         DGP_ERRF0( cErrInvalidParm,
            "can only extract extension data from extension entry.");
    }

    extData = (suColorPositionExtension *)_entry->data;

    *color_tags = extData->color_tags;
    *color_dictinfo = extData->color_dictinfo;

    NORMAL_RETURN;
}

ErrCode suClaimColorTokenPositionExtension( suCacheEntry entry,
		color_tag **color_tags, colorDictInfo **color_dictinfo)
{
    _suCacheEntry *_entry = (_suCacheEntry *)entry;
    _suXiffCache *_xc = (_suXiffCache *)_entry->xc;
    suColorPositionExtension *extData;
    color_tag *_color_tags;
    colorDictInfo *_color_dictinfo;

SU_PROLOGUE_XC( "suClaimColorTokenPositionExtension()", _xc);

DGP_ERR_PASS_CLEANUP(
    suExtractColorTokenPositionExtension( entry, &_color_tags,
    		&_color_dictinfo));

    extData = (suColorPositionExtension *)_entry->data;

    if( _entry->refCount == 1)
    {
        /* No need to deep copy. */
	    *color_tags = _color_tags;
	    *color_dictinfo = _color_dictinfo;

        _entry->refCount = 0;

	    /* Remember not to free the data we are giving away. */
	    extData->color_tags = (color_tag *)NULL;
	    extData->color_dictinfo = (colorDictInfo *)NULL;

DGP_ERR_CHECK_NO_MSG(
            suRemoveEntryInternal( (suXiffCache)_xc, (suCacheEntry)_entry, NULL))
        {
            DGP_WARNF0( "unable to remove dirty cache entry. cache may be inconsistent.");
        }
    }
    else
    {
DGP_CHECK_ALLOC_CLEANUP(
        *color_tags = (color_tag *)DGP_MALLOC( extData->ntag
			* sizeof( color_tag)));
        memcpy( (Dgp_Int8 *)*color_tags, (Dgp_Int8 *)_color_tags,
			extData->ntag * sizeof( color_tag));

DGP_CHECK_ALLOC_CLEANUP(
        *color_dictinfo = (colorDictInfo *)DGP_MALLOC(
			sizeof( **color_dictinfo)));

        (*color_dictinfo)->num_dict_ids = _color_dictinfo->num_dict_ids;
DGP_CHECK_ALLOC_CLEANUP(
        (*color_dictinfo)->dict_ids = (Dgp_UInt32 *)DGP_MALLOC(
			_color_dictinfo->num_dict_ids * sizeof( Dgp_UInt32)));

	    memcpy( (Dgp_Int8 *)(*color_dictinfo)->dict_ids,
	    	(Dgp_Int8 *)_color_dictinfo->dict_ids,
	    	_color_dictinfo->num_dict_ids * sizeof( Dgp_UInt32));

	    /* Release the current reference. */
	    suReleaseReference( (suCacheEntry)_entry);
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if( *color_tags != (color_tag *)NULL) {
        DGP_FREE( (void *)*color_tags);
	    *color_tags = (color_tag *)NULL;
    }
    if( *color_dictinfo != (colorDictInfo *)NULL) {
        if( (*color_dictinfo)->dict_ids != (Dgp_UInt32 *)NULL) {
            DGP_FREE( (void *)(*color_dictinfo)->dict_ids);
	    }
        DGP_FREE( (void *)*color_dictinfo);
	    *color_dictinfo = (colorDictInfo *)NULL;
    }
    ERR_RETURN;
}

ErrCode suGetColorDictionaryUnion( suXiffCache xc, Dgp_Int32 currentPage,
		colorDictInfo *color_dictinfo, suCacheEntry *entry)
{
    _suXiffCache *_xc = (_suXiffCache *)xc;
    _suCacheEntry *_entry = (_suCacheEntry *)NULL;

SU_PROLOGUE_XC( "suGetColorDictionaryUnion()", _xc);
DGP_ASSERT( currentPage >= 0);

DGP_ERR_PASS_CLEANUP(
    suGetEntryInternal( xc, COLOR_DICTIONARY, (void *)color_dictinfo,
    		currentPage, (suCacheEntry *)&_entry));

    _xc->lastUsedEntry = _entry;

    /* Finally, copy over the results. */
    *entry = (suCacheEntry)_entry;

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}


ErrCode suExtractColorDictionaryUnion( suCacheEntry entry, color_token ***toks,
		Dgp_Int32 *numToks)
{
    _suCacheEntry  *_entry = (_suCacheEntry *)entry;
    suColorDictionary *dictData = (suColorDictionary *)NULL;

SU_PROLOGUE_ENTRY( "suExtractColorDictionaryUnion()", _entry);

    if( !isColorDictionaryEntry( _entry)) {
         DGP_ERRF0( cErrInvalidParm,
            "can only extract color dictionary data from color dictionary entry.");
    }

    dictData = (suColorDictionary *)_entry->data;

    *toks = dictData->toks;
    *numToks = dictData->ntok;

    NORMAL_RETURN;
}
/*#endif // 0 */

static Dgp_Int32 suGetValue( suCacheEntry entry, Dgp_Int32 currentPage) {
    _suCacheEntry *_entry = (_suCacheEntry *)entry;

    if( _entry->type == BINARY_TOKENS)
    {
        /* Don't ever throw out the first set of tokens. */
        if( ((suBinaryTokens *)_entry->data)->identifier.dictId == 0)
	{
	    return INT_MAX;
	}
    }

    return (_entry->age * AGE_WEIGHT) + ((_entry->uses - 1) * _entry->uses *
    	MIN( ABS( _entry->lastPage - currentPage), 1)) * USES_WEIGHT;
}


