/*

  19.6.2000 / dhtesq.

  Reconditioning color parts of libsupercache following WB's revision.
  Apparently this file is not required for the black-and-white case.

*/

#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif


#include "libsupercache.h"
#include "core.h"
#include "raw_position_extension.h"

#include "list.h"		/* DgpAdt_List */
#include "util.h"		/* Dgp_Bool */
#include "extension_tags.h"


_suEntryOps rawPositionExtensionOps = { compareRawPositionExtension,
					getRawPositionExtension, 
					destroyRawPositionExtension,
					NULL};

Dgp_Bool
compareRawPositionExtension( void *identifier, suCacheEntry entry) {
    suRawPositionExtension *extData =
    	(suRawPositionExtension *)((_suCacheEntry *)entry)->data;
    suPositionExtensionIdentifier *extId =
        (suPositionExtensionIdentifier *)identifier;

    SU_DEBUG2( "compareRawPositionExtension(): %d ?= %d\n",
    			extId->page, extData->identifier.page);
    SU_DEBUG2( "                                  %d ?= %d\n",
    			extId->layer, extData->identifier.layer);
    SU_DEBUG2( "                                  %d ?= %d\n",
    			extId->image, extData->identifier.image);
    SU_DEBUG2( "                                  %d ?= %d\n",
    			extId->extTag, extData->identifier.extTag);

    if( (extData->identifier.page == extId->page) &&
        	(extData->identifier.layer == extId->layer) &&
		(extData->identifier.image == extId->image) &&
    		(extData->identifier.extTag == extId->extTag)) {
	    return DGP_TRUE;
    } else {
        return DGP_FALSE;
    }
}


ErrCode
getRawPositionExtension( suXiffCache xc, void *identifier, Dgp_Int32 currentPage,
		suCacheEntry *entry) {
    _suXiffCache *_xc = (_suXiffCache *)xc;
    _suCacheEntry *_entry = (_suCacheEntry *)NULL;
    suPositionExtensionIdentifier *extId = (suPositionExtensionIdentifier *)identifier;
    suRawPositionExtension *extData = (suRawPositionExtension *)NULL;
    _suCacheEntry *_posEntry;
    suPositionIdentifier *posId = (suPositionIdentifier *)NULL;
    suTokenPositions *posData;
    Dgp_Int32 which_ext;

SU_PROLOGUE_XC( "getRawPositionExtension()", _xc);

    SU_DEBUG( "Entering getRawPositionExtension()...\n");
    SU_DEBUG2( "Request for raw extension on page: %d  with tag: %d\n",
    		extId->page, extId->extTag);

DGP_ERR_PASS_CLEANUP(
    suInitEntryInternal( xc, RAW_POSITION_EXTENSION,
            (suCacheEntry *)&_entry));

DGP_CHECK_ALLOC_CLEANUP(
    extData = (suRawPositionExtension *)DGP_MALLOC( sizeof( *extData)));

    /* Predictable state... */
    extData->identifier.page = extId->page;
    extData->identifier.layer = extId->layer;
    extData->identifier.image = extId->image;
    extData->identifier.extTag = extId->extTag;

    extData->ext.extension = (void *)NULL;

    /* Get the position block from the cache.  */
DGP_CHECK_ALLOC_CLEANUP(
    posId = (suPositionIdentifier *)DGP_MALLOC( sizeof( *posId)));

    posId->page = extId->page;
    posId->layer = extId->layer;
    posId->image = extId->image;
    posId->extDecoded = ~( 1 << (extId->extTag));
    /* imgHandle field has disappeared;  try zephImage.  dhtesq. */
    posId->zephImage = extId->zephImage;
    /*posId->imgHandle = extId->imgHandle; */

    /* These too.. */
    /*posId->imWidth = extId->imWidth; */
    /*posId->imHeight = extId->imHeight; */

DGP_ERR_PASS_CLEANUP(
    suGetEntryInternal( xc, TOKEN_POSITIONS, (void *)posId, currentPage,
    		(suCacheEntry *)&_posEntry));

    DGP_FREE( (void *)posId);
    posId = (suPositionIdentifier *)NULL;
    posData = (suTokenPositions *)_posEntry->data;

    /* 'next' field has disappeared... try posBlock->count. / dhtesq. */
    /* Find the extension. */
    for(which_ext = 0; which_ext <= posData->posBlock.posCount /* posData->next */; which_ext++)
    {
        if(which_ext == posData->posBlock.posCount /*posData->next */) {
	        /* There is no extension with desired tag!!! */
            DGP_ERR_CLEANUPF0( cErrInvalidParm, "getRawPositionExtension() / no extension with given tag!");

	    } else {
	        if( posData->exts[which_ext].tag == extId->extTag) {
	            break;
            }
	    }
    }

    /* This extension should really still be there. */
DGP_ASSERT( posData->exts[ which_ext].extension != (void *)NULL);

    /* Move the data (...well copy the pointer anyway.) */
    extData->ext.tag = extId->extTag;
    extData->ext.length = posData->exts[ which_ext].length;
    extData->ext.extension = posData->exts[ which_ext].extension;

    /* Clean up and make sure to release the reference to the position block. */
    posData->identifier.extDecoded |= (1 << (extId->extTag));
    posData->exts[ which_ext].extension = (void *)NULL;
    _posEntry->memUsage -= extData->ext.length;
    _xc->currentUsage -= extData->ext.length;
    suReleaseReference( (suCacheEntry)_posEntry);

    _entry->memUsage = extData->ext.length;
    _entry->memUsage += sizeof( *extData);
    _entry->memUsage += sizeof( *_entry);
    _entry->data = (void *)extData;

    /* Tidy up the cache state. */
DGP_ERR_PASS_CLEANUP(
    suAddEntryInternal( xc, (suCacheEntry)_entry));

    /* Copy over the results. */
    *entry = (suCacheEntry)_entry;

    NORMAL_RETURN;

ERR_CLEANUP:
    if( posId == (suPositionIdentifier *)NULL) {
        DGP_FREE( (void *)posId);
    }

    if( extData != (suRawPositionExtension *)NULL) {
      	if( extData->ext.extension != (void *)NULL) {
            DGP_FREE( extData->ext.extension);
      	}
        DGP_FREE( (void *)extData);
    	_entry->data = (void *)NULL;
    }

    if( _entry != (_suCacheEntry *)NULL) {
        /* This now takes an additional argument of Dgp_Bool inCache. */
        /* The entry will only be destroyed if this is true...?  Looks */
        /* like DGP_FALSE is used in cleanup cases. / dhtesq. */
        /*suFreeEntryInternal( xc, (suCacheEntry)_entry); */
        suFreeEntryInternal( xc, (suCacheEntry)_entry, DGP_FALSE);
    }

    ERR_RETURN;
}

void
destroyRawPositionExtension( suXiffCache xc, suCacheEntry entry) {
    _suXiffCache *_xc = (_suXiffCache *)xc;
    _suCacheEntry *_entry = (_suCacheEntry *)entry;
    Dgp_State *dgp_state = _xc->options.dgp_state;
    suRawPositionExtension *extData =
    	(suRawPositionExtension *)_entry->data;

    if( extData != (suRawPositionExtension *)NULL) {
        SU_DEBUG2( "Freeing raw extension for page: %d  with tag: %d\n",
        		extData->identifier.page, extData->identifier.extTag);

        if( extData->ext.extension != (void *)NULL) {
            DGP_FREE( extData->ext.extension);
        }
        DGP_FREE( (void *)extData);
        _entry->data = (void *)NULL;
    }
}
