/* $Header: /project/icons/cvsroot/digipaper/src/libsupercache/sutokposblock.h,v 1.4 2001/01/23 22:48:12 mkrocken Exp $ */

#ifndef SUTOKPOSBLOCK_H
#define SUTOKPOSBLOCK_H

#include "sucache_t.h"
#include "sutokposblock_t.h"

#include "zephimage_t.h"
#include "colorpos.h"
#include "colortoken_t.h"

#ifdef __cplusplus
extern "C"
{
#endif

/*
 * Given a set of unique identifiers, fetch the positions block that
 * corresponds with that image.
 *
 * It is NOT necessary to have called suGetTokenDictionary
 * before calling this function. If you do, however, a cached
 * copy of the Huffman codes is used.
 *
 * Preconditions:
 *  - "xc" is a valid suXiffCache.
 *  - "currentPage" is non-negative.
 *  - "layer" and "image" refer to a valid image. (remember that in
 *      libzephyr the first layer is "0" (the background) and the 
 *      first image is also "0". this applies here as well.)
 *
 * Postconditions:
 *  - on success, "entry" is a pointer to a cache entry containing
 *       the requested position block.
 */
extern ErrCode
suGetTokenPositionBlock( suXiffCache xc, Dgp_Int32 currentPage,
                Dgp_Int32 layer, Dgp_Int32 image, suCacheEntry *pEntry);

/*
 * Same as above, but use the given open image handle
 * rather than opening a new one.
 *
 */
extern ErrCode
suGetTokenPositionBlockH( suXiffCache xc, Dgp_Int32 currentPage,
		zephImageInfo zephImage, suCacheEntry *pEntry);

/*
 * Returns the token positions (with appropriate tokens hanging off
 * of them) for an entry that contains a token position block.
 *
 * Preconditions:
 *  - "entry" was the result of a successful call to
 *      suGetTokenPositionBlock(Ex)(H).
 *
 * Postconditions:
 *  - everything returned by the function continues to be
 *      under the ownership of the cache.
 *  - "boxes" is a pointer to an array of jb2SymbolInstance of size "numBoxes"
 *  - "extTags" is a pointer to an array of Dgp_Int32 of size "numExts"
 */
extern ErrCode
suExtractTokenPositionBlock( suCacheEntry entry,
                 suTokenPositionBlock *pos);


/*
 * As Extract above, but the returned arrays are no longer owned
 * by the cache (ownership is passed to the caller). An implicit
 * call to ReleaseReference is made, as well, so there is no need
 * to Release the entry afterwards.
 *
 * Note that a you may call Extract and then Claim on a single
 * entry (in that order only). Be careful though, the result of
 * Claim may be different than that of Extract- if someone else
 * has an outstanding reference to the entry, then you will
 * receive a copy of the position and extension data, rather than
 * the actual data stored in the cache.
 *
 * Postconditions:
 *  - "entry" is invalid.
 *  - "boxes" is a pointer to an array of jb2SymbolInstance of size "numBoxes"
 *  - "extTags" is a pointer to an array of Dgp_Int32 of size "numExts"
 */
extern ErrCode
suClaimTokenPositionBlock( suCacheEntry entry, 
                 suTokenPositionBlock *pPosBlock);

/*
Added color variants...
dhtesq.
*/
extern ErrCode 
suGetColorTokenPositionExtension( suXiffCache xc, Dgp_Int32 currentPage,
		Dgp_Int32 layer, Dgp_Int32 image, suCacheEntry *entry);

extern ErrCode 
suGetColorTokenPositionExtensionH( suXiffCache xc, Dgp_Int32 currentPage,
		Dgp_Int32 layer, Dgp_Int32 image, zephImageInfo zephImage,
		suCacheEntry *entry);

extern ErrCode 
suExtractColorTokenPositionExtension( suCacheEntry entry,
		color_tag **color_tags, colorDictInfo **color_dictinfo);

extern ErrCode 
suClaimColorTokenPositionExtension( suCacheEntry entry,
		color_tag **color_tags, colorDictInfo **color_dictinfo);

extern ErrCode 
suClaimColorTokenPositionExtension( suCacheEntry entry,
		color_tag **color_tags, colorDictInfo **color_dictinfo);

extern ErrCode 
suGetColorDictionaryUnion( suXiffCache xc, Dgp_Int32 currentPage,
		colorDictInfo *color_dictinfo, suCacheEntry *entry);

extern ErrCode 
suExtractColorDictionaryUnion( suCacheEntry entry, color_token ***toks,
		Dgp_Int32 *numToks);




#ifdef __cplusplus
}
#endif

#endif /* SU_TOKPOSBLOCK_H */
