

#include "core.h"
#include "sucache.h"                /* suReleaseReference */
#include "sutokdict.h"              /* suExtractTokenDictionary */

#include "binary_tokens.h"
#include "token_positions.h"
#include "zephdictionary.h"
#include "zephimage.h"
#include "zephtokenize.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercache/token_positions.c,v 1.33 2003/02/10 22:53:38 mkrocken Exp $");


_suEntryOps tokenPositionOps = { compareTokenPositions, 
                                 getTokenPositions,
                                 destroyTokenPositions,
                                 calculateTokenPositions};

Dgp_Bool
compareTokenPositions( void *identifier, suCacheEntry entry) 
{
  suTokenPositions *posData
    = (suTokenPositions *)((_suCacheEntry *)entry)->data;
  suPositionIdentifier *posId = (suPositionIdentifier *)identifier;

  SU_DEBUG2( "compareTokenPositions(): %d ?= %d\n", posId->page, posData->identifier.page);
  SU_DEBUG2( "                         %d ?= %d\n", posId->layer, posData->identifier.layer);
  SU_DEBUG2( "                         %d ?= %d\n", posId->image, posData->identifier.image);

  if( (posData->identifier.page == posId->page) &&
        (posData->identifier.layer == posId->layer) &&
      (posData->identifier.image == posId->image)) 
    {

      SU_DEBUG2( "\t\t\t %X ?? %X\n",
               posId->extDecoded, posData->identifier.extDecoded);

      if( (posId->extDecoded | posData->identifier.extDecoded) 
          == posId->extDecoded) 
	{

	  return DGP_TRUE;
        }
      else 
	{
          return DGP_FALSE;
        }
    }
  else
    {
      return DGP_FALSE;
    }

}


ErrCode
getTokenPositions( suXiffCache xc, void *identifier, Dgp_Int32 currentPage, 
                  suCacheEntry *entry) 
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)NULL,
    *_dictEntry = (_suCacheEntry *)NULL;
      /**_tokEntry = (_suCacheEntry *)NULL; */
  
  suPositionIdentifier *posId = (suPositionIdentifier *)identifier;
  suTokenPositions *pos = (suTokenPositions *)NULL;
  suBinaryTokensIdentifier binTokId;
  zephDictionaryState zephState;
  
  /*DgpAdt_ListNode ln,notused; */
  Dgp_UInt32 usage = 0, i;
  Dgp_Bool diffClaimSucceeded = DGP_FALSE;
  DgpAdt_List seg_list;
  DgpAdt_ListNode last_original_node = NULL, ln;
  jb2Segment *segment;

SU_PROLOGUE_XC( "getTokenPositions()", _xc);

 SU_DEBUG( "Entering getTokenPositions()...\n");

DGP_ERR_PASS_CLEANUP(
  suInitEntryInternal( xc, TOKEN_POSITIONS, (suCacheEntry *)&_entry ));

  SU_CLAIM_MEMORY( sizeof( *pos));
DGP_CHECK_ALLOC_CLEANUP(
  pos = (suTokenPositions *)DGP_MALLOC( sizeof( *pos)));

  pos->identifier.page = posId->page;
  pos->identifier.layer = posId->layer;
  pos->identifier.image = posId->image;

  /* Set the extDecoded bits to 0. Because right now, none of the
   * extensions have been decoded. */
  pos->identifier.extDecoded = POS_EXTFLAG_NONE;
  
  /* Predictable state... */
  pos->posBlock.pos = (jb2SymbolInstance *)NULL;
  pos->posBlock.posCount = 0;
  pos->posBlock.colorExt = DGP_FALSE;
DGP_ERR_PASS_CLEANUP(
  DgpAdt_liNew (&pos->posBlock.segment_list, dgp_state));
  pos->exts = (Extension *)NULL;

  SU_DEBUG1( "Request for position block on page: %d\n", posId->page);

  /* load all dictionaries into a list of JBIG2 segments */
  binTokId.imageInfo = posId->zephImage;
DGP_ERR_PASS_CLEANUP(
  DgpAdt_liNew (&seg_list, dgp_state));
  for (i = 0; i < posId->zephImage->libCount; i++)
    {
      binTokId.dictId = posId->zephImage->libId[i];

DGP_ERR_PASS_CLEANUP(
       suGetEntryInternal( xc, BINARY_TOKENS, 
			   (void *)&binTokId,
			   currentPage, (suCacheEntry *)&_dictEntry));

DGP_ERR_PASS_CLEANUP(
       suExtractTokenDictionary( (suCacheEntry)_dictEntry, &segment));

DGP_ERR_PASS_CLEANUP(
       DgpAdt_liAdd (seg_list, segment, &last_original_node));

/* make a note that we are referencing this dictionary */
DGP_ERR_PASS_CLEANUP(
       DgpAdt_liAdd (_entry->relations, (void*) _dictEntry, NULL));

    }

  /* read and decode the position block */
  zephState = NULL;
DGP_ERR_PASS_CLEANUP(
  zephReadPositionBlock( posId->zephImage, seg_list, 
                         &pos->posBlock.pos, &pos->posBlock.posCount, 
                         &pos->exts, &usage));

  /* check if we got color extensions */
  if (pos->exts)
  {
    pos->posBlock.colorExt = DGP_TRUE;
  }

  /* put all segments of the position block into a list in the cache structure */
  if (last_original_node)
  {
	  /* we have dictionaries */
	  ln = last_original_node->next;
	  while (ln != DGP_NULLLISTNODE)
	  {
		  segment = (jb2Segment*) ln->userdata;
		  DGP_ERR_PASS_CLEANUP(
			  DgpAdt_liAdd (pos->posBlock.segment_list, segment, NULL));
		  ln = ln->next;
	  }
  }

  /* free the dictionary segment list */
DGP_ERR_PASS_CLEANUP(
    DgpAdt_liFree (seg_list));


  SU_CLAIM_MEMORY(usage)

  /* Use this a flag to remember that the claim/release of the difference
   * between estUsage and usage was successful. */
  diffClaimSucceeded = DGP_TRUE;

  _entry->memUsage += usage;

  /* Make sure to set the usage before calling suAddEntry. */
  _entry->memUsage += sizeof( *pos);
  _entry->data = (void *)pos;
  
  /* Tidy up the cache state. */
DGP_ERR_PASS_CLEANUP(
  suAddEntryInternal( xc, (suCacheEntry)_entry));

  /* And copy back the result. */
  *entry = (suCacheEntry)_entry;

  NORMAL_RETURN;

ERR_CLEANUP:
  
  /* We know that this entry is not in the cache. */
  if( _entry != (_suCacheEntry *)NULL) {
    suFreeEntryInternal( xc, (suCacheEntry)_entry, DGP_FALSE);
  }
  
  ERR_RETURN;
}


void
destroyTokenPositions( suXiffCache xc, suCacheEntry entry) 
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  _suCacheEntry *_entry = (_suCacheEntry *)entry, *_relEntry;
  Dgp_State *dgp_state = _xc->options.dgp_state;
  suTokenPositions *pos = (suTokenPositions *)_entry->data;
  DgpAdt_ListNode relLn, ln;
  /*Dgp_Int32 i, end; */
  jb2Segment *curseg;

DGP_VOID_PROLOGUE( "destroyTokenPositions()", DGP_PERRSTATE);

 if( pos != (suTokenPositions *)NULL) {

   SU_DEBUG1( "Freeing position block for page: %d\n", pos->identifier.page);

   /* First go through each of the relations and references to
    * the current entry. We assume that all these entries are
    * binary token entries. (Only make this pass during normal
    * cleanup, not cache destruction.) */
   if( _entry->uses >= 0) {
     for( relLn = DGPADT_LIFIRST( _entry->relations); relLn != DGP_NULLLISTNODE;
          DgpAdt_liRemAndNext( relLn, &relLn)) {
       
       _relEntry = (_suCacheEntry *)DGPADT_LIGET( relLn);
       if( !isBinaryTokenEntry( _relEntry)) {
         DGP_WARNF0( "relative entry is not a binary token entry!");
       }
              
       /* Decrement the other entry's reference count. */
       suReleaseReference( (suCacheEntry)_relEntry);
     }
   }
   
   /* Now free the memory associated with this entry. */
   DGPADT_FOREACH (ln, pos->posBlock.segment_list) { 
	  curseg = (jb2Segment *)DGPADT_LIGET(ln); 
    jb2FreeSegmentInfo(curseg, dgp_state); 
    DGP_FREE ((void *)curseg); 
   } 
   DgpAdt_liFree(pos->posBlock.segment_list);

   /* this is only part of the memory we had. FIX ME */
   SU_RELEASE_MEMORY_VOID( pos->posBlock.posCount * sizeof( jb2SymbolInstance));
   _entry->memUsage -= pos->posBlock.posCount * sizeof( jb2SymbolInstance);
   
   if(pos->exts != (Extension *)NULL && pos->posBlock.colorExt) {
	   DGP_FREE( (void *)pos->exts->extension);
	   DGP_FREE(pos->exts);
	   /* 26.6.2000 / dhtesq.
	   In putting color support back into the decoder, I'm finding extra bytes left
	   at cleanup, meaning either 1) I've overreported my usage to supercache, or 2)
	   I'm not freeing enough at cleanup.
	   Try adding the extensions and the extension tags to the free agenda...
	   */
	   SU_RELEASE_MEMORY_VOID(sizeof(Extension));
	   _entry->memUsage -= sizeof(Extension);
   }

   DGP_FREE( (void *)pos);
   SU_RELEASE_MEMORY_VOID( sizeof( *pos));
   _entry->memUsage -= sizeof( *pos);
   _entry->data = (void *)NULL;
 }
 
 VOID_RETURN;
}

void
calculateTokenPositions( suXiffCache xc, suCacheEntry entry)
{
  _suXiffCache *_xc = (_suXiffCache *)xc;
  Dgp_State *dgp_state = _xc->options.dgp_state;
  _suCacheEntry *_entry = (_suCacheEntry *)entry, *_relEntry;
  DgpAdt_ListNode ln;
  
DGP_VOID_PROLOGUE( "calculateTokenPositions()", DGP_PERRSTATE);

  /* There is no data that can change size in a position block, but the 
   * bitmaps in the token structures can change. So call calculate on each
   * of the dictionaries of binary tokens that this block references. */

  DGPADT_FOREACH( ln, _entry->relations) {
    _relEntry = (_suCacheEntry *)DGPADT_LIGET( ln);

    if( !isBinaryTokenEntry( _relEntry)) {
      DGP_WARNF0( "relative entry is not a binary token entry!");
    }

    suRecalculateUsageInternal( xc, (suCacheEntry)_relEntry);
  }
}
