/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 1 2000
**
**  Filename: 
**
**  Description of what's implemented in this file:
**    binary cc finding function
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/cc2box.c,v 1.11 2003/01/29 22:47:02 mkrocken Exp $
 *
 */

 /* cc2box.c
   -- original GetBoxes() by ???; modified by Aaron Stump, summer '96 and
      summer '97.
   -- GetColorBoxes() by Aaron Stump, summer '96; modified by Scott
      Cytacki, summer '97

   This file contains a function, GetBoxes(), for calling the IPCore
   CC-finder on a monochrome image, and a function, GetColorBoxes(),
   for calling Scott Cytacki's color CC finder on a qdr file.  

  pff '98: modified to build new type of color extension to position block
*/


#include "types.pub"
typedef struct StrelInfo *StrelInfoPtr;
#define _STREL_INFO_PTR_DEFINED_
#include "memory.pub"
#include "shros.pub"
#include "ipstate.pub"
#define _IPWSTATE_DEFINED_

#include "cc2box.h"
#include "colorCC.h"
#include "list.h"
#include "run2bitmap.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include "uds.h"
#include <color-extension.h>
#include "simple_utils.h"
#include "ipserrs.pub"
#include "shrutils.pub"
#include "iaerror.pub"
#include "shrcc.pub"
#include "dgp_utils.h"
#include "tokenCC.h"
#include "PsgFix.h"


DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercc/cc2box.c,v 1.11 2003/01/29 22:47:02 mkrocken Exp $");

DgpImage_RGB DgpCC_DefaultBgColor = {255, 255, 255};

#define Bit(b)         (1 << (b))
#define BitOn(v, b)    ((v) |= (0x80000000 >> (b)))

#define CC_SEC_MALLOC(_ppcc) \
  DGP_ERR_PASS_CLEANUP(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, DGPCC_CC_POOL_ID, (void **)(_ppcc)))
#define CC_FREE(_pcc) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, DGPCC_CC_POOL_ID, (void *)(_pcc)))

/*
 * This is the callback function for the ipshared function 
 * i_findConnectedComponentRuns.
 *
 * For big CCs, this function converts the runs to another flavor of runs,
 * and throws a descriptor onto a list of outliers (inside (void *)data).
 *
 * For CCs small enough to tokenize, it throws them straight at the tokenizer
 * (libtok-lossless lite), which gives it right back a token number.  Then
 * a descriptor smaller than the one required for a CC is thrown onto a 
 * different list (again, inside (void *)data).
 */
static Int32 
MakeModelsFromRuns(IPState *ipstate, 
                   void *data, 
                   Int32 x, 
                   Int32 y,
                   Int32 w, 
                   Int32 h, 
                   Int32 m, 
                   Int32 nRuns,
                   TCScanLineRun *pRuns)
{
  LlCCData *ccData = (LlCCData *)data;
  Dgp_State *dgp_state = ccData->dgp_state;
  DGP_PROLOGUE("MakeModelsFromRuns()", DGP_PERRSTATE);
  TCScanLineRun *pRunsIndex;
  DgpCC_PosInfo *pInfo;
  DgpCC_PosInfoBlock *pBlock = ccData->pBlock;
  DgpCC_CC *newcc = (DgpCC_CC *)NULL;
  run_elt *tmpruns = (run_elt *)NULL;


  (void)m;
  (void)ipstate;

  if ((w <= ccData->maxWidth) && (h <= ccData->maxHeight)) {

    /* This is something we can tokenize */
    tokpos *newpos;
    LlRunToken rTok;
    run_elt *run;

    rTok.width = (Dgp_Int16)w;
    rTok.height = (Dgp_Int16)h;

    if (ccData->num_boxes >= ccData->num_boxes_alloced) {
      ccData->num_boxes_alloced += ccData->num_boxes_alloced / 2;
      DGP_SEC_REALLOC(ccData->boxes, (tokpos *), (sizeof(tokpos) * ccData->num_boxes_alloced));
    }
    newpos = &(ccData->boxes[ccData->num_boxes]);

    /*
     * Where is this box? Not where it ought to be... there's a magic offset
     * of 32 because we're dealing with an ipshared routine here and sometimes
     * ipshared thinks there's a frame of 32 pixels at the top and left. Yay.
     */
    newpos->xpos = x - 32;
    newpos->ypos = y - 32 + h - 1; /* we're going to shift this box's coordinates
              from specifying the UL corner to the LL
              corner at this point in the execution */
    newpos->extdata = NULL;

    DGP_SEC_MALLOC(tmpruns, (run_elt *), (2 * sizeof(run_elt) * nRuns));
    rTok.runs = tmpruns;
    rTok.nruns = nRuns;

    /*
     * Read out the runs. They were generated bottom-to-top, so reverse
     * them.
     */
    for (run = rTok.runs + 2 * (nRuns - 1), pRunsIndex = pRuns;
         pRunsIndex;
         run -= 2, pRunsIndex = pRunsIndex->pNext) {
      run[0] = (Int16)(pRunsIndex->runLeft - x); /* normalized x-location */
      run[1] = (Int16)(pRunsIndex->runRight - pRunsIndex->runLeft + 1); /* length */
    }

    if(pBlock->num >= pBlock->numAlloced){
      pBlock->numAlloced = pBlock->numAlloced * 3 / 2;
      DGP_SEC_REALLOC(pBlock->poss, (DgpCC_PosInfo *), (pBlock->numAlloced * sizeof(DgpCC_PosInfo)));
    }
   
    pInfo = &(pBlock->poss[pBlock->num]);
    pInfo->page = ccData->pageIndex;
    pInfo->posOffset = ccData->num_boxes;


    rTok.userdata = (void *)pBlock->num;

    DGP_ERR_PASS_CLEANUP(DgpCC_tokenAddCC(dgp_state, ccData->tokstate, ccData->tBlock, &rTok, &(newpos->tok)));
    tmpruns = NULL;

    pBlock->num++;
    ccData->num_boxes++;

    ccData->count++;

  } else {
    
    /*
     * If this model is "too big" then don't build the bitmap -- keep the runs. 
     * Make one big area for them - more efficient. Just store 2 Int16's for 
     * each run - left and right.
     */
    Int16 *run;

    CC_SEC_MALLOC(&newcc);
    newcc->runs = NULL;

    newcc->bitmap.width = (Dgp_UInt16)w;
    newcc->bitmap.height = (Dgp_UInt16)h;
    newcc->bitmap.longsperline =
      (newcc->bitmap.width + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
    newcc->bitmap.bits = (Dgp_UInt32 *)NULL;
    newcc->bitmap.userdata = newcc;

    newcc->runs = (Dgp_Int16 *)NULL;
    newcc->nruns = 0;

    /*
     * Where is this box? Not where it ought to be... there's a magic offset
     * of 32 because we're dealing with an ipshared routine here and sometimes
     * ipshared thinks there's a frame of 32 pixels at the top and left. Yay.
     */
    newcc->x = x - 32;
    newcc->y = y - 32;
  
    DGP_SEC_MALLOC(newcc->runs, (Dgp_Int16 *), (2 * sizeof(*newcc->runs) * nRuns));
    newcc->nruns = nRuns;

    /*
     * Read out the runs. They were generated bottom-to-top, so reverse
     * them.
     */
    for (run = newcc->runs + 2 * (nRuns - 1), pRunsIndex = pRuns;
         pRunsIndex;
         run -= 2, pRunsIndex = pRunsIndex->pNext) {
      run[0] = (Int16)(pRunsIndex->runLeft - x);
      run[1] = (Int16)(pRunsIndex->runRight - x);
    }
    /* Hook it onto the list of outliers */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(ccData->outliers, (void *)newcc, NULL));
    newcc = NULL;
    ccData->count++;

  }

  NORMAL_RETURN;

ERR_CLEANUP:

  if(newcc != NULL) {
    if(newcc->runs != NULL) DGP_FREE(newcc->runs);
    CC_FREE(newcc);
  }

  if(tmpruns != NULL) DGP_FREE(tmpruns);

  ERR_RETURN;
}


/* GetBoxes()
   
   Attempt to read a TIFF file and get the CCs from the resulting image.
   Throw small CCs at the (lossless) tokenizer, and store large CCs as
   outliers.
   
   Parameters:
   Dgp_State *dgp_state  -- a state cookie very similar to an ipstate, it contains
            the memory allocation functions, and error and report printing functions
   LlTokState tokstate    -- a state cookie acquired from a call to tokInit().
   UInt8 *image         -- a pointer to the (framed) image to find CCs in.
   Int32 xSize          -- the width of the image (not including frame)
   Int32 ySize          -- the height of the image (not including frame)
   Int32 bpl            -- the number of bytes per line (including frame)
   Dgp_Int32 *numBoxes        -- the number of tokpos's found will be passed back
            in *numBoxes.
   tokpos **boxes       -- the cc's found.  Null if no cc's where found
   Dgp_Int32 sizeThresh       -- CCs smaller than sizeThresh x sizeThresh will be
            neither tokenized nor put on the DgpAdt_List of outliers. 
   DgpAdt_List *outliers       -- the outliers (CCs which are too big to tokenize) will
            be passed back in *outliers (each outlier will be
      given by a cc struct -- see compress.h for the
      definition of a cc struct).  outliers may be NULL
      (so no DgpAdt_List will be returned in *outliers).
   DgpCC_TokInfoBlock *tBlock -- A vector of TokInfo structs.  Call initTokBlock (tokenCC.h)
      to get this.  GetBoxes puts an offset into the tBlock->toks array in the
      userdata field of each of the classes added to the tokenizer.  The TokInfo 
      at this offset, contains a pointer to the token if it is known, otherwise 
      NULL.  The token will only be known if you set it eariler.  (see the 
      comment in tokenCC.h for more info)
   DgpCC_PosInfoBlock *pBlock -- A vector of PosInfo structs.  Call initPosBlock (tokenCC.h)
      to get start one.  GetBoxes adds an offset into pBlock->poss to the members 
      list of each added class for each cc.  In otherwords, when a new cc is found 
      a new PosInfo is made in the pBlock.  The offset to this new PosInfo is added
      to the members list of the class in the tokenizer.  The PosInfo contains the 
      page index and an offset into the boxes array.
   Dgp_Int32 pageIndex    -- this is value is added to each PosInfo.  It is there so you 
      can have mutilple pages in one pBlock, that coorespond to a different array 
      of positions for each page.
   Dgp_Bool report    -- If true the report printer of the tcc is used to give some
      info about this function.
   
   Return values:
   return DGP_TRUE on success, and DGP_FALSE on failure.
 */
#define INIT_NUM_TOKS 50

DGP_LIBEXPORT ErrCode
DgpLlCCData_init(Dgp_State *dgp_state, 
           LlCCData *llCCData,
          Dgp_Int32 count,
          LlTokState tokstate,
          tokpos *boxes,
          Dgp_Int32 num_boxes,
          Dgp_Int32 num_boxes_alloced,
          DgpAdt_List outliers,
          Dgp_Int32 maxWidth,
          Dgp_Int32 maxHeight,
          DgpCC_TokInfoBlock *tBlock,
          DgpCC_PosInfoBlock *pBlock,
          Dgp_Int32 pageIndex,
          Dgp_Int32 tokMem,
          Dgp_Int32 pageTokMem)
{
  DGP_STATE_PROLOGUE2("DgpLlCCData_init()");

  DGP_CHECK_NULL(llCCData);
  llCCData->count = count;
  llCCData->tokstate = tokstate;
  llCCData->boxes = boxes;
  llCCData->num_boxes = num_boxes;
  llCCData->num_boxes_alloced = num_boxes_alloced;
  llCCData->outliers = outliers;
  llCCData->maxWidth = maxWidth;
  llCCData->maxHeight = maxHeight;
  llCCData->dgp_state = dgp_state;
  llCCData->tBlock = tBlock;
  llCCData->pBlock = pBlock;
  llCCData->pageIndex = pageIndex;
  llCCData->tokMem = tokMem;
  llCCData->pageTokMem = pageTokMem;
  NORMAL_RETURN;
}

ErrCode
DgpCC_getBoxes(Dgp_State *dgp_state, 
               LlTokState tokstate, 
               Dgp_UInt8 *image, 
               Dgp_Int32 xSize, 
               Dgp_Int32 ySize,
               Dgp_Int32 maxTokWidth,
               Dgp_Int32 maxTokHeight,
               Dgp_Int32 bpl, 
               Dgp_Int32 *numBoxes, 
               tokpos **boxes, 
               Dgp_Int32 sizeThresh, 
               DgpAdt_List *outliers, 
               DgpCC_TokInfoBlock *tBlock, 
               DgpCC_PosInfoBlock *pBlock, 
               Dgp_Int32 pageIndex, 
               Dgp_Bool report)
{
  DGP_STATE_PROLOGUE2("DgpCC_getBoxes()");
  
  Int32 returnValue;
  DgpCC_CC *cc;
  DgpAdt_ListNode ln;
  LlCCData ccData = DEFAULT_LLCCDATA;  
  
  /* check some params */
  DGP_ERR_BEGIN(dgp_state);
  DGP_CHECK_NULL(numBoxes);
  DGP_CHECK_NULL(boxes);

  /* validate small pool usage */
  DGP_SMALLPOOL_VALIDATE(DgpCC_CC, DGPCC_CC_POOL_ID);

  /* make sure we actually have work to do */
  if((xSize <= 0) || (ySize <= 0)) {
    *numBoxes = 0;
    *boxes = NULL;
    if(outliers != NULL) *outliers = NULL;
    NORMAL_RETURN;
  }

  /* check the rest of params */
  DGP_CHECK_NULL(tokstate);
  DGP_CHECK_NULL(image);
  DGP_CHECK_NULL(tBlock);
  DGP_CHECK_NULL(pBlock);
  DGP_ERR_GEN((bpl <= 0), cErrInvalidParm);

  /* clear up */
  if(outliers != NULL) *outliers = NULL;
  *numBoxes = 0;
  *boxes = NULL;


#ifdef DBG
  DGP_REPORTF3(report, "GetBoxes() starts with \n\
UInt8 *image = %ld\n\
Int32 xSize  = %d\n\
Int32 ySize  = %d",image,xSize,ySize); 
  DGP_REPORTF1(report, "\
Int32 bpl    = %d", bpl);
#else
  DGP_REPORTF0(report, "DgpCC_getBoxes() starts.");
#endif
    
  /* init cc finder cookie */
  ccData.dgp_state = dgp_state;
  ccData.tBlock = tBlock;
  ccData.pBlock = pBlock;
  ccData.pageIndex = pageIndex;
  
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&(ccData.outliers), dgp_state));
  
  ccData.tokstate = tokstate;
  
  DGP_SEC_MALLOC(ccData.boxes, (tokpos *), 
       sizeof(tokpos) * INIT_SIZE_BOXES);
  ccData.num_boxes_alloced = INIT_SIZE_BOXES;
  
  /* Decide what the cutoff for conversion to bitmap is */
  ccData.maxWidth  = maxTokWidth;
  ccData.maxHeight = maxTokHeight;
  
  /* NOTE that we just effectively turned off the feature whereby
     only those bitmaps smaller than a certain size are eligible
     to be tokenized.  So we'll be trying to tokenize CCs of any size. */
  
  DGP_REPORTF0(report,  "About to call ipcore CC finder.");
  
  /* Get all the CCs */
  DGP_ERR_PASS_CLEANUP(i_findConnectedComponentRuns(dgp_state->ipstate,
           image, (unsigned)bpl, 32, 32,
           xSize, ySize,
           (unsigned)sizeThresh, (unsigned)sizeThresh,
           1,
           (void *)&ccData,
           MakeModelsFromRuns, 
           &returnValue));
  
  DGP_REPORTF1(report,  "Found %d ccs", ccData.count);
  
  /* shrink the array to the min. size required (it might have more tokpos's
     in it than just ccData.num_boxes of them) */
  if (ccData.num_boxes > 0){
    DGP_SEC_REALLOC(ccData.boxes, (tokpos *), ccData.num_boxes * sizeof(tokpos));
  } else {
    DGP_FREE(ccData.boxes);
    ccData.boxes = NULL;
  }
  
  *numBoxes = ccData.num_boxes;

  if (outliers != NULL) {

    /* give them the list */
    *outliers = ccData.outliers;

  } else {

    /* destroy the list...? */
    DGPADT_FOREACH(ln, ccData.outliers) {
      cc = (DgpCC_CC *)DGPADT_LIGET(ln);
      DGP_FREE(cc->runs);
      CC_FREE(cc);
    }
    DgpAdt_liFree(ccData.outliers);

  }
  ccData.outliers = DGP_NULLLIST;
  
  DGP_REPORTF0(report, "GetBoxes() ends (ok).");
  *boxes = ccData.boxes;
  ccData.boxes = NULL;
  
  NORMAL_RETURN;
  
ERR_CLEANUP:

  if (ccData.boxes != NULL) {
    DGP_FREE((void *)ccData.boxes);
  }

  if (ccData.outliers != DGP_NULLLIST) {
    DGPADT_FOREACH(ln, ccData.outliers) {
      cc = (DgpCC_CC *)DGPADT_LIGET(ln);
      if(cc != NULL) {
        if(cc->runs != NULL) DGP_FREE(cc->runs);
        CC_FREE(cc);
      }
    }
    DgpAdt_liFree(ccData.outliers);
  }

  ERR_RETURN;
}
