/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 1, 2000
**
**  Filename: cc2box.h
**
**  Description of what's implemented in this file:
**    function header for cc finding in binary images
**  
**************************************************************************/

/*
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/cc2box.h,v 1.6 2002/10/15 00:19:34 jlin Exp $
 */

#ifndef CC2BOX_H
#define CC2BOX_H

#include "bit_array.h"
#include "list.h"
#include "mmcompress.h"
#include "samptok.h"
#include "tok-lossless.h"
#include "types.pub"
#include "uds.h"
#include <color-extension.h>
#include "dgp_utils.h"
#include "tokenCC.h"

#ifdef __cplusplus
extern "C"
{
#endif

typedef struct _LlCCData {
  Dgp_Int32 count;
  LlTokState tokstate;
  tokpos *boxes;
  Dgp_Int32 num_boxes;
  Dgp_Int32 num_boxes_alloced;
  DgpAdt_List outliers;
  Dgp_Int32 maxWidth;
  Dgp_Int32 maxHeight;
  Dgp_State *dgp_state;
  DgpCC_TokInfoBlock *tBlock;
  DgpCC_PosInfoBlock *pBlock;
  Dgp_Int32 pageIndex;
  Dgp_Int32 tokMem;
  Dgp_Int32 pageTokMem;
} LlCCData;

#define DEFAULT_LLCCDATA {0, NULL, NULL, 0, 0, DGP_NULLLIST, 0, 0, NULL, NULL, NULL, 0, 0, 0}
#define INIT_SIZE_BOXES 5000
#define INIT_SIZE_SAMPBOXES 10

DGP_LIBEXPORT DgpImage_RGB DgpCC_DefaultBgColor;

/* tokBitmap used to be defined and used by the module giving the lossless 
   tokenizer, but now that tokenizer works exclusively with runs, so we
   need to define tokBitmap here ourselves. */
typedef struct {
    Dgp_UInt16 width, height;
    Dgp_UInt32 *bits;
    Dgp_UInt16 longsperline;
    void *userdata;
    } DgpCC_TokBitmap;

typedef struct {
    DgpCC_TokBitmap bitmap;
    Dgp_Int32 x, y;
    Dgp_Int16 *runs;
    Dgp_Int32 nruns;
    } DgpCC_CC;

/* this is the small pool from which DgpCC_CC's will be allocated,
   (outliers are returned using this structure) */
#define DGPCC_CC_POOL_ID DGP_SMALLPOOL_MIN32


/* GetBoxes()
   
   Attempt to read a TIFF file and get the CCs from the resulting image.
   Throw small CCs at the (lossless) tokenizer, and store large CCs as
   outliers.
   
   Parameters:
   TokState tokstate -- a state cookie acquired from a call to tokInit().
   UInt8 *image      -- a pointer to the (framed) image to find CCs in.
   Int32 xSize       -- the width of the image (not including frame)
   Int32 ySize       -- the height of the image (not including frame)
   Int32 bpl         -- the number of bytes per line (including frame)
   Dgp_Int32 *numBoxes     -- the number of tokpos's found will be passed back
                        in *numBoxes.
   tokpos **boxes    -- the cc's found.  Null if no cc's where found
   Dgp_Int32 sizeThresh    -- CCs smaller than sizeThresh x sizeThresh will be
                        neither tokenized nor put on the DgpAdt_List of outliers. 
   DgpAdt_List *outliers    -- the outliers (CCs which are too big to tokenize) will
                        be passed back in *outliers (each outlier will be
      given by a cc struct -- see above for the
      definition of a cc struct).  outliers may be NULL
      (so no DgpAdt_List will be returned in *outliers).
   
   Return values:
   return DGP_TRUE on success, and DGP_FALSE on failure.
 */
DGP_LIBEXPORT ErrCode
DgpCC_getBoxes(Dgp_State *dgp_state, 
               LlTokState tokstate, 
               Dgp_UInt8 *image, 
               Dgp_Int32 xSize, 
               Dgp_Int32 ySize,
               Dgp_Int32 maxTokWidth,
               Dgp_Int32 maxTokHeight,
               Dgp_Int32 bpl, 
               Dgp_Int32 *numBoxes, 
               tokpos **boxes, 
               Dgp_Int32 sizeThresh, 
               DgpAdt_List *outliers, 
               DgpCC_TokInfoBlock *tBlock, 
               DgpCC_PosInfoBlock *pBlock, 
               Dgp_Int32 pageIndex, 
               Dgp_Bool report);

DGP_LIBEXPORT ErrCode
DgpLlCCData_init(Dgp_State *dgp_state, 
           LlCCData *llCCData,
          Dgp_Int32 count,
          LlTokState tokstate,
          tokpos *boxes,
          Dgp_Int32 num_boxes,
          Dgp_Int32 num_boxes_alloced,
          DgpAdt_List outliers,
          Dgp_Int32 maxWidth,
          Dgp_Int32 maxHeight,
          DgpCC_TokInfoBlock *tBlock,
          DgpCC_PosInfoBlock *pBlock,
          Dgp_Int32 pageIndex,
          Dgp_Int32 tokMem,
          Dgp_Int32 pageTokMem);
#ifdef __cplusplus
}
#endif

#endif
