/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 1, 2000
**
**  Filename: colorCC.c
**
**  Description of what's implemented in this file:
**    core functions for color cc finding.
**  
**************************************************************************/

/*
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCC.c,v 1.25 2003/01/27 21:48:39 awilliam Exp $
 */

#include "colorCC-priv.h"
#include <stdlib.h>
#include <sys/types.h>
#include <image.h>
#include "image_defs.h"
#include <list.h>
#include "colorCC.h"
#include "dgp_utils.h"
#include "colorCCstate.h"
#include "simple_utils.h"
#include "run2bitmap.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCC.c,v 1.25 2003/01/27 21:48:39 awilliam Exp $");


static INLINE ErrCode
ccstate_movetoNextLine(CCstate *ccstate, 
                       Run *prev_row_runs,
                       colorCC_state *cookie)
{
  DGP_PROLOGUE("ccstate_movetoNextLine()", NULL);
  
  Dgp_State *dgp_state = cookie->dgp_state;
  DgpAdt_ListNode ln;
  void *ptmp, *first_seg = NULL, *cur_seg;
  Dgp_Int32 num, i, j;
  Dgp_Int32 cur_run_index;
  Dgp_Bool use_runs = !(ccstate_isWashOrSampled(ccstate));
  Dgp_Int32 scrs_pool_id = ((use_runs) ? SCRS_POOL_ID : SCXRS_POOL_ID);


  if(!ccstate_isBackground(ccstate)) {

    /* take care of the second to last line (currently in runs),
       (convert to sc_runs or sc_xruns) */
    if(DGPADT_LILEN(ccstate->prev_run_line) > 0) {

      /* figure out how many new runs there are */
      num = DGPADT_LILEN(ccstate->prev_run_line);
      ln = DGPADT_LIFIRST(ccstate->prev_run_line);
      ptmp = &(first_seg);

      /* convert the second to last line (currently in runs) to
         sc_runs */
      while(num > 0) {

        /* figure out how many runs there are in this segment */
        j = MIN(num, SC_RUN_NUM_RUNS);
        num -= j;

        /* create a new segment */
        SCRS_SEC_MALLOC_WID(scrs_pool_id, &cur_seg);
        scrs_initSeg(((sc_run_seg *)cur_seg), (Dgp_UInt8)j);

        /* add the new segment onto the list */
        *((void **)ptmp) = cur_seg;
        ptmp = &(((sc_run_seg *)cur_seg)->next);

        i = 0;

        if(use_runs) {

          while(i < j) {
      
            /* get the run */
            cur_run_index = (Dgp_Int32)DGPADT_LIGET(ln);
            DGPADT_LIPUT(ln, NULL);
      
            /* convert the run to an sc_run */
            ((sc_run_seg *)cur_seg)->runs[i].xpos = prev_row_runs[cur_run_index].xpos;
            ((sc_run_seg *)cur_seg)->runs[i].length = prev_row_runs[cur_run_index].cccrun.length;

            /* move to next run in list */
            i++;
            ln = DGPADT_LINEXT(ln);

          }

        } else {

          while(i < j) {
      
            /* get the run */
            cur_run_index = (Dgp_Int32)DGPADT_LIGET(ln);
            DGPADT_LIPUT(ln, NULL);
      
            /* convert the run to an sc_xrun */
            ((sc_xrun_seg *)cur_seg)->runs[i].xpos = prev_row_runs[cur_run_index].xpos;
            ((sc_xrun_seg *)cur_seg)->runs[i].length = prev_row_runs[cur_run_index].cccrun.length;
            ((sc_xrun_seg *)cur_seg)->userdata[i] = 
              ((Dgp_UInt32)(prev_row_runs[cur_run_index].cccrun.colorInfo.rgbSamples));

            /* move to next run in list */
            i++;
            ln = DGPADT_LINEXT(ln);

          }

        }
      }
    
      /* add the new sc_run segment list to the end of the sc_lines list */
      DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(ccstate->sc_lines, first_seg, NULL));

    }

  }

  /* clear the old list (in the case of backgrounds, we just
     "lose" this info) */
  DGP_ERR_PASS_CLEANUP(DgpAdt_liClear(ccstate->prev_run_line));


  /* make the last line (in runs) the next to last line */
  ptmp = ccstate->cur_run_line;
  ccstate->cur_run_line = ccstate->prev_run_line;
  ccstate->prev_run_line = ptmp;


  NORMAL_RETURN;

ERR_CLEANUP:

  /* get rid of created sc_xruns */
  scrs_cleanup(&first_seg, (Dgp_Bool)(!use_runs), (Dgp_Bool)(ccstate_isSampled(ccstate)),
    dgp_state);


  ERR_RETURN;
}

ErrCode
DgpCC_init(Dgp_State *dgp_state, 
           DgpCC_ColorCookie *_cookie)
{
  DGP_STATE_PROLOGUE2("DgpCC_init()");

  colorCC_state * cookie = NULL;


  /* check params */
  DGP_ERR_BEGIN(dgp_state);
  DGP_CHECK_NULL(_cookie);
  *_cookie = NULL;

  /* validate our small pool usage */
  DGP_SMALLPOOL_VALIDATE(sc_run_seg, SCRS_POOL_ID);
  DGP_SMALLPOOL_VALIDATE(sc_xrun_seg, SCXRS_POOL_ID);
  DGP_SMALLPOOL_VALIDATE(DgpCC_ColorCC, DGPCC_COLORCC_POOL);

  /* create cookie */
  DGP_SEC_CALLOC(cookie, (colorCC_state *), 1, sizeof(colorCC_state));
  cookie->dgp_state = dgp_state;

  /* init cccmap */
  DGP_ERR_PASS_CLEANUP(cccmap_init(&(cookie->cccmap), cookie));

  /* init run image */
  DGP_ERR_PASS_CLEANUP(rim_init(&(cookie->rim), cookie));


  /* init memory usage settings */
  *_cookie = (DgpCC_ColorCookie)cookie;
  cookie = NULL;

  NORMAL_RETURN;
ERR_CLEANUP:

  if(cookie != NULL) DGP_FREE(cookie);

  ERR_RETURN;

}

ErrCode
DgpCC_cleanup(DgpCC_ColorCookie *_cookie)
{
  DGP_STATE_PROLOGUE2("DgpCC_cleanup()");
  colorCC_state *cookie = NULL;
  Dgp_Free tmp_free = NULL;


  DGP_STOP_WHINE;

  if(_cookie == NULL) NORMAL_RETURN;

  cookie = *_cookie;
  *_cookie = NULL;

  if(cookie == NULL) NORMAL_RETURN;

  tmp_free = cookie->DGP_FREE;

  if(cookie->cccmap != NULL) cccmap_cleanup(&(cookie->cccmap), cookie);
  if(cookie->rim != NULL) rim_cleanup(&(cookie->rim), cookie);

  /* finally, free the cookie */
  tmp_free(cookie);

  NORMAL_RETURN;
}


#define AREA_PERCENT ((Dgp_Float32)(0.10))

ErrCode
DgpCC_initPage(DgpCC_ColorCookie _cookie, 
               Dgp_Int32 width, 
               Dgp_Int32 height, 
               Dgp_UInt16 washThreshold,
               Dgp_Int32 minSampSize)
{
  DGP_STATE_PROLOGUE2("DgpCC_initPage()");
  colorCC_state *cookie = NULL;

  /* check params */
  DGP_CHECK_NULL(_cookie);
  cookie = (colorCC_state *)_cookie;
  DGP_ERR_BEGIN(cookie->dgp_state);
  DGP_ERR_GEN(((width < 0) || (height < 0)), cErrInvalidParm);
  
  cookie->width = width;
  cookie->height = height;
  cookie->minSampSize = minSampSize;
  cookie->wThresh = washThreshold;

  cookie->curLine = 0;

  NORMAL_RETURN;
}


static ErrCode
dgpcc_movetoNextLine(colorCC_state *cookie)
{
  DGP_PROLOGUE("dgpcc_movetoNextLine()", cookie->DGP_PERRSTATE);
  cccMap *cccmap = cookie->cccmap;
  CCstate *ccstate;
  RunImage *rim = cookie->rim;
  RunRow *prev_row;
  Run *prev_row_runs;
  Dgp_Int32 i, num_ccstates;


  prev_row_runs = (rim_getPrevRow(rim))->runs;

  /* the cccmap holds pointers to all ccstates, so we'll use that
     to move each ccstate to the next line, and udate it's "active"
     status.  a ccstate is active if it can still have runs added
     on to it. */

  for(i = 0, num_ccstates = cccmap->num_els; num_ccstates > 0; i++) {
  
    ccstate = cccmap->col[i].ccstate;
  
    /* if this is an empty row, skip it */
    if(ccstate == NULL) continue;
    num_ccstates--;

    if(ccstate_isActive(ccstate)) {

      /* move the ccstate to the next line */
      DGP_ERR_PASS(ccstate_movetoNextLine(ccstate, prev_row_runs, cookie));

      /* check if it's still active */
      if(ccstate->lrY < cookie->curLine) {

        /* set it as newly inactive */
        ccstate_setNewInactive(ccstate);
        cookie->num_new_inactive++;
      }

      /* figure out if it's background (bigger than 10% of the page and
         the background color */
      if((ccstate->color == DEFAULT_BG_COLOR) &&
         ccstate_isNormal(ccstate)) {

        ccstate_makeBackground(ccstate, cookie);

      }      
      
    }


  }

  /* move to the next row in the run image */
  rim_moveNextRow(rim, prev_row);

  /* increment line count in cookie */
  cookie->curLine++;


  NORMAL_RETURN;
}


ErrCode
dgpcc_updateStatus(colorCC_state *cookie)
{
  Dgp_State *dgp_state = cookie->dgp_state;
  DGP_PROLOGUE("dgpcc_updateStatus()", DGP_PERRSTATE);
  cccMap *cccmap = cookie->cccmap;
  CCstate *ccstate;
  DgpCC_ColorCC *new_cc = NULL;
  Dgp_Int32 i;
  Dgp_Bool free_ccstate = DGP_FALSE;


  /* the cccmap holds pointers to all ccstates, so we'll
     use this to move through and check all newly
     inactive ccstates for active neighbors. */


  /* we need to both update ccstates' status and try to remove dead ones */
  for(i = 0; i < cccmap->max_num_els; i++) {
    ccstate = cccmap->col[i].ccstate;

    /* if this is an empty row or the ccstate is active, skip it */
    if((ccstate == NULL) || ccstate_isActive(ccstate)) continue;

    switch(ccstate->state.active) {
    case CCSTATE_NEW_INACTIVE:

      /* it's no longer "new" inactive... */
      ccstate_setInactive(ccstate);

      /* drop through */
    case CCSTATE_INACTIVE:

      /* can we remove this ccstate? */
      if(!ccstate_isPotentialWash(ccstate)) {

        /* remove it from the neighbor map */
        cccmap_rem(cccmap, ccstate->map_index);

        /* if we encounter problems, we've got to free this ccstate */
        free_ccstate = DGP_TRUE;

        if(!ccstate_isBackground(ccstate)) {

            /* make a colorCC out of it */
            DGP_ERR_PASS_CLEANUP(ccstate_convertToColorCC(ccstate, &new_cc, cookie));
        
            if(!colorcc_isWashOrSampled(new_cc)) {
              DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(cookie->new_ccs, new_cc, NULL));
            } else {
              DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(cookie->new_sampled_ccs, new_cc, NULL));
            }
            new_cc = NULL;
        }

        /* ditch it, we don't need it anymore */
        ccstate_cleanup(&ccstate, cookie);

        /* alright, it's gone... */
        free_ccstate = DGP_FALSE;
      }
    }

  }

  /* clear new inactive count */
  cookie->num_new_inactive = 0;

  NORMAL_RETURN;
ERR_CLEANUP:

  if((free_ccstate) && (ccstate != NULL)) {
    ccstate_cleanup(&ccstate, cookie);
  }

  if(new_cc != NULL) {
    if((new_cc->state.sampled) && (new_cc->colorInfo.rgbIm != NULL)) {
      DGPIMAGE_FREE(new_cc->colorInfo.rgbIm);
    }

    if(new_cc->runList != NULL) {
      DGP_FREE(new_cc->runList);
    }

    COLORCC_FREE(new_cc);
  }

  ERR_RETURN;
}


#if 0

/* this function can be used to see the output of the color cc finder */

static ErrCode
saveAllCCstates(colorCC_state *cookie, DgpAdt_List new_ccs, DgpAdt_List new_sampled_ccs)
{
  DGP_PROLOGUE("getFinishedCCstates()", cookie->DGP_PERRSTATE);
  DgpCC_ColorCC *new_cc;
  DgpImage_ColorRGB rgb_im = NULL;
  DgpImage_ColorRGBPixPtr rgb_ptr;
  DgpImage_LongPixPtr l_ptr;
  DgpImage_Packed mask_im = NULL;
  FILE *cc_info = NULL;
  Dgp_Int32 imx, imy, lpl;
  DgpAdt_ListNode ln;
  Dgp_Int32 width, height;

  static Dgp_Int32 wash_num = 0, sample_num = 0, cc_num = 0;

  Dgp_Int8 fname[64];

  
  cc_info = fopen("tokens\\cc_info.txt", "a");
  DGP_CHECK_NULL(cc_info);


  if(new_sampled_ccs != DGP_NULLLIST) {

    DGPADT_FOREACH(ln, new_sampled_ccs) {

      new_cc = (DgpCC_ColorCC *)DGPADT_LIGET(ln);

      width = new_cc->lrX - new_cc->ulX + 1;
      height = new_cc->lrY - new_cc->ulY + 1;
      if(colorcc_isSampled(new_cc)) {
        fprintf(cc_info, "sampcc %03d - x %d, y %d, w %d, h %d\n", sample_num, new_cc->ulX, new_cc->ulY,
          width, height);
      } else {
        fprintf(cc_info, "washcc %03d - x %d, y %d, w %d, h %d\n", wash_num, new_cc->ulX, new_cc->ulY,
          width, height);
      }
      fflush(cc_info);

        /* save the sampled im */
      if(colorcc_isSampled(new_cc)) {
        sprintf(fname, "tokens\\sampcc%03d.ppm", sample_num);
      } else {
        sprintf(fname, "tokens\\washcc%03d.ppm", wash_num);
      }
      DGP_ERR_PASS_CLEANUP(DGPIMAGE_SAVE(new_cc->colorInfo.rgbIm, fname));

      /* create the mask */
      lpl = (DGPIMG_WIDTH(new_cc->colorInfo.rgbIm) + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
      DGP_ERR_PASS_CLEANUP(RunsToUnframedBitmap(&l_ptr, new_cc->runList, new_cc->numRuns, 
        DGPIMG_HEIGHT(new_cc->colorInfo.rgbIm), lpl, 0, cookie->dgp_state));
      DGP_ERR_PASS_CLEANUP(DgpImage_wrap(cookie->dgp_state, l_ptr, 
        DGPIMG_WIDTH(new_cc->colorInfo.rgbIm), 
        DGPIMG_HEIGHT(new_cc->colorInfo.rgbIm), lpl, &mask_im));

      /* save the mask */
      if(colorcc_isSampled(new_cc)) {
        sprintf(fname, "tokens\\sampcc%03dm.pbm", sample_num++);
      } else {
        sprintf(fname, "tokens\\washcc%03dm.pbm", wash_num++);
      }
      DGP_ERR_PASS_CLEANUP(DGPIMAGE_SAVE(mask_im, fname));

      DGPIMAGE_FREE(mask_im);
      mask_im = NULL;

    }
  }

  if(new_ccs != DGP_NULLLIST) {

    DGPADT_FOREACH(ln, new_ccs) {

      Dgp_Int32 cur_run, j;
      Dgp_UInt16 *runs;

      new_cc = (DgpCC_ColorCC *)DGPADT_LIGET(ln);

      width = new_cc->lrX - new_cc->ulX + 1;
      height = new_cc->lrY - new_cc->ulY + 1;
      fprintf(cc_info, "cc %05d - x %d, y %d, w %d, h %d\n", cc_num, new_cc->ulX, new_cc->ulY,
        width, height);
      fflush(cc_info);

      DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(cookie->dgp_state, DGPIMAGE_RGB, (new_cc->lrX - new_cc->ulX + 1),
        (new_cc->lrY - new_cc->ulY + 1), &rgb_im));

      DGP_CHECK_ALLOC_CLEANUP(rgb_im);

      if((new_cc->colorInfo.rgbColor.r == DEFAULT_BG_COLOR_R) &&
         (new_cc->colorInfo.rgbColor.g == DEFAULT_BG_COLOR_G) &&
         (new_cc->colorInfo.rgbColor.b == DEFAULT_BG_COLOR_B)) {

        /* set background color of "white" images */
        memset((void *)DGPIMG_PIXPTR(rgb_im, 0, 0), ~(DEFAULT_BG_COLOR_R), 
          (DGPIMG_WIDTH(rgb_im) * DGPIMG_HEIGHT(rgb_im) * sizeof(DgpImage_RGB)));

      } else {


        /* set background color of other images */
        memset((void *)DGPIMG_PIXPTR(rgb_im, 0, 0), DEFAULT_BG_COLOR_R, 
          (DGPIMG_WIDTH(rgb_im) * DGPIMG_HEIGHT(rgb_im) * sizeof(DgpImage_RGB)));

      }

      /* create cc */
      cur_run = 0;
      imx = 0;
      imy = 0;
      runs = new_cc->runList;
      while(cur_run < new_cc->numRuns) {
        DGP_ASSERT(imy < height);
        DGP_ASSERT(runs[0] + runs[1] <= width);
        rgb_ptr = &(DGPIMG_REF(rgb_im, runs[0], imy));
        for(j = 0; j < runs[1]; j++) {
          *rgb_ptr++ = new_cc->colorInfo.rgbColor;
        }
        imx = runs[0] + runs[1];

        runs += 2;
        cur_run++;
        if(cur_run == new_cc->numRuns) break;
        if(runs[0] <= imx) {
          imy++;
        }
      }

      sprintf(fname, "tokens\\cc%05d.ppm", cc_num++);
      DGP_ERR_PASS_CLEANUP(DGPIMAGE_SAVE(rgb_im, fname));
      DGPIMAGE_FREE(rgb_im);
      rgb_im = NULL;


    }

  }

  fclose(cc_info);
  cc_info = NULL;


  NORMAL_RETURN;

ERR_CLEANUP:

  if(cc_info != NULL) fclose(cc_info);

  ERR_RETURN;
}
#endif


/*
* See the comment in colorCC.h to see what this black box
* is supposed to do.
*
* Here is what it really does.
* A .qdr file consists of two Dgp_Int16 indicating the height and
* with of the image.  Then it contains lists of (header, data)
* pairs seperated by end of line markers.  See readRun which
* describes the encoding of these pairs and markers.
* Each pair is a run of pixels and a color.  The color can
* be either solid or sampled.  The solid color is a 32-bit cmyk color.
* The sampled color is a list of 32-bit cmyk colors for the length
* of the run.
*
* These runs are read from the file and given a label.
* all run's which are touching by 8-connectedness have the sample parent
* label.  The label also contains the the number of runs and the
* bounding box of all runs that are labeled with it.
* The runs of each row are stored in a list.
* If the run is sampled an array of 32-bit values contains its color.
* Also on the is first pass through the image a label is return that
* is pointed to the most number of pixels
*
* The creation of CCC's and the placement of runs in CCC's is done on
* row by row basis.
* First each run of a row is converted:
*  If the label of the run does not have a CCC associated with it
*   a new CCC is created, and added to the list of CCC's.
*   If the run is sampled then the CCC contains
*   an image to store the sampled colors.
*  The length, cc, and a flag of the run are put into an array for the
*   the row.
*  If the run is sampled, the sampled colors are put into the CCC image.
* Second the runs in the row's array are stored into the CCC's that they
*  point to.  If a run meets the filling criteria it is extended.
* The list of CCC's is returned.
*
* modification (stump, 9/21/96): the number of sampled images found will
* be returned in *num_sampled_images.
*
* modification (scytacki, 8/1/97): allow input of image array 
* by generalizing input to void * and stucture that contains
* functions for returning runs.
*/
ErrCode 
DgpCC_doStrip(DgpCC_ColorCookie _cookie, 
              DgpCC_PFNReadRun readRun, 
              DgpCC_RunCookie runCookie, 
              Dgp_Int32 numLines, 
              DgpAdt_List new_ccs,
              DgpAdt_List new_sampled_ccs)
{
  DGP_STATE_PROLOGUE2("ccc_doStrip()");
  colorCC_state *cookie = NULL;
  Dgp_Int32 i;
  Dgp_Allocators *a;


  /* check params */
  DGP_CHECK_NULL(_cookie);
  cookie = (colorCC_state *)_cookie;
  DGP_ERR_BEGIN(cookie->dgp_state);
  DGP_CHECK_NULL(readRun);
  DGP_CHECK_NULL(new_ccs);
  DGP_CHECK_NULL(new_sampled_ccs);

  /* get allocators */
  a = &(cookie->DGP_ALLOCATORS);

  /* uhh, my business is finished, i go now */
  if(numLines <= 0) NORMAL_RETURN;

  cookie->readRun = readRun;
  cookie->runCookie = runCookie;
  cookie->new_ccs = new_ccs;
  cookie->new_sampled_ccs = new_sampled_ccs;
  cookie->is_firstLine = DGP_TRUE;

  for(i = 0; i < numLines; i++) {

    /* read a line of the image */
    DGP_ERR_PASS_CLEANUP(dgpcc_readLine(cookie));

    /* alrightee, move all the ccstates to the next line */
    DGP_ERR_PASS_CLEANUP(dgpcc_movetoNextLine(cookie));

  }

    /* alrightee, move all the ccstates past the last line */
    DGP_ERR_PASS_CLEANUP(dgpcc_movetoNextLine(cookie));

    /* free up some memory in the run image */
    DGP_ERR_PASS_CLEANUP(rim_endPage(cookie->rim, cookie));  

  /* do final heuristics if necessary (wash detection and filling),
     this should get rid of any leftover ccstates. */
    if(cookie->num_new_inactive > 0) {
      if(cookie->wThresh != CCC_NOWASH) {
        DGP_ERR_PASS_CLEANUP(dgpcc_washMerge(cookie));
      }
      DGP_ERR_PASS_CLEANUP(dgpcc_updateStatus(cookie));
    }

    if (cookie->curLine != cookie->height)
    {
      /* correct the current line */
      cookie->curLine--;
    }

#if DGP_DBUG
  if(cookie->curLine > cookie->height) {
    DGP_ASSERT(cookie->cccmap->num_els == 0);
  }
#endif

  /* collapse the map some */
  cccmap_collapse(cookie->cccmap, cookie);

#if DGP_DBUG
  DGP_ASSERT(cccmap_validate(cookie) == cSuccessful);
#endif

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}
