/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 1, 2000
**
**  Filename: colorCC.h
**
**  Description of what's implemented in this file:
**    main header for color cc finder library
**  
**************************************************************************/

/*
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCC.h,v 1.7 2001/06/21 18:50:50 mkrocken Exp $
 */

#ifndef __COLORCC_H
#define __COLORCC_H

#include "dgp_types.h"
#include "image.h"
#include "list.h"
#include "nt_compat.h"
#include "util.h"
#include <dgp_allocators.h>
#include "dgp_state.h"
#include "cccrun_t.h"

#ifdef __cplusplus
extern "C"
{
#endif

#define DGPCC_RGB_FORMAT

/* struct to return color connected component info */
typedef struct {

  /* run list for this colorCC */
  Dgp_UInt16 *runList;
  Dgp_Int32 numRuns;

  /* coordinates for the position of this colorCC */
  Dgp_UInt16 ulX, ulY, lrX, lrY;
  
  /* color value or image (if sampled or wash) */
  union {
    DgpImage_RGB rgbColor;
    DgpImage_ColorRGB rgbIm;
  } colorInfo;

  /* various state flags for this colorCC */
  struct {
    Dgp_UInt8 sampled : 1;  /* is this colorCC sampled */
    Dgp_UInt8 wash    : 1;  /* is this colorCC a wash */
  } state;

} DgpCC_ColorCC;

#define DGPCC_COLORCC_POOL DGP_SMALLPOOL_MIN24

/*
 * Types and a structure to allow any type of input
 * source.  
 */
typedef void *DgpCC_RunCookie;
typedef ErrCode (TC_CDECL *DgpCC_PFNReadRun)(DgpCC_CccRun *, Dgp_Bool *eol, DgpCC_RunCookie, Dgp_State *);


typedef void *DgpCC_ColorCookie;

/*
 * Given an input reader structure findCC() makes a list of CCC's.
 * The maxCC is not included in the list of CCC's.
 * Currently the maxCC is the CCC(Color Connected Component),
 * with the most pixels.  Its ulX,ulY,lrX,lrY are correct.
 * It's runList is NULL.
 * We define a CCC as a 8-connected component.
 * So if we are looking at pixel "O" all          X X X
 * the "X"'s are canidates for being in the       X O X
 * same component.                                X X X
 *
 * The CCC's returned by this function are also filled in.
 * The intension of filling is to cut down on duplicate
 * shape information.  This duplication occurs whenever
 * two CCC's touch.  We don't want to remove all the duplication
 * because we use the shape information to match one CCC with
 * another.  So first the criteria then the outcome.
 * Criteria- we want to elimate duplicate shape info when:
 *  o The bounding box of one CCC is completely inside of another.
 *   and
 *  o The outside CCC is present both to the left and the right
 *     of a run of pixels in the inside CCC.
 *  
 * Outcome - when a run of pixels of the inner CCC meets these
 *  criteria, the run of pixels is added to the outer CCC,
 *  the inner CCC remains the same.
 * We call this a fill.
 *
 * No layering information is necessary to reconstruct the image
 *  from these CCC's.  This is because of the bounding box
 *  stipulation.  The reconstructer only needs to put down
 *  CCC's that contain other CCC's before the contained CCC's.
 *  This can be acomplished by sorting the CCC's by a bounding
 *  box corner and then drawing the CCC's starting from that
 *  same corner of the page.
 * 
 *  modification (stump, 9/21/96): the number of sampled images findCC finds
 *  will be returned as *num_sampled_images.
 */
DGP_LIBEXPORT ErrCode TC_CDECL
DgpCC_init(Dgp_State *dgp_state, 
           DgpCC_ColorCookie *cookie);

#define CCC_NOWASH     ((Dgp_UInt16)0xFFFF)

DGP_LIBEXPORT ErrCode TC_CDECL
DgpCC_initPage(DgpCC_ColorCookie cookie, 
               Dgp_Int32 width, 
               Dgp_Int32 height, 
               Dgp_UInt16 washThreshold,
               Dgp_Int32 minSampSize);

DGP_LIBEXPORT ErrCode TC_CDECL
DgpCC_cleanup(DgpCC_ColorCookie *cookie);


DGP_LIBEXPORT ErrCode TC_CDECL
DgpCC_doStrip(DgpCC_ColorCookie cookie, 
              DgpCC_PFNReadRun readRun, 
              DgpCC_RunCookie runCookie, 
              Dgp_Int32 numLines, 
              DgpAdt_List new_ccs,
              DgpAdt_List new_sampled_ccs);


#ifdef __cplusplus
}
#endif

#endif /* __COLORCC_H */

