/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 1 2000
**
**  Filename: colorCCutil.c
**
**  Description of what's implemented in this file:
**    color connected components utilities
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCstate.c,v 1.26 2003/02/10 22:53:38 mkrocken Exp $
 *
 */

#include "dgp_rcs.h"
#include "colorCCstate.h"
#include "dgp_utils.h"
#include "colorCC-priv.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCstate.c,v 1.26 2003/02/10 22:53:38 mkrocken Exp $");

/* update background's neighbors */
/* 30.10.2000 / dhtesq. */
/* Moved from colorCCutil.h... */
static INLINE void
cccmap_newBackground(cccMap *map, Dgp_Int32 element)
{
  cccMapColData *col = map->col;
  Dgp_Int32 i, num_neighbors = col[element].ccstate->num_neighbors;
  cccMapRowData *row = col[element].row;


  /* done yet? */
  if(num_neighbors <= 0) return;

  /* changing neighbor's data */
  for(i = 0; i < element; i++) {
    if(cccmaprd_isNeighbor(&row[i])) {
      if(cccmaprd_isPotentialWash(&row[i])) {
        cccmaprd_clearPotentialWash(&row[i]);
        col[i].ccstate->num_potential_washes--;
      }

      /* done yet? */
      if(--num_neighbors == 0) return;
    }
  }

  /* still changing neighbor's data */
  for(++i; i < map->max_num_els; i++) {
    if(cccmaprd_isNeighbor(&(col[i].row[element]))) {
      if(cccmaprd_isPotentialWash(&(col[i].row[element]))) {
        cccmaprd_clearPotentialWash(&(col[i].row[element]));
        col[i].ccstate->num_potential_washes--;
      }

      /* done yet? */
      if(--num_neighbors == 0) return;
    }
  }

}


ErrCode
ccstate_init(CCstate **_ccstate, 
             Run *cur_row_runs, 
             Dgp_Int32 cur_run_index,
             colorCC_state *cookie)
{
  Dgp_State *dgp_state = cookie->dgp_state;
  DGP_PROLOGUE("ccstate_init()", DGP_PERRSTATE);
  CCstate *ccstate = NULL;
  Run *cur_run = &cur_row_runs[cur_run_index];


  DGP_ASSERT(_ccstate != NULL);

  /* do heuristics if necessary (wash detection),
     this may get rid of some old ccstates, thereby reducing mem usage. */
  if(cookie->num_new_inactive > DGPCC_MIN_UPDATE_COUNT) {
    if(cookie->wThresh != CCC_NOWASH) {
      DGP_ERR_PASS_CLEANUP(dgpcc_washMerge(cookie));
    } 
    DGP_ERR_PASS_CLEANUP(dgpcc_updateStatus(cookie)); 
  } 

  /* create new ccstate struct */
  DGP_SEC_CALLOC(ccstate, (CCstate *), 1, sizeof(CCstate));

  /* create new lines list */
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&(ccstate->sc_lines), dgp_state));

  /* create current and previous lines list */
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&(ccstate->prev_run_line), dgp_state));
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&(ccstate->cur_run_line), dgp_state));
  
  /* add current run to current run line */
  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(ccstate->cur_run_line, (void *)cur_run_index, NULL));
  cur_run->ccstate = ccstate;

  ccstate->map_index = CCCMAP_INVALID_INDEX;
  ccstate_setActive(ccstate);
  ccstate->ulX = cur_run->xpos;
  ccstate->ulY = ccstate->lrY = (Dgp_UInt16)cookie->curLine;
  ccstate->lrX = cur_run->xpos + 
                 cur_run->cccrun.length - 1;

  ccstate->area = cur_run->cccrun.length;

  if(!run_isSampled(cur_run)) {

    ccstate->color = cur_run->cccrun.colorInfo.rgbColor;
    ccstate_setNormal(ccstate);

  } else {

    ccstate->color = DGPCC_CCSTATE_SAMPLED_COLOR;
    ccstate_setSampled(ccstate);

  }

  *_ccstate = ccstate;
  ccstate = NULL;

  NORMAL_RETURN;
ERR_CLEANUP:

  ccstate_cleanup(&ccstate, cookie);

  ERR_RETURN;
}

void
ccstate_cleanup(CCstate **_ccstate, 
                colorCC_state *cookie)
{
  CCstate *ccstate;
  Dgp_Free tmp_free = cookie->DGP_FREE;
  DgpAdt_ListNode ln;

  if(_ccstate == NULL) return;

  ccstate = *_ccstate;
  
  if(ccstate == NULL) return;
  *_ccstate = NULL;
  
  /* free semi-compact runs */
  if(ccstate->sc_lines != NULL) {
    DGPADT_FOREACH(ln, ccstate->sc_lines) {
      scrs_cleanup(&(DGPADT_LIGET(ln)), (Dgp_Bool)ccstate_isWashOrSampled(ccstate), ccstate_isSampled(ccstate), 
        cookie->dgp_state);
    }
    DgpAdt_liFree(ccstate->sc_lines);
  }

  if(ccstate->prev_run_line != NULL) {
    DgpAdt_liFree(ccstate->prev_run_line);
  }

  if(ccstate->cur_run_line != NULL) {
    DgpAdt_liFree(ccstate->cur_run_line);
  }

  tmp_free(ccstate);
}



static INLINE ErrCode
scrs_merge(sc_run_seg *scr_seg1, 
           sc_run_seg *scr_seg2, 
           Dgp_Bool use_scxr_segs,
           Dgp_Bool has_sampled_data,
           Dgp_State *dgp_state,
           void **ret_scr_seg) 
{
  DGP_PROLOGUE("scrs_merge()", NULL);
  Dgp_UInt16 tmp1, tmp2;
  sc_run_seg *first_scr_seg = NULL;
  sc_run_seg *tmp_scr_seg, **pnext_scr_seg = &first_scr_seg;
  sc_run *pruns1, *pruns2, *tmp_pruns;
  sc_run *pruns1_end, *pruns2_end, *tmp_pruns_end;
  Dgp_UInt32 *puser1 = NULL, *puser2 = NULL, *tmp_puser = NULL;
  Dgp_Bool free_old_segs = DGP_TRUE;
  Dgp_Int32 scrs_pool_id;
  Dgp_UInt32 dummy_userdata[SC_RUN_NUM_RUNS];
  
  
  /* determine the correct allocators to use */
  scrs_pool_id = ((!use_scxr_segs) ? SCRS_POOL_ID : SCXRS_POOL_ID);

  do {

    /* compute last x value for each segment */
    tmp1 = scrs_lastX(scr_seg1);
    tmp2 = scrs_lastX(scr_seg2);

    if(scrs_noOverlap(scr_seg1->runs[0].xpos, tmp1, scr_seg2->runs[0].xpos, tmp2)) {

      /* no overlap, that makes it easy */
      if(scr_seg1->runs[0].xpos < scr_seg2->runs[0].xpos) {

        /* save current */
        tmp_scr_seg = scr_seg1;

        /* move to next */
        scr_seg1 = scr_seg1->next;
        
      } else {

        /* save current */
        tmp_scr_seg = scr_seg2;

        /* move to next */
        scr_seg2 = scr_seg2->next;

      }

      *pnext_scr_seg = tmp_scr_seg;
      tmp_scr_seg->next = NULL;
      pnext_scr_seg = (sc_run_seg **)(&(tmp_scr_seg->next));

    } else {
            
      /* bummer there's some overlap here... */
      tmp_pruns_end = tmp_pruns = NULL;
      pruns1 = scr_seg1->runs;
      pruns1_end = pruns1 + scr_seg1->num_runs;
      pruns2 = scr_seg2->runs;
      pruns2_end = pruns2 + scr_seg2->num_runs;

      /* get userdata pointers if these are extended runs */
      if(use_scxr_segs) {
        puser1 = ((sc_xrun_seg *)scr_seg1)->userdata;
        puser2 = ((sc_xrun_seg *)scr_seg2)->userdata;
      } else {
        puser1 = puser2 = tmp_puser = dummy_userdata;
      }

      /* merge sort the two existing segments into the new segment(s) */
      do {

        if(tmp_pruns_end == tmp_pruns) {

          /* ...create another */
          SCRS_SEC_MALLOC_WID(scrs_pool_id, &tmp_scr_seg);
          scrs_initSeg(tmp_scr_seg, 0);
          tmp_pruns = tmp_scr_seg->runs;
          tmp_pruns_end = tmp_pruns + SC_RUN_NUM_RUNS;
          if(!use_scxr_segs) {
            tmp_puser = dummy_userdata;
          } else {
            tmp_puser = ((sc_xrun_seg *)tmp_scr_seg)->userdata;
          }
        
          /* add it to the list... */
          *pnext_scr_seg = tmp_scr_seg;
          pnext_scr_seg = (sc_run_seg **)(&(tmp_scr_seg->next));

        }

        if(pruns1->xpos < pruns2->xpos) {
          DGP_ASSERT((pruns1->xpos + pruns1->length) <= pruns2->xpos);
          *(tmp_pruns++) = *(pruns1++);
          *(tmp_puser++) = *(puser1++);
        } else {
          DGP_ASSERT((pruns2->xpos + pruns2->length) <= pruns1->xpos);
          *(tmp_pruns++) = *(pruns2++);
          *(tmp_puser++) = *(puser2++);
        }

        tmp_scr_seg->num_runs++;

      } while((pruns1 < pruns1_end) && (pruns2 < pruns2_end));

      /* clean up remaining old segments */
      if(pruns2 == pruns2_end) {

        /* get rid of used up segment */
        tmp_scr_seg = scr_seg2;
        scr_seg2 = scr_seg2->next;
        SCRS_FREE_WID(scrs_pool_id, tmp_scr_seg);

        /* get current state of other segment */
        scr_seg1->num_runs = pruns1_end - pruns1;
        tmp_scr_seg = scr_seg1;
        tmp_pruns = pruns1;
        tmp_pruns_end = pruns1_end;
        tmp_puser = puser1;

       } else {
        
        /* get rid of used up segment */
        tmp_scr_seg = scr_seg1;
        scr_seg1 = scr_seg1->next;
        SCRS_FREE_WID(scrs_pool_id, tmp_scr_seg);

        /* get current state of other segment */
        scr_seg2->num_runs = pruns2_end - pruns2;
        tmp_scr_seg = scr_seg2;
        tmp_pruns = pruns2;
        tmp_pruns_end = pruns2_end;
        tmp_puser = puser2;

      }

      /* fix other segment */
      pruns1 = tmp_scr_seg->runs;
      if(!use_scxr_segs) {

        while(tmp_pruns < tmp_pruns_end) {
          *(pruns1++) = *(tmp_pruns++);
        }

      } else {

        puser1 = ((sc_xrun_seg *)tmp_scr_seg)->userdata;
        while(tmp_pruns < tmp_pruns_end) {
          *(pruns1++) = *(tmp_pruns++);
          *(puser1++) = *(tmp_puser++);
        }

      }

    }

  } while((scr_seg1 != NULL) && (scr_seg2 != NULL));

  /* next, tack on any leftovers (tmp_scr_seg keeps a pointer to the first
     segment of the old list which is tacked on to the end of the new list,
     for the next step) */
  if(scr_seg1 != NULL) {
    *pnext_scr_seg = tmp_scr_seg = scr_seg1;
  } else if(scr_seg2 != NULL) {
    *pnext_scr_seg = tmp_scr_seg = scr_seg2;
  } else {
    tmp_scr_seg = NULL;
  }
  
  free_old_segs = DGP_FALSE;

  /* finally, do a quick run through the new runs and try to compact them,
     (we only go up to the last segment that was merged, anything tacked on
      to the end in the previous step we skip, assuming that it's already
      been compacted). */
  scr_seg1 = first_scr_seg;
  scr_seg2 = scr_seg1->next;
  while((scr_seg2 != NULL) && (scr_seg1 != tmp_scr_seg)) {

    if((scr_seg1->num_runs + ((sc_run_seg *)scr_seg1->next)->num_runs) <= SC_RUN_NUM_RUNS) {
      
      /* combine the next segment with the current */
      pruns1 = scr_seg1->runs + scr_seg1->num_runs;
      pruns2 = scr_seg2->runs;
      tmp_pruns = pruns2 + scr_seg2->num_runs;
      if(!use_scxr_segs) {
        
        while(pruns2 < tmp_pruns) {
          *(pruns1++) = *(pruns2++);
        }

      } else {

        puser1 = ((sc_xrun_seg *)scr_seg1)->userdata + scr_seg1->num_runs;
        puser2 = ((sc_xrun_seg *)scr_seg2)->userdata;
        while(pruns2 < tmp_pruns) {
          *(pruns1++) = *(pruns2++);
          *(puser1++) = *(puser2++);
        }

      }

      /* update count */
      scr_seg1->num_runs += scr_seg2->num_runs;
    
      /* remove the next segment */
      scr_seg1->next = scr_seg2->next;
      SCRS_FREE_WID(scrs_pool_id, scr_seg2);
   
    } else {

      /* move current segment pointer */
      scr_seg1 = scr_seg2;

    }

    /* get next segment pointer */
    scr_seg2 = scr_seg1->next;
  }

  /* return the first segment of the new segment list */
  *ret_scr_seg = first_scr_seg;
  first_scr_seg = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  /* cleanup our memory */
  if(free_old_segs) {
    scrs_cleanup((void **)&scr_seg1, use_scxr_segs, has_sampled_data, dgp_state);
    scrs_cleanup((void **)&scr_seg2, use_scxr_segs, has_sampled_data, dgp_state);
  }
  scrs_cleanup((void **)&first_scr_seg, use_scxr_segs, has_sampled_data, dgp_state);

  ERR_RETURN;
}



#define PREV_RUN_LINE 0
#define CUR_RUN_LINE 1


/* merge two cc_states together */
ErrCode
ccstate_merge(CCstate *ccstate_dst, 
              CCstate **_ccstate_src, 
              Run *_cur_row_runs,
              Run *_prev_row_runs,
              colorCC_state *cookie)
{
  DGP_PROLOGUE("ccstate_merge()", cookie->DGP_PERRSTATE);

  Dgp_Int32 i, j;
  DgpAdt_List dst_sc_list, src_sc_list;
  DgpAdt_ListNode ln1, ln2;
  DgpAdt_List dst_run_lines = DGP_NULLLIST;
  DgpAdt_List src_run_lines = DGP_NULLLIST;
  Run *cur_row_runs = NULL;
  void *tmpptr = NULL, *cur_seg1, *cur_seg2;
  CCstate *ccstate_src = *_ccstate_src;
  Dgp_Bool use_scxr_segs = ccstate_isWashOrSampled(ccstate_dst);
  Dgp_Bool has_sampled_data = ccstate_isSampled(ccstate_dst);



  /* check out some assumptions */
  DGP_ASSERT(ccstate_dst != ccstate_src);
  DGP_ASSERT(ccstate_isSampled(ccstate_dst) == ccstate_isSampled(ccstate_src));
  DGP_ASSERT(ccstate_isWash(ccstate_dst) == ccstate_isWash(ccstate_src));

  /* mine! */
  *_ccstate_src = NULL;

  /* if either one of these is background, we can cheat */
  if(ccstate_isBackground(ccstate_dst) || ccstate_isBackground(ccstate_src)) {

    if(!ccstate_isBackground(ccstate_src)) {
      ccstate_makeBackground(ccstate_src, cookie);
    } else if(!ccstate_isBackground(ccstate_dst)) {
      ccstate_makeBackground(ccstate_dst, cookie);
    }

    /* now cheat on the cur and prev line merging (these don't need to be
       in order, we just need to update the pointers) */

    /* correct ccstate for prev line pointers */
    DGPADT_FOREACH(ln1, ccstate_src->prev_run_line) {
      _prev_row_runs[((Dgp_Int32)(DGPADT_LIGET(ln1)))].ccstate = ccstate_dst;
    }

    DgpAdt_liApp(ccstate_dst->prev_run_line, ccstate_src->prev_run_line);

    ccstate_src->prev_run_line = DGP_NULLLIST;

    /* correct ccstate for cur line pointers */
    DGPADT_FOREACH(ln1, ccstate_src->cur_run_line) {
      _cur_row_runs[((Dgp_Int32)(DGPADT_LIGET(ln1)))].ccstate = ccstate_dst;
    }

    DgpAdt_liApp(ccstate_dst->cur_run_line, ccstate_src->cur_run_line);

    ccstate_src->cur_run_line = DGP_NULLLIST;

    /* now skip ahead */
    goto DONE_MERGING_BACKGROUNDS;
  }


  /* first, take care of the semi-compact run images if necessary */
  if(DGPADT_LILEN(ccstate_src->sc_lines) > 0) {

    if(DGPADT_LILEN(ccstate_dst->sc_lines) > 0) {

      /* determine src and dst lists for merged sc_runs */
      if(ccstate_dst->ulY <= ccstate_src->ulY) {

        j = ccstate_src->ulY - ccstate_dst->ulY;
        
        dst_sc_list = ccstate_dst->sc_lines;
        src_sc_list = ccstate_src->sc_lines;

      } else {

        j = ccstate_dst->ulY - ccstate_src->ulY;

        dst_sc_list = ccstate_src->sc_lines;
        src_sc_list = ccstate_dst->sc_lines;
        
      }

      ln1 = DGPADT_LIFIRST(dst_sc_list);
      ln2 = DGPADT_LIFIRST(src_sc_list);

      /* skip over non-overlapping sc_runs in dst list */
      for(i = 0; i < j; i++) {
        ln1 = DGPADT_LINEXT(ln1);
      }

      i = DGPADT_LILEN(dst_sc_list) - j;
      j = MIN((Dgp_Int32)DGPADT_LILEN(src_sc_list), i);
    
      /* now, merge the overlapping sc_runs */
      for(i = 0; i < j; i++) {
      
        DGP_ASSERT(DGPADT_LIGET(ln1) != NULL);
        DGP_ASSERT(DGPADT_LIGET(ln2) != NULL);

        /* remove the current lines from each ccstate */
        cur_seg1 = DGPADT_LIGET(ln1);
        DGPADT_LIPUT(ln1, NULL);
        cur_seg2 = DGPADT_LIGET(ln2);
        DGPADT_LIPUT(ln2, NULL);


#ifdef DGP_DBUG
        {
          Dgp_Int32 num1, num2, num_ret;
          void *tmp_seg;

          num1 = 0;
          num2 = 0;

          tmp_seg = cur_seg1;
          while(tmp_seg != NULL) {
            num1 += ((sc_run_seg *)tmp_seg)->num_runs;
            tmp_seg = ((sc_run_seg *)tmp_seg)->next;
          }

          tmp_seg = cur_seg2;
          while(tmp_seg != NULL) {
            num2 += ((sc_run_seg *)tmp_seg)->num_runs;
            tmp_seg = ((sc_run_seg *)tmp_seg)->next;
          }
#endif

        /* merge the current lines */
        tmpptr = NULL;
        DGP_ERR_PASS_CLEANUP(scrs_merge((sc_run_seg *)cur_seg1, (sc_run_seg *)cur_seg2, 
          use_scxr_segs, has_sampled_data, cookie->dgp_state, &tmpptr));
      
        
#ifdef DGP_DBUG
          
          num_ret = 0;

          tmp_seg = tmpptr;
          while(tmp_seg != NULL) {
            num_ret += ((sc_run_seg *)tmp_seg)->num_runs;
            tmp_seg = ((sc_run_seg *)tmp_seg)->next;
          }

          if((num1 + num2) != num_ret) {
            DGP_ASSERT((num1 + num2) == num_ret);
          }
        }
#endif

        /* add to dst list */
        DGPADT_LIPUT(ln1, tmpptr);

        /* move to next line */
        ln1 = DGPADT_LINEXT(ln1);
        ln2 = DGPADT_LINEXT(ln2);

      }

      /* finally, tack on any non-overlapping lines at the bottom */
      while(ln2 != NULL) {
        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(dst_sc_list, DGPADT_LIGET(ln2), NULL));
        DGPADT_LIPUT(ln2, NULL);
        ln2 = DGPADT_LINEXT(ln2);
      }

      /* now, add the new sc_list to ccstate_dst */
      ccstate_dst->sc_lines = dst_sc_list;
      dst_sc_list = DGP_NULLLIST;

      /* free old list */
      DgpAdt_liFree(src_sc_list);
      ccstate_src->sc_lines = DGP_NULLLIST;
      src_sc_list = DGP_NULLLIST;

    } else {

      /* oops, ccstate_dst doesn't have semi-compact runs, that makes it easy */
      DgpAdt_liFree(ccstate_dst->sc_lines);
      ccstate_dst->sc_lines = ccstate_src->sc_lines;
      ccstate_src->sc_lines = DGP_NULLLIST;
  
    }

  }

  /* get references to the prev row info */
  dst_run_lines = ccstate_dst->prev_run_line;
  ccstate_dst->prev_run_line = DGP_NULLLIST;

  src_run_lines = ccstate_src->prev_run_line;
  ccstate_src->prev_run_line = DGP_NULLLIST;

  cur_row_runs = _prev_row_runs;

  /* now, combine runs in current and previous run lists */
  for(i = 0; i < 2; i++) {

    if(DGPADT_LILEN(src_run_lines) > 0) {

      /* correct ccstate pointers */
      DGPADT_FOREACH(ln1, src_run_lines) {
        cur_row_runs[((Dgp_Int32)(DGPADT_LIGET(ln1)))].ccstate = ccstate_dst;
      }

      if(DGPADT_LILEN(dst_run_lines) > 0) {
      
        /* merge the two lists */
        ln1 = DGPADT_LIFIRST(dst_run_lines);
        ln2 = DGPADT_LIFIRST(src_run_lines);

        while((ln1 != DGP_NULLLISTNODE) && (ln2 != DGP_NULLLISTNODE)) {

          if(cur_row_runs[((Dgp_Int32)DGPADT_LIGET(ln2))].xpos > 
             cur_row_runs[((Dgp_Int32)DGPADT_LIGET(ln1))].xpos) {
            ln1 = DGPADT_LINEXT(ln1);
          } else {
            DGP_ERR_PASS_CLEANUP(DgpAdt_liAddAfter(dst_run_lines, DGPADT_LIPREV(ln1), 
              DGPADT_LIGET(ln2), NULL));
            ln2 = DGPADT_LINEXT(ln2);
          }

        }

        /* tack the rest of the list on... */
        while(ln2 != DGP_NULLLISTNODE) {
          DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(dst_run_lines, DGPADT_LIGET(ln2), NULL));
          ln2 = DGPADT_LINEXT(ln2);
        }

      } else {

        /* just move over the new list */
        tmpptr = dst_run_lines;
        dst_run_lines = src_run_lines;
        src_run_lines = (DgpAdt_List)tmpptr;

      }
    }

    /* free src list */
    DgpAdt_liFree(src_run_lines);
    src_run_lines = DGP_NULLLIST;
    
    if(i == PREV_RUN_LINE) {
      
      /* save new prev row info */
      ccstate_dst->prev_run_line = dst_run_lines;
      
      /* get current row info */
      dst_run_lines = ccstate_dst->cur_run_line;
      ccstate_dst->cur_run_line = DGP_NULLLIST;

      src_run_lines = ccstate_src->cur_run_line;
      ccstate_src->cur_run_line = DGP_NULLLIST;

      cur_row_runs = _cur_row_runs;

    } else {

      /* save new current row info */
      ccstate_dst->cur_run_line = dst_run_lines;
      dst_run_lines = DGP_NULLLIST;

    }

  }


  /* if we are merging backgrounds, we can skip the complicated algorithms above
     and come directly here */
DONE_MERGING_BACKGROUNDS:


  /* now, adjust the bounds */
  ccstate_dst->ulX = MIN(ccstate_dst->ulX, ccstate_src->ulX);
  ccstate_dst->ulY = MIN(ccstate_dst->ulY, ccstate_src->ulY);
  ccstate_dst->lrX = MAX(ccstate_dst->lrX, ccstate_src->lrX);
  ccstate_dst->lrY = MAX(ccstate_dst->lrY, ccstate_src->lrY);

  /* adjust area */
  ccstate_dst->area += ccstate_src->area;

  /* this should be equal... */
  DGP_ASSERT(ccstate_isBackground(ccstate_dst) == 
             ccstate_isBackground(ccstate_src));

  /* also, merge the info in the neighbor map */
  cccmap_combine(cookie->cccmap, ccstate_dst->map_index, ccstate_src->map_index);
  
  /* finally, get rid of the old ccstate */
  ccstate_cleanup(&ccstate_src, cookie);


  NORMAL_RETURN;

ERR_CLEANUP:

  if(dst_run_lines != DGP_NULLLIST) DgpAdt_liFree(dst_run_lines);

  if(src_run_lines != DGP_NULLLIST) DgpAdt_liFree(src_run_lines);

  if(ccstate_src != NULL) ccstate_cleanup(&ccstate_src, cookie);

  ERR_RETURN;
}


ErrCode
ccstate_makeBackground(CCstate *ccstate, 
                       colorCC_state *cookie)
{
  Dgp_State *dgp_state = cookie->dgp_state;
  DGP_PROLOGUE("ccstate_makeBackground()", DGP_PERRSTATE);
  sc_run_seg *cur_seg, *tmp_seg;
  DgpAdt_ListNode ln;


  /* this should only ever happen to solid color ccstates */
  DGP_ASSERT(!ccstate_isWashOrSampled(ccstate));
  
  /* set background flag */
  ccstate_setBackground(ccstate);

  /* update wash counts */
  cccmap_newBackground(cookie->cccmap, ccstate->map_index);
  ccstate->num_potential_washes = 0;

  /* ditch all the sc_runs */
  if(DGPADT_LILEN(ccstate->sc_lines) > 0) {
    
    DGPADT_FOREACH(ln, ccstate->sc_lines) {
      cur_seg = (sc_run_seg *)DGPADT_LIGET(ln);
      DGPADT_LIPUT(ln, NULL);

      while(cur_seg != NULL) {
        tmp_seg = cur_seg->next;
        SCRS_FREE(cur_seg);
        cur_seg = tmp_seg;
      }
    }

    /* ditch the sc list */
    DgpAdt_liFree(ccstate->sc_lines);
  
    /* replace with empty list */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&(ccstate->sc_lines), cookie->dgp_state));

  }

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}



static INLINE ErrCode
scrs_convertToWash(sc_run_seg *scr_seg, 
                   Dgp_UInt32 scr_seg_color,
                   void **ret_scxr_seg,
                   Dgp_State *dgp_state) 
{
  DGP_PROLOGUE("scrs_convertToWash()", NULL);
  
  Dgp_UInt32 *userdata;
  sc_xrun_seg *first_scxr_seg = NULL;
  sc_xrun_seg **pnext_scxr_seg = &first_scxr_seg;  
  void *tmp_seg;

  while(scr_seg != NULL) {
  
    /* create new sc_xrun_seg */
    SCXRS_SEC_MALLOC(&tmp_seg);

    /* copy sc_run_seg data to sc_xrun_seg */
    *((sc_run_seg *)tmp_seg) = *scr_seg;
    ((sc_run_seg *)tmp_seg)->next = NULL;

    /* make sure the code below works... */
    DGP_ASSERT(SC_RUN_NUM_RUNS == 4);

    /* set the userdata fields for each seg to the color of this seg */
    userdata = ((sc_xrun_seg *)tmp_seg)->userdata;
    switch(((sc_xrun_seg *)tmp_seg)->num_runs) {
    case 4:
      *(userdata++) = scr_seg_color;
      /* drop through */
    case 3:
      *(userdata++) = scr_seg_color;
      /* drop through */
    case 2:
      *(userdata++) = scr_seg_color;
      /* drop through */
    case 1:
      *userdata = scr_seg_color;
      break;
    }

    /* add this sc_xrun_seg to the new list */
    *pnext_scxr_seg = tmp_seg;
    pnext_scxr_seg = (sc_xrun_seg **)&(((sc_xrun_seg *)tmp_seg)->next);
    tmp_seg = NULL;

    /* remove the old sc_run_seg */
    tmp_seg = scr_seg->next;
    SCRS_FREE(scr_seg);
    scr_seg = tmp_seg;
  
  }

  *ret_scxr_seg = first_scxr_seg;
  first_scxr_seg = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  while(first_scxr_seg != NULL) {
    tmp_seg = first_scxr_seg->next;
    SCXRS_FREE(first_scxr_seg);
    first_scxr_seg = tmp_seg;
  }

  while(scr_seg != NULL) {
    tmp_seg = scr_seg->next;
    SCRS_FREE(scr_seg);
    scr_seg = tmp_seg;
  }

  ERR_RETURN;
}


ErrCode
ccstate_makeWash(CCstate *ccstate,
                      colorCC_state *cookie)
{
  DGP_PROLOGUE("ccstate_makeWash()", NULL);
  Dgp_State *dgp_state = cookie->dgp_state;
  DgpAdt_List cur_lines = ccstate->sc_lines, new_lines;
  DgpAdt_ListNode ln;
  sc_run_seg *scr_seg;
  sc_xrun_seg *scxr_seg = NULL;
  void *tmp_seg;


  /* set other ccstate info to indicate that it is a wash */
  ccstate_setWash(ccstate);
  
  /* basically, we just need to convert the sc_runs to sc_xruns */
  if(DGPADT_LILEN(cur_lines) > 0) {

    ccstate->sc_lines = DGP_NULLLIST;

    /* create new list */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&new_lines, dgp_state));
    ccstate->sc_lines = new_lines;

    DGPADT_FOREACH(ln, cur_lines) {
      scr_seg = (sc_run_seg *)DGPADT_LIGET(ln);
      DGPADT_LIPUT(ln, NULL);

      /* convert this line of sc_runs to sc_xruns */
      DGP_ERR_PASS_CLEANUP(scrs_convertToWash(scr_seg, ccstate->color,
					      (void **)&scxr_seg, dgp_state));

      /* add new line of sc_xruns to new list */
      DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(new_lines, scxr_seg, NULL));
      scxr_seg = NULL;
    }

    /* ditch the old list */
    DgpAdt_liFree(cur_lines);
    cur_lines = DGP_NULLLIST;

  }

  /* clear the old color value */
  ccstate->color = 0;

  NORMAL_RETURN;

ERR_CLEANUP:

  while(scxr_seg != NULL) {
    tmp_seg = scxr_seg->next;
    SCXRS_FREE(scxr_seg);
    scxr_seg = tmp_seg;
  }

  if(cur_lines != DGP_NULLLIST) {
    DGPADT_FOREACH(ln, cur_lines) {
      scr_seg = (sc_run_seg *)DGPADT_LIGET(ln);
      while(scr_seg != NULL) {
	      tmp_seg = scr_seg->next;
	      SCRS_FREE(scr_seg);
	      scr_seg = tmp_seg;
      }
    }
    DgpAdt_liFree(cur_lines);
  }

  ERR_RETURN;
}


ErrCode
ccstate_convertToColorCC(CCstate *ccstate, 
                         DgpCC_ColorCC **_new_cc, 
                         colorCC_state *cookie)
{
  DGP_PROLOGUE("ccstate_convertToColorCC()", NULL);
  Dgp_State *dgp_state = cookie->dgp_state;
  DgpAdt_ListNode ln;
  Dgp_Int32 num_runs;
  sc_run_seg *scr_seg = NULL, *tmp_seg;
  Dgp_UInt32 *userdata;
  sc_run *sc_runs;
  Dgp_UInt16 ypos, *run_list, xoff, xpos;
  DgpImage_RGB *im_row, *im_ptr, *end_ptr;
  DgpImage_RGB wash_col;
  DgpCC_ColorCC *new_cc = NULL;;


  /* valid ccstate? */
  DGP_ASSERT(ccstate != NULL);

  /* we should never get background ccstates */
  DGP_ASSERT(!ccstate_isBackground(ccstate));

  /* create the new cc */
  COLORCC_SEC_MALLOC(&new_cc);
  memset(new_cc, 0, sizeof(DgpCC_ColorCC));

  /* copy the easy stuff */
  new_cc->ulX = ccstate->ulX;
  new_cc->ulY = ccstate->ulY;
  new_cc->lrX = ccstate->lrX;
  new_cc->lrY = ccstate->lrY;

  /* set state flags */
  new_cc->state.sampled = ccstate_isSampled(ccstate);
  new_cc->state.wash = ccstate_isWash(ccstate);

  /* there shouldn't be anything in these two lists */
  DGP_ASSERT(DGPADT_LILEN(ccstate->cur_run_line) == 0);
  DGP_ASSERT(DGPADT_LILEN(ccstate->prev_run_line) == 0);

  /* there had better be something here... */
  DGP_ASSERT(DGPADT_LILEN(ccstate->sc_lines) > 0);
  
  /* now, count the number of runs in this ccstate */
  num_runs = 0;
  DGPADT_FOREACH(ln, ccstate->sc_lines) {
    tmp_seg = (sc_run_seg *)DGPADT_LIGET(ln);

    while(tmp_seg != NULL) {
      num_runs += tmp_seg->num_runs;
      tmp_seg = tmp_seg->next;
    }
  }

  /* create run list */
  DGP_SEC_MALLOC(run_list, (Dgp_UInt16 *), (sizeof(Dgp_UInt16) * 2 * num_runs));
  new_cc->runList = run_list;
  new_cc->numRuns = num_runs;

  ypos = 0;
  xoff = ccstate->ulX;

  /* copy over the color info (either a single value or an image */
  if(ccstate_isNormal(ccstate)) {

    /* this is a solid color ccstate */

    /* copy the color value */
    lrgb_copyToSampling(new_cc->colorInfo.rgbColor, ccstate->color);

    /* fill in run list */
    DGPADT_FOREACH(ln, ccstate->sc_lines) {
      scr_seg = (sc_run_seg *)DGPADT_LIGET(ln);
      DGPADT_LIPUT(ln, NULL);

      while(scr_seg != NULL) {
        
        /* if this value changes this code below breaks... */
        DGP_ASSERT(SC_RUN_NUM_RUNS == 4);
        sc_runs = scr_seg->runs;

        /* copy the image info */
        switch(scr_seg->num_runs) {
        case 4:

          /* run info */
          *(run_list++) = sc_runs->xpos - xoff;
          *(run_list++) = sc_runs->length;
          sc_runs++;

          /* drop through */
        case 3:

          /* run info */
          *(run_list++) = sc_runs->xpos - xoff;
          *(run_list++) = sc_runs->length;
          sc_runs++;

          /* drop through */
        case 2:

          /* run info */
          *(run_list++) = sc_runs->xpos - xoff;
          *(run_list++) = sc_runs->length;
          sc_runs++;

          /* drop through */
        case 1:

          /* run info */
          *(run_list++) = sc_runs->xpos - xoff;
          *(run_list++) = sc_runs->length;

          break;
        }

        /* move to next segment */
        tmp_seg = scr_seg;
        scr_seg = scr_seg->next;

        /* free old seg */
        SCRS_FREE(tmp_seg);
      }

    }
  
  } else {

    /* create an image */
#ifdef DGP_DBUG
    DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_RGB,
				      ccstate->lrX - ccstate->ulX + 1,
				      ccstate->lrY - ccstate->ulY + 1,
				      (void **)&(new_cc->colorInfo.rgbIm)));
#else
    DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_RGB | DGPIMAGE_DONTCLEAR,
				      ccstate->lrX - ccstate->ulX + 1,
				      ccstate->lrY - ccstate->ulY + 1,
				      (void **)&(new_cc->colorInfo.rgbIm)));
#endif

    if(ccstate_isSampled(ccstate)) {

      /* fill in run list and sampled image values */
      DGPADT_FOREACH(ln, ccstate->sc_lines) {
        scr_seg = (sc_run_seg *)DGPADT_LIGET(ln);
        DGPADT_LIPUT(ln, NULL);
        
        im_row = DGPIMG_PIXPTR(new_cc->colorInfo.rgbIm, 0, ypos);

        while(scr_seg != NULL) {
          
          /* if this value changes this code below breaks... */
          DGP_ASSERT(SC_RUN_NUM_RUNS == 4);
          userdata = ((sc_xrun_seg *)scr_seg)->userdata;
          sc_runs = scr_seg->runs;

          /* copy the image info and run info */
          switch(scr_seg->num_runs) {
          case 4:

            /* image info */
            xpos = sc_runs->xpos - xoff;
            memcpy((im_row + xpos), ((DgpImage_RGB *)(*userdata)), 
                   (sc_runs->length * sizeof(DgpImage_RGB)));
            DGP_FREE((void *)(*userdata));
            *userdata++ = 0;

            /* run info */
            *(run_list++) = xpos;
            *(run_list++) = sc_runs->length;
            sc_runs++;

            /* drop through */
          case 3:

            /* image info */
            xpos = sc_runs->xpos - xoff;
            memcpy((im_row + xpos), ((DgpImage_RGB *)(*userdata)), 
                   (sc_runs->length * sizeof(DgpImage_RGB)));
            DGP_FREE((void *)(*userdata));
            *userdata++ = 0;

            /* run info */
            *(run_list++) = xpos;
            *(run_list++) = sc_runs->length;
            sc_runs++;

            /* drop through */
          case 2:

            /* image info */
            xpos = sc_runs->xpos - xoff;
            memcpy((im_row + xpos), ((DgpImage_RGB *)(*userdata)), 
                   (sc_runs->length * sizeof(DgpImage_RGB)));
            DGP_FREE((void *)(*userdata));
            *userdata++ = 0;

            /* run info */
            *(run_list++) = xpos;
            *(run_list++) = sc_runs->length;
            sc_runs++;

            /* drop through */
          case 1:

            /* image info */
            xpos = sc_runs->xpos - xoff;
            memcpy((im_row + xpos), ((DgpImage_RGB *)(*userdata)), 
                   (sc_runs->length * sizeof(DgpImage_RGB)));
            DGP_FREE((void *)(*userdata));
            *userdata = 0;

            /* run info */
            *(run_list++) = xpos;
            *(run_list++) = sc_runs->length;

            break;
          }

          /* move to next segment */
          tmp_seg = scr_seg;
          scr_seg = scr_seg->next;

          /* free old seg */
          SCXRS_FREE(tmp_seg);
        }

        ypos++;
      }

    } else {

      /* this ccstate is a wash... */

      Dgp_Int32 cur_xpos;

      /* fill in run list and wash image values */
      DGPADT_FOREACH(ln, ccstate->sc_lines) {
        scr_seg = (sc_run_seg *)DGPADT_LIGET(ln);
        DGPADT_LIPUT(ln, NULL);
        
        im_row = DGPIMG_PIXPTR(new_cc->colorInfo.rgbIm, 0, ypos);
        
        /* things are done slightly differently for washes because the
           runs can be touching, in which case we need to combine them,
           otherwise things further down the road will go puke! */
        cur_xpos = -1;

        while(scr_seg != NULL) {
          
          /* if this value changes this code below breaks... */
          DGP_ASSERT(SC_RUN_NUM_RUNS == 4);
          userdata = ((sc_xrun_seg *)scr_seg)->userdata;
          sc_runs = scr_seg->runs;

          /* copy the image info */
          switch(scr_seg->num_runs) {
          case 4:

            /* image info */
            lrgb_copyToSampling(wash_col, *userdata);
            xpos = sc_runs->xpos - xoff;
            im_ptr = im_row + xpos;
            end_ptr = im_ptr + sc_runs->length;
            while(im_ptr < end_ptr) {
              *im_ptr++ = wash_col;
            }
            userdata++;

            /* run info */
            if(cur_xpos == (Dgp_Int32)xpos) {
              
              /* yes, this is legal! */
              run_list[-1] += sc_runs->length;

            } else {

              *(run_list++) = xpos;
              *(run_list++) = sc_runs->length;

            }
            cur_xpos = (Dgp_Int32)(xpos + sc_runs->length);
            sc_runs++;

            /* drop through */
          case 3:

            /* image info */
            lrgb_copyToSampling(wash_col, *userdata);
            xpos = sc_runs->xpos - xoff;
            im_ptr = im_row + xpos;
            end_ptr = im_ptr + sc_runs->length;
            while(im_ptr < end_ptr) {
              *im_ptr++ = wash_col;
            }
            userdata++;

            /* run info */
            if(cur_xpos == (Dgp_Int32)xpos) {
              
              /* yes, this is legal! */
              run_list[-1] += sc_runs->length;

            } else {

              *(run_list++) = xpos;
              *(run_list++) = sc_runs->length;

            }
            cur_xpos = (Dgp_Int32)(xpos + sc_runs->length);
            sc_runs++;

            /* drop through */
          case 2:

            /* image info */
            lrgb_copyToSampling(wash_col, *userdata);
            xpos = sc_runs->xpos - xoff;
            im_ptr = im_row + xpos;
            end_ptr = im_ptr + sc_runs->length;
            while(im_ptr < end_ptr) {
              *im_ptr++ = wash_col;
            }
            userdata++;

            /* run info */
            if(cur_xpos == (Dgp_Int32)xpos) {
              
              /* yes, this is legal! */
              run_list[-1] += sc_runs->length;

            } else {

              *(run_list++) = xpos;
              *(run_list++) = sc_runs->length;

            }
            cur_xpos = (Dgp_Int32)(xpos + sc_runs->length);
            sc_runs++;

            /* drop through */
          case 1:

            /* image info */
            lrgb_copyToSampling(wash_col, *userdata);
            xpos = sc_runs->xpos - xoff;
            im_ptr = im_row + xpos;
            end_ptr = im_ptr + sc_runs->length;
            while(im_ptr < end_ptr) {
              *im_ptr++ = wash_col;
            }

            /* run info */
            if(cur_xpos == (Dgp_Int32)xpos) {
              
              /* yes, this is legal! */
              run_list[-1] += sc_runs->length;

            } else {

              *(run_list++) = xpos;
              *(run_list++) = sc_runs->length;

            }
            cur_xpos = (Dgp_Int32)(xpos + sc_runs->length);
            break;
          }

          /* move to next segment */
          tmp_seg = scr_seg;
          scr_seg = scr_seg->next;

          /* free old seg */
          SCXRS_FREE(tmp_seg);
        }

        ypos++;

      }

      /* we may end up with a different number of runs than
         we originally anticipated for washes, so
         we're going to recompute this value */
      new_cc->numRuns = (run_list - new_cc->runList) / 2;

      if(new_cc->numRuns < num_runs) {

        /* reduce mem usage */
        DGP_SEC_REALLOC(new_cc->runList, (Dgp_UInt16 *), (sizeof(Dgp_UInt16) * 2 * new_cc->numRuns));

      }

    }

  }

  /* ditch run list */
  DgpAdt_liFree(ccstate->sc_lines);
  ccstate->sc_lines = DGP_NULLLIST;

  /* return the new cc */
  *_new_cc = new_cc;
  new_cc = NULL;

  
  NORMAL_RETURN;

ERR_CLEANUP:

  if(scr_seg != NULL) {
    scrs_cleanup((void **)&scr_seg, (Dgp_Bool)ccstate_isWashOrSampled(ccstate), 
      (Dgp_Bool)ccstate_isWash(ccstate), dgp_state);
  }

  if(new_cc != NULL) {
    if((new_cc->state.sampled) && (new_cc->colorInfo.rgbIm != NULL)) {
      DGPIMAGE_FREE(new_cc->colorInfo.rgbIm);
      new_cc->colorInfo.rgbIm = NULL;
    }

    if(new_cc->runList != NULL) {
      DGP_FREE(new_cc->runList);
      new_cc->runList = NULL;
    }

    COLORCC_FREE(new_cc);
  }

  ERR_RETURN;
}


/********
 * tempCCstate
 ********/
typedef struct {
  /* the connected component info for this ccstate */
  Dgp_UInt16 ulX, ulY, lrX, lrY;
  DgpAdt_List sc_lines;
  void **pnext_seg;

  /* color, for solid color ccstates, otherwise 0 */
  Dgp_UInt32 color;


} tempCCstate;

#define TCCS_POOL_ID DGP_SMALLPOOL_MIN20

#define TCCS_SEC_MALLOC(_pptccs) \
  DGP_ERR_PASS_CLEANUP(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, TCCS_POOL_ID, (void **)(_pptccs)))
#define TCCS_FREE(_ptccs) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, TCCS_POOL_ID, (void *)(_ptccs)))


/********
 * segListNode
 ********/
typedef struct _segListNode {
  struct _segListNode *next;
  sc_run_seg *seg;
  tempCCstate *tccstate;
} segListNode;

#define SEGLN_POOL_ID DGP_SMALLPOOL_MIN16

#define SEGLN_SEC_MALLOC(_ppsegln) \
  DGP_ERR_PASS_CLEANUP(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, SEGLN_POOL_ID, (void **)(_ppsegln)))
#define SEGLN_FREE(_psegln) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, SEGLN_POOL_ID, (void *)(_psegln)))


ErrCode
ccstate_disassembleSampledCC(CCstate *ccstate, 
                             colorCC_state *cookie)
{
  Dgp_State *dgp_state = cookie->dgp_state;
  DGP_PROLOGUE("ccstate_disassembleSampledCC()", DGP_PERRSTATE);
  DgpAdt_List new_tccstates = DGP_NULLLIST;
  tempCCstate *tccstate;
  DgpCC_ColorCC *new_cc = NULL;

  /* run info */
  Dgp_UInt32 seg_color;
  Dgp_UInt16 ypos;
  sc_run_seg *new_seg = NULL;
  
  /* seg info */
  sc_xrun_seg *cur_seg = NULL;
  Dgp_Int32 cur_pos;
  Dgp_UInt16 xoff;

  void *tmpptr;
  DgpAdt_ListNode cur_sc_line, ln, ln2;

  segListNode *segln, *cur_segln, *cur_seglist = NULL, *prev_seglist = NULL;
  segListNode **pcur_seglist_next = &cur_seglist;

  
  Dgp_Bool is_sampled = (Dgp_Bool)ccstate_isSampled(ccstate);


  /* validate smallpool usage */
  DGP_SMALLPOOL_VALIDATE(tempCCstate, TCCS_POOL_ID);
  DGP_SMALLPOOL_VALIDATE(segListNode, SEGLN_POOL_ID);

  /* create a list */
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&new_tccstates, dgp_state));

  /* to make the algorithm below simpler, we're going to add an extra 
     "empty" row onto the bottom of the ccstate's list of sc_lines */
  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(ccstate->sc_lines, NULL, NULL));

  ypos = ccstate->ulY;
  cur_sc_line = DGPADT_LIFIRST(ccstate->sc_lines);

  while(cur_sc_line != DGP_NULLLISTNODE) {

    /* get next line */
    cur_seg = (sc_xrun_seg *)DGPADT_LIGET(cur_sc_line);
    DGPADT_LIPUT(cur_sc_line, NULL);
    cur_pos = 0;
    xoff = 0;

    while(cur_seg != NULL) {

      /* create new sc_run_seg */
      SCRS_SEC_MALLOC(&new_seg);
      scrs_initSeg(new_seg, 1);

      /* create new seg list node */
      SEGLN_SEC_MALLOC(&cur_segln);
      cur_segln->next = NULL;
      cur_segln->tccstate = NULL;
      cur_segln->seg = new_seg;

      /* stick it on the current list */
      *pcur_seglist_next = cur_segln;
      pcur_seglist_next = &(cur_segln->next);

      /* now, read next "run" */
      if(is_sampled) {

        /* convert the sampled stuff to separate runs */

        new_seg->runs[0].xpos = cur_seg->runs[cur_pos].xpos + xoff;
        seg_color = ((Dgp_UInt32 *)cur_seg->userdata[cur_pos])[xoff++];
        new_seg->runs[0].length = 1;

        /* figure out how long this run is */
        while((xoff < cur_seg->runs[cur_pos].length) &&
              (((Dgp_UInt32 *)cur_seg->userdata[cur_pos])[xoff] == seg_color)) {
          xoff++;
          new_seg->runs[0].length++;
        }

        /* move to next run if we are at the end of this one */
        if(xoff == cur_seg->runs[cur_pos].length) {

          /* free sampled data */
          DGP_FREE((void *)(cur_seg->userdata[cur_pos]));
          cur_seg->userdata[cur_pos] = (Dgp_UInt32)NULL;

          cur_pos++;
          xoff = 0;
        }

      } else {

        /* convert the wash to separate runs */

        new_seg->runs[0].xpos = cur_seg->runs[cur_pos].xpos;
        seg_color = cur_seg->userdata[cur_pos];
        new_seg->runs[0].length = cur_seg->runs[cur_pos].length;

        /* move to next run */
        cur_pos++;

      }

      /* figure out which tempccstate to add this run to */
      segln = prev_seglist;
      while(segln != NULL) {
        
        if((scrs_touches(segln->seg->runs[0].xpos, 
                        (segln->seg->runs[0].xpos + segln->seg->runs[0].length),
                        cur_seg->runs[0].xpos,
                        (cur_seg->runs[0].xpos + cur_seg->runs[0].length))) &&
           (segln->tccstate->color == seg_color)) {

          tccstate = segln->tccstate;

          *(tccstate->pnext_seg) = (void *)new_seg;
          tccstate->pnext_seg = (void **)(&(new_seg->next));

          /* update x bounds */
          tccstate->ulX = MIN(tccstate->ulX, new_seg->runs[0].xpos); \
          tccstate->lrX = MAX((tccstate->lrX), (new_seg->runs[0].xpos + new_seg->runs[0].length - 1)); \

          new_seg = NULL;

          /* seg this seg's tccstate */
          cur_segln->tccstate = tccstate;

          /* we're done */
          break;
        }

        /* move to next segln */
        segln = segln->next;
      }

      /* didn't match anything? */
      if(new_seg != NULL) {
        
        /* create new tccstate */
        TCCS_SEC_MALLOC(&tccstate);
        tccstate->sc_lines = DGP_NULLLIST;
        tccstate->color = seg_color;

        /* set bounds */
        tccstate->ulX = new_seg->runs[0].xpos;
        tccstate->ulY = tccstate->lrY = ypos;
        tccstate->lrX = new_seg->runs[0].xpos + new_seg->runs[0].length - 1;

        /* add this tccstate to the list */
        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(new_tccstates, tccstate, NULL));

        /* add this seg to the tccstate */
        DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&(tccstate->sc_lines), dgp_state));
        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(tccstate->sc_lines, new_seg, NULL));
        tccstate->pnext_seg = (void **)(&(new_seg->next));
        new_seg = NULL;

        /* seg this seg's tccstate */
        cur_segln->tccstate = tccstate;

      }
      
      /* end of this seg? */
      if(cur_pos == cur_seg->num_runs) {

        /* move to next seg */
        tmpptr = cur_seg;
        cur_seg = cur_seg->next;

        /* free old seg */
        SCXRS_FREE(tmpptr);

        cur_pos = 0;
        xoff = 0;

      }

    }

    /* move all the temp ccstates to the next line */
    segln = cur_seglist;
    while(segln != NULL) {

      if(segln->tccstate->lrY == ypos) {

        /* move this tccstate to the next line */
        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(segln->tccstate->sc_lines, NULL, &ln2));
        segln->tccstate->pnext_seg = (void **)(&(DGPADT_LIGET(ln2)));
        
        segln->tccstate->lrY = ypos + 1;

      }

      segln = segln->next;
    }

    /* move to next line of original ccstate */
    cur_sc_line = DGPADT_LINEXT(cur_sc_line);
    ypos++;

    /* move cur seg list to prev */
    tmpptr = prev_seglist;
    prev_seglist = cur_seglist;
    cur_seglist = tmpptr;
    pcur_seglist_next = &cur_seglist;

    /* free old prev seg list */
    while(cur_seglist != NULL) {
      tmpptr = cur_seglist;
      cur_seglist = cur_seglist->next;

      SEGLN_FREE(tmpptr);
    }

  }

  /* these should both be empty when we're done */
  DGP_ASSERT(prev_seglist == NULL);
  DGP_ASSERT(cur_seglist == NULL);

  /* we're going to hijack this old ccstate */
  DgpAdt_liFree(ccstate->sc_lines);
  ccstate->sc_lines = DGP_NULLLIST;
  ccstate_clearSampled(ccstate);
  ccstate_clearWash(ccstate);
  ccstate_clearBackground(ccstate);
  ccstate_setNormal(ccstate);

  /* now make all of the tccstates into color_ccs */
  DGPADT_FOREACH(ln, new_tccstates) {

    tccstate = (tempCCstate *)DGPADT_LIGET(ln);
    DGPADT_LIPUT(ln, NULL);
    
    /* copy over the ccstate info */
    ccstate->ulX = tccstate->ulX;
    ccstate->ulY = tccstate->ulY;
    ccstate->lrX = tccstate->lrX;
    ccstate->color = tccstate->color;

    /* all of the tccstates have an extra line on their bottom,
       so make sure to account for that */
    ccstate->lrY = tccstate->lrY - 1;
    ccstate->sc_lines = tccstate->sc_lines;
    tccstate->sc_lines = NULL;
    DGP_ASSERT(DGPADT_LIGET(DGPADT_LILAST(ccstate->sc_lines)) == NULL);
    DgpAdt_liRem(DGPADT_LILAST(ccstate->sc_lines));

    /* ditch the tccstate */
    TCCS_FREE(tccstate);

    /* make a colorCC out of the ccstate */
    DGP_ERR_PASS_CLEANUP(ccstate_convertToColorCC(ccstate, &new_cc, cookie));

    /* add it to the new_ccs list */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(cookie->new_ccs, new_cc, NULL));
    new_cc = NULL;

  }

  /* final cleanup */
  DgpAdt_liFree(new_tccstates);
  new_tccstates = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(new_seg != NULL) SCRS_FREE(new_seg);

  if(cur_seg != NULL) {
    scrs_cleanup((void **)&cur_seg, DGP_TRUE, is_sampled, dgp_state);
  }

  while(cur_seglist != NULL) {
    tmpptr = cur_seglist;
    cur_seglist = cur_seglist->next;
    SEGLN_FREE(tmpptr);
  }

  while(prev_seglist != NULL) {
    tmpptr = prev_seglist;
    prev_seglist = prev_seglist->next;
    SEGLN_FREE(tmpptr);
  }

  if(new_tccstates != NULL) {
    DGPADT_FOREACH(ln, new_tccstates) {
      tccstate = DGPADT_LIGET(ln);
      if(tccstate != NULL) {
        if(tccstate->sc_lines != DGP_NULLLIST) {
          DGPADT_FOREACH(ln2, tccstate->sc_lines) {
            tmpptr = DGPADT_LIGET(ln2);
            if(tmpptr != NULL) scrs_cleanup(&tmpptr, DGP_FALSE, DGP_FALSE, dgp_state);
          }
          DgpAdt_liFree(tccstate->sc_lines);
        }
        TCCS_FREE(tccstate);
      }
    }
    DgpAdt_liFree(new_tccstates);
  }

  if(new_cc != NULL) {
    if(new_cc->runList != NULL) {
      DGP_FREE(new_cc->runList);
    }

    COLORCC_FREE(new_cc);
  }

  ERR_RETURN;
}

#if 0
static INLINE ErrCode
scrs_fill(sc_run_seg *scr_seg1, 
           sc_run_seg *scr_seg2, 
           Dgp_Bool seg2_touching_segs,
           Dgp_State *dgp_state,
           void **ret_scr_seg) 
{
  DGP_PROLOGUE("scrs_fill()", NULL);
  Dgp_UInt16 tmp1;
  sc_run_seg *first_scr_seg = NULL;
  sc_run_seg *tmp_scr_seg, **pnext_scr_seg, *scr_seg2_copy = NULL;
  sc_run *pruns1, *pruns2, *tmp_pruns;
  sc_run *pruns1_end, *pruns2_end, *tmp_pruns_end;
  Dgp_UInt16 cur_xpos;
  void *tmp_ptr;

  
  /* for filling, we consume scr_seg1, but not scr_seg2 */

  if(seg2_touching_segs) {

    /* uh-oh, we're filling under something like a 
       wash where the runs touch.  this can't happen with normal runs,
       so we have to go through and combine any touching runs while 
       making a copy (believe me, it simplifies the code below).  sigh... */
    pnext_scr_seg = &scr_seg2_copy;
    tmp_pruns_end = tmp_pruns = NULL;
    pruns1 = scr_seg2->runs;
    pruns2 = pruns1 + scr_seg2->num_runs;
    do {

      if(tmp_pruns_end == tmp_pruns) {

        /* new segment */
        SCRS_SEC_MALLOC(&tmp_scr_seg);
        scrs_initSeg(tmp_scr_seg, 0);
        tmp_pruns = tmp_scr_seg->runs;
        tmp_pruns_end = tmp_pruns + SC_RUN_NUM_RUNS;

        /* plug it into the list */
        *pnext_scr_seg = tmp_scr_seg;
        pnext_scr_seg = (sc_run_seg **)(&(tmp_scr_seg->next));

      }

      /* fill in current xpos value */
      cur_xpos = pruns1->xpos + pruns1->length;
      tmp_pruns->xpos = pruns1->xpos;
      pruns1++;

      /* consume touching segments */
      do {

        while((pruns1 < pruns2) && 
              (cur_xpos == pruns1->xpos)) {
          cur_xpos += pruns1->length;
          pruns1++;
        }

        /* finished this segment? */
        if(pruns1 == pruns2) {
          scr_seg2 = scr_seg2->next;

          if(scr_seg2 == NULL) break;

          pruns1 = scr_seg2->runs;
          pruns2 = pruns1 + scr_seg2->num_runs;
        }

      } while(cur_xpos == pruns1->xpos);

      /* save new length */
      tmp_pruns->length = cur_xpos - tmp_pruns->xpos;
      tmp_pruns++;
      tmp_scr_seg->num_runs++;

    } while(scr_seg2 != NULL);

    /* now replace the original scr_seg2 with the copy */
    scr_seg2 = scr_seg2_copy;

  }


  /* (re)set new list 'next' pointer */
  pnext_scr_seg = &first_scr_seg;

  /* first, take care of any segments in scr_seg1 which come before any
     segments in scr_seg2 */

  /* compute last x value for scr_seg1 */
  tmp1 = scrs_lastX(scr_seg1);
  while(scrs_lessThan(tmp1, scr_seg2->runs[0].xpos)) {

    /* save current */
    tmp_scr_seg = scr_seg1;

    /* move to next */
    scr_seg1 = scr_seg1->next;

    /* plug this seg into the list */
    *pnext_scr_seg = tmp_scr_seg;
    tmp_scr_seg->next = NULL;
    pnext_scr_seg = (sc_run_seg **)(&(tmp_scr_seg->next));

    /* compute last x value for scr_seg1 */
    tmp1 = scrs_lastX(scr_seg1);
  }

  /* now, combine runs from scr_seg1 and scr_seg2 until we run out
     of runs in scr_seg2, taking into account the
     fact that some of the runs may touch, and must be combined */
  tmp_pruns_end = tmp_pruns = NULL;
  pruns1 = scr_seg1->runs;
  pruns1_end = pruns1 + scr_seg1->num_runs;
  pruns2 = scr_seg2->runs;
  pruns2_end = pruns2 + scr_seg2->num_runs;
  do {
  
    if(tmp_pruns_end == tmp_pruns) {

      /* new segment */
      SCRS_SEC_MALLOC(&tmp_scr_seg);
      scrs_initSeg(tmp_scr_seg, 0);
      tmp_pruns = tmp_scr_seg->runs;
      tmp_pruns_end = tmp_pruns + SC_RUN_NUM_RUNS;

      /* plug it into the list */
      *pnext_scr_seg = tmp_scr_seg;
      pnext_scr_seg = (sc_run_seg **)(&(tmp_scr_seg->next));

    }

    /* fill in current xpos value */
    if(pruns1->xpos < pruns2->xpos) {
      cur_xpos = pruns1->xpos + pruns1->length;
      tmp_pruns->xpos = pruns1->xpos;
      pruns1++;
    } else {
      cur_xpos = pruns2->xpos + pruns2->length;
      tmp_pruns->xpos = pruns2->xpos;
      pruns2++;
    }

    /* consume touching segments */
    do {

      while((pruns2 < pruns2_end) && (pruns1 < pruns1_end)) {

        /* figure out if there is a touching run */
        if((pruns1->xpos < pruns2->xpos) && (cur_xpos == pruns1->xpos)) {
          cur_xpos += pruns1->length;
          pruns1++;
        } else if(cur_xpos == pruns2->xpos) {
          cur_xpos += pruns2->length;
          pruns2++;
        } else {
          /* no more touching runs */
          goto NO_MORE_TOUCHING_RUNS;
        }

      }

      /* finished a segment? */
      if(pruns2 == pruns2_end) {

        scr_seg2 = scr_seg2->next;

        if(scr_seg2 == NULL) break;

        pruns2 = scr_seg2->runs;
        pruns2_end = pruns2 + scr_seg2->num_runs;

      } else {

        tmp_ptr = scr_seg1;
        scr_seg1 = scr_seg1->next;

        /* free used up segment */
        SCRS_FREE(tmp_ptr);

        /* this should never happen in this loop */
        DGP_ASSERT(scr_seg1 != NULL);

        pruns1 = scr_seg1->runs;
        pruns1_end = pruns1 + scr_seg1->num_runs;

      }


    } while(DGP_TRUE);

    /* jump here when no more touching runs */
NO_MORE_TOUCHING_RUNS:

    /* save new length */
    tmp_pruns->length = cur_xpos - tmp_pruns->xpos;
    tmp_pruns++;
    tmp_scr_seg->num_runs++;

  } while(scr_seg2 != NULL);

  /* it's possible that there could be one more touching run
     left in scr_seg1 */
  if(cur_xpos == pruns1->xpos) {
    tmp_pruns[-1].length += pruns1->length;
    pruns1++;
  }

  /* fix up this segment if it has any runs left,
     otherwise, ditch it */
  if(pruns1 < pruns1_end) {

    /* copy remaining runs to the beginning of the run array */
    tmp_pruns = scr_seg1->runs;
    scr_seg1->num_runs = pruns1_end - pruns1;
    while(pruns1 < pruns1_end) {
      *(tmp_pruns++) = *(pruns1++);
    }

  } else {

    /* get next seg */
    tmp_ptr = scr_seg1;
    scr_seg1 = scr_seg1->next;

    /* free old seg */
    SCRS_FREE(tmp_ptr);

  }

  /* finally, tack on any leftovers */
  *pnext_scr_seg = scr_seg1;
  scr_seg1 = NULL;
  
  /* since we don't consume scr_seg2, if we made a copy, 
     we need to free it */
  while(scr_seg2_copy != NULL) {

    tmp_ptr = scr_seg2_copy;
    scr_seg2_copy = scr_seg2_copy->next;

    SCRS_FREE(tmp_ptr);

  }

  /* return the first segment of the new segment list */
  *ret_scr_seg = first_scr_seg;
  first_scr_seg = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  /* cleanup our memory */
  scrs_cleanup((void **)&scr_seg1, DGP_FALSE, DGP_FALSE, dgp_state);
  scrs_cleanup((void **)&scr_seg2_copy, DGP_FALSE, DGP_FALSE, dgp_state);
  scrs_cleanup((void **)&first_scr_seg, DGP_FALSE, DGP_FALSE, dgp_state);

  ERR_RETURN;
}
#endif

