/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 1, 2000
**
**  Filename: colorCCstate.h
**
**  Description of what's implemented in this file:
**    ccstate functions & macros header
**  
**************************************************************************/

/*
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCstate.h,v 1.15 2001/06/21 21:14:05 mkrocken Exp $
 */


#ifndef __COLORCCSTATE_H
#define __COLORCCSTATE_H

#include "dgp_state.h"
#include "colorCCutil.h"
#include "dgp_utils.h"
#include "image_defs.h"

#ifdef __cplusplus
extern "C"
{
#endif


ErrCode
ccstate_init(CCstate **_ccstate, 
             Run *cur_row_runs, 
             Dgp_Int32 cur_run_index, 
             colorCC_state *cookie);

void
ccstate_cleanup(CCstate **_ccstate, 
                colorCC_state *cookie);

ErrCode
ccstate_merge(CCstate *ccstate_dst, 
              CCstate **_ccstate_src, 
              Run *_cur_row_runs,
              Run *_prev_row_runs,
              colorCC_state *cookie); 

ErrCode
ccstate_makeBackground(CCstate *ccstate, 
                       colorCC_state *cookie);

ErrCode
ccstate_makeWash(CCstate *ccstate,
                 colorCC_state *cookie);

ErrCode
ccstate_disassembleSampledCC(CCstate *ccstate, 
                             colorCC_state *cookie);


ErrCode
ccstate_convertToColorCC(CCstate *ccstate, 
                         DgpCC_ColorCC **_new_cc, 
                         colorCC_state *cookie);


/* some macros to set states */
#define ccstate_setNormal(ccstate) \
do { \
  (ccstate)->state.wash = DGPCC_FLAG_FALSE; \
  (ccstate)->state.sampled = DGPCC_FLAG_FALSE; \
  (ccstate)->state.background = DGPCC_FLAG_FALSE; \
} while(0)

#define ccstate_setWash(ccstate) \
do { \
  (ccstate)->state.wash = DGPCC_FLAG_TRUE; \
} while(0)

#define ccstate_clearWash(ccstate) \
do { \
  (ccstate)->state.wash = DGPCC_FLAG_FALSE; \
} while(0)

#define ccstate_setSampled(ccstate) \
do { \
  (ccstate)->state.sampled = DGPCC_FLAG_TRUE; \
} while(0)

#define ccstate_clearSampled(ccstate) \
do { \
  (ccstate)->state.sampled = DGPCC_FLAG_FALSE; \
} while(0)

#define ccstate_setBackground(ccstate) \
do { \
  (ccstate)->state.background = DGPCC_FLAG_TRUE; \
} while(0)

#define ccstate_clearBackground(ccstate) \
do { \
  (ccstate)->state.background = DGPCC_FLAG_FALSE; \
} while(0)

#define ccstate_setActive(ccstate) \
do { \
  (ccstate)->state.active = CCSTATE_ACTIVE; \
} while(0)

#define ccstate_setNewInactive(ccstate) \
do { \
  (ccstate)->state.active = CCSTATE_NEW_INACTIVE; \
} while(0)

#define ccstate_setInactive(ccstate) \
do { \
  (ccstate)->state.active = CCSTATE_INACTIVE; \
} while(0)


#define ccstate_isPotentialWash(ccstate) ((ccstate)->num_potential_washes > 0)


#define ccstate_approxArea(ccstate) \
  ((ccstate)->area)

#define ccstate_updateBounds(ccstate, run, ypos) \
do { \
  (ccstate)->lrY = (ypos); \
  (ccstate)->ulX = MIN((ccstate)->ulX, (run)->xpos); \
  (ccstate)->lrX = MAX(((ccstate)->lrX), ((run)->xpos + (run)->cccrun.length - 1)); \
} while(0)

#define ccstate_addRun(_ccstate, run, run_index, ypos) \
do { \
  (_ccstate)->area += (run)->cccrun.length; \
  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd((_ccstate)->cur_run_line, (void *)(run_index), NULL)); \
  (run)->ccstate = (_ccstate); \
  ccstate_updateBounds((_ccstate), (run), (Dgp_UInt16)(ypos)); \
} while(0)


#ifdef __cplusplus
}
#endif

#endif /* __COLORCCSTATE_H */
