/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 1 2000
**
**  Filename: colorCCutil.c
**
**  Description of what's implemented in this file:
**    color connected components utilities
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCutil.c,v 1.13 2003/01/27 21:48:39 awilliam Exp $
 *
 */

#include "dgp_rcs.h"
#include "colorCCutil.h"
#include "dgp_utils.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCutil.c,v 1.13 2003/01/27 21:48:39 awilliam Exp $");


#ifdef DGP_DBUG
ErrCode
cccmap_validate(colorCC_state *cookie)
{
  DGP_PROLOGUE("cccmap_init()", cookie->DGP_PERRSTATE);
  cccMap *map = cookie->cccmap;
  Dgp_Int32 i, j, num_els = 0;
  cccMapColData *col = map->col;
  cccMapRowData *row1;
  Dgp_UInt32 num_neighbors, num_potential_washes;
  
  DGP_STOP_WHINE;
  for(i = (map->max_num_els - 1); i >= 0; i--) {

    if(col[i].ccstate != NULL) {
  
      num_els++;

      DGP_ASSERT(col[i].ccstate->map_index == i);

      num_neighbors = 0;
      num_potential_washes = 0;

      row1 = col[i].row;
      for(j = 0; j < i; j++) {
        if(cccmaprd_isNeighbor(&row1[j])) {
          DGP_ASSERT(col[j].ccstate != NULL);
          num_neighbors++;
        
          if(cccmaprd_isPotentialWash(&row1[j])) {
            num_potential_washes++;
          }
        }
      }

      for(++j; j < map->max_num_els; j++) {
        if(cccmaprd_isNeighbor(&(col[j].row[i]))) {
          num_neighbors++;

          if(cccmaprd_isPotentialWash(&(col[j].row[i]))) {
            num_potential_washes++;
          }
        }
      }

      if(col[i].ccstate->num_neighbors != num_neighbors) {
        printf("ccstate at %d thinks it has %d neighbors, really has %d\n\tneighbors: ",
          (int)i, (int)col[i].ccstate->num_neighbors, (int)num_neighbors);

        for(j = 0; j < i; j++) {
          if(cccmaprd_isNeighbor(&row1[j])) {
            printf("%d ", (int)j);
          }
        }

        for(++j; j < map->max_num_els; j++) {
          if(cccmaprd_isNeighbor(&(col[j].row[i]))) {
            printf("%d ", (int)j);
          }
        }

        printf("\n\n");


        DGP_ASSERT(col[i].ccstate->num_neighbors == num_neighbors);
      }

      if(col[i].ccstate->num_potential_washes != num_potential_washes) {
        printf("ccstate at %d thinks it has %d potential wash neighbors, really has %d\n\tpotential wash neighbors: ",
          (int)i, (int)col[i].ccstate->num_potential_washes, (int)num_potential_washes);

        for(j = 0; j < i; j++) {
          if(cccmaprd_isNeighbor(&row1[j]) && cccmaprd_isPotentialWash(&row1[j])) {
            printf("%d ", (int)j);
          }
        }

        for(++j; j < map->max_num_els; j++) {
          if(cccmaprd_isNeighbor(&(col[j].row[i])) && cccmaprd_isPotentialWash(&(col[j].row[i]))) {
            printf("%d ", (int)j);
          }
        }

        printf("\n\n");


        DGP_ASSERT(col[i].ccstate->num_potential_washes == num_potential_washes);
      }

    } else {

      row1 = col[i].row;
      for(j = 0; j < i; j++) {
        DGP_ASSERT(!cccmaprd_isNeighbor(&row1[j]));
      }

      DGP_ASSERT(i >= map->cached_index);
    }

  }

  DGP_ASSERT(num_els == map->num_els);

  NORMAL_RETURN;

}
#endif


ErrCode
cccmap_init(cccMap **_map,
            colorCC_state *cookie)
{
  DGP_PROLOGUE("cccmap_init()", cookie->DGP_PERRSTATE);
  cccMap *map = NULL;
  Dgp_Int32 i;
  Dgp_Allocators *a = &(cookie->DGP_ALLOCATORS);

  DGP_ASSERT(_map != NULL);
  *_map = NULL;

  /* create the map */
  DGP_SEC_ACALLOC(map, (cccMap *), 1, sizeof(cccMap));

  /* create the column array */
  DGP_SEC_ACALLOC(map->col, (void *), CCCMAP_DEF_SIZE, sizeof(cccMapColData));

  /* create each row */
  for(i = 1; i < CCCMAP_DEF_SIZE; i++) {
    DGP_SEC_ACALLOC(map->col[i].row, (void *), i, sizeof(cccMapRowData));
  }

  map->max_num_els = CCCMAP_DEF_SIZE;
  
  *_map = map;
  map = NULL;


  NORMAL_RETURN;

ERR_CLEANUP:

  cccmap_cleanup(&map, cookie);

  ERR_RETURN;
}

void
cccmap_cleanup(cccMap **_map, colorCC_state *cookie)
{
  cccMap *map;
  Dgp_Int32 i;
  Dgp_Free tmp_free;

  if(_map == NULL) return;

  map = *_map;

  if(map == NULL) return;
  *_map = NULL;

  tmp_free = cookie->DGP_FREE;

  if(map->col != NULL) {
    for(i = 1; i < map->max_num_els; i++) {
      if(map->col[i].row != NULL) {
        tmp_free(map->col[i].row);
      }
    }
    tmp_free(map->col);
  }

  tmp_free(map);

}


ErrCode
cccmap_expand(cccMap *map, colorCC_state *cookie)
{
  DGP_PROLOGUE("cccmap_expand()", cookie->DGP_PERRSTATE);
  Dgp_Int32 i, expand_els = map->max_num_els + CCCMAP_INCREMENT;
  Dgp_Allocators *a = &(cookie->DGP_ALLOCATORS);
  
  /* add more elements */
  DGP_SEC_AREALLOC(map->col, (cccMapColData *), (expand_els * sizeof(cccMapColData)));

  /* clear new elements */
  memset(map->col + map->max_num_els, 0, (CCCMAP_INCREMENT * sizeof(cccMapColData)));

  /* for each new element, add a row of data*/
  for(i = map->max_num_els; i < expand_els; i++) {
    DGP_SEC_ACALLOC(map->col[i].row, (void *), i, sizeof(cccMapRowData));
  }

  /* update map's max number of elements */
  map->max_num_els = expand_els;

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


void
cccmap_collapse(cccMap *map, colorCC_state *cookie)
{
  Dgp_Int32 i, j, k, collapse_els;
  cccMapColData *col = map->col;
  cccMapRowData *row1, *row2;
  Dgp_UInt32 num_neighbors;
  Dgp_Allocators *a = &(cookie->DGP_ALLOCATORS);


  /* see if we should collapse the map some */
  if((map->max_num_els > CCCMAP_DEF_SIZE) &&
     ((map->max_num_els - map->num_els) >= CCCMAP_INCREMENT)) {

    if(map->num_els > CCCMAP_DEF_SIZE) {
      /* shave off map size by increments of CCCMAP_INCREMENT */
      collapse_els = CCCMAP_DEF_SIZE + (((Dgp_Int32)((map->num_els - CCCMAP_DEF_SIZE + CCCMAP_INCREMENT - 1) /
        CCCMAP_INCREMENT)) * CCCMAP_INCREMENT);
    } else {
      /* we don't want to make it smaller than the default size */
      collapse_els = CCCMAP_DEF_SIZE;
    }

    i = map->cached_index;
    j = map->max_num_els - 1;

    /* only do the loop if there are some ccstates in the map */
    if(map->num_els > 0) {

      do {

        /* find first potential element to be moved */
        while(col[j].ccstate == NULL) j--;

        /* only move if we have to */
        if(j < collapse_els) break;

        /* we have to move, find first open slot below
           the collapse level */
        while(col[i].ccstate != NULL) i++;

        /* move the row data */
        row1 = col[i].row;
        row2 = col[j].row;
        num_neighbors = col[j].ccstate->num_neighbors;
        for(k = 0; k < i; k++) {
          if(cccmaprd_isNeighbor(&row2[k])) {
            row1[k] = row2[k];
            cccmaprd_clearFlags(&row2[k]);
            
            /* done yet? */
            if(--num_neighbors == 0) goto DONE_NEIGHBOR_MOVE;
          }
        }
        for(++k; k < j; k++) {
          if(cccmaprd_isNeighbor(&row2[k])) {
            col[k].row[i] = row2[k];
            cccmaprd_clearFlags(&row2[k]);
            
            /* done yet? */
            if(--num_neighbors == 0) goto DONE_NEIGHBOR_MOVE;
          }
        }

        /* we jump here after we've moved all the neighbors */
DONE_NEIGHBOR_MOVE:

        /* move column data */
        col[i].ccstate = col[j].ccstate;

        /* set new index for the element, or the
           element won't be able to find itself. */
        col[i].ccstate->map_index = i;

        /* indicate the old element is empty */
        col[j].ccstate = NULL;

        /* move past the elements we just
           looked at */
        i++;
        j--;

      } while(j >= collapse_els);

    }

    map->cached_index = i;

    /* now, get rid of extraneous memory */
    for(k = collapse_els; k < map->max_num_els; k++) {
      DGP_AFREE(col[k].row);
      col[k].row = NULL;
    }

    /* update map's max number of elements */
    map->max_num_els = collapse_els;

  }

  return;
}


/* remove an element from the map */
void
cccmap_rem(cccMap *map, Dgp_Int32 element)
{
  DGP_VOID_PROLOGUE("cccmap_rem()", NULL);
  cccMapColData *col = map->col;
  Dgp_Int32 i, num_neighbors = col[element].ccstate->num_neighbors;
  cccMapRowData *row = col[element].row;


  /* this ccstate had better be done with */
  DGP_VOID_ASSERT(col[element].ccstate->num_potential_washes == 0);

  /* done yet? */
  if(num_neighbors <= 0) goto NO_MORE_NEIGHBORS;

  /* removing neighbor info from other elements's data */
  for(i = 0; i < element; i++) {
    if(cccmaprd_isNeighbor(&row[i])) {
      cccmaprd_clearFlags(&row[i]);
      col[i].ccstate->num_neighbors--;

      /* done yet? */
      if(--num_neighbors == 0) goto NO_MORE_NEIGHBORS;
    }
  }

  /* still removing neighbor info from other elements's data */
  for(++i; i < map->max_num_els; i++) {
    if(cccmaprd_isNeighbor(&(col[i].row[element]))) {
      cccmaprd_clearFlags(&(col[i].row[element]));
      col[i].ccstate->num_neighbors--;

      /* done yet? */
      if(--num_neighbors == 0) goto NO_MORE_NEIGHBORS;
    }
  }

NO_MORE_NEIGHBORS:

  /* this element has been removed */
  col[element].ccstate->num_neighbors = 0;
  col[element].ccstate->map_index = CCCMAP_INVALID_INDEX;
  col[element].ccstate = NULL;
  map->num_els--;

  /* change cached index */
  if(element < map->cached_index) map->cached_index = element;

}


/* combine two elements in the map */
void
cccmap_combine(cccMap *map, Dgp_Int32 dst, Dgp_Int32 src)
{
  DGP_VOID_PROLOGUE("cccmap_combine()", NULL);
  cccMapColData *col = map->col;
  Dgp_Int32 k;
  Dgp_Int32 lesser, greater;
  cccMapRowData *row1, *row2;
  Dgp_UInt32 src_num_neighbors, dst_num_neighbors;
  Dgp_UInt16 dst_num_potential_washes, dst_potential;
  Dgp_UInt16 distance;
  Dgp_Bool swapped = DGP_FALSE;


  DGP_VOID_ASSERT(dst != src);
  DGP_VOID_ASSERT(ccstate_isBackground(col[dst].ccstate) == ccstate_isBackground(col[src].ccstate));

  if(dst < src) {
    lesser = dst;
    greater = src;
  } else {
    lesser = src;
    greater = dst;
  }

  /* we want to move the element with the least neighbors, so
     swap if the src has more neighbors (we'll fix it all in
     the end, don't worry... */
  if(col[src].ccstate->num_neighbors > col[dst].ccstate->num_neighbors) {

    /* swap indexes */
    k = src; src = dst; dst = k;
    swapped = DGP_TRUE;

  }

  dst_num_neighbors = col[dst].ccstate->num_neighbors;
  dst_num_potential_washes = col[dst].ccstate->num_potential_washes;
  src_num_neighbors = col[src].ccstate->num_neighbors;
  row1 = col[src].row;

  /* done yet? */
  if(src_num_neighbors == 0) goto NO_MORE_NEIGHBORS;

  /* combine src's neighbor info into dst's info */
  row2 = col[dst].row;
  for(k = 0; k < lesser; k++) {
    /* row1 = src, row2 = dst */
    if(cccmaprd_isNeighbor(&row1[k])) {
      distance = row1[k].distance;
      if(!cccmaprd_isNeighbor(&row2[k])) {
        row2[k].distance = distance;
        ccmaprd_copyFlags(&row2[k], &row1[k]);
        dst_num_neighbors++;
        dst_num_potential_washes += (Dgp_UInt16)(cccmaprd_isPotentialWash(&row1[k]));
      } else {
        col[k].ccstate->num_neighbors--;

        if(cccmaprd_isPotentialWash(&row1[k])) {
          dst_potential = (Dgp_UInt16)(cccmaprd_isPotentialWash(&row2[k]));
          dst_num_potential_washes += 1 - dst_potential;
          col[k].ccstate->num_potential_washes -= dst_potential;
        }

        ccmaprd_mergeFlags(&row2[k], &row1[k]);
        if(row2[k].distance < distance) row2[k].distance = distance;
      }
      cccmaprd_clearFlags(&row1[k]);

      /* done yet? */
      if(--src_num_neighbors == 0) goto NO_MORE_NEIGHBORS;
    }
  }

  if(dst == lesser) {

    /* continue combining src's neighbor info into dst's info */
    for(++k; k < greater; k++) {
      row2 = col[k].row;
      /* row1 = src, row2 = dst */
      if(cccmaprd_isNeighbor(&row1[k])) {
        distance = row1[k].distance;
        if(!cccmaprd_isNeighbor(&row2[dst])) {
          row2[dst].distance = distance;
          ccmaprd_copyFlags(&row2[dst], &row1[k]);
          dst_num_neighbors++;
          dst_num_potential_washes += (Dgp_UInt16)(cccmaprd_isPotentialWash(&row1[k]));
        } else {
          col[k].ccstate->num_neighbors--;

          if(cccmaprd_isPotentialWash(&row1[k])) {
            dst_potential = (Dgp_UInt16)(cccmaprd_isPotentialWash(&row2[dst]));
            dst_num_potential_washes += 1 - dst_potential;
            col[k].ccstate->num_potential_washes -= dst_potential;
          }

          ccmaprd_mergeFlags(&row2[dst], &row1[k]);
          if(row2[dst].distance < distance) row2[dst].distance = distance;
        }
        cccmaprd_clearFlags(&row1[k]);

        /* done yet? */
        if(--src_num_neighbors == 0) goto NO_MORE_NEIGHBORS;
      }
    }

  } else {

    row1 = col[dst].row;

    /* continue combining src's neighbor info into dst's info */
    for(++k; k < greater; k++) {
      row2 = col[k].row;
      /* row1 = dst, row2 = src */
      if(cccmaprd_isNeighbor(&row2[src])) {
        distance = row2[src].distance;
        if(!cccmaprd_isNeighbor(&row1[k])) {
          row1[k].distance = distance;
          ccmaprd_copyFlags(&row1[k], &row2[src]);
          dst_num_neighbors++;
          dst_num_potential_washes += (Dgp_UInt16)(cccmaprd_isPotentialWash(&row2[src]));
        } else {
          col[k].ccstate->num_neighbors--;

          if(cccmaprd_isPotentialWash(&row2[src])) {
            dst_potential = (Dgp_UInt16)(cccmaprd_isPotentialWash(&row1[k]));
            dst_num_potential_washes += 1 - dst_potential;
            col[k].ccstate->num_potential_washes -= dst_potential;
          }

          ccmaprd_mergeFlags(&row1[k], &row2[src]);
          if(row1[k].distance < distance) row1[k].distance = distance;
        }
        cccmaprd_clearFlags(&row2[src]);

        /* done yet? */
        if(--src_num_neighbors == 0) goto NO_MORE_NEIGHBORS;
      }
    }

  }

  /* are src and dst neighbors? */
  if(cccmaprd_isNeighbor(&row1[lesser])) {
    dst_num_potential_washes -= cccmaprd_isPotentialWash(&row1[lesser]);
    cccmaprd_clearFlags(&row1[lesser]);
    dst_num_neighbors--;

    /* done yet? */
    if(--src_num_neighbors == 0) goto NO_MORE_NEIGHBORS;
  }

  /* still combining src's neighbor info into dst's info */
  for(++k; k < map->max_num_els; k++) {
    row1 = col[k].row;
    if(cccmaprd_isNeighbor(&row1[src])) {
      distance = row1[src].distance;
      if(!cccmaprd_isNeighbor(&row1[dst])) {
        row1[dst].distance = distance;
        ccmaprd_copyFlags(&row1[dst], &row1[src]);
        dst_num_neighbors++;
        dst_num_potential_washes += (Dgp_UInt16)(cccmaprd_isPotentialWash(&row1[src]));
      } else {
        col[k].ccstate->num_neighbors--;

        if(cccmaprd_isPotentialWash(&row1[src])) {
          dst_potential = (Dgp_UInt16)(cccmaprd_isPotentialWash(&row1[dst]));
          dst_num_potential_washes += 1 - dst_potential;
          col[k].ccstate->num_potential_washes -= dst_potential;
        }

        ccmaprd_mergeFlags(&row1[dst], &row1[src]);
        if(row1[dst].distance < distance) row1[dst].distance = distance;
      }
      cccmaprd_clearFlags(&row1[src]);

      /* done yet? */
      if(--src_num_neighbors == 0) goto NO_MORE_NEIGHBORS;
    }
  }

NO_MORE_NEIGHBORS:

  /* element dst has been combined with src */
  if(swapped) {
    void *_data;

    /* swap the data info */
    _data = col[dst].ccstate;
    col[dst].ccstate = col[src].ccstate;
    col[src].ccstate = _data;

    /* set new index */
    col[dst].ccstate->map_index = dst;

  }

  /* change the cached value */
  if(src < map->cached_index) map->cached_index = src;

  /* update dst */
  col[dst].ccstate->num_neighbors = dst_num_neighbors;
  col[dst].ccstate->num_potential_washes = dst_num_potential_washes;

  /* invalidate the src */
  col[src].ccstate->map_index = CCCMAP_INVALID_INDEX;
  col[src].ccstate->num_neighbors = 0;
  col[src].ccstate = NULL;
  map->num_els--;
}


ErrCode
rim_init(RunImage **_rim, colorCC_state *cookie)
{
  DGP_PROLOGUE("rim_init()", cookie->DGP_PERRSTATE);
  Dgp_Allocators *a = &(cookie->DGP_ALLOCATORS);
  RunImage *rim = NULL;

  
  DGP_ASSERT(_rim != NULL);
    
  /* create the run image structure */
  DGP_SEC_ACALLOC(rim, (RunImage *), 1, sizeof(RunImage));
  rim->cur_row = rim->rows;
  rim->prev_row = rim->rows + 1;
  
  /* set up the rows */
  DGP_SEC_AMALLOC(rim->rows[0].runs, (Run *), (RIM_DEF_SIZE * sizeof(Run)));
  rim->rows[0].max_runs = RIM_DEF_SIZE;
  DGP_SEC_AMALLOC(rim->rows[1].runs, (Run *), (RIM_DEF_SIZE * sizeof(Run)));
  rim->rows[1].max_runs = RIM_DEF_SIZE;

  *_rim = rim;
  rim = NULL;

  NORMAL_RETURN;
ERR_CLEANUP:

  rim_cleanup(&rim, cookie);

  ERR_RETURN;
}


ErrCode
rim_endPage(RunImage *rim, colorCC_state *cookie)
{
  DGP_PROLOGUE("rim_initPage()", cookie->DGP_PERRSTATE);
  Dgp_Allocators *a = &(cookie->DGP_ALLOCATORS);

  if(rim->rows[0].max_runs > RIM_DEF_SIZE) {
    DGP_SEC_AREALLOC(rim->rows[0].runs, (Run *), (RIM_DEF_SIZE * sizeof(Run)));
    rim->rows[0].max_runs = RIM_DEF_SIZE;
  }
  if(rim->rows[1].max_runs > RIM_DEF_SIZE) {
    DGP_SEC_AREALLOC(rim->rows[1].runs, (Run *), (RIM_DEF_SIZE * sizeof(Run)));
    rim->rows[1].max_runs = RIM_DEF_SIZE;
  }

  rim->rows[0].num_runs = 0;
  rim->rows[1].num_runs = 0;

  NORMAL_RETURN;

ERR_CLEANUP:

  /* no cleanup necessary */
  ERR_RETURN;
}


void
rim_cleanup(RunImage **_rim, colorCC_state *cookie)
{
  RunImage *rim;
  Dgp_Free tmp_free;

  if(_rim == NULL) return;

  rim = *_rim;
  
  if(rim == NULL) return;
  *_rim = NULL;
  
  tmp_free = cookie->DGP_FREE;

  if(rim->rows[0].runs != NULL) {
    tmp_free(rim->rows[0].runs);
  }
  if(rim->rows[1].runs != NULL) {
    tmp_free(rim->rows[1].runs);
  }

  tmp_free(rim);
}


void
scrs_cleanup(void **_scr_seg, 
             Dgp_Bool use_scxr_segs, 
             Dgp_Bool free_userdata,
             Dgp_State *dgp_state)
{
  sc_run_seg *scr_seg = *_scr_seg;
  void *tmp_seg;
  int i;

  *_scr_seg = NULL;

  if(!use_scxr_segs) {
    
    while(scr_seg != NULL) {
      tmp_seg = scr_seg->next;
      SCRS_FREE(scr_seg);
      scr_seg = tmp_seg;
    }

  } else {
    
    while(scr_seg != NULL) {
      tmp_seg = scr_seg->next;

      if(free_userdata) {
        for(i = 0; i < scr_seg->num_runs; i++) {
          if((void *)(((sc_xrun_seg *)scr_seg)->userdata[i]) != NULL) {
            DGP_FREE((void *)(((sc_xrun_seg *)scr_seg)->userdata[i]));
          }
        }
      }

      SCXRS_FREE(scr_seg);
      scr_seg = tmp_seg;
    }

  }

}
