/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 1, 2000
**
**  Filename: colorCCutil.h
**
**  Description of what's implemented in this file:
**    color connected components utilities header
**  
**************************************************************************/

/*
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCutil.h,v 1.15 2001/06/21 18:50:50 mkrocken Exp $
 */


#ifndef __COLORCCUTIL_H
#define __COLORCCUTIL_H

#include "dgp_state.h"
#include "colorCCutil_t.h"
#include "dgp_utils.h"

#ifdef __cplusplus
extern "C"
{
#endif

#define DGPCC_FLAG_TRUE  1
#define DGPCC_FLAG_FALSE 0


/***
 * some color space macros 
 ***/

/* mask to clear extraneous bits in an Dgp_UInt32 */
#define RGB_LONG_MASK 0x00FFFFFF

/*                         XXrrggbb */
#define DEFAULT_BG_COLOR 0x00FFFFFF

#define DEFAULT_BG_COLOR_R 0xFF
#define DEFAULT_BG_COLOR_G 0xFF
#define DEFAULT_BG_COLOR_B 0xFF

#define lrgb_copyToSampling(sampling, color) \
do { \
  (sampling).r = (Dgp_UInt8)(DGPCC_LRGB_R(color)); \
  (sampling).g = (Dgp_UInt8)(DGPCC_LRGB_G(color)); \
  (sampling).b = (Dgp_UInt8)(DGPCC_LRGB_B(color)); \
} while(0)


/********
 * cccMap stuff
 ********/

#define cccmaprd_setNeighbor(prow_data) \
do { \
  (prow_data)->state.flags.neighbor = DGPCC_FLAG_TRUE; \
} while (0)

#define cccmaprd_setPotentialWash(prow_data) \
do { \
  (prow_data)->state.flags.potential_wash = DGPCC_FLAG_TRUE; \
} while (0)

#define cccmaprd_clearPotentialWash(prow_data) \
do { \
  (prow_data)->state.flags.potential_wash = DGPCC_FLAG_FALSE; \
} while (0)

#define ccmaprd_mergeFlags(prow_data_dst, prow_data_src) \
do { \
  (prow_data_dst)->state.all |= (prow_data_src)->state.all; \
} while(0)

#define ccmaprd_copyFlags(prow_data_dst, prow_data_src) \
do { \
  (prow_data_dst)->state.all = (prow_data_src)->state.all; \
} while(0)

#define cccmaprd_clearFlags(prow_data) \
do { \
  (prow_data)->state.all = 0; \
} while(0)


#ifdef DGP_DBUG
ErrCode
cccmap_validate(colorCC_state *cookie);
#endif

ErrCode
cccmap_init(cccMap **map,
            colorCC_state *cookie);

void
cccmap_cleanup(cccMap **map, colorCC_state *cookie);

ErrCode
cccmap_expand(cccMap *map, colorCC_state *cookie);

void
cccmap_collapse(cccMap *map, colorCC_state *cookie);


/* remove an element from the map */
void
cccmap_rem(cccMap *map, Dgp_Int32 element);

/* combine two elements in the map */
void
cccmap_combine(cccMap *map, Dgp_Int32 dst, Dgp_Int32 src);





/* add new data to the map.  if there is room, put it in an existing
   space.  otherwise, expand the map and put it in a new space */
#define cccmap_add(map, _ccstate, _i) \
do { \
  DGP_ASSERT(_ccstate != NULL); \
  if((map)->num_els == (map)->max_num_els) { \
    DGP_ERR_PASS_CLEANUP(cccmap_expand((map), cookie)); \
    (_i) = (map)->num_els; \
  } else { \
    (_i) = (map)->cached_index; \
    while((map)->col[_i].ccstate != NULL) (_i)++; \
  } \
  (map)->cached_index = (_i) + 1; \
  (_ccstate)->map_index = (_i); \
  (map)->col[_i].ccstate = (_ccstate); \
  (map)->num_els++; \
} while(0)


/* update i and j neighbor info, indicating that they are touching */
#define cccmap_updateNeighbor(map, prow_data, _i, _j, ccstate_i, ccstate_j, _distance) \
do { \
  DGP_ASSERT((_i) != CCCMAP_INVALID_INDEX); \
  DGP_ASSERT((_j) != CCCMAP_INVALID_INDEX); \
  (prow_data) = (((_i) < (_j)) ? &((map)->col[_j].row[_i]) : &((map)->col[_i].row[_j])); \
  if(!cccmaprd_isNeighbor(prow_data)) { \
    cccmaprd_setNeighbor(prow_data); \
    (ccstate_i)->num_neighbors++; \
    (ccstate_j)->num_neighbors++; \
    (prow_data)->distance = (_distance); \
    if((prow_data)->distance < cookie->wThresh) { \
      cccmaprd_setPotentialWash(prow_data); \
      (ccstate_i)->num_potential_washes++; \
      (ccstate_j)->num_potential_washes++; \
    } \
  } \
} while(0)


/* get the data describing the relationship between two ccstates */
#define cccmap_getRowData(map, _i, _j) \
 (((_i) > (_j)) ? &((map)->col[_i].row[_j]) : &((map)->col[_j].row[_i]))


/********
 * Run stuff
 ********/

#define run_runsMatch(run1, run2) \
  ((((run1)->cccrun.colorInfo.rgbColor == (run2)->cccrun.colorInfo.rgbColor) && \
    !run_isSampled(run1) && !run_isSampled(run2)) || \
   (run_isSampled(run1) && run_isSampled(run2)))


 /********
 * semi-compact run stuff
 ********/

#define SCRS_SEC_MALLOC(_ppseg) \
  DGP_ERR_PASS_CLEANUP(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, SCRS_POOL_ID, (void **)(_ppseg)))
#define SCRS_FREE(_pseg) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, SCRS_POOL_ID, (void *)(_pseg)))

#define SCXRS_SEC_MALLOC(_ppseg) \
  DGP_ERR_PASS_CLEANUP(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, SCXRS_POOL_ID, (void **)(_ppseg)))
#define SCXRS_FREE(_pseg) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, SCXRS_POOL_ID, (void *)(_pseg)))


#define SCRS_SEC_MALLOC_WID(_pool_id, _ppseg) \
  DGP_ERR_PASS_CLEANUP(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, _pool_id, (void **)(_ppseg)))
#define SCRS_FREE_WID(_pool_id, _pseg) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, _pool_id, (void *)(_pseg)))


ErrCode scrs_malloc(Dgp_State *dgp_state, sc_run_seg **ret_seg);
void scrs_free(Dgp_State *dgp_state, sc_run_seg *old_scr_seg);

ErrCode scxrs_malloc(Dgp_State *dgp_state, sc_xrun_seg **ret_seg);
void scxrs_free(Dgp_State *dgp_state, sc_xrun_seg *old_scxr_seg);

void
scrs_cleanup(void **_scr_seg, 
             Dgp_Bool use_scxr_segs, 
             Dgp_Bool free_userdata,
             Dgp_State *dgp_state);

#define scrs_touches(scrs1_xstart, scrs1_xend, scrs2_xstart, scrs2_xend) \
  ((((scrs2_xstart) <= (scrs1_xend)) && ((scrs1_xend) <= (scrs2_xend))) || \
   (((scrs1_xstart) <= (scrs2_xend)) && ((scrs2_xend) <= (scrs1_xend))))

#define scrs_lessThan(scrs1_xend, scrs2_xstart) \
  ((scrs1_xend) < (scrs2_xstart))

#define scrs_noOverlap(scrs1_xstart, scrs1_xend, scrs2_xstart, scrs2_xend) \
  (((scrs1_xend) <= (scrs2_xstart)) || ((scrs2_xend) <= (scrs1_xstart)))

#define scrs_lastX(scr_seg) \
  ((scr_seg)->runs[(scr_seg)->num_runs-1].xpos + (scr_seg)->runs[(scr_seg)->num_runs-1].length)

#define scrs_initSeg(scr_seg, _num_runs) \
do { \
  (scr_seg)->num_runs = _num_runs; \
  (scr_seg)->next = NULL; \
} while(0)



/********
 * RunImage stuff
 ********/

#define rim_moveNextRow(rim, pcurrow) \
do { \
  (pcurrow) = (rim)->prev_row; \
  (rim)->prev_row = (rim)->cur_row; \
  (pcurrow)->num_runs = 0; \
  (rim)->cur_row = (pcurrow); \
} while(0)

#define rim_getNewRun(rim, pcurrow, pcurrowruns, newrun_index) \
do { \
  if((pcurrow)->num_runs == (pcurrow)->max_runs) { \
    DGP_SEC_AREALLOC((pcurrow)->runs, (Run *), \
      (((pcurrow)->max_runs + RIM_INCREMENT) * sizeof(Run))); \
    (pcurrow)->max_runs += RIM_INCREMENT; \
    (pcurrowruns) = (pcurrow)->runs; \
  } \
  (newrun_index) = ((pcurrow)->num_runs)++; \
} while(0)

#define rim_deleteLastRun(pcurrow) \
do { \
  (pcurrow)->num_runs--; \
} while(0)

#define rim_getPrevRow(rim) ((rim)->prev_row)
#define rim_getCurRow(rim)  ((rim)->cur_row)


ErrCode
rim_init(RunImage **rim, colorCC_state *cookie);

ErrCode
rim_endPage(RunImage *rim, colorCC_state *cookie);

void
rim_cleanup(RunImage **rim, colorCC_state *cookie);


/********
 * DgpCC_ColorCC stuff
 ********/

/* get colorCC states */
#define colorcc_isSampled(colorcc) ((colorcc)->state.sampled)
#define colorcc_isWash(colorcc) ((colorcc)->state.wash)
#define colorcc_isWashOrSampled(colorcc) (colorcc_isWash(colorcc) || colorcc_isSampled(colorcc))

#define COLORCC_SEC_MALLOC(_ppnewcc) \
  DGP_ERR_PASS_CLEANUP(dgp_state->sm_a.mallocSmall(dgp_state->sm_a.smallCookie, DGPCC_COLORCC_POOL, (void **)(_ppnewcc)))

#define COLORCC_FREE(_pnewcc) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, DGPCC_COLORCC_POOL, (void *)(_pnewcc)))




#ifdef __cplusplus
}
#endif

#endif /* __COLORCCUTIL_H */
