/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 31 2000
**
**  Filename: colorCCwash.c
**
**  Description of what's implemented in this file:
**    functions to do wash detection during color cc finding.
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCwash.c,v 1.9 2003/01/27 21:48:39 awilliam Exp $
 *
 */

#include "colorCC-priv.h"
#include "dgp_utils.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercc/colorCCwash.c,v 1.9 2003/01/27 21:48:39 awilliam Exp $");


static INLINE ErrCode
dgpcc_doWashMerge1(CCstate *ccstate,
                   Dgp_Int32 cur, 
                   Dgp_Int32 first, 
                   Dgp_Int32 last, 
	                 cccMapColData *col,
                   colorCC_state *cookie)
{
  DGP_PROLOGUE("dgpcc_doWashMerge1()", NULL);
  Dgp_Int32 j;
  RunImage *rim = cookie->rim;
  Run *cur_row_runs, *prev_row_runs;
  cccMapRowData *row;
  CCstate *tmp_ccstate;
  Dgp_Bool not_wash = DGP_TRUE;


  /* this shouldn't be a wash yet... */
  DGP_ASSERT(!ccstate_isWash(ccstate));

  cur_row_runs = (rim_getCurRow(rim))->runs;
  prev_row_runs = (rim_getPrevRow(rim))->runs;

  do {

    /* check for inactive neighbors, who are potential washes
       with me, and who's color distance from me is within the bounds */
    row = col[cur].row;
    for(j = first; j < cur; j++) {
      if(cccmaprd_isPotentialWash(&row[j]) && 
	       !(ccstate_isActive(col[j].ccstate))) {

        /* clear potential wash status */
        cccmaprd_clearPotentialWash(&row[j]);
        col[j].ccstate->num_potential_washes--;
        ccstate->num_potential_washes--;

	      if(row[j].distance < cookie->wThresh) {
	        /* we've found a wash neighbor */
	        goto FOUND_WASH_NEIGHBOR;
	      }

        if(!ccstate_isPotentialWash(ccstate)) NORMAL_RETURN;
      }
    }


    for(++j; j <= last; j++) {
      row = col[j].row;
      if(cccmaprd_isPotentialWash(&row[cur]) && 
	     !(ccstate_isActive(col[j].ccstate))) {

        /* clear potential wash status */
	      cccmaprd_clearPotentialWash(&row[cur]);
        col[j].ccstate->num_potential_washes--;
        ccstate->num_potential_washes--;

        if(row[cur].distance < cookie->wThresh) {
	        /* we've found a wash neighbor */
	        goto FOUND_WASH_NEIGHBOR;
	      }

        if(!ccstate_isPotentialWash(ccstate)) NORMAL_RETURN;
      }
    }

    /* didn't find any (more) matches, we're done */
    NORMAL_RETURN;

  FOUND_WASH_NEIGHBOR:
    
    /* we've found a wash match */
    tmp_ccstate = col[j].ccstate;

    if(not_wash) {
      DGP_ERR_PASS(ccstate_makeWash(ccstate, cookie));
      not_wash = DGP_FALSE;
    }

    if(!ccstate_isWash(tmp_ccstate)) {
      DGP_ERR_PASS(ccstate_makeWash(tmp_ccstate, cookie));
    }

    DGP_ERR_PASS(ccstate_merge(ccstate, &tmp_ccstate, cur_row_runs, prev_row_runs, cookie));

    /* get index again (it may have moved in the map) */
    cur = ccstate->map_index;

  } while(ccstate_isPotentialWash(ccstate));


  NORMAL_RETURN;
}



ErrCode
dgpcc_washMerge(colorCC_state *cookie)
{
  DGP_PROLOGUE("dgpcc_washMerge()", cookie->DGP_PERRSTATE);
  cccMap *cccmap = cookie->cccmap;
  CCstate *ccstate;
  Dgp_Int32 i, first = -1, last = cccmap->max_num_els;
  cccMapColData *col = cccmap->col;

  /* get the bounds of the cccmap */
  while(col[++first].ccstate == NULL);
  while(col[--last].ccstate == NULL);

  /* the cccmap holds pointers to all ccstates, so we'll
     use this to move through and check all newly
     inactive ccstates for active neighbors. */

  for(i = first; i <= last; i++) {

    ccstate = col[i].ccstate;
  
    /* if this is an empty row, the ccstate is not newly inactive, 
       it couldn't possibly be a wash, or it's background, already a wash, or sampled,
       skip it */
    if((ccstate == NULL) || !ccstate_isNewInactive(ccstate)
       || !ccstate_isPotentialWash(ccstate) || !ccstate_isNormal(ccstate)) continue;

    /* this is the "replaceable" wash merging algorithm 
       (it will probably need to be updated as difficulties
        appear with the current algorithm) */
    DGP_ERR_PASS(dgpcc_doWashMerge1(ccstate, i, first, last, col, cookie));
  }


  NORMAL_RETURN;
}
