/*************************************************************************
**
**  DigiPaper Codebase
**    
**  James Ahlborn, January 1 2000
**
**  Filename: tokenCC.c
**
**  Description of what's implemented in this file:
**    common interface that is used by the lossless token compression
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libsupercc/tokenCC.c,v 1.4 2003/01/27 21:48:39 awilliam Exp $
 *
 */

#include "tokenCC.h"
#include <tok-lossless.h>
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libsupercc/tokenCC.c,v 1.4 2003/01/27 21:48:39 awilliam Exp $");


ErrCode
DgpCC_initPosBlock(Dgp_State *dgp_state, DgpCC_PosInfoBlock **block, Dgp_Int32 num)
{
  DGP_STATE_PROLOGUE2("DgpCC_initPosBlock()");

  DgpCC_PosInfoBlock *pBlock = NULL;

  /* validate params */
  DGP_ERR_BEGIN(dgp_state);
  DGP_CHECK_NULL(block);
  *block = NULL;

  DGP_SEC_CALLOC(pBlock, (DgpCC_PosInfoBlock *), 1, sizeof(DgpCC_PosInfoBlock));

  DGP_SEC_MALLOC(pBlock->poss, (DgpCC_PosInfo *), sizeof(DgpCC_PosInfo) * num);
  pBlock->numAlloced = num;
  
  *block = pBlock;
  pBlock = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(pBlock != NULL) DgpCC_cleanupPosBlock(dgp_state, pBlock);

  ERR_RETURN;
}


ErrCode
DgpCC_cleanupPosBlock(Dgp_State *dgp_state, DgpCC_PosInfoBlock *block)
{
  DGP_STATE_PROLOGUE2("DgpCC_cleanupPosBlock()");
  Dgp_Free tmp_free;

  if(block == NULL) NORMAL_RETURN;
  
  DGP_ERR_BEGIN(dgp_state);

  tmp_free = DGP_FREE;

  if(block->poss != NULL) tmp_free(block->poss);
  tmp_free(block);

  NORMAL_RETURN;
}


ErrCode 
DgpCC_initTokBlock(Dgp_State *dgp_state, DgpCC_TokInfoBlock **block, Dgp_Int32 numToks)
{
  DGP_STATE_PROLOGUE2("DgpCC_initTokBlock()");

  DgpCC_TokInfoBlock *tBlock = NULL;

  /* validate params */
  DGP_ERR_BEGIN(dgp_state);
  DGP_CHECK_NULL(block);
  *block = NULL;

  DGP_SEC_CALLOC(tBlock, (DgpCC_TokInfoBlock *), 1, sizeof(DgpCC_TokInfoBlock));

  DGP_SEC_MALLOC(tBlock->toks, (DgpCC_TokInfo *), sizeof(DgpCC_TokInfo) * numToks);
  tBlock->numAlloced = numToks;

  *block = tBlock;
  tBlock = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(tBlock != NULL) DgpCC_cleanupTokBlock(dgp_state, tBlock);

  ERR_RETURN;
}

ErrCode
DgpCC_cleanupTokBlock(Dgp_State *dgp_state, DgpCC_TokInfoBlock *block)
{
  DGP_STATE_PROLOGUE2("DgpCC_cleanupTokBlock()");
  Dgp_Free tmp_free;

  if(block == NULL) NORMAL_RETURN;
  
  DGP_ERR_BEGIN(dgp_state);

  tmp_free = DGP_FREE;

  if(block->toks != NULL) tmp_free(block->toks);
  tmp_free(block);

  NORMAL_RETURN;
}


ErrCode
DgpCC_tokenAddCC(Dgp_State			*dgp_state, 
				 LlTokState			tokstate, 
				 DgpCC_TokInfoBlock *tBlock, 
				 LlRunToken			*rTok, 
                 token				**tok)
{
  DGP_STATE_PROLOGUE2("DgpCC_tokenAddCC()");

  Dgp_Int32 oldOffset;


  /* check some params */
  DGP_ERR_BEGIN(dgp_state);
  DGP_CHECK_NULL(tokstate);
  DGP_CHECK_NULL(tBlock);
  DGP_CHECK_NULL(rTok);
  DGP_CHECK_NULL(tok);
  
  *tok = NULL;

  if(tBlock->num >= tBlock->numAlloced){
    tBlock->numAlloced = tBlock->numAlloced * 3 / 2;
    DGP_SEC_REALLOC(tBlock->toks, (DgpCC_TokInfo *), tBlock->numAlloced * sizeof(DgpCC_TokInfo));
  }
  
  /* 
   * This is a bit of a hack
   * oldOffset can be one of three things:
   *
   * 1. if the token definitely matches a previous token then oldOffset will be
   * the userdata set when that token was added
   *
   * 2. if the token definitely doesn't match anything then oldOffset will be 
   * the the userdata set on this call
   *
   * 3. if the tokenizer can't say right now whether the token matches or doesn't
   * match then oldOffset will be 0
   *
   * This code makes sense for the first two cases, but its a bit silly for the
   * last case.  In reality the DgpLlAddToken called for a given tokBlock will
   * have some predictable behavior:  
   * One version of DgpLlAddToken will only do the first 2 cases.
   * The other version will only do the last case.  
   * 
   * If the last case is the only one executed for a given tok block, 
   * there will be only one valid tokenInfo for the block [0].  It will
   * have a null token* thus giving all the positions null token* which is
   * just what we want
   */

  tBlock->toks[tBlock->num].token = NULL;

  DGP_ERR_PASS_CLEANUP(
           DgpLlAddToken(tokstate, rTok, (void *)tBlock->num, (void **)&oldOffset));
    
  if(oldOffset == tBlock->num){
    tBlock->num++;
  } 

  *tok = tBlock->toks[oldOffset].token;
  
  NORMAL_RETURN;

ERR_CLEANUP:
  ERR_RETURN;
}
