#ifndef __BASIC_STRUCTS_H__
#define __BASIC_STRUCTS_H__

#include "image.h"
#include <math.h>
#include "resolution.h"
#include "util.h"
typedef struct
{
  Dgp_Int32 cx;
  Dgp_Int32 cy;
}Dgp_Size;

typedef struct
{
  Dgp_Int32 l;
  Dgp_Int32 t;
  Dgp_Int32 r;
  Dgp_Int32 b;
}Dgp_Rect;


INLINE Dgp_Rect rect_init(Dgp_Int32 t, Dgp_Int32 l, Dgp_Int32 b, Dgp_Int32 r)
{
  Dgp_Rect out;
  out.t=t;
  out.l=l;
  out.b=b;
  out.r=r;
  return out;
}

INLINE Dgp_Int32 rect_width(const Dgp_Rect r)
{
  return abs(r.r-r.l)+1;
}

INLINE Dgp_Int32 rect_height(const Dgp_Rect r)
{
  return abs(r.b-r.t)+1;
}

INLINE Dgp_Bool rect_do_rects_overlap(const Dgp_Rect r1, const Dgp_Rect r2)
{
  if((r1.l <= r2.r) && (r2.l<=r1.r) && (r1.t<=r2.b) && (r2.t<=r1.b))
    return DGP_TRUE;
  else
    return DGP_FALSE;

}

INLINE void rect_set_empty(Dgp_Rect *r)
{
  r->t=r->b=r->l=r->r=0;
}


INLINE Dgp_Rect rect_intersection(const Dgp_Rect r1, const Dgp_Rect r2)
{
  Dgp_Rect ret;
  if(!rect_do_rects_overlap(r1,r2))
    rect_set_empty(&ret);
  else
    {
      ret.t=MAX(r1.t,r2.t);
      ret.l=MAX(r1.l,r2.l);
      ret.b=MIN(r1.b,r2.b);
      ret.r=MIN(r1.r,r2.r);
    }

  return ret;
}

INLINE Dgp_Bool rect_is_empty(const Dgp_Rect r)
{
  if((r.l==r.r) || (r.t==r.b))
    return DGP_TRUE;
  else
    return DGP_FALSE;
}

INLINE Dgp_Bool rect_are_same(const Dgp_Rect r1,const Dgp_Rect r2)
{
  if((r1.l==r2.l) && (r1.b==r2.b)
     && (r1.t==r2.t) && (r1.r==r2.r))
    return DGP_TRUE;
  else
    return DGP_FALSE;
}

INLINE Dgp_Bool rect_is_pt_in(const Dgp_Rect r,const DgpImage_Point pt)
{
  Dgp_Bool hIn=DGP_FALSE,vIn=DGP_FALSE;
  if(pt.x>=r.l && pt.x<=r.r)
    hIn=DGP_TRUE;
  if(pt.y>=r.t && pt.y<=r.b)
    vIn=DGP_TRUE;
  return (hIn && vIn);
}

INLINE Dgp_Bool rect_is_point_in(const Dgp_Rect r, Dgp_Int32 x, Dgp_Int32 y)
{
  Dgp_Bool hIn=DGP_FALSE,vIn=DGP_FALSE;
  if(x>=r.l && x<=r.r)
    hIn=DGP_TRUE;
  if(y>=r.t && y<=r.b)
    vIn=DGP_TRUE;
  return (hIn && vIn);
}

INLINE Dgp_Bool rect_is_rect_in(const Dgp_Rect r, const Dgp_Rect test)
{
  return (rect_is_point_in(r,test.l,test.t) && rect_is_point_in(r,test.r,test.b));
}

INLINE Dgp_Rect rect_inflate(Dgp_Rect r, Dgp_Int32 xInflate, Dgp_Int32 yInflate)
{
  r.r+=xInflate;
  r.l-=xInflate;
  r.t-=yInflate;
  r.b+=yInflate;
  return r;
}

INLINE Dgp_Rect rect_offset(Dgp_Rect r, Dgp_Int32 xoff, Dgp_Int32 yoff)
{
  r.r+=xoff;
  r.l+=xoff;
  r.t+=yoff;
  r.b+=yoff;
  return r;
}

INLINE Dgp_Rect  rect_rotate_ext(Dgp_Rect r, Dgp_Int8 orient, Dgp_Int32 pageWidth, Dgp_Int32 pageHeight)
{
    Dgp_Rect tr;
    tr = r;
    switch(orient){
    case 0:
        break;
    case 1:
        r.l = tr.t;
        r.t = pageHeight - tr.r -1;
        r.r = tr.b;
        r.b = pageHeight - tr.l -1;
        break;
    case 2:
        r.l = pageWidth - tr.r -1;
        r.t = pageHeight - tr.b -1;
        r.r = pageWidth - tr.l -1;
        r.b = pageHeight - tr.t -1;
        break;
    case 3:
        r.l = pageWidth - tr.b -1;
        r.t = tr.l;
        r.r = pageWidth - tr.t -1;
        r.b = tr.r;
        break;
    default:
        break;
    }
    return r;
}
INLINE Dgp_Rect rect_zoom( Dgp_Rect r, DgpCoord_Resolution zoom){
		r.b = (Dgp_Int32)(r.b * DgpCoord_float(zoom.yRes));
		r.t = (Dgp_Int32)(r.t * DgpCoord_float(zoom.yRes));
		r.l = (Dgp_Int32)(r.l * DgpCoord_float(zoom.xRes));
		r.r = (Dgp_Int32)(r.r * DgpCoord_float(zoom.xRes));
        return r;
}
INLINE Dgp_Rect rect_inverse_zoom( Dgp_Rect r, DgpCoord_Resolution zoom){
		r.b = (Dgp_Int32)(r.b / DgpCoord_float(zoom.yRes));
		r.t = (Dgp_Int32)(r.t / DgpCoord_float(zoom.yRes));
		r.l = (Dgp_Int32)(r.l / DgpCoord_float(zoom.xRes));
		r.r = (Dgp_Int32)(r.r / DgpCoord_float(zoom.xRes));
        return r;
}

INLINE void rect_fix(Dgp_Rect *r){
    Dgp_Int32 temp;
    temp = r->b;
    r->b = MAX(r->t, r->b);
    r->t = MIN(r->t, temp);
    temp = r->l;
    r->l = MIN(r->l, r->r);
    r->r = MAX(r->r, temp);
}
#endif /*__BASIC_STRUCTS_H__ */
