/*
   @(#)$Header: /project/icons/cvsroot/digipaper/src/libtc_util/insertbox.c,v 1.32 2003/01/27 21:48:39 awilliam Exp $ 
*/

/* insertbox.c

   code extracted into this file from tuncomp.c by
   Aaron Stump, Spring '97

   This file contains all the functions tuncomp.c wants for blatting tokens
   onto a page.  It contains the lower-level routines for blatting monochrome,
   color, and sampled color tokens.  These functions are called by the 
   middle-level functions in insertboxes.c.
*/

#include "bit_array.h"
#include "g4.h"
#include "hash_table.h"
#include "image.h"
#include "insertbox.h"
#include "mmcompress.h"
#include "image_defs.h"
#include "simple_utils.h"
#include <stdlib.h>
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtc_util/insertbox.c,v 1.32 2003/01/27 21:48:39 awilliam Exp $");

#define	BITS_UCHAR	8
#define	BITS_ULONG	((int)(BITS_UCHAR * sizeof(Dgp_UInt32)))

/* Used by InsertBoxRGB_scaled(): */
#define GRAY_CODE_WHITE 0
#define GRAY_CODE_LGRAY (0 | (1 << 1))
#define GRAY_CODE_DGRAY (1 | (1 << 1))
#define GRAY_CODE_BLACK 1 

#define BLAT_SCALED_MONO_WHITE 255
#define BLAT_SCALED_MONO_LGRAY 170
#define BLAT_SCALED_MONO_DGRAY 85
#define BLAT_SCALED_MONO_BLACK 0

/* this nasty little macro extracts the (x mod 32)'th bit from the
   word containing the x'th pixel in the row mb. */
#define model_bit_on(mb,x) (((mb)[((x) / BITS_ULONG)] >> (BITS_ULONG - ((x) % BITS_ULONG) - 1)) & 0x1)

/* The color returned will be closer to c1 than c2 */
static DgpImage_RGB 
get_rgb_intermediate(DgpImage_RGB c1, DgpImage_RGB c2) {
  DgpImage_RGB ret;
  
  ret.r = (Dgp_UInt8)(((((int)c1.r) * 2) + c2.r)/3);
  ret.g = (Dgp_UInt8)(((((int)c1.g) * 2) + c2.g)/3);
  ret.b = (Dgp_UInt8)(((((int)c1.b) * 2) + c2.b)/3); 

  return ret;
}

/*
 * Stick the given token in at the given position.
 * Where the bitmap of the token says to draw a pixel, draw
 * the appropriate pixel of the sampled_image
 */
void
InsertSampledBox(DgpImage_ColorRGB im, DgpImage_ColorRGB sampled_image, tokpos *box,
		 token *model)
{
  DGP_VOID_PROLOGUE("InsertSampledBox()", NULL);
  Dgp_UInt32 w, h;
  Dgp_Int32 x,xend,y,yend, x_mb;
  Dgp_UInt32 sx,sy, smaxx; /* sampled image's x, y, and maxx */
  int temp;
  Dgp_UInt32 *mb;

  /* clip the model if it would go off the right or bottom edge of the image */
  w = model->width;
  smaxx = DGPIMG_XMAX(sampled_image);
  if ((temp = DGPIMG_XMAX(im) - (box->xpos + w - 1)) < 0) {
    w += temp;
    smaxx += temp; /* adjust sampled image's maxx, too */
  }

  h = model->height;
  if ((temp = DGPIMG_YMAX(im) - (box->ypos + h - 1)) < 0) {
    h += temp;
  }

  DGP_VOID_ASSERT(w <= (Dgp_UInt32)DGPIMG_WIDTH(sampled_image));
  DGP_VOID_ASSERT(h <= (Dgp_UInt32)DGPIMG_HEIGHT(sampled_image));

  /* Now for all (x,y) in the model, set im(x,y) to rbgcolor when and only 
     when the corresponding pixel in the model's bitmap is on */
  mb = model->bits;
  for (y = box->ypos, yend = box->ypos + h, sy = 0; y < yend; sy++, y++) {
    for (x_mb = w - 1, x = box->xpos + w - 1, sx = smaxx, xend = box->xpos; 
	 x >= xend; 
	 x--,
	   x_mb--, sx--) {
      if (model_bit_on(mb,x_mb)) 
	DGPIMG_REF(im,x,y) = DGPIMG_REF(sampled_image,sx,sy);
    }
    mb += model->longsperline;
  }

  VOID_RETURN;
}

/*
  Draw the appropriate pixel of the sampled_image wherever either
  token's bitmap has an on pixel.
 */
void
InsertSampledBox_scaled(DgpImage_ColorRGB im, DgpImage_ColorRGB sampled_image, tokpos *box,
			token *model, token *model2)
{
  Dgp_UInt32 w, h;
  Dgp_Int32 x,xend,y,yend, x_mb;
  Dgp_UInt32 sx,sy, smaxx; /* sampled image's x, y, and maxx */
  int temp, gray_code;
  int longsperline = model->longsperline;
  Dgp_UInt32 *mb,*mb2;

  /* clip the model if it would go off the right or bottom edge of the image */
  w = model->width;
  smaxx = DGPIMG_XMAX(sampled_image);
  if ((temp = DGPIMG_XMAX(im) - (box->xpos + w - 1)) < 0) {
    w += temp;
    smaxx += temp; /* adjust sampled image's maxx, too */
  }
  h = model->height;
  if ((temp = DGPIMG_YMAX(im) - (box->ypos + h - 1)) < 0) {
    h += temp;
  }

  /* when we scale down the sampled images, sometimes the dimensions
     of the sampled image's mask and the sampled image
     get rounded a little differently.  So we'll just take the
     intersection of the bounding boxes. */
  w = MIN(w,(Dgp_UInt32)DGPIMG_WIDTH(sampled_image));
  h = MIN(h,(Dgp_UInt32)DGPIMG_HEIGHT(sampled_image));

  /* Now for all (x,y) in the model, set im(x,y) to rbgcolor when and only 
     when the corresponding pixel in either model's bitmap is on */
  mb  = model->bits;
  mb2 = model2->bits;
  for (y = box->ypos, yend = box->ypos + h, sy = 0; y < yend; sy++, y++) {
    for (x_mb = w - 1, x = box->xpos + w - 1, sx = smaxx, xend = box->xpos; 
	 x >= xend; 
	 x--,
	   x_mb--, sx--) {

      gray_code = model_bit_on(mb,x_mb) | (model_bit_on(mb2,x_mb) << 1);
      if (gray_code == GRAY_CODE_WHITE)
	continue;
      else if (gray_code == GRAY_CODE_BLACK)
	DGPIMG_REF(im,x,y) = DGPIMG_REF(sampled_image,sx,sy);
      else if (gray_code == GRAY_CODE_DGRAY)
	DGPIMG_REF(im,x,y) = get_rgb_intermediate(DGPIMG_REF(sampled_image,sx,sy),
					     DGPIMG_REF(im,x,y)); 
      else 
	/* gray_code == GRAY_CODE_LGRAY */
	DGPIMG_REF(im,x,y) = get_rgb_intermediate(DGPIMG_REF(im,x,y),
					     DGPIMG_REF(sampled_image,sx,sy)); 

    }
    mb += longsperline;
    mb2 += longsperline;
  }
}

/*
 * Stick the given token in at the given position with the given color.
 */
void
InsertBoxRGB(DgpImage_ColorRGB im, tokpos *box, token *model, DgpImage_RGB rgbcolor)
{
  Dgp_UInt32 w, h;
  Dgp_Int32 x,xend,y,yend, x_mb;
  int temp;
  Dgp_UInt32 *mb;

  /* clip the model if it would go off the right or bottom edge of the image */
  w = model->width;
  if ((temp = DGPIMG_XMAX(im) - (box->xpos + w - 1)) < 0)
    w += temp;
  
  h = model->height;
  if ((temp = DGPIMG_YMAX(im) - (box->ypos + h - 1)) < 0)
    h += temp;

  /* Now for all (x,y) in the model, set im(x,y) to rbgcolor when and only 
     when the corresponding pixel in the model's bitmap is on */
  mb = model->bits;
  for (y = box->ypos, yend = box->ypos + h; y < yend; y++) {
    for (x_mb = w - 1, x = box->xpos + w - 1, xend = box->xpos; 
	 x>= xend; 
	 x--,
	   x_mb--) {
      if (model_bit_on(mb,x_mb))
	DGPIMG_REF(im,x,y) = rgbcolor;
    }
    mb += model->longsperline;
  }
}

#ifdef DEBUG
static void
write_bitmap(Dgp_Int8 *filename, Dgp_UInt32 *im, int pitch_uDgp_Int32s, int height) {
  FILE *im_out = fopen(filename,"w");
  fprintf(im_out,"P4 %d %d\n",pitch_uDgp_Int32s*32,height);
  fwrite(im,4*pitch_uDgp_Int32s * height, 1, im_out);
  fclose(im_out);
}
#endif

/*  Blat the given scaled token onto the page with the given color.

    Wherever the scaled token has black, we draw the given cmyk color.
    At any pixel where the scaled token has one of its grays, we draw 
    in a color intermediate (in proportion to the intensity of the gray)
    between the background color at that pixel and the given cmyk color.
    Such intermediate values are determined by calling 
    get_rgb_intermediate().

    im     -- the image to blat the token onto.
    box    -- the position telling where the token should be blatted.
    model1 -- contains the first bitplane of the scaled token.
    model2 -- contains the second bitplane of the scaled token.
    cmyk   -- the color to blat the scaled token in.

    NOTE model1 and model2 must have the same dimensions (but we won't check
    for that condition).
 */
void
InsertBoxRGB_scaled(DgpImage_ColorRGB im, tokpos *box, token *model1, token *model2,
		    DgpImage_RGB rgbcolor)
{
  Dgp_UInt32 w, h; 
  Dgp_Int32 x,xend,y,yend, x_mb;
  int temp;
  Dgp_UInt32 *mb1, *mb2;
  int gray_code;
  int longsperline = model1->longsperline;
  
  /* clip the model if it would go off the right or bottom edge of the image */
  w = model1->width;
  if ((temp = DGPIMG_XMAX(im) - (box->xpos + w - 1)) < 0)
    w += temp;
  
  h = model1->height;
  if ((temp = DGPIMG_YMAX(im) - (box->ypos + h - 1)) < 0)
    h += temp;

  /* Now for all (x,y) in the model, set im(x,y) to rbgcolor when and only 
     when the corresponding pixel in the model's bitmap is on */
  mb1 = model1->bits;
  mb2 = model2->bits;
  for (y = box->ypos, yend = box->ypos + h; y < yend; y++) {
    for (x_mb = w - 1, x = box->xpos + w - 1, xend = box->xpos; 
	 x>= xend; 
	 x--,
	   x_mb--) {
      gray_code = model_bit_on(mb1,x_mb) | (model_bit_on(mb2,x_mb) << 1);
      if (gray_code == GRAY_CODE_WHITE)
	continue;
      else if (gray_code == GRAY_CODE_BLACK)
	DGPIMG_REF(im,x,y) = rgbcolor;
      else if (gray_code == GRAY_CODE_DGRAY)
	DGPIMG_REF(im,x,y) = get_rgb_intermediate(rgbcolor,DGPIMG_REF(im,x,y)); 
      else 
	/* gray_code == GRAY_CODE_LGRAY */
	DGPIMG_REF(im,x,y) = get_rgb_intermediate(DGPIMG_REF(im,x,y),rgbcolor); 
    }
    mb1 += longsperline;
    mb2 += longsperline;
  }
}

/*  Blat the given scaled token onto the page.

    im     -- the image to blat the token onto.
    box    -- the position telling where the token should be blatted.
    model1 -- contains the first bitplane of the scaled token.
    model2 -- contains the second bitplane of the scaled token.

    NOTE model1 and model2 must have the same dimensions (but we won't check
    for that condition).
 */
void
InsertBoxScaled(DgpImage_Gray im, tokpos *box, token *model1, token *model2)
{
  Dgp_UInt32 w, h;
  Dgp_Int32 x,xend,y,yend, x_mb;
  int temp;
  Dgp_UInt32 *mb1, *mb2;
  int gray_code;
  int longsperline = model1->longsperline;

  /* clip the model if it would go off the right or bottom edge of the image */
  w = model1->width;
  if ((temp = DGPIMG_XMAX(im) - (box->xpos + w - 1)) < 0)
    w += temp;
  
  h = model1->height;
  if ((temp = DGPIMG_YMAX(im) - (box->ypos + h - 1)) < 0)
    h += temp;

  /* Now for all (x,y) in the model, set im(x,y) to rbgcolor when and only 
     when the corresponding pixel in the model's bitmap is on */
  mb1 = model1->bits;
  mb2 = model2->bits;
  for (y = box->ypos, yend = box->ypos + h; y < yend; y++) {
    for (x_mb = w - 1, x = box->xpos + w - 1, xend = box->xpos; 
	 x>= xend; 
	 x--,
	   x_mb--) {
#if 0
      if (model_bit_on(mb1,x_mb))
	DGPIMG_REF(im,x,y) = BLAT_SCALED_MONO_BLACK;
#else
      gray_code  = model_bit_on(mb1,x_mb);
      gray_code |= (model_bit_on(mb2,x_mb)) << 1;
      if (gray_code == GRAY_CODE_WHITE)
	continue;
      else if (gray_code == GRAY_CODE_BLACK)
	DGPIMG_REF(im,x,y) = BLAT_SCALED_MONO_BLACK;
      else if (gray_code == GRAY_CODE_DGRAY)
	DGPIMG_REF(im,x,y) = BLAT_SCALED_MONO_DGRAY; 
      else 
	/* gray_code == GRAY_CODE_LGRAY */
	DGPIMG_REF(im,x,y) = BLAT_SCALED_MONO_LGRAY; 
#endif
    }
    mb1 += longsperline;
    mb2 += longsperline;
  }
}


/* This function takes a single-use token and calls g4Decode() to decompress
   its bitmap. It also sets all model's fields to be what the backend wants 
   them to be. It passes back a pointer to what model->bits used to
   point to (which the caller needs to free). */
ErrCode
decompressSingleUseToken(token *model, void **old_bits, Dgp_State *dgp_state) {
  DGP_PROLOGUE("decompressSingleUseToken()", DGP_PERRSTATE);
  DgpImage_Packed im = NULL;

  DGP_ERR_PASS_CLEANUP(DgpImage_new(dgp_state, DGPIMAGE_PACKED,model->width,model->height, (void**) &im));

  DGP_ERR_PASS_CLEANUP(InsertCompressedBoxPacked(im,0,0,model->width,model->height,
				  model->bits, model->uses, DGP_FALSE, dgp_state));

  *old_bits = model->bits;
  model->bits = DGPIMG_PIXPTR(im,0,0);
  model->longsperline = (Dgp_UInt16)model->uses;
  model->uses = 0;

  NORMAL_RETURN;

ERR_CLEANUP:

  if (im)
    DgpImage_free(im);
  *old_bits = NULL;

  ERR_RETURN;
}

/* This function takes a single-use token and calls g4Decode() to decompress
   its bitmap. It also sets all model's fields to be what the backend wants 
   them to be. It passes back a pointer to what model->bits used to
   point to (which the caller needs to free). */
ErrCode
decompressSingleUseTokenC(token *model, void **old_bits, void *decode_cookie,
                          Dgp_State *dgp_state) {
  DGP_PROLOGUE("decompressSingleUseTokenC()", DGP_PERRSTATE);
  DgpImage_Packed im = NULL;

  DGP_ERR_PASS_CLEANUP(DgpImage_new(dgp_state, DGPIMAGE_PACKED,model->width,model->height, (void **)&im));

  DGP_ERR_PASS_CLEANUP(InsertCompressedBoxPackedC(im,0,0,model->width,model->height,
				  model->bits, model->uses, DGP_FALSE, decode_cookie, dgp_state));

  *old_bits = model->bits;
  model->bits = DGPIMG_PIXPTR(im,0,0);
  model->longsperline = (Dgp_UInt16)model->uses;
  model->uses = 0;

  NORMAL_RETURN;

ERR_CLEANUP:

  if (im)
    DgpImage_free(im);
  *old_bits = NULL;

  ERR_RETURN;
}

/* InsertCompressedBoxPacked()

   Parameters:

   im          -- the image to insert the g4-compressed bitmap into
   (xpos,ypos) -- the location in im at which to insert the
                  g4-compressed bitmap 
   width       -- the width of the g4-compressed bitmap
   height      -- the height of the g4-compressed bitmap
   bits        -- the g4-compressed bitmap
   size_bits   -- the length of the g4-compressed bitmap in bytes
   framed      -- whether or not im is framed. 

   Return value:
   This returns DGP_TRUE iff g4Decode() did.
*/
ErrCode
InsertCompressedBoxPacked(DgpImage_Packed im, int xpos, int ypos,
			  int width, int height, void *bits,
			  size_t size_bits, Dgp_Bool framed,
			  Dgp_State *dgp_state)
{
  DGP_PROLOGUE("InsertCompressedBoxPacked()", DGP_PERRSTATE);

  /* Uncompress the bitmap directly into the image.

     Here's why the 1st and 7th arguments to g4Decode() are as they are.  

     1. The first parameter is a pointer to the part of the image at
        which to start inserting the bitmap.  So we have to get a
	pointer to the appropriate row of the image (taking into account
	that the image might be framed).
     2. The seventh parameter is (I'm pretty sure) the maximum y-coordinate
        that any pixel touched by g4Decode() can have.  Since we gave
	g4Decode() a pointer to the image at the ypos'th row, we need to
	subtract ypos from the height of the image (minus 1) to give the max
	y-coordinate to touch, given the pointer we gave g4Decode() for the
	image. */
  DGP_ERR_PASS(g4Decode((Dgp_UInt32 *)DGPIMG_GETROW(im, ypos + (framed ? 32 : 0)),
		  width, height,
		  (int)(DGPIMG_GETROWWIDTH(im) / sizeof(Dgp_UInt32)),
		  xpos + (framed ? 32 : 0), DGPIMG_WIDTH(im) - 1,
		  DGPIMG_HEIGHT(im) - ypos - (framed ? 32 : 0) - 1,
		  (Dgp_UInt8 *)bits, size_bits, DGP_FALSE, (size_t *)NULL,
		  dgp_state));

  NORMAL_RETURN;
}


/* InsertCompressedBoxPacked()

   Parameters:

   im          -- the image to insert the g4-compressed bitmap into
   (xpos,ypos) -- the location in im at which to insert the
                  g4-compressed bitmap 
   width       -- the width of the g4-compressed bitmap
   height      -- the height of the g4-compressed bitmap
   bits        -- the g4-compressed bitmap
   size_bits   -- the length of the g4-compressed bitmap in bytes
   framed      -- whether or not im is framed. 
   decode_cookie -- g4 decode cookie
   dgp_state   -- digipaper state

   Return value:
   This returns DGP_TRUE iff g4Decode() did.
*/
ErrCode
InsertCompressedBoxPackedC(DgpImage_Packed im, int xpos, int ypos,
			  int width, int height, void *bits,
			  size_t size_bits, Dgp_Bool framed, void *decode_cookie,
			  Dgp_State *dgp_state)
{
  DGP_PROLOGUE("InsertCompressedBoxPackedC()", DGP_PERRSTATE);

  /* Uncompress the bitmap directly into the image.

     Here's why the 1st and 7th arguments to g4Decode() are as they are.  

     1. The first parameter is a pointer to the part of the image at
        which to start inserting the bitmap.  So we have to get a
	pointer to the appropriate row of the image (taking into account
	that the image might be framed).
     2. The seventh parameter is (I'm pretty sure) the maximum y-coordinate
        that any pixel touched by g4Decode() can have.  Since we gave
	g4Decode() a pointer to the image at the ypos'th row, we need to
	subtract ypos from the height of the image (minus 1) to give the max
	y-coordinate to touch, given the pointer we gave g4Decode() for the
	image. */
  DGP_ERR_PASS(g4DecodeC(decode_cookie, (Dgp_UInt32 *)DGPIMG_GETROW(im, ypos + (framed ? 32 : 0)),
		  width, height,
		  (int)(DGPIMG_GETROWWIDTH(im) / sizeof(Dgp_UInt32)),
		  xpos + (framed ? 32 : 0), DGPIMG_WIDTH(im) - 1,
		  DGPIMG_HEIGHT(im) - ypos - (framed ? 32 : 0) - 1,
		  (Dgp_UInt8 *)bits, size_bits, DGP_FALSE, (size_t *)NULL));

  NORMAL_RETURN;
}

/*
 * Stick the given token in at the given position.
 */
void
InsertBoxPacked(DgpImage_Packed im, tokpos *box, token *model, Dgp_Bool framed)
{
  int l, i, j;
  int row_width;		/* width of each row in Dgp_UInt32s */
  Dgp_UInt32 *iptr, *mptr, carry;
  int rshift, lshift;
  Dgp_Bool extraword;
  
  l = box->xpos / BITS_ULONG;	/* which Dgp_Int32 to start at */
  row_width = model->longsperline;
  if (framed) {
    l++;
  }
  
  rshift = (box->xpos % BITS_ULONG);
  lshift = BITS_ULONG - rshift;
  if (model->width + rshift - row_width * BITS_ULONG > 0) {
    extraword = DGP_TRUE;
  }
  else {
    extraword = DGP_FALSE;
  }
  
  if (rshift) {
    for (j = box->ypos;
	 j < MIN(box->ypos + model->height, DGPIMG_YMAX(im)+1); j++) {
      iptr = (Dgp_UInt32 *)DGPIMG_GETROW(im, j + (framed ? 32 : 0)) + l;
      mptr = &(model->bits[(j - box->ypos) * model->longsperline]);
      
      carry = 0;
      for (i = 0; i < row_width; i++) {
	*iptr++ ^= ((*mptr >> rshift) | carry);
	carry = *mptr++ << lshift;
      }
      
      if (extraword) {
	/* Still some bits in carry */
	*iptr ^= carry;
      }
    }
  }
  
  else {
    /* No shifting! */
    for (j = box->ypos;
	 j < MIN(box->ypos + model->height, DGPIMG_YMAX(im)+1); j++) {
      iptr = (Dgp_UInt32 *)DGPIMG_GETROW(im, j + (framed ? 32 : 0)) + l;
      mptr = &(model->bits[(j - box->ypos) * model->longsperline]);
      
      for (i = 0; i < row_width; i++)
	*(iptr++) ^= *(mptr++);
    }
  }
  
}

/* Insert the residual bitmap (as defined by the params resid*) into
   the DgpImage_Packed im. 
*/
void
InsertResidual(DgpImage_Packed im, void *residualbits, int residlen,
	       int residx, int residy, int residw, int residh,
	       Dgp_Bool framed, Dgp_State *dgp_state)
{
  if (residlen == 0) {
    /*
     * It was stored uncompressed.  residualbits points at the uncompressed
     * bitmap, set up for Dgp_Int32-by-Dgp_Int32 access.
     */
    token dummytok;
    tokpos dummypos;
    dummytok.width = residw;
    dummytok.height = residh;
    dummytok.longsperline = (residw + BITS_ULONG - 1) / BITS_ULONG;
    dummytok.bits = (Dgp_UInt32 *)residualbits;

    dummypos.xpos = residx;
    dummypos.ypos = residy;

    InsertBoxPacked(im, &dummypos, &dummytok, framed);
  }
  else {
    /* It was stored compressed - uncompress it directly into the image */
    if (!InsertCompressedBoxPacked(im, residx, residy, residw, residh,
				   residualbits, residlen, framed, dgp_state)) {
      fprintf(stderr, "problem uncompressing residual\n");
      exit(1);
    }
  }
}
