/*
   @(#)$Header: /project/icons/cvsroot/digipaper/src/libtc_util/insertbox.h,v 1.15 2000/09/30 01:25:56 dtracy Exp $ 
*/

#ifndef INSERTBOX_H
#define INSERTBOX_H

#include "bit_array.h"
#include "hash_table.h"
#include "image.h"
#include "mmcompress.h"
#include "dgp_err_control.h"
#include "dgp_state.h"

#ifdef __cplusplus
extern "C"
{
#endif

/*
 * Stick the given token in at the given position.
 * Where the bitmap of the token says to draw a pixel, draw
 * the appropriate pixel of the sampled_image
 */
DGP_LIBEXPORT void
InsertSampledBox(DgpImage_ColorRGB im, DgpImage_ColorRGB sampled_image, tokpos *box,
		 token *model);
/*
  Draw the appropriate pixel of the sampled_image wherever either
  token's bitmap has an on pixel.
 */
DGP_LIBEXPORT void
InsertSampledBox_scaled(DgpImage_ColorRGB im, DgpImage_ColorRGB sampled_image, tokpos *box,
			token *model, token *model2);
/*
 * Stick the given token in at the given position with the given color.
 */
DGP_LIBEXPORT void
InsertBoxRGB(DgpImage_ColorRGB im, tokpos *box, token *model, DgpImage_RGB rgbcolor);


/*  Blat the given scaled token onto the page with the given color.

    Wherever the scaled token has black, we draw the given cmyk color.
    At any pixel where the scaled token has one of its grays, we draw 
    in a color intermediate (in proportion to the intensity of the gray)
    between the background color at that pixel and the given cmyk color.
    Such intermediate values are determined by calling 
    get_rgb_intermediate().

    im     -- the image to blat the token onto.
    box    -- the position telling where the token should be blatted.
    model1 -- contains the first bitplane of the scaled token.
    model2 -- contains the second bitplane of the scaled token.
    cmyk   -- the color to blat the scaled token in.

    NOTE model1 and model2 must have the same dimensions (but we won't check
    for that condition).
 */
DGP_LIBEXPORT void
InsertBoxRGB_scaled(DgpImage_ColorRGB im, tokpos *box, token *model1, token *model2,
		    DgpImage_RGB rgbcolor);

/*  Blat the given scaled token onto the page.

    im     -- the image to blat the token onto.
    box    -- the position telling where the token should be blatted.
    model1 -- contains the first bitplane of the scaled token.
    model2 -- contains the second bitplane of the scaled token.

    NOTE model1 and model2 must have the same dimensions (but we won't check
    for that condition).
 */
DGP_LIBEXPORT void
InsertBoxScaled(DgpImage_Gray im, tokpos *box, token *model1, token *model2);

/*
 * Stick the given token in at the given position.
 * Where the bitmap of the token says to draw a pixel, draw
 * the appropriate pixel of the sampled_image
 */
DGP_LIBEXPORT void
InsertSampledBox(DgpImage_ColorRGB im, DgpImage_ColorRGB sampled_image, tokpos *box,
		 token *model);

/*
 * Stick the given token in at the given position.
 */
DGP_LIBEXPORT void
InsertBoxPacked(DgpImage_Packed im, tokpos *box, token *model, Dgp_Bool framed);

/* InsertCompressedBoxPacked()

   Parameters:

   im          -- the image to insert the g4-compressed bitmap into
   (xpos,ypos) -- the location in im at which to insert the
                  g4-compressed bitmap 
   width       -- the width of the g4-compressed bitmap
   height      -- the height of the g4-compressed bitmap
   bits        -- the g4-compressed bitmap
   size_bits   -- the length of the g4-compressed bitmap in bytes
   framed      -- whether or not im is framed. 

   Return value:
   This returns DGP_TRUE iff g4Decode() did.
*/
DGP_LIBEXPORT ErrCode
InsertCompressedBoxPacked(DgpImage_Packed im, int xpos, int ypos,
			  int width, int height, void *bits,
			  size_t size_bits, Dgp_Bool framed,
			  Dgp_State *dgp_state);

DGP_LIBEXPORT ErrCode
InsertCompressedBoxPackedC(DgpImage_Packed im, int xpos, int ypos,
			  int width, int height, void *bits,
			  size_t size_bits, Dgp_Bool framed, void *decode_cookie,
			  Dgp_State *dgp_state);

/* This function takes a single-use token and calls g4Decode() to decompress
   its bitmap. It also sets all model's fields to be what the backend wants 
   them to be. It passes back a pointer to what model->bits used to
   point to (which the caller needs to free). */
DGP_LIBEXPORT ErrCode
decompressSingleUseToken(token *model, void **old_bits, Dgp_State *dgp_state);

DGP_LIBEXPORT ErrCode
decompressSingleUseTokenC(token *model, void **old_bits, void *decode_cookie,
                         Dgp_State *dgp_state);

/* Insert the residual bitmap (as defined by the params resid*) into
   the DgpImage_Packed im. 
*/
DGP_LIBEXPORT void
InsertResidual(DgpImage_Packed im, void *residualbits, int residlen,
	       int residx, int residy, int residw, int residh,
	       Dgp_Bool framed, Dgp_State *dgp_state);

#ifdef __cplusplus
}
#endif

#endif
