/*******************************************************
 * ip-tiffio.c
 * $Header: /project/icons/cvsroot/digipaper/src/libtc_util/ip-tiffio.c,v 1.16 2003/02/10 22:53:39 mkrocken Exp $
 ******************************************************/
#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>

#include "libiow.pub"
#include "libutilw.pub"
#include "libpixrw.pub"
#include "rotatop.pub"
#include "iotiff.pub"
#include "defines.pub"
#include "color.pub"
#include "image.h"
#include "image_defs.h"
#include "ip-tiffio.h"
#include "uds.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtc_util/ip-tiffio.c,v 1.16 2003/02/10 22:53:39 mkrocken Exp $");


/*******************************************************************
 * Read the TIFF file. Stick its X and Y dimensions into *xSizePtr and
 * *ySizePtr; stick the bytes-per-line count in *bplPtr. Return the pointer
 * to the bits themselves. Die if you can't do this...
 *
 * WARNING: The bitmap has a "frame" around it: a bunch of extra rows at
 * the top and bottom, and some extra bits at the left and right (as well as
 * possible alignment padding on the right). The xsize and ysize values
 * do *not* include this frame; the bytes-per-line value does. The pointer
 * returned is the top left corner of the bitmap *including* the frame; it
 * can be passed to free() to get rid of it.
 *
 * The frame is 32 rows top and bottom, 32 bits left and right.
 *
 * It also returns a PIXR which holds the image; when you want to fry the
 * image, call w_destroyPixr on it and the bitmap will go as well.
 *********************************************************************/
ErrCode
readTIFF(	Dgp_Int8 *name, 
		int rot, 
		Int32 *xSizePtr, 
		Int32 *ySizePtr,
	 	Int32 *bplPtr, 
		PIXR **_readPixr, 
		Dgp_State *dgp_state)
{
  DGP_PROLOGUE("readTIFF()", DGP_PERRSTATE);
  PIXR *tempPixr = (PIXR *)NULL;
  PIXR *readPixr = (PIXR *)NULL;
  PIXR *pPrD = (PIXR *) NULL;
  IPWState *ipwstate = DGP_PIPWSTATE;
  ColorMap *pMap;

  
  DGP_CHECK_NULL(_readPixr);
  

  DGP_ERR_PASS_CLEANUP(w_read(ipwstate, name, 1, &readPixr));

  /* Rotate the pixr if need */
  rot = rot%4;
  if(rot != 0) {

    tempPixr = readPixr;
    readPixr = NULL;

    switch(rot){
    case 1:
      DGP_ERR_PASS_CLEANUP(w_rotate90(ipwstate, tempPixr, &readPixr));
      break;
    case 2:
      DGP_ERR_PASS_CLEANUP(w_rotate180(ipwstate, tempPixr, &readPixr));
      break;
    case 3:
      DGP_ERR_PASS_CLEANUP(w_rotate270(ipwstate, tempPixr, &readPixr));
      break;
    }

    /* and don't need the image any more  - toast the PIXR */
    w_destroyPixr(ipwstate, &tempPixr);
  }

  /* If this is a indexed color pix, convert it to RGB,  */
  /* since IPCore does not handle indexed color right */
  DGP_ERR_PASS_CLEANUP(w_pixrGetColorMap (ipwstate, readPixr, &pMap));
  if(pMap){ 
    DGP_ERR_PASS_CLEANUP(w_indexedToRGB (ipwstate, readPixr, NULL, pMap, &pPrD));
    DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipwstate, &readPixr));
    readPixr = pPrD; 
    pPrD = (PIXR *)NULL;

  }

  (void)w_pixrGetWidth(ipwstate, readPixr, xSizePtr);
  (void)w_pixrGetHeight(ipwstate, readPixr, ySizePtr);
  (void)w_pixrGetBpl(ipwstate, readPixr, bplPtr);

  *_readPixr = readPixr;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(readPixr != NULL) w_destroyPixr(ipwstate, &readPixr);
  if(tempPixr != NULL) w_destroyPixr(ipwstate, &tempPixr);
  if(pPrD != NULL) w_destroyPixr(ipwstate, &pPrD);

  ERR_RETURN;
}
/*******************************************************************
 * Read the mutipage TIFF file. Stick its X and Y dimensions into *xSizePtr and
 * *ySizePtr; stick the bytes-per-line count in *bplPtr. Return the pointer
 * to the bits themselves. Die if you can't do this...
 *
 * WARNING: The bitmap has a "frame" around it: a bunch of extra rows at
 * the top and bottom, and some extra bits at the left and right (as well as
 * possible alignment padding on the right). The xsize and ysize values
 * do *not* include this frame; the bytes-per-line value does. The pointer
 * returned is the top left corner of the bitmap *including* the frame; it
 * can be passed to free() to get rid of it.
 *
 * The frame is 32 rows top and bottom, 32 bits left and right.
 *
 * It also returns a PIXR which holds the image; when you want to fry the
 * image, call w_destroyPixr on it and the bitmap will go as well.
 *********************************************************************/
ErrCode
readTIFFEx(	Dgp_Int8  *file, 
		Dgp_Int32 rot, 
		Dgp_Int32 pageNum,
		Dgp_Int32 *xSizePtr, 
		Dgp_Int32 *ySizePtr,
	 	Dgp_Int32 *bplPtr, 
		PIXR **_readPixr, 
		Dgp_State *dgp_state)
{
  DGP_PROLOGUE("readTIFFEx()", DGP_PERRSTATE);
  PIXR *tempPixr = (PIXR *)NULL;
  PIXR *readPixr = (PIXR *)NULL;
  PIXR *pPrD = (PIXR *) NULL;
  IPWState *ipwstate = DGP_PIPWSTATE;
  ColorMap *pMap = NULL;
  FH fHeader;

  DGP_CHECK_NULL(_readPixr);
  fHeader.filename = file; 
  DGP_ERR_PASS_CLEANUP(w_readTiff(	ipwstate, 
					&fHeader, 
					pageNum, 
					cFalse,
					&readPixr));

  /* Rotate the pixr if need */
  rot = rot%4;
  if(rot != 0 && readPixr != (PIXR *) NULL) {

    tempPixr = readPixr;
    readPixr = NULL;

    switch(rot){
    case 1:
      DGP_ERR_PASS_CLEANUP(w_rotate90(ipwstate, tempPixr, &readPixr));
      break;
    case 2:
      DGP_ERR_PASS_CLEANUP(w_rotate180(ipwstate, tempPixr, &readPixr));
      break;
    case 3:
      DGP_ERR_PASS_CLEANUP(w_rotate270(ipwstate, tempPixr, &readPixr));
      break;
    }

    /* and don't need the image any more  - toast the PIXR */
    w_destroyPixr(ipwstate, &tempPixr);
  }

  if(readPixr != (PIXR *) NULL){
      (void)w_pixrGetWidth(ipwstate, readPixr, xSizePtr);
      (void)w_pixrGetHeight(ipwstate, readPixr, ySizePtr);
      (void)w_pixrGetBpl(ipwstate, readPixr, bplPtr);

      /* If this is a indexed color pix, convert it to RGB,  */
      /* since IPCore does not handle indexed color right */
      DGP_ERR_PASS_CLEANUP(w_pixrGetColorMap (ipwstate, readPixr, &pMap));
      if(pMap){ 
        w_indexedToRGB (ipwstate, readPixr, NULL, pMap, &pPrD);
        w_destroyPixr(ipwstate, &readPixr);
        readPixr = pPrD; 
        pPrD = (PIXR *)NULL;

      }
  }
  *_readPixr = readPixr;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(readPixr != NULL) w_destroyPixr(ipwstate, &readPixr);
  if(tempPixr != NULL) w_destroyPixr(ipwstate, &tempPixr);
  if(pPrD != NULL) w_destroyPixr(ipwstate, &pPrD);
  ERR_RETURN;
}


