/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtc_util/run2bitmap.c,v 1.16 2003/02/10 22:53:39 mkrocken Exp $
 *
 */

#include "run2bitmap.h"

#ifndef VXWORKS
#include <malloc.h>
#endif  /* !(VXWORKS) */

#include "util.h"
#include "tok-lossless.h" 
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtc_util/run2bitmap.c,v 1.16 2003/02/10 22:53:39 mkrocken Exp $");

#define BitOn(v, b)        ((v) |= (0x80000000 >> (b)))
#define CharBitOn(v, b)    ((v) |= (0x80 >> (b)))

#define	BITS_UCHAR	8

#define GetRunRight(r,i) \
(RunX((r),(i)) + RunLen((r),(i)) - 1)
#define RunToBitmapLookup(b, e)   (lookup[((b) * DGP_BITS_UINT32) + (e)])

/*
 * This lookup table at location lookup[begin * 32 + end] returns the 
 * 32 bit number with ones from the begin'th to the end'th bit, inclusive.
 */
Dgp_UInt32 *
MakeLongRunToBitmapLookup(void)
{
  static Dgp_UInt32 table[DGP_BITS_UINT32 * DGP_BITS_UINT32] = {0};
  static Dgp_Bool calcd = DGP_FALSE;
  int begin, end;
  Dgp_UInt32 *tp, *tprev;
  int b;
  
  if (calcd) {
    return(table);
  }
  calcd = DGP_TRUE;
  
  for (begin = 0; begin < DGP_BITS_UINT32; begin++) {
    for (end = begin; end < DGP_BITS_UINT32; end++) {
      tp    = table + (begin*DGP_BITS_UINT32 + end);
      tprev = table + (end*DGP_BITS_UINT32 + begin);
      for (b = begin; b <= end; b++) {
	BitOn(*tp,    b);
	BitOn(*tprev, b);
      }
    }
  }

  return(table);
}

#if 0
/*
 * This lookup table at location lookup[begin * 8 + end] returns the 
 * 8 bit number with ones from the begin'th to the end'th bit, inclusive.
 */
Dgp_UInt8 *
MakeCharRunToBitmapLookup(void)
{
  static Dgp_UInt8 table[BITS_UCHAR * BITS_UCHAR] = {0};
  static Dgp_Bool calcd = DGP_FALSE;
  int begin, end;
  Dgp_UInt8 *tp, *tprev;
  int b;
  
  if (calcd) {
    return(table);
  }
  calcd = DGP_TRUE;
  
  for (begin = 0; begin < BITS_UCHAR; begin++) {
    for (end = begin; end < BITS_UCHAR; end++) {
      tp    = table + (begin*BITS_UCHAR + end);
      tprev = table + (end*BITS_UCHAR + begin);
      for (b = begin; b <= end; b++) {
	CharBitOn(*tp,    b);
	CharBitOn(*tprev, b);
      }
    }
  }

  return(table);
}
#endif

/* Given some runs (in the form required by tokAddToken), make the appropriate
   unframed bitmap */
ErrCode
RunsToUnframedBitmap(Dgp_UInt32 **_bits, void *_runs, int num_runs, 
		     int height, int Dgp_Int32sperline, int x0, Dgp_State *dgp_state)
{
  DGP_PROLOGUE("RunsToUnframedBitmap()", DGP_PERRSTATE);

  run_elt *runs = _runs;
  Dgp_UInt32 *mp;
  static Dgp_UInt32 *lookup = (Dgp_UInt32 *)NULL;
  int lastRunRight, bitEnd, bitStart, runLeft, temp;
  int i, b;
  int run_len;
  Dgp_UInt32 *bits_end;
  Dgp_UInt32 *bits = NULL;
  Dgp_Allocators *a = &(dgp_state->a);
  
  if (!lookup) {
    DGP_CHECK_NULL_CLEANUP(lookup = MakeLongRunToBitmapLookup());
  }
  
  
  /* allocate the image so it will not have a frame */
  DGP_CHECK_ALLOC_CLEANUP(bits =
    (Dgp_UInt32 *)(a->calloc)((size_t)Dgp_Int32sperline * height,
			    sizeof(Dgp_UInt32)));

  bits_end = bits + Dgp_Int32sperline*height;
  lastRunRight = -100; /* anything that won't allow execution to
			pass through the first guard in the following
			loop the first time through it */

  for (mp = bits,i = 0; i < num_runs; i++) {
    
    runLeft = RunX(runs,i);
    run_len = RunLen(runs,i);
    
    if ((runLeft <= lastRunRight + 1) && (run_len>0)) {
      /* we should skip to the next line in memory (by the principle that
	 we wrap our runs around to the next line just by making the x-location
	 appropriate for the start of the line). */
      mp += Dgp_Int32sperline; /* NOTE that we are moving forward
			     through memory rather than backward
			     (as we do in RunsToBitmap), since
			     we've already reversed the order
			     of the runs. */
    } else if (run_len == 0) {
      /* we need to check to see whether or not the given run encodes a
	 request to skip forward to some y. */

	/* the length of this run is 0 -- it must indicate a change 
	   in y-position */
	mp = bits + Dgp_Int32sperline * runLeft;
	lastRunRight=-100; /*reset the implicit skip switch if we did an explicit skip */
	continue;
    }

    lastRunRight = GetRunRight(runs,i);

    bitStart = runLeft - x0;
    bitEnd   = lastRunRight - x0;

    if ((b = bitStart / DGP_BITS_UINT32) == (temp = bitEnd / DGP_BITS_UINT32)) {
      /* this is a Dgp_Int16 run which we can lookup in one call */
      /*make it stop before the assert so I can look at whats */
      /*wrong. */
	
      DGP_ASSERT(&mp[b] < bits_end);
      mp[b] |=
	RunToBitmapLookup(bitStart % DGP_BITS_UINT32, bitEnd % DGP_BITS_UINT32);
    }
    else {
      /* this run extends through more than one uDgp_Int32 */
      mp[b] |= RunToBitmapLookup(bitStart % DGP_BITS_UINT32, DGP_BITS_UINT32 - 1);
      for (b = bitStart / DGP_BITS_UINT32 + 1; b < temp; b++) {
	DGP_ASSERT(&mp[b] < bits_end);
	mp[b] = ~(0ul);
      }
      DGP_ASSERT(&mp[b] < bits_end);
      mp[b] |= RunToBitmapLookup(0, bitEnd % DGP_BITS_UINT32);
    }
  }

  *_bits = bits;
  bits = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(bits) (a->free)(bits);

  ERR_RETURN;
}
