/*
   @(#)$Header: /project/icons/cvsroot/digipaper/src/libtc_util/simple_utils.c,v 1.39 2003/02/10 22:53:39 mkrocken Exp $ 
*/

/* util.c

   Aaron Stump, Spring '97

   This file defines some little utilities that might be generally
   useful.
*/

#if 0
#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif
#endif /* 0 */

#include "simple_utils.h"
#include "image_defs.h"
#include "insertbox.h"
#include <stdio.h>
#include <stdlib.h>
#include "util.h"

#define HAVE_BOOLEAN
#include "libutilw.pub"
#include "libpixrw.pub"
#include "libiow.pub"
#include "libscalw.pub"
#include "PsgFix.h"


DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtc_util/simple_utils.c,v 1.39 2003/02/10 22:53:39 mkrocken Exp $");


/* Find the maximum element in an array of integers. size is the number
   of elements in the array, and must be > 0 (although no error checking
   will be performed). */
Dgp_Int32 
find_max_of_ints(Dgp_Int32 *array, Dgp_Int32 size) {
  Dgp_Int32 *ptr,*end; /* counter */
  Dgp_Int32 max = array[0];

  ptr = array+1;
  end = array + size;
  for (;ptr < end; ptr++)
    if (*ptr > max)
      max = *ptr;
  return max;
}

#if 0
/* This DictExtBuilder (see writeXiff.h) treats its extcookie as an 
   Extension (see readXiff.h), and just passes back that Extension's info
   for the extension it's been asked to build. */
ErrCode
dummy_DictExtBuilder(token_extension *sorted_extdata,
		     Dgp_Int32 ntok,
		     void *extcookie,
		     void **extension,
		     size_t *extlen) {
  Extension *e = (Extension *)extcookie;

  (void)sorted_extdata;
  (void)ntok;

  *extension = e->extension;
  *extlen    = e->length;
  return DGP_TRUE;
}

/* This PosExtBuilder (see writeXiff.h) treats its extcookie as an 
   Extension (see readXiff.h), and just passes back that Extension's info
   for the extension it's been asked to build. */
ErrCode
dummy_PosExtBuilder(tokpos_extension *sorted_extdata,
		    Dgp_Int32 ntok,
		    void *extcookie,
		    void **extension,
		    size_t *extlen) {
  Extension *e = (Extension *)extcookie;

  (void)sorted_extdata;
  (void)ntok;

  *extension = e->extension;
  *extlen    = e->length;
  return DGP_TRUE;
}
#endif

/* decompressAnyCompressedTokens()

   This function goes through an array of tokens and uncompresses
   any single-use tokens (single-use tokens have their bitmaps 
   g4-compressed). If it runs into problems uncompressing a token, 
   it may bail before it's uncompressed them all.  You need to
   supply a routine to free the compressed data after the bitmap
   has been uncompressed out of it. */
ErrCode 
decompressAnyCompressedTokens(token *toks, Dgp_Int32 ntok, Dgp_State *dgp_state) 
{
  DGP_PROLOGUE("decompressAnyCompressedTokens()", DGP_PERRSTATE);

  Dgp_Int32 i;
  void *old_bits;

  for (i = 0; i < ntok; i++) {
    if ((toks[i].longsperline == 0) && (toks[i].width > 0)) {
      DGP_ERR_PASS(decompressSingleUseToken(&toks[i],&old_bits, dgp_state)); 
      (DGP_FREE)(old_bits);
    }
  }
  
  NORMAL_RETURN;
}

/*
 * When converting a framed packed image to a pixr the pixr might shift offset
 * to align it on a 8 byte boundary.  So if you want to convert a packed image
 * to a framed pixr you should create the packed image with this function
 * it will take into account possible 64 bit padding and adding an extra 4 bytes 
 * at the end of the image buffer incase the image needs to be shifted.
 */
ErrCode
pimNewFrameAligned(Dgp_State *dgp_state, Dgp_Int32 width, Dgp_Int32 height, 
			       Dgp_Int32 *offset, DgpImage_Packed *_pim)
{
  DGP_PROLOGUE("pDgpImage_newFrameAligned()", DGP_PERRSTATE);
	Int32 bufSize;
	Int32 bpl;
	Int32 offsetTo00;
	Dgp_UInt32 *imData = NULL;
  IPWState *ipwstate = DGP_PIPWSTATE;
  DgpImage_Packed pim = NULL;
  Dgp_Allocators *a = &(dgp_state->a);
	

	DGP_ERR_PASS_CLEANUP(w_pixrImageProperties(ipwstate, width, height, 1,
		cFrameBits, &bpl, &bufSize));


	DGP_CHECK_ALLOC_CLEANUP(imData = (a->calloc)(bufSize, 1));

	DGP_ERR_PASS_CLEANUP(w_pixrImageOffset(ipwstate, (Dgp_UInt8 *)imData, bpl, 1, cFrameBits,
					&offsetTo00));

  *offset = (offsetTo00 -
				cFrameBits /8 -	/* back to left edge */
				(cFrameBits * bpl)) * /*back to the top */
				8; /*give the value in pixels */

  DGP_ERR_PASS_CLEANUP(DgpImage_wrap(dgp_state, imData, width + cFrameBits * 2, height + cFrameBits * 2, 
		bpl / 4, &pim));
  imData = NULL;

  *_pim = pim;
  pim = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(pim) {
    DgpImage_unwrap(pim, &imData);
  }
  if(imData) (a->free)(imData);

  ERR_RETURN;
}

/*****************************************
 * Convert DgpImage to pixr and wrap it arround TCImage
 */
ErrCode 
tc_convertToPixr(
Dgp_State     *dgp_state, 
void *image, 
PIXR  **pixr)
{
	DGP_PROLOGUE ("tc_convertToPixr", DGP_PERRSTATE);
    DgpImage_Packed img = (DgpImage_Packed) image;

    DGP_ERR_GEN(!img, cErrInvalidParm);
    DGP_ERR_GEN(!pixr, cErrInvalidParm);

    switch(DGPIMG_TYPE(img)){
    case DGPIMAGE_PACKED:
        DGP_ERR_PASS_CLEANUP(noFramedPackedToPixr(
                                    dgp_state,
                                    DGPIMG_PIXPTR(img,  DGPIMG_XBASE(img),  DGPIMG_YBASE(img)),
                                    DGPIMG_WIDTH(img),
                                    DGPIMG_HEIGHT(img),
                                    DGPIMG_GETROWWIDTH(img)/ sizeof(Dgp_Int32),
                                    pixr));
        break;
#ifndef DGP_FAXMODE
    case DGPIMAGE_RGB:
        DGP_ERR_PASS_CLEANUP(rgbimtorgbpixr(dgp_state, (DgpImage_ColorRGB) img,pixr));
        break;
#endif
    case DGPIMAGE_GRAY:
    default:
        printf("do not support this format");

    }
    
    NORMAL_RETURN;

ERR_CLEANUP:

	ERR_RETURN;

}

ErrCode
noFramedPackedToPixr(
              Dgp_State *dgp_state,
              Dgp_Int32 *bits,
              Dgp_Int32 width,
              Dgp_Int32 height,
              Dgp_Int32 longsperline,
              PIXR **pixr)
{
    Dgp_Int32 i, j, bpl;
    Dgp_Int32 *ptr,*srcPtr, *destPtr;
    IPWState    *ipwstate = dgp_state->ipwstate;
    Dgp_Int32 pixrlpl;

   	DGP_PROLOGUE ("packedToPixr", DGP_PERRSTATE);

    DGP_ERR_PASS_CLEANUP(w_createPixr(	
                                      DGP_PIPWSTATE, 
                                      width, 
                                      height, 
                                      1, 
                                      Int32_TO_ScaledInt1616((Int32)300), 
                                      Int32_TO_ScaledInt1616((Int32)300), 
                                      cFrameBits, 
                                      pixr));
    DGP_ERR_PASS_CLEANUP(w_pixrGetBpl(ipwstate, *pixr, &bpl));
    DGP_ERR_PASS_CLEANUP(w_pixrGetImage(ipwstate, *pixr, (Dgp_UInt8 **)&ptr));
    srcPtr = bits;
    destPtr = ptr;
    pixrlpl = bpl / sizeof(Dgp_Int32);
    for(j= 0; j < height; j ++){
        srcPtr = bits + longsperline * j;
        destPtr = ptr + pixrlpl * j;
        for(i = 0; i < longsperline; i++){
            *destPtr = *srcPtr;
            srcPtr ++;
            destPtr ++;
        }
     }
    NORMAL_RETURN;

ERR_CLEANUP:
    if(pixr) w_destroyPixr(DGP_PIPWSTATE, pixr);
    ERR_RETURN;
}

/*
 * Make a pixr from a packedimage;
 * The frame should be part of the packedimage.
 * So if you need to create a pixr with a frame the packedimage (im) sould be
 * (DGPIMG_WIDTH(im) + 2*frame) by  (DGPIMG_HEIGHT(im) + 2*frame)
 *
 *  !!!NOTE!!! if you are making a framed image you should use pixroffset
 *    to figure out if you need an extra x offset for the image
 */
ErrCode
packedtopixr(Dgp_State *dgp_state, DgpImage_Packed im, Dgp_Int32 frame, PIXR **pixr) 
{
  DGP_PROLOGUE("packedtopixr()", DGP_PERRSTATE);

	Dgp_Int32 width, height;
	Dgp_UInt32 *imData;
	Dgp_Int32 bpl;
  IPWState *ipwstate = DGP_PIPWSTATE;

	width = DGPIMG_WIDTH(im);
	height = DGPIMG_HEIGHT(im);
	bpl = DGPIMG_GETROWWIDTH(im);

	DgpImage_unwrap(im, &imData);

	DGP_ERR_PASS_CLEANUP(w_createPixrStructOnly(ipwstate, width - 2*frame, height - 2*frame, 1, 0, 0, frame,
		(UInt8 *) imData, pixr));

	DGP_ERR_PASS_CLEANUP(w_pixrSetBpl(ipwstate, *pixr, bpl));

	/* Might need to set the endian order based on which machine we are on. */
	/* The DgpImage_Packed is done by Dgp_Int32s.  But it seems like the Pixrs are done */
	/* by bytes.  Howerever I can set the endian of the bytes.  */

	NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

#ifndef DGP_FAXMODE

typedef struct {
    Dgp_Float32 r, g, b;
    } RGBDgp_Float32;
/*
 * convert a CMYK color to an RGB
 * we do the shifting to allow this to work on
 * either endian machines
 */
void 
cmyk2rgb(Dgp_UInt32 cmyk, DgpImage_RGB *outCol)
{
  Dgp_Int32 c, m, y, k;
  Dgp_Float64 r, g, b;
  /*
   * The (very approximate) RGB reflectances of the four ink colours
   */
  RGBDgp_Float32 Cink = {0, 1, 1};
  RGBDgp_Float32 Mink = {1, 0, 1};
  RGBDgp_Float32 Yink = {1, 1, 0};
  RGBDgp_Float32 Kink = {0, 0, 0};

  c = (cmyk >> (DGP_BITS_UINT8 * 3)) & 0xff;
  m = (cmyk >> (DGP_BITS_UINT8 * 2)) & 0xff;
  y = (cmyk >> (DGP_BITS_UINT8 * 1)) & 0xff;
  k = (cmyk >> (DGP_BITS_UINT8 * 0)) & 0xff;
  
  r = g = b = 1.;

  r *= (1 - (c/(Dgp_Float64)255) * (1 - Cink.r));
  g *= (1 - (c/(Dgp_Float64)255) * (1 - Cink.g));
  b *= (1 - (c/(Dgp_Float64)255) * (1 - Cink.b));
  
  r *= (1 - (m/(Dgp_Float64)255) * (1 - Mink.r));
  g *= (1 - (m/(Dgp_Float64)255) * (1 - Mink.g));
  b *= (1 - (m/(Dgp_Float64)255) * (1 - Mink.b));
  
  r *= (1 - (y/(Dgp_Float64)255) * (1 - Yink.r));
  g *= (1 - (y/(Dgp_Float64)255) * (1 - Yink.g));
  b *= (1 - (y/(Dgp_Float64)255) * (1 - Yink.b));
  
  r *= (1 - (k/(Dgp_Float64)255) * (1 - Kink.r));
  g *= (1 - (k/(Dgp_Float64)255) * (1 - Kink.g));
  b *= (1 - (k/(Dgp_Float64)255) * (1 - Kink.b));

  outCol->r = (int)(r * 255 + .5);
  outCol->g = (int)(g * 255 + .5);
  outCol->b = (int)(b * 255 + .5);
}


/* convert DgpImage_ColorRGB image to rgb PIXR */



/* convert CMYK Dgp_Int32 image to rgb PIXR using our broken cmyk2rgb() */
ErrCode
limtorgbpixr(Dgp_State *dgp_state, DgpImage_Long lim, PIXR **pixr) 
{
  DGP_PROLOGUE("limtorgbpixr()", DGP_PERRSTATE);

  PIXR *red, *green, *blue;
  UInt8 *rptr, *gptr, *bptr;
  DgpImage_LongPixPtr lptr;
  Dgp_Int32 width, height;
  Dgp_Int32 j, i;
  DgpImage_RGB rgb_val;
  IPWState *ipwstate = DGP_PIPWSTATE;

  width = DGPIMG_WIDTH(lim);
  height = DGPIMG_HEIGHT(lim);
  
  DGP_ERR_PASS_CLEANUP(w_createPixrExtended(ipwstate, width, height, 8, 0, 0, cFrameBits, 3,
			  1, NULL, pixr));
  
  w_pixrGetChannel(ipwstate, *pixr, 0, &red);
  w_pixrGetChannel(ipwstate, *pixr, 1, &green);
  w_pixrGetChannel(ipwstate, *pixr, 2, &blue);

  lptr = DGPIMG_PIXPTR(lim, DGPIMG_XBASE(lim), DGPIMG_YBASE(lim));

  for(j = 0; j < height; j++) {
    w_getLine(ipwstate, red, j, &rptr);
    w_getLine(ipwstate, green, j, &gptr);
    w_getLine(ipwstate, blue, j, &bptr);
    for(i = 0; i < width; i++) {
      cmyk2rgb(*(lptr++), &rgb_val);
      *(rptr++) = rgb_val.r;
      *(gptr++) = rgb_val.g;
      *(bptr++) = rgb_val.b;
    }
  }
  /* guess what? IPCORE got this screwed up view of the world so we might have to flip the */
  /* byte order of the pixer */
  w_flip(ipwstate, *pixr);

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

DgpImage_ColorRGB 
limtorgbim(DgpImage_Long lim) {
  Dgp_Int32 y,x,ymax,xmax;
  DgpImage_ColorRGB rgbim;
  DgpImage_RGB rgbvalue;

  DgpImage_newOffset(lim->dgp_state, DGPIMAGE_RGB,DGPIMG_WIDTH(lim),
				     DGPIMG_HEIGHT(lim),
				     DGPIMG_XBASE(lim), 
				     DGPIMG_YBASE(lim), (void **)&rgbim);
  if (rgbim == (DgpImage_ColorRGB)NULL) {
    /* Couldn't allocate a new DgpImage_ColorRGB */
    return (DgpImage_ColorRGB)NULL;
  }

  ymax = DGPIMG_YMAX(lim);
  xmax = DGPIMG_XMAX(lim);
  for (y = DGPIMG_YBASE(lim); y <= ymax; y++) 
    for (x = DGPIMG_XBASE(lim); x <= xmax; x++) {
      cmyk2rgb(DGPIMG_REF(lim,x,y),&rgbvalue);
      DGPIMG_REF(rgbim,x,y) = rgbvalue;
    }
  
  return rgbim;
}

/* convert rgb PIXR wjr im */
ErrCode
pixrtorgb(Dgp_State *dgp_state, PIXR *pixr, DgpImage_ColorRGB *_im) 
{ 
  DGP_PROLOGUE("pixrtorgb()", DGP_PERRSTATE);

  PIXR *red, *green, *blue;
  UInt8 *rptr, *gptr, *bptr;
  DgpImage_ColorRGBPixPtr dptr;
  Int32 width, height;
  Dgp_Int32 i, j;
  DgpImage_ColorRGB im = NULL;
  IPWState *ipwstate = DGP_PIPWSTATE;


  /* guess what? IPCORE got this screwed up view of the world so we might have to flip the */
  /* byte order of the pixer */
  w_flip(ipwstate, pixr);

  w_pixrGetWidth(ipwstate, pixr, &width);
  w_pixrGetHeight(ipwstate, pixr, &height);
  
  DGP_ERR_PASS_CLEANUP(DgpImage_new(dgp_state, DGPIMAGE_RGB, width, height, (void **)&im));
  
  w_pixrGetChannel(ipwstate, pixr, 0, &red);
  w_pixrGetChannel(ipwstate, pixr, 1, &green);
  w_pixrGetChannel(ipwstate, pixr, 2, &blue);

  dptr = DGPIMG_PIXPTR(im, 0, 0);

  for(j = 0; j < height; j++) {
    w_getLine(ipwstate, red, j, &rptr);
    w_getLine(ipwstate, green, j, &gptr);
    w_getLine(ipwstate, blue, j, &bptr);
    for(i = 0; i < width; i++) {
	  dptr->r = *(rptr++);
      dptr->g = *(gptr++);
      dptr->b = *(bptr++);
	  dptr++;
    }
  }

  /* guess what? IPCORE got this screwed up view of the world so we might have to flip the */
  /* byte order of the pixer */
  w_flip(ipwstate, pixr);

  *_im = im;
  im = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(im == NULL) DgpImage_free(im);

  ERR_RETURN;
}


/* scale_sampled_image()
   
   Parameters:
   DgpImage_ColorRGB curim      -- the image to scale 
   DgpImage_ColorRGB *_scaledim -- the scaled image will be returned in *_scaledim
   Dgp_Float32 scale_factor  -- the amount to scale down by

   Return value:
   DGP_FALSE on error, DGP_TRUE otherwise.
*/
static ErrCode
scale_sampled_image(Dgp_State *dgp_state,
		    DgpImage_ColorRGB curim, DgpImage_ColorRGB *_scaledim, Dgp_Float32 scale_factor) 
{
  DGP_PROLOGUE("scale_sampled_image()", DGP_PERRSTATE);
  DgpImage_ColorRGB scaledim = NULL;
  PIXR *pixr = NULL;
  PIXR *pixrScaled = NULL;
  Dgp_Int32 x,y;
  Dgp_Int32 w,h;
  Dgp_Int32 neww,newh; /* width and height of the scaled version of curim */
  PIXR *red, *green, *blue;
  UInt8 *red_ptr, *green_ptr, *blue_ptr;
  DgpImage_RGB rgb;
  IPWState *ipwstate = DGP_PIPWSTATE;


  neww = (int)((w=DGPIMG_WIDTH(curim)) / scale_factor);
  newh = (int)((h=DGPIMG_HEIGHT(curim)) / scale_factor);

  /* create a new libwjr image to hold the scaled version of curim */
  DGP_ERR_PASS_CLEANUP(DgpImage_new(dgp_state, DGPIMAGE_RGB,neww, newh, (void **)&scaledim));

  /* Create a new pixr, with 3 8-bit channels */
  DGP_ERR_PASS_CLEANUP(w_createPixrExtended(ipwstate, w,h,
			   8, /* the depth of each channel */
			   0, 0, cFrameBits, 
			   3,cColorRGB,NULL, &pixr));
    
  w_pixrGetChannel(ipwstate, pixr, 0, &red);
  w_pixrGetChannel(ipwstate, pixr, 1, &green);
  w_pixrGetChannel(ipwstate, pixr, 2, &blue);

    /* copy DgpImage_ColorRGB into pixr */
  for (y = 0; y < h; y++) {
    w_getLine(ipwstate, red, y, &red_ptr);
    w_getLine(ipwstate, green, y, &green_ptr);
    w_getLine(ipwstate, blue, y, &blue_ptr);
    for (x = 0; x < w; x++,  
	   red_ptr++, green_ptr++, blue_ptr++) {
      *red_ptr    = (rgb = DGPIMG_REF(curim,x,y)).r;
      *green_ptr = rgb.g;
      *blue_ptr  = rgb.b;
    }
  }
    
  /* scale the original pixr down to have dimensions [neww x newh] */
  DGP_ERR_PASS_CLEANUP(w_scaleLinearSize(ipwstate, pixr, neww, newh,
			&pixrScaled));

  /* copy scaled pixr into DgpImage_ColorRGB */
  w_pixrGetChannel(ipwstate, pixrScaled, 0, &red);
  w_pixrGetChannel(ipwstate, pixrScaled, 1, &green);
  w_pixrGetChannel(ipwstate, pixrScaled, 2, &blue);

  for (y = 0; y < newh; y++) {
    w_getLine(ipwstate, red, y, &red_ptr);
    w_getLine(ipwstate, green, y, &green_ptr);
    w_getLine(ipwstate, blue, y, &blue_ptr);
    for (x = 0; x < neww; x++,  
	   red_ptr++, green_ptr++, blue_ptr++) {
      rgb.r = *red_ptr;
      rgb.g = *green_ptr;
      rgb.b = *blue_ptr;
      DGPIMG_REF(scaledim,x,y) = rgb;
    }
  }

  /* free pixrs */
  DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipwstate, &pixr));

  DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipwstate, &pixrScaled));

  *_scaledim = scaledim;
  scaledim = NULL;

  NORMAL_RETURN;
ERR_CLEANUP:
  if (pixr)
    w_destroyPixr(ipwstate,&pixr);
  if (pixrScaled)
    w_destroyPixr(ipwstate,&pixrScaled);
  if (scaledim)
    DgpImage_free(scaledim);
  ERR_RETURN;
}

/* Scale the given 24-bit images (ie, those in the array sampims) down 
   by the given scale_factor.  The scaled down image will have the
   original's width and height divided by scale_factor.  An array of
   the scaled versions will be passed back in *scaled_sampims.  Use
   the given suite of Dgp_Allocators to allocate memory whenever we have a choice
   of what Dgp_Allocators to use.

   Return DGP_FALSE on internal error, and DGP_TRUE otherwise. */
ErrCode
scale_sampled_images(DgpImage_ColorRGB *sampims, Dgp_Int32 num_sampims, 
		     Dgp_Float32 scale_factor,
		     DgpImage_ColorRGB **_scaled_sampims, Dgp_State *dgp_state) 
{
  DGP_PROLOGUE("scale_sampled_images()", DGP_PERRSTATE);
  DgpImage_ColorRGB *scaled_sampims = NULL;
  Dgp_Int32 i=0,j; 
  IPWState *ipwstate = DGP_PIPWSTATE;
  Dgp_Allocators *a = &(dgp_state->a);


  DGP_CHECK_ALLOC(scaled_sampims = (DgpImage_ColorRGB *)
	  (a->malloc)(num_sampims * (sizeof(*scaled_sampims))));

  for (i = 0; i < num_sampims; i++) {
    /* for each sampled image we were given */
    
    DGP_ERR_PASS_CLEANUP(scale_sampled_image(ipwstate,sampims[i],&scaled_sampims[i],
			     scale_factor)) ;

  } /* end for (i=0; ... ; ...) */

  *_scaled_sampims = scaled_sampims;
  scaled_sampims = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:
  if (scaled_sampims) {

    /* free all scaled images created so far */
    for (j = 0; j < i; j++) {
      DgpImage_free(scaled_sampims[j]);
    }
    (a->free)(scaled_sampims);
  }
  ERR_RETURN;
}
  
/* Scale the 24-bit images in the given DgpAdt_List down by the given scale_factor.
   The scaled down image will have the original's width and height
   divided by scale_factor.  A DgpAdt_List of the scaled versions will be
   passed back in *scaled_sampims.  Use the given suite of Dgp_Allocators
   to allocate memory whenever we have a choice of what Dgp_Allocators to use.

   Return DGP_FALSE on internal error, and DGP_TRUE otherwise. */
ErrCode
scale_sampled_imagesL(DgpAdt_List sampims, Dgp_Float32 scale_factor,
		      DgpAdt_List *_scaled_sampims, Dgp_State *dgp_state) 
{
  DGP_PROLOGUE("scale_sampled_imagesL()", DGP_PERRSTATE);

  DgpAdt_List scaled_sampims = NULL;
  DgpAdt_ListNode ln, prev = NULL;
  IPWState *ipwstate = DGP_PIPWSTATE;
  DgpImage_ColorRGB scaledim = NULL;


  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&scaled_sampims, dgp_state));

  DGPADT_FOREACH(ln,sampims) {
    /* for each sampled image we were given */
    
    DGP_ERR_PASS_CLEANUP(scale_sampled_image(ipwstate,DGPADT_LIGET(ln),&scaledim,scale_factor));

    /* we've got to wind up with a list of images that was in the same
       order as the original list, or our JPEG extension for these
       scaled JPEGs will come out wrong. */
    if (prev == NULL) 
      DgpAdt_liAdd(scaled_sampims,scaledim, &prev);
    else 
      DgpAdt_liAddAfter(scaled_sampims,prev,scaledim, &prev);

  } /* end for (i=0; ... ; ...) */

  *_scaled_sampims = scaled_sampims;
  scaled_sampims = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:
  if (scaled_sampims) {

    /* free all scaled images created so far */
    DGPADT_FOREACH(ln,scaled_sampims)
      DgpImage_free((DgpImage_ColorRGB)DGPADT_LIGET(ln));
    DgpAdt_liFree(scaled_sampims);
  }
  ERR_RETURN;
}



ErrCode
rgbimtorgbpixr(Dgp_State *dgp_state, DgpImage_ColorRGB rgbim, PIXR **pixr) 
{
  DGP_PROLOGUE("limtorgbpixr()", DGP_PERRSTATE);

  PIXR *red, *green, *blue;
  UInt8 *rptr, *gptr, *bptr;
  DgpImage_ColorRGBPixPtr rgbptr;
  Dgp_Int32 width, height;
  Dgp_Int32 j, i;
  IPWState *ipwstate = DGP_PIPWSTATE;

  width = DGPIMG_WIDTH(rgbim);
  height = DGPIMG_HEIGHT(rgbim);
  
  DGP_ERR_PASS_CLEANUP(w_createPixrExtendedNoInit(ipwstate, width, height, 8, 0, 0, cFrameBits, 3,
			  1, NULL, pixr));
  
  w_pixrGetChannel(ipwstate, *pixr, 0, &red);
  w_pixrGetChannel(ipwstate, *pixr, 1, &green);
  w_pixrGetChannel(ipwstate, *pixr, 2, &blue);

  rgbptr = DGPIMG_PIXPTR(rgbim, DGPIMG_XBASE(rgbim), DGPIMG_YBASE(rgbim));

  for(j = 0; j < height; j++) {
    w_getLine(ipwstate, red, j, &rptr);
    w_getLine(ipwstate, green, j, &gptr);
    w_getLine(ipwstate, blue, j, &bptr);
    for(i = 0; i < width; i++) {
      *(rptr++) = rgbptr->r;
      *(gptr++) = rgbptr->g;
      *(bptr++) = rgbptr->b;
      rgbptr++;
    }
  }
  /* guess what? IPCORE got this screwed up view of the world so we might have to flip the */
  /* byte order of the pixer */
  w_flip(ipwstate, *pixr);

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}
#endif
