/*
   @(#)$Header: /project/icons/cvsroot/digipaper/src/libtc_util/simple_utils.h,v 1.30 2002/02/21 01:55:28 jlin Exp $ 
*/

/* util.h
   
   Aaron Stump, Spring '97 

   This file defines prototypes for some little utilities that might
   be generally useful.
*/

#ifndef TC_UTIL_H
#define TC_UTIL_H

/* includes, in alphabetical order */
#include "dgp_allocators.h"
#include "image.h"
#include "libiow.pub"
#include "list.h"
#include "mmcompress.h"
#include "util.h"
#include "dgp_state.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* Find the maximum element in an array of integers. size is the number
   of elements in the array, and must be > 0 (although no error checking
   will be performed). */
extern Dgp_Int32 
find_max_of_ints(Dgp_Int32 *array, Dgp_Int32 size);

#if 0
/* This DictExtBuilder (see writeXiff.h) treats its extcookie as an 
   Extension (see readXiff.h), and just passes back that Extension's info
   for the extension it's been asked to build. */
extern ErrCode
dummy_DictExtBuilder(token_extension *sorted_extdata,
		     Dgp_Int32 ntok,
		     void *extcookie,
		     void **extension,
		     size_t *extlen);

/* This PosExtBuilder (see writeXiff.h) treats its extcookie as an 
   Extension (see readXiff.h), and just passes back that Extension's info
   for the extension it's been asked to build. */
extern ErrCode
dummy_PosExtBuilder(tokpos_extension *sorted_extdata,
		    Dgp_Int32 ntok,
		    void *extcookie,
		    void **extension,
		    size_t *extlen);
#endif

/* decompressAnyCompressedTokens()

   This function goes through an array of tokens and uncompresses
   any single-use tokens (single-use tokens have their bitmaps 
   g4-compressed). If it runs into problems uncompressing a token, 
   it may bail before it's uncompressed them all.  You need to
   supply a routine to free the compressed data after the bitmap
   has been uncompressed out of it. */
DGP_LIBEXPORT ErrCode 
decompressAnyCompressedTokens(token *toks, Dgp_Int32 ntok, Dgp_State *dgp_state);


/*
 * convert a CMYK color to an RGB
 * we do the shifting to allow this to work on
 * either endian machines
 */
extern void 
cmyk2rgb(Dgp_UInt32 cmyk, DgpImage_RGB *outCol);

ErrCode
pixrtorgb(Dgp_State *dgp_state, PIXR *pixr, DgpImage_ColorRGB *im);

extern ErrCode
pimNewFrameAligned(Dgp_State *dgp_state, Dgp_Int32 width, Dgp_Int32 height, 
			       Dgp_Int32 *offset, DgpImage_Packed *pim);

ErrCode
packedtopixr(Dgp_State *dgp_state, DgpImage_Packed im, Dgp_Int32 frame, PIXR **pixr);

DGP_LIBEXPORT ErrCode 
tc_convertToPixr(Dgp_State     *dgp_state, void *image, PIXR  **pixr);

ErrCode
noFramedPackedToPixr(
              Dgp_State *dgp_state,
              Dgp_Int32 *bits,
              Dgp_Int32 width,
              Dgp_Int32 height,
              Dgp_Int32 longsperline,
              PIXR **pixr);

extern DgpImage_ColorRGB 
limtorgbim(DgpImage_Long lim);

extern ErrCode
limtorgbpixr(Dgp_State *dgp_state, DgpImage_Long lim, PIXR **pixr);

extern ErrCode
rgbimtorgbpixr(Dgp_State *dgp_state, DgpImage_ColorRGB lim, PIXR **pixr);


/* decompressAnyCompressedTokens()

   This function goes through an array of tokens and uncompresses
   any single-use tokens (single-use tokens have their bitmaps 
   g4-compressed). If it runs into problems uncompressing a token, 
   it may bail before it's uncompressed them all.  You need to
   supply a routine to free the compressed data after the bitmap
   has been uncompressed out of it. */
extern ErrCode 
decompressAnyCompressedTokens(token *toks, Dgp_Int32 ntok, Dgp_State *dgp_state);

/* Scale the given 24-bit images (ie, those in the array sampims) down 
   by the given scale_factor.  The scaled down image will have the
   original's width and height divided by scale_factor.  An array of
   the scaled versions will be passed back in *scaled_sampims.  Use
   the given suite of Dgp_Allocators to allocate memory whenever we have a choice
   of what Dgp_Allocators to use.

   Return DGP_FALSE on internal error, and DGP_TRUE otherwise. */
extern ErrCode
scale_sampled_images(DgpImage_ColorRGB *sampims, Dgp_Int32 num_sampims, 
		     Dgp_Float32 scale_factor,
		     DgpImage_ColorRGB **_scaled_sampims, Dgp_State *dgp_state);

/* Scale the 24-bit images in the given DgpAdt_List down by the given scale_factor.
   The scaled down image will have the original's width and height
   divided by scale_factor.  A DgpAdt_List of the scaled versions will be
   passed back in *scaled_sampims.  Use the given suite of Dgp_Allocators
   to allocate memory whenever we have a choice of what Dgp_Allocators to use.

   Return DGP_FALSE on internal error, and DGP_TRUE otherwise. */
extern ErrCode
scale_sampled_imagesL(DgpAdt_List sampims, Dgp_Float32 scale_factor,
		      DgpAdt_List *_scaled_sampims, Dgp_State *dgp_state);

#ifdef __cplusplus
}
#endif

#endif



