/************************************************** 
 * libtc_util/tc-tiffio.c
 * $Header: /project/icons/cvsroot/digipaper/src/libtc_util/tc-tiffio.c,v 1.16 2003/02/10 22:53:39 mkrocken Exp $
 **************************************************/
#include "dgp_state.h"
#include <errno.h>
#include "dgp_utils.h"
#include "uds.h"

/* include-libwjr*/
#include <util.h>

/* include-ipcore*/
#include <libutilw.pub>
#include <libpixrw.pub>
#include <librastw.pub>
#include <libiow.pub>
#include <props.pub>
#include <frames.pub>
#include "image.h"
#include "image_defs.h"
#ifdef _WIN32
typedef unsigned char u_char;
typedef unsigned short u_short;
typedef unsigned int u_int;
typedef unsigned long u_long;
#endif
#include "tc-tiffio.h"
#include "ip-tiffio.h"
#include "tiffiop.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtc_util/tc-tiffio.c,v 1.16 2003/02/10 22:53:39 mkrocken Exp $");

/* define addLogo function */
extern ErrCode Dgp_addXeroxLogo (PIXR *image, Dgp_State *dgp_state);

/******************************************************* 
 * Read a tiff file and return the image memory block 
 ********************************************************/
DGP_LIBEXPORT ErrCode
tc_readTiff(            Dgp_State * dgp_state,	/* Dgp state cookie	*/
                        Dgp_Int8 *tiffFile,		/* name of tiff file	*/
                        Dgp_Int32 rotation,           /* rotation 0-3         */
                        TCImagePtr *_tcImagePtr)/* image Ptr(with frame)*/
{
    DGPERR_INIT(tc_readTiff);
    PIXR *readPixr = NULL;
	TCImagePtr tcImage;
    ScaledInt1616 xres, yres;
    IPWState *ipwstate = DGP_PIPWSTATE;
	Dgp_Int32 xSize=0,ySize=0,bpl=0,channel,bpp=0;
	UInt8 *image;
	DGP_SEC_MALLOC(tcImage, (TCImage *), sizeof(TCImage));
        DGP_ERR_PASS_CLEANUP(readTIFF(  tiffFile,
                                        rotation,
                                        (Int32 *)&xSize,
                                        (Int32 *)&ySize,
                                        (Int32 *)&bpl,
                                        &readPixr,
                                        dgp_state));
	tcImage->xSize = xSize;
	tcImage->ySize = ySize;
	tcImage->bpl = bpl;
    w_pixrGetDepth(ipwstate, readPixr, (Int32 *)&bpp);
	tcImage->bpp = bpp;
    w_pixrGetChannelCount(ipwstate, readPixr,(Int32 *)&channel);
	tcImage->channel = channel;
	w_pixrGetResolutionDPI(ipwstate, readPixr,&xres, &yres);
	tcImage->xRes = xres >> 16;
	tcImage->yRes = yres >> 16;
    (void)w_pixrGetImageAll(ipwstate, readPixr, ((UInt8 **)&image));
	tcImage->image = image;
	tcImage->readPixrPtr = (TCPIXRPtr)readPixr;
	*_tcImagePtr = tcImage;

    NORMAL_RETURN;

ERR_CLEANUP:
    /* Need to add frees for all allocations. */
	tc_freeImage(dgp_state,tcImage);
        ERR_RETURN;
}

DGP_LIBEXPORT ErrCode
tc_freeImage(	Dgp_State *dgp_state,
		TCImagePtr tcImagePtr){
        DGPERR_INIT(tc_readTiff);
        IPWState *ipwstate = DGP_PIPWSTATE;
	if(tcImagePtr){
		if(tcImagePtr->readPixrPtr){
			DGP_ERR_PASS_CLEANUP(w_destroyPixr(ipwstate,
					(PIXR**)&tcImagePtr->readPixrPtr));
		}
		DGP_FREE(tcImagePtr);
	}
        NORMAL_RETURN;

ERR_CLEANUP:
        /* Need to add frees for all allocations*/

        ERR_RETURN;
}


/******************************************************* 
 * Read a tiff uds and return the image memory block 
 ********************************************************/
DGP_LIBEXPORT ErrCode
tc_readTiffEx(            
Dgp_State * dgp_state,	/* Dgp state cookie	*/
Dgp_Int8 *tiffFile,		/* name of tiff file	*/
Dgp_Int32 rotation,           /* rotation 0-3         */
Dgp_Int32 pagenum,
TCImagePtr *_tcImagePtr)/* image Ptr(with frame)*/

{
    DGPERR_INIT(tc_readTiff);
    PIXR *readPixr=NULL;
	TCImagePtr tcImage = NULL;
    ScaledInt1616 xres, yres;
    IPWState *ipwstate = DGP_PIPWSTATE;
	Dgp_Int32 xSize=0,ySize=0,bpl=0,channel,bpp=0;
	UInt8 *image;
    DGP_ERR_PASS_CLEANUP(readTIFFEx(  tiffFile,
                                      rotation,
                                      pagenum,
                                      (Int32 *)&xSize,
                                      (Int32 *)&ySize,
                                      (Int32 *)&bpl,
                                      &readPixr,
                                      dgp_state));
	if(readPixr == NULL){ /* No image at pageNum specified... */
		*_tcImagePtr = NULL;
	}
	else{

	  Dgp_addXeroxLogo (readPixr, dgp_state);

        DGP_SEC_MALLOC(tcImage, (TCImage *), sizeof(TCImage));
        tcImage->xSize = xSize;
        tcImage->ySize = ySize;
        tcImage->bpl = bpl;
        w_pixrGetDepth(ipwstate, readPixr, (Int32 *)&bpp);
        tcImage->bpp = bpp;
        w_pixrGetChannelCount(ipwstate, readPixr,(Int32 *)&channel);
        tcImage->channel = channel;
        w_pixrGetResolutionDPI(ipwstate, readPixr,&xres, &yres);
        tcImage->xRes = xres >> 16;
        tcImage->yRes = yres >> 16;
        (void)w_pixrGetImageAll(ipwstate, readPixr, ((UInt8 **)&image));
        tcImage->image = image;
        tcImage->readPixrPtr = (TCPIXRPtr)readPixr;
        *_tcImagePtr = tcImage;
    }

    
    NORMAL_RETURN;

ERR_CLEANUP:
        /* Need to add frees for all allocations*/
	if (tcImage)
		tc_freeImage(dgp_state,tcImage);
    ERR_RETURN;
}

/* Right now this function only output color TIF. We use this */
/* function directly instead of ipcore to bypass the unnecessary */
/* twice interleaved RGB and non-interleaved RGB conversion. */
ErrCode
tc_writeColorTiffDirectory( 
               Dgp_State *dgp_state,
               void *image,
               Dgp_Int32 depth,
               Dgp_Int32 channels,
               Dgp_Int32 xdpi,
               Dgp_Int32 ydpi,
               Dgp_Bool  singleStrip,
               Dgp_Int32 stripSize,
               TCTIFFPtr tiff)

{
    DGP_PROLOGUE("tc_writeTiffDirectory()", DGP_PERRSTATE);
    Dgp_Int32 row, imWidth, imHeight;
    Dgp_Int32 tiffBytes, rowsPerStrip, stripNum=0;
    Dgp_UInt8 *buf;
    DgpImage_ColorRGB pageIm = (DgpImage_ColorRGB) image;
    TIFF *tif = (TCTIFFPtr) tiff;

    DGP_ERR_GEN(!image, cErrNullPtr);
    DGP_ERR_GEN(stripSize <0, cErrNullPtr);
    DGP_ERR_GEN(channels != 3 || depth != 8, cErrTypeNotSupported);

    imWidth = DGPIMG_WIDTH(pageIm);
    imHeight = DGPIMG_HEIGHT(pageIm);

    stripSize = MIN(imHeight, stripSize);

    TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, imWidth );
    TIFFSetField(tif, TIFFTAG_IMAGELENGTH, imHeight );
    TIFFSetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, depth);
    TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, channels);
     
    
    
    if (xdpi > 0 &&	ydpi > 0)
    {
    /* dpi has been set.  Note that we have to pass the ADDRESS
    * of the floating pt value of the resolution to TIFFSetField
    * instead of the float itself.  That's because all floats
    * passed as variable args are converted to doubles and
    * acc tries to read them as double without bothering to
    * align them on 64-bit boundaries first, so gets a Bus Error.
    * We changed TIFFSetField so that when setting the tags
    * that expect a float, they expect the address of a float
    * instead.
        */
        float xdpiFloat, ydpiFloat;
        xdpiFloat = (float)(xdpi);
        ydpiFloat = (float)(ydpi);
        TIFFSetField(tif, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
        TIFFSetField(tif, TIFFTAG_XRESOLUTION, &xdpiFloat);
        TIFFSetField(tif, TIFFTAG_YRESOLUTION, &ydpiFloat);
    }
      
    
    TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
    tiffBytes = imWidth * 3;
    
    /* check for TIFF Scanline size discrepancy */
    DGP_ERR_GEN(TIFFScanlineSize(tif) != tiffBytes, cErrInternal);
    
    rowsPerStrip = (singleStrip ? imHeight : stripSize);
    tiffBytes *= rowsPerStrip;

    TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, rowsPerStrip);

    /* we want to write one strip at a time instead of one line at a time */
    /* to save time */
    tif->tif_row = 0;
    for (row = 0; row < imHeight; row +=rowsPerStrip){
        buf = (Dgp_UInt8 *)DGPIMG_PIXPTR(pageIm, 0, row);
        /*DGP_ERR_GEN(TIFFWriteScanline(tif, buf, row, 0) < 0,
                cErrIO_WriteError); */
        DGP_ERR_GEN(TIFFWriteRawStrip(tif, stripNum, buf, tiffBytes) <0, cErrIO_WriteError);
        stripNum++;
        tif->tif_row += rowsPerStrip;
        /* quit if we just did the last strip */
        if (stripNum >= (int) tif->tif_dir.td_nstrips)
              break;
    }
    
#ifndef PSG
    TIFFCloseStruct(tif); 	/* still keep the tif structure*/
#else
    TIFFClose(tif); /*IPCore-PSG doesn't define TIFFCloseStruct.   */
    /*XXX IPCORE check if this breaks. */
#endif
    NORMAL_RETURN;
    
}

ErrCode
tc_writeTiffHeader (
    Dgp_State       *dgp_state,
    Dgp_Int8       *fname,
    TCTIFFPtr	   *tif)
{
    DGP_PROLOGUE("tc_writeTiffHeader()", DGP_PERRSTATE);

    DGP_ERR_GEN(fname == NULL, cErrNullPtr);
    DGP_ERR_GEN(tif == NULL, cErrNullPtr);
    /* --ejaquith added 28sep95--
     * --DDavies moved 11/18/96--
     * We need this to pass our little IPWState ptr around THROUGH the Leffler
     * code.  This happens when you are Packbits or CCITT compressing
     * because we (eventually) call IPWrap routines (outside of Leffler)
     * which need our IPWState ptr.  We also use the IPWState to
     * print error messages.
     * So the solution is to set this up in the TIFFOpen() call!
     */
    *((TIFF **)tif) = TIFFOpen((void *)DGP_PIPWSTATE, fname, "w");
    DGP_ERR_GEN(*tif == NULL, cErrIO_WriteError);
    NORMAL_RETURN;
}

void
tc_cleanupTiff( TCTIFFPtr *tiff)
{
    TIFF *tif = *((TIFF**)tiff);
        if (tif->tif_cleanup)
                (*tif->tif_cleanup)(tif);
        TIFFFreeDirectory(tif);
        if (tif->tif_rawdata && (tif->tif_flags&TIFF_MYBUFFER))
                _TIFFfree(tif, tif->tif_rawdata);
        if (isMapped(tif))
                TIFFUnmapFileContents(tif, tif->tif_base, tif->tif_size);
        (void) TIFFCloseFile(tif);
        _TIFFfree(tif, (char *)tif);
        *tiff = NULL;
}

void
tc_finishTiff( TCTIFFPtr *tiff)
{
    TIFF *tif = *((TIFF **)tiff);
    (void) TIFFCloseFile(tif);
    _TIFFfree(tif, (char *)tif);
    *tiff = NULL;
}


