/******************************************
 * libtc_util/tc-tiffio.h
 * $Header: /project/icons/cvsroot/digipaper/src/libtc_util/tc-tiffio.h,v 1.7 2002/02/21 01:55:28 jlin Exp $
 *****************************************/
#ifndef _TC_TIFFIO_H
#define _TC_TIFFIO_H


#include "dgp_err_control.h"
#include "dgp_state.h"
#include "uds.h"

#ifdef __cplusplus
extern "C"
{
#endif

typedef struct PIXR *TCPIXRPtr;
typedef struct tiff *TCTIFFPtr;

typedef struct{
	Dgp_Int32 xSize;		/* width (returned)     */
	Dgp_Int32 ySize;		/* height (returned)    */
	Dgp_Int32 xRes;		/* horizontal resolution*/
	Dgp_Int32 yRes;		/* vertical resolution  */
	Dgp_Int32 bpl;		/* byte per line        */
	Dgp_Int32 bpp;		/* bit perl pixel       */
	Dgp_Int32 channel;		/* number of channels	*/
	UInt8 *image;
	TCPIXRPtr readPixrPtr;
}TCImage;

typedef TCImage *TCImagePtr;

DGP_LIBEXPORT ErrCode
tc_readTiff(            Dgp_State * dgp_state,  /* Dgp state cookie     */
                        Dgp_Int8 *tiffFile,     /* name of tiff file    */
                        Dgp_Int32 rotation,           /* rotation 0-3         */
                        TCImagePtr *tcImagePtr);/* image Ptr(with frame)*/
DGP_LIBEXPORT ErrCode
tc_readTiffEx(          Dgp_State * dgp_state,  /* Dgp state cookie     */
                        Dgp_Int8 *tiffFile,     /* name of tiff file    */
                        Dgp_Int32 rotation,           /* rotation 0-3         */
            			Dgp_Int32 pagenum,
                        TCImagePtr *tcImagePtr);/* image Ptr(with frame)*/
DGP_LIBEXPORT ErrCode
tc_freeImage(		Dgp_State *dgp_state,
			        TCImagePtr tcImagePtr);

ErrCode
tc_writeTiffHeader (
    Dgp_State       *dgp_state,
    Dgp_Int8       *fname,
    TCTIFFPtr	    *tif);

ErrCode
tc_writeColorTiffDirectory( 
               Dgp_State *dgp_state,
               void *pageIm,
               Dgp_Int32 depth,
               Dgp_Int32 channels,
               Dgp_Int32 xdpi,
               Dgp_Int32 ydpi,
               Dgp_Bool  singleStrip,
               Dgp_Int32 stripSize,
               TCTIFFPtr tif);

void
tc_cleanupTiff( TCTIFFPtr *tiff);

void
tc_finishTiff( TCTIFFPtr *tiff);


#ifdef __cplusplus
}
#endif

#endif



/******************************************************
 * Running Log
 * $Log: tc-tiffio.h,v $
 * Revision 1.7  2002/02/21 01:55:28  jlin
 * add functions to handle write color tiff
 *
 * Revision 1.6  2001/04/18 23:11:10  jlin
 * initUds
 *
 * Revision 1.5  2000/10/11  18:47:12  dtracy
 * Fixed bug:  decoder was writing 8bpp b&w tiffs.
 *
 * Revision 1.4  2000/04/21 22:08:50  jlin
 * Add tc_readTiffEx to support multi-page tiff read
 *
 * Revision 1.3  2000/03/24  22:22:39  dtracy
 * VxWorks-friendly modifications.
 *
 * Revision 1.2  2000/03/04 00:22:40  mkrocken
 *
 * merged repositories
 *
 * Revision 1.2  2000/02/23 22:23:20  mkrocken
 *
 * changes for new util libs
 *
 * Revision 1.1  2000/02/18 22:06:24  mkrocken
 * initial revision
 *
 * Revision 1.1  1999/12/29 20:12:41  jlin
 * Initial version
 *
 *****************************************************/
