/*
  	Copyright  2000 Xerox Corporation. All rights reserved.
*/


#include "halftone.h"
#include "tclydoc.h"
#include "jtokapi_p.h"
#include "jtokapi.h"

#include "tok-lossy.h"
#include "deskew.h"
#include "compressUtil.h"

#include "memorydatasink.h"
#include "segment.h"
#include "jbig2apiCB.h"
#include "writejbigCB.h"

#include "encode-colors.h"
#include "smooth_token.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtclydoc/apicalls.c,v 1.44 2003/02/11 01:31:44 mkrocken Exp $");

static ErrCode
api_createPosBlock (DgpApi_State *state, PosExtension *posExt);

static void
pageCleanup( 	DgpAdt_List firstOutlier, 
				Dgp_Int32 numOutliers, 
				DgpAdt_List skewedToken, 
				DgpAdt_List firstNewBitmap, 
				Dgp_Int32 currentPage, 
				token *llToks,
				Dgp_Int32 numLlToks,
				Dgp_State *dgp_state);

ErrCode
DgpApi_startSection (void* handle)
{

  DgpApi_State *state = (DgpApi_State*) handle;
  Dgp_State *dgp_state = state->tout->dgp_state;
  DGP_STATE_PROLOGUE ("DgpApi_startSection", dgp_state);

  TCLYSectionBlockPtr sbPtr = NULL;
  DgpTdc_SectionBlock *llSb = NULL;
  DGP_ASSERT (state->handle_check == (size_t) state);

  /* if we are not in a section, we have to start one */
  if (state->status == DGPAPI_NO_SECTION) {
	  /* is there a previous block ? If so, get a pointer to it */
	  if (state->OutputPageCount > 0){
		  sbPtr = (TCLYSectionBlockPtr) state->tdoc->curBlock;
		  llSb = state->llTdoc->curBlock;
	  }
	  /* start a new section */
	  DGP_ERR_PASS_CLEANUP(tcly_doc_startSection(	(TCLYDocCookie) state->tdoc, 
													sbPtr, 
													state->pgNum,
													state->verbose));

	  /* If the previous section ended with a dictionary clear, lossless tokenizer has */
	  /* some mop-up to do. */
	  if(llSb && state->llTdoc->llTokState == NULLLLTOKSTATE){
		  if(llSb->classesFound != DGP_NULLLIST) {
			  DgpLlTokFreeOutput(llSb->classesFound, dgp_state);
		  }		  
		  llSb->classesFound = DGP_NULLLIST;
	  }
	  DGP_ERR_PASS_CLEANUP(DgpTdc_startSection((DgpTdc_Cookie)state->llTdoc, llSb, state->llTdoc->llTokenizerOptions->verbose));

	  /* set the flag and reset the page count */
	  state->status = DGPAPI_IN_SECTION;
	  state->OutputPageCount = 0;
  }

  /* add the new page */
  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;

}
/* monochrome image */
ErrCode
DgpApi_addMonoBuffer (void* handle, 
                      Dgp_UInt8 *buf, Dgp_Int32 bufWidth, Dgp_Int32 bufHeight, 
                      Dgp_Int32 bpl, Dgp_Int32 pageHeight, 
                      Dgp_Int32 xRes, Dgp_Int32 yRes,
                      Dgp_Bool firstBuf, Dgp_Bool lastBuf, Dgp_Bool *overflow)
{
    DgpApi_State *state = (DgpApi_State*) handle;
    Dgp_State *dgp_state = state->tout->dgp_state;
    DGP_STATE_PROLOGUE ("DgpApi_addMonoPage", dgp_state);
    
    DGP_ASSERT (state->handle_check == (size_t) state);
    /* Check the current dictionary memory to see if it is close to  */
    /* the upper limit. if it is, overflow is set to true; this value is used so that */
    /* it will exit from add page loop and go to end section */
    DgpApi_checkTokMemoryBond(handle, overflow);
    
    if(*overflow != DGP_TRUE){
        if(firstBuf){
           DGP_ERR_PASS_CLEANUP( tcly_startPage(     
                                (TCLYDocCookie) state->tdoc, 
                                (DgpTdc_Cookie) state->llTdoc,
                                state->pgNum,
                                bufWidth,
                                pageHeight,
                                xRes,
                                yRes));
        }
        DGP_ERR_PASS_CLEANUP(tcly_encBufferInPage(          
                                (TCLYDocCookie) state->tdoc, 
                                buf, 
                                bufWidth,
                                bufHeight,
                                bpl,
                                32,
                                lastBuf,
                                DGP_FALSE));
        if(lastBuf){
            DGP_ERR_PASS_CLEANUP(tcly_endPage(         
                                (TCLYDocCookie) state->tdoc, 
                                (DgpTdc_Cookie) state->llTdoc,
                                DGP_TRUE,
                                DGP_TRUE));
            *overflow = state->overflow = state->tdoc->curBlock->overflow;   
        }
        
    }
    NORMAL_RETURN;
    
ERR_CLEANUP:
    
    ERR_RETURN;
}


ErrCode
DgpApi_addMonoPage (void* handle, DgpApi_PageData *data, Dgp_Bool * overflow)
{

    DgpApi_State *state = (DgpApi_State*) handle;
    Dgp_State *dgp_state = state->tout->dgp_state;
    DGP_STATE_PROLOGUE ("DgpApi_addMonoPage", dgp_state);
    
    DGP_ASSERT (state->handle_check == (size_t) state);


    /* Check the current dictionary memory to see if it is close to  */
    /* the upper limit. if it is, overflow is set to true; this value is used so that */
    /* it will exit from add page loop and go to end section */
    DgpApi_checkTokMemoryBond(handle, overflow);
    
    if(*overflow != DGP_TRUE){
        /* add the new page */
        DGP_ERR_PASS_CLEANUP(tcly_doc_addMonoPage(	data->image,
            (TCLYDocCookie) state->tdoc, 
            (DgpTdc_Cookie) state->llTdoc,
            data->width, 
            data->height,
            data->bpl, 
            data->xRes, 
            data->yRes, 
            state->pgNum, /* page number */
            DGP_TRUE,
							DGP_FALSE)); 
        state->pgNum++;
        *overflow = state->overflow = state->tdoc->curBlock->overflow;   
    }


    NORMAL_RETURN;
    
ERR_CLEANUP:
    
    ERR_RETURN;
}

#ifndef DGP_FAXMODE

ErrCode
DgpApi_addColorTokens(void* handle, sc_cc *text_ccs, Dgp_Int32 numccs,
                       DgpApi_PageData *data, Dgp_Bool *overflow, DgpAdt_List fgImages)
{
    
    DgpApi_State *state = (DgpApi_State*) handle;
    Dgp_State *dgp_state = state->tout->dgp_state;
    DGP_STATE_PROLOGUE ("DgpApi_addColorTokens", dgp_state);

	/* disable deskewing for color images */
	state->tdoc->deskew_cookie = NULL;
    
	DGP_ERR_PASS(DgpApi_addColorTokensSumGray (handle, text_ccs, numccs, data, NULL, overflow,
        fgImages));
    NORMAL_RETURN;

}

ErrCode
DgpApi_addColorTokensSumGray (void* handle, sc_cc *text_ccs, Dgp_Int32 numccs,
                       DgpApi_PageData *data, DgpImage_Gray grayImage, Dgp_Bool *overflow,
                       DgpAdt_List fgImages)
{
    
    DgpApi_State *state = (DgpApi_State*) handle;
    Dgp_State *dgp_state = state->tout->dgp_state;
    DgpAdt_List tokList = NULL;
    Dgp_Int32 fgScale = 8, fgQual = 70;
    DGP_STATE_PROLOGUE ("DgpApi_addColorTokensSumGray", dgp_state);
    
    DGP_ASSERT (state->handle_check == (size_t) state);
    
    DgpApi_checkTokMemoryBond(handle, overflow);
    
    if(*overflow != DGP_TRUE){
        /* add the new page */
        DGP_ERR_PASS_CLEANUP(tcly_doc_addColorTokens(	
            text_ccs, 
            numccs,
            (TCLYDocCookie) state->tdoc, 
            (DgpTdc_Cookie)state->llTdoc,
            data->width, 
            data->height,
            data->bpl, 
            data->xRes, 
            data->yRes, 
            state->pgNum, /* page number */
            fgQual,
            fgScale,
            DGP_FALSE,
            fgImages));    
        
        state->pgNum++;
        state->overflow = *overflow = state->tdoc->curBlock->overflow;
        if(grayImage) {
            DGP_ERR_PASS_CLEANUP(tokPartial(state->tdoc->tokstate, &tokList));
            DGP_ERR_PASS_CLEANUP(sumGrayInstances( state->tout->dgp_state, 
                grayImage, 
                tokList, 
                state->tdoc->curPage.boxes,
                state->tdoc->curPage.pgNum));
            DGP_ERR_PASS_CLEANUP(tokResumeKeepMember(state->tdoc->tokstate, tokList));
        }
    }


    NORMAL_RETURN;
    
ERR_CLEANUP:
    
    ERR_RETURN;
}
#endif

ErrCode DgpApi_checkTokMemoryBond(void *handle, Dgp_Bool *overflow)
{
    DgpApi_State *state = (DgpApi_State*) handle;
    DGP_PROLOGUE ("DgpApi_checkTokMemoryBond", NULL);
    Dgp_Int32 tokMem;
    
    
    DGP_ASSERT (state->handle_check == (size_t) state);
    DGP_CHECK_NULL(state->tdoc);
    DGP_CHECK_NULL(state->llTdoc);
    DGP_CHECK_NULL(state->tdoc->curBlock);
    DGP_CHECK_NULL(state->llTdoc->curBlock);
    
    state->overflow = DGP_FALSE;
    *overflow = DGP_FALSE;
    
    if(state->tdoc->memBond > 0){
        tokMem =  state->tdoc->curBlock->tokMem + state->llTdoc->curBlock->tokMem;
        if(tokMem >= state->tdoc->memBond){
            state->overflow = DGP_TRUE;
            *overflow = DGP_TRUE;
            
        }else if(tokMem >= state->tdoc->memBond * 0.9){ /* should use tclydoc_getTokMem(); */
            /*its close to overflow but not yet overflow. Since we didn't add it to the list */
            /* we don't need to rollback this page either.             */
            *overflow = DGP_TRUE;
            state->overflow = DGP_FALSE;
        }
    }
    NORMAL_RETURN;
    
}


ErrCode
DgpApi_endSectionGetDict (void* handle, unsigned char dictopt, DgpApi_Jbig2Data *data)
{
    DgpApi_State *state = (DgpApi_State*) handle;
    Dgp_Bool final = DGP_TRUE;
    DGP_PROLOGUE ("DgpApi_endSectionGetDict", NULL);
    DGP_ERR_PASS(DgpApi_endSection((void*) handle, dictopt));
    if (dictopt == DGP_CLEAR_DICT){
        final = DGP_TRUE;
    } else if(dictopt == DGP_INCR_DICT &&  state->tdoc->memBond <0){
        final = DGP_FALSE;
    }else if (dictopt == DGP_INCR_DICT &&  state->tdoc->memBond >0){
        final = DGP_TRUE;
    }
    DGP_ERR_PASS(DgpApi_getDict((void*) handle, final, data));
    NORMAL_RETURN;
}

ErrCode
DgpApi_endSection(void* handle, unsigned char dictopt)
{
    DgpApi_State *state = (DgpApi_State*) handle;
    TCLYSectionBlockPtr sb= NULL;
    DGP_PROLOGUE ("DgpApi_endSectionGet", NULL);
    Dgp_Bool final=DGP_TRUE;
    DgpTdc_SectionBlock *llSb = NULL;
    DGP_ASSERT (state->handle_check == (size_t) state);
    
    /* are we clearing the tokenizer */
    if (dictopt == DGP_CLEAR_DICT){
        final = DGP_TRUE;
        state->tout->final = DGP_TRUE;
    } else if(dictopt == DGP_INCR_DICT &&  state->tdoc->memBond <0){
        final = DGP_FALSE;
        state->tout->final = DGP_FALSE;
    }else if (dictopt == DGP_INCR_DICT &&  state->tdoc->memBond >0){
        /* if membond is set. The we just don't keep dictionary after a incremetal */
        /* section, since this will cause complication in rollback imediately after */
        /* a inc section. We treat incremental same as clear */
        final = DGP_TRUE;
        state->tout->final = DGP_TRUE;
    }
    
    /* Lossless part too... */
    llSb = state->llTdoc->curBlock;
    if(state->overflow == DGP_TRUE){
        /* if this section need to roll back one page due to mermory overflow */
        DGP_ERR_PASS_CLEANUP(tcly_doc_endSectionRollBack(	
                                                        (TCLYDocCookie) state->tdoc,
                                                        &sb,
                                                        state->verbose));
        DGP_ERR_PASS_CLEANUP(DgpTdc_endSectionRollBack( 
                                                        (DgpTdc_Cookie)state->llTdoc, 
                                                        &llSb, 
                                                        state->llTdoc->llTokenizerOptions->verbose));
        state->overflow = DGP_FALSE;
        state->pgNum--;
    }else{    /* end the section */
        DGP_ERR_PASS_CLEANUP(tcly_doc_endSection(	
                                                    (TCLYDocCookie) state->tdoc,
                                                    &sb, 
                                                    final,
                                                    state->verbose));
        DGP_ERR_PASS_CLEANUP(DgpTdc_endSection(
                                                (DgpTdc_Cookie)state->llTdoc, 
                                                &llSb, 
                                                final, 
                                                state->llTdoc->llTokenizerOptions->verbose));
    }
    /* change status flag */
    state->status = DGPAPI_NO_SECTION;


    NORMAL_RETURN;
    
ERR_CLEANUP:
    ERR_RETURN;
}


ErrCode
DgpApi_getDict(void* handle, Dgp_Bool final, DgpApi_Jbig2Data *data){
    DgpApi_State *state = (DgpApi_State*) handle;
    DGP_PROLOGUE ("DgpApi_getDict", NULL);
    TCLYSectionBlock *sb = NULL;
    DgpTdc_SectionBlock *llSb = NULL;
    token *toks;
    DgpAdt_ListNode retnode;
    Jb2State *jb2state;
    
    DGP_ASSERT (state->handle_check == (size_t) state);

    sb = state->tdoc->curBlock; 
    state->tout->curPage = sb->pages;
    llSb = state->llTdoc->curBlock;

    /* create dictionary and return list of tokens */	
    DGP_ERR_PASS_CLEANUP(writeTokOutput( 	state->tout, 
        sb, 
        llSb,
        (DeskewStruct *)state->tdoc->deskew_cookie,
        &toks,
        final));
    
    /* copy the dictionary to the output buffer */
    jb2state = (Jb2State*) state->tout->outputState;
    DGP_ERR_PASS_CLEANUP(getJb2CompressedData (jb2state, (Dgp_UInt8 **)&data->data, &data->data_size));
    
    /* if list of tokens is not empty, add list to TCLYDocState */
    /* Along with the token array, store the number of lossless tokens in every other node */
    /* so we can tailor the cleanup to the token type. */
    if ( toks != (token *)NULL) {
        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(state->tdoc->tokDgpAdt_List, toks, &retnode));
        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(state->tdoc->tokDgpAdt_List, (void *)sb->llNonSingletonCount, &retnode));
    }
    
    /**********
    
      Prepare for creating the position blocks
      
    **********/
    
    /*
    * First, run through the box lists, adjusting the coordinates
    * and tagging each box with its token.
    */
    adjustCoords(sb->classesFound);
    /* Presently this call doesn't do anything, because the classesFound list has already */
    /* been destroyed in the call to createNewTokens(). */
    /*adjustCoords(llSb->classesFound); */
    
    /* This is a list of token structs for tokens that are deskewed.
    * Because a token is only deskewed on a per page basis, we
    * (in deskew.c) create a new token struct that holds the 
    * new bitmap, etc. Evenually, though, we need all the tokens
    * in the residual dictionary to be in a contiguous array.
    * So the bitmap/data is copied (again) from the new token
    * to one that is part of such an array. Then the new token
    * needs to be thrown out and so it goes here. */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&state->skewedTokens, state->tout->dgp_state));
    
    /* This is a list of cc's who bitmaps need to be freed 
    * after each boxcc's page residual dictinonary has been written.
    * It currently only contains cc that belong to singletons
    * that are deskewed. Because singletons are actually part
    * of a class, we encounter them in a (seemingly) random
    * order. So we can't really deskew on a per-page basis.
    * So we keep track of things here.
    */     
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&state->newBitmaps, state->tout->dgp_state));
    
    /* Deskew all the tokens if desired. */
    if(state->tdoc->deskew_cookie) {
#ifdef DGP_FAXMODE
    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else
        DGP_ERR_PASS_CLEANUPF0(deskewTokens(	(DeskewStruct *)state->tdoc->deskew_cookie,
            state->newBitmaps,
            sb), 
            "error while deskewing pages.");
#endif
    }
    NORMAL_RETURN;
    
ERR_CLEANUP:
    ERR_RETURN;
}

ErrCode
DgpApi_redTokNoise(void* handle){
    DgpApi_State *state = (DgpApi_State*) handle;
    Dgp_State *dgp_state = state->tout->dgp_state;
    DGP_PROLOGUE ("DgpApi_redTokNoise", NULL);
    TCLYSectionBlock *sb = NULL;
    sb = state->tdoc->curBlock; 
    DGP_ERR_PASS(smoothToken(dgp_state, sb->classesFound));
    NORMAL_RETURN;

}

ErrCode
DgpApi_cleanSection (void* handle){
    DgpApi_State *state = (DgpApi_State*) handle;
    DGP_PROLOGUE ("DgpApi_cleanSection", NULL);
    TCLYDocCookie tdoc = (TCLYDocCookie) state->tdoc;

    DGP_ASSERT(handle);
    DGP_ERR_PASS(tcly_doc_cleanSection(tdoc, state->verbose));

  NORMAL_RETURN;
}

ErrCode
DgpApi_getNextPosBlock (void* handle, DgpApi_Jbig2Data *data)
{
  DgpApi_State *state = (DgpApi_State*) handle;
  Dgp_State *dgp_state = state->tout->dgp_state;
  DGP_STATE_PROLOGUE ("DgpApi_getNextPosBlock", dgp_state);
  Jb2State *jb2state;
  PosExtension posExt;
  void* extdata = NULL;

  DGP_ASSERT (state->handle_check == (size_t) state);

  if (state->tout->curPage->color)
  {
#ifdef DGP_FAXMODE
    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else
    /* set up extension data structure */
    posExt.extlen = 0;
    posExt.extdata = &extdata;
    posExt.builder = (PosExtBuilder)encodeJbig2Colors;
#endif
  }

  /* create position block */
  DGP_ERR_PASS_CLEANUP(api_createPosBlock (state, &posExt));

  /* created one more output page */
  state->OutputPageCount++;

  /* write it to output buffer */
  jb2state = (Jb2State*) state->tout->outputState;
  DGP_ERR_PASS_CLEANUP(getJb2CompressedData (jb2state, (Dgp_UInt8 **)&data->data,
					     &data->data_size));


  /* create extensions */
  if (state->tout->curPage->color)
  {
    data->extData_size = posExt.extlen;
    DGP_CHECK_ALLOC (data->data = (Dgp_UInt8*) state->tout->dgp_state->a.realloc (data->data, data->data_size + posExt.extlen));
    memcpy(data->data + data->data_size, *(posExt.extdata), posExt.extlen);
    DGP_FREE (*(posExt.extdata));
  }
  else
  {
    data->extData_size = 0;
  }

  /* do some clean up after we put out all pages */
  if (state->OutputPageCount == state->tdoc->curBlock->numPages)
  {
    
  /* There should be nothing left here, it all got freed by
    * pageCleanup. */
    DGP_ASSERT( DGPADT_LILEN( state->skewedTokens) == 0);
    DgpAdt_liFree( state->skewedTokens);
    DGP_ASSERT( DGPADT_LILEN( state->newBitmaps) == 0);
    DgpAdt_liFree( state->newBitmaps);
    
    
    /* do some cleanup */
    if (state->tdoc->curBlock->final){
      emptyOutputState(state->tout);
    }
    
    if (state->tout->final)
    {
      state->tout->callbacks->freeOutputDictblock (state->tout->dictState, state->tout->outputState, NULL);
    } 
  }

  /* set the current page pointer to the next page */
  state->tout->curPage++;

	NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

static int
sortByY (const void* p1, const void* p2)
{
	const tokpos *tokp1, *tokp2;

	tokp1 = (const tokpos*) p1;
	tokp2 = (const tokpos*) p2;

	if (tokp1->ypos > tokp2->ypos) return 1;
	if (tokp1->ypos < tokp2->ypos) return -1;
	return 0;
}

static ErrCode
api_createPosBlock (DgpApi_State *state, PosExtension *posExt)
{
  Dgp_State *dgp_state = state->tout->dgp_state;
  DGP_PROLOGUE("DgpApi_createPosBlock()", DGP_PERRSTATE);
  TCLYOutCookie toc = (TCLYOutCookie) state->tout;
  TCLYSectionBlock *sb = state->tdoc->curBlock;
  DgpTdc_SectionBlock *llSb = state->llTdoc->curBlock;
  DeskewStruct *deskew_cookie = (DeskewStruct *) state->tdoc->deskew_cookie;
  Dgp_Bool g4_toks = toc->g4_toks;
  Dgp_Int32 i = state->OutputPageCount, j, ntok = 0, nLlTok = 0;
  token *page_toks = (token *)NULL;
  tokpos *pos = (tokpos *)NULL;
  Dgp_UInt32 numPos, starty, ntok_sec1;
  void *oldDictstate = (void *)NULL;
  OutputCallbacks *ocb = toc->callbacks;	
  DgpImage_Packed imageCC;
  Dgp_Bool isColor = DGP_FALSE;

  if (state->tout->curPage->color) isColor = DGP_TRUE;

  /* Hold on to the current state of the dictionary so that
   * after we write the per-page dictionary and the page
   * position block, we can revert the dictionary state
   * to the global state. */
  oldDictstate = toc->dictState;

  /* Deskew the outliers on this page. */
  if(deskew_cookie) {
#ifdef DGP_FAXMODE
    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else
	  DGP_ERR_PASS_CLEANUPF0(deskewOutliers(deskew_cookie, 
											sb->pages[i]),
		"error while deskewing outliers.");
#endif
  }
  /* Residual dictionary size has already been measured (when we wrote the section dictionary). */
  ntok = sb->pages[i].residualTokenCount;
  nLlTok = llSb->pages[i].numToks;
  /* llSb->pages[i].toks has already been allocated in makeTokens to account for  */
  /* both lossless and lossy per page tokens */
  page_toks = llSb->pages[i].toks;

  /* start with Page Info segment */
  ocb->userData = &isColor;
  DGP_ERR_PASS_CLEANUP (writeJb2PageInfo(
				 		 sb->pages[i].width,
  						 sb->pages[i].height,
  						 sb->pages[i].xRes,
  						 sb->pages[i].yRes,
						 toc->outputState,
						 ocb->userData));
  ocb->userData = NULL;

  if(ntok + nLlTok > 0) {
	  /* Space for per-page tokens was allocated in call to CreateTokensNew(), when we generated */
	  /* the lossless bitmaps. */
	  DGP_ERR_PASS_CLEANUPF0(writeResDict(	toc,
											sb->pages[i],	
											llSb->pages[i],
											state->skewedTokens, 
											ntok,
											nLlTok,
											page_toks,
											g4_toks),
		"error while writing residual dictionary.");
  }

  /* Clean up the per page bitmaps. This include any bitmaps for
   * outliers on this page, as well as deskewed token classes. */
  /* And lossless tokens too... */
  pageCleanup( 	sb->pages[i].outliers, 
				sb->pages[i].numOutliers, 
				state->skewedTokens, 
				state->newBitmaps, 
				sb->pages[i].pgNum, 
				llSb->pages[i].toks,
				llSb->pages[i].numToks,
				toc->dgp_state);
  /* Now deskew the token positions. */
if( deskew_cookie) {
#ifdef DGP_FAXMODE
    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else
	  DGP_ERR_PASS_CLEANUPF0(deskewPositions( 	deskew_cookie, 
												sb->pages[i]),
		"error while deskewing token positions.");
#endif
  }

  /* Finally, create the position block for this page. */
  makePosArray(	DGPADT_LIFIRST(sb->pages[i].boxes), 
				&llSb->pages[i],
				&pos, 
				sb->pages[i].numBoxes, 
				&numPos, 
				sb->pages[i].width, 
				sb->pages[i].height,
				toc->dgp_state);

  /* sort the positions so we can divide them into the 2 sections */
  ntok_sec1 = 0;
  if (pos) {
      qsort ((void*) pos, numPos, sizeof (tokpos), sortByY);
      while (pos[ntok_sec1].ypos <= sb->pages[i].sectionHeight && ntok_sec1 < numPos) 
	  ntok_sec1++;
  }

  /* Write the position block to the file. */
  /* This part is moved into writeOutputPageDictionary callback
  * to give the user flexibily of creating page segment
  * DGP_ERR_PASS_CLEANUPF0(ocb-> newOutputPage(	toc->outputState, 
  *							DGP_FALSE,
  *							ocb->userData),
  *			   "error creating new page.");
  */

  if (isColor){
#ifdef DGP_FAXMODE
    DGP_ERR_CLEANUPF0(cErrInternal, "Invalid image compression type");
#else      /* link boxes extdata with color_tags */
      DGP_ERR_PASS_CLEANUP(initColors((color_cookie **)(&(posExt->cookie)), 
		  DGP_FALSE, toc->dgp_state));

      if (numPos > 0) DGP_ERR_PASS_CLEANUPF0(ocb->writeOutputPosBlock(	pos, 
	  	numPos, 
		sb->pages[i].width,
  		sb->pages[i].height,
  		sb->pages[i].xRes,
  		sb->pages[i].yRes,
  		posExt, 
  		0, 
  		toc->dictState, 
  		DGP_TRUE,
  		toc->outputState,
  		NULL),
  		"error writing output (color) position block.");

	  cleanupColors(posExt->cookie);
#endif
    }
    else{
		starty = 0;
      DGP_ERR_PASS_CLEANUPF0(ocb->writeOutputPosBlock(	pos, 
	  	ntok_sec1, 
		sb->pages[i].width,
  		sb->pages[i].sectionHeight,
  		sb->pages[i].xRes,
  		sb->pages[i].yRes,
  		NULL, 
  		sb->pages[i].pgNum + 1, /* used to get correct page number */ 
  		toc->dictState, 
  		DGP_FALSE, /* used to indicate 1st section */
  		toc->outputState,
  		&starty),
  		"error writing output position block.");

      /* only put halftones to generic region if we converted them to gray images. */
      /* otherwise, we have already put them into per page dictionary */
      if (state->tdoc->grayHalftone && sb->pages[i].halftones && sb->pages[i].halftones->nelt > 0) {
          DGP_ERR_PASS_CLEANUP(
              encodeHalftoneJBig2 (sb->pages[i].imageinfo,
              sb->pages[i].halftones, 
              ((Jb2State *)toc->outputState)->jb2f, 
              1, /* section */
              sb->pages[i].pgNum + 1,
              toc->dgp_state));	
       }
     /*     } else {
              DGP_ERR_PASS_CLEANUPF0(
                  ocb->writeOutputBinaryPixr(	sb->pages[i].imageinfo,
                  SELECTOR_PLANE, 1 // section , 
                  sb->pages[i].halftones, 
                  toc->outputState,
                  ocb->userData),
                  "error writing halftones.");
          }
      }*/
	
    if(sb->pages[i].largeCCList != DGP_NULLLIST){
		DgpAdt_ListNode ln;
        /* encode large CC as generic regions */
        ln = DGPADT_LIFIRST(sb->pages[i].largeCCList);
	    while(ln) {
            Dgp_Int32 ccX, ccY;
		    dgplssy_cc *newcc = (dgplssy_cc *) ln->userdata;
            imageCC = NULL;
			if (newcc && newcc->y < sb->pages[i].sectionHeight)
			{
				if (newcc->runs)
		            convertCC((void**)&newcc, &imageCC, &ccX, &ccY, dgp_state);
				else
				{
					DgpImage_wrap (dgp_state, newcc->bitmap.bits, newcc->bitmap.width, 
					newcc->bitmap.height, newcc->bitmap.longsperline, &imageCC);
					ccX = newcc->x;
					ccY = newcc->y - newcc->bitmap.height + 1;
				}

	            DGP_ERR_PASS_CLEANUPF0(
		            writeJb2BinaryCC(	
                                    imageCC,
						            SELECTOR_PLANE,
                                    ccX,
                                    ccY,
						            toc->outputState,
						            ocb->userData),
						            "error write large CC");
			    if(imageCC)  DgpImage_free(imageCC);
				imageCC = NULL;
				DgpAdt_liRemAndNext(ln, &ln);
			}
			else
			{
				ln = ln->next;
			}
		}
		if(sb->pages[i].largeCCList->nelt == 0) 
		{
			DgpAdt_liFree (sb->pages[i].largeCCList);
			sb->pages[i].largeCCList = DGP_NULLLIST;
		}
    } 


	/* write the end of stripe segment */
	DGP_ERR_PASS_CLEANUP(writeJb2EOStripe(sb->pages[i].sectionHeight,
			sb->pages[i].pgNum + 1, toc->outputState));
		

	/* write the second stripe */
      for (j = ntok_sec1; j < (Dgp_Int32)numPos; j++)
		  pos[j].ypos -=  (Dgp_Int16) sb->pages[i].sectionHeight;

	    starty = sb->pages[i].sectionHeight;
		DGP_ERR_PASS_CLEANUPF0(ocb->writeOutputPosBlock( &pos[ntok_sec1], 
	  	numPos - ntok_sec1, 
		sb->pages[i].width,
  		sb->pages[i].height - sb->pages[i].sectionHeight,
  		sb->pages[i].xRes,
  		sb->pages[i].yRes,
  		NULL, 
  		sb->pages[i].pgNum + 1, /* used to get correct page number */ 
  		toc->dictState, 
  		DGP_TRUE, /* used to indicate second section */
  		toc->outputState,
  		&starty),
  		"error writing output position block.");

        /* only put halftones to generic region if we converted them to gray images. */
        /* otherwise, we have already put them into per page dictionary */
        if (state->tdoc->grayHalftone && sb->pages[i].halftones && sb->pages[i].halftones->nelt > 0) {
            DGP_ERR_PASS_CLEANUP(
                encodeHalftoneJBig2 ( sb->pages[i].imageinfo,
                sb->pages[i].halftones, 
                ((Jb2State *)toc->outputState)->jb2f, 
                2, /*section,  */
                sb->pages[i].pgNum + 1,
                toc->dgp_state));		
        } 
        /*
            else {
                DGP_ERR_PASS_CLEANUPF0(
                    ocb->writeOutputBinaryPixr(	sb->pages[i].imageinfo,
                    SELECTOR_PLANE, 2 // section,
                    sb->pages[i].halftones, 
                    toc->outputState,
                    ocb->userData),
                    "error writing halftones.");
            }
            DgpAdt_liFree (sb->pages[i].halftones);
            sb->pages[i].halftones = DGP_NULLLIST;
        }*/
        
    if(sb->pages[i].largeCCList != DGP_NULLLIST){
		DgpAdt_ListNode ln;
        /* encode large CC as generic regions */
        ln = DGPADT_LIFIRST(sb->pages[i].largeCCList);
	    while(ln) {
            Dgp_Int32 ccX, ccY;
		    dgplssy_cc *newcc = (dgplssy_cc *) ln->userdata;
            imageCC = NULL;
			if (newcc && newcc->y >= sb->pages[i].sectionHeight)
			{
				if (newcc->runs)
		            convertCC((void**)&newcc, &imageCC, &ccX, &ccY, dgp_state);
				else
				{
					DgpImage_wrap (dgp_state, newcc->bitmap.bits, newcc->bitmap.width, 
					newcc->bitmap.height, newcc->bitmap.longsperline, &imageCC);
					ccX = newcc->x;
					ccY = newcc->y - newcc->bitmap.height + 1;
				}

                DGP_ERR_PASS_CLEANUPF0(
                    writeJb2BinaryCC(	
                                    imageCC,
						            SELECTOR_PLANE,
                                    ccX,
                                    ccY,
						            toc->outputState,
						            ocb->userData),
						            "error write large CC");
                if(imageCC)  DgpImage_free(imageCC);
                imageCC = NULL;
                DgpAdt_liRemAndNext(ln, &ln);
            }
			else
			{
				ln = ln->next;
			}

    	}
        DgpAdt_liFree (sb->pages[i].largeCCList);
	    sb->pages[i].largeCCList = DGP_NULLLIST;
    } 
	
	}

   /*
	* Since the new dictState from writeResDict() is specific to this 
	* page only, we return the dictblock to its state before we wrote
	* this page's posBlock.
	*/
    if((oldDictstate != NULL) && (oldDictstate != toc->dictState)) {
		ocb->freeOutputDictblock(	toc->dictState, 
			toc->outputState,
			ocb->userData);
		toc->dictState = oldDictstate;
    }
	
    /* All done with the position array */
    if (pos) DGP_FREE((void *)pos);
    llSb->pages[i].boxes = (tokpos *)NULL;
	
   /* Same goes for these tokens. They belong to outliers and
	* deskewed tokens, and so were only referenced by this page.
	* i.e. once the per-page dictionary is tossed and the position
	* block is written out, these can be tossed as well. 
	*/
    if (page_toks) DGP_FREE((void *)page_toks);
    llSb->pages[i].toks = (token *)NULL;
		
    /* finish this page */
    DGP_ERR_PASS_CLEANUPF0(ocb->finishOutputPage( 	toc->outputState,
		ocb->userData),
		"error closing page.");

	NORMAL_RETURN;

ERR_CLEANUP:

	ERR_RETURN;	
}
  
/* copied from libtclydoc/compress.c */
/* free any bitmaps created for the outliers and new deskewed tokens */
static void
pageCleanup( 	DgpAdt_List outliers, 
				Dgp_Int32 numOutliers, 
				DgpAdt_List skewedTokens, 
				DgpAdt_List newBitmaps, 
				Dgp_Int32 currentPage, 
				token *llToks,
				Dgp_Int32 numLlToks,
				Dgp_State *dgp_state)
{
  DGP_VOID_PROLOGUE("pageCleanup()", DGP_PERRSTATE);
  DgpAdt_ListNode ln;
  dgplssy_cc *boxcc;
  token *tok;
  Dgp_Int32 i;
  Dgp_Allocators *a = &(dgp_state->a);
    
  /* Throw out any bitmaps of outliers on this page. */
  if(outliers != DGP_NULLLIST){
      ln = DGPADT_LIFIRST(outliers);
      for( i = 0; i < numOutliers; i++, ln = DGPADT_LINEXT( ln)) {
        boxcc = (dgplssy_cc *)DGPADT_LIGET( ln);
        DGP_VOID_ASSERT( boxcc != (dgplssy_cc *)NULL);
      
           
        /* All outliers on this page have had their runs 
         * converted to bitmaps by this time. So just free 
         * the bitmaps. */
        (a->free)( (void *)boxcc->bitmap.bits);
      }  
  }
  
  /* Throw out all tokens in "skewedTokens" as they all are dead symbols-
   * these tokens are not referenced from the symbol dictionaries, nor 
   * are there any remaining cc's that point to them. */
  if(skewedTokens != DGP_NULLLIST){
     ln = DGPADT_LIFIRST(skewedTokens);
      while( ln != DGP_NULLLISTNODE) {
        tok = (token *)DGPADT_LIGET( ln);
        DGP_VOID_ASSERT( tok != (token *)NULL);
        DGP_VOID_ASSERT( tok->committed == DGP_FALSE);

        /* This token can be freed since of the relevent data
         * has been copied to another token struct. Also no
         * one points to it anymore. (Everyone points to the copy.) 
         * The dictionary has been written and we own the bitmap, 
         * so that can be tossed as well. */

        (a->free)( (void *)tok->extdata);	   
        (a->free)( (void *)tok->bits);
        (a->free)( (void *)tok);
        DgpAdt_liRemAndNext( ln, &ln); 

      }
  }
  if(newBitmaps != DGP_NULLLIST){
      ln = DGPADT_LIFIRST(newBitmaps);
      while( ln != DGP_NULLLISTNODE) {
        boxcc = (dgplssy_cc *)DGPADT_LIGET( ln);
      
        DGP_VOID_ASSERT( boxcc != (dgplssy_cc *)NULL);
        if( boxcc->pgnum == currentPage){
		    (a->free)( (void *)boxcc->bitmap.bits);  
		    DgpAdt_liRemAndNext( ln, &ln); 
	    }
        else{
		    ln = DGPADT_LINEXT( ln);
	    }
      }
  }
  /* Get rid of lossless this-page-only token bitmaps too... */
  for(i = 0; i < numLlToks; i++){
	  DGP_FREE(llToks[i].bits);
  }

  return;
}

