  
#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif

#include "jbig2apiCB.h"
/*#include <tc-tools-exec.h> */
#include <dgp_sys_err_control.h>
/* #include "tc-tiffio.h" */
#include "jtokapi_p.h"
#include "jtokapi.h"
#include <stdio.h>
#include "dgp_err_printers.h"
#include "PsgFix.h"

#if 0
#ifdef DGP_DBUG

/*********************************************
 * function to output debug errors to a file 
 **********************************************/
static void 
dgpErrPrinter(	void *cookie, 
		const char *file_name, 
		const char *funct_name, 
		Dgp_Int32 line_number, 
		ErrCode error, 
		const char *format, ...)
{
  va_list(arglist);
  FILE *outfile;

  if(cookie) {
    va_start(arglist, format);
    if(error == DGP_REPORT) {
      outfile = ((statusOutputFiles *)cookie)->rptfile;
    } else {
      outfile = ((statusOutputFiles *)cookie)->errfile;
    }

    /* we're done if the outfile is NULL */
    if(outfile == NULL) return;

    if(error == DGP_REPORT) {
      fprintf(outfile, "Report -- ");
    } else if(error == DGP_WARNING) {
      fprintf(outfile, "Warning -- ");
    } else {
      fprintf(outfile, "Error %d, \"%s\", line %d, %s -- ", 
			(int)error, 
			file_name, 
			(int)line_number, 
			funct_name);
    }
    if(format != NULL) vfprintf(outfile, format, arglist);
    fprintf(outfile, "\n");
  }
}

#else /* DGP_DBUG is not defined */

/*************************************************************
 * function to output nondebug errors to a file 
 **************************************************************/
static void 
dgpErrPrinter(void *cookie, ErrCode error, const char *format, ...)
{
  va_list(arglist);
  FILE *outfile;

  if(cookie) {
    va_start(arglist, format);
    if(error == DGP_REPORT) {
      outfile = ((statusOutputFiles *)cookie)->rptfile;
    } else {
      outfile = ((statusOutputFiles *)cookie)->errfile;
    }

    /* we're done if the outfile is NULL */
    if(outfile == NULL) return;

    if(error == DGP_REPORT) {
      fprintf(outfile, "Report -- ");
    } else if(error == DGP_WARNING) {
      fprintf(outfile, "Warning -- ");
    } else {
      fprintf(outfile, "Error %d -- ", (int)error);
    }
    if(format != NULL) vfprintf(outfile, format, arglist);
    fprintf(outfile, "\n");
  }
}

#endif

#endif

ErrCode
DgpApi_init (void**					handle, 
			 Dgp_Allocators			*a, 
			 Dgp_SmallAllocators	*sm_a, 
             DgpApi_Options			*opts,
			 tokOptions				*passedTokOptions,
			 LlTokOptions			*passedLlTokOptions,
			 DgpTdc_JpegCompInfo	*passedJpegComp)
{

  Dgp_Int32 numPgs, num_dicts;
  TCLYOutCookie tout = NULL;
  Dgp_State *dgp_state = NULL;
  TCLYDocCookie tdc = NULL;
  DgpTdc_Cookie llTdc = NULL;
  Dgp_Bool report = opts->verbose;
  tokOptions *tokopt = NULL;
  LlTokOptions *llTokOpt = NULL;
  OutputCallbacks *cb = NULL;
  DgpApi_State *api_state = NULL;
  uds junk;
  DgpTdc_JpegCompInfo jpegComp;		/* Includes quality & scale factor. */

#ifndef EQUATOR_CA_ENV
  DGP_MAIN_PROLOGUE("DgpApi_init()", stderr, Dgp_fileErrPrinter);
#else
  Dgp_ErrState *__PERR_STATE__, temp;
  PROLOGUE;
  Dgp_UInt8* __FUNCT_NAME__ = "DgpApi_init()";
  temp.userCookie = stderr;
  temp.errPrinter = Dgp_fileErrPrinter;
  __PERR_STATE__ = &temp;
#endif

  /* Init dgp state */
  DGP_ERR_PASS_CLEANUP(Dgp_initDgpState(&dgp_state, 
                                        a, 
                                        sm_a,
                                        stderr,
                                        Dgp_fileErrPrinter,
                                        DGP_TRUE, 
                                        DGP_TRUE));
  DGP_SET_ERR_STATE(DGP_PERRSTATE);


  /* allocate State struct */
  DGP_CHECK_ALLOC_CLEANUP (api_state = DGP_MALLOC (sizeof (*api_state)));
  api_state->handle_check = (size_t) api_state;
  DGP_CHECK_ALLOC_CLEANUP (cb = DGP_MALLOC (sizeof (OutputCallbacks)));
  DGP_CHECK_ALLOC_CLEANUP (tokopt = DGP_MALLOC (sizeof (tokOptions)));
  DGP_CHECK_ALLOC_CLEANUP (llTokOpt = DGP_MALLOC (sizeof (LlTokOptions)));

  /*
   * Setup callback functions
   */
  cb->createOutputState = (CreateOutputState) createJb2StateAPICB;
  cb->closeOutputState = (CloseOutputState) closeJb2StateAPICB;
  cb->writeOutputPageDictionary = (WriteOutputPageDictionary) writeJb2PageDictionaryCB;
  cb->finishOutputPage = (FinishOutputPage) finishJb2PageCB;
  cb->writeOutputPosBlock = (WriteOutputPosBlock) writeJb2PosBlockCB;
  cb->writeOutputDictionary = (WriteOutputDictionary) writeJb2DictionaryCB;
  cb->writeOutputBinaryPixr = (WriteOutputBinaryPixr) writeJb2BinaryPixrCB;
  cb->freeOutputDictblock = (FreeOutputDictblock) freeJb2DictblockCB;
  cb->userData = NULL;
   
  /* Initialize parameters to default settings if caller hasn't supplied an alternative. */
  if(passedTokOptions == NULL){
	  tokopt->dgp_state			 = dgp_state;
	  tokopt->fracHausd			 = 0.97f;
	  tokopt->maxDense           = 3;
	  tokopt->widthDiff			 = 1;
	  tokopt->heightDiff		 = 1;
	  tokopt->commitThresh       = 10000;
	  tokopt->freezeHistThresh   = 10000;
	  tokopt->makeHistThresh     = 10;
	  tokopt->returnHistThresh   = 0;
	  tokopt->nneigh             = 8;
	  tokopt->xshift             = 3;
	  tokopt->xslopechange       = 6;
	  tokopt->newslope           = 2.0;
	  tokopt->dilationRadius     = 1;
	  tokopt->lowerDilationLimit = 6;
	  tokopt->flags              = TOK_OPT_ALIGN | TOK_OPT_RETURNOFFSET;
	  tokopt->mergeCallback = tokopt->callbackCookie = NULL;
  } else {
	  tokopt->dgp_state			 = dgp_state;
	  tokopt->fracHausd			 = passedTokOptions->fracHausd;
	  tokopt->maxDense           = passedTokOptions->maxDense;
	  tokopt->widthDiff			 = passedTokOptions->widthDiff;
	  tokopt->heightDiff		 = passedTokOptions->heightDiff;
	  tokopt->commitThresh       = passedTokOptions->commitThresh;
	  tokopt->freezeHistThresh   = passedTokOptions->freezeHistThresh;
	  tokopt->makeHistThresh     = passedTokOptions->makeHistThresh;
	  tokopt->returnHistThresh   = passedTokOptions->returnHistThresh;
	  tokopt->nneigh             = passedTokOptions->nneigh;
	  tokopt->xshift             = passedTokOptions->xshift;
	  tokopt->xslopechange       = passedTokOptions->xslopechange;
	  tokopt->newslope           = passedTokOptions->newslope;
	  tokopt->dilationRadius     = passedTokOptions->dilationRadius;
	  tokopt->lowerDilationLimit = passedTokOptions->lowerDilationLimit;
	  tokopt->flags              = TOK_OPT_ALIGN | TOK_OPT_RETURNOFFSET;
	  tokopt->mergeCallback = tokopt->callbackCookie = NULL;
  }
  if(passedLlTokOptions == NULL){
	  /* I don't think we can get the options in a usable state without knowing something about the document, */
	  /* but we'll try... */
	  llTokOpt->dgp_state		 = dgp_state;
	  llTokOpt->flags			 = 0;
	  llTokOpt->verbose			 = opts->verbose;
	  llTokOpt->ascii			 = DGP_FALSE;
	  llTokOpt->color			 = DGP_FALSE;
	  llTokOpt->halftone		 = DGP_FALSE;
	  llTokOpt->lossy			 = DGP_FALSE;
	  llTokOpt->maxSamptokH		 = -1;
	  llTokOpt->maxSamptokW		 = -1;
	  llTokOpt->noascii			 = DGP_TRUE;
	  llTokOpt->numStrips		 = 1;
	  llTokOpt->sizeThresh		 = 1;
	  llTokOpt->washThresh		 = 301;
	  llTokOpt->tokenizerParams	 = NULL;
	  llTokOpt->colorTokenizerParams = NULL;
  } else {
	  llTokOpt->dgp_state		 = passedLlTokOptions->dgp_state;
	  llTokOpt->flags			 = passedLlTokOptions->flags;
	  llTokOpt->verbose			 = passedLlTokOptions->verbose;
	  llTokOpt->ascii			 = passedLlTokOptions->ascii;
	  llTokOpt->color			 = passedLlTokOptions->color;
	  llTokOpt->halftone		 = passedLlTokOptions->halftone;
	  llTokOpt->lossy			 = passedLlTokOptions->lossy;
	  llTokOpt->maxSamptokH		 = passedLlTokOptions->maxSamptokH;
	  llTokOpt->maxSamptokW		 = passedLlTokOptions->maxSamptokW;
	  llTokOpt->noascii			 = passedLlTokOptions->noascii;
	  llTokOpt->numStrips		 = passedLlTokOptions->numStrips;
	  llTokOpt->sizeThresh		 = passedLlTokOptions->sizeThresh;
	  llTokOpt->washThresh		 = passedLlTokOptions->washThresh;
	  llTokOpt->tokenizerParams	 = passedLlTokOptions->tokenizerParams;
	  llTokOpt->colorTokenizerParams = passedLlTokOptions->colorTokenizerParams;
  }
  if(passedJpegComp == NULL){
	  jpegComp.jpQuality		 = 75;
	  jpegComp.scaleFactor		 = 1.0f;
  } else {
	  jpegComp.jpQuality		 = passedJpegComp->jpQuality;
	  jpegComp.scaleFactor		 = passedJpegComp->scaleFactor;
  }


  /* create TCLYDOC cookie */
  DGP_ERR_PASS_CLEANUP(tcly_doc_init(	dgp_state, 
                                        &tdc, 
                                        DGP_TRUE, 
                                        opts->grayHalftone, 
                                        opts->sizeThresh, 
                                        opts->forceRes,
                                        opts->bitmapSizeLimit,
                                        tokopt,
                                        opts->deskew,
                                        opts->skewAccuracy,
                                        opts->baselineOrient,
                                        opts->membond,
                                        report));
  api_state->tdoc = (TCLYDocState*) tdc;

  /* 16.4.2001 / dhtesq. */
  /* Prepare the lossless tokenizer too;  we'll use it for small tokens, and eventually */
  /* for a fully lossless mode. */
  DGP_ERR_PASS_CLEANUP(DgpTdc_initDoc(	dgp_state, 
										&llTdc, 
										llTokOpt->lossy,
										llTokOpt->color,
										llTokOpt->halftone,
										llTokOpt->sizeThresh,
										(Dgp_UInt16)llTokOpt->washThresh,
										DGPTDC_SAMP_COMPTYPE_JPEG, 
										&jpegComp, 
										llTokOpt->tokenizerParams,
										llTokOpt->colorTokenizerParams,
										llTokOpt->verbose));
  ((tdoc_state *)llTdc)->llTokenizerOptions = llTokOpt;
  api_state->llTdoc = (tdoc_state *)llTdc;
  
  /* create TCLYOUT cookie */
  numPgs = 1;
  num_dicts = 2;
  DGP_ERR_PASS_CLEANUP(tcly_out_init(	dgp_state, 
				     &tout, 
				     numPgs, 
				     opts->pdfOut, 
				     &junk,
				     opts->classThresh,
				     opts->writeclasses,
				     opts->g4_toks,
				     opts->jbig2Profile,
				     report,
				     cb));
  api_state->tout = tout;

  /* set status to no Section */
  api_state->status = DGPAPI_NO_SECTION;
  api_state->overflow = DGP_FALSE;
  *handle = (void*) api_state;
  
  /* set page count to zero */
  api_state->OutputPageCount = 0;
  api_state->pgNum = 0;

  api_state->verbose = opts->verbose;

  NORMAL_RETURN;

ERR_CLEANUP:
  if(api_state) DGP_FREE(api_state);
  if(cb) DGP_FREE(cb);
  if(tokopt) DGP_FREE(tokopt);
  if(llTokOpt) DGP_FREE(llTokOpt);

  ERR_RETURN;
}



ErrCode
DgpApi_close (void* handle)
{
  DgpApi_State *state = (DgpApi_State*) handle;
  Dgp_State *dgp_state = state->tout->dgp_state;
  TCLYSectionBlockPtr sbPtr = (TCLYSectionBlockPtr)state->tdoc->curBlock;
  DgpTdc_SectionBlock *llSb = state->llTdoc->curBlock;
  DGP_STATE_PROLOGUE ("DgpApi_close", dgp_state);
  OutputCallbacks *cb;
  DgpTdc_Cookie llTdoc = (DgpTdc_Cookie)(state->llTdoc);

  DGP_ERR_PASS(api_checkHandle (state));

  DGP_FREE(state->llTdoc->llTokenizerOptions);
  state->llTdoc->llTokenizerOptions = NULL;
  DGP_ERR_CHECK_NO_MSG(DgpTdc_endDoc(&llTdoc, llSb)) {
    DGP_WARNF0("Problem cleaning up; continuing anyway.");
    DGP_CLEAR_CUR_ERROR;
  }

  DGP_FREE (state->tdoc->tokOpt);
  state->tdoc->tokOpt = NULL;
  DGP_ERR_CHECK_NO_MSG(tcly_doc_end((TCLYDocCookie) state->tdoc, sbPtr)) {
    DGP_WARNF0("Problem cleaning up; continuing anyway.");
    DGP_CLEAR_CUR_ERROR;
  }

  cb = state->tout->callbacks;
  DGP_ERR_CHECK_NO_MSG(tcly_out_end(state->tout)) {
    DGP_WARNF0("Problem cleaning up; continuing anyway.");
    DGP_CLEAR_CUR_ERROR;
  }
  DGP_FREE (cb);
  cb = NULL;

  DGP_FREE (state);
  state = NULL;

  Dgp_cleanupDgpState(&dgp_state);
	
  NORMAL_RETURN;
}


INLINE ErrCode
api_checkHandle (DgpApi_State* handle)
{
	if (handle->handle_check == (size_t) handle)
		return 0;

	return cErrAssertFailure;
}
