#include <stdio.h>
#include <libpixrw.pub>
#include <libjpegw.pub>
#include <rescale.pub>
#include <libiow.pub>
#include <appsinc.h>
#include <iw_scale.pub> 
#include <fill.pub>
#include <rast.pub>

/* err state */
#include "dgp_state.h"
#include "dgp_err_printers.h"
#include "dgp_sys_err_control.h"

/* relate to color fg */
#include "image_defs.h"
#include "basic_structs.h"
#include "image.h"
#include "blit.h"
#include "blit_rgb.h"
#include "scaleBlit.h"
#include "compress.h"
#include "scancolor.h"
#include "mrc2file.h"

#include "colorfg.h"

#include "PsgFix.h"

#ifndef JC_DEFAULT
#define JC_DEFAULT 0 /*defined in parc IPCORE, but not IPCORE-PSG */
#endif

typedef struct {
  Dgp_State *dgp_state;
  Dgp_Int8 *buf;
  Dgp_UInt32 size;
  Dgp_UInt32 alloced;
} jcookie;

#ifdef PSG /*IPCORE-PSG doesn't use *nwritten, so include a version of this without. */
static ErrCode 
jpeg_Compress_callback(IPWState *ipwstate, 
                              void *cookie, 
                              Dgp_UInt8 *buf, 
                              Dgp_UInt32 nbytes)
{
  jcookie *jc = (jcookie *)cookie;
  DGP_PROLOGUE("jpeg_Compress_callback()", jc->DGP_PERRSTATE);
  Dgp_UInt32 size;
  Dgp_Int8 *tmpptr;

  (void)ipwstate;

  size = jc->size + nbytes;
  if(size > jc->alloced){
    jc->alloced = (size * 3) / 2;
    DGP_CHECK_ALLOC(tmpptr = (Dgp_Int8 *)(jc->DGP_REALLOC)(jc->buf, jc->alloced));
    jc->buf = tmpptr;
  }

  memcpy(jc->buf + jc->size, buf, nbytes);
  jc->size += nbytes;

  NORMAL_RETURN;
}
#endif
#ifndef PSG
static ErrCode 
jpeg_Compress_callback(IPWState *ipwstate, 
                              void *cookie, 
                              Dgp_UInt8 *buf, 
                              Dgp_UInt32 nbytes,
                              Dgp_UInt32 *nwritten)
{
  jcookie *jc = (jcookie *)cookie;
  DGP_PROLOGUE("jpeg_Compress_callback()", jc->DGP_PERRSTATE);
  Dgp_UInt32 size;
  Dgp_Int8 *tmpptr;

  (void)ipwstate;

  size = jc->size + nbytes;
  if(size > jc->alloced){
    jc->alloced = (size * 3) / 2;
    DGP_CHECK_ALLOC(tmpptr = (Dgp_Int8 *)(jc->DGP_REALLOC)(jc->buf, jc->alloced));
    jc->buf = tmpptr;
  }

  memcpy(jc->buf + jc->size, buf, nbytes);
  jc->size += nbytes;

  *nwritten = nbytes;

  NORMAL_RETURN;
}
#endif

ErrCode
getColorFg(
Dgp_State *dgp_state,
DgpAdt_List boxes,
Dgp_Int32 width,
Dgp_Int32 height,
Dgp_Int32 xRes,
Dgp_Int32 yRes,
Dgp_Int32 jpegQual,
Dgp_Int32 scaleFac,
DgpAdt_List fgImages)
{
    DGP_STATE_PROLOGUE ("getColorFg", dgp_state);

    /* variabl related to color fg */
    Dgp_Int32 rhc = 0, rhf = 0, nr = 0, rhfLast = 0, rhcLast = 0;
    DgpImage_ColorRGB cImage = NULL, fgImage = NULL;
    Dgp_Rect region, boxRect;
    Dgp_Int32 cBpl = 0, fgWidth = 0, fgHeight = 0;
    Dgp_Int32 i = 0,j = 0, num = 0;
	dgplssy_cc *box = NULL;
	DgpAdt_ListNode ln = NULL;
    DgpImage_RGB *cImagePtr = NULL, *fgImagePtr = NULL;

    /* Jpeg Compress */
    PIXR *fgPixr = NULL, *maskPixr = NULL, *finalPixr = NULL;
    void *ipstate = (void*) dgp_state->ipstate;
    jcookie jc;
    Dgp_CompressedImage *jpeg;
    Dgp_Int8 *tmpptr;
     
    /* create new fg image */
    fgWidth = width / scaleFac;
    fgHeight = height / scaleFac;

    /* Create a foregournd pixr in forground resolution */
    DGP_ERR_PASS_CLEANUP(w_createPixrExtended(	ipstate,
					fgWidth,
					fgHeight,
					8,
					Int32_TO_ScaledInt1616(xRes/scaleFac),
					Int32_TO_ScaledInt1616(yRes/scaleFac), 
					cFrameBits,
					3,
					cColorRGB,
					NULL,
					&finalPixr));

    /*split a large region into sub sections */
    if(height> 5000) {
        rhc = height / 4;
        rhf = fgHeight / 4;
        rhcLast = height - 3*rhc;
        rhfLast = fgHeight - 3*rhf;
        nr = 4;
    } else if(height> 3000){
        rhc =height / 2;
        rhf = fgHeight / 2;
        rhcLast = height - rhc;
        rhfLast = fgHeight - rhf;
        nr = 2;
    } else{
        rhc = height;
        rhf = fgHeight;
        rhcLast = rhc;
        rhfLast = rhf;
        nr = 1;
    }

    /*draw color blocks onto fg */
    for(num = 0; num< nr; num++){
        Dgp_Int32 rHeight = 0, rfgHeight = 0;
        Dgp_Rect rectIntersect;
        
        rHeight = rhc;
        rfgHeight = rhf;

       /*change region height if it's the last one */
        if(num == nr -1){
            rHeight = rhcLast;
            rfgHeight = rhfLast;
        }

        /* create drawing region rectangle */
        region.t = num * rhc; 
        region.l = 0;
        region.b = region.t + rHeight -1;
        region.r = width -1;

        /*create fg DgpImage  */
        DgpImage_newAligned ((Dgp_State *)dgp_state, DGPIMAGE_RGB, 
                        width, rHeight, (void **)(&cImage));
        cImagePtr = (DgpImage_RGB *)DGPIMG_PIXPTR(cImage,0,0);     
        cBpl =  DGPIMG_GETROWWIDTH((DgpImage_Gray)cImage);

        /*set fgDgpImage to white */
        memset(cImagePtr, 0xff, rHeight * cBpl);

        /* draw color box on to fg DgpImage */
        for (ln = DGPADT_LIFIRST(boxes); ln != DGP_NULLLISTNODE; ) {
		    box = (dgplssy_cc *)DGPADT_LIGET(ln);
		    DGP_ASSERT(box != (dgplssy_cc *)NULL);
		    DGP_ASSERT(box->bitmap.bits != (Dgp_UInt32 *)NULL);

            /* create rectangle for each color box */
            boxRect.l = box->x; 
            boxRect.t = box->y;
            boxRect.r = box->x + box->bitmap.width -1;
            boxRect.b = box->y + box->bitmap.height -1;
            
            /* get intersection of color box and the drawing region */
            rectIntersect = rect_intersection( boxRect, region); 
            if(!rect_is_empty(rectIntersect) ){
                
                 /* draw solid color box  */
                 DgpBlat_blitSetRGB( 
                        box->color,
                        cImagePtr,
                        cBpl,
                        region.l, region.t,  
                        rectIntersect.l , rectIntersect.t,
                        rect_width(rectIntersect), rect_height(rectIntersect));
            }
		    ln = DGPADT_LINEXT(ln);
        }

		/* Create a foregournd pixr in original resolution */
		DGP_ERR_PASS_CLEANUP(w_createPixrExtended(	ipstate,
						width,
						rHeight,
						8,
						Int32_TO_ScaledInt1616(xRes),
						Int32_TO_ScaledInt1616(yRes), 
						cFrameBits,
						3,
						cColorRGB,
						NULL,
						&fgPixr));

		/* Create a mask pixr. */
		DGP_ERR_PASS_CLEANUP(w_createPixr(	ipstate,
						width,
						rHeight,
						1,
						Int32_TO_ScaledInt1616(xRes),
						Int32_TO_ScaledInt1616(yRes), 
						cFrameBits,
						&maskPixr));

		{
			  PIXR	*pixr_r = NULL, *pixr_g = NULL, *pixr_b = NULL;
			  DgpImage_RGB			*src_end;
			  Dgp_UInt8				*dst_r, *dst_g, *dst_b;
			  UInt32* dst_mask, *line_mask;

			  DGP_ERR_PASS_CLEANUP(w_pixrGetChannel(ipstate, fgPixr, 0, &pixr_r));
			  DGP_ERR_PASS_CLEANUP(w_pixrGetChannel(ipstate, fgPixr, 1, &pixr_g));
			  DGP_ERR_PASS_CLEANUP(w_pixrGetChannel(ipstate, fgPixr, 2, &pixr_b));
    
			/* Convert interleaved DgpImage  into non-interleaved pixr. */
			for (i = 0; i < rHeight; i++) {
				fgImagePtr = DGPIMG_PIXPTR(cImage, 0, i);
				src_end = DGPIMG_PIXPTR(cImage, width, i);
				w_getLine(ipstate,  pixr_r, i, &dst_r);
				w_getLine(ipstate,  pixr_g, i, &dst_g);
				w_getLine(ipstate,  pixr_b, i, &dst_b);
				for (j=0; fgImagePtr < src_end; fgImagePtr++, j++, dst_r++, dst_g++, dst_b++) {
				  *dst_r = fgImagePtr->r;
				  *dst_g = fgImagePtr->g;
				  *dst_b = fgImagePtr->b;
				}
			}
			/* Byte swap for little-endian platforms. */
			DGP_ERR_PASS_CLEANUP(w_flip(ipstate, fgPixr));
            
            /*w_writeTiff(ipstate, fgPixr, "c:\\temp\\fg.tif",COMPRESSION_NONE ,cTrue, 0);			 */
            /* create mask */
			for (i = 0; i < rHeight; i++) {
				w_getLine(ipstate, maskPixr,i, (UInt8 **)(&line_mask));
				fgImagePtr = DGPIMG_PIXPTR(cImage, 0, i);
				src_end = DGPIMG_PIXPTR(cImage, width, i);
				for (j=0; fgImagePtr < src_end; fgImagePtr++, j++) {
				  if(!(fgImagePtr->r == 255 && fgImagePtr->g ==255 && fgImagePtr->b == 255)){
					  dst_mask = line_mask + j/32;
					  *dst_mask = *dst_mask | (0x00000001 << (31 - j % 32));
				  }

				}
			}
		}
        if(cImage) DgpImage_free(cImage);

		/*grow forground color box until they merge with each other */
        /* with mask define the orginal fg region */
		w_grayPropagate4Mask (ipstate, maskPixr, cTrue, 255, fgPixr);
		if(maskPixr) w_destroyPixr(ipstate, &maskPixr);
		maskPixr = fgPixr;
		fgPixr = NULL;

		/*scale down, now that we have merged color fg */
		w_scale (ipstate, maskPixr, fgWidth, rfgHeight, 
							cFalse, IP_MED_QUALITY, &fgPixr);
		if(maskPixr) w_destroyPixr(ipstate, &maskPixr);

		/*copy the fgPixr into finalPixr */
		w_rasterop (ipstate, 0, 0, fgPixr,			/*src */
							PIX_SRC,				/*op */
							0, rhf * num, fgWidth, rfgHeight, finalPixr /*dst */
							);
		/*w_writeTiff(ipstate, finalPixr, "c:\\temp\\final.tif",COMPRESSION_NONE ,cTrue, 0); */
		if(fgPixr) w_destroyPixr(ipstate, &fgPixr);

	}

    /* jpeg compress fgground image */
    DGP_CHECK_ALLOC (jpeg = DGP_MALLOC (sizeof(*jpeg)));
    jpeg->xres = xRes / scaleFac; /*ScaledInt1616_TO_Int32(tmp_xres); */
    jpeg->yres = yRes / scaleFac; /*ScaledInt1616_TO_Int32(tmp_yres); */
    jpeg->width = fgWidth;
    jpeg->height = fgHeight;
    jpeg->x = 0;
    jpeg->y = 0;
    jpeg->quality = jpegQual;

    /* set up compression cookie */
    jc.dgp_state = dgp_state;
    jc.size = 0;
    jc.alloced = jpeg->width * jpeg->height / 10;
    DGP_CHECK_ALLOC (jc.buf = DGP_MALLOC(jc.alloced));

    DGP_ERR_PASS_CLEANUP(
          w_jCompress_callback(ipstate, finalPixr, jpeg_Compress_callback, 
           (void *)&jc,  jpegQual, JC_DEFAULT));
    if(finalPixr) w_destroyPixr(ipstate, &finalPixr);

    if (jc.alloced > jc.size){
        DGP_CHECK_ALLOC(tmpptr = (Dgp_Int8 *)DGP_REALLOC(jc.buf, jc.size));
        jc.buf = tmpptr;
    }

    /* save forground */
    jpeg->layer = 1;
    jpeg->data = jc.buf;
    jpeg->data_size = jc.size;
    jpeg->extData_size = 0; 

    DGP_ERR_PASS (DgpAdt_liAdd (fgImages, (void*) jpeg, NULL));

    NORMAL_RETURN;

ERR_CLEANUP:
    if(cImage) DgpImage_free(cImage);
    if(fgImage) DgpImage_free(fgImage);
    if(fgPixr) w_destroyPixr(ipstate, &fgPixr);
    if(finalPixr) w_destroyPixr(ipstate, &finalPixr);

	ERR_RETURN; 
}
