/************************************************************
 * deskew.h
 * $Header: /project/icons/cvsroot/digipaper/src/libtclydoc/deskew.h,v 1.6 2003/02/10 22:53:40 mkrocken Exp $
 ***********************************************************/

#ifndef __DESKEW_H
#define __DESKEW_H

#include "libpixrw.pub"
#include "util.h"
#include "list.h"
#include "mmcompress.h"
#include "dgp_state.h"
#include "tclydoc.h"
#include "tclyout.h"
#include "tclypriv.h"

#define BO_DOWN   0
#define BO_UP     1
#define BO_LEFT   2
#define BO_RIGHT  3


/* defaults value for skew accuracy */
#define DEF_SKEW_ACCURACY 2

/* Used for deskewed tokens. */
typedef struct {
    void *oldExtdata;
    token *newToken;
} extdataFork;

typedef struct {
  Dgp_Int32		baselineOrient;
  Int32         skewAccuracy;

  /* Added 25.4.2001 / dhtesq. */
  /* Context for countTokensToBeDeskewed().  These used to be locals in deskewTokens(). */
  Dgp_Int32		*rotateDims;
  Dgp_Int32		minRotateDim;

  Dgp_State     *dgp_state;
} DeskewStruct;


/* create the deskew structure, giving it the number of pages in the file */
extern ErrCode
deskewInit(	DeskewStruct **deskew_cookie, 
		Dgp_Int32 skewAccuracy, 
		Dgp_Int32 baselineOrient, 
           	Dgp_State *dgp_state);

/* clean up any memory used in the deskew structure */
extern void
deskewCleanup(DeskewStruct *deskew_cookie);

/* find the skew angle for a page */
extern ErrCode
getSkewAngle(	DeskewStruct *deskew_cookie, 
		PIXR *pixr,
		Float32 *angle);

/* Deskew a set of token classes. */
extern ErrCode
deskewTokens(	DeskewStruct *deskew_cookie, 
	        DgpAdt_List newBitmaps, 
		TCLYSectionBlock *sb);

/*
 * 25.4.2001 / dhtesq.
 * Since we're going to be putting lossy, lossless, and deskewed tokens into one contiguous array
 * for libjbig2, we need to know the size of the token array beforehand.  This function counts up
 * the number of nonsingleton deskewed tokens in a section we'll have to make space for, but doesn't do any
 * deskewing.  Call this after completing a section, before writing the global dictionary.
 *
 * With some extra ambition, this function could prepare a list of tokens to be deskewed, saving some
 * work in deskewTokens().
 */
DGP_LIBEXPORT ErrCode
countTokensToBeDeskewed(DeskewStruct		*deskewCookie,
						TCLYSectionBlock	*sb);

/* Deskew all the outliers on a given page. */
extern ErrCode
deskewOutliers(DeskewStruct *deskew_cookie, 
		TCLYPageInfo cPage);

/* Deskew the position of all the cc's on a given page. */
extern ErrCode
deskewPositions( DeskewStruct *deskew_cookie, 
		TCLYPageInfo cpage);
	    

#endif /* __DESKEW_H */


/*******************************************************************
 * Running Log
 * $Log: deskew.h,v $
 * Revision 1.6  2003/02/10 22:53:40  mkrocken
 * Replaced C++ style comments with C style comments (OSG request)
 *
 * Revision 1.5  2001/04/26 22:23:45  dtracy
 * Extensive modifications for unification with lossless tokenizer.
 *
 * Revision 1.4  2001/02/09 23:46:01  jlin
 * add halftone detection and convert to gray
 *
 * Revision 1.3  2000/10/11  18:47:12  dtracy
 * Fixed bug:  decoder was writing 8bpp b&w tiffs.
 *
 * Revision 1.2  2000/03/04 00:22:40  mkrocken
 *
 * merged repositories
 *
 * Revision 1.1  2000/02/17 22:23:19  mkrocken
 * Initial Revision
 *
 * Revision 1.1  1999/12/29 19:49:29  jlin
 * Initial version of deskew.h for libtclydoc
 *
 *******************************************************************/
