/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Pedro Felzenszschwalb? (sp?), 19nn - 1998?
**
**  Filename:  getBoxes.c
**
**  Prepares bitmaps for tokenizer based on the output of libscancolor.
**
**
**	Modification history:
**
**	Dillon H. Tracy esq. / Feb. 2000
**		o  Implemented liberr exception handling.
**		o  Search on 'dhtesq' for some of this coder's comments & changes.
**  
**************************************************************************/

#include <stdlib.h>
#include "list.h"
#include "tok-lossy.h"
#include "image.h"
#include "image_defs.h"
#include "dgp_state.h"
#include "scancolor.h"
#include "compress.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtclydoc/getColorBoxes.c,v 1.7 2003/02/10 22:53:40 mkrocken Exp $");

ErrCode
getColorBoxes(Dgp_State		*dgp_state, 
			  sc_cc			*text_ccs, 
			  Dgp_Int32		numccs,
              Dgp_Int32		pgnum, 
			  DgpAdt_List	*_list) {
	DGP_PROLOGUE("getBoxes()", DGP_PERRSTATE);

	dgplssy_cc	        *box = (dgplssy_cc *)NULL;
	Dgp_Int32			i = 0;
	DgpAdt_List		    boxes = NULL; 
	DgpAdt_ListNode	    retnode;	

	DGP_ERR_PASS(DgpAdt_liNew(&boxes, dgp_state));

	/* 'numccs' is now the count of text-candidate connected components. */
	/*	We regard them as being suitable for boxing. */
	for (i = 0; i < numccs; i++) {
		/*	Grab some memory for the box. */
		DGPLSSY_CC_SEC_MALLOC(&box);
		box->pgnum = pgnum;
		box->runs = NULL;
		box->nruns = 0;

		/*	These coordinates used to be minX and minY of ComponentInfo{}, so correspond to the */
		/*	upper left of the box. */
		box->x = text_ccs[i].x;
		box->y = text_ccs[i].y;

		box->color = text_ccs[i].color;

		box->bitmap.width = (Dgp_UInt16)DGPIMG_WIDTH(text_ccs[i].image);
		box->bitmap.height = (Dgp_UInt16)DGPIMG_HEIGHT(text_ccs[i].image);

		box->bitmap.longsperline = (box->bitmap.width + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;

		/*	Tokens do not appear to use the packed image formalism, but the idea is the same. */
		/*	Except for this protocol difference, is there anything keeping us from pointing */
		/*	bitmap.bits at the existing text_ccs.image?  Straight memcpy() works. */
		/*	Possible performance gain? / dhtesq. */
		DGP_CHECK_ALLOC_CLEANUP(box->bitmap.bits = DGP_MALLOC(sizeof(Dgp_Int32) * box->bitmap.longsperline * box->bitmap.height));

		/*	Copy token image into allocated bitmap. */
		memcpy( box->bitmap.bits, DGPIMG_GETROW(text_ccs[i].image, 0),
				sizeof(Dgp_Int32) * box->bitmap.longsperline * box->bitmap.height);

		/*	We're about to lose this pointer, so stow it somewhere. */
		box->bitmap.userdata = box;

		DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(boxes, box, &retnode));

		/*	Done with original image;  free. */
		if(text_ccs[i].image) 
            DGP_ERR_PASS_CLEANUP(DgpImage_free(text_ccs[i].image));
        text_ccs[i].image = NULL;

		/*	Re-null the box for error handling. */
		box = NULL;
	}

	/*	Done with the text components array. */
	DGP_FREE(text_ccs);
	*_list = boxes;
	NORMAL_RETURN;

ERR_CLEANUP:
	
	/*DGP_ERR_PASS(DgpAdt_liFree(boxes)); */
	if(box != NULL){
		if(box->bitmap.bits != NULL){
			free(box->bitmap.bits);
		}
		DGPLSSY_CC_FREE(box);
	}
	/*	Free remaining text_ccs' images. */
	for( ; i < numccs; i++){
	if(text_ccs[i].image)
        DGP_ERR_PASS(DgpImage_free(text_ccs[i].image));
	}

	free(text_ccs);
	*_list = boxes;
	ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode freeColorBoxList(Dgp_State *dgp_state, DgpAdt_List boxes)
**
**
**  Frees the constituent boxes in the passed list of boxes, and then frees the passed list.
**	Does not free the boxes' bitmaps -- see below.
**
**
**	Arguments:
**		o  boxes is the list to be freed.
**
**	Return value:
**		o  An ErrCode.
**		   
**
**	Modification history:
**
**	Dillon H. Tracy esq. / February 2000
**		o  Implemented liberr exception handling;  return type modified.
**  
**/

ErrCode
freeColorBoxList(Dgp_State *dgp_state, DgpAdt_List boxes)
{
	DGP_PROLOGUE("freeBoxList()", DGP_PERRSTATE);
    DgpAdt_ListNode	ln;
    dgplssy_cc			    *box;
    
    /*
     * The box list contains a bunch of ccs.  We don't own their
     * bitmaps - see below.
     */
    DGPADT_FOREACH (ln, boxes) {
		box = (dgplssy_cc *)DGPADT_LIGET(ln);
		/*
		* We don't own box->member.bits anymore - the tokenizer has them.
		* Unless, that is, we'd turned on TOK_OPT_COPYBITMAPS.  In that
		* case, writeClasses should have freed/NULLed it.  It might no
		* longer be NULL because writePages messes around with
		* them.  In either case, it's not our problem now.
		*/	
		/* We do _not_ free(box->bitmap.bits); */
		DGPLSSY_CC_FREE(box);
	}
    DGP_ERR_PASS(DgpAdt_liFree(boxes));
	NORMAL_RETURN;
}
/*************************************************************************/
