
#ifndef JTOKAPI
#define JTOKAPI

#include "dgp_err_control.h"
#include "dgp_allocators.h"
#include "scancolor.h"
#include "tok-lossy.h"

#include "tok-lossless.h"
#include "token-compress.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef void* DgpApi_Handle;

#define DGP_CLEAR_DICT 0
#define DGP_INCR_DICT 1
#define DGP_CLEAR_NOMERGE 2
#define BO_DOWN   0
#define BO_UP     1
#define BO_LEFT   2
#define BO_RIGHT  3
#define DEFAULT_RES 300

/* defaults value for skew accuracy */
#define DEF_SKEW_ACCURACY 2

typedef struct {
  char* data;				/* pointer to data buffer */
  unsigned long data_size;  /* size of buffer */
  unsigned long extData_size;
} DgpApi_Jbig2Data;

typedef struct {
    Dgp_Int32 sizeThresh;
    Dgp_Int32 classThresh;
    Dgp_Int32 skewAccuracy;
    Dgp_Int32 baselineOrient;
    Dgp_Int32 rotation;
    Dgp_Bool  grayHalftone;
    Dgp_Int32 forceRes;
    Dgp_Bool writeclasses;
    Dgp_Bool instrument;
    Dgp_Bool deskew;
    Dgp_Bool g4_toks;
    Dgp_Int32 jbig2Profile;
    Dgp_Bool bitmapSizeLimit;
    Dgp_Int32 membond;
	/* Added 23.4.2001 / dhtesq. */
	Dgp_Bool verbose;
    /* added 28.5.2002 */
    Dgp_Bool pdfOut;
} DgpApi_Options;

typedef struct {
	unsigned char *image;	/* image pointer	*/
	Dgp_Int32 width;		/* width of this page	*/
	Dgp_Int32 height;		/* height of this page	*/
	Dgp_Int32 bpl;		/* bytes per line	*/
	Dgp_Int32 xRes;		/* horizontal resolution */
	Dgp_Int32 yRes;		/* vertical resolution 	*/ 
} DgpApi_PageData;

/* get a handle to a tokenizer/jbig2encoder object 
   pass in your favorite allocators 

  11 April 2001 / dhtesq.
  Added tokOptions{} for more flexibility;  reverts to old defaults if you
  pass in NULL.
*/

ErrCode
DgpApi_init (void**					handle, 
			 Dgp_Allocators			*a, 
			 Dgp_SmallAllocators	*sm_a, 
             DgpApi_Options			*opts,
			 tokOptions				*passedTokOptions,
			 LlTokOptions			*passedLlTokOptions,
			 DgpTdc_JpegCompInfo	*passedJpegComp);

/* feed a page of binary data to the tokenizer
   the format of the input data is compatible with the IPCore Tiff reader */
ErrCode
DgpApi_addMonoPage (DgpApi_Handle handle, DgpApi_PageData *page, Dgp_Bool *overflow);

ErrCode
DgpApi_addColorTokens (DgpApi_Handle handle, sc_cc *text_ccs, Dgp_Int32 numccs,
                     DgpApi_PageData *data, Dgp_Bool *overflow, DgpAdt_List fgImages);
ErrCode
DgpApi_addColorTokensSumGray (DgpApi_Handle handle, sc_cc *text_ccs, Dgp_Int32 numccs,
                     DgpApi_PageData *data, DgpImage_Gray grayImage, Dgp_Bool *overflow, DgpAdt_List fgImages);


/* end the section and get the section dictionary
   The encoded dictionary is returned in a DgpApi_Jbig2Data struct.
   The caller must free the allocated memory.
   dictopt can be either DGP_CLEAR_DICT or DGP_INCR_DICT */
ErrCode
DgpApi_endSection (DgpApi_Handle handle, unsigned char dictopt); 
ErrCode
DgpApi_getDict(DgpApi_Handle handle, Dgp_Bool final, DgpApi_Jbig2Data *data);
ErrCode
DgpApi_endSectionGetDict (DgpApi_Handle handle, unsigned char dictopt, DgpApi_Jbig2Data *data);
/*
 * Reduce token noise for color image
 */
ErrCode
DgpApi_redTokNoise(DgpApi_Handle handle);

/* Check the memory of the dictionary
 */
ErrCode DgpApi_checkTokMemoryBond(void *handle, Dgp_Bool *overflow);
/* returns the position block for a page
   Can only be called after DgpApi_endSectionGetDict has been called
   Must be called as many times as DgpApi_addMonoPage was called in this section */
ErrCode
DgpApi_getNextPosBlock (DgpApi_Handle handle, DgpApi_Jbig2Data *posBlock);

/* close handle and free all resources */
ErrCode
DgpApi_close (DgpApi_Handle handle);

/* start a section*/
ErrCode
DgpApi_startSection (void* handle);

/* clean up on section*/
ErrCode
DgpApi_cleanSection (void* handle);

/* monochrome image */
ErrCode
DgpApi_addMonoBuffer (void* handle, 
                      Dgp_UInt8 *buf, Dgp_Int32 bufWidth, Dgp_Int32 bufHeight, 
                      Dgp_Int32 bpl, Dgp_Int32 pageHeight, 
                      Dgp_Int32 xRes, Dgp_Int32 yRes,
                      Dgp_Bool firstBuf, Dgp_Bool lastBuf, Dgp_Bool *overflow);

#ifdef __cplusplus
}
#endif
#endif

/* USAGE:

  First create a handle by calling DgpApi_init

  Then call DgpApi_addMonoPage as often as needed with one page of binary data at a time
  After all pages have been passed to the tokenizer, or if memory becomes tight, call
  DgpApi_endSectionGetDict. This will close the current section and create a symbol dictionary
  for this section. If dictopt is set to DGP_CLEAR_DICT the dictionary is deleted from
  memory. If dictopt is DGP_INCR_DICT, then the dictionary is kept in memory, and will be used
  for the next pages. When DgpApi_endSectionGetDict is called again, the new dictionary returned
  by the function only contains tokens that have not been saved in a previous dictionary. 

  After calling DgpApi_endSectionGetDict you must call DgpApi_getNextPosBlock as many times as
  you called DgpApi_addMonoPage for this section. Thus if you passed 3 pages by calling
  DgpApi_addMonoPage 3 times, you must call DgpApi_getNextPosBlock 3 times after you closed the
  section. Each call will return the encoded data for one page. The pages are returned in the
  order they were passed in.

  After all encoded position blockes have been retrieved, you can call DgpApi_addMonoPage again
  to start a new section. Or if there are no more pages to be encoded, call DgpApi_close to free
  all remaining memory.

 */

