#include "types.pub"
#include "ipscore.prv"
#include "frames.pub"
#include "iaerror.pub"
#include "defines.pub"

/* prototyes */
#include "shrcc.pub"
#include "shrcc.prv"
#ifdef __GNUC__
#include "ansiprot.h"
#endif
#include "linecc.h"

ErrCode
tcly_findBufCC (
                      IPState    *pIPState,
                      CCState    *pCCState,
                      Int8* buf,
                      Bool  lastBuf,
                      Int32 startX,
                      Int32 endX,
                      Int32 startY,
                      Int32 endY,
                      Int32 accuHeight,
                      Int32 sectionHeight,
                      Int32 bpl,
                      Int32	minWidth,
                      Int32	minHeight,
                      const Int32	connectType,
                      void       *pRectData,
                      Int32 (CDECL *rectReturnProc) (
                                  IPState    *pIPState,
                                  void       *,
                                  Int32	x,
                                  Int32	y,
                                  Int32	w,
                                  Int32	h,
                                  Int32	m,
                                  Int32	nRuns,
                                  TCScanLineRun *pRuns),
                      Int32      *pRectReturnCode)
{
    PROLOGUE;
    TCBlockHeader	*pSymbolHeap;       /* heap for CCSymbol structs */
    TCBlockHeader	*pRunHeap;	    /* heap for TCScanLineRun structs */
    TCScanLineRun	*pLine0 = NULL;	    /* ptr to older scan line */
    TCScanLineRun	*pLine1 = NULL;	    /* ptr to newer scan line */
    TCScanLineRun	*pLine0Left, *pLine0Right; /* ptrs to left and right
						                        * ends of a joined section
						                        * on line 0 */
    TCScanLineRun	*pLine1Run;		/* ptr to a run in line 1 */
    Int32		 currentLine;		/* current line number */
    CCSymbol		*pCurrentSym, *pNewSym;	/* ptrs to symbols */
    Int32		 nextSrcX, nextDestX;	/* used when merging lists
						                 * of runs */
    TCScanLineRun	*pSrcRun, *pDestRun;	/* used to point to list
						                    * of runs to be merged */
    TCScanLineRun	*pPrevRun;		        /* points to run immediately
						                     * before pSrcRun or pDestRun
						                     * during searches */
    Int8 *lineHeadPtr = NULL;
    Int endYY;

	/* see if args are legal */
    if ((connectType > 1) || (pCCState == NULL) ||
	(pCCState->pRunHeap == NULL) || (pCCState->pSymbolHeap == NULL) ||(rectReturnProc == NULL))
	return ia_invalidParm;
	/* Set local heap pointers.  These are both easier to use and
	 * increase similarity between this routine and
	 * i_findConnectedComponentBoundingBoxes.
	 */
    pSymbolHeap = pCCState->pSymbolHeap;
    pRunHeap = pCCState->pRunHeap;

	/* similarly, get the pointers to the existing lists of allocated
	 * runs.  They may be left over from a previous call or were
	 * initialized by i_createCCState.
	 */
    pLine0 = pCCState->pLine0;
    pLine1 = pCCState->pLine1;


	/* MAIN LOOP */
	/* In this loop, we get new set of runs; process all the joins
	 * in the upper line; then assign symbols to all the runs
	 * in the lower line.
	 * We do one too many lines do the last line is actually retired.
	 * In the last line, we just say there are no runs.
	 */
    if(lastBuf)
        endYY = endY + accuHeight + 1;
    else 
        endYY = endY + accuHeight;
    for (currentLine = startY + accuHeight; currentLine <= endYY;  currentLine++)
    {
        /* exchange the structures for the current and previous lines */
        pLine0Left = pLine1;
        pLine1 = pLine0;
        pLine0 = pLine0Left;
        lineHeadPtr = buf + (currentLine - accuHeight) * bpl;
        if( (currentLine - startY + accuHeight) == sectionHeight - 1){
            pLine1->pNext->runLeft = cVeryLargeNum;	/* make a blank line */
            pLine1->pNext->runRight = cVeryLargeNum;
        }else if (currentLine <= endY + accuHeight) {
            __error__ =  i_findAllRunsInLine(pIPState,
                            lineHeadPtr,
                            endX,
                            startX,
                            pLine1->pNext, 
                            pRunHeap);	

            if (__error__ != ia_successful)
            {
                __error__ = ia_callbackError;
                goto findCCRReturn;
            }
        }else if(lastBuf) {
            pLine1->pNext->runLeft = cVeryLargeNum;	/* make a blank line */
            pLine1->pNext->runRight = cVeryLargeNum;
        }
        
        /* ----- FIRST PASS -------- */
        
        /* loop through the runs in line 0 looking for runs without
        * symbols or runs that are joined on line 1 */
        for (pLine0Left = pLine0, pLine0Right = pLine0->pNext,
            pLine1Run = pLine1, pNewSym = NULL;
        pLine0Right->runLeft < cVeryLargeNum;
        pLine0Left = pLine0Right, pLine0Right = pLine0Right->pNext)
        {
            pCurrentSym = pNewSym;
            pNewSym = (CCSymbol *)pLine0Right->pClient;
            while ((pNewSym != NULL) && (pNewSym->pNewSymbol != NULL))
            pNewSym = pNewSym->pNewSymbol;		/* find the real symbol
                                                * associated with
            * this run */
            
            /* try to find a run that joins pLine0Left and pLine0Right */
            while (pLine1Run->runRight < pLine0Right->runLeft - connectType)
                pLine1Run = pLine1Run->pNext;
            
            if ((pLine1Run->runLeft <= pLine0Left->runRight + connectType) &&
                (pLine1Run->runRight >= pLine0Right->runLeft - connectType))
            {
            /* pLine0Left should be joined to pLine0Right.  Are
            * they already part of the same symbol?  Note that
            * we will only detect a join between two real runs
            * (not the terminating runs on the ends) since
            * the leftmost terminating run as a big negative
            * coordinate and all real runs have positive coords.
                */
                if (pCurrentSym != pNewSym)	/* need to do a join */
                {
                    if (pNewSym == NULL)
                        /* cool, just add this new thing to the existing sym */
                    {
                        pNewSym = pCurrentSym;
                        pLine0Right->pClient = pCurrentSym;
                        pCurrentSym->numActiveRuns++;
                        pCurrentSym->numRuns++;
                        pCurrentSym->mass += pLine0Right->runRight -
                            pLine0Right->runLeft + 1;
                        if (pCurrentSym->maxX < pLine0Right->runRight)
                            pCurrentSym->maxX = pLine0Right->runRight;
                    }
                    else
                    {				/* need to join two real syms */
                                    /* Keep the taller symbol, or the one of the
                                    * left if they're the same height.  Why does
                                    * the test below work?  If they're different
                                    * heights, the first clause works.  If they`re
                                    * the same height and that height is 1, we
                                    * took care of that case above.  Therefore,
                                    * the second clause only comes into play when
                                    * they are the same height and the height is
                                    * greater than 1, therefore, both symbols
                                    * already have non-null pRunHead pointers.
                        */
                        if ((pNewSym->minY < pCurrentSym->minY) ||
                            ((pNewSym->minY == pCurrentSym->minY) &&
                            (pNewSym->pRunHead != NULL) &&
                            (pNewSym->pRunHead->runLeft <
                            pCurrentSym->pRunHead->runLeft)))
                            {	/* pNewSym will survive, so exchange
                                pNewSym with pCurrentSym.
                            */
                            CCSymbol	*pTempSym;
                            pTempSym = pNewSym;
                            pNewSym = pCurrentSym;
                            pCurrentSym = pTempSym;
                        }
                        
                        pNewSym->pNewSymbol = pCurrentSym;
                        if (pCurrentSym->minX > pNewSym->minX)
                            pCurrentSym->minX = pNewSym->minX;
                        if (pCurrentSym->maxX < pNewSym->maxX)
                            pCurrentSym->maxX = pNewSym->maxX;
                        if (pCurrentSym->minY > pNewSym->minY)
                            pCurrentSym->minY = pNewSym->minY;
                        if (pCurrentSym->maxY < pNewSym->maxY)
                            pCurrentSym->maxY = pNewSym->maxY;
                        pCurrentSym->numActiveRuns += pNewSym->numActiveRuns;
                        pCurrentSym->numRuns += pNewSym->numRuns;
                        pCurrentSym->mass += pNewSym->mass;
                        
                        /* Now for the fun part.  We need to merge the
                        * lists of runs for the two symbols.
                        * Since the runs point from the lower
                        * right to the upper left, we start
                        * merging from the bottom and work up.
                        * Since we chose pCurrentSym to have the
                        * topmost, leftmost run, we're sure that
                        * we'll run of source runs before we
                        * run out of dest runs.
                        *
                        * NOTE: we don't put runs onto the list
                        * for a symbol until we're done with line 0.
                        * If a symbol, say pNewSym, started one
                        * line ago, no runs have been put on its
                        * list, so pNewSym->pRunTail == NULL.
                        * Therefore, we shouldn't touch it.
                        */
                        pSrcRun = pNewSym->pRunTail;
                        pDestRun = pCurrentSym->pRunTail;
                        /* We want pPrevRun to point somewhere harmless
                        * so that we can set pPrevRun->pNext without
                        * checking to see if pPrevRun is ok all the
                        * time.  We know that we'll never get to
                        * the first run in the destination symbol,
                        * so this is a harmless place.  After we're
                        * done with the merge, we set pCurrentSym->pRunTail
                        * to the value that was stored in
                        * pCurrentSym->runHead->pNext since it
                        * will have been set to point to the
                        * rightmost, bottomost run.
                        */
                        pPrevRun = pCurrentSym->pRunHead;
                        
                        /* This is the loop that does the main merging.
                        * Within a line, we will alternate some runs
                        * from the source with some runs from the
                        * dest.  Scan lines may start and end with
                        * any combination of source and dest runs.
                        * It is this consideration that leads us to
                        * a fairly general loop.
                        *
                        * pPrevRun points to the last run in the merged
                        * list so far.
                        * pSrcRun points to the next source run.
                        * pDestRun points to the next dest run.
                        * nextSrcX is the value we give to the
                        * left coord of the next source run.  This is
                        * set to -cVeryLargeNum at the left end of the
                        * source line.  Remember that we work from
                        * right to left in each scan line.
                        * Ditto for nextDestX.
                        * At the top of the loop, we see if BOTH
                        * nextSrcX and nextDestX are equal to
                        * -cVeryLargeNum.  If so, we're at the left end of
                        * a scan line, so set them to the left coords
                        * of pSrcRun and pDestRun respectively.
                        * Then decide which is larger, nextSrcX or
                        * nextDestX.  If the source is larger, the
                        * next source run goes into the list next.
                        * make pPrevRun->pNext point to it.  The
                        * similar thing happens if the dest is to the
                        * right instead of the source.  Then we follow
                        * whichever list of runs (source or dest) we've
                        * chosen until we find the end of the scan
                        * line or we've gone past the right endpoint of
                        * the set of runs we didn't follow.  We leave
                        * pPrevRun pointing at the last run before
                        * the one that went too far.  If we ran off
                        * the end of the scan line, we set nextSrcX
                        * (if we followed the source) or nextDestX
                        * (if we followed the dest) to -cVeryLargeNum.
                        * Otherwise, we set it to the left coord of
                        * the run that was too far to the right.
                        * We then return to the top of the loop.
                        * This will cause us to follow the other set
                        * of runs and so on.
                        * Note that we need not test for pDestRun
                        * being NULL because we're guaranteed to run
                        * out of source runs before we run out of
                        * dest runs.
                        */
                        nextSrcX = -cVeryLargeNum;
                        nextDestX = -cVeryLargeNum;
                        while (pSrcRun != NULL)
                        {
                        /* Are we at the left end of a scan line?
                        * If so, we had better have a run for both
                        * source and dest (no need to check for NULL)
                            */
                            if ((nextSrcX == -cVeryLargeNum) &&
                                (nextDestX == -cVeryLargeNum))
                            {
                                nextSrcX = pSrcRun->runLeft;
                                nextDestX = pDestRun->runLeft;
                            }
                            
                            /* Should we follow source runs or dest
                            * runs? */
                            if (nextSrcX > nextDestX)
                            {	/* follow source runs */
                                pPrevRun->pNext = pSrcRun;
                                
                                if (nextDestX > -cVeryLargeNum)
                                {	/* we're looking for the last src run,
                                    * or the end of the src line, or
                                    * a gap ending the bunch of
                                    * contiguous source runs that fit
                                    * before the next dest run.
                                    */
                                    while (cTrue)
                                    {
                                        pPrevRun = pSrcRun;
                                        pSrcRun = pSrcRun->pNext;
                                        if ((pSrcRun == NULL) ||
                                            (pSrcRun->runRight >=
                                            pPrevRun->runLeft - 1))
                                        {
                                            nextSrcX = -cVeryLargeNum;
                                            break;
                                        }
                                        if (pSrcRun->runLeft < nextDestX)
                                        {
                                            nextSrcX = pSrcRun->runLeft;
                                            break;
                                        }
                                    }
                                }
                                else
                                {	/* we're looking for the last src run,
                                    * or the end of the src line */
                                    do
                                    {
                                        pPrevRun = pSrcRun;
                                        pSrcRun = pSrcRun->pNext;
                                    } while ((pSrcRun != NULL) &&
                                        (pSrcRun->runRight <
                                        pPrevRun->runLeft - 1));
                                    nextSrcX = -cVeryLargeNum;
                                }
                            }
                            else	/* follow dest runs */
                            {
                                pPrevRun->pNext = pDestRun;
                                
                                if (nextSrcX > -cVeryLargeNum)
                                {	/* we're looking for the end of the
                                    * dest line, or
                                    * a gap ending the bunch of
                                    * contiguous source runs that fit
                                    * before the next source run.
                                    */
                                    while (cTrue)
                                    {
                                        pPrevRun = pDestRun;
                                        pDestRun = pDestRun->pNext;
                                        if (pDestRun->runRight >=
                                            pPrevRun->runLeft - 1)
                                        {
                                            nextDestX = -cVeryLargeNum;
                                            break;
                                        }
                                        if (pDestRun->runLeft < nextSrcX)
                                        {
                                            nextDestX = pDestRun->runLeft;
                                            break;
                                        }
                                    }
                                }
                                else
                                {	/* we're looking for the end of
                                    * the dest line */
                                    do
                                    {
                                        pPrevRun = pDestRun;
                                        pDestRun = pDestRun->pNext;
                                    } while (pDestRun->runRight <
                                        pPrevRun->runLeft - 1);
                                    nextDestX = -cVeryLargeNum;
                                }
                            }
                            }	/* end of loop that merges the rest of the
                                * runs in the two symbols
                        */
                        
                        /* We exit this loop with pPrevRun pointing
                        * to the last run from the source.  Make it
                        * point to the next run from the dest.
                        */
                        pPrevRun->pNext = pDestRun;
                        
                        pNewSym = pCurrentSym;	/* merge these so that the
                                                * "pCurrentSym = pNewSym"
                                                * statement at the top of
                                                * the loop works.
                        */
                        /* set the symbol's end pointer to point to
                        * the new end of the list of runs.  This was
                        * set the to point to the very first run in
                        * the list because we initialized pPrevRun
                        * to be pCurrentRun->pRunHead.
                        * However, if there weren't really any runs
                        * attached to pNewSym, we have done nothing
                        * to any of the pointers, so don't try to
                        * undo it (pCurrentSym->pRunHead still
                        * equals NULL!).
                        */
                        if (pCurrentSym->pRunHead->pNext != NULL)
                        {
                            pCurrentSym->pRunTail =
                                pCurrentSym->pRunHead->pNext;
                            pCurrentSym->pRunHead->pNext = NULL;
                        }
                        
                        }	/* End of case where two existing symbols are
                            * joined.
                            */
                            }	/* End of case where joined runs do not already
                                * belong to the same symbol (one might below to
                                * to no symbol at all yet).
                                */
        }	/* end of case where there was a join */
        
        /* If this run has no symbol, give it one */
        if (pNewSym == NULL)
        {
            __error__ = i_allocateStruct (pIPState, pSymbolHeap,
                (void **) &pNewSym);
            if (__error__ != ia_successful)
                goto findCCRReturn;
            
            pNewSym->minX = pLine0Right->runLeft;
            pNewSym->maxX = pLine0Right->runRight;
            pNewSym->minY = currentLine - 1;
            pNewSym->maxY = 0;			/* not used here */
            pNewSym->pRunHead = pNewSym->pRunTail = NULL;
            pNewSym->numActiveRuns = 1;
            pNewSym->numRuns = 1;
            pNewSym->mass = pLine0Right->runRight -
                pLine0Right->runLeft + 1;
            pNewSym->pNewSymbol = NULL;
            pLine0Right->pClient = pNewSym;
        }
        }	/* End of first pass through pLine0.  When we exit this
            * loop, each run has a symbol and all joins have happened.
            */
            
            /* ----- SECOND PASS -------- */
            
            /* Now we go through the runs again, determining how to change
            * the number of active runs for each symbol and copying symbol
            * pointers to runs on line 1 that overlap runs on line 0.
            */
            for (pLine0Left = pLine0->pNext, pLine1Run = pLine1->pNext;
            pLine0Left->runLeft < cVeryLargeNum;
            pLine0Left = pLine0Right)
            {
            /* First, we go through all the obsolete symbols associated
            * with pLine0Left; decrementing numActiveRuns for each one.
            * If the count goes to 0, we free it.
                */
                for (pCurrentSym = (CCSymbol *)pLine0Left->pClient;
                pCurrentSym->pNewSymbol != NULL;
                )
                {
                    pCurrentSym->numActiveRuns--;
                    if (pCurrentSym->numActiveRuns <= 0)
                    {
                        pNewSym = pCurrentSym;
                        pCurrentSym = pCurrentSym->pNewSymbol;
                        i_freeStruct (pIPState, pSymbolHeap, (void **) &pNewSym);
                    }
                    else
                        pCurrentSym = pCurrentSym->pNewSymbol;
                    
                }
                
                /* At this point, pCurrentSym points to the real symbol for
                * pLine0Left.  Move this run to the list maintained for its
                * symbol.  We need to do this before calling the
                * client's callback proc, so we might as well do it now.
                */
                pLine0Right = pLine0Left->pNext;
                pLine0Left->pNext = pCurrentSym->pRunTail;
                if (pLine0Left->pNext == NULL)
                    pCurrentSym->pRunHead = pLine0Left;
                pCurrentSym->pRunTail = pLine0Left;
                
                /* Next, try to find an overlapping run on line 1 */
                while (pLine1Run->runRight < pLine0Left->runLeft - connectType)
                    pLine1Run = pLine1Run->pNext;
                
                if (pLine1Run->runLeft <= pLine0Left->runRight + connectType)
                {
                /* In this case, we know there is a run under this
                * one, so the real symbol to which this run belongs
                * cannot end here.
                * At this point, pCurrentSym points to the real
                * symbol for this run. We know it cannot end since
                * there is a run under it.  If that run has not yet
                * been tagged, we don't change
                * this symbol's number of active runs (one above
                * becomes one below) but do tag the run below
                * with this symbol.  Also, update the symbol's
                * bounding box using this run's coords.
                * If it has already been tagged, decrement the
                * number of active runs for this symbol to
                * account for the join.
                    */
                    if (pLine1Run->pClient == NULL)
                    {
                        pLine1Run->pClient = pCurrentSym;
                        if (pCurrentSym->minX > pLine1Run->runLeft)
                            pCurrentSym->minX = pLine1Run->runLeft;
                        pCurrentSym->mass += pLine1Run->runRight -
                            pLine1Run->runLeft + 1;
                        pCurrentSym->numRuns++;
                    }
                    else
                    {
                        pCurrentSym->numActiveRuns--;
                    }
                    
                    /* Now tag any more runs on line 1 that overlap
                    * this run in line 0.
                    */
                    while (pLine1Run->pNext->runLeft <=
                        pLine0Left->runRight + connectType)
                    {
                        pLine1Run = pLine1Run->pNext;
                        pLine1Run->pClient = pCurrentSym;
                        pCurrentSym->numActiveRuns++;
                        pCurrentSym->numRuns++;
                        pCurrentSym->mass += pLine1Run->runRight -
                            pLine1Run->runLeft + 1;
                    }
                    
                    /* now that we're at the right end of the runs
                    * that overlap this run, fix up the symbol's maxX.
                    */
                    if (pCurrentSym->maxX < pLine1Run->runRight)
                        pCurrentSym->maxX = pLine1Run->runRight;
                    
                }/*else if(currentLine == accuHeight + endY){ */
                /*} */
                else	/* there is no overlapping run below this one */
                {
                    /* see if the real symbol is done */
                    pCurrentSym->numActiveRuns--;
                    /* if it went to 0, set the lower y coordinate and
                    * return the bounding box */
                    if (pCurrentSym->numActiveRuns <= 0)
                    {
                    /* only tell the client about boxes that have
                    * adequate dimensions.  Note that since we did not
                    * subtract the frame above, we don't need to
                    * add it back in now.
                        */
                        if ((pCurrentSym->maxX - pCurrentSym->minX + 1 >= minWidth)
                            &&
                            (currentLine - pCurrentSym->minY >= minHeight))
                        {
                            __error__ =
                                rectReturnProc(
                                pIPState,
                                pRectData,
                                pCurrentSym->minX,
                                pCurrentSym->minY,
                                pCurrentSym->maxX - pCurrentSym->minX + 1,
                                currentLine - pCurrentSym->minY,
                                pCurrentSym->mass,
                                pCurrentSym->numRuns,
                                pCurrentSym->pRunTail);
                            if (__error__ != ia_successful)
                                goto findCCRBadCallback;
                        }
                        
                        /* free all the runs associated with this symbol */
                        for (pSrcRun = pCurrentSym->pRunTail;
                        pSrcRun != NULL;
                        pSrcRun = pDestRun)
                        {
                            pDestRun = pSrcRun->pNext;
                            i_freeStruct (pIPState, pRunHeap, (void **) &pSrcRun);
                        }
                        /* and free the symbol itself */
                        i_freeStruct (pIPState, pSymbolHeap,
                            (void **) &pCurrentSym);
                    }
                    }	/* end of case where there is not run on line1 overlapping
                        * current run on line 0
                */
    }	/* end of second pass through runs on line 0 */
    
        /* At this point, pLine0Left is pointing to the last buffer
        * run in line 0.  Attach it to the first buffer run
        * to make line 0 legally empty.
        */
        pLine0->pNext = pLine0Left;
        
    }	/* end of loop through the entire image */
    
    /* Things must have worked out ok */
    __error__ = ia_successful;
    goto findCCRReturn;
    
    /* Something bad happened.  Free the heaps and return */
findCCRBadCallback:
    *pRectReturnCode = __error__;
    __error__ = ia_callbackError;
    
findCCRReturn:
/* Remember the pointers to the runs that have already been
* allocated.  We will want to use these the next time
* this routine is called.  This is somewhat faster than
* putting them all back into the heaps and starting over
* on the next call.
* But aren't the lists always empty, or don't they only
* ever have just the dummies in them?  Well, if we were
* called immediately after the client called
* i_findBoundingBoxesForSomeCCs, then the lists weren't
* originally empty, but they are now.  We should remember
* that.
    */
    pCCState->pLine0 = pLine0;
    pCCState->pLine1 = pLine1;
   
    NORMAL_RETURN;
}
