#include "compress.h"

#include "math.h"

#include "tok-lossy.h"

/* include-libwjr*/
#include <util.h>
#include <insertbox.h>
#include <simple_utils.h>

#include "image.h"
#include "image_defs.h"

#include "strel.pub"
#include "shrip.pub"
#include "thin.pub"
#include "shrip.prv"
#include "props.pub"
#include "PsgFix.h"

/* prototype */
#include "smooth_token.h"

#define SEARCH  3
#define FRAME  12
#define MBRFRAME 6

ErrCode
packedToPixr(
              Dgp_State *dgp_state,
              Dgp_Int32 *bits,
              Dgp_Int32 width,
              Dgp_Int32 height,
              Dgp_Int32 longsperline,
              PIXR **pixr)
{
    Dgp_Int32 j, bpl;
    Dgp_Int32 *ptr,*srcPtr, *destPtr;
    IPWState    *ipwstate = dgp_state->ipwstate;
    Dgp_Int32 pixrlpl;

   	DGP_PROLOGUE ("packedToPixr", DGP_PERRSTATE);

    DGP_ERR_PASS_CLEANUP(w_createPixr(	
                                      DGP_PIPWSTATE, 
                                      width, 
                                      height, 
                                      1, 
                                      Int32_TO_ScaledInt1616((Int32)300), 
                                      Int32_TO_ScaledInt1616((Int32)300), 
                                      cFrameBits, 
                                      pixr));
    DGP_ERR_PASS_CLEANUP(w_pixrGetBpl(ipwstate, *pixr, &bpl));
    DGP_ERR_PASS_CLEANUP(w_pixrGetImage(ipwstate, *pixr, (Dgp_UInt8 **)&ptr));
    srcPtr = bits;
    destPtr = ptr;
    pixrlpl = bpl / sizeof(Dgp_Int32);
    for(j= 0; j < height; j ++){
        srcPtr += longsperline ;
        destPtr += pixrlpl;
        memcpy(destPtr, srcPtr, longsperline*sizeof(Dgp_Int32));
     }
    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

ErrCode
getImgDblAverage(Dgp_State *dgp_state, 
           DgpImage_Double image, 
           Dgp_Int32 count)
{
    DGP_PROLOGUE ("getImgDbAverage", DGP_PERRSTATE);
    Dgp_Int32 i, j;

    Dgp_Int32 width = DGPIMG_WIDTH(image);
    Dgp_Int32 height = DGPIMG_HEIGHT(image);

    DGP_ERR_GEN(!image, cErrInvalidParm);
    DGP_ERR_GEN( count == 0, cErrInvalidParm);
    for( j =0;  j < height ; j++){
        for (i = 0; i < width; i++){ 
            DGPIMG_REF(image, i, j) /= count;
        }
    }
    /*
    for( j =0;  j < height ; j++){
        for (i = 0; i < width; i++){ 
            val = DGPIMG_REF(image, i, j);
            if(max < val) max = val;
            if(min > val) min = val;
        }
    }
    for( j =0;  j < height ; j++){
        for (i = 0; i < width; i++){ 
            val =  DGPIMG_REF(image, i, j) / (max - min) * 255.0;
            DGPIMG_REF(image, i, j) = val;
        }
    }*/
    NORMAL_RETURN;
}
void
packedClear ( DgpImage_Packed image)
{
  Dgp_Int32 *ptr;
  Dgp_Int32 height = DGPIMG_HEIGHT(image);
  Dgp_Int32 bpl = DGPIMG_GETROWWIDTH(image);
  ptr = DGPIMG_PIXPTR(image,0,0);
  memset((void*)ptr, 0, height * bpl);
}


ErrCode
packedAddFrame(   
            Dgp_State *dgp_state,
            DgpImage_Packed image,
            Dgp_Int32 frame,
            DgpImage_Packed *newimage)
{
    DGP_PROLOGUE ("packedAddFrame", DGP_PERRSTATE);
    Dgp_Int32 i, j;

    Dgp_Int32 width = DGPIMG_WIDTH(image);
    Dgp_Int32 height = DGPIMG_HEIGHT(image);

    DGP_ERR_GEN(!newimage, cErrInvalidParm);
    DGP_ERR_GEN(!image, cErrInvalidParm);

    DGP_ERR_PASS_CLEANUP(DgpImage_new(dgp_state, DGPIMAGE_PACKED, width + 2*frame, height + 2*frame, (void **)newimage));

    /* clear image */
    packedClear(*newimage);

   /*copy data */
    for( j =0;  j < height; j++){
        for (i = 0; i < width;  i++){ 
            if( DGPIMG_GETBIT(image, i, j))
                DGPIMG_SETBIT(*newimage, i + frame , j+frame);
        }
    }

    NORMAL_RETURN;
ERR_CLEANUP:
    if(*newimage) DGPIMAGE_FREE(*newimage);
    *newimage = NULL;
    ERR_RETURN;

}

/*
 * Get new examplar by thresholding the summed gray image
 */
ErrCode
getExemplar(
            Dgp_State *dgp_state,
            DgpImage_Double sum,
            DgpImage_Packed dltExemplar,
            DgpImage_Packed thinExemplar,
            Dgp_Int32 minWidth,
            Dgp_Int32 minHeight,
            DgpImage_Packed framedExemplar,
            DgpImage_Packed *newExemplar)
{
   	DGP_PROLOGUE ("getExemplar", DGP_PERRSTATE);
    
    Dgp_Int32 i=0,j=0, j0=0, ii=0,jj=0;
    Dgp_Int32 exmpWidth, exmpHeight;
    Dgp_Int32 fcount=0, bcount=0;
    Dgp_Float64 forground=0.0, background=0.0, threshold = 0, aveThreshold = 0 ;
    Dgp_Int32 minX = 0xffff, minY = 0xffff, maxX = 0, maxY = 0;
    Dgp_Float64 val;
    Dgp_Int32 height, width;
    Dgp_Int32 xalign, yalign;
    
    DGP_ERR_GEN(!sum, cErrInvalidParm);
    DGP_ERR_GEN(!dltExemplar, cErrInvalidParm);
    DGP_ERR_GEN(!framedExemplar, cErrInvalidParm);

    height = DGPIMG_HEIGHT(sum); 
    width = DGPIMG_WIDTH(sum);

    xalign = (width - minWidth) /2;
    yalign = (height - minHeight) /2;
    /*Get average threshold */
    for( j = 1 ;  j < minHeight -1; j++){
       for (i = 1 ; i < minWidth -1 ; i++){
            ii = i + xalign;
            jj = j + yalign;
            val =  DGPIMG_REF(sum, ii, jj);
            if(DGPIMG_GETBIT(thinExemplar, ii, jj)) { 
                forground += val;
                fcount++;
            } else if( val !=0 &&
                        ((DGPIMG_GETBIT(dltExemplar, ii-2, jj) && !DGPIMG_GETBIT(dltExemplar, ii, jj)) ||
                        (DGPIMG_GETBIT(dltExemplar, ii+2, jj) && !DGPIMG_GETBIT(dltExemplar, ii, jj))) ){
                    background += val; 
                    bcount++;
            }
        }
    }
    if(fcount == 0) threshold = 0xffff;
    else if(bcount ==0) threshold = 0;
    else  aveThreshold = (forground/fcount + background/bcount) *.5;

    for( j =0;  j < minHeight; j++){
        forground =0;
        background =0;
        fcount =0;
        bcount = 0;
        for(j0 = j-2; j0 <= j+2; j0++){
            for (i = 0; i < minWidth; i++){
                ii = i + xalign;
                jj = j0 + yalign;
                val =  DGPIMG_REF(sum, ii, jj);
                if(DGPIMG_GETBIT(thinExemplar, ii, jj)) { 
                    forground += val;
                    fcount++;
                }else if( val !=0 &&
                        ((DGPIMG_GETBIT(dltExemplar, ii-2, jj) && !DGPIMG_GETBIT(dltExemplar, ii, jj)) ||
                        (DGPIMG_GETBIT(dltExemplar, ii+2, jj) && !DGPIMG_GETBIT(dltExemplar, ii, jj))) ){
                    background += val; 
                    bcount++;
                }
            }
        }
        if(fcount == 0) threshold = aveThreshold;
        else if(bcount ==0) threshold = aveThreshold;
        else threshold = (forground/fcount + background/bcount) *.5;
        for (i = 0; i < minWidth; i++){
            ii = i + xalign;
            jj = j + yalign;
            val =  DGPIMG_REF(sum, ii, jj);
            if((val > threshold) && DGPIMG_GETBIT(dltExemplar, ii, jj))
                DGPIMG_SETBIT(framedExemplar, ii, jj);
        }
    }


    /* Find new bounding box */
    for( j =0;  j < height; j++){
        for (i = 0; i < width; i++){ 
            if(DGPIMG_GETBIT(framedExemplar, i, j)){
                if(minX >= i) minX = i;
                if(minY >= j) minY = j;
                if(maxX <= i) maxX = i;
                if(maxY <= j) maxY = j;
            }
        }
    }
    if(minX < maxX && minY < maxY){
        DGP_ERR_PASS_CLEANUP(
        DgpImage_new(dgp_state, DGPIMAGE_PACKED, maxX - minX + 1, maxY - minY + 1, (void *)newExemplar));
        packedClear(*newExemplar);

        exmpHeight = DGPIMG_HEIGHT(*newExemplar); 
        exmpWidth = DGPIMG_WIDTH(*newExemplar);
        for( j =0;  j < exmpHeight; j++){
            for (i = 0; i < exmpWidth; i++){ 
                if(DGPIMG_GETBIT(framedExemplar, i + minX, j + minY)){
                    DGPIMG_SETBIT(*newExemplar, i, j);
                }
            }
        }
    }else *newExemplar = NULL;

NORMAL_RETURN;
ERR_CLEANUP:
   if(newExemplar) DGPIMAGE_FREE(newExemplar);
    ERR_RETURN;

}


/*
 * Polinomial interpolation (exact form) ( from numerical recipe)
 */
void
pInter1dExact(
Dgp_Float64 *xa,
Dgp_Float64 *ya,
Dgp_Int32 n,
Dgp_Float64 x,
Dgp_Float64 *y,
Dgp_Float64 *dy)
{
    Dgp_Float64 temp1, temp2;
    Dgp_Int32 i, j;

    *y =0;
    for(j=0; j < n; j++){
        temp1 = 1;
        temp2 = 1;
        for(i = 0; i < n; i++){
            if(i != j){
                temp1 *= (x - xa[i]);
                temp2 *= (xa[j] - xa[i]);
            }
        }
        *y += temp1 * ya[j] /temp2;
    }
    *dy = 0;
}

/*
 * Polinomial interpolation (recursive)( from numerical recipe, seems have problem in result)
 */
ErrCode
pInter1d(
Dgp_State * dgp_state,
Dgp_Float64 *xa0,
Dgp_Float64 *ya0,
Dgp_Int32 n,
Dgp_Float64 x,
Dgp_Float64 *y,
Dgp_Float64 *dy)
{
    Dgp_Int32 i, m, ns =1;
    Dgp_Float64 den, dif, dift, ho, hp, w;
    Dgp_Float64 *c=NULL, *d=NULL, * xa=NULL, *ya=NULL;

   	DGP_PROLOGUE ("pInter1d", DGP_PERRSTATE);
    
    DGP_ERR_GEN(!xa0, cErrInvalidParm);
    DGP_ERR_GEN(!ya0, cErrInvalidParm);
    DGP_ERR_GEN(n<0, cErrInvalidParm);

    DGP_CHECK_ALLOC_CLEANUP(c = (Dgp_Float64 *)DGP_MALLOC( (n+1) * sizeof(Dgp_Float64)));
    DGP_CHECK_ALLOC_CLEANUP(d = (Dgp_Float64 *)DGP_MALLOC( (n+1) * sizeof(Dgp_Float64)));

    DGP_CHECK_ALLOC_CLEANUP(xa = (Dgp_Float64 *)DGP_MALLOC( (n+1) * sizeof(Dgp_Float64)));
    DGP_CHECK_ALLOC_CLEANUP(ya = (Dgp_Float64 *)DGP_MALLOC( (n+1) * sizeof(Dgp_Float64)));
    for(i=1; i <=n; i++){
        xa[i] = xa0[i-1];
        ya[i] = ya0[i-1];
    }
    dif = fabs(x-xa[1]);
    ns = 1;
    /* Here we find the index ns of the closest table entry */
    for(i=1; i <= n; i++){
        if( ( dift = fabs(x - xa[i])) < dif){
            ns =i;
            dif = dift;
        }
        /* intialize table c's and d's */
        c[i] = ya[i];
        d[i] = ya[i];
    }

    /* This is the initial appoximation to y */
    *y= ya[ns--];

    /* For each clolumn of the tableau, */
    /* we loop over the current c's and d's and update them */
    for(m = 1; m < n; m++){
        for(i = 1; i < n - m; i++){
            ho = xa[i] - x;
            hp = xa[i + m] - x;
            w = c[i+1] - d[i];
            DGP_ERR_GEN_CLEANUP((den = ho - hp) == 0.0, cErrInternal); 
            den = w / den;
            d[i] = hp * den;
            c[i] = ho * den;
        }
        *y += (*dy = (2 * ns < (n-m) ? c[ns+1] : d[ns--]));
    }
    DGP_FREE(c);
    DGP_FREE(d);
    DGP_FREE(xa);
    DGP_FREE(ya);
    NORMAL_RETURN;

ERR_CLEANUP:
    
    if(d) DGP_FREE(d);
    if(c) DGP_FREE(c);
    if(xa) DGP_FREE(xa);
    if(ya) DGP_FREE(ya);
    ERR_RETURN;
}

/*
 * Polinomial interpolation for 2d ( from numerical recipe)
 */
ErrCode
pInter2d(
Dgp_State * dgp_state,
Dgp_Float64 *x1a,
Dgp_Float64 *x2a,
Dgp_Float64 **ya,
Dgp_Int32 m,
Dgp_Int32 n,
Dgp_Float64 x1,
Dgp_Float64 x2,
Dgp_Float64 *y,
Dgp_Float64 *dy)
{
   	DGP_PROLOGUE ("pInter2d", DGP_PERRSTATE);
    int j;
    Dgp_Float64 *ymtmp;
    
    DGP_ERR_GEN(!x1a, cErrInvalidParm);
    DGP_ERR_GEN(!x2a, cErrInvalidParm);
    DGP_ERR_GEN(!ya, cErrInvalidParm);
    DGP_ERR_GEN(!*ya, cErrInvalidParm);
    DGP_ERR_GEN(n<0, cErrInvalidParm);
    DGP_ERR_GEN(m<0, cErrInvalidParm);

    DGP_CHECK_ALLOC_CLEANUP(ymtmp = (Dgp_Float64 *)DGP_MALLOC( m * sizeof(Dgp_Float64)));

    for(j=0; j < m; j++){
        pInter1dExact(x2a, ya[j], n, x2, &ymtmp[j], dy);
    }
    pInter1dExact(x1a, ymtmp, m, x1, y, dy);
    DGP_FREE(ymtmp);

    NORMAL_RETURN;

ERR_CLEANUP:
    
    if(ymtmp) DGP_FREE(ymtmp);
    ERR_RETURN;

}
 
/*Ailgn each gray instances with binary exemplar by subpixel */
ErrCode
optAlignSub(Dgp_State *dgp_state,
             DgpImage_Gray cim,
             DgpImage_Packed exemplar,
             Dgp_Int32 x,
             Dgp_Int32 y,
             Dgp_Int32 mbrWidth,
             Dgp_Int32 mbrHeight,
             Dgp_Float64 *dx,
             Dgp_Float64 *dy,
             Dgp_Bool *reversVedio
             )
{
    Dgp_Int32 dbWidth, dbHeight;
    Dgp_Int32 i0,j0,ii,jj, i, j;
    Dgp_Float64 ** corelation = NULL;
    Dgp_Float64 interp2dFinal[2*SEARCH -1][2*SEARCH-1];
    Dgp_Float64 xx[SEARCH], yy[SEARCH];
    Dgp_Float64 max = 0, dif=0;
    Dgp_Float64 val = 0, sum = 0;
    Dgp_Bool reverse = DGP_FALSE;
    Dgp_Float64 forground =0, background=0;
    Dgp_Int32 fcount =1, bcount=1, xalign, yalign;
    Dgp_Int32  index[SEARCH];
   	
    DGP_PROLOGUE ("optAlignSub", DGP_PERRSTATE);
    
    dbWidth = DGPIMG_WIDTH(exemplar);
    dbHeight = DGPIMG_HEIGHT(exemplar);

    xalign = (dbWidth - mbrWidth)/2;
    yalign = (dbHeight - mbrHeight)/2 ;

    /* find out if this is reverse vedio*/
    for (j = 0; j < mbrHeight; j++) {
        for (i = 0; i < mbrWidth; i++) {
            val = DGPIMG_REF(cim, x + i , y + j);
            jj = j + yalign;
            ii = i + xalign;
            if( DGPIMG_GETBIT(exemplar, ii, jj)) {
                forground += val;
                fcount ++;
            } else {
                background += val;
                bcount ++;
            }
        }
    }

    if(forground / fcount > background /bcount ) reverse = DGP_TRUE;

    DGP_CHECK_ALLOC_CLEANUP(corelation = (Dgp_Float64 **)DGP_MALLOC( SEARCH * sizeof(Dgp_Float64 *)));
    for (j0 = 0; j0 < SEARCH; j0++) {
        DGP_CHECK_ALLOC_CLEANUP(corelation[j0] = (Dgp_Float64 *)DGP_MALLOC( SEARCH * sizeof(Dgp_Float64)));
    }


    /* Align gray instance with exemplar, move gray instance around SEARCH * SEARCH grid,  */
    /* find corelation for each move */
    for (j0 = 0; j0 < SEARCH; j0++) {
        index[j0] = j0;
        for (i0 = 0; i0 < SEARCH; i0++) {
            sum = 0.0;
            for (j = 0; j <mbrHeight; j++) {
                for (i = 0; i < mbrWidth; i++) {
                    val = DGPIMG_REF(cim, x + i , y + j);
                    if(reverse != DGP_TRUE) val = 255 - val;
                    ii = i + xalign + i0 - SEARCH/2;
                    jj = j + yalign + j0 - SEARCH/2;
                    if( DGPIMG_GETBIT(exemplar, ii, jj)){
                        sum += val;
                    }
                }
            }
            corelation[j0][i0] = sum;
        }
    }


    /* use interpolation to double the number of points in corelation[][] */
    for(j=0; j<SEARCH; j++){
        xx[j] = j;
        yy[j] = j;
        for(i=0; i<SEARCH; i++){
            interp2dFinal[j*2][i*2] = corelation[j][i];
        }
    }


    for(j=0; j< 2*SEARCH-1; j++){
        for(i=0; i< 2* SEARCH-1; i++){
            dif = 0;
            if(i%2 !=0 || j%2 !=0)
                DGP_ERR_PASS( pInter2d( dgp_state, 
                                        xx, yy, 
                                        corelation, 
                                        SEARCH, SEARCH, 
                                        j/2.0, i /2.0, /* row index, col index  */
                                        &interp2dFinal[j][i] ,
                                        &dif));
        }
    }


    i0 = SEARCH -1;
    j0 = SEARCH -1;
    /* Find max value to decide offset */
    for(j = 0; j < 2 * SEARCH - 1; j++){
        for(i=0; i< 2 * SEARCH - 1; i++){
            if(max < interp2dFinal[j][i]){
                i0 = i;
                j0 = j;
                max = interp2dFinal[j][i];
            }
        }
    }

    *dx = (i0 - (SEARCH-1))/2.0;
    *dy = (j0 - (SEARCH-1))/2.0;
    *reversVedio = reverse;

    if(corelation){
        for (j0 = 0; j0 < SEARCH; j0++) {
            if(corelation[j0]) DGP_FREE(corelation[j0]);
        }
        DGP_FREE(corelation);
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    if(corelation){
        for (j0 = 0; j0 < SEARCH; j0++) {
            if(corelation[j0]) DGP_FREE(corelation[j0]);
        }
        DGP_FREE(corelation);
    }
    ERR_RETURN;

}

ErrCode
createSum(
           Dgp_State *dgp_state,
           DgpImage_Gray cim,
           DgpImage_Double sum,
           DgpImage_Packed exemplar,
           Dgp_Int32 index,
           Dgp_Int32 x,
           Dgp_Int32 y,
           Dgp_Int32 mbrWidth,
           Dgp_Int32 mbrHeight)
{
   	DGP_PROLOGUE ("createHist", DGP_PERRSTATE);
    Dgp_Int32 i,j, ii,jj;
    Dgp_Int32 xalign, yalign;
    Dgp_Float64 dx=0.0, dy=0.0, mean = 0.0;
    Dgp_Int32 fdx = 0, fdy = 0 ;
    Dgp_Float64 val;
    Dgp_Bool reverse = DGP_FALSE;

    index = index;
    /* Enlarge member area by MBRFRAME */
    if(x >= MBRFRAME && y >= MBRFRAME){
        x -= MBRFRAME;
        y -= MBRFRAME;
        mbrWidth += 2 * MBRFRAME;
        mbrHeight += 2 * MBRFRAME;
    }

    /*Sub pixel alignment */
    DGP_ERR_PASS( optAlignSub(dgp_state, cim, exemplar, x, y, mbrWidth, mbrHeight,  &dx, &dy , &reverse));

    /* get mean and standard deviation of the image */
    for (j = 0; j < mbrHeight ; j++) {
        for (i = 0; i < mbrWidth; i++) {
            val =  DGPIMG_REF(cim,i + x, y + j); 
            if(reverse != DGP_TRUE) val = 255 - val; 
            mean += val;
        }
    }
    mean /= (mbrHeight * mbrWidth);

    /* std is not used here
    for (j = 0; j < mbrHeight ; j++) {
        for (i = 0; i < mbrWidth; i++) {
            val =  DGPIMG_REF(cim,i + x, y + j); 
            if(reverse != DGP_TRUE) val = 255 - val; 
            temp = val - mean;
            std += temp*temp;
        }
    }
    std = sqrt(std) / (mbrHeight * mbrWidth -1);
    */

    /* get alignment relative to center */
    xalign = DGPIMG_WIDTH(sum);
    yalign = DGPIMG_HEIGHT(sum);
    xalign = (xalign - mbrWidth)/2;
    yalign = (yalign - mbrHeight)/2;

    /* add the nomalized image into sum. val = (val - mean) + 255,  */
    /* consider sub pixel alignment */
    if( dx == 0.0) 
        fdx = (Dgp_Int32) dx;
    else 
        fdx =  (Dgp_Int32)(dx / fabs(dx) * floor(fabs(dx)));
    if(dy == 0.0) 
        fdy = (Dgp_Int32) dy;
    else
        fdy =  (Dgp_Int32)(dy / fabs(dy) *floor(fabs(dy)));
    dx -= fdx;
    dy -= fdy;
    for (j = 0; j < mbrHeight ; j++) {
        for (i = 0; i < mbrWidth; i++) {
            val =  DGPIMG_REF(cim, i + x, y + j); 
            if(reverse != DGP_TRUE) {
                val = 255.0 - val;
            }    
            val =  val - mean + 255;      /* normalize by subtract mean */
            ii = i + xalign + fdx ; 
            jj = j + yalign + fdy;
            DGPIMG_REF(sum, ii,                         jj)                     += val / 4.0;
            DGPIMG_REF(sum, ii + (Dgp_Int32)(dx *2),    jj)                     += val / 4.0;     
            DGPIMG_REF(sum, ii,                         jj + (Dgp_Int32)(dy *2))+= val / 4.0;
            DGPIMG_REF(sum, ii + (Dgp_Int32)(dx *2),    jj + (Dgp_Int32)(dy *2))+= val / 4.0;
            
        }
    }
    NORMAL_RETURN;
}


ErrCode
sumGrayInstances(
Dgp_State *dgp_state, 
DgpImage_Gray cim,
DgpAdt_List tokList,
DgpAdt_List ccList,
Dgp_Int32 pageNum)
{
    DgpImage_Double sum = NULL;
    DgpImage_Packed exemplar = NULL, framedExemplar = NULL;
    DgpAdt_ListNode ln, memberln;
    dgplssy_cc *membercc;
    tokClass *c;
    Dgp_Int32  sumWidth, sumHeight, mbrWidth, mbrHeight;
    tokOffset * offset;
    Dgp_UInt32 *retval;
    GrayToken *tokGray = NULL;

   	DGP_PROLOGUE ("smoothToken", DGP_PERRSTATE);

    DGP_ERR_GEN(!tokList, cErrInvalidParm);
    DGP_ERR_GEN(!ccList, cErrInvalidParm);

#ifdef DGP_DDBUG
            {
                DgpImage_save(cim, "h:\\gray.pnm");
            }
#endif
    /* Loup through all clases... */
    DGPADT_FOREACH (ln, tokList) {
        c = (tokClass *) DGPADT_LIGET(ln);
        DGP_ASSERT(c != NULL);
        if(DGPADT_LILEN(c->members) > 0){
            /* create image to hold sum of the gray instances */
            sumWidth = c->exemplar.width + 2 * FRAME;
            sumHeight = c->exemplar.height + 2 * FRAME;
            if(!c->userdata){
                DGP_ERR_PASS_CLEANUP(DgpImage_new(dgp_state, DGPIMAGE_DOUBLE, sumWidth, sumHeight, (void **)&sum));
                DGP_CHECK_ALLOC_CLEANUP(tokGray = (GrayToken *)DGP_MALLOC(sizeof(GrayToken)));
                tokGray->width = 0xffff;
                tokGray->height = 0xffff;
                tokGray->index = 0;
                tokGray->image = sum;
                tokGray->userdata = c;
                sum = NULL;
            }else tokGray = (GrayToken *) c->userdata;
            /* wrap exemplar in DgpImage_Packed and convert it to Pixr */
            DGP_ERR_PASS_CLEANUP(
                DgpImage_wrap(dgp_state, c->exemplar.bits, c->exemplar.width, c->exemplar.height, c->exemplar.longsperline, &exemplar));
            DGP_ERR_PASS_CLEANUP(packedAddFrame(dgp_state, exemplar, FRAME, &framedExemplar));
            DGP_ERR_PASS_CLEANUP(DgpImage_unwrap(exemplar, (Dgp_UInt32 **)&retval));
            retval = NULL;
            /* loop through this class and add gray instances into sum, do alignment */
            DGPADT_FOREACH (memberln, c->members) {
                offset = (tokOffset *)DGPADT_LIGET(memberln);
                DGP_ASSERT(offset != (tokOffset *)NULL);
                membercc = (dgplssy_cc *)offset->userdata;
                DGP_ASSERT(membercc != (dgplssy_cc *)NULL);
                if(membercc->pgnum == pageNum){
                    mbrWidth = membercc->bitmap.width;
                    mbrHeight = membercc->bitmap.height;
                    if(tokGray->width > mbrWidth) tokGray->width = mbrWidth;
                    if(tokGray->height > mbrHeight) tokGray->height = mbrHeight;
                    DGP_ERR_PASS_CLEANUP(createSum(dgp_state, cim, tokGray->image, framedExemplar, tokGray->index, membercc->x, membercc->y, mbrWidth, mbrHeight)); 
                    tokGray->index++;
                }
            }
            /* point this class's userdata to it's gray average  */
            c->userdata = (void *)tokGray;
            /* Free resouse for this class */
            DGPIMAGE_FREE(framedExemplar);
            framedExemplar = NULL;
            tokGray = NULL;
        }
    }
    NORMAL_RETURN;

ERR_CLEANUP:
        if(sum) DGPIMAGE_FREE(sum);
        sum = NULL;
        if(framedExemplar) DGPIMAGE_FREE(framedExemplar);
        framedExemplar = NULL;
        if(exemplar) DgpImage_unwrap(exemplar, &retval);
        exemplar = NULL;
        if(tokGray) DGP_FREE(tokGray);
        retval = NULL;
	ERR_RETURN;
}

ErrCode
smoothToken(
Dgp_State *dgp_state, 
DgpAdt_List tokList)
{

    DgpImage_Packed exemplar = NULL, dilatedExemplar = NULL, framedExemplar = NULL, newExemplar = NULL, thinExemplar = NULL;
    DgpAdt_ListNode ln;
    tokClass *c;
    Dgp_Int32 dtWidth, dtHeight, dtBpl, sumWidth, sumHeight;
    Dgp_UInt32 *retval, *ptr;
    Dgp_Int32 minWidth = 0xffff, minHeight = 0xffff;
    PIXR *pixr = NULL, *dilatedPixr = NULL, *thinPixr = NULL;
    STREL *pStrel2 = NULL;
    GrayToken *tokGray;

   	DGP_PROLOGUE ("smoothToken", DGP_PERRSTATE);

    DGP_ERR_GEN(!tokList, cErrInvalidParm);

   
    /* create structure element */
    DGP_ERR_PASS_CLEANUP (w_initStrelByName(DGP_PIPWSTATE, strel2));
    DGP_ERR_PASS_CLEANUP (w_getStrel(DGP_PIPWSTATE, strel2, &pStrel2));

    /* Loup through all clases... */
    DGPADT_FOREACH (ln, tokList) {
        c = (tokClass *) DGPADT_LIGET(ln);
        minWidth = 0xffff;
        minHeight = 0xffff;
        DGP_ASSERT(c != NULL);
        if(c->userdata){
            tokGray = (GrayToken *) c->userdata;
            DGP_ASSERT(tokGray->image);
            /* create image to hold sum of the gray instances */
            sumWidth  = DGPIMG_WIDTH(tokGray->image);
            sumHeight = DGPIMG_HEIGHT(tokGray->image);
            /* get average gray img */
            getImgDblAverage(dgp_state, tokGray->image, tokGray->index);
            /* wrap exemplar in DgpImage_Packed and convert it to Pixr */
            DGP_ERR_PASS_CLEANUP(
                DgpImage_wrap(dgp_state, c->exemplar.bits, c->exemplar.width, c->exemplar.height, c->exemplar.longsperline, &exemplar));
            DGP_ERR_PASS_CLEANUP(packedAddFrame(dgp_state, exemplar, FRAME, &framedExemplar));
            DGP_ERR_PASS_CLEANUP(
                packedToPixr( dgp_state, 
                                DGPIMG_PIXPTR(framedExemplar, 0, 0),
                                DGPIMG_WIDTH(framedExemplar), 
                                DGPIMG_HEIGHT(framedExemplar), 
                                DGPIMG_GETROWWIDTH(framedExemplar) / sizeof(Dgp_Int32), 
                                &pixr));
           
            /* Dilate the exemplar */
            DGP_ERR_PASS_CLEANUP(w_dilate(DGP_PIPWSTATE, pixr, pStrel2, &dilatedPixr));
            DGP_ERR_PASS( w_thin8 (DGP_PIPWSTATE, pixr,  (PIXR **)&thinPixr));
#ifdef DGP_DDBUG
            {
                DgpImage_Gray sumGray = NULL;
                DgpImage_imcDtoGAuto(dgp_state, tokGray->image, &sumGray);
                DgpImage_save(sumGray, "h:\\sum.pnm");
                DgpImage_free(sumGray);
  	            DGP_ERR_PASS_CLEANUP(
                    w_writeTiff(DGP_PIPWSTATE, pixr, "exemplar.tif", COMPRESSION_CCITTFAX4 ,cTrue, 0));
  	            DGP_ERR_PASS_CLEANUP(
                    w_writeTiff(DGP_PIPWSTATE, dilatedPixr, "dilate.tif", COMPRESSION_CCITTFAX4 ,cTrue, 0));
  	            DGP_ERR_PASS_CLEANUP(
                    w_writeTiff(DGP_PIPWSTATE, thinPixr, "thin.tif", COMPRESSION_CCITTFAX4 ,cTrue, 0));
            }
#endif
            DGP_ERR_PASS( w_pixrGetWidth(DGP_PIPWSTATE, dilatedPixr, &dtWidth));
	        DGP_ERR_PASS( w_pixrGetHeight(DGP_PIPWSTATE, dilatedPixr, &dtHeight));
	        DGP_ERR_PASS( w_pixrGetBpl(DGP_PIPWSTATE, dilatedPixr, &dtBpl));
            DGP_ERR_PASS_CLEANUP(w_pixrGetImage(DGP_PIPWSTATE, thinPixr, (Dgp_UInt8 **) &ptr));
            DGP_ERR_PASS_CLEANUP(
                DgpImage_wrap(dgp_state, ptr, dtWidth, dtHeight, dtBpl / sizeof(Int32), &thinExemplar));
            ptr = NULL;
            /* wrap dilated exemplar in DgpImage_Packed  */
            DGP_ERR_PASS_CLEANUP(w_pixrGetImage(DGP_PIPWSTATE, dilatedPixr, (Dgp_UInt8 **) &ptr));
            DGP_ERR_PASS_CLEANUP(
                DgpImage_wrap(dgp_state, ptr, dtWidth, dtHeight, dtBpl / sizeof(Int32), &dilatedExemplar));

            /* release resource */
            DGP_ERR_PASS_CLEANUP(w_destroyPixr(DGP_PIPWSTATE, &pixr));
            pixr = NULL;
            packedClear (framedExemplar);
            /* create a new exemplar based on summed gray image, dltExemplar */
            getExemplar(    dgp_state, 
                            tokGray->image, 
                            dilatedExemplar, 
                            thinExemplar, 
                            tokGray->width + 2 * MBRFRAME, 
                            tokGray->height + 2* MBRFRAME, 
                            framedExemplar, 
                            &newExemplar);

            /*copy newExemplar to exemplar */
            DgpImage_unwrap(exemplar, &retval);
            if(newExemplar){
#ifdef DGP_DDBUG
            {
             DGP_ERR_PASS_CLEANUP(
                 packedToPixr( dgp_state, 
                                DGPIMG_PIXPTR(newExemplar, 0, 0),
                                DGPIMG_WIDTH(newExemplar), 
                                DGPIMG_HEIGHT(newExemplar), 
                                DGPIMG_GETROWWIDTH(newExemplar) / sizeof(Dgp_Int32), 
                                &pixr));

  	            DGP_ERR_PASS_CLEANUP(
                    w_writeTiff(DGP_PIPWSTATE, pixr, "newExemplar.tif", COMPRESSION_CCITTFAX4 ,cTrue, 0));
                /* clear resource */
                DGP_ERR_PASS_CLEANUP(w_destroyPixr(DGP_PIPWSTATE, &pixr));
                pixr = NULL;
            }
#endif
                DGP_FREE(c->exemplar.bits);
                c->exemplar.width = (Dgp_UInt8)DGPIMG_WIDTH(newExemplar);
                c->exemplar.height = (Dgp_UInt8)DGPIMG_HEIGHT(newExemplar);
                c->exemplar.longsperline = DGPIMG_GETROWWIDTH(newExemplar)/sizeof(Dgp_Int32);
                DgpImage_unwrap(newExemplar, &retval);
                c->exemplar.bits = retval; 
            }
            /* Free resouse for this class */
            DGPIMAGE_FREE(framedExemplar);
            framedExemplar = NULL;
            DGPIMAGE_FREE(tokGray->image);
            DGP_FREE(tokGray);
            c->userdata = NULL;
            DgpImage_unwrap(dilatedExemplar, &retval);
            dilatedExemplar = NULL;
            DGP_ERR_PASS_CLEANUP(w_destroyPixr(DGP_PIPWSTATE, &dilatedPixr));
            dilatedPixr = NULL;
            DgpImage_unwrap(thinExemplar, &retval);
            thinExemplar = NULL;
            DGP_ERR_PASS_CLEANUP(w_destroyPixr(DGP_PIPWSTATE, &thinPixr));
            thinPixr = NULL;
        }
    }
    NORMAL_RETURN;

ERR_CLEANUP:
        if(dilatedExemplar) DgpImage_unwrap(dilatedExemplar, &retval);
        dilatedExemplar = NULL;
        if(dilatedPixr) w_destroyPixr(DGP_PIPWSTATE, &dilatedPixr);
        dilatedPixr = NULL;
        if(framedExemplar) DGPIMAGE_FREE(framedExemplar);
        framedExemplar = NULL;
        if(thinExemplar) DgpImage_unwrap(thinExemplar, &retval);
        thinExemplar = NULL;
        if(thinPixr) w_destroyPixr(DGP_PIPWSTATE, &thinPixr);
        thinPixr = NULL;
        if(pixr) w_destroyPixr(DGP_PIPWSTATE, &pixr);
        pixr = NULL;
        if(exemplar) DgpImage_unwrap(exemplar, &retval);
        exemplar = NULL;
        if(newExemplar) DGPIMAGE_FREE(newExemplar);
        newExemplar = NULL;
        retval = NULL;
	ERR_RETURN;

}
