/**********************************************************
* tclydoc.c
* $Header: /project/icons/cvsroot/digipaper/src/libtclydoc/tclydoc.c,v 1.45 2003/02/10 22:53:40 mkrocken Exp $
*********************************************************/
#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif

/* include-ipcore*/
#include <libutilw.pub>
#include <libpixrw.pub>
#include <librastw.pub>
#include <libiow.pub>
#include <props.pub>
#include <frames.pub>

#include "tclydoc.h"
#include "tok-lossy.h"
#include "tclyout.h"
#include "tclypriv.h"
#include "uds.h"


#include "compress.h"
#include "tiff_cc2box.h"
#include "deskew.h"
#include "getColorBoxes.h"

#include "dgp_err_printers.h"
#include "dgp_sys_err_control.h"
#include "util.h"

#include <errno.h>
#include "dgp_utils.h"
#include <math.h>

#include "blit.h"
#include "image_defs.h"

#include "insertbox.h"
#include "simple_utils.h"

#include "tokclass.h"
#include "token-compress.h"
#include "token-compress-priv.h"
#include "halftone.h"
#include "colorfg.h"
#include "PsgFix.h"

/* macro to free a cc */
#define CC_FREE(_pcc) \
  (dgp_state->sm_a.freeSmall(dgp_state->sm_a.smallCookie, DGPLSSY_CC_POOL_ID, (void *)(_pcc)))

#define addPage(curPage, sb) \
	if((sb)->numPages >= (sb)->numPagesAlloced){ \
	Dgp_Int32 temp_int =	((sb)->numPages * 3) / 2; \
	DGP_SEC_REALLOC((sb)->pages, (TCLYPageInfo *), temp_int *sizeof(TCLYPageInfo)); \
	(sb)->numPagesAlloced = temp_int; \
	} \
	curPage.index = sb->numPages; \
	sb->pages[sb->numPages] = curPage; \
    sb->numPages++;

/***************************************************************
 * Init the TCLYDocCookie
 * The errPrinter must be valid
 ***************************************************************/

DGP_LIBEXPORT ErrCode
tcly_doc_init(
			  Dgp_State *dgp_state,  
			  TCLYDocCookie *_cookie, 
			  Dgp_Bool lossy, 
			  Dgp_Bool grayHalftone,
			  Dgp_Int32 sizeThreshold,  
			  Dgp_Int32 forceResolution,
			  Dgp_Bool maxTokenSize,
			  tokOptions *tokOpt, 
			  Dgp_Bool deskew,
			  Dgp_Int32 skewAccuracy,
			  Dgp_Int32 baselineOrient,
              Dgp_Int32 membond,
			  Dgp_Bool report)
{
	DGPERR_INIT(tcly_doc_init);
	TCLYDocState  *cookie;
	DGP_ERR_BEGIN(dgp_state);
	
	DGP_SEC_MALLOC(cookie, (TCLYDocState *), sizeof(TCLYDocState));
	cookie->dps 		= dgp_state;
	cookie->lossy 		= lossy;
	cookie->deskew 		= deskew;
	cookie->sizeThreshold 	= sizeThreshold;
	cookie->tokOpt 		= tokOpt;
    cookie->maxTokenSize = maxTokenSize;
	cookie->grayHalftone 	= grayHalftone;
	cookie->tokstate 	= NULLTOKSTATE;
	cookie->curBlock 	= NULL;
	cookie->tokDgpAdt_List		= DGP_NULLLIST;
    cookie->memBond = membond;
	DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&cookie->tokDgpAdt_List, dgp_state));
	cookie->forceResolution = forceResolution;
        cookie->deskew_cookie = (TCLYDeskewCookie)NULL;

	/* validate small pool usage */
	DGP_SMALLPOOL_VALIDATE(dgplssy_cc, DGPLSSY_CC_POOL_ID);

#ifndef DGP_FAXMODE
	if (deskew) 
		DGP_ERR_PASS_CLEANUP(	
			deskewInit((DeskewStruct **)(&(cookie->deskew_cookie)), 
			skewAccuracy, 
			baselineOrient, 
			dgp_state));
#endif
	DGP_REPORTF0(report, "starting document");
	
	*_cookie = (TCLYDocCookie) cookie;
	
	NORMAL_RETURN;
	
ERR_CLEANUP:
	if(cookie->tokDgpAdt_List != DGP_NULLLIST) DgpAdt_liFree(cookie->tokDgpAdt_List);
	cookie->tokDgpAdt_List		= DGP_NULLLIST;
	if(cookie->deskew_cookie) DGP_FREE(cookie->deskew_cookie);
	if(cookie) DGP_FREE(cookie);
	ERR_RETURN;
}

void tcly_doc_pageInit(TCLYPageInfo *curPage){
	curPage->pgNum = 0;
	curPage->index = -1;
	curPage->xRes = 0;
	curPage->yRes = 0;
	curPage->width = 0;
	curPage->height	= 0;
	curPage->numOutliers = 0;
	curPage->outliers = DGP_NULLLIST;
	curPage->boxes 	= DGP_NULLLIST;
	curPage->numBoxes = 0;
	curPage->halftones = DGP_NULLLIST;
    curPage->imageinfo = DGP_NULLLIST;
	curPage->color = DGP_FALSE;
	curPage->skewedTokenPerPage = 0;
	curPage->residualTokenCount = 0;
	curPage->tokMem = 0;
    curPage->pageTokMem =0;
    curPage->largeCCList = DGP_NULLLIST;
    curPage->ccData = NULL;
}

void DgpTdc_pageInit( DgpTdc_PageInfo *curPage){
	curPage->pageNum = 0;
	curPage->index = -1;
	curPage->xRes = 0;
	curPage->yRes = 0;
	curPage->width = 0;
	curPage->height	= 0;
    curPage->numToks =0;
	curPage->outliers = DGP_NULLLIST;
	curPage->boxes 	= NULL;
	curPage->numBoxes = 0;
	curPage->color = DGP_FALSE;
	curPage->tokMem = 0;
    curPage->pageTokMem =0;
	curPage->colors = NULL;
	curPage->largeImages = DGP_NULLLIST;
	curPage->annotations = NULL;
	curPage->num_annotations = 0;
	curPage->cell_size = 0;
	curPage->halftone_boxes = NULL;
	curPage->halftone_numBoxes = 0;

}

/* how big to allocate some of our internal arrays initially (they'll
be reallocated if we get more than this many pages) */
#define INIT_NUM_PAGES 15 
#define INIT_NUM_TOKS  50
#define INIT_NUM_POS   100

/***************************************************************
* Starting to tokenize a section of pages defined by number of
* of incremental dictionary or number of clear dictionary.
***************************************************************/
DGP_LIBEXPORT ErrCode
tcly_doc_startSection(TCLYDocCookie _cookie, 
					  TCLYSectionBlockPtr prev, 
					  Dgp_Int32 docPageNum,
					  Dgp_Bool report)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	Dgp_State *dgp_state = cookie->dps;
	DGPERR_INIT(tcly_doc_startSection);
	TCLYSectionBlock *sb = NULL;
	
	/*DGP_REPORTF0(report, "starting new section"); */
	
	/* Start the tokenizer if it isn't started because:
	1. This is the first section
	2. The last section was final
	*/
	
	if((TCLYSectionBlock *)prev != (TCLYSectionBlock *)NULL){
        sb = (TCLYSectionBlock *) prev;
        DGP_ERR_PASS_CLEANUP(tcly_doc_cleanSection(_cookie,report));
		sb->sectionStartedAtThisPage = docPageNum;
		
	} else {
		if(cookie->tokstate == NULLTOKSTATE){
			/* fire up the main tokenizer */
			DGP_ERR_PASS_CLEANUP(tokInit(	cookie->tokOpt, &(cookie->tokstate)));
			DGP_SEC_MALLOC(	sb, (TCLYSectionBlock *), sizeof(TCLYSectionBlock));
			DGP_SEC_MALLOC(	sb->pages, (TCLYPageInfo *), sizeof(TCLYPageInfo) * INIT_NUM_PAGES);
			sb->numPagesAlloced = INIT_NUM_PAGES; 
			DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&sb->classesFound, dgp_state));

			/* initialize the block*/
			sb->numPages = 0;
            sb->sectionStartedAtThisPage = docPageNum;
            sb->tokMem = 0;
			cookie->curBlock = sb;
			sb = (TCLYSectionBlock *)NULL;
		} else {
			DGP_ERR_CLEANUPF0(cErrInternal, "Need a previous section to restart the tokenizer.");
		}
	}
	
	NORMAL_RETURN;
	
ERR_CLEANUP:
	/* Need to add frees for all allocations*/
	if (sb)
	{
		if(sb->pages) DGP_FREE(sb->pages);
		DGP_FREE(sb);
	}
	ERR_RETURN;
	
}


/***************************************************************
 * Create a new page monocrome page 
 ***************************************************************/
DGP_LIBEXPORT ErrCode
tcly_doc_addMonoPage(UInt8 *image, 
					 TCLYDocCookie _cookie, 
					 DgpTdc_Cookie _llTdoc,
					 Dgp_Int32 width,
					 Dgp_Int32 height,
					 Dgp_Int32 bpl,
					 Dgp_Int32 xRes,
					 Dgp_Int32 yRes,
					 Dgp_Int32 pgNum, 
                     Dgp_Bool noRegion,
					 Dgp_Bool report)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	tdoc_state *llTdoc = (tdoc_state *)_llTdoc;
	Dgp_State *dgp_state = cookie->dps;
	DGPERR_INIT(tcly_doc_addMonoPage);
	
	TCLYPageInfo *curPage = &(cookie->curPage);

    report = report;

    /*initialize parameters */
    DGP_ERR_PASS_CLEANUP( tcly_startPage(     
                        (TCLYDocCookie) cookie, 
                        (DgpTdc_Cookie) llTdoc,
                        pgNum,
                        width,
                        height,
                        xRes,
                        yRes));
	
    /* find ccs and tokenize them */
    /* find halftones list */
    /* deskew if required */
    /* return data in ccData */
    DGP_ERR_PASS_CLEANUP(getBoxesLossy(	
		dgp_state,
		image, 
		curPage->width, 
		curPage->height, 
		bpl, 
        cookie->sizeThreshold,
		cookie->grayHalftone, 
		cookie->deskew_cookie,
		&(curPage->skewAngle),
		&curPage->halftones,
        &curPage->imageinfo,
		NULL,		/* cc finder callback... */
        curPage->ccData,
		(struct timeval *)NULL,
		(struct timeval *)NULL)); 

    /* put related data into a page */
    DGP_ERR_PASS_CLEANUP(tcly_endPage(         
                        (TCLYDocCookie) cookie, 
                        (DgpTdc_Cookie) llTdoc,
                        DGP_TRUE,
                        noRegion));


        NORMAL_RETURN;
	
ERR_CLEANUP:
	ERR_RETURN; 
}

#ifndef DGP_FAXMODE


/***************************************************************
 * add color Tokens page
 ***************************************************************/
DGP_LIBEXPORT ErrCode
tcly_doc_addColorTokens(sc_cc *text_ccs,
				     Dgp_Int32 numccs,
					 TCLYDocCookie _cookie, 
					 DgpTdc_Cookie _llCookie,
					 Dgp_Int32 width,
					 Dgp_Int32 height,
					 Dgp_Int32 bpl,
					 Dgp_Int32 xRes,
					 Dgp_Int32 yRes,
					 Dgp_Int32 pgNum, 
                     Dgp_Int32 fgQual,
                     Dgp_Int32 fgScale,
					 Dgp_Bool report,
                     DgpAdt_List fgImages)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	tdoc_state *llCookie = (tdoc_state *)_llCookie;
	Dgp_State *dgp_state = cookie->dps;
	DgpAdt_ListNode ln, retnode;
	dgplssy_cc *box;
	DGPERR_INIT(tcly_doc_addColorPage);
	TCLYPageInfo *curPage = &(cookie->curPage);
	TCLYSectionBlock *sb = cookie->curBlock;
	DgpTdc_PageInfo *curLlPage = &(llCookie->curPage);
	DgpTdc_SectionBlock *llSb = llCookie->curBlock;
    Dgp_Int32 oldCount, newCount;

	report = report;
        bpl = bpl;	

    tcly_doc_pageInit(curPage);
	curPage->pgNum = pgNum;
	curPage->xRes = xRes;
	curPage->yRes = yRes;
	curPage->width = width;
	curPage->height	= height;
	DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&curPage->outliers, dgp_state));
	curPage->color = DGP_TRUE;
	curPage->numBoxes = numccs;

    DgpTdc_pageInit(curLlPage);
	curLlPage->pageNum = pgNum;
	curLlPage->width = width;
	curLlPage->height = height;
	curLlPage->xRes = xRes;
	curLlPage->yRes = yRes;

	/* Since the cc's have already been found, getColorBoxes is pretty simple. */
	/* It just turns the array of cc's into a list, copying bitmaps. */
	/* Since there's no callback (as in the black and white case), we'll do */
	/* the size check for the lossless tokenizer here. */
	/* */
	/* The concept of an outlier doesn't seem to exist in the color case. */
	DGP_ERR_PASS_CLEANUP(getColorBoxes(	dgp_state,
										text_ccs,
										numccs,
										pgNum,
										&curPage->boxes));
    if(fgImages != DGP_NULLLIST)
        getColorFg(dgp_state, 
                   curPage->boxes, width, height, xRes, yRes, fgQual, fgScale, fgImages);

	/* Force resolution */
	if (cookie->forceResolution!= -1) {
		curPage->xRes = cookie->forceResolution;
		curPage->yRes = cookie->forceResolution;
	}
	DGP_ASSERT(curPage->numBoxes == (Dgp_Int32)DGPADT_LILEN(curPage->boxes));


	for (ln = DGPADT_LIFIRST(curPage->boxes); ln != DGP_NULLLISTNODE; ) {
		box = (dgplssy_cc *)DGPADT_LIGET(ln);
		DGP_ASSERT(box != (dgplssy_cc *)NULL);
		DGP_ASSERT(box->bitmap.bits != (Dgp_UInt32 *)NULL);
		if(box->bitmap.width < cookie->tokOpt->lowerDilationLimit && 
			box->bitmap.height < cookie->tokOpt->lowerDilationLimit){
			/* We'd feed to lossless tokenizer, if we had runs... */
			/*DGP_ERR_PASS_CLEANUP(DgpCC_tokenAddCC(dgp_state, llCookie->llTokState, llSb->tBlock, &rTok, &(newpos->tok))); */
			/* Try adding as outlier... */
			box->tok = NULL;

			/* adjust coordinates */
			box->y += box->bitmap.height - 1;

			DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(curPage->outliers, (void *)box, &retnode));
            if(cookie->memBond >0)
                curPage->pageTokMem += box->bitmap.height * box->bitmap.longsperline * sizeof(Dgp_Int32) + TOKOVERHEAD;			
            /*DGP_ERR_PASS_CLEANUP(tokAddBitmap(cookie->tokstate, &box->bitmap)); */
		} else {
			/* Feed to lossy tokenizer... */
            if(cookie->memBond > 0){
               tokGetCount(cookie->tokstate, &oldCount);
             }
	            DGP_ERR_PASS_CLEANUP(tokAddBitmap(cookie->tokstate, &box->bitmap));
                if(cookie->memBond > 0){
                    tokGetCount(cookie->tokstate, &newCount);
                    if(newCount > oldCount){ /* get the cur tok memory */
                        curPage->tokMem += box->bitmap.height * box->bitmap.longsperline * sizeof(Dgp_Int32) + TOKOVERHEAD;
                    }
                }
		}
        if(cookie->memBond > 0){
            if(sb->tokMem + curPage->tokMem + curPage->pageTokMem +
                llSb->tokMem + curLlPage->tokMem + curLlPage->pageTokMem > cookie->memBond){
                break;
            }
        }
		ln = DGPADT_LINEXT(ln);
	}
	
	/* Make room for token positions... */
	DGP_CHECK_ALLOC_CLEANUP(curLlPage->boxes = (tokpos *)(DGP_MALLOC)(curPage->numBoxes * sizeof(tokpos)));

	/* Residual tokens should include outliers... */
	curPage->numOutliers = DGPADT_LILEN(curPage->outliers);
	curPage->residualTokenCount += curPage->numOutliers;

     /* Add the boxes onto the list we're building */
     addPage(cookie->curPage, sb);
     addLlPage(llCookie->curPage, llSb);

     /* update tokMem count */
    if(cookie->memBond > 0){
        if(sb->tokMem + curPage->tokMem + curPage->pageTokMem +
            llSb->tokMem + curLlPage->tokMem + curLlPage->pageTokMem < cookie->memBond){
            sb->tokMem += curPage->tokMem;
            llSb->tokMem += curLlPage->tokMem;
            sb->overflow = DGP_FALSE;
        }else{
              sb->overflow = DGP_TRUE;
              DGP_ERR_GEN_CLEANUPF0(sb->numPages == 1,cErrUndefined, 
                "Dicionary Memory limit is too small"); 


        }
    }else  sb->overflow = DGP_FALSE;
	NORMAL_RETURN;

ERR_CLEANUP:
	ERR_RETURN; 
}

#endif

/* Two small accessor methods for internal tclydoc data. */
/* Retrieves the box list for the current page in the cookie. */
void
tcly_doc_retrieveBoxList(TCLYDocCookie _cookie,
                         DgpAdt_List *list)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	TCLYPageInfo *curPage = &(cookie->curPage);
    
    *list = curPage->boxes;
    return;
}

/* Retrieves the box count for the current page in the cookie. */
void
tcly_doc_retrieveBoxCount(TCLYDocCookie _cookie,
                          Dgp_Int32 *count)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	TCLYPageInfo *curPage = &(cookie->curPage);
    
    *count = curPage->numBoxes;
    return;
}

/* Accessor for libtok-lossy functions... */
ErrCode
tcly_doc_tokAddBitmap(TCLYDocCookie _cookie,
                            tokBitmap *bitmap)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	Dgp_State *dgp_state = cookie->dps;

	DGPERR_INIT(tcly_doc_tokAddBitmap);

    DGP_ERR_PASS(tokAddBitmap(cookie->tokstate, bitmap));

    NORMAL_RETURN;
}


ErrCode
tcly_doc_retrieveClassList( TCLYDocCookie _cookie,
                            TCLYSectionBlockPtr sbPtr,
                            DgpAdt_List *tokOutput)
{
  TCLYDocState *cookie = (TCLYDocState *)_cookie;
  Dgp_State *dgp_state = cookie->dps;
  TCLYSectionBlock *sb = (TCLYSectionBlock *)sbPtr;
  DGPERR_INIT(tcly_doc_retrieveClassList);
  DGP_STOP_WHINE;

  if(tokOutput != NULL){
    *tokOutput = sb->classesFound;
  }

  NORMAL_RETURN;
}    

/*******************************************************************
* Called by tcly_doc_endSection
******************************************************************/
DGP_LIBEXPORT ErrCode
tcly_doc_cleanSection(TCLYDocCookie _cookie, 
					  Dgp_Bool		report)
{
    TCLYDocState *cookie = (TCLYDocState *)_cookie;
	Dgp_State *dgp_state = cookie->dps;
	DGP_PROLOGUE("tcly_doc_cleanSection()",DGP_PERRSTATE);
	TokState tokstate = cookie->tokstate;
	TCLYSectionBlock *sb = cookie->curBlock;
	DgpAdt_ListNode ln;
	token *toks = NULL;
	tokClass *c;
	Dgp_Int32 llNonSingletonCount;
	Dgp_Int32 i;

	report = report;

	if(sb->final){

        if(sb->overflow == DGP_TRUE){ 
            /* This section has rolledback a page. We need to resume the token */
            /* status and then final it so it won't be included in the next section. */
	        /*
	        * Now let the tokenizer continue.
	        * Some of the classes weren't put into the file dictionary -
	        * we want to get rid of those.  How to tell which?  Those are
	        * the ones whose userdata field is still NULL: if it was
	        * assigned a token number, it gets a non-NULL field.
	        */
            for (ln = DGPADT_LIFIRST(sb->classesFound); ln != DGP_NULLLISTNODE; ) {
                c = (tokClass *)DGPADT_LIGET(ln);
                if (c->userdata == (void *)NULL) {
                    DGP_ASSERT( DGPADT_LILEN(c->members) < 2);
                    tokFreeClass(tokstate, c);
                    DgpAdt_liRemAndNext(ln, &ln);
                }
                else {
                    ln = DGPADT_LINEXT(ln);
                }
            }
            /* Pass what's left back to the tokeniser */
            DGP_ERR_PASS_CLEANUP(tokResume(tokstate, sb->classesFound));
            DGP_ERR_PASS_CLEANUP(tokFinal(tokstate, &(sb->classesFound)));
        }
		/* This clause applies if we just cleared a section. */
		/* Free the output list */
		tokFreeOutput(tokstate, sb->classesFound, DGP_TRUE);
		tokFreeState(tokstate);
		/* Make a new tokenizer state. */
		DGP_ERR_PASS_CLEANUP(tokInit(cookie->tokOpt, &cookie->tokstate));
		tokstate = cookie->tokstate;
		
		/* Why carry around tokens we'll never use again?  Added, dhtesq. */
		ln = DGPADT_LIFIRST(cookie->tokDgpAdt_List);
		while(ln != DGP_NULLLISTNODE){
			toks = (token *)DGPADT_LIGET(ln);
			DGP_ASSERT(toks);
			DgpAdt_liRemAndNext(ln, &ln);
			llNonSingletonCount = (Dgp_Int32)DGPADT_LIGET(ln);
			DgpAdt_liRemAndNext(ln, &ln);

			/* Free lossless bitmaps in here, which the lossy tokenizer does *not* own. */
			for(i = 0; i < llNonSingletonCount; i++){
				DGP_FREE(toks[i].bits);
			}
			DGP_FREE( (void *)toks);
		}
        sb->tokMem = 0;
    } else {
	/*
	* Now let the tokenizer continue.
	* Some of the classes weren't put into the file dictionary -
	* we want to get rid of those.  How to tell which?  Those are
	* the ones whose userdata field is still NULL: if it was
	* assigned a token number, it gets a non-NULL field.
	*/
        for (ln = DGPADT_LIFIRST(sb->classesFound); ln != DGP_NULLLISTNODE; ) {
            c = (tokClass *)DGPADT_LIGET(ln);
            if (c->userdata == (void *)NULL) {
                DGP_ASSERT( DGPADT_LILEN( c->members) < 2);
                tokFreeClass(tokstate, c);
                DgpAdt_liRemAndNext(ln, &ln);
            }
            else {
                ln = DGPADT_LINEXT(ln);
            }
        }
		        
        /* Pass what's left back to the tokeniser */
        DGP_ERR_PASS_CLEANUP(tokResume(tokstate, sb->classesFound));

	}
	sb->classesFound = DGP_NULLLIST;
	/*
	* The pages array will stay allocated, but 
	* the arrays in each page are freed
	*/
	for(i = 0; i < sb->numPages; i++){
		/* Get rid of the current box list - it's been written. Yay! */
		tcly_doc_freePage(&sb->pages[i],DGP_FALSE,	dgp_state);
	}
	sb->numPages = 0;
    sb->llNonSingletonCount = 0;

	DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&sb->classesFound, dgp_state));


    NORMAL_RETURN;
	
ERR_CLEANUP:
	
	ERR_RETURN;
}


/*******************************************************************
* End of a section
******************************************************************/
DGP_LIBEXPORT ErrCode
tcly_doc_endSection(	TCLYDocCookie _cookie, 
						TCLYSectionBlockPtr *retSb,
						Dgp_Bool final, 
						Dgp_Bool report)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	Dgp_State *dgp_state = cookie->dps;
	/*DGPERR_INIT(tcly_doc_endSection); */
	DGP_STATE_PROLOGUE("tcly_doc_endSection()", dgp_state);
	
	TCLYSectionBlock *sb = cookie->curBlock;
	
	/* remove the classes previously found */
	DgpAdt_liFree(sb->classesFound);
	sb->classesFound = NULL;

	sb->final = final;
	if(sb->final){		
		DGP_ERR_PASS_CLEANUP(tokFinal(cookie->tokstate, &(sb->classesFound))); 
	} else {
		DGP_ERR_PASS_CLEANUP(tokPartial(cookie->tokstate, &(sb->classesFound)));
	}

	DGP_REPORTF0(report, "Section closed.");
	
	*retSb = (TCLYSectionBlockPtr) sb;
	
	NORMAL_RETURN;
	
ERR_CLEANUP:
	
	ERR_RETURN;
}
/*******************************************************************
* End of a section
******************************************************************/
DGP_LIBEXPORT ErrCode
tcly_doc_endSectionRollBack
(	TCLYDocCookie _cookie, 
    TCLYSectionBlockPtr *retSb,
	Dgp_Bool report)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	Dgp_State *dgp_state = cookie->dps;
	/*DGPERR_INIT(tcly_doc_endSection); */
    Dgp_Bool isNewClass;
	TCLYSectionBlock *sb = cookie->curBlock;
    TCLYPageInfo curPage = cookie->curPage;
    DgpAdt_ListNode nextClass, nextMember;
    tokClass *c = NULL; 
    tokOffset *offset;
    dgplssy_cc *cc = NULL;
    Dgp_Int32 page = 0;

    DGP_STATE_PROLOGUE("tcly_doc_endSectionRollBack()", dgp_state);
    DGP_CHECK_NULL(sb);
    DGP_CHECK_NULL(cookie->tokstate);

    page = curPage.pgNum;

	/* get a list of classes */
    DGP_ERR_PASS_CLEANUP(tokPartialNoMerge(cookie->tokstate, &(sb->classesFound)));

    /* go through the list of current classes and strip out the classes */
    /* that has members on the current page */
    nextClass = DGPADT_LIFIRST(sb->classesFound);
    while(nextClass != NULL){
        isNewClass = DGP_TRUE;
        c = (tokClass *)DGPADT_LIGET(nextClass);
		nextMember = DGPADT_LIFIRST(c->members);
		while(nextMember != DGP_NULLLISTNODE) {
            offset = (tokOffset *)DGPADT_LIGET(nextMember);
            cc = (dgplssy_cc *)offset->userdata;
			if ( cc ->pgnum == page) {
                DGP_ERR_PASS_CLEANUP(DgpAdt_liRemAndNext(nextMember, &nextMember));
                if(cookie->tokOpt->flags & TOK_OPT_RETURNOFFSET)
                    DGP_FREE(offset);
            } else{
                isNewClass = DGP_FALSE;
			    nextMember = DGPADT_LINEXT(nextMember);
			}
		}/* end nextMember != NULL */
        if(isNewClass == DGP_TRUE){ 
            /* if this class is only relavent to this page */
            /* free this class */
		    DGP_ASSERT( DGPADT_LILEN( c->members) == 0);
            tokFreeClass(cookie->tokstate,c);
            DgpAdt_liRemAndNext(nextClass,&nextClass);
        }else 
            nextClass = DGPADT_LINEXT(nextClass);
    } /* end DGPADT_FOREACH class */
    

    sb->final = DGP_TRUE;
    sb->tokMem = 0;

    tcly_doc_freePage(&cookie->curPage, DGP_TRUE, dgp_state);
    /* make the previous page to be the current page */
    sb->numPages--;
    DGP_ASSERT(sb->numPages > 0);

    cookie->curPage = sb->pages[sb->numPages-1];
	*retSb = (TCLYSectionBlockPtr)sb;

    DGP_REPORTF0(report, "Section closed.");
	
	NORMAL_RETURN;
	
ERR_CLEANUP:
	ERR_RETURN;
}


/************************************************************
* Finish the document:
* free the tokstate if 
* 
* if endSection was not called with final=true 
* return false, 
* Or perhaps this should return a 
* "portable" document block that contains output from 
* all unfinalized tokstates
************************************************************/
DGP_LIBEXPORT ErrCode
tcly_doc_end(TCLYDocCookie			_cookie, 
			 TCLYSectionBlockPtr	_sb)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	/*TokState tokstate = cookie->tokstate; */
	TCLYSectionBlock *sb = (TCLYSectionBlock*) _sb;
	Dgp_State *dgp_state = cookie->dps;
	DeskewStruct *deskew_cookie = (DeskewStruct*)cookie->deskew_cookie;

	Dgp_Int32 i, llNonSingletonCount;
	token * toks;
	/*tokClass* classptr; */
	DgpAdt_ListNode ln;
	/*DgpAdt_ListNode ln, ln2, nextln, nextln2, tempnode, retnode; */
	
	DGP_PROLOGUE("tcly_doc_end()",DGP_PERRSTATE);
	
	DGP_STOP_WHINE;
	
	/* Now we can free any tokens from the global (or incremental)
   	* dictionaries. (We couldn't free them until the _all_ the
   	* position block were written to disk. */
	if(cookie->tokDgpAdt_List != DGP_NULLLIST){
		DGPADT_FOREACH(ln, cookie->tokDgpAdt_List) {
			toks = (token *)DGPADT_LIGET(ln);
			ln = DGPADT_LINEXT(ln);
			llNonSingletonCount = (Dgp_Int32)DGPADT_LIGET(ln);
			/* Free lossless bitmaps in here, which the lossy tokenizer does *not* own. */
			for(i = 0; i < llNonSingletonCount; i++){
				DGP_FREE(toks[i].bits);
			}
			DGP_FREE( (void *)toks);
		}
		DgpAdt_liFree(cookie->tokDgpAdt_List);
		cookie->tokDgpAdt_List = DGP_NULLLIST;
	}
	
	/* cleanup the deskew cookie */
#ifndef DGP_FAXMODE
	deskewCleanup(deskew_cookie);
#endif
	/* Free all the sectionblock stuff*/
	if(sb != NULL){
		if(sb->classesFound != DGP_NULLLIST){
			tokFreeOutput(	cookie->tokstate,
				sb->classesFound, 
				DGP_TRUE);
		}
/*		sb->classesFound = DGP_NULLLIST; */
	}
	/* And free the tokenizer state. */
	if(cookie->tokstate != NULL){
		tokFreeState(cookie->tokstate);
		cookie->tokstate = NULLTOKSTATE;
	}
	
	if(sb != NULL){
		for(i = 0; i < sb->numPages; i++){
			tcly_doc_freePage(	&sb->pages[i], DGP_FALSE, dgp_state );
		}
		DGP_FREE(sb->pages);
		sb->pages = NULL;
		DGP_FREE(sb);
	}
	
	DGP_FREE(cookie);
	
	NORMAL_RETURN;
} 




/*
* Free the box lists.  Remember that the ccs on "outliers"
* are also on "boxes".
*/
void
tcly_doc_freePage(
 TCLYPageInfo *page,
 Dgp_Bool freeBits,
 Dgp_State *dgp_state)
{
	DGP_VOID_PROLOGUE("freeBoxlist()", DGP_PERRSTATE);
	DgpAdt_ListNode ln;
	dgplssy_cc *box;
    PIXR *prTmpIn = NULL;

	/* The outliers are simple - they're just copies*/
    /*if(*outliers != DGP_NULLLIST){
	    DgpAdt_liFree(*outliers);
        *outliers = DGP_NULLLIST;
    }*/
    /* Throw out any bitmaps of outliers on this page. */
    if(page->outliers != DGP_NULLLIST){
      ln = DGPADT_LIFIRST(page->outliers);
      DGPADT_FOREACH (ln, page->outliers) {
        box = (dgplssy_cc *)DGPADT_LIGET( ln);
        DGP_VOID_ASSERT( box != (dgplssy_cc *)NULL);
  
        /* All outliers on this page have had their runs 
         * converted to bitmaps by this time. So just free 
         * the bitmaps. */
        if(box->bitmap.bits && freeBits){
            DGP_FREE( (void *)box->bitmap.bits);
            box->bitmap.bits = NULL;
        }else if (box->runs != (Dgp_Int16 *)NULL) {
			DGP_FREE((void *)box->runs);
            box->runs = NULL;
        }
      }  
      DgpAdt_liFree(page->outliers);
      page->outliers = DGP_NULLLIST;
    }

	/*
	* The box list contains a bunch of ccs.  We don't own their
	* bitmaps - see below.
	*/
    if(page->boxes != DGP_NULLLIST){
	    DGPADT_FOREACH (ln, page->boxes) {
		    box = (dgplssy_cc *)DGPADT_LIGET(ln);
		    DGP_VOID_ASSERT(box != (dgplssy_cc *)NULL);
		    
		    /*
		    * We don't own box->member.bits anymore - the tokeniser had it.
		    * Unless, that is, we'd turned on TOK_OPT_COPYBITMAPS.  In that
		    * case, writeClasses should have freed/NULLed it.  It might no
		    * longer be NULL because writePages messes around with
		    * them.  In either case, it's not our problem now.
		    */
		    CC_FREE(box);
	    }
	    DgpAdt_liFree(page->boxes);
        page->boxes = DGP_NULLLIST;
    }
    if(page->largeCCList != DGP_NULLLIST){
	    DGPADT_FOREACH (ln, page->largeCCList) {
		    box = (dgplssy_cc *)DGPADT_LIGET(ln);
		    DGP_VOID_ASSERT(box != (dgplssy_cc *)NULL);	  
            /* largeCC only have runs */
		    if (box->runs != (Dgp_Int16 *)NULL) {
			    DGP_FREE((void *)box->runs);
		    }
		    CC_FREE(box);
	    }
	    DgpAdt_liFree(page->largeCCList);
        page->largeCCList = DGP_NULLLIST;
    }
    if(page->halftones != DGP_NULLLIST) {
		DGPADT_FOREACH (ln, page->halftones) {
			prTmpIn = (PIXR *)DGPADT_LIGET(ln);
			w_destroyPixr(dgp_state->ipwstate, &prTmpIn);
        }
        DgpAdt_liFree(page->halftones);
        page->halftones = DGP_NULLLIST;
    }

    if(page->imageinfo!= DGP_NULLLIST) {
        void *tempPtr;
		DGPADT_FOREACH (ln, page->imageinfo) {
            tempPtr = (void *)DGPADT_LIGET(ln);
            if(tempPtr) 	DGP_FREE(tempPtr);
        }
        DgpAdt_liFree(page->imageinfo);
        page->imageinfo = DGP_NULLLIST;
    }
}

/* Set the b rightmost (LS) bits of a ulong. b can be 0..32 */
#define	rightbits(b)	(((b) < DGP_BITS_UINT32) ? ((1ul << (b)) - 1) : (Dgp_UInt32)(~0ul))

/*
* Take the bits from the runs and drop them into the image.
*/
static void
fillruns(DgpImage_Packed im, dgplssy_cc *boxcc)
{
	DGP_VOID_PROLOGUE("fillruns()", NULL);
	Dgp_Int16 left, right, lastright;
	Dgp_UInt32 *row;
	Dgp_Int32 rownum = 0;
	Dgp_Int32 l, r;
	Dgp_Int32 lb, rb;
	Dgp_Int16 *runs = boxcc->runs;
	Dgp_Int32 nruns = boxcc->nruns;
	
	left = runs[0];
	right = runs[1];
	runs += 2;
	while (nruns) {
		row = DGPIMG_GETROW(im, rownum);
		while (nruns) {
			/* Fill in the range (left..right, rownum) */
			
			/* What words get touched? */
			l = left / DGP_BITS_UINT32;
			lb = left % DGP_BITS_UINT32;
			r = right / DGP_BITS_UINT32;
			rb = right % DGP_BITS_UINT32;
			if (l == r) {
				DGP_VOID_ASSERT(lb <= rb);
				/* Set bits lb to rb inclusive in row[l] */
				row[l] |= rightbits(DGP_BITS_UINT32 - lb) -
					rightbits(DGP_BITS_UINT32 - rb - 1);
			}
			else {
				row[l] |= rightbits(DGP_BITS_UINT32 - lb);
				for (l++; l < r; l++) {
					row[l] = (Dgp_UInt32)~0ul;
				}
				row[r] |= ~rightbits(DGP_BITS_UINT32 - rb - 1);
			}
			
			nruns--;
			if (nruns) {
				lastright = right;
				left = runs[0];
				right = runs[1];
				runs += 2;
				
				if (lastright + 1 >= left) {
					break;
				}
			}
		}
		rownum++;
	}
}

ErrCode 
convertCC(
          void **membercc, 
          DgpImage_Packed *image,
          Dgp_Int32 *x,
          Dgp_Int32 *y,
          Dgp_State *dgp_state)
{
	DGP_PROLOGUE("tcly_doc_convertCC()",DGP_PERRSTATE);
    dgplssy_cc *newcc = (dgplssy_cc *)(*membercc);

	DGP_ASSERT(membercc != NULL);
	DGP_ASSERT(*membercc != (dgplssy_cc *)NULL);
	DGP_ASSERT(image != NULL);
    
    *x = newcc->x;
    *y = newcc->y;

    DGP_ERR_PASS_CLEANUP(DgpImage_new(dgp_state, DGPIMAGE_PACKED,
		    newcc->bitmap.width, newcc->bitmap.height, (void**) image));
	fillruns(*image, newcc);

    if (newcc->bitmap.bits != (Dgp_UInt32 *)NULL) {
      DGP_FREE((void *)newcc->bitmap.bits);
      newcc->bitmap.bits = NULL;
    }
    if (newcc->runs != (Dgp_Int16 *)NULL) {
        DGP_FREE((void *)newcc->runs);
    }
    DGPLSSY_CC_FREE(newcc);
    *membercc = NULL;
	
    NORMAL_RETURN;
	
ERR_CLEANUP:
    if(*image) DgpImage_free(*image);
    *image = NULL;
	ERR_RETURN;
}

void ccDataCleanup(Dgp_State* dgp_state,
              CCData** ccData)
{
    IPWState *pIPWState = DGP_PIPWSTATE;
    void *heapManager;
    
    if(!ccData) return;
    if(!*ccData) return;
    
    if((*ccData)->pCCState)
#ifndef PSG
        i_destroyCCState (pIPWState, (void **)&((*ccData)->pCCState));
#else
        i_destroyCCState(pIPWState, (void *)(*ccData)->pCCState);
	i_getCCStateHeapManager(pIPWState, (void *)(*ccData)->pCCState, &heapManager);
	DGP_FREE(heapManager);
	DGP_FREE((*ccData)->pCCState);
#endif
    DGP_FREE(*ccData);
    *ccData = NULL;

}

ErrCode	
ccDataInit(Dgp_State* dgp_state, 
           Dgp_Int32 pgNum, 
           Dgp_Int32 memBond,
           Dgp_Int32 xRes, Dgp_Int32 yRes,
           Dgp_Int32 minWidth,
           Dgp_Int32 minHeight,
           Dgp_Int32 sectionHeight,
           TokState  tokstate,
           Dgp_Bool  maxTokenSize,
           Dgp_Int32 pgWidth,
           Dgp_Int32 pgHeight,
           CCData ** _ccData)
{
    CCData *ccData = NULL;
    DgpAdt_List  output = DGP_NULLLIST;
    DgpAdt_List  largeCCs = DGP_NULLLIST;
    IPWState *pIPWState = DGP_PIPWSTATE;
    Dgp_Int32 stateSize, heapMgrSize;
    void* heapManager;
    
    DGPERR_INIT(ccDataCleanUp);

    DGP_CHECK_ALLOC_CLEANUP(ccData = DGP_MALLOC(sizeof(CCData)));

    ccData->boxes = NULL;
    ccData->outliers = NULL;
    ccData->tokMem = 0;
    ccData->pageTokMem = 0;
    ccData->count = 0;
    ccData->dgp_state  = dgp_state;
    ccData->numOutliers = 0;
    ccData->accuHeight = 0;
    ccData->pgnum   = pgNum;
    ccData->membond = memBond;
    ccData->xRes = xRes;
    ccData->yRes = yRes;
    /* Adding code to make small tokens into outliers... */
    ccData->minWidth = minWidth;
    ccData->minHeight = minHeight;
    ccData->sectionHeight = sectionHeight ;
    ccData->pCCState = NULL;
    ccData->tokstate = tokstate;
    
    /* Decide what the cutoff for conversion to bitmap is */
    if (maxTokenSize){
        ccData->maxWidth  = ccData->llCCData.maxWidth = pgWidth / 10;	/* 10% of page width */
        ccData->maxHeight = ccData->llCCData.maxHeight = pgHeight/ 10;	/* 10% of page height */
    }
    else{
        ccData->maxWidth  = pgWidth;
        ccData->maxHeight = pgHeight;
    }

    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&output, dgp_state));
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&largeCCs, dgp_state));
    ccData->boxes = output;
    ccData->outliers   = largeCCs;

#ifndef PSG/*IPCORE-PSG doesn't define i_createCCState */
    DGP_ERR_PASS_CLEANUP( i_createCCState ( pIPWState, 
                                            (void**)&(ccData->pCCState),
                                            i_compressAlloc, i_compressFree));
#else
    
    DGP_ERR_PASS_CLEANUP( i_getCCStateSize( pIPWState, &stateSize));
    DGP_CHECK_ALLOC_CLEANUP(ccData->pCCState = DGP_MALLOC(stateSize));
    DGP_ERR_PASS_CLEANUP( i_getHeapManagerSize( pIPWState, &heapMgrSize));
    DGP_CHECK_ALLOC_CLEANUP( heapManager = DGP_MALLOC(heapMgrSize));
    DGP_ERR_PASS_CLEANUP( i_initializeHeapManager( pIPWState, heapManager));
    DGP_ERR_PASS_CLEANUP( i_initializeCCState( pIPWState,
					       (void*)(ccData->pCCState),
					       heapManager));
					      
#endif

    *_ccData = ccData;
    ccData = NULL;

     NORMAL_RETURN;
	
ERR_CLEANUP:
     if(ccData)
         ccDataCleanup( dgp_state, &ccData);
     *_ccData = NULL;
	 ERR_RETURN; 
}

/***************************************************************
* Monochrome start page
***************************************************************/
DGP_LIBEXPORT ErrCode
tcly_startPage(     
               TCLYDocCookie _cookie, 
               DgpTdc_Cookie _llTdoc,
               Dgp_Int32 pgNum,
               Dgp_Int32 pgWidth,
               Dgp_Int32 pgHeight,
               Dgp_Int32 xRes,
               Dgp_Int32 yRes)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	tdoc_state *llTdoc = (tdoc_state *)_llTdoc;
	Dgp_State *dgp_state = cookie->dps;

	TCLYPageInfo *curPage = &(cookie->curPage);
	DgpTdc_PageInfo *curLlPage = &(llTdoc->curPage);
	DgpTdc_SectionBlock *llSb = llTdoc->curBlock;
    
    DGPERR_INIT(tcly_startPage);

    /*Initialize page and Llpage */
    tcly_doc_pageInit(curPage);
    curPage->pgNum  = pgNum;
    curPage->xRes   = xRes;
    curPage->yRes   = yRes;
    curPage->width  = pgWidth;
    curPage->height	= pgHeight;
    curPage->sectionHeight = 1;
    curPage->ccData = NULL;

    DgpTdc_pageInit(curLlPage);
    curLlPage->pageNum = pgNum;
    curLlPage->width = pgWidth;
    curLlPage->height = pgHeight;
    curLlPage->xRes = xRes;
    curLlPage->yRes = yRes;

    
    
    DGP_ERR_PASS_CLEANUP(ccDataInit(dgp_state, pgNum, 
                                               cookie->memBond,
                                               xRes, yRes,
                                               cookie->tokOpt->lowerDilationLimit,
                                               cookie->tokOpt->lowerDilationLimit,
                                               curPage->sectionHeight,
                                               cookie->tokstate,
                                               cookie->maxTokenSize,
                                               pgWidth,
                                               pgHeight,
                                               (CCData **)(&(curPage->ccData))));
    DGP_ERR_PASS_CLEANUP(DgpLlCCData_init(dgp_state, 
                                          &((CCData *)(curPage->ccData))->llCCData,
                                          0,
                                          (tdoc_state *)llTdoc->llTokState,
                                          NULL,
                                          0,
                                          0,
                                          DGP_NULLLIST,
                                          0,
                                          0,
                                          llSb->tBlock,
                                          llSb->pBlock,
                                          llSb->numPages,
                                          0,
                                          0));
    DGP_SEC_MALLOC(((CCData *)(curPage->ccData))->llCCData.boxes, (tokpos *), sizeof(tokpos) * INIT_SIZE_BOXES);
    ((CCData *)(curPage->ccData))->llCCData.num_boxes_alloced = INIT_SIZE_BOXES;

     NORMAL_RETURN;
	
ERR_CLEANUP:
    if(curPage->ccData)
        ccDataCleanup(dgp_state, (CCData **)(&(curPage->ccData)));
	ERR_RETURN; 
}

/***************************************************************
 * Monochrome encode buffer
 ***************************************************************/
DGP_LIBEXPORT ErrCode
tcly_encBufferInPage(          
		     TCLYDocCookie _cookie, 
                     Dgp_UInt8 *buf, 
                     Dgp_Int32 bufWidth,
                     Dgp_Int32 bufHeight,
					 Dgp_Int32 bpl,
                     Dgp_Int32 frameBits,
                     Dgp_Bool lastBuf,
                     Dgp_Bool report)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	Dgp_State *dgp_state = cookie->dps;
	DGPERR_INIT(tcly_doc_encBuffer);
	
	TCLYPageInfo *curPage = &(cookie->curPage);

    /* if deskewing is desired, attempt to find the angle for the page */
    /* if (cookie->deskew_cookie) {
        DGP_ERR_PASS_CLEANUP(getSkewAngle(cookie->deskew_cookie, buf, &(curPage->skewAngle)));
    }*/
    report = report;
    DGP_ERR_PASS_CLEANUP(getBoxesfromBuffer(	
        dgp_state,
        buf,
        bufWidth,
        bufHeight,
        bpl, 
        frameBits,
        lastBuf,
        cookie->sizeThreshold,
        cookie->sizeThreshold,
        1,
        curPage->ccData));

        NORMAL_RETURN;
	
ERR_CLEANUP:
	ERR_RETURN; 
}

DGP_LIBEXPORT ErrCode
tcly_endPage(          
             TCLYDocCookie _cookie, 
             DgpTdc_Cookie _llTdoc,
             Dgp_Bool newPage,
             Dgp_Bool noRegion)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	tdoc_state *llTdoc = (tdoc_state *)_llTdoc;
	Dgp_State *dgp_state = cookie->dps;
	DGPERR_INIT(tcly_endPage);

    TCLYPageInfo *curPage = &(cookie->curPage);
	DgpTdc_PageInfo *curLlPage = &(llTdoc->curPage);
	TCLYSectionBlock *sb = cookie->curBlock;
	DgpTdc_SectionBlock *llSb = llTdoc->curBlock;

    curPage->tokMem = ((CCData *)(curPage->ccData))->tokMem;
    curPage->pageTokMem =((CCData *)(curPage->ccData))->pageTokMem;
    curPage->outliers = ((CCData *)(curPage->ccData))->outliers;  
    curPage->numOutliers = (Dgp_Int32)DGPADT_LILEN(curPage->outliers);    
    curPage->numBoxes = ((CCData *)(curPage->ccData))->count;
    curPage->boxes = ((CCData *)(curPage->ccData))->boxes;
    curPage->sectionHeight = ((CCData *)(curPage->ccData))->sectionHeight;
    ((CCData *)(curPage->ccData))->boxes = NULL;

    curLlPage->tokMem = ((CCData *)(curPage->ccData))->llCCData.tokMem;
    curLlPage->pageTokMem = ((CCData *)(curPage->ccData))->llCCData.pageTokMem;
    curLlPage->boxes = ((CCData *)(curPage->ccData))->llCCData.boxes;
    curLlPage->numBoxes = ((CCData *)(curPage->ccData))->llCCData.num_boxes;
    
    /* Force resolution */
    if (cookie->forceResolution!= -1) {
        curPage->xRes = cookie->forceResolution;
        curPage->yRes = cookie->forceResolution;
    }
    DGP_ASSERT(curPage->numBoxes == (Dgp_Int32)DGPADT_LILEN(curPage->boxes));
    
    /* Residual tokens should include outliers... */
    curPage->residualTokenCount += curPage->numOutliers;
    
    /*add halftone to boxes so that it can be put into perpage dictionary */
    if ( noRegion && cookie->grayHalftone  != DGP_TRUE &&
            curPage->halftones && curPage->halftones->nelt > 0) {
        DgpAdt_ListNode ln, retnode, lnInfo;
        IPWState *ipwstate = DGP_PIPWSTATE;
        dgplssy_cc *newcc = (dgplssy_cc *)NULL;
        HALFTONEPIC *infoTmp;
        PIXR* prTmp;
        Dgp_Int32 w, h, bpl;
        Dgp_UInt8 *w_image;
        lnInfo = curPage->imageinfo->first;
		DGPADT_FOREACH (ln, curPage->halftones) {
			prTmp = (PIXR *)DGPADT_LIGET(ln);
            infoTmp = (HALFTONEPIC *) DGPADT_LIGET(lnInfo);
            DGP_ERR_PASS_CLEANUP( w_pixrGetWidth (ipwstate, prTmp, &w));
            DGP_ERR_PASS_CLEANUP( w_pixrGetHeight (ipwstate, prTmp, &h));
            DGP_ERR_PASS_CLEANUP( w_pixrGetBpl(ipwstate, prTmp, &bpl));
            w_pixrGetImage(ipwstate, prTmp, &w_image);
            DGPLSSY_CC_SEC_MALLOC(&newcc);
            newcc->x = infoTmp->x;
            /*use LL instread of UL */
            newcc->y = infoTmp->y + h -1;
            newcc->bitmap.width = (Dgp_UInt16)w;
            newcc->bitmap.height = (Dgp_UInt16)h;
            /*newcc->bitmap.longsperline = (w + 31) / 32; */
            newcc->bitmap.longsperline = bpl / 4;
            newcc->bitmap.userdata = newcc;
            newcc->tok = (token *)NULL;
            newcc->runs = (Dgp_Int16 *)NULL;
            newcc->nruns = 0;
            newcc->pgnum = curPage->pgNum;
            newcc->bitmap.bits = (Dgp_UInt32 *)w_image;
    
            DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(curPage->boxes, (void *)newcc, &retnode));
            if(((CCData *)(curPage->ccData))->membond >0){
                curPage->tokMem += bpl * newcc->bitmap.height + TOKOVERHEAD;
            }
            curPage->numBoxes++;
            lnInfo = lnInfo->next;
        }
        /* Residual tokens should include outliers... */
        curPage->residualTokenCount += curPage->halftones->nelt;

    }
    if(newPage){
        /* Add the boxes onto the list we're building */
        addPage(cookie->curPage, sb);
        addLlPage(llTdoc->curPage, llSb);
    }
    /*clean up ccData*/
    if((CCData **)(&(curPage->ccData))) 
        ccDataCleanup(dgp_state, (CCData **)(&(curPage->ccData)));

    /* Update tokMem count*/
    if(cookie->memBond >0){
        if(sb->tokMem + curPage->tokMem + curPage->pageTokMem +
            llSb->tokMem + curLlPage->tokMem + curLlPage->pageTokMem < cookie->memBond){
            sb->tokMem += curPage->tokMem;
            llSb->tokMem += curLlPage->tokMem;
            sb->overflow = DGP_FALSE;
        }else { 
            sb->overflow = DGP_TRUE;
            /*  this singal page has already exceeds the memory limit */
            DGP_ERR_GEN_CLEANUPF0(sb->numPages == 1 && sb->final == DGP_TRUE,cErrUndefined, 
                "Dicionary Memory limit is too small"); 
        }
    }else sb->overflow = DGP_FALSE;
        NORMAL_RETURN;
	
ERR_CLEANUP:
    if((CCData **)(&(curPage->ccData)))
        ccDataCleanup(dgp_state, (CCData **)(&(curPage->ccData)));
	ERR_RETURN; 

}
