/********************************************
 * tclydoc.h
 * $Header: /project/icons/cvsroot/digipaper/src/libtclydoc/tclydoc.h,v 1.22 2003/02/10 22:53:40 mkrocken Exp $
 *******************************************/
#ifndef __TCLYDOC_H
#define __TCLYDOC_H

#ifdef __cplusplus
extern "C"
{
#endif

#include "dgp_state.h"
#include "dgp_err_printers.h"
#include "dgp_utils.h"

#include <uds.h>
#include <tok-lossy.h>
#include <mmcompress.h>

/* include-libwjr */
#include <list.h>
#include <image.h>
#include <util.h>

#include "scancolor.h"

#include "token-compress.h"
/*include "token-compress-priv.h" */
#include "shrcc.pub"

#define DO_HALFTONE 0
#define TOKOVERHEAD 32          /* per token overhead */
#define TOKGENOVERHEAD 8192     /* 8k general overhead */
#define TOKHTOVERHEAD 1024      /* 1k halftone cell dicionary overhead ( based on max sub cell size 8) */
#define TOKARITHOVERHEAD 1024   /* arithmatic encoding overhead 2^template size */

/* 
 * Structure used through out lossy tokenize compression.
 * It has to be initialized through tcly_doc_init.
 * Detail of this structure is in tclypriv.h 
 */

typedef struct TCLYDocState *TCLYDocCookie; 
typedef struct TCLYSectionBlock *TCLYSectionBlockPtr; 
typedef struct PIXR *TCLYPIXRPtr;

/*
 * ---TDOC---
 * The token finding and token matching part of the interface
 *
 * All fuctions have a report arg.  If this is true some trace info
 * will be printed with the rptPrinter in the tc_cookie
 */


/*
 * Initialize TCLYDOCCookie
 * Setting color to  DGP_TRUE, lossy to DGP_FALSE allows the possible 
 * use of the qualities. Currently these params are not used.
 * lossy is DGP_TRUE by default
 */
DGP_LIBEXPORT ErrCode
tcly_doc_init(  Dgp_State *dgp_state,  
		TCLYDocCookie *_cookie, 
		Dgp_Bool lossy, 		/* DGP state cookie            	*/
		Dgp_Bool halftone,		/* flag to deal with halftones  */
		Dgp_Int32 sizeThreshold,  	/* ignore CCs below this size 	*/
		Dgp_Int32 forceResolution,	/* force to use this resoltion  */
		Dgp_Bool maxTokenSize, /* limit the size of tokens ? */
		tokOptions *tokOpt, 	/* options for tc compressor    */
		Dgp_Bool deskew, 	/* turn on or off deskew	*/
		Dgp_Int32 skewAccuracy,	/* parameters related for deskew*/
		Dgp_Int32 baselineOrient,
                Dgp_Int32 membond,
		Dgp_Bool report		/* flag to report debug info 	*/
		);


/*
 * Begin a section
 *   Sections are blocks of pages that have a common dictionary
 * 
 * Use null for prev the first time this is called.
 * If it is called more than once, the old SectionBlock should be passed in.
 * (Especially if the previous section had a partial dictionary)
 */

DGP_LIBEXPORT ErrCode
tcly_doc_startSection(TCLYDocCookie _cookie, 
                      TCLYSectionBlockPtr prev, /* first time to NULL*/
					  Dgp_Int32 docPageNum,
					  Dgp_Bool report);

/*
 * Add a page to the section
 * as usual image must be padded by 32 pixels on each side
 * this is the only function that needs to be called to add 
 * a monochrome page
 */
DGP_LIBEXPORT ErrCode
tcly_doc_addMonoPage( 	UInt8 *image,		/* image pointer	*/
						TCLYDocCookie cookie, 
						DgpTdc_Cookie llTdoc,
                        Dgp_Int32 width,		/* width of this page	*/
                        Dgp_Int32 height,		/* height of this page	*/
                        Dgp_Int32 bpl,		/* byte perl line	*/
                        Dgp_Int32 xRes,		/* horizontal resolution*/
                        Dgp_Int32 yRes,		/* vertical resolution 	*/ 
						Dgp_Int32 pageNumber, 	/* page numger		*/
                        Dgp_Bool noRegion,
						Dgp_Bool report);	/* flag to report debug */


DGP_LIBEXPORT ErrCode
tcly_doc_addColorTokens(sc_cc *text_ccs,
				     Dgp_Int32 numccs,
					 TCLYDocCookie _cookie, 
					 DgpTdc_Cookie _llCookie,
					 Dgp_Int32 width,
					 Dgp_Int32 height,
					 Dgp_Int32 bpl,
					 Dgp_Int32 xRes,
					 Dgp_Int32 yRes,
					 Dgp_Int32 pgNum, 
                     Dgp_Int32 fgQual,
                     Dgp_Int32 fgScale,
					 Dgp_Bool report,
                     DgpAdt_List fgImages);

/* Two small accessor methods for internal tclydoc data. */
/* Retrieves the box list for the current page in the cookie. */
void
tcly_doc_retrieveBoxList(TCLYDocCookie _cookie,
                         DgpAdt_List *list);
/* Retrieves the box count for the current page in the cookie. */
void
tcly_doc_retrieveBoxCount(TCLYDocCookie _cookie,
                          Dgp_Int32 *count);

/* Accessors for libtok-lossy functions... */
ErrCode
tcly_doc_tokAddBitmap(TCLYDocCookie _cookie,
                            tokBitmap *bitmap);
ErrCode
tcly_doc_tokAddBitmapVerify(TCLYDocCookie _cookie,
                            tokBitmap *bitmap,
                            Dgp_Bool *newClassFound,
                            DgpAdt_List *classMatched);

ErrCode
tcly_doc_retrieveClassList( TCLYDocCookie _cookie,
                            TCLYSectionBlockPtr sbPtr,
                            DgpAdt_List *tokOutput);

DGP_LIBEXPORT ErrCode
tcly_doc_addMonoPageVerify(UInt8 *image, 
					 TCLYDocCookie _cookie, 
					 DgpTdc_Cookie _llTdoc,
					 Dgp_Int32 width,
					 Dgp_Int32 height,
					 Dgp_Int32 bpl,
					 Dgp_Int32 xRes,
					 Dgp_Int32 yRes,
					 Dgp_Int32 pgNum, 
    Int32 (CDECL *rectReturnProc) (			/* This is the callback for the cc finder -- customizable. */
		    IPState    *pIPState,
		    void       *,
		    Int32	x,
		    Int32	y,
		    Int32	w,
		    Int32	h,
		    Int32	m,
		    Int32	nRuns,
		    TCScanLineRun *pRuns),
					 Dgp_Bool report,
                     Dgp_Int32 *newClassesFound,
                     Dgp_Int32 *numBoxes,
                     DgpAdt_List *boxList,
                     DgpAdt_List *xrefList,
                     Dgp_Int32 *numMatches,
                     DgpAdt_List *matchList);

/******************************************************************
* Called by tcly_doc_cleanSection to clean previous section block
*******************************************************************/
DGP_LIBEXPORT ErrCode 
tcly_doc_cleanSection(TCLYDocCookie cookie, Dgp_Bool report);

/*
 * Finish the section: this will run tokPartial or tokFinal
 * and put all the necessary information for the section in a 
 * "portable" block
 */
DGP_LIBEXPORT ErrCode
tcly_doc_endSection(	TCLYDocCookie cookie, 
			TCLYSectionBlockPtr *retSb, /* current section block */
			Dgp_Bool final, /* true --  final dict for this section*
					* false -- incremental dict ths sec   */
			Dgp_Bool report); /* flag to report debug */

/* Rollback one page and finish the section. Create a list of classes
 * to store all the classes that's on the last page
 */

DGP_LIBEXPORT ErrCode
tcly_doc_endSectionRollBack
(   TCLYDocCookie _cookie, 
    TCLYSectionBlockPtr *retSb,
    Dgp_Bool report);

/*
 *Finish the document:
 * 
 * Free all the TCLYDOCCookie 
 * if sb is not NULL it will also try
 * to free the section block state.
 */
DGP_LIBEXPORT ErrCode
tcly_doc_end(	TCLYDocCookie cookie, 
		TCLYSectionBlockPtr sb);

/*scanline interfaces */
/***************************************************************
* Monochrome start page
***************************************************************/
DGP_LIBEXPORT ErrCode
tcly_startPage(     
               TCLYDocCookie _cookie, 
               DgpTdc_Cookie _llTdoc,
               Dgp_Int32 pgNum,
               Dgp_Int32 pgWidth,
               Dgp_Int32 pgHeight,
               Dgp_Int32 xRes,
               Dgp_Int32 yRes);
/***************************************************************
 * Monochrome encode buffer
 ***************************************************************/
DGP_LIBEXPORT ErrCode
tcly_encBufferInPage(          
		     TCLYDocCookie _cookie, 
                     Dgp_UInt8 *buf, 
                     Dgp_Int32 bufWidth,
                     Dgp_Int32 bufHeight,
					 Dgp_Int32 bpl,
                     Dgp_Int32 frameBits,
                     Dgp_Bool lastBuf,
					 Dgp_Bool report);
DGP_LIBEXPORT ErrCode
tcly_endPage(          
             TCLYDocCookie _cookie, 
             DgpTdc_Cookie _llTdoc,
             Dgp_Bool newPage,
             Dgp_Bool noRegion);



#ifdef __cplusplus
}
#endif

#endif /*__TCLYDOC_H*/


