/***********************************************************
 * compress.c
 * $Header: /project/icons/cvsroot/digipaper/src/libtclydoc/tclyout.c,v 1.4 2003/01/27 21:48:40 awilliam Exp $
 **********************************************************/

#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif

#include <stdio.h>
#include "dgp_state.h"
#include "dgp_sys_err_control.h"
#include "mmcompress.h"
#include <stdlib.h>
#include "list.h"
#include <sys/types.h>

#ifndef _WIN32
/*#include <netinet/in.h>*/
#endif
#include <errno.h>
#include "g4.h"
#include <limits.h>

#include "compress.h"
#include "compressUtil.h"
#include "deskew.h"

#include "libutilw.pub"
#include "libpixrw.pub"
#include "librastw.pub"
#include "libiow.pub"
#include "PsgFix.h"


/******************************************************************
 * Get rid of all the dictionary state info from this cookie
 *******************************************************************/
void
emptyOutputState(TCLYOutCookie toc)
{
  OutputCallbacks *ocb = toc->callbacks;

  if (toc->dictState != (void *)NULL) {
    ocb->freeOutputDictblock(	toc->dictState, 
				toc->outputState,
				ocb->userData);
    toc->dictState = (void *)NULL;
  }
}

/********************************************************************
* Init TCLYOutCookie
*******************************************************************/
DGP_LIBEXPORT ErrCode
tcly_out_init(	
			  Dgp_State *dgp_state,
			  TCLYOutCookie *_cookie, 
			  Dgp_Int32 numPgs, 
			  Dgp_Bool pdfOut,
			  uds       *stream, 
			  Dgp_Int32 classThresh,
			  Dgp_Int32 writeClasses,
			  Dgp_Bool  g4_toks,
			  Dgp_Int32 jbig2Profile,
              Dgp_Bool  verbose,
			  OutputCallbacks *callbacks)
{
	DGPERR_INIT(tcly_out_init);
	TCLYOutCookie cookie;
	DGP_SEC_MALLOC(cookie, (TCLYOutCookie), sizeof(TCLYOutState));
	cookie->dgp_state	= dgp_state;
	cookie->numPgs 		= numPgs;
	cookie->writeClasses 	= writeClasses;
	cookie->g4_toks 	= g4_toks;
    cookie->verbose     = verbose;
	cookie->stream 		= *stream;
	cookie->classThresh 	= classThresh;
	cookie->callbacks 	= callbacks;
	cookie->outputState 	= (void *) NULL;
	cookie->dictState 	= (void *) NULL;
	
	DGP_ERR_PASS_CLEANUP(
		callbacks->createOutputState(	stream,
		dgp_state,
						jbig2Profile,
		numPgs,
		pdfOut,
		&(cookie->outputState),
		&callbacks->userData));
	*_cookie = cookie;
	
	NORMAL_RETURN;
	
ERR_CLEANUP:
	if(cookie) DGP_FREE(cookie);
	ERR_RETURN;
}


/***************************************************
* End of output.
* Free memory
*****************************************************/
DGP_LIBEXPORT ErrCode
tcly_out_end(	TCLYOutCookie toc)
{
	Dgp_State *dgp_state = toc->dgp_state;
	DGPERR_INIT(tcly_out_end);
	OutputCallbacks *cb =  toc->callbacks;
	if(toc == NULL) NORMAL_RETURN;
	DGP_ERR_PASS_CLEANUP( cb->closeOutputState( 	toc->outputState,
		cb->userData));
	
	if (toc->dictState != (void *)NULL) {
		cb->freeOutputDictblock( 	toc->dictState, 	
			toc->outputState,
			cb->userData);
	}
	/* free the cookie*/
	DGP_FREE(toc);
	
	NORMAL_RETURN;
	
ERR_CLEANUP:
	ERR_RETURN;
	
}

/***********************************************************************
 * Write out a dictionary for all the new tokens in the given output list.
 * - state: state
 * - tokOutput: the output from tokenising the boxes. Must have offsets,
 *	and the userdata of each original tokBitmap must have pointed to the
 *	cc that generated that tokBitmap.
 *	Also, any classes that are carried over from previous calls
 *	to tokPartial (and which were previously passed to
 *    	writeTokOutput) must have their userdata field pointing at
 *	an Dgp_Int32 containing the token number that they were assigned by
 *	writeTokOutput.  No modifications (apart from class tagging, below)
 *	are made to this list.
 *	The userdata field of each tokClass that was written into the
 *	dictionary (usually the non-singletons) will be set to point to
 *	an Dgp_Int32 containing that token's unique token number.
 ************************************************************************/

ErrCode
writeTokOutput(	TCLYOutCookie		toc, 
				TCLYSectionBlock	*sb,
				DgpTdc_SectionBlock *llSb,
				DeskewStruct		*deskewCookie,
				token				**_toks,
				Dgp_Bool			final)
{
  Dgp_State *dgp_state = toc->dgp_state;
  DGP_PROLOGUE( "writeTokOutput()", toc->DGP_PERRSTATE);
  token *toks = (token *)NULL;
  Dgp_Int32 numToks;
  Dgp_Int32 llNonSingletonCount;
  void *newdictstate;
  OutputCallbacks *ocb = toc->callbacks;
  /*
   * Massage the tokOutput list into a bunch of tokens.
   * Also create an array of ints just holding the token numbers,
   * and hook that onto toknums.
   */
  DGP_ERR_PASS_CLEANUPF0(makeTokens(	sb,
										llSb,
										deskewCookie,
										&toks, 
										&numToks,
										&llNonSingletonCount,
										final,
										dgp_state),
	"error while creating tokens.");

  /* if there are no tokens, don't write a dictionary 
   *(and don't change any of the current state)       
   */
  if (numToks + llNonSingletonCount == 0) {
    /* Pass back the tokens, in this case, NULL. */
    *_toks = (token *)NULL;
            
    NORMAL_RETURN;
  }

  /* Write out the dictionary */
  DGP_ERR_PASS_CLEANUPF0(ocb->writeOutputDictionary( 	toks, 
														numToks + llNonSingletonCount,
														toc->dictState,
														&newdictstate,
														(DictExtension *)NULL, 
														0,
														toc->outputState,
														ocb->userData),
	"I had trouble writing out a dictionary...");

  /* Free the old dictionary block */
  if (toc->dictState != (void *)NULL) {
    ocb->freeOutputDictblock(	toc->dictState, 
								toc->outputState,
								ocb->userData);
  }
  /* and give them back the new one */
  toc->dictState = newdictstate;
  
  sb->llNonSingletonCount = llNonSingletonCount;
  /* Pass back the tokens */
  *_toks = toks;
  toks = NULL;
  NORMAL_RETURN;

ERR_CLEANUP:
  if(toks != NULL) (toc->DGP_FREE)(toks);
  ERR_RETURN;
}

