/*************************************************************
 * tclyout.h
 * $Header: /project/icons/cvsroot/digipaper/src/libtclydoc/tclyout.h,v 1.16 2003/02/10 22:53:40 mkrocken Exp $
 ************************************************************/
#ifndef TCLY_OUT_H
#define TCLY_OUT_H


#include <stdio.h>
#include "iostructures.h"
#include "mmcompress.h"
#include "uds.h"
#include "dgp_allocators.h"
#include <color-extension.h>
#include "colorpos.h"
#include "annotation.h"
#include "tclydoc.h"
#include "tclypriv.h"

#ifdef __cplusplus
extern "C"
{
#endif

/*
 * Callback types
 */
typedef ErrCode (* CreateOutputState)(	uds *stream, 
					Dgp_State *dgp_state, 
					Dgp_Int32 jbig2Profile,
					Dgp_Int32 numPgs, 
					Dgp_Bool pdfOut,
	   				void **_outState,
					void *userData);
typedef ErrCode (* CloseOutputState) (	void *_outState,
					void *userData);
typedef ErrCode ( * FinishOutputPage)(	void *_outState,
					void *userData);
typedef ErrCode ( * WriteOutputAnnotations)(	Dgp_Int32 num_annotations,
						Annotation *annotations,
						void *_outState,
						void *userData);
typedef ErrCode (* WriteOutputPosBlock)(	tokpos *pos, 
						Dgp_UInt32 numPos, 
						Dgp_Int32 pgWidth, 
		  				Dgp_Int32 pgHeight, 
						Dgp_Int32 xRes, 
						Dgp_Int32 yRes, 
						PosExtension *exts,
		  				Dgp_Int32 numExts, 
						void *dictblock, 
						Dgp_Bool lossy, 
		  				void *_outState,
						void *userData);
typedef ErrCode (*WriteOutputDictionary)(	token *toks, 
						Dgp_Int32 ntoks, 
						void *oldDictblock, 
						void **newDictblock,
		    				DictExtension *exts, 
						Dgp_Int32 numExts, 
						void *_outState,
						void *userData);
typedef ErrCode (*WriteOutputPageDictionary)(	token *toks, 
						Dgp_Int32 ntoks, 
						Dgp_Int32 pgWidth,
						Dgp_Int32 pgHeight,
						Dgp_Int32 xRes,
						Dgp_Int32 yRes,
						Dgp_Bool ascii_ext,
						void *oldDictblock, 
						void **newDictblock,
		    				DictExtension *exts, 
						Dgp_Int32 numExts, 
						void *_outState,
						void *userData);
typedef ErrCode (*WriteOutputBinaryPixr)(	DgpAdt_List imagePos, 
		    				Dgp_Int32 layer, 
							Dgp_Int32 section,
						DgpAdt_List image, 
						void *outState,
						void *userData);
						
typedef void (*FreeOutputDictblock)(	void *_dictblock, 
					void *_outState,
					void *userData);

/*
 * Application should provide this structure to tcly_out_init
 */
typedef struct {
CreateOutputState       createOutputState;
CloseOutputState        closeOutputState;
FinishOutputPage        finishOutputPage;
WriteOutputAnnotations  writeOutputAnnotations;
WriteOutputPosBlock     writeOutputPosBlock;
WriteOutputDictionary   writeOutputDictionary;
WriteOutputPageDictionary   writeOutputPageDictionary;
WriteOutputBinaryPixr   writeOutputBinaryPixr;
FreeOutputDictblock     freeOutputDictblock;
void 			*userData;
}OutputCallbacks;



/* ---TOUT---
 * The outputting part of the interface
 *
 * If the  report arg is true some trace info
 * will be printed with the rptPrinter in the tc_cookie
 */

/* 
 * Struct to store output related information
 * Detailed defination is at tclypriv.h
 */
typedef struct {
        Dgp_State *dgp_state;    /* DGP State Cookie                     */
        void *outputState;      /* State cookie related different fomat */
        void *dictState;        /* Dictionary state cookie              */
        Dgp_Int32 numPgs;             /* Number of pages                      */
        Dgp_Int32 writeClasses;       /* Flag indicate to write class for debug*/
        Dgp_Int32 classThresh;        /* only write classes having at least this*
                                 * many member                          */
        Dgp_Bool g4_toks;            /*  True  -- Write "Single-use" tokens  *
                                 *  False -- Singletons and outliers    *
                                 *  be kept g4 encoded when decoded.    */
        uds stream;             /* User defined stream                  */
        TCLYPageInfo *curPage;	/* Page to be output		  */
		Dgp_Bool final;			/* if DGP_TRUE, then we can assume that none of the 
								   stuff is needed anymore. */
        Dgp_Bool verbose;
        OutputCallbacks *callbacks;/* Structure hold user defined callbacks*/
}TCLYOutState, *TCLYOutCookie;

/*
 * Prepare TCLYOutCookie for writting a section
 * The numDicts can be an estimate, but numPgs must be known
 * The stream is the where the output will be written.
 */
DGP_LIBEXPORT ErrCode
tcly_out_init(
		Dgp_State *dgp_state, 
		TCLYOutCookie *_cookie, 
		Dgp_Int32 numPgs, 
		Dgp_Bool pdfOut, 
		uds *stream, 
		Dgp_Int32 classThresh,
		Dgp_Int32 writeClasses,
       	Dgp_Bool g4_toks,
		Dgp_Int32 jbig2Profile,
        Dgp_Bool verbose,
		OutputCallbacks *callbacks);

/*
 * Write pos blocks and dictblocks to file
 * The SectionBlock will be altered a bit after this function
 * The resulting sb should be passed back to the TDOC for correct cleanup
 */
#if 0
DGP_LIBEXPORT ErrCode
tcly_out_writeSection(	TCLYOutCookie toc, 
			TCLYDocCookie tdc, 
			Dgp_Bool report);
#endif /* 0 */

/*
 * Close the file and free any state
 */
DGP_LIBEXPORT ErrCode
tcly_out_end(TCLYOutCookie cookie);


#ifdef __cplusplus
}
#endif

#endif




/*******************************************************************
 * Running Log
 * $Log: tclyout.h,v $
 * Revision 1.16  2003/02/10 22:53:40  mkrocken
 * Replaced C++ style comments with C style comments (OSG request)
 *
 * Revision 1.15  2002/05/29 17:37:37  mkrocken
 * add JBIG2 in PDF as a command line option
 *
 *  Modified Files:
 *  	apiinit.c jtokapi.h tclyout.c tclyout.h
 *
 * Revision 1.14  2001/12/19 00:53:17  mkrocken
 * changes for writing 2--stripe JBIG2
 *
 *  Modified Files:
 *  	apicalls.c tclydoc.c tclyout.h tclypriv.h tclyverify.c
 *  	tiff_cc2box.c tiff_cc2box.h
 *
 * Revision 1.13  2001/06/19 22:35:22  mkrocken
 * changed format of generic region encoding callback
 *
 * Revision 1.12  2001/05/29 22:22:53  mkrocken
 * changes to make jbig2 profile a command line option
 *
 * Revision 1.11  2001/05/08 00:42:55  dtracy
 * allocators.h -> dgp_allocators.h throughout code base.
 *
 * Revision 1.10  2001/04/26 22:23:45  dtracy
 * Extensive modifications for unification with lossless tokenizer.
 *
 * Revision 1.9  2001/04/10 23:46:28  dtracy
 * Cleared xiff3 dependencies from codebase. Jpeg-ls tweaks.
 *
 * Revision 1.8  2001/03/13 22:49:26  mkrocken
 * need to access the color tag when writing pages
 *
 * Revision 1.7  2001/02/09 23:46:01  jlin
 * add halftone detection and convert to gray
 *
 * Revision 1.6  2000/10/11  18:47:13  dtracy
 * Fixed bug:  decoder was writing 8bpp b&w tiffs.
 *
 * Revision 1.5  2000/04/13 23:49:36  mkrocken
 * few changes to get jbig2 to clear out dictionaries
 * before this change all jbig2 segments were refering to the
 * first global dictionary, even after the tokenizer was reset.
 *
 * Revision 1.4  2000/03/24 22:23:23  dtracy
 * VxWorks-friendly modifications.
 *
 * Revision 1.3  2000/03/04 00:22:40  mkrocken
 *
 * merged repositories
 *
 * Revision 1.2  2000/02/23 01:01:21  mkrocken
 *
 * changes for new util libs
 *
 * Revision 1.1  2000/02/17 22:23:19  mkrocken
 * Initial Revision
 *
 * Revision 1.2  2000/02/05 00:37:37  jlin
 * Add writeOutputPageDictionaryCB and move newOutputPageCB into it so that user has the flexibility of puting per page dictionary inside of a page or not.
 *
 * Revision 1.1  1999/12/29  19:49:31  jlin
 * Initial version of tclyout.h for libtclydoc
 *
 *******************************************************************/
