/**********************************************
 * tclypriv.h
 * $Header: /project/icons/cvsroot/digipaper/src/libtclydoc/tclypriv.h,v 1.17 2003/02/10 22:53:40 mkrocken Exp $
 **********************************************/
#ifndef __TCLYPRIV_H__
#define __TCLYPRIV_H__

#include "dgp_state.h"

#include <uds.h>
#include <tok-lossy.h>
#include <mmcompress.h>

/* include-libwjr */
#include <list.h>
#include <image.h>
#include <util.h>


#ifdef __cplusplus
extern "C"
{
#endif

#define DO_HALFTONE 0
typedef struct DeskewStruct *TCLYDeskewCookie;

#define getCurPage(tdoc) (tdc->curPage)

typedef struct {
	Dgp_Int32 	pgNum;		/* Page number for current Page 	*/
	Dgp_Int32 	index;    

	Dgp_Bool    color;      /* is this page deeper than 1 bit/pixel ? */
	DgpAdt_List boxes;		/* list of connected component found 	*/
	Dgp_Int32 	numBoxes;	/* number of connnect compnt found 	*/
	DgpAdt_List outliers;	/* list of components found as runs	*/
	Dgp_Int32 	numOutliers;/* number of outliers			*/
    DgpAdt_List largeCCList;/* large cc list that will be put into generic region*/
	Dgp_Int32 	width; 		/* page width				*/
	Dgp_Int32 	height;		/* page height				*/
	Dgp_Int32	sectionHeight; /* height of first section */
	Dgp_Int32 	xRes; 		/* page x resolution			*/
	Dgp_Int32 	yRes;		/* page y resolution			*/
	Float32		skewAngle;	/* skew angle of a page			*/
	Dgp_Int32 	skewedTokenPerPage; /*num of token classes deskewed     */
	Dgp_Int32   tokMem;		/* token memory  related to global dictionary*/
	Dgp_Int32   pageTokMem;		/* token memory related perpage dictionary */
	Dgp_Int32	residualTokenCount; /* Number of tokens on this page NOT going into the section dictionary. / Added dhtesq. */
	DgpAdt_List halftones;	/* halftone image  found		*/
    DgpAdt_List imageinfo;
    void        *ccData;
} TCLYPageInfo;


typedef struct {
	Dgp_UInt16 numMulti;  /* number of cc that are in classes with more than 1 member */
	Dgp_UInt32 numTiny;   /* number of cc that are smaller than 10 pix */
	Dgp_UInt16 numSingleton;  /* number of cc that are unique */
	Dgp_UInt16 numClasses;    /* number of tokens that are not Singletons and not Tiny */
} TCLYTokenStats;


typedef struct {
	Dgp_Bool final;		/* flag for invode tokFinal or tokPartial */
				/* decided by if number of pages for write*
				 * out incremental dictionary or for clear*
				 * dictionary is reached		  */
	Dgp_Int32 numPages;		/* number of pages for a section	  */
	Dgp_Int32 numPagesAlloced;	/* keep track out how many pages allocated*/
	TCLYPageInfo *pages;	/* pages within a section block		  */
	TCLYTokenStats stats;   /* some statistics about the types of tokens */
	DgpAdt_List classesFound;	/* tokens return from tokPartial or tokFinal*/

	Dgp_Int32 llNonSingletonCount;	/* Since lossy and lossless tokens are stored in one array, keep this */
									/* counter so you know where the lossless ones end and the lossy begin. */
	Dgp_Int32 sectionStartedAtThisPage; /* Zero-indexed.  This is the page of the document where the current section starts. */

        Dgp_Int32 tokMem;               /* token memory within this section */
        Dgp_Bool  overflow;               /* token memory within this section */
	/* 17.4.2001 / dhtesq. */
	/* Dragged these in from the corresponding lossless tokenizer struct, less */
	/* duplicate fields. */
#if 0
	/* filled by tdoc */
	DgpCC_PosInfoBlock *pBlock;

	/* Dictionary stuff */
	DgpCC_TokInfoBlock *tBlock;  /*Block containing the TokInfos pointed to by the positions */
	Dgp_Int32 numOldLlToks;		 /* Number of TokInfos in above block that are from previously written sections */
	DgpAdt_List llClassesFound;  /* list of tokClass structs */

	/* Even elements are token arrays of previously written tokens.  Odd elements are the size of
	those arrays.
	*/
	DgpAdt_List llTokBlockList;  
    DgpAdt_List newClassList;
	/* The Images,  */
	Dgp_Int32 numSampIms;
	samptokClass *sampOutput;
	Dgp_Int32 sampCompType;
	void *sampCompInfo;
  
	/* Compress struct for jpeg compression. */
	void *cinfo;

	/* Halftone info */
	DgpAdt_List halfOutput;
#endif /* 0 */
	/*/////////////////////// */
} TCLYSectionBlock;



typedef struct {
	Dgp_State *dps;		/* DGP state cookie 			*/
	tokOptions *tokOpt;	/* options for tc compressor		*/
	TokState tokstate;	/* token state for tc compressor	*/
	Dgp_Bool lossy;		/* flag for lossy compressor (DGP_TRUE)	*/
	Dgp_Bool grayHalftone;	/* convert halftone into gray images		*/
    Dgp_Bool maxTokenSize; /* flag to determine wether or not to limit the size of a token */
	Dgp_Int32 sizeThreshold;	/* CCs below this size are ignored	*/
    Dgp_Int32 forceResolution;	/* force to use this resoltion instead	*/
	Dgp_Bool deskew;		/* turn on or off deskew		*/
	TCLYPageInfo curPage;	/* current page				*/
	TCLYSectionBlock *curBlock;/* section block			*/
	TCLYDeskewCookie deskew_cookie;/* deskew state cookie 		*/
	DgpAdt_List tokDgpAdt_List;		/* tokens write out to dictionary during
	    		                     the whole document			*/
    Dgp_Int32 memBond;
} TCLYDocState;

ErrCode 
convertCC(        
            void **membercc, 
            DgpImage_Packed *image,
            Dgp_Int32 *x,
            Dgp_Int32 *y,
            Dgp_State *dgp_state);
void
tcly_doc_freePage(TCLYPageInfo *page, Dgp_Bool freeBits,  Dgp_State *dgp_state);
#ifdef __cplusplus
}
#endif

#endif /*__TCLYPRIV_H__*/


