#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif

#include "tclydoc.h"
#include "tok-lossy.h"
#include "tclyout.h"
#include "tclypriv.h"
#include "uds.h"

#include "compress.h"

#include "dgp_err_printers.h"
#include "dgp_sys_err_control.h"
#include <errno.h>
#include "dgp_utils.h"
#include "math.h"
#include "image_defs.h"
#include "PsgFix.h"

/* include-libwjr*/
#include <util.h>
#include <insertbox.h>
#include <simple_utils.h>

#include "blit.h"
#include "tokclass.h"

#include "token-compress.h"
#include "token-compress-priv.h"

/* Set the b rightmost (LS) bits of a ulong. b can be 0..32 */
#define	rightbits(b)	(((b) < DGP_BITS_UINT32) ? ((1ul << (b)) - 1) : (Dgp_UInt32)(~0ul))
#define	PAD	5

static void
blat(DgpImage_Packed im, Dgp_Int32 x, Dgp_Int32 y, tokBitmap *bitmap);


/* Blat the bitmap into the image with UL corner at (x,y) */
static void
blat(DgpImage_Packed im, Dgp_Int32 x, Dgp_Int32 y, tokBitmap *bitmap)
{
	Dgp_Int32 l, i, j;
	Dgp_Int32 row_width;		/* width of each row in Dgp_UInt32s */
	Dgp_UInt32 *iptr, *mptr, carry;
	Dgp_Int32 rshift, lshift;
	
	l = x / DGP_BITS_UINT32;	/* which long to start at */
	row_width = (bitmap->width + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
	
	rshift = (x % DGP_BITS_UINT32);
	lshift = DGP_BITS_UINT32 - rshift;
	
	if (rshift) {
		for (j = y;
		j < MIN(y + bitmap->height, DGPIMG_YMAX(im)+1); j++) {
			iptr = DGPIMG_GETROW(im, j) + l;
			mptr = &(bitmap->bits[(j - y) * bitmap->longsperline]);
			
			carry = 0;
			for (i = 0; i < row_width; i++) {
				*iptr |= ((*mptr >> rshift) | carry);
				carry = *mptr << lshift;
				
				mptr++;
				iptr++;
			}
			
			if (bitmap->width + rshift - row_width * DGP_BITS_UINT32 > 0) {
				/* Still some bits in carry */
				*iptr |= carry;
			}
		}
	}
	
	else {
		/* No shifting! */
		for (j = y;
		j < MIN(y + bitmap->height, DGPIMG_YMAX(im)+1); j++) {
			iptr = DGPIMG_GETROW(im, j) + l;
			mptr = &(bitmap->bits[(j - y) * bitmap->longsperline]);
			
			for (i = 0; i < row_width; i++)
				*(iptr++) |= *(mptr++);
		}
	}
}
/*
* Take the bits from the runs and drop them into the image.
*/
static void
fillruns(DgpImage_Packed im, dgplssy_cc *boxcc)
{
	DGP_VOID_PROLOGUE("fillruns()", NULL);
	Dgp_Int16 left, right, lastright;
	Dgp_UInt32 *row;
	Dgp_Int32 rownum = 0;
	Dgp_Int32 l, r;
	Dgp_Int32 lb, rb;
	Dgp_Int16 *runs = boxcc->runs;
	Dgp_Int32 nruns = boxcc->nruns;
	
	left = runs[0];
	right = runs[1];
	runs += 2;
	while (nruns) {
		row = DGPIMG_GETROW(im, rownum);
		while (nruns) {
			/* Fill in the range (left..right, rownum) */
			
			/* What words get touched? */
			l = left / DGP_BITS_UINT32;
			lb = left % DGP_BITS_UINT32;
			r = right / DGP_BITS_UINT32;
			rb = right % DGP_BITS_UINT32;
			if (l == r) {
				DGP_VOID_ASSERT(lb <= rb);
				/* Set bits lb to rb inclusive in row[l] */
				row[l] |= rightbits(DGP_BITS_UINT32 - lb) -
					rightbits(DGP_BITS_UINT32 - rb - 1);
			}
			else {
				row[l] |= rightbits(DGP_BITS_UINT32 - lb);
				for (l++; l < r; l++) {
					row[l] = (Dgp_UInt32)~0ul;
				}
				row[r] |= ~rightbits(DGP_BITS_UINT32 - rb - 1);
			}
			
			nruns--;
			if (nruns) {
				lastright = right;
				left = runs[0];
				right = runs[1];
				runs += 2;
				
				if (lastright + 1 >= left) {
					break;
				}
			}
		}
		rownum++;
	}
}


/*
* Write out the classes. Write each one to a PBM file called
* "/tmp/class-NNN-DDD.pbm".
* NNN is the class number and DDD is the dictionary number.
* Outliers get written as "/tmp/class-NNN-DDD-outlier.pbm".
*/
ErrCode
writeClasses( 	TCLYOutCookie toc,
			 TCLYSectionBlock *sb
			 )
{
	Dgp_State *dgp_state = toc->dgp_state;
	DGP_PROLOGUE("writeClasses()", DGP_PERRSTATE);
	DgpAdt_List tokOutput = sb->classesFound; 
	DgpAdt_ListNode ln, memberln;
	dgplssy_cc *membercc;
	tokClass *class;
	tokOffset *offset;
	Dgp_Int32 i;
	Dgp_Int32 rows, cols;
	Dgp_Int32 n;
	Dgp_Int8 fname[100];
	Dgp_Int32 maxw, maxh;
	DgpImage_Packed classim = NULL;
	Dgp_Int32 r, c;
	Dgp_Allocators *a = &(dgp_state->a);
	
	DGPADT_FOREACH (ln, tokOutput) {
		class = (tokClass *)DGPADT_LIGET(ln);
		DGP_ASSERT(class != (tokClass *)NULL);
		
		n = DGPADT_LILEN(class->members);
		if (n >= toc->classThresh) {
			rows = (Dgp_Int32)sqrt((Dgp_Float64)n);
			cols = 1 + (n + rows - 1) / rows;
			
			maxw = class->exemplar.width;
			maxh = class->exemplar.height;
			DGPADT_FOREACH (memberln, class->members) {
				offset = (tokOffset *)DGPADT_LIGET(memberln);
				DGP_ASSERT(offset != (tokOffset *)NULL);
				
				membercc = (dgplssy_cc *)offset->userdata;
				DGP_ASSERT(membercc != (dgplssy_cc *)NULL);
				
				maxw = MAX(maxw, membercc->bitmap.width);
				maxh = MAX(maxh, membercc->bitmap.height);
			}
			
			/* Add borders */
			maxw += PAD;
			maxh += PAD;
			
			DGP_ERR_PASS_CLEANUP(DgpImage_new(dgp_state, DGPIMAGE_PACKED,
				(unsigned)cols * maxw - PAD,
				(unsigned)rows * maxh - PAD, (void**) &classim));
			
			/* First column contains just the exemplar */
			blat(classim, 0, 0, &class->exemplar);
			
			r = 0;
			c = 1;
			DGPADT_FOREACH (memberln, class->members) {
				offset = (tokOffset *)DGPADT_LIGET(memberln);
				membercc = (dgplssy_cc *)offset->userdata;
				
				blat(classim, c * maxw, r * maxh, &membercc->bitmap);
				r++;
				if (r == rows) {
					r = 0;
					c++;
				}
				/*
				* Toast the member's bitmap now.  We only get to this
				* point when we've set TOK_OPT_COPYBITMAPS, and so we own
				* the bitmap.
				*/
				(a->free)((void *)membercc->bitmap.bits);
				membercc->bitmap.bits = (Dgp_UInt32 *)NULL;
			}
			
			sprintf(fname, "/tmp/classes-%04d-%03d.pbm",
				(int)*(Dgp_Int32 *)class->userdata, (int)sb->pages[0].pgNum);
			
			DGP_ERR_PASS_CLEANUP(DgpImage_save(classim, fname));
			
			DgpImage_free(classim);
			classim = NULL;
		}
	}
	
	/* Now outliers. */
	if (toc->classThresh <= 1) {
		for (i = 0; i< sb->numPages; i++){
			DGPADT_FOREACH (ln, sb->pages[i].outliers) {
				membercc = (dgplssy_cc *)DGPADT_LIGET(ln);
				DGP_ASSERT(membercc != (dgplssy_cc *)NULL);
				DGP_ERR_PASS_CLEANUP(DgpImage_new(
					dgp_state, DGPIMAGE_PACKED,
					membercc->bitmap.width,
					membercc->bitmap.height, (void**) &classim));
				
				fillruns(classim, membercc);
				sprintf(fname, 
					"/tmp/classes-%04d-%03d-outlier.pbm", 
					(int)i,
					(int)sb->pages[0].pgNum);
				
				DGP_ERR_PASS_CLEANUP(DgpImage_save(classim, 
					fname));
				
				DgpImage_free(classim);
				classim = NULL;
				i++;
			}
		}
	}
	
	NORMAL_RETURN;
	
ERR_CLEANUP:
	
	if(classim != NULL) DgpImage_free(classim);
	
	NORMAL_RETURN;
}
