/**********************************************************
* tclydoc.c
* $Header: /project/icons/cvsroot/digipaper/src/libtclydoc/tclyverify.c,v 1.6 2003/02/10 22:53:40 mkrocken Exp $
*********************************************************/
#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif

#include "tclydoc.h"
#include "tok-lossy.h"
#include "tclyout.h"
#include "tclypriv.h"
#include "uds.h"

#include "getColorBoxes.h"

#include "compress.h"
#include "tiff_cc2box.h"
#include "deskew.h"
#include "dgp_err_printers.h"
#include "dgp_sys_err_control.h"
#include <errno.h>
#include "dgp_utils.h"
#include "math.h"
#include "image_defs.h"
#include "PsgFix.h"

/* include-libwjr*/
#include <util.h>

/* include-ipcore*/
#include <libutilw.pub>
#include <libpixrw.pub>
#include <librastw.pub>
#include <libiow.pub>
#include <props.pub>
#include <frames.pub>

#include <insertbox.h>
#include <simple_utils.h>

#include "blit.h"
#include "tokclass.h"

#include "token-compress.h"
#include "token-compress-priv.h"

#define	PAD	5


/* 5.10.2000 / dhtesq. */
/* A cheesy version of addMonoPage() that reports new classes found on the page */
/* via the boolean argument.  You also get a box count and a list of the boxes themselves. */
/* The list remains valid until the section is closed. */
DGP_LIBEXPORT ErrCode
tcly_doc_addMonoPageVerify(UInt8 *image, 
					 TCLYDocCookie _cookie, 
					 DgpTdc_Cookie _llTdoc,
					 Dgp_Int32 width,
					 Dgp_Int32 height,
					 Dgp_Int32 bpl,
					 Dgp_Int32 xRes,
					 Dgp_Int32 yRes,
					 Dgp_Int32 pgNum, 
    Int32 (CDECL *rectReturnProc) (
		    IPState    *pIPState,
		    void       *,
		    Int32	x,
		    Int32	y,
		    Int32	w,
		    Int32	h,
		    Int32	m,
		    Int32	nRuns,
		    TCScanLineRun *pRuns),
					 Dgp_Bool report,
                     Dgp_Int32 *newClassesFound,
                     Dgp_Int32 *numBoxes,
                     DgpAdt_List *boxList,
                     DgpAdt_List *xrefList,
                     Dgp_Int32 *numMatches,
                     DgpAdt_List *matchList)
{
  TCLYDocState *cookie = (TCLYDocState *)_cookie;
  tdoc_state *llTdoc = (tdoc_state *)_llTdoc;
  Dgp_State *dgp_state = cookie->dps;
  DgpAdt_ListNode ln;/*, aln; */
  dgplssy_cc *box; /*, *a_box; */
  DgpAdt_ListNode retnode;
  DgpAdt_List matchingBoxes = NULL, xrefs = NULL;
  Dgp_Bool newClassFound = DGP_FALSE;
  Dgp_Int32 matches = 0;
  TCLYPageInfo *curPage = &(cookie->curPage);
  DgpTdc_PageInfo *curLlPage = &(llTdoc->curPage);
  TCLYSectionBlock *sb = cookie->curBlock;
  DgpTdc_SectionBlock *llSb = llTdoc->curBlock;
  ccXref *xref = NULL;

  DGPERR_INIT(tcly_doc_addMonoPage);
	
  *numBoxes = 0;
  *newClassesFound = 0;

  report = report;
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&matchingBoxes, dgp_state));
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&xrefs, dgp_state));

    /*initialize parameters */
    DGP_ERR_PASS_CLEANUP( tcly_startPage(     
                        (TCLYDocCookie) cookie, 
                        (DgpTdc_Cookie) llTdoc,
                        pgNum,
                        width,
                        height,
                        xRes,
                        yRes));
    /* find ccs and tokenize them */
    /* find halftones list */
    /* deskew if required */
    /* return data in ccData */
    DGP_ERR_PASS_CLEANUP(getBoxesLossy(	
		dgp_state,
		image, 
		curPage->width, 
		curPage->height, 
		bpl, 
        cookie->sizeThreshold,
		cookie->grayHalftone, 
		cookie->deskew_cookie,
		&(curPage->skewAngle),
		&curPage->halftones,
        &curPage->imageinfo,
		rectReturnProc,		/* cc finder callback... */
        curPage->ccData,
		(struct timeval *)NULL,
		(struct timeval *)NULL));

    /* put related data into a page */
    DGP_ERR_PASS_CLEANUP(tcly_endPage(         
                        (TCLYDocCookie) cookie, 
                        (DgpTdc_Cookie) llTdoc,
                        DGP_FALSE,
                        DGP_FALSE));

	DGP_FREE(curLlPage->boxes);

	DGP_ASSERT(curPage->numBoxes == (Dgp_Int32)DGPADT_LILEN(curPage->boxes));
	for (ln = DGPADT_LIFIRST(curPage->boxes); ln != DGP_NULLLISTNODE; ) {
		box = (dgplssy_cc *)DGPADT_LIGET(ln);
		DGP_ASSERT(box != (dgplssy_cc *)NULL);
		if (box->bitmap.bits != (Dgp_UInt32 *)NULL) {
			DGP_ASSERT(box->runs == (Dgp_Int16 *)NULL);
			/* It's a real thing - feed it down the tokenizer. */

            /* Allocate for a ccXref{}... */
            xref = (ccXref *)DGP_MALLOC(sizeof(*xref));
            xref->box = box;
            xref->userdata = NULL;

			DGP_ERR_PASS_CLEANUP(tokAddBitmapVerify(cookie->tokstate, 
				&box->bitmap, &newClassFound, (DgpAdt_List *)&xref->classSiblings));
            DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(xrefs, (void *)xref, &retnode));

			ln = DGPADT_LINEXT(ln);
            if(newClassFound){
                /*DGP_REPORTF4(report, "New class: %d x %d @ %d, %d", box->bitmap.width, box->bitmap.height, box->x, box->y); */
                (*newClassesFound)++;
            } else {
				matches++;
				DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(matchingBoxes, (void *)box, &retnode));
            }
            newClassFound = DGP_FALSE;
        } else {
           /*
            * It's got runs - put it on the outliers list.
			* Don't pull it off the boxes list, though... writePages
			* wants it on both.
			*/
            DGP_ASSERT( box->runs != (Dgp_Int16 *)NULL);
            DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(curPage->outliers, (void *)box, &retnode));
            curPage->numOutliers++;
            ln = DGPADT_LINEXT(ln);
        }
	}

    /*DGP_REPORTF1(report, "%d matches found...", matches); */
    if(numMatches) *numMatches = matches;
    if(numBoxes) *numBoxes = curPage->numBoxes;
    if(boxList) *boxList = curPage->boxes;
    if(matchList){
        *matchList = matchingBoxes;
    } else {
        DGP_ERR_PASS(DgpAdt_liFree(matchingBoxes));
    }
    if(xrefList){
        *xrefList = xrefs;
    } else {
        DGPADT_FOREACH(ln, xrefs){
            xref = (ccXref *)DGPADT_LIGET(ln);
            DGP_FREE(xref);
        }
        DGP_ERR_PASS(DgpAdt_liFree(xrefs));
    }
	
    /* add the tokMem to the section */
    /* Update tokMem count*/
    if(cookie->memBond >0){
        if(sb->tokMem + curPage->tokMem + curPage->pageTokMem +
            llSb->tokMem + curLlPage->tokMem + curLlPage->pageTokMem < cookie->memBond){
            sb->tokMem += curPage->tokMem;
            llSb->tokMem += curLlPage->tokMem;
            sb->overflow = DGP_FALSE;
        }else{
            sb->overflow = DGP_TRUE;
        }
    }else sb->overflow = DGP_FALSE;


	NORMAL_RETURN;
	
ERR_CLEANUP:
    if(matchingBoxes) DGP_ERR_PASS(DgpAdt_liFree(matchingBoxes));
    if(xrefs){
        DGPADT_FOREACH(ln, xrefs){
            xref = (ccXref *)DGPADT_LIGET(ln);
            DGP_FREE(xref);
        }
        DGP_ERR_PASS(DgpAdt_liFree(xrefs));
    }
    if(curPage->outliers) DGP_ERR_PASS(DgpAdt_liFree(curPage->outliers));
	
	ERR_RETURN;
}


ErrCode
tcly_doc_tokAddBitmapVerify(TCLYDocCookie _cookie,
                            tokBitmap *bitmap,
                            Dgp_Bool *newClassFound,
                            DgpAdt_List *classMatched)
{
	TCLYDocState *cookie = (TCLYDocState *)_cookie;
	Dgp_State *dgp_state = cookie->dps;
	DGPERR_INIT(tcly_doc_tokAddBitmapVerify);

    DGP_ERR_PASS(tokAddBitmapVerify(cookie->tokstate, 
                                            bitmap, 
                                            newClassFound,
                                            classMatched));

    NORMAL_RETURN;
}

