#ifdef _WIN32
/* include winsock2.h under NT */
#include <winsock2.h>
#endif

#include "dgp_state.h"
#include "compress.h"
#include "deskew.h"
#include "tiff_cc2box.h"
#include "linecc.h"

#include "libiow.pub"
#include "halftone.pub"
#include "libutilw.pub"
#include "libpixrw.pub"
#include <stdlib.h>
#include <stdio.h>


#include "list.h"
#include "iaerror.pub"

#ifdef _WIN32
#include <winsock.h>
#elif defined(VXWORKS) || defined(EQUATOR_CA_ENV)
#include <sys/times.h>
#else
#include <sys/time.h>
#include <sys/resource.h>
#endif  /* (_WIN32) */


#include "halftone.h"
#include "librastw.pub"
#include "dgp_sys_err_control.h"
#include "run2bitmap.h"
#include "reduce.pub"
#include "PsgFix.h"


#define RunToBitmapLookup(b, e)   (lookup[((b) * DGP_BITS_UINT32) + (e)])
#define TOKOVERHEAD 32

/*
* This function takes a list of runs in a connected component and 
* transforms it into a model bitmap.
* It stores the info in *model. The runs are in *pRuns; the lowest
* x coordinate is in x0.
*
* It returns DGP_TRUE on success, DGP_FALSE on failure.
*/
static ErrCode
RunsToBitmap(dgplssy_cc *model, TCScanLineRun *pRuns, Dgp_Int32 x0, Dgp_State *dgp_state)
{
    DGP_PROLOGUE("RunsToBitmap()", DGP_PERRSTATE);
    Dgp_UInt32 *mp;
    static Dgp_UInt32 *lookup = (Dgp_UInt32 *)NULL;
    Dgp_Int32 lastRunLeft, bitEnd, bitStart;
    Dgp_Int32 b;
    Dgp_Int32 y = model->bitmap.height;
    
    if (!lookup) {
        DGP_CHECK_ALLOC_CLEANUP(lookup = MakeLongRunToBitmapLookup());
    }
    
    if(model->bitmap.bits == NULL){
        DGP_CHECK_ALLOC_CLEANUP(model->bitmap.bits = (Dgp_UInt32 *)
            (DGP_CALLOC)((size_t)model->bitmap.longsperline *
            model->bitmap.height,
            sizeof(Dgp_UInt32)));
    }
    
    mp = 	model->bitmap.bits +
        model->bitmap.longsperline * model->bitmap.height;
    lastRunLeft = 0;
    
    while (pRuns) {
        if (lastRunLeft <= pRuns->runRight + 1) {
            mp -= model->bitmap.longsperline;
            y--;
        }
        
        bitStart = pRuns->runLeft - x0;
        bitEnd   = pRuns->runRight - x0;
        
        if ((b = bitStart / DGP_BITS_UINT32) == bitEnd / DGP_BITS_UINT32) {
            mp[b] |=
                RunToBitmapLookup(bitStart % DGP_BITS_UINT32, bitEnd % DGP_BITS_UINT32);
        } else {
            mp[b] |= RunToBitmapLookup(bitStart % DGP_BITS_UINT32, DGP_BITS_UINT32 - 1);
            for (b = bitStart / DGP_BITS_UINT32 + 1; b < bitEnd / DGP_BITS_UINT32; b++) {
                mp[b] = ~(0ul);
            }
            mp[b] |= RunToBitmapLookup(0, bitEnd % DGP_BITS_UINT32);
        }
        
        lastRunLeft = pRuns->runLeft;
        
        pRuns = pRuns->pNext;
    }
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    
    if(model->bitmap.bits != NULL) {
        (DGP_FREE)(model->bitmap.bits);
        model->bitmap.bits = NULL;
    }
    
    ERR_RETURN;
}

/*
* This is the callback function for the ipshared function 
* i_findConnectedComponentRuns.
* It generates models.
*/
static ErrCode 
MakeModelsFromRuns(	IPState *ipwstate, 
                   void *data, 
                   Int32 x, 
                   Int32 y,
                   Int32 w, 
                   Int32 h, 
                   Int32 m, 
                   Int32 nRuns,
                   TCScanLineRun *pRuns)
{
    CCData *ccData = (CCData *)data;
    Dgp_State *dgp_state = ccData->dgp_state;
    Dgp_Int16 * run;
    DgpAdt_ListNode retnode;
    Dgp_Int32 oldCount = 0, newCount = 0;
    
    DGP_PROLOGUE("MakeModelsFromRuns()", DGP_PERRSTATE);
    
#if 1
    TCScanLineRun *pRunsIndex;
    dgplssy_cc *newcc = (dgplssy_cc *)NULL;
    
    DgpCC_PosInfo *pInfo;
    DgpCC_PosInfoBlock *pBlock = ccData->llCCData.pBlock;
    run_elt *tmpruns = (run_elt *)NULL;
    
    
    ipwstate = ipwstate;
    m = m;	/* It's used */
    DGPLSSY_CC_SEC_MALLOC(&newcc);
    
    newcc->bitmap.width = (Dgp_UInt16)w;
    newcc->bitmap.height = (Dgp_UInt16)h;
    newcc->bitmap.longsperline = (newcc->bitmap.width + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
    newcc->bitmap.bits = (Dgp_UInt32 *)NULL;
    newcc->bitmap.userdata = newcc;
    newcc->tok = (token *)NULL;
    
    newcc->runs = (Dgp_Int16 *)NULL;
    newcc->nruns = 0;
    newcc->pgnum = ccData->pgnum;
    /*
    * Where is this box? Not where it ought to be... there's a magic offset
    * of 32 because we're dealing with an ipshared routine here and sometimes
    * ipshared thinks there's a frame of 32 pixels at the top and left. Yay.
    */
    newcc->x = x - 32;
    newcc->y = y - 32;
    if (ccData->second_section)
    {
        newcc->y += ccData->sectionHeight;
    }
    
    /*
    * If this model is "too big" (we need to work on what's too big),
    * then don't build the bitmap -- keep the runs. Make one big area for
    * them - more efficient. Just store 2 Int16's for each run - left and
    * right.
    *
    * 17.4.2001 / dhtesq.
    * Do something similar with "too small" tokens as well, as an experiment 
    * in dealing with flaky symbol matching for smaller cc's.
    *
    */
    if ((w > ccData->maxWidth) || (h > ccData->maxHeight)){       
        
        Dgp_UInt32 bitmapSize = newcc->bitmap.longsperline * h;
        Dgp_UInt32 runRepSize = 2 * sizeof(*newcc->runs) * nRuns;
        
        if (bitmapSize < runRepSize) {
            /* Generate the bitmap */
            DGP_ERR_PASS_CLEANUP(RunsToBitmap(newcc, pRuns, x, dgp_state));
            /* adjust y coordinate */
            newcc->y += newcc->bitmap.height - 1;
        }
        else {
            
            DGP_CHECK_ALLOC_CLEANUP(newcc->runs = (Dgp_Int16 *)(ccData->DGP_MALLOC)(2 * sizeof(*newcc->runs) * nRuns));
            newcc->nruns = nRuns;
            /*
            * Read out the runs. They were generated bottom-to-top, so reverse
            * them.
            */
            for (run = newcc->runs + 2 * (nRuns - 1), pRunsIndex = pRuns; pRunsIndex; run -= 2, pRunsIndex = pRunsIndex->pNext) {
                run[0] = pRunsIndex->runLeft - (Dgp_Int16)x;
                run[1] = pRunsIndex->runRight - (Dgp_Int16)x;
            }
        }
        /* add outliers to outliers list */
        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(ccData->outliers, (void *)newcc, &retnode));
        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(ccData->boxes, (void *)newcc, &retnode));

        if(ccData->membond >0){
                ccData->tokMem += newcc->bitmap.height * newcc->bitmap.longsperline * sizeof(Dgp_Int32) + TOKOVERHEAD;
        }
        newcc = NULL;
        ccData->numOutliers++;
        ccData->count++;
    } else if (w < ccData->minWidth && h < ccData->minHeight) {
        /* Treat this as a lossless token.    */
        tokpos *newpos;
        LlRunToken rTok;
        run_elt *run;
        rTok.width = (Dgp_Int16)w;
        rTok.height = (Dgp_Int16)h;
        
        if (ccData->llCCData.num_boxes >= ccData->llCCData.num_boxes_alloced) {
            ccData->llCCData.num_boxes_alloced += ccData->llCCData.num_boxes_alloced / 2;
            DGP_SEC_REALLOC(ccData->llCCData.boxes, (tokpos *), (sizeof(tokpos) * ccData->llCCData.num_boxes_alloced));
        }
        newpos = &(ccData->llCCData.boxes[ccData->llCCData.num_boxes]);
        
        /*
        * Where is this box? Not where it ought to be... there's a magic offset
        * of 32 because we're dealing with an ipshared routine here and sometimes
        * ipshared thinks there's a frame of 32 pixels at the top and left. Yay.
        */
        newpos->xpos = x - 32;
        newpos->ypos = y - 32 + h - 1; /* we're going to shift this box's coordinates
                                       from specifying the UL corner to the LL
        corner at this point in the execution */
        newpos->extdata = NULL;
        
        if (ccData->second_section) {
            newpos->ypos += (Dgp_Int16) ccData->sectionHeight;
        }
        
        DGP_SEC_MALLOC(tmpruns, (run_elt *), (2 * sizeof(run_elt) * nRuns));
        rTok.runs = tmpruns;
        rTok.nruns = nRuns;
        
        /*
        * Read out the runs. They were generated bottom-to-top, so reverse
        * them.
        */
        for (run = rTok.runs + 2 * (nRuns - 1), pRunsIndex = pRuns; pRunsIndex; run -= 2, pRunsIndex = pRunsIndex->pNext) {
            run[0] = (Int16)(pRunsIndex->runLeft - x); /* normalized x-location */
            run[1] = (Int16)(pRunsIndex->runRight - pRunsIndex->runLeft + 1); /* length */
        }
        
        if(pBlock->num >= pBlock->numAlloced){
            pBlock->numAlloced = pBlock->numAlloced * 3 / 2;
            DGP_SEC_REALLOC(pBlock->poss, (DgpCC_PosInfo *), (pBlock->numAlloced * sizeof(DgpCC_PosInfo)));
        }
        
        pInfo = &(pBlock->poss[pBlock->num]);
        pInfo->page = ccData->llCCData.pageIndex;
        pInfo->posOffset = ccData->llCCData.num_boxes;
        
        
        rTok.userdata = (void *)pBlock->num;
        
        oldCount = ccData->llCCData.tBlock->num;
        DGP_ERR_PASS_CLEANUP(DgpCC_tokenAddCC(dgp_state, ccData->llCCData.tokstate, ccData->llCCData.tBlock, &rTok, &(newpos->tok)));
        if(ccData->membond >0){   
            if(ccData->llCCData.tBlock->num > oldCount){ /* get the cur tok memory */
                ccData->llCCData.tokMem += newcc->bitmap.height * newcc->bitmap.longsperline * sizeof(Dgp_Int32) + TOKOVERHEAD;
            }
        }
        tmpruns = NULL;
        
        pBlock->num++;
        ccData->llCCData.num_boxes++;
        
        ccData->llCCData.count++;
        if (newcc != (dgplssy_cc *)NULL) {
            DGPLSSY_CC_FREE(newcc);
            newcc = NULL;
        }
        
    } else {
        /* Otherwise, generate the bitmap */
        DGP_ERR_PASS_CLEANUP(RunsToBitmap(newcc, pRuns, x, dgp_state));
        /* Hook it onto the list */
        DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(ccData->boxes, (void *)newcc, &retnode));
        /* Feed it down the tokenizer... */
        if(ccData->membond >0){
            tokGetCount(ccData->tokstate, &oldCount);
        }
        DGP_ERR_PASS_CLEANUP(tokAddBitmap(ccData->tokstate, &newcc->bitmap));
        if(ccData->membond >0){
            tokGetCount(ccData->tokstate, &newCount);
            if(newCount > oldCount){ /* get the cur tok memory */
                ccData->tokMem += newcc->bitmap.height * newcc->bitmap.longsperline * sizeof(Dgp_Int32) + TOKOVERHEAD;
            }
        }
        newcc = NULL;	
        ccData->count++;
    }
    
    NORMAL_RETURN;   /* Fall through and clean up (sometimes newcc becomes memory leak D.A.) */
    
ERR_CLEANUP:
    
    if (newcc != (dgplssy_cc *)NULL) {
        if (newcc->bitmap.bits != (Dgp_UInt32 *)NULL) {
            (ccData->DGP_FREE)((void *)newcc->bitmap.bits);
            newcc->bitmap.bits = NULL;
        }
        if (newcc->runs != (Dgp_Int16 *)NULL) {
            (ccData->DGP_FREE)((void *)newcc->runs);
        }
        DGPLSSY_CC_FREE(newcc);
    }
    
    if(tmpruns != NULL) DGP_FREE(tmpruns);
    
    ERR_RETURN;
#else
    ccData->count++;
    NORMAL_RETURN;
#endif
}


/*
* Initial motivation for this was dgpver, which needs to match scan tokens
* without adding them to the tokenizer's internal class table.  You can pass
* this callback as a parameter to getBoxesLossy() if you want to use it...
*/
ErrCode 
MakeModelsFromRunsWithoutTellingTokenizer(	IPState *ipwstate, 
                                          void *data, 
                                          Int32 x, 
                                          Int32 y,
                                          Int32 w, 
                                          Int32 h, 
                                          Int32 m, 
                                          Int32 nRuns,
                                          TCScanLineRun *pRuns)
{
    CCData *ccData = (CCData *)data;
    Dgp_State *dgp_state = ccData->dgp_state;
    DgpAdt_ListNode retnode;
    DGP_PROLOGUE("MakeModelsFromRunsWithoutTellingTokenizer()", DGP_PERRSTATE);
#if 1
    /*TCScanLineRun *pRunsIndex; */
    /*Int16 *run; */
    dgplssy_cc *newcc = (dgplssy_cc *)NULL;
    
    /*DgpCC_PosInfo *pInfo; */
    run_elt *tmpruns = (run_elt *)NULL;
    
    m = m;	/* It's used */
    ipwstate = ipwstate;
    nRuns = nRuns;

    DGPLSSY_CC_SEC_MALLOC(&newcc);
    
    newcc->bitmap.width = (Dgp_UInt16)w;
    newcc->bitmap.height = (Dgp_UInt16)h;
    newcc->bitmap.longsperline = (newcc->bitmap.width + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
    newcc->bitmap.bits = (Dgp_UInt32 *)NULL;
    newcc->bitmap.userdata = newcc;
    newcc->tok = (token *)NULL;
    
    newcc->runs = (Dgp_Int16 *)NULL;
    newcc->nruns = 0;
    newcc->pgnum = ccData->pgnum;
    /*
    * Where is this box? Not where it ought to be... there's a magic offset
    * of 32 because we're dealing with an ipshared routine here and sometimes
    * ipshared thinks there's a frame of 32 pixels at the top and left. Yay.
    */
    newcc->x = x - 32;
    newcc->y = y - 32;
    
    /* Generate the bitmap. */
    DGP_ERR_PASS_CLEANUP(RunsToBitmap(newcc, pRuns, x, dgp_state));
    /* Hook it onto the list */
    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(ccData->boxes, (void *)newcc, &retnode));
    
    newcc = NULL;	
    ccData->count++;
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    
    if (newcc != (dgplssy_cc *)NULL) {
        if (newcc->bitmap.bits != (Dgp_UInt32 *)NULL) {
            (ccData->DGP_FREE)((void *)newcc->bitmap.bits);
            newcc->bitmap.bits = NULL;
        }
        if (newcc->runs != (Dgp_Int16 *)NULL) {
            (ccData->DGP_FREE)((void *)newcc->runs);
        }
        (ccData->DGP_FREE)(newcc);
    }
    
    if(tmpruns != NULL) DGP_FREE(tmpruns);
    
    ERR_RETURN;
#else
    ccData->count++;
    NORMAL_RETURN;
#endif
}

ErrCode
getSectionHeight(
                 Dgp_State *dgp_state,
                 PIXR* readPixr,
                 Dgp_Int32 pgWidth,
                 Dgp_Int32* sectionHeight)
                 
{
    PIXR *reducedPixr = NULL;
    Dgp_Int32 min, min_pos;
    RECTANGL rect;
    Dgp_Int32 *count_array = NULL;
    IPWState *ipwstate = DGP_PIPWSTATE;
    Dgp_Int32 width, height,y, count, val;

    DGP_PROLOGUE("getSectionHeight()", DGP_PERRSTATE);

    /* reduce by a factor of 8 and find line closest to the center that has no */
    /* black pixels in it */
    DGP_ERR_PASS_CLEANUP(
        w_reduce (ipwstate, readPixr, 1, 1, 1, 0, &reducedPixr));
    DGP_ERR_PASS_CLEANUP(
        w_pixrGetWidth (ipwstate, reducedPixr, &width));
    DGP_ERR_PASS_CLEANUP(
        w_pixrGetHeight (ipwstate, reducedPixr, &height));
    rect.w = width;
    rect.h = 1;
    rect.x = 0;
    count_array = DGP_MALLOC (height * sizeof(Dgp_Int32));
    
    for (y = 0; y < height; y++)
    {
        rect.y = y;
        DGP_ERR_PASS_CLEANUP(
            w_numberBitsRect (ipwstate, reducedPixr, &rect, &count));
        count_array[y] = count;
    }
    
    /* check for inverted images */
    val = width * 9 / 10;
    if (count_array[0] >= val && count_array[height-1] >= val)
    {
        *sectionHeight = 1;
        NORMAL_RETURN;
    }

    min = count_array[height/2];
    min_pos = height /2;
    /* find the minimum in some reasonable range */
    for (y = height/2; y >= 0; y--)
    {
        if (count_array[y] < min)
        {
            min = count_array[y];
            min_pos = y;
        }
        if (count_array[height - y - 1] < min)
        {
            min = count_array[height - y - 1];
            min_pos = height - y - 1;
        }
    }
    
    *sectionHeight = min_pos * 8;
    
    w_destroyPixr (ipwstate, &reducedPixr);
    
    /* rectangle of critical section */
    rect.w = pgWidth;
    rect.h = 8;
    rect.x = 0;
    rect.y = *sectionHeight;
    w_rectOf (ipwstate, readPixr, &rect, &reducedPixr);
    rect.h = 1;
    rect.x = 0;
    
    for (y = 0; y < 8; y++)
    {
        rect.y = y;
        w_numberBitsRect (ipwstate, reducedPixr, &rect, &count);
        count_array[y] = count;
    }
    min = count_array[0];
    min_pos = 0;
    /* find the minimum in some reasonable range */
    for (y = 1; y < 8; y++)
    {
        /* printf ("%d  %d\n", y, count_array[y]); */
        if (count_array[y] < min)
        {
            min = count_array[y];
            min_pos = y;
        }
    }
    *sectionHeight += min_pos;
    
    if(count_array) DGP_FREE (count_array);
    if(reducedPixr) w_destroyPixr (ipwstate, &reducedPixr);
    NORMAL_RETURN;
ERR_CLEANUP:
    if(reducedPixr)  w_destroyPixr (ipwstate, &reducedPixr);
    if(count_array)     DGP_FREE (count_array);

    ERR_RETURN;
}

/*
* Read a bunch of tiffFiles. For each one, get the CC's. Return a DgpAdt_List
* containing all of them. Put the number of boxes from each page in
* numBoxes[j], page sizes in pg*[j], and total boxes in *totboxes.
* The x and y values of each CC are based on the upper left corner.
*
* Each element of the DgpAdt_List that is returned contains a pointer to a cc.
* That cc can either contain a CC represented as a bitmap, or a CC represented
* as a set of runs. You can tell which it is by looking at the runs pointer;
* if it's non-NULL, it's as runs. Each cc contains a tokBitmap that contains
* correct data (except that the bits pointer will be NULL if the CC is
* represented as runs.
* The userdata value of each cc's tokBitmap is set to point to the cc.
* All the ccs are allocated with DGPLSSY_CC_SEC_MALLOC() and should be 
* freed with DGPLSSY_DGPLSSY_CC_FREE().
* If readtime or cctime is not NULL, then they should be incremented by
* the time taken in the appropriate portion of this routine.
*
* 16.5.2001 / dhtesq.
* Added customizable callback for cc finder.  Defaults to the usual thing if
* you pass NULL.
*/

ErrCode
getBoxesLossy(   
              Dgp_State *dgp_state,
              UInt8 *image,
              Dgp_Int32 pgWidth,
              Dgp_Int32 pgHeight,
              Dgp_Int32 bpl,
              Dgp_Int32 sizeThresh,
              Dgp_Int32 grayHalftone,
              void *deskew_cookie,
              Float32 *skewAngle,
              DgpAdt_List *halfton_images,       
              DgpAdt_List *imageInfo,
              Int32 (CDECL *rectReturnProc) (
                  IPState    *pIPState,
                  void       *,
                  Int32	x,
                  Int32	y,
                  Int32	w,
                  Int32	h,
                  Int32	m,
                  Int32	nRuns,
                  TCScanLineRun *pRuns),
              CCData *ccData,
              struct timeval *readtime,
              struct timeval *cctime)
{
    DGP_PROLOGUE("getBoxesLossy()", DGP_PERRSTATE);
    Int32 newbpl;
    Int32 returnValue;
    
    DgpAdt_List second_iminfo = DGP_NULLLIST;  
    DgpAdt_ListNode ln;
    Dgp_Bool reverseVideo = DGP_FALSE;
    
    RECTANGL rect;
    
#if !(defined(_WIN32) || defined(VXWORKS) || defined(EQUATOR_CA_ENV))
    struct rusage oldru, newru;
#endif

    PIXR *readPixr = (PIXR *)NULL, *reversePixr = NULL,
        *halftonePixr = NULL,  *botSectionPixr = NULL;
    
    /*for finding the halftones*/
    PIXR *halftone_mask = (PIXR *)NULL;
    UInt8 cell_size;
    IPWState *ipwstate = DGP_PIPWSTATE;
    UInt8 *w_image = NULL;
    
    DGP_CHECK_NULL(halfton_images);
    DGP_CHECK_NULL(imageInfo);
    DGP_CHECK_NULL(image);
    DGP_CHECK_NULL(ccData);
    DGP_CHECK_NULL(ccData->pCCState);

    /* Default to usual callback if passed NULL. */
    if(rectReturnProc == NULL) rectReturnProc = MakeModelsFromRuns;
    
    /* validate small pool usage */
    DGP_SMALLPOOL_VALIDATE(DgpCC_CC, DGPCC_CC_POOL_ID);
    
    DGP_ERR_PASS_CLEANUP(w_createPixr(	
        ipwstate, 
        (Int32)pgWidth, 
        (Int32)pgHeight, 
        1, 
        Int32_TO_ScaledInt1616((Int32)(ccData->xRes)), 
        Int32_TO_ScaledInt1616((Int32)(ccData->yRes)), 
        32, 
        &readPixr));
    DGP_ERR_PASS_CLEANUP(w_pixrGetBpl(ipwstate, readPixr, &newbpl));
    DGP_ERR_GEN_CLEANUP(!(newbpl == bpl),cErrAssertFailure);
    DGP_ERR_PASS_CLEANUP(w_pixrGetImageAll(ipwstate, readPixr, &w_image));
    
    /* Note w_pixrGetHeight exclude frame*/
    DGP_CHECK_ALLOC_CLEANUP(memcpy(w_image, image, (pgHeight +64 )* bpl));
    
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(halfton_images, dgp_state));
    
    
#if !(defined(_WIN32) || defined(VXWORKS) || defined(EQUATOR_CA_ENV))
    if (readtime != (struct timeval *)NULL) {
        DGP_SYS_ERR_GEN_CLEANUP(IS_NEG_ONE(getrusage(RUSAGE_SELF, &oldru)));
    }
#endif
    
    /* if deskewing is desired, attempt to find the angle for the page */
    if (deskew_cookie) {
        DGP_ERR_PASS_CLEANUP(getSkewAngle(deskew_cookie, readPixr, skewAngle));
    }
    
    /* if the page is of reasonable size (in pixels), find an optimal splitting line */
    if (pgHeight > 500 && !deskew_cookie) {
        getSectionHeight(dgp_state, readPixr, pgWidth, &(ccData->sectionHeight));    
    }
    else {
        /* if the page is small or deskewed, just split after the first line */
        ccData->sectionHeight = 1;
    }
    
   
    /* rectangle of critical section */
    rect.w = pgWidth;
    rect.h = pgHeight - ccData->sectionHeight;
    rect.x = 0;
    rect.y = ccData->sectionHeight;
    
    DGP_ERR_PASS_CLEANUP(w_rectOf (ipwstate, readPixr, &rect, &botSectionPixr));
    DGP_ERR_PASS_CLEANUP(w_cutRectIP (ipwstate, 0, 0, pgWidth, ccData->sectionHeight, readPixr)); 
    
    /*w_writeTiff (ipwstate, readPixr, "top.tif", COMPRESSION_NONE , cTrue, 0);
    w_writeTiff (ipwstate, botSectionPixr, "bot.tif", COMPRESSION_NONE , cTrue, 0); */
    
    /* do halftone detection */
    
    /*
    *Ok, trying to run halftones through the lossy tokenizer takes forever
    *Use GetAllHAlftoneInfo to find the halftones, mask them out 
    *and return the halftone portion of the image in *halfton_images
    */
#ifndef  DGP_FAXMODE
    if(ccData->sectionHeight > 64){
        /* we only do segmentation if the strip height is above certain value */
        DGP_ERR_PASS_CLEANUP(GetAllHalftoneInfo(dgp_state, 
            readPixr, 
            &halftone_mask, 
            imageInfo, 
            &cell_size,
            &reverseVideo));
        if(reverseVideo == DGP_TRUE){
            w_invert(ipwstate, readPixr, &reversePixr);
            w_destroyPixr(ipwstate, &readPixr);
            readPixr = reversePixr;
            reversePixr = NULL;
        }

        if(*imageInfo){
            if (*halfton_images == DGP_NULLLIST)
                DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(halfton_images, dgp_state));
        
        
            /*now, create the PIXR that will contain the halftone image*/
            DGP_ERR_PASS_CLEANUP(w_copy(ipwstate, readPixr, &halftonePixr)); 
            /*mask out the halftone region in the source image.*/
            DGP_ERR_PASS_CLEANUP(w_rasterop(ipwstate, 
                0, 
                0, 
                halftone_mask, 
                PIX_NOT(PIX_SRC)&PIX_DST, 
                0, 
                0, 
                pgWidth, 
                ccData->sectionHeight, 
                readPixr));
        
            /*mask out the text in the halftone image*/
            DGP_ERR_PASS_CLEANUP(w_rasterop(	ipwstate, 
                0, 
                0, 
                halftone_mask, 
                PIX_SRC&PIX_DST, 
                0, 
                0, 
                pgWidth, 
                ccData->sectionHeight, 
                halftonePixr));
        
        
#ifdef DGP_HALFTONE_DBUG
        DGP_ERR_PASS_CLEANUP( w_writeTiff (ipwstate, readPixr, "text.tif", COMPRESSION_NONE , cTrue, 0));
        DGP_ERR_PASS_CLEANUP( w_writeTiff (ipwstate, halftonePixr, "halftonePixr.tif", COMPRESSION_NONE , cTrue, 0));
#endif
            if (grayHalftone) {
                DGP_ERR_PASS_CLEANUP(compressHalftoneToGray(dgp_state, halftonePixr, *imageInfo, cell_size, *halfton_images));
            
            
                if (halftonePixr) w_destroyPixr(ipwstate, &halftonePixr);   
            }
            else {
                DGP_ERR_PASS_CLEANUP(makeListofHalftoneRegions(dgp_state, halftonePixr, *imageInfo, *halfton_images));
            }
            if (halftonePixr) w_destroyPixr(ipwstate, &halftonePixr);
            if (halftone_mask) w_destroyPixr(ipwstate, &halftone_mask);   
        
            ln = DGPADT_LIFIRST (*imageInfo);
            while (ln != NULL)
            {
                ((pHALFTONEPIC)DGPADT_LIGET(ln))->section = 1;
                /* halftone detection code can change the size of the image. If it did, */
                /* we need to adjust the pixr */
                if (((pHALFTONEPIC)DGPADT_LIGET(ln))->y + 
                    ((pHALFTONEPIC)DGPADT_LIGET(ln))->h > ccData->sectionHeight)
                    ((pHALFTONEPIC)DGPADT_LIGET(ln))->h = ccData->sectionHeight -
                    ((pHALFTONEPIC)DGPADT_LIGET(ln))->y;
                if (((pHALFTONEPIC)DGPADT_LIGET(ln))->x + 
                    ((pHALFTONEPIC)DGPADT_LIGET(ln))->w > pgWidth)
                    ((pHALFTONEPIC)DGPADT_LIGET(ln))->w = pgWidth -
                    ((pHALFTONEPIC)DGPADT_LIGET(ln))->x;
            
                ln = ln->next;
            }
        }
    }
    if(pgHeight - ccData->sectionHeight > 64){
        /* work on the second section of the page */
        DGP_ERR_PASS_CLEANUP(GetAllHalftoneInfo(dgp_state, 
            botSectionPixr, 
            &halftone_mask, 
            &second_iminfo, 
            &cell_size,
            &reverseVideo));
        if(reverseVideo == DGP_TRUE){
            w_invert(ipwstate, botSectionPixr, &reversePixr);
            w_destroyPixr(ipwstate, &botSectionPixr);
            botSectionPixr = reversePixr;
            reversePixr = NULL;
        }

        if(second_iminfo){
            if (*halfton_images == DGP_NULLLIST)
                DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(halfton_images, dgp_state));
        
            /*now, create the PIXR that will contain the halftone image*/
            DGP_ERR_PASS_CLEANUP(w_copy(ipwstate, botSectionPixr, &halftonePixr)); 
            /*mask out the halftone region in the source image.*/
            DGP_ERR_PASS_CLEANUP(w_rasterop(ipwstate, 
                0, 
                0, 
                halftone_mask, 
                PIX_NOT(PIX_SRC)&PIX_DST, 
                0, 
                0, 
                pgWidth, 
                pgHeight - ccData->sectionHeight, 
                botSectionPixr));
        
            /*DGP_ERR_PASS_CLEANUP( */
            /*w_writeTiff (ipwstate, botSectionPixr, "text.tif", COMPRESSION_NONE , cTrue, 0)); */
            /*mask out the text in the halftone image*/
            DGP_ERR_PASS_CLEANUP(w_rasterop(	ipwstate, 
                0, 
                0, 
                halftone_mask, 
                PIX_SRC&PIX_DST, 
                0, 
                0, 
                pgWidth, 
                pgHeight - ccData->sectionHeight, 
                halftonePixr));
            /*DGP_ERR_PASS_CLEANUP( */
            /*w_writeTiff (ipwstate, halftonePixr, "halftonelayer.tif", COMPRESSION_NONE , cTrue, 0); */
            if (grayHalftone) {
                DGP_ERR_PASS_CLEANUP(compressHalftoneToGray(dgp_state, 
                                                            halftonePixr, 
                                                            second_iminfo, 
                                                            cell_size, 
                                                            *halfton_images));
                if (halftonePixr) w_destroyPixr(ipwstate, &halftonePixr);   
            }
            else
            {
                DGP_ERR_PASS_CLEANUP(makeListofHalftoneRegions(dgp_state, 
                                                                halftonePixr, 
                                                                second_iminfo, 
                                                                *halfton_images));
            }
        
            if (halftonePixr) w_destroyPixr(ipwstate, &halftonePixr);
            if (halftone_mask) w_destroyPixr(ipwstate, &halftone_mask);   
        
            ln = DGPADT_LIFIRST (second_iminfo);
            if (*imageInfo == NULL) DgpAdt_liNew (imageInfo, dgp_state);
            while (ln != NULL)
            {
                DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(*imageInfo, ln->userdata, NULL));
                ((pHALFTONEPIC)DGPADT_LIGET(ln))->section = 2;
                ((pHALFTONEPIC)DGPADT_LIGET(ln))->y += ccData->sectionHeight;
                /* halftone detection code can change the size of the image. If it did, */
                /* we need to adjust the pixr */
                if (((pHALFTONEPIC)DGPADT_LIGET(ln))->y + 
                    ((pHALFTONEPIC)DGPADT_LIGET(ln))->h > pgHeight)
                    ((pHALFTONEPIC)DGPADT_LIGET(ln))->h = pgHeight -
                    ((pHALFTONEPIC)DGPADT_LIGET(ln))->y;
                if (((pHALFTONEPIC)DGPADT_LIGET(ln))->x + 
                    ((pHALFTONEPIC)DGPADT_LIGET(ln))->w > pgWidth)
                    ((pHALFTONEPIC)DGPADT_LIGET(ln))->w = pgWidth -
                    ((pHALFTONEPIC)DGPADT_LIGET(ln))->x;
            
                DGP_ERR_PASS_CLEANUP(DgpAdt_liRemAndNext (ln, &ln));
            }
            DgpAdt_liFree (second_iminfo);
        
        }
    }
#endif	
    
    /* Get all the CCs */
#if !(defined(_WIN32) || defined(VXWORKS) || defined(EQUATOR_CA_ENV))
    if (cctime != (struct timeval *)NULL) {
        DGP_SYS_ERR_GEN_CLEANUP(IS_NEG_ONE(getrusage(RUSAGE_SELF, &oldru)));
    }
#endif
    
    ccData->second_section = DGP_FALSE;
    DGP_ERR_PASS_CLEANUP(w_pixrGetImageAll(ipwstate, readPixr, &w_image));
    DGP_ERR_PASS_CLEANUP(i_findConnectedComponentRuns(	
        ipwstate, 
        w_image, 
        (unsigned)bpl, 
        32, 
        32,
        pgWidth, 
        ccData->sectionHeight,
        (unsigned)sizeThresh, 
        (unsigned)sizeThresh,
        1,
        (void *)ccData,
        rectReturnProc,
        &returnValue));
    
    DGP_ERR_PASS_CLEANUP(w_pixrGetImageAll(ipwstate, botSectionPixr, &w_image));
    ccData->second_section = DGP_TRUE;
    DGP_ERR_PASS_CLEANUP(i_findConnectedComponentRuns(	
        ipwstate, 
        w_image, 
        (unsigned)bpl, 
        32, 
        32,
        pgWidth, 
        pgHeight - ccData->sectionHeight,
        (unsigned)sizeThresh, 
        (unsigned)sizeThresh,
        1,
        (void *)ccData,
        rectReturnProc,
        &returnValue));
#ifdef	INSTRUMENT
    printf(" found %d ccs\n", ccData->count);
#endif
    
#if !(defined(_WIN32) || defined(VXWORKS) || defined(EQUATOR_CA_ENV))
    if (cctime != (struct timeval *)NULL) {
        DGP_SYS_ERR_GEN_CLEANUP(IS_NEG_ONE(getrusage(RUSAGE_SELF, &newru)));
        cctime->tv_sec += newru.ru_utime.tv_sec - oldru.ru_utime.tv_sec;
        cctime->tv_usec += newru.ru_utime.tv_usec - oldru.ru_utime.tv_usec;
    }
#endif
    
    /* shrink the array to the min. size required (it might have more tokpos's */
    /*   in it than just ccData->num_boxes of them)  */
    /* Au contraire!  For lossless integration, lock down this memory so we can */
    /* use it for the lossy tokpos's later.  Jbig2 interface requires continuous arrays. */
    if(grayHalftone){
        if (ccData->llCCData.num_boxes + ccData->count  > ccData->llCCData.num_boxes_alloced){ 
            ccData->llCCData.num_boxes_alloced = ccData->llCCData.num_boxes + ccData->count;
            DGP_SEC_REALLOC(ccData->llCCData.boxes, (tokpos *), 
                             ccData->llCCData.num_boxes_alloced * sizeof(tokpos));
        } else if( ccData->llCCData.num_boxes + ccData->count <=0){
            DGP_FREE(ccData->llCCData.boxes);
            ccData->llCCData.boxes = NULL;
        }
    }   else {      
        if (ccData->llCCData.num_boxes + ccData->count + DGPADT_LILEN(*halfton_images) > 
                    ccData->llCCData.num_boxes_alloced){ 
            ccData->llCCData.num_boxes_alloced = ccData->llCCData.num_boxes + ccData->count + DGPADT_LILEN(*halfton_images);
            DGP_SEC_REALLOC(ccData->llCCData.boxes, (tokpos *), 
                             ccData->llCCData.num_boxes_alloced * sizeof(tokpos));
        } else if( ccData->llCCData.num_boxes + ccData->count + DGPADT_LILEN(*halfton_images) <=0){
            DGP_FREE(ccData->llCCData.boxes);
            ccData->llCCData.boxes = NULL;
        }
    }
    /* and don't need the image any more  - toast the PIXR */
    if(readPixr)        w_destroyPixr(ipwstate, &readPixr);
    if(botSectionPixr)  w_destroyPixr(ipwstate, &botSectionPixr);
    
    /* check lossless outliers; should be freed by here. */
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    
    if (readPixr)         w_destroyPixr(ipwstate, &readPixr);
    if (halftone_mask)    w_destroyPixr(ipwstate, &halftone_mask);   
    if (halftonePixr)     w_destroyPixr(ipwstate, &halftonePixr);
    if (reversePixr)      w_destroyPixr(ipwstate, &reversePixr);   
    if (ccData->llCCData.boxes) {
        DGP_FREE((void *)ccData->llCCData.boxes);
        ccData->llCCData.boxes = NULL;
    }
    
    ERR_RETURN;
}



/***************************************************************************
 *
 */

ErrCode
getBoxesfromBuffer(   
              Dgp_State  *dgp_state,
              UInt8      *buffer,
              Dgp_Int32  bufWidth,
              Dgp_Int32  bufHeight,
              Dgp_Int32  bpl,
              Dgp_Int32  frameBits,
              Dgp_Bool   lastBuffer,
              Dgp_Int32  minWidth,
              Dgp_Int32  minHeight,
              Dgp_Int32  connectType,
              CCData *ccData)

{
    DGP_PROLOGUE("getBoxesLossy()", DGP_PERRSTATE);
    IPWState *pIPWState = DGP_PIPWSTATE;

    DGP_CHECK_NULL(buffer);
    DGP_CHECK_NULL(ccData);
    DGP_CHECK_NULL(ccData->pCCState);
   
    /* validate small pool usage */
    DGP_SMALLPOOL_VALIDATE(DgpCC_CC, DGPCC_CC_POOL_ID);
                
    DGP_ERR_PASS_CLEANUP(tcly_findBufCC(
                                        pIPWState,
                                        ccData->pCCState,
                                        buffer,
                                        lastBuffer,
                                        frameBits,                  /*startX */
                                        frameBits + bufWidth -1,    /*endX */
                                        frameBits,                  /* startY */
                                        frameBits + bufHeight -1,   /*endY */
                                        ccData->accuHeight,
                                        ccData->sectionHeight,
                                        bpl,
                                        minWidth,
                                        minHeight,
                                        connectType,
                                        (void *)ccData,
                                        MakeModelsFromRuns,
                                        NULL));

    ccData->accuHeight += bufHeight;
    /* shrink the array to the min. size required (it might have more tokpos's */
    /*   in it than just ccData.num_boxes of them)  */
    /* Au contraire!  For lossless integration, lock down this memory so we can */
    /* use it for the lossy tokpos's later.  Jbig2 interface requires continuous arrays. */
    if(lastBuffer){
        if (ccData->llCCData.num_boxes + ccData->count >    ccData->llCCData.num_boxes_alloced){ 
            ccData->llCCData.num_boxes_alloced = ccData->llCCData.num_boxes + ccData->count ;
            DGP_SEC_REALLOC(ccData->llCCData.boxes, (tokpos *), 
                             ccData->llCCData.num_boxes_alloced * sizeof(tokpos));
        } else if( ccData->llCCData.num_boxes + ccData->count  <=0){
            DGP_FREE(ccData->llCCData.boxes);
            ccData->llCCData.boxes = NULL;
        }
    }
    NORMAL_RETURN;
    
ERR_CLEANUP:
    
    ERR_RETURN;
}

