/* $Header: /project/icons/cvsroot/digipaper/src/libtok-lossless/tok-lossless.h,v 1.9 2003/02/11 01:31:44 mkrocken Exp $ */

/*
 * Declarations for the top-level tokenisation routines.
 */

#ifndef	TOK_LOSSLESS_H
#define	TOK_LOSSLESS_H

#include <sys/types.h>
#include "list.h"
#include "util.h"
#include "image.h"
#include "mmcompress.h"
#include "dgp_state.h"

/*#include "token-compress.h" */
/*#include "token-compress-priv.h" */

#ifdef __cplusplus
extern "C"
{
#endif

/* The type of a tokeniser state cookie */
typedef void * LlTokState;

/* and the only defined value */
#define	NULLLLTOKSTATE ((LlTokState)0)

/* All the options the tokeniser can take */
typedef struct {
  Dgp_UInt32 flags;	
  Dgp_State *dgp_state;

/* 16.4.2001 / dhtesq. */
/* Added parameters to solve some problems in the *lossy* tokenizer; */
/* it may be some time before these are well-integrated into the lossless version. */
  Dgp_Int32 sizeThresh;
  Dgp_Int32 maxSamptokW;
  Dgp_Int32 maxSamptokH;
  Dgp_Int32 washThresh;
  Dgp_Int32 numStrips;

/*  DgpTdc_JpegCompInfo jpegComp;		// Includes quality & scale factor. */

  Dgp_Bool lossy;					/* (?) */
  Dgp_Bool color;
  Dgp_Bool ascii;
  Dgp_Bool noascii;
  Dgp_Bool halftone;
  Dgp_Bool verbose;

  void *tokenizerParams;
  void *colorTokenizerParams;

} LlTokOptions;

/* The options bit values */
#define	LLTOK_OPT_COPYRUNS	0x01	/* Do I copy the runs? */

/* abstraction for runs: */
typedef Dgp_UInt16 run_elt;
#define RunX(r,i)   ((r)[2*(i)])
#define RunLen(r,i) ((r)[2*(i) + 1])

/* What you get back a DgpAdt_List of */
typedef struct _LlTokClass {
  run_elt *runs;       /* The array of runs representing the token */
  Dgp_Int32 num_runs;        /* the number of runs in exemplar (exemplar must
			  have 2*num_runs elements) */
  Dgp_Int32 width;
  Dgp_Int32 height;
  void *userdata;
  DgpAdt_List members;
  Dgp_UInt32 num_instances;  /* The number of instances of this token */
} *LlTokClass;

typedef struct {
  run_elt *runs;
  Dgp_Int32 nruns;
  Dgp_Int16 width, height;
  void *userdata;
} LlRunToken;


/* The routines */
DGP_LIBEXPORT ErrCode 
DgpLlTokInit(LlTokState *state, LlTokOptions *options);

DGP_LIBEXPORT ErrCode 
DgpLlAddToken(	 LlTokState		state, 
				 LlRunToken		*rTok, 
				 void			*new_data, 
				 void			**old_data);

DGP_LIBEXPORT ErrCode 
DgpLlTokFinal(LlTokState		state, 
			  DgpAdt_List		*output);

DGP_LIBEXPORT void 
DgpLlTokFreeOutput(DgpAdt_List	output, 
				   Dgp_State	*dgp_state);

DGP_LIBEXPORT void 
DgpLlTokAbort(LlTokState state);

DGP_LIBEXPORT ErrCode 
DgpLlTokPartial(LlTokState		state, 
				DgpAdt_List		*output);

DGP_LIBEXPORT ErrCode 
DgpLlTokResume(LlTokState		state, 
			   DgpAdt_List		output);

DGP_LIBEXPORT void 
DgpLlTokSetDefaults(LlTokOptions *options);

DGP_LIBEXPORT Dgp_Int32 
DgpLlTokMemUsage(LlTokState state);

DGP_LIBEXPORT void 
DgpLlTokFreeClass(LlTokClass	c, 
				  Dgp_State		*dgp_state);


#ifdef __cplusplus
}
#endif

#endif
