/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtok-lossless/tokll.c,v 1.5 2003/02/11 01:31:44 mkrocken Exp $
 *
 */
#include "tok-lossless.h"
#include "tokprivate.h"
#if !defined(VXWORKS) && !defined(EQUATOR_CA_ENV)
#include <memory.h>
#endif
#include <math.h>
#include "sm_image.h"
#include "util.h"
#include "hash_table.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtok-lossless/tokll.c,v 1.5 2003/02/11 01:31:44 mkrocken Exp $");

static void 
compute_width_and_height(run_elt	*runs, 
						 Dgp_Int32	num_runs, 
						 Dgp_Int16	*width, 
						 Dgp_Int16	*height);
static ErrCode
tokRunsToClass(LlTokClass	*_new, 
			   run_elt		*runs, 
			   Dgp_Int32	num_runs, 
			   options		*opt,
			   Dgp_Int32	width, 
			   Dgp_Int32	height, 
			   Dgp_State	*dgp_state);


static ErrCode 
addClassTable(ClassTable	classes, 
			  LlTokClass	c, 
			  Dgp_State		*dgp_state);

static void 
freeState(intTokState *state);

static ErrCode 
copy_runs(run_elt	**_new_runs, 
		  run_elt	*runs, 
		  Dgp_Int32 num_runs, 
		  Dgp_State *dgp_state);


/*
 * Free up a class, including all the bits hung off it.
 */
#define TOK_FREECLASS(_c) \
do { \
  if ((_c)->runs != (run_elt *)NULL) { \
	  (DGP_FREE)((void *)(_c)->runs); \
  } \
  TOKCLASS_FREE((_c)); \
} while(0)

void
DgpLlTokFreeClass(LlTokClass c, 
             Dgp_State *dgp_state)
{
  TOK_FREECLASS(c);
}

/*Just hashing on the width and height of the token */
/*is way too slow for halftone tokens, because there */
/*all the same width and height. So instead of keeping */
/*lists at each width and height, we are going to */
/*keep a hash table. Ugh... I know it's going to */
/*be ugly, but the lossless tokenizer is really very */
/*dumb right now. I would rip apart and redo the whole */
/*thing, but I really don't feel like it. */

/*the following functions are necessary for the */
/*hash table */

/*
 * Do the arrays of runs match exactly? (Don't check width, height, or
 * num_runs)
 */
INLINE Dgp_Bool
ExactMatchRuns(run_elt *model1, run_elt *model2, Dgp_Int32 num_runs)
{
  run_elt *r1, *r1end, *r2;

  for (r1 = model1, r2 = model2, r1end = model1 + 2*num_runs;
       r1 < r1end;
       r1++, r2++)
    if (*r1 != *r2)
      return DGP_FALSE;

  return DGP_TRUE;
}


Dgp_Bool isequal(void *cookie, DgpAdt_ht_elt e1, DgpAdt_ht_elt e2) {

  LlTokClass class1 = (LlTokClass)e1;
  LlTokClass class2 = (LlTokClass)e2;
  cookie = cookie;

  /* compare the number of runs */
  if (class1->num_runs != class2->num_runs) {
    return DGP_FALSE;
  }

  return ExactMatchRuns(class1->runs, class2->runs, class1->num_runs);

}

/*I am going to design this hashing function so that it */
/*is geared mostly towards separating halftones well. */
/*The regular tokens seem to separate out pretty */
/*well anyway by the w x h measure. Let's */
/*try mass initially. Mass seemed to improve it */
/*pretty damn well. Lets go further. If two */
/*tokens have the same pixel mass, they are */
/*either the same, which is great, or their */
/*mass is distributed differently. */
/*It is fairly easy and cheap to determine whether there is */
/*more mass on the right or left of the token. Why don't */
/*compute that as well. */

#define MASSBITS 8    /*we are going to use 8 bits for */
		      /*keeping masses. Everything with mass */
		      /*1-255 will have that value in the lower */
		      /*8 bits of the hash. If the mass is bigger, */
		      /*it will be 0. The next bit (9) */
		      /*will have left/right distribution info */

#define MAXMASS 255   /*dependant on MASSBITS obviously */
#define MAXRUNS 15
#define MAXXLOC 15
#define MEMBOUND 379  /*We mod the hash value with this. */

Dgp_UInt16 hash(void *cookie, DgpAdt_ht_elt e) {

  LlTokClass class1 = (LlTokClass)e;
  Dgp_UInt32 mass = 0;
  /*Dgp_UInt32 left = 0; */
  /*Dgp_UInt32 right = 0; */
  Dgp_UInt16 hashval = 0;
  Dgp_UInt16 firstx;
  /*Dgp_UInt32 finx; */
  Dgp_Int32 i;

  /*Dgp_UInt32 tokmiddle = class1->width/2; */

  cookie = cookie;

  firstx=RunX(class1->runs, 0);

  for (i = 0; i < class1->num_runs; i++) {

      mass+=RunLen(class1->runs, i);

      /*now we need to decide how much of the */
      /*weight of this run falls to the left or the right of the middle */
#if 0
      if (RunLen(class1->runs, i)) {

	  if (RunX(class1->runs, i)<tokmiddle) {
	      /*either the weight falls all to the left */
	      /*or it is split by the boundary and falls to both */

	      /*what x coord does the run end at. */
	      finx = RunX(class1->runs, i) + RunLen(class1->runs, i) - 1;

	      if (finx < tokmiddle) {
		  /*it all falls to the left */
		  left += RunLen(class1->runs, i);
	      } else {
		  /*there is some on the right and the left */
		  left += tokmiddle - RunX(class1->runs, i);
		  right+= finx - tokmiddle + 1;
	      }

	  } else {
	      /*all the weight falls to the right. */
	      right+= RunLen(class1->runs, i);
	  }
      }
#endif /* 0 */

  }

  /*just a sanity check */
  /*assert(left+right==mass); */

  /*if mass is bigger than MAXMASS, just hash to 0 */
  if (mass > MAXMASS) return 0;

  /*mass is within our range, mark it in the high order byte of the hashvalue */
  hashval |= (Dgp_UInt16)(mass << 8);

  /*in the low order 4 bits, record the number of runs */
  if (class1->num_runs <= MAXRUNS) hashval |= (Dgp_UInt16)class1->num_runs;

  /*in the next 4, mark the x location of the first run */
  if (firstx <= MAXXLOC) hashval |= (Dgp_UInt16)(firstx << 4);

  /*mark the left/right bit. 1 if more on the left. 0 if more on the right or equal */
  /*if (left > right) hashval|=0x10000; */
   /*  else  hashval|=0; */

  return hashval % MEMBOUND;

}

Dgp_Int32
DgpLlTokMemUsage(LlTokState state)
{
  intTokState *intstate = (intTokState *)state;

  return intstate->memUsage;
}


ErrCode
DgpLlTokInit(LlTokState *_state, LlTokOptions *options)
{
  DGP_STATE_PROLOGUE2("DgpLlTokInit()");
  intTokState *newTokState;
  LlTokClass newClass = (LlTokClass)NULL;


  /* check params */
  DGP_CHECK_NULL(options);
  DGP_ERR_BEGIN(options->dgp_state);
  DGP_CHECK_NULL(_state);

  /* check small mem usage */
  DGP_SMALLPOOL_VALIDATE(struct _LlTokClass, TOKCLASS_POOL_ID);

  /* Now we can start allocating memory. */
  DGP_CHECK_ALLOC_CLEANUP(newTokState = (options->DGP_MALLOC)(sizeof(intTokState)));
  newTokState->dgp_state = options->dgp_state;
  newTokState->classes.elts = NULL;

  /* Copy over the options */
  newTokState->options.flags = options->flags;

  /* Initialize the class table */
  DGP_ERR_PASS_CLEANUP(DgpAdt_smNew(&(newTokState->classes.elts), DGP_TRUE, 0, options->dgp_state));

  newTokState->classes.count = 0;
  newTokState->returned_classes = DGP_FALSE;
  newTokState->cachedClass = newClass = &(newTokState->dummyClass);
  newTokState->memUsage = 0;
  newClass->width = 0;
  newClass->height = 0;
  newClass->num_instances = 0;
  newClass->userdata = (void *)NULL;
  newClass->num_runs = 0;
  newClass->runs = NULL;

  *_state = newTokState;
  newTokState = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if(newTokState != NULL) {
    if(newTokState->classes.elts != NULL) DgpAdt_smFree(newTokState->classes.elts);
    (options->DGP_FREE)(newTokState);
  }

  ERR_RETURN;
}

/* tokAddCC

   Args:
   LlTokState state -- state cookie generated by DgpLlTokInit()

   run_elt *runs  -- An array of run_elt's.  We're requiring this array to
		     conform to the following:

		     0. RELATIVE COORDINATES:
		        x- and y-locations are given relative to the bounding
		        box of the runs which are being passed in; they're not
			given in terms of where they'd actually be drawn on
			the page they were taken from.

		     1. X-LOCATION AND LENGTH:
		        Each run is given by x-location followed by length
                        of the run (in pixels).  The i'th run has x-location
			runs[2*i] and length runs[2*i + 1].

		     2. SKIPPING TO THE NEXT LINE:
		        If run j+1 has an x location less than or equal to
		        one plus the x-location plus the length of run j
			(ie, one plus the x-location of the right-most
			pixel of run j), it will be drawn on the next line
			(on the line with y-location equal to one plus
			the y-location of run j).

		     3. SKIPPING AHEAD TO ANY LINE:
		        If a run has length 0, then it's x-location will be
		        understood to denote a y-delta to jump.  So
			runs immediately following will be drawn starting
			at that new y-location.  The new y-location must be more
			than 1 greater than the current y-location, unless
			you can't use the mechanism in (2) because
			the next run's x-location is > the end of the previous
			run + 1. (for a token with multiple CCs)

			(otherwise, to skip
			to the next line, use the mechanism described in (2)
			above).

		     4. ORDERING OF RUNS:
		        For every pair of runs A and B, if A precedes B in
		        the array of run_elt's, A must precede B in the image
			in the following sense:
			i.  if A and B are to be drawn on the same line, A's
			    x-location is less than B's.
			ii. if A and B are on different lines, A's y-location
			    is less than B's.

		     NOTE that this representation allows for tokens which
		     consist of one or more than one CC.

		     NOTE that one of the goals of this representation is
		     to ensure that two tokens that would have identical
		     bitmaps if they were rendered from their runs will
		     be represented identically.

		     NOTE: as long as you generate all your runs in the same
		     way, you shouldn't have to worry too much about the
		     above requirements (I'm not going to check them explicitly
		     or anything).  The point is just that in theory, one
		     has to be pretty picky about what arrays of runs are
		     allowed if one wants to ensure that no token can be
		     represented in more than one way.

   Dgp_Int32 num_runs   -- The number of runs.  Note that this must equal twice
                     the number of run_elt's in runs, since each run costs
		     two run_elt's: one for the x-location and one for the
		     length
   Dgp_Int32 *width
   Dgp_Int32 *height    -- If *width or *height is -1, DgpLlAddToken will compute
                     the width and height of the bounding box of the token.
		     Otherwise, it will use *width and *height as the
		     dimensions of the bounding box of the token.

   void *new_data
   void **old_data-- If a matching class is found, then *old_data is set to
                     the userdata field of that class. If a matching class
		     is not found, then a new class is created and its
		     userdata field is set to new_data; *old_data is then also
		     set to new_data.

   Return value:
   This function returns DGP_TRUE on success (whether or not a class was matched).
   On error, it returns DGP_FALSE.

   IMPORTANT:
   If you didn't tell DgpLlTokInit that you wanted DgpLlAddToken to make copies
   of the tokens you hand it, runs may be invalid after this function returns.
*/
#define MAX_CACHED_RUNS 5
ErrCode
DgpLlAddToken(LlTokState state, LlRunToken *rTok, void *new_data, void **old_data)
{
  intTokState *intstate = (intTokState *)state;
  Dgp_State *dgp_state = intstate->dgp_state;
  DGP_PROLOGUE("DgpLlAddToken()", DGP_PERRSTATE);
  ClassTable classes;
  LlTokClass newClass = (LlTokClass)NULL;
  LlTokClass checkClass = (LlTokClass)NULL;
  DgpAdt_HashTable whTable=NULL;
  Dgp_Int32 width, height;
  run_elt *runs = rTok->runs; 
  Dgp_Int32 num_runs = rTok->nruns; 
  Dgp_Int16 *_width = &(rTok->width); 
  Dgp_Int16 *_height = &(rTok->height);
  DgpAdt_ListNode retnode;

  /* check if it is cached; */
  checkClass = intstate->cachedClass;
  if((checkClass->num_runs == num_runs) &&
	  ExactMatchRuns(checkClass->runs, runs, num_runs)){
		/* We got a hit.  prepare for hyper-speed. */

	  /*increment its use count */
	  checkClass->num_instances++;

	  /*free the runs */
	  (DGP_FREE)((void *)runs);

	  /*and return its userdata */
	  *old_data = checkClass->userdata;

	  /* add the per cc information */
	  if(checkClass->members!= DGP_NULLLIST){
		DgpAdt_ListNode retnode;
		  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(checkClass->members, rTok->userdata, &retnode));
	  }

	  NORMAL_RETURN;
  }

  checkClass = (LlTokClass)NULL;
  /*
   * Convert it to our internal representation.  It either copies it or
   * takes ownership, according to options->flags.
   */
  DGP_ASSERT(runs != (run_elt *)NULL);
  DGP_ASSERT(num_runs >= 0);
  if ((*_width == -1) || (*_height == -1))
    compute_width_and_height(runs, num_runs, _width, _height);

  width = *_width;
  height = *_height;

  DGP_ERR_PASS_CLEANUP(tokRunsToClass(&newClass, runs, num_runs, &(intstate->options), width, height, dgp_state));

  /* Set the new class's userdata field. */
  newClass->userdata = new_data;

  /*the table that contains the DgpAdt_HashTables for each set */
  /*of tokens with the same width and height */
  classes = &(intstate->classes);

  DGP_ERR_PASS_CLEANUP(DgpAdt_smGet(classes->elts,(Dgp_UInt16)width,(Dgp_UInt16)height,(void **)&whTable));

  if (whTable == NULL) {
	/* No DgpAdt_HashTable there */

    DGP_ERR_PASS_CLEANUP(DgpAdt_htNew(&whTable, NULL, isequal, hash, NULL, 0, dgp_state));

    DGP_ERR_CHECK_NO_MSG(DgpAdt_smSet(classes->elts, (Dgp_UInt16)width, (Dgp_UInt16)height, (void *)whTable)) {
	    DgpAdt_htFree(whTable);
      DGP_GOTO_CLEANUP;
	  }

  }

  /* Try to add it */
  DGP_ERR_PASS_CLEANUP(DgpAdt_htAddToTable(whTable, (DgpAdt_ht_elt)newClass, (DgpAdt_ht_elt *)&checkClass));

  if (checkClass) {
	  /*the element is already in the table */

	  /*increment its use count */
	  checkClass->num_instances++;

	  /*free what the temporary class we allocate */
	  TOK_FREECLASS(newClass);

	  if(num_runs < MAX_CACHED_RUNS)
		intstate->cachedClass = checkClass;

	  /*and return its userdata */
	  *old_data = checkClass->userdata;

	  if(checkClass->members != DGP_NULLLIST){
		DgpAdt_ListNode retnode;
		  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(checkClass->members, rTok->userdata, &retnode));
	  }

	  NORMAL_RETURN;

  } else {
	  /* record our memory usage */
	  intstate->memUsage = num_runs * sizeof(run_elt) + sizeof(*newClass);
  }

  if(num_runs < MAX_CACHED_RUNS)
	  intstate->cachedClass = newClass;

  classes->count++;

  /* All done - it's added. */
  *old_data = new_data;

  /* We have new class that is real create it's member list and
   * add the member data
   */
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&newClass->members, dgp_state));
  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(newClass->members, rTok->userdata, &retnode));

  NORMAL_RETURN;

ERR_CLEANUP:

  /* Free the stuff we've allocated so far */

  if (newClass != (LlTokClass)NULL) {
    TOK_FREECLASS(newClass);
  }

  if (checkClass) {
    TOK_FREECLASS(checkClass);
  }


  ERR_RETURN;
}

/* return the classes of tokens in a DgpAdt_List.  Free the LlTokState */
ErrCode
DgpLlTokFinal(LlTokState state, DgpAdt_List *_output)
{
  intTokState *intstate = (intTokState *)state;
  DGP_PROLOGUE("DgpLlTokFinal()", intstate->DGP_PERRSTATE);
  ClassTable classes;
  DgpAdt_List classlist = DGP_NULLLIST;
  DgpAdt_SMImage elts;
  DgpAdt_HashTable whTable;
  LlTokClass in_table;
  Dgp_UInt16 w, h;


  DGP_CHECK_NULL(_output);

  classes = &(intstate->classes);

  if (classes->elts == (DgpAdt_SMImage)NULL) {
    /* They never gave us any bitmaps at all... */
    DGP_ASSERT(classes->count == 0);
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&classlist, intstate->dgp_state));

    /* Get rid of the state */
    freeState(intstate);

    NORMAL_RETURN;
  }

  /* Generate the final output */
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&classlist, intstate->dgp_state));

  elts = intstate->classes.elts;

  DGPADT_SM_FOREACH(elts,w, h) {

    /* Don't call smGet() unless the memory for the point we want
       is allocated (because otherwise, smGet() will allocate it
       for us on the fly -- not a good way to cut down on the
       amount of memory we're using). */
    if (DgpAdt_smIsAllocated(elts,w,h))
      DGP_ERR_PASS_CLEANUP(DgpAdt_smGet(elts,w,h,(void **)&whTable));
    else
      whTable = NULL;

    if (whTable != NULL) {
      DgpAdt_htResetGetNextElement(whTable);

      DGP_ERR_PASS_CLEANUP(DgpAdt_htGetNextElement(whTable, (DgpAdt_ht_elt *)&in_table));

      while(in_table) {
	      /* Hook that class onto the list we're building */
		DgpAdt_ListNode retnode;
		  DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(classlist, in_table, &retnode));
	      DGP_ERR_PASS_CLEANUP(DgpAdt_htGetNextElement(whTable, (DgpAdt_ht_elt *)&in_table));
      }
    }
  }

  intstate->returned_classes = DGP_TRUE;
  /* Get rid of the state */
  freeState(intstate);

  *_output = classlist;
  classlist = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  /* Free any partially constructed things */

  if (classlist != DGP_NULLLIST)
    DgpAdt_liFree(classlist);

  ERR_RETURN;
}

void
DgpLlTokFreeOutput(DgpAdt_List classlist, Dgp_State *dgp_state)
{
  DGP_VOID_PROLOGUE("DgpLlTokFreeOutput()", DGP_PERRSTATE);
  DgpAdt_ListNode ln;
  LlTokClass cur_class;

  DGPADT_FOREACH (ln, classlist) {
    cur_class = (LlTokClass)DGPADT_LIGET(ln);
    DGP_VOID_ASSERT(cur_class != (LlTokClass)NULL);
    DGP_VOID_ASSERT(cur_class->runs != (run_elt *)NULL);

    /* TOK_FREECLASS() frees the runs */
    TOK_FREECLASS(cur_class);
  }
  DgpAdt_liFree(classlist);

  VOID_RETURN;
}

void
DgpLlTokAbort(LlTokState state)
{
  intTokState *intstate = (intTokState *)state;

  /* Free up all the stuff hung off the state */
  freeState(intstate);
}

/*
 * Copy the bitmap if the options tell you to; otherwise, use the original
 * one. Compute the hash.
 *
 * Return DGP_TRUE on success, DGP_FALSE on failure. On failure, no memory
 * has been allocated or freed.
 */
static ErrCode
copy_runs(run_elt **_new_runs, run_elt *runs, Dgp_Int32 num_runs, Dgp_State *dgp_state)
{
  DGP_PROLOGUE("copy_runs()", DGP_PERRSTATE);
  run_elt *new_runs;
  run_elt *r, *rend, *rnew;

  DGP_CHECK_ALLOC_CLEANUP(new_runs =
	  (run_elt *)(DGP_MALLOC)(num_runs * sizeof(run_elt)));

  for (r = runs, rnew = new_runs, rend = runs + 2*num_runs;
       r < rend;
       r++, rnew++)
    *rnew = *r;

  *_new_runs = new_runs;

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}



static ErrCode
tokRunsToClass(LlTokClass *_new, run_elt *runs, Dgp_Int32 num_runs, options *opt, Dgp_Int32 width, Dgp_Int32 height, 
               Dgp_State *dgp_state)
{
  DGP_PROLOGUE("tokRunsToClass()", DGP_PERRSTATE);
  LlTokClass newClass = (LlTokClass)NULL;

  TOKCLASS_SEC_MALLOC(&newClass);

  newClass->width = width;
  newClass->height = height;
  newClass->num_instances = 1;

  newClass->userdata = (void *)NULL;
  newClass->num_runs = num_runs;
  if (opt->flags & LLTOK_OPT_COPYRUNS) {
    /* we should copy the runs */
    DGP_ERR_PASS_CLEANUP(copy_runs(&(newClass->runs), runs, num_runs, dgp_state));
  }
  else {
    /* Just use the pointer we were given */
    newClass->runs = runs;
  }

  *_new = newClass;
  newClass = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  if (newClass != (LlTokClass)NULL) {
    TOKCLASS_FREE(newClass);
  }

  ERR_RETURN;
}

/*
 * Ideas on compute a hash (currently not implemented).
 *
 * Incorporate the following information in some way:
 * 1. Whether the majority of the pixels are above or below the horizontal
 *    mid-line of the token (1 bit)
 * 2. Whether the majority of the pixels are to the left or to the right of
 *    the vertical mid-line of the token (1 bit)
 * 3. Whether the mean displacement in x of all the pixels from the
 *    vertical mid-line is greater than the mean displacement in y of all
 *    the pixels from the horizontal mid-line (1 bit)
 * 4. The number of rows that have the most common number of runs, the
 *    number that have the next most common, etc., up to the 4th most
 *    common (16 bits total)
 * 5. What the most common number of runs is (4 bits)
 * 6. What the number of lines skipped is (4 bits)
 * 7. What the least common length of a run is (5 bits)
 */

/*
 * Free up everything that's hooked into the state. Watch out for NULL
 * pointers - they're valid in most fields.
 */
static void
freeState(intTokState *state)
{
  Dgp_UInt16 w, h;
  DgpAdt_SMImage elts;
  DgpAdt_HashTable whTable;
  LlTokClass in_table;
  Dgp_State *dgp_state;


  if(state == NULL) return;

  dgp_state = state->dgp_state;

  elts = state->classes.elts;
  if (elts != (DgpAdt_SMImage)NULL) {
    DGPADT_SM_FOREACH(elts, w, h) {

      /* Don't call smGet() unless the memory for the point we want
	 is allocated (because otherwise, smGet() will allocate it
	 for us on the fly -- not a good way to cut down on the
	 amount of memory we're using). */
      if (DgpAdt_smIsAllocated(elts,w,h))
	      DgpAdt_smGet(elts,w,h,(void **)&whTable);
      else
	      whTable = NULL;

      if (whTable != NULL) {
	      if (!state->returned_classes) {
	        /* we still have responsibility for these classes, so we
	           need to free them (if we've passed them back to the user
	           via DgpLlTokFinal(), it's his/her responsibility to clean them
	           up). */
	           DgpAdt_htResetGetNextElement(whTable);

	          DgpAdt_htGetNextElement(whTable, (DgpAdt_ht_elt *)&in_table);

	           while(in_table) {
		       /* free this class*/
		      TOK_FREECLASS(in_table);

		      DgpAdt_htGetNextElement(whTable, (DgpAdt_ht_elt *)&in_table);
	           }

	      }
	      DgpAdt_htFree(whTable);
      }


    }

    DgpAdt_smFree(elts);
  }

  (DGP_FREE)((void *)state);
}


static void
compute_width_and_height(run_elt *runs, Dgp_Int32 num_runs, Dgp_Int16 *width,
			 Dgp_Int16 *height) {
  Dgp_Int32 maxy = 0, maxx = 0;
  Dgp_Int32 i;
  Dgp_Int32 prev_x = -1;

  for (i = 0; i < num_runs; i++) {

    if (RunLen(runs,i) == 0) {
      /* skipped ahead in y */
      maxy = RunX(runs,i);
      continue;
    }

    if (RunX(runs,i) < prev_x)
      /* skipped to next line */
      maxy++;

    if ((prev_x = RunX(runs,i) + RunLen(runs,i)) > maxx)
      maxx = prev_x;
  }
  *width  = maxx + 1;
  *height = maxy + 1;
}

void DgpLlTokSetDefaults(LlTokOptions *options) {
    options->flags = 0;
    options->dgp_state = NULL;
}

ErrCode
DgpLlTokPartial(LlTokState state, DgpAdt_List *_output) 
{
  intTokState *intstate = (intTokState *)state;
  DGP_PROLOGUE("DgpLlTokPartial()", intstate->DGP_PERRSTATE);
  ClassTable classes;
  DgpAdt_List classlist = DGP_NULLLIST;
  DgpAdt_SMImage elts;
  DgpAdt_HashTable whTable;
  LlTokClass in_table;
  Dgp_UInt16 w, h;


  DGP_CHECK_NULL(_output);

  classes = &(intstate->classes);

  if (classes->elts == (DgpAdt_SMImage)NULL) {
    /* They never gave us any bitmaps at all... */
    DGP_ASSERT(classes->count == 0);
    DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&classlist, intstate->dgp_state));

    NORMAL_RETURN;
  }

  /* Generate the final output */
  DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&classlist, intstate->dgp_state));

  elts = intstate->classes.elts;

  DGPADT_SM_FOREACH(elts,w,h) {

    /* Don't call smGet() unless the memory for the point we want
       is allocated (because otherwise, smGet() will allocate it
       for us on the fly -- not a good way to cut down on the
       amount of memory we're using). */
    if (DgpAdt_smIsAllocated(elts,w,h)) {
      DGP_ERR_PASS_CLEANUP(DgpAdt_smGet(elts,w,h,(void **)&whTable));
    } else {
      whTable = NULL;
    }

    if (whTable != NULL) {

      DgpAdt_htResetGetNextElement(whTable);

      DGP_ERR_PASS_CLEANUP(DgpAdt_htGetNextElement(whTable, (DgpAdt_ht_elt *)&in_table));

      while(in_table) {
		  DgpAdt_ListNode retnode;
	      /* Hook that class onto the list we're building */
	      DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(classlist, in_table, &retnode));
	      DGP_ERR_PASS_CLEANUP(DgpAdt_htGetNextElement(whTable, (DgpAdt_ht_elt *)&in_table));
      }
      DGP_ERR_PASS_CLEANUP(DgpAdt_smSet(elts,w,h,NULL));
      DgpAdt_htFree(whTable);
    }
  }

#ifdef DEBUG_INCRDICT
  fprintf(stderr,"DgpAdt_List being returned in DgpLlTokPartial\n");

  DGPADT_FOREACH(classln,classlist){
	c = ClassData(classln);
	fprintf(stderr,"%d :: tok 0x%X\n",i,(Dgp_Int32)c->tok);
	i += 1;
  }
  fprintf(stderr,"Length returned from DgpLlTokPartial %d\n",DGPADT_LILEN(classlist));
#endif

  *_output = classlist;
  classlist = NULL;

  NORMAL_RETURN;

ERR_CLEANUP:

  /* Free any partially constructed things */

  if (classlist != DGP_NULLLIST)
    DgpAdt_liFree(classlist);

  ERR_RETURN;
}


static ErrCode
addClassTable(ClassTable classes, LlTokClass c, Dgp_State *dgp_state)
{
  DGP_PROLOGUE("addClassTable()", DGP_PERRSTATE);
  DgpAdt_HashTable whTable = NULL;
  LlTokClass in_table;
  void *e;

  if (classes->elts == (DgpAdt_SMImage)NULL) {
    DGP_ERR_PASS_CLEANUP(DgpAdt_smNew(&(classes->elts), DGP_TRUE, 0, dgp_state));
  }

  DGP_ERR_PASS_CLEANUP(DgpAdt_smGet(classes->elts, (Dgp_UInt16)c->width, (Dgp_UInt16)c->height, &e));

  if (e == NULL) {
    /* No DgpAdt_HashTable there to hold c */
    DGP_ERR_PASS_CLEANUP(DgpAdt_htNew(&whTable, NULL, isequal, hash, NULL, 0, dgp_state));
    e = whTable;

    DGP_ERR_PASS_CLEANUP(DgpAdt_smSet(classes->elts, (Dgp_UInt16)c->width, (Dgp_UInt16)c->height, (void *)whTable));
    /* We no Dgp_Int32er own whTable - don't free it on cleanup */
    whTable = NULL;
  }

  DGP_ERR_PASS_CLEANUP(DgpAdt_htAddToTable((DgpAdt_HashTable)e, (void *)c, (DgpAdt_ht_elt *)&in_table));

  NORMAL_RETURN;

ERR_CLEANUP:

  if (whTable != NULL) {
    DgpAdt_htFree(whTable);
  }

  ERR_RETURN;
} /* end of addClassTable */

ErrCode 
DgpLlTokResume(LlTokState state, DgpAdt_List inlist){
  intTokState *intstate = (intTokState *)state;
  DGP_PROLOGUE("DgpLlTokResume()", intstate->DGP_PERRSTATE);
  LlTokClass newclass;
  DgpAdt_ListNode listnode;

  if (inlist == DGP_NULLLIST)
    goto jobdone;

#ifdef DEBUG_INCRDICT
  fprintf(stderr,"In DgpLlTokResume.. adding \n");
#endif

  /* reset cached class */
  intstate->cachedClass = &(intstate->dummyClass);

  DGPADT_FOREACH(listnode,inlist){
    newclass = (LlTokClass)ClassData(listnode);
    DGP_ASSERT(newclass != (LlTokClass)NULL);
    DGP_ERR_PASS(addClassTable(&(intstate->classes),newclass, intstate->dgp_state));
  } /* end of the DGPADT_FOREACH loop */

  jobdone:
    DgpAdt_liFree(inlist);
    NORMAL_RETURN;
}
