/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/bitfiddle.h,v 1.3 2000/03/04 00:22:41 mkrocken Exp $
 *
 */

#ifndef	BITFIDDLE_H

#define	BITFIDDLE_H

/*
 * What this macro does:
 * It's called with 3 params:
 *	v - a uDgp_Int32 value
 *	b - a bit index specifying a bit in v
 *	nb - a bit index specifying a bit in the output.
 * Its job is to extract bit number b (numbered from the MSB, sigh) from
 * v, and shift it so that it appears in bit nb (numbered from the LSB) of
 * the result.
 *
 * nb must be between 0 and 31 inclusive.
 * b might be between -2 and 31. Yay. b<0 should be treated as (b+32) - whoever
 * called us should have set up the v value so that this is right.
 */

/*
 * How it does what it does:
 * There are two ways to do the basic operation:
 *	1) Shift bit b to the LSB, mask off the LSB, shift the LSB to bit nb.
 *	2) Mask off bit b; shift bit b to bit nb.
 *	(We could also shift b to nb and mask off nb, but that's pretty
 *	 similar).
 *
 * Now, it turns out that gcc -O2 generates different instruction sequences
 * for these, if the operands are constants (and they'd better be, or this
 * will generate truly lousy code). So we switch...
 *
 * Another tweak is that the shift-then-mask method has to shift left sometimes
 * and right other times, so it uses a conditional expression; since the
 * thing the expression tests is constant, one branch will always be dead code.
 * But that's not good enough: the dead branch still contains an illegal
 * operation (shift by a negative number), and the compiler complains. So
 * we stick an ABS() around the shift distance... shuts the compiler up and
 * doesn't change the code. Sigh.
 */

/* The shift-mask-shift version */
#define	MoveAndMaskSMS(v, b, nb)					\
	((((v) >> (31 - ((b) & 31))) & 1) << (nb))

/* The mask-shift version */
#define	MoveAndMaskMS(v, b, nb)						\
	(((31 - ((b) & 31)) > (nb)) ?					\
	 (((v) & (1ul << (31 - ((b) & 31)))) >> ABS((31 - ((b) & 31)) - (nb))) : \
	 (((v) & (1ul << (31 - ((b) & 31)))) << ABS((nb) - (31 - ((b) & 31)))))

/*
 * The hybrid version. The rules here are:
 * SMS produces better code if (31 - ((b) & 31)) >= 12 and nb < 12,
 * or if b == 0, or if nb == 31. MS generates better code in other cases.
 *
 * Where did the magic number 12 come from, you ask? Well, it's because the
 * SPARC has a 12-bit immediate field... numbers over that have to be
 * synthesised into registers before they can be used. If we have enough
 * registers, we can re-use them, but...
 *
 * Of course, this means that on anything but a SPARC, and any compiler
 * but gcc, this may not generate the best code. Life's hard.
 */
#define	MoveAndMask(v, b, nb)					\
	(((b == 0) || ((nb) == 31) ||				\
	  (((31 - ((b) & 31)) >= 12) && ((nb) < 12))) ?		\
	 MoveAndMaskSMS((v), (b), (nb)) :			\
	 MoveAndMaskMS((v), (b), (nb)))

#endif
