/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/classtable.c,v 1.14 2003/02/10 22:53:41 mkrocken Exp $
 *
 */

#ifndef _WIN32

/*#define TOKCLASS_TIMING */
#ifdef TOKCLASS_TIMING
#define calc_time( new_usage, old_usage) \
    (((new_usage).ru_utime.tv_sec - (old_usage).ru_utime.tv_sec) * 1000000L + \
    ((new_usage).ru_utime.tv_usec - (old_usage).ru_utime.tv_usec)) / 1000000.0
#include <sys/time.h>
#include <sys/resource.h>
#endif

#endif

#include "util.h"
#include "tokprivate.h"
#include "class.h"
#include "classtable.h"
#include "image_defs.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/classtable.c,v 1.14 2003/02/10 22:53:41 mkrocken Exp $");

/*
 * This file contains routines that operate over the entire table of
 * classes.
 */

static ErrCode expandTable(tokClassTable classes, Dgp_Int32 w, Dgp_Int32 h, 
                           tokOptions *options);
static ErrCode scanBin(TokState state, tokClassTable classes, DgpAdt_List bin, 
                       Dgp_Int32 checkStyle);
static ErrCode scanTwoBins(TokState state, tokClassTable classes, DgpAdt_List bina, 
                           DgpAdt_List binb, Dgp_Int32 checkStyle);
static ErrCode scanTable(TokState state, tokClassTable classes, 
                         Dgp_Int32 checkStyle);
static ErrCode regularise(tokClassTable classes, Dgp_Int32 checkStyle, 
			  tokOptions *opt);

/*
 * Styles for scanTable
 */
#define	SCAN_CHECKALL		0	/* Check all things */
#define	SCAN_CHECKNOSINGLETONS	1	/* Check only those with >1 member */
#define	SCAN_CHECKSINGLETONS	2	/* Check only those with 1 member */
#define	SCAN_CHECKSYNTH		3	/* Check only when one has synth ex */

/*
 * Publically declared functions.
 */
ErrCode 
tokCTExpandTable(TokState state, tokClassTable classes, Dgp_Int32 w, Dgp_Int32 h)
{
    _tokState *intstate = (_tokState *)state;
    return expandTable( classes, w, h, &(intstate->options));
}


/*
 * Now functions internal to this library:
 */

/*
 * Add the given class to the table.
 */
ErrCode
tokCTAddToTable(tokClassTable classes, tokCClass c, tokOptions *options)
{
    Dgp_State *dgp_state = options->dgp_state;
    DgpAdt_List newBin = DGP_NULLLIST;
    DgpAdt_ListNode retnode;
    tokCBoxType *exemplar = c->exemplar;

DGP_PROLOGUE("tokCTAddToTable()", DGP_PERRSTATE);

    if ((classes->elts == (DgpImage_Ptr)NULL) ||
	    (DGPIMG_WIDTH(classes->elts) <= exemplar->width) ||
	    (DGPIMG_HEIGHT(classes->elts) <= exemplar->height)) {
DGP_ERR_PASS_CLEANUP(
	expandTable(classes, exemplar->width, exemplar->height, options));
    }

    if ((DgpAdt_List)tokCTRef(classes, exemplar->width, exemplar->height)
            == DGP_NULLLIST) {
	/* No bin there to hold it */
	DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&newBin, options->dgp_state));
	tokCTRef(classes, exemplar->width, exemplar->height) = (void *)newBin;
	/* We no Dgp_Int32er own newBin - don't free it on cleanup */
	newBin = DGP_NULLLIST;
    }

    DGP_ERR_PASS_CLEANUP (DgpAdt_liAdd((DgpAdt_List)tokCTRef(classes, exemplar->width, exemplar->height),
	      (void *)c, &retnode));

    classes->count++;

    NORMAL_RETURN;

ERR_CLEANUP:

    if (newBin != DGP_NULLLIST) {
	DgpAdt_liFree(newBin);
    }

    ERR_RETURN;
}

/*
 * Expand the table enough to fit in a box with width, height (w,h).
 * Create the table if it didn't exist before.
 */
static ErrCode
expandTable(tokClassTable classes, Dgp_Int32 w, Dgp_Int32 h, tokOptions *options)
{
    Dgp_State *dgp_state = options->dgp_state;
    Dgp_Int32 oldWidth, oldHeight;
    Dgp_Int32 newWidth, newHeight;
    Dgp_Int32 i, j;
    DgpImage_Ptr newTable = (DgpImage_Ptr)NULL;

DGP_PROLOGUE("expandTable()", DGP_PERRSTATE);

    if (classes->elts == (DgpImage_Ptr)NULL) {
	oldWidth = oldHeight = 0;
	newWidth = w + 1;
	newHeight = h + 1;
    }
    else {
	oldWidth = DGPIMG_WIDTH(classes->elts);
	oldHeight = DGPIMG_HEIGHT(classes->elts);
	newWidth = MAX(oldWidth, w + 1);
	newHeight = MAX(oldHeight, h + 1);
    }

    /* Might want to put some extra border? */
DGP_ERR_PASS_CLEANUP(
    DgpImage_new(dgp_state, DGPIMAGE_PTR | DGPIMAGE_DONTCLEAR,
			       (unsigned)newWidth, (unsigned)newHeight, (void **)&newTable));

    /* Copy the old info */
    for (i = 0; i < oldHeight; i++) {
	for (j = 0; j < oldWidth; j++) {
	    DGPIMG_REF(newTable, j, i) = DGPIMG_REF(classes->elts, j, i);
        }
    }

    /* Clear the new places */
    for (i = oldHeight; i < newHeight; i++) {
	for (j = 0; j < newWidth; j++) {
	    DGPIMG_REF(newTable, j, i) = (void *)DGP_NULLLIST;
        }
    }
    for (i = 0; i < oldHeight; i++) {
	for (j = oldWidth; j < newWidth; j++) {
	    DGPIMG_REF(newTable, j, i) = (void *)DGP_NULLLIST;
        }
    }

    if (classes->elts != (DgpImage_Ptr)NULL) {
	DgpImage_free(classes->elts);
    }
    classes->elts = newTable;

    NORMAL_RETURN;

ERR_CLEANUP:
    if (newTable != (DgpImage_Ptr)NULL) {
	DgpImage_free(newTable);
    }

    ERR_RETURN;
}

/*
 * Generate exemplars for everything in the class table.
 */
ErrCode 
generateAllExemplars(TokState state, tokClassTable classes)
{
    tokOptions *opt = &(((_tokState *)state)->options);
    Dgp_State *dgp_state = opt->dgp_state;
    DgpAdt_ListNode classi;
    DgpAdt_List curlist;
    Dgp_Int32 w, h;

DGP_PROLOGUE("generateAllExemplars()", DGP_PERRSTATE);

    for (h = 0; h < DGPIMG_HEIGHT(classes->elts); h++) {
	for (w = 0; w < DGPIMG_WIDTH(classes->elts); w++) {
	    curlist = (DgpAdt_List)tokCTRef(classes, w, h);
	    if (curlist != DGP_NULLLIST) {
		DGPADT_FOREACH (classi, curlist) {
DGP_ERR_PASS_CLEANUP(
		    tokCGenerateExemplar(state, tokCClassData(classi)));
                }
            }
        }
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

/*
 * Scan the table, calling maybeMergeClasses on all kinds of pairs of classes.
 *
 * First, scan the table calling it within each bin. Then re-scan calling it
 * between classes within pairs of bins; use CalcOrder to find what pairs
 * of bins are good.
 */
static ErrCode 
scanTable(TokState state, tokClassTable classes, Dgp_Int32 checkStyle)
{
    _tokState *intstate = (_tokState *)state;
    tokOptions *opt = &(intstate->options);
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_Int32 w, h;
    DgpAdt_List curlist;
    DgpAdt_List otherlist;
    Dgp_Int32 *order;
    Dgp_Int32 ordernum;
    Dgp_Int32 i;

DGP_PROLOGUE("scanTable()", DGP_PERRSTATE);

    /* Get the scan order */
    if( intstate->search_order == (Dgp_Int32 *)NULL) {
DGP_ERR_PASS_CLEANUP(
        tokCTCalcOrder(opt->widthDiff, opt->heightDiff, 
                       &(intstate->search_count), &(intstate->search_order),
                       opt));
    }
    order = intstate->search_order;
    ordernum = intstate->search_count;

    /* Fix up sizes and set status bits */
DGP_ERR_PASS_CLEANUP(
    regularise(classes, checkStyle, opt));

    for (h = 0; h < DGPIMG_HEIGHT(classes->elts); h++) {
	for (w = 0; w < DGPIMG_WIDTH(classes->elts); w++) {
	    curlist = (DgpAdt_List)tokCTRef(classes, w, h);
	    if (curlist != DGP_NULLLIST) {
DGP_ERR_PASS_CLEANUP(
		scanBin(state, classes, curlist, checkStyle));
            }
        }
    }

    /* Ignore order[0] - it's this bin */
    for (i = 1; i < ordernum; i++) {
	for (h = MAX(-order[2 * i + 1], 0);
	     h < MIN(DGPIMG_HEIGHT(classes->elts),
		     DGPIMG_HEIGHT(classes->elts) - order[2 * i + 1]);
	     h++) {
	    for (w = MAX(-order[2 * i], 0);
		 w < MIN(DGPIMG_WIDTH(classes->elts),
			 DGPIMG_WIDTH(classes->elts) - order[2 * i]);
		 w++) {
		curlist = (DgpAdt_List)tokCTRef(classes, w, h);
		otherlist = (DgpAdt_List)tokCTRef(classes, w + order[2 * i],
					h + order[2 * i + 1]);
		if ((curlist != DGP_NULLLIST) && (otherlist != DGP_NULLLIST) &&
		    (curlist != otherlist) /*HACK*/) {
DGP_ERR_PASS_CLEANUP(
		    scanTwoBins(state, classes, curlist, otherlist, 
                                checkStyle));
                }
            }
        }
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

/*
 * This should be called before scanTable: it goes through the table,
 * makes sure each class is in the right bin (when an exemplar changes, its
 * size may change as well, which could mean it's no Dgp_Int32er in the correct
 * place in the table), upgrades all TOKCLASS_OKSOFAR to TOKCLASS_MERGECLEAN,
 * and marks everything TOKCLASS_OKSOFAR.
 *
 * Actually, it only does that status bit fiddling if checkStyle is
 * SCAN_CHECKALL, since that's the only version that does all the checking.
 */
static ErrCode
regularise(tokClassTable classes, Dgp_Int32 checkStyle, tokOptions *opt)
{
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_Int32 w, h;
    DgpAdt_ListNode classi, moveln;
    DgpAdt_List curlist;
    DgpAdt_List movelist = DGP_NULLLIST;
    tokCClass c;

DGP_PROLOGUE("regularise()", DGP_PERRSTATE);

    /*
     * This is a bit messy.  See, the problem is that you can't just
     * do the move and the status bit fiddling at the same time, or some
     * class could get TOKCLASS_OKSOFAR, be moved to another bin, and then
     * get upgraded to TOKCLASS_MERGECLEAN - without going through a merge.
     * So we optimise for the common case: moves are rare, upgrades are
     * common.  Upgrade everything, and keep track of all the classes that
     * should be moved.  Once all the upgrades are done, do the moves.
     * Be careful that in case of failure, you haven't corrupted/lost
     * anything.
     *
     * Also note that any class that gets moved automatically gets
     * MERGECLEAN cleared - it may have been compared against everything
     * visible from its old location, but the new location brings new
     * classes into view; this merge is therefore not clean for it.
     */
    DGP_ERR_PASS_CLEANUP (DgpAdt_liNew(&movelist, dgp_state));

    for (h = 0; h < DGPIMG_HEIGHT(classes->elts); h++) {
	for (w = 0; w < DGPIMG_WIDTH(classes->elts); w++) {
	    curlist = (DgpAdt_List)tokCTRef(classes, w, h);
	    if (curlist != DGP_NULLLIST) {
		DGPADT_FOREACH (classi, curlist) {
		    c = tokCClassData(classi);

		    if (checkStyle == SCAN_CHECKALL) {
			if (c->status & TOKCLASS_OKSOFAR) {
			    c->status |= TOKCLASS_MERGECLEAN;
			    }
			c->status |= TOKCLASS_OKSOFAR;
			}

		    if ((c->exemplar->width != w) ||
			(c->exemplar->height != h)) {
				DgpAdt_ListNode retnode;
			/* It's a mover - hook the DgpAdt_ListNode onto movelist */
			c->status &= ~TOKCLASS_MERGECLEAN;
			DGP_ERR_PASS_CLEANUP (
				DgpAdt_liAdd(movelist, (void *)classi, &retnode));
                    }
                }
            }
        }
    }

    /* Now do the moves */
    DGPADT_FOREACH (moveln, movelist) {
	classi = (DgpAdt_ListNode)DGPADT_LIGET(moveln);

	/* Add it to its proper bin */
DGP_ERR_PASS_CLEANUP(
	tokCTAddToTable(classes, tokCClassData(classi), opt));

	/* and unhook it from the wrong bin */
	DgpAdt_liRem(classi);
	/* tokCTAddToTable thought it was a new class */
	classes->count--;
    }
    DgpAdt_liFree(movelist);

    NORMAL_RETURN;

ERR_CLEANUP:

    if (movelist != DGP_NULLLIST) {
	DgpAdt_liFree(movelist);
    }

    ERR_RETURN;
}

/*
 * Scan through all the classes in the table, making sure that each one's
 * members matches the exemplar.
 */
static ErrCode
checkAllClasses(tokClassTable classes, tokOptions *opt)
{
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_Int32 w, h;
    DgpAdt_ListNode classi;
    DgpAdt_List curlist;

DGP_PROLOGUE("checkAllClasses()", DGP_PERRSTATE);

    for (h = 0; h < DGPIMG_HEIGHT(classes->elts); h++) {
	for (w = 0; w < DGPIMG_WIDTH(classes->elts); w++) {
	    curlist = (DgpAdt_List)tokCTRef(classes, w, h);
	    if (curlist != DGP_NULLLIST) {
		DGPADT_FOREACH (classi, curlist) {
DGP_ERR_PASS_CLEANUP(
		    tokCCheckMembers(classes, tokCClassData(classi), opt));
                }
            }
        }
    }

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

/*
 * Do the final merge of the classes. Well, it's not really "final" since
 * one can get forced at any point by tokMerge...
 *
 * Try to merge all the classes within each bin, then check between bins.
 * Generate synthetic exemplars for each class after every pass as you do
 * this. Also do rejection on each class after generating the exemplars.
 * Iterate until the total number of classes stops decreasing.
 */
ErrCode
tokCTFinalMerge(TokState state, tokClassTable classes)
{
    _tokState *intstate = (_tokState *)state;
    Dgp_State *dgp_state = intstate->options.dgp_state;
    tokOptions *opt = &(intstate->options);
    Dgp_Int32 lastcount;

#ifdef TOKCLASS_TIMING
    struct rusage old_usage, new_usage;
#endif

DGP_PROLOGUE("tokCTFinalMerge()", DGP_PERRSTATE);

#ifdef TOKCLASS_TIMING
    printf( "begin final merge\n");
    printf( "initial \tclasses: %d\n", classes->count);
    
    if (getrusage(RUSAGE_SELF, &old_usage) < 0) {
	printf( "getrusage failed\n");
    }   
#endif

    /* First, do a merge pass */
DGP_ERR_PASS_CLEANUP(
    scanTable(state, classes, SCAN_CHECKALL));
	
#ifdef TOKCLASS_TIMING
    if (getrusage(RUSAGE_SELF, &new_usage) < 0) {
    	printf( "getrusage failed\n");
    } 
    
    printf( "scan all \tclasses: %d \ttime: %6.3f\n", 
    	classes->count, calc_time( new_usage, old_usage));    	
	
    if (getrusage(RUSAGE_SELF, &old_usage) < 0) {
    	printf( "getrusage failed\n");
    }   
#endif
    
    /* Generate the exemplars */
DGP_ERR_PASS_CLEANUP(
    generateAllExemplars(state, classes));

#ifdef TOKCLASS_TIMING
    if (getrusage(RUSAGE_SELF, &new_usage) < 0) {
    	printf( "getrusage failed\n");
    } 
    
    printf( "generate \tclasses: %d \ttime: %6.3f\n", 
    	classes->count, calc_time( new_usage, old_usage)); 
	
    if (getrusage(RUSAGE_SELF, &old_usage) < 0) {
    	printf( "getrusage failed\n");
    }   
#endif

    
    /* and throw stuff out */
DGP_ERR_PASS_CLEANUP(
    checkAllClasses(classes, opt));

#ifdef TOKCLASS_TIMING
    if (getrusage(RUSAGE_SELF, &new_usage) < 0) {
    	printf( "getrusage failed\n");
    } 
    
    printf( "check \t\tclasses: %d \ttime: %6.3f\n", 
    	classes->count, calc_time( new_usage, old_usage)); 
	
    if (getrusage(RUSAGE_SELF, &old_usage) < 0) {
    	printf( "getrusage failed\n");
    }   
#endif

    /*
     * There are now lots of singletons around. Do some special merge passes
     * with them.
     */
    /* Re-generate the exemplars */
DGP_ERR_PASS_CLEANUP(
    generateAllExemplars(state, classes));
	
#ifdef TOKCLASS_TIMING
    if (getrusage(RUSAGE_SELF, &new_usage) < 0) {
    	printf( "getrusage failed\n");
    } 
    
    printf( "generate \tclasses: %d \ttime: %6.3f\n", 
    	classes->count, calc_time( new_usage, old_usage)); 
	
    if (getrusage(RUSAGE_SELF, &old_usage) < 0) {
    	printf( "getrusage failed\n");
    }   
#endif
    
    /* and do a merge pass */
DGP_ERR_PASS_CLEANUP(
    scanTable(state, classes, SCAN_CHECKNOSINGLETONS));
	
#ifdef TOKCLASS_TIMING
    if (getrusage(RUSAGE_SELF, &new_usage) < 0) {
    	printf( "getrusage failed\n");
    } 
    
    printf( "scan nosingle\tclasses: %d \ttime: %6.3f\n", 
    	classes->count, calc_time( new_usage, old_usage)); 
	
    if (getrusage(RUSAGE_SELF, &old_usage) < 0) {
    	printf( "getrusage failed\n");
    }   
#endif

    /* and do a merge pass */
DGP_ERR_PASS_CLEANUP(
    scanTable(state, classes, SCAN_CHECKSINGLETONS));
	
#ifdef TOKCLASS_TIMING
    if (getrusage(RUSAGE_SELF, &new_usage) < 0) {
    	printf( "getrusage failed\n");
    } 
    
    printf( "scan single\tclasses: %d \ttime: %6.3f\n", 
    	classes->count, calc_time( new_usage, old_usage)); 
	
#endif

    /*
     * Here we split and do custom merging if that was specified
     */
    if( opt->mergeCallback != NULL)
    {
    	Dgp_Int32 w, h;
	DgpAdt_List list;
	DgpAdt_ListNode ln;
	tokCClass c;
	
	/* First make sure that all classes are in the correct bins */
	regularise(classes, SCAN_CHECKALL, opt);
	
	/* Now clear all the status bits that are not reserved. */
	for (h = 0; h < DGPIMG_HEIGHT(classes->elts); h++) {
	    for (w = 0; w < DGPIMG_WIDTH(classes->elts); w++) {
	    	list = (DgpAdt_List)tokCTRef( classes, w, h);
	    	DGPADT_FOREACH( ln, list) {
	    	    c = tokCClassData( ln);
    		    c->status &= TOKCLASS_RESERVED_BITS;		
	    	}
	    }
	}

DGP_ERR_PASS_CLEANUP(
	(opt->mergeCallback)(classes, state, opt->callbackCookie));
    }
    else
    {
	/*
	 * Now iterate until it converges.
	 */
	do {
	lastcount = classes->count;
	
#ifdef TOKCLASS_TIMING	
    if (getrusage(RUSAGE_SELF, &old_usage) < 0) {
    	printf( "getrusage failed\n");
    }   
#endif

DGP_ERR_PASS_CLEANUP(	
	scanTable(state, classes, SCAN_CHECKALL));

#ifdef TOKCLASS_TIMING
    if (getrusage(RUSAGE_SELF, &new_usage) < 0) {
    	printf( "getrusage failed\n");
    } 
    
    printf( "scan all \tclasses: %d \ttime: %6.3f\n", 
    	classes->count, calc_time( new_usage, old_usage)); 
	
    if (getrusage(RUSAGE_SELF, &old_usage) < 0) {
    	printf( "getrusage failed\n");
    }   
#endif

DGP_ERR_PASS_CLEANUP(
        generateAllExemplars(state, classes));
	    
#ifdef TOKCLASS_TIMING
    if (getrusage(RUSAGE_SELF, &new_usage) < 0) {
    	printf( "getrusage failed\n");
    } 
    
    printf( "generate \tclasses: %d \ttime: %6.3f\n", 
    	classes->count, calc_time( new_usage, old_usage)); 
	
#endif
	} while (lastcount > classes->count);

	/*
	* Now do some more merges...
	* you'd think you were done, but that's not true in some cases.  Some of
	* the merges might have caused classes to be committed (especially
	* when there were immutable classes), and as part of that, some
	* members might have been cut loose.  When that happens, they're not
	* checked against the bin they land in (this would make infinite
	* recursion a possibility), so we need to do a few merges now.
	*/
	
#ifdef TOKCLASS_TIMING
    if (getrusage(RUSAGE_SELF, &old_usage) < 0) {
    	printf( "getrusage failed\n");
    }   
#endif

DGP_ERR_PASS_CLEANUP(
	scanTable(state, classes, SCAN_CHECKALL));

	
#ifdef TOKCLASS_TIMING
    if (getrusage(RUSAGE_SELF, &new_usage) < 0) {
    	printf( "getrusage failed\n");
    } 
    
    printf( "scan all \tclasses: %d \ttime: %6.3f\n", 
    	classes->count, calc_time( new_usage, old_usage)); 
    printf( "merge complete\n");
	
#endif

	/* It's converged - we're done. */
    }

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

/*
 * Do final alignment on all the classes: find the best shift for each
 * member.
 */
ErrCode
tokCTDoAlign(tokClassTable classes, tokOptions *opt)
{
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_Int32 w, h;
    DgpAdt_ListNode classi;
    DgpAdt_List curlist;

DGP_PROLOGUE("tokCTDoAlign()", DGP_PERRSTATE);

    for (h = 0; h < DGPIMG_HEIGHT(classes->elts); h++) {
	for (w = 0; w < DGPIMG_WIDTH(classes->elts); w++) {
	    curlist = (DgpAdt_List)tokCTRef(classes, w, h);
	    if (curlist != DGP_NULLLIST) {
		DGPADT_FOREACH (classi, curlist) {
DGP_ERR_PASS_CLEANUP(
		    tokCAlignClass(tokCClassData(classi), opt));
                }
            }
        }
    }
    
    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

/*
 * Calculate the order to scan the neighbors, given a w*2+1 by h*2+1 window on 
 * the class table.
 *
 * The order is concentric diamonds outwards from the centre. The returned
 * array consists of alternating (x,y); *num is the number of pairs. The centre
 * point is not included. Only points in the "past" when the table is scanned
 * in reading order (increasing H; within equal H, increasing W) are included.
 *
 * It should only be called with one set of (w,h) values: once you call it
 * with certain values, you must always call it with the same values
 * afterwards.
 *
 */
ErrCode
tokCTCalcOrder(Dgp_Int32 w, Dgp_Int32 h, Dgp_Int32 *num, Dgp_Int32 **_order, tokOptions *opt)
{
    Dgp_State *dgp_state = opt->dgp_state;
    Dgp_Int32 *order;
    Dgp_Int32 dx, dy;
    Dgp_Int32 d;
    Dgp_Int32 i = 0;

DGP_PROLOGUE("tokCTCalcOrder()", DGP_PERRSTATE);

    (*num) = ((w * 2 + 1) * (h * 2 + 1) - 1) / 2 + 1;

DGP_CHECK_ALLOC_CLEANUP(
    order = DGP_MALLOC((*num) * 2 * sizeof(int)));

    /* The centre first */
    order[i++] = 0;
    order[i++] = 0;

    /* Now do the diamonds outwards */
    for (d = 1; d <= w + h; d++) {
	if (d <= w) {
	    order[i++] = -d;
	    order[i++] = 0;
	    }

	for (dx = d - 1; dx > 0; dx--) {
	    dy = d - dx;

	    /* Are we within the window? */
	    if ((ABS(dy) <= h) && (ABS(dx) <= w)) {
		order[i++] = -dx;
		order[i++] = -dy;
		order[i++] = dx;
		order[i++] = -dy;
		}
	    }

	if (d <= h) {
	    order[i++] = 0;
	    order[i++] = -d;
	    }
	}
  
    DGP_ASSERT(i == *num * 2);
    *_order = order;

    NORMAL_RETURN;

ERR_CLEANUP:
    
    ERR_RETURN;
}

/*
 * Scan a bin, merging any things that match according to checkStyle.
 *
 * If checkStyle is SCAN_CHECKALL, check every pair of classes in the bin.
 * If checkStyle is SCAN_CHECKNOSINGLETONS, only check pairs of classes
 *	where at least one class has >1 element.
 * If checkStyle is SCAN_CHECKSINGLETONS, only check pairs of classes where
 *	both have exactly one element.
 * If checkStyle is SCAN_CHECKSYNTH, only check pairs of classes where
 *	at least one class has a synthetic exemplar.
 *
 * Don't bother to compare two classes that are both marked TOKCLASS_MERGECLEAN.
 */
static ErrCode
scanBin(TokState state, tokClassTable classes, DgpAdt_List bin, Dgp_Int32 checkStyle)
{
    tokOptions *opt = &(((_tokState *)state)->options);
    Dgp_State *dgp_state = opt->dgp_state;
    DgpAdt_ListNode ln1, ln2;
    tokCClass a, b;
    Dgp_Bool merged;
    Dgp_Int32 toscan;
    Dgp_Int32 i;

DGP_PROLOGUE("scanBin()", DGP_PERRSTATE);

    i = 0;
    DGPADT_FOREACH (ln1, bin) {
	i++;
	a = tokCClassData(ln1);

	if ((checkStyle == SCAN_CHECKSINGLETONS) &&
	    (DGPADT_LILEN(a->members) > 1)) {
	    continue;
        }

	/*
	 * Try merging everything into class a.
	 * Only try to merge "toscan" elements: this looks at the current
	 * bin size, subtracts off this class's index in the bin, and sees
	 * how many things are currently in the bin.  It then only checks
	 * that many entries in the bin.  In some cases, this might not be
	 * the entire bin: suppose that some stuff gets merged in, the
	 * class gets committed, and some members get rejected back into
	 * this bin.  We don't want to check them against this class again:
	 * they shouldn't match (they got rejected after all), and even if
	 * due to some odd happenstance they do match (it can happen in some
	 * odd cases, like where every member gets rejected), then checking
	 * them again can cause an infinite loop.
	 */
	toscan = DGPADT_LILEN(bin) - i;
	for (ln2 = DGPADT_LINEXT(ln1); toscan > 0; toscan--) {
	    DGP_ASSERT(ln2 != DGP_NULLLISTNODE);
	    b = tokCClassData(ln2);

	    if ((checkStyle == SCAN_CHECKNOSINGLETONS) &&
		(DGPADT_LILEN(a->members) == 1) && (DGPADT_LILEN(b->members) == 1)) {
		ln2 = DGPADT_LINEXT(ln2);
		continue;
            }
	    else if ((checkStyle == SCAN_CHECKSINGLETONS) &&
		(DGPADT_LILEN(b->members) > 1)) {
		ln2 = DGPADT_LINEXT(ln2);
		continue;
            }
	    else if ((checkStyle == SCAN_CHECKSYNTH) &&
		     (!(a->status & TOKCLASS_EXSYNTHETIC)) &&
		     (!(b->status & TOKCLASS_EXSYNTHETIC))) {
		ln2 = DGPADT_LINEXT(ln2);
		continue;
            }

	    if ((a->status & TOKCLASS_MERGECLEAN) &&
		(b->status & TOKCLASS_MERGECLEAN)) {
/*		DGP_WARNF3("NO compare %08lx(%2d,%2d)",
		      a, a->exemplar->width, a->exemplar->height);
		DGP_WARNF3(" with %08lx(%2d,%2d)\n",
		      b, b->exemplar->width, b->exemplar->height); */
		ln2 = DGPADT_LINEXT(ln2);
		continue;
            }

DGP_ERR_CHECK_NO_MSG(
            tokCMaybeMergeClasses(state, classes, a, b, &merged)) {
		/*
		 * A failure down the line. We have to recover...
		 * be might have been merged into a (merged will tell us).
		 */
		if (merged) {
		    DgpAdt_liRem(ln2);
		    classes->count--;
                }
                DGP_ERR_CLEANUP(DGP_CUR_ERROR);
            }
	    if (merged) {
		DgpAdt_liRemAndNext(ln2, &ln2);
		classes->count--;
            }
	    else {
		ln2 = DGPADT_LINEXT(ln2);
            }
        }
    }
    
    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}

/*
 * Scan two bins, merging any things that match according to checkStyle.
 * Merge things from binb into things in bina.
 *
 * If checkStyle is SCAN_CHECKALL, check every pair of classes in the bins.
 * If checkStyle is SCAN_CHECKNOSINGLETONS, only check pairs of classes
 *	where at least one class has >1 element.
 * If checkStyle is SCAN_CHECKSINGLETONS, only check pairs of classes where
 *	both have exactly one element.
 * If checkStyle is SCAN_CHECKSYNTH, only check pairs of classes where
 *	at least one class has a synthetic exemplar.
 */
static ErrCode
scanTwoBins(TokState state, tokClassTable classes, DgpAdt_List bina, DgpAdt_List binb,
            Dgp_Int32 checkStyle)
{
    tokOptions *opt = &(((_tokState *)state)->options);
    Dgp_State *dgp_state = opt->dgp_state;
    DgpAdt_ListNode ln1, ln2;
    tokCClass a, b;
    Dgp_Bool merged;
    Dgp_Int32 toscan;

DGP_PROLOGUE("scanTwoBins()", DGP_PERRSTATE);

    DGPADT_FOREACH (ln1, bina) {
	a = tokCClassData(ln1);

	if ((checkStyle == SCAN_CHECKSINGLETONS) &&
	    (DGPADT_LILEN(a->members) > 1)) {
	    continue;
        }

	/* Try merging everything into class a */
	toscan = DGPADT_LILEN(binb);
	for (ln2 = DGPADT_LIFIRST(binb); toscan > 0; toscan--) {
	    DGP_ASSERT(ln2 != DGP_NULLLISTNODE);
	    b = tokCClassData(ln2);

	    if ((checkStyle == SCAN_CHECKNOSINGLETONS) &&
		(DGPADT_LILEN(a->members) == 1) && (DGPADT_LILEN(b->members) == 1)) {
		ln2 = DGPADT_LINEXT(ln2);
		continue;
            }
	    else if ((checkStyle == SCAN_CHECKSINGLETONS) &&
		     (DGPADT_LILEN(b->members) > 1)) {
		ln2 = DGPADT_LINEXT(ln2);
		continue;
            }
	    else if ((checkStyle == SCAN_CHECKSYNTH) &&
		     (!(a->status & TOKCLASS_EXSYNTHETIC)) &&
		     (!(b->status & TOKCLASS_EXSYNTHETIC))) {
		ln2 = DGPADT_LINEXT(ln2);
		continue;
            }

	    if ((a->status & TOKCLASS_MERGECLEAN) &&
		(b->status & TOKCLASS_MERGECLEAN)) {
/*		DGP_WARNF3("NO compare %08lx(%2d,%2d) ",
		      a, a->exemplar->width, a->exemplar->height);
		DGP_WARNF3("with %08lx(%2d,%2d)\n",
		      b, b->exemplar->width, b->exemplar->height); */
		ln2 = DGPADT_LINEXT(ln2);
		continue;
            }

DGP_ERR_CHECK_NO_MSG(
            tokCMaybeMergeClasses(state, classes, a, b, &merged)) {
		/*
		 * A failure down the line. We have to recover...
		 * b might have been merged into a (merged will tell us).
		 */
		if (merged) {
		    DgpAdt_liRem(ln2);
		    classes->count--;
                }
                DGP_ERR_CLEANUP(DGP_CUR_ERROR);
            }
	    if (merged) {
		DgpAdt_liRemAndNext(ln2, &ln2);
		classes->count--;
            }
	    else {
		ln2 = DGPADT_LINEXT(ln2);
            }
        }
    }
    
    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}
