/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/cross.c,v 1.28 2003/01/27 21:48:41 awilliam Exp $
 *
 */

/*----------------------------------------------------------------------------
			   Michael Leventon
			   PARC Summer 1994
----------------------------------------------------------------------------*/

#include <stdlib.h>
#include "tok-lossy.h"
#include "tokprivate.h"
#include "bitfiddle.h"
#include "miscalloc.h"
#include "bitcounts.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/cross.c,v 1.28 2003/01/27 21:48:41 awilliam Exp $");

#ifdef	DBUG
#include "dbug.h"
extern Dgp_Int32 curimbitmap, maximbitmap;
#endif

/* Array used to make bit counting in a Dgp_Int32 word FAST.  The number of bits
   set in l is bitcounts[HIGH_PART(l)] + bitcounts[LOW_PART(l)]. */
#define HIGH_PART(l) ((l) >> (DGP_BITS_UINT32 / 2))             /* hi half of l */
#define LOW_PART(l)  ((l) & ((1 << (DGP_BITS_UINT32 / 2)) - 1)) /* low half */
							   /* of l */

/* Stuff in dilatelookup.c */
extern Dgp_UInt8 tpdilateLookup[1<<13];
extern Dgp_UInt8 rad1dilateLookup[1<<13];
extern Dgp_UInt8 radroot2dilateLookup[1<<13];

/*
 * Useful things to know:
 * Bits within a word are stored MSB-is-left (lowest x),
 * LSB-is-right (highest x). The leftmost bit is therefore Bit(31), but
 * x == 0 (mod 32).
 *
 * The dilation lookup table is accessed as
 * dilatelookup[env] - 0 or 1 means that is the output (we're actually
 * inverting as we dilate, so the values in the table might be flipped
 * from what you're expecting). 2 means that the output is 0, but that the
 * original pixel was on, and should be classified as an "edge pixel".
 * The env index into the table is made up of the current bit and 12 bits
 * surrounding it, packed into a 13-bit value. The locations of the bits
 * in this value are:
 *         0
 *     1   2   3
 * 4   5   6   7   8
 *     9  10  11
 *        12
 * so the bit two rows above the current bit gives the LSB value of env, etc.
 */

/*
 * Some macro definitions to cover up the bit ordering shown above.
 * The bits listed here are the "leading edge" of the mask (see DoBit),
 * and a couple more that show up various places.
 */
#define	BIT_NN	0
#define	BIT_NW	1
#define	BIT_WW	4
#define	BIT_W	5
#define	BIT_SW	9
#define	BIT_SS	12

/*
 * A macro for general bit-within-uDgp_Int32 access. Numbering is LSB==0.
 */
#define Bit(b)     (1ul << (b))

/*
 * This macro handles the lookup. bits holds the current environment.
 * m_uu holds the Dgp_Int32word two rows up from the current position; m_u is the
 * Dgp_Int32word above; m, m_d and m_dd are what you'd expect.
 * outbits holds the result (the dilated inverted result Dgp_Int32word), and is
 * built LSB-to-MSB (i.e. right to left) by filling in the MSB and shifting
 * right.
 * bits is assumed to be set up correctly for the position we've just finished
 * with. Shifting it left is equivalent to moving the current position left.
 * (see the bit numbering above), except for bits 0, 1, 4, 9, 12, which
 * need to be reloaded.
 */
#define DoBit(b)							\
    bits <<= 1;								\
    bits  &= 0x1FFF & ~(Bit(BIT_NN) | Bit(BIT_NW) | Bit(BIT_WW) | Bit(BIT_SW) | Bit(BIT_SS));           \
    bits  |= MoveAndMask(m_uu, b,   BIT_NN);				\
    bits  |= MoveAndMask(m_u,  b-1, BIT_NW);				\
    bits  |= MoveAndMask(m,    b-2, BIT_WW);				\
    bits  |= MoveAndMask(m_d,  b-1, BIT_SW);				\
    bits  |= MoveAndMask(m_dd, b,   BIT_SS);				\
    outbits |= lookup[bits] << 31;					\
    *edgecount += lookup[bits] >> 1;					\
    if (b>0) outbits >>= 1;

/*
 * This is basically DoBit(1), except that we don't load up BIT_WW;
 * it's used at the left end of the row.
 */
#define DoBit1_zeros							\
    bits <<= 1;								\
    bits  &= 0x1FFF & ~(Bit(BIT_NN) | Bit(BIT_NW) | Bit(BIT_WW) | Bit(BIT_SW) | Bit(BIT_SS));           \
    bits  |= MoveAndMask(m_uu, 1, BIT_NN);				\
    bits  |= MoveAndMask(m_u,  0, BIT_NW);				\
    bits  |= MoveAndMask(m_d,  0, BIT_SW);				\
    bits  |= MoveAndMask(m_dd, 1, BIT_SS);				\
    outbits |= lookup[bits] << 31;					\
    *edgecount += lookup[bits] >> 1;					\
    outbits >>= 1;

/*
 * This is basically DoBit(0), except that we don't load up BIT_*W;
 * it's used at the left end of the row.
 */
#define DoBit0_zeros							\
    bits <<= 1;								\
    bits  &= 0x1FFF & ~(Bit(BIT_NN) | Bit(BIT_NW) | Bit(BIT_WW) | Bit(BIT_SW) | Bit(BIT_SS));           \
    bits  |= MoveAndMask(m_uu, 0, BIT_NN);				\
    bits  |= MoveAndMask(m_dd, 0, BIT_SS);				\
    outbits |= lookup[bits] << 31;					\
    *edgecount += lookup[bits] >> 1;

/*
 * This macro starts up the shift process as we enter a row from the right.
 * It loads in the first (leftmost) two columns of the environment,
 * generates a bit in outbits, and continues. Special cases are handled
 * separately. Note that this can't generate an edgecount since the centre
 * bit is never set, and edge pixels are based on that.
 */
#define SetupBit(b)							\
    bits  = MoveAndMask(m,   b - 2, BIT_WW);				\
    bits |= MoveAndMask(m,   b - 1, BIT_W);				\
    bits |= MoveAndMask(m_u, b - 1, BIT_NW);				\
    bits |= MoveAndMask(m_d, b - 1, BIT_SW);				\
    outbits |= lookup[bits] << 31;					\
    outbits >>= 1;

/*
 * This routine handles a row of the dilation. The stuff it needs set up:
 * wordsleft: number of Dgp_Int32words in each model row.
 * bitsinlastword: number of valid bits in the last Dgp_Int32word. 0 means 32.
 *	If this is 0, then the row pointed at by ip needs to be 1 Dgp_Int32er
 *	than wordsleft.
 *
 * mp_* need to be loaded to point to the word *just past* the
 * last Dgp_Int32word in the appropriate row of the model; ditto ip.
 *
 * mp_* can be set to NULL, indicating that that row of the model doesn't
 * exist and should be treated as zero.
 *
 * 12.4.2001 / dhtesq.
 * Profiling shows this function eating 8% of the execution time for a one page encode,
 * good for second place.  Lots of hits, but inlining doesn't help much.
 *
 */
static void
do_row(Dgp_UInt32 *mp_uu, Dgp_UInt32 *mp_u, Dgp_UInt32 *mp, Dgp_UInt32 *mp_d, Dgp_UInt32 *mp_dd,
       Dgp_UInt32 *ip, Dgp_Int32 wordsleft, Dgp_Int32 bitsinlastword, Dgp_Int32 *edgecount,
       const Dgp_UInt8 *lookup)
{
    Dgp_UInt32 m_uu, m_u, m, m_d, m_dd;
    Dgp_UInt32 outbits;			/* The output word we're building */
    Dgp_Int32 bits = 0;

    /* Load up the last word. */
    m = *--mp;

    m_uu = m_u = m_d = m_dd = 0;

    if (mp_uu != (Dgp_UInt32 *)NULL) {
	m_uu = *--mp_uu;
	}

    if (mp_u != (Dgp_UInt32 *)NULL) {
	m_u = *--mp_u;
	}

    if (mp_d != (Dgp_UInt32 *)NULL) {
	m_d = *--mp_d;
	}

    if (mp_dd != (Dgp_UInt32 *)NULL) {
	m_dd = *--mp_dd;
	}

    /*
     * OK - we're going to set wordsleft to the number of words we haven't
     * touched yet (we're about to touch the last word, so don't
     * count it).
     * Load up outbits with 0s so that the rightmost word is padded
     * with 0s.
     */
    wordsleft--;

    /* Clear the output buffer */
    outbits = 0;

    /*
     * This switch loads up "bits" properly. It generates the
     * "extra" bit in the image (the rightmost one), and loads up the
     * environment with *some* of the bits:    .
     *                                       X . .
     *                                     X X . . .
     *                                       X . .
     *                                         .
     * with the other ones (".") set to zero. Why those? So that the
     * first DoBit shifts appropriately to start on the second
     * (from right) image bit (same position as the rightmost model
     * bit).
     */
    switch (bitsinlastword) {
      case 0: {
	/*
	 * 0 really means 32 bits in the last word: it's full.
	 * We want to set up the state as though we'd just done the bit
	 * to the right of the last word (which we'll do now).
	 */
	bits  = MoveAndMask(m,   30, BIT_WW);
	bits |= MoveAndMask(m,   31, BIT_W);
	bits |= MoveAndMask(m_u, 31, BIT_NW);
	bits |= MoveAndMask(m_d, 31, BIT_SW);

	/*
	 * A full last word in the model means an extra word
	 * in the image.
	 */
	/*assert (irw == mrw + 1);*/
	/*
	 * Copy the last bit of the model into the last bit of the
	 * image. This is the dilation of the rightmost bit of the
	 * model.
	 */
	*--ip = (lookup[bits] << 31);

	/* Correct for the 0/32 confusion */
	wordsleft++;
	break;
	}

      case 1: {
	/*
	 * Only one bit left in the rightmost model word.
	 * Load up the bits as required. We need to start on the
	 * next word. There had better be a next word...
	 */
	/*assert(wordsleft);*/

	bits  = MoveAndMask(m,   0, BIT_W);
	bits |= MoveAndMask(m_u, 0, BIT_NW);
	bits |= MoveAndMask(m_d, 0, BIT_SW);

	m = *--mp;
	bits |= MoveAndMask(m,   31, BIT_WW);

	/*
	 * Do the dilation for the rightmost image bit.
	 */
	outbits |= lookup[bits] << 31;
	outbits >>= 1;
	break;
	}

      /*
       * All the remaining cases are simple. Yay. Do a SetupBit on each
       * one, which will initialise bits and generate the first bit
       * of outbits.
       */
      case 2:  SetupBit(2);  break;
      case 3:  SetupBit(3);  break;
      case 4:  SetupBit(4);  break;
      case 5:  SetupBit(5);  break;
      case 6:  SetupBit(6);  break;
      case 7:  SetupBit(7);  break;
      case 8:  SetupBit(8);  break;
      case 9:  SetupBit(9);  break;
      case 10: SetupBit(10); break;
      case 11: SetupBit(11); break;
      case 12: SetupBit(12); break;
      case 13: SetupBit(13); break;
      case 14: SetupBit(14); break;
      case 15: SetupBit(15); break;
      case 16: SetupBit(16); break;
      case 17: SetupBit(17); break;
      case 18: SetupBit(18); break;
      case 19: SetupBit(19); break;
      case 20: SetupBit(20); break;
      case 21: SetupBit(21); break;
      case 22: SetupBit(22); break;
      case 23: SetupBit(23); break;
      case 24: SetupBit(24); break;
      case 25: SetupBit(25); break;
      case 26: SetupBit(26); break;
      case 27: SetupBit(27); break;
      case 28: SetupBit(28); break;
      case 29: SetupBit(29); break;
      case 30: SetupBit(30); break;
      case 31: SetupBit(31); break;
      }

    /* Now finish off the first word. */
    switch (bitsinlastword)  {
      /* Note that these all fall through. */

      case 31: DoBit(30);
      case 30: DoBit(29);
      case 29: DoBit(28);
      case 28: DoBit(27);
      case 27: DoBit(26);
      case 26: DoBit(25);
      case 25: DoBit(24);
      case 24: DoBit(23);
      case 23: DoBit(22);
      case 22: DoBit(21);
      case 21: DoBit(20);
      case 20: DoBit(19);
      case 19: DoBit(18);
      case 18: DoBit(17);
      case 17: DoBit(16);
      case 16: DoBit(15);
      case 15: DoBit(14);
      case 14: DoBit(13);
      case 13: DoBit(12);
      case 12: DoBit(11);
      case 11: DoBit(10);
      case 10: DoBit(9);
      case 9:  DoBit(8);
      case 8:  DoBit(7);
      case 7:  DoBit(6);
      case 6:  DoBit(5);
      case 5:  DoBit(4);
      case 4:  DoBit(3);
      case 3:  DoBit(2);
      case 2: {
	/*
	 * Need to reload m.
	 */
	if (wordsleft) {
	    m = *--mp;
	    DoBit(1);
	    }
	else {
	    DoBit1_zeros;
	    }
	}

      case 1: {
	/*
	 * Need to reload m_u and m_d
	 */
	if (wordsleft) {
	    if (mp_u != (Dgp_UInt32 *)NULL) {
		m_u = *--mp_u;
		}
	    if (mp_d != (Dgp_UInt32 *)NULL) {
		m_d = *--mp_d;
		}
	    DoBit(0);
	    if (mp_uu != (Dgp_UInt32 *)NULL) {
		m_uu = *--mp_uu;
		}
	    if (mp_dd != (Dgp_UInt32 *)NULL) {
		m_dd = *--mp_dd;
		}
	    }
	else {
	    DoBit0_zeros;
	    }
	*--ip = outbits;
	}
	/*
	 * Note that there's no clause for 0. This switch is just
	 * to handle partial words, and there aren't any there.
	 */
	}

    /*
     * OK - into the main part.
     * The state here is that we have wordsleft words left to do.
     * The m* variables are all loaded up
     * properly to start a new word, as is bits.
     */
    for (; wordsleft; wordsleft--) {
	outbits = 0;

	/* Do all the easy stuff. */
	DoBit(31); DoBit(30); DoBit(29); DoBit(28); DoBit(27);
	DoBit(26); DoBit(25); DoBit(24); DoBit(23); DoBit(22);
	DoBit(21); DoBit(20); DoBit(19); DoBit(18); DoBit(17);
	DoBit(16); DoBit(15); DoBit(14); DoBit(13); DoBit(12);
	DoBit(11); DoBit(10); DoBit(9);  DoBit(8);  DoBit(7);
	DoBit(6);  DoBit(5);  DoBit(4);  DoBit(3);  DoBit(2);

	/* Start reloading, if possible */
	if (wordsleft > 1) {
	    /* Need to reload m 2 before the end of the word, for WW */
	    m = *--mp;
	    DoBit(1);

	    /* m_u and m_d need be reloaded 1 early, for NW and SW */
	    if (mp_u != (Dgp_UInt32 *)NULL) {
		m_u = *--mp_u;
		}
	    if (mp_d != (Dgp_UInt32 *)NULL) {
		m_d = *--mp_d;
		}
	    DoBit(0);

	    /* and finally m_uu and m_dd for the next word */
	    if (mp_uu != (Dgp_UInt32 *)NULL) {
		m_uu = *--mp_uu;
		}
	    if (mp_dd != (Dgp_UInt32 *)NULL) {
		m_dd = *--mp_dd;
		}
	    }
	else {
	    DoBit1_zeros;
	    DoBit0_zeros;
	    }
	*--ip = outbits;
	}
    }


/*  
 * The following code implements the fast cross-dilate, according
 *  to the rules specified in the lookup table (see makedil.c).
 *
 * It gets fed a model (input) and an image (output) to put the results into.
 * The image is two rows taller than the model (one on top, one on bottom),
 * and *one* bit wider than the model (one bit on the right); it's not any
 * wider on the left because then to compare the two, we'd need to shift
 * one or the other.
 *
 * It stores the number of edge points in the model in *edgecount.
 */
DGP_LIBEXPORT void
TableCrossDilate(Dgp_UInt32 *inputbits, Dgp_Int32 input_row_width, Dgp_Int32 input_width,
		 Dgp_Int32 input_height, Dgp_Int32 *edgecount,
		 Dgp_UInt32 *outputbits, Dgp_Int32 output_row_width,
		 const Dgp_UInt8 *lookup, tokOptions *opt)
{
    Dgp_UInt32 *ip;					/* Output pointer */
    Dgp_UInt32 *mp_uu, *mp_u, *mp, *mp_d, *mp_dd;	/* Input pointers */
    const Dgp_Int32 mrw = input_row_width;
    const Dgp_Int32 irw = output_row_width;
    const Dgp_UInt8 bitsinlastword = input_width & 31;
    Dgp_Int32 mrowused = (input_width + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
    Dgp_Int32 irowused = (input_width + 1 + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
    Dgp_UInt32 *mpend;
    Dgp_UInt32 left;

DGP_VOID_PROLOGUE( "TableCrossDilate()", opt->DGP_PERRSTATE);

    *edgecount = 0;

    /* Please don't hand me anything I haven't figured out how to deal with. */
    DGP_VOID_ASSERT(input_width >= 2);

    /*
     * do_row wants pointers that start just beyond the end of the rows.
     * Since there might be slop at the ends of the rows (mrw might be
     * larger than mrowused, for example), we position the pointers at
     * the first word of slop (if any), or beyond the end of the row (if there
     * isn't any slop).
     */

    if (input_height < 4) {
	/*
	 * Sigh - tiny model means it's all edge cases. Handle them.
	 */

	switch (input_height) {
	  case 3: {
	    mp_uu = inputbits + mrowused;
	    mp_u = mp_uu + mrw;
	    mp = mp_d = mp_dd = mp_u + mrw;
	    ip = outputbits + 3 * irw + irowused;
	    
	    /* Bottom row */
	    do_row(mp_uu, mp_u, mp, (Dgp_UInt32 *)NULL, (Dgp_UInt32 *)NULL,
		   ip, mrowused, bitsinlastword, edgecount, lookup);
	    mp_u -= mrw;
	    mp -= mrw;
	    ip -= irw;

	    /* Middle row; row below exists, but row 2 up doesn't. */
	    do_row((Dgp_UInt32 *)NULL, mp_u, mp, mp_d, (Dgp_UInt32 *)NULL,
		   ip, mrowused, bitsinlastword, edgecount, lookup);
	    mp -= mrw;
	    mp_d -= mrw;
	    ip -= irw;

	    /* Top row */
	    do_row((Dgp_UInt32 *)NULL, (Dgp_UInt32 *)NULL, mp, mp_d, mp_dd,
		   ip, mrowused, bitsinlastword, edgecount, lookup);
	    break;
	    }

	  case 2: {
	    mp_u = inputbits + mrowused;
	    mp = mp_d = mp_u + mrw;
	    ip = outputbits + 2 * irw + irowused;
	    
	    /* Bottom row */
	    do_row((Dgp_UInt32 *)NULL, mp_u, mp, (Dgp_UInt32 *)NULL, (Dgp_UInt32 *)NULL,
		   ip, mrowused, bitsinlastword, edgecount, lookup);
	    mp -= mrw;
	    ip -= irw;

	    /* Top row */
	    do_row((Dgp_UInt32 *)NULL, (Dgp_UInt32 *)NULL, mp, mp_d, (Dgp_UInt32 *)NULL,
		   ip, mrowused, bitsinlastword, edgecount, lookup);
	    break;
	    }

	  case 1: {
	    mp = inputbits + mrowused;
	    ip = outputbits + irw + irowused;
	    do_row((Dgp_UInt32 *)NULL, (Dgp_UInt32 *)NULL, mp, (Dgp_UInt32 *)NULL, (Dgp_UInt32 *)NULL,
		   ip, mrowused, bitsinlastword, edgecount, lookup);
	    break;
	    }

	  }
	}
    else {
	/* General case for a large enough image. */

	/*
	 * Start up the pointers at the very end; they'll come into
	 * play staggered.
	 */
	mp_uu = inputbits + (input_height - 3) * mrw + mrowused;
	mp_u = mp_uu + mrw;
	mp = mp_d = mp_dd = mp_u + mrw;
	ip = outputbits + input_height * irw + irowused;

	/* Last row; rows below do not exist. */
	do_row(mp_uu, mp_u, mp, (Dgp_UInt32 *)NULL, (Dgp_UInt32 *)NULL,
	       ip, mrowused, bitsinlastword, edgecount, lookup);
	mp_uu -= mrw;
	mp_u -= mrw;
	mp -= mrw;
	ip -= irw;

	/* Second last row; row below exists. */
	do_row(mp_uu, mp_u, mp, mp_d, (Dgp_UInt32 *)NULL,
	       ip, mrowused, bitsinlastword, edgecount, lookup);
	mp_uu -= mrw;
	mp_u -= mrw;
	mp -= mrw;
	mp_d -= mrw;
	ip -= irw;

	/*
	 * Middle rows; all rows exist. We're done when the uu pointer
	 * is just about to run off the image.
	 */
	while (mp_uu > inputbits) {
	    do_row(mp_uu, mp_u, mp, mp_d, mp_dd,
		   ip, mrowused, bitsinlastword, edgecount, lookup);
	    mp_uu -= mrw;
	    mp_u -= mrw;
	    mp -= mrw;
	    mp_d -= mrw;
	    mp_dd -= mrw;
	    ip -= irw;
	    }

	/* Second from top row - row two above does not exist. */
	do_row((Dgp_UInt32 *)NULL, mp_u, mp, mp_d, mp_dd,
	       ip, mrowused, bitsinlastword, edgecount, lookup);
	mp -= mrw;
	mp_d -= mrw;
	mp_dd -= mrw;
	ip -= irw;

	/* Top row - row above doesn't exist either */
	do_row((Dgp_UInt32 *)NULL, (Dgp_UInt32 *)NULL, mp, mp_d, mp_dd,
	       ip, mrowused, bitsinlastword, edgecount, lookup);
	}

    /*
     * Do extra row at the top of the image.
     * Basically, just make a copy of the top row of the model.
     * This is equivalent to dilating the top row up one.
     * EXCEPTION: Look at lookup to see if it's equal to
     * radroot2dilateLookup. If it is, we have to do something more
     * interesting.
     */
    if (lookup != radroot2dilateLookup) {
	for (mp = inputbits, ip = outputbits, mpend = mp + mrowused;
	     mp != mpend; ip++, mp++) {
	    *ip = *mp;
	    }
	if (irowused != mrowused) {
	    /* Do the extra word. */
	    *ip = 0;
	    }

	/* Do extra bottom row */
	for (mp = inputbits + mrw * (input_height - 1), 
	     ip = outputbits + irw * (input_height + 1), 
	     mpend = mp + mrowused;
	     mp != mpend; ip++, mp++) {
	    *ip = *mp;
	    }
	if (irowused != mrowused) {
	    *ip = 0;
	    }
	}
    else {
	/* The top row */
	left = 0;
	for (mp = inputbits, ip = outputbits, mpend = mp + mrowused;
	     mp != mpend; mp++, ip++) {
	    *ip = left | (*mp) | ((*mp) << 1) | ((*mp) >> 1);
	    left = *mp << (DGP_BITS_UINT32 - 1);
	    }
	if (irowused != mrowused) {
	    /* Do the extra word. */
	    *ip = left;
	    }

	left = 0;
	for (mp = inputbits + mrw * (input_height - 1), 
	     ip = outputbits + irw * (input_height + 1), 
	     mpend = mp + mrowused;
	     mp != mpend; mp++, ip++) {
	    *ip = left | (*mp) | ((*mp) << 1) | ((*mp) >> 1);
	    left = *mp << (DGP_BITS_UINT32 - 1);
	    }
	if (irowused != mrowused) {
	    /* Do the extra word. */
	    *ip = left;
	    }
	}
    }

/*
 * Count how many pixels are turned on in the image. While we're doing this,
 * we invert the image since that's how it is most useful.
 */
static Dgp_Int32
countAndInvertImagePixels(tokCImageInfo *image)
{
    Dgp_Int32 loop;
    Dgp_Int32 count = 0;
    Dgp_Int32 area = image->ysize * image->row_width;

    for (loop = 0; loop < area; loop++) {
	count += bitcounts[HIGH_PART(image->bits[loop])] +
	    bitcounts[LOW_PART(image->bits[loop])];  
	image->bits[loop] = ~(image->bits[loop]);
	}
  
    return(count);
    }

/*
 * Generate the dilated version of the model. Use the TP-dilate. Actually,
 * this produces an inverted version of the dilated model.
 */
ErrCode
MakeImCrossFromMod(tokCModelInfo *model, tokOptions *opt, 
                   tokCImageInfo ** result)
{
    Dgp_State *dgp_state = opt->dgp_state;
    tokCImageInfo *image = (tokCImageInfo *)NULL;
    Dgp_Int32 radius = opt->mockRadius;	/* Use the one with the indicators */
    Dgp_UInt32 *buffer = (Dgp_UInt32 *)NULL;
    Dgp_Int32 i;
    Dgp_UInt32 *p, *q;
    Dgp_Int32 x, y;
    Dgp_Int32 edgecount;
    Dgp_UInt8 *lookup = tpdilateLookup;

  DGP_PROLOGUE("MakeImCrossFromMod()", DGP_PERRSTATE);

  TOKIMG_SEC_MALLOC(&image);

    if (opt->flags & TOK_OPT_NONTPDIL) {
	/* Oh boy - a regular dilate. */
	lookup = rad1dilateLookup;
	if (radius == 10) {
	    radius = 1;
        }
	else if (radius == 20) {
	    radius = 2;
        }
	else if (radius == 14) {
	    radius = 1;
	    lookup = radroot2dilateLookup;
        }
    }

	/*
	 * 11 April 2001 / dhtesq.
	 * Assessing strategies for dealing with Digipaper's small token errors.
	 * WR's xshift, xnewslope, etc. parameters apply after dilation, so frequently
	 * permit matches even when requiring there be no (0 pixels) mismatches.
	 *
	 * Try setting the dilation radius to zero for tokens below a size threshold.
	 */
	/*
	if(model->xsize < opt->lowerDilationLimit && model->ysize < opt->lowerDilationLimit){
		radius = 0;
	}
	*/
    /*
     * Make the image be just like the model, except it has a radius pixel
     * border on the top, bottom, and right sides (the model should already
     * have had a radius pixel border put in on the left).
     */
    image->xsize = model->xsize + radius;
    image->ysize = model->ysize + 2*radius;
    image->row_width = (image->xsize + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;
    image->npts = 0;
  
DGP_CHECK_ALLOC_CLEANUP(
    image->bits = (Dgp_UInt32 *)(DGP_MALLOC)(image->row_width *
					       image->ysize * sizeof(Dgp_UInt32)));

#ifdef	DBUG
    curimbitmap += image->row_width * image->ysize * sizeof(Dgp_UInt32);
    if (curimbitmap > maximbitmap) {
	maximbitmap = curimbitmap;
	DEBUG("image bitmap memory peak %d bytes, brk %d\n", maximbitmap, 
              sbrk(0));
    }
#endif

    if (radius > 0) {
DGP_CHECK_ALLOC_CLEANUP(
	buffer = (Dgp_UInt32 *)(DGP_MALLOC)(image->row_width *
					      image->ysize * sizeof(Dgp_UInt32)));
	
	/*
	 * We're going to call TableCrossDilate a number of times (radius
	 * times, to be exact), pingponging betweek image->bits and buffer.
	 * We want the final result to end up in image->bits.
	 */
	if (radius & 1) {
	    /* Make the first copy into buffer */
	    for (y = 0; y < model->ysize; y++) {
		for (x = 0, p = model->bits + y * model->row_width,
		     q = buffer + y * image->row_width;
		     x < model->row_width;
		     x++) {
		    *q++ = *p++;
                }
            }
        }
	else {
	    /* Make the first copy into image->bits */
	    for (y = 0; y < model->ysize; y++) {
		for (x = 0, p = model->bits + y * model->row_width,
		     q = image->bits + y * image->row_width;
		     x < model->row_width;
		     x++) {
		    *q++ = *p++;
                }
            }
        }
        
	/* Now do the dilations */
	for (i = 0; i < radius; i++) {
	    if ((radius - i) & 1) {
		/* Odd passes: dilate buffer into image->bits */
		TableCrossDilate(buffer, image->row_width, model->xsize + i,
				 model->ysize + 2 * i, &edgecount,
				 image->bits, image->row_width, lookup, opt);
            }
	    else {
		/* Even passes: dilate image->bits into buffer */
		TableCrossDilate(image->bits, image->row_width,
				 model->xsize + i,
				 model->ysize + 2 * i, &edgecount,
				 buffer, image->row_width, lookup, opt);
            }

	    if (i == 0) {
		/* We just counted the model's edge pixels */
		model->nedgepts = edgecount;
            }
        }
    }
    else {
	Dgp_Int32 fake_row_size, fake_width, fake_height;
	DGP_ASSERT(radius == 0);

	/*
	 * We just need to copy it over; the inversion happens below.
	 * However, we also need to determine the number of edge points.
	 * The easiest way to do this is to run it through a dilation.
	 * We'll use the straight radius-1 dilation.  Of course,
	 * to do this, we need to have a little extra slop room.
	 * The general dilation machinery *might* handle everything
	 * properly if I didn't allocate this slop, but it's been long
	 * enough since I wrote it that I don't trust my current
	 * reading of the code.
	 */
	fake_width = model->xsize + 1;
	fake_height = model->ysize + 2;
	fake_row_size = (fake_width + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32;

	DGP_CHECK_ALLOC_CLEANUP(buffer = (Dgp_UInt32 *)(DGP_MALLOC)(fake_row_size * fake_height * sizeof(Dgp_UInt32)));

	/* Feed in a width of at least 2 for the model or it gets unhappy. */
	TableCrossDilate(model->bits, model->row_width,
			 MAX(2, model->xsize),
			 model->ysize, &edgecount,
			 buffer, fake_row_size, rad1dilateLookup, opt);
	/* All that just for a side effect... */

	model->nedgepts = edgecount;

	/* Copy it over */
	(void)memcpy(image->bits, model->bits,
		     model->row_width * model->ysize * sizeof(Dgp_UInt32));
	}

    if (buffer != (Dgp_UInt32 *)NULL) {
	DGP_FREE((void *)buffer);
    }

    image->npts = countAndInvertImagePixels(image);
    *result = image;

    NORMAL_RETURN;

ERR_CLEANUP:

    if (image != (tokCImageInfo *)NULL) {
	TOKIMG_FREE(image);
    }

    if (buffer != (Dgp_UInt32 *)NULL) {
	DGP_FREE((void *)buffer);
    }

    ERR_RETURN;
}
