/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/hauschex.c,v 1.33 2003/02/10 22:53:41 mkrocken Exp $
 *
 */

/*
 * Originally written by Michael Leventon, PARC Summer 1994.
 * Substantially hacked since then by wjr.
 */

#include "tokprivate.h"
#include "hauschex.h"
#include "bitfiddle.h"
#include "bitcounts.h"
#if !defined(VXWORKS) && !defined(EQUATOR_CA_ENV)
#include <memory.h>
#endif  /* (VXWORKS) */

#ifdef	_WIN32
#include <malloc.h>
#elif defined(VXWORKS) || defined(EQUATOR_CA_ENV)
#else
#include <alloca.h>
#endif  /* (_WINDOWS) */

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/hauschex.c,v 1.33 2003/02/10 22:53:41 mkrocken Exp $");

/* Give me the b leftmost bits of l */
#define	TOPBITS(l, b)	((l) & ((Dgp_UInt32)-(Dgp_Int32)(1ul << (DGP_BITS_UINT32 - (b)))))

/* Break down a 32-bit word into two 16-bit lumps, for bit counting */
#define HIGH_PART(l) ((l) >> (DGP_BITS_UINT32 / 2))             /* hi half of l */
#define LOW_PART(l)  ((l) & ((1 << (DGP_BITS_UINT32 / 2)) - 1)) /* low half */
							   /* of l */

static Dgp_Int32 ScanMismatchImage(Dgp_UInt32* tmp, Dgp_Int32 width, Dgp_Int32 xsize, Dgp_Int32 ysize);

/*
 * This function returns the minimum number of edge pixels necessary to
 * allow a one pixel mismatch.  This is used when determining whether
 * or not to shift the model wrt the image in the final merging pass.
 */
Dgp_Int32
MinEdgePixForMismatch(tokOptions *p)
{
    return((Dgp_Int32)((1 + p->xshift) * (1.0 / (1.0 - p->fracHausd))));
}

/*
 * This is the function that tells us how many mismatched pixels are allowed,
 * given the fracHausd parameter and the number of edge pixels in the model. 
 * Right now, it is a piecewise linear function.  It stays zero for a while
 * (determined by xshift), then changes to slope 1.  Then (at xslopechange),
 * it changes to slope newslope. It isn't allowed to exceed
 * (1-fracHausd)*nedgepts (what it would have been if there was no nonlinear
 * initial segment to the curve).
 */
static Dgp_Int32
MaxMismatchedAllowed(Dgp_Int32 nedgepts, tokOptions *p)
{
    Dgp_Int32 temp, temp2;
  
    temp = (Dgp_Int32)((1 - p->fracHausd) * nedgepts);  
  
    if (temp - p->xshift < 0) {
	return(0);
    }

    if (temp < p->xslopechange) {
	return(temp - p->xshift);
    }
  
    temp2 = (Dgp_Int32)((p->xslopechange - p->xshift) +	(p->newslope * (temp - p->xslopechange)));

    return(MIN(temp, temp2));
}

/*
 * This function returns the max number of mismatched pixels in any 3x3 
 * window of the "mismatch image", or the image created by anding model
 * with ~image. The model is shifted by (tx,ty) wrt the image; ty has
 * already been adjusted for the implicit offset of the image.
 *
 * It builds a bitmap (the same size as the model, since that bounds where
 * the bits can be set) of error pixels (in model but not image), then
 * scans that bitmap.
 */
static Dgp_Int32
GetMismatchDensity(tokCImageInfo *image, tokCModelInfo *model, Dgp_Int32 tx, Dgp_Int32 ty,
		   tokOptions *opt)
{
    Dgp_Int32 maxdensity;
    Dgp_Int32 modelsize;
    Dgp_Int32 miny, maxy;
    Dgp_Int32 minrw;
    Dgp_UInt32 *ip, *mp, *tmp;
    Dgp_UInt32 *irow, *mrow, *trow;
    Dgp_UInt32 left, cur, val;
    Dgp_Int32 mrw, irw;
    Dgp_Int32 adjtx;
    Dgp_Int32 x, y;

DGP_VOID_PROLOGUE( "GetMismatchDensity()", opt->DGP_PERRSTATE);

    /* How much memory? */
    modelsize = model->ysize * model->row_width;

    /* Adjust tx,ty so that they represent a shift of image wrt model */
    tx = -tx;
    ty = -ty;

    /* Set up some stuff */
    irw = image->row_width;
    mrw = model->row_width;
    if (tx < 0) {
	DGP_VOID_ASSERT(tx > -DGP_BITS_UINT32);  /* Don't handle this case yet */
	/* Discount the image word lying partially off the left */
	minrw = MIN(irw - 1, mrw);
	adjtx = tx + DGP_BITS_UINT32;
    }
    else {
	DGP_VOID_ASSERT(tx < DGP_BITS_UINT32);
	minrw = MIN(irw, mrw);
	adjtx = tx;
    }

    /* How much of the model is overlapped? */
    miny = MAX(ty, 0);
    maxy = MIN(model->ysize, image->ysize + ty);
    maxy = MAX(maxy, 0);

    {
    Dgp_UInt32 *tmp_storage;

    /* NT has _alloca but no alloca */
#ifdef	_WIN32
#define	alloca	_alloca
#endif

/* I can't find a stack allocator, so for the time being I'm going to put this on the heap. */
/* 14.3.2000 / dhtesq. */
#if defined(VXWORKS) || defined(EQUATOR_CA_ENV)
#define alloca malloc
#endif  /* (VXWORKS) */

    tmp_storage = (Dgp_UInt32 *)alloca(sizeof(Dgp_UInt32)*modelsize);
#if defined(VXWORKS) || defined(EQUATOR_CA_ENV)
    DGP_VOID_ASSERT(tmp_storage);
#endif

     /* This isn't really necessary */
/*    memset((Dgp_UInt8*)tmp_storage, 0, sizeof(Dgp_UInt32) * modelsize);*/

    /* Do this in a nice general manner */

    /* Fill in the top and bottom bits that aren't overlapped */
    for (y = 0, mrow = model->bits, trow = tmp_storage;
	 y < miny;
	 y++, mrow += mrw, trow += mrw) {
	for (x = 0, mp = mrow, tmp = trow; x < mrw; x++, mp++, tmp++) {
	    *tmp = *mp;
        }
    }

    for (y = maxy, mrow = model->bits + mrw * y, trow = tmp_storage + mrw * y;
	 y < model->ysize;
	 y++, mrow += mrw, trow += mrw) {
	for (x = 0, mp = mrow, tmp = trow; x < mrw; x++, mp++, tmp++) {
	    *tmp = *mp;
        }
    }

    /* Now the overlapped middle */
    for (y = miny, irow = image->bits + irw * (y - ty),
	 mrow = model->bits + mrw * y, trow = tmp_storage + mrw * y;
	 y < maxy;
	 y++, irow += irw, mrow += mrw, trow += mrw) {
	ip = irow;
	mp = mrow;
	tmp = trow;

	/* Load up (from image) the partial word extending off the left */
	if (tx < 0) {
	    left = (*ip++) << (-tx);
        }
	else {
	    left = 0;
        }

	/*
	 * Handle the horizontal overlap. At any moment, ip points at
	 * the word overlapping *mp, and extending off its right (or the single
	 * overlapping word if tx==0). left contains the low-order bits from
	 * the word to the left of that, shifted up to the right place.
	 */
	for (x = 0; x < minrw; x++) {
	    cur = *ip++;
	    val = left | (cur >> adjtx);
	    *tmp++ = val & *mp++;
	    if (adjtx) {
		left = cur << (DGP_BITS_UINT32 - adjtx);
            }
	    else {
		/* x << 32 is undefined, actually; not guaranteed to be 0 */
		left = 0;
            }
        }

	/* Now the rest of the row, since we've run out of image */
	val = left | (~0ul >> adjtx);
	for (; x < mrw; x++) {
	    /* image defaults to all-1s outside its borders */
	    *tmp++ = val & *mp++;
	    val = ~0ul;
        }
    }

  
    maxdensity = ScanMismatchImage(tmp_storage, model->row_width, 
				   model->xsize, model->ysize);
#if defined(VXWORKS) || defined(EQUATOR_CA_ENV)
    free(tmp_storage);
#endif
    }

    return (maxdensity); 
}

#define Bit(b)     (1 << (b))

/*  Move the bit at location b (counting from the left) to location k 
    (counting from the right) */

#define SetupBit(b)                                                           \
  bits  = (Dgp_UInt16)MoveAndMask(up, b-1, 0);                                \
  bits |= MoveAndMask(up, b,   1);                                           \
  bits |= MoveAndMask(me, b-1, 3);                                           \
  bits |= MoveAndMask(me, b,   4);                                           \
  bits |= MoveAndMask(dn, b-1, 6);                                           \
  bits |= MoveAndMask(dn, b,   7);

#define DoBit(b)                                                              \
  bits <<= 1;                                                                 \
  bits &= 0x1FF & ~((1 << 3) | (1 << 6));                                     \
  bits |= MoveAndMask(up, b-1, 0);                                           \
  bits |= MoveAndMask(me, b-1, 3);                                           \
  bits |= MoveAndMask(dn, b-1, 6);                                           \
  max = MAX(max, DensityLookup[bits]);

#define ShiftWordLeft                                                         \
  up = *(--ptr_up);                                                           \
  me = *(--ptr);                                                              \
  dn = *(--ptr_dn);

/*
 * The heart of the code that calcs the max density in a 3x3 bitmap.
 */
static Dgp_Int32
ScanMismatchImage(Dgp_UInt32* tmp, Dgp_Int32 width, Dgp_Int32 xsize, Dgp_Int32 ysize)
{
    Dgp_Int32 max = 0;
    Dgp_Int32 y;
    Dgp_UInt32 *ptr_up;
    Dgp_UInt32 *ptr    = tmp + width - 1;	/* Set to last word of row 0 */
    Dgp_UInt32 *ptr_dn = ptr + width;	/* Set to last word of row 1 */
    Dgp_UInt32 up;
    Dgp_UInt32 me = *ptr;
    Dgp_UInt32 dn = *ptr_dn;
    const Dgp_UInt8 *DensityLookup = bitcounts;
    Dgp_Int32 wordsleft;
    Dgp_UInt16 bits = 0;

    for (y = 1; y < ysize-1; y++) {
	ptr_up = tmp + y*width - 1;
	ptr    = ptr_up + width;
	ptr_dn = ptr + width;

	up = *ptr_up;
	me = *ptr;
	dn = *ptr_dn;

	wordsleft = width;

	switch(xsize & 31) {
	  case 0:  
        SetupBit(31);
        break;
	  case 1:  
	    bits  = (Dgp_UInt16)MoveAndMask(up, 0, 1);
	    bits |= MoveAndMask(me, 0, 4);
	    bits |= MoveAndMask(dn, 0, 7);
	    ShiftWordLeft;
	    bits |= MoveAndMask(up, 31, 0);
	    bits |= MoveAndMask(me, 31, 3);
	    bits |= MoveAndMask(dn, 31, 6);
	    wordsleft--;
	    break;   
	  case 2:  SetupBit(1);  break;
	  case 3:  SetupBit(2);  break;
	  case 4:  SetupBit(3);  break;
	  case 5:  SetupBit(4);  break;
	  case 6:  SetupBit(5);  break;
	  case 7:  SetupBit(6);  break;
	  case 8:  SetupBit(7);  break;
	  case 9:  SetupBit(8);  break;
	  case 10:  SetupBit(9);  break;
	  case 11:  SetupBit(10);  break;
	  case 12:  SetupBit(11);  break;
	  case 13:  SetupBit(12);  break;
	  case 14:  SetupBit(13);  break;
	  case 15:  SetupBit(14);  break;
	  case 16:  SetupBit(15);  break;
	  case 17:  SetupBit(16);  break;
	  case 18:  SetupBit(17);  break;
	  case 19:  SetupBit(18);  break;
	  case 20:  SetupBit(19);  break;
	  case 21:  SetupBit(20);  break;
	  case 22:  SetupBit(21);  break;
	  case 23:  SetupBit(22);  break;
	  case 24:  SetupBit(23);  break;
	  case 25:  SetupBit(24);  break;
	  case 26:  SetupBit(25);  break;
	  case 27:  SetupBit(26);  break;
	  case 28:  SetupBit(27);  break;
	  case 29:  SetupBit(28);  break;
	  case 30:  SetupBit(29);  break;
	  case 31:  SetupBit(30);  break;
	    }

	switch(xsize & 31) {
	      case  1: DoBit(31);
	      case  0: DoBit(30);
	      case 31: DoBit(29);
	      case 30: DoBit(28);
	      case 29: DoBit(27);
	      case 28: DoBit(26);
	      case 27: DoBit(25);
	      case 26: DoBit(24);
	      case 25: DoBit(23);
	      case 24: DoBit(22);
	      case 23: DoBit(21);
	      case 22: DoBit(20);
	      case 21: DoBit(19);
	      case 20: DoBit(18);
	      case 19: DoBit(17);
	      case 18: DoBit(16);
	      case 17: DoBit(15);
	      case 16: DoBit(14);
	      case 15: DoBit(13);
	      case 14: DoBit(12);
	      case 13: DoBit(11);
	      case 12: DoBit(10);
	      case 11: DoBit( 9);
	      case 10: DoBit( 8);
	      case  9: DoBit( 7);
	      case  8: DoBit( 6);
	      case  7: DoBit( 5);
	      case  6: DoBit( 4);
	      case  5: DoBit( 3);
	      case  4: DoBit( 2);
	      case  3: DoBit( 1);
	      case  2: 
		if (!(--wordsleft>0)) {
		    goto END_OF_LINE;
		    }
		ShiftWordLeft;
		DoBit(32);	/* Really bit 0, but I hate negative numbers */
	    }

	for (;;) {
	      DoBit(31);
	      DoBit(30);
	      DoBit(29);
	      DoBit(28);
	      DoBit(27);
	      DoBit(26);
	      DoBit(25);
	      DoBit(24);
	      DoBit(23);
	      DoBit(22);
	      DoBit(21);
	      DoBit(20);
	      DoBit(19);
	      DoBit(18);
	      DoBit(17);
	      DoBit(16);
	      DoBit(15);
	      DoBit(14);
	      DoBit(13);
	      DoBit(12);
	      DoBit(11);
	      DoBit(10);
	      DoBit( 9);
	      DoBit( 8);
	      DoBit( 7);
	      DoBit( 6);
	      DoBit( 5);
	      DoBit( 4);
	      DoBit( 3);
	      DoBit( 2);
	      DoBit( 1);
	      
		if (!(--wordsleft>0)) {
		    goto END_OF_LINE;
		    }
		ShiftWordLeft;
		DoBit(32);	/* Really bit 0, but I hate negative numbers */
	    }


      END_OF_LINE:;
	}
    return(max);
    }

/*
 * Calculate the number of mismatched points between two models. 
 * Xor two models WITHOUT dilation.
 * Shift model2 by (x,y).
 */
static Dgp_Int32
ModelNumMismatched(tokCModelInfo *model1, tokCModelInfo *model2,
		   Dgp_Int16 tx, Dgp_Int16 ty, Dgp_Int32 MaxMismatched,
		   tokOptions *opt)
{
    Dgp_Int32 x, y;
    Dgp_UInt32 *m1p, *m2p;
    Dgp_UInt32 *m1row, *m2row;
    Dgp_UInt32 tmp, left, cur;
    Dgp_Int32 m1startskip, m2startskip;
    Dgp_Int32 m1rw, m2rw;
    Dgp_Int32 minrw;
    Dgp_Int32 miny, maxy;
    Dgp_Int32 mismatched = 0;

DGP_VOID_PROLOGUE( "ModelNumMismatched()", opt->DGP_PERRSTATE);

    m1rw = model1->row_width;
    m2rw = model2->row_width;

    if (ty > 0) {
	/* Count the number of points in model1 in rows 0..ty-1 */
	for (y = 0, m1row = model1->bits;
	     y < MIN(ty, model1->ysize);
	     y++, m1row += m1rw) {
	    for (x = 0, m1p = m1row; x < m1rw; x++, m1p++) {
		tmp = *m1p;
		mismatched += bitcounts[HIGH_PART(tmp)] +
		    bitcounts[LOW_PART(tmp)];
		}
	    if (mismatched > MaxMismatched) {
		return(mismatched);
		}
	    }
	}
    else if (ty < 0) {
	/* Count the number of points in model2 in rows 0..-ty-1 */
	for (y = 0, m2row = model2->bits;
	     y < MIN(-ty, model2->ysize);
	     y++, m2row += m2rw) {
	    for (x = 0, m2p = m2row; x < m2rw; x++, m2p++) {
		tmp = *m2p;
		mismatched += bitcounts[HIGH_PART(tmp)] +
		    bitcounts[LOW_PART(tmp)];
		}
	    if (mismatched > MaxMismatched) {
		return(mismatched);
		}
	    }
	}

    /* How high do we go? */
    miny = MAX(ty, 0);
    maxy = MIN(model1->ysize, model2->ysize + ty);
    maxy = MAX(maxy, 0);

    /* Count the stuff at the end early, for easy termination */
    if (maxy < model1->ysize) {
	/* Count the number of points in model1 in rows maxy..ysize-1 */
	for (y = maxy, m1row = model1->bits + y * m1rw;
	     y < model1->ysize;
	     y++, m1row += m1rw) {
	    for (x = 0, m1p = m1row; x < m1rw; x++, m1p++) {
		tmp = *m1p;
		mismatched += bitcounts[HIGH_PART(tmp)] +
		    bitcounts[LOW_PART(tmp)];
		}
	    if (mismatched > MaxMismatched) {
		    return(mismatched);
		}
	    }
	}
    else if (maxy < model2->ysize + ty) {
	/* Count the number of points in model in rows maxy-ty..ysize-1 */
	for (y = maxy - ty, m2row = model2->bits + y * m2rw;
	     y < model2->ysize;
	     y++, m2row += m2rw) {
	    for (x = 0, m2p = m2row; x < m2rw; x++, m2p++) {
		tmp = *m2p;
		mismatched += bitcounts[HIGH_PART(tmp)] +
		    bitcounts[LOW_PART(tmp)];
		}
	    if (mismatched > MaxMismatched) {
		    return(mismatched);
		}
	    }
	}

    /* Now, just the overlapping bit. */

    /* Check the horizontal overlap */
    if (tx > DGP_BITS_UINT32) {
	/* At least one Dgp_Int32 to skip at the start of each row */
	m1startskip = tx / DGP_BITS_UINT32;
	m2startskip = 0;
	}
    else if (tx < -DGP_BITS_UINT32) {
	m1startskip = 0;
	m2startskip = (-tx) / DGP_BITS_UINT32;
	}
    else {
	m1startskip = m2startskip = 0;
	}

    minrw = MIN(m1rw, m2rw);

    /* Special-case some common cases */
    if (tx == 0) {
	/* No horizontal shift */
	if (m1rw == m2rw) {
	    /* No horizonal shift and matching row widths - such luxury! */
	    for (y = miny, m1row = model1->bits + m1rw * y,
		 m2row = model2->bits + m2rw * (y - ty);
		 y < maxy;
		 y++, m1row += m1rw, m2row += m2rw) {
		for (x = 0, m1p = m1row, m2p = m2row;
		     x < minrw;
		     x++, m1p++, m2p++) {
		    tmp = *m1p ^ *m2p;
		    mismatched += bitcounts[HIGH_PART(tmp)] +
			bitcounts[LOW_PART(tmp)];
		    }
		if (mismatched > MaxMismatched) {
		    return(mismatched);
		    }
		}
	    }
	else if (m1rw < m2rw) {
	    /* A bit extra on the right of model2. */
	    for (y = miny, m1row = model1->bits + m1rw * y,
		 m2row = model2->bits + m2rw * (y - ty);
		 y < maxy;
		 y++, m1row += m1rw, m2row += m2rw) {
		for (x = 0, m1p = m1row, m2p = m2row;
		     x < minrw;
		     x++, m1p++, m2p++) {
		    tmp = *m1p ^ *m2p;
		    mismatched += bitcounts[HIGH_PART(tmp)] +
			bitcounts[LOW_PART(tmp)];
		    }
		for (; x < m2rw; x++, m2p++) {
		    tmp = *m2p;
		    mismatched += bitcounts[HIGH_PART(tmp)] +
			bitcounts[LOW_PART(tmp)];
		    }
		if (mismatched > MaxMismatched) {
		    return(mismatched);
		    }
		}
	    }
	else {
	    /* A bit extra on the right of model1 */
	    for (y = miny, m1row = model1->bits + m1rw * y,
		 m2row = model2->bits + m2rw * (y - ty);
		 y < maxy;
		 y++, m1row += m1rw, m2row += m2rw) {
		for (x = 0, m1p = m1row, m2p = m2row;
		     x < minrw;
		     x++, m1p++, m2p++) {
		    tmp = *m1p ^ *m2p;
		    mismatched += bitcounts[HIGH_PART(tmp)] +
			bitcounts[LOW_PART(tmp)];
		    }
		for (; x < m1rw; x++, m1p++) {
		    tmp = *m1p;
		    mismatched += bitcounts[HIGH_PART(tmp)] +
			bitcounts[LOW_PART(tmp)];
		    }
		if (mismatched > MaxMismatched) {
		    return(mismatched);
		    }
		}
	    }
	}
    /* Not a simple case. How about a mostly-simple case? */
    else if ((m1startskip == 0) && (m2startskip == 0)) {
	/* At least there aren't any words to skip at the left */
	if (tx > 0) {
	    /* Need to skip some bits at the left of m1 each row */
	    for (y = miny, m1row = model1->bits + m1rw * y,
		 m2row = model2->bits + m2rw * (y - ty);
		 y < maxy;
		 y++, m1row += m1rw, m2row += m2rw) {
		left = *m1row;
		tmp = TOPBITS(left, tx);
		mismatched += bitcounts[HIGH_PART(tmp)] +
		    bitcounts[LOW_PART(tmp)];
		left <<= tx;

		/*
		 * The inner loop. The state here:
		 * x and m1p track each other, referring to the x'th word
		 * of the current row of model1. m2p lags these by 1: it
		 * always refers to the x-1'th word of model2.
		 *
		 * At the beginning of each iteration, left contains some
		 * bits that haven't been dealt with yet. They should be
		 * combined with the top part of *m2p. The bottom part of
		 * *m2p should be combined with the top part of *m1p, and
		 * the rest of *m1p should be held over in left for the
		 * next iteration.
		 */
		for (x = 1, m1p = m1row + 1, m2p = m2row;
		     x < minrw;
		     x++, m1p++, m2p++) {
		    cur = *m1p;
		    tmp = left | (cur >> (DGP_BITS_UINT32 - tx));
		    tmp ^= *m2p;
		    mismatched += bitcounts[HIGH_PART(tmp)] +
			bitcounts[LOW_PART(tmp)];
		    left = cur << tx;
		    }

		/*
		 * Process the bits still in left.
		 * *m2p is always valid, since it lagged x by 1, and x only
		 * went up to minrw.
		 */
		if (minrw == m1rw) {
		    /*
		     * Maybe some words of model2 left to count, and model1 is
		     * done. If we take what's in left, checked against the
		     * next word in model2, then we're synchronised: minrw was
		     * limited by model1, so every bit of model1 will have been
		     * counted, as will the first minrw words of model2.
		     */
		    DGP_VOID_ASSERT(minrw <= m2rw);

		    tmp = left ^ *m2p++;
		    mismatched += bitcounts[HIGH_PART(tmp)] +
			bitcounts[LOW_PART(tmp)];
		    
		    for (; x < m2rw; x++, m2p++) {
			tmp = *m2p;
			mismatched += bitcounts[HIGH_PART(tmp)] +
			    bitcounts[LOW_PART(tmp)];
			}
		    }
		else {
		    /*
		     * model2 was the determining factor in minrw.
		     * We need to pull stuff out of one more word of model1,
		     * then handle the remaining words of model1.
		     */
		    DGP_VOID_ASSERT(minrw < m1rw);

		    cur = *m1p;
		    tmp = left | (cur >> (DGP_BITS_UINT32 - tx));
		    tmp ^= *m2p;
		    mismatched += bitcounts[HIGH_PART(tmp)] +
			bitcounts[LOW_PART(tmp)];
		    left = cur << tx;
		    mismatched += bitcounts[HIGH_PART(left)] +
			bitcounts[LOW_PART(left)];
		    for (x++, m1p++; x < m1rw; x++, m1p++) {
			tmp = *m1p;
			mismatched += bitcounts[HIGH_PART(tmp)] +
			    bitcounts[LOW_PART(tmp)];
			}
		    }

		/* Terminate? */
		if (mismatched > MaxMismatched) {
		    return(mismatched);
		    }

		}
	    }
	else {
	    /*
	     * Exact same logic, except that we negate tx and switch
	     * the roles of model1 and model2.
	     */
	    tx = -tx;
	    for (y = miny, m1row = model1->bits + m1rw * y,
		 m2row = model2->bits + m2rw * (y - ty);
		 y < maxy;
		 y++, m1row += m1rw, m2row += m2rw) {
		left = *m2row;
		tmp = TOPBITS(left, tx);
		mismatched += bitcounts[HIGH_PART(tmp)] +
		    bitcounts[LOW_PART(tmp)];
		left <<= tx;

		for (x = 1, m2p = m2row + 1, m1p = m1row;
		     x < minrw;
		     x++, m2p++, m1p++) {
		    cur = *m2p;
		    tmp = left | (cur >> (DGP_BITS_UINT32 - tx));
		    tmp ^= *m1p;
		    mismatched += bitcounts[HIGH_PART(tmp)] +
			bitcounts[LOW_PART(tmp)];
		    left = cur << tx;
		    }

		if (minrw == m2rw) {
		    DGP_VOID_ASSERT(minrw <= m1rw);

		    tmp = left ^ *m1p++;
		    mismatched += bitcounts[HIGH_PART(tmp)] +
			bitcounts[LOW_PART(tmp)];
		    
		    for (; x < m1rw; x++, m1p++) {
			tmp = *m1p;
			mismatched += bitcounts[HIGH_PART(tmp)] +
			    bitcounts[LOW_PART(tmp)];
			}
		    }
		else {
		    DGP_VOID_ASSERT(minrw < m2rw);

		    cur = *m2p;
		    tmp = left | (cur >> (DGP_BITS_UINT32 - tx));
		    tmp ^= *m1p;
		    mismatched += bitcounts[HIGH_PART(tmp)] +
			bitcounts[LOW_PART(tmp)];
		    left = cur << tx;
		    mismatched += bitcounts[HIGH_PART(left)] +
			bitcounts[LOW_PART(left)];
		    for (x++, m2p++; x < m2rw; x++, m2p++) {
			tmp = *m2p;
			mismatched += bitcounts[HIGH_PART(tmp)] +
			    bitcounts[LOW_PART(tmp)];
			}
		    }

		if (mismatched > MaxMismatched) {
		    return(mismatched);
		    }

		}
	    }
	}
    else {
	/*
	 * General case.
	 * Don't worry about speed - just make the code simple.
	 */
	DGP_PANIC("unwritten code");	/* Doesn't get much simpler... */
	}

    return(mismatched);
    }

/*
 * This function, when called with a model and an image, counts the number
 * of mismatched model points at model shift (tx,ty) wrt to the image.  
 *
 * Remember that the image has an implicit y offset of +1 because its
 * top row corresponds to the -1'th row of the thing that it was created
 * from. Also, the image is inverted. We therefore count bits in
 * (offset(model) & image) - these are exactly the model bits lying outside
 * the dilated thing-that-made-image.
 *
 * The logic here is the same as ModelNumMismatched, but we count (m&i) bits,
 * not (m1^m2) bits; this changes the edge cases since any bits in the
 * image lying outside the translated model should be ignored (default model
 * bit val is 0), but any bits in the translated model lying outside the image
 * should be counted (default image bit val is 1, since it's inverted).
 *
 * 12.4.2001 / dhtesq.
 * Profiling shows this function eating ~10% of the execution time for a single-page encode,
 * good for first place.  Hit count is highest by far.  Inlining doesn't do much though.
 * Action?
 *
 */
static Dgp_Int32
NumMismatched(tokCImageInfo         *image, 
              tokCModelInfo         *model,
              Dgp_Int16             tx, 
              Dgp_Int16             ty, 
              Dgp_Int32             MaxMismatched,
              tokOptions            *opt)
{
    Dgp_Int32 x, y;
    Dgp_UInt32 *ip, *mp;
    Dgp_UInt32 *irow, *mrow;
    Dgp_UInt32 tmp, left, cur;
    Dgp_Int32 istartskip, mstartskip;
    Dgp_Int32 irw, mrw;
    Dgp_Int32 minrw;
    Dgp_Int32 miny, maxy;
    Dgp_Int32 mismatched = 0;
    Dgp_Int32 maxdensity;
    Dgp_Int32 realtx = tx;	/* We trash tx later... */

	DGP_VOID_PROLOGUE( "NumMismatched()", opt->DGP_PERRSTATE);

    /* 
     * For a match,  M-OV <= (1-f)E = MaxMismatched
     * Since         OV <= N (where N=num image points),
     * this implies  M-N <= (1-f)E  for a match, so
     * if M-N > MaxMismatched, there's no way to have a match: 
     */
    if (model->npts - image->npts > MaxMismatched) {
		return(model->npts);
	}

    /*
     * Increment the position of the model with respect to the image by the
     * dilation radius since the image is dilationRadius pixels "down" by
     * default.
     */
    ty += (Dgp_Int16)opt->dilationRadius;

    irw = image->row_width;
    mrw = model->row_width;

    if (ty < 0) {
		/* Count the number of points in model in rows 0..-ty-1 */
		for (y = 0, mrow = model->bits; y < MIN(-ty, model->ysize); y++, mrow += mrw) {
			for (x = 0, mp = mrow; x < mrw; x++, mp++) {
				tmp = *mp;
				mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
			}
			if (mismatched > MaxMismatched) {
				return(mismatched);
			}
	    }
	}

    /* How high do we go? */
    miny = MAX(ty, 0);
    maxy = MIN(image->ysize, model->ysize + ty);
    maxy = MAX(maxy, 0);

    /* Count the stuff at the end early, for easy termination */
    if (maxy < model->ysize + ty) {
		/* Count the number of points in model in rows maxy-ty..ysize-1 */
		for (y = maxy - ty, mrow = model->bits + y * mrw; y < model->ysize; y++, mrow += mrw) {
			for (x = 0, mp = mrow; x < mrw; x++, mp++) {
				tmp = *mp;
				mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
			}
			if (mismatched > MaxMismatched) {
				return(mismatched);
			}
	    }
	}

    /* Now, just the overlapping bit. */

    /* Check the horizontal overlap */
    if (tx > DGP_BITS_UINT32) {
		/* At least one Dgp_Int32 to skip at the start of each row */
		istartskip = tx / DGP_BITS_UINT32;
		mstartskip = 0;
	}
    else if (tx < -DGP_BITS_UINT32) {
		istartskip = 0;
		mstartskip = (-tx) / DGP_BITS_UINT32;
	}
    else {
		istartskip = mstartskip = 0;
	}

    minrw = MIN(irw, mrw);

    /* Special-case some common cases */
    if (tx == 0) {
		/* No horizontal shift */
		if (irw >= mrw) {
			/* No horizonal shift and matching row widths - such luxury! */
			for (y = miny, irow = image->bits + irw * y, mrow = model->bits + mrw * (y - ty); y < maxy; y++, irow += irw, mrow += mrw) {
				for (x = 0, ip = irow, mp = mrow; x < minrw; x++, ip++, mp++) {
					tmp = *ip & *mp;
					mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
				}
				if (mismatched > MaxMismatched) {
					return(mismatched);
				}
			}
	    }
		else {
			/* A bit extra on the right of model. */
			for (y = miny, irow = image->bits + irw * y, mrow = model->bits + mrw * (y - ty); y < maxy; y++, irow += irw, mrow += mrw) {
				for (x = 0, ip = irow, mp = mrow; x < minrw; x++, ip++, mp++) {
					tmp = *ip & *mp;
					mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
				}
				for (; x < mrw; x++, mp++) {
					tmp = *mp;
					mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
				}
				if (mismatched > MaxMismatched) {
					return(mismatched);
				}
			}
	    }
	}
    /* Not a simple case. How about a mostly-simple case? */
    else if ((istartskip == 0) && (mstartskip == 0)) {
		/* At least there aren't any words to skip at the left */
		if (tx > 0) {
			/* Need to skip some bits at the left of i each row */
			for (y = miny, irow = image->bits + irw * y, mrow = model->bits + mrw * (y - ty); y < maxy; y++, irow += irw, mrow += mrw) {
				left = *irow;
				left <<= tx;

				/*
				* The inner loop. The state here:
				* x and ip track each other, referring to the x'th word
				* of the current row of image. mp lags these by 1: it
				* always refers to the x-1'th word of model.
				*
				* At the beginning of each iteration, left contains some
				* bits that haven't been dealt with yet. They should be
				* combined with the top part of *mp. The bottom part of
				* *mp should be combined with the top part of *ip, and
				* the rest of *ip should be held over in left for the
				* next iteration.
				*/
				for (x = 1, ip = irow + 1, mp = mrow; x < minrw; x++, ip++, mp++) {
					cur = *ip;
					tmp = left | (cur >> (DGP_BITS_UINT32 - tx));
					tmp &= *mp;
					mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
					left = cur << tx;
				}

				/*
				* Process the bits still in left.
				* *mp is always valid, since it lagged x by 1, and x only
				* went up to minrw.
				*/
				if (minrw == irw) {
				/*
				* Maybe some words of model left to count, and image is
				* done. If we take what's in left, checked against the
				* next word in model, then we're synchronised: minrw was
				* limited by image, so every bit of image will have been
				* counted, as will the first minrw words of model.
				*/
					DGP_VOID_ASSERT(minrw <= mrw);

					tmp = left & *mp++;
					mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
		    
					for (; x < mrw; x++, mp++) {
						tmp = *mp;
						mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
					}
				}
				else {
				/*
				* model was the determining factor in minrw.
				* We need to pull stuff out of one more word of image,
				* then we're done.
				*/
					DGP_VOID_ASSERT(minrw < irw);

					cur = *ip;
					tmp = left | (cur >> (DGP_BITS_UINT32 - tx));
					tmp &= *mp;
					mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
				}
				
				/* Terminate? */
				if (mismatched > MaxMismatched) {
					return(mismatched);
				}
			}
		}
		else {
		/*
	     * Exact same logic, except that we negate tx and switch
	     * the roles of image and model.
	     */
			tx = -tx;
			for (y = miny, irow = image->bits + irw * y, mrow = model->bits + mrw * (y - ty); y < maxy; y++, irow += irw, mrow += mrw) {
				left = *mrow;
				tmp = TOPBITS(left, tx);
				mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
				left <<= tx;

				for (x = 1, mp = mrow + 1, ip = irow; x < minrw; x++, mp++, ip++) {
					cur = *mp;
					tmp = left | (cur >> (DGP_BITS_UINT32 - tx));
					tmp &= *ip;
					mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
					left = cur << tx;
				}
				
				if (minrw == mrw) {
					DGP_VOID_ASSERT(minrw <= irw);
					
					tmp = left & *ip++;
					mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
				}
				else {
					DGP_VOID_ASSERT(minrw < mrw);
					
					cur = *mp;
					tmp = left | (cur >> (DGP_BITS_UINT32 - tx));
					tmp &= *ip;
					mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
					left = cur << tx;
					mismatched += bitcounts[HIGH_PART(left)] + bitcounts[LOW_PART(left)];
					for (x++, mp++; x < mrw; x++, mp++) {
						tmp = *mp;
						mismatched += bitcounts[HIGH_PART(tmp)] + bitcounts[LOW_PART(tmp)];
					}
				}
				
				if (mismatched > MaxMismatched) {
					return(mismatched);
				}				
			}
	    }
	}
    else {
	/*
	 * General case.
	 * Don't worry about speed - just make the code simple.
	 */
		DGP_PANIC("unwritten code");	/* Doesn't get much simpler... */
	}

    /* Do density checking. */

    if ((mismatched > opt->maxDense) && (mismatched <= MaxMismatched) && (model->xsize > 2)) {
		maxdensity = GetMismatchDensity(image, model, realtx, ty, opt);
		if (maxdensity > opt->maxDense) {
			return(model->npts); 
	    }
	}

    /* mismatched is now total number of mismatched points */
    return (mismatched); 
}

/*
 * The translations tried by the matching routines.  This should be set up
 * so that the first translation is (0,0), the next 8 are the other +-1
 * translations, and the next 16 are the +-2 translations.
 */
static Dgp_Int32 trans[25][2] = {{0,0}, {-1,0}, {1,0}, {0,-1}, {0,1},
				  {-1,-1}, {1,1}, {1,-1}, {-1,1},
				  {-2,0}, {2,0}, {0,-2}, {0,2},
				  {-2,-2}, {2,2}, {2,-2}, {-2,2},
				  {-2,-1}, {2,1}, {1,-2}, {-1,2},
				  {-2,1}, {2,-1}, {-1,-2}, {1,2}};

/*
 * Determine if the two bitmaps match, by comparing model1 with image2,
 * and image2 with model1.  model2 is offset by (tx,ty) wrt model1.
 * Try a few matches around (*tx, *ty) and report back the best one.
 */
Dgp_Bool
EdgeHausdorffMatch(tokCModelInfo            *model1, 
                   tokCImageInfo            *image1,
                   tokCModelInfo            *model2, 
                   tokCImageInfo            *image2,
                   Dgp_Int32                *tx, 
                   Dgp_Int32                *ty,
                   tokOptions               *opt)
{
    Dgp_Int32 nummissed1, nummissed2;
    Dgp_Int32 maxmissed1, maxmissed2;
    Dgp_Int32 nneigh = opt->nneigh;
    Dgp_Int32 i;
    Dgp_Int32 bestmissed = -1;
    Dgp_Int32 besti = -1;
    Dgp_Int32 basetx = *tx, basety = *ty;
  
    maxmissed1 = MaxMismatchedAllowed(model1->nedgepts, opt);
    maxmissed2 = MaxMismatchedAllowed(model2->nedgepts, opt);
    for (i = 0; i <= nneigh; i++) {

	nummissed2 = NumMismatched(image1, 
                               model2, 
                               (Dgp_Int16)(basetx + trans[i][0]), 
                               (Dgp_Int16)(basety + trans[i][1]),
                               maxmissed2, 
                               opt);

	if (nummissed2 > maxmissed2) {
	    continue;
	    }

	nummissed1 = NumMismatched(image2, 
                               model1,
                               (Dgp_Int16)(-basetx - trans[i][0]),
                               (Dgp_Int16)(-basety - trans[i][1]),
                               maxmissed1, 
                               opt);

	if (nummissed1 > maxmissed1) {
	    continue;
	    }

	if ((besti == -1) || (bestmissed > MAX(nummissed1, nummissed2))) {
	    besti = i;
	    bestmissed = MAX(nummissed1, nummissed2);
	    }
	}

    if (besti == -1) {
	return(DGP_FALSE);
	}

    (*tx) = basetx + trans[besti][0];
    (*ty) = basety + trans[besti][1];

    return(DGP_TRUE);
    }

/*
 * Compare the models to each other WITHOUT dilation to determine the
 * best translation.
 * The initial values of *tx and *ty say which translations should be
 * checked: that translation and its eight neighbours (these are translations
 * of model2 wrt model1). The final values of *tx and *ty are the best
 * translation found.
 */
void
ShiftEdgeHausdorffMatch(tokCModelInfo           *model1, 
                        tokCImageInfo           *image1,
                        tokCModelInfo           *model2, 
                        tokCImageInfo           *image2,
                        Dgp_Int32               *tx,
                        Dgp_Int32               *ty,
                        tokOptions              *opt){
    Dgp_Int32 nneigh = opt->nneigh;
    Dgp_Int32 i;
    Dgp_Int32 nummissed1;
    Dgp_Int32 besti = -1;
    Dgp_Int32 minmissed = 0;
    Dgp_Int32 basetx = *tx, basety = *ty;

    opt = opt;
    image1 = image1;
    image2 = image2;

    besti = 0;
    minmissed = model1->npts;

    for (i = 0; i <= nneigh; i++) {
	nummissed1 = ModelNumMismatched(model1, 
                                    model2, 
                                    (Dgp_Int16)(basetx + trans[i][0]),
                                    (Dgp_Int16)(basety + trans[i][1]),
                                    minmissed, 
                                    opt);
	if (nummissed1 < minmissed) {
	    minmissed = nummissed1;
	    besti = i;
	    }
	}

    (*tx) = basetx + trans[besti][0];
    (*ty) = basety + trans[besti][1];
    }
