/* 
 * 
 * $Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/miscalloc.c,v 1.16 2000/05/16 21:53:23 jahlborn Exp $ 
 * 
 */ 
 
/* 
 * A whole bunch of alloc/free routines for various datatypes which 
 * use the Chunk library routines. 
 */ 
#include "util.h" 
#include "tokprivate.h" 
#include "miscalloc.h" 
 
DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/miscalloc.c,v 1.16 2000/05/16 21:53:23 jahlborn Exp $"); 
 
 
/* 
 * Get rid of all the stuff hung off a tokCPixelSum. 
 */ 
void 
PixelSumFree(tokCPixelSum *ps, tokOptions *opt) 
{ 
    Dgp_State *dgp_state = opt->dgp_state; 
 
    if (ps->sumisbyte) { 
	if (ps->sum.bytesum != (DgpImage_Gray)NULL) { 
#ifdef	DBUG 
	    curhist -= (DGPIMG_WIDTH(ps->sum.bytesum) + sizeof(Dgp_Int8 *))  
               * DGPIMG_HEIGHT(ps->sum.bytesum) * sizeof(Dgp_Int8)  
               + sizeof(*ps->sum.bytesum); 
#endif 
	    DgpImage_free(ps->sum.bytesum); 
	    } 
	} 
    else { 
	if (ps->sum.Shortsum != (DgpImage_Short)NULL) { 
#ifdef	DBUG 
	    curhist -= (DGPIMG_WIDTH(ps->sum.Shortsum) + sizeof(Dgp_Int16 *))  
               * DGPIMG_HEIGHT(ps->sum.Shortsum) * sizeof(Dgp_Int16)  
               + sizeof(*ps->sum.Shortsum); 
#endif 
	    DgpImage_free(ps->sum.Shortsum); 
	    } 
	} 
    TOKPXL_FREE(ps); 
    } 
 
/* 
 * Free a class, including all the stuff hung off it. 
 */ 
void 
ClassFree(tokCClass c, tokOptions *opt) 
{ 
    Dgp_State *dgp_state = opt->dgp_state; 
    if (c->exemplar) { 
	FreeExemplar(c, opt); 
	} 
    if (c->members != DGP_NULLLIST) { 
	DgpAdt_liFree(c->members); 
	} 
    if (c->modelsum) { 
	PixelSumFree(c->modelsum, opt); 
	} 
    TOKCLASS_FREE(c); 
    } 
 
/* 
 * Free up a model, including all the bits hung off it. 
 */ 
void 
FreeModel(tokCModelInfo *model, tokOptions *opt) 
{ 
    Dgp_State *dgp_state = opt->dgp_state; 
  DGP_VOID_PROLOGUE( "FreeModel()", DGP_PERRSTATE); 
 
    DGP_VOID_ASSERT(model != (tokCModelInfo *)NULL); 
    if (model->bits != (Dgp_UInt32 *)NULL) { 
	(DGP_FREE)((void *)model->bits); 
#ifdef	DBUG 
	curbitmap -= model->row_width * model->ysize * sizeof(Dgp_UInt32); 
#endif 
	} 
 
    TOKMDL_FREE(model); 
    } 
 
/* 
 * Free up an image, including all the bits hung off it. 
 */ 
void 
FreeImage(tokCImageInfo *image, tokOptions *opt) 
{ 
    Dgp_State *dgp_state = opt->dgp_state; 
  DGP_VOID_PROLOGUE( "FreeImage()", DGP_PERRSTATE); 
 
    DGP_VOID_ASSERT(image != (tokCImageInfo *)NULL); 
    if (image->bits != (Dgp_UInt32 *)NULL) { 
	(DGP_FREE)((void *)image->bits); 
#ifdef	DBUG 
	curimbitmap -= image->row_width * image->ysize * sizeof(Dgp_UInt32); 
#endif 
	} 
 
    TOKIMG_FREE(image); 
    } 
 
/* 
 * Free up a box, including all the bits hung off it. 
 */ 
void 
FreeBox(tokCBoxType *box, tokOptions *opt) 
{ 
    Dgp_State *dgp_state = opt->dgp_state; 
    if (box->model != (tokCModelInfo *)NULL) { 
	FreeModel(box->model, opt); 
	} 
    if (box->image != (tokCImageInfo *)NULL) { 
	FreeImage(box->image, opt); 
	} 
    TOKBOX_FREE(box); 
    } 
 
/* 
 * Get rid of the exemplar, if it was synthetic. 
 */ 
void 
FreeExemplar(tokCClass class, tokOptions *opt) 
{ 
    if (class->status & TOKCLASS_EXSYNTHETIC) { 
	FreeBox(class->exemplar, opt); 
	class->exemplar = (tokCBoxType *)NULL; 
	class->status &= ~(TOKCLASS_EXSYNTHETIC | TOKCLASS_EXCLEAN); 
	} 
    } 
