/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtok-lossy/tok-lossy.h,v 1.10 2003/02/05 22:19:43 mkrocken Exp $
 *
 */

/*
 * Declarations for the top-level tokenisation routines.
 */

#ifndef	TOK_LOSSY_H

#define	TOK_LOSSY_H

/*#include <sys/types.h> */
#include "dgp_state.h"
#include "list.h"
#include "util.h"
#include "image.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* The type of a tokeniser state cookie */
typedef struct _tokState * TokState;

/* and the only defined value */
#define	NULLTOKSTATE ((TokState)NULL)

/* The type of the class table sent to the custom merger */
typedef struct _tokClassTable * tokClassTable;

/* All the options the tokeniser can take */
typedef struct {
    Dgp_Float32 fracHausd;		/* Fraction for a Hausdorff match */
    Dgp_Int32 maxDense;		/* Max density in a 3x3 error region */
    Dgp_Int32 widthDiff;		/* Max. allowable width and height */
    Dgp_Int32 heightDiff;		/* differences for 2 boxes to match */
    Dgp_Int32 commitThresh;		/* How big a class gets before committing */
    Dgp_Int32 freezeHistThresh;	/* How big a class freezes its histogram */
    Dgp_Int32 makeHistThresh;		/* How big a class gets a histogram */
    Dgp_Int32 returnHistThresh;	/* How big a class returns a histogram */
    Dgp_Int32 nneigh;			/* Shifts tried are in neighborhood this big */
    Dgp_Int32 xshift;			/* These three params control allowable */
    Dgp_Int32 xslopechange;		/* errors for small tokens by adding a */
	
	/* 12.4.2001 / dhtesq. */
	/* Added this parameter to address small token mismatch issues.  The idea */
	/* is not to dilate models smaller than this. */
	Dgp_Int32 lowerDilationLimit;

    Dgp_Float32 newslope;		/* nonlinear portion to the error curve */
    Dgp_Int32 dilationRadius;		/* how much to dilate */
    Dgp_Int32 mockRadius;
    Dgp_UInt32 flags;	/* Various options bits */

    Dgp_State *dgp_state;    /* Digipaper state cookie. */

    /* Callback for custom class merging. Set to NULL for hausdorff-like
     * matching. Should return DGP_FALSE on fatal error. See tokclass.h for details. */
    ErrCode (*mergeCallback)(tokClassTable, TokState, void *);
    void *callbackCookie;	/* userdata for above callback */

} tokOptions;

/* The options bit values */
#define	TOK_OPT_COPYBITMAPS	0x01	/* Do I copy the bitmaps? */
#define	TOK_OPT_RETURNHIST	0x02	/* Do I return a histogram? */
/* 0x04 used to be TOK_OPT_COMMITKEEPHIST */
#define	TOK_OPT_RETURNOFFSET	0x08	/* Do I produce tokOffsets? */
#define	TOK_OPT_ALIGN		0x10	/* Do I shift for best alignment? */
#define	TOK_OPT_NONTPDIL	0x20	/*
					 * Do I use a non-topology-preserving
					 * dilation?
					 */

/* The things you feed down the tokeniser */
typedef struct {
    Dgp_UInt16 width, height;
    Dgp_UInt32 *bits;
    Dgp_UInt16 longsperline;
    void *userdata;
} tokBitmap;

/* What you get back a DgpAdt_List of */
typedef struct {
    tokBitmap exemplar;
    Dgp_Bool histIsByte;
    Dgp_Int32 histcount;
    union {
	    DgpImage_Gray byteHist;
	    DgpImage_Short shortHist;
	} histogram;
    DgpAdt_List members;
    void *userdata;
} tokClass;

/* If you specify TOK_OPT_RETURNOFFSET, you get these back at the end */
typedef struct {
    Dgp_Int16 xOffset, yOffset;
    void *userdata;
} tokOffset;

/* The routines */
extern ErrCode tokInit(tokOptions *options, TokState *state);
extern void tokSetDefaults(tokOptions *options);
extern ErrCode tokAddBitmap(TokState state, tokBitmap *bitmap);
extern ErrCode tokGetCount(TokState state, Dgp_Int32 *count);
extern ErrCode tokAddBitmapVerify(TokState state, tokBitmap *bitmap, Dgp_Bool *newClassFound, DgpAdt_List *classMatched);
extern ErrCode tokMerge(TokState state);
extern ErrCode tokPartial(TokState state, DgpAdt_List *output);
extern ErrCode tokPartialNoMerge(TokState state, DgpAdt_List *output);
extern ErrCode tokFinalNoMerge(TokState state, DgpAdt_List *output);
extern ErrCode tokResume(TokState state, DgpAdt_List output);
extern ErrCode tokFinal(TokState state, DgpAdt_List *output);
extern void tokFreeState(TokState state);
extern ErrCode tokFreeOutput(TokState state, DgpAdt_List output, Dgp_Bool hasOffsets);
extern ErrCode tokResumeKeepMember(TokState state, DgpAdt_List output);
extern ErrCode tokFreeClass(TokState state, tokClass *classes);
extern void tokAbort(TokState state);


#ifdef __cplusplus
}
#endif

#endif
