/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Author:  ?;  199n-9n
**
**  Filename: compile-tables.c.
**
**  Puts together Huffman tables;  output looks like a tex file...? / dhtesq.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Cleaned up after MK's script, which implements most of the dgp_types, et al.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtokcomp/compile-tables.c,v 1.7 2003/02/10 22:53:42 mkrocken Exp $
 *
 */

#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
/*  Removed 24.2.00 / dhtesq. */
/*#include "undecl-stdio.h" */
#include "dgp_err_control.h"
#include "util.h"
#include "huff.h"
#include "token_t.h"
#include "mmcompressprivate.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtokcomp/compile-tables.c,v 1.7 2003/02/10 22:53:42 mkrocken Exp $");


/*************************************************************************
**
**  Function:  static void printcode(Dgp_UInt32 code, Dgp_Int32 codelen)
**
**
**	Arguments:
**		o  code...
**      o  codelen..
**
**	Return value:
**		o  None.
**
**	Modification history:
**      o  None.
**  
**/

static void
printcode(Dgp_UInt32 code, Dgp_Int32 codelen) {
  Dgp_Int32 i;

  for (i = 0; i < codelen; i++) {
    if (code & (1ul << (codelen - i - 1))) {
      putchar('1');
    } else {
      putchar('0');
    }
  }
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static void printTable(encoder *enc, Dgp_Int8 *tablename, Dgp_Int8 *caption)
**
**
**	Arguments:
**		o  enc...
**      o  tablename...
**      o  caption..
**
**	Return value:
**		o  None.
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  Assertions converted to DGP_VOID_ASSERT's.
**  
**/

static void
printTable(encoder *enc, Dgp_Int8 *tablename, Dgp_Int8 *caption) {
  DGP_VOID_PROLOGUE("printTable()", NULL);
  Dgp_Int32 i;

  printf("\\begin{table}\n");
  printf("\\centerline{\n");
  printf("\\begin{tabular}{|l|l|}\n");
  printf("\\hline\n");
  printf("Value & Encoding \\\\ \\hline\n");

  for (i = 0; i < enc->ncodes; i++) {
    if (enc->codes[i].from == INT_MIN) {
      DGP_VOID_ASSERT(enc->codes[i].nbits < 0);
      printf("$-\\infty \\ldots %d$ & ", enc->codes[i].upto);
      printcode(enc->codes[i].prefix, enc->codes[i].prefixbits);
      printf(" + $ (%d - \\mbox{value})$ encoded as in Table~\\ref{table:bofs}", enc->codes[i].upto);
    } else if (enc->codes[i].upto == INT_MAX) {
      DGP_VOID_ASSERT(enc->codes[i].nbits < 0);
      printf("$%d \\ldots \\infty$ & ", enc->codes[i].from);
      printcode(enc->codes[i].prefix, enc->codes[i].prefixbits);
      printf(" + $ (\\mbox{value} - %d)$ encoded as in Table~\\ref{table:bofs}", enc->codes[i].from);
    } else {
      DGP_VOID_ASSERT(enc->codes[i].nbits >= 0);

      if (enc->codes[i].from != enc->codes[i].upto) {
          printf("$%d \\ldots %d$ & ", enc->codes[i].from, enc->codes[i].upto);
      } else {
          printf("$%d$ &", enc->codes[i].from);
      }

      printcode(enc->codes[i].prefix, enc->codes[i].prefixbits);

      if (enc->codes[i].nbits > enc->codes[i].prefixbits) {
          if (enc->codes[i].from < 0) {
              printf(" + $ (\\mbox{value} + %d)$ encoded as %d bit%s", -enc->codes[i].from,
                  enc->codes[i].nbits - enc->codes[i].prefixbits,
                  ((enc->codes[i].nbits != enc->codes[i].prefixbits + 1) ? "s" : ""));
          } else if (enc->codes[i].from == 0) {
              printf(" + $ \\mbox{value}$ encoded as %d bit%s", enc->codes[i].nbits - enc->codes[i].prefixbits,
                  ((enc->codes[i].nbits != enc->codes[i].prefixbits + 1) ? "s" : ""));
          } else {
              printf(" + $ (\\mbox{value} - %d)$ encoded as %d bit%s", enc->codes[i].from,
                  enc->codes[i].nbits - enc->codes[i].prefixbits,
                  ((enc->codes[i].nbits != enc->codes[i].prefixbits + 1) ? "s" : ""));
          }
      }
    }
    if (i != enc->ncodes - 1) {
      printf("\\\\\n");
    }
  }

  if (enc->EOSbits > 0) {
    printf("\\\\\n");
    printf("END & ");
    printcode(enc->EOScode, enc->EOSbits);
  }

  printf("\\\\ \\hline\n");
  printf("\\end{tabular}\n");
  printf("}\n");
  printf("\\caption{%s}\n", caption);
  printf("\\label{%s}\n", tablename);
  printf("\\end{table}\n");
  return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  Dgp_Int32 main(Dgp_Int32 argc, Dgp_Int8 **argv)
**
**
**	Arguments:
**		o  argc, argv...
**
**	Return value:
**		o  None.
**
**	Modification history:
**      o  None.
**  
**/

Dgp_Int32
main(Dgp_Int32 argc, Dgp_Int8 **argv) {
  printf("%% tables generated by $Header: /project/icons/cvsroot/digipaper/src/libtokcomp/compile-tables.c,v 1.7 2003/02/10 22:53:42 mkrocken Exp $\";\n");

  printTable(&mmWHCoders[0], "table:wh0",
	     "Width/height encoding table");

  printTable(&mmUseCoders[0], "table:use0",
	     "Use count encoding table 0");
  printTable(&mmUseCoders[1], "table:use1",
	     "Use count encoding table 1");

  printTable(&mmFirstXCoders[0], "table:firstx0",
	     "First X encoding table 0");
  printTable(&mmFirstXCoders[1], "table:firstx1",
	     "First X encoding table 1");

  printTable(&mmDeltaXCoders[0], "table:deltax0",
	     "Delta X encoding table 0");
  printTable(&mmDeltaXCoders[1], "table:deltax1",
	     "Delta X encoding table 1");
  printTable(&mmDeltaXCoders[2], "table:deltax2",
	     "Delta X encoding table 2");

  printTable(&mmDeltaYCoders[0], "table:deltay0",
	     "Delta Y encoding table 0");
  printTable(&mmDeltaYCoders[1], "table:deltay1",
	     "Delta Y encoding table 1");
  printTable(&mmDeltaYCoders[2], "table:deltay2",
	     "Delta Y encoding table 2");

  return(0);
}
/*************************************************************************/
