/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Author:  ?;  199n-9n
**
**  Filename:  mmcompress.c.
**
**  Various functions for accessing bitfiles & compressing token bitmaps (?).
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Cleaned up after MK's script, which implements most of the dgp_types, et al.
**      o  Scattered liberr compliance issues.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtokcomp/mmcompress.c,v 1.64 2003/02/10 22:53:42 mkrocken Exp $
 *
 */



/*#include <stdlib.h>       Used? / dhtesq. */
#include <limits.h>
/*#include <stddef.h>       Used? / dhtesq. */
/*#include <string.h>       Used? / dhtesq. */
/*#include <stdio.h>        Used? / dhtesq. */
/*#include <sys/types.h>    Used? / dhtesq. */
/*#include <stddef.h>       Used? / dhtesq. */

#include "dgp_err_control.h"
#include "dgp_state.h"
#include "g4.h"
/*#include "util.h"         Used? / dhtesq. */
#include "huff.h"
#include "list.h"
#include "priq.h"
/*  Removed 22.2.00 / dhtesq. */
/*#include "dbug.h" */
/*#include "undecl-stdio.h" */
#include "token_t.h"
#include "tokpos_t.h"
#include "mmcompress.h"
#include "mmcompressprivate.h"
/*#include "dgp_allocators.h"   Used? / dhtesq. */
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtokcomp/mmcompress.c,v 1.64 2003/02/10 22:53:42 mkrocken Exp $");


/* A structure for bit-level access to a file. */
/*  Not to be confused with decompression structure of the same name -- see mmuncompress.c. / dhtesq. */
typedef struct {
  Dgp_UInt32 buf;
  Dgp_Int32 bitsinbuf;
  Dgp_UInt8 *outbuf;
  size_t outbufleft;
} bitfile;

/* A buffer for G4. */
#define  G4BUFLEN  ((size_t)50000)

/* Some guesses at byte sizes for stuff */
#define  TOKG4COMP      2  /* Guess at G4 compression for tokens */
#define  TOKG4SIZE      2  /* How many bytes token size might take */
#define  TOKOVERHEAD    6  /* Overhead per dictionary block */
#define  POSSIZE        4  /* How many bytes per position */
#define  POSOVERHEAD    10 /* Overhead per position block */

/*
 * How many bytes do we have to allow for the g4 length?
 * When we're about to write a g4-compressed block, we precede it
 * with its length; we need to have a bound on how many bytes
 * writing out that length can take.  The length itself can take up
 * to 39 bits to encode, plus there can be 32 bits in the bitfile buffer
 * (need to be subtracted from the bytes left count), for a total of 71
 * bits.
 */
#define  MAXG4LENBYTES  9

#define  LOTSOFPOS  3000  /* Treat big pages specially. */

#ifdef  INSTRUMENT
/* Some instrumentation for file size breakdown */
typedef struct {
  Dgp_Int32 overhead;
  Dgp_Int32 padding;
  Dgp_Int32 bitmapsize;
  Dgp_Int32 bitmaps;
  Dgp_Int32 height;
  Dgp_Int32 width;
  Dgp_Int32 uses;
} dict_bituses;

typedef struct {
  Dgp_Int32 overhead;
  Dgp_Int32 padding;
  Dgp_Int32 firstx;
  Dgp_Int32 deltax;
  Dgp_Int32 deltay;
  Dgp_Int32 y;
  Dgp_Int32 tokhuff;
  Dgp_Int32 toknum;
} pos_bituses;
#endif

/* Routines to access bitfiles */

/*************************************************************************
**
**  Function:  static Dgp_Bool writeCode(Dgp_UInt32 *code, Dgp_Int32 codelen, bitfile *f)
**
**
**  Write the given code, held in the *low-order* codelen bits of code
**  to the given file. Return DGP_TRUE on success, DGP_FALSE on failure. codelen
**  can be up to 31. code is written MSB-to-LSB. The non-significant
**  bits of code (those not the codelen LSBs) must be zero.
**
**	Arguments:
**		o  code...
**		o  codelen...
**		o  f...
**
**	Return value:
**		o  Boolean success/failure.
**
**	Modification history:
**      o  None.
**  
**/

static Dgp_Bool
writeCode(Dgp_UInt32 code, Dgp_Int32 codelen, bitfile *f)
{
  DGP_VOID_PROLOGUE("writeCode()", NULL);
  Dgp_Int32 i;
  DGP_VOID_ASSERT(f != (bitfile *)NULL);

  while (codelen) {
    /* Can we fit into the current buf? */
    if (f->bitsinbuf + codelen <= BITS_ULONG) {
      f->buf |= (code << (BITS_ULONG - f->bitsinbuf - codelen));
      f->bitsinbuf += codelen;
      codelen = 0;
    }
    else {
      /* Put as much as we can into the buffer */
      f->buf |= (code >> (codelen - (BITS_ULONG - f->bitsinbuf)));
      codelen -= BITS_ULONG - f->bitsinbuf;
      f->bitsinbuf = BITS_ULONG;
      /* Mask off what we've got left */
      code &= (1ul << codelen) - 1;
    }

    if (f->bitsinbuf == BITS_ULONG) {
      if (f->outbufleft < sizeof(Dgp_UInt32)) {
        return(DGP_FALSE);
      }
      for (i = 0; i < (Dgp_Int32)sizeof(Dgp_UInt32); i++) {
        *f->outbuf++ = (Dgp_UInt8)((f->buf >> (BITS_ULONG - BITS_UCHAR)) & 0xff);
        f->buf <<= BITS_UCHAR;
      }
      f->outbufleft -= sizeof(Dgp_UInt32);
      f->buf = 0;
      f->bitsinbuf = 0;
    }
  }
  return(DGP_TRUE);
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  static Dgp_Bool flushBitfile(bitfile *f)
**
**
**  Write out anything in the passed buffer.
**
**	Arguments:
**		o  f points to the bitfile to be flushed.
**
**	Return value:
**		o  Boolean success/failure.
**
**	Modification history:
**      o  None.
**  
**/

static Dgp_Bool
flushBitfile(bitfile *f) {
  DGP_VOID_PROLOGUE("flushBitfile()", NULL);

  DGP_VOID_ASSERT(f != (bitfile *)NULL);

  while (f->bitsinbuf > 0) {
    if (f->outbufleft == 0) {
      return(DGP_FALSE);
    }
    *f->outbuf++ = (Dgp_UInt8)((f->buf >> (BITS_ULONG - BITS_UCHAR)) & 0xff);
    f->buf <<= BITS_UCHAR;
    f->bitsinbuf -= BITS_UCHAR;
    f->outbufleft--;
  }
  f->buf = 0;
  f->bitsinbuf = 0;

  return(DGP_TRUE);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool writeValue(Dgp_UInt32 x, bitfile *f [, Dgp_Int32 *nbits])
**
**
**  Write out a value to the file. The encoding is:
**  Value       Code
**  0-127       0+(value as 7 bits)
**  128-1151    10+(value-128 as 10 bits)
**  1152-32767  11+(value as 15 bits)
**  32768-      1100000+(value as 32 bits)
**
**	Arguments:
**		o  x is the value to be written.
**      o  f is the bitfile to be written to.
**      o  (#ifdef INSTRUMENT) nbits...
**
**	Return value:
**		o  Boolean success/failure.
**
**	Modification history:
**      o  None.
**  
**/

static Dgp_Bool
writeValue(Dgp_UInt32 x, bitfile *f
#ifdef  INSTRUMENT
                                   , Dgp_Int32 *nbits
#endif
                                                     ){
#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("writeValue() / %d", x);
#endif
/*  VDEBUG("wv(%08x)\n", x, 50); */
  if (x < 128) {
#ifdef  INSTRUMENT
    *nbits += 8;
#endif
    return(writeCode(0, 1, f) &&    /* 0 */
     writeCode(x, 7, f));    /* value as 7 bits */
  } else if (x < 1152) {
#ifdef  INSTRUMENT
    *nbits += 12;
#endif
    return(writeCode(2, 2, f) &&    /* 10 */
     writeCode(x-128, 10, f));  /* value-128 as 10 bits */
  } else if (x < 32768) {
#ifdef  INSTRUMENT
    *nbits += 17;
#endif
    return(writeCode(3, 2, f) &&    /* 11 */
     writeCode(x, 15, f));    /* value as 15 bits */
  } else {
#ifdef  INSTRUMENT
    *nbits += 39;
#endif
    return(writeCode(0x60, 7, f) &&    /* 1100000 */
     writeCode(x, 32, f));    /* value as 32 bits */
  }
  /*  Not reached. */
}
/*************************************************************************/


/* Routines that deal with encoding */

/*************************************************************************
**
**  Function:  static Dgp_Int32 encodeValCount(Dgp_Int32 val)
**
**
**  Encode this using the special "value" escape coding, and say how many bits
**  are needed.
**
**	Arguments:
**		o  val is the value to be encoded (?).
**
**	Return value:
**		o  A signed integer indicating the number of bits required for encoding (?).
**
**	Modification history:
**      o  None.
**  
**/

static Dgp_Int32
encodeValCount(Dgp_Int32 val){
  if (val < 128) {
    return(8);
  } else if (val < 1152) {
    return(12);
  } else if (val < 32768) {
    return(17);
  } else {
    return(39);
  }
  /*  Not reached. */
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static INLINE codeinfo *findEncode(Dgp_Int32 val, codeinfo *codes, Dgp_Int32 ncodes)
**
**  Apparently does a binary search on the passed array to find the member corresponding to the passed
**  value...
**
**	Arguments:
**		o  val...
**      o  codes is an array of codeinfo{}...
**      o  ncodes is length(codes).
**
**	Return value:
**		o  A pointer to a codeinfo{}.
**
**	Modification history:
**      o  None.
**  
**/

static INLINE codeinfo *
findEncode(Dgp_Int32 val, codeinfo *codes, Dgp_Int32 ncodes) {
  Dgp_Int32 low = 0, high = ncodes - 1, mid;
  while (low < high) {
    mid = (low + high) / 2;
    if (val > codes[mid].upto) {
      low = mid + 1;
    }
    else if (val < codes[mid].from) {
      high = mid - 1;
    }
    else {
      return(&codes[mid]);
    }
  }
  return(&codes[low]);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Int32 encodeCount(Dgp_Int32 val, encoder *enc)
**
**
**  Figure out how many bits it takes to encode value in the encoder given
**  by enc.
**
**	Arguments:
**		o  val...
**      o  enc points to the encoder to be used.
**
**	Return value:
**		o  An integer indicating the number of bits required for encoding.
**
**	Modification history:
**      o  None.
**  
**/

static Dgp_Int32
encodeCount(Dgp_Int32 val, encoder *enc) {
  DGP_VOID_PROLOGUE("encodeCount()", NULL);
  codeinfo *cur;

  cur = findEncode(val, enc->codes, enc->ncodes);

  if (cur->nbits == -1) {
    /* Special case... watch out for infinities. */
    if (cur->from == INT_MIN) {
      return(cur->prefixbits + encodeValCount(cur->upto - val));
    }
    else {
      DGP_VOID_ASSERT(cur->upto == INT_MAX);
      return(cur->prefixbits + encodeValCount(val - cur->from));
    }
  }
  else {
    return(cur->nbits);
  }
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool writeEncoded( Dgp_Int32      val, 
**                                           bitfile        *f, 
**                                           codeinfo       *codes, 
**                                           Dgp_Int32      ncodes
**                                        [, Dgp_Int32      *nbits])
**
**
**  Actually do the encoding and write it out to the given file.
**
**	Arguments:
**		o  val...
**      o  f references the file to be written to.
**      o  codes is an array of codeinfo{}...
**      o  ncodes is length(codes)?
**      o  (#ifdef INSTRUMENT) nbits is...
**
**	Return value:
**		o  Passes success/failure value from writeCode().
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  Return type changed to Dgp_Bool...
**  
**/

static Dgp_Bool
writeEncoded(Dgp_Int32 val, bitfile *f, codeinfo *codes, Dgp_Int32 ncodes
#ifdef  INSTRUMENT
                                                                         , Dgp_Int32 *nbits
#endif
                                                                                           ){
  DGP_VOID_PROLOGUE("writeEncoded()", NULL);
  codeinfo *cur;

  cur = findEncode(val, codes, ncodes);

  if (cur->nbits == -1) {
    /* Special case... watch out for infinities. */
    if (cur->from == INT_MIN) {
#ifdef  INSTRUMENT
      *nbits += cur->prefixbits;
#endif
      return(writeCode(cur->prefix, cur->prefixbits, f) && 
             writeValue((unsigned)(cur->upto - val), f
#ifdef  INSTRUMENT
                                                      , nbits
#endif
                                                             ));
    } else {
      DGP_VOID_ASSERT(cur->upto == INT_MAX);
#ifdef  INSTRUMENT
      *nbits += cur->prefixbits;
#endif
      return(writeCode(cur->prefix, cur->prefixbits, f) &&
             writeValue((unsigned)(val - cur->from), f
#ifdef  INSTRUMENT
                                                      , nbits
#endif
                                                             ));
    }
  } else {
    /* Make sure they didn't mess up the coder. */
    DGP_VOID_ASSERT(val - cur->from < (1 << (cur->nbits - cur->prefixbits)));
#ifdef  INSTRUMENT
    *nbits += cur->nbits;
#endif
    return(writeCode(cur->prefix, cur->prefixbits, f) &&
           writeCode((unsigned)(val - cur->from), cur->nbits - cur->prefixbits, f));
  }
  /*  Not reached. */
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool writeWHDelta(Dgp_UInt32 delta, bitfile *f [, Dgp_Int32 *nbits])
**
**
**  Write a width or height delta value. Height delta values should have
**  had 1 subtracted before being passed here.
**
**	Arguments:
**		o  delta...
**      o  f references the file to be written to.
**      o  (#ifdef INSTRUMENT) nbits is...
**
**	Return value:
**		o  Passes success/failure value from writeEncoded().
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  dbug.h dependency exchanged for DGP_REPORT...
**  
**/

static Dgp_Bool
writeWHDelta(Dgp_UInt32 delta, bitfile *f
#ifdef  INSTRUMENT
                                         , Dgp_Int32 *nbits
#endif
                                                           ){
#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("writeWHDelta() / %d\n", delta);
#endif
/*  VDEBUG("wwhd(%d)\n", delta, 50); */
  return(writeEncoded((Dgp_Int32)delta, f, mmWHCoders[0].codes, mmWHCoders[0].ncodes
#ifdef  INSTRUMENT
                                                                                    , nbits
#endif
                                                                                           ));
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool writeWHEnd(bitfile *f [, Dgp_Int32 *nbits])
**
**
**  Write a width or height delta end code.
**
**	Arguments:
**      o  f references the file to be written to.
**      o  (#ifdef INSTRUMENT) nbits is...
**
**	Return value:
**		o  Passes success/failure value from writeCode().
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  dbug.h dependency exchanged for DGP_REPORT...
**  
**/

static Dgp_Bool
writeWHEnd(bitfile *f
#ifdef  INSTRUMENT
                     , Dgp_Int32 *nbits
#endif
                                       ) {
#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF0("writeWHEnd()\n");
#endif
/*  VDEBUG("wwhend()\n", 0, 50); */
#ifdef  INSTRUMENT
  *nbits += 7;
#endif
  return(writeCode(0x77, 7, f));    /* 1110111 */
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool writeUseCount(Dgp_UInt32         count, 
**                                           bitfile            *f,
**                                           Dgp_Int32          useEncoding
**                                        [, Dgp_Int32          *nbits])
**
**
**  Write a use count.
**
**	Arguments:
**      o  count...
**      o  f references the file to be written to.
**      o  useEncoding is an index for mmUseCoders.
**      o  (#ifdef INSTRUMENT) nbits is...
**
**	Return value:
**		o  Passes success/failure value from writeEncoded().
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  dbug.h dependency exchanged for DGP_REPORT...
**  
**/

static Dgp_Bool
writeUseCount(Dgp_UInt32 count, bitfile *f, Dgp_Int32 useEncoding
#ifdef  INSTRUMENT
                                                                 , Dgp_Int32 *nbits
#endif
                                                                                   ){
#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("writeUseCount() / %d\n", count);
#endif
/*  VDEBUG("wuc(%d)\n", count, 50); */
  return(writeEncoded((Dgp_Int32)count, f, mmUseCoders[useEncoding].codes, 
         mmUseCoders[useEncoding].ncodes
#ifdef  INSTRUMENT
                                        , nbits
#endif
                                               ));
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool writeYDelta(Dgp_UInt32       delta, 
**                                         bitfile          *f,
**                                         Dgp_Int32        deltaYEncoding
**                                      [, Dgp_Int32        *nbits])
**
**
**  Write out an inter-strip y delta.  Use the table from
**  mmDeltaYCoders[deltaYEncoding].
**
**	Arguments:
**      o  delta...
**      o  f references the file to be written to.
**      o  deltaYEncoding is an index for mmDeltaYCoders.
**      o  (#ifdef INSTRUMENT) nbits is...
**
**	Return value:
**		o  Passes success/failure value from writeEncoded().
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  dbug.h dependency exchanged for DGP_REPORT...
**  
**/

static Dgp_Bool
writeYDelta(Dgp_UInt32 delta, bitfile *f, Dgp_Int32 deltaYEncoding
#ifdef  INSTRUMENT
                                                                  , Dgp_Int32 *nbits
#endif
                                                                                    ){
#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("writeYDelta() / %d\n", delta);
#endif
/*  VDEBUG("wyd(%d)\n", delta, 50); */
  return(writeEncoded((Dgp_Int32)delta, f, mmDeltaYCoders[deltaYEncoding].codes,
         mmDeltaYCoders[deltaYEncoding].ncodes
#ifdef  INSTRUMENT
                                              , nbits
#endif
                                                     ));
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool writeYVal(Dgp_Int32      y, 
**                                       bitfile        *f, 
**                                       Dgp_Int32      logstripsize
**                                    [, Dgp_Int32      *nbits])
**
**
**  Write out an intra-strip y value - just the low order bits.
**
**	Arguments:
**      o  y...
**      o  f references the file to be written to.
**      o  logstripsize...
**      o  (#ifdef INSTRUMENT) nbits is...
**
**	Return value:
**		o  Passes success/failure value from writeCode().
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  dbug.h dependency exchanged for DGP_REPORT...
**  
**/

static Dgp_Bool
writeYVal(Dgp_Int32 y, bitfile *f, Dgp_Int32 logstripsize
#ifdef  INSTRUMENT
                                                         , Dgp_Int32 *nbits
#endif
                                                                           ){
#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("writeYVal() / %d\n", ((Dgp_UInt32)y & ((1ul << logstripsize) - 1)));
#endif
/*  VDEBUG("wyv(%ld)\n", ((Dgp_UInt32)y & ((1ul << logstripsize) - 1)), 50); */

#ifdef  INSTRUMENT
  *nbits += logstripsize;
#endif
  return(writeCode(((Dgp_UInt32)y & ((1ul << logstripsize) - 1)), logstripsize, f));
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool writeFirstX( Dgp_Int32       x, 
**                                          bitfile         *f, 
**                                          Dgp_Int32       firstXEncoding,
**                                       [, Dgp_Int32      *nbits])
**
**
**  Write out the first x value in a strip.
**
**	Arguments:
**      o  x...
**      o  f references the file to be written to.
**      o  firstXEncoding...
**      o  (#ifdef INSTRUMENT) nbits is...
**
**	Return value:
**		o  Passes success/failure value from writeEncoded().
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  dbug.h dependency exchanged for DGP_REPORT...
**  
**/

static Dgp_Bool
writeFirstX(Dgp_Int32 x, bitfile *f, Dgp_Int32 firstXEncoding
#ifdef  INSTRUMENT
                                                             , Dgp_Int32 *nbits
#endif
                                                                               ){
#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("writeFirstX() / %d\n", x);
#endif
/*  VDEBUG("wfx(%d)\n", x, 50); */
  return(writeEncoded(x, f,
          mmFirstXCoders[firstXEncoding].codes,
          mmFirstXCoders[firstXEncoding].ncodes
#ifdef  INSTRUMENT
                                               , nbits
#endif
                                                      ));
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool writeXDelta(Dgp_Int32        delta, 
**                                         bitfile          *f,
**                                         Dgp_Int32        deltaXEncoding
**                                      [, Dgp_Int32        *nbits])
**
**
**  Write out an intra-strip x delta.
**
**	Arguments:
**      o  delta...
**      o  f references the file to be written to.
**      o  deltaXEncoding...
**      o  (#ifdef INSTRUMENT) nbits is...
**
**	Return value:
**		o  Passes success/failure value from writeEncoded().
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  dbug.h dependency exchanged for DGP_REPORT...
**  
**/

static Dgp_Bool
writeXDelta(Dgp_Int32 delta, bitfile *f, Dgp_Int32 deltaXEncoding
#ifdef  INSTRUMENT
                                                                 , Dgp_Int32 *nbits
#endif
                                                                                   ){
#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("writeXDelta() / %d\n", delta);
#endif
/*  VDEBUG("wxd(%d)\n", delta, 50); */
  return(writeEncoded(delta, f,
          mmDeltaXCoders[deltaXEncoding].codes,
          mmDeltaXCoders[deltaXEncoding].ncodes
#ifdef  INSTRUMENT
                                               , nbits
#endif
                                                      ));
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool writeXDelta(bitfile          *f,
**                                         Dgp_Int32        deltaXEncoding
**                                      [, Dgp_Int32        *nbits])
**
**
**  Write out an end-of-row code.
**
**	Arguments:
**      o  f references the file to be written to.
**      o  deltaXEncoding...
**      o  (#ifdef INSTRUMENT) nbits is...
**
**	Return value:
**		o  Passes success/failure value from writeCode().
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  dbug.h dependency exchanged for DGP_REPORT...
**  
**/

static Dgp_Bool
writeEndOfRow(bitfile *f, Dgp_Int32 deltaXEncoding
#ifdef  INSTRUMENT
                                                  , Dgp_Int32 *nbits
#endif
                                                                    ) {
#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF0("writeEndOfRow()\n");
#endif
/*  VDEBUG("weor()\n", 0, 50); */
#ifdef  INSTRUMENT
  *nbits += mmDeltaXCoders[deltaXEncoding].EOSbits;
#endif
  return(writeCode(mmDeltaXCoders[deltaXEncoding].EOScode,
                   mmDeltaXCoders[deltaXEncoding].EOSbits, f));
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static void memcpy_shuffling(Dgp_UInt32         *bits, 
**                                          Dgp_UInt8          *buf,
**                                          Dgp_Int32          row_width, 
**                                          Dgp_Int32          Dgp_Int8_row_width, 
**                                          Dgp_Int32          height)
**
**
**  bits is an array of height rows of Dgp_Int32words, each row being row_width
**  longs wide.  We want to copy this, byte-by-byte, into *buf, copying
**  only Dgp_Int8_row_width from each row.
**
**	Arguments:
**      o  bits...
**      o  buf is the destination buffer.
**      o  row_width is in words.
**      o  Dgp_Int8_row_width is in bytes.
**      o  height is in rows.
**
**	Return value:
**		o  None.
**
**	Modification history:
**      o  None.
**  
**/

static void
memcpy_shuffling(Dgp_UInt32         *bits, 
                 Dgp_UInt8          *buf,
                 Dgp_Int32          row_width, 
                 Dgp_Int32          Dgp_Int8_row_width, 
                 Dgp_Int32          height){
  DGP_VOID_PROLOGUE("memcpy_shuffling()", NULL);
  Dgp_Int32         x, y, x2;
  Dgp_UInt32        tmp;
  Dgp_UInt32        *p;
  Dgp_UInt8         *q;

  DGP_VOID_ASSERT((Dgp_Int32)sizeof(Dgp_UInt32) * row_width >= Dgp_Int8_row_width);
  DGP_VOID_ASSERT((Dgp_Int32)sizeof(Dgp_UInt32) * (row_width - 1) < Dgp_Int8_row_width);

  for (y = 0; y < height; y++) {
    p = bits + y * row_width;
    q = buf + y * Dgp_Int8_row_width;
    for (x = 0; x < row_width - 1; x++) {
      tmp = *p++;
      for (x2 = 0; x2 < (Dgp_Int32)sizeof(Dgp_UInt32); x2++) {
        *q++ = (Dgp_UInt8)(tmp >> (BITS_ULONG - (x2 + 1) * BITS_UCHAR));
      }
    }
    tmp = *p++;
    for (x2 = 0; x2 < Dgp_Int8_row_width - (Dgp_Int32)sizeof(Dgp_UInt32) * (row_width - 1); x2++) {
      *q++ = (Dgp_UInt8)(tmp >> (BITS_ULONG - (x2 + 1) * BITS_UCHAR));
    }
  }
  return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool placeToken(token *t, Dgp_UInt32 *bits, Dgp_Int32 row_width, Dgp_Int32 xpos)
**
**
**  Plonk the given token into the given bit array, offset in x by the given
**  amount. The destination part of the bit array should have been cleared.
**  The destination bitmap's height should match t's height. row_width is
**  the width of each row of the destination bitmap in longs.
**
**  We assume that the token's bitmap is zero outside its borders (the
**  padding bits are clear).
**
**	Arguments:
**      o  t...
**      o  bits...
**      o  row_width is in words...
**      o  xpos...
**
**	Return value:
**		o  Boolean;  currently no control paths return DGP_FALSE.
**
**	Modification history:
**      o  None.
**  
**/

static Dgp_Bool
placeToken(token *t, Dgp_UInt32 *bits, Dgp_Int32 row_width, Dgp_Int32 xpos){
  Dgp_Int32         i, j;
  Dgp_UInt32        tmp;
  Dgp_UInt32        *p;
  Dgp_UInt32        *q;
  Dgp_Int32         Dgp_Int32pos;
  Dgp_Int32         mrow_width;
  Dgp_Int32         bitsinlastDgp_Int32;

  /* How many longs in the model actually contain data? */
  mrow_width = (t->width + (BITS_ULONG - 1)) / BITS_ULONG;
  bitsinlastDgp_Int32 = BITS_ULONG - ((mrow_width * BITS_ULONG) - t->width);

  Dgp_Int32pos = xpos / BITS_ULONG;
  xpos = xpos % BITS_ULONG;
    
  for (i = 0; i < t->height; i++) {
    p = t->bits + i * t->longsperline;
    q = bits + Dgp_Int32pos + i * row_width;

    /* Go through Dgp_Int32-by-Dgp_Int32 */
    if (xpos != 0) {
      for (j = 0; j < mrow_width - 1; j++) {
        tmp = *p++;
        *q++ |= tmp >> xpos;
        *q |= tmp << (BITS_ULONG - xpos);
      }
      tmp = *p++;
      if (xpos + bitsinlastDgp_Int32 > BITS_ULONG) {
        *q++ |= tmp >> xpos;
        *q |= tmp << (BITS_ULONG - xpos);
      } else {
        *q |= tmp >> xpos;
      }
    } else {
      /* No shifting - yay! */
      for (j = 0; j < mrow_width; j++) {
        *q++ |= *p++;
      }
    }
  }
  return(DGP_TRUE);
}
/*************************************************************************/

#define TCMP_BUF_CHECK(cond) \
  DGP_ERR_GEN_CLEANUP_NO_MSG((cond), cErrBufferExceeded)


/*************************************************************************
**
**  Function:  static ErrCode writeg4Class(token      **hclass, 
**                                         Dgp_Int32  ntokens, 
**                                         bitfile    *f,
**                                        [dict_bituses *bituses,]
**                                         void       *g4Encoder, 
**                                         void       **encbuf, 
**                                         size_t     *encbuflen,
**                                         Dgp_State  *dgp_state)
**
**
**  Glom all the tokens in a height class together and write them out as a
**  single g4-compressed blob. The tokens are at hclass[0]..hclass[ntokens-1].
**
**	Arguments:
**      o  hclass...
**      o  ntokens is length(hclass)?...
**      o  f references the destination bitfile...
**      o  (#ifdef INSTRUMENT) bituses...
**      o  g4Encoder...
**      o  encbuf...
**      o  encbuflen...
**      o  dgp_state references a prepared Dgp_State{}.
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

static ErrCode
writeg4Class(token      **hclass, 
             Dgp_Int32  ntokens, 
             bitfile    *f,
#ifdef  INSTRUMENT
             dict_bituses *bituses,
#endif
             void       *g4Encoder, 
             void       **encbuf, 
             size_t     *encbuflen,
             Dgp_State  *dgp_state){
  DGP_PROLOGUE("writeg4Class()", DGP_PERRSTATE);
  Dgp_Int32         i;
  Dgp_Int32         classwidth;
  Dgp_Int32         classheight;
  Dgp_Int32         tpos;
  Dgp_UInt32        *classbits = (Dgp_UInt32 *)NULL;
  size_t            Dgp_Int8_row_width;
  Dgp_Int32         row_width;
  size_t            outlen;
  size_t            spaceleft;
  Dgp_Allocators    *a = DGP_PALLOCATORS;


  DGP_ASSERT(hclass != (token **)NULL);
  DGP_ASSERT(f != (bitfile *)NULL);
  DGP_ASSERT(ntokens >= 1);

  classheight = hclass[0]->height;

  /* Figure out the total width of the class */
  classwidth = 0;
  for (i = 0; i < ntokens; i++) {
    DGP_ASSERT(hclass[i]->height == classheight);
    classwidth += hclass[i]->width;
  }
  /* How many longs?... */
  row_width = (classwidth + (BITS_ULONG - 1)) / BITS_ULONG;
  /* ...and how many bytes if I only pad to the next byte? */
  Dgp_Int8_row_width = (classwidth + (BITS_UCHAR - 1)) / BITS_UCHAR;
    
  /* Allocate the plonking place... */
  DGP_CHECK_ALLOC_CLEANUP(classbits = (Dgp_UInt32 *)(a->calloc)((size_t)classheight * row_width, sizeof(*classbits)));
    
  /* ...and start plonking. */
  tpos = 0;
  for (i = 0; i < ntokens; i++) {
    TCMP_BUF_CHECK(!placeToken(hclass[i], classbits, row_width, tpos));
    tpos += hclass[i]->width;
  }

#if 0
  {
    FILE *fp;
    Dgp_Int8 fname[100];
    sprintf(fname, "/tmp/hc%03dx%02d.pbm", classwidth, classheight);
    fp = fopen(fname, "w");
    fprintf(fp, "P4 %d %d\n", row_width * BITS_ULONG, classheight);
    fwrite((Dgp_Int8 *)classbits, row_width * sizeof(Dgp_UInt32), classheight, fp);
    fclose(fp);
  }
#endif

  /* This is our maximum outlen for the encoding of this data. */
  outlen = Dgp_Int8_row_width * classheight;

  /* Check to see if our encbuf is big enough. */
  if(*encbuflen < outlen) {
    if(*encbuf != NULL) {
      (a->free)(*encbuf);
      *encbuf = NULL;
    }
    *encbuflen = outlen;
    DGP_CHECK_ALLOC_CLEANUP(*encbuf = (a->malloc)(outlen));
  }

  /*
   * This is where it gets messy. We need to write out the size of the G4
   * output, followed by the G4 output itself. But we find out the size
   * by doing the encoding - we need to either do the encoding twice or
   * stash it away somewhere. g4buf is available for stashing, but if
   * it's not big enough, we need to do the encode-twice bit, which
   * actually ends up being 2.5 encodings (twice to write it out, and one
   * partial failed one).
   *
   * We do have one out: if the G4-encoded bitmap is larger than the
   * original bitmap, we have leave to store the uncompressed bitmap,
   * flagging this with an "encoded size" value of zero.
   */

  /* Encode the mess */

  /* Make sure we have some minumum space. */
  TCMP_BUF_CHECK(f->outbufleft <= MAXG4LENBYTES);

  /* A slightly conservative ERR_CLEANUP. */

  spaceleft = f->outbufleft - MAXG4LENBYTES;

  /* Now do the g4 encoding. */
  DGP_ERR_CHECK_NO_MSG(g4EncodeC(   g4Encoder, 
                                    classbits, 
                                    classwidth, 
                                    classheight, 
                                    row_width,
                                    *encbuf,
                                    &outlen,
                                    DGP_FALSE)) {
  
    /* Hmm, we seem to have gotten an error (maybe it was too big, 
       maybe somethign else, either way...), so, we'll just try to
       stick this thing in the file unencoded. */
  
    DGP_CLEAR_CUR_ERROR;

    /* Make sure we have space for the unencoded data... */
    TCMP_BUF_CHECK(spaceleft < Dgp_Int8_row_width * classheight);
  
#ifdef  INSTRUMENT
    TCMP_BUF_CHECK(!writeValue(0u, f , &bituses->bitmapsize));
#else
    TCMP_BUF_CHECK(!writeValue(0u, f));
#endif
#ifdef  INSTRUMENT
    bituses->padding += (BITS_ULONG - f->bitsinbuf) % BITS_UCHAR;
#endif
    TCMP_BUF_CHECK(!flushBitfile(f));

    /* Copy the bitmap across. */
    memcpy_shuffling(classbits, f->outbuf, row_width,
         Dgp_Int8_row_width, classheight);
    outlen = Dgp_Int8_row_width * classheight;
  
  } else {

    /* oooo good, the g4 encoding worked, now lets try to copy
       it into our output buffer. */

    /* Make sure we have space for the unencoded data... */
    TCMP_BUF_CHECK(spaceleft < outlen);

    /* Oh good, there's enough room, let's copy. */

    /* Write out the total length and sync to the next byte. */
#ifdef  INSTRUMENT
    TCMP_BUF_CHECK(!writeValue(outlen, f , &bituses->bitmapsize));
#else
    TCMP_BUF_CHECK(!writeValue(outlen, f));
#endif
#ifdef  INSTRUMENT
    bituses->padding += (BITS_ULONG - f->bitsinbuf) % BITS_UCHAR;
#endif
    TCMP_BUF_CHECK(!flushBitfile(f));

    /* ...and copy it across. */
    memcpy(f->outbuf, *encbuf, outlen);

  }


#ifdef DGP_MMTOK_DBUG
  DGP_REPORTF3("writeg4Class() / encoded %dx%d as %d bytes\n", classwidth, classheight, outlen);
#endif

  /* How much was written? Adjust the buffer pointers. */
#ifdef  INSTRUMENT
  bituses->bitmaps += outlen * 8;
#endif
  f->outbuf += outlen;
  DGP_ASSERT(outlen <= f->outbufleft);
  f->outbufleft -= outlen;

  (a->free)((void *)classbits);

  NORMAL_RETURN;

ERR_CLEANUP:

  if (classbits != (Dgp_UInt32 *)NULL) {
    (a->free)((void *)classbits);
  }

  if(*encbuf == NULL) *encbuflen = 0;

  DGP_WARNF0("writeg4Class bailing\n");

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Int32 tokGT(const void *p1, const void *p2)
**
**
**  Used to order the tokens. Order them by height; within height classes,
**  by width; within equivalent sizes by class number. This function decides
**  if p1 > p2.
**  The pointers are token * pointers.
**
**  No longer orders by class number. (Since there is no such thing.)
**
**	Arguments:
**      o  p1, p2...
**
**	Return value:
**      o  An signed integer:  1 for p1 first, -1 for p2 first, 0 for a draw.
**
**	Modification history:
**      o  30.10.2000 / dhtesq.  Consts removed.
**  
**/

typedef const token *tokPtr;
static int
tokGT(const void *p1, const void *p2){
  /*token c1 = **(token **)p1; */
  /*token c2 = **(token **)p2; */
  const token *c1 = *((const tokPtr *)p1);
  const token *c2 = *((const tokPtr *)p2);
    
  if ( c1->height > c2->height) {
    return 1;
  } else if( c1->height == c2->height) {
    if( c1->width > c2->width) {
      return 1;
    } else if( c1->width == c2->width) {
      if( c1->uses > c2->uses) {
        return -1;
      } else if( c1->uses == c2->uses) {
        return 0;
      } else { 
        return 1;
      }
    } else {
      return -1;
    }
  } else {
    return -1;
  }
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode tcmp_encodeTokDict(void         *dstate, 
**                                        token        *tokens, 
**                                        Dgp_Int32    ntok,
**                                        ErrCode      (*extbuilder)(token         **sorted_tokens,
**                                                                   Dgp_Int32     ntok, 
**                                                                   void          *extcookie,
**                                                                   void          **extension,
**                                                                   size_t        *extlen,
**                                                                   Dgp_State     *dgp_state),
**                                        void         *extcookie,
**                                        void         *outbuf, 
**                                        size_t       *outbuflen,
**                                        Dgp_State    *dgp_state)
**
**
**  Write out a token dictionary. Organize them by size and write out all
**  the ones of a certain height together etc. Ignore any token that is
**  marked "committed".
**
**  Generate and assign the Huffman codes for each token. Also build the
**  decompression tables.
**
**  Update the dictionary state block.
**
**	Arguments:
**      o  dstate...
**      o  tokens is an array of token{}.
**      o  ntok is length(tokens)?
**      o  extbuilder is...
**      o  extcookie...
**      o  outbuf...
**      o  outbuflen...
**      o  dgp_state points to a prepared Dgp_State{}.
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

ErrCode
tcmp_encodeTokDict(void         *dstate, 
                   token        *tokens, 
                   Dgp_Int32    ntok,
                   ErrCode      (*extbuilder)(token         **sorted_tokens,
                                              Dgp_Int32     ntok, 
                                              void          *extcookie,
                                              void          **extension,
                                              size_t        *extlen,
                                              Dgp_State     *dgp_state),
                   void         *extcookie,
                   void         *outbuf, 
                   size_t       *outbuflen,
                   Dgp_State    *dgp_state){
  DGP_PROLOGUE("tcmp_encodeTokDict()", DGP_PERRSTATE);
  bitfile           f;
  token             *t;
  Dgp_Int32         oldheight;
  Dgp_Int32         oldwidth;
  Dgp_Int32         i, j;
  token             **ordered = (token **)NULL;
  void              *extension;
  size_t            extensionlen;
  Dgp_Int32         classstart;
  Dgp_Int32         newntok;
  Dgp_Bool          singletons;
  Dgp_Allocators    *a = DGP_PALLOCATORS;

  void              *encbuf = NULL;
  size_t            encbuflen = 0;
    
  /* G4 Encoding Cookie. */
  void *g4Encoder = (void *)NULL;

#ifdef  INSTRUMENT
  dict_bituses bituses;
  bituses.overhead = bituses.padding = bituses.bitmapsize = bituses.bitmaps =
    bituses.height = bituses.width = bituses.uses = 0;
#endif

  /* Set up the output "file". */
  f.outbuf = (Dgp_UInt8 *)outbuf;
  f.outbufleft = *outbuflen;
  f.buf = 0;
  f.bitsinbuf = 0;
    
  /* Initialize the g4encoder. */
  /* Yo, this is kinda lame that we use stderr, but the whole rest of
   * this function uses DEBUG which is just stderr too... */
  DGP_ERR_PASS_CLEANUPF0(EncodeCreate(&g4Encoder, dgp_state, NULL), "tcmp_encodeTokDict() / Failed to create encoder.\n");
    
  /*
   * How many of the tokens are new?
   * Also, do we have to encode any use counts of 0 or 1?
   */
  newntok = 0;
  singletons = DGP_FALSE;
  for (i = 0; i < ntok; i++) {
    if (!tokens[i].committed) {
      newntok++;
      if (tokens[i].uses < 2) {
        singletons = DGP_TRUE;
      }
    }
  }
  
  DGP_CHECK_ALLOC_CLEANUP(ordered = (token **)(a->calloc)(newntok, sizeof(*ordered)));
    
  /* Read the new tokens into the "ordered" array for width/height sort. */
  for (i = 0, j = 0; i < ntok; i++) {
    if (!tokens[i].committed) {
      ordered[j] = &tokens[i];
      j++;
    }
  }
  
  /* Now sort the array. */
  qsort((void *)ordered, (unsigned)newntok, sizeof(token *), tokGT);
    
  /* Build the extension */
  if (extbuilder != (ErrCode (*)(token **, Dgp_Int32, void *, void **, size_t *, Dgp_State *))NULL) {
    DGP_ERR_PASS_CLEANUP((*extbuilder)(ordered, newntok, extcookie, &extension, &extensionlen, dgp_state));
  } else {
    extension = (void *)NULL;
    extensionlen = 0;
  }

  /* Now start writing stuff out. */

  /* Write out the token count. */
#ifdef  INSTRUMENT
  TCMP_BUF_CHECK(!writeValue((unsigned)newntok, &f , &bituses.overhead));
#else
  TCMP_BUF_CHECK(!writeValue((unsigned)newntok, &f));
#endif

  /*
   * Write out whether or not there were singletons, and any other flags
   * we might have at the start of the dictionary.
   * Right now, it's just a single bit: which useEncoding table did
   * we use (0 or 1)?
   */
#ifdef  INSTRUMENT
  bituses.overhead++;
#endif
  if (singletons) {
    TCMP_BUF_CHECK(!writeCode(0ul, 1, &f));
  } else {
    TCMP_BUF_CHECK(!writeCode(1ul, 1, &f));
  }

  /* Create default encoding buffer. */
  encbuflen = G4BUFLEN;
  DGP_CHECK_ALLOC_CLEANUP(encbuf = (a->malloc)(encbuflen));

  /* Write out all the class info, in the appropriate order. */
  oldheight = oldwidth = 0;
  classstart = 0;
  for (i = 0; i < newntok; i++) {
    t = ordered[i];
    if (t->height != oldheight) {
      /* New height class - write out the delta from the previous one. */
      DGP_ASSERT(t->height > oldheight);

      /* Write an end code for the previous height class, if any. */
      if (oldheight != 0) {
#ifdef  INSTRUMENT
  TCMP_BUF_CHECK(!writeWHEnd(&f , &bituses.width));
#else
  TCMP_BUF_CHECK(!writeWHEnd(&f));
#endif
  /* Write out the previous class g4-compressed */
#ifdef  INSTRUMENT
  DGP_ERR_PASS_CLEANUP_NO_MSG(writeg4Class(ordered + classstart,
                                           i - classstart,
                                           &f, 
                                           &bituses,
                                           g4Encoder,
                                           &encbuf, 
                                           &encbuflen, 
                                           dgp_state));
#else
  DGP_ERR_PASS_CLEANUP_NO_MSG(writeg4Class(ordered + classstart,
                                           i - classstart,
                                           &f,
                                           g4Encoder,
                                           &encbuf,
                                           &encbuflen,
                                           dgp_state));
#endif
      }

      /*
       * Write out the height delta from the previous class.
       * It's always at least 1, so subtract that.
       */
#ifdef  INSTRUMENT
      TCMP_BUF_CHECK(!writeWHDelta((unsigned)t->height - oldheight - 1, &f, &bituses.height));
#else
      TCMP_BUF_CHECK(!writeWHDelta((unsigned)t->height - oldheight - 1, &f));
#endif
      oldheight = t->height;
      oldwidth = 0;
      classstart = i;
    }

    /* We're now within a height class. Write out the width delta. */
    DGP_ASSERT(t->width >= oldwidth);
#ifdef  INSTRUMENT
    TCMP_BUF_CHECK(!writeWHDelta((unsigned)t->width - oldwidth, &f,
         &bituses.width));
#else
    TCMP_BUF_CHECK(!writeWHDelta((unsigned)t->width - oldwidth, &f));
#endif
    oldwidth = t->width;

    /* Write out the use count, using the appropriate table */
#ifdef  INSTRUMENT
    TCMP_BUF_CHECK(!writeUseCount(t->uses, &f, (singletons ? 0 : 1), &bituses.uses));
#else
    TCMP_BUF_CHECK(!writeUseCount(t->uses, &f, (singletons ? 0 : 1)));
#endif
  }
  /*
   * Write out an end token for the last height class
   */
#ifdef  INSTRUMENT
  TCMP_BUF_CHECK(!writeWHEnd(&f , &bituses.width));
#else
  TCMP_BUF_CHECK(!writeWHEnd(&f));
#endif

  /* Write out the last class g4-compressed */
#ifdef  INSTRUMENT
  DGP_ERR_PASS_CLEANUP_NO_MSG(writeg4Class(ordered + classstart,
                                           i - classstart,
                                           &f, 
                                           &bituses, 
                                           g4Encoder, 
                                           &encbuf, 
                                           &encbuflen,
                                           dgp_state));
#else
  DGP_ERR_PASS_CLEANUP_NO_MSG(writeg4Class(ordered + classstart,
                                           i - classstart, 
                                           &f,
                                           g4Encoder,
                                           &encbuf, 
                                           &encbuflen,
                                           dgp_state));
#endif

  /* Write out an end token to show it was the last height class */
#ifdef  INSTRUMENT
  TCMP_BUF_CHECK(!writeWHEnd(&f , &bituses.height));
#else
  TCMP_BUF_CHECK(!writeWHEnd(&f));
#endif

  /* cleanup encoding buffer */
  if(encbuf != NULL) {
    (a->free)(encbuf);
    encbuf = NULL;
    encbuflen = 0;
  }

  /*
   * Write out the extensions section.
   */

  /* We don't write the extensionlen anymore. */
#if 0
#ifdef  INSTRUMENT
  TCMP_BUF_CHECK(!writeValue((unsigned)extensionlen, &f , &bituses.overhead));
#else
  TCMP_BUF_CHECK(!writeValue((unsigned)extensionlen, &f));
#endif
#endif

  /* Done with this bit-packed part of the file */
#ifdef  INSTRUMENT
  bituses.padding += (BITS_ULONG - f.bitsinbuf) % BITS_UCHAR;
#endif
  TCMP_BUF_CHECK(!flushBitfile(&f));
    
  /* Copy in the extension block if any */
  if (extension != (void *)NULL) {
    TCMP_BUF_CHECK(f.outbufleft < extensionlen);
    memcpy(f.outbuf, extension, (unsigned)extensionlen);
    f.outbuf += extensionlen;
    f.outbufleft -= extensionlen;
  }
  else {
    DGP_ASSERT(extensionlen == 0);
  }
  
  /*
   * Now update the dictionary state block. We carefully haven't touched
   * it yet because any of the stuff above could have failed.
   */
  DGP_ERR_PASS_CLEANUPF0(internal_convertTokensToState(ntok, dstate, newntok, ordered),
       "internal convert bailed");

  /* Done with the encoder. */
  EncodeCleanUp( &g4Encoder);
  
  /* Done with the ordered table */
  (a->free)((void *)ordered);


#ifdef  INSTRUMENT
  fprintf(stderr, "Bit allocations for dictionary (%d tokens):\n\t\t\t\t%d overhead\n\t\t\t\t%d padding\n\t\t\t\t%d bitmap size\n\t\t\t\t%d bitmaps\n\t\t\t\t%d height class height\n\t\t\t\t%d token width\n\t\t\t\t%d token uses\n",
    ntok, bituses.overhead, bituses.padding, bituses.bitmapsize, bituses.bitmaps, bituses.height, bituses.width, bituses.uses);
#endif

  /* How much space did we use? */
  *outbuflen -= f.outbufleft;
  NORMAL_RETURN;

ERR_CLEANUP:

  if(encbuf != NULL) {
    (a->free)(encbuf);
  }

  if( g4Encoder != (void *)NULL) {
    EncodeCleanUp( &g4Encoder);
  }

  if (ordered != (token **)NULL) {
    (a->free)((void *)ordered);
  }

  /* Should guess how much more space they should feed us. FIX */
  DGP_WARNF0("tcmp_encodeTokDict() bailing for more memory...\n");
  *outbuflen += 50000;

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  size_t tcmp_estimateEncodedTokDictSize(void *dstate, token *tokens, Dgp_Int32 ntok)
**
**
**  This routine takes a dictionary state and a bunch of tokens and figures out
**  how much memory (in bytes) is required to encode those tokens. This is only
**  an estimate - see tcmp_encodeTokDict().
**
**	Arguments:
**      o  dstate...
**      o  tokens...
**      o  ntok is length(tokens)...
**
**	Return value:
**		o  Estimate, in bytes.
**
**	Modification history:
**      o  None.
**  
**/

size_t
tcmp_estimateEncodedTokDictSize(void *dstate, token *tokens, Dgp_Int32 ntok){
  Dgp_Int32 i;
  size_t totsize = 0;

  /* They're used, really. */
  /* Not clear where;  but this is to gag compiler whining. */
  dstate = dstate;

  for (i = 0; i < ntok; i++) {
    if (!tokens[i].committed) {
      /* Going to encode this one. */
      totsize += TOKG4SIZE + tokens[i].width * tokens[i].height / BITS_UCHAR / TOKG4COMP;
    }
  }

  /* Account for some overhead */
  totsize += TOKOVERHEAD;

#ifdef DGP_MMTOK_DBUG
  DGP_REPORTF1("tcmp_estimateEncodedTokDictSize() / estimating %ld bytes for dict\n", totsize);
#endif
  return(totsize);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Int32 comparePos1(const void *p1, const void *p2)
**
**
**  Compare the two positions, in reading order. Return 1 if p1 comes 
**  after p2, -1 if it comes before. 0 if they have the same position.
** 
**  p1 and p2 are tokpos **, that is pointers to pointers to tokens.
**
**	Arguments:
**      o  p1, p2...
**
**	Return value:
**		o  Signed integer comparison value.
**
**	Modification history:
**      o  None.
**  
**/

typedef const tokpos *tokposPtr;
static int
comparePos1(const void *p1, const void *p2) {
  /*tokpos pos1 = **(tokpos **)p1; */
  /*tokpos pos2 = **(tokpos **)p2; */
  const tokpos *pos1 = *((const tokposPtr *)p1);
  const tokpos *pos2 = *((const tokposPtr *)p2);
  Dgp_Int32 s1 = pos1->ypos;
  Dgp_Int32 s2 = pos2->ypos;

  if ( s1 > s2) {
    return 1;
  } else if( s1 == s2) {
    if( pos1->xpos > pos2->xpos) {
      return 1;
    } else if( pos1->xpos == pos2->xpos) {
      return 0;
    } else {
      return -1;
    }
  } else {
    return -1;
  }
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool comparePos1Bool(void *p1, void *p2)
**
**
**  Same thing as comparePos1(), but returning a Dgp_Bool (DGP_TRUE means ">"), and the Y coords
**  are known to match.
**
**  p1 and p2 are tokpos *.
**
**
**	Arguments:
**      o  p1, p2...
**
**	Return value:
**		o  Boolean comparison value;  no ties.
**
**	Modification history:
**      o  None.
**  
**/

static Dgp_Bool
comparePos1Bool(void *p1, void *p2) {
  DGP_VOID_PROLOGUE("comparePos1Bool()", NULL);
  tokpos pos1 = *(tokpos *)p1;
  tokpos pos2 = *(tokpos *)p2;
    
  DGP_VOID_ASSERT( pos1.ypos == pos2.ypos);

  if ((pos1.xpos > pos2.xpos) || (pos1.xpos == pos2.xpos)) {
    return(DGP_TRUE);
  } else {
    return(DGP_FALSE);
  }
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Int32 comparePos2(const void *p1, const void *p2)
**
**
**  Compare the two positions, in reading order, but chopping the lower bit
**  off their y-positions. 1 if p1 comes after p2, -1 if it comes before. 
**  0 for ties.
** 
**  p1 and p2 are tokpos **.
**
**
**	Arguments:
**      o  p1, p2...
**
**	Return value:
**		o  Signed integer comparison value.
**
**	Modification history:
**      o  None.
**  
**/

static int
comparePos2(const void *p1, const void *p2) {
  const tokpos *pos1 = *(const tokposPtr *)p1;
  const tokpos *pos2 = *(const tokposPtr *)p2;
  Dgp_Int32 s1 = pos1->ypos / 2;
  Dgp_Int32 s2 = pos2->ypos / 2;

  if ( s1 > s2) {
    return 1;
  } else if( s1 == s2) {
    if( pos1->xpos > pos2->xpos) {
      return 1;
    } else if( pos1->xpos == pos2->xpos) {
      return 0;
    } else {
      return -1;
    }
  } else {
    return -1;
  }
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Int32 comparePos4(const void *p1, const void *p2)
**
**
**  Same as comparePos2(), but ignore the bottom 2 bits of the y position.
**
**	Arguments:
**      o  p1, p2...
**
**	Return value:
**		o  Signed integer comparison value.
**
**	Modification history:
**      o  None.
**  
**/

static int
comparePos4(const void *p1, const void *p2) {
  const tokpos *pos1 = *(const tokposPtr *)p1;
  const tokpos *pos2 = *(const tokposPtr *)p2;
  Dgp_Int32 s1 = pos1->ypos / 4;
  Dgp_Int32 s2 = pos2->ypos / 4;

  if ( s1 > s2) {
    return 1;
  } else if( s1 == s2) {
    if( pos1->xpos > pos2->xpos) {
      return 1;
    } else if( pos1->xpos == pos2->xpos) {
      return 0;
    } else {
      return -1;
    }
  } else {
    return -1;
  }
}
/*************************************************************************/

#define  MAXSTRIP   4
#define  MAXMODE    16

/* A table of the sorting functions. */
static int (*compares[3])(const void *, const void *) = {
  comparePos1, comparePos2, comparePos4
};

/*************************************************************************
**
**  Function:  static Dgp_Int32 countPosEncoding(tokpos         **sortpos, 
**                                               Dgp_Int32      npos,
**                                               Dgp_Int32      highx, 
**                                               Dgp_Int32      highy,
**                                               Dgp_Int32      stripheight, 
**                                               Dgp_Int32      logstripheight, 
**                                               Dgp_Bool       transposed,
**                                               Dgp_Int32      *modex, 
**                                               Dgp_Int32      *firstXEncoding,
**                                               Dgp_Int32      *deltaXEncoding, 
**                                               Dgp_Int32      *deltaYEncoding)
**
**
**  Figure out how many bits it takes to encode the position info
**  of the tokens pointed to by sortpos[0..npos-1].  They should be grouped
**  into strips of height stripheight.
**
**  If transposed is DGP_TRUE, then the xpos and ypos values in each tokpos
**  have been switched, and the reference coordinate is the upper left.
**  This means we have to space forward by the height of each token, not
**  its width.
** 
**	Arguments:
**      o  sortpos...
**      o  npos...
**      o  highx, highy...
**      o  stripheight...
**      o  logstripheight...
**      o  transposed...
**      o  modex...
**      o  firstXEncoding...
**      o  deltaXEncoding...
**      o  deltaYEncoding...
**
**	Return value:
**		o  Total number of bits counted.
**
**	Modification history:
**      o  None.
**  
**/

static Dgp_Int32
countPosEncoding(tokpos         **sortpos, 
                 Dgp_Int32      npos,
                 Dgp_Int32      highx, 
                 Dgp_Int32      highy,
                 Dgp_Int32      stripheight, 
                 Dgp_Int32      logstripheight, 
                 Dgp_Bool       transposed,
                 Dgp_Int32      *modex, 
                 Dgp_Int32      *firstXEncoding,
                 Dgp_Int32      *deltaXEncoding, 
                 Dgp_Int32      *deltaYEncoding) {
  DGP_VOID_PROLOGUE("countPosEncoding()", NULL);
  Dgp_Int32 i, j;
  Dgp_Int32 modehist[MAXMODE];
  Dgp_Int32 deltax;
  Dgp_Int32 mode;
  Dgp_Int32 modeval;
  Dgp_Int32 firstxbits[NFIRSTX];
  Dgp_Int32 deltaxbits[NDELTAX];
  Dgp_Int32 deltaybits[NDELTAY];
  Dgp_Int32 firstx;
  Dgp_Int32 lastx;
  Dgp_Int32 lasty;
  Dgp_Int32 bestbits;
  Dgp_Int32 totbits;

  /* Please don't complain they're unused. */
  highx = highx;
  highy = highy;
  stripheight = stripheight;

  /* First thing: scan for the modal deltax value */
  for (i = 0; i < MAXMODE; i++) {
    modehist[i] = 0;
  }

  for (i = 1; i < npos; i++) {
    if ((sortpos[i - 1]->ypos >> logstripheight) == (sortpos[i]->ypos >> logstripheight)) {
      /* They're in the same strip - accumulate in the mode histogram */
      deltax = sortpos[i]->xpos - (sortpos[i - 1]->xpos + (transposed ? sortpos[ i - 1]->tok->height :
               sortpos[ i - 1]->tok->width));
      if ((deltax >= 0) && (deltax < MAXMODE)) {
        modehist[deltax]++;
      }
    }
  }

  modeval = modehist[0];
  mode = 0;
  for (i = 1; i < MAXMODE; i++) {
    if (modehist[i] > modeval) {
      modeval = modehist[i];
      mode = i;
    }
  }
  *modex = mode;

  /* Got the modal deltax value. Now scan, building bitcounts. */
  for (i = 0; i < NFIRSTX; i++) {
    firstxbits[i] = 0;
  }
  for (i = 0; i < NDELTAX; i++) {
    deltaxbits[i] = 0;
  }
  for (i = 0; i < NDELTAY; i++) {
    deltaybits[i] = 0;
  }

  lasty = -1;
  firstx = lastx = 0;
  for (i = 0; i < npos; i++) {
    if ((i == 0) || ((sortpos[i]->ypos >> logstripheight) != (lasty >> logstripheight))) {
      DGP_VOID_ASSERT(lasty < sortpos[i]->ypos);
      /* Start of a new strip. */

      /*
       * Encode the first x value as a delta off the previous firstx
       * We're actually folding the bits needed to encode the y position
       * within the strip into the xbits values as well.
       * We're also coding the EOS bits into the firstxbits.
       */
      for (j = 0; j < NFIRSTX; j++) {
        firstxbits[j] += encodeCount(sortpos[i]->xpos - firstx, &mmFirstXCoders[j]) +
                         logstripheight + mmDeltaXCoders[j].EOSbits;
      }
      firstx = sortpos[i]->xpos;
      lastx = sortpos[i]->xpos + (transposed ? sortpos[ i]->tok->height : sortpos[ i]->tok->width);

      /* Encode deltay. */
      for (j = 0; j < NDELTAY; j++) {
        deltaybits[j] += encodeCount((sortpos[i]->ypos >> logstripheight) - (lasty >> logstripheight), &mmDeltaYCoders[j]);
      }
      /* lasty should always refer to the top of the strip. */
      lasty = (sortpos[i]->ypos >> logstripheight) << logstripheight;
    } else {
      /* Within a strip. */
      deltax = sortpos[i]->xpos - lastx - mode;
      for (j = 0; j < NDELTAX; j++) {
        deltaxbits[j] += encodeCount(deltax, &mmDeltaXCoders[j]) + logstripheight;
      }
      lastx = sortpos[i]->xpos + (transposed ? sortpos[ i]->tok->height : sortpos[ i]->tok->width);
    }
  }

  /* Found all the bit counts. */
  bestbits = firstxbits[0];
  *firstXEncoding = 0;
  for (i = 1; i < NFIRSTX; i++) {
    if (firstxbits[i] < bestbits) {
      bestbits = firstxbits[i];
      *firstXEncoding = i;
    }
  }
  totbits = bestbits;

  bestbits = deltaxbits[0];
  *deltaXEncoding = 0;
  for (i = 1; i < NDELTAX; i++) {
    if (deltaxbits[i] < bestbits) {
      bestbits = deltaxbits[i];
      *deltaXEncoding = i;
    }
  }
  totbits += bestbits;

  bestbits = deltaybits[0];
  *deltaYEncoding = 0;
  for (i = 1; i < NDELTAY; i++) {
    if (deltaybits[i] < bestbits) {
      bestbits = deltaybits[i];
      *deltaYEncoding = i;
    }
  }
  totbits += bestbits;

  return(totbits);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static ErrCode binSort1(tokpos           **sortpos, 
**                                     Dgp_Int32        npos, 
**                                     Dgp_Int32        maxx, 
**                                     Dgp_Int32        maxy, 
**                                     Dgp_State        *dgp_state)
**
**
**  Bin sort the whole thing by Y, X and token number.
** 
**	Arguments:
**      o  sortpos...
**      o  npos...
**      o  maxx, maxy...
**      o  dgp_state points to prepared Dgp_State{}.
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  Scattered liberr compliance issues, particularly following revised coding conventions.
**  
**/

static ErrCode
binSort1(tokpos **sortpos, Dgp_Int32 npos, Dgp_Int32 maxx, Dgp_Int32 maxy, Dgp_State *dgp_state) {
  DGP_PROLOGUE("binSort1()", DGP_PERRSTATE);
  tokpos            **sortdup = (tokpos **)NULL;
  DgpAdt_List       *ybins = (DgpAdt_List *)NULL;
  DgpAdt_List       *xbins = (DgpAdt_List *)NULL;
  DgpAdt_ListNode   retnode;
  Dgp_Int32         i, j;
  Dgp_Int32         sortnum;
  DgpAdt_ListNode   ln;
  DgpAdt_PriQ       binpq = DGP_NULLPRIQ;
  DgpAdt_PriQNode   pqn, retpriqn;
  Dgp_Allocators    *a = DGP_PALLOCATORS;


  /* Duplicate the stuff to be sorted. */
  DGP_CHECK_ALLOC_CLEANUP(sortdup = (tokpos **)(a->malloc)(npos * sizeof(*sortdup)));
  memcpy(sortdup, sortpos, npos * sizeof(*sortdup));
    
  /* Create the Y bins */
  DGP_CHECK_ALLOC_CLEANUP(ybins = (DgpAdt_List *)(a->malloc)(maxy * sizeof(*ybins)));

  /* Initialise them to empty, for clean recovery */
  for (i = 0; i < maxy; i++) {
    ybins[i] = DGP_NULLLIST;
  }
    
  /* and now set them all to empty lists */
  for (i = 0; i < maxy; i++) {
      DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&ybins[i], dgp_state));
  }
    
  /* Create the X bins */
  DGP_CHECK_ALLOC_CLEANUP(xbins = (DgpAdt_List *)(a->malloc)(maxx * sizeof(*xbins)));

  /* Initialise them to empty, for clean recovery */
  for (i = 0; i < maxx; i++) {
    xbins[i] = DGP_NULLLIST;
  }
    
  /* and now set them all to empty lists */
  for (i = 0; i < maxx; i++) {
      DGP_ERR_PASS_CLEANUP(DgpAdt_liNew(&xbins[i], dgp_state));
  }
    
  /* Fill in the Y bins */
  for (i = 0; i < npos; i++) {
      DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(ybins[sortpos[i]->ypos], (void *)sortdup[i], &retnode));
  }
    
  /* Now start reading them out */
  sortnum = 0;
  DGP_ERR_PASS_CLEANUP(DgpAdt_pqNew(comparePos1Bool, dgp_state, &binpq));

  for (i = 0; i < maxy; i++) {
    if (DGPADT_LILEN(ybins[i]) > 1) {

      /* We need to sort the things in the bin */
      if (DGPADT_LILEN(ybins[i]) < 10) {
        /* There aren't many - toss them into a pq */
        DGPADT_FOREACH (ln, ybins[i]) {
            DGP_ERR_PASS_CLEANUP(DgpAdt_pqAdd(binpq, DGPADT_LIGET(ln), &retpriqn));
        }
    
        while ((pqn = DGPADT_PQFIRST(binpq)) != DGP_NULLPRIQNODE) {
            sortpos[sortnum++] = (tokpos *)DGPADT_PQGET(pqn);
            DGP_ERR_PASS_CLEANUP(DgpAdt_pqRem(pqn));
        }
      } else {
              
        /* Do a secondary sort on X. */
        DGPADT_FOREACH (ln, ybins[i]) {
            DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(xbins[((tokpos *)DGPADT_LIGET(ln))->xpos], DGPADT_LIGET(ln), &retnode));
        }

        for (j = 0; j < maxx; j++) {
            /* Read through and clean out this x bin. */
            if (DGPADT_LILEN(xbins[j]) > 0) {
                if (DGPADT_LILEN(xbins[j]) > 1) {
                    /* Multiple things at this (x,y). */
                    for (ln = DGPADT_LIFIRST(xbins[j]); ln != DGP_NULLLISTNODE; ) {
                        DGP_ERR_PASS_CLEANUP(DgpAdt_pqAdd(binpq, DGPADT_LIGET(ln), &retpriqn));
                        DGP_ERR_PASS_CLEANUP(DgpAdt_liRemAndNext(ln, &ln));
                    }
    
                    while ((pqn = DGPADT_PQFIRST(binpq)) != DGP_NULLPRIQNODE) {
                        sortpos[sortnum++] = (tokpos *)DGPADT_PQGET(pqn);
                        DGP_ERR_PASS_CLEANUP(DgpAdt_pqRem(pqn));
                    }
                } else {
                    /* Only one thing here. */
                    sortpos[sortnum++] = (tokpos *)DGPADT_LIGET(DGPADT_LIFIRST(xbins[j]));
                    DGP_ERR_PASS_CLEANUP(DgpAdt_liRem(DGPADT_LIFIRST(xbins[j])));
                }
            }
        }
      }
    } else if (DGPADT_LILEN(ybins[i]) == 1) {
      sortpos[sortnum++] = (tokpos *)DGPADT_LIGET(DGPADT_LIFIRST(ybins[i]));
    }
  }
  DGP_ASSERT(sortnum == npos);

  for (i = 0; i < maxy; i++) {
    DgpAdt_liFree(ybins[i]);
  }
  (a->free)((void *)ybins);
  for (i = 0; i < maxx; i++) {
    DgpAdt_liFree(xbins[i]);
  }
  (a->free)((void *)xbins);
  (a->free)((void *)sortdup);
  DGP_ERR_PASS_CLEANUP(DgpAdt_pqFree(binpq));

  NORMAL_RETURN;

ERR_CLEANUP:

  if (sortdup != (tokpos **)NULL) {
    (a->free)((void *)sortdup);
  }

  if (ybins != (DgpAdt_List *)NULL) {
    for (i = 0; i < maxy; i++) {
      if (ybins[i] != DGP_NULLLIST) {
          DgpAdt_liFree(ybins[i]);
      }
    }
    (a->free)((void *)ybins);
  }

  if (xbins != (DgpAdt_List *)NULL) {
    for (i = 0; i < maxx; i++) {
      if (xbins[i] != DGP_NULLLIST) {
        DgpAdt_liFree(xbins[i]);
      }
    }
    (a->free)((void *)xbins);
  }

  if (binpq != DGP_NULLPRIQ) {
    DGP_ERR_PASS(DgpAdt_pqFree(binpq));
  }

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static ErrCode decidePosEncodingLots(tokpos            *positions, 
**                                                  Dgp_Int32         npos,
**                                                  Dgp_Int32         lowx, 
**                                                  Dgp_Int32         lowy, 
**                                                  Dgp_Int32         highx, 
**                                                  Dgp_Int32         highy,
**                                                  Dgp_Bool          *transpose,
**                                                  Dgp_Int32         *modex, 
**                                                  Dgp_Int32         *stripsize,
**                                                  Dgp_Int32         *logstripsize,
**                                                  Dgp_Int32         *firstXEncoding,
**                                                  Dgp_Int32         *deltaXEncoding,
**                                                  Dgp_Int32         *deltaYEncoding,
**                                                  tokpos            ***sortpos, 
**                                                  Dgp_State         *dgp_state)
**
**
**  Decide how the positions should be encoded, in the case that there are
**  lots of positions.  This takes the same set of params as decidePosEncoding,
**  except that it assumes that sortpos has already been allocated, and lowx
**  and lowy have already been subtracted off positions[...].
**
**  It uses some heuristics to decide the optimum encoding:
**  - it histograms the X and Y coordinates of the positions, and looks for
**    the highest peak, and decides the transposition based on this.
**  - 
**
** 
**	Arguments:
**      o  positions...
**      o  npos...
**      o  lowx, lowy...
**      o  highx, highy...
**      o  transpose...
**      o  modex...
**      o  stripsize...
**      o  logstripsize...
**      o  firstXEncoding...
**      o  deltaXEncoding...
**      o  deltaYEncoding...
**      o  sortpos...
**      o  dgp_state points to prepared Dgp_State{}.
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

static ErrCode
decidePosEncodingLots(tokpos            *positions, 
                      Dgp_Int32         npos,
                      Dgp_Int32         lowx, 
                      Dgp_Int32         lowy, 
                      Dgp_Int32         highx, 
                      Dgp_Int32         highy,
                      Dgp_Bool          *transpose,
                      Dgp_Int32         *modex, 
                      Dgp_Int32         *stripsize,
                      Dgp_Int32         *logstripsize,
                      Dgp_Int32         *firstXEncoding,
                      Dgp_Int32         *deltaXEncoding,
                      Dgp_Int32         *deltaYEncoding,
                      tokpos            ***sortpos, 
                      Dgp_State         *dgp_state){
  DGP_PROLOGUE("decidePosEncodingLots()", DGP_PERRSTATE);
  Dgp_Int32 *xhist = NULL;
  Dgp_Int32 *yhist = NULL;
  Dgp_Int32 maxhist;
  Dgp_Int32 i;
  Dgp_Allocators *a = DGP_PALLOCATORS;


  DGP_CHECK_ALLOC_CLEANUP(xhist = (Dgp_Int32 *)(a->calloc)((highx - lowx), sizeof(*xhist)));
  DGP_CHECK_ALLOC_CLEANUP(yhist = (Dgp_Int32 *)(a->calloc)((highy - lowy), sizeof(*yhist)));

  for (i = 0; i < npos; i++) {
    xhist[positions[i].xpos]++;
    yhist[positions[i].ypos]++;
  }

  /*
   * Look for the highest peak.  If it's in the X histogram, then we should
   * transpose the page.
   */
  maxhist = 0;
  *transpose = DGP_FALSE;
  for (i = 0; i < highy - lowy; i++) {
    if (yhist[i] > maxhist) {
      maxhist = yhist[i];
    }
  }

  for (i = 0; i < highx - lowx; i++) {
    if (xhist[i] > maxhist) {
      *transpose = DGP_TRUE;
      break;
    }
  }

  if (*transpose) {
    /* Switch the X and Y coords, for later sorting */
    for (i = 0; i < npos; i++) {
      Dgp_Int32 tmp = positions[i].xpos;
      /* Store the upper left corner, not the lower left */
      positions[i].xpos = positions[i].ypos - (positions[i].tok->height - 1);
      positions[i].ypos = (Dgp_Int16)tmp;
    }
  }

  /* Now decide the strip size. */
  /* For now, just pick a strip size of 1. */
  *stripsize = 1;
  *logstripsize = 0;

  /* Fill in *sortpos... */
  for (i = 0; i < npos; i++) {
    (*sortpos)[i] = &positions[i];
  }

  /* ...and sort it... */
  DGP_ERR_PASS_CLEANUP(binSort1(*sortpos, npos, *transpose ? (highy - lowy) : (highx - lowx),
                                *transpose ? (highx - lowx) : (highy - lowy), dgp_state));
  
  /* ...and find the best tables. */
  (void)countPosEncoding(*sortpos, npos, 
                         *transpose ? highy : highx,
                         *transpose ? highx : highy,
                         *stripsize, 
                         *logstripsize, 
                         *transpose,
                         modex, 
                         firstXEncoding,
                         deltaXEncoding,
                         deltaYEncoding);

  if (*transpose) {
    /* Switch the X and Y coords back */
    for (i = 0; i < npos; i++) {
      Dgp_Int32 tmp = positions[i].ypos;
      /* Store the lower left corner, not the upper left */
      positions[i].ypos = positions[i].xpos + (positions[i].tok->height - 1);
      positions[i].xpos = (Dgp_Int16)tmp;
    }
  }
    
  (a->free)((void *)xhist);
  (a->free)((void *)yhist);
    
  NORMAL_RETURN;

ERR_CLEANUP:
  if (xhist != (Dgp_Int32 *)NULL) {
    (a->free)((void *)xhist);
  }

  if (yhist != (Dgp_Int32 *)NULL) {
    (a->free)((void *)yhist);
  }

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static ErrCode decidePosEncoding(tokpos            *positions, 
**                                              Dgp_Int32         npos,
**                                              Dgp_Int32         lowx, 
**                                              Dgp_Int32         lowy, 
**                                              Dgp_Int32         highx, 
**                                              Dgp_Int32         highy,
**                                              Dgp_Bool          *transpose,
**                                              Dgp_Int32         *modex, 
**                                              Dgp_Int32         *stripsize, 
**                                              Dgp_Int32         *logstripsize,
**                                              Dgp_Int32         *firstXEncoding,
**                                              Dgp_Int32         *deltaXEncoding, 
**                                              Dgp_Int32         *deltaYEncoding,
**                                              tokpos            ***sortpos, 
**                                              Dgp_State         *dgp_state)
**
**
**  Decide how the positions should be encoded. Figure out which firstX
**  table, deltaX table, and deltaY table need to be used. Figure out the
**  strip height and the most common deltax value.
**
**  The strip height affects all the other things, so for each strip height
**  we need to re-sort the token positions and find the best encoding tables.
**
**  We return in *sortpos a table of the sorted positions.  The caller should
**  free *sortpos.
**
** 
**	Arguments:
**      o  positions...
**      o  npos...
**      o  lowx, lowy...
**      o  highx, highy...
**      o  transpose...
**      o  modex...
**      o  stripsize...
**      o  logstripsize...
**      o  firstXEncoding...
**      o  deltaXEncoding...
**      o  deltaYEncoding...
**      o  sortpos...
**      o  dgp_state points to prepared Dgp_State{}.
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

static ErrCode
decidePosEncoding(tokpos            *positions, 
                  Dgp_Int32         npos,
                  Dgp_Int32         lowx, 
                  Dgp_Int32         lowy, 
                  Dgp_Int32         highx, 
                  Dgp_Int32         highy,
                  Dgp_Bool          *transpose,
                  Dgp_Int32         *modex, 
                  Dgp_Int32         *stripsize, 
                  Dgp_Int32         *logstripsize,
                  Dgp_Int32         *firstXEncoding,
                  Dgp_Int32         *deltaXEncoding, 
                  Dgp_Int32         *deltaYEncoding,
                  tokpos            ***sortpos, 
                  Dgp_State         *dgp_state) {
  DGP_PROLOGUE("decidePosEncoding()", DGP_PERRSTATE);
  Dgp_Int32             i, j, k;
  Dgp_Int32             bestbits, curbits;
  Dgp_Int32             curModex, curFirstXEncoding, curDeltaXEncoding, curDeltaYEncoding;
  Dgp_Int32             bestj;
  Dgp_Allocators        *a = DGP_PALLOCATORS;

  *sortpos = (tokpos **)NULL;

  /* Get the ordering buffer */
  DGP_CHECK_ALLOC_CLEANUP(*sortpos = (tokpos **)(a->calloc)(npos, sizeof(**sortpos)));

  /*
   * Run through the tokens and adjust all the y positions down by
   * lowy.  We're going to undo this at the end of the routine.  Why?
   * Well, we'd like the encoding to be independent of lowy, and if we
   * don't make our y coordinates start at 0, we'll have problems if they
   * change lowy: the dividing lines between strips will change.
   *
   * We also subtract off lowx since otherwise the encodings for firstx
   * values won't be right.
   */
  for (k = 0; k < npos; k++) {
    positions[k].ypos -= (Dgp_Int16)lowy;
    positions[k].xpos -= (Dgp_Int16)lowx;
  }

  if (npos > LOTSOFPOS) {
    /* Handle large pages specially */
    bestbits = -1;
    DGP_ERR_PASS_CLEANUP(decidePosEncodingLots(positions, 
                                               npos, 
                                               lowx, 
                                               lowy,
                                               highx, 
                                               highy, 
                                               transpose,
                                               modex,
                                               stripsize,
                                               logstripsize, 
                                               firstXEncoding,
                                               deltaXEncoding,
                                               deltaYEncoding,
                                               sortpos, 
                                               dgp_state));
  } else {

    /* Check out the power of 2 height strips */
    bestbits = 0;
    bestj = 0;
    *modex = *firstXEncoding = *deltaXEncoding = *deltaYEncoding = 0;
    *transpose = DGP_FALSE;
    for (i = 1, j = 0; i <= MAXSTRIP; i <<= 1, j++) {
      /* Sort according to compares[j]... */
  
      for (k = 0; k < npos; k++) {
        (*sortpos)[k] = &positions[k];
      }

      qsort((void *)(*sortpos), (unsigned)npos, sizeof(tokpos *), compares[j]);

      /* ...and find the best tables */
      curbits = countPosEncoding(*sortpos, 
                                 npos, 
                                 highx, 
                                 highy,
                                 i, 
                                 j, 
                                 DGP_FALSE, 
                                 &curModex, 
                                 &curFirstXEncoding,
                                 &curDeltaXEncoding, 
                                 &curDeltaYEncoding);
      if ((j == 0) || (curbits < bestbits)) {
          *transpose = DGP_FALSE;
          *modex = curModex;
          *stripsize = i;
          *logstripsize = j;
          *firstXEncoding = curFirstXEncoding;
          *deltaXEncoding = curDeltaXEncoding;
          *deltaYEncoding = curDeltaYEncoding;
          bestbits = curbits;
          bestj = j;
      }
    }
      
      

    /* Switch the X and Y coords and try again. */
    for (k = 0; k < npos; k++) {
      Dgp_Int32 tmp = positions[k].xpos;
      /* Store the upper left corner, not the lower left */
      positions[k].xpos = positions[k].ypos - (positions[k].tok->height -1);
      positions[k].ypos = (Dgp_Int16)tmp;
    }

    for (i = 1, j = 0; i <= MAXSTRIP; i <<= 1, j++) {
      /* Sort according to compares[j]... */
  
      for (k = 0; k < npos; k++) {
          (*sortpos)[k] = &positions[k];
      }

      qsort((void *)(*sortpos), (unsigned)npos, sizeof(tokpos *), compares[j]);

      /* ...and find the best tables. */
      curbits = countPosEncoding(*sortpos, 
                                 npos, 
                                 highx, 
                                 highy,
                                 i, 
                                 j, 
                                 DGP_TRUE, 
                                 &curModex, 
                                 &curFirstXEncoding,
                                 &curDeltaXEncoding, 
                                 &curDeltaYEncoding);
      if (curbits < bestbits) {
          *transpose = DGP_TRUE;
          *modex = curModex;
          *stripsize = i;
          *logstripsize = j;
          *firstXEncoding = curFirstXEncoding;
          *deltaXEncoding = curDeltaXEncoding;
          *deltaYEncoding = curDeltaYEncoding;
          bestbits = curbits;
          bestj = j;
      }
    }
    
    /* Rebuild the sortpos table for the best encoding if required. */
    if (*transpose) {
      if (*stripsize != MAXSTRIP) {
        for (k = 0; k < npos; k++) {
            (*sortpos)[k] = &positions[k];
        }     

        qsort((void *)(*sortpos), (unsigned)npos, sizeof(tokpos *), compares[bestj]);
      }
    }

    /* Switch the coordinates back. */
    for (k = 0; k < npos; k++) {
      Dgp_Int32 tmp = positions[k].ypos;
      /* Store the lower left corner, not the upper left */
      positions[k].ypos = positions[k].xpos + (positions[k].tok->height - 1);
      positions[k].xpos = (Dgp_Int16)tmp;
    }

    /* Rebuild the sortpos table if we chose a non-transposed encoding. */
    if (!*transpose) {
      for (k = 0; k < npos; k++) {
        (*sortpos)[k] = &positions[k];
      }

      qsort((void *)(*sortpos), (unsigned)npos, sizeof(tokpos *), compares[bestj]);
    }
  }

  /* Undo the position shift. */
  for (k = 0; k < npos; k++) {
    positions[k].ypos += (Dgp_Int16)lowy;
    positions[k].xpos += (Dgp_Int16)lowx;
  }

  /* Don't free *sortpos - our caller gets it. */

#ifdef  INSTRUMENT
  fprintf(stderr, "decided: transpose = %d, stripsize = %d => modex = %d, firstX = %d, deltaX = %d, deltaY = %d, bits = %d\n",
    *transpose ? 1 : 0,
    *stripsize, *modex, *firstXEncoding,
    *deltaXEncoding, *deltaYEncoding, bestbits);
#endif
  NORMAL_RETURN;

ERR_CLEANUP:

  if (*sortpos != (tokpos **)NULL) {
    free((void *)*sortpos);
  }

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool writeTokenHuffTable(dictstate           *state, 
**                                               [ tokstate            **tokbynum, 
**                                                 Dgp_Int32           mintoknum, ]?
**                                                 tokpos              *positions, 
**                                                 Dgp_Int32           npos, 
**                                                 bitfile             *f
**                                              [, pos_bituses         *uses]
**
**
**  Looks decommissioned... / dhtesq.
** 
**	Arguments:
**      o  state...
**      o  tokbynum?...
**      o  mintoknum?...
**      o  positions...
**      o  npos...
**      o  f...
**      o  (#ifdef INSTRUMENT) uses...
**
**	Return value:
**		o  ErrCode.
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  Returns ErrCode but was prototyped as boolean.  I junk this?
**  
**/

#if 0
static ErrCode
writeTokenHuffTable(dictstate           *state, 
                 /* tokstate            **tokbynum, 
                    Dgp_Int32           mintoknum, */
                    tokpos              *positions, 
                    Dgp_Int32           npos, 
                    bitfile             *f
#ifdef  INSTRUMENT
                  , pos_bituses         *uses
#endif
                                             ) {
  DGP_PROLOGUE("writeTokenHuffTable()", NULL);
  Dgp_Int32         i, j;
  Dgp_Int32         nused;
  Huff              *usecodes = (Huff *)NULL;
  Huff              lencodes[35];
  Dgp_Int32         lenuses[35];
  Dgp_Int32         lenmap[35];
  Dgp_Int32         currun, curlen;

  for (i = 0; i < state->ntok; i++) {
    state->toks[i].uses = 0;
  }

  for (i = 0; i < npos; i++) {
    /* (tokbynum[positions[i].toknum - mintoknum])->uses++; */
    positions[i].tok->uses++;
  }

  DGP_CHECK_ALLOC_CLEANUP(usecodes = (Huff *)(a->malloc)(state->ntok * sizeof(*usecodes)));

  /*
   * Write in the use codes for the tokens that are used at all.
   */
  for (i = 0, j = 0; i < state->ntok; i++) {
    if (state->toks[i].uses > 0) {
      usecodes[j++].value = state->toks[i].uses;
    }
  }
  nused = j;

  /* Figure out the Huffman codes for each token */
  DGP_ERR_GEN_CLEANUP((!huffmanCode(usecodes, nused)), cErrInternal);
    
  /* and copy those into the dictionary */
  for (i = 0, j = 0; i < state->ntok; i++) {
    if (state->toks[i].uses > 0) {
      state->toks[i].code = usecodes[j].code;
      state->toks[i].codelen = usecodes[j].codelen;
      j++;
    }
  }

#ifdef  DGP_MMTOK_DBUG
  {
    Dgp_Int32 m1 = 0, m2 = 0, c = 0;
    Dgp_Int32 hist[33];

    for (i = 0; i < state->ntok; i++) {
      c += state->toks[i].uses;
      m1 += state->toks[i].uses * state->toks[i].codelen;
    }
    for (i = 0; i < npos; i++) {
      /* m2 += tokbynum[positions[i].toknum - mintoknum]->codelen; */
      m2 += positions[i].tok->codelen;
    }
    DGP_REPORTF2("npos = %d, c = %d,", npos, c);
    DGP_REPORTF2("m1 = %d, m2 = %d\n", m1, m2);

    for (i = 0; i < 33; i++) {
      hist[i] = 0;
    }
    for (i = 0; i < state->ntok; i++) {
      if (state->toks[i].uses) {
          hist[state->toks[i].codelen]++;
      }
      else {
          hist[0]++;
      }
    }
    for (i = 0; i < 33; i++) {
        DGP_REPORTF2("%2d:  %4d\n", i, hist[i]);      
    }

  }
#endif

  /* Now we want to write out the codes. */
  for (i = 0; i < 35; i++) {
    lenuses[i] = 0;
  }

  /*
   * We're going to write out the use counts for each token in this
   * position block, in order to build a Huffman tree to decode the
   * token numbers.  Well, we really only need to build a Huffman tree
   * from the use counts, then write out the lengths of the codes assigned;
   * the decoder can look at those lengths and infer what the Huffman tree
   * was.
   * These use count code lengths get run-coded; the run codes
   * themselves get Huffman-coded.  This block figures out what the
   * run code occurrence counts are, to build that Huffman table.
   *
   * The codes used in the run coding are:
   * 0-31:    a use count code length (0 means "unused")
   * 32+xx:    repeat the previous code length xx+3 times
   * 33+xxx:    repeat a code length of 0 xxx+3 times
   * 34+xxxxxxx:  repeat a code length of 0 xxx+11 times
   *
   * The run-encoding state machine is:
   * Look at the next use count code length to be encoded.
   * If it is not zero, and doesn't match the previous one, terminate the
   * current run and write it out.
   * If it is not zero and does match the previous one, start a new run
   * if there isn't one active; if there is one active, add to it; if it
   * maxes out (max run len=6), output it and close it.
   * If it is zero and doesn't match the previous one, terminate the
   * current run (if any), and start a new one.
   * If it is zero and does match the previous one, add to the current run;
   * if the run maxes out (max run len=138), output it and close it.
   *
   * We count the number of times each code in this run code occurs,
   * build a Huffman table, and write out the Huffman table as 34 4-bit
   * values (the code lengths in that table).
   */
  currun = 0;
  curlen = -1;
  for (i = 0, j = 0; i < state->ntok; i++) {
    if (state->toks[i].uses > 0) {
      if (usecodes[j].codelen == curlen) {
          currun++;
          if (currun == 6) {
            lenuses[32]++;
            currun = 0;
          }
      } else {
        if (curlen > 0) {
            if (currun > 2) {
                lenuses[32]++;
            } else {
                lenuses[curlen] += currun;
            }
        } else {
            if (currun < 3) {
                lenuses[0] += currun;
            } else if (currun < 11) {
                lenuses[33]++;
            } else {
                lenuses[34]++;
            }
        }
        curlen = usecodes[j].codelen;
        currun = 0;
        lenuses[curlen]++;
      }
      j++;
    } else {
      if (curlen == 0) {
          if (currun == 137) {
              lenuses[34]++;
              currun = 0;
          } else {
              currun++;
          }
      } else {
          if (currun > 0) {
              if (currun > 2) {
                  lenuses[32]++;
              } else {
                  lenuses[curlen] += currun;
              }
          }
          currun = 1;
          curlen = 0;
      }
    }
  }

  /* Finish off any unclosed runs. */
  if (currun > 0) {
    if (curlen == 0) {
      if (currun > 10) {
          lenuses[34]++;
      } else if (currun > 2) {
          lenuses[33]++;
      } else {
          lenuses[0] += currun;
      }
    } else {
        if (currun > 2) {
            lenuses[32]++;
        } else {
            lenuses[curlen] += currun;
        }
    }
  }

  /* Eliminate unused codes */
  for (i = 0, j = 0; i < 35; i++) {
    if (lenuses[i] > 0) {
      lencodes[j].value = lenuses[i];
      lenmap[i] = j++;
    } else {
      lenmap[i] = -1;
    }
  }

  /*
   * lencodes now holds the use counts for the run-length coding that
   * will be used to transmit the token code bit lengths.
   */
  DGP_ERR_GEN_CLEANUP((!huffmanCode(lencodes, j)), cErrInternal);

  /* Now we have the Huffman code for the run-length codes - write it out. */
  for (i = 0; i < 35; i++) {
    if (lenmap[i] < 0) {
      /* This one wasn't used at all. */
      DGP_ASSERT(lenuses[i] == 0);
      TCMP_BUF_CHECK(!writeCode(0ul, 4, f));
#ifdef  INSTRUMENT
      uses->tokhuff += 4;
#endif
    } else {
      /* This one got used - make sure it didn't get a >8bit code */
      DGP_ASSERT(lencodes[lenmap[i]].codelen <= 15);
      TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[i]].codelen, 4, f));
#ifdef  INSTRUMENT
      uses->tokhuff += 4;
#endif
    }
  }

  /* Now redo the run-coding, but writing this time */
  currun = 0;
  curlen = -1;
  for (i = 0, j = 0; i < state->ntok; i++) {
    if (state->toks[i].uses > 0) {
      if (usecodes[j].codelen == curlen) {
          currun++;
          if (currun == 6) {
              TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[32]].code, lencodes[lenmap[32]].codelen, f));
              TCMP_BUF_CHECK(!writeCode(currun - 3, 2, f));
#ifdef  INSTRUMENT
              uses->tokhuff += lencodes[lenmap[32]].codelen + 2;
#endif
              currun = 0;
          }
      } else {
          if (curlen > 0) {
              if (currun > 2) {
                  TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[32]].code, lencodes[lenmap[32]].codelen, f));
                  TCMP_BUF_CHECK(!writeCode(currun - 3, 2, f));
#ifdef  INSTRUMENT
                  uses->tokhuff += lencodes[lenmap[32]].codelen + 2;
#endif
              } else {
                  for (; currun > 0; currun--) {
                      TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[curlen]].code, lencodes[lenmap[curlen]].codelen, f));
#ifdef  INSTRUMENT
                      uses->tokhuff += lencodes[lenmap[curlen]].codelen;
#endif
                  }
              }
          } else {
              if (currun < 3) {
                  for (; currun > 0; currun--) {
                      TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[0]].code, lencodes[lenmap[0]].codelen, f));
#ifdef  INSTRUMENT
                      uses->tokhuff += lencodes[lenmap[0]].codelen + 2;
#endif
                  }
              } else if (currun < 11) {
                  TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[33]].code, lencodes[lenmap[33]].codelen, f));
                  TCMP_BUF_CHECK(!writeCode(currun - 3, 3, f));
#ifdef  INSTRUMENT
                  uses->tokhuff += lencodes[lenmap[33]].codelen + 3;
#endif
              } else {
                  TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[34]].code, lencodes[lenmap[34]].codelen, f));
                  TCMP_BUF_CHECK(!writeCode(currun - 11, 7, f));
#ifdef  INSTRUMENT
                  uses->tokhuff += lencodes[lenmap[34]].codelen + 7;
#endif
              }
          }

  curlen = usecodes[j].codelen;
  currun = 0;
  TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[curlen]].code, lencodes[lenmap[curlen]].codelen, f));
#ifdef  INSTRUMENT
  uses->tokhuff += lencodes[lenmap[curlen]].codelen;
#endif
      }
      j++;
    } else {
        if (curlen == 0) {
            if (currun == 137) {
                TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[34]].code, lencodes[lenmap[34]].codelen, f));
                TCMP_BUF_CHECK(!writeCode(currun - 11, 7, f));
#ifdef  INSTRUMENT
                uses->tokhuff += lencodes[lenmap[34]].codelen + 7;
#endif
                currun = 0;
            } else {
                currun++;
            }
        } else {
            if (currun > 0) {
                if (currun > 2) {
                    TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[32]].code, lencodes[lenmap[32]].codelen, f));
                    TCMP_BUF_CHECK(!writeCode(currun - 3, 2, f));
#ifdef  INSTRUMENT
                    uses->tokhuff += lencodes[lenmap[32]].codelen + 2;
#endif
                } else {
                    for (; currun > 0; currun--) {
                        TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[curlen]].code, lencodes[lenmap[curlen]].codelen, f));
#ifdef  INSTRUMENT
                        uses->tokhuff += lencodes[lenmap[curlen]].codelen;
#endif
                    }
                }
            }
            currun = 1;
            curlen = 0;
        }
    }
  }

  /* Finish off any unclosed runs */
  if (currun > 0) {
    if (curlen == 0) {
      if (currun > 10) {
          TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[34]].code, lencodes[lenmap[34]].codelen, f));
          TCMP_BUF_CHECK(!writeCode(currun - 11, 7, f));
#ifdef  INSTRUMENT
          uses->tokhuff += lencodes[lenmap[34]].codelen + 7;
#endif
      } else if (currun > 2) {
          TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[33]].code, lencodes[lenmap[33]].codelen, f));
          TCMP_BUF_CHECK(!writeCode(currun - 3, 3, f));
#ifdef  INSTRUMENT
          uses->tokhuff += lencodes[lenmap[33]].codelen + 3;
#endif
      } else {
          for (; currun > 0; currun--) {
              TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[0]].code, lencodes[lenmap[0]].codelen, f));
#ifdef  INSTRUMENT
              uses->tokhuff += lencodes[lenmap[0]].codelen;
#endif
          }
      }
    } else {
        if (currun > 2) {
            TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[32]].code, lencodes[lenmap[32]].codelen, f));
            TCMP_BUF_CHECK(!writeCode(currun - 3, 2, f));
#ifdef  INSTRUMENT
            uses->tokhuff += lencodes[lenmap[32]].codelen + 2;
#endif
        } else {
            for (; currun > 0; currun--) {
                TCMP_BUF_CHECK(!writeCode(lencodes[lenmap[curlen]].code, lencodes[lenmap[curlen]].codelen, f));
#ifdef  INSTRUMENT
                uses->tokhuff += lencodes[lenmap[curlen]].codelen;
#endif
            }
        }
    }
  }

  /* That's all - the dictionary state has the codes in it. */
  free((void *)usecodes);

  NORMAL_RETURN;

ERR_CLEANUP:

  if (usecodes != (Huff *)NULL) {
    free((void *)usecodes);
  }
  ERR_RETURN;
}
#endif /* #ifdef 0 */
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode tcmp_encodeTokPositions(void        *dstate, 
**                                             tokpos      *positions, 
**                                             Dgp_Int32   npos,
**                                             Dgp_Int32   lowx, 
**                                             Dgp_Int32   lowy, 
**                                             Dgp_Int32   highx, 
**                                             Dgp_Int32   highy,
**                                             ErrCode     (*extbuilder)(tokpos        **sorted_tokens,
**                                                                                     Dgp_Int32     ntok,
**                                                                                     void          *extcookie,
**                                                                                     void          **extension,
**                                                                                     size_t        *extlen,
**                                                                                     Dgp_State     *dgp_state),
**                                             void        *extcookie,
**                                             void        *outbuf, 
**                                             size_t      *outbuflen, 
**                                             Dgp_State   *dgp_state)
**
**
**  Write out all the boxes. First, compute the final positions
**  of each box, making sure that we're not going to ask the
**  decompressor to render off the edge of the bitmap. Then
**  write out each box's position as a delta off the previous box,
**  and the Huffman code for its class number.
** 
** 
**	Arguments:
**      o  dstate...
**      o  positions...
**      o  npos...
**      o  lowx, lowy...
**      o  highx, highy...
**      o  extbuilder...
**      o  extcookie...
**      o  outbuf...
**      o  outbuflen...
**      o  dgp_state points to a prepared Dgp_State{}.
**
**	Return value:
**		o  ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

ErrCode
tcmp_encodeTokPositions(void        *dstate, 
                        tokpos      *positions, 
                        Dgp_Int32   npos,
                        Dgp_Int32   lowx, 
                        Dgp_Int32   lowy, 
                        Dgp_Int32   highx, 
                        Dgp_Int32   highy,
                        ErrCode     (*extbuilder)(tokpos        **sorted_tokens,
                                                  Dgp_Int32     ntok,
                                                  void          *extcookie,
                                                  void          **extension,
                                                  size_t        *extlen,
                                                  Dgp_State     *dgp_state),
                        void        *extcookie,
                        void        *outbuf, 
                        size_t      *outbuflen, 
                        Dgp_State   *dgp_state) {
  DGP_PROLOGUE("tcmp_encodeTokPositions()", DGP_PERRSTATE);
  dictstate         *state = dstate;
  bitfile           f;
  Dgp_Int32         i;
  tokpos            **sortpos = (tokpos **)NULL;
  Dgp_Int32         lastx, lasty, prevfirstx;
  Dgp_Int32         curx, cury;
  tokstate          *curtok;
  Dgp_Int32         ntok = 0;
  tokstate          *toks = (tokstate *)NULL;
  Dgp_Int32         modex;
  Dgp_Int32         stripsize, logstripsize;
  Dgp_Int32         firstXEncoding, deltaXEncoding, deltaYEncoding;
  Dgp_Bool          transpose;
  void              *extension;
  size_t            extensionlen;
  void              **userdata_copy = (void **)NULL;
  token             *tok;
  Dgp_Allocators    *a = DGP_PALLOCATORS;
#ifdef  INSTRUMENT
  pos_bituses bituses;
  bituses.overhead = bituses.padding = bituses.firstx = bituses.deltax =
    bituses.deltay = bituses.y = bituses.tokhuff = bituses.toknum = 0;
#endif

#ifdef DGP_MMTOK_DBUG
  DGP_REPORTF1("tcmp_encodeTokPositions() / Writing %d boxes\n", npos);
#endif

  /* Set up the output "file". */
  f.outbuf = (Dgp_UInt8 *)outbuf;
  f.outbufleft = *outbuflen;
  f.buf = 0;
  f.bitsinbuf = 0;

  /* You can't encode something relative to nothing. */

  /*
   * Special case where you can: if state is NULL and there are no
   * token positions.
   */

  if (state == (dictstate *)NULL) {
    tokpos *dummy;
    DGP_ASSERT(npos == 0);

    /*
     * We're just going to write out some fixed values here - it's easier
     * than hacking all the code below to recognise the state==NULL case.
     */
    /* Position count. */
#ifdef  INSTRUMENT
    TCMP_BUF_CHECK(!writeValue((unsigned)npos, &f , &bituses.overhead));
#else
    TCMP_BUF_CHECK(!writeValue((unsigned)npos, &f));
#endif
    /* Encoding choices - use 0 everywhere */
    /* Modal delta X value */
#ifdef  INSTRUMENT
    bituses.overhead += 4;
#endif
    TCMP_BUF_CHECK(!writeCode(0u, 4, &f));

    /* Next, 2 bits of (strip height-1) */
#ifdef  INSTRUMENT
    bituses.overhead += 2;
#endif
    TCMP_BUF_CHECK(!writeCode(0u, 2, &f));

    /* Now 2 bits each of encoder info */
#ifdef  INSTRUMENT
    bituses.overhead += 6;
#endif
    TCMP_BUF_CHECK((!writeCode(0u, 2, &f)) ||
       (!writeCode(0u, 2, &f)) ||
       (!writeCode(0u, 2, &f)));

    /* Next, 1 bit of "transposed?" */
#ifdef  INSTRUMENT
    bituses.overhead += 1;
#endif
    TCMP_BUF_CHECK(!writeCode(0u, 1, &f));
  
    /*
     * Of course, it can't all be that simple - we need to call the
     * extension builder with the "sorted" tokpos array...
     */
    if (extbuilder != (ErrCode (*)(tokpos **, Dgp_Int32, void *, void **, size_t *, Dgp_State *))NULL) {
      /* Fake a sorted tokpos array so we're not passing NULL around. */
      sortpos = &dummy;
      DGP_ERR_PASS_CLEANUP((*extbuilder)(sortpos, 
                                         npos, 
                                         extcookie, 
                                         &extension, 
                                         &extensionlen, 
                                         dgp_state));
      sortpos = (tokpos **)NULL;
    }
    else {
      extension = (void *)NULL;
      extensionlen = 0;
    }

    /* Write out some BIG y delta */
#ifdef  INSTRUMENT
    TCMP_BUF_CHECK(!writeYDelta((unsigned)(highy + 1), &f, 0,&bituses.deltay));
#else
    TCMP_BUF_CHECK(!writeYDelta((unsigned)(highy + 1), &f, 0));
#endif
#ifdef  INSTRUMENT
    bituses.padding += (BITS_ULONG - f.bitsinbuf) % BITS_UCHAR;
#endif
    TCMP_BUF_CHECK(!flushBitfile(&f));
    
    /* Copy in the extension block if any */
    if (extension != (void *)NULL) {
      TCMP_BUF_CHECK(f.outbufleft < extensionlen);
      memcpy(f.outbuf, extension, (unsigned)extensionlen);
      f.outbuf += extensionlen;
      f.outbufleft -= extensionlen;
    }
    else {
      DGP_ASSERT(extensionlen == 0);
    }

    /* How much space did we use? */
    *outbuflen -= f.outbufleft;
    NORMAL_RETURN;
  }

  ntok = state->ntok;
  toks = state->toks;
  DGP_ASSERT((ntok > 0) || (npos == 0));

  /* Make sure the Huffman codes are valid */
  DGP_ERR_PASS_CLEANUP(tcmpi_ensureHuffmanCodes(state, dgp_state));

    
  /* Since we've removed the toknum field, we need to hijack
   * the userdata field for a moment. Copy the data for safekeeping
   * and in the meantime, keep the Huffman code and other relevant
   * info there (in the form of tokstate). Note that we allocate
   * an array the size of the number of tokens _in_the_dictionary_
   * since this is how many pointers we will be bashing. */
  DGP_CHECK_ALLOC_CLEANUP(userdata_copy = (void **)(a->malloc)(ntok * sizeof(*userdata_copy)));
     
  /* Copy the data over. */
  for(i = 0; i < ntok; i++) {
    tok = toks[ i].tok;
    userdata_copy[i] = tok->extdata;
    tok->extdata = (void *)&toks[ i];
  }
        
  /* Now start writing them out. */

  /* First thing is the number of positions */
#ifdef  INSTRUMENT
  TCMP_BUF_CHECK(!writeValue((unsigned)npos, &f , &bituses.overhead));
#else
  TCMP_BUF_CHECK(!writeValue((unsigned)npos, &f));
#endif

  /* Now we decide what encoding strategy we'll use */
  DGP_ERR_PASS_CLEANUP(decidePosEncoding(positions, 
                                         npos,
                                         lowx,   lowy,
                                         highx,  highy,
                                         &transpose,
                                         &modex, 
                                         &stripsize, 
                                         &logstripsize,
                                         &firstXEncoding, 
                                         &deltaXEncoding,
                                         &deltaYEncoding, 
                                         &sortpos, 
                                         dgp_state));

  /*
   * That sorted the positions into strips for us (how nice of it!).
   * The strips divisions lie on multiples of stripsize over lowy
   * (i.e., strips can start at lowy, lowy+stripsize, lowy+2*stripsize
   * etc).  We encode y-lowy for each token.
   *
   * If *transpose was set, then the best encoding was found when the
   * top left corner was used as the reference position, and the X and Y
   * coordinates were swapped (i.e. strips are vertical, not horizontal etc).
   */

  /* Write out the status info indicating what encoders we used. */
  /* First, 4 bits of modal x value */
#ifdef  INSTRUMENT
  bituses.overhead += 4;
#endif
  TCMP_BUF_CHECK(!writeCode((unsigned)modex, 4, &f));

  /* Next, 2 bits of (strip height-1) */
#ifdef  INSTRUMENT
  bituses.overhead += 2;
#endif
  TCMP_BUF_CHECK(!writeCode((unsigned)(stripsize - 1), 2, &f));

  /* Now 2 bits each of encoder info */
#ifdef  INSTRUMENT
  bituses.overhead += 6;
#endif
  TCMP_BUF_CHECK((!writeCode((unsigned)firstXEncoding, 2, &f)) ||
                 (!writeCode((unsigned)deltaXEncoding, 2, &f)) ||
                 (!writeCode((unsigned)deltaYEncoding, 2, &f)));

  /* Next, 1 bit of "transposed?" */
#ifdef  INSTRUMENT
  bituses.overhead += 1;
#endif
  TCMP_BUF_CHECK(!writeCode(transpose ? 1u : 0u, 1, &f));

#if 0
  DGP_ERR_PASS_CLEANUP(writeTokenHuffTable(state, /* tokbynum, mintoknum, */
                                                  positions, npos, &f
#ifdef  INSTRUMENT
                                                                     , &bituses
#endif
                                                                               ));
#endif

  /* Before we get too far in, build the extension */
  if (extbuilder != (ErrCode (*)(tokpos **, Dgp_Int32, void *, void **, size_t *, Dgp_State *))NULL) {
    DGP_ERR_PASS_CLEANUP((*extbuilder)(sortpos, 
                                       npos, 
                                       extcookie,
                                       &extension,
                                       &extensionlen,
                                       dgp_state));
  } else {
    extension = (void *)NULL;
    extensionlen = 0;
  }

  lasty = -1;
  lastx = prevfirstx = 0;

  for (i = 0; i < npos; i++) {
    curx = sortpos[i]->xpos - lowx;
    cury = sortpos[i]->ypos - lowy;
    /* We need the tokstate for each tokpos, but surprise! It's
     * hidden in the extdata!! */
    curtok = (tokstate *)sortpos[i]->tok->extdata;

    if (transpose) {
      Dgp_Int32 tmp = curx;
      curx = cury - (curtok->tok->height - 1);
      cury = tmp;
    }

    if ((lastx < 0) || ((lasty >> logstripsize) != (cury >> logstripsize))) {
      DGP_ASSERT(lasty < cury);
      /* A new strip. */

      /* Terminate the previous strip */
      if (lasty != -1) {
#ifdef  INSTRUMENT
  TCMP_BUF_CHECK(!writeEndOfRow(&f, deltaXEncoding , &bituses.deltax));
#else
  TCMP_BUF_CHECK(!writeEndOfRow(&f, deltaXEncoding));
#endif
      }

      /* How far to the next strip? */
#ifdef  INSTRUMENT
      TCMP_BUF_CHECK(!writeYDelta((unsigned)(cury >> logstripsize) - (lasty >> logstripsize),
                                             &f, 
                                             deltaYEncoding, 
                                             &bituses.deltay));
#else
      TCMP_BUF_CHECK(!writeYDelta((unsigned)(cury >> logstripsize) -  (lasty >> logstripsize), 
                                             &f,
                                             deltaYEncoding));
#endif

      /* Write out the first x, y pos, and token number */
#ifdef  INSTRUMENT
      TCMP_BUF_CHECK(!writeFirstX(curx - prevfirstx, &f, firstXEncoding, &bituses.firstx));
#else
      TCMP_BUF_CHECK(!writeFirstX(curx - prevfirstx, &f, firstXEncoding));
#endif
      prevfirstx = curx;

      if (logstripsize) {
#ifdef  INSTRUMENT
  TCMP_BUF_CHECK(!(writeYVal(cury, &f, logstripsize , &bituses.y)));
#else
  TCMP_BUF_CHECK(!(writeYVal(cury, &f, logstripsize)));
#endif
      }

#ifdef  INSTRUMENT
      bituses.toknum += curtok->codelen;
#endif
      TCMP_BUF_CHECK(!writeCode(curtok->code, curtok->codelen, &f));
    } else {
      /* Write out the y position */
      /* Write out the x delta and token number */
#ifdef  INSTRUMENT
      TCMP_BUF_CHECK(!writeXDelta(curx - lastx - modex, &f, deltaXEncoding, &bituses.deltax));
#else
      TCMP_BUF_CHECK(!writeXDelta(curx - lastx - modex, &f, deltaXEncoding));
#endif

      if (logstripsize) {
#ifdef  INSTRUMENT
  TCMP_BUF_CHECK(!(writeYVal(cury, &f, logstripsize , &bituses.y)));
#else
  TCMP_BUF_CHECK(!(writeYVal(cury, &f, logstripsize)));
#endif
      }

#ifdef  INSTRUMENT
      bituses.toknum += curtok->codelen;
#endif
      TCMP_BUF_CHECK(!writeCode(curtok->code, curtok->codelen, &f));
    }

    lastx = curx + (transpose ? curtok->tok->height : curtok->tok->width);
    lasty = (cury >> logstripsize) << logstripsize;
  }

  /* Terminate the last strip. */
  if (lasty != -1) {        /* OK, so I'm paranoid. */
#ifdef  INSTRUMENT
    TCMP_BUF_CHECK(!writeEndOfRow(&f, deltaXEncoding , &bituses.deltax));
#else
    TCMP_BUF_CHECK(!writeEndOfRow(&f, deltaXEncoding));
#endif
  }
  /* Write out some BIG y delta */
#ifdef  INSTRUMENT
  TCMP_BUF_CHECK(!writeYDelta((unsigned)(transpose ? highx : highy) -lasty + 1, &f, deltaYEncoding , &bituses.deltay));
#else
  TCMP_BUF_CHECK(!writeYDelta((unsigned)(transpose ? highx : highy) -lasty + 1, &f, deltaYEncoding));
#endif

  /*
   * Write out the extensions section.
   */

  /* (We don't write extensionlen anymore.) */
#if 0
#ifdef  INSTRUMENT
  TCMP_BUF_CHECK(!writeValue((unsigned)extensionlen, &f , &bituses.overhead));
#else
  TCMP_BUF_CHECK(!writeValue((unsigned)extensionlen, &f));
#endif
#endif

  /* Done with this bit-packed part of the file. */
#ifdef  INSTRUMENT
  bituses.padding += (BITS_ULONG - f.bitsinbuf) % BITS_UCHAR;
#endif
  TCMP_BUF_CHECK(!flushBitfile(&f));
    
  /* Copy in the extension block, if any. */
  if (extension != (void *)NULL) {
    TCMP_BUF_CHECK(f.outbufleft < extensionlen);
    memcpy(f.outbuf, extension, (unsigned)extensionlen);
    f.outbuf += extensionlen;
    f.outbufleft -= extensionlen;
  } else {
    DGP_ASSERT(extensionlen == 0);
  }

  (a->free)((void *)sortpos);
    
  /* Copy the data back. */
  for(i = 0; i < ntok; i++) {
    toks[i].tok->extdata = userdata_copy[i];
  }
  (a->free)( userdata_copy);

#ifdef  INSTRUMENT
  fprintf(stderr, "Bit allocations for positions (%d positions):\n\t\t\t\t%d overhead\n\t\t\t\t%d padding\n\t\t\t\t%d first X\n\t\t\t\t%d delta X\n\t\t\t\t%d delta Y\n\t\t\t\t%d Y\n\t\t\t\t%d token Huffman tables\n\t\t\t\t%d token numbers\n",
    npos, bituses.overhead, bituses.padding, bituses.firstx, bituses.deltax, bituses.deltay, bituses.y, bituses.tokhuff, bituses.toknum);
#endif

  /* How much space did we use? */
  *outbuflen -= f.outbufleft;
  NORMAL_RETURN;

ERR_CLEANUP:

  if (sortpos != (tokpos **)NULL) {
    (a->free)((void *)sortpos);
  }
  
  if( userdata_copy != (void *)NULL) {
    /* Copy the data back. */
    for(i = 0; i < ntok; i++) {
      toks[i].tok->extdata = userdata_copy[i];
    }
    (a->free)( userdata_copy);
  }

  /* Should guess how much more space they should feed us. FIX */
#ifdef DGP_DBUG
  DGP_WARNF0("tcmp_encodeTokPositions() bailing for more memory\n");
#endif
  *outbuflen += 50000;

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  size_t tcmp_estimateEncodedTokPosSize(void *dstate, tokpos *positions, Dgp_Int32 npos)
**
**
**  This routine takes a dictionary state (which must have been properly filled
**  in by encodeTokDict; one which has just been initialized by
**  initializeTokDictState isn't enough), and a bunch of token positions. It
**  figures out an estimate of how much space is required to encode those
**  positions and returns that.
**
** 
**	Arguments:
**      o  dstate...
**      o  positions...
**      o  npos...
**
**	Return value:
**		o  An integral number of bytes.
**
**	Modification history:
**      o  None.
**  
**/

size_t
tcmp_estimateEncodedTokPosSize(void *dstate, tokpos *positions, Dgp_Int32 npos) {
  /* Stop compiler complaints. */
  dstate = dstate;
  positions = positions;

#ifdef DGP_MMTOK_DBUG
  DGP_REPORTF1("tcmp_estimateEncodedTokPosSize() / estimating %d bytes for pos\n", POSOVERHEAD + POSSIZE * npos);
#endif
  return(POSOVERHEAD + POSSIZE * npos);
}
/*************************************************************************/

