/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Author:  ?;  199n-9n
**
**  Filename:  mmcompress.h.
**
**  Types & prototypes for accessing bitfiles & compressing token bitmaps (?).
**  Prototypes' source is in mmcompress.c, mmtokdict.c.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  All #includes pushed up to source files.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtokcomp/mmcompress.h,v 1.23 2003/02/05 22:23:56 mkrocken Exp $
 *
 */

#ifndef  MMCOMPRESS_H
#define  MMCOMPRESS_H


#include "util.h"
#include <sys/types.h>
#include <stddef.h>
#include "dgp_err_control.h"
#include "dgp_state.h"

#include "token_t.h"
#include "tokpos_t.h"

#ifdef __cplusplus
extern "C"
{
#endif


DGP_LIBEXPORT size_t 
tcmp_estimateTokDictStateSize(void *oldstate, token *tokens, Dgp_Int32 ntok);

/* Estimate for a set of "ntok" uncommitted tokens. */
DGP_LIBEXPORT size_t 
tcmp_estimateTokDictStateSizeEx(void *oldstate, Dgp_Int32 ntok);

DGP_LIBEXPORT void 
tcmp_initializeTokDictState(void *oldstate, void *newstatemem, size_t memsize);

DGP_LIBEXPORT size_t 
tcmp_estimateEncodedTokDictSize(void *dstate, token *tokens, Dgp_Int32 ntok);

DGP_LIBEXPORT ErrCode 
tcmp_encodeTokDict(void         *dstate, 
                   token        *tokens, 
                   Dgp_Int32    ntok,         
                   ErrCode      (*extbuilder)(token     **sorted_toks,
                                              Dgp_Int32 ntok,
                                              void      *extcookie,
                                              void      **extension,
                                              size_t    *extlen,
                                              Dgp_State *dgp_state),
                   void         *extcookie,
                   void         *outbuf, 
                   size_t       *outbuflen, 
                   Dgp_State    *dgp_state);

DGP_LIBEXPORT size_t 
tcmp_estimateEncodedTokPosSize(void *dstate, tokpos *positions, Dgp_Int32 npos);

DGP_LIBEXPORT ErrCode 
tcmp_encodeTokPositions(void        *dstate, 
                        tokpos      *positions,
                        Dgp_Int32   npos,
                        Dgp_Int32   lowx, 
                        Dgp_Int32   lowy,
                        Dgp_Int32   highx, 
                        Dgp_Int32   highy,
                        ErrCode     (*extbuilder)(tokpos    **sorted_tokpos,
                                                  Dgp_Int32 ntok,
                                                  void      *extcookie,
                                                  void      **extension,
                                                  size_t    *extlen,
                                                  Dgp_State *dgp_state),
                        void        *extcookie,
                        void        *outbuf, 
                        size_t      *outbuflen, 
                        Dgp_State   *dgp_state);

DGP_LIBEXPORT ErrCode 
tcmp_getTokenCount(void         *buffer, 
                   size_t       buflen, 
                   Dgp_State    *dgp_state,
                   size_t       *count);

DGP_LIBEXPORT ErrCode
tcmp_decodeTokDict(token        *tokens, 
                   void         *buffer,
                   size_t       buflen, 
                   Dgp_State    *dgp_state);

DGP_LIBEXPORT ErrCode 
tcmp_decodeTokDictMUsage(token      *tokens, 
                         void       *buffer,
                         size_t     buflen, 
                         Dgp_State  *dgp_state,
                         Dgp_Int32  *tok_usage);    


/* Decodes the token structures only. Note that the "bits" and "extdata"
 * fields are used for private data; you shouldn't modify them until after
 * calling tcmp_decodeTokDictBitmaps. */
DGP_LIBEXPORT ErrCode
tcmp_decodeTokDictTokens(token      *tokens, 
                         void       *buffer,
                         size_t     buflen, 
                         Dgp_State  *dgp_state);

/* Estimates the total amount of memory needed to decode the bitmaps
 * (excluding buffers for the original compressed data). */
DGP_LIBEXPORT Dgp_UInt32 
tcmp_estimateTokDictBitmapUsage(token *tokens, Dgp_Int32 ntokens);

/* Sets the "extdata" field of each token to the initial allocation size
 * of the bitmap for that token. */
DGP_LIBEXPORT ErrCode
tcmp_decodeTokDictBitmaps(token     *tokens, 
                          Dgp_Int32 ntokens,
                          void      *buffer, 
                          size_t    buflen,
                          Dgp_State *dgp_state);


DGP_LIBEXPORT ErrCode 
tcmp_convertTokensToState(token     *tokens, 
                          Dgp_Int32 ntok,
                          void      *dstate, 
                          Dgp_State *dgp_state);


/*  From mmokdict.c... / dhtesq. */
DGP_LIBEXPORT size_t 
tcmp_estimateMergedTokDictStateSize(void **states, Dgp_Int32 nstates);

DGP_LIBEXPORT ErrCode
tcmp_mergeTokDictStates(void        *newstate, 
                        size_t      newsize,
                        void        **states, 
                        Dgp_Int32   nstates, 
                        Dgp_State   *dgp_state);

DGP_LIBEXPORT ErrCode 
tcmp_getTokPosCount(void        *buffer, 
                    size_t      buflen,
                    Dgp_State   *dgp_state, 
                    size_t      *count);

DGP_LIBEXPORT ErrCode
tcmp_decodeTokPositions(tokpos      *positions,
                        Dgp_Int32   lowx, 
                        Dgp_Int32   lowy,
                        Dgp_Int32   highx, 
                        Dgp_Int32   highy,
                        void        *buffer,
                        size_t      buflen,
                        void        *dstate, 
                        Dgp_State   *dgp_state);

#ifdef __cplusplus
}
#endif

#endif
