/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Author:  ?;  199n-9n
**
**  Filename:  mmcompressprivate.h.
**
**  Types & prototypes for accessing bitfiles & compressing token bitmaps (?);
**  private storage.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  All #includes pushed up to source files.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtokcomp/mmcompressprivate.h,v 1.21 2001/05/08 00:42:55 dtracy Exp $
 *
 */

#ifndef  MMCOMPRESS_PRIVATE_H

#define  MMCOMPRESS_PRIVATE_H

/*
Moved up to source files. / dhtesq. 
UNDO (MK) */

#include "huff.h"
#include <sys/types.h>
#include "mmcompress.h"
#include "dgp_allocators.h"


#ifdef __cplusplus
extern "C"
{
#endif

/* define this to get debug output from mmtokdict */
/*#define DGP_MMTOK_DBUG 1*/


/* The internal information stored for each token. */
typedef struct {
    token *tok;         /* The token itself. */
    Dgp_UInt32 code;    /* The current Huffman code... */
    Dgp_UInt8 codelen;  /* ...and code length. */
    } tokstate;

/*
 * The internal dictionary state type.
 * It has:
 *  ntok - a number of tokens (this is not necessarily equal to the
 *         maximum capacity of the block of memory; it is the number
 *         of tokens *currently* contained in the block)
 *  capacity - the actual capacity of the block of memory
 *  toks - a pointer (to a part of memory in the same block)
 *         to the token states
 *  codesValid - whether the Huffman codes have actually been computed
 *         yet.  They're computed lazily because many dictionaries
 *         (all the per-page ones) are never used by themselves,
 *         but only after being merged.
 *  codeindexblock - an array of ints used in the decoding of Huffman
 *              codes. This code index block is subdivided into
 *       groups (the codeindex array), but the total size
 *       of the groups is ntok ints.
 *  firstcodes, codeindex: arrays used in the decoding
 *
 * The tables are built so that if a code is read with length l,
 * and its value is firstcodes[l] + v, then its
 * index in the dictionary state block (*NOT* its toknum value) is
 * codeindex[l][v].
 *
 * The layout is:
 *  dictstate
 *  capacity * tokstate
 *  capacity * Dgp_Int32
 */
typedef struct {
    Dgp_Int32 ntok;
    Dgp_Int32 capacity;
    tokstate *toks;
    Dgp_Bool codesValid;
    Dgp_Int32 *codeindexblock;
    Dgp_UInt32 firstcodes[DGP_MAX_CODELEN + 1];
    Dgp_Int32 *codeindex[DGP_MAX_CODELEN + 1];
    } dictstate;

/* A couple of useful things. */
#define  BITS_UCHAR  8
#define  BITS_ULONG  ((Dgp_Int32)(BITS_UCHAR * sizeof(Dgp_UInt32)))

/* Whatever globals we need. */
extern Dgp_Allocators tokCompGlobals;

/*
 * Stuff to do with the encoders (in mmencoder.c).
 */

/*
 * The encoders themselves.  Each one is an array whose entries say "encode
 * all the values up to (and including) upto using nbits, using a prefixbits
 * length prefix, and then nbits-prefixbits bits drawn from the value minus
 * the upto of the previous entry".  A nbits value of -1 is special; it means
 * "encode the value using the special 'value' encoder".
 */
typedef struct {
    Dgp_Int32 from;
    Dgp_Int32 upto;
    Dgp_Int32 nbits;
    Dgp_Int32 prefixbits;
    Dgp_UInt32 prefix;
    } codeinfo;

/*
 * An encoder.
 */
typedef struct {
    Dgp_Int32 EOSbits;
    unsigned EOScode;
    codeinfo *codes;
    Dgp_Int32 ncodes;
    } encoder;

/*
 * A decoder.  It's a state machine: from state x, read nbits bits into
 * "val" and transition to state outstate+val.
 * If outstate == 0, then this is a terminal state. Read nbits bits (or a
 * "value" if nbits < 0), and return val+offset (or offset-val if nbits < 0
 * and offset < 0).  nbits == 0 is valid; it means just return offset.
 * If this is a terminal state, nbits < 0, and offset == 0, then this is a
 * special code (some kind of end code).
 */
typedef struct {
    Dgp_Int8 nbits;
    Dgp_UInt8 outstate;
    Dgp_Int16 offset;
    } decodeState;

typedef decodeState *decoder;

/* Forward routine declarations */
ErrCode tcmpi_ensureHuffmanCodes(dictstate *state, Dgp_State *dgp_state);

ErrCode internal_convertTokensToState(Dgp_Int32         ntok,
                                      void              *dstate, 
                                      Dgp_Int32         newntok,
                                      token             **ordered);

/* Declarations for the encoders. */
#define  NDELTAX  3
extern encoder mmDeltaXCoders[NDELTAX];
extern decoder mmDeltaXDecoders[NDELTAX];
#define  NFIRSTX  2
extern encoder mmFirstXCoders[NFIRSTX];
extern decoder mmFirstXDecoders[NFIRSTX];
#define  NDELTAY  3
extern encoder mmDeltaYCoders[NDELTAY];
extern decoder mmDeltaYDecoders[NDELTAY];
#define  NWH  1
extern encoder mmWHCoders[NWH];
extern decoder mmWHDecoders[NWH];
#define  NUSE  2
extern encoder mmUseCoders[NUSE];
extern decoder mmUseDecoders[NUSE];

#ifdef __cplusplus
}
#endif

#endif
