/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Author:  ?;  199n-9n
**
**  Filename:  mmencoder.c
**
**
**  This file contains a bunch of encoders.
**
**	Modification history:
**      o  None.
**
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtokcomp/mmencoder.c,v 1.6 2000/03/04 00:22:43 mkrocken Exp $
 *
 */

#include <limits.h>
#include "dgp_types.h"
#include "huff.h"
#include "token_t.h"
#include "mmcompressprivate.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtokcomp/mmencoder.c,v 1.6 2000/03/04 00:22:43 mkrocken Exp $");


/*
 * The deltax coders.
 */

/*
 * A 300dpi encoder. Encoding is:
 *  value      bits  fields    prefix
 *  EOS      2  2    00
 *  -infinity..-15    19  11;v    11111001110
 *  -14..-8      10  7;3    1111100
 *  -7..-6      10  9;1    111111110
 *  -5..-4      9  8;1    11111110
 *  -3      9  9    111111111
 *  -2      7  7    1111101
 *  -1      4  4    1010
 *  0      3  3    010
 *  1      3  3    011
 *  2      5  5    11010
 *  3      6  6    111010
 *  4-19      7  3;4    100
 *  20-21      7  6;1    111011
 *  22-37      8  4;4    1011
 *  38-69      9  4;5    1100
 *  70-133      11  5;6    11011
 *  134-261      12  5;7    11100
 *  262-389      13  6;7    111100
 *  390-645      15  7;8    1111110
 *  646-1669    16  6;10    111101
 *  1670-      19  11;v    11111001111
 *
 * Note that the two end codes actually steal a bit of the encoding space
 * from another code group (-14..-8 in this case: that has a 7-bit prefix
 * of 1111100, but doesn't use value 111, so codes starting 1111100111 are
 * available).  This is done in many of the encoders.
 */

static codeinfo deltax300[] = {
    { INT_MIN, -15, -1, 11, 0x7ce },
    { -14, -8, 10, 7, 0x7c },
    { -7, -6, 10, 9, 0x1fe },
    { -5, -4, 9, 8, 0xfe },
    { -3, -3, 9, 9, 0x1ff },
    { -2, -2, 7, 7, 0x7d },
    { -1, -1, 4, 4, 0xa },
    { 0, 1, 3, 2, 0x1 },
    { 2, 2, 5, 5, 0x1a },
    { 3, 3, 6, 6, 0x3a },
    { 4, 19, 7, 3, 0x4 },
    { 20, 21, 7, 6, 0x3b },
    { 22, 37, 8, 4, 0xb },
    { 38, 69, 9, 4, 0xc },
    { 70, 133, 11, 5, 0x1b },
    { 134, 261, 12, 5, 0x1c },
    { 262, 389, 13, 6, 0x3c },
    { 390, 645, 15, 7, 0x7e },
    { 646, 1669, 16, 6, 0x3d },
    { 1670, INT_MAX, -1, 11, 0x7cf }
    };

#define  NDX300  (sizeof(deltax300)/sizeof(deltax300[0]))

/*
 * The 300dpi coder, scaled by 2 and with an extra bit tacked on to the
 * end of all the codes.
 *  value      bits  fields    prefix
 *  EOS      2  2    00
 *  -infinity..-30    19  11;v    11111001110
 *  -29..-16    11  7;4    1111100
 *  -15..-12    11  9;2    111111110
 *  -11..-8      10  8;2    11111110
 *  -7..-6      10  9;1    111111111
 *  -5..-4      8  7;1    1111101
 *  -3..-2      5  4;1    1010
 *  -1..0      4  3;1    010
 *  1..2      4  3;1    011
 *  3..4      6  5;1    11010
 *  5..6      7  6;1    111010
 *  7..38      8  3;5    100
 *  39..42      8  6;2    111011
 *  43..74      9  4;5    1011
 *  75..138      10  4;6    1100
 *  139..266    12  5;7    11011
 *  267..522    13  5;8    11100
 *  523..778    14  6;8    111100
 *  779..1290    16  7;9    1111110
 *  1291..3338    17  6;11    111101
 *  3339-      19  11;v    11111001111
 */

static codeinfo deltax3002[] = {
    { INT_MIN, -30, -1, 11, 0x7ce },
    { -29, -16, 11, 7, 0x7c },
    { -15, -12, 11, 9, 0x1fe },
    { -11, -8, 10, 8, 0xfe },
    { -7, -6, 10, 9, 0x1ff },
    { -5, -4, 8, 7, 0x7d },
    { -3, -2, 5, 4, 0xa },
    { -1, 0, 4, 3, 0x2 },
    { 1, 2, 4, 3, 0x3 },
    { 3, 4, 6, 5, 0x1a },
    { 5, 6, 7, 6, 0x3a },
    { 7, 38, 8, 3, 0x4 },
    { 39, 42, 8, 6, 0x3b },
    { 43, 74, 9, 4, 0xb },
    { 75, 138, 10, 4, 0xc },
    { 139, 266, 12, 5, 0x1b },
    { 267, 522, 13, 5, 0x1c },
    { 523, 778, 14, 6, 0x3c },
    { 779, 1290, 16, 7, 0x7e },
    { 1291, 3338, 17, 6, 0x3d },
    { 3339, INT_MAX, -1, 11, 0x7cf }
    };

#define  NDX3002  (sizeof(deltax3002)/sizeof(deltax3002[0]))

/*
 * An encoder tuned a bit more for 600dpi.
 *  value      bits  fields    prefix
 *  EOS      2  2    00
 *  -infinity..-20    18  10;v    1101101110
 *  -19..-6      10  6;4    110110
 *  -5      8  8    11111110
 *  -4      7  7    1111100
 *  -3      5  5    11000
 *  -2..1      4  2;2    01
 *  2      5  5    11001
 *  3      6  6    110111
 *  4      7  7    1111101
 *  5      8  8    11111111
 *  6..69      8  2;6    10
 *  70..101      10  5;5    11010
 *  102..133    11  6;5    111000
 *  134..197    12  6;6    111001
 *  198..325    13  6;7    111010
 *  326..581    14  6;8    111011
 *  582..1093    15  6;9    111100
 *  1094..2117    16  6;10    111101
 *  2118..4165    18  7;11    1111110
 *  4166..infinity    18  10;v    1101101111
 */

static codeinfo deltax600[] = {
    { INT_MIN, -20, -1, 10, 0x36e },
    { -19, -6, 10, 6, 0x36 },
    { -5, -5, 8, 8, 0xfe },
    { -4, -4, 7, 7, 0x7c },
    { -3, -3, 5, 5, 0x18 },
    { -2, 1, 4, 2, 0x1 },
    { 2, 2, 5, 5, 0x19 },
    { 3, 3, 6, 6, 0x37 },
    { 4, 4, 7, 7, 0x7d },
    { 5, 5, 8, 8, 0xff },
    { 6, 69, 8, 2, 0x2 },
    { 70, 101, 10, 5, 0x1a },
    { 102, 133, 11, 6, 0x38 },
    { 134, 197, 12, 6, 0x39 },
    { 198, 325, 13, 6, 0x3a },
    { 326, 581, 14, 6, 0x3b },
    { 582, 1093, 15, 6, 0x3c },
    { 1094, 2117, 16, 6, 0x3d },
    { 2118, 4165, 18, 7, 0x7e },
    { 4166, INT_MAX, -1, 10, 0x36f }
    };

#define  NDX600  (sizeof(deltax600)/sizeof(deltax600[0]))

encoder mmDeltaXCoders[] = {
    { 2, 0, deltax300, NDX300 },
    { 2, 0, deltax3002, NDX3002 },
    { 2, 0, deltax600, NDX600 }
    };

/*
 * The firstx coders.
 */
/*
 * A 300dpi coder.
 *  value      bits  fields    prefix
 *  -infinity..-2048  18  10;v    1110111110
 *  -2047..-1024    14  4;10    1010
 *  -1023..-512    13  4;9    1011
 *  -511..-256    12  4;8    1100
 *  -255..-128    11  4;7    1101
 *  -127..-64    11  5;6    11110
 *  -63..-32    10  5;5    11111
 *  -31..-1      9  4;5    1110
 *  0..127      9  2;7    00
 *  128..255    10  3;7    010
 *  256..511    11  3;8    011
 *  512..1023    13  4;9    1000
 *  1024..2047    14  4;10    1001
 *  2048..infinity    18  10;v    1110111111
 */
static codeinfo firstx300[] = {
    { INT_MIN, -2048, -1, 10, 0x3be },
    { -2047, -1024, 14, 4, 0xa },
    { -1023, -512, 13, 4, 0xb },
    { -511, -256, 12, 4, 0xc },
    { -255, -128, 11, 4, 0xd },
    { -127, -64, 11, 5, 0x1e },
    { -63, -32, 10, 5, 0x1f },
    { -31, -1, 9, 4, 0xe },
    { 0, 127, 9, 2, 0x0 },
    { 128, 255, 10, 3, 0x2 },
    { 256, 511, 11, 3, 0x3 },
    { 512, 1023, 13, 4, 0x8 },
    { 1024, 2047, 14, 4, 0x9 },
    { 2048, INT_MAX, -1, 10, 0x3bf }
    };

#define  NFIRSTX300  (sizeof(firstx300)/sizeof(firstx300[0]))

/*
 * A 600dpi coder.
 *  value      bits  fields    prefix
 *  -infinity..-1024  18  10;v    1011111110
 *  -1023..-512    12  3;9    000
 *  -511..-256    11  3;8    001
 *  -255..-128    11  4;7    1010
 *  -127..-64    11  5;6    11100
 *  -63..-32    10  5;5    11101
 *  -31..-1      9  4;5    1011
 *  0..31      9  4;5    1100
 *  32..63      10  5;5    11110
 *  64..127      11  5;6    11111
 *  128..255    11  4;7    1101
 *  256..511    11  3;8    010
 *  512..1023    12  3;9    011
 *  1024..2047    13  3;10    100
 *  2048..infinity    18  10;v    1011111111
 */
static codeinfo firstx600[] = {
    { INT_MIN, -1024, -1, 10, 0x2fe },
    { -1023, -512, 12, 3, 0x0 },
    { -511, -256, 11, 3, 0x1 },
    { -255, -128, 11, 4, 0xa },
    { -127, -64, 11, 5, 0x1c },
    { -63, -32, 10, 5, 0x1d },
    { -31, -1, 9, 4, 0xb },
    { 0, 31, 9, 4, 0xc },
    { 32, 63, 10, 5, 0x1e },
    { 64, 127, 11, 5, 0x1f },
    { 128, 255, 11, 4, 0xd },
    { 256, 511, 11, 3, 0x2 },
    { 512, 1023, 12, 3, 0x3 },
    { 1024, 2047, 13, 3, 0x4 },
    { 2048, INT_MAX, -1, 10, 0x2ff }
    };

#define  NFIRSTX600  (sizeof(firstx600)/sizeof(firstx600[0]))

encoder mmFirstXCoders[] = {
    { 0, 0, firstx300, NFIRSTX300 },
    { 0, 0, firstx600, NFIRSTX600 }
    };

/*
 * The y strip encoders.
 */

/*
 * Version 1:
 *  value      bits  fields    prefix
 *  1      1  1    0
 *  2..3      3  2;s-2(1)  10
 *  4      4  4    1100
 *  5-6      5  4;s-5(1)  1101
 *  7-8      6  5;s-7(1)  11100
 *  9-12      7  5;s-9(2)  11101
 *  13-16      8  6;s-13(2)  111100
 *  17-20      9  7;s-17(3)  1111010
 *  21-28      10  7;s-21(3)  1111011
 *  29-44      11  7;s-29(4)  1111100
 *  45-76      12  7;s-45(5)  1111101
 *  77-140      13  7;s-77(6)  1111110
 *  141-      15  7;s-141(v)  1111111
 */
static codeinfo deltay1[] = {
    { 1, 1, 1, 1, 0x0 },
    { 2, 3, 3, 2, 0x2 },
    { 4, 4, 4, 4, 0xc },
    { 5, 6, 5, 4, 0xd },
    { 7, 8, 6, 5, 0x1c },
    { 9, 12, 7, 5, 0x1d },
    { 13, 16, 8, 6, 0x3c },
    { 17, 20, 9, 7, 0x7a },
    { 21, 28, 10, 7, 0x7b },
    { 29, 44, 11, 7, 0x7c },
    { 45, 76, 12, 7, 0x7d },
    { 77, 140, 13, 7, 0x7e },
    { 141, INT_MAX, -1, 7, 0x7f }
    };
#define  NDELTAY1  (sizeof(deltay1)/sizeof(deltay1[0]))

/*
 * Version 2:
 *  1      1  1    0
 *  2      2  2    10
 *  3..4      4  3;s-3(1)  110
 *  5      5  5    11100
 *  6-7      6  5;s-6(1)  11101
 *  8-9      7  6;s-8(1)  111100
 *  10      7  7    1111010
 *  11-12      8  7;s-11(1)  1111011
 *  13-16      9  7;s-13(2)  1111100
 *  17-24      10  7;s-17(3)  1111101
 *  25-40      11  7;s-25(4)  1111110
 *  41-72      13  8;s-41(5)  11111110
 *  73-      16  8;s-73(v)  11111111
 */ 
static codeinfo deltay2[] = {
    { 1, 1, 1, 1, 0x0 },
    { 2, 2, 2, 2, 0x2 },
    { 3, 4, 4, 3, 0x6 },
    { 5, 5, 5, 5, 0x1c },
    { 6, 7, 6, 5, 0x1d },
    { 8, 9, 7, 6, 0x3c },
    { 10, 10, 7, 7, 0x7a },
    { 11, 12, 8, 7, 0x7b },
    { 13, 16, 9, 7, 0x7c },
    { 17, 24, 10, 7, 0x7d },
    { 25, 40, 11, 7, 0x7e },
    { 41, 72, 13, 8, 0xfe },
    { 73, INT_MAX, -1, 8, 0xff }
    };
#define  NDELTAY2  (sizeof(deltay2)/sizeof(deltay2[0]))

/*
 * Version 3:
 *  value      bits  fields    prefix
 *  1      1  1    0
 *  2      3  3    100
 *  3      4  4    1100
 *  4      5  5    11100
 *  5..6      5  4;1    1101
 *  7..13      6  3;3    101
 *  14..15      7  6;1    111010
 *  16..19      8  6;2    111011
 *  20..27      9  6;3    111100
 *  28..43      10  6;4    111101
 *  44..75      11  6;5    111110
 *  76..139      12  6;6    111111
 *  140-      14  6;v    101111
 */
static codeinfo deltay3[] = {
    { 1, 1, 1, 1, 0x0 },
    { 2, 2, 3, 3, 0x4 },
    { 3, 3, 4, 4, 0xc },
    { 4, 4, 5, 5, 0x1c },
    { 5, 6, 5, 4, 0xd },
    { 7, 13, 6, 3, 0x5 },
    { 14, 15, 7, 6, 0x3a },
    { 16, 19, 8, 6, 0x3b },
    { 20, 27, 9, 6, 0x3c },
    { 28, 43, 10, 6, 0x3d },
    { 44, 75, 11, 6, 0x3e },
    { 76, 139, 12, 6, 0x3f },
    { 140, INT_MAX, -1, 6, 0x2f }
    };
#define  NDELTAY3  (sizeof(deltay3)/sizeof(deltay3[0]))

encoder mmDeltaYCoders[] = {
    { 0, 0, deltay1, NDELTAY1 },
    { 0, 0, deltay2, NDELTAY2 },
    { 0, 0, deltay3, NDELTAY3 }
    };

/*
 * A couple of other encoders:
 * width/height deltas for the tokens,
 * use counts for the tokens.
 */

/*
 * The width/height delta encoder
 *  value      bits  fields    prefix
 *  0      1  1    0
 *  1      2  2    10
 *  2      3  3    110
 *  3..9      7  4    1110
 *  10..infinity      4;v    1111
 *   END      7  7    1110111
 */

static codeinfo whEncoder[] = {
    { 0, 0, 1, 1, 0x0 },
    { 1, 1, 2, 2, 0x2 },
    { 2, 2, 3, 3, 0x6 },
    { 3, 9, 7, 4, 0xe },
    { 10, INT_MAX, -1, 4, 0xf }
    };
#define  NWH1  (sizeof(whEncoder)/sizeof(whEncoder[0]))

encoder mmWHCoders[] = {
    { 7, 0x77, whEncoder, NWH1 }
    };

/*
 * The use count encoders
 *
 * This one is used when there are use counts of 0 or 1 present in the
 * file (i.e., the singletons and outliers haven't been shuffled off
 * elsewhere).
 *  value      bits  fields    prefix
 *  0      3  3    100
 *  1      1  1    0
 *  2      3  3    101
 *  3..34      8  3;5    110
 *  35..infinity      3;v    111
 */
static codeinfo useEncoder1[] = {
    { 0, 0, 3, 3, 0x4 },
    { 1, 1, 1, 1, 0x0 },
    { 2, 2, 3, 3, 0x5 },
    { 3, 34, 8, 3, 0x6 },
    { 35, INT_MAX, -1, 3, 0x7 }
    };
#define  NUSE1  (sizeof(useEncoder1)/sizeof(useEncoder1[0]))

/*
 * This one is used when the smallest use count in the dictionary is 2.
 *  value      bits  fields    prefix
 *  2      1  1    0
 *  3      3  3    100
 *  4      4  4    1010
 *  5..6      5  4;1    1011
 *  7..10      6  4;2    1100
 *  11..14      7  5;2    11110
 *  15..30      8  4;4    1101
 *  31..94      10  4;6    1110
 *  95..infinity      5;v    11111
 */
static codeinfo useEncoder2[] = {
    { 2, 2, 1, 1, 0x0 },
    { 3, 3, 3, 3, 0x4 },
    { 4, 4, 4, 4, 0xa },
    { 5, 6, 5, 4, 0xb },
    { 7, 10, 6, 4, 0xc },
    { 11, 14, 7, 5, 0x1e },
    { 15, 30, 8, 4, 0xd },
    { 31, 94, 10, 4, 0xe },
    { 95, INT_MAX, -1, 5, 0x1f }
    };
#define  NUSE2  (sizeof(useEncoder2)/sizeof(useEncoder2[0]))

encoder mmUseCoders[] = {
    { 0, 0, useEncoder1, NUSE1 },
    { 0, 0, useEncoder2, NUSE2 }
    };
