/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Author:  ?;  199n-9n
**
**  Filename:  mmtokdict.c.
**
**  Manipulation and Huffman encoding of the token dictionaries...
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Cleaned up after MK's script, which implements most of the dgp_types, et al.
**      o  Scattered liberr compliance issues.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtokcomp/mmtokdict.c,v 1.34 2003/02/10 22:53:42 mkrocken Exp $
 *
 */

#include <stdlib.h>
#if !defined(VXWORKS) && !defined(EQUATOR_CA_ENV)
#include <memory.h>
#endif
#include "dgp_state.h"
#include "dgp_err_control.h"
#include "huff.h"
/*#include "token_t.h" */
#include "tokpos_t.h"
#include "mmcompress.h"
#include "mmcompressprivate.h"
#include "PsgFix.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtokcomp/mmtokdict.c,v 1.34 2003/02/10 22:53:42 mkrocken Exp $");


/*************************************************************************
**
**  Function:  size_t tcmp_estimateTokDictStateSize(void *oldstate, token *tokens, Dgp_Int32 ntok)
**
**
**  Given a dictionary state block ("oldstate"), which may be NULL, and a bunch
**  of tokens, this routine figures out how much dictionary state memory (in
**  bytes) is required to hold the new tokens *plus* the stuff in the old
**  state. This is what it returns.
**
**
**	Arguments:
**		o  oldstate...
**		o  tokens is an array of token{}...
**		o  ntok...
**
**	Return value:
**		o  Estimate of dictionary state size, in bytes.
**
**	Modification history:
**      o  None.
**  
**/

size_t
tcmp_estimateTokDictStateSize(void *oldstate, token *tokens, Dgp_Int32 ntok) {
  Dgp_Int32 i, j;
  dictstate *state = oldstate;
    
  /* How many new tokens? */
  j = 0;
  for (i = 0; i < ntok; i++) {
    if (!tokens[i].committed) {
      j++;
    }
  }

  if (state != (dictstate *)NULL) {
    j += state->ntok;
  }

  return(sizeof(dictstate) + j * (sizeof(tokstate) + sizeof(Dgp_Int32)));
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  size_t tcmp_estimateTokDictStateSizeEx(void *oldstate, Dgp_Int32 ntok)
**
**  
**  See tcmp_estimateTokDictStateSize();  this one appears not to check if the tokens
**  have been committed already.
**
**	Arguments:
**		o  oldstate...
**		o  ntok...
**
**	Return value:
**		o  Estimate of dictionary state size, in bytes.
**
**	Modification history:
**      o  None.
**  
**/

size_t
tcmp_estimateTokDictStateSizeEx(void *oldstate, Dgp_Int32 ntok) {
  Dgp_Int32 j;
  dictstate *state = oldstate;

  j = ntok;
  
  if( state != (dictstate *)NULL) {
    j += state->ntok;
  }
  
  return (sizeof(dictstate) + j * (sizeof(tokstate) + sizeof(Dgp_Int32)));
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  size_t tcmp_estimateMergedTokDictStateSize(void **states, Dgp_Int32 nstates)
**
**
**  Given a bunch of dictionary state blocks, this routine figures out how
**  much dictionary state memory (in bytes) is required to hold the union
**  of those previous states.  This is what it returns.
**
**	Arguments:
**		o  states is an array of (dictstate *)...
**		o  nstates is length(states)...
**
**	Return value:
**		o  Estimate of dictionary state size union, in bytes.
**
**	Modification history:
**      o  None.
**  
**/

size_t
tcmp_estimateMergedTokDictStateSize(void **states, Dgp_Int32 nstates) {
  Dgp_Int32 i;
  Dgp_Int32 ntok = 0;
  dictstate *state;
    
  /* How many new tokens? */
  for (i = 0; i < nstates; i++) {
    state = (dictstate *)states[i];
    ntok += state->ntok;
  }

  return(sizeof(dictstate) + ntok * (sizeof(tokstate) + sizeof(Dgp_Int32)));
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode tcmpi_ensureHuffmanCodes(dictstate *state, Dgp_State *dgp_state)
**
**
**  Given a dictionary state, make sure that the Huffman codes have been
**  computed.
**
**  This routine computes the Huffman codes and Huffman decoding tables.
**  The tables are built so that if a code is read with length l,
**  and its value is firstcodes[l] + v, then its
**  index in the dictionary state block (*NOT* its toknum value) is
**  codeindex[l][v].
**
**	Arguments:
**		o  state...
**		o  dgp_state points to a prepared Dgp_State{}.
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

ErrCode
tcmpi_ensureHuffmanCodes(dictstate *state, Dgp_State *dgp_state) {
  DGP_PROLOGUE("tcmpi_ensureHuffmanCodes()", DGP_PERRSTATE);
  DgpEncode_Huff    *usecodes = (DgpEncode_Huff *)NULL;
  Dgp_UInt32        firstcodes[DGP_MAX_CODELEN + 1];
  Dgp_Int32         *codeindex[DGP_MAX_CODELEN + 1];
  Dgp_Int32         counts[DGP_MAX_CODELEN + 1];
  Dgp_Int32         *p;
  Dgp_Int32         ntok;
  tokstate          *toks;
  Dgp_Int32         i, j;
  Dgp_Allocators    *a = DGP_PALLOCATORS;

  ntok = state->ntok;
  toks = state->toks;

  if (state->codesValid) {
    NORMAL_RETURN;
  }

  /* Get into a known state initially */
  for (i = 0; i <= DGP_MAX_CODELEN; i++) {
    codeindex[i] = (Dgp_Int32 *)NULL;
  }

  DGP_CHECK_ALLOC_CLEANUP(usecodes = (DgpEncode_Huff *)(a->malloc)(ntok * sizeof(*usecodes)));

  /*
   * Write in the use codes for the tokens. Remember, tokens with
   * use count of zero really have a count of one.
   */
  for (i = 0; i < ntok; i++) {
    usecodes[i].value = toks[i].tok->uses ? toks[i].tok->uses : 1;
  }

  /* Figure out the Huffman codes for each token */
  DGP_ERR_PASS_CLEANUP(DgpEncode_huffmanCode(dgp_state, usecodes, ntok));
  /*DGP_ERR_GEN_CLEANUP((!huffmanCode(usecodes, ntok)), cErrInternal); */

  /* Build the decoding tables */
  DGP_ERR_PASS_CLEANUP(DgpEncode_huffBuildDecodeTables(dgp_state, usecodes, ntok, firstcodes, codeindex));
  /*DGP_ERR_GEN_CLEANUP((!huffBuildDecodeTables(usecodes, ntok, firstcodes, codeindex)), cErrInternal); */

  /* That was the last chance for failure. */

  /* Figure out the counts for each code length... */
  for (i = DGP_MAX_CODELEN; i > 0; i--) {
    counts[i] = 2 * firstcodes[i - 1] - firstcodes[i];
  }
  if (ntok == 1) {
    /* This is the only way to have a zero-length code. */
    counts[0] = 1;
  } else {
    counts[0] = 0;
  }

  /* ...and copy them into the dictstate block. */
  for (i = 0, p = state->codeindexblock;
       i <= DGP_MAX_CODELEN;
       p += counts[i], i++) {
    if (counts[i]) {
      /* If counts is zero, can't bank on valid pointers. */

      /* Do the mapping to indices into the toks array... */
      for (j = 0; j < counts[i]; j++) {
          /*
           * At this point:
           * codeindex[j] is an index in the usecodes array.
           * Because of the way that we built the usecodes array,
           * each usecodes entry corresponds to the toks[]
           * entry with the same index. Therefore, the index in the
           * toks array we want is just codeindex[j].
           */
          p[j] = codeindex[i][j];
      }
      state->codeindex[i] = p;
      
      (a->free)(codeindex[i]);
      codeindex[i] = (Dgp_Int32 *)NULL;
    }
    state->firstcodes[i] = firstcodes[i];
  }

  /* ...and Huffman codes for all the tokens. */
  for (i = 0; i < ntok; i++) {
    toks[i].code = usecodes[i].code;
    /*  Added type cast here;  dangerous? */
    toks[i].codelen = (Dgp_UInt8)usecodes[i].codelen;
  }

  /* Free stuff. */
  (a->free)((void *)usecodes);

  for (i = 0; i <= DGP_MAX_CODELEN; i++) {
    if (codeindex[i] != (Dgp_Int32 *)NULL) {
      (a->free)((void *)codeindex[i]);
    }
  }

  /* The codes are now valid - yay! */
  state->codesValid = DGP_TRUE;

  NORMAL_RETURN;

ERR_CLEANUP:

  if (usecodes != (DgpEncode_Huff *)NULL) {
    (a->free)((void *)usecodes);
  }

  for (i = 0; i <= DGP_MAX_CODELEN; i++) {
    if (codeindex[i] != (Dgp_Int32 *)NULL) {
      (a->free)((void *)codeindex[i]);
    }
  }

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  void tcmp_initializeTokDictState(void *oldstate, void *newstatemem, size_t memsize)
**
**
**
**  Given a dictionary state block ("oldstate"), some newly
**  allocated dictionary state memory, and the size of that memory block,
**  initialize the new state block, and copy over all the stuff from the old
**  state block. The old state block should then be discarded.
**
**  oldstate and newstatemem can be the same pointer, but should not otherwise
**  overlap.
**
**  oldstate can be NULL.
**
**	Arguments:
**		o  oldstate...
**      o  newstatemem...
**		o  memsize...
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

void
tcmp_initializeTokDictState(void *oldstate, void *newstatemem, size_t memsize) {
  DGP_VOID_PROLOGUE("tcmp_initializeTokDictState()", NULL);
  dictstate         *olddstate = oldstate;
  dictstate         *newdstate = newstatemem;
  Dgp_Int32         ntokcapacity;
  tokstate          *oldtoks;
  Dgp_Int32         *oldcodeindexblock;
  Dgp_Int32         i;

  /*
   * How many tokens can the new state memory hold?
   * This is its capacity, not the number it's currently holding;
   * that number is olddstate->ntok.
   */
  ntokcapacity = (memsize - sizeof(dictstate)) / (sizeof(tokstate) + sizeof(Dgp_Int32));
    
  /* Where did they get that memory size? */
  DGP_VOID_ASSERT(ntokcapacity * (sizeof(tokstate) + sizeof(Dgp_Int32)) + sizeof(dictstate) == memsize);

  /* Make sure all the old stuff still fits. */
  DGP_VOID_ASSERT((oldstate == (void *)NULL) || (ntokcapacity >= olddstate->capacity));

  /* Patch up the pointers. */
  /* Cross your fingers and hope for alignment... */
  newdstate->toks = (tokstate *)((Dgp_Int8 *)newstatemem + sizeof(dictstate));
  newdstate->codeindexblock = (Dgp_Int32 *)((Dgp_Int8 *)newdstate->toks + ntokcapacity * sizeof(tokstate));

  if (oldstate == newstatemem) {
    /*
     * The block may have moved... what probably happened is
     * that they did a realloc() and passed in the value returned
     * as both "old" and "new" (this is quite legal). realloc()
     * might have extended the memory area in place, or copied
     * it over to a different location.
     * We can't trust oldtoks and oldcodeindexblock, since they
     * could refer to the block's old location.
     * The token info block shouldn't have to be moved, but the
     * codeindexblock may have to be. Worry about overlaps.
     *
     * Of course, if the old codes weren't valid, it doesn't matter
     * what the contents of codeindex was.
     */
    newdstate->codesValid = olddstate->codesValid;

    if (olddstate->codesValid) {
      oldtoks = (tokstate *)((Dgp_Int8 *)olddstate + sizeof(dictstate));
      DGP_VOID_ASSERT(oldtoks == newdstate->toks);
      oldcodeindexblock = (Dgp_Int32 *)((Dgp_Int8 *)oldtoks + olddstate->capacity * sizeof(tokstate));
      if (ntokcapacity > olddstate->capacity) {
          /*
           * Copy top down - gets the overlap right.
           * We only need to copy ntok values, not necessarily
           * capacity.
           */
          for (i = olddstate->ntok - 1; i >= 0; i--) {
            newdstate->codeindexblock[i] = oldcodeindexblock[i];
          }
      }

      /*
       * Fix up the codeindex pointers - should be the same offsets
       * into the block as they were before.
       */
      for (i = 0; i <= DGP_MAX_CODELEN; i++) {
          newdstate->codeindex[i] = newdstate->codeindexblock + (newdstate->codeindex[i] - oldcodeindexblock);
      }
    }

    /*
     * The rest of the dictstate block header (ntok and firstcodes)
     * shouldn't need to be touched - realloc will have copied them.
     */
  } else if (oldstate == (void *)NULL) {
    /* Just do an init. */
    newdstate->ntok = 0;
    newdstate->codesValid = DGP_FALSE;
    for (i = 0; i <= DGP_MAX_CODELEN; i++) {
      /* These pointers are uninitialised so far. */
      newdstate->codeindex[i] = (Dgp_Int32 *)NULL;
    }
  } else {
    /* Copy it across then patch pointers. */
    /* Can't overlap, so memcpy is safe. */
    /* Get the pointers to the old memory - they're still valid */
    oldtoks = olddstate->toks;
    oldcodeindexblock = olddstate->codeindexblock;
    memcpy(newdstate->toks, oldtoks, olddstate->ntok * sizeof(tokstate));
    memcpy(newdstate->codeindexblock, oldcodeindexblock, olddstate->ntok * sizeof(Dgp_Int32));

    DGP_VOID_ASSERT( olddstate);
    DGP_VOID_ASSERT( olddstate->firstcodes);

    /* Fix up the offsets. Copy firstcodes[] while you're at it. */
    for (i = 0; i <= DGP_MAX_CODELEN; i++) {
      newdstate->firstcodes[i] = olddstate->firstcodes[i];
      newdstate->codeindex[i] = newdstate->codeindexblock + (olddstate->codeindex[i] - oldcodeindexblock);
    }
    newdstate->ntok = olddstate->ntok;
    newdstate->codesValid = olddstate->codesValid;
  }

  /* Record the new capacity. */
  newdstate->capacity = ntokcapacity;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode tcmp_mergeTokDictStates(void        *newstate, 
**                                             size_t      newsize,
**                                             void        **states,
**                                             Dgp_Int32   nstates, 
**                                             Dgp_State   *dgp_state)
**
**
**  Given a bunch of dictionary state blocks, some newly
**  allocated dictionary state memory, and the size of that memory block,
**  initialize the new state block, and copy over all the stuff from the old
**  state blocks.
**
**	Arguments:
**		o  newstate...
**      o  newsize...
**		o  states...
**      o  nstates...
**      o  dgp_state...
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

ErrCode
tcmp_mergeTokDictStates(void        *newstate, 
                        size_t      newsize,
                        void        **states,
                        Dgp_Int32   nstates, 
                        Dgp_State   *dgp_state) {
  DGP_PROLOGUE("tcmp_mergeTokDictStates()", DGP_PERRSTATE);

  dictstate         *olddstate;
  dictstate         *newdstate = newstate;
  Dgp_Int32         ntokcapacity;
  Dgp_Int32         ntok;
  tokstate          *toks;
  Dgp_Int32         i, j, n;

  DGP_STOP_WHINE;
  /*
   * How many tokens can the new state memory hold?
   */
  ntokcapacity = (newsize - sizeof(dictstate)) / (sizeof(tokstate) + sizeof(Dgp_Int32));

  /* Where did they get that memory size? */
  DGP_ASSERT(ntokcapacity * (sizeof(tokstate) + sizeof(Dgp_Int32)) + sizeof(dictstate) == newsize);

  newdstate->capacity = ntokcapacity;

  /* How many new tokens? */
  ntok = 0;
  for (i = 0; i < nstates; i++) {
    olddstate = (dictstate *)states[i];
    ntok += olddstate->ntok;
  }

  /* Make sure all the old stuff still fits. */
  DGP_ASSERT(ntokcapacity >= ntok);

  /* Patch up the pointers. */
  /* Cross your fingers and hope for alignment... */
  newdstate->toks = toks = (tokstate *)((Dgp_Int8 *)newstate + sizeof(dictstate));
  newdstate->codeindexblock = (Dgp_Int32 *)((Dgp_Int8 *)newdstate->toks + ntokcapacity * sizeof(tokstate));

  /* Copy all the tokens across to the new block. */
  n = 0;
  for (i = 0; i < nstates; i++) {
    olddstate = (dictstate *)states[i];

    for (j = 0; j < olddstate->ntok; j++, n++) {
      toks[n] = olddstate->toks[j];
    }

  }

  /*
   * Don't build the Huffman decoding tables - just indicate that
   * they're invalid.  Special case: if we're "merging" a single
   * dictionary block (sigh, people call this routine to do that...) then
   * we can copy the codes if they were valid before.
   */
  if ((nstates == 1) && (((dictstate *)states[0])->codesValid == DGP_TRUE)) {
    olddstate = (dictstate *)states[0];
    /*
     * The codes and codelen entries in toks have already been copied.
     * We just have to copy the decoding tables.
     */
    memcpy(newdstate->codeindexblock, olddstate->codeindexblock, olddstate->ntok * sizeof(Dgp_Int32));

    /* Fix up the offsets. Copy firstcodes[] while you're at it. */
    for (i = 0; i <= DGP_MAX_CODELEN; i++) {
      newdstate->firstcodes[i] = olddstate->firstcodes[i];
      newdstate->codeindex[i] = newdstate->codeindexblock + (olddstate->codeindex[i] - olddstate->codeindexblock);
    }
  
    newdstate->codesValid = DGP_TRUE;
  } else {
    newdstate->codesValid = DGP_FALSE;
  }

  /* Update count of tokens in state */
  newdstate->ntok = ntok;

  NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode internal_convertTokensToState(Dgp_Int32         ntok,
**                                                   void              *dstate, 
**                                                   Dgp_Int32         newntok,
**                                                   token             **ordered)
**
**
**  This actually does the work of tcmp_convertTokensToState(); it's a separate
**  routine because it's also called from tcmp_encodeTokDict(), where newntok is
**  already known so it shouldn't be recomputed. Also the ordering of the
**  *new* tokens is significant... a mapping for them can be found in
**  ordered (i.e., ordered[i] is the i'th token in some canonical ordering,
**  which might not be the ordering in the dstate->toks array). The ordering
**  of the old tokens is not as hard, since they're written into the
**  dictionary state block in the canonical ordering.
**
**	Arguments:
**		o  ntok...
**      o  dstate...
**		o  newntok...
**      o  ordered...
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**      o  None -- should this take a Dgp_State{}, given the return type?
**  
**/

ErrCode
internal_convertTokensToState(Dgp_Int32         ntok,
                              void              *dstate, 
                              Dgp_Int32         newntok,
                              token             **ordered) {
  Dgp_Int32         i;
  dictstate         *state = dstate;
  tokstate          *oldtoks = state->toks;
  Dgp_Int32         oldntok;

  /* Total number of tokens now. */
  oldntok = state->ntok;
  ntok = oldntok + newntok;

  /* Update count of tokens in state. */
  state->ntok = ntok;
  /* Remember all the info we need, and mark these things as "committed". */
  for (i = 0; i < newntok; i++) {
    /* This is where we do the permutation by ordered. */
    oldtoks[i + oldntok].tok = ordered[i];
    ordered[i]->committed = DGP_TRUE;
  }

  /* The state's Huffman codes are now invalid. */
  state->codesValid = DGP_FALSE;

#ifdef  DGP_MMTOK_DBUG
  for (i = 0; i < ntok; i++) {
      DGP_REPORTF2("internal_convertTokensToState() / i = %3d, w = %5d,", i, oldtoks[i].width);
      DGP_REPORTF2(" h = %5d, c = %08lx", oldtoks[i].height, oldtoks[i].code);
      DGP_REPORTF2("(%2d), u = %3d\n", oldtoks[i].codelen, oldtoks[i].uses);
  }
#endif

  return(cSuccessful);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode tcmp_convertTokensToState(token         *tokens, 
**                                               Dgp_Int32     ntok, 
**                                               void          *dstate,
**                                               Dgp_State     *dgp_state)
**
**
**  Given a bunch of tokens and a dictionary state (that had better have
**  room and be properly initialized), find the uncommitted tokens and
**  record the pertinent bits about them in *dstate, mark them committed,
**  and update the Huffman codes stored (for *all* the tokens) in *dstate.
**
**  We still need to compute an "ordered" map for these, but we don't
**  change their order, just collect the uncommitted tokens (they don't
**  have to be adjacent in *tokens).
**
**	Arguments:
**		o  tokens is the array to be converted.
**      o  ntok is length(tokens).
**		o  dstate...
**      o  dgp_state references a prepared Dgp_State{}...
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

ErrCode
tcmp_convertTokensToState(token         *tokens, 
                          Dgp_Int32     ntok, 
                          void          *dstate,
                          Dgp_State     *dgp_state) {
  DGP_PROLOGUE("tcmp_convertTokensToState()", DGP_PERRSTATE);
  Dgp_Int32         i, j;
  token             **ordered = (token **)NULL;
  Dgp_Int32         newntok;
  Dgp_Allocators    *a = DGP_PALLOCATORS;

  /* How many of the tokens are new? */
  newntok = 0;
  for (i = 0; i < ntok; i++) {
    if (!tokens[i].committed) {
      newntok++;
    }
  }

  DGP_CHECK_ALLOC_CLEANUP(ordered = (token **)(a->malloc)(newntok * sizeof(*ordered)));

  /* Read the new tokens into the "ordered" array. */
  for (i = 0, j = 0; i < ntok; i++) {
    if (!tokens[i].committed) {
      ordered[j]= &tokens[i];
      j++;
    }
  }

  DGP_ERR_PASS_CLEANUP(internal_convertTokensToState(ntok, dstate, newntok, ordered));

  /* Done with the ordered table */
  (a->free)((void *)ordered);

  NORMAL_RETURN;

ERR_CLEANUP:

  if (ordered != (token **)NULL) {
    (a->free)((void *)ordered);
  }

  ERR_RETURN;
}
/*************************************************************************/

#if 0
Dgp_Allocators tokCompGlobals = {
  malloc,
  free,
  calloc,
  realloc
};

/*************************************************************************
**
**  Function:  static void *calloc_internal(size_t nelem, size_t elsize)
**
**
**  Obsolete?  Currently compiled out;  doesn't look any different from rtl
**  version. / dhtesq.
**
**	Arguments:
**		o  nelem...
**      o  elsize...
**
**	Return value:
**		o  None.
**
**	Modification history:
**      o  None.
**  
**/

static void *
calloc_internal(size_t nelem, size_t elsize) {
  size_t totsize = nelem * elsize;
  void *area = malloc(totsize);

  if (area != (void *)0) {
    (void)memset(area, 0, totsize);
  }
  return(area);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  void tcmp_setAllocator(void      * (*newMalloc)(size_t), 
**                                    void      (*newFree)(void *),
**                                    void      * (*newCalloc)(size_t, size_t),
**                                    void      * (*newRealloc)(void *, size_t))
**
**
**  Obsolete?  Currently compiled out. / dhtesq.
**
**	Arguments:
**		o  newMalloc...
**      o  newFree...
**      o  newCalloc...
**      o  newRealloc...
**
**	Return value:
**		o  None.
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  Removed call to libwjr_set_alloc(), since libwjr is disappearing in the current overhaul.
**      It looks like this function is dead anyway.
**  
**/

void
tcmp_setAllocator(void      * (*newMalloc)(size_t), 
                  void      (*newFree)(void *),
                  void      * (*newCalloc)(size_t, size_t),
                  void      * (*newRealloc)(void *, size_t)) {
  a->malloc = newMalloc;
  a->free = newFree;
  a->calloc = newCalloc;
  a->realloc = newRealloc;

  if (a->calloc == (void * (*)(size_t, size_t))NULL) {
    a->calloc = calloc_internal;
  }
  /*
   * Tell libwjr to use those Dgp_Allocators too. Let it synthesize its own
   * calloc, though, since if we give it ours, then some other module
   * calling libwjr_set_alloc can't hope to pass it the same routines.
   */
  /*  Took this out, since libwjr is disappearing. / dhtesq. */
  /* libwjr_set_alloc(newMalloc, newFree, newCalloc); */
  return;
}
/*************************************************************************/
#endif
