/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Author:  ?;  199n-9n
**
**  Filename:  mmuncompress.c.
**
**  Various functions for decompressing and manipulating token bitmaps (?).
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Cleaned up after MK's script, which implements most of the dgp_types, et al.
**      o  Scattered liberr compliance issues.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/src/libtokcomp/mmuncompress.c,v 1.52 2003/02/10 22:53:42 mkrocken Exp $
 *
 */

#include <stdlib.h>
#include <sys/types.h>
#include <string.h>  /* for MSDEV */
#include <stdio.h>
#include "dgp_err_control.h"
#include "dgp_state.h"
#include "g4.h"
/*  Removed 22.2.00 / dhtesq. */
/*#include "undecl-stdio.h" */
/*#include "dbug.h" */
#include "huff.h"
#include "token_t.h"
#include "tokpos_t.h"
#include "mmcompress.h"
#include "mmcompressprivate.h"
#include "PsgFix.h"


DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libtokcomp/mmuncompress.c,v 1.52 2003/02/10 22:53:42 mkrocken Exp $");

/* A structure for bit-level access to a file. */
typedef struct {
  Dgp_UInt8 buf;
  Dgp_Int32 bitsinbuf;
  Dgp_UInt8 *inbuf;
  size_t inbufleft;
} bitfile;

#ifdef  INSTRUMENT
/* Some instrumentation for file size breakdown. */
typedef struct {
  Dgp_Int32 overhead;
  Dgp_Int32 padding;
  Dgp_Int32 bitmapsize;
  Dgp_Int32 bitmaps;
  Dgp_Int32 height;
  Dgp_Int32 width;
  Dgp_Int32 uses;
} dict_bituses;

typedef struct {
  Dgp_Int32 overhead;
  Dgp_Int32 padding;
  Dgp_Int32 firstx;
  Dgp_Int32 deltax;
  Dgp_Int32 deltay;
  Dgp_Int32 y;
  Dgp_Int32 toknum;
} pos_bituses;
#endif

/* Routines to access bitfiles. */

/*************************************************************************
**
**  Function:  static Dgp_Bool readCode(Dgp_UInt32 *code, Dgp_Int32 codelen, bitfile *f)
**
**
**  Read a few bits from a bitfile. Pack them into the codelen LSBs
**  of *code. Return DGP_TRUE on success, DGP_FALSE on failure.
**
**  The f->bitsinbuf LSBs of f->buf are the valid ones, since it gets
**  read out from the MSB down.
**
**	Arguments:
**		o  code...
**		o  codelen...
**		o  f...
**
**	Return value:
**		o  Boolean success/failure.
**
**	Modification history:
**      17.8.2000 / dhtesq.
**      o  Decoder profiling suggests this is a good place for an inline...
**  
**/

static INLINE Dgp_Bool
readCode(Dgp_UInt32 *code, Dgp_Int32 codelen, bitfile *f) {
  DGP_VOID_PROLOGUE("readCode()", NULL);

  DGP_VOID_ASSERT(code != (Dgp_UInt32 *)NULL);
  DGP_VOID_ASSERT(f != (bitfile *)NULL);

  *code = 0;
  while (codelen) {
    if (f->bitsinbuf >= codelen) {
      /* Enough in the buffer for us. */
      *code |= (f->buf >> (f->bitsinbuf - codelen)) & ((1ul << codelen) - 1);
      f->bitsinbuf -= codelen;
      return(DGP_TRUE);
    }
    /* Eat what's in the buffer and get more. */
    if (f->bitsinbuf) {
      *code |= (f->buf & ((1ul << f->bitsinbuf) - 1)) << (codelen - f->bitsinbuf);
      codelen -= f->bitsinbuf;
    }
    if (!f->inbufleft) {
      return(DGP_FALSE);
    }
    f->buf = *f->inbuf++;
    f->inbufleft--;
    f->bitsinbuf = BITS_UCHAR;
  }
  return(DGP_TRUE);
}
/*************************************************************************/

/* Read n bits into *v, bailing in case of failure */
#define  readChecking(v, n, f)        \
do {            \
    if (!readCode((v), (n), (f))) {    \
	    return(DGP_FALSE);        \
	}          \
} while(0)


/*************************************************************************
**
**  Function:  static Dgp_Bool readValue(Dgp_UInt32 *x, bitfile *f [, Dgp_Int32 *nbits])
**
**
**  Read a value from the file.
** 
**	Arguments:
**		o  x...
**		o  f...
**		o  (#ifdef INSTRUMENT) nbits...
**
**	Return value:
**		o  Boolean;  currently no control paths can return DGP_FALSE.
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  dbug.h dependency converted to a protected DGP_REPORT.
**  
**/

static Dgp_Bool
readValue(Dgp_UInt32 *x, bitfile *f
#ifdef  INSTRUMENT
, Dgp_Int32 *nbits
#endif
)
{
    Dgp_UInt32 v;

    readChecking(&v, 1, f);
    if (v == 0) {        /* 0 */
        readChecking(&v, 7, f);      /* value as 7 bits */

#ifdef  INSTRUMENT
        *nbits += 8;
#endif
        *x = v;
    } else {
        readChecking(&v, 1, f);
        if (!v) {        /* 10 */
            readChecking(&v, 10, f);    /* value-127 as 10 bits */
            *x = v + 128;
#ifdef  INSTRUMENT
            *nbits += 12;
#endif
        } else {          /* 11 */
            readChecking(&v, 15, f);    /* value as 15 bits */
      
            if (v < 1024) {
            	/*
            	 * The file really contains a 32-bit value. We just read
            	 * the top 10 bits of it (the first 5 of the 15 we got were
            	 * zero to flag us to this).
            	 */
            	*x = v << 22;
            	readChecking(&v, 22, f);
            	*x |= v;
#ifdef  INSTRUMENT
            	*nbits += 39;
#endif
            } else {
            	*x = v;
#ifdef  INSTRUMENT
            	*nbits += 17;
#endif
            }
        }
    }

#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("readValue() / (%08x)\n", *x);
#endif
/*  VDEBUG("rv(%08x)\n", *x, 50); */
  return(DGP_TRUE);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool readTokenCode(  Dgp_UInt32      *code,
**                                             Dgp_Int32       *codelen, 
**                                             Dgp_Int32       mincode,
**                                             Dgp_UInt32      *firstcodes, 
**                                             bitfile         *f
**                                	        [, Dgp_Int32       *nbits])
**
**
**  Read a token Huffman code.
** 
**	Arguments:
**		o  code...
**		o  codelen...
**		o  mincode...
**      o  firstcodes...
**      o  f...
**      o  (#ifdef INSTRUMENT) nbits...
**
**	Return value:
**		o  Boolean success/failure.
**
**	Modification history:
**      17.8.2000 / dhtesq.
**      o  Decoder profiling suggests this is a good place for an inline...
**  
**/

static INLINE Dgp_Bool
readTokenCode(  Dgp_UInt32      *code,
                Dgp_Int32       *codelen, 
                Dgp_Int32       mincode,
        	    Dgp_UInt32      *firstcodes, 
                bitfile         *f
#ifdef  INSTRUMENT
	          , Dgp_Int32       *nbits
#endif
                        	          ) {
  DGP_VOID_PROLOGUE("readTokenCode()", NULL);
  Dgp_UInt32 v;

  /* Get a code of the minimum length */
  if (!readCode(code, mincode, f)) {
    goto ERR_CLEANUP;
  }
  *codelen = mincode;
  while (*code < firstcodes[*codelen]) {
    /* Add another bit and try again */
    if (!readCode(&v, 1, f)) {
      goto ERR_CLEANUP;
    }
    *code = ((*code) << 1) | v;
    (*codelen)++;
    DGP_VOID_ASSERT(*codelen <= DGP_MAX_CODELEN);
  }

#ifdef  INSTRUMENT
  *nbits += *codelen;
#endif
  return(DGP_TRUE);

ERR_CLEANUP:
  return(DGP_FALSE);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool readEncoded(Dgp_Int32       *val, 
**                                         Dgp_Bool        *end, 
**                                         bitfile         *f, 
**                                         decoder         dec
**                                      [, Dgp_Int32       *inbits])
**
**
**  Read a code using the given decoding table.  Stick the result in *val and
**  *end.
**
**	Arguments:
**		o  val...
**		o  end...
**      o  f...
**      o  dec...
**      o  (#ifdef INSTRUMENT) nbits...
**
**	Return value:
**		o  Boolean success/failure.
**
**	Modification history:
**      17.8.2000 / dhtesq.
**      o  Decoder profiling suggests this is a good place for an inline...
**  
**/

static INLINE Dgp_Bool
readEncoded(Dgp_Int32       *val, 
            Dgp_Bool        *end, 
            bitfile         *f, 
            decoder         dec
#ifdef  INSTRUMENT
          , Dgp_Int32       *inbits
#endif
                                   ) {
  DGP_VOID_PROLOGUE("readEncoded()", NULL);
  Dgp_UInt32 v;
  Dgp_UInt32 vi;
  Dgp_Int32 state = 0;
  Dgp_Int32 nbits;
  Dgp_Int32 outstate;
  Dgp_Int32 offset;

  *end = DGP_FALSE;
  for (;;) {
    nbits = dec[state].nbits;
    outstate = dec[state].outstate;
    offset = dec[state].offset;
    if (outstate == 0) {
      /*
       * It's a terminal code. Read nbits bits, add that to offset,
       * and return it.
       * Special cases:
       * nbits < 0: either this is an end code (flagged by offset==0),
       * or a "value" encoded number.  If it's a value, then offset < 0
       * and offset > 0 need to be treated differently, since in one case
       * the value counts down from offset, and in the other it counts
       * up.
       */
      if (nbits < 0) {
        if (offset == 0) {
	    *end = DGP_TRUE;
	    *val = 0;
	    return(DGP_TRUE);
        } else if (offset > 0) {
	        if (!readValue(&vi, f
#ifdef  INSTRUMENT
			                     , inbits
#endif
                            			 )) {
        	    goto ERR_CLEANUP;
	        }
            *val = vi + offset;
            return(DGP_TRUE);
        } else {
    	  if (!readValue(&vi, f
#ifdef  INSTRUMENT
	                           , inbits
#endif
                                       )) {
    	    goto ERR_CLEANUP;
	      }
    	  *val = offset - vi;
	      return(DGP_TRUE);
        }
      } else if (nbits > 0) {
    	/* Read some bits, add to offset, and return. */
	    readChecking(&v, nbits, f);
#ifdef  INSTRUMENT
	    *inbits += nbits;
#endif
    	*val = v + offset;
	    return(DGP_TRUE);
      } else {
    	/* Just return offset */
	    *val = offset;
    	return(DGP_TRUE);
      }
    } else {
      /*
       * A non-terminal code - read some bits and make a state
       * transition
       */
      DGP_VOID_ASSERT(nbits > 0);
      readChecking(&v, nbits, f);
#ifdef  INSTRUMENT
      *inbits += nbits;
#endif
      state = outstate + v;
    }
  }
  /*NOTREACHED*/

ERR_CLEANUP:
  return(DGP_FALSE);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool readWHDelta(Dgp_UInt32      *d, 
**                                         Dgp_Bool        *end, 
**                                         bitfile         *f
**                                      [, Dgp_Int32       *nbits])
**
**
**  Read a width or height delta from the file. If there's an end code
**  instead of a real delta, set *end to DGP_TRUE; set it to DGP_FALSE and
**  put the delta in *d otherwise.
**
**	Arguments:
**		o  d...
**		o  end...
**      o  f...
**      o  (#ifdef INSTRUMENT) nbits...
**
**	Return value:
**		o  Boolean success/failure.
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  dbug.h dependency converted to a protected DGP_REPORT.
**  
**/

static Dgp_Bool
readWHDelta(Dgp_UInt32      *d, 
            Dgp_Bool        *end, 
            bitfile         *f
#ifdef  INSTRUMENT
          , Dgp_Int32       *nbits
#endif
                                  ) {
  Dgp_Int32 v = -1;

  if (!readEncoded(&v, end, f, mmWHDecoders[0]
#ifdef  INSTRUMENT
                                              , nbits
#endif
                                                     )) {
    return(DGP_FALSE);
  }

#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("readWHDelta() / %d\n", v);
#endif
/*  VDEBUG(*end?"rwhend()\n":"rwhd(%d)\n", v, 50); */

  *d = v;
  return(DGP_TRUE);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool readUseCount(Dgp_UInt32      *count, 
**                                          bitfile         *f,
**                                          Dgp_Int32       useEncoding,
**                                   	 [, Dgp_Int32       *nbits])
**
**  Read a use count.
**
**	Arguments:
**		o  count...
**		o  f...
**      o  useEncoding...
**      o  (#ifdef INSTRUMENT) nbits...
**
**	Return value:
**		o  Boolean success/failure.
**
**	Modification history:
**      o  None.
**  
**/

static Dgp_Bool
readUseCount(Dgp_UInt32 *count, bitfile *f, Dgp_Int32 useEncoding
#ifdef  INSTRUMENT
                                                        	     , Dgp_Int32 *nbits
#endif
	                                                                               ) {
  Dgp_Int32     v = -1;
  Dgp_Bool      dummy;

  if (!readEncoded(&v, &dummy, f, mmUseDecoders[useEncoding]
#ifdef  INSTRUMENT
		                                                    , nbits
#endif
                                                                   )) {
    return(DGP_FALSE);
  }

#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("readUseCount() / %d\n", v);
#endif
/* VDEBUG("ruc(%d)\n", v, 50); */
  *count = v;
#ifdef  VALPRINT
  printf("U %d\n", *count);
#endif
  return(DGP_TRUE);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool readYDelta(  Dgp_UInt32      *delta, 
**                                          bitfile         *f,
**                                          Dgp_Int32       deltaYEncoding
**                                       [, Dgp_Int32       *nbits])
**
**  Read in an inter-row y delta.
**
**
**	Arguments:
**		o  delta...
**		o  f...
**      o  deltaYEncoding...
**      o  (#ifdef INSTRUMENT) nbits...
**
**	Return value:
**		o  Boolean success/failure.
**
**	Modification history:
**      o  None.
**  
**/

static Dgp_Bool
readYDelta(Dgp_UInt32 *delta, bitfile *f, Dgp_Int32 deltaYEncoding
#ifdef  INSTRUMENT
                                                                  , Dgp_Int32 *nbits
#endif
                                                                                    ) {
  Dgp_Int32     v;
  Dgp_Bool      dummy;

  if (!readEncoded(&v, &dummy, f, mmDeltaYDecoders[deltaYEncoding]
#ifdef  INSTRUMENT
                                                                  , nbits
#endif
                                                                         )) {
    return(DGP_FALSE);
  }
#ifdef DGP_MMTOK_DBUG
  DGP_REPORTF1("readYDelta() / %d\n", v);
#endif
/*  VDEBUG("ryd(%d)\n", v, 50); */
  *delta = v;
#ifdef  VALPRINT
  printf("DY %d\n", *delta);
#endif
  return(DGP_TRUE);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool readYVal(Dgp_UInt32      *y, 
**                                      bitfile         *f, 
**                                      Dgp_Int32       logstripsize
**                                   [, Dgp_Int32       *nbits])
**
**  Read in an intra-row y position.
**
**	Arguments:
**		o  y...
**		o  f...
**      o  logstripsize...
**      o  (#ifdef INSTRUMENT) nbits...
**
**	Return value:
**		o  Boolean success/failure.
**
**	Modification history:
**      o  None.
**  
**/

static Dgp_Bool
readYVal(Dgp_UInt32 *y, bitfile *f, Dgp_Int32 logstripsize
#ifdef  INSTRUMENT
                                                          , Dgp_Int32 *nbits
#endif
	                                                                        ) {
  Dgp_UInt32 v = -1;

  if (!readCode(&v, logstripsize, f)) {
    return(DGP_FALSE);
  }
#ifdef  INSTRUMENT
  *nbits += logstripsize;
#endif

#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("readYVal() / %d\n", v);
#endif
/*  VDEBUG("ryv(%d)\n", v, 50); */

  *y = v;
#ifdef  VALPRINT
  printf("Y %d\n", *y);
#endif
  return(DGP_TRUE);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static Dgp_Bool readXDelta(  Dgp_Int32       *delta, 
**                                          Dgp_Bool        *end, 
**                                          bitfile         *f,
**                                          Dgp_Int32       deltaXEncoding
**                                       [, Dgp_Int32       *nbits])
**
**
**  Read in an intra-row x delta, or an end-of-row code.
**
**	Arguments:
**		o  delta...
**		o  end...
**      o  f...
**      o  deltaXEncoding...
**      o  (#ifdef INSTRUMENT) nbits...
**
**	Return value:
**		o  Boolean success/failure.
**
**	Modification history:
**      17.8.2000 / dhtesq.
**      o  Decoder profiling suggests this is a good place for an inline...
**  
**/

static INLINE Dgp_Bool
readXDelta(Dgp_Int32 *delta, Dgp_Bool *end, bitfile *f, Dgp_Int32 deltaXEncoding
#ifdef  INSTRUMENT
                                                                                , Dgp_Int32 *nbits
#endif
	                                                                                              ) {
  Dgp_Int32 v = -1;

  if (!readEncoded(&v, end, f, mmDeltaXDecoders[deltaXEncoding]
#ifdef  INSTRUMENT
                                                    		   , nbits
#endif
                                                                      )) {
    return(DGP_FALSE);
  }

#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("readXDelta() / %d\n", v);
#endif
/*  VDEBUG(*end?"reor()\n":"rxd(%d)\n", v, 50); */

  *delta = v;
#ifdef  VALPRINT
  if (*end) {
    printf("DXE\n");
  }
  else {
    printf("DX %d\n", *delta);
  }
#endif
  return(DGP_TRUE);
}
/*************************************************************************/

/*************************************************************************
**
**  Function:  static Dgp_Bool readFirstX(  Dgp_Int32       *firstx, 
**                                          bitfile         *f,
**                                          Dgp_Int32       firstXEncoding
**                                       [, Dgp_Int32       *nbits])
**
**
**  Read the first x value from a row.
**
**	Arguments:
**		o  firstx...
**		o  f...
**      o  firstXEncoding...
**      o  (#ifdef INSTRUMENT) nbits...
**
**	Return value:
**		o  Boolean success/failure.
**
**	Modification history:
**      o  None.
**  
**/

static Dgp_Bool
readFirstX(Dgp_Int32 *firstx, bitfile *f, Dgp_Int32 firstXEncoding
#ifdef  INSTRUMENT
                                                                  , Dgp_Int32 *nbits
#endif
                                                                                    ){
  Dgp_Bool dummy;

  if (!readEncoded(firstx, &dummy, f, mmFirstXDecoders[firstXEncoding]
#ifdef  INSTRUMENT
		   , nbits
#endif
		   )) {
    return(DGP_FALSE);
  }

#ifdef DGP_MMTOK_DBUG
    DGP_REPORTF1("readFirstX() / %d\n", *firstx);
#endif
/*  VDEBUG("rfx(%d)\n", *firstx, 50); */

#ifdef  VALPRINT
  printf("FX %d\n", *firstx);
#endif
  return(DGP_TRUE);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static void readoutModel(Dgp_UInt32     *src, 
**                                      Dgp_Int32      bitoffset, 
**                                      Dgp_Int32      srcwidth, 
**                                      Dgp_Int32      height,
**                                      Dgp_UInt32     *dest, 
**                                      Dgp_Int32      destwidth)
**
**
**  Read out destwidth bits from bitoffset bits into the image stored in
**  src, and write them to dest (filling from the MSB as always). Everything
**  should be height rows high, and the rows should be padded to the nearest
**  Dgp_Int32.
**
**  Both the source and destination are supposed to be accessed Dgp_Int32-by-Dgp_Int32.
**
**	Arguments:
**		o  src...
**		o  bitoffset...
**      o  srcwidth...
**      o  height...
**      o  dest...
**      o  destwidth...
**
**	Return value:
**		o  None.
**
**	Modification history:
**      o  None.
**  
**/

static void
readoutModel(Dgp_UInt32     *src, 
             Dgp_Int32      bitoffset, 
             Dgp_Int32      srcwidth, 
             Dgp_Int32      height,
    	     Dgp_UInt32     *dest, 
             Dgp_Int32      destwidth) {
  Dgp_Int32         i, j;
  Dgp_Int32         src_row_width = (srcwidth + BITS_ULONG - 1) / BITS_ULONG;
  Dgp_Int32         dest_row_width = (destwidth + BITS_ULONG - 1) / BITS_ULONG;
  Dgp_Int32         Dgp_Int32offset;
  Dgp_UInt32        *p;
  Dgp_UInt32        *q;
  Dgp_UInt32        left;
  Dgp_Int32         bitsinlastDgp_Int32;
  Dgp_UInt32        lastmask;

  Dgp_Int32offset = bitoffset / BITS_ULONG;
  bitoffset = bitoffset % BITS_ULONG;

  /*
   * How many bits are valid in the last Dgp_Int32 of dest?
   * We want to make sure they're cleared in the output, so we mask them.
   * Set lastmask to a value with the leftmost bitsinlastDgp_Int32 bits set.
   */
  bitsinlastDgp_Int32 = destwidth % BITS_ULONG;
  if (bitsinlastDgp_Int32 == 0) {
    bitsinlastDgp_Int32 = BITS_ULONG;
    lastmask = ~0ul;
  }
  else {
    lastmask =
      ((1ul << bitsinlastDgp_Int32) - 1) << (BITS_ULONG - bitsinlastDgp_Int32);
  }

  src += Dgp_Int32offset;
  for (i = 0; i < height; i++) {
    /* Find the row bases */
    p = src + i * src_row_width;
    q = dest + i * dest_row_width;

    /* Go through Dgp_Int32-by-Dgp_Int32 */
    if (bitoffset != 0) {
      left = *p++ << bitoffset;
      for (j = 0; j < dest_row_width - 1; j++) {
    	*q++ = left | (*p >> (BITS_ULONG - bitoffset));
	    left = *p++ << bitoffset;
      }
      /* Do we need to touch another source Dgp_Int32? */
      if (bitoffset + bitsinlastDgp_Int32 > BITS_ULONG) {
    	*q++ = (left | (*p >> (BITS_ULONG - bitoffset))) & lastmask;
      }
      else {
	    *q++ = left & lastmask;
      }
    }
    else {
      /* No shifting - yay! */
      for (j = 0; j < dest_row_width; j++) {
    	*q++ = *p++;
      }
      /* Mask off the last Dgp_Int32 properly */
      *(q - 1) &= lastmask;
    }
  }
  return;
}
/*************************************************************************/


#define TCMP_READ_CHECK(cond) \
  DGP_ERR_GEN_CLEANUP_NO_MSG((cond), cErrBufferExceeded)


/*************************************************************************
**
**  Function:  ErrCode tcmp_getTokenCount(void *buffer, size_t buflen, Dgp_State *dgp_state, size_t *count)
**
**
**  This routine takes an encoded dictionary block and returns the number of
**  tokens encoded in it. If the block contains a dictionary clearing code, it
**  returns zero. In this case, there is an (as-yet undefined) procedure to be
**  followed.
** 
**	Arguments:
**		o  buffer...
**		o  buflen...
**      o  dgp_state points to a prepared Dgp_State{}.
**      o  count...
**
**	Return value:
**		o  ErrCode;  throws cErrBufferExceeded on failure of readValue().
**
**	Modification history:
**      o  None.
**  
**/

ErrCode
tcmp_getTokenCount(void *buffer, size_t buflen, Dgp_State *dgp_state, size_t *count) {
  DGP_PROLOGUE("tcmp_getTokenCount()", DGP_PERRSTATE);
  bitfile       f;
#ifdef  INSTRUMENT
  Dgp_Int32     dummy = 0;
#endif

  f.inbuf = buffer;
  f.inbufleft = buflen;
  f.bitsinbuf = 0;
  f.buf = 0;
    
#ifdef  INSTRUMENT
  TCMP_READ_CHECK(!readValue((Dgp_UInt32 *)count, &f , &dummy));
#else
  TCMP_READ_CHECK(!readValue((Dgp_UInt32 *)count, &f));
#endif

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static ErrCode readHCTokens(token          *tokens, 
**                                         Dgp_Int32      *pmodelnum, 
**                                         Dgp_Int32      *pheight,
**                                         Dgp_Int32      useEncoding, 
**                                         bitfile        *pf, 
**                                         Dgp_Int32      *pclasswidth,
**                                         Dgp_Bool       *pend, 
**                                         Dgp_State      *dgp_state
**                                      [, dict_bituses   *pbituses])
**
**
**  Assumes that "pf" is a the offset of the delta height value for this class.
**  Read the token fields from a single height class.
**  Note that the "bits" field of a token struct remained undefined.
** 
**	Arguments:
**		o  tokens...
**		o  pmodelnum...
**      o  pheight...
**      o  useEncoding...
**      o  pf...
**      o  pclasswidth...
**      o  pend...
**      o  dgp_state points to a prepared Dgp_State{}.
**      o  (#ifdef INSTRUMENT) pbituses
**
**	Return value:
**		o  ErrCode;  throws cErrBufferExceeded on failure of readWHDelta() or readUseCount().
**
**	Modification history:
**      o  None.
**  
**/

static
ErrCode readHCTokens(token          *tokens, 
                     Dgp_Int32      *pmodelnum, 
                     Dgp_Int32      *pheight,
                     Dgp_Int32      useEncoding, 
                     bitfile        *pf, 
                     Dgp_Int32      *pclasswidth,
                     Dgp_Bool       *pend, 
                     Dgp_State      *dgp_state
#ifdef  INSTRUMENT
                   , dict_bituses   *pbituses
#endif
                                             ) {
  Dgp_Int32     modelnum = *pmodelnum;
  Dgp_Int32     height = *pheight;
  Dgp_Int32     width;
  Dgp_Int32     classwidth;
  Dgp_UInt32    v;
  Dgp_Bool      end;
 
#ifdef  INSTRUMENT
  dict_bituses bituses = *pbituses;
#endif

DGP_PROLOGUE( "readHCTokens()", DGP_PERRSTATE);

  /* Read the delta height for this class */
#ifdef  INSTRUMENT
  TCMP_READ_CHECK(!readWHDelta(&v, &end, pf , &bituses.height));
#else
  TCMP_READ_CHECK(!readWHDelta(&v, &end, pf));
#endif

  if (end) {
    /* Done - no more height classes */
#ifdef  VALPRINT
    printf("HE\n");
#endif
    *pend = DGP_TRUE;
    NORMAL_RETURN;
  }
#ifdef  VALPRINT
  printf("H %d\n", v + 1);
#endif

  /* Read in the size, use count info for the models */
  height += v + 1;
  width = 0;
  classwidth = 0;

  for (;;) {
#ifdef  INSTRUMENT
    TCMP_READ_CHECK(!readWHDelta(&v, &end, pf , &bituses.width));
#else
    TCMP_READ_CHECK(!readWHDelta(&v, &end, pf));
#endif
    if (end) {
#ifdef  VALPRINT
      printf("WE\n");
#endif
      break;
    }
#ifdef  VALPRINT
    printf("W %d\n", v);
#endif
    width += v;
    classwidth += width;

#ifdef  INSTRUMENT
    TCMP_READ_CHECK(!readUseCount(&v, pf, useEncoding , &bituses.uses));
#else
    TCMP_READ_CHECK(!readUseCount(&v, pf, useEncoding));
#endif

    tokens[modelnum].width = (Dgp_UInt16)width;
    tokens[modelnum].height = (Dgp_UInt16)height;
    tokens[modelnum].uses = v;
    tokens[modelnum].longsperline = (Dgp_UInt16)((tokens[modelnum].width + BITS_ULONG - 1) / BITS_ULONG);
    tokens[modelnum].committed = DGP_FALSE;
    tokens[modelnum].userflags = 0;
    tokens[modelnum].extdata = (void *)NULL;
    modelnum++;
  }

  *pmodelnum = modelnum;
  *pheight = height;
  *pclasswidth = classwidth;
  *pend = DGP_FALSE;

  NORMAL_RETURN;

ERR_CLEANUP:
  
  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  static ErrCode readHCBitmaps(token             *tokens, 
**                                          Dgp_Int32         firstmodelnum, 
**                                          Dgp_Int32         lastmodelnum,
**                                          Dgp_Int32         height, 
**                                          Dgp_Int32         classwidth,
**                                          bitfile           *pf, 
**                                          void              *g4Decoder, 
**                                          Dgp_State         *dgp_state
**                                       [, dict_bituses      *pbituses])
**
**
**  Assumes that "pf" is at the offset of the size of the compressed 
**  image strip for the height class starting at token "firstmodelnum" and
**  continuing for until (but not including) "lastmodelnum". "height" is the
**  height of this class.
**
** 
**	Arguments:
**		o  tokens...
**		o  firstmodelnum...
**      o  lastmodelnum...
**      o  height...
**      o  classwidth...
**      o  pf...
**      o  g4Decoder...
**      o  dgp_state points to a prepared Dgp_State{}.
**      o  (#ifdef INSTRUMENT) pbituses
**
**	Return value:
**		o  ErrCode on any of various failures.
**
**	Modification history:
**      o  None.
**  
**/

static
ErrCode readHCBitmaps(token             *tokens, 
                      Dgp_Int32         firstmodelnum, 
                      Dgp_Int32         lastmodelnum,
                      Dgp_Int32         height, 
                      Dgp_Int32         classwidth,
                      bitfile           *pf, 
                      void              *g4Decoder, 
                      Dgp_State         *dgp_state

#ifdef  INSTRUMENT
                    , dict_bituses *pbituses
#endif
                                            ) {
  Dgp_Int32         g4len;
  Dgp_UInt32        *classbits = (Dgp_UInt32 *)NULL;
  Dgp_Int32         classrow_width;
  Dgp_UInt32        v;
  Dgp_Int32         i, classstart = firstmodelnum, bitoffset;
  Dgp_UInt32        tmp;
  Dgp_Allocators    *a = DGP_PALLOCATORS;

    DGP_PROLOGUE( "readHCBitmaps()", DGP_PERRSTATE);

  /* Read in the g4 length */
#ifdef  INSTRUMENT
  TCMP_READ_CHECK(!readValue(&v, pf , &pbituses->bitmapsize));
#else
  TCMP_READ_CHECK(!readValue(&v, pf));
#endif
#ifdef  VALPRINT
  printf("BL %d\n", v);
#endif
  g4len = v;

  /* Flush the partial byte in the buffer... */
#ifdef  INSTRUMENT
  pbituses->padding += pf->bitsinbuf;
#endif
  pf->bitsinbuf = 0;
  pf->buf = 0;

  /* Allocate the bitmap to decode into... */
  classrow_width = (classwidth + BITS_ULONG - 1) / BITS_ULONG;
  DGP_CHECK_ALLOC_CLEANUP(classbits = (Dgp_UInt32 *)(a->calloc)((size_t)height, 
                                      (size_t)classrow_width * sizeof(*classbits)));

  /* ...and decode it */
  if (g4len == 0) {
    /*
     * It's stored as a bitmap. Simple... except that it was
     * stored for byte-by-byte access, and we want Dgp_Int32-by-Dgp_Int32...
     */
    Dgp_Int32 byte_width = (classwidth + BITS_UCHAR - 1) / BITS_UCHAR;

    /* Make sure there's as much data there as people claim,
       (lousy encoding bugs...) */
    TCMP_READ_CHECK(pf->inbufleft < (size_t)(byte_width * height));

    for (i = 0; i < height; i++) {
      Dgp_UInt32        *p = classbits + i * classrow_width;
      Dgp_UInt8         *p2;
      Dgp_UInt32        *q;
      Dgp_Int32         ix, ix2;

      memcpy((Dgp_Int8 *)p, pf->inbuf + i * byte_width, (size_t)byte_width);
      for (ix = byte_width, p2 = (Dgp_Int8 *)p + byte_width;
           ix < classrow_width * (Dgp_Int32)sizeof(Dgp_UInt32);
           ix++) {
        *p2++ = 0;
      }
      /* Shuffle the bytes. */
      for (ix = 0, p2 = (Dgp_Int8 *)p, q = p;
           ix < classrow_width;
           ix++) {
        tmp = 0;
        for (ix2 = 0; ix2 < (Dgp_Int32)sizeof(Dgp_UInt32); ix2++) {
          tmp |= (*p2++) << (BITS_ULONG - (ix2 + 1) * BITS_UCHAR);
        }
        *q++ = tmp;
      }
    }

    pf->inbuf += height * byte_width;
    pf->inbufleft -= height * byte_width;
#ifdef  INSTRUMENT
    pbituses->bitmaps += height * byte_width * BITS_UCHAR;
#endif
  }
  else {
    /* It's stored as G4-compressed. */
    
    /* Make sure there's as much data as there should be. */
    TCMP_READ_CHECK(pf->inbufleft < (size_t)g4len);
    
      /*
       * If the class is a single token, and that token is single-use,
       * then return the G4-compressed data rather than the bitmap.
       * Flag this by setting longsperline=0, and put the length of
       * the G4 data into uses (the real value of uses is zero in this
       * case, so we can "safely" re-use the field).
       */
      if ((classstart == lastmodelnum - 1) && (tokens[classstart].uses == 0)) {
	      /* Here's the special case - don't do anything here */
      } else {
    	DGP_ERR_PASS_CLEANUP(g4DecodeC( g4Decoder, 
                                        classbits, 
                                        classwidth, 
                                        height, 
                                        classrow_width, 
                                        0, 
                                        classwidth - 1, 
                                        height - 1,
                                        pf->inbuf,
                                        g4len, 
                                        DGP_FALSE, 
                                        (size_t *)NULL));
      }
      pf->inbuf += g4len;
      pf->inbufleft -= g4len;
#ifdef  INSTRUMENT
      pbituses->bitmaps += g4len * BITS_UCHAR;
#endif
    }

#ifdef  VALPRINT
    {
      Dgp_Int32 x, y;
      Dgp_Bool white;
      Dgp_Int32 runstart;

      for (y = 0; y < height; y++) {
    	printf("B ");

	    white = DGP_TRUE;
    	runstart = 0;
	    for (x = 0; x < classwidth; x++) {
    	  if (classbits[y * classrow_width + x / BITS_ULONG] & (1ul << (BITS_ULONG - 1 - (x % BITS_ULONG)))) {
	        if (white) {
    	      printf("%d ", x - runstart);
	          white = DGP_FALSE;
	          runstart = x;
    	    }
	      } else {
    	    if (!white) {
	          printf("%d ", x - runstart);
	          white = DGP_TRUE;
    	      runstart = x;
	        }
    	  }
        }
	    printf("%d\n", x - runstart);
      }
    }
#endif

  /* Now we just need to read out all the tokens. */
  bitoffset = 0;
  if ((classstart == lastmodelnum - 1) && (tokens[classstart].uses == 0)) {
    /* Duplicate the G4 data and set the special flags. */
    tokens[classstart].longsperline = 0;
    tokens[classstart].uses = g4len;
    tokens[classstart].extdata = (void *)g4len;
    DGP_CHECK_ALLOC_CLEANUP(tokens[classstart].bits = (a->malloc)((size_t)g4len));
    memcpy((void *)tokens[classstart].bits, (void *)(pf->inbuf - g4len),
           g4len);
  } else {
    for (i = classstart; i < lastmodelnum; i++) {
      /* Use the allocator they gave us for the bitmaps. */
      if (lastmodelnum == classstart + 1) {
        /* There's only one model; no need to copy. */
        DGP_ASSERT(tokens[i].longsperline == classrow_width);
        DGP_ASSERT(classwidth == tokens[i].width);
        
        tokens[i].extdata = (void *)(tokens[i].height * tokens[i].longsperline);
        tokens[i].bits = classbits;
        
        classbits = (Dgp_UInt32 *)NULL;
        bitoffset = classwidth;
      } else {
        DGP_CHECK_ALLOC_CLEANUP(tokens[i].bits = (Dgp_UInt32 *)(a->malloc)((unsigned)height * 
                                                 tokens[i].longsperline * sizeof(Dgp_UInt32)));
        
        tokens[i].extdata = (void *)(tokens[i].height * tokens[i].longsperline);
        readoutModel(classbits, bitoffset, classwidth, height,
                     tokens[i].bits, tokens[i].width);
        bitoffset += tokens[i].width;
        
      }
    }
    DGP_ASSERT(bitoffset == classwidth);
    
    /* Done with this height class. */
    if (classbits != (Dgp_UInt32 *)NULL) {
      /* We didn't steal it... */
      (a->free)((void *)classbits);
      classbits = (Dgp_UInt32 *)NULL;
    }
  }

  NORMAL_RETURN;

ERR_CLEANUP:

  if (classbits != (Dgp_UInt32 *)NULL) {
    (a->free)((void *)classbits);
  }

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode tcmp_decodeTokDict(token *tokens, void *buffer, size_t buflen, Dgp_State *dgp_state)
**
**
**  This routine decodes the stuff in buffer, writing the results to *tokens
**  (which should have the correct number of entries). It allocates the memory
**  for the bitmaps using allocator. It returns
**  DGP_TRUE if it was successful, DGP_FALSE if it failed.
**
**  For some single-use tokens, it might return the G4-compressed bitmap in
**  bits, the bitmap length in uses, and flag this special case with
**  longsperline==0.
**
**  At the moment, it doesn't do any of the PointerFixup magic. Maybe it should
**  eventually.
**
** 
**	Arguments:
**		o  tokens...
**		o  buffer...
**      o  buflen...
**      o  dgp_state points to a prepared Dgp_State{}.
**
**	Return value:
**		o  ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

ErrCode tcmp_decodeTokDict(token *tokens, void *buffer, size_t buflen, Dgp_State *dgp_state) {
  DGP_PROLOGUE("tcmp_decodeTokDictMUsage()", DGP_PERRSTATE);
  Dgp_UInt32        ntokens;
  bitfile           f;
  Dgp_Int32         oldheight;
  Dgp_Bool          end;
  Dgp_Int32         classwidth;
  Dgp_Int32         modelnum, oldmodelnum;
  Dgp_Int32         useEncoding;
  Dgp_UInt32        tmp;

  /* G4 Decoder Cookie. */
  void *g4Decoder = (void *)NULL;

#ifdef  INSTRUMENT
  dict_bituses bituses;
  bituses.overhead = bituses.padding = bituses.bitmapsize = bituses.bitmaps =
    bituses.height = bituses.width = bituses.uses = 0;
#endif
    
  /* Initialize the encoder. */
  DGP_ERR_PASS_CLEANUP(DecodeCreate(&g4Decoder, dgp_state, NULL));

  f.inbuf = buffer;
  f.inbufleft = buflen;
  f.bitsinbuf = 0;
  f.buf = 0;
    
#ifdef  INSTRUMENT
  TCMP_READ_CHECK(!readValue(&ntokens, &f , &bituses.overhead));
#else
  TCMP_READ_CHECK(!readValue(&ntokens, &f));
#endif
#ifdef  VALPRINT
  printf("NS %d\n", ntokens);
#endif

  /*
   * Read out the use-count encoding.
   */
  TCMP_READ_CHECK(!readCode(&tmp, 1, &f));
#ifdef  INSTRUMENT
  bituses.overhead++;
#endif
#ifdef  VALPRINT
  printf("UC %ld\n", tmp);
#endif
  useEncoding = tmp;

  oldheight = 0;
  modelnum = 0;

  for (;;) {
    oldmodelnum = modelnum; 

#ifdef  INSTRUMENT
DGP_ERR_PASS_CLEANUP(
    readHCTokens( tokens, &modelnum, &oldheight, useEncoding, &f, &classwidth,
                  &end, dgp_state, &bituses));
#else
DGP_ERR_PASS_CLEANUP(
    readHCTokens( tokens, &modelnum, &oldheight, useEncoding, &f, &classwidth,
                  &end, dgp_state));
#endif

    if( end) {
      break;
    }

#ifdef  INSTRUMENT
DGP_ERR_PASS_CLEANUP(
    readHCBitmaps( tokens, oldmodelnum, modelnum, oldheight, classwidth, &f, 
                   g4Decoder, dgp_state, &bituses));
#else
DGP_ERR_PASS_CLEANUP(
    readHCBitmaps( tokens, oldmodelnum, modelnum, oldheight, classwidth, &f, 
                   g4Decoder, dgp_state));
#endif
  }
    
#ifdef  INSTRUMENT
  fprintf(stderr, "Bit allocations for dictionary:\t%d overhead\n\t\t\t\t%d padding\n\t\t\t\t%d bitmap size\n\t\t\t\t%d bitmaps\n\t\t\t\t%d height class height\n\t\t\t\t%d token width\n\t\t\t\t%d token uses\n",
	  bituses.overhead, bituses.padding, bituses.bitmapsize, bituses.bitmaps, bituses.height, bituses.width, bituses.uses);
#endif

  /* Done with this... */
  DecodeCleanUp( &g4Decoder);

  NORMAL_RETURN;

ERR_CLEANUP:
  
  if( g4Decoder != (void *)NULL) {
    DecodeCleanUp( &g4Decoder);
  }

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode tcmp_decodeTokDictMUsage(token          *tokens, 
**                                              void           *buffer, 
**                                              size_t         buflen,
**                                              Dgp_State      *dgp_state, 
**                                              Dgp_Int32      *tok_usage)
**
**
**  This routine decodes the stuff in buffer, writing the results to *tokens
**  (which should have the correct number of entries). It allocates the memory
**  for the bitmaps using allocator. It returns
**  DGP_TRUE if it was successful, DGP_FALSE if it failed.
**
**  For some single-use tokens, it might return the G4-compressed bitmap in
**  bits, the bitmap length in uses, and flag this special case with
**  longsperline==0.
**
**  At the moment, it doesn't do any of the PointerFixup magic. Maybe it should
**  eventually.
**
** 
**	Arguments:
**		o  tokens...
**		o  buffer...
**      o  buflen...
**      o  dgp_state points to a prepared Dgp_State{}.
**      o  tok_usage...
**
**	Return value:
**		o  ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

ErrCode
tcmp_decodeTokDictMUsage(token          *tokens, 
                         void           *buffer, 
                         size_t         buflen,
                         Dgp_State      *dgp_state, 
                         Dgp_Int32      *tok_usage) {
  Dgp_Int32 usage;
  Dgp_Int32 ntokens;

DGP_PROLOGUE( "tcmp_decodeTokDictMUsage()", DGP_PERRSTATE);

DGP_ERR_PASS(tcmp_decodeTokDict( tokens, buffer, buflen, dgp_state));

DGP_ERR_PASS(tcmp_getTokenCount( buffer, buflen, dgp_state, (size_t *)&ntokens));

  usage = ntokens * sizeof( token); 
  usage += tcmp_estimateTokDictBitmapUsage( tokens, ntokens);

  *tok_usage = usage;

  NORMAL_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode tcmp_decodeTokDictTokens(token      *tokens, 
**                                              void       *buffer, 
**                                              size_t     buflen,
**                                              Dgp_State  *dgp_state)
**
**  ??
**
**	Arguments:
**		o  tokens...
**		o  buffer...
**      o  buflen...
**      o  dgp_state points to a prepared Dgp_State{}.
**
**	Return value:
**		o  ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

ErrCode
tcmp_decodeTokDictTokens(token      *tokens, 
                         void       *buffer, 
                         size_t     buflen,
                         Dgp_State  *dgp_state) {
  Dgp_UInt32        ntokens;
  bitfile           f;
  Dgp_Int32         oldheight;
  Dgp_Int32         classwidth;
  Dgp_UInt32        v;
  Dgp_Bool          end;  
  Dgp_Int32         modelnum, oldmodelnum;
  Dgp_Int32         useEncoding;
  Dgp_UInt32        tmp;

#ifdef  INSTRUMENT
  dict_bituses bituses;
#endif

DGP_PROLOGUE( "tcmp_decodeTokDictTokens()", DGP_PERRSTATE);

#ifdef  INSTRUMENT
  bituses.overhead = bituses.padding = bituses.bitmapsize = bituses.bitmaps =
    bituses.height = bituses.width = bituses.uses = 0;
#endif

DGP_ASSERT( sizeof( Dgp_UInt32 *) == sizeof( Dgp_Int32));
DGP_ASSERT( sizeof( void *) == sizeof( size_t));

  f.inbuf = buffer;
  f.inbufleft = buflen;
  f.bitsinbuf = 0;
  f.buf = 0;

#ifdef  INSTRUMENT
  TCMP_READ_CHECK(!readValue(&ntokens, &f , &bituses.overhead));
#else
  TCMP_READ_CHECK(!readValue(&ntokens, &f));
#endif
#ifdef  VALPRINT
  printf("NS %d\n", ntokens);
#endif

  /*
   * Read out the use-count encoding.
   */
  TCMP_READ_CHECK(!readCode(&tmp, 1, &f));
#ifdef  INSTRUMENT
  bituses.overhead++;
#endif
#ifdef  VALPRINT
  printf("UC %ld\n", tmp);
#endif
  useEncoding = tmp;

  oldheight = 0;
  modelnum = 0;

  for (;;) {

    oldmodelnum = modelnum; 

#ifdef  INSTRUMENT
DGP_ERR_PASS_CLEANUP(
    readHCTokens( tokens, &modelnum, &oldheight, useEncoding, &f, &classwidth,
                  &end, dgp_state, &bituses));
#else
DGP_ERR_PASS_CLEANUP(
    readHCTokens( tokens, &modelnum, &oldheight, useEncoding, &f, &classwidth,
                  &end, dgp_state));
#endif

    if( end) {
      break;
    }

    /* Save the length of this class and the number of tokens in it. */
    tokens[ oldmodelnum].bits = (Dgp_UInt32 *)modelnum;

    /* We only need to keep track of the total width of the height class
     * if there is more than one token in the class. */
    if( modelnum > oldmodelnum + 1) {
      tokens[ oldmodelnum + 1].bits = (Dgp_UInt32 *)classwidth;
    }
    /* Save the offset of the image size within the buffer. */
    tokens[ oldmodelnum].extdata = (void *)((f.inbufleft * BITS_UCHAR)
                                            + f.bitsinbuf);
    
    /* Now read the size of the compressed image for this height class. */
#ifdef  INSTRUMENT
    TCMP_READ_CHECK(!readValue(&v, &f, &bituses.bitmapsize));
#else
    TCMP_READ_CHECK(!readValue(&v, &f));
#endif
#ifdef  VALPRINT
    printf("BL %d\n", v);
#endif

    /* Skip over the bitmap/g4 data, on to the next height class. */
    f.bitsinbuf = 0;
    f.buf = 0;  
    if( v == 0) {
      Dgp_Int32 buflen = oldheight * ((classwidth + BITS_UCHAR - 1) / BITS_UCHAR);
      f.inbuf += buflen;
      f.inbufleft -= buflen;
    }
    else {
      f.inbuf += v;
      f.inbufleft -= v;
    }
  }

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  Dgp_UInt32 tcmp_estimateTokDictBitmapUsage( token *tokens, Dgp_Int32 ntokens)
**
**
**  Returns the number of bytes (words? / dhtesq.) which will be allocated to make room for the 
**  bitmaps associated with tokens in this dictionary. 
**
**	Arguments:
**		o  tokens is the array whose bitmaps you'll want allocated.
**		o  ntokens is length(tokens).
**
**	Return value:
**		o  Number of bytes (words? / dhtesq.) required for bitmaps for passed array of tokens,
**         as an unsigned long.
**
**	Modification history:
**      o  None.
**  
**/

Dgp_UInt32 
tcmp_estimateTokDictBitmapUsage( token *tokens, Dgp_Int32 ntokens) {
  Dgp_Int32     i;
  Dgp_UInt32    total = 0, current;

  for( i = 0; i < ntokens; i++) {
    if( tokens[ i].longsperline == 0) {
      /* Bitmap will be stored as G4 data. */
      current = tokens[ i].uses;
    }
    else {
      current = tokens[ i].height * tokens[ i].longsperline;
    }
    total += current;
  }
  return total;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode tcmp_decodeTokDictBitmaps(token         *tokens, 
**                                               Dgp_Int32     ntokens, 
**                                               void          *buffer,
**                                               size_t        buflen, 
**                                               Dgp_State     *dgp_state)
**
**
**  Given a set of tokens with valid height, width, use, etc. fields, read
**  the dictionary image data from the file. The function mallocs memory
**  using dgp_state.
**
**
**	Arguments:
**		o  tokens is an array of tokens whose bitmaps are to be decoded.
**		o  ntokens is length(tokens)?
**      o  buffer is the destination for the image data?
**      o  buflen is length(buffer)?
**      o  dgp_state points to a prepared Dgp_State{}.
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

ErrCode
tcmp_decodeTokDictBitmaps(token         *tokens, 
                          Dgp_Int32     ntokens, 
                          void          *buffer,
                          size_t        buflen, 
                          Dgp_State     *dgp_state) {
  bitfile   f;
  Dgp_Int32 modelnum, oldmodelnum = 0, classwidth;
  size_t    bitsleft;

  /* G4 Decoder Cookie. */
  void *g4Decoder = (void *)NULL;

#ifdef  INSTRUMENT
  dict_bituses bituses;
#endif

DGP_PROLOGUE( "tcmp_decodeTokDictBitmaps()", DGP_PERRSTATE);

#ifdef  INSTRUMENT
  bituses.overhead = bituses.padding = bituses.bitmapsize = bituses.bitmaps =
    bituses.height = bituses.width = bituses.uses = 0;
#endif

  /* Initialize the encoder. */
DGP_ERR_PASS_CLEANUP( DecodeCreate( &g4Decoder, dgp_state, NULL));

  /* As we go, NULL out any pointers that we were sneakily using for any
   * other purposes. This makes cleaning up after ourselves much easier. */
  
  while( oldmodelnum < ntokens) {

    /* We've stored the index of the first token of next height class here. */
    modelnum = (Dgp_Int32)tokens[ oldmodelnum].bits;
    tokens[ oldmodelnum].bits = (Dgp_UInt32 *)NULL;
    
    /* If there is only one token in this height class, then the classwidth
     * is easy. Otherwise, it's stored in the second token in the class. */
    if( modelnum == oldmodelnum + 1) {
      classwidth = tokens[ oldmodelnum].width;
    } else {
      classwidth = (Dgp_Int32)tokens[ oldmodelnum + 1].bits;
      tokens[ oldmodelnum + 1].bits = (Dgp_UInt32 *)NULL;
    }

    /* Skip to the offset of the compressed image for this height class. */
    bitsleft = (size_t)tokens[ oldmodelnum].extdata;
    tokens[ oldmodelnum].extdata = (void *)NULL;

    f.inbuf = (Dgp_UInt8 *)buffer + (buflen - ((bitsleft + BITS_UCHAR - 1) / BITS_UCHAR));
    f.inbufleft = (bitsleft + BITS_UCHAR - 1) / BITS_UCHAR;
    if( bitsleft % BITS_UCHAR > 0) {
      f.buf = *(f.inbuf);
      f.bitsinbuf = bitsleft % BITS_UCHAR;      
      f.inbuf++;
      f.inbufleft--;
    } else {
      f.bitsinbuf = 0;
    }

#ifdef  INSTRUMENT
DGP_ERR_PASS_CLEANUP(
    readHCBitmaps( tokens, oldmodelnum, modelnum, 
                   tokens[ oldmodelnum].height, classwidth, &f, 
                   g4Decoder, dgp_state, &bituses));
#else
DGP_ERR_PASS_CLEANUP(
    readHCBitmaps( tokens, oldmodelnum, modelnum, 
                   tokens[ oldmodelnum].height, classwidth, &f, 
                   g4Decoder, dgp_state));
#endif

    oldmodelnum = modelnum; 
  }

  /* Done with this... */
  DecodeCleanUp( &g4Decoder);

  NORMAL_RETURN;

ERR_CLEANUP:

  if( g4Decoder != (void *)NULL) {
    DecodeCleanUp( &g4Decoder);
  } 

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode tcmp_getTokPosCount( void        *buffer, 
**                                          size_t      buflen, 
**                                          Dgp_State   *dgp_state, 
**                                          size_t      *count)
**
**
**  This routine takes an encoded token position block and returns the number
**  of positions encoded in it.
**
**	Arguments:
**		o  buffer...
**      o  buflen is length(buffer)?
**      o  dgp_state points to a prepared Dgp_State{}.
**		o  count...
**
**	Return value:
**		o  An ErrCode.
**
**	Modification history:
**      o  None.
**  
**/

ErrCode
tcmp_getTokPosCount(void *buffer, size_t buflen, Dgp_State *dgp_state, size_t *count) {
  DGP_PROLOGUE("tcmp_getTokPosCount()", DGP_PERRSTATE);
  bitfile f;
#ifdef  INSTRUMENT
  Dgp_Int32 dummy = 0;
#endif

  f.inbuf = buffer;
  f.inbufleft = buflen;
  f.bitsinbuf = 0;
  f.buf = 0;
    
#ifdef  INSTRUMENT
  TCMP_READ_CHECK(!readValue((Dgp_UInt32 *)count, &f , &dummy));
#else
  TCMP_READ_CHECK(!readValue((Dgp_UInt32 *)count, &f));
#endif


  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  ErrCode tcmp_decodeTokPositions(tokpos      *positions,
**                                             Dgp_Int32   lowx,
**                                             Dgp_Int32   lowy, 
**                                             Dgp_Int32   highx,
**                                             Dgp_Int32   highy,
**                                             void        *buffer,
**                                             size_t      buflen, 
**                                             void        *dstate, 
**                                             Dgp_State   *dgp_state)
**
**
**  This routine decodes the stuff in buffer and writes the results to
**  *positions. It returns an ErrCode on failure. The page
**  coordinates are as in encodeTokPositions above; some higher layer of the
**  file needs to keep track of them.
**
**	Arguments:
**		o  positions is the destination for the decoded data
**      o  lowx, lowy...
**      o  highx, highy...
**      o  buffer points to the source (encoded) positions.
**      o  buflen is length(buffer).
**      o  dstate...
**      o  dgp_state points to a prepared Dgp_State{}.
**
**	Return value:
**		o  An ErrCode;  contains assertions.
**
**	Modification history:
**      o  None.
**  
**/

ErrCode
tcmp_decodeTokPositions(tokpos      *positions,
                        Dgp_Int32   lowx,
                        Dgp_Int32   lowy, 
                        Dgp_Int32   highx,
                        Dgp_Int32   highy,
                        void        *buffer,
                        size_t      buflen, 
                        void        *dstate, 
                        Dgp_State   *dgp_state) {
  DGP_PROLOGUE("tcmp_decodeTokPositions()", DGP_PERRSTATE);

  bitfile           f;
  Dgp_Int32         i;
  Dgp_UInt32        code;
  Dgp_Int32         codelen;
  Dgp_Int32         mincode;
  Dgp_Int32         lastx, lasty, prevfirstx;
  Dgp_Int32         cury;
  Dgp_UInt32        ydelta;
  Dgp_Int32         xdelta;
  Dgp_Bool          end;
  Dgp_Int32         boxnum;
  Dgp_Int32         numBoxes;
  Dgp_UInt32        *firstcodes;
  Dgp_Int32         **codeindex;
  dictstate         *state = (dictstate *)dstate;
  tokstate          *t;
  Dgp_Int32         modex;
  Dgp_Int32         stripsize, logstripsize = 0;
  Dgp_Int32         firstXEncoding, deltaXEncoding, deltaYEncoding;
  Dgp_UInt32        tmp;
  Dgp_UInt32        iv;
  Dgp_Bool          transpose;
#ifdef  INSTRUMENT
  pos_bituses       bituses;
  bituses.overhead = bituses.padding = bituses.firstx = bituses.deltax =
    bituses.deltay = bituses.y = bituses.toknum = 0;
#endif

  f.inbuf = buffer;
  f.inbufleft = buflen;
  f.bitsinbuf = 0;
  f.buf = 0;
    
  /* Special case: state is NULL */
  if (state == (dictstate *)NULL) {
    /* There had better not be any positions in this block. */
    /* How many positions are there? */
#ifdef  INSTRUMENT
    TCMP_READ_CHECK(!readValue(&numBoxes, &f , &bituses.overhead));
#else
    TCMP_READ_CHECK(!readValue(&numBoxes, &f));
#endif
    DGP_ASSERT(numBoxes == 0);
#ifdef  VALPRINT
    printf("NP %d\n", numBoxes);
#endif
    /* No point in reading anything else. */
    NORMAL_RETURN;
  }

  /* Look up the decoding tables, making sure they're there first. */
  DGP_ERR_PASS_CLEANUP(tcmpi_ensureHuffmanCodes(state, dgp_state));

  firstcodes = state->firstcodes;
  codeindex = state->codeindex;

  /* What's the shortest code length actually used? */
  mincode = DGP_MAX_CODELEN;
  for (i = 0; i < state->ntok; i++) {
    mincode = MIN(mincode, state->toks[i].codelen);
  }

  /* Start reading */

  /* How many positions are there? */
#ifdef  INSTRUMENT
  TCMP_READ_CHECK(!readValue(&numBoxes, &f , &bituses.overhead));
#else
  TCMP_READ_CHECK(!readValue(&numBoxes, &f));
#endif
#ifdef  VALPRINT
  printf("NP %d\n", numBoxes);
#endif

  /* Read out some stuff to do with the exact encoding */

  /* 4 bits of modal x value */
  TCMP_READ_CHECK(!readCode(&tmp, 4, &f));
  modex = tmp;
#ifdef  INSTRUMENT
  bituses.overhead += 4;
#endif
#ifdef  VALPRINT
  printf("MX %d\n", modex);
#endif

  /* 2 bits of (strip height - 1) */
  TCMP_READ_CHECK(!readCode(&tmp, 2, &f));
  stripsize = tmp + 1;
#ifdef  INSTRUMENT
  bituses.overhead += 2;
#endif
#ifdef  VALPRINT
  printf("SS %d\n", stripsize);
#endif
  if (stripsize == 1) {
    logstripsize = 0;
  }
  else if (stripsize == 2) {
    logstripsize = 1;
  }
  else if (stripsize == 4) {
    logstripsize = 2;
  }
  else {
    DGP_ASSERT((stripsize == 1) || (stripsize == 2) || (stripsize == 4));
  }

  /* 2 bits each of encoder info */
  TCMP_READ_CHECK(!readCode(&tmp, 2, &f));
  firstXEncoding = tmp;
#ifdef  INSTRUMENT
  bituses.overhead += 2;
#endif
#ifdef  VALPRINT
  printf("FXC %ld\n", tmp);
#endif

  TCMP_READ_CHECK(!readCode(&tmp, 2, &f));
  deltaXEncoding = tmp;
#ifdef  INSTRUMENT
  bituses.overhead += 2;
#endif
#ifdef  VALPRINT
  printf("DXC %ld\n", tmp);
#endif

  TCMP_READ_CHECK(!readCode(&tmp, 2, &f));
  deltaYEncoding = tmp;
#ifdef  INSTRUMENT
  bituses.overhead += 2;
#endif
#ifdef  VALPRINT
  printf("DYC %ld\n", tmp);
#endif

  /* One bit of "transposed?" */
  TCMP_READ_CHECK(!readCode(&tmp, 1, &f));
  if (tmp != 0) {
    transpose = DGP_TRUE;
  }
  else {
    transpose = DGP_FALSE;
  }
#ifdef  INSTRUMENT
  bituses.overhead++;
#endif
#ifdef  VALPRINT
  printf("TRC %ld\n", tmp);
#endif

  /* Figure out where the imaginary strip above the top is */
  lasty = ((-1 >> logstripsize) << logstripsize);
  lastx = prevfirstx = 0;
  boxnum = 0;
  for (;;) {
#ifdef  INSTRUMENT
    TCMP_READ_CHECK(!readYDelta(&ydelta, &f, deltaYEncoding, &bituses.deltay));
#else
    TCMP_READ_CHECK(!readYDelta(&ydelta, &f, deltaYEncoding));
#endif

    lasty += (ydelta << logstripsize);

    if (lasty >= (transpose ? highx : highy)) {
      break;
    }

    /* Read the first token */
#ifdef  INSTRUMENT
    TCMP_READ_CHECK(!readFirstX(&iv, &f, firstXEncoding , &bituses.firstx));
#else
    TCMP_READ_CHECK(!readFirstX(&iv, &f, firstXEncoding));
#endif
    lastx = iv + prevfirstx;
    prevfirstx = lastx;

    /* Read y position if there should be one */
    if (logstripsize) {
#ifdef  INSTRUMENT
      TCMP_READ_CHECK(!readYVal(&iv, &f, logstripsize , &bituses.y));
#else
      TCMP_READ_CHECK(!readYVal(&iv, &f, logstripsize));
#endif
      cury = iv;
    }
    else {
      cury = 0;
    }

#ifdef  INSTRUMENT
    TCMP_READ_CHECK(!readTokenCode(&code, &codelen, mincode, firstcodes, &f,
				   &bituses.toknum));
#else
    TCMP_READ_CHECK(!readTokenCode(&code, &codelen, mincode, firstcodes, &f));
#endif
  
    /* Find out the info of the token referred to here. */
    t = &(state->toks[codeindex[codelen][code - firstcodes[codelen]]]);
    DGP_ASSERT(t->codelen == codelen);
    DGP_ASSERT(t->code == code);
    positions[boxnum].tok = t->tok;
    positions[boxnum].extdata = (void *)NULL;
  
#ifdef  VALPRINT
    /* printf("SN %d\n", t->toknum); */
#endif

    if (transpose) {
      /* The upper left coordinate was stored as (Y,X) */
      positions[boxnum].xpos = (Dgp_Int16)(lasty + cury + lowx);
      positions[boxnum].ypos = (Dgp_Int16)(lastx + lowy + t->tok->height - 1);
      lastx += t->tok->height;
    }
    else {
      /* The lower left coordinate was stored as (X,Y) */
      positions[boxnum].xpos = (Dgp_Int16)(lastx + lowx);
      positions[boxnum].ypos = (Dgp_Int16)(lasty + cury + lowy);
      lastx += t->tok->width;
    }

    boxnum++;

    for (;;) {
#ifdef  INSTRUMENT
      TCMP_READ_CHECK(!readXDelta(&xdelta, &end, &f, deltaXEncoding,
				  &bituses.deltax));
#else
      TCMP_READ_CHECK(!readXDelta(&xdelta, &end, &f, deltaXEncoding));
#endif
      xdelta += modex;
      lastx += xdelta;

      if (end) {
	break;
      }

      /* Read y position if there should be one */
      if (logstripsize) {
#ifdef  INSTRUMENT
	TCMP_READ_CHECK(!readYVal(&iv, &f, logstripsize , &bituses.y));
#else
	TCMP_READ_CHECK(!readYVal(&iv, &f, logstripsize));
#endif
	cury = iv;
      }
      else {
	cury = 0;
      }

      /* Read a token code. */

#ifdef  INSTRUMENT
      TCMP_READ_CHECK(!readTokenCode(&code, &codelen, mincode, firstcodes, &f,
				     &bituses.toknum));
#else
      TCMP_READ_CHECK(!readTokenCode(&code, &codelen, mincode, firstcodes,&f));
#endif
      
      /* Find out the info of the token referred to here. */
      t = &(state->toks[codeindex[codelen][code - firstcodes[codelen]]]);
      DGP_ASSERT(t->codelen == codelen);
      DGP_ASSERT(t->code == code);
      positions[boxnum].tok = t->tok;
      positions[boxnum].extdata = (void *)NULL;
#ifdef  VALPRINT
      /*      printf("SN %d\n", t->toknum); */
#endif

      if (transpose) {
	/* The upper left coordinate was stored as (Y,X) */
	positions[boxnum].xpos = (Dgp_Int16)(lasty + cury + lowx);
	positions[boxnum].ypos = (Dgp_Int16)(lastx + lowy + t->tok->height - 1);
	lastx += t->tok->height;
      }
      else {
	/* The lower left coordinate was stored as (X,Y) */
	positions[boxnum].xpos = (Dgp_Int16)(lastx + lowx);
	positions[boxnum].ypos = (Dgp_Int16)(lasty + cury + lowy);
	lastx += t->tok->width;
      }

      boxnum++;
    }
  }
  DGP_ASSERT(boxnum == numBoxes);

  /* Read (and skip) the dictionary extension section */

  /* extensionlen is not here anymore */
#if 0
#ifdef  INSTRUMENT
  TCMP_READ_CHECK(!readValue(&iv, &f , &bituses.overhead));
#else
  TCMP_READ_CHECK(!readValue(&iv, &f));
#endif
#ifdef  INSTRUMENT
  bituses.padding += f.bitsinbuf;
#endif

#ifdef  VALPRINT
  printf("PE %d\n", iv);
#endif
  if (extensionlen != (size_t *)NULL) {
    *extensionlen = iv;
    *extension = f.inbuf;
  }
#endif

#ifdef  INSTRUMENT
  fprintf(stderr, "Bit allocations for positions:\t%d overhead\n\t\t\t\t%d padding\n\t\t\t\t%d first X\n\t\t\t\t%d delta X\n\t\t\t\t%d delta Y\n\t\t\t\t%d Y\n\t\t\t\t%d token numbers\n",
	  bituses.overhead, bituses.padding, bituses.firstx, bituses.deltax, bituses.deltay, bituses.y, bituses.toknum);
#endif

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}
/*************************************************************************/



