
#include "image_defs.h"
#include <winsock2.h>
#include <stdio.h>
#include "santa.h"
#include <santa_page.h>
#include <image.h>
#include <windef.h>
#include <mmsystem.h>
#include "winprint.h"
#include "santa_t.h"
#include "orientation.h"

typedef struct {
    Dgp_Int32 planes;
    Dgp_Int32 bpp;
} BmpColorType;

/* color type referenced by santaPageType */
static BmpColorType colorTypes[2] = {{1,1},{3, 8}};

ErrCode
doRender(santaDoc *pdoc,
         Dgp_Int32 pgNum,
         Dgp_Int8 orient,
         Dgp_Int32 *PgType,
         void      **pPgImage,
         Dgp_Int32 *pgWidth,
         Dgp_Int32 *pgHeight,
         Dgp_Int32 *pgXRes,
         Dgp_Int32 *pgYRes,
         Dgp_State *dgp_state);

#define READ_BUFFER_SIZE 4096


ErrCode
printRAW(Dgp_Int8 *inFile, 
         Dgp_Int8 *prtName, 
         Dgp_Int8 *portName, 
         Dgp_Bool report, 
         Dgp_State *dgp_state)
{
    DGP_PROLOGUE("printRAW()", DGP_PERRSTATE);
    DOC_INFO_1  DocInfo;
    HANDLE hPrinter = NULL;
    Dgp_Int32 cbRead, cbWritten;
    BYTE pReadBuffer[READ_BUFFER_SIZE];
    HANDLE hFile = NULL;
    BOOL readRet;
    Dgp_Int32 prn_status = STATUS_NONE;
    
    
    DGP_W32_ERR_GEN_CLEANUP(!OpenPrinter(prtName, &hPrinter, NULL));
    
    DocInfo.pDocName    = inFile;     /* Document name */
                                      /* If we are using this as a pass through
                                      type function we don't know the the output file for the 
                                      spooler.  So we might need to write this as a processor
    but hopefully we can get away with this */
    DocInfo.pOutputFile = NULL;   
    DocInfo.pDatatype   = "RAW";     /* Document data type */
    
    /* Let the printer know we are starting a new document */
    DGP_W32_ERR_GEN_CLEANUP(!StartDocPrinter(hPrinter, 1, (LPBYTE)&DocInfo));
    prn_status = STATUS_PAGE;
    
    
    DGP_W32_ERR_GEN_CLEANUP((hFile = CreateFile(	inFile, 
        GENERIC_READ, 
        FILE_SHARE_READ, 
        NULL, 
        OPEN_EXISTING,
        FILE_ATTRIBUTE_NORMAL, 
        NULL)) == NULL);
    
    while((readRet = ReadFile(hFile, pReadBuffer, READ_BUFFER_SIZE, &cbRead, NULL)) &&
        cbRead) {
        
        /*Need to open file and read and write chunks */
        DGP_W32_ERR_GEN_CLEANUP(!WritePrinter(hPrinter, pReadBuffer, cbRead, &cbWritten));
    }
    
    DGP_W32_ERR_GEN_CLEANUP(readRet == 0);
    
    CloseHandle(hFile);
    hFile = NULL;
    
    EndDocPrinter(hPrinter);
    prn_status = STATUS_NONE;
    
    ClosePrinter(hPrinter);
    hPrinter = NULL;
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    
    if(hFile != NULL) CloseHandle(hFile);
    
    if(hPrinter != NULL) {
        if(prn_status == STATUS_DOC) EndDocPrinter(hPrinter);
        ClosePrinter(hPrinter);
    }
    
    ERR_RETURN;
}


/* we don't let the user aboard printing. hahahah evil. very evil */
BOOL CALLBACK AbortProc(HDC hdc, Dgp_Int32 nCode) {
    return DGP_TRUE;
}

ErrCode
PrintDialog(Dgp_State *dgp_state, 
            HDC *hDC, 
            Dgp_Int32 *firstPage, 
            Dgp_Int32 *lastPage,
            Dgp_Bool *canceled) 
{
    DGP_PROLOGUE("PrintDialog()", DGP_PERRSTATE);
    
    PRINTDLG pd;
    DWORD dlg_err;
    
    
    /* Initialize the PRINTDLG members.   */
    pd.lStructSize = sizeof(PRINTDLG); 
    pd.hDevMode = (HANDLE) NULL; 
    pd.hDevNames = (HANDLE) NULL; 
    pd.Flags = PD_RETURNDC | PD_NOSELECTION; 
    pd.hwndOwner = NULL; /* hwnd;  */
    pd.hDC = (HDC) NULL; 
    pd.nFromPage =  *firstPage+1; 
    pd.nToPage = *lastPage+1;
    pd.nMinPage = *firstPage+1; 
    pd.nMaxPage = *lastPage+1; 
    pd.nCopies = 1; 
    pd.hInstance = (HANDLE) NULL; 
    pd.lCustData = 0L; 
    pd.lpfnPrintHook = (LPPRINTHOOKPROC) NULL; 
    pd.lpfnSetupHook = (LPSETUPHOOKPROC) NULL; 
    pd.lpPrintTemplateName = (LPSTR) NULL; 
    pd.lpSetupTemplateName = (LPSTR)  NULL; 
    pd.hPrintTemplate = (HANDLE) NULL; 
    pd.hSetupTemplate = (HANDLE) NULL;
    
    *canceled = DGP_FALSE;
    
    /* Display the PRINT dialog box.   */
    if(!PrintDlg(&pd)) {
        if((dlg_err = CommDlgExtendedError()) != 0) {
            DGP_ERR_CLEANUP(dlg_err + cDgpW32);
        } else {
            *canceled = DGP_TRUE;
            DGP_GOTO_CLEANUP;
        }
    }
    
    if (pd.Flags & PD_PAGENUMS) {
        *firstPage = pd.nFromPage - 1;
        *lastPage = pd.nToPage - 1;
    }
    *hDC = pd.hDC;
    pd.hDC = NULL;
    
    if(pd.hDevMode != NULL) GlobalFree(pd.hDevMode);
    if(pd.hDevNames != NULL) GlobalFree(pd.hDevNames);
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    
    if(pd.hDevMode != NULL) GlobalFree(pd.hDevMode);
    if(pd.hDevNames != NULL) GlobalFree(pd.hDevNames);
    if(pd.hDC != NULL) DeleteDC(pd.hDC);
    
    ERR_RETURN;
}


/* windows bitmaps have 'wrong' byte order and 1s and 0s are swaped. */
void
fixDgpImage_Packed(DgpImage_Packed im) {
    Dgp_Int32 width = DGPIMG_WIDTH(im);
    Dgp_Int32 height = DGPIMG_HEIGHT(im);
    Dgp_Int32 y;
    Dgp_UInt32 *ptr, *end, rev;
    Dgp_Int32 Dgp_Int32sperline = DGPIMG_GETROWWIDTH(im)/sizeof(Dgp_Int32);
    
    for (y = 0; y < height; y++) {
        for(ptr = DGPIMG_GETROW(im, y), end = ptr + Dgp_Int32sperline; ptr < end; ptr++) {
            rev = ~*ptr;
            *ptr = ((rev & 0xff) << 24) | ((rev & 0xff00) << 8) | ((rev & 0xff0000) >> 8) | ((rev & 0xff000000) >> 24);
        }
    }
}

/* windows bitmaps want bgr, not rgb */
void
fixDgpImage_ColorRGB(DgpImage_ColorRGB im) {
    Dgp_Int32 width = DGPIMG_WIDTH(im);
    Dgp_Int32 height = DGPIMG_HEIGHT(im);
    Dgp_Int32 y;
    DgpImage_RGB *ptr, *end;
    Dgp_UInt8 tmp;
    
    for (y = 0; y < height; y++) {
        for(ptr = (DgpImage_RGB *)DGPIMG_PIXPTR((DgpImage_ColorRGB)im, 0, y), end = ptr + width; ptr < end; ptr++) {
            tmp = ptr->r;
            ptr->r = ptr->b;
            ptr->b = tmp;
        }
    }
}


ErrCode
printDoc(
         Dgp_Int8 *docName, 
         uds *docuds, 
         Dgp_Int32 name_offset, 
         HDC _hDC, 
         Dgp_Int32 firstPage,
         Dgp_Int32 lastPage,
         Dgp_Bool report,
         Dgp_Bool createJob, 
         Dgp_State *dgp_state) 
         
{
    DGP_PROLOGUE("printDoc()", DGP_PERRSTATE);
    DOCINFO di;
    HDC hDC = NULL;
    Dgp_Int32 logPelsX, logPelsY;
    Dgp_Int32 dcWidth, dcHeight;
    santaDoc *pDoc = NULL;
    Dgp_Int32 i;
    santaPageType PgType;
    void *pPgImage = NULL;
    Dgp_Int32 pgWidth, pgHeight, pgXRes, pgYRes;
    Dgp_Int32 dstWidth, dstHeight;
    HBITMAP bitmap = NULL;
    void *hdcMem = NULL;
    Dgp_Int32 dc_status = STATUS_NONE;
    BITMAPINFOHEADER bmInfo = {sizeof(BITMAPINFOHEADER), 0, 0, 1, 0, BI_RGB, 0, 0, 0, 0, 0};
    Dgp_Float32 scalex, scaley;
    HGDIOBJ tmpObj;
    Dgp_Bool canceled = DGP_FALSE;
    
    
    /*open xiff document*/
    DGP_ERR_PASS_CLEANUPF1(santaOpenDocumentEx(	docuds, 
        1 << (DGP_BITS_UINT32 - 2),
        santaDocOptsDefault,
        dgp_state, 
        &pDoc),
        "error opening document: %s",docName);
    
    if(_hDC == NULL) {
        firstPage = 0;
        lastPage = pDoc->npages - 1;
        
        DGP_ERR_PASS_CLEANUP(PrintDialog(dgp_state, &hDC, &firstPage, &lastPage, &canceled));
        if(canceled) {
            DGP_REPORTF0(report, "user canceled print job");
            NORMAL_RETURN;
        }
    } else {
        hDC = _hDC;
        if(firstPage < 0 || firstPage > pDoc->npages -1) firstPage =0;
        if(lastPage > pDoc->npages - 1 || lastPage < 0) 
            lastPage = pDoc->npages - 1;
    }    
    
    DGP_W32_ERR_GEN_CLEANUPF0(!(GetDeviceCaps(hDC, RASTERCAPS) & RC_BITBLT),
        "printer cannot display bitmaps");
    
    /* retrieve the number of pixels-per-logical-inch in the  */
    /* horizontal and vertical directions of the printer */
    logPelsX = GetDeviceCaps(hDC, LOGPIXELSX); 
    logPelsY = GetDeviceCaps(hDC, LOGPIXELSY); 
    DGP_REPORTF2(report, "printer resolution: %d, %d", logPelsX, logPelsY);
    
    /* set pixels-per-meter in bitmap info */
    bmInfo.biXPelsPerMeter = (LONG)((logPelsX / (Dgp_Float32)2.54) * 1000);
    bmInfo.biYPelsPerMeter = (LONG)((logPelsY / (Dgp_Float32)2.54) * 1000);
    
    /* get size of printer page */
    dcWidth = GetDeviceCaps(hDC, HORZRES);
    dcHeight = GetDeviceCaps(hDC, VERTRES);
    DGP_REPORTF2(report, "printable area in pixels: %d, %d", dcWidth, dcHeight);
    
    /* register the application's AbortProc function with GDI.  */
    SetAbortProc(hDC, (ABORTPROC)AbortProc); 
    
    /* initialize the members of a DOCINFO structure.   */
    di.cbSize = sizeof(DOCINFO);     
    di.lpszDocName = docName + name_offset; 
    di.lpszOutput = (LPTSTR) NULL;     
    di.lpszDatatype = (LPTSTR) NULL; 
    di.fwType = 0;
    
    if( createJob == DGP_TRUE )
    {
        char vIn[] = "Xerox CEF Driver";

        if( ExtEscape(hDC, QUERYESCSUPPORT, sizeof(vIn), vIn, 0, NULL) )
        {
            ExtEscape( hDC, 0xdeadbeaf, strlen(docName), docName, 0, NULL );
            
            StartDoc(hDC, &di);
            EndDoc(hDC);

            goto SKIP_PRINT;
        }
    }

    /* start print job */
    if( createJob == DGP_TRUE )
    {
        DGP_W32_ERR_GEN_CLEANUPF0((StartDoc(hDC, &di) <= 0),
            "error starting document");
    }

    dc_status = STATUS_DOC;
    
    for (i = firstPage; i <= lastPage; i++) {
        
        /* start a new page */
        DGP_W32_ERR_GEN_CLEANUPF0((StartPage(hDC) <= 0), 
            "error starting page");
        dc_status = STATUS_PAGE;

        /* render page */
        /* TODO - check if page is too big memory-wise and do it in strips... */
        DGP_ERR_PASS_CLEANUPF1(doRender(	pDoc, 
            i, 
            ORIENT_RD,
            &PgType, 
            &pPgImage, 
            &pgWidth, 
            &pgHeight, 
            &pgXRes, 
            &pgYRes,
            dgp_state),
            "error rendering page: %d", i);
        
        /*   DgpImage_save((DgpImage_Packed)pPgImage, "temp.pnm"); */
        
        if(report) {
            DGP_REPORTF2(report, "printing page. width: %d, height: %d", pgWidth, pgHeight);
            DGP_REPORTF2(report, "xres: %d, yres: %d", pgXRes, pgYRes);
        }
        
        /* scale image down if necessary */
        dstWidth = (pgWidth*logPelsX)/pgXRes;
        dstHeight = (pgHeight*logPelsY)/pgYRes;
        
        if((dstWidth > dcWidth) || (dstHeight > dcHeight)) {
            scalex = dstWidth / (Dgp_Float32)dcWidth;
            scaley = dstHeight / (Dgp_Float32)dcHeight;
            
            if(scalex > scaley) {
                dstWidth = dcWidth;
                dstHeight = (Dgp_Int32)(dstHeight / scalex);
            } else {
                dstWidth = (Dgp_Int32)(dstWidth / scaley);
                dstHeight = dcHeight;
            }
        }
        
        /* do crazy windows stuff to blat it in the printer device content */
        switch(PgType) {
        case santaPageColor:
            
            /* need bgr, not rgb... */
            fixDgpImage_ColorRGB(pPgImage);
            
            if((pgWidth % sizeof(DWORD)) != 0) {
                DgpImage_ColorRGB pTmpImage = NULL;
                DgpImage_RGB *psrc, *pdst;
                Dgp_Int32 i;
                Dgp_Int32 newWidth = ((Dgp_Int32)((pgWidth + sizeof(DWORD) - 1) / sizeof(DWORD))) * sizeof(DWORD);
                
                
                DGP_ERR_PASS_CLEANUP(DGPIMAGE_NEW(dgp_state, DGPIMAGE_RGB, newWidth, pgHeight, 
                    &pTmpImage));
                
                for(i = 0; i < pgHeight; i++) {
                    psrc = DGPIMG_PIXPTR(((DgpImage_ColorRGB)pPgImage), 0, i);
                    pdst = DGPIMG_PIXPTR(pTmpImage, 0, i);
                    
                    memcpy(pdst, psrc, (sizeof(DgpImage_RGB) * pgWidth));
                    
                    /* clear the invalid bytes */
                    memset(pdst + pgWidth, 0xFF, (sizeof(DgpImage_RGB) * (newWidth - pgWidth)));
                }
                
                DGPIMAGE_FREE(((DgpImage_ColorRGB)pPgImage));
                pPgImage = pTmpImage;
                pTmpImage = NULL;
                pgWidth = newWidth;
            }
            /* initialize DIB header */
            bmInfo.biBitCount = (Dgp_UInt16)(colorTypes[santaPageColor-1].planes * colorTypes[santaPageColor-1].bpp);
            bmInfo.biWidth = pgWidth;
            bmInfo.biHeight = -(pgHeight);
            
            DGP_CHECK_NULL_CLEANUPF0((bitmap = CreateCompatibleBitmap(hDC, pgWidth, pgHeight)),
                "error creating color bitmap");
            
            DGP_W32_ERR_GEN_CLEANUP(SetDIBits(NULL, bitmap, 0, pgHeight, (void *)DGPIMG_PIXPTR((DgpImage_ColorRGB)pPgImage, 0, 0),
                (BITMAPINFO *)(&bmInfo), DIB_RGB_COLORS) == 0);
            break;
            
        case santaPageMono:
            
            /* windows bitmaps have 'wrong' byte order and 1s and 0s are swaped. */
            fixDgpImage_Packed(pPgImage);   
            
            DGP_CHECK_NULL_CLEANUPF0((bitmap = 
                CreateBitmap((( pgWidth + DGP_BITS_UINT32 - 1) / DGP_BITS_UINT32)*DGP_BITS_UINT32, 
                pgHeight, 
                colorTypes[santaPageMono-1].planes, 
                colorTypes[santaPageMono-1].planes, 
                DGPIMG_GETROW((DgpImage_Packed)pPgImage, 0))), "error creating mono bitmap");
            break;
            
        }
        
        DGP_W32_ERR_GEN_CLEANUP((hdcMem = CreateCompatibleDC(hDC)) == NULL);  
        DGP_W32_ERR_GEN_CLEANUPF0(((tmpObj = SelectObject(hdcMem, bitmap)) == NULL),
            "error selecting bitmap into device context");
        
        if((dstWidth != pgWidth) && (dstHeight != pgHeight)) {
            /* stretch and blat! */
            DGP_W32_ERR_GEN_CLEANUPF0(!StretchBlt(hDC, 0, 0, dstWidth, dstHeight,
                hdcMem, 0, 0, pgWidth, pgHeight, SRCCOPY),
                "error blatting image");
        } else {
            /* ooo, perfect fit! */
            DGP_W32_ERR_GEN_CLEANUPF0(!BitBlt(hDC, 0, 0, dstWidth, dstHeight,
                hdcMem, 0, 0, SRCCOPY),
                "error blatting image");
        }
        
        /* replace whatever we removed... */
        SelectObject(hdcMem, tmpObj);
        
        /* Delete the memory DC.  */
        DeleteDC(hdcMem);
        hdcMem = NULL;
        
        /* Delete bitmap */
        DeleteObject(bitmap);
        bitmap = NULL;
        
        /* Delete image */
        switch(PgType) {
        case santaPageColor:
            DGPIMAGE_FREE(((DgpImage_ColorRGB)pPgImage));
            break;
        case santaPageMono:
            DGPIMAGE_FREE(((DgpImage_Packed)pPgImage));
            break;
        }
        pPgImage = NULL;
        
        /* end page */
        DGP_W32_ERR_GEN_CLEANUPF0((EndPage(hDC) <= 0),
            "error ending page");
        dc_status = STATUS_DOC;
  }
  
  /* end job */
  if(createJob == DGP_TRUE){
      DGP_W32_ERR_GEN_CLEANUPF0((EndDoc(hDC) <= 0),
          "error ending document");
  }

SKIP_PRINT:

  dc_status = STATUS_NONE;
  
  
  /* close xiff document */
  DGP_ERR_PASS_CLEANUPF1(santaCloseDocumentEx(&pDoc,dgp_state),
      "error closing document: %s",docName);
  pDoc = NULL;
  
  DGP_REPORTF0(report, "done printing");
  
  /* should I free some things in the pd? */
  if(_hDC == NULL) DeleteDC(hDC);
  hDC = NULL;
  
  NORMAL_RETURN;
  
ERR_CLEANUP:
  
  if(hdcMem != NULL) DeleteDC(hdcMem);
  
  if(bitmap != NULL) DeleteObject(bitmap);
  
  if(pPgImage != NULL) {
      switch(PgType) {
      case santaPageColor:
          DgpImage_free(pPgImage);
          break;
      case santaPageMono:
          DgpImage_free(pPgImage);
          break;
      }
  }
  
  if(pDoc != NULL) santaCloseDocumentEx(&pDoc,dgp_state);
  
  if((_hDC == NULL) && (hDC != NULL)) {
      switch(dc_status) {
      case STATUS_PAGE:
          EndPage(hDC);
          /* drop through */
      case STATUS_DOC:
          if(createJob == DGP_TRUE)   EndDoc(hDC);
          break;
      }
      DeleteDC(hDC);
  }
  
  ERR_RETURN;
}




ErrCode
doRender(santaDoc *pdoc, 
         Dgp_Int32 pgNum, 
         Dgp_Int8  orient,
         Dgp_Int32 *PgType, 
         void 	   **pPgImage, 
         Dgp_Int32 *pgWidth, 
         Dgp_Int32 *pgHeight, 
         Dgp_Int32 *pgXRes, 
         Dgp_Int32 *pgYRes,
         Dgp_State *dgp_state)
{
    DGP_PROLOGUE("doRender()", DGP_PERRSTATE);
    santaPage *ppage;
    Dgp_Int32 x,y,w,h;
    Dgp_Int32 inputtype,dstBpl=-1;
    void *out= NULL;
    void *dst= NULL;
    Dgp_Int32 nw = 0, nh = 0;
    FILE *outgoing = NULL;
    
    DGP_ERR_PASS_CLEANUPF1(santaOpenPage(pdoc, pgNum, dgp_state, &ppage),
        "couldn't open page %d", pgNum);
    ppage->dst_res = ppage->res;

    x = 0;
    y = 0;
    w = ppage->width;
    h = ppage->height;
    nw = (DgpCoord_isRotated(orient, ORIENT_RD) ? h : w);
    nh = (DgpCoord_isRotated(orient, ORIENT_RD) ? w : h);
    
    if (ppage->page_type == santaPageColor){
        inputtype = santaInputRGB;
        *PgType = santaPageColor;
    } else if (ppage->page_type == santaPageMono){
        inputtype = santaInputMonoPacked;
        *PgType = santaPageMono;
    }
    if (inputtype == santaInputRGB){
        DGP_ERR_PASS(DGPIMAGE_NEW(dgp_state, DGPIMAGE_RGB, nw, nh, &out));
        dst = (void *)DGPIMG_PIXPTR((DgpImage_ColorRGB)out, 0, 0);
        if (nh != 0){
            dstBpl = (char *)DGPIMG_PIXPTR((DgpImage_ColorRGB)out, 0, 1) - 
                (char *)DGPIMG_PIXPTR((DgpImage_ColorRGB)out, 0, 0);
        }
        else {
            dstBpl = 0;
        }
    } else if (inputtype == santaInputMonoPacked){
        DGP_ERR_PASS(DGPIMAGE_NEW(dgp_state, DGPIMAGE_PACKED, nw, nh, &out));
        dst = (void *)DGPIMG_GETROW((DgpImage_Packed)out, 0);
        dstBpl = DGPIMG_GETROWWIDTH((DgpImage_Packed)out);
    }
    
    DGP_ERR_PASS(santaRenderInStrips(	pdoc, 
        ppage, 
        orient, 
        1 << (DGP_BITS_UINT32 - 2), 
        x, y, 
        w, h,
        santaJobOptsDefault, 
        inputtype,
        dstBpl, 
        dst, 
        out, 
        DGP_TRUE,
        dgp_state,
        &outgoing,
        DGP_FALSE,
        DGP_SHOWALL_MASK));
    *pgWidth  = ppage->width;
    *pgHeight = ppage->height;
    *pgXRes   = DgpCoord_floor(ppage->res.xRes);
    *pgYRes   = DgpCoord_floor(ppage->res.yRes);
    *pPgImage = out;
    out = NULL;
    DGP_ERR_PASS(santaClosePage(pdoc, &ppage, dgp_state));
    NORMAL_RETURN;
ERR_CLEANUP:
    if(ppage)
        santaClosePage(pdoc, &ppage, dgp_state);
    ERR_RETURN;
}

