/* $Header: /project/icons/cvsroot/digipaper/src/libzephyr/dictionary.c,v 1.4 2003/02/10 22:53:43 mkrocken Exp $ */

#include "zephdictionary.h"
#include "zephprivate.h"

#include "dgp_xif_err_control.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libzephyr/dictionary.c,v 1.4 2003/02/10 22:53:43 mkrocken Exp $");


ErrCode
zephOpenDictionary( zephDocumentInfo docInfo, Dgp_UInt32 libCount,
                    Dgp_UInt32 *libIds, zephDictionaryInfo *pDictInfo)
{
  zephDictionaryInfo dict = (zephDictionaryInfo)NULL;
  _zephPrivateDictionaryInfo *_priv = (_zephPrivateDictionaryInfo *)NULL;
  _zephPrivateDocumentInfo *_privDoc;
/*  XDoc *xDoc; */
  Dgp_UInt32 i;

ZEPH_PROLOGUE_DOC( "zephOpenDictionary()", docInfo);

DGP_ERR_GEN( libCount > 0 && libIds == (Dgp_UInt32 *)NULL, cErrInvalidParm);
DGP_ERR_GEN( pDictInfo == (zephDictionaryInfo *)NULL, cErrInvalidParm);

  _privDoc = (_zephPrivateDocumentInfo *)docInfo->privateInfo;

DGP_CHECK_ALLOC_CLEANUP(
  dict = (zephDictionaryInfo)DGP_MALLOC( sizeof( *dict) + sizeof( *_priv)));
  _priv = (_zephPrivateDictionaryInfo *)((Dgp_UInt8 *)dict 
                                         + sizeof( *dict));

  dict->privateInfo = (zephPrivateDictionaryInfo)_priv;
  dict->libIds = (Dgp_UInt32 *)NULL;
  dict->version = ZEPH_BACKENDDICT_VERSION_NONE;

  _priv->docInfo = docInfo;
  _priv->dgp_state = dgp_state;
  _priv->libSizes = (Dgp_UInt32 *)NULL;
  _priv->bufferSizes = (Dgp_UInt32 *)NULL;
  _priv->buffers = (Dgp_UInt8 **)NULL;

  dict->libCount = libCount;

  /* Next malloc space for all the fields. */
DGP_CHECK_ALLOC_CLEANUP(
  dict->libIds = (Dgp_UInt32 *)DGP_MALLOC( libCount 
                                              * sizeof( Dgp_UInt32)));
DGP_CHECK_ALLOC_CLEANUP(
  _priv->libSizes = (Dgp_UInt32 *)DGP_MALLOC( libCount 
                                                 * sizeof( Dgp_UInt32)));
DGP_CHECK_ALLOC_CLEANUP(
  _priv->bufferSizes = (Dgp_UInt32 *)DGP_MALLOC( libCount 
                                                    * sizeof( Dgp_UInt32)));
DGP_CHECK_ALLOC_CLEANUP( 
  _priv->buffers = (Dgp_UInt8 **)DGP_MALLOC( libCount
                                                * sizeof( Dgp_UInt8 *)));

/*  xDoc = (XDoc *)_privDoc->IO_doc_handle; */

  /* Get the symlib offsets. */
  for( i = 0; i < libCount; i++) {
    dict->libIds[ i] = libIds[ i];

    /* 30.8.2000 / dhtesq. */
    /* Used to call ipcore function here to get symlib offsets and sizes, */
    /* but this appears to be impossible in tfx.  Will fill in libsizes */
    /* when we read the position block, using the memory allocated here. */

    _priv->bufferSizes[ i] = 0;
    _priv->buffers[ i] = (Dgp_UInt8 *)NULL;
  }

  dict->userFlags = 0;
  *pDictInfo = dict;

  NORMAL_RETURN;

ERR_CLEANUP:
  if( dict != (zephDictionaryInfo)NULL) {
    if( dict->libIds != (Dgp_UInt32 *)NULL) {
      DGP_FREE( (void *)dict->libIds);
    }
    if( _priv->libSizes != (Dgp_UInt32 *)NULL) {
      DGP_FREE( (void *)_priv->libSizes);
    }
    if( _priv->bufferSizes != (Dgp_UInt32 *)NULL) {
      DGP_FREE( (void *)_priv->bufferSizes);
    }
    if( _priv->buffers != (Dgp_UInt8 **)NULL) {
      DGP_FREE( (void *)_priv->buffers);
    }
    DGP_FREE( (void *)dict);
  }

  ERR_RETURN;
}


ErrCode
zephSplitDictionary( zephDictionaryInfo dictInfo, Dgp_UInt32 which,
                     zephDictionaryInfo *pDictInfo)
{
  zephDictionaryInfo dict = (zephDictionaryInfo)NULL;
  _zephPrivateDictionaryInfo *_priv = (_zephPrivateDictionaryInfo *)NULL;
  _zephPrivateDictionaryInfo *_privOld;

ZEPH_PROLOGUE_DICT( "zephSplitDictionary()", dictInfo);

DGP_ERR_GEN( which >= dictInfo->libCount, cErrInvalidParm);
DGP_ERR_GEN( pDictInfo == (zephDictionaryInfo *)NULL, cErrInvalidParm);

  _privOld = (_zephPrivateDictionaryInfo *)dictInfo->privateInfo;

DGP_CHECK_ALLOC_CLEANUP(
  dict = (zephDictionaryInfo)DGP_MALLOC( sizeof( *dict) + sizeof( *_priv)));
  _priv = (_zephPrivateDictionaryInfo *)((Dgp_UInt8 *)dict
                                        + sizeof( *dict));

  dict->privateInfo = (zephPrivateDictionaryInfo)_priv;
  dict->libIds = (Dgp_UInt32 *)NULL;
  _priv->docInfo = _privOld->docInfo;
  _priv->dgp_state = dgp_state;
  _priv->libSizes = (Dgp_UInt32 *)NULL;

  dict->libCount = 1;


DGP_CHECK_ALLOC_CLEANUP(
  dict->libIds = (Dgp_UInt32 *)DGP_MALLOC( sizeof( Dgp_UInt32)));
DGP_CHECK_ALLOC_CLEANUP(
  _priv->libSizes = (Dgp_UInt32 *)DGP_MALLOC( sizeof( Dgp_UInt32)));
DGP_CHECK_ALLOC_CLEANUP(
  _priv->bufferSizes = (Dgp_UInt32 *)DGP_MALLOC( sizeof( Dgp_UInt32)));
DGP_CHECK_ALLOC_CLEANUP( 
  _priv->buffers = (Dgp_UInt8 **)DGP_MALLOC( sizeof( Dgp_UInt8 *)));

  dict->libIds[ 0] = dictInfo->libIds[ which];
  dict->version = dictInfo->version;

  _priv->libSizes[ 0] = _privOld->libSizes[ which];
  _priv->bufferSizes[ 0] = 0;
  _priv->buffers[ 0] = (Dgp_UInt8 *)NULL;

  dict->userFlags = 0;
  *pDictInfo = dict;

  NORMAL_RETURN;

ERR_CLEANUP:
  if( dict != (zephDictionaryInfo)NULL) {
    if( dict->libIds != (Dgp_UInt32 *)NULL) {
      DGP_FREE( (void *)dict->libIds);
    }
    if( _priv->libSizes != (Dgp_UInt32 *)NULL) {
      DGP_FREE( (void *)_priv->libSizes);
    }
    if( _priv->bufferSizes != (Dgp_UInt32 *)NULL) {
      DGP_FREE( (void *)_priv->bufferSizes);
    }
    if( _priv->buffers != (Dgp_UInt8 **)NULL) {
      DGP_FREE( (void *)_priv->buffers);
    }
    DGP_FREE( (void *)dict);
  }

  ERR_RETURN;
}


ErrCode
zephMergeDictionaries( zephDocumentInfo docInfo, 
                       Dgp_UInt32 dictInfoCount, 
                       zephDictionaryInfo *dictInfos, 
                       zephDictionaryInfo *pDictInfo)
{
  zephDictionaryInfo dict = (zephDictionaryInfo)NULL;
  _zephPrivateDictionaryInfo *_priv = (_zephPrivateDictionaryInfo *)NULL;
  _zephPrivateDictionaryInfo *_privOld;
  Dgp_UInt32 n, i, j, k, m;
  Dgp_Bool unique = DGP_TRUE;

ZEPH_PROLOGUE_DOC( "zephMergeDictionaries()", docInfo);

DGP_ERR_GEN( dictInfoCount > 0 && dictInfos == (zephDictionaryInfo *)NULL, 
             cErrInvalidParm);

DGP_CHECK_ALLOC_CLEANUP(
  dict = (zephDictionaryInfo)DGP_MALLOC( sizeof( *dict) + sizeof( *_priv)));
  _priv = (_zephPrivateDictionaryInfo *)((Dgp_UInt8 *)dict 
                                        + sizeof( *dict));

  dict->privateInfo = (zephPrivateDictionaryInfo)_priv;
  dict->libIds = (Dgp_UInt32 *)NULL;
  dict->version = ZEPH_BACKENDDICT_VERSION_NONE;

  _priv->docInfo = docInfo;
  _priv->dgp_state = dgp_state;
  _priv->libSizes = (Dgp_UInt32 *)NULL;
  _priv->bufferSizes = (Dgp_UInt32 *)NULL;
  _priv->buffers = (Dgp_UInt8 **)NULL;

  if( dictInfoCount > 0) {
    dict->version = dictInfos[ 0]->version;
  }

  for( i = 0, n = 0; i < dictInfoCount; i++) {
    if( dictInfos[ i]->version != dict->version) {
      DGP_ERR_CLEANUPF0( cErrDecodeFailure,
                       "Cannot merge dictionaries with different versions.");
    }

    for( j = 0; j < dictInfos[ i]->libCount; j++) {
      for( k = 0; k < i; k++) {
        /* First compare to all other dictionaries. */
        for( m = 0; m < dictInfos[ k]->libCount; m++) {
          if( dictInfos[ i]->libIds[ j] == dictInfos[ k]->libIds[ m]) {
            unique = DGP_FALSE;
          }
        }
      }
      /* Now compare to previous symlibs in this dictionary. */
      for( m = 0; m < j; m++) {
        if( dictInfos[ i]->libIds[ j] == dictInfos[ i]->libIds[ m]) {
          unique = DGP_FALSE;
        }
      }
      if( unique) {
        n++;
      }

      unique = DGP_TRUE;
    }
  }

  dict->libCount = n;

DGP_CHECK_ALLOC_CLEANUP(
  dict->libIds = (Dgp_UInt32 *)DGP_MALLOC( n * sizeof( Dgp_UInt32)));
DGP_CHECK_ALLOC_CLEANUP(
  _priv->libSizes = (Dgp_UInt32 *)DGP_MALLOC( n * sizeof( Dgp_UInt32)));
DGP_CHECK_ALLOC_CLEANUP(
  _priv->bufferSizes = (Dgp_UInt32 *)DGP_MALLOC( n 
                                                  * sizeof( Dgp_UInt32 *)));
DGP_CHECK_ALLOC_CLEANUP( 
  _priv->buffers = (Dgp_UInt8 **)DGP_MALLOC(n * sizeof( Dgp_UInt8 *)));

  if( dict->libCount > 0) {
    for( i = 0, n = 0; i < dictInfoCount; i++) {
      for( j = 0; j < dictInfos[ i]->libCount; j++) {
        for( k = 0; k < i; k++) {
          for( m = 0; m < dictInfos[ k]->libCount; m++) {
            if( dictInfos[ i]->libIds[ j] == dictInfos[ k]->libIds[ m]) {
              unique = DGP_FALSE;
            }
          }
        }
        for( m = 0; m < j; m++) {
          if( dictInfos[ i]->libIds[ j] == dictInfos[ i]->libIds[ m]) {
            unique = DGP_FALSE;
          }
        }
        if( unique) {
          dict->libIds[ n] = dictInfos[ i]->libIds[ j];
          _privOld = (_zephPrivateDictionaryInfo *)dictInfos[ i]->privateInfo;
          _priv->libSizes[ n] = _privOld->libSizes[ j];
          _priv->bufferSizes[ n] = _privOld->bufferSizes[ j];
          _priv->buffers[ n] = _privOld->buffers[ j];
          n++;
        }
        
        unique = DGP_TRUE;
      }  
    }
  }

  for( i = 0; i < dictInfoCount; i++) {
DGP_ERR_PASS_CLEANUP(
    zephCloseDictionary( &dictInfos[ i]));
  }

  dict->userFlags = 0;
  *pDictInfo = dict;

  NORMAL_RETURN;

ERR_CLEANUP:
  if( dict != (zephDictionaryInfo)NULL) {
    if( dict->libIds != (Dgp_UInt32 *)NULL) {
      DGP_FREE( (void *)dict->libIds);
    }
    DGP_FREE( (void *)dict);
  }

  for( i = 0; i < dictInfoCount; i++) {
    zephCloseDictionary( &dictInfos[ i]);
  }

  ERR_RETURN;
}


ErrCode
zephCloseDictionary( zephDictionaryInfo *pDictInfo)
{
  zephDictionaryInfo dict;
  _zephPrivateDictionaryInfo *_priv;
  Dgp_UInt32 i;

ZEPH_PROLOGUE_DICT( "zephCloseDictionary()", *pDictInfo);

  dict = *pDictInfo;
  _priv = (_zephPrivateDictionaryInfo *)dict->privateInfo; 

  for( i = 0; i < dict->libCount; i++) {
    if( _priv->buffers[ i] != (Dgp_UInt8 *)NULL) {
      DGP_FREE( (void *)_priv->buffers[ i]);
    }
  }

  DGP_FREE( (void *)dict->libIds);
  DGP_FREE( (void *)_priv->libSizes);
  DGP_FREE( (void *)_priv->bufferSizes);
  DGP_FREE( (void *)_priv->buffers);
  DGP_FREE( (void *)dict);
  
  *pDictInfo = (zephDictionaryInfo)NULL;

  NORMAL_RETURN;
} 


Dgp_UInt32
zephGetDictionaryMemoryUsage( zephDictionaryInfo dictInfo)
{
  _zephPrivateDictionaryInfo *_priv;
  Dgp_UInt32 usage, i;

  if( dictInfo == (zephDictionaryInfo)NULL) {
    return 0;
  }

  _priv = (_zephPrivateDictionaryInfo *)dictInfo->privateInfo;

  usage = sizeof( *dictInfo) 
    + 4 * dictInfo->libCount * sizeof( Dgp_UInt32)
    + dictInfo->libCount * sizeof( Dgp_UInt8 *)
    + sizeof( _zephPrivateDictionaryInfo);

  for( i = 0; i < dictInfo->libCount; i++) {
    usage += _priv->bufferSizes[ i];
  }
  
  return usage;
}
