/* $Header: /project/icons/cvsroot/digipaper/src/libzephyr/document.c,v 1.9 2003/02/10 22:53:44 mkrocken Exp $ */

#include "zephdocument.h"
#include "zephprivate.h"
#include "iowrap.h"
#include "segment.h"

#include "dgp_xif_err_control.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libzephyr/document.c,v 1.9 2003/02/10 22:53:44 mkrocken Exp $");


ErrCode
zephOpenDocument( Dgp_State *dgp_state, uds *u, zephDocumentInfo *pDocInfo)
{
  zephDocumentInfo doc = (zephDocumentInfo)NULL;
  _zephPrivateDocumentInfo *_priv = (_zephPrivateDocumentInfo *)NULL;
  Dgp_Int32 client_id = 0, options = 0;
  /*Dgp_Int8 *author_block = (Dgp_Int8 *)NULL; */
  FileID *fileID;

  DGP_STATE_PROLOGUE( "zephOpenDocument()", dgp_state);

  DGP_ERR_GEN( u == (uds *)NULL, cErrInvalidParm);
  DGP_ERR_GEN( pDocInfo == (zephDocumentInfo *)NULL, cErrInvalidParm);

  /* Allocate some stuff. */
  DGP_CHECK_ALLOC_CLEANUP(fileID = DGP_MALLOC( sizeof(FileID)));
  DGP_CHECK_ALLOC_CLEANUP(doc = DGP_MALLOC( sizeof( *doc) + sizeof( *_priv)));
  _priv = (_zephPrivateDocumentInfo *)((Dgp_UInt8 *)doc + sizeof( *doc));

  doc->privateInfo = (zephPrivateDocumentInfo)_priv;
  _priv->IO_instance_id = 0;
  _priv->IO_doc_handle = 0;
  _priv->dgp_state = dgp_state;
  /* 30.4.2001 / dhtesq. */
  /* Leaking fileID;  record pointer. */
  _priv->fileID = fileID;

  fileID->fp = u->fp;
  fileID->pos = u->ftell (u->fp);
  fileID->u = u;

  /* Start up the MRCIO session. */
  IO_InitInstance( client_id, &_priv->IO_instance_id);
/*DGP_XIF_ERR_PASS_CLEANUP( 
  IO_SetMallocFuncs( _priv->IO_instance_id, (void *)&(DGP_ALLOCATORS), 
                       DGP_MALLOC, DGP_FREE));*/

DGP_XIF_ERR_PASS_CLEANUP(
  IO_OpenDocumentRead(_priv->IO_instance_id, (Int32) fileID,
	    options, (IO_READFUNC)fileRead, (IO_WRITEFUNC)fileWrite, fileSeek, fileFileSize, fileTell,
	    &_priv->IO_doc_handle));


DGP_XIF_ERR_PASS_CLEANUP(
  IO_GetDocInfo( _priv->IO_doc_handle, &doc->version, &doc->revision, 
                  &doc->pageCount));

  doc->userFlags = 0;

  *pDocInfo = doc;
  NORMAL_RETURN;

ERR_CLEANUP:

  if ( _priv->IO_doc_handle != 0)
	IO_CloseDocument( _priv->IO_doc_handle);
  
  if ( _priv->IO_instance_id != 0)
	IO_EndInstance( _priv->IO_instance_id);

  if( doc != (zephDocumentInfo)NULL) {
    DGP_FREE( (void *)doc);
  }

  ERR_RETURN;
}

ErrCode
zephCloseDocument( zephDocumentInfo *pDocInfo)
{
  _zephPrivateDocumentInfo *_priv;

ZEPH_PROLOGUE_DOC("zephCloseDocument()", *pDocInfo);

  _priv = (_zephPrivateDocumentInfo *)(*pDocInfo)->privateInfo;

DGP_XIF_ERR_PASS(
  IO_CloseDocument( _priv->IO_doc_handle));

DGP_XIF_ERR_PASS(
  IO_EndInstance( _priv->IO_instance_id));

  /* 30.4.2001 / dhtesq. */
  /* Leaking fileID... */
  DGP_FREE(_priv->fileID);
  
  DGP_FREE( (void *)(*pDocInfo));

  *pDocInfo = (zephDocumentInfo)NULL;
  NORMAL_RETURN;
}

Dgp_UInt32
zephGetDocumentMemoryUsage( zephDocumentInfo docInfo)
{
  if( docInfo == (zephDocumentInfo)NULL) {
    return 0;
  }

  return sizeof( *docInfo)
    + sizeof( _zephPrivateDocumentInfo);
}

/*extern ErrCode
zephGetSymlibCount( zephDocumentInfo docInfo, Dgp_UInt32 *libCount)
{
  _zephPrivateDocumentInfo *_priv;
  XDoc *xDoc;

ZEPH_PROLOGUE_DOC("zephGetSymlibCount()", docInfo);

  _priv = (_zephPrivateDocumentInfo *)docInfo->privateInfo;

  xDoc = (XDoc *)_priv->IO_doc_handle;

  *libCount = xDoc->symlibTableHeader.symlibCount;
 
  NORMAL_RETURN;
} */
