/* $Header: /project/icons/cvsroot/digipaper/src/libzephyr/group4.c,v 1.12 2003/02/10 22:53:46 mkrocken Exp $ */

#include "endianness.h"
#include "group4.h"

#include "zephconst.h"
#include "g4.h"

#ifdef ZEPH_INCLUDE_GROUP4

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libzephyr/group4.c,v 1.12 2003/02/10 22:53:46 mkrocken Exp $");


typedef struct {
  Dgp_State *dgp_state;
  uds u;
  IO_IMGHANDLE image_handle;
  Dgp_UInt32 compSize;

  Dgp_Int32 srcW, srcH;
  DgpCoord_Resolution *srcRes;

  Dgp_UInt32 g4InternalSize;

  Dgp_Bool has_multiStrip;
  Dgp_UInt32 stripHeight;
  Dgp_UInt32 strip_num;

} zephG4State;


ErrCode
zephG4Open( Dgp_State *dgp_state,
            zephImageInfo image, 
            Dgp_Int32 stripNum, Dgp_Int32 srcH,
            DgpCoord_Resolution *srcRes, 
            void **decodeState)
{
  zephG4State *state = (zephG4State *)NULL;

DGP_STATE_PROLOGUE( "zephG4Open()", dgp_state);

DGP_CHECK_ALLOC_CLEANUP(
  state = DGP_MALLOC( sizeof( *state)));

  state->dgp_state = dgp_state;
  state->image_handle = ((_zephPrivateImageInfo*) image->privateInfo)->IO_image_handle;
  state->compSize = image->compSize;

  state->srcW = image->width;
  state->srcH = srcH;
  state->srcRes = srcRes;
  state->strip_num = stripNum;

  if (image->height != image->stripHeight)
  {
	  state->has_multiStrip = DGP_TRUE;
	  state->stripHeight = image->stripHeight;
  }
  else
  {
	  state->has_multiStrip = DGP_FALSE;
  }

  *decodeState = (void *)state;

  NORMAL_RETURN;

ERR_CLEANUP:
  if( state != (zephG4State *)NULL) {
    DGP_FREE( (void *)state);
  }
  ERR_RETURN;
}

Dgp_UInt32
zephG4Usage( void *decodeState)
{
  if( decodeState != (void *)NULL) {
    return sizeof( zephG4State);
  }
  else {
    return 0;
  }
}

ErrCode
zephG4Abilities( void *decodeState,
                 Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
                 zephReadType *type,
                 DgpCoord_Orientation *orient,
                 DgpCoord_Resolution *dstRes,
                 Dgp_Int32 *dstX, Dgp_Int32 *dstY,
                 Dgp_Int32 *dstW, Dgp_Int32 *dstH,
                 Dgp_UInt32 *maxMemRequired,
                 Dgp_Int32 srcOriginX, Dgp_Int32 srcOriginY,
                 Dgp_UInt8 *cantDo)
{
  zephG4State *state = (zephG4State *)decodeState;
  
  /* FIXME unused params: */
  dstMinX = dstMinX;
  dstMinY = dstMinY;
  srcOriginX = srcOriginX;
  srcOriginY = srcOriginY;
  
  if( *type != ZEPHRT_MONO_PACKED && *type != ZEPHRT_RGB_INTERLEAVED) {
    *type = ZEPHRT_MONO_PACKED;
    *cantDo |= ZEPH_CANT_DO_TYPE;
  }

  if( *orient != ORIENT_RD) {
    *orient = ORIENT_RD;
    *cantDo |= ZEPH_CANT_DO_ORIENT;
  }

  if( !(DgpCoord_equal( state->srcRes->xRes, dstRes->xRes)
        && DgpCoord_equal( state->srcRes->yRes, dstRes->yRes))) {
    dstRes->xRes = state->srcRes->xRes;
    dstRes->yRes = state->srcRes->yRes;
    *cantDo |= ZEPH_CANT_DO_RESOLUTION;
  }

  /* no scaling and no offset in x */
  if( *dstX != 0 || *dstW != state->srcW)
  {
	*dstX = 0;
    *dstW = state->srcW;
    *cantDo |= ZEPH_CANT_DO_RECT;
  }

  /* This is my best estimation based on the code in libg4. */
  state->g4InternalSize = /* ExpandData */ 52
    + /* IndexSequence */ 12
    + /* IndexSequence */ 12
    + /* ref sequence */ ((*dstW + 4) * sizeof( int))
    + /* line sequence */ ((*dstW + 4) * sizeof( int));

  /* We require at least the size of the compressed image plus 
   * FIXME whatever the actual decoder needs (jamez!!). */
  if( *maxMemRequired < state->compSize + state->g4InternalSize) {
    *maxMemRequired = state->compSize + state->g4InternalSize;
    *cantDo |= ZEPH_CANT_DO_MEMORY;
  }

  return cSuccessful;
}

ErrCode
zephG4Read( void *decodeState, 
            void *dst, Dgp_Int32 dstBpl,
            Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
            zephReadType type,
            DgpCoord_Orientation orient,
            DgpCoord_Resolution *dstRes,
            Dgp_Int32 dstX, Dgp_Int32 dstY,
            Dgp_Int32 dstW, Dgp_Int32 dstH,
            Dgp_UInt32 *maxMemRequired,
            Dgp_Int32 srcOriginX, Dgp_Int32 srcOriginY,
			void		*trans_src,
			Dgp_Int32	trans_bpl,
            Dgp_Int32 nominal_page_width, Dgp_Int32 nominal_page_height,
            Dgp_Int32 nominal_strip_height, Dgp_Int32 scaled_strip_height)
{
  zephG4State *state = (zephG4State *)decodeState;
  Dgp_State *dgp_state = state->dgp_state;
  Dgp_UInt8 *buffer = (Dgp_UInt8 *)NULL;
  Dgp_UInt32 bytesFilled = 0;
  Dgp_Int32 rowOffset;

DGP_STATE_PROLOGUE( "zephG4Read()", dgp_state);

  dstMinX = dstMinX; dstMinY = dstMinY;
  nominal_page_width = nominal_page_width;
  nominal_page_height = nominal_page_height;
  type = type; orient = orient; dstRes = dstRes;
  dstX = dstX; 
  maxMemRequired = maxMemRequired;
  srcOriginX = srcOriginX; srcOriginY = srcOriginY;

  DGP_ASSERT( *maxMemRequired >= state->compSize + state->g4InternalSize);

  if (type == ZEPHRT_MONO_PACKED)
  {

  /* Check to make sure dst is on a word boundary. */
  if( (Dgp_UInt32)dst % 4 != 0) {
    DGP_ERRF0( cErrInvalidParm, "The G4 decoder is unable to handle "
               "destinations which do not start on a word boundary."); 
  }

  /* Check to make sure that there are an integer number of words in a row. */
  if( dstBpl % 4 != 0) {
    DGP_ERRF0( cErrInvalidParm, "The G4 decoder handles only "
               "row destinations whose length is a multiple of four bytes."); 
  }
  }

  DGP_CHECK_ALLOC_CLEANUP(buffer = DGP_MALLOC( state->compSize));

  if (state->has_multiStrip)
  {
	rowOffset = state->strip_num * state->stripHeight;
  }
  else
  {
	rowOffset = 0;
  }

  DGP_ERR_PASS_CLEANUP(IO_ImageReadStart(state->image_handle, 0 /* Flags */));
  DGP_ERR_PASS_CLEANUP(IO_CompressedReadStrip (state->image_handle,
					       rowOffset,
					       state->srcH,
					       &buffer,
					       &state->compSize,
					       &bytesFilled));

  /* Close the I/O handle to indicate that we read the data. */
  IO_ImageReadFinish (state->image_handle);
	  
  if (type == ZEPHRT_MONO_PACKED)
  {
	if (dstY == 0 && dstH == state->srcH)
	{
DGP_ERR_PASS_CLEANUP(
	  g4Decode( (Dgp_UInt32 *) dst, dstW, dstH, dstBpl / 4, 0,
            dstW - 1, dstH - 1, buffer, state->compSize, 
            DGP_TRUE, (size_t *)NULL, state->dgp_state)); 
	}
	else
	{
	  Dgp_UInt32 *temp_buf;
	  Dgp_UInt8 *copy_buf;
	  Dgp_UInt8 *dst_ptr = (Dgp_UInt8*) dst;
	  DGP_CHECK_ALLOC_CLEANUP(temp_buf = DGP_CALLOC(dstBpl * state->srcH, 1));

	  /* decode into a temporary buffer */
DGP_ERR_PASS_CLEANUP(
	  g4Decode(temp_buf, dstW, state->srcH, dstBpl / 4, 0,
            dstW - 1, state->srcH - 1, buffer, state->compSize, 
            DGP_TRUE, (size_t *)NULL, state->dgp_state)); 

	  /* copy requested rows */
	  copy_buf = (Dgp_UInt8*) temp_buf + dstY * dstBpl;	  
	  memcpy (dst_ptr, copy_buf, dstH * dstBpl);
      DGP_FREE (temp_buf);
	}
  }
  else if (type == ZEPHRT_RGB_INTERLEAVED)
  {
	  Dgp_UInt32 *temp_buf;
	  Dgp_UInt8 *dst_ptr = (Dgp_UInt8*) dst;
	  Dgp_UInt8 mask, *col_ptr, *row_ptr, *dst_row_ptr, *dst_col_ptr;
	  Dgp_Int32 h, w;
	  Dgp_Int32 wpl = state->srcW / 32 + 1;
	  DGP_CHECK_ALLOC_CLEANUP(temp_buf = DGP_CALLOC(wpl * state->srcH, 4 ));

	  /* decode into a temporary buffer */
DGP_ERR_PASS_CLEANUP(
	  g4Decode(temp_buf, dstW, state->srcH, wpl, 0,
            state->srcW - 1, state->srcH - 1, buffer, state->compSize, 
            DGP_TRUE, (size_t *)NULL, state->dgp_state)); 

/* the g4 decoder works with words, thus need to check endianness */
#ifdef _LITTLE_ENDIAN
	  row_ptr = (Dgp_UInt8*) temp_buf + dstY * wpl * 4;
	  for (h = 0; h < wpl * dstH; h++, row_ptr += 4)
	  {
		mask = *row_ptr;
		*row_ptr = *(row_ptr + 3);
		*(row_ptr + 3) = mask;

		mask = *(row_ptr + 1);
		*(row_ptr + 1) = *(row_ptr + 2);
		*(row_ptr + 2) = mask;
	  }
#endif

	  /* copy each bit into 3 bytes in the RGB destination buffer */
	  row_ptr = (Dgp_UInt8*) temp_buf + dstY * wpl * 4;
	  dst_row_ptr = (Dgp_UInt8*) dst_ptr;
	  for (h = 0; h < dstH; h++, row_ptr += wpl * 4, dst_row_ptr += dstBpl)
	  {
		  w = 0;
		  col_ptr = row_ptr;
		  dst_col_ptr = dst_row_ptr;
		  for (mask = 0x80; w < dstW; w++, mask >>=1)
		  {
			  if (!mask)
			  {
				  mask = 0x80;
				  col_ptr++;
			  }
			  if (mask & *col_ptr)
			  {
				 *dst_col_ptr = 0x00;
				 dst_col_ptr++;
				 *dst_col_ptr = 0x00;
				 dst_col_ptr++;
				 *dst_col_ptr = 0x00;
				 dst_col_ptr++;
			  }
			  else
			  {
				  dst_col_ptr += 3;
			  }
		  }
	  }
	DGP_FREE (temp_buf);
  }

  DGP_FREE( (void *)buffer); 

  NORMAL_RETURN;

ERR_CLEANUP:
  if( buffer != (Dgp_UInt8 *)NULL) {
    DGP_FREE( (void *)buffer);
  }
  ERR_RETURN;
}


ErrCode 
zephG4Close( void *decodeState)
{
  zephG4State *state = (zephG4State *)decodeState;
  Dgp_State *dgp_state = state->dgp_state;
  
  DGP_FREE( (void *)state);

  return cSuccessful;
}

zephDecodeOps zephG4Ops = { zephG4Open, 
                            zephG4Usage,
                            zephG4Abilities,
                            zephG4Read,
                            zephG4Close };

#else /* ZEPH_INCLUDE_GROUP4 */

#include "nop.h"

zephDecodeOps zephG4Ops = { zephNopOpen,
                            zephNopUsage,
                            zephNopAbilities,
                            zephNopRead,
                            zephNopClose };

#endif /* ZEPH_INCLUDE_GROUP4 */

