/* $Header: /project/icons/cvsroot/digipaper/src/libzephyr/image.c,v 1.21 2003/02/10 22:53:46 mkrocken Exp $ */

#include "zephimage.h"
#include "zephdictionary.h"
#include "zephprivate.h"

#include "jpeg.h"
#include "group4.h"
#include "group3_1D.h"
#include "group3_2D.h"

#include "dgp_xif_err_control.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libzephyr/image.c,v 1.21 2003/02/10 22:53:46 mkrocken Exp $");


ErrCode
zephOpenImage( zephPageInfo pageInfo, Dgp_UInt32 layerNum, 
               Dgp_UInt32 imageNum, zephImageInfo *pImageInfo)
{  zephImageInfo image = (zephImageInfo)NULL;
  _zephPrivateImageInfo *_priv = (_zephPrivateImageInfo *)NULL;
  _zephPrivatePageInfo *_privPage;
  _zephPrivateDocumentInfo *_privDoc;
  zephImageInfo mask = (zephImageInfo)NULL;
  _zephPrivateImageInfo *_privMask = (_zephPrivateImageInfo *)NULL;

  Dgp_UInt32 *libIds = (Dgp_UInt32 *)NULL;
  Dgp_UInt32 *libSizes;

  Dgp_UInt32 xRes, yRes;
  Dgp_UInt32 mask_count;
  Dgp_UInt32 i;
  Dgp_Int32 currentHeight;

ZEPH_PROLOGUE_PAGE( "zephOpenImage()", pageInfo);

DGP_ERR_GEN( layerNum >= pageInfo->layerCount, cErrInvalidParm);
DGP_ERR_GEN( imageNum >= pageInfo->imageCounts[ layerNum], cErrInvalidParm);
DGP_ERR_GEN( pImageInfo == (zephImageInfo *)NULL, cErrInvalidParm);

  _privPage = (_zephPrivatePageInfo *)pageInfo->privateInfo;
  _privDoc = (_zephPrivateDocumentInfo *)_privPage->docInfo->privateInfo;

DGP_CHECK_ALLOC_CLEANUP(
  image = DGP_MALLOC( sizeof( *image) + sizeof( *_priv)));
  _priv = (_zephPrivateImageInfo *)((Dgp_UInt8 *)image + sizeof( *image));

  image->privateInfo = (zephPrivateImageInfo)_priv;
  image->layerNum = layerNum;
  image->imageNum = imageNum;
  image->maskInfo = (zephImageInfo)NULL;
  image->libId = (Dgp_UInt32 *)NULL;
  image->colorTokImgExt = NULL;

  _priv->pageInfo = pageInfo;
  _priv->dgp_state = dgp_state;
  _priv->stripCount = 0;
  _priv->decodeState = (void **)NULL;

DGP_XIF_ERR_PASS_CLEANUP(
  IO_GetImageInfo( _privPage->IO_page_handle, (layerNum + 1), imageNum,
                    &image->xPos, &image->yPos, &xRes, &yRes, &image->width,
                    &image->height, &image->stripHeight, &image->compType,
                    &image->colorType, NULL, &mask_count, 
                    &_priv->IO_image_handle));

DGP_XIF_ERR_PASS_CLEANUP(
  IO_GetImageStats( _priv->IO_image_handle, &image->compSize, 
                     &image->libCount, &libIds, 
                     (Dgp_Int32 **)&libSizes));

/* 30.8.2000 / dhtesq. */
/* Above call seems to misread the libCount... */

  if( image->libCount > 0) { 
DGP_CHECK_ALLOC_CLEANUP(
    image->libId = (Dgp_UInt32 *)DGP_MALLOC( image->libCount
                                                 * sizeof( Dgp_UInt32)));
DGP_CHECK_ALLOC_CLEANUP(
    image->libSize = (Dgp_UInt32 *)DGP_MALLOC(image->libCount * sizeof( Dgp_UInt32)));

    for( i = 0; i < image->libCount; i++) {
      image->libId[ i] = libIds[ i];
      image->libSize[ i] = libSizes[ i];
    }

  } else {
	  image->libSize = NULL;
  }

  image->res.xRes = DgpCoord_eqi( xRes);
  image->res.yRes = DgpCoord_eqi( yRes);
  image->userFlags = 0;

  /* Now compute the number of strips in the image. */
  _priv->stripCount = (image->height + image->stripHeight - 1) 
                        / image->stripHeight;

  /* Setup space for strip information. */

DGP_CHECK_ALLOC_CLEANUP(
  _priv->decodeState = (void **)DGP_CALLOC(1, _priv->stripCount * sizeof( void *)));

  for( i = 0; i < _priv->stripCount; i++) {
    _priv->decodeState[ i] = (void *)NULL;
  }

  currentHeight = image->stripHeight;

  for( i = 0; i < _priv->stripCount; i++) {
    /* For each strip we must find the offset and size, as well as initialize 
     * the proper decoder. */
    if( i == _priv->stripCount - 1) {
      /* This is the last strip, so the height can be smaller than the rest. */
      currentHeight = image->height - (i * image->stripHeight);
    }

    /* Compression type specific. */
    switch( image->compType)
    {
    case IO_COMP_JPEG:
    
DGP_ERR_PASS_CLEANUP(
      zephJpegOps.open( _priv->dgp_state, image, 
                        image->width, 
                        image->height, &image->res, 
                        &_priv->decodeState[ i]));
      break;
      
    case IO_COMP_MMR:

DGP_ERR_PASS_CLEANUP(
      zephG4Ops.open( _priv->dgp_state,  image,
                      i, 
                      currentHeight, &image->res, 
                      &_priv->decodeState[ i]));
      break; 

    /* Cases added by Eli */
    case IO_COMP_MH:

DGP_ERR_PASS_CLEANUP(
      zephG3_1DOps.open( _priv->dgp_state,  image,
                      image->width, 
                      currentHeight, &image->res, 
                      &_priv->decodeState[ i]));
      break; 

    case IO_COMP_MR:

DGP_ERR_PASS_CLEANUP(
      zephG3_2DOps.open( _priv->dgp_state,  image,
                      image->width, 
                      currentHeight, &image->res, 
                      &_priv->decodeState[ i]));
      break; 
    /* Cases added by Eli */

    case IO_COMP_TOKENIZATION:
    case IO_COMP_LLTOKENIZATION:
    case IO_COMP_JBIG2:


      break;

    default:
      DGP_ERR_CLEANUPF0( cErrDecodeFailure, "Unknown compression type."); 
    }
  }

  /* Now open the mask image. */
  image->maskInfo = mask;
  if( mask_count > 0) {
    if( mask_count > 1) {
     DGP_WARNF0( "Multiple transparency masks defined. Using only the first.");
    }

DGP_CHECK_ALLOC_CLEANUP(
    mask = (zephImageInfo)DGP_MALLOC( sizeof( *mask) + sizeof( *_privMask)));
    _privMask = (_zephPrivateImageInfo *)((Dgp_UInt8 *)mask
                                           + sizeof( *mask)); 

    mask->privateInfo = (zephPrivateImageInfo)_privMask;
    mask->layerNum = layerNum;
    mask->imageNum = imageNum;
    mask->maskInfo = (zephImageInfo)NULL;
    mask->libCount = 0;
    mask->libId = (Dgp_UInt32 *)NULL;
    
    _privMask->pageInfo = pageInfo;
    _privMask->dgp_state = dgp_state;
    _privMask->stripCount = 0;
    _privMask->decodeState = (void **)NULL;

DGP_XIF_ERR_PASS_CLEANUP(
    IO_GetMaskImageInfo( _priv->IO_image_handle, 0, &mask->xPos,
                        &mask->yPos, &xRes, &yRes, &mask->width, &mask->height,
                        &mask->stripHeight, &mask->compType, 
                        &_privMask->IO_image_handle));
 
DGP_XIF_ERR_PASS_CLEANUP(
    IO_GetImageStats( _privMask->IO_image_handle, &mask->compSize, 
                       &mask->libCount, &libIds, 
                       (Dgp_Int32 **)&libSizes));

    mask->res.xRes = DgpCoord_eqi( xRes);
    mask->res.yRes = DgpCoord_eqi( yRes);
    mask->userFlags = 0;

    /* Now compute the number of strips in the mask. */
    _privMask->stripCount = (mask->height + mask->stripHeight - 1) 
                          / mask->stripHeight;

  /* Setup space for strip information. */

DGP_CHECK_ALLOC_CLEANUP(
    _privMask->decodeState = (void **)DGP_MALLOC( _priv->stripCount 
                                              * sizeof( void *)));

    for( i = 0; i < _privMask->stripCount; i++) {
      _privMask->decodeState[ i] = (void *)NULL;
    }
 
    for( i = 0; i < _privMask->stripCount; i++) {
      /* For each strip we must find the offset and size, as well as 
       * initialize the proper decoder. */
      
      /* Compression type specific. */
      switch( mask->compType)
      {
      case IO_COMP_MMR:

DGP_ERR_PASS_CLEANUP(
        zephG4Ops.open( _privMask->dgp_state, 
                        mask, 
                        mask->width, mask->stripHeight, 
                        &mask->res, &_privMask->decodeState[ i]));
        break;
  
      /* Cases added by Eli */
      case IO_COMP_MH:

DGP_ERR_PASS_CLEANUP(
        zephG3_1DOps.open( _privMask->dgp_state, 
                        mask, 
                        mask->width, mask->stripHeight, 
                        &mask->res, &_privMask->decodeState[ i]));
        break;
  
      case IO_COMP_MR:

DGP_ERR_PASS_CLEANUP(
        zephG3_2DOps.open( _privMask->dgp_state, 
                        mask, 
                        mask->width, mask->stripHeight, 
                        &mask->res, &_privMask->decodeState[ i]));
        break;
      /* Cases added by Eli */

      default:
        DGP_ERR_CLEANUPF0( cErrDecodeFailure,"Unknown mask compression type.");
      }
    }

    image->maskInfo = mask; 
  } 

  *pImageInfo = image;
  NORMAL_RETURN;


ERR_CLEANUP:
  if( image != (zephImageInfo)NULL) {
    if( _priv != (_zephPrivateImageInfo *)NULL) {
      if( _priv->stripCount > 0) {
        if( _priv->decodeState != (void **)NULL) {
          switch( image->compType)
          {
          case IO_COMP_JPEG:
            for( i = 0; i < _priv->stripCount; i++) {
              if( _priv->decodeState[ i] != (void *)NULL) {
                zephJpegOps.close( _priv->decodeState[ i]);
                _priv->decodeState[ i] =(void *)NULL;
              }
            }
            break;
          
          case IO_COMP_MMR:
            for( i = 0; i < _priv->stripCount; i++) {
              if( _priv->decodeState[ i] != (void *)NULL) {
                zephG4Ops.close( _priv->decodeState[ i]);
                _priv->decodeState[ i] = (void *)NULL;
              }
            }
            break;
            
          /* Cases added by Eli */
          case IO_COMP_MH:
            for( i = 0; i < _priv->stripCount; i++) {
              if( _priv->decodeState[ i] != (void *)NULL) {
                zephG3_1DOps.close( _priv->decodeState[ i]);
                _priv->decodeState[ i] = (void *)NULL;
              }
            }
            break;
            
          case IO_COMP_MR:
            for( i = 0; i < _priv->stripCount; i++) {
              if( _priv->decodeState[ i] != (void *)NULL) {
                zephG3_2DOps.close( _priv->decodeState[ i]);
                _priv->decodeState[ i] = (void *)NULL;
              }
            }
            break;
          /* Cases added by Eli */

          case IO_COMP_TOKENIZATION:
          case IO_COMP_LLTOKENIZATION:
            
            break;
          default:
            DGP_WARNF1( "Unknown compression type: &u.", image->compType);
          }
          DGP_FREE( (void *)_priv->decodeState);
        }
      }
    }
    if( image->libId != (Dgp_UInt32 *)NULL) {
      DGP_FREE( (void *)image->libId);
    }
    DGP_FREE( (void *)image);
  }

  if( mask != (zephImageInfo)NULL) {
    if( _privMask != (_zephPrivateImageInfo *)NULL) {
      if( _privMask->stripCount > 0) {
        if( _privMask->decodeState != (void **)NULL) {
          switch( mask->compType)
          {
          case IO_COMP_MMR:
            for( i = 0; i < _privMask->stripCount; i++) {
              if( _privMask->decodeState[ i] != (void *)NULL) {
                zephG4Ops.close( _privMask->decodeState[ i]);
              }
            }
            break;

          /* Cases added by Eli */
          case IO_COMP_MH:
            for( i = 0; i < _privMask->stripCount; i++) {
              if( _privMask->decodeState[ i] != (void *)NULL) {
                zephG3_1DOps.close( _privMask->decodeState[ i]);
              }
            }
            break;

          case IO_COMP_MR:
            for( i = 0; i < _privMask->stripCount; i++) {
              if( _privMask->decodeState[ i] != (void *)NULL) {
                zephG3_2DOps.close( _privMask->decodeState[ i]);
              }
            }
            break;
          /* Cases added by Eli */

          case IO_COMP_TOKENIZATION:
          case IO_COMP_LLTOKENIZATION:
            
            break;
          default:
            DGP_WARNF1( "Unknown mask compression type: %u.", mask->compType);
          }
          DGP_FREE( (void *)_privMask->decodeState);
        }
      }
    }    
    if( mask->libId != (Dgp_UInt32 *)NULL) {
      DGP_FREE( (void *)mask->libId);
    }
    DGP_FREE( (void *)mask);
  }

  ERR_RETURN;
}
  

ErrCode
zephCloseImage( zephImageInfo *pImageInfo)
{
  zephImageInfo image;
  _zephPrivateImageInfo *_priv;
  Dgp_UInt32 i;

ZEPH_PROLOGUE_IMAGE("zephCloseImage()", *pImageInfo);

  image = *pImageInfo;
  _priv = (_zephPrivateImageInfo *)image->privateInfo;

  /* Free the mask if necessary. */
  if( image->maskInfo != (zephImageInfo)NULL) {
    _zephPrivateImageInfo *_privMask 
      = (_zephPrivateImageInfo *)image->maskInfo->privateInfo;

    if( _privMask->decodeState != (void **)NULL) {
      switch( image->maskInfo->compType)
      {
      case IO_COMP_MMR:
        for( i = 0; i < _privMask->stripCount; i++) {
          if( _privMask->decodeState[ i] != (void *)NULL) {
            zephG4Ops.close( _privMask->decodeState[ i]);
          }
        }
        break;
        
      /* Cases added by Eli */
      case IO_COMP_MH:
        for( i = 0; i < _privMask->stripCount; i++) {
          if( _privMask->decodeState[ i] != (void *)NULL) {
            zephG3_1DOps.close( _privMask->decodeState[ i]);
          }
        }
        break;
        
      case IO_COMP_MR:
        for( i = 0; i < _privMask->stripCount; i++) {
          if( _privMask->decodeState[ i] != (void *)NULL) {
            zephG3_2DOps.close( _privMask->decodeState[ i]);
          }
        }
        break;
      /* Cases added by Eli */
        
      case IO_COMP_TOKENIZATION:
      case IO_COMP_LLTOKENIZATION:
	  case IO_COMP_JBIG2:
        
        break;
      default:
        DGP_WARNF1( "Unknown mask compression type: %u.", 
                    image->maskInfo->compType);
      }
      DGP_FREE( (void *)_privMask->decodeState);
    }
    
    if( image->maskInfo->libId != (Dgp_UInt32 *)NULL) {
      DGP_FREE( (void *)image->maskInfo->libId);
    }
    DGP_FREE( (void *)image->maskInfo);
  }

  if( _priv->decodeState != (void **)NULL) {
    switch( image->compType)
    {
    case IO_COMP_JPEG:
      for( i = 0; i < _priv->stripCount; i++) {
        if( _priv->decodeState[ i] != (void *)NULL) {
          zephJpegOps.close( _priv->decodeState[ i]);
        }
      }
      break;
        
    case IO_COMP_MMR:
      for( i = 0; i < _priv->stripCount; i++) {
        if( _priv->decodeState[ i] != (void *)NULL) {
          zephG4Ops.close( _priv->decodeState[ i]);
        }
      }
      break;
      
        
    /* Cases added by Eli */
    case IO_COMP_MH:
      for( i = 0; i < _priv->stripCount; i++) {
        if( _priv->decodeState[ i] != (void *)NULL) {
          zephG3_1DOps.close( _priv->decodeState[ i]);
        }
      }
      break;
      
    case IO_COMP_MR:
      for( i = 0; i < _priv->stripCount; i++) {
        if( _priv->decodeState[ i] != (void *)NULL) {
          zephG3_2DOps.close( _priv->decodeState[ i]);
        }
      }
      break;
    /* Cases added by Eli */
      
    case IO_COMP_TOKENIZATION:
    case IO_COMP_LLTOKENIZATION:
	case IO_COMP_JBIG2:
      
      break;
    default:
      DGP_WARNF1( "Unknown compression type: &u.", image->compType);
    }
    DGP_FREE( (void *)_priv->decodeState);
  }
  
  if( image->libId != (Dgp_UInt32 *)NULL) {
    DGP_FREE( (void *)image->libId);
  }

  /* 30.4.2001 / dhtesq. */
  /* Leaks... */
  if(image->libSize) DGP_FREE(image->libSize);
  if(image->colorTokImgExt){
	  if(image->colorTokImgExt->sorted_positions) DGP_FREE(image->colorTokImgExt->sorted_positions);
	  DGP_FREE(image->colorTokImgExt);
  }

  DGP_FREE( (void *)image);

  *pImageInfo = (zephImageInfo)NULL;
  NORMAL_RETURN;
}


Dgp_UInt32
zephGetImageMemoryUsage( zephImageInfo imageInfo)
{
  Dgp_UInt32 decoderSize = 0, i;
  _zephPrivateImageInfo *_priv;

  if( imageInfo == (zephImageInfo)NULL) {
    return 0;
  }

  _priv = (_zephPrivateImageInfo *)imageInfo->privateInfo;

  switch( imageInfo->compType)
  {
  case IO_COMP_JPEG:
    for( i = 0; i < _priv->stripCount; i++) {
      decoderSize += zephJpegOps.usage( _priv->decodeState[ i]);
    }
    break;

  case IO_COMP_MMR:
    for( i = 0; i < _priv->stripCount; i++) {
      decoderSize += zephG4Ops.usage( _priv->decodeState[ i]);
    }
    break;

  /* Cases added by Eli */
  case IO_COMP_MH:
    for( i = 0; i < _priv->stripCount; i++) {
      decoderSize += zephG3_1DOps.usage( _priv->decodeState[ i]);
    }
    break;

  case IO_COMP_MR:
    for( i = 0; i < _priv->stripCount; i++) {
      decoderSize += zephG3_2DOps.usage( _priv->decodeState[ i]);
    }
    break;
  /* Cases added by Eli */

  case IO_COMP_TOKENIZATION:
  case IO_COMP_LLTOKENIZATION:

    break;
  }

  return sizeof( *imageInfo)
    + sizeof( Dgp_UInt32) * imageInfo->libCount
    + sizeof( _zephPrivateImageInfo)
/*    + sizeof( TiffFields)
    + sizeof( IO_IMGINFO) */
    + 34 /* Hack. Yes, I know. This is bad, but this is the size of XImg. */
    + (2 * sizeof( Dgp_UInt32) + sizeof( void *)) * _priv->stripCount
    + decoderSize;
}


typedef struct {
  void *dst;
  Dgp_Int32 dstBpl;
  Dgp_Int32 dstMinX,  dstMinY;
  zephReadType type;
  DgpCoord_Orientation orient;
  DgpCoord_Resolution *dstRes;
  Dgp_Int32 dstX, dstY;
  Dgp_Int32 dstW, dstH;
  Dgp_UInt32 maxMemRequired;
  Dgp_Int32 srcOriginX, srcOriginY;

  /* 7.8.2000 / dhtesq.  Adding transparency mask support. */
  void *trans_src;
  Dgp_Int32 trans_bpl;
} LazyReadState;

typedef LazyReadState _zephReadState;


ErrCode
zephReadImageAbilities( zephImageInfo imageInfo,
                       Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
                       zephReadType *type,
                       DgpCoord_Orientation *orient,
                       DgpCoord_Resolution *dstRes,
                       Dgp_Int32 *dstX, Dgp_Int32 *dstY,
                       Dgp_Int32 *dstW, Dgp_Int32 *dstH,
                       Dgp_UInt32 *maxMemRequired,
                       Dgp_Int32 srcOriginX, Dgp_Int32 srcOriginY,
                       Dgp_UInt8 *cantDo)
{
  _zephPrivateImageInfo *_priv;
  Dgp_UInt32 i, strip1, strip2;
  zephReadType currentType, originalType;
  DgpCoord_Orientation currentOrient, originalOrient;
  DgpCoord_Resolution currentDstRes;
  Dgp_Int32 currentDstX, currentDstY, currentDstW, currentDstH;
  Dgp_UInt32 currentMemRequired, originalMemRequired;
  Dgp_UInt8 currentCantDo;
  Dgp_Bool rectOkay;
  /* Attempting to handle image scaling cases...are any of these compressed? / dhtesq. */
  Dgp_Int32 scaled_strip_height;
  Dgp_Float32 scale_factor;

ZEPH_PROLOGUE_IMAGE("zephReadImageAbilities()", imageInfo);

  _priv = (_zephPrivateImageInfo *)imageInfo->privateInfo;
  *cantDo = 0;

  /* FIXME this is here because everything else that happens in this */
  /* function is so completely wrong with respect to scaling that i */
  /* thought it best to just not even try. --spoons */
  /* */
  /* Trying.../ dhtesq. */
  /*DGP_ASSERT( DgpCoord_equal( dstRes->xRes, imageInfo->res.xRes) */
  /*            && DgpCoord_equal( dstRes->yRes, imageInfo->res.yRes)); */
  scale_factor = DgpCoord_float(dstRes->xRes) / DgpCoord_float(imageInfo->res.xRes);

  /* If the desired type is compressed, then do our own thing. */
  if( *type == ZEPHRT_COMPRESSED) {
    if( *orient != ORIENT_RD) {
      *orient = ORIENT_RD;
      *cantDo |= ZEPH_CANT_DO_ORIENT;
    }
    if( !(DgpCoord_equal( dstRes->xRes, imageInfo->res.xRes)
          && DgpCoord_equal( dstRes->yRes, imageInfo->res.yRes))) {
      dstRes->xRes = imageInfo->res.xRes;
      dstRes->yRes = imageInfo->res.yRes;
      *cantDo |= ZEPH_CANT_DO_RESOLUTION;
    }

    /* Find the image strip that contains (dstX,dstY). This is actually
     * pretty easy since we already required that there is no rotation and
     * no scale. */
    strip1 = *dstY / imageInfo->stripHeight;

    /* Now we require that we only read that strip-- no more, no less.
     * Check if this is the last strip. */
    if( strip1 == _priv->stripCount - 1) {
      /* The strip might not be the full strip height, so check. */
      Dgp_Int32 height = (Dgp_Int32)(imageInfo->height - 
                             strip1 * imageInfo->stripHeight);

      if( !((*dstX == 0) && (*dstY == (Dgp_Int32)(strip1 * imageInfo->stripHeight))
            && (*dstW == (Dgp_Int32)imageInfo->width) && (*dstH == (Dgp_Int32)height))) {
        *dstX = 0;
        *dstY = strip1 * imageInfo->stripHeight;
        *dstW = imageInfo->width;
        *dstH = height;
        *cantDo |= ZEPH_CANT_DO_RECT;
      }
    }
    else {
      /* If it's not the last strip then it must be full height. */
      if( !((*dstX == 0) && (*dstY == (Dgp_Int32)(strip1 * imageInfo->stripHeight))
           && (*dstW == (Dgp_Int32)imageInfo->width)
           && (*dstH == (Dgp_Int32)imageInfo->stripHeight))) {
        *dstX = 0;
        *dstY = strip1 * imageInfo->stripHeight;
        *dstW = imageInfo->width;
        *dstH = imageInfo->stripHeight;
        *cantDo |= ZEPH_CANT_DO_RECT;
      }
    }

    if( *maxMemRequired < sizeof( LazyReadState)) {
      *maxMemRequired = sizeof( LazyReadState);
      *cantDo |= ZEPH_CANT_DO_MEMORY;
    }

    /* We're outta here. */
    NORMAL_RETURN;
  } 

  /* Okay, so it's not a compressed read. Now figure out how many strips
   * the desired region intersects. Then negotiate each intersection with the
   * decoder and find the "greatest common denominator" between the different
   * intersections.
   *
   * strip1 is the first strip (starting with 0) that intersects the desired
   * region of the destination. strip2 is the last strip that does so. So we
   * have to deal with strips from strip1 to strip2, inclusive.
   */
  /* FIXME scaling */
  /* Scaled strip height will typically be higher than nominal since we'll be */
  /* reading an artificially downsampled jpeg. / dhtesq. */
  scaled_strip_height = (Dgp_Int32)(imageInfo->stripHeight * scale_factor);

  strip1 = *dstY / scaled_strip_height;
  strip2 = (*dstY + *dstH - 1) / scaled_strip_height;
  /*strip2 = 0; */

  /* Set up the defaults. */
  currentType = originalType = *type;
  currentOrient = originalOrient = *orient;
  currentDstRes.xRes = dstRes->xRes;
  currentDstRes.yRes = dstRes->yRes;
  currentDstX = *dstX;
  currentDstW = *dstW;

  if( *maxMemRequired < sizeof( LazyReadState)) {
    *maxMemRequired = sizeof( LazyReadState);
    *cantDo |= ZEPH_CANT_DO_MEMORY;
  }
  /* Since each strip will be rendered sequentially, we only need to keep
   * track of the _largest_ amount of memory need for any one strip. */
  currentMemRequired = originalMemRequired
                         = (*maxMemRequired - sizeof( LazyReadState));
  currentCantDo = 0;

  /* So. We going to go through every strip at least once. If the decoder
   * doesn't complain about the rectangle, then we do it exactly once.
   * In this case, use "&" to find the GCD for type and orient; if the decoder
   * gets wishy-washy about resolution then just default to native; required
   * memory for any one strip is always just the max.
   *
   * If we have problems with the rectangle, then we have problems. We need
   * to keep iterating until we find a width that the decoder can decode every
   * strip into. So what we do is just keep widening the rectangle until
   * everyone is happy.
   */
  for(i = strip1, rectOkay = DGP_TRUE; i <= strip2; i++, currentCantDo = 0) {
    
    /* Set up y-coordinate and the height. These depend on what strip we
     * are currently working on and how it relates to the output rectangle.
     * We have four possibilities.
     *   1) The entire rect lies within a single strip. (i == strip1 == strip2)
     *   2) The rectangle starts in the current strip and extends beyond
     *      it to at least one other strip. (i == strip1 && i != strip2)
     *   3) The rectangle ends in the current strip but began in a previous
     *      strip above the current one. (i != strip! && i == strip2)
     *   4) The rectangle neither starts nor stops in the current strip.
     *      (i != strip1 && i != strip2) 
     */
    /* FIXME dstMinX, dstMinY, srcOriginX, srcOriginY */
    if( i == strip1) {
      currentDstY = *dstY - (i * scaled_strip_height);  
      if( i == strip2) {
        currentDstH = *dstH;
      }
      else {
        currentDstH = ((i + 1) * scaled_strip_height) - *dstY;
      }
    }
    else {
      currentDstY = 0;
      if( i == strip2) {
        currentDstH = (*dstY + *dstH) - (i * scaled_strip_height);
      }
      else {
        currentDstH = scaled_strip_height;
      }
    }
    
    DGP_ASSERT( currentDstY + currentDstH <= (Dgp_Int32)scaled_strip_height);
    
    DGP_ASSERT( ((i * (Dgp_Int32)scaled_strip_height) + currentDstY + currentDstH) 
                  <= (imageInfo->height * scale_factor));
    

    /* Okay, setup is done. Now call the decoder abilities function for this
     * strip. Let's get an opinion. */
    switch( imageInfo->compType)
    {
    case IO_COMP_JPEG:
DGP_ERR_PASS_CLEANUP(
      zephJpegOps.abilities( _priv->decodeState[ i],
                             dstMinX, dstMinY,
                             &currentType,
                             &currentOrient,
                             &currentDstRes,
                             &currentDstX, &currentDstY,
                             &currentDstW, &currentDstH,
                             &currentMemRequired,
                             srcOriginX, srcOriginY,
                             &currentCantDo));
      break;
   
    case IO_COMP_MMR:
DGP_ERR_PASS_CLEANUP(
      zephG4Ops.abilities( _priv->decodeState[ i],
                           dstMinX, dstMinY,
                           &currentType,
                           &currentOrient,
                           &currentDstRes,
                           &currentDstX, &currentDstY,
                           &currentDstW, &currentDstH,
                           &currentMemRequired,
                           srcOriginX, srcOriginY,
                           &currentCantDo));
      break;
    
    /* Cases added by Eli */
    case IO_COMP_MH:
DGP_ERR_PASS_CLEANUP(
      zephG3_1DOps.abilities( _priv->decodeState[ i],
                           dstMinX, dstMinY,
                           &currentType,
                           &currentOrient,
                           &currentDstRes,
                           &currentDstX, &currentDstY,
                           &currentDstW, &currentDstH,
                           &currentMemRequired,
                           srcOriginX, srcOriginY,
                           &currentCantDo));
      break;
    
    case IO_COMP_MR:
DGP_ERR_PASS_CLEANUP(
      zephG3_2DOps.abilities( _priv->decodeState[ i],
                           dstMinX, dstMinY,
                           &currentType,
                           &currentOrient,
                           &currentDstRes,
                           &currentDstX, &currentDstY,
                           &currentDstW, &currentDstH,
                           &currentMemRequired,
                           srcOriginX, srcOriginY,
                           &currentCantDo));
      break;
    /* Cases added by Eli */
    
    case IO_COMP_TOKENIZATION:
    case IO_COMP_LLTOKENIZATION:
      DGP_ERR_CLEANUPF0( cErrDecodeFailure, 
              "Cannot read tokenized images. Use the zephtokenize interface.");

      break;
    default:
      DGP_ERR_CLEANUPF0( cErrDecodeFailure, "Unknown compression type."); 
    }
    
    /* Right. Now if something isn't okay, figure out what and do the right
     * thing. */
    if( currentCantDo > 0) {
      
      /* TYPE */
      if( (currentCantDo & ZEPH_CANT_DO_TYPE) > 0) {
        if( (*cantDo & ZEPH_CANT_DO_TYPE) > 0) {
          *type &= currentType;
          currentType = originalType;
          DGP_ASSERT( *type > 0);
        }
        else {
          *type = currentType;
          *cantDo |= ZEPH_CANT_DO_TYPE;
        }
      }
      
      /* ORIENT */
      if( (currentCantDo & ZEPH_CANT_DO_ORIENT) > 0) {
        if( (*cantDo & ZEPH_CANT_DO_ORIENT) > 0) {
          *orient &= currentOrient;
          currentOrient = originalOrient;
          DGP_ASSERT( *orient > 0);
        }
        else {
          *orient = currentOrient;
          *cantDo |= ZEPH_CANT_DO_ORIENT;
        }
      }

      /* RESOLUTION */
      if( (currentCantDo & ZEPH_CANT_DO_RESOLUTION) > 0) {
        if( (*cantDo & ZEPH_CANT_DO_RESOLUTION) > 0) {
          /* If this strip and the previous strip (that failed) don't agree
           * on what resolution they can handle, then just use the native
           * image resolution, since the decoder _must_ handle this case. */
          if( !(DgpCoord_equal( dstRes->xRes, currentDstRes.xRes)
                && DgpCoord_equal( dstRes->yRes, currentDstRes.yRes))) {
            dstRes->xRes = imageInfo->res.xRes;
            dstRes->yRes = imageInfo->res.yRes;
          }
          /* else They are agree on what they can do. Good. */
        }
        else {
          dstRes->xRes = currentDstRes.xRes;
          dstRes->yRes = currentDstRes.yRes;
          *cantDo |= ZEPH_CANT_DO_RESOLUTION;
        }
      }

      /* RECT */
      if( (currentCantDo & ZEPH_CANT_DO_RECT) > 0) {
        /* We don't care if someone else already complained, we will too. */
        DGP_ASSERT( currentDstX <= *dstX);
        *dstX = currentDstX;

        DGP_ASSERT( currentDstW >= *dstW);
        *dstW = currentDstW;

        if( i == strip1) {

          if( i == strip2) {
            DGP_ASSERT( currentDstH >= *dstH);
            *dstH = currentDstH;
          }
          /* */

          else {
            /* Here we assert that the decoder hasn't moved the lower bound
             * for this strip. */
            DGP_ASSERT( currentDstH == (Dgp_Int32)(((i + 1)* imageInfo->stripHeight)
                           - (currentDstY + (i * imageInfo->stripHeight))));
            /* Now, since the lower bound of the desired rectangle is on 
             * another strip, we must extend the height by the change in
             * dstY (always relative to the current strip). */
            *dstH += (*dstY - (i * imageInfo->stripHeight)) - currentDstY;
          }
          DGP_ASSERT( currentDstY <= (Dgp_Int32)(*dstY 
                                              - (i * imageInfo->stripHeight)));
          *dstY = currentDstY + (i * imageInfo->stripHeight);
        }
        else {
          DGP_ASSERT( currentDstY == 0);
          if( i == strip2) {
            DGP_ASSERT( currentDstH >= (Dgp_Int32)((*dstY + *dstH) 
                                              - (i * imageInfo->stripHeight)));
            *dstH = currentDstH + (i * imageInfo->stripHeight) - *dstY;
          }
          else {
            DGP_ASSERT( currentDstH == (Dgp_Int32)imageInfo->stripHeight);
          }
        }

        /* Set this flag so we run through all the strips again. */
        rectOkay = DGP_FALSE;
        *cantDo |= ZEPH_CANT_DO_RECT;
      }

      /* MEMORY */
      if( (currentCantDo & ZEPH_CANT_DO_MEMORY) > 0) {
        *maxMemRequired = currentMemRequired + sizeof( LazyReadState);
        *cantDo |= ZEPH_CANT_DO_MEMORY;
      }
    } /* end if( currentCantDo > 0) { */

    /* After the last strip, make sure everything is cool. */
    if( i == strip2 && !rectOkay && strip1 != strip2) {
      /* Someone is making trouble about the rectangle, so we have to 
       * do it again. */
      i = strip1 - 1;
      rectOkay = DGP_TRUE;
    }

  } /* end for */

  /* Add ability to read compressed. */
  if( (*cantDo & ZEPH_CANT_DO_TYPE) > 0) {
    *type |= ZEPHRT_COMPRESSED;
  }
  
  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}


ErrCode
zephReadImageStart( zephImageInfo imageInfo, 
                       void *dst, Dgp_Int32 dstBpl,
                       void *trans_src,
                       Dgp_Int32 trans_bpl,
                       Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
                       zephReadType type,
                       DgpCoord_Orientation orient,
                       DgpCoord_Resolution *dstRes, 
                       Dgp_Int32 dstX, Dgp_Int32 dstY,
                       Dgp_Int32 dstW, Dgp_Int32 dstH,
                       Dgp_UInt32 *maxMemRequired,
                       Dgp_Int32 srcOriginX, Dgp_Int32 srcOriginY,
                       zephReadState *pReadState)
{
  LazyReadState *lazy = (LazyReadState *)NULL;
  /*Dgp_UInt8 cantDo = 0; */

ZEPH_PROLOGUE_IMAGE("zephReadImageStart()", imageInfo);

  /* Make sure that nobody is trying to pull anything on us. */
  /* ...
     Since we've already checked the abilities at this point, I'm not sure
     what this is for.  At any rate, the initialization call into the jpeg
     library confuses the file pointer -- the jpeg code assumes 4kB of the file
     has already been read when it starts decoding, but in fact the file offset
     will have been set to the start of the jpeg data -- bad.  
     
     It would certainly be safer to allow multiple readAbilities calls into the
     jpeg library,...future fix.

  
  DGP_ASSERT( (zephReadImageAbilities( imageInfo, dstMinX, dstMinY, &type,
                                       &orient, dstRes, &dstX, &dstY, 
                                       &dstW, &dstH, maxMemRequired,
                                       srcOriginX, srcOriginY, &cantDo),
               cantDo == 0));
   */
  

DGP_CHECK_ALLOC_CLEANUP(
  lazy = DGP_MALLOC( sizeof( *lazy)));

  lazy->dst = dst;
  lazy->dstBpl = dstBpl;
  lazy->trans_src = trans_src;
  lazy->trans_bpl = trans_bpl;
  lazy->dstMinX = dstMinX;
  lazy->dstMinY = dstMinY;
  lazy->type = type;
  lazy->orient = orient;
  lazy->dstRes = dstRes;
  lazy->dstX = dstX;
  lazy->dstY = dstY;
  lazy->dstW = dstW;
  lazy->dstH = dstH;
  lazy->maxMemRequired = *maxMemRequired;
  lazy->srcOriginX = srcOriginX;
  lazy->srcOriginY = srcOriginY;

  *pReadState = (zephReadState)lazy;
  NORMAL_RETURN;
  
ERR_CLEANUP:
  if( lazy != (LazyReadState *)NULL) {
    DGP_FREE( (void *)lazy);
  }

  ERR_RETURN;
}

ErrCode
zephReadImageEnd( zephImageInfo imageInfo, 
                       zephReadState *pReadState,
                       Dgp_UInt32 *maxMemRequired,
                       DgpCoord_Resolution *rres,
                       Dgp_Int32 nominal_page_width,
                       Dgp_Int32 nominal_page_height,
                       Dgp_Int32 nominal_strip_width,
                       Dgp_Int32 nominal_strip_height)
{
  LazyReadState *lazy = (LazyReadState *)*pReadState;
  _zephPrivateImageInfo *_priv;
  _zephPrivateDocumentInfo *_privDoc;
  Dgp_UInt32 i, strip1, strip2;
  void *currentDst = NULL;
  Dgp_Int32 currentDstY, currentDstH;
  Dgp_UInt32 currentMemRequired;
  Dgp_Int32 scaled_strip_height;
  Dgp_Float32 scale_factor;
  /*uds *u; */
  void *cmyk[4];

ZEPH_PROLOGUE_IMAGE("zephReadImageEnd()", imageInfo);
 
DGP_ERR_GEN( pReadState == (zephReadState *)NULL, cErrInvalidParm);

  _priv = (_zephPrivateImageInfo *)imageInfo->privateInfo;
  _privDoc = (_zephPrivateDocumentInfo *)((_zephPrivatePageInfo *)
                (_priv->pageInfo->privateInfo))->docInfo->privateInfo;
  
  if( lazy->type == ZEPHRT_COMPRESSED) {
    strip1 = (lazy->dstH - 1) / imageInfo->stripHeight;       /* ...was lazy->dstY. 27.9.2000 / dhtesq. */

	/* 5.3.2001 / dhtesq. */
	/* Can we do compressed reads without this block? */
	/*
DGP_ERR_GEN_CLEANUP(
    u->fseek( u->fp, _priv->fileOffsets[ strip1], SEEK_SET),
      cErrIO_SeekFailure);

DGP_ERR_GEN_CLEANUP(
    u->fread( (void *)lazy->dst, _priv->stripByteCounts[ strip1], 
              1, u->fp) != 1, 
      cErrIO_ReadError);
	  */
    
    /*DGP_FREE( (void *)*pReadState); */
    /**pReadState = (zephReadState)NULL; */
    NORMAL_RETURN;
  }

  /* Okay, so it's not a compressed read. We have more work to do then. */
  /* */
  /* 12.6.2000 / dhtesq. */
  /* Trying to fix the scaling... */
  scale_factor = DgpCoord_float(rres->xRes) / DgpCoord_float(imageInfo->res.xRes);
  scaled_strip_height = (Dgp_Int32)((Dgp_Float32)imageInfo->stripHeight * scale_factor);

  strip1 = lazy->dstY / scaled_strip_height;
  strip2 = (lazy->dstY + lazy->dstH - 1) / scaled_strip_height;

  DGP_ASSERT( lazy->maxMemRequired >= sizeof( LazyReadState));
  currentMemRequired = (lazy->maxMemRequired - sizeof( LazyReadState));

  for( i = strip1; i <= strip2; i++) {

    /* FIXME dstMinX, dstMinY, srcOriginX, srcOriginY and scaling */
    if( i == strip1) {
      currentDstY = lazy->dstY - (i * scaled_strip_height);  
      if( i == strip2) {
        currentDstH = lazy->dstH;
      }
      else {
        currentDstH = ((i + 1) * scaled_strip_height) - lazy->dstY;
      }
    }
    else {
      currentDstY = 0;
      if( i == strip2) {
        currentDstH = (lazy->dstY + lazy->dstH) - (i * scaled_strip_height);
      }
      else {
        currentDstH = scaled_strip_height;
      }
    }

    /* Find the destination. */
    /* FIXME this will also need serious help to handle scaling */
    switch( lazy->type) {
    case ZEPHRT_MONO_PACKED:
    case ZEPHRT_MONO_8BIT_UNPACKED:
    case ZEPHRT_RGB_INTERLEAVED:
      currentDst = (void *)((Dgp_UInt8 *)lazy->dst 
                + (((currentDstY + (i * scaled_strip_height)) - lazy->dstY) 
                    * lazy->dstBpl));
      break;

    case ZEPHRT_CYMK_1BIT_SEPARATIONS:
      /* FIXME */
    case ZEPHRT_CMYK_8BIT_SEPARATIONS:
        /* 5.9.2000 / dhtesq.  Implementing... */
        cmyk[0] = (void *)(((Dgp_UInt8 **)lazy->dst)[0] + (((currentDstY + (i * scaled_strip_height)) - lazy->dstY) * lazy->dstBpl));
        cmyk[1] = (void *)(((Dgp_UInt8 **)lazy->dst)[1] + (((currentDstY + (i * scaled_strip_height)) - lazy->dstY) * lazy->dstBpl));
        cmyk[2] = (void *)(((Dgp_UInt8 **)lazy->dst)[2] + (((currentDstY + (i * scaled_strip_height)) - lazy->dstY) * lazy->dstBpl));
        cmyk[3] = (void *)(((Dgp_UInt8 **)lazy->dst)[3] + (((currentDstY + (i * scaled_strip_height)) - lazy->dstY) * lazy->dstBpl));
        currentDst = (void *)cmyk;
      break;
    }
      
	   DGP_ASSERT( currentDstY + currentDstH <= scaled_strip_height);
		DGP_ASSERT( ((i * scaled_strip_height) + currentDstY + currentDstH) 
                  <= (imageInfo->height * scale_factor));

    switch( imageInfo->compType)
    {
    case IO_COMP_JPEG:
      
DGP_ERR_CHECK_NO_MSG(
      zephJpegOps.read( _priv->decodeState[ i], 
                        currentDst, lazy->dstBpl,
                        lazy->dstMinX, lazy->dstMinY, 
                        lazy->type, lazy->orient, lazy->dstRes,
                        lazy->dstX, currentDstY, lazy->dstW, currentDstH,
                        &currentMemRequired, 
                        lazy->srcOriginX, lazy->srcOriginY,
                        lazy->trans_src,
                        lazy->trans_bpl,
                        nominal_page_width, nominal_page_height,
                        nominal_strip_width, nominal_strip_height))
      {
        if( DGP_ERROR_IS( cErrInsufficientMem)) {
          *maxMemRequired = currentMemRequired + sizeof( LazyReadState);
          DGP_ERR_CLEANUP( cErrInsufficientMem);
        }
        else {
          DGP_GOTO_CLEANUP;
        }
      }
      break;

    case IO_COMP_MMR:

DGP_ERR_CHECK_NO_MSG(
      zephG4Ops.read( _priv->decodeState[ i],
                      currentDst, lazy->dstBpl,
                      lazy->dstMinX, lazy->dstMinY, 
                      lazy->type, lazy->orient, lazy->dstRes,
                      lazy->dstX, currentDstY, lazy->dstW, currentDstH,
                      &currentMemRequired, 
                      lazy->srcOriginX, lazy->srcOriginY,
                      lazy->trans_src,
                      lazy->trans_bpl,
                      nominal_page_width, nominal_page_height,
                      nominal_strip_width, nominal_strip_height))

      {
        if( DGP_ERROR_IS( cErrInsufficientMem)) {
          *maxMemRequired = currentMemRequired + sizeof( LazyReadState);
          DGP_ERR_CLEANUP( cErrInsufficientMem);
        }
        else {
          DGP_GOTO_CLEANUP;
        }
      }
      break;

    /* Cases added by Eli */
    case IO_COMP_MH:

DGP_ERR_CHECK_NO_MSG(
      zephG3_1DOps.read( _priv->decodeState[ i],
                      currentDst, lazy->dstBpl,
                      lazy->dstMinX, lazy->dstMinY, 
                      lazy->type, lazy->orient, lazy->dstRes,
                      lazy->dstX, currentDstY, lazy->dstW, currentDstH,
                      &currentMemRequired, 
                      lazy->srcOriginX, lazy->srcOriginY,
                      lazy->trans_src,
                      lazy->trans_bpl,
                      nominal_page_width, nominal_page_height,
                      nominal_strip_width, nominal_strip_height))
      {
        if( DGP_ERROR_IS( cErrInsufficientMem)) {
          *maxMemRequired = currentMemRequired + sizeof( LazyReadState);
          DGP_ERR_CLEANUP( cErrInsufficientMem);
        }
        else {
          DGP_GOTO_CLEANUP;
        }
      }
      break;

    case IO_COMP_MR:

DGP_ERR_CHECK_NO_MSG(
      zephG3_2DOps.read( _priv->decodeState[ i],
                      currentDst, lazy->dstBpl,
                      lazy->dstMinX, lazy->dstMinY, 
                      lazy->type, lazy->orient, lazy->dstRes,
                      lazy->dstX, currentDstY, lazy->dstW, currentDstH,
                      &currentMemRequired, 
                      lazy->srcOriginX, lazy->srcOriginY,
                      lazy->trans_src,
                      lazy->trans_bpl,
                      nominal_page_width, nominal_page_height,
                      nominal_strip_width, nominal_strip_height))
      {
        if( DGP_ERROR_IS( cErrInsufficientMem)) {
          *maxMemRequired = currentMemRequired + sizeof( LazyReadState);
          DGP_ERR_CLEANUP( cErrInsufficientMem);
        }
        else {
          DGP_GOTO_CLEANUP;
        }
      }
      break;
    /* Cases added by Eli */

    case IO_COMP_TOKENIZATION:
    case IO_COMP_LLTOKENIZATION:
      DGP_ERR_CLEANUPF0( cErrDecodeFailure, 
              "Cannot read tokenized images. Use the zephtokenize interface.");

      break;

    default:
      DGP_ERR_CLEANUPF0( cErrDecodeFailure, "Unknown compression type.");
    }

  } /* end for */

  DGP_FREE( (void *)*pReadState);
  *pReadState = (zephReadState)NULL;
  NORMAL_RETURN;

ERR_CLEANUP:

  if( *pReadState != (zephReadState)NULL) {
    DGP_FREE( (void *)*pReadState);
  }
  
  ERR_RETURN;
}


ErrCode
zephGetDictionaryInfo( zephImageInfo imageInfo, 
                          zephDictionaryInfo *pDictInfo)
{
  _zephPrivateImageInfo *_priv;
  _zephPrivatePageInfo *_privPage;
  
ZEPH_PROLOGUE_IMAGE( "zephGetDictionaryInfo()", imageInfo);

  _priv = (_zephPrivateImageInfo *)imageInfo->privateInfo;
  _privPage = (_zephPrivatePageInfo *)_priv->pageInfo->privateInfo;

DGP_ERR_PASS(
  zephOpenDictionary( _privPage->docInfo, imageInfo->libCount, 
                      imageInfo->libId, pDictInfo));

  NORMAL_RETURN;
}


ErrCode
zephCalcPagePosition( zephImageInfo imageInfo, Dgp_Int32 *pX, Dgp_Int32 *pY, 
                       Dgp_Int32 *pW, Dgp_Int32 *pH)
{
  zephPageInfo page;
ZEPH_PROLOGUE_IMAGE( "zephCalcPagePosition()", imageInfo);

  page = ((_zephPrivateImageInfo *)imageInfo->privateInfo)->pageInfo;
  
  *pX = imageInfo->xPos;
  *pY = imageInfo->yPos;

  /* 12.3.2001 / dhtesq. */
  /* Adjusted these formulas -- weren't returning correct dimensions of subsampled images. */
  *pW = DgpCoord_floor( DgpCoord_muli( DgpCoord_div( page->res.xRes, 
                                      imageInfo->res.xRes),
                             (imageInfo->width)));
  *pH = DgpCoord_floor( DgpCoord_muli( DgpCoord_div( page->res.yRes, 
                                      imageInfo->res.yRes),
                             (imageInfo->height)));

  NORMAL_RETURN;
}
