/* $Header: /project/icons/cvsroot/digipaper/src/libzephyr/jpeg.c,v 1.19 2003/02/10 22:53:46 mkrocken Exp $ */

#include "jpeg.h"

#define NEED_UNIQUE_EXTERNAL_NAMES
#include "jpeg6blib.h"
#include "jpeg_interface.h"
#include "intellibuffer.h"

#ifdef ZEPH_INCLUDE_JPEG

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libzephyr/jpeg.c,v 1.19 2003/02/10 22:53:46 mkrocken Exp $");


typedef struct {
  Dgp_State *dgp_state;
  uds u;
  IO_IMGHANDLE image_handle;
  Dgp_UInt32 compSize;

  Dgp_Int32 srcW;
  Dgp_Int32 srcH;
  DgpCoord_Resolution *srcRes;
  Dgp_Bool mask;

  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr err;

  Dgp_Int32 mem_estimate;
  Dgp_Int16 color_space;

} zephJpegState;


ErrCode
zephJpegOpen( Dgp_State *dgp_state,
                       zephImageInfo image, 
                       Dgp_Int32 srcW, Dgp_Int32 srcH,
                       DgpCoord_Resolution *srcRes, 
                       void **decodeState)
{
  zephJpegState *jstate = (zephJpegState *)NULL;
  Dgp_UInt32 is_valid;

#ifdef I_WANT_THE_JPEGS
  FILE *fp = NULL;
  long current_pos;
  char *the_compressed_jpeg;
#endif

DGP_PROLOGUE( "zephJpegOpen()", (void *)dgp_state);

DGP_CHECK_ALLOC_CLEANUP( 
  jstate = (zephJpegState *)DGP_MALLOC( sizeof( *jstate)));

  jstate->dgp_state = dgp_state; 
  jstate->image_handle = ((_zephPrivateImageInfo*) image->privateInfo)->IO_image_handle;
  jstate->srcW = srcW;
  jstate->srcH = srcH;
  jstate->srcRes = srcRes;
  jstate->mem_estimate = -1;
  jstate->compSize = image->compSize;
  jstate->u.fp = NULL;
  if (image->maskInfo)
    jstate->mask = DGP_TRUE;
  else
    jstate->mask = DGP_FALSE;
    

  jstate->cinfo.err = jpeg_std_error( &jstate->err);
  jstate->cinfo.mem = NULL;
  jstate->cinfo.marker = NULL;
  jstate->cinfo.inputctl = NULL;
  DGP_ERR_PASS ( DGP_JPEG_init_jdecompress ( &jstate->cinfo, dgp_state));

  /* find out the color space of the compressed Jpeg data */
  DGP_ERR_PASS_CLEANUP(IO_GetPhotometricInterpretation ( jstate->image_handle,
	  &(jstate->color_space), &is_valid));
  if (!is_valid) jstate->color_space = 0;

  *decodeState = (void *)jstate;

#ifdef I_WANT_THE_JPEGS
  fp = fopen("d:\\testbed\\test.jpg", "wb");
  if(!fp){
      printf("I couldn't open the stupid file...\n");
  }
  the_compressed_jpeg = (char *)DGP_MALLOC(compSize);
  current_pos = u->ftell(u->fp);
  u->fseek(u->fp, offset, SEEK_SET);
  u->fread(the_compressed_jpeg, compSize, 1, u->fp);
  u->fseek(u->fp, current_pos, SEEK_SET);
  fwrite((void *)the_compressed_jpeg, compSize, 1, fp);
  fclose(fp);
  DGP_FREE(the_compressed_jpeg);
#endif

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

Dgp_UInt32
zephJpegUsage( void *decodeState)
{
  zephJpegState *jstate = (zephJpegState *)decodeState;

  return jstate->mem_estimate;
}

ErrCode
zephJpegAbilities( void *decodeState,
                       Dgp_Int32 dstMinX, Dgp_Int32 dstMinY,
                       zephReadType *type,
                       DgpCoord_Orientation *orient,
                       DgpCoord_Resolution *dstRes,
                       Dgp_Int32 *dstX, Dgp_Int32 *dstY,
                       Dgp_Int32 *dstW, Dgp_Int32 *dstH,
                       Dgp_UInt32 *maxMemRequired,
                       Dgp_Int32 srcOriginX, Dgp_Int32 srcOriginY,
                       Dgp_UInt8 *cantDo)

{
  zephJpegState *jstate = (zephJpegState *)decodeState;
  Dgp_State *dgp_state;
  DGP_JPEGImageDef image_def;
  Dgp_UInt32 result;
  Dgp_UInt8* buf;
  Dgp_UInt32 buf_size = 0, bytesFilled;

DGP_PROLOGUE( "zephJpegAbilities()", NULL);
DGP_ERR_GEN( jstate == (zephJpegState *)NULL, cErrInvalidParm);
  dgp_state = jstate->dgp_state;
DGP_ERR_BEGIN( dgp_state);

  *maxMemRequired = *maxMemRequired; 

  image_def.imgCoOrds.dstMinX = dstMinX;
  image_def.imgCoOrds.dstMinY = dstMinY;

  /*
  RR seems to have worked in page coordinates strictly, whereas
  WB's dsts are in image coordinates.  Try converting to page.
  / dhtesq.
  
  image_def.imgCoOrds.dstX = *dstX;
  image_def.imgCoOrds.dstY = *dstY;
  */
  image_def.imgCoOrds.dstX = srcOriginX + *dstX;
  image_def.imgCoOrds.dstY = srcOriginY + *dstY;

  image_def.imgCoOrds.srcOriginX = srcOriginX;
  image_def.imgCoOrds.srcOriginY = srcOriginY;

  image_def.imgCoOrds.dstW = *dstW;
  image_def.imgCoOrds.dstH = *dstH;

  /* These two don't really have meaning. */
  image_def.imgCoOrds.bytesPerRow = *dstW;
  image_def.imgCoOrds.numRows = jstate->srcH;

  image_def.imgResolution.srcRes.xRes = jstate->srcRes->xRes;
  image_def.imgResolution.srcRes.yRes = jstate->srcRes->yRes;

  image_def.imgResolution.destRes.xRes = dstRes->xRes;
  image_def.imgResolution.destRes.yRes = dstRes->yRes; 

  /*
  image_def.imgResolution.destRes.xRes = DgpCoord_eqi( 100);
  image_def.imgResolution.destRes.yRes = DgpCoord_eqi( 100); */

  if( *type == ZEPHRT_MONO_PACKED) {
    image_def.imgType = DGP_PACKED;
  }
  else if( *type == ZEPHRT_CMYK_8BIT_SEPARATIONS) {
    image_def.imgType = DGP_CMYK;
  }
  else if( *type == ZEPHRT_RGB_INTERLEAVED) {
    image_def.imgType = DGP_RGB;
  }
  else {
    *type = ZEPHRT_MONO_PACKED | ZEPHRT_CMYK_8BIT_SEPARATIONS 
      | ZEPHRT_RGB_INTERLEAVED;
    *cantDo |= ZEPH_CANT_DO_TYPE;
    NORMAL_RETURN;
  }

  if (jstate->image_handle)
  {
  /* read the compressed JPEG block */
  DGP_ERR_PASS_CLEANUP(
		       IO_ImageReadStart(jstate->image_handle, 0 /* Flags */));
  
  buf_size = jstate->compSize;
  DGP_CHECK_ALLOC_CLEANUP(
			    buf = DGP_MALLOC (buf_size));
  DGP_ERR_PASS_CLEANUP(
		 IO_CompressedReadStrip (jstate->image_handle,
					       0, /* row offset */
					       jstate->srcH,
					       &buf,
					       &buf_size,
					       &bytesFilled));

  /* close the I/O handle to indicate that we read the data */
  IO_ImageReadFinish (jstate->image_handle);
  jstate->image_handle = 0;

  if (bytesFilled > 0)
  {

  DGP_ERR_PASS_CLEANUP(
    DgpIbuf_mopen((void**) &jstate->u.fp, 
	      buf, buf_size, DGPIBUF_READ, dgp_state));

  jstate->u.fread = (fread_func) DgpIbuf_fread;
  jstate->u.fseek = (fseek_func) DgpIbuf_fseek;
  jstate->u.ftell = (ftell_func) DgpIbuf_ftell;
  jstate->u.fwrite = (fwrite_func) NULL;
  jstate->compSize = buf_size;
  }
  else
  {
	  jstate->u.fp = NULL;
  }
  }
  else
  {
    /* we already read the buffer, so just set the file pointer back to the origin */
    jstate->u.fseek (jstate->u.fp, 0, SEEK_SET);
  }


  image_def.imgSrc.u = &jstate->u;
  image_def.imgSrc.size = jstate->compSize;

  /*
  20.6.2000 / dhtesq.
  When I came across these, they did not correspond with the definitions in
  orientation_t.h
  
  switch( *orient) {
  case 0x00:
    image_def.imgOrientation = ORIENT_RD; break;
  case 0x01:
    image_def.imgOrientation = ORIENT_UR; break;
  case 0x02:
    image_def.imgOrientation = ORIENT_RU; break;
  case 0x04:
    image_def.imgOrientation = ORIENT_LU; break;
  case 0x08:
    image_def.imgOrientation = ORIENT_DL; break;
  case 0x10:
    image_def.imgOrientation = ORIENT_DR; break;
  case 0x20:
    image_def.imgOrientation = ORIENT_UL; break;
  case 0x40:
    image_def.imgOrientation = ORIENT_LD; break;
  default:
    DGP_ERRF0( cErrInvalidParm, "Desired orientation must be a single value.");
  }
  */
  /* Try this instead...keep in mind libjpeg6b (it appears) cannot do  */
  /* "inverting" orientations like ORIENT_DR. / dhtesq. */
  image_def.imgOrientation = *orient;

  /*
DGP_ERR_GEN_CLEANUP(
    jstate->u->fseek( jstate->u->fp, jstate->offset, SEEK_SET), 
    cErrIO_SeekFailure); */

  if (jstate->u.fp)
  {
  DGP_ERR_PASS (DGP_JPEG_check_jdecompress (&image_def, &result, &jstate->cinfo, 
					   dgp_state));
  jstate->mem_estimate = image_def.memData.mem_estimate;
  }
  else
  {
	  result = 0;
	  jstate->mem_estimate = 0;
  }


  if( result > 0) {
    if( (result & DGP_IMAGE_TYPE_BIT) > 0) {
      switch( image_def.imgType) {
      case DGP_PACKED:
        *type = ZEPHRT_MONO_PACKED;
        break;
      case DGP_CMYK:
        *type = ZEPHRT_CMYK_8BIT_SEPARATIONS;
        break;
      case DGP_RGB:
        *type = ZEPHRT_RGB_INTERLEAVED;
        break;
      default:
        DGP_ERR_CLEANUPF0( cErrDecodeFailure,
                           "No possible render type.");
      }

      *cantDo |= ZEPH_CANT_DO_TYPE;
    }

    if( (result & DGP_ORIENTATION_BIT) > 0) {
      switch( image_def.imgOrientation) {
      case ORIENT_RD:
        *orient = 0x00; break;
      case ORIENT_UR:
        *orient = 0x01; break;
      case ORIENT_RU:
        *orient = 0x02; break;
      case ORIENT_LU:
        *orient = 0x04; break;
      case ORIENT_DL:
        *orient = 0x08; break;
      case ORIENT_DR:
        *orient = 0x10; break;
      case ORIENT_UL:
        *orient = 0x20; break;
      case ORIENT_LD:
        *orient = 0x40; break;
      }

      *cantDo |= ZEPH_CANT_DO_ORIENT;
    }

    if( (result & DGP_RESOLUTION_BIT) > 0) {

      dstRes->xRes = image_def.imgResolution.destRes.xRes;
      dstRes->yRes = image_def.imgResolution.destRes.yRes;

      *cantDo |= ZEPH_CANT_DO_RESOLUTION;
    }

    if( (result & DGP_COORD_DIFF_BIT) > 0) {
      *dstX = image_def.imgCoOrds.dstX;
      *dstY = image_def.imgCoOrds.dstY;
      *dstW = image_def.imgCoOrds.dstW;
      *dstH = image_def.imgCoOrds.dstH;

      *cantDo |= ZEPH_CANT_DO_RECT;
    }

    if( (result & DGP_SZ_BIT) > 0) {
      DGP_ERR_CLEANUPF0( cErrDecodeFailure, 
                       "Render abilities failed because of 'BYTES_PER_ROW'.");
    }

    NORMAL_RETURN;
  }
  
  /* We made it!! */

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;
}

#ifdef I_WANT_TO_TIME_THE_JPEG
#include <time.h>
#endif

ErrCode
zephJpegRead(void		*decodeState, 
			 void		*dst, 
			 Dgp_Int32	dstBpl,
			 Dgp_Int32	dstMinX, Dgp_Int32 dstMinY,
			 zephReadType type,
			 DgpCoord_Orientation orient,
			 DgpCoord_Resolution *dstRes,
			 Dgp_Int32 dstX, Dgp_Int32 dstY,
			 Dgp_Int32 dstW, Dgp_Int32 dstH,
			 Dgp_UInt32 *maxMemRequired,
			 Dgp_Int32 srcOriginX, Dgp_Int32 srcOriginY,
			 void		*trans_src,
			 Dgp_Int32	trans_bpl,
			 Dgp_Int32	nominal_page_width, Dgp_Int32 nominal_page_height,
			 Dgp_Int32	nominal_strip_width, Dgp_Int32 nominal_strip_height)
{
  zephJpegState *jstate = (zephJpegState *)decodeState;
  Dgp_State *dgp_state;
  DGP_JPEGImageDef image_def;
  dgp_dest_buf dst_buf;

#ifdef I_WANT_TO_TIME_THE_JPEG
  clock_t start, finish;
  double duration;
#endif

DGP_PROLOGUE( "zephJpegRead()", NULL);
DGP_ERR_GEN( jstate == (zephJpegState *)NULL, cErrInvalidParm);
  dgp_state = jstate->dgp_state;
DGP_ERR_BEGIN( dgp_state);

  maxMemRequired = maxMemRequired;
 
  image_def.imgCoOrds.dstMinX = dstMinX;
  image_def.imgCoOrds.dstMinY = dstMinY;
  /*
  RR seems to have worked in page coordinates strictly, whereas
  WB's dsts are in image coordinates.  Try converting to page.
  Also have to stick in the nominal page size so jpeg library can reorient... / dhtesq.
  
  image_def.imgCoOrds.dstX = dstX;
  image_def.imgCoOrds.dstY = dstY;
  */

  image_def.imgCoOrds.dstX = srcOriginX + dstX;
  image_def.imgCoOrds.dstY = srcOriginY + dstY;

  image_def.imgCoOrds.srcOriginX = srcOriginX;
  image_def.imgCoOrds.srcOriginY = srcOriginY;

  /* Added... / dhtesq. */
  image_def.imgCoOrds.nominal_page_width = nominal_page_width;
  image_def.imgCoOrds.nominal_page_height = nominal_page_height;
  image_def.imgCoOrds.nominal_strip_width = nominal_strip_width;
  image_def.imgCoOrds.nominal_strip_height = nominal_strip_height;


  image_def.imgCoOrds.dstW = dstW;
  image_def.imgCoOrds.dstH = dstH;
  
  /* These two don't really have meaning. */
  image_def.imgCoOrds.bytesPerRow = dstBpl;
  image_def.imgCoOrds.numRows = jstate->srcH;

  /* Added these two so I could rotate coordinates in the jpeg library... / dhtesq. */
  image_def.imgResolution.srcRes.xRes = jstate->srcRes->xRes;
  image_def.imgResolution.srcRes.yRes = jstate->srcRes->yRes;

  image_def.imgResolution.destRes.xRes = dstRes->xRes;
  image_def.imgResolution.destRes.yRes = dstRes->yRes; 

  /*
  image_def.imgResolution.destRes.xRes = DgpCoord_eqi( 100);
  image_def.imgResolution.destRes.yRes = DgpCoord_eqi( 100); */

  if( type == ZEPHRT_MONO_PACKED) {
    image_def.imgType = DGP_PACKED;
    dst_buf.other_buf.buffer = (Dgp_Int8 *)dst;
    dst_buf.other_buf.trans_mask = trans_src;
    dst_buf.other_buf.trans_bpl = trans_bpl;
  }
  else if( type == ZEPHRT_CMYK_8BIT_SEPARATIONS) {
    image_def.imgType = DGP_CMYK;
    dst_buf.cmyk_buf.c = ((void **)dst)[0];
    dst_buf.cmyk_buf.m = ((void **)dst)[1];
    dst_buf.cmyk_buf.y = ((void **)dst)[2];
    dst_buf.cmyk_buf.k = ((void **)dst)[3];
    dst_buf.cmyk_buf.trans_mask = trans_src;
    dst_buf.cmyk_buf.trans_bpl = trans_bpl;
  }
  else if( type == ZEPHRT_RGB_INTERLEAVED) {
    image_def.imgType = DGP_RGB;
    dst_buf.other_buf.buffer = (Dgp_Int8 *)dst;
    dst_buf.other_buf.trans_mask = trans_src;
    dst_buf.other_buf.trans_bpl = trans_bpl;
  }
  else {
    DGP_ERR_CLEANUPF0( cErrDecodeFailure, "Unable to render given type.");
  }

  image_def.imgSrc.u = &jstate->u;
  image_def.imgSrc.size = jstate->compSize;

  /*
  20.6.2000 / dhtesq.
  Like in the readAbilities function, these orientations don't correspond
  to the definitions in orientation_t.h.

  switch( orient) {
  case 0x00:
    image_def.imgOrientation = ORIENT_RD; break;
  case 0x01:
    image_def.imgOrientation = ORIENT_UR; break;
  case 0x02:
    image_def.imgOrientation = ORIENT_RU; break;
  case 0x04:
    image_def.imgOrientation = ORIENT_LU; break;
  case 0x08:
    image_def.imgOrientation = ORIENT_DL; break;
  case 0x10:
    image_def.imgOrientation = ORIENT_DR; break;
  case 0x20:
    image_def.imgOrientation = ORIENT_UL; break;
  case 0x40:
    image_def.imgOrientation = ORIENT_LD; break;
  default:
    DGP_ERRF0( cErrInvalidParm, "Desired orientation must be a single value.");
  }

  */
  image_def.imgOrientation = orient;

/*
I'm finding this seek damages the file position pointer for libjpeg6b, which expects
1 4kB chunk to have already been read at this point.  Note that independent of this seek,
zephyr's second call to zephReadImageAbilities also results in a file pointer reset
to the offset (but that seek appears to be in the jpeg library).  So that is commented 
out too.  Think about reinstating this/these if problems. / dhtesq.

DGP_ERR_GEN_CLEANUP(
    jstate->u->fseek( jstate->u->fp, jstate->offset, SEEK_SET), 
*/

	/* try to set a reasonable color space */
	if (jstate->cinfo.num_components == 1) /* check for gray */
		  jstate->cinfo.jpeg_color_space = JCS_GRAYSCALE;
	else if (jstate->color_space == 6)
		jstate->cinfo.jpeg_color_space = JCS_YCbCr;
	else if (jstate->color_space == 10)
		  jstate->cinfo.jpeg_color_space = JCS_ITULAB;
	else /* default to YCbCr */
		jstate->cinfo.jpeg_color_space = JCS_YCbCr;

#ifdef I_WANT_TO_TIME_THE_JPEG
  start = clock();
#endif

  if (jstate->u.fp)
  {
  DGP_ERR_PASS_CLEANUP(
  Dgp_JPEG_decompress_jpeg ( &image_def, &dst_buf, &jstate->cinfo, dgp_state));
  }

#ifdef I_WANT_TO_TIME_THE_JPEG
  finish = clock();
  duration = (double)(finish - start) / CLOCKS_PER_SEC;
  printf("%f seconds for %d x %d...", duration, dstW, dstH);
#endif

  /*
  11.7.2000 / dhtesq.
  If we're doing a band decompression, we might need to decompress this pig again in a subsequent band.

  The way the jpeg interface is constructed, the library will reset its own state after a decompress.
  I think the easiest way around this is to just reinitialize.  Or?
  */

  /*
  switch(orient){
    case ORIENT_RD:
        if(dstMinY + nominal_strip_height < srcOriginY + jstate->srcH){
            DGP_ERR_PASS_CLEANUP(DGP_JPEG_init_jdecompress(&jstate->cinfo, dgp_state));
        }
        break;
    case ORIENT_LU:
        if(dstMinY > srcOriginY){
            DGP_ERR_PASS_CLEANUP(DGP_JPEG_init_jdecompress(&jstate->cinfo, dgp_state));
        }
        break;
    case ORIENT_DL:
        if(dstMinX > srcOriginX){
            DGP_ERR_PASS_CLEANUP(DGP_JPEG_init_jdecompress(&jstate->cinfo, dgp_state));
        }
        break;
    case ORIENT_UR:
        if(dstMinX + nominal_strip_height < srcOriginX + jstate->srcW){
            DGP_ERR_PASS_CLEANUP(DGP_JPEG_init_jdecompress(&jstate->cinfo, dgp_state));
        }

        break;
    default:
        DGP_ERRF0(cErrInvalidParm, "Invalid orientation.");
  } */


  /* 1.5.2001 / dhtesq. */
  /* This allocates, indirectly.  Since we do one of these on open and another on read, */
  /* we get twice as many as we need, and leak. */
  DGP_ERR_PASS_CLEANUP(DGP_JPEG_init_jdecompress(&jstate->cinfo, dgp_state));

  NORMAL_RETURN;

ERR_CLEANUP:

  ERR_RETURN;  
}


ErrCode 
zephJpegClose( void *decodeState)
{
  zephJpegState *jstate = (zephJpegState *)decodeState;
  Dgp_State *dgp_state;

  DGP_PROLOGUE( "zephJpegClose()", NULL);
  DGP_ERR_GEN( jstate == (zephJpegState *)NULL, cErrInvalidParm);
  dgp_state = jstate->dgp_state;
  DGP_ERR_BEGIN( dgp_state);

  /* Free intellibuffer */
  if(jstate->u.fp){
	DgpIbuf_mclose ((void**) &jstate->u.fp);
  }

  Dgp_JPEG_destroy_jdecompress (&jstate->cinfo, dgp_state); 
 
  DGP_FREE( jstate);

  NORMAL_RETURN;
}

zephDecodeOps zephJpegOps = { zephJpegOpen, 
                              zephJpegUsage,
                              zephJpegAbilities, 
                              zephJpegRead,
                              zephJpegClose };

#else /* ZEPH_INCLUDE_JPEG */

#include "nop.h"

zephDecodeOps zephJpegOps = { zephNopOpen,
                              zephNopUsage,
                              zephNopAbilities,
                              zephNopRead,
                              zephNopClose };

#endif /* ZEPH_INCLUDE_JPEG */

