/* $Header: /project/icons/cvsroot/digipaper/src/libzephyr/page.c,v 1.8 2001/02/24 01:42:00 dtracy Exp $ */

#include "zephpage.h"
#include "zephimage.h"
#include "zephprivate.h"

#include "dgp_xif_err_control.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libzephyr/page.c,v 1.8 2001/02/24 01:42:00 dtracy Exp $");

/*
14.11.2000 / dhtesq.
Moved to zephpage_t.h...
#define ZEPH_MAINIMAGE_LAYER     1
#define ZEPH_MAINIMAGE_NUMBER    0
*/


ErrCode
zephOpenPage( zephDocumentInfo docInfo, Dgp_UInt16 pageNum, 
                       zephPageInfo *pPageInfo)
{
  zephPageInfo page = (zephPageInfo)NULL;
  _zephPrivatePageInfo *_priv;
  _zephPrivateDocumentInfo *_privDoc;
  Dgp_UInt32 *imageCounts = (Dgp_UInt32 *)NULL;
  Dgp_UInt32 i;
  Dgp_UInt32 xoffset, yoffset, xresolution, yresolution, width, height;
  IO_COMPRESSION compression;
  IO_COLORTYPE xColorType;
  Dgp_UInt8 pPalette;
  IO_IMGHANDLE inIoImg;

ZEPH_PROLOGUE_DOC( "zephOpenPage()", docInfo);

DGP_ERR_GEN( pPageInfo == (zephPageInfo *)NULL, cErrInvalidParm); 
   
  _privDoc = (_zephPrivateDocumentInfo *)docInfo->privateInfo;

DGP_CHECK_ALLOC_CLEANUP(
  page = DGP_MALLOC( sizeof( *page) + sizeof( *_priv)));
  _priv = (_zephPrivatePageInfo *)((Dgp_UInt8 *)page + sizeof( *page));

  page->privateInfo = (zephPrivatePageInfo)_priv;
  page->pageNum = pageNum;
  page->imageCounts = (Dgp_UInt32 *)NULL;
  page->width = page->height = 0;
  _priv->dgp_state = dgp_state;
  _priv->docInfo = docInfo;

DGP_XIF_ERR_PASS_CLEANUP(
  IO_GetPageInfo( _privDoc->IO_doc_handle, pageNum, &imageCounts, 
                   &page->layerCount, &_priv->annotationCount, 
                   &_priv->IO_page_handle));

  /* Copy the image counts to our array. */
DGP_CHECK_ALLOC_CLEANUP(
  page->imageCounts = (Dgp_UInt32 *)DGP_MALLOC( page->layerCount 
                                                   * sizeof( Dgp_UInt32)));

/* get some information about this page by looking at its layers */
  for( i = 0; i < page->layerCount; i++) {
    page->imageCounts[ i] = imageCounts[ i];
  }

  /* assume that the page size and resolution is given by layer 2 */
  if (imageCounts[1] == 1)
    {
      DGP_XIF_ERR_PASS_CLEANUP(
	    IO_GetLayerInfo(_priv->IO_page_handle, 2, &xoffset, &yoffset,
			    &xresolution, &yresolution,
			    &width, &height, &compression,
			    &xColorType, &pPalette, &inIoImg));
      
      if (xoffset == 0 && yoffset == 0)
	{	    
	  page->width = width;
	  page->height = height;
	  page->res.xRes = DgpCoord_eqi (xresolution);
	  page->res.yRes = DgpCoord_eqi (yresolution);
	}
    }
  
  page->userFlags = 0;
   
  *pPageInfo = page;
  NORMAL_RETURN;

ERR_CLEANUP:
  if( page != (zephPageInfo)NULL) {
    if( page->imageCounts != (Dgp_UInt32 *)NULL) {
      DGP_FREE( (void *)page->imageCounts);
    }
    DGP_FREE( (void *)page);
  }
  
  ERR_RETURN;
}

ErrCode
zephClosePage( zephPageInfo *pPageInfo)
{
ZEPH_PROLOGUE_PAGE("zephClosePage()", *pPageInfo);

  DGP_FREE( (void *)(*pPageInfo)->imageCounts);
  DGP_FREE( (void *)(*pPageInfo));

  *pPageInfo = (zephPageInfo)NULL;
  NORMAL_RETURN;  
}

Dgp_UInt32
zephGetPageMemoryUsage( zephPageInfo pageInfo)
{
  if( pageInfo == (zephPageInfo)NULL) {
    return 0;
  }

  return sizeof( *pageInfo)
    + sizeof( _zephPrivatePageInfo)
    + (pageInfo->layerCount * sizeof( Dgp_UInt32));
}
