/* $Header: /project/icons/cvsroot/digipaper/src/libzephyr/tokenize.c,v 1.17 2003/02/10 22:53:46 mkrocken Exp $ */
#include "zephtokenize.h"
#include "zephprivate.h"
/*#include "zephdictionary.h" */
#include "decodebytes.h"
#include "decode-colors.h"
#include "jb2mmuncompress.h"

/* copied from mrcio/iolib/tifffx/tfxtiff.h */
#define T88OPTION_COLOURTAGS 0x00000040

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/src/libzephyr/tokenize.c,v 1.17 2003/02/10 22:53:46 mkrocken Exp $");

#define ZEPH_INITIAL_DICT_EXT_COUNT   5

typedef struct { 
  Dgp_State *dgp_state;
  zephDictionaryInfo dictInfo;
  void *backend;

} _zephDictionaryState;

ErrCode
zephReadBinaryDictionary ( zephImageInfo zephImage, Dgp_UInt32 dictId, 
			   jb2Segment **segment, Dgp_UInt32 *usage )
{
  _zephPrivateImageInfo *_privImage = NULL;
  _zephPrivatePageInfo *_privPage = NULL;
  _zephPrivateDocumentInfo *_privDoc = NULL;

  zephDocumentInfo Doc;
  /*Dgp_UInt32 bufSize = 0; */
  Dgp_UInt8 *dictBuf = NULL;
  Dgp_UInt32 symDictBytes, symDictBlockBytes, decoderMemory;
  Dgp_UInt16 lastPageUsed;


ZEPH_PROLOGUE_IMAGE( "zephReadBinaryDictionaryAll()", zephImage);
    
  DGP_ERR_GEN( segment == NULL
           , cErrInvalidParm);
  DGP_ERR_GEN( usage == NULL
           , cErrInvalidParm);


  /* initialize */
  _privImage = (_zephPrivateImageInfo *) zephImage->privateInfo;
  _privPage = (_zephPrivatePageInfo *) _privImage->pageInfo->privateInfo;
  _privDoc = (_zephPrivateDocumentInfo *) _privPage->docInfo->privateInfo;
  Doc = (zephDocumentInfo) _privPage->docInfo;
        
    /* read the compressed dictionary */
    DGP_ERR_PASS_CLEANUP(
      IO_GetSymbolDict (_privDoc->IO_doc_handle,
      dictId,
      &dictBuf,
      &symDictBytes,
      &symDictBlockBytes,
      &lastPageUsed,
      &decoderMemory));
    
    /* Decode the dictionary, return a pointer to the segment */
    DGP_ERR_PASS_CLEANUP(
      jb2_decodeTokDictSeg(segment, (void *)dictBuf, (size_t)symDictBytes, 
      dgp_state)) ;
    
    IO_Free((IO_INSTHANDLE)_privDoc->IO_instance_id, (void **)&dictBuf);
    
    /* Fixme */
    *usage = 100;

  NORMAL_RETURN;

ERR_CLEANUP:
  if( dictBuf != (Dgp_UInt8 *)NULL) {
    IO_Free((IO_INSTHANDLE)_privDoc->IO_instance_id, (void **)&dictBuf);
  }

  ERR_RETURN;
}

ErrCode
zephReadPositionBlock( zephImageInfo imageInfo, DgpAdt_List seg_list,
                      jb2SymbolInstance **sym, Dgp_UInt32 *n_sym, 
                      Extension **exts, Dgp_UInt32 *memUsage)
{
    _zephPrivateImageInfo *_priv;
    _zephPrivateDocumentInfo *_privDoc;
    /*_zephPrivateDictionaryInfo *_privDict; */
    /*_zephDictionaryState *state; */
    zephDocumentInfo _Doc;
    Dgp_UInt8 *buf = (Dgp_UInt8 *)NULL;
    /*Dgp_UInt32 usage = 0; */
    Dgp_UInt32 imgFlags = 0,  bytesFilled;
    Dgp_UInt32 total_buf_size = 0, posBlockSize = 0;
    Dgp_UInt8 *pPosBlock;
    Dgp_UInt32 bufSizeNoHeader;
    Dgp_UInt16 dictCount;
    Dgp_UInt32 *dictIDs;
    
    Dgp_UInt32 nboxes;
    
    Dgp_Bool have_extensions = DGP_FALSE;
    IO_IMGHANDLE pioImgHandle = 0;/*NULL; */

    ZEPH_PROLOGUE_IMAGE( "zephReadPositionBlock()", imageInfo);
    
    DGP_ERR_GEN( imageInfo->compType != IO_COMP_TOKENIZATION &&
	       imageInfo->compType != IO_COMP_LLTOKENIZATION &&
           imageInfo->compType != IO_COMP_JBIG2
           , cErrInvalidParm);
    DGP_ERR_GEN( seg_list == DGP_NULLLIST
           , cErrInvalidParm);
    DGP_ERR_GEN( memUsage == NULL
           , cErrInvalidParm);
    DGP_ERR_GEN( exts == NULL
           , cErrInvalidParm);
    DGP_ERR_GEN( n_sym == NULL
           , cErrInvalidParm);
    
    _priv = (_zephPrivateImageInfo *)imageInfo->privateInfo;
    _Doc = (zephDocumentInfo)((_zephPrivatePageInfo *)_priv->pageInfo->privateInfo)->docInfo;
    _privDoc = (_zephPrivateDocumentInfo *)_Doc->privateInfo;
        

    
    /* if this is a mulilayer image, check if there are color tags  */
    if(_priv->pageInfo->layerCount >= (imageInfo->layerNum + 2))
    {

        UInt32 pdwXOffset;
        UInt32 pdwYOffset;
        UInt32 pdwXResolution;
        UInt32 pdwYResolution;
        UInt32 pdwWidth;
        UInt32 pdwHeight;
        UInt32 pdwBestStripHeight;
        IO_COMPRESSION pioCompression;
        IO_COLORTYPE pioColorType;
        UInt32 pdwMaskImages;
        IO_PAGEHANDLE pioPageHandle = 0;/*NULL; */
        UInt32 pdwT88Options =0;
        Bool pbExists;

        pioPageHandle = ((_zephPrivatePageInfo *)_priv->pageInfo->privateInfo)->IO_page_handle;
        DGP_ERR_PASS_CLEANUP(
            IO_GetImageInfo(
            pioPageHandle,
            imageInfo->layerNum + 2, 
            0, /* only check the first image in the next layer */
            &pdwXOffset,
            &pdwYOffset,
            &pdwXResolution,
            &pdwYResolution,
            &pdwWidth,
            &pdwHeight,
            &pdwBestStripHeight,
            &pioCompression,
            &pioColorType,
            NULL,
            &pdwMaskImages,
            &pioImgHandle));

        if (pioImgHandle && pioCompression == IO_COMP_JBIG2) 
        {

            DGP_ERR_PASS_CLEANUP(
            IO_GetT88Options (
              pioImgHandle,
              &pdwT88Options,
              &pbExists));

            /* if we have don't have color tags in this layer, use the even layer */
            if((pdwT88Options & T88OPTION_COLOURTAGS) == 0)
            {
              pioImgHandle = _priv->IO_image_handle;
            }
            else    /* get the real position block size */
            {
                have_extensions = DGP_TRUE;
                DGP_ERR_PASS_CLEANUP(
                    IO_GetCompressedStripSize (
                    _priv->IO_image_handle,
                    0, /* row offset */
                    &posBlockSize));
            }
         }      
    }
    else
    { /* if this is a binary image, read from the even layer */
      pioImgHandle = _priv->IO_image_handle;
    }
    
    
    /* read the compressed position block */
    DGP_ERR_PASS_CLEANUP(
        IO_ImageReadStart(pioImgHandle, imgFlags));
    
    DGP_ERR_PASS_CLEANUP(
        IO_CompressedReadStrip (pioImgHandle,
        0, /* row offset */
        imageInfo->height,
        &buf,
        &total_buf_size,
        &bytesFilled));
    
    if (!have_extensions)
    {
      posBlockSize = total_buf_size;
    }

    /* strip the header */
    DGP_ERR_PASS_CLEANUP(
        IO_GetDictListFromPosBlock (_priv->IO_image_handle,
        buf,
        posBlockSize,
        &dictCount,
        &dictIDs,
        &pPosBlock,
        &bufSizeNoHeader));  
    
     
    DGP_ERR_PASS_CLEANUP(
        IO_ImageReadFinish(pioImgHandle));

    /* make sure we all agree how many dictionaries we need */
    DGP_ASSERT (dictCount == imageInfo->libCount);
        
    /* free dictionary arrays */
    if (dictCount > 0)
    {
        IO_Free ((IO_INSTHANDLE)_privDoc->IO_instance_id, (void **)&dictIDs);
    }
            
    /* decode position block */
    DGP_ERR_PASS_CLEANUP(
        jb2_decodeTokPositions( sym, 0, 0, 0, 0,
				pPosBlock, posBlockSize - (pPosBlock - buf), seg_list, &nboxes, dgp_state));
    
            
    /* copy the uncompressed color extensions, if needed */
    if (have_extensions && posBlockSize < total_buf_size)
    {
      Dgp_UInt8 *pos;

      DGP_CHECK_ALLOC_CLEANUP(*exts = (Extension *)DGP_MALLOC(sizeof(Extension)));
      *memUsage += sizeof(Extension); 
      pos = buf + posBlockSize;
      (*exts)->length = total_buf_size - posBlockSize;
      DGP_CHECK_ALLOC_CLEANUP((*exts)->extension = DGP_MALLOC((*exts)->length));
      memcpy((*exts)->extension, pos, (*exts)->length);
      (*exts)->tag = COLOR_EXT_TAG;
    }
    
    IO_Free((IO_INSTHANDLE)_privDoc->IO_instance_id, (void **)&buf);
    
    *memUsage += nboxes * sizeof (jb2SymbolInstance);
    *n_sym = nboxes;
        
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if( buf != (Dgp_UInt8 *)NULL) {
        IO_Free((IO_INSTHANDLE)_privDoc->IO_instance_id, (void **)&buf);
    }
    
    ERR_RETURN;
}


ErrCode
zephDecodeColorPositionExtension( zephImageInfo imageInfo,
                                 Extension *ext, Dgp_Int32 *ntags, color_tag **tags, 
                                 Dgp_UInt32 *memUsage)
{
    _zephPrivateImageInfo *_priv;
    color_cookie *colorState = (color_cookie *)NULL;
    /*Dgp_UInt32 usage = 0; */
    Dgp_UInt8 *buf;
    
    ZEPH_PROLOGUE_IMAGE( "zephDecodeColorPositionExtension()", imageInfo);
    DGP_ERR_GEN( ext == (Extension *)NULL, cErrInvalidParm);
    DGP_ERR_GEN( tags == (color_tag **)NULL, cErrInvalidParm);
    DGP_ERR_GEN( memUsage == (Dgp_UInt32*)NULL, cErrInvalidParm);
    
    _priv = (_zephPrivateImageInfo *)imageInfo->privateInfo;
    buf = (Dgp_UInt8 *)ext->extension;
    
    DGP_ERR_PASS_CLEANUP(
        initColors( &colorState, DGP_FALSE, dgp_state));
    DGP_ERR_PASS_CLEANUP(
        decodeJbig2Colors(
        colorState, 
        buf,
        ext->length,
        tags, 
        ntags));
    
    /* cleanup */
    cleanupColors( colorState);
    colorState = (color_cookie *)NULL;
    DGP_FREE (buf);
    ext->extension = NULL;
    ext->length = 0;

    *memUsage += *ntags * sizeof( color_tag);
    
    NORMAL_RETURN;
    
ERR_CLEANUP:
    if( colorState != (color_cookie *)NULL) {
        cleanupColors( colorState);
    }
    
    ERR_RETURN;
}

ErrCode
zephReadBinaryAnnotations( zephImageInfo imageInfo, Dgp_UInt8** data, Dgp_UInt32 *size)
{
  _zephPrivateImageInfo *_priv;
  _zephPrivateDocumentInfo *_privDoc;
  zephDocumentInfo _Doc;
  Dgp_UInt8 *buf = (Dgp_UInt8 *)NULL, **symDict = NULL;
  Dgp_UInt32 imgFlags = 0, buf_size, *symDictBytes = NULL, bytesFilled;
  Dgp_UInt8 *pPosBlock;
  Dgp_UInt32 PosBlockSize;
  DgpAdt_List seg_list = DGP_NULLLIST;
  Dgp_UInt16 dictCount;
  Dgp_UInt32 *dictIDs;
  Dgp_UInt32 *symDictBlockBytes = NULL; /*, decoderMemory; */
  Dgp_Int32 i = 0;
  
  
  ZEPH_PROLOGUE_IMAGE( "zephReadBinaryAnnotations()", imageInfo);
  
  DGP_ERR_GEN( imageInfo->compType != IO_COMP_TOKENIZATION &&
	       imageInfo->compType != IO_COMP_LLTOKENIZATION &&
	       imageInfo->compType != IO_COMP_JBIG2
	       , cErrInvalidParm);
  /*DGP_ERR_GEN( dictState == (zephDictionaryState)NULL, cErrInvalidParm); */
  
  _priv = (_zephPrivateImageInfo *)imageInfo->privateInfo;
  _Doc = (zephDocumentInfo)((_zephPrivatePageInfo *)_priv->
			    pageInfo->privateInfo)->docInfo;
  _privDoc = (_zephPrivateDocumentInfo *)_Doc->privateInfo;
  /*_privDict = (_zephPrivateDictionaryInfo *)(((_zephDictionaryState *)dictState)->dictInfo->privateInfo); */

  /*state = (_zephDictionaryState *)dictState; */
  
  /* read the compressed position block */
  DGP_ERR_PASS_CLEANUP(
		       IO_ImageReadStart(_priv->IO_image_handle, imgFlags));
  
  DGP_ERR_PASS_CLEANUP(
		       IO_CompressedReadStrip (_priv->IO_image_handle,
					       0, /* row offset */
					       imageInfo->height,
					       &buf,
					       &buf_size,
					       &bytesFilled));


  /* get the list of required dictionaries to decode this block */
  DGP_ERR_PASS_CLEANUP(
		       IO_GetDictListFromPosBlock (_priv->IO_image_handle,
						   buf,
						   buf_size,
						   &dictCount,
						   &dictIDs,
						   &pPosBlock,
						   &PosBlockSize));  
  
    
  DGP_ERR_PASS_CLEANUP(
		       IO_ImageReadFinish(_priv->IO_image_handle));
  
  	
    /* free dictionary arrays */
    if (dictCount > 0)
    {
        IO_Free ((IO_INSTHANDLE)_privDoc->IO_instance_id, (void **)&dictIDs);
    }
  
  /* get ascii data from position block */
  DGP_ERR_PASS_CLEANUP(
    jb2_decodeBinExtension( data, size,
			   pPosBlock, PosBlockSize, seg_list, dgp_state));
    
  IO_Free((IO_INSTHANDLE)_privDoc->IO_instance_id, (void **)&buf);


  NORMAL_RETURN;

ERR_CLEANUP:

  if( buf != (Dgp_UInt8 *)NULL) {
      IO_Free((IO_INSTHANDLE)_privDoc->IO_instance_id, (void **)&buf);
  }

  ERR_RETURN;
}

