/* $Header: /project/icons/cvsroot/digipaper/src/libzephyr/zephprivate.h,v 1.6 2003/02/10 22:53:46 mkrocken Exp $ */

#ifndef ZEPHPRIVATE_H
#define ZEPHPRIVATE_H

#include "dgp_state.h"

#include "zephdocument_t.h"
#include "zephpage_t.h"
#include "zephdictionary_t.h"
#include "zephimage_t.h"

#include "zephdecode.h"

#include "iowrap.h"

/* Used in the dictionary and tokenize modules. */
#define ZEPH_BACKENDDICT_VERSION_OFFSET            0
#define ZEPH_BACKENDDICT_VERSION_DGPAP             0
#define ZEPH_BACKENDDICT_VERSION_COLOR             1
#define ZEPH_BACKENDDICT_VERSION_JBIG2             2
#define ZEPH_BACKENDDICT_VERSION_NONE              255

#define ZEPH_BACKENDDICT_SIZE_OFFSET               1
#define ZEPH_BACKENDDICT_DATA_OFFSET               5

#define ZEPH_BACKENDPOSB_VERSION_OFFSET            0
#define ZEPH_BACKENDPOSB_VERSION_CURRENT           0

#define ZEPH_BACKENDPOSB_SIZE_OFFSET               1
#define ZEPH_BACKENDPOSB_LIB_COUNT_OFFSET          5
/* Added for color extensions.  Corresponds to POSBLOCK_DICTS_OFFSET */
/* in the old decoder. / dhtesq. */
#define ZEPH_BACKENDPOSB_DICTS_OFFSET              7

#define ZEPH_COLOR_POS_EXT_VERSION_OFFSET          0 
#define ZEPH_COLOR_POS_EXT_VERSION_CURRENT         0
#define ZEPH_COLOR_POS_EXT_LIB_COUNT_OFFSET        1


typedef struct {
  Dgp_State *dgp_state;

  IO_INSTHANDLE IO_instance_id;
  IO_DOCHANDLE IO_doc_handle;

  /* 30.4.2001 / dhtesq. */
  /* Leaking fileID;  record pointer. */
  FileID *fileID;
} _zephPrivateDocumentInfo;


typedef struct {
  zephDocumentInfo docInfo;
  Dgp_State *dgp_state;

  IO_PAGEHANDLE IO_page_handle;
  Dgp_UInt16 annotationCount;

} _zephPrivatePageInfo;


typedef struct {
  zephDocumentInfo docInfo;
  Dgp_State *dgp_state;

  Dgp_UInt32 *libSizes;    /* Compressed sizes. */

  Dgp_UInt32 *bufferSizes;
  Dgp_UInt8 **buffers;     /* Holds a temporary buffer. */

} _zephPrivateDictionaryInfo;


typedef struct {
  zephPageInfo pageInfo;
  Dgp_State *dgp_state;

  IO_IMGHANDLE IO_image_handle;
  Dgp_UInt32 stripCount;
  void **decodeState;

} _zephPrivateImageInfo;


#define ZEPH_PROLOGUE_DOC( funct_name, doc) \
  Dgp_State *dgp_state; \
  DGP_PROLOGUE( funct_name, NULL); \
  DGP_ERR_GEN( (doc) == (zephDocumentInfo)NULL, cErrInvalidParm); \
  DGP_ERR_GEN( (doc)->privateInfo == (zephPrivateDocumentInfo)NULL, \
               cErrInvalidParm); \
  dgp_state = ( (_zephPrivateDocumentInfo *)(doc)->privateInfo)->dgp_state; \
  DGP_ERR_GEN( dgp_state == (Dgp_State *)NULL, cErrInvalidParm); \
  DGP_ERR_BEGIN( dgp_state);

#define ZEPH_PROLOGUE_PAGE( funct_name, page) \
  Dgp_State *dgp_state; \
  DGP_PROLOGUE( funct_name, NULL); \
  DGP_ERR_GEN( (page) == (zephPageInfo)NULL, cErrInvalidParm); \
  DGP_ERR_GEN( (page)->privateInfo == (zephPrivatePageInfo)NULL, \
               cErrInvalidParm); \
  dgp_state = ( (_zephPrivatePageInfo *)(page)->privateInfo)->dgp_state; \
  DGP_ERR_GEN( dgp_state == (Dgp_State *)NULL, cErrInvalidParm); \
  DGP_ERR_BEGIN (dgp_state);

#define ZEPH_PROLOGUE_DICT( funct_name, dict) \
  Dgp_State *dgp_state; \
  DGP_PROLOGUE( funct_name, NULL); \
  DGP_ERR_GEN( (dict) == (zephDictionaryInfo)NULL, cErrInvalidParm); \
  DGP_ERR_GEN( (dict)->privateInfo == (zephPrivateDictionaryInfo)NULL, \
               cErrInvalidParm); \
  dgp_state = ((_zephPrivateDictionaryInfo *)(dict)->privateInfo)->dgp_state; \
  DGP_ERR_GEN( dgp_state == (Dgp_State *)NULL, cErrInvalidParm); \
  DGP_ERR_BEGIN( dgp_state);

#define ZEPH_PROLOGUE_IMAGE( funct_name, image) \
  Dgp_State *dgp_state; \
  DGP_PROLOGUE( funct_name, NULL); \
  DGP_ERR_GEN( (image) == (zephImageInfo)NULL, cErrInvalidParm); \
  DGP_ERR_GEN( (image)->privateInfo == (zephPrivateImageInfo)NULL, \
               cErrInvalidParm); \
  dgp_state = ( (_zephPrivateImageInfo *)(image)->privateInfo)->dgp_state; \
  DGP_ERR_GEN( dgp_state == (Dgp_State *)NULL, cErrInvalidParm); \
  DGP_ERR_BEGIN( dgp_state);


typedef struct {
  zephDecodeOpen open;
  zephDecodeUsage usage;
  zephDecodeAbilities abilities;
  zephDecodeRead read;
  zephDecodeClose close;

} zephDecodeOps;




#endif /* ZEPHPRIVATE_H */
