/* $Header: /project/icons/cvsroot/digipaper/src/shared/extension_tags.h,v 1.1 2001/04/10 23:46:29 dtracy Exp $ */


/* extension_tags.h

   Aaron Stump, Spring '97

   I.   In digipaper files, dictionary blocks and position blocks can both
        have extensions.  As we develop new kinds of extensions, we want to
        have some mechanism for keeping track of all the different kinds we
        come up with.  Our current solution is just to store a tag
        with each extension that tells what kind of extension it is.  This
        file defines the tags for the kinds of extensions we currently support.
   
   II.  If you want to add a new tag, just add your definition below.
        Also, please describe the extension you're adding a tag for in 
        part III (for position block extensions) or part IV (for dictionary
	block extensions). 

   III. Descriptions of extensions to the position block
        (referenced by name of tag):
      
        A. COLOR_EXT_TAG
	   1. Description
	      Color extensions to the position block are our way of adding
	      support for color tokens.  A color extension tells what
	      color (or what JPEG) goes with
	      each position.
	   2. Format
	      For a detailed description of the format of this kind of
	      extension, see the comment at the start of 
	      tokenize/src/libcolcomp/decode-colors.c.
	   3. Programs using this extension:
	      i.  tcomp-lossless-color
              ii. tuncomp-all

        B. SHIFTS_EXT_TAG (obsolete)
	   1. Description
	      Shift extensions are used in scaled token-compressed files.
	      A shift extension tells for each position, whether the token
	      at that position should be shifted one pixel.
	   2. Format
	      For a detailed description of the format of this kind of
	      extension, see the comment at the start of 
	      tokenize/src/libscalepos/decode-shifts.c
	   3. Programs using this extension:

   IV.  Descriptions of extensions to the dictionary block
        (referenced by name of tag):

        B. SCALEDTOK_EXT_TAG
	   1. Description
	      We use these kind of extensions in scaled token-compressed 
	      documents.  We scale the tokens to 2-bit gray, and extensions
	      of this kind hold the tokens for one of the two bitplanes.
	   2. Format
	   3. Programs using this extension
*/

#ifndef EXTENSION_TAGS_H
#define EXTENSION_TAGS_H

#ifdef __cplusplus
extern "C"
{
#endif

/* Tags for extensions to position blocks: */
#define COLOR_EXT_TAG   0
#define SHIFTS_EXT_TAG  3
#define TEXT_EXT_TAG    4


/* Tags for extensions to dictionary blocks: */
#define SCALEDTOK_EXT_TAG  2

#ifdef __cplusplus
}
#endif

#endif
