#ifndef UDS_H
#define UDS_H

#include <stdio.h>
#include "nt_compat.h"
#include "dgp_types.h"
#include "util.h"
#include "dgp_state.h"

#ifdef __cplusplus
extern "C"
{
#endif

/* These typedef's give names to the types of the function pointers that 
   libxiff will need for user-defined IO.  The parameter lists for these
   functions are the same (except that they take a (void *) instead of a 
   (FILE *) and have the same meanings as those for fseek(), ftell(), 
   fread() and fwrite(), respectively. */
typedef Dgp_Int32 (TC_CDECL *fseek_func) (void *, Dgp_Int32, Dgp_Int32);
typedef Dgp_Int32 (TC_CDECL *ftell_func) (void *);
typedef size_t (TC_CDECL *fread_func) (void *, size_t, size_t, void *);
typedef size_t (TC_CDECL *fwrite_func)(void *, size_t, size_t, void *);
typedef Dgp_Int32 (TC_CDECL *feof_func)(void *);

typedef struct {
  void *fp;
  fseek_func  fseek;
  ftell_func  ftell;
  fread_func  fread;
  fwrite_func fwrite;
  /* Added feof(). 27.3.2001 / dhtesq.  I changed the name to ffeof() because on some systems */
  /* it's implemented as a macro. */
  feof_func   ffeof;
} uds;

DGP_LIBEXPORT Dgp_Int32 Dgp_initUds( Dgp_State *dgp_state, uds *u, FILE *file);

#ifdef __cplusplus
}
#endif

#endif

