#ifndef MrcFile_h_Included
#define MrcFile_h_Included


#include "TfxFile.h"
#include "TfxPage.h"

class Jbig2Dict;
class MrcIo;
/*************************************************************************



*************************************************************************/

class MrcFile : public TfxFile
{
public:

    MrcFile( const char* file );

    virtual ~MrcFile( void );

    Asd::UInt32 numPages( void );

    TfxPage * getPage( Asd::UInt32 page );
    virtual Asd::Point32 getPageSize (Asd::UInt32 page );

    /* If page is not loaded already, load it.  This is not necessary */
    /* as calling getPage( page ) will force the load anyway. */
    void loadPage( Asd::UInt32 page );

    /* If the current page is loaded, free up the resources associated */
    /* with it. */
    void purgePage( Asd::UInt32 page ) { delete m_pages[page]; m_pages[page] = NULL; }

    /* Call purgePage() for each page. */
    void purgeAllPages( void ) ;

	/* flag to decide if the buffer require BGR or RGB */
    void setBGR(bool flag) { m_BGR = flag;}
    bool isBGR() { return m_BGR;}

	/* set and get dictionary */
    void setDict(Jbig2Dict* dict) { m_jbig2Dict = dict;}
    Jbig2Dict* getDict() { return m_jbig2Dict;}

protected:

    FILE * m_file;

    /* m_pages should quickly be initialized to (NULL,0) for  */
    /* every page.  Page objects can be created as they are viewed.     */
    /* */
    /* The first time a page is acceesed, the offset into the file */
    /* to the beginning of the page's IFD is placed into the second */
    /* element of the pair.  Then a new page object is created.  The */
    /* pointer to that object is stored as the first member of the pair. */
    /* In principle they can be removed as well if memory needs to be */
    /* freed up. */
    std::vector<MrcPage *> m_pages;
    MrcIo* m_pIO;
    Jbig2Dict* m_jbig2Dict;
    
    bool m_BGR;

    /* these do most of the actual work of reading the compressed data */
    void loadData ( Asd::UInt32 page );
    std::vector<TfxColorBand*> readColorBands ( Asd::UInt32 page, Asd::UInt32 layer );

private:

};


#endif
/*#ifndef MrcFile_h_Included */
