
#ifndef MrcIo_h_Included
#define MrcIo_h_Included

#include "TfxIo.h"
#include <vector>

class DictInfo;

namespace mrcio {

/* typedef for normal file access functions */
typedef Asd::Int32 (*fseek_func) (void *, Asd::Int32, Asd::Int32);
typedef Asd::Int32 (*ftell_func) (void *);
typedef size_t (*fread_func) (void *, size_t, size_t, void *);
typedef size_t (*fwrite_func)(void *, size_t, size_t, void *);
typedef Asd::Int32 (*feof_func)(void *);

/* universal device structure */
typedef struct {
  void *fp;
  fseek_func  fseek;
  ftell_func  ftell;
  fread_func  fread;
  fwrite_func fwrite;
} uds;

/* a mrcio required structure */
typedef struct {
    FILE *fp;
    Asd::Int32 pos;
    uds   *u;
} FileID;

/* custom file access functions, calling stdio lib functions */
Asd::Int32 
fileRead(Asd::UInt32 fileID, Asd::Int8 *buffer, Asd::Int32 byteCount);
Asd::Int32 
fileWrite(Asd::UInt32 fileID, const Asd::Int8 *buffer, Asd::Int32 byteCount);
Asd::Int32 
fileSeek(Asd::UInt32 fileID, Asd::Int32 offset);
Asd::Int32 
fileFileSize(Asd::UInt32 fileID);
Asd::Int32 
fileTell(Asd::UInt32 fileID);
#include "rcs.pub"
#include "iomrc.h"
}
/* An implementation based on Dave A's Mrcio library */
class MrcIo : public TfxIo
{
public:
	MrcIo ( ::FILE* fp);
	virtual ~MrcIo ();

	/*////////////  File/Document level ////////////////// */
	/* how many pages are in this document */
	virtual Asd::UInt32 docNPages () {return m_pageCount;}
	/* get the compressed dictionary */
	virtual Asd::UInt8* dictData (Asd::UInt32 dictID);
    virtual bool dictInfo (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image, DictInfo &dictInfo);
    /* what color space is used in this layer */
    virtual ColorSpace layerColorSpace (Asd::UInt32 page, Asd::UInt32 layer);

	/*////////////  Page level ///////////////// */
	/* how many layers are in this page */
	virtual Asd::UInt32 pageNLayers (Asd::UInt32 pageNumber);

	/*////////////  Layer level //////////////// */
	/* how many images in this layer */
	virtual Asd::UInt32 layerNImages (Asd::UInt32 page, Asd::UInt32 layer);
	/* how is this layer compressed */
	virtual CompressionMethod layerCompression (Asd::UInt32 page, Asd::UInt32 layer);
	/* origin of this layer */
	virtual Asd::Point32 layerOrigin (Asd::UInt32 page, Asd::UInt32 layer);
	/* size of this layer */
	virtual Asd::Point32 layerSize (Asd::UInt32 page, Asd::UInt32 layer);
	/* resolution of this layer */
	virtual Asd::Point32 layerDPI (Asd::UInt32 page, Asd::UInt32 layer);

	/*////////////  Image level ////////////////// */
	/* get the compressed data */
	virtual Asd::UInt8* imageData (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image);
	Asd::UInt8* MrcIo::jbig2Data (Asd::UInt32 page, Asd::UInt32 layer, 
			                        Asd::UInt32 image, 	Asd::UInt32& dataSize);
    Asd::UInt8* colorExtensionData ( Asd::UInt32 page, Asd::UInt32 layer, 
                                        Asd::UInt32 image,  Asd::UInt32& dataSize);
	/* size of the compressed data */
	virtual Asd::UInt32 imageDataSize (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image);
	/* origin of this image */
	virtual Asd::Point32 imageOrigin (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image);
	/* size of this image in pixels */
	virtual Asd::Point32 imageSize (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image);

    /* check if an image is written in multiple strips */
    bool isStriped (Asd::UInt32 page, Asd::UInt32 layer);
    /* get special decode parameters */
    std::vector<Asd::Float32> MrcIo::layerDecodeParam (Asd::UInt32 page, Asd::UInt32 layer);

protected:

    /* return a page handle */
	Asd::UInt32 openPage (Asd::UInt32 pageNumber, Asd::UInt32* layerCount, Asd::UInt32** imageCount);
    /* return an image handle */
    Asd::UInt32 openImage (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image);	

private:

	/* mrcio handles */
	mrcio::IO_INSTHANDLE m_instanceID;
	mrcio::IO_DOCHANDLE m_hDoc;

	/* data storage */
	Asd::UInt32 m_pageCount;

    /* file pointer */
    mrcio::FileID m_fileID;
};


#endif
