#ifndef TfxFile_h_Included
#define TfxFile_h_Included

#include "AsdTypes.h"

class TfxPage;
class TifPage;
class Jbig2Dict;

/*************************************************************************
 
 TfxFile represents a Tiff-fx file.  Since this class has relatively
 little to it, my feeling is that a different base class can be built
 to handle non-files.

 You initialize it with a file name.  This file is opened immediately so as
 to lock it for the duration of its lifespan.

 After that, you can tell it what page you are interested and ask for 
 that page's TfxPage object.  Pages are 0-based.

 The dictionary objects are held here because JBIG2 dictionaries are global
 to the file.

*************************************************************************/

class TfxFile
{
public:

    TfxFile( void ) {}

    virtual ~TfxFile( void );

    virtual Asd::UInt32 numPages( void ) = 0;

    virtual TfxPage * getPage( Asd::UInt32 page ) = 0;

    /* If page is not loaded already, load it.  This is not necessary */
    /* as calling getPage( page ) will force the load anyway. */
    virtual void loadPage( Asd::UInt32 page ) = 0;

    /* If the current page is loaded, free up the resources associated */
    /* with it. */
    virtual void purgePage( Asd::UInt32 page ) = 0;

    /* Call purgePage() for each page. */
    virtual void purgeAllPages( void ) = 0;

    virtual Asd::Point32 getPageSize( Asd::UInt32 page ) = 0;


protected:
private:

};



#endif
/*#ifndef TfxFile_h_Included */
