#include "StdAfx.h"

#ifdef _WIN32
// Warning 4786 appears to be benign, and we get tons of them,
// so we disabled it.
#pragma warning(disable:4786)
#endif


#include "MrcIo.h"
#include "Jbig2Dict.h"

MrcIo::MrcIo ( ::FILE* fp) :
m_instanceID (0),
m_hDoc (0),
m_pageCount (0)
{
	
	// create an instance of the IO libs
	Asd::Int32 status = mrcio::IO_InitInstance(0, &m_instanceID); 

	// set up the required structs for opening a document
	Asd::UInt32 options = 0;
	mrcio::uds *uds = new mrcio::uds;
	uds->fread = (mrcio::fread_func)::fread;
	uds->ftell = (mrcio::ftell_func)::ftell;
	uds->fseek = (mrcio::fseek_func)::fseek;
	m_fileID.fp = fp;
	m_fileID.pos = ftell (fp);
	m_fileID.u = uds;

	// open the document and set the document handle
	if (!status)
		status = mrcio::IO_OpenDocumentRead(m_instanceID, (Asd::Int32) &m_fileID,
		options, (mrcio::IO_READFUNC)mrcio::fileRead, (mrcio::IO_WRITEFUNC)mrcio::fileWrite, 
		mrcio::fileSeek, mrcio::fileFileSize, mrcio::fileTell, &m_hDoc);

	// get the number of pages in the document
	Asd::UInt8 version, revision;
	Asd::UInt16 pageCount;

	if (!status) status = mrcio::IO_GetDocInfo( m_hDoc, &version, &revision, &pageCount);
	m_pageCount = (Asd::UInt32) pageCount;

	// make sure everything went well
	if (status) 
	{
		m_hDoc = 0;
		m_pageCount = 0;
	}
}

MrcIo::~MrcIo ( void )
{
	// close document and instance
	mrcio::IO_CloseDocument( m_hDoc);
	mrcio::IO_EndInstance(m_instanceID); 

    // delete the file access struc
    delete m_fileID.u;
}

Asd::UInt32 MrcIo::openPage (Asd::UInt32 pageNumber, Asd::UInt32* layerCount, Asd::UInt32** imageCount) 
{ 
  
  // check if this is a valid page
  if (!m_hDoc || pageNumber > m_pageCount) return 0;

	// open the page and get all information
  Asd::UInt32 hPage;
  Asd::UInt16 annotationCount;

  if (mrcio::IO_GetPageInfo( m_hDoc, pageNumber, imageCount, 
                   layerCount, &annotationCount, &hPage)) return 0;

  return hPage;
}

Asd::UInt32 MrcIo::pageNLayers (Asd::UInt32 pageNumber)
{
    Asd::UInt32 layerCount, *imageCount = NULL;
    
    // get the number of layers
    if (openPage (pageNumber, &layerCount, &imageCount))
        return layerCount;
    return 0;
}
	
Asd::UInt8* MrcIo::dictData (Asd::UInt32 dictID)
{

  Asd::UInt8 *dictBuf = NULL;
  Asd::UInt32 symDictBytes, symDictBlockBytes, decoderMemory;
  Asd::UInt16 lastPageUsed;
        
  // read the compressed dictionary
  if (mrcio::IO_GetSymbolDict (m_hDoc, dictID, &dictBuf,
      &symDictBytes, &symDictBlockBytes, &lastPageUsed, &decoderMemory)) return NULL;
  
  // copy the data so we don't owe Dave's libs any memory
  Asd::UInt8* buffer = new Asd::UInt8 [symDictBlockBytes];
  if (!buffer) return NULL;
  memcpy ((void*) buffer, (const void*) dictBuf, symDictBlockBytes);
  mrcio::IO_Free((mrcio::IO_INSTHANDLE) m_instanceID, (void **)&dictBuf);

  return buffer;
}

Asd::UInt32 MrcIo::layerNImages (Asd::UInt32 page, Asd::UInt32 layer)
{
    Asd::UInt32 layerCount = 0;
    Asd::UInt32* imageCount;
    Asd::UInt32 index = layer - 1;
    // get the number of images in this layer
    Asd::UInt32 hPage = openPage (page, &layerCount, &imageCount);

    // check parameters
    if (hPage == 0 || index >= layerCount || imageCount[index] == 0) return 0;

    // any number larger than 1 is the correct answer
    if (imageCount[index] > 1)
        return imageCount[index];

    // there is one image
    const Asd::UInt32 image = 0;

    // check for multi -- striped images
    mrcio::IO_IMGHANDLE himage;
    Asd::UInt32 xPos, yPos, xRes, yRes, width, height, stripHeight, mask;
    mrcio::IO_COMPRESSION compType;
    mrcio::IO_COLORTYPE colorType;

    mrcio::IO_GetImageInfo(hPage, layer, image,
            &xPos, &yPos, &xRes, &yRes, &width, &height, &stripHeight, 
            &compType, &colorType, NULL, &mask, &himage);

    // only one image in this layer
    if (stripHeight == height) return 1;

    // calculate the number of strips in the layer
    return (height + stripHeight - 1) / stripHeight;
}

CompressionMethod MrcIo::layerCompression (Asd::UInt32 page, Asd::UInt32 layer)
{
    Asd::UInt32 layerCount, *imageCount = NULL;
    Asd::UInt32 hPage = openPage (page, &layerCount, &imageCount);
    
    // check parameters
    if (!hPage || layer > layerCount) return UnknownCompression;

    // variables needed to call layerInfo
    Asd::UInt32 xoffset, yoffset, xresolution, yresolution, width, height;
    mrcio::IO_COMPRESSION compression;
    mrcio::IO_COLORTYPE xColorType;
    Asd::UInt8 pPalette;  
    mrcio::IO_IMGHANDLE inIoImg;

    if (mrcio::IO_GetLayerInfo(hPage, layer, &xoffset, &yoffset,
        &xresolution, &yresolution,
        &width, &height, &compression,
        &xColorType, &pPalette, &inIoImg)) return UnknownCompression;

    // convert mrcio's compression type to our enum
    switch (compression)
    {
    case IO_COMP_JPEG: 
        return JPEG;
        break;
    case IO_COMP_JBIG2: 
        return JBIG2;
        break;
    case IO_COMP_MMR: 
        return MMR;
        break;
    case IO_COMP_MH: 
        return MH;
        break;
    case IO_COMP_MR: 
        return MR;
        break;
    case IO_COMP_UNKNOWN: 
    default:
        return UnknownCompression;
    }
}

ColorSpace MrcIo::layerColorSpace (Asd::UInt32 page, Asd::UInt32 layer)
{
    Asd::UInt32 layerCount, *imageCount = NULL;
    Asd::UInt32 hPage = openPage (page, &layerCount, &imageCount);
    
    // check parameters
    if (!hPage || layer > layerCount) return Default;

    // variables needed to call layerInfo
    Asd::UInt32 xoffset, yoffset, xresolution, yresolution, width, height;
    mrcio::IO_COMPRESSION compression;
    mrcio::IO_COLORTYPE xColorType;
    Asd::UInt8 pPalette;  
    mrcio::IO_IMGHANDLE inIoImg;

    if (mrcio::IO_GetLayerInfo(hPage, layer, &xoffset, &yoffset,
        &xresolution, &yresolution,
        &width, &height, &compression,
        &xColorType, &pPalette, &inIoImg)) return Default;

    Asd::UInt16 colorSpace;
    Asd::UInt32 known;
    if (mrcio::IO_GetPhotometricInterpretation (inIoImg, &colorSpace, &known)) return Default;
    if (!known) return Default;
    
    // convert mrcio's color Space to our enum (see tfxtiff.h)
    switch (colorSpace)
    {
    case 0: 
        return WhiteIsZero;
        break;
    case 1:
        return BlackIsZero;
        break;
    case 2:
        return RGB;
        break;
    case 5:
        return CMYK;
        break;
    case 6:
        return YCbCr;
        break;
    case 8:
        return CIELab;
        break;
    case 9:
        return ICCLab;
        break;
    case 10:
        return ITULab;
        break;
    default:
        return Default;
    }
}

std::vector<Asd::Float32> MrcIo::layerDecodeParam (Asd::UInt32 page, Asd::UInt32 layer)
{
    Asd::UInt32 layerCount, *imageCount = NULL;
    Asd::UInt32 hPage = openPage (page, &layerCount, &imageCount);
    std::vector<Asd::Float32> vec(0);

    // check parameters
    if (!hPage || layer > layerCount) return vec;

    // variables needed to call layerInfo
    Asd::UInt32 xoffset, yoffset, xresolution, yresolution, width, height;
    mrcio::IO_COMPRESSION compression;
    mrcio::IO_COLORTYPE xColorType;
    Asd::UInt8 pPalette;  
    mrcio::IO_IMGHANDLE inIoImg;

    if (mrcio::IO_GetLayerInfo(hPage, layer, &xoffset, &yoffset,
        &xresolution, &yresolution,
        &width, &height, &compression,
        &xColorType, &pPalette, &inIoImg)) return vec;

    Asd::Int32 *values;
    Asd::UInt32 nValues, exists;
    if (mrcio::IO_GetDecode (inIoImg, &values, &nValues, &exists)) return vec;

    // copy values into vector
    if (exists)
    {
        for (Asd::UInt32 i = 0; i < nValues * 2; i+=2) 
            vec.push_back ((Asd::Float32) values[i] / (Asd::Float32) values[i+1]);
    }

    return vec;
}

Asd::Point32 MrcIo::layerOrigin (Asd::UInt32 page, Asd::UInt32 layer)
{
    Asd::UInt32 layerCount, *imageCount = NULL;
    Asd::UInt32 hPage = openPage (page, &layerCount, &imageCount);
    
    // check parameters
    if (!hPage || layer > layerCount) return Asd::Point32 (-1, -1);

    // variables needed to call layerInfo
    Asd::UInt32 xoffset, yoffset, xresolution, yresolution, width, height;
    mrcio::IO_COMPRESSION compression;
    mrcio::IO_COLORTYPE xColorType;
    Asd::UInt8 pPalette;  
    mrcio::IO_IMGHANDLE inIoImg;

    if (mrcio::IO_GetLayerInfo(hPage, layer, &xoffset, &yoffset,
        &xresolution, &yresolution,
        &width, &height, &compression,
        &xColorType, &pPalette, &inIoImg)) return Asd::Point32 (-1, -1);

    return Asd::Point32 (xoffset, yoffset);
}

Asd::Point32 MrcIo::layerDPI (Asd::UInt32 page, Asd::UInt32 layer)
{
    Asd::UInt32 layerCount, *imageCount = NULL;
    Asd::UInt32 hPage = openPage (page, &layerCount, &imageCount);
    
    // check parameters
    if (!hPage || layer > layerCount) return Asd::Point32 (-1, -1);

    // variables needed to call layerInfo
    Asd::UInt32 xoffset, yoffset, xresolution, yresolution, width, height;
    mrcio::IO_COMPRESSION compression;
    mrcio::IO_COLORTYPE xColorType;
    Asd::UInt8 pPalette;  
    mrcio::IO_IMGHANDLE inIoImg;

    if (mrcio::IO_GetLayerInfo(hPage, layer, &xoffset, &yoffset,
        &xresolution, &yresolution,
        &width, &height, &compression,
        &xColorType, &pPalette, &inIoImg)) return Asd::Point32 (-1, -1);

    return Asd::Point32 (xresolution, yresolution);
}

Asd::Point32 MrcIo::layerSize (Asd::UInt32 page, Asd::UInt32 layer)
{
    Asd::UInt32 layerCount, *imageCount = NULL;
    Asd::UInt32 hPage = openPage (page, &layerCount, &imageCount);
    
    // check parameters
    if (!hPage || layer > layerCount) return Asd::Point32 (-1, -1);

    // variables needed to call layerInfo
    Asd::UInt32 xoffset, yoffset, xresolution, yresolution, width, height;
    mrcio::IO_COMPRESSION compression;
    mrcio::IO_COLORTYPE xColorType;
    Asd::UInt8 pPalette;  
    mrcio::IO_IMGHANDLE inIoImg;

    // get image info
    if (mrcio::IO_GetLayerInfo(hPage, layer, &xoffset, &yoffset,
        &xresolution, &yresolution,
        &width, &height, &compression,
        &xColorType, &pPalette, &inIoImg)) return Asd::Point32 (-1, -1);

    return Asd::Point32 (width, height);
}

Asd::UInt32 MrcIo::openImage (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image)
{
    Asd::UInt32 layerCount, *imageCount = NULL;
    Asd::UInt32 hPage = openPage (page, &layerCount, &imageCount);
    bool strip = false;
    
    // check parameters
    if (!hPage || layer > layerCount) return 0;
    // check image number, watch for striped images
    if (imageCount[layer - 1] > 1 && image >= imageCount[layer - 1]) return 0;
    if (imageCount[layer - 1] == 1 && image >= imageCount[layer - 1]) strip = true;

    mrcio::IO_IMGHANDLE himage;
    Asd::UInt32 xPos, yPos, xRes, yRes, width, height, stripHeight, mask;
    mrcio::IO_COMPRESSION compType;
    mrcio::IO_COLORTYPE colorType;

    // for striped images the index must be 0
    if (strip) image = 0;

    if (mrcio::IO_GetImageInfo(hPage, layer, image,
            &xPos, &yPos, &xRes, &yRes, &width, &height, &stripHeight, 
            &compType, &colorType, NULL, &mask, &himage)) return 0;

    // if strip is TRUE, the image must be striped
    if (strip)
    {
        if (height > stripHeight) 
            return himage;
        else
            return 0;
    }

    return himage;
}

bool MrcIo::isStriped (Asd::UInt32 page, Asd::UInt32 layer)
{
    Asd::UInt32 layerCount, *imageCount = NULL;
    Asd::UInt32 hPage = openPage (page, &layerCount, &imageCount);
    
    // check parameters
    if (!hPage || layer > layerCount) return false;
    // number of images must be 1
    if (imageCount[layer - 1] != 1) return false;

    mrcio::IO_IMGHANDLE himage;
    Asd::UInt32 xPos, yPos, xRes, yRes, width, height, stripHeight, mask;
    mrcio::IO_COMPRESSION compType;
    mrcio::IO_COLORTYPE colorType;

    if (mrcio::IO_GetImageInfo(hPage, layer, 0,
            &xPos, &yPos, &xRes, &yRes, &width, &height, &stripHeight, 
            &compType, &colorType, NULL, &mask, &himage)) return false;

    if (height > stripHeight) 
        return true;

    return false;
}

Asd::Point32 MrcIo::imageOrigin (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image)
{
    Asd::UInt32 layerCount, *imageCount = NULL;
    Asd::UInt32 hPage = openPage (page, &layerCount, &imageCount);
    Asd::UInt32 image_internal = image;
    bool strip = false;

    // check parameters
    if (!hPage || layer > layerCount) return Asd::Point32 (-1, -1);
    if (imageCount[layer - 1] == 1)
    {
        if (isStriped (page, layer)) strip = true;
        else
            if (image > 0) return Asd::Point32 (-1, -1);
    }
    if (imageCount[layer - 1] > 1 && image >= imageCount[layer - 1]) return Asd::Point32 (-1, -1);

    mrcio::IO_IMGHANDLE himage;
    Asd::UInt32 xPos, yPos, xRes, yRes, width, height, stripHeight, mask;
    mrcio::IO_COMPRESSION compType;
    mrcio::IO_COLORTYPE colorType;

    if (strip) image_internal = 0;
    if (mrcio::IO_GetImageInfo(hPage, layer, image_internal,
            &xPos, &yPos, &xRes, &yRes, &width, &height, &stripHeight, 
            &compType, &colorType, NULL, &mask, &himage)) return Asd::Point32 (-1, -1);

    // adjust yPos if image is striped
    if (strip)
    {
        yPos += image * stripHeight;
        if (yPos > height) return Asd::Point32 (-1, -1);
    }

    return Asd::Point32 (xPos, yPos);
    
}

Asd::Point32 MrcIo::imageSize (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image)
{
    Asd::UInt32 layerCount, *imageCount = NULL;
    Asd::UInt32 hPage = openPage (page, &layerCount, &imageCount);
    Asd::UInt32 image_internal = image;
    bool strip = false;

    // check parameters
    if (!hPage || layer > layerCount) return Asd::Point32 (-1, -1);
    if (imageCount[layer - 1] == 1)
    {
        if (isStriped (page, layer)) strip = true;
        else
            if (image > 0) return Asd::Point32 (-1, -1);
    }
    if (imageCount[layer - 1] > 1 && image >= imageCount[layer - 1]) return Asd::Point32 (-1, -1);
    
    mrcio::IO_IMGHANDLE himage;
    Asd::UInt32 xPos, yPos, xRes, yRes, width, height, stripHeight, mask;
    mrcio::IO_COMPRESSION compType;
    mrcio::IO_COLORTYPE colorType;

    if (strip) image_internal = 0;
    if (mrcio::IO_GetImageInfo(hPage, layer, image_internal,
            &xPos, &yPos, &xRes, &yRes, &width, &height, &stripHeight, 
            &compType, &colorType, NULL, &mask, &himage)) return Asd::Point32 (-1, -1);

    // correct the last strip
    if (strip && (image + 1) * stripHeight > height)
    {
        stripHeight = height - image * stripHeight;
    }

    return Asd::Point32 (width, stripHeight);
}

Asd::UInt32 MrcIo::imageDataSize (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image)
{
    Asd::UInt32 iHandle = openImage (page, layer, image);
    bool isStrip = isStriped  (page, layer);
    Asd::UInt32 compSize = 0;

    if (!iHandle) return 0;

    if (isStrip)
    {
        Asd::Point32 offset = imageOrigin (page, layer, image);
        if (mrcio::IO_GetCompressedStripSize( iHandle, offset.getY(), &compSize)) return 0;
    }
    else
    {

        Asd::UInt32 libCount, *libId;
        Asd::Int32 *libSize;
        if (mrcio::IO_GetImageStats( iHandle, &compSize, &libCount, &libId, &libSize )) return 0;
    }

    return compSize;
}


bool MrcIo::dictInfo (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image, DictInfo &dictInfo){
    Asd::UInt32 iHandle = openImage (page, layer, image);
    bool isStrip = isStriped  (page, layer);
    Asd::UInt32 compSize = 0;
    Asd::UInt32 libCount, *libIds;
    Asd::Int32 *libSizes;
    if (!iHandle) return false;
    
    mrcio::IO_GetImageStats( iHandle, &compSize, &libCount, &libIds, &libSizes );
    if(libCount > 0) { 
		dictInfo.setInfo(libCount, (Asd::UInt32*)libSizes, libIds);
    } 
    return true;
}


Asd::UInt8* MrcIo::imageData (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image)
{
    Asd::UInt32 iHandle = openImage (page, layer, image);
    bool isStrip = isStriped  (page, layer);
    Asd::UInt8* buffer = NULL;

    if (!iHandle) return NULL;

    // get the size of the compressed image
    Asd::UInt32 compSize = imageDataSize (page, layer, image);
    if (compSize == 0) return NULL;

    // allocate the buffer
    buffer = new Asd::UInt8 [compSize];
    if (!buffer) return NULL;

    // read the data
    Asd::UInt32 bytesFilled, offsetY = 0;
    Asd::Point32 size = imageSize (page, layer, image);

    if (isStrip)
    {
        Asd::Point32 offset = imageOrigin (page, layer, image);
        offsetY = offset.getY();
    }

    // read the data
    Asd::Int32 status = mrcio::IO_ImageReadStart(iHandle, 0 /* Flags */);
    status += mrcio::IO_CompressedReadStrip (iHandle, offsetY, size.getY(),
        &buffer, &compSize, &bytesFilled);
    status += mrcio::IO_ImageReadFinish (iHandle);

    // check if all went well
    if (bytesFilled != compSize)
        status += 1;

    if (status) 
    {
        delete [] buffer;
        return NULL;
    }

    return buffer;
}

Asd::UInt8* MrcIo::colorExtensionData (
       Asd::UInt32 page, 
       Asd::UInt32 layer, 
       Asd::UInt32 image, 
       Asd::UInt32& dataSize)
{
    Asd::UInt32 iHandle = openImage (page, layer, image);
    
    if (!iHandle) return NULL;
    
    Asd::UInt32 pdwT88Options;
    Asd::UInt32 pbExists;
    mrcio::IO_GetT88Options ( iHandle, &pdwT88Options, &pbExists);
    if(!pbExists) return NULL;
    
    // get selector handle
    Asd::UInt32 iSelHandle = openImage(page, layer-1, image);
    if (!iSelHandle) return NULL;
    
    Asd::UInt32 posBlockSize;
    Asd::UInt32 total_buf_size;
    Asd::UInt32 status = 0;
 
    status += mrcio::IO_GetCompressedStripSize (
        iHandle,
        0, /* row offset */
        &total_buf_size);
   
    status += mrcio::IO_GetCompressedStripSize (
        iSelHandle,
        0, /* row offset */
        &posBlockSize);
    if (posBlockSize >= total_buf_size) 
        return NULL;
    
    if(status) return NULL;
    
    Asd::UInt8 *buf= new Asd::UInt8[total_buf_size];
    if(!buf) return NULL;

    // read the data
    Asd::UInt32 bytesFilled, offsetY = 0;
    Asd::Point32 size = imageSize (page, layer, image);
    bool isStrip = isStriped  (page, layer);
    if (isStrip)
    {
        Asd::Point32 offset = imageOrigin (page, layer, image);
        offsetY = offset.getY();
    }
    status += mrcio::IO_ImageReadStart(iHandle, 0);
    
    status += mrcio::IO_CompressedReadStrip (
        iHandle,
        0, /* row offset */
        size.getY(),
        &buf,
        &total_buf_size,
        &bytesFilled);
    
    status += mrcio::IO_ImageReadFinish(iHandle);

    /* copy the uncompressed color extensions, if needed */
    Asd::UInt8 *ext;
    dataSize = total_buf_size - posBlockSize;
    ext = new Asd::UInt8[dataSize];  
    memcpy(ext, buf + posBlockSize, dataSize);
    
    if(buf) delete buf;
    
    return ext;


}
Asd::UInt8* MrcIo::jbig2Data (
Asd::UInt32 page, 
Asd::UInt32 layer, 
Asd::UInt32 image, 
Asd::UInt32& dataSize)
{
    Asd::UInt32 iHandle = openImage (page, layer, image);
    bool isStrip = isStriped  (page, layer);
    Asd::UInt8* buffer = NULL;

    if (!iHandle) return NULL;

    // get the size of the compressed image
    Asd::UInt32 compSize = imageDataSize (page, layer, image);
    if (compSize == 0) return NULL;

    // allocate the buffer
    buffer = new Asd::UInt8 [compSize];
    if (!buffer) return NULL;

    // read the data
    Asd::UInt32 bytesFilled, offsetY = 0;
    Asd::Point32 size = imageSize (page, layer, image);

    if (isStrip)
    {
        Asd::Point32 offset = imageOrigin (page, layer, image);
        offsetY = offset.getY();
    }

    // read the data
    Asd::Int32 status = mrcio::IO_ImageReadStart(iHandle, 0 /* Flags */);
    status += mrcio::IO_CompressedReadStrip (iHandle, offsetY, size.getY(),
        &buffer, &compSize, &bytesFilled);
	Asd::UInt16 dictCount;
	Asd::UInt32* dictIDs;
	Asd::UInt8* databuffer = NULL;
    status += mrcio::IO_GetDictListFromPosBlock (iHandle,
											  buffer,
											  compSize,
											  &dictCount,
											  &dictIDs,
											  &databuffer,
											  &dataSize);
    status += mrcio::IO_ImageReadFinish (iHandle);

	// free dictionary arrays 
	if (dictCount > 0)
	{
		mrcio::IO_Free (m_instanceID, (void **)&dictIDs);
	}

    // check if all went well
    if (bytesFilled != compSize)
        status += 1;
    if (status) 
    {
		delete [] buffer;
        return NULL;
    }

	// copy the data without the header
	Asd::UInt8* realData = new Asd::UInt8 [dataSize];
	memcpy ((void*) realData, (const void*) databuffer, dataSize); 

	delete [] buffer;
	
	return realData;
}



/* mrcio implementation stuff */

Asd::Int32
mrcio::fileRead(
    Asd::UInt32 fileID,
    Asd::Int8 *buffer,
    Asd::Int32 byteCount
)
{
    FileID *file = (FileID *)fileID;
    int bytes;
    uds *u = (uds *) file->u;
    Asd::Int32 totalRead = 0;

    /* Do my own buffered read */
    while ((bytes = u->fread(buffer, sizeof(Asd::Int8), byteCount, file->fp)) > 0) {
	byteCount -= bytes;
	buffer += bytes;
	totalRead += (Asd::UInt32)bytes;
	if (! byteCount) break;
    }
    file->pos += totalRead;
    if (bytes < 0) {
	return bytes;	/* Error */
    }
    if (byteCount > 0) {
	return totalRead;	/* Error (not all read) put breakpoint here */
    }
    return totalRead;		
}


Asd::Int32
mrcio::fileWrite(
    Asd::UInt32 fileID,
    const Asd::Int8 *buffer,
    Asd::Int32 byteCount
)
{
    FileID *file = (FileID *)fileID;
    int bytes;
    uds *u = file->u;
    Asd::Int32 totalWritten = 0;

    /* Do my own buffered write */
    while ((bytes = u->fwrite((Asd::Int8 *)buffer, sizeof(Asd::Int8), byteCount, file->fp)) > 0) {
	byteCount -= bytes;
	buffer += bytes;
	totalWritten += (Asd::UInt32)bytes;
	if (! byteCount) break;
    }
    file->pos += totalWritten;
    if (bytes < 0) {
	return bytes;	/* Error */
    }
    return totalWritten;		
}


Asd::Int32
mrcio::fileSeek(
    Asd::UInt32 fileID,
    Asd::Int32 offset
)
{
    FileID *file = (FileID *)fileID;
    long retval;
    uds *u = file->u;

    retval = u->fseek(file->fp, offset, SEEK_SET);

    if (retval != 0) {
	return (Asd::UInt32)0;
    }
    return file->pos = (Asd::Int32)offset;
}


Asd::Int32
mrcio::fileTell(
    Asd::UInt32 fileID
)
{
    FileID *file = (FileID *)fileID;
    uds *u = file->u;

    return file->pos;
}


Asd::Int32
mrcio::fileFileSize(
    Asd::UInt32 fileID
)
{
    FileID *file = (FileID *)fileID;
    Asd::Int32 retval, newpos;
    uds *u = file->u;

    retval = u->fseek(file->fp, 0, SEEK_END);
    if (retval != 0) {
	return (Asd::Int32)0;
    }

    newpos = u->ftell(file->fp);
    file->pos = newpos;
    return file->pos;
}

