#ifndef TfxIo_h_Included
#define TfxIo_h_Included

#include "AsdRect.h"
#include "MiscConst.h"

/* An object to abstract the IO for Tiff-fx. */
/* This is a pure interface. At this point it is not complete, because there are no */
/* methods to deal with global dictionaries. */

class TfxIo 
{
public:
	TfxIo () {};
	virtual ~TfxIo () {};

	/*////////////  File/Document level ////////////////// */
	/* how many pages are in this document */
	virtual Asd::UInt32 docNPages () = 0;
	/* get the compressed dictionary */
	virtual Asd::UInt8* dictData (Asd::UInt32 dictID) = 0;

	/*////////////  Page level ///////////////// */
	/* how many layers are in this page */
	virtual Asd::UInt32 pageNLayers (Asd::UInt32 pageNumber) = 0;
	/* get size of the page in pixels, given by layer 2 */
	Asd::Point32 pageSize (Asd::UInt32 pageNumber) 
	{
		return layerSize (pageNumber, 2);
	}

	/*////////////  Layer level //////////////// */
	/* how many images in this layer */
	virtual Asd::UInt32 layerNImages (Asd::UInt32 page, Asd::UInt32 layer) = 0;
	/* how is this layer compressed */
	virtual CompressionMethod layerCompression (Asd::UInt32 page, Asd::UInt32 layer) = 0;
	/* what color space is used in this layer */
	virtual ColorSpace layerColorSpace (Asd::UInt32 page, Asd::UInt32 layer) = 0;
	/* origin of this layer */
	virtual Asd::Point32 layerOrigin (Asd::UInt32 page, Asd::UInt32 layer) = 0;
	/* size of this layer */
	virtual Asd::Point32 layerSize (Asd::UInt32 page, Asd::UInt32 layer) = 0;
	/* resolution of this layer */
	virtual Asd::Point32 layerDPI (Asd::UInt32 page, Asd::UInt32 layer) = 0;

	/*////////////  Image level ////////////////// */
	/* get the compressed data */
	virtual Asd::UInt8* imageData (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image) = 0;
	/* size of the compressed data */
	virtual Asd::UInt32 imageDataSize (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image) = 0;
	/* origin of this image */
	virtual Asd::Point32 imageOrigin (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image) = 0;
	/* size of this image in pixels */
	virtual Asd::Point32 imageSize (Asd::UInt32 page, Asd::UInt32 layer, Asd::UInt32 image) = 0;

protected:

private:

};


#endif
