
#include "StdAfx.h"

#include "TfxPage.h"
#include "AdapterFactory.h"
#include "DibImgBuf.h"
#ifndef REALOS
#include "IntelImgBuf.h"
#endif

#include <math.h>

MrcPage::~MrcPage ()
{
    if (m_bg) delete m_bg;
    m_bg = NULL;

    for (std::vector<TfxSelLayer*>::iterator layer = m_sel.begin(); 
    layer != m_sel.end(); layer++)
    {
        delete (*layer);
        *layer = NULL;
    }
}

    
bool MrcPage::render( ImageAdapter * fullBuf, const Asd::Rect & region, const Asd::Point32 & scale )
{

    bool status = true;

    AdapterFactory fac;
    
    if (region.isEmpty()) return status;
    
    Asd::Rect Roi (region);

    ColorDepth depth = fullBuf->depth();

    // render the background ..
    if (depth == COLOR_24_BIT && m_bg)
    {

        // check if the coordinates are aligned correctly
        while (Roi.top() % m_bg->getResolution().getY() != 0) Roi.offsetY (1);
        while (Roi.left() % m_bg->getResolution().getX() != 0) Roi.offsetX (1);

        // Get the intersection roi in scaled page coord
        Asd::Rect pageClipRoi = m_bg->getPageRoi(Roi,scale);  

        // Get the roi in actual scaled layer coord
        // if the requested scale is not power of 2, change to the nearest power of 2
        Asd::Rect layerRoi = m_bg->getLayerRenderRoi(pageClipRoi, scale); 

        // Get the actual scale of layer (power of 2, since jpeg can only scale in that way)
        bool scaleChanged = false;
        Asd::Point32 layerScale = m_bg->getLayerScalePower2(scale, scaleChanged);

        if(!layerRoi.isEmpty()){
            // Adjust the pageClipRoi relative to the top of the buffer passed in 
            // so that the buffer can be cropped correctly
            pageClipRoi.offsetXY (-Roi.left(), -Roi.top());
            ImageAdapter* ttBgBuf = fac.createSub (pageClipRoi, fullBuf);

            if(scaleChanged){
                // if the scale is not power of 2 compare to background res
                // Render into temp buffer using scale in power of 2
                DibImgBuf tempDibBuf;
                ImageAdapter* tempBgBuf = fac.createInstance (COLOR_24_BIT, layerRoi, &tempDibBuf);
                tempBgBuf->fillWhite();
                status = m_bg->render(tempBgBuf, layerRoi, layerScale);
#ifndef REALOS
                if (getColorSpace() == Default && m_bg->getColorSpace() == ITULab)
                    tempBgBuf->ITULab2RGB ();
#endif               
                //strech the background into the requested size
                if( m_bg->getResolution().getX() <= scale.getX() ||
                    m_bg->getResolution().getY() <= scale.getY() ||
                    m_bg->getResolution().getX() % 2 != 0 || m_bg->getResolution().getY() % 2 != 0) {
                     tempBgBuf->scaleToSize(ttBgBuf);
                } else {
                    Asd::Point32 tempscale(m_bg->getResolution().getX()/scale.getX(),
                                       m_bg->getResolution().getY()/scale.getY());             
                    tempBgBuf->scaleUp(tempscale, ttBgBuf);
                }
            }else{
                status = m_bg->render(ttBgBuf, layerRoi, layerScale);
#ifndef REALOS
                if (getColorSpace() == Default && m_bg->getColorSpace() == ITULab)
                    ttBgBuf->ITULab2RGB ();
#endif
            }
            if(!status) return status;
        }

    }

    if (depth == COLOR_24_BIT)
    {
        // render all sel/fg pairs
        for (std::vector<TfxSelLayer*>::iterator layer = m_sel.begin(); 
             layer != m_sel.end(); layer++)
        {
            status = (*layer)->render(fullBuf, Roi, scale);
            if (!status) return status;
        }
    }
    else
    {
        // render all sel planes
        for (std::vector<TfxSelLayer*>::iterator layer = m_sel.begin(); 
             layer != m_sel.end(); layer++)
        {
            status = (*layer)->renderSel(fullBuf, Roi, scale);
            if (!status) return status;
        }
    }

    return status;

}


/*
 *  Render the requested region to a File
 */
bool MrcPage::renderToFile( ImageBuf * buf, const Asd::Rect & region, 
                           Asd::Rational<Asd::Int32> zoom, const char * filename, 
                           const char * outtype, const Asd::Int32 & xRes, const Asd::Int32 & yRes)
{

    AdapterFactory fac;
	ImageAdapter *outBuf = NULL;
    
    // allocate the correct adpter, based on the color depth of this page
    ColorDepth depth = getColorDepth ();
    Asd::Rational<Asd::Int32> oneZoom (1);

    if (depth == MONOCHROME && zoom != oneZoom)
    {
        // gray scale
        outBuf = fac.createInstance(COLOR_8_BIT, region, buf);
    }
    else
    {
        // monochrome or 24bit color
        outBuf = fac.createInstance(depth, region, buf);
    }

    if (depth == COLOR_24_BIT)
    {
        // need to set buffer to white (RGB)
        outBuf->fillWhite();
    }
    
	if(!renderInternal( outBuf,  region, zoom, 0)) return false;
	if(!outBuf->save(filename, outtype, xRes, yRes)) return false;

	return true;

}

/*
 * Internal render function
 */
bool MrcPage::renderInternal( ImageAdapter * outBuf, const Asd::Rect & region, 
                              Asd::Rational<Asd::Int32> zoom, Asd::Int32 layer)

{

    AdapterFactory fac;
    ColorDepth bufdepth = outBuf->depth();

    // convert zoom to power of 2
    Asd::UInt32 zoomBase2=zoom.toPower2();

    // FIXME need better estimate of maximum zoom
    if (bufdepth== COLOR_24_BIT)
    {
        Asd::UInt32 maxZoom = zoomBase2;

        if (m_bg)
        {
            Asd::Point32 res = m_bg->getResolution ();
            maxZoom = res.getX() * 8;
        }
        
        // check all fg layers
        for (std::vector<TfxSelLayer*>::iterator layer = m_sel.begin(); 
             layer != m_sel.end(); layer++)
        {
            Asd::Point32 res = (*layer)->getFgResolution ();
            if (res.getX() * 8 < maxZoom) maxZoom = res.getX() * 8;
        }
        
        if (maxZoom < zoomBase2) zoomBase2 = maxZoom;
    }       

    // decode into a temporary buffer, if necessary
    Asd::Rational<Asd::Int32> tempzoom (zoomBase2);
    Asd::Point32 zoomBase(zoomBase2, zoomBase2);
    if (tempzoom != zoom)
    {

        // scale the image BB to the zoom size
        Asd::Rect scaledImageRoi (region);
        scaledImageRoi.scaleUp (zoom.num());
        scaledImageRoi.scaleDown (zoom.den());
        scaledImageRoi.scaleDownMax (zoomBase2, zoomBase2);

        DibImgBuf tempBuf;
        ColorDepth depth = bufdepth;
        if (zoomBase2 == 1 && depth == COLOR_8_BIT) depth = MONOCHROME;
        ImageAdapter *ibuf = fac.createInstance(depth, scaledImageRoi, &tempBuf);

        // set to default color
        if (layer <= 1)                
            ibuf->fillWhite();            
        else if (layer % 2 == 1)                
            ibuf->fillBlack();
        
        bool status = false;
        if (layer == 0)
            status = render (ibuf, scaledImageRoi, zoomBase);
        if (layer == 1)
            if (m_bg) status = m_bg->render (ibuf, scaledImageRoi, zoomBase);
        if (layer > 1)
        {
            Asd::Int32 selfg = layer / 2 - 1;
            if (m_sel.size() >= selfg)
            {
                if (layer % 2)
                    status = m_sel[selfg]->renderFg (ibuf, scaledImageRoi, zoomBase);
                else
                    status = m_sel[selfg]->renderSel (ibuf, scaledImageRoi, zoomBase);
            }
        } 
        if (!status) return false;

        ibuf->scaleToSize (outBuf);
    }
    else
    {
        bool status = false;
        if (layer == 0)
            status = render (outBuf, region, zoomBase);
        if (layer == 1)
            if (m_bg) status = m_bg->render (outBuf, region, zoomBase);
        if (layer > 1)
        {
            Asd::Int32 selfg = layer / 2 - 1;
            if (m_sel.size() >= selfg)
            {
                if (layer % 2)
                    status = m_sel[selfg]->renderFg (outBuf, region, zoomBase);
                else
                    status = m_sel[selfg]->renderSel (outBuf, region, zoomBase);
            }
        }
        if (!status) return false;
    }

    return true;
}

Asd::UInt32 MrcPage::getNumLayers( void ) const
{
    Asd::UInt32 count = 1;

    count += m_sel.size() * 2;
    
    return count;
}

ColorDepth MrcPage::getColorDepth( void )
{
    ColorDepth depth = MONOCHROME;

    if (m_bg)
    {
        depth = COLOR_24_BIT;
        return depth;
    }

    // check all sel/fg pairs for a foreground layer
    for (std::vector<TfxSelLayer*>::iterator layer = m_sel.begin(); 
         layer != m_sel.end(); layer++)
    {
        Asd::Point32 size = (*layer)->getFgSize ();
        if (size.getX() > 0) 
        {
            depth = COLOR_24_BIT;
            return depth;
        }
    }

    return depth;
}

bool MrcPage::render( ImageBuf * buf, const Asd::Rect & region, Asd::Float64 zoom, Asd::Int32 layer)
{

    const Asd::UInt32 denom = 210; // 2 * 3 * 5 * 7

    AdapterFactory fac;
    ImageAdapter *fullBuf;
    
    if (getNumLayers() < layer) return false;

    // allocate the correct adpter, based on the color depth of this page
    ColorDepth depth = getColorDepth ();
    Asd::Rect roi(region);
    if (layer > 0)
    {
        if (layer % 2 == 0) depth = MONOCHROME;
        else
        {
            Asd::Int32 resolution = 1;
            if (layer == 1)
            {
                if (m_bg) resolution = m_bg->getResolution().getX();
            }
            else             
                resolution = m_sel[layer / 2 - 1]->getFgResolution().getX();
            
            zoom /= (Asd::Float64) resolution;
        }
    }

    // convert float to rational
    Asd::UInt32 numer = (Asd::UInt32) floor(denom * zoom + 0.5); 
    Asd::Rational<Asd::Int32> ratZoom (numer, denom);
    Asd::Rational<Asd::Int32> oneZoom (1);

    if (depth == MONOCHROME && ratZoom != oneZoom)
    {
        // gray scale
        fullBuf = fac.createInstance(COLOR_8_BIT, roi, buf);
    }
    else
    {
        // monochrome or 24bit color
        fullBuf = fac.createInstance(depth, roi, buf);
    }

    // set to default color  
    if (layer <= 1)        
        fullBuf->fillWhite();
    else if (layer % 2 == 1)            
        fullBuf->fillBlack();
    
    if(!renderInternal(fullBuf, roi, ratZoom, layer)) return false;

#ifndef REALOS
    if (fullBuf->depth() != COLOR_24_BIT) fullBuf->flipBytes();
    else 
    {
        if (getColorSpace () == ITULab ||
            ( layer == 1 && m_bg->getColorSpace() == ITULab ) ||
            ( layer > 2 && m_sel[layer / 2 - 1]->getFgColorSpace() == ITULab) ) 
        {
            if (m_bg && m_bg->numDecParams() == 6)
                fullBuf->ITULab2RGB(m_bg->getDecParam(0), m_bg->getDecParam(1), m_bg->getDecParam(2),
                                m_bg->getDecParam(3), m_bg->getDecParam(4), m_bg->getDecParam(5));
            else if (m_sel[0] && m_sel[0]->numDecParams() == 6)
                fullBuf->ITULab2RGB(m_sel[0]->getDecParam(0), m_sel[0]->getDecParam(1), m_sel[0]->getDecParam(2),
                                m_sel[0]->getDecParam(3), m_sel[0]->getDecParam(4), m_sel[0]->getDecParam(5));
            else
                fullBuf->ITULab2RGB();
        }
    }
#endif

		return true;
}

bool MrcPage::renderBW( ImageBuf * buf, const Asd::Rect & region, Asd::Float64 zoom, Asd::Int32 layer)
{

    const Asd::UInt32 denom = 210; // 2 * 3 * 5 * 7

    AdapterFactory fac;
    ImageAdapter *fullBuf;
    DibImgBuf dib;
    
    if (getNumLayers() < layer) return false;

    // allocate the correct adpter, based on the color depth of this page
    ColorDepth depth = getColorDepth ();
    Asd::Rect roi(region);
    if (layer > 0)
    {
        if (layer % 2 == 0) depth = MONOCHROME;
        else
        {
            Asd::Int32 resolution = 1;
            if (layer == 1)
            {
                if (m_bg) resolution = m_bg->getResolution().getX();
            }
            else             
                resolution = m_sel[layer / 2 - 1]->getFgResolution().getX();
            
            zoom /= (Asd::Float64) resolution;
        }
    }

    // convert float to rational
    Asd::UInt32 numer = (Asd::UInt32) floor(denom * zoom + 0.5); 
    Asd::Rational<Asd::Int32> ratZoom (numer, denom);
    Asd::Rational<Asd::Int32> oneZoom (1);

    if (depth == MONOCHROME && ratZoom != oneZoom)
        depth = COLOR_8_BIT;

    if (depth > MONOCHROME)
    {
        // gray or 24bit color
        fullBuf = fac.createInstance(depth, roi, &dib);
    }
    else
    {
        // monochrome
        fullBuf = fac.createInstance(depth, roi, buf);
    }

    // set to default color 
    if (layer <= 1)            
        fullBuf->fillWhite();        
    else if (layer % 2 == 1)         
        fullBuf->fillBlack();
    
    if(!renderInternal(fullBuf, roi, ratZoom, layer)) return false;

#ifndef REALOS
    if (depth != COLOR_24_BIT) fullBuf->flipBytes();
    else 
    {
        if (getColorSpace () == ITULab ||
            ( layer == 1 && m_bg->getColorSpace() == ITULab ) ||
            ( layer > 2 && m_sel[layer / 2 - 1]->getFgColorSpace() == ITULab) ) 
        {
            if (m_bg && m_bg->numDecParams() == 6)
                fullBuf->ITULab2RGB(m_bg->getDecParam(0), m_bg->getDecParam(1), m_bg->getDecParam(2),
                                m_bg->getDecParam(3), m_bg->getDecParam(4), m_bg->getDecParam(5));
            else if (m_sel[0] && m_sel[0]->numDecParams() == 6)
                fullBuf->ITULab2RGB(m_sel[0]->getDecParam(0), m_sel[0]->getDecParam(1), m_sel[0]->getDecParam(2),
                                m_sel[0]->getDecParam(3), m_sel[0]->getDecParam(4), m_sel[0]->getDecParam(5));
            else
                fullBuf->ITULab2RGB();
        }

        fullBuf->convertToBW (buf);
    }

    if (depth == COLOR_8_BIT)
    {
        // convert from an IPCore adapter to Intel
        Asd::Rect bandRoi(fullBuf->width(), fullBuf->height());
        IntelImgBuf intelBuf (bandRoi, *fullBuf);
        intelBuf.convertToBW (buf);
    }

#endif

		return true;
}


ColorSpace MrcPage::getColorSpace ( void ) 
{
    ColorSpace colorSpace = Default;
    Asd::UInt32 initialized = 0;

    if (m_bg) 
    {
        colorSpace = m_bg->getColorSpace();    
        initialized = 1;
    }

    Asd::Int32 nLayer = 0;   
    while (nLayer < m_sel.size())    
    {    
        if (m_sel[nLayer])
        {
            if (initialized)
            {
                // if there different color spaces are used in the file, indicate that by
                // having a "Default" color space
                if (m_sel[nLayer]->getFgColorSpace() != colorSpace)
                {
                    colorSpace = Default;
                    break;
                }
            }
            else
            {
                colorSpace = m_sel[nLayer]->getFgColorSpace();    
                initialized = 1;
            }
        }
        nLayer++;
    }

    return colorSpace;
}
