#include "StdAfx.h"

#include "TifFile.h"
#include "TifPage.h"

namespace ipcore
{
#include "ip-tiffio.h"
#include "dgp_err_handle.h"
};


TifFile::TifFile( const char* file )
:
 m_fileName(file),
 m_dgpState(NULL),
 m_tiff(NULL)
{
	ipcore::Dgp_initDgpState( &m_dgpState, NULL, NULL, stderr,
		ipcore::Dgp_fileErrPrinter, true, true );

	if( m_dgpState )
	{
		m_tiff = ipcore::TIFFOpen( m_dgpState->ipwstate, file, "r" );
	}
	

	for(ipcore::uint32 page=0; ipcore::TIFFSetDirectory(m_tiff, page); ++page)
	{
		TifPage * tp = new TifPage( this, page );
	    m_pages.push_back( tp );
	}
}

TifFile::~TifFile( void )
{
	purgeAllPages();

	ipcore::TIFFClose( m_tiff ); m_tiff = NULL;

	ipcore::Dgp_cleanupDgpState( &m_dgpState ); m_dgpState = NULL;
}


Asd::Point32 TifFile::getPageSize( Asd::UInt32 page )
{
	Asd::Point32 size;

	if( m_tiff && (page < numPages()) )
	{
	
		ipcore::TIFFSetDirectory( m_tiff, page );

		ipcore::uint32 wid = 0;
		ipcore::uint32 len = 0;

		ipcore::TIFFGetField( m_tiff, TIFFTAG_IMAGEWIDTH,  &wid );
		ipcore::TIFFGetField( m_tiff, TIFFTAG_IMAGELENGTH, &len );	

		size.setXY( wid, len );
	}

	return size;
}

ColorDepth TifFile::getPageColorDepth( Asd::UInt32 page )
{
	ColorDepth depth = COLOR_DEPTH_ERROR;

	if( m_tiff && (page < numPages()) )
	{
		ipcore::TIFFSetDirectory( m_tiff, page );

		ipcore::uint16 bps = 0;
		ipcore::uint16 spp = 0;

		ipcore::TIFFGetField( m_tiff, TIFFTAG_BITSPERSAMPLE,   &bps );
		ipcore::TIFFGetField( m_tiff, TIFFTAG_SAMPLESPERPIXEL, &spp );
	
		depth = int2ColorDepth( bps * spp );
	}

	return depth;
}

void TifFile::purgeAllPages( void )
{
    // delete all pages and set to NULL
    for( std::vector<TifPage*>::iterator page = m_pages.begin(); page != m_pages.end(); ++page )
    {
        delete (*page);
        *page = NULL;
    }
}
