#include "StdAfx.h"


#include "TifPage.h"
#include "TifFile.h"

namespace ipcore
{
	#include "ip-tiffio.h"
	#include "dgp_err_handle.h"
	#include "rescale.pub"
	#include "prxform.pub"
};

#ifndef REALOS
#include "IntelImgBuf.h"
#endif

#ifdef PSG
#undef cFalse //originally defined as ((Bool) 0) which make the compiler complain
#define cFalse 0
#endif

TifPage::TifPage( TifFile * file, Asd::UInt32 page )
:
 m_tifFile( file ),
 m_pageNum( page )
{
}

TifPage::~TifPage( void )
{
}

Asd::Point32 TifPage::getSize( void ) const
{ 
	Asd::Point32 size;

	if( m_tifFile )
	{
		size = m_tifFile->getPageSize( m_pageNum );
	}

	return size;
}

ColorDepth TifPage::getColorDepth( void )
{
	return  m_tifFile? m_tifFile->getPageColorDepth(m_pageNum) : 
	                   COLOR_DEPTH_ERROR;
}



bool TifPage::
render( ImageBuf * buf, const Asd::Rect & region, Asd::Float64 zoom, Asd::Int32 /*layer*/ )
{
	ipcore::TIFF * const tifPtr = m_tifFile? m_tifFile->m_tiff : NULL;

	ipcore::Dgp_State * dgp_state = m_tifFile? m_tifFile->m_dgpState : NULL;

	if( buf && tifPtr && tifPtr && dgp_state )
	{		
		const size_t nameBufLen = m_tifFile->m_fileName.length() + 1;

		ipcore::Dgp_Int8 * name = new ipcore::Dgp_Int8[nameBufLen];

		{
			const char * strPtr = m_tifFile->m_fileName.data();

			for( size_t ii=0; ii < nameBufLen; ++ii, ++strPtr )
			{
				name[ii] = *strPtr;
			}

			name[nameBufLen - 1] = '\0'; // Just to make sure.
		}

		ipcore::Dgp_Int32 rot   = 0;
		ipcore::Dgp_Int32 xSize = 0;
		ipcore::Dgp_Int32 ySize = 0;
		ipcore::Dgp_Int32 bpl   = 0;

		ipcore::PIXR * pixr = NULL;
		ipcore::PIXR * tmpPixr = NULL;
		 
		ipcore::readTIFFEx(	name, rot, m_pageNum + 1, &xSize, &ySize, &bpl, &pixr, dgp_state );

		{
			ipcore::Int32 pixrChannelCount = 0;
			ipcore::Int32 pixrDepth = 0;

			ipcore::w_pixrGetDepth( dgp_state->ipwstate, pixr, &pixrDepth );
			ipcore::w_pixrGetChannelCount( dgp_state->ipwstate, pixr, &pixrChannelCount );
		
			ColorDepth cd = int2ColorDepth( pixrChannelCount * pixrDepth );

			assert( cd );

			if( zoom > 1.0 )
			{
				if( cd == MONOCHROME )
				{
					cd = COLOR_8_BIT;
					ipcore::w_depth1To8( dgp_state->ipwstate, pixr, true, &tmpPixr );

					if( tmpPixr )
					{
						ipcore::w_destroyPixr( dgp_state->ipwstate, &pixr );
						pixr = tmpPixr;
						tmpPixr = NULL;
					}
				}

				xSize /= zoom;
				ySize /= zoom;

				ipcore::w_scaleLinearSize( dgp_state->ipwstate, pixr, xSize, ySize, &tmpPixr );

				if( tmpPixr )
				{
					ipcore::w_destroyPixr( dgp_state->ipwstate, &pixr );
					pixr = tmpPixr;
					tmpPixr = NULL;
				}
			}

			ipcore::w_cutRectIP( dgp_state->ipwstate, region.left(), region.top(),
				region.width(), region.height(), pixr );

			buf->resize( region.size(), cd, sizeof(Asd::UInt32) );

			if( (cd == MONOCHROME) || (cd == COLOR_8_BIT) )
			{
				ipcore::w_pixrToBMP( dgp_state->ipwstate, pixr, cFalse, 0, 0, buf->data() );
			}
			else if( cd == COLOR_24_BIT )
			{
				ipcore::w_RGBPixrsToBMP( dgp_state->ipwstate, pixr, 0, 0, buf->data() );
			}

		}

		DibImgBuf * dib = dynamic_cast<DibImgBuf *>(buf);

		if( dib )
		{
			dib->setBottomUp();
		}
		
		ipcore::w_destroyPixr( dgp_state->ipwstate, &pixr );

		delete [] name; name = NULL;

		return true;
	}	
    
    return false;
}

bool TifPage::
renderBW( ImageBuf * buf, const Asd::Rect & region, Asd::Float64 zoom, Asd::Int32 /*layer*/ )
{
#ifndef REALOS
	DibImgBuf clrRendBuf;

	render( &clrRendBuf, region, zoom, 0 );

	if( clrRendBuf.depth() == MONOCHROME )
	{
		buf->copyFrom( clrRendBuf );
	}
	else
	{
		DibImgBuf clrCopyBuf;

		AdapterFactory fact;

		ImageAdapter * iaPtr = fact.createInstance( clrRendBuf.depth(), region, &clrCopyBuf );

		assert( iaPtr );

		Asd::UInt32 crbh = clrRendBuf.height();
		Asd::UInt32 ccbh = clrCopyBuf.height();

		//assert( clrRendBuf.height() == clrCopyBuf.height() );

		if( clrRendBuf.dataSize() == clrCopyBuf.dataSize() )
		{
			memcpy( clrCopyBuf.data(), clrRendBuf.data(), clrCopyBuf.dataSize() );
		}
		else
		{
			const Asd::UInt32 rows = std::_cpp_min(clrRendBuf.height(), clrCopyBuf.height());

			const Asd::UInt32 rendLen = clrRendBuf.bpl();
			const Asd::UInt32 copyLen = clrCopyBuf.bpl();
			const Asd::UInt32 rowLen = std::_cpp_min( rendLen, copyLen );

			Asd::UInt8 * rendPtr = clrRendBuf.data();
			Asd::UInt8 * copyPtr = clrCopyBuf.data();

			for( Asd::UInt32 row=0; row < rows; ++row )
			{
				memcpy( copyPtr, rendPtr, rowLen );	

				copyPtr += copyLen;
				rendPtr += rendLen;
			}
		}

		iaPtr->convertToBW( buf );
#ifndef REALOS
		if( clrCopyBuf.depth() == COLOR_8_BIT )
		{
			// convert from an IPCore adapter to Intel
			Asd::Rect roi( iaPtr->width(), iaPtr->height() );
			IntelImgBuf intelBuf( roi, *iaPtr );
			intelBuf.convertToBW( buf );
		}
#endif
	}
	

	DibImgBuf * dib = dynamic_cast<DibImgBuf *>(buf);

	if( dib )
	{
		dib->setBottomUp();
	}

#endif
	return true;
}
