#include "StdAfx.h"
#include "TfxBand.h"
#include "MrcFile.h"
#include "DibImgBuf.h"
#include <algorithm>
#include <vector>

#include "AsdSilx.h"

namespace AsdSilx{
#include "decode-colors.h"
#include "uncomp-jbig.h"
#include "memorydatasource.h"
#include "segment.h"
#include "list.h"
#include "blit_rgb.h"
#include "colorBlit.h"
}
using namespace AsdSilx;

class ColorBox{
public:
    ColorBox(const Asd::Rect& rect, Asd::UInt32* bits, const Asd::Int32& lpl,
        const Asd::Int32& r, const Asd::Int32& g, const Asd::Int32& b):
    m_box(rect), m_bits(bits), m_lpl(lpl), m_r(r), m_g(g), m_b(b){};
    ColorBox():m_bits(NULL), m_lpl(0), m_r(0), m_g(0), m_b(0){};
    const Asd::Rect& getBox(){ return m_box;}
    
    bool operator <( ColorBox& cb2 ) const { 
        Asd::Rect box = cb2.getBox();
        if( (m_box.top() < box.top()) ||
            (m_box.left() < box.left() && m_box.top() == box.top()) ||
            (m_box.left() == box.left() && m_box.top() == box.top() &&
            m_box.width()< box.width() && m_box.height() < box.height())
            )
            return true;
        else return false;
    }
    bool operator ==( ColorBox& cb2 ) const { 
        Asd::Rect box = cb2.getBox();
        return (m_box == box);
    }
    bool operator >( ColorBox& cb2 ) const { 
        Asd::Rect box = cb2.getBox();
        if( (m_box.top() > box.top()) ||
            (m_box.left() > box.left() && m_box.top() == box.top()) ||
            (m_box.left() == box.left() && m_box.top() == box.top() &&
            m_box.width() > box.width() && m_box.height() > box.height())
            )
            return true;
        else return false;
    }
    void draw(Asd::Rect& roi){
        
    }
    Asd::Int32 getR(){ return m_r;}
    Asd::Int32 getG(){ return m_g;}
    Asd::Int32 getB(){ return m_b;}
    
    Asd::UInt32* getBits(){   return m_bits;}
    Asd::Int32 getlpl()   {   return m_lpl;}
    Asd::Rect getOriginalBox(){ return m_box;};

    void scaleBox(Asd::Point32 scale){
        m_box.scaleDownMax(scale.getX(), scale.getY());   
    }
    
    
private:
    Asd::Rect m_box;
    Asd::UInt32 *m_bits;
    Asd::Int32 m_lpl;
    Asd::Int32 m_r;
    Asd::Int32 m_g;
    Asd::Int32 m_b;
};

typedef std::vector<ColorBox> CBList;

bool TfxColorTokenBand::
render(ImageAdapter * buffer, const Asd::Rect & region, const Asd::Point32 & scale)     
{
    
    if(!m_data) return true;
    if(!m_selBand) return true;
    if(!buffer) return true;
    
    TfxLayer* pLayer = getLayer();
    if(!pLayer) return false;
    MrcFile *pFile = pLayer->getFile();
    if(!pFile) return false;
    
    ((TfxJBig2Band*)m_selBand)->decodeJBig2(true, false, NULL);
    
    jb2info* info = (jb2info *)(((TfxJBig2Band*)m_selBand)->getJb2Info());
    if(!info) return false;
    
    Dgp_State* dgp_state= info->dgp_state;
    if(!dgp_state) return false;
    
    
    // check what we need to do
    bool needDownScale = false;
    bool needClip = false;
    if (scale.getX() < -1 ||  scale.getY() < -1) 
    {
        return false;
    }
    if (scale.getX() > 1 ||  scale.getY() > 1) 
    {
        needDownScale = true;
    }
    if (region.width() < getSize().getX() / scale.getX() || 
        region.height() < getSize().getY() / scale.getY() )
    {
        needClip = true;
    }
    
    Asd::Rect roi(region);
    Asd::Rect scaledRoi(region);

    // scale roi to orignal page resolution
    roi.scaleUp(scale.getX(), scale.getY());

    //split a large region into sub sections
    Asd::Int32 rh = 0;
    Asd::Int32 h =  0;
    std::vector<Asd::Rect> subRoi, scaledSubRoi;
    if(roi.height() > 5000) {
        rh =region.height() / 4;
        h  = rh * scale.getY();
        subRoi.push_back(Asd::Rect(roi.top(),         roi.left(), roi.top() + h -1,    roi.right()));
        subRoi.push_back(Asd::Rect(roi.top() + h,     roi.left(), roi.top() + 2*h -1,  roi.right()));
        subRoi.push_back(Asd::Rect(roi.top() + 2 * h, roi.left(), roi.top() + 3*h -1,  roi.right()));
        subRoi.push_back(Asd::Rect(roi.top() + 3 * h, roi.left(), roi.bottom(),        roi.right()));

        scaledSubRoi.push_back(Asd::Rect(region.top(),          region.left(),  region.top() + rh -1,       region.right()));
        scaledSubRoi.push_back(Asd::Rect(region.top() + rh,     region.left(),  region.top() + 2 * rh -1,   region.right()));
        scaledSubRoi.push_back(Asd::Rect(region.top() + 2 * rh, region.left(),  region.top() + 3 * rh -1,   region.right()));
        scaledSubRoi.push_back(Asd::Rect(region.top() + 3 * rh, region.left(),  region.bottom(),            region.right()));
    } else if( roi.height()> 3000){
        rh =region.height() / 2;
        h  = rh * scale.getY();
        subRoi.push_back(Asd::Rect(roi.top(),     roi.left(), roi.top() + h -1,    roi.right()));
        subRoi.push_back(Asd::Rect(roi.top() + h, roi.left(), roi.bottom(),        roi.right()));

        scaledSubRoi.push_back(Asd::Rect(region.top(),  region.left(),      region.top() + rh -1, region.right()));
        scaledSubRoi.push_back(Asd::Rect(region.top() + rh, region.left(),  region.bottom(),      region.right()));
    } else{
        subRoi.push_back(Asd::Rect(roi));
        scaledSubRoi.push_back(Asd::Rect(region));
    }
    // create a buffer of original resolution
    AdapterFactory fac;
    DibImgBuf* dibFg = new DibImgBuf();
    ImageAdapter *fgBuf;
    fgBuf = fac.createInstance (COLOR_24_BIT, subRoi[0], dibFg);

    CBList::iterator  iter1;
    CBList cbList;
    
    //decode color tags 
    color_tag *tags = NULL;
    Dgp_Int32 ntags = 0;
    color_cookie *colorState = (color_cookie *)NULL;
    SILX_ERR_CHECK(
        initColors( &colorState, DGP_FALSE, dgp_state));
    SILX_ERR_CHECK( decodeJbig2Colors(  colorState, 
        m_data,
        m_dataSize,
        &tags, 
        &ntags));
    SILX_ERR_CHECK(ntags <= 0)
        
        
    // render the color boxes into buffer
    // using token bits as a indicator for overlapping boxes
    
    // looking for region band
    jb2Segment *segment;
    jb2RegSegment *reginfo; 
    Dgp_Int32 ninstances; 
    jb2SymbolInstance *instances; 
    Dgp_Int32 i; 
    
    ninstances = 0;
    instances = NULL;
    reginfo = NULL;
    if (info->activesegs != DGP_NULLLIST) {
        DgpAdt_ListNode ln;
        DGPADT_FOREACH (ln, info->activesegs) {
            segment = (jb2Segment *)DGPADT_LIGET(ln);
            if(segment && (segment->segtype == JB2SEGTYPE_SYMREG_INT ||
                segment->segtype == JB2SEGTYPE_SYMREG_IMM ||
                segment->segtype == JB2SEGTYPE_SYMREG_IMMLL )){
                ninstances = segment->data.symreg.ninstances; 
                instances = segment->data.symreg.instances; 
                reginfo = &segment->reginfo; 
                for(i =0; i<ninstances; i++){
                    Asd::Rect box((instances[i].sym)->width,(instances[i].sym)->height);
                    box.offsetXY(instances[i].xpos, instances[i].ypos);

                    //create a colorBox object, attatch the object with the token bits
                    ColorBox cb(  
                        box,(instances[i].sym)->bits,
                        ( instances[i].sym)->longsperline,
                        ( tags[i].value).rgb_color.r, 
                        ( tags[i].value).rgb_color.g,
                        ( tags[i].value).rgb_color.b );     
                    
                    box = cb.getBox();
                    Asd::Rect boxRoi = box.intersection(roi);
                    
                    // add the box to a list
                    if(!boxRoi.isEmpty())
                        cbList.push_back(cb);
                    
                }
            }
        }
    }
    
    if(tags){
        DGP_FREE(tags);
        tags = NULL;
    }
    for(i=0; i< subRoi.size(); i++){
        if(i == subRoi.size()-1){
            delete dibFg;
            dibFg = new DibImgBuf();
            fgBuf = fac.createInstance (COLOR_24_BIT, subRoi[i], dibFg);
        }
        memset ((void*) dibFg->data(), 0x00, dibFg->dataSize());

        for ( iter1 = cbList.begin(); iter1 != cbList.end(); iter1++ ) {   
            Asd::Rect box((*iter1).getBox());
            Asd::Rect boxRoi(box.intersection(subRoi[i]));
            if(!boxRoi.isEmpty()){
                DgpImage_RGB val;
                if(pFile->isBGR()){
                    val.r = (*iter1).getB(); 
                    val.g = (*iter1).getG();
                    val.b = (*iter1).getR();
                }else{
                    val.r = (*iter1).getR();
                    val.g = (*iter1).getG();
                    val.b = (*iter1).getB();
                }
                boxRoi.offsetXY(-subRoi[i].left(), -subRoi[i].top());
                DgpBlat_colorBlitBinaryRGB((*iter1).getBits(),  //src
                               (*iter1).getlpl() * sizeof(Dgp_UInt32),
                               box.left() - subRoi[i].left(), box.top() - subRoi[i].top(),
                               (DgpImage_RGB* )fgBuf->data(),
                               fgBuf->bpl(),
                               0, 0, 
                               boxRoi.left(), boxRoi.top(),
                               boxRoi.width(), boxRoi.height(),
                               val);
            }
        }
        // generate image buffer starting at roi.top
        Asd::Rect rect(scaledSubRoi[i]);
        rect.offsetXY(-scaledRoi.left(), -scaledRoi.top());
        ImageAdapter *ibuf = fac.createSub (rect, buffer);
        fgBuf->scaleToSize(ibuf);
    }

    // cleanup
    delete dibFg;
    cleanupColors(colorState);
    colorState = (color_cookie *)NULL;
    
    return true;
    
ERR_CLEANUP:
    if(tags){
        DGP_FREE(tags);
        tags = NULL;
    }
    if(dibFg) delete dibFg;
    if(colorState) cleanupColors(colorState);
    return false;
}

