#include "StdAfx.h"

#include "TfxBand.h"
#include "SimpImgBuf.h"
#include "IPCoreImgBuf.h"

// IPCORE G4 libs
namespace ipcore
{
#include "frames.pub"
#include "shrcmprs.pub"
#include "ccitt.pub"
#include "mmr.pub"

#include "geomadt.h"
#include "props.pub"

#include "rast.pub"

#include "reduce.pub"

#include "prxform.pub"

#include "rescale.pub"

#include "shrpixr.pub"
}

#include "AsdMath.h"

#ifdef PSG
#undef cFalse //formerly defined as ((Bool) 0) which causes a parse error
#define cFalse 0
#endif

using namespace AsdMath;

bool TfxG4Band::render ( ImageAdapter * buf, const Asd::Rect & region, 
						  const Asd::Point32 & scale )
{

    const Asd::Int32 framePix = 32;

	// this implementation is for IPCore Image Buffers only
    IPCoreImgBuf* ibuf = dynamic_cast<IPCoreImgBuf*> (buf);
    if (!ibuf) return false;

    // ipcore setup
    ipcore::IPWState *pIPWState = ibuf->getState();
    Asd::Int32 status = 0;

    // check what we need to do
    bool needDownScale = false;
    bool needClip = false;
    if (scale.getX() < -1 ||  scale.getY() < -1) 
    {
        return false;
    }
    if (scale.getX() > 1 ||  scale.getY() > 1) 
    {
        needDownScale = true;
    }
    if (region.width() < getSize().getX() / scale.getX() ||
        region.height() < getSize().getY() / scale.getY() )
    {
        needClip = true;
    }

    // allocate temporary PIXR
    ipcore::PIXR* tempBuf = NULL;
    Asd::UInt8 *data;
    Asd::Int32 bpl;
    status += ipcore::w_createPixr (pIPWState, getSize().getX(), getSize().getY(), MONOCHROME,
        ipcore::pixrRes, ipcore::pixrRes, framePix, &tempBuf);
    status += ipcore::w_pixrGetImage (pIPWState, tempBuf, &data);
    status += ipcore::w_pixrGetBpl (pIPWState, tempBuf, &bpl);

    // find first and last row of the band that we need
    Asd::UInt32 firstRow = 0;
    Asd::UInt32 nRow = getSize().getY();

	// allocate decode structures
    Asd::Int32 stateBlockSize;
    status += ipcore::i_getSizeOfCCITTDecompressionInfo( pIPWState,
	    getSize().getX() + 2 * framePix, &stateBlockSize);
    void *compInfo = (void*) new Asd::UInt8 [stateBlockSize];
    status += ipcore::i_initializeCCITTDecompressionInfo(
	    pIPWState,
	    compInfo, getSize().getX() + 2 * framePix, 1,
	    cNoPad, stateBlockSize);

    // clear the image
    status += ipcore::w_clear (pIPWState, tempBuf);

    // set up for decode
    Asd::UInt32 imgExh = 0, cmpExh = 0;
    Asd::UInt8 *iPtr = data - framePix / 8;
    Asd::UInt8 *cPtr = m_data;
    Asd::Int32 bytesLeft = m_dataSize, bytesUsed, pct;
    bool contImg = 0, contCmp = 0;
    
    // call the decoder
    do {
#ifndef PSG
        status += ipcore::i_decompressRowsMMR (pIPWState, compInfo,
	    iPtr, firstRow,
        nRow, bpl, cFalse, cPtr, bytesLeft, contImg,
	    contCmp, 1, &imgExh, &cmpExh, &pct, &bytesUsed);
#else //compiler complains because Bool is typed differently in IPCORE-PSG and IPCORE-PARC
	status += ipcore::i_decompressRowsMMR (pIPWState, compInfo,
					       iPtr, firstRow,
					       nRow, bpl, cFalse, cPtr, bytesLeft, contImg,
					       contCmp, 1, ((int *) &imgExh), ((int *) &cmpExh), &pct, &bytesUsed);
#endif
    	bytesLeft -= bytesUsed;

	    if (imgExh == 1 && cmpExh == 0) {

	        /* Okay, we can live with it, but it's wrong */

	    } else if (cmpExh == 1 && imgExh == 0) {

            return false;
        }
    	contImg = 1;
	    contCmp = 1;
    } while (imgExh == 0 && cmpExh == 0 && bytesLeft > 0);

    if (status) return false;

    // clip in with and height
    if (needClip)
    {
        ipcore::PIXR *clipTempBuf = NULL;

        Asd::UInt32 clipWidth = region.width() * scale.getX();
        Asd::UInt32 clipHeight = region.height() * scale.getY();
        status += ipcore::w_createPixr (pIPWState, clipWidth, clipHeight, MONOCHROME,
            ipcore::pixrRes, ipcore::pixrRes, framePix, &clipTempBuf);

        status += ipcore::w_rasterop ( pIPWState, region.left() * scale.getX(), 
            region.top() * scale.getY(), 
            tempBuf, PIX_SRC, 0, 0, clipWidth, clipHeight,
            clipTempBuf);

        status += ipcore::w_destroyPixr(pIPWState, &tempBuf);
        tempBuf = clipTempBuf;
    }

    if (status) return false;

    // max downsample with this method is 8x ...  FIXME
    if (needDownScale)
    {
        Asd::UInt32* scaleTable;
        Asd::UInt32 resScaleX = 1, resScaleY = 1;
        Asd::UInt32 initialScale = MIN(scale.getX(), scale.getY());
        if (initialScale > 8) initialScale = 8;

        if (initialScale < scale.getX())
        {
            resScaleX = scale.getX() / initialScale;
        }
        if (initialScale < scale.getY())
        {
            resScaleY = scale.getY() / initialScale;
        }

        // IPCore convention
        initialScale *= 10;

        status += ipcore::w_makeScaleToGray4bppTable (pIPWState, initialScale, &scaleTable);
        ipcore::PIXR *tempPixr = NULL;
        status += ipcore::w_scaleToGray4bpp (pIPWState, tempBuf, initialScale,
            scaleTable, &tempPixr);
//        IPW_FREE (scaleTable, cAllocatorData);
     	status += ipcore::w_destroyPixr(pIPWState, &tempBuf); 

        if (resScaleY == 1 && resScaleX == 1)
        {
            status += ipcore::w_depth4To8 (pIPWState, tempPixr, &tempBuf);
        	status += ipcore::w_destroyPixr(pIPWState, &tempPixr); 
        }
        else
        {
            // do a linear resize to fit into the ROI
            ipcore::PIXR *ttempPixr = NULL;

            status += ipcore::w_depth4To8 (pIPWState, tempPixr, &ttempPixr);
        	status += ipcore::w_destroyPixr(pIPWState, &tempPixr); 

            status += ipcore::w_scaleLinearSize (pIPWState, ttempPixr, region.width(),
                region.height(), &tempBuf);
        	status += ipcore::w_destroyPixr(pIPWState, &ttempPixr); 
        }
    }

    if (status) return false;

    // copy the data into the output buffer
    ipcore::PIXR* ttempPixr = ibuf->pixr ();

    Asd::Int32 ws, hs, depth, bpls, bpld;
    Asd::UInt8 *pImages, *pImaged;
    status += ipcore::w_pixrGetDepth(pIPWState, tempBuf, &depth);
    status += ipcore::w_pixrFramedWidth(pIPWState, tempBuf, &ws);
	status += ipcore::w_pixrFramedHeight(pIPWState, tempBuf, &hs);
	status += ipcore::w_pixrGetImageAll(pIPWState, tempBuf, &pImages);
	status += ipcore::w_pixrGetImage(pIPWState, ttempPixr, &pImaged);
	status += ipcore::w_pixrGetBpl(pIPWState, tempBuf, &bpls);
	status += ipcore::w_pixrGetBpl(pIPWState, ttempPixr, &bpld);

	/* blit the inner part of the src, exclusive of cFrameBits,
	 * into the dest.  */
	status += ipcore::i_removeFrame (pIPWState, pImages,
					 ws, hs, depth, bpls,
					 pImaged, bpld);

    // cleanup
    delete [] compInfo;
    status += ipcore::w_destroyPixr(pIPWState, &tempBuf);

    // check if there were any errors
    if (status) return false;

    return true;
}
