#include "StdAfx.h"

#ifdef _WIN32
// Warning 4786 appears to be benign, and we get tons of them,
// so we disabled it.
#pragma warning(disable:4786)
#endif

#include "TfxBand.h"
#include "AsdMath.h"

#include "IPCoreImgBuf.h"
namespace ipcore
{
    #include "libadtw.pub"
    #include "librastw.pub"
    #include "libpixrw.pub"
    #include "libscalw.pub"
    #include "libipw.pub"
    
    // silx header that relate to ipcore
    #include "image.h"
    #include "image_defs.h"
    #include "simple_utils.h"
    #include "../libjbig2/uncomp-jbig.h"
    
#define SILX_ERR_CHECK(status)  \
    if ((status) != cSuccessful){ \
    goto ERR_CLEANUP; \
    } \
    
#define SILX_STATUS_CHECK(status)  \
    if (!(status)){ \
    goto ERR_CLEANUP; \
    }
	
#define SILX_NULL_CHECK(status)  \
    if (!(status)){ \
	goto ERR_CLEANUP; \
	}
}

#include "MrcFile.h"
#include "TfxLayer.h"
#include "Jbig2Dict.h"
#include "PsgFix.h"

using namespace AsdMath;
using namespace std;
using namespace ipcore;

ErrCode freeImage( Dgp_State * dgp_state,
                  DgpImage_Packed* dgpImage)
{
    DGP_PROLOGUE("freeImage()",DGP_PERRSTATE);
    DGP_CHECK_NULL(dgpImage);
    if(*dgpImage) {
            Dgp_UInt32 *retval;
            DgpImage_unwrapAll (*dgpImage, &retval);
            *dgpImage= NULL;
    }
    NORMAL_RETURN;

}
ErrCode allocImage( 
                   Dgp_State * dgp_state,
                   Dgp_Int32 width, Dgp_Int32 height,
                   DgpImage_Packed* dgpImage, void** usrImage)
{
    DGP_PROLOGUE("allocImage()",DGP_PERRSTATE);
    PIXR* pixr = NULL;
    DgpImage_Packed image = NULL;
    Dgp_UInt8* ptr = NULL;
    IPWState* pIPWState;
    Dgp_Int32 bpl;

    DGP_CHECK_NULL(dgp_state);
    DGP_CHECK_NULL(usrImage);
    DGP_CHECK_NULL(dgpImage);
    pIPWState = dgp_state->ipwstate;

    DGP_ERR_PASS_CLEANUP( ipcore::w_createPixr (pIPWState, width, height, MONOCHROME,
            ipcore::pixrRes, ipcore::pixrRes, 32, &pixr));
	DGP_ERR_PASS_CLEANUP( ipcore::w_pixrGetImage(pIPWState, pixr, &ptr));
	DGP_ERR_PASS_CLEANUP( ipcore::w_pixrGetBpl(pIPWState, pixr, &bpl));

    DGP_ERR_PASS_CLEANUP( DgpImage_wrapAll( dgp_state, 
                                            (Dgp_UInt32*)ptr, 
                                            DGPIMAGE_PACKED, 
                                            width, height, 
                                            bpl, (void**)(&image) ));
    *usrImage = (void *) pixr;
    *dgpImage = image;
    NORMAL_RETURN;
ERR_CLEANUP:
	if(pixr)  ipcore::w_destroyPixr(pIPWState, &pixr);
    if(image) {
            Dgp_UInt32 *retval;
            DgpImage_unwrapAll (image, &retval);
    }
    ERR_RETURN;
}

bool TfxJBig2Band::decodeJBig2(bool keepAll, bool render, void** srcPixr)
{

    // validation

    TfxLayer* pLayer = getLayer();
    if(!pLayer) return false;

    MrcFile *pFile = pLayer->getFile();
    if(!pFile) return false;

    Jbig2Dict *pDict = pFile->getDict();
    if(!pDict) return false;

	// decode info setup
	jb2info *passedInfo = (jb2info *)(pDict->getJb2Info());
	if (!passedInfo) return false;
    
    Dgp_State *dgp_state = passedInfo->dgp_state;
	if (!dgp_state) return false;

    // First we need to decode the global dictionary if it is not already decoded
    for(int i = 0; i < m_dictInfo.getLibCount(); i++){
       bool status = decodeSymDict(pDict, (m_dictInfo.getLibIds())[i], (m_dictInfo.getLibSizes())[i]);
       if(!status) return false;
    }
	
    if(m_jb2Info) cleanupJb2Info((jb2info**)(&m_jb2Info));

	// create an SilxImageBuf and passed it to jb2 processor to render on
	SILX_ERR_CHECK(
		initJb2Info((jb2info**)&m_jb2Info,DGP_FALSE, NULL, NULL, DGP_TRUE, 6500, 
                        keepAll, render, allocImage, freeImage, dgp_state));

	//Copy segments (dictionary) from passedInfo to curInfo
    SILX_ERR_CHECK(copySegmentList((jb2info*)m_jb2Info, passedInfo));
	// Process all segments
    SILX_ERR_CHECK(processSegmentsFromData((jb2info*)m_jb2Info, m_data, m_dataSize, srcPixr));

    return true;
ERR_CLEANUP:
	if(m_jb2Info) cleanupJb2Info((jb2info**)&m_jb2Info);
    return false;
}


bool TfxJBig2Band::render ( 
						   ImageAdapter * buf, 
						   const Asd::Rect & region,
						   const Asd::Point32 & scale)
{   
    // check what we need to do
    bool needDownScale = false;    
    bool needClip = false;
    if (scale.getX() < -1 ||  scale.getY() < -1) {
        return false;
    }
    if (scale.getX() > 1 ||  scale.getY() > 1){
        needDownScale = true;
    }
    if (region.width() < getSize().getX() / scale.getX() ||
        region.height() < getSize().getY() / scale.getY() ) {
        needClip = true;
    }

    // this implementation is for IPCore Image Buffers only
    IPCoreImgBuf* ibuf = dynamic_cast<IPCoreImgBuf*> (buf);
    if (!ibuf) return false;

    // ipcore setup
    ipcore::IPWState *pIPWState = NULL;
	const Asd::Int32 framePix = 32;
	ipcore::PIXR* srcPixr = NULL;
    ipcore::PIXR *clipTempBuf = NULL;  
    ipcore::PIXR *ttempPixr = NULL;
    ipcore::PIXR *tempPixr = NULL;
		
    //decodeJBig2
    SILX_ERR_CHECK(decodeJBig2(false, true, (void**)&srcPixr) == false);

    Dgp_State *dgp_state;
    dgp_state = ((jb2info*)m_jb2Info)->dgp_state;
	if (!dgp_state) return false;
    pIPWState = dgp_state->ipwstate;

	if(m_jb2Info) cleanupJb2Info((jb2info**)&m_jb2Info);

    // clip in with and height
    if (needClip)
    {
        Asd::UInt32 clipWidth;
		clipWidth = region.width() * scale.getX();
        Asd::UInt32 clipHeight; 
		clipHeight = region.height() * scale.getY();
        SILX_ERR_CHECK( ipcore::w_createPixr (pIPWState, clipWidth, clipHeight, MONOCHROME,
            ipcore::pixrRes, ipcore::pixrRes, framePix, &clipTempBuf));
		
        SILX_ERR_CHECK(ipcore::w_rasterop ( pIPWState, region.left() * scale.getX(), 
            region.top() * scale.getY(), 
            srcPixr, PIX_SRC, 0, 0, clipWidth, clipHeight,
            clipTempBuf));
		
        SILX_ERR_CHECK( ipcore::w_destroyPixr(pIPWState, &srcPixr));
        srcPixr = clipTempBuf;
    }
	
	// max downsample with this method is 8x ...  FIXME
    if (needDownScale)
    {
        Asd::UInt32* scaleTable;
        Asd::UInt32 resScaleX, resScaleY;
		resScaleX = 1;
		resScaleY = 1;
        Asd::UInt32 initialScale ;
		initialScale = MIN(scale.getX(), scale.getY());
        if (initialScale > 8) initialScale = 8;
		
        if (initialScale < scale.getX())
        {
            resScaleX = scale.getX() / initialScale;
        }
        if (initialScale < scale.getY())
        {
            resScaleY = scale.getY() / initialScale;
        }
		
        // IPCore convention
        initialScale *= 10;
        SILX_ERR_CHECK( ipcore::w_makeScaleToGray4bppTable (pIPWState, initialScale, &scaleTable));
        SILX_ERR_CHECK( ipcore::w_scaleToGray4bpp (pIPWState, srcPixr, initialScale,
            scaleTable, &tempPixr));
		SILX_ERR_CHECK( ipcore::w_destroyPixr(pIPWState, &srcPixr)); 
		
        if (resScaleY == 1 && resScaleX == 1) {
			SILX_ERR_CHECK( ipcore::w_depth4To8 (pIPWState, tempPixr, &srcPixr)); 
			SILX_ERR_CHECK( ipcore::w_destroyPixr(pIPWState, &tempPixr)); 
		}	else {
			// do a linear resize to fit into the ROI
			ttempPixr = NULL;
			SILX_ERR_CHECK( ipcore::w_depth4To8 (pIPWState, tempPixr, &ttempPixr));
			SILX_ERR_CHECK( ipcore::w_destroyPixr(pIPWState, &tempPixr)); 
			
			SILX_ERR_CHECK( ipcore::w_scaleLinearSize (pIPWState, ttempPixr, region.width(),
				region.height(), &srcPixr));
			SILX_ERR_CHECK( ipcore::w_destroyPixr(pIPWState, &ttempPixr)); 
		}
	}

	// copy the data into the output buffer
	ttempPixr = ibuf->pixr ();
	Asd::Int32 ws, hs, depth, bpls, bpld;  
	Asd::UInt8 *pImages, *pImaged;
	SILX_ERR_CHECK( ipcore::w_pixrGetDepth(pIPWState, srcPixr, &depth));
	SILX_ERR_CHECK( ipcore::w_pixrFramedWidth(pIPWState, srcPixr, &ws));
	SILX_ERR_CHECK( ipcore::w_pixrFramedHeight(pIPWState, srcPixr, &hs));
	SILX_ERR_CHECK( ipcore::w_pixrGetImageAll(pIPWState, srcPixr, &pImages));
	SILX_ERR_CHECK( ipcore::w_pixrGetImage(pIPWState, ttempPixr, &pImaged));
	SILX_ERR_CHECK( ipcore::w_pixrGetBpl(pIPWState, srcPixr, &bpls));
	SILX_ERR_CHECK( ipcore::w_pixrGetBpl(pIPWState, ttempPixr, &bpld));
	
	/* blit the inner part of the src, exclusive of cFrameBits,
	 * into the dest.  */
	SILX_ERR_CHECK(ipcore::i_removeFrame (pIPWState, pImages,
					 ws, hs, depth, bpls,
					 pImaged, bpld));

	// cleanup
	SILX_ERR_CHECK( ipcore::w_destroyPixr(pIPWState, &srcPixr));
	
	return true;
	
ERR_CLEANUP:
	if(tempPixr)    ipcore::w_destroyPixr(pIPWState, &tempPixr);
	if(clipTempBuf) ipcore::w_destroyPixr(pIPWState, &clipTempBuf);
	if(srcPixr)     ipcore::w_destroyPixr(pIPWState, &srcPixr);
	if(ttempPixr)   ipcore::w_destroyPixr(pIPWState, &ttempPixr);
	if(m_jb2Info)   cleanupJb2Info((jb2info**)&m_jb2Info);
	return false;
}
	
bool TfxJBig2Band::decodeSymDict(Jbig2Dict* pDict, Asd::Int32 dictId, Asd::Int32 dictSize){

    // validation
    if(!pDict) return false;
	jb2info* info = (jb2info *)pDict->getJb2Info();
    if(!info) return false;

    // Get the dictionary data
    DictBufList::iterator where;
    Asd::UInt8* dictData = NULL;
    where = (pDict->getDictBufList()).find(dictId);
    if( where != (pDict->getDictBufList()).end()){
         dictData = where->second;
    } else 
		return false;
	SILX_ERR_CHECK(processDictFromData(info, dictData, dictSize));
    
	return true;

ERR_CLEANUP:
	return false;
}

TfxJBig2Band::~TfxJBig2Band()
{
    if(m_jb2Info) cleanupJb2Info((jb2info**)&m_jb2Info);
}


