#include "StdAfx.h"
#include "SilxImgBuf.h" 
#include "TfxBand.h" 
#include "MrcFile.h"
#include "TfxLayer.h"
#include "MiscConst.h"

#include "AsdSilx.h"
namespace AsdSilx{
#include "jpeg6blib.h"
#include "jpeg_interface.h"
};

using namespace AsdSilx;


bool TfxJpegBand::render ( ImageAdapter * buf, const Asd::Rect & region, 
						  const Asd::Point32 & scale )
{
	
	// this implementation is for IPCore Image Buffers only
    SilxImgBuf* sbuf = dynamic_cast<SilxImgBuf*> (buf);
    if (!sbuf) return false;
    
    TfxLayer* pLayer ;
    MrcFile* pFile;
    
    // check scale
    Asd::Int32 ratioX = scale.getX();
    Asd::Int32 ratioY = ratioX;
    if(ratioX ==0) return (false);
    
    //Dgp Stuff
    Dgp_State *dgp_state = (Dgp_State *)(sbuf->getUserState());
    struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr err;
    DGP_JPEGImageDef image_def;
    dgp_dest_buf dst_buf;
    ColorSpace color_space = YCbCr;
    Dgp_UInt32 ret_val = 0;
	
    // check what we need to do
    Dgp_UInt32 status = cSuccessful;
    bool needUpScale = false;
	
    if (ratioX < -1 ||  ratioY < -1) 
    {
        needUpScale = true;
        ratioX = abs(ratioX);
        ratioY = abs(ratioY);
    }
	// Allocate the IJL JPEG_CORE_PROPERTIES structure.
    cinfo.err = jpeg_std_error ( &err);
    cinfo.mem = NULL;
    cinfo.marker = NULL;
    cinfo.inputctl = NULL;
    SILX_ERR_CHECK(DGP_JPEG_init_jdecompress(&cinfo, dgp_state));

    // we need the jpeg library switch from rgb to bgr for viewing
    pLayer = getLayer();
    if(pLayer) {
        color_space = pLayer->getColorSpace(); 
        pFile = pLayer->getFile();
        if(pFile->isBGR())
            cinfo.BGR = 1;
        else 
            cinfo.BGR = 0;
    }else{
        cinfo.BGR = 0;
        color_space = YCbCr;
    }
    // We will let the page object handle merging and scale up, 
    // We only render jpeg image in it's orignal resolution or less
    if(needUpScale){
        // dstW, dstH, dstX, dstY, indicate the required region in rendered resolution
        // since the "region" is in scaled page resolution, and we only render upto
        // band resolution, we have to do a convert
        image_def.imgCoOrds.dstW = region.width() / ratioX;
        image_def.imgCoOrds.dstH = region.height() / ratioY;
        image_def.imgCoOrds.dstX = region.left() / ratioX;  
        image_def.imgCoOrds.dstY = region.top() / ratioY;
		
        // srcOrigionX and Y represent the origin of this band in rendered res
        image_def.imgCoOrds.srcOriginY = 0;
        image_def.imgCoOrds.srcOriginX = 0;
        ratioX = 1;
        ratioY = 1;
    }else {
		
        // dstW, dstH, dstX, dstY, indicate the required region in rendered resolution
        image_def.imgCoOrds.dstW = region.width();
        image_def.imgCoOrds.dstX = region.left();  
        image_def.imgCoOrds.dstH = region.height();
        image_def.imgCoOrds.dstY = region.top();  
		
        // srcOrigionX and Y represent the origin of this band in rendered res
        image_def.imgCoOrds.srcOriginX = 0;
        image_def.imgCoOrds.srcOriginY = 0;
    }
		
    // the src and dst res is used to figure out scale.
    image_def.imgResolution.srcRes.xRes = DgpCoord_eqi(ratioX);
    image_def.imgResolution.srcRes.yRes = DgpCoord_eqi(ratioY);
    image_def.imgResolution.destRes.xRes = DgpCoord_eqi(1);
    image_def.imgResolution.destRes.yRes = DgpCoord_eqi(1);
	
    // Since the input buffer was already adjusted to the jpeg image location
    // so the jpeg will always start at 0
    image_def.imgCoOrds.dstMinX = image_def.imgCoOrds.dstX;    
    image_def.imgCoOrds.dstMinY = image_def.imgCoOrds.dstY;  
	
    //Buffer bpl and height. 
    image_def.imgCoOrds.bytesPerRow = sbuf->bpl();
    image_def.imgCoOrds.numRows = sbuf->height();;
	
	// Only use for figuring out rotated image coordinates
    image_def.imgCoOrds.nominal_page_width = 0xffff;
    image_def.imgCoOrds.nominal_page_height = 0xffff;
    image_def.imgCoOrds.nominal_strip_width = 0xffff;
    image_def.imgCoOrds.nominal_strip_height = 0xffff;
	
    image_def.imgSrc.u = NULL;
    image_def.imgSrc.buf = m_data;
	
    image_def.imgSrc.size = getMemoryUsage();
    image_def.imgOrientation = ORIENT_RD;
	
    switch (sbuf->depth()){
    case MONOCHROME:
        image_def.imgType = DGP_PACKED;
    case COLOR_32_BIT:
        image_def.imgType = DGP_CMYK;
        dst_buf.cmyk_buf.c = ((Dgp_UInt8 **)sbuf->data())[0];
        dst_buf.cmyk_buf.m = ((Dgp_UInt8 **)sbuf->data())[1];
        dst_buf.cmyk_buf.y = ((Dgp_UInt8 **)sbuf->data())[2];
        dst_buf.cmyk_buf.k = ((Dgp_UInt8 **)sbuf->data())[3];
        break;
    default:
        image_def.imgType = DGP_RGB;
        dst_buf.other_buf.buffer = (Dgp_Int8 *)sbuf->data();
        break;
        
    }
	
    SILX_ERR_CHECK(DGP_JPEG_check_jdecompress(&image_def,&ret_val,&cinfo,dgp_state)); 
    SILX_ERR_CHECK(ret_val);
	
    // We only decode this layer. We don't do merge here
    dst_buf.other_buf.trans_mask = NULL;
    dst_buf.other_buf.trans_bpl = 0;

    // try to set a reasonable color space
	if (cinfo.num_components == 1) // check for gray
        cinfo.jpeg_color_space = JCS_GRAYSCALE;
	else {
		switch(color_space){
		case RGB:
			cinfo.jpeg_color_space = JCS_RGB;
			break;
		case YCbCr:
			cinfo.jpeg_color_space = JCS_YCbCr;
			break;
		case CIELab:
			cinfo.jpeg_color_space = JCS_CIELAB;
			break;
		case ICCLab:
			cinfo.jpeg_color_space = JCS_ICCLAB;
			break;
		case ITULab:
			cinfo.jpeg_color_space = JCS_ITULAB;
            break;
		default:
	        cinfo.jpeg_color_space = JCS_YCbCr;		
            break;
		}
	}
    SILX_ERR_CHECK ( Dgp_JPEG_decompress_jpeg(&image_def,&dst_buf,&cinfo,dgp_state));
    SILX_ERR_CHECK ( Dgp_JPEG_destroy_jdecompress(&cinfo,dgp_state));
	
    return true;
	
ERR_CLEANUP:
	
    Dgp_JPEG_destroy_jdecompress(&cinfo,dgp_state);
    return false;
	
}
