#ifndef AsdMath_h_Included
#define AsdMath_h_Included



#include <math.h>

namespace AsdMath
{
    template <class T> inline T ident( const T & ) { return T(1); }

    template <class T> inline T Abs( const T & n )
    {
        return (n < Asd::zero(n))? -n : n;
    }

    template <class NUM> NUM sgn( const NUM & n )
    {
        return (n >= zero(n))? ident(n) : -ident(n);
    }

    inline int    Abs( const    int & n ) { return ::abs(n); }
    inline float  Abs( const  float & n ) { return  (float) fabs(n); }
    inline double Abs( const double & n ) { return  fabs(n); }

    /* gdc() uses Euclids method for finding the greatest common divisor of */
    /* two integers.  I must have integer semantics.  If both n and m */
    /* are 0, we return 0.  If one of m or n is 0, we return the other. */
    template <class INT> inline INT gcd( INT m, INT n )
    {
        if( m )
        {
            do
            {
                INT r = n % m;
                n = m;
                m = r;
            }
            while( m );
        }

        /* The above algorithm "works" regardless of the signs of n and m, */
        /* but could produce a negative answer.  We must flip the sign in */
        /* this case. */
        
        return Abs(n);
    }

    template <class INT> INT lcm( const INT & i, const INT & j )
    {
        if( i && j )
        {
            return Abs( (i*j) / gcd(i,j) );
        }
        else
        {
            return 0;
        }
    }


   inline double round( double v ) { return floor( v + 0.5 ); }

   template <class T> inline T MAX( const T& a, const T& b){
 	return ( a < b ) ? b : a;
   }
   template <class T> inline T MIN( const T& a, const T& b){
 	return ( a < b ) ? a : b;
   }
};


#endif
/*AsdMath_h_Included */
