#ifndef MiscConst_h_Included
#define MiscConst_h_Included

#include <assert.h>

/* ColorDepth is a set of constanst representing the color depth of an image. */
/* It is useful for situations where an object is some sort of bitmap but */
/* various color depths are allowed.  Please note that the values here */
/* are chosen carefully.  Except COLOR_DEPTH_ERROR, these values all */
/* correspond exactly to the possible values of biBitCount in a Windows */
/* device independent bitmap.  This is deliberate, and you are expected */
/* to program accordingly.  That also means that you connot willy-nilly */
/* change the values of the constants here. */
/* */
/* COLOR_DEPTH_ERROR is 0 for a reason, although the reason is different. */
/* 0 means that the number of bits-per-pixel is specified or is implied */
/* by the JPEG or PNG format in a Windows DIB.  That is why we make it */
/* an error here.  If you do not want to support compression, this */
/* makes the error implicit.  Even if you do, you might want a 24-bit */
/* jpeg to return COLOR_24_BIT. */

enum ColorDepth
{
    COLOR_DEPTH_ERROR  =  0,
    MONOCHROME         =  1,
    COLOR_4_BIT        =  4,
    COLOR_8_BIT        =  8,
    COLOR_24_BIT       = 24,
    COLOR_32_BIT       = 32
};

template <class INT> inline
ColorDepth int2ColorDepth( const INT & i )
{
	switch( i )
	{
		case  1:
		case  4:
		case  8: 
		case 24:
		case 32:

			return ColorDepth(i);

		default:
			
			assert( NULL && COLOR_DEPTH_ERROR );

			break;
	}

	return COLOR_DEPTH_ERROR;
}

enum CompressionMethod
{
	UnknownCompression   = 0,
	JPEG      = 1,
	JBIG2     = 2,
	MMR       = 3,
	MR        = 4,
	MH        = 5
};

enum ColorSpace
{
	Default     = 0,
    WhiteIsZero = 1,
    BlackIsZero = 2,
    RGB         = 3,
    CMYK        = 4,
    YCbCr       = 5,
    CIELab      = 6,
    ICCLab      = 7,
    ITULab      = 8,
    Gray        = 9
};

const int RGB_PIXELSIZE   = 3;
const int MASK_TRANSPARENT = 0xff;
const int MASK_OPAQUE = 0x00;

enum Orientation
{
    Orient_Error = 360,
    Orient_0   =   0,
    Orient_90  =  90,
    Orient_180 = 180,
    Orient_270 = 270
};

template <class INT> inline
Orientation Int2Orient( INT ii )
{
    ii %= 360;

    if( ii < 0 )
    {
        ii += 360;
    }

    assert( ii >=   0 );
    assert( ii <  360 );

    switch( ii )
    {
        case   0 : return Orient_0;
        case  90 : return Orient_90;
        case 180 : return Orient_180;
        case 270 : return Orient_270;

        default:  break;
    };

    return Orient_Error;
}

inline bool OrientIsValid( Orientation orient )
{
    switch( orient )
    {
        case Orient_0:
        case Orient_90:
        case Orient_180:
        case Orient_270: return true;

        default: break;
    };

    return false;
}

#endif /*#ifndef MiscConst_h_Included */
