/*
 *
 * $Header: /project/icons/cvsroot/digipaper/tools/libjbig2.help/stdhuff.c,v 1.3 2000/03/07 23:08:46 mkrocken Exp $
 *
 */

#include "util.h"
#include "hufftable.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/tools/libjbig2.help/stdhuff.c,v 1.3 2000/03/07 23:08:46 mkrocken Exp $");


/* Standard Huffman tables */

#define	NLINES(lines)	(sizeof(lines) / sizeof(lines[0]))

/* Table A - generic values */
static table_line stdhuffA_lines[] = {
    { 0x0,	    0,	1,	 4}, /* 0...15 as 0+4 */
    { 0x2,	   16,	2,	 8}, /* 16...271 as 10+8 */
    { 0x6,	  272,	3,	16}, /* 272...65807 as 110+16 */
    { 0x7,	65808,	3,	32}  /* 65808...infinity as 111+32 - upper */
    };
hufftable_spec stdhuffA = {
    DGP_FALSE, DGP_TRUE, DGP_FALSE,
    NLINES(stdhuffA_lines), stdhuffA_lines
    };

static table_line stdhuffDW1_lines[] = {
    { 0x0,	0,	1,	0 }, /* 0 as 0 */
    { 0x2,	1,	2,	0 }, /* 1 as 10 */
    { 0x6,	2,	3,	0 }, /* 2 as 110 */
    { 0xe,	3,	4,	3 }, /* 3...10 as 1110+3 */
    { 0x1e,	11,	5,	6 }, /* 11...74 as 11110+6 */
    { 0x3e,	75,	6,	32 }, /* 75...infinity as 111110+32 - upper */
    { 0x3f,	0,	6,	0 }, /* OOB as 111111 */
};
hufftable_spec stdhuffDW1 = {
    DGP_FALSE, DGP_TRUE, DGP_TRUE,
    NLINES(stdhuffDW1_lines), stdhuffDW1_lines
    };


static table_line stdhuffDW2_lines[] = {
    { 0xfe,	-256,	8,	8 }, /* -256...-1 as 11111110+8 */
    { 0x0,	0,	1,	0 }, /* 0 as 0 */
    { 0x2,	1,	2,	0 }, /* 1 as 10 */
    { 0x6,	2,	3,	0 }, /* 2 as 110 */
    { 0xe,	3,	4,	3 }, /* 3...10 as 1110+3 */
    { 0x1e,	11,	5,	6 }, /* 11...74 as 11110+6 */
    { 0xff,	-257,	8,	32 }, /* -infinity...-257 as 11111111+32 - lower */
    { 0x7e,	75,	7,	32 }, /* 75...infinity as 1111110+32 - upper */
    { 0x3e,	0,	6,	0 }, /* OOB as 111110 */
};
hufftable_spec stdhuffDW2 = {
    DGP_TRUE, DGP_TRUE, DGP_TRUE,
    NLINES(stdhuffDW2_lines), stdhuffDW2_lines
    };


static table_line stdhuffDH1_lines[] = {
    { 0x0,	1,	1,	0 }, /* 1 as 0 */
    { 0x2,	2,	2,	0 }, /* 2 as 10 */
    { 0x6,	3,	3,	0 }, /* 3 as 110 */
    { 0xe,	4,	4,	3 }, /* 4...11 as 1110+3 */
    { 0x1e,	12,	5,	6 }, /* 12...75 as 11110+6 */
    { 0x1f,	76,	5,	32 }, /* 76...infinity as 11111+32 - upper */
};
hufftable_spec stdhuffDH1 = {
    DGP_FALSE, DGP_TRUE, DGP_FALSE,
    NLINES(stdhuffDH1_lines), stdhuffDH1_lines
    };


static table_line stdhuffDH2_lines[] = {
    { 0x7e,	-255,	7,	8 }, /* -255...0 as 1111110+8 */
    { 0x0,	1,	1,	0 }, /* 1 as 0 */
    { 0x2,	2,	2,	0 }, /* 2 as 10 */
    { 0x6,	3,	3,	0 }, /* 3 as 110 */
    { 0xe,	4,	4,	3 }, /* 4...11 as 1110+3 */
    { 0x1e,	12,	5,	6 }, /* 12...75 as 11110+6 */
    { 0x7f,	-256,	7,	32 }, /* -infinity...-256 as 1111111+32 - lower */
    { 0x3e,	76,	6,	32 }, /* 76...infinity as 111110+32 - upper */
};
hufftable_spec stdhuffDH2 = {
    DGP_TRUE, DGP_TRUE, DGP_FALSE,
    NLINES(stdhuffDH2_lines), stdhuffDH2_lines
    };


static table_line stdhuffFS1_lines[] = {
    { 0x1c,	-2048,	5,	10 }, /* -2048...-1025 as 11100+10 */
    { 0x8,	-1024,	4,	9 }, /* -1024...-513 as 1000+9 */
    { 0x9,	-512,	4,	8 }, /* -512...-257 as 1001+8 */
    { 0xa,	-256,	4,	7 }, /* -256...-129 as 1010+7 */
    { 0x1d,	-128,	5,	6 }, /* -128...-65 as 11101+6 */
    { 0x1e,	-64,	5,	5 }, /* -64...-33 as 11110+5 */
    { 0xb,	-32,	4,	5 }, /* -32...-1 as 1011+5 */
    { 0x0,	0,	2,	7 }, /* 0...127 as 00+7 */
    { 0x2,	128,	3,	7 }, /* 128...255 as 010+7 */
    { 0x3,	256,	3,	8 }, /* 256...511 as 011+8 */
    { 0xc,	512,	4,	9 }, /* 512...1023 as 1100+9 */
    { 0xd,	1024,	4,	10 }, /* 1024...2047 as 1101+10 */
    { 0x3e,	-2049,	6,	32 }, /* -infinity...-2049 as 111110+32 - lower */
    { 0x3f,	2048,	6,	32 }, /* 2048...infinity as 111111+32 - upper */
};
hufftable_spec stdhuffFS1 = {
    DGP_TRUE, DGP_TRUE, DGP_FALSE,
    NLINES(stdhuffFS1_lines), stdhuffFS1_lines
    };


static table_line stdhuffFS2_lines[] = {
    { 0x8,	-1024,	4,	9 }, /* -1024...-513 as 1000+9 */
    { 0x0,	-512,	3,	8 }, /* -512...-257 as 000+8 */
    { 0x9,	-256,	4,	7 }, /* -256...-129 as 1001+7 */
    { 0x1a,	-128,	5,	6 }, /* -128...-65 as 11010+6 */
    { 0x1b,	-64,	5,	5 }, /* -64...-32 as 11011+5 */
    { 0xa,	-32,	4,	5 }, /* -32...-1 as 1010+5 */
    { 0xb,	0,	4,	5 }, /* 0...31 as 1011+5 */
    { 0x1c,	32,	5,	5 }, /* 32...63 as 11100+5 */
    { 0x1d,	64,	5,	6 }, /* 64...127 as 11101+6 */
    { 0xc,	128,	4,	7 }, /* 128...255 as 1100+7 */
    { 0x1,	256,	3,	8 }, /* 256...511 as 001+8 */
    { 0x2,	512,	3,	9 }, /* 512...1023 as 010+9 */
    { 0x3,	1024,	3,	10 }, /* 1024...2047 as 011+10 */
    { 0x1e,	-1025,	5,	32 }, /* -infinity...-1025 as 11110+32 - lower */
    { 0x1f,	2048,	5,	32 }, /* 2048...infinity as 11111+32 - upper */
};
hufftable_spec stdhuffFS2 = {
    DGP_TRUE, DGP_TRUE, DGP_FALSE,
    NLINES(stdhuffFS2_lines), stdhuffFS2_lines
    };


static table_line stdhuffDS1_lines[] = {
    { 0xfc,	-15,	8,	3 }, /* -15...-8 as 11111100+3 */
    { 0x1fc,	-7,	9,	1 }, /* -7...-6 as 111111100+1 */
    { 0xfd,	-5,	8,	1 }, /* -5...-4 as 11111101+1 */
    { 0x1fd,	-3,	9,	0 }, /* -3 as 111111101 */
    { 0x7c,	-2,	7,	0 }, /* -2 as 1111100 */
    { 0xa,	-1,	4,	0 }, /* -1 as 1010 */
    { 0x0,	0,	2,	1 }, /* 0...1 as 00+1 */
    { 0x1a,	2,	5,	0 }, /* 2 as 11010 */
    { 0x3a,	3,	6,	0 }, /* 3 as 111010 */
    { 0x4,	4,	3,	4 }, /* 4...19 as 100+4 */
    { 0x3b,	20,	6,	1 }, /* 20...21 as 111011+1 */
    { 0xb,	22,	4,	4 }, /* 22...37 as 1011+4 */
    { 0xc,	38,	4,	5 }, /* 38...69 as 1100+5 */
    { 0x1b,	70,	5,	6 }, /* 70...133 as 11011+6 */
    { 0x1c,	134,	5,	7 }, /* 134...261 as 11100+7 */
    { 0x3c,	262,	6,	7 }, /* 262...389 as 111100+7 */
    { 0x7d,	390,	7,	8 }, /* 390...645 as 1111101+8 */
    { 0x3d,	646,	6,	10 }, /* 646...1669 as 111101+10 */
    { 0x1fe,	-16,	9,	32 }, /* -infinity...-16 as 111111110+32 - lower */
    { 0x1ff,	1670,	9,	32 }, /* 1670...infinity as 111111111+32 - upper */
    { 0x1,	0,	2,	0 }, /* OOB as 01 */
};
hufftable_spec stdhuffDS1 = {
    DGP_TRUE, DGP_TRUE, DGP_TRUE,
    NLINES(stdhuffDS1_lines), stdhuffDS1_lines
    };


static table_line stdhuffDS2_lines[] = {
    { 0xfc,	-31,	8,	4 }, /* -31...-16 as 11111100+4 */
    { 0x1fc,	-15,	9,	2 }, /* -15...-12 as 111111100+2 */
    { 0xfd,	-11,	8,	2 }, /* -11...-8 as 11111101+2 */
    { 0x1fd,	-7,	9,	1 }, /* -7...-6 as 111111101+1 */
    { 0x7c,	-5,	7,	1 }, /* -5...-4 as 1111100+1 */
    { 0xa,	-3,	4,	1 }, /* -3...-2 as 1010+1 */
    { 0x2,	-1,	3,	1 }, /* -1...0 as 010+1 */
    { 0x3,	1,	3,	1 }, /* 1...2 as 011+1 */
    { 0x1a,	3,	5,	1 }, /* 3...4 as 11010+1 */
    { 0x3a,	5,	6,	1 }, /* 5...6 as 111010+1 */
    { 0x4,	7,	3,	5 }, /* 7...38 as 100+5 */
    { 0x3b,	39,	6,	2 }, /* 39...42 as 111011+2 */
    { 0xb,	43,	4,	5 }, /* 43...74 as 1011+5 */
    { 0xc,	75,	4,	6 }, /* 75...138 as 1100+6 */
    { 0x1b,	139,	5,	7 }, /* 139...266 as 11011+7 */
    { 0x1c,	267,	5,	8 }, /* 267...522 as 11100+8 */
    { 0x3c,	523,	6,	8 }, /* 523...778 as 111100+8 */
    { 0x7d,	779,	7,	9 }, /* 779...1290 as 1111101+9 */
    { 0x3d,	1291,	6,	11 }, /* 1291...3338 as 111101+11 */
    { 0x1fe,	-32,	9,	32 }, /* -infinity...-32 as 111111110+32 - lower */
    { 0x1ff,	3339,	9,	32 }, /* 3339...infinity as 111111111+32 - upper */
    { 0x0,	0,	2,	0 }, /* OOB as 00 */
};
hufftable_spec stdhuffDS2 = {
    DGP_TRUE, DGP_TRUE, DGP_TRUE,
    NLINES(stdhuffDS2_lines), stdhuffDS2_lines
    };


static table_line stdhuffDS3_lines[] = {
    { 0x7a,	-21,	7,	4 }, /* -21...-6 as 1111010+4 */
    { 0xfc,	-5,	8,	0 }, /* -5 as 11111100 */
    { 0x7b,	-4,	7,	0 }, /* -4 as 1111011 */
    { 0x18,	-3,	5,	0 }, /* -3 as 11000 */
    { 0x0,	-2,	2,	2 }, /* -2...1 as 00+2 */
    { 0x19,	2,	5,	0 }, /* 2 as 11001 */
    { 0x36,	3,	6,	0 }, /* 3 as 110110 */
    { 0x7c,	4,	7,	0 }, /* 4 as 1111100 */
    { 0xfd,	5,	8,	0 }, /* 5 as 11111101 */
    { 0x1,	6,	2,	6 }, /* 6...69 as 01+6 */
    { 0x1a,	70,	5,	5 }, /* 70...101 as 11010+5 */
    { 0x37,	102,	6,	5 }, /* 102...133 as 110111+5 */
    { 0x38,	134,	6,	6 }, /* 134...197 as 111000+6 */
    { 0x39,	198,	6,	7 }, /* 198...325 as 111001+7 */
    { 0x3a,	326,	6,	8 }, /* 326...581 as 111010+8 */
    { 0x3b,	582,	6,	9 }, /* 582...1093 as 111011+9 */
    { 0x3c,	1094,	6,	10 }, /* 1094...2117 as 111100+10 */
    { 0x7d,	2118,	7,	11 }, /* 2118...4165 as 1111101+11 */
    { 0xfe,	-22,	8,	32 }, /* -infinity...-22 as 11111110+32 - lower */
    { 0xff,	4166,	8,	32 }, /* 4166...infinity as 11111111+32 - upper */
    { 0x2,	0,	2,	0 }, /* OOB as 10 */
};
hufftable_spec stdhuffDS3 = {
    DGP_TRUE, DGP_TRUE, DGP_TRUE,
    NLINES(stdhuffDS3_lines), stdhuffDS3_lines
    };


static table_line stdhuffDT1_lines[] = {
    { 0x0,	1,	1,	0 }, /* 1 as 0 */
    { 0x2,	2,	2,	1 }, /* 2...3 as 10+1 */
    { 0xc,	4,	4,	0 }, /* 4 as 1100 */
    { 0xd,	5,	4,	1 }, /* 5...6 as 1101+1 */
    { 0x1c,	7,	5,	1 }, /* 7...8 as 11100+1 */
    { 0x1d,	9,	5,	2 }, /* 9...12 as 11101+2 */
    { 0x3c,	13,	6,	2 }, /* 13...16 as 111100+2 */
    { 0x7a,	17,	7,	2 }, /* 17...20 as 1111010+2 */
    { 0x7b,	21,	7,	3 }, /* 21...28 as 1111011+3 */
    { 0x7c,	29,	7,	4 }, /* 29...44 as 1111100+4 */
    { 0x7d,	45,	7,	5 }, /* 45...76 as 1111101+5 */
    { 0x7e,	77,	7,	6 }, /* 77...140 as 1111110+6 */
    { 0x7f,	141,	7,	32 }, /* 141...infinity as 1111111+32 - upper */
};
hufftable_spec stdhuffDT1 = {
    DGP_FALSE, DGP_TRUE, DGP_FALSE,
    NLINES(stdhuffDT1_lines), stdhuffDT1_lines
    };


static table_line stdhuffDT2_lines[] = {
    { 0x0,	1,	1,	0 }, /* 1 as 0 */
    { 0x2,	2,	2,	0 }, /* 2 as 10 */
    { 0x6,	3,	3,	1 }, /* 3...4 as 110+1 */
    { 0x1c,	5,	5,	0 }, /* 5 as 11100 */
    { 0x1d,	6,	5,	1 }, /* 6...7 as 11101+1 */
    { 0x3c,	8,	6,	1 }, /* 8...9 as 111100+1 */
    { 0x7a,	10,	7,	0 }, /* 10 as 1111010 */
    { 0x7b,	11,	7,	1 }, /* 11...12 as 1111011+1 */
    { 0x7c,	13,	7,	2 }, /* 13...16 as 1111100+2 */
    { 0x7d,	17,	7,	3 }, /* 17...24 as 1111101+3 */
    { 0x7e,	25,	7,	4 }, /* 25...40 as 1111110+4 */
    { 0xfe,	41,	8,	5 }, /* 41...72 as 11111110+5 */
    { 0xff,	73,	8,	32 }, /* 73...infinity as 11111111+32 - upper */
};
hufftable_spec stdhuffDT2 = {
    DGP_FALSE, DGP_TRUE, DGP_FALSE,
    NLINES(stdhuffDT2_lines), stdhuffDT2_lines
    };


static table_line stdhuffDT3_lines[] = {
    { 0x0,	1,	1,	0 }, /* 1 as 0 */
    { 0x4,	2,	3,	0 }, /* 2 as 100 */
    { 0xc,	3,	4,	0 }, /* 3 as 1100 */
    { 0x1c,	4,	5,	0 }, /* 4 as 11100 */
    { 0xd,	5,	4,	1 }, /* 5...6 as 1101+1 */
    { 0x5,	7,	3,	3 }, /* 7...14 as 101+3 */
    { 0x3a,	15,	6,	1 }, /* 15...16 as 111010+1 */
    { 0x3b,	17,	6,	2 }, /* 17...20 as 111011+2 */
    { 0x3c,	21,	6,	3 }, /* 21...28 as 111100+3 */
    { 0x3d,	29,	6,	4 }, /* 29...44 as 111101+4 */
    { 0x3e,	45,	6,	5 }, /* 45...76 as 111110+5 */
    { 0x7e,	77,	7,	6 }, /* 77...140 as 1111110+6 */
    { 0x7f,	141,	7,	32 }, /* 141...infinity as 1111111+32 - upper */
};
hufftable_spec stdhuffDT3 = {
    DGP_FALSE, DGP_TRUE, DGP_FALSE,
    NLINES(stdhuffDT3_lines), stdhuffDT3_lines
    };


static table_line stdhuffDT3_neg_lines[] = {
    { 0x0,	1,	2,	0 }, /* 1 as 00 */
    { 0x4,	2,	3,	0 }, /* 2 as 100 */
    { 0xc,	3,	4,	0 }, /* 3 as 1100 */
    { 0x1c,	4,	5,	0 }, /* 4 as 11100 */
    { 0xd,	5,	4,	1 }, /* 5...6 as 1101+1 */
    { 0x5,	7,	3,	3 }, /* 7...14 as 101+3 */
    { 0x3a,	15,	6,	1 }, /* 15...16 as 111010+1 */
    { 0x3b,	17,	6,	2 }, /* 17...20 as 111011+2 */
    { 0x3c,	21,	6,	3 }, /* 21...28 as 111100+3 */
    { 0x3d,	29,	6,	4 }, /* 29...44 as 111101+4 */
    { 0x3e,	45,	6,	5 }, /* 45...76 as 111110+5 */
    { 0x7e,	77,	7,	6 }, /* 77...140 as 1111110+6 */
    { 0x1,	0,	2,	32 }, /* -infinity...0 as 01+32 - lower */
    { 0x7f,	141,	7,	32 }, /* 141...infinity as 1111111+32 - upper */
};
hufftable_spec stdhuffDT3_neg = {
    DGP_TRUE, DGP_TRUE, DGP_FALSE,
    NLINES(stdhuffDT3_neg_lines), stdhuffDT3_neg_lines
    };


static table_line stdhuffRD1_lines[] = {
    { 0x4,	-2,	3,	0 }, /* -2 as 100 */
    { 0x5,	-1,	3,	0 }, /* -1 as 101 */
    { 0x0,	0,	1,	0 }, /* 0 as 0 */
    { 0x6,	1,	3,	0 }, /* 1 as 110 */
    { 0x7,	2,	3,	0 }, /* 2 as 111 */
};
hufftable_spec stdhuffRD1 = {
    DGP_FALSE, DGP_FALSE, DGP_FALSE,
    NLINES(stdhuffRD1_lines), stdhuffRD1_lines
    };


static table_line stdhuffRD2_lines[] = {
    { 0x7c,	-24,	7,	4 }, /* -24...-9 as 1111100+4 */
    { 0x3c,	-8,	6,	2 }, /* -8...-5 as 111100+2 */
    { 0x1c,	-4,	5,	1 }, /* -4...-3 as 11100+1 */
    { 0xc,	-2,	4,	0 }, /* -2 as 1100 */
    { 0x4,	-1,	3,	0 }, /* -1 as 100 */
    { 0x0,	0,	1,	0 }, /* 0 as 0 */
    { 0x5,	1,	3,	0 }, /* 1 as 101 */
    { 0xd,	2,	4,	0 }, /* 2 as 1101 */
    { 0x1d,	3,	5,	1 }, /* 3...4 as 11101+1 */
    { 0x3d,	5,	6,	2 }, /* 5...8 as 111101+2 */
    { 0x7d,	9,	7,	4 }, /* 9...24 as 1111101+4 */
    { 0x7e,	-25,	7,	32 }, /* -infinity...-25 as 1111110+32 - lower */
    { 0x7f,	25,	7,	32 }, /* 25...infinity as 1111111+32 - upper */
};
hufftable_spec stdhuffRD2 = {
    DGP_TRUE, DGP_TRUE, DGP_FALSE,
    NLINES(stdhuffRD2_lines), stdhuffRD2_lines
    };
