/*
 *
 * $Header: /project/icons/cvsroot/digipaper/tools/libjbig2.util/tuncomp-jbig.c,v 1.4 2001/01/30 18:16:50 mkrocken Exp $
 *
 */

/*
 * This is going to be a JBIG2 decompressor: read segments,
 * produce pages, write them out.
 */

#define DEF_FILE_ERROR_OUTPUT

#include "jb2file.h"
#include "segment.h"
#include "symbol.h"
#include "datasource.h"
#include "datasink.h"
#include "segment_util.h"
#include <stdio.h>
// #include "undecl-stdio.h"
#include <stdlib.h>
#include "list.h"
// #include "registry.h"
#include "blit.h"
#include "dgp_err_printers.h"

/* for extensions */
#include "decextseg.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/tools/libjbig2.util/tuncomp-jbig.c,v 1.4 2001/01/30 18:16:50 mkrocken Exp $");

static char usage[] = "usage: %s [-o outformat] [-v] [-noout] filename.jbig2\n";

/* How large is our input buffer? */
#define	BUFSIZE	4096

/* The magic JBIG2 file header: M-C-W JB2 CR LF C-Z LF */
unsigned char jb2header[] = { 0x97, 0x4a, 0x42, 0x32, 0x0d, 0x0a, 0x1a, 0x0a };

/* Segment type names */
unsigned char *typenames[] = {
    "Symbol dictionary", "INVALID", "INVALID", "INVALID",
    "Intermediate symbol region", "INVALID", "Immediate symbol region", "Immediate lossless symbol region",
    "INVALID", "INVALID", "INVALID", "INVALID",
    "INVALID", "INVALID", "INVALID", "INVALID",

    "Halftone dictionary", "INVALID", "INVALID", "INVALID",
    "Intermediate halftone region", "INVALID", "Immediate halftone region", "Immediate lossless halftone region",
    "INVALID", "INVALID", "INVALID", "INVALID",
    "INVALID", "INVALID", "INVALID", "INVALID",

    "INVALID", "INVALID", "INVALID", "INVALID",
    "Intermediate generic region", "INVALID", "Immediate generic region", "Immediate lossless generic region",
    "Intermediate generic refinement region", "INVALID", "Immediate generic refinement region", "Immediate lossless generic refinement region",
    "INVALID", "INVALID", "INVALID", "INVALID",

    "Page information", "End of page", "End of stripe", "End of file",
    "Profiles", "Table", "INVALID", "INVALID",
    "INVALID", "INVALID", "INVALID", "INVALID",
    "INVALID", "INVALID", "Extension", "INVALID"
    };

/* Info we maintain about the file */
typedef struct {
    jb2File f;
    DgpAdt_List activesegs;
    Dgp_Bool sequential;
    DgpImage_Packed curpage;
    jb2PageInfoSegment pageinfo;
    Dgp_Bool dunnosize;
    DgpAdt_List regions;
    int laststripe;
    int lastregion;
    int pagenum;
    int npages;
    FILE *input;
    char *fname;
    char *format;
    DataSource data;
    struct _DataSource _data;
    char buf[BUFSIZE];
    size_t bufsize;
    size_t bytesread;
    jb2Segment **segs;
    int nsegs;
    Dgp_Bool verbose;
    Dgp_Bool noout;
    Dgp_State *dgp_state;
    } jb2info;

static ErrCode processSimpleSegment(jb2Segment *segment, jb2info *info);
static ErrCode processSymRegSegment(jb2Segment *segment, jb2info *info);
static ErrCode processGenRegSegment(jb2Segment *segment, jb2info *info);
static ErrCode processHTRegSegment(jb2Segment *segment, jb2info *info);
static ErrCode processGenRefRegSegment(jb2Segment *segment, jb2info *info);
static ErrCode processPageInfoSegment(jb2Segment *segment, jb2info *info);
static ErrCode processEOPageSegment(jb2Segment *segmente, jb2info *info);
static ErrCode processEOStripeSegment(jb2Segment *segment, jb2info *info);
static ErrCode processExtensionSegment(jb2Segment *segment, jb2info *info);

/* Most of the segments can be handled simply */
#define	processSymDictSegment	processSimpleSegment
#define	processEOFileSegment	processSimpleSegment
#define	processHuffTableSegment	processSimpleSegment
#define	processHTDictSegment	processSimpleSegment
#define	processProfilesSegment	processSimpleSegment

static ErrCode (*processors [64])(jb2Segment *, jb2info *) = {
    processSymDictSegment, NULL, NULL, NULL,
    processSymRegSegment, NULL, processSymRegSegment, processSymRegSegment,
    NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL,

    processHTDictSegment, NULL, NULL, NULL,
    processHTRegSegment, NULL, processHTRegSegment, processHTRegSegment,
    NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL,

    NULL, NULL, NULL, NULL,
    processGenRegSegment, NULL, processGenRegSegment, processGenRegSegment,
    processGenRefRegSegment, NULL, processGenRefRegSegment, processGenRefRegSegment,
    NULL, NULL, NULL, NULL,

    processPageInfoSegment, processEOPageSegment, processEOStripeSegment, processEOFileSegment,
    processProfilesSegment, processHuffTableSegment, NULL, NULL,
    NULL, NULL, NULL, NULL,
    NULL, NULL, processExtensionSegment, NULL
    };
    
static void dumpheader(jb2Segment *curseg);

/* Decode a simple (no action needed after decoding) segment */
static ErrCode
processSimpleSegment(jb2Segment *segment, jb2info *info)
{
  DGP_PROLOGUE("processSimpleSegment()", info->DGP_PERRSTATE);

  Dgp_Allocators *a = &(info->DGP_ALLOCATORS);
    jb2Segment **refsegs = (jb2Segment **)NULL;
    int oldcount;

    DGP_REPORTF1((info->verbose), "decoding segment %d\n", segment->segnum);

    /* Marshal the referenced segments */
    DGP_ERR_PASS_CLEANUPF0(jb2MarshalSegments(info->f, segment,
					      info->activesegs, &refsegs),
			   "can't marshal segments\n");

    /* How many bytes consumed so far? */
    oldcount = info->bytesread + info->_data.bufpos;

    /* Call the decoding routine */
    DGP_ERR_PASS_CLEANUPF0(jb2DecodeSegment(info->f, info->data, segment,
					    refsegs),
			   "can't decode segment\n");

    /* Check bytes */
    DGP_ERR_GEN_CLEANUPF2((info->bytesread + info->_data.bufpos - oldcount !=
			   segment->datalen), cErrDecodeFailure,
	"segment claimed %d bytes but consumed %d\n",
		(int)segment->datalen,
		(int)(info->bytesread + info->_data.bufpos - oldcount));

    /* Don't need refsegs any more */
    if (refsegs != (jb2Segment **)NULL) {
	(a->free)((void *)refsegs);
	}

    /* That's all we have to do. */
    NORMAL_RETURN;

  ERR_CLEANUP:
    if (refsegs != (jb2Segment **)NULL) {
	(a->free)((void *)refsegs);
	}
    ERR_RETURN;
    }

static ErrCode
renderRegion(jb2Segment *segment, jb2info *info) 
{
  DGP_PROLOGUE("renderRegion()", info->DGP_PERRSTATE);

  Dgp_Allocators *a = &(info->DGP_ALLOCATORS);
    jb2RegSegment *clone = (jb2RegSegment *)NULL;
    DgpImage_Packed cloneim = (DgpImage_Packed)NULL;

    DGP_ASSERT(isRegionSegtype(segment->segtype));

    /* Check the page flags - make sure this is valid */
    DGP_ERR_GEN_CLEANUPF0(((!info->pageinfo.may_have_auxbuf) &&
			 (!isImmediateRegion(segment->segtype))),
			cErrDecodeFailure,
			"intermediate region on a page that can't have aux buffers\n");

    DGP_ERR_GEN_CLEANUPF0(((!info->pageinfo.may_refine) &&
			   (isRefinementRegion(segment->segtype))),
			  cErrDecodeFailure,
	"refinement region on a page that can't have refinement\n");

    DGP_ERR_GEN_CLEANUPF0(((!info->pageinfo.may_override_combop) &&
	(segment->reginfo.combop != info->pageinfo.combop)),
			  cErrDecodeFailure,
	"region combination operator doesn't match page combination operator\non a page where they must match\n");

    /* Check the page striping */
    if (info->pageinfo.striped) {
      DGP_ERR_GEN_CLEANUPF0(((int)segment->reginfo.regypos < info->laststripe),
	cErrDecodeFailure,
	    "region goes into previous stripe\n");

    DGP_ERR_GEN_CLEANUPF0(((int)(segment->reginfo.regypos +
				 segment->reginfo.regheight - 1) >
			   info->laststripe + (int)info->pageinfo.max_stripe),
			  cErrDecodeFailure,
			  "region makes stripe too high\n");

	/* Remember the last row touched */
	info->lastregion = MAX(info->lastregion,
			       (int)(segment->reginfo.regypos +
				     segment->reginfo.regheight - 1));
	}

    if (isImmediateRegion(segment->segtype)) {
	/* Render it into the page */

	if (info->dunnosize) {
	  DgpAdt_ListNode retnode;

	    /* Can't render into the page because we don't know its height */
	  DGP_ERR_PASS_CLEANUP(jb2RenderRegion(info->f, segment));

	    /*
	     * Now clone and stash the rendered region so we're not at
	     * the mercy of the segment GC logic
	     */
	    clone = (jb2RegSegment *)(a->malloc)(sizeof(*clone));
	    DGP_CHECK_ALLOC_CLEANUP(clone);
	    *clone = segment->reginfo;

	    DGP_ERR_PASS_CLEANUP(DgpImage_dup(segment->reginfo.im, DGP_FALSE, &cloneim));
	
	    clone->im = cloneim;

	    DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(info->regions, (void *)clone, &retnode));
	    }
	else {
	    /* Just render it into the page */
	  DGP_ERR_PASS_CLEANUP(jb2RenderRegionIntoPage(info->f, segment,
						       info->curpage));
	    }
	}
    else {
	/* Just render it. */

      DGP_ERR_PASS_CLEANUP(jb2RenderRegion(info->f, segment));
	}
    
    NORMAL_RETURN;

  ERR_CLEANUP:
    if (clone != (jb2RegSegment *)NULL) {
	(a->free)((void *)clone);
	}

    if (cloneim != (DgpImage_Packed)NULL) {
	DgpImage_free(cloneim);
	}

    ERR_RETURN;	
    }

/*
 * Process a symbol region segment: decode it, mess around with the
 * appropriate buffers.
 */
static ErrCode
processSymRegSegment(jb2Segment *segment, jb2info *info)
{
  DGP_PROLOGUE("processSymRegSegment()", info->DGP_PERRSTATE);

    /* Step one: just do the segment decode */
    DGP_ERR_PASS(processSimpleSegment(segment, info));

    DGP_ERR_PASS(renderRegion(segment, info));

    NORMAL_RETURN;
    }

/*
 * Process a generic region segment: decode it, mess around with the
 * appropriate buffers.
 */
static ErrCode
processGenRegSegment(jb2Segment *segment, jb2info *info)
{
  DGP_PROLOGUE("processGenRegSegment()", info->DGP_PERRSTATE);

    /* Step one: just do the segment decode */
  DGP_ERR_PASS(processSimpleSegment(segment, info));

  DGP_ERR_PASS(renderRegion(segment, info));

    NORMAL_RETURN;
    }

/*
 * Process a halftone region segment
 */
static ErrCode
processHTRegSegment(jb2Segment *segment, jb2info *info)
{
  DGP_PROLOGUE("processHTRegSegment()", info->DGP_PERRSTATE);

    /* Step one: just do the segment decode */
  DGP_ERR_PASS(processSimpleSegment(segment, info));

  DGP_ERR_PASS(renderRegion(segment, info));

    NORMAL_RETURN;
    }

/*
 * Process a generic region segment: decode it, mess around with the
 * appropriate buffers.
 */
static ErrCode
processGenRefRegSegment(jb2Segment *segment, jb2info *info)
{
  DGP_PROLOGUE("processGenRefRegSegment()", info->DGP_PERRSTATE);

    /* Step one: just do the segment decode */
  DGP_ERR_PASS(processSimpleSegment(segment, info));

  DGP_ERR_GENF0((info->pageinfo.striped), cErrInvalidParm,
		"unimplemented combo of refinement and striping\n");

    DGP_ERR_PASS(renderRegion(segment, info));

    NORMAL_RETURN;
    }

/*
 * Process a page info segment: decode it, create the page image
 */
static ErrCode
processPageInfoSegment(jb2Segment *segment, jb2info *info)
{
  DGP_PROLOGUE("processPageInfoSegment()", info->DGP_PERRSTATE);

    /* Step one: just do the segment decode */
  DGP_ERR_PASS_CLEANUP(processSimpleSegment(segment, info));

    /* Be careful - may not know page size yet */
    if ((segment->data.page.striped) &&
	(segment->data.page.height == 0xffffffff)) {
	/* Page height unknown */
	info->curpage = (DgpImage_Packed)NULL;
	info->dunnosize = DGP_TRUE;
	/* Need a list of regions */
	DgpAdt_liNew(&info->regions, info->dgp_state);
	DGP_CHECK_ALLOC_CLEANUP(info->regions);
	}
    else {
	info->dunnosize = DGP_FALSE;
	DgpImage_new(info->dgp_state, DGPIMAGE_PACKED, segment->data.page.width,
			       segment->data.page.height, &info->curpage);
	DGP_CHECK_ALLOC_CLEANUP(info->curpage);
	info->regions = DGP_NULLLIST;

	if (segment->data.page.defpixel != 0) {
	    DgpBlat_blitFillPacked(info->curpage);
	    }
	}

    info->pagenum = segment->pagenum;
    info->pageinfo = segment->data.page;
    info->laststripe = -1;
    info->lastregion = -1;

    NORMAL_RETURN;

  ERR_CLEANUP:
    ERR_RETURN;
    }

/*
 * Process an end-of-page segment: decode it, write out the page image,
 * free the image
 */
static ErrCode
processEOPageSegment(jb2Segment *segment, jb2info *info)
{
  DGP_PROLOGUE("processEOPageSegment()", info->DGP_PERRSTATE);

  Dgp_Allocators *a = &(info->DGP_ALLOCATORS);
    char fname[1000];

    /* Step one: just do the segment decode */
    DGP_ERR_PASS_CLEANUP(processSimpleSegment(segment, info));

    if (info->dunnosize) {
	DgpAdt_ListNode ln;

	/*
	 * We didn't know the page size.  The last EOStripe segment's
	 * end row is the page end row.
	 */
	DgpImage_new(info->dgp_state, DGPIMAGE_PACKED,
					   info->pageinfo.width,
					   info->laststripe + 1, &info->curpage);

	if (info->pageinfo.defpixel != 0) {
	    DgpBlat_blitFillPacked(info->curpage);
	    }

	/* Now we want to draw the list of segments into the page */
	DGPADT_FOREACH (ln, info->regions) {
	    jb2RegSegment *curreg = (jb2RegSegment *)DGPADT_LIGET(ln);

	    DgpBlat_blitPackedOp(info->curpage, curreg->im,
			 curreg->regxpos, curreg->regypos,
			 jb2Minterms[curreg->combop]);
	    DgpImage_free(curreg->im);
	    (a->free)((void *)curreg);
	    }
	DgpImage_free(info->regions);
	info->regions = DGP_NULLLIST;
	}	    

    /* Save and free info->curpage */
    if (!info->noout) {
	if (info->format != (char *)NULL) {
	    sprintf(fname, info->format, info->pagenum - 1);
	    }
	else {
	    sprintf(fname, "%s.p%04d.pbm", info->fname, info->pagenum - 1);
	    }

	DGP_ERR_PASS_CLEANUP(DgpImage_save(info->curpage, fname));
	}
    DgpImage_free(info->curpage);
    info->curpage = (DgpImage_Packed)NULL;

    NORMAL_RETURN;

  ERR_CLEANUP:
    ERR_RETURN;
    }

static ErrCode
processEOStripeSegment(jb2Segment *segment, jb2info *info)
{
  DGP_PROLOGUE("processEOStripeSegment()", info->DGP_PERRSTATE);

    /* Should we even have one of these? */
  DGP_ERR_GEN_CLEANUPF0((!info->pageinfo.striped), cErrDecodeFailure,
	"non-striped page has EOStripe segment\n");

    /* Decode the stripe */
    DGP_ERR_PASS_CLEANUP(processSimpleSegment(segment, info));

    DGP_ERR_GEN_CLEANUPF0(((int)segment->data.eostripe.last_row <
			info->laststripe), cErrDecodeFailure,
      "stripe goes backwards\n");

    DGP_ERR_GEN_CLEANUPF0(((int)segment->data.eostripe.last_row -
			 info->laststripe > info->pageinfo.max_stripe),
			cErrDecodeFailure, "stripe size is too large\n");

    DGP_ERR_GEN_CLEANUPF0(((int)segment->data.eostripe.last_row <
			  info->lastregion),  cErrDecodeFailure,
			 "stripe doesn't include all regions within stripe\n");

    info->laststripe = segment->data.eostripe.last_row;
    NORMAL_RETURN;

  ERR_CLEANUP:
    ERR_RETURN;
    }

static ErrCode
processExtensionSegment(jb2Segment *segment, jb2info *info)
{
  DGP_PROLOGUE("processExtensionSegment()", info->DGP_PERRSTATE);

    /* Step one: just do the segment decode */
  DGP_ERR_PASS_CLEANUP(processSimpleSegment(segment, info));

    /* Handle anything we recognise */
    switch(segment->data.extension.type) {
      case EXT_ASCIICOMMENT: {
	if (info->verbose) {
	    int i;
	    jb2Extension_ASCIIComment *comment =
		(jb2Extension_ASCIIComment *)segment->data.extension.data;

	    printf("ASCII comment in segment %d\n", segment->segnum);
	    for (i = 0; i < comment->npairs; i++) {
		printf("%s\t|%s\n", comment->names[i], comment->values[i]);
		}
	    }
	  }
	}

    NORMAL_RETURN;

  ERR_CLEANUP:
    ERR_RETURN;
    }

/* Callback for the DataSource */
static ErrCode
moredata(DataSource f)
{
  jb2info *info = (jb2info *)f->userdata;
  DGP_PROLOGUE("moredata()", info->DGP_PERRSTATE);

    size_t nbytes;

    /* Note that we've consumed the previous buffer */
    info->bytesread += info->bufsize;

    /* Read BUFSIZE bytes */
    nbytes = fread(info->buf, 1, BUFSIZE, info->input);

    /* no more data */
    DGP_ERR_GEN((nbytes == 0), cErrIO_ReadError);

    info->bufsize = nbytes;
    info->_data.buf = info->buf;
    info->_data.bufsize = nbytes;
    info->_data.bufpos = 0;

    NORMAL_RETURN;
    }
    

/* Set up info's DataSource, given the open file */
static void
setupds(jb2info *info)
{
    info->bytesread = 0;
    info->bufsize = 0;
    info->_data.buf = info->buf;
    info->_data.bufsize = 0;
    info->_data.bufpos = 0;
    info->_data.userdata = (void *)info;
    info->_data.more_data = moredata;
    info->_data.dgp_state = info->dgp_state;
    info->data = &info->_data;
    }

/* Allocate *newseg; read a header into it. */
static ErrCode
readheader(jb2info *info, jb2Segment **newseg)
{
  DGP_PROLOGUE("readheader()", info->DGP_PERRSTATE);

  Dgp_Allocators *a = &(info->DGP_ALLOCATORS);
    
  /* Allocate a new segment */
    *newseg = (jb2Segment *)(a->malloc)(sizeof(**newseg));
    DGP_CHECK_ALLOC_CLEANUP(*newseg);

    /* Read in a header */
    DGP_ERR_PASS_CLEANUP(jb2DecodeSegmentHeader(info->f, info->data, *newseg));

    /* Paranoia check */
    if ((*newseg)->pagenum > info->npages) {
	fprintf(stderr, "segment has pagenum %d > %d\n", (*newseg)->pagenum,
		info->npages);
	}

    if (info->verbose) {
	dumpheader(*newseg);
	}

    NORMAL_RETURN;

  ERR_CLEANUP:
    ERR_RETURN;
    }

/*
 * Read in all the segment headers, until you see an EOF segment.
 * Stick the result in info->segs and info->nsegs.
 */
static ErrCode
readheaders(jb2info *info)
{
  DGP_PROLOGUE("readheaders()", info->DGP_PERRSTATE);

  Dgp_Allocators *a = &(info->DGP_ALLOCATORS);
    DgpAdt_List segs = DGP_NULLLIST;
    jb2Segment *newseg = (jb2Segment *)NULL;
    int nsegs;
    int i;
    DgpAdt_ListNode ln, retnode;

    DgpAdt_liNew(&segs, info->dgp_state);
    DGP_CHECK_ALLOC_CLEANUP(segs);

    for (;;) {
      DGP_ERR_PASS_CLEANUP(readheader(info, &newseg));

	/* Add it to the list */
      DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(segs, (void *)newseg, &retnode));

	/* If it's an EOF segment, exit */
	if (newseg->segtype == JB2SEGTYPE_EOFILE) {
	    break;
	    }
	newseg = (jb2Segment *)NULL;
	}
    newseg = (jb2Segment *)NULL;

    /* Create the info->segs array */
    info->nsegs = nsegs = DGPADT_LILEN(segs);
    info->segs = (jb2Segment **)(a->malloc)(nsegs * sizeof(*info->segs));
    DGP_CHECK_ALLOC_CLEANUP(info->segs);

    /* and read the list into it */
    i = 0;
    DGPADT_FOREACH (ln, segs) {
	info->segs[i++] = (jb2Segment *)DGPADT_LIGET(ln);
	}
    
    NORMAL_RETURN;

  ERR_CLEANUP:

    if (segs != DGP_NULLLIST) {
	DGPADT_FOREACH (ln, segs) {
	    (a->free)(DGPADT_LIGET(ln));
	    }
	DgpAdt_liFree(segs);
	}

    if (newseg != (jb2Segment *)NULL) {
	(a->free)((void *)newseg);
	}

    ERR_RETURN;
    }

/* Print out all the segment header information */
static void
dumpheader(jb2Segment *curseg)
{
    int j;

    fprintf(stderr,
	    "segnum %2d:\tdata length %4d\tpagenum %2d\tnrefsegs %2d (",
	    curseg->segnum, (int)curseg->datalen, curseg->pagenum,
	    curseg->nrefsegs);

    for (j = 0; j < curseg->nrefsegs; j++) {
	fprintf(stderr, "%d%c", curseg->refsegs[j],
		curseg->refretain[j] ? 'T' : 'F');
	if (j != curseg->nrefsegs - 1) {
	    fprintf(stderr, " ");
	    }
	}

    fprintf(stderr, ")\tsegtype %2d (%s)\t%c%c\n",
	    curseg->segtype, typenames[curseg->segtype],
	    curseg->retain ? 'T' : 'F',
	    curseg->defer_nonretain ? 'T' : 'F');
    }

/* Process a single segment */
static ErrCode
processSegment(jb2Segment *segment, jb2info *info)
{
  Dgp_State *dgp_state = info->dgp_state;
  DGP_PROLOGUE("processSegment()", DGP_PERRSTATE);

  Dgp_Allocators *a = DGP_PALLOCATORS;
    int i;
    DgpAdt_ListNode ln, retnode;
    jb2Segment *curseg;

    /* Dispatch on the segment type */
    DGP_ERR_GEN_CLEANUPF1((processors[segment->segtype] ==
			 (ErrCode (*)(jb2Segment *, jb2info *))NULL),
			cErrNullPtr,
			"bad segment type %d\n", segment->segtype);

    DGP_ERR_PASS_CLEANUP((*processors[segment->segtype])(segment, info));
    
    /*
     * Handle the retention info.  If the segment is flagged as "retained",
     * put it on the active segment list.  Next, look at the segment's
     * referenced segments list; for those that are flagged as "not retained",
     * remove them from the active list.
     */
    if (segment->retain) {
      DGP_REPORTF1((info->verbose), "adding segment %d to active list\n",
		    segment->segnum);

      DGP_ERR_PASS_CLEANUP(DgpAdt_liAdd(info->activesegs, (void *)segment, &retnode));
	}

    for (i = 0; i < segment->nrefsegs; i++) {
	if (!segment->refretain[i]) {
	    /* Slow search; I don't much care */
	  DGP_REPORTF1((info->verbose),
		       "deleting segment %d from active list\n",
			segment->refsegs[i]);

	    DGPADT_FOREACH (ln, info->activesegs) {
		curseg = (jb2Segment *)DGPADT_LIGET(ln);
		if (curseg->segnum == segment->refsegs[i]) {
		    DgpAdt_liRem(ln);
		    jb2FreeSegmentInfo(curseg, dgp_state);
		    (a->free)((void *)curseg);
		    break;
		    }
		}
	    DGP_ERR_GEN_CLEANUPF1((ln == DGP_NULLLISTNODE), cErrNullPtr,
				"couldn't locate segment %d on active list\n",
			segment->refsegs[i]);
	    }
	}

    if (!segment->retain) {
	/* Can free it now */
	jb2FreeSegmentInfo(segment, dgp_state);
	(a->free)((void *)segment);
	}

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
    }


/* Process all the segments */
static ErrCode
process(jb2info *info)
{
  DGP_PROLOGUE("process()", info->DGP_PERRSTATE);

    int i;

    for (i = 0; i < info->nsegs; i++) {
      DGP_ERR_PASS(processSegment(info->segs[i], info));
	}
    NORMAL_RETURN;
    }

int
main(int argc, char **argv)
{
  DGP_MAIN_PROLOGUE("main()", stderr, Dgp_fileErrPrinter);
    char *av0 = argv[0];
    char *fname;
    char *format = (char *)NULL;
    jb2info info;
    int i;
    unsigned char cbuf, lbuf[4];
    DgpAdt_ListNode ln;
    Dgp_Bool verbose = DGP_FALSE;
    Dgp_Bool noout = DGP_FALSE;
    size_t bytes;
    Dgp_State *dgp_state = NULL;


    DGP_ERR_PASS_CLEANUP(Dgp_initDgpState(&dgp_state, NULL, NULL, stderr, Dgp_fileErrPrinter,
				     DGP_FALSE, DGP_FALSE));

    while ((argc >= 2) && (argv[1][0] == '-')) {
	switch (argv[1][1]) {
	  case 'o': {
	    if (argc < 3) {
		fprintf(stderr, usage, av0);
		exit(1);
		}
	    format = argv[2];
	    argc -= 2;
	    argv += 2;
	    break;
	    }

	  case 'v': {
	    verbose = DGP_TRUE;
	    argc--;
	    argv++;
	    break;
	    }

	  case 'n': {
	    noout = DGP_TRUE;
	    argc--;
	    argv++;
	    break;
	    }

	  default: {
	    fprintf(stderr, usage, av0);
	    exit(1);
	    }
	  }
	}

    if(argc < 2) {
      fprintf(stderr, usage, av0);
      exit(1);
    }

    info.dgp_state = dgp_state;
    info.format = format;
    info.verbose = verbose;
    info.noout = noout;

    fname = info.fname = argv[1];

    /* NULL out some pointers */
    info.segs = (jb2Segment **)NULL;

    /* Open the file */
    info.input = fopen(fname, "rb");
    DGP_ERR_GEN_CLEANUPF1((info.input == (FILE *)NULL), cErrNullPtr,
      "can't open %s\n", fname);

    /* and the DataSource */
    setupds(&info);

    /* and the jb2File */
    DGP_ERR_PASS_CLEANUPF0(jb2CreateFile((DataSink)NULL, (DataSink)NULL,
					 dgp_state, &(info.f)),
			   "can't create jb2file\n");

    /* and the active list */
    DgpAdt_liNew(&info.activesegs, info.dgp_state);
    DGP_ERR_GEN_CLEANUPF0((info.activesegs == DGP_NULLLIST), cErrNullPtr,
	"can't create activesegs\n");

    /* Read and check the file header */
    for (i = 0; i < (int)sizeof(jb2header); i++) {
	DSGETC(info.data, &cbuf);
	DGP_ERR_GEN_CLEANUPF3((cbuf != jb2header[i]), cErrDecodeFailure,
	    "header mismatch: byte %d is %02x (should be %02x)\n",
		    i, cbuf, jb2header[i]);
	}
	
    /* Read the one-byte flags field */
    DSGETC(info.data, &cbuf);

    /* Everything but the first 2 bits should be zero */
    DGP_ERR_GEN_CLEANUPF1(((cbuf & ~0x03) != 0), cErrNoImpl,
			  "unimplemented file option %02x\n", cbuf);

    if (cbuf & 0x01) {
	info.sequential = DGP_TRUE;
	}
    else {
	info.sequential = DGP_FALSE;
	}

    if (cbuf & 0x02)
      {
	/* some arbitrary large number */
	info.npages = 65000;
      }
    else
      {
	/* Read the four-byte page count field */
	DSREAD(info.data, 4, lbuf, &bytes);

	/* and get the page count */
	info.npages = (lbuf[0] << 24) | (lbuf[1] << 16) | (lbuf[2] << 8) | lbuf[3];
      }

    if (!info.sequential) {
	/* Random access file - read all the headers */
	DGP_ERR_PASS_CLEANUPF0(readheaders(&info),
	  "problem reading headers\n");

	/* Deal with the segment bodies */
	DGP_ERR_PASS_CLEANUPF0(process(&info),
			       "problem processing\n");
	}
    else {
	jb2Segment *curseg;
	int val;

	/* Sequential file - process one segment at a time */
	for (;;) {
	    /* Read the next segment header, if any */

	    /* Check that we haven't reached EOF */
	  (void)dsPeek(info.data, &val);
	    if (val == EOF) {
		/* Yes - we've hit EOF */
		break;
		}

	    DGP_ERR_CHECK_NO_MSG(readheader(&info, &curseg)) {
		/* Must have been EOF */
		break;
		}

	    /* End early if we saw an EOF segment */
	    if (curseg->segtype == JB2SEGTYPE_EOFILE) {
		/* No point in processing it */
		(DGP_FREE)((void *)curseg);
		break;
		}

	    /* Deal with the segment body */
	    DGP_ERR_PASS_CLEANUP(processSegment(curseg, &info));
	    }
	}

    /* Free whatever we've got */
    if (info.segs != (jb2Segment **)NULL) {
	(DGP_FREE)((void *)info.segs);
	}
    jb2FreeFile(info.f);
    DGPADT_FOREACH (ln, info.activesegs) {
	jb2FreeSegmentInfo((jb2Segment *)DGPADT_LIGET(ln), dgp_state);
	(DGP_FREE)(DGPADT_LIGET(ln));
	}
    DgpAdt_liFree(info.activesegs);
    fclose(info.input);

    Dgp_cleanupDgpState(&dgp_state);

    MAIN_NORMAL_RETURN;

  ERR_CLEANUP:

    MAIN_ERR_RETURN;
    }
