/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-1995, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/

/*************************************************************************
**
**  DigiPaper Codebase
**  
**	  
**  Author:  ?;  199n-9n
**
**  Filename: compile-decoder.c.
**
**  Puts together the source for the Huffman decode tables on the fly (?) / dhtesq.
**
**
**	Modification history:
**		Dillon H. Tracy esq. / Feb. 2000
**		o  Cleaned up after MK's script, which implements most of the dgp_types, et al.
**      o  Scattered liberr compliance issues.
**		o  Search on 'dhtesq' for some of this coder's comments and modifications.
**
**  
**************************************************************************/

/*
 *
 * $Header: /project/icons/cvsroot/digipaper/tools/libtokcomp.help/compile-decoder.c,v 1.2 2000/03/08 23:34:02 dtracy Exp $
 *
 */

#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include "dgp_types.h"
#include "dgp_err_control.h"
//  Removed 21.2.00 / dhtesq.
//#include "undecl-stdio.h"
#include "util.h"
#include "huff.h"
#include "token_t.h"
#include "mmcompressprivate.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/tools/libtokcomp.help/compile-decoder.c,v 1.2 2000/03/08 23:34:02 dtracy Exp $");



/*************************************************************************
**
**  Function:  Dgp_Bool match(Dgp_UInt32 prefix, Dgp_Int32 prefixlen, Dgp_UInt32 code, Dgp_Int32 codelen)
**
**
**  Does the prefix match the initial part of the given code?
**
**
**	Arguments:
**		o  prefix...
**      o  prefixlen...
**      o  code...
**      o  codelen..
**
**	Return value:
**		o  Boolean, true on a match.
**
**	Modification history:
**      o  None.
**  
**/

Dgp_Bool
match(Dgp_UInt32 prefix, Dgp_Int32 prefixlen, Dgp_UInt32 code, Dgp_Int32 codelen) {
  if (prefixlen <= codelen) {
    if ((code >> (codelen - prefixlen)) == prefix) {
      return(DGP_TRUE);
    }
  }
  return(DGP_FALSE);
}
/*************************************************************************/

Dgp_Int8 stateformat[] = "    {%2d, %2d, %3d },\t\t/* %2d: %s%s %s */\n";

typedef struct {
  Dgp_Int32 nbits;
  Dgp_Int32 outstate;
  Dgp_Int32 offset;
  Dgp_Int8 c1[100], c2[100], c3[100];
} entry;

//  What to do with this? / dhtesq.
entry buf[1000];

/*************************************************************************
**
**  Function:  void enter(Dgp_Int32     nbits, 
**                        Dgp_Int32     outstate, 
**                        Dgp_Int32     offset, 
**                        Dgp_Int32     statenum,
**                        Dgp_Int8      *c1, 
**                        Dgp_Int8      *c2, 
**                        Dgp_Int8      *c3)
**
**
**  Fill in various fields in the <statenum> element of buf.
**
**
**	Arguments:
**		o  nbits...
**      o  outstate...
**      o  offset...
**      o  statenum is the index of the entry to be filled.
**      o  c1, c2, c3...
**
**	Return value:
**		o  None.
**
**	Modification history:
**      o  None.
**  
**/

void
enter(Dgp_Int32     nbits, 
      Dgp_Int32     outstate, 
      Dgp_Int32     offset, 
      Dgp_Int32     statenum,
      Dgp_Int8      *c1, 
      Dgp_Int8      *c2, 
      Dgp_Int8      *c3) {
  buf[statenum].nbits = nbits;
  buf[statenum].outstate = outstate;
  buf[statenum].offset = offset;
  strcpy(buf[statenum].c1, c1);
  strcpy(buf[statenum].c2, c2);
  strcpy(buf[statenum].c3, c3);
  return;
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  Dgp_Int8 *printcode(Dgp_UInt32 code, Dgp_Int32 codelen)
**
**
**  Converts the passed code to a character representation.
**
**	Arguments:
**		o  code...
**      o  codelen...
**
**	Return value:
**		o  Character array, stored statically here.
**
**	Modification history:
**      o  None.
**  
**/

Dgp_Int8 *
printcode(Dgp_UInt32 code, Dgp_Int32 codelen) {
  static Dgp_Int8 buf[100];
  Dgp_Int32 i;

  for (i = 0; i < codelen; i++) {
    if (code & (1ul << (codelen - i - 1))) {
      buf[i] = '1';
    }
    else {
      buf[i] = '0';
    }
  }
  buf[i] = '\0';
  return(buf);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  Dgp_Int8 *printx(Dgp_Int32 nx)
**
**
**  Generates a string full of <nx> 'x' characters.
**
**	Arguments:
**		o  nx...
**
**	Return value:
**		o  Character array, stored statically here.
**
**	Modification history:
**      o  None.
**  
**/

Dgp_Int8 *
printx(Dgp_Int32 nx) {
  static Dgp_Int8 buf[100];
  Dgp_Int32 i;

  for (i = 0; i < nx; i++) {
    buf[i] = 'x';
  }
  buf[i] = '\0';
  return(buf);
}    
/*************************************************************************/


/*************************************************************************
**
**  Function:  Dgp_Int8 *printchange(Dgp_Int32 nx, Dgp_Int32 nextstate)
**
**
**  Generates a string containing <nextstate> and <nx> 'x' characters.
**  For reporting only? / dhtesq.
**
**	Arguments:
**		o  nx...
**      o  nextstate...
**
**	Return value:
**		o  Character array, stored statically here.
**
**	Modification history:
**      o  None.
**  
**/

Dgp_Int8 *
printchange(Dgp_Int32 nx, Dgp_Int32 nextstate) {
  static Dgp_Int8 buf[100];
  sprintf(buf, "-> %d+%s", nextstate, printx(nx));
  return(buf);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  Dgp_Int8 *printval(Dgp_Int32 nx, Dgp_Int32 value)
**
**
**  Generates some report string. ? / dhtesq.
**
**	Arguments:
**		o  nx...
**      o  value...
**
**	Return value:
**		o  Character array, stored statically here.
**
**	Modification history:
**      o  None.
**  
**/

Dgp_Int8 *
printval(Dgp_Int32 nx, Dgp_Int32 value) {
  static Dgp_Int8 buf[100];
  if (nx >= 0) {
    if (value >= 0) {
      if (nx > 0) {
        sprintf(buf, "= %s+%d", printx(nx), value);
      } else {
        sprintf(buf, "= %d", value);
      }
    } else {
      sprintf(buf, "= %s%d", printx(nx), value);
    }
  } else {
    if (value >= 0) {
      sprintf(buf, "= v+%d", value);
    } else {
      sprintf(buf, "= -v%d", value);
    }
  }
  return(buf);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  Dgp_Int32 compileRec(encoder      *enc, 
**                                  Dgp_Int32    topstate, 
**                                  Dgp_Int32    firststate,
**                                  Dgp_UInt32   prefix, 
**                                  Dgp_Int32    prefixlen)
**
**
**  Used to generate Huffman decode tables on the fly? / dhtesq.
**
**	Arguments:
**		o  enc...
**      o  topstate...
**      o  firststate...
**      o  prefix...
**      o  prefixlen...
**
**	Return value:
**		o  firststate;  called recursively.
**
**	Modification history:
**      Dillon H. Tracy esq. / Feb. 2000
**      o  Contains assertions, but these won't work anymore without a PROLOGUE.  I inserted a DGP_VOID_PROLOGUE
**         that should do the trick although this function is not strictly void.
**  
**/

Dgp_Int32
compileRec(encoder      *enc, 
           Dgp_Int32    topstate, 
           Dgp_Int32    firststate,
           Dgp_UInt32   prefix, 
           Dgp_Int32    prefixlen) {
  DGP_VOID_PROLOGUE("compileRec()", NULL);
  Dgp_Int32         minbits;
  Dgp_Int32         i;
  Dgp_Bool          matchEOS;
  Dgp_Int32         nextra;
  Dgp_Bool          found = DGP_FALSE;
  codeinfo          *code;

  if ((enc->EOSbits > 0) && match(prefix, prefixlen, enc->EOScode, enc->EOSbits)) {
    /* The EOS code matches the prefix. */
    minbits = enc->EOSbits - prefixlen;
    matchEOS = DGP_TRUE;
    code = (codeinfo *)NULL;
    found = DGP_TRUE;
  } else {
    minbits = 32;
    matchEOS = DGP_FALSE;
    code = (codeinfo *)NULL;
  }

  /* How many bits can we read before the next choice? */
  for (i = 0; i < enc->ncodes; i++) {
    if (match(prefix, prefixlen, enc->codes[i].prefix, enc->codes[i].prefixbits)) {
      found = DGP_TRUE;
      if (enc->codes[i].prefixbits - prefixlen < minbits) {
          minbits = enc->codes[i].prefixbits - prefixlen;
          code = &enc->codes[i];
      }
    }
  }
      
  DGP_VOID_ASSERT(found);

  if (minbits == 0) {
    /* There's an exact match. */
    nextra = 0;
    for (i = 0; i < enc->ncodes; i++) {
      if (match(prefix, prefixlen, enc->codes[i].prefix, enc->codes[i].prefixbits) && 
          (enc->codes[i].prefixbits != prefixlen)) {
          /* There are extra matches. */
          nextra++;
      }
    }
    if ((enc->EOSbits > 0) && match(prefix, prefixlen, enc->EOScode, enc->EOSbits) &&
        enc->EOSbits != prefixlen) {
      nextra++;
    }

    if (nextra == 0) {
      /* Just the single match. Yay! */
      if (code == (codeinfo *)NULL) {
          /* An end code... */
          DGP_VOID_ASSERT(matchEOS);
          enter(-1, 0, 0, topstate, printcode(prefix, prefixlen), "", "= EOS");
      } else {
          /* A regular code... */
          DGP_VOID_ASSERT(code->prefixbits == prefixlen);
          DGP_VOID_ASSERT(code->prefix == prefix);
    
          if (code->from == INT_MIN) {
              /* A special value... */
              DGP_VOID_ASSERT(code->nbits < 0);
              enter(-1, 0, code->upto, topstate, printcode(prefix, prefixlen), "v", printval(-1, code->upto));
          } else if (code->upto == INT_MAX) {
              DGP_VOID_ASSERT(code->nbits < 0);
              enter(-1, 0, code->from, topstate, printcode(prefix, prefixlen), "v", printval(-1, code->from));
          } else {
              enter(code->nbits - prefixlen, 
                    0, 
                    code->from, 
                    topstate, 
                    printcode(prefix, prefixlen), 
                    printx(code->nbits - prefixlen), 
                    printval(code->nbits - prefixlen,
                    code->from));
          }
      }
    } else {
      /* Multiple matches. Joy. */
      DGP_VOID_ASSERT(code->prefixbits == prefixlen);
      DGP_VOID_ASSERT(code->prefix == prefix);
      
      minbits = code->nbits - code->prefixbits;
      enter(minbits,
            firststate, 
            0,
            topstate,
            printcode(prefix, prefixlen), 
            printx(minbits),
            printchange(minbits, firststate));
      for (i = code->from; i <= code->upto; i++) {
          enter(0, 
                0,
                i, 
                firststate + i - code->from,
                printcode((prefix << minbits) + (i - code->from),
                code->nbits), 
                "", 
                printval(0, i));
      }
      topstate = firststate + code->upto - code->from + 1;
      firststate = firststate + (1 << minbits);
      for (i = code->upto + 1; i < code->from + (1 << minbits); i++) {
          firststate = compileRec(enc, 
                                  topstate + i - (code->upto + 1), 
                                  firststate,
                                  (prefix << minbits) + (i - code->from),
                                  prefixlen + minbits);
      }
    }
  } else {
    /* Read minbits bits and decide based on it. */
    enter(minbits, 
          firststate, 
          0,
          topstate, 
          printcode(prefix, prefixlen), 
          printx(minbits),
          printchange(minbits, firststate));
    topstate = firststate;
    firststate = firststate + (1 << minbits);
    for (i = 0; i < (1 << minbits); i++) {
      firststate = compileRec(enc, 
                              topstate + i,
                              firststate,
                              (prefix << minbits) + i,
                              prefixlen + minbits);
    }
  }

  return(firststate);
}
/*************************************************************************/


/*************************************************************************
**
**  Function:  void compileCoder(encoder *enc, Dgp_Int8 *name)
**
**
**  Generates Huffman decode tables on the fly? / dhtesq.
**
**	Arguments:
**		o  enc...
**      o  name...
**
**	Return value:
**		o  None.
**
**	Modification history:
**      o  None.
**  
**/

void
compileCoder(encoder *enc, Dgp_Int8 *name) {
  Dgp_Int32 nstates;
  Dgp_Int32 i;

  nstates = compileRec(enc, 0, 1, 0, 0);
  printf("decodeState %s[] = {\n", name);
  for (i = 0; i < nstates; i++) {
    printf(stateformat, buf[i].nbits, buf[i].outstate, buf[i].offset, i, buf[i].c1, buf[i].c2, buf[i].c3);
  }
  printf("    };\n\n");
}


/*************************************************************************
**
**  Function:  Dgp_Int32 main(Dgp_Int32 argc, Dgp_Int8 **argv)
**
**
**  Generates source for Huffman decode tables.
**
**	Arguments:
**		o  argc, argv...
**
**	Return value:
**		o  Integer;  only control path returns (0) currently.
**
**	Modification history:
**      o  None.
**  
**/

Dgp_Int32
main(Dgp_Int32 argc, Dgp_Int8 **argv) {
  argc = argc; argv = argv;         // (Suppress compiler whine.)
  printf("#include \"../mmcompressprivate.h\"\n");
  printf("DGP_RCSID(\"@(#)decode tables generated by $Header: /project/icons/cvsroot/digipaper/tools/libtokcomp.help/compile-decoder.c,v 1.2 2000/03/08 23:34:02 dtracy Exp $\");\n");

  compileCoder(&mmDeltaXCoders[0], "mmDeltaXDecoder0");
  compileCoder(&mmDeltaXCoders[1], "mmDeltaXDecoder1");
  compileCoder(&mmDeltaXCoders[2], "mmDeltaXDecoder2");

  printf("decoder mmDeltaXDecoders[NDELTAX] = {\n");
  printf("    mmDeltaXDecoder0,\n");
  printf("    mmDeltaXDecoder1,\n");
  printf("    mmDeltaXDecoder2,\n");
  printf("    };\n\n");

  compileCoder(&mmFirstXCoders[0], "mmFirstXDecoder0");
  compileCoder(&mmFirstXCoders[1], "mmFirstXDecoder1");

  printf("decoder mmFirstXDecoders[NFIRSTX] = {\n");
  printf("    mmFirstXDecoder0,\n");
  printf("    mmFirstXDecoder1,\n");
  printf("    };\n\n");

  compileCoder(&mmDeltaYCoders[0], "mmDeltaYDecoder0");
  compileCoder(&mmDeltaYCoders[1], "mmDeltaYDecoder1");
  compileCoder(&mmDeltaYCoders[2], "mmDeltaYDecoder2");

  printf("decoder mmDeltaYDecoders[NDELTAY] = {\n");
  printf("    mmDeltaYDecoder0,\n");
  printf("    mmDeltaYDecoder1,\n");
  printf("    mmDeltaYDecoder2,\n");
  printf("    };\n\n");

  compileCoder(&mmWHCoders[0], "mmWHDecoder0");

  printf("decoder mmWHDecoders[NWH] = {\n");
  printf("    mmWHDecoder0,\n");
  printf("    };\n\n");

  compileCoder(&mmUseCoders[0], "mmUseDecoder0");
  compileCoder(&mmUseCoders[1], "mmUseDecoder1");

  printf("decoder mmUseDecoders[NUSE] = {\n");
  printf("    mmUseDecoder0,\n");
  printf("    mmUseDecoder1,\n");
  printf("    };\n\n");

  return(0);
}
/*************************************************************************/
