/*
 *
 * $Header: /project/icons/cvsroot/digipaper/tools/libtokcomp.util/compile-tables.c,v 1.1 2000/02/24 20:29:24 jahlborn Exp $
 *
 */

#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include "undecl-stdio.h"
#include "util.h"
#include "mmcompressprivate.h"

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/tools/libtokcomp.util/compile-tables.c,v 1.1 2000/02/24 20:29:24 jahlborn Exp $");

static void
printcode(unsigned long code, int codelen)
{
  int i;

  for (i = 0; i < codelen; i++) {
    if (code & (1ul << (codelen - i - 1))) {
      putchar('1');
    }
    else {
      putchar('0');
    }
  }
}

static void
printTable(encoder *enc, char *tablename, char *caption)
{
  int i;

  printf("\\begin{table}\n");
  printf("\\centerline{\n");
  printf("\\begin{tabular}{|l|l|}\n");
  printf("\\hline\n");
  printf("Value & Encoding \\\\ \\hline\n");

  for (i = 0; i < enc->ncodes; i++) {
    if (enc->codes[i].from == INT_MIN) {
      assert(enc->codes[i].nbits < 0);
      printf("$-\\infty \\ldots %d$ & ", enc->codes[i].upto);
      printcode(enc->codes[i].prefix, enc->codes[i].prefixbits);
      printf(" + $ (%d - \\mbox{value})$ encoded as in Table~\\ref{table:bofs}",
	     enc->codes[i].upto);
    }
    else if (enc->codes[i].upto == INT_MAX) {
      assert(enc->codes[i].nbits < 0);
      printf("$%d \\ldots \\infty$ & ", enc->codes[i].from);
      printcode(enc->codes[i].prefix, enc->codes[i].prefixbits);
      printf(" + $ (\\mbox{value} - %d)$ encoded as in Table~\\ref{table:bofs}",
	     enc->codes[i].from);
    }
    else {
      assert(enc->codes[i].nbits >= 0);

      if (enc->codes[i].from != enc->codes[i].upto) {
	printf("$%d \\ldots %d$ & ", enc->codes[i].from,
	       enc->codes[i].upto);
      }
      else {
	printf("$%d$ &", enc->codes[i].from);
      }

      printcode(enc->codes[i].prefix, enc->codes[i].prefixbits);

      if (enc->codes[i].nbits > enc->codes[i].prefixbits) {
	if (enc->codes[i].from < 0) {
	  printf(" + $ (\\mbox{value} + %d)$ encoded as %d bit%s",
		 -enc->codes[i].from,
		 enc->codes[i].nbits - enc->codes[i].prefixbits,
		 ((enc->codes[i].nbits !=
		   enc->codes[i].prefixbits + 1) ? "s" : ""));
	}
	else if (enc->codes[i].from == 0) {
	  printf(" + $ \\mbox{value}$ encoded as %d bit%s",
		 enc->codes[i].nbits - enc->codes[i].prefixbits,
		 ((enc->codes[i].nbits !=
		   enc->codes[i].prefixbits + 1) ? "s" : ""));
	}
	else {
	  printf(" + $ (\\mbox{value} - %d)$ encoded as %d bit%s",
		 enc->codes[i].from,
		 enc->codes[i].nbits - enc->codes[i].prefixbits,
		 ((enc->codes[i].nbits !=
		   enc->codes[i].prefixbits + 1) ? "s" : ""));
	}
      }
    }

    if (i != enc->ncodes - 1) {
      printf("\\\\\n");
    }
  }

  if (enc->EOSbits > 0) {
    printf("\\\\\n");
    printf("END & ");
    printcode(enc->EOScode, enc->EOSbits);
  }

  printf("\\\\ \\hline\n");
  printf("\\end{tabular}\n");
  printf("}\n");
  printf("\\caption{%s}\n", caption);
  printf("\\label{%s}\n", tablename);
  printf("\\end{table}\n");
}

int
main(int argc, char **argv) {
  printf("%% tables generated by $Header: /project/icons/cvsroot/digipaper/tools/libtokcomp.util/compile-tables.c,v 1.1 2000/02/24 20:29:24 jahlborn Exp $\";\n");

  printTable(&mmWHCoders[0], "table:wh0",
	     "Width/height encoding table");

  printTable(&mmUseCoders[0], "table:use0",
	     "Use count encoding table 0");
  printTable(&mmUseCoders[1], "table:use1",
	     "Use count encoding table 1");

  printTable(&mmFirstXCoders[0], "table:firstx0",
	     "First X encoding table 0");
  printTable(&mmFirstXCoders[1], "table:firstx1",
	     "First X encoding table 1");

  printTable(&mmDeltaXCoders[0], "table:deltax0",
	     "Delta X encoding table 0");
  printTable(&mmDeltaXCoders[1], "table:deltax1",
	     "Delta X encoding table 1");
  printTable(&mmDeltaXCoders[2], "table:deltax2",
	     "Delta X encoding table 2");

  printTable(&mmDeltaYCoders[0], "table:deltay0",
	     "Delta Y encoding table 0");
  printTable(&mmDeltaYCoders[1], "table:deltay1",
	     "Delta Y encoding table 1");
  printTable(&mmDeltaYCoders[2], "table:deltay2",
	     "Delta Y encoding table 2");

  return(0);
}
