

// for safety reasons we'll include this first
#ifdef _WIN32
#include <winsock2.h>
#endif 

#include <stdio.h>
#include "dgp_err_control.h"
#include "dgp_err_printers.h"
#include "qdr.h"
#include "colorCC.h"
#include "image.h"
#include "image_Defs.h"


// Need to set files to binary
#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif 

DGP_RCSID("@(#)$Header: /project/icons/cvsroot/digipaper/tools/tcomp-lossless.util/qdrtoppm.c,v 1.6 2000/06/28 15:46:52 jahlborn Exp $");


void 
fillRun(DgpCC_CccRun *run, Dgp_Int32 x, Dgp_Int32 y,
	DgpImage_ColorRGB im, DgpImage_Packed bim)
{
  Dgp_Int32 i;

  if(run->sampled) {

    memcpy(DGPIMG_PIXPTR(im, x, y), run->colorInfo.rgbSamples, run->length * sizeof(DgpImage_RGB));

    if(bim != NULL) {
      for(i=0; i < run->length; i++){
        DGPIMG_SETBIT(bim, x+i, y);
      }
    }

    run->sampled = DGP_FALSE;

  } else {

    for(i=x;i< x+run->length;i++){
      // convert to rgb
			
      //DGPIMG_REF(im,i,y) = run->colorInfo.rgbColor;
      (DGPIMG_REF(im,i,y)).r = (Dgp_UInt8)(DGPCC_LRGB_R(run->colorInfo.rgbColor));
      (DGPIMG_REF(im,i,y)).g = (Dgp_UInt8)(DGPCC_LRGB_G(run->colorInfo.rgbColor));
      (DGPIMG_REF(im,i,y)).b = (Dgp_UInt8)(DGPCC_LRGB_B(run->colorInfo.rgbColor));
    }
  }
}

void
printUsage()
{
  printf("usage: qdrtoppm [options] [infile]\n\t-o  prefix of output file names\n");
  exit(EXIT_FAILURE);
}

Dgp_Int32
main(Dgp_Int32 argc, char **argv)
{
  DGP_MAIN_PROLOGUE("main()", stderr, Dgp_fileErrPrinter);

  FILE *inFp = NULL, *outFp = NULL;
  uds u;
  Dgp_Int32 color_type;
  Dgp_Int32 width, height;
  Dgp_Int32 xres, yres;
  Dgp_Int32 x, y;
  DgpImage_ColorRGB outIm = NULL;
  DgpImage_Packed outbIm = NULL;
  DgpCC_CccRun run;
  Dgp_Int8 outname[64] = "";
  Dgp_Int8 boutname[64] = "";
  Dgp_Bool close_in = FALSE, close_out = FALSE, do_bout = FALSE;
  DgpCC_PFNReadRun run_reader;

#ifdef _WIN32
  Dgp_Int32 result;
#endif

  Dgp_Bool eol;
  Dgp_State *dgp_state = NULL;

  
  /* init dgp state */
  DGP_ERR_PASS(Dgp_initDgpState(&dgp_state, NULL, NULL, stderr, Dgp_fileErrPrinter,
    FALSE, FALSE));
  DGP_SET_ERR_STATE(DGP_PERRSTATE);  

  // set files instream
  if(argc > 2) {
    argv++;
    argc--;
 
    while(argc > 1) {

      if(strcmp(argv[0], "-o") == 0) {
	      argv++;
        argc--;

        if(argc < 1) {
          printUsage();
        }

        sprintf(outname, "%s.ppm", argv[0]);
        sprintf(boutname, "%s.pbm", argv[0]);
        
        argv++;
        argc--;

        do_bout = TRUE;

        outFp = fopen(outname, "wb");
        if(outFp == NULL) {
          fprintf(stderr, "Couldn't open: %s\n", argv[1]);
          MAIN_ERR_RETURN;
        }
        close_out = TRUE;

        if(argc > 0) {
          inFp = fopen(argv[0], "r");
          close_in = TRUE;
        } else {
          inFp = stdin;
        }

      } else {
        printUsage();
        MAIN_ERR_RETURN;
      }


    }

  } else if(argc == 2) {
    if(argv[1][0] == '-') {
      printUsage();
      MAIN_ERR_RETURN;
    }

    inFp = fopen(argv[1], "r");
    if(inFp == NULL){
      fprintf(stderr, "Couldn't open: %s\n", argv[1]);
      MAIN_ERR_RETURN;
    }
    outFp = stdout;
  } else if (argc == 1) {
    inFp = stdin;
    outFp = stdout;
  } else {
    printUsage();
  }

  // Need to set files to binary in Windows
#ifdef _WIN32

  result = _setmode( _fileno( inFp ), _O_BINARY );   
  if( result == -1 )
    perror( "Cannot set mode" );

  result = _setmode( _fileno( outFp ), _O_BINARY );   
  if( result == -1 )
    perror( "Cannot set mode" );
#endif 

  // init instream
  u.fp = inFp;
  u.fseek  = (fseek_func)fseek;
  u.ftell  = (ftell_func)ftell;
  u.fread  = (fread_func)fread;
  u.fwrite = (fwrite_func)fwrite;

  DGP_ERR_PASS(readHeaderQDR(&color_type, &width, &height, 
    &xres, &yres, &u, dgp_state));

  switch(color_type) {
  case QDR_COLORTYPE_RGB:
    run_reader = readRunQDR_rgb;
    break;
  case QDR_COLORTYPE_CMYK:
    run_reader = readRunQDR_cmyk;
    break;
  case QDR2_COLORTYPE_RGB:
    run_reader = readRunQDR2_rgb;
    break;
  }

  fprintf(stderr, "Width: %d, Height: %d, Xres: %d, Yres: %d\n", width, height,
    xres, yres);

  DGP_ERR_PASS(DGPIMAGE_NEW(dgp_state, DGPIMAGE_RGB | DGPIMAGE_DONTCLEAR, width, height, &outIm));
  if(outIm == NULL) {
    fprintf(stderr, "could not open output image\n");
    MAIN_ERR_RETURN;
  }

  if(do_bout) {
    DGP_ERR_PASS(DGPIMAGE_NEW(dgp_state, DGPIMAGE_PACKED, width, height, &outbIm));
    if(outbIm == NULL) {
      fprintf(stderr, "could not open binary output image\n");
      MAIN_ERR_RETURN;
    }
  }

  x = 0;
  y = 0;
  eol = DGP_FALSE;
  run.sampled = DGP_FALSE;
  while(DGP_TRUE) {
    DGP_ERR_PASS(run_reader(&run, &eol, (DgpCC_RunCookie)&u, dgp_state));

    if(eol) {
      eol = DGP_FALSE;
      // new line marker 
      y++;
      x = 0;
      if(y==height) {
	      break;
      }
      continue;
    }
		
    fillRun(&run, x, y, outIm, outbIm);
    x += run.length;
  }

  DgpImage_pSaveF(outIm, outFp);

  if(do_bout) {
    DGPIMAGE_SAVE(outbIm, boutname);
  }

  if(close_in) {
    fclose(inFp);
    inFp = NULL;
  }

  if(close_out) {
    fclose(outFp);
    outFp = NULL;
  }

  Dgp_cleanupDgpState(&dgp_state);

  MAIN_NORMAL_RETURN;
}
