/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxread.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxread.c:	TIFF-FX READ source
 *
 *	Handles read-only TIFF-FX handling.
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tfxread.c,v $; $Revision: 1.3 $; $Date: 2003/02/11 01:31:42 $");

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "props.pub"
#include "frames.pub"
#include "imageref.pub"
#include "shrpixr.pub"
#include "shrpixr.prv"
#include "shrcmprs.pub"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif
#endif	/* NOIPCORE */

#include "tfximage.h"
#include "tfxrtiff.h"
#include "tfxrtfx.h"
#include "tfxrpriv.h"

#ifndef NOIPCORE
#include "tfxrjpeg.h"
#include "tfx.pub"

#else	/* NOIPCORE */
#ifndef NODGP
#define NODGP
#endif
#endif	/* NOIPCORE */

#ifndef TFX_READONLY
#include "tfxwtfx.h"
#include "tfxwtiff.h"
#endif /* TFX_READONLY */

#include "tfxrjb2.h"	/* Has the header structure */

#if !defined(NODGP) && !defined(NOIPCORE)
#define _IPWSTATE_DEFINED_  /* Avoid duplicate definition */
/*#include "jtokapi.h" */
#include "decode_api.h"
#endif	/* NODGP */


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Initialize a TIFF-FX instance, used to identify the TIFF-FX state.
 *
 * LongDescription:::
 *  TFX_InitInstance() initializes a TIFF-FX instance, used to store file I/O
 *  routines connecting the instance to a handle, for maintaining the
 *  malloc/free functions, and for retaining the default authoring description.
 *
 * Input:::
 *  dwClientInstID::	Client ID assigned to this instance
 *
 * Output:::
 *  pxInstHandle::		Pointer to the instance handle returned
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_InitInstance(
    UInt32 dwClientInstID,
    IO_INSTHANDLE *pxInstHandle
)
{
    TFX_INSTANCE *instData = NULL;
    TfxMemoryTally *tally = NULL;
    PROLOGUE;

    /* Return nothing if failure */
    *pxInstHandle = (IO_INSTHANDLE)instData;

    /* Here's the only calls to the system malloc */
    IO_CHECKMALLOC(instData = (TFX_INSTANCE *)malloc(sizeof(*instData) +
				       sizeof(*tally)));
    tally = (void *)((char *)instData + sizeof(*instData));
    
    memset(instData, 0, sizeof(*instData));
    memset(tally, 0, sizeof(*tally));
    
    /* Set defaults */
    instData->handleID = (UInt32)instData;	/* Consistency check */
    instData->handleType = TFX_HANDLE_TYPE;
    instData->nextDocID = 0;	/* Doc unique IDs */
    instData->tfxMagicNum = TFX_MAGIC_NUM;
    
    /* Preserve the default ID for switching back and forth between the
     * client ID and the reserved default */
    instData->defaultMallocID = (void *)tally;
    
    /* Save user identifier */
    instData->clientInstID = dwClientInstID;
    
    /* Initialize the malloc functions to the defaults */
    XERR_PASS_CLEANUP(TFX_SetMallocFuncs((IO_INSTHANDLE)instData,
			    instData->defaultMallocID, wp_TfxMalloc,
			    wp_TfxFree));
    
    /* Return what we did, obscured from the client */
    *pxInstHandle = (IO_INSTHANDLE)instData;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (instData) {
	free(instData);
    }
    if (tally) {
	free(tally);
    }
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Declare a malloc and free function to be used for memory calls.
 *
 * LongDescription:::
 *  TFX_SetMallocFuncs() declares a malloc and free function to be used in
 *  all subsequent memory calls. This allows the client to provide their own
 *  memory management. Without this call, default internal routines are used
 *  instead, which may use the system malloc()/free() pair.
 *
 * Input:::
 *  xInst::			TIFF-FX instance handle, returned from
 *				TFX_InitInstance()
 *  dwClientMallocID::		Identifier pass back to functions when called
 *  xMalloc::			Memory allocator function
 *  xFree::			Memory free'r function
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetMallocFuncs(
    IO_INSTHANDLE xInst,
    void *dwClientMallocID,
    IO_MALLOCFUNC xMalloc,
    IO_FREEFUNC xFree
)
{
    TFX_INSTANCE *instData = (TFX_INSTANCE *)xInst;
    PROLOGUE;
    
    TFX_CheckHandle(xInst);	/* Must be valid, even though not used */

    if (xMalloc && xFree) {	/* Must both be set to use clients */
	instData->malloc = xMalloc;
	instData->free = xFree;
	instData->mallocID = dwClientMallocID;
    } else if (!xMalloc || !xFree) {
	__error__ = IO_BADPARAMETER;
	goto ERR_CLEANUP;
    } else {
	
	/* NULL functions indicate to use the defaults */
	instData->malloc = wp_TfxMalloc;
	instData->free = wp_TfxFree;
	instData->mallocID = instData->defaultMallocID;
    }
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the "TIFF-FX instance" ID associated with a TIFF-FX document handle
 *
 * LongDescription:::
 *  TFX_EndInstance() returns the "TIFF-FX instance" ID previously associated
 *  with a TIFF-FX document handle.
 *
 * Input:::
 *  tfxDocHandle::		Active document handle
 *  pxInstHandle::		Pointer to the TIFF-FX instance handle to be
 *				returned, originally returned from
 *				TFX_InitInstance(), and to be passed in to
 *				TFX_EndInstance().
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetInstance(
    IO_DOCHANDLE tfxDocHandle,
    IO_INSTHANDLE *pxInstHandle
)
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    PROLOGUE;
    
    TFX_CheckHandle(tfxDocHandle);

    *pxInstHandle = (IO_INSTHANDLE)TFX_DOCINST(tfxDoc);
    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  End a previously estalished TIFF-FX instance.
 *
 * LongDescription:::
 *  TFX_EndInstance() ends a previously estalished TIFF-FX instance. It is an
 *  error to end an instance with a current document open (a call to
 *  TFX_CloseDocument() is required first).
 *
 * Input:::
 *  xInst::			TIFF-FX instance handle, returned from
 *				TFX_InitInstance()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_EndInstance(
    IO_INSTHANDLE xInst
)
{
    TFX_INSTANCE *instData = (TFX_INSTANCE *)xInst;
    PROLOGUE;
    
    TFX_CheckHandle(xInst);

#ifndef NOIPCORE    
    /* End IPCore instance */
    if (instData->pIPWState && instData->myIPWState) {
	XERR_PASS_CLEANUP(w_IPWStateTerm((IPWState **)&instData->pIPWState));
    }
#endif
    
    /* Check known TIFF-FX memory for memory leaks */
    if (XERR_IGNORE(wp_TfxMemoryCheck((TfxMemoryTally *)instData->defaultMallocID))) {
	/* Assume something wasn't called correctly */
	__error__ = IO_SEQUENCE;
	/* Contine on with dealloc */
    }
    
    /* Due to potential realloc() calls, some operating systems allow this
     * pointer to propogate. Invalidate before we free it. */
    memset((void *)instData->defaultMallocID, 0, sizeof(TfxMemoryTally));
    memset(instData, 0, sizeof(*instData));

    if (instData->defaultMallocID) {
	free((void *)instData->defaultMallocID);
    }
    free(instData);
    
    if (__error__ != IO_NOERROR) {
	goto ERR_CLEANUP;    /* This is silly, but now we reference the label */
    }
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Check a file for a header identifying it as a currenly supported TIFF-FX.
 *
 * LongDescription:::
 *  TFX_IsTfx() reads enough of a file to identify it as a currently
 *  support TIFF-FX file. The client must only provide enough input parameters
 *  to allow reading and roaming the file to make such a decision. The client
 *  ID is a parameter passed back through the I/O functions, such that any of
 *  these supporting functions could use the ID to identify a file descriptor
 *  or file pointer, or even possibly a structure with this and a file state.
 *  The TIFF-FX file's version and revision are returned if the file is a TIFF-FX,
 *  and are zeroed otherwise. IO_NOSUPPORT is returned if not supported.
 *  IO_BADFILEFORMAT is returned if not a TIFF-FX.
 *
 * Input:::
 *  xInst::			TIFF-FX instance handle, returned from
 *				TFX_InitInstance()
 *  dwClientID::		Client ID, passed back through all I/O funcs
 *  xReadFunc::			Client file read function
 *  xSeekFunc::			Client file seek function
 *  xSizeFunc::			Client file size function
 *
 * Output:::
 *  pdwVersion::		Pointer to the version
 *  pdwRevision::		Pointer to the revision
 *  pdwOptions::		Pointer to the format options
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_IsTfx(
	IO_INSTHANDLE xInst,
	Int32 dwClientID,
	IO_READFUNC xReadFunc,
	IO_SEEKFUNC xSeekFunc,
	IO_SIZEFUNC xSizeFunc,
	UInt8 *pVersion,
	UInt8 *pRevision,
	UInt32 *pdwOptions
)
{
    TFX_FILE tfxFile;
    UInt32 firstIFD = 0UL;
    PROLOGUE;
    
    /* Return nothing until something found */
    *pVersion = 0UL;
    *pRevision = 0UL;
    *pdwOptions = 0UL;
    
    /* Check the parameters */
    XERR_GEN_CLEANUP(xInst == 0UL, IO_BADPARAMETER);
    
    /* Setup a file structure for temporary use */
    memset(&tfxFile, 0, sizeof(TFX_FILE));
    tfxFile.handleID = (UInt32)&tfxFile;
    tfxFile.instance = (TFX_INSTANCE *)xInst;
    TFX_FILETOKEN(&tfxFile)->dwClientFileID = dwClientID;
    TFX_FILETOKEN(&tfxFile)->FileRead = xReadFunc;
    TFX_FILETOKEN(&tfxFile)->FileSeek = xSeekFunc;
    TFX_FILETOKEN(&tfxFile)->FileSize = xSizeFunc;
    
    /* Check to see if it's a TIFF-FX file, using temporary tfxFile token */
    XERR_IGNORE(__error__ = wp_TfxIsTfx(&tfxFile, pVersion, pRevision, &firstIFD));
    
 ERR_CLEANUP:
    
    wp_TfxFileUnLock(TFX_FILETOKEN(&tfxFile));
    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Open a document for read only
 *
 * LongDescription:::
 *  TFX_OpenDocumentRead() begins a session of reading a document. The
 *  client must already have opened the file at the operating system level,
 *  so that the read and seek methods in the file token are available.
 *  This function will perform any initialization required for reading
 *  the file, including reading enough header information to verify
 *  that the file is in a format supported by this version of
 *  the TIFF-FX library.
 *  <BR>
 *  A document handle is returned through the third parameter. This handle
 *  is passed to all subsequent function calls that use the document, and
 *  will encompass the I/O functions passed in here.
 *
 * Input:::
 *  xInstHandle::		TIFF-FX instance handle, returned from
 *				TFX_InitInstance()
 *  dwClientID::		Input client ID
 *  xReadFunc::			Client file read function
 *  xWriteFunc::		Client file write function
 *  xSeekFunc::			Client file seek function
 *  xSizeFunc::			Client file size function
 *  xTellFunc::			Client file tell function
 *
 * Output:::
 *  pTfxDocHandle::		Document handle used to refer to this
 *				page in other functions
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_OpenDocumentRead(
	IO_INSTHANDLE xInstHandle,
	Int32 dwClientID,
	UInt32 options,
	IO_READFUNC xReadFunc,
	IO_WRITEFUNC xWriteFunc,
	IO_SEEKFUNC xSeekFunc,
	IO_SIZEFUNC xSizeFunc,
	IO_TELLFUNC xTellFunc,
	IO_DOCHANDLE *pTfxDocHandle
)
{
    TFX_INSTANCE *instData = (TFX_INSTANCE *)xInstHandle;
    TFX_DOC *tfxDoc = NULL;
    TFX_FILE *tfxFile = NULL;
    UInt8 version = 0, revision = 0;
    UInt32 firstIFD = 0UL;
    PROLOGUE;

    /* Clear the results for now */
    *pTfxDocHandle = 0UL;

    TFX_CheckHandle(xInstHandle);

    /* Check for proper calling sequence */


    /* Create a new document (handle) */
    IO_CHECKMALLOC(tfxDoc = (TFX_DOC *)TFX_MALLOC(sizeof(TFX_DOC)));
    memset(tfxDoc, 0, sizeof(*tfxDoc));
    tfxDoc->handleID = (UInt32)tfxDoc;	/* Self identity */
    tfxDoc->handleType = TFX_HANDLE_TYPE;
    tfxDoc->docID = ++instData->nextDocID;/* ID unique to this library instance
					 */

    /* Create a link back to the instance */
    IO_CHECKMALLOC(tfxFile = (TFX_FILE *)TFX_MALLOC(sizeof(TFX_FILE)));
    memset(tfxFile, 0, sizeof(*tfxFile));
    tfxFile->handleID = (UInt32)tfxFile;	/* Self identity */

    /* Setup the file handle */
    tfxFile->instance = instData;
    TFX_FILETOKEN(tfxFile)->dwClientFileID = dwClientID;
    TFX_FILETOKEN(tfxFile)->FileRead = xReadFunc;
    TFX_FILETOKEN(tfxFile)->FileWrite = xWriteFunc;
    TFX_FILETOKEN(tfxFile)->FileSeek = xSeekFunc;
    TFX_FILETOKEN(tfxFile)->FileSize = xSizeFunc;
    TFX_FILETOKEN(tfxFile)->FileTell = xTellFunc;
    
    /* Retain the file handle */
    tfxDoc->tfxFile = tfxFile;
    
    /* Wait until I have full control over the file */
    XERR_PASS_CLEANUP(wp_TfxFileLock(TFX_FILETOKEN(tfxFile)));
    
    /* Check the TIFF-FX header */
    __error__ = wp_TfxIsTfx(tfxFile, &version, &revision, &firstIFD);
    switch ((int)__error__) {
    case IO_NOERROR:
	break;
    case IO_NOTTIFF:
	/* Not even a valid TIF */
	goto ERR_CLEANUP;
#if 0
    case TFX_NOTTIFFFX:
	/* Not a valid TIFF-FX, but potentially readable. Continue on but prepare
	 * for later failure. */
	__error__ = IO_NOERROR;
	break;
#endif
    case IO_NOSUPPORT:
	/* Version not supported */
	goto ERR_CLEANUP;
    default:
	/* Other error (read/seek/etc.) */
	goto ERR_CLEANUP;
    }
    
    /* Initialize */
    tfxDoc->seqStatus |= TFX_DOC_SEQ_READSTART;
    tfxDoc->readWrite = READ_ALLOWED;
    tfxDoc->docInfo.version = version;
    tfxDoc->docInfo.revision = revision;

    if (firstIFD) {
	/* Start a linked list of IFD offsets */
	IO_CHECKMALLOC(tfxDoc->ifdList = (TFX_IFDOFFSET_ITEM *)
	    TFX_MALLOC(sizeof(*tfxDoc->ifdList)));
	memset(tfxDoc->ifdList, 0, sizeof(*tfxDoc->ifdList));
	tfxDoc->ifdList->ifdOffset = firstIFD;
    }

    tfxDoc->readWrite = READ_ALLOWED;
    
    /* Return the result */
    *pTfxDocHandle = (UInt32)tfxDoc;
    
    XERR_PASS_CLEANUP(wp_TfxFileUnLock(TFX_FILETOKEN(tfxFile)));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (tfxDoc) {
	TFX_FREE(tfxDoc);
    }
    if (tfxFile) {
	TFX_FREE(tfxFile);
    }
    
    if (tfxFile) {
	wp_TfxFileUnLock(TFX_FILETOKEN(tfxFile));
	tfxFile = NULL;
    }

    /* Avoid compiler warning */
    options = 0;

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the basic content of the document.
 *
 * LongDescription:::
 *  TFX_GetDocInfo() returns the document specific content, including: TIFF-FX
 *  version, TIFF-FX revision, number of pages, authoring description, and the
 *  reference page.
 *
 * Input:::
 *  tfxDocHandle::		Input document handle
 *
 * Output:::
 *  pdwVersion::		Version of TIFF-FX
 *  pdwRevision::		Revision of this version of TIFF-FX
 *  pdwNPages::			Number of pages in file
 *  ppAuthorBlock::		Author string (NULL if none)
 *  pwAuthorBlockBytes::	Size of author string
 *  pdwReferencePageOffset::	Reference page offset (-1 if none)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetDocInfo(
	IO_DOCHANDLE tfxDocHandle,
	UInt8 *pVersion,
	UInt8 *pRevision,
	UInt16 *pwNPages
)
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    TFX_DOCINFO *docInfo;
    ErrCode warning = 0;
    TFX_IMG *primImg;
    PROLOGUE;
    
    TFX_CheckHandle(tfxDocHandle);
    
    /* Check for proper calling sequence */
    
    /* Clear the inputs */
    *pVersion = *pRevision = 0;
    *pwNPages = 0;
    
    /* If it has already been read, don't read again */
    if ((tfxDoc->seqStatus & TFX_DOC_SEQ_GETINFO) == 0) {
	UInt32 *pImageCount = NULL;
	UInt32 layers;
	UInt16 annotations;
	IO_PAGEHANDLE pageHandle;

	/* Read through the first primary IFD */
	XERR_PASS_CLEANUP(wp_TfxTraversePrimaryImages(tfxDoc, 1));

	/* If we have a page table, and the page count is > 0, and the
	 * number of entries per table is 0, something is wrong */
/*	XERR_GEN_CLEANUP(tfxDoc->pageTableHeader.pageCount > 0 && */
/*			 tfxDoc->pageTableHeader.entriesPerTable == 0, */
/*			 IO_BADFILEFORMAT); */
	
	/* Update the status flags */
	tfxDoc->seqStatus |= TFX_DOC_SEQ_GETINFO | TFX_DOC_SEQ_READSTART; /* | TFX_DOC_SEQ_W_PAGETABLE; */

	/* The document information is burried in the GlobalParametersIFD
         * found in the first IFD (well, it's recommended that it reside there).
	 * Check to see if one exists. */
	XERR_IGNORE(warning = TFX_GetPageInfo(tfxDocHandle, 0, &pImageCount, &layers,
	    &annotations, &pageHandle));
	XERR_GEN_CLEANUP(warning && warning < IO_WARNINGS, warning);

	/* If the page count isn't found on the first IFD, read through
	 * the rest of the primary IFDs to obtain the page count */
	primImg = (TFX_IMG *)((TFX_PAGE *)pageHandle)->pageInfo.tfxPrimImgHandle;
	XERR_GEN_CLEANUP(primImg == NULL, IO_INTERNAL_ERROR);
	if (primImg->fields.PageNumber[1] == 0) {
	    XERR_PASS_CLEANUP(wp_TfxTraversePrimaryImages(tfxDoc, 0));
	} else {
	    tfxDoc->docInfo.wNPages = primImg->fields.PageNumber[1];
	}

	/* Read the GlobalParametersIFD if it's there */
#if 0
	if (0) {

	    /* Read the page extension tags */
	    XERR_PASS_CLEANUP(wp_TfxReadExtensions(tfxDoc));
	
	    /* Read the extension data (pagetable, reference page, author, etc. */
	    XERR_PASS_CLEANUP(wp_TfxReadTagData(tfxDoc));
	}
#endif

    }
    docInfo = &tfxDoc->docInfo;

    /* Set the sequence */
    tfxDoc->seqStatus |= TFX_DOC_SEQ_READSTART;


    /* Return the results */
    *pVersion = docInfo->version;
    *pRevision = docInfo->revision;
    *pwNPages = docInfo->wNPages;

    if (__error__ == 0) {
	__error__ = warning;	/* If a warning exists, use if no error */
    }
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the basic content of the page
 *
 * LongDescription:::
 *  TFX_GetPageInfo() returns the page specific content, including:
 *  image layer counts, number of layers present, annotation count, and a
 *  handle used to reference this page for other functions.
 *
 * Input:::
 *  tfxDocHandle::		Document handle
 *  dwPageOffset::		Page offset within the document (-1 for the
 *				last page, including incompleted last page)
 *
 * Output:::
 *  ppdwImageCounts::		Number of images for each image layer (array,
 *				this may be NULL, in which case no array is
 *				returned)
 *  pdwLayersPresent::		Number of image layers present (indicates size
 *				of array pdwImageLayers[], may have missing
 *				layers)
 *  pdwAnnotationCount::	Number of annotations on page
 *  pTfxPageHandle::		Page handle, used to identify this page
 *				in the page level routines
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetPageInfo(
	IO_DOCHANDLE tfxDocHandle,
	Int32 dwPageOffset,
	UInt32 **ppdwImageCounts,
	UInt32 *pdwLayersPresent,
	UInt16 *pwAnnotationCount,
	IO_PAGEHANDLE *pTfxPageHandle
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    TFX_PAGE *tfxPage = NULL;
    TFX_PAGE *searchPage = NULL;
    TFX_IMG *tfxImg = NULL;
    UInt8 flags;
    Int32 ifdOffset;
    TFX_PAGEINFO tfxPageInfo;
    UInt32 xOffset, yOffset, xRes, yRes, width, height;
    UInt32 bestStripHeight;
    IO_COMPRESSION compression;
    IO_COLORTYPE colorType;
    UInt8 palette[256*3];
    IO_IMGHANDLE primImgHandle;
    ErrCode warning = 0;
    PROLOGUE;
    
    TFX_CheckHandle(tfxDocHandle);

    /* Return nothing by default */
    if (ppdwImageCounts) {
	*ppdwImageCounts = NULL;
    }
    *pdwLayersPresent = 0;
    *pwAnnotationCount = 0;
    *pTfxPageHandle = 0;

    /* Check for proper calling sequence and parameters... */

    /* Must have the document open for read or write */
    XERR_GEN_CLEANUP(((tfxDoc->seqStatus &
	(TFX_DOC_SEQ_READSTART | TFX_DOC_SEQ_WRITESTART)) == 0),
	IO_SEQUENCE);

    /* Must have gotten the document info (either from TFX_GetDocInfo() or
     * from TFX_OpenDocument... */
    XERR_GEN_CLEANUP((tfxDoc->seqStatus & TFX_DOC_SEQ_GETINFO) == 0,
		     IO_SEQUENCE);

    /* The page should be a valid existing page, or a page that was just
     * started in the write process (TFX_AddPageStart was called). */
    if (dwPageOffset < 0) {
	dwPageOffset = tfxDoc->docInfo.wNPages - 1;
    }
#if 0
    /* We don't know the number of pages in a TIFF-FX file, until we've
     * traversed all of the primary IFDs */
    XERR_GEN_CLEANUP(dwPageOffset >= tfxDoc->pageTableHeader.pageCount &&
	dwPageOffset >= tfxDoc->docInfo.wNPages, IO_BADPARAMETER);
#endif
    
    /* Clear the outputs */
    *pwAnnotationCount = 0;
    *pTfxPageHandle = 0UL;
    memset(&tfxPageInfo, 0, sizeof(tfxPageInfo));
    
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);
    
    /* Look for a previous read of this page */
    searchPage = tfxDoc->pages;
    if (searchPage != NULL) {
	do {
	    if (searchPage->pageInfo.dwPageOffset == dwPageOffset) {
		
		/* Found, this page already visited */
		tfxPage = searchPage;
		break;
	    }
	} while ((searchPage = searchPage->next) != NULL);
    }
    
    if (searchPage == NULL || searchPage->pageInfo.dwPageOffset != dwPageOffset) {
	
	/* Make sure we have the page table (IFD offset) for the correct page */
	XERR_PASS_CLEANUP(wp_TfxGetPageIFD(tfxDoc, dwPageOffset, &flags, &ifdOffset));
	
	/* There are no pages yet.  Return gracefully (could be requesting doc info
	 * on an empty document */
	XERR_GEN_CLEANUP(ifdOffset == 0, IO_NOERROR);
	XERR_GEN_CLEANUP(ifdOffset < 8, IO_BADFILEFORMAT);
	
	/* Begin construction of the page data */
	IO_CHECKMALLOC(tfxPage = (TFX_PAGE *)TFX_MALLOC(sizeof(TFX_PAGE)));
	memset(tfxPage, 0, sizeof(TFX_PAGE));
	tfxPage->handleID = (UInt32)tfxPage;	/* self identity */
	tfxPage->handleType = TFX_HANDLE_TYPE;
	tfxPage->tfxDoc = tfxDoc;
	tfxPage->pageInfo.tfxPageHandle = (UInt32)tfxPage;
	tfxPage->pageInfo.dwPageOffset = dwPageOffset;
	
	/* Begin construction of the primary image data */
	XERR_PASS_CLEANUP(wp_TfxNewImg(tfxPage,
				    (TFX_IMG **)&tfxPage->pageInfo.tfxPrimImgHandle,
				    ifdOffset));

	/* Note: it is possible to have the primary image with a layer 1 or layer 3
	 * designation, however, no other layers and no other images within this layer
	 * would be allowed.
	 * For now, let's assume this is layer 2, first image.  If marked otherwise,
	 * the IFD field will change it. */
	((TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle)->imgInfo.dwImgLayer = 2;
	((TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle)->imgInfo.dwImgOffset = 0;
	((TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle)->imgInfo.tfxImgType = IO_IMGTYPE_PRIMARY;

	/* Keep in image list also (note, it will be pointed to by both the
	 * primImgHandle and the img list */
/*	tfxPage->tfxImgs = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle; */
	
	/* Attach it to the appropriate spot in linked list */
	searchPage = tfxDoc->pages;
	if (searchPage == NULL) {
	    tfxDoc->pages = tfxPage;
	} else {
	    do {
		if (! searchPage->next ||
		    searchPage->next->pageInfo.dwPageOffset > dwPageOffset) {
		    
		    /* Found the right spot to insert this page (after) */
		    break;
		}
	    } while ((searchPage = searchPage->next) != NULL);

	    /* If at the beginning, insert as first */
	    if (searchPage == tfxDoc->pages &&
		searchPage->pageInfo.dwPageOffset > dwPageOffset) {
		tfxPage->next = searchPage;
		searchPage->prev = tfxPage;
		tfxDoc->pages = tfxPage;
	    } else {
	    
		/* Insert inbetween searchPage and searchPage->next */
		tfxPage->next = searchPage->next;
		tfxPage->prev = searchPage;
		if (searchPage->next) {
		    searchPage->next->prev = tfxPage;
		}
		searchPage->next = tfxPage;
	    }
	}
    }

    XERR_GEN_CLEANUP(tfxPage == NULL, IO_INTERNAL_ERROR);

    /* We have the page (either already read or not), read if not */
    if ((tfxPage->seqStatus & TFX_PAGE_SEQ_GETINFO) == 0) {

	/* If the primary image has not been added yet, return counts of zero */
#if 0
	if ((tfxPage->pageInfo.dwLayersPresent == 0 &&
	    tfxDoc->pageTableHeader.pageCount != tfxDoc->docInfo.wNPages) ||
	    tfxPage->pageInfo.tfxPrimImgHandle == 0) {
	    *pTfxPageHandle = tfxPage->pageInfo.tfxPageHandle;
	    NORMAL_RETURN;
	}
#else
	if (tfxPage->pageInfo.dwLayersPresent == 0 &&
	    tfxPage->pageInfo.tfxPrimImgHandle == 0) {
	    *pTfxPageHandle = tfxPage->pageInfo.tfxPageHandle;
	    NORMAL_RETURN;
	}
#endif
	
	/* Set the sequence */
        tfxPage->seqStatus |= TFX_PAGE_SEQ_READSTART;

	/* Parse through the page hierarchy, and tally construction stats */
	
	/* Go to the primary IFD and read the tags */
	tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;
	XERR_IGNORE(warning = TFX_GetImageInfo((UInt32)tfxPage,
	    0, 0, &xOffset, &yOffset, &xRes, &yRes, &width, &height,
	    &bestStripHeight, &compression, &colorType,
	    palette, &primImgHandle));
	XERR_GEN_CLEANUP(warning && warning < IO_WARNINGS, warning);

	/* The annotation data is handled specially. Version 2.x stores the
	 * annotation data as an array of offsets, then stores the annotation
	 * characteristics in a header in front of the data at each of those
	 * offsets. During the TIFF field parsing, the annotation count was
	 * obtained directly from the IFD itself, if the version is 2.x.
	 * Otherwise, the count is contained within the annotation table header.
	 * Let's read it now, so we can pass that count back.
	 */
	XERR_PASS_CLEANUP(wp_TfxGetAnnotTableHeader(tfxPage));
	tfxPage->pageInfo.wAnnotationCount = tfxPage->annotTableHeader.annotCount;

	tfxPage->seqStatus |= TFX_PAGE_SEQ_GETINFO;
    } /* Info not already gotten */

    /* Copy out the page information
     */
    /* Make a copy of the real layer image counts, so that the client cannot
     * muck with the actual list */
    if (tfxPage->extImageCounts) {
	TFX_FREE(tfxPage->extImageCounts);
	tfxPage->extImageCounts = NULL;
    }

    /* Copy the image counts, for each layer, to the output */
    XERR_GEN_CLEANUP(tfxPage->pageInfo.dwLayersPresent == 0, IO_INTERNAL_ERROR);
    IO_CHECKMALLOC(tfxPage->extImageCounts = 
	(UInt32 *)TFX_MALLOC(sizeof(UInt32) * tfxPage->pageInfo.dwLayersPresent));
    memcpy(tfxPage->extImageCounts, tfxPage->pageInfo.pdwImageCounts,
	sizeof(UInt32) * tfxPage->pageInfo.dwLayersPresent);
    if (ppdwImageCounts) {
	*ppdwImageCounts = tfxPage->extImageCounts;
    }
    *pdwLayersPresent = tfxPage->pageInfo.dwLayersPresent;
    *pwAnnotationCount = tfxPage->pageInfo.wAnnotationCount;
    *pTfxPageHandle = tfxPage->pageInfo.tfxPageHandle;

    if (__error__ == 0) {
	__error__ = warning;	/* If a warning exists, use if no error */
    }
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image description, including MRC image counts; indicate support.
 *
 * LongDescription:::
 *  TFX_GetImageInfo() gets the description of an image: position relative
 *  to the Primary image, resolution, size, stored strip size, stored
 *  compression type, color type, palette (if any).
 *  <BR>
 *  Returned is an image handle to be used to reference this image in later
 *  calls. A call to TFX_CloseDocument() renders this image handle invalid.
 *  <BR>
 *  This function also serves as the mechanism to indicate current support.
 *  If the characteristics of this image are unsupported, then an error is
 *  returned: IO_NOSUPPORT.
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  dwImgLayer::		Image layer (1, 2, ...), 0 if we don't know
 *				what the primary IFD contains, and that's
 *				what we're asking for
 *  dwImgOffset::		Image offset for image within that layer (0,
 *				1, ...)
 *
 * Output:::
 *  pdwXOffset::		X location in page coordinates
 *  pdwYOffset::		Y location in page coordinates
 *  pdwXResolution::		X Resolution of image in dpm
 *  pdwYResolution::		Y Resolution of image in dpm
 *  pdwWidth::			Width in image's X resolution units
 *  pdwHeight::			Height in image's Y resolution units
 *  pdwBestStripHeight::	For best performance, pass strips with size
 *				and boundary that are multiples of this (strips
 *				are stored with this size)
 *  pxCompression::		Compression type for src/dst
 *  pxColorType::		Color type (see IO_COLORTYPE
 *  pPalette::			Color palette if IO_COLORTYPE is
 *				IO_COLORTYPE_COLOR4 or IO_COLORTYPE_COLOR8.
				Palette is RGB triples (chunky): RGBRGB...
 *  pTftfxImgHandle::		Returned image handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetImageInfo(
    IO_PAGEHANDLE tfxPageHandle,
    UInt32 dwImgLayer,
    UInt32 dwImgOffset,
    UInt32 *pdwXOffset,
    UInt32 *pdwYOffset,
    UInt32 *pdwXResolution,
    UInt32 *pdwYResolution,
    UInt32 *pdwWidth,
    UInt32 *pdwHeight,
    UInt32 *pdwBestStripHeight,
    IO_COMPRESSION *pxCompression,
    IO_COLORTYPE *pxColorType,
    UInt8 *pPalette,
    IO_IMGHANDLE *pTftfxImgHandle
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    TFX_IMG *tfxPrimImg, *tfxImg = NULL;
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_DOC *tfxDoc = NULL;
    TFX_IMGINFO *imgInfo;
    UInt32 layer;
    UInt32 *prevImageCounts = NULL;
    UInt32 prevLayersPresent;
    PROLOGUE;
    
    TFX_CheckHandle(tfxPageHandle);
    
    /* Check for proper calling sequence */
    
    /* The layer number must be accurate:
     *  > 1 and <= layers present
     * The image offset must be accurate:
     *  < # of images in layer
     */
    XERR_GEN_CLEANUP((dwImgLayer < 1 && tfxPage->pageInfo.dwLayersPresent > 0)
	|| (tfxPage->pageInfo.dwLayersPresent > 0 &&
	 (dwImgLayer > tfxPage->pageInfo.dwLayersPresent ||
	 tfxPage->pageInfo.pdwImageCounts == NULL ||
	 tfxPage->pageInfo.pdwImageCounts[dwImgLayer - 1] <= dwImgOffset)),
	IO_BADPARAMETER);
    
    tfxDoc = tfxPage->tfxDoc;
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);
    
    /* Clear the results */
    *pdwXOffset = *pdwYOffset = *pdwXResolution = *pdwYResolution = 0UL;
    *pdwWidth = *pdwHeight = *pdwBestStripHeight = 0UL;
    *pxCompression = IO_COMP_UNKNOWN;
    *pxColorType = IO_COLORTYPE_NONE;
    *pTftfxImgHandle = 0UL;
    if (pPalette) {
	memset(pPalette, 0, 256*3);
    }
    
    /* Haven't gotten the image info for the parent image yet, get now */
    tfxPrimImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;
    XERR_GEN_CLEANUP(tfxPrimImg == NULL, IO_SEQUENCE);
/*    XERR_GEN_CLEANUP(tfxPrimImg->imgInfo.dwImgLayer != 2, IO_INTERNAL_ERROR); */
    if ((tfxPrimImg->seqStatus & TFX_IMG_SEQ_GETINFO) == 0) {
	
	/* This really gets the info for an individual image */
	XERR_WARNOK_CLEANUP(wp_TfxGetImageDesc(tfxPrimImg));

    }

    /* Use the primary image layer, if it wasn't known */
    if (dwImgLayer == 0) {
	dwImgLayer = tfxPrimImg->imgInfo.dwImgLayer;
    }

    /* Use the primary image, if requested */
    if (dwImgLayer == tfxPrimImg->imgInfo.dwImgLayer && dwImgOffset == 0) {
	    
	/* We're already referencing the primary image */
	tfxImg = tfxPrimImg;
    } else {
	if (tfxPage->tfxImgs == NULL) {

	    /* If we have previously used the MRContent TIFF field to
	     * explain the page content, let's clear it now, and let
	     * the true content guide the values. */
	    prevLayersPresent = tfxPage->pageInfo.dwLayersPresent;
	    prevImageCounts = tfxPage->pageInfo.pdwImageCounts;
	    tfxPage->pageInfo.pdwImageCounts = NULL;
	    tfxPage->pageInfo.dwLayersPresent = 0;

	    /* Initialize the list with the Primary image, which already
	     * exists */
	    XERR_PASS_CLEANUP(wp_TfxIncrementImageCounts(tfxPage,
		tfxPrimImg->imgInfo.dwImgLayer, 1));

	    /* Now read the actual SubIFDs list */
	    XERR_PASS_CLEANUP(wp_TfxTraverseSubImages(tfxPrimImg));

	    /* Compare to what we had. This is really a sanity check; we could
	     * simply return, and hope the client requests the information
	     * again after the first image is read, but that's too obscure. */
	    XERR_GEN_CLEANUP(tfxPage->pageInfo.dwLayersPresent != prevLayersPresent,
		IO_BADFILEFORMAT);
	    for (layer = 0; layer < prevLayersPresent; layer++) {
		XERR_GEN_CLEANUP(tfxPage->pageInfo.pdwImageCounts[layer] !=
		    prevImageCounts[layer], IO_BADFILEFORMAT);
	    }

	    TFX_FREE(prevImageCounts);
	    prevImageCounts = NULL;

	}

	/* Search for the image layer and order (linked list is maintained
         * in sequential order).*/
	tfxImg = tfxPage->tfxImgs;
	while (tfxImg->imgInfo.dwImgLayer < dwImgLayer ||
	    (tfxImg->imgInfo.dwImgLayer == dwImgLayer &&
	    tfxImg->imgInfo.dwImgOffset < dwImgOffset)) {

	    /* The image with this offset doesn't exist */
	    XERR_GEN_CLEANUP(((tfxImg = tfxImg->next) == NULL), IO_INTERNAL_ERROR);
	}
    }

    /* Check to see that the requested image is actually found */
    XERR_GEN_CLEANUP((tfxImg == NULL), IO_BADPARAMETER);
    switch (tfxImg->tfxPage->tfxDoc->docInfo.profile) {
    case TFX_PROFILE_M:
    case TFX_PROFILE_M1:
    case TFX_PROFILE_COH_M:
    case TFX_PROFILE_COH_M1:
	XERR_GEN_CLEANUP(tfxImg->imgInfo.dwImgLayer != dwImgLayer ||
	    tfxImg->imgInfo.dwImgOffset != dwImgOffset, IO_BADFILEFORMAT);
    }
    
    /* Now get the info for that image */
    XERR_WARNOK_CLEANUP(wp_TfxGetImageDesc(tfxImg));
    imgInfo = &tfxImg->imgInfo;

    /* If variable-sized strips, keep them variable */
    if (imgInfo->dwBestStripHeight == 0) {
	XERR_IGNORE(TFX_UseVariableStripHeights(tfxImg->handleID));
    }
    
    /* Doublecheck most of the information, to make sure it is valid */
    {
	UInt32 tmpLayer = imgInfo->dwImgLayer;
	UInt32 tmpOffset = imgInfo->dwImgOffset;
	UInt32 tmpHeight = imgInfo->dwHeight;
	UInt32 tmpStripHeight = imgInfo->dwBestStripHeight;
	IO_COMPRESSION tmpCompression = imgInfo->xCompression;

	XERR_IGNORE(__error__ = TFX_CheckImageInfo((IO_PAGEHANDLE)tfxImg->tfxPage,
	    &tmpLayer, &tmpOffset,
	    &imgInfo->xColorType,
	    &imgInfo->dwXOffset, &imgInfo->dwYOffset,
	    &imgInfo->dwXResolution, &imgInfo->dwYResolution,
	    &imgInfo->dwWidth, &tmpHeight,
	    &tmpStripHeight, imgInfo->pPalette,
	    &tmpCompression, &imgInfo->xColorSpace));

	/*  -- NOTE -- NOTE -- NOTE -- NOTE -- NOTE -- NOTE -- NOTE -- */
	/* It's questionable whether the width/height to the edge of the page
	 * should be propogated to the caller.  If this causes problems,
	 * I guess we'll have to take it out: */
	imgInfo->dwHeight = tmpHeight;
    }

    /* If we're using variable-sized strips, reset best strip height */
    if (tfxImg->seqStatus & TFX_IMG_SEQ_USEVSS) {
	imgInfo->dwBestStripHeight = 0;
    }

    /* Let's make a last-minute check to see if the parameters
     * are consistent with the page */
    if (tfxImg != tfxPrimImg && tfxPrimImg != NULL) {
	XERR_GEN_CLEANUP(imgInfo->dwXOffset > tfxPrimImg->imgInfo.dwWidth,
	    IO_BADFILEFORMAT);
	XERR_GEN_CLEANUP(imgInfo->dwYOffset > tfxPrimImg->imgInfo.dwHeight,
	    IO_BADFILEFORMAT);
#if 0
	/* HACK: since we know what check we're about to make below, let's
	 * adjust the width and/or height to go up to the edge of the page,
	 * instead of erroring */
	/* Hack undone, see -- NOTE -- above */
	if (imgInfo->dwXOffset + (imgInfo->dwWidth - 1) *
	    tfxPrimImg->imgInfo.dwXResolution / imgInfo->dwXResolution >=
	    tfxPrimImg->imgInfo.dwWidth) {
	    imgInfo->dwWidth = ((tfxPrimImg->imgInfo.dwWidth - imgInfo->dwXOffset)
		+ (tfxPrimImg->imgInfo.dwXResolution / imgInfo->dwXResolution - 1))
		/ (tfxPrimImg->imgInfo.dwXResolution / imgInfo->dwXResolution);
	}
	if (imgInfo->dwYOffset + (imgInfo->dwHeight - 1) *
	    tfxPrimImg->imgInfo.dwYResolution / imgInfo->dwYResolution >=
	    tfxPrimImg->imgInfo.dwHeight) {
	    imgInfo->dwHeight = ((tfxPrimImg->imgInfo.dwHeight - imgInfo->dwYOffset)
		+ (tfxPrimImg->imgInfo.dwYResolution / imgInfo->dwYResolution - 1))
		/ (tfxPrimImg->imgInfo.dwYResolution / imgInfo->dwYResolution);
	}
#endif

	/* Note: given the resolution differences, the last subimage pixel
	 * may overhang the edge of the page, meaning the next-to-last subimage
	 * pixel must not quite go up to the edge. That check made here: */
	XERR_GEN_CLEANUP(imgInfo->dwXOffset + (imgInfo->dwWidth - 1) *
	    tfxPrimImg->imgInfo.dwXResolution / imgInfo->dwXResolution >=
	    tfxPrimImg->imgInfo.dwWidth, IO_BADFILEFORMAT);
	XERR_GEN_CLEANUP(imgInfo->dwYOffset + (imgInfo->dwHeight - 1) *
	    tfxPrimImg->imgInfo.dwYResolution / imgInfo->dwYResolution >=
	    tfxPrimImg->imgInfo.dwHeight, IO_BADFILEFORMAT);
    }

    /* Check other consistencies */
    if (imgInfo->dwXResolution == 0) {
	imgInfo->dwXResolution = 300;
    }
    if (imgInfo->dwYResolution == 0) {
	imgInfo->dwYResolution = 300;
    }
    if (imgInfo->dwBestStripHeight == 0 && tfxImg->fields.StripRowCounts == NULL) {
	imgInfo->dwBestStripHeight = imgInfo->dwHeight;
    }

    /* File must be corrupt if not at least reasonable */
    XERR_GEN_CLEANUP(imgInfo->dwWidth == 0, IO_BADFILEFORMAT);
    XERR_GEN_CLEANUP(imgInfo->dwHeight == 0, IO_BADFILEFORMAT);

    /* If we can't handle the compression, fail now.
     * Note, during the original read of the IFD, if the compression
     * isn't supported, the image is pruned from the list. */
    XERR_GEN_CLEANUP(imgInfo->xCompression == IO_COMP_UNKNOWN,
	IO_BADFILEFORMAT);

    /* Check parameters for what we expect to be currently supported
     * (Note: we can't catch everything, but this has come to bite us
     * already as color tokenization was unanticipated.)
     */
    switch (imgInfo->xColorType) {
    case IO_COLORTYPE_BINARY:
	switch (imgInfo->xCompression) {
	case IO_COMP_TOKENIZATION:
	case IO_COMP_LLTOKENIZATION:
	case IO_COMP_JBIG2:
	case IO_COMP_LLJBIG2:
	case IO_COMP_MH:
	case IO_COMP_MR:
	case IO_COMP_MMR:
	case IO_COMP_JBIG:
	case IO_COMP_NOC:
	case IO_COMP_OTHER:
	    break;
	default:
	    __error__ = IO_WARN_NOTCOMPLIANT;
	}
	break;
    case IO_COLORTYPE_GRAY4:
	switch (imgInfo->xCompression) {
	case IO_COMP_JPEG:
	case IO_COMP_NOC:
	case IO_COMP_OTHER:
	    break;
	default:
	    __error__ = IO_WARN_NOTCOMPLIANT;
	}
	break;
    case IO_COLORTYPE_GRAY8:
	switch (imgInfo->xCompression) {
	case IO_COMP_JPEG:
	case IO_COMP_NOC:
	case IO_COMP_OTHER:
	    break;
	default:
	    __error__ = IO_WARN_NOTCOMPLIANT;
	}
	break;
    case IO_COLORTYPE_COLOR4:
    case IO_COLORTYPE_COLOR8:
	switch (imgInfo->xCompression) {
	case IO_COMP_JPEG:
	case IO_COMP_NOC:
	case IO_COMP_OTHER:
	    break;
	default:
	    __error__ = IO_WARN_NOTCOMPLIANT;
	}
	break;
    case IO_COLORTYPE_COLOR24:
	switch (imgInfo->xCompression) {
	case IO_COMP_JPEG:
	case IO_COMP_JBIG2:
	case IO_COMP_OTHER:
	    break;
	case IO_COMP_NOC:
	    switch (tfxImg->tfxPage->tfxDoc->docInfo.profile) {
	    case TFX_PROFILE_COH_C:
	    case TFX_PROFILE_COH_M:
	    case TFX_PROFILE_COH_M1:
		break;
	    default:
		__error__ = IO_WARN_NOTCOMPLIANT;
	    }
	    break;
	default:
	    __error__ = IO_WARN_NOTCOMPLIANT;
	}
	break;
    default:
	__error__ = IO_WARN_NOSUPPORT;
    }
    
    /* Copy out the image information */
    *pdwXOffset = imgInfo->dwXOffset;
    *pdwYOffset = imgInfo->dwYOffset;
    *pdwXResolution = imgInfo->dwXResolution;
    *pdwYResolution = imgInfo->dwYResolution;
    *pdwWidth = imgInfo->dwWidth;
    *pdwHeight = imgInfo->dwHeight;
    *pdwBestStripHeight = imgInfo->dwBestStripHeight;
    *pxCompression = imgInfo->xCompression;
    *pxColorType = imgInfo->xColorType;
    if (imgInfo->pPalette && pPalette) {
	memcpy(pPalette, imgInfo->pPalette, 256*3);
    }
    *pTftfxImgHandle = (IO_IMGHANDLE)tfxImg;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (prevImageCounts) {
	TFX_FREE(prevImageCounts);
	prevImageCounts = NULL;
    }
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get some image stats (compressed size, token dictionary ID/size)
 *
 * LongDescription:::
 *  TFX_GetImageStats() accumulates the total compressed image size, as
 *  stored in the file, and if appropriate, the associated token dictionary
 *  ID and size.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *
 * Output:::
 *  pdwCmpSize::		Total size of the image, compressed
 *  pdwDictCount::		If tokenized, the number of token dictionaries
 *  ppdwDictIDs::		If tokenized, the dictionary ID list
 *  ppdwDictSizes::		If tokenized, the sizes of the token
 *				dictionaries
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetImageStats(
    IO_IMGHANDLE tfxImgHandle,
    UInt32 *pdwCmpSize,
    UInt32 *pdwDictCount,
    UInt32 **ppdwDictIDs,
    Int32 **ppdwDictSizes
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_DOC *tfxDoc = NULL;
    UInt8 *pCmpData = NULL;
    Int32 filePos, dictLocation;
    UInt32 cmpBytes, cmpTotal = 0;
    UInt16 dictNum, dictCount;
    UInt32 *dictIDs = NULL;
    Int32 *dictSizes = NULL;
    UInt8 *tundraData;
    UInt32 tundraBytes;
    UInt32 row, stripHeight;
    UInt16 compression;
    UInt16 lastPageUsed;
    UInt32 decoderMemory;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImgHandle);
    
    /* Check for proper calling sequence */
    
    tfxDoc = tfxImg->tfxPage->tfxDoc;
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);
    
    /* Clear the results */
    *pdwCmpSize = *pdwDictCount = 0UL;
    *ppdwDictIDs = NULL;
    *ppdwDictSizes = NULL;

    if (tfxImg->imgInfo.xCompression == IO_COMP_TOKENIZATION
	|| tfxImg->imgInfo.xCompression == IO_COMP_LLTOKENIZATION
	|| tfxImg->imgInfo.xCompression == IO_COMP_JBIG2
	|| tfxImg->imgInfo.xCompression == IO_COMP_LLJBIG2) {

        /* Get the token dictionary information... */

	/* Read enough of the token stream to extract the dictionary ID */
	XERR_GEN_CLEANUP(tfxImg->fields.StripsPerImage < 1, IO_INTERNAL_ERROR);
	filePos = tfxImg->fields.StripOffsets[0];
	cmpBytes = tfxImg->fields.StripByteCounts[0];

	/* We could try to avoid reading the entire data stream in, but we'd
	 * have to either interpret a portion of the data first (which goes
	 * against having the interpretation all in one place), or make some
	 * assumptions about the size of the header. I did the later and got
	 * bit. Since it's compressed, just allocate and read the whole thing.
	 */
	XERR_GEN_CLEANUP(cmpBytes == 0, IO_BADFILEFORMAT);
	IO_CHECKMALLOC(pCmpData = TFX_MALLOC(cmpBytes));
	TFX_SEEK(tfxFile, filePos);
	TFX_READ(tfxFile, pCmpData, (Int32)cmpBytes);

	/* Get the dictionary ID from the compressed data stream */
	XERR_PASS_CLEANUP(TFX_GetDictListFromPosBlock(tfxImgHandle, pCmpData,
	    cmpBytes, &dictCount, &dictIDs, &tundraData, &tundraBytes));
	*pdwDictCount = dictCount;

	/* Allocate an array for the dictionary sizes */
	if (dictCount) {
	    IO_CHECKMALLOC(dictSizes = TFX_MALLOC(dictCount * sizeof(UInt32)));
	}

	/* From the IDs, get the file location and byte count */
	for (dictNum = 0; dictNum < dictCount; dictNum++) {
	    XERR_PASS_CLEANUP(wp_TfxGetResourceOffset(tfxImg->tfxPage->tfxDoc,
		dictIDs[dictNum], &dictLocation, &dictSizes[dictNum],
		&compression, &lastPageUsed, &decoderMemory));
	}
    }

    /* Loop through each strip, and request the compressed data size */
    XERR_GEN_CLEANUP(tfxImg->imgInfo.dwBestStripHeight < 1, IO_INTERNAL_ERROR);
    stripHeight = tfxImg->imgInfo.dwBestStripHeight;
    for (row = 0; row < tfxImg->imgInfo.dwHeight; row += stripHeight) {
	if (row + stripHeight > tfxImg->imgInfo.dwHeight) {
	    stripHeight = tfxImg->imgInfo.dwHeight - row;
	}
	XERR_PASS_CLEANUP(TFX_GetCompressedStripSize(tfxImgHandle, row,
						     &cmpBytes));
	cmpTotal += cmpBytes;
    }

    /* Preserve the arrays in structures that will persist until the image
     * is freed */
    if (tfxImg->dictList) {
	TFX_FREE(tfxImg->dictList);
    }
    tfxImg->dictList = dictIDs;
    dictIDs = NULL;
    if (tfxImg->dictSizes) {
	TFX_FREE(tfxImg->dictSizes);
    }
    tfxImg->dictSizes = dictSizes;
    dictSizes = NULL;
    if (pCmpData) {
	TFX_FREE(pCmpData);
	pCmpData = NULL;
    }

    /* Copy out the image information */
    *pdwCmpSize = cmpTotal;
    *ppdwDictIDs = tfxImg->dictList;
    *ppdwDictSizes = tfxImg->dictSizes;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (pCmpData) {
	TFX_FREE(pCmpData);
    }
    if (dictIDs) {
	TFX_FREE(dictIDs);
    }
    if (dictSizes) {
	TFX_FREE(dictSizes);
    }
    
    ERR_RETURN;
}


#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image information for a transparency mask.
 *
 * LongDescription:::
 *  TFX_GetMaskImageInfo() returns the same information about an image as does
 *  TFX_GetImageInfo() (see above), but does so for a transparency mask
 *  attached to a non-binary image.
 *
 * Input:::
 *  tfxParentImgHandle::		Parent image handle
 *  dwImgOffset::		Image offset (should always be 0 for now)
 *
 * Output:::
 *  pdwXOffset::		X location in page coordinates
 *  pdwYOffset::		Y location in page coordinates
 *  pdwXResolution::		X Resolution of image in dpm
 *  pdwYResolution::		Y Resolution of image in dpm
 *  pdwWidth::			Width in image's X resolution units
 *  pdwHeight::			Height in image's Y resolution units
 *  pdwBestStripHeight::	For best performance, pass strips with size
 *				and boundary that are multiples of this (strips
 *				are stored with this size)
 *  pxCompression::		Compression type for src/dst
 *  pTfxImgHandle::		Returned image handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TFX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetMaskImageInfo(
	IO_IMGHANDLE tfxParentImgHandle,
	UInt32 dwImgOffset,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	IO_COMPRESSION *pxCompression,
	IO_IMGHANDLE *pTfxImgHandle
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_IMG *tfxParentImg = (TFX_IMG *)tfxParentImgHandle;
    TFX_IMG *tfxImg = NULL;
    TFX_IMGINFO *imgInfo;
    PROLOGUE;
    
    /* Clear the results */
    *pdwXOffset = *pdwYOffset = *pdwXResolution = *pdwYResolution = 0UL;
    *pdwWidth = *pdwHeight = *pdwBestStripHeight = 0UL;
    *pxCompression = IO_COMP_UNKNOWN;
    *pTfxImgHandle = 0UL;
    
    TFX_CheckHandle(tfxParentImgHandle);
    
    /* Check for proper calling sequence */


    instData = TFX_IMGINST(tfxParentImg);
    tfxFile = TFX_IMGFILE(tfxParentImg);
    
    /* Haven't gotten the image info for the parent image yet, get now */
    if ((tfxParentImg->seqStatus & TFX_IMG_SEQ_GETINFO) == 0) {
	
	/* This really gets the info for an individual image */
	XERR_PASS_CLEANUP(wp_TfxGetImageDesc(tfxParentImg));
    }
    
    /* Search for the first image of the type requested */
    XERR_GEN_CLEANUP((! tfxParentImg->imgInfo.dwMaskImages),
		     IO_BADPARAMETER);
    if (tfxParentImg->xMaskImg == NULL) {
	XERR_PASS_CLEANUP(wp_TfxTraverseSubImages(tfxParentImg));
    }
    tfxImg = tfxParentImg->xMaskImg;

    /* Check to see that the requested image is actually found */
    XERR_GEN_CLEANUP((tfxImg == NULL), IO_BADPARAMETER);
    
    /* Search for the Nth image of that type */
    while (dwImgOffset > tfxImg->imgInfo.dwImgOffset) {

	/* The image with this offset doesn't exist */
	XERR_GEN_CLEANUP(((tfxImg = tfxImg->next) == NULL),
		 IO_BADPARAMETER);
    }
    
    /* Now get the info for that image */
    XERR_PASS_CLEANUP(wp_TfxGetImageDesc(tfxImg));
    
    /* Copy out the image information */
    imgInfo = &tfxImg->imgInfo;
    *pdwXOffset = imgInfo->dwXOffset;
    *pdwYOffset = imgInfo->dwYOffset;
    *pdwXResolution = imgInfo->dwXResolution;
    *pdwYResolution = imgInfo->dwYResolution;
    *pdwWidth = imgInfo->dwWidth;
    *pdwHeight = imgInfo->dwHeight;
    *pdwBestStripHeight = imgInfo->dwBestStripHeight;
    *pxCompression = imgInfo->xCompression;
    *pTfxImgHandle = (IO_IMGHANDLE)tfxImg;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image information for a rendering hints image.
 *
 * LongDescription:::
 *  TFX_GetHintsImageInfo() returns the same information about an image as does
 *  TFX_GetImageInfo() (see above), but does so for a rendering hints image,
 *  attached to an image.
 *
 * Input:::
 *  tfxParentImgHandle::	Parent image handle
 *  dwImgOffset::		Image offset (should always be 0 for now)
 *
 * Output:::
 *  pdwXOffset::		X location in page coordinates
 *  pdwYOffset::		Y location in page coordinates
 *  pdwXResolution::		X Resolution of image in dpm
 *  pdwYResolution::		Y Resolution of image in dpm
 *  pdwWidth::			Width in image's X resolution units
 *  pdwHeight::			Height in image's Y resolution units
 *  pdwBestStripHeight::	For best performance, pass strips with size
 *				and boundary that are multiples of this (strips
 *				are stored with this size)
 *  pxCompression::		Compression type for src
 *  pTfxImgHandle::		Returned image handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TFX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetHintsImageInfo(
	IO_IMGHANDLE tfxParentImgHandle,
	UInt32 dwImgOffset,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	IO_COMPRESSION *pxCompression,
	IO_IMGHANDLE *pTfxImgHandle
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_IMG *tfxParentImg = (TFX_IMG *)tfxParentImgHandle;
    TFX_IMG *tfxImg = NULL;
    TFX_IMGINFO *imgInfo;
    PROLOGUE;
    
    /* Clear the results */
    *pdwXOffset = *pdwYOffset = *pdwXResolution = *pdwYResolution = 0UL;
    *pdwWidth = *pdwHeight = *pdwBestStripHeight = 0UL;
    *pxCompression = IO_COMP_UNKNOWN;
    *pTfxImgHandle = 0UL;
    
    TFX_CheckHandle(tfxParentImgHandle);
    
    /* Check for proper calling sequence */


    instData = TFX_IMGINST(tfxParentImg);
    tfxFile = TFX_IMGFILE(tfxParentImg);
    
    /* Haven't gotten the image info for the parent image yet, get now */
    if ((tfxParentImg->seqStatus & TFX_IMG_SEQ_GETINFO) == 0) {
	
	/* This really gets the info for an individual image */
	XERR_PASS_CLEANUP(wp_TfxGetImageDesc(tfxParentImg));
    }
    
    /* Search for the offset requested */
    XERR_GEN_CLEANUP((! tfxParentImg->imgInfo.dwHintsImages),
		     IO_BADPARAMETER);
    if (tfxParentImg->xHintsImgs == NULL) {
	XERR_PASS_CLEANUP(wp_TfxTraverseHintsImages(tfxParentImg));
    }
    tfxImg = tfxParentImg->xHintsImgs;
    while (tfxImg && tfxImg->imgInfo.dwImgOffset != dwImgOffset) {
	tfxImg = tfxImg->next;
    }

    /* Check to see that the requested image is actually found */
    XERR_GEN_CLEANUP((tfxImg == NULL), IO_BADPARAMETER);
    
    /* Search for the Nth image of that type */
    while (dwImgOffset > tfxImg->imgInfo.dwImgOffset) {

	/* The image with this offset doesn't exist */
	XERR_GEN_CLEANUP(((tfxImg = tfxImg->next) == NULL),
		 IO_BADPARAMETER);
    }
    
    /* Now get the info for that image */
    XERR_PASS_CLEANUP(wp_TfxGetImageDesc(tfxImg));
    
    /* Copy out the image information */
    imgInfo = &tfxImg->imgInfo;
    *pdwXOffset = imgInfo->dwXOffset;
    *pdwYOffset = imgInfo->dwYOffset;
    *pdwXResolution = imgInfo->dwXResolution;
    *pdwYResolution = imgInfo->dwYResolution;
    *pdwWidth = imgInfo->dwWidth;
    *pdwHeight = imgInfo->dwHeight;
    *pdwBestStripHeight = imgInfo->dwBestStripHeight;
    *pxCompression = imgInfo->xCompression;
    *pTfxImgHandle = (IO_IMGHANDLE)tfxImg;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}



/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read enough information to determine the profile of the TIFF-FX file
 *  (profile F, C, M, T, ...)
 *
 * LongDescription:::
 *  TFX_GetDocProfile() looks through the document content and makes a
 *  determination as to which of the TIFF-FX profiles it adheres to, then
 *  returns that as a file type.
 *
 * Input:::
 *  tfxDocHandle::		Input document handle
 *
 * Output:::
 *  pdwProfile::		TIFF-FX profile (see iomrc.h DOCTYPE_...)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetDocProfile(
	IO_DOCHANDLE tfxDocHandle,
	DOCTYPE *pdwProfile
)
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    TFX_PAGE *tfxPage;
    TFX_IMG *tfxImg;
    IO_PAGEHANDLE firstPageHandle;
    IO_IMGHANDLE firstPrimaryImageHandle;
    Bool isCoherentTIFF = 0;
    UInt8 version;
    UInt8 revision;
    UInt16 nPages;
    UInt32 *layerImages;
    UInt32 layerCount;
    UInt16 annotCount;
    UInt32 layer, offset;
    UInt32 xOffset, yOffset;
    UInt32 xRes, yRes;
    UInt32 width, height;
    UInt32 bestStripHeight;
    IO_COMPRESSION compression;
    IO_COLORTYPE colorType;
    UInt8 palette[768];
    ErrCode warning = 0;
    PROLOGUE;
    
    TFX_CheckHandle(tfxDocHandle);
    
    /* Check for proper calling sequence */
    
    /* Clear the inputs */
    *pdwProfile = DOCTYPE_TIFFFX_F;	/* Assume most fundamental type */

    /* Get enough information to resolve the first page, and it's primary IFD */
    XERR_IGNORE(warning = TFX_GetDocInfo(tfxDocHandle, &version, &revision,
	&nPages));
    XERR_GEN_CLEANUP(warning && warning < IO_WARNINGS, warning);
    if (nPages < 1) {
	*pdwProfile = DOCTYPE_UNKNOWN;
	return IO_NOERROR;
    }
    XERR_IGNORE(warning = TFX_GetPageInfo(tfxDocHandle, 0, &layerImages,
	&layerCount, &annotCount, &firstPageHandle));
    XERR_GEN_CLEANUP(warning && warning < IO_WARNINGS, warning);

    tfxPage = (TFX_PAGE *)firstPageHandle;
    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    /* 1. Read the reported profile from the FaxProfile field in the GPIFD.
     * The following values should be found:
     * 0 - does not conform to a profile defined for TIFF for facsimile 
     * 1 - minimal black & white lossless, Profile S 
     * 2 - extended black & white lossless, Profile F 
     * 3 - lossless JBIG black & white, Profile J 
     * 4 - lossy color and grayscale, Profile C 
     * 5 - lossless color and grayscale, Profile L 
     * 6 - Mixed Raster Content, Profile M 
     * 7 - lossy JBIG2 black & white, Profile T
     * 255 - JBIG2 & Mixed Raster Content, Profile M1 (E1.x)
     */
    if (tfxImg->fields.found.FaxProfile) {
	/* Map FaxProfile to a profile */
	switch (tfxImg->fields.gpIFD.FaxProfile) {
	case 0:
	    /* Continue on with our investigation.
	     * Assume we'll hit a Coherent TIFF flavor.
	     */
	    break;
	case 1:
	    *pdwProfile = DOCTYPE_TIFFFX_S;
	    break;
	case 2:
	    *pdwProfile = DOCTYPE_TIFFFX_F;
	    break;
	case 3:
	    *pdwProfile = DOCTYPE_TIFFFX_J;
	    break;
	case 4:
	    *pdwProfile = DOCTYPE_TIFFFX_C;
	    break;
	case 5:
	    *pdwProfile = DOCTYPE_TIFFFX_L;
	    break;
	case 6:
	    *pdwProfile = DOCTYPE_TIFFFX_M;
	    break;
	case 7:
	    *pdwProfile = DOCTYPE_TIFFFX_T;
	    break;
	case 255:
	    *pdwProfile = DOCTYPE_TIFFFX_M1;
	    break;
	}
    }

    /* 2. Check for multiple layers, promote to MRC */
    while (1) {
	UInt32 numOccupiedLayers = 0;

	/* Is there more than one layer? */
	if (layerCount > 1) {
	    for (layer = 0; layer < layerCount; layer++) {
		if (layerImages[layer] > 0) {
		    numOccupiedLayers++;
		    /* Multiple image on a layer forces MRC also */
		    if (layerImages[layer] > 1) {
			switch (*pdwProfile) {
			    /* Don't demote these */
			case DOCTYPE_TIFFFX_M1:
			case DOCTYPE_COHTIFF_M1:
			    break;
			default:
			    *pdwProfile = DOCTYPE_TIFFFX_M;
			}
			break;
		    }
		}
	    }
	    if (numOccupiedLayers > 1) {
		switch (*pdwProfile) {
    		    /* Don't demote these */
		case DOCTYPE_TIFFFX_M1:
		case DOCTYPE_COHTIFF_M1:
		    break;
		default:
		    *pdwProfile = DOCTYPE_TIFFFX_M;
		}
		break;
	    }
	}
	break;
    }

    /* Determine the layer where the Primary IFD is attached */
    layer = tfxImg->imgInfo.dwImgLayer;
    offset = tfxImg->imgInfo.dwImgOffset;

    /* Read the primary IFD */
    XERR_IGNORE(warning = TFX_GetImageInfo(firstPageHandle, layer, offset,
	&xOffset, &yOffset, &xRes, &yRes, &width, &height,
	&bestStripHeight, &compression, &colorType,
	palette, &firstPrimaryImageHandle));
    XERR_GEN_CLEANUP(warning && warning < IO_WARNINGS, warning);

    /* 3. Check if GPIFD contains shared resources, promote to Tokenization */
    if (tfxImg->tfxPage->tfxDoc->firstResourceTableLocation > 0) {
	switch (*pdwProfile) {
	case DOCTYPE_TIFFFX_S:
	case DOCTYPE_TIFFFX_F:
	    *pdwProfile = DOCTYPE_TIFFFX_T;
	    break;
	case DOCTYPE_TIFFFX_C:
	case DOCTYPE_TIFFFX_M:
	    *pdwProfile = DOCTYPE_TIFFFX_M1;
	    break;
	case DOCTYPE_COHTIFF_M:
	    *pdwProfile = DOCTYPE_COHTIFF_M1;
	    break;
	case DOCTYPE_TIFFFX_T:
	case DOCTYPE_TIFFFX_M1:
	    break;
	default:
	    XERR_GEN_CLEANUP(1, IO_INTERNAL_ERROR);
	}
    }

    /* 4. Look for color */
    switch (tfxImg->imgInfo.xColorType) {
    case IO_COLORTYPE_BINARY:
	break;
    default:
	switch (*pdwProfile) {
	case DOCTYPE_TIFFFX_S:
	case DOCTYPE_TIFFFX_F:
	    *pdwProfile = DOCTYPE_TIFFFX_C;
	    break;
	default:
	    break;
	}
    }

    /* 5. Look for Coherent TIFF specific items */
    {
	/* a. Unallowed resolutions */
	switch (xRes) {
	case 200:
	case 300:
	case 400:
	    break;
	default:
	    isCoherentTIFF = 1;
	}
    }

    /* Promote to Coherent TIFF if coherent TIFF behavior */
    if (isCoherentTIFF) {
	switch (*pdwProfile) {
	case DOCTYPE_TIFFFX_F:
	    *pdwProfile = DOCTYPE_COHTIFF_B;
	    break;
	case DOCTYPE_TIFFFX_C:
	    *pdwProfile = DOCTYPE_COHTIFF_C;
	    break;
	case DOCTYPE_TIFFFX_M:
	    *pdwProfile = DOCTYPE_COHTIFF_M;
	    break;
	case DOCTYPE_TIFFFX_M1:
	case DOCTYPE_TIFFFX_T:
	    *pdwProfile = DOCTYPE_COHTIFF_M1;
	    break;
	default:
	    XERR_GEN_CLEANUP(1, IO_INTERNAL_ERROR);
	}
    }

    /* Propogate the information to the document itself */
    switch (*pdwProfile) {
    case DOCTYPE_TIFFFX_S:		/* Profile S: standard B&W */
	tfxDoc->docInfo.profile = TFX_PROFILE_S;
	break;
    case DOCTYPE_TIFFFX_F:		/* Profile F: Fax B&W */
	tfxDoc->docInfo.profile = TFX_PROFILE_F;
	break;
    case DOCTYPE_TIFFFX_J:		/* Profile J: JBIG */
	tfxDoc->docInfo.profile = TFX_PROFILE_J;
	break;
    case DOCTYPE_TIFFFX_C:		/* Profile C: Contone */
	tfxDoc->docInfo.profile = TFX_PROFILE_C;
	break;
    case DOCTYPE_TIFFFX_L:		/* Profile L: Color JBIG */
	tfxDoc->docInfo.profile = TFX_PROFILE_L;
	break;
    case DOCTYPE_TIFFFX_M:		/* Profile M: MRC */
	tfxDoc->docInfo.profile = TFX_PROFILE_M;
	break;
    case DOCTYPE_TIFFFX_T:		/* Profile T: Tokenization (JBIG2) */
	tfxDoc->docInfo.profile = TFX_PROFILE_T;
	break;
    case DOCTYPE_TIFFFX_M1:		/* Profile M1: MRC + JBIG2 */
	tfxDoc->docInfo.profile = TFX_PROFILE_M1;
	break;
    case DOCTYPE_COHTIFF_B:		/* Coherent TIFF Profile B: Binary */
	tfxDoc->docInfo.profile = TFX_PROFILE_COH_B;
	break;
    case DOCTYPE_COHTIFF_C:		/* Coherent TIFF Profile C: Contone */
	tfxDoc->docInfo.profile = TFX_PROFILE_COH_C;
	break;
    case DOCTYPE_COHTIFF_M:		/* Coherent TIFF Profile M: MRC */
	tfxDoc->docInfo.profile = TFX_PROFILE_COH_M;
	break;
    case DOCTYPE_COHTIFF_M1:		/* Coherent TIFF Profile M1: MRC + JBIG2 */
	tfxDoc->docInfo.profile = TFX_PROFILE_COH_M1;
	break;
    default:
	XERR_GEN_CLEANUP(1, IO_INTERNAL_ERROR);
    }

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Specify the format of, and prepare to read image data.
 *
 * LongDescription:::
 *  TFX_ImageReadStart() prepares to read image data, and is used to specify
 *  the format of the data to be returned. These values are only stored for
 *  now, but used during later calls to TFX_ImageReadStrip().
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *  dwFlags::			See IO_IMGFLAGS_...
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_ImageReadStart(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwFlags
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_IMGINFO *pTftfxImgInfo;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImgHandle);
    pTftfxImgInfo = &tfxImg->imgInfo;
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxImg->seqStatus & TFX_IMG_SEQ_READSTATUS) !=
		      TFX_IMG_SEQ_READFINISH), IO_SEQUENCE);

#if !defined(TFX_READONLY) && !defined(NOIPCORE) && !defined(NODGP)
    /* If input file is in read/write mode, flush the page before we
     * can read it */
    if ((tfxImg->tfxPage->tfxDoc->readWrite & WRITE_ALLOWED) &&
	tfxImg->tfxPage->seqStatus & (TFX_PAGE_SEQ_TOKENSTART)) {
	XERR_PASS_CLEANUP(TFX_WriteTokenImages((IO_DOCHANDLE)tfxImg->tfxPage->tfxDoc));
    }
#endif /* TFX_READONLY && NOIPCORE */

    /* Store the parameters */
    tfxImg->imgFlags = dwFlags;
    
    /* Update the sequence status */
    tfxImg->seqStatus |= TFX_IMG_SEQ_READSTART;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}



/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get information about the image strip
 *
 * LongDescription:::
 *  wp_TfxGetStripInfo() retrieves the strip offset, strip start, and strip
 *  height, of a strip at a given row offset.
 *
 * Input:::
 *  tfxImg::			Valid image
 *  dwImageRowOffset::		The row offset in the image proper
 *
 * Output:::
 *  pdwStripOffset::		The offset of the strip in the collection
 *  pdwStripRow::		The offset of the first row in the strip
 *  pdwStripHeight::		The height of the strip
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
wp_TfxGetStripInfo(
		   TFX_IMG *tfxImg,
		   UInt32 rowOffset,
		   UInt32 *pdwStripOffset,
		   UInt32 *pdwStripRow,
		   UInt32 *pdwStripHeight
		   )
{
    TFX_STRIP *strip;
    UInt32 stripRow = 0;
    UInt32 stripOffset = 0;
    PROLOGUE;

    TFX_CheckHandle(tfxImg);

    /* Make sure we're being asked for something reasonable */
    if (rowOffset >= tfxImg->imgInfo.dwHeight) {
	rowOffset = tfxImg->imgInfo.dwHeight - 1;
    }

    /* If the strips are fixed in size, compute everything */
    if (tfxImg->imgInfo.dwBestStripHeight > 0) {
	*pdwStripOffset = rowOffset / tfxImg->imgInfo.dwBestStripHeight;
	*pdwStripRow = *pdwStripOffset * tfxImg->imgInfo.dwBestStripHeight;
	*pdwStripHeight = tfxImg->imgInfo.dwBestStripHeight;
    } else if (tfxImg->fields.StripsPerImage > 0) {
	/* The strips are variable in size, and are stored in the file */
	XERR_GEN(tfxImg->fields.StripRowCounts == NULL, IO_INTERNAL_ERROR);

	while (stripOffset < (tfxImg->fields.StripsPerImage - 1) &&
	    (stripRow + tfxImg->fields.StripRowCounts[stripOffset]) <= rowOffset) {
	    stripRow += tfxImg->fields.StripRowCounts[stripOffset];
	    stripOffset++;
	}

	/* If we don't have the strip yet, they the user has probably added stripe
	 * data out of order */
	XERR_GEN(rowOffset < stripRow, IO_SEQUENCE);

	/* Give them what they're looking for */
	*pdwStripOffset = stripOffset;
	*pdwStripRow = stripRow;
	*pdwStripHeight = tfxImg->fields.StripRowCounts[stripOffset];
    } else {
	/* We must be in the middle of generating strips for the primary IFD,
	 * but haven't collected them all to actually write them out */
	strip = tfxImg->strips.firstStrip;
	while (strip && (stripRow + strip->rows) <= rowOffset && strip->next) {
	    stripRow += strip->rows;
	    stripOffset++;
	}

	/* If we don't have the strip yet, they the user has probably added stripe
	 * data out of order */
	XERR_GEN(rowOffset < stripRow, IO_SEQUENCE);
	XERR_GEN(strip == NULL, IO_INTERNAL_ERROR);

	/* Give them what they're looking for */
	*pdwStripOffset = stripOffset;
	*pdwStripRow = stripRow;
	*pdwStripHeight = strip->rows;
    }

    /* Make sure we haven't gone past the end of the image */
    if (*pdwStripRow + *pdwStripHeight > tfxImg->imgInfo.dwHeight) {
	*pdwStripHeight = tfxImg->imgInfo.dwHeight - *pdwStripRow;
    }

    NORMAL_RETURN;
}



/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get information about the image stripe
 *
 * LongDescription:::
 *  TFX_GetStripeInfo() retrieves the stripe offset, stripe start, and stripe
 *  height, of a stripe at a given row offset.  Note that all dimension are
 *  in units of the page, or primary IFD.
 *
 * Input:::
 *  tfxImg::			Valid image
 *  dwImageRowOffset::		The row offset in the image proper
 *
 * Output:::
 *  pdwStripeOffset::		The offset of the stripe in the collection
 *  pdwStripeRow::		The offset of the first row in the stripe
 *  pdwStripeHeight::		The height of the stripe
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetStripeInfo(
		   IO_PAGEHANDLE ioPageHandle,
		   UInt32 rowOffset,
		   UInt32 *pdwStripeOffset,
		   UInt32 *pdwStripeRow,
		   UInt32 *pdwStripeHeight
		   )
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)ioPageHandle;
    TFX_IMG *tfxImg;
    PROLOGUE;

    TFX_CheckHandle(tfxPage);
    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    /* Get the information from the primary IFD */
    XERR_PASS(wp_TfxGetStripInfo(tfxImg, rowOffset, pdwStripeOffset,
	pdwStripeRow, pdwStripeHeight));

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the number of bytes of the compressed strip
 *
 * LongDescription:::
 *  TFX_GetCompressedStripSize() retrieves the size of the strip in compressed
 *  bytes.  It includes the JPEGTables, if appropriate, that would get
 *  automatically added in during TFX_CompressedReadStrip().
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *  dwImageRowOffset::		The row offset in the image proper
 *
 * Output:::
 *  pdwBytes::			The number of bytes for the compressed strip
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetCompressedStripSize(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwImageRowOffset,
	UInt32 *pdwBytes
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    UInt32 offset;
    UInt32 stripeOffset, stripeRow, stripeHeight;
    PROLOGUE;

    /* Default is to return nothing */
    *pdwBytes = 0L;

    /* Check the image handle */
    TFX_CheckHandle(tfxImgHandle);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((tfxImg->seqStatus & TFX_IMG_SEQ_GETINFO) == 0, IO_SEQUENCE);

    /* Get the strip offset */
    XERR_PASS_CLEANUP(wp_TfxGetStripInfo(tfxImg, dwImageRowOffset,
	&stripeOffset, &stripeRow, &stripeHeight));

    /* Strip must start where asked */
    XERR_GEN_CLEANUP(stripeRow != dwImageRowOffset, IO_BADPARAMETER);

    /* Get the size and offset of this strip */
    XERR_PASS_CLEANUP(wp_TfxGetCompressedSize(tfxImg, dwImageRowOffset,
	&offset, pdwBytes));

    /* If the strip requires the JPEGTables to be complete, add it in */
    if (tfxImg->qtableBufsize > 4) {
	switch (tfxImg->imgInfo.xCompression) {
	case IO_COMP_JPEG:
	    pdwBytes += tfxImg->qtableBufsize - 4;	/* Discount the SOI/EOI markers */
	    break;
	}
    }

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read an arbitrary strip in compressed form
 *
 * LongDescription:::
 *  TFX_CompressedReadStrip() returns a strip of compressed data from a
 *  stored strip.  The strip size must match that stored in the file, as
 *  indicated by the dwBestStripHeight parameter passed back from
 *  TFX_GetImageInfo().
 *  <BR>
 *  The JPEGTables, if appropriate, would automatically be inserted into the
 *  compressed data stream.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *  dwImgRowOffset::		Source row offset of the image proper
 *  dwStripHeight::		Number of rows to fill
 *
 * Output:::
 *  ppCmpBuffer::		Pointer to pointer to preallocated
 *				compressed data buffer (may be updated)
 *  pdwCmpBufsize::		Pointer to the number of bytes in buffer
 *  pdwBytesFilled::		Number of bytes successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data from this
 *  buffer as needed, while reading into the intermediate buffer as needed.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_CompressedReadStrip(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 **ppCmpBuffer,
	UInt32 *pdwCmpBufsize,
	UInt32 *pdwBytesFilled
)
{
    TFX_INSTANCE *instData;
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_IMGINFO *pTftfxImgInfo;
    UInt32 stripOffset, stripRow, stripHeight;
    PROLOGUE;

    /* Default to return nothing */
    *pdwBytesFilled = 0UL;
    
    TFX_CheckHandle(tfxImgHandle);
    
    instData = TFX_IMGINST(tfxImg);
    pTftfxImgInfo = &tfxImg->imgInfo;
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxImg->seqStatus & TFX_IMG_SEQ_READSTATUS) !=
		      TFX_IMG_SEQ_READSTART), IO_SEQUENCE);

    /* Get the strip offset */
    XERR_PASS_CLEANUP(wp_TfxGetStripInfo(tfxImg, dwImgRowOffset,
	&stripOffset, &stripRow, &stripHeight));

    /* Strip must start where asked */
    XERR_GEN_CLEANUP(stripRow != dwImgRowOffset, IO_BADPARAMETER);

    /* Strip must be the height asked */
    XERR_GEN_CLEANUP(stripHeight != dwStripHeight, IO_BADPARAMETER);

    /* Read the strip */
    XERR_PASS_CLEANUP(wp_TfxCompressedReadStrip(tfxImg, dwImgRowOffset,
	ppCmpBuffer, pdwCmpBufsize, pdwBytesFilled));


    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Interpret strip image data that contains the list of shared resources
 *
 * LongDescription:::
 *  TFX_GetDictListFromPosBlock() reads the header placed before the JBIG2
 *  position block. The dictionary count, dictionary ID(s), pointer to the
 *  JBIG2 position block, and JBIG2 position block byte count are returned.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle
 *  pCmpData::			Symbol compressed data stream
 *  dwCmpBytes::		Number of bytes in the data stream
 *
 * Output:::
 *  ppdwDictIDList::		Dictionary identifier list, TO BE FREED BY
 *				CLIENT
 *  pdwDictCount::		Number of dictionaries associated with the
 *				position block
 *  ppPositionBlock::		Pointer to the embedded JBIG2 position block
 *				(byte count of the header is implied by the
 *				delta from pCmpData)
 *  pdwPositionBytes::		Number of bytes for the embedded position
 *				block (note, with position block extensions,
 *				the byte count may not add up to the token
 *				stream byte count)
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetDictListFromPosBlock(
    IO_IMGHANDLE tfxImgHandle,
    UInt8 *pCmpData,
    UInt32 dwCmpBytes,
    UInt16 *pdwDictCount,
    UInt32 **ppdwDictIDList,
    UInt8 **ppPositionBlock,
    UInt32 *pdwPositionBytes
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    UInt8 *ptr, positionVersion;
    UInt16 dictNum, dictCount;
    UInt32 positionBytes;
    UInt32 dictID, *newIDs = NULL;
#define CHECK_FOR_PRE_SIZE_MOVE
#ifdef CHECK_FOR_PRE_SIZE_MOVE
    UInt32 offsetDictIDs;
#endif
    PROLOGUE;

    /* Return nothing initally */
    *pdwDictCount = 0;
    *ppdwDictIDList = NULL;
    *ppPositionBlock = pCmpData;
    *pdwPositionBytes = dwCmpBytes;

    /* Check parameters */
    TFX_CheckHandle(tfxImgHandle);
    tfxFile = TFX_IMGFILE(tfxImg);

    XERR_GEN(dwCmpBytes < 7, IO_BADPARAMETER);
    instData = tfxFile->instance;

    /* Start reading the header, also stored in tfxwpub.c */
    ptr = pCmpData;
    memcpy(&positionVersion, ptr + cTfxTCPosPosBlockVersion,
	   sizeof(positionVersion));
    memcpy(&dictCount, ptr + cTfxTCPosPosBlockDictCount, sizeof(dictCount));
    TFX_SWAPSHORT(tfxFile, &dictCount);

#ifdef CHECK_FOR_PRE_SIZE_MOVE
    /* If the old format is found, don't attempt to retrieve the positionBytes */
    if (cTfxTCPosPosBlockDictIDs + dictCount * sizeof(dictID) > dwCmpBytes) {
	positionBytes = dwCmpBytes + 1;
    } else {
#endif
    memcpy(&positionBytes, ptr + cTfxTCPosPosBlockSize(dictCount), sizeof(positionBytes));
    TFX_SWAPLONG(tfxFile, &positionBytes);
#ifdef CHECK_FOR_PRE_SIZE_MOVE
    }
#endif

    /* Make a sanity check on the dictionary count and pos block bytes.
     * Note: it used to be that the JBIG2 (un-embedded) position block size
     * preceded the resource list: dictionary count + dictionary IDs.
     * If we're smart, we can see if we're reading an old file, and automatically
     * recover from the out-date. */
#ifndef CHECK_FOR_PRE_SIZE_MOVE
    XERR_GEN_CLEANUP(cTfxTCPosPosBlockDictIDs + dictCount * sizeof(dictID) > dwCmpBytes,
	IO_BADFILEFORMAT);
#else
    if (cTfxTCPosPosBlockDictIDs + dictCount * sizeof(dictID) > dwCmpBytes
	|| (positionBytes != dwCmpBytes - cTfxTCPosBlockHeaderSize(dictCount)
	    && (tfxImg->fields.T88Options & T88OPTION_COLOURTAGS) ==0)) {
	UInt32 offsetPosBytes = cTfxTCPosPosBlockVersion + 1;
	UInt32 offsetDictCount = offsetPosBytes + 4;
	offsetDictIDs = offsetDictCount + 2;

	memcpy(&dictCount, ptr + offsetDictCount, sizeof(dictCount));
	TFX_SWAPSHORT(tfxFile, &dictCount);

	memcpy(&positionBytes, ptr + offsetPosBytes, sizeof(positionBytes));
	TFX_SWAPLONG(tfxFile, &positionBytes);

	/* Make the sanity check */
	XERR_GEN_CLEANUP(offsetDictIDs + dictCount * sizeof(dictID) > dwCmpBytes,
	    IO_BADFILEFORMAT);
	XERR_GEN_CLEANUP(positionBytes < dwCmpBytes - cTfxTCPosBlockHeaderSize(dictCount),
	    IO_BADFILEFORMAT);
	XERR_GEN_CLEANUP(positionBytes > dwCmpBytes, IO_BADFILEFORMAT);

    } else {
	offsetDictIDs = cTfxTCPosPosBlockDictIDs;
    }
#endif

    /* Return these values */
    *pdwDictCount = dictCount;
    *pdwPositionBytes = positionBytes;

    /* If no external token dictionaries, it/they may be embedded in
     * the position block */
    if (dictCount > 0) {

	/* Allocate a list of new dictionary IDs */
	IO_CHECKMALLOC(newIDs = TFX_MALLOC(dictCount * sizeof(UInt32)));

	/* Accumulate the IDs in a list */
	for (dictNum = 0; dictNum < dictCount; dictNum++) {
#ifdef CHECK_FOR_PRE_SIZE_MOVE
	    memcpy (&dictID, ptr + offsetDictIDs + dictNum * sizeof(dictID),
		sizeof(UInt32));
#else
	    memcpy (&dictID, ptr + cTfxTCPosPosBlockDictIDs + dictNum * sizeof(dictID),
		sizeof(UInt32));
#endif
	    TFX_SWAPLONG(tfxFile, &dictID);
	    newIDs[dictNum] = dictID;
	}
	
	/* Return the results */
	*ppdwDictIDList = newIDs;
	newIDs = NULL;
    }
    *ppPositionBlock = ptr + cTfxTCPosBlockHeaderSize(dictCount);

    NORMAL_RETURN;

ERR_CLEANUP:
    if (newIDs) {
	TFX_FREE(newIDs);
    }
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read a token dictionary
 *
 * LongDescription:::
 *  TFX_GetSymbolDict() reads the symbol dictionary into memory.
 *  <BR>
 *  If the dictionary contains extensions, then the number of bytes in the
 *  entire block will be larger than that of the back-end dictionary.
 *
 * Input:::
 *  tfxDocHandle::		Document handle
 *  dictID::			Dictionary identifier
 *
 * Output:::
 *  ppDictBlock::		Pointer to dictionary memory returned
 *				(client is responsible for free-ing it)
 *  pdwDictBytes::		Number of bytes in back-end dictionary
 *  pdwDictBlockBytes::		Number of bytes in whole block (may include
 *				extensions)
 *  pwLastPageUsed::		Pointer to the last page used (ordinal: 1, 2,
 *				...), 0 means unknown
 *  pdwDecoderMemory::		Pointer to the decoder memory required to hold
 *				the dictionary, 0 means unknown
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetSymbolDict(
	IO_DOCHANDLE tfxDocHandle,
	UInt32 dictID,
	UInt8 **ppDictBlock,
	UInt32 *pdwDictBytes,
	UInt32 *pdwDictBlockBytes,
	UInt16 *pwLastPageUsed,
	UInt32 *pdwDecoderMemory
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    TFX_FILE *tfxFile;
    Int32 dictBytes;
    UInt32 dictBlockBytes;
    UInt8 *dictHeader = NULL;
    UInt8 *dictBlock = NULL;
    Int32 offset;
    UInt8 dictVersion;
    UInt32 dictSize;
    UInt16 resourceType;    /* Should this be long? */
    UInt16 lastPage;
    UInt32 decMem;
    PROLOGUE;

    /* Return nothing, for now */
    *ppDictBlock = NULL;
    *pdwDictBytes = 0;
    *pwLastPageUsed = 0;
    *pdwDecoderMemory = 0UL;

    TFX_CheckHandle(tfxDoc);

    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);

    /* Get the offset and size of the resource with this ID */
    XERR_PASS_CLEANUP(wp_TfxGetResourceOffset(tfxDoc, dictID, &offset,
	&dictBytes, &resourceType, &lastPage, &decMem));

    /* Make sure it's the right resource type (compression) */
    XERR_GEN_CLEANUP(resourceType != TFX_RESOURCE_JBIG2, IO_BADFILEFORMAT);

    dictBlockBytes = dictBytes - cTfxTCDictHeaderSize;
    XERR_GEN_CLEANUP(dictBytes < (Int32)cTfxTCDictHeaderSize,
	IO_BADFILEFORMAT);

    /* Allocate a buffer for the JBIG2 dictionary header */
    IO_CHECKMALLOC(dictHeader = TFX_MALLOC(cTfxTCDictHeaderSize));

    /* Read the dictionary */
    TFX_SEEK(tfxFile, offset);
    TFX_READ(tfxFile, dictHeader, cTfxTCDictHeaderSize);

    /* Extract the header */
    memcpy(&dictVersion, dictHeader + cTfxTCPosDictVersion,
	   sizeof(dictVersion));    /* Version should be 3 */
    memcpy(&dictSize, dictHeader + cTfxTCPosDictSize, sizeof(dictSize));
    TFX_SWAPLONG(tfxFile, &dictSize);

    /* The JBIG2 dictionary block should be at most the rest of the block */
    XERR_GEN_CLEANUP(dictSize > dictBlockBytes, IO_BADFILEFORMAT);

    /* Check the version for support */
    XERR_GEN_CLEANUP(dictVersion > cTfxTCDictVersion, IO_NOSUPPORT);

    /* Free the header */
    TFX_FREE(dictHeader), dictHeader = NULL;

    /* Allocate a buffer for the JBIG2 dictionary */
    IO_CHECKMALLOC(dictBlock = TFX_MALLOC(dictBlockBytes));

    /* Read the JBIG2 dictionary */
    TFX_READ(tfxFile, dictBlock, (Int32)dictBlockBytes);

    /* Return the dictionary */
    *ppDictBlock = dictBlock;
    *pdwDictBytes = dictSize;
    *pdwDictBlockBytes = dictBlockBytes;
    *pwLastPageUsed = lastPage;
    *pdwDecoderMemory = decMem;

    NORMAL_RETURN;

 ERR_CLEANUP:

    if (dictHeader) {
	TFX_FREE(dictHeader);
	dictHeader = NULL;
    }
    if (dictBlock) {
	TFX_FREE(dictBlock);
	dictBlock = NULL;
    }
    NORMAL_RETURN;
}


#ifndef NOIPCORE
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy image data from the intermediate image buffer to the user's buffer
 *
 * LongDescription:::
 *  wp_CopyFromTempImage() copies data stored in the intermediate image buffer,
 *  to the user's image buffer. The number of rows filled is returned.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image, contains
 *				intermediate image buffer
 *
 * Output:::
 *  ppImgChannels::		Destination image channel(s)
 *  dstBPL::			Destination image bytes per line (for one channel)
 *  pdwImgRowOffset::		Pointer to image proper row offset, updated for
 *				outer loop
 *  pdwDstImgRowOffset::	Pointer to destination row offset, updated for
 *				outer loop
 *  pdwRowsToRead::		Pointer to the number of rows in outer loop to
 *				read
 *  pdwTotalRowsFilled::	Pointer to the total number of rows filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_CopyFromTempImage(
    TFX_IMG *tfxImg,
    UInt8 **ppImgChannels,
    Int32 dstBPL,
    UInt32 *pdwImgRowOffset,
    UInt32 *pdwDstImgRowOffset,
    UInt32 *pdwRowsToRead,
    UInt32 *pdwTotalRowsFilled
)
{
    UInt8 *srcPtr, *dstPtr;
    Int32 byteCount;
    IPWState *pIPWState;
    Int32 channel, channels;
    PIXR *srcPixr, *pSubPixr;
    Int32 srcBPL;
    UInt32 rowsFilled = 0UL;
    UInt32 rowsAvailable;
    PROLOGUE;

    /* Default to nothing happened */
    *pdwTotalRowsFilled = 0UL;
    
    if (tfxImg->tmpPixr == NULL || *pdwImgRowOffset < tfxImg->tmpPixrRow ||
	*pdwImgRowOffset >= tfxImg->tmpPixrRow + tfxImg->tmpPixrRows) {
	NORMAL_RETURN;
    }

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    srcPixr = tfxImg->tmpPixr;
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, srcPixr, &channels));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, srcPixr, &srcBPL));
    byteCount = (tfxImg->minBytesPerRow + 3) & ~3;
    if (byteCount > dstBPL) {
	byteCount = dstBPL;		/* This might be a problem */
    }
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, srcPixr, channel,
					  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &srcPtr));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, srcPixr, &srcPtr));
	}
	srcPtr += (*pdwImgRowOffset - tfxImg->tmpPixrRow) * srcBPL;
	dstPtr = ppImgChannels[channel] + *pdwDstImgRowOffset * dstBPL;
	rowsAvailable = tfxImg->tmpPixrRows - (*pdwImgRowOffset - tfxImg->tmpPixrRow);
	if (rowsAvailable > *pdwRowsToRead) {
	    rowsAvailable = *pdwRowsToRead;
	}

	/* Copy the user's image to the PIXR image */
	while (rowsAvailable--) {
	    memcpy(dstPtr, srcPtr, byteCount);
	    srcPtr += srcBPL;
	    dstPtr += dstBPL;
	    rowsFilled++;
	}
    } /* for each channel */
    rowsFilled /= channels;
    *pdwImgRowOffset += rowsFilled;
    *pdwDstImgRowOffset += rowsFilled;
    *pdwRowsToRead -= rowsFilled;
    (*pdwTotalRowsFilled) += rowsFilled;

    /* Convert image order according to user's specifications */

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Read an arbitrary strip of image data.
 *
 * LongDescription:::
 *  TFX_ImageReadStrip() returns a strip of image data from anywhere within
 *  the image. Appropriate buffering may be necessary to read stored strips
 *  of data required to fill the requested rows. Each stored strip is
 *  decompressed entirely before the image data is buffered out.
 *  <BR>
 *  Image data is always top-to-bottom, and left-to-right. If the client
 *  does not choose to read the entire image in as a single strip, it is
 *  advisable to use a strip height that is a multiple of "dwBestStripHeight",
 *  such that not intermediate buffering is required for partial stored strips.
 *  <BR>
 *  This routine is really a wrapper on top of w_TfxReadStrip().
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *  dwImgRowOffset::		Source row offset of the image proper
 *  dwStripHeight::		Number of rows to fill
 *
 * Output:::
 *  ppImagePlanes::		Pointer to 1-3 image planes (channels)
 *  dwBytesPerRow::		Number of bytes per row of image data in each
 *				plane
 *  dwBufferRowOffset::		Destination row offset in the image planes
 *  pdwRowsFilled::		Number of rows successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data from this
 *  buffer as needed, while reading into the intermediate buffer as needed.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_ImageReadStrip(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 **ppImagePlanes,
	UInt32 dwBytesPerRow,
	UInt32 dwBufferRowOffset,
	UInt32 *pdwRowsFilled
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_IMGINFO *pTftfxImgInfo;
    UInt32 totalRowsFilled = 0UL;
    ColorType ctype;
    IPWState *pIPWState;
    Int32 tmpPixrHeight = 0;
    UInt32 rowsFilled = 0UL;
    Int32 readRows;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;
    
    TFX_CheckHandle(tfxImgHandle);
    
    pTftfxImgInfo = &tfxImg->imgInfo;
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxImg->seqStatus & TFX_IMG_SEQ_READSTATUS) !=
		      TFX_IMG_SEQ_READSTART), IO_SEQUENCE);
    
    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    /* If we already have an intermediate PIXR, get its height */
    if (tfxImg->tmpPixr != NULL) {
	XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, tfxImg->tmpPixr, &tmpPixrHeight));
    } /* If no intermediate image buffer */

    /* Loop until the user's image buffer (region) is filled */
    while (dwStripHeight) {

	/* Check for overfill */
	if (dwImgRowOffset + dwStripHeight > pTftfxImgInfo->dwHeight) {
	    dwStripHeight = pTftfxImgInfo->dwHeight - dwImgRowOffset;
	}
	if (dwStripHeight <= 0) {
	    break;
	}

	/* Fill any left over data in the intermediate image */
	XERR_PASS_CLEANUP(wp_CopyFromTempImage(tfxImg, ppImagePlanes, dwBytesPerRow,
	    &dwImgRowOffset, &dwBufferRowOffset, &dwStripHeight, &rowsFilled));
	totalRowsFilled += rowsFilled;

	/* Break out of loop if we're finished */
	if (dwStripHeight == 0) {
	    break;
	}

	/* Compute the number of rows to read */
	if (pTftfxImgInfo->dwBestStripHeight <= 0) {
	    UInt32 stripOffset = 0;
	    UInt32 stripRow = 0;

	    /* Look for the next variable-sized strip */
	    while (stripRow < dwImgRowOffset &&
		stripOffset < tfxImg->fields.StripsPerImage) {
		stripRow += tfxImg->fields.StripRowCounts[stripOffset];
		stripOffset++;
	    }
	    XERR_GEN_CLEANUP(stripRow != dwImgRowOffset, IO_SEQUENCE);
	    XERR_GEN_CLEANUP(stripOffset >= tfxImg->fields.StripsPerImage,
		IO_BADFILEFORMAT);
	    readRows = tfxImg->fields.StripRowCounts[stripOffset];
	} else {
	    readRows = pTftfxImgInfo->dwBestStripHeight;
	}
	if (readRows > (Int32)pTftfxImgInfo->dwHeight - (Int32)dwImgRowOffset) {
	    readRows = pTftfxImgInfo->dwHeight - dwImgRowOffset;
	}
		
	/* If rows requested are the strip size */
	if (tmpPixrHeight < readRows) {
	    XERR_GEN_CLEANUP(tfxImg->tmpPixrRow + tfxImg->tmpPixrRows > dwImgRowOffset,
		IO_INTERNAL_ERROR);
	    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &tfxImg->tmpPixr));
	    tfxImg->tmpPixr = NULL;
	}

	/* If we don't have an intermediate PIXR to handle buffering,
         * create it now */
        if (tfxImg->tmpPixr == NULL) {
	    ScaledInt1616 xres, yres;
		
	    /* Prepare a temporary PIXR */
	    xres = Int32_TO_ScaledInt1616(pTftfxImgInfo->dwXResolution);
	    yres = Int32_TO_ScaledInt1616(pTftfxImgInfo->dwYResolution);
	    switch ((int)pTftfxImgInfo->xColorType) {
	    case IO_COLORTYPE_COLOR24:
		ctype = cColorRGB;
		break;
	    case IO_COLORTYPE_BINARY:
	    case IO_COLORTYPE_GRAY4:
	    case IO_COLORTYPE_COLOR4:
	    case IO_COLORTYPE_GRAY8:
	    case IO_COLORTYPE_COLOR8:
		ctype = cColorGray;
		break;
	    default:
		goto ERR_CLEANUP;
	    }
	    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, pTftfxImgInfo->dwWidth,
		readRows, IO_BPPE(pTftfxImgInfo->xColorType),
		xres, yres, 32, IO_CHANNELS(pTftfxImgInfo->xColorType),
		ctype, NULL, &tfxImg->tmpPixr));
	    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, tfxImg->tmpPixr, &tmpPixrHeight));
	} /* If no intermediate image buffer */

	/* Read a TIFF strip from the file if intermediate PIXR isn't filled with
	 * the next requested image line */
	if (tfxImg->tmpPixrRows == 0 || tfxImg->tmpPixrRow > dwImgRowOffset ||
	    tfxImg->tmpPixrRow + tfxImg->tmpPixrRows <= dwImgRowOffset) {
	    XERR_PASS_CLEANUP(w_TfxReadStrip(pIPWState, tfxImgHandle,
		dwImgRowOffset, readRows, tfxImg->tmpPixr, 0, &rowsFilled));
	    if (rowsFilled < 1) {
		break;
	    }
	    tfxImg->tmpPixrRow = dwImgRowOffset;
	    tfxImg->tmpPixrRows = rowsFilled;
	}

	/* Fill any new data in the intermediate image */
	XERR_PASS_CLEANUP(wp_CopyFromTempImage(tfxImg, ppImagePlanes, dwBytesPerRow,
	    &dwImgRowOffset, &dwBufferRowOffset, &dwStripHeight, &rowsFilled));
	totalRowsFilled += rowsFilled;
    }

    /* Output the total number of rows in user's image filled */
    *pdwRowsFilled = totalRowsFilled;
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}
#endif /* NOIPCORE */


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do necessary cleanup for the image read sequence.
 *
 * LongDescription:::
 *  TFX_ImageReadFinish() notifies the TIFF-FX library that the client is
 *  finished reading an image. Every call to TFX_ImageReadStart must be
 *  eventually followed by a call to this function.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_ImageReadFinish(
    IO_IMGHANDLE tfxImgHandle
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_IMGINFO *pTftfxImgInfo;
    TFX_MISCMEM *mmptr, *next_mm;
    TFX_INSTANCE *instData;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImgHandle);
    
    pTftfxImgInfo = &tfxImg->imgInfo;
    instData = TFX_IMGINST(tfxImg);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxImg->seqStatus & TFX_IMG_SEQ_READSTATUS) !=
		      TFX_IMG_SEQ_READSTART), IO_SEQUENCE);
    
    /* Free intermediate data */
    if (tfxImg->cmpBuffer) {
	TFX_FREE(tfxImg->cmpBuffer);
	tfxImg->cmpBuffer = NULL, tfxImg->cmpBufsize = tfxImg->cmpBytes = 0UL;
    }
#ifndef NOIPCORE
    if (tfxImg->tmpPixr) {
	IPWState *pIPWState;

	/* Reference the IPWrap state */
	XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

	XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &tfxImg->tmpPixr));
	tfxImg->tmpPixrRow = 0;
	tfxImg->tmpPixrRows = 0;
    }
#endif /* NOIPCORE */
    if ((mmptr = tfxImg->tmpCmpData) != NULL) {
	while (mmptr) {
	    if (mmptr->mem) {
		TFX_FREE(mmptr->mem);
	    }
	    next_mm = mmptr->next;
	    TFX_FREE(mmptr);
	    mmptr = next_mm;
	}
	tfxImg->tmpCmpData = NULL;
    }
    
    /* Reset the read status */
    tfxImg->seqStatus &= ~TFX_IMG_SEQ_READSTATUS;	/* Clear read status */
    tfxImg->seqStatus |= TFX_IMG_SEQ_READFINISH;		/* Set to finished */
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image description for the image region of a binary "even" layer
 *
 * LongDescription:::
 *  TFX_GetLayerInfo() gets the description of a single image that could
 *  represent the entire layer of a binary even-numbered layer: position
 *  relative to the Primary image, resolution, and size.
 *  <BR>
 *  Returned is an image handle to be used to reference this image in later
 *  calls to TFX_ImageReadLayer(). A call to TFX_CloseDocument() renders this
 *  image handle invalid.
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  dwImgLayer::		Image layer (1, 2, ...)
 *
 * Output:::
 *  pdwXOffset::		Pointer to the X location in page coordinates
 *  pdwYOffset::		Pointer to the Y location in page coordinates
 *  pdwXResolution::		Pointer to the X Resolution of image in dpm
 *  pdwYResolution::		Pointer to the Y Resolution of image in dpm
 *  pdwWidth::			Pointer to the width in image's X resolution
 *				units
 *  pdwHeight::			Pointer to the height in image's Y resolution
 *				units
 *  pxCompression::		Pointer to the compression type of all images,
 *				or IO_COMP_UNKNOWN if not just one
 *  pxColorType::		Color type (see IO_COLORTYPE
 *  pPalette::			Color palette if IO_COLORTYPE is
 *				IO_COLORTYPE_COLOR4 or IO_COLORTYPE_COLOR8.
				Palette is RGB triples (chunky): RGBRGB...
 *  pTftfxImgHandle::		Returned image handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetLayerInfo(
	IO_PAGEHANDLE tfxPageHandle,
	UInt32 dwImgLayer,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	IO_COMPRESSION *pxCompression,
	IO_COLORTYPE *pxColorType,
	UInt8 *pPalette,
	IO_IMGHANDLE *pTftfxImgHandle
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    UInt32 imgNum;
    UInt32 xOffset, yOffset, xRes, yRes, width, height;
    UInt32 bestStripHeight;
    IO_COMPRESSION compression;
    IO_COLORTYPE colorType;
    UInt8 palette[768];
    IO_IMGHANDLE imgHandle;
    TFX_IMG *primImg;
    UInt32 pageXRes, pageYRes;
    PROLOGUE;
    
    TFX_CheckHandle(tfxPageHandle);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxPage->seqStatus &
	(TFX_PAGE_SEQ_READSTATUS | TFX_PAGE_SEQ_WRITESTATUS)) == 0),
	IO_SEQUENCE);
    
    XERR_GEN_CLEANUP(dwImgLayer < 1 ||
	(tfxPage->pageInfo.dwLayersPresent > 0 &&
	(dwImgLayer > tfxPage->pageInfo.dwLayersPresent ||
	tfxPage->pageInfo.pdwImageCounts == NULL ||
	tfxPage->pageInfo.pdwImageCounts[dwImgLayer - 1] <= 0)),
	IO_BADPARAMETER);
    
    instData = TFX_PAGEINST(tfxPage);
    tfxFile = TFX_PAGEFILE(tfxPage);
    
    /* Clear the results */
    *pdwXOffset = *pdwYOffset = *pdwXResolution = *pdwYResolution = 0UL;
    *pdwWidth = *pdwHeight = 0UL;
    *pxCompression = IO_COMP_UNKNOWN;
    *pxColorType = IO_COLORTYPE_NONE;

    /* Get the page resolution */
    primImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;
    pageXRes = primImg->imgInfo.dwXResolution;
    pageYRes = primImg->imgInfo.dwYResolution;

    /* Loop through the images on that layer, looking for the total
     * dimensions */
    for (imgNum = 0; imgNum < tfxPage->pageInfo.pdwImageCounts[dwImgLayer-1];
    imgNum++) {

	/* Get the image information for this image */
	XERR_WARNOK_CLEANUP(TFX_GetImageInfo(tfxPageHandle, dwImgLayer,
	    imgNum, &xOffset, &yOffset, &xRes, &yRes, &width, &height,
	    &bestStripHeight, &compression, &colorType,
	    palette, &imgHandle));
    
	/* Copy out the image information if it's the first image */
	if (imgNum == 0) {
	    *pdwXOffset = xOffset;
	    *pdwYOffset = yOffset;
	    *pdwXResolution = xRes;
	    *pdwYResolution = yRes;
	    *pdwWidth = width;
	    *pdwHeight = height;
	    /* Use the first image of the layer as the reference */
	    *pTftfxImgHandle = imgHandle;
	    *pxCompression = compression;
	    *pxColorType = colorType;
	    switch (colorType) {
	    case IO_COLORTYPE_COLOR4:
	    case IO_COLORTYPE_COLOR8:
		XERR_GEN_CLEANUP(pPalette == NULL, IO_BADPARAMETER);
		memcpy(pPalette, palette, (1 << IO_BPPE(colorType)) * 3);
		break;
	    }
	} else {

	    /* Check the parameters that should be consistent.
	     * Note: this may not really mean a bad file structure/content;
	     * it may just mean that different characteristics among the
	     * various images within the layer.
	     * If these errors occur, use TFX_GetImageInfo() instead,
	     * and deal with each image separately. */
	    XERR_GEN_CLEANUP(xRes != *pdwXResolution, IO_NOSUPPORT);
	    XERR_GEN_CLEANUP(yRes != *pdwYResolution, IO_NOSUPPORT);
	    XERR_GEN_CLEANUP(colorType != *pxColorType, IO_NOSUPPORT);
	    /* Palettes should be identical */
	    switch (colorType) {
	    case IO_COLORTYPE_COLOR4:
	    case IO_COLORTYPE_COLOR8:
		XERR_GEN_CLEANUP(memcmp(pPalette, palette, (1 << IO_BPPE(colorType)) * 3),
		    IO_BADPARAMETER);
		break;
	    }

	    /* If the dimensions grow the size of the layer image, update
	    * the dimensions.
	    * Note that we'll keep track of the offset in page resolution, and
	    * the size in this layers' dimensions. */
	    if (xOffset < *pdwXOffset) {
		*pdwWidth += *pdwXOffset - xOffset * xRes / pageXRes;
		*pdwXOffset = xOffset;
	    }
	    if (yOffset < *pdwYOffset) {
		*pdwHeight += *pdwYOffset - yOffset * yRes / pageYRes;
		*pdwYOffset = yOffset;
	    }
	    if ((xOffset - *pdwXOffset) * xRes / pageXRes + width > *pdwWidth) {
		*pdwWidth = (xOffset - *pdwXOffset) * xRes / pageXRes + width;
	    }
	    if ((yOffset - *pdwYOffset) * yRes / pageYRes + height > *pdwHeight) {
		*pdwHeight = (yOffset - *pdwYOffset) * yRes / pageYRes + height ;
	    }

	    /* Add in the compression type to the list (unique bits) */
	    *pxCompression |= compression;
	}
    }  /* For each image in the layer */
	
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


#ifndef NOIPCORE
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the image data used to represent one image for an even layer
 *
 * LongDescription:::
 *  TFX_ImageReadLayer() returns the entire image associated with all the
 *  images belonging to an even numbered layer, as described by the image
 *  return via TFX_GetLayerInfo().
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetLayerInfo()
 *  dwImageHeight::		Number of rows to fill
 *
 * Output:::
 *  ppImageBuffer::		Pointer to one or more image planes
 *  dwBytesPerRow::		Number of bytes per row of image data
 *  pdwRowsFilled::		Number of rows successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function allows piecing together the various images separated by
 *  into overlapping or abutting same-width images (it does NOT take care
 *  of creating a mask for other conditions).
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_ImageReadLayer(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwImageHeight,
	UInt8 **ppImageBuffer,
	UInt32 dwBytesPerRow,
	UInt32 *pdwRowsFilled
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    IO_COMPRESSION compression;
    UInt32 layerXOffset, layerYOffset, layerXRes, layerYRes;
    UInt32 layerWidth, layerHeight;
    IO_COLORTYPE colorType;
    UInt8 palette[768];
    IPWState *pIPWState;
    ScaledInt1616 xres, yres;
    PIXR *pPixr = NULL;
    UInt32 row, rowsFilled, bpl;
    Int32 srcBpl, dstBpl;
    UInt8 *src, *dst;
    ColorType ctype = cColorGray;
    UInt32 channel, channels = 1;
    PROLOGUE;

    /* Default to return nothing */
    *pdwRowsFilled = 0UL;
    
    TFX_CheckHandle(tfxImgHandle);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxImg->tfxPage->seqStatus & TFX_PAGE_SEQ_READSTATUS) !=
		      TFX_PAGE_SEQ_READSTART), IO_SEQUENCE);

    /* Get the characteristics of the layer image */
    XERR_PASS_CLEANUP(TFX_GetLayerInfo((IO_PAGEHANDLE)tfxImg->tfxPage,
	tfxImg->imgInfo.dwImgLayer, &layerXOffset, &layerYOffset,
	&layerXRes, &layerYRes, &layerWidth, &layerHeight, &compression,
	&colorType, &palette[0], &tfxImgHandle));

    /* Note, nothing supported for paletted layers.  Palettes would have to be
     * compared, and a common one devised.  None of this even attempted.
     */
    XERR_GEN_CLEANUP(colorType == IO_COLORTYPE_COLOR8, IO_NOSUPPORT);

    /* There should be some consistency */
    XERR_GEN_CLEANUP(tfxImgHandle != (IO_IMGHANDLE)tfxImg, IO_INTERNAL_ERROR);
    XERR_GEN_CLEANUP(layerHeight != dwImageHeight, IO_BADPARAMETER);
    XERR_GEN_CLEANUP(dwBytesPerRow >= layerWidth * 32, IO_BADPARAMETER);
    XERR_GEN_CLEANUP(ppImageBuffer == NULL, IO_BADPARAMETER);
    XERR_GEN_CLEANUP(*ppImageBuffer == NULL, IO_BADPARAMETER);

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    /* Create a PIXR to hold the composite image */
    xres = Int32_TO_ScaledInt1616(layerXRes);
    yres = Int32_TO_ScaledInt1616(layerYRes);
    if (colorType == IO_COLORTYPE_COLOR24) {
	ctype = cColorRGB;
	channels = IO_CHANNELS(colorType);
    }
    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, layerWidth,
    	layerHeight, IO_BPPE(colorType), xres, yres, 32,
	channels, ctype, NULL, &pPixr));

    /* Populate the PIXR with the image data */
    XERR_PASS_CLEANUP(w_TfxImageReadLayer(pIPWState, tfxImgHandle, pPixr,
	&rowsFilled));
    XERR_GEN_CLEANUP(rowsFilled != layerHeight, IO_INTERNAL_ERROR);

    /* Copy the PIXR image data to the client image */
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &srcBpl));
    dstBpl = dwBytesPerRow;
    bpl = MIN(srcBpl, dstBpl);
    for (channel = 0; channel < channels; channel++) {
	PIXR *pSubPixr;

	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
		&pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &src));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr, &src));
	}
	dst = ppImageBuffer[channel];
	for (row = 0; row < layerHeight; row++) {
	    memcpy(dst, src, bpl);
	    dst += dstBpl;
	    src += srcBpl;
	}
    }
    
    /* Cleanup */
    XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &pPixr));
    
    /* Output the total number of rows in user's image filled */
    *pdwRowsFilled = layerHeight;

    NORMAL_RETURN;
    
 ERR_CLEANUP:

    w_destroyPixr(pIPWState, &pPixr);
    ERR_RETURN;
}
#endif /* NOIPCORE */


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the characteristics of an annotation
 *
 * LongDescription:::
 *  TFX_GetAnnotationInfo() reads the data necessary to describe an
 *  annotation. This includes the annotation: type, position, flags, size,
 *  and number of bytes of raw data.
 *
 * Input:::
 *  tfxPageHandle::		Valid page handle, as returned from
 *				TFX_GetPageInfo()
 *  wAnnotOffset::		Absolute offset of this annotation on the
 *				page
 *
 * Output:::
 *  pwAnnotType::		Pointer to the annotation type
 *  pdwXOffset::		Pointer to the X offset
 *  pdwYOffset::		Pointer to the Y offset
 *  pdwWidth::			Pointer to the width
 *  pdwHeight::			Pointer to the height
 *  pdwFlags::			Pointer to the flags
 *  pdwAnnotLength::		Pointer to the number of data bytes
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetAnnotationInfo(
	IO_PAGEHANDLE tfxPageHandle,
	UInt16 wAnnotOffset,
	UInt16 *pwAnnotType,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwFlags,
	UInt32 *pdwAnnotLength
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_ANNOTTABLE *annotTable;
    TFX_ANNOTTABLEENTRY *annot;
    UInt32 entryOffset;
    PROLOGUE;
    
    TFX_CheckHandle(tfxPageHandle);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((tfxPage->seqStatus & TFX_PAGE_SEQ_GETINFO) == 0,
		      IO_SEQUENCE);

    /* Read the annotation headers, and extract the data for this particular
     * annotation. Note that TIFF-FX 3.0 and higher will extract a set of
     * annotation descriptions all at once, when reading an annotation table.
     * For TIFF-FX 2.0, the information isn't grouped in a table, and has to
     * be read individually */
    XERR_PASS(wp_TfxGetAnnotTable(tfxPage, wAnnotOffset, &annotTable));

    /* Return the results */
    entryOffset = wAnnotOffset % tfxPage->annotTableHeader.entriesPerTable;
    annot = &annotTable->entries[entryOffset];
    *pwAnnotType = annot->type;
    *pdwXOffset = annot->xOffset;
    *pdwYOffset = annot->yOffset;
    *pdwWidth = annot->width;
    *pdwHeight = annot->height;
    *pdwFlags = annot->flags;
    *pdwAnnotLength = annot->bytes;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the characteristics of an annotation
 *
 * LongDescription:::
 *  TFX_GetAnnotationData() reads the data associated with an annotation.
 *
 * Input:::
 *  tfxPageHandle::		Valid page handle, as returned from
 *				TFX_GetPageInfo()
 *  wAnnotOffset::		Absolute offset of this annotation on the
 *				page
 *
 * Output:::
 *  dwBytes::			Number of bytes to retrieve (shouldn't exceed
 *				the stored amount)
 *  pData::			Buffer to occupy annotation data
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetAnnotationData(
	IO_PAGEHANDLE tfxPageHandle,
	UInt16 wAnnotOffset,
	UInt32 dwBytes,
	UInt8 *pData
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    PROLOGUE;
    
    TFX_CheckHandle(tfxPageHandle);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((tfxPage->seqStatus & TFX_PAGE_SEQ_GETINFO) == 0,
		      IO_SEQUENCE);

    /* Read the annotation data */
    XERR_PASS(wp_TfxGetAnnotData(tfxPage, wAnnotOffset, dwBytes, pData));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the AnnotationOffsets characteristic for the page
 *
 * LongDescription:::
 *  TFX_GetAnnotationOffsets() retrieves the AnnotationOffsets field
 *  for a page
 *
 * Input:::
 *  tfxPageHandle::		Valid page handle, as returned from
 *				TFX_GetPageInfo()
 *
 * Output:::
 *  pdwAnnotationOffsets::	Pointer to the AnnotationOffsets value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetAnnotationOffsets(
    IO_PAGEHANDLE tfxPageHandle,
	UInt32 *pdwAnnotationOffsets,
	Bool *pbExists
    )
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_IMG *primImg;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* Reference the primary IFD */
    primImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;
    TFX_CheckHandle(primImg);

    /* If found, and read, transfer the data */
    if (primImg->fields.found.AnnotationOffsets) {
	*pdwAnnotationOffsets = primImg->fields.AnnotationOffsets[0];
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Author characteristic for the document
 *
 * LongDescription:::
 *  TFX_GetAuthor() retrieves the Author field for a document
 *
 * Input:::
 *  tfxDocHandle::		Valid document handle, as returned from
 *				TFX_GetDocumentInfo()
 *
 * Output:::
 *  ppAuthor::			Pointer to the Author array returned
 *  pwAuthorBytes::		Pointer to the number of author bytes
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetAuthor(
    IO_DOCHANDLE tfxDocHandle,
	const Int8 **ppAuthor,
	UInt16 *pwAuthorBytes,
	Bool *pbExists
    )
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxDocHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxDoc->Author && tfxDoc->AuthorBytes > 0) {
	*ppAuthor = tfxDoc->Author;
	*pwAuthorBytes = tfxDoc->AuthorBytes;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the BitsPerSample characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetBitsPerSample() retrieves the BitsPerSample field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  ppwBitsPerSample::		Pointer to the BitsPerSample array returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetBitsPerSample(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 **ppwBitsPerSample,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.BitsPerSample) {
	*ppwBitsPerSample = tfxImg->fields.BitsPerSample;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ChromaPositioning characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetChromaPositioning() retrieves the ChromaPositioning field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pwChromaPositioning::	Pointer to the ChromaPositioning value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetChromaPositioning(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 *pwChromaPositioning,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.ChromaPositioning) {
	*pwChromaPositioning = tfxImg->fields.ChromaPositioning;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ChromaSubSampling characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetChromaSubSampling() retrieves the ChromaSubSampling field
 *  for an image.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pwChromaSubSamplingHor::	Pointer to the Horizontal ChromaSubSampling
 *				value returned
 *  pwChromaSubSamplingVer::	Pointer to the Vertical ChromaSubSampling
 *				value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetChromaSubSampling(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 *pwChromaSubSamplingHor,
	UInt16 *pwChromaSubSamplingVer,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.ChromaSubSampling) {
	*pwChromaSubSamplingHor = tfxImg->fields.ChromaSubSamplingHor;
	*pwChromaSubSamplingVer = tfxImg->fields.ChromaSubSamplingVer;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the CodingMethods characteristic for the document
 *
 * LongDescription:::
 *  TFX_GetCodingMethods() retrieves the CodingMethods field
 *  for a document.
 *
 * Input:::
 *  tfxPageHandle::		Valid page handle, as returned from
 *				TFX_GetPageInfo()
 *
 * Output:::
 *  pdwCodingMethods::		Pointer to the CodingMethods value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetCodingMethods(
    IO_DOCHANDLE tfxDocHandle,
	UInt32 *pdwCodingMethods,
	Bool *pbExists
    )
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxDocHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxDoc->codingMethods != 0) {
	*pdwCodingMethods = tfxDoc->codingMethods;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Compression characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetCompression() retrieves the Compression field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pwCompression::		Pointer to the Compression value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetCompression(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 *pwCompression,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.Compression) {
	*pwCompression = tfxImg->fields.Compression;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the DateTime characteristic for the page
 *
 * LongDescription:::
 *  TFX_GetDateTime() retrieves the DateTime field
 *  for a page
 *
 * Input:::
 *  tfxPageHandle::		Valid page handle, as returned from
 *				TFX_GetPageInfo()
 *
 * Output:::
 *  ppDateTime::		Pointer to the DateTime array returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetDateTime(
    IO_PAGEHANDLE tfxPageHandle,
	Int8 **ppDateTime,
	Bool *pbExists
    )
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxPage->DateTime != NULL) {
	*ppDateTime = tfxPage->DateTime;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Decode characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetDecode() retrieves the Decode field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  ppdwDecode::		Pointer to the Decode array returned
 *  pdwDecodeValues::		Pointer to the Decode value count
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetDecode(
    IO_IMGHANDLE tfxImgHandle,
	Int32 **ppdwDecode,
	UInt32 *pdwDecodeValues,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.Decode) {
	*ppdwDecode = tfxImg->fields.Decode;
	*pdwDecodeValues = tfxImg->fields.DecodeValues;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the DocumentName characteristic for the doc
 *
 * LongDescription:::
 *  TFX_GetDocumentName() retrieves the DocumentName field
 *  for a doc
 *
 * Input:::
 *  tfxDocHandle::		Valid doc handle, as returned from
 *				TFX_GetDocInfo()
 *
 * Output:::
 *  ppDocumentName::		Pointer to the DocumentName array returned
 *  pdwDocumentNameBytes::	Pointer to the DocumentName length
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetDocumentName(
    IO_DOCHANDLE tfxDocHandle,
	Int8 **ppDocumentName,
	UInt32 *pdwDocumentNameBytes,
	Bool *pbExists
    )
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxDocHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxDoc->DocumentName != NULL) {
	*ppDocumentName = tfxDoc->DocumentName;
	*pdwDocumentNameBytes = tfxDoc->DocumentNameBytes;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the FaxProfile characteristic for the page
 *
 * LongDescription:::
 *  TFX_GetFaxProfile() retrieves the FaxProfile field
 *  for a page
 *
 * Input:::
 *  tfxPageHandle::		Valid page handle, as returned from
 *				TFX_GetPageInfo()
 *
 * Output:::
 *  pFaxProfile::		Pointer to the FaxProfile value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetFaxProfile(
    IO_PAGEHANDLE tfxPageHandle,
	UInt8 *pFaxProfile,
	Bool *pbExists
    )
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxPage->FaxProfile != 0) {
	*pFaxProfile = tfxPage->FaxProfile;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the FillOrder characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetFillOrder() retrieves the FillOrder field for an image.<BR>
 *  Note that the FillOrder should only be important when passing compressed
 *  data back and forth, and will normally be important when the bit order of
 *  the data is not MSB-to-LSB.
 *
 * Input:::
 *  tfxImgHandle::		Valid page handle, as returned from
 *				TFX_GetPageInfo()
 *
 * Output:::
 *  pwFillOrder::		Pointer to the FillOrder value returned
 *				  1 = MSB to LSB (default)
 *				  2 = LSB to MSB
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetFillOrder(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 *pwFillOrder,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.FillOrder) {
	*pwFillOrder = tfxImg->fields.FillOrder;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the HostComputer characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetHostComputer() retrieves the HostComputer field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  ppHostComputer::		Pointer to the HostComputer array returned
 *  pdwHostComputerLength::	Pointer to the HostComputer length
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetHostComputer(
    IO_IMGHANDLE tfxImgHandle,
	Int8 **ppHostComputer,
	UInt32 *pdwHostComputerLength,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.HostComputer) {
	*ppHostComputer = tfxImg->fields.HostComputer;
	*pdwHostComputerLength = tfxImg->fields.HostComputerLength;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the IFDBaseColor characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetIFDBaseColor() retrieves the IFDBaseColor field for an image.
 *  <BR>
 *  Note, IFDBaseColor is only used when the first extension set is being used.
 *  In this case, both the IFDBaseColor and ImageBaseColor may be used, but the
 *  ImageBaseColor field may be omitted if the color image is not interpretable
 *  by a base TIFF-FX reader (for example it has JPEGTables).
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  ppwImageBaseColor::		Pointer to the ImageBaseColor array returned
 *  pdwImageBaseColorValues::	Pointer to the ImageBaseColor length
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetIFDBaseColor(
    IO_IMGHANDLE tfxImgHandle,
	Int32 **ppdwIFDBaseColor,
	UInt32 *pdwIFDBaseColorValues,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.IFDBaseColor) {
	*ppdwIFDBaseColor = tfxImg->fields.IFDBaseColor;
	*pdwIFDBaseColorValues = tfxImg->fields.IFDBaseColorValues;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ImageBaseColor characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetImageBaseColor() retrieves the ImageBaseColor field for an image.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  ppwImageBaseColor::		Pointer to the ImageBaseColor array returned
 *  pdwImageBaseColorValues::	Pointer to the ImageBaseColor length
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  ImageBaseColor should be added to or replaced by IFDBaseColor when writing
 *  a TIFF-FX extension.
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetImageBaseColor(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 **ppImageBaseColor,
	UInt32 *pdwImageBaseColorValues,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.ImageBaseColor) {
	*ppImageBaseColor = tfxImg->fields.ImageBaseColor;
	*pdwImageBaseColorValues = tfxImg->fields.ImageBaseColorValues;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ImageDescription characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetImageDescription() retrieves the ImageDescription field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  ppImageDescription::	Pointer to the ImageDescription array returned
 *  pdwImageDescriptionLength::	Pointer to the ImageDescription length
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetImageDescription(
    IO_IMGHANDLE tfxImgHandle,
	Int8 **ppImageDescription,
	UInt32 *pdwImageDescriptionLength,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.ImageDescription) {
	*ppImageDescription = tfxImg->fields.ImageDescription;
	*pdwImageDescriptionLength = tfxImg->fields.ImageDescriptionLength;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the T88Options characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetT4Options() retrieves the T88Options field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pdwT4Options::		Pointer to the T88Options value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetT88Options(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 *pdwT88Options,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.T88Options) {
	*pdwT88Options = tfxImg->fields.T88Options;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ModeNumber characteristic for the page
 *
 * LongDescription:::
 *  TFX_GetModeNumber() retrieves the ModeNumber field
 *  for a page
 *
 * Input:::
 *  tfxPageHandle::		Valid page handle, as returned from
 *				TFX_GetPageInfo()
 *
 * Output:::
 *  pModeNumber::		Pointer to the ModeNumber value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetModeNumber(
    IO_PAGEHANDLE tfxPageHandle,
	UInt8 *pModeNumber,
	Bool *pbExists
    )
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxPage->ModeNumber != 0) {
	*pModeNumber = tfxPage->ModeNumber;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Orientation characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetOrientation() retrieves the Orientation field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pwOrientation::		Pointer to the Orientation value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetOrientation(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 *pwOrientation,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.Orientation) {
	*pwOrientation = tfxImg->fields.Orientation;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the PageNumber characteristic for the page
 *
 * LongDescription:::
 *  TFX_GetPageNumber() retrieves the PageNumber field
 *  for a page
 *
 * Input:::
 *  tfxPageHandle::		Valid page handle, as returned from
 *				TFX_GetPageInfo()
 *
 * Output:::
 *  pwPageNumber::		Pointer to the PageNumber value returned
 *  pwPageCount::		Pointer to the PageCount value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetPageNumber(
    IO_PAGEHANDLE tfxPageHandle,
	UInt16 *pwPageNumber,
	UInt16 *pwPageCount,
	Bool *pbExists
    )
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxPage->PageCount != 0) {
	*pwPageNumber = tfxPage->PageNumber;
	*pwPageCount = tfxPage->PageCount;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the PhotometricInterpretation characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetPhotometricInterpretation() retrieves the PhotometricInterpretation field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pwPhotometricInterpretation::  Pointer to the PhotometricInterpretation
 *				value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetPhotometricInterpretation(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 *pwPhotometricInterpretation,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.PhotometricInterpretation) {
	*pwPhotometricInterpretation = tfxImg->fields.PhotometricInterpretation;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the PlanarConfiguration characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetPlanarConfiguration() retrieves the PlanarConfiguration field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pwPlanarConfiguration::	Pointer to the PlanarConfiguration
 *				value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetPlanarConfiguration(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 *pwPlanarConfiguration,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.PlanarConfiguration) {
	*pwPlanarConfiguration = tfxImg->fields.PlanarConfiguration;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ProfileType characteristic for the page
 *
 * LongDescription:::
 *  TFX_GetProfileType() retrieves the ProfileType field
 *  for a page
 *
 * Input:::
 *  tfxPageHandle::		Valid page handle, as returned from
 *				TFX_GetPageInfo()
 *
 * Output:::
 *  pdwProfileType::		Pointer to the ProfileType value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetProfileType(
    IO_PAGEHANDLE tfxPageHandle,
	UInt32 *pdwProfileType,
	Bool *pbExists
    )
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxPage->ProfileType != 0) {
	*pdwProfileType = tfxPage->ProfileType;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ReferencePage characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetReferencePage() retrieves the ReferencePage field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pdwReferencePage::		Pointer to the ReferencePage value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetReferencePage(
    IO_DOCHANDLE tfxDocHandle,
	UInt32 *pdwReferencePage,
	Bool *pbExists
    )
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    TFX_IMG *tfxImg = NULL;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxDocHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* Make sure the first IFD has been read */
    if ((tfxDoc->seqStatus & TFX_DOC_SEQ_GETINFO) == 0) {
	DOCTYPE dwProfile;

	XERR_PASS(TFX_GetDocProfile(tfxDocHandle, &dwProfile));
    }

    if ((tfxDoc->seqStatus & TFX_DOC_SEQ_GPEXISTS)
	&& (tfxDoc->seqStatus & TFX_DOC_SEQ_REFPAGESET)) {
	
	/* Examine the first IFD */
	if (tfxDoc->pages && tfxDoc->pages->pageInfo.tfxPrimImgHandle) {
	    tfxImg = (TFX_IMG *)tfxDoc->pages->pageInfo.tfxPrimImgHandle;
	}
    }

    /* If found, and read, transfer the data */
    if (tfxImg && tfxImg->fields.found.ReferencePage) {

	/* Stored value is page number (ordinal), not offset */
	*pdwReferencePage = tfxImg->fields.gpIFD.ReferencePage - 1;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the RenderingHints characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetRenderingHints() retrieves the RenderingHints field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  ppRenderingHints::	Pointer to the RenderingHints array returned
 *  pdwRenderingHintsLength::	Pointer to the RenderingHints length
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetRenderingHints(
    IO_IMGHANDLE tfxImgHandle,
	Int8 **ppRenderingHints,
	UInt32 *pdwRenderingHintsLength,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.RenderingHints) {
	*ppRenderingHints = tfxImg->fields.RenderingHints;
	*pdwRenderingHintsLength = tfxImg->fields.RenderingHintsLength;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the count of RenderingHints image belonging to an image
 *
 * LongDescription:::
 *  TFX_GetRenderingHintsImageCount() checks to see if a RenderingHints
 *  image exists, and returns the number of images found.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pdwCount::			Pointer to the count of RenderingHints images
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetRenderingHintsImageCount(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 *pdwCount
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pdwCount = 0;

    /* If found, read (if necessary) and transfer the data */
    if (tfxImg->fields.found.RenderingHintsIFD) {
	if (tfxImg->xHintsImgs == NULL) {

	    /* Read the IFD */
	    XERR_WARNOK_CLEANUP(wp_TfxTraverseHintsImages(tfxImg));

#if 0
	    /* There should only be one IFD */
	    if (tfxImg->imgInfo.dwHintsImages > 1) {
		__error__ = IO_WARN_NOSUPPORT;
	    }
#endif
	}

	*pdwCount = tfxImg->imgInfo.dwHintsImages;
    }

ERR_CLEANUP:

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the ResolutionUnit characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetResolutionUnit() retrieves the ResolutionUnit field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pwResolutionUnit::		Pointer to the ResolutionUnit value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetResolutionUnit(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 *pwResolutionUnit,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.ResolutionUnit) {
	*pwResolutionUnit = tfxImg->fields.ResolutionUnit;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the RowsPerStrip characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetRowsPerStrip() retrieves the RowsPerStrip field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pdwRowsPerStrip::		Pointer to the RowsPerStrip value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetRowsPerStrip(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 *pdwRowsPerStrip,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.RowsPerStrip) {
	*pdwRowsPerStrip = tfxImg->fields.RowsPerStrip;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the SamplesPerPixel characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetSamplesPerPixel() retrieves the SamplesPerPixel field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pdwSamplesPerPixel::	Pointer to the SamplesPerPixel value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetSamplesPerPixel(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 *pdwSamplesPerPixel,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.SamplesPerPixel) {
	*pdwSamplesPerPixel = tfxImg->fields.SamplesPerPixel;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Software characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetSoftware() retrieves the Software field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  ppSoftware::		Pointer to the Software array returned
 *  pdwSoftwareLength::		Pointer to the Software length
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetSoftware(
    IO_IMGHANDLE tfxImgHandle,
	Int8 **ppSoftware,
	UInt32 *pdwSoftwareLength,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.Software) {
	*ppSoftware = tfxImg->fields.Software;
	*pdwSoftwareLength = tfxImg->fields.SoftwareLength;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the TargetPrinter characteristic for the page
 *
 * LongDescription:::
 *  TFX_GetTargetPrinter() retrieves the TargetPrinter field
 *  for a page
 *
 * Input:::
 *  tfxPageHandle::		Valid page handle, as returned from
 *				TFX_GetPageInfo()
 *
 * Output:::
 *  ppTargetPrinter::		Pointer to the TargetPrinter array returned
 *  pdwTargetPrinterLength::	Pointer to the TargetPrinter length
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetTargetPrinter(
    IO_PAGEHANDLE tfxPageHandle,
	Int8 **ppTargetPrinter,
	UInt32 *pdwTargetPrinterLength,
	Bool *pbExists
    )
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxPage->TargetPrinter != NULL) {
	*ppTargetPrinter = tfxPage->TargetPrinter;
	*pdwTargetPrinterLength = tfxPage->TargetPrinterLength;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Thresholding characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetThresholding() retrieves the Thresholding field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pwThresholding::		Pointer to the Thresholding array returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetThresholding(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 *pwThresholding,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.Thresholding) {
	*pwThresholding = tfxImg->fields.Thresholding;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the T4Options characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetT4Options() retrieves the T4Options field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pdwT4Options::		Pointer to the T4Options value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetT4Options(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 *pdwT4Options,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.T4Options) {
	*pdwT4Options = tfxImg->fields.T4Options;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the T6Options characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetT6Options() retrieves the T6Options field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pdwT6Options::		Pointer to the T6Options value returned
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetT6Options(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 *pdwT6Options,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.T6Options) {
	*pdwT6Options = tfxImg->fields.T6Options;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the VersionYear characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetVersionYear() retrieves the VersionYear field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  ppVersionYear::		Pointer to the VersionYear array returned
 *  pdwVersionYearLength::	Pointer to the VersionYear length
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetVersionYear(
    IO_IMGHANDLE tfxImgHandle,
	UInt8 **ppVersionYear,
	UInt32 *pdwVersionYearLength,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.VersionYear) {
	*ppVersionYear = tfxImg->fields.gpIFD.VersionYear;
	*pdwVersionYearLength = tfxImg->fields.gpIFD.VersionYearLength;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the YcrCbCoefficients characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetYcrCbCoefficients() retrieves the YcrCbCoefficients field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  ppdwYcrCbCoefficients::	Pointer to the YcrCbCoefficients array returned
 *  pdwYcrCbCoefficientsValues::	Pointer to the YcrCbCoefficients length
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetYCbCrCoefficients(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 **ppdwYCbCrCoefficients,
	UInt32 *pdwYCbCrCoefficientsValues,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.YCbCrCoefficients) {
	/* This could be a bad thing, typecasting a TFX_RATIONAL to a UInt32 array: */
	*ppdwYCbCrCoefficients = (UInt32 *)tfxImg->fields.YCbCrCoefficients;
	*pdwYCbCrCoefficientsValues = tfxImg->fields.YCbCrCoefficientsValues;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Get the Predictor characteristic for the image
 *
 * LongDescription:::
 *  TFX_GetPredictor() retrieves the Predictor field
 *  for an image
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *
 * Output:::
 *  pwPredictor::		Pointer to the Predictor 
 *  pbExists::			Pointer to boolean: if the field doesn't
 *				exist, then *pbExists will be 0, otherwise 1.
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_GetPredictor(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 *pwPredictor,
	Bool *pbExists
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Mark as not found, until proven otherwise */
    *pbExists = 0;

    /* If found, and read, transfer the data */
    if (tfxImg->fields.found.Predictor) {
	*pwPredictor = tfxImg->fields.Predictor;
	*pbExists = 1;
    }

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do the tasks necessary to complete communication to/from a document.
 *
 * LongDescription:::
 *  TFX_CloseDocument() will, if necessary, complete the processes associated
 *  with the storage of a partially written document, including completion of
 *  the Tokenization process (see TFX_WriteTokenDictionary() and
 *  TFX_WriteTokenImage()). All internal data associated with an open
 *  document is freed.
 *
 * Input:::
 *  tfxDocHandle::		Active document handle
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_CloseDocument(
    IO_DOCHANDLE tfxDocHandle
)
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    TFX_INSTANCE *instData;
#ifndef NOIPCORE
    IPWState *pIPWState;
#endif
    PROLOGUE;

#ifndef NOIPCORE
    /* Get an pointer to the IPState, it's needed below */
    XERR_PASS (wp_TfxIPWState (tfxDoc, &pIPWState));
#endif
    
    TFX_CheckHandle(tfxDocHandle);
    instData = TFX_DOCINST(tfxDoc);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxDoc->seqStatus &
		       (TFX_DOC_SEQ_READSTART | TFX_DOC_SEQ_WRITESTART)) == 0),
		     IO_SEQUENCE);

#ifndef TFX_READONLY

    if (tfxDoc->readWrite & WRITE_ALLOWED) {

#if !defined(NOIPCORE) && !defined(NODGP)
        /* Flush the token dictionary and resolve and write the images
	 * that are in a tokenized state. */
	XERR_PASS_CLEANUP(TFX_WriteTokenImages(tfxDocHandle));
#endif	/* NOIPCORE */

	switch (tfxDoc->docInfo.profile) {
	case TFX_PROFILE_M:
	case TFX_PROFILE_T:
	case TFX_PROFILE_M1:
	case TFX_PROFILE_COH_M:
	case TFX_PROFILE_COH_M1:

	    /* Make sure the first IFD has the GlobalParametersIFD */
	    if (tfxDoc->docInfo.wNPages > 0) {
		TFX_IMG *firstImg = (TFX_IMG *)tfxDoc->pages->pageInfo.tfxPrimImgHandle;

		/* If the GPIFD isn't on the first page, and either there was
		 * no request to not write the GPIFD, or there has been a
		 * shared resource written (requiring override of the request),
		 * then write the GPIFD on the first IFD.
		 */
		if (firstImg && firstImg->fields.gpIFD.fieldCount == 0 &&
		    ((tfxDoc->seqStatus & TFX_DOC_SEQ_NOGPIFDREQ) == 0 ||
		     tfxDoc->sharedResourcesLocation > 0)) {
		    TFX_PAGE *tfxPage;
		    TFX_IMG *tfxImg;

		    /* Mark the first page as needing rewritten */
		    firstImg->fields.IFDrewriteNeeded = 1;

		    /* Mark any other primary IFD that contains a GPIFD as
		    * needing rewritten also */
		    tfxPage = tfxDoc->pages;
		    while (tfxPage) {
			tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;
			if (tfxImg != firstImg && tfxImg->fields.gpIFD.fieldCount > 0) {
			    tfxImg->fields.IFDrewriteNeeded = 1;
			}
			tfxPage = tfxPage->next;
		    }
		}
	    }
	}

	/* If any fields have been added to IFDs since they were written, rewrite the
	 * IFDs now */
	XERR_PASS_CLEANUP(wp_TfxRewriteIFDs(tfxDocHandle));
    
	/* Make sure the single-page vs. multi-page bit in NewSubfileType is correct */
/*	XERR_PASS_CLEANUP(wp_TfxTestMultipageBit(tfxDocHandle)); */
    
	/* Make sure the single-page vs. multi-page bit in NewSubfileType is correct */
	XERR_PASS_CLEANUP(wp_TfxTestCodingMethods(tfxDocHandle));
    
	/* Make a pass through the symbol compressed images, and make sure the
	 * sure the shared resources entries make reference to the last page used.
	 */
/*	XERR_PASS_CLEANUP(wp_TfxSetLastPageUsed(tfxDocHandle)); */
    }

#endif /* ! TFX_READONLY */


    /* Fall through to do the rest of the cleanup */
    
 ERR_CLEANUP:

    /* Free up the shared data */
#ifndef NODGP
    /* Close both the read and write tokenization processes */
    wp_TfxCloseTokenization(tfxDoc);
#endif	/* NODGP */

    /* Clear the status, document state reset */
    tfxDoc->seqStatus = 0;

    if (tfxDoc) {
	wp_TfxDocFree(&tfxDoc);
    }
    
    NORMAL_RETURN;
}


#ifndef NOIPCORE
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Decompress an image strip as given from TFX_ImageReadCompressed()
 *
 * LongDescription:::
 *  TFX_Decompress() decompresses an image strip returned from
 *  TFX_ImageReadCompressed().
 *
 * PrivateDescription:::
 *  This is a front-end to a structure oriented routine w_TfxDecompress().
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_GetImageInfo()
 *  xCompression::		See IO_COMPRESSION types
 *  xColorType::		Color type of image data
 *  dwFlags::			Destination image format (see IO_IMGFLAGS_...)
 *  pSrc::			Source buffer (compressed)
 *  dwBytesInSrc::		Byte count of source buffer
 *
 * Output:::
 *  dwRowOffset::		Destination row offset
 *  dwWidth::			Columns of COLORTYPE in dst image
 *  dwHeight::			Rows in dst image
 *  pDst[3]::			Pointer to destination image buffer channels
 *  dwBytesPerRow::		Bytes per row in image buffer
 *  pdwRowsFilled::		Number of rows successfully filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_Decompress(
	IO_IMGHANDLE tfxImgHandle,
	IO_COMPRESSION xCompression,
	IO_COLORTYPE xColorType,
	UInt32 dwFlags,
	UInt8 *pSrc,
	UInt32 dwBytesInSrc,
	UInt32 dwRowOffset,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt8 *pDst[3],
	UInt32 dwBytesPerRow,
	UInt32 *pdwRowsFilled
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    ScaledInt1616 xres, yres;
    Int32 rowsLeft;
    UInt32 rowsFilled = 0UL;
    Int32 pixrBpl;
    UInt32 bytesPerRow;
    IPWState *pIPWState;
    PIXR *pPixr = NULL, *pSubPixr;
    ColorType ctype = cColorNone;
    UInt8 *pUserImage, *pPixrImage;
    Int32 channel, channels;
    PROLOGUE;

    *pdwRowsFilled = 0UL;
    TFX_CheckHandle(tfxImgHandle);

    /* Get an pointer to the IPState, it's needed below */
    XERR_PASS (wp_TfxIPWState (tfxImg->tfxPage->tfxDoc, &pIPWState));

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    /* Check to see if there's anything to do */
    if (dwBytesInSrc == 0) {
	NORMAL_RETURN;
    }

    /* Check buffers and sizes */
    XERR_GEN_CLEANUP(pSrc == NULL || pDst == NULL || pDst[0] == NULL ||
	dwBytesPerRow == 0, IO_BADPARAMETER);
    XERR_GEN_CLEANUP(dwRowOffset >= tfxImg->imgInfo.dwHeight, IO_BADPARAMETER);
    
    /* Create a PIXR image to hold the image */
    xres = Int32_TO_ScaledInt1616(tfxImg->imgInfo.dwXResolution);  /* Any resolution is fine, I guess */
    yres = Int32_TO_ScaledInt1616(tfxImg->imgInfo.dwYResolution);  /* Any resolution is fine, I guess */
    switch (xColorType) {
    case IO_COLORTYPE_COLOR24:
	ctype = cColorRGB;
	break;
    case IO_COLORTYPE_BINARY:
    case IO_COLORTYPE_GRAY4:
    case IO_COLORTYPE_COLOR4:
    case IO_COLORTYPE_GRAY8:
    case IO_COLORTYPE_COLOR8:
	ctype = cColorGray;
	break;
    default:
	break;
    }
    channels = IO_CHANNELS(xColorType);
    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, dwWidth, dwHeight,
	IO_BPPE(xColorType), xres, yres, 32, channels, ctype, NULL, &pPixr));

    /* Call the underlying decompression routine */
    XERR_PASS_CLEANUP(w_TfxDecompress(pIPWState, tfxImgHandle, xCompression,
	xColorType, dwFlags, pSrc, dwBytesInSrc, 0, dwWidth, dwHeight, pPixr,
	&rowsFilled));
    
    /* Copy the data to the user's buffer */
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &pixrBpl));
    bytesPerRow = dwBytesPerRow;
    if ((Int32)bytesPerRow > pixrBpl) {
	bytesPerRow = pixrBpl;
    }
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
					  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &pPixrImage));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr, &pPixrImage));
	}

	rowsLeft = rowsFilled;
	pUserImage = pDst[channel] + dwRowOffset * dwBytesPerRow;
	while (rowsLeft > 0) {
	    memcpy(pUserImage, pPixrImage, bytesPerRow);
	    pUserImage += dwBytesPerRow;
	    pPixrImage += pixrBpl;
	    rowsLeft--;
	}
    }

    *pdwRowsFilled = rowsFilled;
    w_destroyPixr(pIPWState, &pPixr);

    NORMAL_RETURN;
    
 ERR_CLEANUP:

    w_destroyPixr(pIPWState, &pPixr);
    
    ERR_RETURN;
}
#endif /* NOIPCORE */


typedef struct Pairs {
    UInt32 a;
    UInt32 b;
} Pairs;


/*
 * Find the nearest resolution pair
 */
static void
findNearestPair(UInt32 inA, UInt32 inB, UInt32 *outA, UInt32 *outB,
		struct Pairs *list, UInt32 listItems)
{
    UInt32 item;
    UInt32 best = 0;
    UInt32 smallestErr = 0xffffffff;
    Int32 adiff, bdiff;
    UInt32 err;

    for (item = 0; item < listItems; item++) {
	adiff = inA - list[item].a;
	bdiff = inB - list[item].b;
	err = (adiff * adiff) + (bdiff * bdiff);
	if (err < smallestErr) {
	    best = item;
	    smallestErr = err;
	    if (smallestErr == 0) break;
	}
    }

    /* Return the best */
    *outA = list[best].a;
    *outB = list[best].b;
}


/*
 * Find the nearest resolution pair
 */
static void
findNearest(UInt32 in, UInt32 *out, UInt32 *list, UInt32 listItems)
{
    UInt32 item;
    UInt32 best = 0;
    UInt32 smallestErr = 0xffffffff;
    Int32 diff;
    UInt32 err;

    for (item = 0; item < listItems; item++) {
	diff = in - list[item];
	err = (diff * diff);
	if (err < smallestErr) {
	    best = item;
	    smallestErr = err;
	    if (smallestErr == 0) break;
	}
    }

    /* Return the best */
    *out = list[best];
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Check and adjust (if necessary) the image info
 *
 * LongDescription:::
 *  TFX_CheckImageInfo() verifies the consistency of the image info based
 *  on the TIFF-FX version, and compression type. Changes are made to the
 *  parameters based on what TFX_AddImage(), TFX_AddMaskImage(), and/or
 *  TFX_ImageWriteStart() would do, or what TFX_TokenizeImage() would do.
 *  <BR>
 *  Note, if this function catches inconsistent, unsupported, or non-compliant
 *  parameters, it returns a warning only:
 *	IO_WARN_NOSUPPORT or IO_WARN_NOTCOMPLIANT
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  pdwImgLayer::		Image layer in MRC (1, 2, ...)
 *  pdwImgOffset::		Image rank within the layer (0, 1, 2, ...)
 *  pxColorType::		Color type
 *  pdwXOffset::		X pixel offset in Primary image coordinates
 *  pdwYOffset::		Y pixel offset in Primary image coordinates
 *  pdwXResolution::		Resolution in the X direction
 *  pdwYResolution::		Resolution in the Y direction
 *  pdwWidth::			Width
 *  pdwHeight::			Height
 *  pdwBestStripHeight::	Height of strips stored
 *  pPalette::			Palette (if IO_COLORTYPE_COLOR8 or COLOR4)
 *  pxCompression::		Type of compression
 *  pxColorSpace::		Stored color space
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_CheckImageInfo(
	IO_PAGEHANDLE tfxPageHandle,
	UInt32 *pdwImgLayer,
	UInt32 *pdwImgOffset,
	IO_COLORTYPE *pxColorType,
	UInt32 *pdwXOffset,
	UInt32 *pdwYOffset,
	UInt32 *pdwXResolution,
	UInt32 *pdwYResolution,
	UInt32 *pdwWidth,
	UInt32 *pdwHeight,
	UInt32 *pdwBestStripHeight,
	UInt8 *pPalette,
	IO_COMPRESSION *pxCompression,
	IO_COLORSPACE *pxColorSpace
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_IMG *tfxParentImg, *primaryImg = NULL;
    UInt32 newXResolution, newYResolution;
    UInt32 newWidth;
    IO_COMPRESSION newCompression = *pxCompression;
    ErrCode warning = IO_NOERROR;
    Bool thisIsPrimaryImage = 0;
    Bool multiStripSubIFD = 0;	    /* Will we enable multi-strip SubIFD mode? */
    PROLOGUE;

    TFX_CheckHandle(tfxPageHandle);

    if (*pdwWidth == 0 || *pdwHeight == 0) {
	goto ERR_CLEANUP;
    }

    /* Either we need a primary image, or we're adding it now */
    if (*pdwImgLayer != 2 && tfxPage->pageInfo.tfxPrimImgHandle == 0 &&
	(*pdwImgLayer < 1 || *pdwImgLayer > 2)) {
	/* Could be 1 or 2:
	 * 1 if there's only a single color layer;
	 * 2 if binary (should be the first layer added if there's color also) */
	if (*pxColorType == IO_COLORTYPE_BINARY) {
	    *pdwImgLayer = 2;
	} else {
	    *pdwImgLayer = 1;
	}
	*pdwImgOffset = 0;	/* Make first in layer */
    }
    tfxParentImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    /* Page size and resolution must conform to a strict set,
     * if it's true TIFF-FX.
     * Note: I could do one of two things here:
     * 1. complain (LOUDLY) if the width and resolutions aren't compliant
     * or
     * 2. bump the width and resolutions to the nearest compliant values
     *
     * It's unclear which is the best tactic, but ideally, if the resolution
     * or width is really to be adjusted, then image processing is necessary,
     * and this is a file format library only (with some small exceptions
     * for some strange decompression handling).
     *
     * So what should I do.  How about I bump the values, but return a result
     * of NON-COMPLIANCE (IO_WARN_NOTCOMPLIANT).
     */
    if (tfxPage->pageInfo.tfxPrimImgHandle == 0 || (tfxParentImg &&
	tfxParentImg->imgInfo.dwImgLayer == *pdwImgLayer &&
	tfxParentImg->imgInfo.dwImgOffset == *pdwImgOffset)) {

	/* This must be the primary image, which defines the page
	 * dimensions and resolutions */

	/* Keep the adjusted values in new..., then adjust the return
	 * value if they aren't what was input */
	newXResolution = *pdwXResolution;
	newYResolution = *pdwYResolution;
	newWidth = *pdwWidth;

	/* Make sure the resolution is compliant */
	switch (tfxPage->tfxDoc->docInfo.profile) {
	case TFX_PROFILE_S:
	    {
		struct Pairs res[] = {{200, 100}, {204, 98}, {200, 200}, {204, 196}};
		findNearestPair(*pdwXResolution, *pdwYResolution, &newXResolution,
		    &newYResolution, res, sizeof(res) / sizeof(*res));
	    }
	    break;
	case TFX_PROFILE_F:
	case TFX_PROFILE_J:
	case TFX_PROFILE_T:
	    {
		struct Pairs res[] = {{200, 100}, {204, 98}, {200, 200}, {204, 196},
		{204, 391}, {300, 300}, {300, 600}, {408, 391}, {400, 400}, {400, 800},
		{600, 600}, {600, 1200}, {1200, 1200}};
		findNearestPair(*pdwXResolution, *pdwYResolution, &newXResolution,
		    &newYResolution, res, sizeof(res) / sizeof(*res));
	    }
	    break;
	case TFX_PROFILE_C:
	case TFX_PROFILE_L:
	case TFX_PROFILE_M:	/* Primary IFD only, determined already */
	case TFX_PROFILE_M1:	/* Primary IFD only, determined already */
	    {
		struct Pairs res[] = {{100, 100}, {200, 200}, {204, 196}, {300, 300},
		{400, 400}, {600, 600}, {1200, 1200}};
		findNearestPair(*pdwXResolution, *pdwYResolution, &newXResolution,
		    &newYResolution, res, sizeof(res) / sizeof(*res));
	    }
	    break;
	}

	/* Make sure the width is one of those required.
	 * If not, map to the dimension SMALLER (truncates the image size) */
	switch (tfxPage->tfxDoc->docInfo.profile) {
	case TFX_PROFILE_S:
	    newWidth = 1728;
	    break;
	case TFX_PROFILE_F:
	case TFX_PROFILE_L:
	case TFX_PROFILE_T:
	    switch (newXResolution) {
	    case 200:
	    case 204:
		{
		    UInt32 widths[] = {1728, 2048, 2432};
		    findNearest(*pdwWidth, &newWidth, widths,
			sizeof(widths) / sizeof(*widths));
		}
		break;
	    case 300:
		{
		    UInt32 widths[] = {2592, 3072, 3648};
		    findNearest(*pdwWidth, &newWidth, widths,
			sizeof(widths) / sizeof(*widths));
		}
		break;
	    case 400:
	    case 408:
		{
		    UInt32 widths[] = {3456, 4096, 4864};
		    findNearest(*pdwWidth, &newWidth, widths,
			sizeof(widths) / sizeof(*widths));
		}
		break;
	    case 600:
		{
		    UInt32 widths[] = {5184, 6144, 7296};
		    findNearest(*pdwWidth, &newWidth, widths,
			sizeof(widths) / sizeof(*widths));
		}
		break;
	    case 1200:
		{
		    UInt32 widths[] = {10363, 12288, 14592};
		    findNearest(*pdwWidth, &newWidth, widths,
			sizeof(widths) / sizeof(*widths));
		}
		break;
	    }
	    break;
	case TFX_PROFILE_J:
	case TFX_PROFILE_C:
	case TFX_PROFILE_M:	/* Primary IFD only, determined already */
	case TFX_PROFILE_M1:	/* Primary IFD only, determined already */
	    switch (newXResolution) {
	    case 100:
		{
		    UInt32 widths[] = {864, 1024, 1216};
		    findNearest(*pdwWidth, &newWidth, widths,
			sizeof(widths) / sizeof(*widths));
		}
		break;
	    case 200:
	    case 204:
		{
		    UInt32 widths[] = {1728, 2048, 2432};
		    findNearest(*pdwWidth, &newWidth, widths,
			sizeof(widths) / sizeof(*widths));
		}
		break;
	    case 300:
		{
		    UInt32 widths[] = {2592, 3072, 3648};
		    findNearest(*pdwWidth, &newWidth, widths,
			sizeof(widths) / sizeof(*widths));
		}
		break;
	    case 400:
		{
		    UInt32 widths[] = {3456, 4096, 4864};
		    findNearest(*pdwWidth, &newWidth, widths,
			sizeof(widths) / sizeof(*widths));
		}
		break;
	    case 600:
		{
		    UInt32 widths[] = {5184, 6144, 7296};
		    findNearest(*pdwWidth, &newWidth, widths,
			sizeof(widths) / sizeof(*widths));
		}
		break;
	    case 1200:
		{
		    UInt32 widths[] = {10363, 12288, 14592};
		    findNearest(*pdwWidth, &newWidth, widths,
			sizeof(widths) / sizeof(*widths));
		}
		break;
	    }
	    break;
	} /* End of TIFF-FX profiles */

	/* If there's a difference, make the update, and keep around a
	 * new return value */
	 if (*pdwXResolution != newXResolution) {
/*	     *pdwXResolution = newXResolution; */
	     warning = IO_WARN_NOTCOMPLIANT;
	 }
	 if (*pdwYResolution != newYResolution) {
/*	     *pdwYResolution = newYResolution; */
	     warning = IO_WARN_NOTCOMPLIANT;
	 }
	 if (*pdwWidth != newWidth) {
/*	     *pdwWidth = newWidth; */
	     warning = IO_WARN_NOTCOMPLIANT;
	 }

	 /* Make sure the page resolution is okay */
	 if (*pdwXResolution == 0) {
	     *pdwXResolution = 300;	/* My default */
	 }
	 if (*pdwYResolution == 0) {
	     *pdwYResolution = 300;	/* My default */
	 }

	 thisIsPrimaryImage = 1;

    } else { /* This is no the primary IFD, keep it around */
	primaryImg = (TFX_IMG *) tfxPage->pageInfo.tfxPrimImgHandle;
    } /* End of primary IFD check */

    /* Make some page consistency checks */
    if (tfxParentImg != NULL) {
	if (*pdwXResolution == 0) {
	    *pdwXResolution = tfxParentImg->imgInfo.dwXResolution;
	}
	if (*pdwYResolution == 0) {
	    *pdwYResolution = tfxParentImg->imgInfo.dwYResolution;
	}
	if (*pdwXOffset > tfxParentImg->imgInfo.dwWidth) {
	    *pdwXOffset = 0;
	}
	if (*pdwYOffset > tfxParentImg->imgInfo.dwHeight) {
	    *pdwYOffset = 0;
	}

	/* Note: given the resolution differences, the last subimage pixel
	 * may overhang the edge of the page, but not have a pixel completely
	 * outside of the page */
	if (*pdwXOffset + (*pdwWidth - 1) * tfxParentImg->imgInfo.dwXResolution
	    / *pdwXResolution >= tfxParentImg->imgInfo.dwWidth) {
	    if (tfxParentImg->imgInfo.dwXResolution >= *pdwXResolution) {
		*pdwWidth = (tfxParentImg->imgInfo.dwWidth - *pdwXOffset +
		    tfxParentImg->imgInfo.dwXResolution / *pdwXResolution - 1) *
		    *pdwXResolution / tfxParentImg->imgInfo.dwXResolution;
	    } else {
		*pdwWidth = ((tfxParentImg->imgInfo.dwWidth - *pdwXOffset) *
		    *pdwXResolution + tfxParentImg->imgInfo.dwXResolution - 1) /
		    tfxParentImg->imgInfo.dwXResolution;
	    }
	}
	if (*pdwYOffset + (*pdwHeight - 1) * tfxParentImg->imgInfo.dwYResolution
	    / *pdwYResolution >= tfxParentImg->imgInfo.dwHeight) {
	    if (tfxParentImg->imgInfo.dwYResolution >= *pdwYResolution) {
		*pdwHeight = (tfxParentImg->imgInfo.dwHeight - *pdwYOffset +
		    tfxParentImg->imgInfo.dwYResolution / *pdwYResolution - 1) *
		    *pdwYResolution / tfxParentImg->imgInfo.dwYResolution;
	    } else {
		*pdwHeight = ((tfxParentImg->imgInfo.dwHeight - *pdwYOffset) *
		    *pdwYResolution + tfxParentImg->imgInfo.dwYResolution - 1) /
		    tfxParentImg->imgInfo.dwYResolution;
	    }
	}
    } else {
	if (*pdwXOffset != 0) {
	    *pdwXOffset = 0;
	}
	if (*pdwYOffset != 0) {
	    *pdwYOffset = 0;
	}
    }

    /* Adjust the compression type, and strip size, if needed */
    switch ((int)*pxColorType) {
    case IO_COLORTYPE_BINARY:
	switch ((int)newCompression) {
	case IO_COMP_NOC:
	case IO_COMP_MMR:
	case IO_COMP_MR:
	case IO_COMP_MH:
	case IO_COMP_OTHER:
	    break;  /* These are valid */
	case IO_COMP_TOKENIZATION:
	case IO_COMP_LLTOKENIZATION:
	case IO_COMP_JBIG2:
	case IO_COMP_LLJBIG2:
	    /* With JBIG2, you are better off compressing the entire page. */
	    if (*pdwBestStripHeight == 0) {
		*pdwBestStripHeight = *pdwHeight;
	    }
	    break;  /* Otherwise valid */
	case IO_COMP_TRCE:
	    if (tfxPage->tfxDoc->docInfo.version < 3 /* CHANGE ME TO 4 */) {
		newCompression = IO_COMP_JPEG;   /* Only this Supported */
	    }
	    *pdwBestStripHeight = *pdwHeight;
	    break;
	default:
	    newCompression = IO_COMP_MMR; /* The default */
	}
	break;
    case IO_COLORTYPE_GRAY4:
    case IO_COLORTYPE_GRAY8:
    case IO_COLORTYPE_COLOR24:
	switch ((int)newCompression) {
	case IO_COMP_NOC:
	case IO_COMP_JPEG:
	case IO_COMP_TOKENIZATION:
	case IO_COMP_LLTOKENIZATION:
	case IO_COMP_JBIG2:
	case IO_COMP_LLJBIG2:
	case IO_COMP_OTHER:
	    break;  /* These are valid */
	default:
	    newCompression = IO_COMP_JPEG; /* The default */
	}
	break;
    case IO_COLORTYPE_COLOR4:
    case IO_COLORTYPE_COLOR8:
	switch ((int)newCompression) {
	case IO_COMP_NOC:
	case IO_COMP_OTHER:
	    break;  /* These are valid */
	default:
	    newCompression = IO_COMP_NOC; /* The default */
	}
	XERR_GEN_CLEANUP(pPalette == NULL, IO_BADPARAMETER);
	break;
    default:
	break;
    }

    if (*pdwBestStripHeight == 0) {

	/* General rule (in TIFF) is that it shouldn't be more than 8K
	 * uncompressed, but we'll use 64K for non-binary */
	switch (*pxColorType) {
	case IO_COLORTYPE_COLOR24:
	    *pdwBestStripHeight = 65536 /
		(*pdwWidth * IO_BPP(*pxColorType) / 8);

	    /* Round up to the next multiple of 16 */
	    *pdwBestStripHeight = (*pdwBestStripHeight + 15) & ~15;
	    break;
	case IO_COLORTYPE_GRAY4:
	case IO_COLORTYPE_COLOR4:
	case IO_COLORTYPE_GRAY8:
	case IO_COLORTYPE_COLOR8:
	    *pdwBestStripHeight = 65536 /
		(*pdwWidth * IO_BPP(*pxColorType) / 8);

	    /* Round up to the next multiple of 8 */
	    *pdwBestStripHeight = (*pdwBestStripHeight + 7) & ~7;
	    break;
	default:
	    /* TIFF recommends this:
	    *pdwBestStripHeight = 8192 /
		(*pdwWidth * IO_BPP(*pxColorType) / 8);
	     * but primary IFDs dictate stripe boundaries, which
	     * determine how many SubIFDs the other layer are broken
	     * into.  So I recommend something a lot higher, like: */
	    *pdwBestStripHeight = 65536 /
		(*pdwWidth * IO_BPP(*pxColorType) / 8);
	    /* Although to be quite honest, you'll get the best compression
	     * by compressing the entire image in one strip/stripe */
	}

	/* If truncated to nothing, use the full image */
	if (*pdwBestStripHeight == 0) {
	    *pdwBestStripHeight = *pdwHeight;
	}
    }

    /* Make sure the compression type is supported */
    switch (tfxPage->tfxDoc->docInfo.profile) {
    case TFX_PROFILE_S:		/* Profile S: standard B&W */
	/* This profile is not supported */
	XERR_GEN_CLEANUP(1, IO_NOSUPPORT);
	break;
    case TFX_PROFILE_F:		/* Profile F: Fax B&W */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (newCompression) {
	    case IO_COMP_MH:
	    case IO_COMP_MR:
	    case IO_COMP_MMR:
	    case IO_COMP_OTHER:
		break;	/* These are okay */
	    default:
		newCompression = IO_COMP_MMR;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_C:		/* Profile C: Contone */
	switch (*pxColorType) {
	case IO_COLORTYPE_GRAY4:
	case IO_COLORTYPE_GRAY8:
	case IO_COLORTYPE_COLOR8:
	case IO_COLORTYPE_COLOR24:
	    switch (newCompression) {
	    case IO_COMP_JPEG:
	    case IO_COMP_OTHER:
		break;	/* These are okay */
	    default:
		newCompression = IO_COMP_JPEG;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_M:		/* Profile M: MRC */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (newCompression) {
	    case IO_COMP_MH:
	    case IO_COMP_MR:
	    case IO_COMP_MMR:
	    case IO_COMP_OTHER:
		break;	/* These are okay */
	    default:
		newCompression = IO_COMP_MMR;
	    }
	    break;
	case IO_COLORTYPE_GRAY4:
	case IO_COLORTYPE_GRAY8:
	case IO_COLORTYPE_COLOR8:
	case IO_COLORTYPE_COLOR24:
	    switch (newCompression) {
	    case IO_COMP_JPEG:
	    case IO_COMP_OTHER:
		break;	/* These are okay */
	    default:
		newCompression = IO_COMP_JPEG;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_T:		/* Profile T: Tokenization (JBIG2, Binary) */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (newCompression) {
	    case IO_COMP_MH:
	    case IO_COMP_MR:
	    case IO_COMP_MMR:
	    case IO_COMP_TOKENIZATION:
	    case IO_COMP_LLTOKENIZATION:
	    case IO_COMP_JBIG2:
	    case IO_COMP_LLJBIG2:
	    case IO_COMP_OTHER:
		break;	/* These are okay */
	    default:
		newCompression = IO_COMP_MMR;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_M1:		/* Profile M1: MRC + JBIG2 */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (newCompression) {
	    case IO_COMP_MH:
	    case IO_COMP_MR:
	    case IO_COMP_MMR:
	    case IO_COMP_TOKENIZATION:
	    case IO_COMP_LLTOKENIZATION:
	    case IO_COMP_JBIG2:
	    case IO_COMP_LLJBIG2:
	    case IO_COMP_OTHER:
		break;	/* These are okay */
	    default:
		newCompression = IO_COMP_MMR;
	    }
	    break;
	case IO_COLORTYPE_GRAY4:
	case IO_COLORTYPE_GRAY8:
	case IO_COLORTYPE_COLOR8:
	case IO_COLORTYPE_COLOR24:
	    switch (newCompression) {
	    case IO_COMP_JPEG:
	    case IO_COMP_OTHER:
		break;	/* These are okay */
	    default:
		newCompression = IO_COMP_JPEG;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_COH_B:		/* Coherent TIFF Profile B: Binary */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (newCompression) {
	    case IO_COMP_NOC:
	    case IO_COMP_MH:
	    case IO_COMP_MR:
	    case IO_COMP_MMR:
	    case IO_COMP_OTHER:
		break;	/* These are okay */
	    default:
		newCompression = IO_COMP_MMR;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_COH_C:		/* Coherent TIFF Profile C: Contone */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (newCompression) {
	    case IO_COMP_NOC:
	    case IO_COMP_MH:
	    case IO_COMP_MR:
	    case IO_COMP_MMR:
	    case IO_COMP_OTHER:
		break;	/* These are okay */
	    default:
		newCompression = IO_COMP_MMR;
	    }
	    break;
	case IO_COLORTYPE_GRAY4:
	case IO_COLORTYPE_GRAY8:
	case IO_COLORTYPE_COLOR8:
	case IO_COLORTYPE_COLOR24:
	    switch (newCompression) {
	    case IO_COMP_NOC:
	    case IO_COMP_JPEG:
	    case IO_COMP_OTHER:
		break;	/* These are okay */
	    default:
		newCompression = IO_COMP_JPEG;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_COH_M:	/* Coherent TIFF Profile M: MRC */
    case TFX_PROFILE_COH_M1:	/* Coherent TIFF Profile M1: MRC + JBIG2 */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (newCompression) {
	    case IO_COMP_MH:
	    case IO_COMP_MR:
	    case IO_COMP_MMR:
	    case IO_COMP_OTHER:
		break;	/* These are okay */
	    case IO_COMP_TOKENIZATION:
	    case IO_COMP_LLTOKENIZATION:
	    case IO_COMP_JBIG2:
	    case IO_COMP_LLJBIG2:
		if (tfxPage->tfxDoc->docInfo.profile == TFX_PROFILE_COH_M) {
		    newCompression = IO_COMP_MMR;
		}
		break;	/* These are okay */
	    default:
		newCompression = IO_COMP_MMR;
	    }
	    break;
	case IO_COLORTYPE_GRAY4:
	case IO_COLORTYPE_GRAY8:
	case IO_COLORTYPE_COLOR8:
	case IO_COLORTYPE_COLOR24:
	    switch (newCompression) {
	    case IO_COMP_NOC:
	    case IO_COMP_JPEG:
	    case IO_COMP_OTHER:
		break;	/* These are okay */
	    default:
		newCompression = IO_COMP_JPEG;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    }

    /* If the compression isn't what the user asked for, make sure they
     * haven't forced it */
    if (newCompression != *pxCompression) {
	TFX_IMG *thisImg = NULL;

	/* Oh dang, we're gunna have to search through the page images to find
	 * this one.  Start with the primary image, then go through the subimages. */
	if (tfxParentImg && tfxParentImg->imgInfo.dwImgLayer == *pdwImgLayer &&
	    tfxParentImg->imgInfo.dwImgOffset == *pdwImgOffset) {
	    thisImg = tfxParentImg;
	} else {
	    thisImg = tfxPage->tfxImgs;
	    while (thisImg && thisImg->imgInfo.dwImgLayer != *pdwImgLayer &&
		thisImg->imgInfo.dwImgOffset != *pdwImgOffset) {
		thisImg = thisImg->next;
	    }
	}

	/* Use the recommended compression if:
	 * 1 - there is no requested compression
	 * 2 - the requested compression type is non-compliant AND
	 *     this isn't an image that has already been read (can't
	 *     change it if it's already something else)
	 */
	if (*pxCompression == IO_COMP_UNKNOWN ||
	    (*pxCompression != IO_COMP_OTHER &&
	    ! (thisImg && thisImg->fields.found.Compression))) {
	    *pxCompression = newCompression;
	}
    }

    /* Check the page STRIPES.  Note, if the Primary IFD is broken into TIFF
     * strips, then those strips dictate stripes across the page.  Those stripes
     * are artificial boundaries that no SubIFD may extend beyond.  However, due
     * to resolution differences, it's possible for a perimeter pixel to straddle
     * the stripe boundary.
     *
     * The exception to the stripe rules are TIFF-FX M E1.5 and XCTIFF M profiles,
     * which allow a single SubIFD to span multiple stripes, as long as the strips
     * line-up with the stripes EXACTLY!
     */
    if (tfxParentImg && (tfxParentImg->imgInfo.dwImgLayer != *pdwImgLayer
	|| tfxParentImg->imgInfo.dwImgOffset != *pdwImgOffset)) {
	Int32 scale = tfxParentImg->imgInfo.dwYResolution / *pdwYResolution;
	UInt32 firstStripeOffset, lastStripeOffset;
	UInt32 primaryRow;
	UInt32 startOfFirstStripe, firstStripeRows;
	UInt32 startOfLastStripe, lastStripeRows;

	/* Scale should be positive, and greater than one.  This equates to:
	 * the Primary IFD must the highest resolution. */
	XERR_GEN_CLEANUP(scale < 1, IO_BADFILEFORMAT);

	/* The image is also supposed to be an integral multiple of the
	 * Primary IFD resolution.  Just produce a warning if not compliant. */
	if (tfxParentImg->imgInfo.dwYResolution % *pdwYResolution != 0) {
	    warning = IO_WARN_NOTCOMPLIANT;
	}

	/* Compute the stripe offset where this image begins (the Primary
	 * IFD scanline where the first scanline ends), and the stripe
	 * offset where this image ends (the stripe where the last scanline
	 * starts, even though it may hang off the end).
	 */
	primaryRow = *pdwYOffset + scale - 1;
	XERR_PASS_CLEANUP(TFX_GetStripeInfo((IO_PAGEHANDLE)tfxPage,
	    primaryRow, &firstStripeOffset, &startOfFirstStripe,
	    &firstStripeRows));
	primaryRow = *pdwYOffset + *pdwHeight * scale - scale + 1;
	XERR_PASS_CLEANUP(TFX_GetStripeInfo((IO_PAGEHANDLE)tfxPage,
	    primaryRow, &lastStripeOffset, &startOfLastStripe,
	    &lastStripeRows));

	/* If different stripe, and we're doing a write here (can't dispute
	 * what's there if we're reading), then see if we should adjust the
	 * height */
	if (firstStripeOffset != lastStripeOffset &&
	    (tfxPage->tfxDoc->readWrite & WRITE_ALLOWED)) {
	    UInt32 newHeight;

	    /* See if we fall into the category of a single SubIFD that can span
	     * all stripes, with strips aligning to the stripes */
	    switch (tfxPage->tfxDoc->docInfo.profile) {
	    case TFX_PROFILE_M1:
	    case TFX_PROFILE_COH_M:
	    case TFX_PROFILE_COH_M1:
		/* Yep.
		 * We will have to wait to decide whether variable-sized strips
		 * are to be used until ImageWriteStart is called. */
		if (tfxPage->seqStatus & TFX_PAGE_SEQ_MULTISTRIPSUBIFDS) {
		    multiStripSubIFD = 1;

		    /* For now, assume the strips line-up with the primary IFD
		     */
		    *pdwBestStripHeight = ((firstStripeRows - *pdwYOffset) + scale - 1) / scale;
		    break;
		}
		/* If not satisfied, then we must use single-strip IFDs,
		 * possibly more than one. */
	    default:

		/* Compute the height allowed in order to keep the image within
		 * the first stripe */
		if (firstStripeOffset == lastStripeOffset) {
		    /* It's the last stripe */
		    newHeight = (tfxParentImg->imgInfo.dwHeight -
			*pdwYOffset + scale - 1) / scale;
		} else {
		    UInt32 prePixels = 0;
		    if (*pdwYOffset < startOfFirstStripe) {
			prePixels = startOfFirstStripe - *pdwYOffset;
		    }
		    newHeight = ((startOfFirstStripe + firstStripeRows)
			- *pdwYOffset + scale - 1 - prePixels) / scale;
		}

		/* If the image is too big, then the computation above must be wrong */
		XERR_GEN_CLEANUP(*pdwYOffset + newHeight * scale - scale + 1 > tfxParentImg->imgInfo.dwHeight,
		    IO_INTERNAL_ERROR);

		if (newHeight < *pdwHeight) {
		    *pdwHeight = newHeight;
		}
	    }
	}
    }

    /* If the non-primary image isn't multi-strip, it must be a single strip */
    if (! thisIsPrimaryImage && ! multiStripSubIFD) {
	*pdwBestStripHeight = *pdwHeight;
    }

    /* Always limit strips to the image height */
    if (*pdwBestStripHeight > *pdwHeight) {
	*pdwBestStripHeight = *pdwHeight;
    }

    /* Check the color space */
    switch (tfxPage->tfxDoc->docInfo.profile) {
    case TFX_PROFILE_S:		/* Profile S: standard B&W */
	/* This profile is not supported */
	XERR_GEN_CLEANUP(1, IO_NOSUPPORT);	/* This is a serious problem */
	break;
    case TFX_PROFILE_F:		/* Profile F: Fax B&W */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (*pxColorSpace) {
	    case IO_COLORSPACE_WHITEISZERO:
	    case IO_COLORSPACE_BLACKISZERO:
		break;	/* These are okay */
	    default:
		*pxColorSpace = IO_COLORSPACE_WHITEISZERO;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_C:		/* Profile C: Contone */
	switch (*pxColorType) {
	case IO_COLORTYPE_GRAY4:
	case IO_COLORTYPE_GRAY8:
	case IO_COLORTYPE_COLOR8:
	case IO_COLORTYPE_COLOR24:
	    switch (*pxColorSpace) {
	    case IO_COLORSPACE_YCBCR:
	    case IO_COLORSPACE_CIELAB:
	    case IO_COLORSPACE_ICCLAB:
		warning = IO_WARN_NOTCOMPLIANT;
		break;	/* These are non-compliant */
	    case IO_COLORSPACE_ITULAB:
		break;	/* These are okay */
	    default:
		*pxColorSpace = IO_COLORSPACE_ITULAB;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_M:		/* Profile M: MRC */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (*pxColorSpace) {
	    case IO_COLORSPACE_WHITEISZERO:
	    case IO_COLORSPACE_BLACKISZERO:
		break;	/* These are okay */
	    default:
		*pxColorSpace = IO_COLORSPACE_WHITEISZERO;
	    }
	    break;
	case IO_COLORTYPE_GRAY4:
	case IO_COLORTYPE_GRAY8:
	case IO_COLORTYPE_COLOR8:
	case IO_COLORTYPE_COLOR24:
	    switch (*pxColorSpace) {
	    case IO_COLORSPACE_YCBCR:
	    case IO_COLORSPACE_CIELAB:
	    case IO_COLORSPACE_ICCLAB:
		warning = IO_WARN_NOTCOMPLIANT;
		break;	/* These are non-compliant */
	    case IO_COLORSPACE_ITULAB:
		break;	/* These are okay */
	    default:
		*pxColorSpace = IO_COLORSPACE_ITULAB;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_T:		/* Profile T: Tokenization (JBIG2, Binary) */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (*pxColorSpace) {
	    case IO_COLORSPACE_WHITEISZERO:
	    case IO_COLORSPACE_BLACKISZERO:
		break;	/* These are okay */
	    default:
		*pxColorSpace = IO_COLORSPACE_WHITEISZERO;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_M1:		/* Profile M1: MRC + JBIG2 */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (*pxColorSpace) {
	    case IO_COLORSPACE_WHITEISZERO:
	    case IO_COLORSPACE_BLACKISZERO:
		break;	/* These are okay */
	    default:
		*pxColorSpace = IO_COLORSPACE_WHITEISZERO;
	    }
	    break;
	case IO_COLORTYPE_GRAY4:
	case IO_COLORTYPE_GRAY8:
	case IO_COLORTYPE_COLOR8:
	case IO_COLORTYPE_COLOR24:
	    switch (*pxColorSpace) {
	    case IO_COLORSPACE_YCBCR:
	    case IO_COLORSPACE_CIELAB:
	    case IO_COLORSPACE_ICCLAB:
		warning = IO_WARN_NOTCOMPLIANT;
		break;	/* These are non-compliant */
	    case IO_COLORSPACE_ITULAB:
		break;	/* These are okay */
	    default:
		*pxColorSpace = IO_COLORSPACE_ITULAB;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_COH_B:		/* Coherent TIFF Profile B: Binary */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (*pxColorSpace) {
	    case IO_COLORSPACE_WHITEISZERO:
	    case IO_COLORSPACE_BLACKISZERO:
		break;	/* These are okay */
	    default:
		*pxColorSpace = IO_COLORSPACE_WHITEISZERO;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_COH_C:		/* Coherent TIFF Profile C: Contone */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (*pxColorSpace) {
	    case IO_COLORSPACE_WHITEISZERO:
	    case IO_COLORSPACE_BLACKISZERO:
		break;	/* These are okay */
	    default:
		*pxColorSpace = IO_COLORSPACE_WHITEISZERO;
	    }
	    break;
	case IO_COLORTYPE_GRAY4:
	case IO_COLORTYPE_GRAY8:
	case IO_COLORTYPE_COLOR8:
	case IO_COLORTYPE_COLOR24:
	    switch (*pxColorSpace) {
	    case IO_COLORSPACE_RGB:
	    case IO_COLORSPACE_CMYK:
	    case IO_COLORSPACE_YCBCR:
	    case IO_COLORSPACE_CIELAB:
	    case IO_COLORSPACE_ICCLAB:
	    case IO_COLORSPACE_ITULAB:
		break;	/* These are okay */
	    default:
		*pxColorSpace = IO_COLORSPACE_ITULAB;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    case TFX_PROFILE_COH_M:	/* Coherent TIFF Profile M: MRC */
    case TFX_PROFILE_COH_M1:	/* Coherent TIFF Profile M1: MRC + JBIG2 */
	switch (*pxColorType) {
	case IO_COLORTYPE_BINARY:
	    switch (*pxColorSpace) {
	    case IO_COLORSPACE_WHITEISZERO:
	    case IO_COLORSPACE_BLACKISZERO:
		break;	/* These are okay */
	    default:
		*pxColorSpace = IO_COLORSPACE_WHITEISZERO;
	    }
	    break;
	case IO_COLORTYPE_GRAY4:
	case IO_COLORTYPE_GRAY8:
	case IO_COLORTYPE_COLOR8:
	case IO_COLORTYPE_COLOR24:
	    switch (*pxColorSpace) {
	    case IO_COLORSPACE_YCBCR:
	    case IO_COLORSPACE_ICCLAB:
	    case IO_COLORSPACE_ITULAB:
		break;	/* These are okay */
	    default:
		*pxColorSpace = IO_COLORSPACE_ITULAB;
	    }
	    break;
	default:
	    /* This image is not supported in this profile */
	    XERR_GEN_CLEANUP(1, IO_WARN_NOSUPPORT);
	}
	break;
    }

    if (warning)
	__error__ = warning;
    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Use multi-strip SubIFD feature
 *
 * LongDescription:::
 *  IO_UseMultiStripSubIFDs() requests that the document use single multi-strip
 *  SubIFDs, when adding images that span across stripe boundaries.
 *  If images in the page have already been added, then they are not affected.
 *  Only images added to the page after this call will be affected.<BR>
 *  There is currently no way of turning this feature off for a page, once it
 *  has been turned on.
 *
 * Input:::
 *  ioPageHandle::		Valid page handle, as returned from
 *				IO_AddPage()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_UseMultiStripSubIFDs(
	IO_PAGEHANDLE tfxPageHandle
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    PROLOGUE;

    TFX_CheckHandle(tfxPage);

    tfxPage->seqStatus |= TFX_PAGE_SEQ_MULTISTRIPSUBIFDS;

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Do not write the GPIFD field
 *
 * LongDescription:::
 *  TFX_UseNoGPIFD() requests that the document NOT have the GPIFD written.
 *  If the first IFD has yet to be written, and there's no need for a GPIFD,
 *  then it won't be written.  If later the GPIFD is required, the IFD of the
 *  first page will be rewritten with the GPIFD included.<BR>
 *  If the first IFD is already written, then the GPIFD is NOT removed from it.
 *
 * Input:::
 *  tfxDoc::			Valid document handle
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_UseNoGPIFD(
	IO_DOCHANDLE tfxDocHandle
)
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    PROLOGUE;

    TFX_CheckHandle(tfxDoc);

    tfxDoc->seqStatus |= TFX_DOC_SEQ_NOGPIFDREQ;

    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Use variable-sized strips
 *
 * LongDescription:::
 *  TFX_UseVariableStripHeights() sets the output image into a mode where
 *  variable-sized strips are used (StripRowCounts vs. RowsPerStrip).
 *
 * Input:::
 *  tfxImg::			Valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_UseVariableStripHeights(
		   IO_IMGHANDLE tfxImgHandle
		   )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    TFX_CheckHandle(tfxImg);

    /* Make sure we haven't already written the IFD */
    XERR_GEN(tfxImg->seqStatus & TFX_IMG_SEQ_WRITESTART, IO_SEQUENCE);

    tfxImg->seqStatus |= TFX_IMG_SEQ_USEVSS;

    NORMAL_RETURN;
}


/*
$Log: tfxread.c,v $
Revision 1.3  2003/02/11 01:31:42  mkrocken
fixed problems with comments caused by previous commit

Revision 1.2  2003/02/10 22:53:31  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:11  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.45  2002/02/06 19:28:45  dabercro
Use the page count from the PageNumber in the first IFD, if non-zero.
Otherwise, traverse the IFD chain to obtain the page count.

Revision 1.44  2001/12/26 21:23:44  dabercro
Mark as released

Revision 1.43  2001/11/07 21:44:16  dabercro
Make base color signed long, to accomodate CIELAB.
Refined NODGP code carve-out.

Revision 1.42  2001/10/30 20:05:07  dabercro
Add in writing of MH and MR images.

Revision 1.41  2001/09/18 21:10:11  dabercro
Make work with JBIG2

Revision 1.40  2001/08/22 19:51:04  dabercro
Added support for new field: IFDBaseColor.
Improved logic for profile detection.

Revision 1.39  2001/07/20 23:16:37  dabercro
Don't adjust the height for single-strip SubIFDs if we're reading (we can't
dispute what's already there).
Automatically account for the JPEGTables data in the compressed data
stream.

Revision 1.38  2001/06/27 18:57:03  dabercro
Borrow flag names from IO_ layer, added uncompressed support.

Revision 1.37  2001/06/01 18:10:54  dabercro
Remove unused parameter check.

Revision 1.36  2001/05/15 20:08:02  dabercro
Check for old header byte discrepency when we know it doesn't include
colour tags.

Revision 1.35  2001/05/10 18:52:00  dabercro
Expect the byte count of the position block to exactly fill the TIFF stream.
One test file failed the former check (wave.xif).

Revision 1.34  2001/05/04 23:13:03  dabercro
Expand the "old header" check by examining the pos block byte count also

Revision 1.33  2001/05/03 16:43:10  dabercro
Added a new function UseMultiStripSubIFDs.  Rethink logic about
automatically generating multi-strip SubIFDs.

Revision 1.32  2001/05/01 17:12:04  dabercro
Fix pos byte extraction.

Revision 1.31  2001/05/01 00:55:59  dabercro
Add UseNoGPIFD.  Start implementation for multi-strip SubIFDs.
Fix JBIG2 image strip header to relocate position block byte count.

Revision 1.30  2001/04/02 18:25:03  dabercro
Add HostComputer.

Revision 1.29  2001/03/27 22:15:44  dabercro
Allow a position block with 0 dictionaries.

Revision 1.28  2001/03/22 20:17:39  dabercro
Add support for new TIFF field: Predictor.

Revision 1.27  2001/03/22 17:49:47  dabercro
Use default in case statement.

Revision 1.26  2001/03/13 02:59:12  dabercro
Remove lossy-dither and wavelet as valid compression types.

Revision 1.25  2001/03/13 02:18:27  dabercro
Only reinsert GPIFD when appropriate profile.
Return NOSUPPORT instead of BADFILEFORM when checking
resolution, colortype, etc.

Revision 1.24  2001/01/19 23:56:08  dabercro
Include 4-bit gray as valid.
Avoid redifinition of NODGP.

Revision 1.23  2001/01/18 00:21:30  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.22  2000/12/11 22:21:13  dabercro
Fix wording in comments.

Revision 1.21  2000/11/18 00:53:26  dabercro
Avoid cross polenation of data handles across sublibraries
(XIF, MRC, TFX), especially for CopyImage.  Added a handle
type to structure, now checked also.

Revision 1.20  2000/11/17 23:57:35  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.19  2000/11/02 22:21:21  dabercro
Additions for Rendering Hints.

Revision 1.18  2000/10/16 18:22:53  dabercro
Major adds to support stripes, 0-byte strips, and ImageBaseColor.

Revision 1.17  2000/09/27 21:13:21  dabercro
Updates to handle strips and stripes.  Fix to layer strip reading.

Revision 1.16  2000/09/21 21:19:14  dabercro
Add variable-sized strips: StripRowCounts.

Revision 1.15  2000/09/15 22:11:37  dabercro
Use consistent definitions, from iomrc.h.

Revision 1.14  2000/08/30 23:28:19  dabercro
Check for both JBIG2 and Tokenization.

Revision 1.13  2000/08/29 22:35:19  dabercro
Allow non-compliance is some case, but still continue.
Disable turning off the multi-page bit of NewSubfileType.

Revision 1.12  2000/08/25 23:49:29  dabercro
Annotations must be marked for write before the IFD is written,
otherwise they're not.
Mark as tokenized profile only if shared resource table is really
there (not just the field).

Revision 1.11  2000/08/22 01:03:23  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.10  2000/08/02 00:21:40  dabercro
Complain about 24-bit uncompressed data as being
non-compliant for TIFF-FX.

Revision 1.9  2000/08/01 00:01:41  dabercro
Rename JBIG2Options to T88Options.
When merging images in the same layer, compare palettes.
Updated support for JBIG1 profiles.

Revision 1.8  2000/06/07 00:00:16  dabercro
Continued work, lots of stuff.

Revision 1.7  2000/05/10 20:20:25  dabercro
Comment fix

Revision 1.6  2000/05/09 01:52:58  dabercro
Added FillOrder field support.
Updated JBIG2 dictionary front-end.
Support for JBIG2 dictionary extensions, updated parameter to
reflect byte count for extensions.
Write fields that are explicitly set, regardless of default values.
Check if document is single-page when closing, reflect in
NewSubfileType.

Revision 1.5  2000/04/12 19:06:09  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.4  2000/03/11 02:06:34  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.3  2000/02/29 00:19:42  dabercro
Incremental improvements, profile distinctions,etc.

Revision 1.2  2000/02/08 21:25:33  dabercro
Disable JBIG2 (for now), updates to make use of GlobalParametersIFD for symbol libraries.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.45  1998/10/28 20:36:04  aber
aber on Wed Oct 28 12:04:08 PST 1998

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Enable compression of 8-bit gray using Wavelet compression.

Revision 1.44  1998/10/20 18:17:59  aber
aber on Tue Oct 20 11:17:04 PDT 1998

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

When checking for compression consistency, allow lossy dither compression
for binary images (forgot to include it in last checkin).

Revision 1.43  1998/08/28 20:30:57  aber
tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Detect the inconsistency of the unsupported color tokenization, and
return a NOSUPPORT error for tokenized images that aren't binary. This
was practically impossible to anticipate, but should be gracefully
handled in all future releases.

Revision 1.42  1998/08/04 00:21:32  aber
tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Support multiple dictionaries when token compressing, used to retain
single-use tokens in the same dictionary -- to be used when color
tokenizing.

Revision 1.41  1998/07/30 22:39:42  aber
tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Prototype updates for signed-ness.

Revision 1.40  1998/07/27 19:05:53  aber
aber on Mon Jul 27 12:04:31 PDT 1998

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Temporary fix for stuff I'm about to check in.

Revision 1.39  1998/07/27 17:36:16  aber
aber on Mon Jul 27 10:35:43 PDT 1998

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Use private memory tally structure.

Revision 1.38  1998/05/05 07:15:21  aber
aber on Tue May 5 00:15:21 PDT 1998

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Additions necessary to fill in the never completed functions
TFX_Compress and TFX_Decompress.

Revision 1.37  1998/04/06 22:43:45  aber
aber on Mon Apr 6 15:43:44 PDT 1998

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Bug fix: check resolution difference when checking the image size of
the image during read.
Also, check for unanticipated condition of the page resolution being
SMALLER than than the subimage resolution. Adjust subimage dimensions
accordingly.

Revision 1.36  1998/03/25 23:25:14  aber
aber on Wed Mar 25 15:25:13 PST 1998

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Use a unique document ID for each document handle. This allows tagging
each document so that during "quick copy", the token dictionaries are
copy only once, and only once (in the case of two destination documents
somehow ending up with the same memory location for the structure data).

Revision 1.35  1998/02/09 20:21:23  aber
aber on Mon Feb 9 12:21:23 PST 1998

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Signed/unsigned casting to make MSVC happy.

Revision 1.34  1998/01/20 03:20:56  aber
aber on Mon Jan 19 19:18:52 PST 1998

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Before reading a tokenized image, make sure it's flushed to disk.

Revision 1.33  1998/01/16 00:39:45  aber
aber on Thu Jan 15 16:38:01 PST 1998

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Allow page inserts.

Revision 1.32  1998/01/14 00:16:25  ddavies
tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.31  1997/11/24 21:06:32  aber
aber on Mon Nov 24 13:04:42 PST 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Change calling sequence requirements for GetAnnotationData
to have previously gotten the page parameters.
(Was requiring to have been in read mode, but is a valid operation
just after writing).

Revision 1.30  1997/11/19 21:42:10  aber
aber on Wed Nov 19 13:41:32 PST 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Allow getting annotation info while writing.

Revision 1.29  1997/11/18 23:37:01  aber
aber on Tue Nov 18 15:36:16 PST 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Add new function: TFX_WriteTokenImages() -- which will complete
the tokenization process, write the token dictionary, and write
all token images which haven't been stored yet.
Add annotation table element: flags.
Add annotation header element: bytes per element.

Revision 1.28  1997/11/17 19:25:58  aber
aber on Mon Nov 17 11:25:58 PST 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Allow negative page offsets to automatically append the page to
the end of the document.

Revision 1.27  1997/11/14 02:08:57  aber
aber on Thu Nov 13 18:08:57 PST 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Use signed value for reference page; clarify that it's a page offset,
not an ordinal. Store as ordinal (was only reading as ordinal).

Revision 1.26  1997/11/01 20:49:08  aber
aber on Sat Nov 1 12:48:29 PST 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Force buffering strips exactly equal to TIFF strip size -- allows
exact buffer for wavelet strip.

Revision 1.25  1997/10/24 21:38:54  aber
aber on Fri Oct 24 14:38:53 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: More exhaustive checking on calling sequence.

Revision 1.24  1997/10/23 18:53:44  aber
aber on Thu Oct 23 11:53:18 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

> should be >= when checking for a valid page offset.

Revision 1.23  1997/10/16 16:15:47  aber
aber on Thu Oct 16 09:15:47 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Remove carriage-returns (left over from PC).

Revision 1.22  1997/10/15 22:14:13  aber
aber on Wed Oct 15 15:12:07 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Store the page table in TIFF-FX 2 file, if no pages have been saved.

Revision 1.21  1997/10/14 17:01:44  aber
aber on Tue Oct 14 10:00:57 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Assumption made about the number of dictionaries referenced by
a single page was wrong. Assumption removed.

Revision 1.20  1997/10/01 23:35:04  aber
aber on Wed Oct 1 16:34:06 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Simple cleanup, doesn't affect anything relevant.

Revision 1.19  1997/08/29 17:27:59  aber
aber on Fri Aug 29 10:27:59 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Adjust parameter types to match what's actually stored in the file.

Revision 1.18  1997/08/13 22:48:18  aber
aber on Wed Aug 13 15:48:18 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Edit fix via Pedro (XResolution should be Yresolution).

Revision 1.17  1997/08/13 22:18:44  aber
aber on Wed Aug 13 15:18:32 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Allow page info query when writing.

Revision 1.16  1997/08/07 21:27:43  aber
aber on Thu Aug 7 14:27:43 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Make provisions for a 4-byte dictionary block size, instead of 2-bytes
spelled out in Williams table. Also make provisions for multiple
dictionaries per page.

Revision 1.15  1997/07/29 23:49:41  aber
aber on Tue Jul 29 16:47:16 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Add function TFX_GetInstance(), to retrieve the TIFF-FX instance associated
with a document handle. Used to detach the instance when closing the
document.
Added function TFX_GetImageStats(), which retrieves the total compressed
size of the image, and the size of the token dictionary, if tokenized.
Allow requests for compressed data size without calling TFX_ImageReadStart().

Revision 1.14  1997/07/15 23:42:09  aber
aber on Tue Jul 15 16:41:19 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Some buffering fixes.

Revision 1.13  1997/07/02 01:03:59  aber
aber on Tue Jul 1 18:01:39 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Make memory functions match IPWrap. Expand the return values for
IsTfx to indicate: no support, not a TIFF, not a TIFF-FX, and read/seek error.
Complete generic function TFX_ImageReadLayer(), which buffers a PIXR and
calls w_TfxImageReadLayer().

Revision 1.12  1997/06/20 17:43:37  aber
aber on Fri Jun 20 10:43:12 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

More checking for corrupt TIFF-FX files.

Revision 1.11  1997/06/18 00:41:28  aber
tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Updated prototype for TFX_GetLayerInfo.

Revision 1.10  1997/06/17 01:29:15  aber
tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: TIFF-FX 2.0 support. Annotation bug fix.

Revision 1.9  1997/06/12 00:49:42  aber
aber on Wed Jun 11 17:49:11 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

Fix row check for strip alignment.

Revision 1.8  1997/06/10 01:26:37  aber
tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.6  1997/05/30 22:11:31  aber
aber on Fri May 30 15:11:31 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Update documentation for web browsing. TFX_ functions are left
with no "BeginDocumentation" tag, since w_ and wp_ functions are expected.

Revision 1.5  1997/05/30 00:50:58  aber
tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:02  aber
tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:09  aber
aber on Mon May 19 17:54:09 PDT 1997

tfxread.c -> /product/ipcore/ipwrap/src/RCS/tfxread.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:34  aber
Initial revision

*/
