/*****************************************************************
 *  Trade Secret of Xerox Corporation.                           *
 *  Copyright (c) 1991-2001, Xerox Corporation.                  *
 *  All rights reserved.                                         *
 *  Copyright protection claimed includes all forms and matters  *
 *  of copyrightable material and information now allowed by     *
 *  statutory or judicial law or hereafter granted, including    *
 *  without limitation, material generated from the software     *
 *  programs which are displayed on the screen such as icons,    *
 *  screen display looks, etc.                                   *
 *****************************************************************/


/************************************************************************
 * tfxwrite.c
 *
 *	TIFF-FX write of TIFF-FX specific "stuff"
 *
 * Functions defined:
 *
 *
 * Notes:
 *  Comments to be added later.
 ************************************************************************/

/*	tfxwrite.c:	TIFF-FX WRITE source
 *
 *	TIFF-FX write-only routines
 *
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifndef NOIPCORE
#if defined(__GNUC__) || defined(_WIN32)
#include "ansiprot.h"
#endif

IP_RCSINFO(RCSInfo, "$RCSfile: tfxwrite.c,v $; $Revision: 1.3 $; $Date: 2003/02/11 01:31:43 $");

/* IPCore */
#ifdef _MSC_VER
#pragma pack(push, PRE_TIFFFX, 1)
#endif
#include "ipwcore.prv"
#include "defines.pub"
#include "geomadt.h"
#include "props.pub"
#include "frames.pub"
#include "ccitt.pub"
#include "mmr.pub"
#include "shrcmprs.prv"
#include "imageref.pub"
#include "shrpixr.pub"
#include "shrpixr.prv"
#ifdef _MSC_VER
#pragma pack(pop, PRE_TIFFFX)
#endif
#endif	/* NOIPCORE */

#include "tfx.h"
#include "tfxfile.h"
#include "tfximage.h"
#include "tfxwtiff.h"
#include "tfxwtfx.h"
#include "tfxwpriv.h"
#include "tfxrtfx.h"
#include "tfxwpub.h"
#include "tfxrpriv.h"
#include "tfxconv.h"
#include "tfxrjb2.h"

#ifndef NOIPCORE
#include "tfx.pub"
#include "tfxwjb2.h"

#else	/* NOIPCORE */
#ifndef NODGP
#define NODGP
#endif
#endif	/* NOIPCORE */

#ifndef NODGP
#define _IPWSTATE_DEFINED_  /* Avoid duplicate definition */
#include "jtokapi.h"
#endif	/* NODGP */


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Create a new TIFF-FX document
 *
 * LongDescription:::
 *  TFX_CreateDocument() creates a new TIFF-FX document, and prepares to write
 *  data to it. The client must have already created the file at the operating
 *  system level, so that the read, write, and seek methods are available.
 *  <BR>
 *  A handle representing the document is passed back through a parameter.
 *  This handle is passed in to all subsequent TFX_  function calls, and will
 *  encompass the IO_INSTHANDLE passed in here.
 *
 * Input:::
 *  xInstHandle::		TIFF-FX instance handle, returned from
 *				TFX_InitInstance()
 *  dwClientID::		Client ID, passed back through all I/O funcs
 *  profile::			Profile of TIFF-FX to create (see tfx.h)
 *  version::			Version of TIFF-FX document to create
 *  options::			File create options
 *  xReadFunc::			Client file read function
 *  xWriteFunc::		Client file write function
 *  xSeekFunc::			Client file seek function
 *  xSizeFunc::			Client file size function
 *  xTellFunc::			Client file tell function
 *
 * Output:::
 *  pTfxDocHandle::		Pointer to document handle
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_CreateDocument(
	IO_INSTHANDLE xInstHandle,
	Int32 dwClientID,
	UInt8 version,
	UInt32 options,
	IO_READFUNC xReadFunc,
	IO_WRITEFUNC xWriteFunc,
	IO_SEEKFUNC xSeekFunc,
	IO_SIZEFUNC xSizeFunc,
	IO_TELLFUNC xTellFunc,
	IO_DOCHANDLE *pTfxDocHandle
)
{
    TFX_INSTANCE *instData = (TFX_INSTANCE *)xInstHandle;
    TFX_DOC *tfxDoc = NULL;
    TFX_FILE *tfxFile = NULL;
    UInt8 revision;
    PROLOGUE;
    
    /* Clear the results for now */
    *pTfxDocHandle = 0UL;
    
    TFX_CheckHandle(xInstHandle);
    
    /* Check for proper calling sequence */

    
    /* Create a new document (handle) */
    IO_CHECKMALLOC(tfxDoc = (TFX_DOC *)TFX_MALLOC(sizeof(TFX_DOC)));
    memset(tfxDoc, 0, sizeof(*tfxDoc));
    tfxDoc->handleID = (UInt32)tfxDoc;	/* Self identity */
    tfxDoc->handleType = TFX_HANDLE_TYPE;
    tfxDoc->docID = ++instData->nextDocID;/* ID unique to this library instance
					 */
    
    /* Create a link back to the instance */
    IO_CHECKMALLOC(tfxFile = (TFX_FILE *)TFX_MALLOC(sizeof(TFX_FILE)));
    memset(tfxFile, 0, sizeof(*tfxFile));
    tfxFile->handleID = (UInt32)tfxFile;	/* Self identity */
    
    /* Setup the file handle */
    tfxFile->instance = instData;
    TFX_FILETOKEN(tfxFile)->dwClientFileID = dwClientID;
    TFX_FILETOKEN(tfxFile)->FileRead = xReadFunc;
    TFX_FILETOKEN(tfxFile)->FileWrite = xWriteFunc;
    TFX_FILETOKEN(tfxFile)->FileSeek = xSeekFunc;
    TFX_FILETOKEN(tfxFile)->FileSize = xSizeFunc;
    TFX_FILETOKEN(tfxFile)->FileTell = xTellFunc;
    
    /* Retain the file handle */
    tfxDoc->tfxFile = tfxFile;
    
    /* Wait until I have full control over the file */
    XERR_PASS_CLEANUP(wp_TfxFileLock(TFX_FILETOKEN(tfxFile)));
    
    /* Create the TIFF-FX header */
    switch (version) {
    case 7:
	revision = 0;
	break;
    default:
	version = CURRENT_TFX_VERSION;
	revision = CURRENT_TFX_REVISION;
	break;
    }
    XERR_PASS_CLEANUP(wp_TfxWriteHeader(tfxDoc, &version, &revision));

    /* Initialize */
    tfxDoc->seqStatus |= TFX_DOC_SEQ_WRITESTART;
    tfxDoc->readWrite = WRITE_ALLOWED;
    tfxDoc->docInfo.version = version;
    tfxDoc->docInfo.revision = revision;
    tfxDoc->docInfo.profile = (UInt8)(options & TFX_PROFILES);
    
    /* We could store a page table now with the following call:
     *     XERR_PASS_CLEANUP(wp_TfxSetPageIFD(tfxDoc, 0UL, 0, 0UL, NULL));
     * but instead the page table is added when the first page is written,
     * or when the document is closed (in the case of TIFF-FX 3 only, with 0 pages).
     * The reason it's not done here, is that TIFF-FX 3+ files will not have
     * a page table attached when no pages exist (more compact), and the
     * number of page table entries per table is allowed to be set up until
     * the first page is added (see TFX_SetPageTableEntryCount()).
     */
    
    /* Return the result */
    *pTfxDocHandle = (UInt32)tfxDoc;
    
    XERR_PASS_CLEANUP(wp_TfxFileUnLock(TFX_FILETOKEN(tfxFile)));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    wp_TfxFileUnLock(TFX_FILETOKEN(tfxFile));
    if (tfxDoc) {
	TFX_FREE(tfxDoc);
    }
    if (tfxFile) {
	TFX_FREE(tfxFile);
    }
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Open a document for read and write
 *
 * LongDescription:::
 *  TFX_OpenDocumentReadWrite() begins a session of reading and writing
 *  an existing document. The client must already have opened the file at
 *  the operating system level, so that the read and seek methods in the
 *  file token are available.
 *  This function will perform any initialization required for reading
 *  the file, including reading enough header information to verify
 *  that the file is in a format supported by this version of
 *  the TIFF-FX library.
 *  <BR>
 *  A document handle is returned through the third parameter. This handle
 *  is passed to all subsequent function calls that use the document, and
 *  will encompass the I/O functions passed in here.
 *
 * Input:::
 *  xInstHandle::		TIFF-FX instance handle, returned from
 *				TFX_InitInstance()
 *  dwClientID::		Input client ID
 *  version::			Version of TIFF-FX document to create
 *  options::			File create options
 *  xReadFunc::			Client file read function
 *  xWriteFunc::		Client file write function
 *  xSeekFunc::			Client file seek function
 *  xSizeFunc::			Client file size function
 *  xTellFunc::			Client file tell function
 *
 * Output:::
 *  pTfxDocHandle::		Document handle used to refer to this
 *				page in other functions
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_OpenDocumentReadWrite(
	IO_INSTHANDLE xInstHandle,
	Int32 dwClientID,
	UInt8 version,
	UInt32 options,
	IO_READFUNC xReadFunc,
	IO_WRITEFUNC xWriteFunc,
	IO_SEEKFUNC xSeekFunc,
	IO_SIZEFUNC xSizeFunc,
	IO_SIZEFUNC xTellFunc,
	IO_DOCHANDLE *pTfxDocHandle
)
{
    TFX_INSTANCE *instData = (TFX_INSTANCE *)xInstHandle;
    TFX_DOC *tfxDoc;
    TFX_FILE tfxFile;
    UInt8 revision;
    UInt16 nPages;
    Int32 fileSize = 0;
    PROLOGUE;

    /* Clear results for now */
    *pTfxDocHandle = 0UL;

    TFX_CheckHandle(xInstHandle);

    /* Create a link back to the instance */
    memset(&tfxFile, 0, sizeof(TFX_FILE));
    tfxFile.handleID = (UInt32)&tfxFile;	/* Self identity */

    /* Setup the file handle */
    tfxFile.instance = instData;
    TFX_FILETOKEN(&tfxFile)->dwClientFileID = dwClientID;
    TFX_FILETOKEN(&tfxFile)->FileRead = xReadFunc;
    TFX_FILETOKEN(&tfxFile)->FileWrite = xWriteFunc;
    TFX_FILETOKEN(&tfxFile)->FileSeek = xSeekFunc;
    TFX_FILETOKEN(&tfxFile)->FileSize = xSizeFunc;
    TFX_FILETOKEN(&tfxFile)->FileTell = xTellFunc;
    
    /* Check the file size, if it's zero, create the document */
    TFX_FILESIZE(&tfxFile, fileSize);
    if (fileSize == 0) {
	XERR_PASS_CLEANUP(TFX_CreateDocument(xInstHandle, dwClientID,
	    version, options, xReadFunc, xWriteFunc,
	    xSeekFunc, xSizeFunc, xTellFunc, pTfxDocHandle));
    } else {

	/* Open the document read-only */
	XERR_PASS_CLEANUP(TFX_OpenDocumentRead(xInstHandle, dwClientID,
	    options, xReadFunc, xWriteFunc, xSeekFunc, xSizeFunc,
	    xTellFunc, pTfxDocHandle));
    }

    /* Reference the document */
    tfxDoc = (TFX_DOC *)*pTfxDocHandle;

    /* Change the status from read to read/write */
    tfxDoc->readWrite |= READ_ALLOWED | WRITE_ALLOWED;
    tfxDoc->seqStatus |= TFX_DOC_SEQ_READSTART | TFX_DOC_SEQ_WRITESTART;

    /* Get the document information */
    XERR_PASS_CLEANUP(TFX_GetDocInfo(*pTfxDocHandle, &version, &revision,
	&nPages));

    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}



/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy an individual image from one TIFF-FX to another (leave compressed)
 *
 * LongDescription:::
 *  TFX_CopyImage() copies an individual image from one TIFF-FX to another,
 *  leaving the image data compressed during the copy.  This allows a much
 *  faster copy, since the data is not decompressed and recompressed.  It also
 *  means that the use of a lossy algorithm (like JPEG) will not produce added
 *  lossy affects by recompressing with additional or different loss.
 *  <BR>
 *  Note: this function will NOT keep track of version inconsistencies (like
 *  the fact that most profiles do not support tokenization).
 *
 * Input:::
 *  xSrcImgHandle::		Source image handle, returned from
 *				TFX_GetImageInfo()
 *  xDstParentImgHandle::	Destination image handle for parent image (if
 *				source image is a hints image)
 *  xDstPageHandle::		Destination page handle, returned from
 *				TFX_AddPageStart() or TFX_GetPageInfo()
 *  dwImgLayer::		Image layer in MRC
 *  dwXOffset::			X pixel offset in Primary image coordinates
 *  dwYOffset::			Y pixel offset in Primary image coordinates
 *  dwBestStripHeight::		Height of strips stored
 *
 * Output:::
 *  pxDstImgHandle::		Pointer to resulting image handle, can be used
 *				to later reference this image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_CopyImage(
	IO_IMGHANDLE xSrcImgHandle,
	IO_IMGHANDLE xDstParentImgHandle,
	IO_PAGEHANDLE xDstPageHandle,
	UInt32 dwImgLayer,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwBestStripHeight,
	IO_IMGHANDLE *pxDstImgHandle
)
{
    TFX_IMG *xSrcImg, *xDstParentImg = NULL;
    TFX_PAGE *xDstPage;
    TFX_DOC *xSrcDoc, *xDstDoc;
    TFX_IMGINFO *pxSrcImgInfo, xDstImgInfo;
    UInt32 *pDstPageImageCounts = NULL, dstPageLayersPresent = 0;
    UInt16 wAnnotationCount;
    IO_PAGEHANDLE xOutPage;
    PROLOGUE;

    TFX_CheckHandle(xSrcImgHandle);
    TFX_CheckHandle(xDstPageHandle);

    xSrcImg = (TFX_IMG *)xSrcImgHandle;
    xDstPage = (TFX_PAGE *)xDstPageHandle;
    pxSrcImgInfo = &xSrcImg->imgInfo;

    /* If image is a hints image, check the parent handle */
    if (pxSrcImgInfo->tfxImgType == IO_IMGTYPE_HINTS) {
	TFX_CheckHandle(xDstParentImgHandle);
	xDstParentImg = (TFX_IMG *)xDstParentImgHandle;
    }

    /* Access the documents */
    xSrcDoc = xSrcImg->tfxPage->tfxDoc;
    xDstDoc = xDstPage->tfxDoc;

    /* Get (make sure it's fresh) the page info */
    if (xDstDoc->docInfo.wNPages > xDstPage->tfxDoc->docInfo.wNPages) {
	XERR_PASS_CLEANUP(TFX_GetPageInfo((IO_DOCHANDLE)xDstDoc,
	    xDstPage->pageInfo.dwPageOffset,
	    &pDstPageImageCounts, &dstPageLayersPresent,
	    &wAnnotationCount, &xOutPage));
    }

    /* Create the resulting image info structure */
    memcpy(&xDstImgInfo, pxSrcImgInfo, sizeof(xDstImgInfo));

    /* These items will be different */
    xDstImgInfo.dwImgLayer = dwImgLayer;
    xDstImgInfo.dwXOffset = dwXOffset;
    xDstImgInfo.dwYOffset = dwYOffset;
    xDstImgInfo.dwBestStripHeight = dwBestStripHeight;
/*    xDstImgInfo.dwMaskImages = 0; */
    xDstImgInfo.dwHintsImages = 0;
    xDstImgInfo.dwImgOffset = 0;

    /* Call lower level routine to do the copy */
    XERR_PASS_CLEANUP(wp_TfxCopyImage(xSrcImg, xDstParentImg, xDstPage,
	&xDstImgInfo, 0, 0));

    /* Return the new image handle */
    *pxDstImgHandle = (IO_IMGHANDLE)xDstImgInfo.tfxImgHandle;

    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copies the entire contents of a page from one TIFF-FX to another
 *
 * LongDescription:::
 *  TFX_CopyPage() copies a specified page from one TIFF-FX document to
 *  another. The page can be an arbitrary page in the source document,
 *  but must be the next available page in the destination document
 *  (no page insertions).  The image data is left compressed during the
 *  copy, enabling fast copies.  Version differences are NOT fixed in
 *  this copy. If the compression stream of the source image data is of
 *  an outdated format, the stream is not updated.
 *  <BR>
 *  This function can be used to facilitate page insertion, deletion and
 *  reordering; but you must place it in a loop with the appropriate page
 *  offsets to sequentially specify the new destination page order.
 *
 * Input:::
 *  xSrcDocHandle::		Source document handle, returned from
 *				TFX_OpenDocumentRead() or
 *				TFX_OpenDocumentReadWrite()
 *  xDstDocHandle::		Destination document handle, returned from
 *				TFX_CreateDocument() or
 *				TFX_OpenDocumentReadWrite()
 *  dwSrcPageOffset::		Source document offset of page to copy
 *  dwDstPageOffset::		Destination document offset of page to add
 *				(this should be the end of the document).
 *				-1 indicates the next available page.
 *  copyAnnotations::		Boolean for copy the page annotations
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_CopyPage(
	IO_DOCHANDLE xSrcDocHandle,
	IO_DOCHANDLE xDstDocHandle,
	Int32 dwSrcPageOffset,
	Int32 dwDstPageOffset,
	Bool copyAnnotations
)
{
    TFX_DOC *xSrcDoc, *xDstDoc;
    PROLOGUE;

    TFX_CheckHandle(xSrcDocHandle);
    TFX_CheckHandle(xDstDocHandle);

    xSrcDoc = (TFX_DOC *)xSrcDocHandle;
    xDstDoc = (TFX_DOC *)xDstDocHandle;

    /* Allow negative number to indicate the next available page */
    if (dwDstPageOffset < 0) {
	dwDstPageOffset = xDstDoc->docInfo.wNPages;
    }

    XERR_PASS_CLEANUP(wp_TfxCopyPage(xSrcDoc, dwSrcPageOffset,
	xDstDoc, dwDstPageOffset, copyAnnotations));

    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copies the entire document from one file to another that is
 *  stream-ordered
 *
 * LongDescription:::
 *  TFX_CopyDoc() copies a all known file fragments from one file to
 *  another, creating an output file that has stream-ordered data.
 *  Note, streaming order is a "recommendation" for TIFF-FX; this
 *  function is used to facilitate the rewrite of a TIFF-FX file that
 *  may not be in streamed order.
 *  <BR>
 *  All image fragments will remain compressed, but the order in which
 *  they reside in the document may change.
 *  <BR>
 *  Note, using TFX_CopyPage() may not be sufficient to create streaming
 *  order.
 *
 * Input:::
 *  xSrcDocHandle::		Source document handle, returned from
 *				TFX_OpenDocumentRead() or
 *				TFX_OpenDocumentReadWrite()
 *  xDstDocHandle::		Destination document handle, returned from
 *				TFX_CreateDocument() or
 *				TFX_OpenDocumentReadWrite()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_CopyDoc(
	IO_DOCHANDLE xSrcDocHandle,
	IO_DOCHANDLE xDstDocHandle
)
{
    TFX_DOC *xSrcDoc, *xDstDoc;
    UInt16 srcPageOffset, srcPageCount;
    UInt16 dstPageOffset, dstPageCount;
    UInt8 srcVersion, srcRevision;
    UInt8 dstVersion, dstRevision;
    DOCTYPE docProfile;
    PROLOGUE;

    TFX_CheckHandle(xSrcDocHandle);
    TFX_CheckHandle(xDstDocHandle);

    xSrcDoc = (TFX_DOC *)xSrcDocHandle;
    xDstDoc = (TFX_DOC *)xDstDocHandle;

    /* Get the source page count */
    XERR_PASS_CLEANUP(TFX_GetDocInfo(xSrcDocHandle, &srcVersion, &srcRevision,
	&srcPageCount));

    /* Check the destination page count */
    XERR_PASS_CLEANUP(TFX_GetDocInfo(xDstDocHandle, &dstVersion, &dstRevision,
	&dstPageCount));

    /* Determine the profile of the source document */
    XERR_PASS_CLEANUP(TFX_GetDocProfile(xSrcDocHandle, &docProfile));

    /* Make sure we're still consistent */
    if (dstPageCount > 0) {

	/* Check the profile of the destination document */
	switch (xDstDoc->docInfo.profile) {
	case TFX_PROFILE_S:		/* Profile S: standard B&W */
	case TFX_PROFILE_F:		/* Profile F: Fax B&W */
	case TFX_PROFILE_C:		/* Profile C: Contone */
	case TFX_PROFILE_M:		/* Profile M: MRC */
	case TFX_PROFILE_T:		/* Profile T: Tokenization (JBIG2, Binary) */
	case TFX_PROFILE_M1:		/* Profile M1: MRC + JBIG2 */
	case TFX_PROFILE_COH_B:		/* Coherent TIFF Profile B: Binary */
	case TFX_PROFILE_COH_C:		/* Coherent TIFF Profile C: Contone */
	case TFX_PROFILE_COH_M:		/* Coherent TIFF Profile M: MRC */
	case TFX_PROFILE_COH_M1:	/* Coherent TIFF Profile M1: MRC + JBIG2 */

	    /* Later we can get more sophisticated in "upgrading" the destination
	     * document to a different profile.  For now, complain. */
	    XERR_GEN_CLEANUP(xSrcDoc->docInfo.profile != xDstDoc->docInfo.profile,
		IO_NOSUPPORT);
	}
    } else {
	/* Blindly change the profile.  This could cause trouble later, maybe.
	 * Theoretically, if nothing has really been written, then there should
	 * be nothing to distinguish it's profile, yet. */
	xDstDoc->docInfo.profile = xSrcDoc->docInfo.profile;
    }

    /* Copy all the pages */
    dstPageOffset = dstPageCount;
    for (srcPageOffset = 0; srcPageOffset < srcPageCount; srcPageOffset++) {
	XERR_PASS_CLEANUP(wp_TfxCopyPage(xSrcDoc, srcPageOffset,
	    xDstDoc, dstPageOffset, 1));
	dstPageOffset++;
    }

    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}

 
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Begin the process of adding a new page
 *
 * LongDescription:::
 *  TFX_AddPageStart() adds a new page to a multipage document. Page offsets
 *  start at zero, but setting dwPageNumber < 0 causes the new page to be
 *  added at the end of the document. The operation will fail if a page
 *  is added beyond the next available page at the end of the document, i.e.
 *  {0, 1, 2, 7} is not a valid page sequence, and the attempt to add page
 *  offset 7 would fail.
 *  <BR>
 *  Note that when adding pages A, B, and C, each with offset = 0, will result
 *  in the pages having a reverse page order -- the final page offsets will be
 *  2, 1, and 0.
 *  <BR>
 *  The page handle returned from this routine is used in later calls to
 *  attach images and annotations to this page.
 *  <BR>
 *  TFX_AddPageStart() may be followed by adding a Primary image using
 *  TFX_AddImage(). 
 *
 * Input:::
 *  tfxDocHandle::		Document handle
 *  dwPageOffset::		Offset of page to add, should be up to the
 *				existing page count, or negative, to indicate
 *				the end of the document
 *  pTfxPageHandle::		Page handle returned
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_AddPageStart(
	IO_DOCHANDLE tfxDocHandle,
	Int32 dwPageOffset,
	IO_PAGEHANDLE *pTfxPageHandle
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    TFX_PAGE *tfxPage = NULL;
    TFX_PAGE *searchPage = NULL, *prevPage;
/*    UInt8 flags = 0; */
    Int32 pageOffset;
    Bool inserting = 0;
/*    TFX_PAGETABLE *pt; */
/*    Int32 tableOffset; */
    PROLOGUE;
    
    TFX_CheckHandle(tfxDocHandle);
    
    /* Check for proper calling sequence */
    XERR_GEN((tfxDoc->seqStatus & TFX_DOC_SEQ_WRITESTART) == 0,
	IO_SEQUENCE);

    /* Allow negative number to indicate the next available page */
    if (dwPageOffset < 0) {
	dwPageOffset = tfxDoc->docInfo.wNPages;
/*	dwPageOffset = tfxDoc->pageTableHeader.pageCount; */
    }

    /* Pages must be inserted, or appended */
    XERR_GEN(dwPageOffset > tfxDoc->docInfo.wNPages,
	IO_BADPARAMETER);
/*    XERR_GEN(dwPageOffset > tfxDoc->pageTableHeader.pageCount, */
/*	IO_BADPARAMETER); */
    
    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);

#if 0
    if ((tfxDoc->readWrite & READ_ALLOWED) &&
	tfxDoc->pageTableHeader.entriesPerTable > 0) {

	/* Search *PAST* the last page table that is stored in memory */
	pt = tfxDoc->firstPageTable;
	tableOffset = 0;
	if (pt != NULL) {
	    do {
		/* Pages tables should ALWAYS be sequential */
		XERR_GEN_CLEANUP(tableOffset + 1 != pt->ordinal,
		    IO_INTERNAL_ERROR);
		
		tableOffset++;
	    } while ((pt = pt->next) != NULL);
	}

	/* If we haven't yet read the last page table entry, we must
	 * read it now, since we'll either add to the last table, add
	 * a new table, or shift the tables (in the case of an insert)
	 */
	if (tableOffset <= tfxDoc->pageTableHeader.pageCount /
	    tfxDoc->pageTableHeader.entriesPerTable) {
	    XERR_PASS_CLEANUP(wp_TfxReadPageTable(tfxDoc, 0UL, NULL));
	}
    }
#endif
    if (tfxDoc->readWrite & READ_ALLOWED) {
	/* Make sure we have the page info for all the pages up to the
	 * place where we'll add the new one. */
	UInt32 imageLayers;
	UInt16 annotations;
	IO_PAGEHANDLE pageHandle;

	for (pageOffset = 0; pageOffset < dwPageOffset; pageOffset++) {
	    XERR_WARNOK_CLEANUP(TFX_GetPageInfo((IO_DOCHANDLE)tfxDoc,
		(UInt32)pageOffset, NULL, &imageLayers, &annotations,
		&pageHandle));
	}

	/* Keep track of whether or not we're inserting */
	if (dwPageOffset < tfxDoc->docInfo.wNPages) {
	    inserting = 1;
	}
    }
    
    /* Look for a previous read/create of this page */
    searchPage = tfxDoc->pages;
    pageOffset = 0;
    if (searchPage != NULL) {
	do {
	    /* Found, this page already exists.  We'll be inserting this page */
	    if (searchPage->pageInfo.dwPageOffset == dwPageOffset) {
		inserting = 1;
		break;
	    }

	    pageOffset++;
	} while ((searchPage = searchPage->next) != NULL);
    }
    
    if (searchPage == NULL ||
	searchPage->pageInfo.dwPageOffset != dwPageOffset || inserting) {
	
	/* Begin construction of the page data */
	IO_CHECKMALLOC(tfxPage = (TFX_PAGE *)TFX_MALLOC(sizeof(TFX_PAGE)));
	memset(tfxPage, 0, sizeof(TFX_PAGE));
	tfxPage->handleID = (UInt32)tfxPage;	/* self identity */
	tfxPage->handleType = TFX_HANDLE_TYPE;
	tfxPage->tfxDoc = tfxDoc;
	tfxPage->pageInfo.tfxPageHandle = (UInt32)tfxPage;
	tfxPage->pageInfo.dwPageOffset = dwPageOffset;

#if 0
	/* This can get tricky here: write a made-up offset for the last page.
	 * If this causes a new page table to be written, then the following
	 * call to FileLongAlign() won't return a bogus value once the call
	 * to TfxSetPageIFD() that follows is called. */
	if ((tfxDoc->pageTableHeader.entriesPerTable == 0UL) ||
	    ((tfxDoc->docInfo.wNPages %
	    tfxDoc->pageTableHeader.entriesPerTable == 0UL) &&
	    (tfxDoc->docInfo.wNPages / tfxDoc->pageTableHeader.entriesPerTable >=
	    tfxDoc->pageTableHeader.lastOrdinal))) {
	    XERR_PASS_CLEANUP(wp_TfxSetPageIFD(tfxDoc, tfxDoc->docInfo.wNPages,
		flags, 0UL, tfxPage));
	}
#endif

	/* Attach it to the appropriate spot in linked list */
	prevPage = NULL;
	searchPage = tfxDoc->pages;
	if (searchPage == NULL) {
	    tfxDoc->pages = tfxPage;
	} else {
	    do {
		if (searchPage->pageInfo.dwPageOffset >= dwPageOffset) {

		    /* Only = when inserting */
		    XERR_GEN_CLEANUP(inserting != 1 &&
			searchPage->pageInfo.dwPageOffset == dwPageOffset,
			IO_INTERNAL_ERROR);
	    
		    /* Found the right spot to insert this page before */
		    break;
		}
		prevPage = searchPage;
	    } while ((searchPage = searchPage->next) != NULL);
	    
	    /* Insert before searchPage */
	    tfxPage->prev = prevPage;
	    if (prevPage) {	/* There's a previous page */
		prevPage->next = tfxPage;
	    } else {
		tfxDoc->pages = tfxPage;
	    }
	    tfxPage->next = searchPage;
	    if (searchPage) {	/* There's a page which will follow */
		searchPage->prev = tfxPage;

		/* Update the page offset of all those that follow */
		while (searchPage) {
		    searchPage->pageInfo.dwPageOffset++;
		    searchPage = searchPage->next;
		}
	    }
	}

	/* Update the page count */
	tfxDoc->docInfo.wNPages++;
    }

#if 0
    /* If the page was inserted, rewrite the affected page table entries */
    if (inserting) {
	tfxDoc = tfxDoc;
/*	XERR_PASS_CLEANUP(wp_TfxRewritePageTableEntries(tfxDoc, tfxPage)); */
    }
#endif

    /* Return the page handle */
    *pTfxPageHandle = (IO_PAGEHANDLE)tfxPage;

    /* Update status */
    tfxPage->seqStatus |= TFX_PAGE_SEQ_WRITESTART;
    tfxDoc->seqStatus |= TFX_DOC_SEQ_GETINFO; /* | TFX_DOC_SEQ_W_PAGETABLE; */

    tfxPage = NULL;
	
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (tfxPage) {
	TFX_FREE(tfxPage);
	tfxPage = NULL;
    }
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Complete the construction of a page, tokenization still pending
 *
 * LongDescription:::
 *  TFX_AddPageFinish() completes the general construction of a page, although
 *  the completion of the tokenization process later may actually store the
 *  data for the Primary IFD, and may also add other subimages for Layer 2.
 *
 * Input:::
 *  tfxPageHandle::		Page handle of page to complete
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_AddPageFinish(
	IO_PAGEHANDLE tfxPageHandle
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    PROLOGUE;

    TFX_CheckHandle(tfxPageHandle);
    
    /* Check for proper calling sequence */

    /* We must have started to write this page */
    XERR_GEN_CLEANUP((tfxPage->seqStatus & TFX_PAGE_SEQ_WRITESTART) == 0,
	IO_SEQUENCE);

    /* Mark as completely written */
    if ((tfxPage->seqStatus & TFX_PAGE_SEQ_TOKENSTART) == 0) {
	tfxPage->seqStatus &= ~TFX_PAGE_SEQ_WRITESTATUS; /* Clear write status */
	tfxPage->seqStatus |= TFX_PAGE_SEQ_WRITEFINISH;  /* Set to finished */
	tfxPage->seqStatus |= TFX_PAGE_SEQ_READSTART;	/* Mark as readable */
    } else {
	/* Mark as finshed, but pending tokenization completion */
	tfxPage->seqStatus |= TFX_PAGE_SEQ_WRITEPENDING;  /* Pending tokenization */
    }	

    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Delete an entire page from the document (including all images on the page)
 *
 * LongDescription:::
 *  TFX_DeletePage() deletes (removes all references to) an entire page in
 *  the file.  The image data is not physically removed (as the image is
 *  likely to be inbetween other file fragments), but the references to the
 *  page are removed.  This includes: page count in the page table header,
 *  removing the primary IFD from the page table (and shifting subsequent
 *  page references to accomodate the removal), and relinking the previous
 *  page to the subsequent page via the NextIFD IFD entry.
 *  <BR>
 *  Note that although the page is not physically removed, since the
 *  references to it are removed (and lost), there is no way to 'undo' a
 *  page deletion.
 *  <BR>
 *  The intent is to allow physically editing the file, with anticipation
 *  that later the file will be copied (see TFX_CopyPage() and
 *  TFX_CopyImage()), allowing the disowned file fragments to remain behind.
 *  <BR>
 *  Note also that the page offsets of all subsequent pages will automatically
 *  be shifted down one (previous page offsets will now be wrong, only for
 *  pages that followed this one).
 *
 * Input:::
 *  tfxPageHandle::		Page handle of the page to be removed, as
 *				returned from TFX_GetPageInfo()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_DeletePage(
	IO_PAGEHANDLE tfxPageHandle
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_DOC *tfxDoc;
    UInt8 flags;
    Int32 ifdOffset;
    TFX_PAGE *nextPage;	/* Page left in deleted pages place */
    PROLOGUE;

    TFX_CheckHandle(tfxPageHandle);
    tfxDoc = tfxPage->tfxDoc;

    /* Make sure we've read all the page tables (since the internal structures
     * are actually stored back out in the revised page table). */
/*    XERR_PASS_CLEANUP(wp_TfxGetPageIFD(tfxDoc, tfxDoc->pageTableHeader.pageCount - 1, */
/*	&flags, &ifdOffset)); */
    XERR_PASS_CLEANUP(wp_TfxGetPageIFD(tfxDoc, tfxDoc->docInfo.wNPages - 1,
	&flags, &ifdOffset));

#if 1
    /* It's unclear that this is needed, and it will cause a lot of
     * time to be spent reading through all the IFDs */
    /* If we know we're inserting, make sure we have the page info
     * for all the pages, since we'll be shifting their ordinals */
/*    if (tfxPage->pageInfo.dwPageOffset < tfxDoc->pageTableHeader.pageCount) { */
    if (tfxPage->pageInfo.dwPageOffset < tfxDoc->docInfo.wNPages) {
	Int32 pageNum;
	UInt32 imageLayers;
	UInt16 annotations;
	IO_PAGEHANDLE pageHandle;

/*	for (pageNum = 0; pageNum < tfxDoc->pageTableHeader.pageCount; pageNum++) { */
	for (pageNum = 0; pageNum < tfxDoc->docInfo.wNPages; pageNum++) {
	    XERR_PASS_CLEANUP(TFX_GetPageInfo((IO_DOCHANDLE)tfxDoc,
		(UInt32)pageNum, NULL, &imageLayers, &annotations,
		&pageHandle));
	}
    }
#endif

    /* Actually remove the page from the linked list */
    if (tfxPage->prev) {
	tfxPage->prev->next = nextPage = tfxPage->next;
    } else {
	tfxDoc->pages = nextPage = tfxPage->next;
    }
    if (tfxPage->next) {
	tfxPage->next->prev = tfxPage->prev;
    }
    tfxPage->prev = tfxPage->next = NULL;

    /* Remove the page from existance */
    XERR_PASS_CLEANUP(wp_TfxPageFree(&tfxPage, 0));

    /* Revise the page offsets of all succeeding pages */
    tfxPage = nextPage;
    while (tfxPage != NULL) {
	tfxPage->pageInfo.dwPageOffset--;
	tfxPage = tfxPage->next;
    }

#if 0
    /* Shift the page table entries by one */
    XERR_PASS_CLEANUP(wp_TfxRewritePageTableEntries(tfxDoc, nextPage == NULL?
	tfxDoc->pages: nextPage));
#endif

    /* Update the page count */
    tfxDoc->docInfo.wNPages--;
/*    tfxDoc->pageTableHeader.pageCount--; */
/*    XERR_PASS_CLEANUP(wp_TfxWritePageTableHeader(tfxDoc)); */
/*    XERR_GEN_CLEANUP(tfxDoc->docInfo.wNPages != tfxDoc->pageTableHeader.pageCount, */
/*    	IO_INTERNAL_ERROR); */

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add an image to a page
 *
 * LongDescription:::
 *  TFX_AddImage() adds a new image to a page. Image should not be a
 *  transparancy mask, TFX_AddMaskImage() is used for handling this.
 *  <BR>
 *  Note that every empty page must start with a Primary image which defines
 *  the boundaries of the page, both in width and height.  IO_BADPARAMETER
 *  will be returned if the first image added is not a primary image (image
 *  layer == 2).
 *  <BR>
 *  An image handle is returned, that is used in later calls to three
 *  functions: TFX_ImageWriteStart(), TFX_ImageWriteStrip(), and
 *  TFX_ImageWriteFinish(); or to the two functions:
 *  TFX_ImageTokenizeImage(), and TFX_WriteTokenImage(); or the function:
 *  TFX_ImageTokenizeImages().
 *
 * Input:::
 *  tfxPageHandle::		Page to attach to
 *  dwImgLayer::		Image layer in MRC
 *  xColorType::		Color type
 *  dwXOffset::			X pixel offset in Primary image coordinates
 *  dwYOffset::			Y pixel offset in Primary image coordinates
 *  dwXResolution::		Resolution in the X direction
 *  dwYResolution::		Resolution in the Y direction
 *  dwWidth::			Width
 *  dwHeight::			Height
 *  dwBestStripHeight::		Height of strips stored
 *  pPalette::			Palette (if IO_COLORTYPE_COLOR8 or COLOR4)
 *
 * Output:::
 *  pImgHandle::		Pointer to image handle, used to later
 *				reference this image
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_AddImage(
	IO_PAGEHANDLE tfxPageHandle,
	UInt32 dwImgLayer,
	IO_COLORTYPE xColorType,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwXResolution,
	UInt32 dwYResolution,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt32 dwBestStripHeight,
	UInt8 *pPalette,
	IO_IMGHANDLE *pImgHandle
)
{
    TFX_INSTANCE *instData;
    TFX_FILE *tfxFile;
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;	/* Page */
    TFX_IMG *tfxParentImg = NULL;		/* Parent image */
    Int32 ifdOffset = 0;
/*    UInt8 flags = 0; */
    TFX_IMG *curImg;			/* Current image */
    TFX_IMG *newImg = NULL;		/* Unlinked image */
    TFX_IMG *prevImg, *nextImg;		/* Prev/next image in ordered link list */
    IO_COMPRESSION xCompression;
    UInt32 dwImgOffset;
    IO_COLORSPACE xColorSpace = IO_COLORSPACE_DEFAULT;
    PROLOGUE;

    /* Default return value */
    *pImgHandle = 0UL;

    TFX_CheckHandle(tfxPageHandle);
    instData = TFX_PAGEINST(tfxPage);
    tfxFile = TFX_PAGEFILE(tfxPage);

    /* Either we need a primary image, or we're adding it now */
/*    XERR_GEN(dwImgLayer != 2 && tfxPage->pageInfo.tfxPrimImgHandle == 0, */
/*		     IO_SEQUENCE); */
    tfxParentImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    /* Check for proper calling sequence:
     * 1 - not writing this page
     * 2 - primary image not written for this page, and this isn't
     *     the primary image
     */
    XERR_GEN((((tfxPage->seqStatus & TFX_PAGE_SEQ_WRITESTART) == 0) ||
	(((tfxPage->seqStatus & TFX_PAGE_SEQ_WRITEPRIMARY) == 0)
	&& (dwImgLayer != 1) && (dwImgLayer != 2))), IO_SEQUENCE);

    /* Does this fit the profile */
    switch (tfxPage->tfxDoc->docInfo.profile) {
    case TFX_PROFILE_M:		/* Profile M: MRC */

    case TFX_PROFILE_M1:		/* Profile M1: MRC + JBIG2 */

    case TFX_PROFILE_COH_M:		/* Coherent TIFF Profile M: MRC */
    case TFX_PROFILE_COH_M1:	/* Coherent TIFF Profile M1: MRC + JBIG2 */
	break;
    default:
	/* This is a non-primary image being added as an MRC sublayer, but it's
	 * not an MRC profile */
	XERR_GEN_CLEANUP(tfxParentImg != NULL, IO_BADPARAMETER);
    }

    /* Check for version support: only layer 2 offset 0 valid in even layers */
    XERR_GEN(tfxPage->tfxDoc->docInfo.version < 3 &&
	(dwImgLayer % 2) == 0 &&		    /* Even numbered layer */
	(dwImgLayer != 2 || tfxParentImg != NULL),    /* Not primary image */
	IO_NOSUPPORT);

    /* Make some page consistency checks */
    xCompression = IO_COMP_UNKNOWN;
    dwImgOffset = 0UL;
    XERR_IGNORE(__error__ = TFX_CheckImageInfo(tfxPageHandle, &dwImgLayer,
	&dwImgOffset, &xColorType, &dwXOffset, &dwYOffset,
	&dwXResolution, &dwYResolution,	&dwWidth, &dwHeight,
	&dwBestStripHeight, pPalette, &xCompression, &xColorSpace));
    XERR_GEN_CLEANUP(__error__ && __error__ < IO_WARNINGS, __error__);

    /* Align long (for all TIFF IFDs) */
/*    XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &ifdOffset)); */

    /* Allocate a new image */
    XERR_PASS_CLEANUP(wp_TfxNewImg(tfxPage, &newImg, ifdOffset));
    newImg->imgInfo.tfxImgType = IO_IMGTYPE_LAYER;
    newImg->imgInfo.dwImgLayer = dwImgLayer;
    newImg->imgInfo.dwImgOffset = 0UL;

    /* Retain only the transferable information */
    newImg->imgInfo.dwXOffset = dwXOffset;
    newImg->imgInfo.dwYOffset = dwYOffset;
    newImg->imgInfo.dwXResolution = dwXResolution;
    newImg->imgInfo.dwYResolution = dwYResolution;
    newImg->imgInfo.dwWidth = dwWidth;
    newImg->imgInfo.dwHeight = dwHeight;
    newImg->imgInfo.dwBestStripHeight = dwBestStripHeight;
    newImg->imgInfo.xCompression = IO_COMP_UNKNOWN;
    newImg->imgInfo.xColorType = xColorType;
    newImg->imgInfo.xColorSpace = xColorSpace;
    switch ((int)xColorType) {
    case IO_COLORTYPE_COLOR4:
    case IO_COLORTYPE_COLOR8:
	XERR_GEN_CLEANUP((pPalette == NULL),
		 IO_BADPARAMETER);
	{
	    UInt32 paletteSize;

	    paletteSize = (1 << IO_BPPE(xColorType)) * 3;
	    memcpy(newImg->imgInfo.pPalette, pPalette, paletteSize);
	}
    default:
	break;
    }

    /* If this is the primary image, link it to the page */
    if (tfxParentImg == NULL) {

	/* TAKE THIS OUT LATER, debug writing after IFD is queued
	 * TAKE THIS OUT LATER, debug writing after IFD is queued
	 * TAKE THIS OUT LATER, debug writing after IFD is queued
	 * TAKE THIS OUT LATER, debug writing after IFD is queued
	 * TAKE THIS OUT LATER, debug writing after IFD is queued
	 * TAKE THIS OUT LATER, debug writing after IFD is queued
	 * TAKE THIS OUT LATER, debug writing after IFD is queued
	 */
	/* Make a place-holder for the IFD, and begin queuing mode */
/*	XERR_PASS_CLEANUP(wp_TfxQueueIFD(newImg)); */

	/* Assign this one as the primary image */
	tfxPage->pageInfo.tfxPrimImgHandle = (UInt32)(curImg = newImg);
	newImg = NULL;
	curImg->imgInfo.tfxImgType = IO_IMGTYPE_PRIMARY;

#if 0
	/* In the case of an inserted page, shift the page table entries to
	 * accomodate the insertion of this page. */
	if (tfxPage->next) {
	    XERR_PASS_CLEANUP(wp_TfxRewritePageTableEntries(tfxPage->tfxDoc, tfxPage));
	}

	/* Make sure we have the page table for the correct page */
	XERR_PASS_CLEANUP(wp_TfxGeneratePageFlags(tfxPage, &flags));
	XERR_PASS_CLEANUP(wp_TfxSetPageIFD(tfxPage->tfxDoc, tfxPage->pageInfo.dwPageOffset,
					flags, ifdOffset, tfxPage));
#endif

	/* Start off with a image count of 1 for this layer (usually layer 2) */
	XERR_PASS_CLEANUP(wp_TfxIncrementImageCounts(tfxPage, dwImgLayer, 1));

	/* Update the sequence status */
	tfxPage->seqStatus |= TFX_PAGE_SEQ_WRITEPRIMARY | TFX_PAGE_SEQ_GETINFO;
	
    } else {	/* Otherwise link it to the parent image */
	XERR_PASS_CLEANUP(wp_TfxIncrementImageCounts(tfxPage, dwImgLayer, 1));
	newImg->imgInfo.dwImgOffset =
	    tfxPage->pageInfo.pdwImageCounts[dwImgLayer - 1] - 1;
	if (tfxPage->tfxImgs == NULL) {
	    tfxPage->tfxImgs = curImg = newImg;
	    newImg = NULL;
	} else {

	    /* Search for the proper point to insert *before* */
	    prevImg = nextImg = tfxPage->tfxImgs;
	    while (nextImg &&
		(nextImg->imgInfo.dwImgLayer < dwImgLayer ||
		(nextImg->imgInfo.dwImgLayer == dwImgLayer &&
		 nextImg->imgInfo.dwImgOffset < newImg->imgInfo.dwImgOffset)
		)) {
		prevImg = nextImg;
		nextImg = nextImg->next;
	    }

	    /* It is an error to insert (for now) */
	    XERR_GEN_CLEANUP(nextImg && nextImg->imgInfo.dwImgLayer == dwImgLayer &&
		nextImg->imgInfo.dwImgOffset == newImg->imgInfo.dwImgOffset,
		IO_INTERNAL_ERROR);

	    /* Link this image to the one that is now the next one */
	    if (nextImg) {
		newImg->next = nextImg;
		nextImg->prev = newImg;
	    }

	    /* Link prev image to this one */
	    if (prevImg == nextImg) {

		/* Link as new first image */
		tfxPage->tfxImgs = newImg;
	    } else {
		prevImg->next = newImg;
		newImg->prev = prevImg;

	    }
	    curImg = newImg;
	    newImg = NULL;
	}
/*	XERR_PASS_CLEANUP(wp_TifFXUpdateILContent(tfxParentImg)); */

	/* Attach this image to its parent */
/*	XERR_PASS_CLEANUP(wp_TifFXAttachChildIFD(tfxParentImg, curImg, ifdOffset)); */
    }

    /* Return the image added */
    *pImgHandle = (UInt32)curImg;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (newImg) {
	TFX_FREE(newImg);
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Delete an image from a page
 *
 * LongDescription:::
 *  TFX_DeleteImage() deletes (removes all references to) an single image on
 *  a page.  The image data is not physically removed (as the image is
 *  likely to be inbetween other file fragments), but the references to the
 *  image are removed.  This includes: MRC layer image counts and relinking
 *  the previous image to the subsequent image via the NextIFD IFD entry.
 *  <BR>
 *  Note that although the image is not physically removed, since the
 *  references to it are removed (and lost), there is no way to 'undo' an
 *  image deletion.
 *  <BR>
 *  The intent is to allow physically editing the file, with anticipation
 *  that later the file will be copied (see TFX_CopyPage() and
 *  TFX_CopyImage()), allowing the disowned file fragments to remain behind.
 *
 * Input:::
 *  tfxImgHandle::		Image handle of the image to be removed, as
 *				returned from TFX_GetImageInfo()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_DeleteImage(
	IO_IMGHANDLE tfxImgHandle
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;	/* Image */
    PROLOGUE;

    TFX_CheckHandle(tfxImgHandle);

    /* Check for proper calling sequence */
    XERR_GEN((tfxImg->tfxPage->tfxDoc->readWrite & WRITE_ALLOWED) == 0,
	     IO_SEQUENCE);

    /* Let the TIFF routine do the dirty work */
    XERR_PASS_CLEANUP(wp_TifFXDeleteImage(tfxImg));

    NORMAL_RETURN;

 ERR_CLEANUP:

    ERR_RETURN;
}


#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add a transparency mask to a non-binary image
 *
 * LongDescription:::
 *  TFX_AddMaskImage() is like TFX_AddImage(), except that instead of adding
 *  an image to a page, we are adding a transparency mask image to a parent
 *  non-binary image.  The image layer is inherited from the parent image.
 *  This image must have the same X/Y location, width, height and resolutions
 *  as the parent image.  This image MUST also be binary, with 0 indicating
 *  the transparent  pixels, and 1 indicating the opaque pixels of the parent.
 *  <BR>
 *  An image handle is returned, that is used in later calls to three
 *  functions: TFX_ImageWriteStart(), TFX_ImageWriteStrip(), and
 *  TFX_ImageWriteFinish().
 *
 * Input:::
 *  tfxParentImgHandle::		Parent image to attach to
 *  dwBestStripHeight::		For best performance, write strips with
 *				height a multiple of this
 *
 * Output:::
 *  pImgHandle::		Returned image handle for image to be
 *				added
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_AddMaskImage(
	IO_IMGHANDLE tfxParentImgHandle,
	UInt32 dwBestStripHeight,
	IO_IMGHANDLE *pImgHandle
)
{
    TFX_FILE *tfxFile;
    TFX_PAGE *tfxPage;
    TFX_IMG *tfxParentImg = (TFX_IMG *)tfxParentImgHandle;/* Parent image */
    Int32 ifdOffset;
    TFX_IMG *curImg;			/* Current mask image */
    TFX_IMG *newImg = NULL;		/* Unlinked mask image */
    PROLOGUE;

    /* Default return value */
    *pImgHandle = 0UL;

    TFX_CheckHandle(tfxParentImgHandle);

    tfxPage = tfxParentImg->tfxPage;

    /* Mask image cannot be applied to an even layered image, with the
     * special exception of the primary image (this exception does not
     * normally conform to TIFF-FX 3.0, but this library will allow it). */
    XERR_GEN_CLEANUP(((tfxParentImg->imgInfo.dwImgLayer & 1) != 1) &&
		     (tfxParentImg->imgInfo.tfxImgType != IO_IMGTYPE_PRIMARY),
		     IO_BADPARAMETER);
    /* Image must not already have a mask */
    XERR_GEN_CLEANUP((tfxParentImg->seqStatus & TFX_IMG_SEQ_MASKADDED) != 0,
		     IO_BADPARAMETER);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((((tfxPage->seqStatus & TFX_PAGE_SEQ_WRITESTART) == 0) ||
		      (tfxPage->seqStatus & TFX_PAGE_SEQ_WRITEPENDING) ||
	((tfxPage->seqStatus & TFX_PAGE_SEQ_WRITEPRIMARY) == 0)),
		     IO_SEQUENCE);

    if (dwBestStripHeight == 0) {

	/* General rule is that it shouldn't be more than 8K uncompressed */
	dwBestStripHeight = (8192 / (tfxParentImg->imgInfo.dwWidth *
	    IO_BPP(IO_COLORTYPE_BINARY) / 8));
    }

    tfxFile = TFX_PAGEFILE(tfxPage);

    /* Align long (for all TIFF IFDs) */
    XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &ifdOffset));

    /* Allocate a new image */
    XERR_PASS_CLEANUP(wp_TfxNewImg(tfxPage, &newImg, ifdOffset));
    newImg->imgInfo.tfxImgType = IO_IMGTYPE_MASK;
    newImg->imgInfo.dwImgLayer = 0;	/* Not in a layer */
    newImg->imgInfo.dwImgOffset = 0UL;

    /* Retain only the transferable information */
    newImg->imgInfo.dwXOffset = tfxParentImg->imgInfo.dwXOffset;
    newImg->imgInfo.dwYOffset = tfxParentImg->imgInfo.dwYOffset;
    newImg->imgInfo.dwXResolution = tfxParentImg->imgInfo.dwXResolution;
    newImg->imgInfo.dwYResolution = tfxParentImg->imgInfo.dwYResolution;
    newImg->imgInfo.dwWidth = tfxParentImg->imgInfo.dwWidth;
    newImg->imgInfo.dwHeight = tfxParentImg->imgInfo.dwHeight;
    newImg->imgInfo.dwBestStripHeight = dwBestStripHeight;
    newImg->imgInfo.xCompression = IO_COMP_UNKNOWN;
    newImg->imgInfo.xColorType = IO_COLORTYPE_BINARY;

    /* Write out the IFD */
/*    XERR_PASS_CLEANUP(wp_TfxQueueIFD(newImg)); */

    /* Link it to the parent image */
    if (tfxParentImg->xMaskImg == NULL) {
	tfxParentImg->xMaskImg = curImg = newImg;
	newImg = NULL;
    } else {
	/* Only one transparency mask is allowed */
	__error__ = IO_BADPARAMETER;
	goto ERR_CLEANUP;
    }

    /* Attach this image to its parent */
    XERR_PASS_CLEANUP(wp_TifFXAttachChildIFD(tfxParentImg, curImg, ifdOffset));

    /* For now, only one attached mask is allowed */
    tfxParentImg->imgInfo.dwMaskImages++;
    XERR_GEN_CLEANUP(tfxParentImg->imgInfo.dwMaskImages != 1, IO_INTERNAL_ERROR);

    /* Mark as having a mask */
    tfxParentImg->seqStatus |= TFX_IMG_SEQ_MASKADDED;

    /* Return the image added */
    *pImgHandle = (UInt32)curImg;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (newImg) {
	TFX_INSTANCE *instData;

	if (tfxParentImg) {
	    instData = TFX_IMGINST(tfxParentImg);
	    TFX_FREE(newImg);
	}
    }

     ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the RenderingHints string for the current image
 *
 * LongDescription:::
 *  TFX_AddRenderingHints() sets the RenderingHints field for the
 *  current image.
 *  If the field has already been written, the entire IFD will be rewritten
 *  when the document is closed; since the IFD field count does not accomodate
 *  this field initially.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  pRenderingHints::		RenderingHints string to set
 *  dwRenderingHintsBytes::	Bytes in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_AddRenderingHints(
    IO_IMGHANDLE tfxImgHandle,
	char *pRenderingHintsString,
	UInt32 dwRenderingHintsBytes
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_INSTANCE *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    instData = TFX_IMGINST(tfxImg);

    
    if (tfxImg->fields.dataTransferred) {

	/* See if the field is already written, and rewrite the data if it has */
	if (tfxImg->fieldLocation.RenderingHints) {
	    TFX_FILE *tfxFile;
	    UInt32 lval;
	    Int32 location;

	    tfxFile = TFX_IMGFILE(tfxImg);

	    XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &location));
	    TFX_WRITE(tfxFile, pRenderingHintsString,
		(Int32)dwRenderingHintsBytes);

	    /* Update the data location */
	    TFX_SEEK(tfxFile, tfxImg->fieldLocation.RenderingHints - sizeof(long));
	    lval = dwRenderingHintsBytes;
	    TFX_SWAPLONG(tfxFile, &lval);
	    TFX_WRITE(tfxFile, &lval, sizeof(lval));
	    lval = location;
	    TFX_SWAPLONG(tfxFile, &lval);
	    TFX_WRITE(tfxFile, &lval, sizeof(lval));
	} else {

	    /* It's already been written to the file, queue the field
	     * to be rewritten when the document is closed. */
	    tfxImg->fields.IFDrewriteNeeded = 1;
	}
    }

    /* Allocate storage for the string */
    if (tfxImg->fields.RenderingHints != NULL) {
	TFX_FREE(tfxImg->fields.RenderingHints);
    }
    IO_CHECKMALLOC(tfxImg->fields.RenderingHints =
	TFX_MALLOC(dwRenderingHintsBytes + 1));
    tfxImg->fields.RenderingHints[dwRenderingHintsBytes] = '\0';

    /* Copy the data */
    memcpy(tfxImg->fields.RenderingHints, pRenderingHintsString,
	dwRenderingHintsBytes);
    tfxImg->fields.RenderingHintsLength = dwRenderingHintsBytes;

    /* Mark it as available for write */
    tfxImg->fields.found.RenderingHints = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add a rendering hints subimage to a parent image
 *
 * LongDescription:::
 *  TFX_AddHintsImage() is like TFX_AddImage(), except that instead of adding
 *  an image to a page, we are adding a rendering hints image to a parent
 *  image.  The image layer is inherited from the parent image.
 *  This image should have the same X/Y location, width, height and resolutions
 *  as the parent image; although they are parameters here, in case I'm wrong.
 *  <BR>
 *  An image handle is returned, that is used in later calls to three
 *  functions: TFX_ImageWriteStart(), TFX_ImageWriteStrip() or
 *  TFX_CompressedWriteStrip(), and TFX_ImageWriteFinish().
 *
 * Input:::
 *  tfxParentImgHandle::	Parent image to attach to
 *  dwBestStripHeight::		For best performance, write strips with
 *				height a multiple of this
 *
 * Output:::
 *  pImgHandle::		Returned image handle for image to be
 *				added
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_AddHintsImage(
	IO_IMGHANDLE tfxParentImgHandle,
	UInt32 dwBpp,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwXResolution,
	UInt32 dwYResolution,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt32 dwBestStripHeight,
	IO_IMGHANDLE *pImgHandle
)
{
    TFX_FILE *tfxFile;
    TFX_PAGE *tfxPage;
    TFX_IMG *tfxParentImg = (TFX_IMG *)tfxParentImgHandle;/* Parent image */
    Int32 ifdOffset;
    TFX_IMG *curImg = NULL;
    TFX_IMG *newImg = NULL;
    TFX_IMGINFO *parentInfo;
    PROLOGUE;

    /* Default return value */
    *pImgHandle = 0UL;

    TFX_CheckHandle(tfxParentImgHandle);

    tfxPage = tfxParentImg->tfxPage;

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP((((tfxPage->seqStatus & TFX_PAGE_SEQ_WRITESTART) == 0) ||
		      (tfxPage->seqStatus & TFX_PAGE_SEQ_WRITEPENDING) ||
	((tfxPage->seqStatus & TFX_PAGE_SEQ_WRITEPRIMARY) == 0)),
		     IO_SEQUENCE);

    /* Image ought to have the same resolution as the
     * parent image... right? */
    parentInfo = &tfxParentImg->imgInfo;
    XERR_GEN_CLEANUP(parentInfo->dwXResolution != dwXResolution, IO_BADPARAMETER);
    XERR_GEN_CLEANUP(parentInfo->dwYResolution != dwYResolution, IO_BADPARAMETER);

    if (dwBestStripHeight == 0) {

	/* General rule is that it shouldn't be more than 8K uncompressed */
	dwBestStripHeight = (8192 / (tfxParentImg->imgInfo.dwWidth *
	    dwBpp / 8));
    }

    tfxFile = TFX_PAGEFILE(tfxPage);

    /* Align long (for all TIFF IFDs) */
    /* This isn't correct, if the IFD is written later.  This assumes the */
    /* IFD is written immediately after this routine is called. */
/*    XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &ifdOffset)); */
    ifdOffset = 0;

    /* Allocate a new image */
    XERR_PASS_CLEANUP(wp_TfxNewImg(tfxPage, &newImg, ifdOffset));
    newImg->imgInfo.tfxImgType = IO_IMGTYPE_HINTS;
    newImg->imgInfo.dwImgLayer = 0;	/* Not in a layer */
    newImg->imgInfo.dwImgOffset = 0UL;

    /* Retain only the transferable information */
    newImg->imgInfo.dwXOffset = dwXOffset;
    newImg->imgInfo.dwYOffset = dwYOffset;
    newImg->imgInfo.dwXResolution = dwXResolution;
    newImg->imgInfo.dwYResolution = dwYResolution;
    newImg->imgInfo.dwWidth = dwWidth;
    newImg->imgInfo.dwHeight = dwHeight;
    newImg->imgInfo.dwBestStripHeight = dwBestStripHeight;
    newImg->imgInfo.xCompression = IO_COMP_UNKNOWN;
    switch (dwBpp) {
    case 1:
	newImg->imgInfo.xColorType = IO_COLORTYPE_BINARY;
	break;
    case 8:
	newImg->imgInfo.xColorType = IO_COLORTYPE_GRAY8;
	break;
    default:
	XERR_PASS_CLEANUP(IO_NOSUPPORT);
    }
	

    /* Write out the IFD */
/*    XERR_PASS_CLEANUP(wp_TfxQueueIFD(newImg)); */

    /* If this is the first sub-IFD, attatch it to the parent directly */
    if (tfxParentImg->fields.RenderingHintsIFD == 0) {
    
	/* Attach this image to its parent */
        if (tfxParentImg->fields.dataTransferred) {

	    /* See if the field is already written, and rewrite the data if it has */
	    if (tfxParentImg->fields.found.RenderingHintsIFD) {
		TFX_FILE *tfxFile;
		UInt32 lval;

		tfxFile = TFX_IMGFILE(tfxParentImg);

		/* Update the data location */
		TFX_SEEK(tfxFile, tfxParentImg->fieldLocation.RenderingHintsIFD);
		lval = ifdOffset;
		TFX_SWAPLONG(tfxFile, &lval);
		TFX_WRITE(tfxFile, &lval, sizeof(lval));
	    } else {
		
		/* It's already been written to the file, queue the field
		 * to be rewritten when the document is closed. */
		tfxParentImg->fields.IFDrewriteNeeded = 1;
	    }
	}

	/* Keep the value of the RenderingHintsIFD offset */
	tfxParentImg->fields.RenderingHintsIFD = ifdOffset;
	tfxParentImg->fields.found.RenderingHintsIFD = 1;

	/* Identify the parent, for when the true IFD is written (if later) */
	newImg->parentImg = tfxParentImg;

	/* Link it in */
	tfxParentImg->xHintsImgs = curImg = newImg;
	newImg = NULL;	/* Keep from freeing it, now that it's attached to a list */
    } else {
	TFX_IMG *chainImg = NULL;

	/* Attach to the NextIFD of the last in the chain */

	/* Make sure the chain has been traversed */
	XERR_WARNOK_CLEANUP(wp_TfxTraverseHintsImages(tfxParentImg));

	/* Search for the last one */
	chainImg = tfxParentImg->xHintsImgs;
	while (chainImg->next) {
	    chainImg = chainImg->next;
	}

	/* Append to the chain */
	XERR_PASS_CLEANUP(wp_TifFXWriteNextIFD(newImg,
	    chainImg, 0UL, ifdOffset));

	/* To make multiple hints images work, we need to keep track of the image
	 * that points to this image, so that we can write the RenderingHintsIFD
	 * field IFD location when it is really written. */
	newImg->parentImg = chainImg;

	/* Link it in */
	chainImg->next = curImg = newImg;
	newImg = NULL;	/* Keep from freeing it, now that it's attached to a list */
    }

    /* For now, only one attached hints image is allowed */
    tfxParentImg->imgInfo.dwHintsImages++;

    /* Return the image added */
    *pImgHandle = (UInt32)curImg;

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (newImg) {
	TFX_INSTANCE *instData;

	if (tfxParentImg) {
	    instData = TFX_IMGINST(tfxParentImg);
	    TFX_FREE(newImg);
	}
    }

     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Prepare to write image data to the TIFF-FX image
 *
 * LongDescription:::
 *  TFX_ImageWriteStart() initiates the write operation for an image. The
 *  image handle is that returned by TFX_AddImage(). This call will be
 *  followed by one or more calls to TFX_ImageWriteStrip(), and a closing
 *  call to TFX_ImageWriteFinish.
 *  <BR>
 *  pCompression points to the desired compression method. If the compression
 *  type is invalid or unsupported, the actual compression to be used is
 *  returned through this pointer. Note: if the image is the color plane for
 *  later use in color tokenization, the data is not actually written, but is
 *  stored for later extraction and assignment to symbols.
 *  <BR>
 *  If JPEG compression is used, pJPEGQuality points to the desired JPEG
 *  quality (1-100). If the JPEG quality changes from what is desired to what
 *  will be used, the actual value is passed back through the pointer. An
 *  unreasonable quality factor will yield a different output than passed in.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_AddImage()
 *  dwFlags::			Bit combination (OR'd) of (defined in iowmrc.h): <UL>
 *				  <LI>IO_IMGFLAGS_WhiteIsZero <EM>(default)</EM></LI>
 *				  <LI>IO_IMGFLAGS_BlackIsZero</LI>
 *				  <LI>IO_IMGFLAGS_ColorPlanar <EM>(default)</EM></LI>
 *				  <LI>IO_IMGFLAGS_ColorChunky</LI>
 *				  <LI>IO_IMGFLAGS_SubSampChroma <EM>(default)</EM></LI>
 *				  <LI>IO_IMGFLAGS_NoSubSampChroma</LI>
 *				</UL> 
 *  pxCompression::		Pointer to the requested compression type.
 *				IO_COMP_UNKNOWN yields default.
 *				The compression type used is passed back
 *				through this pointer
 *  pdwJPEGQuality::		Pointer to the requested JPEG quality (0-100)
 *				  0 - default used and passed back
 *				 75 - good quality, best compression
 *				 90 - better quality, good compression
 *				100 - best quality, okay compression
 *				The JPEG quality used is passed back
 *				through this pointer
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_ImageWriteStart(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwFlags, 
	IO_COMPRESSION *pxCompression,
	UInt32 *pdwJPEGQuality,
	IO_COLORSPACE *pxColorSpace
)	  
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_IMGINFO *pTfxImgInfo;
    UInt32 tmpBestStripHeight;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImgHandle);
    pTfxImgInfo = &tfxImg->imgInfo;
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxImg->seqStatus & TFX_IMG_SEQ_WRITESTATUS) !=
	TFX_IMG_SEQ_WRITEFINISH), IO_SEQUENCE);
    
    /* Store the flags parameters passed */
    tfxImg->imgFlags = dwFlags;
    pTfxImgInfo->xCompression = *pxCompression;
    tfxImg->JPEGquality = *pdwJPEGQuality;
    pTfxImgInfo->xColorSpace = *pxColorSpace;

    /* Check to see if compression options are valid */
    /* Make some page consistency checks */
    /* Note, it should be safe to ignore the recommendation for BestStripHeight,
     * since it was set originally in AddImage(). If it's been monkied with,
     * there's a good chance that we're doing a "quick copy", and that we
     * have no choice for picking the strip sizes.  This, in turn, can happen
     * when we think we have a TIFF-FX Profile M file, when in fact we have
     * an XCTIFF (any M profile) file instead. */
    tmpBestStripHeight = pTfxImgInfo->dwBestStripHeight;
    XERR_IGNORE(__error__ = TFX_CheckImageInfo((IO_PAGEHANDLE)tfxImg->tfxPage,
	&pTfxImgInfo->dwImgLayer, &pTfxImgInfo->dwImgOffset,
	&pTfxImgInfo->xColorType,
	&pTfxImgInfo->dwXOffset, &pTfxImgInfo->dwYOffset,
	&pTfxImgInfo->dwXResolution, &pTfxImgInfo->dwYResolution,
	&pTfxImgInfo->dwWidth, &pTfxImgInfo->dwHeight,
	&tmpBestStripHeight, pTfxImgInfo->pPalette,
	&pTfxImgInfo->xCompression, &pTfxImgInfo->xColorSpace));
    XERR_GEN_CLEANUP(__error__ && __error__ < IO_WARNINGS, __error__);

    /* See if we fall into the category of a single SubIFD that can span
     * all stripes, with strips aligning to the stripes */
    switch (tfxImg->tfxPage->tfxDoc->docInfo.profile) {
    case TFX_PROFILE_M1:
    case TFX_PROFILE_COH_M:
    case TFX_PROFILE_COH_M1:
        /* If the parent IFD has variable sized strips, then this one better
	 * also (unless it only spans two stripes, in which case the second
	 * strip can't be larger than the first).
	 * Also, if the YPosition places the image inside a stripe, then we
	 * might HAVE to use variable-sized strips.
	 * Note we couldn't really make the decision whether to use variable-
	 * height strips until we knew the primary IFD was written (did it
	 * have variable-height strips also?). */
	{
	    Bool useVariableStrips = 0;	    /* Assume we don't have to, for now */
	    UInt32 firstStripHeight = 0;
	    UInt32 secondStripHeight = 0;
	    TFX_IMG *primImg = (TFX_IMG *)tfxImg->tfxPage->pageInfo.tfxPrimImgHandle;

	    /* If the page uses variable-sized strips,
	     * we must use variable-sized strips */
	    if (tfxImg != primImg && (primImg->seqStatus & TFX_IMG_SEQ_USEVSS)) {
		useVariableStrips = 1;
	    }

	    /* Compute the size of the first strip */
	    firstStripHeight = 0;

	    /* Compute the size of the second strip (if any) */
	    secondStripHeight = 0;

	    /* If the first strip is smaller than the second strip,
	     * we must use variable-sized strips */

	    /* If the first strip is smaller than the second strip,
	     * we must use variable-sized strips */

	    /* Did we find a condition where variable-sized strips are required */
	    if (useVariableStrips) {
		tfxImg->seqStatus |= TFX_IMG_SEQ_USEVSS;
	    }
	}
	break;
    }

    /* If we're using variable-sized strips, reset best strip height */
    if (tfxImg->seqStatus & TFX_IMG_SEQ_USEVSS) {
	pTfxImgInfo->dwBestStripHeight = 0;
    }
 
    switch ((int)pTfxImgInfo->xCompression) {
    case IO_COMP_JPEG:
	if (tfxImg->JPEGquality == 0) {
	    tfxImg->JPEGquality = 75;	/* The default */
	} else if (tfxImg->JPEGquality > 100) {
	    tfxImg->JPEGquality = 100;	/* The max */
	} else if (tfxImg->JPEGquality < 1) {
	    tfxImg->JPEGquality = 1;	/* My hard-coded min */
	}
	break;
    default:
	tfxImg->JPEGquality = 0;
    }

    /* Now write the real IFD */
    XERR_PASS_CLEANUP(wp_TifFXWriteIFD(tfxImg));

    /* If this is a Layer image, or a Mask image, link it to the SubIFD chain of the parent */
    switch (tfxImg->imgInfo.tfxImgType) {
    case IO_IMGTYPE_LAYER:
    case IO_IMGTYPE_MASK:
	{
	    TFX_IMG *tfxPrimImg = (TFX_IMG *)tfxImg->tfxPage->pageInfo.tfxPrimImgHandle;

	    if (tfxPrimImg != tfxImg) {

		/* Attach this image to its parent */
		XERR_PASS_CLEANUP(wp_TifFXAttachChildIFD(tfxPrimImg, tfxImg, tfxImg->ifdOffset));
	    }
	}
    }

    switch (tfxImg->imgInfo.tfxImgType) {
    case IO_IMGTYPE_PRIMARY:
	{
	    TFX_PAGE *tfxPage = tfxImg->tfxPage;
	    UInt8 flags = 0;

	    /* Make sure we have the page table for the correct page */
/*	    XERR_PASS_CLEANUP(wp_TfxGeneratePageFlags(tfxPage, &flags)); */
	    XERR_PASS_CLEANUP(wp_TfxSetPageIFD(tfxPage->tfxDoc, tfxPage->pageInfo.dwPageOffset,
		flags, tfxImg->ifdOffset, tfxPage));
	}
	break;
    case IO_IMGTYPE_HINTS:
	/* There should be a parent image */
	XERR_GEN_CLEANUP(tfxImg->parentImg == NULL, IO_SEQUENCE);

	/* If there's already a hints image written to the parent image, then
	 * we need rewrite the location of this IFD.  The 'parentImg' should be
	 * filled in with the correct image linked "from". */
	if (tfxImg->parentImg->fieldLocation.RenderingHintsIFD) {
	    TFX_FILE *tfxFile;
	    UInt32 tempLong = tfxImg->ifdOffset;

	    tfxFile = TFX_IMGFILE(tfxImg);
	    TFX_SEEK(tfxFile, tfxImg->parentImg->fieldLocation.RenderingHintsIFD);
	    TFX_SWAPLONG(tfxFile, &tempLong);
	    TFX_WRITE(tfxFile, &tempLong, sizeof(tempLong));
	}
	tfxImg->parentImg->fields.RenderingHintsIFD = tfxImg->ifdOffset;
	break;
    }


    /* Return the compression to be used */
    *pxCompression = pTfxImgInfo->xCompression;
    *pdwJPEGQuality = tfxImg->JPEGquality;

    /* Update the sequence status */
    tfxImg->seqStatus |= TFX_IMG_SEQ_WRITESTART | TFX_IMG_SEQ_GETINFO;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


#ifndef NOIPCORE	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Copy user image buffer to intermediate TIFF buffer (PIXR)
 *
 * LongDescription:::
 *  wp_CopyToTempImage() copies fragments of the user buffer necessary to buffer
 *  up a TIFF strip, such that a whole TIFF strip can be accumulated before the
 *  write. The temporary PIXR tfxImg->tmpPixr is used as this intermediate
 *  buffer.
 *
 * Input:::
 *  tfxImg::			Data structure of valid image
 *  ppImgChannels::		Pointer to array of 1 to 3 image channels to
 *				copy from
 *  srcBPL::			Source image bytes per line (for one channel)
 *  pdwBufferRowOffset::	Pointer to source image row offset, updated for
 *				outer loop
 *  pdwImgRowOffset::		Pointer to image proper row offset, updated for
 *				outer loop
 *  pdwRowsToWrite::		Pointer to the number of rows in outer loop to
 *				write
 *  pdwTotalRowsFilled::	Pointer to the total number of rows buffered
 *				out
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
static ErrCode CDECL
wp_CopyToTempImage(
    TFX_IMG *tfxImg,
    UInt8 **ppImgChannels,
    Int32 srcBPL,
    UInt32 *pdwBufferRowOffset,
    UInt32 *pdwImgRowOffset,
    UInt32 *pdwRowsToWrite,
    UInt32 *pdwTotalRowsFilled
)
{
    UInt8 *srcPtr, *dstPtr;
    UInt32 byteCount;
    IPWState *pIPWState;
    Int32 channel, channels;
    PIXR *dstPixr, *pSubPixr;
    Int32 dstBPL = 0L;
    UInt32 rowsFilled = 0UL;
    UInt32 rowsAvailable;
    Int32 pixrRow, pixrHeight;
    PROLOGUE;

    /* Default to return nothing */
    *pdwTotalRowsFilled = 0UL;

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    /* Examine image features of the destination PIXR */
    dstPixr = tfxImg->tmpPixr;
    XERR_PASS_CLEANUP(w_pixrGetChannelCount(pIPWState, dstPixr, &channels));
    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, dstPixr, &pixrHeight));
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, dstPixr, &dstBPL));
    byteCount = (tfxImg->minBytesPerRow + 3) & ~3;
    XERR_GEN_CLEANUP((Int32)byteCount > dstBPL, IO_BADPARAMETER);

    /* Check to see if nothing needs to be done */
    if (tfxImg->tmpPixr == NULL || *pdwImgRowOffset < tfxImg->tmpPixrRow ||
	*pdwImgRowOffset >= tfxImg->tmpPixrRow + pixrHeight) {
	NORMAL_RETURN;
    }

    /* Make sure we haven't been asked to skip lines */
    pixrRow = *pdwImgRowOffset - tfxImg->tmpPixrRow;
    XERR_GEN_CLEANUP((UInt32)pixrRow != tfxImg->tmpPixrRows, IO_SEQUENCE);
    
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, dstPixr, channel,
					  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &dstPtr));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, dstPixr, &dstPtr));
	}
	srcPtr = ppImgChannels[channel] + *pdwBufferRowOffset * srcBPL;
	dstPtr += pixrRow * dstBPL;
	rowsAvailable = pixrHeight - pixrRow;
	if (rowsAvailable > *pdwRowsToWrite) {
	    rowsAvailable = *pdwRowsToWrite;
	}

	/* Copy the PIXR image to the user's image */
	while (rowsAvailable--) {
	    memcpy(dstPtr, srcPtr, byteCount);
	    srcPtr += srcBPL;
	    dstPtr += dstBPL;
	    rowsFilled++;
	}
    } /* for each channel */
    rowsFilled /= channels;
    tfxImg->tmpPixrRows += rowsFilled;	    /* Update intermediate PIXR row count */
    *pdwBufferRowOffset += rowsFilled;	    /* Update external source buffer row */
    *pdwImgRowOffset += rowsFilled;	    /* Update external image proper row */
    *pdwRowsToWrite -= rowsFilled;	    /* Update external "rows left" */
    (*pdwTotalRowsFilled) = rowsFilled;	    /* Return the count of rows filled */
    
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compress and write a strip of image data
 *
 * LongDescription:::
 *  TFX_ImageWriteStrip() compresses and writes a strip of image data. If
 *  necessary, the image data is broken up into TIFF strips, and stored as
 *  separate strips.  The last strip may be stored until the next block of
 *  data is added to complete a TIFF strip.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_AddImage()
 *  ppImagePlanes::		Pointer to image buffer(s)
 *  dwBytesPerRow::		Bytes per image row
 *  dwBufferRowOffset::		Row offset in image of the strip to write
 *  dwStripHeight::		Number of rows in PIXR to write
 *  dwImgRowOffset::		Row offset in image proper to write strip
 *
 * Output:::
 *  pRowsFilled::		Number of rows written
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  This function currently is written on top of another function that does
 *  the actual work. In this function we setup an intermediate image buffer
 *  the size of the "stored" strip height, and copy image data to this
 *  buffer as needed, and writing the intermediate buffer when filled.
 *  <BR>
 *  Part of this function should be moved into a less public file.
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_ImageWriteStrip(
	IO_IMGHANDLE tfxImgHandle,
	UInt8 **ppImagePlanes,
	UInt32 dwBytesPerRow,
	UInt32 dwBufferRowOffset,
	UInt32 dwStripHeight,
	UInt32 dwImgRowOffset,
	UInt32 *pRowsFilled
)
{
    TFX_INSTANCE *instData;
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_IMGINFO *pTfxImgInfo;
    IPWState *pIPWState;
    UInt32 channels;
    Int32 tmpPixrHeight = 0;
    ColorType ctype;
    UInt32 rowsFilled = 0UL;
    UInt32 totalRowsFilled = 0UL;
    Int32 writeRows;
    PROLOGUE;

    /* Default to return nothing */
    *pRowsFilled = 0UL;
    
    TFX_CheckHandle(tfxImgHandle);

    instData = TFX_IMGINST(tfxImg);    
    pTfxImgInfo = &tfxImg->imgInfo;
    channels = IO_CHANNELS(pTfxImgInfo->xColorType);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxImg->seqStatus & TFX_IMG_SEQ_WRITESTATUS) !=
		     TFX_IMG_SEQ_WRITESTART), IO_SEQUENCE);

    /* Check for proper compression mode */
    XERR_GEN_CLEANUP(((tfxImg->imgFlags & IO_IMGFLAGS_Compression) !=
	IO_IMGFLAGS_Uncompressed), IO_SEQUENCE);
/* IPCORE */
    /* Reference the IPWrap state */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    /* Get the temporary PIXR height, if it exists */
    if (tfxImg->tmpPixr != NULL) {
	XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, tfxImg->tmpPixr,
					  &tmpPixrHeight));
    }
/* IPCORE */
    /* Loop until the user's image buffer (region) is emptied */
    while (dwStripHeight) {

	/* Check for overfill */
	if (dwImgRowOffset + dwStripHeight > pTfxImgInfo->dwHeight) {
	    dwStripHeight = pTfxImgInfo->dwHeight - dwImgRowOffset;
	}
	if (dwStripHeight <= 0) {
	    break;
	}

	/* Compute the number of strips to be written (to match TIFF strip) */
	if (pTfxImgInfo->dwBestStripHeight) {
	    writeRows = pTfxImgInfo->dwBestStripHeight;
	} else {
	    writeRows = dwStripHeight;
	}
	if ((UInt32)writeRows > pTfxImgInfo->dwHeight - tfxImg->tmpPixrRow) {
	    writeRows = pTfxImgInfo->dwHeight - tfxImg->tmpPixrRow;
	}

	/* If we don't have an intermediate PIXR to handle buffering,
	 * create it now */
	if (tfxImg->tmpPixr == NULL) {
	    ScaledInt1616 xres, yres;
		
	    /* Prepare a temporary PIXR */
	    xres = Int32_TO_ScaledInt1616(pTfxImgInfo->dwXResolution);
	    yres = Int32_TO_ScaledInt1616(pTfxImgInfo->dwYResolution);
	    switch ((int)pTfxImgInfo->xColorType) {
	    case IO_COLORTYPE_COLOR24:
		ctype = cColorRGB;
		break;
	    case IO_COLORTYPE_BINARY:
	    case IO_COLORTYPE_GRAY4:
	    case IO_COLORTYPE_COLOR4:
	    case IO_COLORTYPE_GRAY8:
	    case IO_COLORTYPE_COLOR8:
		ctype = cColorGray;
		break;
	    default:
		goto ERR_CLEANUP;
	    }
	    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, pTfxImgInfo->dwWidth,
		writeRows, IO_BPPE(pTfxImgInfo->xColorType),
		xres, yres, 32, IO_CHANNELS(pTfxImgInfo->xColorType),
		ctype, NULL, &tfxImg->tmpPixr));
	    XERR_PASS_CLEANUP(w_pixrGetHeight(pIPWState, tfxImg->tmpPixr, &tmpPixrHeight));
	} /* If no intermediate image buffer */

	/* Fill up the intermediate image if we haven't already, update:
	 *	buffer row offset	    dwBufferRowOffset
	 *	image row offset	    dwImgRowOffset
	 *	rows left to buffer out	    dwStripHeight
	 * We update:
	 *	number of rows in buffer    tfxImg->tmpPixrRows
	 */
	XERR_PASS_CLEANUP(wp_CopyToTempImage(tfxImg, ppImagePlanes, dwBytesPerRow,
	    &dwBufferRowOffset, &dwImgRowOffset, &dwStripHeight, &rowsFilled));
	totalRowsFilled += rowsFilled;

	/* If the intermediate buffer is full, write out the strip */
	if (tfxImg->tmpPixrRows == pTfxImgInfo->dwBestStripHeight ||
	    tfxImg->tmpPixrRow + tfxImg->tmpPixrRows == pTfxImgInfo->dwHeight ||
	    pTfxImgInfo->dwBestStripHeight <= 0) {

	    /* Convert image order from user's specifications to PIXR */

	    /*********** for now we assume the user has passed in PIXR image
	     *********** order
	     ***********/

	    /* Write out the PIXR (should now be one full TIFF strip) */
	    XERR_PASS_CLEANUP(w_TfxWriteStrip(pIPWState, tfxImgHandle,
		tfxImg->tmpPixr, 0, writeRows, tfxImg->tmpPixrRow,
		&rowsFilled));
	    tfxImg->tmpPixrRow = dwImgRowOffset;
	    tfxImg->tmpPixrRows = 0;
	}
    }

    /* Return the number of rows written out */
    *pRowsFilled = totalRowsFilled;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}
#endif	/* NOIPCORE */

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Write a strip of compressed image data
 *
 * LongDescription:::
 *  TFX_CompressedWriteStrip() writes a strip of already compressed image
 *  data. The strip *must* be the strip height as given in the parameter
 *  dwBestStripHeight during the call to TFX_AddImage(), except for the
 *  very last strip which may fill up to the end of the image.  The strip
 *  *must* also lie on a strip boundary.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_AddImage()
 *  dwImgRowOffset::		Row offset in image proper to write strip
 *  dwStripHeight::		Number of rows in PIXR to write
 *  pCmpBuffer::		Pointer to the compressed image buffer
 *  dwCmpBytes::		Bytes of compressed data found in buffer
 *
 * Output:::
 *  pdwBytesStored::		Number of bytes written
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_CompressedWriteStrip(
	IO_IMGHANDLE tfxImgHandle,
	UInt32 dwImgRowOffset,
	UInt32 dwStripHeight,
	UInt8 *pCmpBuffer,
	UInt32 dwCmpBytes,
	UInt32 *pdwBytesStored
)
{
    TFX_INSTANCE *instData;
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_IMGINFO *pTfxImgInfo;
    UInt8 *tmpCmp = NULL;
    UInt32 tmpCmpBytes = 0UL, tmpCmpBufsize = 0UL;
    PROLOGUE;

    /* Default to return nothing */
    *pdwBytesStored = 0UL;

    TFX_CheckHandle(tfxImgHandle);

    instData = TFX_IMGINST(tfxImg);    
    pTfxImgInfo = &tfxImg->imgInfo;
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxImg->seqStatus & TFX_IMG_SEQ_WRITESTATUS) !=
	TFX_IMG_SEQ_WRITESTART), IO_SEQUENCE);

    /* Check for proper compression mode */
#if 0
/* Must change this later */
/*    XERR_GEN_CLEANUP(((tfxImg->imgFlags & IO_IMGFLAGS_Compression) != */
/*	IO_IMGFLAGS_Compressed), IO_SEQUENCE); */
#endif

    /* Keep track of the row */
    tfxImg->tmpPixrRow = dwImgRowOffset;

    if (pTfxImgInfo->dwBestStripHeight > 0) {
	/* We *must* have the data in a compressed strip already */
/*	XERR_GEN_CLEANUP((dwStripHeight != pTfxImgInfo->dwBestStripHeight) && */
/*	    (dwStripHeight + dwImgRowOffset != pTfxImgInfo->dwHeight), */
/*	    IO_BADPARAMETER); */
	/* We *should* have the data in a compressed strip already */
	XERR_GEN_CLEANUP((dwStripHeight != pTfxImgInfo->dwBestStripHeight) &&
	    (dwStripHeight + dwImgRowOffset < pTfxImgInfo->dwHeight),
	    IO_BADPARAMETER);

	/* The strip must lie on a strip boundary */
	XERR_GEN_CLEANUP(dwImgRowOffset % pTfxImgInfo->dwBestStripHeight != 0,
	    IO_BADPARAMETER);
    } else {
	/* Keep track of the variable-sized strip */
	tfxImg->tmpPixrRows = dwStripHeight;
    }
    
    /* Internal routines use the compressed buffer attached to the
     * image handle. Repoint previously used values to the ones passed
     * in to this function, then restore later. Ya, ugly, but I haven't
     * had time to consider a better alternative. */
    tmpCmp = tfxImg->cmpBuffer;
    tmpCmpBytes = tfxImg->cmpBytes;
    tmpCmpBufsize = tfxImg->cmpBufsize;

    /* Repoint internal buffer to this one */
    tfxImg->cmpBuffer = pCmpBuffer;
    tfxImg->cmpBytes = dwCmpBytes;
    if (tfxImg->cmpBufsize < tfxImg->cmpBytes) {
	tfxImg->cmpBufsize = tfxImg->cmpBytes;
    }
    XERR_PASS_CLEANUP(wp_TifFXWriteStrip(tfxImg));

    /* Restore the internal state */
    tfxImg->cmpBuffer = tmpCmp;
    tfxImg->cmpBytes = tmpCmpBytes;
    tfxImg->cmpBufsize = tmpCmpBufsize;

    /* Update the status (used as a check for number of rows written,
     * see TFX_ImageWriteFinish()) */
    tfxImg->tmpPixrRow += dwStripHeight;

    /* Return the number of bytes written */
    *pdwBytesStored = dwCmpBytes;

    /* Reset the mode */
#if 0
/*    tfxImg->imgFlags &= ~IO_IMGFLAGS_Compressed; */
#endif

	
    NORMAL_RETURN;
    
 ERR_CLEANUP:
    if (tmpCmp) {
	tfxImg->cmpBuffer = tmpCmp;
	tfxImg->cmpBytes = tmpCmpBytes;
	tfxImg->cmpBufsize = tmpCmpBufsize;
    }
    
    ERR_RETURN;
}

	
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add a compressed symbol dictionary to the file
 *
 * LongDescription:::
 *  TFX_StoreSymbolDictionary() adds a compressed symbol dictionary to the
 *  file, adding an entry in the shared resources.
 *  <BR>
 *  If the dictionary contains extensions, then the number of bytes in the
 *  entire block will be larger than that of the back-end dictionary.
 *
 * Input:::
 *  tfxDocHandle::		Document handle
 *  pDictBlock::		JBIG2 compressed dictionary
 *  dwDictBytes::		Number of bytes in back-end dictionary
 *  dwDictBlockBytes::		Number of bytes in whole block (may include
 *				extensions)
 *
 * Output:::
 *  pdwDictID::			Returned dictionary ID
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_StoreSymbolDictionary(
	IO_DOCHANDLE tfxDocHandle,
	UInt8 *pDictBlock,
	UInt32 dwDictBytes,
	UInt32 dwDictBlockBytes,
	UInt32 *pdwDictID
)
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    UInt8 *dictHeader = NULL;
    Int32 offset;
    UInt8 dictVersion;
    UInt32 tempLong;
    PROLOGUE;


    TFX_CheckHandle(tfxDoc);

    instData = TFX_DOCINST(tfxDoc);
    tfxFile = TFX_DOCFILE(tfxDoc);

    /* Check sequence */

    /* Check parameters */
    XERR_GEN_CLEANUP(dwDictBlockBytes < dwDictBytes, IO_BADPARAMETER);

    /* Allocate a buffer for the dictionary and header */
    IO_CHECKMALLOC(dictHeader = (UInt8 *)TFX_MALLOC(cTfxTCDictHeaderSize));

    /* Add the version to the header */
    dictVersion = cTfxTCDictVersion;
    memcpy(dictHeader + cTfxTCPosDictVersion, &dictVersion,
	sizeof(dictVersion));

    /* Store the size of the JBIG2 dictionary in the header */
    tempLong = dwDictBytes;
    TFX_SWAPLONG(tfxFile, &tempLong);
    memcpy(dictHeader + cTfxTCPosDictSize, &tempLong, sizeof(tempLong));

    /* Store the header in the file */
    XERR_PASS_CLEANUP(wp_TfxFileLongAlign(tfxFile, &offset));
    TFX_WRITE(tfxFile, dictHeader, cTfxTCDictHeaderSize);

    /* Store the data in the file */
    TFX_WRITE(tfxFile, pDictBlock, (Int32)dwDictBlockBytes);

    /* Store the overall size in the shared resources library */
    XERR_PASS_CLEANUP(wp_TfxSetResourceOffset(tfxDoc,
	cTfxTCDictHeaderSize + dwDictBlockBytes, offset, TFX_TIFFCOMP_JBIG2,
	0, 0, &tfxDoc->nextDocTokenID));

    /* Return the dict ID */
    *pdwDictID = tfxDoc->nextDocTokenID++;

    /* Fall through and clean up the temporary buffer */
ERR_CLEANUP:

    if (dictHeader) {
	TFX_FREE(dictHeader);
	dictHeader = NULL;
    }
    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Store the JBIG2 compressed image (position block)
 *
 * LongDescription:::
 *  TFX_StorePositionBlock() places the appropriate wrapper in front of a JBIG2
 *  "position block" or compressed image data, and stores the data as a single
 *  TIFF strip.
 *  <BR>
 *  The symbol dictionary or dictionaries must have been written before this
 *  call, via TFX_StoreSymbolDictionary().  Each stored dictionary is given
 *  an ID, which can be used in the dictIDList[], in the order that the
 *  dictionaries are to be concatenated.
 *
 * Input:::
 *  tfxImgHandle::		Handle to valid image
 *  pPosBlock::			JBIG2 (embedded) position block
 *  dwPosBytes::		Bytes in posBlock[]
 *  dwDictIDList::		List of dictionary IDs associated with this
 *				position block
 *  wDictCount::		Number of dictionary IDs in dwDictIDList[]
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *  The format of the header is:
 *	1 byte	- version number
 *	4 bytes	- position block byte count
 *	2 bytes	- Number of dictionaries associated with the position block
 *	4 bytes	- First dictionary ID
 *	...	- Next dictionary ID
 *      followed by the embedded position block
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_StorePositionBlock(
	IO_IMGHANDLE tfxImgHandle,
	UInt8 *pPosBlock,
	UInt32 dwPosBytes,
	UInt32 *dwDictIDList,
	UInt16 wDictCount
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_FILE *tfxFile;
    TFX_DOC *tfxDoc;
    UInt8 *posStrip = NULL;
    UInt32 headerSize;
    UInt8 posBlockVersion;
    UInt16 tempShort;
    UInt32 dict, tempLong;
    UInt32 binaryCompBytes = 0;
    PROLOGUE;

    TFX_CheckHandle(tfxImg);
    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);
    tfxDoc = tfxImg->tfxPage->tfxDoc;

    /* Check parameters */
    XERR_GEN_CLEANUP(dwDictIDList == NULL, IO_BADPARAMETER);

    /* It's possible to fill in a 0-byte entry, so the nextDocTokenID is not
     * always up-to-date.
     */
#if 0
    /* Check the dictionary IDs */
    for (dict = 0; dict < wDictCount; dict++) {

	/* Shouldn't exceed what's already been stored */
	XERR_GEN_CLEANUP(dwDictIDList[dict] >= tfxDoc->nextDocTokenID, IO_BADPARAMETER);
    }
#endif

    /* Allocate a buffer for the position block header */
    headerSize = cTfxTCPosBlockHeaderSize(wDictCount);
    IO_CHECKMALLOC(posStrip = (UInt8 *)TFX_MALLOC(headerSize + dwPosBytes));
    memset(posStrip, 0, headerSize);

    /* Set to the current position block version */
    posBlockVersion = cTfxTCPositionVersion;
    memcpy(posStrip + cTfxTCPosPosBlockVersion, &posBlockVersion,
	sizeof(posBlockVersion));

    /* Add the dictionary count */
    tempShort = wDictCount;
    TFX_SWAPSHORT(tfxFile, &tempShort);
    memcpy(posStrip + cTfxTCPosPosBlockDictCount, &tempShort,
	sizeof(tempShort));
	    
    /* Add the dictionary ID(s) to the header */
    for (dict = 0; dict < wDictCount; dict++) {
	tempLong = dwDictIDList[dict];
	TFX_SWAPLONG(tfxFile, &tempLong);
	memcpy(posStrip + cTfxTCPosPosBlockDictIDs +
	    dict * sizeof(tempLong), &tempLong,
	    sizeof(tempLong));
    }

    /* Add the position block size to the header */
    tempLong = dwPosBytes;
    TFX_SWAPLONG(tfxFile, &tempLong);
    memcpy(posStrip + cTfxTCPosPosBlockSize(wDictCount), &tempLong,
	sizeof(tempLong));

    /* Copy the position block */
    memcpy(posStrip + headerSize, pPosBlock, dwPosBytes);

    /* Attach to the image (for writing a compressed strip) */
    tfxImg->cmpBuffer = posStrip;
    tfxImg->cmpBufsize = tfxImg->cmpBytes = headerSize + dwPosBytes;

    /* This should have been written already */
    XERR_GEN_CLEANUP(tfxImg->reservedFieldCount < 1, IO_INTERNAL_ERROR);
/*    XERR_PASS_CLEANUP(wp_TifFXWriteIFD(tfxImg)); */

    /* Write the buffer to the file */
    if (tfxImg->cmpBytes) {
	XERR_PASS_CLEANUP(wp_TifFXWriteStrip(tfxImg));
    }

    /* Complete the write for this image (do cleanup and memory freeing) */
    if (tfxImg->cmpBuffer) {
	TFX_FREE(tfxImg->cmpBuffer);
	posStrip = tfxImg->cmpBuffer = NULL, tfxImg->cmpBufsize = tfxImg->cmpBytes = 0UL;
    }

    /* Attach the FG image using the same compressed image stream, and
     * removed the color section from the size of the stream just added */
    if (binaryCompBytes) {
	IO_IMGHANDLE fgImg;

	XERR_PASS_CLEANUP(TFX_ShareColorTokenImage((IO_IMGHANDLE)tfxImg,
	    binaryCompBytes, IO_COLORTYPE_COLOR24, &fgImg));
    }

    /* Update the status (used as a check for number of rows written,
     * see TFX_ImageWriteFinish()).
     * Assume it's the entire image (I should probably allow for JBIG2 TIFF strips). */
    tfxImg->tmpPixrRow += tfxImg->imgInfo.dwHeight;

    /* Preserve the header size, in case a color extension is added
     * (TFX_ShareColorTokenImage()). */
    tfxImg->headerBytes = headerSize;


    /* Fall through to free up any temporary buffers */

 ERR_CLEANUP:

    if (posStrip) {
	TFX_FREE(posStrip);
	posStrip = NULL;
    }
    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Complete an image write
 *
 * LongDescription:::
 *  TFX_ImageWriteFinish() cleans up the write operation state and performs
 *  consistecy checks.  Fails if correct number of rows not written.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_AddImage()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_ImageWriteFinish(
	IO_IMGHANDLE tfxImgHandle
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_MISCMEM *mmptr, *next_mm;
    TFX_INSTANCE *instData;
    PROLOGUE;
    
    TFX_CheckHandle(tfxImgHandle);
    
    instData = TFX_IMGINST(tfxImg);
    
    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxImg->seqStatus & TFX_IMG_SEQ_WRITESTATUS) !=
		      TFX_IMG_SEQ_WRITESTART), IO_SEQUENCE);
    
    /* Free intermediate data */
    if (tfxImg->cmpBuffer) {
	TFX_FREE(tfxImg->cmpBuffer);
	tfxImg->cmpBuffer = NULL, tfxImg->cmpBufsize = tfxImg->cmpBytes = 0UL;
    }
#ifndef NOIPCORE
    if (tfxImg->tmpPixr) {
	IPWState *pIPWState;

	/* Reference the IPWrap state */
	XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

	XERR_PASS_CLEANUP(w_destroyPixr(pIPWState, &tfxImg->tmpPixr));
    }
#endif /* NOIPCORE */
    if ((mmptr = tfxImg->tmpCmpData) != NULL) {
	while (mmptr) {
	    next_mm = mmptr->next;
	    TFX_FREE(mmptr);
	    mmptr = next_mm;
	}
	tfxImg->tmpCmpData = NULL;
    }

    /* If we're writing variable-sized strips, finish them */
    if (tfxImg->fields.StripsPerImage == 0 && tfxImg->strips.stripCount > 1) {
	XERR_PASS_CLEANUP(wp_TifFXWriteStripSizes(tfxImg));
    }
    
    /* Reset the read status */
    tfxImg->seqStatus &= ~TFX_IMG_SEQ_WRITESTATUS;	/* Clear write status */
    tfxImg->seqStatus |= TFX_IMG_SEQ_WRITEFINISH;	/* Set to finished */

    /* Now check to see if the right number of rows was written */
/*    XERR_GEN(tfxImg->tmpPixrRow != tfxImg->imgInfo.dwHeight, IO_SEQUENCE); */
    /* Found some problem TIFF files that have a bigger last strip than needed.
     * Also found that some 0-byte strips may not cause the row count to
     * hit the end.  The only thing we can do is not make this check: */
/*    XERR_GEN(tfxImg->tmpPixrRow < tfxImg->imgInfo.dwHeight, IO_SEQUENCE); */

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
}


#if !defined(NOIPCORE) && !defined(NODGP)
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Store a binary image internally using tokenization
 *
 * LongDescription:::
 *  TFX_TokenizeImage() takes as input an entire binary image and performs
 *  the necessary tasks to add the page to the existing (or created) token
 *  dictionary. All image data is left undisturbed, and the tokenized version
 *  of the entire page can be reconstructed from tokens found in the dictionary,
 *  when completed, albeit with the lossy artifacts of the algorithm.
 *  <BR>
 *  Note, the image is not stored off at this time, since the token dictionary
 *  will not be complete until the last page is tokenized (see
 *  TFX_WriteTokenImage() and TFX_WriteTokenDictionary(), or
 *  TFX_WriteTokenImages()).
 *  <BR>
 *  Also, if a sequence of calls to TFX_TokenizeImage() is interrupted by a
 *  call to either TFX_WriteTokenImage(), TFX_WriteTokenDictionary() or
 *  TFX_WriteTokenImages(), then the next call to TFX_TokenizeImage()
 *  following those calls will cause the previous token dictionary to be
 *  flushed (if not), and a new one to be initialized.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_AddImage()
 *  pImgData::			One channel only.  Note, it must point to the
 *				first actual pixel, but must be part of a
 *				bigger image that includes a 32-pixel frame,
 *				plus necessary padding to make each scanline
 *				long word aligned (JBIG2 requirements)
 *  dwBytesPerRow::		Bytes per row
 *  dwFlags::			Tokenization flags (see TOKENIZE_... in tfx.h)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_TokenizeImage(
	IO_IMGHANDLE tfxImgHandle,
	UInt8 *pImgData,
	UInt32 dwBytesPerRow,
	UInt32 dwFlags
)
{
    TFX_INSTANCE *instData = NULL;
    TFX_FILE *tfxFile;
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    UInt8 *imgptr = NULL;
    IO_COLORSPACE colorSpace = IO_COLORSPACE_DEFAULT;
    PROLOGUE;

    TFX_CheckHandle(tfxImgHandle);

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(((tfxImg->seqStatus & TFX_IMG_SEQ_WRITESTATUS) !=
	TFX_IMG_SEQ_WRITEFINISH), IO_SEQUENCE);
    XERR_GEN_CLEANUP(tfxImg->imgInfo.tfxImgType != IO_IMGTYPE_PRIMARY,
		     IO_BADPARAMETER);
#if 0	/* To be added shortly */
    XERR_GEN_CLEANUP((dwFlags & TOKENIZE_COLORTOO) &&
		     ((tfxImg->tfxPage->seqStatus & TFX_PAGE_SEQ_TKCOLORADDED) == 0)
		     || (tfxImg->tfxPage->seqStatus & TFX_PAGE_SEQ_MULTIDICT) == 0),
		     IO_SEQUENCE);
    XERR_GEN_CLEANUP((dwFlags & TOKENIZE_COLORTOO) &&
		     tfxImg->tfxPage->fgImg == NULL, IO_INTERNAL_ERROR);
#endif

    if (tfxImg->tfxPage->tfxDoc->docInfo.version < 3 ||
	(dwFlags & TOKENIZE_ON) == TOKENIZE_OFF) {

	/* Tokenization not supported for this version (or not enabled),
	 * let's be nice and store it G4 */
	{
	    UInt8 *imagePlanes[3];
	    UInt32 rowsFilled;
	    IO_COMPRESSION compression = IO_COMP_MMR;
	    UInt32 quality = 0UL;

	    /* Begin writing */
	    XERR_PASS_CLEANUP(TFX_ImageWriteStart(tfxImgHandle, 0UL, &compression,
		&quality, &colorSpace));

	    /* Write the whole strip */
	    imagePlanes[0] = pImgData;
	    XERR_PASS_CLEANUP(TFX_ImageWriteStrip(tfxImgHandle, imagePlanes,
		dwBytesPerRow, 0, tfxImg->imgInfo.dwHeight, 0, &rowsFilled));

	    /* Finish writing */
	    XERR_PASS_CLEANUP(TFX_ImageWriteFinish(tfxImgHandle));
	}

	NORMAL_RETURN;
    }

    /* MRC will only tolerate one compression scheme per scanline per layer,
     * so we must use multi-dict mode (residual is folded into JBIG2 stream) */
    tfxImg->tfxPage->seqStatus |= TFX_PAGE_SEQ_MULTIDICTS;

    /* Transfer the color tokenization mode to the page */
    if (dwFlags & TOKENIZE_COLORTOO) {
	tfxImg->tfxPage->seqStatus |= TFX_PAGE_SEQ_TKCOLORADDED;
#if 0	/* To be added shortly */
	tfxImg->tfxPage->tfxDoc->dictFlags |= TOKEN_COLORADDED;
#endif
    }

    /* This is always true for MRC */
    tfxImg->tfxPage->tfxDoc->seqStatus |= TFX_DOC_SEQ_MULTIDICTS;

#ifdef NODGP
    if (dwFlags & TOKENIZE_LOSSLESS) {
	tfxImg->imgInfo.xCompression = IO_COMP_LLTOKENIZATION;
    } else {
	tfxImg->imgInfo.xCompression = IO_COMP_TOKENIZATION;
    }
#else
    if (dwFlags & TOKENIZE_LOSSLESS) {
	tfxImg->imgInfo.xCompression = IO_COMP_LLJBIG2;
    } else {
	tfxImg->imgInfo.xCompression = IO_COMP_JBIG2;
    }
#endif

    instData = TFX_IMGINST(tfxImg);
    tfxFile = TFX_IMGFILE(tfxImg);

    /* Check the current tokenization status */
    if (tfxImg->tfxPage->tfxDoc->seqStatus & TFX_DOC_SEQ_TOKENSTART) {

	/* If this is the case, we must save the previous tokenized
	 * images before proceeding. This is a bad state to be in,
	 * and should have already been taken care of.
	 */
	if (tfxImg->tfxPage->tfxDoc->seqStatus & TFX_DOC_SEQ_TOKENDICTSAVED) {
	    XERR_PASS_CLEANUP(TFX_WriteTokenImages((IO_DOCHANDLE)
						   tfxImg->tfxPage->tfxDoc));
	}
    }

    /* Tokenize the image */
    XERR_PASS_CLEANUP(wp_TfxTokenizeImage(tfxImg, pImgData, dwBytesPerRow,
	dwFlags));

    /* Update the sequence status */
    tfxImg->seqStatus |= TFX_IMG_SEQ_GETINFO | TFX_IMG_SEQ_WRITESTART;

    /* This should always be TRUE */
    if (tfxImg->imgInfo.tfxImgType == IO_IMGTYPE_PRIMARY) {
	tfxImg->tfxPage->seqStatus |= TFX_PAGE_SEQ_WRITEPRIMARY;
    }	

    NORMAL_RETURN;

 ERR_CLEANUP:
    if (imgptr) {
	TFX_FREE(imgptr);
    }
    NORMAL_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Store all previously tokenized images, and token dictionary
 *
 * LongDescription:::
 *  TFX_WriteTokenImages() is a replacement to having to loop through
 *  TFX_WriteTokenImage(), since the list of images that have been tokenized
 *  but not resolved to a finished token dictionary are maintained via the
 *  document handle. This function will also call TFX_WriteTokenDictionary().
 *  See the description for both of these functions for further explanation.
 *
 * Input:::
 *  tfxDocHandle::		Valid document handle, returned from
 *				TFX_CreateDocument() or
 *				TFX_OpenDocumentReadWrite()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_WriteTokenImages(
	IO_DOCHANDLE tfxDocHandle
)
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    IPWState *pIPWState;
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);

    /* Get shared IPCore state struct */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxDoc, &pIPWState));

    /* Store the symbol IDs & positions, and the residual image, for each
     * tokenized (but unresolved) image. */
    if (tfxDoc->seqStatus & (TFX_DOC_SEQ_TOKENSTART)) {
	TFX_PAGE *tfxPage;
	UInt32 page, pages;
	TFX_IMG *tfxImg;

	/* Write the token dictionary, if it hasn't already been written */
	XERR_PASS_CLEANUP(TFX_WriteTokenDictionary(tfxDocHandle));

	/* For each page that has been tokenized, store off the symbol IDs
	 * & posisions, and the residual image */
/*	pages = tfxDoc->pageTableHeader.pageCount; */
	pages = tfxDoc->docInfo.wNPages;
	tfxPage = tfxDoc->pages;
	for (page = 0; page < pages; page++) {
	    if (tfxPage == NULL) {
		break;
	    }
	    if (tfxPage->seqStatus & (TFX_PAGE_SEQ_TOKENSTART)) {
		tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;
		XERR_PASS_CLEANUP(TFX_WriteTokenImage((UInt32)tfxImg));
    	    }
	    tfxPage = tfxPage->next;
	}

	/* Clear the tokenization state */
	tfxDoc->seqStatus &=
	    ~(TFX_DOC_SEQ_TOKENDICTSAVED | TFX_DOC_SEQ_TOKENSTART |
	    TFX_DOC_SEQ_MULTIDICTS | TFX_DOC_SEQ_SHAREDDICT);
    }

    /* Fall through to do the rest of the cleanup */
 ERR_CLEANUP:
    
    /* Free up the shared data */
    if (tfxDoc->tokEncodeHandle) {
    	DgpApi_close(tfxDoc->tokEncodeHandle);
	tfxDoc->tokEncodeHandle = 0;
    }

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Store a previously tokenized image
 *
 * LongDescription:::
 *  TFX_WriteTokenImage() is the partner to TFX_TokenizeImage(), which
 *  previously converted an image to tokens, and preserved it internally.
 *  This call resolves the image into token IDs and positions, and a residual
 *  image, and stores the data for both.  The token IDs and positions are
 *  stored in the token dictionary, which must be resolved (cannot be added
 *  to anymore) and is consolidated (like symbols mapped together) and stored.
 *  Then symbols from the image that repeat are converted to a compressed
 *  tokenization stream and stored. Finally, the symbols that do not repeat
 *  are retained as a separate image and stored as run-lengths (MMR), the
 *  residual image -- a SubIFD.
 *  <BR>
 *  This function should only be called after all of the binary images which
 *  are to share the same token dictionary have been tokenized. Implied is the
 *  need to store the token dictionary, which may have either taken place via
 *  a call to TFX_WriteTokenDictionary(), or is done automatically during the
 *  first call to TFX_WriteTokenImage().
 *  <BR>
 *  Internally, the token IDs are resolved (if not already from a previous call
 *  to TFX_WriteTokenImage() or TFX_WriteTokenDictionary(), and the image
 *  fragments with a count across the pages of 1 (this page only) are removed
 *  from the token dictionary, and placed back on a blank image of the original
 *  size, and stored as a residual image. The remaining tokens (with count 2 or
 *  greater), are then stored with their locations, using the Tundra format, as
 *  proposed for JBIG-2, and indices are assigned Huffman codes, or Lempel-Ziv
 *  compressed.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_AddImage()
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_WriteTokenImage(
	IO_IMGHANDLE tfxImgHandle
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_PAGE *tfxPage;
    PROLOGUE;

    TFX_CheckHandle(tfxImgHandle);
    tfxPage = tfxImg->tfxPage;

    /* Write the token dictionary, if it hasn't already been written */
    XERR_PASS_CLEANUP(TFX_WriteTokenDictionary((UInt32)tfxPage->tfxDoc));

    /* If color tokenization is used, or residuals aren't used,
     * store off the unique symbols in a separate dictionary */
    if (tfxPage->seqStatus & TFX_PAGE_SEQ_MULTIDICTS) {
	XERR_PASS_CLEANUP(wp_TfxWriteTokenDictionary(tfxPage->tfxDoc, tfxPage));
    }

    /* Store the symbol IDs & positions, and the residual image */
    XERR_PASS_CLEANUP(wp_TfxWriteTokenImage(tfxImg));

    /* Clear the tokenization state */
    tfxImg->seqStatus &= ~TFX_IMG_SEQ_TOKENSTART;
    tfxPage->seqStatus &= ~TFX_PAGE_SEQ_TOKENSTART;

    /* If appropriate mark as completely written */
    if (tfxPage->seqStatus & TFX_PAGE_SEQ_WRITEPENDING) {
	tfxPage->seqStatus &= ~TFX_PAGE_SEQ_WRITESTATUS;
	tfxPage->seqStatus |= TFX_PAGE_SEQ_WRITEFINISH;
	tfxPage->seqStatus |= TFX_PAGE_SEQ_READSTART;	/* Mark as readable */
    }

    /* Check to see if it was the last page to be stored for the given
     * dictionary */
    if (tfxPage->tfxDoc->seqStatus & (TFX_DOC_SEQ_TOKENSTART)) {
	TFX_PAGE *pTfxPage;
	UInt32 page, pages;
	Bool pagesLeft = 0;

	/* Check each page */
/*	pages = tfxPage->tfxDoc->pageTableHeader.pageCount; */
	pages = tfxPage->tfxDoc->docInfo.wNPages;
	pTfxPage = tfxPage->tfxDoc->pages;
	for (page = 0; page < pages; page++) {
	    if (pTfxPage == NULL) {
		break;
	    }
	    if (pTfxPage->seqStatus & (TFX_PAGE_SEQ_TOKENSTART)) {
		pagesLeft++;
		break;
    	    }
	    pTfxPage = pTfxPage->next;
	}

	/* If all the tokenized images have been stored, reallow tokenization
	 */
	if (pagesLeft == 0) {
	    /* Clear the tokenization state */
	    tfxPage->tfxDoc->seqStatus &=
		~(TFX_DOC_SEQ_TOKENDICTSAVED | TFX_DOC_SEQ_TOKENSTART |
		TFX_DOC_SEQ_MULTIDICTS | TFX_DOC_SEQ_SHAREDDICT);
	}
    }
    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Store the dictionary of accumulated tokens
 *
 * LongDescription:::
 *  TFX_WriteTokenDictionary() resolves a minimal set of token dictionary
 *  symbols by remapping similar symbols into a fairly distinct set, and
 *  stores the resulting token dictionary using the Tundra format, as
 *  proposed for JBIG-2.
 *  <BR>
 *  The token dictionary will consist of all repeated symbols found during
 *  the calls to TFX_TokenizeImage() for all images passed in either since
 *  the last call to TFX_WriteTokenImage(), TFX_WriteTokenDictionary() or
 *  TFX_WriteTokenImages().
 *  
 * Input:::
 *  tfxDocHandle::		Document handle
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_WriteTokenDictionary(
	IO_DOCHANDLE tfxDocHandle
)
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);

    if ((tfxDoc->seqStatus & (TFX_DOC_SEQ_TOKENSTART)) == 0) {
	NORMAL_RETURN;	/* No tokenization ongoing */
    }
    if (tfxDoc->seqStatus & (TFX_DOC_SEQ_TOKENDICTSAVED)) {
	NORMAL_RETURN;	/* Already saved */
    }

    /* If the document has an open token dictionary, create a consolidation
     * mapping table */
#ifdef NODGP
    XERR_PASS_CLEANUP(wp_TfxConsolidateTokens(tfxDoc));
#endif

    /* Write the consolidated dictionary */
    XERR_PASS_CLEANUP(wp_TfxWriteTokenDictionary(tfxDoc, NULL));

    /* Mark as stored (you cannot tokenize after this) */
    tfxDoc->seqStatus |= TFX_DOC_SEQ_TOKENDICTSAVED;

    NORMAL_RETURN;

 ERR_CLEANUP:

     ERR_RETURN;
}
#endif /* NOIPCORE && NODGP */


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set (write or rewrite) the authoring block
 *
 * LongDescription:::
 *  TFX_SetAuthor() sets the authoring information block. If the authoring
 *  block already exists in the file, the block is rewritten to the new block.
 *  The data will overwrite the previous block if the old block was as big or
 *  bigger than the new block, otherwise it is appended to the file.
 *
 * Input:::
 *  tfxDocHandle::		Document handle
 *  pAuthorBlock::		Pointer to the authoring block of data
 *  wAuthorBytes::		Number of bytes in authoring block
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetAuthor(
	IO_DOCHANDLE tfxDocHandle,
	Int8 *pAuthorBlock,
	UInt16 wAuthorBytes
)
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);

    XERR_PASS(wp_TfxSetAuthor(tfxDoc, pAuthorBlock, wAuthorBytes));

    NORMAL_RETURN;
} 


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set (write or rewrite) the reference page offset
 *
 * LongDescription:::
 *  TFX_SetReferencePage() sets the reference page offset. If the reference
 *  page is already stored in the file, it is overwritten.
 *
 * Input:::
 *  tfxDocHandle::		Document handle
 *  dwReferencePageOffset::	Reference page offset
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetReferencePage(
	IO_DOCHANDLE tfxDocHandle,
	Int32 dwReferencePageOffset
)
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxDocHandle);

    /* Set the reference page at the document level */
/*    tfxDoc->refPageOffset = dwReferencePageOffset; */

    /* Now call the lower level routine to do the dirty work */
    XERR_PASS_CLEANUP(wp_TfxSetReferencePage(tfxDoc, dwReferencePageOffset));

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
} 


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the number of entries each annotation table will contain.
 *
 * LongDescription:::
 *  TFX_SetAnnotTableEntryCount() sets the number of annotation table entries
 *  to be used, before the annotation table header is actually written. This
 *  routine can be called at any reasonable time, but the count passed in will
 *  only apply to new annotation table headers created, and only when an
 *  annotation is actually saved.
 *
 * Input:::
 *  tfxDoc::			Data structure of open (just created) document
 *  wEntriesPerTable::		Number of entries per table to use
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetAnnotTableEntryCount(
	IO_DOCHANDLE tfxDocHandle,
	UInt16 wEntriesPerTable
)
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);
    
    /* If version 2, there is no annotation table */
    if (tfxDoc->docInfo.version < 3) {
	NORMAL_RETURN;
    }

    /* For a minimum value */
    if (wEntriesPerTable < 3) {
	wEntriesPerTable = 3;
    }      

    /* Will be used in any new annotation tables written later */
    tfxDoc->annotsPerTable = wEntriesPerTable;

    NORMAL_RETURN;
}

#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the number of entries each page table will contain.
 *
 * LongDescription:::
 *  TFX_SetPageTableEntryCount() sets the number of page table entries
 *  to be used, before the page table header is actually written. It is
 *  best to call this routine just after the call to TFX_CreateDocument(),
 *  although you should be able to call it any time before the first page
 *  is actually started.
 *
 * Input:::
 *  tfxDoc::			Data structure of open (just created) document
 *  wEntriesPerTable::		Number of entries per table to use
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetPageTableEntryCount(
	IO_DOCHANDLE tfxDocHandle,
	UInt16 wEntriesPerTable
)
{
#if 1
    return IO_NOSUPPORT;
#else
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);
    
    /* For a minimum value */
    if (wEntriesPerTable < 10) {
	wEntriesPerTable = 10;
    }      

    /* Check for proper calling sequence */
    XERR_GEN_CLEANUP(tfxDoc->seqStatus & TFX_DOC_SEQ_W_PAGETABLE, IO_SEQUENCE);

    /* If the file exists, we cannot go back and change the value...
     * unless it was never set in the first place */
    XERR_GEN_CLEANUP((tfxDoc->readWrite & READ_ALLOWED) &&
		     tfxDoc->pageTableHeader.entriesPerTable != 0, IO_NOSUPPORT);

    tfxDoc->pageTableHeader.entriesPerTable = wEntriesPerTable;

    NORMAL_RETURN;
    
 ERR_CLEANUP:
    
    ERR_RETURN;
#endif
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the number of entries each shared resources table will contain.
 *
 * LongDescription:::
 *  TFX_SetResourceTableEntryCount() sets the number of shared resources table
 *  entries to be used, before the symbol library table header is actually
 *  written.
 *  It is best to call this routine just after the call to
 *  TFX_CreateDocument(), although you should be able to call it any time
 *  before the first token dictionary is actually written (up to the call
 *  to TFX_WriteTokenImage(), TFX_WriteTokenDictionary(),
 *  TFX_WriteTokenImages(), or TFX_CloseDocument()).
 *
 * Input:::
 *  tfxDoc::			Data structure of open (just created) document
 *  wEntriesPerTable::		Number of entries per table to use
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetResourceTableEntryCount(
	IO_DOCHANDLE tfxDocHandle,
	UInt16 wEntriesPerTable
)
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    PROLOGUE;

    TFX_CheckHandle(tfxDocHandle);

    /* For a minimum value */
    if (wEntriesPerTable < 1) {
	wEntriesPerTable = 1;
    }      

    /* Note: we used to keep track of when the resource table header was
     * written, but now we don't have to, since each table has it's own
     * unique count */
    /* Check for proper calling sequence */
/*    XERR_GEN_CLEANUP(tfxDoc->seqStatus & TFX_DOC_SEQ_W_RESOURCETABLE, IO_SEQUENCE); */

    /* If the file exists, we cannot go back and change the value...
     * unless it was never set in the first place */
/*    XERR_GEN_CLEANUP((tfxDoc->readWrite & READ_ALLOWED) && */
/*		     tfxDoc->resourceTableHeader.entriesPerTable != 0, IO_NOSUPPORT); */

    tfxDoc->resourcesPerTable = wEntriesPerTable;

    NORMAL_RETURN;
    
/* ERR_CLEANUP: */
    
/*    ERR_RETURN; */
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Add an annotation to a page
 *
 * LongDescription:::
 *  TFX_AddAnnotation() adds an annotation to a page
 *
 * Input:::
 *  tfxPageHandle::		Active page handle
 *  wAnnotType::		Type of annotation
 *  dwXOffset::			Offset in X direction in page coordinates
 *  dwYOffset::			Offset in Y direction in page coordinates
 *  dwWidth::			Width in page coordinates
 *  dwHeight::			Height in page coordinates
 *  pBuf::			Annotation data
 *  dwAnnotBytes::		Length of annotation data, in bytes
 *
 * Output:::
 *  pdwAnnotID::		Annotation ID assigned to this annotation
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_AddAnnotation(
	IO_PAGEHANDLE tfxPageHandle,
	UInt16 wAnnotType,
	UInt32 dwXOffset,
	UInt32 dwYOffset,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt32 dwFlags,
	UInt8 *pBuf,
	UInt32 dwAnnotBytes,
	UInt16 *pdwAnnotID
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_ANNOTTABLE *annotTable;
    PROLOGUE;

    TFX_CheckHandle(tfxPageHandle);

    /* Make sure that we've read the appropriate table that the annotation
     * will be attached to */
    if ((tfxPage->tfxDoc->readWrite & READ_ALLOWED) &&
	tfxPage->annotTableHeader.annotCount > 0) {
	XERR_PASS(wp_TfxGetAnnotTable(tfxPage,
	    (UInt16)(tfxPage->annotTableHeader.annotCount - 1), &annotTable));
    }

    /* Now add the annotation */
    XERR_PASS(wp_TfxAddAnnotation(tfxPage, wAnnotType, dwXOffset, dwYOffset,
	dwWidth, dwHeight, dwFlags, pBuf, dwAnnotBytes, pdwAnnotID));

    NORMAL_RETURN;
} 


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Deletes an annotation from a page
 *
 * LongDescription:::
 *  TFX_DeleteAnnotation() deletes an annotation from a page. Note that the
 *  remaining subsequent annotations will automatically shift down one
 *  position.
 *
 * Input:::
 *  tfxPageHandle::		Page handle, returned from TFX_GetPageInfo()
 *  dwAnnotOffset::		Offset (or ID) of annotation to delete (0, 1, ...)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_DeleteAnnotation(
	IO_PAGEHANDLE tfxPageHandle,
	UInt16 dwAnnotOffset
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    UInt16 oldCount = 0, newCount = 0;
    PROLOGUE;

    /* Check page handle */
    TFX_CheckHandle(tfxPageHandle);

    /* Get the old annotation count */
    oldCount = tfxPage->annotTableHeader.annotCount;

    /* Make sure there's one to delete, and that the requested one is there */
    XERR_GEN_CLEANUP(dwAnnotOffset >= oldCount, IO_BADPARAMETER);

    /* Delete the annotation (just remove the reference, data is left in
     * the file) */
    XERR_PASS_CLEANUP(wp_TfxDeleteAnnotation(tfxPage, dwAnnotOffset, &newCount));

    XERR_GEN_CLEANUP(newCount != oldCount - 1, IO_INTERNAL_ERROR);
    XERR_GEN_CLEANUP(newCount != tfxPage->annotTableHeader.annotCount,
	IO_INTERNAL_ERROR);
    XERR_GEN_CLEANUP(newCount != tfxPage->pageInfo.wAnnotationCount,
	IO_INTERNAL_ERROR);

    NORMAL_RETURN;

ERR_CLEANUP:
    ERR_RETURN;
}


#ifndef NOIPCORE
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Compress a strip of image data
 *
 * LongDescription:::
 *  TFX_Compress() compresses a single image strip. Used as a standalone
 *  compressor, allows an application to retrieve the image data from file
 *  in raw compressed form, and decode individual strips on the fly as needed.
 *  <BR>
 *  Source image is compressed into the destination buffer. If the destination
 *  buffer is filled and would overflow, IO_BUFFEREXCEEDED is returned.
 *
 * Input:::
 *  tfxImgHandle::		Valid image handle, as returned from
 *				TFX_AddImage()
 *  xCompression::		See IO_COMPRESSION types
 *  pdwQuality::		Pointer to desired quality (actual quality
 *				returned), see TFX_ImageWriteStart() for more
 *  xColorType::		Type of input image color
 *  xColorSpace::		Color space to convert to during compression
 *				IO_COLORSPACE_<c> where <c> is one of:
 *				    CMYK, YCBCR, CIELAB, ICCLAB, or ITULAB
 *  dwFlags::			IO_IMGFLAGS_... combo
 *  pSrc[3]::			Source buffer channels
 *  dwBytesPerRow::		Bytes per row in image buffer
 *  dwRowOffset::		Row offset in image of data to compress
 *  dwWidth::			Columns of COLORTYPE in dst image
 *  dwHeight::			Rows in dst image
 *
 * Output:::
 *  ppDst::			Pointer to the destination compressed data buffer
 *  pdwBytesInDst::		Pointer to the byte count of destination buffer
 *  pdwBytesFilled::		Actual bytes in destination filled
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_Compress(
	IO_IMGHANDLE tfxImgHandle,
	IO_COMPRESSION xCompression,
	UInt32 *pdwQuality,
	IO_COLORTYPE xColorType,
	IO_COLORSPACE xColorSpace,
	UInt32 dwFlags,
	UInt8 *pSrc[3],
	UInt32 dwBytesPerRow,
	UInt32 dwRowOffset,
	UInt32 dwWidth,
	UInt32 dwHeight,
	UInt8 **ppDst,
	UInt32 *pdwBytesInDst,
	UInt32 *pdwBytesFilled
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    ScaledInt1616 xres, yres;
    Int32 rowsLeft;
    Int32 pixrBpl;
    UInt32 bytesPerRow;
    IPWState *pIPWState;
    PIXR *pPixr = NULL, *pSubPixr;
    ColorType ctype = cColorNone;
    UInt8 *pUserImage, *pPixrImage;
    Int32 channel, channels;
    PROLOGUE;

    *pdwBytesFilled = 0UL;
    TFX_CheckHandle(tfxImgHandle);

    /* Borrow the quality passed in */
    tfxImg->JPEGquality = *pdwQuality;

    switch ((int)xCompression) {
    case IO_COMP_JPEG:
	if (tfxImg->JPEGquality == 0) {
	    tfxImg->JPEGquality = 75;	/* The default */
	} else if (tfxImg->JPEGquality > 100) {
	    tfxImg->JPEGquality = 100;	/* The max */
	} else if (tfxImg->JPEGquality < 1) {
	    tfxImg->JPEGquality = 1;	/* My hard-coded min */
	}
	break;
    default:
	tfxImg->JPEGquality = 0;
    }

    /* Get an pointer to the IPState, it's needed below */
    XERR_PASS (wp_TfxIPWState (tfxImg->tfxPage->tfxDoc, &pIPWState));

    /* Access the common IPWState */
    XERR_PASS_CLEANUP(wp_TfxIPWState(tfxImg->tfxPage->tfxDoc, &pIPWState));

    /* Check buffers and sizes */
    XERR_GEN_CLEANUP(pSrc == NULL || pSrc[0] == NULL || ppDst == NULL ||
	dwBytesPerRow == 0, IO_BADPARAMETER);
    XERR_GEN_CLEANUP(dwRowOffset >= tfxImg->imgInfo.dwHeight, IO_BADPARAMETER);
    
    /* Create a PIXR image to hold the image */
    xres = Int32_TO_ScaledInt1616(tfxImg->imgInfo.dwXResolution);  /* Any resolution is fine, I guess */
    yres = Int32_TO_ScaledInt1616(tfxImg->imgInfo.dwYResolution);  /* Any resolution is fine, I guess */
    switch (xColorType) {
    case IO_COLORTYPE_COLOR24:
	ctype = cColorRGB;
	break;
    case IO_COLORTYPE_BINARY:
    case IO_COLORTYPE_GRAY4:
    case IO_COLORTYPE_COLOR4:
    case IO_COLORTYPE_GRAY8:
    case IO_COLORTYPE_COLOR8:
	ctype = cColorGray;
	break;
    default:
	break;
    }
    channels = IO_CHANNELS(xColorType);
    XERR_PASS_CLEANUP(w_createPixrExtended(pIPWState, dwWidth, dwHeight,
	IO_BPPE(xColorType), xres, yres, 32, channels, ctype, NULL, &pPixr));

    /* Copy the data to the user's buffer */
    XERR_PASS_CLEANUP(w_pixrGetBpl(pIPWState, pPixr, &pixrBpl));
    bytesPerRow = dwBytesPerRow;
    if ((Int32)bytesPerRow > pixrBpl) {
	bytesPerRow = (UInt32)pixrBpl;
    }
    for (channel = 0; channel < channels; channel++) {
	if (channels > 1) {
	    pSubPixr = NULL;
	    XERR_PASS_CLEANUP(w_pixrGetChannel(pIPWState, pPixr, channel,
					  &pSubPixr));
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pSubPixr, &pPixrImage));
	} else {
	    XERR_PASS_CLEANUP(w_pixrGetImage(pIPWState, pPixr, &pPixrImage));
	}

	rowsLeft = dwHeight;
	pUserImage = pSrc[channel] + dwRowOffset * dwBytesPerRow;
	while (rowsLeft > 0) {
	    memcpy(pPixrImage, pUserImage, bytesPerRow);
	    pUserImage += dwBytesPerRow;
	    pPixrImage += pixrBpl;
	    rowsLeft--;
	}
    }

    /* Call the underlying compression routine */
    XERR_PASS_CLEANUP(w_TfxCompress(pIPWState, tfxImgHandle, xCompression,
	xColorType, xColorSpace, dwFlags, pPixr, 0, dwWidth, dwHeight, ppDst,
	pdwBytesInDst, pdwBytesFilled));
    
    w_destroyPixr(pIPWState, &pPixr);

    NORMAL_RETURN;
    
 ERR_CLEANUP:

    w_destroyPixr(pIPWState, &pPixr);
    
    ERR_RETURN;
}
#endif /* NOIPCORE */


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the AnnotationOffsets for the current page
 *
 * LongDescription:::
 *  TFX_SetAnnotationOffsets() sets the AnnotationOffsets field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  dwAnnotationOffsets::	AnnotationOffsets value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetAnnotationOffsets(
    IO_PAGEHANDLE tfxPageHandle,
	UInt32 dwAnnotationOffsets
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_IMG *tfxImg;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);
    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    XERR_GEN_CLEANUP(tfxImg == NULL, IO_SEQUENCE);

    /* Make sure it's not already been written to the file */
	/* This is incomplete: if setting, and it's already been written,
	 * then I could go back and update it.
	 * But for this particular field, this is just a place-holder
	 * to indicate that the field WILL be written, before the IFD is
	 * actually occupied. */
/*    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE); */

    /* Copy the data */
/*    tfxImg->fields.AnnotationOffsets = dwAnnotationOffsets; */
    dwAnnotationOffsets = 0;	/* Avoid compiler warning, use the variable */

    /* Mark it as available for write */
    tfxImg->fields.found.AnnotationOffsets = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}



#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set (write or rewrite) the authoring block
 *
 * LongDescription:::
 *  TFX_SetAuthor() sets the authoring information block. If the authoring block
 *  already exists in the file, the block is rewritten to the new block. The
 *  data will overwrite the previous block if the old block was as big or
 *  bigger than the new block, otherwise it is appended to the file.
 *
 * Input:::
 *  tfxDocHandle::		Document handle
 *  pAuthorBlock::		Pointer to the authoring block of data
 *  wAuthorBytes::		Number of bytes in authoring block
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetAuthor(
    IO_DOCHANDLE tfxDocHandle,
	Int8 *pAuthorBlock,
	UInt16 wAuthorBytes
    )
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
    TFX_PAGE *tfxPage;
    TFX_IMG *tfxImg;
    TFX_INSTANCE *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxDocHandle);
    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    XERR_GEN_CLEANUP(tfxImg == NULL, IO_SEQUENCE);


    instData = TFX_DOCINST(tfxDoc);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Allocate storage for the string */
    if (tfxImg->fields.Author != NULL) {
	TFX_FREE(tfxImg->fields.Author);
    }
    IO_CHECKMALLOC(tfxImg->fields.Author = TFX_MALLOC(wAuthorBytes + 1));
    tfxImg->fields.Author[dwAuthorBytes] = '\0';

    /* Copy the data */
    memcpy(tfxImg->fields.Author, pAuthorBlock, wAuthorBytes);

    /* Mark it as available for write */
    tfxImg->fields.found.Author = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}
#endif


#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the BitsPerSample for the current image
 *
 * LongDescription:::
 *  TFX_SetBitsPerSample() sets the BitsPerSample field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  dwBitsPerSample::		BitsPerSample value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetBitsPerSample(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 dwBitsPerSample,
	UInt32 dwBitsPerSampleValues
)
{
    *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_INSTANCE *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    instData = TFX_IMGINST(tfxImg);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Allocate storage for the string */
    if (tfxImg->fields.BitsPerSample != NULL) {
	TFX_FREE(tfxImg->fields.BitsPerSample);
    }
    IO_CHECKMALLOC(tfxImg->fields.BitsPerSample = TFX_MALLOC(IMG));

    /* Copy the data */
    memcpy(tfxImg->fields.BitsPerSample, dwBitsPerSample,
	dwBitsPerSampleValues);

    /* Mark it as available for write */
    tfxImg->fields.found.BitsPerSample = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ChromaPositioning for the current image
 *
 * LongDescription:::
 *  TFX_SetChromaPositioning() sets the ChromaPositioning field for the
 *  current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  chromaPositioning::		ChromaPositioning value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetChromaPositioning(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 dwPositioning
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.ChromaPositioning = dwPositioning;

    /* The value should really be 1, the only supported value.  I guess
     * if we're here, there must be a good reason for setting
     * it to something different. */

    /* Mark it as available for write */
    tfxImg->fields.found.ChromaPositioning = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ChromaSubSampling for the current image
 *
 * LongDescription:::
 *  TFX_SetChromaSubSampling() sets the ChromaSubSampling field for the
 *  current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  dwChromaSubSamplingHor::	Horizontal ChromaSubSampling value to set
 *  dwChromaSubSamplingVer::	Vertical ChromaSubSampling value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetChromaSubSampling(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 dwChromaSubSamplingHor,
	UInt16 dwChromaSubSamplingVer
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.ChromaSubSamplingHor = dwChromaSubSamplingHor;
    tfxImg->fields.ChromaSubSamplingVer = dwChromaSubSamplingVer;

    /* Mark it as available for write */
    tfxImg->fields.found.ChromaSubSampling = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


#if 0
/* This is automatically generated.  It shouldn't be specifically set.
 * But if it were, this is how it would be done.
 */
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the CodingMethods for the current page
 *
 * LongDescription:::
 *  TFX_SetCodingMethods() sets the CodingMethods field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  dwCodingMethods::		CodingMethods value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetCodingMethods(
    IO_PAGEHANDLE tfxPageHandle,
	UInt32 dwCodingMethods
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_IMG *tfxImg;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);
    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    XERR_GEN_CLEANUP(tfxImg == NULL, IO_SEQUENCE);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.gpIFD.CodingMethods = dwCodingMethods;

    /* Mark it as available for write */
    tfxImg->fields.found.CodingMethods = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the Compression for the current image
 *
 * LongDescription:::
 *  TFX_SetCompression() sets the Compression
 *  field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  wCompression::		Compression value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetCompression(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wCompression
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.Compression = wCompression;

    /* Mark it as available for write */
    tfxImg->fields.found.Compression = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the DateTime string for the current page
 *
 * LongDescription:::
 *  TFX_SetDateTime() sets the DateTime field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *  <BR>
 *  Note, the format for the DateTime field is fixed, and should contain
 *  no more bytes than is expected for the string.  Format is:
 *  "YYYY:MM:DD HH:MM:SS".
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  pDateTime::			DateTime string to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetDateTime(
    IO_PAGEHANDLE tfxPageHandle,
	Int8 *pDateTime
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_IMG *tfxImg;
    TFX_INSTANCE *instData;
    UInt32 len;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);
    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    XERR_GEN_CLEANUP(tfxImg == NULL, IO_SEQUENCE);

    instData = TFX_PAGEINST(tfxPage);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Allocate storage for the string */
    len = strlen((char *)pDateTime) + 1;	/* Include the terminator */
    if (len > 20) len = 20;
    if (tfxImg->fields.DateTime != NULL) {
	TFX_FREE(tfxImg->fields.DateTime);
    }
    IO_CHECKMALLOC(tfxImg->fields.DateTime = TFX_MALLOC(len + 1));
    tfxImg->fields.DateTime[len] = '\0';

    /* Keep the length, even though it's constant */
    tfxImg->fields.DateTimeLength = len;

    /* Copy the data */
    memcpy(tfxImg->fields.DateTime, pDateTime, len);

    /* Mark it as available for write */
    tfxImg->fields.found.DateTime = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the Decode values for the current image
 *
 * LongDescription:::
 *  TFX_SetDecode() sets the Decode field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  pdwDecode::			Decode array to set: signed rationals
 *				(signed longs) for Lmin, Lmax, amin, amax,
 *				bmin, and bmax,
 *  dwDecodeValues::		Number of Decode values (2 for L*,
 *				6 for L*a*b*)
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetDecode(
    IO_IMGHANDLE tfxImgHandle,
	Int32 *pdwDecode,
	UInt32 dwDecodeValues
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure there aren't too many values */
    XERR_GEN_CLEANUP(dwDecodeValues >
	(sizeof(tfxImg->fields.Decode) / sizeof(tfxImg->fields.Decode[0])),
	IO_BADPARAMETER);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    memcpy(tfxImg->fields.Decode, pdwDecode, dwDecodeValues * 2 *
	sizeof(*pdwDecode));

    /* Mark it as available for write */
    tfxImg->fields.found.Decode = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the DocumentName string for the current doc
 *
 * LongDescription:::
 *  TFX_SetDocumentName() sets the DocumentName field for the current DOC.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxDocHandle::		Doc handle
 *  pDocumentName::		DocumentName string to set
 *  dwDocumentNameBytes::	Bytes in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetDocumentName(
    IO_DOCHANDLE tfxDocHandle,
	Int8 *pDocumentName,
	UInt32 dwDocumentNameBytes
    )
{
    TFX_DOC *tfxDoc = (TFX_DOC *)tfxDocHandle;
/*    TFX_INSTANCE *instData; */
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxDocHandle);

#if 1
    XERR_GEN_CLEANUP(1, IO_NOSUPPORT);
    /* Avoid compiler warnings */
    tfxDoc = tfxDoc;
    pDocumentName = pDocumentName;
    dwDocumentNameBytes = dwDocumentNameBytes;
#else
    instData = TFX_DOCINST(tfxDoc);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Allocate storage for the string */
    if (tfxImg->fields.DocumentName != NULL) {
	TFX_FREE(tfxImg->fields.DocumentName);
    }
    IO_CHECKMALLOC(tfxImg->fields.DocumentName =
	TFX_MALLOC(dwDocumentNameBytes));
    tfxImg->fields.DocumentName[dwDocumentNameBytes] = '\0';

    /* Copy the data */
    memcpy(tfxImg->fields.DocumentName, pDocumentName, dwDocumentNameBytes);

    /* Mark it as available for write */
    tfxImg->fields.found.DocumentName = 1;
#endif
    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the FaxProfile for the current page
 *
 * LongDescription:::
 *  TFX_SetFaxProfile() sets the FaxProfile field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  faxProfile::		FaxProfile value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetFaxProfile(
    IO_PAGEHANDLE tfxPageHandle,
	UInt8 faxProfile
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_IMG *tfxImg;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);
    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    XERR_GEN_CLEANUP(tfxImg == NULL, IO_SEQUENCE);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.gpIFD.FaxProfile = faxProfile;

    /* Mark it as available for write */
    tfxImg->fields.found.FaxProfile = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the FillOrder for the current image
 *
 * LongDescription:::
 *  TFX_SetFillOrder() sets the FillOrder field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.<BR>
 *  Note that the FillOrder should only be set when passing compressed
 *  data back and forth, and the bit order of the data is not MSB-to-LSB.
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  wFillOrder::		FillOrder value to set:
 *				  1 = MSB to LSB (default)
 *				  2 = LSB to MSB
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetFillOrder(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wFillOrder
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.FillOrder = wFillOrder;

    /* Mark it as available for write */
    tfxImg->fields.found.FillOrder = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the HostComputer for the current image
 *
 * LongDescription:::
 *  TFX_SetHostComputer() sets the HostComputer field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::			Image handle
 *  pHostComputer::			HostComputer string to set
 *  dwHostComputerBytes::		Bytes in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  Elizabeth Espinosa-Pardo, based on TFX_SetSoftware
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetHostComputer(
    IO_IMGHANDLE tfxImgHandle,
	Int8 *pHostComputer,
	UInt32 dwHostComputerBytes
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_INSTANCE *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    instData = TFX_IMGINST(tfxImg);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Allocate storage for the string */
    if (tfxImg->fields.HostComputer != NULL) {
	TFX_FREE(tfxImg->fields.HostComputer);
    }
    IO_CHECKMALLOC(tfxImg->fields.HostComputer = TFX_MALLOC(dwHostComputerBytes + 1));
	tfxImg->fields.HostComputer[dwHostComputerBytes] = '\0';

    /* Copy the data */
    memcpy(tfxImg->fields.HostComputer, pHostComputer, dwHostComputerBytes);

    /* Mark it as available for write */
    tfxImg->fields.found.HostComputer = 1;
	tfxImg->fields.HostComputerLength = dwHostComputerBytes; 

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the IFDBaseColor for the current image
 *
 * LongDescription:::
 *  IO_SetIFDBaseColor() sets the IFDBaseColor field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *  <BR>
 *  Note, IFDBaseColor is only used when the first extension set is being used.
 *  In this case, both the IFDBaseColor and ImageBaseColor may be used, but the
 *  ImageBaseColor field may be omitted if the color image is not interpretable
 *  by a base TIFF-FX reader (for example it has JPEGTables).
 *
 * Input:::
 *  ioImgHandle::		Image handle
 *  *pIFDBaseColor::		IFDBaseColor array to set.  This is a color
 *				in the color characteristics described by the
 *				IFD: PhotometricInterpretation, SamplesPerPixel,
 *				BitsPerSample, and if appropriate: Decode,
 *				Indexed and/or ICCProfiles.
 *  dwIFDBaseColorValues::	Number of color values: should be
 *				SamplesPerPixel
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetIFDBaseColor(
    IO_IMGHANDLE tfxImgHandle,
	Int32 *pdwIFDBaseColor,
	UInt32 dwIFDBaseColorValues
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_FILE *tfxFile;
    UInt32 colorIndex;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);
    tfxFile = TFX_IMGFILE(tfxImg);

    /* Shouldn't be more than 4 values */
    XERR_GEN_CLEANUP(dwIFDBaseColorValues >
	(sizeof(tfxImg->fields.IFDBaseColor) /
	 sizeof(tfxImg->fields.IFDBaseColor[0])), 
	IO_BADPARAMETER);

    /* If the IFD has already been written... */
    if (tfxImg->fields.dataTransferred) {

	/* ... and there was an IFDBaseColor field stored, then store this
	 * over the one there. */
	if (tfxImg->fieldDataLocation.IFDBaseColor) {

	    /* Must have the same number of components */
	    XERR_GEN_CLEANUP(dwIFDBaseColorValues != tfxImg->fields.IFDBaseColorValues,
		IO_BADPARAMETER);

	    /* Store the new values over the old ones */
	    TFX_SEEK(tfxFile, tfxImg->fieldDataLocation.IFDBaseColor);
	    switch (tfxImg->fields.IFDBaseColorType) {
	    case IFDBYTE:
	    {
		unsigned char cval;
		
		for (colorIndex = 0; colorIndex < dwIFDBaseColorValues; colorIndex++) {
		    cval = (unsigned char)pdwIFDBaseColor[colorIndex];
		    TFX_WRITE(tfxFile, &cval, sizeof(cval));
		}
	    }
	    break;
	    case IFDSHORT:
	    {
		unsigned short sval;

		for (colorIndex = 0; colorIndex < dwIFDBaseColorValues; colorIndex++) {
		    sval = (unsigned short)pdwIFDBaseColor[colorIndex];
		    TFX_SWAPSHORT(tfxFile, &sval);
		    TFX_WRITE(tfxFile, &sval, sizeof(sval));
		}
	    }
	    break;
	    case IFDSSHORT:
	    {
		short sval;

		for (colorIndex = 0; colorIndex < dwIFDBaseColorValues; colorIndex++) {
		    sval = (short)pdwIFDBaseColor[colorIndex];
		    TFX_SWAPSHORT(tfxFile, &sval);
		    TFX_WRITE(tfxFile, &sval, sizeof(sval));
		}
	    }
	    break;
	    }
	} else {
	    /* ... and there was NO IFDBaseColor field stored, then we
	     * need to rewrite the field list with the field added */
	    tfxImg->fields.IFDrewriteNeeded = 1;
	}
    }

    /* Copy the data */
    for (colorIndex = 0; colorIndex < dwIFDBaseColorValues; colorIndex++) {
	tfxImg->fields.IFDBaseColor[colorIndex] = pdwIFDBaseColor[colorIndex];
    }

    /* Mark it as available for write */
    tfxImg->fields.found.IFDBaseColor = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}




/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ImageBaseColor for the current image
 *
 * LongDescription:::
 *  TFX_SetImageBaseColor() sets the ImageBaseColor field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  *pImageBaseColor::		ImageBaseColor array to set.  This is an
 *				ITULAB color, 3 items, for L*, a*, and b*.
 *				If the TIFF field Decode exists, then it
 *				specifies the gammt used for this ITULAB
 *				color also; otherwise, the default gamut
 *				applies: L* [0,100], b* [-85,85],
 *				and a* [-75,125].  Note, the Decode gamut
 *				applies even if the PhotometricInterpretation
 *				is NOT ITULAB.
 *				To convert between 8-bit RGB and 8-bit ITULAB
 *				values, use IO_sRGB2ITULAB() and
 *				IO_ITULAB2sRGB().
 *  dwImageBaseColorValues::	Number of color values: 1 if L*, 3 if color.
 *				Note, this should be the same as
 *				SamplesPerPixel.
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetImageBaseColor(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 *pImageBaseColor,
	UInt32 dwImageBaseColorValues
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_FILE *tfxFile;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);
    tfxFile = TFX_IMGFILE(tfxImg);

    /* Shouldn't be more than 3 values */
    XERR_GEN_CLEANUP(dwImageBaseColorValues >
	(sizeof(tfxImg->fields.ImageBaseColor) /
	 sizeof(tfxImg->fields.ImageBaseColor[0])), 
	IO_BADPARAMETER);

    /* If the IFD has already been written... */
    if (tfxImg->fields.dataTransferred) {

	/* ... and there was an ImageBaseColor field stored, then store this
	 * over the one there. */
	if (tfxImg->fieldDataLocation.ImageBaseColor) {
	    UInt32 colorIndex;
	    UInt16 sval;

	    /* Must have the same number of components */
	    XERR_GEN_CLEANUP(dwImageBaseColorValues != tfxImg->fields.ImageBaseColorValues,
		IO_BADPARAMETER);

	    /* Store the new values over the old ones */
	    TFX_SEEK(tfxFile, tfxImg->fieldDataLocation.ImageBaseColor);
	    for (colorIndex = 0; colorIndex < dwImageBaseColorValues; colorIndex++) {
		sval = pImageBaseColor[colorIndex];
		TFX_SWAPSHORT(tfxFile, &sval);
		TFX_WRITE(tfxFile, &sval, sizeof(sval));
	    }
	} else {
	    /* ... and there was NO ImageBaseColor field stored, then we
	     * need to rewrite the field list with the field added */
	    tfxImg->fields.IFDrewriteNeeded = 1;
	}
    }

    /* Copy the data */
    memcpy(tfxImg->fields.ImageBaseColor, pImageBaseColor,
	dwImageBaseColorValues * sizeof(tfxImg->fields.ImageBaseColor[0]));

    /* Mark it as available for write */
    tfxImg->fields.found.ImageBaseColor = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ImageDescription string for the current image
 *
 * LongDescription:::
 *  TFX_SetImageDescription() sets the ImageDescription field for the
 *  current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  pImageDescription::		ImageDescription string to set
 *  dwImageDescriptionBytes::	Bytes in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetImageDescription(
    IO_IMGHANDLE tfxImgHandle,
	Int8 *pImageDescription,
	UInt32 dwImageDescriptionBytes
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_INSTANCE *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    instData = TFX_IMGINST(tfxImg);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Allocate storage for the string */
    if (tfxImg->fields.ImageDescription != NULL) {
	TFX_FREE(tfxImg->fields.ImageDescription);
    }
    IO_CHECKMALLOC(tfxImg->fields.ImageDescription =
	TFX_MALLOC(dwImageDescriptionBytes + 1));
    tfxImg->fields.ImageDescription[dwImageDescriptionBytes] = '\0';

    /* Copy the data */
    memcpy(tfxImg->fields.ImageDescription, pImageDescription,
	dwImageDescriptionBytes);
    tfxImg->fields.ImageDescriptionLength = dwImageDescriptionBytes;

    /* Mark it as available for write */
    tfxImg->fields.found.ImageDescription = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the T88Options for the current image
 *
 * LongDescription:::
 *  TFX_SetT88Options() sets the T88Options field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  dwT88Options::		T88Options value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetT88Options(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 dwT88Options
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.T88Options = dwT88Options;

    /* Mark it as available for write */
    tfxImg->fields.found.T88Options = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ModeNumber for the current page
 *
 * LongDescription:::
 *  TFX_SetModeNumber() sets the ModeNumber field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  modeNumber::		ModeNumber value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetModeNumber(
    IO_PAGEHANDLE tfxPageHandle,
	UInt8 modeNumber
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_IMG *tfxImg;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);
    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    XERR_GEN_CLEANUP(tfxImg == NULL, IO_SEQUENCE);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.gpIFD.ModeNumber = modeNumber;

    /* Mark it as available for write */
    tfxImg->fields.found.ModeNumber = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the Orientation for the current image
 *
 * LongDescription:::
 *  TFX_SetOrientation() sets the Orientation field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  dwOrientation::		Orientation value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetOrientation(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 dwOrientation
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.Orientation = dwOrientation;

    /* Mark it as available for write */
    tfxImg->fields.found.Orientation = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the PageNumber for the current page
 *
 * LongDescription:::
 *  TFX_SetPageNumber() sets the PageNumber field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *  <BR>
 *  Note, the TIFF PageNumber field has two values, 1: the page offset,
 *  and 2: the page count.
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  dwPageOffset::		PageOffset value to set
 *  dwPageCount::		PageCount value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetPageNumber(
    IO_PAGEHANDLE tfxPageHandle,
	UInt16 dwPageOffset,
	UInt16 dwPageCount
    )
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_IMG *tfxImg;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);
    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    XERR_GEN_CLEANUP(tfxImg == NULL, IO_SEQUENCE);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.PageNumber[0] = dwPageOffset;
    tfxImg->fields.PageNumber[1] = dwPageCount;

    /* Mark it as available for write */
    tfxImg->fields.found.PageNumber = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the PhotometricInterpretation for the current image
 *
 * LongDescription:::
 *  TFX_SetPhotometricInterpretation() sets the PhotometricInterpretation
 *  field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  wPhotometricInterpretation::	PhotometricInterpretation value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetPhotometricInterpretation(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wPhotometricInterpretation
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.PhotometricInterpretation = wPhotometricInterpretation;

    /* Mark it as available for write */
    tfxImg->fields.found.PhotometricInterpretation = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the PlanarConfiguration for the current image
 *
 * LongDescription:::
 *  TFX_SetPlanarConfiguration() sets the PlanarConfiguration field for the
 *  current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  wPlanarConfiguration::	PlanarConfiguration value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetPlanarConfiguration(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wPlanarConfiguration
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.PlanarConfiguration = wPlanarConfiguration;

    /* Mark it as available for write */
    tfxImg->fields.found.PlanarConfiguration = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ProfileType for the current page
 *
 * LongDescription:::
 *  TFX_SetProfileType() sets the ProfileType field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  dwProfileType::		ProfileType value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetProfileType(
    IO_PAGEHANDLE tfxPageHandle,
	UInt32 dwProfileType
)
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_IMG *tfxImg;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);
    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    XERR_GEN_CLEANUP(tfxImg == NULL, IO_SEQUENCE);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.gpIFD.ProfileType = dwProfileType;

    /* Mark it as available for write */
    tfxImg->fields.found.ProfileType = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the ResolutionUnit for the current image
 *
 * LongDescription:::
 *  TFX_SetResolutionUnit() sets the ResolutionUnit field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  dwResolutionUnit::		ResolutionUnit value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetResolutionUnit(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 dwResolutionUnit
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.ResolutionUnit = dwResolutionUnit;

    /* Mark it as available for write */
    tfxImg->fields.found.ResolutionUnit = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the RowsPerStrip for the current image
 *
 * LongDescription:::
 *  TFX_SetRowsPerStrip() sets the RowsPerStrip field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  dwRowsPerStrip::		RowsPerStrip value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetRowsPerStrip(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 dwRowsPerStrip
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.RowsPerStrip = dwRowsPerStrip;

    /* Mark it as available for write */
    tfxImg->fields.found.RowsPerStrip = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the SamplesPerPixel for the current image
 *
 * LongDescription:::
 *  TFX_SetSamplesPerPixel() sets the SamplesPerPixel field for the current
 *  image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  samplesPerPixel::		SamplesPerPixel array to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetSamplesPerPixel(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 dwSamplesPerPixel
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.SamplesPerPixel = dwSamplesPerPixel;

    /* Mark it as available for write */
    tfxImg->fields.found.SamplesPerPixel = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the Software for the current image
 *
 * LongDescription:::
 *  TFX_SetSoftware() sets the Software field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  pSoftware::			Software string to set
 *  dwSoftwareBytes::		Bytes in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetSoftware(
    IO_IMGHANDLE tfxImgHandle,
	Int8 *pSoftware,
	UInt32 dwSoftwareBytes
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_INSTANCE *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    instData = TFX_IMGINST(tfxImg);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Allocate storage for the string */
    if (tfxImg->fields.Software != NULL) {
	TFX_FREE(tfxImg->fields.Software);
    }
    IO_CHECKMALLOC(tfxImg->fields.Software = TFX_MALLOC(dwSoftwareBytes + 1));
    tfxImg->fields.Software[dwSoftwareBytes] = '\0';

    /* Copy the data */
    memcpy(tfxImg->fields.Software, pSoftware, dwSoftwareBytes);
    tfxImg->fields.SoftwareLength = dwSoftwareBytes;

    /* Mark it as available for write */
    tfxImg->fields.found.Software = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}

/* Dup */
#if 0
/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the TargetPrinter for the current page
 *
 * LongDescription:::
 *  TFX_SetTargetPrinter() sets the TargetPrinter field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  pTargetPrinter::		TargetPrinter string to set
 *  dwTargetPrinterBytes::	Bytes in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetTargetPrinter(
    IO_PAGEHANDLE tfxPageHandle,
	Int8 *pTargetPrinter,
	UInt32 dwTargetPrinterBytes
    )
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_IMG *tfxImg;
    TFX_INSTANCE *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);
    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    XERR_GEN_CLEANUP(tfxImg == NULL, IO_SEQUENCE);

    instData = TFX_PAGEINST(tfxPage);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Allocate storage for the string */
    if (tfxImg->fields.TargetPrinter != NULL) {
	TFX_FREE(tfxImg->fields.TargetPrinter);
    }
    IO_CHECKMALLOC(tfxImg->fields.TargetPrinter =
	TFX_MALLOC(dwTargetPrinterBytes));
    tfxImg->fields.TargetPrinter[dwTargetPrinterBytes] = '\0';

    /* Copy the data */
    memcpy(tfxImg->fields.TargetPrinter, pTargetPrinter, dwTargetPrinterBytes);

    /* Mark it as available for write */
    tfxImg->fields.found.TargetPrinter = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}
#endif


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the TargetPrinter for the current page
 *
 * LongDescription:::
 *  TFX_SetTargetPrinter() sets the TargetPrinter field for the current PAGE.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxPageHandle::		Page handle
 *  pTargetPrinter::		TargetPrinter string to set
 *  dwTargetPrinterBytes::	Bytes in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetTargetPrinter(
    IO_PAGEHANDLE tfxPageHandle,
	Int8 *pTargetPrinter,
	UInt32 dwTargetPrinterBytes
    )
{
    TFX_PAGE *tfxPage = (TFX_PAGE *)tfxPageHandle;
    TFX_IMG *tfxImg;
    TFX_INSTANCE *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxPageHandle);
    tfxImg = (TFX_IMG *)tfxPage->pageInfo.tfxPrimImgHandle;

    XERR_GEN_CLEANUP(tfxImg == NULL, IO_SEQUENCE);


    instData = TFX_PAGEINST(tfxPage);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Allocate storage for the string */
    if (tfxImg->fields.TargetPrinter != NULL) {
	TFX_FREE(tfxImg->fields.TargetPrinter);
    }
    IO_CHECKMALLOC(tfxImg->fields.TargetPrinter =
	TFX_MALLOC(dwTargetPrinterBytes + 1));
    tfxImg->fields.TargetPrinter[dwTargetPrinterBytes] = '\0';

    /* Copy the data */
    memcpy(tfxImg->fields.TargetPrinter, pTargetPrinter, dwTargetPrinterBytes);

    /* Mark it as available for write */
    tfxImg->fields.found.TargetPrinter = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;

}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the Thresholding string for the current image
 *
 * LongDescription:::
 *  TFX_SetThresholding() sets the Thresholding field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  wThresholding::		Thresholding value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetThresholding(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wThresholding
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_INSTANCE *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    instData = TFX_IMGINST(tfxImg);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.Thresholding = wThresholding;

    /* Mark it as available for write */
    tfxImg->fields.found.Thresholding = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the T4Options for the current image
 *
 * LongDescription:::
 *  TFX_SetT4Options() sets the T4Options field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  dwT4Options::			T4Options value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetT4Options(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 dwT4Options
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.T4Options = dwT4Options;

    /* Mark it as available for write */
    tfxImg->fields.found.T4Options = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the T6Options for the current image
 *
 * LongDescription:::
 *  TFX_SetT6Options() sets the T6Options field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  dwT6Options::		T6Options value to set
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetT6Options(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 dwT6Options
)
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.T6Options = dwT6Options;

    /* Mark it as available for write */
    tfxImg->fields.found.T6Options = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the VersionYear byte-string for the current image
 *
 * LongDescription:::
 *  TFX_SetVersionYear() sets the VersionYear field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  pVersionYear::		VersionYear byte-string to set.  Note, this
 *				is NOT an ASCII string, but an actual
 *				byte-value sequence.
 *  dwVersionYearBytes::	Byte values in string
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetVersionYear(
    IO_IMGHANDLE tfxImgHandle,
	UInt8 *pVersionYear,
	UInt32 dwVersionYearBytes
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_INSTANCE *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    instData = TFX_IMGINST(tfxImg);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Check the parameters */
    XERR_GEN_CLEANUP(pVersionYear == NULL || dwVersionYearBytes > 4,
	IO_BADPARAMETER);

    /* Copy the data */
    memcpy(tfxImg->fields.gpIFD.VersionYear, pVersionYear, dwVersionYearBytes);

    /* Mark it as available for write */
    tfxImg->fields.found.VersionYear = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the YCbCrCoefficients for the current image
 *
 * LongDescription:::
 *  TFX_SetYCbCrCoefficients() sets the YCbCrCoefficients field for the
 *  current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  pdwYCbCrCoefficients::	YCbCrCoefficients value to set
 *  dwYCbCrCoefficientValues::	Number of values
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetYCbCrCoefficients(
    IO_IMGHANDLE tfxImgHandle,
	UInt32 *pdwYCbCrCoefficients,
	UInt32 dwYCbCrCoefficientValues
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_INSTANCE  *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    instData = TFX_IMGINST(tfxImg);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Allocate storage for the string */
    if (tfxImg->fields.YCbCrCoefficients != NULL) {
	TFX_FREE(tfxImg->fields.YCbCrCoefficients);
    }
    IO_CHECKMALLOC(tfxImg->fields.YCbCrCoefficients =
	TFX_MALLOC(dwYCbCrCoefficientValues));

    /* Copy the data */
    memcpy(tfxImg->fields.YCbCrCoefficients, pdwYCbCrCoefficients,
	dwYCbCrCoefficientValues);

    /* Mark it as available for write */
    tfxImg->fields.found.YCbCrCoefficients = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Set the Predictor for the current image
 *
 * LongDescription:::
 *  TFX_SetPredictor() sets the Predictor field for the current image.
 *  If the field has already been written, an error is returned; otherwise
 *  it is queued for write when the IFD is actually written.
 *
 * Input:::
 *  tfxImgHandle::		Image handle
 *  wPredictor::		Predictor value
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  IO
 *
 * Keywords:::
 *  file io, IO
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_SetPredictor(
    IO_IMGHANDLE tfxImgHandle,
	UInt16 wPredictor
    )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    TFX_INSTANCE  *instData;
    PROLOGUE;

    /* Make sure the handle is valid */
    TFX_CheckHandle(tfxImgHandle);

    instData = TFX_IMGINST(tfxImg);

    /* Make sure it's not already been written to the file */
    XERR_GEN_CLEANUP(tfxImg->fields.dataTransferred, IO_SEQUENCE);

    /* Copy the data */
    tfxImg->fields.Predictor = wPredictor;

    /* Mark it as available for write */
    tfxImg->fields.found.Predictor = 1;

    NORMAL_RETURN;

ERR_CLEANUP:

    ERR_RETURN;
}


/******************************************************************************
 * BeginDocumentation:::
 *
 * OneLineDescription:::
 *  Use color strips, instead of compressed strips
 *
 * LongDescription:::
 *  TFX_UseColorStrips() sets the output image into a mode where strips of a
 *  solid color are marked with a single color, instead of being compressed.
 *  In such cases, if ImageBaseColor has been set, then only strips with the
 *  solid color that matches ImageBaseColor will be left uncompressed (0 byte
 *  strips).  Otherwise, the first strip with a solid color will set the
 *  precedence for the color marked for the entire image.
 *  <BR>
 *  Note, if ImageBaseColor hasn't been set before TFX_ImageWriteStart() is
 *  called, the IFD will be rewritten when the document is closed; so that the
 *  new field will be added.
 *
 * Input:::
 *  tfxImg::			Valid image
 *
 * Output:::
 *
 * Returns:::
 *  error status
 *
 * Notes:::
 *
 * FunctionCategory:::
 *  TIFF-FX
 *
 * Keywords:::
 *  file io, MRC, TIFF-FX
 *
 * Level:::
 *  heuristic
 *
 * Maturity:::
 *  released
 *
 * Authors/Modifications:::
 *  aber::
 *
 * Citations:::
 *
 * EndDocumentation:::
 *****************************************************************************/
ErrCode CDECL
TFX_UseColorStrips(
		   IO_IMGHANDLE tfxImgHandle
		   )
{
    TFX_IMG *tfxImg = (TFX_IMG *)tfxImgHandle;
    PROLOGUE;

    TFX_CheckHandle(tfxImg);

    /* Make sure we haven't already written the strips */
    XERR_GEN(tfxImg->seqStatus & TFX_IMG_SEQ_WRITEFINISH, IO_SEQUENCE);

    tfxImg->seqStatus |= TFX_IMG_SEQ_USECOLORSTRIPS;

    NORMAL_RETURN;
}


/*
$Log: tfxwrite.c,v $
Revision 1.3  2003/02/11 01:31:43  mkrocken
fixed problems with comments caused by previous commit

Revision 1.2  2003/02/10 22:53:32  mkrocken
Replaced C++ style comments with C style comments (OSG request)

Revision 1.1  2003/01/03 01:41:14  mkrocken
added Dave A's mrcio to the digipaper code base

Revision 1.46  2001/12/26 21:23:45  dabercro
Mark as released

Revision 1.45  2001/11/07 21:49:35  dabercro
Allow IFDBaseColor to be SSHORT.  Keep track of which type it is.
Refine code carve-out for NODGP.

Revision 1.44  2001/10/30 20:01:53  dabercro
Ignore warnings when page is non-compliant

Revision 1.43  2001/09/18 21:10:11  dabercro
Make work with JBIG2

Revision 1.42  2001/08/22 19:49:34  dabercro
Added support for new field: IFDBaseColor

Revision 1.41  2001/06/27 18:44:11  dabercro
Borrow flag names from IO_ layer

Revision 1.40  2001/05/07 19:23:56  dabercro
Fix "typo" in 8-bit hints image creation.

Revision 1.39  2001/05/03 16:43:10  dabercro
Added a new function UseMultiStripSubIFDs.  Rethink logic about
automatically generating multi-strip SubIFDs.

Revision 1.38  2001/05/01 01:25:35  dabercro
Should not have written all the strips already.

Revision 1.37  2001/05/01 00:58:31  dabercro
Fix location of hints image, if it's not added right after ImageWriteStart.
Make planar configuration a short.

Revision 1.36  2001/04/02 18:25:04  dabercro
Add HostComputer.

Revision 1.35  2001/03/22 20:17:39  dabercro
Add support for new TIFF field: Predictor.

Revision 1.34  2001/03/13 02:59:13  dabercro
Remove lossy-dither and wavelet as valid compression types.

Revision 1.33  2001/02/24 03:27:29  dabercro
Use char* for string to rendering hints.

Revision 1.32  2001/01/19 23:57:11  dabercro
Avoid redifinition of NODGP.

Revision 1.31  2001/01/18 00:21:32  dabercro
Lots of global changes.  Added colorspace parameter to be able
to pick between the color spaces.  Added rendering hint support.
Added lots of stuff to be able to carve away IPCore (no codec).
Started work to make JBIG2 the symbol compressor (is still the
IPCore one). Lots of work on makefiles.  Some function moves
to make read-only libraries have all the right parts. Support for
specifying ITULAB gammut on encode.  More support for
keeping IO_/TFX_/XIF_/MRC_ function handles separate.
Lots of compiler warnings fixed for GNU compiler.  Probably
other things as well, but this is all I can think of right now.

Revision 1.30  2000/12/11 22:23:24  dabercro
Modifications to quick-copy in recommended TIFF-FX file order.

Revision 1.29  2000/11/18 00:53:26  dabercro
Avoid cross polenation of data handles across sublibraries
(XIF, MRC, TFX), especially for CopyImage.  Added a handle
type to structure, now checked also.

Revision 1.28  2000/11/17 23:57:35  dabercro
Added rendering hints read and other miscellaneous changes.

Revision 1.27  2000/11/02 22:21:21  dabercro
Additions for Rendering Hints.

Revision 1.26  2000/10/19 17:59:10  dabercro
Return the FG image handle when adding an image for the
color JBIG2 component.

Revision 1.25  2000/10/16 18:22:54  dabercro
Major adds to support stripes, 0-byte strips, and ImageBaseColor.

Revision 1.24  2000/09/27 21:56:45  dabercro
Variable strips and stripe work.

Revision 1.23  2000/09/22 21:50:21  dabercro
Variable-sized strips now handle multiple layers.

Revision 1.22  2000/09/21 21:19:15  dabercro
Add variable-sized strips: StripRowCounts.

Revision 1.21  2000/09/15 22:12:32  dabercro
Use consistent definitions, from iomrc.h. Other bug fixes.

Revision 1.20  2000/08/30 23:29:07  dabercro
Check for both JBIG2 and Tokenization.

Revision 1.19  2000/08/30 19:39:00  dabercro
It's a little nicer to return a warning, that indicates an image could
be skipped, instead of a serious error... when no support for a
particular profile.

Revision 1.18  2000/08/29 22:36:38  dabercro
Allow other compression types.

Revision 1.17  2000/08/25 23:54:36  dabercro
Annotations must be marked for write before the IFD is written.
Identify all symbol compressed formats when isolating symbol
compression.

Revision 1.16  2000/08/22 01:03:24  dabercro
Changes and adds for UNIX support: Solaris and Linux.

Revision 1.15  2000/08/08 00:27:00  dabercro
Populate the software-length item.

Revision 1.14  2000/08/04 21:56:23  dabercro
Revise logic for page insertions without a page table.

Revision 1.13  2000/08/03 18:25:08  dabercro
Don't return error when it's IO_NOERROR

Revision 1.12  2000/08/03 00:05:57  dabercro
Ignore the warning returned from CheckImageInfo.

Revision 1.11  2000/08/02 22:42:44  dabercro
Return a warning only, if not compliant.  Let the non-compliant
values persist (until political decisions about this are resolved).

Revision 1.10  2000/08/02 00:22:23  dabercro
Comments.  Keep date/time length.

Revision 1.9  2000/08/01 00:11:10  dabercro
Check for resolution and page width compliance for TIFF-FX;
return warning if not compliant, and change to compliant values.
T88Options name change.

Revision 1.8  2000/06/07 00:00:17  dabercro
Continued work, lots of stuff.

Revision 1.7  2000/05/10 22:49:25  dabercro
Make sure compression type matches profile.

Revision 1.6  2000/05/10 20:21:05  dabercro
Maintain strip position when writing compressed position block

Revision 1.5  2000/05/09 01:52:58  dabercro
Added FillOrder field support.
Updated JBIG2 dictionary front-end.
Support for JBIG2 dictionary extensions, updated parameter to
reflect byte count for extensions.
Write fields that are explicitly set, regardless of default values.
Check if document is single-page when closing, reflect in
NewSubfileType.

Revision 1.4  2000/04/12 19:06:09  dabercro
Added many new TIFF fields to support the various profiles.
Added new functions to pass JBIG2 data directly.

Revision 1.3  2000/03/11 02:06:35  dabercro
Milestone checkin: CIELAB/ICCLAB/ITULAB work.  Support for major TIFF fields.  Successful conversion of most of the Genoa documents to XIF.

Revision 1.2  2000/02/08 21:25:34  dabercro
Disable JBIG2 (for now), updates to make use of GlobalParametersIFD for symbol libraries.

Revision 1.1.1.1  2000/02/02 00:40:51  dabercro
Initial checkin of MRC I/O package, include XIFF, TIFF-FX, Coherent TIFF, and T.44 formats.

Revision 1.58  1998/08/04 00:21:37  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Support multiple dictionaries when token compressing, used to retain
single-use tokens in the same dictionary -- to be used when color
tokenizing.

Revision 1.57  1998/07/30 22:39:46  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Prototype updates for signed-ness.

Revision 1.56  1998/05/11 18:25:00  aber
aber on Mon May 11 11:24:09 PDT 1998

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Do not propogate the source layer number when copying images compressed.
Allow the user specified image layer to be assigned.

Revision 1.55  1998/05/05 17:22:40  aber
aber on Tue May 5 10:22:11 PDT 1998

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Added a parameter to TFX_Compress to allow changing the JPEG/Wavelet
quality.

Revision 1.54  1998/05/05 07:15:29  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Additions necessary to fill in the never completed functions
TFX_Compress and TFX_Decompress.

Revision 1.53  1998/04/06 22:43:47  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Bug fix: check resolution difference when checking the image size of
the image during read.
Also, check for unanticipated condition of the page resolution being
SMALLER than than the subimage resolution. Adjust subimage dimensions
accordingly.

Revision 1.52  1998/04/02 00:09:52  aber
aber on Wed Apr 1 16:09:51 PST 1998

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Allow attaching a transparency mask to the primary image, special request.

Revision 1.51  1998/03/25 23:25:17  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Use a unique document ID for each document handle. This allows tagging
each document so that during "quick copy", the token dictionaries are
copy only once, and only once (in the case of two destination documents
somehow ending up with the same memory location for the structure data).

Revision 1.50  1998/02/24 01:41:35  aber
aber on Mon Feb 23 17:40:25 PST 1998

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Fix bounding problem, where the FG/BG image could be larger than the
selector image (accounting for the resolution differences). Adjust
the FG/BG image size, rather than allow the X/Y position to go negative.

Revision 1.49  1998/01/20 03:23:55  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Expand read/write mode to do a create, if the file doesn't already exist.
Fix read/write problems. Proto update for MixedRasterContent renaming.

Revision 1.48  1998/01/16 00:43:12  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Allow page inserts.
To be done later: the page table is regenerated from the linked-list
of pages. This should really be regenerated from the page table itself.

Revision 1.47  1998/01/14 00:16:40  ddavies
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Add IPWState parameter to all IPShared routines and all callbacks.

Revision 1.46  1997/12/18 00:19:24  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Keep track of when a page is completed (pending resolved tokenization),
and is completely written (able to read back).
Add new function TFX_DeletePage() (unwritten) and TFX_DeleteImage().

Revision 1.45  1997/11/27 00:59:56  aber
aber on Wed Nov 26 16:59:39 PST 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Fix PC complaints.

Revision 1.44  1997/11/26 23:26:58  aber
aber on Wed Nov 26 15:26:07 PST 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Remove stipulation that a mask image must mask a non-binary image.
Force tokenization of primary image ONLY.

Revision 1.43  1997/11/25 01:20:37  aber
aber on Mon Nov 24 17:20:08 PST 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Reset the tokenization state after the last tokenized image is
stored; this allows tokenization to start all over again with
a fresh token dictionary.

Revision 1.42  1997/11/24 21:12:05  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Add new function TFX_CheckImageInfo() that is used to localize the
requirements of the supported compression types, and the strip height
that matches the TIFF-FX version and compression type.
This allows the check to be made before TFX_CopyImage() is called,
so that a strip size change will force recompression, instead of
writing a bad strip height out (say when going from a TIFF-FX 3.0 to 2.0,
the strip height of a JPEG image MUST be the image height).

Revision 1.41  1997/11/20 02:54:29  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Resolve (once again) that TIFF-FX 2.0 files cannot have multi-tiled JPEG images.

Revision 1.40  1997/11/19 20:23:53  aber
aber on Wed Nov 19 12:22:54 PST 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Bug fix: when adding an image to a linked list, order by both image
layer, then order within the layer.

Revision 1.39  1997/11/18 23:46:34  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Add new parameter to TFX_CopyPage() to allow (disable) annotation
copy.
Add new function TFX_WriteTokenImages() to loop through the pages that
have been tokenized but not written yet. Automatically store the token
dictionary (if not stored yet).
Add flags parameter to annotation functions.

Revision 1.38  1997/11/17 19:26:02  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Allow negative page offsets to automatically append the page to
the end of the document.

Revision 1.37  1997/11/14 02:09:03  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Use signed value for reference page; clarify that it's a page offset,
not an ordinal. Store as ordinal (was only reading as ordinal).

Revision 1.36  1997/11/11 18:12:43  aber
aber on Tue Nov 11 10:12:05 PST 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

In TFX_CopyImage(), assign transparency mask image offset to be 0, even
though it's not used.

Revision 1.35  1997/11/07 01:59:49  aber
aber on Thu Nov 6 17:58:50 PST 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Keep track of rows buffered, not rows written, when buffering data out.

Revision 1.34  1997/11/07 01:13:09  aber
aber on Thu Nov 6 17:12:46 PST 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Fix logic for new function: TFX_CopyImage().

Revision 1.33  1997/11/06 22:52:17  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

New function added: TFX_CopyImage(). This allows copying an individual
image compressed. Formerly only the entire page could be copied as a whole,
while the individual images were copied compressed. This allows an editor
(like Pagis) to copy selected images compressed, then copy over the remaining
edited images over with edits.
This is the "rough draft" of this function, it may need work (I need a
test environment).

Revision 1.32  1997/11/01 20:59:59  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Remove stipulation that wavelet compression requires the full image.
Add strip buffering for wavelet compression.

Revision 1.31  1997/10/24 21:39:00  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: More exhaustive checking on calling sequence.

Revision 1.30  1997/10/16 19:04:50  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Allow GetImageInfo any time after AddImage.

Revision 1.29  1997/10/15 21:36:52  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Modify parameters to TfxSetPageIFD. Update page count in document during
page add (wasn't being updated at all).

Revision 1.28  1997/10/01 23:41:29  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Recommend strip height a multiple of 8 when 4/8-bit gray/color.
Always check that the strip is not larger than the image.
Don't force JPEG compression when NOC (uncompressed) requested.
Do recommend a full-height strip when using JPEG on a TIFF-FX 2.0 file,
but allow 3.0 files to vary.

Revision 1.27  1997/09/19 00:17:34  aber
aber on Thu Sep 18 17:15:37 PDT 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

When writing, allow TFX_GetPageInfo() call.
Set minimum entries per table for page (10), symlib (2), and annotation
(3).

Revision 1.26  1997/08/29 17:27:33  aber
aber on Fri Aug 29 10:27:01 PDT 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Add 3 new functions to preset the size of the tables: annot table,
page table, and symlib table.
Adjust parameter types to match what's actually stored in the file.

Revision 1.25  1997/08/27 00:17:10  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Update strip sizes to be 64k (instead of whole image, or 8k TIFF
recommendation), and round up to the next multiple of 16. Don't
let it exceed the image height.

Revision 1.24  1997/08/15 00:11:12  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Slight code cleanup.

Revision 1.23  1997/08/13 22:48:20  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Edit fix via Pedro (XResolution should be Yresolution).

Revision 1.22  1997/07/30 01:08:57  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

If strip height not specified by client, use TIFF 8K computation. If JPEG
or Wavelet, force strip to be the entire image (really bad compression if
we use the 8K rule). Mark tokenized images as having image characteristics
set.

Revision 1.21  1997/07/24 00:37:29  aber
aber on Wed Jul 23 17:33:36 PDT 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Add check for existance of a mask when adding one (one shouldn't
already exist).

Revision 1.20  1997/07/22 22:43:04  aber
aber on Tue Jul 22 15:42:10 PDT 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Append the page to the document when adding page zero,
carries over from XFILE.

Revision 1.19  1997/07/15 23:44:13  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Change some parameters to isolate TFX_ImageWriteStrip from w_TfxImageWriteStrip.

Revision 1.18  1997/07/14 23:23:59  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

If no recommended strip size, use the TIFF recommendation of 8K.

Revision 1.17  1997/07/10 22:02:30  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Use 16-bit ID for annotations. Support annotation deletion.

Revision 1.16  1997/07/02 01:06:48  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Add sequence checks for tokenization.

Revision 1.15  1997/06/26 22:32:10  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

If the TIFF-FX 2.0 file is to store a JPEG image, it must be all one tile.
Override the tile height to include the entire image (to be buffered).

Revision 1.14  1997/06/23 22:59:39  aber
aber on Mon Jun 23 15:59:15 PDT 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Update function call to match prototype update for palette support.

Revision 1.13  1997/06/23 22:48:08  aber
aber on Mon Jun 23 15:48:08 PDT 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Sync up image pointers so that lossy dither works right.

Revision 1.12  1997/06/20 17:49:13  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Added comments. Rethink the control parameters to strip buffering.

Revision 1.11  1997/06/17 22:36:28  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Check new flag bit for tokenization on. Define renames.

Revision 1.9  1997/06/10 01:27:00  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Addition of lossy-dither, lossless tokenization, and a
function to "paste" the even numbered layers back together:
TFX_ImageReadLayer().

Revision 1.7  1997/05/30 22:12:44  aber
aber on Fri May 30 15:12:32 PDT 1997

tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

Variable rename to match documentation.

Revision 1.6  1997/05/30 22:11:34  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Update documentation for web browsing. TFX_ functions are left
with no "BeginDocumentation" tag, since w_ and wp_ functions are expected.

Revision 1.5  1997/05/30 00:51:18  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Function name changes to accomodate wp_ private routine convention.
Additional routines to support page copying without decompression
and recompression. Documentation updates. Read-write mode started.
Some bug fixes via Andrew.

Revision 1.4  1997/05/20 22:29:21  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: More conformance with ANSI, used -pedantic -Wstrict-prototypes
-Wuninitialized -Wunused -Wtraditional gcc flags to attempt to catch
anything else the Mac might complain about.

Revision 1.3  1997/05/20 00:54:29  aber
tfxwrite.c -> /product/ipcore/ipwrap/src/RCS/tfxwrite.c,v

CommonCheck: Fixed unused variables, prototype mismatches (CDECL), and variables
potentially used before being set.

Revision 1.1  1997/05/17 00:46:47  aber
Initial revision

*/
